#=======================================================================
#    ____  ____  _____              _    ____ ___   ____
#   |  _ \|  _ \|  ___|  _   _     / \  |  _ \_ _| |___ \
#   | |_) | | | | |_    (_) (_)   / _ \ | |_) | |    __) |
#   |  __/| |_| |  _|    _   _   / ___ \|  __/| |   / __/
#   |_|   |____/|_|     (_) (_) /_/   \_\_|  |___| |_____|
#
#   A Perl Module Chain to faciliate the Creation and Modification
#   of High-Quality "Portable Document Format (PDF)" Files.
#
#   Copyright 1999-2004 Alfred Reibenschuh <areibens@cpan.org>.
#
#=======================================================================
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the
#   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#   Boston, MA 02111-1307, USA.
#
#   $Id: korean.pm,v 1.7 2004/11/22 01:05:08 fredo Exp $
#
#=======================================================================

$cmap->{korean}={
    'ccs' => [ 
        'Adobe',    # registry 
        'Korea1',   # ordering 
        2,          # supplement
    ],
    'cmap' => { # perl unicode maps to adobe cmap (TBD)
        'ident'         =>  [   'ident',        'Identity'      ],
    },
    'g2u' => [
        0x0000, # G+0  U+0000  controlNUL,  <control>
        0x0020, # G+1  U+0020  space,  SPACE
        0x0021, # G+2  U+0021  exclam,  EXCLAMATION MARK
        0x0022, # G+3  U+0022  quotedbl,  QUOTATION MARK
        0x0023, # G+4  U+0023  numbersign,  NUMBER SIGN
        0x0024, # G+5  U+0024  dollar,  DOLLAR SIGN
        0x0025, # G+6  U+0025  percent,  PERCENT SIGN
        0x0026, # G+7  U+0026  ampersand,  AMPERSAND
        0x0027, # G+8  U+0027  quotesingle,  APOSTROPHE
        0x0028, # G+9  U+0028  parenleft,  LEFT PARENTHESIS
        0x0029, # G+10  U+0029  parenright,  RIGHT PARENTHESIS
        0x002A, # G+11  U+002A  asterisk,  ASTERISK
        0x002B, # G+12  U+002B  plus,  PLUS SIGN
        0x002C, # G+13  U+002C  comma,  COMMA
        0x002D, # G+14  U+002D  hyphen,  HYPHEN-MINUS
        0x002E, # G+15  U+002E  period,  FULL STOP
        0x002F, # G+16  U+002F  slash,  SOLIDUS
        0x0030, # G+17  U+0030  zero,  DIGIT ZERO
        0x0031, # G+18  U+0031  one,  DIGIT ONE
        0x0032, # G+19  U+0032  two,  DIGIT TWO
        0x0033, # G+20  U+0033  three,  DIGIT THREE
        0x0034, # G+21  U+0034  four,  DIGIT FOUR
        0x0035, # G+22  U+0035  five,  DIGIT FIVE
        0x0036, # G+23  U+0036  six,  DIGIT SIX
        0x0037, # G+24  U+0037  seven,  DIGIT SEVEN
        0x0038, # G+25  U+0038  eight,  DIGIT EIGHT
        0x0039, # G+26  U+0039  nine,  DIGIT NINE
        0x003A, # G+27  U+003A  colon,  COLON
        0x003B, # G+28  U+003B  semicolon,  SEMICOLON
        0x003C, # G+29  U+003C  less,  LESS-THAN SIGN
        0x003D, # G+30  U+003D  equal,  EQUALS SIGN
        0x003E, # G+31  U+003E  greater,  GREATER-THAN SIGN
        0x003F, # G+32  U+003F  question,  QUESTION MARK
        0x0040, # G+33  U+0040  at,  COMMERCIAL AT
        0x0041, # G+34  U+0041  A,  LATIN CAPITAL LETTER A
        0x0042, # G+35  U+0042  B,  LATIN CAPITAL LETTER B
        0x0043, # G+36  U+0043  C,  LATIN CAPITAL LETTER C
        0x0044, # G+37  U+0044  D,  LATIN CAPITAL LETTER D
        0x0045, # G+38  U+0045  E,  LATIN CAPITAL LETTER E
        0x0046, # G+39  U+0046  F,  LATIN CAPITAL LETTER F
        0x0047, # G+40  U+0047  G,  LATIN CAPITAL LETTER G
        0x0048, # G+41  U+0048  H,  LATIN CAPITAL LETTER H
        0x0049, # G+42  U+0049  I,  LATIN CAPITAL LETTER I
        0x004A, # G+43  U+004A  J,  LATIN CAPITAL LETTER J
        0x004B, # G+44  U+004B  K,  LATIN CAPITAL LETTER K
        0x004C, # G+45  U+004C  L,  LATIN CAPITAL LETTER L
        0x004D, # G+46  U+004D  M,  LATIN CAPITAL LETTER M
        0x004E, # G+47  U+004E  N,  LATIN CAPITAL LETTER N
        0x004F, # G+48  U+004F  O,  LATIN CAPITAL LETTER O
        0x0050, # G+49  U+0050  P,  LATIN CAPITAL LETTER P
        0x0051, # G+50  U+0051  Q,  LATIN CAPITAL LETTER Q
        0x0052, # G+51  U+0052  R,  LATIN CAPITAL LETTER R
        0x0053, # G+52  U+0053  S,  LATIN CAPITAL LETTER S
        0x0054, # G+53  U+0054  T,  LATIN CAPITAL LETTER T
        0x0055, # G+54  U+0055  U,  LATIN CAPITAL LETTER U
        0x0056, # G+55  U+0056  V,  LATIN CAPITAL LETTER V
        0x0057, # G+56  U+0057  W,  LATIN CAPITAL LETTER W
        0x0058, # G+57  U+0058  X,  LATIN CAPITAL LETTER X
        0x0059, # G+58  U+0059  Y,  LATIN CAPITAL LETTER Y
        0x005A, # G+59  U+005A  Z,  LATIN CAPITAL LETTER Z
        0x005B, # G+60  U+005B  bracketleft,  LEFT SQUARE BRACKET
        0x005C, # G+61  U+005C  backslash,  REVERSE SOLIDUS
        0x005D, # G+62  U+005D  bracketright,  RIGHT SQUARE BRACKET
        0x005E, # G+63  U+005E  asciicircum,  CIRCUMFLEX ACCENT
        0x005F, # G+64  U+005F  underscore,  LOW LINE
        0x0060, # G+65  U+0060  grave,  GRAVE ACCENT
        0x0061, # G+66  U+0061  a,  LATIN SMALL LETTER A
        0x0062, # G+67  U+0062  b,  LATIN SMALL LETTER B
        0x0063, # G+68  U+0063  c,  LATIN SMALL LETTER C
        0x0064, # G+69  U+0064  d,  LATIN SMALL LETTER D
        0x0065, # G+70  U+0065  e,  LATIN SMALL LETTER E
        0x0066, # G+71  U+0066  f,  LATIN SMALL LETTER F
        0x0067, # G+72  U+0067  g,  LATIN SMALL LETTER G
        0x0068, # G+73  U+0068  h,  LATIN SMALL LETTER H
        0x0069, # G+74  U+0069  i,  LATIN SMALL LETTER I
        0x006A, # G+75  U+006A  j,  LATIN SMALL LETTER J
        0x006B, # G+76  U+006B  k,  LATIN SMALL LETTER K
        0x006C, # G+77  U+006C  l,  LATIN SMALL LETTER L
        0x006D, # G+78  U+006D  m,  LATIN SMALL LETTER M
        0x006E, # G+79  U+006E  n,  LATIN SMALL LETTER N
        0x006F, # G+80  U+006F  o,  LATIN SMALL LETTER O
        0x0070, # G+81  U+0070  p,  LATIN SMALL LETTER P
        0x0071, # G+82  U+0071  q,  LATIN SMALL LETTER Q
        0x0072, # G+83  U+0072  r,  LATIN SMALL LETTER R
        0x0073, # G+84  U+0073  s,  LATIN SMALL LETTER S
        0x0074, # G+85  U+0074  t,  LATIN SMALL LETTER T
        0x0075, # G+86  U+0075  u,  LATIN SMALL LETTER U
        0x0076, # G+87  U+0076  v,  LATIN SMALL LETTER V
        0x0077, # G+88  U+0077  w,  LATIN SMALL LETTER W
        0x0078, # G+89  U+0078  x,  LATIN SMALL LETTER X
        0x0079, # G+90  U+0079  y,  LATIN SMALL LETTER Y
        0x007A, # G+91  U+007A  z,  LATIN SMALL LETTER Z
        0x007B, # G+92  U+007B  braceleft,  LEFT CURLY BRACKET
        0x007C, # G+93  U+007C  bar,  VERTICAL LINE
        0x007D, # G+94  U+007D  braceright,  RIGHT CURLY BRACKET
        0x007E, # G+95  U+007E  asciitilde,  TILDE
        0x20A9, # G+96  U+20A9  won,  WON SIGN
        0x2010, # G+97  U+2010  hyphentwo,  HYPHEN
        0x00A9, # G+98  U+00A9  copyright,  COPYRIGHT SIGN
        0x0000, # G+99  U+0000  controlNUL,  <control>
        0x0000, # G+100  U+0000  controlNUL,  <control>
        0x3000, # G+101  U+3000  ideographicspace,  IDEOGRAPHIC SPACE
        0x3001, # G+102  U+3001  ideographiccomma,  IDEOGRAPHIC COMMA
        0x3002, # G+103  U+3002  ideographicperiod,  IDEOGRAPHIC FULL STOP
        0x00B7, # G+104  U+00B7  periodcentered,  MIDDLE DOT
        0x2025, # G+105  U+2025  twodotenleader,  TWO DOT LEADER
        0x2026, # G+106  U+2026  ellipsis,  HORIZONTAL ELLIPSIS
        0x00A8, # G+107  U+00A8  dieresis,  DIAERESIS
        0x3003, # G+108  U+3003  dittomark,  DITTO MARK
        0x00AD, # G+109  U+00AD  hyphen,  SOFT HYPHEN
        0x2015, # G+110  U+2015  afii00208,  HORIZONTAL BAR
        0x2225, # G+111  U+2225  parallel,  PARALLEL TO
        0xFF3C, # G+112  U+FF3C  backslashmonospace,  FULLWIDTH REVERSE SOLIDUS
        0x223C, # G+113  U+223C  similar,  TILDE OPERATOR
        0x2018, # G+114  U+2018  quoteleft,  LEFT SINGLE QUOTATION MARK
        0x2019, # G+115  U+2019  quoteright,  RIGHT SINGLE QUOTATION MARK
        0x201C, # G+116  U+201C  quotedblleft,  LEFT DOUBLE QUOTATION MARK
        0x201D, # G+117  U+201D  quotedblright,  RIGHT DOUBLE QUOTATION MARK
        0x3014, # G+118  U+3014  tortoiseshellbracketleft,  LEFT TORTOISE SHELL BRACKET
        0x3015, # G+119  U+3015  tortoiseshellbracketright,  RIGHT TORTOISE SHELL BRACKET
        0x3008, # G+120  U+3008  anglebracketleft,  LEFT ANGLE BRACKET
        0x3009, # G+121  U+3009  anglebracketright,  RIGHT ANGLE BRACKET
        0x300A, # G+122  U+300A  dblanglebracketleft,  LEFT DOUBLE ANGLE BRACKET
        0x300B, # G+123  U+300B  dblanglebracketright,  RIGHT DOUBLE ANGLE BRACKET
        0x300C, # G+124  U+300C  cornerbracketleft,  LEFT CORNER BRACKET
        0x300D, # G+125  U+300D  cornerbracketright,  RIGHT CORNER BRACKET
        0x300E, # G+126  U+300E  whitecornerbracketleft,  LEFT WHITE CORNER BRACKET
        0x300F, # G+127  U+300F  whitecornerbracketright,  RIGHT WHITE CORNER BRACKET
        0x3010, # G+128  U+3010  blacklenticularbracketleft,  LEFT BLACK LENTICULAR BRACKET
        0x3011, # G+129  U+3011  blacklenticularbracketright,  RIGHT BLACK LENTICULAR BRACKET
        0x00B1, # G+130  U+00B1  plusminus,  PLUS-MINUS SIGN
        0x00D7, # G+131  U+00D7  multiply,  MULTIPLICATION SIGN
        0x00F7, # G+132  U+00F7  divide,  DIVISION SIGN
        0x2260, # G+133  U+2260  notequal,  NOT EQUAL TO
        0x2264, # G+134  U+2264  lessequal,  LESS-THAN OR EQUAL TO
        0x2265, # G+135  U+2265  greaterequal,  GREATER-THAN OR EQUAL TO
        0x221E, # G+136  U+221E  infinity,  INFINITY
        0x2234, # G+137  U+2234  therefore,  THEREFORE
        0x00B0, # G+138  U+00B0  degree,  DEGREE SIGN
        0x2032, # G+139  U+2032  minute,  PRIME
        0x2033, # G+140  U+2033  second,  DOUBLE PRIME
        0x2103, # G+141  U+2103  centigrade,  DEGREE CELSIUS
        0x212B, # G+142  U+212B  angstrom,  ANGSTROM SIGN
        0xFFE0, # G+143  U+FFE0  centmonospace,  FULLWIDTH CENT SIGN
        0xFFE1, # G+144  U+FFE1  sterlingmonospace,  FULLWIDTH POUND SIGN
        0xFFE5, # G+145  U+FFE5  yenmonospace,  FULLWIDTH YEN SIGN
        0x2642, # G+146  U+2642  male,  MALE SIGN
        0x2640, # G+147  U+2640  female,  FEMALE SIGN
        0x2220, # G+148  U+2220  angle,  ANGLE
        0x22A5, # G+149  U+22A5  perpendicular,  UP TACK
        0x2312, # G+150  U+2312  arc,  ARC
        0x2202, # G+151  U+2202  partialdiff,  PARTIAL DIFFERENTIAL
        0x2207, # G+152  U+2207  gradient,  NABLA
        0x2261, # G+153  U+2261  equivalence,  IDENTICAL TO
        0x2252, # G+154  U+2252  approxequalorimage,  APPROXIMATELY EQUAL TO OR THE IMAGE OF
        0x00A7, # G+155  U+00A7  section,  SECTION SIGN
        0x203B, # G+156  U+203B  referencemark,  REFERENCE MARK
        0x2606, # G+157  U+2606  whitestar,  WHITE STAR
        0x2605, # G+158  U+2605  blackstar,  BLACK STAR
        0x25CB, # G+159  U+25CB  circle,  WHITE CIRCLE
        0x25CF, # G+160  U+25CF  H18533,  BLACK CIRCLE
        0x25CE, # G+161  U+25CE  bullseye,  BULLSEYE
        0x25C7, # G+162  U+25C7  whitediamond,  WHITE DIAMOND
        0x25C6, # G+163  U+25C6  blackdiamond,  BLACK DIAMOND
        0x25A1, # G+164  U+25A1  H22073,  WHITE SQUARE
        0x25A0, # G+165  U+25A0  filledbox,  BLACK SQUARE
        0x25B3, # G+166  U+25B3  whiteuppointingtriangle,  WHITE UP-POINTING TRIANGLE
        0x25B2, # G+167  U+25B2  triagup,  BLACK UP-POINTING TRIANGLE
        0x25BD, # G+168  U+25BD  whitedownpointingtriangle,  WHITE DOWN-POINTING TRIANGLE
        0x25BC, # G+169  U+25BC  triagdn,  BLACK DOWN-POINTING TRIANGLE
        0x2192, # G+170  U+2192  arrowright,  RIGHTWARDS ARROW
        0x2190, # G+171  U+2190  arrowleft,  LEFTWARDS ARROW
        0x2191, # G+172  U+2191  arrowup,  UPWARDS ARROW
        0x2193, # G+173  U+2193  arrowdown,  DOWNWARDS ARROW
        0x2194, # G+174  U+2194  arrowboth,  LEFT RIGHT ARROW
        0x3013, # G+175  U+3013  getamark,  GETA MARK
        0x226A, # G+176  U+226A  muchless,  MUCH LESS-THAN
        0x226B, # G+177  U+226B  muchgreater,  MUCH GREATER-THAN
        0x221A, # G+178  U+221A  radical,  SQUARE ROOT
        0x223D, # G+179  U+223D  reversedtilde,  REVERSED TILDE
        0x221D, # G+180  U+221D  proportional,  PROPORTIONAL TO
        0x2235, # G+181  U+2235  because,  BECAUSE
        0x222B, # G+182  U+222B  integral,  INTEGRAL
        0x222C, # G+183  U+222C  dblintegral,  DOUBLE INTEGRAL
        0x2208, # G+184  U+2208  element,  ELEMENT OF
        0x220B, # G+185  U+220B  suchthat,  CONTAINS AS MEMBER
        0x2286, # G+186  U+2286  reflexsubset,  SUBSET OF OR EQUAL TO
        0x2287, # G+187  U+2287  reflexsuperset,  SUPERSET OF OR EQUAL TO
        0x2282, # G+188  U+2282  propersubset,  SUBSET OF
        0x2283, # G+189  U+2283  propersuperset,  SUPERSET OF
        0x222A, # G+190  U+222A  union,  UNION
        0x2229, # G+191  U+2229  intersection,  INTERSECTION
        0x2227, # G+192  U+2227  logicaland,  LOGICAL AND
        0x2228, # G+193  U+2228  logicalor,  LOGICAL OR
        0xFFE2, # G+194  U+FFE2  uniFFE2,  FULLWIDTH NOT SIGN
        0x21D2, # G+195  U+21D2  arrowdblright,  RIGHTWARDS DOUBLE ARROW
        0x21D4, # G+196  U+21D4  arrowdblboth,  LEFT RIGHT DOUBLE ARROW
        0x2200, # G+197  U+2200  universal,  FOR ALL
        0x2203, # G+198  U+2203  existential,  THERE EXISTS
        0x00B4, # G+199  U+00B4  acute,  ACUTE ACCENT
        0xFF5E, # G+200  U+FF5E  asciitildemonospace,  FULLWIDTH TILDE
        0x02C7, # G+201  U+02C7  caron,  CARON
        0x02D8, # G+202  U+02D8  breve,  BREVE
        0x02DD, # G+203  U+02DD  hungarumlaut,  DOUBLE ACUTE ACCENT
        0x02DA, # G+204  U+02DA  ring,  RING ABOVE
        0x02D9, # G+205  U+02D9  dotaccent,  DOT ABOVE
        0x00B8, # G+206  U+00B8  cedilla,  CEDILLA
        0x02DB, # G+207  U+02DB  ogonek,  OGONEK
        0x00A1, # G+208  U+00A1  exclamdown,  INVERTED EXCLAMATION MARK
        0x00BF, # G+209  U+00BF  questiondown,  INVERTED QUESTION MARK
        0x02D0, # G+210  U+02D0  colontriangularmod,  MODIFIER LETTER TRIANGULAR COLON
        0x222E, # G+211  U+222E  contourintegral,  CONTOUR INTEGRAL
        0x2211, # G+212  U+2211  summation,  N-ARY SUMMATION
        0x220F, # G+213  U+220F  product,  N-ARY PRODUCT
        0x00A4, # G+214  U+00A4  currency,  CURRENCY SIGN
        0x2109, # G+215  U+2109  fahrenheit,  DEGREE FAHRENHEIT
        0x2030, # G+216  U+2030  perthousand,  PER MILLE SIGN
        0x25C1, # G+217  U+25C1  whiteleftpointingtriangle,  WHITE LEFT-POINTING TRIANGLE
        0x25C0, # G+218  U+25C0  blackleftpointingtriangle,  BLACK LEFT-POINTING TRIANGLE
        0x25B7, # G+219  U+25B7  whiterightpointingtriangle,  WHITE RIGHT-POINTING TRIANGLE
        0x25B6, # G+220  U+25B6  blackrightpointingtriangle,  BLACK RIGHT-POINTING TRIANGLE
        0x2664, # G+221  U+2664  spadesuitwhite,  WHITE SPADE SUIT
        0x2660, # G+222  U+2660  spade,  BLACK SPADE SUIT
        0x2661, # G+223  U+2661  heartsuitwhite,  WHITE HEART SUIT
        0x2665, # G+224  U+2665  heart,  BLACK HEART SUIT
        0x2667, # G+225  U+2667  clubsuitwhite,  WHITE CLUB SUIT
        0x2663, # G+226  U+2663  club,  BLACK CLUB SUIT
        0x2299, # G+227  U+2299  circleot,  CIRCLED DOT OPERATOR
        0x25C8, # G+228  U+25C8  whitediamondcontainingblacksmalldiamond,  WHITE DIAMOND CONTAINING BLACK SMALL DIAMOND
        0x25A3, # G+229  U+25A3  squarewhitewithsmallblack,  WHITE SQUARE CONTAINING BLACK SMALL SQUARE
        0x25D0, # G+230  U+25D0  circlewithlefthalfblack,  CIRCLE WITH LEFT HALF BLACK
        0x25D1, # G+231  U+25D1  circlewithrighthalfblack,  CIRCLE WITH RIGHT HALF BLACK
        0x2592, # G+232  U+2592  shade,  MEDIUM SHADE
        0x25A4, # G+233  U+25A4  squarehorizontalfill,  SQUARE WITH HORIZONTAL FILL
        0x25A5, # G+234  U+25A5  squareverticalfill,  SQUARE WITH VERTICAL FILL
        0x25A8, # G+235  U+25A8  squareupperrighttolowerleftfill,  SQUARE WITH UPPER RIGHT TO LOWER LEFT FILL
        0x25A7, # G+236  U+25A7  squareupperlefttolowerrightfill,  SQUARE WITH UPPER LEFT TO LOWER RIGHT FILL
        0x25A6, # G+237  U+25A6  squareorthogonalcrosshatchfill,  SQUARE WITH ORTHOGONAL CROSSHATCH FILL
        0x25A9, # G+238  U+25A9  squarediagonalcrosshatchfill,  SQUARE WITH DIAGONAL CROSSHATCH FILL
        0x2668, # G+239  U+2668  hotsprings,  HOT SPRINGS
        0x260F, # G+240  U+260F  whitetelephone,  WHITE TELEPHONE
        0x260E, # G+241  U+260E  telephoneblack,  BLACK TELEPHONE
        0x261C, # G+242  U+261C  pointingindexleftwhite,  WHITE LEFT POINTING INDEX
        0x261E, # G+243  U+261E  pointingindexrightwhite,  WHITE RIGHT POINTING INDEX
        0x00B6, # G+244  U+00B6  paragraph,  PILCROW SIGN
        0x2020, # G+245  U+2020  dagger,  DAGGER
        0x2021, # G+246  U+2021  daggerdbl,  DOUBLE DAGGER
        0x2195, # G+247  U+2195  arrowupdn,  UP DOWN ARROW
        0x2197, # G+248  U+2197  arrowupright,  NORTH EAST ARROW
        0x2199, # G+249  U+2199  arrowdownleft,  SOUTH WEST ARROW
        0x2196, # G+250  U+2196  arrowupleft,  NORTH WEST ARROW
        0x2198, # G+251  U+2198  arrowdownright,  SOUTH EAST ARROW
        0x266D, # G+252  U+266D  musicflatsign,  MUSIC FLAT SIGN
        0x2669, # G+253  U+2669  quarternote,  QUARTER NOTE
        0x266A, # G+254  U+266A  musicalnote,  EIGHTH NOTE
        0x266C, # G+255  U+266C  beamedsixteenthnotes,  BEAMED SIXTEENTH NOTES
        0x327F, # G+256  U+327F  koreanstandardsymbol,  KOREAN STANDARD SYMBOL
        0x321C, # G+257  U+321C  cieucuparenkorean,  PARENTHESIZED HANGUL CIEUC U
        0x2116, # G+258  U+2116  afii61352,  NUMERO SIGN
        0x33C7, # G+259  U+33C7  cosquare,  SQUARE CO
        0x2122, # G+260  U+2122  trademark,  TRADE MARK SIGN
        0x33C2, # G+261  U+33C2  amsquare,  SQUARE AM
        0x33D8, # G+262  U+33D8  pmsquare,  SQUARE PM
        0x2121, # G+263  U+2121  telephone,  TELEPHONE SIGN
        0xFF01, # G+264  U+FF01  exclammonospace,  FULLWIDTH EXCLAMATION MARK
        0xFF02, # G+265  U+FF02  quotedblmonospace,  FULLWIDTH QUOTATION MARK
        0xFF03, # G+266  U+FF03  numbersignmonospace,  FULLWIDTH NUMBER SIGN
        0xFF04, # G+267  U+FF04  dollarmonospace,  FULLWIDTH DOLLAR SIGN
        0xFF05, # G+268  U+FF05  percentmonospace,  FULLWIDTH PERCENT SIGN
        0xFF06, # G+269  U+FF06  ampersandmonospace,  FULLWIDTH AMPERSAND
        0xFF07, # G+270  U+FF07  quotesinglemonospace,  FULLWIDTH APOSTROPHE
        0xFF08, # G+271  U+FF08  parenleftmonospace,  FULLWIDTH LEFT PARENTHESIS
        0xFF09, # G+272  U+FF09  parenrightmonospace,  FULLWIDTH RIGHT PARENTHESIS
        0xFF0A, # G+273  U+FF0A  asteriskmonospace,  FULLWIDTH ASTERISK
        0xFF0B, # G+274  U+FF0B  plusmonospace,  FULLWIDTH PLUS SIGN
        0xFF0C, # G+275  U+FF0C  commamonospace,  FULLWIDTH COMMA
        0xFF0D, # G+276  U+FF0D  hyphenmonospace,  FULLWIDTH HYPHEN-MINUS
        0xFF0E, # G+277  U+FF0E  periodmonospace,  FULLWIDTH FULL STOP
        0xFF0F, # G+278  U+FF0F  slashmonospace,  FULLWIDTH SOLIDUS
        0xFF10, # G+279  U+FF10  zeromonospace,  FULLWIDTH DIGIT ZERO
        0xFF11, # G+280  U+FF11  onemonospace,  FULLWIDTH DIGIT ONE
        0xFF12, # G+281  U+FF12  twomonospace,  FULLWIDTH DIGIT TWO
        0xFF13, # G+282  U+FF13  threemonospace,  FULLWIDTH DIGIT THREE
        0xFF14, # G+283  U+FF14  fourmonospace,  FULLWIDTH DIGIT FOUR
        0xFF15, # G+284  U+FF15  fivemonospace,  FULLWIDTH DIGIT FIVE
        0xFF16, # G+285  U+FF16  sixmonospace,  FULLWIDTH DIGIT SIX
        0xFF17, # G+286  U+FF17  sevenmonospace,  FULLWIDTH DIGIT SEVEN
        0xFF18, # G+287  U+FF18  eightmonospace,  FULLWIDTH DIGIT EIGHT
        0xFF19, # G+288  U+FF19  ninemonospace,  FULLWIDTH DIGIT NINE
        0xFF1A, # G+289  U+FF1A  colonmonospace,  FULLWIDTH COLON
        0xFF1B, # G+290  U+FF1B  semicolonmonospace,  FULLWIDTH SEMICOLON
        0xFF1C, # G+291  U+FF1C  lessmonospace,  FULLWIDTH LESS-THAN SIGN
        0xFF1D, # G+292  U+FF1D  equalmonospace,  FULLWIDTH EQUALS SIGN
        0xFF1E, # G+293  U+FF1E  greatermonospace,  FULLWIDTH GREATER-THAN SIGN
        0xFF1F, # G+294  U+FF1F  questionmonospace,  FULLWIDTH QUESTION MARK
        0xFF20, # G+295  U+FF20  atmonospace,  FULLWIDTH COMMERCIAL AT
        0xFF21, # G+296  U+FF21  Amonospace,  FULLWIDTH LATIN CAPITAL LETTER A
        0xFF22, # G+297  U+FF22  Bmonospace,  FULLWIDTH LATIN CAPITAL LETTER B
        0xFF23, # G+298  U+FF23  Cmonospace,  FULLWIDTH LATIN CAPITAL LETTER C
        0xFF24, # G+299  U+FF24  Dmonospace,  FULLWIDTH LATIN CAPITAL LETTER D
        0xFF25, # G+300  U+FF25  Emonospace,  FULLWIDTH LATIN CAPITAL LETTER E
        0xFF26, # G+301  U+FF26  Fmonospace,  FULLWIDTH LATIN CAPITAL LETTER F
        0xFF27, # G+302  U+FF27  Gmonospace,  FULLWIDTH LATIN CAPITAL LETTER G
        0xFF28, # G+303  U+FF28  Hmonospace,  FULLWIDTH LATIN CAPITAL LETTER H
        0xFF29, # G+304  U+FF29  Imonospace,  FULLWIDTH LATIN CAPITAL LETTER I
        0xFF2A, # G+305  U+FF2A  Jmonospace,  FULLWIDTH LATIN CAPITAL LETTER J
        0xFF2B, # G+306  U+FF2B  Kmonospace,  FULLWIDTH LATIN CAPITAL LETTER K
        0xFF2C, # G+307  U+FF2C  Lmonospace,  FULLWIDTH LATIN CAPITAL LETTER L
        0xFF2D, # G+308  U+FF2D  Mmonospace,  FULLWIDTH LATIN CAPITAL LETTER M
        0xFF2E, # G+309  U+FF2E  Nmonospace,  FULLWIDTH LATIN CAPITAL LETTER N
        0xFF2F, # G+310  U+FF2F  Omonospace,  FULLWIDTH LATIN CAPITAL LETTER O
        0xFF30, # G+311  U+FF30  Pmonospace,  FULLWIDTH LATIN CAPITAL LETTER P
        0xFF31, # G+312  U+FF31  Qmonospace,  FULLWIDTH LATIN CAPITAL LETTER Q
        0xFF32, # G+313  U+FF32  Rmonospace,  FULLWIDTH LATIN CAPITAL LETTER R
        0xFF33, # G+314  U+FF33  Smonospace,  FULLWIDTH LATIN CAPITAL LETTER S
        0xFF34, # G+315  U+FF34  Tmonospace,  FULLWIDTH LATIN CAPITAL LETTER T
        0xFF35, # G+316  U+FF35  Umonospace,  FULLWIDTH LATIN CAPITAL LETTER U
        0xFF36, # G+317  U+FF36  Vmonospace,  FULLWIDTH LATIN CAPITAL LETTER V
        0xFF37, # G+318  U+FF37  Wmonospace,  FULLWIDTH LATIN CAPITAL LETTER W
        0xFF38, # G+319  U+FF38  Xmonospace,  FULLWIDTH LATIN CAPITAL LETTER X
        0xFF39, # G+320  U+FF39  Ymonospace,  FULLWIDTH LATIN CAPITAL LETTER Y
        0xFF3A, # G+321  U+FF3A  Zmonospace,  FULLWIDTH LATIN CAPITAL LETTER Z
        0xFF3B, # G+322  U+FF3B  bracketleftmonospace,  FULLWIDTH LEFT SQUARE BRACKET
        0xFFE6, # G+323  U+FFE6  wonmonospace,  FULLWIDTH WON SIGN
        0xFF3D, # G+324  U+FF3D  bracketrightmonospace,  FULLWIDTH RIGHT SQUARE BRACKET
        0xFF3E, # G+325  U+FF3E  asciicircummonospace,  FULLWIDTH CIRCUMFLEX ACCENT
        0xFF3F, # G+326  U+FF3F  underscoremonospace,  FULLWIDTH LOW LINE
        0xFF40, # G+327  U+FF40  gravemonospace,  FULLWIDTH GRAVE ACCENT
        0xFF41, # G+328  U+FF41  amonospace,  FULLWIDTH LATIN SMALL LETTER A
        0xFF42, # G+329  U+FF42  bmonospace,  FULLWIDTH LATIN SMALL LETTER B
        0xFF43, # G+330  U+FF43  cmonospace,  FULLWIDTH LATIN SMALL LETTER C
        0xFF44, # G+331  U+FF44  dmonospace,  FULLWIDTH LATIN SMALL LETTER D
        0xFF45, # G+332  U+FF45  emonospace,  FULLWIDTH LATIN SMALL LETTER E
        0xFF46, # G+333  U+FF46  fmonospace,  FULLWIDTH LATIN SMALL LETTER F
        0xFF47, # G+334  U+FF47  gmonospace,  FULLWIDTH LATIN SMALL LETTER G
        0xFF48, # G+335  U+FF48  hmonospace,  FULLWIDTH LATIN SMALL LETTER H
        0xFF49, # G+336  U+FF49  imonospace,  FULLWIDTH LATIN SMALL LETTER I
        0xFF4A, # G+337  U+FF4A  jmonospace,  FULLWIDTH LATIN SMALL LETTER J
        0xFF4B, # G+338  U+FF4B  kmonospace,  FULLWIDTH LATIN SMALL LETTER K
        0xFF4C, # G+339  U+FF4C  lmonospace,  FULLWIDTH LATIN SMALL LETTER L
        0xFF4D, # G+340  U+FF4D  mmonospace,  FULLWIDTH LATIN SMALL LETTER M
        0xFF4E, # G+341  U+FF4E  nmonospace,  FULLWIDTH LATIN SMALL LETTER N
        0xFF4F, # G+342  U+FF4F  omonospace,  FULLWIDTH LATIN SMALL LETTER O
        0xFF50, # G+343  U+FF50  pmonospace,  FULLWIDTH LATIN SMALL LETTER P
        0xFF51, # G+344  U+FF51  qmonospace,  FULLWIDTH LATIN SMALL LETTER Q
        0xFF52, # G+345  U+FF52  rmonospace,  FULLWIDTH LATIN SMALL LETTER R
        0xFF53, # G+346  U+FF53  smonospace,  FULLWIDTH LATIN SMALL LETTER S
        0xFF54, # G+347  U+FF54  tmonospace,  FULLWIDTH LATIN SMALL LETTER T
        0xFF55, # G+348  U+FF55  umonospace,  FULLWIDTH LATIN SMALL LETTER U
        0xFF56, # G+349  U+FF56  vmonospace,  FULLWIDTH LATIN SMALL LETTER V
        0xFF57, # G+350  U+FF57  wmonospace,  FULLWIDTH LATIN SMALL LETTER W
        0xFF58, # G+351  U+FF58  xmonospace,  FULLWIDTH LATIN SMALL LETTER X
        0xFF59, # G+352  U+FF59  ymonospace,  FULLWIDTH LATIN SMALL LETTER Y
        0xFF5A, # G+353  U+FF5A  zmonospace,  FULLWIDTH LATIN SMALL LETTER Z
        0xFF5B, # G+354  U+FF5B  braceleftmonospace,  FULLWIDTH LEFT CURLY BRACKET
        0xFF5C, # G+355  U+FF5C  barmonospace,  FULLWIDTH VERTICAL LINE
        0xFF5D, # G+356  U+FF5D  bracerightmonospace,  FULLWIDTH RIGHT CURLY BRACKET
        0xFFE3, # G+357  U+FFE3  macronmonospace,  FULLWIDTH MACRON
        0x3131, # G+358  U+3131  kiyeokkorean,  HANGUL LETTER KIYEOK
        0x3132, # G+359  U+3132  ssangkiyeokkorean,  HANGUL LETTER SSANGKIYEOK
        0x3133, # G+360  U+3133  kiyeoksioskorean,  HANGUL LETTER KIYEOK-SIOS
        0x3134, # G+361  U+3134  nieunkorean,  HANGUL LETTER NIEUN
        0x3135, # G+362  U+3135  nieuncieuckorean,  HANGUL LETTER NIEUN-CIEUC
        0x3136, # G+363  U+3136  nieunhieuhkorean,  HANGUL LETTER NIEUN-HIEUH
        0x3137, # G+364  U+3137  tikeutkorean,  HANGUL LETTER TIKEUT
        0x3138, # G+365  U+3138  ssangtikeutkorean,  HANGUL LETTER SSANGTIKEUT
        0x3139, # G+366  U+3139  rieulkorean,  HANGUL LETTER RIEUL
        0x313A, # G+367  U+313A  rieulkiyeokkorean,  HANGUL LETTER RIEUL-KIYEOK
        0x313B, # G+368  U+313B  rieulmieumkorean,  HANGUL LETTER RIEUL-MIEUM
        0x313C, # G+369  U+313C  rieulpieupkorean,  HANGUL LETTER RIEUL-PIEUP
        0x313D, # G+370  U+313D  rieulsioskorean,  HANGUL LETTER RIEUL-SIOS
        0x313E, # G+371  U+313E  rieulthieuthkorean,  HANGUL LETTER RIEUL-THIEUTH
        0x313F, # G+372  U+313F  rieulphieuphkorean,  HANGUL LETTER RIEUL-PHIEUPH
        0x3140, # G+373  U+3140  rieulhieuhkorean,  HANGUL LETTER RIEUL-HIEUH
        0x3141, # G+374  U+3141  mieumkorean,  HANGUL LETTER MIEUM
        0x3142, # G+375  U+3142  pieupkorean,  HANGUL LETTER PIEUP
        0x3143, # G+376  U+3143  ssangpieupkorean,  HANGUL LETTER SSANGPIEUP
        0x3144, # G+377  U+3144  pieupsioskorean,  HANGUL LETTER PIEUP-SIOS
        0x3145, # G+378  U+3145  sioskorean,  HANGUL LETTER SIOS
        0x3146, # G+379  U+3146  ssangsioskorean,  HANGUL LETTER SSANGSIOS
        0x3147, # G+380  U+3147  ieungkorean,  HANGUL LETTER IEUNG
        0x3148, # G+381  U+3148  cieuckorean,  HANGUL LETTER CIEUC
        0x3149, # G+382  U+3149  ssangcieuckorean,  HANGUL LETTER SSANGCIEUC
        0x314A, # G+383  U+314A  chieuchkorean,  HANGUL LETTER CHIEUCH
        0x314B, # G+384  U+314B  khieukhkorean,  HANGUL LETTER KHIEUKH
        0x314C, # G+385  U+314C  thieuthkorean,  HANGUL LETTER THIEUTH
        0x314D, # G+386  U+314D  phieuphkorean,  HANGUL LETTER PHIEUPH
        0x314E, # G+387  U+314E  hieuhkorean,  HANGUL LETTER HIEUH
        0x314F, # G+388  U+314F  akorean,  HANGUL LETTER A
        0x3150, # G+389  U+3150  aekorean,  HANGUL LETTER AE
        0x3151, # G+390  U+3151  yakorean,  HANGUL LETTER YA
        0x3152, # G+391  U+3152  yaekorean,  HANGUL LETTER YAE
        0x3153, # G+392  U+3153  eokorean,  HANGUL LETTER EO
        0x3154, # G+393  U+3154  ekorean,  HANGUL LETTER E
        0x3155, # G+394  U+3155  yeokorean,  HANGUL LETTER YEO
        0x3156, # G+395  U+3156  yekorean,  HANGUL LETTER YE
        0x3157, # G+396  U+3157  okorean,  HANGUL LETTER O
        0x3158, # G+397  U+3158  wakorean,  HANGUL LETTER WA
        0x3159, # G+398  U+3159  waekorean,  HANGUL LETTER WAE
        0x315A, # G+399  U+315A  oekorean,  HANGUL LETTER OE
        0x315B, # G+400  U+315B  yokorean,  HANGUL LETTER YO
        0x315C, # G+401  U+315C  ukorean,  HANGUL LETTER U
        0x315D, # G+402  U+315D  weokorean,  HANGUL LETTER WEO
        0x315E, # G+403  U+315E  wekorean,  HANGUL LETTER WE
        0x315F, # G+404  U+315F  wikorean,  HANGUL LETTER WI
        0x3160, # G+405  U+3160  yukorean,  HANGUL LETTER YU
        0x3161, # G+406  U+3161  eukorean,  HANGUL LETTER EU
        0x3162, # G+407  U+3162  yikorean,  HANGUL LETTER YI
        0x3163, # G+408  U+3163  ikorean,  HANGUL LETTER I
        0x3165, # G+409  U+3165  ssangnieunkorean,  HANGUL LETTER SSANGNIEUN
        0x3166, # G+410  U+3166  nieuntikeutkorean,  HANGUL LETTER NIEUN-TIKEUT
        0x3167, # G+411  U+3167  nieunsioskorean,  HANGUL LETTER NIEUN-SIOS
        0x3168, # G+412  U+3168  nieunpansioskorean,  HANGUL LETTER NIEUN-PANSIOS
        0x3169, # G+413  U+3169  rieulkiyeoksioskorean,  HANGUL LETTER RIEUL-KIYEOK-SIOS
        0x316A, # G+414  U+316A  rieultikeutkorean,  HANGUL LETTER RIEUL-TIKEUT
        0x316B, # G+415  U+316B  rieulpieupsioskorean,  HANGUL LETTER RIEUL-PIEUP-SIOS
        0x316C, # G+416  U+316C  rieulpansioskorean,  HANGUL LETTER RIEUL-PANSIOS
        0x316D, # G+417  U+316D  rieulyeorinhieuhkorean,  HANGUL LETTER RIEUL-YEORINHIEUH
        0x316E, # G+418  U+316E  mieumpieupkorean,  HANGUL LETTER MIEUM-PIEUP
        0x316F, # G+419  U+316F  mieumsioskorean,  HANGUL LETTER MIEUM-SIOS
        0x3170, # G+420  U+3170  mieumpansioskorean,  HANGUL LETTER MIEUM-PANSIOS
        0x3171, # G+421  U+3171  kapyeounmieumkorean,  HANGUL LETTER KAPYEOUNMIEUM
        0x3172, # G+422  U+3172  pieupkiyeokkorean,  HANGUL LETTER PIEUP-KIYEOK
        0x3173, # G+423  U+3173  pieuptikeutkorean,  HANGUL LETTER PIEUP-TIKEUT
        0x3174, # G+424  U+3174  pieupsioskiyeokkorean,  HANGUL LETTER PIEUP-SIOS-KIYEOK
        0x3175, # G+425  U+3175  pieupsiostikeutkorean,  HANGUL LETTER PIEUP-SIOS-TIKEUT
        0x3176, # G+426  U+3176  pieupcieuckorean,  HANGUL LETTER PIEUP-CIEUC
        0x3177, # G+427  U+3177  pieupthieuthkorean,  HANGUL LETTER PIEUP-THIEUTH
        0x3178, # G+428  U+3178  kapyeounpieupkorean,  HANGUL LETTER KAPYEOUNPIEUP
        0x3179, # G+429  U+3179  kapyeounssangpieupkorean,  HANGUL LETTER KAPYEOUNSSANGPIEUP
        0x317A, # G+430  U+317A  sioskiyeokkorean,  HANGUL LETTER SIOS-KIYEOK
        0x317B, # G+431  U+317B  siosnieunkorean,  HANGUL LETTER SIOS-NIEUN
        0x317C, # G+432  U+317C  siostikeutkorean,  HANGUL LETTER SIOS-TIKEUT
        0x317D, # G+433  U+317D  siospieupkorean,  HANGUL LETTER SIOS-PIEUP
        0x317E, # G+434  U+317E  sioscieuckorean,  HANGUL LETTER SIOS-CIEUC
        0x317F, # G+435  U+317F  pansioskorean,  HANGUL LETTER PANSIOS
        0x3180, # G+436  U+3180  ssangieungkorean,  HANGUL LETTER SSANGIEUNG
        0x3181, # G+437  U+3181  yesieungkorean,  HANGUL LETTER YESIEUNG
        0x3182, # G+438  U+3182  yesieungsioskorean,  HANGUL LETTER YESIEUNG-SIOS
        0x3183, # G+439  U+3183  yesieungpansioskorean,  HANGUL LETTER YESIEUNG-PANSIOS
        0x3184, # G+440  U+3184  kapyeounphieuphkorean,  HANGUL LETTER KAPYEOUNPHIEUPH
        0x3185, # G+441  U+3185  ssanghieuhkorean,  HANGUL LETTER SSANGHIEUH
        0x3186, # G+442  U+3186  yeorinhieuhkorean,  HANGUL LETTER YEORINHIEUH
        0x3187, # G+443  U+3187  yoyakorean,  HANGUL LETTER YO-YA
        0x3188, # G+444  U+3188  yoyaekorean,  HANGUL LETTER YO-YAE
        0x3189, # G+445  U+3189  yoikorean,  HANGUL LETTER YO-I
        0x318A, # G+446  U+318A  yuyeokorean,  HANGUL LETTER YU-YEO
        0x318B, # G+447  U+318B  yuyekorean,  HANGUL LETTER YU-YE
        0x318C, # G+448  U+318C  yuikorean,  HANGUL LETTER YU-I
        0x318D, # G+449  U+318D  araeakorean,  HANGUL LETTER ARAEA
        0x318E, # G+450  U+318E  araeaekorean,  HANGUL LETTER ARAEAE
        0x2170, # G+451  U+2170  oneroman,  SMALL ROMAN NUMERAL ONE
        0x2171, # G+452  U+2171  tworoman,  SMALL ROMAN NUMERAL TWO
        0x2172, # G+453  U+2172  threeroman,  SMALL ROMAN NUMERAL THREE
        0x2173, # G+454  U+2173  fourroman,  SMALL ROMAN NUMERAL FOUR
        0x2174, # G+455  U+2174  fiveroman,  SMALL ROMAN NUMERAL FIVE
        0x2175, # G+456  U+2175  sixroman,  SMALL ROMAN NUMERAL SIX
        0x2176, # G+457  U+2176  sevenroman,  SMALL ROMAN NUMERAL SEVEN
        0x2177, # G+458  U+2177  eightroman,  SMALL ROMAN NUMERAL EIGHT
        0x2178, # G+459  U+2178  nineroman,  SMALL ROMAN NUMERAL NINE
        0x2179, # G+460  U+2179  tenroman,  SMALL ROMAN NUMERAL TEN
        0x2160, # G+461  U+2160  Oneroman,  ROMAN NUMERAL ONE
        0x2161, # G+462  U+2161  Tworoman,  ROMAN NUMERAL TWO
        0x2162, # G+463  U+2162  Threeroman,  ROMAN NUMERAL THREE
        0x2163, # G+464  U+2163  Fourroman,  ROMAN NUMERAL FOUR
        0x2164, # G+465  U+2164  Fiveroman,  ROMAN NUMERAL FIVE
        0x2165, # G+466  U+2165  Sixroman,  ROMAN NUMERAL SIX
        0x2166, # G+467  U+2166  Sevenroman,  ROMAN NUMERAL SEVEN
        0x2167, # G+468  U+2167  Eightroman,  ROMAN NUMERAL EIGHT
        0x2168, # G+469  U+2168  Nineroman,  ROMAN NUMERAL NINE
        0x2169, # G+470  U+2169  Tenroman,  ROMAN NUMERAL TEN
        0x0391, # G+471  U+0391  Alpha,  GREEK CAPITAL LETTER ALPHA
        0x0392, # G+472  U+0392  Beta,  GREEK CAPITAL LETTER BETA
        0x0393, # G+473  U+0393  Gamma,  GREEK CAPITAL LETTER GAMMA
        0x0394, # G+474  U+0394  Delta,  GREEK CAPITAL LETTER DELTA
        0x0395, # G+475  U+0395  Epsilon,  GREEK CAPITAL LETTER EPSILON
        0x0396, # G+476  U+0396  Zeta,  GREEK CAPITAL LETTER ZETA
        0x0397, # G+477  U+0397  Eta,  GREEK CAPITAL LETTER ETA
        0x0398, # G+478  U+0398  Theta,  GREEK CAPITAL LETTER THETA
        0x0399, # G+479  U+0399  Iota,  GREEK CAPITAL LETTER IOTA
        0x039A, # G+480  U+039A  Kappa,  GREEK CAPITAL LETTER KAPPA
        0x039B, # G+481  U+039B  Lambda,  GREEK CAPITAL LETTER LAMDA
        0x039C, # G+482  U+039C  Mu,  GREEK CAPITAL LETTER MU
        0x039D, # G+483  U+039D  Nu,  GREEK CAPITAL LETTER NU
        0x039E, # G+484  U+039E  Xi,  GREEK CAPITAL LETTER XI
        0x039F, # G+485  U+039F  Omicron,  GREEK CAPITAL LETTER OMICRON
        0x03A0, # G+486  U+03A0  Pi,  GREEK CAPITAL LETTER PI
        0x03A1, # G+487  U+03A1  Rho,  GREEK CAPITAL LETTER RHO
        0x03A3, # G+488  U+03A3  Sigma,  GREEK CAPITAL LETTER SIGMA
        0x03A4, # G+489  U+03A4  Tau,  GREEK CAPITAL LETTER TAU
        0x03A5, # G+490  U+03A5  Upsilon,  GREEK CAPITAL LETTER UPSILON
        0x03A6, # G+491  U+03A6  Phi,  GREEK CAPITAL LETTER PHI
        0x03A7, # G+492  U+03A7  Chi,  GREEK CAPITAL LETTER CHI
        0x03A8, # G+493  U+03A8  Psi,  GREEK CAPITAL LETTER PSI
        0x03A9, # G+494  U+03A9  Omega,  GREEK CAPITAL LETTER OMEGA
        0x03B1, # G+495  U+03B1  alpha,  GREEK SMALL LETTER ALPHA
        0x03B2, # G+496  U+03B2  beta,  GREEK SMALL LETTER BETA
        0x03B3, # G+497  U+03B3  gamma,  GREEK SMALL LETTER GAMMA
        0x03B4, # G+498  U+03B4  delta,  GREEK SMALL LETTER DELTA
        0x03B5, # G+499  U+03B5  epsilon,  GREEK SMALL LETTER EPSILON
        0x03B6, # G+500  U+03B6  zeta,  GREEK SMALL LETTER ZETA
        0x03B7, # G+501  U+03B7  eta,  GREEK SMALL LETTER ETA
        0x03B8, # G+502  U+03B8  theta,  GREEK SMALL LETTER THETA
        0x03B9, # G+503  U+03B9  iota,  GREEK SMALL LETTER IOTA
        0x03BA, # G+504  U+03BA  kappa,  GREEK SMALL LETTER KAPPA
        0x03BB, # G+505  U+03BB  lambda,  GREEK SMALL LETTER LAMDA
        0x03BC, # G+506  U+03BC  mu,  GREEK SMALL LETTER MU
        0x03BD, # G+507  U+03BD  nu,  GREEK SMALL LETTER NU
        0x03BE, # G+508  U+03BE  xi,  GREEK SMALL LETTER XI
        0x03BF, # G+509  U+03BF  omicron,  GREEK SMALL LETTER OMICRON
        0x03C0, # G+510  U+03C0  pi,  GREEK SMALL LETTER PI
        0x03C1, # G+511  U+03C1  rho,  GREEK SMALL LETTER RHO
        0x03C3, # G+512  U+03C3  sigma,  GREEK SMALL LETTER SIGMA
        0x03C4, # G+513  U+03C4  tau,  GREEK SMALL LETTER TAU
        0x03C5, # G+514  U+03C5  upsilon,  GREEK SMALL LETTER UPSILON
        0x03C6, # G+515  U+03C6  phi,  GREEK SMALL LETTER PHI
        0x03C7, # G+516  U+03C7  chi,  GREEK SMALL LETTER CHI
        0x03C8, # G+517  U+03C8  psi,  GREEK SMALL LETTER PSI
        0x03C9, # G+518  U+03C9  omega,  GREEK SMALL LETTER OMEGA
        0x2500, # G+519  U+2500  SF100000,  BOX DRAWINGS LIGHT HORIZONTAL
        0x2502, # G+520  U+2502  SF110000,  BOX DRAWINGS LIGHT VERTICAL
        0x250C, # G+521  U+250C  SF010000,  BOX DRAWINGS LIGHT DOWN AND RIGHT
        0x2510, # G+522  U+2510  SF030000,  BOX DRAWINGS LIGHT DOWN AND LEFT
        0x2518, # G+523  U+2518  SF040000,  BOX DRAWINGS LIGHT UP AND LEFT
        0x2514, # G+524  U+2514  SF020000,  BOX DRAWINGS LIGHT UP AND RIGHT
        0x251C, # G+525  U+251C  SF080000,  BOX DRAWINGS LIGHT VERTICAL AND RIGHT
        0x252C, # G+526  U+252C  SF060000,  BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
        0x2524, # G+527  U+2524  SF090000,  BOX DRAWINGS LIGHT VERTICAL AND LEFT
        0x2534, # G+528  U+2534  SF070000,  BOX DRAWINGS LIGHT UP AND HORIZONTAL
        0x253C, # G+529  U+253C  SF050000,  BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
        0x2501, # G+530  U+2501  uni2501,  BOX DRAWINGS HEAVY HORIZONTAL
        0x2503, # G+531  U+2503  uni2503,  BOX DRAWINGS HEAVY VERTICAL
        0x250F, # G+532  U+250F  uni250F,  BOX DRAWINGS HEAVY DOWN AND RIGHT
        0x2513, # G+533  U+2513  uni2513,  BOX DRAWINGS HEAVY DOWN AND LEFT
        0x251B, # G+534  U+251B  uni251B,  BOX DRAWINGS HEAVY UP AND LEFT
        0x2517, # G+535  U+2517  uni2517,  BOX DRAWINGS HEAVY UP AND RIGHT
        0x2523, # G+536  U+2523  uni2523,  BOX DRAWINGS HEAVY VERTICAL AND RIGHT
        0x2533, # G+537  U+2533  uni2533,  BOX DRAWINGS HEAVY DOWN AND HORIZONTAL
        0x252B, # G+538  U+252B  uni252B,  BOX DRAWINGS HEAVY VERTICAL AND LEFT
        0x253B, # G+539  U+253B  uni253B,  BOX DRAWINGS HEAVY UP AND HORIZONTAL
        0x254B, # G+540  U+254B  uni254B,  BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL
        0x2520, # G+541  U+2520  uni2520,  BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT
        0x252F, # G+542  U+252F  uni252F,  BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY
        0x2528, # G+543  U+2528  uni2528,  BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT
        0x2537, # G+544  U+2537  uni2537,  BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY
        0x253F, # G+545  U+253F  uni253F,  BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY
        0x251D, # G+546  U+251D  uni251D,  BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY
        0x2530, # G+547  U+2530  uni2530,  BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT
        0x2525, # G+548  U+2525  uni2525,  BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY
        0x2538, # G+549  U+2538  uni2538,  BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT
        0x2542, # G+550  U+2542  uni2542,  BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT
        0x2512, # G+551  U+2512  uni2512,  BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT
        0x2511, # G+552  U+2511  uni2511,  BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY
        0x251A, # G+553  U+251A  uni251A,  BOX DRAWINGS UP HEAVY AND LEFT LIGHT
        0x2519, # G+554  U+2519  uni2519,  BOX DRAWINGS UP LIGHT AND LEFT HEAVY
        0x2516, # G+555  U+2516  uni2516,  BOX DRAWINGS UP HEAVY AND RIGHT LIGHT
        0x2515, # G+556  U+2515  uni2515,  BOX DRAWINGS UP LIGHT AND RIGHT HEAVY
        0x250E, # G+557  U+250E  uni250E,  BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT
        0x250D, # G+558  U+250D  uni250D,  BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY
        0x251E, # G+559  U+251E  uni251E,  BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT
        0x251F, # G+560  U+251F  uni251F,  BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT
        0x2521, # G+561  U+2521  uni2521,  BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY
        0x2522, # G+562  U+2522  uni2522,  BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY
        0x2526, # G+563  U+2526  uni2526,  BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT
        0x2527, # G+564  U+2527  uni2527,  BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT
        0x2529, # G+565  U+2529  uni2529,  BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY
        0x252A, # G+566  U+252A  uni252A,  BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY
        0x252D, # G+567  U+252D  uni252D,  BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT
        0x252E, # G+568  U+252E  uni252E,  BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT
        0x2531, # G+569  U+2531  uni2531,  BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY
        0x2532, # G+570  U+2532  uni2532,  BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY
        0x2535, # G+571  U+2535  uni2535,  BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT
        0x2536, # G+572  U+2536  uni2536,  BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT
        0x2539, # G+573  U+2539  uni2539,  BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY
        0x253A, # G+574  U+253A  uni253A,  BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY
        0x253D, # G+575  U+253D  uni253D,  BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT
        0x253E, # G+576  U+253E  uni253E,  BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT
        0x2540, # G+577  U+2540  uni2540,  BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT
        0x2541, # G+578  U+2541  uni2541,  BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT
        0x2543, # G+579  U+2543  uni2543,  BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT
        0x2544, # G+580  U+2544  uni2544,  BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT
        0x2545, # G+581  U+2545  uni2545,  BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT
        0x2546, # G+582  U+2546  uni2546,  BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT
        0x2547, # G+583  U+2547  uni2547,  BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY
        0x2548, # G+584  U+2548  uni2548,  BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY
        0x2549, # G+585  U+2549  uni2549,  BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY
        0x254A, # G+586  U+254A  uni254A,  BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY
        0x3395, # G+587  U+3395  mulsquare,  SQUARE MU L
        0x3396, # G+588  U+3396  mlsquare,  SQUARE ML
        0x3397, # G+589  U+3397  dlsquare,  SQUARE DL
        0x2113, # G+590  U+2113  afii61289,  SCRIPT SMALL L
        0x3398, # G+591  U+3398  klsquare,  SQUARE KL
        0x33C4, # G+592  U+33C4  squarecc,  SQUARE CC
        0x33A3, # G+593  U+33A3  mmcubedsquare,  SQUARE MM CUBED
        0x33A4, # G+594  U+33A4  cmcubedsquare,  SQUARE CM CUBED
        0x33A5, # G+595  U+33A5  mcubedsquare,  SQUARE M CUBED
        0x33A6, # G+596  U+33A6  kmcubedsquare,  SQUARE KM CUBED
        0x3399, # G+597  U+3399  fmsquare,  SQUARE FM
        0x339A, # G+598  U+339A  nmsquare,  SQUARE NM
        0x339B, # G+599  U+339B  mumsquare,  SQUARE MU M
        0x339C, # G+600  U+339C  squaremm,  SQUARE MM
        0x339D, # G+601  U+339D  squarecm,  SQUARE CM
        0x339E, # G+602  U+339E  squarekm,  SQUARE KM
        0x339F, # G+603  U+339F  mmsquaredsquare,  SQUARE MM SQUARED
        0x33A0, # G+604  U+33A0  cmsquaredsquare,  SQUARE CM SQUARED
        0x33A1, # G+605  U+33A1  squaremsquared,  SQUARE M SQUARED
        0x33A2, # G+606  U+33A2  kmsquaredsquare,  SQUARE KM SQUARED
        0x33CA, # G+607  U+33CA  hasquare,  SQUARE HA
        0x338D, # G+608  U+338D  mugsquare,  SQUARE MU G
        0x338E, # G+609  U+338E  squaremg,  SQUARE MG
        0x338F, # G+610  U+338F  squarekg,  SQUARE KG
        0x33CF, # G+611  U+33CF  ktsquare,  SQUARE KT
        0x3388, # G+612  U+3388  calsquare,  SQUARE CAL
        0x3389, # G+613  U+3389  kcalsquare,  SQUARE KCAL
        0x33C8, # G+614  U+33C8  dbsquare,  SQUARE DB
        0x33A7, # G+615  U+33A7  moverssquare,  SQUARE M OVER S
        0x33A8, # G+616  U+33A8  moverssquaredsquare,  SQUARE M OVER S SQUARED
        0x33B0, # G+617  U+33B0  pssquare,  SQUARE PS
        0x33B1, # G+618  U+33B1  nssquare,  SQUARE NS
        0x33B2, # G+619  U+33B2  mussquare,  SQUARE MU S
        0x33B3, # G+620  U+33B3  mssquare,  SQUARE MS
        0x33B4, # G+621  U+33B4  pvsquare,  SQUARE PV
        0x33B5, # G+622  U+33B5  nvsquare,  SQUARE NV
        0x33B6, # G+623  U+33B6  muvsquare,  SQUARE MU V
        0x33B7, # G+624  U+33B7  mvsquare,  SQUARE MV
        0x33B8, # G+625  U+33B8  kvsquare,  SQUARE KV
        0x33B9, # G+626  U+33B9  mvmegasquare,  SQUARE MV MEGA
        0x3380, # G+627  U+3380  paampssquare,  SQUARE PA AMPS
        0x3381, # G+628  U+3381  nasquare,  SQUARE NA
        0x3382, # G+629  U+3382  muasquare,  SQUARE MU A
        0x3383, # G+630  U+3383  masquare,  SQUARE MA
        0x3384, # G+631  U+3384  kasquare,  SQUARE KA
        0x33BA, # G+632  U+33BA  pwsquare,  SQUARE PW
        0x33BB, # G+633  U+33BB  nwsquare,  SQUARE NW
        0x33BC, # G+634  U+33BC  muwsquare,  SQUARE MU W
        0x33BD, # G+635  U+33BD  mwsquare,  SQUARE MW
        0x33BE, # G+636  U+33BE  kwsquare,  SQUARE KW
        0x33BF, # G+637  U+33BF  mwmegasquare,  SQUARE MW MEGA
        0x3390, # G+638  U+3390  Hzsquare,  SQUARE HZ
        0x3391, # G+639  U+3391  khzsquare,  SQUARE KHZ
        0x3392, # G+640  U+3392  mhzsquare,  SQUARE MHZ
        0x3393, # G+641  U+3393  ghzsquare,  SQUARE GHZ
        0x3394, # G+642  U+3394  thzsquare,  SQUARE THZ
        0x2126, # G+643  U+2126  Omega,  OHM SIGN
        0x33C0, # G+644  U+33C0  kohmsquare,  SQUARE K OHM
        0x33C1, # G+645  U+33C1  mohmsquare,  SQUARE M OHM
        0x338A, # G+646  U+338A  pfsquare,  SQUARE PF
        0x338B, # G+647  U+338B  nfsquare,  SQUARE NF
        0x338C, # G+648  U+338C  mufsquare,  SQUARE MU F
        0x33D6, # G+649  U+33D6  molsquare,  SQUARE MOL
        0x33C5, # G+650  U+33C5  cdsquare,  SQUARE CD
        0x33AD, # G+651  U+33AD  radsquare,  SQUARE RAD
        0x33AE, # G+652  U+33AE  radoverssquare,  SQUARE RAD OVER S
        0x33AF, # G+653  U+33AF  radoverssquaredsquare,  SQUARE RAD OVER S SQUARED
        0x33DB, # G+654  U+33DB  srsquare,  SQUARE SR
        0x33A9, # G+655  U+33A9  pasquare,  SQUARE PA
        0x33AA, # G+656  U+33AA  kpasquare,  SQUARE KPA
        0x33AB, # G+657  U+33AB  mpasquare,  SQUARE MPA
        0x33AC, # G+658  U+33AC  gpasquare,  SQUARE GPA
        0x33DD, # G+659  U+33DD  wbsquare,  SQUARE WB
        0x33D0, # G+660  U+33D0  lmsquare,  SQUARE LM
        0x33D3, # G+661  U+33D3  lxsquare,  SQUARE LX
        0x33C3, # G+662  U+33C3  bqsquare,  SQUARE BQ
        0x33C9, # G+663  U+33C9  gysquare,  SQUARE GY
        0x33DC, # G+664  U+33DC  svsquare,  SQUARE SV
        0x33C6, # G+665  U+33C6  coverkgsquare,  SQUARE C OVER KG
        0x00C6, # G+666  U+00C6  AE,  LATIN CAPITAL LETTER AE
        0x00D0, # G+667  U+00D0  Eth,  LATIN CAPITAL LETTER ETH
        0x00AA, # G+668  U+00AA  ordfeminine,  FEMININE ORDINAL INDICATOR
        0x0126, # G+669  U+0126  Hbar,  LATIN CAPITAL LETTER H WITH STROKE
        0x0132, # G+670  U+0132  IJ,  LATIN CAPITAL LIGATURE IJ
        0x013F, # G+671  U+013F  Ldot,  LATIN CAPITAL LETTER L WITH MIDDLE DOT
        0x0141, # G+672  U+0141  Lslash,  LATIN CAPITAL LETTER L WITH STROKE
        0x00D8, # G+673  U+00D8  Oslash,  LATIN CAPITAL LETTER O WITH STROKE
        0x0152, # G+674  U+0152  OE,  LATIN CAPITAL LIGATURE OE
        0x00BA, # G+675  U+00BA  ordmasculine,  MASCULINE ORDINAL INDICATOR
        0x00DE, # G+676  U+00DE  Thorn,  LATIN CAPITAL LETTER THORN
        0x0166, # G+677  U+0166  Tbar,  LATIN CAPITAL LETTER T WITH STROKE
        0x014A, # G+678  U+014A  Eng,  LATIN CAPITAL LETTER ENG
        0x3260, # G+679  U+3260  kiyeokcirclekorean,  CIRCLED HANGUL KIYEOK
        0x3261, # G+680  U+3261  nieuncirclekorean,  CIRCLED HANGUL NIEUN
        0x3262, # G+681  U+3262  tikeutcirclekorean,  CIRCLED HANGUL TIKEUT
        0x3263, # G+682  U+3263  rieulcirclekorean,  CIRCLED HANGUL RIEUL
        0x3264, # G+683  U+3264  mieumcirclekorean,  CIRCLED HANGUL MIEUM
        0x3265, # G+684  U+3265  pieupcirclekorean,  CIRCLED HANGUL PIEUP
        0x3266, # G+685  U+3266  sioscirclekorean,  CIRCLED HANGUL SIOS
        0x3267, # G+686  U+3267  ieungcirclekorean,  CIRCLED HANGUL IEUNG
        0x3268, # G+687  U+3268  cieuccirclekorean,  CIRCLED HANGUL CIEUC
        0x3269, # G+688  U+3269  chieuchcirclekorean,  CIRCLED HANGUL CHIEUCH
        0x326A, # G+689  U+326A  khieukhcirclekorean,  CIRCLED HANGUL KHIEUKH
        0x326B, # G+690  U+326B  thieuthcirclekorean,  CIRCLED HANGUL THIEUTH
        0x326C, # G+691  U+326C  phieuphcirclekorean,  CIRCLED HANGUL PHIEUPH
        0x326D, # G+692  U+326D  hieuhcirclekorean,  CIRCLED HANGUL HIEUH
        0x326E, # G+693  U+326E  kiyeokacirclekorean,  CIRCLED HANGUL KIYEOK A
        0x326F, # G+694  U+326F  nieunacirclekorean,  CIRCLED HANGUL NIEUN A
        0x3270, # G+695  U+3270  tikeutacirclekorean,  CIRCLED HANGUL TIKEUT A
        0x3271, # G+696  U+3271  rieulacirclekorean,  CIRCLED HANGUL RIEUL A
        0x3272, # G+697  U+3272  mieumacirclekorean,  CIRCLED HANGUL MIEUM A
        0x3273, # G+698  U+3273  pieupacirclekorean,  CIRCLED HANGUL PIEUP A
        0x3274, # G+699  U+3274  siosacirclekorean,  CIRCLED HANGUL SIOS A
        0x3275, # G+700  U+3275  ieungacirclekorean,  CIRCLED HANGUL IEUNG A
        0x3276, # G+701  U+3276  cieucacirclekorean,  CIRCLED HANGUL CIEUC A
        0x3277, # G+702  U+3277  chieuchacirclekorean,  CIRCLED HANGUL CHIEUCH A
        0x3278, # G+703  U+3278  khieukhacirclekorean,  CIRCLED HANGUL KHIEUKH A
        0x3279, # G+704  U+3279  thieuthacirclekorean,  CIRCLED HANGUL THIEUTH A
        0x327A, # G+705  U+327A  phieuphacirclekorean,  CIRCLED HANGUL PHIEUPH A
        0x327B, # G+706  U+327B  hieuhacirclekorean,  CIRCLED HANGUL HIEUH A
        0x24D0, # G+707  U+24D0  acircle,  CIRCLED LATIN SMALL LETTER A
        0x24D1, # G+708  U+24D1  bcircle,  CIRCLED LATIN SMALL LETTER B
        0x24D2, # G+709  U+24D2  ccircle,  CIRCLED LATIN SMALL LETTER C
        0x24D3, # G+710  U+24D3  dcircle,  CIRCLED LATIN SMALL LETTER D
        0x24D4, # G+711  U+24D4  ecircle,  CIRCLED LATIN SMALL LETTER E
        0x24D5, # G+712  U+24D5  fcircle,  CIRCLED LATIN SMALL LETTER F
        0x24D6, # G+713  U+24D6  gcircle,  CIRCLED LATIN SMALL LETTER G
        0x24D7, # G+714  U+24D7  hcircle,  CIRCLED LATIN SMALL LETTER H
        0x24D8, # G+715  U+24D8  icircle,  CIRCLED LATIN SMALL LETTER I
        0x24D9, # G+716  U+24D9  jcircle,  CIRCLED LATIN SMALL LETTER J
        0x24DA, # G+717  U+24DA  kcircle,  CIRCLED LATIN SMALL LETTER K
        0x24DB, # G+718  U+24DB  lcircle,  CIRCLED LATIN SMALL LETTER L
        0x24DC, # G+719  U+24DC  mcircle,  CIRCLED LATIN SMALL LETTER M
        0x24DD, # G+720  U+24DD  ncircle,  CIRCLED LATIN SMALL LETTER N
        0x24DE, # G+721  U+24DE  ocircle,  CIRCLED LATIN SMALL LETTER O
        0x24DF, # G+722  U+24DF  pcircle,  CIRCLED LATIN SMALL LETTER P
        0x24E0, # G+723  U+24E0  qcircle,  CIRCLED LATIN SMALL LETTER Q
        0x24E1, # G+724  U+24E1  rcircle,  CIRCLED LATIN SMALL LETTER R
        0x24E2, # G+725  U+24E2  scircle,  CIRCLED LATIN SMALL LETTER S
        0x24E3, # G+726  U+24E3  tcircle,  CIRCLED LATIN SMALL LETTER T
        0x24E4, # G+727  U+24E4  ucircle,  CIRCLED LATIN SMALL LETTER U
        0x24E5, # G+728  U+24E5  vcircle,  CIRCLED LATIN SMALL LETTER V
        0x24E6, # G+729  U+24E6  wcircle,  CIRCLED LATIN SMALL LETTER W
        0x24E7, # G+730  U+24E7  xcircle,  CIRCLED LATIN SMALL LETTER X
        0x24E8, # G+731  U+24E8  ycircle,  CIRCLED LATIN SMALL LETTER Y
        0x24E9, # G+732  U+24E9  zcircle,  CIRCLED LATIN SMALL LETTER Z
        0x2460, # G+733  U+2460  onecircle,  CIRCLED DIGIT ONE
        0x2461, # G+734  U+2461  twocircle,  CIRCLED DIGIT TWO
        0x2462, # G+735  U+2462  threecircle,  CIRCLED DIGIT THREE
        0x2463, # G+736  U+2463  fourcircle,  CIRCLED DIGIT FOUR
        0x2464, # G+737  U+2464  fivecircle,  CIRCLED DIGIT FIVE
        0x2465, # G+738  U+2465  sixcircle,  CIRCLED DIGIT SIX
        0x2466, # G+739  U+2466  sevencircle,  CIRCLED DIGIT SEVEN
        0x2467, # G+740  U+2467  eightcircle,  CIRCLED DIGIT EIGHT
        0x2468, # G+741  U+2468  ninecircle,  CIRCLED DIGIT NINE
        0x2469, # G+742  U+2469  tencircle,  CIRCLED NUMBER TEN
        0x246A, # G+743  U+246A  elevencircle,  CIRCLED NUMBER ELEVEN
        0x246B, # G+744  U+246B  twelvecircle,  CIRCLED NUMBER TWELVE
        0x246C, # G+745  U+246C  thirteencircle,  CIRCLED NUMBER THIRTEEN
        0x246D, # G+746  U+246D  fourteencircle,  CIRCLED NUMBER FOURTEEN
        0x246E, # G+747  U+246E  fifteencircle,  CIRCLED NUMBER FIFTEEN
        0x00BD, # G+748  U+00BD  onehalf,  VULGAR FRACTION ONE HALF
        0x2153, # G+749  U+2153  onethird,  VULGAR FRACTION ONE THIRD
        0x2154, # G+750  U+2154  twothirds,  VULGAR FRACTION TWO THIRDS
        0x00BC, # G+751  U+00BC  onequarter,  VULGAR FRACTION ONE QUARTER
        0x00BE, # G+752  U+00BE  threequarters,  VULGAR FRACTION THREE QUARTERS
        0x215B, # G+753  U+215B  oneeighth,  VULGAR FRACTION ONE EIGHTH
        0x215C, # G+754  U+215C  threeeighths,  VULGAR FRACTION THREE EIGHTHS
        0x215D, # G+755  U+215D  fiveeighths,  VULGAR FRACTION FIVE EIGHTHS
        0x215E, # G+756  U+215E  seveneighths,  VULGAR FRACTION SEVEN EIGHTHS
        0x00E6, # G+757  U+00E6  ae,  LATIN SMALL LETTER AE
        0x0111, # G+758  U+0111  dcroat,  LATIN SMALL LETTER D WITH STROKE
        0x00F0, # G+759  U+00F0  eth,  LATIN SMALL LETTER ETH
        0x0127, # G+760  U+0127  hbar,  LATIN SMALL LETTER H WITH STROKE
        0x0131, # G+761  U+0131  dotlessi,  LATIN SMALL LETTER DOTLESS I
        0x0133, # G+762  U+0133  ij,  LATIN SMALL LIGATURE IJ
        0x0138, # G+763  U+0138  kgreenlandic,  LATIN SMALL LETTER KRA
        0x0140, # G+764  U+0140  ldot,  LATIN SMALL LETTER L WITH MIDDLE DOT
        0x0142, # G+765  U+0142  lslash,  LATIN SMALL LETTER L WITH STROKE
        0x00F8, # G+766  U+00F8  oslash,  LATIN SMALL LETTER O WITH STROKE
        0x0153, # G+767  U+0153  oe,  LATIN SMALL LIGATURE OE
        0x00DF, # G+768  U+00DF  germandbls,  LATIN SMALL LETTER SHARP S
        0x00FE, # G+769  U+00FE  thorn,  LATIN SMALL LETTER THORN
        0x0167, # G+770  U+0167  tbar,  LATIN SMALL LETTER T WITH STROKE
        0x014B, # G+771  U+014B  eng,  LATIN SMALL LETTER ENG
        0x0149, # G+772  U+0149  napostrophe,  LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
        0x3200, # G+773  U+3200  kiyeokparenkorean,  PARENTHESIZED HANGUL KIYEOK
        0x3201, # G+774  U+3201  nieunparenkorean,  PARENTHESIZED HANGUL NIEUN
        0x3202, # G+775  U+3202  tikeutparenkorean,  PARENTHESIZED HANGUL TIKEUT
        0x3203, # G+776  U+3203  rieulparenkorean,  PARENTHESIZED HANGUL RIEUL
        0x3204, # G+777  U+3204  mieumparenkorean,  PARENTHESIZED HANGUL MIEUM
        0x3205, # G+778  U+3205  pieupparenkorean,  PARENTHESIZED HANGUL PIEUP
        0x3206, # G+779  U+3206  siosparenkorean,  PARENTHESIZED HANGUL SIOS
        0x3207, # G+780  U+3207  ieungparenkorean,  PARENTHESIZED HANGUL IEUNG
        0x3208, # G+781  U+3208  cieucparenkorean,  PARENTHESIZED HANGUL CIEUC
        0x3209, # G+782  U+3209  chieuchparenkorean,  PARENTHESIZED HANGUL CHIEUCH
        0x320A, # G+783  U+320A  khieukhparenkorean,  PARENTHESIZED HANGUL KHIEUKH
        0x320B, # G+784  U+320B  thieuthparenkorean,  PARENTHESIZED HANGUL THIEUTH
        0x320C, # G+785  U+320C  phieuphparenkorean,  PARENTHESIZED HANGUL PHIEUPH
        0x320D, # G+786  U+320D  hieuhparenkorean,  PARENTHESIZED HANGUL HIEUH
        0x320E, # G+787  U+320E  kiyeokaparenkorean,  PARENTHESIZED HANGUL KIYEOK A
        0x320F, # G+788  U+320F  nieunaparenkorean,  PARENTHESIZED HANGUL NIEUN A
        0x3210, # G+789  U+3210  tikeutaparenkorean,  PARENTHESIZED HANGUL TIKEUT A
        0x3211, # G+790  U+3211  rieulaparenkorean,  PARENTHESIZED HANGUL RIEUL A
        0x3212, # G+791  U+3212  mieumaparenkorean,  PARENTHESIZED HANGUL MIEUM A
        0x3213, # G+792  U+3213  pieupaparenkorean,  PARENTHESIZED HANGUL PIEUP A
        0x3214, # G+793  U+3214  siosaparenkorean,  PARENTHESIZED HANGUL SIOS A
        0x3215, # G+794  U+3215  ieungaparenkorean,  PARENTHESIZED HANGUL IEUNG A
        0x3216, # G+795  U+3216  cieucaparenkorean,  PARENTHESIZED HANGUL CIEUC A
        0x3217, # G+796  U+3217  chieuchaparenkorean,  PARENTHESIZED HANGUL CHIEUCH A
        0x3218, # G+797  U+3218  khieukhaparenkorean,  PARENTHESIZED HANGUL KHIEUKH A
        0x3219, # G+798  U+3219  thieuthaparenkorean,  PARENTHESIZED HANGUL THIEUTH A
        0x321A, # G+799  U+321A  phieuphaparenkorean,  PARENTHESIZED HANGUL PHIEUPH A
        0x321B, # G+800  U+321B  hieuhaparenkorean,  PARENTHESIZED HANGUL HIEUH A
        0x249C, # G+801  U+249C  aparen,  PARENTHESIZED LATIN SMALL LETTER A
        0x249D, # G+802  U+249D  bparen,  PARENTHESIZED LATIN SMALL LETTER B
        0x249E, # G+803  U+249E  cparen,  PARENTHESIZED LATIN SMALL LETTER C
        0x249F, # G+804  U+249F  dparen,  PARENTHESIZED LATIN SMALL LETTER D
        0x24A0, # G+805  U+24A0  eparen,  PARENTHESIZED LATIN SMALL LETTER E
        0x24A1, # G+806  U+24A1  fparen,  PARENTHESIZED LATIN SMALL LETTER F
        0x24A2, # G+807  U+24A2  gparen,  PARENTHESIZED LATIN SMALL LETTER G
        0x24A3, # G+808  U+24A3  hparen,  PARENTHESIZED LATIN SMALL LETTER H
        0x24A4, # G+809  U+24A4  iparen,  PARENTHESIZED LATIN SMALL LETTER I
        0x24A5, # G+810  U+24A5  jparen,  PARENTHESIZED LATIN SMALL LETTER J
        0x24A6, # G+811  U+24A6  kparen,  PARENTHESIZED LATIN SMALL LETTER K
        0x24A7, # G+812  U+24A7  lparen,  PARENTHESIZED LATIN SMALL LETTER L
        0x24A8, # G+813  U+24A8  mparen,  PARENTHESIZED LATIN SMALL LETTER M
        0x24A9, # G+814  U+24A9  nparen,  PARENTHESIZED LATIN SMALL LETTER N
        0x24AA, # G+815  U+24AA  oparen,  PARENTHESIZED LATIN SMALL LETTER O
        0x24AB, # G+816  U+24AB  pparen,  PARENTHESIZED LATIN SMALL LETTER P
        0x24AC, # G+817  U+24AC  qparen,  PARENTHESIZED LATIN SMALL LETTER Q
        0x24AD, # G+818  U+24AD  rparen,  PARENTHESIZED LATIN SMALL LETTER R
        0x24AE, # G+819  U+24AE  sparen,  PARENTHESIZED LATIN SMALL LETTER S
        0x24AF, # G+820  U+24AF  tparen,  PARENTHESIZED LATIN SMALL LETTER T
        0x24B0, # G+821  U+24B0  uparen,  PARENTHESIZED LATIN SMALL LETTER U
        0x24B1, # G+822  U+24B1  vparen,  PARENTHESIZED LATIN SMALL LETTER V
        0x24B2, # G+823  U+24B2  wparen,  PARENTHESIZED LATIN SMALL LETTER W
        0x24B3, # G+824  U+24B3  xparen,  PARENTHESIZED LATIN SMALL LETTER X
        0x24B4, # G+825  U+24B4  yparen,  PARENTHESIZED LATIN SMALL LETTER Y
        0x24B5, # G+826  U+24B5  zparen,  PARENTHESIZED LATIN SMALL LETTER Z
        0x2474, # G+827  U+2474  oneparen,  PARENTHESIZED DIGIT ONE
        0x2475, # G+828  U+2475  twoparen,  PARENTHESIZED DIGIT TWO
        0x2476, # G+829  U+2476  threeparen,  PARENTHESIZED DIGIT THREE
        0x2477, # G+830  U+2477  fourparen,  PARENTHESIZED DIGIT FOUR
        0x2478, # G+831  U+2478  fiveparen,  PARENTHESIZED DIGIT FIVE
        0x2479, # G+832  U+2479  sixparen,  PARENTHESIZED DIGIT SIX
        0x247A, # G+833  U+247A  sevenparen,  PARENTHESIZED DIGIT SEVEN
        0x247B, # G+834  U+247B  eightparen,  PARENTHESIZED DIGIT EIGHT
        0x247C, # G+835  U+247C  nineparen,  PARENTHESIZED DIGIT NINE
        0x247D, # G+836  U+247D  tenparen,  PARENTHESIZED NUMBER TEN
        0x247E, # G+837  U+247E  elevenparen,  PARENTHESIZED NUMBER ELEVEN
        0x247F, # G+838  U+247F  twelveparen,  PARENTHESIZED NUMBER TWELVE
        0x2480, # G+839  U+2480  thirteenparen,  PARENTHESIZED NUMBER THIRTEEN
        0x2481, # G+840  U+2481  fourteenparen,  PARENTHESIZED NUMBER FOURTEEN
        0x2482, # G+841  U+2482  fifteenparen,  PARENTHESIZED NUMBER FIFTEEN
        0x00B9, # G+842  U+00B9  onesuperior,  SUPERSCRIPT ONE
        0x00B2, # G+843  U+00B2  twosuperior,  SUPERSCRIPT TWO
        0x00B3, # G+844  U+00B3  threesuperior,  SUPERSCRIPT THREE
        0x2074, # G+845  U+2074  foursuperior,  SUPERSCRIPT FOUR
        0x207F, # G+846  U+207F  nsuperior,  SUPERSCRIPT LATIN SMALL LETTER N
        0x2081, # G+847  U+2081  oneinferior,  SUBSCRIPT ONE
        0x2082, # G+848  U+2082  twoinferior,  SUBSCRIPT TWO
        0x2083, # G+849  U+2083  threeinferior,  SUBSCRIPT THREE
        0x2084, # G+850  U+2084  fourinferior,  SUBSCRIPT FOUR
        0x3041, # G+851  U+3041  asmallhiragana,  HIRAGANA LETTER SMALL A
        0x3042, # G+852  U+3042  ahiragana,  HIRAGANA LETTER A
        0x3043, # G+853  U+3043  ismallhiragana,  HIRAGANA LETTER SMALL I
        0x3044, # G+854  U+3044  ihiragana,  HIRAGANA LETTER I
        0x3045, # G+855  U+3045  usmallhiragana,  HIRAGANA LETTER SMALL U
        0x3046, # G+856  U+3046  uhiragana,  HIRAGANA LETTER U
        0x3047, # G+857  U+3047  esmallhiragana,  HIRAGANA LETTER SMALL E
        0x3048, # G+858  U+3048  ehiragana,  HIRAGANA LETTER E
        0x3049, # G+859  U+3049  osmallhiragana,  HIRAGANA LETTER SMALL O
        0x304A, # G+860  U+304A  ohiragana,  HIRAGANA LETTER O
        0x304B, # G+861  U+304B  kahiragana,  HIRAGANA LETTER KA
        0x304C, # G+862  U+304C  gahiragana,  HIRAGANA LETTER GA
        0x304D, # G+863  U+304D  kihiragana,  HIRAGANA LETTER KI
        0x304E, # G+864  U+304E  gihiragana,  HIRAGANA LETTER GI
        0x304F, # G+865  U+304F  kuhiragana,  HIRAGANA LETTER KU
        0x3050, # G+866  U+3050  guhiragana,  HIRAGANA LETTER GU
        0x3051, # G+867  U+3051  kehiragana,  HIRAGANA LETTER KE
        0x3052, # G+868  U+3052  gehiragana,  HIRAGANA LETTER GE
        0x3053, # G+869  U+3053  kohiragana,  HIRAGANA LETTER KO
        0x3054, # G+870  U+3054  gohiragana,  HIRAGANA LETTER GO
        0x3055, # G+871  U+3055  sahiragana,  HIRAGANA LETTER SA
        0x3056, # G+872  U+3056  zahiragana,  HIRAGANA LETTER ZA
        0x3057, # G+873  U+3057  sihiragana,  HIRAGANA LETTER SI
        0x3058, # G+874  U+3058  zihiragana,  HIRAGANA LETTER ZI
        0x3059, # G+875  U+3059  suhiragana,  HIRAGANA LETTER SU
        0x305A, # G+876  U+305A  zuhiragana,  HIRAGANA LETTER ZU
        0x305B, # G+877  U+305B  sehiragana,  HIRAGANA LETTER SE
        0x305C, # G+878  U+305C  zehiragana,  HIRAGANA LETTER ZE
        0x305D, # G+879  U+305D  sohiragana,  HIRAGANA LETTER SO
        0x305E, # G+880  U+305E  zohiragana,  HIRAGANA LETTER ZO
        0x305F, # G+881  U+305F  tahiragana,  HIRAGANA LETTER TA
        0x3060, # G+882  U+3060  dahiragana,  HIRAGANA LETTER DA
        0x3061, # G+883  U+3061  tihiragana,  HIRAGANA LETTER TI
        0x3062, # G+884  U+3062  dihiragana,  HIRAGANA LETTER DI
        0x3063, # G+885  U+3063  tusmallhiragana,  HIRAGANA LETTER SMALL TU
        0x3064, # G+886  U+3064  tuhiragana,  HIRAGANA LETTER TU
        0x3065, # G+887  U+3065  duhiragana,  HIRAGANA LETTER DU
        0x3066, # G+888  U+3066  tehiragana,  HIRAGANA LETTER TE
        0x3067, # G+889  U+3067  dehiragana,  HIRAGANA LETTER DE
        0x3068, # G+890  U+3068  tohiragana,  HIRAGANA LETTER TO
        0x3069, # G+891  U+3069  dohiragana,  HIRAGANA LETTER DO
        0x306A, # G+892  U+306A  nahiragana,  HIRAGANA LETTER NA
        0x306B, # G+893  U+306B  nihiragana,  HIRAGANA LETTER NI
        0x306C, # G+894  U+306C  nuhiragana,  HIRAGANA LETTER NU
        0x306D, # G+895  U+306D  nehiragana,  HIRAGANA LETTER NE
        0x306E, # G+896  U+306E  nohiragana,  HIRAGANA LETTER NO
        0x306F, # G+897  U+306F  hahiragana,  HIRAGANA LETTER HA
        0x3070, # G+898  U+3070  bahiragana,  HIRAGANA LETTER BA
        0x3071, # G+899  U+3071  pahiragana,  HIRAGANA LETTER PA
        0x3072, # G+900  U+3072  hihiragana,  HIRAGANA LETTER HI
        0x3073, # G+901  U+3073  bihiragana,  HIRAGANA LETTER BI
        0x3074, # G+902  U+3074  pihiragana,  HIRAGANA LETTER PI
        0x3075, # G+903  U+3075  huhiragana,  HIRAGANA LETTER HU
        0x3076, # G+904  U+3076  buhiragana,  HIRAGANA LETTER BU
        0x3077, # G+905  U+3077  puhiragana,  HIRAGANA LETTER PU
        0x3078, # G+906  U+3078  hehiragana,  HIRAGANA LETTER HE
        0x3079, # G+907  U+3079  behiragana,  HIRAGANA LETTER BE
        0x307A, # G+908  U+307A  pehiragana,  HIRAGANA LETTER PE
        0x307B, # G+909  U+307B  hohiragana,  HIRAGANA LETTER HO
        0x307C, # G+910  U+307C  bohiragana,  HIRAGANA LETTER BO
        0x307D, # G+911  U+307D  pohiragana,  HIRAGANA LETTER PO
        0x307E, # G+912  U+307E  mahiragana,  HIRAGANA LETTER MA
        0x307F, # G+913  U+307F  mihiragana,  HIRAGANA LETTER MI
        0x3080, # G+914  U+3080  muhiragana,  HIRAGANA LETTER MU
        0x3081, # G+915  U+3081  mehiragana,  HIRAGANA LETTER ME
        0x3082, # G+916  U+3082  mohiragana,  HIRAGANA LETTER MO
        0x3083, # G+917  U+3083  yasmallhiragana,  HIRAGANA LETTER SMALL YA
        0x3084, # G+918  U+3084  yahiragana,  HIRAGANA LETTER YA
        0x3085, # G+919  U+3085  yusmallhiragana,  HIRAGANA LETTER SMALL YU
        0x3086, # G+920  U+3086  yuhiragana,  HIRAGANA LETTER YU
        0x3087, # G+921  U+3087  yosmallhiragana,  HIRAGANA LETTER SMALL YO
        0x3088, # G+922  U+3088  yohiragana,  HIRAGANA LETTER YO
        0x3089, # G+923  U+3089  rahiragana,  HIRAGANA LETTER RA
        0x308A, # G+924  U+308A  rihiragana,  HIRAGANA LETTER RI
        0x308B, # G+925  U+308B  ruhiragana,  HIRAGANA LETTER RU
        0x308C, # G+926  U+308C  rehiragana,  HIRAGANA LETTER RE
        0x308D, # G+927  U+308D  rohiragana,  HIRAGANA LETTER RO
        0x308E, # G+928  U+308E  wasmallhiragana,  HIRAGANA LETTER SMALL WA
        0x308F, # G+929  U+308F  wahiragana,  HIRAGANA LETTER WA
        0x3090, # G+930  U+3090  wihiragana,  HIRAGANA LETTER WI
        0x3091, # G+931  U+3091  wehiragana,  HIRAGANA LETTER WE
        0x3092, # G+932  U+3092  wohiragana,  HIRAGANA LETTER WO
        0x3093, # G+933  U+3093  nhiragana,  HIRAGANA LETTER N
        0x30A1, # G+934  U+30A1  asmallkatakana,  KATAKANA LETTER SMALL A
        0x30A2, # G+935  U+30A2  akatakana,  KATAKANA LETTER A
        0x30A3, # G+936  U+30A3  ismallkatakana,  KATAKANA LETTER SMALL I
        0x30A4, # G+937  U+30A4  ikatakana,  KATAKANA LETTER I
        0x30A5, # G+938  U+30A5  usmallkatakana,  KATAKANA LETTER SMALL U
        0x30A6, # G+939  U+30A6  ukatakana,  KATAKANA LETTER U
        0x30A7, # G+940  U+30A7  esmallkatakana,  KATAKANA LETTER SMALL E
        0x30A8, # G+941  U+30A8  ekatakana,  KATAKANA LETTER E
        0x30A9, # G+942  U+30A9  osmallkatakana,  KATAKANA LETTER SMALL O
        0x30AA, # G+943  U+30AA  okatakana,  KATAKANA LETTER O
        0x30AB, # G+944  U+30AB  kakatakana,  KATAKANA LETTER KA
        0x30AC, # G+945  U+30AC  gakatakana,  KATAKANA LETTER GA
        0x30AD, # G+946  U+30AD  kikatakana,  KATAKANA LETTER KI
        0x30AE, # G+947  U+30AE  gikatakana,  KATAKANA LETTER GI
        0x30AF, # G+948  U+30AF  kukatakana,  KATAKANA LETTER KU
        0x30B0, # G+949  U+30B0  gukatakana,  KATAKANA LETTER GU
        0x30B1, # G+950  U+30B1  kekatakana,  KATAKANA LETTER KE
        0x30B2, # G+951  U+30B2  gekatakana,  KATAKANA LETTER GE
        0x30B3, # G+952  U+30B3  kokatakana,  KATAKANA LETTER KO
        0x30B4, # G+953  U+30B4  gokatakana,  KATAKANA LETTER GO
        0x30B5, # G+954  U+30B5  sakatakana,  KATAKANA LETTER SA
        0x30B6, # G+955  U+30B6  zakatakana,  KATAKANA LETTER ZA
        0x30B7, # G+956  U+30B7  sikatakana,  KATAKANA LETTER SI
        0x30B8, # G+957  U+30B8  zikatakana,  KATAKANA LETTER ZI
        0x30B9, # G+958  U+30B9  sukatakana,  KATAKANA LETTER SU
        0x30BA, # G+959  U+30BA  zukatakana,  KATAKANA LETTER ZU
        0x30BB, # G+960  U+30BB  sekatakana,  KATAKANA LETTER SE
        0x30BC, # G+961  U+30BC  zekatakana,  KATAKANA LETTER ZE
        0x30BD, # G+962  U+30BD  sokatakana,  KATAKANA LETTER SO
        0x30BE, # G+963  U+30BE  zokatakana,  KATAKANA LETTER ZO
        0x30BF, # G+964  U+30BF  takatakana,  KATAKANA LETTER TA
        0x30C0, # G+965  U+30C0  dakatakana,  KATAKANA LETTER DA
        0x30C1, # G+966  U+30C1  tikatakana,  KATAKANA LETTER TI
        0x30C2, # G+967  U+30C2  dikatakana,  KATAKANA LETTER DI
        0x30C3, # G+968  U+30C3  tusmallkatakana,  KATAKANA LETTER SMALL TU
        0x30C4, # G+969  U+30C4  tukatakana,  KATAKANA LETTER TU
        0x30C5, # G+970  U+30C5  dukatakana,  KATAKANA LETTER DU
        0x30C6, # G+971  U+30C6  tekatakana,  KATAKANA LETTER TE
        0x30C7, # G+972  U+30C7  dekatakana,  KATAKANA LETTER DE
        0x30C8, # G+973  U+30C8  tokatakana,  KATAKANA LETTER TO
        0x30C9, # G+974  U+30C9  dokatakana,  KATAKANA LETTER DO
        0x30CA, # G+975  U+30CA  nakatakana,  KATAKANA LETTER NA
        0x30CB, # G+976  U+30CB  nikatakana,  KATAKANA LETTER NI
        0x30CC, # G+977  U+30CC  nukatakana,  KATAKANA LETTER NU
        0x30CD, # G+978  U+30CD  nekatakana,  KATAKANA LETTER NE
        0x30CE, # G+979  U+30CE  nokatakana,  KATAKANA LETTER NO
        0x30CF, # G+980  U+30CF  hakatakana,  KATAKANA LETTER HA
        0x30D0, # G+981  U+30D0  bakatakana,  KATAKANA LETTER BA
        0x30D1, # G+982  U+30D1  pakatakana,  KATAKANA LETTER PA
        0x30D2, # G+983  U+30D2  hikatakana,  KATAKANA LETTER HI
        0x30D3, # G+984  U+30D3  bikatakana,  KATAKANA LETTER BI
        0x30D4, # G+985  U+30D4  pikatakana,  KATAKANA LETTER PI
        0x30D5, # G+986  U+30D5  hukatakana,  KATAKANA LETTER HU
        0x30D6, # G+987  U+30D6  bukatakana,  KATAKANA LETTER BU
        0x30D7, # G+988  U+30D7  pukatakana,  KATAKANA LETTER PU
        0x30D8, # G+989  U+30D8  hekatakana,  KATAKANA LETTER HE
        0x30D9, # G+990  U+30D9  bekatakana,  KATAKANA LETTER BE
        0x30DA, # G+991  U+30DA  pekatakana,  KATAKANA LETTER PE
        0x30DB, # G+992  U+30DB  hokatakana,  KATAKANA LETTER HO
        0x30DC, # G+993  U+30DC  bokatakana,  KATAKANA LETTER BO
        0x30DD, # G+994  U+30DD  pokatakana,  KATAKANA LETTER PO
        0x30DE, # G+995  U+30DE  makatakana,  KATAKANA LETTER MA
        0x30DF, # G+996  U+30DF  mikatakana,  KATAKANA LETTER MI
        0x30E0, # G+997  U+30E0  mukatakana,  KATAKANA LETTER MU
        0x30E1, # G+998  U+30E1  mekatakana,  KATAKANA LETTER ME
        0x30E2, # G+999  U+30E2  mokatakana,  KATAKANA LETTER MO
        0x30E3, # G+1000  U+30E3  yasmallkatakana,  KATAKANA LETTER SMALL YA
        0x30E4, # G+1001  U+30E4  yakatakana,  KATAKANA LETTER YA
        0x30E5, # G+1002  U+30E5  yusmallkatakana,  KATAKANA LETTER SMALL YU
        0x30E6, # G+1003  U+30E6  yukatakana,  KATAKANA LETTER YU
        0x30E7, # G+1004  U+30E7  yosmallkatakana,  KATAKANA LETTER SMALL YO
        0x30E8, # G+1005  U+30E8  yokatakana,  KATAKANA LETTER YO
        0x30E9, # G+1006  U+30E9  rakatakana,  KATAKANA LETTER RA
        0x30EA, # G+1007  U+30EA  rikatakana,  KATAKANA LETTER RI
        0x30EB, # G+1008  U+30EB  rukatakana,  KATAKANA LETTER RU
        0x30EC, # G+1009  U+30EC  rekatakana,  KATAKANA LETTER RE
        0x30ED, # G+1010  U+30ED  rokatakana,  KATAKANA LETTER RO
        0x30EE, # G+1011  U+30EE  wasmallkatakana,  KATAKANA LETTER SMALL WA
        0x30EF, # G+1012  U+30EF  wakatakana,  KATAKANA LETTER WA
        0x30F0, # G+1013  U+30F0  wikatakana,  KATAKANA LETTER WI
        0x30F1, # G+1014  U+30F1  wekatakana,  KATAKANA LETTER WE
        0x30F2, # G+1015  U+30F2  wokatakana,  KATAKANA LETTER WO
        0x30F3, # G+1016  U+30F3  nkatakana,  KATAKANA LETTER N
        0x30F4, # G+1017  U+30F4  vukatakana,  KATAKANA LETTER VU
        0x30F5, # G+1018  U+30F5  kasmallkatakana,  KATAKANA LETTER SMALL KA
        0x30F6, # G+1019  U+30F6  kesmallkatakana,  KATAKANA LETTER SMALL KE
        0x0410, # G+1020  U+0410  afii10017,  CYRILLIC CAPITAL LETTER A
        0x0411, # G+1021  U+0411  afii10018,  CYRILLIC CAPITAL LETTER BE
        0x0412, # G+1022  U+0412  afii10019,  CYRILLIC CAPITAL LETTER VE
        0x0413, # G+1023  U+0413  afii10020,  CYRILLIC CAPITAL LETTER GHE
        0x0414, # G+1024  U+0414  afii10021,  CYRILLIC CAPITAL LETTER DE
        0x0415, # G+1025  U+0415  afii10022,  CYRILLIC CAPITAL LETTER IE
        0x0401, # G+1026  U+0401  afii10023,  CYRILLIC CAPITAL LETTER IO
        0x0416, # G+1027  U+0416  afii10024,  CYRILLIC CAPITAL LETTER ZHE
        0x0417, # G+1028  U+0417  afii10025,  CYRILLIC CAPITAL LETTER ZE
        0x0418, # G+1029  U+0418  afii10026,  CYRILLIC CAPITAL LETTER I
        0x0419, # G+1030  U+0419  afii10027,  CYRILLIC CAPITAL LETTER SHORT I
        0x041A, # G+1031  U+041A  afii10028,  CYRILLIC CAPITAL LETTER KA
        0x041B, # G+1032  U+041B  afii10029,  CYRILLIC CAPITAL LETTER EL
        0x041C, # G+1033  U+041C  afii10030,  CYRILLIC CAPITAL LETTER EM
        0x041D, # G+1034  U+041D  afii10031,  CYRILLIC CAPITAL LETTER EN
        0x041E, # G+1035  U+041E  afii10032,  CYRILLIC CAPITAL LETTER O
        0x041F, # G+1036  U+041F  afii10033,  CYRILLIC CAPITAL LETTER PE
        0x0420, # G+1037  U+0420  afii10034,  CYRILLIC CAPITAL LETTER ER
        0x0421, # G+1038  U+0421  afii10035,  CYRILLIC CAPITAL LETTER ES
        0x0422, # G+1039  U+0422  afii10036,  CYRILLIC CAPITAL LETTER TE
        0x0423, # G+1040  U+0423  afii10037,  CYRILLIC CAPITAL LETTER U
        0x0424, # G+1041  U+0424  afii10038,  CYRILLIC CAPITAL LETTER EF
        0x0425, # G+1042  U+0425  afii10039,  CYRILLIC CAPITAL LETTER HA
        0x0426, # G+1043  U+0426  afii10040,  CYRILLIC CAPITAL LETTER TSE
        0x0427, # G+1044  U+0427  afii10041,  CYRILLIC CAPITAL LETTER CHE
        0x0428, # G+1045  U+0428  afii10042,  CYRILLIC CAPITAL LETTER SHA
        0x0429, # G+1046  U+0429  afii10043,  CYRILLIC CAPITAL LETTER SHCHA
        0x042A, # G+1047  U+042A  afii10044,  CYRILLIC CAPITAL LETTER HARD SIGN
        0x042B, # G+1048  U+042B  afii10045,  CYRILLIC CAPITAL LETTER YERU
        0x042C, # G+1049  U+042C  afii10046,  CYRILLIC CAPITAL LETTER SOFT SIGN
        0x042D, # G+1050  U+042D  afii10047,  CYRILLIC CAPITAL LETTER E
        0x042E, # G+1051  U+042E  afii10048,  CYRILLIC CAPITAL LETTER YU
        0x042F, # G+1052  U+042F  afii10049,  CYRILLIC CAPITAL LETTER YA
        0x0430, # G+1053  U+0430  afii10065,  CYRILLIC SMALL LETTER A
        0x0431, # G+1054  U+0431  afii10066,  CYRILLIC SMALL LETTER BE
        0x0432, # G+1055  U+0432  afii10067,  CYRILLIC SMALL LETTER VE
        0x0433, # G+1056  U+0433  afii10068,  CYRILLIC SMALL LETTER GHE
        0x0434, # G+1057  U+0434  afii10069,  CYRILLIC SMALL LETTER DE
        0x0435, # G+1058  U+0435  afii10070,  CYRILLIC SMALL LETTER IE
        0x0451, # G+1059  U+0451  afii10071,  CYRILLIC SMALL LETTER IO
        0x0436, # G+1060  U+0436  afii10072,  CYRILLIC SMALL LETTER ZHE
        0x0437, # G+1061  U+0437  afii10073,  CYRILLIC SMALL LETTER ZE
        0x0438, # G+1062  U+0438  afii10074,  CYRILLIC SMALL LETTER I
        0x0439, # G+1063  U+0439  afii10075,  CYRILLIC SMALL LETTER SHORT I
        0x043A, # G+1064  U+043A  afii10076,  CYRILLIC SMALL LETTER KA
        0x043B, # G+1065  U+043B  afii10077,  CYRILLIC SMALL LETTER EL
        0x043C, # G+1066  U+043C  afii10078,  CYRILLIC SMALL LETTER EM
        0x043D, # G+1067  U+043D  afii10079,  CYRILLIC SMALL LETTER EN
        0x043E, # G+1068  U+043E  afii10080,  CYRILLIC SMALL LETTER O
        0x043F, # G+1069  U+043F  afii10081,  CYRILLIC SMALL LETTER PE
        0x0440, # G+1070  U+0440  afii10082,  CYRILLIC SMALL LETTER ER
        0x0441, # G+1071  U+0441  afii10083,  CYRILLIC SMALL LETTER ES
        0x0442, # G+1072  U+0442  afii10084,  CYRILLIC SMALL LETTER TE
        0x0443, # G+1073  U+0443  afii10085,  CYRILLIC SMALL LETTER U
        0x0444, # G+1074  U+0444  afii10086,  CYRILLIC SMALL LETTER EF
        0x0445, # G+1075  U+0445  afii10087,  CYRILLIC SMALL LETTER HA
        0x0446, # G+1076  U+0446  afii10088,  CYRILLIC SMALL LETTER TSE
        0x0447, # G+1077  U+0447  afii10089,  CYRILLIC SMALL LETTER CHE
        0x0448, # G+1078  U+0448  afii10090,  CYRILLIC SMALL LETTER SHA
        0x0449, # G+1079  U+0449  afii10091,  CYRILLIC SMALL LETTER SHCHA
        0x044A, # G+1080  U+044A  afii10092,  CYRILLIC SMALL LETTER HARD SIGN
        0x044B, # G+1081  U+044B  afii10093,  CYRILLIC SMALL LETTER YERU
        0x044C, # G+1082  U+044C  afii10094,  CYRILLIC SMALL LETTER SOFT SIGN
        0x044D, # G+1083  U+044D  afii10095,  CYRILLIC SMALL LETTER E
        0x044E, # G+1084  U+044E  afii10096,  CYRILLIC SMALL LETTER YU
        0x044F, # G+1085  U+044F  afii10097,  CYRILLIC SMALL LETTER YA
        0xAC00, # G+1086  U+AC00  uniAC00,  HANGUL SYLLABLE-AC00
        0xAC01, # G+1087  U+AC01  uniAC01,  HANGUL SYLLABLE-AC01
        0xAC04, # G+1088  U+AC04  uniAC04,  HANGUL SYLLABLE-AC04
        0xAC07, # G+1089  U+AC07  uniAC07,  HANGUL SYLLABLE-AC07
        0xAC08, # G+1090  U+AC08  uniAC08,  HANGUL SYLLABLE-AC08
        0xAC09, # G+1091  U+AC09  uniAC09,  HANGUL SYLLABLE-AC09
        0xAC0A, # G+1092  U+AC0A  uniAC0A,  HANGUL SYLLABLE-AC0A
        0xAC10, # G+1093  U+AC10  uniAC10,  HANGUL SYLLABLE-AC10
        0xAC11, # G+1094  U+AC11  uniAC11,  HANGUL SYLLABLE-AC11
        0xAC12, # G+1095  U+AC12  uniAC12,  HANGUL SYLLABLE-AC12
        0xAC13, # G+1096  U+AC13  uniAC13,  HANGUL SYLLABLE-AC13
        0xAC14, # G+1097  U+AC14  uniAC14,  HANGUL SYLLABLE-AC14
        0xAC15, # G+1098  U+AC15  uniAC15,  HANGUL SYLLABLE-AC15
        0xAC16, # G+1099  U+AC16  uniAC16,  HANGUL SYLLABLE-AC16
        0xAC17, # G+1100  U+AC17  uniAC17,  HANGUL SYLLABLE-AC17
        0xAC19, # G+1101  U+AC19  uniAC19,  HANGUL SYLLABLE-AC19
        0xAC1A, # G+1102  U+AC1A  uniAC1A,  HANGUL SYLLABLE-AC1A
        0xAC1B, # G+1103  U+AC1B  uniAC1B,  HANGUL SYLLABLE-AC1B
        0xAC1C, # G+1104  U+AC1C  uniAC1C,  HANGUL SYLLABLE-AC1C
        0xAC1D, # G+1105  U+AC1D  uniAC1D,  HANGUL SYLLABLE-AC1D
        0xAC20, # G+1106  U+AC20  uniAC20,  HANGUL SYLLABLE-AC20
        0xAC24, # G+1107  U+AC24  uniAC24,  HANGUL SYLLABLE-AC24
        0xAC2C, # G+1108  U+AC2C  uniAC2C,  HANGUL SYLLABLE-AC2C
        0xAC2D, # G+1109  U+AC2D  uniAC2D,  HANGUL SYLLABLE-AC2D
        0xAC2F, # G+1110  U+AC2F  uniAC2F,  HANGUL SYLLABLE-AC2F
        0xAC30, # G+1111  U+AC30  uniAC30,  HANGUL SYLLABLE-AC30
        0xAC31, # G+1112  U+AC31  uniAC31,  HANGUL SYLLABLE-AC31
        0xAC38, # G+1113  U+AC38  uniAC38,  HANGUL SYLLABLE-AC38
        0xAC39, # G+1114  U+AC39  uniAC39,  HANGUL SYLLABLE-AC39
        0xAC3C, # G+1115  U+AC3C  uniAC3C,  HANGUL SYLLABLE-AC3C
        0xAC40, # G+1116  U+AC40  uniAC40,  HANGUL SYLLABLE-AC40
        0xAC4B, # G+1117  U+AC4B  uniAC4B,  HANGUL SYLLABLE-AC4B
        0xAC4D, # G+1118  U+AC4D  uniAC4D,  HANGUL SYLLABLE-AC4D
        0xAC54, # G+1119  U+AC54  uniAC54,  HANGUL SYLLABLE-AC54
        0xAC58, # G+1120  U+AC58  uniAC58,  HANGUL SYLLABLE-AC58
        0xAC5C, # G+1121  U+AC5C  uniAC5C,  HANGUL SYLLABLE-AC5C
        0xAC70, # G+1122  U+AC70  uniAC70,  HANGUL SYLLABLE-AC70
        0xAC71, # G+1123  U+AC71  uniAC71,  HANGUL SYLLABLE-AC71
        0xAC74, # G+1124  U+AC74  uniAC74,  HANGUL SYLLABLE-AC74
        0xAC77, # G+1125  U+AC77  uniAC77,  HANGUL SYLLABLE-AC77
        0xAC78, # G+1126  U+AC78  uniAC78,  HANGUL SYLLABLE-AC78
        0xAC7A, # G+1127  U+AC7A  uniAC7A,  HANGUL SYLLABLE-AC7A
        0xAC80, # G+1128  U+AC80  uniAC80,  HANGUL SYLLABLE-AC80
        0xAC81, # G+1129  U+AC81  uniAC81,  HANGUL SYLLABLE-AC81
        0xAC83, # G+1130  U+AC83  uniAC83,  HANGUL SYLLABLE-AC83
        0xAC84, # G+1131  U+AC84  uniAC84,  HANGUL SYLLABLE-AC84
        0xAC85, # G+1132  U+AC85  uniAC85,  HANGUL SYLLABLE-AC85
        0xAC86, # G+1133  U+AC86  uniAC86,  HANGUL SYLLABLE-AC86
        0xAC89, # G+1134  U+AC89  uniAC89,  HANGUL SYLLABLE-AC89
        0xAC8A, # G+1135  U+AC8A  uniAC8A,  HANGUL SYLLABLE-AC8A
        0xAC8B, # G+1136  U+AC8B  uniAC8B,  HANGUL SYLLABLE-AC8B
        0xAC8C, # G+1137  U+AC8C  uniAC8C,  HANGUL SYLLABLE-AC8C
        0xAC90, # G+1138  U+AC90  uniAC90,  HANGUL SYLLABLE-AC90
        0xAC94, # G+1139  U+AC94  uniAC94,  HANGUL SYLLABLE-AC94
        0xAC9C, # G+1140  U+AC9C  uniAC9C,  HANGUL SYLLABLE-AC9C
        0xAC9D, # G+1141  U+AC9D  uniAC9D,  HANGUL SYLLABLE-AC9D
        0xAC9F, # G+1142  U+AC9F  uniAC9F,  HANGUL SYLLABLE-AC9F
        0xACA0, # G+1143  U+ACA0  uniACA0,  HANGUL SYLLABLE-ACA0
        0xACA1, # G+1144  U+ACA1  uniACA1,  HANGUL SYLLABLE-ACA1
        0xACA8, # G+1145  U+ACA8  uniACA8,  HANGUL SYLLABLE-ACA8
        0xACA9, # G+1146  U+ACA9  uniACA9,  HANGUL SYLLABLE-ACA9
        0xACAA, # G+1147  U+ACAA  uniACAA,  HANGUL SYLLABLE-ACAA
        0xACAC, # G+1148  U+ACAC  uniACAC,  HANGUL SYLLABLE-ACAC
        0xACAF, # G+1149  U+ACAF  uniACAF,  HANGUL SYLLABLE-ACAF
        0xACB0, # G+1150  U+ACB0  uniACB0,  HANGUL SYLLABLE-ACB0
        0xACB8, # G+1151  U+ACB8  uniACB8,  HANGUL SYLLABLE-ACB8
        0xACB9, # G+1152  U+ACB9  uniACB9,  HANGUL SYLLABLE-ACB9
        0xACBB, # G+1153  U+ACBB  uniACBB,  HANGUL SYLLABLE-ACBB
        0xACBC, # G+1154  U+ACBC  uniACBC,  HANGUL SYLLABLE-ACBC
        0xACBD, # G+1155  U+ACBD  uniACBD,  HANGUL SYLLABLE-ACBD
        0xACC1, # G+1156  U+ACC1  uniACC1,  HANGUL SYLLABLE-ACC1
        0xACC4, # G+1157  U+ACC4  uniACC4,  HANGUL SYLLABLE-ACC4
        0xACC8, # G+1158  U+ACC8  uniACC8,  HANGUL SYLLABLE-ACC8
        0xACCC, # G+1159  U+ACCC  uniACCC,  HANGUL SYLLABLE-ACCC
        0xACD5, # G+1160  U+ACD5  uniACD5,  HANGUL SYLLABLE-ACD5
        0xACD7, # G+1161  U+ACD7  uniACD7,  HANGUL SYLLABLE-ACD7
        0xACE0, # G+1162  U+ACE0  uniACE0,  HANGUL SYLLABLE-ACE0
        0xACE1, # G+1163  U+ACE1  uniACE1,  HANGUL SYLLABLE-ACE1
        0xACE4, # G+1164  U+ACE4  uniACE4,  HANGUL SYLLABLE-ACE4
        0xACE7, # G+1165  U+ACE7  uniACE7,  HANGUL SYLLABLE-ACE7
        0xACE8, # G+1166  U+ACE8  uniACE8,  HANGUL SYLLABLE-ACE8
        0xACEA, # G+1167  U+ACEA  uniACEA,  HANGUL SYLLABLE-ACEA
        0xACEC, # G+1168  U+ACEC  uniACEC,  HANGUL SYLLABLE-ACEC
        0xACEF, # G+1169  U+ACEF  uniACEF,  HANGUL SYLLABLE-ACEF
        0xACF0, # G+1170  U+ACF0  uniACF0,  HANGUL SYLLABLE-ACF0
        0xACF1, # G+1171  U+ACF1  uniACF1,  HANGUL SYLLABLE-ACF1
        0xACF3, # G+1172  U+ACF3  uniACF3,  HANGUL SYLLABLE-ACF3
        0xACF5, # G+1173  U+ACF5  uniACF5,  HANGUL SYLLABLE-ACF5
        0xACF6, # G+1174  U+ACF6  uniACF6,  HANGUL SYLLABLE-ACF6
        0xACFC, # G+1175  U+ACFC  uniACFC,  HANGUL SYLLABLE-ACFC
        0xACFD, # G+1176  U+ACFD  uniACFD,  HANGUL SYLLABLE-ACFD
        0xAD00, # G+1177  U+AD00  uniAD00,  HANGUL SYLLABLE-AD00
        0xAD04, # G+1178  U+AD04  uniAD04,  HANGUL SYLLABLE-AD04
        0xAD06, # G+1179  U+AD06  uniAD06,  HANGUL SYLLABLE-AD06
        0xAD0C, # G+1180  U+AD0C  uniAD0C,  HANGUL SYLLABLE-AD0C
        0xAD0D, # G+1181  U+AD0D  uniAD0D,  HANGUL SYLLABLE-AD0D
        0xAD0F, # G+1182  U+AD0F  uniAD0F,  HANGUL SYLLABLE-AD0F
        0xAD11, # G+1183  U+AD11  uniAD11,  HANGUL SYLLABLE-AD11
        0xAD18, # G+1184  U+AD18  uniAD18,  HANGUL SYLLABLE-AD18
        0xAD1C, # G+1185  U+AD1C  uniAD1C,  HANGUL SYLLABLE-AD1C
        0xAD20, # G+1186  U+AD20  uniAD20,  HANGUL SYLLABLE-AD20
        0xAD29, # G+1187  U+AD29  uniAD29,  HANGUL SYLLABLE-AD29
        0xAD2C, # G+1188  U+AD2C  uniAD2C,  HANGUL SYLLABLE-AD2C
        0xAD2D, # G+1189  U+AD2D  uniAD2D,  HANGUL SYLLABLE-AD2D
        0xAD34, # G+1190  U+AD34  uniAD34,  HANGUL SYLLABLE-AD34
        0xAD35, # G+1191  U+AD35  uniAD35,  HANGUL SYLLABLE-AD35
        0xAD38, # G+1192  U+AD38  uniAD38,  HANGUL SYLLABLE-AD38
        0xAD3C, # G+1193  U+AD3C  uniAD3C,  HANGUL SYLLABLE-AD3C
        0xAD44, # G+1194  U+AD44  uniAD44,  HANGUL SYLLABLE-AD44
        0xAD45, # G+1195  U+AD45  uniAD45,  HANGUL SYLLABLE-AD45
        0xAD47, # G+1196  U+AD47  uniAD47,  HANGUL SYLLABLE-AD47
        0xAD49, # G+1197  U+AD49  uniAD49,  HANGUL SYLLABLE-AD49
        0xAD50, # G+1198  U+AD50  uniAD50,  HANGUL SYLLABLE-AD50
        0xAD54, # G+1199  U+AD54  uniAD54,  HANGUL SYLLABLE-AD54
        0xAD58, # G+1200  U+AD58  uniAD58,  HANGUL SYLLABLE-AD58
        0xAD61, # G+1201  U+AD61  uniAD61,  HANGUL SYLLABLE-AD61
        0xAD63, # G+1202  U+AD63  uniAD63,  HANGUL SYLLABLE-AD63
        0xAD6C, # G+1203  U+AD6C  uniAD6C,  HANGUL SYLLABLE-AD6C
        0xAD6D, # G+1204  U+AD6D  uniAD6D,  HANGUL SYLLABLE-AD6D
        0xAD70, # G+1205  U+AD70  uniAD70,  HANGUL SYLLABLE-AD70
        0xAD73, # G+1206  U+AD73  uniAD73,  HANGUL SYLLABLE-AD73
        0xAD74, # G+1207  U+AD74  uniAD74,  HANGUL SYLLABLE-AD74
        0xAD75, # G+1208  U+AD75  uniAD75,  HANGUL SYLLABLE-AD75
        0xAD76, # G+1209  U+AD76  uniAD76,  HANGUL SYLLABLE-AD76
        0xAD7B, # G+1210  U+AD7B  uniAD7B,  HANGUL SYLLABLE-AD7B
        0xAD7C, # G+1211  U+AD7C  uniAD7C,  HANGUL SYLLABLE-AD7C
        0xAD7D, # G+1212  U+AD7D  uniAD7D,  HANGUL SYLLABLE-AD7D
        0xAD7F, # G+1213  U+AD7F  uniAD7F,  HANGUL SYLLABLE-AD7F
        0xAD81, # G+1214  U+AD81  uniAD81,  HANGUL SYLLABLE-AD81
        0xAD82, # G+1215  U+AD82  uniAD82,  HANGUL SYLLABLE-AD82
        0xAD88, # G+1216  U+AD88  uniAD88,  HANGUL SYLLABLE-AD88
        0xAD89, # G+1217  U+AD89  uniAD89,  HANGUL SYLLABLE-AD89
        0xAD8C, # G+1218  U+AD8C  uniAD8C,  HANGUL SYLLABLE-AD8C
        0xAD90, # G+1219  U+AD90  uniAD90,  HANGUL SYLLABLE-AD90
        0xAD9C, # G+1220  U+AD9C  uniAD9C,  HANGUL SYLLABLE-AD9C
        0xAD9D, # G+1221  U+AD9D  uniAD9D,  HANGUL SYLLABLE-AD9D
        0xADA4, # G+1222  U+ADA4  uniADA4,  HANGUL SYLLABLE-ADA4
        0xADB7, # G+1223  U+ADB7  uniADB7,  HANGUL SYLLABLE-ADB7
        0xADC0, # G+1224  U+ADC0  uniADC0,  HANGUL SYLLABLE-ADC0
        0xADC1, # G+1225  U+ADC1  uniADC1,  HANGUL SYLLABLE-ADC1
        0xADC4, # G+1226  U+ADC4  uniADC4,  HANGUL SYLLABLE-ADC4
        0xADC8, # G+1227  U+ADC8  uniADC8,  HANGUL SYLLABLE-ADC8
        0xADD0, # G+1228  U+ADD0  uniADD0,  HANGUL SYLLABLE-ADD0
        0xADD1, # G+1229  U+ADD1  uniADD1,  HANGUL SYLLABLE-ADD1
        0xADD3, # G+1230  U+ADD3  uniADD3,  HANGUL SYLLABLE-ADD3
        0xADDC, # G+1231  U+ADDC  uniADDC,  HANGUL SYLLABLE-ADDC
        0xADE0, # G+1232  U+ADE0  uniADE0,  HANGUL SYLLABLE-ADE0
        0xADE4, # G+1233  U+ADE4  uniADE4,  HANGUL SYLLABLE-ADE4
        0xADF8, # G+1234  U+ADF8  uniADF8,  HANGUL SYLLABLE-ADF8
        0xADF9, # G+1235  U+ADF9  uniADF9,  HANGUL SYLLABLE-ADF9
        0xADFC, # G+1236  U+ADFC  uniADFC,  HANGUL SYLLABLE-ADFC
        0xADFF, # G+1237  U+ADFF  uniADFF,  HANGUL SYLLABLE-ADFF
        0xAE00, # G+1238  U+AE00  uniAE00,  HANGUL SYLLABLE-AE00
        0xAE01, # G+1239  U+AE01  uniAE01,  HANGUL SYLLABLE-AE01
        0xAE08, # G+1240  U+AE08  uniAE08,  HANGUL SYLLABLE-AE08
        0xAE09, # G+1241  U+AE09  uniAE09,  HANGUL SYLLABLE-AE09
        0xAE0B, # G+1242  U+AE0B  uniAE0B,  HANGUL SYLLABLE-AE0B
        0xAE0D, # G+1243  U+AE0D  uniAE0D,  HANGUL SYLLABLE-AE0D
        0xAE14, # G+1244  U+AE14  uniAE14,  HANGUL SYLLABLE-AE14
        0xAE30, # G+1245  U+AE30  uniAE30,  HANGUL SYLLABLE-AE30
        0xAE31, # G+1246  U+AE31  uniAE31,  HANGUL SYLLABLE-AE31
        0xAE34, # G+1247  U+AE34  uniAE34,  HANGUL SYLLABLE-AE34
        0xAE37, # G+1248  U+AE37  uniAE37,  HANGUL SYLLABLE-AE37
        0xAE38, # G+1249  U+AE38  uniAE38,  HANGUL SYLLABLE-AE38
        0xAE3A, # G+1250  U+AE3A  uniAE3A,  HANGUL SYLLABLE-AE3A
        0xAE40, # G+1251  U+AE40  uniAE40,  HANGUL SYLLABLE-AE40
        0xAE41, # G+1252  U+AE41  uniAE41,  HANGUL SYLLABLE-AE41
        0xAE43, # G+1253  U+AE43  uniAE43,  HANGUL SYLLABLE-AE43
        0xAE45, # G+1254  U+AE45  uniAE45,  HANGUL SYLLABLE-AE45
        0xAE46, # G+1255  U+AE46  uniAE46,  HANGUL SYLLABLE-AE46
        0xAE4A, # G+1256  U+AE4A  uniAE4A,  HANGUL SYLLABLE-AE4A
        0xAE4C, # G+1257  U+AE4C  uniAE4C,  HANGUL SYLLABLE-AE4C
        0xAE4D, # G+1258  U+AE4D  uniAE4D,  HANGUL SYLLABLE-AE4D
        0xAE4E, # G+1259  U+AE4E  uniAE4E,  HANGUL SYLLABLE-AE4E
        0xAE50, # G+1260  U+AE50  uniAE50,  HANGUL SYLLABLE-AE50
        0xAE54, # G+1261  U+AE54  uniAE54,  HANGUL SYLLABLE-AE54
        0xAE56, # G+1262  U+AE56  uniAE56,  HANGUL SYLLABLE-AE56
        0xAE5C, # G+1263  U+AE5C  uniAE5C,  HANGUL SYLLABLE-AE5C
        0xAE5D, # G+1264  U+AE5D  uniAE5D,  HANGUL SYLLABLE-AE5D
        0xAE5F, # G+1265  U+AE5F  uniAE5F,  HANGUL SYLLABLE-AE5F
        0xAE60, # G+1266  U+AE60  uniAE60,  HANGUL SYLLABLE-AE60
        0xAE61, # G+1267  U+AE61  uniAE61,  HANGUL SYLLABLE-AE61
        0xAE65, # G+1268  U+AE65  uniAE65,  HANGUL SYLLABLE-AE65
        0xAE68, # G+1269  U+AE68  uniAE68,  HANGUL SYLLABLE-AE68
        0xAE69, # G+1270  U+AE69  uniAE69,  HANGUL SYLLABLE-AE69
        0xAE6C, # G+1271  U+AE6C  uniAE6C,  HANGUL SYLLABLE-AE6C
        0xAE70, # G+1272  U+AE70  uniAE70,  HANGUL SYLLABLE-AE70
        0xAE78, # G+1273  U+AE78  uniAE78,  HANGUL SYLLABLE-AE78
        0xAE79, # G+1274  U+AE79  uniAE79,  HANGUL SYLLABLE-AE79
        0xAE7B, # G+1275  U+AE7B  uniAE7B,  HANGUL SYLLABLE-AE7B
        0xAE7C, # G+1276  U+AE7C  uniAE7C,  HANGUL SYLLABLE-AE7C
        0xAE7D, # G+1277  U+AE7D  uniAE7D,  HANGUL SYLLABLE-AE7D
        0xAE84, # G+1278  U+AE84  uniAE84,  HANGUL SYLLABLE-AE84
        0xAE85, # G+1279  U+AE85  uniAE85,  HANGUL SYLLABLE-AE85
        0xAE8C, # G+1280  U+AE8C  uniAE8C,  HANGUL SYLLABLE-AE8C
        0xAEBC, # G+1281  U+AEBC  uniAEBC,  HANGUL SYLLABLE-AEBC
        0xAEBD, # G+1282  U+AEBD  uniAEBD,  HANGUL SYLLABLE-AEBD
        0xAEBE, # G+1283  U+AEBE  uniAEBE,  HANGUL SYLLABLE-AEBE
        0xAEC0, # G+1284  U+AEC0  uniAEC0,  HANGUL SYLLABLE-AEC0
        0xAEC4, # G+1285  U+AEC4  uniAEC4,  HANGUL SYLLABLE-AEC4
        0xAECC, # G+1286  U+AECC  uniAECC,  HANGUL SYLLABLE-AECC
        0xAECD, # G+1287  U+AECD  uniAECD,  HANGUL SYLLABLE-AECD
        0xAECF, # G+1288  U+AECF  uniAECF,  HANGUL SYLLABLE-AECF
        0xAED0, # G+1289  U+AED0  uniAED0,  HANGUL SYLLABLE-AED0
        0xAED1, # G+1290  U+AED1  uniAED1,  HANGUL SYLLABLE-AED1
        0xAED8, # G+1291  U+AED8  uniAED8,  HANGUL SYLLABLE-AED8
        0xAED9, # G+1292  U+AED9  uniAED9,  HANGUL SYLLABLE-AED9
        0xAEDC, # G+1293  U+AEDC  uniAEDC,  HANGUL SYLLABLE-AEDC
        0xAEE8, # G+1294  U+AEE8  uniAEE8,  HANGUL SYLLABLE-AEE8
        0xAEEB, # G+1295  U+AEEB  uniAEEB,  HANGUL SYLLABLE-AEEB
        0xAEED, # G+1296  U+AEED  uniAEED,  HANGUL SYLLABLE-AEED
        0xAEF4, # G+1297  U+AEF4  uniAEF4,  HANGUL SYLLABLE-AEF4
        0xAEF8, # G+1298  U+AEF8  uniAEF8,  HANGUL SYLLABLE-AEF8
        0xAEFC, # G+1299  U+AEFC  uniAEFC,  HANGUL SYLLABLE-AEFC
        0xAF07, # G+1300  U+AF07  uniAF07,  HANGUL SYLLABLE-AF07
        0xAF08, # G+1301  U+AF08  uniAF08,  HANGUL SYLLABLE-AF08
        0xAF0D, # G+1302  U+AF0D  uniAF0D,  HANGUL SYLLABLE-AF0D
        0xAF10, # G+1303  U+AF10  uniAF10,  HANGUL SYLLABLE-AF10
        0xAF2C, # G+1304  U+AF2C  uniAF2C,  HANGUL SYLLABLE-AF2C
        0xAF2D, # G+1305  U+AF2D  uniAF2D,  HANGUL SYLLABLE-AF2D
        0xAF30, # G+1306  U+AF30  uniAF30,  HANGUL SYLLABLE-AF30
        0xAF32, # G+1307  U+AF32  uniAF32,  HANGUL SYLLABLE-AF32
        0xAF34, # G+1308  U+AF34  uniAF34,  HANGUL SYLLABLE-AF34
        0xAF3C, # G+1309  U+AF3C  uniAF3C,  HANGUL SYLLABLE-AF3C
        0xAF3D, # G+1310  U+AF3D  uniAF3D,  HANGUL SYLLABLE-AF3D
        0xAF3F, # G+1311  U+AF3F  uniAF3F,  HANGUL SYLLABLE-AF3F
        0xAF41, # G+1312  U+AF41  uniAF41,  HANGUL SYLLABLE-AF41
        0xAF42, # G+1313  U+AF42  uniAF42,  HANGUL SYLLABLE-AF42
        0xAF43, # G+1314  U+AF43  uniAF43,  HANGUL SYLLABLE-AF43
        0xAF48, # G+1315  U+AF48  uniAF48,  HANGUL SYLLABLE-AF48
        0xAF49, # G+1316  U+AF49  uniAF49,  HANGUL SYLLABLE-AF49
        0xAF50, # G+1317  U+AF50  uniAF50,  HANGUL SYLLABLE-AF50
        0xAF5C, # G+1318  U+AF5C  uniAF5C,  HANGUL SYLLABLE-AF5C
        0xAF5D, # G+1319  U+AF5D  uniAF5D,  HANGUL SYLLABLE-AF5D
        0xAF64, # G+1320  U+AF64  uniAF64,  HANGUL SYLLABLE-AF64
        0xAF65, # G+1321  U+AF65  uniAF65,  HANGUL SYLLABLE-AF65
        0xAF79, # G+1322  U+AF79  uniAF79,  HANGUL SYLLABLE-AF79
        0xAF80, # G+1323  U+AF80  uniAF80,  HANGUL SYLLABLE-AF80
        0xAF84, # G+1324  U+AF84  uniAF84,  HANGUL SYLLABLE-AF84
        0xAF88, # G+1325  U+AF88  uniAF88,  HANGUL SYLLABLE-AF88
        0xAF90, # G+1326  U+AF90  uniAF90,  HANGUL SYLLABLE-AF90
        0xAF91, # G+1327  U+AF91  uniAF91,  HANGUL SYLLABLE-AF91
        0xAF95, # G+1328  U+AF95  uniAF95,  HANGUL SYLLABLE-AF95
        0xAF9C, # G+1329  U+AF9C  uniAF9C,  HANGUL SYLLABLE-AF9C
        0xAFB8, # G+1330  U+AFB8  uniAFB8,  HANGUL SYLLABLE-AFB8
        0xAFB9, # G+1331  U+AFB9  uniAFB9,  HANGUL SYLLABLE-AFB9
        0xAFBC, # G+1332  U+AFBC  uniAFBC,  HANGUL SYLLABLE-AFBC
        0xAFC0, # G+1333  U+AFC0  uniAFC0,  HANGUL SYLLABLE-AFC0
        0xAFC7, # G+1334  U+AFC7  uniAFC7,  HANGUL SYLLABLE-AFC7
        0xAFC8, # G+1335  U+AFC8  uniAFC8,  HANGUL SYLLABLE-AFC8
        0xAFC9, # G+1336  U+AFC9  uniAFC9,  HANGUL SYLLABLE-AFC9
        0xAFCB, # G+1337  U+AFCB  uniAFCB,  HANGUL SYLLABLE-AFCB
        0xAFCD, # G+1338  U+AFCD  uniAFCD,  HANGUL SYLLABLE-AFCD
        0xAFCE, # G+1339  U+AFCE  uniAFCE,  HANGUL SYLLABLE-AFCE
        0xAFD4, # G+1340  U+AFD4  uniAFD4,  HANGUL SYLLABLE-AFD4
        0xAFDC, # G+1341  U+AFDC  uniAFDC,  HANGUL SYLLABLE-AFDC
        0xAFE8, # G+1342  U+AFE8  uniAFE8,  HANGUL SYLLABLE-AFE8
        0xAFE9, # G+1343  U+AFE9  uniAFE9,  HANGUL SYLLABLE-AFE9
        0xAFF0, # G+1344  U+AFF0  uniAFF0,  HANGUL SYLLABLE-AFF0
        0xAFF1, # G+1345  U+AFF1  uniAFF1,  HANGUL SYLLABLE-AFF1
        0xAFF4, # G+1346  U+AFF4  uniAFF4,  HANGUL SYLLABLE-AFF4
        0xAFF8, # G+1347  U+AFF8  uniAFF8,  HANGUL SYLLABLE-AFF8
        0xB000, # G+1348  U+B000  uniB000,  HANGUL SYLLABLE-B000
        0xB001, # G+1349  U+B001  uniB001,  HANGUL SYLLABLE-B001
        0xB004, # G+1350  U+B004  uniB004,  HANGUL SYLLABLE-B004
        0xB00C, # G+1351  U+B00C  uniB00C,  HANGUL SYLLABLE-B00C
        0xB010, # G+1352  U+B010  uniB010,  HANGUL SYLLABLE-B010
        0xB014, # G+1353  U+B014  uniB014,  HANGUL SYLLABLE-B014
        0xB01C, # G+1354  U+B01C  uniB01C,  HANGUL SYLLABLE-B01C
        0xB01D, # G+1355  U+B01D  uniB01D,  HANGUL SYLLABLE-B01D
        0xB028, # G+1356  U+B028  uniB028,  HANGUL SYLLABLE-B028
        0xB044, # G+1357  U+B044  uniB044,  HANGUL SYLLABLE-B044
        0xB045, # G+1358  U+B045  uniB045,  HANGUL SYLLABLE-B045
        0xB048, # G+1359  U+B048  uniB048,  HANGUL SYLLABLE-B048
        0xB04A, # G+1360  U+B04A  uniB04A,  HANGUL SYLLABLE-B04A
        0xB04C, # G+1361  U+B04C  uniB04C,  HANGUL SYLLABLE-B04C
        0xB04E, # G+1362  U+B04E  uniB04E,  HANGUL SYLLABLE-B04E
        0xB053, # G+1363  U+B053  uniB053,  HANGUL SYLLABLE-B053
        0xB054, # G+1364  U+B054  uniB054,  HANGUL SYLLABLE-B054
        0xB055, # G+1365  U+B055  uniB055,  HANGUL SYLLABLE-B055
        0xB057, # G+1366  U+B057  uniB057,  HANGUL SYLLABLE-B057
        0xB059, # G+1367  U+B059  uniB059,  HANGUL SYLLABLE-B059
        0xB05D, # G+1368  U+B05D  uniB05D,  HANGUL SYLLABLE-B05D
        0xB07C, # G+1369  U+B07C  uniB07C,  HANGUL SYLLABLE-B07C
        0xB07D, # G+1370  U+B07D  uniB07D,  HANGUL SYLLABLE-B07D
        0xB080, # G+1371  U+B080  uniB080,  HANGUL SYLLABLE-B080
        0xB084, # G+1372  U+B084  uniB084,  HANGUL SYLLABLE-B084
        0xB08C, # G+1373  U+B08C  uniB08C,  HANGUL SYLLABLE-B08C
        0xB08D, # G+1374  U+B08D  uniB08D,  HANGUL SYLLABLE-B08D
        0xB08F, # G+1375  U+B08F  uniB08F,  HANGUL SYLLABLE-B08F
        0xB091, # G+1376  U+B091  uniB091,  HANGUL SYLLABLE-B091
        0xB098, # G+1377  U+B098  uniB098,  HANGUL SYLLABLE-B098
        0xB099, # G+1378  U+B099  uniB099,  HANGUL SYLLABLE-B099
        0xB09A, # G+1379  U+B09A  uniB09A,  HANGUL SYLLABLE-B09A
        0xB09C, # G+1380  U+B09C  uniB09C,  HANGUL SYLLABLE-B09C
        0xB09F, # G+1381  U+B09F  uniB09F,  HANGUL SYLLABLE-B09F
        0xB0A0, # G+1382  U+B0A0  uniB0A0,  HANGUL SYLLABLE-B0A0
        0xB0A1, # G+1383  U+B0A1  uniB0A1,  HANGUL SYLLABLE-B0A1
        0xB0A2, # G+1384  U+B0A2  uniB0A2,  HANGUL SYLLABLE-B0A2
        0xB0A8, # G+1385  U+B0A8  uniB0A8,  HANGUL SYLLABLE-B0A8
        0xB0A9, # G+1386  U+B0A9  uniB0A9,  HANGUL SYLLABLE-B0A9
        0xB0AB, # G+1387  U+B0AB  uniB0AB,  HANGUL SYLLABLE-B0AB
        0xB0AC, # G+1388  U+B0AC  uniB0AC,  HANGUL SYLLABLE-B0AC
        0xB0AD, # G+1389  U+B0AD  uniB0AD,  HANGUL SYLLABLE-B0AD
        0xB0AE, # G+1390  U+B0AE  uniB0AE,  HANGUL SYLLABLE-B0AE
        0xB0AF, # G+1391  U+B0AF  uniB0AF,  HANGUL SYLLABLE-B0AF
        0xB0B1, # G+1392  U+B0B1  uniB0B1,  HANGUL SYLLABLE-B0B1
        0xB0B3, # G+1393  U+B0B3  uniB0B3,  HANGUL SYLLABLE-B0B3
        0xB0B4, # G+1394  U+B0B4  uniB0B4,  HANGUL SYLLABLE-B0B4
        0xB0B5, # G+1395  U+B0B5  uniB0B5,  HANGUL SYLLABLE-B0B5
        0xB0B8, # G+1396  U+B0B8  uniB0B8,  HANGUL SYLLABLE-B0B8
        0xB0BC, # G+1397  U+B0BC  uniB0BC,  HANGUL SYLLABLE-B0BC
        0xB0C4, # G+1398  U+B0C4  uniB0C4,  HANGUL SYLLABLE-B0C4
        0xB0C5, # G+1399  U+B0C5  uniB0C5,  HANGUL SYLLABLE-B0C5
        0xB0C7, # G+1400  U+B0C7  uniB0C7,  HANGUL SYLLABLE-B0C7
        0xB0C8, # G+1401  U+B0C8  uniB0C8,  HANGUL SYLLABLE-B0C8
        0xB0C9, # G+1402  U+B0C9  uniB0C9,  HANGUL SYLLABLE-B0C9
        0xB0D0, # G+1403  U+B0D0  uniB0D0,  HANGUL SYLLABLE-B0D0
        0xB0D1, # G+1404  U+B0D1  uniB0D1,  HANGUL SYLLABLE-B0D1
        0xB0D4, # G+1405  U+B0D4  uniB0D4,  HANGUL SYLLABLE-B0D4
        0xB0D8, # G+1406  U+B0D8  uniB0D8,  HANGUL SYLLABLE-B0D8
        0xB0E0, # G+1407  U+B0E0  uniB0E0,  HANGUL SYLLABLE-B0E0
        0xB0E5, # G+1408  U+B0E5  uniB0E5,  HANGUL SYLLABLE-B0E5
        0xB108, # G+1409  U+B108  uniB108,  HANGUL SYLLABLE-B108
        0xB109, # G+1410  U+B109  uniB109,  HANGUL SYLLABLE-B109
        0xB10B, # G+1411  U+B10B  uniB10B,  HANGUL SYLLABLE-B10B
        0xB10C, # G+1412  U+B10C  uniB10C,  HANGUL SYLLABLE-B10C
        0xB110, # G+1413  U+B110  uniB110,  HANGUL SYLLABLE-B110
        0xB112, # G+1414  U+B112  uniB112,  HANGUL SYLLABLE-B112
        0xB113, # G+1415  U+B113  uniB113,  HANGUL SYLLABLE-B113
        0xB118, # G+1416  U+B118  uniB118,  HANGUL SYLLABLE-B118
        0xB119, # G+1417  U+B119  uniB119,  HANGUL SYLLABLE-B119
        0xB11B, # G+1418  U+B11B  uniB11B,  HANGUL SYLLABLE-B11B
        0xB11C, # G+1419  U+B11C  uniB11C,  HANGUL SYLLABLE-B11C
        0xB11D, # G+1420  U+B11D  uniB11D,  HANGUL SYLLABLE-B11D
        0xB123, # G+1421  U+B123  uniB123,  HANGUL SYLLABLE-B123
        0xB124, # G+1422  U+B124  uniB124,  HANGUL SYLLABLE-B124
        0xB125, # G+1423  U+B125  uniB125,  HANGUL SYLLABLE-B125
        0xB128, # G+1424  U+B128  uniB128,  HANGUL SYLLABLE-B128
        0xB12C, # G+1425  U+B12C  uniB12C,  HANGUL SYLLABLE-B12C
        0xB134, # G+1426  U+B134  uniB134,  HANGUL SYLLABLE-B134
        0xB135, # G+1427  U+B135  uniB135,  HANGUL SYLLABLE-B135
        0xB137, # G+1428  U+B137  uniB137,  HANGUL SYLLABLE-B137
        0xB138, # G+1429  U+B138  uniB138,  HANGUL SYLLABLE-B138
        0xB139, # G+1430  U+B139  uniB139,  HANGUL SYLLABLE-B139
        0xB140, # G+1431  U+B140  uniB140,  HANGUL SYLLABLE-B140
        0xB141, # G+1432  U+B141  uniB141,  HANGUL SYLLABLE-B141
        0xB144, # G+1433  U+B144  uniB144,  HANGUL SYLLABLE-B144
        0xB148, # G+1434  U+B148  uniB148,  HANGUL SYLLABLE-B148
        0xB150, # G+1435  U+B150  uniB150,  HANGUL SYLLABLE-B150
        0xB151, # G+1436  U+B151  uniB151,  HANGUL SYLLABLE-B151
        0xB154, # G+1437  U+B154  uniB154,  HANGUL SYLLABLE-B154
        0xB155, # G+1438  U+B155  uniB155,  HANGUL SYLLABLE-B155
        0xB158, # G+1439  U+B158  uniB158,  HANGUL SYLLABLE-B158
        0xB15C, # G+1440  U+B15C  uniB15C,  HANGUL SYLLABLE-B15C
        0xB160, # G+1441  U+B160  uniB160,  HANGUL SYLLABLE-B160
        0xB178, # G+1442  U+B178  uniB178,  HANGUL SYLLABLE-B178
        0xB179, # G+1443  U+B179  uniB179,  HANGUL SYLLABLE-B179
        0xB17C, # G+1444  U+B17C  uniB17C,  HANGUL SYLLABLE-B17C
        0xB180, # G+1445  U+B180  uniB180,  HANGUL SYLLABLE-B180
        0xB182, # G+1446  U+B182  uniB182,  HANGUL SYLLABLE-B182
        0xB188, # G+1447  U+B188  uniB188,  HANGUL SYLLABLE-B188
        0xB189, # G+1448  U+B189  uniB189,  HANGUL SYLLABLE-B189
        0xB18B, # G+1449  U+B18B  uniB18B,  HANGUL SYLLABLE-B18B
        0xB18D, # G+1450  U+B18D  uniB18D,  HANGUL SYLLABLE-B18D
        0xB192, # G+1451  U+B192  uniB192,  HANGUL SYLLABLE-B192
        0xB193, # G+1452  U+B193  uniB193,  HANGUL SYLLABLE-B193
        0xB194, # G+1453  U+B194  uniB194,  HANGUL SYLLABLE-B194
        0xB198, # G+1454  U+B198  uniB198,  HANGUL SYLLABLE-B198
        0xB19C, # G+1455  U+B19C  uniB19C,  HANGUL SYLLABLE-B19C
        0xB1A8, # G+1456  U+B1A8  uniB1A8,  HANGUL SYLLABLE-B1A8
        0xB1CC, # G+1457  U+B1CC  uniB1CC,  HANGUL SYLLABLE-B1CC
        0xB1D0, # G+1458  U+B1D0  uniB1D0,  HANGUL SYLLABLE-B1D0
        0xB1D4, # G+1459  U+B1D4  uniB1D4,  HANGUL SYLLABLE-B1D4
        0xB1DC, # G+1460  U+B1DC  uniB1DC,  HANGUL SYLLABLE-B1DC
        0xB1DD, # G+1461  U+B1DD  uniB1DD,  HANGUL SYLLABLE-B1DD
        0xB1DF, # G+1462  U+B1DF  uniB1DF,  HANGUL SYLLABLE-B1DF
        0xB1E8, # G+1463  U+B1E8  uniB1E8,  HANGUL SYLLABLE-B1E8
        0xB1E9, # G+1464  U+B1E9  uniB1E9,  HANGUL SYLLABLE-B1E9
        0xB1EC, # G+1465  U+B1EC  uniB1EC,  HANGUL SYLLABLE-B1EC
        0xB1F0, # G+1466  U+B1F0  uniB1F0,  HANGUL SYLLABLE-B1F0
        0xB1F9, # G+1467  U+B1F9  uniB1F9,  HANGUL SYLLABLE-B1F9
        0xB1FB, # G+1468  U+B1FB  uniB1FB,  HANGUL SYLLABLE-B1FB
        0xB1FD, # G+1469  U+B1FD  uniB1FD,  HANGUL SYLLABLE-B1FD
        0xB204, # G+1470  U+B204  uniB204,  HANGUL SYLLABLE-B204
        0xB205, # G+1471  U+B205  uniB205,  HANGUL SYLLABLE-B205
        0xB208, # G+1472  U+B208  uniB208,  HANGUL SYLLABLE-B208
        0xB20B, # G+1473  U+B20B  uniB20B,  HANGUL SYLLABLE-B20B
        0xB20C, # G+1474  U+B20C  uniB20C,  HANGUL SYLLABLE-B20C
        0xB214, # G+1475  U+B214  uniB214,  HANGUL SYLLABLE-B214
        0xB215, # G+1476  U+B215  uniB215,  HANGUL SYLLABLE-B215
        0xB217, # G+1477  U+B217  uniB217,  HANGUL SYLLABLE-B217
        0xB219, # G+1478  U+B219  uniB219,  HANGUL SYLLABLE-B219
        0xB220, # G+1479  U+B220  uniB220,  HANGUL SYLLABLE-B220
        0xB234, # G+1480  U+B234  uniB234,  HANGUL SYLLABLE-B234
        0xB23C, # G+1481  U+B23C  uniB23C,  HANGUL SYLLABLE-B23C
        0xB258, # G+1482  U+B258  uniB258,  HANGUL SYLLABLE-B258
        0xB25C, # G+1483  U+B25C  uniB25C,  HANGUL SYLLABLE-B25C
        0xB260, # G+1484  U+B260  uniB260,  HANGUL SYLLABLE-B260
        0xB268, # G+1485  U+B268  uniB268,  HANGUL SYLLABLE-B268
        0xB269, # G+1486  U+B269  uniB269,  HANGUL SYLLABLE-B269
        0xB274, # G+1487  U+B274  uniB274,  HANGUL SYLLABLE-B274
        0xB275, # G+1488  U+B275  uniB275,  HANGUL SYLLABLE-B275
        0xB27C, # G+1489  U+B27C  uniB27C,  HANGUL SYLLABLE-B27C
        0xB284, # G+1490  U+B284  uniB284,  HANGUL SYLLABLE-B284
        0xB285, # G+1491  U+B285  uniB285,  HANGUL SYLLABLE-B285
        0xB289, # G+1492  U+B289  uniB289,  HANGUL SYLLABLE-B289
        0xB290, # G+1493  U+B290  uniB290,  HANGUL SYLLABLE-B290
        0xB291, # G+1494  U+B291  uniB291,  HANGUL SYLLABLE-B291
        0xB294, # G+1495  U+B294  uniB294,  HANGUL SYLLABLE-B294
        0xB298, # G+1496  U+B298  uniB298,  HANGUL SYLLABLE-B298
        0xB299, # G+1497  U+B299  uniB299,  HANGUL SYLLABLE-B299
        0xB29A, # G+1498  U+B29A  uniB29A,  HANGUL SYLLABLE-B29A
        0xB2A0, # G+1499  U+B2A0  uniB2A0,  HANGUL SYLLABLE-B2A0
        0xB2A1, # G+1500  U+B2A1  uniB2A1,  HANGUL SYLLABLE-B2A1
        0xB2A3, # G+1501  U+B2A3  uniB2A3,  HANGUL SYLLABLE-B2A3
        0xB2A5, # G+1502  U+B2A5  uniB2A5,  HANGUL SYLLABLE-B2A5
        0xB2A6, # G+1503  U+B2A6  uniB2A6,  HANGUL SYLLABLE-B2A6
        0xB2AA, # G+1504  U+B2AA  uniB2AA,  HANGUL SYLLABLE-B2AA
        0xB2AC, # G+1505  U+B2AC  uniB2AC,  HANGUL SYLLABLE-B2AC
        0xB2B0, # G+1506  U+B2B0  uniB2B0,  HANGUL SYLLABLE-B2B0
        0xB2B4, # G+1507  U+B2B4  uniB2B4,  HANGUL SYLLABLE-B2B4
        0xB2C8, # G+1508  U+B2C8  uniB2C8,  HANGUL SYLLABLE-B2C8
        0xB2C9, # G+1509  U+B2C9  uniB2C9,  HANGUL SYLLABLE-B2C9
        0xB2CC, # G+1510  U+B2CC  uniB2CC,  HANGUL SYLLABLE-B2CC
        0xB2D0, # G+1511  U+B2D0  uniB2D0,  HANGUL SYLLABLE-B2D0
        0xB2D2, # G+1512  U+B2D2  uniB2D2,  HANGUL SYLLABLE-B2D2
        0xB2D8, # G+1513  U+B2D8  uniB2D8,  HANGUL SYLLABLE-B2D8
        0xB2D9, # G+1514  U+B2D9  uniB2D9,  HANGUL SYLLABLE-B2D9
        0xB2DB, # G+1515  U+B2DB  uniB2DB,  HANGUL SYLLABLE-B2DB
        0xB2DD, # G+1516  U+B2DD  uniB2DD,  HANGUL SYLLABLE-B2DD
        0xB2E2, # G+1517  U+B2E2  uniB2E2,  HANGUL SYLLABLE-B2E2
        0xB2E4, # G+1518  U+B2E4  uniB2E4,  HANGUL SYLLABLE-B2E4
        0xB2E5, # G+1519  U+B2E5  uniB2E5,  HANGUL SYLLABLE-B2E5
        0xB2E6, # G+1520  U+B2E6  uniB2E6,  HANGUL SYLLABLE-B2E6
        0xB2E8, # G+1521  U+B2E8  uniB2E8,  HANGUL SYLLABLE-B2E8
        0xB2EB, # G+1522  U+B2EB  uniB2EB,  HANGUL SYLLABLE-B2EB
        0xB2EC, # G+1523  U+B2EC  uniB2EC,  HANGUL SYLLABLE-B2EC
        0xB2ED, # G+1524  U+B2ED  uniB2ED,  HANGUL SYLLABLE-B2ED
        0xB2EE, # G+1525  U+B2EE  uniB2EE,  HANGUL SYLLABLE-B2EE
        0xB2EF, # G+1526  U+B2EF  uniB2EF,  HANGUL SYLLABLE-B2EF
        0xB2F3, # G+1527  U+B2F3  uniB2F3,  HANGUL SYLLABLE-B2F3
        0xB2F4, # G+1528  U+B2F4  uniB2F4,  HANGUL SYLLABLE-B2F4
        0xB2F5, # G+1529  U+B2F5  uniB2F5,  HANGUL SYLLABLE-B2F5
        0xB2F7, # G+1530  U+B2F7  uniB2F7,  HANGUL SYLLABLE-B2F7
        0xB2F8, # G+1531  U+B2F8  uniB2F8,  HANGUL SYLLABLE-B2F8
        0xB2F9, # G+1532  U+B2F9  uniB2F9,  HANGUL SYLLABLE-B2F9
        0xB2FA, # G+1533  U+B2FA  uniB2FA,  HANGUL SYLLABLE-B2FA
        0xB2FB, # G+1534  U+B2FB  uniB2FB,  HANGUL SYLLABLE-B2FB
        0xB2FF, # G+1535  U+B2FF  uniB2FF,  HANGUL SYLLABLE-B2FF
        0xB300, # G+1536  U+B300  uniB300,  HANGUL SYLLABLE-B300
        0xB301, # G+1537  U+B301  uniB301,  HANGUL SYLLABLE-B301
        0xB304, # G+1538  U+B304  uniB304,  HANGUL SYLLABLE-B304
        0xB308, # G+1539  U+B308  uniB308,  HANGUL SYLLABLE-B308
        0xB310, # G+1540  U+B310  uniB310,  HANGUL SYLLABLE-B310
        0xB311, # G+1541  U+B311  uniB311,  HANGUL SYLLABLE-B311
        0xB313, # G+1542  U+B313  uniB313,  HANGUL SYLLABLE-B313
        0xB314, # G+1543  U+B314  uniB314,  HANGUL SYLLABLE-B314
        0xB315, # G+1544  U+B315  uniB315,  HANGUL SYLLABLE-B315
        0xB31C, # G+1545  U+B31C  uniB31C,  HANGUL SYLLABLE-B31C
        0xB354, # G+1546  U+B354  uniB354,  HANGUL SYLLABLE-B354
        0xB355, # G+1547  U+B355  uniB355,  HANGUL SYLLABLE-B355
        0xB356, # G+1548  U+B356  uniB356,  HANGUL SYLLABLE-B356
        0xB358, # G+1549  U+B358  uniB358,  HANGUL SYLLABLE-B358
        0xB35B, # G+1550  U+B35B  uniB35B,  HANGUL SYLLABLE-B35B
        0xB35C, # G+1551  U+B35C  uniB35C,  HANGUL SYLLABLE-B35C
        0xB35E, # G+1552  U+B35E  uniB35E,  HANGUL SYLLABLE-B35E
        0xB35F, # G+1553  U+B35F  uniB35F,  HANGUL SYLLABLE-B35F
        0xB364, # G+1554  U+B364  uniB364,  HANGUL SYLLABLE-B364
        0xB365, # G+1555  U+B365  uniB365,  HANGUL SYLLABLE-B365
        0xB367, # G+1556  U+B367  uniB367,  HANGUL SYLLABLE-B367
        0xB369, # G+1557  U+B369  uniB369,  HANGUL SYLLABLE-B369
        0xB36B, # G+1558  U+B36B  uniB36B,  HANGUL SYLLABLE-B36B
        0xB36E, # G+1559  U+B36E  uniB36E,  HANGUL SYLLABLE-B36E
        0xB370, # G+1560  U+B370  uniB370,  HANGUL SYLLABLE-B370
        0xB371, # G+1561  U+B371  uniB371,  HANGUL SYLLABLE-B371
        0xB374, # G+1562  U+B374  uniB374,  HANGUL SYLLABLE-B374
        0xB378, # G+1563  U+B378  uniB378,  HANGUL SYLLABLE-B378
        0xB380, # G+1564  U+B380  uniB380,  HANGUL SYLLABLE-B380
        0xB381, # G+1565  U+B381  uniB381,  HANGUL SYLLABLE-B381
        0xB383, # G+1566  U+B383  uniB383,  HANGUL SYLLABLE-B383
        0xB384, # G+1567  U+B384  uniB384,  HANGUL SYLLABLE-B384
        0xB385, # G+1568  U+B385  uniB385,  HANGUL SYLLABLE-B385
        0xB38C, # G+1569  U+B38C  uniB38C,  HANGUL SYLLABLE-B38C
        0xB390, # G+1570  U+B390  uniB390,  HANGUL SYLLABLE-B390
        0xB394, # G+1571  U+B394  uniB394,  HANGUL SYLLABLE-B394
        0xB3A0, # G+1572  U+B3A0  uniB3A0,  HANGUL SYLLABLE-B3A0
        0xB3A1, # G+1573  U+B3A1  uniB3A1,  HANGUL SYLLABLE-B3A1
        0xB3A8, # G+1574  U+B3A8  uniB3A8,  HANGUL SYLLABLE-B3A8
        0xB3AC, # G+1575  U+B3AC  uniB3AC,  HANGUL SYLLABLE-B3AC
        0xB3C4, # G+1576  U+B3C4  uniB3C4,  HANGUL SYLLABLE-B3C4
        0xB3C5, # G+1577  U+B3C5  uniB3C5,  HANGUL SYLLABLE-B3C5
        0xB3C8, # G+1578  U+B3C8  uniB3C8,  HANGUL SYLLABLE-B3C8
        0xB3CB, # G+1579  U+B3CB  uniB3CB,  HANGUL SYLLABLE-B3CB
        0xB3CC, # G+1580  U+B3CC  uniB3CC,  HANGUL SYLLABLE-B3CC
        0xB3CE, # G+1581  U+B3CE  uniB3CE,  HANGUL SYLLABLE-B3CE
        0xB3D0, # G+1582  U+B3D0  uniB3D0,  HANGUL SYLLABLE-B3D0
        0xB3D4, # G+1583  U+B3D4  uniB3D4,  HANGUL SYLLABLE-B3D4
        0xB3D5, # G+1584  U+B3D5  uniB3D5,  HANGUL SYLLABLE-B3D5
        0xB3D7, # G+1585  U+B3D7  uniB3D7,  HANGUL SYLLABLE-B3D7
        0xB3D9, # G+1586  U+B3D9  uniB3D9,  HANGUL SYLLABLE-B3D9
        0xB3DB, # G+1587  U+B3DB  uniB3DB,  HANGUL SYLLABLE-B3DB
        0xB3DD, # G+1588  U+B3DD  uniB3DD,  HANGUL SYLLABLE-B3DD
        0xB3E0, # G+1589  U+B3E0  uniB3E0,  HANGUL SYLLABLE-B3E0
        0xB3E4, # G+1590  U+B3E4  uniB3E4,  HANGUL SYLLABLE-B3E4
        0xB3E8, # G+1591  U+B3E8  uniB3E8,  HANGUL SYLLABLE-B3E8
        0xB3FC, # G+1592  U+B3FC  uniB3FC,  HANGUL SYLLABLE-B3FC
        0xB410, # G+1593  U+B410  uniB410,  HANGUL SYLLABLE-B410
        0xB418, # G+1594  U+B418  uniB418,  HANGUL SYLLABLE-B418
        0xB41C, # G+1595  U+B41C  uniB41C,  HANGUL SYLLABLE-B41C
        0xB420, # G+1596  U+B420  uniB420,  HANGUL SYLLABLE-B420
        0xB428, # G+1597  U+B428  uniB428,  HANGUL SYLLABLE-B428
        0xB429, # G+1598  U+B429  uniB429,  HANGUL SYLLABLE-B429
        0xB42B, # G+1599  U+B42B  uniB42B,  HANGUL SYLLABLE-B42B
        0xB434, # G+1600  U+B434  uniB434,  HANGUL SYLLABLE-B434
        0xB450, # G+1601  U+B450  uniB450,  HANGUL SYLLABLE-B450
        0xB451, # G+1602  U+B451  uniB451,  HANGUL SYLLABLE-B451
        0xB454, # G+1603  U+B454  uniB454,  HANGUL SYLLABLE-B454
        0xB458, # G+1604  U+B458  uniB458,  HANGUL SYLLABLE-B458
        0xB460, # G+1605  U+B460  uniB460,  HANGUL SYLLABLE-B460
        0xB461, # G+1606  U+B461  uniB461,  HANGUL SYLLABLE-B461
        0xB463, # G+1607  U+B463  uniB463,  HANGUL SYLLABLE-B463
        0xB465, # G+1608  U+B465  uniB465,  HANGUL SYLLABLE-B465
        0xB46C, # G+1609  U+B46C  uniB46C,  HANGUL SYLLABLE-B46C
        0xB480, # G+1610  U+B480  uniB480,  HANGUL SYLLABLE-B480
        0xB488, # G+1611  U+B488  uniB488,  HANGUL SYLLABLE-B488
        0xB49D, # G+1612  U+B49D  uniB49D,  HANGUL SYLLABLE-B49D
        0xB4A4, # G+1613  U+B4A4  uniB4A4,  HANGUL SYLLABLE-B4A4
        0xB4A8, # G+1614  U+B4A8  uniB4A8,  HANGUL SYLLABLE-B4A8
        0xB4AC, # G+1615  U+B4AC  uniB4AC,  HANGUL SYLLABLE-B4AC
        0xB4B5, # G+1616  U+B4B5  uniB4B5,  HANGUL SYLLABLE-B4B5
        0xB4B7, # G+1617  U+B4B7  uniB4B7,  HANGUL SYLLABLE-B4B7
        0xB4B9, # G+1618  U+B4B9  uniB4B9,  HANGUL SYLLABLE-B4B9
        0xB4C0, # G+1619  U+B4C0  uniB4C0,  HANGUL SYLLABLE-B4C0
        0xB4C4, # G+1620  U+B4C4  uniB4C4,  HANGUL SYLLABLE-B4C4
        0xB4C8, # G+1621  U+B4C8  uniB4C8,  HANGUL SYLLABLE-B4C8
        0xB4D0, # G+1622  U+B4D0  uniB4D0,  HANGUL SYLLABLE-B4D0
        0xB4D5, # G+1623  U+B4D5  uniB4D5,  HANGUL SYLLABLE-B4D5
        0xB4DC, # G+1624  U+B4DC  uniB4DC,  HANGUL SYLLABLE-B4DC
        0xB4DD, # G+1625  U+B4DD  uniB4DD,  HANGUL SYLLABLE-B4DD
        0xB4E0, # G+1626  U+B4E0  uniB4E0,  HANGUL SYLLABLE-B4E0
        0xB4E3, # G+1627  U+B4E3  uniB4E3,  HANGUL SYLLABLE-B4E3
        0xB4E4, # G+1628  U+B4E4  uniB4E4,  HANGUL SYLLABLE-B4E4
        0xB4E6, # G+1629  U+B4E6  uniB4E6,  HANGUL SYLLABLE-B4E6
        0xB4EC, # G+1630  U+B4EC  uniB4EC,  HANGUL SYLLABLE-B4EC
        0xB4ED, # G+1631  U+B4ED  uniB4ED,  HANGUL SYLLABLE-B4ED
        0xB4EF, # G+1632  U+B4EF  uniB4EF,  HANGUL SYLLABLE-B4EF
        0xB4F1, # G+1633  U+B4F1  uniB4F1,  HANGUL SYLLABLE-B4F1
        0xB4F8, # G+1634  U+B4F8  uniB4F8,  HANGUL SYLLABLE-B4F8
        0xB514, # G+1635  U+B514  uniB514,  HANGUL SYLLABLE-B514
        0xB515, # G+1636  U+B515  uniB515,  HANGUL SYLLABLE-B515
        0xB518, # G+1637  U+B518  uniB518,  HANGUL SYLLABLE-B518
        0xB51B, # G+1638  U+B51B  uniB51B,  HANGUL SYLLABLE-B51B
        0xB51C, # G+1639  U+B51C  uniB51C,  HANGUL SYLLABLE-B51C
        0xB524, # G+1640  U+B524  uniB524,  HANGUL SYLLABLE-B524
        0xB525, # G+1641  U+B525  uniB525,  HANGUL SYLLABLE-B525
        0xB527, # G+1642  U+B527  uniB527,  HANGUL SYLLABLE-B527
        0xB528, # G+1643  U+B528  uniB528,  HANGUL SYLLABLE-B528
        0xB529, # G+1644  U+B529  uniB529,  HANGUL SYLLABLE-B529
        0xB52A, # G+1645  U+B52A  uniB52A,  HANGUL SYLLABLE-B52A
        0xB530, # G+1646  U+B530  uniB530,  HANGUL SYLLABLE-B530
        0xB531, # G+1647  U+B531  uniB531,  HANGUL SYLLABLE-B531
        0xB534, # G+1648  U+B534  uniB534,  HANGUL SYLLABLE-B534
        0xB538, # G+1649  U+B538  uniB538,  HANGUL SYLLABLE-B538
        0xB540, # G+1650  U+B540  uniB540,  HANGUL SYLLABLE-B540
        0xB541, # G+1651  U+B541  uniB541,  HANGUL SYLLABLE-B541
        0xB543, # G+1652  U+B543  uniB543,  HANGUL SYLLABLE-B543
        0xB544, # G+1653  U+B544  uniB544,  HANGUL SYLLABLE-B544
        0xB545, # G+1654  U+B545  uniB545,  HANGUL SYLLABLE-B545
        0xB54B, # G+1655  U+B54B  uniB54B,  HANGUL SYLLABLE-B54B
        0xB54C, # G+1656  U+B54C  uniB54C,  HANGUL SYLLABLE-B54C
        0xB54D, # G+1657  U+B54D  uniB54D,  HANGUL SYLLABLE-B54D
        0xB550, # G+1658  U+B550  uniB550,  HANGUL SYLLABLE-B550
        0xB554, # G+1659  U+B554  uniB554,  HANGUL SYLLABLE-B554
        0xB55C, # G+1660  U+B55C  uniB55C,  HANGUL SYLLABLE-B55C
        0xB55D, # G+1661  U+B55D  uniB55D,  HANGUL SYLLABLE-B55D
        0xB55F, # G+1662  U+B55F  uniB55F,  HANGUL SYLLABLE-B55F
        0xB560, # G+1663  U+B560  uniB560,  HANGUL SYLLABLE-B560
        0xB561, # G+1664  U+B561  uniB561,  HANGUL SYLLABLE-B561
        0xB5A0, # G+1665  U+B5A0  uniB5A0,  HANGUL SYLLABLE-B5A0
        0xB5A1, # G+1666  U+B5A1  uniB5A1,  HANGUL SYLLABLE-B5A1
        0xB5A4, # G+1667  U+B5A4  uniB5A4,  HANGUL SYLLABLE-B5A4
        0xB5A8, # G+1668  U+B5A8  uniB5A8,  HANGUL SYLLABLE-B5A8
        0xB5AA, # G+1669  U+B5AA  uniB5AA,  HANGUL SYLLABLE-B5AA
        0xB5AB, # G+1670  U+B5AB  uniB5AB,  HANGUL SYLLABLE-B5AB
        0xB5B0, # G+1671  U+B5B0  uniB5B0,  HANGUL SYLLABLE-B5B0
        0xB5B1, # G+1672  U+B5B1  uniB5B1,  HANGUL SYLLABLE-B5B1
        0xB5B3, # G+1673  U+B5B3  uniB5B3,  HANGUL SYLLABLE-B5B3
        0xB5B4, # G+1674  U+B5B4  uniB5B4,  HANGUL SYLLABLE-B5B4
        0xB5B5, # G+1675  U+B5B5  uniB5B5,  HANGUL SYLLABLE-B5B5
        0xB5BB, # G+1676  U+B5BB  uniB5BB,  HANGUL SYLLABLE-B5BB
        0xB5BC, # G+1677  U+B5BC  uniB5BC,  HANGUL SYLLABLE-B5BC
        0xB5BD, # G+1678  U+B5BD  uniB5BD,  HANGUL SYLLABLE-B5BD
        0xB5C0, # G+1679  U+B5C0  uniB5C0,  HANGUL SYLLABLE-B5C0
        0xB5C4, # G+1680  U+B5C4  uniB5C4,  HANGUL SYLLABLE-B5C4
        0xB5CC, # G+1681  U+B5CC  uniB5CC,  HANGUL SYLLABLE-B5CC
        0xB5CD, # G+1682  U+B5CD  uniB5CD,  HANGUL SYLLABLE-B5CD
        0xB5CF, # G+1683  U+B5CF  uniB5CF,  HANGUL SYLLABLE-B5CF
        0xB5D0, # G+1684  U+B5D0  uniB5D0,  HANGUL SYLLABLE-B5D0
        0xB5D1, # G+1685  U+B5D1  uniB5D1,  HANGUL SYLLABLE-B5D1
        0xB5D8, # G+1686  U+B5D8  uniB5D8,  HANGUL SYLLABLE-B5D8
        0xB5EC, # G+1687  U+B5EC  uniB5EC,  HANGUL SYLLABLE-B5EC
        0xB610, # G+1688  U+B610  uniB610,  HANGUL SYLLABLE-B610
        0xB611, # G+1689  U+B611  uniB611,  HANGUL SYLLABLE-B611
        0xB614, # G+1690  U+B614  uniB614,  HANGUL SYLLABLE-B614
        0xB618, # G+1691  U+B618  uniB618,  HANGUL SYLLABLE-B618
        0xB625, # G+1692  U+B625  uniB625,  HANGUL SYLLABLE-B625
        0xB62C, # G+1693  U+B62C  uniB62C,  HANGUL SYLLABLE-B62C
        0xB634, # G+1694  U+B634  uniB634,  HANGUL SYLLABLE-B634
        0xB648, # G+1695  U+B648  uniB648,  HANGUL SYLLABLE-B648
        0xB664, # G+1696  U+B664  uniB664,  HANGUL SYLLABLE-B664
        0xB668, # G+1697  U+B668  uniB668,  HANGUL SYLLABLE-B668
        0xB69C, # G+1698  U+B69C  uniB69C,  HANGUL SYLLABLE-B69C
        0xB69D, # G+1699  U+B69D  uniB69D,  HANGUL SYLLABLE-B69D
        0xB6A0, # G+1700  U+B6A0  uniB6A0,  HANGUL SYLLABLE-B6A0
        0xB6A4, # G+1701  U+B6A4  uniB6A4,  HANGUL SYLLABLE-B6A4
        0xB6AB, # G+1702  U+B6AB  uniB6AB,  HANGUL SYLLABLE-B6AB
        0xB6AC, # G+1703  U+B6AC  uniB6AC,  HANGUL SYLLABLE-B6AC
        0xB6B1, # G+1704  U+B6B1  uniB6B1,  HANGUL SYLLABLE-B6B1
        0xB6D4, # G+1705  U+B6D4  uniB6D4,  HANGUL SYLLABLE-B6D4
        0xB6F0, # G+1706  U+B6F0  uniB6F0,  HANGUL SYLLABLE-B6F0
        0xB6F4, # G+1707  U+B6F4  uniB6F4,  HANGUL SYLLABLE-B6F4
        0xB6F8, # G+1708  U+B6F8  uniB6F8,  HANGUL SYLLABLE-B6F8
        0xB700, # G+1709  U+B700  uniB700,  HANGUL SYLLABLE-B700
        0xB701, # G+1710  U+B701  uniB701,  HANGUL SYLLABLE-B701
        0xB705, # G+1711  U+B705  uniB705,  HANGUL SYLLABLE-B705
        0xB728, # G+1712  U+B728  uniB728,  HANGUL SYLLABLE-B728
        0xB729, # G+1713  U+B729  uniB729,  HANGUL SYLLABLE-B729
        0xB72C, # G+1714  U+B72C  uniB72C,  HANGUL SYLLABLE-B72C
        0xB72F, # G+1715  U+B72F  uniB72F,  HANGUL SYLLABLE-B72F
        0xB730, # G+1716  U+B730  uniB730,  HANGUL SYLLABLE-B730
        0xB738, # G+1717  U+B738  uniB738,  HANGUL SYLLABLE-B738
        0xB739, # G+1718  U+B739  uniB739,  HANGUL SYLLABLE-B739
        0xB73B, # G+1719  U+B73B  uniB73B,  HANGUL SYLLABLE-B73B
        0xB744, # G+1720  U+B744  uniB744,  HANGUL SYLLABLE-B744
        0xB748, # G+1721  U+B748  uniB748,  HANGUL SYLLABLE-B748
        0xB74C, # G+1722  U+B74C  uniB74C,  HANGUL SYLLABLE-B74C
        0xB754, # G+1723  U+B754  uniB754,  HANGUL SYLLABLE-B754
        0xB755, # G+1724  U+B755  uniB755,  HANGUL SYLLABLE-B755
        0xB760, # G+1725  U+B760  uniB760,  HANGUL SYLLABLE-B760
        0xB764, # G+1726  U+B764  uniB764,  HANGUL SYLLABLE-B764
        0xB768, # G+1727  U+B768  uniB768,  HANGUL SYLLABLE-B768
        0xB770, # G+1728  U+B770  uniB770,  HANGUL SYLLABLE-B770
        0xB771, # G+1729  U+B771  uniB771,  HANGUL SYLLABLE-B771
        0xB773, # G+1730  U+B773  uniB773,  HANGUL SYLLABLE-B773
        0xB775, # G+1731  U+B775  uniB775,  HANGUL SYLLABLE-B775
        0xB77C, # G+1732  U+B77C  uniB77C,  HANGUL SYLLABLE-B77C
        0xB77D, # G+1733  U+B77D  uniB77D,  HANGUL SYLLABLE-B77D
        0xB780, # G+1734  U+B780  uniB780,  HANGUL SYLLABLE-B780
        0xB784, # G+1735  U+B784  uniB784,  HANGUL SYLLABLE-B784
        0xB78C, # G+1736  U+B78C  uniB78C,  HANGUL SYLLABLE-B78C
        0xB78D, # G+1737  U+B78D  uniB78D,  HANGUL SYLLABLE-B78D
        0xB78F, # G+1738  U+B78F  uniB78F,  HANGUL SYLLABLE-B78F
        0xB790, # G+1739  U+B790  uniB790,  HANGUL SYLLABLE-B790
        0xB791, # G+1740  U+B791  uniB791,  HANGUL SYLLABLE-B791
        0xB792, # G+1741  U+B792  uniB792,  HANGUL SYLLABLE-B792
        0xB796, # G+1742  U+B796  uniB796,  HANGUL SYLLABLE-B796
        0xB797, # G+1743  U+B797  uniB797,  HANGUL SYLLABLE-B797
        0xB798, # G+1744  U+B798  uniB798,  HANGUL SYLLABLE-B798
        0xB799, # G+1745  U+B799  uniB799,  HANGUL SYLLABLE-B799
        0xB79C, # G+1746  U+B79C  uniB79C,  HANGUL SYLLABLE-B79C
        0xB7A0, # G+1747  U+B7A0  uniB7A0,  HANGUL SYLLABLE-B7A0
        0xB7A8, # G+1748  U+B7A8  uniB7A8,  HANGUL SYLLABLE-B7A8
        0xB7A9, # G+1749  U+B7A9  uniB7A9,  HANGUL SYLLABLE-B7A9
        0xB7AB, # G+1750  U+B7AB  uniB7AB,  HANGUL SYLLABLE-B7AB
        0xB7AC, # G+1751  U+B7AC  uniB7AC,  HANGUL SYLLABLE-B7AC
        0xB7AD, # G+1752  U+B7AD  uniB7AD,  HANGUL SYLLABLE-B7AD
        0xB7B4, # G+1753  U+B7B4  uniB7B4,  HANGUL SYLLABLE-B7B4
        0xB7B5, # G+1754  U+B7B5  uniB7B5,  HANGUL SYLLABLE-B7B5
        0xB7B8, # G+1755  U+B7B8  uniB7B8,  HANGUL SYLLABLE-B7B8
        0xB7C7, # G+1756  U+B7C7  uniB7C7,  HANGUL SYLLABLE-B7C7
        0xB7C9, # G+1757  U+B7C9  uniB7C9,  HANGUL SYLLABLE-B7C9
        0xB7EC, # G+1758  U+B7EC  uniB7EC,  HANGUL SYLLABLE-B7EC
        0xB7ED, # G+1759  U+B7ED  uniB7ED,  HANGUL SYLLABLE-B7ED
        0xB7F0, # G+1760  U+B7F0  uniB7F0,  HANGUL SYLLABLE-B7F0
        0xB7F4, # G+1761  U+B7F4  uniB7F4,  HANGUL SYLLABLE-B7F4
        0xB7FC, # G+1762  U+B7FC  uniB7FC,  HANGUL SYLLABLE-B7FC
        0xB7FD, # G+1763  U+B7FD  uniB7FD,  HANGUL SYLLABLE-B7FD
        0xB7FF, # G+1764  U+B7FF  uniB7FF,  HANGUL SYLLABLE-B7FF
        0xB800, # G+1765  U+B800  uniB800,  HANGUL SYLLABLE-B800
        0xB801, # G+1766  U+B801  uniB801,  HANGUL SYLLABLE-B801
        0xB807, # G+1767  U+B807  uniB807,  HANGUL SYLLABLE-B807
        0xB808, # G+1768  U+B808  uniB808,  HANGUL SYLLABLE-B808
        0xB809, # G+1769  U+B809  uniB809,  HANGUL SYLLABLE-B809
        0xB80C, # G+1770  U+B80C  uniB80C,  HANGUL SYLLABLE-B80C
        0xB810, # G+1771  U+B810  uniB810,  HANGUL SYLLABLE-B810
        0xB818, # G+1772  U+B818  uniB818,  HANGUL SYLLABLE-B818
        0xB819, # G+1773  U+B819  uniB819,  HANGUL SYLLABLE-B819
        0xB81B, # G+1774  U+B81B  uniB81B,  HANGUL SYLLABLE-B81B
        0xB81D, # G+1775  U+B81D  uniB81D,  HANGUL SYLLABLE-B81D
        0xB824, # G+1776  U+B824  uniB824,  HANGUL SYLLABLE-B824
        0xB825, # G+1777  U+B825  uniB825,  HANGUL SYLLABLE-B825
        0xB828, # G+1778  U+B828  uniB828,  HANGUL SYLLABLE-B828
        0xB82C, # G+1779  U+B82C  uniB82C,  HANGUL SYLLABLE-B82C
        0xB834, # G+1780  U+B834  uniB834,  HANGUL SYLLABLE-B834
        0xB835, # G+1781  U+B835  uniB835,  HANGUL SYLLABLE-B835
        0xB837, # G+1782  U+B837  uniB837,  HANGUL SYLLABLE-B837
        0xB838, # G+1783  U+B838  uniB838,  HANGUL SYLLABLE-B838
        0xB839, # G+1784  U+B839  uniB839,  HANGUL SYLLABLE-B839
        0xB840, # G+1785  U+B840  uniB840,  HANGUL SYLLABLE-B840
        0xB844, # G+1786  U+B844  uniB844,  HANGUL SYLLABLE-B844
        0xB851, # G+1787  U+B851  uniB851,  HANGUL SYLLABLE-B851
        0xB853, # G+1788  U+B853  uniB853,  HANGUL SYLLABLE-B853
        0xB85C, # G+1789  U+B85C  uniB85C,  HANGUL SYLLABLE-B85C
        0xB85D, # G+1790  U+B85D  uniB85D,  HANGUL SYLLABLE-B85D
        0xB860, # G+1791  U+B860  uniB860,  HANGUL SYLLABLE-B860
        0xB864, # G+1792  U+B864  uniB864,  HANGUL SYLLABLE-B864
        0xB86C, # G+1793  U+B86C  uniB86C,  HANGUL SYLLABLE-B86C
        0xB86D, # G+1794  U+B86D  uniB86D,  HANGUL SYLLABLE-B86D
        0xB86F, # G+1795  U+B86F  uniB86F,  HANGUL SYLLABLE-B86F
        0xB871, # G+1796  U+B871  uniB871,  HANGUL SYLLABLE-B871
        0xB878, # G+1797  U+B878  uniB878,  HANGUL SYLLABLE-B878
        0xB87C, # G+1798  U+B87C  uniB87C,  HANGUL SYLLABLE-B87C
        0xB88D, # G+1799  U+B88D  uniB88D,  HANGUL SYLLABLE-B88D
        0xB8A8, # G+1800  U+B8A8  uniB8A8,  HANGUL SYLLABLE-B8A8
        0xB8B0, # G+1801  U+B8B0  uniB8B0,  HANGUL SYLLABLE-B8B0
        0xB8B4, # G+1802  U+B8B4  uniB8B4,  HANGUL SYLLABLE-B8B4
        0xB8B8, # G+1803  U+B8B8  uniB8B8,  HANGUL SYLLABLE-B8B8
        0xB8C0, # G+1804  U+B8C0  uniB8C0,  HANGUL SYLLABLE-B8C0
        0xB8C1, # G+1805  U+B8C1  uniB8C1,  HANGUL SYLLABLE-B8C1
        0xB8C3, # G+1806  U+B8C3  uniB8C3,  HANGUL SYLLABLE-B8C3
        0xB8C5, # G+1807  U+B8C5  uniB8C5,  HANGUL SYLLABLE-B8C5
        0xB8CC, # G+1808  U+B8CC  uniB8CC,  HANGUL SYLLABLE-B8CC
        0xB8D0, # G+1809  U+B8D0  uniB8D0,  HANGUL SYLLABLE-B8D0
        0xB8D4, # G+1810  U+B8D4  uniB8D4,  HANGUL SYLLABLE-B8D4
        0xB8DD, # G+1811  U+B8DD  uniB8DD,  HANGUL SYLLABLE-B8DD
        0xB8DF, # G+1812  U+B8DF  uniB8DF,  HANGUL SYLLABLE-B8DF
        0xB8E1, # G+1813  U+B8E1  uniB8E1,  HANGUL SYLLABLE-B8E1
        0xB8E8, # G+1814  U+B8E8  uniB8E8,  HANGUL SYLLABLE-B8E8
        0xB8E9, # G+1815  U+B8E9  uniB8E9,  HANGUL SYLLABLE-B8E9
        0xB8EC, # G+1816  U+B8EC  uniB8EC,  HANGUL SYLLABLE-B8EC
        0xB8F0, # G+1817  U+B8F0  uniB8F0,  HANGUL SYLLABLE-B8F0
        0xB8F8, # G+1818  U+B8F8  uniB8F8,  HANGUL SYLLABLE-B8F8
        0xB8F9, # G+1819  U+B8F9  uniB8F9,  HANGUL SYLLABLE-B8F9
        0xB8FB, # G+1820  U+B8FB  uniB8FB,  HANGUL SYLLABLE-B8FB
        0xB8FD, # G+1821  U+B8FD  uniB8FD,  HANGUL SYLLABLE-B8FD
        0xB904, # G+1822  U+B904  uniB904,  HANGUL SYLLABLE-B904
        0xB918, # G+1823  U+B918  uniB918,  HANGUL SYLLABLE-B918
        0xB920, # G+1824  U+B920  uniB920,  HANGUL SYLLABLE-B920
        0xB93C, # G+1825  U+B93C  uniB93C,  HANGUL SYLLABLE-B93C
        0xB93D, # G+1826  U+B93D  uniB93D,  HANGUL SYLLABLE-B93D
        0xB940, # G+1827  U+B940  uniB940,  HANGUL SYLLABLE-B940
        0xB944, # G+1828  U+B944  uniB944,  HANGUL SYLLABLE-B944
        0xB94C, # G+1829  U+B94C  uniB94C,  HANGUL SYLLABLE-B94C
        0xB94F, # G+1830  U+B94F  uniB94F,  HANGUL SYLLABLE-B94F
        0xB951, # G+1831  U+B951  uniB951,  HANGUL SYLLABLE-B951
        0xB958, # G+1832  U+B958  uniB958,  HANGUL SYLLABLE-B958
        0xB959, # G+1833  U+B959  uniB959,  HANGUL SYLLABLE-B959
        0xB95C, # G+1834  U+B95C  uniB95C,  HANGUL SYLLABLE-B95C
        0xB960, # G+1835  U+B960  uniB960,  HANGUL SYLLABLE-B960
        0xB968, # G+1836  U+B968  uniB968,  HANGUL SYLLABLE-B968
        0xB969, # G+1837  U+B969  uniB969,  HANGUL SYLLABLE-B969
        0xB96B, # G+1838  U+B96B  uniB96B,  HANGUL SYLLABLE-B96B
        0xB96D, # G+1839  U+B96D  uniB96D,  HANGUL SYLLABLE-B96D
        0xB974, # G+1840  U+B974  uniB974,  HANGUL SYLLABLE-B974
        0xB975, # G+1841  U+B975  uniB975,  HANGUL SYLLABLE-B975
        0xB978, # G+1842  U+B978  uniB978,  HANGUL SYLLABLE-B978
        0xB97C, # G+1843  U+B97C  uniB97C,  HANGUL SYLLABLE-B97C
        0xB984, # G+1844  U+B984  uniB984,  HANGUL SYLLABLE-B984
        0xB985, # G+1845  U+B985  uniB985,  HANGUL SYLLABLE-B985
        0xB987, # G+1846  U+B987  uniB987,  HANGUL SYLLABLE-B987
        0xB989, # G+1847  U+B989  uniB989,  HANGUL SYLLABLE-B989
        0xB98A, # G+1848  U+B98A  uniB98A,  HANGUL SYLLABLE-B98A
        0xB98D, # G+1849  U+B98D  uniB98D,  HANGUL SYLLABLE-B98D
        0xB98E, # G+1850  U+B98E  uniB98E,  HANGUL SYLLABLE-B98E
        0xB9AC, # G+1851  U+B9AC  uniB9AC,  HANGUL SYLLABLE-B9AC
        0xB9AD, # G+1852  U+B9AD  uniB9AD,  HANGUL SYLLABLE-B9AD
        0xB9B0, # G+1853  U+B9B0  uniB9B0,  HANGUL SYLLABLE-B9B0
        0xB9B4, # G+1854  U+B9B4  uniB9B4,  HANGUL SYLLABLE-B9B4
        0xB9BC, # G+1855  U+B9BC  uniB9BC,  HANGUL SYLLABLE-B9BC
        0xB9BD, # G+1856  U+B9BD  uniB9BD,  HANGUL SYLLABLE-B9BD
        0xB9BF, # G+1857  U+B9BF  uniB9BF,  HANGUL SYLLABLE-B9BF
        0xB9C1, # G+1858  U+B9C1  uniB9C1,  HANGUL SYLLABLE-B9C1
        0xB9C8, # G+1859  U+B9C8  uniB9C8,  HANGUL SYLLABLE-B9C8
        0xB9C9, # G+1860  U+B9C9  uniB9C9,  HANGUL SYLLABLE-B9C9
        0xB9CC, # G+1861  U+B9CC  uniB9CC,  HANGUL SYLLABLE-B9CC
        0xB9CE, # G+1862  U+B9CE  uniB9CE,  HANGUL SYLLABLE-B9CE
        0xB9CF, # G+1863  U+B9CF  uniB9CF,  HANGUL SYLLABLE-B9CF
        0xB9D0, # G+1864  U+B9D0  uniB9D0,  HANGUL SYLLABLE-B9D0
        0xB9D1, # G+1865  U+B9D1  uniB9D1,  HANGUL SYLLABLE-B9D1
        0xB9D2, # G+1866  U+B9D2  uniB9D2,  HANGUL SYLLABLE-B9D2
        0xB9D8, # G+1867  U+B9D8  uniB9D8,  HANGUL SYLLABLE-B9D8
        0xB9D9, # G+1868  U+B9D9  uniB9D9,  HANGUL SYLLABLE-B9D9
        0xB9DB, # G+1869  U+B9DB  uniB9DB,  HANGUL SYLLABLE-B9DB
        0xB9DD, # G+1870  U+B9DD  uniB9DD,  HANGUL SYLLABLE-B9DD
        0xB9DE, # G+1871  U+B9DE  uniB9DE,  HANGUL SYLLABLE-B9DE
        0xB9E1, # G+1872  U+B9E1  uniB9E1,  HANGUL SYLLABLE-B9E1
        0xB9E3, # G+1873  U+B9E3  uniB9E3,  HANGUL SYLLABLE-B9E3
        0xB9E4, # G+1874  U+B9E4  uniB9E4,  HANGUL SYLLABLE-B9E4
        0xB9E5, # G+1875  U+B9E5  uniB9E5,  HANGUL SYLLABLE-B9E5
        0xB9E8, # G+1876  U+B9E8  uniB9E8,  HANGUL SYLLABLE-B9E8
        0xB9EC, # G+1877  U+B9EC  uniB9EC,  HANGUL SYLLABLE-B9EC
        0xB9F4, # G+1878  U+B9F4  uniB9F4,  HANGUL SYLLABLE-B9F4
        0xB9F5, # G+1879  U+B9F5  uniB9F5,  HANGUL SYLLABLE-B9F5
        0xB9F7, # G+1880  U+B9F7  uniB9F7,  HANGUL SYLLABLE-B9F7
        0xB9F8, # G+1881  U+B9F8  uniB9F8,  HANGUL SYLLABLE-B9F8
        0xB9F9, # G+1882  U+B9F9  uniB9F9,  HANGUL SYLLABLE-B9F9
        0xB9FA, # G+1883  U+B9FA  uniB9FA,  HANGUL SYLLABLE-B9FA
        0xBA00, # G+1884  U+BA00  uniBA00,  HANGUL SYLLABLE-BA00
        0xBA01, # G+1885  U+BA01  uniBA01,  HANGUL SYLLABLE-BA01
        0xBA08, # G+1886  U+BA08  uniBA08,  HANGUL SYLLABLE-BA08
        0xBA15, # G+1887  U+BA15  uniBA15,  HANGUL SYLLABLE-BA15
        0xBA38, # G+1888  U+BA38  uniBA38,  HANGUL SYLLABLE-BA38
        0xBA39, # G+1889  U+BA39  uniBA39,  HANGUL SYLLABLE-BA39
        0xBA3C, # G+1890  U+BA3C  uniBA3C,  HANGUL SYLLABLE-BA3C
        0xBA40, # G+1891  U+BA40  uniBA40,  HANGUL SYLLABLE-BA40
        0xBA42, # G+1892  U+BA42  uniBA42,  HANGUL SYLLABLE-BA42
        0xBA48, # G+1893  U+BA48  uniBA48,  HANGUL SYLLABLE-BA48
        0xBA49, # G+1894  U+BA49  uniBA49,  HANGUL SYLLABLE-BA49
        0xBA4B, # G+1895  U+BA4B  uniBA4B,  HANGUL SYLLABLE-BA4B
        0xBA4D, # G+1896  U+BA4D  uniBA4D,  HANGUL SYLLABLE-BA4D
        0xBA4E, # G+1897  U+BA4E  uniBA4E,  HANGUL SYLLABLE-BA4E
        0xBA53, # G+1898  U+BA53  uniBA53,  HANGUL SYLLABLE-BA53
        0xBA54, # G+1899  U+BA54  uniBA54,  HANGUL SYLLABLE-BA54
        0xBA55, # G+1900  U+BA55  uniBA55,  HANGUL SYLLABLE-BA55
        0xBA58, # G+1901  U+BA58  uniBA58,  HANGUL SYLLABLE-BA58
        0xBA5C, # G+1902  U+BA5C  uniBA5C,  HANGUL SYLLABLE-BA5C
        0xBA64, # G+1903  U+BA64  uniBA64,  HANGUL SYLLABLE-BA64
        0xBA65, # G+1904  U+BA65  uniBA65,  HANGUL SYLLABLE-BA65
        0xBA67, # G+1905  U+BA67  uniBA67,  HANGUL SYLLABLE-BA67
        0xBA68, # G+1906  U+BA68  uniBA68,  HANGUL SYLLABLE-BA68
        0xBA69, # G+1907  U+BA69  uniBA69,  HANGUL SYLLABLE-BA69
        0xBA70, # G+1908  U+BA70  uniBA70,  HANGUL SYLLABLE-BA70
        0xBA71, # G+1909  U+BA71  uniBA71,  HANGUL SYLLABLE-BA71
        0xBA74, # G+1910  U+BA74  uniBA74,  HANGUL SYLLABLE-BA74
        0xBA78, # G+1911  U+BA78  uniBA78,  HANGUL SYLLABLE-BA78
        0xBA83, # G+1912  U+BA83  uniBA83,  HANGUL SYLLABLE-BA83
        0xBA84, # G+1913  U+BA84  uniBA84,  HANGUL SYLLABLE-BA84
        0xBA85, # G+1914  U+BA85  uniBA85,  HANGUL SYLLABLE-BA85
        0xBA87, # G+1915  U+BA87  uniBA87,  HANGUL SYLLABLE-BA87
        0xBA8C, # G+1916  U+BA8C  uniBA8C,  HANGUL SYLLABLE-BA8C
        0xBAA8, # G+1917  U+BAA8  uniBAA8,  HANGUL SYLLABLE-BAA8
        0xBAA9, # G+1918  U+BAA9  uniBAA9,  HANGUL SYLLABLE-BAA9
        0xBAAB, # G+1919  U+BAAB  uniBAAB,  HANGUL SYLLABLE-BAAB
        0xBAAC, # G+1920  U+BAAC  uniBAAC,  HANGUL SYLLABLE-BAAC
        0xBAB0, # G+1921  U+BAB0  uniBAB0,  HANGUL SYLLABLE-BAB0
        0xBAB2, # G+1922  U+BAB2  uniBAB2,  HANGUL SYLLABLE-BAB2
        0xBAB8, # G+1923  U+BAB8  uniBAB8,  HANGUL SYLLABLE-BAB8
        0xBAB9, # G+1924  U+BAB9  uniBAB9,  HANGUL SYLLABLE-BAB9
        0xBABB, # G+1925  U+BABB  uniBABB,  HANGUL SYLLABLE-BABB
        0xBABD, # G+1926  U+BABD  uniBABD,  HANGUL SYLLABLE-BABD
        0xBAC4, # G+1927  U+BAC4  uniBAC4,  HANGUL SYLLABLE-BAC4
        0xBAC8, # G+1928  U+BAC8  uniBAC8,  HANGUL SYLLABLE-BAC8
        0xBAD8, # G+1929  U+BAD8  uniBAD8,  HANGUL SYLLABLE-BAD8
        0xBAD9, # G+1930  U+BAD9  uniBAD9,  HANGUL SYLLABLE-BAD9
        0xBAFC, # G+1931  U+BAFC  uniBAFC,  HANGUL SYLLABLE-BAFC
        0xBB00, # G+1932  U+BB00  uniBB00,  HANGUL SYLLABLE-BB00
        0xBB04, # G+1933  U+BB04  uniBB04,  HANGUL SYLLABLE-BB04
        0xBB0D, # G+1934  U+BB0D  uniBB0D,  HANGUL SYLLABLE-BB0D
        0xBB0F, # G+1935  U+BB0F  uniBB0F,  HANGUL SYLLABLE-BB0F
        0xBB11, # G+1936  U+BB11  uniBB11,  HANGUL SYLLABLE-BB11
        0xBB18, # G+1937  U+BB18  uniBB18,  HANGUL SYLLABLE-BB18
        0xBB1C, # G+1938  U+BB1C  uniBB1C,  HANGUL SYLLABLE-BB1C
        0xBB20, # G+1939  U+BB20  uniBB20,  HANGUL SYLLABLE-BB20
        0xBB29, # G+1940  U+BB29  uniBB29,  HANGUL SYLLABLE-BB29
        0xBB2B, # G+1941  U+BB2B  uniBB2B,  HANGUL SYLLABLE-BB2B
        0xBB34, # G+1942  U+BB34  uniBB34,  HANGUL SYLLABLE-BB34
        0xBB35, # G+1943  U+BB35  uniBB35,  HANGUL SYLLABLE-BB35
        0xBB36, # G+1944  U+BB36  uniBB36,  HANGUL SYLLABLE-BB36
        0xBB38, # G+1945  U+BB38  uniBB38,  HANGUL SYLLABLE-BB38
        0xBB3B, # G+1946  U+BB3B  uniBB3B,  HANGUL SYLLABLE-BB3B
        0xBB3C, # G+1947  U+BB3C  uniBB3C,  HANGUL SYLLABLE-BB3C
        0xBB3D, # G+1948  U+BB3D  uniBB3D,  HANGUL SYLLABLE-BB3D
        0xBB3E, # G+1949  U+BB3E  uniBB3E,  HANGUL SYLLABLE-BB3E
        0xBB44, # G+1950  U+BB44  uniBB44,  HANGUL SYLLABLE-BB44
        0xBB45, # G+1951  U+BB45  uniBB45,  HANGUL SYLLABLE-BB45
        0xBB47, # G+1952  U+BB47  uniBB47,  HANGUL SYLLABLE-BB47
        0xBB49, # G+1953  U+BB49  uniBB49,  HANGUL SYLLABLE-BB49
        0xBB4D, # G+1954  U+BB4D  uniBB4D,  HANGUL SYLLABLE-BB4D
        0xBB4F, # G+1955  U+BB4F  uniBB4F,  HANGUL SYLLABLE-BB4F
        0xBB50, # G+1956  U+BB50  uniBB50,  HANGUL SYLLABLE-BB50
        0xBB54, # G+1957  U+BB54  uniBB54,  HANGUL SYLLABLE-BB54
        0xBB58, # G+1958  U+BB58  uniBB58,  HANGUL SYLLABLE-BB58
        0xBB61, # G+1959  U+BB61  uniBB61,  HANGUL SYLLABLE-BB61
        0xBB63, # G+1960  U+BB63  uniBB63,  HANGUL SYLLABLE-BB63
        0xBB6C, # G+1961  U+BB6C  uniBB6C,  HANGUL SYLLABLE-BB6C
        0xBB88, # G+1962  U+BB88  uniBB88,  HANGUL SYLLABLE-BB88
        0xBB8C, # G+1963  U+BB8C  uniBB8C,  HANGUL SYLLABLE-BB8C
        0xBB90, # G+1964  U+BB90  uniBB90,  HANGUL SYLLABLE-BB90
        0xBBA4, # G+1965  U+BBA4  uniBBA4,  HANGUL SYLLABLE-BBA4
        0xBBA8, # G+1966  U+BBA8  uniBBA8,  HANGUL SYLLABLE-BBA8
        0xBBAC, # G+1967  U+BBAC  uniBBAC,  HANGUL SYLLABLE-BBAC
        0xBBB4, # G+1968  U+BBB4  uniBBB4,  HANGUL SYLLABLE-BBB4
        0xBBB7, # G+1969  U+BBB7  uniBBB7,  HANGUL SYLLABLE-BBB7
        0xBBC0, # G+1970  U+BBC0  uniBBC0,  HANGUL SYLLABLE-BBC0
        0xBBC4, # G+1971  U+BBC4  uniBBC4,  HANGUL SYLLABLE-BBC4
        0xBBC8, # G+1972  U+BBC8  uniBBC8,  HANGUL SYLLABLE-BBC8
        0xBBD0, # G+1973  U+BBD0  uniBBD0,  HANGUL SYLLABLE-BBD0
        0xBBD3, # G+1974  U+BBD3  uniBBD3,  HANGUL SYLLABLE-BBD3
        0xBBF8, # G+1975  U+BBF8  uniBBF8,  HANGUL SYLLABLE-BBF8
        0xBBF9, # G+1976  U+BBF9  uniBBF9,  HANGUL SYLLABLE-BBF9
        0xBBFC, # G+1977  U+BBFC  uniBBFC,  HANGUL SYLLABLE-BBFC
        0xBBFF, # G+1978  U+BBFF  uniBBFF,  HANGUL SYLLABLE-BBFF
        0xBC00, # G+1979  U+BC00  uniBC00,  HANGUL SYLLABLE-BC00
        0xBC02, # G+1980  U+BC02  uniBC02,  HANGUL SYLLABLE-BC02
        0xBC08, # G+1981  U+BC08  uniBC08,  HANGUL SYLLABLE-BC08
        0xBC09, # G+1982  U+BC09  uniBC09,  HANGUL SYLLABLE-BC09
        0xBC0B, # G+1983  U+BC0B  uniBC0B,  HANGUL SYLLABLE-BC0B
        0xBC0C, # G+1984  U+BC0C  uniBC0C,  HANGUL SYLLABLE-BC0C
        0xBC0D, # G+1985  U+BC0D  uniBC0D,  HANGUL SYLLABLE-BC0D
        0xBC0F, # G+1986  U+BC0F  uniBC0F,  HANGUL SYLLABLE-BC0F
        0xBC11, # G+1987  U+BC11  uniBC11,  HANGUL SYLLABLE-BC11
        0xBC14, # G+1988  U+BC14  uniBC14,  HANGUL SYLLABLE-BC14
        0xBC15, # G+1989  U+BC15  uniBC15,  HANGUL SYLLABLE-BC15
        0xBC16, # G+1990  U+BC16  uniBC16,  HANGUL SYLLABLE-BC16
        0xBC17, # G+1991  U+BC17  uniBC17,  HANGUL SYLLABLE-BC17
        0xBC18, # G+1992  U+BC18  uniBC18,  HANGUL SYLLABLE-BC18
        0xBC1B, # G+1993  U+BC1B  uniBC1B,  HANGUL SYLLABLE-BC1B
        0xBC1C, # G+1994  U+BC1C  uniBC1C,  HANGUL SYLLABLE-BC1C
        0xBC1D, # G+1995  U+BC1D  uniBC1D,  HANGUL SYLLABLE-BC1D
        0xBC1E, # G+1996  U+BC1E  uniBC1E,  HANGUL SYLLABLE-BC1E
        0xBC1F, # G+1997  U+BC1F  uniBC1F,  HANGUL SYLLABLE-BC1F
        0xBC24, # G+1998  U+BC24  uniBC24,  HANGUL SYLLABLE-BC24
        0xBC25, # G+1999  U+BC25  uniBC25,  HANGUL SYLLABLE-BC25
        0xBC27, # G+2000  U+BC27  uniBC27,  HANGUL SYLLABLE-BC27
        0xBC29, # G+2001  U+BC29  uniBC29,  HANGUL SYLLABLE-BC29
        0xBC2D, # G+2002  U+BC2D  uniBC2D,  HANGUL SYLLABLE-BC2D
        0xBC30, # G+2003  U+BC30  uniBC30,  HANGUL SYLLABLE-BC30
        0xBC31, # G+2004  U+BC31  uniBC31,  HANGUL SYLLABLE-BC31
        0xBC34, # G+2005  U+BC34  uniBC34,  HANGUL SYLLABLE-BC34
        0xBC38, # G+2006  U+BC38  uniBC38,  HANGUL SYLLABLE-BC38
        0xBC40, # G+2007  U+BC40  uniBC40,  HANGUL SYLLABLE-BC40
        0xBC41, # G+2008  U+BC41  uniBC41,  HANGUL SYLLABLE-BC41
        0xBC43, # G+2009  U+BC43  uniBC43,  HANGUL SYLLABLE-BC43
        0xBC44, # G+2010  U+BC44  uniBC44,  HANGUL SYLLABLE-BC44
        0xBC45, # G+2011  U+BC45  uniBC45,  HANGUL SYLLABLE-BC45
        0xBC49, # G+2012  U+BC49  uniBC49,  HANGUL SYLLABLE-BC49
        0xBC4C, # G+2013  U+BC4C  uniBC4C,  HANGUL SYLLABLE-BC4C
        0xBC4D, # G+2014  U+BC4D  uniBC4D,  HANGUL SYLLABLE-BC4D
        0xBC50, # G+2015  U+BC50  uniBC50,  HANGUL SYLLABLE-BC50
        0xBC5D, # G+2016  U+BC5D  uniBC5D,  HANGUL SYLLABLE-BC5D
        0xBC84, # G+2017  U+BC84  uniBC84,  HANGUL SYLLABLE-BC84
        0xBC85, # G+2018  U+BC85  uniBC85,  HANGUL SYLLABLE-BC85
        0xBC88, # G+2019  U+BC88  uniBC88,  HANGUL SYLLABLE-BC88
        0xBC8B, # G+2020  U+BC8B  uniBC8B,  HANGUL SYLLABLE-BC8B
        0xBC8C, # G+2021  U+BC8C  uniBC8C,  HANGUL SYLLABLE-BC8C
        0xBC8E, # G+2022  U+BC8E  uniBC8E,  HANGUL SYLLABLE-BC8E
        0xBC94, # G+2023  U+BC94  uniBC94,  HANGUL SYLLABLE-BC94
        0xBC95, # G+2024  U+BC95  uniBC95,  HANGUL SYLLABLE-BC95
        0xBC97, # G+2025  U+BC97  uniBC97,  HANGUL SYLLABLE-BC97
        0xBC99, # G+2026  U+BC99  uniBC99,  HANGUL SYLLABLE-BC99
        0xBC9A, # G+2027  U+BC9A  uniBC9A,  HANGUL SYLLABLE-BC9A
        0xBCA0, # G+2028  U+BCA0  uniBCA0,  HANGUL SYLLABLE-BCA0
        0xBCA1, # G+2029  U+BCA1  uniBCA1,  HANGUL SYLLABLE-BCA1
        0xBCA4, # G+2030  U+BCA4  uniBCA4,  HANGUL SYLLABLE-BCA4
        0xBCA7, # G+2031  U+BCA7  uniBCA7,  HANGUL SYLLABLE-BCA7
        0xBCA8, # G+2032  U+BCA8  uniBCA8,  HANGUL SYLLABLE-BCA8
        0xBCB0, # G+2033  U+BCB0  uniBCB0,  HANGUL SYLLABLE-BCB0
        0xBCB1, # G+2034  U+BCB1  uniBCB1,  HANGUL SYLLABLE-BCB1
        0xBCB3, # G+2035  U+BCB3  uniBCB3,  HANGUL SYLLABLE-BCB3
        0xBCB4, # G+2036  U+BCB4  uniBCB4,  HANGUL SYLLABLE-BCB4
        0xBCB5, # G+2037  U+BCB5  uniBCB5,  HANGUL SYLLABLE-BCB5
        0xBCBC, # G+2038  U+BCBC  uniBCBC,  HANGUL SYLLABLE-BCBC
        0xBCBD, # G+2039  U+BCBD  uniBCBD,  HANGUL SYLLABLE-BCBD
        0xBCC0, # G+2040  U+BCC0  uniBCC0,  HANGUL SYLLABLE-BCC0
        0xBCC4, # G+2041  U+BCC4  uniBCC4,  HANGUL SYLLABLE-BCC4
        0xBCCD, # G+2042  U+BCCD  uniBCCD,  HANGUL SYLLABLE-BCCD
        0xBCCF, # G+2043  U+BCCF  uniBCCF,  HANGUL SYLLABLE-BCCF
        0xBCD0, # G+2044  U+BCD0  uniBCD0,  HANGUL SYLLABLE-BCD0
        0xBCD1, # G+2045  U+BCD1  uniBCD1,  HANGUL SYLLABLE-BCD1
        0xBCD5, # G+2046  U+BCD5  uniBCD5,  HANGUL SYLLABLE-BCD5
        0xBCD8, # G+2047  U+BCD8  uniBCD8,  HANGUL SYLLABLE-BCD8
        0xBCDC, # G+2048  U+BCDC  uniBCDC,  HANGUL SYLLABLE-BCDC
        0xBCF4, # G+2049  U+BCF4  uniBCF4,  HANGUL SYLLABLE-BCF4
        0xBCF5, # G+2050  U+BCF5  uniBCF5,  HANGUL SYLLABLE-BCF5
        0xBCF6, # G+2051  U+BCF6  uniBCF6,  HANGUL SYLLABLE-BCF6
        0xBCF8, # G+2052  U+BCF8  uniBCF8,  HANGUL SYLLABLE-BCF8
        0xBCFC, # G+2053  U+BCFC  uniBCFC,  HANGUL SYLLABLE-BCFC
        0xBD04, # G+2054  U+BD04  uniBD04,  HANGUL SYLLABLE-BD04
        0xBD05, # G+2055  U+BD05  uniBD05,  HANGUL SYLLABLE-BD05
        0xBD07, # G+2056  U+BD07  uniBD07,  HANGUL SYLLABLE-BD07
        0xBD09, # G+2057  U+BD09  uniBD09,  HANGUL SYLLABLE-BD09
        0xBD10, # G+2058  U+BD10  uniBD10,  HANGUL SYLLABLE-BD10
        0xBD14, # G+2059  U+BD14  uniBD14,  HANGUL SYLLABLE-BD14
        0xBD24, # G+2060  U+BD24  uniBD24,  HANGUL SYLLABLE-BD24
        0xBD2C, # G+2061  U+BD2C  uniBD2C,  HANGUL SYLLABLE-BD2C
        0xBD40, # G+2062  U+BD40  uniBD40,  HANGUL SYLLABLE-BD40
        0xBD48, # G+2063  U+BD48  uniBD48,  HANGUL SYLLABLE-BD48
        0xBD49, # G+2064  U+BD49  uniBD49,  HANGUL SYLLABLE-BD49
        0xBD4C, # G+2065  U+BD4C  uniBD4C,  HANGUL SYLLABLE-BD4C
        0xBD50, # G+2066  U+BD50  uniBD50,  HANGUL SYLLABLE-BD50
        0xBD58, # G+2067  U+BD58  uniBD58,  HANGUL SYLLABLE-BD58
        0xBD59, # G+2068  U+BD59  uniBD59,  HANGUL SYLLABLE-BD59
        0xBD64, # G+2069  U+BD64  uniBD64,  HANGUL SYLLABLE-BD64
        0xBD68, # G+2070  U+BD68  uniBD68,  HANGUL SYLLABLE-BD68
        0xBD80, # G+2071  U+BD80  uniBD80,  HANGUL SYLLABLE-BD80
        0xBD81, # G+2072  U+BD81  uniBD81,  HANGUL SYLLABLE-BD81
        0xBD84, # G+2073  U+BD84  uniBD84,  HANGUL SYLLABLE-BD84
        0xBD87, # G+2074  U+BD87  uniBD87,  HANGUL SYLLABLE-BD87
        0xBD88, # G+2075  U+BD88  uniBD88,  HANGUL SYLLABLE-BD88
        0xBD89, # G+2076  U+BD89  uniBD89,  HANGUL SYLLABLE-BD89
        0xBD8A, # G+2077  U+BD8A  uniBD8A,  HANGUL SYLLABLE-BD8A
        0xBD90, # G+2078  U+BD90  uniBD90,  HANGUL SYLLABLE-BD90
        0xBD91, # G+2079  U+BD91  uniBD91,  HANGUL SYLLABLE-BD91
        0xBD93, # G+2080  U+BD93  uniBD93,  HANGUL SYLLABLE-BD93
        0xBD95, # G+2081  U+BD95  uniBD95,  HANGUL SYLLABLE-BD95
        0xBD99, # G+2082  U+BD99  uniBD99,  HANGUL SYLLABLE-BD99
        0xBD9A, # G+2083  U+BD9A  uniBD9A,  HANGUL SYLLABLE-BD9A
        0xBD9C, # G+2084  U+BD9C  uniBD9C,  HANGUL SYLLABLE-BD9C
        0xBDA4, # G+2085  U+BDA4  uniBDA4,  HANGUL SYLLABLE-BDA4
        0xBDB0, # G+2086  U+BDB0  uniBDB0,  HANGUL SYLLABLE-BDB0
        0xBDB8, # G+2087  U+BDB8  uniBDB8,  HANGUL SYLLABLE-BDB8
        0xBDD4, # G+2088  U+BDD4  uniBDD4,  HANGUL SYLLABLE-BDD4
        0xBDD5, # G+2089  U+BDD5  uniBDD5,  HANGUL SYLLABLE-BDD5
        0xBDD8, # G+2090  U+BDD8  uniBDD8,  HANGUL SYLLABLE-BDD8
        0xBDDC, # G+2091  U+BDDC  uniBDDC,  HANGUL SYLLABLE-BDDC
        0xBDE9, # G+2092  U+BDE9  uniBDE9,  HANGUL SYLLABLE-BDE9
        0xBDF0, # G+2093  U+BDF0  uniBDF0,  HANGUL SYLLABLE-BDF0
        0xBDF4, # G+2094  U+BDF4  uniBDF4,  HANGUL SYLLABLE-BDF4
        0xBDF8, # G+2095  U+BDF8  uniBDF8,  HANGUL SYLLABLE-BDF8
        0xBE00, # G+2096  U+BE00  uniBE00,  HANGUL SYLLABLE-BE00
        0xBE03, # G+2097  U+BE03  uniBE03,  HANGUL SYLLABLE-BE03
        0xBE05, # G+2098  U+BE05  uniBE05,  HANGUL SYLLABLE-BE05
        0xBE0C, # G+2099  U+BE0C  uniBE0C,  HANGUL SYLLABLE-BE0C
        0xBE0D, # G+2100  U+BE0D  uniBE0D,  HANGUL SYLLABLE-BE0D
        0xBE10, # G+2101  U+BE10  uniBE10,  HANGUL SYLLABLE-BE10
        0xBE14, # G+2102  U+BE14  uniBE14,  HANGUL SYLLABLE-BE14
        0xBE1C, # G+2103  U+BE1C  uniBE1C,  HANGUL SYLLABLE-BE1C
        0xBE1D, # G+2104  U+BE1D  uniBE1D,  HANGUL SYLLABLE-BE1D
        0xBE1F, # G+2105  U+BE1F  uniBE1F,  HANGUL SYLLABLE-BE1F
        0xBE44, # G+2106  U+BE44  uniBE44,  HANGUL SYLLABLE-BE44
        0xBE45, # G+2107  U+BE45  uniBE45,  HANGUL SYLLABLE-BE45
        0xBE48, # G+2108  U+BE48  uniBE48,  HANGUL SYLLABLE-BE48
        0xBE4C, # G+2109  U+BE4C  uniBE4C,  HANGUL SYLLABLE-BE4C
        0xBE4E, # G+2110  U+BE4E  uniBE4E,  HANGUL SYLLABLE-BE4E
        0xBE54, # G+2111  U+BE54  uniBE54,  HANGUL SYLLABLE-BE54
        0xBE55, # G+2112  U+BE55  uniBE55,  HANGUL SYLLABLE-BE55
        0xBE57, # G+2113  U+BE57  uniBE57,  HANGUL SYLLABLE-BE57
        0xBE59, # G+2114  U+BE59  uniBE59,  HANGUL SYLLABLE-BE59
        0xBE5A, # G+2115  U+BE5A  uniBE5A,  HANGUL SYLLABLE-BE5A
        0xBE5B, # G+2116  U+BE5B  uniBE5B,  HANGUL SYLLABLE-BE5B
        0xBE60, # G+2117  U+BE60  uniBE60,  HANGUL SYLLABLE-BE60
        0xBE61, # G+2118  U+BE61  uniBE61,  HANGUL SYLLABLE-BE61
        0xBE64, # G+2119  U+BE64  uniBE64,  HANGUL SYLLABLE-BE64
        0xBE68, # G+2120  U+BE68  uniBE68,  HANGUL SYLLABLE-BE68
        0xBE6A, # G+2121  U+BE6A  uniBE6A,  HANGUL SYLLABLE-BE6A
        0xBE70, # G+2122  U+BE70  uniBE70,  HANGUL SYLLABLE-BE70
        0xBE71, # G+2123  U+BE71  uniBE71,  HANGUL SYLLABLE-BE71
        0xBE73, # G+2124  U+BE73  uniBE73,  HANGUL SYLLABLE-BE73
        0xBE74, # G+2125  U+BE74  uniBE74,  HANGUL SYLLABLE-BE74
        0xBE75, # G+2126  U+BE75  uniBE75,  HANGUL SYLLABLE-BE75
        0xBE7B, # G+2127  U+BE7B  uniBE7B,  HANGUL SYLLABLE-BE7B
        0xBE7C, # G+2128  U+BE7C  uniBE7C,  HANGUL SYLLABLE-BE7C
        0xBE7D, # G+2129  U+BE7D  uniBE7D,  HANGUL SYLLABLE-BE7D
        0xBE80, # G+2130  U+BE80  uniBE80,  HANGUL SYLLABLE-BE80
        0xBE84, # G+2131  U+BE84  uniBE84,  HANGUL SYLLABLE-BE84
        0xBE8C, # G+2132  U+BE8C  uniBE8C,  HANGUL SYLLABLE-BE8C
        0xBE8D, # G+2133  U+BE8D  uniBE8D,  HANGUL SYLLABLE-BE8D
        0xBE8F, # G+2134  U+BE8F  uniBE8F,  HANGUL SYLLABLE-BE8F
        0xBE90, # G+2135  U+BE90  uniBE90,  HANGUL SYLLABLE-BE90
        0xBE91, # G+2136  U+BE91  uniBE91,  HANGUL SYLLABLE-BE91
        0xBE98, # G+2137  U+BE98  uniBE98,  HANGUL SYLLABLE-BE98
        0xBE99, # G+2138  U+BE99  uniBE99,  HANGUL SYLLABLE-BE99
        0xBEA8, # G+2139  U+BEA8  uniBEA8,  HANGUL SYLLABLE-BEA8
        0xBED0, # G+2140  U+BED0  uniBED0,  HANGUL SYLLABLE-BED0
        0xBED1, # G+2141  U+BED1  uniBED1,  HANGUL SYLLABLE-BED1
        0xBED4, # G+2142  U+BED4  uniBED4,  HANGUL SYLLABLE-BED4
        0xBED7, # G+2143  U+BED7  uniBED7,  HANGUL SYLLABLE-BED7
        0xBED8, # G+2144  U+BED8  uniBED8,  HANGUL SYLLABLE-BED8
        0xBEE0, # G+2145  U+BEE0  uniBEE0,  HANGUL SYLLABLE-BEE0
        0xBEE3, # G+2146  U+BEE3  uniBEE3,  HANGUL SYLLABLE-BEE3
        0xBEE4, # G+2147  U+BEE4  uniBEE4,  HANGUL SYLLABLE-BEE4
        0xBEE5, # G+2148  U+BEE5  uniBEE5,  HANGUL SYLLABLE-BEE5
        0xBEEC, # G+2149  U+BEEC  uniBEEC,  HANGUL SYLLABLE-BEEC
        0xBF01, # G+2150  U+BF01  uniBF01,  HANGUL SYLLABLE-BF01
        0xBF08, # G+2151  U+BF08  uniBF08,  HANGUL SYLLABLE-BF08
        0xBF09, # G+2152  U+BF09  uniBF09,  HANGUL SYLLABLE-BF09
        0xBF18, # G+2153  U+BF18  uniBF18,  HANGUL SYLLABLE-BF18
        0xBF19, # G+2154  U+BF19  uniBF19,  HANGUL SYLLABLE-BF19
        0xBF1B, # G+2155  U+BF1B  uniBF1B,  HANGUL SYLLABLE-BF1B
        0xBF1C, # G+2156  U+BF1C  uniBF1C,  HANGUL SYLLABLE-BF1C
        0xBF1D, # G+2157  U+BF1D  uniBF1D,  HANGUL SYLLABLE-BF1D
        0xBF40, # G+2158  U+BF40  uniBF40,  HANGUL SYLLABLE-BF40
        0xBF41, # G+2159  U+BF41  uniBF41,  HANGUL SYLLABLE-BF41
        0xBF44, # G+2160  U+BF44  uniBF44,  HANGUL SYLLABLE-BF44
        0xBF48, # G+2161  U+BF48  uniBF48,  HANGUL SYLLABLE-BF48
        0xBF50, # G+2162  U+BF50  uniBF50,  HANGUL SYLLABLE-BF50
        0xBF51, # G+2163  U+BF51  uniBF51,  HANGUL SYLLABLE-BF51
        0xBF55, # G+2164  U+BF55  uniBF55,  HANGUL SYLLABLE-BF55
        0xBF94, # G+2165  U+BF94  uniBF94,  HANGUL SYLLABLE-BF94
        0xBFB0, # G+2166  U+BFB0  uniBFB0,  HANGUL SYLLABLE-BFB0
        0xBFC5, # G+2167  U+BFC5  uniBFC5,  HANGUL SYLLABLE-BFC5
        0xBFCC, # G+2168  U+BFCC  uniBFCC,  HANGUL SYLLABLE-BFCC
        0xBFCD, # G+2169  U+BFCD  uniBFCD,  HANGUL SYLLABLE-BFCD
        0xBFD0, # G+2170  U+BFD0  uniBFD0,  HANGUL SYLLABLE-BFD0
        0xBFD4, # G+2171  U+BFD4  uniBFD4,  HANGUL SYLLABLE-BFD4
        0xBFDC, # G+2172  U+BFDC  uniBFDC,  HANGUL SYLLABLE-BFDC
        0xBFDF, # G+2173  U+BFDF  uniBFDF,  HANGUL SYLLABLE-BFDF
        0xBFE1, # G+2174  U+BFE1  uniBFE1,  HANGUL SYLLABLE-BFE1
        0xC03C, # G+2175  U+C03C  uniC03C,  HANGUL SYLLABLE-C03C
        0xC051, # G+2176  U+C051  uniC051,  HANGUL SYLLABLE-C051
        0xC058, # G+2177  U+C058  uniC058,  HANGUL SYLLABLE-C058
        0xC05C, # G+2178  U+C05C  uniC05C,  HANGUL SYLLABLE-C05C
        0xC060, # G+2179  U+C060  uniC060,  HANGUL SYLLABLE-C060
        0xC068, # G+2180  U+C068  uniC068,  HANGUL SYLLABLE-C068
        0xC069, # G+2181  U+C069  uniC069,  HANGUL SYLLABLE-C069
        0xC090, # G+2182  U+C090  uniC090,  HANGUL SYLLABLE-C090
        0xC091, # G+2183  U+C091  uniC091,  HANGUL SYLLABLE-C091
        0xC094, # G+2184  U+C094  uniC094,  HANGUL SYLLABLE-C094
        0xC098, # G+2185  U+C098  uniC098,  HANGUL SYLLABLE-C098
        0xC0A0, # G+2186  U+C0A0  uniC0A0,  HANGUL SYLLABLE-C0A0
        0xC0A1, # G+2187  U+C0A1  uniC0A1,  HANGUL SYLLABLE-C0A1
        0xC0A3, # G+2188  U+C0A3  uniC0A3,  HANGUL SYLLABLE-C0A3
        0xC0A5, # G+2189  U+C0A5  uniC0A5,  HANGUL SYLLABLE-C0A5
        0xC0AC, # G+2190  U+C0AC  uniC0AC,  HANGUL SYLLABLE-C0AC
        0xC0AD, # G+2191  U+C0AD  uniC0AD,  HANGUL SYLLABLE-C0AD
        0xC0AF, # G+2192  U+C0AF  uniC0AF,  HANGUL SYLLABLE-C0AF
        0xC0B0, # G+2193  U+C0B0  uniC0B0,  HANGUL SYLLABLE-C0B0
        0xC0B3, # G+2194  U+C0B3  uniC0B3,  HANGUL SYLLABLE-C0B3
        0xC0B4, # G+2195  U+C0B4  uniC0B4,  HANGUL SYLLABLE-C0B4
        0xC0B5, # G+2196  U+C0B5  uniC0B5,  HANGUL SYLLABLE-C0B5
        0xC0B6, # G+2197  U+C0B6  uniC0B6,  HANGUL SYLLABLE-C0B6
        0xC0BC, # G+2198  U+C0BC  uniC0BC,  HANGUL SYLLABLE-C0BC
        0xC0BD, # G+2199  U+C0BD  uniC0BD,  HANGUL SYLLABLE-C0BD
        0xC0BF, # G+2200  U+C0BF  uniC0BF,  HANGUL SYLLABLE-C0BF
        0xC0C0, # G+2201  U+C0C0  uniC0C0,  HANGUL SYLLABLE-C0C0
        0xC0C1, # G+2202  U+C0C1  uniC0C1,  HANGUL SYLLABLE-C0C1
        0xC0C5, # G+2203  U+C0C5  uniC0C5,  HANGUL SYLLABLE-C0C5
        0xC0C8, # G+2204  U+C0C8  uniC0C8,  HANGUL SYLLABLE-C0C8
        0xC0C9, # G+2205  U+C0C9  uniC0C9,  HANGUL SYLLABLE-C0C9
        0xC0CC, # G+2206  U+C0CC  uniC0CC,  HANGUL SYLLABLE-C0CC
        0xC0D0, # G+2207  U+C0D0  uniC0D0,  HANGUL SYLLABLE-C0D0
        0xC0D8, # G+2208  U+C0D8  uniC0D8,  HANGUL SYLLABLE-C0D8
        0xC0D9, # G+2209  U+C0D9  uniC0D9,  HANGUL SYLLABLE-C0D9
        0xC0DB, # G+2210  U+C0DB  uniC0DB,  HANGUL SYLLABLE-C0DB
        0xC0DC, # G+2211  U+C0DC  uniC0DC,  HANGUL SYLLABLE-C0DC
        0xC0DD, # G+2212  U+C0DD  uniC0DD,  HANGUL SYLLABLE-C0DD
        0xC0E4, # G+2213  U+C0E4  uniC0E4,  HANGUL SYLLABLE-C0E4
        0xC0E5, # G+2214  U+C0E5  uniC0E5,  HANGUL SYLLABLE-C0E5
        0xC0E8, # G+2215  U+C0E8  uniC0E8,  HANGUL SYLLABLE-C0E8
        0xC0EC, # G+2216  U+C0EC  uniC0EC,  HANGUL SYLLABLE-C0EC
        0xC0F4, # G+2217  U+C0F4  uniC0F4,  HANGUL SYLLABLE-C0F4
        0xC0F5, # G+2218  U+C0F5  uniC0F5,  HANGUL SYLLABLE-C0F5
        0xC0F7, # G+2219  U+C0F7  uniC0F7,  HANGUL SYLLABLE-C0F7
        0xC0F9, # G+2220  U+C0F9  uniC0F9,  HANGUL SYLLABLE-C0F9
        0xC100, # G+2221  U+C100  uniC100,  HANGUL SYLLABLE-C100
        0xC104, # G+2222  U+C104  uniC104,  HANGUL SYLLABLE-C104
        0xC108, # G+2223  U+C108  uniC108,  HANGUL SYLLABLE-C108
        0xC110, # G+2224  U+C110  uniC110,  HANGUL SYLLABLE-C110
        0xC115, # G+2225  U+C115  uniC115,  HANGUL SYLLABLE-C115
        0xC11C, # G+2226  U+C11C  uniC11C,  HANGUL SYLLABLE-C11C
        0xC11D, # G+2227  U+C11D  uniC11D,  HANGUL SYLLABLE-C11D
        0xC11E, # G+2228  U+C11E  uniC11E,  HANGUL SYLLABLE-C11E
        0xC11F, # G+2229  U+C11F  uniC11F,  HANGUL SYLLABLE-C11F
        0xC120, # G+2230  U+C120  uniC120,  HANGUL SYLLABLE-C120
        0xC123, # G+2231  U+C123  uniC123,  HANGUL SYLLABLE-C123
        0xC124, # G+2232  U+C124  uniC124,  HANGUL SYLLABLE-C124
        0xC126, # G+2233  U+C126  uniC126,  HANGUL SYLLABLE-C126
        0xC127, # G+2234  U+C127  uniC127,  HANGUL SYLLABLE-C127
        0xC12C, # G+2235  U+C12C  uniC12C,  HANGUL SYLLABLE-C12C
        0xC12D, # G+2236  U+C12D  uniC12D,  HANGUL SYLLABLE-C12D
        0xC12F, # G+2237  U+C12F  uniC12F,  HANGUL SYLLABLE-C12F
        0xC130, # G+2238  U+C130  uniC130,  HANGUL SYLLABLE-C130
        0xC131, # G+2239  U+C131  uniC131,  HANGUL SYLLABLE-C131
        0xC136, # G+2240  U+C136  uniC136,  HANGUL SYLLABLE-C136
        0xC138, # G+2241  U+C138  uniC138,  HANGUL SYLLABLE-C138
        0xC139, # G+2242  U+C139  uniC139,  HANGUL SYLLABLE-C139
        0xC13C, # G+2243  U+C13C  uniC13C,  HANGUL SYLLABLE-C13C
        0xC140, # G+2244  U+C140  uniC140,  HANGUL SYLLABLE-C140
        0xC148, # G+2245  U+C148  uniC148,  HANGUL SYLLABLE-C148
        0xC149, # G+2246  U+C149  uniC149,  HANGUL SYLLABLE-C149
        0xC14B, # G+2247  U+C14B  uniC14B,  HANGUL SYLLABLE-C14B
        0xC14C, # G+2248  U+C14C  uniC14C,  HANGUL SYLLABLE-C14C
        0xC14D, # G+2249  U+C14D  uniC14D,  HANGUL SYLLABLE-C14D
        0xC154, # G+2250  U+C154  uniC154,  HANGUL SYLLABLE-C154
        0xC155, # G+2251  U+C155  uniC155,  HANGUL SYLLABLE-C155
        0xC158, # G+2252  U+C158  uniC158,  HANGUL SYLLABLE-C158
        0xC15C, # G+2253  U+C15C  uniC15C,  HANGUL SYLLABLE-C15C
        0xC164, # G+2254  U+C164  uniC164,  HANGUL SYLLABLE-C164
        0xC165, # G+2255  U+C165  uniC165,  HANGUL SYLLABLE-C165
        0xC167, # G+2256  U+C167  uniC167,  HANGUL SYLLABLE-C167
        0xC168, # G+2257  U+C168  uniC168,  HANGUL SYLLABLE-C168
        0xC169, # G+2258  U+C169  uniC169,  HANGUL SYLLABLE-C169
        0xC170, # G+2259  U+C170  uniC170,  HANGUL SYLLABLE-C170
        0xC174, # G+2260  U+C174  uniC174,  HANGUL SYLLABLE-C174
        0xC178, # G+2261  U+C178  uniC178,  HANGUL SYLLABLE-C178
        0xC185, # G+2262  U+C185  uniC185,  HANGUL SYLLABLE-C185
        0xC18C, # G+2263  U+C18C  uniC18C,  HANGUL SYLLABLE-C18C
        0xC18D, # G+2264  U+C18D  uniC18D,  HANGUL SYLLABLE-C18D
        0xC18E, # G+2265  U+C18E  uniC18E,  HANGUL SYLLABLE-C18E
        0xC190, # G+2266  U+C190  uniC190,  HANGUL SYLLABLE-C190
        0xC194, # G+2267  U+C194  uniC194,  HANGUL SYLLABLE-C194
        0xC196, # G+2268  U+C196  uniC196,  HANGUL SYLLABLE-C196
        0xC19C, # G+2269  U+C19C  uniC19C,  HANGUL SYLLABLE-C19C
        0xC19D, # G+2270  U+C19D  uniC19D,  HANGUL SYLLABLE-C19D
        0xC19F, # G+2271  U+C19F  uniC19F,  HANGUL SYLLABLE-C19F
        0xC1A1, # G+2272  U+C1A1  uniC1A1,  HANGUL SYLLABLE-C1A1
        0xC1A5, # G+2273  U+C1A5  uniC1A5,  HANGUL SYLLABLE-C1A5
        0xC1A8, # G+2274  U+C1A8  uniC1A8,  HANGUL SYLLABLE-C1A8
        0xC1A9, # G+2275  U+C1A9  uniC1A9,  HANGUL SYLLABLE-C1A9
        0xC1AC, # G+2276  U+C1AC  uniC1AC,  HANGUL SYLLABLE-C1AC
        0xC1B0, # G+2277  U+C1B0  uniC1B0,  HANGUL SYLLABLE-C1B0
        0xC1BD, # G+2278  U+C1BD  uniC1BD,  HANGUL SYLLABLE-C1BD
        0xC1C4, # G+2279  U+C1C4  uniC1C4,  HANGUL SYLLABLE-C1C4
        0xC1C8, # G+2280  U+C1C8  uniC1C8,  HANGUL SYLLABLE-C1C8
        0xC1CC, # G+2281  U+C1CC  uniC1CC,  HANGUL SYLLABLE-C1CC
        0xC1D4, # G+2282  U+C1D4  uniC1D4,  HANGUL SYLLABLE-C1D4
        0xC1D7, # G+2283  U+C1D7  uniC1D7,  HANGUL SYLLABLE-C1D7
        0xC1D8, # G+2284  U+C1D8  uniC1D8,  HANGUL SYLLABLE-C1D8
        0xC1E0, # G+2285  U+C1E0  uniC1E0,  HANGUL SYLLABLE-C1E0
        0xC1E4, # G+2286  U+C1E4  uniC1E4,  HANGUL SYLLABLE-C1E4
        0xC1E8, # G+2287  U+C1E8  uniC1E8,  HANGUL SYLLABLE-C1E8
        0xC1F0, # G+2288  U+C1F0  uniC1F0,  HANGUL SYLLABLE-C1F0
        0xC1F1, # G+2289  U+C1F1  uniC1F1,  HANGUL SYLLABLE-C1F1
        0xC1F3, # G+2290  U+C1F3  uniC1F3,  HANGUL SYLLABLE-C1F3
        0xC1FC, # G+2291  U+C1FC  uniC1FC,  HANGUL SYLLABLE-C1FC
        0xC1FD, # G+2292  U+C1FD  uniC1FD,  HANGUL SYLLABLE-C1FD
        0xC200, # G+2293  U+C200  uniC200,  HANGUL SYLLABLE-C200
        0xC204, # G+2294  U+C204  uniC204,  HANGUL SYLLABLE-C204
        0xC20C, # G+2295  U+C20C  uniC20C,  HANGUL SYLLABLE-C20C
        0xC20D, # G+2296  U+C20D  uniC20D,  HANGUL SYLLABLE-C20D
        0xC20F, # G+2297  U+C20F  uniC20F,  HANGUL SYLLABLE-C20F
        0xC211, # G+2298  U+C211  uniC211,  HANGUL SYLLABLE-C211
        0xC218, # G+2299  U+C218  uniC218,  HANGUL SYLLABLE-C218
        0xC219, # G+2300  U+C219  uniC219,  HANGUL SYLLABLE-C219
        0xC21C, # G+2301  U+C21C  uniC21C,  HANGUL SYLLABLE-C21C
        0xC21F, # G+2302  U+C21F  uniC21F,  HANGUL SYLLABLE-C21F
        0xC220, # G+2303  U+C220  uniC220,  HANGUL SYLLABLE-C220
        0xC228, # G+2304  U+C228  uniC228,  HANGUL SYLLABLE-C228
        0xC229, # G+2305  U+C229  uniC229,  HANGUL SYLLABLE-C229
        0xC22B, # G+2306  U+C22B  uniC22B,  HANGUL SYLLABLE-C22B
        0xC22D, # G+2307  U+C22D  uniC22D,  HANGUL SYLLABLE-C22D
        0xC22F, # G+2308  U+C22F  uniC22F,  HANGUL SYLLABLE-C22F
        0xC231, # G+2309  U+C231  uniC231,  HANGUL SYLLABLE-C231
        0xC232, # G+2310  U+C232  uniC232,  HANGUL SYLLABLE-C232
        0xC234, # G+2311  U+C234  uniC234,  HANGUL SYLLABLE-C234
        0xC248, # G+2312  U+C248  uniC248,  HANGUL SYLLABLE-C248
        0xC250, # G+2313  U+C250  uniC250,  HANGUL SYLLABLE-C250
        0xC251, # G+2314  U+C251  uniC251,  HANGUL SYLLABLE-C251
        0xC254, # G+2315  U+C254  uniC254,  HANGUL SYLLABLE-C254
        0xC258, # G+2316  U+C258  uniC258,  HANGUL SYLLABLE-C258
        0xC260, # G+2317  U+C260  uniC260,  HANGUL SYLLABLE-C260
        0xC265, # G+2318  U+C265  uniC265,  HANGUL SYLLABLE-C265
        0xC26C, # G+2319  U+C26C  uniC26C,  HANGUL SYLLABLE-C26C
        0xC26D, # G+2320  U+C26D  uniC26D,  HANGUL SYLLABLE-C26D
        0xC270, # G+2321  U+C270  uniC270,  HANGUL SYLLABLE-C270
        0xC274, # G+2322  U+C274  uniC274,  HANGUL SYLLABLE-C274
        0xC27C, # G+2323  U+C27C  uniC27C,  HANGUL SYLLABLE-C27C
        0xC27D, # G+2324  U+C27D  uniC27D,  HANGUL SYLLABLE-C27D
        0xC27F, # G+2325  U+C27F  uniC27F,  HANGUL SYLLABLE-C27F
        0xC281, # G+2326  U+C281  uniC281,  HANGUL SYLLABLE-C281
        0xC288, # G+2327  U+C288  uniC288,  HANGUL SYLLABLE-C288
        0xC289, # G+2328  U+C289  uniC289,  HANGUL SYLLABLE-C289
        0xC290, # G+2329  U+C290  uniC290,  HANGUL SYLLABLE-C290
        0xC298, # G+2330  U+C298  uniC298,  HANGUL SYLLABLE-C298
        0xC29B, # G+2331  U+C29B  uniC29B,  HANGUL SYLLABLE-C29B
        0xC29D, # G+2332  U+C29D  uniC29D,  HANGUL SYLLABLE-C29D
        0xC2A4, # G+2333  U+C2A4  uniC2A4,  HANGUL SYLLABLE-C2A4
        0xC2A5, # G+2334  U+C2A5  uniC2A5,  HANGUL SYLLABLE-C2A5
        0xC2A8, # G+2335  U+C2A8  uniC2A8,  HANGUL SYLLABLE-C2A8
        0xC2AC, # G+2336  U+C2AC  uniC2AC,  HANGUL SYLLABLE-C2AC
        0xC2AD, # G+2337  U+C2AD  uniC2AD,  HANGUL SYLLABLE-C2AD
        0xC2B4, # G+2338  U+C2B4  uniC2B4,  HANGUL SYLLABLE-C2B4
        0xC2B5, # G+2339  U+C2B5  uniC2B5,  HANGUL SYLLABLE-C2B5
        0xC2B7, # G+2340  U+C2B7  uniC2B7,  HANGUL SYLLABLE-C2B7
        0xC2B9, # G+2341  U+C2B9  uniC2B9,  HANGUL SYLLABLE-C2B9
        0xC2DC, # G+2342  U+C2DC  uniC2DC,  HANGUL SYLLABLE-C2DC
        0xC2DD, # G+2343  U+C2DD  uniC2DD,  HANGUL SYLLABLE-C2DD
        0xC2E0, # G+2344  U+C2E0  uniC2E0,  HANGUL SYLLABLE-C2E0
        0xC2E3, # G+2345  U+C2E3  uniC2E3,  HANGUL SYLLABLE-C2E3
        0xC2E4, # G+2346  U+C2E4  uniC2E4,  HANGUL SYLLABLE-C2E4
        0xC2EB, # G+2347  U+C2EB  uniC2EB,  HANGUL SYLLABLE-C2EB
        0xC2EC, # G+2348  U+C2EC  uniC2EC,  HANGUL SYLLABLE-C2EC
        0xC2ED, # G+2349  U+C2ED  uniC2ED,  HANGUL SYLLABLE-C2ED
        0xC2EF, # G+2350  U+C2EF  uniC2EF,  HANGUL SYLLABLE-C2EF
        0xC2F1, # G+2351  U+C2F1  uniC2F1,  HANGUL SYLLABLE-C2F1
        0xC2F6, # G+2352  U+C2F6  uniC2F6,  HANGUL SYLLABLE-C2F6
        0xC2F8, # G+2353  U+C2F8  uniC2F8,  HANGUL SYLLABLE-C2F8
        0xC2F9, # G+2354  U+C2F9  uniC2F9,  HANGUL SYLLABLE-C2F9
        0xC2FB, # G+2355  U+C2FB  uniC2FB,  HANGUL SYLLABLE-C2FB
        0xC2FC, # G+2356  U+C2FC  uniC2FC,  HANGUL SYLLABLE-C2FC
        0xC300, # G+2357  U+C300  uniC300,  HANGUL SYLLABLE-C300
        0xC308, # G+2358  U+C308  uniC308,  HANGUL SYLLABLE-C308
        0xC309, # G+2359  U+C309  uniC309,  HANGUL SYLLABLE-C309
        0xC30C, # G+2360  U+C30C  uniC30C,  HANGUL SYLLABLE-C30C
        0xC30D, # G+2361  U+C30D  uniC30D,  HANGUL SYLLABLE-C30D
        0xC313, # G+2362  U+C313  uniC313,  HANGUL SYLLABLE-C313
        0xC314, # G+2363  U+C314  uniC314,  HANGUL SYLLABLE-C314
        0xC315, # G+2364  U+C315  uniC315,  HANGUL SYLLABLE-C315
        0xC318, # G+2365  U+C318  uniC318,  HANGUL SYLLABLE-C318
        0xC31C, # G+2366  U+C31C  uniC31C,  HANGUL SYLLABLE-C31C
        0xC324, # G+2367  U+C324  uniC324,  HANGUL SYLLABLE-C324
        0xC325, # G+2368  U+C325  uniC325,  HANGUL SYLLABLE-C325
        0xC328, # G+2369  U+C328  uniC328,  HANGUL SYLLABLE-C328
        0xC329, # G+2370  U+C329  uniC329,  HANGUL SYLLABLE-C329
        0xC345, # G+2371  U+C345  uniC345,  HANGUL SYLLABLE-C345
        0xC368, # G+2372  U+C368  uniC368,  HANGUL SYLLABLE-C368
        0xC369, # G+2373  U+C369  uniC369,  HANGUL SYLLABLE-C369
        0xC36C, # G+2374  U+C36C  uniC36C,  HANGUL SYLLABLE-C36C
        0xC370, # G+2375  U+C370  uniC370,  HANGUL SYLLABLE-C370
        0xC372, # G+2376  U+C372  uniC372,  HANGUL SYLLABLE-C372
        0xC378, # G+2377  U+C378  uniC378,  HANGUL SYLLABLE-C378
        0xC379, # G+2378  U+C379  uniC379,  HANGUL SYLLABLE-C379
        0xC37C, # G+2379  U+C37C  uniC37C,  HANGUL SYLLABLE-C37C
        0xC37D, # G+2380  U+C37D  uniC37D,  HANGUL SYLLABLE-C37D
        0xC384, # G+2381  U+C384  uniC384,  HANGUL SYLLABLE-C384
        0xC388, # G+2382  U+C388  uniC388,  HANGUL SYLLABLE-C388
        0xC38C, # G+2383  U+C38C  uniC38C,  HANGUL SYLLABLE-C38C
        0xC3C0, # G+2384  U+C3C0  uniC3C0,  HANGUL SYLLABLE-C3C0
        0xC3D8, # G+2385  U+C3D8  uniC3D8,  HANGUL SYLLABLE-C3D8
        0xC3D9, # G+2386  U+C3D9  uniC3D9,  HANGUL SYLLABLE-C3D9
        0xC3DC, # G+2387  U+C3DC  uniC3DC,  HANGUL SYLLABLE-C3DC
        0xC3DF, # G+2388  U+C3DF  uniC3DF,  HANGUL SYLLABLE-C3DF
        0xC3E0, # G+2389  U+C3E0  uniC3E0,  HANGUL SYLLABLE-C3E0
        0xC3E2, # G+2390  U+C3E2  uniC3E2,  HANGUL SYLLABLE-C3E2
        0xC3E8, # G+2391  U+C3E8  uniC3E8,  HANGUL SYLLABLE-C3E8
        0xC3E9, # G+2392  U+C3E9  uniC3E9,  HANGUL SYLLABLE-C3E9
        0xC3ED, # G+2393  U+C3ED  uniC3ED,  HANGUL SYLLABLE-C3ED
        0xC3F4, # G+2394  U+C3F4  uniC3F4,  HANGUL SYLLABLE-C3F4
        0xC3F5, # G+2395  U+C3F5  uniC3F5,  HANGUL SYLLABLE-C3F5
        0xC3F8, # G+2396  U+C3F8  uniC3F8,  HANGUL SYLLABLE-C3F8
        0xC408, # G+2397  U+C408  uniC408,  HANGUL SYLLABLE-C408
        0xC410, # G+2398  U+C410  uniC410,  HANGUL SYLLABLE-C410
        0xC424, # G+2399  U+C424  uniC424,  HANGUL SYLLABLE-C424
        0xC42C, # G+2400  U+C42C  uniC42C,  HANGUL SYLLABLE-C42C
        0xC430, # G+2401  U+C430  uniC430,  HANGUL SYLLABLE-C430
        0xC434, # G+2402  U+C434  uniC434,  HANGUL SYLLABLE-C434
        0xC43C, # G+2403  U+C43C  uniC43C,  HANGUL SYLLABLE-C43C
        0xC43D, # G+2404  U+C43D  uniC43D,  HANGUL SYLLABLE-C43D
        0xC448, # G+2405  U+C448  uniC448,  HANGUL SYLLABLE-C448
        0xC464, # G+2406  U+C464  uniC464,  HANGUL SYLLABLE-C464
        0xC465, # G+2407  U+C465  uniC465,  HANGUL SYLLABLE-C465
        0xC468, # G+2408  U+C468  uniC468,  HANGUL SYLLABLE-C468
        0xC46C, # G+2409  U+C46C  uniC46C,  HANGUL SYLLABLE-C46C
        0xC474, # G+2410  U+C474  uniC474,  HANGUL SYLLABLE-C474
        0xC475, # G+2411  U+C475  uniC475,  HANGUL SYLLABLE-C475
        0xC479, # G+2412  U+C479  uniC479,  HANGUL SYLLABLE-C479
        0xC480, # G+2413  U+C480  uniC480,  HANGUL SYLLABLE-C480
        0xC494, # G+2414  U+C494  uniC494,  HANGUL SYLLABLE-C494
        0xC49C, # G+2415  U+C49C  uniC49C,  HANGUL SYLLABLE-C49C
        0xC4B8, # G+2416  U+C4B8  uniC4B8,  HANGUL SYLLABLE-C4B8
        0xC4BC, # G+2417  U+C4BC  uniC4BC,  HANGUL SYLLABLE-C4BC
        0xC4E9, # G+2418  U+C4E9  uniC4E9,  HANGUL SYLLABLE-C4E9
        0xC4F0, # G+2419  U+C4F0  uniC4F0,  HANGUL SYLLABLE-C4F0
        0xC4F1, # G+2420  U+C4F1  uniC4F1,  HANGUL SYLLABLE-C4F1
        0xC4F4, # G+2421  U+C4F4  uniC4F4,  HANGUL SYLLABLE-C4F4
        0xC4F8, # G+2422  U+C4F8  uniC4F8,  HANGUL SYLLABLE-C4F8
        0xC4FA, # G+2423  U+C4FA  uniC4FA,  HANGUL SYLLABLE-C4FA
        0xC4FF, # G+2424  U+C4FF  uniC4FF,  HANGUL SYLLABLE-C4FF
        0xC500, # G+2425  U+C500  uniC500,  HANGUL SYLLABLE-C500
        0xC501, # G+2426  U+C501  uniC501,  HANGUL SYLLABLE-C501
        0xC50C, # G+2427  U+C50C  uniC50C,  HANGUL SYLLABLE-C50C
        0xC510, # G+2428  U+C510  uniC510,  HANGUL SYLLABLE-C510
        0xC514, # G+2429  U+C514  uniC514,  HANGUL SYLLABLE-C514
        0xC51C, # G+2430  U+C51C  uniC51C,  HANGUL SYLLABLE-C51C
        0xC528, # G+2431  U+C528  uniC528,  HANGUL SYLLABLE-C528
        0xC529, # G+2432  U+C529  uniC529,  HANGUL SYLLABLE-C529
        0xC52C, # G+2433  U+C52C  uniC52C,  HANGUL SYLLABLE-C52C
        0xC530, # G+2434  U+C530  uniC530,  HANGUL SYLLABLE-C530
        0xC538, # G+2435  U+C538  uniC538,  HANGUL SYLLABLE-C538
        0xC539, # G+2436  U+C539  uniC539,  HANGUL SYLLABLE-C539
        0xC53B, # G+2437  U+C53B  uniC53B,  HANGUL SYLLABLE-C53B
        0xC53D, # G+2438  U+C53D  uniC53D,  HANGUL SYLLABLE-C53D
        0xC544, # G+2439  U+C544  uniC544,  HANGUL SYLLABLE-C544
        0xC545, # G+2440  U+C545  uniC545,  HANGUL SYLLABLE-C545
        0xC548, # G+2441  U+C548  uniC548,  HANGUL SYLLABLE-C548
        0xC549, # G+2442  U+C549  uniC549,  HANGUL SYLLABLE-C549
        0xC54A, # G+2443  U+C54A  uniC54A,  HANGUL SYLLABLE-C54A
        0xC54C, # G+2444  U+C54C  uniC54C,  HANGUL SYLLABLE-C54C
        0xC54D, # G+2445  U+C54D  uniC54D,  HANGUL SYLLABLE-C54D
        0xC54E, # G+2446  U+C54E  uniC54E,  HANGUL SYLLABLE-C54E
        0xC553, # G+2447  U+C553  uniC553,  HANGUL SYLLABLE-C553
        0xC554, # G+2448  U+C554  uniC554,  HANGUL SYLLABLE-C554
        0xC555, # G+2449  U+C555  uniC555,  HANGUL SYLLABLE-C555
        0xC557, # G+2450  U+C557  uniC557,  HANGUL SYLLABLE-C557
        0xC558, # G+2451  U+C558  uniC558,  HANGUL SYLLABLE-C558
        0xC559, # G+2452  U+C559  uniC559,  HANGUL SYLLABLE-C559
        0xC55D, # G+2453  U+C55D  uniC55D,  HANGUL SYLLABLE-C55D
        0xC55E, # G+2454  U+C55E  uniC55E,  HANGUL SYLLABLE-C55E
        0xC560, # G+2455  U+C560  uniC560,  HANGUL SYLLABLE-C560
        0xC561, # G+2456  U+C561  uniC561,  HANGUL SYLLABLE-C561
        0xC564, # G+2457  U+C564  uniC564,  HANGUL SYLLABLE-C564
        0xC568, # G+2458  U+C568  uniC568,  HANGUL SYLLABLE-C568
        0xC570, # G+2459  U+C570  uniC570,  HANGUL SYLLABLE-C570
        0xC571, # G+2460  U+C571  uniC571,  HANGUL SYLLABLE-C571
        0xC573, # G+2461  U+C573  uniC573,  HANGUL SYLLABLE-C573
        0xC574, # G+2462  U+C574  uniC574,  HANGUL SYLLABLE-C574
        0xC575, # G+2463  U+C575  uniC575,  HANGUL SYLLABLE-C575
        0xC57C, # G+2464  U+C57C  uniC57C,  HANGUL SYLLABLE-C57C
        0xC57D, # G+2465  U+C57D  uniC57D,  HANGUL SYLLABLE-C57D
        0xC580, # G+2466  U+C580  uniC580,  HANGUL SYLLABLE-C580
        0xC584, # G+2467  U+C584  uniC584,  HANGUL SYLLABLE-C584
        0xC587, # G+2468  U+C587  uniC587,  HANGUL SYLLABLE-C587
        0xC58C, # G+2469  U+C58C  uniC58C,  HANGUL SYLLABLE-C58C
        0xC58D, # G+2470  U+C58D  uniC58D,  HANGUL SYLLABLE-C58D
        0xC58F, # G+2471  U+C58F  uniC58F,  HANGUL SYLLABLE-C58F
        0xC591, # G+2472  U+C591  uniC591,  HANGUL SYLLABLE-C591
        0xC595, # G+2473  U+C595  uniC595,  HANGUL SYLLABLE-C595
        0xC597, # G+2474  U+C597  uniC597,  HANGUL SYLLABLE-C597
        0xC598, # G+2475  U+C598  uniC598,  HANGUL SYLLABLE-C598
        0xC59C, # G+2476  U+C59C  uniC59C,  HANGUL SYLLABLE-C59C
        0xC5A0, # G+2477  U+C5A0  uniC5A0,  HANGUL SYLLABLE-C5A0
        0xC5A9, # G+2478  U+C5A9  uniC5A9,  HANGUL SYLLABLE-C5A9
        0xC5B4, # G+2479  U+C5B4  uniC5B4,  HANGUL SYLLABLE-C5B4
        0xC5B5, # G+2480  U+C5B5  uniC5B5,  HANGUL SYLLABLE-C5B5
        0xC5B8, # G+2481  U+C5B8  uniC5B8,  HANGUL SYLLABLE-C5B8
        0xC5B9, # G+2482  U+C5B9  uniC5B9,  HANGUL SYLLABLE-C5B9
        0xC5BB, # G+2483  U+C5BB  uniC5BB,  HANGUL SYLLABLE-C5BB
        0xC5BC, # G+2484  U+C5BC  uniC5BC,  HANGUL SYLLABLE-C5BC
        0xC5BD, # G+2485  U+C5BD  uniC5BD,  HANGUL SYLLABLE-C5BD
        0xC5BE, # G+2486  U+C5BE  uniC5BE,  HANGUL SYLLABLE-C5BE
        0xC5C4, # G+2487  U+C5C4  uniC5C4,  HANGUL SYLLABLE-C5C4
        0xC5C5, # G+2488  U+C5C5  uniC5C5,  HANGUL SYLLABLE-C5C5
        0xC5C6, # G+2489  U+C5C6  uniC5C6,  HANGUL SYLLABLE-C5C6
        0xC5C7, # G+2490  U+C5C7  uniC5C7,  HANGUL SYLLABLE-C5C7
        0xC5C8, # G+2491  U+C5C8  uniC5C8,  HANGUL SYLLABLE-C5C8
        0xC5C9, # G+2492  U+C5C9  uniC5C9,  HANGUL SYLLABLE-C5C9
        0xC5CA, # G+2493  U+C5CA  uniC5CA,  HANGUL SYLLABLE-C5CA
        0xC5CC, # G+2494  U+C5CC  uniC5CC,  HANGUL SYLLABLE-C5CC
        0xC5CE, # G+2495  U+C5CE  uniC5CE,  HANGUL SYLLABLE-C5CE
        0xC5D0, # G+2496  U+C5D0  uniC5D0,  HANGUL SYLLABLE-C5D0
        0xC5D1, # G+2497  U+C5D1  uniC5D1,  HANGUL SYLLABLE-C5D1
        0xC5D4, # G+2498  U+C5D4  uniC5D4,  HANGUL SYLLABLE-C5D4
        0xC5D8, # G+2499  U+C5D8  uniC5D8,  HANGUL SYLLABLE-C5D8
        0xC5E0, # G+2500  U+C5E0  uniC5E0,  HANGUL SYLLABLE-C5E0
        0xC5E1, # G+2501  U+C5E1  uniC5E1,  HANGUL SYLLABLE-C5E1
        0xC5E3, # G+2502  U+C5E3  uniC5E3,  HANGUL SYLLABLE-C5E3
        0xC5E5, # G+2503  U+C5E5  uniC5E5,  HANGUL SYLLABLE-C5E5
        0xC5EC, # G+2504  U+C5EC  uniC5EC,  HANGUL SYLLABLE-C5EC
        0xC5ED, # G+2505  U+C5ED  uniC5ED,  HANGUL SYLLABLE-C5ED
        0xC5EE, # G+2506  U+C5EE  uniC5EE,  HANGUL SYLLABLE-C5EE
        0xC5F0, # G+2507  U+C5F0  uniC5F0,  HANGUL SYLLABLE-C5F0
        0xC5F4, # G+2508  U+C5F4  uniC5F4,  HANGUL SYLLABLE-C5F4
        0xC5F6, # G+2509  U+C5F6  uniC5F6,  HANGUL SYLLABLE-C5F6
        0xC5F7, # G+2510  U+C5F7  uniC5F7,  HANGUL SYLLABLE-C5F7
        0xC5FC, # G+2511  U+C5FC  uniC5FC,  HANGUL SYLLABLE-C5FC
        0xC5FD, # G+2512  U+C5FD  uniC5FD,  HANGUL SYLLABLE-C5FD
        0xC5FE, # G+2513  U+C5FE  uniC5FE,  HANGUL SYLLABLE-C5FE
        0xC5FF, # G+2514  U+C5FF  uniC5FF,  HANGUL SYLLABLE-C5FF
        0xC600, # G+2515  U+C600  uniC600,  HANGUL SYLLABLE-C600
        0xC601, # G+2516  U+C601  uniC601,  HANGUL SYLLABLE-C601
        0xC605, # G+2517  U+C605  uniC605,  HANGUL SYLLABLE-C605
        0xC606, # G+2518  U+C606  uniC606,  HANGUL SYLLABLE-C606
        0xC607, # G+2519  U+C607  uniC607,  HANGUL SYLLABLE-C607
        0xC608, # G+2520  U+C608  uniC608,  HANGUL SYLLABLE-C608
        0xC60C, # G+2521  U+C60C  uniC60C,  HANGUL SYLLABLE-C60C
        0xC610, # G+2522  U+C610  uniC610,  HANGUL SYLLABLE-C610
        0xC618, # G+2523  U+C618  uniC618,  HANGUL SYLLABLE-C618
        0xC619, # G+2524  U+C619  uniC619,  HANGUL SYLLABLE-C619
        0xC61B, # G+2525  U+C61B  uniC61B,  HANGUL SYLLABLE-C61B
        0xC61C, # G+2526  U+C61C  uniC61C,  HANGUL SYLLABLE-C61C
        0xC624, # G+2527  U+C624  uniC624,  HANGUL SYLLABLE-C624
        0xC625, # G+2528  U+C625  uniC625,  HANGUL SYLLABLE-C625
        0xC628, # G+2529  U+C628  uniC628,  HANGUL SYLLABLE-C628
        0xC62C, # G+2530  U+C62C  uniC62C,  HANGUL SYLLABLE-C62C
        0xC62D, # G+2531  U+C62D  uniC62D,  HANGUL SYLLABLE-C62D
        0xC62E, # G+2532  U+C62E  uniC62E,  HANGUL SYLLABLE-C62E
        0xC630, # G+2533  U+C630  uniC630,  HANGUL SYLLABLE-C630
        0xC633, # G+2534  U+C633  uniC633,  HANGUL SYLLABLE-C633
        0xC634, # G+2535  U+C634  uniC634,  HANGUL SYLLABLE-C634
        0xC635, # G+2536  U+C635  uniC635,  HANGUL SYLLABLE-C635
        0xC637, # G+2537  U+C637  uniC637,  HANGUL SYLLABLE-C637
        0xC639, # G+2538  U+C639  uniC639,  HANGUL SYLLABLE-C639
        0xC63B, # G+2539  U+C63B  uniC63B,  HANGUL SYLLABLE-C63B
        0xC640, # G+2540  U+C640  uniC640,  HANGUL SYLLABLE-C640
        0xC641, # G+2541  U+C641  uniC641,  HANGUL SYLLABLE-C641
        0xC644, # G+2542  U+C644  uniC644,  HANGUL SYLLABLE-C644
        0xC648, # G+2543  U+C648  uniC648,  HANGUL SYLLABLE-C648
        0xC650, # G+2544  U+C650  uniC650,  HANGUL SYLLABLE-C650
        0xC651, # G+2545  U+C651  uniC651,  HANGUL SYLLABLE-C651
        0xC653, # G+2546  U+C653  uniC653,  HANGUL SYLLABLE-C653
        0xC654, # G+2547  U+C654  uniC654,  HANGUL SYLLABLE-C654
        0xC655, # G+2548  U+C655  uniC655,  HANGUL SYLLABLE-C655
        0xC65C, # G+2549  U+C65C  uniC65C,  HANGUL SYLLABLE-C65C
        0xC65D, # G+2550  U+C65D  uniC65D,  HANGUL SYLLABLE-C65D
        0xC660, # G+2551  U+C660  uniC660,  HANGUL SYLLABLE-C660
        0xC66C, # G+2552  U+C66C  uniC66C,  HANGUL SYLLABLE-C66C
        0xC66F, # G+2553  U+C66F  uniC66F,  HANGUL SYLLABLE-C66F
        0xC671, # G+2554  U+C671  uniC671,  HANGUL SYLLABLE-C671
        0xC678, # G+2555  U+C678  uniC678,  HANGUL SYLLABLE-C678
        0xC679, # G+2556  U+C679  uniC679,  HANGUL SYLLABLE-C679
        0xC67C, # G+2557  U+C67C  uniC67C,  HANGUL SYLLABLE-C67C
        0xC680, # G+2558  U+C680  uniC680,  HANGUL SYLLABLE-C680
        0xC688, # G+2559  U+C688  uniC688,  HANGUL SYLLABLE-C688
        0xC689, # G+2560  U+C689  uniC689,  HANGUL SYLLABLE-C689
        0xC68B, # G+2561  U+C68B  uniC68B,  HANGUL SYLLABLE-C68B
        0xC68D, # G+2562  U+C68D  uniC68D,  HANGUL SYLLABLE-C68D
        0xC694, # G+2563  U+C694  uniC694,  HANGUL SYLLABLE-C694
        0xC695, # G+2564  U+C695  uniC695,  HANGUL SYLLABLE-C695
        0xC698, # G+2565  U+C698  uniC698,  HANGUL SYLLABLE-C698
        0xC69C, # G+2566  U+C69C  uniC69C,  HANGUL SYLLABLE-C69C
        0xC6A4, # G+2567  U+C6A4  uniC6A4,  HANGUL SYLLABLE-C6A4
        0xC6A5, # G+2568  U+C6A5  uniC6A5,  HANGUL SYLLABLE-C6A5
        0xC6A7, # G+2569  U+C6A7  uniC6A7,  HANGUL SYLLABLE-C6A7
        0xC6A9, # G+2570  U+C6A9  uniC6A9,  HANGUL SYLLABLE-C6A9
        0xC6B0, # G+2571  U+C6B0  uniC6B0,  HANGUL SYLLABLE-C6B0
        0xC6B1, # G+2572  U+C6B1  uniC6B1,  HANGUL SYLLABLE-C6B1
        0xC6B4, # G+2573  U+C6B4  uniC6B4,  HANGUL SYLLABLE-C6B4
        0xC6B8, # G+2574  U+C6B8  uniC6B8,  HANGUL SYLLABLE-C6B8
        0xC6B9, # G+2575  U+C6B9  uniC6B9,  HANGUL SYLLABLE-C6B9
        0xC6BA, # G+2576  U+C6BA  uniC6BA,  HANGUL SYLLABLE-C6BA
        0xC6C0, # G+2577  U+C6C0  uniC6C0,  HANGUL SYLLABLE-C6C0
        0xC6C1, # G+2578  U+C6C1  uniC6C1,  HANGUL SYLLABLE-C6C1
        0xC6C3, # G+2579  U+C6C3  uniC6C3,  HANGUL SYLLABLE-C6C3
        0xC6C5, # G+2580  U+C6C5  uniC6C5,  HANGUL SYLLABLE-C6C5
        0xC6CC, # G+2581  U+C6CC  uniC6CC,  HANGUL SYLLABLE-C6CC
        0xC6CD, # G+2582  U+C6CD  uniC6CD,  HANGUL SYLLABLE-C6CD
        0xC6D0, # G+2583  U+C6D0  uniC6D0,  HANGUL SYLLABLE-C6D0
        0xC6D4, # G+2584  U+C6D4  uniC6D4,  HANGUL SYLLABLE-C6D4
        0xC6DC, # G+2585  U+C6DC  uniC6DC,  HANGUL SYLLABLE-C6DC
        0xC6DD, # G+2586  U+C6DD  uniC6DD,  HANGUL SYLLABLE-C6DD
        0xC6E0, # G+2587  U+C6E0  uniC6E0,  HANGUL SYLLABLE-C6E0
        0xC6E1, # G+2588  U+C6E1  uniC6E1,  HANGUL SYLLABLE-C6E1
        0xC6E8, # G+2589  U+C6E8  uniC6E8,  HANGUL SYLLABLE-C6E8
        0xC6E9, # G+2590  U+C6E9  uniC6E9,  HANGUL SYLLABLE-C6E9
        0xC6EC, # G+2591  U+C6EC  uniC6EC,  HANGUL SYLLABLE-C6EC
        0xC6F0, # G+2592  U+C6F0  uniC6F0,  HANGUL SYLLABLE-C6F0
        0xC6F8, # G+2593  U+C6F8  uniC6F8,  HANGUL SYLLABLE-C6F8
        0xC6F9, # G+2594  U+C6F9  uniC6F9,  HANGUL SYLLABLE-C6F9
        0xC6FD, # G+2595  U+C6FD  uniC6FD,  HANGUL SYLLABLE-C6FD
        0xC704, # G+2596  U+C704  uniC704,  HANGUL SYLLABLE-C704
        0xC705, # G+2597  U+C705  uniC705,  HANGUL SYLLABLE-C705
        0xC708, # G+2598  U+C708  uniC708,  HANGUL SYLLABLE-C708
        0xC70C, # G+2599  U+C70C  uniC70C,  HANGUL SYLLABLE-C70C
        0xC714, # G+2600  U+C714  uniC714,  HANGUL SYLLABLE-C714
        0xC715, # G+2601  U+C715  uniC715,  HANGUL SYLLABLE-C715
        0xC717, # G+2602  U+C717  uniC717,  HANGUL SYLLABLE-C717
        0xC719, # G+2603  U+C719  uniC719,  HANGUL SYLLABLE-C719
        0xC720, # G+2604  U+C720  uniC720,  HANGUL SYLLABLE-C720
        0xC721, # G+2605  U+C721  uniC721,  HANGUL SYLLABLE-C721
        0xC724, # G+2606  U+C724  uniC724,  HANGUL SYLLABLE-C724
        0xC728, # G+2607  U+C728  uniC728,  HANGUL SYLLABLE-C728
        0xC730, # G+2608  U+C730  uniC730,  HANGUL SYLLABLE-C730
        0xC731, # G+2609  U+C731  uniC731,  HANGUL SYLLABLE-C731
        0xC733, # G+2610  U+C733  uniC733,  HANGUL SYLLABLE-C733
        0xC735, # G+2611  U+C735  uniC735,  HANGUL SYLLABLE-C735
        0xC737, # G+2612  U+C737  uniC737,  HANGUL SYLLABLE-C737
        0xC73C, # G+2613  U+C73C  uniC73C,  HANGUL SYLLABLE-C73C
        0xC73D, # G+2614  U+C73D  uniC73D,  HANGUL SYLLABLE-C73D
        0xC740, # G+2615  U+C740  uniC740,  HANGUL SYLLABLE-C740
        0xC744, # G+2616  U+C744  uniC744,  HANGUL SYLLABLE-C744
        0xC74A, # G+2617  U+C74A  uniC74A,  HANGUL SYLLABLE-C74A
        0xC74C, # G+2618  U+C74C  uniC74C,  HANGUL SYLLABLE-C74C
        0xC74D, # G+2619  U+C74D  uniC74D,  HANGUL SYLLABLE-C74D
        0xC74F, # G+2620  U+C74F  uniC74F,  HANGUL SYLLABLE-C74F
        0xC751, # G+2621  U+C751  uniC751,  HANGUL SYLLABLE-C751
        0xC752, # G+2622  U+C752  uniC752,  HANGUL SYLLABLE-C752
        0xC753, # G+2623  U+C753  uniC753,  HANGUL SYLLABLE-C753
        0xC754, # G+2624  U+C754  uniC754,  HANGUL SYLLABLE-C754
        0xC755, # G+2625  U+C755  uniC755,  HANGUL SYLLABLE-C755
        0xC756, # G+2626  U+C756  uniC756,  HANGUL SYLLABLE-C756
        0xC757, # G+2627  U+C757  uniC757,  HANGUL SYLLABLE-C757
        0xC758, # G+2628  U+C758  uniC758,  HANGUL SYLLABLE-C758
        0xC75C, # G+2629  U+C75C  uniC75C,  HANGUL SYLLABLE-C75C
        0xC760, # G+2630  U+C760  uniC760,  HANGUL SYLLABLE-C760
        0xC768, # G+2631  U+C768  uniC768,  HANGUL SYLLABLE-C768
        0xC76B, # G+2632  U+C76B  uniC76B,  HANGUL SYLLABLE-C76B
        0xC774, # G+2633  U+C774  uniC774,  HANGUL SYLLABLE-C774
        0xC775, # G+2634  U+C775  uniC775,  HANGUL SYLLABLE-C775
        0xC778, # G+2635  U+C778  uniC778,  HANGUL SYLLABLE-C778
        0xC77C, # G+2636  U+C77C  uniC77C,  HANGUL SYLLABLE-C77C
        0xC77D, # G+2637  U+C77D  uniC77D,  HANGUL SYLLABLE-C77D
        0xC77E, # G+2638  U+C77E  uniC77E,  HANGUL SYLLABLE-C77E
        0xC783, # G+2639  U+C783  uniC783,  HANGUL SYLLABLE-C783
        0xC784, # G+2640  U+C784  uniC784,  HANGUL SYLLABLE-C784
        0xC785, # G+2641  U+C785  uniC785,  HANGUL SYLLABLE-C785
        0xC787, # G+2642  U+C787  uniC787,  HANGUL SYLLABLE-C787
        0xC788, # G+2643  U+C788  uniC788,  HANGUL SYLLABLE-C788
        0xC789, # G+2644  U+C789  uniC789,  HANGUL SYLLABLE-C789
        0xC78A, # G+2645  U+C78A  uniC78A,  HANGUL SYLLABLE-C78A
        0xC78E, # G+2646  U+C78E  uniC78E,  HANGUL SYLLABLE-C78E
        0xC790, # G+2647  U+C790  uniC790,  HANGUL SYLLABLE-C790
        0xC791, # G+2648  U+C791  uniC791,  HANGUL SYLLABLE-C791
        0xC794, # G+2649  U+C794  uniC794,  HANGUL SYLLABLE-C794
        0xC796, # G+2650  U+C796  uniC796,  HANGUL SYLLABLE-C796
        0xC797, # G+2651  U+C797  uniC797,  HANGUL SYLLABLE-C797
        0xC798, # G+2652  U+C798  uniC798,  HANGUL SYLLABLE-C798
        0xC79A, # G+2653  U+C79A  uniC79A,  HANGUL SYLLABLE-C79A
        0xC7A0, # G+2654  U+C7A0  uniC7A0,  HANGUL SYLLABLE-C7A0
        0xC7A1, # G+2655  U+C7A1  uniC7A1,  HANGUL SYLLABLE-C7A1
        0xC7A3, # G+2656  U+C7A3  uniC7A3,  HANGUL SYLLABLE-C7A3
        0xC7A4, # G+2657  U+C7A4  uniC7A4,  HANGUL SYLLABLE-C7A4
        0xC7A5, # G+2658  U+C7A5  uniC7A5,  HANGUL SYLLABLE-C7A5
        0xC7A6, # G+2659  U+C7A6  uniC7A6,  HANGUL SYLLABLE-C7A6
        0xC7AC, # G+2660  U+C7AC  uniC7AC,  HANGUL SYLLABLE-C7AC
        0xC7AD, # G+2661  U+C7AD  uniC7AD,  HANGUL SYLLABLE-C7AD
        0xC7B0, # G+2662  U+C7B0  uniC7B0,  HANGUL SYLLABLE-C7B0
        0xC7B4, # G+2663  U+C7B4  uniC7B4,  HANGUL SYLLABLE-C7B4
        0xC7BC, # G+2664  U+C7BC  uniC7BC,  HANGUL SYLLABLE-C7BC
        0xC7BD, # G+2665  U+C7BD  uniC7BD,  HANGUL SYLLABLE-C7BD
        0xC7BF, # G+2666  U+C7BF  uniC7BF,  HANGUL SYLLABLE-C7BF
        0xC7C0, # G+2667  U+C7C0  uniC7C0,  HANGUL SYLLABLE-C7C0
        0xC7C1, # G+2668  U+C7C1  uniC7C1,  HANGUL SYLLABLE-C7C1
        0xC7C8, # G+2669  U+C7C8  uniC7C8,  HANGUL SYLLABLE-C7C8
        0xC7C9, # G+2670  U+C7C9  uniC7C9,  HANGUL SYLLABLE-C7C9
        0xC7CC, # G+2671  U+C7CC  uniC7CC,  HANGUL SYLLABLE-C7CC
        0xC7CE, # G+2672  U+C7CE  uniC7CE,  HANGUL SYLLABLE-C7CE
        0xC7D0, # G+2673  U+C7D0  uniC7D0,  HANGUL SYLLABLE-C7D0
        0xC7D8, # G+2674  U+C7D8  uniC7D8,  HANGUL SYLLABLE-C7D8
        0xC7DD, # G+2675  U+C7DD  uniC7DD,  HANGUL SYLLABLE-C7DD
        0xC7E4, # G+2676  U+C7E4  uniC7E4,  HANGUL SYLLABLE-C7E4
        0xC7E8, # G+2677  U+C7E8  uniC7E8,  HANGUL SYLLABLE-C7E8
        0xC7EC, # G+2678  U+C7EC  uniC7EC,  HANGUL SYLLABLE-C7EC
        0xC800, # G+2679  U+C800  uniC800,  HANGUL SYLLABLE-C800
        0xC801, # G+2680  U+C801  uniC801,  HANGUL SYLLABLE-C801
        0xC804, # G+2681  U+C804  uniC804,  HANGUL SYLLABLE-C804
        0xC808, # G+2682  U+C808  uniC808,  HANGUL SYLLABLE-C808
        0xC80A, # G+2683  U+C80A  uniC80A,  HANGUL SYLLABLE-C80A
        0xC810, # G+2684  U+C810  uniC810,  HANGUL SYLLABLE-C810
        0xC811, # G+2685  U+C811  uniC811,  HANGUL SYLLABLE-C811
        0xC813, # G+2686  U+C813  uniC813,  HANGUL SYLLABLE-C813
        0xC815, # G+2687  U+C815  uniC815,  HANGUL SYLLABLE-C815
        0xC816, # G+2688  U+C816  uniC816,  HANGUL SYLLABLE-C816
        0xC81C, # G+2689  U+C81C  uniC81C,  HANGUL SYLLABLE-C81C
        0xC81D, # G+2690  U+C81D  uniC81D,  HANGUL SYLLABLE-C81D
        0xC820, # G+2691  U+C820  uniC820,  HANGUL SYLLABLE-C820
        0xC824, # G+2692  U+C824  uniC824,  HANGUL SYLLABLE-C824
        0xC82C, # G+2693  U+C82C  uniC82C,  HANGUL SYLLABLE-C82C
        0xC82D, # G+2694  U+C82D  uniC82D,  HANGUL SYLLABLE-C82D
        0xC82F, # G+2695  U+C82F  uniC82F,  HANGUL SYLLABLE-C82F
        0xC831, # G+2696  U+C831  uniC831,  HANGUL SYLLABLE-C831
        0xC838, # G+2697  U+C838  uniC838,  HANGUL SYLLABLE-C838
        0xC83C, # G+2698  U+C83C  uniC83C,  HANGUL SYLLABLE-C83C
        0xC840, # G+2699  U+C840  uniC840,  HANGUL SYLLABLE-C840
        0xC848, # G+2700  U+C848  uniC848,  HANGUL SYLLABLE-C848
        0xC849, # G+2701  U+C849  uniC849,  HANGUL SYLLABLE-C849
        0xC84C, # G+2702  U+C84C  uniC84C,  HANGUL SYLLABLE-C84C
        0xC84D, # G+2703  U+C84D  uniC84D,  HANGUL SYLLABLE-C84D
        0xC854, # G+2704  U+C854  uniC854,  HANGUL SYLLABLE-C854
        0xC870, # G+2705  U+C870  uniC870,  HANGUL SYLLABLE-C870
        0xC871, # G+2706  U+C871  uniC871,  HANGUL SYLLABLE-C871
        0xC874, # G+2707  U+C874  uniC874,  HANGUL SYLLABLE-C874
        0xC878, # G+2708  U+C878  uniC878,  HANGUL SYLLABLE-C878
        0xC87A, # G+2709  U+C87A  uniC87A,  HANGUL SYLLABLE-C87A
        0xC880, # G+2710  U+C880  uniC880,  HANGUL SYLLABLE-C880
        0xC881, # G+2711  U+C881  uniC881,  HANGUL SYLLABLE-C881
        0xC883, # G+2712  U+C883  uniC883,  HANGUL SYLLABLE-C883
        0xC885, # G+2713  U+C885  uniC885,  HANGUL SYLLABLE-C885
        0xC886, # G+2714  U+C886  uniC886,  HANGUL SYLLABLE-C886
        0xC887, # G+2715  U+C887  uniC887,  HANGUL SYLLABLE-C887
        0xC88B, # G+2716  U+C88B  uniC88B,  HANGUL SYLLABLE-C88B
        0xC88C, # G+2717  U+C88C  uniC88C,  HANGUL SYLLABLE-C88C
        0xC88D, # G+2718  U+C88D  uniC88D,  HANGUL SYLLABLE-C88D
        0xC894, # G+2719  U+C894  uniC894,  HANGUL SYLLABLE-C894
        0xC89D, # G+2720  U+C89D  uniC89D,  HANGUL SYLLABLE-C89D
        0xC89F, # G+2721  U+C89F  uniC89F,  HANGUL SYLLABLE-C89F
        0xC8A1, # G+2722  U+C8A1  uniC8A1,  HANGUL SYLLABLE-C8A1
        0xC8A8, # G+2723  U+C8A8  uniC8A8,  HANGUL SYLLABLE-C8A8
        0xC8BC, # G+2724  U+C8BC  uniC8BC,  HANGUL SYLLABLE-C8BC
        0xC8BD, # G+2725  U+C8BD  uniC8BD,  HANGUL SYLLABLE-C8BD
        0xC8C4, # G+2726  U+C8C4  uniC8C4,  HANGUL SYLLABLE-C8C4
        0xC8C8, # G+2727  U+C8C8  uniC8C8,  HANGUL SYLLABLE-C8C8
        0xC8CC, # G+2728  U+C8CC  uniC8CC,  HANGUL SYLLABLE-C8CC
        0xC8D4, # G+2729  U+C8D4  uniC8D4,  HANGUL SYLLABLE-C8D4
        0xC8D5, # G+2730  U+C8D5  uniC8D5,  HANGUL SYLLABLE-C8D5
        0xC8D7, # G+2731  U+C8D7  uniC8D7,  HANGUL SYLLABLE-C8D7
        0xC8D9, # G+2732  U+C8D9  uniC8D9,  HANGUL SYLLABLE-C8D9
        0xC8E0, # G+2733  U+C8E0  uniC8E0,  HANGUL SYLLABLE-C8E0
        0xC8E1, # G+2734  U+C8E1  uniC8E1,  HANGUL SYLLABLE-C8E1
        0xC8E4, # G+2735  U+C8E4  uniC8E4,  HANGUL SYLLABLE-C8E4
        0xC8F5, # G+2736  U+C8F5  uniC8F5,  HANGUL SYLLABLE-C8F5
        0xC8FC, # G+2737  U+C8FC  uniC8FC,  HANGUL SYLLABLE-C8FC
        0xC8FD, # G+2738  U+C8FD  uniC8FD,  HANGUL SYLLABLE-C8FD
        0xC900, # G+2739  U+C900  uniC900,  HANGUL SYLLABLE-C900
        0xC904, # G+2740  U+C904  uniC904,  HANGUL SYLLABLE-C904
        0xC905, # G+2741  U+C905  uniC905,  HANGUL SYLLABLE-C905
        0xC906, # G+2742  U+C906  uniC906,  HANGUL SYLLABLE-C906
        0xC90C, # G+2743  U+C90C  uniC90C,  HANGUL SYLLABLE-C90C
        0xC90D, # G+2744  U+C90D  uniC90D,  HANGUL SYLLABLE-C90D
        0xC90F, # G+2745  U+C90F  uniC90F,  HANGUL SYLLABLE-C90F
        0xC911, # G+2746  U+C911  uniC911,  HANGUL SYLLABLE-C911
        0xC918, # G+2747  U+C918  uniC918,  HANGUL SYLLABLE-C918
        0xC92C, # G+2748  U+C92C  uniC92C,  HANGUL SYLLABLE-C92C
        0xC934, # G+2749  U+C934  uniC934,  HANGUL SYLLABLE-C934
        0xC950, # G+2750  U+C950  uniC950,  HANGUL SYLLABLE-C950
        0xC951, # G+2751  U+C951  uniC951,  HANGUL SYLLABLE-C951
        0xC954, # G+2752  U+C954  uniC954,  HANGUL SYLLABLE-C954
        0xC958, # G+2753  U+C958  uniC958,  HANGUL SYLLABLE-C958
        0xC960, # G+2754  U+C960  uniC960,  HANGUL SYLLABLE-C960
        0xC961, # G+2755  U+C961  uniC961,  HANGUL SYLLABLE-C961
        0xC963, # G+2756  U+C963  uniC963,  HANGUL SYLLABLE-C963
        0xC96C, # G+2757  U+C96C  uniC96C,  HANGUL SYLLABLE-C96C
        0xC970, # G+2758  U+C970  uniC970,  HANGUL SYLLABLE-C970
        0xC974, # G+2759  U+C974  uniC974,  HANGUL SYLLABLE-C974
        0xC97C, # G+2760  U+C97C  uniC97C,  HANGUL SYLLABLE-C97C
        0xC988, # G+2761  U+C988  uniC988,  HANGUL SYLLABLE-C988
        0xC989, # G+2762  U+C989  uniC989,  HANGUL SYLLABLE-C989
        0xC98C, # G+2763  U+C98C  uniC98C,  HANGUL SYLLABLE-C98C
        0xC990, # G+2764  U+C990  uniC990,  HANGUL SYLLABLE-C990
        0xC998, # G+2765  U+C998  uniC998,  HANGUL SYLLABLE-C998
        0xC999, # G+2766  U+C999  uniC999,  HANGUL SYLLABLE-C999
        0xC99B, # G+2767  U+C99B  uniC99B,  HANGUL SYLLABLE-C99B
        0xC99D, # G+2768  U+C99D  uniC99D,  HANGUL SYLLABLE-C99D
        0xC9C0, # G+2769  U+C9C0  uniC9C0,  HANGUL SYLLABLE-C9C0
        0xC9C1, # G+2770  U+C9C1  uniC9C1,  HANGUL SYLLABLE-C9C1
        0xC9C4, # G+2771  U+C9C4  uniC9C4,  HANGUL SYLLABLE-C9C4
        0xC9C7, # G+2772  U+C9C7  uniC9C7,  HANGUL SYLLABLE-C9C7
        0xC9C8, # G+2773  U+C9C8  uniC9C8,  HANGUL SYLLABLE-C9C8
        0xC9CA, # G+2774  U+C9CA  uniC9CA,  HANGUL SYLLABLE-C9CA
        0xC9D0, # G+2775  U+C9D0  uniC9D0,  HANGUL SYLLABLE-C9D0
        0xC9D1, # G+2776  U+C9D1  uniC9D1,  HANGUL SYLLABLE-C9D1
        0xC9D3, # G+2777  U+C9D3  uniC9D3,  HANGUL SYLLABLE-C9D3
        0xC9D5, # G+2778  U+C9D5  uniC9D5,  HANGUL SYLLABLE-C9D5
        0xC9D6, # G+2779  U+C9D6  uniC9D6,  HANGUL SYLLABLE-C9D6
        0xC9D9, # G+2780  U+C9D9  uniC9D9,  HANGUL SYLLABLE-C9D9
        0xC9DA, # G+2781  U+C9DA  uniC9DA,  HANGUL SYLLABLE-C9DA
        0xC9DC, # G+2782  U+C9DC  uniC9DC,  HANGUL SYLLABLE-C9DC
        0xC9DD, # G+2783  U+C9DD  uniC9DD,  HANGUL SYLLABLE-C9DD
        0xC9E0, # G+2784  U+C9E0  uniC9E0,  HANGUL SYLLABLE-C9E0
        0xC9E2, # G+2785  U+C9E2  uniC9E2,  HANGUL SYLLABLE-C9E2
        0xC9E4, # G+2786  U+C9E4  uniC9E4,  HANGUL SYLLABLE-C9E4
        0xC9E7, # G+2787  U+C9E7  uniC9E7,  HANGUL SYLLABLE-C9E7
        0xC9EC, # G+2788  U+C9EC  uniC9EC,  HANGUL SYLLABLE-C9EC
        0xC9ED, # G+2789  U+C9ED  uniC9ED,  HANGUL SYLLABLE-C9ED
        0xC9EF, # G+2790  U+C9EF  uniC9EF,  HANGUL SYLLABLE-C9EF
        0xC9F0, # G+2791  U+C9F0  uniC9F0,  HANGUL SYLLABLE-C9F0
        0xC9F1, # G+2792  U+C9F1  uniC9F1,  HANGUL SYLLABLE-C9F1
        0xC9F8, # G+2793  U+C9F8  uniC9F8,  HANGUL SYLLABLE-C9F8
        0xC9F9, # G+2794  U+C9F9  uniC9F9,  HANGUL SYLLABLE-C9F9
        0xC9FC, # G+2795  U+C9FC  uniC9FC,  HANGUL SYLLABLE-C9FC
        0xCA00, # G+2796  U+CA00  uniCA00,  HANGUL SYLLABLE-CA00
        0xCA08, # G+2797  U+CA08  uniCA08,  HANGUL SYLLABLE-CA08
        0xCA09, # G+2798  U+CA09  uniCA09,  HANGUL SYLLABLE-CA09
        0xCA0B, # G+2799  U+CA0B  uniCA0B,  HANGUL SYLLABLE-CA0B
        0xCA0C, # G+2800  U+CA0C  uniCA0C,  HANGUL SYLLABLE-CA0C
        0xCA0D, # G+2801  U+CA0D  uniCA0D,  HANGUL SYLLABLE-CA0D
        0xCA14, # G+2802  U+CA14  uniCA14,  HANGUL SYLLABLE-CA14
        0xCA18, # G+2803  U+CA18  uniCA18,  HANGUL SYLLABLE-CA18
        0xCA29, # G+2804  U+CA29  uniCA29,  HANGUL SYLLABLE-CA29
        0xCA4C, # G+2805  U+CA4C  uniCA4C,  HANGUL SYLLABLE-CA4C
        0xCA4D, # G+2806  U+CA4D  uniCA4D,  HANGUL SYLLABLE-CA4D
        0xCA50, # G+2807  U+CA50  uniCA50,  HANGUL SYLLABLE-CA50
        0xCA54, # G+2808  U+CA54  uniCA54,  HANGUL SYLLABLE-CA54
        0xCA5C, # G+2809  U+CA5C  uniCA5C,  HANGUL SYLLABLE-CA5C
        0xCA5D, # G+2810  U+CA5D  uniCA5D,  HANGUL SYLLABLE-CA5D
        0xCA5F, # G+2811  U+CA5F  uniCA5F,  HANGUL SYLLABLE-CA5F
        0xCA60, # G+2812  U+CA60  uniCA60,  HANGUL SYLLABLE-CA60
        0xCA61, # G+2813  U+CA61  uniCA61,  HANGUL SYLLABLE-CA61
        0xCA68, # G+2814  U+CA68  uniCA68,  HANGUL SYLLABLE-CA68
        0xCA7D, # G+2815  U+CA7D  uniCA7D,  HANGUL SYLLABLE-CA7D
        0xCA84, # G+2816  U+CA84  uniCA84,  HANGUL SYLLABLE-CA84
        0xCA98, # G+2817  U+CA98  uniCA98,  HANGUL SYLLABLE-CA98
        0xCABC, # G+2818  U+CABC  uniCABC,  HANGUL SYLLABLE-CABC
        0xCABD, # G+2819  U+CABD  uniCABD,  HANGUL SYLLABLE-CABD
        0xCAC0, # G+2820  U+CAC0  uniCAC0,  HANGUL SYLLABLE-CAC0
        0xCAC4, # G+2821  U+CAC4  uniCAC4,  HANGUL SYLLABLE-CAC4
        0xCACC, # G+2822  U+CACC  uniCACC,  HANGUL SYLLABLE-CACC
        0xCACD, # G+2823  U+CACD  uniCACD,  HANGUL SYLLABLE-CACD
        0xCACF, # G+2824  U+CACF  uniCACF,  HANGUL SYLLABLE-CACF
        0xCAD1, # G+2825  U+CAD1  uniCAD1,  HANGUL SYLLABLE-CAD1
        0xCAD3, # G+2826  U+CAD3  uniCAD3,  HANGUL SYLLABLE-CAD3
        0xCAD8, # G+2827  U+CAD8  uniCAD8,  HANGUL SYLLABLE-CAD8
        0xCAD9, # G+2828  U+CAD9  uniCAD9,  HANGUL SYLLABLE-CAD9
        0xCAE0, # G+2829  U+CAE0  uniCAE0,  HANGUL SYLLABLE-CAE0
        0xCAEC, # G+2830  U+CAEC  uniCAEC,  HANGUL SYLLABLE-CAEC
        0xCAF4, # G+2831  U+CAF4  uniCAF4,  HANGUL SYLLABLE-CAF4
        0xCB08, # G+2832  U+CB08  uniCB08,  HANGUL SYLLABLE-CB08
        0xCB10, # G+2833  U+CB10  uniCB10,  HANGUL SYLLABLE-CB10
        0xCB14, # G+2834  U+CB14  uniCB14,  HANGUL SYLLABLE-CB14
        0xCB18, # G+2835  U+CB18  uniCB18,  HANGUL SYLLABLE-CB18
        0xCB20, # G+2836  U+CB20  uniCB20,  HANGUL SYLLABLE-CB20
        0xCB21, # G+2837  U+CB21  uniCB21,  HANGUL SYLLABLE-CB21
        0xCB41, # G+2838  U+CB41  uniCB41,  HANGUL SYLLABLE-CB41
        0xCB48, # G+2839  U+CB48  uniCB48,  HANGUL SYLLABLE-CB48
        0xCB49, # G+2840  U+CB49  uniCB49,  HANGUL SYLLABLE-CB49
        0xCB4C, # G+2841  U+CB4C  uniCB4C,  HANGUL SYLLABLE-CB4C
        0xCB50, # G+2842  U+CB50  uniCB50,  HANGUL SYLLABLE-CB50
        0xCB58, # G+2843  U+CB58  uniCB58,  HANGUL SYLLABLE-CB58
        0xCB59, # G+2844  U+CB59  uniCB59,  HANGUL SYLLABLE-CB59
        0xCB5D, # G+2845  U+CB5D  uniCB5D,  HANGUL SYLLABLE-CB5D
        0xCB64, # G+2846  U+CB64  uniCB64,  HANGUL SYLLABLE-CB64
        0xCB78, # G+2847  U+CB78  uniCB78,  HANGUL SYLLABLE-CB78
        0xCB79, # G+2848  U+CB79  uniCB79,  HANGUL SYLLABLE-CB79
        0xCB9C, # G+2849  U+CB9C  uniCB9C,  HANGUL SYLLABLE-CB9C
        0xCBB8, # G+2850  U+CBB8  uniCBB8,  HANGUL SYLLABLE-CBB8
        0xCBD4, # G+2851  U+CBD4  uniCBD4,  HANGUL SYLLABLE-CBD4
        0xCBE4, # G+2852  U+CBE4  uniCBE4,  HANGUL SYLLABLE-CBE4
        0xCBE7, # G+2853  U+CBE7  uniCBE7,  HANGUL SYLLABLE-CBE7
        0xCBE9, # G+2854  U+CBE9  uniCBE9,  HANGUL SYLLABLE-CBE9
        0xCC0C, # G+2855  U+CC0C  uniCC0C,  HANGUL SYLLABLE-CC0C
        0xCC0D, # G+2856  U+CC0D  uniCC0D,  HANGUL SYLLABLE-CC0D
        0xCC10, # G+2857  U+CC10  uniCC10,  HANGUL SYLLABLE-CC10
        0xCC14, # G+2858  U+CC14  uniCC14,  HANGUL SYLLABLE-CC14
        0xCC1C, # G+2859  U+CC1C  uniCC1C,  HANGUL SYLLABLE-CC1C
        0xCC1D, # G+2860  U+CC1D  uniCC1D,  HANGUL SYLLABLE-CC1D
        0xCC21, # G+2861  U+CC21  uniCC21,  HANGUL SYLLABLE-CC21
        0xCC22, # G+2862  U+CC22  uniCC22,  HANGUL SYLLABLE-CC22
        0xCC27, # G+2863  U+CC27  uniCC27,  HANGUL SYLLABLE-CC27
        0xCC28, # G+2864  U+CC28  uniCC28,  HANGUL SYLLABLE-CC28
        0xCC29, # G+2865  U+CC29  uniCC29,  HANGUL SYLLABLE-CC29
        0xCC2C, # G+2866  U+CC2C  uniCC2C,  HANGUL SYLLABLE-CC2C
        0xCC2E, # G+2867  U+CC2E  uniCC2E,  HANGUL SYLLABLE-CC2E
        0xCC30, # G+2868  U+CC30  uniCC30,  HANGUL SYLLABLE-CC30
        0xCC38, # G+2869  U+CC38  uniCC38,  HANGUL SYLLABLE-CC38
        0xCC39, # G+2870  U+CC39  uniCC39,  HANGUL SYLLABLE-CC39
        0xCC3B, # G+2871  U+CC3B  uniCC3B,  HANGUL SYLLABLE-CC3B
        0xCC3C, # G+2872  U+CC3C  uniCC3C,  HANGUL SYLLABLE-CC3C
        0xCC3D, # G+2873  U+CC3D  uniCC3D,  HANGUL SYLLABLE-CC3D
        0xCC3E, # G+2874  U+CC3E  uniCC3E,  HANGUL SYLLABLE-CC3E
        0xCC44, # G+2875  U+CC44  uniCC44,  HANGUL SYLLABLE-CC44
        0xCC45, # G+2876  U+CC45  uniCC45,  HANGUL SYLLABLE-CC45
        0xCC48, # G+2877  U+CC48  uniCC48,  HANGUL SYLLABLE-CC48
        0xCC4C, # G+2878  U+CC4C  uniCC4C,  HANGUL SYLLABLE-CC4C
        0xCC54, # G+2879  U+CC54  uniCC54,  HANGUL SYLLABLE-CC54
        0xCC55, # G+2880  U+CC55  uniCC55,  HANGUL SYLLABLE-CC55
        0xCC57, # G+2881  U+CC57  uniCC57,  HANGUL SYLLABLE-CC57
        0xCC58, # G+2882  U+CC58  uniCC58,  HANGUL SYLLABLE-CC58
        0xCC59, # G+2883  U+CC59  uniCC59,  HANGUL SYLLABLE-CC59
        0xCC60, # G+2884  U+CC60  uniCC60,  HANGUL SYLLABLE-CC60
        0xCC64, # G+2885  U+CC64  uniCC64,  HANGUL SYLLABLE-CC64
        0xCC66, # G+2886  U+CC66  uniCC66,  HANGUL SYLLABLE-CC66
        0xCC68, # G+2887  U+CC68  uniCC68,  HANGUL SYLLABLE-CC68
        0xCC70, # G+2888  U+CC70  uniCC70,  HANGUL SYLLABLE-CC70
        0xCC75, # G+2889  U+CC75  uniCC75,  HANGUL SYLLABLE-CC75
        0xCC98, # G+2890  U+CC98  uniCC98,  HANGUL SYLLABLE-CC98
        0xCC99, # G+2891  U+CC99  uniCC99,  HANGUL SYLLABLE-CC99
        0xCC9C, # G+2892  U+CC9C  uniCC9C,  HANGUL SYLLABLE-CC9C
        0xCCA0, # G+2893  U+CCA0  uniCCA0,  HANGUL SYLLABLE-CCA0
        0xCCA8, # G+2894  U+CCA8  uniCCA8,  HANGUL SYLLABLE-CCA8
        0xCCA9, # G+2895  U+CCA9  uniCCA9,  HANGUL SYLLABLE-CCA9
        0xCCAB, # G+2896  U+CCAB  uniCCAB,  HANGUL SYLLABLE-CCAB
        0xCCAC, # G+2897  U+CCAC  uniCCAC,  HANGUL SYLLABLE-CCAC
        0xCCAD, # G+2898  U+CCAD  uniCCAD,  HANGUL SYLLABLE-CCAD
        0xCCB4, # G+2899  U+CCB4  uniCCB4,  HANGUL SYLLABLE-CCB4
        0xCCB5, # G+2900  U+CCB5  uniCCB5,  HANGUL SYLLABLE-CCB5
        0xCCB8, # G+2901  U+CCB8  uniCCB8,  HANGUL SYLLABLE-CCB8
        0xCCBC, # G+2902  U+CCBC  uniCCBC,  HANGUL SYLLABLE-CCBC
        0xCCC4, # G+2903  U+CCC4  uniCCC4,  HANGUL SYLLABLE-CCC4
        0xCCC5, # G+2904  U+CCC5  uniCCC5,  HANGUL SYLLABLE-CCC5
        0xCCC7, # G+2905  U+CCC7  uniCCC7,  HANGUL SYLLABLE-CCC7
        0xCCC9, # G+2906  U+CCC9  uniCCC9,  HANGUL SYLLABLE-CCC9
        0xCCD0, # G+2907  U+CCD0  uniCCD0,  HANGUL SYLLABLE-CCD0
        0xCCD4, # G+2908  U+CCD4  uniCCD4,  HANGUL SYLLABLE-CCD4
        0xCCE4, # G+2909  U+CCE4  uniCCE4,  HANGUL SYLLABLE-CCE4
        0xCCEC, # G+2910  U+CCEC  uniCCEC,  HANGUL SYLLABLE-CCEC
        0xCCF0, # G+2911  U+CCF0  uniCCF0,  HANGUL SYLLABLE-CCF0
        0xCD01, # G+2912  U+CD01  uniCD01,  HANGUL SYLLABLE-CD01
        0xCD08, # G+2913  U+CD08  uniCD08,  HANGUL SYLLABLE-CD08
        0xCD09, # G+2914  U+CD09  uniCD09,  HANGUL SYLLABLE-CD09
        0xCD0C, # G+2915  U+CD0C  uniCD0C,  HANGUL SYLLABLE-CD0C
        0xCD10, # G+2916  U+CD10  uniCD10,  HANGUL SYLLABLE-CD10
        0xCD18, # G+2917  U+CD18  uniCD18,  HANGUL SYLLABLE-CD18
        0xCD19, # G+2918  U+CD19  uniCD19,  HANGUL SYLLABLE-CD19
        0xCD1B, # G+2919  U+CD1B  uniCD1B,  HANGUL SYLLABLE-CD1B
        0xCD1D, # G+2920  U+CD1D  uniCD1D,  HANGUL SYLLABLE-CD1D
        0xCD24, # G+2921  U+CD24  uniCD24,  HANGUL SYLLABLE-CD24
        0xCD28, # G+2922  U+CD28  uniCD28,  HANGUL SYLLABLE-CD28
        0xCD2C, # G+2923  U+CD2C  uniCD2C,  HANGUL SYLLABLE-CD2C
        0xCD39, # G+2924  U+CD39  uniCD39,  HANGUL SYLLABLE-CD39
        0xCD5C, # G+2925  U+CD5C  uniCD5C,  HANGUL SYLLABLE-CD5C
        0xCD60, # G+2926  U+CD60  uniCD60,  HANGUL SYLLABLE-CD60
        0xCD64, # G+2927  U+CD64  uniCD64,  HANGUL SYLLABLE-CD64
        0xCD6C, # G+2928  U+CD6C  uniCD6C,  HANGUL SYLLABLE-CD6C
        0xCD6D, # G+2929  U+CD6D  uniCD6D,  HANGUL SYLLABLE-CD6D
        0xCD6F, # G+2930  U+CD6F  uniCD6F,  HANGUL SYLLABLE-CD6F
        0xCD71, # G+2931  U+CD71  uniCD71,  HANGUL SYLLABLE-CD71
        0xCD78, # G+2932  U+CD78  uniCD78,  HANGUL SYLLABLE-CD78
        0xCD88, # G+2933  U+CD88  uniCD88,  HANGUL SYLLABLE-CD88
        0xCD94, # G+2934  U+CD94  uniCD94,  HANGUL SYLLABLE-CD94
        0xCD95, # G+2935  U+CD95  uniCD95,  HANGUL SYLLABLE-CD95
        0xCD98, # G+2936  U+CD98  uniCD98,  HANGUL SYLLABLE-CD98
        0xCD9C, # G+2937  U+CD9C  uniCD9C,  HANGUL SYLLABLE-CD9C
        0xCDA4, # G+2938  U+CDA4  uniCDA4,  HANGUL SYLLABLE-CDA4
        0xCDA5, # G+2939  U+CDA5  uniCDA5,  HANGUL SYLLABLE-CDA5
        0xCDA7, # G+2940  U+CDA7  uniCDA7,  HANGUL SYLLABLE-CDA7
        0xCDA9, # G+2941  U+CDA9  uniCDA9,  HANGUL SYLLABLE-CDA9
        0xCDB0, # G+2942  U+CDB0  uniCDB0,  HANGUL SYLLABLE-CDB0
        0xCDC4, # G+2943  U+CDC4  uniCDC4,  HANGUL SYLLABLE-CDC4
        0xCDCC, # G+2944  U+CDCC  uniCDCC,  HANGUL SYLLABLE-CDCC
        0xCDD0, # G+2945  U+CDD0  uniCDD0,  HANGUL SYLLABLE-CDD0
        0xCDE8, # G+2946  U+CDE8  uniCDE8,  HANGUL SYLLABLE-CDE8
        0xCDEC, # G+2947  U+CDEC  uniCDEC,  HANGUL SYLLABLE-CDEC
        0xCDF0, # G+2948  U+CDF0  uniCDF0,  HANGUL SYLLABLE-CDF0
        0xCDF8, # G+2949  U+CDF8  uniCDF8,  HANGUL SYLLABLE-CDF8
        0xCDF9, # G+2950  U+CDF9  uniCDF9,  HANGUL SYLLABLE-CDF9
        0xCDFB, # G+2951  U+CDFB  uniCDFB,  HANGUL SYLLABLE-CDFB
        0xCDFD, # G+2952  U+CDFD  uniCDFD,  HANGUL SYLLABLE-CDFD
        0xCE04, # G+2953  U+CE04  uniCE04,  HANGUL SYLLABLE-CE04
        0xCE08, # G+2954  U+CE08  uniCE08,  HANGUL SYLLABLE-CE08
        0xCE0C, # G+2955  U+CE0C  uniCE0C,  HANGUL SYLLABLE-CE0C
        0xCE14, # G+2956  U+CE14  uniCE14,  HANGUL SYLLABLE-CE14
        0xCE19, # G+2957  U+CE19  uniCE19,  HANGUL SYLLABLE-CE19
        0xCE20, # G+2958  U+CE20  uniCE20,  HANGUL SYLLABLE-CE20
        0xCE21, # G+2959  U+CE21  uniCE21,  HANGUL SYLLABLE-CE21
        0xCE24, # G+2960  U+CE24  uniCE24,  HANGUL SYLLABLE-CE24
        0xCE28, # G+2961  U+CE28  uniCE28,  HANGUL SYLLABLE-CE28
        0xCE30, # G+2962  U+CE30  uniCE30,  HANGUL SYLLABLE-CE30
        0xCE31, # G+2963  U+CE31  uniCE31,  HANGUL SYLLABLE-CE31
        0xCE33, # G+2964  U+CE33  uniCE33,  HANGUL SYLLABLE-CE33
        0xCE35, # G+2965  U+CE35  uniCE35,  HANGUL SYLLABLE-CE35
        0xCE58, # G+2966  U+CE58  uniCE58,  HANGUL SYLLABLE-CE58
        0xCE59, # G+2967  U+CE59  uniCE59,  HANGUL SYLLABLE-CE59
        0xCE5C, # G+2968  U+CE5C  uniCE5C,  HANGUL SYLLABLE-CE5C
        0xCE5F, # G+2969  U+CE5F  uniCE5F,  HANGUL SYLLABLE-CE5F
        0xCE60, # G+2970  U+CE60  uniCE60,  HANGUL SYLLABLE-CE60
        0xCE61, # G+2971  U+CE61  uniCE61,  HANGUL SYLLABLE-CE61
        0xCE68, # G+2972  U+CE68  uniCE68,  HANGUL SYLLABLE-CE68
        0xCE69, # G+2973  U+CE69  uniCE69,  HANGUL SYLLABLE-CE69
        0xCE6B, # G+2974  U+CE6B  uniCE6B,  HANGUL SYLLABLE-CE6B
        0xCE6D, # G+2975  U+CE6D  uniCE6D,  HANGUL SYLLABLE-CE6D
        0xCE74, # G+2976  U+CE74  uniCE74,  HANGUL SYLLABLE-CE74
        0xCE75, # G+2977  U+CE75  uniCE75,  HANGUL SYLLABLE-CE75
        0xCE78, # G+2978  U+CE78  uniCE78,  HANGUL SYLLABLE-CE78
        0xCE7C, # G+2979  U+CE7C  uniCE7C,  HANGUL SYLLABLE-CE7C
        0xCE84, # G+2980  U+CE84  uniCE84,  HANGUL SYLLABLE-CE84
        0xCE85, # G+2981  U+CE85  uniCE85,  HANGUL SYLLABLE-CE85
        0xCE87, # G+2982  U+CE87  uniCE87,  HANGUL SYLLABLE-CE87
        0xCE89, # G+2983  U+CE89  uniCE89,  HANGUL SYLLABLE-CE89
        0xCE90, # G+2984  U+CE90  uniCE90,  HANGUL SYLLABLE-CE90
        0xCE91, # G+2985  U+CE91  uniCE91,  HANGUL SYLLABLE-CE91
        0xCE94, # G+2986  U+CE94  uniCE94,  HANGUL SYLLABLE-CE94
        0xCE98, # G+2987  U+CE98  uniCE98,  HANGUL SYLLABLE-CE98
        0xCEA0, # G+2988  U+CEA0  uniCEA0,  HANGUL SYLLABLE-CEA0
        0xCEA1, # G+2989  U+CEA1  uniCEA1,  HANGUL SYLLABLE-CEA1
        0xCEA3, # G+2990  U+CEA3  uniCEA3,  HANGUL SYLLABLE-CEA3
        0xCEA4, # G+2991  U+CEA4  uniCEA4,  HANGUL SYLLABLE-CEA4
        0xCEA5, # G+2992  U+CEA5  uniCEA5,  HANGUL SYLLABLE-CEA5
        0xCEAC, # G+2993  U+CEAC  uniCEAC,  HANGUL SYLLABLE-CEAC
        0xCEAD, # G+2994  U+CEAD  uniCEAD,  HANGUL SYLLABLE-CEAD
        0xCEC1, # G+2995  U+CEC1  uniCEC1,  HANGUL SYLLABLE-CEC1
        0xCEE4, # G+2996  U+CEE4  uniCEE4,  HANGUL SYLLABLE-CEE4
        0xCEE5, # G+2997  U+CEE5  uniCEE5,  HANGUL SYLLABLE-CEE5
        0xCEE8, # G+2998  U+CEE8  uniCEE8,  HANGUL SYLLABLE-CEE8
        0xCEEB, # G+2999  U+CEEB  uniCEEB,  HANGUL SYLLABLE-CEEB
        0xCEEC, # G+3000  U+CEEC  uniCEEC,  HANGUL SYLLABLE-CEEC
        0xCEF4, # G+3001  U+CEF4  uniCEF4,  HANGUL SYLLABLE-CEF4
        0xCEF5, # G+3002  U+CEF5  uniCEF5,  HANGUL SYLLABLE-CEF5
        0xCEF7, # G+3003  U+CEF7  uniCEF7,  HANGUL SYLLABLE-CEF7
        0xCEF8, # G+3004  U+CEF8  uniCEF8,  HANGUL SYLLABLE-CEF8
        0xCEF9, # G+3005  U+CEF9  uniCEF9,  HANGUL SYLLABLE-CEF9
        0xCF00, # G+3006  U+CF00  uniCF00,  HANGUL SYLLABLE-CF00
        0xCF01, # G+3007  U+CF01  uniCF01,  HANGUL SYLLABLE-CF01
        0xCF04, # G+3008  U+CF04  uniCF04,  HANGUL SYLLABLE-CF04
        0xCF08, # G+3009  U+CF08  uniCF08,  HANGUL SYLLABLE-CF08
        0xCF10, # G+3010  U+CF10  uniCF10,  HANGUL SYLLABLE-CF10
        0xCF11, # G+3011  U+CF11  uniCF11,  HANGUL SYLLABLE-CF11
        0xCF13, # G+3012  U+CF13  uniCF13,  HANGUL SYLLABLE-CF13
        0xCF15, # G+3013  U+CF15  uniCF15,  HANGUL SYLLABLE-CF15
        0xCF1C, # G+3014  U+CF1C  uniCF1C,  HANGUL SYLLABLE-CF1C
        0xCF20, # G+3015  U+CF20  uniCF20,  HANGUL SYLLABLE-CF20
        0xCF24, # G+3016  U+CF24  uniCF24,  HANGUL SYLLABLE-CF24
        0xCF2C, # G+3017  U+CF2C  uniCF2C,  HANGUL SYLLABLE-CF2C
        0xCF2D, # G+3018  U+CF2D  uniCF2D,  HANGUL SYLLABLE-CF2D
        0xCF2F, # G+3019  U+CF2F  uniCF2F,  HANGUL SYLLABLE-CF2F
        0xCF30, # G+3020  U+CF30  uniCF30,  HANGUL SYLLABLE-CF30
        0xCF31, # G+3021  U+CF31  uniCF31,  HANGUL SYLLABLE-CF31
        0xCF38, # G+3022  U+CF38  uniCF38,  HANGUL SYLLABLE-CF38
        0xCF54, # G+3023  U+CF54  uniCF54,  HANGUL SYLLABLE-CF54
        0xCF55, # G+3024  U+CF55  uniCF55,  HANGUL SYLLABLE-CF55
        0xCF58, # G+3025  U+CF58  uniCF58,  HANGUL SYLLABLE-CF58
        0xCF5C, # G+3026  U+CF5C  uniCF5C,  HANGUL SYLLABLE-CF5C
        0xCF64, # G+3027  U+CF64  uniCF64,  HANGUL SYLLABLE-CF64
        0xCF65, # G+3028  U+CF65  uniCF65,  HANGUL SYLLABLE-CF65
        0xCF67, # G+3029  U+CF67  uniCF67,  HANGUL SYLLABLE-CF67
        0xCF69, # G+3030  U+CF69  uniCF69,  HANGUL SYLLABLE-CF69
        0xCF70, # G+3031  U+CF70  uniCF70,  HANGUL SYLLABLE-CF70
        0xCF71, # G+3032  U+CF71  uniCF71,  HANGUL SYLLABLE-CF71
        0xCF74, # G+3033  U+CF74  uniCF74,  HANGUL SYLLABLE-CF74
        0xCF78, # G+3034  U+CF78  uniCF78,  HANGUL SYLLABLE-CF78
        0xCF80, # G+3035  U+CF80  uniCF80,  HANGUL SYLLABLE-CF80
        0xCF85, # G+3036  U+CF85  uniCF85,  HANGUL SYLLABLE-CF85
        0xCF8C, # G+3037  U+CF8C  uniCF8C,  HANGUL SYLLABLE-CF8C
        0xCFA1, # G+3038  U+CFA1  uniCFA1,  HANGUL SYLLABLE-CFA1
        0xCFA8, # G+3039  U+CFA8  uniCFA8,  HANGUL SYLLABLE-CFA8
        0xCFB0, # G+3040  U+CFB0  uniCFB0,  HANGUL SYLLABLE-CFB0
        0xCFC4, # G+3041  U+CFC4  uniCFC4,  HANGUL SYLLABLE-CFC4
        0xCFE0, # G+3042  U+CFE0  uniCFE0,  HANGUL SYLLABLE-CFE0
        0xCFE1, # G+3043  U+CFE1  uniCFE1,  HANGUL SYLLABLE-CFE1
        0xCFE4, # G+3044  U+CFE4  uniCFE4,  HANGUL SYLLABLE-CFE4
        0xCFE8, # G+3045  U+CFE8  uniCFE8,  HANGUL SYLLABLE-CFE8
        0xCFF0, # G+3046  U+CFF0  uniCFF0,  HANGUL SYLLABLE-CFF0
        0xCFF1, # G+3047  U+CFF1  uniCFF1,  HANGUL SYLLABLE-CFF1
        0xCFF3, # G+3048  U+CFF3  uniCFF3,  HANGUL SYLLABLE-CFF3
        0xCFF5, # G+3049  U+CFF5  uniCFF5,  HANGUL SYLLABLE-CFF5
        0xCFFC, # G+3050  U+CFFC  uniCFFC,  HANGUL SYLLABLE-CFFC
        0xD000, # G+3051  U+D000  uniD000,  HANGUL SYLLABLE-D000
        0xD004, # G+3052  U+D004  uniD004,  HANGUL SYLLABLE-D004
        0xD011, # G+3053  U+D011  uniD011,  HANGUL SYLLABLE-D011
        0xD018, # G+3054  U+D018  uniD018,  HANGUL SYLLABLE-D018
        0xD02D, # G+3055  U+D02D  uniD02D,  HANGUL SYLLABLE-D02D
        0xD034, # G+3056  U+D034  uniD034,  HANGUL SYLLABLE-D034
        0xD035, # G+3057  U+D035  uniD035,  HANGUL SYLLABLE-D035
        0xD038, # G+3058  U+D038  uniD038,  HANGUL SYLLABLE-D038
        0xD03C, # G+3059  U+D03C  uniD03C,  HANGUL SYLLABLE-D03C
        0xD044, # G+3060  U+D044  uniD044,  HANGUL SYLLABLE-D044
        0xD045, # G+3061  U+D045  uniD045,  HANGUL SYLLABLE-D045
        0xD047, # G+3062  U+D047  uniD047,  HANGUL SYLLABLE-D047
        0xD049, # G+3063  U+D049  uniD049,  HANGUL SYLLABLE-D049
        0xD050, # G+3064  U+D050  uniD050,  HANGUL SYLLABLE-D050
        0xD054, # G+3065  U+D054  uniD054,  HANGUL SYLLABLE-D054
        0xD058, # G+3066  U+D058  uniD058,  HANGUL SYLLABLE-D058
        0xD060, # G+3067  U+D060  uniD060,  HANGUL SYLLABLE-D060
        0xD06C, # G+3068  U+D06C  uniD06C,  HANGUL SYLLABLE-D06C
        0xD06D, # G+3069  U+D06D  uniD06D,  HANGUL SYLLABLE-D06D
        0xD070, # G+3070  U+D070  uniD070,  HANGUL SYLLABLE-D070
        0xD074, # G+3071  U+D074  uniD074,  HANGUL SYLLABLE-D074
        0xD07C, # G+3072  U+D07C  uniD07C,  HANGUL SYLLABLE-D07C
        0xD07D, # G+3073  U+D07D  uniD07D,  HANGUL SYLLABLE-D07D
        0xD081, # G+3074  U+D081  uniD081,  HANGUL SYLLABLE-D081
        0xD0A4, # G+3075  U+D0A4  uniD0A4,  HANGUL SYLLABLE-D0A4
        0xD0A5, # G+3076  U+D0A5  uniD0A5,  HANGUL SYLLABLE-D0A5
        0xD0A8, # G+3077  U+D0A8  uniD0A8,  HANGUL SYLLABLE-D0A8
        0xD0AC, # G+3078  U+D0AC  uniD0AC,  HANGUL SYLLABLE-D0AC
        0xD0B4, # G+3079  U+D0B4  uniD0B4,  HANGUL SYLLABLE-D0B4
        0xD0B5, # G+3080  U+D0B5  uniD0B5,  HANGUL SYLLABLE-D0B5
        0xD0B7, # G+3081  U+D0B7  uniD0B7,  HANGUL SYLLABLE-D0B7
        0xD0B9, # G+3082  U+D0B9  uniD0B9,  HANGUL SYLLABLE-D0B9
        0xD0C0, # G+3083  U+D0C0  uniD0C0,  HANGUL SYLLABLE-D0C0
        0xD0C1, # G+3084  U+D0C1  uniD0C1,  HANGUL SYLLABLE-D0C1
        0xD0C4, # G+3085  U+D0C4  uniD0C4,  HANGUL SYLLABLE-D0C4
        0xD0C8, # G+3086  U+D0C8  uniD0C8,  HANGUL SYLLABLE-D0C8
        0xD0C9, # G+3087  U+D0C9  uniD0C9,  HANGUL SYLLABLE-D0C9
        0xD0D0, # G+3088  U+D0D0  uniD0D0,  HANGUL SYLLABLE-D0D0
        0xD0D1, # G+3089  U+D0D1  uniD0D1,  HANGUL SYLLABLE-D0D1
        0xD0D3, # G+3090  U+D0D3  uniD0D3,  HANGUL SYLLABLE-D0D3
        0xD0D4, # G+3091  U+D0D4  uniD0D4,  HANGUL SYLLABLE-D0D4
        0xD0D5, # G+3092  U+D0D5  uniD0D5,  HANGUL SYLLABLE-D0D5
        0xD0DC, # G+3093  U+D0DC  uniD0DC,  HANGUL SYLLABLE-D0DC
        0xD0DD, # G+3094  U+D0DD  uniD0DD,  HANGUL SYLLABLE-D0DD
        0xD0E0, # G+3095  U+D0E0  uniD0E0,  HANGUL SYLLABLE-D0E0
        0xD0E4, # G+3096  U+D0E4  uniD0E4,  HANGUL SYLLABLE-D0E4
        0xD0EC, # G+3097  U+D0EC  uniD0EC,  HANGUL SYLLABLE-D0EC
        0xD0ED, # G+3098  U+D0ED  uniD0ED,  HANGUL SYLLABLE-D0ED
        0xD0EF, # G+3099  U+D0EF  uniD0EF,  HANGUL SYLLABLE-D0EF
        0xD0F0, # G+3100  U+D0F0  uniD0F0,  HANGUL SYLLABLE-D0F0
        0xD0F1, # G+3101  U+D0F1  uniD0F1,  HANGUL SYLLABLE-D0F1
        0xD0F8, # G+3102  U+D0F8  uniD0F8,  HANGUL SYLLABLE-D0F8
        0xD10D, # G+3103  U+D10D  uniD10D,  HANGUL SYLLABLE-D10D
        0xD130, # G+3104  U+D130  uniD130,  HANGUL SYLLABLE-D130
        0xD131, # G+3105  U+D131  uniD131,  HANGUL SYLLABLE-D131
        0xD134, # G+3106  U+D134  uniD134,  HANGUL SYLLABLE-D134
        0xD138, # G+3107  U+D138  uniD138,  HANGUL SYLLABLE-D138
        0xD13A, # G+3108  U+D13A  uniD13A,  HANGUL SYLLABLE-D13A
        0xD140, # G+3109  U+D140  uniD140,  HANGUL SYLLABLE-D140
        0xD141, # G+3110  U+D141  uniD141,  HANGUL SYLLABLE-D141
        0xD143, # G+3111  U+D143  uniD143,  HANGUL SYLLABLE-D143
        0xD144, # G+3112  U+D144  uniD144,  HANGUL SYLLABLE-D144
        0xD145, # G+3113  U+D145  uniD145,  HANGUL SYLLABLE-D145
        0xD14C, # G+3114  U+D14C  uniD14C,  HANGUL SYLLABLE-D14C
        0xD14D, # G+3115  U+D14D  uniD14D,  HANGUL SYLLABLE-D14D
        0xD150, # G+3116  U+D150  uniD150,  HANGUL SYLLABLE-D150
        0xD154, # G+3117  U+D154  uniD154,  HANGUL SYLLABLE-D154
        0xD15C, # G+3118  U+D15C  uniD15C,  HANGUL SYLLABLE-D15C
        0xD15D, # G+3119  U+D15D  uniD15D,  HANGUL SYLLABLE-D15D
        0xD15F, # G+3120  U+D15F  uniD15F,  HANGUL SYLLABLE-D15F
        0xD161, # G+3121  U+D161  uniD161,  HANGUL SYLLABLE-D161
        0xD168, # G+3122  U+D168  uniD168,  HANGUL SYLLABLE-D168
        0xD16C, # G+3123  U+D16C  uniD16C,  HANGUL SYLLABLE-D16C
        0xD17C, # G+3124  U+D17C  uniD17C,  HANGUL SYLLABLE-D17C
        0xD184, # G+3125  U+D184  uniD184,  HANGUL SYLLABLE-D184
        0xD188, # G+3126  U+D188  uniD188,  HANGUL SYLLABLE-D188
        0xD1A0, # G+3127  U+D1A0  uniD1A0,  HANGUL SYLLABLE-D1A0
        0xD1A1, # G+3128  U+D1A1  uniD1A1,  HANGUL SYLLABLE-D1A1
        0xD1A4, # G+3129  U+D1A4  uniD1A4,  HANGUL SYLLABLE-D1A4
        0xD1A8, # G+3130  U+D1A8  uniD1A8,  HANGUL SYLLABLE-D1A8
        0xD1B0, # G+3131  U+D1B0  uniD1B0,  HANGUL SYLLABLE-D1B0
        0xD1B1, # G+3132  U+D1B1  uniD1B1,  HANGUL SYLLABLE-D1B1
        0xD1B3, # G+3133  U+D1B3  uniD1B3,  HANGUL SYLLABLE-D1B3
        0xD1B5, # G+3134  U+D1B5  uniD1B5,  HANGUL SYLLABLE-D1B5
        0xD1BA, # G+3135  U+D1BA  uniD1BA,  HANGUL SYLLABLE-D1BA
        0xD1BC, # G+3136  U+D1BC  uniD1BC,  HANGUL SYLLABLE-D1BC
        0xD1C0, # G+3137  U+D1C0  uniD1C0,  HANGUL SYLLABLE-D1C0
        0xD1D8, # G+3138  U+D1D8  uniD1D8,  HANGUL SYLLABLE-D1D8
        0xD1F4, # G+3139  U+D1F4  uniD1F4,  HANGUL SYLLABLE-D1F4
        0xD1F8, # G+3140  U+D1F8  uniD1F8,  HANGUL SYLLABLE-D1F8
        0xD207, # G+3141  U+D207  uniD207,  HANGUL SYLLABLE-D207
        0xD209, # G+3142  U+D209  uniD209,  HANGUL SYLLABLE-D209
        0xD210, # G+3143  U+D210  uniD210,  HANGUL SYLLABLE-D210
        0xD22C, # G+3144  U+D22C  uniD22C,  HANGUL SYLLABLE-D22C
        0xD22D, # G+3145  U+D22D  uniD22D,  HANGUL SYLLABLE-D22D
        0xD230, # G+3146  U+D230  uniD230,  HANGUL SYLLABLE-D230
        0xD234, # G+3147  U+D234  uniD234,  HANGUL SYLLABLE-D234
        0xD23C, # G+3148  U+D23C  uniD23C,  HANGUL SYLLABLE-D23C
        0xD23D, # G+3149  U+D23D  uniD23D,  HANGUL SYLLABLE-D23D
        0xD23F, # G+3150  U+D23F  uniD23F,  HANGUL SYLLABLE-D23F
        0xD241, # G+3151  U+D241  uniD241,  HANGUL SYLLABLE-D241
        0xD248, # G+3152  U+D248  uniD248,  HANGUL SYLLABLE-D248
        0xD25C, # G+3153  U+D25C  uniD25C,  HANGUL SYLLABLE-D25C
        0xD264, # G+3154  U+D264  uniD264,  HANGUL SYLLABLE-D264
        0xD280, # G+3155  U+D280  uniD280,  HANGUL SYLLABLE-D280
        0xD281, # G+3156  U+D281  uniD281,  HANGUL SYLLABLE-D281
        0xD284, # G+3157  U+D284  uniD284,  HANGUL SYLLABLE-D284
        0xD288, # G+3158  U+D288  uniD288,  HANGUL SYLLABLE-D288
        0xD290, # G+3159  U+D290  uniD290,  HANGUL SYLLABLE-D290
        0xD291, # G+3160  U+D291  uniD291,  HANGUL SYLLABLE-D291
        0xD295, # G+3161  U+D295  uniD295,  HANGUL SYLLABLE-D295
        0xD29C, # G+3162  U+D29C  uniD29C,  HANGUL SYLLABLE-D29C
        0xD2A0, # G+3163  U+D2A0  uniD2A0,  HANGUL SYLLABLE-D2A0
        0xD2A4, # G+3164  U+D2A4  uniD2A4,  HANGUL SYLLABLE-D2A4
        0xD2AC, # G+3165  U+D2AC  uniD2AC,  HANGUL SYLLABLE-D2AC
        0xD2B1, # G+3166  U+D2B1  uniD2B1,  HANGUL SYLLABLE-D2B1
        0xD2B8, # G+3167  U+D2B8  uniD2B8,  HANGUL SYLLABLE-D2B8
        0xD2B9, # G+3168  U+D2B9  uniD2B9,  HANGUL SYLLABLE-D2B9
        0xD2BC, # G+3169  U+D2BC  uniD2BC,  HANGUL SYLLABLE-D2BC
        0xD2BF, # G+3170  U+D2BF  uniD2BF,  HANGUL SYLLABLE-D2BF
        0xD2C0, # G+3171  U+D2C0  uniD2C0,  HANGUL SYLLABLE-D2C0
        0xD2C2, # G+3172  U+D2C2  uniD2C2,  HANGUL SYLLABLE-D2C2
        0xD2C8, # G+3173  U+D2C8  uniD2C8,  HANGUL SYLLABLE-D2C8
        0xD2C9, # G+3174  U+D2C9  uniD2C9,  HANGUL SYLLABLE-D2C9
        0xD2CB, # G+3175  U+D2CB  uniD2CB,  HANGUL SYLLABLE-D2CB
        0xD2D4, # G+3176  U+D2D4  uniD2D4,  HANGUL SYLLABLE-D2D4
        0xD2D8, # G+3177  U+D2D8  uniD2D8,  HANGUL SYLLABLE-D2D8
        0xD2DC, # G+3178  U+D2DC  uniD2DC,  HANGUL SYLLABLE-D2DC
        0xD2E4, # G+3179  U+D2E4  uniD2E4,  HANGUL SYLLABLE-D2E4
        0xD2E5, # G+3180  U+D2E5  uniD2E5,  HANGUL SYLLABLE-D2E5
        0xD2F0, # G+3181  U+D2F0  uniD2F0,  HANGUL SYLLABLE-D2F0
        0xD2F1, # G+3182  U+D2F1  uniD2F1,  HANGUL SYLLABLE-D2F1
        0xD2F4, # G+3183  U+D2F4  uniD2F4,  HANGUL SYLLABLE-D2F4
        0xD2F8, # G+3184  U+D2F8  uniD2F8,  HANGUL SYLLABLE-D2F8
        0xD300, # G+3185  U+D300  uniD300,  HANGUL SYLLABLE-D300
        0xD301, # G+3186  U+D301  uniD301,  HANGUL SYLLABLE-D301
        0xD303, # G+3187  U+D303  uniD303,  HANGUL SYLLABLE-D303
        0xD305, # G+3188  U+D305  uniD305,  HANGUL SYLLABLE-D305
        0xD30C, # G+3189  U+D30C  uniD30C,  HANGUL SYLLABLE-D30C
        0xD30D, # G+3190  U+D30D  uniD30D,  HANGUL SYLLABLE-D30D
        0xD30E, # G+3191  U+D30E  uniD30E,  HANGUL SYLLABLE-D30E
        0xD310, # G+3192  U+D310  uniD310,  HANGUL SYLLABLE-D310
        0xD314, # G+3193  U+D314  uniD314,  HANGUL SYLLABLE-D314
        0xD316, # G+3194  U+D316  uniD316,  HANGUL SYLLABLE-D316
        0xD31C, # G+3195  U+D31C  uniD31C,  HANGUL SYLLABLE-D31C
        0xD31D, # G+3196  U+D31D  uniD31D,  HANGUL SYLLABLE-D31D
        0xD31F, # G+3197  U+D31F  uniD31F,  HANGUL SYLLABLE-D31F
        0xD320, # G+3198  U+D320  uniD320,  HANGUL SYLLABLE-D320
        0xD321, # G+3199  U+D321  uniD321,  HANGUL SYLLABLE-D321
        0xD325, # G+3200  U+D325  uniD325,  HANGUL SYLLABLE-D325
        0xD328, # G+3201  U+D328  uniD328,  HANGUL SYLLABLE-D328
        0xD329, # G+3202  U+D329  uniD329,  HANGUL SYLLABLE-D329
        0xD32C, # G+3203  U+D32C  uniD32C,  HANGUL SYLLABLE-D32C
        0xD330, # G+3204  U+D330  uniD330,  HANGUL SYLLABLE-D330
        0xD338, # G+3205  U+D338  uniD338,  HANGUL SYLLABLE-D338
        0xD339, # G+3206  U+D339  uniD339,  HANGUL SYLLABLE-D339
        0xD33B, # G+3207  U+D33B  uniD33B,  HANGUL SYLLABLE-D33B
        0xD33C, # G+3208  U+D33C  uniD33C,  HANGUL SYLLABLE-D33C
        0xD33D, # G+3209  U+D33D  uniD33D,  HANGUL SYLLABLE-D33D
        0xD344, # G+3210  U+D344  uniD344,  HANGUL SYLLABLE-D344
        0xD345, # G+3211  U+D345  uniD345,  HANGUL SYLLABLE-D345
        0xD37C, # G+3212  U+D37C  uniD37C,  HANGUL SYLLABLE-D37C
        0xD37D, # G+3213  U+D37D  uniD37D,  HANGUL SYLLABLE-D37D
        0xD380, # G+3214  U+D380  uniD380,  HANGUL SYLLABLE-D380
        0xD384, # G+3215  U+D384  uniD384,  HANGUL SYLLABLE-D384
        0xD38C, # G+3216  U+D38C  uniD38C,  HANGUL SYLLABLE-D38C
        0xD38D, # G+3217  U+D38D  uniD38D,  HANGUL SYLLABLE-D38D
        0xD38F, # G+3218  U+D38F  uniD38F,  HANGUL SYLLABLE-D38F
        0xD390, # G+3219  U+D390  uniD390,  HANGUL SYLLABLE-D390
        0xD391, # G+3220  U+D391  uniD391,  HANGUL SYLLABLE-D391
        0xD398, # G+3221  U+D398  uniD398,  HANGUL SYLLABLE-D398
        0xD399, # G+3222  U+D399  uniD399,  HANGUL SYLLABLE-D399
        0xD39C, # G+3223  U+D39C  uniD39C,  HANGUL SYLLABLE-D39C
        0xD3A0, # G+3224  U+D3A0  uniD3A0,  HANGUL SYLLABLE-D3A0
        0xD3A8, # G+3225  U+D3A8  uniD3A8,  HANGUL SYLLABLE-D3A8
        0xD3A9, # G+3226  U+D3A9  uniD3A9,  HANGUL SYLLABLE-D3A9
        0xD3AB, # G+3227  U+D3AB  uniD3AB,  HANGUL SYLLABLE-D3AB
        0xD3AD, # G+3228  U+D3AD  uniD3AD,  HANGUL SYLLABLE-D3AD
        0xD3B4, # G+3229  U+D3B4  uniD3B4,  HANGUL SYLLABLE-D3B4
        0xD3B8, # G+3230  U+D3B8  uniD3B8,  HANGUL SYLLABLE-D3B8
        0xD3BC, # G+3231  U+D3BC  uniD3BC,  HANGUL SYLLABLE-D3BC
        0xD3C4, # G+3232  U+D3C4  uniD3C4,  HANGUL SYLLABLE-D3C4
        0xD3C5, # G+3233  U+D3C5  uniD3C5,  HANGUL SYLLABLE-D3C5
        0xD3C8, # G+3234  U+D3C8  uniD3C8,  HANGUL SYLLABLE-D3C8
        0xD3C9, # G+3235  U+D3C9  uniD3C9,  HANGUL SYLLABLE-D3C9
        0xD3D0, # G+3236  U+D3D0  uniD3D0,  HANGUL SYLLABLE-D3D0
        0xD3D8, # G+3237  U+D3D8  uniD3D8,  HANGUL SYLLABLE-D3D8
        0xD3E1, # G+3238  U+D3E1  uniD3E1,  HANGUL SYLLABLE-D3E1
        0xD3E3, # G+3239  U+D3E3  uniD3E3,  HANGUL SYLLABLE-D3E3
        0xD3EC, # G+3240  U+D3EC  uniD3EC,  HANGUL SYLLABLE-D3EC
        0xD3ED, # G+3241  U+D3ED  uniD3ED,  HANGUL SYLLABLE-D3ED
        0xD3F0, # G+3242  U+D3F0  uniD3F0,  HANGUL SYLLABLE-D3F0
        0xD3F4, # G+3243  U+D3F4  uniD3F4,  HANGUL SYLLABLE-D3F4
        0xD3FC, # G+3244  U+D3FC  uniD3FC,  HANGUL SYLLABLE-D3FC
        0xD3FD, # G+3245  U+D3FD  uniD3FD,  HANGUL SYLLABLE-D3FD
        0xD3FF, # G+3246  U+D3FF  uniD3FF,  HANGUL SYLLABLE-D3FF
        0xD401, # G+3247  U+D401  uniD401,  HANGUL SYLLABLE-D401
        0xD408, # G+3248  U+D408  uniD408,  HANGUL SYLLABLE-D408
        0xD41D, # G+3249  U+D41D  uniD41D,  HANGUL SYLLABLE-D41D
        0xD440, # G+3250  U+D440  uniD440,  HANGUL SYLLABLE-D440
        0xD444, # G+3251  U+D444  uniD444,  HANGUL SYLLABLE-D444
        0xD45C, # G+3252  U+D45C  uniD45C,  HANGUL SYLLABLE-D45C
        0xD460, # G+3253  U+D460  uniD460,  HANGUL SYLLABLE-D460
        0xD464, # G+3254  U+D464  uniD464,  HANGUL SYLLABLE-D464
        0xD46D, # G+3255  U+D46D  uniD46D,  HANGUL SYLLABLE-D46D
        0xD46F, # G+3256  U+D46F  uniD46F,  HANGUL SYLLABLE-D46F
        0xD478, # G+3257  U+D478  uniD478,  HANGUL SYLLABLE-D478
        0xD479, # G+3258  U+D479  uniD479,  HANGUL SYLLABLE-D479
        0xD47C, # G+3259  U+D47C  uniD47C,  HANGUL SYLLABLE-D47C
        0xD47F, # G+3260  U+D47F  uniD47F,  HANGUL SYLLABLE-D47F
        0xD480, # G+3261  U+D480  uniD480,  HANGUL SYLLABLE-D480
        0xD482, # G+3262  U+D482  uniD482,  HANGUL SYLLABLE-D482
        0xD488, # G+3263  U+D488  uniD488,  HANGUL SYLLABLE-D488
        0xD489, # G+3264  U+D489  uniD489,  HANGUL SYLLABLE-D489
        0xD48B, # G+3265  U+D48B  uniD48B,  HANGUL SYLLABLE-D48B
        0xD48D, # G+3266  U+D48D  uniD48D,  HANGUL SYLLABLE-D48D
        0xD494, # G+3267  U+D494  uniD494,  HANGUL SYLLABLE-D494
        0xD4A9, # G+3268  U+D4A9  uniD4A9,  HANGUL SYLLABLE-D4A9
        0xD4CC, # G+3269  U+D4CC  uniD4CC,  HANGUL SYLLABLE-D4CC
        0xD4D0, # G+3270  U+D4D0  uniD4D0,  HANGUL SYLLABLE-D4D0
        0xD4D4, # G+3271  U+D4D4  uniD4D4,  HANGUL SYLLABLE-D4D4
        0xD4DC, # G+3272  U+D4DC  uniD4DC,  HANGUL SYLLABLE-D4DC
        0xD4DF, # G+3273  U+D4DF  uniD4DF,  HANGUL SYLLABLE-D4DF
        0xD4E8, # G+3274  U+D4E8  uniD4E8,  HANGUL SYLLABLE-D4E8
        0xD4EC, # G+3275  U+D4EC  uniD4EC,  HANGUL SYLLABLE-D4EC
        0xD4F0, # G+3276  U+D4F0  uniD4F0,  HANGUL SYLLABLE-D4F0
        0xD4F8, # G+3277  U+D4F8  uniD4F8,  HANGUL SYLLABLE-D4F8
        0xD4FB, # G+3278  U+D4FB  uniD4FB,  HANGUL SYLLABLE-D4FB
        0xD4FD, # G+3279  U+D4FD  uniD4FD,  HANGUL SYLLABLE-D4FD
        0xD504, # G+3280  U+D504  uniD504,  HANGUL SYLLABLE-D504
        0xD508, # G+3281  U+D508  uniD508,  HANGUL SYLLABLE-D508
        0xD50C, # G+3282  U+D50C  uniD50C,  HANGUL SYLLABLE-D50C
        0xD514, # G+3283  U+D514  uniD514,  HANGUL SYLLABLE-D514
        0xD515, # G+3284  U+D515  uniD515,  HANGUL SYLLABLE-D515
        0xD517, # G+3285  U+D517  uniD517,  HANGUL SYLLABLE-D517
        0xD53C, # G+3286  U+D53C  uniD53C,  HANGUL SYLLABLE-D53C
        0xD53D, # G+3287  U+D53D  uniD53D,  HANGUL SYLLABLE-D53D
        0xD540, # G+3288  U+D540  uniD540,  HANGUL SYLLABLE-D540
        0xD544, # G+3289  U+D544  uniD544,  HANGUL SYLLABLE-D544
        0xD54C, # G+3290  U+D54C  uniD54C,  HANGUL SYLLABLE-D54C
        0xD54D, # G+3291  U+D54D  uniD54D,  HANGUL SYLLABLE-D54D
        0xD54F, # G+3292  U+D54F  uniD54F,  HANGUL SYLLABLE-D54F
        0xD551, # G+3293  U+D551  uniD551,  HANGUL SYLLABLE-D551
        0xD558, # G+3294  U+D558  uniD558,  HANGUL SYLLABLE-D558
        0xD559, # G+3295  U+D559  uniD559,  HANGUL SYLLABLE-D559
        0xD55C, # G+3296  U+D55C  uniD55C,  HANGUL SYLLABLE-D55C
        0xD560, # G+3297  U+D560  uniD560,  HANGUL SYLLABLE-D560
        0xD565, # G+3298  U+D565  uniD565,  HANGUL SYLLABLE-D565
        0xD568, # G+3299  U+D568  uniD568,  HANGUL SYLLABLE-D568
        0xD569, # G+3300  U+D569  uniD569,  HANGUL SYLLABLE-D569
        0xD56B, # G+3301  U+D56B  uniD56B,  HANGUL SYLLABLE-D56B
        0xD56D, # G+3302  U+D56D  uniD56D,  HANGUL SYLLABLE-D56D
        0xD574, # G+3303  U+D574  uniD574,  HANGUL SYLLABLE-D574
        0xD575, # G+3304  U+D575  uniD575,  HANGUL SYLLABLE-D575
        0xD578, # G+3305  U+D578  uniD578,  HANGUL SYLLABLE-D578
        0xD57C, # G+3306  U+D57C  uniD57C,  HANGUL SYLLABLE-D57C
        0xD584, # G+3307  U+D584  uniD584,  HANGUL SYLLABLE-D584
        0xD585, # G+3308  U+D585  uniD585,  HANGUL SYLLABLE-D585
        0xD587, # G+3309  U+D587  uniD587,  HANGUL SYLLABLE-D587
        0xD588, # G+3310  U+D588  uniD588,  HANGUL SYLLABLE-D588
        0xD589, # G+3311  U+D589  uniD589,  HANGUL SYLLABLE-D589
        0xD590, # G+3312  U+D590  uniD590,  HANGUL SYLLABLE-D590
        0xD5A5, # G+3313  U+D5A5  uniD5A5,  HANGUL SYLLABLE-D5A5
        0xD5C8, # G+3314  U+D5C8  uniD5C8,  HANGUL SYLLABLE-D5C8
        0xD5C9, # G+3315  U+D5C9  uniD5C9,  HANGUL SYLLABLE-D5C9
        0xD5CC, # G+3316  U+D5CC  uniD5CC,  HANGUL SYLLABLE-D5CC
        0xD5D0, # G+3317  U+D5D0  uniD5D0,  HANGUL SYLLABLE-D5D0
        0xD5D2, # G+3318  U+D5D2  uniD5D2,  HANGUL SYLLABLE-D5D2
        0xD5D8, # G+3319  U+D5D8  uniD5D8,  HANGUL SYLLABLE-D5D8
        0xD5D9, # G+3320  U+D5D9  uniD5D9,  HANGUL SYLLABLE-D5D9
        0xD5DB, # G+3321  U+D5DB  uniD5DB,  HANGUL SYLLABLE-D5DB
        0xD5DD, # G+3322  U+D5DD  uniD5DD,  HANGUL SYLLABLE-D5DD
        0xD5E4, # G+3323  U+D5E4  uniD5E4,  HANGUL SYLLABLE-D5E4
        0xD5E5, # G+3324  U+D5E5  uniD5E5,  HANGUL SYLLABLE-D5E5
        0xD5E8, # G+3325  U+D5E8  uniD5E8,  HANGUL SYLLABLE-D5E8
        0xD5EC, # G+3326  U+D5EC  uniD5EC,  HANGUL SYLLABLE-D5EC
        0xD5F4, # G+3327  U+D5F4  uniD5F4,  HANGUL SYLLABLE-D5F4
        0xD5F5, # G+3328  U+D5F5  uniD5F5,  HANGUL SYLLABLE-D5F5
        0xD5F7, # G+3329  U+D5F7  uniD5F7,  HANGUL SYLLABLE-D5F7
        0xD5F9, # G+3330  U+D5F9  uniD5F9,  HANGUL SYLLABLE-D5F9
        0xD600, # G+3331  U+D600  uniD600,  HANGUL SYLLABLE-D600
        0xD601, # G+3332  U+D601  uniD601,  HANGUL SYLLABLE-D601
        0xD604, # G+3333  U+D604  uniD604,  HANGUL SYLLABLE-D604
        0xD608, # G+3334  U+D608  uniD608,  HANGUL SYLLABLE-D608
        0xD610, # G+3335  U+D610  uniD610,  HANGUL SYLLABLE-D610
        0xD611, # G+3336  U+D611  uniD611,  HANGUL SYLLABLE-D611
        0xD613, # G+3337  U+D613  uniD613,  HANGUL SYLLABLE-D613
        0xD614, # G+3338  U+D614  uniD614,  HANGUL SYLLABLE-D614
        0xD615, # G+3339  U+D615  uniD615,  HANGUL SYLLABLE-D615
        0xD61C, # G+3340  U+D61C  uniD61C,  HANGUL SYLLABLE-D61C
        0xD620, # G+3341  U+D620  uniD620,  HANGUL SYLLABLE-D620
        0xD624, # G+3342  U+D624  uniD624,  HANGUL SYLLABLE-D624
        0xD62D, # G+3343  U+D62D  uniD62D,  HANGUL SYLLABLE-D62D
        0xD638, # G+3344  U+D638  uniD638,  HANGUL SYLLABLE-D638
        0xD639, # G+3345  U+D639  uniD639,  HANGUL SYLLABLE-D639
        0xD63C, # G+3346  U+D63C  uniD63C,  HANGUL SYLLABLE-D63C
        0xD640, # G+3347  U+D640  uniD640,  HANGUL SYLLABLE-D640
        0xD645, # G+3348  U+D645  uniD645,  HANGUL SYLLABLE-D645
        0xD648, # G+3349  U+D648  uniD648,  HANGUL SYLLABLE-D648
        0xD649, # G+3350  U+D649  uniD649,  HANGUL SYLLABLE-D649
        0xD64B, # G+3351  U+D64B  uniD64B,  HANGUL SYLLABLE-D64B
        0xD64D, # G+3352  U+D64D  uniD64D,  HANGUL SYLLABLE-D64D
        0xD651, # G+3353  U+D651  uniD651,  HANGUL SYLLABLE-D651
        0xD654, # G+3354  U+D654  uniD654,  HANGUL SYLLABLE-D654
        0xD655, # G+3355  U+D655  uniD655,  HANGUL SYLLABLE-D655
        0xD658, # G+3356  U+D658  uniD658,  HANGUL SYLLABLE-D658
        0xD65C, # G+3357  U+D65C  uniD65C,  HANGUL SYLLABLE-D65C
        0xD667, # G+3358  U+D667  uniD667,  HANGUL SYLLABLE-D667
        0xD669, # G+3359  U+D669  uniD669,  HANGUL SYLLABLE-D669
        0xD670, # G+3360  U+D670  uniD670,  HANGUL SYLLABLE-D670
        0xD671, # G+3361  U+D671  uniD671,  HANGUL SYLLABLE-D671
        0xD674, # G+3362  U+D674  uniD674,  HANGUL SYLLABLE-D674
        0xD683, # G+3363  U+D683  uniD683,  HANGUL SYLLABLE-D683
        0xD685, # G+3364  U+D685  uniD685,  HANGUL SYLLABLE-D685
        0xD68C, # G+3365  U+D68C  uniD68C,  HANGUL SYLLABLE-D68C
        0xD68D, # G+3366  U+D68D  uniD68D,  HANGUL SYLLABLE-D68D
        0xD690, # G+3367  U+D690  uniD690,  HANGUL SYLLABLE-D690
        0xD694, # G+3368  U+D694  uniD694,  HANGUL SYLLABLE-D694
        0xD69D, # G+3369  U+D69D  uniD69D,  HANGUL SYLLABLE-D69D
        0xD69F, # G+3370  U+D69F  uniD69F,  HANGUL SYLLABLE-D69F
        0xD6A1, # G+3371  U+D6A1  uniD6A1,  HANGUL SYLLABLE-D6A1
        0xD6A8, # G+3372  U+D6A8  uniD6A8,  HANGUL SYLLABLE-D6A8
        0xD6AC, # G+3373  U+D6AC  uniD6AC,  HANGUL SYLLABLE-D6AC
        0xD6B0, # G+3374  U+D6B0  uniD6B0,  HANGUL SYLLABLE-D6B0
        0xD6B9, # G+3375  U+D6B9  uniD6B9,  HANGUL SYLLABLE-D6B9
        0xD6BB, # G+3376  U+D6BB  uniD6BB,  HANGUL SYLLABLE-D6BB
        0xD6C4, # G+3377  U+D6C4  uniD6C4,  HANGUL SYLLABLE-D6C4
        0xD6C5, # G+3378  U+D6C5  uniD6C5,  HANGUL SYLLABLE-D6C5
        0xD6C8, # G+3379  U+D6C8  uniD6C8,  HANGUL SYLLABLE-D6C8
        0xD6CC, # G+3380  U+D6CC  uniD6CC,  HANGUL SYLLABLE-D6CC
        0xD6D1, # G+3381  U+D6D1  uniD6D1,  HANGUL SYLLABLE-D6D1
        0xD6D4, # G+3382  U+D6D4  uniD6D4,  HANGUL SYLLABLE-D6D4
        0xD6D7, # G+3383  U+D6D7  uniD6D7,  HANGUL SYLLABLE-D6D7
        0xD6D9, # G+3384  U+D6D9  uniD6D9,  HANGUL SYLLABLE-D6D9
        0xD6E0, # G+3385  U+D6E0  uniD6E0,  HANGUL SYLLABLE-D6E0
        0xD6E4, # G+3386  U+D6E4  uniD6E4,  HANGUL SYLLABLE-D6E4
        0xD6E8, # G+3387  U+D6E8  uniD6E8,  HANGUL SYLLABLE-D6E8
        0xD6F0, # G+3388  U+D6F0  uniD6F0,  HANGUL SYLLABLE-D6F0
        0xD6F5, # G+3389  U+D6F5  uniD6F5,  HANGUL SYLLABLE-D6F5
        0xD6FC, # G+3390  U+D6FC  uniD6FC,  HANGUL SYLLABLE-D6FC
        0xD6FD, # G+3391  U+D6FD  uniD6FD,  HANGUL SYLLABLE-D6FD
        0xD700, # G+3392  U+D700  uniD700,  HANGUL SYLLABLE-D700
        0xD704, # G+3393  U+D704  uniD704,  HANGUL SYLLABLE-D704
        0xD711, # G+3394  U+D711  uniD711,  HANGUL SYLLABLE-D711
        0xD718, # G+3395  U+D718  uniD718,  HANGUL SYLLABLE-D718
        0xD719, # G+3396  U+D719  uniD719,  HANGUL SYLLABLE-D719
        0xD71C, # G+3397  U+D71C  uniD71C,  HANGUL SYLLABLE-D71C
        0xD720, # G+3398  U+D720  uniD720,  HANGUL SYLLABLE-D720
        0xD728, # G+3399  U+D728  uniD728,  HANGUL SYLLABLE-D728
        0xD729, # G+3400  U+D729  uniD729,  HANGUL SYLLABLE-D729
        0xD72B, # G+3401  U+D72B  uniD72B,  HANGUL SYLLABLE-D72B
        0xD72D, # G+3402  U+D72D  uniD72D,  HANGUL SYLLABLE-D72D
        0xD734, # G+3403  U+D734  uniD734,  HANGUL SYLLABLE-D734
        0xD735, # G+3404  U+D735  uniD735,  HANGUL SYLLABLE-D735
        0xD738, # G+3405  U+D738  uniD738,  HANGUL SYLLABLE-D738
        0xD73C, # G+3406  U+D73C  uniD73C,  HANGUL SYLLABLE-D73C
        0xD744, # G+3407  U+D744  uniD744,  HANGUL SYLLABLE-D744
        0xD747, # G+3408  U+D747  uniD747,  HANGUL SYLLABLE-D747
        0xD749, # G+3409  U+D749  uniD749,  HANGUL SYLLABLE-D749
        0xD750, # G+3410  U+D750  uniD750,  HANGUL SYLLABLE-D750
        0xD751, # G+3411  U+D751  uniD751,  HANGUL SYLLABLE-D751
        0xD754, # G+3412  U+D754  uniD754,  HANGUL SYLLABLE-D754
        0xD756, # G+3413  U+D756  uniD756,  HANGUL SYLLABLE-D756
        0xD757, # G+3414  U+D757  uniD757,  HANGUL SYLLABLE-D757
        0xD758, # G+3415  U+D758  uniD758,  HANGUL SYLLABLE-D758
        0xD759, # G+3416  U+D759  uniD759,  HANGUL SYLLABLE-D759
        0xD760, # G+3417  U+D760  uniD760,  HANGUL SYLLABLE-D760
        0xD761, # G+3418  U+D761  uniD761,  HANGUL SYLLABLE-D761
        0xD763, # G+3419  U+D763  uniD763,  HANGUL SYLLABLE-D763
        0xD765, # G+3420  U+D765  uniD765,  HANGUL SYLLABLE-D765
        0xD769, # G+3421  U+D769  uniD769,  HANGUL SYLLABLE-D769
        0xD76C, # G+3422  U+D76C  uniD76C,  HANGUL SYLLABLE-D76C
        0xD770, # G+3423  U+D770  uniD770,  HANGUL SYLLABLE-D770
        0xD774, # G+3424  U+D774  uniD774,  HANGUL SYLLABLE-D774
        0xD77C, # G+3425  U+D77C  uniD77C,  HANGUL SYLLABLE-D77C
        0xD77D, # G+3426  U+D77D  uniD77D,  HANGUL SYLLABLE-D77D
        0xD781, # G+3427  U+D781  uniD781,  HANGUL SYLLABLE-D781
        0xD788, # G+3428  U+D788  uniD788,  HANGUL SYLLABLE-D788
        0xD789, # G+3429  U+D789  uniD789,  HANGUL SYLLABLE-D789
        0xD78C, # G+3430  U+D78C  uniD78C,  HANGUL SYLLABLE-D78C
        0xD790, # G+3431  U+D790  uniD790,  HANGUL SYLLABLE-D790
        0xD798, # G+3432  U+D798  uniD798,  HANGUL SYLLABLE-D798
        0xD799, # G+3433  U+D799  uniD799,  HANGUL SYLLABLE-D799
        0xD79B, # G+3434  U+D79B  uniD79B,  HANGUL SYLLABLE-D79B
        0xD79D, # G+3435  U+D79D  uniD79D,  HANGUL SYLLABLE-D79D
        0x4F3D, # G+3436  U+4F3D  uni4F3D,  CJK UNIFIED IDEOGRAPH-4F3D
        0x4F73, # G+3437  U+4F73  uni4F73,  CJK UNIFIED IDEOGRAPH-4F73
        0x5047, # G+3438  U+5047  uni5047,  CJK UNIFIED IDEOGRAPH-5047
        0x50F9, # G+3439  U+50F9  uni50F9,  CJK UNIFIED IDEOGRAPH-50F9
        0x52A0, # G+3440  U+52A0  uni52A0,  CJK UNIFIED IDEOGRAPH-52A0
        0x53EF, # G+3441  U+53EF  uni53EF,  CJK UNIFIED IDEOGRAPH-53EF
        0x5475, # G+3442  U+5475  uni5475,  CJK UNIFIED IDEOGRAPH-5475
        0x54E5, # G+3443  U+54E5  uni54E5,  CJK UNIFIED IDEOGRAPH-54E5
        0x5609, # G+3444  U+5609  uni5609,  CJK UNIFIED IDEOGRAPH-5609
        0x5AC1, # G+3445  U+5AC1  uni5AC1,  CJK UNIFIED IDEOGRAPH-5AC1
        0x5BB6, # G+3446  U+5BB6  uni5BB6,  CJK UNIFIED IDEOGRAPH-5BB6
        0x6687, # G+3447  U+6687  uni6687,  CJK UNIFIED IDEOGRAPH-6687
        0x67B6, # G+3448  U+67B6  uni67B6,  CJK UNIFIED IDEOGRAPH-67B6
        0x67B7, # G+3449  U+67B7  uni67B7,  CJK UNIFIED IDEOGRAPH-67B7
        0x67EF, # G+3450  U+67EF  uni67EF,  CJK UNIFIED IDEOGRAPH-67EF
        0x6B4C, # G+3451  U+6B4C  uni6B4C,  CJK UNIFIED IDEOGRAPH-6B4C
        0x73C2, # G+3452  U+73C2  uni73C2,  CJK UNIFIED IDEOGRAPH-73C2
        0x75C2, # G+3453  U+75C2  uni75C2,  CJK UNIFIED IDEOGRAPH-75C2
        0x7A3C, # G+3454  U+7A3C  uni7A3C,  CJK UNIFIED IDEOGRAPH-7A3C
        0x82DB, # G+3455  U+82DB  uni82DB,  CJK UNIFIED IDEOGRAPH-82DB
        0x8304, # G+3456  U+8304  uni8304,  CJK UNIFIED IDEOGRAPH-8304
        0x8857, # G+3457  U+8857  uni8857,  CJK UNIFIED IDEOGRAPH-8857
        0x8888, # G+3458  U+8888  uni8888,  CJK UNIFIED IDEOGRAPH-8888
        0x8A36, # G+3459  U+8A36  uni8A36,  CJK UNIFIED IDEOGRAPH-8A36
        0x8CC8, # G+3460  U+8CC8  uni8CC8,  CJK UNIFIED IDEOGRAPH-8CC8
        0x8DCF, # G+3461  U+8DCF  uni8DCF,  CJK UNIFIED IDEOGRAPH-8DCF
        0x8EFB, # G+3462  U+8EFB  uni8EFB,  CJK UNIFIED IDEOGRAPH-8EFB
        0x8FE6, # G+3463  U+8FE6  uni8FE6,  CJK UNIFIED IDEOGRAPH-8FE6
        0x99D5, # G+3464  U+99D5  uni99D5,  CJK UNIFIED IDEOGRAPH-99D5
        0x523B, # G+3465  U+523B  uni523B,  CJK UNIFIED IDEOGRAPH-523B
        0x5374, # G+3466  U+5374  uni5374,  CJK UNIFIED IDEOGRAPH-5374
        0x5404, # G+3467  U+5404  uni5404,  CJK UNIFIED IDEOGRAPH-5404
        0x606A, # G+3468  U+606A  uni606A,  CJK UNIFIED IDEOGRAPH-606A
        0x6164, # G+3469  U+6164  uni6164,  CJK UNIFIED IDEOGRAPH-6164
        0x6BBC, # G+3470  U+6BBC  uni6BBC,  CJK UNIFIED IDEOGRAPH-6BBC
        0x73CF, # G+3471  U+73CF  uni73CF,  CJK UNIFIED IDEOGRAPH-73CF
        0x811A, # G+3472  U+811A  uni811A,  CJK UNIFIED IDEOGRAPH-811A
        0x89BA, # G+3473  U+89BA  uni89BA,  CJK UNIFIED IDEOGRAPH-89BA
        0x89D2, # G+3474  U+89D2  uni89D2,  CJK UNIFIED IDEOGRAPH-89D2
        0x95A3, # G+3475  U+95A3  uni95A3,  CJK UNIFIED IDEOGRAPH-95A3
        0x4F83, # G+3476  U+4F83  uni4F83,  CJK UNIFIED IDEOGRAPH-4F83
        0x520A, # G+3477  U+520A  uni520A,  CJK UNIFIED IDEOGRAPH-520A
        0x58BE, # G+3478  U+58BE  uni58BE,  CJK UNIFIED IDEOGRAPH-58BE
        0x5978, # G+3479  U+5978  uni5978,  CJK UNIFIED IDEOGRAPH-5978
        0x59E6, # G+3480  U+59E6  uni59E6,  CJK UNIFIED IDEOGRAPH-59E6
        0x5E72, # G+3481  U+5E72  uni5E72,  CJK UNIFIED IDEOGRAPH-5E72
        0x5E79, # G+3482  U+5E79  uni5E79,  CJK UNIFIED IDEOGRAPH-5E79
        0x61C7, # G+3483  U+61C7  uni61C7,  CJK UNIFIED IDEOGRAPH-61C7
        0x63C0, # G+3484  U+63C0  uni63C0,  CJK UNIFIED IDEOGRAPH-63C0
        0x6746, # G+3485  U+6746  uni6746,  CJK UNIFIED IDEOGRAPH-6746
        0x67EC, # G+3486  U+67EC  uni67EC,  CJK UNIFIED IDEOGRAPH-67EC
        0x687F, # G+3487  U+687F  uni687F,  CJK UNIFIED IDEOGRAPH-687F
        0x6F97, # G+3488  U+6F97  uni6F97,  CJK UNIFIED IDEOGRAPH-6F97
        0x764E, # G+3489  U+764E  uni764E,  CJK UNIFIED IDEOGRAPH-764E
        0x770B, # G+3490  U+770B  uni770B,  CJK UNIFIED IDEOGRAPH-770B
        0x78F5, # G+3491  U+78F5  uni78F5,  CJK UNIFIED IDEOGRAPH-78F5
        0x7A08, # G+3492  U+7A08  uni7A08,  CJK UNIFIED IDEOGRAPH-7A08
        0x7AFF, # G+3493  U+7AFF  uni7AFF,  CJK UNIFIED IDEOGRAPH-7AFF
        0x7C21, # G+3494  U+7C21  uni7C21,  CJK UNIFIED IDEOGRAPH-7C21
        0x809D, # G+3495  U+809D  uni809D,  CJK UNIFIED IDEOGRAPH-809D
        0x826E, # G+3496  U+826E  uni826E,  CJK UNIFIED IDEOGRAPH-826E
        0x8271, # G+3497  U+8271  uni8271,  CJK UNIFIED IDEOGRAPH-8271
        0x8AEB, # G+3498  U+8AEB  uni8AEB,  CJK UNIFIED IDEOGRAPH-8AEB
        0x9593, # G+3499  U+9593  uni9593,  CJK UNIFIED IDEOGRAPH-9593
        0x4E6B, # G+3500  U+4E6B  uni4E6B,  CJK UNIFIED IDEOGRAPH-4E6B
        0x559D, # G+3501  U+559D  uni559D,  CJK UNIFIED IDEOGRAPH-559D
        0x66F7, # G+3502  U+66F7  uni66F7,  CJK UNIFIED IDEOGRAPH-66F7
        0x6E34, # G+3503  U+6E34  uni6E34,  CJK UNIFIED IDEOGRAPH-6E34
        0x78A3, # G+3504  U+78A3  uni78A3,  CJK UNIFIED IDEOGRAPH-78A3
        0x7AED, # G+3505  U+7AED  uni7AED,  CJK UNIFIED IDEOGRAPH-7AED
        0x845B, # G+3506  U+845B  uni845B,  CJK UNIFIED IDEOGRAPH-845B
        0x8910, # G+3507  U+8910  uni8910,  CJK UNIFIED IDEOGRAPH-8910
        0x874E, # G+3508  U+874E  uni874E,  CJK UNIFIED IDEOGRAPH-874E
        0x97A8, # G+3509  U+97A8  uni97A8,  CJK UNIFIED IDEOGRAPH-97A8
        0x52D8, # G+3510  U+52D8  uni52D8,  CJK UNIFIED IDEOGRAPH-52D8
        0x574E, # G+3511  U+574E  uni574E,  CJK UNIFIED IDEOGRAPH-574E
        0x582A, # G+3512  U+582A  uni582A,  CJK UNIFIED IDEOGRAPH-582A
        0x5D4C, # G+3513  U+5D4C  uni5D4C,  CJK UNIFIED IDEOGRAPH-5D4C
        0x611F, # G+3514  U+611F  uni611F,  CJK UNIFIED IDEOGRAPH-611F
        0x61BE, # G+3515  U+61BE  uni61BE,  CJK UNIFIED IDEOGRAPH-61BE
        0x6221, # G+3516  U+6221  uni6221,  CJK UNIFIED IDEOGRAPH-6221
        0x6562, # G+3517  U+6562  uni6562,  CJK UNIFIED IDEOGRAPH-6562
        0x67D1, # G+3518  U+67D1  uni67D1,  CJK UNIFIED IDEOGRAPH-67D1
        0x6A44, # G+3519  U+6A44  uni6A44,  CJK UNIFIED IDEOGRAPH-6A44
        0x6E1B, # G+3520  U+6E1B  uni6E1B,  CJK UNIFIED IDEOGRAPH-6E1B
        0x7518, # G+3521  U+7518  uni7518,  CJK UNIFIED IDEOGRAPH-7518
        0x75B3, # G+3522  U+75B3  uni75B3,  CJK UNIFIED IDEOGRAPH-75B3
        0x76E3, # G+3523  U+76E3  uni76E3,  CJK UNIFIED IDEOGRAPH-76E3
        0x77B0, # G+3524  U+77B0  uni77B0,  CJK UNIFIED IDEOGRAPH-77B0
        0x7D3A, # G+3525  U+7D3A  uni7D3A,  CJK UNIFIED IDEOGRAPH-7D3A
        0x90AF, # G+3526  U+90AF  uni90AF,  CJK UNIFIED IDEOGRAPH-90AF
        0x9451, # G+3527  U+9451  uni9451,  CJK UNIFIED IDEOGRAPH-9451
        0x9452, # G+3528  U+9452  uni9452,  CJK UNIFIED IDEOGRAPH-9452
        0x9F95, # G+3529  U+9F95  uni9F95,  CJK UNIFIED IDEOGRAPH-9F95
        0x5323, # G+3530  U+5323  uni5323,  CJK UNIFIED IDEOGRAPH-5323
        0x5CAC, # G+3531  U+5CAC  uni5CAC,  CJK UNIFIED IDEOGRAPH-5CAC
        0x7532, # G+3532  U+7532  uni7532,  CJK UNIFIED IDEOGRAPH-7532
        0x80DB, # G+3533  U+80DB  uni80DB,  CJK UNIFIED IDEOGRAPH-80DB
        0x9240, # G+3534  U+9240  uni9240,  CJK UNIFIED IDEOGRAPH-9240
        0x9598, # G+3535  U+9598  uni9598,  CJK UNIFIED IDEOGRAPH-9598
        0x525B, # G+3536  U+525B  uni525B,  CJK UNIFIED IDEOGRAPH-525B
        0x5808, # G+3537  U+5808  uni5808,  CJK UNIFIED IDEOGRAPH-5808
        0x59DC, # G+3538  U+59DC  uni59DC,  CJK UNIFIED IDEOGRAPH-59DC
        0x5CA1, # G+3539  U+5CA1  uni5CA1,  CJK UNIFIED IDEOGRAPH-5CA1
        0x5D17, # G+3540  U+5D17  uni5D17,  CJK UNIFIED IDEOGRAPH-5D17
        0x5EB7, # G+3541  U+5EB7  uni5EB7,  CJK UNIFIED IDEOGRAPH-5EB7
        0x5F3A, # G+3542  U+5F3A  uni5F3A,  CJK UNIFIED IDEOGRAPH-5F3A
        0x5F4A, # G+3543  U+5F4A  uni5F4A,  CJK UNIFIED IDEOGRAPH-5F4A
        0x6177, # G+3544  U+6177  uni6177,  CJK UNIFIED IDEOGRAPH-6177
        0x6C5F, # G+3545  U+6C5F  uni6C5F,  CJK UNIFIED IDEOGRAPH-6C5F
        0x757A, # G+3546  U+757A  uni757A,  CJK UNIFIED IDEOGRAPH-757A
        0x7586, # G+3547  U+7586  uni7586,  CJK UNIFIED IDEOGRAPH-7586
        0x7CE0, # G+3548  U+7CE0  uni7CE0,  CJK UNIFIED IDEOGRAPH-7CE0
        0x7D73, # G+3549  U+7D73  uni7D73,  CJK UNIFIED IDEOGRAPH-7D73
        0x7DB1, # G+3550  U+7DB1  uni7DB1,  CJK UNIFIED IDEOGRAPH-7DB1
        0x7F8C, # G+3551  U+7F8C  uni7F8C,  CJK UNIFIED IDEOGRAPH-7F8C
        0x8154, # G+3552  U+8154  uni8154,  CJK UNIFIED IDEOGRAPH-8154
        0x8221, # G+3553  U+8221  uni8221,  CJK UNIFIED IDEOGRAPH-8221
        0x8591, # G+3554  U+8591  uni8591,  CJK UNIFIED IDEOGRAPH-8591
        0x8941, # G+3555  U+8941  uni8941,  CJK UNIFIED IDEOGRAPH-8941
        0x8B1B, # G+3556  U+8B1B  uni8B1B,  CJK UNIFIED IDEOGRAPH-8B1B
        0x92FC, # G+3557  U+92FC  uni92FC,  CJK UNIFIED IDEOGRAPH-92FC
        0x964D, # G+3558  U+964D  uni964D,  CJK UNIFIED IDEOGRAPH-964D
        0x9C47, # G+3559  U+9C47  uni9C47,  CJK UNIFIED IDEOGRAPH-9C47
        0x4ECB, # G+3560  U+4ECB  uni4ECB,  CJK UNIFIED IDEOGRAPH-4ECB
        0x4EF7, # G+3561  U+4EF7  uni4EF7,  CJK UNIFIED IDEOGRAPH-4EF7
        0x500B, # G+3562  U+500B  uni500B,  CJK UNIFIED IDEOGRAPH-500B
        0x51F1, # G+3563  U+51F1  uni51F1,  CJK UNIFIED IDEOGRAPH-51F1
        0x584F, # G+3564  U+584F  uni584F,  CJK UNIFIED IDEOGRAPH-584F
        0x6137, # G+3565  U+6137  uni6137,  CJK UNIFIED IDEOGRAPH-6137
        0x613E, # G+3566  U+613E  uni613E,  CJK UNIFIED IDEOGRAPH-613E
        0x6168, # G+3567  U+6168  uni6168,  CJK UNIFIED IDEOGRAPH-6168
        0x6539, # G+3568  U+6539  uni6539,  CJK UNIFIED IDEOGRAPH-6539
        0x69EA, # G+3569  U+69EA  uni69EA,  CJK UNIFIED IDEOGRAPH-69EA
        0x6F11, # G+3570  U+6F11  uni6F11,  CJK UNIFIED IDEOGRAPH-6F11
        0x75A5, # G+3571  U+75A5  uni75A5,  CJK UNIFIED IDEOGRAPH-75A5
        0x7686, # G+3572  U+7686  uni7686,  CJK UNIFIED IDEOGRAPH-7686
        0x76D6, # G+3573  U+76D6  uni76D6,  CJK UNIFIED IDEOGRAPH-76D6
        0x7B87, # G+3574  U+7B87  uni7B87,  CJK UNIFIED IDEOGRAPH-7B87
        0x82A5, # G+3575  U+82A5  uni82A5,  CJK UNIFIED IDEOGRAPH-82A5
        0x84CB, # G+3576  U+84CB  uni84CB,  CJK UNIFIED IDEOGRAPH-84CB
        0x93A7, # G+3577  U+93A7  uni93A7,  CJK UNIFIED IDEOGRAPH-93A7
        0x958B, # G+3578  U+958B  uni958B,  CJK UNIFIED IDEOGRAPH-958B
        0x5580, # G+3579  U+5580  uni5580,  CJK UNIFIED IDEOGRAPH-5580
        0x5BA2, # G+3580  U+5BA2  uni5BA2,  CJK UNIFIED IDEOGRAPH-5BA2
        0x5751, # G+3581  U+5751  uni5751,  CJK UNIFIED IDEOGRAPH-5751
        0x7CB3, # G+3582  U+7CB3  uni7CB3,  CJK UNIFIED IDEOGRAPH-7CB3
        0x7FB9, # G+3583  U+7FB9  uni7FB9,  CJK UNIFIED IDEOGRAPH-7FB9
        0x91B5, # G+3584  U+91B5  uni91B5,  CJK UNIFIED IDEOGRAPH-91B5
        0x5028, # G+3585  U+5028  uni5028,  CJK UNIFIED IDEOGRAPH-5028
        0x53BB, # G+3586  U+53BB  uni53BB,  CJK UNIFIED IDEOGRAPH-53BB
        0x5C45, # G+3587  U+5C45  uni5C45,  CJK UNIFIED IDEOGRAPH-5C45
        0x5DE8, # G+3588  U+5DE8  uni5DE8,  CJK UNIFIED IDEOGRAPH-5DE8
        0x62D2, # G+3589  U+62D2  uni62D2,  CJK UNIFIED IDEOGRAPH-62D2
        0x636E, # G+3590  U+636E  uni636E,  CJK UNIFIED IDEOGRAPH-636E
        0x64DA, # G+3591  U+64DA  uni64DA,  CJK UNIFIED IDEOGRAPH-64DA
        0x64E7, # G+3592  U+64E7  uni64E7,  CJK UNIFIED IDEOGRAPH-64E7
        0x6E20, # G+3593  U+6E20  uni6E20,  CJK UNIFIED IDEOGRAPH-6E20
        0x70AC, # G+3594  U+70AC  uni70AC,  CJK UNIFIED IDEOGRAPH-70AC
        0x795B, # G+3595  U+795B  uni795B,  CJK UNIFIED IDEOGRAPH-795B
        0x8DDD, # G+3596  U+8DDD  uni8DDD,  CJK UNIFIED IDEOGRAPH-8DDD
        0x8E1E, # G+3597  U+8E1E  uni8E1E,  CJK UNIFIED IDEOGRAPH-8E1E
        0x907D, # G+3598  U+907D  uni907D,  CJK UNIFIED IDEOGRAPH-907D
        0x9245, # G+3599  U+9245  uni9245,  CJK UNIFIED IDEOGRAPH-9245
        0x92F8, # G+3600  U+92F8  uni92F8,  CJK UNIFIED IDEOGRAPH-92F8
        0x4E7E, # G+3601  U+4E7E  uni4E7E,  CJK UNIFIED IDEOGRAPH-4E7E
        0x4EF6, # G+3602  U+4EF6  uni4EF6,  CJK UNIFIED IDEOGRAPH-4EF6
        0x5065, # G+3603  U+5065  uni5065,  CJK UNIFIED IDEOGRAPH-5065
        0x5DFE, # G+3604  U+5DFE  uni5DFE,  CJK UNIFIED IDEOGRAPH-5DFE
        0x5EFA, # G+3605  U+5EFA  uni5EFA,  CJK UNIFIED IDEOGRAPH-5EFA
        0x6106, # G+3606  U+6106  uni6106,  CJK UNIFIED IDEOGRAPH-6106
        0x6957, # G+3607  U+6957  uni6957,  CJK UNIFIED IDEOGRAPH-6957
        0x8171, # G+3608  U+8171  uni8171,  CJK UNIFIED IDEOGRAPH-8171
        0x8654, # G+3609  U+8654  uni8654,  CJK UNIFIED IDEOGRAPH-8654
        0x8E47, # G+3610  U+8E47  uni8E47,  CJK UNIFIED IDEOGRAPH-8E47
        0x9375, # G+3611  U+9375  uni9375,  CJK UNIFIED IDEOGRAPH-9375
        0x9A2B, # G+3612  U+9A2B  uni9A2B,  CJK UNIFIED IDEOGRAPH-9A2B
        0x4E5E, # G+3613  U+4E5E  uni4E5E,  CJK UNIFIED IDEOGRAPH-4E5E
        0x5091, # G+3614  U+5091  uni5091,  CJK UNIFIED IDEOGRAPH-5091
        0x6770, # G+3615  U+6770  uni6770,  CJK UNIFIED IDEOGRAPH-6770
        0x6840, # G+3616  U+6840  uni6840,  CJK UNIFIED IDEOGRAPH-6840
        0x5109, # G+3617  U+5109  uni5109,  CJK UNIFIED IDEOGRAPH-5109
        0x528D, # G+3618  U+528D  uni528D,  CJK UNIFIED IDEOGRAPH-528D
        0x5292, # G+3619  U+5292  uni5292,  CJK UNIFIED IDEOGRAPH-5292
        0x6AA2, # G+3620  U+6AA2  uni6AA2,  CJK UNIFIED IDEOGRAPH-6AA2
        0x77BC, # G+3621  U+77BC  uni77BC,  CJK UNIFIED IDEOGRAPH-77BC
        0x9210, # G+3622  U+9210  uni9210,  CJK UNIFIED IDEOGRAPH-9210
        0x9ED4, # G+3623  U+9ED4  uni9ED4,  CJK UNIFIED IDEOGRAPH-9ED4
        0x52AB, # G+3624  U+52AB  uni52AB,  CJK UNIFIED IDEOGRAPH-52AB
        0x602F, # G+3625  U+602F  uni602F,  CJK UNIFIED IDEOGRAPH-602F
        0x8FF2, # G+3626  U+8FF2  uni8FF2,  CJK UNIFIED IDEOGRAPH-8FF2
        0x5048, # G+3627  U+5048  uni5048,  CJK UNIFIED IDEOGRAPH-5048
        0x61A9, # G+3628  U+61A9  uni61A9,  CJK UNIFIED IDEOGRAPH-61A9
        0x63ED, # G+3629  U+63ED  uni63ED,  CJK UNIFIED IDEOGRAPH-63ED
        0x64CA, # G+3630  U+64CA  uni64CA,  CJK UNIFIED IDEOGRAPH-64CA
        0x683C, # G+3631  U+683C  uni683C,  CJK UNIFIED IDEOGRAPH-683C
        0x6A84, # G+3632  U+6A84  uni6A84,  CJK UNIFIED IDEOGRAPH-6A84
        0x6FC0, # G+3633  U+6FC0  uni6FC0,  CJK UNIFIED IDEOGRAPH-6FC0
        0x8188, # G+3634  U+8188  uni8188,  CJK UNIFIED IDEOGRAPH-8188
        0x89A1, # G+3635  U+89A1  uni89A1,  CJK UNIFIED IDEOGRAPH-89A1
        0x9694, # G+3636  U+9694  uni9694,  CJK UNIFIED IDEOGRAPH-9694
        0x5805, # G+3637  U+5805  uni5805,  CJK UNIFIED IDEOGRAPH-5805
        0x727D, # G+3638  U+727D  uni727D,  CJK UNIFIED IDEOGRAPH-727D
        0x72AC, # G+3639  U+72AC  uni72AC,  CJK UNIFIED IDEOGRAPH-72AC
        0x7504, # G+3640  U+7504  uni7504,  CJK UNIFIED IDEOGRAPH-7504
        0x7D79, # G+3641  U+7D79  uni7D79,  CJK UNIFIED IDEOGRAPH-7D79
        0x7E6D, # G+3642  U+7E6D  uni7E6D,  CJK UNIFIED IDEOGRAPH-7E6D
        0x80A9, # G+3643  U+80A9  uni80A9,  CJK UNIFIED IDEOGRAPH-80A9
        0x898B, # G+3644  U+898B  uni898B,  CJK UNIFIED IDEOGRAPH-898B
        0x8B74, # G+3645  U+8B74  uni8B74,  CJK UNIFIED IDEOGRAPH-8B74
        0x9063, # G+3646  U+9063  uni9063,  CJK UNIFIED IDEOGRAPH-9063
        0x9D51, # G+3647  U+9D51  uni9D51,  CJK UNIFIED IDEOGRAPH-9D51
        0x6289, # G+3648  U+6289  uni6289,  CJK UNIFIED IDEOGRAPH-6289
        0x6C7A, # G+3649  U+6C7A  uni6C7A,  CJK UNIFIED IDEOGRAPH-6C7A
        0x6F54, # G+3650  U+6F54  uni6F54,  CJK UNIFIED IDEOGRAPH-6F54
        0x7D50, # G+3651  U+7D50  uni7D50,  CJK UNIFIED IDEOGRAPH-7D50
        0x7F3A, # G+3652  U+7F3A  uni7F3A,  CJK UNIFIED IDEOGRAPH-7F3A
        0x8A23, # G+3653  U+8A23  uni8A23,  CJK UNIFIED IDEOGRAPH-8A23
        0x517C, # G+3654  U+517C  uni517C,  CJK UNIFIED IDEOGRAPH-517C
        0x614A, # G+3655  U+614A  uni614A,  CJK UNIFIED IDEOGRAPH-614A
        0x7B9D, # G+3656  U+7B9D  uni7B9D,  CJK UNIFIED IDEOGRAPH-7B9D
        0x8B19, # G+3657  U+8B19  uni8B19,  CJK UNIFIED IDEOGRAPH-8B19
        0x9257, # G+3658  U+9257  uni9257,  CJK UNIFIED IDEOGRAPH-9257
        0x938C, # G+3659  U+938C  uni938C,  CJK UNIFIED IDEOGRAPH-938C
        0x4EAC, # G+3660  U+4EAC  uni4EAC,  CJK UNIFIED IDEOGRAPH-4EAC
        0x4FD3, # G+3661  U+4FD3  uni4FD3,  CJK UNIFIED IDEOGRAPH-4FD3
        0x501E, # G+3662  U+501E  uni501E,  CJK UNIFIED IDEOGRAPH-501E
        0x50BE, # G+3663  U+50BE  uni50BE,  CJK UNIFIED IDEOGRAPH-50BE
        0x5106, # G+3664  U+5106  uni5106,  CJK UNIFIED IDEOGRAPH-5106
        0x52C1, # G+3665  U+52C1  uni52C1,  CJK UNIFIED IDEOGRAPH-52C1
        0x52CD, # G+3666  U+52CD  uni52CD,  CJK UNIFIED IDEOGRAPH-52CD
        0x537F, # G+3667  U+537F  uni537F,  CJK UNIFIED IDEOGRAPH-537F
        0x5770, # G+3668  U+5770  uni5770,  CJK UNIFIED IDEOGRAPH-5770
        0x5883, # G+3669  U+5883  uni5883,  CJK UNIFIED IDEOGRAPH-5883
        0x5E9A, # G+3670  U+5E9A  uni5E9A,  CJK UNIFIED IDEOGRAPH-5E9A
        0x5F91, # G+3671  U+5F91  uni5F91,  CJK UNIFIED IDEOGRAPH-5F91
        0x6176, # G+3672  U+6176  uni6176,  CJK UNIFIED IDEOGRAPH-6176
        0x61AC, # G+3673  U+61AC  uni61AC,  CJK UNIFIED IDEOGRAPH-61AC
        0x64CE, # G+3674  U+64CE  uni64CE,  CJK UNIFIED IDEOGRAPH-64CE
        0x656C, # G+3675  U+656C  uni656C,  CJK UNIFIED IDEOGRAPH-656C
        0x666F, # G+3676  U+666F  uni666F,  CJK UNIFIED IDEOGRAPH-666F
        0x66BB, # G+3677  U+66BB  uni66BB,  CJK UNIFIED IDEOGRAPH-66BB
        0xF901, # G+3678  U+F901  uniF901,  CJK COMPATIBILITY IDEOGRAPH-F901
        0x6897, # G+3679  U+6897  uni6897,  CJK UNIFIED IDEOGRAPH-6897
        0x6D87, # G+3680  U+6D87  uni6D87,  CJK UNIFIED IDEOGRAPH-6D87
        0x7085, # G+3681  U+7085  uni7085,  CJK UNIFIED IDEOGRAPH-7085
        0x70F1, # G+3682  U+70F1  uni70F1,  CJK UNIFIED IDEOGRAPH-70F1
        0x749F, # G+3683  U+749F  uni749F,  CJK UNIFIED IDEOGRAPH-749F
        0x74A5, # G+3684  U+74A5  uni74A5,  CJK UNIFIED IDEOGRAPH-74A5
        0x74CA, # G+3685  U+74CA  uni74CA,  CJK UNIFIED IDEOGRAPH-74CA
        0x75D9, # G+3686  U+75D9  uni75D9,  CJK UNIFIED IDEOGRAPH-75D9
        0x786C, # G+3687  U+786C  uni786C,  CJK UNIFIED IDEOGRAPH-786C
        0x78EC, # G+3688  U+78EC  uni78EC,  CJK UNIFIED IDEOGRAPH-78EC
        0x7ADF, # G+3689  U+7ADF  uni7ADF,  CJK UNIFIED IDEOGRAPH-7ADF
        0x7AF6, # G+3690  U+7AF6  uni7AF6,  CJK UNIFIED IDEOGRAPH-7AF6
        0x7D45, # G+3691  U+7D45  uni7D45,  CJK UNIFIED IDEOGRAPH-7D45
        0x7D93, # G+3692  U+7D93  uni7D93,  CJK UNIFIED IDEOGRAPH-7D93
        0x8015, # G+3693  U+8015  uni8015,  CJK UNIFIED IDEOGRAPH-8015
        0x803F, # G+3694  U+803F  uni803F,  CJK UNIFIED IDEOGRAPH-803F
        0x811B, # G+3695  U+811B  uni811B,  CJK UNIFIED IDEOGRAPH-811B
        0x8396, # G+3696  U+8396  uni8396,  CJK UNIFIED IDEOGRAPH-8396
        0x8B66, # G+3697  U+8B66  uni8B66,  CJK UNIFIED IDEOGRAPH-8B66
        0x8F15, # G+3698  U+8F15  uni8F15,  CJK UNIFIED IDEOGRAPH-8F15
        0x9015, # G+3699  U+9015  uni9015,  CJK UNIFIED IDEOGRAPH-9015
        0x93E1, # G+3700  U+93E1  uni93E1,  CJK UNIFIED IDEOGRAPH-93E1
        0x9803, # G+3701  U+9803  uni9803,  CJK UNIFIED IDEOGRAPH-9803
        0x9838, # G+3702  U+9838  uni9838,  CJK UNIFIED IDEOGRAPH-9838
        0x9A5A, # G+3703  U+9A5A  uni9A5A,  CJK UNIFIED IDEOGRAPH-9A5A
        0x9BE8, # G+3704  U+9BE8  uni9BE8,  CJK UNIFIED IDEOGRAPH-9BE8
        0x4FC2, # G+3705  U+4FC2  uni4FC2,  CJK UNIFIED IDEOGRAPH-4FC2
        0x5553, # G+3706  U+5553  uni5553,  CJK UNIFIED IDEOGRAPH-5553
        0x583A, # G+3707  U+583A  uni583A,  CJK UNIFIED IDEOGRAPH-583A
        0x5951, # G+3708  U+5951  uni5951,  CJK UNIFIED IDEOGRAPH-5951
        0x5B63, # G+3709  U+5B63  uni5B63,  CJK UNIFIED IDEOGRAPH-5B63
        0x5C46, # G+3710  U+5C46  uni5C46,  CJK UNIFIED IDEOGRAPH-5C46
        0x60B8, # G+3711  U+60B8  uni60B8,  CJK UNIFIED IDEOGRAPH-60B8
        0x6212, # G+3712  U+6212  uni6212,  CJK UNIFIED IDEOGRAPH-6212
        0x6842, # G+3713  U+6842  uni6842,  CJK UNIFIED IDEOGRAPH-6842
        0x68B0, # G+3714  U+68B0  uni68B0,  CJK UNIFIED IDEOGRAPH-68B0
        0x68E8, # G+3715  U+68E8  uni68E8,  CJK UNIFIED IDEOGRAPH-68E8
        0x6EAA, # G+3716  U+6EAA  uni6EAA,  CJK UNIFIED IDEOGRAPH-6EAA
        0x754C, # G+3717  U+754C  uni754C,  CJK UNIFIED IDEOGRAPH-754C
        0x7678, # G+3718  U+7678  uni7678,  CJK UNIFIED IDEOGRAPH-7678
        0x78CE, # G+3719  U+78CE  uni78CE,  CJK UNIFIED IDEOGRAPH-78CE
        0x7A3D, # G+3720  U+7A3D  uni7A3D,  CJK UNIFIED IDEOGRAPH-7A3D
        0x7CFB, # G+3721  U+7CFB  uni7CFB,  CJK UNIFIED IDEOGRAPH-7CFB
        0x7E6B, # G+3722  U+7E6B  uni7E6B,  CJK UNIFIED IDEOGRAPH-7E6B
        0x7E7C, # G+3723  U+7E7C  uni7E7C,  CJK UNIFIED IDEOGRAPH-7E7C
        0x8A08, # G+3724  U+8A08  uni8A08,  CJK UNIFIED IDEOGRAPH-8A08
        0x8AA1, # G+3725  U+8AA1  uni8AA1,  CJK UNIFIED IDEOGRAPH-8AA1
        0x8C3F, # G+3726  U+8C3F  uni8C3F,  CJK UNIFIED IDEOGRAPH-8C3F
        0x968E, # G+3727  U+968E  uni968E,  CJK UNIFIED IDEOGRAPH-968E
        0x9DC4, # G+3728  U+9DC4  uni9DC4,  CJK UNIFIED IDEOGRAPH-9DC4
        0x53E4, # G+3729  U+53E4  uni53E4,  CJK UNIFIED IDEOGRAPH-53E4
        0x53E9, # G+3730  U+53E9  uni53E9,  CJK UNIFIED IDEOGRAPH-53E9
        0x544A, # G+3731  U+544A  uni544A,  CJK UNIFIED IDEOGRAPH-544A
        0x5471, # G+3732  U+5471  uni5471,  CJK UNIFIED IDEOGRAPH-5471
        0x56FA, # G+3733  U+56FA  uni56FA,  CJK UNIFIED IDEOGRAPH-56FA
        0x59D1, # G+3734  U+59D1  uni59D1,  CJK UNIFIED IDEOGRAPH-59D1
        0x5B64, # G+3735  U+5B64  uni5B64,  CJK UNIFIED IDEOGRAPH-5B64
        0x5C3B, # G+3736  U+5C3B  uni5C3B,  CJK UNIFIED IDEOGRAPH-5C3B
        0x5EAB, # G+3737  U+5EAB  uni5EAB,  CJK UNIFIED IDEOGRAPH-5EAB
        0x62F7, # G+3738  U+62F7  uni62F7,  CJK UNIFIED IDEOGRAPH-62F7
        0x6537, # G+3739  U+6537  uni6537,  CJK UNIFIED IDEOGRAPH-6537
        0x6545, # G+3740  U+6545  uni6545,  CJK UNIFIED IDEOGRAPH-6545
        0x6572, # G+3741  U+6572  uni6572,  CJK UNIFIED IDEOGRAPH-6572
        0x66A0, # G+3742  U+66A0  uni66A0,  CJK UNIFIED IDEOGRAPH-66A0
        0x67AF, # G+3743  U+67AF  uni67AF,  CJK UNIFIED IDEOGRAPH-67AF
        0x69C1, # G+3744  U+69C1  uni69C1,  CJK UNIFIED IDEOGRAPH-69C1
        0x6CBD, # G+3745  U+6CBD  uni6CBD,  CJK UNIFIED IDEOGRAPH-6CBD
        0x75FC, # G+3746  U+75FC  uni75FC,  CJK UNIFIED IDEOGRAPH-75FC
        0x7690, # G+3747  U+7690  uni7690,  CJK UNIFIED IDEOGRAPH-7690
        0x777E, # G+3748  U+777E  uni777E,  CJK UNIFIED IDEOGRAPH-777E
        0x7A3F, # G+3749  U+7A3F  uni7A3F,  CJK UNIFIED IDEOGRAPH-7A3F
        0x7F94, # G+3750  U+7F94  uni7F94,  CJK UNIFIED IDEOGRAPH-7F94
        0x8003, # G+3751  U+8003  uni8003,  CJK UNIFIED IDEOGRAPH-8003
        0x80A1, # G+3752  U+80A1  uni80A1,  CJK UNIFIED IDEOGRAPH-80A1
        0x818F, # G+3753  U+818F  uni818F,  CJK UNIFIED IDEOGRAPH-818F
        0x82E6, # G+3754  U+82E6  uni82E6,  CJK UNIFIED IDEOGRAPH-82E6
        0x82FD, # G+3755  U+82FD  uni82FD,  CJK UNIFIED IDEOGRAPH-82FD
        0x83F0, # G+3756  U+83F0  uni83F0,  CJK UNIFIED IDEOGRAPH-83F0
        0x85C1, # G+3757  U+85C1  uni85C1,  CJK UNIFIED IDEOGRAPH-85C1
        0x8831, # G+3758  U+8831  uni8831,  CJK UNIFIED IDEOGRAPH-8831
        0x88B4, # G+3759  U+88B4  uni88B4,  CJK UNIFIED IDEOGRAPH-88B4
        0x8AA5, # G+3760  U+8AA5  uni8AA5,  CJK UNIFIED IDEOGRAPH-8AA5
        0x8F9C, # G+3761  U+8F9C  uni8F9C,  CJK UNIFIED IDEOGRAPH-8F9C
        0x932E, # G+3762  U+932E  uni932E,  CJK UNIFIED IDEOGRAPH-932E
        0x96C7, # G+3763  U+96C7  uni96C7,  CJK UNIFIED IDEOGRAPH-96C7
        0x9867, # G+3764  U+9867  uni9867,  CJK UNIFIED IDEOGRAPH-9867
        0x9AD8, # G+3765  U+9AD8  uni9AD8,  CJK UNIFIED IDEOGRAPH-9AD8
        0x9F13, # G+3766  U+9F13  uni9F13,  CJK UNIFIED IDEOGRAPH-9F13
        0x54ED, # G+3767  U+54ED  uni54ED,  CJK UNIFIED IDEOGRAPH-54ED
        0x659B, # G+3768  U+659B  uni659B,  CJK UNIFIED IDEOGRAPH-659B
        0x66F2, # G+3769  U+66F2  uni66F2,  CJK UNIFIED IDEOGRAPH-66F2
        0x688F, # G+3770  U+688F  uni688F,  CJK UNIFIED IDEOGRAPH-688F
        0x7A40, # G+3771  U+7A40  uni7A40,  CJK UNIFIED IDEOGRAPH-7A40
        0x8C37, # G+3772  U+8C37  uni8C37,  CJK UNIFIED IDEOGRAPH-8C37
        0x9D60, # G+3773  U+9D60  uni9D60,  CJK UNIFIED IDEOGRAPH-9D60
        0x56F0, # G+3774  U+56F0  uni56F0,  CJK UNIFIED IDEOGRAPH-56F0
        0x5764, # G+3775  U+5764  uni5764,  CJK UNIFIED IDEOGRAPH-5764
        0x5D11, # G+3776  U+5D11  uni5D11,  CJK UNIFIED IDEOGRAPH-5D11
        0x6606, # G+3777  U+6606  uni6606,  CJK UNIFIED IDEOGRAPH-6606
        0x68B1, # G+3778  U+68B1  uni68B1,  CJK UNIFIED IDEOGRAPH-68B1
        0x68CD, # G+3779  U+68CD  uni68CD,  CJK UNIFIED IDEOGRAPH-68CD
        0x6EFE, # G+3780  U+6EFE  uni6EFE,  CJK UNIFIED IDEOGRAPH-6EFE
        0x7428, # G+3781  U+7428  uni7428,  CJK UNIFIED IDEOGRAPH-7428
        0x889E, # G+3782  U+889E  uni889E,  CJK UNIFIED IDEOGRAPH-889E
        0x9BE4, # G+3783  U+9BE4  uni9BE4,  CJK UNIFIED IDEOGRAPH-9BE4
        0x6C68, # G+3784  U+6C68  uni6C68,  CJK UNIFIED IDEOGRAPH-6C68
        0x9AA8, # G+3785  U+9AA8  uni9AA8,  CJK UNIFIED IDEOGRAPH-9AA8
        0x4F9B, # G+3786  U+4F9B  uni4F9B,  CJK UNIFIED IDEOGRAPH-4F9B
        0x516C, # G+3787  U+516C  uni516C,  CJK UNIFIED IDEOGRAPH-516C
        0x5171, # G+3788  U+5171  uni5171,  CJK UNIFIED IDEOGRAPH-5171
        0x529F, # G+3789  U+529F  uni529F,  CJK UNIFIED IDEOGRAPH-529F
        0x5B54, # G+3790  U+5B54  uni5B54,  CJK UNIFIED IDEOGRAPH-5B54
        0x5DE5, # G+3791  U+5DE5  uni5DE5,  CJK UNIFIED IDEOGRAPH-5DE5
        0x6050, # G+3792  U+6050  uni6050,  CJK UNIFIED IDEOGRAPH-6050
        0x606D, # G+3793  U+606D  uni606D,  CJK UNIFIED IDEOGRAPH-606D
        0x62F1, # G+3794  U+62F1  uni62F1,  CJK UNIFIED IDEOGRAPH-62F1
        0x63A7, # G+3795  U+63A7  uni63A7,  CJK UNIFIED IDEOGRAPH-63A7
        0x653B, # G+3796  U+653B  uni653B,  CJK UNIFIED IDEOGRAPH-653B
        0x73D9, # G+3797  U+73D9  uni73D9,  CJK UNIFIED IDEOGRAPH-73D9
        0x7A7A, # G+3798  U+7A7A  uni7A7A,  CJK UNIFIED IDEOGRAPH-7A7A
        0x86A3, # G+3799  U+86A3  uni86A3,  CJK UNIFIED IDEOGRAPH-86A3
        0x8CA2, # G+3800  U+8CA2  uni8CA2,  CJK UNIFIED IDEOGRAPH-8CA2
        0x978F, # G+3801  U+978F  uni978F,  CJK UNIFIED IDEOGRAPH-978F
        0x4E32, # G+3802  U+4E32  uni4E32,  CJK UNIFIED IDEOGRAPH-4E32
        0x5BE1, # G+3803  U+5BE1  uni5BE1,  CJK UNIFIED IDEOGRAPH-5BE1
        0x6208, # G+3804  U+6208  uni6208,  CJK UNIFIED IDEOGRAPH-6208
        0x679C, # G+3805  U+679C  uni679C,  CJK UNIFIED IDEOGRAPH-679C
        0x74DC, # G+3806  U+74DC  uni74DC,  CJK UNIFIED IDEOGRAPH-74DC
        0x79D1, # G+3807  U+79D1  uni79D1,  CJK UNIFIED IDEOGRAPH-79D1
        0x83D3, # G+3808  U+83D3  uni83D3,  CJK UNIFIED IDEOGRAPH-83D3
        0x8A87, # G+3809  U+8A87  uni8A87,  CJK UNIFIED IDEOGRAPH-8A87
        0x8AB2, # G+3810  U+8AB2  uni8AB2,  CJK UNIFIED IDEOGRAPH-8AB2
        0x8DE8, # G+3811  U+8DE8  uni8DE8,  CJK UNIFIED IDEOGRAPH-8DE8
        0x904E, # G+3812  U+904E  uni904E,  CJK UNIFIED IDEOGRAPH-904E
        0x934B, # G+3813  U+934B  uni934B,  CJK UNIFIED IDEOGRAPH-934B
        0x9846, # G+3814  U+9846  uni9846,  CJK UNIFIED IDEOGRAPH-9846
        0x5ED3, # G+3815  U+5ED3  uni5ED3,  CJK UNIFIED IDEOGRAPH-5ED3
        0x69E8, # G+3816  U+69E8  uni69E8,  CJK UNIFIED IDEOGRAPH-69E8
        0x85FF, # G+3817  U+85FF  uni85FF,  CJK UNIFIED IDEOGRAPH-85FF
        0x90ED, # G+3818  U+90ED  uni90ED,  CJK UNIFIED IDEOGRAPH-90ED
        0x51A0, # G+3819  U+51A0  uni51A0,  CJK UNIFIED IDEOGRAPH-51A0
        0x5B98, # G+3820  U+5B98  uni5B98,  CJK UNIFIED IDEOGRAPH-5B98
        0x5BEC, # G+3821  U+5BEC  uni5BEC,  CJK UNIFIED IDEOGRAPH-5BEC
        0x6163, # G+3822  U+6163  uni6163,  CJK UNIFIED IDEOGRAPH-6163
        0x68FA, # G+3823  U+68FA  uni68FA,  CJK UNIFIED IDEOGRAPH-68FA
        0x6B3E, # G+3824  U+6B3E  uni6B3E,  CJK UNIFIED IDEOGRAPH-6B3E
        0x704C, # G+3825  U+704C  uni704C,  CJK UNIFIED IDEOGRAPH-704C
        0x742F, # G+3826  U+742F  uni742F,  CJK UNIFIED IDEOGRAPH-742F
        0x74D8, # G+3827  U+74D8  uni74D8,  CJK UNIFIED IDEOGRAPH-74D8
        0x7BA1, # G+3828  U+7BA1  uni7BA1,  CJK UNIFIED IDEOGRAPH-7BA1
        0x7F50, # G+3829  U+7F50  uni7F50,  CJK UNIFIED IDEOGRAPH-7F50
        0x83C5, # G+3830  U+83C5  uni83C5,  CJK UNIFIED IDEOGRAPH-83C5
        0x89C0, # G+3831  U+89C0  uni89C0,  CJK UNIFIED IDEOGRAPH-89C0
        0x8CAB, # G+3832  U+8CAB  uni8CAB,  CJK UNIFIED IDEOGRAPH-8CAB
        0x95DC, # G+3833  U+95DC  uni95DC,  CJK UNIFIED IDEOGRAPH-95DC
        0x9928, # G+3834  U+9928  uni9928,  CJK UNIFIED IDEOGRAPH-9928
        0x522E, # G+3835  U+522E  uni522E,  CJK UNIFIED IDEOGRAPH-522E
        0x605D, # G+3836  U+605D  uni605D,  CJK UNIFIED IDEOGRAPH-605D
        0x62EC, # G+3837  U+62EC  uni62EC,  CJK UNIFIED IDEOGRAPH-62EC
        0x9002, # G+3838  U+9002  uni9002,  CJK UNIFIED IDEOGRAPH-9002
        0x4F8A, # G+3839  U+4F8A  uni4F8A,  CJK UNIFIED IDEOGRAPH-4F8A
        0x5149, # G+3840  U+5149  uni5149,  CJK UNIFIED IDEOGRAPH-5149
        0x5321, # G+3841  U+5321  uni5321,  CJK UNIFIED IDEOGRAPH-5321
        0x58D9, # G+3842  U+58D9  uni58D9,  CJK UNIFIED IDEOGRAPH-58D9
        0x5EE3, # G+3843  U+5EE3  uni5EE3,  CJK UNIFIED IDEOGRAPH-5EE3
        0x66E0, # G+3844  U+66E0  uni66E0,  CJK UNIFIED IDEOGRAPH-66E0
        0x6D38, # G+3845  U+6D38  uni6D38,  CJK UNIFIED IDEOGRAPH-6D38
        0x709A, # G+3846  U+709A  uni709A,  CJK UNIFIED IDEOGRAPH-709A
        0x72C2, # G+3847  U+72C2  uni72C2,  CJK UNIFIED IDEOGRAPH-72C2
        0x73D6, # G+3848  U+73D6  uni73D6,  CJK UNIFIED IDEOGRAPH-73D6
        0x7B50, # G+3849  U+7B50  uni7B50,  CJK UNIFIED IDEOGRAPH-7B50
        0x80F1, # G+3850  U+80F1  uni80F1,  CJK UNIFIED IDEOGRAPH-80F1
        0x945B, # G+3851  U+945B  uni945B,  CJK UNIFIED IDEOGRAPH-945B
        0x5366, # G+3852  U+5366  uni5366,  CJK UNIFIED IDEOGRAPH-5366
        0x639B, # G+3853  U+639B  uni639B,  CJK UNIFIED IDEOGRAPH-639B
        0x7F6B, # G+3854  U+7F6B  uni7F6B,  CJK UNIFIED IDEOGRAPH-7F6B
        0x4E56, # G+3855  U+4E56  uni4E56,  CJK UNIFIED IDEOGRAPH-4E56
        0x5080, # G+3856  U+5080  uni5080,  CJK UNIFIED IDEOGRAPH-5080
        0x584A, # G+3857  U+584A  uni584A,  CJK UNIFIED IDEOGRAPH-584A
        0x58DE, # G+3858  U+58DE  uni58DE,  CJK UNIFIED IDEOGRAPH-58DE
        0x602A, # G+3859  U+602A  uni602A,  CJK UNIFIED IDEOGRAPH-602A
        0x6127, # G+3860  U+6127  uni6127,  CJK UNIFIED IDEOGRAPH-6127
        0x62D0, # G+3861  U+62D0  uni62D0,  CJK UNIFIED IDEOGRAPH-62D0
        0x69D0, # G+3862  U+69D0  uni69D0,  CJK UNIFIED IDEOGRAPH-69D0
        0x9B41, # G+3863  U+9B41  uni9B41,  CJK UNIFIED IDEOGRAPH-9B41
        0x5B8F, # G+3864  U+5B8F  uni5B8F,  CJK UNIFIED IDEOGRAPH-5B8F
        0x7D18, # G+3865  U+7D18  uni7D18,  CJK UNIFIED IDEOGRAPH-7D18
        0x80B1, # G+3866  U+80B1  uni80B1,  CJK UNIFIED IDEOGRAPH-80B1
        0x8F5F, # G+3867  U+8F5F  uni8F5F,  CJK UNIFIED IDEOGRAPH-8F5F
        0x4EA4, # G+3868  U+4EA4  uni4EA4,  CJK UNIFIED IDEOGRAPH-4EA4
        0x50D1, # G+3869  U+50D1  uni50D1,  CJK UNIFIED IDEOGRAPH-50D1
        0x54AC, # G+3870  U+54AC  uni54AC,  CJK UNIFIED IDEOGRAPH-54AC
        0x55AC, # G+3871  U+55AC  uni55AC,  CJK UNIFIED IDEOGRAPH-55AC
        0x5B0C, # G+3872  U+5B0C  uni5B0C,  CJK UNIFIED IDEOGRAPH-5B0C
        0x5DA0, # G+3873  U+5DA0  uni5DA0,  CJK UNIFIED IDEOGRAPH-5DA0
        0x5DE7, # G+3874  U+5DE7  uni5DE7,  CJK UNIFIED IDEOGRAPH-5DE7
        0x652A, # G+3875  U+652A  uni652A,  CJK UNIFIED IDEOGRAPH-652A
        0x654E, # G+3876  U+654E  uni654E,  CJK UNIFIED IDEOGRAPH-654E
        0x6821, # G+3877  U+6821  uni6821,  CJK UNIFIED IDEOGRAPH-6821
        0x6A4B, # G+3878  U+6A4B  uni6A4B,  CJK UNIFIED IDEOGRAPH-6A4B
        0x72E1, # G+3879  U+72E1  uni72E1,  CJK UNIFIED IDEOGRAPH-72E1
        0x768E, # G+3880  U+768E  uni768E,  CJK UNIFIED IDEOGRAPH-768E
        0x77EF, # G+3881  U+77EF  uni77EF,  CJK UNIFIED IDEOGRAPH-77EF
        0x7D5E, # G+3882  U+7D5E  uni7D5E,  CJK UNIFIED IDEOGRAPH-7D5E
        0x7FF9, # G+3883  U+7FF9  uni7FF9,  CJK UNIFIED IDEOGRAPH-7FF9
        0x81A0, # G+3884  U+81A0  uni81A0,  CJK UNIFIED IDEOGRAPH-81A0
        0x854E, # G+3885  U+854E  uni854E,  CJK UNIFIED IDEOGRAPH-854E
        0x86DF, # G+3886  U+86DF  uni86DF,  CJK UNIFIED IDEOGRAPH-86DF
        0x8F03, # G+3887  U+8F03  uni8F03,  CJK UNIFIED IDEOGRAPH-8F03
        0x8F4E, # G+3888  U+8F4E  uni8F4E,  CJK UNIFIED IDEOGRAPH-8F4E
        0x90CA, # G+3889  U+90CA  uni90CA,  CJK UNIFIED IDEOGRAPH-90CA
        0x9903, # G+3890  U+9903  uni9903,  CJK UNIFIED IDEOGRAPH-9903
        0x9A55, # G+3891  U+9A55  uni9A55,  CJK UNIFIED IDEOGRAPH-9A55
        0x9BAB, # G+3892  U+9BAB  uni9BAB,  CJK UNIFIED IDEOGRAPH-9BAB
        0x4E18, # G+3893  U+4E18  uni4E18,  CJK UNIFIED IDEOGRAPH-4E18
        0x4E45, # G+3894  U+4E45  uni4E45,  CJK UNIFIED IDEOGRAPH-4E45
        0x4E5D, # G+3895  U+4E5D  uni4E5D,  CJK UNIFIED IDEOGRAPH-4E5D
        0x4EC7, # G+3896  U+4EC7  uni4EC7,  CJK UNIFIED IDEOGRAPH-4EC7
        0x4FF1, # G+3897  U+4FF1  uni4FF1,  CJK UNIFIED IDEOGRAPH-4FF1
        0x5177, # G+3898  U+5177  uni5177,  CJK UNIFIED IDEOGRAPH-5177
        0x52FE, # G+3899  U+52FE  uni52FE,  CJK UNIFIED IDEOGRAPH-52FE
        0x5340, # G+3900  U+5340  uni5340,  CJK UNIFIED IDEOGRAPH-5340
        0x53E3, # G+3901  U+53E3  uni53E3,  CJK UNIFIED IDEOGRAPH-53E3
        0x53E5, # G+3902  U+53E5  uni53E5,  CJK UNIFIED IDEOGRAPH-53E5
        0x548E, # G+3903  U+548E  uni548E,  CJK UNIFIED IDEOGRAPH-548E
        0x5614, # G+3904  U+5614  uni5614,  CJK UNIFIED IDEOGRAPH-5614
        0x5775, # G+3905  U+5775  uni5775,  CJK UNIFIED IDEOGRAPH-5775
        0x57A2, # G+3906  U+57A2  uni57A2,  CJK UNIFIED IDEOGRAPH-57A2
        0x5BC7, # G+3907  U+5BC7  uni5BC7,  CJK UNIFIED IDEOGRAPH-5BC7
        0x5D87, # G+3908  U+5D87  uni5D87,  CJK UNIFIED IDEOGRAPH-5D87
        0x5ED0, # G+3909  U+5ED0  uni5ED0,  CJK UNIFIED IDEOGRAPH-5ED0
        0x61FC, # G+3910  U+61FC  uni61FC,  CJK UNIFIED IDEOGRAPH-61FC
        0x62D8, # G+3911  U+62D8  uni62D8,  CJK UNIFIED IDEOGRAPH-62D8
        0x6551, # G+3912  U+6551  uni6551,  CJK UNIFIED IDEOGRAPH-6551
        0x67B8, # G+3913  U+67B8  uni67B8,  CJK UNIFIED IDEOGRAPH-67B8
        0x67E9, # G+3914  U+67E9  uni67E9,  CJK UNIFIED IDEOGRAPH-67E9
        0x69CB, # G+3915  U+69CB  uni69CB,  CJK UNIFIED IDEOGRAPH-69CB
        0x6B50, # G+3916  U+6B50  uni6B50,  CJK UNIFIED IDEOGRAPH-6B50
        0x6BC6, # G+3917  U+6BC6  uni6BC6,  CJK UNIFIED IDEOGRAPH-6BC6
        0x6BEC, # G+3918  U+6BEC  uni6BEC,  CJK UNIFIED IDEOGRAPH-6BEC
        0x6C42, # G+3919  U+6C42  uni6C42,  CJK UNIFIED IDEOGRAPH-6C42
        0x6E9D, # G+3920  U+6E9D  uni6E9D,  CJK UNIFIED IDEOGRAPH-6E9D
        0x7078, # G+3921  U+7078  uni7078,  CJK UNIFIED IDEOGRAPH-7078
        0x72D7, # G+3922  U+72D7  uni72D7,  CJK UNIFIED IDEOGRAPH-72D7
        0x7396, # G+3923  U+7396  uni7396,  CJK UNIFIED IDEOGRAPH-7396
        0x7403, # G+3924  U+7403  uni7403,  CJK UNIFIED IDEOGRAPH-7403
        0x77BF, # G+3925  U+77BF  uni77BF,  CJK UNIFIED IDEOGRAPH-77BF
        0x77E9, # G+3926  U+77E9  uni77E9,  CJK UNIFIED IDEOGRAPH-77E9
        0x7A76, # G+3927  U+7A76  uni7A76,  CJK UNIFIED IDEOGRAPH-7A76
        0x7D7F, # G+3928  U+7D7F  uni7D7F,  CJK UNIFIED IDEOGRAPH-7D7F
        0x8009, # G+3929  U+8009  uni8009,  CJK UNIFIED IDEOGRAPH-8009
        0x81FC, # G+3930  U+81FC  uni81FC,  CJK UNIFIED IDEOGRAPH-81FC
        0x8205, # G+3931  U+8205  uni8205,  CJK UNIFIED IDEOGRAPH-8205
        0x820A, # G+3932  U+820A  uni820A,  CJK UNIFIED IDEOGRAPH-820A
        0x82DF, # G+3933  U+82DF  uni82DF,  CJK UNIFIED IDEOGRAPH-82DF
        0x8862, # G+3934  U+8862  uni8862,  CJK UNIFIED IDEOGRAPH-8862
        0x8B33, # G+3935  U+8B33  uni8B33,  CJK UNIFIED IDEOGRAPH-8B33
        0x8CFC, # G+3936  U+8CFC  uni8CFC,  CJK UNIFIED IDEOGRAPH-8CFC
        0x8EC0, # G+3937  U+8EC0  uni8EC0,  CJK UNIFIED IDEOGRAPH-8EC0
        0x9011, # G+3938  U+9011  uni9011,  CJK UNIFIED IDEOGRAPH-9011
        0x90B1, # G+3939  U+90B1  uni90B1,  CJK UNIFIED IDEOGRAPH-90B1
        0x9264, # G+3940  U+9264  uni9264,  CJK UNIFIED IDEOGRAPH-9264
        0x92B6, # G+3941  U+92B6  uni92B6,  CJK UNIFIED IDEOGRAPH-92B6
        0x99D2, # G+3942  U+99D2  uni99D2,  CJK UNIFIED IDEOGRAPH-99D2
        0x9A45, # G+3943  U+9A45  uni9A45,  CJK UNIFIED IDEOGRAPH-9A45
        0x9CE9, # G+3944  U+9CE9  uni9CE9,  CJK UNIFIED IDEOGRAPH-9CE9
        0x9DD7, # G+3945  U+9DD7  uni9DD7,  CJK UNIFIED IDEOGRAPH-9DD7
        0x9F9C, # G+3946  U+9F9C  uni9F9C,  CJK UNIFIED IDEOGRAPH-9F9C
        0x570B, # G+3947  U+570B  uni570B,  CJK UNIFIED IDEOGRAPH-570B
        0x5C40, # G+3948  U+5C40  uni5C40,  CJK UNIFIED IDEOGRAPH-5C40
        0x83CA, # G+3949  U+83CA  uni83CA,  CJK UNIFIED IDEOGRAPH-83CA
        0x97A0, # G+3950  U+97A0  uni97A0,  CJK UNIFIED IDEOGRAPH-97A0
        0x97AB, # G+3951  U+97AB  uni97AB,  CJK UNIFIED IDEOGRAPH-97AB
        0x9EB4, # G+3952  U+9EB4  uni9EB4,  CJK UNIFIED IDEOGRAPH-9EB4
        0x541B, # G+3953  U+541B  uni541B,  CJK UNIFIED IDEOGRAPH-541B
        0x7A98, # G+3954  U+7A98  uni7A98,  CJK UNIFIED IDEOGRAPH-7A98
        0x7FA4, # G+3955  U+7FA4  uni7FA4,  CJK UNIFIED IDEOGRAPH-7FA4
        0x88D9, # G+3956  U+88D9  uni88D9,  CJK UNIFIED IDEOGRAPH-88D9
        0x8ECD, # G+3957  U+8ECD  uni8ECD,  CJK UNIFIED IDEOGRAPH-8ECD
        0x90E1, # G+3958  U+90E1  uni90E1,  CJK UNIFIED IDEOGRAPH-90E1
        0x5800, # G+3959  U+5800  uni5800,  CJK UNIFIED IDEOGRAPH-5800
        0x5C48, # G+3960  U+5C48  uni5C48,  CJK UNIFIED IDEOGRAPH-5C48
        0x6398, # G+3961  U+6398  uni6398,  CJK UNIFIED IDEOGRAPH-6398
        0x7A9F, # G+3962  U+7A9F  uni7A9F,  CJK UNIFIED IDEOGRAPH-7A9F
        0x5BAE, # G+3963  U+5BAE  uni5BAE,  CJK UNIFIED IDEOGRAPH-5BAE
        0x5F13, # G+3964  U+5F13  uni5F13,  CJK UNIFIED IDEOGRAPH-5F13
        0x7A79, # G+3965  U+7A79  uni7A79,  CJK UNIFIED IDEOGRAPH-7A79
        0x7AAE, # G+3966  U+7AAE  uni7AAE,  CJK UNIFIED IDEOGRAPH-7AAE
        0x828E, # G+3967  U+828E  uni828E,  CJK UNIFIED IDEOGRAPH-828E
        0x8EAC, # G+3968  U+8EAC  uni8EAC,  CJK UNIFIED IDEOGRAPH-8EAC
        0x5026, # G+3969  U+5026  uni5026,  CJK UNIFIED IDEOGRAPH-5026
        0x5238, # G+3970  U+5238  uni5238,  CJK UNIFIED IDEOGRAPH-5238
        0x52F8, # G+3971  U+52F8  uni52F8,  CJK UNIFIED IDEOGRAPH-52F8
        0x5377, # G+3972  U+5377  uni5377,  CJK UNIFIED IDEOGRAPH-5377
        0x5708, # G+3973  U+5708  uni5708,  CJK UNIFIED IDEOGRAPH-5708
        0x62F3, # G+3974  U+62F3  uni62F3,  CJK UNIFIED IDEOGRAPH-62F3
        0x6372, # G+3975  U+6372  uni6372,  CJK UNIFIED IDEOGRAPH-6372
        0x6B0A, # G+3976  U+6B0A  uni6B0A,  CJK UNIFIED IDEOGRAPH-6B0A
        0x6DC3, # G+3977  U+6DC3  uni6DC3,  CJK UNIFIED IDEOGRAPH-6DC3
        0x7737, # G+3978  U+7737  uni7737,  CJK UNIFIED IDEOGRAPH-7737
        0x53A5, # G+3979  U+53A5  uni53A5,  CJK UNIFIED IDEOGRAPH-53A5
        0x7357, # G+3980  U+7357  uni7357,  CJK UNIFIED IDEOGRAPH-7357
        0x8568, # G+3981  U+8568  uni8568,  CJK UNIFIED IDEOGRAPH-8568
        0x8E76, # G+3982  U+8E76  uni8E76,  CJK UNIFIED IDEOGRAPH-8E76
        0x95D5, # G+3983  U+95D5  uni95D5,  CJK UNIFIED IDEOGRAPH-95D5
        0x673A, # G+3984  U+673A  uni673A,  CJK UNIFIED IDEOGRAPH-673A
        0x6AC3, # G+3985  U+6AC3  uni6AC3,  CJK UNIFIED IDEOGRAPH-6AC3
        0x6F70, # G+3986  U+6F70  uni6F70,  CJK UNIFIED IDEOGRAPH-6F70
        0x8A6D, # G+3987  U+8A6D  uni8A6D,  CJK UNIFIED IDEOGRAPH-8A6D
        0x8ECC, # G+3988  U+8ECC  uni8ECC,  CJK UNIFIED IDEOGRAPH-8ECC
        0x994B, # G+3989  U+994B  uni994B,  CJK UNIFIED IDEOGRAPH-994B
        0x6677, # G+3990  U+6677  uni6677,  CJK UNIFIED IDEOGRAPH-6677
        0x6B78, # G+3991  U+6B78  uni6B78,  CJK UNIFIED IDEOGRAPH-6B78
        0x8CB4, # G+3992  U+8CB4  uni8CB4,  CJK UNIFIED IDEOGRAPH-8CB4
        0x9B3C, # G+3993  U+9B3C  uni9B3C,  CJK UNIFIED IDEOGRAPH-9B3C
        0x53EB, # G+3994  U+53EB  uni53EB,  CJK UNIFIED IDEOGRAPH-53EB
        0x572D, # G+3995  U+572D  uni572D,  CJK UNIFIED IDEOGRAPH-572D
        0x594E, # G+3996  U+594E  uni594E,  CJK UNIFIED IDEOGRAPH-594E
        0x63C6, # G+3997  U+63C6  uni63C6,  CJK UNIFIED IDEOGRAPH-63C6
        0x69FB, # G+3998  U+69FB  uni69FB,  CJK UNIFIED IDEOGRAPH-69FB
        0x73EA, # G+3999  U+73EA  uni73EA,  CJK UNIFIED IDEOGRAPH-73EA
        0x7845, # G+4000  U+7845  uni7845,  CJK UNIFIED IDEOGRAPH-7845
        0x7ABA, # G+4001  U+7ABA  uni7ABA,  CJK UNIFIED IDEOGRAPH-7ABA
        0x7AC5, # G+4002  U+7AC5  uni7AC5,  CJK UNIFIED IDEOGRAPH-7AC5
        0x7CFE, # G+4003  U+7CFE  uni7CFE,  CJK UNIFIED IDEOGRAPH-7CFE
        0x8475, # G+4004  U+8475  uni8475,  CJK UNIFIED IDEOGRAPH-8475
        0x898F, # G+4005  U+898F  uni898F,  CJK UNIFIED IDEOGRAPH-898F
        0x8D73, # G+4006  U+8D73  uni8D73,  CJK UNIFIED IDEOGRAPH-8D73
        0x9035, # G+4007  U+9035  uni9035,  CJK UNIFIED IDEOGRAPH-9035
        0x95A8, # G+4008  U+95A8  uni95A8,  CJK UNIFIED IDEOGRAPH-95A8
        0x52FB, # G+4009  U+52FB  uni52FB,  CJK UNIFIED IDEOGRAPH-52FB
        0x5747, # G+4010  U+5747  uni5747,  CJK UNIFIED IDEOGRAPH-5747
        0x7547, # G+4011  U+7547  uni7547,  CJK UNIFIED IDEOGRAPH-7547
        0x7B60, # G+4012  U+7B60  uni7B60,  CJK UNIFIED IDEOGRAPH-7B60
        0x83CC, # G+4013  U+83CC  uni83CC,  CJK UNIFIED IDEOGRAPH-83CC
        0x921E, # G+4014  U+921E  uni921E,  CJK UNIFIED IDEOGRAPH-921E
        0x6A58, # G+4015  U+6A58  uni6A58,  CJK UNIFIED IDEOGRAPH-6A58
        0x514B, # G+4016  U+514B  uni514B,  CJK UNIFIED IDEOGRAPH-514B
        0x524B, # G+4017  U+524B  uni524B,  CJK UNIFIED IDEOGRAPH-524B
        0x5287, # G+4018  U+5287  uni5287,  CJK UNIFIED IDEOGRAPH-5287
        0x621F, # G+4019  U+621F  uni621F,  CJK UNIFIED IDEOGRAPH-621F
        0x68D8, # G+4020  U+68D8  uni68D8,  CJK UNIFIED IDEOGRAPH-68D8
        0x6975, # G+4021  U+6975  uni6975,  CJK UNIFIED IDEOGRAPH-6975
        0x9699, # G+4022  U+9699  uni9699,  CJK UNIFIED IDEOGRAPH-9699
        0x50C5, # G+4023  U+50C5  uni50C5,  CJK UNIFIED IDEOGRAPH-50C5
        0x52A4, # G+4024  U+52A4  uni52A4,  CJK UNIFIED IDEOGRAPH-52A4
        0x52E4, # G+4025  U+52E4  uni52E4,  CJK UNIFIED IDEOGRAPH-52E4
        0x61C3, # G+4026  U+61C3  uni61C3,  CJK UNIFIED IDEOGRAPH-61C3
        0x65A4, # G+4027  U+65A4  uni65A4,  CJK UNIFIED IDEOGRAPH-65A4
        0x6839, # G+4028  U+6839  uni6839,  CJK UNIFIED IDEOGRAPH-6839
        0x69FF, # G+4029  U+69FF  uni69FF,  CJK UNIFIED IDEOGRAPH-69FF
        0x747E, # G+4030  U+747E  uni747E,  CJK UNIFIED IDEOGRAPH-747E
        0x7B4B, # G+4031  U+7B4B  uni7B4B,  CJK UNIFIED IDEOGRAPH-7B4B
        0x82B9, # G+4032  U+82B9  uni82B9,  CJK UNIFIED IDEOGRAPH-82B9
        0x83EB, # G+4033  U+83EB  uni83EB,  CJK UNIFIED IDEOGRAPH-83EB
        0x89B2, # G+4034  U+89B2  uni89B2,  CJK UNIFIED IDEOGRAPH-89B2
        0x8B39, # G+4035  U+8B39  uni8B39,  CJK UNIFIED IDEOGRAPH-8B39
        0x8FD1, # G+4036  U+8FD1  uni8FD1,  CJK UNIFIED IDEOGRAPH-8FD1
        0x9949, # G+4037  U+9949  uni9949,  CJK UNIFIED IDEOGRAPH-9949
        0x4ECA, # G+4038  U+4ECA  uni4ECA,  CJK UNIFIED IDEOGRAPH-4ECA
        0x5997, # G+4039  U+5997  uni5997,  CJK UNIFIED IDEOGRAPH-5997
        0x64D2, # G+4040  U+64D2  uni64D2,  CJK UNIFIED IDEOGRAPH-64D2
        0x6611, # G+4041  U+6611  uni6611,  CJK UNIFIED IDEOGRAPH-6611
        0x6A8E, # G+4042  U+6A8E  uni6A8E,  CJK UNIFIED IDEOGRAPH-6A8E
        0x7434, # G+4043  U+7434  uni7434,  CJK UNIFIED IDEOGRAPH-7434
        0x7981, # G+4044  U+7981  uni7981,  CJK UNIFIED IDEOGRAPH-7981
        0x79BD, # G+4045  U+79BD  uni79BD,  CJK UNIFIED IDEOGRAPH-79BD
        0x82A9, # G+4046  U+82A9  uni82A9,  CJK UNIFIED IDEOGRAPH-82A9
        0x887E, # G+4047  U+887E  uni887E,  CJK UNIFIED IDEOGRAPH-887E
        0x887F, # G+4048  U+887F  uni887F,  CJK UNIFIED IDEOGRAPH-887F
        0x895F, # G+4049  U+895F  uni895F,  CJK UNIFIED IDEOGRAPH-895F
        0x9326, # G+4050  U+9326  uni9326,  CJK UNIFIED IDEOGRAPH-9326
        0x4F0B, # G+4051  U+4F0B  uni4F0B,  CJK UNIFIED IDEOGRAPH-4F0B
        0x53CA, # G+4052  U+53CA  uni53CA,  CJK UNIFIED IDEOGRAPH-53CA
        0x6025, # G+4053  U+6025  uni6025,  CJK UNIFIED IDEOGRAPH-6025
        0x6271, # G+4054  U+6271  uni6271,  CJK UNIFIED IDEOGRAPH-6271
        0x6C72, # G+4055  U+6C72  uni6C72,  CJK UNIFIED IDEOGRAPH-6C72
        0x7D1A, # G+4056  U+7D1A  uni7D1A,  CJK UNIFIED IDEOGRAPH-7D1A
        0x7D66, # G+4057  U+7D66  uni7D66,  CJK UNIFIED IDEOGRAPH-7D66
        0x4E98, # G+4058  U+4E98  uni4E98,  CJK UNIFIED IDEOGRAPH-4E98
        0x5162, # G+4059  U+5162  uni5162,  CJK UNIFIED IDEOGRAPH-5162
        0x77DC, # G+4060  U+77DC  uni77DC,  CJK UNIFIED IDEOGRAPH-77DC
        0x80AF, # G+4061  U+80AF  uni80AF,  CJK UNIFIED IDEOGRAPH-80AF
        0x4F01, # G+4062  U+4F01  uni4F01,  CJK UNIFIED IDEOGRAPH-4F01
        0x4F0E, # G+4063  U+4F0E  uni4F0E,  CJK UNIFIED IDEOGRAPH-4F0E
        0x5176, # G+4064  U+5176  uni5176,  CJK UNIFIED IDEOGRAPH-5176
        0x5180, # G+4065  U+5180  uni5180,  CJK UNIFIED IDEOGRAPH-5180
        0x55DC, # G+4066  U+55DC  uni55DC,  CJK UNIFIED IDEOGRAPH-55DC
        0x5668, # G+4067  U+5668  uni5668,  CJK UNIFIED IDEOGRAPH-5668
        0x573B, # G+4068  U+573B  uni573B,  CJK UNIFIED IDEOGRAPH-573B
        0x57FA, # G+4069  U+57FA  uni57FA,  CJK UNIFIED IDEOGRAPH-57FA
        0x57FC, # G+4070  U+57FC  uni57FC,  CJK UNIFIED IDEOGRAPH-57FC
        0x5914, # G+4071  U+5914  uni5914,  CJK UNIFIED IDEOGRAPH-5914
        0x5947, # G+4072  U+5947  uni5947,  CJK UNIFIED IDEOGRAPH-5947
        0x5993, # G+4073  U+5993  uni5993,  CJK UNIFIED IDEOGRAPH-5993
        0x5BC4, # G+4074  U+5BC4  uni5BC4,  CJK UNIFIED IDEOGRAPH-5BC4
        0x5C90, # G+4075  U+5C90  uni5C90,  CJK UNIFIED IDEOGRAPH-5C90
        0x5D0E, # G+4076  U+5D0E  uni5D0E,  CJK UNIFIED IDEOGRAPH-5D0E
        0x5DF1, # G+4077  U+5DF1  uni5DF1,  CJK UNIFIED IDEOGRAPH-5DF1
        0x5E7E, # G+4078  U+5E7E  uni5E7E,  CJK UNIFIED IDEOGRAPH-5E7E
        0x5FCC, # G+4079  U+5FCC  uni5FCC,  CJK UNIFIED IDEOGRAPH-5FCC
        0x6280, # G+4080  U+6280  uni6280,  CJK UNIFIED IDEOGRAPH-6280
        0x65D7, # G+4081  U+65D7  uni65D7,  CJK UNIFIED IDEOGRAPH-65D7
        0x65E3, # G+4082  U+65E3  uni65E3,  CJK UNIFIED IDEOGRAPH-65E3
        0x671E, # G+4083  U+671E  uni671E,  CJK UNIFIED IDEOGRAPH-671E
        0x671F, # G+4084  U+671F  uni671F,  CJK UNIFIED IDEOGRAPH-671F
        0x675E, # G+4085  U+675E  uni675E,  CJK UNIFIED IDEOGRAPH-675E
        0x68CB, # G+4086  U+68CB  uni68CB,  CJK UNIFIED IDEOGRAPH-68CB
        0x68C4, # G+4087  U+68C4  uni68C4,  CJK UNIFIED IDEOGRAPH-68C4
        0x6A5F, # G+4088  U+6A5F  uni6A5F,  CJK UNIFIED IDEOGRAPH-6A5F
        0x6B3A, # G+4089  U+6B3A  uni6B3A,  CJK UNIFIED IDEOGRAPH-6B3A
        0x6C23, # G+4090  U+6C23  uni6C23,  CJK UNIFIED IDEOGRAPH-6C23
        0x6C7D, # G+4091  U+6C7D  uni6C7D,  CJK UNIFIED IDEOGRAPH-6C7D
        0x6C82, # G+4092  U+6C82  uni6C82,  CJK UNIFIED IDEOGRAPH-6C82
        0x6DC7, # G+4093  U+6DC7  uni6DC7,  CJK UNIFIED IDEOGRAPH-6DC7
        0x7398, # G+4094  U+7398  uni7398,  CJK UNIFIED IDEOGRAPH-7398
        0x7426, # G+4095  U+7426  uni7426,  CJK UNIFIED IDEOGRAPH-7426
        0x742A, # G+4096  U+742A  uni742A,  CJK UNIFIED IDEOGRAPH-742A
        0x7482, # G+4097  U+7482  uni7482,  CJK UNIFIED IDEOGRAPH-7482
        0x74A3, # G+4098  U+74A3  uni74A3,  CJK UNIFIED IDEOGRAPH-74A3
        0x7578, # G+4099  U+7578  uni7578,  CJK UNIFIED IDEOGRAPH-7578
        0x757F, # G+4100  U+757F  uni757F,  CJK UNIFIED IDEOGRAPH-757F
        0x7881, # G+4101  U+7881  uni7881,  CJK UNIFIED IDEOGRAPH-7881
        0x78EF, # G+4102  U+78EF  uni78EF,  CJK UNIFIED IDEOGRAPH-78EF
        0x7941, # G+4103  U+7941  uni7941,  CJK UNIFIED IDEOGRAPH-7941
        0x7947, # G+4104  U+7947  uni7947,  CJK UNIFIED IDEOGRAPH-7947
        0x7948, # G+4105  U+7948  uni7948,  CJK UNIFIED IDEOGRAPH-7948
        0x797A, # G+4106  U+797A  uni797A,  CJK UNIFIED IDEOGRAPH-797A
        0x7B95, # G+4107  U+7B95  uni7B95,  CJK UNIFIED IDEOGRAPH-7B95
        0x7D00, # G+4108  U+7D00  uni7D00,  CJK UNIFIED IDEOGRAPH-7D00
        0x7DBA, # G+4109  U+7DBA  uni7DBA,  CJK UNIFIED IDEOGRAPH-7DBA
        0x7F88, # G+4110  U+7F88  uni7F88,  CJK UNIFIED IDEOGRAPH-7F88
        0x8006, # G+4111  U+8006  uni8006,  CJK UNIFIED IDEOGRAPH-8006
        0x802D, # G+4112  U+802D  uni802D,  CJK UNIFIED IDEOGRAPH-802D
        0x808C, # G+4113  U+808C  uni808C,  CJK UNIFIED IDEOGRAPH-808C
        0x8A18, # G+4114  U+8A18  uni8A18,  CJK UNIFIED IDEOGRAPH-8A18
        0x8B4F, # G+4115  U+8B4F  uni8B4F,  CJK UNIFIED IDEOGRAPH-8B4F
        0xF900, # G+4116  U+F900  uniF900,  CJK COMPATIBILITY IDEOGRAPH-F900
        0x8D77, # G+4117  U+8D77  uni8D77,  CJK UNIFIED IDEOGRAPH-8D77
        0x9321, # G+4118  U+9321  uni9321,  CJK UNIFIED IDEOGRAPH-9321
        0x9324, # G+4119  U+9324  uni9324,  CJK UNIFIED IDEOGRAPH-9324
        0x98E2, # G+4120  U+98E2  uni98E2,  CJK UNIFIED IDEOGRAPH-98E2
        0x9951, # G+4121  U+9951  uni9951,  CJK UNIFIED IDEOGRAPH-9951
        0x9A0E, # G+4122  U+9A0E  uni9A0E,  CJK UNIFIED IDEOGRAPH-9A0E
        0x9A0F, # G+4123  U+9A0F  uni9A0F,  CJK UNIFIED IDEOGRAPH-9A0F
        0x9A65, # G+4124  U+9A65  uni9A65,  CJK UNIFIED IDEOGRAPH-9A65
        0x9E92, # G+4125  U+9E92  uni9E92,  CJK UNIFIED IDEOGRAPH-9E92
        0x7DCA, # G+4126  U+7DCA  uni7DCA,  CJK UNIFIED IDEOGRAPH-7DCA
        0x4F76, # G+4127  U+4F76  uni4F76,  CJK UNIFIED IDEOGRAPH-4F76
        0x5409, # G+4128  U+5409  uni5409,  CJK UNIFIED IDEOGRAPH-5409
        0x62EE, # G+4129  U+62EE  uni62EE,  CJK UNIFIED IDEOGRAPH-62EE
        0x6854, # G+4130  U+6854  uni6854,  CJK UNIFIED IDEOGRAPH-6854
        0xF90A, # G+4131  U+F90A  uniF90A,  CJK COMPATIBILITY IDEOGRAPH-F90A
        0x55AB, # G+4132  U+55AB  uni55AB,  CJK UNIFIED IDEOGRAPH-55AB
        0x513A, # G+4133  U+513A  uni513A,  CJK UNIFIED IDEOGRAPH-513A
        0x5A1C, # G+4134  U+5A1C  uni5A1C,  CJK UNIFIED IDEOGRAPH-5A1C
        0x61E6, # G+4135  U+61E6  uni61E6,  CJK UNIFIED IDEOGRAPH-61E6
        0x62CF, # G+4136  U+62CF  uni62CF,  CJK UNIFIED IDEOGRAPH-62CF
        0x62FF, # G+4137  U+62FF  uni62FF,  CJK UNIFIED IDEOGRAPH-62FF
        0x90A3, # G+4138  U+90A3  uni90A3,  CJK UNIFIED IDEOGRAPH-90A3
        0x8AFE, # G+4139  U+8AFE  uni8AFE,  CJK UNIFIED IDEOGRAPH-8AFE
        0x6696, # G+4140  U+6696  uni6696,  CJK UNIFIED IDEOGRAPH-6696
        0x7156, # G+4141  U+7156  uni7156,  CJK UNIFIED IDEOGRAPH-7156
        0x96E3, # G+4142  U+96E3  uni96E3,  CJK UNIFIED IDEOGRAPH-96E3
        0x634F, # G+4143  U+634F  uni634F,  CJK UNIFIED IDEOGRAPH-634F
        0x637A, # G+4144  U+637A  uni637A,  CJK UNIFIED IDEOGRAPH-637A
        0x5357, # G+4145  U+5357  uni5357,  CJK UNIFIED IDEOGRAPH-5357
        0x678F, # G+4146  U+678F  uni678F,  CJK UNIFIED IDEOGRAPH-678F
        0x6960, # G+4147  U+6960  uni6960,  CJK UNIFIED IDEOGRAPH-6960
        0x6E73, # G+4148  U+6E73  uni6E73,  CJK UNIFIED IDEOGRAPH-6E73
        0x7537, # G+4149  U+7537  uni7537,  CJK UNIFIED IDEOGRAPH-7537
        0x7D0D, # G+4150  U+7D0D  uni7D0D,  CJK UNIFIED IDEOGRAPH-7D0D
        0x8872, # G+4151  U+8872  uni8872,  CJK UNIFIED IDEOGRAPH-8872
        0x56CA, # G+4152  U+56CA  uni56CA,  CJK UNIFIED IDEOGRAPH-56CA
        0x5A18, # G+4153  U+5A18  uni5A18,  CJK UNIFIED IDEOGRAPH-5A18
        0x4E43, # G+4154  U+4E43  uni4E43,  CJK UNIFIED IDEOGRAPH-4E43
        0x5167, # G+4155  U+5167  uni5167,  CJK UNIFIED IDEOGRAPH-5167
        0xF90C, # G+4156  U+F90C  uniF90C,  CJK COMPATIBILITY IDEOGRAPH-F90C
        0x67F0, # G+4157  U+67F0  uni67F0,  CJK UNIFIED IDEOGRAPH-67F0
        0x8010, # G+4158  U+8010  uni8010,  CJK UNIFIED IDEOGRAPH-8010
        0x5973, # G+4159  U+5973  uni5973,  CJK UNIFIED IDEOGRAPH-5973
        0x5E74, # G+4160  U+5E74  uni5E74,  CJK UNIFIED IDEOGRAPH-5E74
        0x649A, # G+4161  U+649A  uni649A,  CJK UNIFIED IDEOGRAPH-649A
        0x79CA, # G+4162  U+79CA  uni79CA,  CJK UNIFIED IDEOGRAPH-79CA
        0x5FF5, # G+4163  U+5FF5  uni5FF5,  CJK UNIFIED IDEOGRAPH-5FF5
        0x606C, # G+4164  U+606C  uni606C,  CJK UNIFIED IDEOGRAPH-606C
        0x62C8, # G+4165  U+62C8  uni62C8,  CJK UNIFIED IDEOGRAPH-62C8
        0x637B, # G+4166  U+637B  uni637B,  CJK UNIFIED IDEOGRAPH-637B
        0x5BE7, # G+4167  U+5BE7  uni5BE7,  CJK UNIFIED IDEOGRAPH-5BE7
        0x5BD7, # G+4168  U+5BD7  uni5BD7,  CJK UNIFIED IDEOGRAPH-5BD7
        0x52AA, # G+4169  U+52AA  uni52AA,  CJK UNIFIED IDEOGRAPH-52AA
        0x5974, # G+4170  U+5974  uni5974,  CJK UNIFIED IDEOGRAPH-5974
        0x5F29, # G+4171  U+5F29  uni5F29,  CJK UNIFIED IDEOGRAPH-5F29
        0x6012, # G+4172  U+6012  uni6012,  CJK UNIFIED IDEOGRAPH-6012
        0x7459, # G+4173  U+7459  uni7459,  CJK UNIFIED IDEOGRAPH-7459
        0x99D1, # G+4174  U+99D1  uni99D1,  CJK UNIFIED IDEOGRAPH-99D1
        0x6FC3, # G+4175  U+6FC3  uni6FC3,  CJK UNIFIED IDEOGRAPH-6FC3
        0x81BF, # G+4176  U+81BF  uni81BF,  CJK UNIFIED IDEOGRAPH-81BF
        0x8FB2, # G+4177  U+8FB2  uni8FB2,  CJK UNIFIED IDEOGRAPH-8FB2
        0x60F1, # G+4178  U+60F1  uni60F1,  CJK UNIFIED IDEOGRAPH-60F1
        0x8166, # G+4179  U+8166  uni8166,  CJK UNIFIED IDEOGRAPH-8166
        0x5C3F, # G+4180  U+5C3F  uni5C3F,  CJK UNIFIED IDEOGRAPH-5C3F
        0x5AE9, # G+4181  U+5AE9  uni5AE9,  CJK UNIFIED IDEOGRAPH-5AE9
        0x8A25, # G+4182  U+8A25  uni8A25,  CJK UNIFIED IDEOGRAPH-8A25
        0x677B, # G+4183  U+677B  uni677B,  CJK UNIFIED IDEOGRAPH-677B
        0x7D10, # G+4184  U+7D10  uni7D10,  CJK UNIFIED IDEOGRAPH-7D10
        0x80FD, # G+4185  U+80FD  uni80FD,  CJK UNIFIED IDEOGRAPH-80FD
        0x5C3C, # G+4186  U+5C3C  uni5C3C,  CJK UNIFIED IDEOGRAPH-5C3C
        0x6CE5, # G+4187  U+6CE5  uni6CE5,  CJK UNIFIED IDEOGRAPH-6CE5
        0x533F, # G+4188  U+533F  uni533F,  CJK UNIFIED IDEOGRAPH-533F
        0x6EBA, # G+4189  U+6EBA  uni6EBA,  CJK UNIFIED IDEOGRAPH-6EBA
        0x591A, # G+4190  U+591A  uni591A,  CJK UNIFIED IDEOGRAPH-591A
        0x8336, # G+4191  U+8336  uni8336,  CJK UNIFIED IDEOGRAPH-8336
        0x4E39, # G+4192  U+4E39  uni4E39,  CJK UNIFIED IDEOGRAPH-4E39
        0x4EB6, # G+4193  U+4EB6  uni4EB6,  CJK UNIFIED IDEOGRAPH-4EB6
        0x4F46, # G+4194  U+4F46  uni4F46,  CJK UNIFIED IDEOGRAPH-4F46
        0x55AE, # G+4195  U+55AE  uni55AE,  CJK UNIFIED IDEOGRAPH-55AE
        0x5718, # G+4196  U+5718  uni5718,  CJK UNIFIED IDEOGRAPH-5718
        0x58C7, # G+4197  U+58C7  uni58C7,  CJK UNIFIED IDEOGRAPH-58C7
        0x5F56, # G+4198  U+5F56  uni5F56,  CJK UNIFIED IDEOGRAPH-5F56
        0x65B7, # G+4199  U+65B7  uni65B7,  CJK UNIFIED IDEOGRAPH-65B7
        0x65E6, # G+4200  U+65E6  uni65E6,  CJK UNIFIED IDEOGRAPH-65E6
        0x6A80, # G+4201  U+6A80  uni6A80,  CJK UNIFIED IDEOGRAPH-6A80
        0x6BB5, # G+4202  U+6BB5  uni6BB5,  CJK UNIFIED IDEOGRAPH-6BB5
        0x6E4D, # G+4203  U+6E4D  uni6E4D,  CJK UNIFIED IDEOGRAPH-6E4D
        0x77ED, # G+4204  U+77ED  uni77ED,  CJK UNIFIED IDEOGRAPH-77ED
        0x7AEF, # G+4205  U+7AEF  uni7AEF,  CJK UNIFIED IDEOGRAPH-7AEF
        0x7C1E, # G+4206  U+7C1E  uni7C1E,  CJK UNIFIED IDEOGRAPH-7C1E
        0x7DDE, # G+4207  U+7DDE  uni7DDE,  CJK UNIFIED IDEOGRAPH-7DDE
        0x86CB, # G+4208  U+86CB  uni86CB,  CJK UNIFIED IDEOGRAPH-86CB
        0x8892, # G+4209  U+8892  uni8892,  CJK UNIFIED IDEOGRAPH-8892
        0x9132, # G+4210  U+9132  uni9132,  CJK UNIFIED IDEOGRAPH-9132
        0x935B, # G+4211  U+935B  uni935B,  CJK UNIFIED IDEOGRAPH-935B
        0x64BB, # G+4212  U+64BB  uni64BB,  CJK UNIFIED IDEOGRAPH-64BB
        0x6FBE, # G+4213  U+6FBE  uni6FBE,  CJK UNIFIED IDEOGRAPH-6FBE
        0x737A, # G+4214  U+737A  uni737A,  CJK UNIFIED IDEOGRAPH-737A
        0x75B8, # G+4215  U+75B8  uni75B8,  CJK UNIFIED IDEOGRAPH-75B8
        0x9054, # G+4216  U+9054  uni9054,  CJK UNIFIED IDEOGRAPH-9054
        0x5556, # G+4217  U+5556  uni5556,  CJK UNIFIED IDEOGRAPH-5556
        0x574D, # G+4218  U+574D  uni574D,  CJK UNIFIED IDEOGRAPH-574D
        0x61BA, # G+4219  U+61BA  uni61BA,  CJK UNIFIED IDEOGRAPH-61BA
        0x64D4, # G+4220  U+64D4  uni64D4,  CJK UNIFIED IDEOGRAPH-64D4
        0x66C7, # G+4221  U+66C7  uni66C7,  CJK UNIFIED IDEOGRAPH-66C7
        0x6DE1, # G+4222  U+6DE1  uni6DE1,  CJK UNIFIED IDEOGRAPH-6DE1
        0x6E5B, # G+4223  U+6E5B  uni6E5B,  CJK UNIFIED IDEOGRAPH-6E5B
        0x6F6D, # G+4224  U+6F6D  uni6F6D,  CJK UNIFIED IDEOGRAPH-6F6D
        0x6FB9, # G+4225  U+6FB9  uni6FB9,  CJK UNIFIED IDEOGRAPH-6FB9
        0x75F0, # G+4226  U+75F0  uni75F0,  CJK UNIFIED IDEOGRAPH-75F0
        0x8043, # G+4227  U+8043  uni8043,  CJK UNIFIED IDEOGRAPH-8043
        0x81BD, # G+4228  U+81BD  uni81BD,  CJK UNIFIED IDEOGRAPH-81BD
        0x8541, # G+4229  U+8541  uni8541,  CJK UNIFIED IDEOGRAPH-8541
        0x8983, # G+4230  U+8983  uni8983,  CJK UNIFIED IDEOGRAPH-8983
        0x8AC7, # G+4231  U+8AC7  uni8AC7,  CJK UNIFIED IDEOGRAPH-8AC7
        0x8B5A, # G+4232  U+8B5A  uni8B5A,  CJK UNIFIED IDEOGRAPH-8B5A
        0x931F, # G+4233  U+931F  uni931F,  CJK UNIFIED IDEOGRAPH-931F
        0x6C93, # G+4234  U+6C93  uni6C93,  CJK UNIFIED IDEOGRAPH-6C93
        0x7553, # G+4235  U+7553  uni7553,  CJK UNIFIED IDEOGRAPH-7553
        0x7B54, # G+4236  U+7B54  uni7B54,  CJK UNIFIED IDEOGRAPH-7B54
        0x8E0F, # G+4237  U+8E0F  uni8E0F,  CJK UNIFIED IDEOGRAPH-8E0F
        0x905D, # G+4238  U+905D  uni905D,  CJK UNIFIED IDEOGRAPH-905D
        0x5510, # G+4239  U+5510  uni5510,  CJK UNIFIED IDEOGRAPH-5510
        0x5802, # G+4240  U+5802  uni5802,  CJK UNIFIED IDEOGRAPH-5802
        0x5858, # G+4241  U+5858  uni5858,  CJK UNIFIED IDEOGRAPH-5858
        0x5E62, # G+4242  U+5E62  uni5E62,  CJK UNIFIED IDEOGRAPH-5E62
        0x6207, # G+4243  U+6207  uni6207,  CJK UNIFIED IDEOGRAPH-6207
        0x649E, # G+4244  U+649E  uni649E,  CJK UNIFIED IDEOGRAPH-649E
        0x68E0, # G+4245  U+68E0  uni68E0,  CJK UNIFIED IDEOGRAPH-68E0
        0x7576, # G+4246  U+7576  uni7576,  CJK UNIFIED IDEOGRAPH-7576
        0x7CD6, # G+4247  U+7CD6  uni7CD6,  CJK UNIFIED IDEOGRAPH-7CD6
        0x87B3, # G+4248  U+87B3  uni87B3,  CJK UNIFIED IDEOGRAPH-87B3
        0x9EE8, # G+4249  U+9EE8  uni9EE8,  CJK UNIFIED IDEOGRAPH-9EE8
        0x4EE3, # G+4250  U+4EE3  uni4EE3,  CJK UNIFIED IDEOGRAPH-4EE3
        0x5788, # G+4251  U+5788  uni5788,  CJK UNIFIED IDEOGRAPH-5788
        0x576E, # G+4252  U+576E  uni576E,  CJK UNIFIED IDEOGRAPH-576E
        0x5927, # G+4253  U+5927  uni5927,  CJK UNIFIED IDEOGRAPH-5927
        0x5C0D, # G+4254  U+5C0D  uni5C0D,  CJK UNIFIED IDEOGRAPH-5C0D
        0x5CB1, # G+4255  U+5CB1  uni5CB1,  CJK UNIFIED IDEOGRAPH-5CB1
        0x5E36, # G+4256  U+5E36  uni5E36,  CJK UNIFIED IDEOGRAPH-5E36
        0x5F85, # G+4257  U+5F85  uni5F85,  CJK UNIFIED IDEOGRAPH-5F85
        0x6234, # G+4258  U+6234  uni6234,  CJK UNIFIED IDEOGRAPH-6234
        0x64E1, # G+4259  U+64E1  uni64E1,  CJK UNIFIED IDEOGRAPH-64E1
        0x73B3, # G+4260  U+73B3  uni73B3,  CJK UNIFIED IDEOGRAPH-73B3
        0x81FA, # G+4261  U+81FA  uni81FA,  CJK UNIFIED IDEOGRAPH-81FA
        0x888B, # G+4262  U+888B  uni888B,  CJK UNIFIED IDEOGRAPH-888B
        0x8CB8, # G+4263  U+8CB8  uni8CB8,  CJK UNIFIED IDEOGRAPH-8CB8
        0x968A, # G+4264  U+968A  uni968A,  CJK UNIFIED IDEOGRAPH-968A
        0x9EDB, # G+4265  U+9EDB  uni9EDB,  CJK UNIFIED IDEOGRAPH-9EDB
        0x5B85, # G+4266  U+5B85  uni5B85,  CJK UNIFIED IDEOGRAPH-5B85
        0x5FB7, # G+4267  U+5FB7  uni5FB7,  CJK UNIFIED IDEOGRAPH-5FB7
        0x60B3, # G+4268  U+60B3  uni60B3,  CJK UNIFIED IDEOGRAPH-60B3
        0x5012, # G+4269  U+5012  uni5012,  CJK UNIFIED IDEOGRAPH-5012
        0x5200, # G+4270  U+5200  uni5200,  CJK UNIFIED IDEOGRAPH-5200
        0x5230, # G+4271  U+5230  uni5230,  CJK UNIFIED IDEOGRAPH-5230
        0x5716, # G+4272  U+5716  uni5716,  CJK UNIFIED IDEOGRAPH-5716
        0x5835, # G+4273  U+5835  uni5835,  CJK UNIFIED IDEOGRAPH-5835
        0x5857, # G+4274  U+5857  uni5857,  CJK UNIFIED IDEOGRAPH-5857
        0x5C0E, # G+4275  U+5C0E  uni5C0E,  CJK UNIFIED IDEOGRAPH-5C0E
        0x5C60, # G+4276  U+5C60  uni5C60,  CJK UNIFIED IDEOGRAPH-5C60
        0x5CF6, # G+4277  U+5CF6  uni5CF6,  CJK UNIFIED IDEOGRAPH-5CF6
        0x5D8B, # G+4278  U+5D8B  uni5D8B,  CJK UNIFIED IDEOGRAPH-5D8B
        0x5EA6, # G+4279  U+5EA6  uni5EA6,  CJK UNIFIED IDEOGRAPH-5EA6
        0x5F92, # G+4280  U+5F92  uni5F92,  CJK UNIFIED IDEOGRAPH-5F92
        0x60BC, # G+4281  U+60BC  uni60BC,  CJK UNIFIED IDEOGRAPH-60BC
        0x6311, # G+4282  U+6311  uni6311,  CJK UNIFIED IDEOGRAPH-6311
        0x6389, # G+4283  U+6389  uni6389,  CJK UNIFIED IDEOGRAPH-6389
        0x6417, # G+4284  U+6417  uni6417,  CJK UNIFIED IDEOGRAPH-6417
        0x6843, # G+4285  U+6843  uni6843,  CJK UNIFIED IDEOGRAPH-6843
        0x68F9, # G+4286  U+68F9  uni68F9,  CJK UNIFIED IDEOGRAPH-68F9
        0x6AC2, # G+4287  U+6AC2  uni6AC2,  CJK UNIFIED IDEOGRAPH-6AC2
        0x6DD8, # G+4288  U+6DD8  uni6DD8,  CJK UNIFIED IDEOGRAPH-6DD8
        0x6E21, # G+4289  U+6E21  uni6E21,  CJK UNIFIED IDEOGRAPH-6E21
        0x6ED4, # G+4290  U+6ED4  uni6ED4,  CJK UNIFIED IDEOGRAPH-6ED4
        0x6FE4, # G+4291  U+6FE4  uni6FE4,  CJK UNIFIED IDEOGRAPH-6FE4
        0x71FE, # G+4292  U+71FE  uni71FE,  CJK UNIFIED IDEOGRAPH-71FE
        0x76DC, # G+4293  U+76DC  uni76DC,  CJK UNIFIED IDEOGRAPH-76DC
        0x7779, # G+4294  U+7779  uni7779,  CJK UNIFIED IDEOGRAPH-7779
        0x79B1, # G+4295  U+79B1  uni79B1,  CJK UNIFIED IDEOGRAPH-79B1
        0x7A3B, # G+4296  U+7A3B  uni7A3B,  CJK UNIFIED IDEOGRAPH-7A3B
        0x8404, # G+4297  U+8404  uni8404,  CJK UNIFIED IDEOGRAPH-8404
        0x89A9, # G+4298  U+89A9  uni89A9,  CJK UNIFIED IDEOGRAPH-89A9
        0x8CED, # G+4299  U+8CED  uni8CED,  CJK UNIFIED IDEOGRAPH-8CED
        0x8DF3, # G+4300  U+8DF3  uni8DF3,  CJK UNIFIED IDEOGRAPH-8DF3
        0x8E48, # G+4301  U+8E48  uni8E48,  CJK UNIFIED IDEOGRAPH-8E48
        0x9003, # G+4302  U+9003  uni9003,  CJK UNIFIED IDEOGRAPH-9003
        0x9014, # G+4303  U+9014  uni9014,  CJK UNIFIED IDEOGRAPH-9014
        0x9053, # G+4304  U+9053  uni9053,  CJK UNIFIED IDEOGRAPH-9053
        0x90FD, # G+4305  U+90FD  uni90FD,  CJK UNIFIED IDEOGRAPH-90FD
        0x934D, # G+4306  U+934D  uni934D,  CJK UNIFIED IDEOGRAPH-934D
        0x9676, # G+4307  U+9676  uni9676,  CJK UNIFIED IDEOGRAPH-9676
        0x97DC, # G+4308  U+97DC  uni97DC,  CJK UNIFIED IDEOGRAPH-97DC
        0x6BD2, # G+4309  U+6BD2  uni6BD2,  CJK UNIFIED IDEOGRAPH-6BD2
        0x7006, # G+4310  U+7006  uni7006,  CJK UNIFIED IDEOGRAPH-7006
        0x7258, # G+4311  U+7258  uni7258,  CJK UNIFIED IDEOGRAPH-7258
        0x72A2, # G+4312  U+72A2  uni72A2,  CJK UNIFIED IDEOGRAPH-72A2
        0x7368, # G+4313  U+7368  uni7368,  CJK UNIFIED IDEOGRAPH-7368
        0x7763, # G+4314  U+7763  uni7763,  CJK UNIFIED IDEOGRAPH-7763
        0x79BF, # G+4315  U+79BF  uni79BF,  CJK UNIFIED IDEOGRAPH-79BF
        0x7BE4, # G+4316  U+7BE4  uni7BE4,  CJK UNIFIED IDEOGRAPH-7BE4
        0x7E9B, # G+4317  U+7E9B  uni7E9B,  CJK UNIFIED IDEOGRAPH-7E9B
        0x8B80, # G+4318  U+8B80  uni8B80,  CJK UNIFIED IDEOGRAPH-8B80
        0x58A9, # G+4319  U+58A9  uni58A9,  CJK UNIFIED IDEOGRAPH-58A9
        0x60C7, # G+4320  U+60C7  uni60C7,  CJK UNIFIED IDEOGRAPH-60C7
        0x6566, # G+4321  U+6566  uni6566,  CJK UNIFIED IDEOGRAPH-6566
        0x65FD, # G+4322  U+65FD  uni65FD,  CJK UNIFIED IDEOGRAPH-65FD
        0x66BE, # G+4323  U+66BE  uni66BE,  CJK UNIFIED IDEOGRAPH-66BE
        0x6C8C, # G+4324  U+6C8C  uni6C8C,  CJK UNIFIED IDEOGRAPH-6C8C
        0x711E, # G+4325  U+711E  uni711E,  CJK UNIFIED IDEOGRAPH-711E
        0x71C9, # G+4326  U+71C9  uni71C9,  CJK UNIFIED IDEOGRAPH-71C9
        0x8C5A, # G+4327  U+8C5A  uni8C5A,  CJK UNIFIED IDEOGRAPH-8C5A
        0x9813, # G+4328  U+9813  uni9813,  CJK UNIFIED IDEOGRAPH-9813
        0x4E6D, # G+4329  U+4E6D  uni4E6D,  CJK UNIFIED IDEOGRAPH-4E6D
        0x7A81, # G+4330  U+7A81  uni7A81,  CJK UNIFIED IDEOGRAPH-7A81
        0x4EDD, # G+4331  U+4EDD  uni4EDD,  CJK UNIFIED IDEOGRAPH-4EDD
        0x51AC, # G+4332  U+51AC  uni51AC,  CJK UNIFIED IDEOGRAPH-51AC
        0x51CD, # G+4333  U+51CD  uni51CD,  CJK UNIFIED IDEOGRAPH-51CD
        0x52D5, # G+4334  U+52D5  uni52D5,  CJK UNIFIED IDEOGRAPH-52D5
        0x540C, # G+4335  U+540C  uni540C,  CJK UNIFIED IDEOGRAPH-540C
        0x61A7, # G+4336  U+61A7  uni61A7,  CJK UNIFIED IDEOGRAPH-61A7
        0x6771, # G+4337  U+6771  uni6771,  CJK UNIFIED IDEOGRAPH-6771
        0x6850, # G+4338  U+6850  uni6850,  CJK UNIFIED IDEOGRAPH-6850
        0x68DF, # G+4339  U+68DF  uni68DF,  CJK UNIFIED IDEOGRAPH-68DF
        0x6D1E, # G+4340  U+6D1E  uni6D1E,  CJK UNIFIED IDEOGRAPH-6D1E
        0x6F7C, # G+4341  U+6F7C  uni6F7C,  CJK UNIFIED IDEOGRAPH-6F7C
        0x75BC, # G+4342  U+75BC  uni75BC,  CJK UNIFIED IDEOGRAPH-75BC
        0x77B3, # G+4343  U+77B3  uni77B3,  CJK UNIFIED IDEOGRAPH-77B3
        0x7AE5, # G+4344  U+7AE5  uni7AE5,  CJK UNIFIED IDEOGRAPH-7AE5
        0x80F4, # G+4345  U+80F4  uni80F4,  CJK UNIFIED IDEOGRAPH-80F4
        0x8463, # G+4346  U+8463  uni8463,  CJK UNIFIED IDEOGRAPH-8463
        0x9285, # G+4347  U+9285  uni9285,  CJK UNIFIED IDEOGRAPH-9285
        0x515C, # G+4348  U+515C  uni515C,  CJK UNIFIED IDEOGRAPH-515C
        0x6597, # G+4349  U+6597  uni6597,  CJK UNIFIED IDEOGRAPH-6597
        0x675C, # G+4350  U+675C  uni675C,  CJK UNIFIED IDEOGRAPH-675C
        0x6793, # G+4351  U+6793  uni6793,  CJK UNIFIED IDEOGRAPH-6793
        0x75D8, # G+4352  U+75D8  uni75D8,  CJK UNIFIED IDEOGRAPH-75D8
        0x7AC7, # G+4353  U+7AC7  uni7AC7,  CJK UNIFIED IDEOGRAPH-7AC7
        0x8373, # G+4354  U+8373  uni8373,  CJK UNIFIED IDEOGRAPH-8373
        0x8C46, # G+4355  U+8C46  uni8C46,  CJK UNIFIED IDEOGRAPH-8C46
        0x9017, # G+4356  U+9017  uni9017,  CJK UNIFIED IDEOGRAPH-9017
        0x982D, # G+4357  U+982D  uni982D,  CJK UNIFIED IDEOGRAPH-982D
        0x5C6F, # G+4358  U+5C6F  uni5C6F,  CJK UNIFIED IDEOGRAPH-5C6F
        0x81C0, # G+4359  U+81C0  uni81C0,  CJK UNIFIED IDEOGRAPH-81C0
        0x829A, # G+4360  U+829A  uni829A,  CJK UNIFIED IDEOGRAPH-829A
        0x9041, # G+4361  U+9041  uni9041,  CJK UNIFIED IDEOGRAPH-9041
        0x906F, # G+4362  U+906F  uni906F,  CJK UNIFIED IDEOGRAPH-906F
        0x920D, # G+4363  U+920D  uni920D,  CJK UNIFIED IDEOGRAPH-920D
        0x5F97, # G+4364  U+5F97  uni5F97,  CJK UNIFIED IDEOGRAPH-5F97
        0x5D9D, # G+4365  U+5D9D  uni5D9D,  CJK UNIFIED IDEOGRAPH-5D9D
        0x6A59, # G+4366  U+6A59  uni6A59,  CJK UNIFIED IDEOGRAPH-6A59
        0x71C8, # G+4367  U+71C8  uni71C8,  CJK UNIFIED IDEOGRAPH-71C8
        0x767B, # G+4368  U+767B  uni767B,  CJK UNIFIED IDEOGRAPH-767B
        0x7B49, # G+4369  U+7B49  uni7B49,  CJK UNIFIED IDEOGRAPH-7B49
        0x85E4, # G+4370  U+85E4  uni85E4,  CJK UNIFIED IDEOGRAPH-85E4
        0x8B04, # G+4371  U+8B04  uni8B04,  CJK UNIFIED IDEOGRAPH-8B04
        0x9127, # G+4372  U+9127  uni9127,  CJK UNIFIED IDEOGRAPH-9127
        0x9A30, # G+4373  U+9A30  uni9A30,  CJK UNIFIED IDEOGRAPH-9A30
        0xF90B, # G+4374  U+F90B  uniF90B,  CJK COMPATIBILITY IDEOGRAPH-F90B
        0xF90D, # G+4375  U+F90D  uniF90D,  CJK COMPATIBILITY IDEOGRAPH-F90D
        0xF90E, # G+4376  U+F90E  uniF90E,  CJK COMPATIBILITY IDEOGRAPH-F90E
        0xF90F, # G+4377  U+F90F  uniF90F,  CJK COMPATIBILITY IDEOGRAPH-F90F
        0xF910, # G+4378  U+F910  uniF910,  CJK COMPATIBILITY IDEOGRAPH-F910
        0xF911, # G+4379  U+F911  uniF911,  CJK COMPATIBILITY IDEOGRAPH-F911
        0xF912, # G+4380  U+F912  uniF912,  CJK COMPATIBILITY IDEOGRAPH-F912
        0xF913, # G+4381  U+F913  uniF913,  CJK COMPATIBILITY IDEOGRAPH-F913
        0xF915, # G+4382  U+F915  uniF915,  CJK COMPATIBILITY IDEOGRAPH-F915
        0xF916, # G+4383  U+F916  uniF916,  CJK COMPATIBILITY IDEOGRAPH-F916
        0xF917, # G+4384  U+F917  uniF917,  CJK COMPATIBILITY IDEOGRAPH-F917
        0x7D61, # G+4385  U+7D61  uni7D61,  CJK UNIFIED IDEOGRAPH-7D61
        0xF918, # G+4386  U+F918  uniF918,  CJK COMPATIBILITY IDEOGRAPH-F918
        0xF919, # G+4387  U+F919  uniF919,  CJK COMPATIBILITY IDEOGRAPH-F919
        0xF91A, # G+4388  U+F91A  uniF91A,  CJK COMPATIBILITY IDEOGRAPH-F91A
        0xF91B, # G+4389  U+F91B  uniF91B,  CJK COMPATIBILITY IDEOGRAPH-F91B
        0xF91C, # G+4390  U+F91C  uniF91C,  CJK COMPATIBILITY IDEOGRAPH-F91C
        0xF91D, # G+4391  U+F91D  uniF91D,  CJK COMPATIBILITY IDEOGRAPH-F91D
        0x6B12, # G+4392  U+6B12  uni6B12,  CJK UNIFIED IDEOGRAPH-6B12
        0x703E, # G+4393  U+703E  uni703E,  CJK UNIFIED IDEOGRAPH-703E
        0xF91E, # G+4394  U+F91E  uniF91E,  CJK COMPATIBILITY IDEOGRAPH-F91E
        0xF91F, # G+4395  U+F91F  uniF91F,  CJK COMPATIBILITY IDEOGRAPH-F91F
        0xF920, # G+4396  U+F920  uniF920,  CJK COMPATIBILITY IDEOGRAPH-F920
        0x524C, # G+4397  U+524C  uni524C,  CJK UNIFIED IDEOGRAPH-524C
        0x8FA3, # G+4398  U+8FA3  uni8FA3,  CJK UNIFIED IDEOGRAPH-8FA3
        0xF921, # G+4399  U+F921  uniF921,  CJK COMPATIBILITY IDEOGRAPH-F921
        0x64E5, # G+4400  U+64E5  uni64E5,  CJK UNIFIED IDEOGRAPH-64E5
        0x652C, # G+4401  U+652C  uni652C,  CJK UNIFIED IDEOGRAPH-652C
        0x6B16, # G+4402  U+6B16  uni6B16,  CJK UNIFIED IDEOGRAPH-6B16
        0xF922, # G+4403  U+F922  uniF922,  CJK COMPATIBILITY IDEOGRAPH-F922
        0x7C43, # G+4404  U+7C43  uni7C43,  CJK UNIFIED IDEOGRAPH-7C43
        0x7E9C, # G+4405  U+7E9C  uni7E9C,  CJK UNIFIED IDEOGRAPH-7E9C
        0xF923, # G+4406  U+F923  uniF923,  CJK COMPATIBILITY IDEOGRAPH-F923
        0xF924, # G+4407  U+F924  uniF924,  CJK COMPATIBILITY IDEOGRAPH-F924
        0x89BD, # G+4408  U+89BD  uni89BD,  CJK UNIFIED IDEOGRAPH-89BD
        0xF925, # G+4409  U+F925  uniF925,  CJK COMPATIBILITY IDEOGRAPH-F925
        0xF926, # G+4410  U+F926  uniF926,  CJK COMPATIBILITY IDEOGRAPH-F926
        0xF927, # G+4411  U+F927  uniF927,  CJK COMPATIBILITY IDEOGRAPH-F927
        0xF928, # G+4412  U+F928  uniF928,  CJK COMPATIBILITY IDEOGRAPH-F928
        0xF929, # G+4413  U+F929  uniF929,  CJK COMPATIBILITY IDEOGRAPH-F929
        0xF92A, # G+4414  U+F92A  uniF92A,  CJK COMPATIBILITY IDEOGRAPH-F92A
        0xF92B, # G+4415  U+F92B  uniF92B,  CJK COMPATIBILITY IDEOGRAPH-F92B
        0x7405, # G+4416  U+7405  uni7405,  CJK UNIFIED IDEOGRAPH-7405
        0x746F, # G+4417  U+746F  uni746F,  CJK UNIFIED IDEOGRAPH-746F
        0x8782, # G+4418  U+8782  uni8782,  CJK UNIFIED IDEOGRAPH-8782
        0xF92C, # G+4419  U+F92C  uniF92C,  CJK COMPATIBILITY IDEOGRAPH-F92C
        0xF92D, # G+4420  U+F92D  uniF92D,  CJK COMPATIBILITY IDEOGRAPH-F92D
        0x5D0D, # G+4421  U+5D0D  uni5D0D,  CJK UNIFIED IDEOGRAPH-5D0D
        0x5FA0, # G+4422  U+5FA0  uni5FA0,  CJK UNIFIED IDEOGRAPH-5FA0
        0x840A, # G+4423  U+840A  uni840A,  CJK UNIFIED IDEOGRAPH-840A
        0xF92E, # G+4424  U+F92E  uniF92E,  CJK COMPATIBILITY IDEOGRAPH-F92E
        0x63A0, # G+4425  U+63A0  uni63A0,  CJK UNIFIED IDEOGRAPH-63A0
        0x7565, # G+4426  U+7565  uni7565,  CJK UNIFIED IDEOGRAPH-7565
        0x4EAE, # G+4427  U+4EAE  uni4EAE,  CJK UNIFIED IDEOGRAPH-4EAE
        0x5006, # G+4428  U+5006  uni5006,  CJK UNIFIED IDEOGRAPH-5006
        0x5169, # G+4429  U+5169  uni5169,  CJK UNIFIED IDEOGRAPH-5169
        0x51C9, # G+4430  U+51C9  uni51C9,  CJK UNIFIED IDEOGRAPH-51C9
        0x6881, # G+4431  U+6881  uni6881,  CJK UNIFIED IDEOGRAPH-6881
        0x6A11, # G+4432  U+6A11  uni6A11,  CJK UNIFIED IDEOGRAPH-6A11
        0x7CAE, # G+4433  U+7CAE  uni7CAE,  CJK UNIFIED IDEOGRAPH-7CAE
        0x7CB1, # G+4434  U+7CB1  uni7CB1,  CJK UNIFIED IDEOGRAPH-7CB1
        0x7CE7, # G+4435  U+7CE7  uni7CE7,  CJK UNIFIED IDEOGRAPH-7CE7
        0x826F, # G+4436  U+826F  uni826F,  CJK UNIFIED IDEOGRAPH-826F
        0x8AD2, # G+4437  U+8AD2  uni8AD2,  CJK UNIFIED IDEOGRAPH-8AD2
        0x8F1B, # G+4438  U+8F1B  uni8F1B,  CJK UNIFIED IDEOGRAPH-8F1B
        0x91CF, # G+4439  U+91CF  uni91CF,  CJK UNIFIED IDEOGRAPH-91CF
        0x4FB6, # G+4440  U+4FB6  uni4FB6,  CJK UNIFIED IDEOGRAPH-4FB6
        0x5137, # G+4441  U+5137  uni5137,  CJK UNIFIED IDEOGRAPH-5137
        0x52F5, # G+4442  U+52F5  uni52F5,  CJK UNIFIED IDEOGRAPH-52F5
        0x5442, # G+4443  U+5442  uni5442,  CJK UNIFIED IDEOGRAPH-5442
        0x5EEC, # G+4444  U+5EEC  uni5EEC,  CJK UNIFIED IDEOGRAPH-5EEC
        0x616E, # G+4445  U+616E  uni616E,  CJK UNIFIED IDEOGRAPH-616E
        0x623E, # G+4446  U+623E  uni623E,  CJK UNIFIED IDEOGRAPH-623E
        0x65C5, # G+4447  U+65C5  uni65C5,  CJK UNIFIED IDEOGRAPH-65C5
        0x6ADA, # G+4448  U+6ADA  uni6ADA,  CJK UNIFIED IDEOGRAPH-6ADA
        0x6FFE, # G+4449  U+6FFE  uni6FFE,  CJK UNIFIED IDEOGRAPH-6FFE
        0x792A, # G+4450  U+792A  uni792A,  CJK UNIFIED IDEOGRAPH-792A
        0x85DC, # G+4451  U+85DC  uni85DC,  CJK UNIFIED IDEOGRAPH-85DC
        0x8823, # G+4452  U+8823  uni8823,  CJK UNIFIED IDEOGRAPH-8823
        0x95AD, # G+4453  U+95AD  uni95AD,  CJK UNIFIED IDEOGRAPH-95AD
        0x9A62, # G+4454  U+9A62  uni9A62,  CJK UNIFIED IDEOGRAPH-9A62
        0x9A6A, # G+4455  U+9A6A  uni9A6A,  CJK UNIFIED IDEOGRAPH-9A6A
        0x9E97, # G+4456  U+9E97  uni9E97,  CJK UNIFIED IDEOGRAPH-9E97
        0x9ECE, # G+4457  U+9ECE  uni9ECE,  CJK UNIFIED IDEOGRAPH-9ECE
        0x529B, # G+4458  U+529B  uni529B,  CJK UNIFIED IDEOGRAPH-529B
        0x66C6, # G+4459  U+66C6  uni66C6,  CJK UNIFIED IDEOGRAPH-66C6
        0x6B77, # G+4460  U+6B77  uni6B77,  CJK UNIFIED IDEOGRAPH-6B77
        0x701D, # G+4461  U+701D  uni701D,  CJK UNIFIED IDEOGRAPH-701D
        0x792B, # G+4462  U+792B  uni792B,  CJK UNIFIED IDEOGRAPH-792B
        0x8F62, # G+4463  U+8F62  uni8F62,  CJK UNIFIED IDEOGRAPH-8F62
        0x9742, # G+4464  U+9742  uni9742,  CJK UNIFIED IDEOGRAPH-9742
        0x6190, # G+4465  U+6190  uni6190,  CJK UNIFIED IDEOGRAPH-6190
        0x6200, # G+4466  U+6200  uni6200,  CJK UNIFIED IDEOGRAPH-6200
        0x6523, # G+4467  U+6523  uni6523,  CJK UNIFIED IDEOGRAPH-6523
        0x6F23, # G+4468  U+6F23  uni6F23,  CJK UNIFIED IDEOGRAPH-6F23
        0x7149, # G+4469  U+7149  uni7149,  CJK UNIFIED IDEOGRAPH-7149
        0x7489, # G+4470  U+7489  uni7489,  CJK UNIFIED IDEOGRAPH-7489
        0x7DF4, # G+4471  U+7DF4  uni7DF4,  CJK UNIFIED IDEOGRAPH-7DF4
        0x806F, # G+4472  U+806F  uni806F,  CJK UNIFIED IDEOGRAPH-806F
        0x84EE, # G+4473  U+84EE  uni84EE,  CJK UNIFIED IDEOGRAPH-84EE
        0x8F26, # G+4474  U+8F26  uni8F26,  CJK UNIFIED IDEOGRAPH-8F26
        0x9023, # G+4475  U+9023  uni9023,  CJK UNIFIED IDEOGRAPH-9023
        0x934A, # G+4476  U+934A  uni934A,  CJK UNIFIED IDEOGRAPH-934A
        0x51BD, # G+4477  U+51BD  uni51BD,  CJK UNIFIED IDEOGRAPH-51BD
        0x5217, # G+4478  U+5217  uni5217,  CJK UNIFIED IDEOGRAPH-5217
        0x52A3, # G+4479  U+52A3  uni52A3,  CJK UNIFIED IDEOGRAPH-52A3
        0x6D0C, # G+4480  U+6D0C  uni6D0C,  CJK UNIFIED IDEOGRAPH-6D0C
        0x70C8, # G+4481  U+70C8  uni70C8,  CJK UNIFIED IDEOGRAPH-70C8
        0x88C2, # G+4482  U+88C2  uni88C2,  CJK UNIFIED IDEOGRAPH-88C2
        0x5EC9, # G+4483  U+5EC9  uni5EC9,  CJK UNIFIED IDEOGRAPH-5EC9
        0x6582, # G+4484  U+6582  uni6582,  CJK UNIFIED IDEOGRAPH-6582
        0x6BAE, # G+4485  U+6BAE  uni6BAE,  CJK UNIFIED IDEOGRAPH-6BAE
        0x6FC2, # G+4486  U+6FC2  uni6FC2,  CJK UNIFIED IDEOGRAPH-6FC2
        0x7C3E, # G+4487  U+7C3E  uni7C3E,  CJK UNIFIED IDEOGRAPH-7C3E
        0x7375, # G+4488  U+7375  uni7375,  CJK UNIFIED IDEOGRAPH-7375
        0x4EE4, # G+4489  U+4EE4  uni4EE4,  CJK UNIFIED IDEOGRAPH-4EE4
        0x4F36, # G+4490  U+4F36  uni4F36,  CJK UNIFIED IDEOGRAPH-4F36
        0x56F9, # G+4491  U+56F9  uni56F9,  CJK UNIFIED IDEOGRAPH-56F9
        0x5CBA, # G+4492  U+5CBA  uni5CBA,  CJK UNIFIED IDEOGRAPH-5CBA
        0x5DBA, # G+4493  U+5DBA  uni5DBA,  CJK UNIFIED IDEOGRAPH-5DBA
        0x601C, # G+4494  U+601C  uni601C,  CJK UNIFIED IDEOGRAPH-601C
        0x73B2, # G+4495  U+73B2  uni73B2,  CJK UNIFIED IDEOGRAPH-73B2
        0x7B2D, # G+4496  U+7B2D  uni7B2D,  CJK UNIFIED IDEOGRAPH-7B2D
        0x7F9A, # G+4497  U+7F9A  uni7F9A,  CJK UNIFIED IDEOGRAPH-7F9A
        0x7FCE, # G+4498  U+7FCE  uni7FCE,  CJK UNIFIED IDEOGRAPH-7FCE
        0x8046, # G+4499  U+8046  uni8046,  CJK UNIFIED IDEOGRAPH-8046
        0x901E, # G+4500  U+901E  uni901E,  CJK UNIFIED IDEOGRAPH-901E
        0x9234, # G+4501  U+9234  uni9234,  CJK UNIFIED IDEOGRAPH-9234
        0x96F6, # G+4502  U+96F6  uni96F6,  CJK UNIFIED IDEOGRAPH-96F6
        0x9748, # G+4503  U+9748  uni9748,  CJK UNIFIED IDEOGRAPH-9748
        0x9818, # G+4504  U+9818  uni9818,  CJK UNIFIED IDEOGRAPH-9818
        0x9F61, # G+4505  U+9F61  uni9F61,  CJK UNIFIED IDEOGRAPH-9F61
        0x4F8B, # G+4506  U+4F8B  uni4F8B,  CJK UNIFIED IDEOGRAPH-4F8B
        0x6FA7, # G+4507  U+6FA7  uni6FA7,  CJK UNIFIED IDEOGRAPH-6FA7
        0x79AE, # G+4508  U+79AE  uni79AE,  CJK UNIFIED IDEOGRAPH-79AE
        0x91B4, # G+4509  U+91B4  uni91B4,  CJK UNIFIED IDEOGRAPH-91B4
        0x96B7, # G+4510  U+96B7  uni96B7,  CJK UNIFIED IDEOGRAPH-96B7
        0xF92F, # G+4511  U+F92F  uniF92F,  CJK COMPATIBILITY IDEOGRAPH-F92F
        0x6488, # G+4512  U+6488  uni6488,  CJK UNIFIED IDEOGRAPH-6488
        0xF930, # G+4513  U+F930  uniF930,  CJK COMPATIBILITY IDEOGRAPH-F930
        0xF931, # G+4514  U+F931  uniF931,  CJK COMPATIBILITY IDEOGRAPH-F931
        0x6F5E, # G+4515  U+6F5E  uni6F5E,  CJK UNIFIED IDEOGRAPH-6F5E
        0x7018, # G+4516  U+7018  uni7018,  CJK UNIFIED IDEOGRAPH-7018
        0xF932, # G+4517  U+F932  uniF932,  CJK COMPATIBILITY IDEOGRAPH-F932
        0xF933, # G+4518  U+F933  uniF933,  CJK COMPATIBILITY IDEOGRAPH-F933
        0xF934, # G+4519  U+F934  uniF934,  CJK COMPATIBILITY IDEOGRAPH-F934
        0xF935, # G+4520  U+F935  uniF935,  CJK COMPATIBILITY IDEOGRAPH-F935
        0xF936, # G+4521  U+F936  uniF936,  CJK COMPATIBILITY IDEOGRAPH-F936
        0xF937, # G+4522  U+F937  uniF937,  CJK COMPATIBILITY IDEOGRAPH-F937
        0x8F05, # G+4523  U+8F05  uni8F05,  CJK UNIFIED IDEOGRAPH-8F05
        0xF938, # G+4524  U+F938  uniF938,  CJK COMPATIBILITY IDEOGRAPH-F938
        0xF939, # G+4525  U+F939  uniF939,  CJK COMPATIBILITY IDEOGRAPH-F939
        0xF93A, # G+4526  U+F93A  uniF93A,  CJK COMPATIBILITY IDEOGRAPH-F93A
        0x9E75, # G+4527  U+9E75  uni9E75,  CJK UNIFIED IDEOGRAPH-9E75
        0xF93B, # G+4528  U+F93B  uniF93B,  CJK COMPATIBILITY IDEOGRAPH-F93B
        0xF93C, # G+4529  U+F93C  uniF93C,  CJK COMPATIBILITY IDEOGRAPH-F93C
        0xF93D, # G+4530  U+F93D  uniF93D,  CJK COMPATIBILITY IDEOGRAPH-F93D
        0xF93E, # G+4531  U+F93E  uniF93E,  CJK COMPATIBILITY IDEOGRAPH-F93E
        0xF93F, # G+4532  U+F93F  uniF93F,  CJK COMPATIBILITY IDEOGRAPH-F93F
        0xF940, # G+4533  U+F940  uniF940,  CJK COMPATIBILITY IDEOGRAPH-F940
        0x9E93, # G+4534  U+9E93  uni9E93,  CJK UNIFIED IDEOGRAPH-9E93
        0xF941, # G+4535  U+F941  uniF941,  CJK COMPATIBILITY IDEOGRAPH-F941
        0xF942, # G+4536  U+F942  uniF942,  CJK COMPATIBILITY IDEOGRAPH-F942
        0xF943, # G+4537  U+F943  uniF943,  CJK COMPATIBILITY IDEOGRAPH-F943
        0x6727, # G+4538  U+6727  uni6727,  CJK UNIFIED IDEOGRAPH-6727
        0x7027, # G+4539  U+7027  uni7027,  CJK UNIFIED IDEOGRAPH-7027
        0x74CF, # G+4540  U+74CF  uni74CF,  CJK UNIFIED IDEOGRAPH-74CF
        0xF944, # G+4541  U+F944  uniF944,  CJK COMPATIBILITY IDEOGRAPH-F944
        0xF945, # G+4542  U+F945  uniF945,  CJK COMPATIBILITY IDEOGRAPH-F945
        0x5121, # G+4543  U+5121  uni5121,  CJK UNIFIED IDEOGRAPH-5121
        0x7028, # G+4544  U+7028  uni7028,  CJK UNIFIED IDEOGRAPH-7028
        0xF946, # G+4545  U+F946  uniF946,  CJK COMPATIBILITY IDEOGRAPH-F946
        0xF947, # G+4546  U+F947  uniF947,  CJK COMPATIBILITY IDEOGRAPH-F947
        0xF948, # G+4547  U+F948  uniF948,  CJK COMPATIBILITY IDEOGRAPH-F948
        0x8CDA, # G+4548  U+8CDA  uni8CDA,  CJK UNIFIED IDEOGRAPH-8CDA
        0x8CF4, # G+4549  U+8CF4  uni8CF4,  CJK UNIFIED IDEOGRAPH-8CF4
        0xF949, # G+4550  U+F949  uniF949,  CJK COMPATIBILITY IDEOGRAPH-F949
        0x4E86, # G+4551  U+4E86  uni4E86,  CJK UNIFIED IDEOGRAPH-4E86
        0x50DA, # G+4552  U+50DA  uni50DA,  CJK UNIFIED IDEOGRAPH-50DA
        0x5BEE, # G+4553  U+5BEE  uni5BEE,  CJK UNIFIED IDEOGRAPH-5BEE
        0x5ED6, # G+4554  U+5ED6  uni5ED6,  CJK UNIFIED IDEOGRAPH-5ED6
        0x6599, # G+4555  U+6599  uni6599,  CJK UNIFIED IDEOGRAPH-6599
        0x71CE, # G+4556  U+71CE  uni71CE,  CJK UNIFIED IDEOGRAPH-71CE
        0x7642, # G+4557  U+7642  uni7642,  CJK UNIFIED IDEOGRAPH-7642
        0x77AD, # G+4558  U+77AD  uni77AD,  CJK UNIFIED IDEOGRAPH-77AD
        0x804A, # G+4559  U+804A  uni804A,  CJK UNIFIED IDEOGRAPH-804A
        0x84FC, # G+4560  U+84FC  uni84FC,  CJK UNIFIED IDEOGRAPH-84FC
        0x907C, # G+4561  U+907C  uni907C,  CJK UNIFIED IDEOGRAPH-907C
        0x9B27, # G+4562  U+9B27  uni9B27,  CJK UNIFIED IDEOGRAPH-9B27
        0x9F8D, # G+4563  U+9F8D  uni9F8D,  CJK UNIFIED IDEOGRAPH-9F8D
        0xF94A, # G+4564  U+F94A  uniF94A,  CJK COMPATIBILITY IDEOGRAPH-F94A
        0x5A41, # G+4565  U+5A41  uni5A41,  CJK UNIFIED IDEOGRAPH-5A41
        0xF94B, # G+4566  U+F94B  uniF94B,  CJK COMPATIBILITY IDEOGRAPH-F94B
        0xF94C, # G+4567  U+F94C  uniF94C,  CJK COMPATIBILITY IDEOGRAPH-F94C
        0xF94D, # G+4568  U+F94D  uniF94D,  CJK COMPATIBILITY IDEOGRAPH-F94D
        0xF94E, # G+4569  U+F94E  uniF94E,  CJK COMPATIBILITY IDEOGRAPH-F94E
        0x763B, # G+4570  U+763B  uni763B,  CJK UNIFIED IDEOGRAPH-763B
        0xF94F, # G+4571  U+F94F  uniF94F,  CJK COMPATIBILITY IDEOGRAPH-F94F
        0xF950, # G+4572  U+F950  uniF950,  CJK COMPATIBILITY IDEOGRAPH-F950
        0x851E, # G+4573  U+851E  uni851E,  CJK UNIFIED IDEOGRAPH-851E
        0x8938, # G+4574  U+8938  uni8938,  CJK UNIFIED IDEOGRAPH-8938
        0x93E4, # G+4575  U+93E4  uni93E4,  CJK UNIFIED IDEOGRAPH-93E4
        0xF951, # G+4576  U+F951  uniF951,  CJK COMPATIBILITY IDEOGRAPH-F951
        0x5289, # G+4577  U+5289  uni5289,  CJK UNIFIED IDEOGRAPH-5289
        0x65D2, # G+4578  U+65D2  uni65D2,  CJK UNIFIED IDEOGRAPH-65D2
        0x67F3, # G+4579  U+67F3  uni67F3,  CJK UNIFIED IDEOGRAPH-67F3
        0x69B4, # G+4580  U+69B4  uni69B4,  CJK UNIFIED IDEOGRAPH-69B4
        0x6D41, # G+4581  U+6D41  uni6D41,  CJK UNIFIED IDEOGRAPH-6D41
        0x6E9C, # G+4582  U+6E9C  uni6E9C,  CJK UNIFIED IDEOGRAPH-6E9C
        0x700F, # G+4583  U+700F  uni700F,  CJK UNIFIED IDEOGRAPH-700F
        0x7409, # G+4584  U+7409  uni7409,  CJK UNIFIED IDEOGRAPH-7409
        0x7460, # G+4585  U+7460  uni7460,  CJK UNIFIED IDEOGRAPH-7460
        0x7559, # G+4586  U+7559  uni7559,  CJK UNIFIED IDEOGRAPH-7559
        0x7624, # G+4587  U+7624  uni7624,  CJK UNIFIED IDEOGRAPH-7624
        0x786B, # G+4588  U+786B  uni786B,  CJK UNIFIED IDEOGRAPH-786B
        0x8B2C, # G+4589  U+8B2C  uni8B2C,  CJK UNIFIED IDEOGRAPH-8B2C
        0x985E, # G+4590  U+985E  uni985E,  CJK UNIFIED IDEOGRAPH-985E
        0x516D, # G+4591  U+516D  uni516D,  CJK UNIFIED IDEOGRAPH-516D
        0x622E, # G+4592  U+622E  uni622E,  CJK UNIFIED IDEOGRAPH-622E
        0x9678, # G+4593  U+9678  uni9678,  CJK UNIFIED IDEOGRAPH-9678
        0x4F96, # G+4594  U+4F96  uni4F96,  CJK UNIFIED IDEOGRAPH-4F96
        0x502B, # G+4595  U+502B  uni502B,  CJK UNIFIED IDEOGRAPH-502B
        0x5D19, # G+4596  U+5D19  uni5D19,  CJK UNIFIED IDEOGRAPH-5D19
        0x6DEA, # G+4597  U+6DEA  uni6DEA,  CJK UNIFIED IDEOGRAPH-6DEA
        0x7DB8, # G+4598  U+7DB8  uni7DB8,  CJK UNIFIED IDEOGRAPH-7DB8
        0x8F2A, # G+4599  U+8F2A  uni8F2A,  CJK UNIFIED IDEOGRAPH-8F2A
        0x5F8B, # G+4600  U+5F8B  uni5F8B,  CJK UNIFIED IDEOGRAPH-5F8B
        0x6144, # G+4601  U+6144  uni6144,  CJK UNIFIED IDEOGRAPH-6144
        0x6817, # G+4602  U+6817  uni6817,  CJK UNIFIED IDEOGRAPH-6817
        0x9686, # G+4603  U+9686  uni9686,  CJK UNIFIED IDEOGRAPH-9686
        0xF952, # G+4604  U+F952  uniF952,  CJK COMPATIBILITY IDEOGRAPH-F952
        0xF953, # G+4605  U+F953  uniF953,  CJK COMPATIBILITY IDEOGRAPH-F953
        0xF954, # G+4606  U+F954  uniF954,  CJK COMPATIBILITY IDEOGRAPH-F954
        0xF955, # G+4607  U+F955  uniF955,  CJK COMPATIBILITY IDEOGRAPH-F955
        0x695E, # G+4608  U+695E  uni695E,  CJK UNIFIED IDEOGRAPH-695E
        0xF956, # G+4609  U+F956  uniF956,  CJK COMPATIBILITY IDEOGRAPH-F956
        0xF957, # G+4610  U+F957  uniF957,  CJK COMPATIBILITY IDEOGRAPH-F957
        0xF958, # G+4611  U+F958  uniF958,  CJK COMPATIBILITY IDEOGRAPH-F958
        0xF959, # G+4612  U+F959  uniF959,  CJK COMPATIBILITY IDEOGRAPH-F959
        0x4FDA, # G+4613  U+4FDA  uni4FDA,  CJK UNIFIED IDEOGRAPH-4FDA
        0x5229, # G+4614  U+5229  uni5229,  CJK UNIFIED IDEOGRAPH-5229
        0x5398, # G+4615  U+5398  uni5398,  CJK UNIFIED IDEOGRAPH-5398
        0x540F, # G+4616  U+540F  uni540F,  CJK UNIFIED IDEOGRAPH-540F
        0x550E, # G+4617  U+550E  uni550E,  CJK UNIFIED IDEOGRAPH-550E
        0x5C65, # G+4618  U+5C65  uni5C65,  CJK UNIFIED IDEOGRAPH-5C65
        0x60A7, # G+4619  U+60A7  uni60A7,  CJK UNIFIED IDEOGRAPH-60A7
        0x674E, # G+4620  U+674E  uni674E,  CJK UNIFIED IDEOGRAPH-674E
        0x68A8, # G+4621  U+68A8  uni68A8,  CJK UNIFIED IDEOGRAPH-68A8
        0x6D6C, # G+4622  U+6D6C  uni6D6C,  CJK UNIFIED IDEOGRAPH-6D6C
        0x7281, # G+4623  U+7281  uni7281,  CJK UNIFIED IDEOGRAPH-7281
        0x72F8, # G+4624  U+72F8  uni72F8,  CJK UNIFIED IDEOGRAPH-72F8
        0x7406, # G+4625  U+7406  uni7406,  CJK UNIFIED IDEOGRAPH-7406
        0x7483, # G+4626  U+7483  uni7483,  CJK UNIFIED IDEOGRAPH-7483
        0x75E2, # G+4627  U+75E2  uni75E2,  CJK UNIFIED IDEOGRAPH-75E2
        0x7C6C, # G+4628  U+7C6C  uni7C6C,  CJK UNIFIED IDEOGRAPH-7C6C
        0x7F79, # G+4629  U+7F79  uni7F79,  CJK UNIFIED IDEOGRAPH-7F79
        0x7FB8, # G+4630  U+7FB8  uni7FB8,  CJK UNIFIED IDEOGRAPH-7FB8
        0x8389, # G+4631  U+8389  uni8389,  CJK UNIFIED IDEOGRAPH-8389
        0x88CF, # G+4632  U+88CF  uni88CF,  CJK UNIFIED IDEOGRAPH-88CF
        0x88E1, # G+4633  U+88E1  uni88E1,  CJK UNIFIED IDEOGRAPH-88E1
        0x91CC, # G+4634  U+91CC  uni91CC,  CJK UNIFIED IDEOGRAPH-91CC
        0x91D0, # G+4635  U+91D0  uni91D0,  CJK UNIFIED IDEOGRAPH-91D0
        0x96E2, # G+4636  U+96E2  uni96E2,  CJK UNIFIED IDEOGRAPH-96E2
        0x9BC9, # G+4637  U+9BC9  uni9BC9,  CJK UNIFIED IDEOGRAPH-9BC9
        0x541D, # G+4638  U+541D  uni541D,  CJK UNIFIED IDEOGRAPH-541D
        0x6F7E, # G+4639  U+6F7E  uni6F7E,  CJK UNIFIED IDEOGRAPH-6F7E
        0x71D0, # G+4640  U+71D0  uni71D0,  CJK UNIFIED IDEOGRAPH-71D0
        0x7498, # G+4641  U+7498  uni7498,  CJK UNIFIED IDEOGRAPH-7498
        0x85FA, # G+4642  U+85FA  uni85FA,  CJK UNIFIED IDEOGRAPH-85FA
        0x8EAA, # G+4643  U+8EAA  uni8EAA,  CJK UNIFIED IDEOGRAPH-8EAA
        0x96A3, # G+4644  U+96A3  uni96A3,  CJK UNIFIED IDEOGRAPH-96A3
        0x9C57, # G+4645  U+9C57  uni9C57,  CJK UNIFIED IDEOGRAPH-9C57
        0x9E9F, # G+4646  U+9E9F  uni9E9F,  CJK UNIFIED IDEOGRAPH-9E9F
        0x6797, # G+4647  U+6797  uni6797,  CJK UNIFIED IDEOGRAPH-6797
        0x6DCB, # G+4648  U+6DCB  uni6DCB,  CJK UNIFIED IDEOGRAPH-6DCB
        0x7433, # G+4649  U+7433  uni7433,  CJK UNIFIED IDEOGRAPH-7433
        0x81E8, # G+4650  U+81E8  uni81E8,  CJK UNIFIED IDEOGRAPH-81E8
        0x9716, # G+4651  U+9716  uni9716,  CJK UNIFIED IDEOGRAPH-9716
        0x782C, # G+4652  U+782C  uni782C,  CJK UNIFIED IDEOGRAPH-782C
        0x7ACB, # G+4653  U+7ACB  uni7ACB,  CJK UNIFIED IDEOGRAPH-7ACB
        0x7B20, # G+4654  U+7B20  uni7B20,  CJK UNIFIED IDEOGRAPH-7B20
        0x7C92, # G+4655  U+7C92  uni7C92,  CJK UNIFIED IDEOGRAPH-7C92
        0x6469, # G+4656  U+6469  uni6469,  CJK UNIFIED IDEOGRAPH-6469
        0x746A, # G+4657  U+746A  uni746A,  CJK UNIFIED IDEOGRAPH-746A
        0x75F2, # G+4658  U+75F2  uni75F2,  CJK UNIFIED IDEOGRAPH-75F2
        0x78BC, # G+4659  U+78BC  uni78BC,  CJK UNIFIED IDEOGRAPH-78BC
        0x78E8, # G+4660  U+78E8  uni78E8,  CJK UNIFIED IDEOGRAPH-78E8
        0x99AC, # G+4661  U+99AC  uni99AC,  CJK UNIFIED IDEOGRAPH-99AC
        0x9B54, # G+4662  U+9B54  uni9B54,  CJK UNIFIED IDEOGRAPH-9B54
        0x9EBB, # G+4663  U+9EBB  uni9EBB,  CJK UNIFIED IDEOGRAPH-9EBB
        0x5BDE, # G+4664  U+5BDE  uni5BDE,  CJK UNIFIED IDEOGRAPH-5BDE
        0x5E55, # G+4665  U+5E55  uni5E55,  CJK UNIFIED IDEOGRAPH-5E55
        0x6F20, # G+4666  U+6F20  uni6F20,  CJK UNIFIED IDEOGRAPH-6F20
        0x819C, # G+4667  U+819C  uni819C,  CJK UNIFIED IDEOGRAPH-819C
        0x83AB, # G+4668  U+83AB  uni83AB,  CJK UNIFIED IDEOGRAPH-83AB
        0x9088, # G+4669  U+9088  uni9088,  CJK UNIFIED IDEOGRAPH-9088
        0x4E07, # G+4670  U+4E07  uni4E07,  CJK UNIFIED IDEOGRAPH-4E07
        0x534D, # G+4671  U+534D  uni534D,  CJK UNIFIED IDEOGRAPH-534D
        0x5A29, # G+4672  U+5A29  uni5A29,  CJK UNIFIED IDEOGRAPH-5A29
        0x5DD2, # G+4673  U+5DD2  uni5DD2,  CJK UNIFIED IDEOGRAPH-5DD2
        0x5F4E, # G+4674  U+5F4E  uni5F4E,  CJK UNIFIED IDEOGRAPH-5F4E
        0x6162, # G+4675  U+6162  uni6162,  CJK UNIFIED IDEOGRAPH-6162
        0x633D, # G+4676  U+633D  uni633D,  CJK UNIFIED IDEOGRAPH-633D
        0x6669, # G+4677  U+6669  uni6669,  CJK UNIFIED IDEOGRAPH-6669
        0x66FC, # G+4678  U+66FC  uni66FC,  CJK UNIFIED IDEOGRAPH-66FC
        0x6EFF, # G+4679  U+6EFF  uni6EFF,  CJK UNIFIED IDEOGRAPH-6EFF
        0x6F2B, # G+4680  U+6F2B  uni6F2B,  CJK UNIFIED IDEOGRAPH-6F2B
        0x7063, # G+4681  U+7063  uni7063,  CJK UNIFIED IDEOGRAPH-7063
        0x779E, # G+4682  U+779E  uni779E,  CJK UNIFIED IDEOGRAPH-779E
        0x842C, # G+4683  U+842C  uni842C,  CJK UNIFIED IDEOGRAPH-842C
        0x8513, # G+4684  U+8513  uni8513,  CJK UNIFIED IDEOGRAPH-8513
        0x883B, # G+4685  U+883B  uni883B,  CJK UNIFIED IDEOGRAPH-883B
        0x8F13, # G+4686  U+8F13  uni8F13,  CJK UNIFIED IDEOGRAPH-8F13
        0x9945, # G+4687  U+9945  uni9945,  CJK UNIFIED IDEOGRAPH-9945
        0x9C3B, # G+4688  U+9C3B  uni9C3B,  CJK UNIFIED IDEOGRAPH-9C3B
        0x551C, # G+4689  U+551C  uni551C,  CJK UNIFIED IDEOGRAPH-551C
        0x62B9, # G+4690  U+62B9  uni62B9,  CJK UNIFIED IDEOGRAPH-62B9
        0x672B, # G+4691  U+672B  uni672B,  CJK UNIFIED IDEOGRAPH-672B
        0x6CAB, # G+4692  U+6CAB  uni6CAB,  CJK UNIFIED IDEOGRAPH-6CAB
        0x8309, # G+4693  U+8309  uni8309,  CJK UNIFIED IDEOGRAPH-8309
        0x896A, # G+4694  U+896A  uni896A,  CJK UNIFIED IDEOGRAPH-896A
        0x977A, # G+4695  U+977A  uni977A,  CJK UNIFIED IDEOGRAPH-977A
        0x4EA1, # G+4696  U+4EA1  uni4EA1,  CJK UNIFIED IDEOGRAPH-4EA1
        0x5984, # G+4697  U+5984  uni5984,  CJK UNIFIED IDEOGRAPH-5984
        0x5FD8, # G+4698  U+5FD8  uni5FD8,  CJK UNIFIED IDEOGRAPH-5FD8
        0x5FD9, # G+4699  U+5FD9  uni5FD9,  CJK UNIFIED IDEOGRAPH-5FD9
        0x671B, # G+4700  U+671B  uni671B,  CJK UNIFIED IDEOGRAPH-671B
        0x7DB2, # G+4701  U+7DB2  uni7DB2,  CJK UNIFIED IDEOGRAPH-7DB2
        0x7F54, # G+4702  U+7F54  uni7F54,  CJK UNIFIED IDEOGRAPH-7F54
        0x8292, # G+4703  U+8292  uni8292,  CJK UNIFIED IDEOGRAPH-8292
        0x832B, # G+4704  U+832B  uni832B,  CJK UNIFIED IDEOGRAPH-832B
        0x83BD, # G+4705  U+83BD  uni83BD,  CJK UNIFIED IDEOGRAPH-83BD
        0x8F1E, # G+4706  U+8F1E  uni8F1E,  CJK UNIFIED IDEOGRAPH-8F1E
        0x9099, # G+4707  U+9099  uni9099,  CJK UNIFIED IDEOGRAPH-9099
        0x57CB, # G+4708  U+57CB  uni57CB,  CJK UNIFIED IDEOGRAPH-57CB
        0x59B9, # G+4709  U+59B9  uni59B9,  CJK UNIFIED IDEOGRAPH-59B9
        0x5A92, # G+4710  U+5A92  uni5A92,  CJK UNIFIED IDEOGRAPH-5A92
        0x5BD0, # G+4711  U+5BD0  uni5BD0,  CJK UNIFIED IDEOGRAPH-5BD0
        0x6627, # G+4712  U+6627  uni6627,  CJK UNIFIED IDEOGRAPH-6627
        0x679A, # G+4713  U+679A  uni679A,  CJK UNIFIED IDEOGRAPH-679A
        0x6885, # G+4714  U+6885  uni6885,  CJK UNIFIED IDEOGRAPH-6885
        0x6BCF, # G+4715  U+6BCF  uni6BCF,  CJK UNIFIED IDEOGRAPH-6BCF
        0x7164, # G+4716  U+7164  uni7164,  CJK UNIFIED IDEOGRAPH-7164
        0x7F75, # G+4717  U+7F75  uni7F75,  CJK UNIFIED IDEOGRAPH-7F75
        0x8CB7, # G+4718  U+8CB7  uni8CB7,  CJK UNIFIED IDEOGRAPH-8CB7
        0x8CE3, # G+4719  U+8CE3  uni8CE3,  CJK UNIFIED IDEOGRAPH-8CE3
        0x9081, # G+4720  U+9081  uni9081,  CJK UNIFIED IDEOGRAPH-9081
        0x9B45, # G+4721  U+9B45  uni9B45,  CJK UNIFIED IDEOGRAPH-9B45
        0x8108, # G+4722  U+8108  uni8108,  CJK UNIFIED IDEOGRAPH-8108
        0x8C8A, # G+4723  U+8C8A  uni8C8A,  CJK UNIFIED IDEOGRAPH-8C8A
        0x964C, # G+4724  U+964C  uni964C,  CJK UNIFIED IDEOGRAPH-964C
        0x9A40, # G+4725  U+9A40  uni9A40,  CJK UNIFIED IDEOGRAPH-9A40
        0x9EA5, # G+4726  U+9EA5  uni9EA5,  CJK UNIFIED IDEOGRAPH-9EA5
        0x5B5F, # G+4727  U+5B5F  uni5B5F,  CJK UNIFIED IDEOGRAPH-5B5F
        0x6C13, # G+4728  U+6C13  uni6C13,  CJK UNIFIED IDEOGRAPH-6C13
        0x731B, # G+4729  U+731B  uni731B,  CJK UNIFIED IDEOGRAPH-731B
        0x76F2, # G+4730  U+76F2  uni76F2,  CJK UNIFIED IDEOGRAPH-76F2
        0x76DF, # G+4731  U+76DF  uni76DF,  CJK UNIFIED IDEOGRAPH-76DF
        0x840C, # G+4732  U+840C  uni840C,  CJK UNIFIED IDEOGRAPH-840C
        0x51AA, # G+4733  U+51AA  uni51AA,  CJK UNIFIED IDEOGRAPH-51AA
        0x8993, # G+4734  U+8993  uni8993,  CJK UNIFIED IDEOGRAPH-8993
        0x514D, # G+4735  U+514D  uni514D,  CJK UNIFIED IDEOGRAPH-514D
        0x5195, # G+4736  U+5195  uni5195,  CJK UNIFIED IDEOGRAPH-5195
        0x52C9, # G+4737  U+52C9  uni52C9,  CJK UNIFIED IDEOGRAPH-52C9
        0x68C9, # G+4738  U+68C9  uni68C9,  CJK UNIFIED IDEOGRAPH-68C9
        0x6C94, # G+4739  U+6C94  uni6C94,  CJK UNIFIED IDEOGRAPH-6C94
        0x7704, # G+4740  U+7704  uni7704,  CJK UNIFIED IDEOGRAPH-7704
        0x7720, # G+4741  U+7720  uni7720,  CJK UNIFIED IDEOGRAPH-7720
        0x7DBF, # G+4742  U+7DBF  uni7DBF,  CJK UNIFIED IDEOGRAPH-7DBF
        0x7DEC, # G+4743  U+7DEC  uni7DEC,  CJK UNIFIED IDEOGRAPH-7DEC
        0x9762, # G+4744  U+9762  uni9762,  CJK UNIFIED IDEOGRAPH-9762
        0x9EB5, # G+4745  U+9EB5  uni9EB5,  CJK UNIFIED IDEOGRAPH-9EB5
        0x6EC5, # G+4746  U+6EC5  uni6EC5,  CJK UNIFIED IDEOGRAPH-6EC5
        0x8511, # G+4747  U+8511  uni8511,  CJK UNIFIED IDEOGRAPH-8511
        0x51A5, # G+4748  U+51A5  uni51A5,  CJK UNIFIED IDEOGRAPH-51A5
        0x540D, # G+4749  U+540D  uni540D,  CJK UNIFIED IDEOGRAPH-540D
        0x547D, # G+4750  U+547D  uni547D,  CJK UNIFIED IDEOGRAPH-547D
        0x660E, # G+4751  U+660E  uni660E,  CJK UNIFIED IDEOGRAPH-660E
        0x669D, # G+4752  U+669D  uni669D,  CJK UNIFIED IDEOGRAPH-669D
        0x6927, # G+4753  U+6927  uni6927,  CJK UNIFIED IDEOGRAPH-6927
        0x6E9F, # G+4754  U+6E9F  uni6E9F,  CJK UNIFIED IDEOGRAPH-6E9F
        0x76BF, # G+4755  U+76BF  uni76BF,  CJK UNIFIED IDEOGRAPH-76BF
        0x7791, # G+4756  U+7791  uni7791,  CJK UNIFIED IDEOGRAPH-7791
        0x8317, # G+4757  U+8317  uni8317,  CJK UNIFIED IDEOGRAPH-8317
        0x84C2, # G+4758  U+84C2  uni84C2,  CJK UNIFIED IDEOGRAPH-84C2
        0x879F, # G+4759  U+879F  uni879F,  CJK UNIFIED IDEOGRAPH-879F
        0x9169, # G+4760  U+9169  uni9169,  CJK UNIFIED IDEOGRAPH-9169
        0x9298, # G+4761  U+9298  uni9298,  CJK UNIFIED IDEOGRAPH-9298
        0x9CF4, # G+4762  U+9CF4  uni9CF4,  CJK UNIFIED IDEOGRAPH-9CF4
        0x8882, # G+4763  U+8882  uni8882,  CJK UNIFIED IDEOGRAPH-8882
        0x4FAE, # G+4764  U+4FAE  uni4FAE,  CJK UNIFIED IDEOGRAPH-4FAE
        0x5192, # G+4765  U+5192  uni5192,  CJK UNIFIED IDEOGRAPH-5192
        0x52DF, # G+4766  U+52DF  uni52DF,  CJK UNIFIED IDEOGRAPH-52DF
        0x59C6, # G+4767  U+59C6  uni59C6,  CJK UNIFIED IDEOGRAPH-59C6
        0x5E3D, # G+4768  U+5E3D  uni5E3D,  CJK UNIFIED IDEOGRAPH-5E3D
        0x6155, # G+4769  U+6155  uni6155,  CJK UNIFIED IDEOGRAPH-6155
        0x6478, # G+4770  U+6478  uni6478,  CJK UNIFIED IDEOGRAPH-6478
        0x6479, # G+4771  U+6479  uni6479,  CJK UNIFIED IDEOGRAPH-6479
        0x66AE, # G+4772  U+66AE  uni66AE,  CJK UNIFIED IDEOGRAPH-66AE
        0x67D0, # G+4773  U+67D0  uni67D0,  CJK UNIFIED IDEOGRAPH-67D0
        0x6A21, # G+4774  U+6A21  uni6A21,  CJK UNIFIED IDEOGRAPH-6A21
        0x6BCD, # G+4775  U+6BCD  uni6BCD,  CJK UNIFIED IDEOGRAPH-6BCD
        0x6BDB, # G+4776  U+6BDB  uni6BDB,  CJK UNIFIED IDEOGRAPH-6BDB
        0x725F, # G+4777  U+725F  uni725F,  CJK UNIFIED IDEOGRAPH-725F
        0x7261, # G+4778  U+7261  uni7261,  CJK UNIFIED IDEOGRAPH-7261
        0x7441, # G+4779  U+7441  uni7441,  CJK UNIFIED IDEOGRAPH-7441
        0x7738, # G+4780  U+7738  uni7738,  CJK UNIFIED IDEOGRAPH-7738
        0x77DB, # G+4781  U+77DB  uni77DB,  CJK UNIFIED IDEOGRAPH-77DB
        0x8017, # G+4782  U+8017  uni8017,  CJK UNIFIED IDEOGRAPH-8017
        0x82BC, # G+4783  U+82BC  uni82BC,  CJK UNIFIED IDEOGRAPH-82BC
        0x8305, # G+4784  U+8305  uni8305,  CJK UNIFIED IDEOGRAPH-8305
        0x8B00, # G+4785  U+8B00  uni8B00,  CJK UNIFIED IDEOGRAPH-8B00
        0x8B28, # G+4786  U+8B28  uni8B28,  CJK UNIFIED IDEOGRAPH-8B28
        0x8C8C, # G+4787  U+8C8C  uni8C8C,  CJK UNIFIED IDEOGRAPH-8C8C
        0x6728, # G+4788  U+6728  uni6728,  CJK UNIFIED IDEOGRAPH-6728
        0x6C90, # G+4789  U+6C90  uni6C90,  CJK UNIFIED IDEOGRAPH-6C90
        0x7267, # G+4790  U+7267  uni7267,  CJK UNIFIED IDEOGRAPH-7267
        0x76EE, # G+4791  U+76EE  uni76EE,  CJK UNIFIED IDEOGRAPH-76EE
        0x7766, # G+4792  U+7766  uni7766,  CJK UNIFIED IDEOGRAPH-7766
        0x7A46, # G+4793  U+7A46  uni7A46,  CJK UNIFIED IDEOGRAPH-7A46
        0x9DA9, # G+4794  U+9DA9  uni9DA9,  CJK UNIFIED IDEOGRAPH-9DA9
        0x6B7F, # G+4795  U+6B7F  uni6B7F,  CJK UNIFIED IDEOGRAPH-6B7F
        0x6C92, # G+4796  U+6C92  uni6C92,  CJK UNIFIED IDEOGRAPH-6C92
        0x5922, # G+4797  U+5922  uni5922,  CJK UNIFIED IDEOGRAPH-5922
        0x6726, # G+4798  U+6726  uni6726,  CJK UNIFIED IDEOGRAPH-6726
        0x8499, # G+4799  U+8499  uni8499,  CJK UNIFIED IDEOGRAPH-8499
        0x536F, # G+4800  U+536F  uni536F,  CJK UNIFIED IDEOGRAPH-536F
        0x5893, # G+4801  U+5893  uni5893,  CJK UNIFIED IDEOGRAPH-5893
        0x5999, # G+4802  U+5999  uni5999,  CJK UNIFIED IDEOGRAPH-5999
        0x5EDF, # G+4803  U+5EDF  uni5EDF,  CJK UNIFIED IDEOGRAPH-5EDF
        0x63CF, # G+4804  U+63CF  uni63CF,  CJK UNIFIED IDEOGRAPH-63CF
        0x6634, # G+4805  U+6634  uni6634,  CJK UNIFIED IDEOGRAPH-6634
        0x6773, # G+4806  U+6773  uni6773,  CJK UNIFIED IDEOGRAPH-6773
        0x6E3A, # G+4807  U+6E3A  uni6E3A,  CJK UNIFIED IDEOGRAPH-6E3A
        0x732B, # G+4808  U+732B  uni732B,  CJK UNIFIED IDEOGRAPH-732B
        0x7AD7, # G+4809  U+7AD7  uni7AD7,  CJK UNIFIED IDEOGRAPH-7AD7
        0x82D7, # G+4810  U+82D7  uni82D7,  CJK UNIFIED IDEOGRAPH-82D7
        0x9328, # G+4811  U+9328  uni9328,  CJK UNIFIED IDEOGRAPH-9328
        0x52D9, # G+4812  U+52D9  uni52D9,  CJK UNIFIED IDEOGRAPH-52D9
        0x5DEB, # G+4813  U+5DEB  uni5DEB,  CJK UNIFIED IDEOGRAPH-5DEB
        0x61AE, # G+4814  U+61AE  uni61AE,  CJK UNIFIED IDEOGRAPH-61AE
        0x61CB, # G+4815  U+61CB  uni61CB,  CJK UNIFIED IDEOGRAPH-61CB
        0x620A, # G+4816  U+620A  uni620A,  CJK UNIFIED IDEOGRAPH-620A
        0x62C7, # G+4817  U+62C7  uni62C7,  CJK UNIFIED IDEOGRAPH-62C7
        0x64AB, # G+4818  U+64AB  uni64AB,  CJK UNIFIED IDEOGRAPH-64AB
        0x65E0, # G+4819  U+65E0  uni65E0,  CJK UNIFIED IDEOGRAPH-65E0
        0x6959, # G+4820  U+6959  uni6959,  CJK UNIFIED IDEOGRAPH-6959
        0x6B66, # G+4821  U+6B66  uni6B66,  CJK UNIFIED IDEOGRAPH-6B66
        0x6BCB, # G+4822  U+6BCB  uni6BCB,  CJK UNIFIED IDEOGRAPH-6BCB
        0x7121, # G+4823  U+7121  uni7121,  CJK UNIFIED IDEOGRAPH-7121
        0x73F7, # G+4824  U+73F7  uni73F7,  CJK UNIFIED IDEOGRAPH-73F7
        0x755D, # G+4825  U+755D  uni755D,  CJK UNIFIED IDEOGRAPH-755D
        0x7E46, # G+4826  U+7E46  uni7E46,  CJK UNIFIED IDEOGRAPH-7E46
        0x821E, # G+4827  U+821E  uni821E,  CJK UNIFIED IDEOGRAPH-821E
        0x8302, # G+4828  U+8302  uni8302,  CJK UNIFIED IDEOGRAPH-8302
        0x856A, # G+4829  U+856A  uni856A,  CJK UNIFIED IDEOGRAPH-856A
        0x8AA3, # G+4830  U+8AA3  uni8AA3,  CJK UNIFIED IDEOGRAPH-8AA3
        0x8CBF, # G+4831  U+8CBF  uni8CBF,  CJK UNIFIED IDEOGRAPH-8CBF
        0x9727, # G+4832  U+9727  uni9727,  CJK UNIFIED IDEOGRAPH-9727
        0x9D61, # G+4833  U+9D61  uni9D61,  CJK UNIFIED IDEOGRAPH-9D61
        0x58A8, # G+4834  U+58A8  uni58A8,  CJK UNIFIED IDEOGRAPH-58A8
        0x9ED8, # G+4835  U+9ED8  uni9ED8,  CJK UNIFIED IDEOGRAPH-9ED8
        0x5011, # G+4836  U+5011  uni5011,  CJK UNIFIED IDEOGRAPH-5011
        0x520E, # G+4837  U+520E  uni520E,  CJK UNIFIED IDEOGRAPH-520E
        0x543B, # G+4838  U+543B  uni543B,  CJK UNIFIED IDEOGRAPH-543B
        0x554F, # G+4839  U+554F  uni554F,  CJK UNIFIED IDEOGRAPH-554F
        0x6587, # G+4840  U+6587  uni6587,  CJK UNIFIED IDEOGRAPH-6587
        0x6C76, # G+4841  U+6C76  uni6C76,  CJK UNIFIED IDEOGRAPH-6C76
        0x7D0A, # G+4842  U+7D0A  uni7D0A,  CJK UNIFIED IDEOGRAPH-7D0A
        0x7D0B, # G+4843  U+7D0B  uni7D0B,  CJK UNIFIED IDEOGRAPH-7D0B
        0x805E, # G+4844  U+805E  uni805E,  CJK UNIFIED IDEOGRAPH-805E
        0x868A, # G+4845  U+868A  uni868A,  CJK UNIFIED IDEOGRAPH-868A
        0x9580, # G+4846  U+9580  uni9580,  CJK UNIFIED IDEOGRAPH-9580
        0x96EF, # G+4847  U+96EF  uni96EF,  CJK UNIFIED IDEOGRAPH-96EF
        0x52FF, # G+4848  U+52FF  uni52FF,  CJK UNIFIED IDEOGRAPH-52FF
        0x6C95, # G+4849  U+6C95  uni6C95,  CJK UNIFIED IDEOGRAPH-6C95
        0x7269, # G+4850  U+7269  uni7269,  CJK UNIFIED IDEOGRAPH-7269
        0x5473, # G+4851  U+5473  uni5473,  CJK UNIFIED IDEOGRAPH-5473
        0x5A9A, # G+4852  U+5A9A  uni5A9A,  CJK UNIFIED IDEOGRAPH-5A9A
        0x5C3E, # G+4853  U+5C3E  uni5C3E,  CJK UNIFIED IDEOGRAPH-5C3E
        0x5D4B, # G+4854  U+5D4B  uni5D4B,  CJK UNIFIED IDEOGRAPH-5D4B
        0x5F4C, # G+4855  U+5F4C  uni5F4C,  CJK UNIFIED IDEOGRAPH-5F4C
        0x5FAE, # G+4856  U+5FAE  uni5FAE,  CJK UNIFIED IDEOGRAPH-5FAE
        0x672A, # G+4857  U+672A  uni672A,  CJK UNIFIED IDEOGRAPH-672A
        0x68B6, # G+4858  U+68B6  uni68B6,  CJK UNIFIED IDEOGRAPH-68B6
        0x6963, # G+4859  U+6963  uni6963,  CJK UNIFIED IDEOGRAPH-6963
        0x6E3C, # G+4860  U+6E3C  uni6E3C,  CJK UNIFIED IDEOGRAPH-6E3C
        0x6E44, # G+4861  U+6E44  uni6E44,  CJK UNIFIED IDEOGRAPH-6E44
        0x7709, # G+4862  U+7709  uni7709,  CJK UNIFIED IDEOGRAPH-7709
        0x7C73, # G+4863  U+7C73  uni7C73,  CJK UNIFIED IDEOGRAPH-7C73
        0x7F8E, # G+4864  U+7F8E  uni7F8E,  CJK UNIFIED IDEOGRAPH-7F8E
        0x8587, # G+4865  U+8587  uni8587,  CJK UNIFIED IDEOGRAPH-8587
        0x8B0E, # G+4866  U+8B0E  uni8B0E,  CJK UNIFIED IDEOGRAPH-8B0E
        0x8FF7, # G+4867  U+8FF7  uni8FF7,  CJK UNIFIED IDEOGRAPH-8FF7
        0x9761, # G+4868  U+9761  uni9761,  CJK UNIFIED IDEOGRAPH-9761
        0x9EF4, # G+4869  U+9EF4  uni9EF4,  CJK UNIFIED IDEOGRAPH-9EF4
        0x5CB7, # G+4870  U+5CB7  uni5CB7,  CJK UNIFIED IDEOGRAPH-5CB7
        0x60B6, # G+4871  U+60B6  uni60B6,  CJK UNIFIED IDEOGRAPH-60B6
        0x610D, # G+4872  U+610D  uni610D,  CJK UNIFIED IDEOGRAPH-610D
        0x61AB, # G+4873  U+61AB  uni61AB,  CJK UNIFIED IDEOGRAPH-61AB
        0x654F, # G+4874  U+654F  uni654F,  CJK UNIFIED IDEOGRAPH-654F
        0x65FB, # G+4875  U+65FB  uni65FB,  CJK UNIFIED IDEOGRAPH-65FB
        0x65FC, # G+4876  U+65FC  uni65FC,  CJK UNIFIED IDEOGRAPH-65FC
        0x6C11, # G+4877  U+6C11  uni6C11,  CJK UNIFIED IDEOGRAPH-6C11
        0x6CEF, # G+4878  U+6CEF  uni6CEF,  CJK UNIFIED IDEOGRAPH-6CEF
        0x739F, # G+4879  U+739F  uni739F,  CJK UNIFIED IDEOGRAPH-739F
        0x73C9, # G+4880  U+73C9  uni73C9,  CJK UNIFIED IDEOGRAPH-73C9
        0x7DE1, # G+4881  U+7DE1  uni7DE1,  CJK UNIFIED IDEOGRAPH-7DE1
        0x9594, # G+4882  U+9594  uni9594,  CJK UNIFIED IDEOGRAPH-9594
        0x5BC6, # G+4883  U+5BC6  uni5BC6,  CJK UNIFIED IDEOGRAPH-5BC6
        0x871C, # G+4884  U+871C  uni871C,  CJK UNIFIED IDEOGRAPH-871C
        0x8B10, # G+4885  U+8B10  uni8B10,  CJK UNIFIED IDEOGRAPH-8B10
        0x525D, # G+4886  U+525D  uni525D,  CJK UNIFIED IDEOGRAPH-525D
        0x535A, # G+4887  U+535A  uni535A,  CJK UNIFIED IDEOGRAPH-535A
        0x62CD, # G+4888  U+62CD  uni62CD,  CJK UNIFIED IDEOGRAPH-62CD
        0x640F, # G+4889  U+640F  uni640F,  CJK UNIFIED IDEOGRAPH-640F
        0x64B2, # G+4890  U+64B2  uni64B2,  CJK UNIFIED IDEOGRAPH-64B2
        0x6734, # G+4891  U+6734  uni6734,  CJK UNIFIED IDEOGRAPH-6734
        0x6A38, # G+4892  U+6A38  uni6A38,  CJK UNIFIED IDEOGRAPH-6A38
        0x6CCA, # G+4893  U+6CCA  uni6CCA,  CJK UNIFIED IDEOGRAPH-6CCA
        0x73C0, # G+4894  U+73C0  uni73C0,  CJK UNIFIED IDEOGRAPH-73C0
        0x749E, # G+4895  U+749E  uni749E,  CJK UNIFIED IDEOGRAPH-749E
        0x7B94, # G+4896  U+7B94  uni7B94,  CJK UNIFIED IDEOGRAPH-7B94
        0x7C95, # G+4897  U+7C95  uni7C95,  CJK UNIFIED IDEOGRAPH-7C95
        0x7E1B, # G+4898  U+7E1B  uni7E1B,  CJK UNIFIED IDEOGRAPH-7E1B
        0x818A, # G+4899  U+818A  uni818A,  CJK UNIFIED IDEOGRAPH-818A
        0x8236, # G+4900  U+8236  uni8236,  CJK UNIFIED IDEOGRAPH-8236
        0x8584, # G+4901  U+8584  uni8584,  CJK UNIFIED IDEOGRAPH-8584
        0x8FEB, # G+4902  U+8FEB  uni8FEB,  CJK UNIFIED IDEOGRAPH-8FEB
        0x96F9, # G+4903  U+96F9  uni96F9,  CJK UNIFIED IDEOGRAPH-96F9
        0x99C1, # G+4904  U+99C1  uni99C1,  CJK UNIFIED IDEOGRAPH-99C1
        0x4F34, # G+4905  U+4F34  uni4F34,  CJK UNIFIED IDEOGRAPH-4F34
        0x534A, # G+4906  U+534A  uni534A,  CJK UNIFIED IDEOGRAPH-534A
        0x53CD, # G+4907  U+53CD  uni53CD,  CJK UNIFIED IDEOGRAPH-53CD
        0x53DB, # G+4908  U+53DB  uni53DB,  CJK UNIFIED IDEOGRAPH-53DB
        0x62CC, # G+4909  U+62CC  uni62CC,  CJK UNIFIED IDEOGRAPH-62CC
        0x642C, # G+4910  U+642C  uni642C,  CJK UNIFIED IDEOGRAPH-642C
        0x6500, # G+4911  U+6500  uni6500,  CJK UNIFIED IDEOGRAPH-6500
        0x6591, # G+4912  U+6591  uni6591,  CJK UNIFIED IDEOGRAPH-6591
        0x69C3, # G+4913  U+69C3  uni69C3,  CJK UNIFIED IDEOGRAPH-69C3
        0x6CEE, # G+4914  U+6CEE  uni6CEE,  CJK UNIFIED IDEOGRAPH-6CEE
        0x6F58, # G+4915  U+6F58  uni6F58,  CJK UNIFIED IDEOGRAPH-6F58
        0x73ED, # G+4916  U+73ED  uni73ED,  CJK UNIFIED IDEOGRAPH-73ED
        0x7554, # G+4917  U+7554  uni7554,  CJK UNIFIED IDEOGRAPH-7554
        0x7622, # G+4918  U+7622  uni7622,  CJK UNIFIED IDEOGRAPH-7622
        0x76E4, # G+4919  U+76E4  uni76E4,  CJK UNIFIED IDEOGRAPH-76E4
        0x76FC, # G+4920  U+76FC  uni76FC,  CJK UNIFIED IDEOGRAPH-76FC
        0x78D0, # G+4921  U+78D0  uni78D0,  CJK UNIFIED IDEOGRAPH-78D0
        0x78FB, # G+4922  U+78FB  uni78FB,  CJK UNIFIED IDEOGRAPH-78FB
        0x792C, # G+4923  U+792C  uni792C,  CJK UNIFIED IDEOGRAPH-792C
        0x7D46, # G+4924  U+7D46  uni7D46,  CJK UNIFIED IDEOGRAPH-7D46
        0x822C, # G+4925  U+822C  uni822C,  CJK UNIFIED IDEOGRAPH-822C
        0x87E0, # G+4926  U+87E0  uni87E0,  CJK UNIFIED IDEOGRAPH-87E0
        0x8FD4, # G+4927  U+8FD4  uni8FD4,  CJK UNIFIED IDEOGRAPH-8FD4
        0x9812, # G+4928  U+9812  uni9812,  CJK UNIFIED IDEOGRAPH-9812
        0x98EF, # G+4929  U+98EF  uni98EF,  CJK UNIFIED IDEOGRAPH-98EF
        0x52C3, # G+4930  U+52C3  uni52C3,  CJK UNIFIED IDEOGRAPH-52C3
        0x62D4, # G+4931  U+62D4  uni62D4,  CJK UNIFIED IDEOGRAPH-62D4
        0x64A5, # G+4932  U+64A5  uni64A5,  CJK UNIFIED IDEOGRAPH-64A5
        0x6E24, # G+4933  U+6E24  uni6E24,  CJK UNIFIED IDEOGRAPH-6E24
        0x6F51, # G+4934  U+6F51  uni6F51,  CJK UNIFIED IDEOGRAPH-6F51
        0x767C, # G+4935  U+767C  uni767C,  CJK UNIFIED IDEOGRAPH-767C
        0x8DCB, # G+4936  U+8DCB  uni8DCB,  CJK UNIFIED IDEOGRAPH-8DCB
        0x91B1, # G+4937  U+91B1  uni91B1,  CJK UNIFIED IDEOGRAPH-91B1
        0x9262, # G+4938  U+9262  uni9262,  CJK UNIFIED IDEOGRAPH-9262
        0x9AEE, # G+4939  U+9AEE  uni9AEE,  CJK UNIFIED IDEOGRAPH-9AEE
        0x9B43, # G+4940  U+9B43  uni9B43,  CJK UNIFIED IDEOGRAPH-9B43
        0x5023, # G+4941  U+5023  uni5023,  CJK UNIFIED IDEOGRAPH-5023
        0x508D, # G+4942  U+508D  uni508D,  CJK UNIFIED IDEOGRAPH-508D
        0x574A, # G+4943  U+574A  uni574A,  CJK UNIFIED IDEOGRAPH-574A
        0x59A8, # G+4944  U+59A8  uni59A8,  CJK UNIFIED IDEOGRAPH-59A8
        0x5C28, # G+4945  U+5C28  uni5C28,  CJK UNIFIED IDEOGRAPH-5C28
        0x5E47, # G+4946  U+5E47  uni5E47,  CJK UNIFIED IDEOGRAPH-5E47
        0x5F77, # G+4947  U+5F77  uni5F77,  CJK UNIFIED IDEOGRAPH-5F77
        0x623F, # G+4948  U+623F  uni623F,  CJK UNIFIED IDEOGRAPH-623F
        0x653E, # G+4949  U+653E  uni653E,  CJK UNIFIED IDEOGRAPH-653E
        0x65B9, # G+4950  U+65B9  uni65B9,  CJK UNIFIED IDEOGRAPH-65B9
        0x65C1, # G+4951  U+65C1  uni65C1,  CJK UNIFIED IDEOGRAPH-65C1
        0x6609, # G+4952  U+6609  uni6609,  CJK UNIFIED IDEOGRAPH-6609
        0x678B, # G+4953  U+678B  uni678B,  CJK UNIFIED IDEOGRAPH-678B
        0x699C, # G+4954  U+699C  uni699C,  CJK UNIFIED IDEOGRAPH-699C
        0x6EC2, # G+4955  U+6EC2  uni6EC2,  CJK UNIFIED IDEOGRAPH-6EC2
        0x78C5, # G+4956  U+78C5  uni78C5,  CJK UNIFIED IDEOGRAPH-78C5
        0x7D21, # G+4957  U+7D21  uni7D21,  CJK UNIFIED IDEOGRAPH-7D21
        0x80AA, # G+4958  U+80AA  uni80AA,  CJK UNIFIED IDEOGRAPH-80AA
        0x8180, # G+4959  U+8180  uni8180,  CJK UNIFIED IDEOGRAPH-8180
        0x822B, # G+4960  U+822B  uni822B,  CJK UNIFIED IDEOGRAPH-822B
        0x82B3, # G+4961  U+82B3  uni82B3,  CJK UNIFIED IDEOGRAPH-82B3
        0x84A1, # G+4962  U+84A1  uni84A1,  CJK UNIFIED IDEOGRAPH-84A1
        0x868C, # G+4963  U+868C  uni868C,  CJK UNIFIED IDEOGRAPH-868C
        0x8A2A, # G+4964  U+8A2A  uni8A2A,  CJK UNIFIED IDEOGRAPH-8A2A
        0x8B17, # G+4965  U+8B17  uni8B17,  CJK UNIFIED IDEOGRAPH-8B17
        0x90A6, # G+4966  U+90A6  uni90A6,  CJK UNIFIED IDEOGRAPH-90A6
        0x9632, # G+4967  U+9632  uni9632,  CJK UNIFIED IDEOGRAPH-9632
        0x9F90, # G+4968  U+9F90  uni9F90,  CJK UNIFIED IDEOGRAPH-9F90
        0x500D, # G+4969  U+500D  uni500D,  CJK UNIFIED IDEOGRAPH-500D
        0x4FF3, # G+4970  U+4FF3  uni4FF3,  CJK UNIFIED IDEOGRAPH-4FF3
        0x57F9, # G+4971  U+57F9  uni57F9,  CJK UNIFIED IDEOGRAPH-57F9
        0x5F98, # G+4972  U+5F98  uni5F98,  CJK UNIFIED IDEOGRAPH-5F98
        0x62DC, # G+4973  U+62DC  uni62DC,  CJK UNIFIED IDEOGRAPH-62DC
        0x6392, # G+4974  U+6392  uni6392,  CJK UNIFIED IDEOGRAPH-6392
        0x676F, # G+4975  U+676F  uni676F,  CJK UNIFIED IDEOGRAPH-676F
        0x6E43, # G+4976  U+6E43  uni6E43,  CJK UNIFIED IDEOGRAPH-6E43
        0x7119, # G+4977  U+7119  uni7119,  CJK UNIFIED IDEOGRAPH-7119
        0x76C3, # G+4978  U+76C3  uni76C3,  CJK UNIFIED IDEOGRAPH-76C3
        0x80CC, # G+4979  U+80CC  uni80CC,  CJK UNIFIED IDEOGRAPH-80CC
        0x80DA, # G+4980  U+80DA  uni80DA,  CJK UNIFIED IDEOGRAPH-80DA
        0x88F4, # G+4981  U+88F4  uni88F4,  CJK UNIFIED IDEOGRAPH-88F4
        0x88F5, # G+4982  U+88F5  uni88F5,  CJK UNIFIED IDEOGRAPH-88F5
        0x8919, # G+4983  U+8919  uni8919,  CJK UNIFIED IDEOGRAPH-8919
        0x8CE0, # G+4984  U+8CE0  uni8CE0,  CJK UNIFIED IDEOGRAPH-8CE0
        0x8F29, # G+4985  U+8F29  uni8F29,  CJK UNIFIED IDEOGRAPH-8F29
        0x914D, # G+4986  U+914D  uni914D,  CJK UNIFIED IDEOGRAPH-914D
        0x966A, # G+4987  U+966A  uni966A,  CJK UNIFIED IDEOGRAPH-966A
        0x4F2F, # G+4988  U+4F2F  uni4F2F,  CJK UNIFIED IDEOGRAPH-4F2F
        0x4F70, # G+4989  U+4F70  uni4F70,  CJK UNIFIED IDEOGRAPH-4F70
        0x5E1B, # G+4990  U+5E1B  uni5E1B,  CJK UNIFIED IDEOGRAPH-5E1B
        0x67CF, # G+4991  U+67CF  uni67CF,  CJK UNIFIED IDEOGRAPH-67CF
        0x6822, # G+4992  U+6822  uni6822,  CJK UNIFIED IDEOGRAPH-6822
        0x767D, # G+4993  U+767D  uni767D,  CJK UNIFIED IDEOGRAPH-767D
        0x767E, # G+4994  U+767E  uni767E,  CJK UNIFIED IDEOGRAPH-767E
        0x9B44, # G+4995  U+9B44  uni9B44,  CJK UNIFIED IDEOGRAPH-9B44
        0x5E61, # G+4996  U+5E61  uni5E61,  CJK UNIFIED IDEOGRAPH-5E61
        0x6A0A, # G+4997  U+6A0A  uni6A0A,  CJK UNIFIED IDEOGRAPH-6A0A
        0x7169, # G+4998  U+7169  uni7169,  CJK UNIFIED IDEOGRAPH-7169
        0x71D4, # G+4999  U+71D4  uni71D4,  CJK UNIFIED IDEOGRAPH-71D4
        0x756A, # G+5000  U+756A  uni756A,  CJK UNIFIED IDEOGRAPH-756A
        0x7E41, # G+5001  U+7E41  uni7E41,  CJK UNIFIED IDEOGRAPH-7E41
        0x8543, # G+5002  U+8543  uni8543,  CJK UNIFIED IDEOGRAPH-8543
        0x85E9, # G+5003  U+85E9  uni85E9,  CJK UNIFIED IDEOGRAPH-85E9
        0x98DC, # G+5004  U+98DC  uni98DC,  CJK UNIFIED IDEOGRAPH-98DC
        0x4F10, # G+5005  U+4F10  uni4F10,  CJK UNIFIED IDEOGRAPH-4F10
        0x7B4F, # G+5006  U+7B4F  uni7B4F,  CJK UNIFIED IDEOGRAPH-7B4F
        0x7F70, # G+5007  U+7F70  uni7F70,  CJK UNIFIED IDEOGRAPH-7F70
        0x95A5, # G+5008  U+95A5  uni95A5,  CJK UNIFIED IDEOGRAPH-95A5
        0x51E1, # G+5009  U+51E1  uni51E1,  CJK UNIFIED IDEOGRAPH-51E1
        0x5E06, # G+5010  U+5E06  uni5E06,  CJK UNIFIED IDEOGRAPH-5E06
        0x68B5, # G+5011  U+68B5  uni68B5,  CJK UNIFIED IDEOGRAPH-68B5
        0x6C3E, # G+5012  U+6C3E  uni6C3E,  CJK UNIFIED IDEOGRAPH-6C3E
        0x6C4E, # G+5013  U+6C4E  uni6C4E,  CJK UNIFIED IDEOGRAPH-6C4E
        0x6CDB, # G+5014  U+6CDB  uni6CDB,  CJK UNIFIED IDEOGRAPH-6CDB
        0x72AF, # G+5015  U+72AF  uni72AF,  CJK UNIFIED IDEOGRAPH-72AF
        0x7BC4, # G+5016  U+7BC4  uni7BC4,  CJK UNIFIED IDEOGRAPH-7BC4
        0x8303, # G+5017  U+8303  uni8303,  CJK UNIFIED IDEOGRAPH-8303
        0x6CD5, # G+5018  U+6CD5  uni6CD5,  CJK UNIFIED IDEOGRAPH-6CD5
        0x743A, # G+5019  U+743A  uni743A,  CJK UNIFIED IDEOGRAPH-743A
        0x50FB, # G+5020  U+50FB  uni50FB,  CJK UNIFIED IDEOGRAPH-50FB
        0x5288, # G+5021  U+5288  uni5288,  CJK UNIFIED IDEOGRAPH-5288
        0x58C1, # G+5022  U+58C1  uni58C1,  CJK UNIFIED IDEOGRAPH-58C1
        0x64D8, # G+5023  U+64D8  uni64D8,  CJK UNIFIED IDEOGRAPH-64D8
        0x6A97, # G+5024  U+6A97  uni6A97,  CJK UNIFIED IDEOGRAPH-6A97
        0x74A7, # G+5025  U+74A7  uni74A7,  CJK UNIFIED IDEOGRAPH-74A7
        0x7656, # G+5026  U+7656  uni7656,  CJK UNIFIED IDEOGRAPH-7656
        0x78A7, # G+5027  U+78A7  uni78A7,  CJK UNIFIED IDEOGRAPH-78A7
        0x8617, # G+5028  U+8617  uni8617,  CJK UNIFIED IDEOGRAPH-8617
        0x95E2, # G+5029  U+95E2  uni95E2,  CJK UNIFIED IDEOGRAPH-95E2
        0x9739, # G+5030  U+9739  uni9739,  CJK UNIFIED IDEOGRAPH-9739
        0x535E, # G+5031  U+535E  uni535E,  CJK UNIFIED IDEOGRAPH-535E
        0x5F01, # G+5032  U+5F01  uni5F01,  CJK UNIFIED IDEOGRAPH-5F01
        0x8B8A, # G+5033  U+8B8A  uni8B8A,  CJK UNIFIED IDEOGRAPH-8B8A
        0x8FA8, # G+5034  U+8FA8  uni8FA8,  CJK UNIFIED IDEOGRAPH-8FA8
        0x8FAF, # G+5035  U+8FAF  uni8FAF,  CJK UNIFIED IDEOGRAPH-8FAF
        0x908A, # G+5036  U+908A  uni908A,  CJK UNIFIED IDEOGRAPH-908A
        0x5225, # G+5037  U+5225  uni5225,  CJK UNIFIED IDEOGRAPH-5225
        0x77A5, # G+5038  U+77A5  uni77A5,  CJK UNIFIED IDEOGRAPH-77A5
        0x9C49, # G+5039  U+9C49  uni9C49,  CJK UNIFIED IDEOGRAPH-9C49
        0x9F08, # G+5040  U+9F08  uni9F08,  CJK UNIFIED IDEOGRAPH-9F08
        0x4E19, # G+5041  U+4E19  uni4E19,  CJK UNIFIED IDEOGRAPH-4E19
        0x5002, # G+5042  U+5002  uni5002,  CJK UNIFIED IDEOGRAPH-5002
        0x5175, # G+5043  U+5175  uni5175,  CJK UNIFIED IDEOGRAPH-5175
        0x5C5B, # G+5044  U+5C5B  uni5C5B,  CJK UNIFIED IDEOGRAPH-5C5B
        0x5E77, # G+5045  U+5E77  uni5E77,  CJK UNIFIED IDEOGRAPH-5E77
        0x661E, # G+5046  U+661E  uni661E,  CJK UNIFIED IDEOGRAPH-661E
        0x663A, # G+5047  U+663A  uni663A,  CJK UNIFIED IDEOGRAPH-663A
        0x67C4, # G+5048  U+67C4  uni67C4,  CJK UNIFIED IDEOGRAPH-67C4
        0x68C5, # G+5049  U+68C5  uni68C5,  CJK UNIFIED IDEOGRAPH-68C5
        0x70B3, # G+5050  U+70B3  uni70B3,  CJK UNIFIED IDEOGRAPH-70B3
        0x7501, # G+5051  U+7501  uni7501,  CJK UNIFIED IDEOGRAPH-7501
        0x75C5, # G+5052  U+75C5  uni75C5,  CJK UNIFIED IDEOGRAPH-75C5
        0x79C9, # G+5053  U+79C9  uni79C9,  CJK UNIFIED IDEOGRAPH-79C9
        0x7ADD, # G+5054  U+7ADD  uni7ADD,  CJK UNIFIED IDEOGRAPH-7ADD
        0x8F27, # G+5055  U+8F27  uni8F27,  CJK UNIFIED IDEOGRAPH-8F27
        0x9920, # G+5056  U+9920  uni9920,  CJK UNIFIED IDEOGRAPH-9920
        0x9A08, # G+5057  U+9A08  uni9A08,  CJK UNIFIED IDEOGRAPH-9A08
        0x4FDD, # G+5058  U+4FDD  uni4FDD,  CJK UNIFIED IDEOGRAPH-4FDD
        0x5821, # G+5059  U+5821  uni5821,  CJK UNIFIED IDEOGRAPH-5821
        0x5831, # G+5060  U+5831  uni5831,  CJK UNIFIED IDEOGRAPH-5831
        0x5BF6, # G+5061  U+5BF6  uni5BF6,  CJK UNIFIED IDEOGRAPH-5BF6
        0x666E, # G+5062  U+666E  uni666E,  CJK UNIFIED IDEOGRAPH-666E
        0x6B65, # G+5063  U+6B65  uni6B65,  CJK UNIFIED IDEOGRAPH-6B65
        0x6D11, # G+5064  U+6D11  uni6D11,  CJK UNIFIED IDEOGRAPH-6D11
        0x6E7A, # G+5065  U+6E7A  uni6E7A,  CJK UNIFIED IDEOGRAPH-6E7A
        0x6F7D, # G+5066  U+6F7D  uni6F7D,  CJK UNIFIED IDEOGRAPH-6F7D
        0x73E4, # G+5067  U+73E4  uni73E4,  CJK UNIFIED IDEOGRAPH-73E4
        0x752B, # G+5068  U+752B  uni752B,  CJK UNIFIED IDEOGRAPH-752B
        0x83E9, # G+5069  U+83E9  uni83E9,  CJK UNIFIED IDEOGRAPH-83E9
        0x88DC, # G+5070  U+88DC  uni88DC,  CJK UNIFIED IDEOGRAPH-88DC
        0x8913, # G+5071  U+8913  uni8913,  CJK UNIFIED IDEOGRAPH-8913
        0x8B5C, # G+5072  U+8B5C  uni8B5C,  CJK UNIFIED IDEOGRAPH-8B5C
        0x8F14, # G+5073  U+8F14  uni8F14,  CJK UNIFIED IDEOGRAPH-8F14
        0x4F0F, # G+5074  U+4F0F  uni4F0F,  CJK UNIFIED IDEOGRAPH-4F0F
        0x50D5, # G+5075  U+50D5  uni50D5,  CJK UNIFIED IDEOGRAPH-50D5
        0x5310, # G+5076  U+5310  uni5310,  CJK UNIFIED IDEOGRAPH-5310
        0x535C, # G+5077  U+535C  uni535C,  CJK UNIFIED IDEOGRAPH-535C
        0x5B93, # G+5078  U+5B93  uni5B93,  CJK UNIFIED IDEOGRAPH-5B93
        0x5FA9, # G+5079  U+5FA9  uni5FA9,  CJK UNIFIED IDEOGRAPH-5FA9
        0x670D, # G+5080  U+670D  uni670D,  CJK UNIFIED IDEOGRAPH-670D
        0x798F, # G+5081  U+798F  uni798F,  CJK UNIFIED IDEOGRAPH-798F
        0x8179, # G+5082  U+8179  uni8179,  CJK UNIFIED IDEOGRAPH-8179
        0x832F, # G+5083  U+832F  uni832F,  CJK UNIFIED IDEOGRAPH-832F
        0x8514, # G+5084  U+8514  uni8514,  CJK UNIFIED IDEOGRAPH-8514
        0x8907, # G+5085  U+8907  uni8907,  CJK UNIFIED IDEOGRAPH-8907
        0x8986, # G+5086  U+8986  uni8986,  CJK UNIFIED IDEOGRAPH-8986
        0x8F39, # G+5087  U+8F39  uni8F39,  CJK UNIFIED IDEOGRAPH-8F39
        0x8F3B, # G+5088  U+8F3B  uni8F3B,  CJK UNIFIED IDEOGRAPH-8F3B
        0x99A5, # G+5089  U+99A5  uni99A5,  CJK UNIFIED IDEOGRAPH-99A5
        0x9C12, # G+5090  U+9C12  uni9C12,  CJK UNIFIED IDEOGRAPH-9C12
        0x672C, # G+5091  U+672C  uni672C,  CJK UNIFIED IDEOGRAPH-672C
        0x4E76, # G+5092  U+4E76  uni4E76,  CJK UNIFIED IDEOGRAPH-4E76
        0x4FF8, # G+5093  U+4FF8  uni4FF8,  CJK UNIFIED IDEOGRAPH-4FF8
        0x5949, # G+5094  U+5949  uni5949,  CJK UNIFIED IDEOGRAPH-5949
        0x5C01, # G+5095  U+5C01  uni5C01,  CJK UNIFIED IDEOGRAPH-5C01
        0x5CEF, # G+5096  U+5CEF  uni5CEF,  CJK UNIFIED IDEOGRAPH-5CEF
        0x5CF0, # G+5097  U+5CF0  uni5CF0,  CJK UNIFIED IDEOGRAPH-5CF0
        0x6367, # G+5098  U+6367  uni6367,  CJK UNIFIED IDEOGRAPH-6367
        0x68D2, # G+5099  U+68D2  uni68D2,  CJK UNIFIED IDEOGRAPH-68D2
        0x70FD, # G+5100  U+70FD  uni70FD,  CJK UNIFIED IDEOGRAPH-70FD
        0x71A2, # G+5101  U+71A2  uni71A2,  CJK UNIFIED IDEOGRAPH-71A2
        0x742B, # G+5102  U+742B  uni742B,  CJK UNIFIED IDEOGRAPH-742B
        0x7E2B, # G+5103  U+7E2B  uni7E2B,  CJK UNIFIED IDEOGRAPH-7E2B
        0x84EC, # G+5104  U+84EC  uni84EC,  CJK UNIFIED IDEOGRAPH-84EC
        0x8702, # G+5105  U+8702  uni8702,  CJK UNIFIED IDEOGRAPH-8702
        0x9022, # G+5106  U+9022  uni9022,  CJK UNIFIED IDEOGRAPH-9022
        0x92D2, # G+5107  U+92D2  uni92D2,  CJK UNIFIED IDEOGRAPH-92D2
        0x9CF3, # G+5108  U+9CF3  uni9CF3,  CJK UNIFIED IDEOGRAPH-9CF3
        0x4E0D, # G+5109  U+4E0D  uni4E0D,  CJK UNIFIED IDEOGRAPH-4E0D
        0x4ED8, # G+5110  U+4ED8  uni4ED8,  CJK UNIFIED IDEOGRAPH-4ED8
        0x4FEF, # G+5111  U+4FEF  uni4FEF,  CJK UNIFIED IDEOGRAPH-4FEF
        0x5085, # G+5112  U+5085  uni5085,  CJK UNIFIED IDEOGRAPH-5085
        0x5256, # G+5113  U+5256  uni5256,  CJK UNIFIED IDEOGRAPH-5256
        0x526F, # G+5114  U+526F  uni526F,  CJK UNIFIED IDEOGRAPH-526F
        0x5426, # G+5115  U+5426  uni5426,  CJK UNIFIED IDEOGRAPH-5426
        0x5490, # G+5116  U+5490  uni5490,  CJK UNIFIED IDEOGRAPH-5490
        0x57E0, # G+5117  U+57E0  uni57E0,  CJK UNIFIED IDEOGRAPH-57E0
        0x592B, # G+5118  U+592B  uni592B,  CJK UNIFIED IDEOGRAPH-592B
        0x5A66, # G+5119  U+5A66  uni5A66,  CJK UNIFIED IDEOGRAPH-5A66
        0x5B5A, # G+5120  U+5B5A  uni5B5A,  CJK UNIFIED IDEOGRAPH-5B5A
        0x5B75, # G+5121  U+5B75  uni5B75,  CJK UNIFIED IDEOGRAPH-5B75
        0x5BCC, # G+5122  U+5BCC  uni5BCC,  CJK UNIFIED IDEOGRAPH-5BCC
        0x5E9C, # G+5123  U+5E9C  uni5E9C,  CJK UNIFIED IDEOGRAPH-5E9C
        0x6276, # G+5124  U+6276  uni6276,  CJK UNIFIED IDEOGRAPH-6276
        0x6577, # G+5125  U+6577  uni6577,  CJK UNIFIED IDEOGRAPH-6577
        0x65A7, # G+5126  U+65A7  uni65A7,  CJK UNIFIED IDEOGRAPH-65A7
        0x6D6E, # G+5127  U+6D6E  uni6D6E,  CJK UNIFIED IDEOGRAPH-6D6E
        0x6EA5, # G+5128  U+6EA5  uni6EA5,  CJK UNIFIED IDEOGRAPH-6EA5
        0x7236, # G+5129  U+7236  uni7236,  CJK UNIFIED IDEOGRAPH-7236
        0x7B26, # G+5130  U+7B26  uni7B26,  CJK UNIFIED IDEOGRAPH-7B26
        0x7C3F, # G+5131  U+7C3F  uni7C3F,  CJK UNIFIED IDEOGRAPH-7C3F
        0x7F36, # G+5132  U+7F36  uni7F36,  CJK UNIFIED IDEOGRAPH-7F36
        0x8150, # G+5133  U+8150  uni8150,  CJK UNIFIED IDEOGRAPH-8150
        0x8151, # G+5134  U+8151  uni8151,  CJK UNIFIED IDEOGRAPH-8151
        0x819A, # G+5135  U+819A  uni819A,  CJK UNIFIED IDEOGRAPH-819A
        0x8240, # G+5136  U+8240  uni8240,  CJK UNIFIED IDEOGRAPH-8240
        0x8299, # G+5137  U+8299  uni8299,  CJK UNIFIED IDEOGRAPH-8299
        0x83A9, # G+5138  U+83A9  uni83A9,  CJK UNIFIED IDEOGRAPH-83A9
        0x8A03, # G+5139  U+8A03  uni8A03,  CJK UNIFIED IDEOGRAPH-8A03
        0x8CA0, # G+5140  U+8CA0  uni8CA0,  CJK UNIFIED IDEOGRAPH-8CA0
        0x8CE6, # G+5141  U+8CE6  uni8CE6,  CJK UNIFIED IDEOGRAPH-8CE6
        0x8CFB, # G+5142  U+8CFB  uni8CFB,  CJK UNIFIED IDEOGRAPH-8CFB
        0x8D74, # G+5143  U+8D74  uni8D74,  CJK UNIFIED IDEOGRAPH-8D74
        0x8DBA, # G+5144  U+8DBA  uni8DBA,  CJK UNIFIED IDEOGRAPH-8DBA
        0x90E8, # G+5145  U+90E8  uni90E8,  CJK UNIFIED IDEOGRAPH-90E8
        0x91DC, # G+5146  U+91DC  uni91DC,  CJK UNIFIED IDEOGRAPH-91DC
        0x961C, # G+5147  U+961C  uni961C,  CJK UNIFIED IDEOGRAPH-961C
        0x9644, # G+5148  U+9644  uni9644,  CJK UNIFIED IDEOGRAPH-9644
        0x99D9, # G+5149  U+99D9  uni99D9,  CJK UNIFIED IDEOGRAPH-99D9
        0x9CE7, # G+5150  U+9CE7  uni9CE7,  CJK UNIFIED IDEOGRAPH-9CE7
        0xF963, # G+5151  U+F963  uniF963,  CJK COMPATIBILITY IDEOGRAPH-F963
        0x5206, # G+5152  U+5206  uni5206,  CJK UNIFIED IDEOGRAPH-5206
        0x5429, # G+5153  U+5429  uni5429,  CJK UNIFIED IDEOGRAPH-5429
        0x5674, # G+5154  U+5674  uni5674,  CJK UNIFIED IDEOGRAPH-5674
        0x58B3, # G+5155  U+58B3  uni58B3,  CJK UNIFIED IDEOGRAPH-58B3
        0x5954, # G+5156  U+5954  uni5954,  CJK UNIFIED IDEOGRAPH-5954
        0x596E, # G+5157  U+596E  uni596E,  CJK UNIFIED IDEOGRAPH-596E
        0x5FFF, # G+5158  U+5FFF  uni5FFF,  CJK UNIFIED IDEOGRAPH-5FFF
        0x61A4, # G+5159  U+61A4  uni61A4,  CJK UNIFIED IDEOGRAPH-61A4
        0x626E, # G+5160  U+626E  uni626E,  CJK UNIFIED IDEOGRAPH-626E
        0x6610, # G+5161  U+6610  uni6610,  CJK UNIFIED IDEOGRAPH-6610
        0x6C7E, # G+5162  U+6C7E  uni6C7E,  CJK UNIFIED IDEOGRAPH-6C7E
        0x711A, # G+5163  U+711A  uni711A,  CJK UNIFIED IDEOGRAPH-711A
        0x76C6, # G+5164  U+76C6  uni76C6,  CJK UNIFIED IDEOGRAPH-76C6
        0x7C89, # G+5165  U+7C89  uni7C89,  CJK UNIFIED IDEOGRAPH-7C89
        0x7CDE, # G+5166  U+7CDE  uni7CDE,  CJK UNIFIED IDEOGRAPH-7CDE
        0x7D1B, # G+5167  U+7D1B  uni7D1B,  CJK UNIFIED IDEOGRAPH-7D1B
        0x82AC, # G+5168  U+82AC  uni82AC,  CJK UNIFIED IDEOGRAPH-82AC
        0x8CC1, # G+5169  U+8CC1  uni8CC1,  CJK UNIFIED IDEOGRAPH-8CC1
        0x96F0, # G+5170  U+96F0  uni96F0,  CJK UNIFIED IDEOGRAPH-96F0
        0x4F5B, # G+5171  U+4F5B  uni4F5B,  CJK UNIFIED IDEOGRAPH-4F5B
        0x5F17, # G+5172  U+5F17  uni5F17,  CJK UNIFIED IDEOGRAPH-5F17
        0x5F7F, # G+5173  U+5F7F  uni5F7F,  CJK UNIFIED IDEOGRAPH-5F7F
        0x62C2, # G+5174  U+62C2  uni62C2,  CJK UNIFIED IDEOGRAPH-62C2
        0x5D29, # G+5175  U+5D29  uni5D29,  CJK UNIFIED IDEOGRAPH-5D29
        0x670B, # G+5176  U+670B  uni670B,  CJK UNIFIED IDEOGRAPH-670B
        0x68DA, # G+5177  U+68DA  uni68DA,  CJK UNIFIED IDEOGRAPH-68DA
        0x787C, # G+5178  U+787C  uni787C,  CJK UNIFIED IDEOGRAPH-787C
        0x7E43, # G+5179  U+7E43  uni7E43,  CJK UNIFIED IDEOGRAPH-7E43
        0x9D6C, # G+5180  U+9D6C  uni9D6C,  CJK UNIFIED IDEOGRAPH-9D6C
        0x4E15, # G+5181  U+4E15  uni4E15,  CJK UNIFIED IDEOGRAPH-4E15
        0x5099, # G+5182  U+5099  uni5099,  CJK UNIFIED IDEOGRAPH-5099
        0x5315, # G+5183  U+5315  uni5315,  CJK UNIFIED IDEOGRAPH-5315
        0x532A, # G+5184  U+532A  uni532A,  CJK UNIFIED IDEOGRAPH-532A
        0x5351, # G+5185  U+5351  uni5351,  CJK UNIFIED IDEOGRAPH-5351
        0x5983, # G+5186  U+5983  uni5983,  CJK UNIFIED IDEOGRAPH-5983
        0x5A62, # G+5187  U+5A62  uni5A62,  CJK UNIFIED IDEOGRAPH-5A62
        0x5E87, # G+5188  U+5E87  uni5E87,  CJK UNIFIED IDEOGRAPH-5E87
        0x60B2, # G+5189  U+60B2  uni60B2,  CJK UNIFIED IDEOGRAPH-60B2
        0x618A, # G+5190  U+618A  uni618A,  CJK UNIFIED IDEOGRAPH-618A
        0x6249, # G+5191  U+6249  uni6249,  CJK UNIFIED IDEOGRAPH-6249
        0x6279, # G+5192  U+6279  uni6279,  CJK UNIFIED IDEOGRAPH-6279
        0x6590, # G+5193  U+6590  uni6590,  CJK UNIFIED IDEOGRAPH-6590
        0x6787, # G+5194  U+6787  uni6787,  CJK UNIFIED IDEOGRAPH-6787
        0x69A7, # G+5195  U+69A7  uni69A7,  CJK UNIFIED IDEOGRAPH-69A7
        0x6BD4, # G+5196  U+6BD4  uni6BD4,  CJK UNIFIED IDEOGRAPH-6BD4
        0x6BD6, # G+5197  U+6BD6  uni6BD6,  CJK UNIFIED IDEOGRAPH-6BD6
        0x6BD7, # G+5198  U+6BD7  uni6BD7,  CJK UNIFIED IDEOGRAPH-6BD7
        0x6BD8, # G+5199  U+6BD8  uni6BD8,  CJK UNIFIED IDEOGRAPH-6BD8
        0x6CB8, # G+5200  U+6CB8  uni6CB8,  CJK UNIFIED IDEOGRAPH-6CB8
        0x7435, # G+5201  U+7435  uni7435,  CJK UNIFIED IDEOGRAPH-7435
        0x75FA, # G+5202  U+75FA  uni75FA,  CJK UNIFIED IDEOGRAPH-75FA
        0x7812, # G+5203  U+7812  uni7812,  CJK UNIFIED IDEOGRAPH-7812
        0x7891, # G+5204  U+7891  uni7891,  CJK UNIFIED IDEOGRAPH-7891
        0x79D5, # G+5205  U+79D5  uni79D5,  CJK UNIFIED IDEOGRAPH-79D5
        0x79D8, # G+5206  U+79D8  uni79D8,  CJK UNIFIED IDEOGRAPH-79D8
        0x7C83, # G+5207  U+7C83  uni7C83,  CJK UNIFIED IDEOGRAPH-7C83
        0x7DCB, # G+5208  U+7DCB  uni7DCB,  CJK UNIFIED IDEOGRAPH-7DCB
        0x7FE1, # G+5209  U+7FE1  uni7FE1,  CJK UNIFIED IDEOGRAPH-7FE1
        0x80A5, # G+5210  U+80A5  uni80A5,  CJK UNIFIED IDEOGRAPH-80A5
        0x813E, # G+5211  U+813E  uni813E,  CJK UNIFIED IDEOGRAPH-813E
        0x81C2, # G+5212  U+81C2  uni81C2,  CJK UNIFIED IDEOGRAPH-81C2
        0x83F2, # G+5213  U+83F2  uni83F2,  CJK UNIFIED IDEOGRAPH-83F2
        0x871A, # G+5214  U+871A  uni871A,  CJK UNIFIED IDEOGRAPH-871A
        0x88E8, # G+5215  U+88E8  uni88E8,  CJK UNIFIED IDEOGRAPH-88E8
        0x8AB9, # G+5216  U+8AB9  uni8AB9,  CJK UNIFIED IDEOGRAPH-8AB9
        0x8B6C, # G+5217  U+8B6C  uni8B6C,  CJK UNIFIED IDEOGRAPH-8B6C
        0x8CBB, # G+5218  U+8CBB  uni8CBB,  CJK UNIFIED IDEOGRAPH-8CBB
        0x9119, # G+5219  U+9119  uni9119,  CJK UNIFIED IDEOGRAPH-9119
        0x975E, # G+5220  U+975E  uni975E,  CJK UNIFIED IDEOGRAPH-975E
        0x98DB, # G+5221  U+98DB  uni98DB,  CJK UNIFIED IDEOGRAPH-98DB
        0x9F3B, # G+5222  U+9F3B  uni9F3B,  CJK UNIFIED IDEOGRAPH-9F3B
        0x56AC, # G+5223  U+56AC  uni56AC,  CJK UNIFIED IDEOGRAPH-56AC
        0x5B2A, # G+5224  U+5B2A  uni5B2A,  CJK UNIFIED IDEOGRAPH-5B2A
        0x5F6C, # G+5225  U+5F6C  uni5F6C,  CJK UNIFIED IDEOGRAPH-5F6C
        0x658C, # G+5226  U+658C  uni658C,  CJK UNIFIED IDEOGRAPH-658C
        0x6AB3, # G+5227  U+6AB3  uni6AB3,  CJK UNIFIED IDEOGRAPH-6AB3
        0x6BAF, # G+5228  U+6BAF  uni6BAF,  CJK UNIFIED IDEOGRAPH-6BAF
        0x6D5C, # G+5229  U+6D5C  uni6D5C,  CJK UNIFIED IDEOGRAPH-6D5C
        0x6FF1, # G+5230  U+6FF1  uni6FF1,  CJK UNIFIED IDEOGRAPH-6FF1
        0x7015, # G+5231  U+7015  uni7015,  CJK UNIFIED IDEOGRAPH-7015
        0x725D, # G+5232  U+725D  uni725D,  CJK UNIFIED IDEOGRAPH-725D
        0x73AD, # G+5233  U+73AD  uni73AD,  CJK UNIFIED IDEOGRAPH-73AD
        0x8CA7, # G+5234  U+8CA7  uni8CA7,  CJK UNIFIED IDEOGRAPH-8CA7
        0x8CD3, # G+5235  U+8CD3  uni8CD3,  CJK UNIFIED IDEOGRAPH-8CD3
        0x983B, # G+5236  U+983B  uni983B,  CJK UNIFIED IDEOGRAPH-983B
        0x6191, # G+5237  U+6191  uni6191,  CJK UNIFIED IDEOGRAPH-6191
        0x6C37, # G+5238  U+6C37  uni6C37,  CJK UNIFIED IDEOGRAPH-6C37
        0x8058, # G+5239  U+8058  uni8058,  CJK UNIFIED IDEOGRAPH-8058
        0x9A01, # G+5240  U+9A01  uni9A01,  CJK UNIFIED IDEOGRAPH-9A01
        0x4E4D, # G+5241  U+4E4D  uni4E4D,  CJK UNIFIED IDEOGRAPH-4E4D
        0x4E8B, # G+5242  U+4E8B  uni4E8B,  CJK UNIFIED IDEOGRAPH-4E8B
        0x4E9B, # G+5243  U+4E9B  uni4E9B,  CJK UNIFIED IDEOGRAPH-4E9B
        0x4ED5, # G+5244  U+4ED5  uni4ED5,  CJK UNIFIED IDEOGRAPH-4ED5
        0x4F3A, # G+5245  U+4F3A  uni4F3A,  CJK UNIFIED IDEOGRAPH-4F3A
        0x4F3C, # G+5246  U+4F3C  uni4F3C,  CJK UNIFIED IDEOGRAPH-4F3C
        0x4F7F, # G+5247  U+4F7F  uni4F7F,  CJK UNIFIED IDEOGRAPH-4F7F
        0x4FDF, # G+5248  U+4FDF  uni4FDF,  CJK UNIFIED IDEOGRAPH-4FDF
        0x50FF, # G+5249  U+50FF  uni50FF,  CJK UNIFIED IDEOGRAPH-50FF
        0x53F2, # G+5250  U+53F2  uni53F2,  CJK UNIFIED IDEOGRAPH-53F2
        0x53F8, # G+5251  U+53F8  uni53F8,  CJK UNIFIED IDEOGRAPH-53F8
        0x5506, # G+5252  U+5506  uni5506,  CJK UNIFIED IDEOGRAPH-5506
        0x55E3, # G+5253  U+55E3  uni55E3,  CJK UNIFIED IDEOGRAPH-55E3
        0x56DB, # G+5254  U+56DB  uni56DB,  CJK UNIFIED IDEOGRAPH-56DB
        0x58EB, # G+5255  U+58EB  uni58EB,  CJK UNIFIED IDEOGRAPH-58EB
        0x5962, # G+5256  U+5962  uni5962,  CJK UNIFIED IDEOGRAPH-5962
        0x5A11, # G+5257  U+5A11  uni5A11,  CJK UNIFIED IDEOGRAPH-5A11
        0x5BEB, # G+5258  U+5BEB  uni5BEB,  CJK UNIFIED IDEOGRAPH-5BEB
        0x5BFA, # G+5259  U+5BFA  uni5BFA,  CJK UNIFIED IDEOGRAPH-5BFA
        0x5C04, # G+5260  U+5C04  uni5C04,  CJK UNIFIED IDEOGRAPH-5C04
        0x5DF3, # G+5261  U+5DF3  uni5DF3,  CJK UNIFIED IDEOGRAPH-5DF3
        0x5E2B, # G+5262  U+5E2B  uni5E2B,  CJK UNIFIED IDEOGRAPH-5E2B
        0x5F99, # G+5263  U+5F99  uni5F99,  CJK UNIFIED IDEOGRAPH-5F99
        0x601D, # G+5264  U+601D  uni601D,  CJK UNIFIED IDEOGRAPH-601D
        0x6368, # G+5265  U+6368  uni6368,  CJK UNIFIED IDEOGRAPH-6368
        0x659C, # G+5266  U+659C  uni659C,  CJK UNIFIED IDEOGRAPH-659C
        0x65AF, # G+5267  U+65AF  uni65AF,  CJK UNIFIED IDEOGRAPH-65AF
        0x67F6, # G+5268  U+67F6  uni67F6,  CJK UNIFIED IDEOGRAPH-67F6
        0x67FB, # G+5269  U+67FB  uni67FB,  CJK UNIFIED IDEOGRAPH-67FB
        0x68AD, # G+5270  U+68AD  uni68AD,  CJK UNIFIED IDEOGRAPH-68AD
        0x6B7B, # G+5271  U+6B7B  uni6B7B,  CJK UNIFIED IDEOGRAPH-6B7B
        0x6C99, # G+5272  U+6C99  uni6C99,  CJK UNIFIED IDEOGRAPH-6C99
        0x6CD7, # G+5273  U+6CD7  uni6CD7,  CJK UNIFIED IDEOGRAPH-6CD7
        0x6E23, # G+5274  U+6E23  uni6E23,  CJK UNIFIED IDEOGRAPH-6E23
        0x7009, # G+5275  U+7009  uni7009,  CJK UNIFIED IDEOGRAPH-7009
        0x7345, # G+5276  U+7345  uni7345,  CJK UNIFIED IDEOGRAPH-7345
        0x7802, # G+5277  U+7802  uni7802,  CJK UNIFIED IDEOGRAPH-7802
        0x793E, # G+5278  U+793E  uni793E,  CJK UNIFIED IDEOGRAPH-793E
        0x7940, # G+5279  U+7940  uni7940,  CJK UNIFIED IDEOGRAPH-7940
        0x7960, # G+5280  U+7960  uni7960,  CJK UNIFIED IDEOGRAPH-7960
        0x79C1, # G+5281  U+79C1  uni79C1,  CJK UNIFIED IDEOGRAPH-79C1
        0x7BE9, # G+5282  U+7BE9  uni7BE9,  CJK UNIFIED IDEOGRAPH-7BE9
        0x7D17, # G+5283  U+7D17  uni7D17,  CJK UNIFIED IDEOGRAPH-7D17
        0x7D72, # G+5284  U+7D72  uni7D72,  CJK UNIFIED IDEOGRAPH-7D72
        0x8086, # G+5285  U+8086  uni8086,  CJK UNIFIED IDEOGRAPH-8086
        0x820D, # G+5286  U+820D  uni820D,  CJK UNIFIED IDEOGRAPH-820D
        0x838E, # G+5287  U+838E  uni838E,  CJK UNIFIED IDEOGRAPH-838E
        0x84D1, # G+5288  U+84D1  uni84D1,  CJK UNIFIED IDEOGRAPH-84D1
        0x86C7, # G+5289  U+86C7  uni86C7,  CJK UNIFIED IDEOGRAPH-86C7
        0x88DF, # G+5290  U+88DF  uni88DF,  CJK UNIFIED IDEOGRAPH-88DF
        0x8A50, # G+5291  U+8A50  uni8A50,  CJK UNIFIED IDEOGRAPH-8A50
        0x8A5E, # G+5292  U+8A5E  uni8A5E,  CJK UNIFIED IDEOGRAPH-8A5E
        0x8B1D, # G+5293  U+8B1D  uni8B1D,  CJK UNIFIED IDEOGRAPH-8B1D
        0x8CDC, # G+5294  U+8CDC  uni8CDC,  CJK UNIFIED IDEOGRAPH-8CDC
        0x8D66, # G+5295  U+8D66  uni8D66,  CJK UNIFIED IDEOGRAPH-8D66
        0x8FAD, # G+5296  U+8FAD  uni8FAD,  CJK UNIFIED IDEOGRAPH-8FAD
        0x90AA, # G+5297  U+90AA  uni90AA,  CJK UNIFIED IDEOGRAPH-90AA
        0x98FC, # G+5298  U+98FC  uni98FC,  CJK UNIFIED IDEOGRAPH-98FC
        0x99DF, # G+5299  U+99DF  uni99DF,  CJK UNIFIED IDEOGRAPH-99DF
        0x9E9D, # G+5300  U+9E9D  uni9E9D,  CJK UNIFIED IDEOGRAPH-9E9D
        0x524A, # G+5301  U+524A  uni524A,  CJK UNIFIED IDEOGRAPH-524A
        0x6714, # G+5302  U+6714  uni6714,  CJK UNIFIED IDEOGRAPH-6714
        0x5098, # G+5303  U+5098  uni5098,  CJK UNIFIED IDEOGRAPH-5098
        0x522A, # G+5304  U+522A  uni522A,  CJK UNIFIED IDEOGRAPH-522A
        0x5C71, # G+5305  U+5C71  uni5C71,  CJK UNIFIED IDEOGRAPH-5C71
        0x6563, # G+5306  U+6563  uni6563,  CJK UNIFIED IDEOGRAPH-6563
        0x6C55, # G+5307  U+6C55  uni6C55,  CJK UNIFIED IDEOGRAPH-6C55
        0x73CA, # G+5308  U+73CA  uni73CA,  CJK UNIFIED IDEOGRAPH-73CA
        0x7523, # G+5309  U+7523  uni7523,  CJK UNIFIED IDEOGRAPH-7523
        0x759D, # G+5310  U+759D  uni759D,  CJK UNIFIED IDEOGRAPH-759D
        0x7B97, # G+5311  U+7B97  uni7B97,  CJK UNIFIED IDEOGRAPH-7B97
        0x849C, # G+5312  U+849C  uni849C,  CJK UNIFIED IDEOGRAPH-849C
        0x9178, # G+5313  U+9178  uni9178,  CJK UNIFIED IDEOGRAPH-9178
        0x9730, # G+5314  U+9730  uni9730,  CJK UNIFIED IDEOGRAPH-9730
        0x4E77, # G+5315  U+4E77  uni4E77,  CJK UNIFIED IDEOGRAPH-4E77
        0x6492, # G+5316  U+6492  uni6492,  CJK UNIFIED IDEOGRAPH-6492
        0x6BBA, # G+5317  U+6BBA  uni6BBA,  CJK UNIFIED IDEOGRAPH-6BBA
        0x715E, # G+5318  U+715E  uni715E,  CJK UNIFIED IDEOGRAPH-715E
        0x85A9, # G+5319  U+85A9  uni85A9,  CJK UNIFIED IDEOGRAPH-85A9
        0x4E09, # G+5320  U+4E09  uni4E09,  CJK UNIFIED IDEOGRAPH-4E09
        0x6749, # G+5321  U+6749  uni6749,  CJK UNIFIED IDEOGRAPH-6749
        0x68EE, # G+5322  U+68EE  uni68EE,  CJK UNIFIED IDEOGRAPH-68EE
        0x6E17, # G+5323  U+6E17  uni6E17,  CJK UNIFIED IDEOGRAPH-6E17
        0x829F, # G+5324  U+829F  uni829F,  CJK UNIFIED IDEOGRAPH-829F
        0x8518, # G+5325  U+8518  uni8518,  CJK UNIFIED IDEOGRAPH-8518
        0x886B, # G+5326  U+886B  uni886B,  CJK UNIFIED IDEOGRAPH-886B
        0x63F7, # G+5327  U+63F7  uni63F7,  CJK UNIFIED IDEOGRAPH-63F7
        0x6F81, # G+5328  U+6F81  uni6F81,  CJK UNIFIED IDEOGRAPH-6F81
        0x9212, # G+5329  U+9212  uni9212,  CJK UNIFIED IDEOGRAPH-9212
        0x98AF, # G+5330  U+98AF  uni98AF,  CJK UNIFIED IDEOGRAPH-98AF
        0x4E0A, # G+5331  U+4E0A  uni4E0A,  CJK UNIFIED IDEOGRAPH-4E0A
        0x50B7, # G+5332  U+50B7  uni50B7,  CJK UNIFIED IDEOGRAPH-50B7
        0x50CF, # G+5333  U+50CF  uni50CF,  CJK UNIFIED IDEOGRAPH-50CF
        0x511F, # G+5334  U+511F  uni511F,  CJK UNIFIED IDEOGRAPH-511F
        0x5546, # G+5335  U+5546  uni5546,  CJK UNIFIED IDEOGRAPH-5546
        0x55AA, # G+5336  U+55AA  uni55AA,  CJK UNIFIED IDEOGRAPH-55AA
        0x5617, # G+5337  U+5617  uni5617,  CJK UNIFIED IDEOGRAPH-5617
        0x5B40, # G+5338  U+5B40  uni5B40,  CJK UNIFIED IDEOGRAPH-5B40
        0x5C19, # G+5339  U+5C19  uni5C19,  CJK UNIFIED IDEOGRAPH-5C19
        0x5CE0, # G+5340  U+5CE0  uni5CE0,  CJK UNIFIED IDEOGRAPH-5CE0
        0x5E38, # G+5341  U+5E38  uni5E38,  CJK UNIFIED IDEOGRAPH-5E38
        0x5E8A, # G+5342  U+5E8A  uni5E8A,  CJK UNIFIED IDEOGRAPH-5E8A
        0x5EA0, # G+5343  U+5EA0  uni5EA0,  CJK UNIFIED IDEOGRAPH-5EA0
        0x5EC2, # G+5344  U+5EC2  uni5EC2,  CJK UNIFIED IDEOGRAPH-5EC2
        0x60F3, # G+5345  U+60F3  uni60F3,  CJK UNIFIED IDEOGRAPH-60F3
        0x6851, # G+5346  U+6851  uni6851,  CJK UNIFIED IDEOGRAPH-6851
        0x6A61, # G+5347  U+6A61  uni6A61,  CJK UNIFIED IDEOGRAPH-6A61
        0x6E58, # G+5348  U+6E58  uni6E58,  CJK UNIFIED IDEOGRAPH-6E58
        0x723D, # G+5349  U+723D  uni723D,  CJK UNIFIED IDEOGRAPH-723D
        0x7240, # G+5350  U+7240  uni7240,  CJK UNIFIED IDEOGRAPH-7240
        0x72C0, # G+5351  U+72C0  uni72C0,  CJK UNIFIED IDEOGRAPH-72C0
        0x76F8, # G+5352  U+76F8  uni76F8,  CJK UNIFIED IDEOGRAPH-76F8
        0x7965, # G+5353  U+7965  uni7965,  CJK UNIFIED IDEOGRAPH-7965
        0x7BB1, # G+5354  U+7BB1  uni7BB1,  CJK UNIFIED IDEOGRAPH-7BB1
        0x7FD4, # G+5355  U+7FD4  uni7FD4,  CJK UNIFIED IDEOGRAPH-7FD4
        0x88F3, # G+5356  U+88F3  uni88F3,  CJK UNIFIED IDEOGRAPH-88F3
        0x89F4, # G+5357  U+89F4  uni89F4,  CJK UNIFIED IDEOGRAPH-89F4
        0x8A73, # G+5358  U+8A73  uni8A73,  CJK UNIFIED IDEOGRAPH-8A73
        0x8C61, # G+5359  U+8C61  uni8C61,  CJK UNIFIED IDEOGRAPH-8C61
        0x8CDE, # G+5360  U+8CDE  uni8CDE,  CJK UNIFIED IDEOGRAPH-8CDE
        0x971C, # G+5361  U+971C  uni971C,  CJK UNIFIED IDEOGRAPH-971C
        0x585E, # G+5362  U+585E  uni585E,  CJK UNIFIED IDEOGRAPH-585E
        0x74BD, # G+5363  U+74BD  uni74BD,  CJK UNIFIED IDEOGRAPH-74BD
        0x8CFD, # G+5364  U+8CFD  uni8CFD,  CJK UNIFIED IDEOGRAPH-8CFD
        0x55C7, # G+5365  U+55C7  uni55C7,  CJK UNIFIED IDEOGRAPH-55C7
        0x7A61, # G+5366  U+7A61  uni7A61,  CJK UNIFIED IDEOGRAPH-7A61
        0xF96A, # G+5367  U+F96A  uniF96A,  CJK COMPATIBILITY IDEOGRAPH-F96A
        0x8272, # G+5368  U+8272  uni8272,  CJK UNIFIED IDEOGRAPH-8272
        0x7272, # G+5369  U+7272  uni7272,  CJK UNIFIED IDEOGRAPH-7272
        0x751F, # G+5370  U+751F  uni751F,  CJK UNIFIED IDEOGRAPH-751F
        0x7525, # G+5371  U+7525  uni7525,  CJK UNIFIED IDEOGRAPH-7525
        0x7B19, # G+5372  U+7B19  uni7B19,  CJK UNIFIED IDEOGRAPH-7B19
        0x5885, # G+5373  U+5885  uni5885,  CJK UNIFIED IDEOGRAPH-5885
        0x58FB, # G+5374  U+58FB  uni58FB,  CJK UNIFIED IDEOGRAPH-58FB
        0x5DBC, # G+5375  U+5DBC  uni5DBC,  CJK UNIFIED IDEOGRAPH-5DBC
        0x5E8F, # G+5376  U+5E8F  uni5E8F,  CJK UNIFIED IDEOGRAPH-5E8F
        0x5EB6, # G+5377  U+5EB6  uni5EB6,  CJK UNIFIED IDEOGRAPH-5EB6
        0x5F90, # G+5378  U+5F90  uni5F90,  CJK UNIFIED IDEOGRAPH-5F90
        0x6055, # G+5379  U+6055  uni6055,  CJK UNIFIED IDEOGRAPH-6055
        0x6292, # G+5380  U+6292  uni6292,  CJK UNIFIED IDEOGRAPH-6292
        0x637F, # G+5381  U+637F  uni637F,  CJK UNIFIED IDEOGRAPH-637F
        0x654D, # G+5382  U+654D  uni654D,  CJK UNIFIED IDEOGRAPH-654D
        0x6691, # G+5383  U+6691  uni6691,  CJK UNIFIED IDEOGRAPH-6691
        0x66D9, # G+5384  U+66D9  uni66D9,  CJK UNIFIED IDEOGRAPH-66D9
        0x66F8, # G+5385  U+66F8  uni66F8,  CJK UNIFIED IDEOGRAPH-66F8
        0x6816, # G+5386  U+6816  uni6816,  CJK UNIFIED IDEOGRAPH-6816
        0x68F2, # G+5387  U+68F2  uni68F2,  CJK UNIFIED IDEOGRAPH-68F2
        0x7280, # G+5388  U+7280  uni7280,  CJK UNIFIED IDEOGRAPH-7280
        0x745E, # G+5389  U+745E  uni745E,  CJK UNIFIED IDEOGRAPH-745E
        0x7B6E, # G+5390  U+7B6E  uni7B6E,  CJK UNIFIED IDEOGRAPH-7B6E
        0x7D6E, # G+5391  U+7D6E  uni7D6E,  CJK UNIFIED IDEOGRAPH-7D6E
        0x7DD6, # G+5392  U+7DD6  uni7DD6,  CJK UNIFIED IDEOGRAPH-7DD6
        0x7F72, # G+5393  U+7F72  uni7F72,  CJK UNIFIED IDEOGRAPH-7F72
        0x80E5, # G+5394  U+80E5  uni80E5,  CJK UNIFIED IDEOGRAPH-80E5
        0x8212, # G+5395  U+8212  uni8212,  CJK UNIFIED IDEOGRAPH-8212
        0x85AF, # G+5396  U+85AF  uni85AF,  CJK UNIFIED IDEOGRAPH-85AF
        0x897F, # G+5397  U+897F  uni897F,  CJK UNIFIED IDEOGRAPH-897F
        0x8A93, # G+5398  U+8A93  uni8A93,  CJK UNIFIED IDEOGRAPH-8A93
        0x901D, # G+5399  U+901D  uni901D,  CJK UNIFIED IDEOGRAPH-901D
        0x92E4, # G+5400  U+92E4  uni92E4,  CJK UNIFIED IDEOGRAPH-92E4
        0x9ECD, # G+5401  U+9ECD  uni9ECD,  CJK UNIFIED IDEOGRAPH-9ECD
        0x9F20, # G+5402  U+9F20  uni9F20,  CJK UNIFIED IDEOGRAPH-9F20
        0x5915, # G+5403  U+5915  uni5915,  CJK UNIFIED IDEOGRAPH-5915
        0x596D, # G+5404  U+596D  uni596D,  CJK UNIFIED IDEOGRAPH-596D
        0x5E2D, # G+5405  U+5E2D  uni5E2D,  CJK UNIFIED IDEOGRAPH-5E2D
        0x60DC, # G+5406  U+60DC  uni60DC,  CJK UNIFIED IDEOGRAPH-60DC
        0x6614, # G+5407  U+6614  uni6614,  CJK UNIFIED IDEOGRAPH-6614
        0x6673, # G+5408  U+6673  uni6673,  CJK UNIFIED IDEOGRAPH-6673
        0x6790, # G+5409  U+6790  uni6790,  CJK UNIFIED IDEOGRAPH-6790
        0x6C50, # G+5410  U+6C50  uni6C50,  CJK UNIFIED IDEOGRAPH-6C50
        0x6DC5, # G+5411  U+6DC5  uni6DC5,  CJK UNIFIED IDEOGRAPH-6DC5
        0x6F5F, # G+5412  U+6F5F  uni6F5F,  CJK UNIFIED IDEOGRAPH-6F5F
        0x77F3, # G+5413  U+77F3  uni77F3,  CJK UNIFIED IDEOGRAPH-77F3
        0x78A9, # G+5414  U+78A9  uni78A9,  CJK UNIFIED IDEOGRAPH-78A9
        0x84C6, # G+5415  U+84C6  uni84C6,  CJK UNIFIED IDEOGRAPH-84C6
        0x91CB, # G+5416  U+91CB  uni91CB,  CJK UNIFIED IDEOGRAPH-91CB
        0x932B, # G+5417  U+932B  uni932B,  CJK UNIFIED IDEOGRAPH-932B
        0x4ED9, # G+5418  U+4ED9  uni4ED9,  CJK UNIFIED IDEOGRAPH-4ED9
        0x50CA, # G+5419  U+50CA  uni50CA,  CJK UNIFIED IDEOGRAPH-50CA
        0x5148, # G+5420  U+5148  uni5148,  CJK UNIFIED IDEOGRAPH-5148
        0x5584, # G+5421  U+5584  uni5584,  CJK UNIFIED IDEOGRAPH-5584
        0x5B0B, # G+5422  U+5B0B  uni5B0B,  CJK UNIFIED IDEOGRAPH-5B0B
        0x5BA3, # G+5423  U+5BA3  uni5BA3,  CJK UNIFIED IDEOGRAPH-5BA3
        0x6247, # G+5424  U+6247  uni6247,  CJK UNIFIED IDEOGRAPH-6247
        0x657E, # G+5425  U+657E  uni657E,  CJK UNIFIED IDEOGRAPH-657E
        0x65CB, # G+5426  U+65CB  uni65CB,  CJK UNIFIED IDEOGRAPH-65CB
        0x6E32, # G+5427  U+6E32  uni6E32,  CJK UNIFIED IDEOGRAPH-6E32
        0x717D, # G+5428  U+717D  uni717D,  CJK UNIFIED IDEOGRAPH-717D
        0x7401, # G+5429  U+7401  uni7401,  CJK UNIFIED IDEOGRAPH-7401
        0x7444, # G+5430  U+7444  uni7444,  CJK UNIFIED IDEOGRAPH-7444
        0x7487, # G+5431  U+7487  uni7487,  CJK UNIFIED IDEOGRAPH-7487
        0x74BF, # G+5432  U+74BF  uni74BF,  CJK UNIFIED IDEOGRAPH-74BF
        0x766C, # G+5433  U+766C  uni766C,  CJK UNIFIED IDEOGRAPH-766C
        0x79AA, # G+5434  U+79AA  uni79AA,  CJK UNIFIED IDEOGRAPH-79AA
        0x7DDA, # G+5435  U+7DDA  uni7DDA,  CJK UNIFIED IDEOGRAPH-7DDA
        0x7E55, # G+5436  U+7E55  uni7E55,  CJK UNIFIED IDEOGRAPH-7E55
        0x7FA8, # G+5437  U+7FA8  uni7FA8,  CJK UNIFIED IDEOGRAPH-7FA8
        0x817A, # G+5438  U+817A  uni817A,  CJK UNIFIED IDEOGRAPH-817A
        0x81B3, # G+5439  U+81B3  uni81B3,  CJK UNIFIED IDEOGRAPH-81B3
        0x8239, # G+5440  U+8239  uni8239,  CJK UNIFIED IDEOGRAPH-8239
        0x861A, # G+5441  U+861A  uni861A,  CJK UNIFIED IDEOGRAPH-861A
        0x87EC, # G+5442  U+87EC  uni87EC,  CJK UNIFIED IDEOGRAPH-87EC
        0x8A75, # G+5443  U+8A75  uni8A75,  CJK UNIFIED IDEOGRAPH-8A75
        0x8DE3, # G+5444  U+8DE3  uni8DE3,  CJK UNIFIED IDEOGRAPH-8DE3
        0x9078, # G+5445  U+9078  uni9078,  CJK UNIFIED IDEOGRAPH-9078
        0x9291, # G+5446  U+9291  uni9291,  CJK UNIFIED IDEOGRAPH-9291
        0x9425, # G+5447  U+9425  uni9425,  CJK UNIFIED IDEOGRAPH-9425
        0x994D, # G+5448  U+994D  uni994D,  CJK UNIFIED IDEOGRAPH-994D
        0x9BAE, # G+5449  U+9BAE  uni9BAE,  CJK UNIFIED IDEOGRAPH-9BAE
        0x5368, # G+5450  U+5368  uni5368,  CJK UNIFIED IDEOGRAPH-5368
        0x5C51, # G+5451  U+5C51  uni5C51,  CJK UNIFIED IDEOGRAPH-5C51
        0x6954, # G+5452  U+6954  uni6954,  CJK UNIFIED IDEOGRAPH-6954
        0x6CC4, # G+5453  U+6CC4  uni6CC4,  CJK UNIFIED IDEOGRAPH-6CC4
        0x6D29, # G+5454  U+6D29  uni6D29,  CJK UNIFIED IDEOGRAPH-6D29
        0x6E2B, # G+5455  U+6E2B  uni6E2B,  CJK UNIFIED IDEOGRAPH-6E2B
        0x820C, # G+5456  U+820C  uni820C,  CJK UNIFIED IDEOGRAPH-820C
        0x859B, # G+5457  U+859B  uni859B,  CJK UNIFIED IDEOGRAPH-859B
        0x893B, # G+5458  U+893B  uni893B,  CJK UNIFIED IDEOGRAPH-893B
        0x8A2D, # G+5459  U+8A2D  uni8A2D,  CJK UNIFIED IDEOGRAPH-8A2D
        0x8AAA, # G+5460  U+8AAA  uni8AAA,  CJK UNIFIED IDEOGRAPH-8AAA
        0x96EA, # G+5461  U+96EA  uni96EA,  CJK UNIFIED IDEOGRAPH-96EA
        0x9F67, # G+5462  U+9F67  uni9F67,  CJK UNIFIED IDEOGRAPH-9F67
        0x5261, # G+5463  U+5261  uni5261,  CJK UNIFIED IDEOGRAPH-5261
        0x66B9, # G+5464  U+66B9  uni66B9,  CJK UNIFIED IDEOGRAPH-66B9
        0x6BB2, # G+5465  U+6BB2  uni6BB2,  CJK UNIFIED IDEOGRAPH-6BB2
        0x7E96, # G+5466  U+7E96  uni7E96,  CJK UNIFIED IDEOGRAPH-7E96
        0x87FE, # G+5467  U+87FE  uni87FE,  CJK UNIFIED IDEOGRAPH-87FE
        0x8D0D, # G+5468  U+8D0D  uni8D0D,  CJK UNIFIED IDEOGRAPH-8D0D
        0x9583, # G+5469  U+9583  uni9583,  CJK UNIFIED IDEOGRAPH-9583
        0x965D, # G+5470  U+965D  uni965D,  CJK UNIFIED IDEOGRAPH-965D
        0x651D, # G+5471  U+651D  uni651D,  CJK UNIFIED IDEOGRAPH-651D
        0x6D89, # G+5472  U+6D89  uni6D89,  CJK UNIFIED IDEOGRAPH-6D89
        0x71EE, # G+5473  U+71EE  uni71EE,  CJK UNIFIED IDEOGRAPH-71EE
        0x57CE, # G+5474  U+57CE  uni57CE,  CJK UNIFIED IDEOGRAPH-57CE
        0x59D3, # G+5475  U+59D3  uni59D3,  CJK UNIFIED IDEOGRAPH-59D3
        0x5BAC, # G+5476  U+5BAC  uni5BAC,  CJK UNIFIED IDEOGRAPH-5BAC
        0x6027, # G+5477  U+6027  uni6027,  CJK UNIFIED IDEOGRAPH-6027
        0x60FA, # G+5478  U+60FA  uni60FA,  CJK UNIFIED IDEOGRAPH-60FA
        0x6210, # G+5479  U+6210  uni6210,  CJK UNIFIED IDEOGRAPH-6210
        0x661F, # G+5480  U+661F  uni661F,  CJK UNIFIED IDEOGRAPH-661F
        0x665F, # G+5481  U+665F  uni665F,  CJK UNIFIED IDEOGRAPH-665F
        0x7329, # G+5482  U+7329  uni7329,  CJK UNIFIED IDEOGRAPH-7329
        0x73F9, # G+5483  U+73F9  uni73F9,  CJK UNIFIED IDEOGRAPH-73F9
        0x76DB, # G+5484  U+76DB  uni76DB,  CJK UNIFIED IDEOGRAPH-76DB
        0xF96D, # G+5485  U+F96D  uniF96D,  CJK COMPATIBILITY IDEOGRAPH-F96D
        0x7B6C, # G+5486  U+7B6C  uni7B6C,  CJK UNIFIED IDEOGRAPH-7B6C
        0x8056, # G+5487  U+8056  uni8056,  CJK UNIFIED IDEOGRAPH-8056
        0x8072, # G+5488  U+8072  uni8072,  CJK UNIFIED IDEOGRAPH-8072
        0x8165, # G+5489  U+8165  uni8165,  CJK UNIFIED IDEOGRAPH-8165
        0x8AA0, # G+5490  U+8AA0  uni8AA0,  CJK UNIFIED IDEOGRAPH-8AA0
        0x9192, # G+5491  U+9192  uni9192,  CJK UNIFIED IDEOGRAPH-9192
        0x4E16, # G+5492  U+4E16  uni4E16,  CJK UNIFIED IDEOGRAPH-4E16
        0x52E2, # G+5493  U+52E2  uni52E2,  CJK UNIFIED IDEOGRAPH-52E2
        0x6B72, # G+5494  U+6B72  uni6B72,  CJK UNIFIED IDEOGRAPH-6B72
        0x6D17, # G+5495  U+6D17  uni6D17,  CJK UNIFIED IDEOGRAPH-6D17
        0x7A05, # G+5496  U+7A05  uni7A05,  CJK UNIFIED IDEOGRAPH-7A05
        0x7B39, # G+5497  U+7B39  uni7B39,  CJK UNIFIED IDEOGRAPH-7B39
        0x7D30, # G+5498  U+7D30  uni7D30,  CJK UNIFIED IDEOGRAPH-7D30
        0x8CB0, # G+5499  U+8CB0  uni8CB0,  CJK UNIFIED IDEOGRAPH-8CB0
        0x53EC, # G+5500  U+53EC  uni53EC,  CJK UNIFIED IDEOGRAPH-53EC
        0x562F, # G+5501  U+562F  uni562F,  CJK UNIFIED IDEOGRAPH-562F
        0x5851, # G+5502  U+5851  uni5851,  CJK UNIFIED IDEOGRAPH-5851
        0x5BB5, # G+5503  U+5BB5  uni5BB5,  CJK UNIFIED IDEOGRAPH-5BB5
        0x5C0F, # G+5504  U+5C0F  uni5C0F,  CJK UNIFIED IDEOGRAPH-5C0F
        0x5C11, # G+5505  U+5C11  uni5C11,  CJK UNIFIED IDEOGRAPH-5C11
        0x5DE2, # G+5506  U+5DE2  uni5DE2,  CJK UNIFIED IDEOGRAPH-5DE2
        0x6240, # G+5507  U+6240  uni6240,  CJK UNIFIED IDEOGRAPH-6240
        0x6383, # G+5508  U+6383  uni6383,  CJK UNIFIED IDEOGRAPH-6383
        0x6414, # G+5509  U+6414  uni6414,  CJK UNIFIED IDEOGRAPH-6414
        0x662D, # G+5510  U+662D  uni662D,  CJK UNIFIED IDEOGRAPH-662D
        0x68B3, # G+5511  U+68B3  uni68B3,  CJK UNIFIED IDEOGRAPH-68B3
        0x6CBC, # G+5512  U+6CBC  uni6CBC,  CJK UNIFIED IDEOGRAPH-6CBC
        0x6D88, # G+5513  U+6D88  uni6D88,  CJK UNIFIED IDEOGRAPH-6D88
        0x6EAF, # G+5514  U+6EAF  uni6EAF,  CJK UNIFIED IDEOGRAPH-6EAF
        0x701F, # G+5515  U+701F  uni701F,  CJK UNIFIED IDEOGRAPH-701F
        0x70A4, # G+5516  U+70A4  uni70A4,  CJK UNIFIED IDEOGRAPH-70A4
        0x71D2, # G+5517  U+71D2  uni71D2,  CJK UNIFIED IDEOGRAPH-71D2
        0x7526, # G+5518  U+7526  uni7526,  CJK UNIFIED IDEOGRAPH-7526
        0x758F, # G+5519  U+758F  uni758F,  CJK UNIFIED IDEOGRAPH-758F
        0x758E, # G+5520  U+758E  uni758E,  CJK UNIFIED IDEOGRAPH-758E
        0x7619, # G+5521  U+7619  uni7619,  CJK UNIFIED IDEOGRAPH-7619
        0x7B11, # G+5522  U+7B11  uni7B11,  CJK UNIFIED IDEOGRAPH-7B11
        0x7BE0, # G+5523  U+7BE0  uni7BE0,  CJK UNIFIED IDEOGRAPH-7BE0
        0x7C2B, # G+5524  U+7C2B  uni7C2B,  CJK UNIFIED IDEOGRAPH-7C2B
        0x7D20, # G+5525  U+7D20  uni7D20,  CJK UNIFIED IDEOGRAPH-7D20
        0x7D39, # G+5526  U+7D39  uni7D39,  CJK UNIFIED IDEOGRAPH-7D39
        0x852C, # G+5527  U+852C  uni852C,  CJK UNIFIED IDEOGRAPH-852C
        0x856D, # G+5528  U+856D  uni856D,  CJK UNIFIED IDEOGRAPH-856D
        0x8607, # G+5529  U+8607  uni8607,  CJK UNIFIED IDEOGRAPH-8607
        0x8A34, # G+5530  U+8A34  uni8A34,  CJK UNIFIED IDEOGRAPH-8A34
        0x900D, # G+5531  U+900D  uni900D,  CJK UNIFIED IDEOGRAPH-900D
        0x9061, # G+5532  U+9061  uni9061,  CJK UNIFIED IDEOGRAPH-9061
        0x90B5, # G+5533  U+90B5  uni90B5,  CJK UNIFIED IDEOGRAPH-90B5
        0x92B7, # G+5534  U+92B7  uni92B7,  CJK UNIFIED IDEOGRAPH-92B7
        0x97F6, # G+5535  U+97F6  uni97F6,  CJK UNIFIED IDEOGRAPH-97F6
        0x9A37, # G+5536  U+9A37  uni9A37,  CJK UNIFIED IDEOGRAPH-9A37
        0x4FD7, # G+5537  U+4FD7  uni4FD7,  CJK UNIFIED IDEOGRAPH-4FD7
        0x5C6C, # G+5538  U+5C6C  uni5C6C,  CJK UNIFIED IDEOGRAPH-5C6C
        0x675F, # G+5539  U+675F  uni675F,  CJK UNIFIED IDEOGRAPH-675F
        0x6D91, # G+5540  U+6D91  uni6D91,  CJK UNIFIED IDEOGRAPH-6D91
        0x7C9F, # G+5541  U+7C9F  uni7C9F,  CJK UNIFIED IDEOGRAPH-7C9F
        0x7E8C, # G+5542  U+7E8C  uni7E8C,  CJK UNIFIED IDEOGRAPH-7E8C
        0x8B16, # G+5543  U+8B16  uni8B16,  CJK UNIFIED IDEOGRAPH-8B16
        0x8D16, # G+5544  U+8D16  uni8D16,  CJK UNIFIED IDEOGRAPH-8D16
        0x901F, # G+5545  U+901F  uni901F,  CJK UNIFIED IDEOGRAPH-901F
        0x5B6B, # G+5546  U+5B6B  uni5B6B,  CJK UNIFIED IDEOGRAPH-5B6B
        0x5DFD, # G+5547  U+5DFD  uni5DFD,  CJK UNIFIED IDEOGRAPH-5DFD
        0x640D, # G+5548  U+640D  uni640D,  CJK UNIFIED IDEOGRAPH-640D
        0x84C0, # G+5549  U+84C0  uni84C0,  CJK UNIFIED IDEOGRAPH-84C0
        0x905C, # G+5550  U+905C  uni905C,  CJK UNIFIED IDEOGRAPH-905C
        0x98E1, # G+5551  U+98E1  uni98E1,  CJK UNIFIED IDEOGRAPH-98E1
        0xF961, # G+5552  U+F961  uniF961,  CJK COMPATIBILITY IDEOGRAPH-F961
        0x5B8B, # G+5553  U+5B8B  uni5B8B,  CJK UNIFIED IDEOGRAPH-5B8B
        0x609A, # G+5554  U+609A  uni609A,  CJK UNIFIED IDEOGRAPH-609A
        0x677E, # G+5555  U+677E  uni677E,  CJK UNIFIED IDEOGRAPH-677E
        0x6DDE, # G+5556  U+6DDE  uni6DDE,  CJK UNIFIED IDEOGRAPH-6DDE
        0x8A1F, # G+5557  U+8A1F  uni8A1F,  CJK UNIFIED IDEOGRAPH-8A1F
        0x8AA6, # G+5558  U+8AA6  uni8AA6,  CJK UNIFIED IDEOGRAPH-8AA6
        0x9001, # G+5559  U+9001  uni9001,  CJK UNIFIED IDEOGRAPH-9001
        0x980C, # G+5560  U+980C  uni980C,  CJK UNIFIED IDEOGRAPH-980C
        0x5237, # G+5561  U+5237  uni5237,  CJK UNIFIED IDEOGRAPH-5237
        0x7051, # G+5562  U+7051  uni7051,  CJK UNIFIED IDEOGRAPH-7051
        0x788E, # G+5563  U+788E  uni788E,  CJK UNIFIED IDEOGRAPH-788E
        0x9396, # G+5564  U+9396  uni9396,  CJK UNIFIED IDEOGRAPH-9396
        0x8870, # G+5565  U+8870  uni8870,  CJK UNIFIED IDEOGRAPH-8870
        0x91D7, # G+5566  U+91D7  uni91D7,  CJK UNIFIED IDEOGRAPH-91D7
        0x4FEE, # G+5567  U+4FEE  uni4FEE,  CJK UNIFIED IDEOGRAPH-4FEE
        0x53D7, # G+5568  U+53D7  uni53D7,  CJK UNIFIED IDEOGRAPH-53D7
        0x55FD, # G+5569  U+55FD  uni55FD,  CJK UNIFIED IDEOGRAPH-55FD
        0x56DA, # G+5570  U+56DA  uni56DA,  CJK UNIFIED IDEOGRAPH-56DA
        0x5782, # G+5571  U+5782  uni5782,  CJK UNIFIED IDEOGRAPH-5782
        0x58FD, # G+5572  U+58FD  uni58FD,  CJK UNIFIED IDEOGRAPH-58FD
        0x5AC2, # G+5573  U+5AC2  uni5AC2,  CJK UNIFIED IDEOGRAPH-5AC2
        0x5B88, # G+5574  U+5B88  uni5B88,  CJK UNIFIED IDEOGRAPH-5B88
        0x5CAB, # G+5575  U+5CAB  uni5CAB,  CJK UNIFIED IDEOGRAPH-5CAB
        0x5CC0, # G+5576  U+5CC0  uni5CC0,  CJK UNIFIED IDEOGRAPH-5CC0
        0x5E25, # G+5577  U+5E25  uni5E25,  CJK UNIFIED IDEOGRAPH-5E25
        0x6101, # G+5578  U+6101  uni6101,  CJK UNIFIED IDEOGRAPH-6101
        0x620D, # G+5579  U+620D  uni620D,  CJK UNIFIED IDEOGRAPH-620D
        0x624B, # G+5580  U+624B  uni624B,  CJK UNIFIED IDEOGRAPH-624B
        0x6388, # G+5581  U+6388  uni6388,  CJK UNIFIED IDEOGRAPH-6388
        0x641C, # G+5582  U+641C  uni641C,  CJK UNIFIED IDEOGRAPH-641C
        0x6536, # G+5583  U+6536  uni6536,  CJK UNIFIED IDEOGRAPH-6536
        0xF969, # G+5584  U+F969  uniF969,  CJK COMPATIBILITY IDEOGRAPH-F969
        0x6A39, # G+5585  U+6A39  uni6A39,  CJK UNIFIED IDEOGRAPH-6A39
        0x6B8A, # G+5586  U+6B8A  uni6B8A,  CJK UNIFIED IDEOGRAPH-6B8A
        0x6C34, # G+5587  U+6C34  uni6C34,  CJK UNIFIED IDEOGRAPH-6C34
        0x6D19, # G+5588  U+6D19  uni6D19,  CJK UNIFIED IDEOGRAPH-6D19
        0x6F31, # G+5589  U+6F31  uni6F31,  CJK UNIFIED IDEOGRAPH-6F31
        0x71E7, # G+5590  U+71E7  uni71E7,  CJK UNIFIED IDEOGRAPH-71E7
        0x72E9, # G+5591  U+72E9  uni72E9,  CJK UNIFIED IDEOGRAPH-72E9
        0x7378, # G+5592  U+7378  uni7378,  CJK UNIFIED IDEOGRAPH-7378
        0x7407, # G+5593  U+7407  uni7407,  CJK UNIFIED IDEOGRAPH-7407
        0x74B2, # G+5594  U+74B2  uni74B2,  CJK UNIFIED IDEOGRAPH-74B2
        0x7626, # G+5595  U+7626  uni7626,  CJK UNIFIED IDEOGRAPH-7626
        0x7761, # G+5596  U+7761  uni7761,  CJK UNIFIED IDEOGRAPH-7761
        0x79C0, # G+5597  U+79C0  uni79C0,  CJK UNIFIED IDEOGRAPH-79C0
        0x7A57, # G+5598  U+7A57  uni7A57,  CJK UNIFIED IDEOGRAPH-7A57
        0x7AEA, # G+5599  U+7AEA  uni7AEA,  CJK UNIFIED IDEOGRAPH-7AEA
        0x7CB9, # G+5600  U+7CB9  uni7CB9,  CJK UNIFIED IDEOGRAPH-7CB9
        0x7D8F, # G+5601  U+7D8F  uni7D8F,  CJK UNIFIED IDEOGRAPH-7D8F
        0x7DAC, # G+5602  U+7DAC  uni7DAC,  CJK UNIFIED IDEOGRAPH-7DAC
        0x7E61, # G+5603  U+7E61  uni7E61,  CJK UNIFIED IDEOGRAPH-7E61
        0x7F9E, # G+5604  U+7F9E  uni7F9E,  CJK UNIFIED IDEOGRAPH-7F9E
        0x8129, # G+5605  U+8129  uni8129,  CJK UNIFIED IDEOGRAPH-8129
        0x8331, # G+5606  U+8331  uni8331,  CJK UNIFIED IDEOGRAPH-8331
        0x8490, # G+5607  U+8490  uni8490,  CJK UNIFIED IDEOGRAPH-8490
        0x84DA, # G+5608  U+84DA  uni84DA,  CJK UNIFIED IDEOGRAPH-84DA
        0x85EA, # G+5609  U+85EA  uni85EA,  CJK UNIFIED IDEOGRAPH-85EA
        0x8896, # G+5610  U+8896  uni8896,  CJK UNIFIED IDEOGRAPH-8896
        0x8AB0, # G+5611  U+8AB0  uni8AB0,  CJK UNIFIED IDEOGRAPH-8AB0
        0x8B90, # G+5612  U+8B90  uni8B90,  CJK UNIFIED IDEOGRAPH-8B90
        0x8F38, # G+5613  U+8F38  uni8F38,  CJK UNIFIED IDEOGRAPH-8F38
        0x9042, # G+5614  U+9042  uni9042,  CJK UNIFIED IDEOGRAPH-9042
        0x9083, # G+5615  U+9083  uni9083,  CJK UNIFIED IDEOGRAPH-9083
        0x916C, # G+5616  U+916C  uni916C,  CJK UNIFIED IDEOGRAPH-916C
        0x9296, # G+5617  U+9296  uni9296,  CJK UNIFIED IDEOGRAPH-9296
        0x92B9, # G+5618  U+92B9  uni92B9,  CJK UNIFIED IDEOGRAPH-92B9
        0x968B, # G+5619  U+968B  uni968B,  CJK UNIFIED IDEOGRAPH-968B
        0x96A7, # G+5620  U+96A7  uni96A7,  CJK UNIFIED IDEOGRAPH-96A7
        0x96A8, # G+5621  U+96A8  uni96A8,  CJK UNIFIED IDEOGRAPH-96A8
        0x96D6, # G+5622  U+96D6  uni96D6,  CJK UNIFIED IDEOGRAPH-96D6
        0x9700, # G+5623  U+9700  uni9700,  CJK UNIFIED IDEOGRAPH-9700
        0x9808, # G+5624  U+9808  uni9808,  CJK UNIFIED IDEOGRAPH-9808
        0x9996, # G+5625  U+9996  uni9996,  CJK UNIFIED IDEOGRAPH-9996
        0x9AD3, # G+5626  U+9AD3  uni9AD3,  CJK UNIFIED IDEOGRAPH-9AD3
        0x9B1A, # G+5627  U+9B1A  uni9B1A,  CJK UNIFIED IDEOGRAPH-9B1A
        0x53D4, # G+5628  U+53D4  uni53D4,  CJK UNIFIED IDEOGRAPH-53D4
        0x587E, # G+5629  U+587E  uni587E,  CJK UNIFIED IDEOGRAPH-587E
        0x5919, # G+5630  U+5919  uni5919,  CJK UNIFIED IDEOGRAPH-5919
        0x5B70, # G+5631  U+5B70  uni5B70,  CJK UNIFIED IDEOGRAPH-5B70
        0x5BBF, # G+5632  U+5BBF  uni5BBF,  CJK UNIFIED IDEOGRAPH-5BBF
        0x6DD1, # G+5633  U+6DD1  uni6DD1,  CJK UNIFIED IDEOGRAPH-6DD1
        0x6F5A, # G+5634  U+6F5A  uni6F5A,  CJK UNIFIED IDEOGRAPH-6F5A
        0x719F, # G+5635  U+719F  uni719F,  CJK UNIFIED IDEOGRAPH-719F
        0x7421, # G+5636  U+7421  uni7421,  CJK UNIFIED IDEOGRAPH-7421
        0x74B9, # G+5637  U+74B9  uni74B9,  CJK UNIFIED IDEOGRAPH-74B9
        0x8085, # G+5638  U+8085  uni8085,  CJK UNIFIED IDEOGRAPH-8085
        0x83FD, # G+5639  U+83FD  uni83FD,  CJK UNIFIED IDEOGRAPH-83FD
        0x5DE1, # G+5640  U+5DE1  uni5DE1,  CJK UNIFIED IDEOGRAPH-5DE1
        0x5F87, # G+5641  U+5F87  uni5F87,  CJK UNIFIED IDEOGRAPH-5F87
        0x5FAA, # G+5642  U+5FAA  uni5FAA,  CJK UNIFIED IDEOGRAPH-5FAA
        0x6042, # G+5643  U+6042  uni6042,  CJK UNIFIED IDEOGRAPH-6042
        0x65EC, # G+5644  U+65EC  uni65EC,  CJK UNIFIED IDEOGRAPH-65EC
        0x6812, # G+5645  U+6812  uni6812,  CJK UNIFIED IDEOGRAPH-6812
        0x696F, # G+5646  U+696F  uni696F,  CJK UNIFIED IDEOGRAPH-696F
        0x6A53, # G+5647  U+6A53  uni6A53,  CJK UNIFIED IDEOGRAPH-6A53
        0x6B89, # G+5648  U+6B89  uni6B89,  CJK UNIFIED IDEOGRAPH-6B89
        0x6D35, # G+5649  U+6D35  uni6D35,  CJK UNIFIED IDEOGRAPH-6D35
        0x6DF3, # G+5650  U+6DF3  uni6DF3,  CJK UNIFIED IDEOGRAPH-6DF3
        0x73E3, # G+5651  U+73E3  uni73E3,  CJK UNIFIED IDEOGRAPH-73E3
        0x76FE, # G+5652  U+76FE  uni76FE,  CJK UNIFIED IDEOGRAPH-76FE
        0x77AC, # G+5653  U+77AC  uni77AC,  CJK UNIFIED IDEOGRAPH-77AC
        0x7B4D, # G+5654  U+7B4D  uni7B4D,  CJK UNIFIED IDEOGRAPH-7B4D
        0x7D14, # G+5655  U+7D14  uni7D14,  CJK UNIFIED IDEOGRAPH-7D14
        0x8123, # G+5656  U+8123  uni8123,  CJK UNIFIED IDEOGRAPH-8123
        0x821C, # G+5657  U+821C  uni821C,  CJK UNIFIED IDEOGRAPH-821C
        0x8340, # G+5658  U+8340  uni8340,  CJK UNIFIED IDEOGRAPH-8340
        0x84F4, # G+5659  U+84F4  uni84F4,  CJK UNIFIED IDEOGRAPH-84F4
        0x8563, # G+5660  U+8563  uni8563,  CJK UNIFIED IDEOGRAPH-8563
        0x8A62, # G+5661  U+8A62  uni8A62,  CJK UNIFIED IDEOGRAPH-8A62
        0x8AC4, # G+5662  U+8AC4  uni8AC4,  CJK UNIFIED IDEOGRAPH-8AC4
        0x9187, # G+5663  U+9187  uni9187,  CJK UNIFIED IDEOGRAPH-9187
        0x931E, # G+5664  U+931E  uni931E,  CJK UNIFIED IDEOGRAPH-931E
        0x9806, # G+5665  U+9806  uni9806,  CJK UNIFIED IDEOGRAPH-9806
        0x99B4, # G+5666  U+99B4  uni99B4,  CJK UNIFIED IDEOGRAPH-99B4
        0x620C, # G+5667  U+620C  uni620C,  CJK UNIFIED IDEOGRAPH-620C
        0x8853, # G+5668  U+8853  uni8853,  CJK UNIFIED IDEOGRAPH-8853
        0x8FF0, # G+5669  U+8FF0  uni8FF0,  CJK UNIFIED IDEOGRAPH-8FF0
        0x9265, # G+5670  U+9265  uni9265,  CJK UNIFIED IDEOGRAPH-9265
        0x5D07, # G+5671  U+5D07  uni5D07,  CJK UNIFIED IDEOGRAPH-5D07
        0x5D27, # G+5672  U+5D27  uni5D27,  CJK UNIFIED IDEOGRAPH-5D27
        0x5D69, # G+5673  U+5D69  uni5D69,  CJK UNIFIED IDEOGRAPH-5D69
        0x745F, # G+5674  U+745F  uni745F,  CJK UNIFIED IDEOGRAPH-745F
        0x819D, # G+5675  U+819D  uni819D,  CJK UNIFIED IDEOGRAPH-819D
        0x8768, # G+5676  U+8768  uni8768,  CJK UNIFIED IDEOGRAPH-8768
        0x6FD5, # G+5677  U+6FD5  uni6FD5,  CJK UNIFIED IDEOGRAPH-6FD5
        0x62FE, # G+5678  U+62FE  uni62FE,  CJK UNIFIED IDEOGRAPH-62FE
        0x7FD2, # G+5679  U+7FD2  uni7FD2,  CJK UNIFIED IDEOGRAPH-7FD2
        0x8936, # G+5680  U+8936  uni8936,  CJK UNIFIED IDEOGRAPH-8936
        0x8972, # G+5681  U+8972  uni8972,  CJK UNIFIED IDEOGRAPH-8972
        0x4E1E, # G+5682  U+4E1E  uni4E1E,  CJK UNIFIED IDEOGRAPH-4E1E
        0x4E58, # G+5683  U+4E58  uni4E58,  CJK UNIFIED IDEOGRAPH-4E58
        0x50E7, # G+5684  U+50E7  uni50E7,  CJK UNIFIED IDEOGRAPH-50E7
        0x52DD, # G+5685  U+52DD  uni52DD,  CJK UNIFIED IDEOGRAPH-52DD
        0x5347, # G+5686  U+5347  uni5347,  CJK UNIFIED IDEOGRAPH-5347
        0x627F, # G+5687  U+627F  uni627F,  CJK UNIFIED IDEOGRAPH-627F
        0x6607, # G+5688  U+6607  uni6607,  CJK UNIFIED IDEOGRAPH-6607
        0x7E69, # G+5689  U+7E69  uni7E69,  CJK UNIFIED IDEOGRAPH-7E69
        0x8805, # G+5690  U+8805  uni8805,  CJK UNIFIED IDEOGRAPH-8805
        0x965E, # G+5691  U+965E  uni965E,  CJK UNIFIED IDEOGRAPH-965E
        0x4F8D, # G+5692  U+4F8D  uni4F8D,  CJK UNIFIED IDEOGRAPH-4F8D
        0x5319, # G+5693  U+5319  uni5319,  CJK UNIFIED IDEOGRAPH-5319
        0x5636, # G+5694  U+5636  uni5636,  CJK UNIFIED IDEOGRAPH-5636
        0x59CB, # G+5695  U+59CB  uni59CB,  CJK UNIFIED IDEOGRAPH-59CB
        0x5AA4, # G+5696  U+5AA4  uni5AA4,  CJK UNIFIED IDEOGRAPH-5AA4
        0x5C38, # G+5697  U+5C38  uni5C38,  CJK UNIFIED IDEOGRAPH-5C38
        0x5C4E, # G+5698  U+5C4E  uni5C4E,  CJK UNIFIED IDEOGRAPH-5C4E
        0x5C4D, # G+5699  U+5C4D  uni5C4D,  CJK UNIFIED IDEOGRAPH-5C4D
        0x5E02, # G+5700  U+5E02  uni5E02,  CJK UNIFIED IDEOGRAPH-5E02
        0x5F11, # G+5701  U+5F11  uni5F11,  CJK UNIFIED IDEOGRAPH-5F11
        0x6043, # G+5702  U+6043  uni6043,  CJK UNIFIED IDEOGRAPH-6043
        0x65BD, # G+5703  U+65BD  uni65BD,  CJK UNIFIED IDEOGRAPH-65BD
        0x662F, # G+5704  U+662F  uni662F,  CJK UNIFIED IDEOGRAPH-662F
        0x6642, # G+5705  U+6642  uni6642,  CJK UNIFIED IDEOGRAPH-6642
        0x67BE, # G+5706  U+67BE  uni67BE,  CJK UNIFIED IDEOGRAPH-67BE
        0x67F4, # G+5707  U+67F4  uni67F4,  CJK UNIFIED IDEOGRAPH-67F4
        0x731C, # G+5708  U+731C  uni731C,  CJK UNIFIED IDEOGRAPH-731C
        0x77E2, # G+5709  U+77E2  uni77E2,  CJK UNIFIED IDEOGRAPH-77E2
        0x793A, # G+5710  U+793A  uni793A,  CJK UNIFIED IDEOGRAPH-793A
        0x7FC5, # G+5711  U+7FC5  uni7FC5,  CJK UNIFIED IDEOGRAPH-7FC5
        0x8494, # G+5712  U+8494  uni8494,  CJK UNIFIED IDEOGRAPH-8494
        0x84CD, # G+5713  U+84CD  uni84CD,  CJK UNIFIED IDEOGRAPH-84CD
        0x8996, # G+5714  U+8996  uni8996,  CJK UNIFIED IDEOGRAPH-8996
        0x8A66, # G+5715  U+8A66  uni8A66,  CJK UNIFIED IDEOGRAPH-8A66
        0x8A69, # G+5716  U+8A69  uni8A69,  CJK UNIFIED IDEOGRAPH-8A69
        0x8AE1, # G+5717  U+8AE1  uni8AE1,  CJK UNIFIED IDEOGRAPH-8AE1
        0x8C55, # G+5718  U+8C55  uni8C55,  CJK UNIFIED IDEOGRAPH-8C55
        0x8C7A, # G+5719  U+8C7A  uni8C7A,  CJK UNIFIED IDEOGRAPH-8C7A
        0x57F4, # G+5720  U+57F4  uni57F4,  CJK UNIFIED IDEOGRAPH-57F4
        0x5BD4, # G+5721  U+5BD4  uni5BD4,  CJK UNIFIED IDEOGRAPH-5BD4
        0x5F0F, # G+5722  U+5F0F  uni5F0F,  CJK UNIFIED IDEOGRAPH-5F0F
        0x606F, # G+5723  U+606F  uni606F,  CJK UNIFIED IDEOGRAPH-606F
        0x62ED, # G+5724  U+62ED  uni62ED,  CJK UNIFIED IDEOGRAPH-62ED
        0x690D, # G+5725  U+690D  uni690D,  CJK UNIFIED IDEOGRAPH-690D
        0x6B96, # G+5726  U+6B96  uni6B96,  CJK UNIFIED IDEOGRAPH-6B96
        0x6E5C, # G+5727  U+6E5C  uni6E5C,  CJK UNIFIED IDEOGRAPH-6E5C
        0x7184, # G+5728  U+7184  uni7184,  CJK UNIFIED IDEOGRAPH-7184
        0x7BD2, # G+5729  U+7BD2  uni7BD2,  CJK UNIFIED IDEOGRAPH-7BD2
        0x8755, # G+5730  U+8755  uni8755,  CJK UNIFIED IDEOGRAPH-8755
        0x8B58, # G+5731  U+8B58  uni8B58,  CJK UNIFIED IDEOGRAPH-8B58
        0x8EFE, # G+5732  U+8EFE  uni8EFE,  CJK UNIFIED IDEOGRAPH-8EFE
        0x98DF, # G+5733  U+98DF  uni98DF,  CJK UNIFIED IDEOGRAPH-98DF
        0x98FE, # G+5734  U+98FE  uni98FE,  CJK UNIFIED IDEOGRAPH-98FE
        0x4F38, # G+5735  U+4F38  uni4F38,  CJK UNIFIED IDEOGRAPH-4F38
        0x4F81, # G+5736  U+4F81  uni4F81,  CJK UNIFIED IDEOGRAPH-4F81
        0x4FE1, # G+5737  U+4FE1  uni4FE1,  CJK UNIFIED IDEOGRAPH-4FE1
        0x547B, # G+5738  U+547B  uni547B,  CJK UNIFIED IDEOGRAPH-547B
        0x5A20, # G+5739  U+5A20  uni5A20,  CJK UNIFIED IDEOGRAPH-5A20
        0x5BB8, # G+5740  U+5BB8  uni5BB8,  CJK UNIFIED IDEOGRAPH-5BB8
        0x613C, # G+5741  U+613C  uni613C,  CJK UNIFIED IDEOGRAPH-613C
        0x65B0, # G+5742  U+65B0  uni65B0,  CJK UNIFIED IDEOGRAPH-65B0
        0x6668, # G+5743  U+6668  uni6668,  CJK UNIFIED IDEOGRAPH-6668
        0x71FC, # G+5744  U+71FC  uni71FC,  CJK UNIFIED IDEOGRAPH-71FC
        0x7533, # G+5745  U+7533  uni7533,  CJK UNIFIED IDEOGRAPH-7533
        0x795E, # G+5746  U+795E  uni795E,  CJK UNIFIED IDEOGRAPH-795E
        0x7D33, # G+5747  U+7D33  uni7D33,  CJK UNIFIED IDEOGRAPH-7D33
        0x814E, # G+5748  U+814E  uni814E,  CJK UNIFIED IDEOGRAPH-814E
        0x81E3, # G+5749  U+81E3  uni81E3,  CJK UNIFIED IDEOGRAPH-81E3
        0x8398, # G+5750  U+8398  uni8398,  CJK UNIFIED IDEOGRAPH-8398
        0x85AA, # G+5751  U+85AA  uni85AA,  CJK UNIFIED IDEOGRAPH-85AA
        0x85CE, # G+5752  U+85CE  uni85CE,  CJK UNIFIED IDEOGRAPH-85CE
        0x8703, # G+5753  U+8703  uni8703,  CJK UNIFIED IDEOGRAPH-8703
        0x8A0A, # G+5754  U+8A0A  uni8A0A,  CJK UNIFIED IDEOGRAPH-8A0A
        0x8EAB, # G+5755  U+8EAB  uni8EAB,  CJK UNIFIED IDEOGRAPH-8EAB
        0x8F9B, # G+5756  U+8F9B  uni8F9B,  CJK UNIFIED IDEOGRAPH-8F9B
        0x8FC5, # G+5757  U+8FC5  uni8FC5,  CJK UNIFIED IDEOGRAPH-8FC5
        0x5931, # G+5758  U+5931  uni5931,  CJK UNIFIED IDEOGRAPH-5931
        0x5BA4, # G+5759  U+5BA4  uni5BA4,  CJK UNIFIED IDEOGRAPH-5BA4
        0x5BE6, # G+5760  U+5BE6  uni5BE6,  CJK UNIFIED IDEOGRAPH-5BE6
        0x6089, # G+5761  U+6089  uni6089,  CJK UNIFIED IDEOGRAPH-6089
        0x5BE9, # G+5762  U+5BE9  uni5BE9,  CJK UNIFIED IDEOGRAPH-5BE9
        0x5C0B, # G+5763  U+5C0B  uni5C0B,  CJK UNIFIED IDEOGRAPH-5C0B
        0x5FC3, # G+5764  U+5FC3  uni5FC3,  CJK UNIFIED IDEOGRAPH-5FC3
        0x6C81, # G+5765  U+6C81  uni6C81,  CJK UNIFIED IDEOGRAPH-6C81
        0x6DF1, # G+5766  U+6DF1  uni6DF1,  CJK UNIFIED IDEOGRAPH-6DF1
        0x700B, # G+5767  U+700B  uni700B,  CJK UNIFIED IDEOGRAPH-700B
        0x751A, # G+5768  U+751A  uni751A,  CJK UNIFIED IDEOGRAPH-751A
        0x82AF, # G+5769  U+82AF  uni82AF,  CJK UNIFIED IDEOGRAPH-82AF
        0x8AF6, # G+5770  U+8AF6  uni8AF6,  CJK UNIFIED IDEOGRAPH-8AF6
        0x4EC0, # G+5771  U+4EC0  uni4EC0,  CJK UNIFIED IDEOGRAPH-4EC0
        0x5341, # G+5772  U+5341  uni5341,  CJK UNIFIED IDEOGRAPH-5341
        0x96D9, # G+5773  U+96D9  uni96D9,  CJK UNIFIED IDEOGRAPH-96D9
        0x6C0F, # G+5774  U+6C0F  uni6C0F,  CJK UNIFIED IDEOGRAPH-6C0F
        0x4E9E, # G+5775  U+4E9E  uni4E9E,  CJK UNIFIED IDEOGRAPH-4E9E
        0x4FC4, # G+5776  U+4FC4  uni4FC4,  CJK UNIFIED IDEOGRAPH-4FC4
        0x5152, # G+5777  U+5152  uni5152,  CJK UNIFIED IDEOGRAPH-5152
        0x555E, # G+5778  U+555E  uni555E,  CJK UNIFIED IDEOGRAPH-555E
        0x5A25, # G+5779  U+5A25  uni5A25,  CJK UNIFIED IDEOGRAPH-5A25
        0x5CE8, # G+5780  U+5CE8  uni5CE8,  CJK UNIFIED IDEOGRAPH-5CE8
        0x6211, # G+5781  U+6211  uni6211,  CJK UNIFIED IDEOGRAPH-6211
        0x7259, # G+5782  U+7259  uni7259,  CJK UNIFIED IDEOGRAPH-7259
        0x82BD, # G+5783  U+82BD  uni82BD,  CJK UNIFIED IDEOGRAPH-82BD
        0x83AA, # G+5784  U+83AA  uni83AA,  CJK UNIFIED IDEOGRAPH-83AA
        0x86FE, # G+5785  U+86FE  uni86FE,  CJK UNIFIED IDEOGRAPH-86FE
        0x8859, # G+5786  U+8859  uni8859,  CJK UNIFIED IDEOGRAPH-8859
        0x8A1D, # G+5787  U+8A1D  uni8A1D,  CJK UNIFIED IDEOGRAPH-8A1D
        0x963F, # G+5788  U+963F  uni963F,  CJK UNIFIED IDEOGRAPH-963F
        0x96C5, # G+5789  U+96C5  uni96C5,  CJK UNIFIED IDEOGRAPH-96C5
        0x9913, # G+5790  U+9913  uni9913,  CJK UNIFIED IDEOGRAPH-9913
        0x9D09, # G+5791  U+9D09  uni9D09,  CJK UNIFIED IDEOGRAPH-9D09
        0x9D5D, # G+5792  U+9D5D  uni9D5D,  CJK UNIFIED IDEOGRAPH-9D5D
        0x580A, # G+5793  U+580A  uni580A,  CJK UNIFIED IDEOGRAPH-580A
        0x5CB3, # G+5794  U+5CB3  uni5CB3,  CJK UNIFIED IDEOGRAPH-5CB3
        0x5DBD, # G+5795  U+5DBD  uni5DBD,  CJK UNIFIED IDEOGRAPH-5DBD
        0x5E44, # G+5796  U+5E44  uni5E44,  CJK UNIFIED IDEOGRAPH-5E44
        0x60E1, # G+5797  U+60E1  uni60E1,  CJK UNIFIED IDEOGRAPH-60E1
        0x6115, # G+5798  U+6115  uni6115,  CJK UNIFIED IDEOGRAPH-6115
        0x63E1, # G+5799  U+63E1  uni63E1,  CJK UNIFIED IDEOGRAPH-63E1
        0xF914, # G+5800  U+F914  uniF914,  CJK COMPATIBILITY IDEOGRAPH-F914
        0x6E25, # G+5801  U+6E25  uni6E25,  CJK UNIFIED IDEOGRAPH-6E25
        0x9102, # G+5802  U+9102  uni9102,  CJK UNIFIED IDEOGRAPH-9102
        0x9354, # G+5803  U+9354  uni9354,  CJK UNIFIED IDEOGRAPH-9354
        0x984E, # G+5804  U+984E  uni984E,  CJK UNIFIED IDEOGRAPH-984E
        0x9C10, # G+5805  U+9C10  uni9C10,  CJK UNIFIED IDEOGRAPH-9C10
        0x9F77, # G+5806  U+9F77  uni9F77,  CJK UNIFIED IDEOGRAPH-9F77
        0x5B89, # G+5807  U+5B89  uni5B89,  CJK UNIFIED IDEOGRAPH-5B89
        0x5CB8, # G+5808  U+5CB8  uni5CB8,  CJK UNIFIED IDEOGRAPH-5CB8
        0x6309, # G+5809  U+6309  uni6309,  CJK UNIFIED IDEOGRAPH-6309
        0x664F, # G+5810  U+664F  uni664F,  CJK UNIFIED IDEOGRAPH-664F
        0x6848, # G+5811  U+6848  uni6848,  CJK UNIFIED IDEOGRAPH-6848
        0x773C, # G+5812  U+773C  uni773C,  CJK UNIFIED IDEOGRAPH-773C
        0x96C1, # G+5813  U+96C1  uni96C1,  CJK UNIFIED IDEOGRAPH-96C1
        0x978D, # G+5814  U+978D  uni978D,  CJK UNIFIED IDEOGRAPH-978D
        0x9854, # G+5815  U+9854  uni9854,  CJK UNIFIED IDEOGRAPH-9854
        0x9B9F, # G+5816  U+9B9F  uni9B9F,  CJK UNIFIED IDEOGRAPH-9B9F
        0x65A1, # G+5817  U+65A1  uni65A1,  CJK UNIFIED IDEOGRAPH-65A1
        0x8B01, # G+5818  U+8B01  uni8B01,  CJK UNIFIED IDEOGRAPH-8B01
        0x8ECB, # G+5819  U+8ECB  uni8ECB,  CJK UNIFIED IDEOGRAPH-8ECB
        0x95BC, # G+5820  U+95BC  uni95BC,  CJK UNIFIED IDEOGRAPH-95BC
        0x5535, # G+5821  U+5535  uni5535,  CJK UNIFIED IDEOGRAPH-5535
        0x5CA9, # G+5822  U+5CA9  uni5CA9,  CJK UNIFIED IDEOGRAPH-5CA9
        0x5DD6, # G+5823  U+5DD6  uni5DD6,  CJK UNIFIED IDEOGRAPH-5DD6
        0x5EB5, # G+5824  U+5EB5  uni5EB5,  CJK UNIFIED IDEOGRAPH-5EB5
        0x6697, # G+5825  U+6697  uni6697,  CJK UNIFIED IDEOGRAPH-6697
        0x764C, # G+5826  U+764C  uni764C,  CJK UNIFIED IDEOGRAPH-764C
        0x83F4, # G+5827  U+83F4  uni83F4,  CJK UNIFIED IDEOGRAPH-83F4
        0x95C7, # G+5828  U+95C7  uni95C7,  CJK UNIFIED IDEOGRAPH-95C7
        0x58D3, # G+5829  U+58D3  uni58D3,  CJK UNIFIED IDEOGRAPH-58D3
        0x62BC, # G+5830  U+62BC  uni62BC,  CJK UNIFIED IDEOGRAPH-62BC
        0x72CE, # G+5831  U+72CE  uni72CE,  CJK UNIFIED IDEOGRAPH-72CE
        0x9D28, # G+5832  U+9D28  uni9D28,  CJK UNIFIED IDEOGRAPH-9D28
        0x4EF0, # G+5833  U+4EF0  uni4EF0,  CJK UNIFIED IDEOGRAPH-4EF0
        0x592E, # G+5834  U+592E  uni592E,  CJK UNIFIED IDEOGRAPH-592E
        0x600F, # G+5835  U+600F  uni600F,  CJK UNIFIED IDEOGRAPH-600F
        0x663B, # G+5836  U+663B  uni663B,  CJK UNIFIED IDEOGRAPH-663B
        0x6B83, # G+5837  U+6B83  uni6B83,  CJK UNIFIED IDEOGRAPH-6B83
        0x79E7, # G+5838  U+79E7  uni79E7,  CJK UNIFIED IDEOGRAPH-79E7
        0x9D26, # G+5839  U+9D26  uni9D26,  CJK UNIFIED IDEOGRAPH-9D26
        0x5393, # G+5840  U+5393  uni5393,  CJK UNIFIED IDEOGRAPH-5393
        0x54C0, # G+5841  U+54C0  uni54C0,  CJK UNIFIED IDEOGRAPH-54C0
        0x57C3, # G+5842  U+57C3  uni57C3,  CJK UNIFIED IDEOGRAPH-57C3
        0x5D16, # G+5843  U+5D16  uni5D16,  CJK UNIFIED IDEOGRAPH-5D16
        0x611B, # G+5844  U+611B  uni611B,  CJK UNIFIED IDEOGRAPH-611B
        0x66D6, # G+5845  U+66D6  uni66D6,  CJK UNIFIED IDEOGRAPH-66D6
        0x6DAF, # G+5846  U+6DAF  uni6DAF,  CJK UNIFIED IDEOGRAPH-6DAF
        0x788D, # G+5847  U+788D  uni788D,  CJK UNIFIED IDEOGRAPH-788D
        0x827E, # G+5848  U+827E  uni827E,  CJK UNIFIED IDEOGRAPH-827E
        0x9698, # G+5849  U+9698  uni9698,  CJK UNIFIED IDEOGRAPH-9698
        0x9744, # G+5850  U+9744  uni9744,  CJK UNIFIED IDEOGRAPH-9744
        0x5384, # G+5851  U+5384  uni5384,  CJK UNIFIED IDEOGRAPH-5384
        0x627C, # G+5852  U+627C  uni627C,  CJK UNIFIED IDEOGRAPH-627C
        0x6396, # G+5853  U+6396  uni6396,  CJK UNIFIED IDEOGRAPH-6396
        0x6DB2, # G+5854  U+6DB2  uni6DB2,  CJK UNIFIED IDEOGRAPH-6DB2
        0x7E0A, # G+5855  U+7E0A  uni7E0A,  CJK UNIFIED IDEOGRAPH-7E0A
        0x814B, # G+5856  U+814B  uni814B,  CJK UNIFIED IDEOGRAPH-814B
        0x984D, # G+5857  U+984D  uni984D,  CJK UNIFIED IDEOGRAPH-984D
        0x6AFB, # G+5858  U+6AFB  uni6AFB,  CJK UNIFIED IDEOGRAPH-6AFB
        0x7F4C, # G+5859  U+7F4C  uni7F4C,  CJK UNIFIED IDEOGRAPH-7F4C
        0x9DAF, # G+5860  U+9DAF  uni9DAF,  CJK UNIFIED IDEOGRAPH-9DAF
        0x9E1A, # G+5861  U+9E1A  uni9E1A,  CJK UNIFIED IDEOGRAPH-9E1A
        0x4E5F, # G+5862  U+4E5F  uni4E5F,  CJK UNIFIED IDEOGRAPH-4E5F
        0x503B, # G+5863  U+503B  uni503B,  CJK UNIFIED IDEOGRAPH-503B
        0x51B6, # G+5864  U+51B6  uni51B6,  CJK UNIFIED IDEOGRAPH-51B6
        0x591C, # G+5865  U+591C  uni591C,  CJK UNIFIED IDEOGRAPH-591C
        0x60F9, # G+5866  U+60F9  uni60F9,  CJK UNIFIED IDEOGRAPH-60F9
        0x63F6, # G+5867  U+63F6  uni63F6,  CJK UNIFIED IDEOGRAPH-63F6
        0x6930, # G+5868  U+6930  uni6930,  CJK UNIFIED IDEOGRAPH-6930
        0x723A, # G+5869  U+723A  uni723A,  CJK UNIFIED IDEOGRAPH-723A
        0x8036, # G+5870  U+8036  uni8036,  CJK UNIFIED IDEOGRAPH-8036
        0x91CE, # G+5871  U+91CE  uni91CE,  CJK UNIFIED IDEOGRAPH-91CE
        0x5F31, # G+5872  U+5F31  uni5F31,  CJK UNIFIED IDEOGRAPH-5F31
        0x7D04, # G+5873  U+7D04  uni7D04,  CJK UNIFIED IDEOGRAPH-7D04
        0xF974, # G+5874  U+F974  uniF974,  CJK COMPATIBILITY IDEOGRAPH-F974
        0x846F, # G+5875  U+846F  uni846F,  CJK UNIFIED IDEOGRAPH-846F
        0x84BB, # G+5876  U+84BB  uni84BB,  CJK UNIFIED IDEOGRAPH-84BB
        0x85E5, # G+5877  U+85E5  uni85E5,  CJK UNIFIED IDEOGRAPH-85E5
        0x8E8D, # G+5878  U+8E8D  uni8E8D,  CJK UNIFIED IDEOGRAPH-8E8D
        0x4F6F, # G+5879  U+4F6F  uni4F6F,  CJK UNIFIED IDEOGRAPH-4F6F
        0x58E4, # G+5880  U+58E4  uni58E4,  CJK UNIFIED IDEOGRAPH-58E4
        0x5B43, # G+5881  U+5B43  uni5B43,  CJK UNIFIED IDEOGRAPH-5B43
        0x6059, # G+5882  U+6059  uni6059,  CJK UNIFIED IDEOGRAPH-6059
        0x63DA, # G+5883  U+63DA  uni63DA,  CJK UNIFIED IDEOGRAPH-63DA
        0x6518, # G+5884  U+6518  uni6518,  CJK UNIFIED IDEOGRAPH-6518
        0x656D, # G+5885  U+656D  uni656D,  CJK UNIFIED IDEOGRAPH-656D
        0x6698, # G+5886  U+6698  uni6698,  CJK UNIFIED IDEOGRAPH-6698
        0x694A, # G+5887  U+694A  uni694A,  CJK UNIFIED IDEOGRAPH-694A
        0x6A23, # G+5888  U+6A23  uni6A23,  CJK UNIFIED IDEOGRAPH-6A23
        0x6D0B, # G+5889  U+6D0B  uni6D0B,  CJK UNIFIED IDEOGRAPH-6D0B
        0x7001, # G+5890  U+7001  uni7001,  CJK UNIFIED IDEOGRAPH-7001
        0x716C, # G+5891  U+716C  uni716C,  CJK UNIFIED IDEOGRAPH-716C
        0x75D2, # G+5892  U+75D2  uni75D2,  CJK UNIFIED IDEOGRAPH-75D2
        0x760D, # G+5893  U+760D  uni760D,  CJK UNIFIED IDEOGRAPH-760D
        0x79B3, # G+5894  U+79B3  uni79B3,  CJK UNIFIED IDEOGRAPH-79B3
        0x7A70, # G+5895  U+7A70  uni7A70,  CJK UNIFIED IDEOGRAPH-7A70
        0x7F8A, # G+5896  U+7F8A  uni7F8A,  CJK UNIFIED IDEOGRAPH-7F8A
        0x8944, # G+5897  U+8944  uni8944,  CJK UNIFIED IDEOGRAPH-8944
        0x8B93, # G+5898  U+8B93  uni8B93,  CJK UNIFIED IDEOGRAPH-8B93
        0x91C0, # G+5899  U+91C0  uni91C0,  CJK UNIFIED IDEOGRAPH-91C0
        0x967D, # G+5900  U+967D  uni967D,  CJK UNIFIED IDEOGRAPH-967D
        0x990A, # G+5901  U+990A  uni990A,  CJK UNIFIED IDEOGRAPH-990A
        0x5704, # G+5902  U+5704  uni5704,  CJK UNIFIED IDEOGRAPH-5704
        0x5FA1, # G+5903  U+5FA1  uni5FA1,  CJK UNIFIED IDEOGRAPH-5FA1
        0x65BC, # G+5904  U+65BC  uni65BC,  CJK UNIFIED IDEOGRAPH-65BC
        0x6F01, # G+5905  U+6F01  uni6F01,  CJK UNIFIED IDEOGRAPH-6F01
        0x7600, # G+5906  U+7600  uni7600,  CJK UNIFIED IDEOGRAPH-7600
        0x79A6, # G+5907  U+79A6  uni79A6,  CJK UNIFIED IDEOGRAPH-79A6
        0x8A9E, # G+5908  U+8A9E  uni8A9E,  CJK UNIFIED IDEOGRAPH-8A9E
        0x99AD, # G+5909  U+99AD  uni99AD,  CJK UNIFIED IDEOGRAPH-99AD
        0x9B5A, # G+5910  U+9B5A  uni9B5A,  CJK UNIFIED IDEOGRAPH-9B5A
        0x9F6C, # G+5911  U+9F6C  uni9F6C,  CJK UNIFIED IDEOGRAPH-9F6C
        0x5104, # G+5912  U+5104  uni5104,  CJK UNIFIED IDEOGRAPH-5104
        0x61B6, # G+5913  U+61B6  uni61B6,  CJK UNIFIED IDEOGRAPH-61B6
        0x6291, # G+5914  U+6291  uni6291,  CJK UNIFIED IDEOGRAPH-6291
        0x6A8D, # G+5915  U+6A8D  uni6A8D,  CJK UNIFIED IDEOGRAPH-6A8D
        0x81C6, # G+5916  U+81C6  uni81C6,  CJK UNIFIED IDEOGRAPH-81C6
        0x5043, # G+5917  U+5043  uni5043,  CJK UNIFIED IDEOGRAPH-5043
        0x5830, # G+5918  U+5830  uni5830,  CJK UNIFIED IDEOGRAPH-5830
        0x5F66, # G+5919  U+5F66  uni5F66,  CJK UNIFIED IDEOGRAPH-5F66
        0x7109, # G+5920  U+7109  uni7109,  CJK UNIFIED IDEOGRAPH-7109
        0x8A00, # G+5921  U+8A00  uni8A00,  CJK UNIFIED IDEOGRAPH-8A00
        0x8AFA, # G+5922  U+8AFA  uni8AFA,  CJK UNIFIED IDEOGRAPH-8AFA
        0x5B7C, # G+5923  U+5B7C  uni5B7C,  CJK UNIFIED IDEOGRAPH-5B7C
        0x8616, # G+5924  U+8616  uni8616,  CJK UNIFIED IDEOGRAPH-8616
        0x4FFA, # G+5925  U+4FFA  uni4FFA,  CJK UNIFIED IDEOGRAPH-4FFA
        0x513C, # G+5926  U+513C  uni513C,  CJK UNIFIED IDEOGRAPH-513C
        0x56B4, # G+5927  U+56B4  uni56B4,  CJK UNIFIED IDEOGRAPH-56B4
        0x5944, # G+5928  U+5944  uni5944,  CJK UNIFIED IDEOGRAPH-5944
        0x63A9, # G+5929  U+63A9  uni63A9,  CJK UNIFIED IDEOGRAPH-63A9
        0x6DF9, # G+5930  U+6DF9  uni6DF9,  CJK UNIFIED IDEOGRAPH-6DF9
        0x5DAA, # G+5931  U+5DAA  uni5DAA,  CJK UNIFIED IDEOGRAPH-5DAA
        0x696D, # G+5932  U+696D  uni696D,  CJK UNIFIED IDEOGRAPH-696D
        0x5186, # G+5933  U+5186  uni5186,  CJK UNIFIED IDEOGRAPH-5186
        0x4E88, # G+5934  U+4E88  uni4E88,  CJK UNIFIED IDEOGRAPH-4E88
        0x4F59, # G+5935  U+4F59  uni4F59,  CJK UNIFIED IDEOGRAPH-4F59
        0x5982, # G+5936  U+5982  uni5982,  CJK UNIFIED IDEOGRAPH-5982
        0x6B5F, # G+5937  U+6B5F  uni6B5F,  CJK UNIFIED IDEOGRAPH-6B5F
        0x6C5D, # G+5938  U+6C5D  uni6C5D,  CJK UNIFIED IDEOGRAPH-6C5D
        0x74B5, # G+5939  U+74B5  uni74B5,  CJK UNIFIED IDEOGRAPH-74B5
        0x7916, # G+5940  U+7916  uni7916,  CJK UNIFIED IDEOGRAPH-7916
        0x8207, # G+5941  U+8207  uni8207,  CJK UNIFIED IDEOGRAPH-8207
        0x8245, # G+5942  U+8245  uni8245,  CJK UNIFIED IDEOGRAPH-8245
        0x8339, # G+5943  U+8339  uni8339,  CJK UNIFIED IDEOGRAPH-8339
        0x8F3F, # G+5944  U+8F3F  uni8F3F,  CJK UNIFIED IDEOGRAPH-8F3F
        0x8F5D, # G+5945  U+8F5D  uni8F5D,  CJK UNIFIED IDEOGRAPH-8F5D
        0x9918, # G+5946  U+9918  uni9918,  CJK UNIFIED IDEOGRAPH-9918
        0x4EA6, # G+5947  U+4EA6  uni4EA6,  CJK UNIFIED IDEOGRAPH-4EA6
        0x57DF, # G+5948  U+57DF  uni57DF,  CJK UNIFIED IDEOGRAPH-57DF
        0x5F79, # G+5949  U+5F79  uni5F79,  CJK UNIFIED IDEOGRAPH-5F79
        0x6613, # G+5950  U+6613  uni6613,  CJK UNIFIED IDEOGRAPH-6613
        0x75AB, # G+5951  U+75AB  uni75AB,  CJK UNIFIED IDEOGRAPH-75AB
        0x7E79, # G+5952  U+7E79  uni7E79,  CJK UNIFIED IDEOGRAPH-7E79
        0x8B6F, # G+5953  U+8B6F  uni8B6F,  CJK UNIFIED IDEOGRAPH-8B6F
        0x9006, # G+5954  U+9006  uni9006,  CJK UNIFIED IDEOGRAPH-9006
        0x9A5B, # G+5955  U+9A5B  uni9A5B,  CJK UNIFIED IDEOGRAPH-9A5B
        0x56A5, # G+5956  U+56A5  uni56A5,  CJK UNIFIED IDEOGRAPH-56A5
        0x5827, # G+5957  U+5827  uni5827,  CJK UNIFIED IDEOGRAPH-5827
        0x59F8, # G+5958  U+59F8  uni59F8,  CJK UNIFIED IDEOGRAPH-59F8
        0x5A1F, # G+5959  U+5A1F  uni5A1F,  CJK UNIFIED IDEOGRAPH-5A1F
        0x5BB4, # G+5960  U+5BB4  uni5BB4,  CJK UNIFIED IDEOGRAPH-5BB4
        0x5EF6, # G+5961  U+5EF6  uni5EF6,  CJK UNIFIED IDEOGRAPH-5EF6
        0x6350, # G+5962  U+6350  uni6350,  CJK UNIFIED IDEOGRAPH-6350
        0x633B, # G+5963  U+633B  uni633B,  CJK UNIFIED IDEOGRAPH-633B
        0x693D, # G+5964  U+693D  uni693D,  CJK UNIFIED IDEOGRAPH-693D
        0x6C87, # G+5965  U+6C87  uni6C87,  CJK UNIFIED IDEOGRAPH-6C87
        0x6CBF, # G+5966  U+6CBF  uni6CBF,  CJK UNIFIED IDEOGRAPH-6CBF
        0x6D8E, # G+5967  U+6D8E  uni6D8E,  CJK UNIFIED IDEOGRAPH-6D8E
        0x6D93, # G+5968  U+6D93  uni6D93,  CJK UNIFIED IDEOGRAPH-6D93
        0x6DF5, # G+5969  U+6DF5  uni6DF5,  CJK UNIFIED IDEOGRAPH-6DF5
        0x6F14, # G+5970  U+6F14  uni6F14,  CJK UNIFIED IDEOGRAPH-6F14
        0x70DF, # G+5971  U+70DF  uni70DF,  CJK UNIFIED IDEOGRAPH-70DF
        0x7136, # G+5972  U+7136  uni7136,  CJK UNIFIED IDEOGRAPH-7136
        0x7159, # G+5973  U+7159  uni7159,  CJK UNIFIED IDEOGRAPH-7159
        0x71C3, # G+5974  U+71C3  uni71C3,  CJK UNIFIED IDEOGRAPH-71C3
        0x71D5, # G+5975  U+71D5  uni71D5,  CJK UNIFIED IDEOGRAPH-71D5
        0x784F, # G+5976  U+784F  uni784F,  CJK UNIFIED IDEOGRAPH-784F
        0x786F, # G+5977  U+786F  uni786F,  CJK UNIFIED IDEOGRAPH-786F
        0x7B75, # G+5978  U+7B75  uni7B75,  CJK UNIFIED IDEOGRAPH-7B75
        0x7DE3, # G+5979  U+7DE3  uni7DE3,  CJK UNIFIED IDEOGRAPH-7DE3
        0x7E2F, # G+5980  U+7E2F  uni7E2F,  CJK UNIFIED IDEOGRAPH-7E2F
        0x884D, # G+5981  U+884D  uni884D,  CJK UNIFIED IDEOGRAPH-884D
        0x8EDF, # G+5982  U+8EDF  uni8EDF,  CJK UNIFIED IDEOGRAPH-8EDF
        0x925B, # G+5983  U+925B  uni925B,  CJK UNIFIED IDEOGRAPH-925B
        0x9CF6, # G+5984  U+9CF6  uni9CF6,  CJK UNIFIED IDEOGRAPH-9CF6
        0x6085, # G+5985  U+6085  uni6085,  CJK UNIFIED IDEOGRAPH-6085
        0x6D85, # G+5986  U+6D85  uni6D85,  CJK UNIFIED IDEOGRAPH-6D85
        0x71B1, # G+5987  U+71B1  uni71B1,  CJK UNIFIED IDEOGRAPH-71B1
        0x95B1, # G+5988  U+95B1  uni95B1,  CJK UNIFIED IDEOGRAPH-95B1
        0x53AD, # G+5989  U+53AD  uni53AD,  CJK UNIFIED IDEOGRAPH-53AD
        0x67D3, # G+5990  U+67D3  uni67D3,  CJK UNIFIED IDEOGRAPH-67D3
        0x708E, # G+5991  U+708E  uni708E,  CJK UNIFIED IDEOGRAPH-708E
        0x7130, # G+5992  U+7130  uni7130,  CJK UNIFIED IDEOGRAPH-7130
        0x7430, # G+5993  U+7430  uni7430,  CJK UNIFIED IDEOGRAPH-7430
        0x8276, # G+5994  U+8276  uni8276,  CJK UNIFIED IDEOGRAPH-8276
        0x82D2, # G+5995  U+82D2  uni82D2,  CJK UNIFIED IDEOGRAPH-82D2
        0x95BB, # G+5996  U+95BB  uni95BB,  CJK UNIFIED IDEOGRAPH-95BB
        0x9AE5, # G+5997  U+9AE5  uni9AE5,  CJK UNIFIED IDEOGRAPH-9AE5
        0x9E7D, # G+5998  U+9E7D  uni9E7D,  CJK UNIFIED IDEOGRAPH-9E7D
        0x66C4, # G+5999  U+66C4  uni66C4,  CJK UNIFIED IDEOGRAPH-66C4
        0x71C1, # G+6000  U+71C1  uni71C1,  CJK UNIFIED IDEOGRAPH-71C1
        0xF96E, # G+6001  U+F96E  uniF96E,  CJK COMPATIBILITY IDEOGRAPH-F96E
        0x584B, # G+6002  U+584B  uni584B,  CJK UNIFIED IDEOGRAPH-584B
        0x5DB8, # G+6003  U+5DB8  uni5DB8,  CJK UNIFIED IDEOGRAPH-5DB8
        0x5F71, # G+6004  U+5F71  uni5F71,  CJK UNIFIED IDEOGRAPH-5F71
        0x6620, # G+6005  U+6620  uni6620,  CJK UNIFIED IDEOGRAPH-6620
        0x668E, # G+6006  U+668E  uni668E,  CJK UNIFIED IDEOGRAPH-668E
        0x6979, # G+6007  U+6979  uni6979,  CJK UNIFIED IDEOGRAPH-6979
        0x69AE, # G+6008  U+69AE  uni69AE,  CJK UNIFIED IDEOGRAPH-69AE
        0x6C38, # G+6009  U+6C38  uni6C38,  CJK UNIFIED IDEOGRAPH-6C38
        0x6CF3, # G+6010  U+6CF3  uni6CF3,  CJK UNIFIED IDEOGRAPH-6CF3
        0x6E36, # G+6011  U+6E36  uni6E36,  CJK UNIFIED IDEOGRAPH-6E36
        0x6F41, # G+6012  U+6F41  uni6F41,  CJK UNIFIED IDEOGRAPH-6F41
        0x6FDA, # G+6013  U+6FDA  uni6FDA,  CJK UNIFIED IDEOGRAPH-6FDA
        0x701B, # G+6014  U+701B  uni701B,  CJK UNIFIED IDEOGRAPH-701B
        0x702F, # G+6015  U+702F  uni702F,  CJK UNIFIED IDEOGRAPH-702F
        0x7150, # G+6016  U+7150  uni7150,  CJK UNIFIED IDEOGRAPH-7150
        0x71DF, # G+6017  U+71DF  uni71DF,  CJK UNIFIED IDEOGRAPH-71DF
        0x7370, # G+6018  U+7370  uni7370,  CJK UNIFIED IDEOGRAPH-7370
        0x745B, # G+6019  U+745B  uni745B,  CJK UNIFIED IDEOGRAPH-745B
        0x74D4, # G+6020  U+74D4  uni74D4,  CJK UNIFIED IDEOGRAPH-74D4
        0x76C8, # G+6021  U+76C8  uni76C8,  CJK UNIFIED IDEOGRAPH-76C8
        0x7A4E, # G+6022  U+7A4E  uni7A4E,  CJK UNIFIED IDEOGRAPH-7A4E
        0x7E93, # G+6023  U+7E93  uni7E93,  CJK UNIFIED IDEOGRAPH-7E93
        0x82F1, # G+6024  U+82F1  uni82F1,  CJK UNIFIED IDEOGRAPH-82F1
        0x8A60, # G+6025  U+8A60  uni8A60,  CJK UNIFIED IDEOGRAPH-8A60
        0x8FCE, # G+6026  U+8FCE  uni8FCE,  CJK UNIFIED IDEOGRAPH-8FCE
        0x9348, # G+6027  U+9348  uni9348,  CJK UNIFIED IDEOGRAPH-9348
        0x9719, # G+6028  U+9719  uni9719,  CJK UNIFIED IDEOGRAPH-9719
        0x4E42, # G+6029  U+4E42  uni4E42,  CJK UNIFIED IDEOGRAPH-4E42
        0x502A, # G+6030  U+502A  uni502A,  CJK UNIFIED IDEOGRAPH-502A
        0x5208, # G+6031  U+5208  uni5208,  CJK UNIFIED IDEOGRAPH-5208
        0x53E1, # G+6032  U+53E1  uni53E1,  CJK UNIFIED IDEOGRAPH-53E1
        0x66F3, # G+6033  U+66F3  uni66F3,  CJK UNIFIED IDEOGRAPH-66F3
        0x6C6D, # G+6034  U+6C6D  uni6C6D,  CJK UNIFIED IDEOGRAPH-6C6D
        0x6FCA, # G+6035  U+6FCA  uni6FCA,  CJK UNIFIED IDEOGRAPH-6FCA
        0x730A, # G+6036  U+730A  uni730A,  CJK UNIFIED IDEOGRAPH-730A
        0x777F, # G+6037  U+777F  uni777F,  CJK UNIFIED IDEOGRAPH-777F
        0x7A62, # G+6038  U+7A62  uni7A62,  CJK UNIFIED IDEOGRAPH-7A62
        0x82AE, # G+6039  U+82AE  uni82AE,  CJK UNIFIED IDEOGRAPH-82AE
        0x85DD, # G+6040  U+85DD  uni85DD,  CJK UNIFIED IDEOGRAPH-85DD
        0x8602, # G+6041  U+8602  uni8602,  CJK UNIFIED IDEOGRAPH-8602
        0x88D4, # G+6042  U+88D4  uni88D4,  CJK UNIFIED IDEOGRAPH-88D4
        0x8A63, # G+6043  U+8A63  uni8A63,  CJK UNIFIED IDEOGRAPH-8A63
        0x8B7D, # G+6044  U+8B7D  uni8B7D,  CJK UNIFIED IDEOGRAPH-8B7D
        0x8C6B, # G+6045  U+8C6B  uni8C6B,  CJK UNIFIED IDEOGRAPH-8C6B
        0x92B3, # G+6046  U+92B3  uni92B3,  CJK UNIFIED IDEOGRAPH-92B3
        0x9713, # G+6047  U+9713  uni9713,  CJK UNIFIED IDEOGRAPH-9713
        0x9810, # G+6048  U+9810  uni9810,  CJK UNIFIED IDEOGRAPH-9810
        0x4E94, # G+6049  U+4E94  uni4E94,  CJK UNIFIED IDEOGRAPH-4E94
        0x4F0D, # G+6050  U+4F0D  uni4F0D,  CJK UNIFIED IDEOGRAPH-4F0D
        0x4FC9, # G+6051  U+4FC9  uni4FC9,  CJK UNIFIED IDEOGRAPH-4FC9
        0x50B2, # G+6052  U+50B2  uni50B2,  CJK UNIFIED IDEOGRAPH-50B2
        0x5348, # G+6053  U+5348  uni5348,  CJK UNIFIED IDEOGRAPH-5348
        0x543E, # G+6054  U+543E  uni543E,  CJK UNIFIED IDEOGRAPH-543E
        0x5433, # G+6055  U+5433  uni5433,  CJK UNIFIED IDEOGRAPH-5433
        0x55DA, # G+6056  U+55DA  uni55DA,  CJK UNIFIED IDEOGRAPH-55DA
        0x5862, # G+6057  U+5862  uni5862,  CJK UNIFIED IDEOGRAPH-5862
        0x58BA, # G+6058  U+58BA  uni58BA,  CJK UNIFIED IDEOGRAPH-58BA
        0x5967, # G+6059  U+5967  uni5967,  CJK UNIFIED IDEOGRAPH-5967
        0x5A1B, # G+6060  U+5A1B  uni5A1B,  CJK UNIFIED IDEOGRAPH-5A1B
        0x5BE4, # G+6061  U+5BE4  uni5BE4,  CJK UNIFIED IDEOGRAPH-5BE4
        0x609F, # G+6062  U+609F  uni609F,  CJK UNIFIED IDEOGRAPH-609F
        0x61CA, # G+6063  U+61CA  uni61CA,  CJK UNIFIED IDEOGRAPH-61CA
        0x6556, # G+6064  U+6556  uni6556,  CJK UNIFIED IDEOGRAPH-6556
        0x65FF, # G+6065  U+65FF  uni65FF,  CJK UNIFIED IDEOGRAPH-65FF
        0x6664, # G+6066  U+6664  uni6664,  CJK UNIFIED IDEOGRAPH-6664
        0x68A7, # G+6067  U+68A7  uni68A7,  CJK UNIFIED IDEOGRAPH-68A7
        0x6C5A, # G+6068  U+6C5A  uni6C5A,  CJK UNIFIED IDEOGRAPH-6C5A
        0x6FB3, # G+6069  U+6FB3  uni6FB3,  CJK UNIFIED IDEOGRAPH-6FB3
        0x70CF, # G+6070  U+70CF  uni70CF,  CJK UNIFIED IDEOGRAPH-70CF
        0x71AC, # G+6071  U+71AC  uni71AC,  CJK UNIFIED IDEOGRAPH-71AC
        0x7352, # G+6072  U+7352  uni7352,  CJK UNIFIED IDEOGRAPH-7352
        0x7B7D, # G+6073  U+7B7D  uni7B7D,  CJK UNIFIED IDEOGRAPH-7B7D
        0x8708, # G+6074  U+8708  uni8708,  CJK UNIFIED IDEOGRAPH-8708
        0x8AA4, # G+6075  U+8AA4  uni8AA4,  CJK UNIFIED IDEOGRAPH-8AA4
        0x9C32, # G+6076  U+9C32  uni9C32,  CJK UNIFIED IDEOGRAPH-9C32
        0x9F07, # G+6077  U+9F07  uni9F07,  CJK UNIFIED IDEOGRAPH-9F07
        0x5C4B, # G+6078  U+5C4B  uni5C4B,  CJK UNIFIED IDEOGRAPH-5C4B
        0x6C83, # G+6079  U+6C83  uni6C83,  CJK UNIFIED IDEOGRAPH-6C83
        0x7344, # G+6080  U+7344  uni7344,  CJK UNIFIED IDEOGRAPH-7344
        0x7389, # G+6081  U+7389  uni7389,  CJK UNIFIED IDEOGRAPH-7389
        0x923A, # G+6082  U+923A  uni923A,  CJK UNIFIED IDEOGRAPH-923A
        0x6EAB, # G+6083  U+6EAB  uni6EAB,  CJK UNIFIED IDEOGRAPH-6EAB
        0x7465, # G+6084  U+7465  uni7465,  CJK UNIFIED IDEOGRAPH-7465
        0x761F, # G+6085  U+761F  uni761F,  CJK UNIFIED IDEOGRAPH-761F
        0x7A69, # G+6086  U+7A69  uni7A69,  CJK UNIFIED IDEOGRAPH-7A69
        0x7E15, # G+6087  U+7E15  uni7E15,  CJK UNIFIED IDEOGRAPH-7E15
        0x860A, # G+6088  U+860A  uni860A,  CJK UNIFIED IDEOGRAPH-860A
        0x5140, # G+6089  U+5140  uni5140,  CJK UNIFIED IDEOGRAPH-5140
        0x58C5, # G+6090  U+58C5  uni58C5,  CJK UNIFIED IDEOGRAPH-58C5
        0x64C1, # G+6091  U+64C1  uni64C1,  CJK UNIFIED IDEOGRAPH-64C1
        0x74EE, # G+6092  U+74EE  uni74EE,  CJK UNIFIED IDEOGRAPH-74EE
        0x7515, # G+6093  U+7515  uni7515,  CJK UNIFIED IDEOGRAPH-7515
        0x7670, # G+6094  U+7670  uni7670,  CJK UNIFIED IDEOGRAPH-7670
        0x7FC1, # G+6095  U+7FC1  uni7FC1,  CJK UNIFIED IDEOGRAPH-7FC1
        0x9095, # G+6096  U+9095  uni9095,  CJK UNIFIED IDEOGRAPH-9095
        0x96CD, # G+6097  U+96CD  uni96CD,  CJK UNIFIED IDEOGRAPH-96CD
        0x9954, # G+6098  U+9954  uni9954,  CJK UNIFIED IDEOGRAPH-9954
        0x6E26, # G+6099  U+6E26  uni6E26,  CJK UNIFIED IDEOGRAPH-6E26
        0x74E6, # G+6100  U+74E6  uni74E6,  CJK UNIFIED IDEOGRAPH-74E6
        0x7AA9, # G+6101  U+7AA9  uni7AA9,  CJK UNIFIED IDEOGRAPH-7AA9
        0x7AAA, # G+6102  U+7AAA  uni7AAA,  CJK UNIFIED IDEOGRAPH-7AAA
        0x81E5, # G+6103  U+81E5  uni81E5,  CJK UNIFIED IDEOGRAPH-81E5
        0x86D9, # G+6104  U+86D9  uni86D9,  CJK UNIFIED IDEOGRAPH-86D9
        0x8778, # G+6105  U+8778  uni8778,  CJK UNIFIED IDEOGRAPH-8778
        0x8A1B, # G+6106  U+8A1B  uni8A1B,  CJK UNIFIED IDEOGRAPH-8A1B
        0x5A49, # G+6107  U+5A49  uni5A49,  CJK UNIFIED IDEOGRAPH-5A49
        0x5B8C, # G+6108  U+5B8C  uni5B8C,  CJK UNIFIED IDEOGRAPH-5B8C
        0x5B9B, # G+6109  U+5B9B  uni5B9B,  CJK UNIFIED IDEOGRAPH-5B9B
        0x68A1, # G+6110  U+68A1  uni68A1,  CJK UNIFIED IDEOGRAPH-68A1
        0x6900, # G+6111  U+6900  uni6900,  CJK UNIFIED IDEOGRAPH-6900
        0x6D63, # G+6112  U+6D63  uni6D63,  CJK UNIFIED IDEOGRAPH-6D63
        0x73A9, # G+6113  U+73A9  uni73A9,  CJK UNIFIED IDEOGRAPH-73A9
        0x7413, # G+6114  U+7413  uni7413,  CJK UNIFIED IDEOGRAPH-7413
        0x742C, # G+6115  U+742C  uni742C,  CJK UNIFIED IDEOGRAPH-742C
        0x7897, # G+6116  U+7897  uni7897,  CJK UNIFIED IDEOGRAPH-7897
        0x7DE9, # G+6117  U+7DE9  uni7DE9,  CJK UNIFIED IDEOGRAPH-7DE9
        0x7FEB, # G+6118  U+7FEB  uni7FEB,  CJK UNIFIED IDEOGRAPH-7FEB
        0x8118, # G+6119  U+8118  uni8118,  CJK UNIFIED IDEOGRAPH-8118
        0x8155, # G+6120  U+8155  uni8155,  CJK UNIFIED IDEOGRAPH-8155
        0x839E, # G+6121  U+839E  uni839E,  CJK UNIFIED IDEOGRAPH-839E
        0x8C4C, # G+6122  U+8C4C  uni8C4C,  CJK UNIFIED IDEOGRAPH-8C4C
        0x962E, # G+6123  U+962E  uni962E,  CJK UNIFIED IDEOGRAPH-962E
        0x9811, # G+6124  U+9811  uni9811,  CJK UNIFIED IDEOGRAPH-9811
        0x66F0, # G+6125  U+66F0  uni66F0,  CJK UNIFIED IDEOGRAPH-66F0
        0x5F80, # G+6126  U+5F80  uni5F80,  CJK UNIFIED IDEOGRAPH-5F80
        0x65FA, # G+6127  U+65FA  uni65FA,  CJK UNIFIED IDEOGRAPH-65FA
        0x6789, # G+6128  U+6789  uni6789,  CJK UNIFIED IDEOGRAPH-6789
        0x6C6A, # G+6129  U+6C6A  uni6C6A,  CJK UNIFIED IDEOGRAPH-6C6A
        0x738B, # G+6130  U+738B  uni738B,  CJK UNIFIED IDEOGRAPH-738B
        0x502D, # G+6131  U+502D  uni502D,  CJK UNIFIED IDEOGRAPH-502D
        0x5A03, # G+6132  U+5A03  uni5A03,  CJK UNIFIED IDEOGRAPH-5A03
        0x6B6A, # G+6133  U+6B6A  uni6B6A,  CJK UNIFIED IDEOGRAPH-6B6A
        0x77EE, # G+6134  U+77EE  uni77EE,  CJK UNIFIED IDEOGRAPH-77EE
        0x5916, # G+6135  U+5916  uni5916,  CJK UNIFIED IDEOGRAPH-5916
        0x5D6C, # G+6136  U+5D6C  uni5D6C,  CJK UNIFIED IDEOGRAPH-5D6C
        0x5DCD, # G+6137  U+5DCD  uni5DCD,  CJK UNIFIED IDEOGRAPH-5DCD
        0x7325, # G+6138  U+7325  uni7325,  CJK UNIFIED IDEOGRAPH-7325
        0x754F, # G+6139  U+754F  uni754F,  CJK UNIFIED IDEOGRAPH-754F
        0x50E5, # G+6140  U+50E5  uni50E5,  CJK UNIFIED IDEOGRAPH-50E5
        0x51F9, # G+6141  U+51F9  uni51F9,  CJK UNIFIED IDEOGRAPH-51F9
        0x582F, # G+6142  U+582F  uni582F,  CJK UNIFIED IDEOGRAPH-582F
        0x592D, # G+6143  U+592D  uni592D,  CJK UNIFIED IDEOGRAPH-592D
        0x5996, # G+6144  U+5996  uni5996,  CJK UNIFIED IDEOGRAPH-5996
        0x59DA, # G+6145  U+59DA  uni59DA,  CJK UNIFIED IDEOGRAPH-59DA
        0x5BE5, # G+6146  U+5BE5  uni5BE5,  CJK UNIFIED IDEOGRAPH-5BE5
        0x5DA2, # G+6147  U+5DA2  uni5DA2,  CJK UNIFIED IDEOGRAPH-5DA2
        0x62D7, # G+6148  U+62D7  uni62D7,  CJK UNIFIED IDEOGRAPH-62D7
        0x6416, # G+6149  U+6416  uni6416,  CJK UNIFIED IDEOGRAPH-6416
        0x6493, # G+6150  U+6493  uni6493,  CJK UNIFIED IDEOGRAPH-6493
        0x64FE, # G+6151  U+64FE  uni64FE,  CJK UNIFIED IDEOGRAPH-64FE
        0x66DC, # G+6152  U+66DC  uni66DC,  CJK UNIFIED IDEOGRAPH-66DC
        0x6A48, # G+6153  U+6A48  uni6A48,  CJK UNIFIED IDEOGRAPH-6A48
        0x71FF, # G+6154  U+71FF  uni71FF,  CJK UNIFIED IDEOGRAPH-71FF
        0x7464, # G+6155  U+7464  uni7464,  CJK UNIFIED IDEOGRAPH-7464
        0x7A88, # G+6156  U+7A88  uni7A88,  CJK UNIFIED IDEOGRAPH-7A88
        0x7AAF, # G+6157  U+7AAF  uni7AAF,  CJK UNIFIED IDEOGRAPH-7AAF
        0x7E47, # G+6158  U+7E47  uni7E47,  CJK UNIFIED IDEOGRAPH-7E47
        0x7E5E, # G+6159  U+7E5E  uni7E5E,  CJK UNIFIED IDEOGRAPH-7E5E
        0x8000, # G+6160  U+8000  uni8000,  CJK UNIFIED IDEOGRAPH-8000
        0x8170, # G+6161  U+8170  uni8170,  CJK UNIFIED IDEOGRAPH-8170
        0x87EF, # G+6162  U+87EF  uni87EF,  CJK UNIFIED IDEOGRAPH-87EF
        0x8981, # G+6163  U+8981  uni8981,  CJK UNIFIED IDEOGRAPH-8981
        0x8B20, # G+6164  U+8B20  uni8B20,  CJK UNIFIED IDEOGRAPH-8B20
        0x9059, # G+6165  U+9059  uni9059,  CJK UNIFIED IDEOGRAPH-9059
        0x9080, # G+6166  U+9080  uni9080,  CJK UNIFIED IDEOGRAPH-9080
        0x9952, # G+6167  U+9952  uni9952,  CJK UNIFIED IDEOGRAPH-9952
        0x617E, # G+6168  U+617E  uni617E,  CJK UNIFIED IDEOGRAPH-617E
        0x6B32, # G+6169  U+6B32  uni6B32,  CJK UNIFIED IDEOGRAPH-6B32
        0x6D74, # G+6170  U+6D74  uni6D74,  CJK UNIFIED IDEOGRAPH-6D74
        0x7E1F, # G+6171  U+7E1F  uni7E1F,  CJK UNIFIED IDEOGRAPH-7E1F
        0x8925, # G+6172  U+8925  uni8925,  CJK UNIFIED IDEOGRAPH-8925
        0x8FB1, # G+6173  U+8FB1  uni8FB1,  CJK UNIFIED IDEOGRAPH-8FB1
        0x4FD1, # G+6174  U+4FD1  uni4FD1,  CJK UNIFIED IDEOGRAPH-4FD1
        0x50AD, # G+6175  U+50AD  uni50AD,  CJK UNIFIED IDEOGRAPH-50AD
        0x5197, # G+6176  U+5197  uni5197,  CJK UNIFIED IDEOGRAPH-5197
        0x52C7, # G+6177  U+52C7  uni52C7,  CJK UNIFIED IDEOGRAPH-52C7
        0x57C7, # G+6178  U+57C7  uni57C7,  CJK UNIFIED IDEOGRAPH-57C7
        0x5889, # G+6179  U+5889  uni5889,  CJK UNIFIED IDEOGRAPH-5889
        0x5BB9, # G+6180  U+5BB9  uni5BB9,  CJK UNIFIED IDEOGRAPH-5BB9
        0x5EB8, # G+6181  U+5EB8  uni5EB8,  CJK UNIFIED IDEOGRAPH-5EB8
        0x6142, # G+6182  U+6142  uni6142,  CJK UNIFIED IDEOGRAPH-6142
        0x6995, # G+6183  U+6995  uni6995,  CJK UNIFIED IDEOGRAPH-6995
        0x6D8C, # G+6184  U+6D8C  uni6D8C,  CJK UNIFIED IDEOGRAPH-6D8C
        0x6E67, # G+6185  U+6E67  uni6E67,  CJK UNIFIED IDEOGRAPH-6E67
        0x6EB6, # G+6186  U+6EB6  uni6EB6,  CJK UNIFIED IDEOGRAPH-6EB6
        0x7194, # G+6187  U+7194  uni7194,  CJK UNIFIED IDEOGRAPH-7194
        0x7462, # G+6188  U+7462  uni7462,  CJK UNIFIED IDEOGRAPH-7462
        0x7528, # G+6189  U+7528  uni7528,  CJK UNIFIED IDEOGRAPH-7528
        0x752C, # G+6190  U+752C  uni752C,  CJK UNIFIED IDEOGRAPH-752C
        0x8073, # G+6191  U+8073  uni8073,  CJK UNIFIED IDEOGRAPH-8073
        0x8338, # G+6192  U+8338  uni8338,  CJK UNIFIED IDEOGRAPH-8338
        0x84C9, # G+6193  U+84C9  uni84C9,  CJK UNIFIED IDEOGRAPH-84C9
        0x8E0A, # G+6194  U+8E0A  uni8E0A,  CJK UNIFIED IDEOGRAPH-8E0A
        0x9394, # G+6195  U+9394  uni9394,  CJK UNIFIED IDEOGRAPH-9394
        0x93DE, # G+6196  U+93DE  uni93DE,  CJK UNIFIED IDEOGRAPH-93DE
        0x4E8E, # G+6197  U+4E8E  uni4E8E,  CJK UNIFIED IDEOGRAPH-4E8E
        0x4F51, # G+6198  U+4F51  uni4F51,  CJK UNIFIED IDEOGRAPH-4F51
        0x5076, # G+6199  U+5076  uni5076,  CJK UNIFIED IDEOGRAPH-5076
        0x512A, # G+6200  U+512A  uni512A,  CJK UNIFIED IDEOGRAPH-512A
        0x53C8, # G+6201  U+53C8  uni53C8,  CJK UNIFIED IDEOGRAPH-53C8
        0x53CB, # G+6202  U+53CB  uni53CB,  CJK UNIFIED IDEOGRAPH-53CB
        0x53F3, # G+6203  U+53F3  uni53F3,  CJK UNIFIED IDEOGRAPH-53F3
        0x5B87, # G+6204  U+5B87  uni5B87,  CJK UNIFIED IDEOGRAPH-5B87
        0x5BD3, # G+6205  U+5BD3  uni5BD3,  CJK UNIFIED IDEOGRAPH-5BD3
        0x5C24, # G+6206  U+5C24  uni5C24,  CJK UNIFIED IDEOGRAPH-5C24
        0x611A, # G+6207  U+611A  uni611A,  CJK UNIFIED IDEOGRAPH-611A
        0x6182, # G+6208  U+6182  uni6182,  CJK UNIFIED IDEOGRAPH-6182
        0x65F4, # G+6209  U+65F4  uni65F4,  CJK UNIFIED IDEOGRAPH-65F4
        0x725B, # G+6210  U+725B  uni725B,  CJK UNIFIED IDEOGRAPH-725B
        0x7397, # G+6211  U+7397  uni7397,  CJK UNIFIED IDEOGRAPH-7397
        0x7440, # G+6212  U+7440  uni7440,  CJK UNIFIED IDEOGRAPH-7440
        0x76C2, # G+6213  U+76C2  uni76C2,  CJK UNIFIED IDEOGRAPH-76C2
        0x7950, # G+6214  U+7950  uni7950,  CJK UNIFIED IDEOGRAPH-7950
        0x7991, # G+6215  U+7991  uni7991,  CJK UNIFIED IDEOGRAPH-7991
        0x79B9, # G+6216  U+79B9  uni79B9,  CJK UNIFIED IDEOGRAPH-79B9
        0x7D06, # G+6217  U+7D06  uni7D06,  CJK UNIFIED IDEOGRAPH-7D06
        0x7FBD, # G+6218  U+7FBD  uni7FBD,  CJK UNIFIED IDEOGRAPH-7FBD
        0x828B, # G+6219  U+828B  uni828B,  CJK UNIFIED IDEOGRAPH-828B
        0x85D5, # G+6220  U+85D5  uni85D5,  CJK UNIFIED IDEOGRAPH-85D5
        0x865E, # G+6221  U+865E  uni865E,  CJK UNIFIED IDEOGRAPH-865E
        0x8FC2, # G+6222  U+8FC2  uni8FC2,  CJK UNIFIED IDEOGRAPH-8FC2
        0x9047, # G+6223  U+9047  uni9047,  CJK UNIFIED IDEOGRAPH-9047
        0x90F5, # G+6224  U+90F5  uni90F5,  CJK UNIFIED IDEOGRAPH-90F5
        0x91EA, # G+6225  U+91EA  uni91EA,  CJK UNIFIED IDEOGRAPH-91EA
        0x9685, # G+6226  U+9685  uni9685,  CJK UNIFIED IDEOGRAPH-9685
        0x96E8, # G+6227  U+96E8  uni96E8,  CJK UNIFIED IDEOGRAPH-96E8
        0x96E9, # G+6228  U+96E9  uni96E9,  CJK UNIFIED IDEOGRAPH-96E9
        0x52D6, # G+6229  U+52D6  uni52D6,  CJK UNIFIED IDEOGRAPH-52D6
        0x5F67, # G+6230  U+5F67  uni5F67,  CJK UNIFIED IDEOGRAPH-5F67
        0x65ED, # G+6231  U+65ED  uni65ED,  CJK UNIFIED IDEOGRAPH-65ED
        0x6631, # G+6232  U+6631  uni6631,  CJK UNIFIED IDEOGRAPH-6631
        0x682F, # G+6233  U+682F  uni682F,  CJK UNIFIED IDEOGRAPH-682F
        0x715C, # G+6234  U+715C  uni715C,  CJK UNIFIED IDEOGRAPH-715C
        0x7A36, # G+6235  U+7A36  uni7A36,  CJK UNIFIED IDEOGRAPH-7A36
        0x90C1, # G+6236  U+90C1  uni90C1,  CJK UNIFIED IDEOGRAPH-90C1
        0x980A, # G+6237  U+980A  uni980A,  CJK UNIFIED IDEOGRAPH-980A
        0x4E91, # G+6238  U+4E91  uni4E91,  CJK UNIFIED IDEOGRAPH-4E91
        0x6A52, # G+6239  U+6A52  uni6A52,  CJK UNIFIED IDEOGRAPH-6A52
        0x6B9E, # G+6240  U+6B9E  uni6B9E,  CJK UNIFIED IDEOGRAPH-6B9E
        0x6F90, # G+6241  U+6F90  uni6F90,  CJK UNIFIED IDEOGRAPH-6F90
        0x7189, # G+6242  U+7189  uni7189,  CJK UNIFIED IDEOGRAPH-7189
        0x8018, # G+6243  U+8018  uni8018,  CJK UNIFIED IDEOGRAPH-8018
        0x82B8, # G+6244  U+82B8  uni82B8,  CJK UNIFIED IDEOGRAPH-82B8
        0x8553, # G+6245  U+8553  uni8553,  CJK UNIFIED IDEOGRAPH-8553
        0x904B, # G+6246  U+904B  uni904B,  CJK UNIFIED IDEOGRAPH-904B
        0x9695, # G+6247  U+9695  uni9695,  CJK UNIFIED IDEOGRAPH-9695
        0x96F2, # G+6248  U+96F2  uni96F2,  CJK UNIFIED IDEOGRAPH-96F2
        0x97FB, # G+6249  U+97FB  uni97FB,  CJK UNIFIED IDEOGRAPH-97FB
        0x851A, # G+6250  U+851A  uni851A,  CJK UNIFIED IDEOGRAPH-851A
        0x9B31, # G+6251  U+9B31  uni9B31,  CJK UNIFIED IDEOGRAPH-9B31
        0x4E90, # G+6252  U+4E90  uni4E90,  CJK UNIFIED IDEOGRAPH-4E90
        0x718A, # G+6253  U+718A  uni718A,  CJK UNIFIED IDEOGRAPH-718A
        0x96C4, # G+6254  U+96C4  uni96C4,  CJK UNIFIED IDEOGRAPH-96C4
        0x5143, # G+6255  U+5143  uni5143,  CJK UNIFIED IDEOGRAPH-5143
        0x539F, # G+6256  U+539F  uni539F,  CJK UNIFIED IDEOGRAPH-539F
        0x54E1, # G+6257  U+54E1  uni54E1,  CJK UNIFIED IDEOGRAPH-54E1
        0x5713, # G+6258  U+5713  uni5713,  CJK UNIFIED IDEOGRAPH-5713
        0x5712, # G+6259  U+5712  uni5712,  CJK UNIFIED IDEOGRAPH-5712
        0x57A3, # G+6260  U+57A3  uni57A3,  CJK UNIFIED IDEOGRAPH-57A3
        0x5A9B, # G+6261  U+5A9B  uni5A9B,  CJK UNIFIED IDEOGRAPH-5A9B
        0x5AC4, # G+6262  U+5AC4  uni5AC4,  CJK UNIFIED IDEOGRAPH-5AC4
        0x5BC3, # G+6263  U+5BC3  uni5BC3,  CJK UNIFIED IDEOGRAPH-5BC3
        0x6028, # G+6264  U+6028  uni6028,  CJK UNIFIED IDEOGRAPH-6028
        0x613F, # G+6265  U+613F  uni613F,  CJK UNIFIED IDEOGRAPH-613F
        0x63F4, # G+6266  U+63F4  uni63F4,  CJK UNIFIED IDEOGRAPH-63F4
        0x6C85, # G+6267  U+6C85  uni6C85,  CJK UNIFIED IDEOGRAPH-6C85
        0x6D39, # G+6268  U+6D39  uni6D39,  CJK UNIFIED IDEOGRAPH-6D39
        0x6E72, # G+6269  U+6E72  uni6E72,  CJK UNIFIED IDEOGRAPH-6E72
        0x6E90, # G+6270  U+6E90  uni6E90,  CJK UNIFIED IDEOGRAPH-6E90
        0x7230, # G+6271  U+7230  uni7230,  CJK UNIFIED IDEOGRAPH-7230
        0x733F, # G+6272  U+733F  uni733F,  CJK UNIFIED IDEOGRAPH-733F
        0x7457, # G+6273  U+7457  uni7457,  CJK UNIFIED IDEOGRAPH-7457
        0x82D1, # G+6274  U+82D1  uni82D1,  CJK UNIFIED IDEOGRAPH-82D1
        0x8881, # G+6275  U+8881  uni8881,  CJK UNIFIED IDEOGRAPH-8881
        0x8F45, # G+6276  U+8F45  uni8F45,  CJK UNIFIED IDEOGRAPH-8F45
        0x9060, # G+6277  U+9060  uni9060,  CJK UNIFIED IDEOGRAPH-9060
        0x9662, # G+6278  U+9662  uni9662,  CJK UNIFIED IDEOGRAPH-9662
        0x9858, # G+6279  U+9858  uni9858,  CJK UNIFIED IDEOGRAPH-9858
        0x9D1B, # G+6280  U+9D1B  uni9D1B,  CJK UNIFIED IDEOGRAPH-9D1B
        0x6708, # G+6281  U+6708  uni6708,  CJK UNIFIED IDEOGRAPH-6708
        0x8D8A, # G+6282  U+8D8A  uni8D8A,  CJK UNIFIED IDEOGRAPH-8D8A
        0x925E, # G+6283  U+925E  uni925E,  CJK UNIFIED IDEOGRAPH-925E
        0x4F4D, # G+6284  U+4F4D  uni4F4D,  CJK UNIFIED IDEOGRAPH-4F4D
        0x5049, # G+6285  U+5049  uni5049,  CJK UNIFIED IDEOGRAPH-5049
        0x50DE, # G+6286  U+50DE  uni50DE,  CJK UNIFIED IDEOGRAPH-50DE
        0x5371, # G+6287  U+5371  uni5371,  CJK UNIFIED IDEOGRAPH-5371
        0x570D, # G+6288  U+570D  uni570D,  CJK UNIFIED IDEOGRAPH-570D
        0x59D4, # G+6289  U+59D4  uni59D4,  CJK UNIFIED IDEOGRAPH-59D4
        0x5A01, # G+6290  U+5A01  uni5A01,  CJK UNIFIED IDEOGRAPH-5A01
        0x5C09, # G+6291  U+5C09  uni5C09,  CJK UNIFIED IDEOGRAPH-5C09
        0x6170, # G+6292  U+6170  uni6170,  CJK UNIFIED IDEOGRAPH-6170
        0x6690, # G+6293  U+6690  uni6690,  CJK UNIFIED IDEOGRAPH-6690
        0x6E2D, # G+6294  U+6E2D  uni6E2D,  CJK UNIFIED IDEOGRAPH-6E2D
        0x7232, # G+6295  U+7232  uni7232,  CJK UNIFIED IDEOGRAPH-7232
        0x744B, # G+6296  U+744B  uni744B,  CJK UNIFIED IDEOGRAPH-744B
        0x7DEF, # G+6297  U+7DEF  uni7DEF,  CJK UNIFIED IDEOGRAPH-7DEF
        0x80C3, # G+6298  U+80C3  uni80C3,  CJK UNIFIED IDEOGRAPH-80C3
        0x840E, # G+6299  U+840E  uni840E,  CJK UNIFIED IDEOGRAPH-840E
        0x8466, # G+6300  U+8466  uni8466,  CJK UNIFIED IDEOGRAPH-8466
        0x853F, # G+6301  U+853F  uni853F,  CJK UNIFIED IDEOGRAPH-853F
        0x875F, # G+6302  U+875F  uni875F,  CJK UNIFIED IDEOGRAPH-875F
        0x885B, # G+6303  U+885B  uni885B,  CJK UNIFIED IDEOGRAPH-885B
        0x8918, # G+6304  U+8918  uni8918,  CJK UNIFIED IDEOGRAPH-8918
        0x8B02, # G+6305  U+8B02  uni8B02,  CJK UNIFIED IDEOGRAPH-8B02
        0x9055, # G+6306  U+9055  uni9055,  CJK UNIFIED IDEOGRAPH-9055
        0x97CB, # G+6307  U+97CB  uni97CB,  CJK UNIFIED IDEOGRAPH-97CB
        0x9B4F, # G+6308  U+9B4F  uni9B4F,  CJK UNIFIED IDEOGRAPH-9B4F
        0x4E73, # G+6309  U+4E73  uni4E73,  CJK UNIFIED IDEOGRAPH-4E73
        0x4F91, # G+6310  U+4F91  uni4F91,  CJK UNIFIED IDEOGRAPH-4F91
        0x5112, # G+6311  U+5112  uni5112,  CJK UNIFIED IDEOGRAPH-5112
        0x516A, # G+6312  U+516A  uni516A,  CJK UNIFIED IDEOGRAPH-516A
        0x552F, # G+6313  U+552F  uni552F,  CJK UNIFIED IDEOGRAPH-552F
        0x55A9, # G+6314  U+55A9  uni55A9,  CJK UNIFIED IDEOGRAPH-55A9
        0x5B7A, # G+6315  U+5B7A  uni5B7A,  CJK UNIFIED IDEOGRAPH-5B7A
        0x5BA5, # G+6316  U+5BA5  uni5BA5,  CJK UNIFIED IDEOGRAPH-5BA5
        0x5E7C, # G+6317  U+5E7C  uni5E7C,  CJK UNIFIED IDEOGRAPH-5E7C
        0x5E7D, # G+6318  U+5E7D  uni5E7D,  CJK UNIFIED IDEOGRAPH-5E7D
        0x5EBE, # G+6319  U+5EBE  uni5EBE,  CJK UNIFIED IDEOGRAPH-5EBE
        0x60A0, # G+6320  U+60A0  uni60A0,  CJK UNIFIED IDEOGRAPH-60A0
        0x60DF, # G+6321  U+60DF  uni60DF,  CJK UNIFIED IDEOGRAPH-60DF
        0x6108, # G+6322  U+6108  uni6108,  CJK UNIFIED IDEOGRAPH-6108
        0x6109, # G+6323  U+6109  uni6109,  CJK UNIFIED IDEOGRAPH-6109
        0x63C4, # G+6324  U+63C4  uni63C4,  CJK UNIFIED IDEOGRAPH-63C4
        0x6538, # G+6325  U+6538  uni6538,  CJK UNIFIED IDEOGRAPH-6538
        0x6709, # G+6326  U+6709  uni6709,  CJK UNIFIED IDEOGRAPH-6709
        0x67D4, # G+6327  U+67D4  uni67D4,  CJK UNIFIED IDEOGRAPH-67D4
        0x67DA, # G+6328  U+67DA  uni67DA,  CJK UNIFIED IDEOGRAPH-67DA
        0x6961, # G+6329  U+6961  uni6961,  CJK UNIFIED IDEOGRAPH-6961
        0x6962, # G+6330  U+6962  uni6962,  CJK UNIFIED IDEOGRAPH-6962
        0x6CB9, # G+6331  U+6CB9  uni6CB9,  CJK UNIFIED IDEOGRAPH-6CB9
        0x6D27, # G+6332  U+6D27  uni6D27,  CJK UNIFIED IDEOGRAPH-6D27
        0x6E38, # G+6333  U+6E38  uni6E38,  CJK UNIFIED IDEOGRAPH-6E38
        0x6FE1, # G+6334  U+6FE1  uni6FE1,  CJK UNIFIED IDEOGRAPH-6FE1
        0x7336, # G+6335  U+7336  uni7336,  CJK UNIFIED IDEOGRAPH-7336
        0x7337, # G+6336  U+7337  uni7337,  CJK UNIFIED IDEOGRAPH-7337
        0x745C, # G+6337  U+745C  uni745C,  CJK UNIFIED IDEOGRAPH-745C
        0x7531, # G+6338  U+7531  uni7531,  CJK UNIFIED IDEOGRAPH-7531
        0x7652, # G+6339  U+7652  uni7652,  CJK UNIFIED IDEOGRAPH-7652
        0x7DAD, # G+6340  U+7DAD  uni7DAD,  CJK UNIFIED IDEOGRAPH-7DAD
        0x81FE, # G+6341  U+81FE  uni81FE,  CJK UNIFIED IDEOGRAPH-81FE
        0x8438, # G+6342  U+8438  uni8438,  CJK UNIFIED IDEOGRAPH-8438
        0x88D5, # G+6343  U+88D5  uni88D5,  CJK UNIFIED IDEOGRAPH-88D5
        0x8A98, # G+6344  U+8A98  uni8A98,  CJK UNIFIED IDEOGRAPH-8A98
        0x8ADB, # G+6345  U+8ADB  uni8ADB,  CJK UNIFIED IDEOGRAPH-8ADB
        0x8AED, # G+6346  U+8AED  uni8AED,  CJK UNIFIED IDEOGRAPH-8AED
        0x8E30, # G+6347  U+8E30  uni8E30,  CJK UNIFIED IDEOGRAPH-8E30
        0x8E42, # G+6348  U+8E42  uni8E42,  CJK UNIFIED IDEOGRAPH-8E42
        0x904A, # G+6349  U+904A  uni904A,  CJK UNIFIED IDEOGRAPH-904A
        0x903E, # G+6350  U+903E  uni903E,  CJK UNIFIED IDEOGRAPH-903E
        0x907A, # G+6351  U+907A  uni907A,  CJK UNIFIED IDEOGRAPH-907A
        0x9149, # G+6352  U+9149  uni9149,  CJK UNIFIED IDEOGRAPH-9149
        0x91C9, # G+6353  U+91C9  uni91C9,  CJK UNIFIED IDEOGRAPH-91C9
        0x936E, # G+6354  U+936E  uni936E,  CJK UNIFIED IDEOGRAPH-936E
        0x5809, # G+6355  U+5809  uni5809,  CJK UNIFIED IDEOGRAPH-5809
        0x6BD3, # G+6356  U+6BD3  uni6BD3,  CJK UNIFIED IDEOGRAPH-6BD3
        0x8089, # G+6357  U+8089  uni8089,  CJK UNIFIED IDEOGRAPH-8089
        0x80B2, # G+6358  U+80B2  uni80B2,  CJK UNIFIED IDEOGRAPH-80B2
        0x5141, # G+6359  U+5141  uni5141,  CJK UNIFIED IDEOGRAPH-5141
        0x596B, # G+6360  U+596B  uni596B,  CJK UNIFIED IDEOGRAPH-596B
        0x5C39, # G+6361  U+5C39  uni5C39,  CJK UNIFIED IDEOGRAPH-5C39
        0x6F64, # G+6362  U+6F64  uni6F64,  CJK UNIFIED IDEOGRAPH-6F64
        0x73A7, # G+6363  U+73A7  uni73A7,  CJK UNIFIED IDEOGRAPH-73A7
        0x80E4, # G+6364  U+80E4  uni80E4,  CJK UNIFIED IDEOGRAPH-80E4
        0x8D07, # G+6365  U+8D07  uni8D07,  CJK UNIFIED IDEOGRAPH-8D07
        0x9217, # G+6366  U+9217  uni9217,  CJK UNIFIED IDEOGRAPH-9217
        0x958F, # G+6367  U+958F  uni958F,  CJK UNIFIED IDEOGRAPH-958F
        0x807F, # G+6368  U+807F  uni807F,  CJK UNIFIED IDEOGRAPH-807F
        0x620E, # G+6369  U+620E  uni620E,  CJK UNIFIED IDEOGRAPH-620E
        0x701C, # G+6370  U+701C  uni701C,  CJK UNIFIED IDEOGRAPH-701C
        0x7D68, # G+6371  U+7D68  uni7D68,  CJK UNIFIED IDEOGRAPH-7D68
        0x878D, # G+6372  U+878D  uni878D,  CJK UNIFIED IDEOGRAPH-878D
        0x57A0, # G+6373  U+57A0  uni57A0,  CJK UNIFIED IDEOGRAPH-57A0
        0x6069, # G+6374  U+6069  uni6069,  CJK UNIFIED IDEOGRAPH-6069
        0x6147, # G+6375  U+6147  uni6147,  CJK UNIFIED IDEOGRAPH-6147
        0x6BB7, # G+6376  U+6BB7  uni6BB7,  CJK UNIFIED IDEOGRAPH-6BB7
        0x8ABE, # G+6377  U+8ABE  uni8ABE,  CJK UNIFIED IDEOGRAPH-8ABE
        0x9280, # G+6378  U+9280  uni9280,  CJK UNIFIED IDEOGRAPH-9280
        0x96B1, # G+6379  U+96B1  uni96B1,  CJK UNIFIED IDEOGRAPH-96B1
        0x4E59, # G+6380  U+4E59  uni4E59,  CJK UNIFIED IDEOGRAPH-4E59
        0x541F, # G+6381  U+541F  uni541F,  CJK UNIFIED IDEOGRAPH-541F
        0x6DEB, # G+6382  U+6DEB  uni6DEB,  CJK UNIFIED IDEOGRAPH-6DEB
        0x852D, # G+6383  U+852D  uni852D,  CJK UNIFIED IDEOGRAPH-852D
        0x9670, # G+6384  U+9670  uni9670,  CJK UNIFIED IDEOGRAPH-9670
        0x97F3, # G+6385  U+97F3  uni97F3,  CJK UNIFIED IDEOGRAPH-97F3
        0x98EE, # G+6386  U+98EE  uni98EE,  CJK UNIFIED IDEOGRAPH-98EE
        0x63D6, # G+6387  U+63D6  uni63D6,  CJK UNIFIED IDEOGRAPH-63D6
        0x6CE3, # G+6388  U+6CE3  uni6CE3,  CJK UNIFIED IDEOGRAPH-6CE3
        0x9091, # G+6389  U+9091  uni9091,  CJK UNIFIED IDEOGRAPH-9091
        0x51DD, # G+6390  U+51DD  uni51DD,  CJK UNIFIED IDEOGRAPH-51DD
        0x61C9, # G+6391  U+61C9  uni61C9,  CJK UNIFIED IDEOGRAPH-61C9
        0x81BA, # G+6392  U+81BA  uni81BA,  CJK UNIFIED IDEOGRAPH-81BA
        0x9DF9, # G+6393  U+9DF9  uni9DF9,  CJK UNIFIED IDEOGRAPH-9DF9
        0x4F9D, # G+6394  U+4F9D  uni4F9D,  CJK UNIFIED IDEOGRAPH-4F9D
        0x501A, # G+6395  U+501A  uni501A,  CJK UNIFIED IDEOGRAPH-501A
        0x5100, # G+6396  U+5100  uni5100,  CJK UNIFIED IDEOGRAPH-5100
        0x5B9C, # G+6397  U+5B9C  uni5B9C,  CJK UNIFIED IDEOGRAPH-5B9C
        0x610F, # G+6398  U+610F  uni610F,  CJK UNIFIED IDEOGRAPH-610F
        0x61FF, # G+6399  U+61FF  uni61FF,  CJK UNIFIED IDEOGRAPH-61FF
        0x64EC, # G+6400  U+64EC  uni64EC,  CJK UNIFIED IDEOGRAPH-64EC
        0x6905, # G+6401  U+6905  uni6905,  CJK UNIFIED IDEOGRAPH-6905
        0x6BC5, # G+6402  U+6BC5  uni6BC5,  CJK UNIFIED IDEOGRAPH-6BC5
        0x7591, # G+6403  U+7591  uni7591,  CJK UNIFIED IDEOGRAPH-7591
        0x77E3, # G+6404  U+77E3  uni77E3,  CJK UNIFIED IDEOGRAPH-77E3
        0x7FA9, # G+6405  U+7FA9  uni7FA9,  CJK UNIFIED IDEOGRAPH-7FA9
        0x8264, # G+6406  U+8264  uni8264,  CJK UNIFIED IDEOGRAPH-8264
        0x858F, # G+6407  U+858F  uni858F,  CJK UNIFIED IDEOGRAPH-858F
        0x87FB, # G+6408  U+87FB  uni87FB,  CJK UNIFIED IDEOGRAPH-87FB
        0x8863, # G+6409  U+8863  uni8863,  CJK UNIFIED IDEOGRAPH-8863
        0x8ABC, # G+6410  U+8ABC  uni8ABC,  CJK UNIFIED IDEOGRAPH-8ABC
        0x8B70, # G+6411  U+8B70  uni8B70,  CJK UNIFIED IDEOGRAPH-8B70
        0x91AB, # G+6412  U+91AB  uni91AB,  CJK UNIFIED IDEOGRAPH-91AB
        0x4E8C, # G+6413  U+4E8C  uni4E8C,  CJK UNIFIED IDEOGRAPH-4E8C
        0x4EE5, # G+6414  U+4EE5  uni4EE5,  CJK UNIFIED IDEOGRAPH-4EE5
        0x4F0A, # G+6415  U+4F0A  uni4F0A,  CJK UNIFIED IDEOGRAPH-4F0A
        0x5937, # G+6416  U+5937  uni5937,  CJK UNIFIED IDEOGRAPH-5937
        0x59E8, # G+6417  U+59E8  uni59E8,  CJK UNIFIED IDEOGRAPH-59E8
        0x5DF2, # G+6418  U+5DF2  uni5DF2,  CJK UNIFIED IDEOGRAPH-5DF2
        0x5F1B, # G+6419  U+5F1B  uni5F1B,  CJK UNIFIED IDEOGRAPH-5F1B
        0x5F5B, # G+6420  U+5F5B  uni5F5B,  CJK UNIFIED IDEOGRAPH-5F5B
        0x6021, # G+6421  U+6021  uni6021,  CJK UNIFIED IDEOGRAPH-6021
        0x723E, # G+6422  U+723E  uni723E,  CJK UNIFIED IDEOGRAPH-723E
        0x73E5, # G+6423  U+73E5  uni73E5,  CJK UNIFIED IDEOGRAPH-73E5
        0xF962, # G+6424  U+F962  uniF962,  CJK COMPATIBILITY IDEOGRAPH-F962
        0x75CD, # G+6425  U+75CD  uni75CD,  CJK UNIFIED IDEOGRAPH-75CD
        0x79FB, # G+6426  U+79FB  uni79FB,  CJK UNIFIED IDEOGRAPH-79FB
        0x800C, # G+6427  U+800C  uni800C,  CJK UNIFIED IDEOGRAPH-800C
        0x8033, # G+6428  U+8033  uni8033,  CJK UNIFIED IDEOGRAPH-8033
        0x8084, # G+6429  U+8084  uni8084,  CJK UNIFIED IDEOGRAPH-8084
        0x82E1, # G+6430  U+82E1  uni82E1,  CJK UNIFIED IDEOGRAPH-82E1
        0x8351, # G+6431  U+8351  uni8351,  CJK UNIFIED IDEOGRAPH-8351
        0x8CBD, # G+6432  U+8CBD  uni8CBD,  CJK UNIFIED IDEOGRAPH-8CBD
        0x8CB3, # G+6433  U+8CB3  uni8CB3,  CJK UNIFIED IDEOGRAPH-8CB3
        0x9087, # G+6434  U+9087  uni9087,  CJK UNIFIED IDEOGRAPH-9087
        0x98F4, # G+6435  U+98F4  uni98F4,  CJK UNIFIED IDEOGRAPH-98F4
        0x990C, # G+6436  U+990C  uni990C,  CJK UNIFIED IDEOGRAPH-990C
        0x7037, # G+6437  U+7037  uni7037,  CJK UNIFIED IDEOGRAPH-7037
        0x76CA, # G+6438  U+76CA  uni76CA,  CJK UNIFIED IDEOGRAPH-76CA
        0x7FCA, # G+6439  U+7FCA  uni7FCA,  CJK UNIFIED IDEOGRAPH-7FCA
        0x7FCC, # G+6440  U+7FCC  uni7FCC,  CJK UNIFIED IDEOGRAPH-7FCC
        0x7FFC, # G+6441  U+7FFC  uni7FFC,  CJK UNIFIED IDEOGRAPH-7FFC
        0x8B1A, # G+6442  U+8B1A  uni8B1A,  CJK UNIFIED IDEOGRAPH-8B1A
        0x4EBA, # G+6443  U+4EBA  uni4EBA,  CJK UNIFIED IDEOGRAPH-4EBA
        0x4EC1, # G+6444  U+4EC1  uni4EC1,  CJK UNIFIED IDEOGRAPH-4EC1
        0x5203, # G+6445  U+5203  uni5203,  CJK UNIFIED IDEOGRAPH-5203
        0x5370, # G+6446  U+5370  uni5370,  CJK UNIFIED IDEOGRAPH-5370
        0xF99E, # G+6447  U+F99E  uniF99E,  CJK COMPATIBILITY IDEOGRAPH-F99E
        0x56E0, # G+6448  U+56E0  uni56E0,  CJK UNIFIED IDEOGRAPH-56E0
        0x59FB, # G+6449  U+59FB  uni59FB,  CJK UNIFIED IDEOGRAPH-59FB
        0x5BC5, # G+6450  U+5BC5  uni5BC5,  CJK UNIFIED IDEOGRAPH-5BC5
        0x5F15, # G+6451  U+5F15  uni5F15,  CJK UNIFIED IDEOGRAPH-5F15
        0x5FCD, # G+6452  U+5FCD  uni5FCD,  CJK UNIFIED IDEOGRAPH-5FCD
        0x6E6E, # G+6453  U+6E6E  uni6E6E,  CJK UNIFIED IDEOGRAPH-6E6E
        0x7D6A, # G+6454  U+7D6A  uni7D6A,  CJK UNIFIED IDEOGRAPH-7D6A
        0x8335, # G+6455  U+8335  uni8335,  CJK UNIFIED IDEOGRAPH-8335
        0x8693, # G+6456  U+8693  uni8693,  CJK UNIFIED IDEOGRAPH-8693
        0x8A8D, # G+6457  U+8A8D  uni8A8D,  CJK UNIFIED IDEOGRAPH-8A8D
        0x976D, # G+6458  U+976D  uni976D,  CJK UNIFIED IDEOGRAPH-976D
        0x9777, # G+6459  U+9777  uni9777,  CJK UNIFIED IDEOGRAPH-9777
        0x4E00, # G+6460  U+4E00  uni4E00,  CJK UNIFIED IDEOGRAPH-4E00
        0x4F5A, # G+6461  U+4F5A  uni4F5A,  CJK UNIFIED IDEOGRAPH-4F5A
        0x4F7E, # G+6462  U+4F7E  uni4F7E,  CJK UNIFIED IDEOGRAPH-4F7E
        0x58F9, # G+6463  U+58F9  uni58F9,  CJK UNIFIED IDEOGRAPH-58F9
        0x65E5, # G+6464  U+65E5  uni65E5,  CJK UNIFIED IDEOGRAPH-65E5
        0x6EA2, # G+6465  U+6EA2  uni6EA2,  CJK UNIFIED IDEOGRAPH-6EA2
        0x9038, # G+6466  U+9038  uni9038,  CJK UNIFIED IDEOGRAPH-9038
        0x93B0, # G+6467  U+93B0  uni93B0,  CJK UNIFIED IDEOGRAPH-93B0
        0x99B9, # G+6468  U+99B9  uni99B9,  CJK UNIFIED IDEOGRAPH-99B9
        0x4EFB, # G+6469  U+4EFB  uni4EFB,  CJK UNIFIED IDEOGRAPH-4EFB
        0x58EC, # G+6470  U+58EC  uni58EC,  CJK UNIFIED IDEOGRAPH-58EC
        0x598A, # G+6471  U+598A  uni598A,  CJK UNIFIED IDEOGRAPH-598A
        0x59D9, # G+6472  U+59D9  uni59D9,  CJK UNIFIED IDEOGRAPH-59D9
        0x6041, # G+6473  U+6041  uni6041,  CJK UNIFIED IDEOGRAPH-6041
        0x7A14, # G+6474  U+7A14  uni7A14,  CJK UNIFIED IDEOGRAPH-7A14
        0x834F, # G+6475  U+834F  uni834F,  CJK UNIFIED IDEOGRAPH-834F
        0x8CC3, # G+6476  U+8CC3  uni8CC3,  CJK UNIFIED IDEOGRAPH-8CC3
        0x5165, # G+6477  U+5165  uni5165,  CJK UNIFIED IDEOGRAPH-5165
        0x5344, # G+6478  U+5344  twentyhangzhou,  CJK UNIFIED IDEOGRAPH-5344
        0x4ECD, # G+6479  U+4ECD  uni4ECD,  CJK UNIFIED IDEOGRAPH-4ECD
        0x5269, # G+6480  U+5269  uni5269,  CJK UNIFIED IDEOGRAPH-5269
        0x5B55, # G+6481  U+5B55  uni5B55,  CJK UNIFIED IDEOGRAPH-5B55
        0x82BF, # G+6482  U+82BF  uni82BF,  CJK UNIFIED IDEOGRAPH-82BF
        0x4ED4, # G+6483  U+4ED4  uni4ED4,  CJK UNIFIED IDEOGRAPH-4ED4
        0x523A, # G+6484  U+523A  uni523A,  CJK UNIFIED IDEOGRAPH-523A
        0x54A8, # G+6485  U+54A8  uni54A8,  CJK UNIFIED IDEOGRAPH-54A8
        0x59C9, # G+6486  U+59C9  uni59C9,  CJK UNIFIED IDEOGRAPH-59C9
        0x59FF, # G+6487  U+59FF  uni59FF,  CJK UNIFIED IDEOGRAPH-59FF
        0x5B50, # G+6488  U+5B50  uni5B50,  CJK UNIFIED IDEOGRAPH-5B50
        0x5B57, # G+6489  U+5B57  uni5B57,  CJK UNIFIED IDEOGRAPH-5B57
        0x5B5C, # G+6490  U+5B5C  uni5B5C,  CJK UNIFIED IDEOGRAPH-5B5C
        0x6063, # G+6491  U+6063  uni6063,  CJK UNIFIED IDEOGRAPH-6063
        0x6148, # G+6492  U+6148  uni6148,  CJK UNIFIED IDEOGRAPH-6148
        0x6ECB, # G+6493  U+6ECB  uni6ECB,  CJK UNIFIED IDEOGRAPH-6ECB
        0x7099, # G+6494  U+7099  uni7099,  CJK UNIFIED IDEOGRAPH-7099
        0x716E, # G+6495  U+716E  uni716E,  CJK UNIFIED IDEOGRAPH-716E
        0x7386, # G+6496  U+7386  uni7386,  CJK UNIFIED IDEOGRAPH-7386
        0x74F7, # G+6497  U+74F7  uni74F7,  CJK UNIFIED IDEOGRAPH-74F7
        0x75B5, # G+6498  U+75B5  uni75B5,  CJK UNIFIED IDEOGRAPH-75B5
        0x78C1, # G+6499  U+78C1  uni78C1,  CJK UNIFIED IDEOGRAPH-78C1
        0x7D2B, # G+6500  U+7D2B  uni7D2B,  CJK UNIFIED IDEOGRAPH-7D2B
        0x8005, # G+6501  U+8005  uni8005,  CJK UNIFIED IDEOGRAPH-8005
        0x81EA, # G+6502  U+81EA  uni81EA,  CJK UNIFIED IDEOGRAPH-81EA
        0x8328, # G+6503  U+8328  uni8328,  CJK UNIFIED IDEOGRAPH-8328
        0x8517, # G+6504  U+8517  uni8517,  CJK UNIFIED IDEOGRAPH-8517
        0x85C9, # G+6505  U+85C9  uni85C9,  CJK UNIFIED IDEOGRAPH-85C9
        0x8AEE, # G+6506  U+8AEE  uni8AEE,  CJK UNIFIED IDEOGRAPH-8AEE
        0x8CC7, # G+6507  U+8CC7  uni8CC7,  CJK UNIFIED IDEOGRAPH-8CC7
        0x96CC, # G+6508  U+96CC  uni96CC,  CJK UNIFIED IDEOGRAPH-96CC
        0x4F5C, # G+6509  U+4F5C  uni4F5C,  CJK UNIFIED IDEOGRAPH-4F5C
        0x52FA, # G+6510  U+52FA  uni52FA,  CJK UNIFIED IDEOGRAPH-52FA
        0x56BC, # G+6511  U+56BC  uni56BC,  CJK UNIFIED IDEOGRAPH-56BC
        0x65AB, # G+6512  U+65AB  uni65AB,  CJK UNIFIED IDEOGRAPH-65AB
        0x6628, # G+6513  U+6628  uni6628,  CJK UNIFIED IDEOGRAPH-6628
        0x707C, # G+6514  U+707C  uni707C,  CJK UNIFIED IDEOGRAPH-707C
        0x70B8, # G+6515  U+70B8  uni70B8,  CJK UNIFIED IDEOGRAPH-70B8
        0x7235, # G+6516  U+7235  uni7235,  CJK UNIFIED IDEOGRAPH-7235
        0x7DBD, # G+6517  U+7DBD  uni7DBD,  CJK UNIFIED IDEOGRAPH-7DBD
        0x828D, # G+6518  U+828D  uni828D,  CJK UNIFIED IDEOGRAPH-828D
        0x914C, # G+6519  U+914C  uni914C,  CJK UNIFIED IDEOGRAPH-914C
        0x96C0, # G+6520  U+96C0  uni96C0,  CJK UNIFIED IDEOGRAPH-96C0
        0x9D72, # G+6521  U+9D72  uni9D72,  CJK UNIFIED IDEOGRAPH-9D72
        0x5B71, # G+6522  U+5B71  uni5B71,  CJK UNIFIED IDEOGRAPH-5B71
        0x68E7, # G+6523  U+68E7  uni68E7,  CJK UNIFIED IDEOGRAPH-68E7
        0x6B98, # G+6524  U+6B98  uni6B98,  CJK UNIFIED IDEOGRAPH-6B98
        0x6F7A, # G+6525  U+6F7A  uni6F7A,  CJK UNIFIED IDEOGRAPH-6F7A
        0x76DE, # G+6526  U+76DE  uni76DE,  CJK UNIFIED IDEOGRAPH-76DE
        0x5C91, # G+6527  U+5C91  uni5C91,  CJK UNIFIED IDEOGRAPH-5C91
        0x66AB, # G+6528  U+66AB  uni66AB,  CJK UNIFIED IDEOGRAPH-66AB
        0x6F5B, # G+6529  U+6F5B  uni6F5B,  CJK UNIFIED IDEOGRAPH-6F5B
        0x7BB4, # G+6530  U+7BB4  uni7BB4,  CJK UNIFIED IDEOGRAPH-7BB4
        0x7C2A, # G+6531  U+7C2A  uni7C2A,  CJK UNIFIED IDEOGRAPH-7C2A
        0x8836, # G+6532  U+8836  uni8836,  CJK UNIFIED IDEOGRAPH-8836
        0x96DC, # G+6533  U+96DC  uni96DC,  CJK UNIFIED IDEOGRAPH-96DC
        0x4E08, # G+6534  U+4E08  uni4E08,  CJK UNIFIED IDEOGRAPH-4E08
        0x4ED7, # G+6535  U+4ED7  uni4ED7,  CJK UNIFIED IDEOGRAPH-4ED7
        0x5320, # G+6536  U+5320  uni5320,  CJK UNIFIED IDEOGRAPH-5320
        0x5834, # G+6537  U+5834  uni5834,  CJK UNIFIED IDEOGRAPH-5834
        0x58BB, # G+6538  U+58BB  uni58BB,  CJK UNIFIED IDEOGRAPH-58BB
        0x58EF, # G+6539  U+58EF  uni58EF,  CJK UNIFIED IDEOGRAPH-58EF
        0x596C, # G+6540  U+596C  uni596C,  CJK UNIFIED IDEOGRAPH-596C
        0x5C07, # G+6541  U+5C07  uni5C07,  CJK UNIFIED IDEOGRAPH-5C07
        0x5E33, # G+6542  U+5E33  uni5E33,  CJK UNIFIED IDEOGRAPH-5E33
        0x5E84, # G+6543  U+5E84  uni5E84,  CJK UNIFIED IDEOGRAPH-5E84
        0x5F35, # G+6544  U+5F35  uni5F35,  CJK UNIFIED IDEOGRAPH-5F35
        0x638C, # G+6545  U+638C  uni638C,  CJK UNIFIED IDEOGRAPH-638C
        0x66B2, # G+6546  U+66B2  uni66B2,  CJK UNIFIED IDEOGRAPH-66B2
        0x6756, # G+6547  U+6756  uni6756,  CJK UNIFIED IDEOGRAPH-6756
        0x6A1F, # G+6548  U+6A1F  uni6A1F,  CJK UNIFIED IDEOGRAPH-6A1F
        0x6AA3, # G+6549  U+6AA3  uni6AA3,  CJK UNIFIED IDEOGRAPH-6AA3
        0x6B0C, # G+6550  U+6B0C  uni6B0C,  CJK UNIFIED IDEOGRAPH-6B0C
        0x6F3F, # G+6551  U+6F3F  uni6F3F,  CJK UNIFIED IDEOGRAPH-6F3F
        0x7246, # G+6552  U+7246  uni7246,  CJK UNIFIED IDEOGRAPH-7246
        0x7350, # G+6553  U+7350  uni7350,  CJK UNIFIED IDEOGRAPH-7350
        0x748B, # G+6554  U+748B  uni748B,  CJK UNIFIED IDEOGRAPH-748B
        0x7AE0, # G+6555  U+7AE0  uni7AE0,  CJK UNIFIED IDEOGRAPH-7AE0
        0x7CA7, # G+6556  U+7CA7  uni7CA7,  CJK UNIFIED IDEOGRAPH-7CA7
        0x8178, # G+6557  U+8178  uni8178,  CJK UNIFIED IDEOGRAPH-8178
        0x81DF, # G+6558  U+81DF  uni81DF,  CJK UNIFIED IDEOGRAPH-81DF
        0x81E7, # G+6559  U+81E7  uni81E7,  CJK UNIFIED IDEOGRAPH-81E7
        0x838A, # G+6560  U+838A  uni838A,  CJK UNIFIED IDEOGRAPH-838A
        0x846C, # G+6561  U+846C  uni846C,  CJK UNIFIED IDEOGRAPH-846C
        0x8523, # G+6562  U+8523  uni8523,  CJK UNIFIED IDEOGRAPH-8523
        0x8594, # G+6563  U+8594  uni8594,  CJK UNIFIED IDEOGRAPH-8594
        0x85CF, # G+6564  U+85CF  uni85CF,  CJK UNIFIED IDEOGRAPH-85CF
        0x88DD, # G+6565  U+88DD  uni88DD,  CJK UNIFIED IDEOGRAPH-88DD
        0x8D13, # G+6566  U+8D13  uni8D13,  CJK UNIFIED IDEOGRAPH-8D13
        0x91AC, # G+6567  U+91AC  uni91AC,  CJK UNIFIED IDEOGRAPH-91AC
        0x9577, # G+6568  U+9577  uni9577,  CJK UNIFIED IDEOGRAPH-9577
        0x969C, # G+6569  U+969C  uni969C,  CJK UNIFIED IDEOGRAPH-969C
        0x518D, # G+6570  U+518D  uni518D,  CJK UNIFIED IDEOGRAPH-518D
        0x54C9, # G+6571  U+54C9  uni54C9,  CJK UNIFIED IDEOGRAPH-54C9
        0x5728, # G+6572  U+5728  uni5728,  CJK UNIFIED IDEOGRAPH-5728
        0x5BB0, # G+6573  U+5BB0  uni5BB0,  CJK UNIFIED IDEOGRAPH-5BB0
        0x624D, # G+6574  U+624D  uni624D,  CJK UNIFIED IDEOGRAPH-624D
        0x6750, # G+6575  U+6750  uni6750,  CJK UNIFIED IDEOGRAPH-6750
        0x683D, # G+6576  U+683D  uni683D,  CJK UNIFIED IDEOGRAPH-683D
        0x6893, # G+6577  U+6893  uni6893,  CJK UNIFIED IDEOGRAPH-6893
        0x6E3D, # G+6578  U+6E3D  uni6E3D,  CJK UNIFIED IDEOGRAPH-6E3D
        0x6ED3, # G+6579  U+6ED3  uni6ED3,  CJK UNIFIED IDEOGRAPH-6ED3
        0x707D, # G+6580  U+707D  uni707D,  CJK UNIFIED IDEOGRAPH-707D
        0x7E21, # G+6581  U+7E21  uni7E21,  CJK UNIFIED IDEOGRAPH-7E21
        0x88C1, # G+6582  U+88C1  uni88C1,  CJK UNIFIED IDEOGRAPH-88C1
        0x8CA1, # G+6583  U+8CA1  uni8CA1,  CJK UNIFIED IDEOGRAPH-8CA1
        0x8F09, # G+6584  U+8F09  uni8F09,  CJK UNIFIED IDEOGRAPH-8F09
        0x9F4B, # G+6585  U+9F4B  uni9F4B,  CJK UNIFIED IDEOGRAPH-9F4B
        0x9F4E, # G+6586  U+9F4E  uni9F4E,  CJK UNIFIED IDEOGRAPH-9F4E
        0x722D, # G+6587  U+722D  uni722D,  CJK UNIFIED IDEOGRAPH-722D
        0x7B8F, # G+6588  U+7B8F  uni7B8F,  CJK UNIFIED IDEOGRAPH-7B8F
        0x8ACD, # G+6589  U+8ACD  uni8ACD,  CJK UNIFIED IDEOGRAPH-8ACD
        0x931A, # G+6590  U+931A  uni931A,  CJK UNIFIED IDEOGRAPH-931A
        0x4F47, # G+6591  U+4F47  uni4F47,  CJK UNIFIED IDEOGRAPH-4F47
        0x4F4E, # G+6592  U+4F4E  uni4F4E,  CJK UNIFIED IDEOGRAPH-4F4E
        0x5132, # G+6593  U+5132  uni5132,  CJK UNIFIED IDEOGRAPH-5132
        0x5480, # G+6594  U+5480  uni5480,  CJK UNIFIED IDEOGRAPH-5480
        0x59D0, # G+6595  U+59D0  uni59D0,  CJK UNIFIED IDEOGRAPH-59D0
        0x5E95, # G+6596  U+5E95  uni5E95,  CJK UNIFIED IDEOGRAPH-5E95
        0x62B5, # G+6597  U+62B5  uni62B5,  CJK UNIFIED IDEOGRAPH-62B5
        0x6775, # G+6598  U+6775  uni6775,  CJK UNIFIED IDEOGRAPH-6775
        0x696E, # G+6599  U+696E  uni696E,  CJK UNIFIED IDEOGRAPH-696E
        0x6A17, # G+6600  U+6A17  uni6A17,  CJK UNIFIED IDEOGRAPH-6A17
        0x6CAE, # G+6601  U+6CAE  uni6CAE,  CJK UNIFIED IDEOGRAPH-6CAE
        0x6E1A, # G+6602  U+6E1A  uni6E1A,  CJK UNIFIED IDEOGRAPH-6E1A
        0x72D9, # G+6603  U+72D9  uni72D9,  CJK UNIFIED IDEOGRAPH-72D9
        0x732A, # G+6604  U+732A  uni732A,  CJK UNIFIED IDEOGRAPH-732A
        0x75BD, # G+6605  U+75BD  uni75BD,  CJK UNIFIED IDEOGRAPH-75BD
        0x7BB8, # G+6606  U+7BB8  uni7BB8,  CJK UNIFIED IDEOGRAPH-7BB8
        0x7D35, # G+6607  U+7D35  uni7D35,  CJK UNIFIED IDEOGRAPH-7D35
        0x82E7, # G+6608  U+82E7  uni82E7,  CJK UNIFIED IDEOGRAPH-82E7
        0x83F9, # G+6609  U+83F9  uni83F9,  CJK UNIFIED IDEOGRAPH-83F9
        0x8457, # G+6610  U+8457  uni8457,  CJK UNIFIED IDEOGRAPH-8457
        0x85F7, # G+6611  U+85F7  uni85F7,  CJK UNIFIED IDEOGRAPH-85F7
        0x8A5B, # G+6612  U+8A5B  uni8A5B,  CJK UNIFIED IDEOGRAPH-8A5B
        0x8CAF, # G+6613  U+8CAF  uni8CAF,  CJK UNIFIED IDEOGRAPH-8CAF
        0x8E87, # G+6614  U+8E87  uni8E87,  CJK UNIFIED IDEOGRAPH-8E87
        0x9019, # G+6615  U+9019  uni9019,  CJK UNIFIED IDEOGRAPH-9019
        0x90B8, # G+6616  U+90B8  uni90B8,  CJK UNIFIED IDEOGRAPH-90B8
        0x96CE, # G+6617  U+96CE  uni96CE,  CJK UNIFIED IDEOGRAPH-96CE
        0x9F5F, # G+6618  U+9F5F  uni9F5F,  CJK UNIFIED IDEOGRAPH-9F5F
        0x52E3, # G+6619  U+52E3  uni52E3,  CJK UNIFIED IDEOGRAPH-52E3
        0x540A, # G+6620  U+540A  uni540A,  CJK UNIFIED IDEOGRAPH-540A
        0x5AE1, # G+6621  U+5AE1  uni5AE1,  CJK UNIFIED IDEOGRAPH-5AE1
        0x5BC2, # G+6622  U+5BC2  uni5BC2,  CJK UNIFIED IDEOGRAPH-5BC2
        0x6458, # G+6623  U+6458  uni6458,  CJK UNIFIED IDEOGRAPH-6458
        0x6575, # G+6624  U+6575  uni6575,  CJK UNIFIED IDEOGRAPH-6575
        0x6EF4, # G+6625  U+6EF4  uni6EF4,  CJK UNIFIED IDEOGRAPH-6EF4
        0x72C4, # G+6626  U+72C4  uni72C4,  CJK UNIFIED IDEOGRAPH-72C4
        0x7684, # G+6627  U+7684  uni7684,  CJK UNIFIED IDEOGRAPH-7684
        0x7A4D, # G+6628  U+7A4D  uni7A4D,  CJK UNIFIED IDEOGRAPH-7A4D
        0x7B1B, # G+6629  U+7B1B  uni7B1B,  CJK UNIFIED IDEOGRAPH-7B1B
        0x7C4D, # G+6630  U+7C4D  uni7C4D,  CJK UNIFIED IDEOGRAPH-7C4D
        0x7E3E, # G+6631  U+7E3E  uni7E3E,  CJK UNIFIED IDEOGRAPH-7E3E
        0x7FDF, # G+6632  U+7FDF  uni7FDF,  CJK UNIFIED IDEOGRAPH-7FDF
        0x837B, # G+6633  U+837B  uni837B,  CJK UNIFIED IDEOGRAPH-837B
        0x8B2B, # G+6634  U+8B2B  uni8B2B,  CJK UNIFIED IDEOGRAPH-8B2B
        0x8CCA, # G+6635  U+8CCA  uni8CCA,  CJK UNIFIED IDEOGRAPH-8CCA
        0x8D64, # G+6636  U+8D64  uni8D64,  CJK UNIFIED IDEOGRAPH-8D64
        0x8DE1, # G+6637  U+8DE1  uni8DE1,  CJK UNIFIED IDEOGRAPH-8DE1
        0x8E5F, # G+6638  U+8E5F  uni8E5F,  CJK UNIFIED IDEOGRAPH-8E5F
        0x8FEA, # G+6639  U+8FEA  uni8FEA,  CJK UNIFIED IDEOGRAPH-8FEA
        0x8FF9, # G+6640  U+8FF9  uni8FF9,  CJK UNIFIED IDEOGRAPH-8FF9
        0x9069, # G+6641  U+9069  uni9069,  CJK UNIFIED IDEOGRAPH-9069
        0x93D1, # G+6642  U+93D1  uni93D1,  CJK UNIFIED IDEOGRAPH-93D1
        0x4F43, # G+6643  U+4F43  uni4F43,  CJK UNIFIED IDEOGRAPH-4F43
        0x4F7A, # G+6644  U+4F7A  uni4F7A,  CJK UNIFIED IDEOGRAPH-4F7A
        0x50B3, # G+6645  U+50B3  uni50B3,  CJK UNIFIED IDEOGRAPH-50B3
        0x5168, # G+6646  U+5168  uni5168,  CJK UNIFIED IDEOGRAPH-5168
        0x5178, # G+6647  U+5178  uni5178,  CJK UNIFIED IDEOGRAPH-5178
        0x524D, # G+6648  U+524D  uni524D,  CJK UNIFIED IDEOGRAPH-524D
        0x526A, # G+6649  U+526A  uni526A,  CJK UNIFIED IDEOGRAPH-526A
        0x5861, # G+6650  U+5861  uni5861,  CJK UNIFIED IDEOGRAPH-5861
        0x587C, # G+6651  U+587C  uni587C,  CJK UNIFIED IDEOGRAPH-587C
        0x5960, # G+6652  U+5960  uni5960,  CJK UNIFIED IDEOGRAPH-5960
        0x5C08, # G+6653  U+5C08  uni5C08,  CJK UNIFIED IDEOGRAPH-5C08
        0x5C55, # G+6654  U+5C55  uni5C55,  CJK UNIFIED IDEOGRAPH-5C55
        0x5EDB, # G+6655  U+5EDB  uni5EDB,  CJK UNIFIED IDEOGRAPH-5EDB
        0x609B, # G+6656  U+609B  uni609B,  CJK UNIFIED IDEOGRAPH-609B
        0x6230, # G+6657  U+6230  uni6230,  CJK UNIFIED IDEOGRAPH-6230
        0x6813, # G+6658  U+6813  uni6813,  CJK UNIFIED IDEOGRAPH-6813
        0x6BBF, # G+6659  U+6BBF  uni6BBF,  CJK UNIFIED IDEOGRAPH-6BBF
        0x6C08, # G+6660  U+6C08  uni6C08,  CJK UNIFIED IDEOGRAPH-6C08
        0x6FB1, # G+6661  U+6FB1  uni6FB1,  CJK UNIFIED IDEOGRAPH-6FB1
        0x714E, # G+6662  U+714E  uni714E,  CJK UNIFIED IDEOGRAPH-714E
        0x7420, # G+6663  U+7420  uni7420,  CJK UNIFIED IDEOGRAPH-7420
        0x7530, # G+6664  U+7530  uni7530,  CJK UNIFIED IDEOGRAPH-7530
        0x7538, # G+6665  U+7538  uni7538,  CJK UNIFIED IDEOGRAPH-7538
        0x7551, # G+6666  U+7551  uni7551,  CJK UNIFIED IDEOGRAPH-7551
        0x7672, # G+6667  U+7672  uni7672,  CJK UNIFIED IDEOGRAPH-7672
        0x7B4C, # G+6668  U+7B4C  uni7B4C,  CJK UNIFIED IDEOGRAPH-7B4C
        0x7B8B, # G+6669  U+7B8B  uni7B8B,  CJK UNIFIED IDEOGRAPH-7B8B
        0x7BAD, # G+6670  U+7BAD  uni7BAD,  CJK UNIFIED IDEOGRAPH-7BAD
        0x7BC6, # G+6671  U+7BC6  uni7BC6,  CJK UNIFIED IDEOGRAPH-7BC6
        0x7E8F, # G+6672  U+7E8F  uni7E8F,  CJK UNIFIED IDEOGRAPH-7E8F
        0x8A6E, # G+6673  U+8A6E  uni8A6E,  CJK UNIFIED IDEOGRAPH-8A6E
        0x8F3E, # G+6674  U+8F3E  uni8F3E,  CJK UNIFIED IDEOGRAPH-8F3E
        0x8F49, # G+6675  U+8F49  uni8F49,  CJK UNIFIED IDEOGRAPH-8F49
        0x923F, # G+6676  U+923F  uni923F,  CJK UNIFIED IDEOGRAPH-923F
        0x9293, # G+6677  U+9293  uni9293,  CJK UNIFIED IDEOGRAPH-9293
        0x9322, # G+6678  U+9322  uni9322,  CJK UNIFIED IDEOGRAPH-9322
        0x942B, # G+6679  U+942B  uni942B,  CJK UNIFIED IDEOGRAPH-942B
        0x96FB, # G+6680  U+96FB  uni96FB,  CJK UNIFIED IDEOGRAPH-96FB
        0x985A, # G+6681  U+985A  uni985A,  CJK UNIFIED IDEOGRAPH-985A
        0x986B, # G+6682  U+986B  uni986B,  CJK UNIFIED IDEOGRAPH-986B
        0x991E, # G+6683  U+991E  uni991E,  CJK UNIFIED IDEOGRAPH-991E
        0x5207, # G+6684  U+5207  uni5207,  CJK UNIFIED IDEOGRAPH-5207
        0x622A, # G+6685  U+622A  uni622A,  CJK UNIFIED IDEOGRAPH-622A
        0x6298, # G+6686  U+6298  uni6298,  CJK UNIFIED IDEOGRAPH-6298
        0x6D59, # G+6687  U+6D59  uni6D59,  CJK UNIFIED IDEOGRAPH-6D59
        0x7664, # G+6688  U+7664  uni7664,  CJK UNIFIED IDEOGRAPH-7664
        0x7ACA, # G+6689  U+7ACA  uni7ACA,  CJK UNIFIED IDEOGRAPH-7ACA
        0x7BC0, # G+6690  U+7BC0  uni7BC0,  CJK UNIFIED IDEOGRAPH-7BC0
        0x7D76, # G+6691  U+7D76  uni7D76,  CJK UNIFIED IDEOGRAPH-7D76
        0x5360, # G+6692  U+5360  uni5360,  CJK UNIFIED IDEOGRAPH-5360
        0x5CBE, # G+6693  U+5CBE  uni5CBE,  CJK UNIFIED IDEOGRAPH-5CBE
        0x5E97, # G+6694  U+5E97  uni5E97,  CJK UNIFIED IDEOGRAPH-5E97
        0x6F38, # G+6695  U+6F38  uni6F38,  CJK UNIFIED IDEOGRAPH-6F38
        0x70B9, # G+6696  U+70B9  uni70B9,  CJK UNIFIED IDEOGRAPH-70B9
        0x7C98, # G+6697  U+7C98  uni7C98,  CJK UNIFIED IDEOGRAPH-7C98
        0x9711, # G+6698  U+9711  uni9711,  CJK UNIFIED IDEOGRAPH-9711
        0x9B8E, # G+6699  U+9B8E  uni9B8E,  CJK UNIFIED IDEOGRAPH-9B8E
        0x9EDE, # G+6700  U+9EDE  uni9EDE,  CJK UNIFIED IDEOGRAPH-9EDE
        0x63A5, # G+6701  U+63A5  uni63A5,  CJK UNIFIED IDEOGRAPH-63A5
        0x647A, # G+6702  U+647A  uni647A,  CJK UNIFIED IDEOGRAPH-647A
        0x8776, # G+6703  U+8776  uni8776,  CJK UNIFIED IDEOGRAPH-8776
        0x4E01, # G+6704  U+4E01  uni4E01,  CJK UNIFIED IDEOGRAPH-4E01
        0x4E95, # G+6705  U+4E95  uni4E95,  CJK UNIFIED IDEOGRAPH-4E95
        0x4EAD, # G+6706  U+4EAD  uni4EAD,  CJK UNIFIED IDEOGRAPH-4EAD
        0x505C, # G+6707  U+505C  uni505C,  CJK UNIFIED IDEOGRAPH-505C
        0x5075, # G+6708  U+5075  uni5075,  CJK UNIFIED IDEOGRAPH-5075
        0x5448, # G+6709  U+5448  uni5448,  CJK UNIFIED IDEOGRAPH-5448
        0x59C3, # G+6710  U+59C3  uni59C3,  CJK UNIFIED IDEOGRAPH-59C3
        0x5B9A, # G+6711  U+5B9A  uni5B9A,  CJK UNIFIED IDEOGRAPH-5B9A
        0x5E40, # G+6712  U+5E40  uni5E40,  CJK UNIFIED IDEOGRAPH-5E40
        0x5EAD, # G+6713  U+5EAD  uni5EAD,  CJK UNIFIED IDEOGRAPH-5EAD
        0x5EF7, # G+6714  U+5EF7  uni5EF7,  CJK UNIFIED IDEOGRAPH-5EF7
        0x5F81, # G+6715  U+5F81  uni5F81,  CJK UNIFIED IDEOGRAPH-5F81
        0x60C5, # G+6716  U+60C5  uni60C5,  CJK UNIFIED IDEOGRAPH-60C5
        0x633A, # G+6717  U+633A  uni633A,  CJK UNIFIED IDEOGRAPH-633A
        0x653F, # G+6718  U+653F  uni653F,  CJK UNIFIED IDEOGRAPH-653F
        0x6574, # G+6719  U+6574  uni6574,  CJK UNIFIED IDEOGRAPH-6574
        0x65CC, # G+6720  U+65CC  uni65CC,  CJK UNIFIED IDEOGRAPH-65CC
        0x6676, # G+6721  U+6676  uni6676,  CJK UNIFIED IDEOGRAPH-6676
        0x6678, # G+6722  U+6678  uni6678,  CJK UNIFIED IDEOGRAPH-6678
        0x67FE, # G+6723  U+67FE  uni67FE,  CJK UNIFIED IDEOGRAPH-67FE
        0x6968, # G+6724  U+6968  uni6968,  CJK UNIFIED IDEOGRAPH-6968
        0x6A89, # G+6725  U+6A89  uni6A89,  CJK UNIFIED IDEOGRAPH-6A89
        0x6B63, # G+6726  U+6B63  uni6B63,  CJK UNIFIED IDEOGRAPH-6B63
        0x6C40, # G+6727  U+6C40  uni6C40,  CJK UNIFIED IDEOGRAPH-6C40
        0x6DC0, # G+6728  U+6DC0  uni6DC0,  CJK UNIFIED IDEOGRAPH-6DC0
        0x6DE8, # G+6729  U+6DE8  uni6DE8,  CJK UNIFIED IDEOGRAPH-6DE8
        0x6E1F, # G+6730  U+6E1F  uni6E1F,  CJK UNIFIED IDEOGRAPH-6E1F
        0x6E5E, # G+6731  U+6E5E  uni6E5E,  CJK UNIFIED IDEOGRAPH-6E5E
        0x701E, # G+6732  U+701E  uni701E,  CJK UNIFIED IDEOGRAPH-701E
        0x70A1, # G+6733  U+70A1  uni70A1,  CJK UNIFIED IDEOGRAPH-70A1
        0x738E, # G+6734  U+738E  uni738E,  CJK UNIFIED IDEOGRAPH-738E
        0x73FD, # G+6735  U+73FD  uni73FD,  CJK UNIFIED IDEOGRAPH-73FD
        0x753A, # G+6736  U+753A  uni753A,  CJK UNIFIED IDEOGRAPH-753A
        0x775B, # G+6737  U+775B  uni775B,  CJK UNIFIED IDEOGRAPH-775B
        0x7887, # G+6738  U+7887  uni7887,  CJK UNIFIED IDEOGRAPH-7887
        0x798E, # G+6739  U+798E  uni798E,  CJK UNIFIED IDEOGRAPH-798E
        0x7A0B, # G+6740  U+7A0B  uni7A0B,  CJK UNIFIED IDEOGRAPH-7A0B
        0x7A7D, # G+6741  U+7A7D  uni7A7D,  CJK UNIFIED IDEOGRAPH-7A7D
        0x7CBE, # G+6742  U+7CBE  uni7CBE,  CJK UNIFIED IDEOGRAPH-7CBE
        0x7D8E, # G+6743  U+7D8E  uni7D8E,  CJK UNIFIED IDEOGRAPH-7D8E
        0x8247, # G+6744  U+8247  uni8247,  CJK UNIFIED IDEOGRAPH-8247
        0x8A02, # G+6745  U+8A02  uni8A02,  CJK UNIFIED IDEOGRAPH-8A02
        0x8AEA, # G+6746  U+8AEA  uni8AEA,  CJK UNIFIED IDEOGRAPH-8AEA
        0x8C9E, # G+6747  U+8C9E  uni8C9E,  CJK UNIFIED IDEOGRAPH-8C9E
        0x912D, # G+6748  U+912D  uni912D,  CJK UNIFIED IDEOGRAPH-912D
        0x914A, # G+6749  U+914A  uni914A,  CJK UNIFIED IDEOGRAPH-914A
        0x91D8, # G+6750  U+91D8  uni91D8,  CJK UNIFIED IDEOGRAPH-91D8
        0x9266, # G+6751  U+9266  uni9266,  CJK UNIFIED IDEOGRAPH-9266
        0x92CC, # G+6752  U+92CC  uni92CC,  CJK UNIFIED IDEOGRAPH-92CC
        0x9320, # G+6753  U+9320  uni9320,  CJK UNIFIED IDEOGRAPH-9320
        0x9706, # G+6754  U+9706  uni9706,  CJK UNIFIED IDEOGRAPH-9706
        0x9756, # G+6755  U+9756  uni9756,  CJK UNIFIED IDEOGRAPH-9756
        0x975C, # G+6756  U+975C  uni975C,  CJK UNIFIED IDEOGRAPH-975C
        0x9802, # G+6757  U+9802  uni9802,  CJK UNIFIED IDEOGRAPH-9802
        0x9F0E, # G+6758  U+9F0E  uni9F0E,  CJK UNIFIED IDEOGRAPH-9F0E
        0x5236, # G+6759  U+5236  uni5236,  CJK UNIFIED IDEOGRAPH-5236
        0x5291, # G+6760  U+5291  uni5291,  CJK UNIFIED IDEOGRAPH-5291
        0x557C, # G+6761  U+557C  uni557C,  CJK UNIFIED IDEOGRAPH-557C
        0x5824, # G+6762  U+5824  uni5824,  CJK UNIFIED IDEOGRAPH-5824
        0x5E1D, # G+6763  U+5E1D  uni5E1D,  CJK UNIFIED IDEOGRAPH-5E1D
        0x5F1F, # G+6764  U+5F1F  uni5F1F,  CJK UNIFIED IDEOGRAPH-5F1F
        0x608C, # G+6765  U+608C  uni608C,  CJK UNIFIED IDEOGRAPH-608C
        0x63D0, # G+6766  U+63D0  uni63D0,  CJK UNIFIED IDEOGRAPH-63D0
        0x68AF, # G+6767  U+68AF  uni68AF,  CJK UNIFIED IDEOGRAPH-68AF
        0x6FDF, # G+6768  U+6FDF  uni6FDF,  CJK UNIFIED IDEOGRAPH-6FDF
        0x796D, # G+6769  U+796D  uni796D,  CJK UNIFIED IDEOGRAPH-796D
        0x7B2C, # G+6770  U+7B2C  uni7B2C,  CJK UNIFIED IDEOGRAPH-7B2C
        0x81CD, # G+6771  U+81CD  uni81CD,  CJK UNIFIED IDEOGRAPH-81CD
        0x85BA, # G+6772  U+85BA  uni85BA,  CJK UNIFIED IDEOGRAPH-85BA
        0x88FD, # G+6773  U+88FD  uni88FD,  CJK UNIFIED IDEOGRAPH-88FD
        0x8AF8, # G+6774  U+8AF8  uni8AF8,  CJK UNIFIED IDEOGRAPH-8AF8
        0x8E44, # G+6775  U+8E44  uni8E44,  CJK UNIFIED IDEOGRAPH-8E44
        0x918D, # G+6776  U+918D  uni918D,  CJK UNIFIED IDEOGRAPH-918D
        0x9664, # G+6777  U+9664  uni9664,  CJK UNIFIED IDEOGRAPH-9664
        0x969B, # G+6778  U+969B  uni969B,  CJK UNIFIED IDEOGRAPH-969B
        0x973D, # G+6779  U+973D  uni973D,  CJK UNIFIED IDEOGRAPH-973D
        0x984C, # G+6780  U+984C  uni984C,  CJK UNIFIED IDEOGRAPH-984C
        0x9F4A, # G+6781  U+9F4A  uni9F4A,  CJK UNIFIED IDEOGRAPH-9F4A
        0x4FCE, # G+6782  U+4FCE  uni4FCE,  CJK UNIFIED IDEOGRAPH-4FCE
        0x5146, # G+6783  U+5146  uni5146,  CJK UNIFIED IDEOGRAPH-5146
        0x51CB, # G+6784  U+51CB  uni51CB,  CJK UNIFIED IDEOGRAPH-51CB
        0x52A9, # G+6785  U+52A9  uni52A9,  CJK UNIFIED IDEOGRAPH-52A9
        0x5632, # G+6786  U+5632  uni5632,  CJK UNIFIED IDEOGRAPH-5632
        0x5F14, # G+6787  U+5F14  uni5F14,  CJK UNIFIED IDEOGRAPH-5F14
        0x5F6B, # G+6788  U+5F6B  uni5F6B,  CJK UNIFIED IDEOGRAPH-5F6B
        0x63AA, # G+6789  U+63AA  uni63AA,  CJK UNIFIED IDEOGRAPH-63AA
        0x64CD, # G+6790  U+64CD  uni64CD,  CJK UNIFIED IDEOGRAPH-64CD
        0x65E9, # G+6791  U+65E9  uni65E9,  CJK UNIFIED IDEOGRAPH-65E9
        0x6641, # G+6792  U+6641  uni6641,  CJK UNIFIED IDEOGRAPH-6641
        0x66FA, # G+6793  U+66FA  uni66FA,  CJK UNIFIED IDEOGRAPH-66FA
        0x66F9, # G+6794  U+66F9  uni66F9,  CJK UNIFIED IDEOGRAPH-66F9
        0x671D, # G+6795  U+671D  uni671D,  CJK UNIFIED IDEOGRAPH-671D
        0x689D, # G+6796  U+689D  uni689D,  CJK UNIFIED IDEOGRAPH-689D
        0x68D7, # G+6797  U+68D7  uni68D7,  CJK UNIFIED IDEOGRAPH-68D7
        0x69FD, # G+6798  U+69FD  uni69FD,  CJK UNIFIED IDEOGRAPH-69FD
        0x6F15, # G+6799  U+6F15  uni6F15,  CJK UNIFIED IDEOGRAPH-6F15
        0x6F6E, # G+6800  U+6F6E  uni6F6E,  CJK UNIFIED IDEOGRAPH-6F6E
        0x7167, # G+6801  U+7167  uni7167,  CJK UNIFIED IDEOGRAPH-7167
        0x71E5, # G+6802  U+71E5  uni71E5,  CJK UNIFIED IDEOGRAPH-71E5
        0x722A, # G+6803  U+722A  uni722A,  CJK UNIFIED IDEOGRAPH-722A
        0x74AA, # G+6804  U+74AA  uni74AA,  CJK UNIFIED IDEOGRAPH-74AA
        0x773A, # G+6805  U+773A  uni773A,  CJK UNIFIED IDEOGRAPH-773A
        0x7956, # G+6806  U+7956  uni7956,  CJK UNIFIED IDEOGRAPH-7956
        0x795A, # G+6807  U+795A  uni795A,  CJK UNIFIED IDEOGRAPH-795A
        0x79DF, # G+6808  U+79DF  uni79DF,  CJK UNIFIED IDEOGRAPH-79DF
        0x7A20, # G+6809  U+7A20  uni7A20,  CJK UNIFIED IDEOGRAPH-7A20
        0x7A95, # G+6810  U+7A95  uni7A95,  CJK UNIFIED IDEOGRAPH-7A95
        0x7C97, # G+6811  U+7C97  uni7C97,  CJK UNIFIED IDEOGRAPH-7C97
        0x7CDF, # G+6812  U+7CDF  uni7CDF,  CJK UNIFIED IDEOGRAPH-7CDF
        0x7D44, # G+6813  U+7D44  uni7D44,  CJK UNIFIED IDEOGRAPH-7D44
        0x7E70, # G+6814  U+7E70  uni7E70,  CJK UNIFIED IDEOGRAPH-7E70
        0x8087, # G+6815  U+8087  uni8087,  CJK UNIFIED IDEOGRAPH-8087
        0x85FB, # G+6816  U+85FB  uni85FB,  CJK UNIFIED IDEOGRAPH-85FB
        0x86A4, # G+6817  U+86A4  uni86A4,  CJK UNIFIED IDEOGRAPH-86A4
        0x8A54, # G+6818  U+8A54  uni8A54,  CJK UNIFIED IDEOGRAPH-8A54
        0x8ABF, # G+6819  U+8ABF  uni8ABF,  CJK UNIFIED IDEOGRAPH-8ABF
        0x8D99, # G+6820  U+8D99  uni8D99,  CJK UNIFIED IDEOGRAPH-8D99
        0x8E81, # G+6821  U+8E81  uni8E81,  CJK UNIFIED IDEOGRAPH-8E81
        0x9020, # G+6822  U+9020  uni9020,  CJK UNIFIED IDEOGRAPH-9020
        0x906D, # G+6823  U+906D  uni906D,  CJK UNIFIED IDEOGRAPH-906D
        0x91E3, # G+6824  U+91E3  uni91E3,  CJK UNIFIED IDEOGRAPH-91E3
        0x963B, # G+6825  U+963B  uni963B,  CJK UNIFIED IDEOGRAPH-963B
        0x96D5, # G+6826  U+96D5  uni96D5,  CJK UNIFIED IDEOGRAPH-96D5
        0x9CE5, # G+6827  U+9CE5  uni9CE5,  CJK UNIFIED IDEOGRAPH-9CE5
        0x65CF, # G+6828  U+65CF  uni65CF,  CJK UNIFIED IDEOGRAPH-65CF
        0x7C07, # G+6829  U+7C07  uni7C07,  CJK UNIFIED IDEOGRAPH-7C07
        0x8DB3, # G+6830  U+8DB3  uni8DB3,  CJK UNIFIED IDEOGRAPH-8DB3
        0x93C3, # G+6831  U+93C3  uni93C3,  CJK UNIFIED IDEOGRAPH-93C3
        0x5B58, # G+6832  U+5B58  uni5B58,  CJK UNIFIED IDEOGRAPH-5B58
        0x5C0A, # G+6833  U+5C0A  uni5C0A,  CJK UNIFIED IDEOGRAPH-5C0A
        0x5352, # G+6834  U+5352  uni5352,  CJK UNIFIED IDEOGRAPH-5352
        0x62D9, # G+6835  U+62D9  uni62D9,  CJK UNIFIED IDEOGRAPH-62D9
        0x731D, # G+6836  U+731D  uni731D,  CJK UNIFIED IDEOGRAPH-731D
        0x5027, # G+6837  U+5027  uni5027,  CJK UNIFIED IDEOGRAPH-5027
        0x5B97, # G+6838  U+5B97  uni5B97,  CJK UNIFIED IDEOGRAPH-5B97
        0x5F9E, # G+6839  U+5F9E  uni5F9E,  CJK UNIFIED IDEOGRAPH-5F9E
        0x60B0, # G+6840  U+60B0  uni60B0,  CJK UNIFIED IDEOGRAPH-60B0
        0x616B, # G+6841  U+616B  uni616B,  CJK UNIFIED IDEOGRAPH-616B
        0x68D5, # G+6842  U+68D5  uni68D5,  CJK UNIFIED IDEOGRAPH-68D5
        0x6DD9, # G+6843  U+6DD9  uni6DD9,  CJK UNIFIED IDEOGRAPH-6DD9
        0x742E, # G+6844  U+742E  uni742E,  CJK UNIFIED IDEOGRAPH-742E
        0x7A2E, # G+6845  U+7A2E  uni7A2E,  CJK UNIFIED IDEOGRAPH-7A2E
        0x7D42, # G+6846  U+7D42  uni7D42,  CJK UNIFIED IDEOGRAPH-7D42
        0x7D9C, # G+6847  U+7D9C  uni7D9C,  CJK UNIFIED IDEOGRAPH-7D9C
        0x7E31, # G+6848  U+7E31  uni7E31,  CJK UNIFIED IDEOGRAPH-7E31
        0x816B, # G+6849  U+816B  uni816B,  CJK UNIFIED IDEOGRAPH-816B
        0x8E2A, # G+6850  U+8E2A  uni8E2A,  CJK UNIFIED IDEOGRAPH-8E2A
        0x8E35, # G+6851  U+8E35  uni8E35,  CJK UNIFIED IDEOGRAPH-8E35
        0x937E, # G+6852  U+937E  uni937E,  CJK UNIFIED IDEOGRAPH-937E
        0x9418, # G+6853  U+9418  uni9418,  CJK UNIFIED IDEOGRAPH-9418
        0x4F50, # G+6854  U+4F50  uni4F50,  CJK UNIFIED IDEOGRAPH-4F50
        0x5750, # G+6855  U+5750  uni5750,  CJK UNIFIED IDEOGRAPH-5750
        0x5DE6, # G+6856  U+5DE6  uni5DE6,  CJK UNIFIED IDEOGRAPH-5DE6
        0x5EA7, # G+6857  U+5EA7  uni5EA7,  CJK UNIFIED IDEOGRAPH-5EA7
        0x632B, # G+6858  U+632B  uni632B,  CJK UNIFIED IDEOGRAPH-632B
        0x7F6A, # G+6859  U+7F6A  uni7F6A,  CJK UNIFIED IDEOGRAPH-7F6A
        0x4E3B, # G+6860  U+4E3B  uni4E3B,  CJK UNIFIED IDEOGRAPH-4E3B
        0x4F4F, # G+6861  U+4F4F  uni4F4F,  CJK UNIFIED IDEOGRAPH-4F4F
        0x4F8F, # G+6862  U+4F8F  uni4F8F,  CJK UNIFIED IDEOGRAPH-4F8F
        0x505A, # G+6863  U+505A  uni505A,  CJK UNIFIED IDEOGRAPH-505A
        0x59DD, # G+6864  U+59DD  uni59DD,  CJK UNIFIED IDEOGRAPH-59DD
        0x80C4, # G+6865  U+80C4  uni80C4,  CJK UNIFIED IDEOGRAPH-80C4
        0x546A, # G+6866  U+546A  uni546A,  CJK UNIFIED IDEOGRAPH-546A
        0x5468, # G+6867  U+5468  uni5468,  CJK UNIFIED IDEOGRAPH-5468
        0x55FE, # G+6868  U+55FE  uni55FE,  CJK UNIFIED IDEOGRAPH-55FE
        0x594F, # G+6869  U+594F  uni594F,  CJK UNIFIED IDEOGRAPH-594F
        0x5B99, # G+6870  U+5B99  uni5B99,  CJK UNIFIED IDEOGRAPH-5B99
        0x5DDE, # G+6871  U+5DDE  uni5DDE,  CJK UNIFIED IDEOGRAPH-5DDE
        0x5EDA, # G+6872  U+5EDA  uni5EDA,  CJK UNIFIED IDEOGRAPH-5EDA
        0x665D, # G+6873  U+665D  uni665D,  CJK UNIFIED IDEOGRAPH-665D
        0x6731, # G+6874  U+6731  uni6731,  CJK UNIFIED IDEOGRAPH-6731
        0x67F1, # G+6875  U+67F1  uni67F1,  CJK UNIFIED IDEOGRAPH-67F1
        0x682A, # G+6876  U+682A  uni682A,  CJK UNIFIED IDEOGRAPH-682A
        0x6CE8, # G+6877  U+6CE8  uni6CE8,  CJK UNIFIED IDEOGRAPH-6CE8
        0x6D32, # G+6878  U+6D32  uni6D32,  CJK UNIFIED IDEOGRAPH-6D32
        0x6E4A, # G+6879  U+6E4A  uni6E4A,  CJK UNIFIED IDEOGRAPH-6E4A
        0x6F8D, # G+6880  U+6F8D  uni6F8D,  CJK UNIFIED IDEOGRAPH-6F8D
        0x70B7, # G+6881  U+70B7  uni70B7,  CJK UNIFIED IDEOGRAPH-70B7
        0x73E0, # G+6882  U+73E0  uni73E0,  CJK UNIFIED IDEOGRAPH-73E0
        0x7587, # G+6883  U+7587  uni7587,  CJK UNIFIED IDEOGRAPH-7587
        0x7C4C, # G+6884  U+7C4C  uni7C4C,  CJK UNIFIED IDEOGRAPH-7C4C
        0x7D02, # G+6885  U+7D02  uni7D02,  CJK UNIFIED IDEOGRAPH-7D02
        0x7D2C, # G+6886  U+7D2C  uni7D2C,  CJK UNIFIED IDEOGRAPH-7D2C
        0x7DA2, # G+6887  U+7DA2  uni7DA2,  CJK UNIFIED IDEOGRAPH-7DA2
        0x821F, # G+6888  U+821F  uni821F,  CJK UNIFIED IDEOGRAPH-821F
        0x86DB, # G+6889  U+86DB  uni86DB,  CJK UNIFIED IDEOGRAPH-86DB
        0x8A3B, # G+6890  U+8A3B  uni8A3B,  CJK UNIFIED IDEOGRAPH-8A3B
        0x8A85, # G+6891  U+8A85  uni8A85,  CJK UNIFIED IDEOGRAPH-8A85
        0x8D70, # G+6892  U+8D70  uni8D70,  CJK UNIFIED IDEOGRAPH-8D70
        0x8E8A, # G+6893  U+8E8A  uni8E8A,  CJK UNIFIED IDEOGRAPH-8E8A
        0x8F33, # G+6894  U+8F33  uni8F33,  CJK UNIFIED IDEOGRAPH-8F33
        0x9031, # G+6895  U+9031  uni9031,  CJK UNIFIED IDEOGRAPH-9031
        0x914E, # G+6896  U+914E  uni914E,  CJK UNIFIED IDEOGRAPH-914E
        0x9152, # G+6897  U+9152  uni9152,  CJK UNIFIED IDEOGRAPH-9152
        0x9444, # G+6898  U+9444  uni9444,  CJK UNIFIED IDEOGRAPH-9444
        0x99D0, # G+6899  U+99D0  uni99D0,  CJK UNIFIED IDEOGRAPH-99D0
        0x7AF9, # G+6900  U+7AF9  uni7AF9,  CJK UNIFIED IDEOGRAPH-7AF9
        0x7CA5, # G+6901  U+7CA5  uni7CA5,  CJK UNIFIED IDEOGRAPH-7CA5
        0x4FCA, # G+6902  U+4FCA  uni4FCA,  CJK UNIFIED IDEOGRAPH-4FCA
        0x5101, # G+6903  U+5101  uni5101,  CJK UNIFIED IDEOGRAPH-5101
        0x51C6, # G+6904  U+51C6  uni51C6,  CJK UNIFIED IDEOGRAPH-51C6
        0x57C8, # G+6905  U+57C8  uni57C8,  CJK UNIFIED IDEOGRAPH-57C8
        0x5BEF, # G+6906  U+5BEF  uni5BEF,  CJK UNIFIED IDEOGRAPH-5BEF
        0x5CFB, # G+6907  U+5CFB  uni5CFB,  CJK UNIFIED IDEOGRAPH-5CFB
        0x6659, # G+6908  U+6659  uni6659,  CJK UNIFIED IDEOGRAPH-6659
        0x6A3D, # G+6909  U+6A3D  uni6A3D,  CJK UNIFIED IDEOGRAPH-6A3D
        0x6D5A, # G+6910  U+6D5A  uni6D5A,  CJK UNIFIED IDEOGRAPH-6D5A
        0x6E96, # G+6911  U+6E96  uni6E96,  CJK UNIFIED IDEOGRAPH-6E96
        0x6FEC, # G+6912  U+6FEC  uni6FEC,  CJK UNIFIED IDEOGRAPH-6FEC
        0x710C, # G+6913  U+710C  uni710C,  CJK UNIFIED IDEOGRAPH-710C
        0x756F, # G+6914  U+756F  uni756F,  CJK UNIFIED IDEOGRAPH-756F
        0x7AE3, # G+6915  U+7AE3  uni7AE3,  CJK UNIFIED IDEOGRAPH-7AE3
        0x8822, # G+6916  U+8822  uni8822,  CJK UNIFIED IDEOGRAPH-8822
        0x9021, # G+6917  U+9021  uni9021,  CJK UNIFIED IDEOGRAPH-9021
        0x9075, # G+6918  U+9075  uni9075,  CJK UNIFIED IDEOGRAPH-9075
        0x96CB, # G+6919  U+96CB  uni96CB,  CJK UNIFIED IDEOGRAPH-96CB
        0x99FF, # G+6920  U+99FF  uni99FF,  CJK UNIFIED IDEOGRAPH-99FF
        0x8301, # G+6921  U+8301  uni8301,  CJK UNIFIED IDEOGRAPH-8301
        0x4E2D, # G+6922  U+4E2D  uni4E2D,  CJK UNIFIED IDEOGRAPH-4E2D
        0x4EF2, # G+6923  U+4EF2  uni4EF2,  CJK UNIFIED IDEOGRAPH-4EF2
        0x8846, # G+6924  U+8846  uni8846,  CJK UNIFIED IDEOGRAPH-8846
        0x91CD, # G+6925  U+91CD  uni91CD,  CJK UNIFIED IDEOGRAPH-91CD
        0x537D, # G+6926  U+537D  uni537D,  CJK UNIFIED IDEOGRAPH-537D
        0x6ADB, # G+6927  U+6ADB  uni6ADB,  CJK UNIFIED IDEOGRAPH-6ADB
        0x696B, # G+6928  U+696B  uni696B,  CJK UNIFIED IDEOGRAPH-696B
        0x6C41, # G+6929  U+6C41  uni6C41,  CJK UNIFIED IDEOGRAPH-6C41
        0x847A, # G+6930  U+847A  uni847A,  CJK UNIFIED IDEOGRAPH-847A
        0x589E, # G+6931  U+589E  uni589E,  CJK UNIFIED IDEOGRAPH-589E
        0x618E, # G+6932  U+618E  uni618E,  CJK UNIFIED IDEOGRAPH-618E
        0x66FE, # G+6933  U+66FE  uni66FE,  CJK UNIFIED IDEOGRAPH-66FE
        0x62EF, # G+6934  U+62EF  uni62EF,  CJK UNIFIED IDEOGRAPH-62EF
        0x70DD, # G+6935  U+70DD  uni70DD,  CJK UNIFIED IDEOGRAPH-70DD
        0x7511, # G+6936  U+7511  uni7511,  CJK UNIFIED IDEOGRAPH-7511
        0x75C7, # G+6937  U+75C7  uni75C7,  CJK UNIFIED IDEOGRAPH-75C7
        0x7E52, # G+6938  U+7E52  uni7E52,  CJK UNIFIED IDEOGRAPH-7E52
        0x84B8, # G+6939  U+84B8  uni84B8,  CJK UNIFIED IDEOGRAPH-84B8
        0x8B49, # G+6940  U+8B49  uni8B49,  CJK UNIFIED IDEOGRAPH-8B49
        0x8D08, # G+6941  U+8D08  uni8D08,  CJK UNIFIED IDEOGRAPH-8D08
        0x4E4B, # G+6942  U+4E4B  uni4E4B,  CJK UNIFIED IDEOGRAPH-4E4B
        0x53EA, # G+6943  U+53EA  uni53EA,  CJK UNIFIED IDEOGRAPH-53EA
        0x54AB, # G+6944  U+54AB  uni54AB,  CJK UNIFIED IDEOGRAPH-54AB
        0x5730, # G+6945  U+5730  uni5730,  CJK UNIFIED IDEOGRAPH-5730
        0x5740, # G+6946  U+5740  uni5740,  CJK UNIFIED IDEOGRAPH-5740
        0x5FD7, # G+6947  U+5FD7  uni5FD7,  CJK UNIFIED IDEOGRAPH-5FD7
        0x6301, # G+6948  U+6301  uni6301,  CJK UNIFIED IDEOGRAPH-6301
        0x6307, # G+6949  U+6307  uni6307,  CJK UNIFIED IDEOGRAPH-6307
        0x646F, # G+6950  U+646F  uni646F,  CJK UNIFIED IDEOGRAPH-646F
        0x652F, # G+6951  U+652F  uni652F,  CJK UNIFIED IDEOGRAPH-652F
        0x65E8, # G+6952  U+65E8  uni65E8,  CJK UNIFIED IDEOGRAPH-65E8
        0x667A, # G+6953  U+667A  uni667A,  CJK UNIFIED IDEOGRAPH-667A
        0x679D, # G+6954  U+679D  uni679D,  CJK UNIFIED IDEOGRAPH-679D
        0x67B3, # G+6955  U+67B3  uni67B3,  CJK UNIFIED IDEOGRAPH-67B3
        0x6B62, # G+6956  U+6B62  uni6B62,  CJK UNIFIED IDEOGRAPH-6B62
        0x6C60, # G+6957  U+6C60  uni6C60,  CJK UNIFIED IDEOGRAPH-6C60
        0x6C9A, # G+6958  U+6C9A  uni6C9A,  CJK UNIFIED IDEOGRAPH-6C9A
        0x6F2C, # G+6959  U+6F2C  uni6F2C,  CJK UNIFIED IDEOGRAPH-6F2C
        0x77E5, # G+6960  U+77E5  uni77E5,  CJK UNIFIED IDEOGRAPH-77E5
        0x7825, # G+6961  U+7825  uni7825,  CJK UNIFIED IDEOGRAPH-7825
        0x7949, # G+6962  U+7949  uni7949,  CJK UNIFIED IDEOGRAPH-7949
        0x7957, # G+6963  U+7957  uni7957,  CJK UNIFIED IDEOGRAPH-7957
        0x7D19, # G+6964  U+7D19  uni7D19,  CJK UNIFIED IDEOGRAPH-7D19
        0x80A2, # G+6965  U+80A2  uni80A2,  CJK UNIFIED IDEOGRAPH-80A2
        0x8102, # G+6966  U+8102  uni8102,  CJK UNIFIED IDEOGRAPH-8102
        0x81F3, # G+6967  U+81F3  uni81F3,  CJK UNIFIED IDEOGRAPH-81F3
        0x829D, # G+6968  U+829D  uni829D,  CJK UNIFIED IDEOGRAPH-829D
        0x82B7, # G+6969  U+82B7  uni82B7,  CJK UNIFIED IDEOGRAPH-82B7
        0x8718, # G+6970  U+8718  uni8718,  CJK UNIFIED IDEOGRAPH-8718
        0x8A8C, # G+6971  U+8A8C  uni8A8C,  CJK UNIFIED IDEOGRAPH-8A8C
        0x8D04, # G+6972  U+8D04  uni8D04,  CJK UNIFIED IDEOGRAPH-8D04
        0x8DBE, # G+6973  U+8DBE  uni8DBE,  CJK UNIFIED IDEOGRAPH-8DBE
        0x9072, # G+6974  U+9072  uni9072,  CJK UNIFIED IDEOGRAPH-9072
        0x76F4, # G+6975  U+76F4  uni76F4,  CJK UNIFIED IDEOGRAPH-76F4
        0x7A19, # G+6976  U+7A19  uni7A19,  CJK UNIFIED IDEOGRAPH-7A19
        0x7A37, # G+6977  U+7A37  uni7A37,  CJK UNIFIED IDEOGRAPH-7A37
        0x7E54, # G+6978  U+7E54  uni7E54,  CJK UNIFIED IDEOGRAPH-7E54
        0x8077, # G+6979  U+8077  uni8077,  CJK UNIFIED IDEOGRAPH-8077
        0x5507, # G+6980  U+5507  uni5507,  CJK UNIFIED IDEOGRAPH-5507
        0x55D4, # G+6981  U+55D4  uni55D4,  CJK UNIFIED IDEOGRAPH-55D4
        0x5875, # G+6982  U+5875  uni5875,  CJK UNIFIED IDEOGRAPH-5875
        0x632F, # G+6983  U+632F  uni632F,  CJK UNIFIED IDEOGRAPH-632F
        0x6422, # G+6984  U+6422  uni6422,  CJK UNIFIED IDEOGRAPH-6422
        0x6649, # G+6985  U+6649  uni6649,  CJK UNIFIED IDEOGRAPH-6649
        0x664B, # G+6986  U+664B  uni664B,  CJK UNIFIED IDEOGRAPH-664B
        0x686D, # G+6987  U+686D  uni686D,  CJK UNIFIED IDEOGRAPH-686D
        0x699B, # G+6988  U+699B  uni699B,  CJK UNIFIED IDEOGRAPH-699B
        0x6B84, # G+6989  U+6B84  uni6B84,  CJK UNIFIED IDEOGRAPH-6B84
        0x6D25, # G+6990  U+6D25  uni6D25,  CJK UNIFIED IDEOGRAPH-6D25
        0x6EB1, # G+6991  U+6EB1  uni6EB1,  CJK UNIFIED IDEOGRAPH-6EB1
        0x73CD, # G+6992  U+73CD  uni73CD,  CJK UNIFIED IDEOGRAPH-73CD
        0x7468, # G+6993  U+7468  uni7468,  CJK UNIFIED IDEOGRAPH-7468
        0x74A1, # G+6994  U+74A1  uni74A1,  CJK UNIFIED IDEOGRAPH-74A1
        0x755B, # G+6995  U+755B  uni755B,  CJK UNIFIED IDEOGRAPH-755B
        0x75B9, # G+6996  U+75B9  uni75B9,  CJK UNIFIED IDEOGRAPH-75B9
        0x76E1, # G+6997  U+76E1  uni76E1,  CJK UNIFIED IDEOGRAPH-76E1
        0x771E, # G+6998  U+771E  uni771E,  CJK UNIFIED IDEOGRAPH-771E
        0x778B, # G+6999  U+778B  uni778B,  CJK UNIFIED IDEOGRAPH-778B
        0x79E6, # G+7000  U+79E6  uni79E6,  CJK UNIFIED IDEOGRAPH-79E6
        0x7E09, # G+7001  U+7E09  uni7E09,  CJK UNIFIED IDEOGRAPH-7E09
        0x7E1D, # G+7002  U+7E1D  uni7E1D,  CJK UNIFIED IDEOGRAPH-7E1D
        0x81FB, # G+7003  U+81FB  uni81FB,  CJK UNIFIED IDEOGRAPH-81FB
        0x852F, # G+7004  U+852F  uni852F,  CJK UNIFIED IDEOGRAPH-852F
        0x8897, # G+7005  U+8897  uni8897,  CJK UNIFIED IDEOGRAPH-8897
        0x8A3A, # G+7006  U+8A3A  uni8A3A,  CJK UNIFIED IDEOGRAPH-8A3A
        0x8CD1, # G+7007  U+8CD1  uni8CD1,  CJK UNIFIED IDEOGRAPH-8CD1
        0x8EEB, # G+7008  U+8EEB  uni8EEB,  CJK UNIFIED IDEOGRAPH-8EEB
        0xF971, # G+7009  U+F971  uniF971,  CJK COMPATIBILITY IDEOGRAPH-F971
        0x9032, # G+7010  U+9032  uni9032,  CJK UNIFIED IDEOGRAPH-9032
        0x93AD, # G+7011  U+93AD  uni93AD,  CJK UNIFIED IDEOGRAPH-93AD
        0x9663, # G+7012  U+9663  uni9663,  CJK UNIFIED IDEOGRAPH-9663
        0x9673, # G+7013  U+9673  uni9673,  CJK UNIFIED IDEOGRAPH-9673
        0x9707, # G+7014  U+9707  uni9707,  CJK UNIFIED IDEOGRAPH-9707
        0x4F84, # G+7015  U+4F84  uni4F84,  CJK UNIFIED IDEOGRAPH-4F84
        0x53F1, # G+7016  U+53F1  uni53F1,  CJK UNIFIED IDEOGRAPH-53F1
        0x59EA, # G+7017  U+59EA  uni59EA,  CJK UNIFIED IDEOGRAPH-59EA
        0x5AC9, # G+7018  U+5AC9  uni5AC9,  CJK UNIFIED IDEOGRAPH-5AC9
        0x5E19, # G+7019  U+5E19  uni5E19,  CJK UNIFIED IDEOGRAPH-5E19
        0x684E, # G+7020  U+684E  uni684E,  CJK UNIFIED IDEOGRAPH-684E
        0x74C6, # G+7021  U+74C6  uni74C6,  CJK UNIFIED IDEOGRAPH-74C6
        0x75BE, # G+7022  U+75BE  uni75BE,  CJK UNIFIED IDEOGRAPH-75BE
        0x79E9, # G+7023  U+79E9  uni79E9,  CJK UNIFIED IDEOGRAPH-79E9
        0x7A92, # G+7024  U+7A92  uni7A92,  CJK UNIFIED IDEOGRAPH-7A92
        0x81A3, # G+7025  U+81A3  uni81A3,  CJK UNIFIED IDEOGRAPH-81A3
        0x86ED, # G+7026  U+86ED  uni86ED,  CJK UNIFIED IDEOGRAPH-86ED
        0x8CEA, # G+7027  U+8CEA  uni8CEA,  CJK UNIFIED IDEOGRAPH-8CEA
        0x8DCC, # G+7028  U+8DCC  uni8DCC,  CJK UNIFIED IDEOGRAPH-8DCC
        0x8FED, # G+7029  U+8FED  uni8FED,  CJK UNIFIED IDEOGRAPH-8FED
        0x659F, # G+7030  U+659F  uni659F,  CJK UNIFIED IDEOGRAPH-659F
        0x6715, # G+7031  U+6715  uni6715,  CJK UNIFIED IDEOGRAPH-6715
        0x57F7, # G+7032  U+57F7  uni57F7,  CJK UNIFIED IDEOGRAPH-57F7
        0x6F57, # G+7033  U+6F57  uni6F57,  CJK UNIFIED IDEOGRAPH-6F57
        0x7DDD, # G+7034  U+7DDD  uni7DDD,  CJK UNIFIED IDEOGRAPH-7DDD
        0x8F2F, # G+7035  U+8F2F  uni8F2F,  CJK UNIFIED IDEOGRAPH-8F2F
        0x93F6, # G+7036  U+93F6  uni93F6,  CJK UNIFIED IDEOGRAPH-93F6
        0x96C6, # G+7037  U+96C6  uni96C6,  CJK UNIFIED IDEOGRAPH-96C6
        0x5FB5, # G+7038  U+5FB5  uni5FB5,  CJK UNIFIED IDEOGRAPH-5FB5
        0x61F2, # G+7039  U+61F2  uni61F2,  CJK UNIFIED IDEOGRAPH-61F2
        0x6F84, # G+7040  U+6F84  uni6F84,  CJK UNIFIED IDEOGRAPH-6F84
        0x4E14, # G+7041  U+4E14  uni4E14,  CJK UNIFIED IDEOGRAPH-4E14
        0x4F98, # G+7042  U+4F98  uni4F98,  CJK UNIFIED IDEOGRAPH-4F98
        0x501F, # G+7043  U+501F  uni501F,  CJK UNIFIED IDEOGRAPH-501F
        0x53C9, # G+7044  U+53C9  uni53C9,  CJK UNIFIED IDEOGRAPH-53C9
        0x55DF, # G+7045  U+55DF  uni55DF,  CJK UNIFIED IDEOGRAPH-55DF
        0x5D6F, # G+7046  U+5D6F  uni5D6F,  CJK UNIFIED IDEOGRAPH-5D6F
        0x5DEE, # G+7047  U+5DEE  uni5DEE,  CJK UNIFIED IDEOGRAPH-5DEE
        0x6B21, # G+7048  U+6B21  uni6B21,  CJK UNIFIED IDEOGRAPH-6B21
        0x6B64, # G+7049  U+6B64  uni6B64,  CJK UNIFIED IDEOGRAPH-6B64
        0x78CB, # G+7050  U+78CB  uni78CB,  CJK UNIFIED IDEOGRAPH-78CB
        0x7B9A, # G+7051  U+7B9A  uni7B9A,  CJK UNIFIED IDEOGRAPH-7B9A
        0x8E49, # G+7052  U+8E49  uni8E49,  CJK UNIFIED IDEOGRAPH-8E49
        0xF902, # G+7053  U+F902  uniF902,  CJK COMPATIBILITY IDEOGRAPH-F902
        0x906E, # G+7054  U+906E  uni906E,  CJK UNIFIED IDEOGRAPH-906E
        0x6349, # G+7055  U+6349  uni6349,  CJK UNIFIED IDEOGRAPH-6349
        0x643E, # G+7056  U+643E  uni643E,  CJK UNIFIED IDEOGRAPH-643E
        0x7740, # G+7057  U+7740  uni7740,  CJK UNIFIED IDEOGRAPH-7740
        0x7A84, # G+7058  U+7A84  uni7A84,  CJK UNIFIED IDEOGRAPH-7A84
        0x932F, # G+7059  U+932F  uni932F,  CJK UNIFIED IDEOGRAPH-932F
        0x947F, # G+7060  U+947F  uni947F,  CJK UNIFIED IDEOGRAPH-947F
        0x9F6A, # G+7061  U+9F6A  uni9F6A,  CJK UNIFIED IDEOGRAPH-9F6A
        0x64B0, # G+7062  U+64B0  uni64B0,  CJK UNIFIED IDEOGRAPH-64B0
        0x6FAF, # G+7063  U+6FAF  uni6FAF,  CJK UNIFIED IDEOGRAPH-6FAF
        0x71E6, # G+7064  U+71E6  uni71E6,  CJK UNIFIED IDEOGRAPH-71E6
        0x74A8, # G+7065  U+74A8  uni74A8,  CJK UNIFIED IDEOGRAPH-74A8
        0x74DA, # G+7066  U+74DA  uni74DA,  CJK UNIFIED IDEOGRAPH-74DA
        0x7AC4, # G+7067  U+7AC4  uni7AC4,  CJK UNIFIED IDEOGRAPH-7AC4
        0x7C12, # G+7068  U+7C12  uni7C12,  CJK UNIFIED IDEOGRAPH-7C12
        0x7E82, # G+7069  U+7E82  uni7E82,  CJK UNIFIED IDEOGRAPH-7E82
        0x7CB2, # G+7070  U+7CB2  uni7CB2,  CJK UNIFIED IDEOGRAPH-7CB2
        0x7E98, # G+7071  U+7E98  uni7E98,  CJK UNIFIED IDEOGRAPH-7E98
        0x8B9A, # G+7072  U+8B9A  uni8B9A,  CJK UNIFIED IDEOGRAPH-8B9A
        0x8D0A, # G+7073  U+8D0A  uni8D0A,  CJK UNIFIED IDEOGRAPH-8D0A
        0x947D, # G+7074  U+947D  uni947D,  CJK UNIFIED IDEOGRAPH-947D
        0x9910, # G+7075  U+9910  uni9910,  CJK UNIFIED IDEOGRAPH-9910
        0x994C, # G+7076  U+994C  uni994C,  CJK UNIFIED IDEOGRAPH-994C
        0x5239, # G+7077  U+5239  uni5239,  CJK UNIFIED IDEOGRAPH-5239
        0x5BDF, # G+7078  U+5BDF  uni5BDF,  CJK UNIFIED IDEOGRAPH-5BDF
        0x64E6, # G+7079  U+64E6  uni64E6,  CJK UNIFIED IDEOGRAPH-64E6
        0x672D, # G+7080  U+672D  uni672D,  CJK UNIFIED IDEOGRAPH-672D
        0x7D2E, # G+7081  U+7D2E  uni7D2E,  CJK UNIFIED IDEOGRAPH-7D2E
        0x50ED, # G+7082  U+50ED  uni50ED,  CJK UNIFIED IDEOGRAPH-50ED
        0xF96B, # G+7083  U+F96B  uniF96B,  CJK COMPATIBILITY IDEOGRAPH-F96B
        0x5879, # G+7084  U+5879  uni5879,  CJK UNIFIED IDEOGRAPH-5879
        0x6158, # G+7085  U+6158  uni6158,  CJK UNIFIED IDEOGRAPH-6158
        0x6159, # G+7086  U+6159  uni6159,  CJK UNIFIED IDEOGRAPH-6159
        0x61FA, # G+7087  U+61FA  uni61FA,  CJK UNIFIED IDEOGRAPH-61FA
        0x65AC, # G+7088  U+65AC  uni65AC,  CJK UNIFIED IDEOGRAPH-65AC
        0x7AD9, # G+7089  U+7AD9  uni7AD9,  CJK UNIFIED IDEOGRAPH-7AD9
        0x8B92, # G+7090  U+8B92  uni8B92,  CJK UNIFIED IDEOGRAPH-8B92
        0x8B96, # G+7091  U+8B96  uni8B96,  CJK UNIFIED IDEOGRAPH-8B96
        0x5009, # G+7092  U+5009  uni5009,  CJK UNIFIED IDEOGRAPH-5009
        0x5021, # G+7093  U+5021  uni5021,  CJK UNIFIED IDEOGRAPH-5021
        0x5275, # G+7094  U+5275  uni5275,  CJK UNIFIED IDEOGRAPH-5275
        0x5531, # G+7095  U+5531  uni5531,  CJK UNIFIED IDEOGRAPH-5531
        0x5A3C, # G+7096  U+5A3C  uni5A3C,  CJK UNIFIED IDEOGRAPH-5A3C
        0x5EE0, # G+7097  U+5EE0  uni5EE0,  CJK UNIFIED IDEOGRAPH-5EE0
        0x5F70, # G+7098  U+5F70  uni5F70,  CJK UNIFIED IDEOGRAPH-5F70
        0x6134, # G+7099  U+6134  uni6134,  CJK UNIFIED IDEOGRAPH-6134
        0x655E, # G+7100  U+655E  uni655E,  CJK UNIFIED IDEOGRAPH-655E
        0x660C, # G+7101  U+660C  uni660C,  CJK UNIFIED IDEOGRAPH-660C
        0x6636, # G+7102  U+6636  uni6636,  CJK UNIFIED IDEOGRAPH-6636
        0x66A2, # G+7103  U+66A2  uni66A2,  CJK UNIFIED IDEOGRAPH-66A2
        0x69CD, # G+7104  U+69CD  uni69CD,  CJK UNIFIED IDEOGRAPH-69CD
        0x6EC4, # G+7105  U+6EC4  uni6EC4,  CJK UNIFIED IDEOGRAPH-6EC4
        0x6F32, # G+7106  U+6F32  uni6F32,  CJK UNIFIED IDEOGRAPH-6F32
        0x7316, # G+7107  U+7316  uni7316,  CJK UNIFIED IDEOGRAPH-7316
        0x7621, # G+7108  U+7621  uni7621,  CJK UNIFIED IDEOGRAPH-7621
        0x7A93, # G+7109  U+7A93  uni7A93,  CJK UNIFIED IDEOGRAPH-7A93
        0x8139, # G+7110  U+8139  uni8139,  CJK UNIFIED IDEOGRAPH-8139
        0x8259, # G+7111  U+8259  uni8259,  CJK UNIFIED IDEOGRAPH-8259
        0x83D6, # G+7112  U+83D6  uni83D6,  CJK UNIFIED IDEOGRAPH-83D6
        0x84BC, # G+7113  U+84BC  uni84BC,  CJK UNIFIED IDEOGRAPH-84BC
        0x50B5, # G+7114  U+50B5  uni50B5,  CJK UNIFIED IDEOGRAPH-50B5
        0x57F0, # G+7115  U+57F0  uni57F0,  CJK UNIFIED IDEOGRAPH-57F0
        0x5BC0, # G+7116  U+5BC0  uni5BC0,  CJK UNIFIED IDEOGRAPH-5BC0
        0x5BE8, # G+7117  U+5BE8  uni5BE8,  CJK UNIFIED IDEOGRAPH-5BE8
        0x5F69, # G+7118  U+5F69  uni5F69,  CJK UNIFIED IDEOGRAPH-5F69
        0x63A1, # G+7119  U+63A1  uni63A1,  CJK UNIFIED IDEOGRAPH-63A1
        0x7826, # G+7120  U+7826  uni7826,  CJK UNIFIED IDEOGRAPH-7826
        0x7DB5, # G+7121  U+7DB5  uni7DB5,  CJK UNIFIED IDEOGRAPH-7DB5
        0x83DC, # G+7122  U+83DC  uni83DC,  CJK UNIFIED IDEOGRAPH-83DC
        0x8521, # G+7123  U+8521  uni8521,  CJK UNIFIED IDEOGRAPH-8521
        0x91C7, # G+7124  U+91C7  uni91C7,  CJK UNIFIED IDEOGRAPH-91C7
        0x91F5, # G+7125  U+91F5  uni91F5,  CJK UNIFIED IDEOGRAPH-91F5
        0x518A, # G+7126  U+518A  uni518A,  CJK UNIFIED IDEOGRAPH-518A
        0x67F5, # G+7127  U+67F5  uni67F5,  CJK UNIFIED IDEOGRAPH-67F5
        0x7B56, # G+7128  U+7B56  uni7B56,  CJK UNIFIED IDEOGRAPH-7B56
        0x8CAC, # G+7129  U+8CAC  uni8CAC,  CJK UNIFIED IDEOGRAPH-8CAC
        0x51C4, # G+7130  U+51C4  uni51C4,  CJK UNIFIED IDEOGRAPH-51C4
        0x59BB, # G+7131  U+59BB  uni59BB,  CJK UNIFIED IDEOGRAPH-59BB
        0x60BD, # G+7132  U+60BD  uni60BD,  CJK UNIFIED IDEOGRAPH-60BD
        0x8655, # G+7133  U+8655  uni8655,  CJK UNIFIED IDEOGRAPH-8655
        0x501C, # G+7134  U+501C  uni501C,  CJK UNIFIED IDEOGRAPH-501C
        0x5254, # G+7135  U+5254  uni5254,  CJK UNIFIED IDEOGRAPH-5254
        0x5C3A, # G+7136  U+5C3A  uni5C3A,  CJK UNIFIED IDEOGRAPH-5C3A
        0x617D, # G+7137  U+617D  uni617D,  CJK UNIFIED IDEOGRAPH-617D
        0x621A, # G+7138  U+621A  uni621A,  CJK UNIFIED IDEOGRAPH-621A
        0x62D3, # G+7139  U+62D3  uni62D3,  CJK UNIFIED IDEOGRAPH-62D3
        0x64F2, # G+7140  U+64F2  uni64F2,  CJK UNIFIED IDEOGRAPH-64F2
        0x65A5, # G+7141  U+65A5  uni65A5,  CJK UNIFIED IDEOGRAPH-65A5
        0x6ECC, # G+7142  U+6ECC  uni6ECC,  CJK UNIFIED IDEOGRAPH-6ECC
        0x7620, # G+7143  U+7620  uni7620,  CJK UNIFIED IDEOGRAPH-7620
        0x810A, # G+7144  U+810A  uni810A,  CJK UNIFIED IDEOGRAPH-810A
        0x8E60, # G+7145  U+8E60  uni8E60,  CJK UNIFIED IDEOGRAPH-8E60
        0x965F, # G+7146  U+965F  uni965F,  CJK UNIFIED IDEOGRAPH-965F
        0x96BB, # G+7147  U+96BB  uni96BB,  CJK UNIFIED IDEOGRAPH-96BB
        0x4EDF, # G+7148  U+4EDF  uni4EDF,  CJK UNIFIED IDEOGRAPH-4EDF
        0x5343, # G+7149  U+5343  uni5343,  CJK UNIFIED IDEOGRAPH-5343
        0x5598, # G+7150  U+5598  uni5598,  CJK UNIFIED IDEOGRAPH-5598
        0x5929, # G+7151  U+5929  uni5929,  CJK UNIFIED IDEOGRAPH-5929
        0x5DDD, # G+7152  U+5DDD  uni5DDD,  CJK UNIFIED IDEOGRAPH-5DDD
        0x64C5, # G+7153  U+64C5  uni64C5,  CJK UNIFIED IDEOGRAPH-64C5
        0x6CC9, # G+7154  U+6CC9  uni6CC9,  CJK UNIFIED IDEOGRAPH-6CC9
        0x6DFA, # G+7155  U+6DFA  uni6DFA,  CJK UNIFIED IDEOGRAPH-6DFA
        0x7394, # G+7156  U+7394  uni7394,  CJK UNIFIED IDEOGRAPH-7394
        0x7A7F, # G+7157  U+7A7F  uni7A7F,  CJK UNIFIED IDEOGRAPH-7A7F
        0x821B, # G+7158  U+821B  uni821B,  CJK UNIFIED IDEOGRAPH-821B
        0x85A6, # G+7159  U+85A6  uni85A6,  CJK UNIFIED IDEOGRAPH-85A6
        0x8CE4, # G+7160  U+8CE4  uni8CE4,  CJK UNIFIED IDEOGRAPH-8CE4
        0x8E10, # G+7161  U+8E10  uni8E10,  CJK UNIFIED IDEOGRAPH-8E10
        0x9077, # G+7162  U+9077  uni9077,  CJK UNIFIED IDEOGRAPH-9077
        0x91E7, # G+7163  U+91E7  uni91E7,  CJK UNIFIED IDEOGRAPH-91E7
        0x95E1, # G+7164  U+95E1  uni95E1,  CJK UNIFIED IDEOGRAPH-95E1
        0x9621, # G+7165  U+9621  uni9621,  CJK UNIFIED IDEOGRAPH-9621
        0x97C6, # G+7166  U+97C6  uni97C6,  CJK UNIFIED IDEOGRAPH-97C6
        0x51F8, # G+7167  U+51F8  uni51F8,  CJK UNIFIED IDEOGRAPH-51F8
        0x54F2, # G+7168  U+54F2  uni54F2,  CJK UNIFIED IDEOGRAPH-54F2
        0x5586, # G+7169  U+5586  uni5586,  CJK UNIFIED IDEOGRAPH-5586
        0x5FB9, # G+7170  U+5FB9  uni5FB9,  CJK UNIFIED IDEOGRAPH-5FB9
        0x64A4, # G+7171  U+64A4  uni64A4,  CJK UNIFIED IDEOGRAPH-64A4
        0x6F88, # G+7172  U+6F88  uni6F88,  CJK UNIFIED IDEOGRAPH-6F88
        0x7DB4, # G+7173  U+7DB4  uni7DB4,  CJK UNIFIED IDEOGRAPH-7DB4
        0x8F1F, # G+7174  U+8F1F  uni8F1F,  CJK UNIFIED IDEOGRAPH-8F1F
        0x8F4D, # G+7175  U+8F4D  uni8F4D,  CJK UNIFIED IDEOGRAPH-8F4D
        0x9435, # G+7176  U+9435  uni9435,  CJK UNIFIED IDEOGRAPH-9435
        0x50C9, # G+7177  U+50C9  uni50C9,  CJK UNIFIED IDEOGRAPH-50C9
        0x5C16, # G+7178  U+5C16  uni5C16,  CJK UNIFIED IDEOGRAPH-5C16
        0x6CBE, # G+7179  U+6CBE  uni6CBE,  CJK UNIFIED IDEOGRAPH-6CBE
        0x6DFB, # G+7180  U+6DFB  uni6DFB,  CJK UNIFIED IDEOGRAPH-6DFB
        0x751B, # G+7181  U+751B  uni751B,  CJK UNIFIED IDEOGRAPH-751B
        0x77BB, # G+7182  U+77BB  uni77BB,  CJK UNIFIED IDEOGRAPH-77BB
        0x7C3D, # G+7183  U+7C3D  uni7C3D,  CJK UNIFIED IDEOGRAPH-7C3D
        0x7C64, # G+7184  U+7C64  uni7C64,  CJK UNIFIED IDEOGRAPH-7C64
        0x8A79, # G+7185  U+8A79  uni8A79,  CJK UNIFIED IDEOGRAPH-8A79
        0x8AC2, # G+7186  U+8AC2  uni8AC2,  CJK UNIFIED IDEOGRAPH-8AC2
        0x581E, # G+7187  U+581E  uni581E,  CJK UNIFIED IDEOGRAPH-581E
        0x59BE, # G+7188  U+59BE  uni59BE,  CJK UNIFIED IDEOGRAPH-59BE
        0x5E16, # G+7189  U+5E16  uni5E16,  CJK UNIFIED IDEOGRAPH-5E16
        0x6377, # G+7190  U+6377  uni6377,  CJK UNIFIED IDEOGRAPH-6377
        0x7252, # G+7191  U+7252  uni7252,  CJK UNIFIED IDEOGRAPH-7252
        0x758A, # G+7192  U+758A  uni758A,  CJK UNIFIED IDEOGRAPH-758A
        0x776B, # G+7193  U+776B  uni776B,  CJK UNIFIED IDEOGRAPH-776B
        0x8ADC, # G+7194  U+8ADC  uni8ADC,  CJK UNIFIED IDEOGRAPH-8ADC
        0x8CBC, # G+7195  U+8CBC  uni8CBC,  CJK UNIFIED IDEOGRAPH-8CBC
        0x8F12, # G+7196  U+8F12  uni8F12,  CJK UNIFIED IDEOGRAPH-8F12
        0x5EF3, # G+7197  U+5EF3  uni5EF3,  CJK UNIFIED IDEOGRAPH-5EF3
        0x6674, # G+7198  U+6674  uni6674,  CJK UNIFIED IDEOGRAPH-6674
        0x6DF8, # G+7199  U+6DF8  uni6DF8,  CJK UNIFIED IDEOGRAPH-6DF8
        0x807D, # G+7200  U+807D  uni807D,  CJK UNIFIED IDEOGRAPH-807D
        0x83C1, # G+7201  U+83C1  uni83C1,  CJK UNIFIED IDEOGRAPH-83C1
        0x8ACB, # G+7202  U+8ACB  uni8ACB,  CJK UNIFIED IDEOGRAPH-8ACB
        0x9751, # G+7203  U+9751  uni9751,  CJK UNIFIED IDEOGRAPH-9751
        0x9BD6, # G+7204  U+9BD6  uni9BD6,  CJK UNIFIED IDEOGRAPH-9BD6
        0x5243, # G+7205  U+5243  uni5243,  CJK UNIFIED IDEOGRAPH-5243
        0x66FF, # G+7206  U+66FF  uni66FF,  CJK UNIFIED IDEOGRAPH-66FF
        0x6D95, # G+7207  U+6D95  uni6D95,  CJK UNIFIED IDEOGRAPH-6D95
        0x6EEF, # G+7208  U+6EEF  uni6EEF,  CJK UNIFIED IDEOGRAPH-6EEF
        0x7DE0, # G+7209  U+7DE0  uni7DE0,  CJK UNIFIED IDEOGRAPH-7DE0
        0x8AE6, # G+7210  U+8AE6  uni8AE6,  CJK UNIFIED IDEOGRAPH-8AE6
        0x902E, # G+7211  U+902E  uni902E,  CJK UNIFIED IDEOGRAPH-902E
        0x905E, # G+7212  U+905E  uni905E,  CJK UNIFIED IDEOGRAPH-905E
        0x9AD4, # G+7213  U+9AD4  uni9AD4,  CJK UNIFIED IDEOGRAPH-9AD4
        0x521D, # G+7214  U+521D  uni521D,  CJK UNIFIED IDEOGRAPH-521D
        0x527F, # G+7215  U+527F  uni527F,  CJK UNIFIED IDEOGRAPH-527F
        0x54E8, # G+7216  U+54E8  uni54E8,  CJK UNIFIED IDEOGRAPH-54E8
        0x6194, # G+7217  U+6194  uni6194,  CJK UNIFIED IDEOGRAPH-6194
        0x6284, # G+7218  U+6284  uni6284,  CJK UNIFIED IDEOGRAPH-6284
        0x62DB, # G+7219  U+62DB  uni62DB,  CJK UNIFIED IDEOGRAPH-62DB
        0x68A2, # G+7220  U+68A2  uni68A2,  CJK UNIFIED IDEOGRAPH-68A2
        0x6912, # G+7221  U+6912  uni6912,  CJK UNIFIED IDEOGRAPH-6912
        0x695A, # G+7222  U+695A  uni695A,  CJK UNIFIED IDEOGRAPH-695A
        0x6A35, # G+7223  U+6A35  uni6A35,  CJK UNIFIED IDEOGRAPH-6A35
        0x7092, # G+7224  U+7092  uni7092,  CJK UNIFIED IDEOGRAPH-7092
        0x7126, # G+7225  U+7126  uni7126,  CJK UNIFIED IDEOGRAPH-7126
        0x785D, # G+7226  U+785D  uni785D,  CJK UNIFIED IDEOGRAPH-785D
        0x7901, # G+7227  U+7901  uni7901,  CJK UNIFIED IDEOGRAPH-7901
        0x790E, # G+7228  U+790E  uni790E,  CJK UNIFIED IDEOGRAPH-790E
        0x79D2, # G+7229  U+79D2  uni79D2,  CJK UNIFIED IDEOGRAPH-79D2
        0x7A0D, # G+7230  U+7A0D  uni7A0D,  CJK UNIFIED IDEOGRAPH-7A0D
        0x8096, # G+7231  U+8096  uni8096,  CJK UNIFIED IDEOGRAPH-8096
        0x8278, # G+7232  U+8278  uni8278,  CJK UNIFIED IDEOGRAPH-8278
        0x82D5, # G+7233  U+82D5  uni82D5,  CJK UNIFIED IDEOGRAPH-82D5
        0x8349, # G+7234  U+8349  uni8349,  CJK UNIFIED IDEOGRAPH-8349
        0x8549, # G+7235  U+8549  uni8549,  CJK UNIFIED IDEOGRAPH-8549
        0x8C82, # G+7236  U+8C82  uni8C82,  CJK UNIFIED IDEOGRAPH-8C82
        0x8D85, # G+7237  U+8D85  uni8D85,  CJK UNIFIED IDEOGRAPH-8D85
        0x9162, # G+7238  U+9162  uni9162,  CJK UNIFIED IDEOGRAPH-9162
        0x918B, # G+7239  U+918B  uni918B,  CJK UNIFIED IDEOGRAPH-918B
        0x91AE, # G+7240  U+91AE  uni91AE,  CJK UNIFIED IDEOGRAPH-91AE
        0x4FC3, # G+7241  U+4FC3  uni4FC3,  CJK UNIFIED IDEOGRAPH-4FC3
        0x56D1, # G+7242  U+56D1  uni56D1,  CJK UNIFIED IDEOGRAPH-56D1
        0x71ED, # G+7243  U+71ED  uni71ED,  CJK UNIFIED IDEOGRAPH-71ED
        0x77D7, # G+7244  U+77D7  uni77D7,  CJK UNIFIED IDEOGRAPH-77D7
        0x8700, # G+7245  U+8700  uni8700,  CJK UNIFIED IDEOGRAPH-8700
        0x89F8, # G+7246  U+89F8  uni89F8,  CJK UNIFIED IDEOGRAPH-89F8
        0x5BF8, # G+7247  U+5BF8  uni5BF8,  CJK UNIFIED IDEOGRAPH-5BF8
        0x5FD6, # G+7248  U+5FD6  uni5FD6,  CJK UNIFIED IDEOGRAPH-5FD6
        0x6751, # G+7249  U+6751  uni6751,  CJK UNIFIED IDEOGRAPH-6751
        0x90A8, # G+7250  U+90A8  uni90A8,  CJK UNIFIED IDEOGRAPH-90A8
        0x53E2, # G+7251  U+53E2  uni53E2,  CJK UNIFIED IDEOGRAPH-53E2
        0x585A, # G+7252  U+585A  uni585A,  CJK UNIFIED IDEOGRAPH-585A
        0x5BF5, # G+7253  U+5BF5  uni5BF5,  CJK UNIFIED IDEOGRAPH-5BF5
        0x60A4, # G+7254  U+60A4  uni60A4,  CJK UNIFIED IDEOGRAPH-60A4
        0x6181, # G+7255  U+6181  uni6181,  CJK UNIFIED IDEOGRAPH-6181
        0x6460, # G+7256  U+6460  uni6460,  CJK UNIFIED IDEOGRAPH-6460
        0x7E3D, # G+7257  U+7E3D  uni7E3D,  CJK UNIFIED IDEOGRAPH-7E3D
        0x8070, # G+7258  U+8070  uni8070,  CJK UNIFIED IDEOGRAPH-8070
        0x8525, # G+7259  U+8525  uni8525,  CJK UNIFIED IDEOGRAPH-8525
        0x9283, # G+7260  U+9283  uni9283,  CJK UNIFIED IDEOGRAPH-9283
        0x64AE, # G+7261  U+64AE  uni64AE,  CJK UNIFIED IDEOGRAPH-64AE
        0x50AC, # G+7262  U+50AC  uni50AC,  CJK UNIFIED IDEOGRAPH-50AC
        0x5D14, # G+7263  U+5D14  uni5D14,  CJK UNIFIED IDEOGRAPH-5D14
        0x6700, # G+7264  U+6700  uni6700,  CJK UNIFIED IDEOGRAPH-6700
        0x589C, # G+7265  U+589C  uni589C,  CJK UNIFIED IDEOGRAPH-589C
        0x62BD, # G+7266  U+62BD  uni62BD,  CJK UNIFIED IDEOGRAPH-62BD
        0x63A8, # G+7267  U+63A8  uni63A8,  CJK UNIFIED IDEOGRAPH-63A8
        0x690E, # G+7268  U+690E  uni690E,  CJK UNIFIED IDEOGRAPH-690E
        0x6978, # G+7269  U+6978  uni6978,  CJK UNIFIED IDEOGRAPH-6978
        0x6A1E, # G+7270  U+6A1E  uni6A1E,  CJK UNIFIED IDEOGRAPH-6A1E
        0x6E6B, # G+7271  U+6E6B  uni6E6B,  CJK UNIFIED IDEOGRAPH-6E6B
        0x76BA, # G+7272  U+76BA  uni76BA,  CJK UNIFIED IDEOGRAPH-76BA
        0x79CB, # G+7273  U+79CB  uni79CB,  CJK UNIFIED IDEOGRAPH-79CB
        0x82BB, # G+7274  U+82BB  uni82BB,  CJK UNIFIED IDEOGRAPH-82BB
        0x8429, # G+7275  U+8429  uni8429,  CJK UNIFIED IDEOGRAPH-8429
        0x8ACF, # G+7276  U+8ACF  uni8ACF,  CJK UNIFIED IDEOGRAPH-8ACF
        0x8DA8, # G+7277  U+8DA8  uni8DA8,  CJK UNIFIED IDEOGRAPH-8DA8
        0x8FFD, # G+7278  U+8FFD  uni8FFD,  CJK UNIFIED IDEOGRAPH-8FFD
        0x9112, # G+7279  U+9112  uni9112,  CJK UNIFIED IDEOGRAPH-9112
        0x914B, # G+7280  U+914B  uni914B,  CJK UNIFIED IDEOGRAPH-914B
        0x919C, # G+7281  U+919C  uni919C,  CJK UNIFIED IDEOGRAPH-919C
        0x9310, # G+7282  U+9310  uni9310,  CJK UNIFIED IDEOGRAPH-9310
        0x9318, # G+7283  U+9318  uni9318,  CJK UNIFIED IDEOGRAPH-9318
        0x939A, # G+7284  U+939A  uni939A,  CJK UNIFIED IDEOGRAPH-939A
        0x96DB, # G+7285  U+96DB  uni96DB,  CJK UNIFIED IDEOGRAPH-96DB
        0x9A36, # G+7286  U+9A36  uni9A36,  CJK UNIFIED IDEOGRAPH-9A36
        0x9C0D, # G+7287  U+9C0D  uni9C0D,  CJK UNIFIED IDEOGRAPH-9C0D
        0x4E11, # G+7288  U+4E11  uni4E11,  CJK UNIFIED IDEOGRAPH-4E11
        0x755C, # G+7289  U+755C  uni755C,  CJK UNIFIED IDEOGRAPH-755C
        0x795D, # G+7290  U+795D  uni795D,  CJK UNIFIED IDEOGRAPH-795D
        0x7AFA, # G+7291  U+7AFA  uni7AFA,  CJK UNIFIED IDEOGRAPH-7AFA
        0x7B51, # G+7292  U+7B51  uni7B51,  CJK UNIFIED IDEOGRAPH-7B51
        0x7BC9, # G+7293  U+7BC9  uni7BC9,  CJK UNIFIED IDEOGRAPH-7BC9
        0x7E2E, # G+7294  U+7E2E  uni7E2E,  CJK UNIFIED IDEOGRAPH-7E2E
        0x84C4, # G+7295  U+84C4  uni84C4,  CJK UNIFIED IDEOGRAPH-84C4
        0x8E59, # G+7296  U+8E59  uni8E59,  CJK UNIFIED IDEOGRAPH-8E59
        0x8E74, # G+7297  U+8E74  uni8E74,  CJK UNIFIED IDEOGRAPH-8E74
        0x8EF8, # G+7298  U+8EF8  uni8EF8,  CJK UNIFIED IDEOGRAPH-8EF8
        0x9010, # G+7299  U+9010  uni9010,  CJK UNIFIED IDEOGRAPH-9010
        0x6625, # G+7300  U+6625  uni6625,  CJK UNIFIED IDEOGRAPH-6625
        0x693F, # G+7301  U+693F  uni693F,  CJK UNIFIED IDEOGRAPH-693F
        0x7443, # G+7302  U+7443  uni7443,  CJK UNIFIED IDEOGRAPH-7443
        0x51FA, # G+7303  U+51FA  uni51FA,  CJK UNIFIED IDEOGRAPH-51FA
        0x672E, # G+7304  U+672E  uni672E,  CJK UNIFIED IDEOGRAPH-672E
        0x9EDC, # G+7305  U+9EDC  uni9EDC,  CJK UNIFIED IDEOGRAPH-9EDC
        0x5145, # G+7306  U+5145  uni5145,  CJK UNIFIED IDEOGRAPH-5145
        0x5FE0, # G+7307  U+5FE0  uni5FE0,  CJK UNIFIED IDEOGRAPH-5FE0
        0x6C96, # G+7308  U+6C96  uni6C96,  CJK UNIFIED IDEOGRAPH-6C96
        0x87F2, # G+7309  U+87F2  uni87F2,  CJK UNIFIED IDEOGRAPH-87F2
        0x885D, # G+7310  U+885D  uni885D,  CJK UNIFIED IDEOGRAPH-885D
        0x8877, # G+7311  U+8877  uni8877,  CJK UNIFIED IDEOGRAPH-8877
        0x60B4, # G+7312  U+60B4  uni60B4,  CJK UNIFIED IDEOGRAPH-60B4
        0x81B5, # G+7313  U+81B5  uni81B5,  CJK UNIFIED IDEOGRAPH-81B5
        0x8403, # G+7314  U+8403  uni8403,  CJK UNIFIED IDEOGRAPH-8403
        0x8D05, # G+7315  U+8D05  uni8D05,  CJK UNIFIED IDEOGRAPH-8D05
        0x53D6, # G+7316  U+53D6  uni53D6,  CJK UNIFIED IDEOGRAPH-53D6
        0x5439, # G+7317  U+5439  uni5439,  CJK UNIFIED IDEOGRAPH-5439
        0x5634, # G+7318  U+5634  uni5634,  CJK UNIFIED IDEOGRAPH-5634
        0x5A36, # G+7319  U+5A36  uni5A36,  CJK UNIFIED IDEOGRAPH-5A36
        0x5C31, # G+7320  U+5C31  uni5C31,  CJK UNIFIED IDEOGRAPH-5C31
        0x708A, # G+7321  U+708A  uni708A,  CJK UNIFIED IDEOGRAPH-708A
        0x7FE0, # G+7322  U+7FE0  uni7FE0,  CJK UNIFIED IDEOGRAPH-7FE0
        0x805A, # G+7323  U+805A  uni805A,  CJK UNIFIED IDEOGRAPH-805A
        0x8106, # G+7324  U+8106  uni8106,  CJK UNIFIED IDEOGRAPH-8106
        0x81ED, # G+7325  U+81ED  uni81ED,  CJK UNIFIED IDEOGRAPH-81ED
        0x8DA3, # G+7326  U+8DA3  uni8DA3,  CJK UNIFIED IDEOGRAPH-8DA3
        0x9189, # G+7327  U+9189  uni9189,  CJK UNIFIED IDEOGRAPH-9189
        0x9A5F, # G+7328  U+9A5F  uni9A5F,  CJK UNIFIED IDEOGRAPH-9A5F
        0x9DF2, # G+7329  U+9DF2  uni9DF2,  CJK UNIFIED IDEOGRAPH-9DF2
        0x5074, # G+7330  U+5074  uni5074,  CJK UNIFIED IDEOGRAPH-5074
        0x4EC4, # G+7331  U+4EC4  uni4EC4,  CJK UNIFIED IDEOGRAPH-4EC4
        0x53A0, # G+7332  U+53A0  uni53A0,  CJK UNIFIED IDEOGRAPH-53A0
        0x60FB, # G+7333  U+60FB  uni60FB,  CJK UNIFIED IDEOGRAPH-60FB
        0x6E2C, # G+7334  U+6E2C  uni6E2C,  CJK UNIFIED IDEOGRAPH-6E2C
        0x5C64, # G+7335  U+5C64  uni5C64,  CJK UNIFIED IDEOGRAPH-5C64
        0x4F88, # G+7336  U+4F88  uni4F88,  CJK UNIFIED IDEOGRAPH-4F88
        0x5024, # G+7337  U+5024  uni5024,  CJK UNIFIED IDEOGRAPH-5024
        0x55E4, # G+7338  U+55E4  uni55E4,  CJK UNIFIED IDEOGRAPH-55E4
        0x5CD9, # G+7339  U+5CD9  uni5CD9,  CJK UNIFIED IDEOGRAPH-5CD9
        0x5E5F, # G+7340  U+5E5F  uni5E5F,  CJK UNIFIED IDEOGRAPH-5E5F
        0x6065, # G+7341  U+6065  uni6065,  CJK UNIFIED IDEOGRAPH-6065
        0x6894, # G+7342  U+6894  uni6894,  CJK UNIFIED IDEOGRAPH-6894
        0x6CBB, # G+7343  U+6CBB  uni6CBB,  CJK UNIFIED IDEOGRAPH-6CBB
        0x6DC4, # G+7344  U+6DC4  uni6DC4,  CJK UNIFIED IDEOGRAPH-6DC4
        0x71BE, # G+7345  U+71BE  uni71BE,  CJK UNIFIED IDEOGRAPH-71BE
        0x75D4, # G+7346  U+75D4  uni75D4,  CJK UNIFIED IDEOGRAPH-75D4
        0x75F4, # G+7347  U+75F4  uni75F4,  CJK UNIFIED IDEOGRAPH-75F4
        0x7661, # G+7348  U+7661  uni7661,  CJK UNIFIED IDEOGRAPH-7661
        0x7A1A, # G+7349  U+7A1A  uni7A1A,  CJK UNIFIED IDEOGRAPH-7A1A
        0x7A49, # G+7350  U+7A49  uni7A49,  CJK UNIFIED IDEOGRAPH-7A49
        0x7DC7, # G+7351  U+7DC7  uni7DC7,  CJK UNIFIED IDEOGRAPH-7DC7
        0x7DFB, # G+7352  U+7DFB  uni7DFB,  CJK UNIFIED IDEOGRAPH-7DFB
        0x7F6E, # G+7353  U+7F6E  uni7F6E,  CJK UNIFIED IDEOGRAPH-7F6E
        0x81F4, # G+7354  U+81F4  uni81F4,  CJK UNIFIED IDEOGRAPH-81F4
        0x86A9, # G+7355  U+86A9  uni86A9,  CJK UNIFIED IDEOGRAPH-86A9
        0x8F1C, # G+7356  U+8F1C  uni8F1C,  CJK UNIFIED IDEOGRAPH-8F1C
        0x96C9, # G+7357  U+96C9  uni96C9,  CJK UNIFIED IDEOGRAPH-96C9
        0x99B3, # G+7358  U+99B3  uni99B3,  CJK UNIFIED IDEOGRAPH-99B3
        0x9F52, # G+7359  U+9F52  uni9F52,  CJK UNIFIED IDEOGRAPH-9F52
        0x5247, # G+7360  U+5247  uni5247,  CJK UNIFIED IDEOGRAPH-5247
        0x52C5, # G+7361  U+52C5  uni52C5,  CJK UNIFIED IDEOGRAPH-52C5
        0x98ED, # G+7362  U+98ED  uni98ED,  CJK UNIFIED IDEOGRAPH-98ED
        0x89AA, # G+7363  U+89AA  uni89AA,  CJK UNIFIED IDEOGRAPH-89AA
        0x4E03, # G+7364  U+4E03  uni4E03,  CJK UNIFIED IDEOGRAPH-4E03
        0x67D2, # G+7365  U+67D2  uni67D2,  CJK UNIFIED IDEOGRAPH-67D2
        0x6F06, # G+7366  U+6F06  uni6F06,  CJK UNIFIED IDEOGRAPH-6F06
        0x4FB5, # G+7367  U+4FB5  uni4FB5,  CJK UNIFIED IDEOGRAPH-4FB5
        0x5BE2, # G+7368  U+5BE2  uni5BE2,  CJK UNIFIED IDEOGRAPH-5BE2
        0x6795, # G+7369  U+6795  uni6795,  CJK UNIFIED IDEOGRAPH-6795
        0xF972, # G+7370  U+F972  uniF972,  CJK COMPATIBILITY IDEOGRAPH-F972
        0x6D78, # G+7371  U+6D78  uni6D78,  CJK UNIFIED IDEOGRAPH-6D78
        0x741B, # G+7372  U+741B  uni741B,  CJK UNIFIED IDEOGRAPH-741B
        0x7827, # G+7373  U+7827  uni7827,  CJK UNIFIED IDEOGRAPH-7827
        0x91DD, # G+7374  U+91DD  uni91DD,  CJK UNIFIED IDEOGRAPH-91DD
        0x937C, # G+7375  U+937C  uni937C,  CJK UNIFIED IDEOGRAPH-937C
        0x87C4, # G+7376  U+87C4  uni87C4,  CJK UNIFIED IDEOGRAPH-87C4
        0x79E4, # G+7377  U+79E4  uni79E4,  CJK UNIFIED IDEOGRAPH-79E4
        0x7A31, # G+7378  U+7A31  uni7A31,  CJK UNIFIED IDEOGRAPH-7A31
        0x5FEB, # G+7379  U+5FEB  uni5FEB,  CJK UNIFIED IDEOGRAPH-5FEB
        0x4ED6, # G+7380  U+4ED6  uni4ED6,  CJK UNIFIED IDEOGRAPH-4ED6
        0x54A4, # G+7381  U+54A4  uni54A4,  CJK UNIFIED IDEOGRAPH-54A4
        0x553E, # G+7382  U+553E  uni553E,  CJK UNIFIED IDEOGRAPH-553E
        0x58AE, # G+7383  U+58AE  uni58AE,  CJK UNIFIED IDEOGRAPH-58AE
        0x59A5, # G+7384  U+59A5  uni59A5,  CJK UNIFIED IDEOGRAPH-59A5
        0x60F0, # G+7385  U+60F0  uni60F0,  CJK UNIFIED IDEOGRAPH-60F0
        0x6253, # G+7386  U+6253  uni6253,  CJK UNIFIED IDEOGRAPH-6253
        0x62D6, # G+7387  U+62D6  uni62D6,  CJK UNIFIED IDEOGRAPH-62D6
        0x6736, # G+7388  U+6736  uni6736,  CJK UNIFIED IDEOGRAPH-6736
        0x6955, # G+7389  U+6955  uni6955,  CJK UNIFIED IDEOGRAPH-6955
        0x8235, # G+7390  U+8235  uni8235,  CJK UNIFIED IDEOGRAPH-8235
        0x9640, # G+7391  U+9640  uni9640,  CJK UNIFIED IDEOGRAPH-9640
        0x99B1, # G+7392  U+99B1  uni99B1,  CJK UNIFIED IDEOGRAPH-99B1
        0x99DD, # G+7393  U+99DD  uni99DD,  CJK UNIFIED IDEOGRAPH-99DD
        0x502C, # G+7394  U+502C  uni502C,  CJK UNIFIED IDEOGRAPH-502C
        0x5353, # G+7395  U+5353  uni5353,  CJK UNIFIED IDEOGRAPH-5353
        0x5544, # G+7396  U+5544  uni5544,  CJK UNIFIED IDEOGRAPH-5544
        0x577C, # G+7397  U+577C  uni577C,  CJK UNIFIED IDEOGRAPH-577C
        0x6258, # G+7398  U+6258  uni6258,  CJK UNIFIED IDEOGRAPH-6258
        0x64E2, # G+7399  U+64E2  uni64E2,  CJK UNIFIED IDEOGRAPH-64E2
        0x666B, # G+7400  U+666B  uni666B,  CJK UNIFIED IDEOGRAPH-666B
        0x67DD, # G+7401  U+67DD  uni67DD,  CJK UNIFIED IDEOGRAPH-67DD
        0x6FC1, # G+7402  U+6FC1  uni6FC1,  CJK UNIFIED IDEOGRAPH-6FC1
        0x6FEF, # G+7403  U+6FEF  uni6FEF,  CJK UNIFIED IDEOGRAPH-6FEF
        0x7422, # G+7404  U+7422  uni7422,  CJK UNIFIED IDEOGRAPH-7422
        0x7438, # G+7405  U+7438  uni7438,  CJK UNIFIED IDEOGRAPH-7438
        0x8A17, # G+7406  U+8A17  uni8A17,  CJK UNIFIED IDEOGRAPH-8A17
        0x9438, # G+7407  U+9438  uni9438,  CJK UNIFIED IDEOGRAPH-9438
        0x5451, # G+7408  U+5451  uni5451,  CJK UNIFIED IDEOGRAPH-5451
        0x5606, # G+7409  U+5606  uni5606,  CJK UNIFIED IDEOGRAPH-5606
        0x5766, # G+7410  U+5766  uni5766,  CJK UNIFIED IDEOGRAPH-5766
        0x5F48, # G+7411  U+5F48  uni5F48,  CJK UNIFIED IDEOGRAPH-5F48
        0x619A, # G+7412  U+619A  uni619A,  CJK UNIFIED IDEOGRAPH-619A
        0x6B4E, # G+7413  U+6B4E  uni6B4E,  CJK UNIFIED IDEOGRAPH-6B4E
        0x7058, # G+7414  U+7058  uni7058,  CJK UNIFIED IDEOGRAPH-7058
        0x70AD, # G+7415  U+70AD  uni70AD,  CJK UNIFIED IDEOGRAPH-70AD
        0x7DBB, # G+7416  U+7DBB  uni7DBB,  CJK UNIFIED IDEOGRAPH-7DBB
        0x8A95, # G+7417  U+8A95  uni8A95,  CJK UNIFIED IDEOGRAPH-8A95
        0x596A, # G+7418  U+596A  uni596A,  CJK UNIFIED IDEOGRAPH-596A
        0x812B, # G+7419  U+812B  uni812B,  CJK UNIFIED IDEOGRAPH-812B
        0x63A2, # G+7420  U+63A2  uni63A2,  CJK UNIFIED IDEOGRAPH-63A2
        0x7708, # G+7421  U+7708  uni7708,  CJK UNIFIED IDEOGRAPH-7708
        0x803D, # G+7422  U+803D  uni803D,  CJK UNIFIED IDEOGRAPH-803D
        0x8CAA, # G+7423  U+8CAA  uni8CAA,  CJK UNIFIED IDEOGRAPH-8CAA
        0x5854, # G+7424  U+5854  uni5854,  CJK UNIFIED IDEOGRAPH-5854
        0x642D, # G+7425  U+642D  uni642D,  CJK UNIFIED IDEOGRAPH-642D
        0x69BB, # G+7426  U+69BB  uni69BB,  CJK UNIFIED IDEOGRAPH-69BB
        0x5B95, # G+7427  U+5B95  uni5B95,  CJK UNIFIED IDEOGRAPH-5B95
        0x5E11, # G+7428  U+5E11  uni5E11,  CJK UNIFIED IDEOGRAPH-5E11
        0x6E6F, # G+7429  U+6E6F  uni6E6F,  CJK UNIFIED IDEOGRAPH-6E6F
        0x8569, # G+7430  U+8569  uni8569,  CJK UNIFIED IDEOGRAPH-8569
        0x514C, # G+7431  U+514C  uni514C,  CJK UNIFIED IDEOGRAPH-514C
        0x53F0, # G+7432  U+53F0  uni53F0,  CJK UNIFIED IDEOGRAPH-53F0
        0x592A, # G+7433  U+592A  uni592A,  CJK UNIFIED IDEOGRAPH-592A
        0x6020, # G+7434  U+6020  uni6020,  CJK UNIFIED IDEOGRAPH-6020
        0x614B, # G+7435  U+614B  uni614B,  CJK UNIFIED IDEOGRAPH-614B
        0x6B86, # G+7436  U+6B86  uni6B86,  CJK UNIFIED IDEOGRAPH-6B86
        0x6C70, # G+7437  U+6C70  uni6C70,  CJK UNIFIED IDEOGRAPH-6C70
        0x6CF0, # G+7438  U+6CF0  uni6CF0,  CJK UNIFIED IDEOGRAPH-6CF0
        0x7B1E, # G+7439  U+7B1E  uni7B1E,  CJK UNIFIED IDEOGRAPH-7B1E
        0x80CE, # G+7440  U+80CE  uni80CE,  CJK UNIFIED IDEOGRAPH-80CE
        0x82D4, # G+7441  U+82D4  uni82D4,  CJK UNIFIED IDEOGRAPH-82D4
        0x8DC6, # G+7442  U+8DC6  uni8DC6,  CJK UNIFIED IDEOGRAPH-8DC6
        0x90B0, # G+7443  U+90B0  uni90B0,  CJK UNIFIED IDEOGRAPH-90B0
        0x98B1, # G+7444  U+98B1  uni98B1,  CJK UNIFIED IDEOGRAPH-98B1
        0x64C7, # G+7445  U+64C7  uni64C7,  CJK UNIFIED IDEOGRAPH-64C7
        0x6FA4, # G+7446  U+6FA4  uni6FA4,  CJK UNIFIED IDEOGRAPH-6FA4
        0x6491, # G+7447  U+6491  uni6491,  CJK UNIFIED IDEOGRAPH-6491
        0x6504, # G+7448  U+6504  uni6504,  CJK UNIFIED IDEOGRAPH-6504
        0x514E, # G+7449  U+514E  uni514E,  CJK UNIFIED IDEOGRAPH-514E
        0x5410, # G+7450  U+5410  uni5410,  CJK UNIFIED IDEOGRAPH-5410
        0x571F, # G+7451  U+571F  uni571F,  CJK UNIFIED IDEOGRAPH-571F
        0x8A0E, # G+7452  U+8A0E  uni8A0E,  CJK UNIFIED IDEOGRAPH-8A0E
        0x615F, # G+7453  U+615F  uni615F,  CJK UNIFIED IDEOGRAPH-615F
        0x6876, # G+7454  U+6876  uni6876,  CJK UNIFIED IDEOGRAPH-6876
        0x75DB, # G+7455  U+75DB  uni75DB,  CJK UNIFIED IDEOGRAPH-75DB
        0x7B52, # G+7456  U+7B52  uni7B52,  CJK UNIFIED IDEOGRAPH-7B52
        0x7D71, # G+7457  U+7D71  uni7D71,  CJK UNIFIED IDEOGRAPH-7D71
        0x901A, # G+7458  U+901A  uni901A,  CJK UNIFIED IDEOGRAPH-901A
        0x5806, # G+7459  U+5806  uni5806,  CJK UNIFIED IDEOGRAPH-5806
        0x69CC, # G+7460  U+69CC  uni69CC,  CJK UNIFIED IDEOGRAPH-69CC
        0x817F, # G+7461  U+817F  uni817F,  CJK UNIFIED IDEOGRAPH-817F
        0x892A, # G+7462  U+892A  uni892A,  CJK UNIFIED IDEOGRAPH-892A
        0x9000, # G+7463  U+9000  uni9000,  CJK UNIFIED IDEOGRAPH-9000
        0x9839, # G+7464  U+9839  uni9839,  CJK UNIFIED IDEOGRAPH-9839
        0x5078, # G+7465  U+5078  uni5078,  CJK UNIFIED IDEOGRAPH-5078
        0x5957, # G+7466  U+5957  uni5957,  CJK UNIFIED IDEOGRAPH-5957
        0x59AC, # G+7467  U+59AC  uni59AC,  CJK UNIFIED IDEOGRAPH-59AC
        0x6295, # G+7468  U+6295  uni6295,  CJK UNIFIED IDEOGRAPH-6295
        0x900F, # G+7469  U+900F  uni900F,  CJK UNIFIED IDEOGRAPH-900F
        0x9B2A, # G+7470  U+9B2A  uni9B2A,  CJK UNIFIED IDEOGRAPH-9B2A
        0x615D, # G+7471  U+615D  uni615D,  CJK UNIFIED IDEOGRAPH-615D
        0x7279, # G+7472  U+7279  uni7279,  CJK UNIFIED IDEOGRAPH-7279
        0x95D6, # G+7473  U+95D6  uni95D6,  CJK UNIFIED IDEOGRAPH-95D6
        0x5761, # G+7474  U+5761  uni5761,  CJK UNIFIED IDEOGRAPH-5761
        0x5A46, # G+7475  U+5A46  uni5A46,  CJK UNIFIED IDEOGRAPH-5A46
        0x5DF4, # G+7476  U+5DF4  uni5DF4,  CJK UNIFIED IDEOGRAPH-5DF4
        0x628A, # G+7477  U+628A  uni628A,  CJK UNIFIED IDEOGRAPH-628A
        0x64AD, # G+7478  U+64AD  uni64AD,  CJK UNIFIED IDEOGRAPH-64AD
        0x64FA, # G+7479  U+64FA  uni64FA,  CJK UNIFIED IDEOGRAPH-64FA
        0x6777, # G+7480  U+6777  uni6777,  CJK UNIFIED IDEOGRAPH-6777
        0x6CE2, # G+7481  U+6CE2  uni6CE2,  CJK UNIFIED IDEOGRAPH-6CE2
        0x6D3E, # G+7482  U+6D3E  uni6D3E,  CJK UNIFIED IDEOGRAPH-6D3E
        0x722C, # G+7483  U+722C  uni722C,  CJK UNIFIED IDEOGRAPH-722C
        0x7436, # G+7484  U+7436  uni7436,  CJK UNIFIED IDEOGRAPH-7436
        0x7834, # G+7485  U+7834  uni7834,  CJK UNIFIED IDEOGRAPH-7834
        0x7F77, # G+7486  U+7F77  uni7F77,  CJK UNIFIED IDEOGRAPH-7F77
        0x82AD, # G+7487  U+82AD  uni82AD,  CJK UNIFIED IDEOGRAPH-82AD
        0x8DDB, # G+7488  U+8DDB  uni8DDB,  CJK UNIFIED IDEOGRAPH-8DDB
        0x9817, # G+7489  U+9817  uni9817,  CJK UNIFIED IDEOGRAPH-9817
        0x5224, # G+7490  U+5224  uni5224,  CJK UNIFIED IDEOGRAPH-5224
        0x5742, # G+7491  U+5742  uni5742,  CJK UNIFIED IDEOGRAPH-5742
        0x677F, # G+7492  U+677F  uni677F,  CJK UNIFIED IDEOGRAPH-677F
        0x7248, # G+7493  U+7248  uni7248,  CJK UNIFIED IDEOGRAPH-7248
        0x74E3, # G+7494  U+74E3  uni74E3,  CJK UNIFIED IDEOGRAPH-74E3
        0x8CA9, # G+7495  U+8CA9  uni8CA9,  CJK UNIFIED IDEOGRAPH-8CA9
        0x8FA6, # G+7496  U+8FA6  uni8FA6,  CJK UNIFIED IDEOGRAPH-8FA6
        0x9211, # G+7497  U+9211  uni9211,  CJK UNIFIED IDEOGRAPH-9211
        0x962A, # G+7498  U+962A  uni962A,  CJK UNIFIED IDEOGRAPH-962A
        0x516B, # G+7499  U+516B  uni516B,  CJK UNIFIED IDEOGRAPH-516B
        0x53ED, # G+7500  U+53ED  uni53ED,  CJK UNIFIED IDEOGRAPH-53ED
        0x634C, # G+7501  U+634C  uni634C,  CJK UNIFIED IDEOGRAPH-634C
        0x4F69, # G+7502  U+4F69  uni4F69,  CJK UNIFIED IDEOGRAPH-4F69
        0x5504, # G+7503  U+5504  uni5504,  CJK UNIFIED IDEOGRAPH-5504
        0x6096, # G+7504  U+6096  uni6096,  CJK UNIFIED IDEOGRAPH-6096
        0x6557, # G+7505  U+6557  uni6557,  CJK UNIFIED IDEOGRAPH-6557
        0x6C9B, # G+7506  U+6C9B  uni6C9B,  CJK UNIFIED IDEOGRAPH-6C9B
        0x6D7F, # G+7507  U+6D7F  uni6D7F,  CJK UNIFIED IDEOGRAPH-6D7F
        0x724C, # G+7508  U+724C  uni724C,  CJK UNIFIED IDEOGRAPH-724C
        0x72FD, # G+7509  U+72FD  uni72FD,  CJK UNIFIED IDEOGRAPH-72FD
        0x7A17, # G+7510  U+7A17  uni7A17,  CJK UNIFIED IDEOGRAPH-7A17
        0x8987, # G+7511  U+8987  uni8987,  CJK UNIFIED IDEOGRAPH-8987
        0x8C9D, # G+7512  U+8C9D  uni8C9D,  CJK UNIFIED IDEOGRAPH-8C9D
        0x5F6D, # G+7513  U+5F6D  uni5F6D,  CJK UNIFIED IDEOGRAPH-5F6D
        0x6F8E, # G+7514  U+6F8E  uni6F8E,  CJK UNIFIED IDEOGRAPH-6F8E
        0x70F9, # G+7515  U+70F9  uni70F9,  CJK UNIFIED IDEOGRAPH-70F9
        0x81A8, # G+7516  U+81A8  uni81A8,  CJK UNIFIED IDEOGRAPH-81A8
        0x610E, # G+7517  U+610E  uni610E,  CJK UNIFIED IDEOGRAPH-610E
        0xF965, # G+7518  U+F965  uniF965,  CJK COMPATIBILITY IDEOGRAPH-F965
        0x504F, # G+7519  U+504F  uni504F,  CJK UNIFIED IDEOGRAPH-504F
        0x6241, # G+7520  U+6241  uni6241,  CJK UNIFIED IDEOGRAPH-6241
        0x7247, # G+7521  U+7247  uni7247,  CJK UNIFIED IDEOGRAPH-7247
        0x7BC7, # G+7522  U+7BC7  uni7BC7,  CJK UNIFIED IDEOGRAPH-7BC7
        0x7DE8, # G+7523  U+7DE8  uni7DE8,  CJK UNIFIED IDEOGRAPH-7DE8
        0x7FE9, # G+7524  U+7FE9  uni7FE9,  CJK UNIFIED IDEOGRAPH-7FE9
        0x904D, # G+7525  U+904D  uni904D,  CJK UNIFIED IDEOGRAPH-904D
        0x97AD, # G+7526  U+97AD  uni97AD,  CJK UNIFIED IDEOGRAPH-97AD
        0x9A19, # G+7527  U+9A19  uni9A19,  CJK UNIFIED IDEOGRAPH-9A19
        0x8CB6, # G+7528  U+8CB6  uni8CB6,  CJK UNIFIED IDEOGRAPH-8CB6
        0x576A, # G+7529  U+576A  uni576A,  CJK UNIFIED IDEOGRAPH-576A
        0x5E73, # G+7530  U+5E73  uni5E73,  CJK UNIFIED IDEOGRAPH-5E73
        0x67B0, # G+7531  U+67B0  uni67B0,  CJK UNIFIED IDEOGRAPH-67B0
        0x840D, # G+7532  U+840D  uni840D,  CJK UNIFIED IDEOGRAPH-840D
        0x8A55, # G+7533  U+8A55  uni8A55,  CJK UNIFIED IDEOGRAPH-8A55
        0x5420, # G+7534  U+5420  uni5420,  CJK UNIFIED IDEOGRAPH-5420
        0x5B16, # G+7535  U+5B16  uni5B16,  CJK UNIFIED IDEOGRAPH-5B16
        0x5E63, # G+7536  U+5E63  uni5E63,  CJK UNIFIED IDEOGRAPH-5E63
        0x5EE2, # G+7537  U+5EE2  uni5EE2,  CJK UNIFIED IDEOGRAPH-5EE2
        0x5F0A, # G+7538  U+5F0A  uni5F0A,  CJK UNIFIED IDEOGRAPH-5F0A
        0x6583, # G+7539  U+6583  uni6583,  CJK UNIFIED IDEOGRAPH-6583
        0x80BA, # G+7540  U+80BA  uni80BA,  CJK UNIFIED IDEOGRAPH-80BA
        0x853D, # G+7541  U+853D  uni853D,  CJK UNIFIED IDEOGRAPH-853D
        0x9589, # G+7542  U+9589  uni9589,  CJK UNIFIED IDEOGRAPH-9589
        0x965B, # G+7543  U+965B  uni965B,  CJK UNIFIED IDEOGRAPH-965B
        0x4F48, # G+7544  U+4F48  uni4F48,  CJK UNIFIED IDEOGRAPH-4F48
        0x5305, # G+7545  U+5305  uni5305,  CJK UNIFIED IDEOGRAPH-5305
        0x530D, # G+7546  U+530D  uni530D,  CJK UNIFIED IDEOGRAPH-530D
        0x530F, # G+7547  U+530F  uni530F,  CJK UNIFIED IDEOGRAPH-530F
        0x5486, # G+7548  U+5486  uni5486,  CJK UNIFIED IDEOGRAPH-5486
        0x54FA, # G+7549  U+54FA  uni54FA,  CJK UNIFIED IDEOGRAPH-54FA
        0x5703, # G+7550  U+5703  uni5703,  CJK UNIFIED IDEOGRAPH-5703
        0x5E03, # G+7551  U+5E03  uni5E03,  CJK UNIFIED IDEOGRAPH-5E03
        0x6016, # G+7552  U+6016  uni6016,  CJK UNIFIED IDEOGRAPH-6016
        0x629B, # G+7553  U+629B  uni629B,  CJK UNIFIED IDEOGRAPH-629B
        0x62B1, # G+7554  U+62B1  uni62B1,  CJK UNIFIED IDEOGRAPH-62B1
        0x6355, # G+7555  U+6355  uni6355,  CJK UNIFIED IDEOGRAPH-6355
        0x6CE1, # G+7556  U+6CE1  uni6CE1,  CJK UNIFIED IDEOGRAPH-6CE1
        0x6D66, # G+7557  U+6D66  uni6D66,  CJK UNIFIED IDEOGRAPH-6D66
        0x75B1, # G+7558  U+75B1  uni75B1,  CJK UNIFIED IDEOGRAPH-75B1
        0x7832, # G+7559  U+7832  uni7832,  CJK UNIFIED IDEOGRAPH-7832
        0x80DE, # G+7560  U+80DE  uni80DE,  CJK UNIFIED IDEOGRAPH-80DE
        0x812F, # G+7561  U+812F  uni812F,  CJK UNIFIED IDEOGRAPH-812F
        0x82DE, # G+7562  U+82DE  uni82DE,  CJK UNIFIED IDEOGRAPH-82DE
        0x8461, # G+7563  U+8461  uni8461,  CJK UNIFIED IDEOGRAPH-8461
        0x84B2, # G+7564  U+84B2  uni84B2,  CJK UNIFIED IDEOGRAPH-84B2
        0x888D, # G+7565  U+888D  uni888D,  CJK UNIFIED IDEOGRAPH-888D
        0x8912, # G+7566  U+8912  uni8912,  CJK UNIFIED IDEOGRAPH-8912
        0x900B, # G+7567  U+900B  uni900B,  CJK UNIFIED IDEOGRAPH-900B
        0x92EA, # G+7568  U+92EA  uni92EA,  CJK UNIFIED IDEOGRAPH-92EA
        0x98FD, # G+7569  U+98FD  uni98FD,  CJK UNIFIED IDEOGRAPH-98FD
        0x9B91, # G+7570  U+9B91  uni9B91,  CJK UNIFIED IDEOGRAPH-9B91
        0x5E45, # G+7571  U+5E45  uni5E45,  CJK UNIFIED IDEOGRAPH-5E45
        0xFA06, # G+7572  U+FA06  uniFA06,  CJK COMPATIBILITY IDEOGRAPH-FA06
        0x66DD, # G+7573  U+66DD  uni66DD,  CJK UNIFIED IDEOGRAPH-66DD
        0x7011, # G+7574  U+7011  uni7011,  CJK UNIFIED IDEOGRAPH-7011
        0x7206, # G+7575  U+7206  uni7206,  CJK UNIFIED IDEOGRAPH-7206
        0x4FF5, # G+7576  U+4FF5  uni4FF5,  CJK UNIFIED IDEOGRAPH-4FF5
        0x527D, # G+7577  U+527D  uni527D,  CJK UNIFIED IDEOGRAPH-527D
        0x5F6A, # G+7578  U+5F6A  uni5F6A,  CJK UNIFIED IDEOGRAPH-5F6A
        0x6153, # G+7579  U+6153  uni6153,  CJK UNIFIED IDEOGRAPH-6153
        0x6753, # G+7580  U+6753  uni6753,  CJK UNIFIED IDEOGRAPH-6753
        0x6A19, # G+7581  U+6A19  uni6A19,  CJK UNIFIED IDEOGRAPH-6A19
        0x6F02, # G+7582  U+6F02  uni6F02,  CJK UNIFIED IDEOGRAPH-6F02
        0x74E2, # G+7583  U+74E2  uni74E2,  CJK UNIFIED IDEOGRAPH-74E2
        0x7968, # G+7584  U+7968  uni7968,  CJK UNIFIED IDEOGRAPH-7968
        0x8868, # G+7585  U+8868  uni8868,  CJK UNIFIED IDEOGRAPH-8868
        0x8C79, # G+7586  U+8C79  uni8C79,  CJK UNIFIED IDEOGRAPH-8C79
        0x98C7, # G+7587  U+98C7  uni98C7,  CJK UNIFIED IDEOGRAPH-98C7
        0x98C4, # G+7588  U+98C4  uni98C4,  CJK UNIFIED IDEOGRAPH-98C4
        0x9A43, # G+7589  U+9A43  uni9A43,  CJK UNIFIED IDEOGRAPH-9A43
        0x54C1, # G+7590  U+54C1  uni54C1,  CJK UNIFIED IDEOGRAPH-54C1
        0x7A1F, # G+7591  U+7A1F  uni7A1F,  CJK UNIFIED IDEOGRAPH-7A1F
        0x6953, # G+7592  U+6953  uni6953,  CJK UNIFIED IDEOGRAPH-6953
        0x8AF7, # G+7593  U+8AF7  uni8AF7,  CJK UNIFIED IDEOGRAPH-8AF7
        0x8C4A, # G+7594  U+8C4A  uni8C4A,  CJK UNIFIED IDEOGRAPH-8C4A
        0x98A8, # G+7595  U+98A8  uni98A8,  CJK UNIFIED IDEOGRAPH-98A8
        0x99AE, # G+7596  U+99AE  uni99AE,  CJK UNIFIED IDEOGRAPH-99AE
        0x5F7C, # G+7597  U+5F7C  uni5F7C,  CJK UNIFIED IDEOGRAPH-5F7C
        0x62AB, # G+7598  U+62AB  uni62AB,  CJK UNIFIED IDEOGRAPH-62AB
        0x75B2, # G+7599  U+75B2  uni75B2,  CJK UNIFIED IDEOGRAPH-75B2
        0x76AE, # G+7600  U+76AE  uni76AE,  CJK UNIFIED IDEOGRAPH-76AE
        0x88AB, # G+7601  U+88AB  uni88AB,  CJK UNIFIED IDEOGRAPH-88AB
        0x907F, # G+7602  U+907F  uni907F,  CJK UNIFIED IDEOGRAPH-907F
        0x9642, # G+7603  U+9642  uni9642,  CJK UNIFIED IDEOGRAPH-9642
        0x5339, # G+7604  U+5339  uni5339,  CJK UNIFIED IDEOGRAPH-5339
        0x5F3C, # G+7605  U+5F3C  uni5F3C,  CJK UNIFIED IDEOGRAPH-5F3C
        0x5FC5, # G+7606  U+5FC5  uni5FC5,  CJK UNIFIED IDEOGRAPH-5FC5
        0xF968, # G+7607  U+F968  uniF968,  CJK COMPATIBILITY IDEOGRAPH-F968
        0x73CC, # G+7608  U+73CC  uni73CC,  CJK UNIFIED IDEOGRAPH-73CC
        0x7562, # G+7609  U+7562  uni7562,  CJK UNIFIED IDEOGRAPH-7562
        0x758B, # G+7610  U+758B  uni758B,  CJK UNIFIED IDEOGRAPH-758B
        0x7B46, # G+7611  U+7B46  uni7B46,  CJK UNIFIED IDEOGRAPH-7B46
        0x82FE, # G+7612  U+82FE  uni82FE,  CJK UNIFIED IDEOGRAPH-82FE
        0x999D, # G+7613  U+999D  uni999D,  CJK UNIFIED IDEOGRAPH-999D
        0x4E4F, # G+7614  U+4E4F  uni4E4F,  CJK UNIFIED IDEOGRAPH-4E4F
        0x903C, # G+7615  U+903C  uni903C,  CJK UNIFIED IDEOGRAPH-903C
        0x4E0B, # G+7616  U+4E0B  uni4E0B,  CJK UNIFIED IDEOGRAPH-4E0B
        0x4F55, # G+7617  U+4F55  uni4F55,  CJK UNIFIED IDEOGRAPH-4F55
        0x53A6, # G+7618  U+53A6  uni53A6,  CJK UNIFIED IDEOGRAPH-53A6
        0x590F, # G+7619  U+590F  uni590F,  CJK UNIFIED IDEOGRAPH-590F
        0x5EC8, # G+7620  U+5EC8  uni5EC8,  CJK UNIFIED IDEOGRAPH-5EC8
        0x6630, # G+7621  U+6630  uni6630,  CJK UNIFIED IDEOGRAPH-6630
        0x6CB3, # G+7622  U+6CB3  uni6CB3,  CJK UNIFIED IDEOGRAPH-6CB3
        0x7455, # G+7623  U+7455  uni7455,  CJK UNIFIED IDEOGRAPH-7455
        0x8377, # G+7624  U+8377  uni8377,  CJK UNIFIED IDEOGRAPH-8377
        0x8766, # G+7625  U+8766  uni8766,  CJK UNIFIED IDEOGRAPH-8766
        0x8CC0, # G+7626  U+8CC0  uni8CC0,  CJK UNIFIED IDEOGRAPH-8CC0
        0x9050, # G+7627  U+9050  uni9050,  CJK UNIFIED IDEOGRAPH-9050
        0x971E, # G+7628  U+971E  uni971E,  CJK UNIFIED IDEOGRAPH-971E
        0x9C15, # G+7629  U+9C15  uni9C15,  CJK UNIFIED IDEOGRAPH-9C15
        0x58D1, # G+7630  U+58D1  uni58D1,  CJK UNIFIED IDEOGRAPH-58D1
        0x5B78, # G+7631  U+5B78  uni5B78,  CJK UNIFIED IDEOGRAPH-5B78
        0x8650, # G+7632  U+8650  uni8650,  CJK UNIFIED IDEOGRAPH-8650
        0x8B14, # G+7633  U+8B14  uni8B14,  CJK UNIFIED IDEOGRAPH-8B14
        0x9DB4, # G+7634  U+9DB4  uni9DB4,  CJK UNIFIED IDEOGRAPH-9DB4
        0x5BD2, # G+7635  U+5BD2  uni5BD2,  CJK UNIFIED IDEOGRAPH-5BD2
        0x6068, # G+7636  U+6068  uni6068,  CJK UNIFIED IDEOGRAPH-6068
        0x608D, # G+7637  U+608D  uni608D,  CJK UNIFIED IDEOGRAPH-608D
        0x65F1, # G+7638  U+65F1  uni65F1,  CJK UNIFIED IDEOGRAPH-65F1
        0x6C57, # G+7639  U+6C57  uni6C57,  CJK UNIFIED IDEOGRAPH-6C57
        0x6F22, # G+7640  U+6F22  uni6F22,  CJK UNIFIED IDEOGRAPH-6F22
        0x6FA3, # G+7641  U+6FA3  uni6FA3,  CJK UNIFIED IDEOGRAPH-6FA3
        0x701A, # G+7642  U+701A  uni701A,  CJK UNIFIED IDEOGRAPH-701A
        0x7F55, # G+7643  U+7F55  uni7F55,  CJK UNIFIED IDEOGRAPH-7F55
        0x7FF0, # G+7644  U+7FF0  uni7FF0,  CJK UNIFIED IDEOGRAPH-7FF0
        0x9591, # G+7645  U+9591  uni9591,  CJK UNIFIED IDEOGRAPH-9591
        0x9592, # G+7646  U+9592  uni9592,  CJK UNIFIED IDEOGRAPH-9592
        0x9650, # G+7647  U+9650  uni9650,  CJK UNIFIED IDEOGRAPH-9650
        0x97D3, # G+7648  U+97D3  uni97D3,  CJK UNIFIED IDEOGRAPH-97D3
        0x5272, # G+7649  U+5272  uni5272,  CJK UNIFIED IDEOGRAPH-5272
        0x8F44, # G+7650  U+8F44  uni8F44,  CJK UNIFIED IDEOGRAPH-8F44
        0x51FD, # G+7651  U+51FD  uni51FD,  CJK UNIFIED IDEOGRAPH-51FD
        0x542B, # G+7652  U+542B  uni542B,  CJK UNIFIED IDEOGRAPH-542B
        0x54B8, # G+7653  U+54B8  uni54B8,  CJK UNIFIED IDEOGRAPH-54B8
        0x5563, # G+7654  U+5563  uni5563,  CJK UNIFIED IDEOGRAPH-5563
        0x558A, # G+7655  U+558A  uni558A,  CJK UNIFIED IDEOGRAPH-558A
        0x6ABB, # G+7656  U+6ABB  uni6ABB,  CJK UNIFIED IDEOGRAPH-6ABB
        0x6DB5, # G+7657  U+6DB5  uni6DB5,  CJK UNIFIED IDEOGRAPH-6DB5
        0x7DD8, # G+7658  U+7DD8  uni7DD8,  CJK UNIFIED IDEOGRAPH-7DD8
        0x8266, # G+7659  U+8266  uni8266,  CJK UNIFIED IDEOGRAPH-8266
        0x929C, # G+7660  U+929C  uni929C,  CJK UNIFIED IDEOGRAPH-929C
        0x9677, # G+7661  U+9677  uni9677,  CJK UNIFIED IDEOGRAPH-9677
        0x9E79, # G+7662  U+9E79  uni9E79,  CJK UNIFIED IDEOGRAPH-9E79
        0x5408, # G+7663  U+5408  uni5408,  CJK UNIFIED IDEOGRAPH-5408
        0x54C8, # G+7664  U+54C8  uni54C8,  CJK UNIFIED IDEOGRAPH-54C8
        0x76D2, # G+7665  U+76D2  uni76D2,  CJK UNIFIED IDEOGRAPH-76D2
        0x86E4, # G+7666  U+86E4  uni86E4,  CJK UNIFIED IDEOGRAPH-86E4
        0x95A4, # G+7667  U+95A4  uni95A4,  CJK UNIFIED IDEOGRAPH-95A4
        0x95D4, # G+7668  U+95D4  uni95D4,  CJK UNIFIED IDEOGRAPH-95D4
        0x965C, # G+7669  U+965C  uni965C,  CJK UNIFIED IDEOGRAPH-965C
        0x4EA2, # G+7670  U+4EA2  uni4EA2,  CJK UNIFIED IDEOGRAPH-4EA2
        0x4F09, # G+7671  U+4F09  uni4F09,  CJK UNIFIED IDEOGRAPH-4F09
        0x59EE, # G+7672  U+59EE  uni59EE,  CJK UNIFIED IDEOGRAPH-59EE
        0x5AE6, # G+7673  U+5AE6  uni5AE6,  CJK UNIFIED IDEOGRAPH-5AE6
        0x5DF7, # G+7674  U+5DF7  uni5DF7,  CJK UNIFIED IDEOGRAPH-5DF7
        0x6052, # G+7675  U+6052  uni6052,  CJK UNIFIED IDEOGRAPH-6052
        0x6297, # G+7676  U+6297  uni6297,  CJK UNIFIED IDEOGRAPH-6297
        0x676D, # G+7677  U+676D  uni676D,  CJK UNIFIED IDEOGRAPH-676D
        0x6841, # G+7678  U+6841  uni6841,  CJK UNIFIED IDEOGRAPH-6841
        0x6C86, # G+7679  U+6C86  uni6C86,  CJK UNIFIED IDEOGRAPH-6C86
        0x6E2F, # G+7680  U+6E2F  uni6E2F,  CJK UNIFIED IDEOGRAPH-6E2F
        0x7F38, # G+7681  U+7F38  uni7F38,  CJK UNIFIED IDEOGRAPH-7F38
        0x809B, # G+7682  U+809B  uni809B,  CJK UNIFIED IDEOGRAPH-809B
        0x822A, # G+7683  U+822A  uni822A,  CJK UNIFIED IDEOGRAPH-822A
        0x9805, # G+7684  U+9805  uni9805,  CJK UNIFIED IDEOGRAPH-9805
        0x4EA5, # G+7685  U+4EA5  uni4EA5,  CJK UNIFIED IDEOGRAPH-4EA5
        0x5055, # G+7686  U+5055  uni5055,  CJK UNIFIED IDEOGRAPH-5055
        0x54B3, # G+7687  U+54B3  uni54B3,  CJK UNIFIED IDEOGRAPH-54B3
        0x5793, # G+7688  U+5793  uni5793,  CJK UNIFIED IDEOGRAPH-5793
        0x595A, # G+7689  U+595A  uni595A,  CJK UNIFIED IDEOGRAPH-595A
        0x5B69, # G+7690  U+5B69  uni5B69,  CJK UNIFIED IDEOGRAPH-5B69
        0x5BB3, # G+7691  U+5BB3  uni5BB3,  CJK UNIFIED IDEOGRAPH-5BB3
        0x61C8, # G+7692  U+61C8  uni61C8,  CJK UNIFIED IDEOGRAPH-61C8
        0x6977, # G+7693  U+6977  uni6977,  CJK UNIFIED IDEOGRAPH-6977
        0x6D77, # G+7694  U+6D77  uni6D77,  CJK UNIFIED IDEOGRAPH-6D77
        0x7023, # G+7695  U+7023  uni7023,  CJK UNIFIED IDEOGRAPH-7023
        0x87F9, # G+7696  U+87F9  uni87F9,  CJK UNIFIED IDEOGRAPH-87F9
        0x89E3, # G+7697  U+89E3  uni89E3,  CJK UNIFIED IDEOGRAPH-89E3
        0x8A72, # G+7698  U+8A72  uni8A72,  CJK UNIFIED IDEOGRAPH-8A72
        0x8AE7, # G+7699  U+8AE7  uni8AE7,  CJK UNIFIED IDEOGRAPH-8AE7
        0x9082, # G+7700  U+9082  uni9082,  CJK UNIFIED IDEOGRAPH-9082
        0x99ED, # G+7701  U+99ED  uni99ED,  CJK UNIFIED IDEOGRAPH-99ED
        0x9AB8, # G+7702  U+9AB8  uni9AB8,  CJK UNIFIED IDEOGRAPH-9AB8
        0x52BE, # G+7703  U+52BE  uni52BE,  CJK UNIFIED IDEOGRAPH-52BE
        0x6838, # G+7704  U+6838  uni6838,  CJK UNIFIED IDEOGRAPH-6838
        0x5016, # G+7705  U+5016  uni5016,  CJK UNIFIED IDEOGRAPH-5016
        0x5E78, # G+7706  U+5E78  uni5E78,  CJK UNIFIED IDEOGRAPH-5E78
        0x674F, # G+7707  U+674F  uni674F,  CJK UNIFIED IDEOGRAPH-674F
        0x8347, # G+7708  U+8347  uni8347,  CJK UNIFIED IDEOGRAPH-8347
        0xFA08, # G+7709  U+FA08  uniFA08,  CJK COMPATIBILITY IDEOGRAPH-FA08
        0x4EAB, # G+7710  U+4EAB  uni4EAB,  CJK UNIFIED IDEOGRAPH-4EAB
        0x5411, # G+7711  U+5411  uni5411,  CJK UNIFIED IDEOGRAPH-5411
        0x56AE, # G+7712  U+56AE  uni56AE,  CJK UNIFIED IDEOGRAPH-56AE
        0x73E6, # G+7713  U+73E6  uni73E6,  CJK UNIFIED IDEOGRAPH-73E6
        0x9115, # G+7714  U+9115  uni9115,  CJK UNIFIED IDEOGRAPH-9115
        0x97FF, # G+7715  U+97FF  uni97FF,  CJK UNIFIED IDEOGRAPH-97FF
        0x9909, # G+7716  U+9909  uni9909,  CJK UNIFIED IDEOGRAPH-9909
        0x9957, # G+7717  U+9957  uni9957,  CJK UNIFIED IDEOGRAPH-9957
        0x9999, # G+7718  U+9999  uni9999,  CJK UNIFIED IDEOGRAPH-9999
        0x5653, # G+7719  U+5653  uni5653,  CJK UNIFIED IDEOGRAPH-5653
        0x589F, # G+7720  U+589F  uni589F,  CJK UNIFIED IDEOGRAPH-589F
        0x865B, # G+7721  U+865B  uni865B,  CJK UNIFIED IDEOGRAPH-865B
        0x8A31, # G+7722  U+8A31  uni8A31,  CJK UNIFIED IDEOGRAPH-8A31
        0x61B2, # G+7723  U+61B2  uni61B2,  CJK UNIFIED IDEOGRAPH-61B2
        0x6AF6, # G+7724  U+6AF6  uni6AF6,  CJK UNIFIED IDEOGRAPH-6AF6
        0x737B, # G+7725  U+737B  uni737B,  CJK UNIFIED IDEOGRAPH-737B
        0x8ED2, # G+7726  U+8ED2  uni8ED2,  CJK UNIFIED IDEOGRAPH-8ED2
        0x6B47, # G+7727  U+6B47  uni6B47,  CJK UNIFIED IDEOGRAPH-6B47
        0x96AA, # G+7728  U+96AA  uni96AA,  CJK UNIFIED IDEOGRAPH-96AA
        0x9A57, # G+7729  U+9A57  uni9A57,  CJK UNIFIED IDEOGRAPH-9A57
        0x5955, # G+7730  U+5955  uni5955,  CJK UNIFIED IDEOGRAPH-5955
        0x7200, # G+7731  U+7200  uni7200,  CJK UNIFIED IDEOGRAPH-7200
        0x8D6B, # G+7732  U+8D6B  uni8D6B,  CJK UNIFIED IDEOGRAPH-8D6B
        0x9769, # G+7733  U+9769  uni9769,  CJK UNIFIED IDEOGRAPH-9769
        0x4FD4, # G+7734  U+4FD4  uni4FD4,  CJK UNIFIED IDEOGRAPH-4FD4
        0x5CF4, # G+7735  U+5CF4  uni5CF4,  CJK UNIFIED IDEOGRAPH-5CF4
        0x5F26, # G+7736  U+5F26  uni5F26,  CJK UNIFIED IDEOGRAPH-5F26
        0x61F8, # G+7737  U+61F8  uni61F8,  CJK UNIFIED IDEOGRAPH-61F8
        0x665B, # G+7738  U+665B  uni665B,  CJK UNIFIED IDEOGRAPH-665B
        0x6CEB, # G+7739  U+6CEB  uni6CEB,  CJK UNIFIED IDEOGRAPH-6CEB
        0x70AB, # G+7740  U+70AB  uni70AB,  CJK UNIFIED IDEOGRAPH-70AB
        0x7384, # G+7741  U+7384  uni7384,  CJK UNIFIED IDEOGRAPH-7384
        0x73B9, # G+7742  U+73B9  uni73B9,  CJK UNIFIED IDEOGRAPH-73B9
        0x73FE, # G+7743  U+73FE  uni73FE,  CJK UNIFIED IDEOGRAPH-73FE
        0x7729, # G+7744  U+7729  uni7729,  CJK UNIFIED IDEOGRAPH-7729
        0x774D, # G+7745  U+774D  uni774D,  CJK UNIFIED IDEOGRAPH-774D
        0x7D43, # G+7746  U+7D43  uni7D43,  CJK UNIFIED IDEOGRAPH-7D43
        0x7D62, # G+7747  U+7D62  uni7D62,  CJK UNIFIED IDEOGRAPH-7D62
        0x7E23, # G+7748  U+7E23  uni7E23,  CJK UNIFIED IDEOGRAPH-7E23
        0x8237, # G+7749  U+8237  uni8237,  CJK UNIFIED IDEOGRAPH-8237
        0x8852, # G+7750  U+8852  uni8852,  CJK UNIFIED IDEOGRAPH-8852
        0x8CE2, # G+7751  U+8CE2  uni8CE2,  CJK UNIFIED IDEOGRAPH-8CE2
        0x9249, # G+7752  U+9249  uni9249,  CJK UNIFIED IDEOGRAPH-9249
        0x986F, # G+7753  U+986F  uni986F,  CJK UNIFIED IDEOGRAPH-986F
        0x5B51, # G+7754  U+5B51  uni5B51,  CJK UNIFIED IDEOGRAPH-5B51
        0x7A74, # G+7755  U+7A74  uni7A74,  CJK UNIFIED IDEOGRAPH-7A74
        0x8840, # G+7756  U+8840  uni8840,  CJK UNIFIED IDEOGRAPH-8840
        0x9801, # G+7757  U+9801  uni9801,  CJK UNIFIED IDEOGRAPH-9801
        0x5ACC, # G+7758  U+5ACC  uni5ACC,  CJK UNIFIED IDEOGRAPH-5ACC
        0x4FE0, # G+7759  U+4FE0  uni4FE0,  CJK UNIFIED IDEOGRAPH-4FE0
        0x5354, # G+7760  U+5354  uni5354,  CJK UNIFIED IDEOGRAPH-5354
        0x593E, # G+7761  U+593E  uni593E,  CJK UNIFIED IDEOGRAPH-593E
        0x5CFD, # G+7762  U+5CFD  uni5CFD,  CJK UNIFIED IDEOGRAPH-5CFD
        0x633E, # G+7763  U+633E  uni633E,  CJK UNIFIED IDEOGRAPH-633E
        0x6D79, # G+7764  U+6D79  uni6D79,  CJK UNIFIED IDEOGRAPH-6D79
        0x72F9, # G+7765  U+72F9  uni72F9,  CJK UNIFIED IDEOGRAPH-72F9
        0x8105, # G+7766  U+8105  uni8105,  CJK UNIFIED IDEOGRAPH-8105
        0x8107, # G+7767  U+8107  uni8107,  CJK UNIFIED IDEOGRAPH-8107
        0x83A2, # G+7768  U+83A2  uni83A2,  CJK UNIFIED IDEOGRAPH-83A2
        0x92CF, # G+7769  U+92CF  uni92CF,  CJK UNIFIED IDEOGRAPH-92CF
        0x9830, # G+7770  U+9830  uni9830,  CJK UNIFIED IDEOGRAPH-9830
        0x4EA8, # G+7771  U+4EA8  uni4EA8,  CJK UNIFIED IDEOGRAPH-4EA8
        0x5144, # G+7772  U+5144  uni5144,  CJK UNIFIED IDEOGRAPH-5144
        0x5211, # G+7773  U+5211  uni5211,  CJK UNIFIED IDEOGRAPH-5211
        0x578B, # G+7774  U+578B  uni578B,  CJK UNIFIED IDEOGRAPH-578B
        0x5F62, # G+7775  U+5F62  uni5F62,  CJK UNIFIED IDEOGRAPH-5F62
        0x6CC2, # G+7776  U+6CC2  uni6CC2,  CJK UNIFIED IDEOGRAPH-6CC2
        0x6ECE, # G+7777  U+6ECE  uni6ECE,  CJK UNIFIED IDEOGRAPH-6ECE
        0x7005, # G+7778  U+7005  uni7005,  CJK UNIFIED IDEOGRAPH-7005
        0x7050, # G+7779  U+7050  uni7050,  CJK UNIFIED IDEOGRAPH-7050
        0x70AF, # G+7780  U+70AF  uni70AF,  CJK UNIFIED IDEOGRAPH-70AF
        0x7192, # G+7781  U+7192  uni7192,  CJK UNIFIED IDEOGRAPH-7192
        0x73E9, # G+7782  U+73E9  uni73E9,  CJK UNIFIED IDEOGRAPH-73E9
        0xF9AE, # G+7783  U+F9AE  uniF9AE,  CJK COMPATIBILITY IDEOGRAPH-F9AE
        0x834A, # G+7784  U+834A  uni834A,  CJK UNIFIED IDEOGRAPH-834A
        0x87A2, # G+7785  U+87A2  uni87A2,  CJK UNIFIED IDEOGRAPH-87A2
        0x8861, # G+7786  U+8861  uni8861,  CJK UNIFIED IDEOGRAPH-8861
        0x9008, # G+7787  U+9008  uni9008,  CJK UNIFIED IDEOGRAPH-9008
        0x90A2, # G+7788  U+90A2  uni90A2,  CJK UNIFIED IDEOGRAPH-90A2
        0x93A3, # G+7789  U+93A3  uni93A3,  CJK UNIFIED IDEOGRAPH-93A3
        0x99A8, # G+7790  U+99A8  uni99A8,  CJK UNIFIED IDEOGRAPH-99A8
        0x516E, # G+7791  U+516E  uni516E,  CJK UNIFIED IDEOGRAPH-516E
        0x5F57, # G+7792  U+5F57  uni5F57,  CJK UNIFIED IDEOGRAPH-5F57
        0x60E0, # G+7793  U+60E0  uni60E0,  CJK UNIFIED IDEOGRAPH-60E0
        0x6167, # G+7794  U+6167  uni6167,  CJK UNIFIED IDEOGRAPH-6167
        0x66B3, # G+7795  U+66B3  uni66B3,  CJK UNIFIED IDEOGRAPH-66B3
        0x8559, # G+7796  U+8559  uni8559,  CJK UNIFIED IDEOGRAPH-8559
        0x8E4A, # G+7797  U+8E4A  uni8E4A,  CJK UNIFIED IDEOGRAPH-8E4A
        0x91AF, # G+7798  U+91AF  uni91AF,  CJK UNIFIED IDEOGRAPH-91AF
        0x978B, # G+7799  U+978B  uni978B,  CJK UNIFIED IDEOGRAPH-978B
        0x4E4E, # G+7800  U+4E4E  uni4E4E,  CJK UNIFIED IDEOGRAPH-4E4E
        0x4E92, # G+7801  U+4E92  uni4E92,  CJK UNIFIED IDEOGRAPH-4E92
        0x547C, # G+7802  U+547C  uni547C,  CJK UNIFIED IDEOGRAPH-547C
        0x58D5, # G+7803  U+58D5  uni58D5,  CJK UNIFIED IDEOGRAPH-58D5
        0x58FA, # G+7804  U+58FA  uni58FA,  CJK UNIFIED IDEOGRAPH-58FA
        0x597D, # G+7805  U+597D  uni597D,  CJK UNIFIED IDEOGRAPH-597D
        0x5CB5, # G+7806  U+5CB5  uni5CB5,  CJK UNIFIED IDEOGRAPH-5CB5
        0x5F27, # G+7807  U+5F27  uni5F27,  CJK UNIFIED IDEOGRAPH-5F27
        0x6236, # G+7808  U+6236  uni6236,  CJK UNIFIED IDEOGRAPH-6236
        0x6248, # G+7809  U+6248  uni6248,  CJK UNIFIED IDEOGRAPH-6248
        0x660A, # G+7810  U+660A  uni660A,  CJK UNIFIED IDEOGRAPH-660A
        0x6667, # G+7811  U+6667  uni6667,  CJK UNIFIED IDEOGRAPH-6667
        0x6BEB, # G+7812  U+6BEB  uni6BEB,  CJK UNIFIED IDEOGRAPH-6BEB
        0x6D69, # G+7813  U+6D69  uni6D69,  CJK UNIFIED IDEOGRAPH-6D69
        0x6DCF, # G+7814  U+6DCF  uni6DCF,  CJK UNIFIED IDEOGRAPH-6DCF
        0x6E56, # G+7815  U+6E56  uni6E56,  CJK UNIFIED IDEOGRAPH-6E56
        0x6EF8, # G+7816  U+6EF8  uni6EF8,  CJK UNIFIED IDEOGRAPH-6EF8
        0x6F94, # G+7817  U+6F94  uni6F94,  CJK UNIFIED IDEOGRAPH-6F94
        0x6FE0, # G+7818  U+6FE0  uni6FE0,  CJK UNIFIED IDEOGRAPH-6FE0
        0x6FE9, # G+7819  U+6FE9  uni6FE9,  CJK UNIFIED IDEOGRAPH-6FE9
        0x705D, # G+7820  U+705D  uni705D,  CJK UNIFIED IDEOGRAPH-705D
        0x72D0, # G+7821  U+72D0  uni72D0,  CJK UNIFIED IDEOGRAPH-72D0
        0x7425, # G+7822  U+7425  uni7425,  CJK UNIFIED IDEOGRAPH-7425
        0x745A, # G+7823  U+745A  uni745A,  CJK UNIFIED IDEOGRAPH-745A
        0x74E0, # G+7824  U+74E0  uni74E0,  CJK UNIFIED IDEOGRAPH-74E0
        0x7693, # G+7825  U+7693  uni7693,  CJK UNIFIED IDEOGRAPH-7693
        0x795C, # G+7826  U+795C  uni795C,  CJK UNIFIED IDEOGRAPH-795C
        0x7CCA, # G+7827  U+7CCA  uni7CCA,  CJK UNIFIED IDEOGRAPH-7CCA
        0x7E1E, # G+7828  U+7E1E  uni7E1E,  CJK UNIFIED IDEOGRAPH-7E1E
        0x80E1, # G+7829  U+80E1  uni80E1,  CJK UNIFIED IDEOGRAPH-80E1
        0x82A6, # G+7830  U+82A6  uni82A6,  CJK UNIFIED IDEOGRAPH-82A6
        0x846B, # G+7831  U+846B  uni846B,  CJK UNIFIED IDEOGRAPH-846B
        0x84BF, # G+7832  U+84BF  uni84BF,  CJK UNIFIED IDEOGRAPH-84BF
        0x864E, # G+7833  U+864E  uni864E,  CJK UNIFIED IDEOGRAPH-864E
        0x865F, # G+7834  U+865F  uni865F,  CJK UNIFIED IDEOGRAPH-865F
        0x8774, # G+7835  U+8774  uni8774,  CJK UNIFIED IDEOGRAPH-8774
        0x8B77, # G+7836  U+8B77  uni8B77,  CJK UNIFIED IDEOGRAPH-8B77
        0x8C6A, # G+7837  U+8C6A  uni8C6A,  CJK UNIFIED IDEOGRAPH-8C6A
        0x93AC, # G+7838  U+93AC  uni93AC,  CJK UNIFIED IDEOGRAPH-93AC
        0x9800, # G+7839  U+9800  uni9800,  CJK UNIFIED IDEOGRAPH-9800
        0x9865, # G+7840  U+9865  uni9865,  CJK UNIFIED IDEOGRAPH-9865
        0x60D1, # G+7841  U+60D1  uni60D1,  CJK UNIFIED IDEOGRAPH-60D1
        0x6216, # G+7842  U+6216  uni6216,  CJK UNIFIED IDEOGRAPH-6216
        0x9177, # G+7843  U+9177  uni9177,  CJK UNIFIED IDEOGRAPH-9177
        0x5A5A, # G+7844  U+5A5A  uni5A5A,  CJK UNIFIED IDEOGRAPH-5A5A
        0x660F, # G+7845  U+660F  uni660F,  CJK UNIFIED IDEOGRAPH-660F
        0x6DF7, # G+7846  U+6DF7  uni6DF7,  CJK UNIFIED IDEOGRAPH-6DF7
        0x6E3E, # G+7847  U+6E3E  uni6E3E,  CJK UNIFIED IDEOGRAPH-6E3E
        0x743F, # G+7848  U+743F  uni743F,  CJK UNIFIED IDEOGRAPH-743F
        0x9B42, # G+7849  U+9B42  uni9B42,  CJK UNIFIED IDEOGRAPH-9B42
        0x5FFD, # G+7850  U+5FFD  uni5FFD,  CJK UNIFIED IDEOGRAPH-5FFD
        0x60DA, # G+7851  U+60DA  uni60DA,  CJK UNIFIED IDEOGRAPH-60DA
        0x7B0F, # G+7852  U+7B0F  uni7B0F,  CJK UNIFIED IDEOGRAPH-7B0F
        0x54C4, # G+7853  U+54C4  uni54C4,  CJK UNIFIED IDEOGRAPH-54C4
        0x5F18, # G+7854  U+5F18  uni5F18,  CJK UNIFIED IDEOGRAPH-5F18
        0x6C5E, # G+7855  U+6C5E  uni6C5E,  CJK UNIFIED IDEOGRAPH-6C5E
        0x6CD3, # G+7856  U+6CD3  uni6CD3,  CJK UNIFIED IDEOGRAPH-6CD3
        0x6D2A, # G+7857  U+6D2A  uni6D2A,  CJK UNIFIED IDEOGRAPH-6D2A
        0x70D8, # G+7858  U+70D8  uni70D8,  CJK UNIFIED IDEOGRAPH-70D8
        0x7D05, # G+7859  U+7D05  uni7D05,  CJK UNIFIED IDEOGRAPH-7D05
        0x8679, # G+7860  U+8679  uni8679,  CJK UNIFIED IDEOGRAPH-8679
        0x8A0C, # G+7861  U+8A0C  uni8A0C,  CJK UNIFIED IDEOGRAPH-8A0C
        0x9D3B, # G+7862  U+9D3B  uni9D3B,  CJK UNIFIED IDEOGRAPH-9D3B
        0x5316, # G+7863  U+5316  uni5316,  CJK UNIFIED IDEOGRAPH-5316
        0x548C, # G+7864  U+548C  uni548C,  CJK UNIFIED IDEOGRAPH-548C
        0x5B05, # G+7865  U+5B05  uni5B05,  CJK UNIFIED IDEOGRAPH-5B05
        0x6A3A, # G+7866  U+6A3A  uni6A3A,  CJK UNIFIED IDEOGRAPH-6A3A
        0x706B, # G+7867  U+706B  uni706B,  CJK UNIFIED IDEOGRAPH-706B
        0x7575, # G+7868  U+7575  uni7575,  CJK UNIFIED IDEOGRAPH-7575
        0x798D, # G+7869  U+798D  uni798D,  CJK UNIFIED IDEOGRAPH-798D
        0x79BE, # G+7870  U+79BE  uni79BE,  CJK UNIFIED IDEOGRAPH-79BE
        0x82B1, # G+7871  U+82B1  uni82B1,  CJK UNIFIED IDEOGRAPH-82B1
        0x83EF, # G+7872  U+83EF  uni83EF,  CJK UNIFIED IDEOGRAPH-83EF
        0x8A71, # G+7873  U+8A71  uni8A71,  CJK UNIFIED IDEOGRAPH-8A71
        0x8B41, # G+7874  U+8B41  uni8B41,  CJK UNIFIED IDEOGRAPH-8B41
        0x8CA8, # G+7875  U+8CA8  uni8CA8,  CJK UNIFIED IDEOGRAPH-8CA8
        0x9774, # G+7876  U+9774  uni9774,  CJK UNIFIED IDEOGRAPH-9774
        0x64F4, # G+7877  U+64F4  uni64F4,  CJK UNIFIED IDEOGRAPH-64F4
        0x652B, # G+7878  U+652B  uni652B,  CJK UNIFIED IDEOGRAPH-652B
        0x78BA, # G+7879  U+78BA  uni78BA,  CJK UNIFIED IDEOGRAPH-78BA
        0x78BB, # G+7880  U+78BB  uni78BB,  CJK UNIFIED IDEOGRAPH-78BB
        0x7A6B, # G+7881  U+7A6B  uni7A6B,  CJK UNIFIED IDEOGRAPH-7A6B
        0x4E38, # G+7882  U+4E38  uni4E38,  CJK UNIFIED IDEOGRAPH-4E38
        0x559A, # G+7883  U+559A  uni559A,  CJK UNIFIED IDEOGRAPH-559A
        0x5950, # G+7884  U+5950  uni5950,  CJK UNIFIED IDEOGRAPH-5950
        0x5BA6, # G+7885  U+5BA6  uni5BA6,  CJK UNIFIED IDEOGRAPH-5BA6
        0x5E7B, # G+7886  U+5E7B  uni5E7B,  CJK UNIFIED IDEOGRAPH-5E7B
        0x60A3, # G+7887  U+60A3  uni60A3,  CJK UNIFIED IDEOGRAPH-60A3
        0x63DB, # G+7888  U+63DB  uni63DB,  CJK UNIFIED IDEOGRAPH-63DB
        0x6B61, # G+7889  U+6B61  uni6B61,  CJK UNIFIED IDEOGRAPH-6B61
        0x6665, # G+7890  U+6665  uni6665,  CJK UNIFIED IDEOGRAPH-6665
        0x6853, # G+7891  U+6853  uni6853,  CJK UNIFIED IDEOGRAPH-6853
        0x6E19, # G+7892  U+6E19  uni6E19,  CJK UNIFIED IDEOGRAPH-6E19
        0x7165, # G+7893  U+7165  uni7165,  CJK UNIFIED IDEOGRAPH-7165
        0x74B0, # G+7894  U+74B0  uni74B0,  CJK UNIFIED IDEOGRAPH-74B0
        0x7D08, # G+7895  U+7D08  uni7D08,  CJK UNIFIED IDEOGRAPH-7D08
        0x9084, # G+7896  U+9084  uni9084,  CJK UNIFIED IDEOGRAPH-9084
        0x9A69, # G+7897  U+9A69  uni9A69,  CJK UNIFIED IDEOGRAPH-9A69
        0x9C25, # G+7898  U+9C25  uni9C25,  CJK UNIFIED IDEOGRAPH-9C25
        0x6D3B, # G+7899  U+6D3B  uni6D3B,  CJK UNIFIED IDEOGRAPH-6D3B
        0xF904, # G+7900  U+F904  uniF904,  CJK COMPATIBILITY IDEOGRAPH-F904
        0x733E, # G+7901  U+733E  uni733E,  CJK UNIFIED IDEOGRAPH-733E
        0x8C41, # G+7902  U+8C41  uni8C41,  CJK UNIFIED IDEOGRAPH-8C41
        0x95CA, # G+7903  U+95CA  uni95CA,  CJK UNIFIED IDEOGRAPH-95CA
        0x51F0, # G+7904  U+51F0  uni51F0,  CJK UNIFIED IDEOGRAPH-51F0
        0x5E4C, # G+7905  U+5E4C  uni5E4C,  CJK UNIFIED IDEOGRAPH-5E4C
        0x5FA8, # G+7906  U+5FA8  uni5FA8,  CJK UNIFIED IDEOGRAPH-5FA8
        0x604D, # G+7907  U+604D  uni604D,  CJK UNIFIED IDEOGRAPH-604D
        0x60F6, # G+7908  U+60F6  uni60F6,  CJK UNIFIED IDEOGRAPH-60F6
        0x6130, # G+7909  U+6130  uni6130,  CJK UNIFIED IDEOGRAPH-6130
        0x614C, # G+7910  U+614C  uni614C,  CJK UNIFIED IDEOGRAPH-614C
        0x6643, # G+7911  U+6643  uni6643,  CJK UNIFIED IDEOGRAPH-6643
        0x6644, # G+7912  U+6644  uni6644,  CJK UNIFIED IDEOGRAPH-6644
        0x69A5, # G+7913  U+69A5  uni69A5,  CJK UNIFIED IDEOGRAPH-69A5
        0x6CC1, # G+7914  U+6CC1  uni6CC1,  CJK UNIFIED IDEOGRAPH-6CC1
        0x6E5F, # G+7915  U+6E5F  uni6E5F,  CJK UNIFIED IDEOGRAPH-6E5F
        0x6EC9, # G+7916  U+6EC9  uni6EC9,  CJK UNIFIED IDEOGRAPH-6EC9
        0x6F62, # G+7917  U+6F62  uni6F62,  CJK UNIFIED IDEOGRAPH-6F62
        0x714C, # G+7918  U+714C  uni714C,  CJK UNIFIED IDEOGRAPH-714C
        0x749C, # G+7919  U+749C  uni749C,  CJK UNIFIED IDEOGRAPH-749C
        0x7687, # G+7920  U+7687  uni7687,  CJK UNIFIED IDEOGRAPH-7687
        0x7BC1, # G+7921  U+7BC1  uni7BC1,  CJK UNIFIED IDEOGRAPH-7BC1
        0x7C27, # G+7922  U+7C27  uni7C27,  CJK UNIFIED IDEOGRAPH-7C27
        0x8352, # G+7923  U+8352  uni8352,  CJK UNIFIED IDEOGRAPH-8352
        0x8757, # G+7924  U+8757  uni8757,  CJK UNIFIED IDEOGRAPH-8757
        0x9051, # G+7925  U+9051  uni9051,  CJK UNIFIED IDEOGRAPH-9051
        0x968D, # G+7926  U+968D  uni968D,  CJK UNIFIED IDEOGRAPH-968D
        0x9EC3, # G+7927  U+9EC3  uni9EC3,  CJK UNIFIED IDEOGRAPH-9EC3
        0x532F, # G+7928  U+532F  uni532F,  CJK UNIFIED IDEOGRAPH-532F
        0x56DE, # G+7929  U+56DE  uni56DE,  CJK UNIFIED IDEOGRAPH-56DE
        0x5EFB, # G+7930  U+5EFB  uni5EFB,  CJK UNIFIED IDEOGRAPH-5EFB
        0x5F8A, # G+7931  U+5F8A  uni5F8A,  CJK UNIFIED IDEOGRAPH-5F8A
        0x6062, # G+7932  U+6062  uni6062,  CJK UNIFIED IDEOGRAPH-6062
        0x6094, # G+7933  U+6094  uni6094,  CJK UNIFIED IDEOGRAPH-6094
        0x61F7, # G+7934  U+61F7  uni61F7,  CJK UNIFIED IDEOGRAPH-61F7
        0x6666, # G+7935  U+6666  uni6666,  CJK UNIFIED IDEOGRAPH-6666
        0x6703, # G+7936  U+6703  uni6703,  CJK UNIFIED IDEOGRAPH-6703
        0x6A9C, # G+7937  U+6A9C  uni6A9C,  CJK UNIFIED IDEOGRAPH-6A9C
        0x6DEE, # G+7938  U+6DEE  uni6DEE,  CJK UNIFIED IDEOGRAPH-6DEE
        0x6FAE, # G+7939  U+6FAE  uni6FAE,  CJK UNIFIED IDEOGRAPH-6FAE
        0x7070, # G+7940  U+7070  uni7070,  CJK UNIFIED IDEOGRAPH-7070
        0x736A, # G+7941  U+736A  uni736A,  CJK UNIFIED IDEOGRAPH-736A
        0x7E6A, # G+7942  U+7E6A  uni7E6A,  CJK UNIFIED IDEOGRAPH-7E6A
        0x81BE, # G+7943  U+81BE  uni81BE,  CJK UNIFIED IDEOGRAPH-81BE
        0x8334, # G+7944  U+8334  uni8334,  CJK UNIFIED IDEOGRAPH-8334
        0x86D4, # G+7945  U+86D4  uni86D4,  CJK UNIFIED IDEOGRAPH-86D4
        0x8AA8, # G+7946  U+8AA8  uni8AA8,  CJK UNIFIED IDEOGRAPH-8AA8
        0x8CC4, # G+7947  U+8CC4  uni8CC4,  CJK UNIFIED IDEOGRAPH-8CC4
        0x5283, # G+7948  U+5283  uni5283,  CJK UNIFIED IDEOGRAPH-5283
        0x7372, # G+7949  U+7372  uni7372,  CJK UNIFIED IDEOGRAPH-7372
        0x5B96, # G+7950  U+5B96  uni5B96,  CJK UNIFIED IDEOGRAPH-5B96
        0x6A6B, # G+7951  U+6A6B  uni6A6B,  CJK UNIFIED IDEOGRAPH-6A6B
        0x9404, # G+7952  U+9404  uni9404,  CJK UNIFIED IDEOGRAPH-9404
        0x54EE, # G+7953  U+54EE  uni54EE,  CJK UNIFIED IDEOGRAPH-54EE
        0x5686, # G+7954  U+5686  uni5686,  CJK UNIFIED IDEOGRAPH-5686
        0x5B5D, # G+7955  U+5B5D  uni5B5D,  CJK UNIFIED IDEOGRAPH-5B5D
        0x6548, # G+7956  U+6548  uni6548,  CJK UNIFIED IDEOGRAPH-6548
        0x6585, # G+7957  U+6585  uni6585,  CJK UNIFIED IDEOGRAPH-6585
        0x66C9, # G+7958  U+66C9  uni66C9,  CJK UNIFIED IDEOGRAPH-66C9
        0x689F, # G+7959  U+689F  uni689F,  CJK UNIFIED IDEOGRAPH-689F
        0x6D8D, # G+7960  U+6D8D  uni6D8D,  CJK UNIFIED IDEOGRAPH-6D8D
        0x6DC6, # G+7961  U+6DC6  uni6DC6,  CJK UNIFIED IDEOGRAPH-6DC6
        0x723B, # G+7962  U+723B  uni723B,  CJK UNIFIED IDEOGRAPH-723B
        0x80B4, # G+7963  U+80B4  uni80B4,  CJK UNIFIED IDEOGRAPH-80B4
        0x9175, # G+7964  U+9175  uni9175,  CJK UNIFIED IDEOGRAPH-9175
        0x9A4D, # G+7965  U+9A4D  uni9A4D,  CJK UNIFIED IDEOGRAPH-9A4D
        0x4FAF, # G+7966  U+4FAF  uni4FAF,  CJK UNIFIED IDEOGRAPH-4FAF
        0x5019, # G+7967  U+5019  uni5019,  CJK UNIFIED IDEOGRAPH-5019
        0x539A, # G+7968  U+539A  uni539A,  CJK UNIFIED IDEOGRAPH-539A
        0x540E, # G+7969  U+540E  uni540E,  CJK UNIFIED IDEOGRAPH-540E
        0x543C, # G+7970  U+543C  uni543C,  CJK UNIFIED IDEOGRAPH-543C
        0x5589, # G+7971  U+5589  uni5589,  CJK UNIFIED IDEOGRAPH-5589
        0x55C5, # G+7972  U+55C5  uni55C5,  CJK UNIFIED IDEOGRAPH-55C5
        0x5E3F, # G+7973  U+5E3F  uni5E3F,  CJK UNIFIED IDEOGRAPH-5E3F
        0x5F8C, # G+7974  U+5F8C  uni5F8C,  CJK UNIFIED IDEOGRAPH-5F8C
        0x673D, # G+7975  U+673D  uni673D,  CJK UNIFIED IDEOGRAPH-673D
        0x7166, # G+7976  U+7166  uni7166,  CJK UNIFIED IDEOGRAPH-7166
        0x73DD, # G+7977  U+73DD  uni73DD,  CJK UNIFIED IDEOGRAPH-73DD
        0x9005, # G+7978  U+9005  uni9005,  CJK UNIFIED IDEOGRAPH-9005
        0x52DB, # G+7979  U+52DB  uni52DB,  CJK UNIFIED IDEOGRAPH-52DB
        0x52F3, # G+7980  U+52F3  uni52F3,  CJK UNIFIED IDEOGRAPH-52F3
        0x5864, # G+7981  U+5864  uni5864,  CJK UNIFIED IDEOGRAPH-5864
        0x58CE, # G+7982  U+58CE  uni58CE,  CJK UNIFIED IDEOGRAPH-58CE
        0x7104, # G+7983  U+7104  uni7104,  CJK UNIFIED IDEOGRAPH-7104
        0x718F, # G+7984  U+718F  uni718F,  CJK UNIFIED IDEOGRAPH-718F
        0x71FB, # G+7985  U+71FB  uni71FB,  CJK UNIFIED IDEOGRAPH-71FB
        0x85B0, # G+7986  U+85B0  uni85B0,  CJK UNIFIED IDEOGRAPH-85B0
        0x8A13, # G+7987  U+8A13  uni8A13,  CJK UNIFIED IDEOGRAPH-8A13
        0xF9C5, # G+7988  U+F9C5  uniF9C5,  CJK COMPATIBILITY IDEOGRAPH-F9C5
        0x85A8, # G+7989  U+85A8  uni85A8,  CJK UNIFIED IDEOGRAPH-85A8
        0x55A7, # G+7990  U+55A7  uni55A7,  CJK UNIFIED IDEOGRAPH-55A7
        0x6684, # G+7991  U+6684  uni6684,  CJK UNIFIED IDEOGRAPH-6684
        0x714A, # G+7992  U+714A  uni714A,  CJK UNIFIED IDEOGRAPH-714A
        0x8431, # G+7993  U+8431  uni8431,  CJK UNIFIED IDEOGRAPH-8431
        0x5349, # G+7994  U+5349  uni5349,  CJK UNIFIED IDEOGRAPH-5349
        0x5599, # G+7995  U+5599  uni5599,  CJK UNIFIED IDEOGRAPH-5599
        0x6BC1, # G+7996  U+6BC1  uni6BC1,  CJK UNIFIED IDEOGRAPH-6BC1
        0x5F59, # G+7997  U+5F59  uni5F59,  CJK UNIFIED IDEOGRAPH-5F59
        0x5FBD, # G+7998  U+5FBD  uni5FBD,  CJK UNIFIED IDEOGRAPH-5FBD
        0x63EE, # G+7999  U+63EE  uni63EE,  CJK UNIFIED IDEOGRAPH-63EE
        0x6689, # G+8000  U+6689  uni6689,  CJK UNIFIED IDEOGRAPH-6689
        0x7147, # G+8001  U+7147  uni7147,  CJK UNIFIED IDEOGRAPH-7147
        0x8AF1, # G+8002  U+8AF1  uni8AF1,  CJK UNIFIED IDEOGRAPH-8AF1
        0x8F1D, # G+8003  U+8F1D  uni8F1D,  CJK UNIFIED IDEOGRAPH-8F1D
        0x9EBE, # G+8004  U+9EBE  uni9EBE,  CJK UNIFIED IDEOGRAPH-9EBE
        0x4F11, # G+8005  U+4F11  uni4F11,  CJK UNIFIED IDEOGRAPH-4F11
        0x643A, # G+8006  U+643A  uni643A,  CJK UNIFIED IDEOGRAPH-643A
        0x70CB, # G+8007  U+70CB  uni70CB,  CJK UNIFIED IDEOGRAPH-70CB
        0x7566, # G+8008  U+7566  uni7566,  CJK UNIFIED IDEOGRAPH-7566
        0x8667, # G+8009  U+8667  uni8667,  CJK UNIFIED IDEOGRAPH-8667
        0x6064, # G+8010  U+6064  uni6064,  CJK UNIFIED IDEOGRAPH-6064
        0x8B4E, # G+8011  U+8B4E  uni8B4E,  CJK UNIFIED IDEOGRAPH-8B4E
        0x9DF8, # G+8012  U+9DF8  uni9DF8,  CJK UNIFIED IDEOGRAPH-9DF8
        0x5147, # G+8013  U+5147  uni5147,  CJK UNIFIED IDEOGRAPH-5147
        0x51F6, # G+8014  U+51F6  uni51F6,  CJK UNIFIED IDEOGRAPH-51F6
        0x5308, # G+8015  U+5308  uni5308,  CJK UNIFIED IDEOGRAPH-5308
        0x6D36, # G+8016  U+6D36  uni6D36,  CJK UNIFIED IDEOGRAPH-6D36
        0x80F8, # G+8017  U+80F8  uni80F8,  CJK UNIFIED IDEOGRAPH-80F8
        0x9ED1, # G+8018  U+9ED1  uni9ED1,  CJK UNIFIED IDEOGRAPH-9ED1
        0x6615, # G+8019  U+6615  uni6615,  CJK UNIFIED IDEOGRAPH-6615
        0x6B23, # G+8020  U+6B23  uni6B23,  CJK UNIFIED IDEOGRAPH-6B23
        0x7098, # G+8021  U+7098  uni7098,  CJK UNIFIED IDEOGRAPH-7098
        0x75D5, # G+8022  U+75D5  uni75D5,  CJK UNIFIED IDEOGRAPH-75D5
        0x5403, # G+8023  U+5403  uni5403,  CJK UNIFIED IDEOGRAPH-5403
        0x5C79, # G+8024  U+5C79  uni5C79,  CJK UNIFIED IDEOGRAPH-5C79
        0x7D07, # G+8025  U+7D07  uni7D07,  CJK UNIFIED IDEOGRAPH-7D07
        0x8A16, # G+8026  U+8A16  uni8A16,  CJK UNIFIED IDEOGRAPH-8A16
        0x6B20, # G+8027  U+6B20  uni6B20,  CJK UNIFIED IDEOGRAPH-6B20
        0x6B3D, # G+8028  U+6B3D  uni6B3D,  CJK UNIFIED IDEOGRAPH-6B3D
        0x6B46, # G+8029  U+6B46  uni6B46,  CJK UNIFIED IDEOGRAPH-6B46
        0x5438, # G+8030  U+5438  uni5438,  CJK UNIFIED IDEOGRAPH-5438
        0x6070, # G+8031  U+6070  uni6070,  CJK UNIFIED IDEOGRAPH-6070
        0x6D3D, # G+8032  U+6D3D  uni6D3D,  CJK UNIFIED IDEOGRAPH-6D3D
        0x7FD5, # G+8033  U+7FD5  uni7FD5,  CJK UNIFIED IDEOGRAPH-7FD5
        0x8208, # G+8034  U+8208  uni8208,  CJK UNIFIED IDEOGRAPH-8208
        0x50D6, # G+8035  U+50D6  uni50D6,  CJK UNIFIED IDEOGRAPH-50D6
        0x51DE, # G+8036  U+51DE  uni51DE,  CJK UNIFIED IDEOGRAPH-51DE
        0x559C, # G+8037  U+559C  uni559C,  CJK UNIFIED IDEOGRAPH-559C
        0x566B, # G+8038  U+566B  uni566B,  CJK UNIFIED IDEOGRAPH-566B
        0x56CD, # G+8039  U+56CD  uni56CD,  CJK UNIFIED IDEOGRAPH-56CD
        0x59EC, # G+8040  U+59EC  uni59EC,  CJK UNIFIED IDEOGRAPH-59EC
        0x5B09, # G+8041  U+5B09  uni5B09,  CJK UNIFIED IDEOGRAPH-5B09
        0x5E0C, # G+8042  U+5E0C  uni5E0C,  CJK UNIFIED IDEOGRAPH-5E0C
        0x6199, # G+8043  U+6199  uni6199,  CJK UNIFIED IDEOGRAPH-6199
        0x6198, # G+8044  U+6198  uni6198,  CJK UNIFIED IDEOGRAPH-6198
        0x6231, # G+8045  U+6231  uni6231,  CJK UNIFIED IDEOGRAPH-6231
        0x665E, # G+8046  U+665E  uni665E,  CJK UNIFIED IDEOGRAPH-665E
        0x66E6, # G+8047  U+66E6  uni66E6,  CJK UNIFIED IDEOGRAPH-66E6
        0x7199, # G+8048  U+7199  uni7199,  CJK UNIFIED IDEOGRAPH-7199
        0x71B9, # G+8049  U+71B9  uni71B9,  CJK UNIFIED IDEOGRAPH-71B9
        0x71BA, # G+8050  U+71BA  uni71BA,  CJK UNIFIED IDEOGRAPH-71BA
        0x72A7, # G+8051  U+72A7  uni72A7,  CJK UNIFIED IDEOGRAPH-72A7
        0x79A7, # G+8052  U+79A7  uni79A7,  CJK UNIFIED IDEOGRAPH-79A7
        0x7A00, # G+8053  U+7A00  uni7A00,  CJK UNIFIED IDEOGRAPH-7A00
        0x7FB2, # G+8054  U+7FB2  uni7FB2,  CJK UNIFIED IDEOGRAPH-7FB2
        0x8A70, # G+8055  U+8A70  uni8A70,  CJK UNIFIED IDEOGRAPH-8A70
        0x0000, # G+8056  U+0000  controlNUL,  <control>
        0x0000, # G+8057  U+0000  controlNUL,  <control>
        0x0000, # G+8058  U+0000  controlNUL,  <control>
        0x0000, # G+8059  U+0000  controlNUL,  <control>
        0x0000, # G+8060  U+0000  controlNUL,  <control>
        0x0000, # G+8061  U+0000  controlNUL,  <control>
        0x0000, # G+8062  U+0000  controlNUL,  <control>
        0x0000, # G+8063  U+0000  controlNUL,  <control>
        0x0000, # G+8064  U+0000  controlNUL,  <control>
        0x0000, # G+8065  U+0000  controlNUL,  <control>
        0x0000, # G+8066  U+0000  controlNUL,  <control>
        0x0000, # G+8067  U+0000  controlNUL,  <control>
        0x0000, # G+8068  U+0000  controlNUL,  <control>
        0x0000, # G+8069  U+0000  controlNUL,  <control>
        0x0000, # G+8070  U+0000  controlNUL,  <control>
        0x0000, # G+8071  U+0000  controlNUL,  <control>
        0x0000, # G+8072  U+0000  controlNUL,  <control>
        0x0000, # G+8073  U+0000  controlNUL,  <control>
        0x0000, # G+8074  U+0000  controlNUL,  <control>
        0x0000, # G+8075  U+0000  controlNUL,  <control>
        0x0000, # G+8076  U+0000  controlNUL,  <control>
        0x0000, # G+8077  U+0000  controlNUL,  <control>
        0x0000, # G+8078  U+0000  controlNUL,  <control>
        0x0000, # G+8079  U+0000  controlNUL,  <control>
        0x0000, # G+8080  U+0000  controlNUL,  <control>
        0x0000, # G+8081  U+0000  controlNUL,  <control>
        0x0000, # G+8082  U+0000  controlNUL,  <control>
        0x0000, # G+8083  U+0000  controlNUL,  <control>
        0x0000, # G+8084  U+0000  controlNUL,  <control>
        0x0000, # G+8085  U+0000  controlNUL,  <control>
        0x0000, # G+8086  U+0000  controlNUL,  <control>
        0x0000, # G+8087  U+0000  controlNUL,  <control>
        0x0000, # G+8088  U+0000  controlNUL,  <control>
        0x0000, # G+8089  U+0000  controlNUL,  <control>
        0x0000, # G+8090  U+0000  controlNUL,  <control>
        0x0000, # G+8091  U+0000  controlNUL,  <control>
        0x0000, # G+8092  U+0000  controlNUL,  <control>
        0x0000, # G+8093  U+0000  controlNUL,  <control>
        0x0020, # G+8094  U+0020  space,  SPACE
        0x0021, # G+8095  U+0021  exclam,  EXCLAMATION MARK
        0x0022, # G+8096  U+0022  quotedbl,  QUOTATION MARK
        0x0023, # G+8097  U+0023  numbersign,  NUMBER SIGN
        0x0024, # G+8098  U+0024  dollar,  DOLLAR SIGN
        0x0025, # G+8099  U+0025  percent,  PERCENT SIGN
        0x0026, # G+8100  U+0026  ampersand,  AMPERSAND
        0x0027, # G+8101  U+0027  quotesingle,  APOSTROPHE
        0x0028, # G+8102  U+0028  parenleft,  LEFT PARENTHESIS
        0x0029, # G+8103  U+0029  parenright,  RIGHT PARENTHESIS
        0x002A, # G+8104  U+002A  asterisk,  ASTERISK
        0x002B, # G+8105  U+002B  plus,  PLUS SIGN
        0x002C, # G+8106  U+002C  comma,  COMMA
        0x002D, # G+8107  U+002D  hyphen,  HYPHEN-MINUS
        0x002E, # G+8108  U+002E  period,  FULL STOP
        0x002F, # G+8109  U+002F  slash,  SOLIDUS
        0x0030, # G+8110  U+0030  zero,  DIGIT ZERO
        0x0031, # G+8111  U+0031  one,  DIGIT ONE
        0x0032, # G+8112  U+0032  two,  DIGIT TWO
        0x0033, # G+8113  U+0033  three,  DIGIT THREE
        0x0034, # G+8114  U+0034  four,  DIGIT FOUR
        0x0035, # G+8115  U+0035  five,  DIGIT FIVE
        0x0036, # G+8116  U+0036  six,  DIGIT SIX
        0x0037, # G+8117  U+0037  seven,  DIGIT SEVEN
        0x0038, # G+8118  U+0038  eight,  DIGIT EIGHT
        0x0039, # G+8119  U+0039  nine,  DIGIT NINE
        0x003A, # G+8120  U+003A  colon,  COLON
        0x003B, # G+8121  U+003B  semicolon,  SEMICOLON
        0x003C, # G+8122  U+003C  less,  LESS-THAN SIGN
        0x003D, # G+8123  U+003D  equal,  EQUALS SIGN
        0x003E, # G+8124  U+003E  greater,  GREATER-THAN SIGN
        0x003F, # G+8125  U+003F  question,  QUESTION MARK
        0x0040, # G+8126  U+0040  at,  COMMERCIAL AT
        0x0041, # G+8127  U+0041  A,  LATIN CAPITAL LETTER A
        0x0042, # G+8128  U+0042  B,  LATIN CAPITAL LETTER B
        0x0043, # G+8129  U+0043  C,  LATIN CAPITAL LETTER C
        0x0044, # G+8130  U+0044  D,  LATIN CAPITAL LETTER D
        0x0045, # G+8131  U+0045  E,  LATIN CAPITAL LETTER E
        0x0046, # G+8132  U+0046  F,  LATIN CAPITAL LETTER F
        0x0047, # G+8133  U+0047  G,  LATIN CAPITAL LETTER G
        0x0048, # G+8134  U+0048  H,  LATIN CAPITAL LETTER H
        0x0049, # G+8135  U+0049  I,  LATIN CAPITAL LETTER I
        0x004A, # G+8136  U+004A  J,  LATIN CAPITAL LETTER J
        0x004B, # G+8137  U+004B  K,  LATIN CAPITAL LETTER K
        0x004C, # G+8138  U+004C  L,  LATIN CAPITAL LETTER L
        0x004D, # G+8139  U+004D  M,  LATIN CAPITAL LETTER M
        0x004E, # G+8140  U+004E  N,  LATIN CAPITAL LETTER N
        0x004F, # G+8141  U+004F  O,  LATIN CAPITAL LETTER O
        0x0050, # G+8142  U+0050  P,  LATIN CAPITAL LETTER P
        0x0051, # G+8143  U+0051  Q,  LATIN CAPITAL LETTER Q
        0x0052, # G+8144  U+0052  R,  LATIN CAPITAL LETTER R
        0x0053, # G+8145  U+0053  S,  LATIN CAPITAL LETTER S
        0x0054, # G+8146  U+0054  T,  LATIN CAPITAL LETTER T
        0x0055, # G+8147  U+0055  U,  LATIN CAPITAL LETTER U
        0x0056, # G+8148  U+0056  V,  LATIN CAPITAL LETTER V
        0x0057, # G+8149  U+0057  W,  LATIN CAPITAL LETTER W
        0x0058, # G+8150  U+0058  X,  LATIN CAPITAL LETTER X
        0x0059, # G+8151  U+0059  Y,  LATIN CAPITAL LETTER Y
        0x005A, # G+8152  U+005A  Z,  LATIN CAPITAL LETTER Z
        0x005B, # G+8153  U+005B  bracketleft,  LEFT SQUARE BRACKET
        0x005C, # G+8154  U+005C  backslash,  REVERSE SOLIDUS
        0x005D, # G+8155  U+005D  bracketright,  RIGHT SQUARE BRACKET
        0x005E, # G+8156  U+005E  asciicircum,  CIRCUMFLEX ACCENT
        0x005F, # G+8157  U+005F  underscore,  LOW LINE
        0x0060, # G+8158  U+0060  grave,  GRAVE ACCENT
        0x0061, # G+8159  U+0061  a,  LATIN SMALL LETTER A
        0x0062, # G+8160  U+0062  b,  LATIN SMALL LETTER B
        0x0063, # G+8161  U+0063  c,  LATIN SMALL LETTER C
        0x0064, # G+8162  U+0064  d,  LATIN SMALL LETTER D
        0x0065, # G+8163  U+0065  e,  LATIN SMALL LETTER E
        0x0066, # G+8164  U+0066  f,  LATIN SMALL LETTER F
        0x0067, # G+8165  U+0067  g,  LATIN SMALL LETTER G
        0x0068, # G+8166  U+0068  h,  LATIN SMALL LETTER H
        0x0069, # G+8167  U+0069  i,  LATIN SMALL LETTER I
        0x006A, # G+8168  U+006A  j,  LATIN SMALL LETTER J
        0x006B, # G+8169  U+006B  k,  LATIN SMALL LETTER K
        0x006C, # G+8170  U+006C  l,  LATIN SMALL LETTER L
        0x006D, # G+8171  U+006D  m,  LATIN SMALL LETTER M
        0x006E, # G+8172  U+006E  n,  LATIN SMALL LETTER N
        0x006F, # G+8173  U+006F  o,  LATIN SMALL LETTER O
        0x0070, # G+8174  U+0070  p,  LATIN SMALL LETTER P
        0x0071, # G+8175  U+0071  q,  LATIN SMALL LETTER Q
        0x0072, # G+8176  U+0072  r,  LATIN SMALL LETTER R
        0x0073, # G+8177  U+0073  s,  LATIN SMALL LETTER S
        0x0074, # G+8178  U+0074  t,  LATIN SMALL LETTER T
        0x0075, # G+8179  U+0075  u,  LATIN SMALL LETTER U
        0x0076, # G+8180  U+0076  v,  LATIN SMALL LETTER V
        0x0077, # G+8181  U+0077  w,  LATIN SMALL LETTER W
        0x0078, # G+8182  U+0078  x,  LATIN SMALL LETTER X
        0x0079, # G+8183  U+0079  y,  LATIN SMALL LETTER Y
        0x007A, # G+8184  U+007A  z,  LATIN SMALL LETTER Z
        0x007B, # G+8185  U+007B  braceleft,  LEFT CURLY BRACKET
        0x007C, # G+8186  U+007C  bar,  VERTICAL LINE
        0x007D, # G+8187  U+007D  braceright,  RIGHT CURLY BRACKET
        0x007E, # G+8188  U+007E  asciitilde,  TILDE
        0x0000, # G+8189  U+0000  controlNUL,  <control>
        0x0000, # G+8190  U+0000  controlNUL,  <control>
        0x0000, # G+8191  U+0000  controlNUL,  <control>
        0x0000, # G+8192  U+0000  controlNUL,  <control>
        0x0000, # G+8193  U+0000  controlNUL,  <control>
        0x0000, # G+8194  U+0000  controlNUL,  <control>
        0x0000, # G+8195  U+0000  controlNUL,  <control>
        0x0000, # G+8196  U+0000  controlNUL,  <control>
        0x0000, # G+8197  U+0000  controlNUL,  <control>
        0x0000, # G+8198  U+0000  controlNUL,  <control>
        0x0000, # G+8199  U+0000  controlNUL,  <control>
        0x0000, # G+8200  U+0000  controlNUL,  <control>
        0x0000, # G+8201  U+0000  controlNUL,  <control>
        0x0000, # G+8202  U+0000  controlNUL,  <control>
        0x0000, # G+8203  U+0000  controlNUL,  <control>
        0x0000, # G+8204  U+0000  controlNUL,  <control>
        0x0000, # G+8205  U+0000  controlNUL,  <control>
        0x0000, # G+8206  U+0000  controlNUL,  <control>
        0x0000, # G+8207  U+0000  controlNUL,  <control>
        0x0000, # G+8208  U+0000  controlNUL,  <control>
        0x0000, # G+8209  U+0000  controlNUL,  <control>
        0x0000, # G+8210  U+0000  controlNUL,  <control>
        0x0000, # G+8211  U+0000  controlNUL,  <control>
        0x0000, # G+8212  U+0000  controlNUL,  <control>
        0x0000, # G+8213  U+0000  controlNUL,  <control>
        0x0000, # G+8214  U+0000  controlNUL,  <control>
        0x0000, # G+8215  U+0000  controlNUL,  <control>
        0x0000, # G+8216  U+0000  controlNUL,  <control>
        0x0000, # G+8217  U+0000  controlNUL,  <control>
        0x0000, # G+8218  U+0000  controlNUL,  <control>
        0x3016, # G+8219  U+3016  whitelenticularbracketleft,  LEFT WHITE LENTICULAR BRACKET
        0x3017, # G+8220  U+3017  whitelenticularbracketright,  RIGHT WHITE LENTICULAR BRACKET
        0x3018, # G+8221  U+3018  whitetortoiseshellbracketleft,  LEFT WHITE TORTOISE SHELL BRACKET
        0x3019, # G+8222  U+3019  whitetortoiseshellbracketright,  RIGHT WHITE TORTOISE SHELL BRACKET
        0x0000, # G+8223  U+0000  controlNUL,  <control>
        0x0000, # G+8224  U+0000  controlNUL,  <control>
        0x0000, # G+8225  U+0000  controlNUL,  <control>
        0x0000, # G+8226  U+0000  controlNUL,  <control>
        0x0000, # G+8227  U+0000  controlNUL,  <control>
        0x0000, # G+8228  U+0000  controlNUL,  <control>
        0x0000, # G+8229  U+0000  controlNUL,  <control>
        0x0000, # G+8230  U+0000  controlNUL,  <control>
        0x0000, # G+8231  U+0000  controlNUL,  <control>
        0x0000, # G+8232  U+0000  controlNUL,  <control>
        0x0000, # G+8233  U+0000  controlNUL,  <control>
        0x0000, # G+8234  U+0000  controlNUL,  <control>
        0x0000, # G+8235  U+0000  controlNUL,  <control>
        0x0000, # G+8236  U+0000  controlNUL,  <control>
        0x201F, # G+8237  U+201F  uni201F,  DOUBLE HIGH-REVERSED-9 QUOTATION MARK
        0x201B, # G+8238  U+201B  quotereversed,  SINGLE HIGH-REVERSED-9 QUOTATION MARK
        0x207A, # G+8239  U+207A  plussuperior,  SUPERSCRIPT PLUS SIGN
        0x207B, # G+8240  U+207B  uni207B,  SUPERSCRIPT MINUS
        0x0000, # G+8241  U+0000  controlNUL,  <control>
        0x0000, # G+8242  U+0000  controlNUL,  <control>
        0x0000, # G+8243  U+0000  controlNUL,  <control>
        0x0000, # G+8244  U+0000  controlNUL,  <control>
        0x0000, # G+8245  U+0000  controlNUL,  <control>
        0x0000, # G+8246  U+0000  controlNUL,  <control>
        0x0000, # G+8247  U+0000  controlNUL,  <control>
        0x207C, # G+8248  U+207C  equalsuperior,  SUPERSCRIPT EQUALS SIGN
        0x0000, # G+8249  U+0000  controlNUL,  <control>
        0x207D, # G+8250  U+207D  parenleftsuperior,  SUPERSCRIPT LEFT PARENTHESIS
        0x207E, # G+8251  U+207E  parenrightsuperior,  SUPERSCRIPT RIGHT PARENTHESIS
        0x0000, # G+8252  U+0000  controlNUL,  <control>
        0x0000, # G+8253  U+0000  controlNUL,  <control>
        0x0000, # G+8254  U+0000  controlNUL,  <control>
        0x0000, # G+8255  U+0000  controlNUL,  <control>
        0x0000, # G+8256  U+0000  controlNUL,  <control>
        0x0000, # G+8257  U+0000  controlNUL,  <control>
        0x0000, # G+8258  U+0000  controlNUL,  <control>
        0x0000, # G+8259  U+0000  controlNUL,  <control>
        0x0000, # G+8260  U+0000  controlNUL,  <control>
        0x0000, # G+8261  U+0000  controlNUL,  <control>
        0x0000, # G+8262  U+0000  controlNUL,  <control>
        0x0000, # G+8263  U+0000  controlNUL,  <control>
        0x0000, # G+8264  U+0000  controlNUL,  <control>
        0x0000, # G+8265  U+0000  controlNUL,  <control>
        0x0000, # G+8266  U+0000  controlNUL,  <control>
        0x0000, # G+8267  U+0000  controlNUL,  <control>
        0x0000, # G+8268  U+0000  controlNUL,  <control>
        0x0000, # G+8269  U+0000  controlNUL,  <control>
        0x0000, # G+8270  U+0000  controlNUL,  <control>
        0x0000, # G+8271  U+0000  controlNUL,  <control>
        0x0000, # G+8272  U+0000  controlNUL,  <control>
        0x0000, # G+8273  U+0000  controlNUL,  <control>
        0x0000, # G+8274  U+0000  controlNUL,  <control>
        0x02BC, # G+8275  U+02BC  afii57929,  MODIFIER LETTER APOSTROPHE
        0x0000, # G+8276  U+0000  controlNUL,  <control>
        0x0000, # G+8277  U+0000  controlNUL,  <control>
        0x0000, # G+8278  U+0000  controlNUL,  <control>
        0x0000, # G+8279  U+0000  controlNUL,  <control>
        0x0000, # G+8280  U+0000  controlNUL,  <control>
        0x0000, # G+8281  U+0000  controlNUL,  <control>
        0x0000, # G+8282  U+0000  controlNUL,  <control>
        0x0000, # G+8283  U+0000  controlNUL,  <control>
        0x0000, # G+8284  U+0000  controlNUL,  <control>
        0x0000, # G+8285  U+0000  controlNUL,  <control>
        0x0000, # G+8286  U+0000  controlNUL,  <control>
        0x0000, # G+8287  U+0000  controlNUL,  <control>
        0x0000, # G+8288  U+0000  controlNUL,  <control>
        0x0000, # G+8289  U+0000  controlNUL,  <control>
        0x0000, # G+8290  U+0000  controlNUL,  <control>
        0x0000, # G+8291  U+0000  controlNUL,  <control>
        0x0000, # G+8292  U+0000  controlNUL,  <control>
        0x0000, # G+8293  U+0000  controlNUL,  <control>
        0x0000, # G+8294  U+0000  controlNUL,  <control>
        0x0000, # G+8295  U+0000  controlNUL,  <control>
        0x0000, # G+8296  U+0000  controlNUL,  <control>
        0x0000, # G+8297  U+0000  controlNUL,  <control>
        0x0000, # G+8298  U+0000  controlNUL,  <control>
        0x0000, # G+8299  U+0000  controlNUL,  <control>
        0x0000, # G+8300  U+0000  controlNUL,  <control>
        0x0000, # G+8301  U+0000  controlNUL,  <control>
        0x0000, # G+8302  U+0000  controlNUL,  <control>
        0x0000, # G+8303  U+0000  controlNUL,  <control>
        0x0000, # G+8304  U+0000  controlNUL,  <control>
        0x0000, # G+8305  U+0000  controlNUL,  <control>
        0x0000, # G+8306  U+0000  controlNUL,  <control>
        0x0000, # G+8307  U+0000  controlNUL,  <control>
        0x0000, # G+8308  U+0000  controlNUL,  <control>
        0x0000, # G+8309  U+0000  controlNUL,  <control>
        0x0000, # G+8310  U+0000  controlNUL,  <control>
        0x0000, # G+8311  U+0000  controlNUL,  <control>
        0x0000, # G+8312  U+0000  controlNUL,  <control>
        0x0000, # G+8313  U+0000  controlNUL,  <control>
        0x0000, # G+8314  U+0000  controlNUL,  <control>
        0x0000, # G+8315  U+0000  controlNUL,  <control>
        0x0000, # G+8316  U+0000  controlNUL,  <control>
        0x0000, # G+8317  U+0000  controlNUL,  <control>
        0x0000, # G+8318  U+0000  controlNUL,  <control>
        0x0000, # G+8319  U+0000  controlNUL,  <control>
        0x22EE, # G+8320  U+22EE  ellipsisvertical,  VERTICAL ELLIPSIS
        0x2237, # G+8321  U+2237  proportion,  PROPORTION
        0x0000, # G+8322  U+0000  controlNUL,  <control>
        0x0000, # G+8323  U+0000  controlNUL,  <control>
        0x0000, # G+8324  U+0000  controlNUL,  <control>
        0x0000, # G+8325  U+0000  controlNUL,  <control>
        0x0000, # G+8326  U+0000  controlNUL,  <control>
        0x0000, # G+8327  U+0000  controlNUL,  <control>
        0x0000, # G+8328  U+0000  controlNUL,  <control>
        0x0000, # G+8329  U+0000  controlNUL,  <control>
        0x0000, # G+8330  U+0000  controlNUL,  <control>
        0x0000, # G+8331  U+0000  controlNUL,  <control>
        0x0000, # G+8332  U+0000  controlNUL,  <control>
        0x0000, # G+8333  U+0000  controlNUL,  <control>
        0x0000, # G+8334  U+0000  controlNUL,  <control>
        0x0000, # G+8335  U+0000  controlNUL,  <control>
        0x0000, # G+8336  U+0000  controlNUL,  <control>
        0x0000, # G+8337  U+0000  controlNUL,  <control>
        0x0000, # G+8338  U+0000  controlNUL,  <control>
        0x0000, # G+8339  U+0000  controlNUL,  <control>
        0x0000, # G+8340  U+0000  controlNUL,  <control>
        0x0000, # G+8341  U+0000  controlNUL,  <control>
        0x278A, # G+8342  U+278A  onecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ONE
        0x278B, # G+8343  U+278B  twocircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT TWO
        0x278C, # G+8344  U+278C  threecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT THREE
        0x278D, # G+8345  U+278D  fourcircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FOUR
        0x278E, # G+8346  U+278E  fivecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FIVE
        0x278F, # G+8347  U+278F  sixcircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SIX
        0x2790, # G+8348  U+2790  sevencircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SEVEN
        0x2791, # G+8349  U+2791  eightcircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT EIGHT
        0x2792, # G+8350  U+2792  ninecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT NINE
        0x2793, # G+8351  U+2793  a159,  DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN
        0x0000, # G+8352  U+0000  controlNUL,  <control>
        0x0000, # G+8353  U+0000  controlNUL,  <control>
        0x0000, # G+8354  U+0000  controlNUL,  <control>
        0x0000, # G+8355  U+0000  controlNUL,  <control>
        0x0000, # G+8356  U+0000  controlNUL,  <control>
        0x0000, # G+8357  U+0000  controlNUL,  <control>
        0x0000, # G+8358  U+0000  controlNUL,  <control>
        0x0000, # G+8359  U+0000  controlNUL,  <control>
        0x0000, # G+8360  U+0000  controlNUL,  <control>
        0x0000, # G+8361  U+0000  controlNUL,  <control>
        0x0000, # G+8362  U+0000  controlNUL,  <control>
        0x0000, # G+8363  U+0000  controlNUL,  <control>
        0x0000, # G+8364  U+0000  controlNUL,  <control>
        0x0000, # G+8365  U+0000  controlNUL,  <control>
        0x0000, # G+8366  U+0000  controlNUL,  <control>
        0x0000, # G+8367  U+0000  controlNUL,  <control>
        0x0000, # G+8368  U+0000  controlNUL,  <control>
        0x0000, # G+8369  U+0000  controlNUL,  <control>
        0x0000, # G+8370  U+0000  controlNUL,  <control>
        0x0000, # G+8371  U+0000  controlNUL,  <control>
        0x0000, # G+8372  U+0000  controlNUL,  <control>
        0x0000, # G+8373  U+0000  controlNUL,  <control>
        0x0000, # G+8374  U+0000  controlNUL,  <control>
        0x0000, # G+8375  U+0000  controlNUL,  <control>
        0x0000, # G+8376  U+0000  controlNUL,  <control>
        0x0000, # G+8377  U+0000  controlNUL,  <control>
        0x0000, # G+8378  U+0000  controlNUL,  <control>
        0x0000, # G+8379  U+0000  controlNUL,  <control>
        0x0000, # G+8380  U+0000  controlNUL,  <control>
        0x0000, # G+8381  U+0000  controlNUL,  <control>
        0x0000, # G+8382  U+0000  controlNUL,  <control>
        0x0000, # G+8383  U+0000  controlNUL,  <control>
        0x0000, # G+8384  U+0000  controlNUL,  <control>
        0x0000, # G+8385  U+0000  controlNUL,  <control>
        0x0000, # G+8386  U+0000  controlNUL,  <control>
        0x0000, # G+8387  U+0000  controlNUL,  <control>
        0x24B6, # G+8388  U+24B6  Acircle,  CIRCLED LATIN CAPITAL LETTER A
        0x24B7, # G+8389  U+24B7  Bcircle,  CIRCLED LATIN CAPITAL LETTER B
        0x24B8, # G+8390  U+24B8  Ccircle,  CIRCLED LATIN CAPITAL LETTER C
        0x24B9, # G+8391  U+24B9  Dcircle,  CIRCLED LATIN CAPITAL LETTER D
        0x24BA, # G+8392  U+24BA  Ecircle,  CIRCLED LATIN CAPITAL LETTER E
        0x24BB, # G+8393  U+24BB  Fcircle,  CIRCLED LATIN CAPITAL LETTER F
        0x24BC, # G+8394  U+24BC  Gcircle,  CIRCLED LATIN CAPITAL LETTER G
        0x24BD, # G+8395  U+24BD  Hcircle,  CIRCLED LATIN CAPITAL LETTER H
        0x24BE, # G+8396  U+24BE  Icircle,  CIRCLED LATIN CAPITAL LETTER I
        0x24BF, # G+8397  U+24BF  Jcircle,  CIRCLED LATIN CAPITAL LETTER J
        0x24C0, # G+8398  U+24C0  Kcircle,  CIRCLED LATIN CAPITAL LETTER K
        0x24C1, # G+8399  U+24C1  Lcircle,  CIRCLED LATIN CAPITAL LETTER L
        0x24C2, # G+8400  U+24C2  Mcircle,  CIRCLED LATIN CAPITAL LETTER M
        0x24C3, # G+8401  U+24C3  Ncircle,  CIRCLED LATIN CAPITAL LETTER N
        0x24C4, # G+8402  U+24C4  Ocircle,  CIRCLED LATIN CAPITAL LETTER O
        0x24C5, # G+8403  U+24C5  Pcircle,  CIRCLED LATIN CAPITAL LETTER P
        0x24C6, # G+8404  U+24C6  Qcircle,  CIRCLED LATIN CAPITAL LETTER Q
        0x24C7, # G+8405  U+24C7  Rcircle,  CIRCLED LATIN CAPITAL LETTER R
        0x24C8, # G+8406  U+24C8  Scircle,  CIRCLED LATIN CAPITAL LETTER S
        0x24C9, # G+8407  U+24C9  Tcircle,  CIRCLED LATIN CAPITAL LETTER T
        0x24CA, # G+8408  U+24CA  Ucircle,  CIRCLED LATIN CAPITAL LETTER U
        0x24CB, # G+8409  U+24CB  Vcircle,  CIRCLED LATIN CAPITAL LETTER V
        0x24CC, # G+8410  U+24CC  Wcircle,  CIRCLED LATIN CAPITAL LETTER W
        0x24CD, # G+8411  U+24CD  Xcircle,  CIRCLED LATIN CAPITAL LETTER X
        0x24CE, # G+8412  U+24CE  Ycircle,  CIRCLED LATIN CAPITAL LETTER Y
        0x24CF, # G+8413  U+24CF  Zcircle,  CIRCLED LATIN CAPITAL LETTER Z
        0x0000, # G+8414  U+0000  controlNUL,  <control>
        0x0000, # G+8415  U+0000  controlNUL,  <control>
        0x0000, # G+8416  U+0000  controlNUL,  <control>
        0x0000, # G+8417  U+0000  controlNUL,  <control>
        0x0000, # G+8418  U+0000  controlNUL,  <control>
        0x0000, # G+8419  U+0000  controlNUL,  <control>
        0x0000, # G+8420  U+0000  controlNUL,  <control>
        0x0000, # G+8421  U+0000  controlNUL,  <control>
        0x0000, # G+8422  U+0000  controlNUL,  <control>
        0x0000, # G+8423  U+0000  controlNUL,  <control>
        0x0000, # G+8424  U+0000  controlNUL,  <control>
        0x0000, # G+8425  U+0000  controlNUL,  <control>
        0x0000, # G+8426  U+0000  controlNUL,  <control>
        0x0000, # G+8427  U+0000  controlNUL,  <control>
        0x0000, # G+8428  U+0000  controlNUL,  <control>
        0x0000, # G+8429  U+0000  controlNUL,  <control>
        0x0000, # G+8430  U+0000  controlNUL,  <control>
        0x0000, # G+8431  U+0000  controlNUL,  <control>
        0x0000, # G+8432  U+0000  controlNUL,  <control>
        0x0000, # G+8433  U+0000  controlNUL,  <control>
        0x0000, # G+8434  U+0000  controlNUL,  <control>
        0x0000, # G+8435  U+0000  controlNUL,  <control>
        0x0000, # G+8436  U+0000  controlNUL,  <control>
        0x0000, # G+8437  U+0000  controlNUL,  <control>
        0x0000, # G+8438  U+0000  controlNUL,  <control>
        0x0000, # G+8439  U+0000  controlNUL,  <control>
        0x0000, # G+8440  U+0000  controlNUL,  <control>
        0x0000, # G+8441  U+0000  controlNUL,  <control>
        0x0000, # G+8442  U+0000  controlNUL,  <control>
        0x0000, # G+8443  U+0000  controlNUL,  <control>
        0x0000, # G+8444  U+0000  controlNUL,  <control>
        0x0000, # G+8445  U+0000  controlNUL,  <control>
        0x0000, # G+8446  U+0000  controlNUL,  <control>
        0x0000, # G+8447  U+0000  controlNUL,  <control>
        0x0000, # G+8448  U+0000  controlNUL,  <control>
        0x0000, # G+8449  U+0000  controlNUL,  <control>
        0x0000, # G+8450  U+0000  controlNUL,  <control>
        0x0000, # G+8451  U+0000  controlNUL,  <control>
        0x0000, # G+8452  U+0000  controlNUL,  <control>
        0x0000, # G+8453  U+0000  controlNUL,  <control>
        0x0000, # G+8454  U+0000  controlNUL,  <control>
        0x0000, # G+8455  U+0000  controlNUL,  <control>
        0x0000, # G+8456  U+0000  controlNUL,  <control>
        0x0000, # G+8457  U+0000  controlNUL,  <control>
        0x0000, # G+8458  U+0000  controlNUL,  <control>
        0x0000, # G+8459  U+0000  controlNUL,  <control>
        0x0000, # G+8460  U+0000  controlNUL,  <control>
        0x0000, # G+8461  U+0000  controlNUL,  <control>
        0x0000, # G+8462  U+0000  controlNUL,  <control>
        0x0000, # G+8463  U+0000  controlNUL,  <control>
        0x0000, # G+8464  U+0000  controlNUL,  <control>
        0x0000, # G+8465  U+0000  controlNUL,  <control>
        0x0000, # G+8466  U+0000  controlNUL,  <control>
        0x0000, # G+8467  U+0000  controlNUL,  <control>
        0x0000, # G+8468  U+0000  controlNUL,  <control>
        0x0000, # G+8469  U+0000  controlNUL,  <control>
        0x0000, # G+8470  U+0000  controlNUL,  <control>
        0x0000, # G+8471  U+0000  controlNUL,  <control>
        0x0000, # G+8472  U+0000  controlNUL,  <control>
        0x0000, # G+8473  U+0000  controlNUL,  <control>
        0x0000, # G+8474  U+0000  controlNUL,  <control>
        0x227A, # G+8475  U+227A  precedes,  PRECEDES
        0x227B, # G+8476  U+227B  succeeds,  SUCCEEDS
        0x22CE, # G+8477  U+22CE  curlyor,  CURLY LOGICAL OR
        0x22CF, # G+8478  U+22CF  curlyand,  CURLY LOGICAL AND
        0x2280, # G+8479  U+2280  notprecedes,  DOES NOT PRECEDE
        0x2281, # G+8480  U+2281  notsucceeds,  DOES NOT SUCCEED
        0x2270, # G+8481  U+2270  notlessnorequal,  NEITHER LESS-THAN NOR EQUAL TO
        0x2271, # G+8482  U+2271  notgreaternorequal,  NEITHER GREATER-THAN NOR EQUAL TO
        0x2272, # G+8483  U+2272  lessorequivalent,  LESS-THAN OR EQUIVALENT TO
        0x2273, # G+8484  U+2273  greaterorequivalent,  GREATER-THAN OR EQUIVALENT TO
        0x0000, # G+8485  U+0000  controlNUL,  <control>
        0x228A, # G+8486  U+228A  subsetnotequal,  SUBSET OF WITH NOT EQUAL TO
        0x0000, # G+8487  U+0000  controlNUL,  <control>
        0x228B, # G+8488  U+228B  supersetnotequal,  SUPERSET OF WITH NOT EQUAL TO
        0x2276, # G+8489  U+2276  lessorgreater,  LESS-THAN OR GREATER-THAN
        0x2277, # G+8490  U+2277  greaterorless,  GREATER-THAN OR LESS-THAN
        0x2279, # G+8491  U+2279  notgreaternorless,  NEITHER GREATER-THAN NOR LESS-THAN
        0x22DA, # G+8492  U+22DA  lessequalorgreater,  LESS-THAN EQUAL TO OR GREATER-THAN
        0x22DB, # G+8493  U+22DB  greaterequalorless,  GREATER-THAN EQUAL TO OR LESS-THAN
        0x0000, # G+8494  U+0000  controlNUL,  <control>
        0x0000, # G+8495  U+0000  controlNUL,  <control>
        0x0000, # G+8496  U+0000  controlNUL,  <control>
        0x0000, # G+8497  U+0000  controlNUL,  <control>
        0x0000, # G+8498  U+0000  controlNUL,  <control>
        0x2245, # G+8499  U+2245  congruent,  APPROXIMATELY EQUAL TO
        0x2243, # G+8500  U+2243  asymptoticallyequal,  ASYMPTOTICALLY EQUAL TO
        0x2248, # G+8501  U+2248  approxequal,  ALMOST EQUAL TO
        0x0000, # G+8502  U+0000  controlNUL,  <control>
        0x22A4, # G+8503  U+22A4  tackdown,  DOWN TACK
        0x0000, # G+8504  U+0000  controlNUL,  <control>
        0x0000, # G+8505  U+0000  controlNUL,  <control>
        0x0000, # G+8506  U+0000  controlNUL,  <control>
        0x0000, # G+8507  U+0000  controlNUL,  <control>
        0x0000, # G+8508  U+0000  controlNUL,  <control>
        0x0000, # G+8509  U+0000  controlNUL,  <control>
        0x0000, # G+8510  U+0000  controlNUL,  <control>
        0x0000, # G+8511  U+0000  controlNUL,  <control>
        0x0000, # G+8512  U+0000  controlNUL,  <control>
        0x0000, # G+8513  U+0000  controlNUL,  <control>
        0x0000, # G+8514  U+0000  controlNUL,  <control>
        0x0000, # G+8515  U+0000  controlNUL,  <control>
        0x0000, # G+8516  U+0000  controlNUL,  <control>
        0x0000, # G+8517  U+0000  controlNUL,  <control>
        0x0000, # G+8518  U+0000  controlNUL,  <control>
        0x0000, # G+8519  U+0000  controlNUL,  <control>
        0x0000, # G+8520  U+0000  controlNUL,  <control>
        0x0000, # G+8521  U+0000  controlNUL,  <control>
        0x0000, # G+8522  U+0000  controlNUL,  <control>
        0x0000, # G+8523  U+0000  controlNUL,  <control>
        0x0000, # G+8524  U+0000  controlNUL,  <control>
        0x0000, # G+8525  U+0000  controlNUL,  <control>
        0x0000, # G+8526  U+0000  controlNUL,  <control>
        0x0000, # G+8527  U+0000  controlNUL,  <control>
        0x0000, # G+8528  U+0000  controlNUL,  <control>
        0x0000, # G+8529  U+0000  controlNUL,  <control>
        0x0000, # G+8530  U+0000  controlNUL,  <control>
        0x0000, # G+8531  U+0000  controlNUL,  <control>
        0x0000, # G+8532  U+0000  controlNUL,  <control>
        0x0000, # G+8533  U+0000  controlNUL,  <control>
        0x0000, # G+8534  U+0000  controlNUL,  <control>
        0x0000, # G+8535  U+0000  controlNUL,  <control>
        0x0000, # G+8536  U+0000  controlNUL,  <control>
        0x0000, # G+8537  U+0000  controlNUL,  <control>
        0x0000, # G+8538  U+0000  controlNUL,  <control>
        0x0000, # G+8539  U+0000  controlNUL,  <control>
        0x0000, # G+8540  U+0000  controlNUL,  <control>
        0x0000, # G+8541  U+0000  controlNUL,  <control>
        0x0000, # G+8542  U+0000  controlNUL,  <control>
        0x0000, # G+8543  U+0000  controlNUL,  <control>
        0x0000, # G+8544  U+0000  controlNUL,  <control>
        0x0000, # G+8545  U+0000  controlNUL,  <control>
        0x0000, # G+8546  U+0000  controlNUL,  <control>
        0x0000, # G+8547  U+0000  controlNUL,  <control>
        0x0000, # G+8548  U+0000  controlNUL,  <control>
        0x0000, # G+8549  U+0000  controlNUL,  <control>
        0x0000, # G+8550  U+0000  controlNUL,  <control>
        0x0000, # G+8551  U+0000  controlNUL,  <control>
        0x0000, # G+8552  U+0000  controlNUL,  <control>
        0x0000, # G+8553  U+0000  controlNUL,  <control>
        0x0000, # G+8554  U+0000  controlNUL,  <control>
        0x0000, # G+8555  U+0000  controlNUL,  <control>
        0x0000, # G+8556  U+0000  controlNUL,  <control>
        0x0000, # G+8557  U+0000  controlNUL,  <control>
        0x0000, # G+8558  U+0000  controlNUL,  <control>
        0x0000, # G+8559  U+0000  controlNUL,  <control>
        0x0000, # G+8560  U+0000  controlNUL,  <control>
        0x0000, # G+8561  U+0000  controlNUL,  <control>
        0x0000, # G+8562  U+0000  controlNUL,  <control>
        0x0000, # G+8563  U+0000  controlNUL,  <control>
        0x0000, # G+8564  U+0000  controlNUL,  <control>
        0x0000, # G+8565  U+0000  controlNUL,  <control>
        0x0000, # G+8566  U+0000  controlNUL,  <control>
        0x0000, # G+8567  U+0000  controlNUL,  <control>
        0x0000, # G+8568  U+0000  controlNUL,  <control>
        0x0000, # G+8569  U+0000  controlNUL,  <control>
        0x0000, # G+8570  U+0000  controlNUL,  <control>
        0x0000, # G+8571  U+0000  controlNUL,  <control>
        0x0000, # G+8572  U+0000  controlNUL,  <control>
        0x0000, # G+8573  U+0000  controlNUL,  <control>
        0x0000, # G+8574  U+0000  controlNUL,  <control>
        0x0000, # G+8575  U+0000  controlNUL,  <control>
        0x0000, # G+8576  U+0000  controlNUL,  <control>
        0x0000, # G+8577  U+0000  controlNUL,  <control>
        0x0000, # G+8578  U+0000  controlNUL,  <control>
        0x0000, # G+8579  U+0000  controlNUL,  <control>
        0x0000, # G+8580  U+0000  controlNUL,  <control>
        0x0000, # G+8581  U+0000  controlNUL,  <control>
        0x2034, # G+8582  U+2034  tprime,  TRIPLE PRIME
        0x0000, # G+8583  U+0000  controlNUL,  <control>
        0x0000, # G+8584  U+0000  controlNUL,  <control>
        0x0000, # G+8585  U+0000  controlNUL,  <control>
        0x0000, # G+8586  U+0000  controlNUL,  <control>
        0x0000, # G+8587  U+0000  controlNUL,  <control>
        0x0000, # G+8588  U+0000  controlNUL,  <control>
        0x0000, # G+8589  U+0000  controlNUL,  <control>
        0x0000, # G+8590  U+0000  controlNUL,  <control>
        0x0000, # G+8591  U+0000  controlNUL,  <control>
        0x0000, # G+8592  U+0000  controlNUL,  <control>
        0x0000, # G+8593  U+0000  controlNUL,  <control>
        0x266F, # G+8594  U+266F  musicsharpsign,  MUSIC SHARP SIGN
        0x0000, # G+8595  U+0000  controlNUL,  <control>
        0x0000, # G+8596  U+0000  controlNUL,  <control>
        0x0000, # G+8597  U+0000  controlNUL,  <control>
        0x0000, # G+8598  U+0000  controlNUL,  <control>
        0x2042, # G+8599  U+2042  asterism,  ASTERISM
        0x0000, # G+8600  U+0000  controlNUL,  <control>
        0x0000, # G+8601  U+0000  controlNUL,  <control>
        0x0000, # G+8602  U+0000  controlNUL,  <control>
        0x0000, # G+8603  U+0000  controlNUL,  <control>
        0x273D, # G+8604  U+273D  a56,  HEAVY TEARDROP-SPOKED ASTERISK
        0x0000, # G+8605  U+0000  controlNUL,  <control>
        0x0000, # G+8606  U+0000  controlNUL,  <control>
        0x2022, # G+8607  U+2022  bullet,  BULLET
        0x0000, # G+8608  U+0000  controlNUL,  <control>
        0x0000, # G+8609  U+0000  controlNUL,  <control>
        0x0000, # G+8610  U+0000  controlNUL,  <control>
        0x0000, # G+8611  U+0000  controlNUL,  <control>
        0x2039, # G+8612  U+2039  guilsinglleft,  SINGLE LEFT-POINTING ANGLE QUOTATION MARK
        0x203A, # G+8613  U+203A  guilsinglright,  SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
        0x0000, # G+8614  U+0000  controlNUL,  <control>
        0x0000, # G+8615  U+0000  controlNUL,  <control>
        0x0000, # G+8616  U+0000  controlNUL,  <control>
        0x0000, # G+8617  U+0000  controlNUL,  <control>
        0x0000, # G+8618  U+0000  controlNUL,  <control>
        0x0000, # G+8619  U+0000  controlNUL,  <control>
        0x0000, # G+8620  U+0000  controlNUL,  <control>
        0x0000, # G+8621  U+0000  controlNUL,  <control>
        0x0000, # G+8622  U+0000  controlNUL,  <control>
        0x0000, # G+8623  U+0000  controlNUL,  <control>
        0x0000, # G+8624  U+0000  controlNUL,  <control>
        0x0000, # G+8625  U+0000  controlNUL,  <control>
        0x0000, # G+8626  U+0000  controlNUL,  <control>
        0x0000, # G+8627  U+0000  controlNUL,  <control>
        0x0000, # G+8628  U+0000  controlNUL,  <control>
        0x0000, # G+8629  U+0000  controlNUL,  <control>
        0x271A, # G+8630  U+271A  a26,  HEAVY GREEK CROSS
        0x2716, # G+8631  U+2716  a22,  HEAVY MULTIPLICATION X
        0x0000, # G+8632  U+0000  controlNUL,  <control>
        0x25EF, # G+8633  U+25EF  largecircle,  LARGE CIRCLE
        0x0000, # G+8634  U+0000  controlNUL,  <control>
        0x0000, # G+8635  U+0000  controlNUL,  <control>
        0x0000, # G+8636  U+0000  controlNUL,  <control>
        0x2756, # G+8637  U+2756  a79,  BLACK DIAMOND MINUS WHITE X
        0x0000, # G+8638  U+0000  controlNUL,  <control>
        0x25CC, # G+8639  U+25CC  dottedcircle,  DOTTED CIRCLE
        0x0000, # G+8640  U+0000  controlNUL,  <control>
        0x0000, # G+8641  U+0000  controlNUL,  <control>
        0x0000, # G+8642  U+0000  controlNUL,  <control>
        0x0000, # G+8643  U+0000  controlNUL,  <control>
        0x0000, # G+8644  U+0000  controlNUL,  <control>
        0x0000, # G+8645  U+0000  controlNUL,  <control>
        0x0000, # G+8646  U+0000  controlNUL,  <control>
        0x0000, # G+8647  U+0000  controlNUL,  <control>
        0x0000, # G+8648  U+0000  controlNUL,  <control>
        0x0000, # G+8649  U+0000  controlNUL,  <control>
        0x0000, # G+8650  U+0000  controlNUL,  <control>
        0x0000, # G+8651  U+0000  controlNUL,  <control>
        0x0000, # G+8652  U+0000  controlNUL,  <control>
        0x0000, # G+8653  U+0000  controlNUL,  <control>
        0x0000, # G+8654  U+0000  controlNUL,  <control>
        0x0000, # G+8655  U+0000  controlNUL,  <control>
        0x0000, # G+8656  U+0000  controlNUL,  <control>
        0x0000, # G+8657  U+0000  controlNUL,  <control>
        0x0000, # G+8658  U+0000  controlNUL,  <control>
        0x0000, # G+8659  U+0000  controlNUL,  <control>
        0x0000, # G+8660  U+0000  controlNUL,  <control>
        0x0000, # G+8661  U+0000  controlNUL,  <control>
        0x0000, # G+8662  U+0000  controlNUL,  <control>
        0x0000, # G+8663  U+0000  controlNUL,  <control>
        0x262F, # G+8664  U+262F  yinyang,  YIN YANG
        0x0000, # G+8665  U+0000  controlNUL,  <control>
        0x0000, # G+8666  U+0000  controlNUL,  <control>
        0x0000, # G+8667  U+0000  controlNUL,  <control>
        0x0000, # G+8668  U+0000  controlNUL,  <control>
        0x0000, # G+8669  U+0000  controlNUL,  <control>
        0x0000, # G+8670  U+0000  controlNUL,  <control>
        0x3020, # G+8671  U+3020  postalmarkface,  POSTAL MARK FACE
        0x0000, # G+8672  U+0000  controlNUL,  <control>
        0x2776, # G+8673  U+2776  a130,  DINGBAT NEGATIVE CIRCLED DIGIT ONE
        0x2777, # G+8674  U+2777  a131,  DINGBAT NEGATIVE CIRCLED DIGIT TWO
        0x2778, # G+8675  U+2778  a132,  DINGBAT NEGATIVE CIRCLED DIGIT THREE
        0x2779, # G+8676  U+2779  a133,  DINGBAT NEGATIVE CIRCLED DIGIT FOUR
        0x277A, # G+8677  U+277A  a134,  DINGBAT NEGATIVE CIRCLED DIGIT FIVE
        0x277B, # G+8678  U+277B  a135,  DINGBAT NEGATIVE CIRCLED DIGIT SIX
        0x277C, # G+8679  U+277C  a136,  DINGBAT NEGATIVE CIRCLED DIGIT SEVEN
        0x277D, # G+8680  U+277D  a137,  DINGBAT NEGATIVE CIRCLED DIGIT EIGHT
        0x277E, # G+8681  U+277E  a138,  DINGBAT NEGATIVE CIRCLED DIGIT NINE
        0x277F, # G+8682  U+277F  a139,  DINGBAT NEGATIVE CIRCLED NUMBER TEN
        0x0000, # G+8683  U+0000  controlNUL,  <control>
        0x0000, # G+8684  U+0000  controlNUL,  <control>
        0x0000, # G+8685  U+0000  controlNUL,  <control>
        0x0000, # G+8686  U+0000  controlNUL,  <control>
        0x0000, # G+8687  U+0000  controlNUL,  <control>
        0x0000, # G+8688  U+0000  controlNUL,  <control>
        0x0000, # G+8689  U+0000  controlNUL,  <control>
        0x0000, # G+8690  U+0000  controlNUL,  <control>
        0x0000, # G+8691  U+0000  controlNUL,  <control>
        0x0000, # G+8692  U+0000  controlNUL,  <control>
        0x0000, # G+8693  U+0000  controlNUL,  <control>
        0x0000, # G+8694  U+0000  controlNUL,  <control>
        0x0000, # G+8695  U+0000  controlNUL,  <control>
        0x0000, # G+8696  U+0000  controlNUL,  <control>
        0x0000, # G+8697  U+0000  controlNUL,  <control>
        0x0000, # G+8698  U+0000  controlNUL,  <control>
        0x0000, # G+8699  U+0000  controlNUL,  <control>
        0x3012, # G+8700  U+3012  postalmark,  POSTAL MARK
        0x3036, # G+8701  U+3036  circlepostalmark,  CIRCLED POSTAL MARK
        0x0000, # G+8702  U+0000  controlNUL,  <control>
        0x0000, # G+8703  U+0000  controlNUL,  <control>
        0x0000, # G+8704  U+0000  controlNUL,  <control>
        0x0000, # G+8705  U+0000  controlNUL,  <control>
        0x0000, # G+8706  U+0000  controlNUL,  <control>
        0x0000, # G+8707  U+0000  controlNUL,  <control>
        0x0000, # G+8708  U+0000  controlNUL,  <control>
        0x0000, # G+8709  U+0000  controlNUL,  <control>
        0x0000, # G+8710  U+0000  controlNUL,  <control>
        0x0000, # G+8711  U+0000  controlNUL,  <control>
        0x0000, # G+8712  U+0000  controlNUL,  <control>
        0x0000, # G+8713  U+0000  controlNUL,  <control>
        0x0000, # G+8714  U+0000  controlNUL,  <control>
        0x2206, # G+8715  U+2206  Delta,  INCREMENT
        0x0000, # G+8716  U+0000  controlNUL,  <control>
        0x221F, # G+8717  U+221F  orthogonal,  RIGHT ANGLE
        0x0000, # G+8718  U+0000  controlNUL,  <control>
        0x2225, # G+8719  U+2225  parallel,  PARALLEL TO
        0x2226, # G+8720  U+2226  notparallel,  NOT PARALLEL TO
        0x0000, # G+8721  U+0000  controlNUL,  <control>
        0x2253, # G+8722  U+2253  imageorapproximatelyequal,  IMAGE OF OR APPROXIMATELY EQUAL TO
        0x2251, # G+8723  U+2251  geometricallyequal,  GEOMETRICALLY EQUAL TO
        0x2266, # G+8724  U+2266  lessoverequal,  LESS-THAN OVER EQUAL TO
        0x2267, # G+8725  U+2267  greateroverequal,  GREATER-THAN OVER EQUAL TO
        0x2213, # G+8726  U+2213  minusplus,  MINUS-OR-PLUS SIGN
        0x2295, # G+8727  U+2295  circleplus,  CIRCLED PLUS
        0x2296, # G+8728  U+2296  minuscircle,  CIRCLED MINUS
        0x2297, # G+8729  U+2297  circlemultiply,  CIRCLED TIMES
        0x0000, # G+8730  U+0000  controlNUL,  <control>
        0x2314, # G+8731  U+2314  uni2314,  SECTOR
        0x0000, # G+8732  U+0000  controlNUL,  <control>
        0x0000, # G+8733  U+0000  controlNUL,  <control>
        0x2262, # G+8734  U+2262  notidentical,  NOT IDENTICAL TO
        0x0000, # G+8735  U+0000  controlNUL,  <control>
        0x25B1, # G+8736  U+25B1  uni25B1,  WHITE PARALLELOGRAM
        0x0000, # G+8737  U+0000  controlNUL,  <control>
        0x2222, # G+8738  U+2222  angsph,  SPHERICAL ANGLE
        0x2250, # G+8739  U+2250  approaches,  APPROACHES THE LIMIT
        0x0000, # G+8740  U+0000  controlNUL,  <control>
        0x0000, # G+8741  U+0000  controlNUL,  <control>
        0x22A3, # G+8742  U+22A3  tackleft,  LEFT TACK
        0x0000, # G+8743  U+0000  controlNUL,  <control>
        0x0000, # G+8744  U+0000  controlNUL,  <control>
        0x226E, # G+8745  U+226E  notless,  NOT LESS-THAN
        0x226F, # G+8746  U+226F  notgreater,  NOT GREATER-THAN
        0x2285, # G+8747  U+2285  notsuperset,  NOT A SUPERSET OF
        0x2284, # G+8748  U+2284  notsubset,  NOT A SUBSET OF
        0x2209, # G+8749  U+2209  notelement,  NOT AN ELEMENT OF
        0x220C, # G+8750  U+220C  notcontains,  DOES NOT CONTAIN AS MEMBER
        0x22BB, # G+8751  U+22BB  veebar,  XOR
        0x22BC, # G+8752  U+22BC  barwed,  NAND
        0x225A, # G+8753  U+225A  uni225A,  EQUIANGULAR TO
        0x2306, # G+8754  U+2306  Barwed,  PERSPECTIVE
        0x0000, # G+8755  U+0000  controlNUL,  <control>
        0x0000, # G+8756  U+0000  controlNUL,  <control>
        0x0000, # G+8757  U+0000  controlNUL,  <control>
        0x0000, # G+8758  U+0000  controlNUL,  <control>
        0x0000, # G+8759  U+0000  controlNUL,  <control>
        0x0000, # G+8760  U+0000  controlNUL,  <control>
        0x329E, # G+8761  U+329E  ideographicprintcircle,  CIRCLED IDEOGRAPH PRINT
        0x0000, # G+8762  U+0000  controlNUL,  <control>
        0x203C, # G+8763  U+203C  exclamdbl,  DOUBLE EXCLAMATION MARK
        0x0000, # G+8764  U+0000  controlNUL,  <control>
        0x0000, # G+8765  U+0000  controlNUL,  <control>
        0x0000, # G+8766  U+0000  controlNUL,  <control>
        0x0000, # G+8767  U+0000  controlNUL,  <control>
        0x0000, # G+8768  U+0000  controlNUL,  <control>
        0x0000, # G+8769  U+0000  controlNUL,  <control>
        0x0000, # G+8770  U+0000  controlNUL,  <control>
        0x0000, # G+8771  U+0000  controlNUL,  <control>
        0x0000, # G+8772  U+0000  controlNUL,  <control>
        0x0000, # G+8773  U+0000  controlNUL,  <control>
        0x0000, # G+8774  U+0000  controlNUL,  <control>
        0x25E6, # G+8775  U+25E6  openbullet,  WHITE BULLET
        0x0000, # G+8776  U+0000  controlNUL,  <control>
        0x0000, # G+8777  U+0000  controlNUL,  <control>
        0x0000, # G+8778  U+0000  controlNUL,  <control>
        0x25BF, # G+8779  U+25BF  whitedownpointingsmalltriangle,  WHITE DOWN-POINTING SMALL TRIANGLE
        0x25B5, # G+8780  U+25B5  whiteuppointingsmalltriangle,  WHITE UP-POINTING SMALL TRIANGLE
        0x25B9, # G+8781  U+25B9  whiterightpointingsmalltriangle,  WHITE RIGHT-POINTING SMALL TRIANGLE
        0x25C3, # G+8782  U+25C3  whiteleftpointingsmalltriangle,  WHITE LEFT-POINTING SMALL TRIANGLE
        0x0000, # G+8783  U+0000  controlNUL,  <control>
        0x0000, # G+8784  U+0000  controlNUL,  <control>
        0x0000, # G+8785  U+0000  controlNUL,  <control>
        0x0000, # G+8786  U+0000  controlNUL,  <control>
        0x25CA, # G+8787  U+25CA  lozenge,  LOZENGE
        0x3231, # G+8788  U+3231  ideographicstockparen,  PARENTHESIZED IDEOGRAPH STOCK
        0x3239, # G+8789  U+3239  ideographicrepresentparen,  PARENTHESIZED IDEOGRAPH REPRESENT
        0x33CB, # G+8790  U+33CB  HPsquare,  SQUARE HP
        0x246F, # G+8791  U+246F  sixteencircle,  CIRCLED NUMBER SIXTEEN
        0x2470, # G+8792  U+2470  seventeencircle,  CIRCLED NUMBER SEVENTEEN
        0x2471, # G+8793  U+2471  eighteencircle,  CIRCLED NUMBER EIGHTEEN
        0x2472, # G+8794  U+2472  nineteencircle,  CIRCLED NUMBER NINETEEN
        0x2473, # G+8795  U+2473  twentycircle,  CIRCLED NUMBER TWENTY
        0x0000, # G+8796  U+0000  controlNUL,  <control>
        0x0000, # G+8797  U+0000  controlNUL,  <control>
        0x0000, # G+8798  U+0000  controlNUL,  <control>
        0x0000, # G+8799  U+0000  controlNUL,  <control>
        0x0000, # G+8800  U+0000  controlNUL,  <control>
        0x0000, # G+8801  U+0000  controlNUL,  <control>
        0x0000, # G+8802  U+0000  controlNUL,  <control>
        0x0000, # G+8803  U+0000  controlNUL,  <control>
        0x0000, # G+8804  U+0000  controlNUL,  <control>
        0x0000, # G+8805  U+0000  controlNUL,  <control>
        0x0000, # G+8806  U+0000  controlNUL,  <control>
        0x0000, # G+8807  U+0000  controlNUL,  <control>
        0x0000, # G+8808  U+0000  controlNUL,  <control>
        0x0000, # G+8809  U+0000  controlNUL,  <control>
        0x0000, # G+8810  U+0000  controlNUL,  <control>
        0x0000, # G+8811  U+0000  controlNUL,  <control>
        0x0000, # G+8812  U+0000  controlNUL,  <control>
        0x0000, # G+8813  U+0000  controlNUL,  <control>
        0x21D0, # G+8814  U+21D0  arrowdblleft,  LEFTWARDS DOUBLE ARROW
        0x21CF, # G+8815  U+21CF  arrowrightdblstroke,  RIGHTWARDS DOUBLE ARROW WITH STROKE
        0x21CD, # G+8816  U+21CD  arrowleftdblstroke,  LEFTWARDS DOUBLE ARROW WITH STROKE
        0x0000, # G+8817  U+0000  controlNUL,  <control>
        0x0000, # G+8818  U+0000  controlNUL,  <control>
        0x0000, # G+8819  U+0000  controlNUL,  <control>
        0x0000, # G+8820  U+0000  controlNUL,  <control>
        0x0000, # G+8821  U+0000  controlNUL,  <control>
        0x0000, # G+8822  U+0000  controlNUL,  <control>
        0x0000, # G+8823  U+0000  controlNUL,  <control>
        0x0000, # G+8824  U+0000  controlNUL,  <control>
        0x0000, # G+8825  U+0000  controlNUL,  <control>
        0x0000, # G+8826  U+0000  controlNUL,  <control>
        0x0000, # G+8827  U+0000  controlNUL,  <control>
        0x0000, # G+8828  U+0000  controlNUL,  <control>
        0x0000, # G+8829  U+0000  controlNUL,  <control>
        0x0000, # G+8830  U+0000  controlNUL,  <control>
        0x0000, # G+8831  U+0000  controlNUL,  <control>
        0x0000, # G+8832  U+0000  controlNUL,  <control>
        0x0000, # G+8833  U+0000  controlNUL,  <control>
        0x0000, # G+8834  U+0000  controlNUL,  <control>
        0x0000, # G+8835  U+0000  controlNUL,  <control>
        0x0000, # G+8836  U+0000  controlNUL,  <control>
        0x0000, # G+8837  U+0000  controlNUL,  <control>
        0x0000, # G+8838  U+0000  controlNUL,  <control>
        0x0000, # G+8839  U+0000  controlNUL,  <control>
        0x0000, # G+8840  U+0000  controlNUL,  <control>
        0x0000, # G+8841  U+0000  controlNUL,  <control>
        0x0000, # G+8842  U+0000  controlNUL,  <control>
        0x0000, # G+8843  U+0000  controlNUL,  <control>
        0x0000, # G+8844  U+0000  controlNUL,  <control>
        0x0000, # G+8845  U+0000  controlNUL,  <control>
        0x0000, # G+8846  U+0000  controlNUL,  <control>
        0x0000, # G+8847  U+0000  controlNUL,  <control>
        0x0000, # G+8848  U+0000  controlNUL,  <control>
        0x0000, # G+8849  U+0000  controlNUL,  <control>
        0x0000, # G+8850  U+0000  controlNUL,  <control>
        0x0000, # G+8851  U+0000  controlNUL,  <control>
        0x0000, # G+8852  U+0000  controlNUL,  <control>
        0x0000, # G+8853  U+0000  controlNUL,  <control>
        0x21D1, # G+8854  U+21D1  arrowdblup,  UPWARDS DOUBLE ARROW
        0x21D3, # G+8855  U+21D3  arrowdbldown,  DOWNWARDS DOUBLE ARROW
        0x0000, # G+8856  U+0000  controlNUL,  <control>
        0x0000, # G+8857  U+0000  controlNUL,  <control>
        0x0000, # G+8858  U+0000  controlNUL,  <control>
        0x0000, # G+8859  U+0000  controlNUL,  <control>
        0x0000, # G+8860  U+0000  controlNUL,  <control>
        0x0000, # G+8861  U+0000  controlNUL,  <control>
        0x0000, # G+8862  U+0000  controlNUL,  <control>
        0x0000, # G+8863  U+0000  controlNUL,  <control>
        0x21B2, # G+8864  U+21B2  uni21B2,  DOWNWARDS ARROW WITH TIP LEFTWARDS
        0x21B1, # G+8865  U+21B1  rsh,  UPWARDS ARROW WITH TIP RIGHTWARDS
        0x0000, # G+8866  U+0000  controlNUL,  <control>
        0x21B4, # G+8867  U+21B4  uni21B4,  RIGHTWARDS ARROW WITH CORNER DOWNWARDS
        0x21B0, # G+8868  U+21B0  lsh,  UPWARDS ARROW WITH TIP LEFTWARDS
        0x21B3, # G+8869  U+21B3  uni21B3,  DOWNWARDS ARROW WITH TIP RIGHTWARDS
        0x0000, # G+8870  U+0000  controlNUL,  <control>
        0x0000, # G+8871  U+0000  controlNUL,  <control>
        0x0000, # G+8872  U+0000  controlNUL,  <control>
        0x0000, # G+8873  U+0000  controlNUL,  <control>
        0x0000, # G+8874  U+0000  controlNUL,  <control>
        0x0000, # G+8875  U+0000  controlNUL,  <control>
        0x0000, # G+8876  U+0000  controlNUL,  <control>
        0x0000, # G+8877  U+0000  controlNUL,  <control>
        0x0000, # G+8878  U+0000  controlNUL,  <control>
        0x0000, # G+8879  U+0000  controlNUL,  <control>
        0x0000, # G+8880  U+0000  controlNUL,  <control>
        0x0000, # G+8881  U+0000  controlNUL,  <control>
        0x0000, # G+8882  U+0000  controlNUL,  <control>
        0x0000, # G+8883  U+0000  controlNUL,  <control>
        0x21BC, # G+8884  U+21BC  harpoonleftbarbup,  LEFTWARDS HARPOON WITH BARB UPWARDS
        0x21C0, # G+8885  U+21C0  harpoonrightbarbup,  RIGHTWARDS HARPOON WITH BARB UPWARDS
        0x0000, # G+8886  U+0000  controlNUL,  <control>
        0x0000, # G+8887  U+0000  controlNUL,  <control>
        0x0000, # G+8888  U+0000  controlNUL,  <control>
        0x0000, # G+8889  U+0000  controlNUL,  <control>
        0x0000, # G+8890  U+0000  controlNUL,  <control>
        0x0000, # G+8891  U+0000  controlNUL,  <control>
        0x0000, # G+8892  U+0000  controlNUL,  <control>
        0x0000, # G+8893  U+0000  controlNUL,  <control>
        0x0000, # G+8894  U+0000  controlNUL,  <control>
        0x0000, # G+8895  U+0000  controlNUL,  <control>
        0x21C4, # G+8896  U+21C4  arrowrightoverleft,  RIGHTWARDS ARROW OVER LEFTWARDS ARROW
        0x21C5, # G+8897  U+21C5  arrowupleftofdown,  UPWARDS ARROW LEFTWARDS OF DOWNWARDS ARROW
        0x0000, # G+8898  U+0000  controlNUL,  <control>
        0x0000, # G+8899  U+0000  controlNUL,  <control>
        0x0000, # G+8900  U+0000  controlNUL,  <control>
        0x0000, # G+8901  U+0000  controlNUL,  <control>
        0x0000, # G+8902  U+0000  controlNUL,  <control>
        0x0000, # G+8903  U+0000  controlNUL,  <control>
        0x0000, # G+8904  U+0000  controlNUL,  <control>
        0x0000, # G+8905  U+0000  controlNUL,  <control>
        0x0000, # G+8906  U+0000  controlNUL,  <control>
        0x0000, # G+8907  U+0000  controlNUL,  <control>
        0x0000, # G+8908  U+0000  controlNUL,  <control>
        0x0000, # G+8909  U+0000  controlNUL,  <control>
        0x0000, # G+8910  U+0000  controlNUL,  <control>
        0x0000, # G+8911  U+0000  controlNUL,  <control>
        0x0000, # G+8912  U+0000  controlNUL,  <control>
        0x0000, # G+8913  U+0000  controlNUL,  <control>
        0x0000, # G+8914  U+0000  controlNUL,  <control>
        0x0000, # G+8915  U+0000  controlNUL,  <control>
        0x0000, # G+8916  U+0000  controlNUL,  <control>
        0x0000, # G+8917  U+0000  controlNUL,  <control>
        0x0000, # G+8918  U+0000  controlNUL,  <control>
        0x0000, # G+8919  U+0000  controlNUL,  <control>
        0x0000, # G+8920  U+0000  controlNUL,  <control>
        0x0000, # G+8921  U+0000  controlNUL,  <control>
        0x0000, # G+8922  U+0000  controlNUL,  <control>
        0x0000, # G+8923  U+0000  controlNUL,  <control>
        0x0000, # G+8924  U+0000  controlNUL,  <control>
        0x0000, # G+8925  U+0000  controlNUL,  <control>
        0x0000, # G+8926  U+0000  controlNUL,  <control>
        0x0000, # G+8927  U+0000  controlNUL,  <control>
        0x0000, # G+8928  U+0000  controlNUL,  <control>
        0x0000, # G+8929  U+0000  controlNUL,  <control>
        0x0000, # G+8930  U+0000  controlNUL,  <control>
        0x0000, # G+8931  U+0000  controlNUL,  <control>
        0x0000, # G+8932  U+0000  controlNUL,  <control>
        0x0000, # G+8933  U+0000  controlNUL,  <control>
        0x0000, # G+8934  U+0000  controlNUL,  <control>
        0x0000, # G+8935  U+0000  controlNUL,  <control>
        0x0000, # G+8936  U+0000  controlNUL,  <control>
        0x0000, # G+8937  U+0000  controlNUL,  <control>
        0x0000, # G+8938  U+0000  controlNUL,  <control>
        0x0000, # G+8939  U+0000  controlNUL,  <control>
        0x0000, # G+8940  U+0000  controlNUL,  <control>
        0x0000, # G+8941  U+0000  controlNUL,  <control>
        0x0000, # G+8942  U+0000  controlNUL,  <control>
        0x0000, # G+8943  U+0000  controlNUL,  <control>
        0x0000, # G+8944  U+0000  controlNUL,  <control>
        0x0000, # G+8945  U+0000  controlNUL,  <control>
        0x0000, # G+8946  U+0000  controlNUL,  <control>
        0x0000, # G+8947  U+0000  controlNUL,  <control>
        0x0000, # G+8948  U+0000  controlNUL,  <control>
        0x0000, # G+8949  U+0000  controlNUL,  <control>
        0x0000, # G+8950  U+0000  controlNUL,  <control>
        0x0000, # G+8951  U+0000  controlNUL,  <control>
        0x0000, # G+8952  U+0000  controlNUL,  <control>
        0x0000, # G+8953  U+0000  controlNUL,  <control>
        0x0000, # G+8954  U+0000  controlNUL,  <control>
        0x0000, # G+8955  U+0000  controlNUL,  <control>
        0x0000, # G+8956  U+0000  controlNUL,  <control>
        0x0000, # G+8957  U+0000  controlNUL,  <control>
        0x0000, # G+8958  U+0000  controlNUL,  <control>
        0x0000, # G+8959  U+0000  controlNUL,  <control>
        0x0000, # G+8960  U+0000  controlNUL,  <control>
        0x0000, # G+8961  U+0000  controlNUL,  <control>
        0x0000, # G+8962  U+0000  controlNUL,  <control>
        0x0000, # G+8963  U+0000  controlNUL,  <control>
        0x0000, # G+8964  U+0000  controlNUL,  <control>
        0x0000, # G+8965  U+0000  controlNUL,  <control>
        0x0000, # G+8966  U+0000  controlNUL,  <control>
        0x0000, # G+8967  U+0000  controlNUL,  <control>
        0x0000, # G+8968  U+0000  controlNUL,  <control>
        0x0000, # G+8969  U+0000  controlNUL,  <control>
        0x0000, # G+8970  U+0000  controlNUL,  <control>
        0x0000, # G+8971  U+0000  controlNUL,  <control>
        0x0000, # G+8972  U+0000  controlNUL,  <control>
        0x0000, # G+8973  U+0000  controlNUL,  <control>
        0x0000, # G+8974  U+0000  controlNUL,  <control>
        0x0000, # G+8975  U+0000  controlNUL,  <control>
        0x0000, # G+8976  U+0000  controlNUL,  <control>
        0x0000, # G+8977  U+0000  controlNUL,  <control>
        0x0000, # G+8978  U+0000  controlNUL,  <control>
        0x0000, # G+8979  U+0000  controlNUL,  <control>
        0x0000, # G+8980  U+0000  controlNUL,  <control>
        0x0000, # G+8981  U+0000  controlNUL,  <control>
        0x0000, # G+8982  U+0000  controlNUL,  <control>
        0x0000, # G+8983  U+0000  controlNUL,  <control>
        0x0000, # G+8984  U+0000  controlNUL,  <control>
        0x0000, # G+8985  U+0000  controlNUL,  <control>
        0x0000, # G+8986  U+0000  controlNUL,  <control>
        0x0000, # G+8987  U+0000  controlNUL,  <control>
        0x0000, # G+8988  U+0000  controlNUL,  <control>
        0x0000, # G+8989  U+0000  controlNUL,  <control>
        0x0000, # G+8990  U+0000  controlNUL,  <control>
        0x0000, # G+8991  U+0000  controlNUL,  <control>
        0x0000, # G+8992  U+0000  controlNUL,  <control>
        0x0000, # G+8993  U+0000  controlNUL,  <control>
        0x0000, # G+8994  U+0000  controlNUL,  <control>
        0x0000, # G+8995  U+0000  controlNUL,  <control>
        0x0000, # G+8996  U+0000  controlNUL,  <control>
        0x0000, # G+8997  U+0000  controlNUL,  <control>
        0x0000, # G+8998  U+0000  controlNUL,  <control>
        0x0000, # G+8999  U+0000  controlNUL,  <control>
        0x0000, # G+9000  U+0000  controlNUL,  <control>
        0x0000, # G+9001  U+0000  controlNUL,  <control>
        0x0000, # G+9002  U+0000  controlNUL,  <control>
        0x0000, # G+9003  U+0000  controlNUL,  <control>
        0x0000, # G+9004  U+0000  controlNUL,  <control>
        0x0000, # G+9005  U+0000  controlNUL,  <control>
        0x0000, # G+9006  U+0000  controlNUL,  <control>
        0x0000, # G+9007  U+0000  controlNUL,  <control>
        0x0000, # G+9008  U+0000  controlNUL,  <control>
        0x0000, # G+9009  U+0000  controlNUL,  <control>
        0x0000, # G+9010  U+0000  controlNUL,  <control>
        0x0000, # G+9011  U+0000  controlNUL,  <control>
        0x0000, # G+9012  U+0000  controlNUL,  <control>
        0x0000, # G+9013  U+0000  controlNUL,  <control>
        0x0000, # G+9014  U+0000  controlNUL,  <control>
        0x0000, # G+9015  U+0000  controlNUL,  <control>
        0x0000, # G+9016  U+0000  controlNUL,  <control>
        0x0000, # G+9017  U+0000  controlNUL,  <control>
        0x0000, # G+9018  U+0000  controlNUL,  <control>
        0x0000, # G+9019  U+0000  controlNUL,  <control>
        0x0000, # G+9020  U+0000  controlNUL,  <control>
        0x0000, # G+9021  U+0000  controlNUL,  <control>
        0x0000, # G+9022  U+0000  controlNUL,  <control>
        0x0000, # G+9023  U+0000  controlNUL,  <control>
        0x0000, # G+9024  U+0000  controlNUL,  <control>
        0x0000, # G+9025  U+0000  controlNUL,  <control>
        0x0000, # G+9026  U+0000  controlNUL,  <control>
        0x0000, # G+9027  U+0000  controlNUL,  <control>
        0x0000, # G+9028  U+0000  controlNUL,  <control>
        0x0000, # G+9029  U+0000  controlNUL,  <control>
        0x0000, # G+9030  U+0000  controlNUL,  <control>
        0x0000, # G+9031  U+0000  controlNUL,  <control>
        0x0000, # G+9032  U+0000  controlNUL,  <control>
        0x0000, # G+9033  U+0000  controlNUL,  <control>
        0x0000, # G+9034  U+0000  controlNUL,  <control>
        0x0000, # G+9035  U+0000  controlNUL,  <control>
        0x0000, # G+9036  U+0000  controlNUL,  <control>
        0x0000, # G+9037  U+0000  controlNUL,  <control>
        0x0000, # G+9038  U+0000  controlNUL,  <control>
        0x0000, # G+9039  U+0000  controlNUL,  <control>
        0x0000, # G+9040  U+0000  controlNUL,  <control>
        0x0000, # G+9041  U+0000  controlNUL,  <control>
        0x2483, # G+9042  U+2483  sixteenparen,  PARENTHESIZED NUMBER SIXTEEN
        0x2484, # G+9043  U+2484  seventeenparen,  PARENTHESIZED NUMBER SEVENTEEN
        0x2485, # G+9044  U+2485  eighteenparen,  PARENTHESIZED NUMBER EIGHTEEN
        0x2486, # G+9045  U+2486  nineteenparen,  PARENTHESIZED NUMBER NINETEEN
        0x2487, # G+9046  U+2487  twentyparen,  PARENTHESIZED NUMBER TWENTY
        0x0000, # G+9047  U+0000  controlNUL,  <control>
        0x0000, # G+9048  U+0000  controlNUL,  <control>
        0x0000, # G+9049  U+0000  controlNUL,  <control>
        0x0000, # G+9050  U+0000  controlNUL,  <control>
        0x0000, # G+9051  U+0000  controlNUL,  <control>
        0x0000, # G+9052  U+0000  controlNUL,  <control>
        0x0000, # G+9053  U+0000  controlNUL,  <control>
        0x0000, # G+9054  U+0000  controlNUL,  <control>
        0x0000, # G+9055  U+0000  controlNUL,  <control>
        0x0000, # G+9056  U+0000  controlNUL,  <control>
        0x0000, # G+9057  U+0000  controlNUL,  <control>
        0x0000, # G+9058  U+0000  controlNUL,  <control>
        0x0000, # G+9059  U+0000  controlNUL,  <control>
        0x0000, # G+9060  U+0000  controlNUL,  <control>
        0x0000, # G+9061  U+0000  controlNUL,  <control>
        0x0000, # G+9062  U+0000  controlNUL,  <control>
        0x0000, # G+9063  U+0000  controlNUL,  <control>
        0x0000, # G+9064  U+0000  controlNUL,  <control>
        0x0000, # G+9065  U+0000  controlNUL,  <control>
        0x0000, # G+9066  U+0000  controlNUL,  <control>
        0x0000, # G+9067  U+0000  controlNUL,  <control>
        0x0000, # G+9068  U+0000  controlNUL,  <control>
        0x0000, # G+9069  U+0000  controlNUL,  <control>
        0x0000, # G+9070  U+0000  controlNUL,  <control>
        0x0000, # G+9071  U+0000  controlNUL,  <control>
        0x0000, # G+9072  U+0000  controlNUL,  <control>
        0x0000, # G+9073  U+0000  controlNUL,  <control>
        0x0000, # G+9074  U+0000  controlNUL,  <control>
        0x0000, # G+9075  U+0000  controlNUL,  <control>
        0x0000, # G+9076  U+0000  controlNUL,  <control>
        0x0000, # G+9077  U+0000  controlNUL,  <control>
        0x0000, # G+9078  U+0000  controlNUL,  <control>
        0x0000, # G+9079  U+0000  controlNUL,  <control>
        0x3294, # G+9080  U+3294  ideographnamecircle,  CIRCLED IDEOGRAPH NAME
        0x0000, # G+9081  U+0000  controlNUL,  <control>
        0x0000, # G+9082  U+0000  controlNUL,  <control>
        0x0000, # G+9083  U+0000  controlNUL,  <control>
        0x0000, # G+9084  U+0000  controlNUL,  <control>
        0x0000, # G+9085  U+0000  controlNUL,  <control>
        0x0000, # G+9086  U+0000  controlNUL,  <control>
        0x0000, # G+9087  U+0000  controlNUL,  <control>
        0x0000, # G+9088  U+0000  controlNUL,  <control>
        0x0000, # G+9089  U+0000  controlNUL,  <control>
        0x0000, # G+9090  U+0000  controlNUL,  <control>
        0x0000, # G+9091  U+0000  controlNUL,  <control>
        0x0000, # G+9092  U+0000  controlNUL,  <control>
        0x0000, # G+9093  U+0000  controlNUL,  <control>
        0x0000, # G+9094  U+0000  controlNUL,  <control>
        0x0000, # G+9095  U+0000  controlNUL,  <control>
        0x32A5, # G+9096  U+32A5  ideographiccentrecircle,  CIRCLED IDEOGRAPH CENTRE
        0x0000, # G+9097  U+0000  controlNUL,  <control>
        0x0000, # G+9098  U+0000  controlNUL,  <control>
        0x0000, # G+9099  U+0000  controlNUL,  <control>
        0x0000, # G+9100  U+0000  controlNUL,  <control>
        0x0000, # G+9101  U+0000  controlNUL,  <control>
        0x0000, # G+9102  U+0000  controlNUL,  <control>
        0x0000, # G+9103  U+0000  controlNUL,  <control>
        0x0000, # G+9104  U+0000  controlNUL,  <control>
        0x0000, # G+9105  U+0000  controlNUL,  <control>
        0x0000, # G+9106  U+0000  controlNUL,  <control>
        0x0000, # G+9107  U+0000  controlNUL,  <control>
        0x0000, # G+9108  U+0000  controlNUL,  <control>
        0x0000, # G+9109  U+0000  controlNUL,  <control>
        0x0000, # G+9110  U+0000  controlNUL,  <control>
        0x0000, # G+9111  U+0000  controlNUL,  <control>
        0x0000, # G+9112  U+0000  controlNUL,  <control>
        0x0000, # G+9113  U+0000  controlNUL,  <control>
        0x0000, # G+9114  U+0000  controlNUL,  <control>
        0x0000, # G+9115  U+0000  controlNUL,  <control>
        0x0000, # G+9116  U+0000  controlNUL,  <control>
        0x0000, # G+9117  U+0000  controlNUL,  <control>
        0x0000, # G+9118  U+0000  controlNUL,  <control>
        0x0000, # G+9119  U+0000  controlNUL,  <control>
        0x0000, # G+9120  U+0000  controlNUL,  <control>
        0x0000, # G+9121  U+0000  controlNUL,  <control>
        0x0000, # G+9122  U+0000  controlNUL,  <control>
        0x0000, # G+9123  U+0000  controlNUL,  <control>
        0x0000, # G+9124  U+0000  controlNUL,  <control>
        0x0000, # G+9125  U+0000  controlNUL,  <control>
        0x0000, # G+9126  U+0000  controlNUL,  <control>
        0x0000, # G+9127  U+0000  controlNUL,  <control>
        0x0000, # G+9128  U+0000  controlNUL,  <control>
        0x0000, # G+9129  U+0000  controlNUL,  <control>
        0x0000, # G+9130  U+0000  controlNUL,  <control>
        0x0000, # G+9131  U+0000  controlNUL,  <control>
        0x0000, # G+9132  U+0000  controlNUL,  <control>
        0x0000, # G+9133  U+0000  controlNUL,  <control>
        0x0000, # G+9134  U+0000  controlNUL,  <control>
        0x0000, # G+9135  U+0000  controlNUL,  <control>
        0x0000, # G+9136  U+0000  controlNUL,  <control>
        0x0000, # G+9137  U+0000  controlNUL,  <control>
        0x0000, # G+9138  U+0000  controlNUL,  <control>
        0x0000, # G+9139  U+0000  controlNUL,  <control>
        0x0000, # G+9140  U+0000  controlNUL,  <control>
        0x0000, # G+9141  U+0000  controlNUL,  <control>
        0x0000, # G+9142  U+0000  controlNUL,  <control>
        0x0000, # G+9143  U+0000  controlNUL,  <control>
        0x0000, # G+9144  U+0000  controlNUL,  <control>
        0x0000, # G+9145  U+0000  controlNUL,  <control>
        0x0000, # G+9146  U+0000  controlNUL,  <control>
        0x0000, # G+9147  U+0000  controlNUL,  <control>
        0x0000, # G+9148  U+0000  controlNUL,  <control>
        0x0000, # G+9149  U+0000  controlNUL,  <control>
        0x0000, # G+9150  U+0000  controlNUL,  <control>
        0x0000, # G+9151  U+0000  controlNUL,  <control>
        0x0000, # G+9152  U+0000  controlNUL,  <control>
        0x0000, # G+9153  U+0000  controlNUL,  <control>
        0x0000, # G+9154  U+0000  controlNUL,  <control>
        0x0000, # G+9155  U+0000  controlNUL,  <control>
        0x0000, # G+9156  U+0000  controlNUL,  <control>
        0x0000, # G+9157  U+0000  controlNUL,  <control>
        0x0000, # G+9158  U+0000  controlNUL,  <control>
        0x0000, # G+9159  U+0000  controlNUL,  <control>
        0x0000, # G+9160  U+0000  controlNUL,  <control>
        0x0000, # G+9161  U+0000  controlNUL,  <control>
        0x0000, # G+9162  U+0000  controlNUL,  <control>
        0x0000, # G+9163  U+0000  controlNUL,  <control>
        0x0000, # G+9164  U+0000  controlNUL,  <control>
        0x0000, # G+9165  U+0000  controlNUL,  <control>
        0x0000, # G+9166  U+0000  controlNUL,  <control>
        0x0000, # G+9167  U+0000  controlNUL,  <control>
        0x0000, # G+9168  U+0000  controlNUL,  <control>
        0x0000, # G+9169  U+0000  controlNUL,  <control>
        0x0000, # G+9170  U+0000  controlNUL,  <control>
        0x0000, # G+9171  U+0000  controlNUL,  <control>
        0x0000, # G+9172  U+0000  controlNUL,  <control>
        0x0000, # G+9173  U+0000  controlNUL,  <control>
        0x0000, # G+9174  U+0000  controlNUL,  <control>
        0x0000, # G+9175  U+0000  controlNUL,  <control>
        0x0000, # G+9176  U+0000  controlNUL,  <control>
        0x0000, # G+9177  U+0000  controlNUL,  <control>
        0x0000, # G+9178  U+0000  controlNUL,  <control>
        0x0000, # G+9179  U+0000  controlNUL,  <control>
        0x0000, # G+9180  U+0000  controlNUL,  <control>
        0x0000, # G+9181  U+0000  controlNUL,  <control>
        0x0000, # G+9182  U+0000  controlNUL,  <control>
        0x0000, # G+9183  U+0000  controlNUL,  <control>
        0x0000, # G+9184  U+0000  controlNUL,  <control>
        0x0000, # G+9185  U+0000  controlNUL,  <control>
        0x0000, # G+9186  U+0000  controlNUL,  <control>
        0x0000, # G+9187  U+0000  controlNUL,  <control>
        0x0000, # G+9188  U+0000  controlNUL,  <control>
        0x0000, # G+9189  U+0000  controlNUL,  <control>
        0x21E0, # G+9190  U+21E0  arrowdashleft,  LEFTWARDS DASHED ARROW
        0x21E2, # G+9191  U+21E2  arrowdashright,  RIGHTWARDS DASHED ARROW
        0x21E1, # G+9192  U+21E1  arrowdashup,  UPWARDS DASHED ARROW
        0x21E3, # G+9193  U+21E3  arrowdashdown,  DOWNWARDS DASHED ARROW
        0x0000, # G+9194  U+0000  controlNUL,  <control>
        0x0000, # G+9195  U+0000  controlNUL,  <control>
        0x0000, # G+9196  U+0000  controlNUL,  <control>
        0x0000, # G+9197  U+0000  controlNUL,  <control>
        0x21E6, # G+9198  U+21E6  arrowleftwhite,  LEFTWARDS WHITE ARROW
        0x21E8, # G+9199  U+21E8  arrowrightwhite,  RIGHTWARDS WHITE ARROW
        0x21E7, # G+9200  U+21E7  arrowupwhite,  UPWARDS WHITE ARROW
        0x21E9, # G+9201  U+21E9  arrowdownwhite,  DOWNWARDS WHITE ARROW
        0x0000, # G+9202  U+0000  controlNUL,  <control>
        0x0000, # G+9203  U+0000  controlNUL,  <control>
        0x0000, # G+9204  U+0000  controlNUL,  <control>
        0x0000, # G+9205  U+0000  controlNUL,  <control>
        0x0000, # G+9206  U+0000  controlNUL,  <control>
        0x0000, # G+9207  U+0000  controlNUL,  <control>
        0x0000, # G+9208  U+0000  controlNUL,  <control>
        0x0000, # G+9209  U+0000  controlNUL,  <control>
        0x0000, # G+9210  U+0000  controlNUL,  <control>
        0x0000, # G+9211  U+0000  controlNUL,  <control>
        0x0000, # G+9212  U+0000  controlNUL,  <control>
        0x0000, # G+9213  U+0000  controlNUL,  <control>
        0x0000, # G+9214  U+0000  controlNUL,  <control>
        0x0000, # G+9215  U+0000  controlNUL,  <control>
        0x0000, # G+9216  U+0000  controlNUL,  <control>
        0x0000, # G+9217  U+0000  controlNUL,  <control>
        0x0000, # G+9218  U+0000  controlNUL,  <control>
        0x0000, # G+9219  U+0000  controlNUL,  <control>
        0x0000, # G+9220  U+0000  controlNUL,  <control>
        0x0000, # G+9221  U+0000  controlNUL,  <control>
        0x261D, # G+9222  U+261D  pointingindexupwhite,  WHITE UP POINTING INDEX
        0x261F, # G+9223  U+261F  pointingindexdownwhite,  WHITE DOWN POINTING INDEX
        0x0000, # G+9224  U+0000  controlNUL,  <control>
        0x0000, # G+9225  U+0000  controlNUL,  <control>
        0x0000, # G+9226  U+0000  controlNUL,  <control>
        0x0000, # G+9227  U+0000  controlNUL,  <control>
        0x0000, # G+9228  U+0000  controlNUL,  <control>
        0x0000, # G+9229  U+0000  controlNUL,  <control>
        0x0000, # G+9230  U+0000  controlNUL,  <control>
        0x0000, # G+9231  U+0000  controlNUL,  <control>
        0x0000, # G+9232  U+0000  controlNUL,  <control>
        0x0000, # G+9233  U+0000  controlNUL,  <control>
        0x0000, # G+9234  U+0000  controlNUL,  <control>
        0x0000, # G+9235  U+0000  controlNUL,  <control>
        0x0000, # G+9236  U+0000  controlNUL,  <control>
        0x0000, # G+9237  U+0000  controlNUL,  <control>
        0x0000, # G+9238  U+0000  controlNUL,  <control>
        0x0000, # G+9239  U+0000  controlNUL,  <control>
        0x0000, # G+9240  U+0000  controlNUL,  <control>
        0x0000, # G+9241  U+0000  controlNUL,  <control>
        0x0000, # G+9242  U+0000  controlNUL,  <control>
        0x0000, # G+9243  U+0000  controlNUL,  <control>
        0x0000, # G+9244  U+0000  controlNUL,  <control>
        0x0000, # G+9245  U+0000  controlNUL,  <control>
        0x0000, # G+9246  U+0000  controlNUL,  <control>
        0x0000, # G+9247  U+0000  controlNUL,  <control>
        0x0000, # G+9248  U+0000  controlNUL,  <control>
        0x0000, # G+9249  U+0000  controlNUL,  <control>
        0x0000, # G+9250  U+0000  controlNUL,  <control>
        0x0000, # G+9251  U+0000  controlNUL,  <control>
        0x0000, # G+9252  U+0000  controlNUL,  <control>
        0x0000, # G+9253  U+0000  controlNUL,  <control>
        0x0000, # G+9254  U+0000  controlNUL,  <control>
        0x0000, # G+9255  U+0000  controlNUL,  <control>
        0x0000, # G+9256  U+0000  controlNUL,  <control>
        0x0000, # G+9257  U+0000  controlNUL,  <control>
        0x0000, # G+9258  U+0000  controlNUL,  <control>
        0x0000, # G+9259  U+0000  controlNUL,  <control>
        0x0000, # G+9260  U+0000  controlNUL,  <control>
        0x0000, # G+9261  U+0000  controlNUL,  <control>
        0x0000, # G+9262  U+0000  controlNUL,  <control>
        0x0000, # G+9263  U+0000  controlNUL,  <control>
        0x0000, # G+9264  U+0000  controlNUL,  <control>
        0x0000, # G+9265  U+0000  controlNUL,  <control>
        0x0000, # G+9266  U+0000  controlNUL,  <control>
        0x0000, # G+9267  U+0000  controlNUL,  <control>
        0x0000, # G+9268  U+0000  controlNUL,  <control>
        0x0000, # G+9269  U+0000  controlNUL,  <control>
        0x0000, # G+9270  U+0000  controlNUL,  <control>
        0x0000, # G+9271  U+0000  controlNUL,  <control>
        0x0000, # G+9272  U+0000  controlNUL,  <control>
        0x0000, # G+9273  U+0000  controlNUL,  <control>
        0x0000, # G+9274  U+0000  controlNUL,  <control>
        0x0000, # G+9275  U+0000  controlNUL,  <control>
        0x0000, # G+9276  U+0000  controlNUL,  <control>
        0x0000, # G+9277  U+0000  controlNUL,  <control>
        0x0000, # G+9278  U+0000  controlNUL,  <control>
        0x0000, # G+9279  U+0000  controlNUL,  <control>
        0x0000, # G+9280  U+0000  controlNUL,  <control>
        0x0000, # G+9281  U+0000  controlNUL,  <control>
        0x0000, # G+9282  U+0000  controlNUL,  <control>
        0x0000, # G+9283  U+0000  controlNUL,  <control>
        0x0000, # G+9284  U+0000  controlNUL,  <control>
        0x0000, # G+9285  U+0000  controlNUL,  <control>
        0x0000, # G+9286  U+0000  controlNUL,  <control>
        0x0000, # G+9287  U+0000  controlNUL,  <control>
        0x0000, # G+9288  U+0000  controlNUL,  <control>
        0x0000, # G+9289  U+0000  controlNUL,  <control>
        0x0000, # G+9290  U+0000  controlNUL,  <control>
        0x0000, # G+9291  U+0000  controlNUL,  <control>
        0x0000, # G+9292  U+0000  controlNUL,  <control>
        0x0000, # G+9293  U+0000  controlNUL,  <control>
        0x0000, # G+9294  U+0000  controlNUL,  <control>
        0x0000, # G+9295  U+0000  controlNUL,  <control>
        0x0000, # G+9296  U+0000  controlNUL,  <control>
        0x0000, # G+9297  U+0000  controlNUL,  <control>
        0x0000, # G+9298  U+0000  controlNUL,  <control>
        0x0000, # G+9299  U+0000  controlNUL,  <control>
        0x3290, # G+9300  U+3290  ideographsuncircle,  CIRCLED IDEOGRAPH SUN
        0x328A, # G+9301  U+328A  ideographmooncircle,  CIRCLED IDEOGRAPH MOON
        0x328B, # G+9302  U+328B  ideographfirecircle,  CIRCLED IDEOGRAPH FIRE
        0x328C, # G+9303  U+328C  ideographwatercircle,  CIRCLED IDEOGRAPH WATER
        0x328D, # G+9304  U+328D  ideographwoodcircle,  CIRCLED IDEOGRAPH WOOD
        0x328E, # G+9305  U+328E  ideographmetalcircle,  CIRCLED IDEOGRAPH METAL
        0x328F, # G+9306  U+328F  ideographearthcircle,  CIRCLED IDEOGRAPH EARTH
        0x0000, # G+9307  U+0000  controlNUL,  <control>
        0x0000, # G+9308  U+0000  controlNUL,  <control>
        0x0000, # G+9309  U+0000  controlNUL,  <control>
        0x0000, # G+9310  U+0000  controlNUL,  <control>
        0x0000, # G+9311  U+0000  controlNUL,  <control>
        0x0000, # G+9312  U+0000  controlNUL,  <control>
        0x0000, # G+9313  U+0000  controlNUL,  <control>
        0x0000, # G+9314  U+0000  controlNUL,  <control>
        0x0000, # G+9315  U+0000  controlNUL,  <control>
        0x0000, # G+9316  U+0000  controlNUL,  <control>
        0x0000, # G+9317  U+0000  controlNUL,  <control>
        0x0000, # G+9318  U+0000  controlNUL,  <control>
        0x0000, # G+9319  U+0000  controlNUL,  <control>
        0x0000, # G+9320  U+0000  controlNUL,  <control>
        0x0000, # G+9321  U+0000  controlNUL,  <control>
        0x301E, # G+9322  U+301E  quotedblprime,  DOUBLE PRIME QUOTATION MARK
        0x301F, # G+9323  U+301F  uni301F,  LOW DOUBLE PRIME QUOTATION MARK
        0x2036, # G+9324  U+2036  uni2036,  REVERSED DOUBLE PRIME
        0x0000, # G+9325  U+0000  controlNUL,  <control>
        0x2035, # G+9326  U+2035  primereversed,  REVERSED PRIME
        0x0000, # G+9327  U+0000  controlNUL,  <control>
        0x0000, # G+9328  U+0000  controlNUL,  <control>
        0x0000, # G+9329  U+0000  controlNUL,  <control>
        0x30FC, # G+9330  U+30FC  prolongedkana,  KATAKANA-HIRAGANA PROLONGED SOUND MARK
        0x0000, # G+9331  U+0000  controlNUL,  <control>
        0x0000, # G+9332  U+0000  controlNUL,  <control>
        0xAC02, # G+9333  U+AC02  uniAC02,  HANGUL SYLLABLE-AC02
        0xAC03, # G+9334  U+AC03  uniAC03,  HANGUL SYLLABLE-AC03
        0xAC05, # G+9335  U+AC05  uniAC05,  HANGUL SYLLABLE-AC05
        0xAC06, # G+9336  U+AC06  uniAC06,  HANGUL SYLLABLE-AC06
        0xAC0B, # G+9337  U+AC0B  uniAC0B,  HANGUL SYLLABLE-AC0B
        0xAC0C, # G+9338  U+AC0C  uniAC0C,  HANGUL SYLLABLE-AC0C
        0xAC0D, # G+9339  U+AC0D  uniAC0D,  HANGUL SYLLABLE-AC0D
        0xAC0E, # G+9340  U+AC0E  uniAC0E,  HANGUL SYLLABLE-AC0E
        0xAC0F, # G+9341  U+AC0F  uniAC0F,  HANGUL SYLLABLE-AC0F
        0xAC18, # G+9342  U+AC18  uniAC18,  HANGUL SYLLABLE-AC18
        0xAC1E, # G+9343  U+AC1E  uniAC1E,  HANGUL SYLLABLE-AC1E
        0xAC1F, # G+9344  U+AC1F  uniAC1F,  HANGUL SYLLABLE-AC1F
        0xAC21, # G+9345  U+AC21  uniAC21,  HANGUL SYLLABLE-AC21
        0xAC22, # G+9346  U+AC22  uniAC22,  HANGUL SYLLABLE-AC22
        0xAC23, # G+9347  U+AC23  uniAC23,  HANGUL SYLLABLE-AC23
        0xAC25, # G+9348  U+AC25  uniAC25,  HANGUL SYLLABLE-AC25
        0xAC26, # G+9349  U+AC26  uniAC26,  HANGUL SYLLABLE-AC26
        0xAC27, # G+9350  U+AC27  uniAC27,  HANGUL SYLLABLE-AC27
        0xAC28, # G+9351  U+AC28  uniAC28,  HANGUL SYLLABLE-AC28
        0xAC29, # G+9352  U+AC29  uniAC29,  HANGUL SYLLABLE-AC29
        0xAC2A, # G+9353  U+AC2A  uniAC2A,  HANGUL SYLLABLE-AC2A
        0xAC2B, # G+9354  U+AC2B  uniAC2B,  HANGUL SYLLABLE-AC2B
        0xAC2E, # G+9355  U+AC2E  uniAC2E,  HANGUL SYLLABLE-AC2E
        0xAC32, # G+9356  U+AC32  uniAC32,  HANGUL SYLLABLE-AC32
        0xAC33, # G+9357  U+AC33  uniAC33,  HANGUL SYLLABLE-AC33
        0xAC34, # G+9358  U+AC34  uniAC34,  HANGUL SYLLABLE-AC34
        0xAC35, # G+9359  U+AC35  uniAC35,  HANGUL SYLLABLE-AC35
        0xAC36, # G+9360  U+AC36  uniAC36,  HANGUL SYLLABLE-AC36
        0xAC37, # G+9361  U+AC37  uniAC37,  HANGUL SYLLABLE-AC37
        0xAC3A, # G+9362  U+AC3A  uniAC3A,  HANGUL SYLLABLE-AC3A
        0xAC3B, # G+9363  U+AC3B  uniAC3B,  HANGUL SYLLABLE-AC3B
        0xAC3D, # G+9364  U+AC3D  uniAC3D,  HANGUL SYLLABLE-AC3D
        0xAC3E, # G+9365  U+AC3E  uniAC3E,  HANGUL SYLLABLE-AC3E
        0xAC3F, # G+9366  U+AC3F  uniAC3F,  HANGUL SYLLABLE-AC3F
        0xAC41, # G+9367  U+AC41  uniAC41,  HANGUL SYLLABLE-AC41
        0xAC42, # G+9368  U+AC42  uniAC42,  HANGUL SYLLABLE-AC42
        0xAC43, # G+9369  U+AC43  uniAC43,  HANGUL SYLLABLE-AC43
        0xAC44, # G+9370  U+AC44  uniAC44,  HANGUL SYLLABLE-AC44
        0xAC45, # G+9371  U+AC45  uniAC45,  HANGUL SYLLABLE-AC45
        0xAC46, # G+9372  U+AC46  uniAC46,  HANGUL SYLLABLE-AC46
        0xAC47, # G+9373  U+AC47  uniAC47,  HANGUL SYLLABLE-AC47
        0xAC48, # G+9374  U+AC48  uniAC48,  HANGUL SYLLABLE-AC48
        0xAC49, # G+9375  U+AC49  uniAC49,  HANGUL SYLLABLE-AC49
        0xAC4A, # G+9376  U+AC4A  uniAC4A,  HANGUL SYLLABLE-AC4A
        0xAC4C, # G+9377  U+AC4C  uniAC4C,  HANGUL SYLLABLE-AC4C
        0xAC4E, # G+9378  U+AC4E  uniAC4E,  HANGUL SYLLABLE-AC4E
        0xAC4F, # G+9379  U+AC4F  uniAC4F,  HANGUL SYLLABLE-AC4F
        0xAC50, # G+9380  U+AC50  uniAC50,  HANGUL SYLLABLE-AC50
        0xAC51, # G+9381  U+AC51  uniAC51,  HANGUL SYLLABLE-AC51
        0xAC52, # G+9382  U+AC52  uniAC52,  HANGUL SYLLABLE-AC52
        0xAC53, # G+9383  U+AC53  uniAC53,  HANGUL SYLLABLE-AC53
        0xAC55, # G+9384  U+AC55  uniAC55,  HANGUL SYLLABLE-AC55
        0xAC56, # G+9385  U+AC56  uniAC56,  HANGUL SYLLABLE-AC56
        0xAC57, # G+9386  U+AC57  uniAC57,  HANGUL SYLLABLE-AC57
        0xAC59, # G+9387  U+AC59  uniAC59,  HANGUL SYLLABLE-AC59
        0xAC5A, # G+9388  U+AC5A  uniAC5A,  HANGUL SYLLABLE-AC5A
        0xAC5B, # G+9389  U+AC5B  uniAC5B,  HANGUL SYLLABLE-AC5B
        0xAC5D, # G+9390  U+AC5D  uniAC5D,  HANGUL SYLLABLE-AC5D
        0xAC5E, # G+9391  U+AC5E  uniAC5E,  HANGUL SYLLABLE-AC5E
        0xAC5F, # G+9392  U+AC5F  uniAC5F,  HANGUL SYLLABLE-AC5F
        0xAC60, # G+9393  U+AC60  uniAC60,  HANGUL SYLLABLE-AC60
        0xAC61, # G+9394  U+AC61  uniAC61,  HANGUL SYLLABLE-AC61
        0xAC62, # G+9395  U+AC62  uniAC62,  HANGUL SYLLABLE-AC62
        0xAC63, # G+9396  U+AC63  uniAC63,  HANGUL SYLLABLE-AC63
        0xAC64, # G+9397  U+AC64  uniAC64,  HANGUL SYLLABLE-AC64
        0xAC65, # G+9398  U+AC65  uniAC65,  HANGUL SYLLABLE-AC65
        0xAC66, # G+9399  U+AC66  uniAC66,  HANGUL SYLLABLE-AC66
        0xAC67, # G+9400  U+AC67  uniAC67,  HANGUL SYLLABLE-AC67
        0xAC68, # G+9401  U+AC68  uniAC68,  HANGUL SYLLABLE-AC68
        0xAC69, # G+9402  U+AC69  uniAC69,  HANGUL SYLLABLE-AC69
        0xAC6A, # G+9403  U+AC6A  uniAC6A,  HANGUL SYLLABLE-AC6A
        0xAC6B, # G+9404  U+AC6B  uniAC6B,  HANGUL SYLLABLE-AC6B
        0xAC6C, # G+9405  U+AC6C  uniAC6C,  HANGUL SYLLABLE-AC6C
        0xAC6D, # G+9406  U+AC6D  uniAC6D,  HANGUL SYLLABLE-AC6D
        0xAC6E, # G+9407  U+AC6E  uniAC6E,  HANGUL SYLLABLE-AC6E
        0xAC6F, # G+9408  U+AC6F  uniAC6F,  HANGUL SYLLABLE-AC6F
        0xAC72, # G+9409  U+AC72  uniAC72,  HANGUL SYLLABLE-AC72
        0xAC73, # G+9410  U+AC73  uniAC73,  HANGUL SYLLABLE-AC73
        0xAC75, # G+9411  U+AC75  uniAC75,  HANGUL SYLLABLE-AC75
        0xAC76, # G+9412  U+AC76  uniAC76,  HANGUL SYLLABLE-AC76
        0xAC79, # G+9413  U+AC79  uniAC79,  HANGUL SYLLABLE-AC79
        0xAC7B, # G+9414  U+AC7B  uniAC7B,  HANGUL SYLLABLE-AC7B
        0xAC7C, # G+9415  U+AC7C  uniAC7C,  HANGUL SYLLABLE-AC7C
        0xAC7D, # G+9416  U+AC7D  uniAC7D,  HANGUL SYLLABLE-AC7D
        0xAC7E, # G+9417  U+AC7E  uniAC7E,  HANGUL SYLLABLE-AC7E
        0xAC7F, # G+9418  U+AC7F  uniAC7F,  HANGUL SYLLABLE-AC7F
        0xAC82, # G+9419  U+AC82  uniAC82,  HANGUL SYLLABLE-AC82
        0xAC87, # G+9420  U+AC87  uniAC87,  HANGUL SYLLABLE-AC87
        0xAC88, # G+9421  U+AC88  uniAC88,  HANGUL SYLLABLE-AC88
        0xAC8D, # G+9422  U+AC8D  uniAC8D,  HANGUL SYLLABLE-AC8D
        0xAC8E, # G+9423  U+AC8E  uniAC8E,  HANGUL SYLLABLE-AC8E
        0xAC8F, # G+9424  U+AC8F  uniAC8F,  HANGUL SYLLABLE-AC8F
        0xAC91, # G+9425  U+AC91  uniAC91,  HANGUL SYLLABLE-AC91
        0xAC92, # G+9426  U+AC92  uniAC92,  HANGUL SYLLABLE-AC92
        0xAC93, # G+9427  U+AC93  uniAC93,  HANGUL SYLLABLE-AC93
        0xAC95, # G+9428  U+AC95  uniAC95,  HANGUL SYLLABLE-AC95
        0xAC96, # G+9429  U+AC96  uniAC96,  HANGUL SYLLABLE-AC96
        0xAC97, # G+9430  U+AC97  uniAC97,  HANGUL SYLLABLE-AC97
        0xAC98, # G+9431  U+AC98  uniAC98,  HANGUL SYLLABLE-AC98
        0xAC99, # G+9432  U+AC99  uniAC99,  HANGUL SYLLABLE-AC99
        0xAC9A, # G+9433  U+AC9A  uniAC9A,  HANGUL SYLLABLE-AC9A
        0xAC9B, # G+9434  U+AC9B  uniAC9B,  HANGUL SYLLABLE-AC9B
        0xAC9E, # G+9435  U+AC9E  uniAC9E,  HANGUL SYLLABLE-AC9E
        0xACA2, # G+9436  U+ACA2  uniACA2,  HANGUL SYLLABLE-ACA2
        0xACA3, # G+9437  U+ACA3  uniACA3,  HANGUL SYLLABLE-ACA3
        0xACA4, # G+9438  U+ACA4  uniACA4,  HANGUL SYLLABLE-ACA4
        0xACA5, # G+9439  U+ACA5  uniACA5,  HANGUL SYLLABLE-ACA5
        0xACA6, # G+9440  U+ACA6  uniACA6,  HANGUL SYLLABLE-ACA6
        0xACA7, # G+9441  U+ACA7  uniACA7,  HANGUL SYLLABLE-ACA7
        0xACAB, # G+9442  U+ACAB  uniACAB,  HANGUL SYLLABLE-ACAB
        0xACAD, # G+9443  U+ACAD  uniACAD,  HANGUL SYLLABLE-ACAD
        0xACAE, # G+9444  U+ACAE  uniACAE,  HANGUL SYLLABLE-ACAE
        0xACB1, # G+9445  U+ACB1  uniACB1,  HANGUL SYLLABLE-ACB1
        0xACB2, # G+9446  U+ACB2  uniACB2,  HANGUL SYLLABLE-ACB2
        0xACB3, # G+9447  U+ACB3  uniACB3,  HANGUL SYLLABLE-ACB3
        0xACB4, # G+9448  U+ACB4  uniACB4,  HANGUL SYLLABLE-ACB4
        0xACB5, # G+9449  U+ACB5  uniACB5,  HANGUL SYLLABLE-ACB5
        0xACB6, # G+9450  U+ACB6  uniACB6,  HANGUL SYLLABLE-ACB6
        0xACB7, # G+9451  U+ACB7  uniACB7,  HANGUL SYLLABLE-ACB7
        0xACBA, # G+9452  U+ACBA  uniACBA,  HANGUL SYLLABLE-ACBA
        0xACBE, # G+9453  U+ACBE  uniACBE,  HANGUL SYLLABLE-ACBE
        0xACBF, # G+9454  U+ACBF  uniACBF,  HANGUL SYLLABLE-ACBF
        0xACC0, # G+9455  U+ACC0  uniACC0,  HANGUL SYLLABLE-ACC0
        0xACC2, # G+9456  U+ACC2  uniACC2,  HANGUL SYLLABLE-ACC2
        0xACC3, # G+9457  U+ACC3  uniACC3,  HANGUL SYLLABLE-ACC3
        0xACC5, # G+9458  U+ACC5  uniACC5,  HANGUL SYLLABLE-ACC5
        0xACC6, # G+9459  U+ACC6  uniACC6,  HANGUL SYLLABLE-ACC6
        0xACC7, # G+9460  U+ACC7  uniACC7,  HANGUL SYLLABLE-ACC7
        0xACC9, # G+9461  U+ACC9  uniACC9,  HANGUL SYLLABLE-ACC9
        0xACCA, # G+9462  U+ACCA  uniACCA,  HANGUL SYLLABLE-ACCA
        0xACCB, # G+9463  U+ACCB  uniACCB,  HANGUL SYLLABLE-ACCB
        0xACCD, # G+9464  U+ACCD  uniACCD,  HANGUL SYLLABLE-ACCD
        0xACCE, # G+9465  U+ACCE  uniACCE,  HANGUL SYLLABLE-ACCE
        0xACCF, # G+9466  U+ACCF  uniACCF,  HANGUL SYLLABLE-ACCF
        0xACD0, # G+9467  U+ACD0  uniACD0,  HANGUL SYLLABLE-ACD0
        0xACD1, # G+9468  U+ACD1  uniACD1,  HANGUL SYLLABLE-ACD1
        0xACD2, # G+9469  U+ACD2  uniACD2,  HANGUL SYLLABLE-ACD2
        0xACD3, # G+9470  U+ACD3  uniACD3,  HANGUL SYLLABLE-ACD3
        0xACD4, # G+9471  U+ACD4  uniACD4,  HANGUL SYLLABLE-ACD4
        0xACD6, # G+9472  U+ACD6  uniACD6,  HANGUL SYLLABLE-ACD6
        0xACD8, # G+9473  U+ACD8  uniACD8,  HANGUL SYLLABLE-ACD8
        0xACD9, # G+9474  U+ACD9  uniACD9,  HANGUL SYLLABLE-ACD9
        0xACDA, # G+9475  U+ACDA  uniACDA,  HANGUL SYLLABLE-ACDA
        0xACDB, # G+9476  U+ACDB  uniACDB,  HANGUL SYLLABLE-ACDB
        0xACDC, # G+9477  U+ACDC  uniACDC,  HANGUL SYLLABLE-ACDC
        0xACDD, # G+9478  U+ACDD  uniACDD,  HANGUL SYLLABLE-ACDD
        0xACDE, # G+9479  U+ACDE  uniACDE,  HANGUL SYLLABLE-ACDE
        0xACDF, # G+9480  U+ACDF  uniACDF,  HANGUL SYLLABLE-ACDF
        0xACE2, # G+9481  U+ACE2  uniACE2,  HANGUL SYLLABLE-ACE2
        0xACE3, # G+9482  U+ACE3  uniACE3,  HANGUL SYLLABLE-ACE3
        0xACE5, # G+9483  U+ACE5  uniACE5,  HANGUL SYLLABLE-ACE5
        0xACE6, # G+9484  U+ACE6  uniACE6,  HANGUL SYLLABLE-ACE6
        0xACE9, # G+9485  U+ACE9  uniACE9,  HANGUL SYLLABLE-ACE9
        0xACEB, # G+9486  U+ACEB  uniACEB,  HANGUL SYLLABLE-ACEB
        0xACED, # G+9487  U+ACED  uniACED,  HANGUL SYLLABLE-ACED
        0xACEE, # G+9488  U+ACEE  uniACEE,  HANGUL SYLLABLE-ACEE
        0xACF2, # G+9489  U+ACF2  uniACF2,  HANGUL SYLLABLE-ACF2
        0xACF4, # G+9490  U+ACF4  uniACF4,  HANGUL SYLLABLE-ACF4
        0xACF7, # G+9491  U+ACF7  uniACF7,  HANGUL SYLLABLE-ACF7
        0xACF8, # G+9492  U+ACF8  uniACF8,  HANGUL SYLLABLE-ACF8
        0xACF9, # G+9493  U+ACF9  uniACF9,  HANGUL SYLLABLE-ACF9
        0xACFA, # G+9494  U+ACFA  uniACFA,  HANGUL SYLLABLE-ACFA
        0xACFB, # G+9495  U+ACFB  uniACFB,  HANGUL SYLLABLE-ACFB
        0xACFE, # G+9496  U+ACFE  uniACFE,  HANGUL SYLLABLE-ACFE
        0xACFF, # G+9497  U+ACFF  uniACFF,  HANGUL SYLLABLE-ACFF
        0xAD01, # G+9498  U+AD01  uniAD01,  HANGUL SYLLABLE-AD01
        0xAD02, # G+9499  U+AD02  uniAD02,  HANGUL SYLLABLE-AD02
        0xAD03, # G+9500  U+AD03  uniAD03,  HANGUL SYLLABLE-AD03
        0xAD05, # G+9501  U+AD05  uniAD05,  HANGUL SYLLABLE-AD05
        0xAD07, # G+9502  U+AD07  uniAD07,  HANGUL SYLLABLE-AD07
        0xAD08, # G+9503  U+AD08  uniAD08,  HANGUL SYLLABLE-AD08
        0xAD09, # G+9504  U+AD09  uniAD09,  HANGUL SYLLABLE-AD09
        0xAD0A, # G+9505  U+AD0A  uniAD0A,  HANGUL SYLLABLE-AD0A
        0xAD0B, # G+9506  U+AD0B  uniAD0B,  HANGUL SYLLABLE-AD0B
        0xAD0E, # G+9507  U+AD0E  uniAD0E,  HANGUL SYLLABLE-AD0E
        0xAD10, # G+9508  U+AD10  uniAD10,  HANGUL SYLLABLE-AD10
        0xAD12, # G+9509  U+AD12  uniAD12,  HANGUL SYLLABLE-AD12
        0xAD13, # G+9510  U+AD13  uniAD13,  HANGUL SYLLABLE-AD13
        0xAD14, # G+9511  U+AD14  uniAD14,  HANGUL SYLLABLE-AD14
        0xAD15, # G+9512  U+AD15  uniAD15,  HANGUL SYLLABLE-AD15
        0xAD16, # G+9513  U+AD16  uniAD16,  HANGUL SYLLABLE-AD16
        0xAD17, # G+9514  U+AD17  uniAD17,  HANGUL SYLLABLE-AD17
        0xAD19, # G+9515  U+AD19  uniAD19,  HANGUL SYLLABLE-AD19
        0xAD1A, # G+9516  U+AD1A  uniAD1A,  HANGUL SYLLABLE-AD1A
        0xAD1B, # G+9517  U+AD1B  uniAD1B,  HANGUL SYLLABLE-AD1B
        0xAD1D, # G+9518  U+AD1D  uniAD1D,  HANGUL SYLLABLE-AD1D
        0xAD1E, # G+9519  U+AD1E  uniAD1E,  HANGUL SYLLABLE-AD1E
        0xAD1F, # G+9520  U+AD1F  uniAD1F,  HANGUL SYLLABLE-AD1F
        0xAD21, # G+9521  U+AD21  uniAD21,  HANGUL SYLLABLE-AD21
        0xAD22, # G+9522  U+AD22  uniAD22,  HANGUL SYLLABLE-AD22
        0xAD23, # G+9523  U+AD23  uniAD23,  HANGUL SYLLABLE-AD23
        0xAD24, # G+9524  U+AD24  uniAD24,  HANGUL SYLLABLE-AD24
        0xAD25, # G+9525  U+AD25  uniAD25,  HANGUL SYLLABLE-AD25
        0xAD26, # G+9526  U+AD26  uniAD26,  HANGUL SYLLABLE-AD26
        0xAD27, # G+9527  U+AD27  uniAD27,  HANGUL SYLLABLE-AD27
        0xAD28, # G+9528  U+AD28  uniAD28,  HANGUL SYLLABLE-AD28
        0xAD2A, # G+9529  U+AD2A  uniAD2A,  HANGUL SYLLABLE-AD2A
        0xAD2B, # G+9530  U+AD2B  uniAD2B,  HANGUL SYLLABLE-AD2B
        0xAD2E, # G+9531  U+AD2E  uniAD2E,  HANGUL SYLLABLE-AD2E
        0xAD2F, # G+9532  U+AD2F  uniAD2F,  HANGUL SYLLABLE-AD2F
        0xAD30, # G+9533  U+AD30  uniAD30,  HANGUL SYLLABLE-AD30
        0xAD31, # G+9534  U+AD31  uniAD31,  HANGUL SYLLABLE-AD31
        0xAD32, # G+9535  U+AD32  uniAD32,  HANGUL SYLLABLE-AD32
        0xAD33, # G+9536  U+AD33  uniAD33,  HANGUL SYLLABLE-AD33
        0xAD36, # G+9537  U+AD36  uniAD36,  HANGUL SYLLABLE-AD36
        0xAD37, # G+9538  U+AD37  uniAD37,  HANGUL SYLLABLE-AD37
        0xAD39, # G+9539  U+AD39  uniAD39,  HANGUL SYLLABLE-AD39
        0xAD3A, # G+9540  U+AD3A  uniAD3A,  HANGUL SYLLABLE-AD3A
        0xAD3B, # G+9541  U+AD3B  uniAD3B,  HANGUL SYLLABLE-AD3B
        0xAD3D, # G+9542  U+AD3D  uniAD3D,  HANGUL SYLLABLE-AD3D
        0xAD3E, # G+9543  U+AD3E  uniAD3E,  HANGUL SYLLABLE-AD3E
        0xAD3F, # G+9544  U+AD3F  uniAD3F,  HANGUL SYLLABLE-AD3F
        0xAD40, # G+9545  U+AD40  uniAD40,  HANGUL SYLLABLE-AD40
        0xAD41, # G+9546  U+AD41  uniAD41,  HANGUL SYLLABLE-AD41
        0xAD42, # G+9547  U+AD42  uniAD42,  HANGUL SYLLABLE-AD42
        0xAD43, # G+9548  U+AD43  uniAD43,  HANGUL SYLLABLE-AD43
        0xAD46, # G+9549  U+AD46  uniAD46,  HANGUL SYLLABLE-AD46
        0xAD48, # G+9550  U+AD48  uniAD48,  HANGUL SYLLABLE-AD48
        0xAD4A, # G+9551  U+AD4A  uniAD4A,  HANGUL SYLLABLE-AD4A
        0xAD4B, # G+9552  U+AD4B  uniAD4B,  HANGUL SYLLABLE-AD4B
        0xAD4C, # G+9553  U+AD4C  uniAD4C,  HANGUL SYLLABLE-AD4C
        0xAD4D, # G+9554  U+AD4D  uniAD4D,  HANGUL SYLLABLE-AD4D
        0xAD4E, # G+9555  U+AD4E  uniAD4E,  HANGUL SYLLABLE-AD4E
        0xAD4F, # G+9556  U+AD4F  uniAD4F,  HANGUL SYLLABLE-AD4F
        0xAD51, # G+9557  U+AD51  uniAD51,  HANGUL SYLLABLE-AD51
        0xAD52, # G+9558  U+AD52  uniAD52,  HANGUL SYLLABLE-AD52
        0xAD53, # G+9559  U+AD53  uniAD53,  HANGUL SYLLABLE-AD53
        0xAD55, # G+9560  U+AD55  uniAD55,  HANGUL SYLLABLE-AD55
        0xAD56, # G+9561  U+AD56  uniAD56,  HANGUL SYLLABLE-AD56
        0xAD57, # G+9562  U+AD57  uniAD57,  HANGUL SYLLABLE-AD57
        0xAD59, # G+9563  U+AD59  uniAD59,  HANGUL SYLLABLE-AD59
        0xAD5A, # G+9564  U+AD5A  uniAD5A,  HANGUL SYLLABLE-AD5A
        0xAD5B, # G+9565  U+AD5B  uniAD5B,  HANGUL SYLLABLE-AD5B
        0xAD5C, # G+9566  U+AD5C  uniAD5C,  HANGUL SYLLABLE-AD5C
        0xAD5D, # G+9567  U+AD5D  uniAD5D,  HANGUL SYLLABLE-AD5D
        0xAD5E, # G+9568  U+AD5E  uniAD5E,  HANGUL SYLLABLE-AD5E
        0xAD5F, # G+9569  U+AD5F  uniAD5F,  HANGUL SYLLABLE-AD5F
        0xAD60, # G+9570  U+AD60  uniAD60,  HANGUL SYLLABLE-AD60
        0xAD62, # G+9571  U+AD62  uniAD62,  HANGUL SYLLABLE-AD62
        0xAD64, # G+9572  U+AD64  uniAD64,  HANGUL SYLLABLE-AD64
        0xAD65, # G+9573  U+AD65  uniAD65,  HANGUL SYLLABLE-AD65
        0xAD66, # G+9574  U+AD66  uniAD66,  HANGUL SYLLABLE-AD66
        0xAD67, # G+9575  U+AD67  uniAD67,  HANGUL SYLLABLE-AD67
        0xAD68, # G+9576  U+AD68  uniAD68,  HANGUL SYLLABLE-AD68
        0xAD69, # G+9577  U+AD69  uniAD69,  HANGUL SYLLABLE-AD69
        0xAD6A, # G+9578  U+AD6A  uniAD6A,  HANGUL SYLLABLE-AD6A
        0xAD6B, # G+9579  U+AD6B  uniAD6B,  HANGUL SYLLABLE-AD6B
        0xAD6E, # G+9580  U+AD6E  uniAD6E,  HANGUL SYLLABLE-AD6E
        0xAD6F, # G+9581  U+AD6F  uniAD6F,  HANGUL SYLLABLE-AD6F
        0xAD71, # G+9582  U+AD71  uniAD71,  HANGUL SYLLABLE-AD71
        0xAD72, # G+9583  U+AD72  uniAD72,  HANGUL SYLLABLE-AD72
        0xAD77, # G+9584  U+AD77  uniAD77,  HANGUL SYLLABLE-AD77
        0xAD78, # G+9585  U+AD78  uniAD78,  HANGUL SYLLABLE-AD78
        0xAD79, # G+9586  U+AD79  uniAD79,  HANGUL SYLLABLE-AD79
        0xAD7A, # G+9587  U+AD7A  uniAD7A,  HANGUL SYLLABLE-AD7A
        0xAD7E, # G+9588  U+AD7E  uniAD7E,  HANGUL SYLLABLE-AD7E
        0xAD80, # G+9589  U+AD80  uniAD80,  HANGUL SYLLABLE-AD80
        0xAD83, # G+9590  U+AD83  uniAD83,  HANGUL SYLLABLE-AD83
        0xAD84, # G+9591  U+AD84  uniAD84,  HANGUL SYLLABLE-AD84
        0xAD85, # G+9592  U+AD85  uniAD85,  HANGUL SYLLABLE-AD85
        0xAD86, # G+9593  U+AD86  uniAD86,  HANGUL SYLLABLE-AD86
        0xAD87, # G+9594  U+AD87  uniAD87,  HANGUL SYLLABLE-AD87
        0xAD8A, # G+9595  U+AD8A  uniAD8A,  HANGUL SYLLABLE-AD8A
        0xAD8B, # G+9596  U+AD8B  uniAD8B,  HANGUL SYLLABLE-AD8B
        0xAD8D, # G+9597  U+AD8D  uniAD8D,  HANGUL SYLLABLE-AD8D
        0xAD8E, # G+9598  U+AD8E  uniAD8E,  HANGUL SYLLABLE-AD8E
        0xAD8F, # G+9599  U+AD8F  uniAD8F,  HANGUL SYLLABLE-AD8F
        0xAD91, # G+9600  U+AD91  uniAD91,  HANGUL SYLLABLE-AD91
        0xAD92, # G+9601  U+AD92  uniAD92,  HANGUL SYLLABLE-AD92
        0xAD93, # G+9602  U+AD93  uniAD93,  HANGUL SYLLABLE-AD93
        0xAD94, # G+9603  U+AD94  uniAD94,  HANGUL SYLLABLE-AD94
        0xAD95, # G+9604  U+AD95  uniAD95,  HANGUL SYLLABLE-AD95
        0xAD96, # G+9605  U+AD96  uniAD96,  HANGUL SYLLABLE-AD96
        0xAD97, # G+9606  U+AD97  uniAD97,  HANGUL SYLLABLE-AD97
        0xAD98, # G+9607  U+AD98  uniAD98,  HANGUL SYLLABLE-AD98
        0xAD99, # G+9608  U+AD99  uniAD99,  HANGUL SYLLABLE-AD99
        0xAD9A, # G+9609  U+AD9A  uniAD9A,  HANGUL SYLLABLE-AD9A
        0xAD9B, # G+9610  U+AD9B  uniAD9B,  HANGUL SYLLABLE-AD9B
        0xAD9E, # G+9611  U+AD9E  uniAD9E,  HANGUL SYLLABLE-AD9E
        0xAD9F, # G+9612  U+AD9F  uniAD9F,  HANGUL SYLLABLE-AD9F
        0xADA0, # G+9613  U+ADA0  uniADA0,  HANGUL SYLLABLE-ADA0
        0xADA1, # G+9614  U+ADA1  uniADA1,  HANGUL SYLLABLE-ADA1
        0xADA2, # G+9615  U+ADA2  uniADA2,  HANGUL SYLLABLE-ADA2
        0xADA3, # G+9616  U+ADA3  uniADA3,  HANGUL SYLLABLE-ADA3
        0xADA5, # G+9617  U+ADA5  uniADA5,  HANGUL SYLLABLE-ADA5
        0xADA6, # G+9618  U+ADA6  uniADA6,  HANGUL SYLLABLE-ADA6
        0xADA7, # G+9619  U+ADA7  uniADA7,  HANGUL SYLLABLE-ADA7
        0xADA8, # G+9620  U+ADA8  uniADA8,  HANGUL SYLLABLE-ADA8
        0xADA9, # G+9621  U+ADA9  uniADA9,  HANGUL SYLLABLE-ADA9
        0xADAA, # G+9622  U+ADAA  uniADAA,  HANGUL SYLLABLE-ADAA
        0xADAB, # G+9623  U+ADAB  uniADAB,  HANGUL SYLLABLE-ADAB
        0xADAC, # G+9624  U+ADAC  uniADAC,  HANGUL SYLLABLE-ADAC
        0xADAD, # G+9625  U+ADAD  uniADAD,  HANGUL SYLLABLE-ADAD
        0xADAE, # G+9626  U+ADAE  uniADAE,  HANGUL SYLLABLE-ADAE
        0xADAF, # G+9627  U+ADAF  uniADAF,  HANGUL SYLLABLE-ADAF
        0xADB0, # G+9628  U+ADB0  uniADB0,  HANGUL SYLLABLE-ADB0
        0xADB1, # G+9629  U+ADB1  uniADB1,  HANGUL SYLLABLE-ADB1
        0xADB2, # G+9630  U+ADB2  uniADB2,  HANGUL SYLLABLE-ADB2
        0xADB3, # G+9631  U+ADB3  uniADB3,  HANGUL SYLLABLE-ADB3
        0xADB4, # G+9632  U+ADB4  uniADB4,  HANGUL SYLLABLE-ADB4
        0xADB5, # G+9633  U+ADB5  uniADB5,  HANGUL SYLLABLE-ADB5
        0xADB6, # G+9634  U+ADB6  uniADB6,  HANGUL SYLLABLE-ADB6
        0xADB8, # G+9635  U+ADB8  uniADB8,  HANGUL SYLLABLE-ADB8
        0xADB9, # G+9636  U+ADB9  uniADB9,  HANGUL SYLLABLE-ADB9
        0xADBA, # G+9637  U+ADBA  uniADBA,  HANGUL SYLLABLE-ADBA
        0xADBB, # G+9638  U+ADBB  uniADBB,  HANGUL SYLLABLE-ADBB
        0xADBC, # G+9639  U+ADBC  uniADBC,  HANGUL SYLLABLE-ADBC
        0xADBD, # G+9640  U+ADBD  uniADBD,  HANGUL SYLLABLE-ADBD
        0xADBE, # G+9641  U+ADBE  uniADBE,  HANGUL SYLLABLE-ADBE
        0xADBF, # G+9642  U+ADBF  uniADBF,  HANGUL SYLLABLE-ADBF
        0xADC2, # G+9643  U+ADC2  uniADC2,  HANGUL SYLLABLE-ADC2
        0xADC3, # G+9644  U+ADC3  uniADC3,  HANGUL SYLLABLE-ADC3
        0xADC5, # G+9645  U+ADC5  uniADC5,  HANGUL SYLLABLE-ADC5
        0xADC6, # G+9646  U+ADC6  uniADC6,  HANGUL SYLLABLE-ADC6
        0xADC7, # G+9647  U+ADC7  uniADC7,  HANGUL SYLLABLE-ADC7
        0xADC9, # G+9648  U+ADC9  uniADC9,  HANGUL SYLLABLE-ADC9
        0xADCA, # G+9649  U+ADCA  uniADCA,  HANGUL SYLLABLE-ADCA
        0xADCB, # G+9650  U+ADCB  uniADCB,  HANGUL SYLLABLE-ADCB
        0xADCC, # G+9651  U+ADCC  uniADCC,  HANGUL SYLLABLE-ADCC
        0xADCD, # G+9652  U+ADCD  uniADCD,  HANGUL SYLLABLE-ADCD
        0xADCE, # G+9653  U+ADCE  uniADCE,  HANGUL SYLLABLE-ADCE
        0xADCF, # G+9654  U+ADCF  uniADCF,  HANGUL SYLLABLE-ADCF
        0xADD2, # G+9655  U+ADD2  uniADD2,  HANGUL SYLLABLE-ADD2
        0xADD4, # G+9656  U+ADD4  uniADD4,  HANGUL SYLLABLE-ADD4
        0xADD5, # G+9657  U+ADD5  uniADD5,  HANGUL SYLLABLE-ADD5
        0xADD6, # G+9658  U+ADD6  uniADD6,  HANGUL SYLLABLE-ADD6
        0xADD7, # G+9659  U+ADD7  uniADD7,  HANGUL SYLLABLE-ADD7
        0xADD8, # G+9660  U+ADD8  uniADD8,  HANGUL SYLLABLE-ADD8
        0xADD9, # G+9661  U+ADD9  uniADD9,  HANGUL SYLLABLE-ADD9
        0xADDA, # G+9662  U+ADDA  uniADDA,  HANGUL SYLLABLE-ADDA
        0xADDB, # G+9663  U+ADDB  uniADDB,  HANGUL SYLLABLE-ADDB
        0xADDD, # G+9664  U+ADDD  uniADDD,  HANGUL SYLLABLE-ADDD
        0xADDE, # G+9665  U+ADDE  uniADDE,  HANGUL SYLLABLE-ADDE
        0xADDF, # G+9666  U+ADDF  uniADDF,  HANGUL SYLLABLE-ADDF
        0xADE1, # G+9667  U+ADE1  uniADE1,  HANGUL SYLLABLE-ADE1
        0xADE2, # G+9668  U+ADE2  uniADE2,  HANGUL SYLLABLE-ADE2
        0xADE3, # G+9669  U+ADE3  uniADE3,  HANGUL SYLLABLE-ADE3
        0xADE5, # G+9670  U+ADE5  uniADE5,  HANGUL SYLLABLE-ADE5
        0xADE6, # G+9671  U+ADE6  uniADE6,  HANGUL SYLLABLE-ADE6
        0xADE7, # G+9672  U+ADE7  uniADE7,  HANGUL SYLLABLE-ADE7
        0xADE8, # G+9673  U+ADE8  uniADE8,  HANGUL SYLLABLE-ADE8
        0xADE9, # G+9674  U+ADE9  uniADE9,  HANGUL SYLLABLE-ADE9
        0xADEA, # G+9675  U+ADEA  uniADEA,  HANGUL SYLLABLE-ADEA
        0xADEB, # G+9676  U+ADEB  uniADEB,  HANGUL SYLLABLE-ADEB
        0xADEC, # G+9677  U+ADEC  uniADEC,  HANGUL SYLLABLE-ADEC
        0xADED, # G+9678  U+ADED  uniADED,  HANGUL SYLLABLE-ADED
        0xADEE, # G+9679  U+ADEE  uniADEE,  HANGUL SYLLABLE-ADEE
        0xADEF, # G+9680  U+ADEF  uniADEF,  HANGUL SYLLABLE-ADEF
        0xADF0, # G+9681  U+ADF0  uniADF0,  HANGUL SYLLABLE-ADF0
        0xADF1, # G+9682  U+ADF1  uniADF1,  HANGUL SYLLABLE-ADF1
        0xADF2, # G+9683  U+ADF2  uniADF2,  HANGUL SYLLABLE-ADF2
        0xADF3, # G+9684  U+ADF3  uniADF3,  HANGUL SYLLABLE-ADF3
        0xADF4, # G+9685  U+ADF4  uniADF4,  HANGUL SYLLABLE-ADF4
        0xADF5, # G+9686  U+ADF5  uniADF5,  HANGUL SYLLABLE-ADF5
        0xADF6, # G+9687  U+ADF6  uniADF6,  HANGUL SYLLABLE-ADF6
        0xADF7, # G+9688  U+ADF7  uniADF7,  HANGUL SYLLABLE-ADF7
        0xADFA, # G+9689  U+ADFA  uniADFA,  HANGUL SYLLABLE-ADFA
        0xADFB, # G+9690  U+ADFB  uniADFB,  HANGUL SYLLABLE-ADFB
        0xADFD, # G+9691  U+ADFD  uniADFD,  HANGUL SYLLABLE-ADFD
        0xADFE, # G+9692  U+ADFE  uniADFE,  HANGUL SYLLABLE-ADFE
        0xAE02, # G+9693  U+AE02  uniAE02,  HANGUL SYLLABLE-AE02
        0xAE03, # G+9694  U+AE03  uniAE03,  HANGUL SYLLABLE-AE03
        0xAE04, # G+9695  U+AE04  uniAE04,  HANGUL SYLLABLE-AE04
        0xAE05, # G+9696  U+AE05  uniAE05,  HANGUL SYLLABLE-AE05
        0xAE06, # G+9697  U+AE06  uniAE06,  HANGUL SYLLABLE-AE06
        0xAE07, # G+9698  U+AE07  uniAE07,  HANGUL SYLLABLE-AE07
        0xAE0A, # G+9699  U+AE0A  uniAE0A,  HANGUL SYLLABLE-AE0A
        0xAE0C, # G+9700  U+AE0C  uniAE0C,  HANGUL SYLLABLE-AE0C
        0xAE0E, # G+9701  U+AE0E  uniAE0E,  HANGUL SYLLABLE-AE0E
        0xAE0F, # G+9702  U+AE0F  uniAE0F,  HANGUL SYLLABLE-AE0F
        0xAE10, # G+9703  U+AE10  uniAE10,  HANGUL SYLLABLE-AE10
        0xAE11, # G+9704  U+AE11  uniAE11,  HANGUL SYLLABLE-AE11
        0xAE12, # G+9705  U+AE12  uniAE12,  HANGUL SYLLABLE-AE12
        0xAE13, # G+9706  U+AE13  uniAE13,  HANGUL SYLLABLE-AE13
        0xAE15, # G+9707  U+AE15  uniAE15,  HANGUL SYLLABLE-AE15
        0xAE16, # G+9708  U+AE16  uniAE16,  HANGUL SYLLABLE-AE16
        0xAE17, # G+9709  U+AE17  uniAE17,  HANGUL SYLLABLE-AE17
        0xAE18, # G+9710  U+AE18  uniAE18,  HANGUL SYLLABLE-AE18
        0xAE19, # G+9711  U+AE19  uniAE19,  HANGUL SYLLABLE-AE19
        0xAE1A, # G+9712  U+AE1A  uniAE1A,  HANGUL SYLLABLE-AE1A
        0xAE1B, # G+9713  U+AE1B  uniAE1B,  HANGUL SYLLABLE-AE1B
        0xAE1C, # G+9714  U+AE1C  uniAE1C,  HANGUL SYLLABLE-AE1C
        0xAE1D, # G+9715  U+AE1D  uniAE1D,  HANGUL SYLLABLE-AE1D
        0xAE1E, # G+9716  U+AE1E  uniAE1E,  HANGUL SYLLABLE-AE1E
        0xAE1F, # G+9717  U+AE1F  uniAE1F,  HANGUL SYLLABLE-AE1F
        0xAE20, # G+9718  U+AE20  uniAE20,  HANGUL SYLLABLE-AE20
        0xAE21, # G+9719  U+AE21  uniAE21,  HANGUL SYLLABLE-AE21
        0xAE22, # G+9720  U+AE22  uniAE22,  HANGUL SYLLABLE-AE22
        0xAE23, # G+9721  U+AE23  uniAE23,  HANGUL SYLLABLE-AE23
        0xAE24, # G+9722  U+AE24  uniAE24,  HANGUL SYLLABLE-AE24
        0xAE25, # G+9723  U+AE25  uniAE25,  HANGUL SYLLABLE-AE25
        0xAE26, # G+9724  U+AE26  uniAE26,  HANGUL SYLLABLE-AE26
        0xAE27, # G+9725  U+AE27  uniAE27,  HANGUL SYLLABLE-AE27
        0xAE28, # G+9726  U+AE28  uniAE28,  HANGUL SYLLABLE-AE28
        0xAE29, # G+9727  U+AE29  uniAE29,  HANGUL SYLLABLE-AE29
        0xAE2A, # G+9728  U+AE2A  uniAE2A,  HANGUL SYLLABLE-AE2A
        0xAE2B, # G+9729  U+AE2B  uniAE2B,  HANGUL SYLLABLE-AE2B
        0xAE2C, # G+9730  U+AE2C  uniAE2C,  HANGUL SYLLABLE-AE2C
        0xAE2D, # G+9731  U+AE2D  uniAE2D,  HANGUL SYLLABLE-AE2D
        0xAE2E, # G+9732  U+AE2E  uniAE2E,  HANGUL SYLLABLE-AE2E
        0xAE2F, # G+9733  U+AE2F  uniAE2F,  HANGUL SYLLABLE-AE2F
        0xAE32, # G+9734  U+AE32  uniAE32,  HANGUL SYLLABLE-AE32
        0xAE33, # G+9735  U+AE33  uniAE33,  HANGUL SYLLABLE-AE33
        0xAE35, # G+9736  U+AE35  uniAE35,  HANGUL SYLLABLE-AE35
        0xAE36, # G+9737  U+AE36  uniAE36,  HANGUL SYLLABLE-AE36
        0xAE39, # G+9738  U+AE39  uniAE39,  HANGUL SYLLABLE-AE39
        0xAE3B, # G+9739  U+AE3B  uniAE3B,  HANGUL SYLLABLE-AE3B
        0xAE3C, # G+9740  U+AE3C  uniAE3C,  HANGUL SYLLABLE-AE3C
        0xAE3D, # G+9741  U+AE3D  uniAE3D,  HANGUL SYLLABLE-AE3D
        0xAE3E, # G+9742  U+AE3E  uniAE3E,  HANGUL SYLLABLE-AE3E
        0xAE3F, # G+9743  U+AE3F  uniAE3F,  HANGUL SYLLABLE-AE3F
        0xAE42, # G+9744  U+AE42  uniAE42,  HANGUL SYLLABLE-AE42
        0xAE44, # G+9745  U+AE44  uniAE44,  HANGUL SYLLABLE-AE44
        0xAE47, # G+9746  U+AE47  uniAE47,  HANGUL SYLLABLE-AE47
        0xAE48, # G+9747  U+AE48  uniAE48,  HANGUL SYLLABLE-AE48
        0xAE49, # G+9748  U+AE49  uniAE49,  HANGUL SYLLABLE-AE49
        0xAE4B, # G+9749  U+AE4B  uniAE4B,  HANGUL SYLLABLE-AE4B
        0xAE4F, # G+9750  U+AE4F  uniAE4F,  HANGUL SYLLABLE-AE4F
        0xAE51, # G+9751  U+AE51  uniAE51,  HANGUL SYLLABLE-AE51
        0xAE52, # G+9752  U+AE52  uniAE52,  HANGUL SYLLABLE-AE52
        0xAE53, # G+9753  U+AE53  uniAE53,  HANGUL SYLLABLE-AE53
        0xAE55, # G+9754  U+AE55  uniAE55,  HANGUL SYLLABLE-AE55
        0xAE57, # G+9755  U+AE57  uniAE57,  HANGUL SYLLABLE-AE57
        0xAE58, # G+9756  U+AE58  uniAE58,  HANGUL SYLLABLE-AE58
        0xAE59, # G+9757  U+AE59  uniAE59,  HANGUL SYLLABLE-AE59
        0xAE5A, # G+9758  U+AE5A  uniAE5A,  HANGUL SYLLABLE-AE5A
        0xAE5B, # G+9759  U+AE5B  uniAE5B,  HANGUL SYLLABLE-AE5B
        0xAE5E, # G+9760  U+AE5E  uniAE5E,  HANGUL SYLLABLE-AE5E
        0xAE62, # G+9761  U+AE62  uniAE62,  HANGUL SYLLABLE-AE62
        0xAE63, # G+9762  U+AE63  uniAE63,  HANGUL SYLLABLE-AE63
        0xAE64, # G+9763  U+AE64  uniAE64,  HANGUL SYLLABLE-AE64
        0xAE66, # G+9764  U+AE66  uniAE66,  HANGUL SYLLABLE-AE66
        0xAE67, # G+9765  U+AE67  uniAE67,  HANGUL SYLLABLE-AE67
        0xAE6A, # G+9766  U+AE6A  uniAE6A,  HANGUL SYLLABLE-AE6A
        0xAE6B, # G+9767  U+AE6B  uniAE6B,  HANGUL SYLLABLE-AE6B
        0xAE6D, # G+9768  U+AE6D  uniAE6D,  HANGUL SYLLABLE-AE6D
        0xAE6E, # G+9769  U+AE6E  uniAE6E,  HANGUL SYLLABLE-AE6E
        0xAE6F, # G+9770  U+AE6F  uniAE6F,  HANGUL SYLLABLE-AE6F
        0xAE71, # G+9771  U+AE71  uniAE71,  HANGUL SYLLABLE-AE71
        0xAE72, # G+9772  U+AE72  uniAE72,  HANGUL SYLLABLE-AE72
        0xAE73, # G+9773  U+AE73  uniAE73,  HANGUL SYLLABLE-AE73
        0xAE74, # G+9774  U+AE74  uniAE74,  HANGUL SYLLABLE-AE74
        0xAE75, # G+9775  U+AE75  uniAE75,  HANGUL SYLLABLE-AE75
        0xAE76, # G+9776  U+AE76  uniAE76,  HANGUL SYLLABLE-AE76
        0xAE77, # G+9777  U+AE77  uniAE77,  HANGUL SYLLABLE-AE77
        0xAE7A, # G+9778  U+AE7A  uniAE7A,  HANGUL SYLLABLE-AE7A
        0xAE7E, # G+9779  U+AE7E  uniAE7E,  HANGUL SYLLABLE-AE7E
        0xAE7F, # G+9780  U+AE7F  uniAE7F,  HANGUL SYLLABLE-AE7F
        0xAE80, # G+9781  U+AE80  uniAE80,  HANGUL SYLLABLE-AE80
        0xAE81, # G+9782  U+AE81  uniAE81,  HANGUL SYLLABLE-AE81
        0xAE82, # G+9783  U+AE82  uniAE82,  HANGUL SYLLABLE-AE82
        0xAE83, # G+9784  U+AE83  uniAE83,  HANGUL SYLLABLE-AE83
        0xAE86, # G+9785  U+AE86  uniAE86,  HANGUL SYLLABLE-AE86
        0xAE87, # G+9786  U+AE87  uniAE87,  HANGUL SYLLABLE-AE87
        0xAE88, # G+9787  U+AE88  uniAE88,  HANGUL SYLLABLE-AE88
        0xAE89, # G+9788  U+AE89  uniAE89,  HANGUL SYLLABLE-AE89
        0xAE8A, # G+9789  U+AE8A  uniAE8A,  HANGUL SYLLABLE-AE8A
        0xAE8B, # G+9790  U+AE8B  uniAE8B,  HANGUL SYLLABLE-AE8B
        0xAE8D, # G+9791  U+AE8D  uniAE8D,  HANGUL SYLLABLE-AE8D
        0xAE8E, # G+9792  U+AE8E  uniAE8E,  HANGUL SYLLABLE-AE8E
        0xAE8F, # G+9793  U+AE8F  uniAE8F,  HANGUL SYLLABLE-AE8F
        0xAE90, # G+9794  U+AE90  uniAE90,  HANGUL SYLLABLE-AE90
        0xAE91, # G+9795  U+AE91  uniAE91,  HANGUL SYLLABLE-AE91
        0xAE92, # G+9796  U+AE92  uniAE92,  HANGUL SYLLABLE-AE92
        0xAE93, # G+9797  U+AE93  uniAE93,  HANGUL SYLLABLE-AE93
        0xAE94, # G+9798  U+AE94  uniAE94,  HANGUL SYLLABLE-AE94
        0xAE95, # G+9799  U+AE95  uniAE95,  HANGUL SYLLABLE-AE95
        0xAE96, # G+9800  U+AE96  uniAE96,  HANGUL SYLLABLE-AE96
        0xAE97, # G+9801  U+AE97  uniAE97,  HANGUL SYLLABLE-AE97
        0xAE98, # G+9802  U+AE98  uniAE98,  HANGUL SYLLABLE-AE98
        0xAE99, # G+9803  U+AE99  uniAE99,  HANGUL SYLLABLE-AE99
        0xAE9A, # G+9804  U+AE9A  uniAE9A,  HANGUL SYLLABLE-AE9A
        0xAE9B, # G+9805  U+AE9B  uniAE9B,  HANGUL SYLLABLE-AE9B
        0xAE9C, # G+9806  U+AE9C  uniAE9C,  HANGUL SYLLABLE-AE9C
        0xAE9D, # G+9807  U+AE9D  uniAE9D,  HANGUL SYLLABLE-AE9D
        0xAE9E, # G+9808  U+AE9E  uniAE9E,  HANGUL SYLLABLE-AE9E
        0xAE9F, # G+9809  U+AE9F  uniAE9F,  HANGUL SYLLABLE-AE9F
        0xAEA0, # G+9810  U+AEA0  uniAEA0,  HANGUL SYLLABLE-AEA0
        0xAEA1, # G+9811  U+AEA1  uniAEA1,  HANGUL SYLLABLE-AEA1
        0xAEA2, # G+9812  U+AEA2  uniAEA2,  HANGUL SYLLABLE-AEA2
        0xAEA3, # G+9813  U+AEA3  uniAEA3,  HANGUL SYLLABLE-AEA3
        0xAEA4, # G+9814  U+AEA4  uniAEA4,  HANGUL SYLLABLE-AEA4
        0xAEA5, # G+9815  U+AEA5  uniAEA5,  HANGUL SYLLABLE-AEA5
        0xAEA6, # G+9816  U+AEA6  uniAEA6,  HANGUL SYLLABLE-AEA6
        0xAEA7, # G+9817  U+AEA7  uniAEA7,  HANGUL SYLLABLE-AEA7
        0xAEA8, # G+9818  U+AEA8  uniAEA8,  HANGUL SYLLABLE-AEA8
        0xAEA9, # G+9819  U+AEA9  uniAEA9,  HANGUL SYLLABLE-AEA9
        0xAEAA, # G+9820  U+AEAA  uniAEAA,  HANGUL SYLLABLE-AEAA
        0xAEAB, # G+9821  U+AEAB  uniAEAB,  HANGUL SYLLABLE-AEAB
        0xAEAC, # G+9822  U+AEAC  uniAEAC,  HANGUL SYLLABLE-AEAC
        0xAEAD, # G+9823  U+AEAD  uniAEAD,  HANGUL SYLLABLE-AEAD
        0xAEAE, # G+9824  U+AEAE  uniAEAE,  HANGUL SYLLABLE-AEAE
        0xAEAF, # G+9825  U+AEAF  uniAEAF,  HANGUL SYLLABLE-AEAF
        0xAEB0, # G+9826  U+AEB0  uniAEB0,  HANGUL SYLLABLE-AEB0
        0xAEB1, # G+9827  U+AEB1  uniAEB1,  HANGUL SYLLABLE-AEB1
        0xAEB2, # G+9828  U+AEB2  uniAEB2,  HANGUL SYLLABLE-AEB2
        0xAEB3, # G+9829  U+AEB3  uniAEB3,  HANGUL SYLLABLE-AEB3
        0xAEB4, # G+9830  U+AEB4  uniAEB4,  HANGUL SYLLABLE-AEB4
        0xAEB5, # G+9831  U+AEB5  uniAEB5,  HANGUL SYLLABLE-AEB5
        0xAEB6, # G+9832  U+AEB6  uniAEB6,  HANGUL SYLLABLE-AEB6
        0xAEB7, # G+9833  U+AEB7  uniAEB7,  HANGUL SYLLABLE-AEB7
        0xAEB8, # G+9834  U+AEB8  uniAEB8,  HANGUL SYLLABLE-AEB8
        0xAEB9, # G+9835  U+AEB9  uniAEB9,  HANGUL SYLLABLE-AEB9
        0xAEBA, # G+9836  U+AEBA  uniAEBA,  HANGUL SYLLABLE-AEBA
        0xAEBB, # G+9837  U+AEBB  uniAEBB,  HANGUL SYLLABLE-AEBB
        0xAEBF, # G+9838  U+AEBF  uniAEBF,  HANGUL SYLLABLE-AEBF
        0xAEC1, # G+9839  U+AEC1  uniAEC1,  HANGUL SYLLABLE-AEC1
        0xAEC2, # G+9840  U+AEC2  uniAEC2,  HANGUL SYLLABLE-AEC2
        0xAEC3, # G+9841  U+AEC3  uniAEC3,  HANGUL SYLLABLE-AEC3
        0xAEC5, # G+9842  U+AEC5  uniAEC5,  HANGUL SYLLABLE-AEC5
        0xAEC6, # G+9843  U+AEC6  uniAEC6,  HANGUL SYLLABLE-AEC6
        0xAEC7, # G+9844  U+AEC7  uniAEC7,  HANGUL SYLLABLE-AEC7
        0xAEC8, # G+9845  U+AEC8  uniAEC8,  HANGUL SYLLABLE-AEC8
        0xAEC9, # G+9846  U+AEC9  uniAEC9,  HANGUL SYLLABLE-AEC9
        0xAECA, # G+9847  U+AECA  uniAECA,  HANGUL SYLLABLE-AECA
        0xAECB, # G+9848  U+AECB  uniAECB,  HANGUL SYLLABLE-AECB
        0xAECE, # G+9849  U+AECE  uniAECE,  HANGUL SYLLABLE-AECE
        0xAED2, # G+9850  U+AED2  uniAED2,  HANGUL SYLLABLE-AED2
        0xAED3, # G+9851  U+AED3  uniAED3,  HANGUL SYLLABLE-AED3
        0xAED4, # G+9852  U+AED4  uniAED4,  HANGUL SYLLABLE-AED4
        0xAED5, # G+9853  U+AED5  uniAED5,  HANGUL SYLLABLE-AED5
        0xAED6, # G+9854  U+AED6  uniAED6,  HANGUL SYLLABLE-AED6
        0xAED7, # G+9855  U+AED7  uniAED7,  HANGUL SYLLABLE-AED7
        0xAEDA, # G+9856  U+AEDA  uniAEDA,  HANGUL SYLLABLE-AEDA
        0xAEDB, # G+9857  U+AEDB  uniAEDB,  HANGUL SYLLABLE-AEDB
        0xAEDD, # G+9858  U+AEDD  uniAEDD,  HANGUL SYLLABLE-AEDD
        0xAEDE, # G+9859  U+AEDE  uniAEDE,  HANGUL SYLLABLE-AEDE
        0xAEDF, # G+9860  U+AEDF  uniAEDF,  HANGUL SYLLABLE-AEDF
        0xAEE0, # G+9861  U+AEE0  uniAEE0,  HANGUL SYLLABLE-AEE0
        0xAEE1, # G+9862  U+AEE1  uniAEE1,  HANGUL SYLLABLE-AEE1
        0xAEE2, # G+9863  U+AEE2  uniAEE2,  HANGUL SYLLABLE-AEE2
        0xAEE3, # G+9864  U+AEE3  uniAEE3,  HANGUL SYLLABLE-AEE3
        0xAEE4, # G+9865  U+AEE4  uniAEE4,  HANGUL SYLLABLE-AEE4
        0xAEE5, # G+9866  U+AEE5  uniAEE5,  HANGUL SYLLABLE-AEE5
        0xAEE6, # G+9867  U+AEE6  uniAEE6,  HANGUL SYLLABLE-AEE6
        0xAEE7, # G+9868  U+AEE7  uniAEE7,  HANGUL SYLLABLE-AEE7
        0xAEE9, # G+9869  U+AEE9  uniAEE9,  HANGUL SYLLABLE-AEE9
        0xAEEA, # G+9870  U+AEEA  uniAEEA,  HANGUL SYLLABLE-AEEA
        0xAEEC, # G+9871  U+AEEC  uniAEEC,  HANGUL SYLLABLE-AEEC
        0xAEEE, # G+9872  U+AEEE  uniAEEE,  HANGUL SYLLABLE-AEEE
        0xAEEF, # G+9873  U+AEEF  uniAEEF,  HANGUL SYLLABLE-AEEF
        0xAEF0, # G+9874  U+AEF0  uniAEF0,  HANGUL SYLLABLE-AEF0
        0xAEF1, # G+9875  U+AEF1  uniAEF1,  HANGUL SYLLABLE-AEF1
        0xAEF2, # G+9876  U+AEF2  uniAEF2,  HANGUL SYLLABLE-AEF2
        0xAEF3, # G+9877  U+AEF3  uniAEF3,  HANGUL SYLLABLE-AEF3
        0xAEF5, # G+9878  U+AEF5  uniAEF5,  HANGUL SYLLABLE-AEF5
        0xAEF6, # G+9879  U+AEF6  uniAEF6,  HANGUL SYLLABLE-AEF6
        0xAEF7, # G+9880  U+AEF7  uniAEF7,  HANGUL SYLLABLE-AEF7
        0xAEF9, # G+9881  U+AEF9  uniAEF9,  HANGUL SYLLABLE-AEF9
        0xAEFA, # G+9882  U+AEFA  uniAEFA,  HANGUL SYLLABLE-AEFA
        0xAEFB, # G+9883  U+AEFB  uniAEFB,  HANGUL SYLLABLE-AEFB
        0xAEFD, # G+9884  U+AEFD  uniAEFD,  HANGUL SYLLABLE-AEFD
        0xAEFE, # G+9885  U+AEFE  uniAEFE,  HANGUL SYLLABLE-AEFE
        0xAEFF, # G+9886  U+AEFF  uniAEFF,  HANGUL SYLLABLE-AEFF
        0xAF00, # G+9887  U+AF00  uniAF00,  HANGUL SYLLABLE-AF00
        0xAF01, # G+9888  U+AF01  uniAF01,  HANGUL SYLLABLE-AF01
        0xAF02, # G+9889  U+AF02  uniAF02,  HANGUL SYLLABLE-AF02
        0xAF03, # G+9890  U+AF03  uniAF03,  HANGUL SYLLABLE-AF03
        0xAF04, # G+9891  U+AF04  uniAF04,  HANGUL SYLLABLE-AF04
        0xAF05, # G+9892  U+AF05  uniAF05,  HANGUL SYLLABLE-AF05
        0xAF06, # G+9893  U+AF06  uniAF06,  HANGUL SYLLABLE-AF06
        0xAF09, # G+9894  U+AF09  uniAF09,  HANGUL SYLLABLE-AF09
        0xAF0A, # G+9895  U+AF0A  uniAF0A,  HANGUL SYLLABLE-AF0A
        0xAF0B, # G+9896  U+AF0B  uniAF0B,  HANGUL SYLLABLE-AF0B
        0xAF0C, # G+9897  U+AF0C  uniAF0C,  HANGUL SYLLABLE-AF0C
        0xAF0E, # G+9898  U+AF0E  uniAF0E,  HANGUL SYLLABLE-AF0E
        0xAF0F, # G+9899  U+AF0F  uniAF0F,  HANGUL SYLLABLE-AF0F
        0xAF11, # G+9900  U+AF11  uniAF11,  HANGUL SYLLABLE-AF11
        0xAF12, # G+9901  U+AF12  uniAF12,  HANGUL SYLLABLE-AF12
        0xAF13, # G+9902  U+AF13  uniAF13,  HANGUL SYLLABLE-AF13
        0xAF14, # G+9903  U+AF14  uniAF14,  HANGUL SYLLABLE-AF14
        0xAF15, # G+9904  U+AF15  uniAF15,  HANGUL SYLLABLE-AF15
        0xAF16, # G+9905  U+AF16  uniAF16,  HANGUL SYLLABLE-AF16
        0xAF17, # G+9906  U+AF17  uniAF17,  HANGUL SYLLABLE-AF17
        0xAF18, # G+9907  U+AF18  uniAF18,  HANGUL SYLLABLE-AF18
        0xAF19, # G+9908  U+AF19  uniAF19,  HANGUL SYLLABLE-AF19
        0xAF1A, # G+9909  U+AF1A  uniAF1A,  HANGUL SYLLABLE-AF1A
        0xAF1B, # G+9910  U+AF1B  uniAF1B,  HANGUL SYLLABLE-AF1B
        0xAF1C, # G+9911  U+AF1C  uniAF1C,  HANGUL SYLLABLE-AF1C
        0xAF1D, # G+9912  U+AF1D  uniAF1D,  HANGUL SYLLABLE-AF1D
        0xAF1E, # G+9913  U+AF1E  uniAF1E,  HANGUL SYLLABLE-AF1E
        0xAF1F, # G+9914  U+AF1F  uniAF1F,  HANGUL SYLLABLE-AF1F
        0xAF20, # G+9915  U+AF20  uniAF20,  HANGUL SYLLABLE-AF20
        0xAF21, # G+9916  U+AF21  uniAF21,  HANGUL SYLLABLE-AF21
        0xAF22, # G+9917  U+AF22  uniAF22,  HANGUL SYLLABLE-AF22
        0xAF23, # G+9918  U+AF23  uniAF23,  HANGUL SYLLABLE-AF23
        0xAF24, # G+9919  U+AF24  uniAF24,  HANGUL SYLLABLE-AF24
        0xAF25, # G+9920  U+AF25  uniAF25,  HANGUL SYLLABLE-AF25
        0xAF26, # G+9921  U+AF26  uniAF26,  HANGUL SYLLABLE-AF26
        0xAF27, # G+9922  U+AF27  uniAF27,  HANGUL SYLLABLE-AF27
        0xAF28, # G+9923  U+AF28  uniAF28,  HANGUL SYLLABLE-AF28
        0xAF29, # G+9924  U+AF29  uniAF29,  HANGUL SYLLABLE-AF29
        0xAF2A, # G+9925  U+AF2A  uniAF2A,  HANGUL SYLLABLE-AF2A
        0xAF2B, # G+9926  U+AF2B  uniAF2B,  HANGUL SYLLABLE-AF2B
        0xAF2E, # G+9927  U+AF2E  uniAF2E,  HANGUL SYLLABLE-AF2E
        0xAF2F, # G+9928  U+AF2F  uniAF2F,  HANGUL SYLLABLE-AF2F
        0xAF31, # G+9929  U+AF31  uniAF31,  HANGUL SYLLABLE-AF31
        0xAF33, # G+9930  U+AF33  uniAF33,  HANGUL SYLLABLE-AF33
        0xAF35, # G+9931  U+AF35  uniAF35,  HANGUL SYLLABLE-AF35
        0xAF36, # G+9932  U+AF36  uniAF36,  HANGUL SYLLABLE-AF36
        0xAF37, # G+9933  U+AF37  uniAF37,  HANGUL SYLLABLE-AF37
        0xAF38, # G+9934  U+AF38  uniAF38,  HANGUL SYLLABLE-AF38
        0xAF39, # G+9935  U+AF39  uniAF39,  HANGUL SYLLABLE-AF39
        0xAF3A, # G+9936  U+AF3A  uniAF3A,  HANGUL SYLLABLE-AF3A
        0xAF3B, # G+9937  U+AF3B  uniAF3B,  HANGUL SYLLABLE-AF3B
        0xAF3E, # G+9938  U+AF3E  uniAF3E,  HANGUL SYLLABLE-AF3E
        0xAF40, # G+9939  U+AF40  uniAF40,  HANGUL SYLLABLE-AF40
        0xAF44, # G+9940  U+AF44  uniAF44,  HANGUL SYLLABLE-AF44
        0xAF45, # G+9941  U+AF45  uniAF45,  HANGUL SYLLABLE-AF45
        0xAF46, # G+9942  U+AF46  uniAF46,  HANGUL SYLLABLE-AF46
        0xAF47, # G+9943  U+AF47  uniAF47,  HANGUL SYLLABLE-AF47
        0xAF4A, # G+9944  U+AF4A  uniAF4A,  HANGUL SYLLABLE-AF4A
        0xAF4B, # G+9945  U+AF4B  uniAF4B,  HANGUL SYLLABLE-AF4B
        0xAF4C, # G+9946  U+AF4C  uniAF4C,  HANGUL SYLLABLE-AF4C
        0xAF4D, # G+9947  U+AF4D  uniAF4D,  HANGUL SYLLABLE-AF4D
        0xAF4E, # G+9948  U+AF4E  uniAF4E,  HANGUL SYLLABLE-AF4E
        0xAF4F, # G+9949  U+AF4F  uniAF4F,  HANGUL SYLLABLE-AF4F
        0xAF51, # G+9950  U+AF51  uniAF51,  HANGUL SYLLABLE-AF51
        0xAF52, # G+9951  U+AF52  uniAF52,  HANGUL SYLLABLE-AF52
        0xAF53, # G+9952  U+AF53  uniAF53,  HANGUL SYLLABLE-AF53
        0xAF54, # G+9953  U+AF54  uniAF54,  HANGUL SYLLABLE-AF54
        0xAF55, # G+9954  U+AF55  uniAF55,  HANGUL SYLLABLE-AF55
        0xAF56, # G+9955  U+AF56  uniAF56,  HANGUL SYLLABLE-AF56
        0xAF57, # G+9956  U+AF57  uniAF57,  HANGUL SYLLABLE-AF57
        0xAF58, # G+9957  U+AF58  uniAF58,  HANGUL SYLLABLE-AF58
        0xAF59, # G+9958  U+AF59  uniAF59,  HANGUL SYLLABLE-AF59
        0xAF5A, # G+9959  U+AF5A  uniAF5A,  HANGUL SYLLABLE-AF5A
        0xAF5B, # G+9960  U+AF5B  uniAF5B,  HANGUL SYLLABLE-AF5B
        0xAF5E, # G+9961  U+AF5E  uniAF5E,  HANGUL SYLLABLE-AF5E
        0xAF5F, # G+9962  U+AF5F  uniAF5F,  HANGUL SYLLABLE-AF5F
        0xAF60, # G+9963  U+AF60  uniAF60,  HANGUL SYLLABLE-AF60
        0xAF61, # G+9964  U+AF61  uniAF61,  HANGUL SYLLABLE-AF61
        0xAF62, # G+9965  U+AF62  uniAF62,  HANGUL SYLLABLE-AF62
        0xAF63, # G+9966  U+AF63  uniAF63,  HANGUL SYLLABLE-AF63
        0xAF66, # G+9967  U+AF66  uniAF66,  HANGUL SYLLABLE-AF66
        0xAF67, # G+9968  U+AF67  uniAF67,  HANGUL SYLLABLE-AF67
        0xAF68, # G+9969  U+AF68  uniAF68,  HANGUL SYLLABLE-AF68
        0xAF69, # G+9970  U+AF69  uniAF69,  HANGUL SYLLABLE-AF69
        0xAF6A, # G+9971  U+AF6A  uniAF6A,  HANGUL SYLLABLE-AF6A
        0xAF6B, # G+9972  U+AF6B  uniAF6B,  HANGUL SYLLABLE-AF6B
        0xAF6C, # G+9973  U+AF6C  uniAF6C,  HANGUL SYLLABLE-AF6C
        0xAF6D, # G+9974  U+AF6D  uniAF6D,  HANGUL SYLLABLE-AF6D
        0xAF6E, # G+9975  U+AF6E  uniAF6E,  HANGUL SYLLABLE-AF6E
        0xAF6F, # G+9976  U+AF6F  uniAF6F,  HANGUL SYLLABLE-AF6F
        0xAF70, # G+9977  U+AF70  uniAF70,  HANGUL SYLLABLE-AF70
        0xAF71, # G+9978  U+AF71  uniAF71,  HANGUL SYLLABLE-AF71
        0xAF72, # G+9979  U+AF72  uniAF72,  HANGUL SYLLABLE-AF72
        0xAF73, # G+9980  U+AF73  uniAF73,  HANGUL SYLLABLE-AF73
        0xAF74, # G+9981  U+AF74  uniAF74,  HANGUL SYLLABLE-AF74
        0xAF75, # G+9982  U+AF75  uniAF75,  HANGUL SYLLABLE-AF75
        0xAF76, # G+9983  U+AF76  uniAF76,  HANGUL SYLLABLE-AF76
        0xAF77, # G+9984  U+AF77  uniAF77,  HANGUL SYLLABLE-AF77
        0xAF78, # G+9985  U+AF78  uniAF78,  HANGUL SYLLABLE-AF78
        0xAF7A, # G+9986  U+AF7A  uniAF7A,  HANGUL SYLLABLE-AF7A
        0xAF7B, # G+9987  U+AF7B  uniAF7B,  HANGUL SYLLABLE-AF7B
        0xAF7C, # G+9988  U+AF7C  uniAF7C,  HANGUL SYLLABLE-AF7C
        0xAF7D, # G+9989  U+AF7D  uniAF7D,  HANGUL SYLLABLE-AF7D
        0xAF7E, # G+9990  U+AF7E  uniAF7E,  HANGUL SYLLABLE-AF7E
        0xAF7F, # G+9991  U+AF7F  uniAF7F,  HANGUL SYLLABLE-AF7F
        0xAF81, # G+9992  U+AF81  uniAF81,  HANGUL SYLLABLE-AF81
        0xAF82, # G+9993  U+AF82  uniAF82,  HANGUL SYLLABLE-AF82
        0xAF83, # G+9994  U+AF83  uniAF83,  HANGUL SYLLABLE-AF83
        0xAF85, # G+9995  U+AF85  uniAF85,  HANGUL SYLLABLE-AF85
        0xAF86, # G+9996  U+AF86  uniAF86,  HANGUL SYLLABLE-AF86
        0xAF87, # G+9997  U+AF87  uniAF87,  HANGUL SYLLABLE-AF87
        0xAF89, # G+9998  U+AF89  uniAF89,  HANGUL SYLLABLE-AF89
        0xAF8A, # G+9999  U+AF8A  uniAF8A,  HANGUL SYLLABLE-AF8A
        0xAF8B, # G+10000  U+AF8B  uniAF8B,  HANGUL SYLLABLE-AF8B
        0xAF8C, # G+10001  U+AF8C  uniAF8C,  HANGUL SYLLABLE-AF8C
        0xAF8D, # G+10002  U+AF8D  uniAF8D,  HANGUL SYLLABLE-AF8D
        0xAF8E, # G+10003  U+AF8E  uniAF8E,  HANGUL SYLLABLE-AF8E
        0xAF8F, # G+10004  U+AF8F  uniAF8F,  HANGUL SYLLABLE-AF8F
        0xAF92, # G+10005  U+AF92  uniAF92,  HANGUL SYLLABLE-AF92
        0xAF93, # G+10006  U+AF93  uniAF93,  HANGUL SYLLABLE-AF93
        0xAF94, # G+10007  U+AF94  uniAF94,  HANGUL SYLLABLE-AF94
        0xAF96, # G+10008  U+AF96  uniAF96,  HANGUL SYLLABLE-AF96
        0xAF97, # G+10009  U+AF97  uniAF97,  HANGUL SYLLABLE-AF97
        0xAF98, # G+10010  U+AF98  uniAF98,  HANGUL SYLLABLE-AF98
        0xAF99, # G+10011  U+AF99  uniAF99,  HANGUL SYLLABLE-AF99
        0xAF9A, # G+10012  U+AF9A  uniAF9A,  HANGUL SYLLABLE-AF9A
        0xAF9B, # G+10013  U+AF9B  uniAF9B,  HANGUL SYLLABLE-AF9B
        0xAF9D, # G+10014  U+AF9D  uniAF9D,  HANGUL SYLLABLE-AF9D
        0xAF9E, # G+10015  U+AF9E  uniAF9E,  HANGUL SYLLABLE-AF9E
        0xAF9F, # G+10016  U+AF9F  uniAF9F,  HANGUL SYLLABLE-AF9F
        0xAFA0, # G+10017  U+AFA0  uniAFA0,  HANGUL SYLLABLE-AFA0
        0xAFA1, # G+10018  U+AFA1  uniAFA1,  HANGUL SYLLABLE-AFA1
        0xAFA2, # G+10019  U+AFA2  uniAFA2,  HANGUL SYLLABLE-AFA2
        0xAFA3, # G+10020  U+AFA3  uniAFA3,  HANGUL SYLLABLE-AFA3
        0xAFA4, # G+10021  U+AFA4  uniAFA4,  HANGUL SYLLABLE-AFA4
        0xAFA5, # G+10022  U+AFA5  uniAFA5,  HANGUL SYLLABLE-AFA5
        0xAFA6, # G+10023  U+AFA6  uniAFA6,  HANGUL SYLLABLE-AFA6
        0xAFA7, # G+10024  U+AFA7  uniAFA7,  HANGUL SYLLABLE-AFA7
        0xAFA8, # G+10025  U+AFA8  uniAFA8,  HANGUL SYLLABLE-AFA8
        0xAFA9, # G+10026  U+AFA9  uniAFA9,  HANGUL SYLLABLE-AFA9
        0xAFAA, # G+10027  U+AFAA  uniAFAA,  HANGUL SYLLABLE-AFAA
        0xAFAB, # G+10028  U+AFAB  uniAFAB,  HANGUL SYLLABLE-AFAB
        0xAFAC, # G+10029  U+AFAC  uniAFAC,  HANGUL SYLLABLE-AFAC
        0xAFAD, # G+10030  U+AFAD  uniAFAD,  HANGUL SYLLABLE-AFAD
        0xAFAE, # G+10031  U+AFAE  uniAFAE,  HANGUL SYLLABLE-AFAE
        0xAFAF, # G+10032  U+AFAF  uniAFAF,  HANGUL SYLLABLE-AFAF
        0xAFB0, # G+10033  U+AFB0  uniAFB0,  HANGUL SYLLABLE-AFB0
        0xAFB1, # G+10034  U+AFB1  uniAFB1,  HANGUL SYLLABLE-AFB1
        0xAFB2, # G+10035  U+AFB2  uniAFB2,  HANGUL SYLLABLE-AFB2
        0xAFB3, # G+10036  U+AFB3  uniAFB3,  HANGUL SYLLABLE-AFB3
        0xAFB4, # G+10037  U+AFB4  uniAFB4,  HANGUL SYLLABLE-AFB4
        0xAFB5, # G+10038  U+AFB5  uniAFB5,  HANGUL SYLLABLE-AFB5
        0xAFB6, # G+10039  U+AFB6  uniAFB6,  HANGUL SYLLABLE-AFB6
        0xAFB7, # G+10040  U+AFB7  uniAFB7,  HANGUL SYLLABLE-AFB7
        0xAFBA, # G+10041  U+AFBA  uniAFBA,  HANGUL SYLLABLE-AFBA
        0xAFBB, # G+10042  U+AFBB  uniAFBB,  HANGUL SYLLABLE-AFBB
        0xAFBD, # G+10043  U+AFBD  uniAFBD,  HANGUL SYLLABLE-AFBD
        0xAFBE, # G+10044  U+AFBE  uniAFBE,  HANGUL SYLLABLE-AFBE
        0xAFBF, # G+10045  U+AFBF  uniAFBF,  HANGUL SYLLABLE-AFBF
        0xAFC1, # G+10046  U+AFC1  uniAFC1,  HANGUL SYLLABLE-AFC1
        0xAFC2, # G+10047  U+AFC2  uniAFC2,  HANGUL SYLLABLE-AFC2
        0xAFC3, # G+10048  U+AFC3  uniAFC3,  HANGUL SYLLABLE-AFC3
        0xAFC4, # G+10049  U+AFC4  uniAFC4,  HANGUL SYLLABLE-AFC4
        0xAFC5, # G+10050  U+AFC5  uniAFC5,  HANGUL SYLLABLE-AFC5
        0xAFC6, # G+10051  U+AFC6  uniAFC6,  HANGUL SYLLABLE-AFC6
        0xAFCA, # G+10052  U+AFCA  uniAFCA,  HANGUL SYLLABLE-AFCA
        0xAFCC, # G+10053  U+AFCC  uniAFCC,  HANGUL SYLLABLE-AFCC
        0xAFCF, # G+10054  U+AFCF  uniAFCF,  HANGUL SYLLABLE-AFCF
        0xAFD0, # G+10055  U+AFD0  uniAFD0,  HANGUL SYLLABLE-AFD0
        0xAFD1, # G+10056  U+AFD1  uniAFD1,  HANGUL SYLLABLE-AFD1
        0xAFD2, # G+10057  U+AFD2  uniAFD2,  HANGUL SYLLABLE-AFD2
        0xAFD3, # G+10058  U+AFD3  uniAFD3,  HANGUL SYLLABLE-AFD3
        0xAFD5, # G+10059  U+AFD5  uniAFD5,  HANGUL SYLLABLE-AFD5
        0xAFD6, # G+10060  U+AFD6  uniAFD6,  HANGUL SYLLABLE-AFD6
        0xAFD7, # G+10061  U+AFD7  uniAFD7,  HANGUL SYLLABLE-AFD7
        0xAFD8, # G+10062  U+AFD8  uniAFD8,  HANGUL SYLLABLE-AFD8
        0xAFD9, # G+10063  U+AFD9  uniAFD9,  HANGUL SYLLABLE-AFD9
        0xAFDA, # G+10064  U+AFDA  uniAFDA,  HANGUL SYLLABLE-AFDA
        0xAFDB, # G+10065  U+AFDB  uniAFDB,  HANGUL SYLLABLE-AFDB
        0xAFDD, # G+10066  U+AFDD  uniAFDD,  HANGUL SYLLABLE-AFDD
        0xAFDE, # G+10067  U+AFDE  uniAFDE,  HANGUL SYLLABLE-AFDE
        0xAFDF, # G+10068  U+AFDF  uniAFDF,  HANGUL SYLLABLE-AFDF
        0xAFE0, # G+10069  U+AFE0  uniAFE0,  HANGUL SYLLABLE-AFE0
        0xAFE1, # G+10070  U+AFE1  uniAFE1,  HANGUL SYLLABLE-AFE1
        0xAFE2, # G+10071  U+AFE2  uniAFE2,  HANGUL SYLLABLE-AFE2
        0xAFE3, # G+10072  U+AFE3  uniAFE3,  HANGUL SYLLABLE-AFE3
        0xAFE4, # G+10073  U+AFE4  uniAFE4,  HANGUL SYLLABLE-AFE4
        0xAFE5, # G+10074  U+AFE5  uniAFE5,  HANGUL SYLLABLE-AFE5
        0xAFE6, # G+10075  U+AFE6  uniAFE6,  HANGUL SYLLABLE-AFE6
        0xAFE7, # G+10076  U+AFE7  uniAFE7,  HANGUL SYLLABLE-AFE7
        0xAFEA, # G+10077  U+AFEA  uniAFEA,  HANGUL SYLLABLE-AFEA
        0xAFEB, # G+10078  U+AFEB  uniAFEB,  HANGUL SYLLABLE-AFEB
        0xAFEC, # G+10079  U+AFEC  uniAFEC,  HANGUL SYLLABLE-AFEC
        0xAFED, # G+10080  U+AFED  uniAFED,  HANGUL SYLLABLE-AFED
        0xAFEE, # G+10081  U+AFEE  uniAFEE,  HANGUL SYLLABLE-AFEE
        0xAFEF, # G+10082  U+AFEF  uniAFEF,  HANGUL SYLLABLE-AFEF
        0xAFF2, # G+10083  U+AFF2  uniAFF2,  HANGUL SYLLABLE-AFF2
        0xAFF3, # G+10084  U+AFF3  uniAFF3,  HANGUL SYLLABLE-AFF3
        0xAFF5, # G+10085  U+AFF5  uniAFF5,  HANGUL SYLLABLE-AFF5
        0xAFF6, # G+10086  U+AFF6  uniAFF6,  HANGUL SYLLABLE-AFF6
        0xAFF7, # G+10087  U+AFF7  uniAFF7,  HANGUL SYLLABLE-AFF7
        0xAFF9, # G+10088  U+AFF9  uniAFF9,  HANGUL SYLLABLE-AFF9
        0xAFFA, # G+10089  U+AFFA  uniAFFA,  HANGUL SYLLABLE-AFFA
        0xAFFB, # G+10090  U+AFFB  uniAFFB,  HANGUL SYLLABLE-AFFB
        0xAFFC, # G+10091  U+AFFC  uniAFFC,  HANGUL SYLLABLE-AFFC
        0xAFFD, # G+10092  U+AFFD  uniAFFD,  HANGUL SYLLABLE-AFFD
        0xAFFE, # G+10093  U+AFFE  uniAFFE,  HANGUL SYLLABLE-AFFE
        0xAFFF, # G+10094  U+AFFF  uniAFFF,  HANGUL SYLLABLE-AFFF
        0xB002, # G+10095  U+B002  uniB002,  HANGUL SYLLABLE-B002
        0xB003, # G+10096  U+B003  uniB003,  HANGUL SYLLABLE-B003
        0xB005, # G+10097  U+B005  uniB005,  HANGUL SYLLABLE-B005
        0xB006, # G+10098  U+B006  uniB006,  HANGUL SYLLABLE-B006
        0xB007, # G+10099  U+B007  uniB007,  HANGUL SYLLABLE-B007
        0xB008, # G+10100  U+B008  uniB008,  HANGUL SYLLABLE-B008
        0xB009, # G+10101  U+B009  uniB009,  HANGUL SYLLABLE-B009
        0xB00A, # G+10102  U+B00A  uniB00A,  HANGUL SYLLABLE-B00A
        0xB00B, # G+10103  U+B00B  uniB00B,  HANGUL SYLLABLE-B00B
        0xB00D, # G+10104  U+B00D  uniB00D,  HANGUL SYLLABLE-B00D
        0xB00E, # G+10105  U+B00E  uniB00E,  HANGUL SYLLABLE-B00E
        0xB00F, # G+10106  U+B00F  uniB00F,  HANGUL SYLLABLE-B00F
        0xB011, # G+10107  U+B011  uniB011,  HANGUL SYLLABLE-B011
        0xB012, # G+10108  U+B012  uniB012,  HANGUL SYLLABLE-B012
        0xB013, # G+10109  U+B013  uniB013,  HANGUL SYLLABLE-B013
        0xB015, # G+10110  U+B015  uniB015,  HANGUL SYLLABLE-B015
        0xB016, # G+10111  U+B016  uniB016,  HANGUL SYLLABLE-B016
        0xB017, # G+10112  U+B017  uniB017,  HANGUL SYLLABLE-B017
        0xB018, # G+10113  U+B018  uniB018,  HANGUL SYLLABLE-B018
        0xB019, # G+10114  U+B019  uniB019,  HANGUL SYLLABLE-B019
        0xB01A, # G+10115  U+B01A  uniB01A,  HANGUL SYLLABLE-B01A
        0xB01B, # G+10116  U+B01B  uniB01B,  HANGUL SYLLABLE-B01B
        0xB01E, # G+10117  U+B01E  uniB01E,  HANGUL SYLLABLE-B01E
        0xB01F, # G+10118  U+B01F  uniB01F,  HANGUL SYLLABLE-B01F
        0xB020, # G+10119  U+B020  uniB020,  HANGUL SYLLABLE-B020
        0xB021, # G+10120  U+B021  uniB021,  HANGUL SYLLABLE-B021
        0xB022, # G+10121  U+B022  uniB022,  HANGUL SYLLABLE-B022
        0xB023, # G+10122  U+B023  uniB023,  HANGUL SYLLABLE-B023
        0xB024, # G+10123  U+B024  uniB024,  HANGUL SYLLABLE-B024
        0xB025, # G+10124  U+B025  uniB025,  HANGUL SYLLABLE-B025
        0xB026, # G+10125  U+B026  uniB026,  HANGUL SYLLABLE-B026
        0xB027, # G+10126  U+B027  uniB027,  HANGUL SYLLABLE-B027
        0xB029, # G+10127  U+B029  uniB029,  HANGUL SYLLABLE-B029
        0xB02A, # G+10128  U+B02A  uniB02A,  HANGUL SYLLABLE-B02A
        0xB02B, # G+10129  U+B02B  uniB02B,  HANGUL SYLLABLE-B02B
        0xB02C, # G+10130  U+B02C  uniB02C,  HANGUL SYLLABLE-B02C
        0xB02D, # G+10131  U+B02D  uniB02D,  HANGUL SYLLABLE-B02D
        0xB02E, # G+10132  U+B02E  uniB02E,  HANGUL SYLLABLE-B02E
        0xB02F, # G+10133  U+B02F  uniB02F,  HANGUL SYLLABLE-B02F
        0xB030, # G+10134  U+B030  uniB030,  HANGUL SYLLABLE-B030
        0xB031, # G+10135  U+B031  uniB031,  HANGUL SYLLABLE-B031
        0xB032, # G+10136  U+B032  uniB032,  HANGUL SYLLABLE-B032
        0xB033, # G+10137  U+B033  uniB033,  HANGUL SYLLABLE-B033
        0xB034, # G+10138  U+B034  uniB034,  HANGUL SYLLABLE-B034
        0xB035, # G+10139  U+B035  uniB035,  HANGUL SYLLABLE-B035
        0xB036, # G+10140  U+B036  uniB036,  HANGUL SYLLABLE-B036
        0xB037, # G+10141  U+B037  uniB037,  HANGUL SYLLABLE-B037
        0xB038, # G+10142  U+B038  uniB038,  HANGUL SYLLABLE-B038
        0xB039, # G+10143  U+B039  uniB039,  HANGUL SYLLABLE-B039
        0xB03A, # G+10144  U+B03A  uniB03A,  HANGUL SYLLABLE-B03A
        0xB03B, # G+10145  U+B03B  uniB03B,  HANGUL SYLLABLE-B03B
        0xB03C, # G+10146  U+B03C  uniB03C,  HANGUL SYLLABLE-B03C
        0xB03D, # G+10147  U+B03D  uniB03D,  HANGUL SYLLABLE-B03D
        0xB03E, # G+10148  U+B03E  uniB03E,  HANGUL SYLLABLE-B03E
        0xB03F, # G+10149  U+B03F  uniB03F,  HANGUL SYLLABLE-B03F
        0xB040, # G+10150  U+B040  uniB040,  HANGUL SYLLABLE-B040
        0xB041, # G+10151  U+B041  uniB041,  HANGUL SYLLABLE-B041
        0xB042, # G+10152  U+B042  uniB042,  HANGUL SYLLABLE-B042
        0xB043, # G+10153  U+B043  uniB043,  HANGUL SYLLABLE-B043
        0xB046, # G+10154  U+B046  uniB046,  HANGUL SYLLABLE-B046
        0xB047, # G+10155  U+B047  uniB047,  HANGUL SYLLABLE-B047
        0xB049, # G+10156  U+B049  uniB049,  HANGUL SYLLABLE-B049
        0xB04B, # G+10157  U+B04B  uniB04B,  HANGUL SYLLABLE-B04B
        0xB04D, # G+10158  U+B04D  uniB04D,  HANGUL SYLLABLE-B04D
        0xB04F, # G+10159  U+B04F  uniB04F,  HANGUL SYLLABLE-B04F
        0xB050, # G+10160  U+B050  uniB050,  HANGUL SYLLABLE-B050
        0xB051, # G+10161  U+B051  uniB051,  HANGUL SYLLABLE-B051
        0xB052, # G+10162  U+B052  uniB052,  HANGUL SYLLABLE-B052
        0xB056, # G+10163  U+B056  uniB056,  HANGUL SYLLABLE-B056
        0xB058, # G+10164  U+B058  uniB058,  HANGUL SYLLABLE-B058
        0xB05A, # G+10165  U+B05A  uniB05A,  HANGUL SYLLABLE-B05A
        0xB05B, # G+10166  U+B05B  uniB05B,  HANGUL SYLLABLE-B05B
        0xB05C, # G+10167  U+B05C  uniB05C,  HANGUL SYLLABLE-B05C
        0xB05E, # G+10168  U+B05E  uniB05E,  HANGUL SYLLABLE-B05E
        0xB05F, # G+10169  U+B05F  uniB05F,  HANGUL SYLLABLE-B05F
        0xB060, # G+10170  U+B060  uniB060,  HANGUL SYLLABLE-B060
        0xB061, # G+10171  U+B061  uniB061,  HANGUL SYLLABLE-B061
        0xB062, # G+10172  U+B062  uniB062,  HANGUL SYLLABLE-B062
        0xB063, # G+10173  U+B063  uniB063,  HANGUL SYLLABLE-B063
        0xB064, # G+10174  U+B064  uniB064,  HANGUL SYLLABLE-B064
        0xB065, # G+10175  U+B065  uniB065,  HANGUL SYLLABLE-B065
        0xB066, # G+10176  U+B066  uniB066,  HANGUL SYLLABLE-B066
        0xB067, # G+10177  U+B067  uniB067,  HANGUL SYLLABLE-B067
        0xB068, # G+10178  U+B068  uniB068,  HANGUL SYLLABLE-B068
        0xB069, # G+10179  U+B069  uniB069,  HANGUL SYLLABLE-B069
        0xB06A, # G+10180  U+B06A  uniB06A,  HANGUL SYLLABLE-B06A
        0xB06B, # G+10181  U+B06B  uniB06B,  HANGUL SYLLABLE-B06B
        0xB06C, # G+10182  U+B06C  uniB06C,  HANGUL SYLLABLE-B06C
        0xB06D, # G+10183  U+B06D  uniB06D,  HANGUL SYLLABLE-B06D
        0xB06E, # G+10184  U+B06E  uniB06E,  HANGUL SYLLABLE-B06E
        0xB06F, # G+10185  U+B06F  uniB06F,  HANGUL SYLLABLE-B06F
        0xB070, # G+10186  U+B070  uniB070,  HANGUL SYLLABLE-B070
        0xB071, # G+10187  U+B071  uniB071,  HANGUL SYLLABLE-B071
        0xB072, # G+10188  U+B072  uniB072,  HANGUL SYLLABLE-B072
        0xB073, # G+10189  U+B073  uniB073,  HANGUL SYLLABLE-B073
        0xB074, # G+10190  U+B074  uniB074,  HANGUL SYLLABLE-B074
        0xB075, # G+10191  U+B075  uniB075,  HANGUL SYLLABLE-B075
        0xB076, # G+10192  U+B076  uniB076,  HANGUL SYLLABLE-B076
        0xB077, # G+10193  U+B077  uniB077,  HANGUL SYLLABLE-B077
        0xB078, # G+10194  U+B078  uniB078,  HANGUL SYLLABLE-B078
        0xB079, # G+10195  U+B079  uniB079,  HANGUL SYLLABLE-B079
        0xB07A, # G+10196  U+B07A  uniB07A,  HANGUL SYLLABLE-B07A
        0xB07B, # G+10197  U+B07B  uniB07B,  HANGUL SYLLABLE-B07B
        0xB07E, # G+10198  U+B07E  uniB07E,  HANGUL SYLLABLE-B07E
        0xB07F, # G+10199  U+B07F  uniB07F,  HANGUL SYLLABLE-B07F
        0xB081, # G+10200  U+B081  uniB081,  HANGUL SYLLABLE-B081
        0xB082, # G+10201  U+B082  uniB082,  HANGUL SYLLABLE-B082
        0xB083, # G+10202  U+B083  uniB083,  HANGUL SYLLABLE-B083
        0xB085, # G+10203  U+B085  uniB085,  HANGUL SYLLABLE-B085
        0xB086, # G+10204  U+B086  uniB086,  HANGUL SYLLABLE-B086
        0xB087, # G+10205  U+B087  uniB087,  HANGUL SYLLABLE-B087
        0xB088, # G+10206  U+B088  uniB088,  HANGUL SYLLABLE-B088
        0xB089, # G+10207  U+B089  uniB089,  HANGUL SYLLABLE-B089
        0xB08A, # G+10208  U+B08A  uniB08A,  HANGUL SYLLABLE-B08A
        0xB08B, # G+10209  U+B08B  uniB08B,  HANGUL SYLLABLE-B08B
        0xB08E, # G+10210  U+B08E  uniB08E,  HANGUL SYLLABLE-B08E
        0xB090, # G+10211  U+B090  uniB090,  HANGUL SYLLABLE-B090
        0xB092, # G+10212  U+B092  uniB092,  HANGUL SYLLABLE-B092
        0xB093, # G+10213  U+B093  uniB093,  HANGUL SYLLABLE-B093
        0xB094, # G+10214  U+B094  uniB094,  HANGUL SYLLABLE-B094
        0xB095, # G+10215  U+B095  uniB095,  HANGUL SYLLABLE-B095
        0xB096, # G+10216  U+B096  uniB096,  HANGUL SYLLABLE-B096
        0xB097, # G+10217  U+B097  uniB097,  HANGUL SYLLABLE-B097
        0xB09B, # G+10218  U+B09B  uniB09B,  HANGUL SYLLABLE-B09B
        0xB09D, # G+10219  U+B09D  uniB09D,  HANGUL SYLLABLE-B09D
        0xB09E, # G+10220  U+B09E  uniB09E,  HANGUL SYLLABLE-B09E
        0xB0A3, # G+10221  U+B0A3  uniB0A3,  HANGUL SYLLABLE-B0A3
        0xB0A4, # G+10222  U+B0A4  uniB0A4,  HANGUL SYLLABLE-B0A4
        0xB0A5, # G+10223  U+B0A5  uniB0A5,  HANGUL SYLLABLE-B0A5
        0xB0A6, # G+10224  U+B0A6  uniB0A6,  HANGUL SYLLABLE-B0A6
        0xB0A7, # G+10225  U+B0A7  uniB0A7,  HANGUL SYLLABLE-B0A7
        0xB0AA, # G+10226  U+B0AA  uniB0AA,  HANGUL SYLLABLE-B0AA
        0xB0B0, # G+10227  U+B0B0  uniB0B0,  HANGUL SYLLABLE-B0B0
        0xB0B2, # G+10228  U+B0B2  uniB0B2,  HANGUL SYLLABLE-B0B2
        0xB0B6, # G+10229  U+B0B6  uniB0B6,  HANGUL SYLLABLE-B0B6
        0xB0B7, # G+10230  U+B0B7  uniB0B7,  HANGUL SYLLABLE-B0B7
        0xB0B9, # G+10231  U+B0B9  uniB0B9,  HANGUL SYLLABLE-B0B9
        0xB0BA, # G+10232  U+B0BA  uniB0BA,  HANGUL SYLLABLE-B0BA
        0xB0BB, # G+10233  U+B0BB  uniB0BB,  HANGUL SYLLABLE-B0BB
        0xB0BD, # G+10234  U+B0BD  uniB0BD,  HANGUL SYLLABLE-B0BD
        0xB0BE, # G+10235  U+B0BE  uniB0BE,  HANGUL SYLLABLE-B0BE
        0xB0BF, # G+10236  U+B0BF  uniB0BF,  HANGUL SYLLABLE-B0BF
        0xB0C0, # G+10237  U+B0C0  uniB0C0,  HANGUL SYLLABLE-B0C0
        0xB0C1, # G+10238  U+B0C1  uniB0C1,  HANGUL SYLLABLE-B0C1
        0xB0C2, # G+10239  U+B0C2  uniB0C2,  HANGUL SYLLABLE-B0C2
        0xB0C3, # G+10240  U+B0C3  uniB0C3,  HANGUL SYLLABLE-B0C3
        0xB0C6, # G+10241  U+B0C6  uniB0C6,  HANGUL SYLLABLE-B0C6
        0xB0CA, # G+10242  U+B0CA  uniB0CA,  HANGUL SYLLABLE-B0CA
        0xB0CB, # G+10243  U+B0CB  uniB0CB,  HANGUL SYLLABLE-B0CB
        0xB0CC, # G+10244  U+B0CC  uniB0CC,  HANGUL SYLLABLE-B0CC
        0xB0CD, # G+10245  U+B0CD  uniB0CD,  HANGUL SYLLABLE-B0CD
        0xB0CE, # G+10246  U+B0CE  uniB0CE,  HANGUL SYLLABLE-B0CE
        0xB0CF, # G+10247  U+B0CF  uniB0CF,  HANGUL SYLLABLE-B0CF
        0xB0D2, # G+10248  U+B0D2  uniB0D2,  HANGUL SYLLABLE-B0D2
        0xB0D3, # G+10249  U+B0D3  uniB0D3,  HANGUL SYLLABLE-B0D3
        0xB0D5, # G+10250  U+B0D5  uniB0D5,  HANGUL SYLLABLE-B0D5
        0xB0D6, # G+10251  U+B0D6  uniB0D6,  HANGUL SYLLABLE-B0D6
        0xB0D7, # G+10252  U+B0D7  uniB0D7,  HANGUL SYLLABLE-B0D7
        0xB0D9, # G+10253  U+B0D9  uniB0D9,  HANGUL SYLLABLE-B0D9
        0xB0DA, # G+10254  U+B0DA  uniB0DA,  HANGUL SYLLABLE-B0DA
        0xB0DB, # G+10255  U+B0DB  uniB0DB,  HANGUL SYLLABLE-B0DB
        0xB0DC, # G+10256  U+B0DC  uniB0DC,  HANGUL SYLLABLE-B0DC
        0xB0DD, # G+10257  U+B0DD  uniB0DD,  HANGUL SYLLABLE-B0DD
        0xB0DE, # G+10258  U+B0DE  uniB0DE,  HANGUL SYLLABLE-B0DE
        0xB0DF, # G+10259  U+B0DF  uniB0DF,  HANGUL SYLLABLE-B0DF
        0xB0E1, # G+10260  U+B0E1  uniB0E1,  HANGUL SYLLABLE-B0E1
        0xB0E2, # G+10261  U+B0E2  uniB0E2,  HANGUL SYLLABLE-B0E2
        0xB0E3, # G+10262  U+B0E3  uniB0E3,  HANGUL SYLLABLE-B0E3
        0xB0E4, # G+10263  U+B0E4  uniB0E4,  HANGUL SYLLABLE-B0E4
        0xB0E6, # G+10264  U+B0E6  uniB0E6,  HANGUL SYLLABLE-B0E6
        0xB0E7, # G+10265  U+B0E7  uniB0E7,  HANGUL SYLLABLE-B0E7
        0xB0E8, # G+10266  U+B0E8  uniB0E8,  HANGUL SYLLABLE-B0E8
        0xB0E9, # G+10267  U+B0E9  uniB0E9,  HANGUL SYLLABLE-B0E9
        0xB0EA, # G+10268  U+B0EA  uniB0EA,  HANGUL SYLLABLE-B0EA
        0xB0EB, # G+10269  U+B0EB  uniB0EB,  HANGUL SYLLABLE-B0EB
        0xB0EC, # G+10270  U+B0EC  uniB0EC,  HANGUL SYLLABLE-B0EC
        0xB0ED, # G+10271  U+B0ED  uniB0ED,  HANGUL SYLLABLE-B0ED
        0xB0EE, # G+10272  U+B0EE  uniB0EE,  HANGUL SYLLABLE-B0EE
        0xB0EF, # G+10273  U+B0EF  uniB0EF,  HANGUL SYLLABLE-B0EF
        0xB0F0, # G+10274  U+B0F0  uniB0F0,  HANGUL SYLLABLE-B0F0
        0xB0F1, # G+10275  U+B0F1  uniB0F1,  HANGUL SYLLABLE-B0F1
        0xB0F2, # G+10276  U+B0F2  uniB0F2,  HANGUL SYLLABLE-B0F2
        0xB0F3, # G+10277  U+B0F3  uniB0F3,  HANGUL SYLLABLE-B0F3
        0xB0F4, # G+10278  U+B0F4  uniB0F4,  HANGUL SYLLABLE-B0F4
        0xB0F5, # G+10279  U+B0F5  uniB0F5,  HANGUL SYLLABLE-B0F5
        0xB0F6, # G+10280  U+B0F6  uniB0F6,  HANGUL SYLLABLE-B0F6
        0xB0F7, # G+10281  U+B0F7  uniB0F7,  HANGUL SYLLABLE-B0F7
        0xB0F8, # G+10282  U+B0F8  uniB0F8,  HANGUL SYLLABLE-B0F8
        0xB0F9, # G+10283  U+B0F9  uniB0F9,  HANGUL SYLLABLE-B0F9
        0xB0FA, # G+10284  U+B0FA  uniB0FA,  HANGUL SYLLABLE-B0FA
        0xB0FB, # G+10285  U+B0FB  uniB0FB,  HANGUL SYLLABLE-B0FB
        0xB0FC, # G+10286  U+B0FC  uniB0FC,  HANGUL SYLLABLE-B0FC
        0xB0FD, # G+10287  U+B0FD  uniB0FD,  HANGUL SYLLABLE-B0FD
        0xB0FE, # G+10288  U+B0FE  uniB0FE,  HANGUL SYLLABLE-B0FE
        0xB0FF, # G+10289  U+B0FF  uniB0FF,  HANGUL SYLLABLE-B0FF
        0xB100, # G+10290  U+B100  uniB100,  HANGUL SYLLABLE-B100
        0xB101, # G+10291  U+B101  uniB101,  HANGUL SYLLABLE-B101
        0xB102, # G+10292  U+B102  uniB102,  HANGUL SYLLABLE-B102
        0xB103, # G+10293  U+B103  uniB103,  HANGUL SYLLABLE-B103
        0xB104, # G+10294  U+B104  uniB104,  HANGUL SYLLABLE-B104
        0xB105, # G+10295  U+B105  uniB105,  HANGUL SYLLABLE-B105
        0xB106, # G+10296  U+B106  uniB106,  HANGUL SYLLABLE-B106
        0xB107, # G+10297  U+B107  uniB107,  HANGUL SYLLABLE-B107
        0xB10A, # G+10298  U+B10A  uniB10A,  HANGUL SYLLABLE-B10A
        0xB10D, # G+10299  U+B10D  uniB10D,  HANGUL SYLLABLE-B10D
        0xB10E, # G+10300  U+B10E  uniB10E,  HANGUL SYLLABLE-B10E
        0xB10F, # G+10301  U+B10F  uniB10F,  HANGUL SYLLABLE-B10F
        0xB111, # G+10302  U+B111  uniB111,  HANGUL SYLLABLE-B111
        0xB114, # G+10303  U+B114  uniB114,  HANGUL SYLLABLE-B114
        0xB115, # G+10304  U+B115  uniB115,  HANGUL SYLLABLE-B115
        0xB116, # G+10305  U+B116  uniB116,  HANGUL SYLLABLE-B116
        0xB117, # G+10306  U+B117  uniB117,  HANGUL SYLLABLE-B117
        0xB11A, # G+10307  U+B11A  uniB11A,  HANGUL SYLLABLE-B11A
        0xB11E, # G+10308  U+B11E  uniB11E,  HANGUL SYLLABLE-B11E
        0xB11F, # G+10309  U+B11F  uniB11F,  HANGUL SYLLABLE-B11F
        0xB120, # G+10310  U+B120  uniB120,  HANGUL SYLLABLE-B120
        0xB121, # G+10311  U+B121  uniB121,  HANGUL SYLLABLE-B121
        0xB122, # G+10312  U+B122  uniB122,  HANGUL SYLLABLE-B122
        0xB126, # G+10313  U+B126  uniB126,  HANGUL SYLLABLE-B126
        0xB127, # G+10314  U+B127  uniB127,  HANGUL SYLLABLE-B127
        0xB129, # G+10315  U+B129  uniB129,  HANGUL SYLLABLE-B129
        0xB12A, # G+10316  U+B12A  uniB12A,  HANGUL SYLLABLE-B12A
        0xB12B, # G+10317  U+B12B  uniB12B,  HANGUL SYLLABLE-B12B
        0xB12D, # G+10318  U+B12D  uniB12D,  HANGUL SYLLABLE-B12D
        0xB12E, # G+10319  U+B12E  uniB12E,  HANGUL SYLLABLE-B12E
        0xB12F, # G+10320  U+B12F  uniB12F,  HANGUL SYLLABLE-B12F
        0xB130, # G+10321  U+B130  uniB130,  HANGUL SYLLABLE-B130
        0xB131, # G+10322  U+B131  uniB131,  HANGUL SYLLABLE-B131
        0xB132, # G+10323  U+B132  uniB132,  HANGUL SYLLABLE-B132
        0xB133, # G+10324  U+B133  uniB133,  HANGUL SYLLABLE-B133
        0xB136, # G+10325  U+B136  uniB136,  HANGUL SYLLABLE-B136
        0xB13A, # G+10326  U+B13A  uniB13A,  HANGUL SYLLABLE-B13A
        0xB13B, # G+10327  U+B13B  uniB13B,  HANGUL SYLLABLE-B13B
        0xB13C, # G+10328  U+B13C  uniB13C,  HANGUL SYLLABLE-B13C
        0xB13D, # G+10329  U+B13D  uniB13D,  HANGUL SYLLABLE-B13D
        0xB13E, # G+10330  U+B13E  uniB13E,  HANGUL SYLLABLE-B13E
        0xB13F, # G+10331  U+B13F  uniB13F,  HANGUL SYLLABLE-B13F
        0xB142, # G+10332  U+B142  uniB142,  HANGUL SYLLABLE-B142
        0xB143, # G+10333  U+B143  uniB143,  HANGUL SYLLABLE-B143
        0xB145, # G+10334  U+B145  uniB145,  HANGUL SYLLABLE-B145
        0xB146, # G+10335  U+B146  uniB146,  HANGUL SYLLABLE-B146
        0xB147, # G+10336  U+B147  uniB147,  HANGUL SYLLABLE-B147
        0xB149, # G+10337  U+B149  uniB149,  HANGUL SYLLABLE-B149
        0xB14A, # G+10338  U+B14A  uniB14A,  HANGUL SYLLABLE-B14A
        0xB14B, # G+10339  U+B14B  uniB14B,  HANGUL SYLLABLE-B14B
        0xB14C, # G+10340  U+B14C  uniB14C,  HANGUL SYLLABLE-B14C
        0xB14D, # G+10341  U+B14D  uniB14D,  HANGUL SYLLABLE-B14D
        0xB14E, # G+10342  U+B14E  uniB14E,  HANGUL SYLLABLE-B14E
        0xB14F, # G+10343  U+B14F  uniB14F,  HANGUL SYLLABLE-B14F
        0xB152, # G+10344  U+B152  uniB152,  HANGUL SYLLABLE-B152
        0xB153, # G+10345  U+B153  uniB153,  HANGUL SYLLABLE-B153
        0xB156, # G+10346  U+B156  uniB156,  HANGUL SYLLABLE-B156
        0xB157, # G+10347  U+B157  uniB157,  HANGUL SYLLABLE-B157
        0xB159, # G+10348  U+B159  uniB159,  HANGUL SYLLABLE-B159
        0xB15A, # G+10349  U+B15A  uniB15A,  HANGUL SYLLABLE-B15A
        0xB15B, # G+10350  U+B15B  uniB15B,  HANGUL SYLLABLE-B15B
        0xB15D, # G+10351  U+B15D  uniB15D,  HANGUL SYLLABLE-B15D
        0xB15E, # G+10352  U+B15E  uniB15E,  HANGUL SYLLABLE-B15E
        0xB15F, # G+10353  U+B15F  uniB15F,  HANGUL SYLLABLE-B15F
        0xB161, # G+10354  U+B161  uniB161,  HANGUL SYLLABLE-B161
        0xB162, # G+10355  U+B162  uniB162,  HANGUL SYLLABLE-B162
        0xB163, # G+10356  U+B163  uniB163,  HANGUL SYLLABLE-B163
        0xB164, # G+10357  U+B164  uniB164,  HANGUL SYLLABLE-B164
        0xB165, # G+10358  U+B165  uniB165,  HANGUL SYLLABLE-B165
        0xB166, # G+10359  U+B166  uniB166,  HANGUL SYLLABLE-B166
        0xB167, # G+10360  U+B167  uniB167,  HANGUL SYLLABLE-B167
        0xB168, # G+10361  U+B168  uniB168,  HANGUL SYLLABLE-B168
        0xB169, # G+10362  U+B169  uniB169,  HANGUL SYLLABLE-B169
        0xB16A, # G+10363  U+B16A  uniB16A,  HANGUL SYLLABLE-B16A
        0xB16B, # G+10364  U+B16B  uniB16B,  HANGUL SYLLABLE-B16B
        0xB16C, # G+10365  U+B16C  uniB16C,  HANGUL SYLLABLE-B16C
        0xB16D, # G+10366  U+B16D  uniB16D,  HANGUL SYLLABLE-B16D
        0xB16E, # G+10367  U+B16E  uniB16E,  HANGUL SYLLABLE-B16E
        0xB16F, # G+10368  U+B16F  uniB16F,  HANGUL SYLLABLE-B16F
        0xB170, # G+10369  U+B170  uniB170,  HANGUL SYLLABLE-B170
        0xB171, # G+10370  U+B171  uniB171,  HANGUL SYLLABLE-B171
        0xB172, # G+10371  U+B172  uniB172,  HANGUL SYLLABLE-B172
        0xB173, # G+10372  U+B173  uniB173,  HANGUL SYLLABLE-B173
        0xB174, # G+10373  U+B174  uniB174,  HANGUL SYLLABLE-B174
        0xB175, # G+10374  U+B175  uniB175,  HANGUL SYLLABLE-B175
        0xB176, # G+10375  U+B176  uniB176,  HANGUL SYLLABLE-B176
        0xB177, # G+10376  U+B177  uniB177,  HANGUL SYLLABLE-B177
        0xB17A, # G+10377  U+B17A  uniB17A,  HANGUL SYLLABLE-B17A
        0xB17B, # G+10378  U+B17B  uniB17B,  HANGUL SYLLABLE-B17B
        0xB17D, # G+10379  U+B17D  uniB17D,  HANGUL SYLLABLE-B17D
        0xB17E, # G+10380  U+B17E  uniB17E,  HANGUL SYLLABLE-B17E
        0xB17F, # G+10381  U+B17F  uniB17F,  HANGUL SYLLABLE-B17F
        0xB181, # G+10382  U+B181  uniB181,  HANGUL SYLLABLE-B181
        0xB183, # G+10383  U+B183  uniB183,  HANGUL SYLLABLE-B183
        0xB184, # G+10384  U+B184  uniB184,  HANGUL SYLLABLE-B184
        0xB185, # G+10385  U+B185  uniB185,  HANGUL SYLLABLE-B185
        0xB186, # G+10386  U+B186  uniB186,  HANGUL SYLLABLE-B186
        0xB187, # G+10387  U+B187  uniB187,  HANGUL SYLLABLE-B187
        0xB18A, # G+10388  U+B18A  uniB18A,  HANGUL SYLLABLE-B18A
        0xB18C, # G+10389  U+B18C  uniB18C,  HANGUL SYLLABLE-B18C
        0xB18E, # G+10390  U+B18E  uniB18E,  HANGUL SYLLABLE-B18E
        0xB18F, # G+10391  U+B18F  uniB18F,  HANGUL SYLLABLE-B18F
        0xB190, # G+10392  U+B190  uniB190,  HANGUL SYLLABLE-B190
        0xB191, # G+10393  U+B191  uniB191,  HANGUL SYLLABLE-B191
        0xB195, # G+10394  U+B195  uniB195,  HANGUL SYLLABLE-B195
        0xB196, # G+10395  U+B196  uniB196,  HANGUL SYLLABLE-B196
        0xB197, # G+10396  U+B197  uniB197,  HANGUL SYLLABLE-B197
        0xB199, # G+10397  U+B199  uniB199,  HANGUL SYLLABLE-B199
        0xB19A, # G+10398  U+B19A  uniB19A,  HANGUL SYLLABLE-B19A
        0xB19B, # G+10399  U+B19B  uniB19B,  HANGUL SYLLABLE-B19B
        0xB19D, # G+10400  U+B19D  uniB19D,  HANGUL SYLLABLE-B19D
        0xB19E, # G+10401  U+B19E  uniB19E,  HANGUL SYLLABLE-B19E
        0xB19F, # G+10402  U+B19F  uniB19F,  HANGUL SYLLABLE-B19F
        0xB1A0, # G+10403  U+B1A0  uniB1A0,  HANGUL SYLLABLE-B1A0
        0xB1A1, # G+10404  U+B1A1  uniB1A1,  HANGUL SYLLABLE-B1A1
        0xB1A2, # G+10405  U+B1A2  uniB1A2,  HANGUL SYLLABLE-B1A2
        0xB1A3, # G+10406  U+B1A3  uniB1A3,  HANGUL SYLLABLE-B1A3
        0xB1A4, # G+10407  U+B1A4  uniB1A4,  HANGUL SYLLABLE-B1A4
        0xB1A5, # G+10408  U+B1A5  uniB1A5,  HANGUL SYLLABLE-B1A5
        0xB1A6, # G+10409  U+B1A6  uniB1A6,  HANGUL SYLLABLE-B1A6
        0xB1A7, # G+10410  U+B1A7  uniB1A7,  HANGUL SYLLABLE-B1A7
        0xB1A9, # G+10411  U+B1A9  uniB1A9,  HANGUL SYLLABLE-B1A9
        0xB1AA, # G+10412  U+B1AA  uniB1AA,  HANGUL SYLLABLE-B1AA
        0xB1AB, # G+10413  U+B1AB  uniB1AB,  HANGUL SYLLABLE-B1AB
        0xB1AC, # G+10414  U+B1AC  uniB1AC,  HANGUL SYLLABLE-B1AC
        0xB1AD, # G+10415  U+B1AD  uniB1AD,  HANGUL SYLLABLE-B1AD
        0xB1AE, # G+10416  U+B1AE  uniB1AE,  HANGUL SYLLABLE-B1AE
        0xB1AF, # G+10417  U+B1AF  uniB1AF,  HANGUL SYLLABLE-B1AF
        0xB1B0, # G+10418  U+B1B0  uniB1B0,  HANGUL SYLLABLE-B1B0
        0xB1B1, # G+10419  U+B1B1  uniB1B1,  HANGUL SYLLABLE-B1B1
        0xB1B2, # G+10420  U+B1B2  uniB1B2,  HANGUL SYLLABLE-B1B2
        0xB1B3, # G+10421  U+B1B3  uniB1B3,  HANGUL SYLLABLE-B1B3
        0xB1B4, # G+10422  U+B1B4  uniB1B4,  HANGUL SYLLABLE-B1B4
        0xB1B5, # G+10423  U+B1B5  uniB1B5,  HANGUL SYLLABLE-B1B5
        0xB1B6, # G+10424  U+B1B6  uniB1B6,  HANGUL SYLLABLE-B1B6
        0xB1B7, # G+10425  U+B1B7  uniB1B7,  HANGUL SYLLABLE-B1B7
        0xB1B8, # G+10426  U+B1B8  uniB1B8,  HANGUL SYLLABLE-B1B8
        0xB1B9, # G+10427  U+B1B9  uniB1B9,  HANGUL SYLLABLE-B1B9
        0xB1BA, # G+10428  U+B1BA  uniB1BA,  HANGUL SYLLABLE-B1BA
        0xB1BB, # G+10429  U+B1BB  uniB1BB,  HANGUL SYLLABLE-B1BB
        0xB1BC, # G+10430  U+B1BC  uniB1BC,  HANGUL SYLLABLE-B1BC
        0xB1BD, # G+10431  U+B1BD  uniB1BD,  HANGUL SYLLABLE-B1BD
        0xB1BE, # G+10432  U+B1BE  uniB1BE,  HANGUL SYLLABLE-B1BE
        0xB1BF, # G+10433  U+B1BF  uniB1BF,  HANGUL SYLLABLE-B1BF
        0xB1C0, # G+10434  U+B1C0  uniB1C0,  HANGUL SYLLABLE-B1C0
        0xB1C1, # G+10435  U+B1C1  uniB1C1,  HANGUL SYLLABLE-B1C1
        0xB1C2, # G+10436  U+B1C2  uniB1C2,  HANGUL SYLLABLE-B1C2
        0xB1C3, # G+10437  U+B1C3  uniB1C3,  HANGUL SYLLABLE-B1C3
        0xB1C4, # G+10438  U+B1C4  uniB1C4,  HANGUL SYLLABLE-B1C4
        0xB1C5, # G+10439  U+B1C5  uniB1C5,  HANGUL SYLLABLE-B1C5
        0xB1C6, # G+10440  U+B1C6  uniB1C6,  HANGUL SYLLABLE-B1C6
        0xB1C7, # G+10441  U+B1C7  uniB1C7,  HANGUL SYLLABLE-B1C7
        0xB1C8, # G+10442  U+B1C8  uniB1C8,  HANGUL SYLLABLE-B1C8
        0xB1C9, # G+10443  U+B1C9  uniB1C9,  HANGUL SYLLABLE-B1C9
        0xB1CA, # G+10444  U+B1CA  uniB1CA,  HANGUL SYLLABLE-B1CA
        0xB1CB, # G+10445  U+B1CB  uniB1CB,  HANGUL SYLLABLE-B1CB
        0xB1CD, # G+10446  U+B1CD  uniB1CD,  HANGUL SYLLABLE-B1CD
        0xB1CE, # G+10447  U+B1CE  uniB1CE,  HANGUL SYLLABLE-B1CE
        0xB1CF, # G+10448  U+B1CF  uniB1CF,  HANGUL SYLLABLE-B1CF
        0xB1D1, # G+10449  U+B1D1  uniB1D1,  HANGUL SYLLABLE-B1D1
        0xB1D2, # G+10450  U+B1D2  uniB1D2,  HANGUL SYLLABLE-B1D2
        0xB1D3, # G+10451  U+B1D3  uniB1D3,  HANGUL SYLLABLE-B1D3
        0xB1D5, # G+10452  U+B1D5  uniB1D5,  HANGUL SYLLABLE-B1D5
        0xB1D6, # G+10453  U+B1D6  uniB1D6,  HANGUL SYLLABLE-B1D6
        0xB1D7, # G+10454  U+B1D7  uniB1D7,  HANGUL SYLLABLE-B1D7
        0xB1D8, # G+10455  U+B1D8  uniB1D8,  HANGUL SYLLABLE-B1D8
        0xB1D9, # G+10456  U+B1D9  uniB1D9,  HANGUL SYLLABLE-B1D9
        0xB1DA, # G+10457  U+B1DA  uniB1DA,  HANGUL SYLLABLE-B1DA
        0xB1DB, # G+10458  U+B1DB  uniB1DB,  HANGUL SYLLABLE-B1DB
        0xB1DE, # G+10459  U+B1DE  uniB1DE,  HANGUL SYLLABLE-B1DE
        0xB1E0, # G+10460  U+B1E0  uniB1E0,  HANGUL SYLLABLE-B1E0
        0xB1E1, # G+10461  U+B1E1  uniB1E1,  HANGUL SYLLABLE-B1E1
        0xB1E2, # G+10462  U+B1E2  uniB1E2,  HANGUL SYLLABLE-B1E2
        0xB1E3, # G+10463  U+B1E3  uniB1E3,  HANGUL SYLLABLE-B1E3
        0xB1E4, # G+10464  U+B1E4  uniB1E4,  HANGUL SYLLABLE-B1E4
        0xB1E5, # G+10465  U+B1E5  uniB1E5,  HANGUL SYLLABLE-B1E5
        0xB1E6, # G+10466  U+B1E6  uniB1E6,  HANGUL SYLLABLE-B1E6
        0xB1E7, # G+10467  U+B1E7  uniB1E7,  HANGUL SYLLABLE-B1E7
        0xB1EA, # G+10468  U+B1EA  uniB1EA,  HANGUL SYLLABLE-B1EA
        0xB1EB, # G+10469  U+B1EB  uniB1EB,  HANGUL SYLLABLE-B1EB
        0xB1ED, # G+10470  U+B1ED  uniB1ED,  HANGUL SYLLABLE-B1ED
        0xB1EE, # G+10471  U+B1EE  uniB1EE,  HANGUL SYLLABLE-B1EE
        0xB1EF, # G+10472  U+B1EF  uniB1EF,  HANGUL SYLLABLE-B1EF
        0xB1F1, # G+10473  U+B1F1  uniB1F1,  HANGUL SYLLABLE-B1F1
        0xB1F2, # G+10474  U+B1F2  uniB1F2,  HANGUL SYLLABLE-B1F2
        0xB1F3, # G+10475  U+B1F3  uniB1F3,  HANGUL SYLLABLE-B1F3
        0xB1F4, # G+10476  U+B1F4  uniB1F4,  HANGUL SYLLABLE-B1F4
        0xB1F5, # G+10477  U+B1F5  uniB1F5,  HANGUL SYLLABLE-B1F5
        0xB1F6, # G+10478  U+B1F6  uniB1F6,  HANGUL SYLLABLE-B1F6
        0xB1F7, # G+10479  U+B1F7  uniB1F7,  HANGUL SYLLABLE-B1F7
        0xB1F8, # G+10480  U+B1F8  uniB1F8,  HANGUL SYLLABLE-B1F8
        0xB1FA, # G+10481  U+B1FA  uniB1FA,  HANGUL SYLLABLE-B1FA
        0xB1FC, # G+10482  U+B1FC  uniB1FC,  HANGUL SYLLABLE-B1FC
        0xB1FE, # G+10483  U+B1FE  uniB1FE,  HANGUL SYLLABLE-B1FE
        0xB1FF, # G+10484  U+B1FF  uniB1FF,  HANGUL SYLLABLE-B1FF
        0xB200, # G+10485  U+B200  uniB200,  HANGUL SYLLABLE-B200
        0xB201, # G+10486  U+B201  uniB201,  HANGUL SYLLABLE-B201
        0xB202, # G+10487  U+B202  uniB202,  HANGUL SYLLABLE-B202
        0xB203, # G+10488  U+B203  uniB203,  HANGUL SYLLABLE-B203
        0xB206, # G+10489  U+B206  uniB206,  HANGUL SYLLABLE-B206
        0xB207, # G+10490  U+B207  uniB207,  HANGUL SYLLABLE-B207
        0xB209, # G+10491  U+B209  uniB209,  HANGUL SYLLABLE-B209
        0xB20A, # G+10492  U+B20A  uniB20A,  HANGUL SYLLABLE-B20A
        0xB20D, # G+10493  U+B20D  uniB20D,  HANGUL SYLLABLE-B20D
        0xB20E, # G+10494  U+B20E  uniB20E,  HANGUL SYLLABLE-B20E
        0xB20F, # G+10495  U+B20F  uniB20F,  HANGUL SYLLABLE-B20F
        0xB210, # G+10496  U+B210  uniB210,  HANGUL SYLLABLE-B210
        0xB211, # G+10497  U+B211  uniB211,  HANGUL SYLLABLE-B211
        0xB212, # G+10498  U+B212  uniB212,  HANGUL SYLLABLE-B212
        0xB213, # G+10499  U+B213  uniB213,  HANGUL SYLLABLE-B213
        0xB216, # G+10500  U+B216  uniB216,  HANGUL SYLLABLE-B216
        0xB218, # G+10501  U+B218  uniB218,  HANGUL SYLLABLE-B218
        0xB21A, # G+10502  U+B21A  uniB21A,  HANGUL SYLLABLE-B21A
        0xB21B, # G+10503  U+B21B  uniB21B,  HANGUL SYLLABLE-B21B
        0xB21C, # G+10504  U+B21C  uniB21C,  HANGUL SYLLABLE-B21C
        0xB21D, # G+10505  U+B21D  uniB21D,  HANGUL SYLLABLE-B21D
        0xB21E, # G+10506  U+B21E  uniB21E,  HANGUL SYLLABLE-B21E
        0xB21F, # G+10507  U+B21F  uniB21F,  HANGUL SYLLABLE-B21F
        0xB221, # G+10508  U+B221  uniB221,  HANGUL SYLLABLE-B221
        0xB222, # G+10509  U+B222  uniB222,  HANGUL SYLLABLE-B222
        0xB223, # G+10510  U+B223  uniB223,  HANGUL SYLLABLE-B223
        0xB224, # G+10511  U+B224  uniB224,  HANGUL SYLLABLE-B224
        0xB225, # G+10512  U+B225  uniB225,  HANGUL SYLLABLE-B225
        0xB226, # G+10513  U+B226  uniB226,  HANGUL SYLLABLE-B226
        0xB227, # G+10514  U+B227  uniB227,  HANGUL SYLLABLE-B227
        0xB228, # G+10515  U+B228  uniB228,  HANGUL SYLLABLE-B228
        0xB229, # G+10516  U+B229  uniB229,  HANGUL SYLLABLE-B229
        0xB22A, # G+10517  U+B22A  uniB22A,  HANGUL SYLLABLE-B22A
        0xB22B, # G+10518  U+B22B  uniB22B,  HANGUL SYLLABLE-B22B
        0xB22C, # G+10519  U+B22C  uniB22C,  HANGUL SYLLABLE-B22C
        0xB22D, # G+10520  U+B22D  uniB22D,  HANGUL SYLLABLE-B22D
        0xB22E, # G+10521  U+B22E  uniB22E,  HANGUL SYLLABLE-B22E
        0xB22F, # G+10522  U+B22F  uniB22F,  HANGUL SYLLABLE-B22F
        0xB230, # G+10523  U+B230  uniB230,  HANGUL SYLLABLE-B230
        0xB231, # G+10524  U+B231  uniB231,  HANGUL SYLLABLE-B231
        0xB232, # G+10525  U+B232  uniB232,  HANGUL SYLLABLE-B232
        0xB233, # G+10526  U+B233  uniB233,  HANGUL SYLLABLE-B233
        0xB235, # G+10527  U+B235  uniB235,  HANGUL SYLLABLE-B235
        0xB236, # G+10528  U+B236  uniB236,  HANGUL SYLLABLE-B236
        0xB237, # G+10529  U+B237  uniB237,  HANGUL SYLLABLE-B237
        0xB238, # G+10530  U+B238  uniB238,  HANGUL SYLLABLE-B238
        0xB239, # G+10531  U+B239  uniB239,  HANGUL SYLLABLE-B239
        0xB23A, # G+10532  U+B23A  uniB23A,  HANGUL SYLLABLE-B23A
        0xB23B, # G+10533  U+B23B  uniB23B,  HANGUL SYLLABLE-B23B
        0xB23D, # G+10534  U+B23D  uniB23D,  HANGUL SYLLABLE-B23D
        0xB23E, # G+10535  U+B23E  uniB23E,  HANGUL SYLLABLE-B23E
        0xB23F, # G+10536  U+B23F  uniB23F,  HANGUL SYLLABLE-B23F
        0xB240, # G+10537  U+B240  uniB240,  HANGUL SYLLABLE-B240
        0xB241, # G+10538  U+B241  uniB241,  HANGUL SYLLABLE-B241
        0xB242, # G+10539  U+B242  uniB242,  HANGUL SYLLABLE-B242
        0xB243, # G+10540  U+B243  uniB243,  HANGUL SYLLABLE-B243
        0xB244, # G+10541  U+B244  uniB244,  HANGUL SYLLABLE-B244
        0xB245, # G+10542  U+B245  uniB245,  HANGUL SYLLABLE-B245
        0xB246, # G+10543  U+B246  uniB246,  HANGUL SYLLABLE-B246
        0xB247, # G+10544  U+B247  uniB247,  HANGUL SYLLABLE-B247
        0xB248, # G+10545  U+B248  uniB248,  HANGUL SYLLABLE-B248
        0xB249, # G+10546  U+B249  uniB249,  HANGUL SYLLABLE-B249
        0xB24A, # G+10547  U+B24A  uniB24A,  HANGUL SYLLABLE-B24A
        0xB24B, # G+10548  U+B24B  uniB24B,  HANGUL SYLLABLE-B24B
        0xB24C, # G+10549  U+B24C  uniB24C,  HANGUL SYLLABLE-B24C
        0xB24D, # G+10550  U+B24D  uniB24D,  HANGUL SYLLABLE-B24D
        0xB24E, # G+10551  U+B24E  uniB24E,  HANGUL SYLLABLE-B24E
        0xB24F, # G+10552  U+B24F  uniB24F,  HANGUL SYLLABLE-B24F
        0xB250, # G+10553  U+B250  uniB250,  HANGUL SYLLABLE-B250
        0xB251, # G+10554  U+B251  uniB251,  HANGUL SYLLABLE-B251
        0xB252, # G+10555  U+B252  uniB252,  HANGUL SYLLABLE-B252
        0xB253, # G+10556  U+B253  uniB253,  HANGUL SYLLABLE-B253
        0xB254, # G+10557  U+B254  uniB254,  HANGUL SYLLABLE-B254
        0xB255, # G+10558  U+B255  uniB255,  HANGUL SYLLABLE-B255
        0xB256, # G+10559  U+B256  uniB256,  HANGUL SYLLABLE-B256
        0xB257, # G+10560  U+B257  uniB257,  HANGUL SYLLABLE-B257
        0xB259, # G+10561  U+B259  uniB259,  HANGUL SYLLABLE-B259
        0xB25A, # G+10562  U+B25A  uniB25A,  HANGUL SYLLABLE-B25A
        0xB25B, # G+10563  U+B25B  uniB25B,  HANGUL SYLLABLE-B25B
        0xB25D, # G+10564  U+B25D  uniB25D,  HANGUL SYLLABLE-B25D
        0xB25E, # G+10565  U+B25E  uniB25E,  HANGUL SYLLABLE-B25E
        0xB25F, # G+10566  U+B25F  uniB25F,  HANGUL SYLLABLE-B25F
        0xB261, # G+10567  U+B261  uniB261,  HANGUL SYLLABLE-B261
        0xB262, # G+10568  U+B262  uniB262,  HANGUL SYLLABLE-B262
        0xB263, # G+10569  U+B263  uniB263,  HANGUL SYLLABLE-B263
        0xB264, # G+10570  U+B264  uniB264,  HANGUL SYLLABLE-B264
        0xB265, # G+10571  U+B265  uniB265,  HANGUL SYLLABLE-B265
        0xB266, # G+10572  U+B266  uniB266,  HANGUL SYLLABLE-B266
        0xB267, # G+10573  U+B267  uniB267,  HANGUL SYLLABLE-B267
        0xB26A, # G+10574  U+B26A  uniB26A,  HANGUL SYLLABLE-B26A
        0xB26B, # G+10575  U+B26B  uniB26B,  HANGUL SYLLABLE-B26B
        0xB26C, # G+10576  U+B26C  uniB26C,  HANGUL SYLLABLE-B26C
        0xB26D, # G+10577  U+B26D  uniB26D,  HANGUL SYLLABLE-B26D
        0xB26E, # G+10578  U+B26E  uniB26E,  HANGUL SYLLABLE-B26E
        0xB26F, # G+10579  U+B26F  uniB26F,  HANGUL SYLLABLE-B26F
        0xB270, # G+10580  U+B270  uniB270,  HANGUL SYLLABLE-B270
        0xB271, # G+10581  U+B271  uniB271,  HANGUL SYLLABLE-B271
        0xB272, # G+10582  U+B272  uniB272,  HANGUL SYLLABLE-B272
        0xB273, # G+10583  U+B273  uniB273,  HANGUL SYLLABLE-B273
        0xB276, # G+10584  U+B276  uniB276,  HANGUL SYLLABLE-B276
        0xB277, # G+10585  U+B277  uniB277,  HANGUL SYLLABLE-B277
        0xB278, # G+10586  U+B278  uniB278,  HANGUL SYLLABLE-B278
        0xB279, # G+10587  U+B279  uniB279,  HANGUL SYLLABLE-B279
        0xB27A, # G+10588  U+B27A  uniB27A,  HANGUL SYLLABLE-B27A
        0xB27B, # G+10589  U+B27B  uniB27B,  HANGUL SYLLABLE-B27B
        0xB27D, # G+10590  U+B27D  uniB27D,  HANGUL SYLLABLE-B27D
        0xB27E, # G+10591  U+B27E  uniB27E,  HANGUL SYLLABLE-B27E
        0xB27F, # G+10592  U+B27F  uniB27F,  HANGUL SYLLABLE-B27F
        0xB280, # G+10593  U+B280  uniB280,  HANGUL SYLLABLE-B280
        0xB281, # G+10594  U+B281  uniB281,  HANGUL SYLLABLE-B281
        0xB282, # G+10595  U+B282  uniB282,  HANGUL SYLLABLE-B282
        0xB283, # G+10596  U+B283  uniB283,  HANGUL SYLLABLE-B283
        0xB286, # G+10597  U+B286  uniB286,  HANGUL SYLLABLE-B286
        0xB287, # G+10598  U+B287  uniB287,  HANGUL SYLLABLE-B287
        0xB288, # G+10599  U+B288  uniB288,  HANGUL SYLLABLE-B288
        0xB28A, # G+10600  U+B28A  uniB28A,  HANGUL SYLLABLE-B28A
        0xB28B, # G+10601  U+B28B  uniB28B,  HANGUL SYLLABLE-B28B
        0xB28C, # G+10602  U+B28C  uniB28C,  HANGUL SYLLABLE-B28C
        0xB28D, # G+10603  U+B28D  uniB28D,  HANGUL SYLLABLE-B28D
        0xB28E, # G+10604  U+B28E  uniB28E,  HANGUL SYLLABLE-B28E
        0xB28F, # G+10605  U+B28F  uniB28F,  HANGUL SYLLABLE-B28F
        0xB292, # G+10606  U+B292  uniB292,  HANGUL SYLLABLE-B292
        0xB293, # G+10607  U+B293  uniB293,  HANGUL SYLLABLE-B293
        0xB295, # G+10608  U+B295  uniB295,  HANGUL SYLLABLE-B295
        0xB296, # G+10609  U+B296  uniB296,  HANGUL SYLLABLE-B296
        0xB297, # G+10610  U+B297  uniB297,  HANGUL SYLLABLE-B297
        0xB29B, # G+10611  U+B29B  uniB29B,  HANGUL SYLLABLE-B29B
        0xB29C, # G+10612  U+B29C  uniB29C,  HANGUL SYLLABLE-B29C
        0xB29D, # G+10613  U+B29D  uniB29D,  HANGUL SYLLABLE-B29D
        0xB29E, # G+10614  U+B29E  uniB29E,  HANGUL SYLLABLE-B29E
        0xB29F, # G+10615  U+B29F  uniB29F,  HANGUL SYLLABLE-B29F
        0xB2A2, # G+10616  U+B2A2  uniB2A2,  HANGUL SYLLABLE-B2A2
        0xB2A4, # G+10617  U+B2A4  uniB2A4,  HANGUL SYLLABLE-B2A4
        0xB2A7, # G+10618  U+B2A7  uniB2A7,  HANGUL SYLLABLE-B2A7
        0xB2A8, # G+10619  U+B2A8  uniB2A8,  HANGUL SYLLABLE-B2A8
        0xB2A9, # G+10620  U+B2A9  uniB2A9,  HANGUL SYLLABLE-B2A9
        0xB2AB, # G+10621  U+B2AB  uniB2AB,  HANGUL SYLLABLE-B2AB
        0xB2AD, # G+10622  U+B2AD  uniB2AD,  HANGUL SYLLABLE-B2AD
        0xB2AE, # G+10623  U+B2AE  uniB2AE,  HANGUL SYLLABLE-B2AE
        0xB2AF, # G+10624  U+B2AF  uniB2AF,  HANGUL SYLLABLE-B2AF
        0xB2B1, # G+10625  U+B2B1  uniB2B1,  HANGUL SYLLABLE-B2B1
        0xB2B2, # G+10626  U+B2B2  uniB2B2,  HANGUL SYLLABLE-B2B2
        0xB2B3, # G+10627  U+B2B3  uniB2B3,  HANGUL SYLLABLE-B2B3
        0xB2B5, # G+10628  U+B2B5  uniB2B5,  HANGUL SYLLABLE-B2B5
        0xB2B6, # G+10629  U+B2B6  uniB2B6,  HANGUL SYLLABLE-B2B6
        0xB2B7, # G+10630  U+B2B7  uniB2B7,  HANGUL SYLLABLE-B2B7
        0xB2B8, # G+10631  U+B2B8  uniB2B8,  HANGUL SYLLABLE-B2B8
        0xB2B9, # G+10632  U+B2B9  uniB2B9,  HANGUL SYLLABLE-B2B9
        0xB2BA, # G+10633  U+B2BA  uniB2BA,  HANGUL SYLLABLE-B2BA
        0xB2BB, # G+10634  U+B2BB  uniB2BB,  HANGUL SYLLABLE-B2BB
        0xB2BC, # G+10635  U+B2BC  uniB2BC,  HANGUL SYLLABLE-B2BC
        0xB2BD, # G+10636  U+B2BD  uniB2BD,  HANGUL SYLLABLE-B2BD
        0xB2BE, # G+10637  U+B2BE  uniB2BE,  HANGUL SYLLABLE-B2BE
        0xB2BF, # G+10638  U+B2BF  uniB2BF,  HANGUL SYLLABLE-B2BF
        0xB2C0, # G+10639  U+B2C0  uniB2C0,  HANGUL SYLLABLE-B2C0
        0xB2C1, # G+10640  U+B2C1  uniB2C1,  HANGUL SYLLABLE-B2C1
        0xB2C2, # G+10641  U+B2C2  uniB2C2,  HANGUL SYLLABLE-B2C2
        0xB2C3, # G+10642  U+B2C3  uniB2C3,  HANGUL SYLLABLE-B2C3
        0xB2C4, # G+10643  U+B2C4  uniB2C4,  HANGUL SYLLABLE-B2C4
        0xB2C5, # G+10644  U+B2C5  uniB2C5,  HANGUL SYLLABLE-B2C5
        0xB2C6, # G+10645  U+B2C6  uniB2C6,  HANGUL SYLLABLE-B2C6
        0xB2C7, # G+10646  U+B2C7  uniB2C7,  HANGUL SYLLABLE-B2C7
        0xB2CA, # G+10647  U+B2CA  uniB2CA,  HANGUL SYLLABLE-B2CA
        0xB2CB, # G+10648  U+B2CB  uniB2CB,  HANGUL SYLLABLE-B2CB
        0xB2CD, # G+10649  U+B2CD  uniB2CD,  HANGUL SYLLABLE-B2CD
        0xB2CE, # G+10650  U+B2CE  uniB2CE,  HANGUL SYLLABLE-B2CE
        0xB2CF, # G+10651  U+B2CF  uniB2CF,  HANGUL SYLLABLE-B2CF
        0xB2D1, # G+10652  U+B2D1  uniB2D1,  HANGUL SYLLABLE-B2D1
        0xB2D3, # G+10653  U+B2D3  uniB2D3,  HANGUL SYLLABLE-B2D3
        0xB2D4, # G+10654  U+B2D4  uniB2D4,  HANGUL SYLLABLE-B2D4
        0xB2D5, # G+10655  U+B2D5  uniB2D5,  HANGUL SYLLABLE-B2D5
        0xB2D6, # G+10656  U+B2D6  uniB2D6,  HANGUL SYLLABLE-B2D6
        0xB2D7, # G+10657  U+B2D7  uniB2D7,  HANGUL SYLLABLE-B2D7
        0xB2DA, # G+10658  U+B2DA  uniB2DA,  HANGUL SYLLABLE-B2DA
        0xB2DC, # G+10659  U+B2DC  uniB2DC,  HANGUL SYLLABLE-B2DC
        0xB2DE, # G+10660  U+B2DE  uniB2DE,  HANGUL SYLLABLE-B2DE
        0xB2DF, # G+10661  U+B2DF  uniB2DF,  HANGUL SYLLABLE-B2DF
        0xB2E0, # G+10662  U+B2E0  uniB2E0,  HANGUL SYLLABLE-B2E0
        0xB2E1, # G+10663  U+B2E1  uniB2E1,  HANGUL SYLLABLE-B2E1
        0xB2E3, # G+10664  U+B2E3  uniB2E3,  HANGUL SYLLABLE-B2E3
        0xB2E7, # G+10665  U+B2E7  uniB2E7,  HANGUL SYLLABLE-B2E7
        0xB2E9, # G+10666  U+B2E9  uniB2E9,  HANGUL SYLLABLE-B2E9
        0xB2EA, # G+10667  U+B2EA  uniB2EA,  HANGUL SYLLABLE-B2EA
        0xB2F0, # G+10668  U+B2F0  uniB2F0,  HANGUL SYLLABLE-B2F0
        0xB2F1, # G+10669  U+B2F1  uniB2F1,  HANGUL SYLLABLE-B2F1
        0xB2F2, # G+10670  U+B2F2  uniB2F2,  HANGUL SYLLABLE-B2F2
        0xB2F6, # G+10671  U+B2F6  uniB2F6,  HANGUL SYLLABLE-B2F6
        0xB2FC, # G+10672  U+B2FC  uniB2FC,  HANGUL SYLLABLE-B2FC
        0xB2FD, # G+10673  U+B2FD  uniB2FD,  HANGUL SYLLABLE-B2FD
        0xB2FE, # G+10674  U+B2FE  uniB2FE,  HANGUL SYLLABLE-B2FE
        0xB302, # G+10675  U+B302  uniB302,  HANGUL SYLLABLE-B302
        0xB303, # G+10676  U+B303  uniB303,  HANGUL SYLLABLE-B303
        0xB305, # G+10677  U+B305  uniB305,  HANGUL SYLLABLE-B305
        0xB306, # G+10678  U+B306  uniB306,  HANGUL SYLLABLE-B306
        0xB307, # G+10679  U+B307  uniB307,  HANGUL SYLLABLE-B307
        0xB309, # G+10680  U+B309  uniB309,  HANGUL SYLLABLE-B309
        0xB30A, # G+10681  U+B30A  uniB30A,  HANGUL SYLLABLE-B30A
        0xB30B, # G+10682  U+B30B  uniB30B,  HANGUL SYLLABLE-B30B
        0xB30C, # G+10683  U+B30C  uniB30C,  HANGUL SYLLABLE-B30C
        0xB30D, # G+10684  U+B30D  uniB30D,  HANGUL SYLLABLE-B30D
        0xB30E, # G+10685  U+B30E  uniB30E,  HANGUL SYLLABLE-B30E
        0xB30F, # G+10686  U+B30F  uniB30F,  HANGUL SYLLABLE-B30F
        0xB312, # G+10687  U+B312  uniB312,  HANGUL SYLLABLE-B312
        0xB316, # G+10688  U+B316  uniB316,  HANGUL SYLLABLE-B316
        0xB317, # G+10689  U+B317  uniB317,  HANGUL SYLLABLE-B317
        0xB318, # G+10690  U+B318  uniB318,  HANGUL SYLLABLE-B318
        0xB319, # G+10691  U+B319  uniB319,  HANGUL SYLLABLE-B319
        0xB31A, # G+10692  U+B31A  uniB31A,  HANGUL SYLLABLE-B31A
        0xB31B, # G+10693  U+B31B  uniB31B,  HANGUL SYLLABLE-B31B
        0xB31D, # G+10694  U+B31D  uniB31D,  HANGUL SYLLABLE-B31D
        0xB31E, # G+10695  U+B31E  uniB31E,  HANGUL SYLLABLE-B31E
        0xB31F, # G+10696  U+B31F  uniB31F,  HANGUL SYLLABLE-B31F
        0xB320, # G+10697  U+B320  uniB320,  HANGUL SYLLABLE-B320
        0xB321, # G+10698  U+B321  uniB321,  HANGUL SYLLABLE-B321
        0xB322, # G+10699  U+B322  uniB322,  HANGUL SYLLABLE-B322
        0xB323, # G+10700  U+B323  uniB323,  HANGUL SYLLABLE-B323
        0xB324, # G+10701  U+B324  uniB324,  HANGUL SYLLABLE-B324
        0xB325, # G+10702  U+B325  uniB325,  HANGUL SYLLABLE-B325
        0xB326, # G+10703  U+B326  uniB326,  HANGUL SYLLABLE-B326
        0xB327, # G+10704  U+B327  uniB327,  HANGUL SYLLABLE-B327
        0xB328, # G+10705  U+B328  uniB328,  HANGUL SYLLABLE-B328
        0xB329, # G+10706  U+B329  uniB329,  HANGUL SYLLABLE-B329
        0xB32A, # G+10707  U+B32A  uniB32A,  HANGUL SYLLABLE-B32A
        0xB32B, # G+10708  U+B32B  uniB32B,  HANGUL SYLLABLE-B32B
        0xB32C, # G+10709  U+B32C  uniB32C,  HANGUL SYLLABLE-B32C
        0xB32D, # G+10710  U+B32D  uniB32D,  HANGUL SYLLABLE-B32D
        0xB32E, # G+10711  U+B32E  uniB32E,  HANGUL SYLLABLE-B32E
        0xB32F, # G+10712  U+B32F  uniB32F,  HANGUL SYLLABLE-B32F
        0xB330, # G+10713  U+B330  uniB330,  HANGUL SYLLABLE-B330
        0xB331, # G+10714  U+B331  uniB331,  HANGUL SYLLABLE-B331
        0xB332, # G+10715  U+B332  uniB332,  HANGUL SYLLABLE-B332
        0xB333, # G+10716  U+B333  uniB333,  HANGUL SYLLABLE-B333
        0xB334, # G+10717  U+B334  uniB334,  HANGUL SYLLABLE-B334
        0xB335, # G+10718  U+B335  uniB335,  HANGUL SYLLABLE-B335
        0xB336, # G+10719  U+B336  uniB336,  HANGUL SYLLABLE-B336
        0xB337, # G+10720  U+B337  uniB337,  HANGUL SYLLABLE-B337
        0xB338, # G+10721  U+B338  uniB338,  HANGUL SYLLABLE-B338
        0xB339, # G+10722  U+B339  uniB339,  HANGUL SYLLABLE-B339
        0xB33A, # G+10723  U+B33A  uniB33A,  HANGUL SYLLABLE-B33A
        0xB33B, # G+10724  U+B33B  uniB33B,  HANGUL SYLLABLE-B33B
        0xB33C, # G+10725  U+B33C  uniB33C,  HANGUL SYLLABLE-B33C
        0xB33D, # G+10726  U+B33D  uniB33D,  HANGUL SYLLABLE-B33D
        0xB33E, # G+10727  U+B33E  uniB33E,  HANGUL SYLLABLE-B33E
        0xB33F, # G+10728  U+B33F  uniB33F,  HANGUL SYLLABLE-B33F
        0xB340, # G+10729  U+B340  uniB340,  HANGUL SYLLABLE-B340
        0xB341, # G+10730  U+B341  uniB341,  HANGUL SYLLABLE-B341
        0xB342, # G+10731  U+B342  uniB342,  HANGUL SYLLABLE-B342
        0xB343, # G+10732  U+B343  uniB343,  HANGUL SYLLABLE-B343
        0xB344, # G+10733  U+B344  uniB344,  HANGUL SYLLABLE-B344
        0xB345, # G+10734  U+B345  uniB345,  HANGUL SYLLABLE-B345
        0xB346, # G+10735  U+B346  uniB346,  HANGUL SYLLABLE-B346
        0xB347, # G+10736  U+B347  uniB347,  HANGUL SYLLABLE-B347
        0xB348, # G+10737  U+B348  uniB348,  HANGUL SYLLABLE-B348
        0xB349, # G+10738  U+B349  uniB349,  HANGUL SYLLABLE-B349
        0xB34A, # G+10739  U+B34A  uniB34A,  HANGUL SYLLABLE-B34A
        0xB34B, # G+10740  U+B34B  uniB34B,  HANGUL SYLLABLE-B34B
        0xB34C, # G+10741  U+B34C  uniB34C,  HANGUL SYLLABLE-B34C
        0xB34D, # G+10742  U+B34D  uniB34D,  HANGUL SYLLABLE-B34D
        0xB34E, # G+10743  U+B34E  uniB34E,  HANGUL SYLLABLE-B34E
        0xB34F, # G+10744  U+B34F  uniB34F,  HANGUL SYLLABLE-B34F
        0xB350, # G+10745  U+B350  uniB350,  HANGUL SYLLABLE-B350
        0xB351, # G+10746  U+B351  uniB351,  HANGUL SYLLABLE-B351
        0xB352, # G+10747  U+B352  uniB352,  HANGUL SYLLABLE-B352
        0xB353, # G+10748  U+B353  uniB353,  HANGUL SYLLABLE-B353
        0xB357, # G+10749  U+B357  uniB357,  HANGUL SYLLABLE-B357
        0xB359, # G+10750  U+B359  uniB359,  HANGUL SYLLABLE-B359
        0xB35A, # G+10751  U+B35A  uniB35A,  HANGUL SYLLABLE-B35A
        0xB35D, # G+10752  U+B35D  uniB35D,  HANGUL SYLLABLE-B35D
        0xB360, # G+10753  U+B360  uniB360,  HANGUL SYLLABLE-B360
        0xB361, # G+10754  U+B361  uniB361,  HANGUL SYLLABLE-B361
        0xB362, # G+10755  U+B362  uniB362,  HANGUL SYLLABLE-B362
        0xB363, # G+10756  U+B363  uniB363,  HANGUL SYLLABLE-B363
        0xB366, # G+10757  U+B366  uniB366,  HANGUL SYLLABLE-B366
        0xB368, # G+10758  U+B368  uniB368,  HANGUL SYLLABLE-B368
        0xB36A, # G+10759  U+B36A  uniB36A,  HANGUL SYLLABLE-B36A
        0xB36C, # G+10760  U+B36C  uniB36C,  HANGUL SYLLABLE-B36C
        0xB36D, # G+10761  U+B36D  uniB36D,  HANGUL SYLLABLE-B36D
        0xB36F, # G+10762  U+B36F  uniB36F,  HANGUL SYLLABLE-B36F
        0xB372, # G+10763  U+B372  uniB372,  HANGUL SYLLABLE-B372
        0xB373, # G+10764  U+B373  uniB373,  HANGUL SYLLABLE-B373
        0xB375, # G+10765  U+B375  uniB375,  HANGUL SYLLABLE-B375
        0xB376, # G+10766  U+B376  uniB376,  HANGUL SYLLABLE-B376
        0xB377, # G+10767  U+B377  uniB377,  HANGUL SYLLABLE-B377
        0xB379, # G+10768  U+B379  uniB379,  HANGUL SYLLABLE-B379
        0xB37A, # G+10769  U+B37A  uniB37A,  HANGUL SYLLABLE-B37A
        0xB37B, # G+10770  U+B37B  uniB37B,  HANGUL SYLLABLE-B37B
        0xB37C, # G+10771  U+B37C  uniB37C,  HANGUL SYLLABLE-B37C
        0xB37D, # G+10772  U+B37D  uniB37D,  HANGUL SYLLABLE-B37D
        0xB37E, # G+10773  U+B37E  uniB37E,  HANGUL SYLLABLE-B37E
        0xB37F, # G+10774  U+B37F  uniB37F,  HANGUL SYLLABLE-B37F
        0xB382, # G+10775  U+B382  uniB382,  HANGUL SYLLABLE-B382
        0xB386, # G+10776  U+B386  uniB386,  HANGUL SYLLABLE-B386
        0xB387, # G+10777  U+B387  uniB387,  HANGUL SYLLABLE-B387
        0xB388, # G+10778  U+B388  uniB388,  HANGUL SYLLABLE-B388
        0xB389, # G+10779  U+B389  uniB389,  HANGUL SYLLABLE-B389
        0xB38A, # G+10780  U+B38A  uniB38A,  HANGUL SYLLABLE-B38A
        0xB38B, # G+10781  U+B38B  uniB38B,  HANGUL SYLLABLE-B38B
        0xB38D, # G+10782  U+B38D  uniB38D,  HANGUL SYLLABLE-B38D
        0xB38E, # G+10783  U+B38E  uniB38E,  HANGUL SYLLABLE-B38E
        0xB38F, # G+10784  U+B38F  uniB38F,  HANGUL SYLLABLE-B38F
        0xB391, # G+10785  U+B391  uniB391,  HANGUL SYLLABLE-B391
        0xB392, # G+10786  U+B392  uniB392,  HANGUL SYLLABLE-B392
        0xB393, # G+10787  U+B393  uniB393,  HANGUL SYLLABLE-B393
        0xB395, # G+10788  U+B395  uniB395,  HANGUL SYLLABLE-B395
        0xB396, # G+10789  U+B396  uniB396,  HANGUL SYLLABLE-B396
        0xB397, # G+10790  U+B397  uniB397,  HANGUL SYLLABLE-B397
        0xB398, # G+10791  U+B398  uniB398,  HANGUL SYLLABLE-B398
        0xB399, # G+10792  U+B399  uniB399,  HANGUL SYLLABLE-B399
        0xB39A, # G+10793  U+B39A  uniB39A,  HANGUL SYLLABLE-B39A
        0xB39B, # G+10794  U+B39B  uniB39B,  HANGUL SYLLABLE-B39B
        0xB39C, # G+10795  U+B39C  uniB39C,  HANGUL SYLLABLE-B39C
        0xB39D, # G+10796  U+B39D  uniB39D,  HANGUL SYLLABLE-B39D
        0xB39E, # G+10797  U+B39E  uniB39E,  HANGUL SYLLABLE-B39E
        0xB39F, # G+10798  U+B39F  uniB39F,  HANGUL SYLLABLE-B39F
        0xB3A2, # G+10799  U+B3A2  uniB3A2,  HANGUL SYLLABLE-B3A2
        0xB3A3, # G+10800  U+B3A3  uniB3A3,  HANGUL SYLLABLE-B3A3
        0xB3A4, # G+10801  U+B3A4  uniB3A4,  HANGUL SYLLABLE-B3A4
        0xB3A5, # G+10802  U+B3A5  uniB3A5,  HANGUL SYLLABLE-B3A5
        0xB3A6, # G+10803  U+B3A6  uniB3A6,  HANGUL SYLLABLE-B3A6
        0xB3A7, # G+10804  U+B3A7  uniB3A7,  HANGUL SYLLABLE-B3A7
        0xB3A9, # G+10805  U+B3A9  uniB3A9,  HANGUL SYLLABLE-B3A9
        0xB3AA, # G+10806  U+B3AA  uniB3AA,  HANGUL SYLLABLE-B3AA
        0xB3AB, # G+10807  U+B3AB  uniB3AB,  HANGUL SYLLABLE-B3AB
        0xB3AD, # G+10808  U+B3AD  uniB3AD,  HANGUL SYLLABLE-B3AD
        0xB3AE, # G+10809  U+B3AE  uniB3AE,  HANGUL SYLLABLE-B3AE
        0xB3AF, # G+10810  U+B3AF  uniB3AF,  HANGUL SYLLABLE-B3AF
        0xB3B0, # G+10811  U+B3B0  uniB3B0,  HANGUL SYLLABLE-B3B0
        0xB3B1, # G+10812  U+B3B1  uniB3B1,  HANGUL SYLLABLE-B3B1
        0xB3B2, # G+10813  U+B3B2  uniB3B2,  HANGUL SYLLABLE-B3B2
        0xB3B3, # G+10814  U+B3B3  uniB3B3,  HANGUL SYLLABLE-B3B3
        0xB3B4, # G+10815  U+B3B4  uniB3B4,  HANGUL SYLLABLE-B3B4
        0xB3B5, # G+10816  U+B3B5  uniB3B5,  HANGUL SYLLABLE-B3B5
        0xB3B6, # G+10817  U+B3B6  uniB3B6,  HANGUL SYLLABLE-B3B6
        0xB3B7, # G+10818  U+B3B7  uniB3B7,  HANGUL SYLLABLE-B3B7
        0xB3B8, # G+10819  U+B3B8  uniB3B8,  HANGUL SYLLABLE-B3B8
        0xB3B9, # G+10820  U+B3B9  uniB3B9,  HANGUL SYLLABLE-B3B9
        0xB3BA, # G+10821  U+B3BA  uniB3BA,  HANGUL SYLLABLE-B3BA
        0xB3BB, # G+10822  U+B3BB  uniB3BB,  HANGUL SYLLABLE-B3BB
        0xB3BC, # G+10823  U+B3BC  uniB3BC,  HANGUL SYLLABLE-B3BC
        0xB3BD, # G+10824  U+B3BD  uniB3BD,  HANGUL SYLLABLE-B3BD
        0xB3BE, # G+10825  U+B3BE  uniB3BE,  HANGUL SYLLABLE-B3BE
        0xB3BF, # G+10826  U+B3BF  uniB3BF,  HANGUL SYLLABLE-B3BF
        0xB3C0, # G+10827  U+B3C0  uniB3C0,  HANGUL SYLLABLE-B3C0
        0xB3C1, # G+10828  U+B3C1  uniB3C1,  HANGUL SYLLABLE-B3C1
        0xB3C2, # G+10829  U+B3C2  uniB3C2,  HANGUL SYLLABLE-B3C2
        0xB3C3, # G+10830  U+B3C3  uniB3C3,  HANGUL SYLLABLE-B3C3
        0xB3C6, # G+10831  U+B3C6  uniB3C6,  HANGUL SYLLABLE-B3C6
        0xB3C7, # G+10832  U+B3C7  uniB3C7,  HANGUL SYLLABLE-B3C7
        0xB3C9, # G+10833  U+B3C9  uniB3C9,  HANGUL SYLLABLE-B3C9
        0xB3CA, # G+10834  U+B3CA  uniB3CA,  HANGUL SYLLABLE-B3CA
        0xB3CD, # G+10835  U+B3CD  uniB3CD,  HANGUL SYLLABLE-B3CD
        0xB3CF, # G+10836  U+B3CF  uniB3CF,  HANGUL SYLLABLE-B3CF
        0xB3D1, # G+10837  U+B3D1  uniB3D1,  HANGUL SYLLABLE-B3D1
        0xB3D2, # G+10838  U+B3D2  uniB3D2,  HANGUL SYLLABLE-B3D2
        0xB3D3, # G+10839  U+B3D3  uniB3D3,  HANGUL SYLLABLE-B3D3
        0xB3D6, # G+10840  U+B3D6  uniB3D6,  HANGUL SYLLABLE-B3D6
        0xB3D8, # G+10841  U+B3D8  uniB3D8,  HANGUL SYLLABLE-B3D8
        0xB3DA, # G+10842  U+B3DA  uniB3DA,  HANGUL SYLLABLE-B3DA
        0xB3DC, # G+10843  U+B3DC  uniB3DC,  HANGUL SYLLABLE-B3DC
        0xB3DE, # G+10844  U+B3DE  uniB3DE,  HANGUL SYLLABLE-B3DE
        0xB3DF, # G+10845  U+B3DF  uniB3DF,  HANGUL SYLLABLE-B3DF
        0xB3E1, # G+10846  U+B3E1  uniB3E1,  HANGUL SYLLABLE-B3E1
        0xB3E2, # G+10847  U+B3E2  uniB3E2,  HANGUL SYLLABLE-B3E2
        0xB3E3, # G+10848  U+B3E3  uniB3E3,  HANGUL SYLLABLE-B3E3
        0xB3E5, # G+10849  U+B3E5  uniB3E5,  HANGUL SYLLABLE-B3E5
        0xB3E6, # G+10850  U+B3E6  uniB3E6,  HANGUL SYLLABLE-B3E6
        0xB3E7, # G+10851  U+B3E7  uniB3E7,  HANGUL SYLLABLE-B3E7
        0xB3E9, # G+10852  U+B3E9  uniB3E9,  HANGUL SYLLABLE-B3E9
        0xB3EA, # G+10853  U+B3EA  uniB3EA,  HANGUL SYLLABLE-B3EA
        0xB3EB, # G+10854  U+B3EB  uniB3EB,  HANGUL SYLLABLE-B3EB
        0xB3EC, # G+10855  U+B3EC  uniB3EC,  HANGUL SYLLABLE-B3EC
        0xB3ED, # G+10856  U+B3ED  uniB3ED,  HANGUL SYLLABLE-B3ED
        0xB3EE, # G+10857  U+B3EE  uniB3EE,  HANGUL SYLLABLE-B3EE
        0xB3EF, # G+10858  U+B3EF  uniB3EF,  HANGUL SYLLABLE-B3EF
        0xB3F0, # G+10859  U+B3F0  uniB3F0,  HANGUL SYLLABLE-B3F0
        0xB3F1, # G+10860  U+B3F1  uniB3F1,  HANGUL SYLLABLE-B3F1
        0xB3F2, # G+10861  U+B3F2  uniB3F2,  HANGUL SYLLABLE-B3F2
        0xB3F3, # G+10862  U+B3F3  uniB3F3,  HANGUL SYLLABLE-B3F3
        0xB3F4, # G+10863  U+B3F4  uniB3F4,  HANGUL SYLLABLE-B3F4
        0xB3F5, # G+10864  U+B3F5  uniB3F5,  HANGUL SYLLABLE-B3F5
        0xB3F6, # G+10865  U+B3F6  uniB3F6,  HANGUL SYLLABLE-B3F6
        0xB3F7, # G+10866  U+B3F7  uniB3F7,  HANGUL SYLLABLE-B3F7
        0xB3F8, # G+10867  U+B3F8  uniB3F8,  HANGUL SYLLABLE-B3F8
        0xB3F9, # G+10868  U+B3F9  uniB3F9,  HANGUL SYLLABLE-B3F9
        0xB3FA, # G+10869  U+B3FA  uniB3FA,  HANGUL SYLLABLE-B3FA
        0xB3FB, # G+10870  U+B3FB  uniB3FB,  HANGUL SYLLABLE-B3FB
        0xB3FD, # G+10871  U+B3FD  uniB3FD,  HANGUL SYLLABLE-B3FD
        0xB3FE, # G+10872  U+B3FE  uniB3FE,  HANGUL SYLLABLE-B3FE
        0xB3FF, # G+10873  U+B3FF  uniB3FF,  HANGUL SYLLABLE-B3FF
        0xB400, # G+10874  U+B400  uniB400,  HANGUL SYLLABLE-B400
        0xB401, # G+10875  U+B401  uniB401,  HANGUL SYLLABLE-B401
        0xB402, # G+10876  U+B402  uniB402,  HANGUL SYLLABLE-B402
        0xB403, # G+10877  U+B403  uniB403,  HANGUL SYLLABLE-B403
        0xB404, # G+10878  U+B404  uniB404,  HANGUL SYLLABLE-B404
        0xB405, # G+10879  U+B405  uniB405,  HANGUL SYLLABLE-B405
        0xB406, # G+10880  U+B406  uniB406,  HANGUL SYLLABLE-B406
        0xB407, # G+10881  U+B407  uniB407,  HANGUL SYLLABLE-B407
        0xB408, # G+10882  U+B408  uniB408,  HANGUL SYLLABLE-B408
        0xB409, # G+10883  U+B409  uniB409,  HANGUL SYLLABLE-B409
        0xB40A, # G+10884  U+B40A  uniB40A,  HANGUL SYLLABLE-B40A
        0xB40B, # G+10885  U+B40B  uniB40B,  HANGUL SYLLABLE-B40B
        0xB40C, # G+10886  U+B40C  uniB40C,  HANGUL SYLLABLE-B40C
        0xB40D, # G+10887  U+B40D  uniB40D,  HANGUL SYLLABLE-B40D
        0xB40E, # G+10888  U+B40E  uniB40E,  HANGUL SYLLABLE-B40E
        0xB40F, # G+10889  U+B40F  uniB40F,  HANGUL SYLLABLE-B40F
        0xB411, # G+10890  U+B411  uniB411,  HANGUL SYLLABLE-B411
        0xB412, # G+10891  U+B412  uniB412,  HANGUL SYLLABLE-B412
        0xB413, # G+10892  U+B413  uniB413,  HANGUL SYLLABLE-B413
        0xB414, # G+10893  U+B414  uniB414,  HANGUL SYLLABLE-B414
        0xB415, # G+10894  U+B415  uniB415,  HANGUL SYLLABLE-B415
        0xB416, # G+10895  U+B416  uniB416,  HANGUL SYLLABLE-B416
        0xB417, # G+10896  U+B417  uniB417,  HANGUL SYLLABLE-B417
        0xB419, # G+10897  U+B419  uniB419,  HANGUL SYLLABLE-B419
        0xB41A, # G+10898  U+B41A  uniB41A,  HANGUL SYLLABLE-B41A
        0xB41B, # G+10899  U+B41B  uniB41B,  HANGUL SYLLABLE-B41B
        0xB41D, # G+10900  U+B41D  uniB41D,  HANGUL SYLLABLE-B41D
        0xB41E, # G+10901  U+B41E  uniB41E,  HANGUL SYLLABLE-B41E
        0xB41F, # G+10902  U+B41F  uniB41F,  HANGUL SYLLABLE-B41F
        0xB421, # G+10903  U+B421  uniB421,  HANGUL SYLLABLE-B421
        0xB422, # G+10904  U+B422  uniB422,  HANGUL SYLLABLE-B422
        0xB423, # G+10905  U+B423  uniB423,  HANGUL SYLLABLE-B423
        0xB424, # G+10906  U+B424  uniB424,  HANGUL SYLLABLE-B424
        0xB425, # G+10907  U+B425  uniB425,  HANGUL SYLLABLE-B425
        0xB426, # G+10908  U+B426  uniB426,  HANGUL SYLLABLE-B426
        0xB427, # G+10909  U+B427  uniB427,  HANGUL SYLLABLE-B427
        0xB42A, # G+10910  U+B42A  uniB42A,  HANGUL SYLLABLE-B42A
        0xB42C, # G+10911  U+B42C  uniB42C,  HANGUL SYLLABLE-B42C
        0xB42D, # G+10912  U+B42D  uniB42D,  HANGUL SYLLABLE-B42D
        0xB42E, # G+10913  U+B42E  uniB42E,  HANGUL SYLLABLE-B42E
        0xB42F, # G+10914  U+B42F  uniB42F,  HANGUL SYLLABLE-B42F
        0xB430, # G+10915  U+B430  uniB430,  HANGUL SYLLABLE-B430
        0xB431, # G+10916  U+B431  uniB431,  HANGUL SYLLABLE-B431
        0xB432, # G+10917  U+B432  uniB432,  HANGUL SYLLABLE-B432
        0xB433, # G+10918  U+B433  uniB433,  HANGUL SYLLABLE-B433
        0xB435, # G+10919  U+B435  uniB435,  HANGUL SYLLABLE-B435
        0xB436, # G+10920  U+B436  uniB436,  HANGUL SYLLABLE-B436
        0xB437, # G+10921  U+B437  uniB437,  HANGUL SYLLABLE-B437
        0xB438, # G+10922  U+B438  uniB438,  HANGUL SYLLABLE-B438
        0xB439, # G+10923  U+B439  uniB439,  HANGUL SYLLABLE-B439
        0xB43A, # G+10924  U+B43A  uniB43A,  HANGUL SYLLABLE-B43A
        0xB43B, # G+10925  U+B43B  uniB43B,  HANGUL SYLLABLE-B43B
        0xB43C, # G+10926  U+B43C  uniB43C,  HANGUL SYLLABLE-B43C
        0xB43D, # G+10927  U+B43D  uniB43D,  HANGUL SYLLABLE-B43D
        0xB43E, # G+10928  U+B43E  uniB43E,  HANGUL SYLLABLE-B43E
        0xB43F, # G+10929  U+B43F  uniB43F,  HANGUL SYLLABLE-B43F
        0xB440, # G+10930  U+B440  uniB440,  HANGUL SYLLABLE-B440
        0xB441, # G+10931  U+B441  uniB441,  HANGUL SYLLABLE-B441
        0xB442, # G+10932  U+B442  uniB442,  HANGUL SYLLABLE-B442
        0xB443, # G+10933  U+B443  uniB443,  HANGUL SYLLABLE-B443
        0xB444, # G+10934  U+B444  uniB444,  HANGUL SYLLABLE-B444
        0xB445, # G+10935  U+B445  uniB445,  HANGUL SYLLABLE-B445
        0xB446, # G+10936  U+B446  uniB446,  HANGUL SYLLABLE-B446
        0xB447, # G+10937  U+B447  uniB447,  HANGUL SYLLABLE-B447
        0xB448, # G+10938  U+B448  uniB448,  HANGUL SYLLABLE-B448
        0xB449, # G+10939  U+B449  uniB449,  HANGUL SYLLABLE-B449
        0xB44A, # G+10940  U+B44A  uniB44A,  HANGUL SYLLABLE-B44A
        0xB44B, # G+10941  U+B44B  uniB44B,  HANGUL SYLLABLE-B44B
        0xB44C, # G+10942  U+B44C  uniB44C,  HANGUL SYLLABLE-B44C
        0xB44D, # G+10943  U+B44D  uniB44D,  HANGUL SYLLABLE-B44D
        0xB44E, # G+10944  U+B44E  uniB44E,  HANGUL SYLLABLE-B44E
        0xB44F, # G+10945  U+B44F  uniB44F,  HANGUL SYLLABLE-B44F
        0xB452, # G+10946  U+B452  uniB452,  HANGUL SYLLABLE-B452
        0xB453, # G+10947  U+B453  uniB453,  HANGUL SYLLABLE-B453
        0xB455, # G+10948  U+B455  uniB455,  HANGUL SYLLABLE-B455
        0xB456, # G+10949  U+B456  uniB456,  HANGUL SYLLABLE-B456
        0xB457, # G+10950  U+B457  uniB457,  HANGUL SYLLABLE-B457
        0xB459, # G+10951  U+B459  uniB459,  HANGUL SYLLABLE-B459
        0xB45A, # G+10952  U+B45A  uniB45A,  HANGUL SYLLABLE-B45A
        0xB45B, # G+10953  U+B45B  uniB45B,  HANGUL SYLLABLE-B45B
        0xB45C, # G+10954  U+B45C  uniB45C,  HANGUL SYLLABLE-B45C
        0xB45D, # G+10955  U+B45D  uniB45D,  HANGUL SYLLABLE-B45D
        0xB45E, # G+10956  U+B45E  uniB45E,  HANGUL SYLLABLE-B45E
        0xB45F, # G+10957  U+B45F  uniB45F,  HANGUL SYLLABLE-B45F
        0xB462, # G+10958  U+B462  uniB462,  HANGUL SYLLABLE-B462
        0xB464, # G+10959  U+B464  uniB464,  HANGUL SYLLABLE-B464
        0xB466, # G+10960  U+B466  uniB466,  HANGUL SYLLABLE-B466
        0xB467, # G+10961  U+B467  uniB467,  HANGUL SYLLABLE-B467
        0xB468, # G+10962  U+B468  uniB468,  HANGUL SYLLABLE-B468
        0xB469, # G+10963  U+B469  uniB469,  HANGUL SYLLABLE-B469
        0xB46A, # G+10964  U+B46A  uniB46A,  HANGUL SYLLABLE-B46A
        0xB46B, # G+10965  U+B46B  uniB46B,  HANGUL SYLLABLE-B46B
        0xB46D, # G+10966  U+B46D  uniB46D,  HANGUL SYLLABLE-B46D
        0xB46E, # G+10967  U+B46E  uniB46E,  HANGUL SYLLABLE-B46E
        0xB46F, # G+10968  U+B46F  uniB46F,  HANGUL SYLLABLE-B46F
        0xB470, # G+10969  U+B470  uniB470,  HANGUL SYLLABLE-B470
        0xB471, # G+10970  U+B471  uniB471,  HANGUL SYLLABLE-B471
        0xB472, # G+10971  U+B472  uniB472,  HANGUL SYLLABLE-B472
        0xB473, # G+10972  U+B473  uniB473,  HANGUL SYLLABLE-B473
        0xB474, # G+10973  U+B474  uniB474,  HANGUL SYLLABLE-B474
        0xB475, # G+10974  U+B475  uniB475,  HANGUL SYLLABLE-B475
        0xB476, # G+10975  U+B476  uniB476,  HANGUL SYLLABLE-B476
        0xB477, # G+10976  U+B477  uniB477,  HANGUL SYLLABLE-B477
        0xB478, # G+10977  U+B478  uniB478,  HANGUL SYLLABLE-B478
        0xB479, # G+10978  U+B479  uniB479,  HANGUL SYLLABLE-B479
        0xB47A, # G+10979  U+B47A  uniB47A,  HANGUL SYLLABLE-B47A
        0xB47B, # G+10980  U+B47B  uniB47B,  HANGUL SYLLABLE-B47B
        0xB47C, # G+10981  U+B47C  uniB47C,  HANGUL SYLLABLE-B47C
        0xB47D, # G+10982  U+B47D  uniB47D,  HANGUL SYLLABLE-B47D
        0xB47E, # G+10983  U+B47E  uniB47E,  HANGUL SYLLABLE-B47E
        0xB47F, # G+10984  U+B47F  uniB47F,  HANGUL SYLLABLE-B47F
        0xB481, # G+10985  U+B481  uniB481,  HANGUL SYLLABLE-B481
        0xB482, # G+10986  U+B482  uniB482,  HANGUL SYLLABLE-B482
        0xB483, # G+10987  U+B483  uniB483,  HANGUL SYLLABLE-B483
        0xB484, # G+10988  U+B484  uniB484,  HANGUL SYLLABLE-B484
        0xB485, # G+10989  U+B485  uniB485,  HANGUL SYLLABLE-B485
        0xB486, # G+10990  U+B486  uniB486,  HANGUL SYLLABLE-B486
        0xB487, # G+10991  U+B487  uniB487,  HANGUL SYLLABLE-B487
        0xB489, # G+10992  U+B489  uniB489,  HANGUL SYLLABLE-B489
        0xB48A, # G+10993  U+B48A  uniB48A,  HANGUL SYLLABLE-B48A
        0xB48B, # G+10994  U+B48B  uniB48B,  HANGUL SYLLABLE-B48B
        0xB48C, # G+10995  U+B48C  uniB48C,  HANGUL SYLLABLE-B48C
        0xB48D, # G+10996  U+B48D  uniB48D,  HANGUL SYLLABLE-B48D
        0xB48E, # G+10997  U+B48E  uniB48E,  HANGUL SYLLABLE-B48E
        0xB48F, # G+10998  U+B48F  uniB48F,  HANGUL SYLLABLE-B48F
        0xB490, # G+10999  U+B490  uniB490,  HANGUL SYLLABLE-B490
        0xB491, # G+11000  U+B491  uniB491,  HANGUL SYLLABLE-B491
        0xB492, # G+11001  U+B492  uniB492,  HANGUL SYLLABLE-B492
        0xB493, # G+11002  U+B493  uniB493,  HANGUL SYLLABLE-B493
        0xB494, # G+11003  U+B494  uniB494,  HANGUL SYLLABLE-B494
        0xB495, # G+11004  U+B495  uniB495,  HANGUL SYLLABLE-B495
        0xB496, # G+11005  U+B496  uniB496,  HANGUL SYLLABLE-B496
        0xB497, # G+11006  U+B497  uniB497,  HANGUL SYLLABLE-B497
        0xB498, # G+11007  U+B498  uniB498,  HANGUL SYLLABLE-B498
        0xB499, # G+11008  U+B499  uniB499,  HANGUL SYLLABLE-B499
        0xB49A, # G+11009  U+B49A  uniB49A,  HANGUL SYLLABLE-B49A
        0xB49B, # G+11010  U+B49B  uniB49B,  HANGUL SYLLABLE-B49B
        0xB49C, # G+11011  U+B49C  uniB49C,  HANGUL SYLLABLE-B49C
        0xB49E, # G+11012  U+B49E  uniB49E,  HANGUL SYLLABLE-B49E
        0xB49F, # G+11013  U+B49F  uniB49F,  HANGUL SYLLABLE-B49F
        0xB4A0, # G+11014  U+B4A0  uniB4A0,  HANGUL SYLLABLE-B4A0
        0xB4A1, # G+11015  U+B4A1  uniB4A1,  HANGUL SYLLABLE-B4A1
        0xB4A2, # G+11016  U+B4A2  uniB4A2,  HANGUL SYLLABLE-B4A2
        0xB4A3, # G+11017  U+B4A3  uniB4A3,  HANGUL SYLLABLE-B4A3
        0xB4A5, # G+11018  U+B4A5  uniB4A5,  HANGUL SYLLABLE-B4A5
        0xB4A6, # G+11019  U+B4A6  uniB4A6,  HANGUL SYLLABLE-B4A6
        0xB4A7, # G+11020  U+B4A7  uniB4A7,  HANGUL SYLLABLE-B4A7
        0xB4A9, # G+11021  U+B4A9  uniB4A9,  HANGUL SYLLABLE-B4A9
        0xB4AA, # G+11022  U+B4AA  uniB4AA,  HANGUL SYLLABLE-B4AA
        0xB4AB, # G+11023  U+B4AB  uniB4AB,  HANGUL SYLLABLE-B4AB
        0xB4AD, # G+11024  U+B4AD  uniB4AD,  HANGUL SYLLABLE-B4AD
        0xB4AE, # G+11025  U+B4AE  uniB4AE,  HANGUL SYLLABLE-B4AE
        0xB4AF, # G+11026  U+B4AF  uniB4AF,  HANGUL SYLLABLE-B4AF
        0xB4B0, # G+11027  U+B4B0  uniB4B0,  HANGUL SYLLABLE-B4B0
        0xB4B1, # G+11028  U+B4B1  uniB4B1,  HANGUL SYLLABLE-B4B1
        0xB4B2, # G+11029  U+B4B2  uniB4B2,  HANGUL SYLLABLE-B4B2
        0xB4B3, # G+11030  U+B4B3  uniB4B3,  HANGUL SYLLABLE-B4B3
        0xB4B4, # G+11031  U+B4B4  uniB4B4,  HANGUL SYLLABLE-B4B4
        0xB4B6, # G+11032  U+B4B6  uniB4B6,  HANGUL SYLLABLE-B4B6
        0xB4B8, # G+11033  U+B4B8  uniB4B8,  HANGUL SYLLABLE-B4B8
        0xB4BA, # G+11034  U+B4BA  uniB4BA,  HANGUL SYLLABLE-B4BA
        0xB4BB, # G+11035  U+B4BB  uniB4BB,  HANGUL SYLLABLE-B4BB
        0xB4BC, # G+11036  U+B4BC  uniB4BC,  HANGUL SYLLABLE-B4BC
        0xB4BD, # G+11037  U+B4BD  uniB4BD,  HANGUL SYLLABLE-B4BD
        0xB4BE, # G+11038  U+B4BE  uniB4BE,  HANGUL SYLLABLE-B4BE
        0xB4BF, # G+11039  U+B4BF  uniB4BF,  HANGUL SYLLABLE-B4BF
        0xB4C1, # G+11040  U+B4C1  uniB4C1,  HANGUL SYLLABLE-B4C1
        0xB4C2, # G+11041  U+B4C2  uniB4C2,  HANGUL SYLLABLE-B4C2
        0xB4C3, # G+11042  U+B4C3  uniB4C3,  HANGUL SYLLABLE-B4C3
        0xB4C5, # G+11043  U+B4C5  uniB4C5,  HANGUL SYLLABLE-B4C5
        0xB4C6, # G+11044  U+B4C6  uniB4C6,  HANGUL SYLLABLE-B4C6
        0xB4C7, # G+11045  U+B4C7  uniB4C7,  HANGUL SYLLABLE-B4C7
        0xB4C9, # G+11046  U+B4C9  uniB4C9,  HANGUL SYLLABLE-B4C9
        0xB4CA, # G+11047  U+B4CA  uniB4CA,  HANGUL SYLLABLE-B4CA
        0xB4CB, # G+11048  U+B4CB  uniB4CB,  HANGUL SYLLABLE-B4CB
        0xB4CC, # G+11049  U+B4CC  uniB4CC,  HANGUL SYLLABLE-B4CC
        0xB4CD, # G+11050  U+B4CD  uniB4CD,  HANGUL SYLLABLE-B4CD
        0xB4CE, # G+11051  U+B4CE  uniB4CE,  HANGUL SYLLABLE-B4CE
        0xB4CF, # G+11052  U+B4CF  uniB4CF,  HANGUL SYLLABLE-B4CF
        0xB4D1, # G+11053  U+B4D1  uniB4D1,  HANGUL SYLLABLE-B4D1
        0xB4D2, # G+11054  U+B4D2  uniB4D2,  HANGUL SYLLABLE-B4D2
        0xB4D3, # G+11055  U+B4D3  uniB4D3,  HANGUL SYLLABLE-B4D3
        0xB4D4, # G+11056  U+B4D4  uniB4D4,  HANGUL SYLLABLE-B4D4
        0xB4D6, # G+11057  U+B4D6  uniB4D6,  HANGUL SYLLABLE-B4D6
        0xB4D7, # G+11058  U+B4D7  uniB4D7,  HANGUL SYLLABLE-B4D7
        0xB4D8, # G+11059  U+B4D8  uniB4D8,  HANGUL SYLLABLE-B4D8
        0xB4D9, # G+11060  U+B4D9  uniB4D9,  HANGUL SYLLABLE-B4D9
        0xB4DA, # G+11061  U+B4DA  uniB4DA,  HANGUL SYLLABLE-B4DA
        0xB4DB, # G+11062  U+B4DB  uniB4DB,  HANGUL SYLLABLE-B4DB
        0xB4DE, # G+11063  U+B4DE  uniB4DE,  HANGUL SYLLABLE-B4DE
        0xB4DF, # G+11064  U+B4DF  uniB4DF,  HANGUL SYLLABLE-B4DF
        0xB4E1, # G+11065  U+B4E1  uniB4E1,  HANGUL SYLLABLE-B4E1
        0xB4E2, # G+11066  U+B4E2  uniB4E2,  HANGUL SYLLABLE-B4E2
        0xB4E5, # G+11067  U+B4E5  uniB4E5,  HANGUL SYLLABLE-B4E5
        0xB4E7, # G+11068  U+B4E7  uniB4E7,  HANGUL SYLLABLE-B4E7
        0xB4E8, # G+11069  U+B4E8  uniB4E8,  HANGUL SYLLABLE-B4E8
        0xB4E9, # G+11070  U+B4E9  uniB4E9,  HANGUL SYLLABLE-B4E9
        0xB4EA, # G+11071  U+B4EA  uniB4EA,  HANGUL SYLLABLE-B4EA
        0xB4EB, # G+11072  U+B4EB  uniB4EB,  HANGUL SYLLABLE-B4EB
        0xB4EE, # G+11073  U+B4EE  uniB4EE,  HANGUL SYLLABLE-B4EE
        0xB4F0, # G+11074  U+B4F0  uniB4F0,  HANGUL SYLLABLE-B4F0
        0xB4F2, # G+11075  U+B4F2  uniB4F2,  HANGUL SYLLABLE-B4F2
        0xB4F3, # G+11076  U+B4F3  uniB4F3,  HANGUL SYLLABLE-B4F3
        0xB4F4, # G+11077  U+B4F4  uniB4F4,  HANGUL SYLLABLE-B4F4
        0xB4F5, # G+11078  U+B4F5  uniB4F5,  HANGUL SYLLABLE-B4F5
        0xB4F6, # G+11079  U+B4F6  uniB4F6,  HANGUL SYLLABLE-B4F6
        0xB4F7, # G+11080  U+B4F7  uniB4F7,  HANGUL SYLLABLE-B4F7
        0xB4F9, # G+11081  U+B4F9  uniB4F9,  HANGUL SYLLABLE-B4F9
        0xB4FA, # G+11082  U+B4FA  uniB4FA,  HANGUL SYLLABLE-B4FA
        0xB4FB, # G+11083  U+B4FB  uniB4FB,  HANGUL SYLLABLE-B4FB
        0xB4FC, # G+11084  U+B4FC  uniB4FC,  HANGUL SYLLABLE-B4FC
        0xB4FD, # G+11085  U+B4FD  uniB4FD,  HANGUL SYLLABLE-B4FD
        0xB4FE, # G+11086  U+B4FE  uniB4FE,  HANGUL SYLLABLE-B4FE
        0xB4FF, # G+11087  U+B4FF  uniB4FF,  HANGUL SYLLABLE-B4FF
        0xB500, # G+11088  U+B500  uniB500,  HANGUL SYLLABLE-B500
        0xB501, # G+11089  U+B501  uniB501,  HANGUL SYLLABLE-B501
        0xB502, # G+11090  U+B502  uniB502,  HANGUL SYLLABLE-B502
        0xB503, # G+11091  U+B503  uniB503,  HANGUL SYLLABLE-B503
        0xB504, # G+11092  U+B504  uniB504,  HANGUL SYLLABLE-B504
        0xB505, # G+11093  U+B505  uniB505,  HANGUL SYLLABLE-B505
        0xB506, # G+11094  U+B506  uniB506,  HANGUL SYLLABLE-B506
        0xB507, # G+11095  U+B507  uniB507,  HANGUL SYLLABLE-B507
        0xB508, # G+11096  U+B508  uniB508,  HANGUL SYLLABLE-B508
        0xB509, # G+11097  U+B509  uniB509,  HANGUL SYLLABLE-B509
        0xB50A, # G+11098  U+B50A  uniB50A,  HANGUL SYLLABLE-B50A
        0xB50B, # G+11099  U+B50B  uniB50B,  HANGUL SYLLABLE-B50B
        0xB50C, # G+11100  U+B50C  uniB50C,  HANGUL SYLLABLE-B50C
        0xB50D, # G+11101  U+B50D  uniB50D,  HANGUL SYLLABLE-B50D
        0xB50E, # G+11102  U+B50E  uniB50E,  HANGUL SYLLABLE-B50E
        0xB50F, # G+11103  U+B50F  uniB50F,  HANGUL SYLLABLE-B50F
        0xB510, # G+11104  U+B510  uniB510,  HANGUL SYLLABLE-B510
        0xB511, # G+11105  U+B511  uniB511,  HANGUL SYLLABLE-B511
        0xB512, # G+11106  U+B512  uniB512,  HANGUL SYLLABLE-B512
        0xB513, # G+11107  U+B513  uniB513,  HANGUL SYLLABLE-B513
        0xB516, # G+11108  U+B516  uniB516,  HANGUL SYLLABLE-B516
        0xB517, # G+11109  U+B517  uniB517,  HANGUL SYLLABLE-B517
        0xB519, # G+11110  U+B519  uniB519,  HANGUL SYLLABLE-B519
        0xB51A, # G+11111  U+B51A  uniB51A,  HANGUL SYLLABLE-B51A
        0xB51D, # G+11112  U+B51D  uniB51D,  HANGUL SYLLABLE-B51D
        0xB51E, # G+11113  U+B51E  uniB51E,  HANGUL SYLLABLE-B51E
        0xB51F, # G+11114  U+B51F  uniB51F,  HANGUL SYLLABLE-B51F
        0xB520, # G+11115  U+B520  uniB520,  HANGUL SYLLABLE-B520
        0xB521, # G+11116  U+B521  uniB521,  HANGUL SYLLABLE-B521
        0xB522, # G+11117  U+B522  uniB522,  HANGUL SYLLABLE-B522
        0xB523, # G+11118  U+B523  uniB523,  HANGUL SYLLABLE-B523
        0xB526, # G+11119  U+B526  uniB526,  HANGUL SYLLABLE-B526
        0xB52B, # G+11120  U+B52B  uniB52B,  HANGUL SYLLABLE-B52B
        0xB52C, # G+11121  U+B52C  uniB52C,  HANGUL SYLLABLE-B52C
        0xB52D, # G+11122  U+B52D  uniB52D,  HANGUL SYLLABLE-B52D
        0xB52E, # G+11123  U+B52E  uniB52E,  HANGUL SYLLABLE-B52E
        0xB52F, # G+11124  U+B52F  uniB52F,  HANGUL SYLLABLE-B52F
        0xB532, # G+11125  U+B532  uniB532,  HANGUL SYLLABLE-B532
        0xB533, # G+11126  U+B533  uniB533,  HANGUL SYLLABLE-B533
        0xB535, # G+11127  U+B535  uniB535,  HANGUL SYLLABLE-B535
        0xB536, # G+11128  U+B536  uniB536,  HANGUL SYLLABLE-B536
        0xB537, # G+11129  U+B537  uniB537,  HANGUL SYLLABLE-B537
        0xB539, # G+11130  U+B539  uniB539,  HANGUL SYLLABLE-B539
        0xB53A, # G+11131  U+B53A  uniB53A,  HANGUL SYLLABLE-B53A
        0xB53B, # G+11132  U+B53B  uniB53B,  HANGUL SYLLABLE-B53B
        0xB53C, # G+11133  U+B53C  uniB53C,  HANGUL SYLLABLE-B53C
        0xB53D, # G+11134  U+B53D  uniB53D,  HANGUL SYLLABLE-B53D
        0xB53E, # G+11135  U+B53E  uniB53E,  HANGUL SYLLABLE-B53E
        0xB53F, # G+11136  U+B53F  uniB53F,  HANGUL SYLLABLE-B53F
        0xB542, # G+11137  U+B542  uniB542,  HANGUL SYLLABLE-B542
        0xB546, # G+11138  U+B546  uniB546,  HANGUL SYLLABLE-B546
        0xB547, # G+11139  U+B547  uniB547,  HANGUL SYLLABLE-B547
        0xB548, # G+11140  U+B548  uniB548,  HANGUL SYLLABLE-B548
        0xB549, # G+11141  U+B549  uniB549,  HANGUL SYLLABLE-B549
        0xB54A, # G+11142  U+B54A  uniB54A,  HANGUL SYLLABLE-B54A
        0xB54E, # G+11143  U+B54E  uniB54E,  HANGUL SYLLABLE-B54E
        0xB54F, # G+11144  U+B54F  uniB54F,  HANGUL SYLLABLE-B54F
        0xB551, # G+11145  U+B551  uniB551,  HANGUL SYLLABLE-B551
        0xB552, # G+11146  U+B552  uniB552,  HANGUL SYLLABLE-B552
        0xB553, # G+11147  U+B553  uniB553,  HANGUL SYLLABLE-B553
        0xB555, # G+11148  U+B555  uniB555,  HANGUL SYLLABLE-B555
        0xB556, # G+11149  U+B556  uniB556,  HANGUL SYLLABLE-B556
        0xB557, # G+11150  U+B557  uniB557,  HANGUL SYLLABLE-B557
        0xB558, # G+11151  U+B558  uniB558,  HANGUL SYLLABLE-B558
        0xB559, # G+11152  U+B559  uniB559,  HANGUL SYLLABLE-B559
        0xB55A, # G+11153  U+B55A  uniB55A,  HANGUL SYLLABLE-B55A
        0xB55B, # G+11154  U+B55B  uniB55B,  HANGUL SYLLABLE-B55B
        0xB55E, # G+11155  U+B55E  uniB55E,  HANGUL SYLLABLE-B55E
        0xB562, # G+11156  U+B562  uniB562,  HANGUL SYLLABLE-B562
        0xB563, # G+11157  U+B563  uniB563,  HANGUL SYLLABLE-B563
        0xB564, # G+11158  U+B564  uniB564,  HANGUL SYLLABLE-B564
        0xB565, # G+11159  U+B565  uniB565,  HANGUL SYLLABLE-B565
        0xB566, # G+11160  U+B566  uniB566,  HANGUL SYLLABLE-B566
        0xB567, # G+11161  U+B567  uniB567,  HANGUL SYLLABLE-B567
        0xB568, # G+11162  U+B568  uniB568,  HANGUL SYLLABLE-B568
        0xB569, # G+11163  U+B569  uniB569,  HANGUL SYLLABLE-B569
        0xB56A, # G+11164  U+B56A  uniB56A,  HANGUL SYLLABLE-B56A
        0xB56B, # G+11165  U+B56B  uniB56B,  HANGUL SYLLABLE-B56B
        0xB56C, # G+11166  U+B56C  uniB56C,  HANGUL SYLLABLE-B56C
        0xB56D, # G+11167  U+B56D  uniB56D,  HANGUL SYLLABLE-B56D
        0xB56E, # G+11168  U+B56E  uniB56E,  HANGUL SYLLABLE-B56E
        0xB56F, # G+11169  U+B56F  uniB56F,  HANGUL SYLLABLE-B56F
        0xB570, # G+11170  U+B570  uniB570,  HANGUL SYLLABLE-B570
        0xB571, # G+11171  U+B571  uniB571,  HANGUL SYLLABLE-B571
        0xB572, # G+11172  U+B572  uniB572,  HANGUL SYLLABLE-B572
        0xB573, # G+11173  U+B573  uniB573,  HANGUL SYLLABLE-B573
        0xB574, # G+11174  U+B574  uniB574,  HANGUL SYLLABLE-B574
        0xB575, # G+11175  U+B575  uniB575,  HANGUL SYLLABLE-B575
        0xB576, # G+11176  U+B576  uniB576,  HANGUL SYLLABLE-B576
        0xB577, # G+11177  U+B577  uniB577,  HANGUL SYLLABLE-B577
        0xB578, # G+11178  U+B578  uniB578,  HANGUL SYLLABLE-B578
        0xB579, # G+11179  U+B579  uniB579,  HANGUL SYLLABLE-B579
        0xB57A, # G+11180  U+B57A  uniB57A,  HANGUL SYLLABLE-B57A
        0xB57B, # G+11181  U+B57B  uniB57B,  HANGUL SYLLABLE-B57B
        0xB57C, # G+11182  U+B57C  uniB57C,  HANGUL SYLLABLE-B57C
        0xB57D, # G+11183  U+B57D  uniB57D,  HANGUL SYLLABLE-B57D
        0xB57E, # G+11184  U+B57E  uniB57E,  HANGUL SYLLABLE-B57E
        0xB57F, # G+11185  U+B57F  uniB57F,  HANGUL SYLLABLE-B57F
        0xB580, # G+11186  U+B580  uniB580,  HANGUL SYLLABLE-B580
        0xB581, # G+11187  U+B581  uniB581,  HANGUL SYLLABLE-B581
        0xB582, # G+11188  U+B582  uniB582,  HANGUL SYLLABLE-B582
        0xB583, # G+11189  U+B583  uniB583,  HANGUL SYLLABLE-B583
        0xB584, # G+11190  U+B584  uniB584,  HANGUL SYLLABLE-B584
        0xB585, # G+11191  U+B585  uniB585,  HANGUL SYLLABLE-B585
        0xB586, # G+11192  U+B586  uniB586,  HANGUL SYLLABLE-B586
        0xB587, # G+11193  U+B587  uniB587,  HANGUL SYLLABLE-B587
        0xB588, # G+11194  U+B588  uniB588,  HANGUL SYLLABLE-B588
        0xB589, # G+11195  U+B589  uniB589,  HANGUL SYLLABLE-B589
        0xB58A, # G+11196  U+B58A  uniB58A,  HANGUL SYLLABLE-B58A
        0xB58B, # G+11197  U+B58B  uniB58B,  HANGUL SYLLABLE-B58B
        0xB58C, # G+11198  U+B58C  uniB58C,  HANGUL SYLLABLE-B58C
        0xB58D, # G+11199  U+B58D  uniB58D,  HANGUL SYLLABLE-B58D
        0xB58E, # G+11200  U+B58E  uniB58E,  HANGUL SYLLABLE-B58E
        0xB58F, # G+11201  U+B58F  uniB58F,  HANGUL SYLLABLE-B58F
        0xB590, # G+11202  U+B590  uniB590,  HANGUL SYLLABLE-B590
        0xB591, # G+11203  U+B591  uniB591,  HANGUL SYLLABLE-B591
        0xB592, # G+11204  U+B592  uniB592,  HANGUL SYLLABLE-B592
        0xB593, # G+11205  U+B593  uniB593,  HANGUL SYLLABLE-B593
        0xB594, # G+11206  U+B594  uniB594,  HANGUL SYLLABLE-B594
        0xB595, # G+11207  U+B595  uniB595,  HANGUL SYLLABLE-B595
        0xB596, # G+11208  U+B596  uniB596,  HANGUL SYLLABLE-B596
        0xB597, # G+11209  U+B597  uniB597,  HANGUL SYLLABLE-B597
        0xB598, # G+11210  U+B598  uniB598,  HANGUL SYLLABLE-B598
        0xB599, # G+11211  U+B599  uniB599,  HANGUL SYLLABLE-B599
        0xB59A, # G+11212  U+B59A  uniB59A,  HANGUL SYLLABLE-B59A
        0xB59B, # G+11213  U+B59B  uniB59B,  HANGUL SYLLABLE-B59B
        0xB59C, # G+11214  U+B59C  uniB59C,  HANGUL SYLLABLE-B59C
        0xB59D, # G+11215  U+B59D  uniB59D,  HANGUL SYLLABLE-B59D
        0xB59E, # G+11216  U+B59E  uniB59E,  HANGUL SYLLABLE-B59E
        0xB59F, # G+11217  U+B59F  uniB59F,  HANGUL SYLLABLE-B59F
        0xB5A2, # G+11218  U+B5A2  uniB5A2,  HANGUL SYLLABLE-B5A2
        0xB5A3, # G+11219  U+B5A3  uniB5A3,  HANGUL SYLLABLE-B5A3
        0xB5A5, # G+11220  U+B5A5  uniB5A5,  HANGUL SYLLABLE-B5A5
        0xB5A6, # G+11221  U+B5A6  uniB5A6,  HANGUL SYLLABLE-B5A6
        0xB5A7, # G+11222  U+B5A7  uniB5A7,  HANGUL SYLLABLE-B5A7
        0xB5A9, # G+11223  U+B5A9  uniB5A9,  HANGUL SYLLABLE-B5A9
        0xB5AC, # G+11224  U+B5AC  uniB5AC,  HANGUL SYLLABLE-B5AC
        0xB5AD, # G+11225  U+B5AD  uniB5AD,  HANGUL SYLLABLE-B5AD
        0xB5AE, # G+11226  U+B5AE  uniB5AE,  HANGUL SYLLABLE-B5AE
        0xB5AF, # G+11227  U+B5AF  uniB5AF,  HANGUL SYLLABLE-B5AF
        0xB5B2, # G+11228  U+B5B2  uniB5B2,  HANGUL SYLLABLE-B5B2
        0xB5B6, # G+11229  U+B5B6  uniB5B6,  HANGUL SYLLABLE-B5B6
        0xB5B7, # G+11230  U+B5B7  uniB5B7,  HANGUL SYLLABLE-B5B7
        0xB5B8, # G+11231  U+B5B8  uniB5B8,  HANGUL SYLLABLE-B5B8
        0xB5B9, # G+11232  U+B5B9  uniB5B9,  HANGUL SYLLABLE-B5B9
        0xB5BA, # G+11233  U+B5BA  uniB5BA,  HANGUL SYLLABLE-B5BA
        0xB5BE, # G+11234  U+B5BE  uniB5BE,  HANGUL SYLLABLE-B5BE
        0xB5BF, # G+11235  U+B5BF  uniB5BF,  HANGUL SYLLABLE-B5BF
        0xB5C1, # G+11236  U+B5C1  uniB5C1,  HANGUL SYLLABLE-B5C1
        0xB5C2, # G+11237  U+B5C2  uniB5C2,  HANGUL SYLLABLE-B5C2
        0xB5C3, # G+11238  U+B5C3  uniB5C3,  HANGUL SYLLABLE-B5C3
        0xB5C5, # G+11239  U+B5C5  uniB5C5,  HANGUL SYLLABLE-B5C5
        0xB5C6, # G+11240  U+B5C6  uniB5C6,  HANGUL SYLLABLE-B5C6
        0xB5C7, # G+11241  U+B5C7  uniB5C7,  HANGUL SYLLABLE-B5C7
        0xB5C8, # G+11242  U+B5C8  uniB5C8,  HANGUL SYLLABLE-B5C8
        0xB5C9, # G+11243  U+B5C9  uniB5C9,  HANGUL SYLLABLE-B5C9
        0xB5CA, # G+11244  U+B5CA  uniB5CA,  HANGUL SYLLABLE-B5CA
        0xB5CB, # G+11245  U+B5CB  uniB5CB,  HANGUL SYLLABLE-B5CB
        0xB5CE, # G+11246  U+B5CE  uniB5CE,  HANGUL SYLLABLE-B5CE
        0xB5D2, # G+11247  U+B5D2  uniB5D2,  HANGUL SYLLABLE-B5D2
        0xB5D3, # G+11248  U+B5D3  uniB5D3,  HANGUL SYLLABLE-B5D3
        0xB5D4, # G+11249  U+B5D4  uniB5D4,  HANGUL SYLLABLE-B5D4
        0xB5D5, # G+11250  U+B5D5  uniB5D5,  HANGUL SYLLABLE-B5D5
        0xB5D6, # G+11251  U+B5D6  uniB5D6,  HANGUL SYLLABLE-B5D6
        0xB5D7, # G+11252  U+B5D7  uniB5D7,  HANGUL SYLLABLE-B5D7
        0xB5D9, # G+11253  U+B5D9  uniB5D9,  HANGUL SYLLABLE-B5D9
        0xB5DA, # G+11254  U+B5DA  uniB5DA,  HANGUL SYLLABLE-B5DA
        0xB5DB, # G+11255  U+B5DB  uniB5DB,  HANGUL SYLLABLE-B5DB
        0xB5DC, # G+11256  U+B5DC  uniB5DC,  HANGUL SYLLABLE-B5DC
        0xB5DD, # G+11257  U+B5DD  uniB5DD,  HANGUL SYLLABLE-B5DD
        0xB5DE, # G+11258  U+B5DE  uniB5DE,  HANGUL SYLLABLE-B5DE
        0xB5DF, # G+11259  U+B5DF  uniB5DF,  HANGUL SYLLABLE-B5DF
        0xB5E0, # G+11260  U+B5E0  uniB5E0,  HANGUL SYLLABLE-B5E0
        0xB5E1, # G+11261  U+B5E1  uniB5E1,  HANGUL SYLLABLE-B5E1
        0xB5E2, # G+11262  U+B5E2  uniB5E2,  HANGUL SYLLABLE-B5E2
        0xB5E3, # G+11263  U+B5E3  uniB5E3,  HANGUL SYLLABLE-B5E3
        0xB5E4, # G+11264  U+B5E4  uniB5E4,  HANGUL SYLLABLE-B5E4
        0xB5E5, # G+11265  U+B5E5  uniB5E5,  HANGUL SYLLABLE-B5E5
        0xB5E6, # G+11266  U+B5E6  uniB5E6,  HANGUL SYLLABLE-B5E6
        0xB5E7, # G+11267  U+B5E7  uniB5E7,  HANGUL SYLLABLE-B5E7
        0xB5E8, # G+11268  U+B5E8  uniB5E8,  HANGUL SYLLABLE-B5E8
        0xB5E9, # G+11269  U+B5E9  uniB5E9,  HANGUL SYLLABLE-B5E9
        0xB5EA, # G+11270  U+B5EA  uniB5EA,  HANGUL SYLLABLE-B5EA
        0xB5EB, # G+11271  U+B5EB  uniB5EB,  HANGUL SYLLABLE-B5EB
        0xB5ED, # G+11272  U+B5ED  uniB5ED,  HANGUL SYLLABLE-B5ED
        0xB5EE, # G+11273  U+B5EE  uniB5EE,  HANGUL SYLLABLE-B5EE
        0xB5EF, # G+11274  U+B5EF  uniB5EF,  HANGUL SYLLABLE-B5EF
        0xB5F0, # G+11275  U+B5F0  uniB5F0,  HANGUL SYLLABLE-B5F0
        0xB5F1, # G+11276  U+B5F1  uniB5F1,  HANGUL SYLLABLE-B5F1
        0xB5F2, # G+11277  U+B5F2  uniB5F2,  HANGUL SYLLABLE-B5F2
        0xB5F3, # G+11278  U+B5F3  uniB5F3,  HANGUL SYLLABLE-B5F3
        0xB5F4, # G+11279  U+B5F4  uniB5F4,  HANGUL SYLLABLE-B5F4
        0xB5F5, # G+11280  U+B5F5  uniB5F5,  HANGUL SYLLABLE-B5F5
        0xB5F6, # G+11281  U+B5F6  uniB5F6,  HANGUL SYLLABLE-B5F6
        0xB5F7, # G+11282  U+B5F7  uniB5F7,  HANGUL SYLLABLE-B5F7
        0xB5F8, # G+11283  U+B5F8  uniB5F8,  HANGUL SYLLABLE-B5F8
        0xB5F9, # G+11284  U+B5F9  uniB5F9,  HANGUL SYLLABLE-B5F9
        0xB5FA, # G+11285  U+B5FA  uniB5FA,  HANGUL SYLLABLE-B5FA
        0xB5FB, # G+11286  U+B5FB  uniB5FB,  HANGUL SYLLABLE-B5FB
        0xB5FC, # G+11287  U+B5FC  uniB5FC,  HANGUL SYLLABLE-B5FC
        0xB5FD, # G+11288  U+B5FD  uniB5FD,  HANGUL SYLLABLE-B5FD
        0xB5FE, # G+11289  U+B5FE  uniB5FE,  HANGUL SYLLABLE-B5FE
        0xB5FF, # G+11290  U+B5FF  uniB5FF,  HANGUL SYLLABLE-B5FF
        0xB600, # G+11291  U+B600  uniB600,  HANGUL SYLLABLE-B600
        0xB601, # G+11292  U+B601  uniB601,  HANGUL SYLLABLE-B601
        0xB602, # G+11293  U+B602  uniB602,  HANGUL SYLLABLE-B602
        0xB603, # G+11294  U+B603  uniB603,  HANGUL SYLLABLE-B603
        0xB604, # G+11295  U+B604  uniB604,  HANGUL SYLLABLE-B604
        0xB605, # G+11296  U+B605  uniB605,  HANGUL SYLLABLE-B605
        0xB606, # G+11297  U+B606  uniB606,  HANGUL SYLLABLE-B606
        0xB607, # G+11298  U+B607  uniB607,  HANGUL SYLLABLE-B607
        0xB608, # G+11299  U+B608  uniB608,  HANGUL SYLLABLE-B608
        0xB609, # G+11300  U+B609  uniB609,  HANGUL SYLLABLE-B609
        0xB60A, # G+11301  U+B60A  uniB60A,  HANGUL SYLLABLE-B60A
        0xB60B, # G+11302  U+B60B  uniB60B,  HANGUL SYLLABLE-B60B
        0xB60C, # G+11303  U+B60C  uniB60C,  HANGUL SYLLABLE-B60C
        0xB60D, # G+11304  U+B60D  uniB60D,  HANGUL SYLLABLE-B60D
        0xB60E, # G+11305  U+B60E  uniB60E,  HANGUL SYLLABLE-B60E
        0xB60F, # G+11306  U+B60F  uniB60F,  HANGUL SYLLABLE-B60F
        0xB612, # G+11307  U+B612  uniB612,  HANGUL SYLLABLE-B612
        0xB613, # G+11308  U+B613  uniB613,  HANGUL SYLLABLE-B613
        0xB615, # G+11309  U+B615  uniB615,  HANGUL SYLLABLE-B615
        0xB616, # G+11310  U+B616  uniB616,  HANGUL SYLLABLE-B616
        0xB617, # G+11311  U+B617  uniB617,  HANGUL SYLLABLE-B617
        0xB619, # G+11312  U+B619  uniB619,  HANGUL SYLLABLE-B619
        0xB61A, # G+11313  U+B61A  uniB61A,  HANGUL SYLLABLE-B61A
        0xB61B, # G+11314  U+B61B  uniB61B,  HANGUL SYLLABLE-B61B
        0xB61C, # G+11315  U+B61C  uniB61C,  HANGUL SYLLABLE-B61C
        0xB61D, # G+11316  U+B61D  uniB61D,  HANGUL SYLLABLE-B61D
        0xB61E, # G+11317  U+B61E  uniB61E,  HANGUL SYLLABLE-B61E
        0xB61F, # G+11318  U+B61F  uniB61F,  HANGUL SYLLABLE-B61F
        0xB620, # G+11319  U+B620  uniB620,  HANGUL SYLLABLE-B620
        0xB621, # G+11320  U+B621  uniB621,  HANGUL SYLLABLE-B621
        0xB622, # G+11321  U+B622  uniB622,  HANGUL SYLLABLE-B622
        0xB623, # G+11322  U+B623  uniB623,  HANGUL SYLLABLE-B623
        0xB624, # G+11323  U+B624  uniB624,  HANGUL SYLLABLE-B624
        0xB626, # G+11324  U+B626  uniB626,  HANGUL SYLLABLE-B626
        0xB627, # G+11325  U+B627  uniB627,  HANGUL SYLLABLE-B627
        0xB628, # G+11326  U+B628  uniB628,  HANGUL SYLLABLE-B628
        0xB629, # G+11327  U+B629  uniB629,  HANGUL SYLLABLE-B629
        0xB62A, # G+11328  U+B62A  uniB62A,  HANGUL SYLLABLE-B62A
        0xB62B, # G+11329  U+B62B  uniB62B,  HANGUL SYLLABLE-B62B
        0xB62D, # G+11330  U+B62D  uniB62D,  HANGUL SYLLABLE-B62D
        0xB62E, # G+11331  U+B62E  uniB62E,  HANGUL SYLLABLE-B62E
        0xB62F, # G+11332  U+B62F  uniB62F,  HANGUL SYLLABLE-B62F
        0xB630, # G+11333  U+B630  uniB630,  HANGUL SYLLABLE-B630
        0xB631, # G+11334  U+B631  uniB631,  HANGUL SYLLABLE-B631
        0xB632, # G+11335  U+B632  uniB632,  HANGUL SYLLABLE-B632
        0xB633, # G+11336  U+B633  uniB633,  HANGUL SYLLABLE-B633
        0xB635, # G+11337  U+B635  uniB635,  HANGUL SYLLABLE-B635
        0xB636, # G+11338  U+B636  uniB636,  HANGUL SYLLABLE-B636
        0xB637, # G+11339  U+B637  uniB637,  HANGUL SYLLABLE-B637
        0xB638, # G+11340  U+B638  uniB638,  HANGUL SYLLABLE-B638
        0xB639, # G+11341  U+B639  uniB639,  HANGUL SYLLABLE-B639
        0xB63A, # G+11342  U+B63A  uniB63A,  HANGUL SYLLABLE-B63A
        0xB63B, # G+11343  U+B63B  uniB63B,  HANGUL SYLLABLE-B63B
        0xB63C, # G+11344  U+B63C  uniB63C,  HANGUL SYLLABLE-B63C
        0xB63D, # G+11345  U+B63D  uniB63D,  HANGUL SYLLABLE-B63D
        0xB63E, # G+11346  U+B63E  uniB63E,  HANGUL SYLLABLE-B63E
        0xB63F, # G+11347  U+B63F  uniB63F,  HANGUL SYLLABLE-B63F
        0xB640, # G+11348  U+B640  uniB640,  HANGUL SYLLABLE-B640
        0xB641, # G+11349  U+B641  uniB641,  HANGUL SYLLABLE-B641
        0xB642, # G+11350  U+B642  uniB642,  HANGUL SYLLABLE-B642
        0xB643, # G+11351  U+B643  uniB643,  HANGUL SYLLABLE-B643
        0xB644, # G+11352  U+B644  uniB644,  HANGUL SYLLABLE-B644
        0xB645, # G+11353  U+B645  uniB645,  HANGUL SYLLABLE-B645
        0xB646, # G+11354  U+B646  uniB646,  HANGUL SYLLABLE-B646
        0xB647, # G+11355  U+B647  uniB647,  HANGUL SYLLABLE-B647
        0xB649, # G+11356  U+B649  uniB649,  HANGUL SYLLABLE-B649
        0xB64A, # G+11357  U+B64A  uniB64A,  HANGUL SYLLABLE-B64A
        0xB64B, # G+11358  U+B64B  uniB64B,  HANGUL SYLLABLE-B64B
        0xB64C, # G+11359  U+B64C  uniB64C,  HANGUL SYLLABLE-B64C
        0xB64D, # G+11360  U+B64D  uniB64D,  HANGUL SYLLABLE-B64D
        0xB64E, # G+11361  U+B64E  uniB64E,  HANGUL SYLLABLE-B64E
        0xB64F, # G+11362  U+B64F  uniB64F,  HANGUL SYLLABLE-B64F
        0xB650, # G+11363  U+B650  uniB650,  HANGUL SYLLABLE-B650
        0xB651, # G+11364  U+B651  uniB651,  HANGUL SYLLABLE-B651
        0xB652, # G+11365  U+B652  uniB652,  HANGUL SYLLABLE-B652
        0xB653, # G+11366  U+B653  uniB653,  HANGUL SYLLABLE-B653
        0xB654, # G+11367  U+B654  uniB654,  HANGUL SYLLABLE-B654
        0xB655, # G+11368  U+B655  uniB655,  HANGUL SYLLABLE-B655
        0xB656, # G+11369  U+B656  uniB656,  HANGUL SYLLABLE-B656
        0xB657, # G+11370  U+B657  uniB657,  HANGUL SYLLABLE-B657
        0xB658, # G+11371  U+B658  uniB658,  HANGUL SYLLABLE-B658
        0xB659, # G+11372  U+B659  uniB659,  HANGUL SYLLABLE-B659
        0xB65A, # G+11373  U+B65A  uniB65A,  HANGUL SYLLABLE-B65A
        0xB65B, # G+11374  U+B65B  uniB65B,  HANGUL SYLLABLE-B65B
        0xB65C, # G+11375  U+B65C  uniB65C,  HANGUL SYLLABLE-B65C
        0xB65D, # G+11376  U+B65D  uniB65D,  HANGUL SYLLABLE-B65D
        0xB65E, # G+11377  U+B65E  uniB65E,  HANGUL SYLLABLE-B65E
        0xB65F, # G+11378  U+B65F  uniB65F,  HANGUL SYLLABLE-B65F
        0xB660, # G+11379  U+B660  uniB660,  HANGUL SYLLABLE-B660
        0xB661, # G+11380  U+B661  uniB661,  HANGUL SYLLABLE-B661
        0xB662, # G+11381  U+B662  uniB662,  HANGUL SYLLABLE-B662
        0xB663, # G+11382  U+B663  uniB663,  HANGUL SYLLABLE-B663
        0xB665, # G+11383  U+B665  uniB665,  HANGUL SYLLABLE-B665
        0xB666, # G+11384  U+B666  uniB666,  HANGUL SYLLABLE-B666
        0xB667, # G+11385  U+B667  uniB667,  HANGUL SYLLABLE-B667
        0xB669, # G+11386  U+B669  uniB669,  HANGUL SYLLABLE-B669
        0xB66A, # G+11387  U+B66A  uniB66A,  HANGUL SYLLABLE-B66A
        0xB66B, # G+11388  U+B66B  uniB66B,  HANGUL SYLLABLE-B66B
        0xB66C, # G+11389  U+B66C  uniB66C,  HANGUL SYLLABLE-B66C
        0xB66D, # G+11390  U+B66D  uniB66D,  HANGUL SYLLABLE-B66D
        0xB66E, # G+11391  U+B66E  uniB66E,  HANGUL SYLLABLE-B66E
        0xB66F, # G+11392  U+B66F  uniB66F,  HANGUL SYLLABLE-B66F
        0xB670, # G+11393  U+B670  uniB670,  HANGUL SYLLABLE-B670
        0xB671, # G+11394  U+B671  uniB671,  HANGUL SYLLABLE-B671
        0xB672, # G+11395  U+B672  uniB672,  HANGUL SYLLABLE-B672
        0xB673, # G+11396  U+B673  uniB673,  HANGUL SYLLABLE-B673
        0xB674, # G+11397  U+B674  uniB674,  HANGUL SYLLABLE-B674
        0xB675, # G+11398  U+B675  uniB675,  HANGUL SYLLABLE-B675
        0xB676, # G+11399  U+B676  uniB676,  HANGUL SYLLABLE-B676
        0xB677, # G+11400  U+B677  uniB677,  HANGUL SYLLABLE-B677
        0xB678, # G+11401  U+B678  uniB678,  HANGUL SYLLABLE-B678
        0xB679, # G+11402  U+B679  uniB679,  HANGUL SYLLABLE-B679
        0xB67A, # G+11403  U+B67A  uniB67A,  HANGUL SYLLABLE-B67A
        0xB67B, # G+11404  U+B67B  uniB67B,  HANGUL SYLLABLE-B67B
        0xB67C, # G+11405  U+B67C  uniB67C,  HANGUL SYLLABLE-B67C
        0xB67D, # G+11406  U+B67D  uniB67D,  HANGUL SYLLABLE-B67D
        0xB67E, # G+11407  U+B67E  uniB67E,  HANGUL SYLLABLE-B67E
        0xB67F, # G+11408  U+B67F  uniB67F,  HANGUL SYLLABLE-B67F
        0xB680, # G+11409  U+B680  uniB680,  HANGUL SYLLABLE-B680
        0xB681, # G+11410  U+B681  uniB681,  HANGUL SYLLABLE-B681
        0xB682, # G+11411  U+B682  uniB682,  HANGUL SYLLABLE-B682
        0xB683, # G+11412  U+B683  uniB683,  HANGUL SYLLABLE-B683
        0xB684, # G+11413  U+B684  uniB684,  HANGUL SYLLABLE-B684
        0xB685, # G+11414  U+B685  uniB685,  HANGUL SYLLABLE-B685
        0xB686, # G+11415  U+B686  uniB686,  HANGUL SYLLABLE-B686
        0xB687, # G+11416  U+B687  uniB687,  HANGUL SYLLABLE-B687
        0xB688, # G+11417  U+B688  uniB688,  HANGUL SYLLABLE-B688
        0xB689, # G+11418  U+B689  uniB689,  HANGUL SYLLABLE-B689
        0xB68A, # G+11419  U+B68A  uniB68A,  HANGUL SYLLABLE-B68A
        0xB68B, # G+11420  U+B68B  uniB68B,  HANGUL SYLLABLE-B68B
        0xB68C, # G+11421  U+B68C  uniB68C,  HANGUL SYLLABLE-B68C
        0xB68D, # G+11422  U+B68D  uniB68D,  HANGUL SYLLABLE-B68D
        0xB68E, # G+11423  U+B68E  uniB68E,  HANGUL SYLLABLE-B68E
        0xB68F, # G+11424  U+B68F  uniB68F,  HANGUL SYLLABLE-B68F
        0xB690, # G+11425  U+B690  uniB690,  HANGUL SYLLABLE-B690
        0xB691, # G+11426  U+B691  uniB691,  HANGUL SYLLABLE-B691
        0xB692, # G+11427  U+B692  uniB692,  HANGUL SYLLABLE-B692
        0xB693, # G+11428  U+B693  uniB693,  HANGUL SYLLABLE-B693
        0xB694, # G+11429  U+B694  uniB694,  HANGUL SYLLABLE-B694
        0xB695, # G+11430  U+B695  uniB695,  HANGUL SYLLABLE-B695
        0xB696, # G+11431  U+B696  uniB696,  HANGUL SYLLABLE-B696
        0xB697, # G+11432  U+B697  uniB697,  HANGUL SYLLABLE-B697
        0xB698, # G+11433  U+B698  uniB698,  HANGUL SYLLABLE-B698
        0xB699, # G+11434  U+B699  uniB699,  HANGUL SYLLABLE-B699
        0xB69A, # G+11435  U+B69A  uniB69A,  HANGUL SYLLABLE-B69A
        0xB69B, # G+11436  U+B69B  uniB69B,  HANGUL SYLLABLE-B69B
        0xB69E, # G+11437  U+B69E  uniB69E,  HANGUL SYLLABLE-B69E
        0xB69F, # G+11438  U+B69F  uniB69F,  HANGUL SYLLABLE-B69F
        0xB6A1, # G+11439  U+B6A1  uniB6A1,  HANGUL SYLLABLE-B6A1
        0xB6A2, # G+11440  U+B6A2  uniB6A2,  HANGUL SYLLABLE-B6A2
        0xB6A3, # G+11441  U+B6A3  uniB6A3,  HANGUL SYLLABLE-B6A3
        0xB6A5, # G+11442  U+B6A5  uniB6A5,  HANGUL SYLLABLE-B6A5
        0xB6A6, # G+11443  U+B6A6  uniB6A6,  HANGUL SYLLABLE-B6A6
        0xB6A7, # G+11444  U+B6A7  uniB6A7,  HANGUL SYLLABLE-B6A7
        0xB6A8, # G+11445  U+B6A8  uniB6A8,  HANGUL SYLLABLE-B6A8
        0xB6A9, # G+11446  U+B6A9  uniB6A9,  HANGUL SYLLABLE-B6A9
        0xB6AA, # G+11447  U+B6AA  uniB6AA,  HANGUL SYLLABLE-B6AA
        0xB6AD, # G+11448  U+B6AD  uniB6AD,  HANGUL SYLLABLE-B6AD
        0xB6AE, # G+11449  U+B6AE  uniB6AE,  HANGUL SYLLABLE-B6AE
        0xB6AF, # G+11450  U+B6AF  uniB6AF,  HANGUL SYLLABLE-B6AF
        0xB6B0, # G+11451  U+B6B0  uniB6B0,  HANGUL SYLLABLE-B6B0
        0xB6B2, # G+11452  U+B6B2  uniB6B2,  HANGUL SYLLABLE-B6B2
        0xB6B3, # G+11453  U+B6B3  uniB6B3,  HANGUL SYLLABLE-B6B3
        0xB6B4, # G+11454  U+B6B4  uniB6B4,  HANGUL SYLLABLE-B6B4
        0xB6B5, # G+11455  U+B6B5  uniB6B5,  HANGUL SYLLABLE-B6B5
        0xB6B6, # G+11456  U+B6B6  uniB6B6,  HANGUL SYLLABLE-B6B6
        0xB6B7, # G+11457  U+B6B7  uniB6B7,  HANGUL SYLLABLE-B6B7
        0xB6B8, # G+11458  U+B6B8  uniB6B8,  HANGUL SYLLABLE-B6B8
        0xB6B9, # G+11459  U+B6B9  uniB6B9,  HANGUL SYLLABLE-B6B9
        0xB6BA, # G+11460  U+B6BA  uniB6BA,  HANGUL SYLLABLE-B6BA
        0xB6BB, # G+11461  U+B6BB  uniB6BB,  HANGUL SYLLABLE-B6BB
        0xB6BC, # G+11462  U+B6BC  uniB6BC,  HANGUL SYLLABLE-B6BC
        0xB6BD, # G+11463  U+B6BD  uniB6BD,  HANGUL SYLLABLE-B6BD
        0xB6BE, # G+11464  U+B6BE  uniB6BE,  HANGUL SYLLABLE-B6BE
        0xB6BF, # G+11465  U+B6BF  uniB6BF,  HANGUL SYLLABLE-B6BF
        0xB6C0, # G+11466  U+B6C0  uniB6C0,  HANGUL SYLLABLE-B6C0
        0xB6C1, # G+11467  U+B6C1  uniB6C1,  HANGUL SYLLABLE-B6C1
        0xB6C2, # G+11468  U+B6C2  uniB6C2,  HANGUL SYLLABLE-B6C2
        0xB6C3, # G+11469  U+B6C3  uniB6C3,  HANGUL SYLLABLE-B6C3
        0xB6C4, # G+11470  U+B6C4  uniB6C4,  HANGUL SYLLABLE-B6C4
        0xB6C5, # G+11471  U+B6C5  uniB6C5,  HANGUL SYLLABLE-B6C5
        0xB6C6, # G+11472  U+B6C6  uniB6C6,  HANGUL SYLLABLE-B6C6
        0xB6C7, # G+11473  U+B6C7  uniB6C7,  HANGUL SYLLABLE-B6C7
        0xB6C8, # G+11474  U+B6C8  uniB6C8,  HANGUL SYLLABLE-B6C8
        0xB6C9, # G+11475  U+B6C9  uniB6C9,  HANGUL SYLLABLE-B6C9
        0xB6CA, # G+11476  U+B6CA  uniB6CA,  HANGUL SYLLABLE-B6CA
        0xB6CB, # G+11477  U+B6CB  uniB6CB,  HANGUL SYLLABLE-B6CB
        0xB6CC, # G+11478  U+B6CC  uniB6CC,  HANGUL SYLLABLE-B6CC
        0xB6CD, # G+11479  U+B6CD  uniB6CD,  HANGUL SYLLABLE-B6CD
        0xB6CE, # G+11480  U+B6CE  uniB6CE,  HANGUL SYLLABLE-B6CE
        0xB6CF, # G+11481  U+B6CF  uniB6CF,  HANGUL SYLLABLE-B6CF
        0xB6D0, # G+11482  U+B6D0  uniB6D0,  HANGUL SYLLABLE-B6D0
        0xB6D1, # G+11483  U+B6D1  uniB6D1,  HANGUL SYLLABLE-B6D1
        0xB6D2, # G+11484  U+B6D2  uniB6D2,  HANGUL SYLLABLE-B6D2
        0xB6D3, # G+11485  U+B6D3  uniB6D3,  HANGUL SYLLABLE-B6D3
        0xB6D5, # G+11486  U+B6D5  uniB6D5,  HANGUL SYLLABLE-B6D5
        0xB6D6, # G+11487  U+B6D6  uniB6D6,  HANGUL SYLLABLE-B6D6
        0xB6D7, # G+11488  U+B6D7  uniB6D7,  HANGUL SYLLABLE-B6D7
        0xB6D8, # G+11489  U+B6D8  uniB6D8,  HANGUL SYLLABLE-B6D8
        0xB6D9, # G+11490  U+B6D9  uniB6D9,  HANGUL SYLLABLE-B6D9
        0xB6DA, # G+11491  U+B6DA  uniB6DA,  HANGUL SYLLABLE-B6DA
        0xB6DB, # G+11492  U+B6DB  uniB6DB,  HANGUL SYLLABLE-B6DB
        0xB6DC, # G+11493  U+B6DC  uniB6DC,  HANGUL SYLLABLE-B6DC
        0xB6DD, # G+11494  U+B6DD  uniB6DD,  HANGUL SYLLABLE-B6DD
        0xB6DE, # G+11495  U+B6DE  uniB6DE,  HANGUL SYLLABLE-B6DE
        0xB6DF, # G+11496  U+B6DF  uniB6DF,  HANGUL SYLLABLE-B6DF
        0xB6E0, # G+11497  U+B6E0  uniB6E0,  HANGUL SYLLABLE-B6E0
        0xB6E1, # G+11498  U+B6E1  uniB6E1,  HANGUL SYLLABLE-B6E1
        0xB6E2, # G+11499  U+B6E2  uniB6E2,  HANGUL SYLLABLE-B6E2
        0xB6E3, # G+11500  U+B6E3  uniB6E3,  HANGUL SYLLABLE-B6E3
        0xB6E4, # G+11501  U+B6E4  uniB6E4,  HANGUL SYLLABLE-B6E4
        0xB6E5, # G+11502  U+B6E5  uniB6E5,  HANGUL SYLLABLE-B6E5
        0xB6E6, # G+11503  U+B6E6  uniB6E6,  HANGUL SYLLABLE-B6E6
        0xB6E7, # G+11504  U+B6E7  uniB6E7,  HANGUL SYLLABLE-B6E7
        0xB6E8, # G+11505  U+B6E8  uniB6E8,  HANGUL SYLLABLE-B6E8
        0xB6E9, # G+11506  U+B6E9  uniB6E9,  HANGUL SYLLABLE-B6E9
        0xB6EA, # G+11507  U+B6EA  uniB6EA,  HANGUL SYLLABLE-B6EA
        0xB6EB, # G+11508  U+B6EB  uniB6EB,  HANGUL SYLLABLE-B6EB
        0xB6EC, # G+11509  U+B6EC  uniB6EC,  HANGUL SYLLABLE-B6EC
        0xB6ED, # G+11510  U+B6ED  uniB6ED,  HANGUL SYLLABLE-B6ED
        0xB6EE, # G+11511  U+B6EE  uniB6EE,  HANGUL SYLLABLE-B6EE
        0xB6EF, # G+11512  U+B6EF  uniB6EF,  HANGUL SYLLABLE-B6EF
        0xB6F1, # G+11513  U+B6F1  uniB6F1,  HANGUL SYLLABLE-B6F1
        0xB6F2, # G+11514  U+B6F2  uniB6F2,  HANGUL SYLLABLE-B6F2
        0xB6F3, # G+11515  U+B6F3  uniB6F3,  HANGUL SYLLABLE-B6F3
        0xB6F5, # G+11516  U+B6F5  uniB6F5,  HANGUL SYLLABLE-B6F5
        0xB6F6, # G+11517  U+B6F6  uniB6F6,  HANGUL SYLLABLE-B6F6
        0xB6F7, # G+11518  U+B6F7  uniB6F7,  HANGUL SYLLABLE-B6F7
        0xB6F9, # G+11519  U+B6F9  uniB6F9,  HANGUL SYLLABLE-B6F9
        0xB6FA, # G+11520  U+B6FA  uniB6FA,  HANGUL SYLLABLE-B6FA
        0xB6FB, # G+11521  U+B6FB  uniB6FB,  HANGUL SYLLABLE-B6FB
        0xB6FC, # G+11522  U+B6FC  uniB6FC,  HANGUL SYLLABLE-B6FC
        0xB6FD, # G+11523  U+B6FD  uniB6FD,  HANGUL SYLLABLE-B6FD
        0xB6FE, # G+11524  U+B6FE  uniB6FE,  HANGUL SYLLABLE-B6FE
        0xB6FF, # G+11525  U+B6FF  uniB6FF,  HANGUL SYLLABLE-B6FF
        0xB702, # G+11526  U+B702  uniB702,  HANGUL SYLLABLE-B702
        0xB703, # G+11527  U+B703  uniB703,  HANGUL SYLLABLE-B703
        0xB704, # G+11528  U+B704  uniB704,  HANGUL SYLLABLE-B704
        0xB706, # G+11529  U+B706  uniB706,  HANGUL SYLLABLE-B706
        0xB707, # G+11530  U+B707  uniB707,  HANGUL SYLLABLE-B707
        0xB708, # G+11531  U+B708  uniB708,  HANGUL SYLLABLE-B708
        0xB709, # G+11532  U+B709  uniB709,  HANGUL SYLLABLE-B709
        0xB70A, # G+11533  U+B70A  uniB70A,  HANGUL SYLLABLE-B70A
        0xB70B, # G+11534  U+B70B  uniB70B,  HANGUL SYLLABLE-B70B
        0xB70C, # G+11535  U+B70C  uniB70C,  HANGUL SYLLABLE-B70C
        0xB70D, # G+11536  U+B70D  uniB70D,  HANGUL SYLLABLE-B70D
        0xB70E, # G+11537  U+B70E  uniB70E,  HANGUL SYLLABLE-B70E
        0xB70F, # G+11538  U+B70F  uniB70F,  HANGUL SYLLABLE-B70F
        0xB710, # G+11539  U+B710  uniB710,  HANGUL SYLLABLE-B710
        0xB711, # G+11540  U+B711  uniB711,  HANGUL SYLLABLE-B711
        0xB712, # G+11541  U+B712  uniB712,  HANGUL SYLLABLE-B712
        0xB713, # G+11542  U+B713  uniB713,  HANGUL SYLLABLE-B713
        0xB714, # G+11543  U+B714  uniB714,  HANGUL SYLLABLE-B714
        0xB715, # G+11544  U+B715  uniB715,  HANGUL SYLLABLE-B715
        0xB716, # G+11545  U+B716  uniB716,  HANGUL SYLLABLE-B716
        0xB717, # G+11546  U+B717  uniB717,  HANGUL SYLLABLE-B717
        0xB718, # G+11547  U+B718  uniB718,  HANGUL SYLLABLE-B718
        0xB719, # G+11548  U+B719  uniB719,  HANGUL SYLLABLE-B719
        0xB71A, # G+11549  U+B71A  uniB71A,  HANGUL SYLLABLE-B71A
        0xB71B, # G+11550  U+B71B  uniB71B,  HANGUL SYLLABLE-B71B
        0xB71C, # G+11551  U+B71C  uniB71C,  HANGUL SYLLABLE-B71C
        0xB71D, # G+11552  U+B71D  uniB71D,  HANGUL SYLLABLE-B71D
        0xB71E, # G+11553  U+B71E  uniB71E,  HANGUL SYLLABLE-B71E
        0xB71F, # G+11554  U+B71F  uniB71F,  HANGUL SYLLABLE-B71F
        0xB720, # G+11555  U+B720  uniB720,  HANGUL SYLLABLE-B720
        0xB721, # G+11556  U+B721  uniB721,  HANGUL SYLLABLE-B721
        0xB722, # G+11557  U+B722  uniB722,  HANGUL SYLLABLE-B722
        0xB723, # G+11558  U+B723  uniB723,  HANGUL SYLLABLE-B723
        0xB724, # G+11559  U+B724  uniB724,  HANGUL SYLLABLE-B724
        0xB725, # G+11560  U+B725  uniB725,  HANGUL SYLLABLE-B725
        0xB726, # G+11561  U+B726  uniB726,  HANGUL SYLLABLE-B726
        0xB727, # G+11562  U+B727  uniB727,  HANGUL SYLLABLE-B727
        0xB72A, # G+11563  U+B72A  uniB72A,  HANGUL SYLLABLE-B72A
        0xB72B, # G+11564  U+B72B  uniB72B,  HANGUL SYLLABLE-B72B
        0xB72D, # G+11565  U+B72D  uniB72D,  HANGUL SYLLABLE-B72D
        0xB72E, # G+11566  U+B72E  uniB72E,  HANGUL SYLLABLE-B72E
        0xB731, # G+11567  U+B731  uniB731,  HANGUL SYLLABLE-B731
        0xB732, # G+11568  U+B732  uniB732,  HANGUL SYLLABLE-B732
        0xB733, # G+11569  U+B733  uniB733,  HANGUL SYLLABLE-B733
        0xB734, # G+11570  U+B734  uniB734,  HANGUL SYLLABLE-B734
        0xB735, # G+11571  U+B735  uniB735,  HANGUL SYLLABLE-B735
        0xB736, # G+11572  U+B736  uniB736,  HANGUL SYLLABLE-B736
        0xB737, # G+11573  U+B737  uniB737,  HANGUL SYLLABLE-B737
        0xB73A, # G+11574  U+B73A  uniB73A,  HANGUL SYLLABLE-B73A
        0xB73C, # G+11575  U+B73C  uniB73C,  HANGUL SYLLABLE-B73C
        0xB73D, # G+11576  U+B73D  uniB73D,  HANGUL SYLLABLE-B73D
        0xB73E, # G+11577  U+B73E  uniB73E,  HANGUL SYLLABLE-B73E
        0xB73F, # G+11578  U+B73F  uniB73F,  HANGUL SYLLABLE-B73F
        0xB740, # G+11579  U+B740  uniB740,  HANGUL SYLLABLE-B740
        0xB741, # G+11580  U+B741  uniB741,  HANGUL SYLLABLE-B741
        0xB742, # G+11581  U+B742  uniB742,  HANGUL SYLLABLE-B742
        0xB743, # G+11582  U+B743  uniB743,  HANGUL SYLLABLE-B743
        0xB745, # G+11583  U+B745  uniB745,  HANGUL SYLLABLE-B745
        0xB746, # G+11584  U+B746  uniB746,  HANGUL SYLLABLE-B746
        0xB747, # G+11585  U+B747  uniB747,  HANGUL SYLLABLE-B747
        0xB749, # G+11586  U+B749  uniB749,  HANGUL SYLLABLE-B749
        0xB74A, # G+11587  U+B74A  uniB74A,  HANGUL SYLLABLE-B74A
        0xB74B, # G+11588  U+B74B  uniB74B,  HANGUL SYLLABLE-B74B
        0xB74D, # G+11589  U+B74D  uniB74D,  HANGUL SYLLABLE-B74D
        0xB74E, # G+11590  U+B74E  uniB74E,  HANGUL SYLLABLE-B74E
        0xB74F, # G+11591  U+B74F  uniB74F,  HANGUL SYLLABLE-B74F
        0xB750, # G+11592  U+B750  uniB750,  HANGUL SYLLABLE-B750
        0xB751, # G+11593  U+B751  uniB751,  HANGUL SYLLABLE-B751
        0xB752, # G+11594  U+B752  uniB752,  HANGUL SYLLABLE-B752
        0xB753, # G+11595  U+B753  uniB753,  HANGUL SYLLABLE-B753
        0xB756, # G+11596  U+B756  uniB756,  HANGUL SYLLABLE-B756
        0xB757, # G+11597  U+B757  uniB757,  HANGUL SYLLABLE-B757
        0xB758, # G+11598  U+B758  uniB758,  HANGUL SYLLABLE-B758
        0xB759, # G+11599  U+B759  uniB759,  HANGUL SYLLABLE-B759
        0xB75A, # G+11600  U+B75A  uniB75A,  HANGUL SYLLABLE-B75A
        0xB75B, # G+11601  U+B75B  uniB75B,  HANGUL SYLLABLE-B75B
        0xB75C, # G+11602  U+B75C  uniB75C,  HANGUL SYLLABLE-B75C
        0xB75D, # G+11603  U+B75D  uniB75D,  HANGUL SYLLABLE-B75D
        0xB75E, # G+11604  U+B75E  uniB75E,  HANGUL SYLLABLE-B75E
        0xB75F, # G+11605  U+B75F  uniB75F,  HANGUL SYLLABLE-B75F
        0xB761, # G+11606  U+B761  uniB761,  HANGUL SYLLABLE-B761
        0xB762, # G+11607  U+B762  uniB762,  HANGUL SYLLABLE-B762
        0xB763, # G+11608  U+B763  uniB763,  HANGUL SYLLABLE-B763
        0xB765, # G+11609  U+B765  uniB765,  HANGUL SYLLABLE-B765
        0xB766, # G+11610  U+B766  uniB766,  HANGUL SYLLABLE-B766
        0xB767, # G+11611  U+B767  uniB767,  HANGUL SYLLABLE-B767
        0xB769, # G+11612  U+B769  uniB769,  HANGUL SYLLABLE-B769
        0xB76A, # G+11613  U+B76A  uniB76A,  HANGUL SYLLABLE-B76A
        0xB76B, # G+11614  U+B76B  uniB76B,  HANGUL SYLLABLE-B76B
        0xB76C, # G+11615  U+B76C  uniB76C,  HANGUL SYLLABLE-B76C
        0xB76D, # G+11616  U+B76D  uniB76D,  HANGUL SYLLABLE-B76D
        0xB76E, # G+11617  U+B76E  uniB76E,  HANGUL SYLLABLE-B76E
        0xB76F, # G+11618  U+B76F  uniB76F,  HANGUL SYLLABLE-B76F
        0xB772, # G+11619  U+B772  uniB772,  HANGUL SYLLABLE-B772
        0xB774, # G+11620  U+B774  uniB774,  HANGUL SYLLABLE-B774
        0xB776, # G+11621  U+B776  uniB776,  HANGUL SYLLABLE-B776
        0xB777, # G+11622  U+B777  uniB777,  HANGUL SYLLABLE-B777
        0xB778, # G+11623  U+B778  uniB778,  HANGUL SYLLABLE-B778
        0xB779, # G+11624  U+B779  uniB779,  HANGUL SYLLABLE-B779
        0xB77A, # G+11625  U+B77A  uniB77A,  HANGUL SYLLABLE-B77A
        0xB77B, # G+11626  U+B77B  uniB77B,  HANGUL SYLLABLE-B77B
        0xB77E, # G+11627  U+B77E  uniB77E,  HANGUL SYLLABLE-B77E
        0xB77F, # G+11628  U+B77F  uniB77F,  HANGUL SYLLABLE-B77F
        0xB781, # G+11629  U+B781  uniB781,  HANGUL SYLLABLE-B781
        0xB782, # G+11630  U+B782  uniB782,  HANGUL SYLLABLE-B782
        0xB783, # G+11631  U+B783  uniB783,  HANGUL SYLLABLE-B783
        0xB785, # G+11632  U+B785  uniB785,  HANGUL SYLLABLE-B785
        0xB786, # G+11633  U+B786  uniB786,  HANGUL SYLLABLE-B786
        0xB787, # G+11634  U+B787  uniB787,  HANGUL SYLLABLE-B787
        0xB788, # G+11635  U+B788  uniB788,  HANGUL SYLLABLE-B788
        0xB789, # G+11636  U+B789  uniB789,  HANGUL SYLLABLE-B789
        0xB78A, # G+11637  U+B78A  uniB78A,  HANGUL SYLLABLE-B78A
        0xB78B, # G+11638  U+B78B  uniB78B,  HANGUL SYLLABLE-B78B
        0xB78E, # G+11639  U+B78E  uniB78E,  HANGUL SYLLABLE-B78E
        0xB793, # G+11640  U+B793  uniB793,  HANGUL SYLLABLE-B793
        0xB794, # G+11641  U+B794  uniB794,  HANGUL SYLLABLE-B794
        0xB795, # G+11642  U+B795  uniB795,  HANGUL SYLLABLE-B795
        0xB79A, # G+11643  U+B79A  uniB79A,  HANGUL SYLLABLE-B79A
        0xB79B, # G+11644  U+B79B  uniB79B,  HANGUL SYLLABLE-B79B
        0xB79D, # G+11645  U+B79D  uniB79D,  HANGUL SYLLABLE-B79D
        0xB79E, # G+11646  U+B79E  uniB79E,  HANGUL SYLLABLE-B79E
        0xB79F, # G+11647  U+B79F  uniB79F,  HANGUL SYLLABLE-B79F
        0xB7A1, # G+11648  U+B7A1  uniB7A1,  HANGUL SYLLABLE-B7A1
        0xB7A2, # G+11649  U+B7A2  uniB7A2,  HANGUL SYLLABLE-B7A2
        0xB7A3, # G+11650  U+B7A3  uniB7A3,  HANGUL SYLLABLE-B7A3
        0xB7A4, # G+11651  U+B7A4  uniB7A4,  HANGUL SYLLABLE-B7A4
        0xB7A5, # G+11652  U+B7A5  uniB7A5,  HANGUL SYLLABLE-B7A5
        0xB7A6, # G+11653  U+B7A6  uniB7A6,  HANGUL SYLLABLE-B7A6
        0xB7A7, # G+11654  U+B7A7  uniB7A7,  HANGUL SYLLABLE-B7A7
        0xB7AA, # G+11655  U+B7AA  uniB7AA,  HANGUL SYLLABLE-B7AA
        0xB7AE, # G+11656  U+B7AE  uniB7AE,  HANGUL SYLLABLE-B7AE
        0xB7AF, # G+11657  U+B7AF  uniB7AF,  HANGUL SYLLABLE-B7AF
        0xB7B0, # G+11658  U+B7B0  uniB7B0,  HANGUL SYLLABLE-B7B0
        0xB7B1, # G+11659  U+B7B1  uniB7B1,  HANGUL SYLLABLE-B7B1
        0xB7B2, # G+11660  U+B7B2  uniB7B2,  HANGUL SYLLABLE-B7B2
        0xB7B3, # G+11661  U+B7B3  uniB7B3,  HANGUL SYLLABLE-B7B3
        0xB7B6, # G+11662  U+B7B6  uniB7B6,  HANGUL SYLLABLE-B7B6
        0xB7B7, # G+11663  U+B7B7  uniB7B7,  HANGUL SYLLABLE-B7B7
        0xB7B9, # G+11664  U+B7B9  uniB7B9,  HANGUL SYLLABLE-B7B9
        0xB7BA, # G+11665  U+B7BA  uniB7BA,  HANGUL SYLLABLE-B7BA
        0xB7BB, # G+11666  U+B7BB  uniB7BB,  HANGUL SYLLABLE-B7BB
        0xB7BC, # G+11667  U+B7BC  uniB7BC,  HANGUL SYLLABLE-B7BC
        0xB7BD, # G+11668  U+B7BD  uniB7BD,  HANGUL SYLLABLE-B7BD
        0xB7BE, # G+11669  U+B7BE  uniB7BE,  HANGUL SYLLABLE-B7BE
        0xB7BF, # G+11670  U+B7BF  uniB7BF,  HANGUL SYLLABLE-B7BF
        0xB7C0, # G+11671  U+B7C0  uniB7C0,  HANGUL SYLLABLE-B7C0
        0xB7C1, # G+11672  U+B7C1  uniB7C1,  HANGUL SYLLABLE-B7C1
        0xB7C2, # G+11673  U+B7C2  uniB7C2,  HANGUL SYLLABLE-B7C2
        0xB7C3, # G+11674  U+B7C3  uniB7C3,  HANGUL SYLLABLE-B7C3
        0xB7C4, # G+11675  U+B7C4  uniB7C4,  HANGUL SYLLABLE-B7C4
        0xB7C5, # G+11676  U+B7C5  uniB7C5,  HANGUL SYLLABLE-B7C5
        0xB7C6, # G+11677  U+B7C6  uniB7C6,  HANGUL SYLLABLE-B7C6
        0xB7C8, # G+11678  U+B7C8  uniB7C8,  HANGUL SYLLABLE-B7C8
        0xB7CA, # G+11679  U+B7CA  uniB7CA,  HANGUL SYLLABLE-B7CA
        0xB7CB, # G+11680  U+B7CB  uniB7CB,  HANGUL SYLLABLE-B7CB
        0xB7CC, # G+11681  U+B7CC  uniB7CC,  HANGUL SYLLABLE-B7CC
        0xB7CD, # G+11682  U+B7CD  uniB7CD,  HANGUL SYLLABLE-B7CD
        0xB7CE, # G+11683  U+B7CE  uniB7CE,  HANGUL SYLLABLE-B7CE
        0xB7CF, # G+11684  U+B7CF  uniB7CF,  HANGUL SYLLABLE-B7CF
        0xB7D0, # G+11685  U+B7D0  uniB7D0,  HANGUL SYLLABLE-B7D0
        0xB7D1, # G+11686  U+B7D1  uniB7D1,  HANGUL SYLLABLE-B7D1
        0xB7D2, # G+11687  U+B7D2  uniB7D2,  HANGUL SYLLABLE-B7D2
        0xB7D3, # G+11688  U+B7D3  uniB7D3,  HANGUL SYLLABLE-B7D3
        0xB7D4, # G+11689  U+B7D4  uniB7D4,  HANGUL SYLLABLE-B7D4
        0xB7D5, # G+11690  U+B7D5  uniB7D5,  HANGUL SYLLABLE-B7D5
        0xB7D6, # G+11691  U+B7D6  uniB7D6,  HANGUL SYLLABLE-B7D6
        0xB7D7, # G+11692  U+B7D7  uniB7D7,  HANGUL SYLLABLE-B7D7
        0xB7D8, # G+11693  U+B7D8  uniB7D8,  HANGUL SYLLABLE-B7D8
        0xB7D9, # G+11694  U+B7D9  uniB7D9,  HANGUL SYLLABLE-B7D9
        0xB7DA, # G+11695  U+B7DA  uniB7DA,  HANGUL SYLLABLE-B7DA
        0xB7DB, # G+11696  U+B7DB  uniB7DB,  HANGUL SYLLABLE-B7DB
        0xB7DC, # G+11697  U+B7DC  uniB7DC,  HANGUL SYLLABLE-B7DC
        0xB7DD, # G+11698  U+B7DD  uniB7DD,  HANGUL SYLLABLE-B7DD
        0xB7DE, # G+11699  U+B7DE  uniB7DE,  HANGUL SYLLABLE-B7DE
        0xB7DF, # G+11700  U+B7DF  uniB7DF,  HANGUL SYLLABLE-B7DF
        0xB7E0, # G+11701  U+B7E0  uniB7E0,  HANGUL SYLLABLE-B7E0
        0xB7E1, # G+11702  U+B7E1  uniB7E1,  HANGUL SYLLABLE-B7E1
        0xB7E2, # G+11703  U+B7E2  uniB7E2,  HANGUL SYLLABLE-B7E2
        0xB7E3, # G+11704  U+B7E3  uniB7E3,  HANGUL SYLLABLE-B7E3
        0xB7E4, # G+11705  U+B7E4  uniB7E4,  HANGUL SYLLABLE-B7E4
        0xB7E5, # G+11706  U+B7E5  uniB7E5,  HANGUL SYLLABLE-B7E5
        0xB7E6, # G+11707  U+B7E6  uniB7E6,  HANGUL SYLLABLE-B7E6
        0xB7E7, # G+11708  U+B7E7  uniB7E7,  HANGUL SYLLABLE-B7E7
        0xB7E8, # G+11709  U+B7E8  uniB7E8,  HANGUL SYLLABLE-B7E8
        0xB7E9, # G+11710  U+B7E9  uniB7E9,  HANGUL SYLLABLE-B7E9
        0xB7EA, # G+11711  U+B7EA  uniB7EA,  HANGUL SYLLABLE-B7EA
        0xB7EB, # G+11712  U+B7EB  uniB7EB,  HANGUL SYLLABLE-B7EB
        0xB7EE, # G+11713  U+B7EE  uniB7EE,  HANGUL SYLLABLE-B7EE
        0xB7EF, # G+11714  U+B7EF  uniB7EF,  HANGUL SYLLABLE-B7EF
        0xB7F1, # G+11715  U+B7F1  uniB7F1,  HANGUL SYLLABLE-B7F1
        0xB7F2, # G+11716  U+B7F2  uniB7F2,  HANGUL SYLLABLE-B7F2
        0xB7F3, # G+11717  U+B7F3  uniB7F3,  HANGUL SYLLABLE-B7F3
        0xB7F5, # G+11718  U+B7F5  uniB7F5,  HANGUL SYLLABLE-B7F5
        0xB7F6, # G+11719  U+B7F6  uniB7F6,  HANGUL SYLLABLE-B7F6
        0xB7F7, # G+11720  U+B7F7  uniB7F7,  HANGUL SYLLABLE-B7F7
        0xB7F8, # G+11721  U+B7F8  uniB7F8,  HANGUL SYLLABLE-B7F8
        0xB7F9, # G+11722  U+B7F9  uniB7F9,  HANGUL SYLLABLE-B7F9
        0xB7FA, # G+11723  U+B7FA  uniB7FA,  HANGUL SYLLABLE-B7FA
        0xB7FB, # G+11724  U+B7FB  uniB7FB,  HANGUL SYLLABLE-B7FB
        0xB7FE, # G+11725  U+B7FE  uniB7FE,  HANGUL SYLLABLE-B7FE
        0xB802, # G+11726  U+B802  uniB802,  HANGUL SYLLABLE-B802
        0xB803, # G+11727  U+B803  uniB803,  HANGUL SYLLABLE-B803
        0xB804, # G+11728  U+B804  uniB804,  HANGUL SYLLABLE-B804
        0xB805, # G+11729  U+B805  uniB805,  HANGUL SYLLABLE-B805
        0xB806, # G+11730  U+B806  uniB806,  HANGUL SYLLABLE-B806
        0xB80A, # G+11731  U+B80A  uniB80A,  HANGUL SYLLABLE-B80A
        0xB80B, # G+11732  U+B80B  uniB80B,  HANGUL SYLLABLE-B80B
        0xB80D, # G+11733  U+B80D  uniB80D,  HANGUL SYLLABLE-B80D
        0xB80E, # G+11734  U+B80E  uniB80E,  HANGUL SYLLABLE-B80E
        0xB80F, # G+11735  U+B80F  uniB80F,  HANGUL SYLLABLE-B80F
        0xB811, # G+11736  U+B811  uniB811,  HANGUL SYLLABLE-B811
        0xB812, # G+11737  U+B812  uniB812,  HANGUL SYLLABLE-B812
        0xB813, # G+11738  U+B813  uniB813,  HANGUL SYLLABLE-B813
        0xB814, # G+11739  U+B814  uniB814,  HANGUL SYLLABLE-B814
        0xB815, # G+11740  U+B815  uniB815,  HANGUL SYLLABLE-B815
        0xB816, # G+11741  U+B816  uniB816,  HANGUL SYLLABLE-B816
        0xB817, # G+11742  U+B817  uniB817,  HANGUL SYLLABLE-B817
        0xB81A, # G+11743  U+B81A  uniB81A,  HANGUL SYLLABLE-B81A
        0xB81C, # G+11744  U+B81C  uniB81C,  HANGUL SYLLABLE-B81C
        0xB81E, # G+11745  U+B81E  uniB81E,  HANGUL SYLLABLE-B81E
        0xB81F, # G+11746  U+B81F  uniB81F,  HANGUL SYLLABLE-B81F
        0xB820, # G+11747  U+B820  uniB820,  HANGUL SYLLABLE-B820
        0xB821, # G+11748  U+B821  uniB821,  HANGUL SYLLABLE-B821
        0xB822, # G+11749  U+B822  uniB822,  HANGUL SYLLABLE-B822
        0xB823, # G+11750  U+B823  uniB823,  HANGUL SYLLABLE-B823
        0xB826, # G+11751  U+B826  uniB826,  HANGUL SYLLABLE-B826
        0xB827, # G+11752  U+B827  uniB827,  HANGUL SYLLABLE-B827
        0xB829, # G+11753  U+B829  uniB829,  HANGUL SYLLABLE-B829
        0xB82A, # G+11754  U+B82A  uniB82A,  HANGUL SYLLABLE-B82A
        0xB82B, # G+11755  U+B82B  uniB82B,  HANGUL SYLLABLE-B82B
        0xB82D, # G+11756  U+B82D  uniB82D,  HANGUL SYLLABLE-B82D
        0xB82E, # G+11757  U+B82E  uniB82E,  HANGUL SYLLABLE-B82E
        0xB82F, # G+11758  U+B82F  uniB82F,  HANGUL SYLLABLE-B82F
        0xB830, # G+11759  U+B830  uniB830,  HANGUL SYLLABLE-B830
        0xB831, # G+11760  U+B831  uniB831,  HANGUL SYLLABLE-B831
        0xB832, # G+11761  U+B832  uniB832,  HANGUL SYLLABLE-B832
        0xB833, # G+11762  U+B833  uniB833,  HANGUL SYLLABLE-B833
        0xB836, # G+11763  U+B836  uniB836,  HANGUL SYLLABLE-B836
        0xB83A, # G+11764  U+B83A  uniB83A,  HANGUL SYLLABLE-B83A
        0xB83B, # G+11765  U+B83B  uniB83B,  HANGUL SYLLABLE-B83B
        0xB83C, # G+11766  U+B83C  uniB83C,  HANGUL SYLLABLE-B83C
        0xB83D, # G+11767  U+B83D  uniB83D,  HANGUL SYLLABLE-B83D
        0xB83E, # G+11768  U+B83E  uniB83E,  HANGUL SYLLABLE-B83E
        0xB83F, # G+11769  U+B83F  uniB83F,  HANGUL SYLLABLE-B83F
        0xB841, # G+11770  U+B841  uniB841,  HANGUL SYLLABLE-B841
        0xB842, # G+11771  U+B842  uniB842,  HANGUL SYLLABLE-B842
        0xB843, # G+11772  U+B843  uniB843,  HANGUL SYLLABLE-B843
        0xB845, # G+11773  U+B845  uniB845,  HANGUL SYLLABLE-B845
        0xB846, # G+11774  U+B846  uniB846,  HANGUL SYLLABLE-B846
        0xB847, # G+11775  U+B847  uniB847,  HANGUL SYLLABLE-B847
        0xB848, # G+11776  U+B848  uniB848,  HANGUL SYLLABLE-B848
        0xB849, # G+11777  U+B849  uniB849,  HANGUL SYLLABLE-B849
        0xB84A, # G+11778  U+B84A  uniB84A,  HANGUL SYLLABLE-B84A
        0xB84B, # G+11779  U+B84B  uniB84B,  HANGUL SYLLABLE-B84B
        0xB84C, # G+11780  U+B84C  uniB84C,  HANGUL SYLLABLE-B84C
        0xB84D, # G+11781  U+B84D  uniB84D,  HANGUL SYLLABLE-B84D
        0xB84E, # G+11782  U+B84E  uniB84E,  HANGUL SYLLABLE-B84E
        0xB84F, # G+11783  U+B84F  uniB84F,  HANGUL SYLLABLE-B84F
        0xB850, # G+11784  U+B850  uniB850,  HANGUL SYLLABLE-B850
        0xB852, # G+11785  U+B852  uniB852,  HANGUL SYLLABLE-B852
        0xB854, # G+11786  U+B854  uniB854,  HANGUL SYLLABLE-B854
        0xB855, # G+11787  U+B855  uniB855,  HANGUL SYLLABLE-B855
        0xB856, # G+11788  U+B856  uniB856,  HANGUL SYLLABLE-B856
        0xB857, # G+11789  U+B857  uniB857,  HANGUL SYLLABLE-B857
        0xB858, # G+11790  U+B858  uniB858,  HANGUL SYLLABLE-B858
        0xB859, # G+11791  U+B859  uniB859,  HANGUL SYLLABLE-B859
        0xB85A, # G+11792  U+B85A  uniB85A,  HANGUL SYLLABLE-B85A
        0xB85B, # G+11793  U+B85B  uniB85B,  HANGUL SYLLABLE-B85B
        0xB85E, # G+11794  U+B85E  uniB85E,  HANGUL SYLLABLE-B85E
        0xB85F, # G+11795  U+B85F  uniB85F,  HANGUL SYLLABLE-B85F
        0xB861, # G+11796  U+B861  uniB861,  HANGUL SYLLABLE-B861
        0xB862, # G+11797  U+B862  uniB862,  HANGUL SYLLABLE-B862
        0xB863, # G+11798  U+B863  uniB863,  HANGUL SYLLABLE-B863
        0xB865, # G+11799  U+B865  uniB865,  HANGUL SYLLABLE-B865
        0xB866, # G+11800  U+B866  uniB866,  HANGUL SYLLABLE-B866
        0xB867, # G+11801  U+B867  uniB867,  HANGUL SYLLABLE-B867
        0xB868, # G+11802  U+B868  uniB868,  HANGUL SYLLABLE-B868
        0xB869, # G+11803  U+B869  uniB869,  HANGUL SYLLABLE-B869
        0xB86A, # G+11804  U+B86A  uniB86A,  HANGUL SYLLABLE-B86A
        0xB86B, # G+11805  U+B86B  uniB86B,  HANGUL SYLLABLE-B86B
        0xB86E, # G+11806  U+B86E  uniB86E,  HANGUL SYLLABLE-B86E
        0xB870, # G+11807  U+B870  uniB870,  HANGUL SYLLABLE-B870
        0xB872, # G+11808  U+B872  uniB872,  HANGUL SYLLABLE-B872
        0xB873, # G+11809  U+B873  uniB873,  HANGUL SYLLABLE-B873
        0xB874, # G+11810  U+B874  uniB874,  HANGUL SYLLABLE-B874
        0xB875, # G+11811  U+B875  uniB875,  HANGUL SYLLABLE-B875
        0xB876, # G+11812  U+B876  uniB876,  HANGUL SYLLABLE-B876
        0xB877, # G+11813  U+B877  uniB877,  HANGUL SYLLABLE-B877
        0xB879, # G+11814  U+B879  uniB879,  HANGUL SYLLABLE-B879
        0xB87A, # G+11815  U+B87A  uniB87A,  HANGUL SYLLABLE-B87A
        0xB87B, # G+11816  U+B87B  uniB87B,  HANGUL SYLLABLE-B87B
        0xB87D, # G+11817  U+B87D  uniB87D,  HANGUL SYLLABLE-B87D
        0xB87E, # G+11818  U+B87E  uniB87E,  HANGUL SYLLABLE-B87E
        0xB87F, # G+11819  U+B87F  uniB87F,  HANGUL SYLLABLE-B87F
        0xB880, # G+11820  U+B880  uniB880,  HANGUL SYLLABLE-B880
        0xB881, # G+11821  U+B881  uniB881,  HANGUL SYLLABLE-B881
        0xB882, # G+11822  U+B882  uniB882,  HANGUL SYLLABLE-B882
        0xB883, # G+11823  U+B883  uniB883,  HANGUL SYLLABLE-B883
        0xB884, # G+11824  U+B884  uniB884,  HANGUL SYLLABLE-B884
        0xB885, # G+11825  U+B885  uniB885,  HANGUL SYLLABLE-B885
        0xB886, # G+11826  U+B886  uniB886,  HANGUL SYLLABLE-B886
        0xB887, # G+11827  U+B887  uniB887,  HANGUL SYLLABLE-B887
        0xB888, # G+11828  U+B888  uniB888,  HANGUL SYLLABLE-B888
        0xB889, # G+11829  U+B889  uniB889,  HANGUL SYLLABLE-B889
        0xB88A, # G+11830  U+B88A  uniB88A,  HANGUL SYLLABLE-B88A
        0xB88B, # G+11831  U+B88B  uniB88B,  HANGUL SYLLABLE-B88B
        0xB88C, # G+11832  U+B88C  uniB88C,  HANGUL SYLLABLE-B88C
        0xB88E, # G+11833  U+B88E  uniB88E,  HANGUL SYLLABLE-B88E
        0xB88F, # G+11834  U+B88F  uniB88F,  HANGUL SYLLABLE-B88F
        0xB890, # G+11835  U+B890  uniB890,  HANGUL SYLLABLE-B890
        0xB891, # G+11836  U+B891  uniB891,  HANGUL SYLLABLE-B891
        0xB892, # G+11837  U+B892  uniB892,  HANGUL SYLLABLE-B892
        0xB893, # G+11838  U+B893  uniB893,  HANGUL SYLLABLE-B893
        0xB894, # G+11839  U+B894  uniB894,  HANGUL SYLLABLE-B894
        0xB895, # G+11840  U+B895  uniB895,  HANGUL SYLLABLE-B895
        0xB896, # G+11841  U+B896  uniB896,  HANGUL SYLLABLE-B896
        0xB897, # G+11842  U+B897  uniB897,  HANGUL SYLLABLE-B897
        0xB898, # G+11843  U+B898  uniB898,  HANGUL SYLLABLE-B898
        0xB899, # G+11844  U+B899  uniB899,  HANGUL SYLLABLE-B899
        0xB89A, # G+11845  U+B89A  uniB89A,  HANGUL SYLLABLE-B89A
        0xB89B, # G+11846  U+B89B  uniB89B,  HANGUL SYLLABLE-B89B
        0xB89C, # G+11847  U+B89C  uniB89C,  HANGUL SYLLABLE-B89C
        0xB89D, # G+11848  U+B89D  uniB89D,  HANGUL SYLLABLE-B89D
        0xB89E, # G+11849  U+B89E  uniB89E,  HANGUL SYLLABLE-B89E
        0xB89F, # G+11850  U+B89F  uniB89F,  HANGUL SYLLABLE-B89F
        0xB8A0, # G+11851  U+B8A0  uniB8A0,  HANGUL SYLLABLE-B8A0
        0xB8A1, # G+11852  U+B8A1  uniB8A1,  HANGUL SYLLABLE-B8A1
        0xB8A2, # G+11853  U+B8A2  uniB8A2,  HANGUL SYLLABLE-B8A2
        0xB8A3, # G+11854  U+B8A3  uniB8A3,  HANGUL SYLLABLE-B8A3
        0xB8A4, # G+11855  U+B8A4  uniB8A4,  HANGUL SYLLABLE-B8A4
        0xB8A5, # G+11856  U+B8A5  uniB8A5,  HANGUL SYLLABLE-B8A5
        0xB8A6, # G+11857  U+B8A6  uniB8A6,  HANGUL SYLLABLE-B8A6
        0xB8A7, # G+11858  U+B8A7  uniB8A7,  HANGUL SYLLABLE-B8A7
        0xB8A9, # G+11859  U+B8A9  uniB8A9,  HANGUL SYLLABLE-B8A9
        0xB8AA, # G+11860  U+B8AA  uniB8AA,  HANGUL SYLLABLE-B8AA
        0xB8AB, # G+11861  U+B8AB  uniB8AB,  HANGUL SYLLABLE-B8AB
        0xB8AC, # G+11862  U+B8AC  uniB8AC,  HANGUL SYLLABLE-B8AC
        0xB8AD, # G+11863  U+B8AD  uniB8AD,  HANGUL SYLLABLE-B8AD
        0xB8AE, # G+11864  U+B8AE  uniB8AE,  HANGUL SYLLABLE-B8AE
        0xB8AF, # G+11865  U+B8AF  uniB8AF,  HANGUL SYLLABLE-B8AF
        0xB8B1, # G+11866  U+B8B1  uniB8B1,  HANGUL SYLLABLE-B8B1
        0xB8B2, # G+11867  U+B8B2  uniB8B2,  HANGUL SYLLABLE-B8B2
        0xB8B3, # G+11868  U+B8B3  uniB8B3,  HANGUL SYLLABLE-B8B3
        0xB8B5, # G+11869  U+B8B5  uniB8B5,  HANGUL SYLLABLE-B8B5
        0xB8B6, # G+11870  U+B8B6  uniB8B6,  HANGUL SYLLABLE-B8B6
        0xB8B7, # G+11871  U+B8B7  uniB8B7,  HANGUL SYLLABLE-B8B7
        0xB8B9, # G+11872  U+B8B9  uniB8B9,  HANGUL SYLLABLE-B8B9
        0xB8BA, # G+11873  U+B8BA  uniB8BA,  HANGUL SYLLABLE-B8BA
        0xB8BB, # G+11874  U+B8BB  uniB8BB,  HANGUL SYLLABLE-B8BB
        0xB8BC, # G+11875  U+B8BC  uniB8BC,  HANGUL SYLLABLE-B8BC
        0xB8BD, # G+11876  U+B8BD  uniB8BD,  HANGUL SYLLABLE-B8BD
        0xB8BE, # G+11877  U+B8BE  uniB8BE,  HANGUL SYLLABLE-B8BE
        0xB8BF, # G+11878  U+B8BF  uniB8BF,  HANGUL SYLLABLE-B8BF
        0xB8C2, # G+11879  U+B8C2  uniB8C2,  HANGUL SYLLABLE-B8C2
        0xB8C4, # G+11880  U+B8C4  uniB8C4,  HANGUL SYLLABLE-B8C4
        0xB8C6, # G+11881  U+B8C6  uniB8C6,  HANGUL SYLLABLE-B8C6
        0xB8C7, # G+11882  U+B8C7  uniB8C7,  HANGUL SYLLABLE-B8C7
        0xB8C8, # G+11883  U+B8C8  uniB8C8,  HANGUL SYLLABLE-B8C8
        0xB8C9, # G+11884  U+B8C9  uniB8C9,  HANGUL SYLLABLE-B8C9
        0xB8CA, # G+11885  U+B8CA  uniB8CA,  HANGUL SYLLABLE-B8CA
        0xB8CB, # G+11886  U+B8CB  uniB8CB,  HANGUL SYLLABLE-B8CB
        0xB8CD, # G+11887  U+B8CD  uniB8CD,  HANGUL SYLLABLE-B8CD
        0xB8CE, # G+11888  U+B8CE  uniB8CE,  HANGUL SYLLABLE-B8CE
        0xB8CF, # G+11889  U+B8CF  uniB8CF,  HANGUL SYLLABLE-B8CF
        0xB8D1, # G+11890  U+B8D1  uniB8D1,  HANGUL SYLLABLE-B8D1
        0xB8D2, # G+11891  U+B8D2  uniB8D2,  HANGUL SYLLABLE-B8D2
        0xB8D3, # G+11892  U+B8D3  uniB8D3,  HANGUL SYLLABLE-B8D3
        0xB8D5, # G+11893  U+B8D5  uniB8D5,  HANGUL SYLLABLE-B8D5
        0xB8D6, # G+11894  U+B8D6  uniB8D6,  HANGUL SYLLABLE-B8D6
        0xB8D7, # G+11895  U+B8D7  uniB8D7,  HANGUL SYLLABLE-B8D7
        0xB8D8, # G+11896  U+B8D8  uniB8D8,  HANGUL SYLLABLE-B8D8
        0xB8D9, # G+11897  U+B8D9  uniB8D9,  HANGUL SYLLABLE-B8D9
        0xB8DA, # G+11898  U+B8DA  uniB8DA,  HANGUL SYLLABLE-B8DA
        0xB8DB, # G+11899  U+B8DB  uniB8DB,  HANGUL SYLLABLE-B8DB
        0xB8DC, # G+11900  U+B8DC  uniB8DC,  HANGUL SYLLABLE-B8DC
        0xB8DE, # G+11901  U+B8DE  uniB8DE,  HANGUL SYLLABLE-B8DE
        0xB8E0, # G+11902  U+B8E0  uniB8E0,  HANGUL SYLLABLE-B8E0
        0xB8E2, # G+11903  U+B8E2  uniB8E2,  HANGUL SYLLABLE-B8E2
        0xB8E3, # G+11904  U+B8E3  uniB8E3,  HANGUL SYLLABLE-B8E3
        0xB8E4, # G+11905  U+B8E4  uniB8E4,  HANGUL SYLLABLE-B8E4
        0xB8E5, # G+11906  U+B8E5  uniB8E5,  HANGUL SYLLABLE-B8E5
        0xB8E6, # G+11907  U+B8E6  uniB8E6,  HANGUL SYLLABLE-B8E6
        0xB8E7, # G+11908  U+B8E7  uniB8E7,  HANGUL SYLLABLE-B8E7
        0xB8EA, # G+11909  U+B8EA  uniB8EA,  HANGUL SYLLABLE-B8EA
        0xB8EB, # G+11910  U+B8EB  uniB8EB,  HANGUL SYLLABLE-B8EB
        0xB8ED, # G+11911  U+B8ED  uniB8ED,  HANGUL SYLLABLE-B8ED
        0xB8EE, # G+11912  U+B8EE  uniB8EE,  HANGUL SYLLABLE-B8EE
        0xB8EF, # G+11913  U+B8EF  uniB8EF,  HANGUL SYLLABLE-B8EF
        0xB8F1, # G+11914  U+B8F1  uniB8F1,  HANGUL SYLLABLE-B8F1
        0xB8F2, # G+11915  U+B8F2  uniB8F2,  HANGUL SYLLABLE-B8F2
        0xB8F3, # G+11916  U+B8F3  uniB8F3,  HANGUL SYLLABLE-B8F3
        0xB8F4, # G+11917  U+B8F4  uniB8F4,  HANGUL SYLLABLE-B8F4
        0xB8F5, # G+11918  U+B8F5  uniB8F5,  HANGUL SYLLABLE-B8F5
        0xB8F6, # G+11919  U+B8F6  uniB8F6,  HANGUL SYLLABLE-B8F6
        0xB8F7, # G+11920  U+B8F7  uniB8F7,  HANGUL SYLLABLE-B8F7
        0xB8FA, # G+11921  U+B8FA  uniB8FA,  HANGUL SYLLABLE-B8FA
        0xB8FC, # G+11922  U+B8FC  uniB8FC,  HANGUL SYLLABLE-B8FC
        0xB8FE, # G+11923  U+B8FE  uniB8FE,  HANGUL SYLLABLE-B8FE
        0xB8FF, # G+11924  U+B8FF  uniB8FF,  HANGUL SYLLABLE-B8FF
        0xB900, # G+11925  U+B900  uniB900,  HANGUL SYLLABLE-B900
        0xB901, # G+11926  U+B901  uniB901,  HANGUL SYLLABLE-B901
        0xB902, # G+11927  U+B902  uniB902,  HANGUL SYLLABLE-B902
        0xB903, # G+11928  U+B903  uniB903,  HANGUL SYLLABLE-B903
        0xB905, # G+11929  U+B905  uniB905,  HANGUL SYLLABLE-B905
        0xB906, # G+11930  U+B906  uniB906,  HANGUL SYLLABLE-B906
        0xB907, # G+11931  U+B907  uniB907,  HANGUL SYLLABLE-B907
        0xB908, # G+11932  U+B908  uniB908,  HANGUL SYLLABLE-B908
        0xB909, # G+11933  U+B909  uniB909,  HANGUL SYLLABLE-B909
        0xB90A, # G+11934  U+B90A  uniB90A,  HANGUL SYLLABLE-B90A
        0xB90B, # G+11935  U+B90B  uniB90B,  HANGUL SYLLABLE-B90B
        0xB90C, # G+11936  U+B90C  uniB90C,  HANGUL SYLLABLE-B90C
        0xB90D, # G+11937  U+B90D  uniB90D,  HANGUL SYLLABLE-B90D
        0xB90E, # G+11938  U+B90E  uniB90E,  HANGUL SYLLABLE-B90E
        0xB90F, # G+11939  U+B90F  uniB90F,  HANGUL SYLLABLE-B90F
        0xB910, # G+11940  U+B910  uniB910,  HANGUL SYLLABLE-B910
        0xB911, # G+11941  U+B911  uniB911,  HANGUL SYLLABLE-B911
        0xB912, # G+11942  U+B912  uniB912,  HANGUL SYLLABLE-B912
        0xB913, # G+11943  U+B913  uniB913,  HANGUL SYLLABLE-B913
        0xB914, # G+11944  U+B914  uniB914,  HANGUL SYLLABLE-B914
        0xB915, # G+11945  U+B915  uniB915,  HANGUL SYLLABLE-B915
        0xB916, # G+11946  U+B916  uniB916,  HANGUL SYLLABLE-B916
        0xB917, # G+11947  U+B917  uniB917,  HANGUL SYLLABLE-B917
        0xB919, # G+11948  U+B919  uniB919,  HANGUL SYLLABLE-B919
        0xB91A, # G+11949  U+B91A  uniB91A,  HANGUL SYLLABLE-B91A
        0xB91B, # G+11950  U+B91B  uniB91B,  HANGUL SYLLABLE-B91B
        0xB91C, # G+11951  U+B91C  uniB91C,  HANGUL SYLLABLE-B91C
        0xB91D, # G+11952  U+B91D  uniB91D,  HANGUL SYLLABLE-B91D
        0xB91E, # G+11953  U+B91E  uniB91E,  HANGUL SYLLABLE-B91E
        0xB91F, # G+11954  U+B91F  uniB91F,  HANGUL SYLLABLE-B91F
        0xB921, # G+11955  U+B921  uniB921,  HANGUL SYLLABLE-B921
        0xB922, # G+11956  U+B922  uniB922,  HANGUL SYLLABLE-B922
        0xB923, # G+11957  U+B923  uniB923,  HANGUL SYLLABLE-B923
        0xB924, # G+11958  U+B924  uniB924,  HANGUL SYLLABLE-B924
        0xB925, # G+11959  U+B925  uniB925,  HANGUL SYLLABLE-B925
        0xB926, # G+11960  U+B926  uniB926,  HANGUL SYLLABLE-B926
        0xB927, # G+11961  U+B927  uniB927,  HANGUL SYLLABLE-B927
        0xB928, # G+11962  U+B928  uniB928,  HANGUL SYLLABLE-B928
        0xB929, # G+11963  U+B929  uniB929,  HANGUL SYLLABLE-B929
        0xB92A, # G+11964  U+B92A  uniB92A,  HANGUL SYLLABLE-B92A
        0xB92B, # G+11965  U+B92B  uniB92B,  HANGUL SYLLABLE-B92B
        0xB92C, # G+11966  U+B92C  uniB92C,  HANGUL SYLLABLE-B92C
        0xB92D, # G+11967  U+B92D  uniB92D,  HANGUL SYLLABLE-B92D
        0xB92E, # G+11968  U+B92E  uniB92E,  HANGUL SYLLABLE-B92E
        0xB92F, # G+11969  U+B92F  uniB92F,  HANGUL SYLLABLE-B92F
        0xB930, # G+11970  U+B930  uniB930,  HANGUL SYLLABLE-B930
        0xB931, # G+11971  U+B931  uniB931,  HANGUL SYLLABLE-B931
        0xB932, # G+11972  U+B932  uniB932,  HANGUL SYLLABLE-B932
        0xB933, # G+11973  U+B933  uniB933,  HANGUL SYLLABLE-B933
        0xB934, # G+11974  U+B934  uniB934,  HANGUL SYLLABLE-B934
        0xB935, # G+11975  U+B935  uniB935,  HANGUL SYLLABLE-B935
        0xB936, # G+11976  U+B936  uniB936,  HANGUL SYLLABLE-B936
        0xB937, # G+11977  U+B937  uniB937,  HANGUL SYLLABLE-B937
        0xB938, # G+11978  U+B938  uniB938,  HANGUL SYLLABLE-B938
        0xB939, # G+11979  U+B939  uniB939,  HANGUL SYLLABLE-B939
        0xB93A, # G+11980  U+B93A  uniB93A,  HANGUL SYLLABLE-B93A
        0xB93B, # G+11981  U+B93B  uniB93B,  HANGUL SYLLABLE-B93B
        0xB93E, # G+11982  U+B93E  uniB93E,  HANGUL SYLLABLE-B93E
        0xB93F, # G+11983  U+B93F  uniB93F,  HANGUL SYLLABLE-B93F
        0xB941, # G+11984  U+B941  uniB941,  HANGUL SYLLABLE-B941
        0xB942, # G+11985  U+B942  uniB942,  HANGUL SYLLABLE-B942
        0xB943, # G+11986  U+B943  uniB943,  HANGUL SYLLABLE-B943
        0xB945, # G+11987  U+B945  uniB945,  HANGUL SYLLABLE-B945
        0xB946, # G+11988  U+B946  uniB946,  HANGUL SYLLABLE-B946
        0xB947, # G+11989  U+B947  uniB947,  HANGUL SYLLABLE-B947
        0xB948, # G+11990  U+B948  uniB948,  HANGUL SYLLABLE-B948
        0xB949, # G+11991  U+B949  uniB949,  HANGUL SYLLABLE-B949
        0xB94A, # G+11992  U+B94A  uniB94A,  HANGUL SYLLABLE-B94A
        0xB94B, # G+11993  U+B94B  uniB94B,  HANGUL SYLLABLE-B94B
        0xB94D, # G+11994  U+B94D  uniB94D,  HANGUL SYLLABLE-B94D
        0xB94E, # G+11995  U+B94E  uniB94E,  HANGUL SYLLABLE-B94E
        0xB950, # G+11996  U+B950  uniB950,  HANGUL SYLLABLE-B950
        0xB952, # G+11997  U+B952  uniB952,  HANGUL SYLLABLE-B952
        0xB953, # G+11998  U+B953  uniB953,  HANGUL SYLLABLE-B953
        0xB954, # G+11999  U+B954  uniB954,  HANGUL SYLLABLE-B954
        0xB955, # G+12000  U+B955  uniB955,  HANGUL SYLLABLE-B955
        0xB956, # G+12001  U+B956  uniB956,  HANGUL SYLLABLE-B956
        0xB957, # G+12002  U+B957  uniB957,  HANGUL SYLLABLE-B957
        0xB95A, # G+12003  U+B95A  uniB95A,  HANGUL SYLLABLE-B95A
        0xB95B, # G+12004  U+B95B  uniB95B,  HANGUL SYLLABLE-B95B
        0xB95D, # G+12005  U+B95D  uniB95D,  HANGUL SYLLABLE-B95D
        0xB95E, # G+12006  U+B95E  uniB95E,  HANGUL SYLLABLE-B95E
        0xB95F, # G+12007  U+B95F  uniB95F,  HANGUL SYLLABLE-B95F
        0xB961, # G+12008  U+B961  uniB961,  HANGUL SYLLABLE-B961
        0xB962, # G+12009  U+B962  uniB962,  HANGUL SYLLABLE-B962
        0xB963, # G+12010  U+B963  uniB963,  HANGUL SYLLABLE-B963
        0xB964, # G+12011  U+B964  uniB964,  HANGUL SYLLABLE-B964
        0xB965, # G+12012  U+B965  uniB965,  HANGUL SYLLABLE-B965
        0xB966, # G+12013  U+B966  uniB966,  HANGUL SYLLABLE-B966
        0xB967, # G+12014  U+B967  uniB967,  HANGUL SYLLABLE-B967
        0xB96A, # G+12015  U+B96A  uniB96A,  HANGUL SYLLABLE-B96A
        0xB96C, # G+12016  U+B96C  uniB96C,  HANGUL SYLLABLE-B96C
        0xB96E, # G+12017  U+B96E  uniB96E,  HANGUL SYLLABLE-B96E
        0xB96F, # G+12018  U+B96F  uniB96F,  HANGUL SYLLABLE-B96F
        0xB970, # G+12019  U+B970  uniB970,  HANGUL SYLLABLE-B970
        0xB971, # G+12020  U+B971  uniB971,  HANGUL SYLLABLE-B971
        0xB972, # G+12021  U+B972  uniB972,  HANGUL SYLLABLE-B972
        0xB973, # G+12022  U+B973  uniB973,  HANGUL SYLLABLE-B973
        0xB976, # G+12023  U+B976  uniB976,  HANGUL SYLLABLE-B976
        0xB977, # G+12024  U+B977  uniB977,  HANGUL SYLLABLE-B977
        0xB979, # G+12025  U+B979  uniB979,  HANGUL SYLLABLE-B979
        0xB97A, # G+12026  U+B97A  uniB97A,  HANGUL SYLLABLE-B97A
        0xB97B, # G+12027  U+B97B  uniB97B,  HANGUL SYLLABLE-B97B
        0xB97D, # G+12028  U+B97D  uniB97D,  HANGUL SYLLABLE-B97D
        0xB97E, # G+12029  U+B97E  uniB97E,  HANGUL SYLLABLE-B97E
        0xB97F, # G+12030  U+B97F  uniB97F,  HANGUL SYLLABLE-B97F
        0xB980, # G+12031  U+B980  uniB980,  HANGUL SYLLABLE-B980
        0xB981, # G+12032  U+B981  uniB981,  HANGUL SYLLABLE-B981
        0xB982, # G+12033  U+B982  uniB982,  HANGUL SYLLABLE-B982
        0xB983, # G+12034  U+B983  uniB983,  HANGUL SYLLABLE-B983
        0xB986, # G+12035  U+B986  uniB986,  HANGUL SYLLABLE-B986
        0xB988, # G+12036  U+B988  uniB988,  HANGUL SYLLABLE-B988
        0xB98B, # G+12037  U+B98B  uniB98B,  HANGUL SYLLABLE-B98B
        0xB98C, # G+12038  U+B98C  uniB98C,  HANGUL SYLLABLE-B98C
        0xB98F, # G+12039  U+B98F  uniB98F,  HANGUL SYLLABLE-B98F
        0xB990, # G+12040  U+B990  uniB990,  HANGUL SYLLABLE-B990
        0xB991, # G+12041  U+B991  uniB991,  HANGUL SYLLABLE-B991
        0xB992, # G+12042  U+B992  uniB992,  HANGUL SYLLABLE-B992
        0xB993, # G+12043  U+B993  uniB993,  HANGUL SYLLABLE-B993
        0xB994, # G+12044  U+B994  uniB994,  HANGUL SYLLABLE-B994
        0xB995, # G+12045  U+B995  uniB995,  HANGUL SYLLABLE-B995
        0xB996, # G+12046  U+B996  uniB996,  HANGUL SYLLABLE-B996
        0xB997, # G+12047  U+B997  uniB997,  HANGUL SYLLABLE-B997
        0xB998, # G+12048  U+B998  uniB998,  HANGUL SYLLABLE-B998
        0xB999, # G+12049  U+B999  uniB999,  HANGUL SYLLABLE-B999
        0xB99A, # G+12050  U+B99A  uniB99A,  HANGUL SYLLABLE-B99A
        0xB99B, # G+12051  U+B99B  uniB99B,  HANGUL SYLLABLE-B99B
        0xB99C, # G+12052  U+B99C  uniB99C,  HANGUL SYLLABLE-B99C
        0xB99D, # G+12053  U+B99D  uniB99D,  HANGUL SYLLABLE-B99D
        0xB99E, # G+12054  U+B99E  uniB99E,  HANGUL SYLLABLE-B99E
        0xB99F, # G+12055  U+B99F  uniB99F,  HANGUL SYLLABLE-B99F
        0xB9A0, # G+12056  U+B9A0  uniB9A0,  HANGUL SYLLABLE-B9A0
        0xB9A1, # G+12057  U+B9A1  uniB9A1,  HANGUL SYLLABLE-B9A1
        0xB9A2, # G+12058  U+B9A2  uniB9A2,  HANGUL SYLLABLE-B9A2
        0xB9A3, # G+12059  U+B9A3  uniB9A3,  HANGUL SYLLABLE-B9A3
        0xB9A4, # G+12060  U+B9A4  uniB9A4,  HANGUL SYLLABLE-B9A4
        0xB9A5, # G+12061  U+B9A5  uniB9A5,  HANGUL SYLLABLE-B9A5
        0xB9A6, # G+12062  U+B9A6  uniB9A6,  HANGUL SYLLABLE-B9A6
        0xB9A7, # G+12063  U+B9A7  uniB9A7,  HANGUL SYLLABLE-B9A7
        0xB9A8, # G+12064  U+B9A8  uniB9A8,  HANGUL SYLLABLE-B9A8
        0xB9A9, # G+12065  U+B9A9  uniB9A9,  HANGUL SYLLABLE-B9A9
        0xB9AA, # G+12066  U+B9AA  uniB9AA,  HANGUL SYLLABLE-B9AA
        0xB9AB, # G+12067  U+B9AB  uniB9AB,  HANGUL SYLLABLE-B9AB
        0xB9AE, # G+12068  U+B9AE  uniB9AE,  HANGUL SYLLABLE-B9AE
        0xB9AF, # G+12069  U+B9AF  uniB9AF,  HANGUL SYLLABLE-B9AF
        0xB9B1, # G+12070  U+B9B1  uniB9B1,  HANGUL SYLLABLE-B9B1
        0xB9B2, # G+12071  U+B9B2  uniB9B2,  HANGUL SYLLABLE-B9B2
        0xB9B3, # G+12072  U+B9B3  uniB9B3,  HANGUL SYLLABLE-B9B3
        0xB9B5, # G+12073  U+B9B5  uniB9B5,  HANGUL SYLLABLE-B9B5
        0xB9B6, # G+12074  U+B9B6  uniB9B6,  HANGUL SYLLABLE-B9B6
        0xB9B7, # G+12075  U+B9B7  uniB9B7,  HANGUL SYLLABLE-B9B7
        0xB9B8, # G+12076  U+B9B8  uniB9B8,  HANGUL SYLLABLE-B9B8
        0xB9B9, # G+12077  U+B9B9  uniB9B9,  HANGUL SYLLABLE-B9B9
        0xB9BA, # G+12078  U+B9BA  uniB9BA,  HANGUL SYLLABLE-B9BA
        0xB9BB, # G+12079  U+B9BB  uniB9BB,  HANGUL SYLLABLE-B9BB
        0xB9BE, # G+12080  U+B9BE  uniB9BE,  HANGUL SYLLABLE-B9BE
        0xB9C0, # G+12081  U+B9C0  uniB9C0,  HANGUL SYLLABLE-B9C0
        0xB9C2, # G+12082  U+B9C2  uniB9C2,  HANGUL SYLLABLE-B9C2
        0xB9C3, # G+12083  U+B9C3  uniB9C3,  HANGUL SYLLABLE-B9C3
        0xB9C4, # G+12084  U+B9C4  uniB9C4,  HANGUL SYLLABLE-B9C4
        0xB9C5, # G+12085  U+B9C5  uniB9C5,  HANGUL SYLLABLE-B9C5
        0xB9C6, # G+12086  U+B9C6  uniB9C6,  HANGUL SYLLABLE-B9C6
        0xB9C7, # G+12087  U+B9C7  uniB9C7,  HANGUL SYLLABLE-B9C7
        0xB9CA, # G+12088  U+B9CA  uniB9CA,  HANGUL SYLLABLE-B9CA
        0xB9CB, # G+12089  U+B9CB  uniB9CB,  HANGUL SYLLABLE-B9CB
        0xB9CD, # G+12090  U+B9CD  uniB9CD,  HANGUL SYLLABLE-B9CD
        0xB9D3, # G+12091  U+B9D3  uniB9D3,  HANGUL SYLLABLE-B9D3
        0xB9D4, # G+12092  U+B9D4  uniB9D4,  HANGUL SYLLABLE-B9D4
        0xB9D5, # G+12093  U+B9D5  uniB9D5,  HANGUL SYLLABLE-B9D5
        0xB9D6, # G+12094  U+B9D6  uniB9D6,  HANGUL SYLLABLE-B9D6
        0xB9D7, # G+12095  U+B9D7  uniB9D7,  HANGUL SYLLABLE-B9D7
        0xB9DA, # G+12096  U+B9DA  uniB9DA,  HANGUL SYLLABLE-B9DA
        0xB9DC, # G+12097  U+B9DC  uniB9DC,  HANGUL SYLLABLE-B9DC
        0xB9DF, # G+12098  U+B9DF  uniB9DF,  HANGUL SYLLABLE-B9DF
        0xB9E0, # G+12099  U+B9E0  uniB9E0,  HANGUL SYLLABLE-B9E0
        0xB9E2, # G+12100  U+B9E2  uniB9E2,  HANGUL SYLLABLE-B9E2
        0xB9E6, # G+12101  U+B9E6  uniB9E6,  HANGUL SYLLABLE-B9E6
        0xB9E7, # G+12102  U+B9E7  uniB9E7,  HANGUL SYLLABLE-B9E7
        0xB9E9, # G+12103  U+B9E9  uniB9E9,  HANGUL SYLLABLE-B9E9
        0xB9EA, # G+12104  U+B9EA  uniB9EA,  HANGUL SYLLABLE-B9EA
        0xB9EB, # G+12105  U+B9EB  uniB9EB,  HANGUL SYLLABLE-B9EB
        0xB9ED, # G+12106  U+B9ED  uniB9ED,  HANGUL SYLLABLE-B9ED
        0xB9EE, # G+12107  U+B9EE  uniB9EE,  HANGUL SYLLABLE-B9EE
        0xB9EF, # G+12108  U+B9EF  uniB9EF,  HANGUL SYLLABLE-B9EF
        0xB9F0, # G+12109  U+B9F0  uniB9F0,  HANGUL SYLLABLE-B9F0
        0xB9F1, # G+12110  U+B9F1  uniB9F1,  HANGUL SYLLABLE-B9F1
        0xB9F2, # G+12111  U+B9F2  uniB9F2,  HANGUL SYLLABLE-B9F2
        0xB9F3, # G+12112  U+B9F3  uniB9F3,  HANGUL SYLLABLE-B9F3
        0xB9F6, # G+12113  U+B9F6  uniB9F6,  HANGUL SYLLABLE-B9F6
        0xB9FB, # G+12114  U+B9FB  uniB9FB,  HANGUL SYLLABLE-B9FB
        0xB9FC, # G+12115  U+B9FC  uniB9FC,  HANGUL SYLLABLE-B9FC
        0xB9FD, # G+12116  U+B9FD  uniB9FD,  HANGUL SYLLABLE-B9FD
        0xB9FE, # G+12117  U+B9FE  uniB9FE,  HANGUL SYLLABLE-B9FE
        0xB9FF, # G+12118  U+B9FF  uniB9FF,  HANGUL SYLLABLE-B9FF
        0xBA02, # G+12119  U+BA02  uniBA02,  HANGUL SYLLABLE-BA02
        0xBA03, # G+12120  U+BA03  uniBA03,  HANGUL SYLLABLE-BA03
        0xBA04, # G+12121  U+BA04  uniBA04,  HANGUL SYLLABLE-BA04
        0xBA05, # G+12122  U+BA05  uniBA05,  HANGUL SYLLABLE-BA05
        0xBA06, # G+12123  U+BA06  uniBA06,  HANGUL SYLLABLE-BA06
        0xBA07, # G+12124  U+BA07  uniBA07,  HANGUL SYLLABLE-BA07
        0xBA09, # G+12125  U+BA09  uniBA09,  HANGUL SYLLABLE-BA09
        0xBA0A, # G+12126  U+BA0A  uniBA0A,  HANGUL SYLLABLE-BA0A
        0xBA0B, # G+12127  U+BA0B  uniBA0B,  HANGUL SYLLABLE-BA0B
        0xBA0C, # G+12128  U+BA0C  uniBA0C,  HANGUL SYLLABLE-BA0C
        0xBA0D, # G+12129  U+BA0D  uniBA0D,  HANGUL SYLLABLE-BA0D
        0xBA0E, # G+12130  U+BA0E  uniBA0E,  HANGUL SYLLABLE-BA0E
        0xBA0F, # G+12131  U+BA0F  uniBA0F,  HANGUL SYLLABLE-BA0F
        0xBA10, # G+12132  U+BA10  uniBA10,  HANGUL SYLLABLE-BA10
        0xBA11, # G+12133  U+BA11  uniBA11,  HANGUL SYLLABLE-BA11
        0xBA12, # G+12134  U+BA12  uniBA12,  HANGUL SYLLABLE-BA12
        0xBA13, # G+12135  U+BA13  uniBA13,  HANGUL SYLLABLE-BA13
        0xBA14, # G+12136  U+BA14  uniBA14,  HANGUL SYLLABLE-BA14
        0xBA16, # G+12137  U+BA16  uniBA16,  HANGUL SYLLABLE-BA16
        0xBA17, # G+12138  U+BA17  uniBA17,  HANGUL SYLLABLE-BA17
        0xBA18, # G+12139  U+BA18  uniBA18,  HANGUL SYLLABLE-BA18
        0xBA19, # G+12140  U+BA19  uniBA19,  HANGUL SYLLABLE-BA19
        0xBA1A, # G+12141  U+BA1A  uniBA1A,  HANGUL SYLLABLE-BA1A
        0xBA1B, # G+12142  U+BA1B  uniBA1B,  HANGUL SYLLABLE-BA1B
        0xBA1C, # G+12143  U+BA1C  uniBA1C,  HANGUL SYLLABLE-BA1C
        0xBA1D, # G+12144  U+BA1D  uniBA1D,  HANGUL SYLLABLE-BA1D
        0xBA1E, # G+12145  U+BA1E  uniBA1E,  HANGUL SYLLABLE-BA1E
        0xBA1F, # G+12146  U+BA1F  uniBA1F,  HANGUL SYLLABLE-BA1F
        0xBA20, # G+12147  U+BA20  uniBA20,  HANGUL SYLLABLE-BA20
        0xBA21, # G+12148  U+BA21  uniBA21,  HANGUL SYLLABLE-BA21
        0xBA22, # G+12149  U+BA22  uniBA22,  HANGUL SYLLABLE-BA22
        0xBA23, # G+12150  U+BA23  uniBA23,  HANGUL SYLLABLE-BA23
        0xBA24, # G+12151  U+BA24  uniBA24,  HANGUL SYLLABLE-BA24
        0xBA25, # G+12152  U+BA25  uniBA25,  HANGUL SYLLABLE-BA25
        0xBA26, # G+12153  U+BA26  uniBA26,  HANGUL SYLLABLE-BA26
        0xBA27, # G+12154  U+BA27  uniBA27,  HANGUL SYLLABLE-BA27
        0xBA28, # G+12155  U+BA28  uniBA28,  HANGUL SYLLABLE-BA28
        0xBA29, # G+12156  U+BA29  uniBA29,  HANGUL SYLLABLE-BA29
        0xBA2A, # G+12157  U+BA2A  uniBA2A,  HANGUL SYLLABLE-BA2A
        0xBA2B, # G+12158  U+BA2B  uniBA2B,  HANGUL SYLLABLE-BA2B
        0xBA2C, # G+12159  U+BA2C  uniBA2C,  HANGUL SYLLABLE-BA2C
        0xBA2D, # G+12160  U+BA2D  uniBA2D,  HANGUL SYLLABLE-BA2D
        0xBA2E, # G+12161  U+BA2E  uniBA2E,  HANGUL SYLLABLE-BA2E
        0xBA2F, # G+12162  U+BA2F  uniBA2F,  HANGUL SYLLABLE-BA2F
        0xBA30, # G+12163  U+BA30  uniBA30,  HANGUL SYLLABLE-BA30
        0xBA31, # G+12164  U+BA31  uniBA31,  HANGUL SYLLABLE-BA31
        0xBA32, # G+12165  U+BA32  uniBA32,  HANGUL SYLLABLE-BA32
        0xBA33, # G+12166  U+BA33  uniBA33,  HANGUL SYLLABLE-BA33
        0xBA34, # G+12167  U+BA34  uniBA34,  HANGUL SYLLABLE-BA34
        0xBA35, # G+12168  U+BA35  uniBA35,  HANGUL SYLLABLE-BA35
        0xBA36, # G+12169  U+BA36  uniBA36,  HANGUL SYLLABLE-BA36
        0xBA37, # G+12170  U+BA37  uniBA37,  HANGUL SYLLABLE-BA37
        0xBA3A, # G+12171  U+BA3A  uniBA3A,  HANGUL SYLLABLE-BA3A
        0xBA3B, # G+12172  U+BA3B  uniBA3B,  HANGUL SYLLABLE-BA3B
        0xBA3D, # G+12173  U+BA3D  uniBA3D,  HANGUL SYLLABLE-BA3D
        0xBA3E, # G+12174  U+BA3E  uniBA3E,  HANGUL SYLLABLE-BA3E
        0xBA3F, # G+12175  U+BA3F  uniBA3F,  HANGUL SYLLABLE-BA3F
        0xBA41, # G+12176  U+BA41  uniBA41,  HANGUL SYLLABLE-BA41
        0xBA43, # G+12177  U+BA43  uniBA43,  HANGUL SYLLABLE-BA43
        0xBA44, # G+12178  U+BA44  uniBA44,  HANGUL SYLLABLE-BA44
        0xBA45, # G+12179  U+BA45  uniBA45,  HANGUL SYLLABLE-BA45
        0xBA46, # G+12180  U+BA46  uniBA46,  HANGUL SYLLABLE-BA46
        0xBA47, # G+12181  U+BA47  uniBA47,  HANGUL SYLLABLE-BA47
        0xBA4A, # G+12182  U+BA4A  uniBA4A,  HANGUL SYLLABLE-BA4A
        0xBA4C, # G+12183  U+BA4C  uniBA4C,  HANGUL SYLLABLE-BA4C
        0xBA4F, # G+12184  U+BA4F  uniBA4F,  HANGUL SYLLABLE-BA4F
        0xBA50, # G+12185  U+BA50  uniBA50,  HANGUL SYLLABLE-BA50
        0xBA51, # G+12186  U+BA51  uniBA51,  HANGUL SYLLABLE-BA51
        0xBA52, # G+12187  U+BA52  uniBA52,  HANGUL SYLLABLE-BA52
        0xBA56, # G+12188  U+BA56  uniBA56,  HANGUL SYLLABLE-BA56
        0xBA57, # G+12189  U+BA57  uniBA57,  HANGUL SYLLABLE-BA57
        0xBA59, # G+12190  U+BA59  uniBA59,  HANGUL SYLLABLE-BA59
        0xBA5A, # G+12191  U+BA5A  uniBA5A,  HANGUL SYLLABLE-BA5A
        0xBA5B, # G+12192  U+BA5B  uniBA5B,  HANGUL SYLLABLE-BA5B
        0xBA5D, # G+12193  U+BA5D  uniBA5D,  HANGUL SYLLABLE-BA5D
        0xBA5E, # G+12194  U+BA5E  uniBA5E,  HANGUL SYLLABLE-BA5E
        0xBA5F, # G+12195  U+BA5F  uniBA5F,  HANGUL SYLLABLE-BA5F
        0xBA60, # G+12196  U+BA60  uniBA60,  HANGUL SYLLABLE-BA60
        0xBA61, # G+12197  U+BA61  uniBA61,  HANGUL SYLLABLE-BA61
        0xBA62, # G+12198  U+BA62  uniBA62,  HANGUL SYLLABLE-BA62
        0xBA63, # G+12199  U+BA63  uniBA63,  HANGUL SYLLABLE-BA63
        0xBA66, # G+12200  U+BA66  uniBA66,  HANGUL SYLLABLE-BA66
        0xBA6A, # G+12201  U+BA6A  uniBA6A,  HANGUL SYLLABLE-BA6A
        0xBA6B, # G+12202  U+BA6B  uniBA6B,  HANGUL SYLLABLE-BA6B
        0xBA6C, # G+12203  U+BA6C  uniBA6C,  HANGUL SYLLABLE-BA6C
        0xBA6D, # G+12204  U+BA6D  uniBA6D,  HANGUL SYLLABLE-BA6D
        0xBA6E, # G+12205  U+BA6E  uniBA6E,  HANGUL SYLLABLE-BA6E
        0xBA6F, # G+12206  U+BA6F  uniBA6F,  HANGUL SYLLABLE-BA6F
        0xBA72, # G+12207  U+BA72  uniBA72,  HANGUL SYLLABLE-BA72
        0xBA73, # G+12208  U+BA73  uniBA73,  HANGUL SYLLABLE-BA73
        0xBA75, # G+12209  U+BA75  uniBA75,  HANGUL SYLLABLE-BA75
        0xBA76, # G+12210  U+BA76  uniBA76,  HANGUL SYLLABLE-BA76
        0xBA77, # G+12211  U+BA77  uniBA77,  HANGUL SYLLABLE-BA77
        0xBA79, # G+12212  U+BA79  uniBA79,  HANGUL SYLLABLE-BA79
        0xBA7A, # G+12213  U+BA7A  uniBA7A,  HANGUL SYLLABLE-BA7A
        0xBA7B, # G+12214  U+BA7B  uniBA7B,  HANGUL SYLLABLE-BA7B
        0xBA7C, # G+12215  U+BA7C  uniBA7C,  HANGUL SYLLABLE-BA7C
        0xBA7D, # G+12216  U+BA7D  uniBA7D,  HANGUL SYLLABLE-BA7D
        0xBA7E, # G+12217  U+BA7E  uniBA7E,  HANGUL SYLLABLE-BA7E
        0xBA7F, # G+12218  U+BA7F  uniBA7F,  HANGUL SYLLABLE-BA7F
        0xBA80, # G+12219  U+BA80  uniBA80,  HANGUL SYLLABLE-BA80
        0xBA81, # G+12220  U+BA81  uniBA81,  HANGUL SYLLABLE-BA81
        0xBA82, # G+12221  U+BA82  uniBA82,  HANGUL SYLLABLE-BA82
        0xBA86, # G+12222  U+BA86  uniBA86,  HANGUL SYLLABLE-BA86
        0xBA88, # G+12223  U+BA88  uniBA88,  HANGUL SYLLABLE-BA88
        0xBA89, # G+12224  U+BA89  uniBA89,  HANGUL SYLLABLE-BA89
        0xBA8A, # G+12225  U+BA8A  uniBA8A,  HANGUL SYLLABLE-BA8A
        0xBA8B, # G+12226  U+BA8B  uniBA8B,  HANGUL SYLLABLE-BA8B
        0xBA8D, # G+12227  U+BA8D  uniBA8D,  HANGUL SYLLABLE-BA8D
        0xBA8E, # G+12228  U+BA8E  uniBA8E,  HANGUL SYLLABLE-BA8E
        0xBA8F, # G+12229  U+BA8F  uniBA8F,  HANGUL SYLLABLE-BA8F
        0xBA90, # G+12230  U+BA90  uniBA90,  HANGUL SYLLABLE-BA90
        0xBA91, # G+12231  U+BA91  uniBA91,  HANGUL SYLLABLE-BA91
        0xBA92, # G+12232  U+BA92  uniBA92,  HANGUL SYLLABLE-BA92
        0xBA93, # G+12233  U+BA93  uniBA93,  HANGUL SYLLABLE-BA93
        0xBA94, # G+12234  U+BA94  uniBA94,  HANGUL SYLLABLE-BA94
        0xBA95, # G+12235  U+BA95  uniBA95,  HANGUL SYLLABLE-BA95
        0xBA96, # G+12236  U+BA96  uniBA96,  HANGUL SYLLABLE-BA96
        0xBA97, # G+12237  U+BA97  uniBA97,  HANGUL SYLLABLE-BA97
        0xBA98, # G+12238  U+BA98  uniBA98,  HANGUL SYLLABLE-BA98
        0xBA99, # G+12239  U+BA99  uniBA99,  HANGUL SYLLABLE-BA99
        0xBA9A, # G+12240  U+BA9A  uniBA9A,  HANGUL SYLLABLE-BA9A
        0xBA9B, # G+12241  U+BA9B  uniBA9B,  HANGUL SYLLABLE-BA9B
        0xBA9C, # G+12242  U+BA9C  uniBA9C,  HANGUL SYLLABLE-BA9C
        0xBA9D, # G+12243  U+BA9D  uniBA9D,  HANGUL SYLLABLE-BA9D
        0xBA9E, # G+12244  U+BA9E  uniBA9E,  HANGUL SYLLABLE-BA9E
        0xBA9F, # G+12245  U+BA9F  uniBA9F,  HANGUL SYLLABLE-BA9F
        0xBAA0, # G+12246  U+BAA0  uniBAA0,  HANGUL SYLLABLE-BAA0
        0xBAA1, # G+12247  U+BAA1  uniBAA1,  HANGUL SYLLABLE-BAA1
        0xBAA2, # G+12248  U+BAA2  uniBAA2,  HANGUL SYLLABLE-BAA2
        0xBAA3, # G+12249  U+BAA3  uniBAA3,  HANGUL SYLLABLE-BAA3
        0xBAA4, # G+12250  U+BAA4  uniBAA4,  HANGUL SYLLABLE-BAA4
        0xBAA5, # G+12251  U+BAA5  uniBAA5,  HANGUL SYLLABLE-BAA5
        0xBAA6, # G+12252  U+BAA6  uniBAA6,  HANGUL SYLLABLE-BAA6
        0xBAA7, # G+12253  U+BAA7  uniBAA7,  HANGUL SYLLABLE-BAA7
        0xBAAA, # G+12254  U+BAAA  uniBAAA,  HANGUL SYLLABLE-BAAA
        0xBAAD, # G+12255  U+BAAD  uniBAAD,  HANGUL SYLLABLE-BAAD
        0xBAAE, # G+12256  U+BAAE  uniBAAE,  HANGUL SYLLABLE-BAAE
        0xBAAF, # G+12257  U+BAAF  uniBAAF,  HANGUL SYLLABLE-BAAF
        0xBAB1, # G+12258  U+BAB1  uniBAB1,  HANGUL SYLLABLE-BAB1
        0xBAB3, # G+12259  U+BAB3  uniBAB3,  HANGUL SYLLABLE-BAB3
        0xBAB4, # G+12260  U+BAB4  uniBAB4,  HANGUL SYLLABLE-BAB4
        0xBAB5, # G+12261  U+BAB5  uniBAB5,  HANGUL SYLLABLE-BAB5
        0xBAB6, # G+12262  U+BAB6  uniBAB6,  HANGUL SYLLABLE-BAB6
        0xBAB7, # G+12263  U+BAB7  uniBAB7,  HANGUL SYLLABLE-BAB7
        0xBABA, # G+12264  U+BABA  uniBABA,  HANGUL SYLLABLE-BABA
        0xBABC, # G+12265  U+BABC  uniBABC,  HANGUL SYLLABLE-BABC
        0xBABE, # G+12266  U+BABE  uniBABE,  HANGUL SYLLABLE-BABE
        0xBABF, # G+12267  U+BABF  uniBABF,  HANGUL SYLLABLE-BABF
        0xBAC0, # G+12268  U+BAC0  uniBAC0,  HANGUL SYLLABLE-BAC0
        0xBAC1, # G+12269  U+BAC1  uniBAC1,  HANGUL SYLLABLE-BAC1
        0xBAC2, # G+12270  U+BAC2  uniBAC2,  HANGUL SYLLABLE-BAC2
        0xBAC3, # G+12271  U+BAC3  uniBAC3,  HANGUL SYLLABLE-BAC3
        0xBAC5, # G+12272  U+BAC5  uniBAC5,  HANGUL SYLLABLE-BAC5
        0xBAC6, # G+12273  U+BAC6  uniBAC6,  HANGUL SYLLABLE-BAC6
        0xBAC7, # G+12274  U+BAC7  uniBAC7,  HANGUL SYLLABLE-BAC7
        0xBAC9, # G+12275  U+BAC9  uniBAC9,  HANGUL SYLLABLE-BAC9
        0xBACA, # G+12276  U+BACA  uniBACA,  HANGUL SYLLABLE-BACA
        0xBACB, # G+12277  U+BACB  uniBACB,  HANGUL SYLLABLE-BACB
        0xBACC, # G+12278  U+BACC  uniBACC,  HANGUL SYLLABLE-BACC
        0xBACD, # G+12279  U+BACD  uniBACD,  HANGUL SYLLABLE-BACD
        0xBACE, # G+12280  U+BACE  uniBACE,  HANGUL SYLLABLE-BACE
        0xBACF, # G+12281  U+BACF  uniBACF,  HANGUL SYLLABLE-BACF
        0xBAD0, # G+12282  U+BAD0  uniBAD0,  HANGUL SYLLABLE-BAD0
        0xBAD1, # G+12283  U+BAD1  uniBAD1,  HANGUL SYLLABLE-BAD1
        0xBAD2, # G+12284  U+BAD2  uniBAD2,  HANGUL SYLLABLE-BAD2
        0xBAD3, # G+12285  U+BAD3  uniBAD3,  HANGUL SYLLABLE-BAD3
        0xBAD4, # G+12286  U+BAD4  uniBAD4,  HANGUL SYLLABLE-BAD4
        0xBAD5, # G+12287  U+BAD5  uniBAD5,  HANGUL SYLLABLE-BAD5
        0xBAD6, # G+12288  U+BAD6  uniBAD6,  HANGUL SYLLABLE-BAD6
        0xBAD7, # G+12289  U+BAD7  uniBAD7,  HANGUL SYLLABLE-BAD7
        0xBADA, # G+12290  U+BADA  uniBADA,  HANGUL SYLLABLE-BADA
        0xBADB, # G+12291  U+BADB  uniBADB,  HANGUL SYLLABLE-BADB
        0xBADC, # G+12292  U+BADC  uniBADC,  HANGUL SYLLABLE-BADC
        0xBADD, # G+12293  U+BADD  uniBADD,  HANGUL SYLLABLE-BADD
        0xBADE, # G+12294  U+BADE  uniBADE,  HANGUL SYLLABLE-BADE
        0xBADF, # G+12295  U+BADF  uniBADF,  HANGUL SYLLABLE-BADF
        0xBAE0, # G+12296  U+BAE0  uniBAE0,  HANGUL SYLLABLE-BAE0
        0xBAE1, # G+12297  U+BAE1  uniBAE1,  HANGUL SYLLABLE-BAE1
        0xBAE2, # G+12298  U+BAE2  uniBAE2,  HANGUL SYLLABLE-BAE2
        0xBAE3, # G+12299  U+BAE3  uniBAE3,  HANGUL SYLLABLE-BAE3
        0xBAE4, # G+12300  U+BAE4  uniBAE4,  HANGUL SYLLABLE-BAE4
        0xBAE5, # G+12301  U+BAE5  uniBAE5,  HANGUL SYLLABLE-BAE5
        0xBAE6, # G+12302  U+BAE6  uniBAE6,  HANGUL SYLLABLE-BAE6
        0xBAE7, # G+12303  U+BAE7  uniBAE7,  HANGUL SYLLABLE-BAE7
        0xBAE8, # G+12304  U+BAE8  uniBAE8,  HANGUL SYLLABLE-BAE8
        0xBAE9, # G+12305  U+BAE9  uniBAE9,  HANGUL SYLLABLE-BAE9
        0xBAEA, # G+12306  U+BAEA  uniBAEA,  HANGUL SYLLABLE-BAEA
        0xBAEB, # G+12307  U+BAEB  uniBAEB,  HANGUL SYLLABLE-BAEB
        0xBAEC, # G+12308  U+BAEC  uniBAEC,  HANGUL SYLLABLE-BAEC
        0xBAED, # G+12309  U+BAED  uniBAED,  HANGUL SYLLABLE-BAED
        0xBAEE, # G+12310  U+BAEE  uniBAEE,  HANGUL SYLLABLE-BAEE
        0xBAEF, # G+12311  U+BAEF  uniBAEF,  HANGUL SYLLABLE-BAEF
        0xBAF0, # G+12312  U+BAF0  uniBAF0,  HANGUL SYLLABLE-BAF0
        0xBAF1, # G+12313  U+BAF1  uniBAF1,  HANGUL SYLLABLE-BAF1
        0xBAF2, # G+12314  U+BAF2  uniBAF2,  HANGUL SYLLABLE-BAF2
        0xBAF3, # G+12315  U+BAF3  uniBAF3,  HANGUL SYLLABLE-BAF3
        0xBAF4, # G+12316  U+BAF4  uniBAF4,  HANGUL SYLLABLE-BAF4
        0xBAF5, # G+12317  U+BAF5  uniBAF5,  HANGUL SYLLABLE-BAF5
        0xBAF6, # G+12318  U+BAF6  uniBAF6,  HANGUL SYLLABLE-BAF6
        0xBAF7, # G+12319  U+BAF7  uniBAF7,  HANGUL SYLLABLE-BAF7
        0xBAF8, # G+12320  U+BAF8  uniBAF8,  HANGUL SYLLABLE-BAF8
        0xBAF9, # G+12321  U+BAF9  uniBAF9,  HANGUL SYLLABLE-BAF9
        0xBAFA, # G+12322  U+BAFA  uniBAFA,  HANGUL SYLLABLE-BAFA
        0xBAFB, # G+12323  U+BAFB  uniBAFB,  HANGUL SYLLABLE-BAFB
        0xBAFD, # G+12324  U+BAFD  uniBAFD,  HANGUL SYLLABLE-BAFD
        0xBAFE, # G+12325  U+BAFE  uniBAFE,  HANGUL SYLLABLE-BAFE
        0xBAFF, # G+12326  U+BAFF  uniBAFF,  HANGUL SYLLABLE-BAFF
        0xBB01, # G+12327  U+BB01  uniBB01,  HANGUL SYLLABLE-BB01
        0xBB02, # G+12328  U+BB02  uniBB02,  HANGUL SYLLABLE-BB02
        0xBB03, # G+12329  U+BB03  uniBB03,  HANGUL SYLLABLE-BB03
        0xBB05, # G+12330  U+BB05  uniBB05,  HANGUL SYLLABLE-BB05
        0xBB06, # G+12331  U+BB06  uniBB06,  HANGUL SYLLABLE-BB06
        0xBB07, # G+12332  U+BB07  uniBB07,  HANGUL SYLLABLE-BB07
        0xBB08, # G+12333  U+BB08  uniBB08,  HANGUL SYLLABLE-BB08
        0xBB09, # G+12334  U+BB09  uniBB09,  HANGUL SYLLABLE-BB09
        0xBB0A, # G+12335  U+BB0A  uniBB0A,  HANGUL SYLLABLE-BB0A
        0xBB0B, # G+12336  U+BB0B  uniBB0B,  HANGUL SYLLABLE-BB0B
        0xBB0C, # G+12337  U+BB0C  uniBB0C,  HANGUL SYLLABLE-BB0C
        0xBB0E, # G+12338  U+BB0E  uniBB0E,  HANGUL SYLLABLE-BB0E
        0xBB10, # G+12339  U+BB10  uniBB10,  HANGUL SYLLABLE-BB10
        0xBB12, # G+12340  U+BB12  uniBB12,  HANGUL SYLLABLE-BB12
        0xBB13, # G+12341  U+BB13  uniBB13,  HANGUL SYLLABLE-BB13
        0xBB14, # G+12342  U+BB14  uniBB14,  HANGUL SYLLABLE-BB14
        0xBB15, # G+12343  U+BB15  uniBB15,  HANGUL SYLLABLE-BB15
        0xBB16, # G+12344  U+BB16  uniBB16,  HANGUL SYLLABLE-BB16
        0xBB17, # G+12345  U+BB17  uniBB17,  HANGUL SYLLABLE-BB17
        0xBB19, # G+12346  U+BB19  uniBB19,  HANGUL SYLLABLE-BB19
        0xBB1A, # G+12347  U+BB1A  uniBB1A,  HANGUL SYLLABLE-BB1A
        0xBB1B, # G+12348  U+BB1B  uniBB1B,  HANGUL SYLLABLE-BB1B
        0xBB1D, # G+12349  U+BB1D  uniBB1D,  HANGUL SYLLABLE-BB1D
        0xBB1E, # G+12350  U+BB1E  uniBB1E,  HANGUL SYLLABLE-BB1E
        0xBB1F, # G+12351  U+BB1F  uniBB1F,  HANGUL SYLLABLE-BB1F
        0xBB21, # G+12352  U+BB21  uniBB21,  HANGUL SYLLABLE-BB21
        0xBB22, # G+12353  U+BB22  uniBB22,  HANGUL SYLLABLE-BB22
        0xBB23, # G+12354  U+BB23  uniBB23,  HANGUL SYLLABLE-BB23
        0xBB24, # G+12355  U+BB24  uniBB24,  HANGUL SYLLABLE-BB24
        0xBB25, # G+12356  U+BB25  uniBB25,  HANGUL SYLLABLE-BB25
        0xBB26, # G+12357  U+BB26  uniBB26,  HANGUL SYLLABLE-BB26
        0xBB27, # G+12358  U+BB27  uniBB27,  HANGUL SYLLABLE-BB27
        0xBB28, # G+12359  U+BB28  uniBB28,  HANGUL SYLLABLE-BB28
        0xBB2A, # G+12360  U+BB2A  uniBB2A,  HANGUL SYLLABLE-BB2A
        0xBB2C, # G+12361  U+BB2C  uniBB2C,  HANGUL SYLLABLE-BB2C
        0xBB2D, # G+12362  U+BB2D  uniBB2D,  HANGUL SYLLABLE-BB2D
        0xBB2E, # G+12363  U+BB2E  uniBB2E,  HANGUL SYLLABLE-BB2E
        0xBB2F, # G+12364  U+BB2F  uniBB2F,  HANGUL SYLLABLE-BB2F
        0xBB30, # G+12365  U+BB30  uniBB30,  HANGUL SYLLABLE-BB30
        0xBB31, # G+12366  U+BB31  uniBB31,  HANGUL SYLLABLE-BB31
        0xBB32, # G+12367  U+BB32  uniBB32,  HANGUL SYLLABLE-BB32
        0xBB33, # G+12368  U+BB33  uniBB33,  HANGUL SYLLABLE-BB33
        0xBB37, # G+12369  U+BB37  uniBB37,  HANGUL SYLLABLE-BB37
        0xBB39, # G+12370  U+BB39  uniBB39,  HANGUL SYLLABLE-BB39
        0xBB3A, # G+12371  U+BB3A  uniBB3A,  HANGUL SYLLABLE-BB3A
        0xBB3F, # G+12372  U+BB3F  uniBB3F,  HANGUL SYLLABLE-BB3F
        0xBB40, # G+12373  U+BB40  uniBB40,  HANGUL SYLLABLE-BB40
        0xBB41, # G+12374  U+BB41  uniBB41,  HANGUL SYLLABLE-BB41
        0xBB42, # G+12375  U+BB42  uniBB42,  HANGUL SYLLABLE-BB42
        0xBB43, # G+12376  U+BB43  uniBB43,  HANGUL SYLLABLE-BB43
        0xBB46, # G+12377  U+BB46  uniBB46,  HANGUL SYLLABLE-BB46
        0xBB48, # G+12378  U+BB48  uniBB48,  HANGUL SYLLABLE-BB48
        0xBB4A, # G+12379  U+BB4A  uniBB4A,  HANGUL SYLLABLE-BB4A
        0xBB4B, # G+12380  U+BB4B  uniBB4B,  HANGUL SYLLABLE-BB4B
        0xBB4C, # G+12381  U+BB4C  uniBB4C,  HANGUL SYLLABLE-BB4C
        0xBB4E, # G+12382  U+BB4E  uniBB4E,  HANGUL SYLLABLE-BB4E
        0xBB51, # G+12383  U+BB51  uniBB51,  HANGUL SYLLABLE-BB51
        0xBB52, # G+12384  U+BB52  uniBB52,  HANGUL SYLLABLE-BB52
        0xBB53, # G+12385  U+BB53  uniBB53,  HANGUL SYLLABLE-BB53
        0xBB55, # G+12386  U+BB55  uniBB55,  HANGUL SYLLABLE-BB55
        0xBB56, # G+12387  U+BB56  uniBB56,  HANGUL SYLLABLE-BB56
        0xBB57, # G+12388  U+BB57  uniBB57,  HANGUL SYLLABLE-BB57
        0xBB59, # G+12389  U+BB59  uniBB59,  HANGUL SYLLABLE-BB59
        0xBB5A, # G+12390  U+BB5A  uniBB5A,  HANGUL SYLLABLE-BB5A
        0xBB5B, # G+12391  U+BB5B  uniBB5B,  HANGUL SYLLABLE-BB5B
        0xBB5C, # G+12392  U+BB5C  uniBB5C,  HANGUL SYLLABLE-BB5C
        0xBB5D, # G+12393  U+BB5D  uniBB5D,  HANGUL SYLLABLE-BB5D
        0xBB5E, # G+12394  U+BB5E  uniBB5E,  HANGUL SYLLABLE-BB5E
        0xBB5F, # G+12395  U+BB5F  uniBB5F,  HANGUL SYLLABLE-BB5F
        0xBB60, # G+12396  U+BB60  uniBB60,  HANGUL SYLLABLE-BB60
        0xBB62, # G+12397  U+BB62  uniBB62,  HANGUL SYLLABLE-BB62
        0xBB64, # G+12398  U+BB64  uniBB64,  HANGUL SYLLABLE-BB64
        0xBB65, # G+12399  U+BB65  uniBB65,  HANGUL SYLLABLE-BB65
        0xBB66, # G+12400  U+BB66  uniBB66,  HANGUL SYLLABLE-BB66
        0xBB67, # G+12401  U+BB67  uniBB67,  HANGUL SYLLABLE-BB67
        0xBB68, # G+12402  U+BB68  uniBB68,  HANGUL SYLLABLE-BB68
        0xBB69, # G+12403  U+BB69  uniBB69,  HANGUL SYLLABLE-BB69
        0xBB6A, # G+12404  U+BB6A  uniBB6A,  HANGUL SYLLABLE-BB6A
        0xBB6B, # G+12405  U+BB6B  uniBB6B,  HANGUL SYLLABLE-BB6B
        0xBB6D, # G+12406  U+BB6D  uniBB6D,  HANGUL SYLLABLE-BB6D
        0xBB6E, # G+12407  U+BB6E  uniBB6E,  HANGUL SYLLABLE-BB6E
        0xBB6F, # G+12408  U+BB6F  uniBB6F,  HANGUL SYLLABLE-BB6F
        0xBB70, # G+12409  U+BB70  uniBB70,  HANGUL SYLLABLE-BB70
        0xBB71, # G+12410  U+BB71  uniBB71,  HANGUL SYLLABLE-BB71
        0xBB72, # G+12411  U+BB72  uniBB72,  HANGUL SYLLABLE-BB72
        0xBB73, # G+12412  U+BB73  uniBB73,  HANGUL SYLLABLE-BB73
        0xBB74, # G+12413  U+BB74  uniBB74,  HANGUL SYLLABLE-BB74
        0xBB75, # G+12414  U+BB75  uniBB75,  HANGUL SYLLABLE-BB75
        0xBB76, # G+12415  U+BB76  uniBB76,  HANGUL SYLLABLE-BB76
        0xBB77, # G+12416  U+BB77  uniBB77,  HANGUL SYLLABLE-BB77
        0xBB78, # G+12417  U+BB78  uniBB78,  HANGUL SYLLABLE-BB78
        0xBB79, # G+12418  U+BB79  uniBB79,  HANGUL SYLLABLE-BB79
        0xBB7A, # G+12419  U+BB7A  uniBB7A,  HANGUL SYLLABLE-BB7A
        0xBB7B, # G+12420  U+BB7B  uniBB7B,  HANGUL SYLLABLE-BB7B
        0xBB7C, # G+12421  U+BB7C  uniBB7C,  HANGUL SYLLABLE-BB7C
        0xBB7D, # G+12422  U+BB7D  uniBB7D,  HANGUL SYLLABLE-BB7D
        0xBB7E, # G+12423  U+BB7E  uniBB7E,  HANGUL SYLLABLE-BB7E
        0xBB7F, # G+12424  U+BB7F  uniBB7F,  HANGUL SYLLABLE-BB7F
        0xBB80, # G+12425  U+BB80  uniBB80,  HANGUL SYLLABLE-BB80
        0xBB81, # G+12426  U+BB81  uniBB81,  HANGUL SYLLABLE-BB81
        0xBB82, # G+12427  U+BB82  uniBB82,  HANGUL SYLLABLE-BB82
        0xBB83, # G+12428  U+BB83  uniBB83,  HANGUL SYLLABLE-BB83
        0xBB84, # G+12429  U+BB84  uniBB84,  HANGUL SYLLABLE-BB84
        0xBB85, # G+12430  U+BB85  uniBB85,  HANGUL SYLLABLE-BB85
        0xBB86, # G+12431  U+BB86  uniBB86,  HANGUL SYLLABLE-BB86
        0xBB87, # G+12432  U+BB87  uniBB87,  HANGUL SYLLABLE-BB87
        0xBB89, # G+12433  U+BB89  uniBB89,  HANGUL SYLLABLE-BB89
        0xBB8A, # G+12434  U+BB8A  uniBB8A,  HANGUL SYLLABLE-BB8A
        0xBB8B, # G+12435  U+BB8B  uniBB8B,  HANGUL SYLLABLE-BB8B
        0xBB8D, # G+12436  U+BB8D  uniBB8D,  HANGUL SYLLABLE-BB8D
        0xBB8E, # G+12437  U+BB8E  uniBB8E,  HANGUL SYLLABLE-BB8E
        0xBB8F, # G+12438  U+BB8F  uniBB8F,  HANGUL SYLLABLE-BB8F
        0xBB91, # G+12439  U+BB91  uniBB91,  HANGUL SYLLABLE-BB91
        0xBB92, # G+12440  U+BB92  uniBB92,  HANGUL SYLLABLE-BB92
        0xBB93, # G+12441  U+BB93  uniBB93,  HANGUL SYLLABLE-BB93
        0xBB94, # G+12442  U+BB94  uniBB94,  HANGUL SYLLABLE-BB94
        0xBB95, # G+12443  U+BB95  uniBB95,  HANGUL SYLLABLE-BB95
        0xBB96, # G+12444  U+BB96  uniBB96,  HANGUL SYLLABLE-BB96
        0xBB97, # G+12445  U+BB97  uniBB97,  HANGUL SYLLABLE-BB97
        0xBB98, # G+12446  U+BB98  uniBB98,  HANGUL SYLLABLE-BB98
        0xBB99, # G+12447  U+BB99  uniBB99,  HANGUL SYLLABLE-BB99
        0xBB9A, # G+12448  U+BB9A  uniBB9A,  HANGUL SYLLABLE-BB9A
        0xBB9B, # G+12449  U+BB9B  uniBB9B,  HANGUL SYLLABLE-BB9B
        0xBB9C, # G+12450  U+BB9C  uniBB9C,  HANGUL SYLLABLE-BB9C
        0xBB9D, # G+12451  U+BB9D  uniBB9D,  HANGUL SYLLABLE-BB9D
        0xBB9E, # G+12452  U+BB9E  uniBB9E,  HANGUL SYLLABLE-BB9E
        0xBB9F, # G+12453  U+BB9F  uniBB9F,  HANGUL SYLLABLE-BB9F
        0xBBA0, # G+12454  U+BBA0  uniBBA0,  HANGUL SYLLABLE-BBA0
        0xBBA1, # G+12455  U+BBA1  uniBBA1,  HANGUL SYLLABLE-BBA1
        0xBBA2, # G+12456  U+BBA2  uniBBA2,  HANGUL SYLLABLE-BBA2
        0xBBA3, # G+12457  U+BBA3  uniBBA3,  HANGUL SYLLABLE-BBA3
        0xBBA5, # G+12458  U+BBA5  uniBBA5,  HANGUL SYLLABLE-BBA5
        0xBBA6, # G+12459  U+BBA6  uniBBA6,  HANGUL SYLLABLE-BBA6
        0xBBA7, # G+12460  U+BBA7  uniBBA7,  HANGUL SYLLABLE-BBA7
        0xBBA9, # G+12461  U+BBA9  uniBBA9,  HANGUL SYLLABLE-BBA9
        0xBBAA, # G+12462  U+BBAA  uniBBAA,  HANGUL SYLLABLE-BBAA
        0xBBAB, # G+12463  U+BBAB  uniBBAB,  HANGUL SYLLABLE-BBAB
        0xBBAD, # G+12464  U+BBAD  uniBBAD,  HANGUL SYLLABLE-BBAD
        0xBBAE, # G+12465  U+BBAE  uniBBAE,  HANGUL SYLLABLE-BBAE
        0xBBAF, # G+12466  U+BBAF  uniBBAF,  HANGUL SYLLABLE-BBAF
        0xBBB0, # G+12467  U+BBB0  uniBBB0,  HANGUL SYLLABLE-BBB0
        0xBBB1, # G+12468  U+BBB1  uniBBB1,  HANGUL SYLLABLE-BBB1
        0xBBB2, # G+12469  U+BBB2  uniBBB2,  HANGUL SYLLABLE-BBB2
        0xBBB3, # G+12470  U+BBB3  uniBBB3,  HANGUL SYLLABLE-BBB3
        0xBBB5, # G+12471  U+BBB5  uniBBB5,  HANGUL SYLLABLE-BBB5
        0xBBB6, # G+12472  U+BBB6  uniBBB6,  HANGUL SYLLABLE-BBB6
        0xBBB8, # G+12473  U+BBB8  uniBBB8,  HANGUL SYLLABLE-BBB8
        0xBBB9, # G+12474  U+BBB9  uniBBB9,  HANGUL SYLLABLE-BBB9
        0xBBBA, # G+12475  U+BBBA  uniBBBA,  HANGUL SYLLABLE-BBBA
        0xBBBB, # G+12476  U+BBBB  uniBBBB,  HANGUL SYLLABLE-BBBB
        0xBBBC, # G+12477  U+BBBC  uniBBBC,  HANGUL SYLLABLE-BBBC
        0xBBBD, # G+12478  U+BBBD  uniBBBD,  HANGUL SYLLABLE-BBBD
        0xBBBE, # G+12479  U+BBBE  uniBBBE,  HANGUL SYLLABLE-BBBE
        0xBBBF, # G+12480  U+BBBF  uniBBBF,  HANGUL SYLLABLE-BBBF
        0xBBC1, # G+12481  U+BBC1  uniBBC1,  HANGUL SYLLABLE-BBC1
        0xBBC2, # G+12482  U+BBC2  uniBBC2,  HANGUL SYLLABLE-BBC2
        0xBBC3, # G+12483  U+BBC3  uniBBC3,  HANGUL SYLLABLE-BBC3
        0xBBC5, # G+12484  U+BBC5  uniBBC5,  HANGUL SYLLABLE-BBC5
        0xBBC6, # G+12485  U+BBC6  uniBBC6,  HANGUL SYLLABLE-BBC6
        0xBBC7, # G+12486  U+BBC7  uniBBC7,  HANGUL SYLLABLE-BBC7
        0xBBC9, # G+12487  U+BBC9  uniBBC9,  HANGUL SYLLABLE-BBC9
        0xBBCA, # G+12488  U+BBCA  uniBBCA,  HANGUL SYLLABLE-BBCA
        0xBBCB, # G+12489  U+BBCB  uniBBCB,  HANGUL SYLLABLE-BBCB
        0xBBCC, # G+12490  U+BBCC  uniBBCC,  HANGUL SYLLABLE-BBCC
        0xBBCD, # G+12491  U+BBCD  uniBBCD,  HANGUL SYLLABLE-BBCD
        0xBBCE, # G+12492  U+BBCE  uniBBCE,  HANGUL SYLLABLE-BBCE
        0xBBCF, # G+12493  U+BBCF  uniBBCF,  HANGUL SYLLABLE-BBCF
        0xBBD1, # G+12494  U+BBD1  uniBBD1,  HANGUL SYLLABLE-BBD1
        0xBBD2, # G+12495  U+BBD2  uniBBD2,  HANGUL SYLLABLE-BBD2
        0xBBD4, # G+12496  U+BBD4  uniBBD4,  HANGUL SYLLABLE-BBD4
        0xBBD5, # G+12497  U+BBD5  uniBBD5,  HANGUL SYLLABLE-BBD5
        0xBBD6, # G+12498  U+BBD6  uniBBD6,  HANGUL SYLLABLE-BBD6
        0xBBD7, # G+12499  U+BBD7  uniBBD7,  HANGUL SYLLABLE-BBD7
        0xBBD8, # G+12500  U+BBD8  uniBBD8,  HANGUL SYLLABLE-BBD8
        0xBBD9, # G+12501  U+BBD9  uniBBD9,  HANGUL SYLLABLE-BBD9
        0xBBDA, # G+12502  U+BBDA  uniBBDA,  HANGUL SYLLABLE-BBDA
        0xBBDB, # G+12503  U+BBDB  uniBBDB,  HANGUL SYLLABLE-BBDB
        0xBBDC, # G+12504  U+BBDC  uniBBDC,  HANGUL SYLLABLE-BBDC
        0xBBDD, # G+12505  U+BBDD  uniBBDD,  HANGUL SYLLABLE-BBDD
        0xBBDE, # G+12506  U+BBDE  uniBBDE,  HANGUL SYLLABLE-BBDE
        0xBBDF, # G+12507  U+BBDF  uniBBDF,  HANGUL SYLLABLE-BBDF
        0xBBE0, # G+12508  U+BBE0  uniBBE0,  HANGUL SYLLABLE-BBE0
        0xBBE1, # G+12509  U+BBE1  uniBBE1,  HANGUL SYLLABLE-BBE1
        0xBBE2, # G+12510  U+BBE2  uniBBE2,  HANGUL SYLLABLE-BBE2
        0xBBE3, # G+12511  U+BBE3  uniBBE3,  HANGUL SYLLABLE-BBE3
        0xBBE4, # G+12512  U+BBE4  uniBBE4,  HANGUL SYLLABLE-BBE4
        0xBBE5, # G+12513  U+BBE5  uniBBE5,  HANGUL SYLLABLE-BBE5
        0xBBE6, # G+12514  U+BBE6  uniBBE6,  HANGUL SYLLABLE-BBE6
        0xBBE7, # G+12515  U+BBE7  uniBBE7,  HANGUL SYLLABLE-BBE7
        0xBBE8, # G+12516  U+BBE8  uniBBE8,  HANGUL SYLLABLE-BBE8
        0xBBE9, # G+12517  U+BBE9  uniBBE9,  HANGUL SYLLABLE-BBE9
        0xBBEA, # G+12518  U+BBEA  uniBBEA,  HANGUL SYLLABLE-BBEA
        0xBBEB, # G+12519  U+BBEB  uniBBEB,  HANGUL SYLLABLE-BBEB
        0xBBEC, # G+12520  U+BBEC  uniBBEC,  HANGUL SYLLABLE-BBEC
        0xBBED, # G+12521  U+BBED  uniBBED,  HANGUL SYLLABLE-BBED
        0xBBEE, # G+12522  U+BBEE  uniBBEE,  HANGUL SYLLABLE-BBEE
        0xBBEF, # G+12523  U+BBEF  uniBBEF,  HANGUL SYLLABLE-BBEF
        0xBBF0, # G+12524  U+BBF0  uniBBF0,  HANGUL SYLLABLE-BBF0
        0xBBF1, # G+12525  U+BBF1  uniBBF1,  HANGUL SYLLABLE-BBF1
        0xBBF2, # G+12526  U+BBF2  uniBBF2,  HANGUL SYLLABLE-BBF2
        0xBBF3, # G+12527  U+BBF3  uniBBF3,  HANGUL SYLLABLE-BBF3
        0xBBF4, # G+12528  U+BBF4  uniBBF4,  HANGUL SYLLABLE-BBF4
        0xBBF5, # G+12529  U+BBF5  uniBBF5,  HANGUL SYLLABLE-BBF5
        0xBBF6, # G+12530  U+BBF6  uniBBF6,  HANGUL SYLLABLE-BBF6
        0xBBF7, # G+12531  U+BBF7  uniBBF7,  HANGUL SYLLABLE-BBF7
        0xBBFA, # G+12532  U+BBFA  uniBBFA,  HANGUL SYLLABLE-BBFA
        0xBBFB, # G+12533  U+BBFB  uniBBFB,  HANGUL SYLLABLE-BBFB
        0xBBFD, # G+12534  U+BBFD  uniBBFD,  HANGUL SYLLABLE-BBFD
        0xBBFE, # G+12535  U+BBFE  uniBBFE,  HANGUL SYLLABLE-BBFE
        0xBC01, # G+12536  U+BC01  uniBC01,  HANGUL SYLLABLE-BC01
        0xBC03, # G+12537  U+BC03  uniBC03,  HANGUL SYLLABLE-BC03
        0xBC04, # G+12538  U+BC04  uniBC04,  HANGUL SYLLABLE-BC04
        0xBC05, # G+12539  U+BC05  uniBC05,  HANGUL SYLLABLE-BC05
        0xBC06, # G+12540  U+BC06  uniBC06,  HANGUL SYLLABLE-BC06
        0xBC07, # G+12541  U+BC07  uniBC07,  HANGUL SYLLABLE-BC07
        0xBC0A, # G+12542  U+BC0A  uniBC0A,  HANGUL SYLLABLE-BC0A
        0xBC0E, # G+12543  U+BC0E  uniBC0E,  HANGUL SYLLABLE-BC0E
        0xBC10, # G+12544  U+BC10  uniBC10,  HANGUL SYLLABLE-BC10
        0xBC12, # G+12545  U+BC12  uniBC12,  HANGUL SYLLABLE-BC12
        0xBC13, # G+12546  U+BC13  uniBC13,  HANGUL SYLLABLE-BC13
        0xBC19, # G+12547  U+BC19  uniBC19,  HANGUL SYLLABLE-BC19
        0xBC1A, # G+12548  U+BC1A  uniBC1A,  HANGUL SYLLABLE-BC1A
        0xBC20, # G+12549  U+BC20  uniBC20,  HANGUL SYLLABLE-BC20
        0xBC21, # G+12550  U+BC21  uniBC21,  HANGUL SYLLABLE-BC21
        0xBC22, # G+12551  U+BC22  uniBC22,  HANGUL SYLLABLE-BC22
        0xBC23, # G+12552  U+BC23  uniBC23,  HANGUL SYLLABLE-BC23
        0xBC26, # G+12553  U+BC26  uniBC26,  HANGUL SYLLABLE-BC26
        0xBC28, # G+12554  U+BC28  uniBC28,  HANGUL SYLLABLE-BC28
        0xBC2A, # G+12555  U+BC2A  uniBC2A,  HANGUL SYLLABLE-BC2A
        0xBC2B, # G+12556  U+BC2B  uniBC2B,  HANGUL SYLLABLE-BC2B
        0xBC2C, # G+12557  U+BC2C  uniBC2C,  HANGUL SYLLABLE-BC2C
        0xBC2E, # G+12558  U+BC2E  uniBC2E,  HANGUL SYLLABLE-BC2E
        0xBC2F, # G+12559  U+BC2F  uniBC2F,  HANGUL SYLLABLE-BC2F
        0xBC32, # G+12560  U+BC32  uniBC32,  HANGUL SYLLABLE-BC32
        0xBC33, # G+12561  U+BC33  uniBC33,  HANGUL SYLLABLE-BC33
        0xBC35, # G+12562  U+BC35  uniBC35,  HANGUL SYLLABLE-BC35
        0xBC36, # G+12563  U+BC36  uniBC36,  HANGUL SYLLABLE-BC36
        0xBC37, # G+12564  U+BC37  uniBC37,  HANGUL SYLLABLE-BC37
        0xBC39, # G+12565  U+BC39  uniBC39,  HANGUL SYLLABLE-BC39
        0xBC3A, # G+12566  U+BC3A  uniBC3A,  HANGUL SYLLABLE-BC3A
        0xBC3B, # G+12567  U+BC3B  uniBC3B,  HANGUL SYLLABLE-BC3B
        0xBC3C, # G+12568  U+BC3C  uniBC3C,  HANGUL SYLLABLE-BC3C
        0xBC3D, # G+12569  U+BC3D  uniBC3D,  HANGUL SYLLABLE-BC3D
        0xBC3E, # G+12570  U+BC3E  uniBC3E,  HANGUL SYLLABLE-BC3E
        0xBC3F, # G+12571  U+BC3F  uniBC3F,  HANGUL SYLLABLE-BC3F
        0xBC42, # G+12572  U+BC42  uniBC42,  HANGUL SYLLABLE-BC42
        0xBC46, # G+12573  U+BC46  uniBC46,  HANGUL SYLLABLE-BC46
        0xBC47, # G+12574  U+BC47  uniBC47,  HANGUL SYLLABLE-BC47
        0xBC48, # G+12575  U+BC48  uniBC48,  HANGUL SYLLABLE-BC48
        0xBC4A, # G+12576  U+BC4A  uniBC4A,  HANGUL SYLLABLE-BC4A
        0xBC4B, # G+12577  U+BC4B  uniBC4B,  HANGUL SYLLABLE-BC4B
        0xBC4E, # G+12578  U+BC4E  uniBC4E,  HANGUL SYLLABLE-BC4E
        0xBC4F, # G+12579  U+BC4F  uniBC4F,  HANGUL SYLLABLE-BC4F
        0xBC51, # G+12580  U+BC51  uniBC51,  HANGUL SYLLABLE-BC51
        0xBC52, # G+12581  U+BC52  uniBC52,  HANGUL SYLLABLE-BC52
        0xBC53, # G+12582  U+BC53  uniBC53,  HANGUL SYLLABLE-BC53
        0xBC54, # G+12583  U+BC54  uniBC54,  HANGUL SYLLABLE-BC54
        0xBC55, # G+12584  U+BC55  uniBC55,  HANGUL SYLLABLE-BC55
        0xBC56, # G+12585  U+BC56  uniBC56,  HANGUL SYLLABLE-BC56
        0xBC57, # G+12586  U+BC57  uniBC57,  HANGUL SYLLABLE-BC57
        0xBC58, # G+12587  U+BC58  uniBC58,  HANGUL SYLLABLE-BC58
        0xBC59, # G+12588  U+BC59  uniBC59,  HANGUL SYLLABLE-BC59
        0xBC5A, # G+12589  U+BC5A  uniBC5A,  HANGUL SYLLABLE-BC5A
        0xBC5B, # G+12590  U+BC5B  uniBC5B,  HANGUL SYLLABLE-BC5B
        0xBC5C, # G+12591  U+BC5C  uniBC5C,  HANGUL SYLLABLE-BC5C
        0xBC5E, # G+12592  U+BC5E  uniBC5E,  HANGUL SYLLABLE-BC5E
        0xBC5F, # G+12593  U+BC5F  uniBC5F,  HANGUL SYLLABLE-BC5F
        0xBC60, # G+12594  U+BC60  uniBC60,  HANGUL SYLLABLE-BC60
        0xBC61, # G+12595  U+BC61  uniBC61,  HANGUL SYLLABLE-BC61
        0xBC62, # G+12596  U+BC62  uniBC62,  HANGUL SYLLABLE-BC62
        0xBC63, # G+12597  U+BC63  uniBC63,  HANGUL SYLLABLE-BC63
        0xBC64, # G+12598  U+BC64  uniBC64,  HANGUL SYLLABLE-BC64
        0xBC65, # G+12599  U+BC65  uniBC65,  HANGUL SYLLABLE-BC65
        0xBC66, # G+12600  U+BC66  uniBC66,  HANGUL SYLLABLE-BC66
        0xBC67, # G+12601  U+BC67  uniBC67,  HANGUL SYLLABLE-BC67
        0xBC68, # G+12602  U+BC68  uniBC68,  HANGUL SYLLABLE-BC68
        0xBC69, # G+12603  U+BC69  uniBC69,  HANGUL SYLLABLE-BC69
        0xBC6A, # G+12604  U+BC6A  uniBC6A,  HANGUL SYLLABLE-BC6A
        0xBC6B, # G+12605  U+BC6B  uniBC6B,  HANGUL SYLLABLE-BC6B
        0xBC6C, # G+12606  U+BC6C  uniBC6C,  HANGUL SYLLABLE-BC6C
        0xBC6D, # G+12607  U+BC6D  uniBC6D,  HANGUL SYLLABLE-BC6D
        0xBC6E, # G+12608  U+BC6E  uniBC6E,  HANGUL SYLLABLE-BC6E
        0xBC6F, # G+12609  U+BC6F  uniBC6F,  HANGUL SYLLABLE-BC6F
        0xBC70, # G+12610  U+BC70  uniBC70,  HANGUL SYLLABLE-BC70
        0xBC71, # G+12611  U+BC71  uniBC71,  HANGUL SYLLABLE-BC71
        0xBC72, # G+12612  U+BC72  uniBC72,  HANGUL SYLLABLE-BC72
        0xBC73, # G+12613  U+BC73  uniBC73,  HANGUL SYLLABLE-BC73
        0xBC74, # G+12614  U+BC74  uniBC74,  HANGUL SYLLABLE-BC74
        0xBC75, # G+12615  U+BC75  uniBC75,  HANGUL SYLLABLE-BC75
        0xBC76, # G+12616  U+BC76  uniBC76,  HANGUL SYLLABLE-BC76
        0xBC77, # G+12617  U+BC77  uniBC77,  HANGUL SYLLABLE-BC77
        0xBC78, # G+12618  U+BC78  uniBC78,  HANGUL SYLLABLE-BC78
        0xBC79, # G+12619  U+BC79  uniBC79,  HANGUL SYLLABLE-BC79
        0xBC7A, # G+12620  U+BC7A  uniBC7A,  HANGUL SYLLABLE-BC7A
        0xBC7B, # G+12621  U+BC7B  uniBC7B,  HANGUL SYLLABLE-BC7B
        0xBC7C, # G+12622  U+BC7C  uniBC7C,  HANGUL SYLLABLE-BC7C
        0xBC7D, # G+12623  U+BC7D  uniBC7D,  HANGUL SYLLABLE-BC7D
        0xBC7E, # G+12624  U+BC7E  uniBC7E,  HANGUL SYLLABLE-BC7E
        0xBC7F, # G+12625  U+BC7F  uniBC7F,  HANGUL SYLLABLE-BC7F
        0xBC80, # G+12626  U+BC80  uniBC80,  HANGUL SYLLABLE-BC80
        0xBC81, # G+12627  U+BC81  uniBC81,  HANGUL SYLLABLE-BC81
        0xBC82, # G+12628  U+BC82  uniBC82,  HANGUL SYLLABLE-BC82
        0xBC83, # G+12629  U+BC83  uniBC83,  HANGUL SYLLABLE-BC83
        0xBC86, # G+12630  U+BC86  uniBC86,  HANGUL SYLLABLE-BC86
        0xBC87, # G+12631  U+BC87  uniBC87,  HANGUL SYLLABLE-BC87
        0xBC89, # G+12632  U+BC89  uniBC89,  HANGUL SYLLABLE-BC89
        0xBC8A, # G+12633  U+BC8A  uniBC8A,  HANGUL SYLLABLE-BC8A
        0xBC8D, # G+12634  U+BC8D  uniBC8D,  HANGUL SYLLABLE-BC8D
        0xBC8F, # G+12635  U+BC8F  uniBC8F,  HANGUL SYLLABLE-BC8F
        0xBC90, # G+12636  U+BC90  uniBC90,  HANGUL SYLLABLE-BC90
        0xBC91, # G+12637  U+BC91  uniBC91,  HANGUL SYLLABLE-BC91
        0xBC92, # G+12638  U+BC92  uniBC92,  HANGUL SYLLABLE-BC92
        0xBC93, # G+12639  U+BC93  uniBC93,  HANGUL SYLLABLE-BC93
        0xBC96, # G+12640  U+BC96  uniBC96,  HANGUL SYLLABLE-BC96
        0xBC98, # G+12641  U+BC98  uniBC98,  HANGUL SYLLABLE-BC98
        0xBC9B, # G+12642  U+BC9B  uniBC9B,  HANGUL SYLLABLE-BC9B
        0xBC9C, # G+12643  U+BC9C  uniBC9C,  HANGUL SYLLABLE-BC9C
        0xBC9D, # G+12644  U+BC9D  uniBC9D,  HANGUL SYLLABLE-BC9D
        0xBC9E, # G+12645  U+BC9E  uniBC9E,  HANGUL SYLLABLE-BC9E
        0xBC9F, # G+12646  U+BC9F  uniBC9F,  HANGUL SYLLABLE-BC9F
        0xBCA2, # G+12647  U+BCA2  uniBCA2,  HANGUL SYLLABLE-BCA2
        0xBCA3, # G+12648  U+BCA3  uniBCA3,  HANGUL SYLLABLE-BCA3
        0xBCA5, # G+12649  U+BCA5  uniBCA5,  HANGUL SYLLABLE-BCA5
        0xBCA6, # G+12650  U+BCA6  uniBCA6,  HANGUL SYLLABLE-BCA6
        0xBCA9, # G+12651  U+BCA9  uniBCA9,  HANGUL SYLLABLE-BCA9
        0xBCAA, # G+12652  U+BCAA  uniBCAA,  HANGUL SYLLABLE-BCAA
        0xBCAB, # G+12653  U+BCAB  uniBCAB,  HANGUL SYLLABLE-BCAB
        0xBCAC, # G+12654  U+BCAC  uniBCAC,  HANGUL SYLLABLE-BCAC
        0xBCAD, # G+12655  U+BCAD  uniBCAD,  HANGUL SYLLABLE-BCAD
        0xBCAE, # G+12656  U+BCAE  uniBCAE,  HANGUL SYLLABLE-BCAE
        0xBCAF, # G+12657  U+BCAF  uniBCAF,  HANGUL SYLLABLE-BCAF
        0xBCB2, # G+12658  U+BCB2  uniBCB2,  HANGUL SYLLABLE-BCB2
        0xBCB6, # G+12659  U+BCB6  uniBCB6,  HANGUL SYLLABLE-BCB6
        0xBCB7, # G+12660  U+BCB7  uniBCB7,  HANGUL SYLLABLE-BCB7
        0xBCB8, # G+12661  U+BCB8  uniBCB8,  HANGUL SYLLABLE-BCB8
        0xBCB9, # G+12662  U+BCB9  uniBCB9,  HANGUL SYLLABLE-BCB9
        0xBCBA, # G+12663  U+BCBA  uniBCBA,  HANGUL SYLLABLE-BCBA
        0xBCBB, # G+12664  U+BCBB  uniBCBB,  HANGUL SYLLABLE-BCBB
        0xBCBE, # G+12665  U+BCBE  uniBCBE,  HANGUL SYLLABLE-BCBE
        0xBCBF, # G+12666  U+BCBF  uniBCBF,  HANGUL SYLLABLE-BCBF
        0xBCC1, # G+12667  U+BCC1  uniBCC1,  HANGUL SYLLABLE-BCC1
        0xBCC2, # G+12668  U+BCC2  uniBCC2,  HANGUL SYLLABLE-BCC2
        0xBCC3, # G+12669  U+BCC3  uniBCC3,  HANGUL SYLLABLE-BCC3
        0xBCC5, # G+12670  U+BCC5  uniBCC5,  HANGUL SYLLABLE-BCC5
        0xBCC6, # G+12671  U+BCC6  uniBCC6,  HANGUL SYLLABLE-BCC6
        0xBCC7, # G+12672  U+BCC7  uniBCC7,  HANGUL SYLLABLE-BCC7
        0xBCC8, # G+12673  U+BCC8  uniBCC8,  HANGUL SYLLABLE-BCC8
        0xBCC9, # G+12674  U+BCC9  uniBCC9,  HANGUL SYLLABLE-BCC9
        0xBCCA, # G+12675  U+BCCA  uniBCCA,  HANGUL SYLLABLE-BCCA
        0xBCCB, # G+12676  U+BCCB  uniBCCB,  HANGUL SYLLABLE-BCCB
        0xBCCC, # G+12677  U+BCCC  uniBCCC,  HANGUL SYLLABLE-BCCC
        0xBCCE, # G+12678  U+BCCE  uniBCCE,  HANGUL SYLLABLE-BCCE
        0xBCD2, # G+12679  U+BCD2  uniBCD2,  HANGUL SYLLABLE-BCD2
        0xBCD3, # G+12680  U+BCD3  uniBCD3,  HANGUL SYLLABLE-BCD3
        0xBCD4, # G+12681  U+BCD4  uniBCD4,  HANGUL SYLLABLE-BCD4
        0xBCD6, # G+12682  U+BCD6  uniBCD6,  HANGUL SYLLABLE-BCD6
        0xBCD7, # G+12683  U+BCD7  uniBCD7,  HANGUL SYLLABLE-BCD7
        0xBCD9, # G+12684  U+BCD9  uniBCD9,  HANGUL SYLLABLE-BCD9
        0xBCDA, # G+12685  U+BCDA  uniBCDA,  HANGUL SYLLABLE-BCDA
        0xBCDB, # G+12686  U+BCDB  uniBCDB,  HANGUL SYLLABLE-BCDB
        0xBCDD, # G+12687  U+BCDD  uniBCDD,  HANGUL SYLLABLE-BCDD
        0xBCDE, # G+12688  U+BCDE  uniBCDE,  HANGUL SYLLABLE-BCDE
        0xBCDF, # G+12689  U+BCDF  uniBCDF,  HANGUL SYLLABLE-BCDF
        0xBCE0, # G+12690  U+BCE0  uniBCE0,  HANGUL SYLLABLE-BCE0
        0xBCE1, # G+12691  U+BCE1  uniBCE1,  HANGUL SYLLABLE-BCE1
        0xBCE2, # G+12692  U+BCE2  uniBCE2,  HANGUL SYLLABLE-BCE2
        0xBCE3, # G+12693  U+BCE3  uniBCE3,  HANGUL SYLLABLE-BCE3
        0xBCE4, # G+12694  U+BCE4  uniBCE4,  HANGUL SYLLABLE-BCE4
        0xBCE5, # G+12695  U+BCE5  uniBCE5,  HANGUL SYLLABLE-BCE5
        0xBCE6, # G+12696  U+BCE6  uniBCE6,  HANGUL SYLLABLE-BCE6
        0xBCE7, # G+12697  U+BCE7  uniBCE7,  HANGUL SYLLABLE-BCE7
        0xBCE8, # G+12698  U+BCE8  uniBCE8,  HANGUL SYLLABLE-BCE8
        0xBCE9, # G+12699  U+BCE9  uniBCE9,  HANGUL SYLLABLE-BCE9
        0xBCEA, # G+12700  U+BCEA  uniBCEA,  HANGUL SYLLABLE-BCEA
        0xBCEB, # G+12701  U+BCEB  uniBCEB,  HANGUL SYLLABLE-BCEB
        0xBCEC, # G+12702  U+BCEC  uniBCEC,  HANGUL SYLLABLE-BCEC
        0xBCED, # G+12703  U+BCED  uniBCED,  HANGUL SYLLABLE-BCED
        0xBCEE, # G+12704  U+BCEE  uniBCEE,  HANGUL SYLLABLE-BCEE
        0xBCEF, # G+12705  U+BCEF  uniBCEF,  HANGUL SYLLABLE-BCEF
        0xBCF0, # G+12706  U+BCF0  uniBCF0,  HANGUL SYLLABLE-BCF0
        0xBCF1, # G+12707  U+BCF1  uniBCF1,  HANGUL SYLLABLE-BCF1
        0xBCF2, # G+12708  U+BCF2  uniBCF2,  HANGUL SYLLABLE-BCF2
        0xBCF3, # G+12709  U+BCF3  uniBCF3,  HANGUL SYLLABLE-BCF3
        0xBCF7, # G+12710  U+BCF7  uniBCF7,  HANGUL SYLLABLE-BCF7
        0xBCF9, # G+12711  U+BCF9  uniBCF9,  HANGUL SYLLABLE-BCF9
        0xBCFA, # G+12712  U+BCFA  uniBCFA,  HANGUL SYLLABLE-BCFA
        0xBCFB, # G+12713  U+BCFB  uniBCFB,  HANGUL SYLLABLE-BCFB
        0xBCFD, # G+12714  U+BCFD  uniBCFD,  HANGUL SYLLABLE-BCFD
        0xBCFE, # G+12715  U+BCFE  uniBCFE,  HANGUL SYLLABLE-BCFE
        0xBCFF, # G+12716  U+BCFF  uniBCFF,  HANGUL SYLLABLE-BCFF
        0xBD00, # G+12717  U+BD00  uniBD00,  HANGUL SYLLABLE-BD00
        0xBD01, # G+12718  U+BD01  uniBD01,  HANGUL SYLLABLE-BD01
        0xBD02, # G+12719  U+BD02  uniBD02,  HANGUL SYLLABLE-BD02
        0xBD03, # G+12720  U+BD03  uniBD03,  HANGUL SYLLABLE-BD03
        0xBD06, # G+12721  U+BD06  uniBD06,  HANGUL SYLLABLE-BD06
        0xBD08, # G+12722  U+BD08  uniBD08,  HANGUL SYLLABLE-BD08
        0xBD0A, # G+12723  U+BD0A  uniBD0A,  HANGUL SYLLABLE-BD0A
        0xBD0B, # G+12724  U+BD0B  uniBD0B,  HANGUL SYLLABLE-BD0B
        0xBD0C, # G+12725  U+BD0C  uniBD0C,  HANGUL SYLLABLE-BD0C
        0xBD0D, # G+12726  U+BD0D  uniBD0D,  HANGUL SYLLABLE-BD0D
        0xBD0E, # G+12727  U+BD0E  uniBD0E,  HANGUL SYLLABLE-BD0E
        0xBD0F, # G+12728  U+BD0F  uniBD0F,  HANGUL SYLLABLE-BD0F
        0xBD11, # G+12729  U+BD11  uniBD11,  HANGUL SYLLABLE-BD11
        0xBD12, # G+12730  U+BD12  uniBD12,  HANGUL SYLLABLE-BD12
        0xBD13, # G+12731  U+BD13  uniBD13,  HANGUL SYLLABLE-BD13
        0xBD15, # G+12732  U+BD15  uniBD15,  HANGUL SYLLABLE-BD15
        0xBD16, # G+12733  U+BD16  uniBD16,  HANGUL SYLLABLE-BD16
        0xBD17, # G+12734  U+BD17  uniBD17,  HANGUL SYLLABLE-BD17
        0xBD18, # G+12735  U+BD18  uniBD18,  HANGUL SYLLABLE-BD18
        0xBD19, # G+12736  U+BD19  uniBD19,  HANGUL SYLLABLE-BD19
        0xBD1A, # G+12737  U+BD1A  uniBD1A,  HANGUL SYLLABLE-BD1A
        0xBD1B, # G+12738  U+BD1B  uniBD1B,  HANGUL SYLLABLE-BD1B
        0xBD1C, # G+12739  U+BD1C  uniBD1C,  HANGUL SYLLABLE-BD1C
        0xBD1D, # G+12740  U+BD1D  uniBD1D,  HANGUL SYLLABLE-BD1D
        0xBD1E, # G+12741  U+BD1E  uniBD1E,  HANGUL SYLLABLE-BD1E
        0xBD1F, # G+12742  U+BD1F  uniBD1F,  HANGUL SYLLABLE-BD1F
        0xBD20, # G+12743  U+BD20  uniBD20,  HANGUL SYLLABLE-BD20
        0xBD21, # G+12744  U+BD21  uniBD21,  HANGUL SYLLABLE-BD21
        0xBD22, # G+12745  U+BD22  uniBD22,  HANGUL SYLLABLE-BD22
        0xBD23, # G+12746  U+BD23  uniBD23,  HANGUL SYLLABLE-BD23
        0xBD25, # G+12747  U+BD25  uniBD25,  HANGUL SYLLABLE-BD25
        0xBD26, # G+12748  U+BD26  uniBD26,  HANGUL SYLLABLE-BD26
        0xBD27, # G+12749  U+BD27  uniBD27,  HANGUL SYLLABLE-BD27
        0xBD28, # G+12750  U+BD28  uniBD28,  HANGUL SYLLABLE-BD28
        0xBD29, # G+12751  U+BD29  uniBD29,  HANGUL SYLLABLE-BD29
        0xBD2A, # G+12752  U+BD2A  uniBD2A,  HANGUL SYLLABLE-BD2A
        0xBD2B, # G+12753  U+BD2B  uniBD2B,  HANGUL SYLLABLE-BD2B
        0xBD2D, # G+12754  U+BD2D  uniBD2D,  HANGUL SYLLABLE-BD2D
        0xBD2E, # G+12755  U+BD2E  uniBD2E,  HANGUL SYLLABLE-BD2E
        0xBD2F, # G+12756  U+BD2F  uniBD2F,  HANGUL SYLLABLE-BD2F
        0xBD30, # G+12757  U+BD30  uniBD30,  HANGUL SYLLABLE-BD30
        0xBD31, # G+12758  U+BD31  uniBD31,  HANGUL SYLLABLE-BD31
        0xBD32, # G+12759  U+BD32  uniBD32,  HANGUL SYLLABLE-BD32
        0xBD33, # G+12760  U+BD33  uniBD33,  HANGUL SYLLABLE-BD33
        0xBD34, # G+12761  U+BD34  uniBD34,  HANGUL SYLLABLE-BD34
        0xBD35, # G+12762  U+BD35  uniBD35,  HANGUL SYLLABLE-BD35
        0xBD36, # G+12763  U+BD36  uniBD36,  HANGUL SYLLABLE-BD36
        0xBD37, # G+12764  U+BD37  uniBD37,  HANGUL SYLLABLE-BD37
        0xBD38, # G+12765  U+BD38  uniBD38,  HANGUL SYLLABLE-BD38
        0xBD39, # G+12766  U+BD39  uniBD39,  HANGUL SYLLABLE-BD39
        0xBD3A, # G+12767  U+BD3A  uniBD3A,  HANGUL SYLLABLE-BD3A
        0xBD3B, # G+12768  U+BD3B  uniBD3B,  HANGUL SYLLABLE-BD3B
        0xBD3C, # G+12769  U+BD3C  uniBD3C,  HANGUL SYLLABLE-BD3C
        0xBD3D, # G+12770  U+BD3D  uniBD3D,  HANGUL SYLLABLE-BD3D
        0xBD3E, # G+12771  U+BD3E  uniBD3E,  HANGUL SYLLABLE-BD3E
        0xBD3F, # G+12772  U+BD3F  uniBD3F,  HANGUL SYLLABLE-BD3F
        0xBD41, # G+12773  U+BD41  uniBD41,  HANGUL SYLLABLE-BD41
        0xBD42, # G+12774  U+BD42  uniBD42,  HANGUL SYLLABLE-BD42
        0xBD43, # G+12775  U+BD43  uniBD43,  HANGUL SYLLABLE-BD43
        0xBD44, # G+12776  U+BD44  uniBD44,  HANGUL SYLLABLE-BD44
        0xBD45, # G+12777  U+BD45  uniBD45,  HANGUL SYLLABLE-BD45
        0xBD46, # G+12778  U+BD46  uniBD46,  HANGUL SYLLABLE-BD46
        0xBD47, # G+12779  U+BD47  uniBD47,  HANGUL SYLLABLE-BD47
        0xBD4A, # G+12780  U+BD4A  uniBD4A,  HANGUL SYLLABLE-BD4A
        0xBD4B, # G+12781  U+BD4B  uniBD4B,  HANGUL SYLLABLE-BD4B
        0xBD4D, # G+12782  U+BD4D  uniBD4D,  HANGUL SYLLABLE-BD4D
        0xBD4E, # G+12783  U+BD4E  uniBD4E,  HANGUL SYLLABLE-BD4E
        0xBD4F, # G+12784  U+BD4F  uniBD4F,  HANGUL SYLLABLE-BD4F
        0xBD51, # G+12785  U+BD51  uniBD51,  HANGUL SYLLABLE-BD51
        0xBD52, # G+12786  U+BD52  uniBD52,  HANGUL SYLLABLE-BD52
        0xBD53, # G+12787  U+BD53  uniBD53,  HANGUL SYLLABLE-BD53
        0xBD54, # G+12788  U+BD54  uniBD54,  HANGUL SYLLABLE-BD54
        0xBD55, # G+12789  U+BD55  uniBD55,  HANGUL SYLLABLE-BD55
        0xBD56, # G+12790  U+BD56  uniBD56,  HANGUL SYLLABLE-BD56
        0xBD57, # G+12791  U+BD57  uniBD57,  HANGUL SYLLABLE-BD57
        0xBD5A, # G+12792  U+BD5A  uniBD5A,  HANGUL SYLLABLE-BD5A
        0xBD5B, # G+12793  U+BD5B  uniBD5B,  HANGUL SYLLABLE-BD5B
        0xBD5C, # G+12794  U+BD5C  uniBD5C,  HANGUL SYLLABLE-BD5C
        0xBD5D, # G+12795  U+BD5D  uniBD5D,  HANGUL SYLLABLE-BD5D
        0xBD5E, # G+12796  U+BD5E  uniBD5E,  HANGUL SYLLABLE-BD5E
        0xBD5F, # G+12797  U+BD5F  uniBD5F,  HANGUL SYLLABLE-BD5F
        0xBD60, # G+12798  U+BD60  uniBD60,  HANGUL SYLLABLE-BD60
        0xBD61, # G+12799  U+BD61  uniBD61,  HANGUL SYLLABLE-BD61
        0xBD62, # G+12800  U+BD62  uniBD62,  HANGUL SYLLABLE-BD62
        0xBD63, # G+12801  U+BD63  uniBD63,  HANGUL SYLLABLE-BD63
        0xBD65, # G+12802  U+BD65  uniBD65,  HANGUL SYLLABLE-BD65
        0xBD66, # G+12803  U+BD66  uniBD66,  HANGUL SYLLABLE-BD66
        0xBD67, # G+12804  U+BD67  uniBD67,  HANGUL SYLLABLE-BD67
        0xBD69, # G+12805  U+BD69  uniBD69,  HANGUL SYLLABLE-BD69
        0xBD6A, # G+12806  U+BD6A  uniBD6A,  HANGUL SYLLABLE-BD6A
        0xBD6B, # G+12807  U+BD6B  uniBD6B,  HANGUL SYLLABLE-BD6B
        0xBD6C, # G+12808  U+BD6C  uniBD6C,  HANGUL SYLLABLE-BD6C
        0xBD6D, # G+12809  U+BD6D  uniBD6D,  HANGUL SYLLABLE-BD6D
        0xBD6E, # G+12810  U+BD6E  uniBD6E,  HANGUL SYLLABLE-BD6E
        0xBD6F, # G+12811  U+BD6F  uniBD6F,  HANGUL SYLLABLE-BD6F
        0xBD70, # G+12812  U+BD70  uniBD70,  HANGUL SYLLABLE-BD70
        0xBD71, # G+12813  U+BD71  uniBD71,  HANGUL SYLLABLE-BD71
        0xBD72, # G+12814  U+BD72  uniBD72,  HANGUL SYLLABLE-BD72
        0xBD73, # G+12815  U+BD73  uniBD73,  HANGUL SYLLABLE-BD73
        0xBD74, # G+12816  U+BD74  uniBD74,  HANGUL SYLLABLE-BD74
        0xBD75, # G+12817  U+BD75  uniBD75,  HANGUL SYLLABLE-BD75
        0xBD76, # G+12818  U+BD76  uniBD76,  HANGUL SYLLABLE-BD76
        0xBD77, # G+12819  U+BD77  uniBD77,  HANGUL SYLLABLE-BD77
        0xBD78, # G+12820  U+BD78  uniBD78,  HANGUL SYLLABLE-BD78
        0xBD79, # G+12821  U+BD79  uniBD79,  HANGUL SYLLABLE-BD79
        0xBD7A, # G+12822  U+BD7A  uniBD7A,  HANGUL SYLLABLE-BD7A
        0xBD7B, # G+12823  U+BD7B  uniBD7B,  HANGUL SYLLABLE-BD7B
        0xBD7C, # G+12824  U+BD7C  uniBD7C,  HANGUL SYLLABLE-BD7C
        0xBD7D, # G+12825  U+BD7D  uniBD7D,  HANGUL SYLLABLE-BD7D
        0xBD7E, # G+12826  U+BD7E  uniBD7E,  HANGUL SYLLABLE-BD7E
        0xBD7F, # G+12827  U+BD7F  uniBD7F,  HANGUL SYLLABLE-BD7F
        0xBD82, # G+12828  U+BD82  uniBD82,  HANGUL SYLLABLE-BD82
        0xBD83, # G+12829  U+BD83  uniBD83,  HANGUL SYLLABLE-BD83
        0xBD85, # G+12830  U+BD85  uniBD85,  HANGUL SYLLABLE-BD85
        0xBD86, # G+12831  U+BD86  uniBD86,  HANGUL SYLLABLE-BD86
        0xBD8B, # G+12832  U+BD8B  uniBD8B,  HANGUL SYLLABLE-BD8B
        0xBD8C, # G+12833  U+BD8C  uniBD8C,  HANGUL SYLLABLE-BD8C
        0xBD8D, # G+12834  U+BD8D  uniBD8D,  HANGUL SYLLABLE-BD8D
        0xBD8E, # G+12835  U+BD8E  uniBD8E,  HANGUL SYLLABLE-BD8E
        0xBD8F, # G+12836  U+BD8F  uniBD8F,  HANGUL SYLLABLE-BD8F
        0xBD92, # G+12837  U+BD92  uniBD92,  HANGUL SYLLABLE-BD92
        0xBD94, # G+12838  U+BD94  uniBD94,  HANGUL SYLLABLE-BD94
        0xBD96, # G+12839  U+BD96  uniBD96,  HANGUL SYLLABLE-BD96
        0xBD97, # G+12840  U+BD97  uniBD97,  HANGUL SYLLABLE-BD97
        0xBD98, # G+12841  U+BD98  uniBD98,  HANGUL SYLLABLE-BD98
        0xBD9B, # G+12842  U+BD9B  uniBD9B,  HANGUL SYLLABLE-BD9B
        0xBD9D, # G+12843  U+BD9D  uniBD9D,  HANGUL SYLLABLE-BD9D
        0xBD9E, # G+12844  U+BD9E  uniBD9E,  HANGUL SYLLABLE-BD9E
        0xBD9F, # G+12845  U+BD9F  uniBD9F,  HANGUL SYLLABLE-BD9F
        0xBDA0, # G+12846  U+BDA0  uniBDA0,  HANGUL SYLLABLE-BDA0
        0xBDA1, # G+12847  U+BDA1  uniBDA1,  HANGUL SYLLABLE-BDA1
        0xBDA2, # G+12848  U+BDA2  uniBDA2,  HANGUL SYLLABLE-BDA2
        0xBDA3, # G+12849  U+BDA3  uniBDA3,  HANGUL SYLLABLE-BDA3
        0xBDA5, # G+12850  U+BDA5  uniBDA5,  HANGUL SYLLABLE-BDA5
        0xBDA6, # G+12851  U+BDA6  uniBDA6,  HANGUL SYLLABLE-BDA6
        0xBDA7, # G+12852  U+BDA7  uniBDA7,  HANGUL SYLLABLE-BDA7
        0xBDA8, # G+12853  U+BDA8  uniBDA8,  HANGUL SYLLABLE-BDA8
        0xBDA9, # G+12854  U+BDA9  uniBDA9,  HANGUL SYLLABLE-BDA9
        0xBDAA, # G+12855  U+BDAA  uniBDAA,  HANGUL SYLLABLE-BDAA
        0xBDAB, # G+12856  U+BDAB  uniBDAB,  HANGUL SYLLABLE-BDAB
        0xBDAC, # G+12857  U+BDAC  uniBDAC,  HANGUL SYLLABLE-BDAC
        0xBDAD, # G+12858  U+BDAD  uniBDAD,  HANGUL SYLLABLE-BDAD
        0xBDAE, # G+12859  U+BDAE  uniBDAE,  HANGUL SYLLABLE-BDAE
        0xBDAF, # G+12860  U+BDAF  uniBDAF,  HANGUL SYLLABLE-BDAF
        0xBDB1, # G+12861  U+BDB1  uniBDB1,  HANGUL SYLLABLE-BDB1
        0xBDB2, # G+12862  U+BDB2  uniBDB2,  HANGUL SYLLABLE-BDB2
        0xBDB3, # G+12863  U+BDB3  uniBDB3,  HANGUL SYLLABLE-BDB3
        0xBDB4, # G+12864  U+BDB4  uniBDB4,  HANGUL SYLLABLE-BDB4
        0xBDB5, # G+12865  U+BDB5  uniBDB5,  HANGUL SYLLABLE-BDB5
        0xBDB6, # G+12866  U+BDB6  uniBDB6,  HANGUL SYLLABLE-BDB6
        0xBDB7, # G+12867  U+BDB7  uniBDB7,  HANGUL SYLLABLE-BDB7
        0xBDB9, # G+12868  U+BDB9  uniBDB9,  HANGUL SYLLABLE-BDB9
        0xBDBA, # G+12869  U+BDBA  uniBDBA,  HANGUL SYLLABLE-BDBA
        0xBDBB, # G+12870  U+BDBB  uniBDBB,  HANGUL SYLLABLE-BDBB
        0xBDBC, # G+12871  U+BDBC  uniBDBC,  HANGUL SYLLABLE-BDBC
        0xBDBD, # G+12872  U+BDBD  uniBDBD,  HANGUL SYLLABLE-BDBD
        0xBDBE, # G+12873  U+BDBE  uniBDBE,  HANGUL SYLLABLE-BDBE
        0xBDBF, # G+12874  U+BDBF  uniBDBF,  HANGUL SYLLABLE-BDBF
        0xBDC0, # G+12875  U+BDC0  uniBDC0,  HANGUL SYLLABLE-BDC0
        0xBDC1, # G+12876  U+BDC1  uniBDC1,  HANGUL SYLLABLE-BDC1
        0xBDC2, # G+12877  U+BDC2  uniBDC2,  HANGUL SYLLABLE-BDC2
        0xBDC3, # G+12878  U+BDC3  uniBDC3,  HANGUL SYLLABLE-BDC3
        0xBDC4, # G+12879  U+BDC4  uniBDC4,  HANGUL SYLLABLE-BDC4
        0xBDC5, # G+12880  U+BDC5  uniBDC5,  HANGUL SYLLABLE-BDC5
        0xBDC6, # G+12881  U+BDC6  uniBDC6,  HANGUL SYLLABLE-BDC6
        0xBDC7, # G+12882  U+BDC7  uniBDC7,  HANGUL SYLLABLE-BDC7
        0xBDC8, # G+12883  U+BDC8  uniBDC8,  HANGUL SYLLABLE-BDC8
        0xBDC9, # G+12884  U+BDC9  uniBDC9,  HANGUL SYLLABLE-BDC9
        0xBDCA, # G+12885  U+BDCA  uniBDCA,  HANGUL SYLLABLE-BDCA
        0xBDCB, # G+12886  U+BDCB  uniBDCB,  HANGUL SYLLABLE-BDCB
        0xBDCC, # G+12887  U+BDCC  uniBDCC,  HANGUL SYLLABLE-BDCC
        0xBDCD, # G+12888  U+BDCD  uniBDCD,  HANGUL SYLLABLE-BDCD
        0xBDCE, # G+12889  U+BDCE  uniBDCE,  HANGUL SYLLABLE-BDCE
        0xBDCF, # G+12890  U+BDCF  uniBDCF,  HANGUL SYLLABLE-BDCF
        0xBDD0, # G+12891  U+BDD0  uniBDD0,  HANGUL SYLLABLE-BDD0
        0xBDD1, # G+12892  U+BDD1  uniBDD1,  HANGUL SYLLABLE-BDD1
        0xBDD2, # G+12893  U+BDD2  uniBDD2,  HANGUL SYLLABLE-BDD2
        0xBDD3, # G+12894  U+BDD3  uniBDD3,  HANGUL SYLLABLE-BDD3
        0xBDD6, # G+12895  U+BDD6  uniBDD6,  HANGUL SYLLABLE-BDD6
        0xBDD7, # G+12896  U+BDD7  uniBDD7,  HANGUL SYLLABLE-BDD7
        0xBDD9, # G+12897  U+BDD9  uniBDD9,  HANGUL SYLLABLE-BDD9
        0xBDDA, # G+12898  U+BDDA  uniBDDA,  HANGUL SYLLABLE-BDDA
        0xBDDB, # G+12899  U+BDDB  uniBDDB,  HANGUL SYLLABLE-BDDB
        0xBDDD, # G+12900  U+BDDD  uniBDDD,  HANGUL SYLLABLE-BDDD
        0xBDDE, # G+12901  U+BDDE  uniBDDE,  HANGUL SYLLABLE-BDDE
        0xBDDF, # G+12902  U+BDDF  uniBDDF,  HANGUL SYLLABLE-BDDF
        0xBDE0, # G+12903  U+BDE0  uniBDE0,  HANGUL SYLLABLE-BDE0
        0xBDE1, # G+12904  U+BDE1  uniBDE1,  HANGUL SYLLABLE-BDE1
        0xBDE2, # G+12905  U+BDE2  uniBDE2,  HANGUL SYLLABLE-BDE2
        0xBDE3, # G+12906  U+BDE3  uniBDE3,  HANGUL SYLLABLE-BDE3
        0xBDE4, # G+12907  U+BDE4  uniBDE4,  HANGUL SYLLABLE-BDE4
        0xBDE5, # G+12908  U+BDE5  uniBDE5,  HANGUL SYLLABLE-BDE5
        0xBDE6, # G+12909  U+BDE6  uniBDE6,  HANGUL SYLLABLE-BDE6
        0xBDE7, # G+12910  U+BDE7  uniBDE7,  HANGUL SYLLABLE-BDE7
        0xBDE8, # G+12911  U+BDE8  uniBDE8,  HANGUL SYLLABLE-BDE8
        0xBDEA, # G+12912  U+BDEA  uniBDEA,  HANGUL SYLLABLE-BDEA
        0xBDEB, # G+12913  U+BDEB  uniBDEB,  HANGUL SYLLABLE-BDEB
        0xBDEC, # G+12914  U+BDEC  uniBDEC,  HANGUL SYLLABLE-BDEC
        0xBDED, # G+12915  U+BDED  uniBDED,  HANGUL SYLLABLE-BDED
        0xBDEE, # G+12916  U+BDEE  uniBDEE,  HANGUL SYLLABLE-BDEE
        0xBDEF, # G+12917  U+BDEF  uniBDEF,  HANGUL SYLLABLE-BDEF
        0xBDF1, # G+12918  U+BDF1  uniBDF1,  HANGUL SYLLABLE-BDF1
        0xBDF2, # G+12919  U+BDF2  uniBDF2,  HANGUL SYLLABLE-BDF2
        0xBDF3, # G+12920  U+BDF3  uniBDF3,  HANGUL SYLLABLE-BDF3
        0xBDF5, # G+12921  U+BDF5  uniBDF5,  HANGUL SYLLABLE-BDF5
        0xBDF6, # G+12922  U+BDF6  uniBDF6,  HANGUL SYLLABLE-BDF6
        0xBDF7, # G+12923  U+BDF7  uniBDF7,  HANGUL SYLLABLE-BDF7
        0xBDF9, # G+12924  U+BDF9  uniBDF9,  HANGUL SYLLABLE-BDF9
        0xBDFA, # G+12925  U+BDFA  uniBDFA,  HANGUL SYLLABLE-BDFA
        0xBDFB, # G+12926  U+BDFB  uniBDFB,  HANGUL SYLLABLE-BDFB
        0xBDFC, # G+12927  U+BDFC  uniBDFC,  HANGUL SYLLABLE-BDFC
        0xBDFD, # G+12928  U+BDFD  uniBDFD,  HANGUL SYLLABLE-BDFD
        0xBDFE, # G+12929  U+BDFE  uniBDFE,  HANGUL SYLLABLE-BDFE
        0xBDFF, # G+12930  U+BDFF  uniBDFF,  HANGUL SYLLABLE-BDFF
        0xBE01, # G+12931  U+BE01  uniBE01,  HANGUL SYLLABLE-BE01
        0xBE02, # G+12932  U+BE02  uniBE02,  HANGUL SYLLABLE-BE02
        0xBE04, # G+12933  U+BE04  uniBE04,  HANGUL SYLLABLE-BE04
        0xBE06, # G+12934  U+BE06  uniBE06,  HANGUL SYLLABLE-BE06
        0xBE07, # G+12935  U+BE07  uniBE07,  HANGUL SYLLABLE-BE07
        0xBE08, # G+12936  U+BE08  uniBE08,  HANGUL SYLLABLE-BE08
        0xBE09, # G+12937  U+BE09  uniBE09,  HANGUL SYLLABLE-BE09
        0xBE0A, # G+12938  U+BE0A  uniBE0A,  HANGUL SYLLABLE-BE0A
        0xBE0B, # G+12939  U+BE0B  uniBE0B,  HANGUL SYLLABLE-BE0B
        0xBE0E, # G+12940  U+BE0E  uniBE0E,  HANGUL SYLLABLE-BE0E
        0xBE0F, # G+12941  U+BE0F  uniBE0F,  HANGUL SYLLABLE-BE0F
        0xBE11, # G+12942  U+BE11  uniBE11,  HANGUL SYLLABLE-BE11
        0xBE12, # G+12943  U+BE12  uniBE12,  HANGUL SYLLABLE-BE12
        0xBE13, # G+12944  U+BE13  uniBE13,  HANGUL SYLLABLE-BE13
        0xBE15, # G+12945  U+BE15  uniBE15,  HANGUL SYLLABLE-BE15
        0xBE16, # G+12946  U+BE16  uniBE16,  HANGUL SYLLABLE-BE16
        0xBE17, # G+12947  U+BE17  uniBE17,  HANGUL SYLLABLE-BE17
        0xBE18, # G+12948  U+BE18  uniBE18,  HANGUL SYLLABLE-BE18
        0xBE19, # G+12949  U+BE19  uniBE19,  HANGUL SYLLABLE-BE19
        0xBE1A, # G+12950  U+BE1A  uniBE1A,  HANGUL SYLLABLE-BE1A
        0xBE1B, # G+12951  U+BE1B  uniBE1B,  HANGUL SYLLABLE-BE1B
        0xBE1E, # G+12952  U+BE1E  uniBE1E,  HANGUL SYLLABLE-BE1E
        0xBE20, # G+12953  U+BE20  uniBE20,  HANGUL SYLLABLE-BE20
        0xBE21, # G+12954  U+BE21  uniBE21,  HANGUL SYLLABLE-BE21
        0xBE22, # G+12955  U+BE22  uniBE22,  HANGUL SYLLABLE-BE22
        0xBE23, # G+12956  U+BE23  uniBE23,  HANGUL SYLLABLE-BE23
        0xBE24, # G+12957  U+BE24  uniBE24,  HANGUL SYLLABLE-BE24
        0xBE25, # G+12958  U+BE25  uniBE25,  HANGUL SYLLABLE-BE25
        0xBE26, # G+12959  U+BE26  uniBE26,  HANGUL SYLLABLE-BE26
        0xBE27, # G+12960  U+BE27  uniBE27,  HANGUL SYLLABLE-BE27
        0xBE28, # G+12961  U+BE28  uniBE28,  HANGUL SYLLABLE-BE28
        0xBE29, # G+12962  U+BE29  uniBE29,  HANGUL SYLLABLE-BE29
        0xBE2A, # G+12963  U+BE2A  uniBE2A,  HANGUL SYLLABLE-BE2A
        0xBE2B, # G+12964  U+BE2B  uniBE2B,  HANGUL SYLLABLE-BE2B
        0xBE2C, # G+12965  U+BE2C  uniBE2C,  HANGUL SYLLABLE-BE2C
        0xBE2D, # G+12966  U+BE2D  uniBE2D,  HANGUL SYLLABLE-BE2D
        0xBE2E, # G+12967  U+BE2E  uniBE2E,  HANGUL SYLLABLE-BE2E
        0xBE2F, # G+12968  U+BE2F  uniBE2F,  HANGUL SYLLABLE-BE2F
        0xBE30, # G+12969  U+BE30  uniBE30,  HANGUL SYLLABLE-BE30
        0xBE31, # G+12970  U+BE31  uniBE31,  HANGUL SYLLABLE-BE31
        0xBE32, # G+12971  U+BE32  uniBE32,  HANGUL SYLLABLE-BE32
        0xBE33, # G+12972  U+BE33  uniBE33,  HANGUL SYLLABLE-BE33
        0xBE34, # G+12973  U+BE34  uniBE34,  HANGUL SYLLABLE-BE34
        0xBE35, # G+12974  U+BE35  uniBE35,  HANGUL SYLLABLE-BE35
        0xBE36, # G+12975  U+BE36  uniBE36,  HANGUL SYLLABLE-BE36
        0xBE37, # G+12976  U+BE37  uniBE37,  HANGUL SYLLABLE-BE37
        0xBE38, # G+12977  U+BE38  uniBE38,  HANGUL SYLLABLE-BE38
        0xBE39, # G+12978  U+BE39  uniBE39,  HANGUL SYLLABLE-BE39
        0xBE3A, # G+12979  U+BE3A  uniBE3A,  HANGUL SYLLABLE-BE3A
        0xBE3B, # G+12980  U+BE3B  uniBE3B,  HANGUL SYLLABLE-BE3B
        0xBE3C, # G+12981  U+BE3C  uniBE3C,  HANGUL SYLLABLE-BE3C
        0xBE3D, # G+12982  U+BE3D  uniBE3D,  HANGUL SYLLABLE-BE3D
        0xBE3E, # G+12983  U+BE3E  uniBE3E,  HANGUL SYLLABLE-BE3E
        0xBE3F, # G+12984  U+BE3F  uniBE3F,  HANGUL SYLLABLE-BE3F
        0xBE40, # G+12985  U+BE40  uniBE40,  HANGUL SYLLABLE-BE40
        0xBE41, # G+12986  U+BE41  uniBE41,  HANGUL SYLLABLE-BE41
        0xBE42, # G+12987  U+BE42  uniBE42,  HANGUL SYLLABLE-BE42
        0xBE43, # G+12988  U+BE43  uniBE43,  HANGUL SYLLABLE-BE43
        0xBE46, # G+12989  U+BE46  uniBE46,  HANGUL SYLLABLE-BE46
        0xBE47, # G+12990  U+BE47  uniBE47,  HANGUL SYLLABLE-BE47
        0xBE49, # G+12991  U+BE49  uniBE49,  HANGUL SYLLABLE-BE49
        0xBE4A, # G+12992  U+BE4A  uniBE4A,  HANGUL SYLLABLE-BE4A
        0xBE4B, # G+12993  U+BE4B  uniBE4B,  HANGUL SYLLABLE-BE4B
        0xBE4D, # G+12994  U+BE4D  uniBE4D,  HANGUL SYLLABLE-BE4D
        0xBE4F, # G+12995  U+BE4F  uniBE4F,  HANGUL SYLLABLE-BE4F
        0xBE50, # G+12996  U+BE50  uniBE50,  HANGUL SYLLABLE-BE50
        0xBE51, # G+12997  U+BE51  uniBE51,  HANGUL SYLLABLE-BE51
        0xBE52, # G+12998  U+BE52  uniBE52,  HANGUL SYLLABLE-BE52
        0xBE53, # G+12999  U+BE53  uniBE53,  HANGUL SYLLABLE-BE53
        0xBE56, # G+13000  U+BE56  uniBE56,  HANGUL SYLLABLE-BE56
        0xBE58, # G+13001  U+BE58  uniBE58,  HANGUL SYLLABLE-BE58
        0xBE5C, # G+13002  U+BE5C  uniBE5C,  HANGUL SYLLABLE-BE5C
        0xBE5D, # G+13003  U+BE5D  uniBE5D,  HANGUL SYLLABLE-BE5D
        0xBE5E, # G+13004  U+BE5E  uniBE5E,  HANGUL SYLLABLE-BE5E
        0xBE5F, # G+13005  U+BE5F  uniBE5F,  HANGUL SYLLABLE-BE5F
        0xBE62, # G+13006  U+BE62  uniBE62,  HANGUL SYLLABLE-BE62
        0xBE63, # G+13007  U+BE63  uniBE63,  HANGUL SYLLABLE-BE63
        0xBE65, # G+13008  U+BE65  uniBE65,  HANGUL SYLLABLE-BE65
        0xBE66, # G+13009  U+BE66  uniBE66,  HANGUL SYLLABLE-BE66
        0xBE67, # G+13010  U+BE67  uniBE67,  HANGUL SYLLABLE-BE67
        0xBE69, # G+13011  U+BE69  uniBE69,  HANGUL SYLLABLE-BE69
        0xBE6B, # G+13012  U+BE6B  uniBE6B,  HANGUL SYLLABLE-BE6B
        0xBE6C, # G+13013  U+BE6C  uniBE6C,  HANGUL SYLLABLE-BE6C
        0xBE6D, # G+13014  U+BE6D  uniBE6D,  HANGUL SYLLABLE-BE6D
        0xBE6E, # G+13015  U+BE6E  uniBE6E,  HANGUL SYLLABLE-BE6E
        0xBE6F, # G+13016  U+BE6F  uniBE6F,  HANGUL SYLLABLE-BE6F
        0xBE72, # G+13017  U+BE72  uniBE72,  HANGUL SYLLABLE-BE72
        0xBE76, # G+13018  U+BE76  uniBE76,  HANGUL SYLLABLE-BE76
        0xBE77, # G+13019  U+BE77  uniBE77,  HANGUL SYLLABLE-BE77
        0xBE78, # G+13020  U+BE78  uniBE78,  HANGUL SYLLABLE-BE78
        0xBE79, # G+13021  U+BE79  uniBE79,  HANGUL SYLLABLE-BE79
        0xBE7A, # G+13022  U+BE7A  uniBE7A,  HANGUL SYLLABLE-BE7A
        0xBE7E, # G+13023  U+BE7E  uniBE7E,  HANGUL SYLLABLE-BE7E
        0xBE7F, # G+13024  U+BE7F  uniBE7F,  HANGUL SYLLABLE-BE7F
        0xBE81, # G+13025  U+BE81  uniBE81,  HANGUL SYLLABLE-BE81
        0xBE82, # G+13026  U+BE82  uniBE82,  HANGUL SYLLABLE-BE82
        0xBE83, # G+13027  U+BE83  uniBE83,  HANGUL SYLLABLE-BE83
        0xBE85, # G+13028  U+BE85  uniBE85,  HANGUL SYLLABLE-BE85
        0xBE86, # G+13029  U+BE86  uniBE86,  HANGUL SYLLABLE-BE86
        0xBE87, # G+13030  U+BE87  uniBE87,  HANGUL SYLLABLE-BE87
        0xBE88, # G+13031  U+BE88  uniBE88,  HANGUL SYLLABLE-BE88
        0xBE89, # G+13032  U+BE89  uniBE89,  HANGUL SYLLABLE-BE89
        0xBE8A, # G+13033  U+BE8A  uniBE8A,  HANGUL SYLLABLE-BE8A
        0xBE8B, # G+13034  U+BE8B  uniBE8B,  HANGUL SYLLABLE-BE8B
        0xBE8E, # G+13035  U+BE8E  uniBE8E,  HANGUL SYLLABLE-BE8E
        0xBE92, # G+13036  U+BE92  uniBE92,  HANGUL SYLLABLE-BE92
        0xBE93, # G+13037  U+BE93  uniBE93,  HANGUL SYLLABLE-BE93
        0xBE94, # G+13038  U+BE94  uniBE94,  HANGUL SYLLABLE-BE94
        0xBE95, # G+13039  U+BE95  uniBE95,  HANGUL SYLLABLE-BE95
        0xBE96, # G+13040  U+BE96  uniBE96,  HANGUL SYLLABLE-BE96
        0xBE97, # G+13041  U+BE97  uniBE97,  HANGUL SYLLABLE-BE97
        0xBE9A, # G+13042  U+BE9A  uniBE9A,  HANGUL SYLLABLE-BE9A
        0xBE9B, # G+13043  U+BE9B  uniBE9B,  HANGUL SYLLABLE-BE9B
        0xBE9C, # G+13044  U+BE9C  uniBE9C,  HANGUL SYLLABLE-BE9C
        0xBE9D, # G+13045  U+BE9D  uniBE9D,  HANGUL SYLLABLE-BE9D
        0xBE9E, # G+13046  U+BE9E  uniBE9E,  HANGUL SYLLABLE-BE9E
        0xBE9F, # G+13047  U+BE9F  uniBE9F,  HANGUL SYLLABLE-BE9F
        0xBEA0, # G+13048  U+BEA0  uniBEA0,  HANGUL SYLLABLE-BEA0
        0xBEA1, # G+13049  U+BEA1  uniBEA1,  HANGUL SYLLABLE-BEA1
        0xBEA2, # G+13050  U+BEA2  uniBEA2,  HANGUL SYLLABLE-BEA2
        0xBEA3, # G+13051  U+BEA3  uniBEA3,  HANGUL SYLLABLE-BEA3
        0xBEA4, # G+13052  U+BEA4  uniBEA4,  HANGUL SYLLABLE-BEA4
        0xBEA5, # G+13053  U+BEA5  uniBEA5,  HANGUL SYLLABLE-BEA5
        0xBEA6, # G+13054  U+BEA6  uniBEA6,  HANGUL SYLLABLE-BEA6
        0xBEA7, # G+13055  U+BEA7  uniBEA7,  HANGUL SYLLABLE-BEA7
        0xBEA9, # G+13056  U+BEA9  uniBEA9,  HANGUL SYLLABLE-BEA9
        0xBEAA, # G+13057  U+BEAA  uniBEAA,  HANGUL SYLLABLE-BEAA
        0xBEAB, # G+13058  U+BEAB  uniBEAB,  HANGUL SYLLABLE-BEAB
        0xBEAC, # G+13059  U+BEAC  uniBEAC,  HANGUL SYLLABLE-BEAC
        0xBEAD, # G+13060  U+BEAD  uniBEAD,  HANGUL SYLLABLE-BEAD
        0xBEAE, # G+13061  U+BEAE  uniBEAE,  HANGUL SYLLABLE-BEAE
        0xBEAF, # G+13062  U+BEAF  uniBEAF,  HANGUL SYLLABLE-BEAF
        0xBEB0, # G+13063  U+BEB0  uniBEB0,  HANGUL SYLLABLE-BEB0
        0xBEB1, # G+13064  U+BEB1  uniBEB1,  HANGUL SYLLABLE-BEB1
        0xBEB2, # G+13065  U+BEB2  uniBEB2,  HANGUL SYLLABLE-BEB2
        0xBEB3, # G+13066  U+BEB3  uniBEB3,  HANGUL SYLLABLE-BEB3
        0xBEB4, # G+13067  U+BEB4  uniBEB4,  HANGUL SYLLABLE-BEB4
        0xBEB5, # G+13068  U+BEB5  uniBEB5,  HANGUL SYLLABLE-BEB5
        0xBEB6, # G+13069  U+BEB6  uniBEB6,  HANGUL SYLLABLE-BEB6
        0xBEB7, # G+13070  U+BEB7  uniBEB7,  HANGUL SYLLABLE-BEB7
        0xBEB8, # G+13071  U+BEB8  uniBEB8,  HANGUL SYLLABLE-BEB8
        0xBEB9, # G+13072  U+BEB9  uniBEB9,  HANGUL SYLLABLE-BEB9
        0xBEBA, # G+13073  U+BEBA  uniBEBA,  HANGUL SYLLABLE-BEBA
        0xBEBB, # G+13074  U+BEBB  uniBEBB,  HANGUL SYLLABLE-BEBB
        0xBEBC, # G+13075  U+BEBC  uniBEBC,  HANGUL SYLLABLE-BEBC
        0xBEBD, # G+13076  U+BEBD  uniBEBD,  HANGUL SYLLABLE-BEBD
        0xBEBE, # G+13077  U+BEBE  uniBEBE,  HANGUL SYLLABLE-BEBE
        0xBEBF, # G+13078  U+BEBF  uniBEBF,  HANGUL SYLLABLE-BEBF
        0xBEC0, # G+13079  U+BEC0  uniBEC0,  HANGUL SYLLABLE-BEC0
        0xBEC1, # G+13080  U+BEC1  uniBEC1,  HANGUL SYLLABLE-BEC1
        0xBEC2, # G+13081  U+BEC2  uniBEC2,  HANGUL SYLLABLE-BEC2
        0xBEC3, # G+13082  U+BEC3  uniBEC3,  HANGUL SYLLABLE-BEC3
        0xBEC4, # G+13083  U+BEC4  uniBEC4,  HANGUL SYLLABLE-BEC4
        0xBEC5, # G+13084  U+BEC5  uniBEC5,  HANGUL SYLLABLE-BEC5
        0xBEC6, # G+13085  U+BEC6  uniBEC6,  HANGUL SYLLABLE-BEC6
        0xBEC7, # G+13086  U+BEC7  uniBEC7,  HANGUL SYLLABLE-BEC7
        0xBEC8, # G+13087  U+BEC8  uniBEC8,  HANGUL SYLLABLE-BEC8
        0xBEC9, # G+13088  U+BEC9  uniBEC9,  HANGUL SYLLABLE-BEC9
        0xBECA, # G+13089  U+BECA  uniBECA,  HANGUL SYLLABLE-BECA
        0xBECB, # G+13090  U+BECB  uniBECB,  HANGUL SYLLABLE-BECB
        0xBECC, # G+13091  U+BECC  uniBECC,  HANGUL SYLLABLE-BECC
        0xBECD, # G+13092  U+BECD  uniBECD,  HANGUL SYLLABLE-BECD
        0xBECE, # G+13093  U+BECE  uniBECE,  HANGUL SYLLABLE-BECE
        0xBECF, # G+13094  U+BECF  uniBECF,  HANGUL SYLLABLE-BECF
        0xBED2, # G+13095  U+BED2  uniBED2,  HANGUL SYLLABLE-BED2
        0xBED3, # G+13096  U+BED3  uniBED3,  HANGUL SYLLABLE-BED3
        0xBED5, # G+13097  U+BED5  uniBED5,  HANGUL SYLLABLE-BED5
        0xBED6, # G+13098  U+BED6  uniBED6,  HANGUL SYLLABLE-BED6
        0xBED9, # G+13099  U+BED9  uniBED9,  HANGUL SYLLABLE-BED9
        0xBEDA, # G+13100  U+BEDA  uniBEDA,  HANGUL SYLLABLE-BEDA
        0xBEDB, # G+13101  U+BEDB  uniBEDB,  HANGUL SYLLABLE-BEDB
        0xBEDC, # G+13102  U+BEDC  uniBEDC,  HANGUL SYLLABLE-BEDC
        0xBEDD, # G+13103  U+BEDD  uniBEDD,  HANGUL SYLLABLE-BEDD
        0xBEDE, # G+13104  U+BEDE  uniBEDE,  HANGUL SYLLABLE-BEDE
        0xBEDF, # G+13105  U+BEDF  uniBEDF,  HANGUL SYLLABLE-BEDF
        0xBEE1, # G+13106  U+BEE1  uniBEE1,  HANGUL SYLLABLE-BEE1
        0xBEE2, # G+13107  U+BEE2  uniBEE2,  HANGUL SYLLABLE-BEE2
        0xBEE6, # G+13108  U+BEE6  uniBEE6,  HANGUL SYLLABLE-BEE6
        0xBEE7, # G+13109  U+BEE7  uniBEE7,  HANGUL SYLLABLE-BEE7
        0xBEE8, # G+13110  U+BEE8  uniBEE8,  HANGUL SYLLABLE-BEE8
        0xBEE9, # G+13111  U+BEE9  uniBEE9,  HANGUL SYLLABLE-BEE9
        0xBEEA, # G+13112  U+BEEA  uniBEEA,  HANGUL SYLLABLE-BEEA
        0xBEEB, # G+13113  U+BEEB  uniBEEB,  HANGUL SYLLABLE-BEEB
        0xBEED, # G+13114  U+BEED  uniBEED,  HANGUL SYLLABLE-BEED
        0xBEEE, # G+13115  U+BEEE  uniBEEE,  HANGUL SYLLABLE-BEEE
        0xBEEF, # G+13116  U+BEEF  uniBEEF,  HANGUL SYLLABLE-BEEF
        0xBEF0, # G+13117  U+BEF0  uniBEF0,  HANGUL SYLLABLE-BEF0
        0xBEF1, # G+13118  U+BEF1  uniBEF1,  HANGUL SYLLABLE-BEF1
        0xBEF2, # G+13119  U+BEF2  uniBEF2,  HANGUL SYLLABLE-BEF2
        0xBEF3, # G+13120  U+BEF3  uniBEF3,  HANGUL SYLLABLE-BEF3
        0xBEF4, # G+13121  U+BEF4  uniBEF4,  HANGUL SYLLABLE-BEF4
        0xBEF5, # G+13122  U+BEF5  uniBEF5,  HANGUL SYLLABLE-BEF5
        0xBEF6, # G+13123  U+BEF6  uniBEF6,  HANGUL SYLLABLE-BEF6
        0xBEF7, # G+13124  U+BEF7  uniBEF7,  HANGUL SYLLABLE-BEF7
        0xBEF8, # G+13125  U+BEF8  uniBEF8,  HANGUL SYLLABLE-BEF8
        0xBEF9, # G+13126  U+BEF9  uniBEF9,  HANGUL SYLLABLE-BEF9
        0xBEFA, # G+13127  U+BEFA  uniBEFA,  HANGUL SYLLABLE-BEFA
        0xBEFB, # G+13128  U+BEFB  uniBEFB,  HANGUL SYLLABLE-BEFB
        0xBEFC, # G+13129  U+BEFC  uniBEFC,  HANGUL SYLLABLE-BEFC
        0xBEFD, # G+13130  U+BEFD  uniBEFD,  HANGUL SYLLABLE-BEFD
        0xBEFE, # G+13131  U+BEFE  uniBEFE,  HANGUL SYLLABLE-BEFE
        0xBEFF, # G+13132  U+BEFF  uniBEFF,  HANGUL SYLLABLE-BEFF
        0xBF00, # G+13133  U+BF00  uniBF00,  HANGUL SYLLABLE-BF00
        0xBF02, # G+13134  U+BF02  uniBF02,  HANGUL SYLLABLE-BF02
        0xBF03, # G+13135  U+BF03  uniBF03,  HANGUL SYLLABLE-BF03
        0xBF04, # G+13136  U+BF04  uniBF04,  HANGUL SYLLABLE-BF04
        0xBF05, # G+13137  U+BF05  uniBF05,  HANGUL SYLLABLE-BF05
        0xBF06, # G+13138  U+BF06  uniBF06,  HANGUL SYLLABLE-BF06
        0xBF07, # G+13139  U+BF07  uniBF07,  HANGUL SYLLABLE-BF07
        0xBF0A, # G+13140  U+BF0A  uniBF0A,  HANGUL SYLLABLE-BF0A
        0xBF0B, # G+13141  U+BF0B  uniBF0B,  HANGUL SYLLABLE-BF0B
        0xBF0C, # G+13142  U+BF0C  uniBF0C,  HANGUL SYLLABLE-BF0C
        0xBF0D, # G+13143  U+BF0D  uniBF0D,  HANGUL SYLLABLE-BF0D
        0xBF0E, # G+13144  U+BF0E  uniBF0E,  HANGUL SYLLABLE-BF0E
        0xBF0F, # G+13145  U+BF0F  uniBF0F,  HANGUL SYLLABLE-BF0F
        0xBF10, # G+13146  U+BF10  uniBF10,  HANGUL SYLLABLE-BF10
        0xBF11, # G+13147  U+BF11  uniBF11,  HANGUL SYLLABLE-BF11
        0xBF12, # G+13148  U+BF12  uniBF12,  HANGUL SYLLABLE-BF12
        0xBF13, # G+13149  U+BF13  uniBF13,  HANGUL SYLLABLE-BF13
        0xBF14, # G+13150  U+BF14  uniBF14,  HANGUL SYLLABLE-BF14
        0xBF15, # G+13151  U+BF15  uniBF15,  HANGUL SYLLABLE-BF15
        0xBF16, # G+13152  U+BF16  uniBF16,  HANGUL SYLLABLE-BF16
        0xBF17, # G+13153  U+BF17  uniBF17,  HANGUL SYLLABLE-BF17
        0xBF1A, # G+13154  U+BF1A  uniBF1A,  HANGUL SYLLABLE-BF1A
        0xBF1E, # G+13155  U+BF1E  uniBF1E,  HANGUL SYLLABLE-BF1E
        0xBF1F, # G+13156  U+BF1F  uniBF1F,  HANGUL SYLLABLE-BF1F
        0xBF20, # G+13157  U+BF20  uniBF20,  HANGUL SYLLABLE-BF20
        0xBF21, # G+13158  U+BF21  uniBF21,  HANGUL SYLLABLE-BF21
        0xBF22, # G+13159  U+BF22  uniBF22,  HANGUL SYLLABLE-BF22
        0xBF23, # G+13160  U+BF23  uniBF23,  HANGUL SYLLABLE-BF23
        0xBF24, # G+13161  U+BF24  uniBF24,  HANGUL SYLLABLE-BF24
        0xBF25, # G+13162  U+BF25  uniBF25,  HANGUL SYLLABLE-BF25
        0xBF26, # G+13163  U+BF26  uniBF26,  HANGUL SYLLABLE-BF26
        0xBF27, # G+13164  U+BF27  uniBF27,  HANGUL SYLLABLE-BF27
        0xBF28, # G+13165  U+BF28  uniBF28,  HANGUL SYLLABLE-BF28
        0xBF29, # G+13166  U+BF29  uniBF29,  HANGUL SYLLABLE-BF29
        0xBF2A, # G+13167  U+BF2A  uniBF2A,  HANGUL SYLLABLE-BF2A
        0xBF2B, # G+13168  U+BF2B  uniBF2B,  HANGUL SYLLABLE-BF2B
        0xBF2C, # G+13169  U+BF2C  uniBF2C,  HANGUL SYLLABLE-BF2C
        0xBF2D, # G+13170  U+BF2D  uniBF2D,  HANGUL SYLLABLE-BF2D
        0xBF2E, # G+13171  U+BF2E  uniBF2E,  HANGUL SYLLABLE-BF2E
        0xBF2F, # G+13172  U+BF2F  uniBF2F,  HANGUL SYLLABLE-BF2F
        0xBF30, # G+13173  U+BF30  uniBF30,  HANGUL SYLLABLE-BF30
        0xBF31, # G+13174  U+BF31  uniBF31,  HANGUL SYLLABLE-BF31
        0xBF32, # G+13175  U+BF32  uniBF32,  HANGUL SYLLABLE-BF32
        0xBF33, # G+13176  U+BF33  uniBF33,  HANGUL SYLLABLE-BF33
        0xBF34, # G+13177  U+BF34  uniBF34,  HANGUL SYLLABLE-BF34
        0xBF35, # G+13178  U+BF35  uniBF35,  HANGUL SYLLABLE-BF35
        0xBF36, # G+13179  U+BF36  uniBF36,  HANGUL SYLLABLE-BF36
        0xBF37, # G+13180  U+BF37  uniBF37,  HANGUL SYLLABLE-BF37
        0xBF38, # G+13181  U+BF38  uniBF38,  HANGUL SYLLABLE-BF38
        0xBF39, # G+13182  U+BF39  uniBF39,  HANGUL SYLLABLE-BF39
        0xBF3A, # G+13183  U+BF3A  uniBF3A,  HANGUL SYLLABLE-BF3A
        0xBF3B, # G+13184  U+BF3B  uniBF3B,  HANGUL SYLLABLE-BF3B
        0xBF3C, # G+13185  U+BF3C  uniBF3C,  HANGUL SYLLABLE-BF3C
        0xBF3D, # G+13186  U+BF3D  uniBF3D,  HANGUL SYLLABLE-BF3D
        0xBF3E, # G+13187  U+BF3E  uniBF3E,  HANGUL SYLLABLE-BF3E
        0xBF3F, # G+13188  U+BF3F  uniBF3F,  HANGUL SYLLABLE-BF3F
        0xBF42, # G+13189  U+BF42  uniBF42,  HANGUL SYLLABLE-BF42
        0xBF43, # G+13190  U+BF43  uniBF43,  HANGUL SYLLABLE-BF43
        0xBF45, # G+13191  U+BF45  uniBF45,  HANGUL SYLLABLE-BF45
        0xBF46, # G+13192  U+BF46  uniBF46,  HANGUL SYLLABLE-BF46
        0xBF47, # G+13193  U+BF47  uniBF47,  HANGUL SYLLABLE-BF47
        0xBF49, # G+13194  U+BF49  uniBF49,  HANGUL SYLLABLE-BF49
        0xBF4A, # G+13195  U+BF4A  uniBF4A,  HANGUL SYLLABLE-BF4A
        0xBF4B, # G+13196  U+BF4B  uniBF4B,  HANGUL SYLLABLE-BF4B
        0xBF4C, # G+13197  U+BF4C  uniBF4C,  HANGUL SYLLABLE-BF4C
        0xBF4D, # G+13198  U+BF4D  uniBF4D,  HANGUL SYLLABLE-BF4D
        0xBF4E, # G+13199  U+BF4E  uniBF4E,  HANGUL SYLLABLE-BF4E
        0xBF4F, # G+13200  U+BF4F  uniBF4F,  HANGUL SYLLABLE-BF4F
        0xBF52, # G+13201  U+BF52  uniBF52,  HANGUL SYLLABLE-BF52
        0xBF53, # G+13202  U+BF53  uniBF53,  HANGUL SYLLABLE-BF53
        0xBF54, # G+13203  U+BF54  uniBF54,  HANGUL SYLLABLE-BF54
        0xBF56, # G+13204  U+BF56  uniBF56,  HANGUL SYLLABLE-BF56
        0xBF57, # G+13205  U+BF57  uniBF57,  HANGUL SYLLABLE-BF57
        0xBF58, # G+13206  U+BF58  uniBF58,  HANGUL SYLLABLE-BF58
        0xBF59, # G+13207  U+BF59  uniBF59,  HANGUL SYLLABLE-BF59
        0xBF5A, # G+13208  U+BF5A  uniBF5A,  HANGUL SYLLABLE-BF5A
        0xBF5B, # G+13209  U+BF5B  uniBF5B,  HANGUL SYLLABLE-BF5B
        0xBF5C, # G+13210  U+BF5C  uniBF5C,  HANGUL SYLLABLE-BF5C
        0xBF5D, # G+13211  U+BF5D  uniBF5D,  HANGUL SYLLABLE-BF5D
        0xBF5E, # G+13212  U+BF5E  uniBF5E,  HANGUL SYLLABLE-BF5E
        0xBF5F, # G+13213  U+BF5F  uniBF5F,  HANGUL SYLLABLE-BF5F
        0xBF60, # G+13214  U+BF60  uniBF60,  HANGUL SYLLABLE-BF60
        0xBF61, # G+13215  U+BF61  uniBF61,  HANGUL SYLLABLE-BF61
        0xBF62, # G+13216  U+BF62  uniBF62,  HANGUL SYLLABLE-BF62
        0xBF63, # G+13217  U+BF63  uniBF63,  HANGUL SYLLABLE-BF63
        0xBF64, # G+13218  U+BF64  uniBF64,  HANGUL SYLLABLE-BF64
        0xBF65, # G+13219  U+BF65  uniBF65,  HANGUL SYLLABLE-BF65
        0xBF66, # G+13220  U+BF66  uniBF66,  HANGUL SYLLABLE-BF66
        0xBF67, # G+13221  U+BF67  uniBF67,  HANGUL SYLLABLE-BF67
        0xBF68, # G+13222  U+BF68  uniBF68,  HANGUL SYLLABLE-BF68
        0xBF69, # G+13223  U+BF69  uniBF69,  HANGUL SYLLABLE-BF69
        0xBF6A, # G+13224  U+BF6A  uniBF6A,  HANGUL SYLLABLE-BF6A
        0xBF6B, # G+13225  U+BF6B  uniBF6B,  HANGUL SYLLABLE-BF6B
        0xBF6C, # G+13226  U+BF6C  uniBF6C,  HANGUL SYLLABLE-BF6C
        0xBF6D, # G+13227  U+BF6D  uniBF6D,  HANGUL SYLLABLE-BF6D
        0xBF6E, # G+13228  U+BF6E  uniBF6E,  HANGUL SYLLABLE-BF6E
        0xBF6F, # G+13229  U+BF6F  uniBF6F,  HANGUL SYLLABLE-BF6F
        0xBF70, # G+13230  U+BF70  uniBF70,  HANGUL SYLLABLE-BF70
        0xBF71, # G+13231  U+BF71  uniBF71,  HANGUL SYLLABLE-BF71
        0xBF72, # G+13232  U+BF72  uniBF72,  HANGUL SYLLABLE-BF72
        0xBF73, # G+13233  U+BF73  uniBF73,  HANGUL SYLLABLE-BF73
        0xBF74, # G+13234  U+BF74  uniBF74,  HANGUL SYLLABLE-BF74
        0xBF75, # G+13235  U+BF75  uniBF75,  HANGUL SYLLABLE-BF75
        0xBF76, # G+13236  U+BF76  uniBF76,  HANGUL SYLLABLE-BF76
        0xBF77, # G+13237  U+BF77  uniBF77,  HANGUL SYLLABLE-BF77
        0xBF78, # G+13238  U+BF78  uniBF78,  HANGUL SYLLABLE-BF78
        0xBF79, # G+13239  U+BF79  uniBF79,  HANGUL SYLLABLE-BF79
        0xBF7A, # G+13240  U+BF7A  uniBF7A,  HANGUL SYLLABLE-BF7A
        0xBF7B, # G+13241  U+BF7B  uniBF7B,  HANGUL SYLLABLE-BF7B
        0xBF7C, # G+13242  U+BF7C  uniBF7C,  HANGUL SYLLABLE-BF7C
        0xBF7D, # G+13243  U+BF7D  uniBF7D,  HANGUL SYLLABLE-BF7D
        0xBF7E, # G+13244  U+BF7E  uniBF7E,  HANGUL SYLLABLE-BF7E
        0xBF7F, # G+13245  U+BF7F  uniBF7F,  HANGUL SYLLABLE-BF7F
        0xBF80, # G+13246  U+BF80  uniBF80,  HANGUL SYLLABLE-BF80
        0xBF81, # G+13247  U+BF81  uniBF81,  HANGUL SYLLABLE-BF81
        0xBF82, # G+13248  U+BF82  uniBF82,  HANGUL SYLLABLE-BF82
        0xBF83, # G+13249  U+BF83  uniBF83,  HANGUL SYLLABLE-BF83
        0xBF84, # G+13250  U+BF84  uniBF84,  HANGUL SYLLABLE-BF84
        0xBF85, # G+13251  U+BF85  uniBF85,  HANGUL SYLLABLE-BF85
        0xBF86, # G+13252  U+BF86  uniBF86,  HANGUL SYLLABLE-BF86
        0xBF87, # G+13253  U+BF87  uniBF87,  HANGUL SYLLABLE-BF87
        0xBF88, # G+13254  U+BF88  uniBF88,  HANGUL SYLLABLE-BF88
        0xBF89, # G+13255  U+BF89  uniBF89,  HANGUL SYLLABLE-BF89
        0xBF8A, # G+13256  U+BF8A  uniBF8A,  HANGUL SYLLABLE-BF8A
        0xBF8B, # G+13257  U+BF8B  uniBF8B,  HANGUL SYLLABLE-BF8B
        0xBF8C, # G+13258  U+BF8C  uniBF8C,  HANGUL SYLLABLE-BF8C
        0xBF8D, # G+13259  U+BF8D  uniBF8D,  HANGUL SYLLABLE-BF8D
        0xBF8E, # G+13260  U+BF8E  uniBF8E,  HANGUL SYLLABLE-BF8E
        0xBF8F, # G+13261  U+BF8F  uniBF8F,  HANGUL SYLLABLE-BF8F
        0xBF90, # G+13262  U+BF90  uniBF90,  HANGUL SYLLABLE-BF90
        0xBF91, # G+13263  U+BF91  uniBF91,  HANGUL SYLLABLE-BF91
        0xBF92, # G+13264  U+BF92  uniBF92,  HANGUL SYLLABLE-BF92
        0xBF93, # G+13265  U+BF93  uniBF93,  HANGUL SYLLABLE-BF93
        0xBF95, # G+13266  U+BF95  uniBF95,  HANGUL SYLLABLE-BF95
        0xBF96, # G+13267  U+BF96  uniBF96,  HANGUL SYLLABLE-BF96
        0xBF97, # G+13268  U+BF97  uniBF97,  HANGUL SYLLABLE-BF97
        0xBF98, # G+13269  U+BF98  uniBF98,  HANGUL SYLLABLE-BF98
        0xBF99, # G+13270  U+BF99  uniBF99,  HANGUL SYLLABLE-BF99
        0xBF9A, # G+13271  U+BF9A  uniBF9A,  HANGUL SYLLABLE-BF9A
        0xBF9B, # G+13272  U+BF9B  uniBF9B,  HANGUL SYLLABLE-BF9B
        0xBF9C, # G+13273  U+BF9C  uniBF9C,  HANGUL SYLLABLE-BF9C
        0xBF9D, # G+13274  U+BF9D  uniBF9D,  HANGUL SYLLABLE-BF9D
        0xBF9E, # G+13275  U+BF9E  uniBF9E,  HANGUL SYLLABLE-BF9E
        0xBF9F, # G+13276  U+BF9F  uniBF9F,  HANGUL SYLLABLE-BF9F
        0xBFA0, # G+13277  U+BFA0  uniBFA0,  HANGUL SYLLABLE-BFA0
        0xBFA1, # G+13278  U+BFA1  uniBFA1,  HANGUL SYLLABLE-BFA1
        0xBFA2, # G+13279  U+BFA2  uniBFA2,  HANGUL SYLLABLE-BFA2
        0xBFA3, # G+13280  U+BFA3  uniBFA3,  HANGUL SYLLABLE-BFA3
        0xBFA4, # G+13281  U+BFA4  uniBFA4,  HANGUL SYLLABLE-BFA4
        0xBFA5, # G+13282  U+BFA5  uniBFA5,  HANGUL SYLLABLE-BFA5
        0xBFA6, # G+13283  U+BFA6  uniBFA6,  HANGUL SYLLABLE-BFA6
        0xBFA7, # G+13284  U+BFA7  uniBFA7,  HANGUL SYLLABLE-BFA7
        0xBFA8, # G+13285  U+BFA8  uniBFA8,  HANGUL SYLLABLE-BFA8
        0xBFA9, # G+13286  U+BFA9  uniBFA9,  HANGUL SYLLABLE-BFA9
        0xBFAA, # G+13287  U+BFAA  uniBFAA,  HANGUL SYLLABLE-BFAA
        0xBFAB, # G+13288  U+BFAB  uniBFAB,  HANGUL SYLLABLE-BFAB
        0xBFAC, # G+13289  U+BFAC  uniBFAC,  HANGUL SYLLABLE-BFAC
        0xBFAD, # G+13290  U+BFAD  uniBFAD,  HANGUL SYLLABLE-BFAD
        0xBFAE, # G+13291  U+BFAE  uniBFAE,  HANGUL SYLLABLE-BFAE
        0xBFAF, # G+13292  U+BFAF  uniBFAF,  HANGUL SYLLABLE-BFAF
        0xBFB1, # G+13293  U+BFB1  uniBFB1,  HANGUL SYLLABLE-BFB1
        0xBFB2, # G+13294  U+BFB2  uniBFB2,  HANGUL SYLLABLE-BFB2
        0xBFB3, # G+13295  U+BFB3  uniBFB3,  HANGUL SYLLABLE-BFB3
        0xBFB4, # G+13296  U+BFB4  uniBFB4,  HANGUL SYLLABLE-BFB4
        0xBFB5, # G+13297  U+BFB5  uniBFB5,  HANGUL SYLLABLE-BFB5
        0xBFB6, # G+13298  U+BFB6  uniBFB6,  HANGUL SYLLABLE-BFB6
        0xBFB7, # G+13299  U+BFB7  uniBFB7,  HANGUL SYLLABLE-BFB7
        0xBFB8, # G+13300  U+BFB8  uniBFB8,  HANGUL SYLLABLE-BFB8
        0xBFB9, # G+13301  U+BFB9  uniBFB9,  HANGUL SYLLABLE-BFB9
        0xBFBA, # G+13302  U+BFBA  uniBFBA,  HANGUL SYLLABLE-BFBA
        0xBFBB, # G+13303  U+BFBB  uniBFBB,  HANGUL SYLLABLE-BFBB
        0xBFBC, # G+13304  U+BFBC  uniBFBC,  HANGUL SYLLABLE-BFBC
        0xBFBD, # G+13305  U+BFBD  uniBFBD,  HANGUL SYLLABLE-BFBD
        0xBFBE, # G+13306  U+BFBE  uniBFBE,  HANGUL SYLLABLE-BFBE
        0xBFBF, # G+13307  U+BFBF  uniBFBF,  HANGUL SYLLABLE-BFBF
        0xBFC0, # G+13308  U+BFC0  uniBFC0,  HANGUL SYLLABLE-BFC0
        0xBFC1, # G+13309  U+BFC1  uniBFC1,  HANGUL SYLLABLE-BFC1
        0xBFC2, # G+13310  U+BFC2  uniBFC2,  HANGUL SYLLABLE-BFC2
        0xBFC3, # G+13311  U+BFC3  uniBFC3,  HANGUL SYLLABLE-BFC3
        0xBFC4, # G+13312  U+BFC4  uniBFC4,  HANGUL SYLLABLE-BFC4
        0xBFC6, # G+13313  U+BFC6  uniBFC6,  HANGUL SYLLABLE-BFC6
        0xBFC7, # G+13314  U+BFC7  uniBFC7,  HANGUL SYLLABLE-BFC7
        0xBFC8, # G+13315  U+BFC8  uniBFC8,  HANGUL SYLLABLE-BFC8
        0xBFC9, # G+13316  U+BFC9  uniBFC9,  HANGUL SYLLABLE-BFC9
        0xBFCA, # G+13317  U+BFCA  uniBFCA,  HANGUL SYLLABLE-BFCA
        0xBFCB, # G+13318  U+BFCB  uniBFCB,  HANGUL SYLLABLE-BFCB
        0xBFCE, # G+13319  U+BFCE  uniBFCE,  HANGUL SYLLABLE-BFCE
        0xBFCF, # G+13320  U+BFCF  uniBFCF,  HANGUL SYLLABLE-BFCF
        0xBFD1, # G+13321  U+BFD1  uniBFD1,  HANGUL SYLLABLE-BFD1
        0xBFD2, # G+13322  U+BFD2  uniBFD2,  HANGUL SYLLABLE-BFD2
        0xBFD3, # G+13323  U+BFD3  uniBFD3,  HANGUL SYLLABLE-BFD3
        0xBFD5, # G+13324  U+BFD5  uniBFD5,  HANGUL SYLLABLE-BFD5
        0xBFD6, # G+13325  U+BFD6  uniBFD6,  HANGUL SYLLABLE-BFD6
        0xBFD7, # G+13326  U+BFD7  uniBFD7,  HANGUL SYLLABLE-BFD7
        0xBFD8, # G+13327  U+BFD8  uniBFD8,  HANGUL SYLLABLE-BFD8
        0xBFD9, # G+13328  U+BFD9  uniBFD9,  HANGUL SYLLABLE-BFD9
        0xBFDA, # G+13329  U+BFDA  uniBFDA,  HANGUL SYLLABLE-BFDA
        0xBFDB, # G+13330  U+BFDB  uniBFDB,  HANGUL SYLLABLE-BFDB
        0xBFDD, # G+13331  U+BFDD  uniBFDD,  HANGUL SYLLABLE-BFDD
        0xBFDE, # G+13332  U+BFDE  uniBFDE,  HANGUL SYLLABLE-BFDE
        0xBFE0, # G+13333  U+BFE0  uniBFE0,  HANGUL SYLLABLE-BFE0
        0xBFE2, # G+13334  U+BFE2  uniBFE2,  HANGUL SYLLABLE-BFE2
        0xBFE3, # G+13335  U+BFE3  uniBFE3,  HANGUL SYLLABLE-BFE3
        0xBFE4, # G+13336  U+BFE4  uniBFE4,  HANGUL SYLLABLE-BFE4
        0xBFE5, # G+13337  U+BFE5  uniBFE5,  HANGUL SYLLABLE-BFE5
        0xBFE6, # G+13338  U+BFE6  uniBFE6,  HANGUL SYLLABLE-BFE6
        0xBFE7, # G+13339  U+BFE7  uniBFE7,  HANGUL SYLLABLE-BFE7
        0xBFE8, # G+13340  U+BFE8  uniBFE8,  HANGUL SYLLABLE-BFE8
        0xBFE9, # G+13341  U+BFE9  uniBFE9,  HANGUL SYLLABLE-BFE9
        0xBFEA, # G+13342  U+BFEA  uniBFEA,  HANGUL SYLLABLE-BFEA
        0xBFEB, # G+13343  U+BFEB  uniBFEB,  HANGUL SYLLABLE-BFEB
        0xBFEC, # G+13344  U+BFEC  uniBFEC,  HANGUL SYLLABLE-BFEC
        0xBFED, # G+13345  U+BFED  uniBFED,  HANGUL SYLLABLE-BFED
        0xBFEE, # G+13346  U+BFEE  uniBFEE,  HANGUL SYLLABLE-BFEE
        0xBFEF, # G+13347  U+BFEF  uniBFEF,  HANGUL SYLLABLE-BFEF
        0xBFF0, # G+13348  U+BFF0  uniBFF0,  HANGUL SYLLABLE-BFF0
        0xBFF1, # G+13349  U+BFF1  uniBFF1,  HANGUL SYLLABLE-BFF1
        0xBFF2, # G+13350  U+BFF2  uniBFF2,  HANGUL SYLLABLE-BFF2
        0xBFF3, # G+13351  U+BFF3  uniBFF3,  HANGUL SYLLABLE-BFF3
        0xBFF4, # G+13352  U+BFF4  uniBFF4,  HANGUL SYLLABLE-BFF4
        0xBFF5, # G+13353  U+BFF5  uniBFF5,  HANGUL SYLLABLE-BFF5
        0xBFF6, # G+13354  U+BFF6  uniBFF6,  HANGUL SYLLABLE-BFF6
        0xBFF7, # G+13355  U+BFF7  uniBFF7,  HANGUL SYLLABLE-BFF7
        0xBFF8, # G+13356  U+BFF8  uniBFF8,  HANGUL SYLLABLE-BFF8
        0xBFF9, # G+13357  U+BFF9  uniBFF9,  HANGUL SYLLABLE-BFF9
        0xBFFA, # G+13358  U+BFFA  uniBFFA,  HANGUL SYLLABLE-BFFA
        0xBFFB, # G+13359  U+BFFB  uniBFFB,  HANGUL SYLLABLE-BFFB
        0xBFFC, # G+13360  U+BFFC  uniBFFC,  HANGUL SYLLABLE-BFFC
        0xBFFD, # G+13361  U+BFFD  uniBFFD,  HANGUL SYLLABLE-BFFD
        0xBFFE, # G+13362  U+BFFE  uniBFFE,  HANGUL SYLLABLE-BFFE
        0xBFFF, # G+13363  U+BFFF  uniBFFF,  HANGUL SYLLABLE-BFFF
        0xC000, # G+13364  U+C000  uniC000,  HANGUL SYLLABLE-C000
        0xC001, # G+13365  U+C001  uniC001,  HANGUL SYLLABLE-C001
        0xC002, # G+13366  U+C002  uniC002,  HANGUL SYLLABLE-C002
        0xC003, # G+13367  U+C003  uniC003,  HANGUL SYLLABLE-C003
        0xC004, # G+13368  U+C004  uniC004,  HANGUL SYLLABLE-C004
        0xC005, # G+13369  U+C005  uniC005,  HANGUL SYLLABLE-C005
        0xC006, # G+13370  U+C006  uniC006,  HANGUL SYLLABLE-C006
        0xC007, # G+13371  U+C007  uniC007,  HANGUL SYLLABLE-C007
        0xC008, # G+13372  U+C008  uniC008,  HANGUL SYLLABLE-C008
        0xC009, # G+13373  U+C009  uniC009,  HANGUL SYLLABLE-C009
        0xC00A, # G+13374  U+C00A  uniC00A,  HANGUL SYLLABLE-C00A
        0xC00B, # G+13375  U+C00B  uniC00B,  HANGUL SYLLABLE-C00B
        0xC00C, # G+13376  U+C00C  uniC00C,  HANGUL SYLLABLE-C00C
        0xC00D, # G+13377  U+C00D  uniC00D,  HANGUL SYLLABLE-C00D
        0xC00E, # G+13378  U+C00E  uniC00E,  HANGUL SYLLABLE-C00E
        0xC00F, # G+13379  U+C00F  uniC00F,  HANGUL SYLLABLE-C00F
        0xC010, # G+13380  U+C010  uniC010,  HANGUL SYLLABLE-C010
        0xC011, # G+13381  U+C011  uniC011,  HANGUL SYLLABLE-C011
        0xC012, # G+13382  U+C012  uniC012,  HANGUL SYLLABLE-C012
        0xC013, # G+13383  U+C013  uniC013,  HANGUL SYLLABLE-C013
        0xC014, # G+13384  U+C014  uniC014,  HANGUL SYLLABLE-C014
        0xC015, # G+13385  U+C015  uniC015,  HANGUL SYLLABLE-C015
        0xC016, # G+13386  U+C016  uniC016,  HANGUL SYLLABLE-C016
        0xC017, # G+13387  U+C017  uniC017,  HANGUL SYLLABLE-C017
        0xC018, # G+13388  U+C018  uniC018,  HANGUL SYLLABLE-C018
        0xC019, # G+13389  U+C019  uniC019,  HANGUL SYLLABLE-C019
        0xC01A, # G+13390  U+C01A  uniC01A,  HANGUL SYLLABLE-C01A
        0xC01B, # G+13391  U+C01B  uniC01B,  HANGUL SYLLABLE-C01B
        0xC01C, # G+13392  U+C01C  uniC01C,  HANGUL SYLLABLE-C01C
        0xC01D, # G+13393  U+C01D  uniC01D,  HANGUL SYLLABLE-C01D
        0xC01E, # G+13394  U+C01E  uniC01E,  HANGUL SYLLABLE-C01E
        0xC01F, # G+13395  U+C01F  uniC01F,  HANGUL SYLLABLE-C01F
        0xC020, # G+13396  U+C020  uniC020,  HANGUL SYLLABLE-C020
        0xC021, # G+13397  U+C021  uniC021,  HANGUL SYLLABLE-C021
        0xC022, # G+13398  U+C022  uniC022,  HANGUL SYLLABLE-C022
        0xC023, # G+13399  U+C023  uniC023,  HANGUL SYLLABLE-C023
        0xC024, # G+13400  U+C024  uniC024,  HANGUL SYLLABLE-C024
        0xC025, # G+13401  U+C025  uniC025,  HANGUL SYLLABLE-C025
        0xC026, # G+13402  U+C026  uniC026,  HANGUL SYLLABLE-C026
        0xC027, # G+13403  U+C027  uniC027,  HANGUL SYLLABLE-C027
        0xC028, # G+13404  U+C028  uniC028,  HANGUL SYLLABLE-C028
        0xC029, # G+13405  U+C029  uniC029,  HANGUL SYLLABLE-C029
        0xC02A, # G+13406  U+C02A  uniC02A,  HANGUL SYLLABLE-C02A
        0xC02B, # G+13407  U+C02B  uniC02B,  HANGUL SYLLABLE-C02B
        0xC02C, # G+13408  U+C02C  uniC02C,  HANGUL SYLLABLE-C02C
        0xC02D, # G+13409  U+C02D  uniC02D,  HANGUL SYLLABLE-C02D
        0xC02E, # G+13410  U+C02E  uniC02E,  HANGUL SYLLABLE-C02E
        0xC02F, # G+13411  U+C02F  uniC02F,  HANGUL SYLLABLE-C02F
        0xC030, # G+13412  U+C030  uniC030,  HANGUL SYLLABLE-C030
        0xC031, # G+13413  U+C031  uniC031,  HANGUL SYLLABLE-C031
        0xC032, # G+13414  U+C032  uniC032,  HANGUL SYLLABLE-C032
        0xC033, # G+13415  U+C033  uniC033,  HANGUL SYLLABLE-C033
        0xC034, # G+13416  U+C034  uniC034,  HANGUL SYLLABLE-C034
        0xC035, # G+13417  U+C035  uniC035,  HANGUL SYLLABLE-C035
        0xC036, # G+13418  U+C036  uniC036,  HANGUL SYLLABLE-C036
        0xC037, # G+13419  U+C037  uniC037,  HANGUL SYLLABLE-C037
        0xC038, # G+13420  U+C038  uniC038,  HANGUL SYLLABLE-C038
        0xC039, # G+13421  U+C039  uniC039,  HANGUL SYLLABLE-C039
        0xC03A, # G+13422  U+C03A  uniC03A,  HANGUL SYLLABLE-C03A
        0xC03B, # G+13423  U+C03B  uniC03B,  HANGUL SYLLABLE-C03B
        0xC03D, # G+13424  U+C03D  uniC03D,  HANGUL SYLLABLE-C03D
        0xC03E, # G+13425  U+C03E  uniC03E,  HANGUL SYLLABLE-C03E
        0xC03F, # G+13426  U+C03F  uniC03F,  HANGUL SYLLABLE-C03F
        0xC040, # G+13427  U+C040  uniC040,  HANGUL SYLLABLE-C040
        0xC041, # G+13428  U+C041  uniC041,  HANGUL SYLLABLE-C041
        0xC042, # G+13429  U+C042  uniC042,  HANGUL SYLLABLE-C042
        0xC043, # G+13430  U+C043  uniC043,  HANGUL SYLLABLE-C043
        0xC044, # G+13431  U+C044  uniC044,  HANGUL SYLLABLE-C044
        0xC045, # G+13432  U+C045  uniC045,  HANGUL SYLLABLE-C045
        0xC046, # G+13433  U+C046  uniC046,  HANGUL SYLLABLE-C046
        0xC047, # G+13434  U+C047  uniC047,  HANGUL SYLLABLE-C047
        0xC048, # G+13435  U+C048  uniC048,  HANGUL SYLLABLE-C048
        0xC049, # G+13436  U+C049  uniC049,  HANGUL SYLLABLE-C049
        0xC04A, # G+13437  U+C04A  uniC04A,  HANGUL SYLLABLE-C04A
        0xC04B, # G+13438  U+C04B  uniC04B,  HANGUL SYLLABLE-C04B
        0xC04C, # G+13439  U+C04C  uniC04C,  HANGUL SYLLABLE-C04C
        0xC04D, # G+13440  U+C04D  uniC04D,  HANGUL SYLLABLE-C04D
        0xC04E, # G+13441  U+C04E  uniC04E,  HANGUL SYLLABLE-C04E
        0xC04F, # G+13442  U+C04F  uniC04F,  HANGUL SYLLABLE-C04F
        0xC050, # G+13443  U+C050  uniC050,  HANGUL SYLLABLE-C050
        0xC052, # G+13444  U+C052  uniC052,  HANGUL SYLLABLE-C052
        0xC053, # G+13445  U+C053  uniC053,  HANGUL SYLLABLE-C053
        0xC054, # G+13446  U+C054  uniC054,  HANGUL SYLLABLE-C054
        0xC055, # G+13447  U+C055  uniC055,  HANGUL SYLLABLE-C055
        0xC056, # G+13448  U+C056  uniC056,  HANGUL SYLLABLE-C056
        0xC057, # G+13449  U+C057  uniC057,  HANGUL SYLLABLE-C057
        0xC059, # G+13450  U+C059  uniC059,  HANGUL SYLLABLE-C059
        0xC05A, # G+13451  U+C05A  uniC05A,  HANGUL SYLLABLE-C05A
        0xC05B, # G+13452  U+C05B  uniC05B,  HANGUL SYLLABLE-C05B
        0xC05D, # G+13453  U+C05D  uniC05D,  HANGUL SYLLABLE-C05D
        0xC05E, # G+13454  U+C05E  uniC05E,  HANGUL SYLLABLE-C05E
        0xC05F, # G+13455  U+C05F  uniC05F,  HANGUL SYLLABLE-C05F
        0xC061, # G+13456  U+C061  uniC061,  HANGUL SYLLABLE-C061
        0xC062, # G+13457  U+C062  uniC062,  HANGUL SYLLABLE-C062
        0xC063, # G+13458  U+C063  uniC063,  HANGUL SYLLABLE-C063
        0xC064, # G+13459  U+C064  uniC064,  HANGUL SYLLABLE-C064
        0xC065, # G+13460  U+C065  uniC065,  HANGUL SYLLABLE-C065
        0xC066, # G+13461  U+C066  uniC066,  HANGUL SYLLABLE-C066
        0xC067, # G+13462  U+C067  uniC067,  HANGUL SYLLABLE-C067
        0xC06A, # G+13463  U+C06A  uniC06A,  HANGUL SYLLABLE-C06A
        0xC06B, # G+13464  U+C06B  uniC06B,  HANGUL SYLLABLE-C06B
        0xC06C, # G+13465  U+C06C  uniC06C,  HANGUL SYLLABLE-C06C
        0xC06D, # G+13466  U+C06D  uniC06D,  HANGUL SYLLABLE-C06D
        0xC06E, # G+13467  U+C06E  uniC06E,  HANGUL SYLLABLE-C06E
        0xC06F, # G+13468  U+C06F  uniC06F,  HANGUL SYLLABLE-C06F
        0xC070, # G+13469  U+C070  uniC070,  HANGUL SYLLABLE-C070
        0xC071, # G+13470  U+C071  uniC071,  HANGUL SYLLABLE-C071
        0xC072, # G+13471  U+C072  uniC072,  HANGUL SYLLABLE-C072
        0xC073, # G+13472  U+C073  uniC073,  HANGUL SYLLABLE-C073
        0xC074, # G+13473  U+C074  uniC074,  HANGUL SYLLABLE-C074
        0xC075, # G+13474  U+C075  uniC075,  HANGUL SYLLABLE-C075
        0xC076, # G+13475  U+C076  uniC076,  HANGUL SYLLABLE-C076
        0xC077, # G+13476  U+C077  uniC077,  HANGUL SYLLABLE-C077
        0xC078, # G+13477  U+C078  uniC078,  HANGUL SYLLABLE-C078
        0xC079, # G+13478  U+C079  uniC079,  HANGUL SYLLABLE-C079
        0xC07A, # G+13479  U+C07A  uniC07A,  HANGUL SYLLABLE-C07A
        0xC07B, # G+13480  U+C07B  uniC07B,  HANGUL SYLLABLE-C07B
        0xC07C, # G+13481  U+C07C  uniC07C,  HANGUL SYLLABLE-C07C
        0xC07D, # G+13482  U+C07D  uniC07D,  HANGUL SYLLABLE-C07D
        0xC07E, # G+13483  U+C07E  uniC07E,  HANGUL SYLLABLE-C07E
        0xC07F, # G+13484  U+C07F  uniC07F,  HANGUL SYLLABLE-C07F
        0xC080, # G+13485  U+C080  uniC080,  HANGUL SYLLABLE-C080
        0xC081, # G+13486  U+C081  uniC081,  HANGUL SYLLABLE-C081
        0xC082, # G+13487  U+C082  uniC082,  HANGUL SYLLABLE-C082
        0xC083, # G+13488  U+C083  uniC083,  HANGUL SYLLABLE-C083
        0xC084, # G+13489  U+C084  uniC084,  HANGUL SYLLABLE-C084
        0xC085, # G+13490  U+C085  uniC085,  HANGUL SYLLABLE-C085
        0xC086, # G+13491  U+C086  uniC086,  HANGUL SYLLABLE-C086
        0xC087, # G+13492  U+C087  uniC087,  HANGUL SYLLABLE-C087
        0xC088, # G+13493  U+C088  uniC088,  HANGUL SYLLABLE-C088
        0xC089, # G+13494  U+C089  uniC089,  HANGUL SYLLABLE-C089
        0xC08A, # G+13495  U+C08A  uniC08A,  HANGUL SYLLABLE-C08A
        0xC08B, # G+13496  U+C08B  uniC08B,  HANGUL SYLLABLE-C08B
        0xC08C, # G+13497  U+C08C  uniC08C,  HANGUL SYLLABLE-C08C
        0xC08D, # G+13498  U+C08D  uniC08D,  HANGUL SYLLABLE-C08D
        0xC08E, # G+13499  U+C08E  uniC08E,  HANGUL SYLLABLE-C08E
        0xC08F, # G+13500  U+C08F  uniC08F,  HANGUL SYLLABLE-C08F
        0xC092, # G+13501  U+C092  uniC092,  HANGUL SYLLABLE-C092
        0xC093, # G+13502  U+C093  uniC093,  HANGUL SYLLABLE-C093
        0xC095, # G+13503  U+C095  uniC095,  HANGUL SYLLABLE-C095
        0xC096, # G+13504  U+C096  uniC096,  HANGUL SYLLABLE-C096
        0xC097, # G+13505  U+C097  uniC097,  HANGUL SYLLABLE-C097
        0xC099, # G+13506  U+C099  uniC099,  HANGUL SYLLABLE-C099
        0xC09A, # G+13507  U+C09A  uniC09A,  HANGUL SYLLABLE-C09A
        0xC09B, # G+13508  U+C09B  uniC09B,  HANGUL SYLLABLE-C09B
        0xC09C, # G+13509  U+C09C  uniC09C,  HANGUL SYLLABLE-C09C
        0xC09D, # G+13510  U+C09D  uniC09D,  HANGUL SYLLABLE-C09D
        0xC09E, # G+13511  U+C09E  uniC09E,  HANGUL SYLLABLE-C09E
        0xC09F, # G+13512  U+C09F  uniC09F,  HANGUL SYLLABLE-C09F
        0xC0A2, # G+13513  U+C0A2  uniC0A2,  HANGUL SYLLABLE-C0A2
        0xC0A4, # G+13514  U+C0A4  uniC0A4,  HANGUL SYLLABLE-C0A4
        0xC0A6, # G+13515  U+C0A6  uniC0A6,  HANGUL SYLLABLE-C0A6
        0xC0A7, # G+13516  U+C0A7  uniC0A7,  HANGUL SYLLABLE-C0A7
        0xC0A8, # G+13517  U+C0A8  uniC0A8,  HANGUL SYLLABLE-C0A8
        0xC0A9, # G+13518  U+C0A9  uniC0A9,  HANGUL SYLLABLE-C0A9
        0xC0AA, # G+13519  U+C0AA  uniC0AA,  HANGUL SYLLABLE-C0AA
        0xC0AB, # G+13520  U+C0AB  uniC0AB,  HANGUL SYLLABLE-C0AB
        0xC0AE, # G+13521  U+C0AE  uniC0AE,  HANGUL SYLLABLE-C0AE
        0xC0B1, # G+13522  U+C0B1  uniC0B1,  HANGUL SYLLABLE-C0B1
        0xC0B2, # G+13523  U+C0B2  uniC0B2,  HANGUL SYLLABLE-C0B2
        0xC0B7, # G+13524  U+C0B7  uniC0B7,  HANGUL SYLLABLE-C0B7
        0xC0B8, # G+13525  U+C0B8  uniC0B8,  HANGUL SYLLABLE-C0B8
        0xC0B9, # G+13526  U+C0B9  uniC0B9,  HANGUL SYLLABLE-C0B9
        0xC0BA, # G+13527  U+C0BA  uniC0BA,  HANGUL SYLLABLE-C0BA
        0xC0BB, # G+13528  U+C0BB  uniC0BB,  HANGUL SYLLABLE-C0BB
        0xC0BE, # G+13529  U+C0BE  uniC0BE,  HANGUL SYLLABLE-C0BE
        0xC0C2, # G+13530  U+C0C2  uniC0C2,  HANGUL SYLLABLE-C0C2
        0xC0C3, # G+13531  U+C0C3  uniC0C3,  HANGUL SYLLABLE-C0C3
        0xC0C4, # G+13532  U+C0C4  uniC0C4,  HANGUL SYLLABLE-C0C4
        0xC0C6, # G+13533  U+C0C6  uniC0C6,  HANGUL SYLLABLE-C0C6
        0xC0C7, # G+13534  U+C0C7  uniC0C7,  HANGUL SYLLABLE-C0C7
        0xC0CA, # G+13535  U+C0CA  uniC0CA,  HANGUL SYLLABLE-C0CA
        0xC0CB, # G+13536  U+C0CB  uniC0CB,  HANGUL SYLLABLE-C0CB
        0xC0CD, # G+13537  U+C0CD  uniC0CD,  HANGUL SYLLABLE-C0CD
        0xC0CE, # G+13538  U+C0CE  uniC0CE,  HANGUL SYLLABLE-C0CE
        0xC0CF, # G+13539  U+C0CF  uniC0CF,  HANGUL SYLLABLE-C0CF
        0xC0D1, # G+13540  U+C0D1  uniC0D1,  HANGUL SYLLABLE-C0D1
        0xC0D2, # G+13541  U+C0D2  uniC0D2,  HANGUL SYLLABLE-C0D2
        0xC0D3, # G+13542  U+C0D3  uniC0D3,  HANGUL SYLLABLE-C0D3
        0xC0D4, # G+13543  U+C0D4  uniC0D4,  HANGUL SYLLABLE-C0D4
        0xC0D5, # G+13544  U+C0D5  uniC0D5,  HANGUL SYLLABLE-C0D5
        0xC0D6, # G+13545  U+C0D6  uniC0D6,  HANGUL SYLLABLE-C0D6
        0xC0D7, # G+13546  U+C0D7  uniC0D7,  HANGUL SYLLABLE-C0D7
        0xC0DA, # G+13547  U+C0DA  uniC0DA,  HANGUL SYLLABLE-C0DA
        0xC0DE, # G+13548  U+C0DE  uniC0DE,  HANGUL SYLLABLE-C0DE
        0xC0DF, # G+13549  U+C0DF  uniC0DF,  HANGUL SYLLABLE-C0DF
        0xC0E0, # G+13550  U+C0E0  uniC0E0,  HANGUL SYLLABLE-C0E0
        0xC0E1, # G+13551  U+C0E1  uniC0E1,  HANGUL SYLLABLE-C0E1
        0xC0E2, # G+13552  U+C0E2  uniC0E2,  HANGUL SYLLABLE-C0E2
        0xC0E3, # G+13553  U+C0E3  uniC0E3,  HANGUL SYLLABLE-C0E3
        0xC0E6, # G+13554  U+C0E6  uniC0E6,  HANGUL SYLLABLE-C0E6
        0xC0E7, # G+13555  U+C0E7  uniC0E7,  HANGUL SYLLABLE-C0E7
        0xC0E9, # G+13556  U+C0E9  uniC0E9,  HANGUL SYLLABLE-C0E9
        0xC0EA, # G+13557  U+C0EA  uniC0EA,  HANGUL SYLLABLE-C0EA
        0xC0EB, # G+13558  U+C0EB  uniC0EB,  HANGUL SYLLABLE-C0EB
        0xC0ED, # G+13559  U+C0ED  uniC0ED,  HANGUL SYLLABLE-C0ED
        0xC0EE, # G+13560  U+C0EE  uniC0EE,  HANGUL SYLLABLE-C0EE
        0xC0EF, # G+13561  U+C0EF  uniC0EF,  HANGUL SYLLABLE-C0EF
        0xC0F0, # G+13562  U+C0F0  uniC0F0,  HANGUL SYLLABLE-C0F0
        0xC0F1, # G+13563  U+C0F1  uniC0F1,  HANGUL SYLLABLE-C0F1
        0xC0F2, # G+13564  U+C0F2  uniC0F2,  HANGUL SYLLABLE-C0F2
        0xC0F3, # G+13565  U+C0F3  uniC0F3,  HANGUL SYLLABLE-C0F3
        0xC0F6, # G+13566  U+C0F6  uniC0F6,  HANGUL SYLLABLE-C0F6
        0xC0F8, # G+13567  U+C0F8  uniC0F8,  HANGUL SYLLABLE-C0F8
        0xC0FA, # G+13568  U+C0FA  uniC0FA,  HANGUL SYLLABLE-C0FA
        0xC0FB, # G+13569  U+C0FB  uniC0FB,  HANGUL SYLLABLE-C0FB
        0xC0FC, # G+13570  U+C0FC  uniC0FC,  HANGUL SYLLABLE-C0FC
        0xC0FD, # G+13571  U+C0FD  uniC0FD,  HANGUL SYLLABLE-C0FD
        0xC0FE, # G+13572  U+C0FE  uniC0FE,  HANGUL SYLLABLE-C0FE
        0xC0FF, # G+13573  U+C0FF  uniC0FF,  HANGUL SYLLABLE-C0FF
        0xC101, # G+13574  U+C101  uniC101,  HANGUL SYLLABLE-C101
        0xC102, # G+13575  U+C102  uniC102,  HANGUL SYLLABLE-C102
        0xC103, # G+13576  U+C103  uniC103,  HANGUL SYLLABLE-C103
        0xC105, # G+13577  U+C105  uniC105,  HANGUL SYLLABLE-C105
        0xC106, # G+13578  U+C106  uniC106,  HANGUL SYLLABLE-C106
        0xC107, # G+13579  U+C107  uniC107,  HANGUL SYLLABLE-C107
        0xC109, # G+13580  U+C109  uniC109,  HANGUL SYLLABLE-C109
        0xC10A, # G+13581  U+C10A  uniC10A,  HANGUL SYLLABLE-C10A
        0xC10B, # G+13582  U+C10B  uniC10B,  HANGUL SYLLABLE-C10B
        0xC10C, # G+13583  U+C10C  uniC10C,  HANGUL SYLLABLE-C10C
        0xC10D, # G+13584  U+C10D  uniC10D,  HANGUL SYLLABLE-C10D
        0xC10E, # G+13585  U+C10E  uniC10E,  HANGUL SYLLABLE-C10E
        0xC10F, # G+13586  U+C10F  uniC10F,  HANGUL SYLLABLE-C10F
        0xC111, # G+13587  U+C111  uniC111,  HANGUL SYLLABLE-C111
        0xC112, # G+13588  U+C112  uniC112,  HANGUL SYLLABLE-C112
        0xC113, # G+13589  U+C113  uniC113,  HANGUL SYLLABLE-C113
        0xC114, # G+13590  U+C114  uniC114,  HANGUL SYLLABLE-C114
        0xC116, # G+13591  U+C116  uniC116,  HANGUL SYLLABLE-C116
        0xC117, # G+13592  U+C117  uniC117,  HANGUL SYLLABLE-C117
        0xC118, # G+13593  U+C118  uniC118,  HANGUL SYLLABLE-C118
        0xC119, # G+13594  U+C119  uniC119,  HANGUL SYLLABLE-C119
        0xC11A, # G+13595  U+C11A  uniC11A,  HANGUL SYLLABLE-C11A
        0xC11B, # G+13596  U+C11B  uniC11B,  HANGUL SYLLABLE-C11B
        0xC121, # G+13597  U+C121  uniC121,  HANGUL SYLLABLE-C121
        0xC122, # G+13598  U+C122  uniC122,  HANGUL SYLLABLE-C122
        0xC125, # G+13599  U+C125  uniC125,  HANGUL SYLLABLE-C125
        0xC128, # G+13600  U+C128  uniC128,  HANGUL SYLLABLE-C128
        0xC129, # G+13601  U+C129  uniC129,  HANGUL SYLLABLE-C129
        0xC12A, # G+13602  U+C12A  uniC12A,  HANGUL SYLLABLE-C12A
        0xC12B, # G+13603  U+C12B  uniC12B,  HANGUL SYLLABLE-C12B
        0xC12E, # G+13604  U+C12E  uniC12E,  HANGUL SYLLABLE-C12E
        0xC132, # G+13605  U+C132  uniC132,  HANGUL SYLLABLE-C132
        0xC133, # G+13606  U+C133  uniC133,  HANGUL SYLLABLE-C133
        0xC134, # G+13607  U+C134  uniC134,  HANGUL SYLLABLE-C134
        0xC135, # G+13608  U+C135  uniC135,  HANGUL SYLLABLE-C135
        0xC137, # G+13609  U+C137  uniC137,  HANGUL SYLLABLE-C137
        0xC13A, # G+13610  U+C13A  uniC13A,  HANGUL SYLLABLE-C13A
        0xC13B, # G+13611  U+C13B  uniC13B,  HANGUL SYLLABLE-C13B
        0xC13D, # G+13612  U+C13D  uniC13D,  HANGUL SYLLABLE-C13D
        0xC13E, # G+13613  U+C13E  uniC13E,  HANGUL SYLLABLE-C13E
        0xC13F, # G+13614  U+C13F  uniC13F,  HANGUL SYLLABLE-C13F
        0xC141, # G+13615  U+C141  uniC141,  HANGUL SYLLABLE-C141
        0xC142, # G+13616  U+C142  uniC142,  HANGUL SYLLABLE-C142
        0xC143, # G+13617  U+C143  uniC143,  HANGUL SYLLABLE-C143
        0xC144, # G+13618  U+C144  uniC144,  HANGUL SYLLABLE-C144
        0xC145, # G+13619  U+C145  uniC145,  HANGUL SYLLABLE-C145
        0xC146, # G+13620  U+C146  uniC146,  HANGUL SYLLABLE-C146
        0xC147, # G+13621  U+C147  uniC147,  HANGUL SYLLABLE-C147
        0xC14A, # G+13622  U+C14A  uniC14A,  HANGUL SYLLABLE-C14A
        0xC14E, # G+13623  U+C14E  uniC14E,  HANGUL SYLLABLE-C14E
        0xC14F, # G+13624  U+C14F  uniC14F,  HANGUL SYLLABLE-C14F
        0xC150, # G+13625  U+C150  uniC150,  HANGUL SYLLABLE-C150
        0xC151, # G+13626  U+C151  uniC151,  HANGUL SYLLABLE-C151
        0xC152, # G+13627  U+C152  uniC152,  HANGUL SYLLABLE-C152
        0xC153, # G+13628  U+C153  uniC153,  HANGUL SYLLABLE-C153
        0xC156, # G+13629  U+C156  uniC156,  HANGUL SYLLABLE-C156
        0xC157, # G+13630  U+C157  uniC157,  HANGUL SYLLABLE-C157
        0xC159, # G+13631  U+C159  uniC159,  HANGUL SYLLABLE-C159
        0xC15A, # G+13632  U+C15A  uniC15A,  HANGUL SYLLABLE-C15A
        0xC15B, # G+13633  U+C15B  uniC15B,  HANGUL SYLLABLE-C15B
        0xC15D, # G+13634  U+C15D  uniC15D,  HANGUL SYLLABLE-C15D
        0xC15E, # G+13635  U+C15E  uniC15E,  HANGUL SYLLABLE-C15E
        0xC15F, # G+13636  U+C15F  uniC15F,  HANGUL SYLLABLE-C15F
        0xC160, # G+13637  U+C160  uniC160,  HANGUL SYLLABLE-C160
        0xC161, # G+13638  U+C161  uniC161,  HANGUL SYLLABLE-C161
        0xC162, # G+13639  U+C162  uniC162,  HANGUL SYLLABLE-C162
        0xC163, # G+13640  U+C163  uniC163,  HANGUL SYLLABLE-C163
        0xC166, # G+13641  U+C166  uniC166,  HANGUL SYLLABLE-C166
        0xC16A, # G+13642  U+C16A  uniC16A,  HANGUL SYLLABLE-C16A
        0xC16B, # G+13643  U+C16B  uniC16B,  HANGUL SYLLABLE-C16B
        0xC16C, # G+13644  U+C16C  uniC16C,  HANGUL SYLLABLE-C16C
        0xC16D, # G+13645  U+C16D  uniC16D,  HANGUL SYLLABLE-C16D
        0xC16E, # G+13646  U+C16E  uniC16E,  HANGUL SYLLABLE-C16E
        0xC16F, # G+13647  U+C16F  uniC16F,  HANGUL SYLLABLE-C16F
        0xC171, # G+13648  U+C171  uniC171,  HANGUL SYLLABLE-C171
        0xC172, # G+13649  U+C172  uniC172,  HANGUL SYLLABLE-C172
        0xC173, # G+13650  U+C173  uniC173,  HANGUL SYLLABLE-C173
        0xC175, # G+13651  U+C175  uniC175,  HANGUL SYLLABLE-C175
        0xC176, # G+13652  U+C176  uniC176,  HANGUL SYLLABLE-C176
        0xC177, # G+13653  U+C177  uniC177,  HANGUL SYLLABLE-C177
        0xC179, # G+13654  U+C179  uniC179,  HANGUL SYLLABLE-C179
        0xC17A, # G+13655  U+C17A  uniC17A,  HANGUL SYLLABLE-C17A
        0xC17B, # G+13656  U+C17B  uniC17B,  HANGUL SYLLABLE-C17B
        0xC17C, # G+13657  U+C17C  uniC17C,  HANGUL SYLLABLE-C17C
        0xC17D, # G+13658  U+C17D  uniC17D,  HANGUL SYLLABLE-C17D
        0xC17E, # G+13659  U+C17E  uniC17E,  HANGUL SYLLABLE-C17E
        0xC17F, # G+13660  U+C17F  uniC17F,  HANGUL SYLLABLE-C17F
        0xC180, # G+13661  U+C180  uniC180,  HANGUL SYLLABLE-C180
        0xC181, # G+13662  U+C181  uniC181,  HANGUL SYLLABLE-C181
        0xC182, # G+13663  U+C182  uniC182,  HANGUL SYLLABLE-C182
        0xC183, # G+13664  U+C183  uniC183,  HANGUL SYLLABLE-C183
        0xC184, # G+13665  U+C184  uniC184,  HANGUL SYLLABLE-C184
        0xC186, # G+13666  U+C186  uniC186,  HANGUL SYLLABLE-C186
        0xC187, # G+13667  U+C187  uniC187,  HANGUL SYLLABLE-C187
        0xC188, # G+13668  U+C188  uniC188,  HANGUL SYLLABLE-C188
        0xC189, # G+13669  U+C189  uniC189,  HANGUL SYLLABLE-C189
        0xC18A, # G+13670  U+C18A  uniC18A,  HANGUL SYLLABLE-C18A
        0xC18B, # G+13671  U+C18B  uniC18B,  HANGUL SYLLABLE-C18B
        0xC18F, # G+13672  U+C18F  uniC18F,  HANGUL SYLLABLE-C18F
        0xC191, # G+13673  U+C191  uniC191,  HANGUL SYLLABLE-C191
        0xC192, # G+13674  U+C192  uniC192,  HANGUL SYLLABLE-C192
        0xC193, # G+13675  U+C193  uniC193,  HANGUL SYLLABLE-C193
        0xC195, # G+13676  U+C195  uniC195,  HANGUL SYLLABLE-C195
        0xC197, # G+13677  U+C197  uniC197,  HANGUL SYLLABLE-C197
        0xC198, # G+13678  U+C198  uniC198,  HANGUL SYLLABLE-C198
        0xC199, # G+13679  U+C199  uniC199,  HANGUL SYLLABLE-C199
        0xC19A, # G+13680  U+C19A  uniC19A,  HANGUL SYLLABLE-C19A
        0xC19B, # G+13681  U+C19B  uniC19B,  HANGUL SYLLABLE-C19B
        0xC19E, # G+13682  U+C19E  uniC19E,  HANGUL SYLLABLE-C19E
        0xC1A0, # G+13683  U+C1A0  uniC1A0,  HANGUL SYLLABLE-C1A0
        0xC1A2, # G+13684  U+C1A2  uniC1A2,  HANGUL SYLLABLE-C1A2
        0xC1A3, # G+13685  U+C1A3  uniC1A3,  HANGUL SYLLABLE-C1A3
        0xC1A4, # G+13686  U+C1A4  uniC1A4,  HANGUL SYLLABLE-C1A4
        0xC1A6, # G+13687  U+C1A6  uniC1A6,  HANGUL SYLLABLE-C1A6
        0xC1A7, # G+13688  U+C1A7  uniC1A7,  HANGUL SYLLABLE-C1A7
        0xC1AA, # G+13689  U+C1AA  uniC1AA,  HANGUL SYLLABLE-C1AA
        0xC1AB, # G+13690  U+C1AB  uniC1AB,  HANGUL SYLLABLE-C1AB
        0xC1AD, # G+13691  U+C1AD  uniC1AD,  HANGUL SYLLABLE-C1AD
        0xC1AE, # G+13692  U+C1AE  uniC1AE,  HANGUL SYLLABLE-C1AE
        0xC1AF, # G+13693  U+C1AF  uniC1AF,  HANGUL SYLLABLE-C1AF
        0xC1B1, # G+13694  U+C1B1  uniC1B1,  HANGUL SYLLABLE-C1B1
        0xC1B2, # G+13695  U+C1B2  uniC1B2,  HANGUL SYLLABLE-C1B2
        0xC1B3, # G+13696  U+C1B3  uniC1B3,  HANGUL SYLLABLE-C1B3
        0xC1B4, # G+13697  U+C1B4  uniC1B4,  HANGUL SYLLABLE-C1B4
        0xC1B5, # G+13698  U+C1B5  uniC1B5,  HANGUL SYLLABLE-C1B5
        0xC1B6, # G+13699  U+C1B6  uniC1B6,  HANGUL SYLLABLE-C1B6
        0xC1B7, # G+13700  U+C1B7  uniC1B7,  HANGUL SYLLABLE-C1B7
        0xC1B8, # G+13701  U+C1B8  uniC1B8,  HANGUL SYLLABLE-C1B8
        0xC1B9, # G+13702  U+C1B9  uniC1B9,  HANGUL SYLLABLE-C1B9
        0xC1BA, # G+13703  U+C1BA  uniC1BA,  HANGUL SYLLABLE-C1BA
        0xC1BB, # G+13704  U+C1BB  uniC1BB,  HANGUL SYLLABLE-C1BB
        0xC1BC, # G+13705  U+C1BC  uniC1BC,  HANGUL SYLLABLE-C1BC
        0xC1BE, # G+13706  U+C1BE  uniC1BE,  HANGUL SYLLABLE-C1BE
        0xC1BF, # G+13707  U+C1BF  uniC1BF,  HANGUL SYLLABLE-C1BF
        0xC1C0, # G+13708  U+C1C0  uniC1C0,  HANGUL SYLLABLE-C1C0
        0xC1C1, # G+13709  U+C1C1  uniC1C1,  HANGUL SYLLABLE-C1C1
        0xC1C2, # G+13710  U+C1C2  uniC1C2,  HANGUL SYLLABLE-C1C2
        0xC1C3, # G+13711  U+C1C3  uniC1C3,  HANGUL SYLLABLE-C1C3
        0xC1C5, # G+13712  U+C1C5  uniC1C5,  HANGUL SYLLABLE-C1C5
        0xC1C6, # G+13713  U+C1C6  uniC1C6,  HANGUL SYLLABLE-C1C6
        0xC1C7, # G+13714  U+C1C7  uniC1C7,  HANGUL SYLLABLE-C1C7
        0xC1C9, # G+13715  U+C1C9  uniC1C9,  HANGUL SYLLABLE-C1C9
        0xC1CA, # G+13716  U+C1CA  uniC1CA,  HANGUL SYLLABLE-C1CA
        0xC1CB, # G+13717  U+C1CB  uniC1CB,  HANGUL SYLLABLE-C1CB
        0xC1CD, # G+13718  U+C1CD  uniC1CD,  HANGUL SYLLABLE-C1CD
        0xC1CE, # G+13719  U+C1CE  uniC1CE,  HANGUL SYLLABLE-C1CE
        0xC1CF, # G+13720  U+C1CF  uniC1CF,  HANGUL SYLLABLE-C1CF
        0xC1D0, # G+13721  U+C1D0  uniC1D0,  HANGUL SYLLABLE-C1D0
        0xC1D1, # G+13722  U+C1D1  uniC1D1,  HANGUL SYLLABLE-C1D1
        0xC1D2, # G+13723  U+C1D2  uniC1D2,  HANGUL SYLLABLE-C1D2
        0xC1D3, # G+13724  U+C1D3  uniC1D3,  HANGUL SYLLABLE-C1D3
        0xC1D5, # G+13725  U+C1D5  uniC1D5,  HANGUL SYLLABLE-C1D5
        0xC1D6, # G+13726  U+C1D6  uniC1D6,  HANGUL SYLLABLE-C1D6
        0xC1D9, # G+13727  U+C1D9  uniC1D9,  HANGUL SYLLABLE-C1D9
        0xC1DA, # G+13728  U+C1DA  uniC1DA,  HANGUL SYLLABLE-C1DA
        0xC1DB, # G+13729  U+C1DB  uniC1DB,  HANGUL SYLLABLE-C1DB
        0xC1DC, # G+13730  U+C1DC  uniC1DC,  HANGUL SYLLABLE-C1DC
        0xC1DD, # G+13731  U+C1DD  uniC1DD,  HANGUL SYLLABLE-C1DD
        0xC1DE, # G+13732  U+C1DE  uniC1DE,  HANGUL SYLLABLE-C1DE
        0xC1DF, # G+13733  U+C1DF  uniC1DF,  HANGUL SYLLABLE-C1DF
        0xC1E1, # G+13734  U+C1E1  uniC1E1,  HANGUL SYLLABLE-C1E1
        0xC1E2, # G+13735  U+C1E2  uniC1E2,  HANGUL SYLLABLE-C1E2
        0xC1E3, # G+13736  U+C1E3  uniC1E3,  HANGUL SYLLABLE-C1E3
        0xC1E5, # G+13737  U+C1E5  uniC1E5,  HANGUL SYLLABLE-C1E5
        0xC1E6, # G+13738  U+C1E6  uniC1E6,  HANGUL SYLLABLE-C1E6
        0xC1E7, # G+13739  U+C1E7  uniC1E7,  HANGUL SYLLABLE-C1E7
        0xC1E9, # G+13740  U+C1E9  uniC1E9,  HANGUL SYLLABLE-C1E9
        0xC1EA, # G+13741  U+C1EA  uniC1EA,  HANGUL SYLLABLE-C1EA
        0xC1EB, # G+13742  U+C1EB  uniC1EB,  HANGUL SYLLABLE-C1EB
        0xC1EC, # G+13743  U+C1EC  uniC1EC,  HANGUL SYLLABLE-C1EC
        0xC1ED, # G+13744  U+C1ED  uniC1ED,  HANGUL SYLLABLE-C1ED
        0xC1EE, # G+13745  U+C1EE  uniC1EE,  HANGUL SYLLABLE-C1EE
        0xC1EF, # G+13746  U+C1EF  uniC1EF,  HANGUL SYLLABLE-C1EF
        0xC1F2, # G+13747  U+C1F2  uniC1F2,  HANGUL SYLLABLE-C1F2
        0xC1F4, # G+13748  U+C1F4  uniC1F4,  HANGUL SYLLABLE-C1F4
        0xC1F5, # G+13749  U+C1F5  uniC1F5,  HANGUL SYLLABLE-C1F5
        0xC1F6, # G+13750  U+C1F6  uniC1F6,  HANGUL SYLLABLE-C1F6
        0xC1F7, # G+13751  U+C1F7  uniC1F7,  HANGUL SYLLABLE-C1F7
        0xC1F8, # G+13752  U+C1F8  uniC1F8,  HANGUL SYLLABLE-C1F8
        0xC1F9, # G+13753  U+C1F9  uniC1F9,  HANGUL SYLLABLE-C1F9
        0xC1FA, # G+13754  U+C1FA  uniC1FA,  HANGUL SYLLABLE-C1FA
        0xC1FB, # G+13755  U+C1FB  uniC1FB,  HANGUL SYLLABLE-C1FB
        0xC1FE, # G+13756  U+C1FE  uniC1FE,  HANGUL SYLLABLE-C1FE
        0xC1FF, # G+13757  U+C1FF  uniC1FF,  HANGUL SYLLABLE-C1FF
        0xC201, # G+13758  U+C201  uniC201,  HANGUL SYLLABLE-C201
        0xC202, # G+13759  U+C202  uniC202,  HANGUL SYLLABLE-C202
        0xC203, # G+13760  U+C203  uniC203,  HANGUL SYLLABLE-C203
        0xC205, # G+13761  U+C205  uniC205,  HANGUL SYLLABLE-C205
        0xC206, # G+13762  U+C206  uniC206,  HANGUL SYLLABLE-C206
        0xC207, # G+13763  U+C207  uniC207,  HANGUL SYLLABLE-C207
        0xC208, # G+13764  U+C208  uniC208,  HANGUL SYLLABLE-C208
        0xC209, # G+13765  U+C209  uniC209,  HANGUL SYLLABLE-C209
        0xC20A, # G+13766  U+C20A  uniC20A,  HANGUL SYLLABLE-C20A
        0xC20B, # G+13767  U+C20B  uniC20B,  HANGUL SYLLABLE-C20B
        0xC20E, # G+13768  U+C20E  uniC20E,  HANGUL SYLLABLE-C20E
        0xC210, # G+13769  U+C210  uniC210,  HANGUL SYLLABLE-C210
        0xC212, # G+13770  U+C212  uniC212,  HANGUL SYLLABLE-C212
        0xC213, # G+13771  U+C213  uniC213,  HANGUL SYLLABLE-C213
        0xC214, # G+13772  U+C214  uniC214,  HANGUL SYLLABLE-C214
        0xC215, # G+13773  U+C215  uniC215,  HANGUL SYLLABLE-C215
        0xC216, # G+13774  U+C216  uniC216,  HANGUL SYLLABLE-C216
        0xC217, # G+13775  U+C217  uniC217,  HANGUL SYLLABLE-C217
        0xC21A, # G+13776  U+C21A  uniC21A,  HANGUL SYLLABLE-C21A
        0xC21B, # G+13777  U+C21B  uniC21B,  HANGUL SYLLABLE-C21B
        0xC21D, # G+13778  U+C21D  uniC21D,  HANGUL SYLLABLE-C21D
        0xC21E, # G+13779  U+C21E  uniC21E,  HANGUL SYLLABLE-C21E
        0xC221, # G+13780  U+C221  uniC221,  HANGUL SYLLABLE-C221
        0xC222, # G+13781  U+C222  uniC222,  HANGUL SYLLABLE-C222
        0xC223, # G+13782  U+C223  uniC223,  HANGUL SYLLABLE-C223
        0xC224, # G+13783  U+C224  uniC224,  HANGUL SYLLABLE-C224
        0xC225, # G+13784  U+C225  uniC225,  HANGUL SYLLABLE-C225
        0xC226, # G+13785  U+C226  uniC226,  HANGUL SYLLABLE-C226
        0xC227, # G+13786  U+C227  uniC227,  HANGUL SYLLABLE-C227
        0xC22A, # G+13787  U+C22A  uniC22A,  HANGUL SYLLABLE-C22A
        0xC22C, # G+13788  U+C22C  uniC22C,  HANGUL SYLLABLE-C22C
        0xC22E, # G+13789  U+C22E  uniC22E,  HANGUL SYLLABLE-C22E
        0xC230, # G+13790  U+C230  uniC230,  HANGUL SYLLABLE-C230
        0xC233, # G+13791  U+C233  uniC233,  HANGUL SYLLABLE-C233
        0xC235, # G+13792  U+C235  uniC235,  HANGUL SYLLABLE-C235
        0xC236, # G+13793  U+C236  uniC236,  HANGUL SYLLABLE-C236
        0xC237, # G+13794  U+C237  uniC237,  HANGUL SYLLABLE-C237
        0xC238, # G+13795  U+C238  uniC238,  HANGUL SYLLABLE-C238
        0xC239, # G+13796  U+C239  uniC239,  HANGUL SYLLABLE-C239
        0xC23A, # G+13797  U+C23A  uniC23A,  HANGUL SYLLABLE-C23A
        0xC23B, # G+13798  U+C23B  uniC23B,  HANGUL SYLLABLE-C23B
        0xC23C, # G+13799  U+C23C  uniC23C,  HANGUL SYLLABLE-C23C
        0xC23D, # G+13800  U+C23D  uniC23D,  HANGUL SYLLABLE-C23D
        0xC23E, # G+13801  U+C23E  uniC23E,  HANGUL SYLLABLE-C23E
        0xC23F, # G+13802  U+C23F  uniC23F,  HANGUL SYLLABLE-C23F
        0xC240, # G+13803  U+C240  uniC240,  HANGUL SYLLABLE-C240
        0xC241, # G+13804  U+C241  uniC241,  HANGUL SYLLABLE-C241
        0xC242, # G+13805  U+C242  uniC242,  HANGUL SYLLABLE-C242
        0xC243, # G+13806  U+C243  uniC243,  HANGUL SYLLABLE-C243
        0xC244, # G+13807  U+C244  uniC244,  HANGUL SYLLABLE-C244
        0xC245, # G+13808  U+C245  uniC245,  HANGUL SYLLABLE-C245
        0xC246, # G+13809  U+C246  uniC246,  HANGUL SYLLABLE-C246
        0xC247, # G+13810  U+C247  uniC247,  HANGUL SYLLABLE-C247
        0xC249, # G+13811  U+C249  uniC249,  HANGUL SYLLABLE-C249
        0xC24A, # G+13812  U+C24A  uniC24A,  HANGUL SYLLABLE-C24A
        0xC24B, # G+13813  U+C24B  uniC24B,  HANGUL SYLLABLE-C24B
        0xC24C, # G+13814  U+C24C  uniC24C,  HANGUL SYLLABLE-C24C
        0xC24D, # G+13815  U+C24D  uniC24D,  HANGUL SYLLABLE-C24D
        0xC24E, # G+13816  U+C24E  uniC24E,  HANGUL SYLLABLE-C24E
        0xC24F, # G+13817  U+C24F  uniC24F,  HANGUL SYLLABLE-C24F
        0xC252, # G+13818  U+C252  uniC252,  HANGUL SYLLABLE-C252
        0xC253, # G+13819  U+C253  uniC253,  HANGUL SYLLABLE-C253
        0xC255, # G+13820  U+C255  uniC255,  HANGUL SYLLABLE-C255
        0xC256, # G+13821  U+C256  uniC256,  HANGUL SYLLABLE-C256
        0xC257, # G+13822  U+C257  uniC257,  HANGUL SYLLABLE-C257
        0xC259, # G+13823  U+C259  uniC259,  HANGUL SYLLABLE-C259
        0xC25A, # G+13824  U+C25A  uniC25A,  HANGUL SYLLABLE-C25A
        0xC25B, # G+13825  U+C25B  uniC25B,  HANGUL SYLLABLE-C25B
        0xC25C, # G+13826  U+C25C  uniC25C,  HANGUL SYLLABLE-C25C
        0xC25D, # G+13827  U+C25D  uniC25D,  HANGUL SYLLABLE-C25D
        0xC25E, # G+13828  U+C25E  uniC25E,  HANGUL SYLLABLE-C25E
        0xC25F, # G+13829  U+C25F  uniC25F,  HANGUL SYLLABLE-C25F
        0xC261, # G+13830  U+C261  uniC261,  HANGUL SYLLABLE-C261
        0xC262, # G+13831  U+C262  uniC262,  HANGUL SYLLABLE-C262
        0xC263, # G+13832  U+C263  uniC263,  HANGUL SYLLABLE-C263
        0xC264, # G+13833  U+C264  uniC264,  HANGUL SYLLABLE-C264
        0xC266, # G+13834  U+C266  uniC266,  HANGUL SYLLABLE-C266
        0xC267, # G+13835  U+C267  uniC267,  HANGUL SYLLABLE-C267
        0xC268, # G+13836  U+C268  uniC268,  HANGUL SYLLABLE-C268
        0xC269, # G+13837  U+C269  uniC269,  HANGUL SYLLABLE-C269
        0xC26A, # G+13838  U+C26A  uniC26A,  HANGUL SYLLABLE-C26A
        0xC26B, # G+13839  U+C26B  uniC26B,  HANGUL SYLLABLE-C26B
        0xC26E, # G+13840  U+C26E  uniC26E,  HANGUL SYLLABLE-C26E
        0xC26F, # G+13841  U+C26F  uniC26F,  HANGUL SYLLABLE-C26F
        0xC271, # G+13842  U+C271  uniC271,  HANGUL SYLLABLE-C271
        0xC272, # G+13843  U+C272  uniC272,  HANGUL SYLLABLE-C272
        0xC273, # G+13844  U+C273  uniC273,  HANGUL SYLLABLE-C273
        0xC275, # G+13845  U+C275  uniC275,  HANGUL SYLLABLE-C275
        0xC276, # G+13846  U+C276  uniC276,  HANGUL SYLLABLE-C276
        0xC277, # G+13847  U+C277  uniC277,  HANGUL SYLLABLE-C277
        0xC278, # G+13848  U+C278  uniC278,  HANGUL SYLLABLE-C278
        0xC279, # G+13849  U+C279  uniC279,  HANGUL SYLLABLE-C279
        0xC27A, # G+13850  U+C27A  uniC27A,  HANGUL SYLLABLE-C27A
        0xC27B, # G+13851  U+C27B  uniC27B,  HANGUL SYLLABLE-C27B
        0xC27E, # G+13852  U+C27E  uniC27E,  HANGUL SYLLABLE-C27E
        0xC280, # G+13853  U+C280  uniC280,  HANGUL SYLLABLE-C280
        0xC282, # G+13854  U+C282  uniC282,  HANGUL SYLLABLE-C282
        0xC283, # G+13855  U+C283  uniC283,  HANGUL SYLLABLE-C283
        0xC284, # G+13856  U+C284  uniC284,  HANGUL SYLLABLE-C284
        0xC285, # G+13857  U+C285  uniC285,  HANGUL SYLLABLE-C285
        0xC286, # G+13858  U+C286  uniC286,  HANGUL SYLLABLE-C286
        0xC287, # G+13859  U+C287  uniC287,  HANGUL SYLLABLE-C287
        0xC28A, # G+13860  U+C28A  uniC28A,  HANGUL SYLLABLE-C28A
        0xC28B, # G+13861  U+C28B  uniC28B,  HANGUL SYLLABLE-C28B
        0xC28C, # G+13862  U+C28C  uniC28C,  HANGUL SYLLABLE-C28C
        0xC28D, # G+13863  U+C28D  uniC28D,  HANGUL SYLLABLE-C28D
        0xC28E, # G+13864  U+C28E  uniC28E,  HANGUL SYLLABLE-C28E
        0xC28F, # G+13865  U+C28F  uniC28F,  HANGUL SYLLABLE-C28F
        0xC291, # G+13866  U+C291  uniC291,  HANGUL SYLLABLE-C291
        0xC292, # G+13867  U+C292  uniC292,  HANGUL SYLLABLE-C292
        0xC293, # G+13868  U+C293  uniC293,  HANGUL SYLLABLE-C293
        0xC294, # G+13869  U+C294  uniC294,  HANGUL SYLLABLE-C294
        0xC295, # G+13870  U+C295  uniC295,  HANGUL SYLLABLE-C295
        0xC296, # G+13871  U+C296  uniC296,  HANGUL SYLLABLE-C296
        0xC297, # G+13872  U+C297  uniC297,  HANGUL SYLLABLE-C297
        0xC299, # G+13873  U+C299  uniC299,  HANGUL SYLLABLE-C299
        0xC29A, # G+13874  U+C29A  uniC29A,  HANGUL SYLLABLE-C29A
        0xC29C, # G+13875  U+C29C  uniC29C,  HANGUL SYLLABLE-C29C
        0xC29E, # G+13876  U+C29E  uniC29E,  HANGUL SYLLABLE-C29E
        0xC29F, # G+13877  U+C29F  uniC29F,  HANGUL SYLLABLE-C29F
        0xC2A0, # G+13878  U+C2A0  uniC2A0,  HANGUL SYLLABLE-C2A0
        0xC2A1, # G+13879  U+C2A1  uniC2A1,  HANGUL SYLLABLE-C2A1
        0xC2A2, # G+13880  U+C2A2  uniC2A2,  HANGUL SYLLABLE-C2A2
        0xC2A3, # G+13881  U+C2A3  uniC2A3,  HANGUL SYLLABLE-C2A3
        0xC2A6, # G+13882  U+C2A6  uniC2A6,  HANGUL SYLLABLE-C2A6
        0xC2A7, # G+13883  U+C2A7  uniC2A7,  HANGUL SYLLABLE-C2A7
        0xC2A9, # G+13884  U+C2A9  uniC2A9,  HANGUL SYLLABLE-C2A9
        0xC2AA, # G+13885  U+C2AA  uniC2AA,  HANGUL SYLLABLE-C2AA
        0xC2AB, # G+13886  U+C2AB  uniC2AB,  HANGUL SYLLABLE-C2AB
        0xC2AE, # G+13887  U+C2AE  uniC2AE,  HANGUL SYLLABLE-C2AE
        0xC2AF, # G+13888  U+C2AF  uniC2AF,  HANGUL SYLLABLE-C2AF
        0xC2B0, # G+13889  U+C2B0  uniC2B0,  HANGUL SYLLABLE-C2B0
        0xC2B1, # G+13890  U+C2B1  uniC2B1,  HANGUL SYLLABLE-C2B1
        0xC2B2, # G+13891  U+C2B2  uniC2B2,  HANGUL SYLLABLE-C2B2
        0xC2B3, # G+13892  U+C2B3  uniC2B3,  HANGUL SYLLABLE-C2B3
        0xC2B6, # G+13893  U+C2B6  uniC2B6,  HANGUL SYLLABLE-C2B6
        0xC2B8, # G+13894  U+C2B8  uniC2B8,  HANGUL SYLLABLE-C2B8
        0xC2BA, # G+13895  U+C2BA  uniC2BA,  HANGUL SYLLABLE-C2BA
        0xC2BB, # G+13896  U+C2BB  uniC2BB,  HANGUL SYLLABLE-C2BB
        0xC2BC, # G+13897  U+C2BC  uniC2BC,  HANGUL SYLLABLE-C2BC
        0xC2BD, # G+13898  U+C2BD  uniC2BD,  HANGUL SYLLABLE-C2BD
        0xC2BE, # G+13899  U+C2BE  uniC2BE,  HANGUL SYLLABLE-C2BE
        0xC2BF, # G+13900  U+C2BF  uniC2BF,  HANGUL SYLLABLE-C2BF
        0xC2C0, # G+13901  U+C2C0  uniC2C0,  HANGUL SYLLABLE-C2C0
        0xC2C1, # G+13902  U+C2C1  uniC2C1,  HANGUL SYLLABLE-C2C1
        0xC2C2, # G+13903  U+C2C2  uniC2C2,  HANGUL SYLLABLE-C2C2
        0xC2C3, # G+13904  U+C2C3  uniC2C3,  HANGUL SYLLABLE-C2C3
        0xC2C4, # G+13905  U+C2C4  uniC2C4,  HANGUL SYLLABLE-C2C4
        0xC2C5, # G+13906  U+C2C5  uniC2C5,  HANGUL SYLLABLE-C2C5
        0xC2C6, # G+13907  U+C2C6  uniC2C6,  HANGUL SYLLABLE-C2C6
        0xC2C7, # G+13908  U+C2C7  uniC2C7,  HANGUL SYLLABLE-C2C7
        0xC2C8, # G+13909  U+C2C8  uniC2C8,  HANGUL SYLLABLE-C2C8
        0xC2C9, # G+13910  U+C2C9  uniC2C9,  HANGUL SYLLABLE-C2C9
        0xC2CA, # G+13911  U+C2CA  uniC2CA,  HANGUL SYLLABLE-C2CA
        0xC2CB, # G+13912  U+C2CB  uniC2CB,  HANGUL SYLLABLE-C2CB
        0xC2CC, # G+13913  U+C2CC  uniC2CC,  HANGUL SYLLABLE-C2CC
        0xC2CD, # G+13914  U+C2CD  uniC2CD,  HANGUL SYLLABLE-C2CD
        0xC2CE, # G+13915  U+C2CE  uniC2CE,  HANGUL SYLLABLE-C2CE
        0xC2CF, # G+13916  U+C2CF  uniC2CF,  HANGUL SYLLABLE-C2CF
        0xC2D0, # G+13917  U+C2D0  uniC2D0,  HANGUL SYLLABLE-C2D0
        0xC2D1, # G+13918  U+C2D1  uniC2D1,  HANGUL SYLLABLE-C2D1
        0xC2D2, # G+13919  U+C2D2  uniC2D2,  HANGUL SYLLABLE-C2D2
        0xC2D3, # G+13920  U+C2D3  uniC2D3,  HANGUL SYLLABLE-C2D3
        0xC2D4, # G+13921  U+C2D4  uniC2D4,  HANGUL SYLLABLE-C2D4
        0xC2D5, # G+13922  U+C2D5  uniC2D5,  HANGUL SYLLABLE-C2D5
        0xC2D6, # G+13923  U+C2D6  uniC2D6,  HANGUL SYLLABLE-C2D6
        0xC2D7, # G+13924  U+C2D7  uniC2D7,  HANGUL SYLLABLE-C2D7
        0xC2D8, # G+13925  U+C2D8  uniC2D8,  HANGUL SYLLABLE-C2D8
        0xC2D9, # G+13926  U+C2D9  uniC2D9,  HANGUL SYLLABLE-C2D9
        0xC2DA, # G+13927  U+C2DA  uniC2DA,  HANGUL SYLLABLE-C2DA
        0xC2DB, # G+13928  U+C2DB  uniC2DB,  HANGUL SYLLABLE-C2DB
        0xC2DE, # G+13929  U+C2DE  uniC2DE,  HANGUL SYLLABLE-C2DE
        0xC2DF, # G+13930  U+C2DF  uniC2DF,  HANGUL SYLLABLE-C2DF
        0xC2E1, # G+13931  U+C2E1  uniC2E1,  HANGUL SYLLABLE-C2E1
        0xC2E2, # G+13932  U+C2E2  uniC2E2,  HANGUL SYLLABLE-C2E2
        0xC2E5, # G+13933  U+C2E5  uniC2E5,  HANGUL SYLLABLE-C2E5
        0xC2E6, # G+13934  U+C2E6  uniC2E6,  HANGUL SYLLABLE-C2E6
        0xC2E7, # G+13935  U+C2E7  uniC2E7,  HANGUL SYLLABLE-C2E7
        0xC2E8, # G+13936  U+C2E8  uniC2E8,  HANGUL SYLLABLE-C2E8
        0xC2E9, # G+13937  U+C2E9  uniC2E9,  HANGUL SYLLABLE-C2E9
        0xC2EA, # G+13938  U+C2EA  uniC2EA,  HANGUL SYLLABLE-C2EA
        0xC2EE, # G+13939  U+C2EE  uniC2EE,  HANGUL SYLLABLE-C2EE
        0xC2F0, # G+13940  U+C2F0  uniC2F0,  HANGUL SYLLABLE-C2F0
        0xC2F2, # G+13941  U+C2F2  uniC2F2,  HANGUL SYLLABLE-C2F2
        0xC2F3, # G+13942  U+C2F3  uniC2F3,  HANGUL SYLLABLE-C2F3
        0xC2F4, # G+13943  U+C2F4  uniC2F4,  HANGUL SYLLABLE-C2F4
        0xC2F5, # G+13944  U+C2F5  uniC2F5,  HANGUL SYLLABLE-C2F5
        0xC2F7, # G+13945  U+C2F7  uniC2F7,  HANGUL SYLLABLE-C2F7
        0xC2FA, # G+13946  U+C2FA  uniC2FA,  HANGUL SYLLABLE-C2FA
        0xC2FD, # G+13947  U+C2FD  uniC2FD,  HANGUL SYLLABLE-C2FD
        0xC2FE, # G+13948  U+C2FE  uniC2FE,  HANGUL SYLLABLE-C2FE
        0xC2FF, # G+13949  U+C2FF  uniC2FF,  HANGUL SYLLABLE-C2FF
        0xC301, # G+13950  U+C301  uniC301,  HANGUL SYLLABLE-C301
        0xC302, # G+13951  U+C302  uniC302,  HANGUL SYLLABLE-C302
        0xC303, # G+13952  U+C303  uniC303,  HANGUL SYLLABLE-C303
        0xC304, # G+13953  U+C304  uniC304,  HANGUL SYLLABLE-C304
        0xC305, # G+13954  U+C305  uniC305,  HANGUL SYLLABLE-C305
        0xC306, # G+13955  U+C306  uniC306,  HANGUL SYLLABLE-C306
        0xC307, # G+13956  U+C307  uniC307,  HANGUL SYLLABLE-C307
        0xC30A, # G+13957  U+C30A  uniC30A,  HANGUL SYLLABLE-C30A
        0xC30B, # G+13958  U+C30B  uniC30B,  HANGUL SYLLABLE-C30B
        0xC30E, # G+13959  U+C30E  uniC30E,  HANGUL SYLLABLE-C30E
        0xC30F, # G+13960  U+C30F  uniC30F,  HANGUL SYLLABLE-C30F
        0xC310, # G+13961  U+C310  uniC310,  HANGUL SYLLABLE-C310
        0xC311, # G+13962  U+C311  uniC311,  HANGUL SYLLABLE-C311
        0xC312, # G+13963  U+C312  uniC312,  HANGUL SYLLABLE-C312
        0xC316, # G+13964  U+C316  uniC316,  HANGUL SYLLABLE-C316
        0xC317, # G+13965  U+C317  uniC317,  HANGUL SYLLABLE-C317
        0xC319, # G+13966  U+C319  uniC319,  HANGUL SYLLABLE-C319
        0xC31A, # G+13967  U+C31A  uniC31A,  HANGUL SYLLABLE-C31A
        0xC31B, # G+13968  U+C31B  uniC31B,  HANGUL SYLLABLE-C31B
        0xC31D, # G+13969  U+C31D  uniC31D,  HANGUL SYLLABLE-C31D
        0xC31E, # G+13970  U+C31E  uniC31E,  HANGUL SYLLABLE-C31E
        0xC31F, # G+13971  U+C31F  uniC31F,  HANGUL SYLLABLE-C31F
        0xC320, # G+13972  U+C320  uniC320,  HANGUL SYLLABLE-C320
        0xC321, # G+13973  U+C321  uniC321,  HANGUL SYLLABLE-C321
        0xC322, # G+13974  U+C322  uniC322,  HANGUL SYLLABLE-C322
        0xC323, # G+13975  U+C323  uniC323,  HANGUL SYLLABLE-C323
        0xC326, # G+13976  U+C326  uniC326,  HANGUL SYLLABLE-C326
        0xC327, # G+13977  U+C327  uniC327,  HANGUL SYLLABLE-C327
        0xC32A, # G+13978  U+C32A  uniC32A,  HANGUL SYLLABLE-C32A
        0xC32B, # G+13979  U+C32B  uniC32B,  HANGUL SYLLABLE-C32B
        0xC32C, # G+13980  U+C32C  uniC32C,  HANGUL SYLLABLE-C32C
        0xC32D, # G+13981  U+C32D  uniC32D,  HANGUL SYLLABLE-C32D
        0xC32E, # G+13982  U+C32E  uniC32E,  HANGUL SYLLABLE-C32E
        0xC32F, # G+13983  U+C32F  uniC32F,  HANGUL SYLLABLE-C32F
        0xC330, # G+13984  U+C330  uniC330,  HANGUL SYLLABLE-C330
        0xC331, # G+13985  U+C331  uniC331,  HANGUL SYLLABLE-C331
        0xC332, # G+13986  U+C332  uniC332,  HANGUL SYLLABLE-C332
        0xC333, # G+13987  U+C333  uniC333,  HANGUL SYLLABLE-C333
        0xC334, # G+13988  U+C334  uniC334,  HANGUL SYLLABLE-C334
        0xC335, # G+13989  U+C335  uniC335,  HANGUL SYLLABLE-C335
        0xC336, # G+13990  U+C336  uniC336,  HANGUL SYLLABLE-C336
        0xC337, # G+13991  U+C337  uniC337,  HANGUL SYLLABLE-C337
        0xC338, # G+13992  U+C338  uniC338,  HANGUL SYLLABLE-C338
        0xC339, # G+13993  U+C339  uniC339,  HANGUL SYLLABLE-C339
        0xC33A, # G+13994  U+C33A  uniC33A,  HANGUL SYLLABLE-C33A
        0xC33B, # G+13995  U+C33B  uniC33B,  HANGUL SYLLABLE-C33B
        0xC33C, # G+13996  U+C33C  uniC33C,  HANGUL SYLLABLE-C33C
        0xC33D, # G+13997  U+C33D  uniC33D,  HANGUL SYLLABLE-C33D
        0xC33E, # G+13998  U+C33E  uniC33E,  HANGUL SYLLABLE-C33E
        0xC33F, # G+13999  U+C33F  uniC33F,  HANGUL SYLLABLE-C33F
        0xC340, # G+14000  U+C340  uniC340,  HANGUL SYLLABLE-C340
        0xC341, # G+14001  U+C341  uniC341,  HANGUL SYLLABLE-C341
        0xC342, # G+14002  U+C342  uniC342,  HANGUL SYLLABLE-C342
        0xC343, # G+14003  U+C343  uniC343,  HANGUL SYLLABLE-C343
        0xC344, # G+14004  U+C344  uniC344,  HANGUL SYLLABLE-C344
        0xC346, # G+14005  U+C346  uniC346,  HANGUL SYLLABLE-C346
        0xC347, # G+14006  U+C347  uniC347,  HANGUL SYLLABLE-C347
        0xC348, # G+14007  U+C348  uniC348,  HANGUL SYLLABLE-C348
        0xC349, # G+14008  U+C349  uniC349,  HANGUL SYLLABLE-C349
        0xC34A, # G+14009  U+C34A  uniC34A,  HANGUL SYLLABLE-C34A
        0xC34B, # G+14010  U+C34B  uniC34B,  HANGUL SYLLABLE-C34B
        0xC34C, # G+14011  U+C34C  uniC34C,  HANGUL SYLLABLE-C34C
        0xC34D, # G+14012  U+C34D  uniC34D,  HANGUL SYLLABLE-C34D
        0xC34E, # G+14013  U+C34E  uniC34E,  HANGUL SYLLABLE-C34E
        0xC34F, # G+14014  U+C34F  uniC34F,  HANGUL SYLLABLE-C34F
        0xC350, # G+14015  U+C350  uniC350,  HANGUL SYLLABLE-C350
        0xC351, # G+14016  U+C351  uniC351,  HANGUL SYLLABLE-C351
        0xC352, # G+14017  U+C352  uniC352,  HANGUL SYLLABLE-C352
        0xC353, # G+14018  U+C353  uniC353,  HANGUL SYLLABLE-C353
        0xC354, # G+14019  U+C354  uniC354,  HANGUL SYLLABLE-C354
        0xC355, # G+14020  U+C355  uniC355,  HANGUL SYLLABLE-C355
        0xC356, # G+14021  U+C356  uniC356,  HANGUL SYLLABLE-C356
        0xC357, # G+14022  U+C357  uniC357,  HANGUL SYLLABLE-C357
        0xC358, # G+14023  U+C358  uniC358,  HANGUL SYLLABLE-C358
        0xC359, # G+14024  U+C359  uniC359,  HANGUL SYLLABLE-C359
        0xC35A, # G+14025  U+C35A  uniC35A,  HANGUL SYLLABLE-C35A
        0xC35B, # G+14026  U+C35B  uniC35B,  HANGUL SYLLABLE-C35B
        0xC35C, # G+14027  U+C35C  uniC35C,  HANGUL SYLLABLE-C35C
        0xC35D, # G+14028  U+C35D  uniC35D,  HANGUL SYLLABLE-C35D
        0xC35E, # G+14029  U+C35E  uniC35E,  HANGUL SYLLABLE-C35E
        0xC35F, # G+14030  U+C35F  uniC35F,  HANGUL SYLLABLE-C35F
        0xC360, # G+14031  U+C360  uniC360,  HANGUL SYLLABLE-C360
        0xC361, # G+14032  U+C361  uniC361,  HANGUL SYLLABLE-C361
        0xC362, # G+14033  U+C362  uniC362,  HANGUL SYLLABLE-C362
        0xC363, # G+14034  U+C363  uniC363,  HANGUL SYLLABLE-C363
        0xC364, # G+14035  U+C364  uniC364,  HANGUL SYLLABLE-C364
        0xC365, # G+14036  U+C365  uniC365,  HANGUL SYLLABLE-C365
        0xC366, # G+14037  U+C366  uniC366,  HANGUL SYLLABLE-C366
        0xC367, # G+14038  U+C367  uniC367,  HANGUL SYLLABLE-C367
        0xC36A, # G+14039  U+C36A  uniC36A,  HANGUL SYLLABLE-C36A
        0xC36B, # G+14040  U+C36B  uniC36B,  HANGUL SYLLABLE-C36B
        0xC36D, # G+14041  U+C36D  uniC36D,  HANGUL SYLLABLE-C36D
        0xC36E, # G+14042  U+C36E  uniC36E,  HANGUL SYLLABLE-C36E
        0xC36F, # G+14043  U+C36F  uniC36F,  HANGUL SYLLABLE-C36F
        0xC371, # G+14044  U+C371  uniC371,  HANGUL SYLLABLE-C371
        0xC373, # G+14045  U+C373  uniC373,  HANGUL SYLLABLE-C373
        0xC374, # G+14046  U+C374  uniC374,  HANGUL SYLLABLE-C374
        0xC375, # G+14047  U+C375  uniC375,  HANGUL SYLLABLE-C375
        0xC376, # G+14048  U+C376  uniC376,  HANGUL SYLLABLE-C376
        0xC377, # G+14049  U+C377  uniC377,  HANGUL SYLLABLE-C377
        0xC37A, # G+14050  U+C37A  uniC37A,  HANGUL SYLLABLE-C37A
        0xC37B, # G+14051  U+C37B  uniC37B,  HANGUL SYLLABLE-C37B
        0xC37E, # G+14052  U+C37E  uniC37E,  HANGUL SYLLABLE-C37E
        0xC37F, # G+14053  U+C37F  uniC37F,  HANGUL SYLLABLE-C37F
        0xC380, # G+14054  U+C380  uniC380,  HANGUL SYLLABLE-C380
        0xC381, # G+14055  U+C381  uniC381,  HANGUL SYLLABLE-C381
        0xC382, # G+14056  U+C382  uniC382,  HANGUL SYLLABLE-C382
        0xC383, # G+14057  U+C383  uniC383,  HANGUL SYLLABLE-C383
        0xC385, # G+14058  U+C385  uniC385,  HANGUL SYLLABLE-C385
        0xC386, # G+14059  U+C386  uniC386,  HANGUL SYLLABLE-C386
        0xC387, # G+14060  U+C387  uniC387,  HANGUL SYLLABLE-C387
        0xC389, # G+14061  U+C389  uniC389,  HANGUL SYLLABLE-C389
        0xC38A, # G+14062  U+C38A  uniC38A,  HANGUL SYLLABLE-C38A
        0xC38B, # G+14063  U+C38B  uniC38B,  HANGUL SYLLABLE-C38B
        0xC38D, # G+14064  U+C38D  uniC38D,  HANGUL SYLLABLE-C38D
        0xC38E, # G+14065  U+C38E  uniC38E,  HANGUL SYLLABLE-C38E
        0xC38F, # G+14066  U+C38F  uniC38F,  HANGUL SYLLABLE-C38F
        0xC390, # G+14067  U+C390  uniC390,  HANGUL SYLLABLE-C390
        0xC391, # G+14068  U+C391  uniC391,  HANGUL SYLLABLE-C391
        0xC392, # G+14069  U+C392  uniC392,  HANGUL SYLLABLE-C392
        0xC393, # G+14070  U+C393  uniC393,  HANGUL SYLLABLE-C393
        0xC394, # G+14071  U+C394  uniC394,  HANGUL SYLLABLE-C394
        0xC395, # G+14072  U+C395  uniC395,  HANGUL SYLLABLE-C395
        0xC396, # G+14073  U+C396  uniC396,  HANGUL SYLLABLE-C396
        0xC397, # G+14074  U+C397  uniC397,  HANGUL SYLLABLE-C397
        0xC398, # G+14075  U+C398  uniC398,  HANGUL SYLLABLE-C398
        0xC399, # G+14076  U+C399  uniC399,  HANGUL SYLLABLE-C399
        0xC39A, # G+14077  U+C39A  uniC39A,  HANGUL SYLLABLE-C39A
        0xC39B, # G+14078  U+C39B  uniC39B,  HANGUL SYLLABLE-C39B
        0xC39C, # G+14079  U+C39C  uniC39C,  HANGUL SYLLABLE-C39C
        0xC39D, # G+14080  U+C39D  uniC39D,  HANGUL SYLLABLE-C39D
        0xC39E, # G+14081  U+C39E  uniC39E,  HANGUL SYLLABLE-C39E
        0xC39F, # G+14082  U+C39F  uniC39F,  HANGUL SYLLABLE-C39F
        0xC3A0, # G+14083  U+C3A0  uniC3A0,  HANGUL SYLLABLE-C3A0
        0xC3A1, # G+14084  U+C3A1  uniC3A1,  HANGUL SYLLABLE-C3A1
        0xC3A2, # G+14085  U+C3A2  uniC3A2,  HANGUL SYLLABLE-C3A2
        0xC3A3, # G+14086  U+C3A3  uniC3A3,  HANGUL SYLLABLE-C3A3
        0xC3A4, # G+14087  U+C3A4  uniC3A4,  HANGUL SYLLABLE-C3A4
        0xC3A5, # G+14088  U+C3A5  uniC3A5,  HANGUL SYLLABLE-C3A5
        0xC3A6, # G+14089  U+C3A6  uniC3A6,  HANGUL SYLLABLE-C3A6
        0xC3A7, # G+14090  U+C3A7  uniC3A7,  HANGUL SYLLABLE-C3A7
        0xC3A8, # G+14091  U+C3A8  uniC3A8,  HANGUL SYLLABLE-C3A8
        0xC3A9, # G+14092  U+C3A9  uniC3A9,  HANGUL SYLLABLE-C3A9
        0xC3AA, # G+14093  U+C3AA  uniC3AA,  HANGUL SYLLABLE-C3AA
        0xC3AB, # G+14094  U+C3AB  uniC3AB,  HANGUL SYLLABLE-C3AB
        0xC3AC, # G+14095  U+C3AC  uniC3AC,  HANGUL SYLLABLE-C3AC
        0xC3AD, # G+14096  U+C3AD  uniC3AD,  HANGUL SYLLABLE-C3AD
        0xC3AE, # G+14097  U+C3AE  uniC3AE,  HANGUL SYLLABLE-C3AE
        0xC3AF, # G+14098  U+C3AF  uniC3AF,  HANGUL SYLLABLE-C3AF
        0xC3B0, # G+14099  U+C3B0  uniC3B0,  HANGUL SYLLABLE-C3B0
        0xC3B1, # G+14100  U+C3B1  uniC3B1,  HANGUL SYLLABLE-C3B1
        0xC3B2, # G+14101  U+C3B2  uniC3B2,  HANGUL SYLLABLE-C3B2
        0xC3B3, # G+14102  U+C3B3  uniC3B3,  HANGUL SYLLABLE-C3B3
        0xC3B4, # G+14103  U+C3B4  uniC3B4,  HANGUL SYLLABLE-C3B4
        0xC3B5, # G+14104  U+C3B5  uniC3B5,  HANGUL SYLLABLE-C3B5
        0xC3B6, # G+14105  U+C3B6  uniC3B6,  HANGUL SYLLABLE-C3B6
        0xC3B7, # G+14106  U+C3B7  uniC3B7,  HANGUL SYLLABLE-C3B7
        0xC3B8, # G+14107  U+C3B8  uniC3B8,  HANGUL SYLLABLE-C3B8
        0xC3B9, # G+14108  U+C3B9  uniC3B9,  HANGUL SYLLABLE-C3B9
        0xC3BA, # G+14109  U+C3BA  uniC3BA,  HANGUL SYLLABLE-C3BA
        0xC3BB, # G+14110  U+C3BB  uniC3BB,  HANGUL SYLLABLE-C3BB
        0xC3BC, # G+14111  U+C3BC  uniC3BC,  HANGUL SYLLABLE-C3BC
        0xC3BD, # G+14112  U+C3BD  uniC3BD,  HANGUL SYLLABLE-C3BD
        0xC3BE, # G+14113  U+C3BE  uniC3BE,  HANGUL SYLLABLE-C3BE
        0xC3BF, # G+14114  U+C3BF  uniC3BF,  HANGUL SYLLABLE-C3BF
        0xC3C1, # G+14115  U+C3C1  uniC3C1,  HANGUL SYLLABLE-C3C1
        0xC3C2, # G+14116  U+C3C2  uniC3C2,  HANGUL SYLLABLE-C3C2
        0xC3C3, # G+14117  U+C3C3  uniC3C3,  HANGUL SYLLABLE-C3C3
        0xC3C4, # G+14118  U+C3C4  uniC3C4,  HANGUL SYLLABLE-C3C4
        0xC3C5, # G+14119  U+C3C5  uniC3C5,  HANGUL SYLLABLE-C3C5
        0xC3C6, # G+14120  U+C3C6  uniC3C6,  HANGUL SYLLABLE-C3C6
        0xC3C7, # G+14121  U+C3C7  uniC3C7,  HANGUL SYLLABLE-C3C7
        0xC3C8, # G+14122  U+C3C8  uniC3C8,  HANGUL SYLLABLE-C3C8
        0xC3C9, # G+14123  U+C3C9  uniC3C9,  HANGUL SYLLABLE-C3C9
        0xC3CA, # G+14124  U+C3CA  uniC3CA,  HANGUL SYLLABLE-C3CA
        0xC3CB, # G+14125  U+C3CB  uniC3CB,  HANGUL SYLLABLE-C3CB
        0xC3CC, # G+14126  U+C3CC  uniC3CC,  HANGUL SYLLABLE-C3CC
        0xC3CD, # G+14127  U+C3CD  uniC3CD,  HANGUL SYLLABLE-C3CD
        0xC3CE, # G+14128  U+C3CE  uniC3CE,  HANGUL SYLLABLE-C3CE
        0xC3CF, # G+14129  U+C3CF  uniC3CF,  HANGUL SYLLABLE-C3CF
        0xC3D0, # G+14130  U+C3D0  uniC3D0,  HANGUL SYLLABLE-C3D0
        0xC3D1, # G+14131  U+C3D1  uniC3D1,  HANGUL SYLLABLE-C3D1
        0xC3D2, # G+14132  U+C3D2  uniC3D2,  HANGUL SYLLABLE-C3D2
        0xC3D3, # G+14133  U+C3D3  uniC3D3,  HANGUL SYLLABLE-C3D3
        0xC3D4, # G+14134  U+C3D4  uniC3D4,  HANGUL SYLLABLE-C3D4
        0xC3D5, # G+14135  U+C3D5  uniC3D5,  HANGUL SYLLABLE-C3D5
        0xC3D6, # G+14136  U+C3D6  uniC3D6,  HANGUL SYLLABLE-C3D6
        0xC3D7, # G+14137  U+C3D7  uniC3D7,  HANGUL SYLLABLE-C3D7
        0xC3DA, # G+14138  U+C3DA  uniC3DA,  HANGUL SYLLABLE-C3DA
        0xC3DB, # G+14139  U+C3DB  uniC3DB,  HANGUL SYLLABLE-C3DB
        0xC3DD, # G+14140  U+C3DD  uniC3DD,  HANGUL SYLLABLE-C3DD
        0xC3DE, # G+14141  U+C3DE  uniC3DE,  HANGUL SYLLABLE-C3DE
        0xC3E1, # G+14142  U+C3E1  uniC3E1,  HANGUL SYLLABLE-C3E1
        0xC3E3, # G+14143  U+C3E3  uniC3E3,  HANGUL SYLLABLE-C3E3
        0xC3E4, # G+14144  U+C3E4  uniC3E4,  HANGUL SYLLABLE-C3E4
        0xC3E5, # G+14145  U+C3E5  uniC3E5,  HANGUL SYLLABLE-C3E5
        0xC3E6, # G+14146  U+C3E6  uniC3E6,  HANGUL SYLLABLE-C3E6
        0xC3E7, # G+14147  U+C3E7  uniC3E7,  HANGUL SYLLABLE-C3E7
        0xC3EA, # G+14148  U+C3EA  uniC3EA,  HANGUL SYLLABLE-C3EA
        0xC3EB, # G+14149  U+C3EB  uniC3EB,  HANGUL SYLLABLE-C3EB
        0xC3EC, # G+14150  U+C3EC  uniC3EC,  HANGUL SYLLABLE-C3EC
        0xC3EE, # G+14151  U+C3EE  uniC3EE,  HANGUL SYLLABLE-C3EE
        0xC3EF, # G+14152  U+C3EF  uniC3EF,  HANGUL SYLLABLE-C3EF
        0xC3F0, # G+14153  U+C3F0  uniC3F0,  HANGUL SYLLABLE-C3F0
        0xC3F1, # G+14154  U+C3F1  uniC3F1,  HANGUL SYLLABLE-C3F1
        0xC3F2, # G+14155  U+C3F2  uniC3F2,  HANGUL SYLLABLE-C3F2
        0xC3F3, # G+14156  U+C3F3  uniC3F3,  HANGUL SYLLABLE-C3F3
        0xC3F6, # G+14157  U+C3F6  uniC3F6,  HANGUL SYLLABLE-C3F6
        0xC3F7, # G+14158  U+C3F7  uniC3F7,  HANGUL SYLLABLE-C3F7
        0xC3F9, # G+14159  U+C3F9  uniC3F9,  HANGUL SYLLABLE-C3F9
        0xC3FA, # G+14160  U+C3FA  uniC3FA,  HANGUL SYLLABLE-C3FA
        0xC3FB, # G+14161  U+C3FB  uniC3FB,  HANGUL SYLLABLE-C3FB
        0xC3FC, # G+14162  U+C3FC  uniC3FC,  HANGUL SYLLABLE-C3FC
        0xC3FD, # G+14163  U+C3FD  uniC3FD,  HANGUL SYLLABLE-C3FD
        0xC3FE, # G+14164  U+C3FE  uniC3FE,  HANGUL SYLLABLE-C3FE
        0xC3FF, # G+14165  U+C3FF  uniC3FF,  HANGUL SYLLABLE-C3FF
        0xC400, # G+14166  U+C400  uniC400,  HANGUL SYLLABLE-C400
        0xC401, # G+14167  U+C401  uniC401,  HANGUL SYLLABLE-C401
        0xC402, # G+14168  U+C402  uniC402,  HANGUL SYLLABLE-C402
        0xC403, # G+14169  U+C403  uniC403,  HANGUL SYLLABLE-C403
        0xC404, # G+14170  U+C404  uniC404,  HANGUL SYLLABLE-C404
        0xC405, # G+14171  U+C405  uniC405,  HANGUL SYLLABLE-C405
        0xC406, # G+14172  U+C406  uniC406,  HANGUL SYLLABLE-C406
        0xC407, # G+14173  U+C407  uniC407,  HANGUL SYLLABLE-C407
        0xC409, # G+14174  U+C409  uniC409,  HANGUL SYLLABLE-C409
        0xC40A, # G+14175  U+C40A  uniC40A,  HANGUL SYLLABLE-C40A
        0xC40B, # G+14176  U+C40B  uniC40B,  HANGUL SYLLABLE-C40B
        0xC40C, # G+14177  U+C40C  uniC40C,  HANGUL SYLLABLE-C40C
        0xC40D, # G+14178  U+C40D  uniC40D,  HANGUL SYLLABLE-C40D
        0xC40E, # G+14179  U+C40E  uniC40E,  HANGUL SYLLABLE-C40E
        0xC40F, # G+14180  U+C40F  uniC40F,  HANGUL SYLLABLE-C40F
        0xC411, # G+14181  U+C411  uniC411,  HANGUL SYLLABLE-C411
        0xC412, # G+14182  U+C412  uniC412,  HANGUL SYLLABLE-C412
        0xC413, # G+14183  U+C413  uniC413,  HANGUL SYLLABLE-C413
        0xC414, # G+14184  U+C414  uniC414,  HANGUL SYLLABLE-C414
        0xC415, # G+14185  U+C415  uniC415,  HANGUL SYLLABLE-C415
        0xC416, # G+14186  U+C416  uniC416,  HANGUL SYLLABLE-C416
        0xC417, # G+14187  U+C417  uniC417,  HANGUL SYLLABLE-C417
        0xC418, # G+14188  U+C418  uniC418,  HANGUL SYLLABLE-C418
        0xC419, # G+14189  U+C419  uniC419,  HANGUL SYLLABLE-C419
        0xC41A, # G+14190  U+C41A  uniC41A,  HANGUL SYLLABLE-C41A
        0xC41B, # G+14191  U+C41B  uniC41B,  HANGUL SYLLABLE-C41B
        0xC41C, # G+14192  U+C41C  uniC41C,  HANGUL SYLLABLE-C41C
        0xC41D, # G+14193  U+C41D  uniC41D,  HANGUL SYLLABLE-C41D
        0xC41E, # G+14194  U+C41E  uniC41E,  HANGUL SYLLABLE-C41E
        0xC41F, # G+14195  U+C41F  uniC41F,  HANGUL SYLLABLE-C41F
        0xC420, # G+14196  U+C420  uniC420,  HANGUL SYLLABLE-C420
        0xC421, # G+14197  U+C421  uniC421,  HANGUL SYLLABLE-C421
        0xC422, # G+14198  U+C422  uniC422,  HANGUL SYLLABLE-C422
        0xC423, # G+14199  U+C423  uniC423,  HANGUL SYLLABLE-C423
        0xC425, # G+14200  U+C425  uniC425,  HANGUL SYLLABLE-C425
        0xC426, # G+14201  U+C426  uniC426,  HANGUL SYLLABLE-C426
        0xC427, # G+14202  U+C427  uniC427,  HANGUL SYLLABLE-C427
        0xC428, # G+14203  U+C428  uniC428,  HANGUL SYLLABLE-C428
        0xC429, # G+14204  U+C429  uniC429,  HANGUL SYLLABLE-C429
        0xC42A, # G+14205  U+C42A  uniC42A,  HANGUL SYLLABLE-C42A
        0xC42B, # G+14206  U+C42B  uniC42B,  HANGUL SYLLABLE-C42B
        0xC42D, # G+14207  U+C42D  uniC42D,  HANGUL SYLLABLE-C42D
        0xC42E, # G+14208  U+C42E  uniC42E,  HANGUL SYLLABLE-C42E
        0xC42F, # G+14209  U+C42F  uniC42F,  HANGUL SYLLABLE-C42F
        0xC431, # G+14210  U+C431  uniC431,  HANGUL SYLLABLE-C431
        0xC432, # G+14211  U+C432  uniC432,  HANGUL SYLLABLE-C432
        0xC433, # G+14212  U+C433  uniC433,  HANGUL SYLLABLE-C433
        0xC435, # G+14213  U+C435  uniC435,  HANGUL SYLLABLE-C435
        0xC436, # G+14214  U+C436  uniC436,  HANGUL SYLLABLE-C436
        0xC437, # G+14215  U+C437  uniC437,  HANGUL SYLLABLE-C437
        0xC438, # G+14216  U+C438  uniC438,  HANGUL SYLLABLE-C438
        0xC439, # G+14217  U+C439  uniC439,  HANGUL SYLLABLE-C439
        0xC43A, # G+14218  U+C43A  uniC43A,  HANGUL SYLLABLE-C43A
        0xC43B, # G+14219  U+C43B  uniC43B,  HANGUL SYLLABLE-C43B
        0xC43E, # G+14220  U+C43E  uniC43E,  HANGUL SYLLABLE-C43E
        0xC43F, # G+14221  U+C43F  uniC43F,  HANGUL SYLLABLE-C43F
        0xC440, # G+14222  U+C440  uniC440,  HANGUL SYLLABLE-C440
        0xC441, # G+14223  U+C441  uniC441,  HANGUL SYLLABLE-C441
        0xC442, # G+14224  U+C442  uniC442,  HANGUL SYLLABLE-C442
        0xC443, # G+14225  U+C443  uniC443,  HANGUL SYLLABLE-C443
        0xC444, # G+14226  U+C444  uniC444,  HANGUL SYLLABLE-C444
        0xC445, # G+14227  U+C445  uniC445,  HANGUL SYLLABLE-C445
        0xC446, # G+14228  U+C446  uniC446,  HANGUL SYLLABLE-C446
        0xC447, # G+14229  U+C447  uniC447,  HANGUL SYLLABLE-C447
        0xC449, # G+14230  U+C449  uniC449,  HANGUL SYLLABLE-C449
        0xC44A, # G+14231  U+C44A  uniC44A,  HANGUL SYLLABLE-C44A
        0xC44B, # G+14232  U+C44B  uniC44B,  HANGUL SYLLABLE-C44B
        0xC44C, # G+14233  U+C44C  uniC44C,  HANGUL SYLLABLE-C44C
        0xC44D, # G+14234  U+C44D  uniC44D,  HANGUL SYLLABLE-C44D
        0xC44E, # G+14235  U+C44E  uniC44E,  HANGUL SYLLABLE-C44E
        0xC44F, # G+14236  U+C44F  uniC44F,  HANGUL SYLLABLE-C44F
        0xC450, # G+14237  U+C450  uniC450,  HANGUL SYLLABLE-C450
        0xC451, # G+14238  U+C451  uniC451,  HANGUL SYLLABLE-C451
        0xC452, # G+14239  U+C452  uniC452,  HANGUL SYLLABLE-C452
        0xC453, # G+14240  U+C453  uniC453,  HANGUL SYLLABLE-C453
        0xC454, # G+14241  U+C454  uniC454,  HANGUL SYLLABLE-C454
        0xC455, # G+14242  U+C455  uniC455,  HANGUL SYLLABLE-C455
        0xC456, # G+14243  U+C456  uniC456,  HANGUL SYLLABLE-C456
        0xC457, # G+14244  U+C457  uniC457,  HANGUL SYLLABLE-C457
        0xC458, # G+14245  U+C458  uniC458,  HANGUL SYLLABLE-C458
        0xC459, # G+14246  U+C459  uniC459,  HANGUL SYLLABLE-C459
        0xC45A, # G+14247  U+C45A  uniC45A,  HANGUL SYLLABLE-C45A
        0xC45B, # G+14248  U+C45B  uniC45B,  HANGUL SYLLABLE-C45B
        0xC45C, # G+14249  U+C45C  uniC45C,  HANGUL SYLLABLE-C45C
        0xC45D, # G+14250  U+C45D  uniC45D,  HANGUL SYLLABLE-C45D
        0xC45E, # G+14251  U+C45E  uniC45E,  HANGUL SYLLABLE-C45E
        0xC45F, # G+14252  U+C45F  uniC45F,  HANGUL SYLLABLE-C45F
        0xC460, # G+14253  U+C460  uniC460,  HANGUL SYLLABLE-C460
        0xC461, # G+14254  U+C461  uniC461,  HANGUL SYLLABLE-C461
        0xC462, # G+14255  U+C462  uniC462,  HANGUL SYLLABLE-C462
        0xC463, # G+14256  U+C463  uniC463,  HANGUL SYLLABLE-C463
        0xC466, # G+14257  U+C466  uniC466,  HANGUL SYLLABLE-C466
        0xC467, # G+14258  U+C467  uniC467,  HANGUL SYLLABLE-C467
        0xC469, # G+14259  U+C469  uniC469,  HANGUL SYLLABLE-C469
        0xC46A, # G+14260  U+C46A  uniC46A,  HANGUL SYLLABLE-C46A
        0xC46B, # G+14261  U+C46B  uniC46B,  HANGUL SYLLABLE-C46B
        0xC46D, # G+14262  U+C46D  uniC46D,  HANGUL SYLLABLE-C46D
        0xC46E, # G+14263  U+C46E  uniC46E,  HANGUL SYLLABLE-C46E
        0xC46F, # G+14264  U+C46F  uniC46F,  HANGUL SYLLABLE-C46F
        0xC470, # G+14265  U+C470  uniC470,  HANGUL SYLLABLE-C470
        0xC471, # G+14266  U+C471  uniC471,  HANGUL SYLLABLE-C471
        0xC472, # G+14267  U+C472  uniC472,  HANGUL SYLLABLE-C472
        0xC473, # G+14268  U+C473  uniC473,  HANGUL SYLLABLE-C473
        0xC476, # G+14269  U+C476  uniC476,  HANGUL SYLLABLE-C476
        0xC477, # G+14270  U+C477  uniC477,  HANGUL SYLLABLE-C477
        0xC478, # G+14271  U+C478  uniC478,  HANGUL SYLLABLE-C478
        0xC47A, # G+14272  U+C47A  uniC47A,  HANGUL SYLLABLE-C47A
        0xC47B, # G+14273  U+C47B  uniC47B,  HANGUL SYLLABLE-C47B
        0xC47C, # G+14274  U+C47C  uniC47C,  HANGUL SYLLABLE-C47C
        0xC47D, # G+14275  U+C47D  uniC47D,  HANGUL SYLLABLE-C47D
        0xC47E, # G+14276  U+C47E  uniC47E,  HANGUL SYLLABLE-C47E
        0xC47F, # G+14277  U+C47F  uniC47F,  HANGUL SYLLABLE-C47F
        0xC481, # G+14278  U+C481  uniC481,  HANGUL SYLLABLE-C481
        0xC482, # G+14279  U+C482  uniC482,  HANGUL SYLLABLE-C482
        0xC483, # G+14280  U+C483  uniC483,  HANGUL SYLLABLE-C483
        0xC484, # G+14281  U+C484  uniC484,  HANGUL SYLLABLE-C484
        0xC485, # G+14282  U+C485  uniC485,  HANGUL SYLLABLE-C485
        0xC486, # G+14283  U+C486  uniC486,  HANGUL SYLLABLE-C486
        0xC487, # G+14284  U+C487  uniC487,  HANGUL SYLLABLE-C487
        0xC488, # G+14285  U+C488  uniC488,  HANGUL SYLLABLE-C488
        0xC489, # G+14286  U+C489  uniC489,  HANGUL SYLLABLE-C489
        0xC48A, # G+14287  U+C48A  uniC48A,  HANGUL SYLLABLE-C48A
        0xC48B, # G+14288  U+C48B  uniC48B,  HANGUL SYLLABLE-C48B
        0xC48C, # G+14289  U+C48C  uniC48C,  HANGUL SYLLABLE-C48C
        0xC48D, # G+14290  U+C48D  uniC48D,  HANGUL SYLLABLE-C48D
        0xC48E, # G+14291  U+C48E  uniC48E,  HANGUL SYLLABLE-C48E
        0xC48F, # G+14292  U+C48F  uniC48F,  HANGUL SYLLABLE-C48F
        0xC490, # G+14293  U+C490  uniC490,  HANGUL SYLLABLE-C490
        0xC491, # G+14294  U+C491  uniC491,  HANGUL SYLLABLE-C491
        0xC492, # G+14295  U+C492  uniC492,  HANGUL SYLLABLE-C492
        0xC493, # G+14296  U+C493  uniC493,  HANGUL SYLLABLE-C493
        0xC495, # G+14297  U+C495  uniC495,  HANGUL SYLLABLE-C495
        0xC496, # G+14298  U+C496  uniC496,  HANGUL SYLLABLE-C496
        0xC497, # G+14299  U+C497  uniC497,  HANGUL SYLLABLE-C497
        0xC498, # G+14300  U+C498  uniC498,  HANGUL SYLLABLE-C498
        0xC499, # G+14301  U+C499  uniC499,  HANGUL SYLLABLE-C499
        0xC49A, # G+14302  U+C49A  uniC49A,  HANGUL SYLLABLE-C49A
        0xC49B, # G+14303  U+C49B  uniC49B,  HANGUL SYLLABLE-C49B
        0xC49D, # G+14304  U+C49D  uniC49D,  HANGUL SYLLABLE-C49D
        0xC49E, # G+14305  U+C49E  uniC49E,  HANGUL SYLLABLE-C49E
        0xC49F, # G+14306  U+C49F  uniC49F,  HANGUL SYLLABLE-C49F
        0xC4A0, # G+14307  U+C4A0  uniC4A0,  HANGUL SYLLABLE-C4A0
        0xC4A1, # G+14308  U+C4A1  uniC4A1,  HANGUL SYLLABLE-C4A1
        0xC4A2, # G+14309  U+C4A2  uniC4A2,  HANGUL SYLLABLE-C4A2
        0xC4A3, # G+14310  U+C4A3  uniC4A3,  HANGUL SYLLABLE-C4A3
        0xC4A4, # G+14311  U+C4A4  uniC4A4,  HANGUL SYLLABLE-C4A4
        0xC4A5, # G+14312  U+C4A5  uniC4A5,  HANGUL SYLLABLE-C4A5
        0xC4A6, # G+14313  U+C4A6  uniC4A6,  HANGUL SYLLABLE-C4A6
        0xC4A7, # G+14314  U+C4A7  uniC4A7,  HANGUL SYLLABLE-C4A7
        0xC4A8, # G+14315  U+C4A8  uniC4A8,  HANGUL SYLLABLE-C4A8
        0xC4A9, # G+14316  U+C4A9  uniC4A9,  HANGUL SYLLABLE-C4A9
        0xC4AA, # G+14317  U+C4AA  uniC4AA,  HANGUL SYLLABLE-C4AA
        0xC4AB, # G+14318  U+C4AB  uniC4AB,  HANGUL SYLLABLE-C4AB
        0xC4AC, # G+14319  U+C4AC  uniC4AC,  HANGUL SYLLABLE-C4AC
        0xC4AD, # G+14320  U+C4AD  uniC4AD,  HANGUL SYLLABLE-C4AD
        0xC4AE, # G+14321  U+C4AE  uniC4AE,  HANGUL SYLLABLE-C4AE
        0xC4AF, # G+14322  U+C4AF  uniC4AF,  HANGUL SYLLABLE-C4AF
        0xC4B0, # G+14323  U+C4B0  uniC4B0,  HANGUL SYLLABLE-C4B0
        0xC4B1, # G+14324  U+C4B1  uniC4B1,  HANGUL SYLLABLE-C4B1
        0xC4B2, # G+14325  U+C4B2  uniC4B2,  HANGUL SYLLABLE-C4B2
        0xC4B3, # G+14326  U+C4B3  uniC4B3,  HANGUL SYLLABLE-C4B3
        0xC4B4, # G+14327  U+C4B4  uniC4B4,  HANGUL SYLLABLE-C4B4
        0xC4B5, # G+14328  U+C4B5  uniC4B5,  HANGUL SYLLABLE-C4B5
        0xC4B6, # G+14329  U+C4B6  uniC4B6,  HANGUL SYLLABLE-C4B6
        0xC4B7, # G+14330  U+C4B7  uniC4B7,  HANGUL SYLLABLE-C4B7
        0xC4B9, # G+14331  U+C4B9  uniC4B9,  HANGUL SYLLABLE-C4B9
        0xC4BA, # G+14332  U+C4BA  uniC4BA,  HANGUL SYLLABLE-C4BA
        0xC4BB, # G+14333  U+C4BB  uniC4BB,  HANGUL SYLLABLE-C4BB
        0xC4BD, # G+14334  U+C4BD  uniC4BD,  HANGUL SYLLABLE-C4BD
        0xC4BE, # G+14335  U+C4BE  uniC4BE,  HANGUL SYLLABLE-C4BE
        0xC4BF, # G+14336  U+C4BF  uniC4BF,  HANGUL SYLLABLE-C4BF
        0xC4C0, # G+14337  U+C4C0  uniC4C0,  HANGUL SYLLABLE-C4C0
        0xC4C1, # G+14338  U+C4C1  uniC4C1,  HANGUL SYLLABLE-C4C1
        0xC4C2, # G+14339  U+C4C2  uniC4C2,  HANGUL SYLLABLE-C4C2
        0xC4C3, # G+14340  U+C4C3  uniC4C3,  HANGUL SYLLABLE-C4C3
        0xC4C4, # G+14341  U+C4C4  uniC4C4,  HANGUL SYLLABLE-C4C4
        0xC4C5, # G+14342  U+C4C5  uniC4C5,  HANGUL SYLLABLE-C4C5
        0xC4C6, # G+14343  U+C4C6  uniC4C6,  HANGUL SYLLABLE-C4C6
        0xC4C7, # G+14344  U+C4C7  uniC4C7,  HANGUL SYLLABLE-C4C7
        0xC4C8, # G+14345  U+C4C8  uniC4C8,  HANGUL SYLLABLE-C4C8
        0xC4C9, # G+14346  U+C4C9  uniC4C9,  HANGUL SYLLABLE-C4C9
        0xC4CA, # G+14347  U+C4CA  uniC4CA,  HANGUL SYLLABLE-C4CA
        0xC4CB, # G+14348  U+C4CB  uniC4CB,  HANGUL SYLLABLE-C4CB
        0xC4CC, # G+14349  U+C4CC  uniC4CC,  HANGUL SYLLABLE-C4CC
        0xC4CD, # G+14350  U+C4CD  uniC4CD,  HANGUL SYLLABLE-C4CD
        0xC4CE, # G+14351  U+C4CE  uniC4CE,  HANGUL SYLLABLE-C4CE
        0xC4CF, # G+14352  U+C4CF  uniC4CF,  HANGUL SYLLABLE-C4CF
        0xC4D0, # G+14353  U+C4D0  uniC4D0,  HANGUL SYLLABLE-C4D0
        0xC4D1, # G+14354  U+C4D1  uniC4D1,  HANGUL SYLLABLE-C4D1
        0xC4D2, # G+14355  U+C4D2  uniC4D2,  HANGUL SYLLABLE-C4D2
        0xC4D3, # G+14356  U+C4D3  uniC4D3,  HANGUL SYLLABLE-C4D3
        0xC4D4, # G+14357  U+C4D4  uniC4D4,  HANGUL SYLLABLE-C4D4
        0xC4D5, # G+14358  U+C4D5  uniC4D5,  HANGUL SYLLABLE-C4D5
        0xC4D6, # G+14359  U+C4D6  uniC4D6,  HANGUL SYLLABLE-C4D6
        0xC4D7, # G+14360  U+C4D7  uniC4D7,  HANGUL SYLLABLE-C4D7
        0xC4D8, # G+14361  U+C4D8  uniC4D8,  HANGUL SYLLABLE-C4D8
        0xC4D9, # G+14362  U+C4D9  uniC4D9,  HANGUL SYLLABLE-C4D9
        0xC4DA, # G+14363  U+C4DA  uniC4DA,  HANGUL SYLLABLE-C4DA
        0xC4DB, # G+14364  U+C4DB  uniC4DB,  HANGUL SYLLABLE-C4DB
        0xC4DC, # G+14365  U+C4DC  uniC4DC,  HANGUL SYLLABLE-C4DC
        0xC4DD, # G+14366  U+C4DD  uniC4DD,  HANGUL SYLLABLE-C4DD
        0xC4DE, # G+14367  U+C4DE  uniC4DE,  HANGUL SYLLABLE-C4DE
        0xC4DF, # G+14368  U+C4DF  uniC4DF,  HANGUL SYLLABLE-C4DF
        0xC4E0, # G+14369  U+C4E0  uniC4E0,  HANGUL SYLLABLE-C4E0
        0xC4E1, # G+14370  U+C4E1  uniC4E1,  HANGUL SYLLABLE-C4E1
        0xC4E2, # G+14371  U+C4E2  uniC4E2,  HANGUL SYLLABLE-C4E2
        0xC4E3, # G+14372  U+C4E3  uniC4E3,  HANGUL SYLLABLE-C4E3
        0xC4E4, # G+14373  U+C4E4  uniC4E4,  HANGUL SYLLABLE-C4E4
        0xC4E5, # G+14374  U+C4E5  uniC4E5,  HANGUL SYLLABLE-C4E5
        0xC4E6, # G+14375  U+C4E6  uniC4E6,  HANGUL SYLLABLE-C4E6
        0xC4E7, # G+14376  U+C4E7  uniC4E7,  HANGUL SYLLABLE-C4E7
        0xC4E8, # G+14377  U+C4E8  uniC4E8,  HANGUL SYLLABLE-C4E8
        0xC4EA, # G+14378  U+C4EA  uniC4EA,  HANGUL SYLLABLE-C4EA
        0xC4EB, # G+14379  U+C4EB  uniC4EB,  HANGUL SYLLABLE-C4EB
        0xC4EC, # G+14380  U+C4EC  uniC4EC,  HANGUL SYLLABLE-C4EC
        0xC4ED, # G+14381  U+C4ED  uniC4ED,  HANGUL SYLLABLE-C4ED
        0xC4EE, # G+14382  U+C4EE  uniC4EE,  HANGUL SYLLABLE-C4EE
        0xC4EF, # G+14383  U+C4EF  uniC4EF,  HANGUL SYLLABLE-C4EF
        0xC4F2, # G+14384  U+C4F2  uniC4F2,  HANGUL SYLLABLE-C4F2
        0xC4F3, # G+14385  U+C4F3  uniC4F3,  HANGUL SYLLABLE-C4F3
        0xC4F5, # G+14386  U+C4F5  uniC4F5,  HANGUL SYLLABLE-C4F5
        0xC4F6, # G+14387  U+C4F6  uniC4F6,  HANGUL SYLLABLE-C4F6
        0xC4F7, # G+14388  U+C4F7  uniC4F7,  HANGUL SYLLABLE-C4F7
        0xC4F9, # G+14389  U+C4F9  uniC4F9,  HANGUL SYLLABLE-C4F9
        0xC4FB, # G+14390  U+C4FB  uniC4FB,  HANGUL SYLLABLE-C4FB
        0xC4FC, # G+14391  U+C4FC  uniC4FC,  HANGUL SYLLABLE-C4FC
        0xC4FD, # G+14392  U+C4FD  uniC4FD,  HANGUL SYLLABLE-C4FD
        0xC4FE, # G+14393  U+C4FE  uniC4FE,  HANGUL SYLLABLE-C4FE
        0xC502, # G+14394  U+C502  uniC502,  HANGUL SYLLABLE-C502
        0xC503, # G+14395  U+C503  uniC503,  HANGUL SYLLABLE-C503
        0xC504, # G+14396  U+C504  uniC504,  HANGUL SYLLABLE-C504
        0xC505, # G+14397  U+C505  uniC505,  HANGUL SYLLABLE-C505
        0xC506, # G+14398  U+C506  uniC506,  HANGUL SYLLABLE-C506
        0xC507, # G+14399  U+C507  uniC507,  HANGUL SYLLABLE-C507
        0xC508, # G+14400  U+C508  uniC508,  HANGUL SYLLABLE-C508
        0xC509, # G+14401  U+C509  uniC509,  HANGUL SYLLABLE-C509
        0xC50A, # G+14402  U+C50A  uniC50A,  HANGUL SYLLABLE-C50A
        0xC50B, # G+14403  U+C50B  uniC50B,  HANGUL SYLLABLE-C50B
        0xC50D, # G+14404  U+C50D  uniC50D,  HANGUL SYLLABLE-C50D
        0xC50E, # G+14405  U+C50E  uniC50E,  HANGUL SYLLABLE-C50E
        0xC50F, # G+14406  U+C50F  uniC50F,  HANGUL SYLLABLE-C50F
        0xC511, # G+14407  U+C511  uniC511,  HANGUL SYLLABLE-C511
        0xC512, # G+14408  U+C512  uniC512,  HANGUL SYLLABLE-C512
        0xC513, # G+14409  U+C513  uniC513,  HANGUL SYLLABLE-C513
        0xC515, # G+14410  U+C515  uniC515,  HANGUL SYLLABLE-C515
        0xC516, # G+14411  U+C516  uniC516,  HANGUL SYLLABLE-C516
        0xC517, # G+14412  U+C517  uniC517,  HANGUL SYLLABLE-C517
        0xC518, # G+14413  U+C518  uniC518,  HANGUL SYLLABLE-C518
        0xC519, # G+14414  U+C519  uniC519,  HANGUL SYLLABLE-C519
        0xC51A, # G+14415  U+C51A  uniC51A,  HANGUL SYLLABLE-C51A
        0xC51B, # G+14416  U+C51B  uniC51B,  HANGUL SYLLABLE-C51B
        0xC51D, # G+14417  U+C51D  uniC51D,  HANGUL SYLLABLE-C51D
        0xC51E, # G+14418  U+C51E  uniC51E,  HANGUL SYLLABLE-C51E
        0xC51F, # G+14419  U+C51F  uniC51F,  HANGUL SYLLABLE-C51F
        0xC520, # G+14420  U+C520  uniC520,  HANGUL SYLLABLE-C520
        0xC521, # G+14421  U+C521  uniC521,  HANGUL SYLLABLE-C521
        0xC522, # G+14422  U+C522  uniC522,  HANGUL SYLLABLE-C522
        0xC523, # G+14423  U+C523  uniC523,  HANGUL SYLLABLE-C523
        0xC524, # G+14424  U+C524  uniC524,  HANGUL SYLLABLE-C524
        0xC525, # G+14425  U+C525  uniC525,  HANGUL SYLLABLE-C525
        0xC526, # G+14426  U+C526  uniC526,  HANGUL SYLLABLE-C526
        0xC527, # G+14427  U+C527  uniC527,  HANGUL SYLLABLE-C527
        0xC52A, # G+14428  U+C52A  uniC52A,  HANGUL SYLLABLE-C52A
        0xC52B, # G+14429  U+C52B  uniC52B,  HANGUL SYLLABLE-C52B
        0xC52D, # G+14430  U+C52D  uniC52D,  HANGUL SYLLABLE-C52D
        0xC52E, # G+14431  U+C52E  uniC52E,  HANGUL SYLLABLE-C52E
        0xC52F, # G+14432  U+C52F  uniC52F,  HANGUL SYLLABLE-C52F
        0xC531, # G+14433  U+C531  uniC531,  HANGUL SYLLABLE-C531
        0xC532, # G+14434  U+C532  uniC532,  HANGUL SYLLABLE-C532
        0xC533, # G+14435  U+C533  uniC533,  HANGUL SYLLABLE-C533
        0xC534, # G+14436  U+C534  uniC534,  HANGUL SYLLABLE-C534
        0xC535, # G+14437  U+C535  uniC535,  HANGUL SYLLABLE-C535
        0xC536, # G+14438  U+C536  uniC536,  HANGUL SYLLABLE-C536
        0xC537, # G+14439  U+C537  uniC537,  HANGUL SYLLABLE-C537
        0xC53A, # G+14440  U+C53A  uniC53A,  HANGUL SYLLABLE-C53A
        0xC53C, # G+14441  U+C53C  uniC53C,  HANGUL SYLLABLE-C53C
        0xC53E, # G+14442  U+C53E  uniC53E,  HANGUL SYLLABLE-C53E
        0xC53F, # G+14443  U+C53F  uniC53F,  HANGUL SYLLABLE-C53F
        0xC540, # G+14444  U+C540  uniC540,  HANGUL SYLLABLE-C540
        0xC541, # G+14445  U+C541  uniC541,  HANGUL SYLLABLE-C541
        0xC542, # G+14446  U+C542  uniC542,  HANGUL SYLLABLE-C542
        0xC543, # G+14447  U+C543  uniC543,  HANGUL SYLLABLE-C543
        0xC546, # G+14448  U+C546  uniC546,  HANGUL SYLLABLE-C546
        0xC547, # G+14449  U+C547  uniC547,  HANGUL SYLLABLE-C547
        0xC54B, # G+14450  U+C54B  uniC54B,  HANGUL SYLLABLE-C54B
        0xC54F, # G+14451  U+C54F  uniC54F,  HANGUL SYLLABLE-C54F
        0xC550, # G+14452  U+C550  uniC550,  HANGUL SYLLABLE-C550
        0xC551, # G+14453  U+C551  uniC551,  HANGUL SYLLABLE-C551
        0xC552, # G+14454  U+C552  uniC552,  HANGUL SYLLABLE-C552
        0xC556, # G+14455  U+C556  uniC556,  HANGUL SYLLABLE-C556
        0xC55A, # G+14456  U+C55A  uniC55A,  HANGUL SYLLABLE-C55A
        0xC55B, # G+14457  U+C55B  uniC55B,  HANGUL SYLLABLE-C55B
        0xC55C, # G+14458  U+C55C  uniC55C,  HANGUL SYLLABLE-C55C
        0xC55F, # G+14459  U+C55F  uniC55F,  HANGUL SYLLABLE-C55F
        0xC562, # G+14460  U+C562  uniC562,  HANGUL SYLLABLE-C562
        0xC563, # G+14461  U+C563  uniC563,  HANGUL SYLLABLE-C563
        0xC565, # G+14462  U+C565  uniC565,  HANGUL SYLLABLE-C565
        0xC566, # G+14463  U+C566  uniC566,  HANGUL SYLLABLE-C566
        0xC567, # G+14464  U+C567  uniC567,  HANGUL SYLLABLE-C567
        0xC569, # G+14465  U+C569  uniC569,  HANGUL SYLLABLE-C569
        0xC56A, # G+14466  U+C56A  uniC56A,  HANGUL SYLLABLE-C56A
        0xC56B, # G+14467  U+C56B  uniC56B,  HANGUL SYLLABLE-C56B
        0xC56C, # G+14468  U+C56C  uniC56C,  HANGUL SYLLABLE-C56C
        0xC56D, # G+14469  U+C56D  uniC56D,  HANGUL SYLLABLE-C56D
        0xC56E, # G+14470  U+C56E  uniC56E,  HANGUL SYLLABLE-C56E
        0xC56F, # G+14471  U+C56F  uniC56F,  HANGUL SYLLABLE-C56F
        0xC572, # G+14472  U+C572  uniC572,  HANGUL SYLLABLE-C572
        0xC576, # G+14473  U+C576  uniC576,  HANGUL SYLLABLE-C576
        0xC577, # G+14474  U+C577  uniC577,  HANGUL SYLLABLE-C577
        0xC578, # G+14475  U+C578  uniC578,  HANGUL SYLLABLE-C578
        0xC579, # G+14476  U+C579  uniC579,  HANGUL SYLLABLE-C579
        0xC57A, # G+14477  U+C57A  uniC57A,  HANGUL SYLLABLE-C57A
        0xC57B, # G+14478  U+C57B  uniC57B,  HANGUL SYLLABLE-C57B
        0xC57E, # G+14479  U+C57E  uniC57E,  HANGUL SYLLABLE-C57E
        0xC57F, # G+14480  U+C57F  uniC57F,  HANGUL SYLLABLE-C57F
        0xC581, # G+14481  U+C581  uniC581,  HANGUL SYLLABLE-C581
        0xC582, # G+14482  U+C582  uniC582,  HANGUL SYLLABLE-C582
        0xC583, # G+14483  U+C583  uniC583,  HANGUL SYLLABLE-C583
        0xC585, # G+14484  U+C585  uniC585,  HANGUL SYLLABLE-C585
        0xC586, # G+14485  U+C586  uniC586,  HANGUL SYLLABLE-C586
        0xC588, # G+14486  U+C588  uniC588,  HANGUL SYLLABLE-C588
        0xC589, # G+14487  U+C589  uniC589,  HANGUL SYLLABLE-C589
        0xC58A, # G+14488  U+C58A  uniC58A,  HANGUL SYLLABLE-C58A
        0xC58B, # G+14489  U+C58B  uniC58B,  HANGUL SYLLABLE-C58B
        0xC58E, # G+14490  U+C58E  uniC58E,  HANGUL SYLLABLE-C58E
        0xC590, # G+14491  U+C590  uniC590,  HANGUL SYLLABLE-C590
        0xC592, # G+14492  U+C592  uniC592,  HANGUL SYLLABLE-C592
        0xC593, # G+14493  U+C593  uniC593,  HANGUL SYLLABLE-C593
        0xC594, # G+14494  U+C594  uniC594,  HANGUL SYLLABLE-C594
        0xC596, # G+14495  U+C596  uniC596,  HANGUL SYLLABLE-C596
        0xC599, # G+14496  U+C599  uniC599,  HANGUL SYLLABLE-C599
        0xC59A, # G+14497  U+C59A  uniC59A,  HANGUL SYLLABLE-C59A
        0xC59B, # G+14498  U+C59B  uniC59B,  HANGUL SYLLABLE-C59B
        0xC59D, # G+14499  U+C59D  uniC59D,  HANGUL SYLLABLE-C59D
        0xC59E, # G+14500  U+C59E  uniC59E,  HANGUL SYLLABLE-C59E
        0xC59F, # G+14501  U+C59F  uniC59F,  HANGUL SYLLABLE-C59F
        0xC5A1, # G+14502  U+C5A1  uniC5A1,  HANGUL SYLLABLE-C5A1
        0xC5A2, # G+14503  U+C5A2  uniC5A2,  HANGUL SYLLABLE-C5A2
        0xC5A3, # G+14504  U+C5A3  uniC5A3,  HANGUL SYLLABLE-C5A3
        0xC5A4, # G+14505  U+C5A4  uniC5A4,  HANGUL SYLLABLE-C5A4
        0xC5A5, # G+14506  U+C5A5  uniC5A5,  HANGUL SYLLABLE-C5A5
        0xC5A6, # G+14507  U+C5A6  uniC5A6,  HANGUL SYLLABLE-C5A6
        0xC5A7, # G+14508  U+C5A7  uniC5A7,  HANGUL SYLLABLE-C5A7
        0xC5A8, # G+14509  U+C5A8  uniC5A8,  HANGUL SYLLABLE-C5A8
        0xC5AA, # G+14510  U+C5AA  uniC5AA,  HANGUL SYLLABLE-C5AA
        0xC5AB, # G+14511  U+C5AB  uniC5AB,  HANGUL SYLLABLE-C5AB
        0xC5AC, # G+14512  U+C5AC  uniC5AC,  HANGUL SYLLABLE-C5AC
        0xC5AD, # G+14513  U+C5AD  uniC5AD,  HANGUL SYLLABLE-C5AD
        0xC5AE, # G+14514  U+C5AE  uniC5AE,  HANGUL SYLLABLE-C5AE
        0xC5AF, # G+14515  U+C5AF  uniC5AF,  HANGUL SYLLABLE-C5AF
        0xC5B0, # G+14516  U+C5B0  uniC5B0,  HANGUL SYLLABLE-C5B0
        0xC5B1, # G+14517  U+C5B1  uniC5B1,  HANGUL SYLLABLE-C5B1
        0xC5B2, # G+14518  U+C5B2  uniC5B2,  HANGUL SYLLABLE-C5B2
        0xC5B3, # G+14519  U+C5B3  uniC5B3,  HANGUL SYLLABLE-C5B3
        0xC5B6, # G+14520  U+C5B6  uniC5B6,  HANGUL SYLLABLE-C5B6
        0xC5B7, # G+14521  U+C5B7  uniC5B7,  HANGUL SYLLABLE-C5B7
        0xC5BA, # G+14522  U+C5BA  uniC5BA,  HANGUL SYLLABLE-C5BA
        0xC5BF, # G+14523  U+C5BF  uniC5BF,  HANGUL SYLLABLE-C5BF
        0xC5C0, # G+14524  U+C5C0  uniC5C0,  HANGUL SYLLABLE-C5C0
        0xC5C1, # G+14525  U+C5C1  uniC5C1,  HANGUL SYLLABLE-C5C1
        0xC5C2, # G+14526  U+C5C2  uniC5C2,  HANGUL SYLLABLE-C5C2
        0xC5C3, # G+14527  U+C5C3  uniC5C3,  HANGUL SYLLABLE-C5C3
        0xC5CB, # G+14528  U+C5CB  uniC5CB,  HANGUL SYLLABLE-C5CB
        0xC5CD, # G+14529  U+C5CD  uniC5CD,  HANGUL SYLLABLE-C5CD
        0xC5CF, # G+14530  U+C5CF  uniC5CF,  HANGUL SYLLABLE-C5CF
        0xC5D2, # G+14531  U+C5D2  uniC5D2,  HANGUL SYLLABLE-C5D2
        0xC5D3, # G+14532  U+C5D3  uniC5D3,  HANGUL SYLLABLE-C5D3
        0xC5D5, # G+14533  U+C5D5  uniC5D5,  HANGUL SYLLABLE-C5D5
        0xC5D6, # G+14534  U+C5D6  uniC5D6,  HANGUL SYLLABLE-C5D6
        0xC5D7, # G+14535  U+C5D7  uniC5D7,  HANGUL SYLLABLE-C5D7
        0xC5D9, # G+14536  U+C5D9  uniC5D9,  HANGUL SYLLABLE-C5D9
        0xC5DA, # G+14537  U+C5DA  uniC5DA,  HANGUL SYLLABLE-C5DA
        0xC5DB, # G+14538  U+C5DB  uniC5DB,  HANGUL SYLLABLE-C5DB
        0xC5DC, # G+14539  U+C5DC  uniC5DC,  HANGUL SYLLABLE-C5DC
        0xC5DD, # G+14540  U+C5DD  uniC5DD,  HANGUL SYLLABLE-C5DD
        0xC5DE, # G+14541  U+C5DE  uniC5DE,  HANGUL SYLLABLE-C5DE
        0xC5DF, # G+14542  U+C5DF  uniC5DF,  HANGUL SYLLABLE-C5DF
        0xC5E2, # G+14543  U+C5E2  uniC5E2,  HANGUL SYLLABLE-C5E2
        0xC5E4, # G+14544  U+C5E4  uniC5E4,  HANGUL SYLLABLE-C5E4
        0xC5E6, # G+14545  U+C5E6  uniC5E6,  HANGUL SYLLABLE-C5E6
        0xC5E7, # G+14546  U+C5E7  uniC5E7,  HANGUL SYLLABLE-C5E7
        0xC5E8, # G+14547  U+C5E8  uniC5E8,  HANGUL SYLLABLE-C5E8
        0xC5E9, # G+14548  U+C5E9  uniC5E9,  HANGUL SYLLABLE-C5E9
        0xC5EA, # G+14549  U+C5EA  uniC5EA,  HANGUL SYLLABLE-C5EA
        0xC5EB, # G+14550  U+C5EB  uniC5EB,  HANGUL SYLLABLE-C5EB
        0xC5EF, # G+14551  U+C5EF  uniC5EF,  HANGUL SYLLABLE-C5EF
        0xC5F1, # G+14552  U+C5F1  uniC5F1,  HANGUL SYLLABLE-C5F1
        0xC5F2, # G+14553  U+C5F2  uniC5F2,  HANGUL SYLLABLE-C5F2
        0xC5F3, # G+14554  U+C5F3  uniC5F3,  HANGUL SYLLABLE-C5F3
        0xC5F5, # G+14555  U+C5F5  uniC5F5,  HANGUL SYLLABLE-C5F5
        0xC5F8, # G+14556  U+C5F8  uniC5F8,  HANGUL SYLLABLE-C5F8
        0xC5F9, # G+14557  U+C5F9  uniC5F9,  HANGUL SYLLABLE-C5F9
        0xC5FA, # G+14558  U+C5FA  uniC5FA,  HANGUL SYLLABLE-C5FA
        0xC5FB, # G+14559  U+C5FB  uniC5FB,  HANGUL SYLLABLE-C5FB
        0xC602, # G+14560  U+C602  uniC602,  HANGUL SYLLABLE-C602
        0xC603, # G+14561  U+C603  uniC603,  HANGUL SYLLABLE-C603
        0xC604, # G+14562  U+C604  uniC604,  HANGUL SYLLABLE-C604
        0xC609, # G+14563  U+C609  uniC609,  HANGUL SYLLABLE-C609
        0xC60A, # G+14564  U+C60A  uniC60A,  HANGUL SYLLABLE-C60A
        0xC60B, # G+14565  U+C60B  uniC60B,  HANGUL SYLLABLE-C60B
        0xC60D, # G+14566  U+C60D  uniC60D,  HANGUL SYLLABLE-C60D
        0xC60E, # G+14567  U+C60E  uniC60E,  HANGUL SYLLABLE-C60E
        0xC60F, # G+14568  U+C60F  uniC60F,  HANGUL SYLLABLE-C60F
        0xC611, # G+14569  U+C611  uniC611,  HANGUL SYLLABLE-C611
        0xC612, # G+14570  U+C612  uniC612,  HANGUL SYLLABLE-C612
        0xC613, # G+14571  U+C613  uniC613,  HANGUL SYLLABLE-C613
        0xC614, # G+14572  U+C614  uniC614,  HANGUL SYLLABLE-C614
        0xC615, # G+14573  U+C615  uniC615,  HANGUL SYLLABLE-C615
        0xC616, # G+14574  U+C616  uniC616,  HANGUL SYLLABLE-C616
        0xC617, # G+14575  U+C617  uniC617,  HANGUL SYLLABLE-C617
        0xC61A, # G+14576  U+C61A  uniC61A,  HANGUL SYLLABLE-C61A
        0xC61D, # G+14577  U+C61D  uniC61D,  HANGUL SYLLABLE-C61D
        0xC61E, # G+14578  U+C61E  uniC61E,  HANGUL SYLLABLE-C61E
        0xC61F, # G+14579  U+C61F  uniC61F,  HANGUL SYLLABLE-C61F
        0xC620, # G+14580  U+C620  uniC620,  HANGUL SYLLABLE-C620
        0xC621, # G+14581  U+C621  uniC621,  HANGUL SYLLABLE-C621
        0xC622, # G+14582  U+C622  uniC622,  HANGUL SYLLABLE-C622
        0xC623, # G+14583  U+C623  uniC623,  HANGUL SYLLABLE-C623
        0xC626, # G+14584  U+C626  uniC626,  HANGUL SYLLABLE-C626
        0xC627, # G+14585  U+C627  uniC627,  HANGUL SYLLABLE-C627
        0xC629, # G+14586  U+C629  uniC629,  HANGUL SYLLABLE-C629
        0xC62A, # G+14587  U+C62A  uniC62A,  HANGUL SYLLABLE-C62A
        0xC62B, # G+14588  U+C62B  uniC62B,  HANGUL SYLLABLE-C62B
        0xC62F, # G+14589  U+C62F  uniC62F,  HANGUL SYLLABLE-C62F
        0xC631, # G+14590  U+C631  uniC631,  HANGUL SYLLABLE-C631
        0xC632, # G+14591  U+C632  uniC632,  HANGUL SYLLABLE-C632
        0xC636, # G+14592  U+C636  uniC636,  HANGUL SYLLABLE-C636
        0xC638, # G+14593  U+C638  uniC638,  HANGUL SYLLABLE-C638
        0xC63A, # G+14594  U+C63A  uniC63A,  HANGUL SYLLABLE-C63A
        0xC63C, # G+14595  U+C63C  uniC63C,  HANGUL SYLLABLE-C63C
        0xC63D, # G+14596  U+C63D  uniC63D,  HANGUL SYLLABLE-C63D
        0xC63E, # G+14597  U+C63E  uniC63E,  HANGUL SYLLABLE-C63E
        0xC63F, # G+14598  U+C63F  uniC63F,  HANGUL SYLLABLE-C63F
        0xC642, # G+14599  U+C642  uniC642,  HANGUL SYLLABLE-C642
        0xC643, # G+14600  U+C643  uniC643,  HANGUL SYLLABLE-C643
        0xC645, # G+14601  U+C645  uniC645,  HANGUL SYLLABLE-C645
        0xC646, # G+14602  U+C646  uniC646,  HANGUL SYLLABLE-C646
        0xC647, # G+14603  U+C647  uniC647,  HANGUL SYLLABLE-C647
        0xC649, # G+14604  U+C649  uniC649,  HANGUL SYLLABLE-C649
        0xC64A, # G+14605  U+C64A  uniC64A,  HANGUL SYLLABLE-C64A
        0xC64B, # G+14606  U+C64B  uniC64B,  HANGUL SYLLABLE-C64B
        0xC64C, # G+14607  U+C64C  uniC64C,  HANGUL SYLLABLE-C64C
        0xC64D, # G+14608  U+C64D  uniC64D,  HANGUL SYLLABLE-C64D
        0xC64E, # G+14609  U+C64E  uniC64E,  HANGUL SYLLABLE-C64E
        0xC64F, # G+14610  U+C64F  uniC64F,  HANGUL SYLLABLE-C64F
        0xC652, # G+14611  U+C652  uniC652,  HANGUL SYLLABLE-C652
        0xC656, # G+14612  U+C656  uniC656,  HANGUL SYLLABLE-C656
        0xC657, # G+14613  U+C657  uniC657,  HANGUL SYLLABLE-C657
        0xC658, # G+14614  U+C658  uniC658,  HANGUL SYLLABLE-C658
        0xC659, # G+14615  U+C659  uniC659,  HANGUL SYLLABLE-C659
        0xC65A, # G+14616  U+C65A  uniC65A,  HANGUL SYLLABLE-C65A
        0xC65B, # G+14617  U+C65B  uniC65B,  HANGUL SYLLABLE-C65B
        0xC65E, # G+14618  U+C65E  uniC65E,  HANGUL SYLLABLE-C65E
        0xC65F, # G+14619  U+C65F  uniC65F,  HANGUL SYLLABLE-C65F
        0xC661, # G+14620  U+C661  uniC661,  HANGUL SYLLABLE-C661
        0xC662, # G+14621  U+C662  uniC662,  HANGUL SYLLABLE-C662
        0xC663, # G+14622  U+C663  uniC663,  HANGUL SYLLABLE-C663
        0xC664, # G+14623  U+C664  uniC664,  HANGUL SYLLABLE-C664
        0xC665, # G+14624  U+C665  uniC665,  HANGUL SYLLABLE-C665
        0xC666, # G+14625  U+C666  uniC666,  HANGUL SYLLABLE-C666
        0xC667, # G+14626  U+C667  uniC667,  HANGUL SYLLABLE-C667
        0xC668, # G+14627  U+C668  uniC668,  HANGUL SYLLABLE-C668
        0xC669, # G+14628  U+C669  uniC669,  HANGUL SYLLABLE-C669
        0xC66A, # G+14629  U+C66A  uniC66A,  HANGUL SYLLABLE-C66A
        0xC66B, # G+14630  U+C66B  uniC66B,  HANGUL SYLLABLE-C66B
        0xC66D, # G+14631  U+C66D  uniC66D,  HANGUL SYLLABLE-C66D
        0xC66E, # G+14632  U+C66E  uniC66E,  HANGUL SYLLABLE-C66E
        0xC670, # G+14633  U+C670  uniC670,  HANGUL SYLLABLE-C670
        0xC672, # G+14634  U+C672  uniC672,  HANGUL SYLLABLE-C672
        0xC673, # G+14635  U+C673  uniC673,  HANGUL SYLLABLE-C673
        0xC674, # G+14636  U+C674  uniC674,  HANGUL SYLLABLE-C674
        0xC675, # G+14637  U+C675  uniC675,  HANGUL SYLLABLE-C675
        0xC676, # G+14638  U+C676  uniC676,  HANGUL SYLLABLE-C676
        0xC677, # G+14639  U+C677  uniC677,  HANGUL SYLLABLE-C677
        0xC67A, # G+14640  U+C67A  uniC67A,  HANGUL SYLLABLE-C67A
        0xC67B, # G+14641  U+C67B  uniC67B,  HANGUL SYLLABLE-C67B
        0xC67D, # G+14642  U+C67D  uniC67D,  HANGUL SYLLABLE-C67D
        0xC67E, # G+14643  U+C67E  uniC67E,  HANGUL SYLLABLE-C67E
        0xC67F, # G+14644  U+C67F  uniC67F,  HANGUL SYLLABLE-C67F
        0xC681, # G+14645  U+C681  uniC681,  HANGUL SYLLABLE-C681
        0xC682, # G+14646  U+C682  uniC682,  HANGUL SYLLABLE-C682
        0xC683, # G+14647  U+C683  uniC683,  HANGUL SYLLABLE-C683
        0xC684, # G+14648  U+C684  uniC684,  HANGUL SYLLABLE-C684
        0xC685, # G+14649  U+C685  uniC685,  HANGUL SYLLABLE-C685
        0xC686, # G+14650  U+C686  uniC686,  HANGUL SYLLABLE-C686
        0xC687, # G+14651  U+C687  uniC687,  HANGUL SYLLABLE-C687
        0xC68A, # G+14652  U+C68A  uniC68A,  HANGUL SYLLABLE-C68A
        0xC68C, # G+14653  U+C68C  uniC68C,  HANGUL SYLLABLE-C68C
        0xC68E, # G+14654  U+C68E  uniC68E,  HANGUL SYLLABLE-C68E
        0xC68F, # G+14655  U+C68F  uniC68F,  HANGUL SYLLABLE-C68F
        0xC690, # G+14656  U+C690  uniC690,  HANGUL SYLLABLE-C690
        0xC691, # G+14657  U+C691  uniC691,  HANGUL SYLLABLE-C691
        0xC692, # G+14658  U+C692  uniC692,  HANGUL SYLLABLE-C692
        0xC693, # G+14659  U+C693  uniC693,  HANGUL SYLLABLE-C693
        0xC696, # G+14660  U+C696  uniC696,  HANGUL SYLLABLE-C696
        0xC697, # G+14661  U+C697  uniC697,  HANGUL SYLLABLE-C697
        0xC699, # G+14662  U+C699  uniC699,  HANGUL SYLLABLE-C699
        0xC69A, # G+14663  U+C69A  uniC69A,  HANGUL SYLLABLE-C69A
        0xC69B, # G+14664  U+C69B  uniC69B,  HANGUL SYLLABLE-C69B
        0xC69D, # G+14665  U+C69D  uniC69D,  HANGUL SYLLABLE-C69D
        0xC69E, # G+14666  U+C69E  uniC69E,  HANGUL SYLLABLE-C69E
        0xC69F, # G+14667  U+C69F  uniC69F,  HANGUL SYLLABLE-C69F
        0xC6A0, # G+14668  U+C6A0  uniC6A0,  HANGUL SYLLABLE-C6A0
        0xC6A1, # G+14669  U+C6A1  uniC6A1,  HANGUL SYLLABLE-C6A1
        0xC6A2, # G+14670  U+C6A2  uniC6A2,  HANGUL SYLLABLE-C6A2
        0xC6A3, # G+14671  U+C6A3  uniC6A3,  HANGUL SYLLABLE-C6A3
        0xC6A6, # G+14672  U+C6A6  uniC6A6,  HANGUL SYLLABLE-C6A6
        0xC6A8, # G+14673  U+C6A8  uniC6A8,  HANGUL SYLLABLE-C6A8
        0xC6AA, # G+14674  U+C6AA  uniC6AA,  HANGUL SYLLABLE-C6AA
        0xC6AB, # G+14675  U+C6AB  uniC6AB,  HANGUL SYLLABLE-C6AB
        0xC6AC, # G+14676  U+C6AC  uniC6AC,  HANGUL SYLLABLE-C6AC
        0xC6AD, # G+14677  U+C6AD  uniC6AD,  HANGUL SYLLABLE-C6AD
        0xC6AE, # G+14678  U+C6AE  uniC6AE,  HANGUL SYLLABLE-C6AE
        0xC6AF, # G+14679  U+C6AF  uniC6AF,  HANGUL SYLLABLE-C6AF
        0xC6B2, # G+14680  U+C6B2  uniC6B2,  HANGUL SYLLABLE-C6B2
        0xC6B3, # G+14681  U+C6B3  uniC6B3,  HANGUL SYLLABLE-C6B3
        0xC6B5, # G+14682  U+C6B5  uniC6B5,  HANGUL SYLLABLE-C6B5
        0xC6B6, # G+14683  U+C6B6  uniC6B6,  HANGUL SYLLABLE-C6B6
        0xC6B7, # G+14684  U+C6B7  uniC6B7,  HANGUL SYLLABLE-C6B7
        0xC6BB, # G+14685  U+C6BB  uniC6BB,  HANGUL SYLLABLE-C6BB
        0xC6BC, # G+14686  U+C6BC  uniC6BC,  HANGUL SYLLABLE-C6BC
        0xC6BD, # G+14687  U+C6BD  uniC6BD,  HANGUL SYLLABLE-C6BD
        0xC6BE, # G+14688  U+C6BE  uniC6BE,  HANGUL SYLLABLE-C6BE
        0xC6BF, # G+14689  U+C6BF  uniC6BF,  HANGUL SYLLABLE-C6BF
        0xC6C2, # G+14690  U+C6C2  uniC6C2,  HANGUL SYLLABLE-C6C2
        0xC6C4, # G+14691  U+C6C4  uniC6C4,  HANGUL SYLLABLE-C6C4
        0xC6C6, # G+14692  U+C6C6  uniC6C6,  HANGUL SYLLABLE-C6C6
        0xC6C7, # G+14693  U+C6C7  uniC6C7,  HANGUL SYLLABLE-C6C7
        0xC6C8, # G+14694  U+C6C8  uniC6C8,  HANGUL SYLLABLE-C6C8
        0xC6C9, # G+14695  U+C6C9  uniC6C9,  HANGUL SYLLABLE-C6C9
        0xC6CA, # G+14696  U+C6CA  uniC6CA,  HANGUL SYLLABLE-C6CA
        0xC6CB, # G+14697  U+C6CB  uniC6CB,  HANGUL SYLLABLE-C6CB
        0xC6CE, # G+14698  U+C6CE  uniC6CE,  HANGUL SYLLABLE-C6CE
        0xC6CF, # G+14699  U+C6CF  uniC6CF,  HANGUL SYLLABLE-C6CF
        0xC6D1, # G+14700  U+C6D1  uniC6D1,  HANGUL SYLLABLE-C6D1
        0xC6D2, # G+14701  U+C6D2  uniC6D2,  HANGUL SYLLABLE-C6D2
        0xC6D3, # G+14702  U+C6D3  uniC6D3,  HANGUL SYLLABLE-C6D3
        0xC6D5, # G+14703  U+C6D5  uniC6D5,  HANGUL SYLLABLE-C6D5
        0xC6D6, # G+14704  U+C6D6  uniC6D6,  HANGUL SYLLABLE-C6D6
        0xC6D7, # G+14705  U+C6D7  uniC6D7,  HANGUL SYLLABLE-C6D7
        0xC6D8, # G+14706  U+C6D8  uniC6D8,  HANGUL SYLLABLE-C6D8
        0xC6D9, # G+14707  U+C6D9  uniC6D9,  HANGUL SYLLABLE-C6D9
        0xC6DA, # G+14708  U+C6DA  uniC6DA,  HANGUL SYLLABLE-C6DA
        0xC6DB, # G+14709  U+C6DB  uniC6DB,  HANGUL SYLLABLE-C6DB
        0xC6DE, # G+14710  U+C6DE  uniC6DE,  HANGUL SYLLABLE-C6DE
        0xC6DF, # G+14711  U+C6DF  uniC6DF,  HANGUL SYLLABLE-C6DF
        0xC6E2, # G+14712  U+C6E2  uniC6E2,  HANGUL SYLLABLE-C6E2
        0xC6E3, # G+14713  U+C6E3  uniC6E3,  HANGUL SYLLABLE-C6E3
        0xC6E4, # G+14714  U+C6E4  uniC6E4,  HANGUL SYLLABLE-C6E4
        0xC6E5, # G+14715  U+C6E5  uniC6E5,  HANGUL SYLLABLE-C6E5
        0xC6E6, # G+14716  U+C6E6  uniC6E6,  HANGUL SYLLABLE-C6E6
        0xC6E7, # G+14717  U+C6E7  uniC6E7,  HANGUL SYLLABLE-C6E7
        0xC6EA, # G+14718  U+C6EA  uniC6EA,  HANGUL SYLLABLE-C6EA
        0xC6EB, # G+14719  U+C6EB  uniC6EB,  HANGUL SYLLABLE-C6EB
        0xC6ED, # G+14720  U+C6ED  uniC6ED,  HANGUL SYLLABLE-C6ED
        0xC6EE, # G+14721  U+C6EE  uniC6EE,  HANGUL SYLLABLE-C6EE
        0xC6EF, # G+14722  U+C6EF  uniC6EF,  HANGUL SYLLABLE-C6EF
        0xC6F1, # G+14723  U+C6F1  uniC6F1,  HANGUL SYLLABLE-C6F1
        0xC6F2, # G+14724  U+C6F2  uniC6F2,  HANGUL SYLLABLE-C6F2
        0xC6F3, # G+14725  U+C6F3  uniC6F3,  HANGUL SYLLABLE-C6F3
        0xC6F4, # G+14726  U+C6F4  uniC6F4,  HANGUL SYLLABLE-C6F4
        0xC6F5, # G+14727  U+C6F5  uniC6F5,  HANGUL SYLLABLE-C6F5
        0xC6F6, # G+14728  U+C6F6  uniC6F6,  HANGUL SYLLABLE-C6F6
        0xC6F7, # G+14729  U+C6F7  uniC6F7,  HANGUL SYLLABLE-C6F7
        0xC6FA, # G+14730  U+C6FA  uniC6FA,  HANGUL SYLLABLE-C6FA
        0xC6FB, # G+14731  U+C6FB  uniC6FB,  HANGUL SYLLABLE-C6FB
        0xC6FC, # G+14732  U+C6FC  uniC6FC,  HANGUL SYLLABLE-C6FC
        0xC6FE, # G+14733  U+C6FE  uniC6FE,  HANGUL SYLLABLE-C6FE
        0xC6FF, # G+14734  U+C6FF  uniC6FF,  HANGUL SYLLABLE-C6FF
        0xC700, # G+14735  U+C700  uniC700,  HANGUL SYLLABLE-C700
        0xC701, # G+14736  U+C701  uniC701,  HANGUL SYLLABLE-C701
        0xC702, # G+14737  U+C702  uniC702,  HANGUL SYLLABLE-C702
        0xC703, # G+14738  U+C703  uniC703,  HANGUL SYLLABLE-C703
        0xC706, # G+14739  U+C706  uniC706,  HANGUL SYLLABLE-C706
        0xC707, # G+14740  U+C707  uniC707,  HANGUL SYLLABLE-C707
        0xC709, # G+14741  U+C709  uniC709,  HANGUL SYLLABLE-C709
        0xC70A, # G+14742  U+C70A  uniC70A,  HANGUL SYLLABLE-C70A
        0xC70B, # G+14743  U+C70B  uniC70B,  HANGUL SYLLABLE-C70B
        0xC70D, # G+14744  U+C70D  uniC70D,  HANGUL SYLLABLE-C70D
        0xC70E, # G+14745  U+C70E  uniC70E,  HANGUL SYLLABLE-C70E
        0xC70F, # G+14746  U+C70F  uniC70F,  HANGUL SYLLABLE-C70F
        0xC710, # G+14747  U+C710  uniC710,  HANGUL SYLLABLE-C710
        0xC711, # G+14748  U+C711  uniC711,  HANGUL SYLLABLE-C711
        0xC712, # G+14749  U+C712  uniC712,  HANGUL SYLLABLE-C712
        0xC713, # G+14750  U+C713  uniC713,  HANGUL SYLLABLE-C713
        0xC716, # G+14751  U+C716  uniC716,  HANGUL SYLLABLE-C716
        0xC718, # G+14752  U+C718  uniC718,  HANGUL SYLLABLE-C718
        0xC71A, # G+14753  U+C71A  uniC71A,  HANGUL SYLLABLE-C71A
        0xC71B, # G+14754  U+C71B  uniC71B,  HANGUL SYLLABLE-C71B
        0xC71C, # G+14755  U+C71C  uniC71C,  HANGUL SYLLABLE-C71C
        0xC71D, # G+14756  U+C71D  uniC71D,  HANGUL SYLLABLE-C71D
        0xC71E, # G+14757  U+C71E  uniC71E,  HANGUL SYLLABLE-C71E
        0xC71F, # G+14758  U+C71F  uniC71F,  HANGUL SYLLABLE-C71F
        0xC722, # G+14759  U+C722  uniC722,  HANGUL SYLLABLE-C722
        0xC723, # G+14760  U+C723  uniC723,  HANGUL SYLLABLE-C723
        0xC725, # G+14761  U+C725  uniC725,  HANGUL SYLLABLE-C725
        0xC726, # G+14762  U+C726  uniC726,  HANGUL SYLLABLE-C726
        0xC727, # G+14763  U+C727  uniC727,  HANGUL SYLLABLE-C727
        0xC729, # G+14764  U+C729  uniC729,  HANGUL SYLLABLE-C729
        0xC72A, # G+14765  U+C72A  uniC72A,  HANGUL SYLLABLE-C72A
        0xC72B, # G+14766  U+C72B  uniC72B,  HANGUL SYLLABLE-C72B
        0xC72C, # G+14767  U+C72C  uniC72C,  HANGUL SYLLABLE-C72C
        0xC72D, # G+14768  U+C72D  uniC72D,  HANGUL SYLLABLE-C72D
        0xC72E, # G+14769  U+C72E  uniC72E,  HANGUL SYLLABLE-C72E
        0xC72F, # G+14770  U+C72F  uniC72F,  HANGUL SYLLABLE-C72F
        0xC732, # G+14771  U+C732  uniC732,  HANGUL SYLLABLE-C732
        0xC734, # G+14772  U+C734  uniC734,  HANGUL SYLLABLE-C734
        0xC736, # G+14773  U+C736  uniC736,  HANGUL SYLLABLE-C736
        0xC738, # G+14774  U+C738  uniC738,  HANGUL SYLLABLE-C738
        0xC739, # G+14775  U+C739  uniC739,  HANGUL SYLLABLE-C739
        0xC73A, # G+14776  U+C73A  uniC73A,  HANGUL SYLLABLE-C73A
        0xC73B, # G+14777  U+C73B  uniC73B,  HANGUL SYLLABLE-C73B
        0xC73E, # G+14778  U+C73E  uniC73E,  HANGUL SYLLABLE-C73E
        0xC73F, # G+14779  U+C73F  uniC73F,  HANGUL SYLLABLE-C73F
        0xC741, # G+14780  U+C741  uniC741,  HANGUL SYLLABLE-C741
        0xC742, # G+14781  U+C742  uniC742,  HANGUL SYLLABLE-C742
        0xC743, # G+14782  U+C743  uniC743,  HANGUL SYLLABLE-C743
        0xC745, # G+14783  U+C745  uniC745,  HANGUL SYLLABLE-C745
        0xC746, # G+14784  U+C746  uniC746,  HANGUL SYLLABLE-C746
        0xC747, # G+14785  U+C747  uniC747,  HANGUL SYLLABLE-C747
        0xC748, # G+14786  U+C748  uniC748,  HANGUL SYLLABLE-C748
        0xC749, # G+14787  U+C749  uniC749,  HANGUL SYLLABLE-C749
        0xC74B, # G+14788  U+C74B  uniC74B,  HANGUL SYLLABLE-C74B
        0xC74E, # G+14789  U+C74E  uniC74E,  HANGUL SYLLABLE-C74E
        0xC750, # G+14790  U+C750  uniC750,  HANGUL SYLLABLE-C750
        0xC759, # G+14791  U+C759  uniC759,  HANGUL SYLLABLE-C759
        0xC75A, # G+14792  U+C75A  uniC75A,  HANGUL SYLLABLE-C75A
        0xC75B, # G+14793  U+C75B  uniC75B,  HANGUL SYLLABLE-C75B
        0xC75D, # G+14794  U+C75D  uniC75D,  HANGUL SYLLABLE-C75D
        0xC75E, # G+14795  U+C75E  uniC75E,  HANGUL SYLLABLE-C75E
        0xC75F, # G+14796  U+C75F  uniC75F,  HANGUL SYLLABLE-C75F
        0xC761, # G+14797  U+C761  uniC761,  HANGUL SYLLABLE-C761
        0xC762, # G+14798  U+C762  uniC762,  HANGUL SYLLABLE-C762
        0xC763, # G+14799  U+C763  uniC763,  HANGUL SYLLABLE-C763
        0xC764, # G+14800  U+C764  uniC764,  HANGUL SYLLABLE-C764
        0xC765, # G+14801  U+C765  uniC765,  HANGUL SYLLABLE-C765
        0xC766, # G+14802  U+C766  uniC766,  HANGUL SYLLABLE-C766
        0xC767, # G+14803  U+C767  uniC767,  HANGUL SYLLABLE-C767
        0xC769, # G+14804  U+C769  uniC769,  HANGUL SYLLABLE-C769
        0xC76A, # G+14805  U+C76A  uniC76A,  HANGUL SYLLABLE-C76A
        0xC76C, # G+14806  U+C76C  uniC76C,  HANGUL SYLLABLE-C76C
        0xC76D, # G+14807  U+C76D  uniC76D,  HANGUL SYLLABLE-C76D
        0xC76E, # G+14808  U+C76E  uniC76E,  HANGUL SYLLABLE-C76E
        0xC76F, # G+14809  U+C76F  uniC76F,  HANGUL SYLLABLE-C76F
        0xC770, # G+14810  U+C770  uniC770,  HANGUL SYLLABLE-C770
        0xC771, # G+14811  U+C771  uniC771,  HANGUL SYLLABLE-C771
        0xC772, # G+14812  U+C772  uniC772,  HANGUL SYLLABLE-C772
        0xC773, # G+14813  U+C773  uniC773,  HANGUL SYLLABLE-C773
        0xC776, # G+14814  U+C776  uniC776,  HANGUL SYLLABLE-C776
        0xC777, # G+14815  U+C777  uniC777,  HANGUL SYLLABLE-C777
        0xC779, # G+14816  U+C779  uniC779,  HANGUL SYLLABLE-C779
        0xC77A, # G+14817  U+C77A  uniC77A,  HANGUL SYLLABLE-C77A
        0xC77B, # G+14818  U+C77B  uniC77B,  HANGUL SYLLABLE-C77B
        0xC77F, # G+14819  U+C77F  uniC77F,  HANGUL SYLLABLE-C77F
        0xC780, # G+14820  U+C780  uniC780,  HANGUL SYLLABLE-C780
        0xC781, # G+14821  U+C781  uniC781,  HANGUL SYLLABLE-C781
        0xC782, # G+14822  U+C782  uniC782,  HANGUL SYLLABLE-C782
        0xC786, # G+14823  U+C786  uniC786,  HANGUL SYLLABLE-C786
        0xC78B, # G+14824  U+C78B  uniC78B,  HANGUL SYLLABLE-C78B
        0xC78C, # G+14825  U+C78C  uniC78C,  HANGUL SYLLABLE-C78C
        0xC78D, # G+14826  U+C78D  uniC78D,  HANGUL SYLLABLE-C78D
        0xC78F, # G+14827  U+C78F  uniC78F,  HANGUL SYLLABLE-C78F
        0xC792, # G+14828  U+C792  uniC792,  HANGUL SYLLABLE-C792
        0xC793, # G+14829  U+C793  uniC793,  HANGUL SYLLABLE-C793
        0xC795, # G+14830  U+C795  uniC795,  HANGUL SYLLABLE-C795
        0xC799, # G+14831  U+C799  uniC799,  HANGUL SYLLABLE-C799
        0xC79B, # G+14832  U+C79B  uniC79B,  HANGUL SYLLABLE-C79B
        0xC79C, # G+14833  U+C79C  uniC79C,  HANGUL SYLLABLE-C79C
        0xC79D, # G+14834  U+C79D  uniC79D,  HANGUL SYLLABLE-C79D
        0xC79E, # G+14835  U+C79E  uniC79E,  HANGUL SYLLABLE-C79E
        0xC79F, # G+14836  U+C79F  uniC79F,  HANGUL SYLLABLE-C79F
        0xC7A2, # G+14837  U+C7A2  uniC7A2,  HANGUL SYLLABLE-C7A2
        0xC7A7, # G+14838  U+C7A7  uniC7A7,  HANGUL SYLLABLE-C7A7
        0xC7A8, # G+14839  U+C7A8  uniC7A8,  HANGUL SYLLABLE-C7A8
        0xC7A9, # G+14840  U+C7A9  uniC7A9,  HANGUL SYLLABLE-C7A9
        0xC7AA, # G+14841  U+C7AA  uniC7AA,  HANGUL SYLLABLE-C7AA
        0xC7AB, # G+14842  U+C7AB  uniC7AB,  HANGUL SYLLABLE-C7AB
        0xC7AE, # G+14843  U+C7AE  uniC7AE,  HANGUL SYLLABLE-C7AE
        0xC7AF, # G+14844  U+C7AF  uniC7AF,  HANGUL SYLLABLE-C7AF
        0xC7B1, # G+14845  U+C7B1  uniC7B1,  HANGUL SYLLABLE-C7B1
        0xC7B2, # G+14846  U+C7B2  uniC7B2,  HANGUL SYLLABLE-C7B2
        0xC7B3, # G+14847  U+C7B3  uniC7B3,  HANGUL SYLLABLE-C7B3
        0xC7B5, # G+14848  U+C7B5  uniC7B5,  HANGUL SYLLABLE-C7B5
        0xC7B6, # G+14849  U+C7B6  uniC7B6,  HANGUL SYLLABLE-C7B6
        0xC7B7, # G+14850  U+C7B7  uniC7B7,  HANGUL SYLLABLE-C7B7
        0xC7B8, # G+14851  U+C7B8  uniC7B8,  HANGUL SYLLABLE-C7B8
        0xC7B9, # G+14852  U+C7B9  uniC7B9,  HANGUL SYLLABLE-C7B9
        0xC7BA, # G+14853  U+C7BA  uniC7BA,  HANGUL SYLLABLE-C7BA
        0xC7BB, # G+14854  U+C7BB  uniC7BB,  HANGUL SYLLABLE-C7BB
        0xC7BE, # G+14855  U+C7BE  uniC7BE,  HANGUL SYLLABLE-C7BE
        0xC7C2, # G+14856  U+C7C2  uniC7C2,  HANGUL SYLLABLE-C7C2
        0xC7C3, # G+14857  U+C7C3  uniC7C3,  HANGUL SYLLABLE-C7C3
        0xC7C4, # G+14858  U+C7C4  uniC7C4,  HANGUL SYLLABLE-C7C4
        0xC7C5, # G+14859  U+C7C5  uniC7C5,  HANGUL SYLLABLE-C7C5
        0xC7C6, # G+14860  U+C7C6  uniC7C6,  HANGUL SYLLABLE-C7C6
        0xC7C7, # G+14861  U+C7C7  uniC7C7,  HANGUL SYLLABLE-C7C7
        0xC7CA, # G+14862  U+C7CA  uniC7CA,  HANGUL SYLLABLE-C7CA
        0xC7CB, # G+14863  U+C7CB  uniC7CB,  HANGUL SYLLABLE-C7CB
        0xC7CD, # G+14864  U+C7CD  uniC7CD,  HANGUL SYLLABLE-C7CD
        0xC7CF, # G+14865  U+C7CF  uniC7CF,  HANGUL SYLLABLE-C7CF
        0xC7D1, # G+14866  U+C7D1  uniC7D1,  HANGUL SYLLABLE-C7D1
        0xC7D2, # G+14867  U+C7D2  uniC7D2,  HANGUL SYLLABLE-C7D2
        0xC7D3, # G+14868  U+C7D3  uniC7D3,  HANGUL SYLLABLE-C7D3
        0xC7D4, # G+14869  U+C7D4  uniC7D4,  HANGUL SYLLABLE-C7D4
        0xC7D5, # G+14870  U+C7D5  uniC7D5,  HANGUL SYLLABLE-C7D5
        0xC7D6, # G+14871  U+C7D6  uniC7D6,  HANGUL SYLLABLE-C7D6
        0xC7D7, # G+14872  U+C7D7  uniC7D7,  HANGUL SYLLABLE-C7D7
        0xC7D9, # G+14873  U+C7D9  uniC7D9,  HANGUL SYLLABLE-C7D9
        0xC7DA, # G+14874  U+C7DA  uniC7DA,  HANGUL SYLLABLE-C7DA
        0xC7DB, # G+14875  U+C7DB  uniC7DB,  HANGUL SYLLABLE-C7DB
        0xC7DC, # G+14876  U+C7DC  uniC7DC,  HANGUL SYLLABLE-C7DC
        0xC7DE, # G+14877  U+C7DE  uniC7DE,  HANGUL SYLLABLE-C7DE
        0xC7DF, # G+14878  U+C7DF  uniC7DF,  HANGUL SYLLABLE-C7DF
        0xC7E0, # G+14879  U+C7E0  uniC7E0,  HANGUL SYLLABLE-C7E0
        0xC7E1, # G+14880  U+C7E1  uniC7E1,  HANGUL SYLLABLE-C7E1
        0xC7E2, # G+14881  U+C7E2  uniC7E2,  HANGUL SYLLABLE-C7E2
        0xC7E3, # G+14882  U+C7E3  uniC7E3,  HANGUL SYLLABLE-C7E3
        0xC7E5, # G+14883  U+C7E5  uniC7E5,  HANGUL SYLLABLE-C7E5
        0xC7E6, # G+14884  U+C7E6  uniC7E6,  HANGUL SYLLABLE-C7E6
        0xC7E7, # G+14885  U+C7E7  uniC7E7,  HANGUL SYLLABLE-C7E7
        0xC7E9, # G+14886  U+C7E9  uniC7E9,  HANGUL SYLLABLE-C7E9
        0xC7EA, # G+14887  U+C7EA  uniC7EA,  HANGUL SYLLABLE-C7EA
        0xC7EB, # G+14888  U+C7EB  uniC7EB,  HANGUL SYLLABLE-C7EB
        0xC7ED, # G+14889  U+C7ED  uniC7ED,  HANGUL SYLLABLE-C7ED
        0xC7EE, # G+14890  U+C7EE  uniC7EE,  HANGUL SYLLABLE-C7EE
        0xC7EF, # G+14891  U+C7EF  uniC7EF,  HANGUL SYLLABLE-C7EF
        0xC7F0, # G+14892  U+C7F0  uniC7F0,  HANGUL SYLLABLE-C7F0
        0xC7F1, # G+14893  U+C7F1  uniC7F1,  HANGUL SYLLABLE-C7F1
        0xC7F2, # G+14894  U+C7F2  uniC7F2,  HANGUL SYLLABLE-C7F2
        0xC7F3, # G+14895  U+C7F3  uniC7F3,  HANGUL SYLLABLE-C7F3
        0xC7F4, # G+14896  U+C7F4  uniC7F4,  HANGUL SYLLABLE-C7F4
        0xC7F5, # G+14897  U+C7F5  uniC7F5,  HANGUL SYLLABLE-C7F5
        0xC7F6, # G+14898  U+C7F6  uniC7F6,  HANGUL SYLLABLE-C7F6
        0xC7F7, # G+14899  U+C7F7  uniC7F7,  HANGUL SYLLABLE-C7F7
        0xC7F8, # G+14900  U+C7F8  uniC7F8,  HANGUL SYLLABLE-C7F8
        0xC7F9, # G+14901  U+C7F9  uniC7F9,  HANGUL SYLLABLE-C7F9
        0xC7FA, # G+14902  U+C7FA  uniC7FA,  HANGUL SYLLABLE-C7FA
        0xC7FB, # G+14903  U+C7FB  uniC7FB,  HANGUL SYLLABLE-C7FB
        0xC7FC, # G+14904  U+C7FC  uniC7FC,  HANGUL SYLLABLE-C7FC
        0xC7FD, # G+14905  U+C7FD  uniC7FD,  HANGUL SYLLABLE-C7FD
        0xC7FE, # G+14906  U+C7FE  uniC7FE,  HANGUL SYLLABLE-C7FE
        0xC7FF, # G+14907  U+C7FF  uniC7FF,  HANGUL SYLLABLE-C7FF
        0xC802, # G+14908  U+C802  uniC802,  HANGUL SYLLABLE-C802
        0xC803, # G+14909  U+C803  uniC803,  HANGUL SYLLABLE-C803
        0xC805, # G+14910  U+C805  uniC805,  HANGUL SYLLABLE-C805
        0xC806, # G+14911  U+C806  uniC806,  HANGUL SYLLABLE-C806
        0xC807, # G+14912  U+C807  uniC807,  HANGUL SYLLABLE-C807
        0xC809, # G+14913  U+C809  uniC809,  HANGUL SYLLABLE-C809
        0xC80B, # G+14914  U+C80B  uniC80B,  HANGUL SYLLABLE-C80B
        0xC80C, # G+14915  U+C80C  uniC80C,  HANGUL SYLLABLE-C80C
        0xC80D, # G+14916  U+C80D  uniC80D,  HANGUL SYLLABLE-C80D
        0xC80E, # G+14917  U+C80E  uniC80E,  HANGUL SYLLABLE-C80E
        0xC80F, # G+14918  U+C80F  uniC80F,  HANGUL SYLLABLE-C80F
        0xC812, # G+14919  U+C812  uniC812,  HANGUL SYLLABLE-C812
        0xC814, # G+14920  U+C814  uniC814,  HANGUL SYLLABLE-C814
        0xC817, # G+14921  U+C817  uniC817,  HANGUL SYLLABLE-C817
        0xC818, # G+14922  U+C818  uniC818,  HANGUL SYLLABLE-C818
        0xC819, # G+14923  U+C819  uniC819,  HANGUL SYLLABLE-C819
        0xC81A, # G+14924  U+C81A  uniC81A,  HANGUL SYLLABLE-C81A
        0xC81B, # G+14925  U+C81B  uniC81B,  HANGUL SYLLABLE-C81B
        0xC81E, # G+14926  U+C81E  uniC81E,  HANGUL SYLLABLE-C81E
        0xC81F, # G+14927  U+C81F  uniC81F,  HANGUL SYLLABLE-C81F
        0xC821, # G+14928  U+C821  uniC821,  HANGUL SYLLABLE-C821
        0xC822, # G+14929  U+C822  uniC822,  HANGUL SYLLABLE-C822
        0xC823, # G+14930  U+C823  uniC823,  HANGUL SYLLABLE-C823
        0xC825, # G+14931  U+C825  uniC825,  HANGUL SYLLABLE-C825
        0xC826, # G+14932  U+C826  uniC826,  HANGUL SYLLABLE-C826
        0xC827, # G+14933  U+C827  uniC827,  HANGUL SYLLABLE-C827
        0xC828, # G+14934  U+C828  uniC828,  HANGUL SYLLABLE-C828
        0xC829, # G+14935  U+C829  uniC829,  HANGUL SYLLABLE-C829
        0xC82A, # G+14936  U+C82A  uniC82A,  HANGUL SYLLABLE-C82A
        0xC82B, # G+14937  U+C82B  uniC82B,  HANGUL SYLLABLE-C82B
        0xC82E, # G+14938  U+C82E  uniC82E,  HANGUL SYLLABLE-C82E
        0xC830, # G+14939  U+C830  uniC830,  HANGUL SYLLABLE-C830
        0xC832, # G+14940  U+C832  uniC832,  HANGUL SYLLABLE-C832
        0xC833, # G+14941  U+C833  uniC833,  HANGUL SYLLABLE-C833
        0xC834, # G+14942  U+C834  uniC834,  HANGUL SYLLABLE-C834
        0xC835, # G+14943  U+C835  uniC835,  HANGUL SYLLABLE-C835
        0xC836, # G+14944  U+C836  uniC836,  HANGUL SYLLABLE-C836
        0xC837, # G+14945  U+C837  uniC837,  HANGUL SYLLABLE-C837
        0xC839, # G+14946  U+C839  uniC839,  HANGUL SYLLABLE-C839
        0xC83A, # G+14947  U+C83A  uniC83A,  HANGUL SYLLABLE-C83A
        0xC83B, # G+14948  U+C83B  uniC83B,  HANGUL SYLLABLE-C83B
        0xC83D, # G+14949  U+C83D  uniC83D,  HANGUL SYLLABLE-C83D
        0xC83E, # G+14950  U+C83E  uniC83E,  HANGUL SYLLABLE-C83E
        0xC83F, # G+14951  U+C83F  uniC83F,  HANGUL SYLLABLE-C83F
        0xC841, # G+14952  U+C841  uniC841,  HANGUL SYLLABLE-C841
        0xC842, # G+14953  U+C842  uniC842,  HANGUL SYLLABLE-C842
        0xC843, # G+14954  U+C843  uniC843,  HANGUL SYLLABLE-C843
        0xC844, # G+14955  U+C844  uniC844,  HANGUL SYLLABLE-C844
        0xC845, # G+14956  U+C845  uniC845,  HANGUL SYLLABLE-C845
        0xC846, # G+14957  U+C846  uniC846,  HANGUL SYLLABLE-C846
        0xC847, # G+14958  U+C847  uniC847,  HANGUL SYLLABLE-C847
        0xC84A, # G+14959  U+C84A  uniC84A,  HANGUL SYLLABLE-C84A
        0xC84B, # G+14960  U+C84B  uniC84B,  HANGUL SYLLABLE-C84B
        0xC84E, # G+14961  U+C84E  uniC84E,  HANGUL SYLLABLE-C84E
        0xC84F, # G+14962  U+C84F  uniC84F,  HANGUL SYLLABLE-C84F
        0xC850, # G+14963  U+C850  uniC850,  HANGUL SYLLABLE-C850
        0xC851, # G+14964  U+C851  uniC851,  HANGUL SYLLABLE-C851
        0xC852, # G+14965  U+C852  uniC852,  HANGUL SYLLABLE-C852
        0xC853, # G+14966  U+C853  uniC853,  HANGUL SYLLABLE-C853
        0xC855, # G+14967  U+C855  uniC855,  HANGUL SYLLABLE-C855
        0xC856, # G+14968  U+C856  uniC856,  HANGUL SYLLABLE-C856
        0xC857, # G+14969  U+C857  uniC857,  HANGUL SYLLABLE-C857
        0xC858, # G+14970  U+C858  uniC858,  HANGUL SYLLABLE-C858
        0xC859, # G+14971  U+C859  uniC859,  HANGUL SYLLABLE-C859
        0xC85A, # G+14972  U+C85A  uniC85A,  HANGUL SYLLABLE-C85A
        0xC85B, # G+14973  U+C85B  uniC85B,  HANGUL SYLLABLE-C85B
        0xC85C, # G+14974  U+C85C  uniC85C,  HANGUL SYLLABLE-C85C
        0xC85D, # G+14975  U+C85D  uniC85D,  HANGUL SYLLABLE-C85D
        0xC85E, # G+14976  U+C85E  uniC85E,  HANGUL SYLLABLE-C85E
        0xC85F, # G+14977  U+C85F  uniC85F,  HANGUL SYLLABLE-C85F
        0xC860, # G+14978  U+C860  uniC860,  HANGUL SYLLABLE-C860
        0xC861, # G+14979  U+C861  uniC861,  HANGUL SYLLABLE-C861
        0xC862, # G+14980  U+C862  uniC862,  HANGUL SYLLABLE-C862
        0xC863, # G+14981  U+C863  uniC863,  HANGUL SYLLABLE-C863
        0xC864, # G+14982  U+C864  uniC864,  HANGUL SYLLABLE-C864
        0xC865, # G+14983  U+C865  uniC865,  HANGUL SYLLABLE-C865
        0xC866, # G+14984  U+C866  uniC866,  HANGUL SYLLABLE-C866
        0xC867, # G+14985  U+C867  uniC867,  HANGUL SYLLABLE-C867
        0xC868, # G+14986  U+C868  uniC868,  HANGUL SYLLABLE-C868
        0xC869, # G+14987  U+C869  uniC869,  HANGUL SYLLABLE-C869
        0xC86A, # G+14988  U+C86A  uniC86A,  HANGUL SYLLABLE-C86A
        0xC86B, # G+14989  U+C86B  uniC86B,  HANGUL SYLLABLE-C86B
        0xC86C, # G+14990  U+C86C  uniC86C,  HANGUL SYLLABLE-C86C
        0xC86D, # G+14991  U+C86D  uniC86D,  HANGUL SYLLABLE-C86D
        0xC86E, # G+14992  U+C86E  uniC86E,  HANGUL SYLLABLE-C86E
        0xC86F, # G+14993  U+C86F  uniC86F,  HANGUL SYLLABLE-C86F
        0xC872, # G+14994  U+C872  uniC872,  HANGUL SYLLABLE-C872
        0xC873, # G+14995  U+C873  uniC873,  HANGUL SYLLABLE-C873
        0xC875, # G+14996  U+C875  uniC875,  HANGUL SYLLABLE-C875
        0xC876, # G+14997  U+C876  uniC876,  HANGUL SYLLABLE-C876
        0xC877, # G+14998  U+C877  uniC877,  HANGUL SYLLABLE-C877
        0xC879, # G+14999  U+C879  uniC879,  HANGUL SYLLABLE-C879
        0xC87B, # G+15000  U+C87B  uniC87B,  HANGUL SYLLABLE-C87B
        0xC87C, # G+15001  U+C87C  uniC87C,  HANGUL SYLLABLE-C87C
        0xC87D, # G+15002  U+C87D  uniC87D,  HANGUL SYLLABLE-C87D
        0xC87E, # G+15003  U+C87E  uniC87E,  HANGUL SYLLABLE-C87E
        0xC87F, # G+15004  U+C87F  uniC87F,  HANGUL SYLLABLE-C87F
        0xC882, # G+15005  U+C882  uniC882,  HANGUL SYLLABLE-C882
        0xC884, # G+15006  U+C884  uniC884,  HANGUL SYLLABLE-C884
        0xC888, # G+15007  U+C888  uniC888,  HANGUL SYLLABLE-C888
        0xC889, # G+15008  U+C889  uniC889,  HANGUL SYLLABLE-C889
        0xC88A, # G+15009  U+C88A  uniC88A,  HANGUL SYLLABLE-C88A
        0xC88E, # G+15010  U+C88E  uniC88E,  HANGUL SYLLABLE-C88E
        0xC88F, # G+15011  U+C88F  uniC88F,  HANGUL SYLLABLE-C88F
        0xC890, # G+15012  U+C890  uniC890,  HANGUL SYLLABLE-C890
        0xC891, # G+15013  U+C891  uniC891,  HANGUL SYLLABLE-C891
        0xC892, # G+15014  U+C892  uniC892,  HANGUL SYLLABLE-C892
        0xC893, # G+15015  U+C893  uniC893,  HANGUL SYLLABLE-C893
        0xC895, # G+15016  U+C895  uniC895,  HANGUL SYLLABLE-C895
        0xC896, # G+15017  U+C896  uniC896,  HANGUL SYLLABLE-C896
        0xC897, # G+15018  U+C897  uniC897,  HANGUL SYLLABLE-C897
        0xC898, # G+15019  U+C898  uniC898,  HANGUL SYLLABLE-C898
        0xC899, # G+15020  U+C899  uniC899,  HANGUL SYLLABLE-C899
        0xC89A, # G+15021  U+C89A  uniC89A,  HANGUL SYLLABLE-C89A
        0xC89B, # G+15022  U+C89B  uniC89B,  HANGUL SYLLABLE-C89B
        0xC89C, # G+15023  U+C89C  uniC89C,  HANGUL SYLLABLE-C89C
        0xC89E, # G+15024  U+C89E  uniC89E,  HANGUL SYLLABLE-C89E
        0xC8A0, # G+15025  U+C8A0  uniC8A0,  HANGUL SYLLABLE-C8A0
        0xC8A2, # G+15026  U+C8A2  uniC8A2,  HANGUL SYLLABLE-C8A2
        0xC8A3, # G+15027  U+C8A3  uniC8A3,  HANGUL SYLLABLE-C8A3
        0xC8A4, # G+15028  U+C8A4  uniC8A4,  HANGUL SYLLABLE-C8A4
        0xC8A5, # G+15029  U+C8A5  uniC8A5,  HANGUL SYLLABLE-C8A5
        0xC8A6, # G+15030  U+C8A6  uniC8A6,  HANGUL SYLLABLE-C8A6
        0xC8A7, # G+15031  U+C8A7  uniC8A7,  HANGUL SYLLABLE-C8A7
        0xC8A9, # G+15032  U+C8A9  uniC8A9,  HANGUL SYLLABLE-C8A9
        0xC8AA, # G+15033  U+C8AA  uniC8AA,  HANGUL SYLLABLE-C8AA
        0xC8AB, # G+15034  U+C8AB  uniC8AB,  HANGUL SYLLABLE-C8AB
        0xC8AC, # G+15035  U+C8AC  uniC8AC,  HANGUL SYLLABLE-C8AC
        0xC8AD, # G+15036  U+C8AD  uniC8AD,  HANGUL SYLLABLE-C8AD
        0xC8AE, # G+15037  U+C8AE  uniC8AE,  HANGUL SYLLABLE-C8AE
        0xC8AF, # G+15038  U+C8AF  uniC8AF,  HANGUL SYLLABLE-C8AF
        0xC8B0, # G+15039  U+C8B0  uniC8B0,  HANGUL SYLLABLE-C8B0
        0xC8B1, # G+15040  U+C8B1  uniC8B1,  HANGUL SYLLABLE-C8B1
        0xC8B2, # G+15041  U+C8B2  uniC8B2,  HANGUL SYLLABLE-C8B2
        0xC8B3, # G+15042  U+C8B3  uniC8B3,  HANGUL SYLLABLE-C8B3
        0xC8B4, # G+15043  U+C8B4  uniC8B4,  HANGUL SYLLABLE-C8B4
        0xC8B5, # G+15044  U+C8B5  uniC8B5,  HANGUL SYLLABLE-C8B5
        0xC8B6, # G+15045  U+C8B6  uniC8B6,  HANGUL SYLLABLE-C8B6
        0xC8B7, # G+15046  U+C8B7  uniC8B7,  HANGUL SYLLABLE-C8B7
        0xC8B8, # G+15047  U+C8B8  uniC8B8,  HANGUL SYLLABLE-C8B8
        0xC8B9, # G+15048  U+C8B9  uniC8B9,  HANGUL SYLLABLE-C8B9
        0xC8BA, # G+15049  U+C8BA  uniC8BA,  HANGUL SYLLABLE-C8BA
        0xC8BB, # G+15050  U+C8BB  uniC8BB,  HANGUL SYLLABLE-C8BB
        0xC8BE, # G+15051  U+C8BE  uniC8BE,  HANGUL SYLLABLE-C8BE
        0xC8BF, # G+15052  U+C8BF  uniC8BF,  HANGUL SYLLABLE-C8BF
        0xC8C0, # G+15053  U+C8C0  uniC8C0,  HANGUL SYLLABLE-C8C0
        0xC8C1, # G+15054  U+C8C1  uniC8C1,  HANGUL SYLLABLE-C8C1
        0xC8C2, # G+15055  U+C8C2  uniC8C2,  HANGUL SYLLABLE-C8C2
        0xC8C3, # G+15056  U+C8C3  uniC8C3,  HANGUL SYLLABLE-C8C3
        0xC8C5, # G+15057  U+C8C5  uniC8C5,  HANGUL SYLLABLE-C8C5
        0xC8C6, # G+15058  U+C8C6  uniC8C6,  HANGUL SYLLABLE-C8C6
        0xC8C7, # G+15059  U+C8C7  uniC8C7,  HANGUL SYLLABLE-C8C7
        0xC8C9, # G+15060  U+C8C9  uniC8C9,  HANGUL SYLLABLE-C8C9
        0xC8CA, # G+15061  U+C8CA  uniC8CA,  HANGUL SYLLABLE-C8CA
        0xC8CB, # G+15062  U+C8CB  uniC8CB,  HANGUL SYLLABLE-C8CB
        0xC8CD, # G+15063  U+C8CD  uniC8CD,  HANGUL SYLLABLE-C8CD
        0xC8CE, # G+15064  U+C8CE  uniC8CE,  HANGUL SYLLABLE-C8CE
        0xC8CF, # G+15065  U+C8CF  uniC8CF,  HANGUL SYLLABLE-C8CF
        0xC8D0, # G+15066  U+C8D0  uniC8D0,  HANGUL SYLLABLE-C8D0
        0xC8D1, # G+15067  U+C8D1  uniC8D1,  HANGUL SYLLABLE-C8D1
        0xC8D2, # G+15068  U+C8D2  uniC8D2,  HANGUL SYLLABLE-C8D2
        0xC8D3, # G+15069  U+C8D3  uniC8D3,  HANGUL SYLLABLE-C8D3
        0xC8D6, # G+15070  U+C8D6  uniC8D6,  HANGUL SYLLABLE-C8D6
        0xC8D8, # G+15071  U+C8D8  uniC8D8,  HANGUL SYLLABLE-C8D8
        0xC8DA, # G+15072  U+C8DA  uniC8DA,  HANGUL SYLLABLE-C8DA
        0xC8DB, # G+15073  U+C8DB  uniC8DB,  HANGUL SYLLABLE-C8DB
        0xC8DC, # G+15074  U+C8DC  uniC8DC,  HANGUL SYLLABLE-C8DC
        0xC8DD, # G+15075  U+C8DD  uniC8DD,  HANGUL SYLLABLE-C8DD
        0xC8DE, # G+15076  U+C8DE  uniC8DE,  HANGUL SYLLABLE-C8DE
        0xC8DF, # G+15077  U+C8DF  uniC8DF,  HANGUL SYLLABLE-C8DF
        0xC8E2, # G+15078  U+C8E2  uniC8E2,  HANGUL SYLLABLE-C8E2
        0xC8E3, # G+15079  U+C8E3  uniC8E3,  HANGUL SYLLABLE-C8E3
        0xC8E5, # G+15080  U+C8E5  uniC8E5,  HANGUL SYLLABLE-C8E5
        0xC8E6, # G+15081  U+C8E6  uniC8E6,  HANGUL SYLLABLE-C8E6
        0xC8E7, # G+15082  U+C8E7  uniC8E7,  HANGUL SYLLABLE-C8E7
        0xC8E8, # G+15083  U+C8E8  uniC8E8,  HANGUL SYLLABLE-C8E8
        0xC8E9, # G+15084  U+C8E9  uniC8E9,  HANGUL SYLLABLE-C8E9
        0xC8EA, # G+15085  U+C8EA  uniC8EA,  HANGUL SYLLABLE-C8EA
        0xC8EB, # G+15086  U+C8EB  uniC8EB,  HANGUL SYLLABLE-C8EB
        0xC8EC, # G+15087  U+C8EC  uniC8EC,  HANGUL SYLLABLE-C8EC
        0xC8ED, # G+15088  U+C8ED  uniC8ED,  HANGUL SYLLABLE-C8ED
        0xC8EE, # G+15089  U+C8EE  uniC8EE,  HANGUL SYLLABLE-C8EE
        0xC8EF, # G+15090  U+C8EF  uniC8EF,  HANGUL SYLLABLE-C8EF
        0xC8F0, # G+15091  U+C8F0  uniC8F0,  HANGUL SYLLABLE-C8F0
        0xC8F1, # G+15092  U+C8F1  uniC8F1,  HANGUL SYLLABLE-C8F1
        0xC8F2, # G+15093  U+C8F2  uniC8F2,  HANGUL SYLLABLE-C8F2
        0xC8F3, # G+15094  U+C8F3  uniC8F3,  HANGUL SYLLABLE-C8F3
        0xC8F4, # G+15095  U+C8F4  uniC8F4,  HANGUL SYLLABLE-C8F4
        0xC8F6, # G+15096  U+C8F6  uniC8F6,  HANGUL SYLLABLE-C8F6
        0xC8F7, # G+15097  U+C8F7  uniC8F7,  HANGUL SYLLABLE-C8F7
        0xC8F8, # G+15098  U+C8F8  uniC8F8,  HANGUL SYLLABLE-C8F8
        0xC8F9, # G+15099  U+C8F9  uniC8F9,  HANGUL SYLLABLE-C8F9
        0xC8FA, # G+15100  U+C8FA  uniC8FA,  HANGUL SYLLABLE-C8FA
        0xC8FB, # G+15101  U+C8FB  uniC8FB,  HANGUL SYLLABLE-C8FB
        0xC8FE, # G+15102  U+C8FE  uniC8FE,  HANGUL SYLLABLE-C8FE
        0xC8FF, # G+15103  U+C8FF  uniC8FF,  HANGUL SYLLABLE-C8FF
        0xC901, # G+15104  U+C901  uniC901,  HANGUL SYLLABLE-C901
        0xC902, # G+15105  U+C902  uniC902,  HANGUL SYLLABLE-C902
        0xC903, # G+15106  U+C903  uniC903,  HANGUL SYLLABLE-C903
        0xC907, # G+15107  U+C907  uniC907,  HANGUL SYLLABLE-C907
        0xC908, # G+15108  U+C908  uniC908,  HANGUL SYLLABLE-C908
        0xC909, # G+15109  U+C909  uniC909,  HANGUL SYLLABLE-C909
        0xC90A, # G+15110  U+C90A  uniC90A,  HANGUL SYLLABLE-C90A
        0xC90B, # G+15111  U+C90B  uniC90B,  HANGUL SYLLABLE-C90B
        0xC90E, # G+15112  U+C90E  uniC90E,  HANGUL SYLLABLE-C90E
        0xC910, # G+15113  U+C910  uniC910,  HANGUL SYLLABLE-C910
        0xC912, # G+15114  U+C912  uniC912,  HANGUL SYLLABLE-C912
        0xC913, # G+15115  U+C913  uniC913,  HANGUL SYLLABLE-C913
        0xC914, # G+15116  U+C914  uniC914,  HANGUL SYLLABLE-C914
        0xC915, # G+15117  U+C915  uniC915,  HANGUL SYLLABLE-C915
        0xC916, # G+15118  U+C916  uniC916,  HANGUL SYLLABLE-C916
        0xC917, # G+15119  U+C917  uniC917,  HANGUL SYLLABLE-C917
        0xC919, # G+15120  U+C919  uniC919,  HANGUL SYLLABLE-C919
        0xC91A, # G+15121  U+C91A  uniC91A,  HANGUL SYLLABLE-C91A
        0xC91B, # G+15122  U+C91B  uniC91B,  HANGUL SYLLABLE-C91B
        0xC91C, # G+15123  U+C91C  uniC91C,  HANGUL SYLLABLE-C91C
        0xC91D, # G+15124  U+C91D  uniC91D,  HANGUL SYLLABLE-C91D
        0xC91E, # G+15125  U+C91E  uniC91E,  HANGUL SYLLABLE-C91E
        0xC91F, # G+15126  U+C91F  uniC91F,  HANGUL SYLLABLE-C91F
        0xC920, # G+15127  U+C920  uniC920,  HANGUL SYLLABLE-C920
        0xC921, # G+15128  U+C921  uniC921,  HANGUL SYLLABLE-C921
        0xC922, # G+15129  U+C922  uniC922,  HANGUL SYLLABLE-C922
        0xC923, # G+15130  U+C923  uniC923,  HANGUL SYLLABLE-C923
        0xC924, # G+15131  U+C924  uniC924,  HANGUL SYLLABLE-C924
        0xC925, # G+15132  U+C925  uniC925,  HANGUL SYLLABLE-C925
        0xC926, # G+15133  U+C926  uniC926,  HANGUL SYLLABLE-C926
        0xC927, # G+15134  U+C927  uniC927,  HANGUL SYLLABLE-C927
        0xC928, # G+15135  U+C928  uniC928,  HANGUL SYLLABLE-C928
        0xC929, # G+15136  U+C929  uniC929,  HANGUL SYLLABLE-C929
        0xC92A, # G+15137  U+C92A  uniC92A,  HANGUL SYLLABLE-C92A
        0xC92B, # G+15138  U+C92B  uniC92B,  HANGUL SYLLABLE-C92B
        0xC92D, # G+15139  U+C92D  uniC92D,  HANGUL SYLLABLE-C92D
        0xC92E, # G+15140  U+C92E  uniC92E,  HANGUL SYLLABLE-C92E
        0xC92F, # G+15141  U+C92F  uniC92F,  HANGUL SYLLABLE-C92F
        0xC930, # G+15142  U+C930  uniC930,  HANGUL SYLLABLE-C930
        0xC931, # G+15143  U+C931  uniC931,  HANGUL SYLLABLE-C931
        0xC932, # G+15144  U+C932  uniC932,  HANGUL SYLLABLE-C932
        0xC933, # G+15145  U+C933  uniC933,  HANGUL SYLLABLE-C933
        0xC935, # G+15146  U+C935  uniC935,  HANGUL SYLLABLE-C935
        0xC936, # G+15147  U+C936  uniC936,  HANGUL SYLLABLE-C936
        0xC937, # G+15148  U+C937  uniC937,  HANGUL SYLLABLE-C937
        0xC938, # G+15149  U+C938  uniC938,  HANGUL SYLLABLE-C938
        0xC939, # G+15150  U+C939  uniC939,  HANGUL SYLLABLE-C939
        0xC93A, # G+15151  U+C93A  uniC93A,  HANGUL SYLLABLE-C93A
        0xC93B, # G+15152  U+C93B  uniC93B,  HANGUL SYLLABLE-C93B
        0xC93C, # G+15153  U+C93C  uniC93C,  HANGUL SYLLABLE-C93C
        0xC93D, # G+15154  U+C93D  uniC93D,  HANGUL SYLLABLE-C93D
        0xC93E, # G+15155  U+C93E  uniC93E,  HANGUL SYLLABLE-C93E
        0xC93F, # G+15156  U+C93F  uniC93F,  HANGUL SYLLABLE-C93F
        0xC940, # G+15157  U+C940  uniC940,  HANGUL SYLLABLE-C940
        0xC941, # G+15158  U+C941  uniC941,  HANGUL SYLLABLE-C941
        0xC942, # G+15159  U+C942  uniC942,  HANGUL SYLLABLE-C942
        0xC943, # G+15160  U+C943  uniC943,  HANGUL SYLLABLE-C943
        0xC944, # G+15161  U+C944  uniC944,  HANGUL SYLLABLE-C944
        0xC945, # G+15162  U+C945  uniC945,  HANGUL SYLLABLE-C945
        0xC946, # G+15163  U+C946  uniC946,  HANGUL SYLLABLE-C946
        0xC947, # G+15164  U+C947  uniC947,  HANGUL SYLLABLE-C947
        0xC948, # G+15165  U+C948  uniC948,  HANGUL SYLLABLE-C948
        0xC949, # G+15166  U+C949  uniC949,  HANGUL SYLLABLE-C949
        0xC94A, # G+15167  U+C94A  uniC94A,  HANGUL SYLLABLE-C94A
        0xC94B, # G+15168  U+C94B  uniC94B,  HANGUL SYLLABLE-C94B
        0xC94C, # G+15169  U+C94C  uniC94C,  HANGUL SYLLABLE-C94C
        0xC94D, # G+15170  U+C94D  uniC94D,  HANGUL SYLLABLE-C94D
        0xC94E, # G+15171  U+C94E  uniC94E,  HANGUL SYLLABLE-C94E
        0xC94F, # G+15172  U+C94F  uniC94F,  HANGUL SYLLABLE-C94F
        0xC952, # G+15173  U+C952  uniC952,  HANGUL SYLLABLE-C952
        0xC953, # G+15174  U+C953  uniC953,  HANGUL SYLLABLE-C953
        0xC955, # G+15175  U+C955  uniC955,  HANGUL SYLLABLE-C955
        0xC956, # G+15176  U+C956  uniC956,  HANGUL SYLLABLE-C956
        0xC957, # G+15177  U+C957  uniC957,  HANGUL SYLLABLE-C957
        0xC959, # G+15178  U+C959  uniC959,  HANGUL SYLLABLE-C959
        0xC95A, # G+15179  U+C95A  uniC95A,  HANGUL SYLLABLE-C95A
        0xC95B, # G+15180  U+C95B  uniC95B,  HANGUL SYLLABLE-C95B
        0xC95C, # G+15181  U+C95C  uniC95C,  HANGUL SYLLABLE-C95C
        0xC95D, # G+15182  U+C95D  uniC95D,  HANGUL SYLLABLE-C95D
        0xC95E, # G+15183  U+C95E  uniC95E,  HANGUL SYLLABLE-C95E
        0xC95F, # G+15184  U+C95F  uniC95F,  HANGUL SYLLABLE-C95F
        0xC962, # G+15185  U+C962  uniC962,  HANGUL SYLLABLE-C962
        0xC964, # G+15186  U+C964  uniC964,  HANGUL SYLLABLE-C964
        0xC965, # G+15187  U+C965  uniC965,  HANGUL SYLLABLE-C965
        0xC966, # G+15188  U+C966  uniC966,  HANGUL SYLLABLE-C966
        0xC967, # G+15189  U+C967  uniC967,  HANGUL SYLLABLE-C967
        0xC968, # G+15190  U+C968  uniC968,  HANGUL SYLLABLE-C968
        0xC969, # G+15191  U+C969  uniC969,  HANGUL SYLLABLE-C969
        0xC96A, # G+15192  U+C96A  uniC96A,  HANGUL SYLLABLE-C96A
        0xC96B, # G+15193  U+C96B  uniC96B,  HANGUL SYLLABLE-C96B
        0xC96D, # G+15194  U+C96D  uniC96D,  HANGUL SYLLABLE-C96D
        0xC96E, # G+15195  U+C96E  uniC96E,  HANGUL SYLLABLE-C96E
        0xC96F, # G+15196  U+C96F  uniC96F,  HANGUL SYLLABLE-C96F
        0xC971, # G+15197  U+C971  uniC971,  HANGUL SYLLABLE-C971
        0xC972, # G+15198  U+C972  uniC972,  HANGUL SYLLABLE-C972
        0xC973, # G+15199  U+C973  uniC973,  HANGUL SYLLABLE-C973
        0xC975, # G+15200  U+C975  uniC975,  HANGUL SYLLABLE-C975
        0xC976, # G+15201  U+C976  uniC976,  HANGUL SYLLABLE-C976
        0xC977, # G+15202  U+C977  uniC977,  HANGUL SYLLABLE-C977
        0xC978, # G+15203  U+C978  uniC978,  HANGUL SYLLABLE-C978
        0xC979, # G+15204  U+C979  uniC979,  HANGUL SYLLABLE-C979
        0xC97A, # G+15205  U+C97A  uniC97A,  HANGUL SYLLABLE-C97A
        0xC97B, # G+15206  U+C97B  uniC97B,  HANGUL SYLLABLE-C97B
        0xC97D, # G+15207  U+C97D  uniC97D,  HANGUL SYLLABLE-C97D
        0xC97E, # G+15208  U+C97E  uniC97E,  HANGUL SYLLABLE-C97E
        0xC97F, # G+15209  U+C97F  uniC97F,  HANGUL SYLLABLE-C97F
        0xC980, # G+15210  U+C980  uniC980,  HANGUL SYLLABLE-C980
        0xC981, # G+15211  U+C981  uniC981,  HANGUL SYLLABLE-C981
        0xC982, # G+15212  U+C982  uniC982,  HANGUL SYLLABLE-C982
        0xC983, # G+15213  U+C983  uniC983,  HANGUL SYLLABLE-C983
        0xC984, # G+15214  U+C984  uniC984,  HANGUL SYLLABLE-C984
        0xC985, # G+15215  U+C985  uniC985,  HANGUL SYLLABLE-C985
        0xC986, # G+15216  U+C986  uniC986,  HANGUL SYLLABLE-C986
        0xC987, # G+15217  U+C987  uniC987,  HANGUL SYLLABLE-C987
        0xC98A, # G+15218  U+C98A  uniC98A,  HANGUL SYLLABLE-C98A
        0xC98B, # G+15219  U+C98B  uniC98B,  HANGUL SYLLABLE-C98B
        0xC98D, # G+15220  U+C98D  uniC98D,  HANGUL SYLLABLE-C98D
        0xC98E, # G+15221  U+C98E  uniC98E,  HANGUL SYLLABLE-C98E
        0xC98F, # G+15222  U+C98F  uniC98F,  HANGUL SYLLABLE-C98F
        0xC991, # G+15223  U+C991  uniC991,  HANGUL SYLLABLE-C991
        0xC992, # G+15224  U+C992  uniC992,  HANGUL SYLLABLE-C992
        0xC993, # G+15225  U+C993  uniC993,  HANGUL SYLLABLE-C993
        0xC994, # G+15226  U+C994  uniC994,  HANGUL SYLLABLE-C994
        0xC995, # G+15227  U+C995  uniC995,  HANGUL SYLLABLE-C995
        0xC996, # G+15228  U+C996  uniC996,  HANGUL SYLLABLE-C996
        0xC997, # G+15229  U+C997  uniC997,  HANGUL SYLLABLE-C997
        0xC99A, # G+15230  U+C99A  uniC99A,  HANGUL SYLLABLE-C99A
        0xC99C, # G+15231  U+C99C  uniC99C,  HANGUL SYLLABLE-C99C
        0xC99E, # G+15232  U+C99E  uniC99E,  HANGUL SYLLABLE-C99E
        0xC99F, # G+15233  U+C99F  uniC99F,  HANGUL SYLLABLE-C99F
        0xC9A0, # G+15234  U+C9A0  uniC9A0,  HANGUL SYLLABLE-C9A0
        0xC9A1, # G+15235  U+C9A1  uniC9A1,  HANGUL SYLLABLE-C9A1
        0xC9A2, # G+15236  U+C9A2  uniC9A2,  HANGUL SYLLABLE-C9A2
        0xC9A3, # G+15237  U+C9A3  uniC9A3,  HANGUL SYLLABLE-C9A3
        0xC9A4, # G+15238  U+C9A4  uniC9A4,  HANGUL SYLLABLE-C9A4
        0xC9A5, # G+15239  U+C9A5  uniC9A5,  HANGUL SYLLABLE-C9A5
        0xC9A6, # G+15240  U+C9A6  uniC9A6,  HANGUL SYLLABLE-C9A6
        0xC9A7, # G+15241  U+C9A7  uniC9A7,  HANGUL SYLLABLE-C9A7
        0xC9A8, # G+15242  U+C9A8  uniC9A8,  HANGUL SYLLABLE-C9A8
        0xC9A9, # G+15243  U+C9A9  uniC9A9,  HANGUL SYLLABLE-C9A9
        0xC9AA, # G+15244  U+C9AA  uniC9AA,  HANGUL SYLLABLE-C9AA
        0xC9AB, # G+15245  U+C9AB  uniC9AB,  HANGUL SYLLABLE-C9AB
        0xC9AC, # G+15246  U+C9AC  uniC9AC,  HANGUL SYLLABLE-C9AC
        0xC9AD, # G+15247  U+C9AD  uniC9AD,  HANGUL SYLLABLE-C9AD
        0xC9AE, # G+15248  U+C9AE  uniC9AE,  HANGUL SYLLABLE-C9AE
        0xC9AF, # G+15249  U+C9AF  uniC9AF,  HANGUL SYLLABLE-C9AF
        0xC9B0, # G+15250  U+C9B0  uniC9B0,  HANGUL SYLLABLE-C9B0
        0xC9B1, # G+15251  U+C9B1  uniC9B1,  HANGUL SYLLABLE-C9B1
        0xC9B2, # G+15252  U+C9B2  uniC9B2,  HANGUL SYLLABLE-C9B2
        0xC9B3, # G+15253  U+C9B3  uniC9B3,  HANGUL SYLLABLE-C9B3
        0xC9B4, # G+15254  U+C9B4  uniC9B4,  HANGUL SYLLABLE-C9B4
        0xC9B5, # G+15255  U+C9B5  uniC9B5,  HANGUL SYLLABLE-C9B5
        0xC9B6, # G+15256  U+C9B6  uniC9B6,  HANGUL SYLLABLE-C9B6
        0xC9B7, # G+15257  U+C9B7  uniC9B7,  HANGUL SYLLABLE-C9B7
        0xC9B8, # G+15258  U+C9B8  uniC9B8,  HANGUL SYLLABLE-C9B8
        0xC9B9, # G+15259  U+C9B9  uniC9B9,  HANGUL SYLLABLE-C9B9
        0xC9BA, # G+15260  U+C9BA  uniC9BA,  HANGUL SYLLABLE-C9BA
        0xC9BB, # G+15261  U+C9BB  uniC9BB,  HANGUL SYLLABLE-C9BB
        0xC9BC, # G+15262  U+C9BC  uniC9BC,  HANGUL SYLLABLE-C9BC
        0xC9BD, # G+15263  U+C9BD  uniC9BD,  HANGUL SYLLABLE-C9BD
        0xC9BE, # G+15264  U+C9BE  uniC9BE,  HANGUL SYLLABLE-C9BE
        0xC9BF, # G+15265  U+C9BF  uniC9BF,  HANGUL SYLLABLE-C9BF
        0xC9C2, # G+15266  U+C9C2  uniC9C2,  HANGUL SYLLABLE-C9C2
        0xC9C3, # G+15267  U+C9C3  uniC9C3,  HANGUL SYLLABLE-C9C3
        0xC9C5, # G+15268  U+C9C5  uniC9C5,  HANGUL SYLLABLE-C9C5
        0xC9C6, # G+15269  U+C9C6  uniC9C6,  HANGUL SYLLABLE-C9C6
        0xC9C9, # G+15270  U+C9C9  uniC9C9,  HANGUL SYLLABLE-C9C9
        0xC9CB, # G+15271  U+C9CB  uniC9CB,  HANGUL SYLLABLE-C9CB
        0xC9CC, # G+15272  U+C9CC  uniC9CC,  HANGUL SYLLABLE-C9CC
        0xC9CD, # G+15273  U+C9CD  uniC9CD,  HANGUL SYLLABLE-C9CD
        0xC9CE, # G+15274  U+C9CE  uniC9CE,  HANGUL SYLLABLE-C9CE
        0xC9CF, # G+15275  U+C9CF  uniC9CF,  HANGUL SYLLABLE-C9CF
        0xC9D2, # G+15276  U+C9D2  uniC9D2,  HANGUL SYLLABLE-C9D2
        0xC9D4, # G+15277  U+C9D4  uniC9D4,  HANGUL SYLLABLE-C9D4
        0xC9D7, # G+15278  U+C9D7  uniC9D7,  HANGUL SYLLABLE-C9D7
        0xC9D8, # G+15279  U+C9D8  uniC9D8,  HANGUL SYLLABLE-C9D8
        0xC9DB, # G+15280  U+C9DB  uniC9DB,  HANGUL SYLLABLE-C9DB
        0xC9DE, # G+15281  U+C9DE  uniC9DE,  HANGUL SYLLABLE-C9DE
        0xC9DF, # G+15282  U+C9DF  uniC9DF,  HANGUL SYLLABLE-C9DF
        0xC9E1, # G+15283  U+C9E1  uniC9E1,  HANGUL SYLLABLE-C9E1
        0xC9E3, # G+15284  U+C9E3  uniC9E3,  HANGUL SYLLABLE-C9E3
        0xC9E5, # G+15285  U+C9E5  uniC9E5,  HANGUL SYLLABLE-C9E5
        0xC9E6, # G+15286  U+C9E6  uniC9E6,  HANGUL SYLLABLE-C9E6
        0xC9E8, # G+15287  U+C9E8  uniC9E8,  HANGUL SYLLABLE-C9E8
        0xC9E9, # G+15288  U+C9E9  uniC9E9,  HANGUL SYLLABLE-C9E9
        0xC9EA, # G+15289  U+C9EA  uniC9EA,  HANGUL SYLLABLE-C9EA
        0xC9EB, # G+15290  U+C9EB  uniC9EB,  HANGUL SYLLABLE-C9EB
        0xC9EE, # G+15291  U+C9EE  uniC9EE,  HANGUL SYLLABLE-C9EE
        0xC9F2, # G+15292  U+C9F2  uniC9F2,  HANGUL SYLLABLE-C9F2
        0xC9F3, # G+15293  U+C9F3  uniC9F3,  HANGUL SYLLABLE-C9F3
        0xC9F4, # G+15294  U+C9F4  uniC9F4,  HANGUL SYLLABLE-C9F4
        0xC9F5, # G+15295  U+C9F5  uniC9F5,  HANGUL SYLLABLE-C9F5
        0xC9F6, # G+15296  U+C9F6  uniC9F6,  HANGUL SYLLABLE-C9F6
        0xC9F7, # G+15297  U+C9F7  uniC9F7,  HANGUL SYLLABLE-C9F7
        0xC9FA, # G+15298  U+C9FA  uniC9FA,  HANGUL SYLLABLE-C9FA
        0xC9FB, # G+15299  U+C9FB  uniC9FB,  HANGUL SYLLABLE-C9FB
        0xC9FD, # G+15300  U+C9FD  uniC9FD,  HANGUL SYLLABLE-C9FD
        0xC9FE, # G+15301  U+C9FE  uniC9FE,  HANGUL SYLLABLE-C9FE
        0xC9FF, # G+15302  U+C9FF  uniC9FF,  HANGUL SYLLABLE-C9FF
        0xCA01, # G+15303  U+CA01  uniCA01,  HANGUL SYLLABLE-CA01
        0xCA02, # G+15304  U+CA02  uniCA02,  HANGUL SYLLABLE-CA02
        0xCA03, # G+15305  U+CA03  uniCA03,  HANGUL SYLLABLE-CA03
        0xCA04, # G+15306  U+CA04  uniCA04,  HANGUL SYLLABLE-CA04
        0xCA05, # G+15307  U+CA05  uniCA05,  HANGUL SYLLABLE-CA05
        0xCA06, # G+15308  U+CA06  uniCA06,  HANGUL SYLLABLE-CA06
        0xCA07, # G+15309  U+CA07  uniCA07,  HANGUL SYLLABLE-CA07
        0xCA0A, # G+15310  U+CA0A  uniCA0A,  HANGUL SYLLABLE-CA0A
        0xCA0E, # G+15311  U+CA0E  uniCA0E,  HANGUL SYLLABLE-CA0E
        0xCA0F, # G+15312  U+CA0F  uniCA0F,  HANGUL SYLLABLE-CA0F
        0xCA10, # G+15313  U+CA10  uniCA10,  HANGUL SYLLABLE-CA10
        0xCA11, # G+15314  U+CA11  uniCA11,  HANGUL SYLLABLE-CA11
        0xCA12, # G+15315  U+CA12  uniCA12,  HANGUL SYLLABLE-CA12
        0xCA13, # G+15316  U+CA13  uniCA13,  HANGUL SYLLABLE-CA13
        0xCA15, # G+15317  U+CA15  uniCA15,  HANGUL SYLLABLE-CA15
        0xCA16, # G+15318  U+CA16  uniCA16,  HANGUL SYLLABLE-CA16
        0xCA17, # G+15319  U+CA17  uniCA17,  HANGUL SYLLABLE-CA17
        0xCA19, # G+15320  U+CA19  uniCA19,  HANGUL SYLLABLE-CA19
        0xCA1A, # G+15321  U+CA1A  uniCA1A,  HANGUL SYLLABLE-CA1A
        0xCA1B, # G+15322  U+CA1B  uniCA1B,  HANGUL SYLLABLE-CA1B
        0xCA1C, # G+15323  U+CA1C  uniCA1C,  HANGUL SYLLABLE-CA1C
        0xCA1D, # G+15324  U+CA1D  uniCA1D,  HANGUL SYLLABLE-CA1D
        0xCA1E, # G+15325  U+CA1E  uniCA1E,  HANGUL SYLLABLE-CA1E
        0xCA1F, # G+15326  U+CA1F  uniCA1F,  HANGUL SYLLABLE-CA1F
        0xCA20, # G+15327  U+CA20  uniCA20,  HANGUL SYLLABLE-CA20
        0xCA21, # G+15328  U+CA21  uniCA21,  HANGUL SYLLABLE-CA21
        0xCA22, # G+15329  U+CA22  uniCA22,  HANGUL SYLLABLE-CA22
        0xCA23, # G+15330  U+CA23  uniCA23,  HANGUL SYLLABLE-CA23
        0xCA24, # G+15331  U+CA24  uniCA24,  HANGUL SYLLABLE-CA24
        0xCA25, # G+15332  U+CA25  uniCA25,  HANGUL SYLLABLE-CA25
        0xCA26, # G+15333  U+CA26  uniCA26,  HANGUL SYLLABLE-CA26
        0xCA27, # G+15334  U+CA27  uniCA27,  HANGUL SYLLABLE-CA27
        0xCA28, # G+15335  U+CA28  uniCA28,  HANGUL SYLLABLE-CA28
        0xCA2A, # G+15336  U+CA2A  uniCA2A,  HANGUL SYLLABLE-CA2A
        0xCA2B, # G+15337  U+CA2B  uniCA2B,  HANGUL SYLLABLE-CA2B
        0xCA2C, # G+15338  U+CA2C  uniCA2C,  HANGUL SYLLABLE-CA2C
        0xCA2D, # G+15339  U+CA2D  uniCA2D,  HANGUL SYLLABLE-CA2D
        0xCA2E, # G+15340  U+CA2E  uniCA2E,  HANGUL SYLLABLE-CA2E
        0xCA2F, # G+15341  U+CA2F  uniCA2F,  HANGUL SYLLABLE-CA2F
        0xCA30, # G+15342  U+CA30  uniCA30,  HANGUL SYLLABLE-CA30
        0xCA31, # G+15343  U+CA31  uniCA31,  HANGUL SYLLABLE-CA31
        0xCA32, # G+15344  U+CA32  uniCA32,  HANGUL SYLLABLE-CA32
        0xCA33, # G+15345  U+CA33  uniCA33,  HANGUL SYLLABLE-CA33
        0xCA34, # G+15346  U+CA34  uniCA34,  HANGUL SYLLABLE-CA34
        0xCA35, # G+15347  U+CA35  uniCA35,  HANGUL SYLLABLE-CA35
        0xCA36, # G+15348  U+CA36  uniCA36,  HANGUL SYLLABLE-CA36
        0xCA37, # G+15349  U+CA37  uniCA37,  HANGUL SYLLABLE-CA37
        0xCA38, # G+15350  U+CA38  uniCA38,  HANGUL SYLLABLE-CA38
        0xCA39, # G+15351  U+CA39  uniCA39,  HANGUL SYLLABLE-CA39
        0xCA3A, # G+15352  U+CA3A  uniCA3A,  HANGUL SYLLABLE-CA3A
        0xCA3B, # G+15353  U+CA3B  uniCA3B,  HANGUL SYLLABLE-CA3B
        0xCA3C, # G+15354  U+CA3C  uniCA3C,  HANGUL SYLLABLE-CA3C
        0xCA3D, # G+15355  U+CA3D  uniCA3D,  HANGUL SYLLABLE-CA3D
        0xCA3E, # G+15356  U+CA3E  uniCA3E,  HANGUL SYLLABLE-CA3E
        0xCA3F, # G+15357  U+CA3F  uniCA3F,  HANGUL SYLLABLE-CA3F
        0xCA40, # G+15358  U+CA40  uniCA40,  HANGUL SYLLABLE-CA40
        0xCA41, # G+15359  U+CA41  uniCA41,  HANGUL SYLLABLE-CA41
        0xCA42, # G+15360  U+CA42  uniCA42,  HANGUL SYLLABLE-CA42
        0xCA43, # G+15361  U+CA43  uniCA43,  HANGUL SYLLABLE-CA43
        0xCA44, # G+15362  U+CA44  uniCA44,  HANGUL SYLLABLE-CA44
        0xCA45, # G+15363  U+CA45  uniCA45,  HANGUL SYLLABLE-CA45
        0xCA46, # G+15364  U+CA46  uniCA46,  HANGUL SYLLABLE-CA46
        0xCA47, # G+15365  U+CA47  uniCA47,  HANGUL SYLLABLE-CA47
        0xCA48, # G+15366  U+CA48  uniCA48,  HANGUL SYLLABLE-CA48
        0xCA49, # G+15367  U+CA49  uniCA49,  HANGUL SYLLABLE-CA49
        0xCA4A, # G+15368  U+CA4A  uniCA4A,  HANGUL SYLLABLE-CA4A
        0xCA4B, # G+15369  U+CA4B  uniCA4B,  HANGUL SYLLABLE-CA4B
        0xCA4E, # G+15370  U+CA4E  uniCA4E,  HANGUL SYLLABLE-CA4E
        0xCA4F, # G+15371  U+CA4F  uniCA4F,  HANGUL SYLLABLE-CA4F
        0xCA51, # G+15372  U+CA51  uniCA51,  HANGUL SYLLABLE-CA51
        0xCA52, # G+15373  U+CA52  uniCA52,  HANGUL SYLLABLE-CA52
        0xCA53, # G+15374  U+CA53  uniCA53,  HANGUL SYLLABLE-CA53
        0xCA55, # G+15375  U+CA55  uniCA55,  HANGUL SYLLABLE-CA55
        0xCA56, # G+15376  U+CA56  uniCA56,  HANGUL SYLLABLE-CA56
        0xCA57, # G+15377  U+CA57  uniCA57,  HANGUL SYLLABLE-CA57
        0xCA58, # G+15378  U+CA58  uniCA58,  HANGUL SYLLABLE-CA58
        0xCA59, # G+15379  U+CA59  uniCA59,  HANGUL SYLLABLE-CA59
        0xCA5A, # G+15380  U+CA5A  uniCA5A,  HANGUL SYLLABLE-CA5A
        0xCA5B, # G+15381  U+CA5B  uniCA5B,  HANGUL SYLLABLE-CA5B
        0xCA5E, # G+15382  U+CA5E  uniCA5E,  HANGUL SYLLABLE-CA5E
        0xCA62, # G+15383  U+CA62  uniCA62,  HANGUL SYLLABLE-CA62
        0xCA63, # G+15384  U+CA63  uniCA63,  HANGUL SYLLABLE-CA63
        0xCA64, # G+15385  U+CA64  uniCA64,  HANGUL SYLLABLE-CA64
        0xCA65, # G+15386  U+CA65  uniCA65,  HANGUL SYLLABLE-CA65
        0xCA66, # G+15387  U+CA66  uniCA66,  HANGUL SYLLABLE-CA66
        0xCA67, # G+15388  U+CA67  uniCA67,  HANGUL SYLLABLE-CA67
        0xCA69, # G+15389  U+CA69  uniCA69,  HANGUL SYLLABLE-CA69
        0xCA6A, # G+15390  U+CA6A  uniCA6A,  HANGUL SYLLABLE-CA6A
        0xCA6B, # G+15391  U+CA6B  uniCA6B,  HANGUL SYLLABLE-CA6B
        0xCA6C, # G+15392  U+CA6C  uniCA6C,  HANGUL SYLLABLE-CA6C
        0xCA6D, # G+15393  U+CA6D  uniCA6D,  HANGUL SYLLABLE-CA6D
        0xCA6E, # G+15394  U+CA6E  uniCA6E,  HANGUL SYLLABLE-CA6E
        0xCA6F, # G+15395  U+CA6F  uniCA6F,  HANGUL SYLLABLE-CA6F
        0xCA70, # G+15396  U+CA70  uniCA70,  HANGUL SYLLABLE-CA70
        0xCA71, # G+15397  U+CA71  uniCA71,  HANGUL SYLLABLE-CA71
        0xCA72, # G+15398  U+CA72  uniCA72,  HANGUL SYLLABLE-CA72
        0xCA73, # G+15399  U+CA73  uniCA73,  HANGUL SYLLABLE-CA73
        0xCA74, # G+15400  U+CA74  uniCA74,  HANGUL SYLLABLE-CA74
        0xCA75, # G+15401  U+CA75  uniCA75,  HANGUL SYLLABLE-CA75
        0xCA76, # G+15402  U+CA76  uniCA76,  HANGUL SYLLABLE-CA76
        0xCA77, # G+15403  U+CA77  uniCA77,  HANGUL SYLLABLE-CA77
        0xCA78, # G+15404  U+CA78  uniCA78,  HANGUL SYLLABLE-CA78
        0xCA79, # G+15405  U+CA79  uniCA79,  HANGUL SYLLABLE-CA79
        0xCA7A, # G+15406  U+CA7A  uniCA7A,  HANGUL SYLLABLE-CA7A
        0xCA7B, # G+15407  U+CA7B  uniCA7B,  HANGUL SYLLABLE-CA7B
        0xCA7C, # G+15408  U+CA7C  uniCA7C,  HANGUL SYLLABLE-CA7C
        0xCA7E, # G+15409  U+CA7E  uniCA7E,  HANGUL SYLLABLE-CA7E
        0xCA7F, # G+15410  U+CA7F  uniCA7F,  HANGUL SYLLABLE-CA7F
        0xCA80, # G+15411  U+CA80  uniCA80,  HANGUL SYLLABLE-CA80
        0xCA81, # G+15412  U+CA81  uniCA81,  HANGUL SYLLABLE-CA81
        0xCA82, # G+15413  U+CA82  uniCA82,  HANGUL SYLLABLE-CA82
        0xCA83, # G+15414  U+CA83  uniCA83,  HANGUL SYLLABLE-CA83
        0xCA85, # G+15415  U+CA85  uniCA85,  HANGUL SYLLABLE-CA85
        0xCA86, # G+15416  U+CA86  uniCA86,  HANGUL SYLLABLE-CA86
        0xCA87, # G+15417  U+CA87  uniCA87,  HANGUL SYLLABLE-CA87
        0xCA88, # G+15418  U+CA88  uniCA88,  HANGUL SYLLABLE-CA88
        0xCA89, # G+15419  U+CA89  uniCA89,  HANGUL SYLLABLE-CA89
        0xCA8A, # G+15420  U+CA8A  uniCA8A,  HANGUL SYLLABLE-CA8A
        0xCA8B, # G+15421  U+CA8B  uniCA8B,  HANGUL SYLLABLE-CA8B
        0xCA8C, # G+15422  U+CA8C  uniCA8C,  HANGUL SYLLABLE-CA8C
        0xCA8D, # G+15423  U+CA8D  uniCA8D,  HANGUL SYLLABLE-CA8D
        0xCA8E, # G+15424  U+CA8E  uniCA8E,  HANGUL SYLLABLE-CA8E
        0xCA8F, # G+15425  U+CA8F  uniCA8F,  HANGUL SYLLABLE-CA8F
        0xCA90, # G+15426  U+CA90  uniCA90,  HANGUL SYLLABLE-CA90
        0xCA91, # G+15427  U+CA91  uniCA91,  HANGUL SYLLABLE-CA91
        0xCA92, # G+15428  U+CA92  uniCA92,  HANGUL SYLLABLE-CA92
        0xCA93, # G+15429  U+CA93  uniCA93,  HANGUL SYLLABLE-CA93
        0xCA94, # G+15430  U+CA94  uniCA94,  HANGUL SYLLABLE-CA94
        0xCA95, # G+15431  U+CA95  uniCA95,  HANGUL SYLLABLE-CA95
        0xCA96, # G+15432  U+CA96  uniCA96,  HANGUL SYLLABLE-CA96
        0xCA97, # G+15433  U+CA97  uniCA97,  HANGUL SYLLABLE-CA97
        0xCA99, # G+15434  U+CA99  uniCA99,  HANGUL SYLLABLE-CA99
        0xCA9A, # G+15435  U+CA9A  uniCA9A,  HANGUL SYLLABLE-CA9A
        0xCA9B, # G+15436  U+CA9B  uniCA9B,  HANGUL SYLLABLE-CA9B
        0xCA9C, # G+15437  U+CA9C  uniCA9C,  HANGUL SYLLABLE-CA9C
        0xCA9D, # G+15438  U+CA9D  uniCA9D,  HANGUL SYLLABLE-CA9D
        0xCA9E, # G+15439  U+CA9E  uniCA9E,  HANGUL SYLLABLE-CA9E
        0xCA9F, # G+15440  U+CA9F  uniCA9F,  HANGUL SYLLABLE-CA9F
        0xCAA0, # G+15441  U+CAA0  uniCAA0,  HANGUL SYLLABLE-CAA0
        0xCAA1, # G+15442  U+CAA1  uniCAA1,  HANGUL SYLLABLE-CAA1
        0xCAA2, # G+15443  U+CAA2  uniCAA2,  HANGUL SYLLABLE-CAA2
        0xCAA3, # G+15444  U+CAA3  uniCAA3,  HANGUL SYLLABLE-CAA3
        0xCAA4, # G+15445  U+CAA4  uniCAA4,  HANGUL SYLLABLE-CAA4
        0xCAA5, # G+15446  U+CAA5  uniCAA5,  HANGUL SYLLABLE-CAA5
        0xCAA6, # G+15447  U+CAA6  uniCAA6,  HANGUL SYLLABLE-CAA6
        0xCAA7, # G+15448  U+CAA7  uniCAA7,  HANGUL SYLLABLE-CAA7
        0xCAA8, # G+15449  U+CAA8  uniCAA8,  HANGUL SYLLABLE-CAA8
        0xCAA9, # G+15450  U+CAA9  uniCAA9,  HANGUL SYLLABLE-CAA9
        0xCAAA, # G+15451  U+CAAA  uniCAAA,  HANGUL SYLLABLE-CAAA
        0xCAAB, # G+15452  U+CAAB  uniCAAB,  HANGUL SYLLABLE-CAAB
        0xCAAC, # G+15453  U+CAAC  uniCAAC,  HANGUL SYLLABLE-CAAC
        0xCAAD, # G+15454  U+CAAD  uniCAAD,  HANGUL SYLLABLE-CAAD
        0xCAAE, # G+15455  U+CAAE  uniCAAE,  HANGUL SYLLABLE-CAAE
        0xCAAF, # G+15456  U+CAAF  uniCAAF,  HANGUL SYLLABLE-CAAF
        0xCAB0, # G+15457  U+CAB0  uniCAB0,  HANGUL SYLLABLE-CAB0
        0xCAB1, # G+15458  U+CAB1  uniCAB1,  HANGUL SYLLABLE-CAB1
        0xCAB2, # G+15459  U+CAB2  uniCAB2,  HANGUL SYLLABLE-CAB2
        0xCAB3, # G+15460  U+CAB3  uniCAB3,  HANGUL SYLLABLE-CAB3
        0xCAB4, # G+15461  U+CAB4  uniCAB4,  HANGUL SYLLABLE-CAB4
        0xCAB5, # G+15462  U+CAB5  uniCAB5,  HANGUL SYLLABLE-CAB5
        0xCAB6, # G+15463  U+CAB6  uniCAB6,  HANGUL SYLLABLE-CAB6
        0xCAB7, # G+15464  U+CAB7  uniCAB7,  HANGUL SYLLABLE-CAB7
        0xCAB8, # G+15465  U+CAB8  uniCAB8,  HANGUL SYLLABLE-CAB8
        0xCAB9, # G+15466  U+CAB9  uniCAB9,  HANGUL SYLLABLE-CAB9
        0xCABA, # G+15467  U+CABA  uniCABA,  HANGUL SYLLABLE-CABA
        0xCABB, # G+15468  U+CABB  uniCABB,  HANGUL SYLLABLE-CABB
        0xCABE, # G+15469  U+CABE  uniCABE,  HANGUL SYLLABLE-CABE
        0xCABF, # G+15470  U+CABF  uniCABF,  HANGUL SYLLABLE-CABF
        0xCAC1, # G+15471  U+CAC1  uniCAC1,  HANGUL SYLLABLE-CAC1
        0xCAC2, # G+15472  U+CAC2  uniCAC2,  HANGUL SYLLABLE-CAC2
        0xCAC3, # G+15473  U+CAC3  uniCAC3,  HANGUL SYLLABLE-CAC3
        0xCAC5, # G+15474  U+CAC5  uniCAC5,  HANGUL SYLLABLE-CAC5
        0xCAC6, # G+15475  U+CAC6  uniCAC6,  HANGUL SYLLABLE-CAC6
        0xCAC7, # G+15476  U+CAC7  uniCAC7,  HANGUL SYLLABLE-CAC7
        0xCAC8, # G+15477  U+CAC8  uniCAC8,  HANGUL SYLLABLE-CAC8
        0xCAC9, # G+15478  U+CAC9  uniCAC9,  HANGUL SYLLABLE-CAC9
        0xCACA, # G+15479  U+CACA  uniCACA,  HANGUL SYLLABLE-CACA
        0xCACB, # G+15480  U+CACB  uniCACB,  HANGUL SYLLABLE-CACB
        0xCACE, # G+15481  U+CACE  uniCACE,  HANGUL SYLLABLE-CACE
        0xCAD0, # G+15482  U+CAD0  uniCAD0,  HANGUL SYLLABLE-CAD0
        0xCAD2, # G+15483  U+CAD2  uniCAD2,  HANGUL SYLLABLE-CAD2
        0xCAD4, # G+15484  U+CAD4  uniCAD4,  HANGUL SYLLABLE-CAD4
        0xCAD5, # G+15485  U+CAD5  uniCAD5,  HANGUL SYLLABLE-CAD5
        0xCAD6, # G+15486  U+CAD6  uniCAD6,  HANGUL SYLLABLE-CAD6
        0xCAD7, # G+15487  U+CAD7  uniCAD7,  HANGUL SYLLABLE-CAD7
        0xCADA, # G+15488  U+CADA  uniCADA,  HANGUL SYLLABLE-CADA
        0xCADB, # G+15489  U+CADB  uniCADB,  HANGUL SYLLABLE-CADB
        0xCADC, # G+15490  U+CADC  uniCADC,  HANGUL SYLLABLE-CADC
        0xCADD, # G+15491  U+CADD  uniCADD,  HANGUL SYLLABLE-CADD
        0xCADE, # G+15492  U+CADE  uniCADE,  HANGUL SYLLABLE-CADE
        0xCADF, # G+15493  U+CADF  uniCADF,  HANGUL SYLLABLE-CADF
        0xCAE1, # G+15494  U+CAE1  uniCAE1,  HANGUL SYLLABLE-CAE1
        0xCAE2, # G+15495  U+CAE2  uniCAE2,  HANGUL SYLLABLE-CAE2
        0xCAE3, # G+15496  U+CAE3  uniCAE3,  HANGUL SYLLABLE-CAE3
        0xCAE4, # G+15497  U+CAE4  uniCAE4,  HANGUL SYLLABLE-CAE4
        0xCAE5, # G+15498  U+CAE5  uniCAE5,  HANGUL SYLLABLE-CAE5
        0xCAE6, # G+15499  U+CAE6  uniCAE6,  HANGUL SYLLABLE-CAE6
        0xCAE7, # G+15500  U+CAE7  uniCAE7,  HANGUL SYLLABLE-CAE7
        0xCAE8, # G+15501  U+CAE8  uniCAE8,  HANGUL SYLLABLE-CAE8
        0xCAE9, # G+15502  U+CAE9  uniCAE9,  HANGUL SYLLABLE-CAE9
        0xCAEA, # G+15503  U+CAEA  uniCAEA,  HANGUL SYLLABLE-CAEA
        0xCAEB, # G+15504  U+CAEB  uniCAEB,  HANGUL SYLLABLE-CAEB
        0xCAED, # G+15505  U+CAED  uniCAED,  HANGUL SYLLABLE-CAED
        0xCAEE, # G+15506  U+CAEE  uniCAEE,  HANGUL SYLLABLE-CAEE
        0xCAEF, # G+15507  U+CAEF  uniCAEF,  HANGUL SYLLABLE-CAEF
        0xCAF0, # G+15508  U+CAF0  uniCAF0,  HANGUL SYLLABLE-CAF0
        0xCAF1, # G+15509  U+CAF1  uniCAF1,  HANGUL SYLLABLE-CAF1
        0xCAF2, # G+15510  U+CAF2  uniCAF2,  HANGUL SYLLABLE-CAF2
        0xCAF3, # G+15511  U+CAF3  uniCAF3,  HANGUL SYLLABLE-CAF3
        0xCAF5, # G+15512  U+CAF5  uniCAF5,  HANGUL SYLLABLE-CAF5
        0xCAF6, # G+15513  U+CAF6  uniCAF6,  HANGUL SYLLABLE-CAF6
        0xCAF7, # G+15514  U+CAF7  uniCAF7,  HANGUL SYLLABLE-CAF7
        0xCAF8, # G+15515  U+CAF8  uniCAF8,  HANGUL SYLLABLE-CAF8
        0xCAF9, # G+15516  U+CAF9  uniCAF9,  HANGUL SYLLABLE-CAF9
        0xCAFA, # G+15517  U+CAFA  uniCAFA,  HANGUL SYLLABLE-CAFA
        0xCAFB, # G+15518  U+CAFB  uniCAFB,  HANGUL SYLLABLE-CAFB
        0xCAFC, # G+15519  U+CAFC  uniCAFC,  HANGUL SYLLABLE-CAFC
        0xCAFD, # G+15520  U+CAFD  uniCAFD,  HANGUL SYLLABLE-CAFD
        0xCAFE, # G+15521  U+CAFE  uniCAFE,  HANGUL SYLLABLE-CAFE
        0xCAFF, # G+15522  U+CAFF  uniCAFF,  HANGUL SYLLABLE-CAFF
        0xCB00, # G+15523  U+CB00  uniCB00,  HANGUL SYLLABLE-CB00
        0xCB01, # G+15524  U+CB01  uniCB01,  HANGUL SYLLABLE-CB01
        0xCB02, # G+15525  U+CB02  uniCB02,  HANGUL SYLLABLE-CB02
        0xCB03, # G+15526  U+CB03  uniCB03,  HANGUL SYLLABLE-CB03
        0xCB04, # G+15527  U+CB04  uniCB04,  HANGUL SYLLABLE-CB04
        0xCB05, # G+15528  U+CB05  uniCB05,  HANGUL SYLLABLE-CB05
        0xCB06, # G+15529  U+CB06  uniCB06,  HANGUL SYLLABLE-CB06
        0xCB07, # G+15530  U+CB07  uniCB07,  HANGUL SYLLABLE-CB07
        0xCB09, # G+15531  U+CB09  uniCB09,  HANGUL SYLLABLE-CB09
        0xCB0A, # G+15532  U+CB0A  uniCB0A,  HANGUL SYLLABLE-CB0A
        0xCB0B, # G+15533  U+CB0B  uniCB0B,  HANGUL SYLLABLE-CB0B
        0xCB0C, # G+15534  U+CB0C  uniCB0C,  HANGUL SYLLABLE-CB0C
        0xCB0D, # G+15535  U+CB0D  uniCB0D,  HANGUL SYLLABLE-CB0D
        0xCB0E, # G+15536  U+CB0E  uniCB0E,  HANGUL SYLLABLE-CB0E
        0xCB0F, # G+15537  U+CB0F  uniCB0F,  HANGUL SYLLABLE-CB0F
        0xCB11, # G+15538  U+CB11  uniCB11,  HANGUL SYLLABLE-CB11
        0xCB12, # G+15539  U+CB12  uniCB12,  HANGUL SYLLABLE-CB12
        0xCB13, # G+15540  U+CB13  uniCB13,  HANGUL SYLLABLE-CB13
        0xCB15, # G+15541  U+CB15  uniCB15,  HANGUL SYLLABLE-CB15
        0xCB16, # G+15542  U+CB16  uniCB16,  HANGUL SYLLABLE-CB16
        0xCB17, # G+15543  U+CB17  uniCB17,  HANGUL SYLLABLE-CB17
        0xCB19, # G+15544  U+CB19  uniCB19,  HANGUL SYLLABLE-CB19
        0xCB1A, # G+15545  U+CB1A  uniCB1A,  HANGUL SYLLABLE-CB1A
        0xCB1B, # G+15546  U+CB1B  uniCB1B,  HANGUL SYLLABLE-CB1B
        0xCB1C, # G+15547  U+CB1C  uniCB1C,  HANGUL SYLLABLE-CB1C
        0xCB1D, # G+15548  U+CB1D  uniCB1D,  HANGUL SYLLABLE-CB1D
        0xCB1E, # G+15549  U+CB1E  uniCB1E,  HANGUL SYLLABLE-CB1E
        0xCB1F, # G+15550  U+CB1F  uniCB1F,  HANGUL SYLLABLE-CB1F
        0xCB22, # G+15551  U+CB22  uniCB22,  HANGUL SYLLABLE-CB22
        0xCB23, # G+15552  U+CB23  uniCB23,  HANGUL SYLLABLE-CB23
        0xCB24, # G+15553  U+CB24  uniCB24,  HANGUL SYLLABLE-CB24
        0xCB25, # G+15554  U+CB25  uniCB25,  HANGUL SYLLABLE-CB25
        0xCB26, # G+15555  U+CB26  uniCB26,  HANGUL SYLLABLE-CB26
        0xCB27, # G+15556  U+CB27  uniCB27,  HANGUL SYLLABLE-CB27
        0xCB28, # G+15557  U+CB28  uniCB28,  HANGUL SYLLABLE-CB28
        0xCB29, # G+15558  U+CB29  uniCB29,  HANGUL SYLLABLE-CB29
        0xCB2A, # G+15559  U+CB2A  uniCB2A,  HANGUL SYLLABLE-CB2A
        0xCB2B, # G+15560  U+CB2B  uniCB2B,  HANGUL SYLLABLE-CB2B
        0xCB2C, # G+15561  U+CB2C  uniCB2C,  HANGUL SYLLABLE-CB2C
        0xCB2D, # G+15562  U+CB2D  uniCB2D,  HANGUL SYLLABLE-CB2D
        0xCB2E, # G+15563  U+CB2E  uniCB2E,  HANGUL SYLLABLE-CB2E
        0xCB2F, # G+15564  U+CB2F  uniCB2F,  HANGUL SYLLABLE-CB2F
        0xCB30, # G+15565  U+CB30  uniCB30,  HANGUL SYLLABLE-CB30
        0xCB31, # G+15566  U+CB31  uniCB31,  HANGUL SYLLABLE-CB31
        0xCB32, # G+15567  U+CB32  uniCB32,  HANGUL SYLLABLE-CB32
        0xCB33, # G+15568  U+CB33  uniCB33,  HANGUL SYLLABLE-CB33
        0xCB34, # G+15569  U+CB34  uniCB34,  HANGUL SYLLABLE-CB34
        0xCB35, # G+15570  U+CB35  uniCB35,  HANGUL SYLLABLE-CB35
        0xCB36, # G+15571  U+CB36  uniCB36,  HANGUL SYLLABLE-CB36
        0xCB37, # G+15572  U+CB37  uniCB37,  HANGUL SYLLABLE-CB37
        0xCB38, # G+15573  U+CB38  uniCB38,  HANGUL SYLLABLE-CB38
        0xCB39, # G+15574  U+CB39  uniCB39,  HANGUL SYLLABLE-CB39
        0xCB3A, # G+15575  U+CB3A  uniCB3A,  HANGUL SYLLABLE-CB3A
        0xCB3B, # G+15576  U+CB3B  uniCB3B,  HANGUL SYLLABLE-CB3B
        0xCB3C, # G+15577  U+CB3C  uniCB3C,  HANGUL SYLLABLE-CB3C
        0xCB3D, # G+15578  U+CB3D  uniCB3D,  HANGUL SYLLABLE-CB3D
        0xCB3E, # G+15579  U+CB3E  uniCB3E,  HANGUL SYLLABLE-CB3E
        0xCB3F, # G+15580  U+CB3F  uniCB3F,  HANGUL SYLLABLE-CB3F
        0xCB40, # G+15581  U+CB40  uniCB40,  HANGUL SYLLABLE-CB40
        0xCB42, # G+15582  U+CB42  uniCB42,  HANGUL SYLLABLE-CB42
        0xCB43, # G+15583  U+CB43  uniCB43,  HANGUL SYLLABLE-CB43
        0xCB44, # G+15584  U+CB44  uniCB44,  HANGUL SYLLABLE-CB44
        0xCB45, # G+15585  U+CB45  uniCB45,  HANGUL SYLLABLE-CB45
        0xCB46, # G+15586  U+CB46  uniCB46,  HANGUL SYLLABLE-CB46
        0xCB47, # G+15587  U+CB47  uniCB47,  HANGUL SYLLABLE-CB47
        0xCB4A, # G+15588  U+CB4A  uniCB4A,  HANGUL SYLLABLE-CB4A
        0xCB4B, # G+15589  U+CB4B  uniCB4B,  HANGUL SYLLABLE-CB4B
        0xCB4D, # G+15590  U+CB4D  uniCB4D,  HANGUL SYLLABLE-CB4D
        0xCB4E, # G+15591  U+CB4E  uniCB4E,  HANGUL SYLLABLE-CB4E
        0xCB4F, # G+15592  U+CB4F  uniCB4F,  HANGUL SYLLABLE-CB4F
        0xCB51, # G+15593  U+CB51  uniCB51,  HANGUL SYLLABLE-CB51
        0xCB52, # G+15594  U+CB52  uniCB52,  HANGUL SYLLABLE-CB52
        0xCB53, # G+15595  U+CB53  uniCB53,  HANGUL SYLLABLE-CB53
        0xCB54, # G+15596  U+CB54  uniCB54,  HANGUL SYLLABLE-CB54
        0xCB55, # G+15597  U+CB55  uniCB55,  HANGUL SYLLABLE-CB55
        0xCB56, # G+15598  U+CB56  uniCB56,  HANGUL SYLLABLE-CB56
        0xCB57, # G+15599  U+CB57  uniCB57,  HANGUL SYLLABLE-CB57
        0xCB5A, # G+15600  U+CB5A  uniCB5A,  HANGUL SYLLABLE-CB5A
        0xCB5B, # G+15601  U+CB5B  uniCB5B,  HANGUL SYLLABLE-CB5B
        0xCB5C, # G+15602  U+CB5C  uniCB5C,  HANGUL SYLLABLE-CB5C
        0xCB5E, # G+15603  U+CB5E  uniCB5E,  HANGUL SYLLABLE-CB5E
        0xCB5F, # G+15604  U+CB5F  uniCB5F,  HANGUL SYLLABLE-CB5F
        0xCB60, # G+15605  U+CB60  uniCB60,  HANGUL SYLLABLE-CB60
        0xCB61, # G+15606  U+CB61  uniCB61,  HANGUL SYLLABLE-CB61
        0xCB62, # G+15607  U+CB62  uniCB62,  HANGUL SYLLABLE-CB62
        0xCB63, # G+15608  U+CB63  uniCB63,  HANGUL SYLLABLE-CB63
        0xCB65, # G+15609  U+CB65  uniCB65,  HANGUL SYLLABLE-CB65
        0xCB66, # G+15610  U+CB66  uniCB66,  HANGUL SYLLABLE-CB66
        0xCB67, # G+15611  U+CB67  uniCB67,  HANGUL SYLLABLE-CB67
        0xCB68, # G+15612  U+CB68  uniCB68,  HANGUL SYLLABLE-CB68
        0xCB69, # G+15613  U+CB69  uniCB69,  HANGUL SYLLABLE-CB69
        0xCB6A, # G+15614  U+CB6A  uniCB6A,  HANGUL SYLLABLE-CB6A
        0xCB6B, # G+15615  U+CB6B  uniCB6B,  HANGUL SYLLABLE-CB6B
        0xCB6C, # G+15616  U+CB6C  uniCB6C,  HANGUL SYLLABLE-CB6C
        0xCB6D, # G+15617  U+CB6D  uniCB6D,  HANGUL SYLLABLE-CB6D
        0xCB6E, # G+15618  U+CB6E  uniCB6E,  HANGUL SYLLABLE-CB6E
        0xCB6F, # G+15619  U+CB6F  uniCB6F,  HANGUL SYLLABLE-CB6F
        0xCB70, # G+15620  U+CB70  uniCB70,  HANGUL SYLLABLE-CB70
        0xCB71, # G+15621  U+CB71  uniCB71,  HANGUL SYLLABLE-CB71
        0xCB72, # G+15622  U+CB72  uniCB72,  HANGUL SYLLABLE-CB72
        0xCB73, # G+15623  U+CB73  uniCB73,  HANGUL SYLLABLE-CB73
        0xCB74, # G+15624  U+CB74  uniCB74,  HANGUL SYLLABLE-CB74
        0xCB75, # G+15625  U+CB75  uniCB75,  HANGUL SYLLABLE-CB75
        0xCB76, # G+15626  U+CB76  uniCB76,  HANGUL SYLLABLE-CB76
        0xCB77, # G+15627  U+CB77  uniCB77,  HANGUL SYLLABLE-CB77
        0xCB7A, # G+15628  U+CB7A  uniCB7A,  HANGUL SYLLABLE-CB7A
        0xCB7B, # G+15629  U+CB7B  uniCB7B,  HANGUL SYLLABLE-CB7B
        0xCB7C, # G+15630  U+CB7C  uniCB7C,  HANGUL SYLLABLE-CB7C
        0xCB7D, # G+15631  U+CB7D  uniCB7D,  HANGUL SYLLABLE-CB7D
        0xCB7E, # G+15632  U+CB7E  uniCB7E,  HANGUL SYLLABLE-CB7E
        0xCB7F, # G+15633  U+CB7F  uniCB7F,  HANGUL SYLLABLE-CB7F
        0xCB80, # G+15634  U+CB80  uniCB80,  HANGUL SYLLABLE-CB80
        0xCB81, # G+15635  U+CB81  uniCB81,  HANGUL SYLLABLE-CB81
        0xCB82, # G+15636  U+CB82  uniCB82,  HANGUL SYLLABLE-CB82
        0xCB83, # G+15637  U+CB83  uniCB83,  HANGUL SYLLABLE-CB83
        0xCB84, # G+15638  U+CB84  uniCB84,  HANGUL SYLLABLE-CB84
        0xCB85, # G+15639  U+CB85  uniCB85,  HANGUL SYLLABLE-CB85
        0xCB86, # G+15640  U+CB86  uniCB86,  HANGUL SYLLABLE-CB86
        0xCB87, # G+15641  U+CB87  uniCB87,  HANGUL SYLLABLE-CB87
        0xCB88, # G+15642  U+CB88  uniCB88,  HANGUL SYLLABLE-CB88
        0xCB89, # G+15643  U+CB89  uniCB89,  HANGUL SYLLABLE-CB89
        0xCB8A, # G+15644  U+CB8A  uniCB8A,  HANGUL SYLLABLE-CB8A
        0xCB8B, # G+15645  U+CB8B  uniCB8B,  HANGUL SYLLABLE-CB8B
        0xCB8C, # G+15646  U+CB8C  uniCB8C,  HANGUL SYLLABLE-CB8C
        0xCB8D, # G+15647  U+CB8D  uniCB8D,  HANGUL SYLLABLE-CB8D
        0xCB8E, # G+15648  U+CB8E  uniCB8E,  HANGUL SYLLABLE-CB8E
        0xCB8F, # G+15649  U+CB8F  uniCB8F,  HANGUL SYLLABLE-CB8F
        0xCB90, # G+15650  U+CB90  uniCB90,  HANGUL SYLLABLE-CB90
        0xCB91, # G+15651  U+CB91  uniCB91,  HANGUL SYLLABLE-CB91
        0xCB92, # G+15652  U+CB92  uniCB92,  HANGUL SYLLABLE-CB92
        0xCB93, # G+15653  U+CB93  uniCB93,  HANGUL SYLLABLE-CB93
        0xCB94, # G+15654  U+CB94  uniCB94,  HANGUL SYLLABLE-CB94
        0xCB95, # G+15655  U+CB95  uniCB95,  HANGUL SYLLABLE-CB95
        0xCB96, # G+15656  U+CB96  uniCB96,  HANGUL SYLLABLE-CB96
        0xCB97, # G+15657  U+CB97  uniCB97,  HANGUL SYLLABLE-CB97
        0xCB98, # G+15658  U+CB98  uniCB98,  HANGUL SYLLABLE-CB98
        0xCB99, # G+15659  U+CB99  uniCB99,  HANGUL SYLLABLE-CB99
        0xCB9A, # G+15660  U+CB9A  uniCB9A,  HANGUL SYLLABLE-CB9A
        0xCB9B, # G+15661  U+CB9B  uniCB9B,  HANGUL SYLLABLE-CB9B
        0xCB9D, # G+15662  U+CB9D  uniCB9D,  HANGUL SYLLABLE-CB9D
        0xCB9E, # G+15663  U+CB9E  uniCB9E,  HANGUL SYLLABLE-CB9E
        0xCB9F, # G+15664  U+CB9F  uniCB9F,  HANGUL SYLLABLE-CB9F
        0xCBA0, # G+15665  U+CBA0  uniCBA0,  HANGUL SYLLABLE-CBA0
        0xCBA1, # G+15666  U+CBA1  uniCBA1,  HANGUL SYLLABLE-CBA1
        0xCBA2, # G+15667  U+CBA2  uniCBA2,  HANGUL SYLLABLE-CBA2
        0xCBA3, # G+15668  U+CBA3  uniCBA3,  HANGUL SYLLABLE-CBA3
        0xCBA4, # G+15669  U+CBA4  uniCBA4,  HANGUL SYLLABLE-CBA4
        0xCBA5, # G+15670  U+CBA5  uniCBA5,  HANGUL SYLLABLE-CBA5
        0xCBA6, # G+15671  U+CBA6  uniCBA6,  HANGUL SYLLABLE-CBA6
        0xCBA7, # G+15672  U+CBA7  uniCBA7,  HANGUL SYLLABLE-CBA7
        0xCBA8, # G+15673  U+CBA8  uniCBA8,  HANGUL SYLLABLE-CBA8
        0xCBA9, # G+15674  U+CBA9  uniCBA9,  HANGUL SYLLABLE-CBA9
        0xCBAA, # G+15675  U+CBAA  uniCBAA,  HANGUL SYLLABLE-CBAA
        0xCBAB, # G+15676  U+CBAB  uniCBAB,  HANGUL SYLLABLE-CBAB
        0xCBAC, # G+15677  U+CBAC  uniCBAC,  HANGUL SYLLABLE-CBAC
        0xCBAD, # G+15678  U+CBAD  uniCBAD,  HANGUL SYLLABLE-CBAD
        0xCBAE, # G+15679  U+CBAE  uniCBAE,  HANGUL SYLLABLE-CBAE
        0xCBAF, # G+15680  U+CBAF  uniCBAF,  HANGUL SYLLABLE-CBAF
        0xCBB0, # G+15681  U+CBB0  uniCBB0,  HANGUL SYLLABLE-CBB0
        0xCBB1, # G+15682  U+CBB1  uniCBB1,  HANGUL SYLLABLE-CBB1
        0xCBB2, # G+15683  U+CBB2  uniCBB2,  HANGUL SYLLABLE-CBB2
        0xCBB3, # G+15684  U+CBB3  uniCBB3,  HANGUL SYLLABLE-CBB3
        0xCBB4, # G+15685  U+CBB4  uniCBB4,  HANGUL SYLLABLE-CBB4
        0xCBB5, # G+15686  U+CBB5  uniCBB5,  HANGUL SYLLABLE-CBB5
        0xCBB6, # G+15687  U+CBB6  uniCBB6,  HANGUL SYLLABLE-CBB6
        0xCBB7, # G+15688  U+CBB7  uniCBB7,  HANGUL SYLLABLE-CBB7
        0xCBB9, # G+15689  U+CBB9  uniCBB9,  HANGUL SYLLABLE-CBB9
        0xCBBA, # G+15690  U+CBBA  uniCBBA,  HANGUL SYLLABLE-CBBA
        0xCBBB, # G+15691  U+CBBB  uniCBBB,  HANGUL SYLLABLE-CBBB
        0xCBBC, # G+15692  U+CBBC  uniCBBC,  HANGUL SYLLABLE-CBBC
        0xCBBD, # G+15693  U+CBBD  uniCBBD,  HANGUL SYLLABLE-CBBD
        0xCBBE, # G+15694  U+CBBE  uniCBBE,  HANGUL SYLLABLE-CBBE
        0xCBBF, # G+15695  U+CBBF  uniCBBF,  HANGUL SYLLABLE-CBBF
        0xCBC0, # G+15696  U+CBC0  uniCBC0,  HANGUL SYLLABLE-CBC0
        0xCBC1, # G+15697  U+CBC1  uniCBC1,  HANGUL SYLLABLE-CBC1
        0xCBC2, # G+15698  U+CBC2  uniCBC2,  HANGUL SYLLABLE-CBC2
        0xCBC3, # G+15699  U+CBC3  uniCBC3,  HANGUL SYLLABLE-CBC3
        0xCBC4, # G+15700  U+CBC4  uniCBC4,  HANGUL SYLLABLE-CBC4
        0xCBC5, # G+15701  U+CBC5  uniCBC5,  HANGUL SYLLABLE-CBC5
        0xCBC6, # G+15702  U+CBC6  uniCBC6,  HANGUL SYLLABLE-CBC6
        0xCBC7, # G+15703  U+CBC7  uniCBC7,  HANGUL SYLLABLE-CBC7
        0xCBC8, # G+15704  U+CBC8  uniCBC8,  HANGUL SYLLABLE-CBC8
        0xCBC9, # G+15705  U+CBC9  uniCBC9,  HANGUL SYLLABLE-CBC9
        0xCBCA, # G+15706  U+CBCA  uniCBCA,  HANGUL SYLLABLE-CBCA
        0xCBCB, # G+15707  U+CBCB  uniCBCB,  HANGUL SYLLABLE-CBCB
        0xCBCC, # G+15708  U+CBCC  uniCBCC,  HANGUL SYLLABLE-CBCC
        0xCBCD, # G+15709  U+CBCD  uniCBCD,  HANGUL SYLLABLE-CBCD
        0xCBCE, # G+15710  U+CBCE  uniCBCE,  HANGUL SYLLABLE-CBCE
        0xCBCF, # G+15711  U+CBCF  uniCBCF,  HANGUL SYLLABLE-CBCF
        0xCBD0, # G+15712  U+CBD0  uniCBD0,  HANGUL SYLLABLE-CBD0
        0xCBD1, # G+15713  U+CBD1  uniCBD1,  HANGUL SYLLABLE-CBD1
        0xCBD2, # G+15714  U+CBD2  uniCBD2,  HANGUL SYLLABLE-CBD2
        0xCBD3, # G+15715  U+CBD3  uniCBD3,  HANGUL SYLLABLE-CBD3
        0xCBD5, # G+15716  U+CBD5  uniCBD5,  HANGUL SYLLABLE-CBD5
        0xCBD6, # G+15717  U+CBD6  uniCBD6,  HANGUL SYLLABLE-CBD6
        0xCBD7, # G+15718  U+CBD7  uniCBD7,  HANGUL SYLLABLE-CBD7
        0xCBD8, # G+15719  U+CBD8  uniCBD8,  HANGUL SYLLABLE-CBD8
        0xCBD9, # G+15720  U+CBD9  uniCBD9,  HANGUL SYLLABLE-CBD9
        0xCBDA, # G+15721  U+CBDA  uniCBDA,  HANGUL SYLLABLE-CBDA
        0xCBDB, # G+15722  U+CBDB  uniCBDB,  HANGUL SYLLABLE-CBDB
        0xCBDC, # G+15723  U+CBDC  uniCBDC,  HANGUL SYLLABLE-CBDC
        0xCBDD, # G+15724  U+CBDD  uniCBDD,  HANGUL SYLLABLE-CBDD
        0xCBDE, # G+15725  U+CBDE  uniCBDE,  HANGUL SYLLABLE-CBDE
        0xCBDF, # G+15726  U+CBDF  uniCBDF,  HANGUL SYLLABLE-CBDF
        0xCBE0, # G+15727  U+CBE0  uniCBE0,  HANGUL SYLLABLE-CBE0
        0xCBE1, # G+15728  U+CBE1  uniCBE1,  HANGUL SYLLABLE-CBE1
        0xCBE2, # G+15729  U+CBE2  uniCBE2,  HANGUL SYLLABLE-CBE2
        0xCBE3, # G+15730  U+CBE3  uniCBE3,  HANGUL SYLLABLE-CBE3
        0xCBE5, # G+15731  U+CBE5  uniCBE5,  HANGUL SYLLABLE-CBE5
        0xCBE6, # G+15732  U+CBE6  uniCBE6,  HANGUL SYLLABLE-CBE6
        0xCBE8, # G+15733  U+CBE8  uniCBE8,  HANGUL SYLLABLE-CBE8
        0xCBEA, # G+15734  U+CBEA  uniCBEA,  HANGUL SYLLABLE-CBEA
        0xCBEB, # G+15735  U+CBEB  uniCBEB,  HANGUL SYLLABLE-CBEB
        0xCBEC, # G+15736  U+CBEC  uniCBEC,  HANGUL SYLLABLE-CBEC
        0xCBED, # G+15737  U+CBED  uniCBED,  HANGUL SYLLABLE-CBED
        0xCBEE, # G+15738  U+CBEE  uniCBEE,  HANGUL SYLLABLE-CBEE
        0xCBEF, # G+15739  U+CBEF  uniCBEF,  HANGUL SYLLABLE-CBEF
        0xCBF0, # G+15740  U+CBF0  uniCBF0,  HANGUL SYLLABLE-CBF0
        0xCBF1, # G+15741  U+CBF1  uniCBF1,  HANGUL SYLLABLE-CBF1
        0xCBF2, # G+15742  U+CBF2  uniCBF2,  HANGUL SYLLABLE-CBF2
        0xCBF3, # G+15743  U+CBF3  uniCBF3,  HANGUL SYLLABLE-CBF3
        0xCBF4, # G+15744  U+CBF4  uniCBF4,  HANGUL SYLLABLE-CBF4
        0xCBF5, # G+15745  U+CBF5  uniCBF5,  HANGUL SYLLABLE-CBF5
        0xCBF6, # G+15746  U+CBF6  uniCBF6,  HANGUL SYLLABLE-CBF6
        0xCBF7, # G+15747  U+CBF7  uniCBF7,  HANGUL SYLLABLE-CBF7
        0xCBF8, # G+15748  U+CBF8  uniCBF8,  HANGUL SYLLABLE-CBF8
        0xCBF9, # G+15749  U+CBF9  uniCBF9,  HANGUL SYLLABLE-CBF9
        0xCBFA, # G+15750  U+CBFA  uniCBFA,  HANGUL SYLLABLE-CBFA
        0xCBFB, # G+15751  U+CBFB  uniCBFB,  HANGUL SYLLABLE-CBFB
        0xCBFC, # G+15752  U+CBFC  uniCBFC,  HANGUL SYLLABLE-CBFC
        0xCBFD, # G+15753  U+CBFD  uniCBFD,  HANGUL SYLLABLE-CBFD
        0xCBFE, # G+15754  U+CBFE  uniCBFE,  HANGUL SYLLABLE-CBFE
        0xCBFF, # G+15755  U+CBFF  uniCBFF,  HANGUL SYLLABLE-CBFF
        0xCC00, # G+15756  U+CC00  uniCC00,  HANGUL SYLLABLE-CC00
        0xCC01, # G+15757  U+CC01  uniCC01,  HANGUL SYLLABLE-CC01
        0xCC02, # G+15758  U+CC02  uniCC02,  HANGUL SYLLABLE-CC02
        0xCC03, # G+15759  U+CC03  uniCC03,  HANGUL SYLLABLE-CC03
        0xCC04, # G+15760  U+CC04  uniCC04,  HANGUL SYLLABLE-CC04
        0xCC05, # G+15761  U+CC05  uniCC05,  HANGUL SYLLABLE-CC05
        0xCC06, # G+15762  U+CC06  uniCC06,  HANGUL SYLLABLE-CC06
        0xCC07, # G+15763  U+CC07  uniCC07,  HANGUL SYLLABLE-CC07
        0xCC08, # G+15764  U+CC08  uniCC08,  HANGUL SYLLABLE-CC08
        0xCC09, # G+15765  U+CC09  uniCC09,  HANGUL SYLLABLE-CC09
        0xCC0A, # G+15766  U+CC0A  uniCC0A,  HANGUL SYLLABLE-CC0A
        0xCC0B, # G+15767  U+CC0B  uniCC0B,  HANGUL SYLLABLE-CC0B
        0xCC0E, # G+15768  U+CC0E  uniCC0E,  HANGUL SYLLABLE-CC0E
        0xCC0F, # G+15769  U+CC0F  uniCC0F,  HANGUL SYLLABLE-CC0F
        0xCC11, # G+15770  U+CC11  uniCC11,  HANGUL SYLLABLE-CC11
        0xCC12, # G+15771  U+CC12  uniCC12,  HANGUL SYLLABLE-CC12
        0xCC13, # G+15772  U+CC13  uniCC13,  HANGUL SYLLABLE-CC13
        0xCC15, # G+15773  U+CC15  uniCC15,  HANGUL SYLLABLE-CC15
        0xCC16, # G+15774  U+CC16  uniCC16,  HANGUL SYLLABLE-CC16
        0xCC17, # G+15775  U+CC17  uniCC17,  HANGUL SYLLABLE-CC17
        0xCC18, # G+15776  U+CC18  uniCC18,  HANGUL SYLLABLE-CC18
        0xCC19, # G+15777  U+CC19  uniCC19,  HANGUL SYLLABLE-CC19
        0xCC1A, # G+15778  U+CC1A  uniCC1A,  HANGUL SYLLABLE-CC1A
        0xCC1B, # G+15779  U+CC1B  uniCC1B,  HANGUL SYLLABLE-CC1B
        0xCC1E, # G+15780  U+CC1E  uniCC1E,  HANGUL SYLLABLE-CC1E
        0xCC1F, # G+15781  U+CC1F  uniCC1F,  HANGUL SYLLABLE-CC1F
        0xCC20, # G+15782  U+CC20  uniCC20,  HANGUL SYLLABLE-CC20
        0xCC23, # G+15783  U+CC23  uniCC23,  HANGUL SYLLABLE-CC23
        0xCC24, # G+15784  U+CC24  uniCC24,  HANGUL SYLLABLE-CC24
        0xCC25, # G+15785  U+CC25  uniCC25,  HANGUL SYLLABLE-CC25
        0xCC26, # G+15786  U+CC26  uniCC26,  HANGUL SYLLABLE-CC26
        0xCC2A, # G+15787  U+CC2A  uniCC2A,  HANGUL SYLLABLE-CC2A
        0xCC2B, # G+15788  U+CC2B  uniCC2B,  HANGUL SYLLABLE-CC2B
        0xCC2D, # G+15789  U+CC2D  uniCC2D,  HANGUL SYLLABLE-CC2D
        0xCC2F, # G+15790  U+CC2F  uniCC2F,  HANGUL SYLLABLE-CC2F
        0xCC31, # G+15791  U+CC31  uniCC31,  HANGUL SYLLABLE-CC31
        0xCC32, # G+15792  U+CC32  uniCC32,  HANGUL SYLLABLE-CC32
        0xCC33, # G+15793  U+CC33  uniCC33,  HANGUL SYLLABLE-CC33
        0xCC34, # G+15794  U+CC34  uniCC34,  HANGUL SYLLABLE-CC34
        0xCC35, # G+15795  U+CC35  uniCC35,  HANGUL SYLLABLE-CC35
        0xCC36, # G+15796  U+CC36  uniCC36,  HANGUL SYLLABLE-CC36
        0xCC37, # G+15797  U+CC37  uniCC37,  HANGUL SYLLABLE-CC37
        0xCC3A, # G+15798  U+CC3A  uniCC3A,  HANGUL SYLLABLE-CC3A
        0xCC3F, # G+15799  U+CC3F  uniCC3F,  HANGUL SYLLABLE-CC3F
        0xCC40, # G+15800  U+CC40  uniCC40,  HANGUL SYLLABLE-CC40
        0xCC41, # G+15801  U+CC41  uniCC41,  HANGUL SYLLABLE-CC41
        0xCC42, # G+15802  U+CC42  uniCC42,  HANGUL SYLLABLE-CC42
        0xCC43, # G+15803  U+CC43  uniCC43,  HANGUL SYLLABLE-CC43
        0xCC46, # G+15804  U+CC46  uniCC46,  HANGUL SYLLABLE-CC46
        0xCC47, # G+15805  U+CC47  uniCC47,  HANGUL SYLLABLE-CC47
        0xCC49, # G+15806  U+CC49  uniCC49,  HANGUL SYLLABLE-CC49
        0xCC4A, # G+15807  U+CC4A  uniCC4A,  HANGUL SYLLABLE-CC4A
        0xCC4B, # G+15808  U+CC4B  uniCC4B,  HANGUL SYLLABLE-CC4B
        0xCC4D, # G+15809  U+CC4D  uniCC4D,  HANGUL SYLLABLE-CC4D
        0xCC4E, # G+15810  U+CC4E  uniCC4E,  HANGUL SYLLABLE-CC4E
        0xCC4F, # G+15811  U+CC4F  uniCC4F,  HANGUL SYLLABLE-CC4F
        0xCC50, # G+15812  U+CC50  uniCC50,  HANGUL SYLLABLE-CC50
        0xCC51, # G+15813  U+CC51  uniCC51,  HANGUL SYLLABLE-CC51
        0xCC52, # G+15814  U+CC52  uniCC52,  HANGUL SYLLABLE-CC52
        0xCC53, # G+15815  U+CC53  uniCC53,  HANGUL SYLLABLE-CC53
        0xCC56, # G+15816  U+CC56  uniCC56,  HANGUL SYLLABLE-CC56
        0xCC5A, # G+15817  U+CC5A  uniCC5A,  HANGUL SYLLABLE-CC5A
        0xCC5B, # G+15818  U+CC5B  uniCC5B,  HANGUL SYLLABLE-CC5B
        0xCC5C, # G+15819  U+CC5C  uniCC5C,  HANGUL SYLLABLE-CC5C
        0xCC5D, # G+15820  U+CC5D  uniCC5D,  HANGUL SYLLABLE-CC5D
        0xCC5E, # G+15821  U+CC5E  uniCC5E,  HANGUL SYLLABLE-CC5E
        0xCC5F, # G+15822  U+CC5F  uniCC5F,  HANGUL SYLLABLE-CC5F
        0xCC61, # G+15823  U+CC61  uniCC61,  HANGUL SYLLABLE-CC61
        0xCC62, # G+15824  U+CC62  uniCC62,  HANGUL SYLLABLE-CC62
        0xCC63, # G+15825  U+CC63  uniCC63,  HANGUL SYLLABLE-CC63
        0xCC65, # G+15826  U+CC65  uniCC65,  HANGUL SYLLABLE-CC65
        0xCC67, # G+15827  U+CC67  uniCC67,  HANGUL SYLLABLE-CC67
        0xCC69, # G+15828  U+CC69  uniCC69,  HANGUL SYLLABLE-CC69
        0xCC6A, # G+15829  U+CC6A  uniCC6A,  HANGUL SYLLABLE-CC6A
        0xCC6B, # G+15830  U+CC6B  uniCC6B,  HANGUL SYLLABLE-CC6B
        0xCC6C, # G+15831  U+CC6C  uniCC6C,  HANGUL SYLLABLE-CC6C
        0xCC6D, # G+15832  U+CC6D  uniCC6D,  HANGUL SYLLABLE-CC6D
        0xCC6E, # G+15833  U+CC6E  uniCC6E,  HANGUL SYLLABLE-CC6E
        0xCC6F, # G+15834  U+CC6F  uniCC6F,  HANGUL SYLLABLE-CC6F
        0xCC71, # G+15835  U+CC71  uniCC71,  HANGUL SYLLABLE-CC71
        0xCC72, # G+15836  U+CC72  uniCC72,  HANGUL SYLLABLE-CC72
        0xCC73, # G+15837  U+CC73  uniCC73,  HANGUL SYLLABLE-CC73
        0xCC74, # G+15838  U+CC74  uniCC74,  HANGUL SYLLABLE-CC74
        0xCC76, # G+15839  U+CC76  uniCC76,  HANGUL SYLLABLE-CC76
        0xCC77, # G+15840  U+CC77  uniCC77,  HANGUL SYLLABLE-CC77
        0xCC78, # G+15841  U+CC78  uniCC78,  HANGUL SYLLABLE-CC78
        0xCC79, # G+15842  U+CC79  uniCC79,  HANGUL SYLLABLE-CC79
        0xCC7A, # G+15843  U+CC7A  uniCC7A,  HANGUL SYLLABLE-CC7A
        0xCC7B, # G+15844  U+CC7B  uniCC7B,  HANGUL SYLLABLE-CC7B
        0xCC7C, # G+15845  U+CC7C  uniCC7C,  HANGUL SYLLABLE-CC7C
        0xCC7D, # G+15846  U+CC7D  uniCC7D,  HANGUL SYLLABLE-CC7D
        0xCC7E, # G+15847  U+CC7E  uniCC7E,  HANGUL SYLLABLE-CC7E
        0xCC7F, # G+15848  U+CC7F  uniCC7F,  HANGUL SYLLABLE-CC7F
        0xCC80, # G+15849  U+CC80  uniCC80,  HANGUL SYLLABLE-CC80
        0xCC81, # G+15850  U+CC81  uniCC81,  HANGUL SYLLABLE-CC81
        0xCC82, # G+15851  U+CC82  uniCC82,  HANGUL SYLLABLE-CC82
        0xCC83, # G+15852  U+CC83  uniCC83,  HANGUL SYLLABLE-CC83
        0xCC84, # G+15853  U+CC84  uniCC84,  HANGUL SYLLABLE-CC84
        0xCC85, # G+15854  U+CC85  uniCC85,  HANGUL SYLLABLE-CC85
        0xCC86, # G+15855  U+CC86  uniCC86,  HANGUL SYLLABLE-CC86
        0xCC87, # G+15856  U+CC87  uniCC87,  HANGUL SYLLABLE-CC87
        0xCC88, # G+15857  U+CC88  uniCC88,  HANGUL SYLLABLE-CC88
        0xCC89, # G+15858  U+CC89  uniCC89,  HANGUL SYLLABLE-CC89
        0xCC8A, # G+15859  U+CC8A  uniCC8A,  HANGUL SYLLABLE-CC8A
        0xCC8B, # G+15860  U+CC8B  uniCC8B,  HANGUL SYLLABLE-CC8B
        0xCC8C, # G+15861  U+CC8C  uniCC8C,  HANGUL SYLLABLE-CC8C
        0xCC8D, # G+15862  U+CC8D  uniCC8D,  HANGUL SYLLABLE-CC8D
        0xCC8E, # G+15863  U+CC8E  uniCC8E,  HANGUL SYLLABLE-CC8E
        0xCC8F, # G+15864  U+CC8F  uniCC8F,  HANGUL SYLLABLE-CC8F
        0xCC90, # G+15865  U+CC90  uniCC90,  HANGUL SYLLABLE-CC90
        0xCC91, # G+15866  U+CC91  uniCC91,  HANGUL SYLLABLE-CC91
        0xCC92, # G+15867  U+CC92  uniCC92,  HANGUL SYLLABLE-CC92
        0xCC93, # G+15868  U+CC93  uniCC93,  HANGUL SYLLABLE-CC93
        0xCC94, # G+15869  U+CC94  uniCC94,  HANGUL SYLLABLE-CC94
        0xCC95, # G+15870  U+CC95  uniCC95,  HANGUL SYLLABLE-CC95
        0xCC96, # G+15871  U+CC96  uniCC96,  HANGUL SYLLABLE-CC96
        0xCC97, # G+15872  U+CC97  uniCC97,  HANGUL SYLLABLE-CC97
        0xCC9A, # G+15873  U+CC9A  uniCC9A,  HANGUL SYLLABLE-CC9A
        0xCC9B, # G+15874  U+CC9B  uniCC9B,  HANGUL SYLLABLE-CC9B
        0xCC9D, # G+15875  U+CC9D  uniCC9D,  HANGUL SYLLABLE-CC9D
        0xCC9E, # G+15876  U+CC9E  uniCC9E,  HANGUL SYLLABLE-CC9E
        0xCC9F, # G+15877  U+CC9F  uniCC9F,  HANGUL SYLLABLE-CC9F
        0xCCA1, # G+15878  U+CCA1  uniCCA1,  HANGUL SYLLABLE-CCA1
        0xCCA2, # G+15879  U+CCA2  uniCCA2,  HANGUL SYLLABLE-CCA2
        0xCCA3, # G+15880  U+CCA3  uniCCA3,  HANGUL SYLLABLE-CCA3
        0xCCA4, # G+15881  U+CCA4  uniCCA4,  HANGUL SYLLABLE-CCA4
        0xCCA5, # G+15882  U+CCA5  uniCCA5,  HANGUL SYLLABLE-CCA5
        0xCCA6, # G+15883  U+CCA6  uniCCA6,  HANGUL SYLLABLE-CCA6
        0xCCA7, # G+15884  U+CCA7  uniCCA7,  HANGUL SYLLABLE-CCA7
        0xCCAA, # G+15885  U+CCAA  uniCCAA,  HANGUL SYLLABLE-CCAA
        0xCCAE, # G+15886  U+CCAE  uniCCAE,  HANGUL SYLLABLE-CCAE
        0xCCAF, # G+15887  U+CCAF  uniCCAF,  HANGUL SYLLABLE-CCAF
        0xCCB0, # G+15888  U+CCB0  uniCCB0,  HANGUL SYLLABLE-CCB0
        0xCCB1, # G+15889  U+CCB1  uniCCB1,  HANGUL SYLLABLE-CCB1
        0xCCB2, # G+15890  U+CCB2  uniCCB2,  HANGUL SYLLABLE-CCB2
        0xCCB3, # G+15891  U+CCB3  uniCCB3,  HANGUL SYLLABLE-CCB3
        0xCCB6, # G+15892  U+CCB6  uniCCB6,  HANGUL SYLLABLE-CCB6
        0xCCB7, # G+15893  U+CCB7  uniCCB7,  HANGUL SYLLABLE-CCB7
        0xCCB9, # G+15894  U+CCB9  uniCCB9,  HANGUL SYLLABLE-CCB9
        0xCCBA, # G+15895  U+CCBA  uniCCBA,  HANGUL SYLLABLE-CCBA
        0xCCBB, # G+15896  U+CCBB  uniCCBB,  HANGUL SYLLABLE-CCBB
        0xCCBD, # G+15897  U+CCBD  uniCCBD,  HANGUL SYLLABLE-CCBD
        0xCCBE, # G+15898  U+CCBE  uniCCBE,  HANGUL SYLLABLE-CCBE
        0xCCBF, # G+15899  U+CCBF  uniCCBF,  HANGUL SYLLABLE-CCBF
        0xCCC0, # G+15900  U+CCC0  uniCCC0,  HANGUL SYLLABLE-CCC0
        0xCCC1, # G+15901  U+CCC1  uniCCC1,  HANGUL SYLLABLE-CCC1
        0xCCC2, # G+15902  U+CCC2  uniCCC2,  HANGUL SYLLABLE-CCC2
        0xCCC3, # G+15903  U+CCC3  uniCCC3,  HANGUL SYLLABLE-CCC3
        0xCCC6, # G+15904  U+CCC6  uniCCC6,  HANGUL SYLLABLE-CCC6
        0xCCC8, # G+15905  U+CCC8  uniCCC8,  HANGUL SYLLABLE-CCC8
        0xCCCA, # G+15906  U+CCCA  uniCCCA,  HANGUL SYLLABLE-CCCA
        0xCCCB, # G+15907  U+CCCB  uniCCCB,  HANGUL SYLLABLE-CCCB
        0xCCCC, # G+15908  U+CCCC  uniCCCC,  HANGUL SYLLABLE-CCCC
        0xCCCD, # G+15909  U+CCCD  uniCCCD,  HANGUL SYLLABLE-CCCD
        0xCCCE, # G+15910  U+CCCE  uniCCCE,  HANGUL SYLLABLE-CCCE
        0xCCCF, # G+15911  U+CCCF  uniCCCF,  HANGUL SYLLABLE-CCCF
        0xCCD1, # G+15912  U+CCD1  uniCCD1,  HANGUL SYLLABLE-CCD1
        0xCCD2, # G+15913  U+CCD2  uniCCD2,  HANGUL SYLLABLE-CCD2
        0xCCD3, # G+15914  U+CCD3  uniCCD3,  HANGUL SYLLABLE-CCD3
        0xCCD5, # G+15915  U+CCD5  uniCCD5,  HANGUL SYLLABLE-CCD5
        0xCCD6, # G+15916  U+CCD6  uniCCD6,  HANGUL SYLLABLE-CCD6
        0xCCD7, # G+15917  U+CCD7  uniCCD7,  HANGUL SYLLABLE-CCD7
        0xCCD8, # G+15918  U+CCD8  uniCCD8,  HANGUL SYLLABLE-CCD8
        0xCCD9, # G+15919  U+CCD9  uniCCD9,  HANGUL SYLLABLE-CCD9
        0xCCDA, # G+15920  U+CCDA  uniCCDA,  HANGUL SYLLABLE-CCDA
        0xCCDB, # G+15921  U+CCDB  uniCCDB,  HANGUL SYLLABLE-CCDB
        0xCCDC, # G+15922  U+CCDC  uniCCDC,  HANGUL SYLLABLE-CCDC
        0xCCDD, # G+15923  U+CCDD  uniCCDD,  HANGUL SYLLABLE-CCDD
        0xCCDE, # G+15924  U+CCDE  uniCCDE,  HANGUL SYLLABLE-CCDE
        0xCCDF, # G+15925  U+CCDF  uniCCDF,  HANGUL SYLLABLE-CCDF
        0xCCE0, # G+15926  U+CCE0  uniCCE0,  HANGUL SYLLABLE-CCE0
        0xCCE1, # G+15927  U+CCE1  uniCCE1,  HANGUL SYLLABLE-CCE1
        0xCCE2, # G+15928  U+CCE2  uniCCE2,  HANGUL SYLLABLE-CCE2
        0xCCE3, # G+15929  U+CCE3  uniCCE3,  HANGUL SYLLABLE-CCE3
        0xCCE5, # G+15930  U+CCE5  uniCCE5,  HANGUL SYLLABLE-CCE5
        0xCCE6, # G+15931  U+CCE6  uniCCE6,  HANGUL SYLLABLE-CCE6
        0xCCE7, # G+15932  U+CCE7  uniCCE7,  HANGUL SYLLABLE-CCE7
        0xCCE8, # G+15933  U+CCE8  uniCCE8,  HANGUL SYLLABLE-CCE8
        0xCCE9, # G+15934  U+CCE9  uniCCE9,  HANGUL SYLLABLE-CCE9
        0xCCEA, # G+15935  U+CCEA  uniCCEA,  HANGUL SYLLABLE-CCEA
        0xCCEB, # G+15936  U+CCEB  uniCCEB,  HANGUL SYLLABLE-CCEB
        0xCCED, # G+15937  U+CCED  uniCCED,  HANGUL SYLLABLE-CCED
        0xCCEE, # G+15938  U+CCEE  uniCCEE,  HANGUL SYLLABLE-CCEE
        0xCCEF, # G+15939  U+CCEF  uniCCEF,  HANGUL SYLLABLE-CCEF
        0xCCF1, # G+15940  U+CCF1  uniCCF1,  HANGUL SYLLABLE-CCF1
        0xCCF2, # G+15941  U+CCF2  uniCCF2,  HANGUL SYLLABLE-CCF2
        0xCCF3, # G+15942  U+CCF3  uniCCF3,  HANGUL SYLLABLE-CCF3
        0xCCF4, # G+15943  U+CCF4  uniCCF4,  HANGUL SYLLABLE-CCF4
        0xCCF5, # G+15944  U+CCF5  uniCCF5,  HANGUL SYLLABLE-CCF5
        0xCCF6, # G+15945  U+CCF6  uniCCF6,  HANGUL SYLLABLE-CCF6
        0xCCF7, # G+15946  U+CCF7  uniCCF7,  HANGUL SYLLABLE-CCF7
        0xCCF8, # G+15947  U+CCF8  uniCCF8,  HANGUL SYLLABLE-CCF8
        0xCCF9, # G+15948  U+CCF9  uniCCF9,  HANGUL SYLLABLE-CCF9
        0xCCFA, # G+15949  U+CCFA  uniCCFA,  HANGUL SYLLABLE-CCFA
        0xCCFB, # G+15950  U+CCFB  uniCCFB,  HANGUL SYLLABLE-CCFB
        0xCCFC, # G+15951  U+CCFC  uniCCFC,  HANGUL SYLLABLE-CCFC
        0xCCFD, # G+15952  U+CCFD  uniCCFD,  HANGUL SYLLABLE-CCFD
        0xCCFE, # G+15953  U+CCFE  uniCCFE,  HANGUL SYLLABLE-CCFE
        0xCCFF, # G+15954  U+CCFF  uniCCFF,  HANGUL SYLLABLE-CCFF
        0xCD00, # G+15955  U+CD00  uniCD00,  HANGUL SYLLABLE-CD00
        0xCD02, # G+15956  U+CD02  uniCD02,  HANGUL SYLLABLE-CD02
        0xCD03, # G+15957  U+CD03  uniCD03,  HANGUL SYLLABLE-CD03
        0xCD04, # G+15958  U+CD04  uniCD04,  HANGUL SYLLABLE-CD04
        0xCD05, # G+15959  U+CD05  uniCD05,  HANGUL SYLLABLE-CD05
        0xCD06, # G+15960  U+CD06  uniCD06,  HANGUL SYLLABLE-CD06
        0xCD07, # G+15961  U+CD07  uniCD07,  HANGUL SYLLABLE-CD07
        0xCD0A, # G+15962  U+CD0A  uniCD0A,  HANGUL SYLLABLE-CD0A
        0xCD0B, # G+15963  U+CD0B  uniCD0B,  HANGUL SYLLABLE-CD0B
        0xCD0D, # G+15964  U+CD0D  uniCD0D,  HANGUL SYLLABLE-CD0D
        0xCD0E, # G+15965  U+CD0E  uniCD0E,  HANGUL SYLLABLE-CD0E
        0xCD0F, # G+15966  U+CD0F  uniCD0F,  HANGUL SYLLABLE-CD0F
        0xCD11, # G+15967  U+CD11  uniCD11,  HANGUL SYLLABLE-CD11
        0xCD12, # G+15968  U+CD12  uniCD12,  HANGUL SYLLABLE-CD12
        0xCD13, # G+15969  U+CD13  uniCD13,  HANGUL SYLLABLE-CD13
        0xCD14, # G+15970  U+CD14  uniCD14,  HANGUL SYLLABLE-CD14
        0xCD15, # G+15971  U+CD15  uniCD15,  HANGUL SYLLABLE-CD15
        0xCD16, # G+15972  U+CD16  uniCD16,  HANGUL SYLLABLE-CD16
        0xCD17, # G+15973  U+CD17  uniCD17,  HANGUL SYLLABLE-CD17
        0xCD1A, # G+15974  U+CD1A  uniCD1A,  HANGUL SYLLABLE-CD1A
        0xCD1C, # G+15975  U+CD1C  uniCD1C,  HANGUL SYLLABLE-CD1C
        0xCD1E, # G+15976  U+CD1E  uniCD1E,  HANGUL SYLLABLE-CD1E
        0xCD1F, # G+15977  U+CD1F  uniCD1F,  HANGUL SYLLABLE-CD1F
        0xCD20, # G+15978  U+CD20  uniCD20,  HANGUL SYLLABLE-CD20
        0xCD21, # G+15979  U+CD21  uniCD21,  HANGUL SYLLABLE-CD21
        0xCD22, # G+15980  U+CD22  uniCD22,  HANGUL SYLLABLE-CD22
        0xCD23, # G+15981  U+CD23  uniCD23,  HANGUL SYLLABLE-CD23
        0xCD25, # G+15982  U+CD25  uniCD25,  HANGUL SYLLABLE-CD25
        0xCD26, # G+15983  U+CD26  uniCD26,  HANGUL SYLLABLE-CD26
        0xCD27, # G+15984  U+CD27  uniCD27,  HANGUL SYLLABLE-CD27
        0xCD29, # G+15985  U+CD29  uniCD29,  HANGUL SYLLABLE-CD29
        0xCD2A, # G+15986  U+CD2A  uniCD2A,  HANGUL SYLLABLE-CD2A
        0xCD2B, # G+15987  U+CD2B  uniCD2B,  HANGUL SYLLABLE-CD2B
        0xCD2D, # G+15988  U+CD2D  uniCD2D,  HANGUL SYLLABLE-CD2D
        0xCD2E, # G+15989  U+CD2E  uniCD2E,  HANGUL SYLLABLE-CD2E
        0xCD2F, # G+15990  U+CD2F  uniCD2F,  HANGUL SYLLABLE-CD2F
        0xCD30, # G+15991  U+CD30  uniCD30,  HANGUL SYLLABLE-CD30
        0xCD31, # G+15992  U+CD31  uniCD31,  HANGUL SYLLABLE-CD31
        0xCD32, # G+15993  U+CD32  uniCD32,  HANGUL SYLLABLE-CD32
        0xCD33, # G+15994  U+CD33  uniCD33,  HANGUL SYLLABLE-CD33
        0xCD34, # G+15995  U+CD34  uniCD34,  HANGUL SYLLABLE-CD34
        0xCD35, # G+15996  U+CD35  uniCD35,  HANGUL SYLLABLE-CD35
        0xCD36, # G+15997  U+CD36  uniCD36,  HANGUL SYLLABLE-CD36
        0xCD37, # G+15998  U+CD37  uniCD37,  HANGUL SYLLABLE-CD37
        0xCD38, # G+15999  U+CD38  uniCD38,  HANGUL SYLLABLE-CD38
        0xCD3A, # G+16000  U+CD3A  uniCD3A,  HANGUL SYLLABLE-CD3A
        0xCD3B, # G+16001  U+CD3B  uniCD3B,  HANGUL SYLLABLE-CD3B
        0xCD3C, # G+16002  U+CD3C  uniCD3C,  HANGUL SYLLABLE-CD3C
        0xCD3D, # G+16003  U+CD3D  uniCD3D,  HANGUL SYLLABLE-CD3D
        0xCD3E, # G+16004  U+CD3E  uniCD3E,  HANGUL SYLLABLE-CD3E
        0xCD3F, # G+16005  U+CD3F  uniCD3F,  HANGUL SYLLABLE-CD3F
        0xCD40, # G+16006  U+CD40  uniCD40,  HANGUL SYLLABLE-CD40
        0xCD41, # G+16007  U+CD41  uniCD41,  HANGUL SYLLABLE-CD41
        0xCD42, # G+16008  U+CD42  uniCD42,  HANGUL SYLLABLE-CD42
        0xCD43, # G+16009  U+CD43  uniCD43,  HANGUL SYLLABLE-CD43
        0xCD44, # G+16010  U+CD44  uniCD44,  HANGUL SYLLABLE-CD44
        0xCD45, # G+16011  U+CD45  uniCD45,  HANGUL SYLLABLE-CD45
        0xCD46, # G+16012  U+CD46  uniCD46,  HANGUL SYLLABLE-CD46
        0xCD47, # G+16013  U+CD47  uniCD47,  HANGUL SYLLABLE-CD47
        0xCD48, # G+16014  U+CD48  uniCD48,  HANGUL SYLLABLE-CD48
        0xCD49, # G+16015  U+CD49  uniCD49,  HANGUL SYLLABLE-CD49
        0xCD4A, # G+16016  U+CD4A  uniCD4A,  HANGUL SYLLABLE-CD4A
        0xCD4B, # G+16017  U+CD4B  uniCD4B,  HANGUL SYLLABLE-CD4B
        0xCD4C, # G+16018  U+CD4C  uniCD4C,  HANGUL SYLLABLE-CD4C
        0xCD4D, # G+16019  U+CD4D  uniCD4D,  HANGUL SYLLABLE-CD4D
        0xCD4E, # G+16020  U+CD4E  uniCD4E,  HANGUL SYLLABLE-CD4E
        0xCD4F, # G+16021  U+CD4F  uniCD4F,  HANGUL SYLLABLE-CD4F
        0xCD50, # G+16022  U+CD50  uniCD50,  HANGUL SYLLABLE-CD50
        0xCD51, # G+16023  U+CD51  uniCD51,  HANGUL SYLLABLE-CD51
        0xCD52, # G+16024  U+CD52  uniCD52,  HANGUL SYLLABLE-CD52
        0xCD53, # G+16025  U+CD53  uniCD53,  HANGUL SYLLABLE-CD53
        0xCD54, # G+16026  U+CD54  uniCD54,  HANGUL SYLLABLE-CD54
        0xCD55, # G+16027  U+CD55  uniCD55,  HANGUL SYLLABLE-CD55
        0xCD56, # G+16028  U+CD56  uniCD56,  HANGUL SYLLABLE-CD56
        0xCD57, # G+16029  U+CD57  uniCD57,  HANGUL SYLLABLE-CD57
        0xCD58, # G+16030  U+CD58  uniCD58,  HANGUL SYLLABLE-CD58
        0xCD59, # G+16031  U+CD59  uniCD59,  HANGUL SYLLABLE-CD59
        0xCD5A, # G+16032  U+CD5A  uniCD5A,  HANGUL SYLLABLE-CD5A
        0xCD5B, # G+16033  U+CD5B  uniCD5B,  HANGUL SYLLABLE-CD5B
        0xCD5D, # G+16034  U+CD5D  uniCD5D,  HANGUL SYLLABLE-CD5D
        0xCD5E, # G+16035  U+CD5E  uniCD5E,  HANGUL SYLLABLE-CD5E
        0xCD5F, # G+16036  U+CD5F  uniCD5F,  HANGUL SYLLABLE-CD5F
        0xCD61, # G+16037  U+CD61  uniCD61,  HANGUL SYLLABLE-CD61
        0xCD62, # G+16038  U+CD62  uniCD62,  HANGUL SYLLABLE-CD62
        0xCD63, # G+16039  U+CD63  uniCD63,  HANGUL SYLLABLE-CD63
        0xCD65, # G+16040  U+CD65  uniCD65,  HANGUL SYLLABLE-CD65
        0xCD66, # G+16041  U+CD66  uniCD66,  HANGUL SYLLABLE-CD66
        0xCD67, # G+16042  U+CD67  uniCD67,  HANGUL SYLLABLE-CD67
        0xCD68, # G+16043  U+CD68  uniCD68,  HANGUL SYLLABLE-CD68
        0xCD69, # G+16044  U+CD69  uniCD69,  HANGUL SYLLABLE-CD69
        0xCD6A, # G+16045  U+CD6A  uniCD6A,  HANGUL SYLLABLE-CD6A
        0xCD6B, # G+16046  U+CD6B  uniCD6B,  HANGUL SYLLABLE-CD6B
        0xCD6E, # G+16047  U+CD6E  uniCD6E,  HANGUL SYLLABLE-CD6E
        0xCD70, # G+16048  U+CD70  uniCD70,  HANGUL SYLLABLE-CD70
        0xCD72, # G+16049  U+CD72  uniCD72,  HANGUL SYLLABLE-CD72
        0xCD73, # G+16050  U+CD73  uniCD73,  HANGUL SYLLABLE-CD73
        0xCD74, # G+16051  U+CD74  uniCD74,  HANGUL SYLLABLE-CD74
        0xCD75, # G+16052  U+CD75  uniCD75,  HANGUL SYLLABLE-CD75
        0xCD76, # G+16053  U+CD76  uniCD76,  HANGUL SYLLABLE-CD76
        0xCD77, # G+16054  U+CD77  uniCD77,  HANGUL SYLLABLE-CD77
        0xCD79, # G+16055  U+CD79  uniCD79,  HANGUL SYLLABLE-CD79
        0xCD7A, # G+16056  U+CD7A  uniCD7A,  HANGUL SYLLABLE-CD7A
        0xCD7B, # G+16057  U+CD7B  uniCD7B,  HANGUL SYLLABLE-CD7B
        0xCD7C, # G+16058  U+CD7C  uniCD7C,  HANGUL SYLLABLE-CD7C
        0xCD7D, # G+16059  U+CD7D  uniCD7D,  HANGUL SYLLABLE-CD7D
        0xCD7E, # G+16060  U+CD7E  uniCD7E,  HANGUL SYLLABLE-CD7E
        0xCD7F, # G+16061  U+CD7F  uniCD7F,  HANGUL SYLLABLE-CD7F
        0xCD80, # G+16062  U+CD80  uniCD80,  HANGUL SYLLABLE-CD80
        0xCD81, # G+16063  U+CD81  uniCD81,  HANGUL SYLLABLE-CD81
        0xCD82, # G+16064  U+CD82  uniCD82,  HANGUL SYLLABLE-CD82
        0xCD83, # G+16065  U+CD83  uniCD83,  HANGUL SYLLABLE-CD83
        0xCD84, # G+16066  U+CD84  uniCD84,  HANGUL SYLLABLE-CD84
        0xCD85, # G+16067  U+CD85  uniCD85,  HANGUL SYLLABLE-CD85
        0xCD86, # G+16068  U+CD86  uniCD86,  HANGUL SYLLABLE-CD86
        0xCD87, # G+16069  U+CD87  uniCD87,  HANGUL SYLLABLE-CD87
        0xCD89, # G+16070  U+CD89  uniCD89,  HANGUL SYLLABLE-CD89
        0xCD8A, # G+16071  U+CD8A  uniCD8A,  HANGUL SYLLABLE-CD8A
        0xCD8B, # G+16072  U+CD8B  uniCD8B,  HANGUL SYLLABLE-CD8B
        0xCD8C, # G+16073  U+CD8C  uniCD8C,  HANGUL SYLLABLE-CD8C
        0xCD8D, # G+16074  U+CD8D  uniCD8D,  HANGUL SYLLABLE-CD8D
        0xCD8E, # G+16075  U+CD8E  uniCD8E,  HANGUL SYLLABLE-CD8E
        0xCD8F, # G+16076  U+CD8F  uniCD8F,  HANGUL SYLLABLE-CD8F
        0xCD90, # G+16077  U+CD90  uniCD90,  HANGUL SYLLABLE-CD90
        0xCD91, # G+16078  U+CD91  uniCD91,  HANGUL SYLLABLE-CD91
        0xCD92, # G+16079  U+CD92  uniCD92,  HANGUL SYLLABLE-CD92
        0xCD93, # G+16080  U+CD93  uniCD93,  HANGUL SYLLABLE-CD93
        0xCD96, # G+16081  U+CD96  uniCD96,  HANGUL SYLLABLE-CD96
        0xCD97, # G+16082  U+CD97  uniCD97,  HANGUL SYLLABLE-CD97
        0xCD99, # G+16083  U+CD99  uniCD99,  HANGUL SYLLABLE-CD99
        0xCD9A, # G+16084  U+CD9A  uniCD9A,  HANGUL SYLLABLE-CD9A
        0xCD9B, # G+16085  U+CD9B  uniCD9B,  HANGUL SYLLABLE-CD9B
        0xCD9D, # G+16086  U+CD9D  uniCD9D,  HANGUL SYLLABLE-CD9D
        0xCD9E, # G+16087  U+CD9E  uniCD9E,  HANGUL SYLLABLE-CD9E
        0xCD9F, # G+16088  U+CD9F  uniCD9F,  HANGUL SYLLABLE-CD9F
        0xCDA0, # G+16089  U+CDA0  uniCDA0,  HANGUL SYLLABLE-CDA0
        0xCDA1, # G+16090  U+CDA1  uniCDA1,  HANGUL SYLLABLE-CDA1
        0xCDA2, # G+16091  U+CDA2  uniCDA2,  HANGUL SYLLABLE-CDA2
        0xCDA3, # G+16092  U+CDA3  uniCDA3,  HANGUL SYLLABLE-CDA3
        0xCDA6, # G+16093  U+CDA6  uniCDA6,  HANGUL SYLLABLE-CDA6
        0xCDA8, # G+16094  U+CDA8  uniCDA8,  HANGUL SYLLABLE-CDA8
        0xCDAA, # G+16095  U+CDAA  uniCDAA,  HANGUL SYLLABLE-CDAA
        0xCDAB, # G+16096  U+CDAB  uniCDAB,  HANGUL SYLLABLE-CDAB
        0xCDAC, # G+16097  U+CDAC  uniCDAC,  HANGUL SYLLABLE-CDAC
        0xCDAD, # G+16098  U+CDAD  uniCDAD,  HANGUL SYLLABLE-CDAD
        0xCDAE, # G+16099  U+CDAE  uniCDAE,  HANGUL SYLLABLE-CDAE
        0xCDAF, # G+16100  U+CDAF  uniCDAF,  HANGUL SYLLABLE-CDAF
        0xCDB1, # G+16101  U+CDB1  uniCDB1,  HANGUL SYLLABLE-CDB1
        0xCDB2, # G+16102  U+CDB2  uniCDB2,  HANGUL SYLLABLE-CDB2
        0xCDB3, # G+16103  U+CDB3  uniCDB3,  HANGUL SYLLABLE-CDB3
        0xCDB4, # G+16104  U+CDB4  uniCDB4,  HANGUL SYLLABLE-CDB4
        0xCDB5, # G+16105  U+CDB5  uniCDB5,  HANGUL SYLLABLE-CDB5
        0xCDB6, # G+16106  U+CDB6  uniCDB6,  HANGUL SYLLABLE-CDB6
        0xCDB7, # G+16107  U+CDB7  uniCDB7,  HANGUL SYLLABLE-CDB7
        0xCDB8, # G+16108  U+CDB8  uniCDB8,  HANGUL SYLLABLE-CDB8
        0xCDB9, # G+16109  U+CDB9  uniCDB9,  HANGUL SYLLABLE-CDB9
        0xCDBA, # G+16110  U+CDBA  uniCDBA,  HANGUL SYLLABLE-CDBA
        0xCDBB, # G+16111  U+CDBB  uniCDBB,  HANGUL SYLLABLE-CDBB
        0xCDBC, # G+16112  U+CDBC  uniCDBC,  HANGUL SYLLABLE-CDBC
        0xCDBD, # G+16113  U+CDBD  uniCDBD,  HANGUL SYLLABLE-CDBD
        0xCDBE, # G+16114  U+CDBE  uniCDBE,  HANGUL SYLLABLE-CDBE
        0xCDBF, # G+16115  U+CDBF  uniCDBF,  HANGUL SYLLABLE-CDBF
        0xCDC0, # G+16116  U+CDC0  uniCDC0,  HANGUL SYLLABLE-CDC0
        0xCDC1, # G+16117  U+CDC1  uniCDC1,  HANGUL SYLLABLE-CDC1
        0xCDC2, # G+16118  U+CDC2  uniCDC2,  HANGUL SYLLABLE-CDC2
        0xCDC3, # G+16119  U+CDC3  uniCDC3,  HANGUL SYLLABLE-CDC3
        0xCDC5, # G+16120  U+CDC5  uniCDC5,  HANGUL SYLLABLE-CDC5
        0xCDC6, # G+16121  U+CDC6  uniCDC6,  HANGUL SYLLABLE-CDC6
        0xCDC7, # G+16122  U+CDC7  uniCDC7,  HANGUL SYLLABLE-CDC7
        0xCDC8, # G+16123  U+CDC8  uniCDC8,  HANGUL SYLLABLE-CDC8
        0xCDC9, # G+16124  U+CDC9  uniCDC9,  HANGUL SYLLABLE-CDC9
        0xCDCA, # G+16125  U+CDCA  uniCDCA,  HANGUL SYLLABLE-CDCA
        0xCDCB, # G+16126  U+CDCB  uniCDCB,  HANGUL SYLLABLE-CDCB
        0xCDCD, # G+16127  U+CDCD  uniCDCD,  HANGUL SYLLABLE-CDCD
        0xCDCE, # G+16128  U+CDCE  uniCDCE,  HANGUL SYLLABLE-CDCE
        0xCDCF, # G+16129  U+CDCF  uniCDCF,  HANGUL SYLLABLE-CDCF
        0xCDD1, # G+16130  U+CDD1  uniCDD1,  HANGUL SYLLABLE-CDD1
        0xCDD2, # G+16131  U+CDD2  uniCDD2,  HANGUL SYLLABLE-CDD2
        0xCDD3, # G+16132  U+CDD3  uniCDD3,  HANGUL SYLLABLE-CDD3
        0xCDD4, # G+16133  U+CDD4  uniCDD4,  HANGUL SYLLABLE-CDD4
        0xCDD5, # G+16134  U+CDD5  uniCDD5,  HANGUL SYLLABLE-CDD5
        0xCDD6, # G+16135  U+CDD6  uniCDD6,  HANGUL SYLLABLE-CDD6
        0xCDD7, # G+16136  U+CDD7  uniCDD7,  HANGUL SYLLABLE-CDD7
        0xCDD8, # G+16137  U+CDD8  uniCDD8,  HANGUL SYLLABLE-CDD8
        0xCDD9, # G+16138  U+CDD9  uniCDD9,  HANGUL SYLLABLE-CDD9
        0xCDDA, # G+16139  U+CDDA  uniCDDA,  HANGUL SYLLABLE-CDDA
        0xCDDB, # G+16140  U+CDDB  uniCDDB,  HANGUL SYLLABLE-CDDB
        0xCDDC, # G+16141  U+CDDC  uniCDDC,  HANGUL SYLLABLE-CDDC
        0xCDDD, # G+16142  U+CDDD  uniCDDD,  HANGUL SYLLABLE-CDDD
        0xCDDE, # G+16143  U+CDDE  uniCDDE,  HANGUL SYLLABLE-CDDE
        0xCDDF, # G+16144  U+CDDF  uniCDDF,  HANGUL SYLLABLE-CDDF
        0xCDE0, # G+16145  U+CDE0  uniCDE0,  HANGUL SYLLABLE-CDE0
        0xCDE1, # G+16146  U+CDE1  uniCDE1,  HANGUL SYLLABLE-CDE1
        0xCDE2, # G+16147  U+CDE2  uniCDE2,  HANGUL SYLLABLE-CDE2
        0xCDE3, # G+16148  U+CDE3  uniCDE3,  HANGUL SYLLABLE-CDE3
        0xCDE4, # G+16149  U+CDE4  uniCDE4,  HANGUL SYLLABLE-CDE4
        0xCDE5, # G+16150  U+CDE5  uniCDE5,  HANGUL SYLLABLE-CDE5
        0xCDE6, # G+16151  U+CDE6  uniCDE6,  HANGUL SYLLABLE-CDE6
        0xCDE7, # G+16152  U+CDE7  uniCDE7,  HANGUL SYLLABLE-CDE7
        0xCDE9, # G+16153  U+CDE9  uniCDE9,  HANGUL SYLLABLE-CDE9
        0xCDEA, # G+16154  U+CDEA  uniCDEA,  HANGUL SYLLABLE-CDEA
        0xCDEB, # G+16155  U+CDEB  uniCDEB,  HANGUL SYLLABLE-CDEB
        0xCDED, # G+16156  U+CDED  uniCDED,  HANGUL SYLLABLE-CDED
        0xCDEE, # G+16157  U+CDEE  uniCDEE,  HANGUL SYLLABLE-CDEE
        0xCDEF, # G+16158  U+CDEF  uniCDEF,  HANGUL SYLLABLE-CDEF
        0xCDF1, # G+16159  U+CDF1  uniCDF1,  HANGUL SYLLABLE-CDF1
        0xCDF2, # G+16160  U+CDF2  uniCDF2,  HANGUL SYLLABLE-CDF2
        0xCDF3, # G+16161  U+CDF3  uniCDF3,  HANGUL SYLLABLE-CDF3
        0xCDF4, # G+16162  U+CDF4  uniCDF4,  HANGUL SYLLABLE-CDF4
        0xCDF5, # G+16163  U+CDF5  uniCDF5,  HANGUL SYLLABLE-CDF5
        0xCDF6, # G+16164  U+CDF6  uniCDF6,  HANGUL SYLLABLE-CDF6
        0xCDF7, # G+16165  U+CDF7  uniCDF7,  HANGUL SYLLABLE-CDF7
        0xCDFA, # G+16166  U+CDFA  uniCDFA,  HANGUL SYLLABLE-CDFA
        0xCDFC, # G+16167  U+CDFC  uniCDFC,  HANGUL SYLLABLE-CDFC
        0xCDFE, # G+16168  U+CDFE  uniCDFE,  HANGUL SYLLABLE-CDFE
        0xCDFF, # G+16169  U+CDFF  uniCDFF,  HANGUL SYLLABLE-CDFF
        0xCE00, # G+16170  U+CE00  uniCE00,  HANGUL SYLLABLE-CE00
        0xCE01, # G+16171  U+CE01  uniCE01,  HANGUL SYLLABLE-CE01
        0xCE02, # G+16172  U+CE02  uniCE02,  HANGUL SYLLABLE-CE02
        0xCE03, # G+16173  U+CE03  uniCE03,  HANGUL SYLLABLE-CE03
        0xCE05, # G+16174  U+CE05  uniCE05,  HANGUL SYLLABLE-CE05
        0xCE06, # G+16175  U+CE06  uniCE06,  HANGUL SYLLABLE-CE06
        0xCE07, # G+16176  U+CE07  uniCE07,  HANGUL SYLLABLE-CE07
        0xCE09, # G+16177  U+CE09  uniCE09,  HANGUL SYLLABLE-CE09
        0xCE0A, # G+16178  U+CE0A  uniCE0A,  HANGUL SYLLABLE-CE0A
        0xCE0B, # G+16179  U+CE0B  uniCE0B,  HANGUL SYLLABLE-CE0B
        0xCE0D, # G+16180  U+CE0D  uniCE0D,  HANGUL SYLLABLE-CE0D
        0xCE0E, # G+16181  U+CE0E  uniCE0E,  HANGUL SYLLABLE-CE0E
        0xCE0F, # G+16182  U+CE0F  uniCE0F,  HANGUL SYLLABLE-CE0F
        0xCE10, # G+16183  U+CE10  uniCE10,  HANGUL SYLLABLE-CE10
        0xCE11, # G+16184  U+CE11  uniCE11,  HANGUL SYLLABLE-CE11
        0xCE12, # G+16185  U+CE12  uniCE12,  HANGUL SYLLABLE-CE12
        0xCE13, # G+16186  U+CE13  uniCE13,  HANGUL SYLLABLE-CE13
        0xCE15, # G+16187  U+CE15  uniCE15,  HANGUL SYLLABLE-CE15
        0xCE16, # G+16188  U+CE16  uniCE16,  HANGUL SYLLABLE-CE16
        0xCE17, # G+16189  U+CE17  uniCE17,  HANGUL SYLLABLE-CE17
        0xCE18, # G+16190  U+CE18  uniCE18,  HANGUL SYLLABLE-CE18
        0xCE1A, # G+16191  U+CE1A  uniCE1A,  HANGUL SYLLABLE-CE1A
        0xCE1B, # G+16192  U+CE1B  uniCE1B,  HANGUL SYLLABLE-CE1B
        0xCE1C, # G+16193  U+CE1C  uniCE1C,  HANGUL SYLLABLE-CE1C
        0xCE1D, # G+16194  U+CE1D  uniCE1D,  HANGUL SYLLABLE-CE1D
        0xCE1E, # G+16195  U+CE1E  uniCE1E,  HANGUL SYLLABLE-CE1E
        0xCE1F, # G+16196  U+CE1F  uniCE1F,  HANGUL SYLLABLE-CE1F
        0xCE22, # G+16197  U+CE22  uniCE22,  HANGUL SYLLABLE-CE22
        0xCE23, # G+16198  U+CE23  uniCE23,  HANGUL SYLLABLE-CE23
        0xCE25, # G+16199  U+CE25  uniCE25,  HANGUL SYLLABLE-CE25
        0xCE26, # G+16200  U+CE26  uniCE26,  HANGUL SYLLABLE-CE26
        0xCE27, # G+16201  U+CE27  uniCE27,  HANGUL SYLLABLE-CE27
        0xCE29, # G+16202  U+CE29  uniCE29,  HANGUL SYLLABLE-CE29
        0xCE2A, # G+16203  U+CE2A  uniCE2A,  HANGUL SYLLABLE-CE2A
        0xCE2B, # G+16204  U+CE2B  uniCE2B,  HANGUL SYLLABLE-CE2B
        0xCE2C, # G+16205  U+CE2C  uniCE2C,  HANGUL SYLLABLE-CE2C
        0xCE2D, # G+16206  U+CE2D  uniCE2D,  HANGUL SYLLABLE-CE2D
        0xCE2E, # G+16207  U+CE2E  uniCE2E,  HANGUL SYLLABLE-CE2E
        0xCE2F, # G+16208  U+CE2F  uniCE2F,  HANGUL SYLLABLE-CE2F
        0xCE32, # G+16209  U+CE32  uniCE32,  HANGUL SYLLABLE-CE32
        0xCE34, # G+16210  U+CE34  uniCE34,  HANGUL SYLLABLE-CE34
        0xCE36, # G+16211  U+CE36  uniCE36,  HANGUL SYLLABLE-CE36
        0xCE37, # G+16212  U+CE37  uniCE37,  HANGUL SYLLABLE-CE37
        0xCE38, # G+16213  U+CE38  uniCE38,  HANGUL SYLLABLE-CE38
        0xCE39, # G+16214  U+CE39  uniCE39,  HANGUL SYLLABLE-CE39
        0xCE3A, # G+16215  U+CE3A  uniCE3A,  HANGUL SYLLABLE-CE3A
        0xCE3B, # G+16216  U+CE3B  uniCE3B,  HANGUL SYLLABLE-CE3B
        0xCE3C, # G+16217  U+CE3C  uniCE3C,  HANGUL SYLLABLE-CE3C
        0xCE3D, # G+16218  U+CE3D  uniCE3D,  HANGUL SYLLABLE-CE3D
        0xCE3E, # G+16219  U+CE3E  uniCE3E,  HANGUL SYLLABLE-CE3E
        0xCE3F, # G+16220  U+CE3F  uniCE3F,  HANGUL SYLLABLE-CE3F
        0xCE40, # G+16221  U+CE40  uniCE40,  HANGUL SYLLABLE-CE40
        0xCE41, # G+16222  U+CE41  uniCE41,  HANGUL SYLLABLE-CE41
        0xCE42, # G+16223  U+CE42  uniCE42,  HANGUL SYLLABLE-CE42
        0xCE43, # G+16224  U+CE43  uniCE43,  HANGUL SYLLABLE-CE43
        0xCE44, # G+16225  U+CE44  uniCE44,  HANGUL SYLLABLE-CE44
        0xCE45, # G+16226  U+CE45  uniCE45,  HANGUL SYLLABLE-CE45
        0xCE46, # G+16227  U+CE46  uniCE46,  HANGUL SYLLABLE-CE46
        0xCE47, # G+16228  U+CE47  uniCE47,  HANGUL SYLLABLE-CE47
        0xCE48, # G+16229  U+CE48  uniCE48,  HANGUL SYLLABLE-CE48
        0xCE49, # G+16230  U+CE49  uniCE49,  HANGUL SYLLABLE-CE49
        0xCE4A, # G+16231  U+CE4A  uniCE4A,  HANGUL SYLLABLE-CE4A
        0xCE4B, # G+16232  U+CE4B  uniCE4B,  HANGUL SYLLABLE-CE4B
        0xCE4C, # G+16233  U+CE4C  uniCE4C,  HANGUL SYLLABLE-CE4C
        0xCE4D, # G+16234  U+CE4D  uniCE4D,  HANGUL SYLLABLE-CE4D
        0xCE4E, # G+16235  U+CE4E  uniCE4E,  HANGUL SYLLABLE-CE4E
        0xCE4F, # G+16236  U+CE4F  uniCE4F,  HANGUL SYLLABLE-CE4F
        0xCE50, # G+16237  U+CE50  uniCE50,  HANGUL SYLLABLE-CE50
        0xCE51, # G+16238  U+CE51  uniCE51,  HANGUL SYLLABLE-CE51
        0xCE52, # G+16239  U+CE52  uniCE52,  HANGUL SYLLABLE-CE52
        0xCE53, # G+16240  U+CE53  uniCE53,  HANGUL SYLLABLE-CE53
        0xCE54, # G+16241  U+CE54  uniCE54,  HANGUL SYLLABLE-CE54
        0xCE55, # G+16242  U+CE55  uniCE55,  HANGUL SYLLABLE-CE55
        0xCE56, # G+16243  U+CE56  uniCE56,  HANGUL SYLLABLE-CE56
        0xCE57, # G+16244  U+CE57  uniCE57,  HANGUL SYLLABLE-CE57
        0xCE5A, # G+16245  U+CE5A  uniCE5A,  HANGUL SYLLABLE-CE5A
        0xCE5B, # G+16246  U+CE5B  uniCE5B,  HANGUL SYLLABLE-CE5B
        0xCE5D, # G+16247  U+CE5D  uniCE5D,  HANGUL SYLLABLE-CE5D
        0xCE5E, # G+16248  U+CE5E  uniCE5E,  HANGUL SYLLABLE-CE5E
        0xCE62, # G+16249  U+CE62  uniCE62,  HANGUL SYLLABLE-CE62
        0xCE63, # G+16250  U+CE63  uniCE63,  HANGUL SYLLABLE-CE63
        0xCE64, # G+16251  U+CE64  uniCE64,  HANGUL SYLLABLE-CE64
        0xCE65, # G+16252  U+CE65  uniCE65,  HANGUL SYLLABLE-CE65
        0xCE66, # G+16253  U+CE66  uniCE66,  HANGUL SYLLABLE-CE66
        0xCE67, # G+16254  U+CE67  uniCE67,  HANGUL SYLLABLE-CE67
        0xCE6A, # G+16255  U+CE6A  uniCE6A,  HANGUL SYLLABLE-CE6A
        0xCE6C, # G+16256  U+CE6C  uniCE6C,  HANGUL SYLLABLE-CE6C
        0xCE6E, # G+16257  U+CE6E  uniCE6E,  HANGUL SYLLABLE-CE6E
        0xCE6F, # G+16258  U+CE6F  uniCE6F,  HANGUL SYLLABLE-CE6F
        0xCE70, # G+16259  U+CE70  uniCE70,  HANGUL SYLLABLE-CE70
        0xCE71, # G+16260  U+CE71  uniCE71,  HANGUL SYLLABLE-CE71
        0xCE72, # G+16261  U+CE72  uniCE72,  HANGUL SYLLABLE-CE72
        0xCE73, # G+16262  U+CE73  uniCE73,  HANGUL SYLLABLE-CE73
        0xCE76, # G+16263  U+CE76  uniCE76,  HANGUL SYLLABLE-CE76
        0xCE77, # G+16264  U+CE77  uniCE77,  HANGUL SYLLABLE-CE77
        0xCE79, # G+16265  U+CE79  uniCE79,  HANGUL SYLLABLE-CE79
        0xCE7A, # G+16266  U+CE7A  uniCE7A,  HANGUL SYLLABLE-CE7A
        0xCE7B, # G+16267  U+CE7B  uniCE7B,  HANGUL SYLLABLE-CE7B
        0xCE7D, # G+16268  U+CE7D  uniCE7D,  HANGUL SYLLABLE-CE7D
        0xCE7E, # G+16269  U+CE7E  uniCE7E,  HANGUL SYLLABLE-CE7E
        0xCE7F, # G+16270  U+CE7F  uniCE7F,  HANGUL SYLLABLE-CE7F
        0xCE80, # G+16271  U+CE80  uniCE80,  HANGUL SYLLABLE-CE80
        0xCE81, # G+16272  U+CE81  uniCE81,  HANGUL SYLLABLE-CE81
        0xCE82, # G+16273  U+CE82  uniCE82,  HANGUL SYLLABLE-CE82
        0xCE83, # G+16274  U+CE83  uniCE83,  HANGUL SYLLABLE-CE83
        0xCE86, # G+16275  U+CE86  uniCE86,  HANGUL SYLLABLE-CE86
        0xCE88, # G+16276  U+CE88  uniCE88,  HANGUL SYLLABLE-CE88
        0xCE8A, # G+16277  U+CE8A  uniCE8A,  HANGUL SYLLABLE-CE8A
        0xCE8B, # G+16278  U+CE8B  uniCE8B,  HANGUL SYLLABLE-CE8B
        0xCE8C, # G+16279  U+CE8C  uniCE8C,  HANGUL SYLLABLE-CE8C
        0xCE8D, # G+16280  U+CE8D  uniCE8D,  HANGUL SYLLABLE-CE8D
        0xCE8E, # G+16281  U+CE8E  uniCE8E,  HANGUL SYLLABLE-CE8E
        0xCE8F, # G+16282  U+CE8F  uniCE8F,  HANGUL SYLLABLE-CE8F
        0xCE92, # G+16283  U+CE92  uniCE92,  HANGUL SYLLABLE-CE92
        0xCE93, # G+16284  U+CE93  uniCE93,  HANGUL SYLLABLE-CE93
        0xCE95, # G+16285  U+CE95  uniCE95,  HANGUL SYLLABLE-CE95
        0xCE96, # G+16286  U+CE96  uniCE96,  HANGUL SYLLABLE-CE96
        0xCE97, # G+16287  U+CE97  uniCE97,  HANGUL SYLLABLE-CE97
        0xCE99, # G+16288  U+CE99  uniCE99,  HANGUL SYLLABLE-CE99
        0xCE9A, # G+16289  U+CE9A  uniCE9A,  HANGUL SYLLABLE-CE9A
        0xCE9B, # G+16290  U+CE9B  uniCE9B,  HANGUL SYLLABLE-CE9B
        0xCE9C, # G+16291  U+CE9C  uniCE9C,  HANGUL SYLLABLE-CE9C
        0xCE9D, # G+16292  U+CE9D  uniCE9D,  HANGUL SYLLABLE-CE9D
        0xCE9E, # G+16293  U+CE9E  uniCE9E,  HANGUL SYLLABLE-CE9E
        0xCE9F, # G+16294  U+CE9F  uniCE9F,  HANGUL SYLLABLE-CE9F
        0xCEA2, # G+16295  U+CEA2  uniCEA2,  HANGUL SYLLABLE-CEA2
        0xCEA6, # G+16296  U+CEA6  uniCEA6,  HANGUL SYLLABLE-CEA6
        0xCEA7, # G+16297  U+CEA7  uniCEA7,  HANGUL SYLLABLE-CEA7
        0xCEA8, # G+16298  U+CEA8  uniCEA8,  HANGUL SYLLABLE-CEA8
        0xCEA9, # G+16299  U+CEA9  uniCEA9,  HANGUL SYLLABLE-CEA9
        0xCEAA, # G+16300  U+CEAA  uniCEAA,  HANGUL SYLLABLE-CEAA
        0xCEAB, # G+16301  U+CEAB  uniCEAB,  HANGUL SYLLABLE-CEAB
        0xCEAE, # G+16302  U+CEAE  uniCEAE,  HANGUL SYLLABLE-CEAE
        0xCEAF, # G+16303  U+CEAF  uniCEAF,  HANGUL SYLLABLE-CEAF
        0xCEB0, # G+16304  U+CEB0  uniCEB0,  HANGUL SYLLABLE-CEB0
        0xCEB1, # G+16305  U+CEB1  uniCEB1,  HANGUL SYLLABLE-CEB1
        0xCEB2, # G+16306  U+CEB2  uniCEB2,  HANGUL SYLLABLE-CEB2
        0xCEB3, # G+16307  U+CEB3  uniCEB3,  HANGUL SYLLABLE-CEB3
        0xCEB4, # G+16308  U+CEB4  uniCEB4,  HANGUL SYLLABLE-CEB4
        0xCEB5, # G+16309  U+CEB5  uniCEB5,  HANGUL SYLLABLE-CEB5
        0xCEB6, # G+16310  U+CEB6  uniCEB6,  HANGUL SYLLABLE-CEB6
        0xCEB7, # G+16311  U+CEB7  uniCEB7,  HANGUL SYLLABLE-CEB7
        0xCEB8, # G+16312  U+CEB8  uniCEB8,  HANGUL SYLLABLE-CEB8
        0xCEB9, # G+16313  U+CEB9  uniCEB9,  HANGUL SYLLABLE-CEB9
        0xCEBA, # G+16314  U+CEBA  uniCEBA,  HANGUL SYLLABLE-CEBA
        0xCEBB, # G+16315  U+CEBB  uniCEBB,  HANGUL SYLLABLE-CEBB
        0xCEBC, # G+16316  U+CEBC  uniCEBC,  HANGUL SYLLABLE-CEBC
        0xCEBD, # G+16317  U+CEBD  uniCEBD,  HANGUL SYLLABLE-CEBD
        0xCEBE, # G+16318  U+CEBE  uniCEBE,  HANGUL SYLLABLE-CEBE
        0xCEBF, # G+16319  U+CEBF  uniCEBF,  HANGUL SYLLABLE-CEBF
        0xCEC0, # G+16320  U+CEC0  uniCEC0,  HANGUL SYLLABLE-CEC0
        0xCEC2, # G+16321  U+CEC2  uniCEC2,  HANGUL SYLLABLE-CEC2
        0xCEC3, # G+16322  U+CEC3  uniCEC3,  HANGUL SYLLABLE-CEC3
        0xCEC4, # G+16323  U+CEC4  uniCEC4,  HANGUL SYLLABLE-CEC4
        0xCEC5, # G+16324  U+CEC5  uniCEC5,  HANGUL SYLLABLE-CEC5
        0xCEC6, # G+16325  U+CEC6  uniCEC6,  HANGUL SYLLABLE-CEC6
        0xCEC7, # G+16326  U+CEC7  uniCEC7,  HANGUL SYLLABLE-CEC7
        0xCEC8, # G+16327  U+CEC8  uniCEC8,  HANGUL SYLLABLE-CEC8
        0xCEC9, # G+16328  U+CEC9  uniCEC9,  HANGUL SYLLABLE-CEC9
        0xCECA, # G+16329  U+CECA  uniCECA,  HANGUL SYLLABLE-CECA
        0xCECB, # G+16330  U+CECB  uniCECB,  HANGUL SYLLABLE-CECB
        0xCECC, # G+16331  U+CECC  uniCECC,  HANGUL SYLLABLE-CECC
        0xCECD, # G+16332  U+CECD  uniCECD,  HANGUL SYLLABLE-CECD
        0xCECE, # G+16333  U+CECE  uniCECE,  HANGUL SYLLABLE-CECE
        0xCECF, # G+16334  U+CECF  uniCECF,  HANGUL SYLLABLE-CECF
        0xCED0, # G+16335  U+CED0  uniCED0,  HANGUL SYLLABLE-CED0
        0xCED1, # G+16336  U+CED1  uniCED1,  HANGUL SYLLABLE-CED1
        0xCED2, # G+16337  U+CED2  uniCED2,  HANGUL SYLLABLE-CED2
        0xCED3, # G+16338  U+CED3  uniCED3,  HANGUL SYLLABLE-CED3
        0xCED4, # G+16339  U+CED4  uniCED4,  HANGUL SYLLABLE-CED4
        0xCED5, # G+16340  U+CED5  uniCED5,  HANGUL SYLLABLE-CED5
        0xCED6, # G+16341  U+CED6  uniCED6,  HANGUL SYLLABLE-CED6
        0xCED7, # G+16342  U+CED7  uniCED7,  HANGUL SYLLABLE-CED7
        0xCED8, # G+16343  U+CED8  uniCED8,  HANGUL SYLLABLE-CED8
        0xCED9, # G+16344  U+CED9  uniCED9,  HANGUL SYLLABLE-CED9
        0xCEDA, # G+16345  U+CEDA  uniCEDA,  HANGUL SYLLABLE-CEDA
        0xCEDB, # G+16346  U+CEDB  uniCEDB,  HANGUL SYLLABLE-CEDB
        0xCEDC, # G+16347  U+CEDC  uniCEDC,  HANGUL SYLLABLE-CEDC
        0xCEDD, # G+16348  U+CEDD  uniCEDD,  HANGUL SYLLABLE-CEDD
        0xCEDE, # G+16349  U+CEDE  uniCEDE,  HANGUL SYLLABLE-CEDE
        0xCEDF, # G+16350  U+CEDF  uniCEDF,  HANGUL SYLLABLE-CEDF
        0xCEE0, # G+16351  U+CEE0  uniCEE0,  HANGUL SYLLABLE-CEE0
        0xCEE1, # G+16352  U+CEE1  uniCEE1,  HANGUL SYLLABLE-CEE1
        0xCEE2, # G+16353  U+CEE2  uniCEE2,  HANGUL SYLLABLE-CEE2
        0xCEE3, # G+16354  U+CEE3  uniCEE3,  HANGUL SYLLABLE-CEE3
        0xCEE6, # G+16355  U+CEE6  uniCEE6,  HANGUL SYLLABLE-CEE6
        0xCEE7, # G+16356  U+CEE7  uniCEE7,  HANGUL SYLLABLE-CEE7
        0xCEE9, # G+16357  U+CEE9  uniCEE9,  HANGUL SYLLABLE-CEE9
        0xCEEA, # G+16358  U+CEEA  uniCEEA,  HANGUL SYLLABLE-CEEA
        0xCEED, # G+16359  U+CEED  uniCEED,  HANGUL SYLLABLE-CEED
        0xCEEE, # G+16360  U+CEEE  uniCEEE,  HANGUL SYLLABLE-CEEE
        0xCEEF, # G+16361  U+CEEF  uniCEEF,  HANGUL SYLLABLE-CEEF
        0xCEF0, # G+16362  U+CEF0  uniCEF0,  HANGUL SYLLABLE-CEF0
        0xCEF1, # G+16363  U+CEF1  uniCEF1,  HANGUL SYLLABLE-CEF1
        0xCEF2, # G+16364  U+CEF2  uniCEF2,  HANGUL SYLLABLE-CEF2
        0xCEF3, # G+16365  U+CEF3  uniCEF3,  HANGUL SYLLABLE-CEF3
        0xCEF6, # G+16366  U+CEF6  uniCEF6,  HANGUL SYLLABLE-CEF6
        0xCEFA, # G+16367  U+CEFA  uniCEFA,  HANGUL SYLLABLE-CEFA
        0xCEFB, # G+16368  U+CEFB  uniCEFB,  HANGUL SYLLABLE-CEFB
        0xCEFC, # G+16369  U+CEFC  uniCEFC,  HANGUL SYLLABLE-CEFC
        0xCEFD, # G+16370  U+CEFD  uniCEFD,  HANGUL SYLLABLE-CEFD
        0xCEFE, # G+16371  U+CEFE  uniCEFE,  HANGUL SYLLABLE-CEFE
        0xCEFF, # G+16372  U+CEFF  uniCEFF,  HANGUL SYLLABLE-CEFF
        0xCF02, # G+16373  U+CF02  uniCF02,  HANGUL SYLLABLE-CF02
        0xCF03, # G+16374  U+CF03  uniCF03,  HANGUL SYLLABLE-CF03
        0xCF05, # G+16375  U+CF05  uniCF05,  HANGUL SYLLABLE-CF05
        0xCF06, # G+16376  U+CF06  uniCF06,  HANGUL SYLLABLE-CF06
        0xCF07, # G+16377  U+CF07  uniCF07,  HANGUL SYLLABLE-CF07
        0xCF09, # G+16378  U+CF09  uniCF09,  HANGUL SYLLABLE-CF09
        0xCF0A, # G+16379  U+CF0A  uniCF0A,  HANGUL SYLLABLE-CF0A
        0xCF0B, # G+16380  U+CF0B  uniCF0B,  HANGUL SYLLABLE-CF0B
        0xCF0C, # G+16381  U+CF0C  uniCF0C,  HANGUL SYLLABLE-CF0C
        0xCF0D, # G+16382  U+CF0D  uniCF0D,  HANGUL SYLLABLE-CF0D
        0xCF0E, # G+16383  U+CF0E  uniCF0E,  HANGUL SYLLABLE-CF0E
        0xCF0F, # G+16384  U+CF0F  uniCF0F,  HANGUL SYLLABLE-CF0F
        0xCF12, # G+16385  U+CF12  uniCF12,  HANGUL SYLLABLE-CF12
        0xCF14, # G+16386  U+CF14  uniCF14,  HANGUL SYLLABLE-CF14
        0xCF16, # G+16387  U+CF16  uniCF16,  HANGUL SYLLABLE-CF16
        0xCF17, # G+16388  U+CF17  uniCF17,  HANGUL SYLLABLE-CF17
        0xCF18, # G+16389  U+CF18  uniCF18,  HANGUL SYLLABLE-CF18
        0xCF19, # G+16390  U+CF19  uniCF19,  HANGUL SYLLABLE-CF19
        0xCF1A, # G+16391  U+CF1A  uniCF1A,  HANGUL SYLLABLE-CF1A
        0xCF1B, # G+16392  U+CF1B  uniCF1B,  HANGUL SYLLABLE-CF1B
        0xCF1D, # G+16393  U+CF1D  uniCF1D,  HANGUL SYLLABLE-CF1D
        0xCF1E, # G+16394  U+CF1E  uniCF1E,  HANGUL SYLLABLE-CF1E
        0xCF1F, # G+16395  U+CF1F  uniCF1F,  HANGUL SYLLABLE-CF1F
        0xCF21, # G+16396  U+CF21  uniCF21,  HANGUL SYLLABLE-CF21
        0xCF22, # G+16397  U+CF22  uniCF22,  HANGUL SYLLABLE-CF22
        0xCF23, # G+16398  U+CF23  uniCF23,  HANGUL SYLLABLE-CF23
        0xCF25, # G+16399  U+CF25  uniCF25,  HANGUL SYLLABLE-CF25
        0xCF26, # G+16400  U+CF26  uniCF26,  HANGUL SYLLABLE-CF26
        0xCF27, # G+16401  U+CF27  uniCF27,  HANGUL SYLLABLE-CF27
        0xCF28, # G+16402  U+CF28  uniCF28,  HANGUL SYLLABLE-CF28
        0xCF29, # G+16403  U+CF29  uniCF29,  HANGUL SYLLABLE-CF29
        0xCF2A, # G+16404  U+CF2A  uniCF2A,  HANGUL SYLLABLE-CF2A
        0xCF2B, # G+16405  U+CF2B  uniCF2B,  HANGUL SYLLABLE-CF2B
        0xCF2E, # G+16406  U+CF2E  uniCF2E,  HANGUL SYLLABLE-CF2E
        0xCF32, # G+16407  U+CF32  uniCF32,  HANGUL SYLLABLE-CF32
        0xCF33, # G+16408  U+CF33  uniCF33,  HANGUL SYLLABLE-CF33
        0xCF34, # G+16409  U+CF34  uniCF34,  HANGUL SYLLABLE-CF34
        0xCF35, # G+16410  U+CF35  uniCF35,  HANGUL SYLLABLE-CF35
        0xCF36, # G+16411  U+CF36  uniCF36,  HANGUL SYLLABLE-CF36
        0xCF37, # G+16412  U+CF37  uniCF37,  HANGUL SYLLABLE-CF37
        0xCF39, # G+16413  U+CF39  uniCF39,  HANGUL SYLLABLE-CF39
        0xCF3A, # G+16414  U+CF3A  uniCF3A,  HANGUL SYLLABLE-CF3A
        0xCF3B, # G+16415  U+CF3B  uniCF3B,  HANGUL SYLLABLE-CF3B
        0xCF3C, # G+16416  U+CF3C  uniCF3C,  HANGUL SYLLABLE-CF3C
        0xCF3D, # G+16417  U+CF3D  uniCF3D,  HANGUL SYLLABLE-CF3D
        0xCF3E, # G+16418  U+CF3E  uniCF3E,  HANGUL SYLLABLE-CF3E
        0xCF3F, # G+16419  U+CF3F  uniCF3F,  HANGUL SYLLABLE-CF3F
        0xCF40, # G+16420  U+CF40  uniCF40,  HANGUL SYLLABLE-CF40
        0xCF41, # G+16421  U+CF41  uniCF41,  HANGUL SYLLABLE-CF41
        0xCF42, # G+16422  U+CF42  uniCF42,  HANGUL SYLLABLE-CF42
        0xCF43, # G+16423  U+CF43  uniCF43,  HANGUL SYLLABLE-CF43
        0xCF44, # G+16424  U+CF44  uniCF44,  HANGUL SYLLABLE-CF44
        0xCF45, # G+16425  U+CF45  uniCF45,  HANGUL SYLLABLE-CF45
        0xCF46, # G+16426  U+CF46  uniCF46,  HANGUL SYLLABLE-CF46
        0xCF47, # G+16427  U+CF47  uniCF47,  HANGUL SYLLABLE-CF47
        0xCF48, # G+16428  U+CF48  uniCF48,  HANGUL SYLLABLE-CF48
        0xCF49, # G+16429  U+CF49  uniCF49,  HANGUL SYLLABLE-CF49
        0xCF4A, # G+16430  U+CF4A  uniCF4A,  HANGUL SYLLABLE-CF4A
        0xCF4B, # G+16431  U+CF4B  uniCF4B,  HANGUL SYLLABLE-CF4B
        0xCF4C, # G+16432  U+CF4C  uniCF4C,  HANGUL SYLLABLE-CF4C
        0xCF4D, # G+16433  U+CF4D  uniCF4D,  HANGUL SYLLABLE-CF4D
        0xCF4E, # G+16434  U+CF4E  uniCF4E,  HANGUL SYLLABLE-CF4E
        0xCF4F, # G+16435  U+CF4F  uniCF4F,  HANGUL SYLLABLE-CF4F
        0xCF50, # G+16436  U+CF50  uniCF50,  HANGUL SYLLABLE-CF50
        0xCF51, # G+16437  U+CF51  uniCF51,  HANGUL SYLLABLE-CF51
        0xCF52, # G+16438  U+CF52  uniCF52,  HANGUL SYLLABLE-CF52
        0xCF53, # G+16439  U+CF53  uniCF53,  HANGUL SYLLABLE-CF53
        0xCF56, # G+16440  U+CF56  uniCF56,  HANGUL SYLLABLE-CF56
        0xCF57, # G+16441  U+CF57  uniCF57,  HANGUL SYLLABLE-CF57
        0xCF59, # G+16442  U+CF59  uniCF59,  HANGUL SYLLABLE-CF59
        0xCF5A, # G+16443  U+CF5A  uniCF5A,  HANGUL SYLLABLE-CF5A
        0xCF5B, # G+16444  U+CF5B  uniCF5B,  HANGUL SYLLABLE-CF5B
        0xCF5D, # G+16445  U+CF5D  uniCF5D,  HANGUL SYLLABLE-CF5D
        0xCF5E, # G+16446  U+CF5E  uniCF5E,  HANGUL SYLLABLE-CF5E
        0xCF5F, # G+16447  U+CF5F  uniCF5F,  HANGUL SYLLABLE-CF5F
        0xCF60, # G+16448  U+CF60  uniCF60,  HANGUL SYLLABLE-CF60
        0xCF61, # G+16449  U+CF61  uniCF61,  HANGUL SYLLABLE-CF61
        0xCF62, # G+16450  U+CF62  uniCF62,  HANGUL SYLLABLE-CF62
        0xCF63, # G+16451  U+CF63  uniCF63,  HANGUL SYLLABLE-CF63
        0xCF66, # G+16452  U+CF66  uniCF66,  HANGUL SYLLABLE-CF66
        0xCF68, # G+16453  U+CF68  uniCF68,  HANGUL SYLLABLE-CF68
        0xCF6A, # G+16454  U+CF6A  uniCF6A,  HANGUL SYLLABLE-CF6A
        0xCF6B, # G+16455  U+CF6B  uniCF6B,  HANGUL SYLLABLE-CF6B
        0xCF6C, # G+16456  U+CF6C  uniCF6C,  HANGUL SYLLABLE-CF6C
        0xCF6D, # G+16457  U+CF6D  uniCF6D,  HANGUL SYLLABLE-CF6D
        0xCF6E, # G+16458  U+CF6E  uniCF6E,  HANGUL SYLLABLE-CF6E
        0xCF6F, # G+16459  U+CF6F  uniCF6F,  HANGUL SYLLABLE-CF6F
        0xCF72, # G+16460  U+CF72  uniCF72,  HANGUL SYLLABLE-CF72
        0xCF73, # G+16461  U+CF73  uniCF73,  HANGUL SYLLABLE-CF73
        0xCF75, # G+16462  U+CF75  uniCF75,  HANGUL SYLLABLE-CF75
        0xCF76, # G+16463  U+CF76  uniCF76,  HANGUL SYLLABLE-CF76
        0xCF77, # G+16464  U+CF77  uniCF77,  HANGUL SYLLABLE-CF77
        0xCF79, # G+16465  U+CF79  uniCF79,  HANGUL SYLLABLE-CF79
        0xCF7A, # G+16466  U+CF7A  uniCF7A,  HANGUL SYLLABLE-CF7A
        0xCF7B, # G+16467  U+CF7B  uniCF7B,  HANGUL SYLLABLE-CF7B
        0xCF7C, # G+16468  U+CF7C  uniCF7C,  HANGUL SYLLABLE-CF7C
        0xCF7D, # G+16469  U+CF7D  uniCF7D,  HANGUL SYLLABLE-CF7D
        0xCF7E, # G+16470  U+CF7E  uniCF7E,  HANGUL SYLLABLE-CF7E
        0xCF7F, # G+16471  U+CF7F  uniCF7F,  HANGUL SYLLABLE-CF7F
        0xCF81, # G+16472  U+CF81  uniCF81,  HANGUL SYLLABLE-CF81
        0xCF82, # G+16473  U+CF82  uniCF82,  HANGUL SYLLABLE-CF82
        0xCF83, # G+16474  U+CF83  uniCF83,  HANGUL SYLLABLE-CF83
        0xCF84, # G+16475  U+CF84  uniCF84,  HANGUL SYLLABLE-CF84
        0xCF86, # G+16476  U+CF86  uniCF86,  HANGUL SYLLABLE-CF86
        0xCF87, # G+16477  U+CF87  uniCF87,  HANGUL SYLLABLE-CF87
        0xCF88, # G+16478  U+CF88  uniCF88,  HANGUL SYLLABLE-CF88
        0xCF89, # G+16479  U+CF89  uniCF89,  HANGUL SYLLABLE-CF89
        0xCF8A, # G+16480  U+CF8A  uniCF8A,  HANGUL SYLLABLE-CF8A
        0xCF8B, # G+16481  U+CF8B  uniCF8B,  HANGUL SYLLABLE-CF8B
        0xCF8D, # G+16482  U+CF8D  uniCF8D,  HANGUL SYLLABLE-CF8D
        0xCF8E, # G+16483  U+CF8E  uniCF8E,  HANGUL SYLLABLE-CF8E
        0xCF8F, # G+16484  U+CF8F  uniCF8F,  HANGUL SYLLABLE-CF8F
        0xCF90, # G+16485  U+CF90  uniCF90,  HANGUL SYLLABLE-CF90
        0xCF91, # G+16486  U+CF91  uniCF91,  HANGUL SYLLABLE-CF91
        0xCF92, # G+16487  U+CF92  uniCF92,  HANGUL SYLLABLE-CF92
        0xCF93, # G+16488  U+CF93  uniCF93,  HANGUL SYLLABLE-CF93
        0xCF94, # G+16489  U+CF94  uniCF94,  HANGUL SYLLABLE-CF94
        0xCF95, # G+16490  U+CF95  uniCF95,  HANGUL SYLLABLE-CF95
        0xCF96, # G+16491  U+CF96  uniCF96,  HANGUL SYLLABLE-CF96
        0xCF97, # G+16492  U+CF97  uniCF97,  HANGUL SYLLABLE-CF97
        0xCF98, # G+16493  U+CF98  uniCF98,  HANGUL SYLLABLE-CF98
        0xCF99, # G+16494  U+CF99  uniCF99,  HANGUL SYLLABLE-CF99
        0xCF9A, # G+16495  U+CF9A  uniCF9A,  HANGUL SYLLABLE-CF9A
        0xCF9B, # G+16496  U+CF9B  uniCF9B,  HANGUL SYLLABLE-CF9B
        0xCF9C, # G+16497  U+CF9C  uniCF9C,  HANGUL SYLLABLE-CF9C
        0xCF9D, # G+16498  U+CF9D  uniCF9D,  HANGUL SYLLABLE-CF9D
        0xCF9E, # G+16499  U+CF9E  uniCF9E,  HANGUL SYLLABLE-CF9E
        0xCF9F, # G+16500  U+CF9F  uniCF9F,  HANGUL SYLLABLE-CF9F
        0xCFA0, # G+16501  U+CFA0  uniCFA0,  HANGUL SYLLABLE-CFA0
        0xCFA2, # G+16502  U+CFA2  uniCFA2,  HANGUL SYLLABLE-CFA2
        0xCFA3, # G+16503  U+CFA3  uniCFA3,  HANGUL SYLLABLE-CFA3
        0xCFA4, # G+16504  U+CFA4  uniCFA4,  HANGUL SYLLABLE-CFA4
        0xCFA5, # G+16505  U+CFA5  uniCFA5,  HANGUL SYLLABLE-CFA5
        0xCFA6, # G+16506  U+CFA6  uniCFA6,  HANGUL SYLLABLE-CFA6
        0xCFA7, # G+16507  U+CFA7  uniCFA7,  HANGUL SYLLABLE-CFA7
        0xCFA9, # G+16508  U+CFA9  uniCFA9,  HANGUL SYLLABLE-CFA9
        0xCFAA, # G+16509  U+CFAA  uniCFAA,  HANGUL SYLLABLE-CFAA
        0xCFAB, # G+16510  U+CFAB  uniCFAB,  HANGUL SYLLABLE-CFAB
        0xCFAC, # G+16511  U+CFAC  uniCFAC,  HANGUL SYLLABLE-CFAC
        0xCFAD, # G+16512  U+CFAD  uniCFAD,  HANGUL SYLLABLE-CFAD
        0xCFAE, # G+16513  U+CFAE  uniCFAE,  HANGUL SYLLABLE-CFAE
        0xCFAF, # G+16514  U+CFAF  uniCFAF,  HANGUL SYLLABLE-CFAF
        0xCFB1, # G+16515  U+CFB1  uniCFB1,  HANGUL SYLLABLE-CFB1
        0xCFB2, # G+16516  U+CFB2  uniCFB2,  HANGUL SYLLABLE-CFB2
        0xCFB3, # G+16517  U+CFB3  uniCFB3,  HANGUL SYLLABLE-CFB3
        0xCFB4, # G+16518  U+CFB4  uniCFB4,  HANGUL SYLLABLE-CFB4
        0xCFB5, # G+16519  U+CFB5  uniCFB5,  HANGUL SYLLABLE-CFB5
        0xCFB6, # G+16520  U+CFB6  uniCFB6,  HANGUL SYLLABLE-CFB6
        0xCFB7, # G+16521  U+CFB7  uniCFB7,  HANGUL SYLLABLE-CFB7
        0xCFB8, # G+16522  U+CFB8  uniCFB8,  HANGUL SYLLABLE-CFB8
        0xCFB9, # G+16523  U+CFB9  uniCFB9,  HANGUL SYLLABLE-CFB9
        0xCFBA, # G+16524  U+CFBA  uniCFBA,  HANGUL SYLLABLE-CFBA
        0xCFBB, # G+16525  U+CFBB  uniCFBB,  HANGUL SYLLABLE-CFBB
        0xCFBC, # G+16526  U+CFBC  uniCFBC,  HANGUL SYLLABLE-CFBC
        0xCFBD, # G+16527  U+CFBD  uniCFBD,  HANGUL SYLLABLE-CFBD
        0xCFBE, # G+16528  U+CFBE  uniCFBE,  HANGUL SYLLABLE-CFBE
        0xCFBF, # G+16529  U+CFBF  uniCFBF,  HANGUL SYLLABLE-CFBF
        0xCFC0, # G+16530  U+CFC0  uniCFC0,  HANGUL SYLLABLE-CFC0
        0xCFC1, # G+16531  U+CFC1  uniCFC1,  HANGUL SYLLABLE-CFC1
        0xCFC2, # G+16532  U+CFC2  uniCFC2,  HANGUL SYLLABLE-CFC2
        0xCFC3, # G+16533  U+CFC3  uniCFC3,  HANGUL SYLLABLE-CFC3
        0xCFC5, # G+16534  U+CFC5  uniCFC5,  HANGUL SYLLABLE-CFC5
        0xCFC6, # G+16535  U+CFC6  uniCFC6,  HANGUL SYLLABLE-CFC6
        0xCFC7, # G+16536  U+CFC7  uniCFC7,  HANGUL SYLLABLE-CFC7
        0xCFC8, # G+16537  U+CFC8  uniCFC8,  HANGUL SYLLABLE-CFC8
        0xCFC9, # G+16538  U+CFC9  uniCFC9,  HANGUL SYLLABLE-CFC9
        0xCFCA, # G+16539  U+CFCA  uniCFCA,  HANGUL SYLLABLE-CFCA
        0xCFCB, # G+16540  U+CFCB  uniCFCB,  HANGUL SYLLABLE-CFCB
        0xCFCC, # G+16541  U+CFCC  uniCFCC,  HANGUL SYLLABLE-CFCC
        0xCFCD, # G+16542  U+CFCD  uniCFCD,  HANGUL SYLLABLE-CFCD
        0xCFCE, # G+16543  U+CFCE  uniCFCE,  HANGUL SYLLABLE-CFCE
        0xCFCF, # G+16544  U+CFCF  uniCFCF,  HANGUL SYLLABLE-CFCF
        0xCFD0, # G+16545  U+CFD0  uniCFD0,  HANGUL SYLLABLE-CFD0
        0xCFD1, # G+16546  U+CFD1  uniCFD1,  HANGUL SYLLABLE-CFD1
        0xCFD2, # G+16547  U+CFD2  uniCFD2,  HANGUL SYLLABLE-CFD2
        0xCFD3, # G+16548  U+CFD3  uniCFD3,  HANGUL SYLLABLE-CFD3
        0xCFD4, # G+16549  U+CFD4  uniCFD4,  HANGUL SYLLABLE-CFD4
        0xCFD5, # G+16550  U+CFD5  uniCFD5,  HANGUL SYLLABLE-CFD5
        0xCFD6, # G+16551  U+CFD6  uniCFD6,  HANGUL SYLLABLE-CFD6
        0xCFD7, # G+16552  U+CFD7  uniCFD7,  HANGUL SYLLABLE-CFD7
        0xCFD8, # G+16553  U+CFD8  uniCFD8,  HANGUL SYLLABLE-CFD8
        0xCFD9, # G+16554  U+CFD9  uniCFD9,  HANGUL SYLLABLE-CFD9
        0xCFDA, # G+16555  U+CFDA  uniCFDA,  HANGUL SYLLABLE-CFDA
        0xCFDB, # G+16556  U+CFDB  uniCFDB,  HANGUL SYLLABLE-CFDB
        0xCFDC, # G+16557  U+CFDC  uniCFDC,  HANGUL SYLLABLE-CFDC
        0xCFDD, # G+16558  U+CFDD  uniCFDD,  HANGUL SYLLABLE-CFDD
        0xCFDE, # G+16559  U+CFDE  uniCFDE,  HANGUL SYLLABLE-CFDE
        0xCFDF, # G+16560  U+CFDF  uniCFDF,  HANGUL SYLLABLE-CFDF
        0xCFE2, # G+16561  U+CFE2  uniCFE2,  HANGUL SYLLABLE-CFE2
        0xCFE3, # G+16562  U+CFE3  uniCFE3,  HANGUL SYLLABLE-CFE3
        0xCFE5, # G+16563  U+CFE5  uniCFE5,  HANGUL SYLLABLE-CFE5
        0xCFE6, # G+16564  U+CFE6  uniCFE6,  HANGUL SYLLABLE-CFE6
        0xCFE7, # G+16565  U+CFE7  uniCFE7,  HANGUL SYLLABLE-CFE7
        0xCFE9, # G+16566  U+CFE9  uniCFE9,  HANGUL SYLLABLE-CFE9
        0xCFEA, # G+16567  U+CFEA  uniCFEA,  HANGUL SYLLABLE-CFEA
        0xCFEB, # G+16568  U+CFEB  uniCFEB,  HANGUL SYLLABLE-CFEB
        0xCFEC, # G+16569  U+CFEC  uniCFEC,  HANGUL SYLLABLE-CFEC
        0xCFED, # G+16570  U+CFED  uniCFED,  HANGUL SYLLABLE-CFED
        0xCFEE, # G+16571  U+CFEE  uniCFEE,  HANGUL SYLLABLE-CFEE
        0xCFEF, # G+16572  U+CFEF  uniCFEF,  HANGUL SYLLABLE-CFEF
        0xCFF2, # G+16573  U+CFF2  uniCFF2,  HANGUL SYLLABLE-CFF2
        0xCFF4, # G+16574  U+CFF4  uniCFF4,  HANGUL SYLLABLE-CFF4
        0xCFF6, # G+16575  U+CFF6  uniCFF6,  HANGUL SYLLABLE-CFF6
        0xCFF7, # G+16576  U+CFF7  uniCFF7,  HANGUL SYLLABLE-CFF7
        0xCFF8, # G+16577  U+CFF8  uniCFF8,  HANGUL SYLLABLE-CFF8
        0xCFF9, # G+16578  U+CFF9  uniCFF9,  HANGUL SYLLABLE-CFF9
        0xCFFA, # G+16579  U+CFFA  uniCFFA,  HANGUL SYLLABLE-CFFA
        0xCFFB, # G+16580  U+CFFB  uniCFFB,  HANGUL SYLLABLE-CFFB
        0xCFFD, # G+16581  U+CFFD  uniCFFD,  HANGUL SYLLABLE-CFFD
        0xCFFE, # G+16582  U+CFFE  uniCFFE,  HANGUL SYLLABLE-CFFE
        0xCFFF, # G+16583  U+CFFF  uniCFFF,  HANGUL SYLLABLE-CFFF
        0xD001, # G+16584  U+D001  uniD001,  HANGUL SYLLABLE-D001
        0xD002, # G+16585  U+D002  uniD002,  HANGUL SYLLABLE-D002
        0xD003, # G+16586  U+D003  uniD003,  HANGUL SYLLABLE-D003
        0xD005, # G+16587  U+D005  uniD005,  HANGUL SYLLABLE-D005
        0xD006, # G+16588  U+D006  uniD006,  HANGUL SYLLABLE-D006
        0xD007, # G+16589  U+D007  uniD007,  HANGUL SYLLABLE-D007
        0xD008, # G+16590  U+D008  uniD008,  HANGUL SYLLABLE-D008
        0xD009, # G+16591  U+D009  uniD009,  HANGUL SYLLABLE-D009
        0xD00A, # G+16592  U+D00A  uniD00A,  HANGUL SYLLABLE-D00A
        0xD00B, # G+16593  U+D00B  uniD00B,  HANGUL SYLLABLE-D00B
        0xD00C, # G+16594  U+D00C  uniD00C,  HANGUL SYLLABLE-D00C
        0xD00D, # G+16595  U+D00D  uniD00D,  HANGUL SYLLABLE-D00D
        0xD00E, # G+16596  U+D00E  uniD00E,  HANGUL SYLLABLE-D00E
        0xD00F, # G+16597  U+D00F  uniD00F,  HANGUL SYLLABLE-D00F
        0xD010, # G+16598  U+D010  uniD010,  HANGUL SYLLABLE-D010
        0xD012, # G+16599  U+D012  uniD012,  HANGUL SYLLABLE-D012
        0xD013, # G+16600  U+D013  uniD013,  HANGUL SYLLABLE-D013
        0xD014, # G+16601  U+D014  uniD014,  HANGUL SYLLABLE-D014
        0xD015, # G+16602  U+D015  uniD015,  HANGUL SYLLABLE-D015
        0xD016, # G+16603  U+D016  uniD016,  HANGUL SYLLABLE-D016
        0xD017, # G+16604  U+D017  uniD017,  HANGUL SYLLABLE-D017
        0xD019, # G+16605  U+D019  uniD019,  HANGUL SYLLABLE-D019
        0xD01A, # G+16606  U+D01A  uniD01A,  HANGUL SYLLABLE-D01A
        0xD01B, # G+16607  U+D01B  uniD01B,  HANGUL SYLLABLE-D01B
        0xD01C, # G+16608  U+D01C  uniD01C,  HANGUL SYLLABLE-D01C
        0xD01D, # G+16609  U+D01D  uniD01D,  HANGUL SYLLABLE-D01D
        0xD01E, # G+16610  U+D01E  uniD01E,  HANGUL SYLLABLE-D01E
        0xD01F, # G+16611  U+D01F  uniD01F,  HANGUL SYLLABLE-D01F
        0xD020, # G+16612  U+D020  uniD020,  HANGUL SYLLABLE-D020
        0xD021, # G+16613  U+D021  uniD021,  HANGUL SYLLABLE-D021
        0xD022, # G+16614  U+D022  uniD022,  HANGUL SYLLABLE-D022
        0xD023, # G+16615  U+D023  uniD023,  HANGUL SYLLABLE-D023
        0xD024, # G+16616  U+D024  uniD024,  HANGUL SYLLABLE-D024
        0xD025, # G+16617  U+D025  uniD025,  HANGUL SYLLABLE-D025
        0xD026, # G+16618  U+D026  uniD026,  HANGUL SYLLABLE-D026
        0xD027, # G+16619  U+D027  uniD027,  HANGUL SYLLABLE-D027
        0xD028, # G+16620  U+D028  uniD028,  HANGUL SYLLABLE-D028
        0xD029, # G+16621  U+D029  uniD029,  HANGUL SYLLABLE-D029
        0xD02A, # G+16622  U+D02A  uniD02A,  HANGUL SYLLABLE-D02A
        0xD02B, # G+16623  U+D02B  uniD02B,  HANGUL SYLLABLE-D02B
        0xD02C, # G+16624  U+D02C  uniD02C,  HANGUL SYLLABLE-D02C
        0xD02E, # G+16625  U+D02E  uniD02E,  HANGUL SYLLABLE-D02E
        0xD02F, # G+16626  U+D02F  uniD02F,  HANGUL SYLLABLE-D02F
        0xD030, # G+16627  U+D030  uniD030,  HANGUL SYLLABLE-D030
        0xD031, # G+16628  U+D031  uniD031,  HANGUL SYLLABLE-D031
        0xD032, # G+16629  U+D032  uniD032,  HANGUL SYLLABLE-D032
        0xD033, # G+16630  U+D033  uniD033,  HANGUL SYLLABLE-D033
        0xD036, # G+16631  U+D036  uniD036,  HANGUL SYLLABLE-D036
        0xD037, # G+16632  U+D037  uniD037,  HANGUL SYLLABLE-D037
        0xD039, # G+16633  U+D039  uniD039,  HANGUL SYLLABLE-D039
        0xD03A, # G+16634  U+D03A  uniD03A,  HANGUL SYLLABLE-D03A
        0xD03B, # G+16635  U+D03B  uniD03B,  HANGUL SYLLABLE-D03B
        0xD03D, # G+16636  U+D03D  uniD03D,  HANGUL SYLLABLE-D03D
        0xD03E, # G+16637  U+D03E  uniD03E,  HANGUL SYLLABLE-D03E
        0xD03F, # G+16638  U+D03F  uniD03F,  HANGUL SYLLABLE-D03F
        0xD040, # G+16639  U+D040  uniD040,  HANGUL SYLLABLE-D040
        0xD041, # G+16640  U+D041  uniD041,  HANGUL SYLLABLE-D041
        0xD042, # G+16641  U+D042  uniD042,  HANGUL SYLLABLE-D042
        0xD043, # G+16642  U+D043  uniD043,  HANGUL SYLLABLE-D043
        0xD046, # G+16643  U+D046  uniD046,  HANGUL SYLLABLE-D046
        0xD048, # G+16644  U+D048  uniD048,  HANGUL SYLLABLE-D048
        0xD04A, # G+16645  U+D04A  uniD04A,  HANGUL SYLLABLE-D04A
        0xD04B, # G+16646  U+D04B  uniD04B,  HANGUL SYLLABLE-D04B
        0xD04C, # G+16647  U+D04C  uniD04C,  HANGUL SYLLABLE-D04C
        0xD04D, # G+16648  U+D04D  uniD04D,  HANGUL SYLLABLE-D04D
        0xD04E, # G+16649  U+D04E  uniD04E,  HANGUL SYLLABLE-D04E
        0xD04F, # G+16650  U+D04F  uniD04F,  HANGUL SYLLABLE-D04F
        0xD051, # G+16651  U+D051  uniD051,  HANGUL SYLLABLE-D051
        0xD052, # G+16652  U+D052  uniD052,  HANGUL SYLLABLE-D052
        0xD053, # G+16653  U+D053  uniD053,  HANGUL SYLLABLE-D053
        0xD055, # G+16654  U+D055  uniD055,  HANGUL SYLLABLE-D055
        0xD056, # G+16655  U+D056  uniD056,  HANGUL SYLLABLE-D056
        0xD057, # G+16656  U+D057  uniD057,  HANGUL SYLLABLE-D057
        0xD059, # G+16657  U+D059  uniD059,  HANGUL SYLLABLE-D059
        0xD05A, # G+16658  U+D05A  uniD05A,  HANGUL SYLLABLE-D05A
        0xD05B, # G+16659  U+D05B  uniD05B,  HANGUL SYLLABLE-D05B
        0xD05C, # G+16660  U+D05C  uniD05C,  HANGUL SYLLABLE-D05C
        0xD05D, # G+16661  U+D05D  uniD05D,  HANGUL SYLLABLE-D05D
        0xD05E, # G+16662  U+D05E  uniD05E,  HANGUL SYLLABLE-D05E
        0xD05F, # G+16663  U+D05F  uniD05F,  HANGUL SYLLABLE-D05F
        0xD061, # G+16664  U+D061  uniD061,  HANGUL SYLLABLE-D061
        0xD062, # G+16665  U+D062  uniD062,  HANGUL SYLLABLE-D062
        0xD063, # G+16666  U+D063  uniD063,  HANGUL SYLLABLE-D063
        0xD064, # G+16667  U+D064  uniD064,  HANGUL SYLLABLE-D064
        0xD065, # G+16668  U+D065  uniD065,  HANGUL SYLLABLE-D065
        0xD066, # G+16669  U+D066  uniD066,  HANGUL SYLLABLE-D066
        0xD067, # G+16670  U+D067  uniD067,  HANGUL SYLLABLE-D067
        0xD068, # G+16671  U+D068  uniD068,  HANGUL SYLLABLE-D068
        0xD069, # G+16672  U+D069  uniD069,  HANGUL SYLLABLE-D069
        0xD06A, # G+16673  U+D06A  uniD06A,  HANGUL SYLLABLE-D06A
        0xD06B, # G+16674  U+D06B  uniD06B,  HANGUL SYLLABLE-D06B
        0xD06E, # G+16675  U+D06E  uniD06E,  HANGUL SYLLABLE-D06E
        0xD06F, # G+16676  U+D06F  uniD06F,  HANGUL SYLLABLE-D06F
        0xD071, # G+16677  U+D071  uniD071,  HANGUL SYLLABLE-D071
        0xD072, # G+16678  U+D072  uniD072,  HANGUL SYLLABLE-D072
        0xD073, # G+16679  U+D073  uniD073,  HANGUL SYLLABLE-D073
        0xD075, # G+16680  U+D075  uniD075,  HANGUL SYLLABLE-D075
        0xD076, # G+16681  U+D076  uniD076,  HANGUL SYLLABLE-D076
        0xD077, # G+16682  U+D077  uniD077,  HANGUL SYLLABLE-D077
        0xD078, # G+16683  U+D078  uniD078,  HANGUL SYLLABLE-D078
        0xD079, # G+16684  U+D079  uniD079,  HANGUL SYLLABLE-D079
        0xD07A, # G+16685  U+D07A  uniD07A,  HANGUL SYLLABLE-D07A
        0xD07B, # G+16686  U+D07B  uniD07B,  HANGUL SYLLABLE-D07B
        0xD07E, # G+16687  U+D07E  uniD07E,  HANGUL SYLLABLE-D07E
        0xD07F, # G+16688  U+D07F  uniD07F,  HANGUL SYLLABLE-D07F
        0xD080, # G+16689  U+D080  uniD080,  HANGUL SYLLABLE-D080
        0xD082, # G+16690  U+D082  uniD082,  HANGUL SYLLABLE-D082
        0xD083, # G+16691  U+D083  uniD083,  HANGUL SYLLABLE-D083
        0xD084, # G+16692  U+D084  uniD084,  HANGUL SYLLABLE-D084
        0xD085, # G+16693  U+D085  uniD085,  HANGUL SYLLABLE-D085
        0xD086, # G+16694  U+D086  uniD086,  HANGUL SYLLABLE-D086
        0xD087, # G+16695  U+D087  uniD087,  HANGUL SYLLABLE-D087
        0xD088, # G+16696  U+D088  uniD088,  HANGUL SYLLABLE-D088
        0xD089, # G+16697  U+D089  uniD089,  HANGUL SYLLABLE-D089
        0xD08A, # G+16698  U+D08A  uniD08A,  HANGUL SYLLABLE-D08A
        0xD08B, # G+16699  U+D08B  uniD08B,  HANGUL SYLLABLE-D08B
        0xD08C, # G+16700  U+D08C  uniD08C,  HANGUL SYLLABLE-D08C
        0xD08D, # G+16701  U+D08D  uniD08D,  HANGUL SYLLABLE-D08D
        0xD08E, # G+16702  U+D08E  uniD08E,  HANGUL SYLLABLE-D08E
        0xD08F, # G+16703  U+D08F  uniD08F,  HANGUL SYLLABLE-D08F
        0xD090, # G+16704  U+D090  uniD090,  HANGUL SYLLABLE-D090
        0xD091, # G+16705  U+D091  uniD091,  HANGUL SYLLABLE-D091
        0xD092, # G+16706  U+D092  uniD092,  HANGUL SYLLABLE-D092
        0xD093, # G+16707  U+D093  uniD093,  HANGUL SYLLABLE-D093
        0xD094, # G+16708  U+D094  uniD094,  HANGUL SYLLABLE-D094
        0xD095, # G+16709  U+D095  uniD095,  HANGUL SYLLABLE-D095
        0xD096, # G+16710  U+D096  uniD096,  HANGUL SYLLABLE-D096
        0xD097, # G+16711  U+D097  uniD097,  HANGUL SYLLABLE-D097
        0xD098, # G+16712  U+D098  uniD098,  HANGUL SYLLABLE-D098
        0xD099, # G+16713  U+D099  uniD099,  HANGUL SYLLABLE-D099
        0xD09A, # G+16714  U+D09A  uniD09A,  HANGUL SYLLABLE-D09A
        0xD09B, # G+16715  U+D09B  uniD09B,  HANGUL SYLLABLE-D09B
        0xD09C, # G+16716  U+D09C  uniD09C,  HANGUL SYLLABLE-D09C
        0xD09D, # G+16717  U+D09D  uniD09D,  HANGUL SYLLABLE-D09D
        0xD09E, # G+16718  U+D09E  uniD09E,  HANGUL SYLLABLE-D09E
        0xD09F, # G+16719  U+D09F  uniD09F,  HANGUL SYLLABLE-D09F
        0xD0A0, # G+16720  U+D0A0  uniD0A0,  HANGUL SYLLABLE-D0A0
        0xD0A1, # G+16721  U+D0A1  uniD0A1,  HANGUL SYLLABLE-D0A1
        0xD0A2, # G+16722  U+D0A2  uniD0A2,  HANGUL SYLLABLE-D0A2
        0xD0A3, # G+16723  U+D0A3  uniD0A3,  HANGUL SYLLABLE-D0A3
        0xD0A6, # G+16724  U+D0A6  uniD0A6,  HANGUL SYLLABLE-D0A6
        0xD0A7, # G+16725  U+D0A7  uniD0A7,  HANGUL SYLLABLE-D0A7
        0xD0A9, # G+16726  U+D0A9  uniD0A9,  HANGUL SYLLABLE-D0A9
        0xD0AA, # G+16727  U+D0AA  uniD0AA,  HANGUL SYLLABLE-D0AA
        0xD0AB, # G+16728  U+D0AB  uniD0AB,  HANGUL SYLLABLE-D0AB
        0xD0AD, # G+16729  U+D0AD  uniD0AD,  HANGUL SYLLABLE-D0AD
        0xD0AE, # G+16730  U+D0AE  uniD0AE,  HANGUL SYLLABLE-D0AE
        0xD0AF, # G+16731  U+D0AF  uniD0AF,  HANGUL SYLLABLE-D0AF
        0xD0B0, # G+16732  U+D0B0  uniD0B0,  HANGUL SYLLABLE-D0B0
        0xD0B1, # G+16733  U+D0B1  uniD0B1,  HANGUL SYLLABLE-D0B1
        0xD0B2, # G+16734  U+D0B2  uniD0B2,  HANGUL SYLLABLE-D0B2
        0xD0B3, # G+16735  U+D0B3  uniD0B3,  HANGUL SYLLABLE-D0B3
        0xD0B6, # G+16736  U+D0B6  uniD0B6,  HANGUL SYLLABLE-D0B6
        0xD0B8, # G+16737  U+D0B8  uniD0B8,  HANGUL SYLLABLE-D0B8
        0xD0BA, # G+16738  U+D0BA  uniD0BA,  HANGUL SYLLABLE-D0BA
        0xD0BB, # G+16739  U+D0BB  uniD0BB,  HANGUL SYLLABLE-D0BB
        0xD0BC, # G+16740  U+D0BC  uniD0BC,  HANGUL SYLLABLE-D0BC
        0xD0BD, # G+16741  U+D0BD  uniD0BD,  HANGUL SYLLABLE-D0BD
        0xD0BE, # G+16742  U+D0BE  uniD0BE,  HANGUL SYLLABLE-D0BE
        0xD0BF, # G+16743  U+D0BF  uniD0BF,  HANGUL SYLLABLE-D0BF
        0xD0C2, # G+16744  U+D0C2  uniD0C2,  HANGUL SYLLABLE-D0C2
        0xD0C3, # G+16745  U+D0C3  uniD0C3,  HANGUL SYLLABLE-D0C3
        0xD0C5, # G+16746  U+D0C5  uniD0C5,  HANGUL SYLLABLE-D0C5
        0xD0C6, # G+16747  U+D0C6  uniD0C6,  HANGUL SYLLABLE-D0C6
        0xD0C7, # G+16748  U+D0C7  uniD0C7,  HANGUL SYLLABLE-D0C7
        0xD0CA, # G+16749  U+D0CA  uniD0CA,  HANGUL SYLLABLE-D0CA
        0xD0CB, # G+16750  U+D0CB  uniD0CB,  HANGUL SYLLABLE-D0CB
        0xD0CC, # G+16751  U+D0CC  uniD0CC,  HANGUL SYLLABLE-D0CC
        0xD0CD, # G+16752  U+D0CD  uniD0CD,  HANGUL SYLLABLE-D0CD
        0xD0CE, # G+16753  U+D0CE  uniD0CE,  HANGUL SYLLABLE-D0CE
        0xD0CF, # G+16754  U+D0CF  uniD0CF,  HANGUL SYLLABLE-D0CF
        0xD0D2, # G+16755  U+D0D2  uniD0D2,  HANGUL SYLLABLE-D0D2
        0xD0D6, # G+16756  U+D0D6  uniD0D6,  HANGUL SYLLABLE-D0D6
        0xD0D7, # G+16757  U+D0D7  uniD0D7,  HANGUL SYLLABLE-D0D7
        0xD0D8, # G+16758  U+D0D8  uniD0D8,  HANGUL SYLLABLE-D0D8
        0xD0D9, # G+16759  U+D0D9  uniD0D9,  HANGUL SYLLABLE-D0D9
        0xD0DA, # G+16760  U+D0DA  uniD0DA,  HANGUL SYLLABLE-D0DA
        0xD0DB, # G+16761  U+D0DB  uniD0DB,  HANGUL SYLLABLE-D0DB
        0xD0DE, # G+16762  U+D0DE  uniD0DE,  HANGUL SYLLABLE-D0DE
        0xD0DF, # G+16763  U+D0DF  uniD0DF,  HANGUL SYLLABLE-D0DF
        0xD0E1, # G+16764  U+D0E1  uniD0E1,  HANGUL SYLLABLE-D0E1
        0xD0E2, # G+16765  U+D0E2  uniD0E2,  HANGUL SYLLABLE-D0E2
        0xD0E3, # G+16766  U+D0E3  uniD0E3,  HANGUL SYLLABLE-D0E3
        0xD0E5, # G+16767  U+D0E5  uniD0E5,  HANGUL SYLLABLE-D0E5
        0xD0E6, # G+16768  U+D0E6  uniD0E6,  HANGUL SYLLABLE-D0E6
        0xD0E7, # G+16769  U+D0E7  uniD0E7,  HANGUL SYLLABLE-D0E7
        0xD0E8, # G+16770  U+D0E8  uniD0E8,  HANGUL SYLLABLE-D0E8
        0xD0E9, # G+16771  U+D0E9  uniD0E9,  HANGUL SYLLABLE-D0E9
        0xD0EA, # G+16772  U+D0EA  uniD0EA,  HANGUL SYLLABLE-D0EA
        0xD0EB, # G+16773  U+D0EB  uniD0EB,  HANGUL SYLLABLE-D0EB
        0xD0EE, # G+16774  U+D0EE  uniD0EE,  HANGUL SYLLABLE-D0EE
        0xD0F2, # G+16775  U+D0F2  uniD0F2,  HANGUL SYLLABLE-D0F2
        0xD0F3, # G+16776  U+D0F3  uniD0F3,  HANGUL SYLLABLE-D0F3
        0xD0F4, # G+16777  U+D0F4  uniD0F4,  HANGUL SYLLABLE-D0F4
        0xD0F5, # G+16778  U+D0F5  uniD0F5,  HANGUL SYLLABLE-D0F5
        0xD0F6, # G+16779  U+D0F6  uniD0F6,  HANGUL SYLLABLE-D0F6
        0xD0F7, # G+16780  U+D0F7  uniD0F7,  HANGUL SYLLABLE-D0F7
        0xD0F9, # G+16781  U+D0F9  uniD0F9,  HANGUL SYLLABLE-D0F9
        0xD0FA, # G+16782  U+D0FA  uniD0FA,  HANGUL SYLLABLE-D0FA
        0xD0FB, # G+16783  U+D0FB  uniD0FB,  HANGUL SYLLABLE-D0FB
        0xD0FC, # G+16784  U+D0FC  uniD0FC,  HANGUL SYLLABLE-D0FC
        0xD0FD, # G+16785  U+D0FD  uniD0FD,  HANGUL SYLLABLE-D0FD
        0xD0FE, # G+16786  U+D0FE  uniD0FE,  HANGUL SYLLABLE-D0FE
        0xD0FF, # G+16787  U+D0FF  uniD0FF,  HANGUL SYLLABLE-D0FF
        0xD100, # G+16788  U+D100  uniD100,  HANGUL SYLLABLE-D100
        0xD101, # G+16789  U+D101  uniD101,  HANGUL SYLLABLE-D101
        0xD102, # G+16790  U+D102  uniD102,  HANGUL SYLLABLE-D102
        0xD103, # G+16791  U+D103  uniD103,  HANGUL SYLLABLE-D103
        0xD104, # G+16792  U+D104  uniD104,  HANGUL SYLLABLE-D104
        0xD105, # G+16793  U+D105  uniD105,  HANGUL SYLLABLE-D105
        0xD106, # G+16794  U+D106  uniD106,  HANGUL SYLLABLE-D106
        0xD107, # G+16795  U+D107  uniD107,  HANGUL SYLLABLE-D107
        0xD108, # G+16796  U+D108  uniD108,  HANGUL SYLLABLE-D108
        0xD109, # G+16797  U+D109  uniD109,  HANGUL SYLLABLE-D109
        0xD10A, # G+16798  U+D10A  uniD10A,  HANGUL SYLLABLE-D10A
        0xD10B, # G+16799  U+D10B  uniD10B,  HANGUL SYLLABLE-D10B
        0xD10C, # G+16800  U+D10C  uniD10C,  HANGUL SYLLABLE-D10C
        0xD10E, # G+16801  U+D10E  uniD10E,  HANGUL SYLLABLE-D10E
        0xD10F, # G+16802  U+D10F  uniD10F,  HANGUL SYLLABLE-D10F
        0xD110, # G+16803  U+D110  uniD110,  HANGUL SYLLABLE-D110
        0xD111, # G+16804  U+D111  uniD111,  HANGUL SYLLABLE-D111
        0xD112, # G+16805  U+D112  uniD112,  HANGUL SYLLABLE-D112
        0xD113, # G+16806  U+D113  uniD113,  HANGUL SYLLABLE-D113
        0xD114, # G+16807  U+D114  uniD114,  HANGUL SYLLABLE-D114
        0xD115, # G+16808  U+D115  uniD115,  HANGUL SYLLABLE-D115
        0xD116, # G+16809  U+D116  uniD116,  HANGUL SYLLABLE-D116
        0xD117, # G+16810  U+D117  uniD117,  HANGUL SYLLABLE-D117
        0xD118, # G+16811  U+D118  uniD118,  HANGUL SYLLABLE-D118
        0xD119, # G+16812  U+D119  uniD119,  HANGUL SYLLABLE-D119
        0xD11A, # G+16813  U+D11A  uniD11A,  HANGUL SYLLABLE-D11A
        0xD11B, # G+16814  U+D11B  uniD11B,  HANGUL SYLLABLE-D11B
        0xD11C, # G+16815  U+D11C  uniD11C,  HANGUL SYLLABLE-D11C
        0xD11D, # G+16816  U+D11D  uniD11D,  HANGUL SYLLABLE-D11D
        0xD11E, # G+16817  U+D11E  uniD11E,  HANGUL SYLLABLE-D11E
        0xD11F, # G+16818  U+D11F  uniD11F,  HANGUL SYLLABLE-D11F
        0xD120, # G+16819  U+D120  uniD120,  HANGUL SYLLABLE-D120
        0xD121, # G+16820  U+D121  uniD121,  HANGUL SYLLABLE-D121
        0xD122, # G+16821  U+D122  uniD122,  HANGUL SYLLABLE-D122
        0xD123, # G+16822  U+D123  uniD123,  HANGUL SYLLABLE-D123
        0xD124, # G+16823  U+D124  uniD124,  HANGUL SYLLABLE-D124
        0xD125, # G+16824  U+D125  uniD125,  HANGUL SYLLABLE-D125
        0xD126, # G+16825  U+D126  uniD126,  HANGUL SYLLABLE-D126
        0xD127, # G+16826  U+D127  uniD127,  HANGUL SYLLABLE-D127
        0xD128, # G+16827  U+D128  uniD128,  HANGUL SYLLABLE-D128
        0xD129, # G+16828  U+D129  uniD129,  HANGUL SYLLABLE-D129
        0xD12A, # G+16829  U+D12A  uniD12A,  HANGUL SYLLABLE-D12A
        0xD12B, # G+16830  U+D12B  uniD12B,  HANGUL SYLLABLE-D12B
        0xD12C, # G+16831  U+D12C  uniD12C,  HANGUL SYLLABLE-D12C
        0xD12D, # G+16832  U+D12D  uniD12D,  HANGUL SYLLABLE-D12D
        0xD12E, # G+16833  U+D12E  uniD12E,  HANGUL SYLLABLE-D12E
        0xD12F, # G+16834  U+D12F  uniD12F,  HANGUL SYLLABLE-D12F
        0xD132, # G+16835  U+D132  uniD132,  HANGUL SYLLABLE-D132
        0xD133, # G+16836  U+D133  uniD133,  HANGUL SYLLABLE-D133
        0xD135, # G+16837  U+D135  uniD135,  HANGUL SYLLABLE-D135
        0xD136, # G+16838  U+D136  uniD136,  HANGUL SYLLABLE-D136
        0xD137, # G+16839  U+D137  uniD137,  HANGUL SYLLABLE-D137
        0xD139, # G+16840  U+D139  uniD139,  HANGUL SYLLABLE-D139
        0xD13B, # G+16841  U+D13B  uniD13B,  HANGUL SYLLABLE-D13B
        0xD13C, # G+16842  U+D13C  uniD13C,  HANGUL SYLLABLE-D13C
        0xD13D, # G+16843  U+D13D  uniD13D,  HANGUL SYLLABLE-D13D
        0xD13E, # G+16844  U+D13E  uniD13E,  HANGUL SYLLABLE-D13E
        0xD13F, # G+16845  U+D13F  uniD13F,  HANGUL SYLLABLE-D13F
        0xD142, # G+16846  U+D142  uniD142,  HANGUL SYLLABLE-D142
        0xD146, # G+16847  U+D146  uniD146,  HANGUL SYLLABLE-D146
        0xD147, # G+16848  U+D147  uniD147,  HANGUL SYLLABLE-D147
        0xD148, # G+16849  U+D148  uniD148,  HANGUL SYLLABLE-D148
        0xD149, # G+16850  U+D149  uniD149,  HANGUL SYLLABLE-D149
        0xD14A, # G+16851  U+D14A  uniD14A,  HANGUL SYLLABLE-D14A
        0xD14B, # G+16852  U+D14B  uniD14B,  HANGUL SYLLABLE-D14B
        0xD14E, # G+16853  U+D14E  uniD14E,  HANGUL SYLLABLE-D14E
        0xD14F, # G+16854  U+D14F  uniD14F,  HANGUL SYLLABLE-D14F
        0xD151, # G+16855  U+D151  uniD151,  HANGUL SYLLABLE-D151
        0xD152, # G+16856  U+D152  uniD152,  HANGUL SYLLABLE-D152
        0xD153, # G+16857  U+D153  uniD153,  HANGUL SYLLABLE-D153
        0xD155, # G+16858  U+D155  uniD155,  HANGUL SYLLABLE-D155
        0xD156, # G+16859  U+D156  uniD156,  HANGUL SYLLABLE-D156
        0xD157, # G+16860  U+D157  uniD157,  HANGUL SYLLABLE-D157
        0xD158, # G+16861  U+D158  uniD158,  HANGUL SYLLABLE-D158
        0xD159, # G+16862  U+D159  uniD159,  HANGUL SYLLABLE-D159
        0xD15A, # G+16863  U+D15A  uniD15A,  HANGUL SYLLABLE-D15A
        0xD15B, # G+16864  U+D15B  uniD15B,  HANGUL SYLLABLE-D15B
        0xD15E, # G+16865  U+D15E  uniD15E,  HANGUL SYLLABLE-D15E
        0xD160, # G+16866  U+D160  uniD160,  HANGUL SYLLABLE-D160
        0xD162, # G+16867  U+D162  uniD162,  HANGUL SYLLABLE-D162
        0xD163, # G+16868  U+D163  uniD163,  HANGUL SYLLABLE-D163
        0xD164, # G+16869  U+D164  uniD164,  HANGUL SYLLABLE-D164
        0xD165, # G+16870  U+D165  uniD165,  HANGUL SYLLABLE-D165
        0xD166, # G+16871  U+D166  uniD166,  HANGUL SYLLABLE-D166
        0xD167, # G+16872  U+D167  uniD167,  HANGUL SYLLABLE-D167
        0xD169, # G+16873  U+D169  uniD169,  HANGUL SYLLABLE-D169
        0xD16A, # G+16874  U+D16A  uniD16A,  HANGUL SYLLABLE-D16A
        0xD16B, # G+16875  U+D16B  uniD16B,  HANGUL SYLLABLE-D16B
        0xD16D, # G+16876  U+D16D  uniD16D,  HANGUL SYLLABLE-D16D
        0xD16E, # G+16877  U+D16E  uniD16E,  HANGUL SYLLABLE-D16E
        0xD16F, # G+16878  U+D16F  uniD16F,  HANGUL SYLLABLE-D16F
        0xD170, # G+16879  U+D170  uniD170,  HANGUL SYLLABLE-D170
        0xD171, # G+16880  U+D171  uniD171,  HANGUL SYLLABLE-D171
        0xD172, # G+16881  U+D172  uniD172,  HANGUL SYLLABLE-D172
        0xD173, # G+16882  U+D173  uniD173,  HANGUL SYLLABLE-D173
        0xD174, # G+16883  U+D174  uniD174,  HANGUL SYLLABLE-D174
        0xD175, # G+16884  U+D175  uniD175,  HANGUL SYLLABLE-D175
        0xD176, # G+16885  U+D176  uniD176,  HANGUL SYLLABLE-D176
        0xD177, # G+16886  U+D177  uniD177,  HANGUL SYLLABLE-D177
        0xD178, # G+16887  U+D178  uniD178,  HANGUL SYLLABLE-D178
        0xD179, # G+16888  U+D179  uniD179,  HANGUL SYLLABLE-D179
        0xD17A, # G+16889  U+D17A  uniD17A,  HANGUL SYLLABLE-D17A
        0xD17B, # G+16890  U+D17B  uniD17B,  HANGUL SYLLABLE-D17B
        0xD17D, # G+16891  U+D17D  uniD17D,  HANGUL SYLLABLE-D17D
        0xD17E, # G+16892  U+D17E  uniD17E,  HANGUL SYLLABLE-D17E
        0xD17F, # G+16893  U+D17F  uniD17F,  HANGUL SYLLABLE-D17F
        0xD180, # G+16894  U+D180  uniD180,  HANGUL SYLLABLE-D180
        0xD181, # G+16895  U+D181  uniD181,  HANGUL SYLLABLE-D181
        0xD182, # G+16896  U+D182  uniD182,  HANGUL SYLLABLE-D182
        0xD183, # G+16897  U+D183  uniD183,  HANGUL SYLLABLE-D183
        0xD185, # G+16898  U+D185  uniD185,  HANGUL SYLLABLE-D185
        0xD186, # G+16899  U+D186  uniD186,  HANGUL SYLLABLE-D186
        0xD187, # G+16900  U+D187  uniD187,  HANGUL SYLLABLE-D187
        0xD189, # G+16901  U+D189  uniD189,  HANGUL SYLLABLE-D189
        0xD18A, # G+16902  U+D18A  uniD18A,  HANGUL SYLLABLE-D18A
        0xD18B, # G+16903  U+D18B  uniD18B,  HANGUL SYLLABLE-D18B
        0xD18C, # G+16904  U+D18C  uniD18C,  HANGUL SYLLABLE-D18C
        0xD18D, # G+16905  U+D18D  uniD18D,  HANGUL SYLLABLE-D18D
        0xD18E, # G+16906  U+D18E  uniD18E,  HANGUL SYLLABLE-D18E
        0xD18F, # G+16907  U+D18F  uniD18F,  HANGUL SYLLABLE-D18F
        0xD190, # G+16908  U+D190  uniD190,  HANGUL SYLLABLE-D190
        0xD191, # G+16909  U+D191  uniD191,  HANGUL SYLLABLE-D191
        0xD192, # G+16910  U+D192  uniD192,  HANGUL SYLLABLE-D192
        0xD193, # G+16911  U+D193  uniD193,  HANGUL SYLLABLE-D193
        0xD194, # G+16912  U+D194  uniD194,  HANGUL SYLLABLE-D194
        0xD195, # G+16913  U+D195  uniD195,  HANGUL SYLLABLE-D195
        0xD196, # G+16914  U+D196  uniD196,  HANGUL SYLLABLE-D196
        0xD197, # G+16915  U+D197  uniD197,  HANGUL SYLLABLE-D197
        0xD198, # G+16916  U+D198  uniD198,  HANGUL SYLLABLE-D198
        0xD199, # G+16917  U+D199  uniD199,  HANGUL SYLLABLE-D199
        0xD19A, # G+16918  U+D19A  uniD19A,  HANGUL SYLLABLE-D19A
        0xD19B, # G+16919  U+D19B  uniD19B,  HANGUL SYLLABLE-D19B
        0xD19C, # G+16920  U+D19C  uniD19C,  HANGUL SYLLABLE-D19C
        0xD19D, # G+16921  U+D19D  uniD19D,  HANGUL SYLLABLE-D19D
        0xD19E, # G+16922  U+D19E  uniD19E,  HANGUL SYLLABLE-D19E
        0xD19F, # G+16923  U+D19F  uniD19F,  HANGUL SYLLABLE-D19F
        0xD1A2, # G+16924  U+D1A2  uniD1A2,  HANGUL SYLLABLE-D1A2
        0xD1A3, # G+16925  U+D1A3  uniD1A3,  HANGUL SYLLABLE-D1A3
        0xD1A5, # G+16926  U+D1A5  uniD1A5,  HANGUL SYLLABLE-D1A5
        0xD1A6, # G+16927  U+D1A6  uniD1A6,  HANGUL SYLLABLE-D1A6
        0xD1A7, # G+16928  U+D1A7  uniD1A7,  HANGUL SYLLABLE-D1A7
        0xD1A9, # G+16929  U+D1A9  uniD1A9,  HANGUL SYLLABLE-D1A9
        0xD1AA, # G+16930  U+D1AA  uniD1AA,  HANGUL SYLLABLE-D1AA
        0xD1AB, # G+16931  U+D1AB  uniD1AB,  HANGUL SYLLABLE-D1AB
        0xD1AC, # G+16932  U+D1AC  uniD1AC,  HANGUL SYLLABLE-D1AC
        0xD1AD, # G+16933  U+D1AD  uniD1AD,  HANGUL SYLLABLE-D1AD
        0xD1AE, # G+16934  U+D1AE  uniD1AE,  HANGUL SYLLABLE-D1AE
        0xD1AF, # G+16935  U+D1AF  uniD1AF,  HANGUL SYLLABLE-D1AF
        0xD1B2, # G+16936  U+D1B2  uniD1B2,  HANGUL SYLLABLE-D1B2
        0xD1B4, # G+16937  U+D1B4  uniD1B4,  HANGUL SYLLABLE-D1B4
        0xD1B6, # G+16938  U+D1B6  uniD1B6,  HANGUL SYLLABLE-D1B6
        0xD1B7, # G+16939  U+D1B7  uniD1B7,  HANGUL SYLLABLE-D1B7
        0xD1B8, # G+16940  U+D1B8  uniD1B8,  HANGUL SYLLABLE-D1B8
        0xD1B9, # G+16941  U+D1B9  uniD1B9,  HANGUL SYLLABLE-D1B9
        0xD1BB, # G+16942  U+D1BB  uniD1BB,  HANGUL SYLLABLE-D1BB
        0xD1BD, # G+16943  U+D1BD  uniD1BD,  HANGUL SYLLABLE-D1BD
        0xD1BE, # G+16944  U+D1BE  uniD1BE,  HANGUL SYLLABLE-D1BE
        0xD1BF, # G+16945  U+D1BF  uniD1BF,  HANGUL SYLLABLE-D1BF
        0xD1C1, # G+16946  U+D1C1  uniD1C1,  HANGUL SYLLABLE-D1C1
        0xD1C2, # G+16947  U+D1C2  uniD1C2,  HANGUL SYLLABLE-D1C2
        0xD1C3, # G+16948  U+D1C3  uniD1C3,  HANGUL SYLLABLE-D1C3
        0xD1C4, # G+16949  U+D1C4  uniD1C4,  HANGUL SYLLABLE-D1C4
        0xD1C5, # G+16950  U+D1C5  uniD1C5,  HANGUL SYLLABLE-D1C5
        0xD1C6, # G+16951  U+D1C6  uniD1C6,  HANGUL SYLLABLE-D1C6
        0xD1C7, # G+16952  U+D1C7  uniD1C7,  HANGUL SYLLABLE-D1C7
        0xD1C8, # G+16953  U+D1C8  uniD1C8,  HANGUL SYLLABLE-D1C8
        0xD1C9, # G+16954  U+D1C9  uniD1C9,  HANGUL SYLLABLE-D1C9
        0xD1CA, # G+16955  U+D1CA  uniD1CA,  HANGUL SYLLABLE-D1CA
        0xD1CB, # G+16956  U+D1CB  uniD1CB,  HANGUL SYLLABLE-D1CB
        0xD1CC, # G+16957  U+D1CC  uniD1CC,  HANGUL SYLLABLE-D1CC
        0xD1CD, # G+16958  U+D1CD  uniD1CD,  HANGUL SYLLABLE-D1CD
        0xD1CE, # G+16959  U+D1CE  uniD1CE,  HANGUL SYLLABLE-D1CE
        0xD1CF, # G+16960  U+D1CF  uniD1CF,  HANGUL SYLLABLE-D1CF
        0xD1D0, # G+16961  U+D1D0  uniD1D0,  HANGUL SYLLABLE-D1D0
        0xD1D1, # G+16962  U+D1D1  uniD1D1,  HANGUL SYLLABLE-D1D1
        0xD1D2, # G+16963  U+D1D2  uniD1D2,  HANGUL SYLLABLE-D1D2
        0xD1D3, # G+16964  U+D1D3  uniD1D3,  HANGUL SYLLABLE-D1D3
        0xD1D4, # G+16965  U+D1D4  uniD1D4,  HANGUL SYLLABLE-D1D4
        0xD1D5, # G+16966  U+D1D5  uniD1D5,  HANGUL SYLLABLE-D1D5
        0xD1D6, # G+16967  U+D1D6  uniD1D6,  HANGUL SYLLABLE-D1D6
        0xD1D7, # G+16968  U+D1D7  uniD1D7,  HANGUL SYLLABLE-D1D7
        0xD1D9, # G+16969  U+D1D9  uniD1D9,  HANGUL SYLLABLE-D1D9
        0xD1DA, # G+16970  U+D1DA  uniD1DA,  HANGUL SYLLABLE-D1DA
        0xD1DB, # G+16971  U+D1DB  uniD1DB,  HANGUL SYLLABLE-D1DB
        0xD1DC, # G+16972  U+D1DC  uniD1DC,  HANGUL SYLLABLE-D1DC
        0xD1DD, # G+16973  U+D1DD  uniD1DD,  HANGUL SYLLABLE-D1DD
        0xD1DE, # G+16974  U+D1DE  uniD1DE,  HANGUL SYLLABLE-D1DE
        0xD1DF, # G+16975  U+D1DF  uniD1DF,  HANGUL SYLLABLE-D1DF
        0xD1E0, # G+16976  U+D1E0  uniD1E0,  HANGUL SYLLABLE-D1E0
        0xD1E1, # G+16977  U+D1E1  uniD1E1,  HANGUL SYLLABLE-D1E1
        0xD1E2, # G+16978  U+D1E2  uniD1E2,  HANGUL SYLLABLE-D1E2
        0xD1E3, # G+16979  U+D1E3  uniD1E3,  HANGUL SYLLABLE-D1E3
        0xD1E4, # G+16980  U+D1E4  uniD1E4,  HANGUL SYLLABLE-D1E4
        0xD1E5, # G+16981  U+D1E5  uniD1E5,  HANGUL SYLLABLE-D1E5
        0xD1E6, # G+16982  U+D1E6  uniD1E6,  HANGUL SYLLABLE-D1E6
        0xD1E7, # G+16983  U+D1E7  uniD1E7,  HANGUL SYLLABLE-D1E7
        0xD1E8, # G+16984  U+D1E8  uniD1E8,  HANGUL SYLLABLE-D1E8
        0xD1E9, # G+16985  U+D1E9  uniD1E9,  HANGUL SYLLABLE-D1E9
        0xD1EA, # G+16986  U+D1EA  uniD1EA,  HANGUL SYLLABLE-D1EA
        0xD1EB, # G+16987  U+D1EB  uniD1EB,  HANGUL SYLLABLE-D1EB
        0xD1EC, # G+16988  U+D1EC  uniD1EC,  HANGUL SYLLABLE-D1EC
        0xD1ED, # G+16989  U+D1ED  uniD1ED,  HANGUL SYLLABLE-D1ED
        0xD1EE, # G+16990  U+D1EE  uniD1EE,  HANGUL SYLLABLE-D1EE
        0xD1EF, # G+16991  U+D1EF  uniD1EF,  HANGUL SYLLABLE-D1EF
        0xD1F0, # G+16992  U+D1F0  uniD1F0,  HANGUL SYLLABLE-D1F0
        0xD1F1, # G+16993  U+D1F1  uniD1F1,  HANGUL SYLLABLE-D1F1
        0xD1F2, # G+16994  U+D1F2  uniD1F2,  HANGUL SYLLABLE-D1F2
        0xD1F3, # G+16995  U+D1F3  uniD1F3,  HANGUL SYLLABLE-D1F3
        0xD1F5, # G+16996  U+D1F5  uniD1F5,  HANGUL SYLLABLE-D1F5
        0xD1F6, # G+16997  U+D1F6  uniD1F6,  HANGUL SYLLABLE-D1F6
        0xD1F7, # G+16998  U+D1F7  uniD1F7,  HANGUL SYLLABLE-D1F7
        0xD1F9, # G+16999  U+D1F9  uniD1F9,  HANGUL SYLLABLE-D1F9
        0xD1FA, # G+17000  U+D1FA  uniD1FA,  HANGUL SYLLABLE-D1FA
        0xD1FB, # G+17001  U+D1FB  uniD1FB,  HANGUL SYLLABLE-D1FB
        0xD1FC, # G+17002  U+D1FC  uniD1FC,  HANGUL SYLLABLE-D1FC
        0xD1FD, # G+17003  U+D1FD  uniD1FD,  HANGUL SYLLABLE-D1FD
        0xD1FE, # G+17004  U+D1FE  uniD1FE,  HANGUL SYLLABLE-D1FE
        0xD1FF, # G+17005  U+D1FF  uniD1FF,  HANGUL SYLLABLE-D1FF
        0xD200, # G+17006  U+D200  uniD200,  HANGUL SYLLABLE-D200
        0xD201, # G+17007  U+D201  uniD201,  HANGUL SYLLABLE-D201
        0xD202, # G+17008  U+D202  uniD202,  HANGUL SYLLABLE-D202
        0xD203, # G+17009  U+D203  uniD203,  HANGUL SYLLABLE-D203
        0xD204, # G+17010  U+D204  uniD204,  HANGUL SYLLABLE-D204
        0xD205, # G+17011  U+D205  uniD205,  HANGUL SYLLABLE-D205
        0xD206, # G+17012  U+D206  uniD206,  HANGUL SYLLABLE-D206
        0xD208, # G+17013  U+D208  uniD208,  HANGUL SYLLABLE-D208
        0xD20A, # G+17014  U+D20A  uniD20A,  HANGUL SYLLABLE-D20A
        0xD20B, # G+17015  U+D20B  uniD20B,  HANGUL SYLLABLE-D20B
        0xD20C, # G+17016  U+D20C  uniD20C,  HANGUL SYLLABLE-D20C
        0xD20D, # G+17017  U+D20D  uniD20D,  HANGUL SYLLABLE-D20D
        0xD20E, # G+17018  U+D20E  uniD20E,  HANGUL SYLLABLE-D20E
        0xD20F, # G+17019  U+D20F  uniD20F,  HANGUL SYLLABLE-D20F
        0xD211, # G+17020  U+D211  uniD211,  HANGUL SYLLABLE-D211
        0xD212, # G+17021  U+D212  uniD212,  HANGUL SYLLABLE-D212
        0xD213, # G+17022  U+D213  uniD213,  HANGUL SYLLABLE-D213
        0xD214, # G+17023  U+D214  uniD214,  HANGUL SYLLABLE-D214
        0xD215, # G+17024  U+D215  uniD215,  HANGUL SYLLABLE-D215
        0xD216, # G+17025  U+D216  uniD216,  HANGUL SYLLABLE-D216
        0xD217, # G+17026  U+D217  uniD217,  HANGUL SYLLABLE-D217
        0xD218, # G+17027  U+D218  uniD218,  HANGUL SYLLABLE-D218
        0xD219, # G+17028  U+D219  uniD219,  HANGUL SYLLABLE-D219
        0xD21A, # G+17029  U+D21A  uniD21A,  HANGUL SYLLABLE-D21A
        0xD21B, # G+17030  U+D21B  uniD21B,  HANGUL SYLLABLE-D21B
        0xD21C, # G+17031  U+D21C  uniD21C,  HANGUL SYLLABLE-D21C
        0xD21D, # G+17032  U+D21D  uniD21D,  HANGUL SYLLABLE-D21D
        0xD21E, # G+17033  U+D21E  uniD21E,  HANGUL SYLLABLE-D21E
        0xD21F, # G+17034  U+D21F  uniD21F,  HANGUL SYLLABLE-D21F
        0xD220, # G+17035  U+D220  uniD220,  HANGUL SYLLABLE-D220
        0xD221, # G+17036  U+D221  uniD221,  HANGUL SYLLABLE-D221
        0xD222, # G+17037  U+D222  uniD222,  HANGUL SYLLABLE-D222
        0xD223, # G+17038  U+D223  uniD223,  HANGUL SYLLABLE-D223
        0xD224, # G+17039  U+D224  uniD224,  HANGUL SYLLABLE-D224
        0xD225, # G+17040  U+D225  uniD225,  HANGUL SYLLABLE-D225
        0xD226, # G+17041  U+D226  uniD226,  HANGUL SYLLABLE-D226
        0xD227, # G+17042  U+D227  uniD227,  HANGUL SYLLABLE-D227
        0xD228, # G+17043  U+D228  uniD228,  HANGUL SYLLABLE-D228
        0xD229, # G+17044  U+D229  uniD229,  HANGUL SYLLABLE-D229
        0xD22A, # G+17045  U+D22A  uniD22A,  HANGUL SYLLABLE-D22A
        0xD22B, # G+17046  U+D22B  uniD22B,  HANGUL SYLLABLE-D22B
        0xD22E, # G+17047  U+D22E  uniD22E,  HANGUL SYLLABLE-D22E
        0xD22F, # G+17048  U+D22F  uniD22F,  HANGUL SYLLABLE-D22F
        0xD231, # G+17049  U+D231  uniD231,  HANGUL SYLLABLE-D231
        0xD232, # G+17050  U+D232  uniD232,  HANGUL SYLLABLE-D232
        0xD233, # G+17051  U+D233  uniD233,  HANGUL SYLLABLE-D233
        0xD235, # G+17052  U+D235  uniD235,  HANGUL SYLLABLE-D235
        0xD236, # G+17053  U+D236  uniD236,  HANGUL SYLLABLE-D236
        0xD237, # G+17054  U+D237  uniD237,  HANGUL SYLLABLE-D237
        0xD238, # G+17055  U+D238  uniD238,  HANGUL SYLLABLE-D238
        0xD239, # G+17056  U+D239  uniD239,  HANGUL SYLLABLE-D239
        0xD23A, # G+17057  U+D23A  uniD23A,  HANGUL SYLLABLE-D23A
        0xD23B, # G+17058  U+D23B  uniD23B,  HANGUL SYLLABLE-D23B
        0xD23E, # G+17059  U+D23E  uniD23E,  HANGUL SYLLABLE-D23E
        0xD240, # G+17060  U+D240  uniD240,  HANGUL SYLLABLE-D240
        0xD242, # G+17061  U+D242  uniD242,  HANGUL SYLLABLE-D242
        0xD243, # G+17062  U+D243  uniD243,  HANGUL SYLLABLE-D243
        0xD244, # G+17063  U+D244  uniD244,  HANGUL SYLLABLE-D244
        0xD245, # G+17064  U+D245  uniD245,  HANGUL SYLLABLE-D245
        0xD246, # G+17065  U+D246  uniD246,  HANGUL SYLLABLE-D246
        0xD247, # G+17066  U+D247  uniD247,  HANGUL SYLLABLE-D247
        0xD249, # G+17067  U+D249  uniD249,  HANGUL SYLLABLE-D249
        0xD24A, # G+17068  U+D24A  uniD24A,  HANGUL SYLLABLE-D24A
        0xD24B, # G+17069  U+D24B  uniD24B,  HANGUL SYLLABLE-D24B
        0xD24C, # G+17070  U+D24C  uniD24C,  HANGUL SYLLABLE-D24C
        0xD24D, # G+17071  U+D24D  uniD24D,  HANGUL SYLLABLE-D24D
        0xD24E, # G+17072  U+D24E  uniD24E,  HANGUL SYLLABLE-D24E
        0xD24F, # G+17073  U+D24F  uniD24F,  HANGUL SYLLABLE-D24F
        0xD250, # G+17074  U+D250  uniD250,  HANGUL SYLLABLE-D250
        0xD251, # G+17075  U+D251  uniD251,  HANGUL SYLLABLE-D251
        0xD252, # G+17076  U+D252  uniD252,  HANGUL SYLLABLE-D252
        0xD253, # G+17077  U+D253  uniD253,  HANGUL SYLLABLE-D253
        0xD254, # G+17078  U+D254  uniD254,  HANGUL SYLLABLE-D254
        0xD255, # G+17079  U+D255  uniD255,  HANGUL SYLLABLE-D255
        0xD256, # G+17080  U+D256  uniD256,  HANGUL SYLLABLE-D256
        0xD257, # G+17081  U+D257  uniD257,  HANGUL SYLLABLE-D257
        0xD258, # G+17082  U+D258  uniD258,  HANGUL SYLLABLE-D258
        0xD259, # G+17083  U+D259  uniD259,  HANGUL SYLLABLE-D259
        0xD25A, # G+17084  U+D25A  uniD25A,  HANGUL SYLLABLE-D25A
        0xD25B, # G+17085  U+D25B  uniD25B,  HANGUL SYLLABLE-D25B
        0xD25D, # G+17086  U+D25D  uniD25D,  HANGUL SYLLABLE-D25D
        0xD25E, # G+17087  U+D25E  uniD25E,  HANGUL SYLLABLE-D25E
        0xD25F, # G+17088  U+D25F  uniD25F,  HANGUL SYLLABLE-D25F
        0xD260, # G+17089  U+D260  uniD260,  HANGUL SYLLABLE-D260
        0xD261, # G+17090  U+D261  uniD261,  HANGUL SYLLABLE-D261
        0xD262, # G+17091  U+D262  uniD262,  HANGUL SYLLABLE-D262
        0xD263, # G+17092  U+D263  uniD263,  HANGUL SYLLABLE-D263
        0xD265, # G+17093  U+D265  uniD265,  HANGUL SYLLABLE-D265
        0xD266, # G+17094  U+D266  uniD266,  HANGUL SYLLABLE-D266
        0xD267, # G+17095  U+D267  uniD267,  HANGUL SYLLABLE-D267
        0xD268, # G+17096  U+D268  uniD268,  HANGUL SYLLABLE-D268
        0xD269, # G+17097  U+D269  uniD269,  HANGUL SYLLABLE-D269
        0xD26A, # G+17098  U+D26A  uniD26A,  HANGUL SYLLABLE-D26A
        0xD26B, # G+17099  U+D26B  uniD26B,  HANGUL SYLLABLE-D26B
        0xD26C, # G+17100  U+D26C  uniD26C,  HANGUL SYLLABLE-D26C
        0xD26D, # G+17101  U+D26D  uniD26D,  HANGUL SYLLABLE-D26D
        0xD26E, # G+17102  U+D26E  uniD26E,  HANGUL SYLLABLE-D26E
        0xD26F, # G+17103  U+D26F  uniD26F,  HANGUL SYLLABLE-D26F
        0xD270, # G+17104  U+D270  uniD270,  HANGUL SYLLABLE-D270
        0xD271, # G+17105  U+D271  uniD271,  HANGUL SYLLABLE-D271
        0xD272, # G+17106  U+D272  uniD272,  HANGUL SYLLABLE-D272
        0xD273, # G+17107  U+D273  uniD273,  HANGUL SYLLABLE-D273
        0xD274, # G+17108  U+D274  uniD274,  HANGUL SYLLABLE-D274
        0xD275, # G+17109  U+D275  uniD275,  HANGUL SYLLABLE-D275
        0xD276, # G+17110  U+D276  uniD276,  HANGUL SYLLABLE-D276
        0xD277, # G+17111  U+D277  uniD277,  HANGUL SYLLABLE-D277
        0xD278, # G+17112  U+D278  uniD278,  HANGUL SYLLABLE-D278
        0xD279, # G+17113  U+D279  uniD279,  HANGUL SYLLABLE-D279
        0xD27A, # G+17114  U+D27A  uniD27A,  HANGUL SYLLABLE-D27A
        0xD27B, # G+17115  U+D27B  uniD27B,  HANGUL SYLLABLE-D27B
        0xD27C, # G+17116  U+D27C  uniD27C,  HANGUL SYLLABLE-D27C
        0xD27D, # G+17117  U+D27D  uniD27D,  HANGUL SYLLABLE-D27D
        0xD27E, # G+17118  U+D27E  uniD27E,  HANGUL SYLLABLE-D27E
        0xD27F, # G+17119  U+D27F  uniD27F,  HANGUL SYLLABLE-D27F
        0xD282, # G+17120  U+D282  uniD282,  HANGUL SYLLABLE-D282
        0xD283, # G+17121  U+D283  uniD283,  HANGUL SYLLABLE-D283
        0xD285, # G+17122  U+D285  uniD285,  HANGUL SYLLABLE-D285
        0xD286, # G+17123  U+D286  uniD286,  HANGUL SYLLABLE-D286
        0xD287, # G+17124  U+D287  uniD287,  HANGUL SYLLABLE-D287
        0xD289, # G+17125  U+D289  uniD289,  HANGUL SYLLABLE-D289
        0xD28A, # G+17126  U+D28A  uniD28A,  HANGUL SYLLABLE-D28A
        0xD28B, # G+17127  U+D28B  uniD28B,  HANGUL SYLLABLE-D28B
        0xD28C, # G+17128  U+D28C  uniD28C,  HANGUL SYLLABLE-D28C
        0xD28D, # G+17129  U+D28D  uniD28D,  HANGUL SYLLABLE-D28D
        0xD28E, # G+17130  U+D28E  uniD28E,  HANGUL SYLLABLE-D28E
        0xD28F, # G+17131  U+D28F  uniD28F,  HANGUL SYLLABLE-D28F
        0xD292, # G+17132  U+D292  uniD292,  HANGUL SYLLABLE-D292
        0xD293, # G+17133  U+D293  uniD293,  HANGUL SYLLABLE-D293
        0xD294, # G+17134  U+D294  uniD294,  HANGUL SYLLABLE-D294
        0xD296, # G+17135  U+D296  uniD296,  HANGUL SYLLABLE-D296
        0xD297, # G+17136  U+D297  uniD297,  HANGUL SYLLABLE-D297
        0xD298, # G+17137  U+D298  uniD298,  HANGUL SYLLABLE-D298
        0xD299, # G+17138  U+D299  uniD299,  HANGUL SYLLABLE-D299
        0xD29A, # G+17139  U+D29A  uniD29A,  HANGUL SYLLABLE-D29A
        0xD29B, # G+17140  U+D29B  uniD29B,  HANGUL SYLLABLE-D29B
        0xD29D, # G+17141  U+D29D  uniD29D,  HANGUL SYLLABLE-D29D
        0xD29E, # G+17142  U+D29E  uniD29E,  HANGUL SYLLABLE-D29E
        0xD29F, # G+17143  U+D29F  uniD29F,  HANGUL SYLLABLE-D29F
        0xD2A1, # G+17144  U+D2A1  uniD2A1,  HANGUL SYLLABLE-D2A1
        0xD2A2, # G+17145  U+D2A2  uniD2A2,  HANGUL SYLLABLE-D2A2
        0xD2A3, # G+17146  U+D2A3  uniD2A3,  HANGUL SYLLABLE-D2A3
        0xD2A5, # G+17147  U+D2A5  uniD2A5,  HANGUL SYLLABLE-D2A5
        0xD2A6, # G+17148  U+D2A6  uniD2A6,  HANGUL SYLLABLE-D2A6
        0xD2A7, # G+17149  U+D2A7  uniD2A7,  HANGUL SYLLABLE-D2A7
        0xD2A8, # G+17150  U+D2A8  uniD2A8,  HANGUL SYLLABLE-D2A8
        0xD2A9, # G+17151  U+D2A9  uniD2A9,  HANGUL SYLLABLE-D2A9
        0xD2AA, # G+17152  U+D2AA  uniD2AA,  HANGUL SYLLABLE-D2AA
        0xD2AB, # G+17153  U+D2AB  uniD2AB,  HANGUL SYLLABLE-D2AB
        0xD2AD, # G+17154  U+D2AD  uniD2AD,  HANGUL SYLLABLE-D2AD
        0xD2AE, # G+17155  U+D2AE  uniD2AE,  HANGUL SYLLABLE-D2AE
        0xD2AF, # G+17156  U+D2AF  uniD2AF,  HANGUL SYLLABLE-D2AF
        0xD2B0, # G+17157  U+D2B0  uniD2B0,  HANGUL SYLLABLE-D2B0
        0xD2B2, # G+17158  U+D2B2  uniD2B2,  HANGUL SYLLABLE-D2B2
        0xD2B3, # G+17159  U+D2B3  uniD2B3,  HANGUL SYLLABLE-D2B3
        0xD2B4, # G+17160  U+D2B4  uniD2B4,  HANGUL SYLLABLE-D2B4
        0xD2B5, # G+17161  U+D2B5  uniD2B5,  HANGUL SYLLABLE-D2B5
        0xD2B6, # G+17162  U+D2B6  uniD2B6,  HANGUL SYLLABLE-D2B6
        0xD2B7, # G+17163  U+D2B7  uniD2B7,  HANGUL SYLLABLE-D2B7
        0xD2BA, # G+17164  U+D2BA  uniD2BA,  HANGUL SYLLABLE-D2BA
        0xD2BB, # G+17165  U+D2BB  uniD2BB,  HANGUL SYLLABLE-D2BB
        0xD2BD, # G+17166  U+D2BD  uniD2BD,  HANGUL SYLLABLE-D2BD
        0xD2BE, # G+17167  U+D2BE  uniD2BE,  HANGUL SYLLABLE-D2BE
        0xD2C1, # G+17168  U+D2C1  uniD2C1,  HANGUL SYLLABLE-D2C1
        0xD2C3, # G+17169  U+D2C3  uniD2C3,  HANGUL SYLLABLE-D2C3
        0xD2C4, # G+17170  U+D2C4  uniD2C4,  HANGUL SYLLABLE-D2C4
        0xD2C5, # G+17171  U+D2C5  uniD2C5,  HANGUL SYLLABLE-D2C5
        0xD2C6, # G+17172  U+D2C6  uniD2C6,  HANGUL SYLLABLE-D2C6
        0xD2C7, # G+17173  U+D2C7  uniD2C7,  HANGUL SYLLABLE-D2C7
        0xD2CA, # G+17174  U+D2CA  uniD2CA,  HANGUL SYLLABLE-D2CA
        0xD2CC, # G+17175  U+D2CC  uniD2CC,  HANGUL SYLLABLE-D2CC
        0xD2CD, # G+17176  U+D2CD  uniD2CD,  HANGUL SYLLABLE-D2CD
        0xD2CE, # G+17177  U+D2CE  uniD2CE,  HANGUL SYLLABLE-D2CE
        0xD2CF, # G+17178  U+D2CF  uniD2CF,  HANGUL SYLLABLE-D2CF
        0xD2D0, # G+17179  U+D2D0  uniD2D0,  HANGUL SYLLABLE-D2D0
        0xD2D1, # G+17180  U+D2D1  uniD2D1,  HANGUL SYLLABLE-D2D1
        0xD2D2, # G+17181  U+D2D2  uniD2D2,  HANGUL SYLLABLE-D2D2
        0xD2D3, # G+17182  U+D2D3  uniD2D3,  HANGUL SYLLABLE-D2D3
        0xD2D5, # G+17183  U+D2D5  uniD2D5,  HANGUL SYLLABLE-D2D5
        0xD2D6, # G+17184  U+D2D6  uniD2D6,  HANGUL SYLLABLE-D2D6
        0xD2D7, # G+17185  U+D2D7  uniD2D7,  HANGUL SYLLABLE-D2D7
        0xD2D9, # G+17186  U+D2D9  uniD2D9,  HANGUL SYLLABLE-D2D9
        0xD2DA, # G+17187  U+D2DA  uniD2DA,  HANGUL SYLLABLE-D2DA
        0xD2DB, # G+17188  U+D2DB  uniD2DB,  HANGUL SYLLABLE-D2DB
        0xD2DD, # G+17189  U+D2DD  uniD2DD,  HANGUL SYLLABLE-D2DD
        0xD2DE, # G+17190  U+D2DE  uniD2DE,  HANGUL SYLLABLE-D2DE
        0xD2DF, # G+17191  U+D2DF  uniD2DF,  HANGUL SYLLABLE-D2DF
        0xD2E0, # G+17192  U+D2E0  uniD2E0,  HANGUL SYLLABLE-D2E0
        0xD2E1, # G+17193  U+D2E1  uniD2E1,  HANGUL SYLLABLE-D2E1
        0xD2E2, # G+17194  U+D2E2  uniD2E2,  HANGUL SYLLABLE-D2E2
        0xD2E3, # G+17195  U+D2E3  uniD2E3,  HANGUL SYLLABLE-D2E3
        0xD2E6, # G+17196  U+D2E6  uniD2E6,  HANGUL SYLLABLE-D2E6
        0xD2E7, # G+17197  U+D2E7  uniD2E7,  HANGUL SYLLABLE-D2E7
        0xD2E8, # G+17198  U+D2E8  uniD2E8,  HANGUL SYLLABLE-D2E8
        0xD2E9, # G+17199  U+D2E9  uniD2E9,  HANGUL SYLLABLE-D2E9
        0xD2EA, # G+17200  U+D2EA  uniD2EA,  HANGUL SYLLABLE-D2EA
        0xD2EB, # G+17201  U+D2EB  uniD2EB,  HANGUL SYLLABLE-D2EB
        0xD2EC, # G+17202  U+D2EC  uniD2EC,  HANGUL SYLLABLE-D2EC
        0xD2ED, # G+17203  U+D2ED  uniD2ED,  HANGUL SYLLABLE-D2ED
        0xD2EE, # G+17204  U+D2EE  uniD2EE,  HANGUL SYLLABLE-D2EE
        0xD2EF, # G+17205  U+D2EF  uniD2EF,  HANGUL SYLLABLE-D2EF
        0xD2F2, # G+17206  U+D2F2  uniD2F2,  HANGUL SYLLABLE-D2F2
        0xD2F3, # G+17207  U+D2F3  uniD2F3,  HANGUL SYLLABLE-D2F3
        0xD2F5, # G+17208  U+D2F5  uniD2F5,  HANGUL SYLLABLE-D2F5
        0xD2F6, # G+17209  U+D2F6  uniD2F6,  HANGUL SYLLABLE-D2F6
        0xD2F7, # G+17210  U+D2F7  uniD2F7,  HANGUL SYLLABLE-D2F7
        0xD2F9, # G+17211  U+D2F9  uniD2F9,  HANGUL SYLLABLE-D2F9
        0xD2FA, # G+17212  U+D2FA  uniD2FA,  HANGUL SYLLABLE-D2FA
        0xD2FB, # G+17213  U+D2FB  uniD2FB,  HANGUL SYLLABLE-D2FB
        0xD2FC, # G+17214  U+D2FC  uniD2FC,  HANGUL SYLLABLE-D2FC
        0xD2FD, # G+17215  U+D2FD  uniD2FD,  HANGUL SYLLABLE-D2FD
        0xD2FE, # G+17216  U+D2FE  uniD2FE,  HANGUL SYLLABLE-D2FE
        0xD2FF, # G+17217  U+D2FF  uniD2FF,  HANGUL SYLLABLE-D2FF
        0xD302, # G+17218  U+D302  uniD302,  HANGUL SYLLABLE-D302
        0xD304, # G+17219  U+D304  uniD304,  HANGUL SYLLABLE-D304
        0xD306, # G+17220  U+D306  uniD306,  HANGUL SYLLABLE-D306
        0xD307, # G+17221  U+D307  uniD307,  HANGUL SYLLABLE-D307
        0xD308, # G+17222  U+D308  uniD308,  HANGUL SYLLABLE-D308
        0xD309, # G+17223  U+D309  uniD309,  HANGUL SYLLABLE-D309
        0xD30A, # G+17224  U+D30A  uniD30A,  HANGUL SYLLABLE-D30A
        0xD30B, # G+17225  U+D30B  uniD30B,  HANGUL SYLLABLE-D30B
        0xD30F, # G+17226  U+D30F  uniD30F,  HANGUL SYLLABLE-D30F
        0xD311, # G+17227  U+D311  uniD311,  HANGUL SYLLABLE-D311
        0xD312, # G+17228  U+D312  uniD312,  HANGUL SYLLABLE-D312
        0xD313, # G+17229  U+D313  uniD313,  HANGUL SYLLABLE-D313
        0xD315, # G+17230  U+D315  uniD315,  HANGUL SYLLABLE-D315
        0xD317, # G+17231  U+D317  uniD317,  HANGUL SYLLABLE-D317
        0xD318, # G+17232  U+D318  uniD318,  HANGUL SYLLABLE-D318
        0xD319, # G+17233  U+D319  uniD319,  HANGUL SYLLABLE-D319
        0xD31A, # G+17234  U+D31A  uniD31A,  HANGUL SYLLABLE-D31A
        0xD31B, # G+17235  U+D31B  uniD31B,  HANGUL SYLLABLE-D31B
        0xD31E, # G+17236  U+D31E  uniD31E,  HANGUL SYLLABLE-D31E
        0xD322, # G+17237  U+D322  uniD322,  HANGUL SYLLABLE-D322
        0xD323, # G+17238  U+D323  uniD323,  HANGUL SYLLABLE-D323
        0xD324, # G+17239  U+D324  uniD324,  HANGUL SYLLABLE-D324
        0xD326, # G+17240  U+D326  uniD326,  HANGUL SYLLABLE-D326
        0xD327, # G+17241  U+D327  uniD327,  HANGUL SYLLABLE-D327
        0xD32A, # G+17242  U+D32A  uniD32A,  HANGUL SYLLABLE-D32A
        0xD32B, # G+17243  U+D32B  uniD32B,  HANGUL SYLLABLE-D32B
        0xD32D, # G+17244  U+D32D  uniD32D,  HANGUL SYLLABLE-D32D
        0xD32E, # G+17245  U+D32E  uniD32E,  HANGUL SYLLABLE-D32E
        0xD32F, # G+17246  U+D32F  uniD32F,  HANGUL SYLLABLE-D32F
        0xD331, # G+17247  U+D331  uniD331,  HANGUL SYLLABLE-D331
        0xD332, # G+17248  U+D332  uniD332,  HANGUL SYLLABLE-D332
        0xD333, # G+17249  U+D333  uniD333,  HANGUL SYLLABLE-D333
        0xD334, # G+17250  U+D334  uniD334,  HANGUL SYLLABLE-D334
        0xD335, # G+17251  U+D335  uniD335,  HANGUL SYLLABLE-D335
        0xD336, # G+17252  U+D336  uniD336,  HANGUL SYLLABLE-D336
        0xD337, # G+17253  U+D337  uniD337,  HANGUL SYLLABLE-D337
        0xD33A, # G+17254  U+D33A  uniD33A,  HANGUL SYLLABLE-D33A
        0xD33E, # G+17255  U+D33E  uniD33E,  HANGUL SYLLABLE-D33E
        0xD33F, # G+17256  U+D33F  uniD33F,  HANGUL SYLLABLE-D33F
        0xD340, # G+17257  U+D340  uniD340,  HANGUL SYLLABLE-D340
        0xD341, # G+17258  U+D341  uniD341,  HANGUL SYLLABLE-D341
        0xD342, # G+17259  U+D342  uniD342,  HANGUL SYLLABLE-D342
        0xD343, # G+17260  U+D343  uniD343,  HANGUL SYLLABLE-D343
        0xD346, # G+17261  U+D346  uniD346,  HANGUL SYLLABLE-D346
        0xD347, # G+17262  U+D347  uniD347,  HANGUL SYLLABLE-D347
        0xD348, # G+17263  U+D348  uniD348,  HANGUL SYLLABLE-D348
        0xD349, # G+17264  U+D349  uniD349,  HANGUL SYLLABLE-D349
        0xD34A, # G+17265  U+D34A  uniD34A,  HANGUL SYLLABLE-D34A
        0xD34B, # G+17266  U+D34B  uniD34B,  HANGUL SYLLABLE-D34B
        0xD34C, # G+17267  U+D34C  uniD34C,  HANGUL SYLLABLE-D34C
        0xD34D, # G+17268  U+D34D  uniD34D,  HANGUL SYLLABLE-D34D
        0xD34E, # G+17269  U+D34E  uniD34E,  HANGUL SYLLABLE-D34E
        0xD34F, # G+17270  U+D34F  uniD34F,  HANGUL SYLLABLE-D34F
        0xD350, # G+17271  U+D350  uniD350,  HANGUL SYLLABLE-D350
        0xD351, # G+17272  U+D351  uniD351,  HANGUL SYLLABLE-D351
        0xD352, # G+17273  U+D352  uniD352,  HANGUL SYLLABLE-D352
        0xD353, # G+17274  U+D353  uniD353,  HANGUL SYLLABLE-D353
        0xD354, # G+17275  U+D354  uniD354,  HANGUL SYLLABLE-D354
        0xD355, # G+17276  U+D355  uniD355,  HANGUL SYLLABLE-D355
        0xD356, # G+17277  U+D356  uniD356,  HANGUL SYLLABLE-D356
        0xD357, # G+17278  U+D357  uniD357,  HANGUL SYLLABLE-D357
        0xD358, # G+17279  U+D358  uniD358,  HANGUL SYLLABLE-D358
        0xD359, # G+17280  U+D359  uniD359,  HANGUL SYLLABLE-D359
        0xD35A, # G+17281  U+D35A  uniD35A,  HANGUL SYLLABLE-D35A
        0xD35B, # G+17282  U+D35B  uniD35B,  HANGUL SYLLABLE-D35B
        0xD35C, # G+17283  U+D35C  uniD35C,  HANGUL SYLLABLE-D35C
        0xD35D, # G+17284  U+D35D  uniD35D,  HANGUL SYLLABLE-D35D
        0xD35E, # G+17285  U+D35E  uniD35E,  HANGUL SYLLABLE-D35E
        0xD35F, # G+17286  U+D35F  uniD35F,  HANGUL SYLLABLE-D35F
        0xD360, # G+17287  U+D360  uniD360,  HANGUL SYLLABLE-D360
        0xD361, # G+17288  U+D361  uniD361,  HANGUL SYLLABLE-D361
        0xD362, # G+17289  U+D362  uniD362,  HANGUL SYLLABLE-D362
        0xD363, # G+17290  U+D363  uniD363,  HANGUL SYLLABLE-D363
        0xD364, # G+17291  U+D364  uniD364,  HANGUL SYLLABLE-D364
        0xD365, # G+17292  U+D365  uniD365,  HANGUL SYLLABLE-D365
        0xD366, # G+17293  U+D366  uniD366,  HANGUL SYLLABLE-D366
        0xD367, # G+17294  U+D367  uniD367,  HANGUL SYLLABLE-D367
        0xD368, # G+17295  U+D368  uniD368,  HANGUL SYLLABLE-D368
        0xD369, # G+17296  U+D369  uniD369,  HANGUL SYLLABLE-D369
        0xD36A, # G+17297  U+D36A  uniD36A,  HANGUL SYLLABLE-D36A
        0xD36B, # G+17298  U+D36B  uniD36B,  HANGUL SYLLABLE-D36B
        0xD36C, # G+17299  U+D36C  uniD36C,  HANGUL SYLLABLE-D36C
        0xD36D, # G+17300  U+D36D  uniD36D,  HANGUL SYLLABLE-D36D
        0xD36E, # G+17301  U+D36E  uniD36E,  HANGUL SYLLABLE-D36E
        0xD36F, # G+17302  U+D36F  uniD36F,  HANGUL SYLLABLE-D36F
        0xD370, # G+17303  U+D370  uniD370,  HANGUL SYLLABLE-D370
        0xD371, # G+17304  U+D371  uniD371,  HANGUL SYLLABLE-D371
        0xD372, # G+17305  U+D372  uniD372,  HANGUL SYLLABLE-D372
        0xD373, # G+17306  U+D373  uniD373,  HANGUL SYLLABLE-D373
        0xD374, # G+17307  U+D374  uniD374,  HANGUL SYLLABLE-D374
        0xD375, # G+17308  U+D375  uniD375,  HANGUL SYLLABLE-D375
        0xD376, # G+17309  U+D376  uniD376,  HANGUL SYLLABLE-D376
        0xD377, # G+17310  U+D377  uniD377,  HANGUL SYLLABLE-D377
        0xD378, # G+17311  U+D378  uniD378,  HANGUL SYLLABLE-D378
        0xD379, # G+17312  U+D379  uniD379,  HANGUL SYLLABLE-D379
        0xD37A, # G+17313  U+D37A  uniD37A,  HANGUL SYLLABLE-D37A
        0xD37B, # G+17314  U+D37B  uniD37B,  HANGUL SYLLABLE-D37B
        0xD37E, # G+17315  U+D37E  uniD37E,  HANGUL SYLLABLE-D37E
        0xD37F, # G+17316  U+D37F  uniD37F,  HANGUL SYLLABLE-D37F
        0xD381, # G+17317  U+D381  uniD381,  HANGUL SYLLABLE-D381
        0xD382, # G+17318  U+D382  uniD382,  HANGUL SYLLABLE-D382
        0xD383, # G+17319  U+D383  uniD383,  HANGUL SYLLABLE-D383
        0xD385, # G+17320  U+D385  uniD385,  HANGUL SYLLABLE-D385
        0xD386, # G+17321  U+D386  uniD386,  HANGUL SYLLABLE-D386
        0xD387, # G+17322  U+D387  uniD387,  HANGUL SYLLABLE-D387
        0xD388, # G+17323  U+D388  uniD388,  HANGUL SYLLABLE-D388
        0xD389, # G+17324  U+D389  uniD389,  HANGUL SYLLABLE-D389
        0xD38A, # G+17325  U+D38A  uniD38A,  HANGUL SYLLABLE-D38A
        0xD38B, # G+17326  U+D38B  uniD38B,  HANGUL SYLLABLE-D38B
        0xD38E, # G+17327  U+D38E  uniD38E,  HANGUL SYLLABLE-D38E
        0xD392, # G+17328  U+D392  uniD392,  HANGUL SYLLABLE-D392
        0xD393, # G+17329  U+D393  uniD393,  HANGUL SYLLABLE-D393
        0xD394, # G+17330  U+D394  uniD394,  HANGUL SYLLABLE-D394
        0xD395, # G+17331  U+D395  uniD395,  HANGUL SYLLABLE-D395
        0xD396, # G+17332  U+D396  uniD396,  HANGUL SYLLABLE-D396
        0xD397, # G+17333  U+D397  uniD397,  HANGUL SYLLABLE-D397
        0xD39A, # G+17334  U+D39A  uniD39A,  HANGUL SYLLABLE-D39A
        0xD39B, # G+17335  U+D39B  uniD39B,  HANGUL SYLLABLE-D39B
        0xD39D, # G+17336  U+D39D  uniD39D,  HANGUL SYLLABLE-D39D
        0xD39E, # G+17337  U+D39E  uniD39E,  HANGUL SYLLABLE-D39E
        0xD39F, # G+17338  U+D39F  uniD39F,  HANGUL SYLLABLE-D39F
        0xD3A1, # G+17339  U+D3A1  uniD3A1,  HANGUL SYLLABLE-D3A1
        0xD3A2, # G+17340  U+D3A2  uniD3A2,  HANGUL SYLLABLE-D3A2
        0xD3A3, # G+17341  U+D3A3  uniD3A3,  HANGUL SYLLABLE-D3A3
        0xD3A4, # G+17342  U+D3A4  uniD3A4,  HANGUL SYLLABLE-D3A4
        0xD3A5, # G+17343  U+D3A5  uniD3A5,  HANGUL SYLLABLE-D3A5
        0xD3A6, # G+17344  U+D3A6  uniD3A6,  HANGUL SYLLABLE-D3A6
        0xD3A7, # G+17345  U+D3A7  uniD3A7,  HANGUL SYLLABLE-D3A7
        0xD3AA, # G+17346  U+D3AA  uniD3AA,  HANGUL SYLLABLE-D3AA
        0xD3AC, # G+17347  U+D3AC  uniD3AC,  HANGUL SYLLABLE-D3AC
        0xD3AE, # G+17348  U+D3AE  uniD3AE,  HANGUL SYLLABLE-D3AE
        0xD3AF, # G+17349  U+D3AF  uniD3AF,  HANGUL SYLLABLE-D3AF
        0xD3B0, # G+17350  U+D3B0  uniD3B0,  HANGUL SYLLABLE-D3B0
        0xD3B1, # G+17351  U+D3B1  uniD3B1,  HANGUL SYLLABLE-D3B1
        0xD3B2, # G+17352  U+D3B2  uniD3B2,  HANGUL SYLLABLE-D3B2
        0xD3B3, # G+17353  U+D3B3  uniD3B3,  HANGUL SYLLABLE-D3B3
        0xD3B5, # G+17354  U+D3B5  uniD3B5,  HANGUL SYLLABLE-D3B5
        0xD3B6, # G+17355  U+D3B6  uniD3B6,  HANGUL SYLLABLE-D3B6
        0xD3B7, # G+17356  U+D3B7  uniD3B7,  HANGUL SYLLABLE-D3B7
        0xD3B9, # G+17357  U+D3B9  uniD3B9,  HANGUL SYLLABLE-D3B9
        0xD3BA, # G+17358  U+D3BA  uniD3BA,  HANGUL SYLLABLE-D3BA
        0xD3BB, # G+17359  U+D3BB  uniD3BB,  HANGUL SYLLABLE-D3BB
        0xD3BD, # G+17360  U+D3BD  uniD3BD,  HANGUL SYLLABLE-D3BD
        0xD3BE, # G+17361  U+D3BE  uniD3BE,  HANGUL SYLLABLE-D3BE
        0xD3BF, # G+17362  U+D3BF  uniD3BF,  HANGUL SYLLABLE-D3BF
        0xD3C0, # G+17363  U+D3C0  uniD3C0,  HANGUL SYLLABLE-D3C0
        0xD3C1, # G+17364  U+D3C1  uniD3C1,  HANGUL SYLLABLE-D3C1
        0xD3C2, # G+17365  U+D3C2  uniD3C2,  HANGUL SYLLABLE-D3C2
        0xD3C3, # G+17366  U+D3C3  uniD3C3,  HANGUL SYLLABLE-D3C3
        0xD3C6, # G+17367  U+D3C6  uniD3C6,  HANGUL SYLLABLE-D3C6
        0xD3C7, # G+17368  U+D3C7  uniD3C7,  HANGUL SYLLABLE-D3C7
        0xD3CA, # G+17369  U+D3CA  uniD3CA,  HANGUL SYLLABLE-D3CA
        0xD3CB, # G+17370  U+D3CB  uniD3CB,  HANGUL SYLLABLE-D3CB
        0xD3CC, # G+17371  U+D3CC  uniD3CC,  HANGUL SYLLABLE-D3CC
        0xD3CD, # G+17372  U+D3CD  uniD3CD,  HANGUL SYLLABLE-D3CD
        0xD3CE, # G+17373  U+D3CE  uniD3CE,  HANGUL SYLLABLE-D3CE
        0xD3CF, # G+17374  U+D3CF  uniD3CF,  HANGUL SYLLABLE-D3CF
        0xD3D1, # G+17375  U+D3D1  uniD3D1,  HANGUL SYLLABLE-D3D1
        0xD3D2, # G+17376  U+D3D2  uniD3D2,  HANGUL SYLLABLE-D3D2
        0xD3D3, # G+17377  U+D3D3  uniD3D3,  HANGUL SYLLABLE-D3D3
        0xD3D4, # G+17378  U+D3D4  uniD3D4,  HANGUL SYLLABLE-D3D4
        0xD3D5, # G+17379  U+D3D5  uniD3D5,  HANGUL SYLLABLE-D3D5
        0xD3D6, # G+17380  U+D3D6  uniD3D6,  HANGUL SYLLABLE-D3D6
        0xD3D7, # G+17381  U+D3D7  uniD3D7,  HANGUL SYLLABLE-D3D7
        0xD3D9, # G+17382  U+D3D9  uniD3D9,  HANGUL SYLLABLE-D3D9
        0xD3DA, # G+17383  U+D3DA  uniD3DA,  HANGUL SYLLABLE-D3DA
        0xD3DB, # G+17384  U+D3DB  uniD3DB,  HANGUL SYLLABLE-D3DB
        0xD3DC, # G+17385  U+D3DC  uniD3DC,  HANGUL SYLLABLE-D3DC
        0xD3DD, # G+17386  U+D3DD  uniD3DD,  HANGUL SYLLABLE-D3DD
        0xD3DE, # G+17387  U+D3DE  uniD3DE,  HANGUL SYLLABLE-D3DE
        0xD3DF, # G+17388  U+D3DF  uniD3DF,  HANGUL SYLLABLE-D3DF
        0xD3E0, # G+17389  U+D3E0  uniD3E0,  HANGUL SYLLABLE-D3E0
        0xD3E2, # G+17390  U+D3E2  uniD3E2,  HANGUL SYLLABLE-D3E2
        0xD3E4, # G+17391  U+D3E4  uniD3E4,  HANGUL SYLLABLE-D3E4
        0xD3E5, # G+17392  U+D3E5  uniD3E5,  HANGUL SYLLABLE-D3E5
        0xD3E6, # G+17393  U+D3E6  uniD3E6,  HANGUL SYLLABLE-D3E6
        0xD3E7, # G+17394  U+D3E7  uniD3E7,  HANGUL SYLLABLE-D3E7
        0xD3E8, # G+17395  U+D3E8  uniD3E8,  HANGUL SYLLABLE-D3E8
        0xD3E9, # G+17396  U+D3E9  uniD3E9,  HANGUL SYLLABLE-D3E9
        0xD3EA, # G+17397  U+D3EA  uniD3EA,  HANGUL SYLLABLE-D3EA
        0xD3EB, # G+17398  U+D3EB  uniD3EB,  HANGUL SYLLABLE-D3EB
        0xD3EE, # G+17399  U+D3EE  uniD3EE,  HANGUL SYLLABLE-D3EE
        0xD3EF, # G+17400  U+D3EF  uniD3EF,  HANGUL SYLLABLE-D3EF
        0xD3F1, # G+17401  U+D3F1  uniD3F1,  HANGUL SYLLABLE-D3F1
        0xD3F2, # G+17402  U+D3F2  uniD3F2,  HANGUL SYLLABLE-D3F2
        0xD3F3, # G+17403  U+D3F3  uniD3F3,  HANGUL SYLLABLE-D3F3
        0xD3F5, # G+17404  U+D3F5  uniD3F5,  HANGUL SYLLABLE-D3F5
        0xD3F6, # G+17405  U+D3F6  uniD3F6,  HANGUL SYLLABLE-D3F6
        0xD3F7, # G+17406  U+D3F7  uniD3F7,  HANGUL SYLLABLE-D3F7
        0xD3F8, # G+17407  U+D3F8  uniD3F8,  HANGUL SYLLABLE-D3F8
        0xD3F9, # G+17408  U+D3F9  uniD3F9,  HANGUL SYLLABLE-D3F9
        0xD3FA, # G+17409  U+D3FA  uniD3FA,  HANGUL SYLLABLE-D3FA
        0xD3FB, # G+17410  U+D3FB  uniD3FB,  HANGUL SYLLABLE-D3FB
        0xD3FE, # G+17411  U+D3FE  uniD3FE,  HANGUL SYLLABLE-D3FE
        0xD400, # G+17412  U+D400  uniD400,  HANGUL SYLLABLE-D400
        0xD402, # G+17413  U+D402  uniD402,  HANGUL SYLLABLE-D402
        0xD403, # G+17414  U+D403  uniD403,  HANGUL SYLLABLE-D403
        0xD404, # G+17415  U+D404  uniD404,  HANGUL SYLLABLE-D404
        0xD405, # G+17416  U+D405  uniD405,  HANGUL SYLLABLE-D405
        0xD406, # G+17417  U+D406  uniD406,  HANGUL SYLLABLE-D406
        0xD407, # G+17418  U+D407  uniD407,  HANGUL SYLLABLE-D407
        0xD409, # G+17419  U+D409  uniD409,  HANGUL SYLLABLE-D409
        0xD40A, # G+17420  U+D40A  uniD40A,  HANGUL SYLLABLE-D40A
        0xD40B, # G+17421  U+D40B  uniD40B,  HANGUL SYLLABLE-D40B
        0xD40C, # G+17422  U+D40C  uniD40C,  HANGUL SYLLABLE-D40C
        0xD40D, # G+17423  U+D40D  uniD40D,  HANGUL SYLLABLE-D40D
        0xD40E, # G+17424  U+D40E  uniD40E,  HANGUL SYLLABLE-D40E
        0xD40F, # G+17425  U+D40F  uniD40F,  HANGUL SYLLABLE-D40F
        0xD410, # G+17426  U+D410  uniD410,  HANGUL SYLLABLE-D410
        0xD411, # G+17427  U+D411  uniD411,  HANGUL SYLLABLE-D411
        0xD412, # G+17428  U+D412  uniD412,  HANGUL SYLLABLE-D412
        0xD413, # G+17429  U+D413  uniD413,  HANGUL SYLLABLE-D413
        0xD414, # G+17430  U+D414  uniD414,  HANGUL SYLLABLE-D414
        0xD415, # G+17431  U+D415  uniD415,  HANGUL SYLLABLE-D415
        0xD416, # G+17432  U+D416  uniD416,  HANGUL SYLLABLE-D416
        0xD417, # G+17433  U+D417  uniD417,  HANGUL SYLLABLE-D417
        0xD418, # G+17434  U+D418  uniD418,  HANGUL SYLLABLE-D418
        0xD419, # G+17435  U+D419  uniD419,  HANGUL SYLLABLE-D419
        0xD41A, # G+17436  U+D41A  uniD41A,  HANGUL SYLLABLE-D41A
        0xD41B, # G+17437  U+D41B  uniD41B,  HANGUL SYLLABLE-D41B
        0xD41C, # G+17438  U+D41C  uniD41C,  HANGUL SYLLABLE-D41C
        0xD41E, # G+17439  U+D41E  uniD41E,  HANGUL SYLLABLE-D41E
        0xD41F, # G+17440  U+D41F  uniD41F,  HANGUL SYLLABLE-D41F
        0xD420, # G+17441  U+D420  uniD420,  HANGUL SYLLABLE-D420
        0xD421, # G+17442  U+D421  uniD421,  HANGUL SYLLABLE-D421
        0xD422, # G+17443  U+D422  uniD422,  HANGUL SYLLABLE-D422
        0xD423, # G+17444  U+D423  uniD423,  HANGUL SYLLABLE-D423
        0xD424, # G+17445  U+D424  uniD424,  HANGUL SYLLABLE-D424
        0xD425, # G+17446  U+D425  uniD425,  HANGUL SYLLABLE-D425
        0xD426, # G+17447  U+D426  uniD426,  HANGUL SYLLABLE-D426
        0xD427, # G+17448  U+D427  uniD427,  HANGUL SYLLABLE-D427
        0xD428, # G+17449  U+D428  uniD428,  HANGUL SYLLABLE-D428
        0xD429, # G+17450  U+D429  uniD429,  HANGUL SYLLABLE-D429
        0xD42A, # G+17451  U+D42A  uniD42A,  HANGUL SYLLABLE-D42A
        0xD42B, # G+17452  U+D42B  uniD42B,  HANGUL SYLLABLE-D42B
        0xD42C, # G+17453  U+D42C  uniD42C,  HANGUL SYLLABLE-D42C
        0xD42D, # G+17454  U+D42D  uniD42D,  HANGUL SYLLABLE-D42D
        0xD42E, # G+17455  U+D42E  uniD42E,  HANGUL SYLLABLE-D42E
        0xD42F, # G+17456  U+D42F  uniD42F,  HANGUL SYLLABLE-D42F
        0xD430, # G+17457  U+D430  uniD430,  HANGUL SYLLABLE-D430
        0xD431, # G+17458  U+D431  uniD431,  HANGUL SYLLABLE-D431
        0xD432, # G+17459  U+D432  uniD432,  HANGUL SYLLABLE-D432
        0xD433, # G+17460  U+D433  uniD433,  HANGUL SYLLABLE-D433
        0xD434, # G+17461  U+D434  uniD434,  HANGUL SYLLABLE-D434
        0xD435, # G+17462  U+D435  uniD435,  HANGUL SYLLABLE-D435
        0xD436, # G+17463  U+D436  uniD436,  HANGUL SYLLABLE-D436
        0xD437, # G+17464  U+D437  uniD437,  HANGUL SYLLABLE-D437
        0xD438, # G+17465  U+D438  uniD438,  HANGUL SYLLABLE-D438
        0xD439, # G+17466  U+D439  uniD439,  HANGUL SYLLABLE-D439
        0xD43A, # G+17467  U+D43A  uniD43A,  HANGUL SYLLABLE-D43A
        0xD43B, # G+17468  U+D43B  uniD43B,  HANGUL SYLLABLE-D43B
        0xD43C, # G+17469  U+D43C  uniD43C,  HANGUL SYLLABLE-D43C
        0xD43D, # G+17470  U+D43D  uniD43D,  HANGUL SYLLABLE-D43D
        0xD43E, # G+17471  U+D43E  uniD43E,  HANGUL SYLLABLE-D43E
        0xD43F, # G+17472  U+D43F  uniD43F,  HANGUL SYLLABLE-D43F
        0xD441, # G+17473  U+D441  uniD441,  HANGUL SYLLABLE-D441
        0xD442, # G+17474  U+D442  uniD442,  HANGUL SYLLABLE-D442
        0xD443, # G+17475  U+D443  uniD443,  HANGUL SYLLABLE-D443
        0xD445, # G+17476  U+D445  uniD445,  HANGUL SYLLABLE-D445
        0xD446, # G+17477  U+D446  uniD446,  HANGUL SYLLABLE-D446
        0xD447, # G+17478  U+D447  uniD447,  HANGUL SYLLABLE-D447
        0xD448, # G+17479  U+D448  uniD448,  HANGUL SYLLABLE-D448
        0xD449, # G+17480  U+D449  uniD449,  HANGUL SYLLABLE-D449
        0xD44A, # G+17481  U+D44A  uniD44A,  HANGUL SYLLABLE-D44A
        0xD44B, # G+17482  U+D44B  uniD44B,  HANGUL SYLLABLE-D44B
        0xD44C, # G+17483  U+D44C  uniD44C,  HANGUL SYLLABLE-D44C
        0xD44D, # G+17484  U+D44D  uniD44D,  HANGUL SYLLABLE-D44D
        0xD44E, # G+17485  U+D44E  uniD44E,  HANGUL SYLLABLE-D44E
        0xD44F, # G+17486  U+D44F  uniD44F,  HANGUL SYLLABLE-D44F
        0xD450, # G+17487  U+D450  uniD450,  HANGUL SYLLABLE-D450
        0xD451, # G+17488  U+D451  uniD451,  HANGUL SYLLABLE-D451
        0xD452, # G+17489  U+D452  uniD452,  HANGUL SYLLABLE-D452
        0xD453, # G+17490  U+D453  uniD453,  HANGUL SYLLABLE-D453
        0xD454, # G+17491  U+D454  uniD454,  HANGUL SYLLABLE-D454
        0xD455, # G+17492  U+D455  uniD455,  HANGUL SYLLABLE-D455
        0xD456, # G+17493  U+D456  uniD456,  HANGUL SYLLABLE-D456
        0xD457, # G+17494  U+D457  uniD457,  HANGUL SYLLABLE-D457
        0xD458, # G+17495  U+D458  uniD458,  HANGUL SYLLABLE-D458
        0xD459, # G+17496  U+D459  uniD459,  HANGUL SYLLABLE-D459
        0xD45A, # G+17497  U+D45A  uniD45A,  HANGUL SYLLABLE-D45A
        0xD45B, # G+17498  U+D45B  uniD45B,  HANGUL SYLLABLE-D45B
        0xD45D, # G+17499  U+D45D  uniD45D,  HANGUL SYLLABLE-D45D
        0xD45E, # G+17500  U+D45E  uniD45E,  HANGUL SYLLABLE-D45E
        0xD45F, # G+17501  U+D45F  uniD45F,  HANGUL SYLLABLE-D45F
        0xD461, # G+17502  U+D461  uniD461,  HANGUL SYLLABLE-D461
        0xD462, # G+17503  U+D462  uniD462,  HANGUL SYLLABLE-D462
        0xD463, # G+17504  U+D463  uniD463,  HANGUL SYLLABLE-D463
        0xD465, # G+17505  U+D465  uniD465,  HANGUL SYLLABLE-D465
        0xD466, # G+17506  U+D466  uniD466,  HANGUL SYLLABLE-D466
        0xD467, # G+17507  U+D467  uniD467,  HANGUL SYLLABLE-D467
        0xD468, # G+17508  U+D468  uniD468,  HANGUL SYLLABLE-D468
        0xD469, # G+17509  U+D469  uniD469,  HANGUL SYLLABLE-D469
        0xD46A, # G+17510  U+D46A  uniD46A,  HANGUL SYLLABLE-D46A
        0xD46B, # G+17511  U+D46B  uniD46B,  HANGUL SYLLABLE-D46B
        0xD46C, # G+17512  U+D46C  uniD46C,  HANGUL SYLLABLE-D46C
        0xD46E, # G+17513  U+D46E  uniD46E,  HANGUL SYLLABLE-D46E
        0xD470, # G+17514  U+D470  uniD470,  HANGUL SYLLABLE-D470
        0xD471, # G+17515  U+D471  uniD471,  HANGUL SYLLABLE-D471
        0xD472, # G+17516  U+D472  uniD472,  HANGUL SYLLABLE-D472
        0xD473, # G+17517  U+D473  uniD473,  HANGUL SYLLABLE-D473
        0xD474, # G+17518  U+D474  uniD474,  HANGUL SYLLABLE-D474
        0xD475, # G+17519  U+D475  uniD475,  HANGUL SYLLABLE-D475
        0xD476, # G+17520  U+D476  uniD476,  HANGUL SYLLABLE-D476
        0xD477, # G+17521  U+D477  uniD477,  HANGUL SYLLABLE-D477
        0xD47A, # G+17522  U+D47A  uniD47A,  HANGUL SYLLABLE-D47A
        0xD47B, # G+17523  U+D47B  uniD47B,  HANGUL SYLLABLE-D47B
        0xD47D, # G+17524  U+D47D  uniD47D,  HANGUL SYLLABLE-D47D
        0xD47E, # G+17525  U+D47E  uniD47E,  HANGUL SYLLABLE-D47E
        0xD481, # G+17526  U+D481  uniD481,  HANGUL SYLLABLE-D481
        0xD483, # G+17527  U+D483  uniD483,  HANGUL SYLLABLE-D483
        0xD484, # G+17528  U+D484  uniD484,  HANGUL SYLLABLE-D484
        0xD485, # G+17529  U+D485  uniD485,  HANGUL SYLLABLE-D485
        0xD486, # G+17530  U+D486  uniD486,  HANGUL SYLLABLE-D486
        0xD487, # G+17531  U+D487  uniD487,  HANGUL SYLLABLE-D487
        0xD48A, # G+17532  U+D48A  uniD48A,  HANGUL SYLLABLE-D48A
        0xD48C, # G+17533  U+D48C  uniD48C,  HANGUL SYLLABLE-D48C
        0xD48E, # G+17534  U+D48E  uniD48E,  HANGUL SYLLABLE-D48E
        0xD48F, # G+17535  U+D48F  uniD48F,  HANGUL SYLLABLE-D48F
        0xD490, # G+17536  U+D490  uniD490,  HANGUL SYLLABLE-D490
        0xD491, # G+17537  U+D491  uniD491,  HANGUL SYLLABLE-D491
        0xD492, # G+17538  U+D492  uniD492,  HANGUL SYLLABLE-D492
        0xD493, # G+17539  U+D493  uniD493,  HANGUL SYLLABLE-D493
        0xD495, # G+17540  U+D495  uniD495,  HANGUL SYLLABLE-D495
        0xD496, # G+17541  U+D496  uniD496,  HANGUL SYLLABLE-D496
        0xD497, # G+17542  U+D497  uniD497,  HANGUL SYLLABLE-D497
        0xD498, # G+17543  U+D498  uniD498,  HANGUL SYLLABLE-D498
        0xD499, # G+17544  U+D499  uniD499,  HANGUL SYLLABLE-D499
        0xD49A, # G+17545  U+D49A  uniD49A,  HANGUL SYLLABLE-D49A
        0xD49B, # G+17546  U+D49B  uniD49B,  HANGUL SYLLABLE-D49B
        0xD49C, # G+17547  U+D49C  uniD49C,  HANGUL SYLLABLE-D49C
        0xD49D, # G+17548  U+D49D  uniD49D,  HANGUL SYLLABLE-D49D
        0xD49E, # G+17549  U+D49E  uniD49E,  HANGUL SYLLABLE-D49E
        0xD49F, # G+17550  U+D49F  uniD49F,  HANGUL SYLLABLE-D49F
        0xD4A0, # G+17551  U+D4A0  uniD4A0,  HANGUL SYLLABLE-D4A0
        0xD4A1, # G+17552  U+D4A1  uniD4A1,  HANGUL SYLLABLE-D4A1
        0xD4A2, # G+17553  U+D4A2  uniD4A2,  HANGUL SYLLABLE-D4A2
        0xD4A3, # G+17554  U+D4A3  uniD4A3,  HANGUL SYLLABLE-D4A3
        0xD4A4, # G+17555  U+D4A4  uniD4A4,  HANGUL SYLLABLE-D4A4
        0xD4A5, # G+17556  U+D4A5  uniD4A5,  HANGUL SYLLABLE-D4A5
        0xD4A6, # G+17557  U+D4A6  uniD4A6,  HANGUL SYLLABLE-D4A6
        0xD4A7, # G+17558  U+D4A7  uniD4A7,  HANGUL SYLLABLE-D4A7
        0xD4A8, # G+17559  U+D4A8  uniD4A8,  HANGUL SYLLABLE-D4A8
        0xD4AA, # G+17560  U+D4AA  uniD4AA,  HANGUL SYLLABLE-D4AA
        0xD4AB, # G+17561  U+D4AB  uniD4AB,  HANGUL SYLLABLE-D4AB
        0xD4AC, # G+17562  U+D4AC  uniD4AC,  HANGUL SYLLABLE-D4AC
        0xD4AD, # G+17563  U+D4AD  uniD4AD,  HANGUL SYLLABLE-D4AD
        0xD4AE, # G+17564  U+D4AE  uniD4AE,  HANGUL SYLLABLE-D4AE
        0xD4AF, # G+17565  U+D4AF  uniD4AF,  HANGUL SYLLABLE-D4AF
        0xD4B0, # G+17566  U+D4B0  uniD4B0,  HANGUL SYLLABLE-D4B0
        0xD4B1, # G+17567  U+D4B1  uniD4B1,  HANGUL SYLLABLE-D4B1
        0xD4B2, # G+17568  U+D4B2  uniD4B2,  HANGUL SYLLABLE-D4B2
        0xD4B3, # G+17569  U+D4B3  uniD4B3,  HANGUL SYLLABLE-D4B3
        0xD4B4, # G+17570  U+D4B4  uniD4B4,  HANGUL SYLLABLE-D4B4
        0xD4B5, # G+17571  U+D4B5  uniD4B5,  HANGUL SYLLABLE-D4B5
        0xD4B6, # G+17572  U+D4B6  uniD4B6,  HANGUL SYLLABLE-D4B6
        0xD4B7, # G+17573  U+D4B7  uniD4B7,  HANGUL SYLLABLE-D4B7
        0xD4B8, # G+17574  U+D4B8  uniD4B8,  HANGUL SYLLABLE-D4B8
        0xD4B9, # G+17575  U+D4B9  uniD4B9,  HANGUL SYLLABLE-D4B9
        0xD4BA, # G+17576  U+D4BA  uniD4BA,  HANGUL SYLLABLE-D4BA
        0xD4BB, # G+17577  U+D4BB  uniD4BB,  HANGUL SYLLABLE-D4BB
        0xD4BC, # G+17578  U+D4BC  uniD4BC,  HANGUL SYLLABLE-D4BC
        0xD4BD, # G+17579  U+D4BD  uniD4BD,  HANGUL SYLLABLE-D4BD
        0xD4BE, # G+17580  U+D4BE  uniD4BE,  HANGUL SYLLABLE-D4BE
        0xD4BF, # G+17581  U+D4BF  uniD4BF,  HANGUL SYLLABLE-D4BF
        0xD4C0, # G+17582  U+D4C0  uniD4C0,  HANGUL SYLLABLE-D4C0
        0xD4C1, # G+17583  U+D4C1  uniD4C1,  HANGUL SYLLABLE-D4C1
        0xD4C2, # G+17584  U+D4C2  uniD4C2,  HANGUL SYLLABLE-D4C2
        0xD4C3, # G+17585  U+D4C3  uniD4C3,  HANGUL SYLLABLE-D4C3
        0xD4C4, # G+17586  U+D4C4  uniD4C4,  HANGUL SYLLABLE-D4C4
        0xD4C5, # G+17587  U+D4C5  uniD4C5,  HANGUL SYLLABLE-D4C5
        0xD4C6, # G+17588  U+D4C6  uniD4C6,  HANGUL SYLLABLE-D4C6
        0xD4C7, # G+17589  U+D4C7  uniD4C7,  HANGUL SYLLABLE-D4C7
        0xD4C8, # G+17590  U+D4C8  uniD4C8,  HANGUL SYLLABLE-D4C8
        0xD4C9, # G+17591  U+D4C9  uniD4C9,  HANGUL SYLLABLE-D4C9
        0xD4CA, # G+17592  U+D4CA  uniD4CA,  HANGUL SYLLABLE-D4CA
        0xD4CB, # G+17593  U+D4CB  uniD4CB,  HANGUL SYLLABLE-D4CB
        0xD4CD, # G+17594  U+D4CD  uniD4CD,  HANGUL SYLLABLE-D4CD
        0xD4CE, # G+17595  U+D4CE  uniD4CE,  HANGUL SYLLABLE-D4CE
        0xD4CF, # G+17596  U+D4CF  uniD4CF,  HANGUL SYLLABLE-D4CF
        0xD4D1, # G+17597  U+D4D1  uniD4D1,  HANGUL SYLLABLE-D4D1
        0xD4D2, # G+17598  U+D4D2  uniD4D2,  HANGUL SYLLABLE-D4D2
        0xD4D3, # G+17599  U+D4D3  uniD4D3,  HANGUL SYLLABLE-D4D3
        0xD4D5, # G+17600  U+D4D5  uniD4D5,  HANGUL SYLLABLE-D4D5
        0xD4D6, # G+17601  U+D4D6  uniD4D6,  HANGUL SYLLABLE-D4D6
        0xD4D7, # G+17602  U+D4D7  uniD4D7,  HANGUL SYLLABLE-D4D7
        0xD4D8, # G+17603  U+D4D8  uniD4D8,  HANGUL SYLLABLE-D4D8
        0xD4D9, # G+17604  U+D4D9  uniD4D9,  HANGUL SYLLABLE-D4D9
        0xD4DA, # G+17605  U+D4DA  uniD4DA,  HANGUL SYLLABLE-D4DA
        0xD4DB, # G+17606  U+D4DB  uniD4DB,  HANGUL SYLLABLE-D4DB
        0xD4DD, # G+17607  U+D4DD  uniD4DD,  HANGUL SYLLABLE-D4DD
        0xD4DE, # G+17608  U+D4DE  uniD4DE,  HANGUL SYLLABLE-D4DE
        0xD4E0, # G+17609  U+D4E0  uniD4E0,  HANGUL SYLLABLE-D4E0
        0xD4E1, # G+17610  U+D4E1  uniD4E1,  HANGUL SYLLABLE-D4E1
        0xD4E2, # G+17611  U+D4E2  uniD4E2,  HANGUL SYLLABLE-D4E2
        0xD4E3, # G+17612  U+D4E3  uniD4E3,  HANGUL SYLLABLE-D4E3
        0xD4E4, # G+17613  U+D4E4  uniD4E4,  HANGUL SYLLABLE-D4E4
        0xD4E5, # G+17614  U+D4E5  uniD4E5,  HANGUL SYLLABLE-D4E5
        0xD4E6, # G+17615  U+D4E6  uniD4E6,  HANGUL SYLLABLE-D4E6
        0xD4E7, # G+17616  U+D4E7  uniD4E7,  HANGUL SYLLABLE-D4E7
        0xD4E9, # G+17617  U+D4E9  uniD4E9,  HANGUL SYLLABLE-D4E9
        0xD4EA, # G+17618  U+D4EA  uniD4EA,  HANGUL SYLLABLE-D4EA
        0xD4EB, # G+17619  U+D4EB  uniD4EB,  HANGUL SYLLABLE-D4EB
        0xD4ED, # G+17620  U+D4ED  uniD4ED,  HANGUL SYLLABLE-D4ED
        0xD4EE, # G+17621  U+D4EE  uniD4EE,  HANGUL SYLLABLE-D4EE
        0xD4EF, # G+17622  U+D4EF  uniD4EF,  HANGUL SYLLABLE-D4EF
        0xD4F1, # G+17623  U+D4F1  uniD4F1,  HANGUL SYLLABLE-D4F1
        0xD4F2, # G+17624  U+D4F2  uniD4F2,  HANGUL SYLLABLE-D4F2
        0xD4F3, # G+17625  U+D4F3  uniD4F3,  HANGUL SYLLABLE-D4F3
        0xD4F4, # G+17626  U+D4F4  uniD4F4,  HANGUL SYLLABLE-D4F4
        0xD4F5, # G+17627  U+D4F5  uniD4F5,  HANGUL SYLLABLE-D4F5
        0xD4F6, # G+17628  U+D4F6  uniD4F6,  HANGUL SYLLABLE-D4F6
        0xD4F7, # G+17629  U+D4F7  uniD4F7,  HANGUL SYLLABLE-D4F7
        0xD4F9, # G+17630  U+D4F9  uniD4F9,  HANGUL SYLLABLE-D4F9
        0xD4FA, # G+17631  U+D4FA  uniD4FA,  HANGUL SYLLABLE-D4FA
        0xD4FC, # G+17632  U+D4FC  uniD4FC,  HANGUL SYLLABLE-D4FC
        0xD4FE, # G+17633  U+D4FE  uniD4FE,  HANGUL SYLLABLE-D4FE
        0xD4FF, # G+17634  U+D4FF  uniD4FF,  HANGUL SYLLABLE-D4FF
        0xD500, # G+17635  U+D500  uniD500,  HANGUL SYLLABLE-D500
        0xD501, # G+17636  U+D501  uniD501,  HANGUL SYLLABLE-D501
        0xD502, # G+17637  U+D502  uniD502,  HANGUL SYLLABLE-D502
        0xD503, # G+17638  U+D503  uniD503,  HANGUL SYLLABLE-D503
        0xD505, # G+17639  U+D505  uniD505,  HANGUL SYLLABLE-D505
        0xD506, # G+17640  U+D506  uniD506,  HANGUL SYLLABLE-D506
        0xD507, # G+17641  U+D507  uniD507,  HANGUL SYLLABLE-D507
        0xD509, # G+17642  U+D509  uniD509,  HANGUL SYLLABLE-D509
        0xD50A, # G+17643  U+D50A  uniD50A,  HANGUL SYLLABLE-D50A
        0xD50B, # G+17644  U+D50B  uniD50B,  HANGUL SYLLABLE-D50B
        0xD50D, # G+17645  U+D50D  uniD50D,  HANGUL SYLLABLE-D50D
        0xD50E, # G+17646  U+D50E  uniD50E,  HANGUL SYLLABLE-D50E
        0xD50F, # G+17647  U+D50F  uniD50F,  HANGUL SYLLABLE-D50F
        0xD510, # G+17648  U+D510  uniD510,  HANGUL SYLLABLE-D510
        0xD511, # G+17649  U+D511  uniD511,  HANGUL SYLLABLE-D511
        0xD512, # G+17650  U+D512  uniD512,  HANGUL SYLLABLE-D512
        0xD513, # G+17651  U+D513  uniD513,  HANGUL SYLLABLE-D513
        0xD516, # G+17652  U+D516  uniD516,  HANGUL SYLLABLE-D516
        0xD518, # G+17653  U+D518  uniD518,  HANGUL SYLLABLE-D518
        0xD519, # G+17654  U+D519  uniD519,  HANGUL SYLLABLE-D519
        0xD51A, # G+17655  U+D51A  uniD51A,  HANGUL SYLLABLE-D51A
        0xD51B, # G+17656  U+D51B  uniD51B,  HANGUL SYLLABLE-D51B
        0xD51C, # G+17657  U+D51C  uniD51C,  HANGUL SYLLABLE-D51C
        0xD51D, # G+17658  U+D51D  uniD51D,  HANGUL SYLLABLE-D51D
        0xD51E, # G+17659  U+D51E  uniD51E,  HANGUL SYLLABLE-D51E
        0xD51F, # G+17660  U+D51F  uniD51F,  HANGUL SYLLABLE-D51F
        0xD520, # G+17661  U+D520  uniD520,  HANGUL SYLLABLE-D520
        0xD521, # G+17662  U+D521  uniD521,  HANGUL SYLLABLE-D521
        0xD522, # G+17663  U+D522  uniD522,  HANGUL SYLLABLE-D522
        0xD523, # G+17664  U+D523  uniD523,  HANGUL SYLLABLE-D523
        0xD524, # G+17665  U+D524  uniD524,  HANGUL SYLLABLE-D524
        0xD525, # G+17666  U+D525  uniD525,  HANGUL SYLLABLE-D525
        0xD526, # G+17667  U+D526  uniD526,  HANGUL SYLLABLE-D526
        0xD527, # G+17668  U+D527  uniD527,  HANGUL SYLLABLE-D527
        0xD528, # G+17669  U+D528  uniD528,  HANGUL SYLLABLE-D528
        0xD529, # G+17670  U+D529  uniD529,  HANGUL SYLLABLE-D529
        0xD52A, # G+17671  U+D52A  uniD52A,  HANGUL SYLLABLE-D52A
        0xD52B, # G+17672  U+D52B  uniD52B,  HANGUL SYLLABLE-D52B
        0xD52C, # G+17673  U+D52C  uniD52C,  HANGUL SYLLABLE-D52C
        0xD52D, # G+17674  U+D52D  uniD52D,  HANGUL SYLLABLE-D52D
        0xD52E, # G+17675  U+D52E  uniD52E,  HANGUL SYLLABLE-D52E
        0xD52F, # G+17676  U+D52F  uniD52F,  HANGUL SYLLABLE-D52F
        0xD530, # G+17677  U+D530  uniD530,  HANGUL SYLLABLE-D530
        0xD531, # G+17678  U+D531  uniD531,  HANGUL SYLLABLE-D531
        0xD532, # G+17679  U+D532  uniD532,  HANGUL SYLLABLE-D532
        0xD533, # G+17680  U+D533  uniD533,  HANGUL SYLLABLE-D533
        0xD534, # G+17681  U+D534  uniD534,  HANGUL SYLLABLE-D534
        0xD535, # G+17682  U+D535  uniD535,  HANGUL SYLLABLE-D535
        0xD536, # G+17683  U+D536  uniD536,  HANGUL SYLLABLE-D536
        0xD537, # G+17684  U+D537  uniD537,  HANGUL SYLLABLE-D537
        0xD538, # G+17685  U+D538  uniD538,  HANGUL SYLLABLE-D538
        0xD539, # G+17686  U+D539  uniD539,  HANGUL SYLLABLE-D539
        0xD53A, # G+17687  U+D53A  uniD53A,  HANGUL SYLLABLE-D53A
        0xD53B, # G+17688  U+D53B  uniD53B,  HANGUL SYLLABLE-D53B
        0xD53E, # G+17689  U+D53E  uniD53E,  HANGUL SYLLABLE-D53E
        0xD53F, # G+17690  U+D53F  uniD53F,  HANGUL SYLLABLE-D53F
        0xD541, # G+17691  U+D541  uniD541,  HANGUL SYLLABLE-D541
        0xD542, # G+17692  U+D542  uniD542,  HANGUL SYLLABLE-D542
        0xD543, # G+17693  U+D543  uniD543,  HANGUL SYLLABLE-D543
        0xD545, # G+17694  U+D545  uniD545,  HANGUL SYLLABLE-D545
        0xD546, # G+17695  U+D546  uniD546,  HANGUL SYLLABLE-D546
        0xD547, # G+17696  U+D547  uniD547,  HANGUL SYLLABLE-D547
        0xD548, # G+17697  U+D548  uniD548,  HANGUL SYLLABLE-D548
        0xD549, # G+17698  U+D549  uniD549,  HANGUL SYLLABLE-D549
        0xD54A, # G+17699  U+D54A  uniD54A,  HANGUL SYLLABLE-D54A
        0xD54B, # G+17700  U+D54B  uniD54B,  HANGUL SYLLABLE-D54B
        0xD54E, # G+17701  U+D54E  uniD54E,  HANGUL SYLLABLE-D54E
        0xD550, # G+17702  U+D550  uniD550,  HANGUL SYLLABLE-D550
        0xD552, # G+17703  U+D552  uniD552,  HANGUL SYLLABLE-D552
        0xD553, # G+17704  U+D553  uniD553,  HANGUL SYLLABLE-D553
        0xD554, # G+17705  U+D554  uniD554,  HANGUL SYLLABLE-D554
        0xD555, # G+17706  U+D555  uniD555,  HANGUL SYLLABLE-D555
        0xD556, # G+17707  U+D556  uniD556,  HANGUL SYLLABLE-D556
        0xD557, # G+17708  U+D557  uniD557,  HANGUL SYLLABLE-D557
        0xD55A, # G+17709  U+D55A  uniD55A,  HANGUL SYLLABLE-D55A
        0xD55B, # G+17710  U+D55B  uniD55B,  HANGUL SYLLABLE-D55B
        0xD55D, # G+17711  U+D55D  uniD55D,  HANGUL SYLLABLE-D55D
        0xD55E, # G+17712  U+D55E  uniD55E,  HANGUL SYLLABLE-D55E
        0xD55F, # G+17713  U+D55F  uniD55F,  HANGUL SYLLABLE-D55F
        0xD561, # G+17714  U+D561  uniD561,  HANGUL SYLLABLE-D561
        0xD562, # G+17715  U+D562  uniD562,  HANGUL SYLLABLE-D562
        0xD563, # G+17716  U+D563  uniD563,  HANGUL SYLLABLE-D563
        0xD564, # G+17717  U+D564  uniD564,  HANGUL SYLLABLE-D564
        0xD566, # G+17718  U+D566  uniD566,  HANGUL SYLLABLE-D566
        0xD567, # G+17719  U+D567  uniD567,  HANGUL SYLLABLE-D567
        0xD56A, # G+17720  U+D56A  uniD56A,  HANGUL SYLLABLE-D56A
        0xD56C, # G+17721  U+D56C  uniD56C,  HANGUL SYLLABLE-D56C
        0xD56E, # G+17722  U+D56E  uniD56E,  HANGUL SYLLABLE-D56E
        0xD56F, # G+17723  U+D56F  uniD56F,  HANGUL SYLLABLE-D56F
        0xD570, # G+17724  U+D570  uniD570,  HANGUL SYLLABLE-D570
        0xD571, # G+17725  U+D571  uniD571,  HANGUL SYLLABLE-D571
        0xD572, # G+17726  U+D572  uniD572,  HANGUL SYLLABLE-D572
        0xD573, # G+17727  U+D573  uniD573,  HANGUL SYLLABLE-D573
        0xD576, # G+17728  U+D576  uniD576,  HANGUL SYLLABLE-D576
        0xD577, # G+17729  U+D577  uniD577,  HANGUL SYLLABLE-D577
        0xD579, # G+17730  U+D579  uniD579,  HANGUL SYLLABLE-D579
        0xD57A, # G+17731  U+D57A  uniD57A,  HANGUL SYLLABLE-D57A
        0xD57B, # G+17732  U+D57B  uniD57B,  HANGUL SYLLABLE-D57B
        0xD57D, # G+17733  U+D57D  uniD57D,  HANGUL SYLLABLE-D57D
        0xD57E, # G+17734  U+D57E  uniD57E,  HANGUL SYLLABLE-D57E
        0xD57F, # G+17735  U+D57F  uniD57F,  HANGUL SYLLABLE-D57F
        0xD580, # G+17736  U+D580  uniD580,  HANGUL SYLLABLE-D580
        0xD581, # G+17737  U+D581  uniD581,  HANGUL SYLLABLE-D581
        0xD582, # G+17738  U+D582  uniD582,  HANGUL SYLLABLE-D582
        0xD583, # G+17739  U+D583  uniD583,  HANGUL SYLLABLE-D583
        0xD586, # G+17740  U+D586  uniD586,  HANGUL SYLLABLE-D586
        0xD58A, # G+17741  U+D58A  uniD58A,  HANGUL SYLLABLE-D58A
        0xD58B, # G+17742  U+D58B  uniD58B,  HANGUL SYLLABLE-D58B
        0xD58C, # G+17743  U+D58C  uniD58C,  HANGUL SYLLABLE-D58C
        0xD58D, # G+17744  U+D58D  uniD58D,  HANGUL SYLLABLE-D58D
        0xD58E, # G+17745  U+D58E  uniD58E,  HANGUL SYLLABLE-D58E
        0xD58F, # G+17746  U+D58F  uniD58F,  HANGUL SYLLABLE-D58F
        0xD591, # G+17747  U+D591  uniD591,  HANGUL SYLLABLE-D591
        0xD592, # G+17748  U+D592  uniD592,  HANGUL SYLLABLE-D592
        0xD593, # G+17749  U+D593  uniD593,  HANGUL SYLLABLE-D593
        0xD594, # G+17750  U+D594  uniD594,  HANGUL SYLLABLE-D594
        0xD595, # G+17751  U+D595  uniD595,  HANGUL SYLLABLE-D595
        0xD596, # G+17752  U+D596  uniD596,  HANGUL SYLLABLE-D596
        0xD597, # G+17753  U+D597  uniD597,  HANGUL SYLLABLE-D597
        0xD598, # G+17754  U+D598  uniD598,  HANGUL SYLLABLE-D598
        0xD599, # G+17755  U+D599  uniD599,  HANGUL SYLLABLE-D599
        0xD59A, # G+17756  U+D59A  uniD59A,  HANGUL SYLLABLE-D59A
        0xD59B, # G+17757  U+D59B  uniD59B,  HANGUL SYLLABLE-D59B
        0xD59C, # G+17758  U+D59C  uniD59C,  HANGUL SYLLABLE-D59C
        0xD59D, # G+17759  U+D59D  uniD59D,  HANGUL SYLLABLE-D59D
        0xD59E, # G+17760  U+D59E  uniD59E,  HANGUL SYLLABLE-D59E
        0xD59F, # G+17761  U+D59F  uniD59F,  HANGUL SYLLABLE-D59F
        0xD5A0, # G+17762  U+D5A0  uniD5A0,  HANGUL SYLLABLE-D5A0
        0xD5A1, # G+17763  U+D5A1  uniD5A1,  HANGUL SYLLABLE-D5A1
        0xD5A2, # G+17764  U+D5A2  uniD5A2,  HANGUL SYLLABLE-D5A2
        0xD5A3, # G+17765  U+D5A3  uniD5A3,  HANGUL SYLLABLE-D5A3
        0xD5A4, # G+17766  U+D5A4  uniD5A4,  HANGUL SYLLABLE-D5A4
        0xD5A6, # G+17767  U+D5A6  uniD5A6,  HANGUL SYLLABLE-D5A6
        0xD5A7, # G+17768  U+D5A7  uniD5A7,  HANGUL SYLLABLE-D5A7
        0xD5A8, # G+17769  U+D5A8  uniD5A8,  HANGUL SYLLABLE-D5A8
        0xD5A9, # G+17770  U+D5A9  uniD5A9,  HANGUL SYLLABLE-D5A9
        0xD5AA, # G+17771  U+D5AA  uniD5AA,  HANGUL SYLLABLE-D5AA
        0xD5AB, # G+17772  U+D5AB  uniD5AB,  HANGUL SYLLABLE-D5AB
        0xD5AC, # G+17773  U+D5AC  uniD5AC,  HANGUL SYLLABLE-D5AC
        0xD5AD, # G+17774  U+D5AD  uniD5AD,  HANGUL SYLLABLE-D5AD
        0xD5AE, # G+17775  U+D5AE  uniD5AE,  HANGUL SYLLABLE-D5AE
        0xD5AF, # G+17776  U+D5AF  uniD5AF,  HANGUL SYLLABLE-D5AF
        0xD5B0, # G+17777  U+D5B0  uniD5B0,  HANGUL SYLLABLE-D5B0
        0xD5B1, # G+17778  U+D5B1  uniD5B1,  HANGUL SYLLABLE-D5B1
        0xD5B2, # G+17779  U+D5B2  uniD5B2,  HANGUL SYLLABLE-D5B2
        0xD5B3, # G+17780  U+D5B3  uniD5B3,  HANGUL SYLLABLE-D5B3
        0xD5B4, # G+17781  U+D5B4  uniD5B4,  HANGUL SYLLABLE-D5B4
        0xD5B5, # G+17782  U+D5B5  uniD5B5,  HANGUL SYLLABLE-D5B5
        0xD5B6, # G+17783  U+D5B6  uniD5B6,  HANGUL SYLLABLE-D5B6
        0xD5B7, # G+17784  U+D5B7  uniD5B7,  HANGUL SYLLABLE-D5B7
        0xD5B8, # G+17785  U+D5B8  uniD5B8,  HANGUL SYLLABLE-D5B8
        0xD5B9, # G+17786  U+D5B9  uniD5B9,  HANGUL SYLLABLE-D5B9
        0xD5BA, # G+17787  U+D5BA  uniD5BA,  HANGUL SYLLABLE-D5BA
        0xD5BB, # G+17788  U+D5BB  uniD5BB,  HANGUL SYLLABLE-D5BB
        0xD5BC, # G+17789  U+D5BC  uniD5BC,  HANGUL SYLLABLE-D5BC
        0xD5BD, # G+17790  U+D5BD  uniD5BD,  HANGUL SYLLABLE-D5BD
        0xD5BE, # G+17791  U+D5BE  uniD5BE,  HANGUL SYLLABLE-D5BE
        0xD5BF, # G+17792  U+D5BF  uniD5BF,  HANGUL SYLLABLE-D5BF
        0xD5C0, # G+17793  U+D5C0  uniD5C0,  HANGUL SYLLABLE-D5C0
        0xD5C1, # G+17794  U+D5C1  uniD5C1,  HANGUL SYLLABLE-D5C1
        0xD5C2, # G+17795  U+D5C2  uniD5C2,  HANGUL SYLLABLE-D5C2
        0xD5C3, # G+17796  U+D5C3  uniD5C3,  HANGUL SYLLABLE-D5C3
        0xD5C4, # G+17797  U+D5C4  uniD5C4,  HANGUL SYLLABLE-D5C4
        0xD5C5, # G+17798  U+D5C5  uniD5C5,  HANGUL SYLLABLE-D5C5
        0xD5C6, # G+17799  U+D5C6  uniD5C6,  HANGUL SYLLABLE-D5C6
        0xD5C7, # G+17800  U+D5C7  uniD5C7,  HANGUL SYLLABLE-D5C7
        0xD5CA, # G+17801  U+D5CA  uniD5CA,  HANGUL SYLLABLE-D5CA
        0xD5CB, # G+17802  U+D5CB  uniD5CB,  HANGUL SYLLABLE-D5CB
        0xD5CD, # G+17803  U+D5CD  uniD5CD,  HANGUL SYLLABLE-D5CD
        0xD5CE, # G+17804  U+D5CE  uniD5CE,  HANGUL SYLLABLE-D5CE
        0xD5CF, # G+17805  U+D5CF  uniD5CF,  HANGUL SYLLABLE-D5CF
        0xD5D1, # G+17806  U+D5D1  uniD5D1,  HANGUL SYLLABLE-D5D1
        0xD5D3, # G+17807  U+D5D3  uniD5D3,  HANGUL SYLLABLE-D5D3
        0xD5D4, # G+17808  U+D5D4  uniD5D4,  HANGUL SYLLABLE-D5D4
        0xD5D5, # G+17809  U+D5D5  uniD5D5,  HANGUL SYLLABLE-D5D5
        0xD5D6, # G+17810  U+D5D6  uniD5D6,  HANGUL SYLLABLE-D5D6
        0xD5D7, # G+17811  U+D5D7  uniD5D7,  HANGUL SYLLABLE-D5D7
        0xD5DA, # G+17812  U+D5DA  uniD5DA,  HANGUL SYLLABLE-D5DA
        0xD5DC, # G+17813  U+D5DC  uniD5DC,  HANGUL SYLLABLE-D5DC
        0xD5DE, # G+17814  U+D5DE  uniD5DE,  HANGUL SYLLABLE-D5DE
        0xD5DF, # G+17815  U+D5DF  uniD5DF,  HANGUL SYLLABLE-D5DF
        0xD5E0, # G+17816  U+D5E0  uniD5E0,  HANGUL SYLLABLE-D5E0
        0xD5E1, # G+17817  U+D5E1  uniD5E1,  HANGUL SYLLABLE-D5E1
        0xD5E2, # G+17818  U+D5E2  uniD5E2,  HANGUL SYLLABLE-D5E2
        0xD5E3, # G+17819  U+D5E3  uniD5E3,  HANGUL SYLLABLE-D5E3
        0xD5E6, # G+17820  U+D5E6  uniD5E6,  HANGUL SYLLABLE-D5E6
        0xD5E7, # G+17821  U+D5E7  uniD5E7,  HANGUL SYLLABLE-D5E7
        0xD5E9, # G+17822  U+D5E9  uniD5E9,  HANGUL SYLLABLE-D5E9
        0xD5EA, # G+17823  U+D5EA  uniD5EA,  HANGUL SYLLABLE-D5EA
        0xD5EB, # G+17824  U+D5EB  uniD5EB,  HANGUL SYLLABLE-D5EB
        0xD5ED, # G+17825  U+D5ED  uniD5ED,  HANGUL SYLLABLE-D5ED
        0xD5EE, # G+17826  U+D5EE  uniD5EE,  HANGUL SYLLABLE-D5EE
        0xD5EF, # G+17827  U+D5EF  uniD5EF,  HANGUL SYLLABLE-D5EF
        0xD5F0, # G+17828  U+D5F0  uniD5F0,  HANGUL SYLLABLE-D5F0
        0xD5F1, # G+17829  U+D5F1  uniD5F1,  HANGUL SYLLABLE-D5F1
        0xD5F2, # G+17830  U+D5F2  uniD5F2,  HANGUL SYLLABLE-D5F2
        0xD5F3, # G+17831  U+D5F3  uniD5F3,  HANGUL SYLLABLE-D5F3
        0xD5F6, # G+17832  U+D5F6  uniD5F6,  HANGUL SYLLABLE-D5F6
        0xD5F8, # G+17833  U+D5F8  uniD5F8,  HANGUL SYLLABLE-D5F8
        0xD5FA, # G+17834  U+D5FA  uniD5FA,  HANGUL SYLLABLE-D5FA
        0xD5FB, # G+17835  U+D5FB  uniD5FB,  HANGUL SYLLABLE-D5FB
        0xD5FC, # G+17836  U+D5FC  uniD5FC,  HANGUL SYLLABLE-D5FC
        0xD5FD, # G+17837  U+D5FD  uniD5FD,  HANGUL SYLLABLE-D5FD
        0xD5FE, # G+17838  U+D5FE  uniD5FE,  HANGUL SYLLABLE-D5FE
        0xD5FF, # G+17839  U+D5FF  uniD5FF,  HANGUL SYLLABLE-D5FF
        0xD602, # G+17840  U+D602  uniD602,  HANGUL SYLLABLE-D602
        0xD603, # G+17841  U+D603  uniD603,  HANGUL SYLLABLE-D603
        0xD605, # G+17842  U+D605  uniD605,  HANGUL SYLLABLE-D605
        0xD606, # G+17843  U+D606  uniD606,  HANGUL SYLLABLE-D606
        0xD607, # G+17844  U+D607  uniD607,  HANGUL SYLLABLE-D607
        0xD609, # G+17845  U+D609  uniD609,  HANGUL SYLLABLE-D609
        0xD60A, # G+17846  U+D60A  uniD60A,  HANGUL SYLLABLE-D60A
        0xD60B, # G+17847  U+D60B  uniD60B,  HANGUL SYLLABLE-D60B
        0xD60C, # G+17848  U+D60C  uniD60C,  HANGUL SYLLABLE-D60C
        0xD60D, # G+17849  U+D60D  uniD60D,  HANGUL SYLLABLE-D60D
        0xD60E, # G+17850  U+D60E  uniD60E,  HANGUL SYLLABLE-D60E
        0xD60F, # G+17851  U+D60F  uniD60F,  HANGUL SYLLABLE-D60F
        0xD612, # G+17852  U+D612  uniD612,  HANGUL SYLLABLE-D612
        0xD616, # G+17853  U+D616  uniD616,  HANGUL SYLLABLE-D616
        0xD617, # G+17854  U+D617  uniD617,  HANGUL SYLLABLE-D617
        0xD618, # G+17855  U+D618  uniD618,  HANGUL SYLLABLE-D618
        0xD619, # G+17856  U+D619  uniD619,  HANGUL SYLLABLE-D619
        0xD61A, # G+17857  U+D61A  uniD61A,  HANGUL SYLLABLE-D61A
        0xD61B, # G+17858  U+D61B  uniD61B,  HANGUL SYLLABLE-D61B
        0xD61D, # G+17859  U+D61D  uniD61D,  HANGUL SYLLABLE-D61D
        0xD61E, # G+17860  U+D61E  uniD61E,  HANGUL SYLLABLE-D61E
        0xD61F, # G+17861  U+D61F  uniD61F,  HANGUL SYLLABLE-D61F
        0xD621, # G+17862  U+D621  uniD621,  HANGUL SYLLABLE-D621
        0xD622, # G+17863  U+D622  uniD622,  HANGUL SYLLABLE-D622
        0xD623, # G+17864  U+D623  uniD623,  HANGUL SYLLABLE-D623
        0xD625, # G+17865  U+D625  uniD625,  HANGUL SYLLABLE-D625
        0xD626, # G+17866  U+D626  uniD626,  HANGUL SYLLABLE-D626
        0xD627, # G+17867  U+D627  uniD627,  HANGUL SYLLABLE-D627
        0xD628, # G+17868  U+D628  uniD628,  HANGUL SYLLABLE-D628
        0xD629, # G+17869  U+D629  uniD629,  HANGUL SYLLABLE-D629
        0xD62A, # G+17870  U+D62A  uniD62A,  HANGUL SYLLABLE-D62A
        0xD62B, # G+17871  U+D62B  uniD62B,  HANGUL SYLLABLE-D62B
        0xD62C, # G+17872  U+D62C  uniD62C,  HANGUL SYLLABLE-D62C
        0xD62E, # G+17873  U+D62E  uniD62E,  HANGUL SYLLABLE-D62E
        0xD62F, # G+17874  U+D62F  uniD62F,  HANGUL SYLLABLE-D62F
        0xD630, # G+17875  U+D630  uniD630,  HANGUL SYLLABLE-D630
        0xD631, # G+17876  U+D631  uniD631,  HANGUL SYLLABLE-D631
        0xD632, # G+17877  U+D632  uniD632,  HANGUL SYLLABLE-D632
        0xD633, # G+17878  U+D633  uniD633,  HANGUL SYLLABLE-D633
        0xD634, # G+17879  U+D634  uniD634,  HANGUL SYLLABLE-D634
        0xD635, # G+17880  U+D635  uniD635,  HANGUL SYLLABLE-D635
        0xD636, # G+17881  U+D636  uniD636,  HANGUL SYLLABLE-D636
        0xD637, # G+17882  U+D637  uniD637,  HANGUL SYLLABLE-D637
        0xD63A, # G+17883  U+D63A  uniD63A,  HANGUL SYLLABLE-D63A
        0xD63B, # G+17884  U+D63B  uniD63B,  HANGUL SYLLABLE-D63B
        0xD63D, # G+17885  U+D63D  uniD63D,  HANGUL SYLLABLE-D63D
        0xD63E, # G+17886  U+D63E  uniD63E,  HANGUL SYLLABLE-D63E
        0xD63F, # G+17887  U+D63F  uniD63F,  HANGUL SYLLABLE-D63F
        0xD641, # G+17888  U+D641  uniD641,  HANGUL SYLLABLE-D641
        0xD642, # G+17889  U+D642  uniD642,  HANGUL SYLLABLE-D642
        0xD643, # G+17890  U+D643  uniD643,  HANGUL SYLLABLE-D643
        0xD644, # G+17891  U+D644  uniD644,  HANGUL SYLLABLE-D644
        0xD646, # G+17892  U+D646  uniD646,  HANGUL SYLLABLE-D646
        0xD647, # G+17893  U+D647  uniD647,  HANGUL SYLLABLE-D647
        0xD64A, # G+17894  U+D64A  uniD64A,  HANGUL SYLLABLE-D64A
        0xD64C, # G+17895  U+D64C  uniD64C,  HANGUL SYLLABLE-D64C
        0xD64E, # G+17896  U+D64E  uniD64E,  HANGUL SYLLABLE-D64E
        0xD64F, # G+17897  U+D64F  uniD64F,  HANGUL SYLLABLE-D64F
        0xD650, # G+17898  U+D650  uniD650,  HANGUL SYLLABLE-D650
        0xD652, # G+17899  U+D652  uniD652,  HANGUL SYLLABLE-D652
        0xD653, # G+17900  U+D653  uniD653,  HANGUL SYLLABLE-D653
        0xD656, # G+17901  U+D656  uniD656,  HANGUL SYLLABLE-D656
        0xD657, # G+17902  U+D657  uniD657,  HANGUL SYLLABLE-D657
        0xD659, # G+17903  U+D659  uniD659,  HANGUL SYLLABLE-D659
        0xD65A, # G+17904  U+D65A  uniD65A,  HANGUL SYLLABLE-D65A
        0xD65B, # G+17905  U+D65B  uniD65B,  HANGUL SYLLABLE-D65B
        0xD65D, # G+17906  U+D65D  uniD65D,  HANGUL SYLLABLE-D65D
        0xD65E, # G+17907  U+D65E  uniD65E,  HANGUL SYLLABLE-D65E
        0xD65F, # G+17908  U+D65F  uniD65F,  HANGUL SYLLABLE-D65F
        0xD660, # G+17909  U+D660  uniD660,  HANGUL SYLLABLE-D660
        0xD661, # G+17910  U+D661  uniD661,  HANGUL SYLLABLE-D661
        0xD662, # G+17911  U+D662  uniD662,  HANGUL SYLLABLE-D662
        0xD663, # G+17912  U+D663  uniD663,  HANGUL SYLLABLE-D663
        0xD664, # G+17913  U+D664  uniD664,  HANGUL SYLLABLE-D664
        0xD665, # G+17914  U+D665  uniD665,  HANGUL SYLLABLE-D665
        0xD666, # G+17915  U+D666  uniD666,  HANGUL SYLLABLE-D666
        0xD668, # G+17916  U+D668  uniD668,  HANGUL SYLLABLE-D668
        0xD66A, # G+17917  U+D66A  uniD66A,  HANGUL SYLLABLE-D66A
        0xD66B, # G+17918  U+D66B  uniD66B,  HANGUL SYLLABLE-D66B
        0xD66C, # G+17919  U+D66C  uniD66C,  HANGUL SYLLABLE-D66C
        0xD66D, # G+17920  U+D66D  uniD66D,  HANGUL SYLLABLE-D66D
        0xD66E, # G+17921  U+D66E  uniD66E,  HANGUL SYLLABLE-D66E
        0xD66F, # G+17922  U+D66F  uniD66F,  HANGUL SYLLABLE-D66F
        0xD672, # G+17923  U+D672  uniD672,  HANGUL SYLLABLE-D672
        0xD673, # G+17924  U+D673  uniD673,  HANGUL SYLLABLE-D673
        0xD675, # G+17925  U+D675  uniD675,  HANGUL SYLLABLE-D675
        0xD676, # G+17926  U+D676  uniD676,  HANGUL SYLLABLE-D676
        0xD677, # G+17927  U+D677  uniD677,  HANGUL SYLLABLE-D677
        0xD678, # G+17928  U+D678  uniD678,  HANGUL SYLLABLE-D678
        0xD679, # G+17929  U+D679  uniD679,  HANGUL SYLLABLE-D679
        0xD67A, # G+17930  U+D67A  uniD67A,  HANGUL SYLLABLE-D67A
        0xD67B, # G+17931  U+D67B  uniD67B,  HANGUL SYLLABLE-D67B
        0xD67C, # G+17932  U+D67C  uniD67C,  HANGUL SYLLABLE-D67C
        0xD67D, # G+17933  U+D67D  uniD67D,  HANGUL SYLLABLE-D67D
        0xD67E, # G+17934  U+D67E  uniD67E,  HANGUL SYLLABLE-D67E
        0xD67F, # G+17935  U+D67F  uniD67F,  HANGUL SYLLABLE-D67F
        0xD680, # G+17936  U+D680  uniD680,  HANGUL SYLLABLE-D680
        0xD681, # G+17937  U+D681  uniD681,  HANGUL SYLLABLE-D681
        0xD682, # G+17938  U+D682  uniD682,  HANGUL SYLLABLE-D682
        0xD684, # G+17939  U+D684  uniD684,  HANGUL SYLLABLE-D684
        0xD686, # G+17940  U+D686  uniD686,  HANGUL SYLLABLE-D686
        0xD687, # G+17941  U+D687  uniD687,  HANGUL SYLLABLE-D687
        0xD688, # G+17942  U+D688  uniD688,  HANGUL SYLLABLE-D688
        0xD689, # G+17943  U+D689  uniD689,  HANGUL SYLLABLE-D689
        0xD68A, # G+17944  U+D68A  uniD68A,  HANGUL SYLLABLE-D68A
        0xD68B, # G+17945  U+D68B  uniD68B,  HANGUL SYLLABLE-D68B
        0xD68E, # G+17946  U+D68E  uniD68E,  HANGUL SYLLABLE-D68E
        0xD68F, # G+17947  U+D68F  uniD68F,  HANGUL SYLLABLE-D68F
        0xD691, # G+17948  U+D691  uniD691,  HANGUL SYLLABLE-D691
        0xD692, # G+17949  U+D692  uniD692,  HANGUL SYLLABLE-D692
        0xD693, # G+17950  U+D693  uniD693,  HANGUL SYLLABLE-D693
        0xD695, # G+17951  U+D695  uniD695,  HANGUL SYLLABLE-D695
        0xD696, # G+17952  U+D696  uniD696,  HANGUL SYLLABLE-D696
        0xD697, # G+17953  U+D697  uniD697,  HANGUL SYLLABLE-D697
        0xD698, # G+17954  U+D698  uniD698,  HANGUL SYLLABLE-D698
        0xD699, # G+17955  U+D699  uniD699,  HANGUL SYLLABLE-D699
        0xD69A, # G+17956  U+D69A  uniD69A,  HANGUL SYLLABLE-D69A
        0xD69B, # G+17957  U+D69B  uniD69B,  HANGUL SYLLABLE-D69B
        0xD69C, # G+17958  U+D69C  uniD69C,  HANGUL SYLLABLE-D69C
        0xD69E, # G+17959  U+D69E  uniD69E,  HANGUL SYLLABLE-D69E
        0xD6A0, # G+17960  U+D6A0  uniD6A0,  HANGUL SYLLABLE-D6A0
        0xD6A2, # G+17961  U+D6A2  uniD6A2,  HANGUL SYLLABLE-D6A2
        0xD6A3, # G+17962  U+D6A3  uniD6A3,  HANGUL SYLLABLE-D6A3
        0xD6A4, # G+17963  U+D6A4  uniD6A4,  HANGUL SYLLABLE-D6A4
        0xD6A5, # G+17964  U+D6A5  uniD6A5,  HANGUL SYLLABLE-D6A5
        0xD6A6, # G+17965  U+D6A6  uniD6A6,  HANGUL SYLLABLE-D6A6
        0xD6A7, # G+17966  U+D6A7  uniD6A7,  HANGUL SYLLABLE-D6A7
        0xD6A9, # G+17967  U+D6A9  uniD6A9,  HANGUL SYLLABLE-D6A9
        0xD6AA, # G+17968  U+D6AA  uniD6AA,  HANGUL SYLLABLE-D6AA
        0xD6AB, # G+17969  U+D6AB  uniD6AB,  HANGUL SYLLABLE-D6AB
        0xD6AD, # G+17970  U+D6AD  uniD6AD,  HANGUL SYLLABLE-D6AD
        0xD6AE, # G+17971  U+D6AE  uniD6AE,  HANGUL SYLLABLE-D6AE
        0xD6AF, # G+17972  U+D6AF  uniD6AF,  HANGUL SYLLABLE-D6AF
        0xD6B1, # G+17973  U+D6B1  uniD6B1,  HANGUL SYLLABLE-D6B1
        0xD6B2, # G+17974  U+D6B2  uniD6B2,  HANGUL SYLLABLE-D6B2
        0xD6B3, # G+17975  U+D6B3  uniD6B3,  HANGUL SYLLABLE-D6B3
        0xD6B4, # G+17976  U+D6B4  uniD6B4,  HANGUL SYLLABLE-D6B4
        0xD6B5, # G+17977  U+D6B5  uniD6B5,  HANGUL SYLLABLE-D6B5
        0xD6B6, # G+17978  U+D6B6  uniD6B6,  HANGUL SYLLABLE-D6B6
        0xD6B7, # G+17979  U+D6B7  uniD6B7,  HANGUL SYLLABLE-D6B7
        0xD6B8, # G+17980  U+D6B8  uniD6B8,  HANGUL SYLLABLE-D6B8
        0xD6BA, # G+17981  U+D6BA  uniD6BA,  HANGUL SYLLABLE-D6BA
        0xD6BC, # G+17982  U+D6BC  uniD6BC,  HANGUL SYLLABLE-D6BC
        0xD6BD, # G+17983  U+D6BD  uniD6BD,  HANGUL SYLLABLE-D6BD
        0xD6BE, # G+17984  U+D6BE  uniD6BE,  HANGUL SYLLABLE-D6BE
        0xD6BF, # G+17985  U+D6BF  uniD6BF,  HANGUL SYLLABLE-D6BF
        0xD6C0, # G+17986  U+D6C0  uniD6C0,  HANGUL SYLLABLE-D6C0
        0xD6C1, # G+17987  U+D6C1  uniD6C1,  HANGUL SYLLABLE-D6C1
        0xD6C2, # G+17988  U+D6C2  uniD6C2,  HANGUL SYLLABLE-D6C2
        0xD6C3, # G+17989  U+D6C3  uniD6C3,  HANGUL SYLLABLE-D6C3
        0xD6C6, # G+17990  U+D6C6  uniD6C6,  HANGUL SYLLABLE-D6C6
        0xD6C7, # G+17991  U+D6C7  uniD6C7,  HANGUL SYLLABLE-D6C7
        0xD6C9, # G+17992  U+D6C9  uniD6C9,  HANGUL SYLLABLE-D6C9
        0xD6CA, # G+17993  U+D6CA  uniD6CA,  HANGUL SYLLABLE-D6CA
        0xD6CB, # G+17994  U+D6CB  uniD6CB,  HANGUL SYLLABLE-D6CB
        0xD6CD, # G+17995  U+D6CD  uniD6CD,  HANGUL SYLLABLE-D6CD
        0xD6CE, # G+17996  U+D6CE  uniD6CE,  HANGUL SYLLABLE-D6CE
        0xD6CF, # G+17997  U+D6CF  uniD6CF,  HANGUL SYLLABLE-D6CF
        0xD6D0, # G+17998  U+D6D0  uniD6D0,  HANGUL SYLLABLE-D6D0
        0xD6D2, # G+17999  U+D6D2  uniD6D2,  HANGUL SYLLABLE-D6D2
        0xD6D3, # G+18000  U+D6D3  uniD6D3,  HANGUL SYLLABLE-D6D3
        0xD6D5, # G+18001  U+D6D5  uniD6D5,  HANGUL SYLLABLE-D6D5
        0xD6D6, # G+18002  U+D6D6  uniD6D6,  HANGUL SYLLABLE-D6D6
        0xD6D8, # G+18003  U+D6D8  uniD6D8,  HANGUL SYLLABLE-D6D8
        0xD6DA, # G+18004  U+D6DA  uniD6DA,  HANGUL SYLLABLE-D6DA
        0xD6DB, # G+18005  U+D6DB  uniD6DB,  HANGUL SYLLABLE-D6DB
        0xD6DC, # G+18006  U+D6DC  uniD6DC,  HANGUL SYLLABLE-D6DC
        0xD6DD, # G+18007  U+D6DD  uniD6DD,  HANGUL SYLLABLE-D6DD
        0xD6DE, # G+18008  U+D6DE  uniD6DE,  HANGUL SYLLABLE-D6DE
        0xD6DF, # G+18009  U+D6DF  uniD6DF,  HANGUL SYLLABLE-D6DF
        0xD6E1, # G+18010  U+D6E1  uniD6E1,  HANGUL SYLLABLE-D6E1
        0xD6E2, # G+18011  U+D6E2  uniD6E2,  HANGUL SYLLABLE-D6E2
        0xD6E3, # G+18012  U+D6E3  uniD6E3,  HANGUL SYLLABLE-D6E3
        0xD6E5, # G+18013  U+D6E5  uniD6E5,  HANGUL SYLLABLE-D6E5
        0xD6E6, # G+18014  U+D6E6  uniD6E6,  HANGUL SYLLABLE-D6E6
        0xD6E7, # G+18015  U+D6E7  uniD6E7,  HANGUL SYLLABLE-D6E7
        0xD6E9, # G+18016  U+D6E9  uniD6E9,  HANGUL SYLLABLE-D6E9
        0xD6EA, # G+18017  U+D6EA  uniD6EA,  HANGUL SYLLABLE-D6EA
        0xD6EB, # G+18018  U+D6EB  uniD6EB,  HANGUL SYLLABLE-D6EB
        0xD6EC, # G+18019  U+D6EC  uniD6EC,  HANGUL SYLLABLE-D6EC
        0xD6ED, # G+18020  U+D6ED  uniD6ED,  HANGUL SYLLABLE-D6ED
        0xD6EE, # G+18021  U+D6EE  uniD6EE,  HANGUL SYLLABLE-D6EE
        0xD6EF, # G+18022  U+D6EF  uniD6EF,  HANGUL SYLLABLE-D6EF
        0xD6F1, # G+18023  U+D6F1  uniD6F1,  HANGUL SYLLABLE-D6F1
        0xD6F2, # G+18024  U+D6F2  uniD6F2,  HANGUL SYLLABLE-D6F2
        0xD6F3, # G+18025  U+D6F3  uniD6F3,  HANGUL SYLLABLE-D6F3
        0xD6F4, # G+18026  U+D6F4  uniD6F4,  HANGUL SYLLABLE-D6F4
        0xD6F6, # G+18027  U+D6F6  uniD6F6,  HANGUL SYLLABLE-D6F6
        0xD6F7, # G+18028  U+D6F7  uniD6F7,  HANGUL SYLLABLE-D6F7
        0xD6F8, # G+18029  U+D6F8  uniD6F8,  HANGUL SYLLABLE-D6F8
        0xD6F9, # G+18030  U+D6F9  uniD6F9,  HANGUL SYLLABLE-D6F9
        0xD6FA, # G+18031  U+D6FA  uniD6FA,  HANGUL SYLLABLE-D6FA
        0xD6FB, # G+18032  U+D6FB  uniD6FB,  HANGUL SYLLABLE-D6FB
        0xD6FE, # G+18033  U+D6FE  uniD6FE,  HANGUL SYLLABLE-D6FE
        0xD6FF, # G+18034  U+D6FF  uniD6FF,  HANGUL SYLLABLE-D6FF
        0xD701, # G+18035  U+D701  uniD701,  HANGUL SYLLABLE-D701
        0xD702, # G+18036  U+D702  uniD702,  HANGUL SYLLABLE-D702
        0xD703, # G+18037  U+D703  uniD703,  HANGUL SYLLABLE-D703
        0xD705, # G+18038  U+D705  uniD705,  HANGUL SYLLABLE-D705
        0xD706, # G+18039  U+D706  uniD706,  HANGUL SYLLABLE-D706
        0xD707, # G+18040  U+D707  uniD707,  HANGUL SYLLABLE-D707
        0xD708, # G+18041  U+D708  uniD708,  HANGUL SYLLABLE-D708
        0xD709, # G+18042  U+D709  uniD709,  HANGUL SYLLABLE-D709
        0xD70A, # G+18043  U+D70A  uniD70A,  HANGUL SYLLABLE-D70A
        0xD70B, # G+18044  U+D70B  uniD70B,  HANGUL SYLLABLE-D70B
        0xD70C, # G+18045  U+D70C  uniD70C,  HANGUL SYLLABLE-D70C
        0xD70D, # G+18046  U+D70D  uniD70D,  HANGUL SYLLABLE-D70D
        0xD70E, # G+18047  U+D70E  uniD70E,  HANGUL SYLLABLE-D70E
        0xD70F, # G+18048  U+D70F  uniD70F,  HANGUL SYLLABLE-D70F
        0xD710, # G+18049  U+D710  uniD710,  HANGUL SYLLABLE-D710
        0xD712, # G+18050  U+D712  uniD712,  HANGUL SYLLABLE-D712
        0xD713, # G+18051  U+D713  uniD713,  HANGUL SYLLABLE-D713
        0xD714, # G+18052  U+D714  uniD714,  HANGUL SYLLABLE-D714
        0xD715, # G+18053  U+D715  uniD715,  HANGUL SYLLABLE-D715
        0xD716, # G+18054  U+D716  uniD716,  HANGUL SYLLABLE-D716
        0xD717, # G+18055  U+D717  uniD717,  HANGUL SYLLABLE-D717
        0xD71A, # G+18056  U+D71A  uniD71A,  HANGUL SYLLABLE-D71A
        0xD71B, # G+18057  U+D71B  uniD71B,  HANGUL SYLLABLE-D71B
        0xD71D, # G+18058  U+D71D  uniD71D,  HANGUL SYLLABLE-D71D
        0xD71E, # G+18059  U+D71E  uniD71E,  HANGUL SYLLABLE-D71E
        0xD71F, # G+18060  U+D71F  uniD71F,  HANGUL SYLLABLE-D71F
        0xD721, # G+18061  U+D721  uniD721,  HANGUL SYLLABLE-D721
        0xD722, # G+18062  U+D722  uniD722,  HANGUL SYLLABLE-D722
        0xD723, # G+18063  U+D723  uniD723,  HANGUL SYLLABLE-D723
        0xD724, # G+18064  U+D724  uniD724,  HANGUL SYLLABLE-D724
        0xD725, # G+18065  U+D725  uniD725,  HANGUL SYLLABLE-D725
        0xD726, # G+18066  U+D726  uniD726,  HANGUL SYLLABLE-D726
        0xD727, # G+18067  U+D727  uniD727,  HANGUL SYLLABLE-D727
        0xD72A, # G+18068  U+D72A  uniD72A,  HANGUL SYLLABLE-D72A
        0xD72C, # G+18069  U+D72C  uniD72C,  HANGUL SYLLABLE-D72C
        0xD72E, # G+18070  U+D72E  uniD72E,  HANGUL SYLLABLE-D72E
        0xD72F, # G+18071  U+D72F  uniD72F,  HANGUL SYLLABLE-D72F
        0xD730, # G+18072  U+D730  uniD730,  HANGUL SYLLABLE-D730
        0xD731, # G+18073  U+D731  uniD731,  HANGUL SYLLABLE-D731
        0xD732, # G+18074  U+D732  uniD732,  HANGUL SYLLABLE-D732
        0xD733, # G+18075  U+D733  uniD733,  HANGUL SYLLABLE-D733
        0xD736, # G+18076  U+D736  uniD736,  HANGUL SYLLABLE-D736
        0xD737, # G+18077  U+D737  uniD737,  HANGUL SYLLABLE-D737
        0xD739, # G+18078  U+D739  uniD739,  HANGUL SYLLABLE-D739
        0xD73A, # G+18079  U+D73A  uniD73A,  HANGUL SYLLABLE-D73A
        0xD73B, # G+18080  U+D73B  uniD73B,  HANGUL SYLLABLE-D73B
        0xD73D, # G+18081  U+D73D  uniD73D,  HANGUL SYLLABLE-D73D
        0xD73E, # G+18082  U+D73E  uniD73E,  HANGUL SYLLABLE-D73E
        0xD73F, # G+18083  U+D73F  uniD73F,  HANGUL SYLLABLE-D73F
        0xD740, # G+18084  U+D740  uniD740,  HANGUL SYLLABLE-D740
        0xD741, # G+18085  U+D741  uniD741,  HANGUL SYLLABLE-D741
        0xD742, # G+18086  U+D742  uniD742,  HANGUL SYLLABLE-D742
        0xD743, # G+18087  U+D743  uniD743,  HANGUL SYLLABLE-D743
        0xD745, # G+18088  U+D745  uniD745,  HANGUL SYLLABLE-D745
        0xD746, # G+18089  U+D746  uniD746,  HANGUL SYLLABLE-D746
        0xD748, # G+18090  U+D748  uniD748,  HANGUL SYLLABLE-D748
        0xD74A, # G+18091  U+D74A  uniD74A,  HANGUL SYLLABLE-D74A
        0xD74B, # G+18092  U+D74B  uniD74B,  HANGUL SYLLABLE-D74B
        0xD74C, # G+18093  U+D74C  uniD74C,  HANGUL SYLLABLE-D74C
        0xD74D, # G+18094  U+D74D  uniD74D,  HANGUL SYLLABLE-D74D
        0xD74E, # G+18095  U+D74E  uniD74E,  HANGUL SYLLABLE-D74E
        0xD74F, # G+18096  U+D74F  uniD74F,  HANGUL SYLLABLE-D74F
        0xD752, # G+18097  U+D752  uniD752,  HANGUL SYLLABLE-D752
        0xD753, # G+18098  U+D753  uniD753,  HANGUL SYLLABLE-D753
        0xD755, # G+18099  U+D755  uniD755,  HANGUL SYLLABLE-D755
        0xD75A, # G+18100  U+D75A  uniD75A,  HANGUL SYLLABLE-D75A
        0xD75B, # G+18101  U+D75B  uniD75B,  HANGUL SYLLABLE-D75B
        0xD75C, # G+18102  U+D75C  uniD75C,  HANGUL SYLLABLE-D75C
        0xD75D, # G+18103  U+D75D  uniD75D,  HANGUL SYLLABLE-D75D
        0xD75E, # G+18104  U+D75E  uniD75E,  HANGUL SYLLABLE-D75E
        0xD75F, # G+18105  U+D75F  uniD75F,  HANGUL SYLLABLE-D75F
        0xD762, # G+18106  U+D762  uniD762,  HANGUL SYLLABLE-D762
        0xD764, # G+18107  U+D764  uniD764,  HANGUL SYLLABLE-D764
        0xD766, # G+18108  U+D766  uniD766,  HANGUL SYLLABLE-D766
        0xD767, # G+18109  U+D767  uniD767,  HANGUL SYLLABLE-D767
        0xD768, # G+18110  U+D768  uniD768,  HANGUL SYLLABLE-D768
        0xD76A, # G+18111  U+D76A  uniD76A,  HANGUL SYLLABLE-D76A
        0xD76B, # G+18112  U+D76B  uniD76B,  HANGUL SYLLABLE-D76B
        0xD76D, # G+18113  U+D76D  uniD76D,  HANGUL SYLLABLE-D76D
        0xD76E, # G+18114  U+D76E  uniD76E,  HANGUL SYLLABLE-D76E
        0xD76F, # G+18115  U+D76F  uniD76F,  HANGUL SYLLABLE-D76F
        0xD771, # G+18116  U+D771  uniD771,  HANGUL SYLLABLE-D771
        0xD772, # G+18117  U+D772  uniD772,  HANGUL SYLLABLE-D772
        0xD773, # G+18118  U+D773  uniD773,  HANGUL SYLLABLE-D773
        0xD775, # G+18119  U+D775  uniD775,  HANGUL SYLLABLE-D775
        0xD776, # G+18120  U+D776  uniD776,  HANGUL SYLLABLE-D776
        0xD777, # G+18121  U+D777  uniD777,  HANGUL SYLLABLE-D777
        0xD778, # G+18122  U+D778  uniD778,  HANGUL SYLLABLE-D778
        0xD779, # G+18123  U+D779  uniD779,  HANGUL SYLLABLE-D779
        0xD77A, # G+18124  U+D77A  uniD77A,  HANGUL SYLLABLE-D77A
        0xD77B, # G+18125  U+D77B  uniD77B,  HANGUL SYLLABLE-D77B
        0xD77E, # G+18126  U+D77E  uniD77E,  HANGUL SYLLABLE-D77E
        0xD77F, # G+18127  U+D77F  uniD77F,  HANGUL SYLLABLE-D77F
        0xD780, # G+18128  U+D780  uniD780,  HANGUL SYLLABLE-D780
        0xD782, # G+18129  U+D782  uniD782,  HANGUL SYLLABLE-D782
        0xD783, # G+18130  U+D783  uniD783,  HANGUL SYLLABLE-D783
        0xD784, # G+18131  U+D784  uniD784,  HANGUL SYLLABLE-D784
        0xD785, # G+18132  U+D785  uniD785,  HANGUL SYLLABLE-D785
        0xD786, # G+18133  U+D786  uniD786,  HANGUL SYLLABLE-D786
        0xD787, # G+18134  U+D787  uniD787,  HANGUL SYLLABLE-D787
        0xD78A, # G+18135  U+D78A  uniD78A,  HANGUL SYLLABLE-D78A
        0xD78B, # G+18136  U+D78B  uniD78B,  HANGUL SYLLABLE-D78B
        0xD78D, # G+18137  U+D78D  uniD78D,  HANGUL SYLLABLE-D78D
        0xD78E, # G+18138  U+D78E  uniD78E,  HANGUL SYLLABLE-D78E
        0xD78F, # G+18139  U+D78F  uniD78F,  HANGUL SYLLABLE-D78F
        0xD791, # G+18140  U+D791  uniD791,  HANGUL SYLLABLE-D791
        0xD792, # G+18141  U+D792  uniD792,  HANGUL SYLLABLE-D792
        0xD793, # G+18142  U+D793  uniD793,  HANGUL SYLLABLE-D793
        0xD794, # G+18143  U+D794  uniD794,  HANGUL SYLLABLE-D794
        0xD795, # G+18144  U+D795  uniD795,  HANGUL SYLLABLE-D795
        0xD796, # G+18145  U+D796  uniD796,  HANGUL SYLLABLE-D796
        0xD797, # G+18146  U+D797  uniD797,  HANGUL SYLLABLE-D797
        0xD79A, # G+18147  U+D79A  uniD79A,  HANGUL SYLLABLE-D79A
        0xD79C, # G+18148  U+D79C  uniD79C,  HANGUL SYLLABLE-D79C
        0xD79E, # G+18149  U+D79E  uniD79E,  HANGUL SYLLABLE-D79E
        0xD79F, # G+18150  U+D79F  uniD79F,  HANGUL SYLLABLE-D79F
        0xD7A0, # G+18151  U+D7A0  uniD7A0,  HANGUL SYLLABLE-D7A0
        0xD7A1, # G+18152  U+D7A1  uniD7A1,  HANGUL SYLLABLE-D7A1
        0xD7A2, # G+18153  U+D7A2  uniD7A2,  HANGUL SYLLABLE-D7A2
        0xD7A3, # G+18154  U+D7A3  uniD7A3,  HANGUL SYLLABLE-D7A3
    ],
    'u2g' => {
        '32' => '8094'                , # G+8094  U+0020  space,  SPACE
        '33' => '8095'                , # G+8095  U+0021  exclam,  EXCLAMATION MARK
        '34' => '8096'                , # G+8096  U+0022  quotedbl,  QUOTATION MARK
        '35' => '8097'                , # G+8097  U+0023  numbersign,  NUMBER SIGN
        '36' => '8098'                , # G+8098  U+0024  dollar,  DOLLAR SIGN
        '37' => '8099'                , # G+8099  U+0025  percent,  PERCENT SIGN
        '38' => '8100'                , # G+8100  U+0026  ampersand,  AMPERSAND
        '39' => '8101'                , # G+8101  U+0027  quotesingle,  APOSTROPHE
        '40' => '8102'                , # G+8102  U+0028  parenleft,  LEFT PARENTHESIS
        '41' => '8103'                , # G+8103  U+0029  parenright,  RIGHT PARENTHESIS
        '42' => '8104'                , # G+8104  U+002A  asterisk,  ASTERISK
        '43' => '8105'                , # G+8105  U+002B  plus,  PLUS SIGN
        '44' => '8106'                , # G+8106  U+002C  comma,  COMMA
        '45' => '8107'                , # G+8107  U+002D  hyphen,  HYPHEN-MINUS
        '46' => '8108'                , # G+8108  U+002E  period,  FULL STOP
        '47' => '8109'                , # G+8109  U+002F  slash,  SOLIDUS
        '48' => '8110'                , # G+8110  U+0030  zero,  DIGIT ZERO
        '49' => '8111'                , # G+8111  U+0031  one,  DIGIT ONE
        '50' => '8112'                , # G+8112  U+0032  two,  DIGIT TWO
        '51' => '8113'                , # G+8113  U+0033  three,  DIGIT THREE
        '52' => '8114'                , # G+8114  U+0034  four,  DIGIT FOUR
        '53' => '8115'                , # G+8115  U+0035  five,  DIGIT FIVE
        '54' => '8116'                , # G+8116  U+0036  six,  DIGIT SIX
        '55' => '8117'                , # G+8117  U+0037  seven,  DIGIT SEVEN
        '56' => '8118'                , # G+8118  U+0038  eight,  DIGIT EIGHT
        '57' => '8119'                , # G+8119  U+0039  nine,  DIGIT NINE
        '58' => '8120'                , # G+8120  U+003A  colon,  COLON
        '59' => '8121'                , # G+8121  U+003B  semicolon,  SEMICOLON
        '60' => '8122'                , # G+8122  U+003C  less,  LESS-THAN SIGN
        '61' => '8123'                , # G+8123  U+003D  equal,  EQUALS SIGN
        '62' => '8124'                , # G+8124  U+003E  greater,  GREATER-THAN SIGN
        '63' => '8125'                , # G+8125  U+003F  question,  QUESTION MARK
        '64' => '8126'                , # G+8126  U+0040  at,  COMMERCIAL AT
        '65' => '8127'                , # G+8127  U+0041  A,  LATIN CAPITAL LETTER A
        '66' => '8128'                , # G+8128  U+0042  B,  LATIN CAPITAL LETTER B
        '67' => '8129'                , # G+8129  U+0043  C,  LATIN CAPITAL LETTER C
        '68' => '8130'                , # G+8130  U+0044  D,  LATIN CAPITAL LETTER D
        '69' => '8131'                , # G+8131  U+0045  E,  LATIN CAPITAL LETTER E
        '70' => '8132'                , # G+8132  U+0046  F,  LATIN CAPITAL LETTER F
        '71' => '8133'                , # G+8133  U+0047  G,  LATIN CAPITAL LETTER G
        '72' => '8134'                , # G+8134  U+0048  H,  LATIN CAPITAL LETTER H
        '73' => '8135'                , # G+8135  U+0049  I,  LATIN CAPITAL LETTER I
        '74' => '8136'                , # G+8136  U+004A  J,  LATIN CAPITAL LETTER J
        '75' => '8137'                , # G+8137  U+004B  K,  LATIN CAPITAL LETTER K
        '76' => '8138'                , # G+8138  U+004C  L,  LATIN CAPITAL LETTER L
        '77' => '8139'                , # G+8139  U+004D  M,  LATIN CAPITAL LETTER M
        '78' => '8140'                , # G+8140  U+004E  N,  LATIN CAPITAL LETTER N
        '79' => '8141'                , # G+8141  U+004F  O,  LATIN CAPITAL LETTER O
        '80' => '8142'                , # G+8142  U+0050  P,  LATIN CAPITAL LETTER P
        '81' => '8143'                , # G+8143  U+0051  Q,  LATIN CAPITAL LETTER Q
        '82' => '8144'                , # G+8144  U+0052  R,  LATIN CAPITAL LETTER R
        '83' => '8145'                , # G+8145  U+0053  S,  LATIN CAPITAL LETTER S
        '84' => '8146'                , # G+8146  U+0054  T,  LATIN CAPITAL LETTER T
        '85' => '8147'                , # G+8147  U+0055  U,  LATIN CAPITAL LETTER U
        '86' => '8148'                , # G+8148  U+0056  V,  LATIN CAPITAL LETTER V
        '87' => '8149'                , # G+8149  U+0057  W,  LATIN CAPITAL LETTER W
        '88' => '8150'                , # G+8150  U+0058  X,  LATIN CAPITAL LETTER X
        '89' => '8151'                , # G+8151  U+0059  Y,  LATIN CAPITAL LETTER Y
        '90' => '8152'                , # G+8152  U+005A  Z,  LATIN CAPITAL LETTER Z
        '91' => '8153'                , # G+8153  U+005B  bracketleft,  LEFT SQUARE BRACKET
        '92' => '8154'                , # G+8154  U+005C  backslash,  REVERSE SOLIDUS
        '93' => '8155'                , # G+8155  U+005D  bracketright,  RIGHT SQUARE BRACKET
        '94' => '8156'                , # G+8156  U+005E  asciicircum,  CIRCUMFLEX ACCENT
        '95' => '8157'                , # G+8157  U+005F  underscore,  LOW LINE
        '96' => '8158'                , # G+8158  U+0060  grave,  GRAVE ACCENT
        '97' => '8159'                , # G+8159  U+0061  a,  LATIN SMALL LETTER A
        '98' => '8160'                , # G+8160  U+0062  b,  LATIN SMALL LETTER B
        '99' => '8161'                , # G+8161  U+0063  c,  LATIN SMALL LETTER C
        '100' => '8162'               , # G+8162  U+0064  d,  LATIN SMALL LETTER D
        '101' => '8163'               , # G+8163  U+0065  e,  LATIN SMALL LETTER E
        '102' => '8164'               , # G+8164  U+0066  f,  LATIN SMALL LETTER F
        '103' => '8165'               , # G+8165  U+0067  g,  LATIN SMALL LETTER G
        '104' => '8166'               , # G+8166  U+0068  h,  LATIN SMALL LETTER H
        '105' => '8167'               , # G+8167  U+0069  i,  LATIN SMALL LETTER I
        '106' => '8168'               , # G+8168  U+006A  j,  LATIN SMALL LETTER J
        '107' => '8169'               , # G+8169  U+006B  k,  LATIN SMALL LETTER K
        '108' => '8170'               , # G+8170  U+006C  l,  LATIN SMALL LETTER L
        '109' => '8171'               , # G+8171  U+006D  m,  LATIN SMALL LETTER M
        '110' => '8172'               , # G+8172  U+006E  n,  LATIN SMALL LETTER N
        '111' => '8173'               , # G+8173  U+006F  o,  LATIN SMALL LETTER O
        '112' => '8174'               , # G+8174  U+0070  p,  LATIN SMALL LETTER P
        '113' => '8175'               , # G+8175  U+0071  q,  LATIN SMALL LETTER Q
        '114' => '8176'               , # G+8176  U+0072  r,  LATIN SMALL LETTER R
        '115' => '8177'               , # G+8177  U+0073  s,  LATIN SMALL LETTER S
        '116' => '8178'               , # G+8178  U+0074  t,  LATIN SMALL LETTER T
        '117' => '8179'               , # G+8179  U+0075  u,  LATIN SMALL LETTER U
        '118' => '8180'               , # G+8180  U+0076  v,  LATIN SMALL LETTER V
        '119' => '8181'               , # G+8181  U+0077  w,  LATIN SMALL LETTER W
        '120' => '8182'               , # G+8182  U+0078  x,  LATIN SMALL LETTER X
        '121' => '8183'               , # G+8183  U+0079  y,  LATIN SMALL LETTER Y
        '122' => '8184'               , # G+8184  U+007A  z,  LATIN SMALL LETTER Z
        '123' => '8185'               , # G+8185  U+007B  braceleft,  LEFT CURLY BRACKET
        '124' => '8186'               , # G+8186  U+007C  bar,  VERTICAL LINE
        '125' => '8187'               , # G+8187  U+007D  braceright,  RIGHT CURLY BRACKET
        '126' => '8188'               , # G+8188  U+007E  asciitilde,  TILDE
        '161' => '208'                , # G+208  U+00A1  exclamdown,  INVERTED EXCLAMATION MARK
        '164' => '214'                , # G+214  U+00A4  currency,  CURRENCY SIGN
        '167' => '155'                , # G+155  U+00A7  section,  SECTION SIGN
        '168' => '107'                , # G+107  U+00A8  dieresis,  DIAERESIS
        '169' => '98'                 , # G+98  U+00A9  copyright,  COPYRIGHT SIGN
        '170' => '668'                , # G+668  U+00AA  ordfeminine,  FEMININE ORDINAL INDICATOR
        '171' => '176'                , # G+176  U+00AB  guillemotleft,  LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
        '173' => '109'                , # G+109  U+00AD  hyphen,  SOFT HYPHEN
        '176' => '138'                , # G+138  U+00B0  degree,  DEGREE SIGN
        '177' => '130'                , # G+130  U+00B1  plusminus,  PLUS-MINUS SIGN
        '178' => '843'                , # G+843  U+00B2  twosuperior,  SUPERSCRIPT TWO
        '179' => '844'                , # G+844  U+00B3  threesuperior,  SUPERSCRIPT THREE
        '180' => '199'                , # G+199  U+00B4  acute,  ACUTE ACCENT
        '182' => '244'                , # G+244  U+00B6  paragraph,  PILCROW SIGN
        '183' => '104'                , # G+104  U+00B7  periodcentered,  MIDDLE DOT
        '184' => '206'                , # G+206  U+00B8  cedilla,  CEDILLA
        '185' => '842'                , # G+842  U+00B9  onesuperior,  SUPERSCRIPT ONE
        '186' => '675'                , # G+675  U+00BA  ordmasculine,  MASCULINE ORDINAL INDICATOR
        '187' => '177'                , # G+177  U+00BB  guillemotright,  RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
        '188' => '751'                , # G+751  U+00BC  onequarter,  VULGAR FRACTION ONE QUARTER
        '189' => '748'                , # G+748  U+00BD  onehalf,  VULGAR FRACTION ONE HALF
        '190' => '752'                , # G+752  U+00BE  threequarters,  VULGAR FRACTION THREE QUARTERS
        '191' => '209'                , # G+209  U+00BF  questiondown,  INVERTED QUESTION MARK
        '198' => '666'                , # G+666  U+00C6  AE,  LATIN CAPITAL LETTER AE
        '208' => '667'                , # G+667  U+00D0  Eth,  LATIN CAPITAL LETTER ETH
        '215' => '131'                , # G+131  U+00D7  multiply,  MULTIPLICATION SIGN
        '216' => '673'                , # G+673  U+00D8  Oslash,  LATIN CAPITAL LETTER O WITH STROKE
        '222' => '676'                , # G+676  U+00DE  Thorn,  LATIN CAPITAL LETTER THORN
        '223' => '768'                , # G+768  U+00DF  germandbls,  LATIN SMALL LETTER SHARP S
        '230' => '757'                , # G+757  U+00E6  ae,  LATIN SMALL LETTER AE
        '240' => '759'                , # G+759  U+00F0  eth,  LATIN SMALL LETTER ETH
        '247' => '132'                , # G+132  U+00F7  divide,  DIVISION SIGN
        '248' => '766'                , # G+766  U+00F8  oslash,  LATIN SMALL LETTER O WITH STROKE
        '254' => '769'                , # G+769  U+00FE  thorn,  LATIN SMALL LETTER THORN
        '273' => '758'                , # G+758  U+0111  dcroat,  LATIN SMALL LETTER D WITH STROKE
        '294' => '669'                , # G+669  U+0126  Hbar,  LATIN CAPITAL LETTER H WITH STROKE
        '295' => '760'                , # G+760  U+0127  hbar,  LATIN SMALL LETTER H WITH STROKE
        '305' => '761'                , # G+761  U+0131  dotlessi,  LATIN SMALL LETTER DOTLESS I
        '306' => '670'                , # G+670  U+0132  IJ,  LATIN CAPITAL LIGATURE IJ
        '307' => '762'                , # G+762  U+0133  ij,  LATIN SMALL LIGATURE IJ
        '312' => '763'                , # G+763  U+0138  kgreenlandic,  LATIN SMALL LETTER KRA
        '319' => '671'                , # G+671  U+013F  Ldot,  LATIN CAPITAL LETTER L WITH MIDDLE DOT
        '320' => '764'                , # G+764  U+0140  ldot,  LATIN SMALL LETTER L WITH MIDDLE DOT
        '321' => '672'                , # G+672  U+0141  Lslash,  LATIN CAPITAL LETTER L WITH STROKE
        '322' => '765'                , # G+765  U+0142  lslash,  LATIN SMALL LETTER L WITH STROKE
        '329' => '772'                , # G+772  U+0149  napostrophe,  LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
        '330' => '678'                , # G+678  U+014A  Eng,  LATIN CAPITAL LETTER ENG
        '331' => '771'                , # G+771  U+014B  eng,  LATIN SMALL LETTER ENG
        '338' => '674'                , # G+674  U+0152  OE,  LATIN CAPITAL LIGATURE OE
        '339' => '767'                , # G+767  U+0153  oe,  LATIN SMALL LIGATURE OE
        '358' => '677'                , # G+677  U+0166  Tbar,  LATIN CAPITAL LETTER T WITH STROKE
        '359' => '770'                , # G+770  U+0167  tbar,  LATIN SMALL LETTER T WITH STROKE
        '700' => '8275'               , # G+8275  U+02BC  afii57929,  MODIFIER LETTER APOSTROPHE
        '711' => '201'                , # G+201  U+02C7  caron,  CARON
        '720' => '210'                , # G+210  U+02D0  colontriangularmod,  MODIFIER LETTER TRIANGULAR COLON
        '728' => '202'                , # G+202  U+02D8  breve,  BREVE
        '729' => '205'                , # G+205  U+02D9  dotaccent,  DOT ABOVE
        '730' => '204'                , # G+204  U+02DA  ring,  RING ABOVE
        '731' => '207'                , # G+207  U+02DB  ogonek,  OGONEK
        '732' => '200'                , # G+200  U+02DC  tilde,  SMALL TILDE
        '733' => '203'                , # G+203  U+02DD  hungarumlaut,  DOUBLE ACUTE ACCENT
        '913' => '471'                , # G+471  U+0391  Alpha,  GREEK CAPITAL LETTER ALPHA
        '914' => '472'                , # G+472  U+0392  Beta,  GREEK CAPITAL LETTER BETA
        '915' => '473'                , # G+473  U+0393  Gamma,  GREEK CAPITAL LETTER GAMMA
        '916' => '474'                , # G+474  U+0394  Delta,  GREEK CAPITAL LETTER DELTA
        '917' => '475'                , # G+475  U+0395  Epsilon,  GREEK CAPITAL LETTER EPSILON
        '918' => '476'                , # G+476  U+0396  Zeta,  GREEK CAPITAL LETTER ZETA
        '919' => '477'                , # G+477  U+0397  Eta,  GREEK CAPITAL LETTER ETA
        '920' => '478'                , # G+478  U+0398  Theta,  GREEK CAPITAL LETTER THETA
        '921' => '479'                , # G+479  U+0399  Iota,  GREEK CAPITAL LETTER IOTA
        '922' => '480'                , # G+480  U+039A  Kappa,  GREEK CAPITAL LETTER KAPPA
        '923' => '481'                , # G+481  U+039B  Lambda,  GREEK CAPITAL LETTER LAMDA
        '924' => '482'                , # G+482  U+039C  Mu,  GREEK CAPITAL LETTER MU
        '925' => '483'                , # G+483  U+039D  Nu,  GREEK CAPITAL LETTER NU
        '926' => '484'                , # G+484  U+039E  Xi,  GREEK CAPITAL LETTER XI
        '927' => '485'                , # G+485  U+039F  Omicron,  GREEK CAPITAL LETTER OMICRON
        '928' => '486'                , # G+486  U+03A0  Pi,  GREEK CAPITAL LETTER PI
        '929' => '487'                , # G+487  U+03A1  Rho,  GREEK CAPITAL LETTER RHO
        '931' => '488'                , # G+488  U+03A3  Sigma,  GREEK CAPITAL LETTER SIGMA
        '932' => '489'                , # G+489  U+03A4  Tau,  GREEK CAPITAL LETTER TAU
        '933' => '490'                , # G+490  U+03A5  Upsilon,  GREEK CAPITAL LETTER UPSILON
        '934' => '491'                , # G+491  U+03A6  Phi,  GREEK CAPITAL LETTER PHI
        '935' => '492'                , # G+492  U+03A7  Chi,  GREEK CAPITAL LETTER CHI
        '936' => '493'                , # G+493  U+03A8  Psi,  GREEK CAPITAL LETTER PSI
        '937' => '494'                , # G+494  U+03A9  Omega,  GREEK CAPITAL LETTER OMEGA
        '945' => '495'                , # G+495  U+03B1  alpha,  GREEK SMALL LETTER ALPHA
        '946' => '496'                , # G+496  U+03B2  beta,  GREEK SMALL LETTER BETA
        '947' => '497'                , # G+497  U+03B3  gamma,  GREEK SMALL LETTER GAMMA
        '948' => '498'                , # G+498  U+03B4  delta,  GREEK SMALL LETTER DELTA
        '949' => '499'                , # G+499  U+03B5  epsilon,  GREEK SMALL LETTER EPSILON
        '950' => '500'                , # G+500  U+03B6  zeta,  GREEK SMALL LETTER ZETA
        '951' => '501'                , # G+501  U+03B7  eta,  GREEK SMALL LETTER ETA
        '952' => '502'                , # G+502  U+03B8  theta,  GREEK SMALL LETTER THETA
        '953' => '503'                , # G+503  U+03B9  iota,  GREEK SMALL LETTER IOTA
        '954' => '504'                , # G+504  U+03BA  kappa,  GREEK SMALL LETTER KAPPA
        '955' => '505'                , # G+505  U+03BB  lambda,  GREEK SMALL LETTER LAMDA
        '956' => '506'                , # G+506  U+03BC  mu,  GREEK SMALL LETTER MU
        '957' => '507'                , # G+507  U+03BD  nu,  GREEK SMALL LETTER NU
        '958' => '508'                , # G+508  U+03BE  xi,  GREEK SMALL LETTER XI
        '959' => '509'                , # G+509  U+03BF  omicron,  GREEK SMALL LETTER OMICRON
        '960' => '510'                , # G+510  U+03C0  pi,  GREEK SMALL LETTER PI
        '961' => '511'                , # G+511  U+03C1  rho,  GREEK SMALL LETTER RHO
        '963' => '512'                , # G+512  U+03C3  sigma,  GREEK SMALL LETTER SIGMA
        '964' => '513'                , # G+513  U+03C4  tau,  GREEK SMALL LETTER TAU
        '965' => '514'                , # G+514  U+03C5  upsilon,  GREEK SMALL LETTER UPSILON
        '966' => '515'                , # G+515  U+03C6  phi,  GREEK SMALL LETTER PHI
        '967' => '516'                , # G+516  U+03C7  chi,  GREEK SMALL LETTER CHI
        '968' => '517'                , # G+517  U+03C8  psi,  GREEK SMALL LETTER PSI
        '969' => '518'                , # G+518  U+03C9  omega,  GREEK SMALL LETTER OMEGA
        '1025' => '1026'              , # G+1026  U+0401  afii10023,  CYRILLIC CAPITAL LETTER IO
        '1040' => '1020'              , # G+1020  U+0410  afii10017,  CYRILLIC CAPITAL LETTER A
        '1041' => '1021'              , # G+1021  U+0411  afii10018,  CYRILLIC CAPITAL LETTER BE
        '1042' => '1022'              , # G+1022  U+0412  afii10019,  CYRILLIC CAPITAL LETTER VE
        '1043' => '1023'              , # G+1023  U+0413  afii10020,  CYRILLIC CAPITAL LETTER GHE
        '1044' => '1024'              , # G+1024  U+0414  afii10021,  CYRILLIC CAPITAL LETTER DE
        '1045' => '1025'              , # G+1025  U+0415  afii10022,  CYRILLIC CAPITAL LETTER IE
        '1046' => '1027'              , # G+1027  U+0416  afii10024,  CYRILLIC CAPITAL LETTER ZHE
        '1047' => '1028'              , # G+1028  U+0417  afii10025,  CYRILLIC CAPITAL LETTER ZE
        '1048' => '1029'              , # G+1029  U+0418  afii10026,  CYRILLIC CAPITAL LETTER I
        '1049' => '1030'              , # G+1030  U+0419  afii10027,  CYRILLIC CAPITAL LETTER SHORT I
        '1050' => '1031'              , # G+1031  U+041A  afii10028,  CYRILLIC CAPITAL LETTER KA
        '1051' => '1032'              , # G+1032  U+041B  afii10029,  CYRILLIC CAPITAL LETTER EL
        '1052' => '1033'              , # G+1033  U+041C  afii10030,  CYRILLIC CAPITAL LETTER EM
        '1053' => '1034'              , # G+1034  U+041D  afii10031,  CYRILLIC CAPITAL LETTER EN
        '1054' => '1035'              , # G+1035  U+041E  afii10032,  CYRILLIC CAPITAL LETTER O
        '1055' => '1036'              , # G+1036  U+041F  afii10033,  CYRILLIC CAPITAL LETTER PE
        '1056' => '1037'              , # G+1037  U+0420  afii10034,  CYRILLIC CAPITAL LETTER ER
        '1057' => '1038'              , # G+1038  U+0421  afii10035,  CYRILLIC CAPITAL LETTER ES
        '1058' => '1039'              , # G+1039  U+0422  afii10036,  CYRILLIC CAPITAL LETTER TE
        '1059' => '1040'              , # G+1040  U+0423  afii10037,  CYRILLIC CAPITAL LETTER U
        '1060' => '1041'              , # G+1041  U+0424  afii10038,  CYRILLIC CAPITAL LETTER EF
        '1061' => '1042'              , # G+1042  U+0425  afii10039,  CYRILLIC CAPITAL LETTER HA
        '1062' => '1043'              , # G+1043  U+0426  afii10040,  CYRILLIC CAPITAL LETTER TSE
        '1063' => '1044'              , # G+1044  U+0427  afii10041,  CYRILLIC CAPITAL LETTER CHE
        '1064' => '1045'              , # G+1045  U+0428  afii10042,  CYRILLIC CAPITAL LETTER SHA
        '1065' => '1046'              , # G+1046  U+0429  afii10043,  CYRILLIC CAPITAL LETTER SHCHA
        '1066' => '1047'              , # G+1047  U+042A  afii10044,  CYRILLIC CAPITAL LETTER HARD SIGN
        '1067' => '1048'              , # G+1048  U+042B  afii10045,  CYRILLIC CAPITAL LETTER YERU
        '1068' => '1049'              , # G+1049  U+042C  afii10046,  CYRILLIC CAPITAL LETTER SOFT SIGN
        '1069' => '1050'              , # G+1050  U+042D  afii10047,  CYRILLIC CAPITAL LETTER E
        '1070' => '1051'              , # G+1051  U+042E  afii10048,  CYRILLIC CAPITAL LETTER YU
        '1071' => '1052'              , # G+1052  U+042F  afii10049,  CYRILLIC CAPITAL LETTER YA
        '1072' => '1053'              , # G+1053  U+0430  afii10065,  CYRILLIC SMALL LETTER A
        '1073' => '1054'              , # G+1054  U+0431  afii10066,  CYRILLIC SMALL LETTER BE
        '1074' => '1055'              , # G+1055  U+0432  afii10067,  CYRILLIC SMALL LETTER VE
        '1075' => '1056'              , # G+1056  U+0433  afii10068,  CYRILLIC SMALL LETTER GHE
        '1076' => '1057'              , # G+1057  U+0434  afii10069,  CYRILLIC SMALL LETTER DE
        '1077' => '1058'              , # G+1058  U+0435  afii10070,  CYRILLIC SMALL LETTER IE
        '1078' => '1060'              , # G+1060  U+0436  afii10072,  CYRILLIC SMALL LETTER ZHE
        '1079' => '1061'              , # G+1061  U+0437  afii10073,  CYRILLIC SMALL LETTER ZE
        '1080' => '1062'              , # G+1062  U+0438  afii10074,  CYRILLIC SMALL LETTER I
        '1081' => '1063'              , # G+1063  U+0439  afii10075,  CYRILLIC SMALL LETTER SHORT I
        '1082' => '1064'              , # G+1064  U+043A  afii10076,  CYRILLIC SMALL LETTER KA
        '1083' => '1065'              , # G+1065  U+043B  afii10077,  CYRILLIC SMALL LETTER EL
        '1084' => '1066'              , # G+1066  U+043C  afii10078,  CYRILLIC SMALL LETTER EM
        '1085' => '1067'              , # G+1067  U+043D  afii10079,  CYRILLIC SMALL LETTER EN
        '1086' => '1068'              , # G+1068  U+043E  afii10080,  CYRILLIC SMALL LETTER O
        '1087' => '1069'              , # G+1069  U+043F  afii10081,  CYRILLIC SMALL LETTER PE
        '1088' => '1070'              , # G+1070  U+0440  afii10082,  CYRILLIC SMALL LETTER ER
        '1089' => '1071'              , # G+1071  U+0441  afii10083,  CYRILLIC SMALL LETTER ES
        '1090' => '1072'              , # G+1072  U+0442  afii10084,  CYRILLIC SMALL LETTER TE
        '1091' => '1073'              , # G+1073  U+0443  afii10085,  CYRILLIC SMALL LETTER U
        '1092' => '1074'              , # G+1074  U+0444  afii10086,  CYRILLIC SMALL LETTER EF
        '1093' => '1075'              , # G+1075  U+0445  afii10087,  CYRILLIC SMALL LETTER HA
        '1094' => '1076'              , # G+1076  U+0446  afii10088,  CYRILLIC SMALL LETTER TSE
        '1095' => '1077'              , # G+1077  U+0447  afii10089,  CYRILLIC SMALL LETTER CHE
        '1096' => '1078'              , # G+1078  U+0448  afii10090,  CYRILLIC SMALL LETTER SHA
        '1097' => '1079'              , # G+1079  U+0449  afii10091,  CYRILLIC SMALL LETTER SHCHA
        '1098' => '1080'              , # G+1080  U+044A  afii10092,  CYRILLIC SMALL LETTER HARD SIGN
        '1099' => '1081'              , # G+1081  U+044B  afii10093,  CYRILLIC SMALL LETTER YERU
        '1100' => '1082'              , # G+1082  U+044C  afii10094,  CYRILLIC SMALL LETTER SOFT SIGN
        '1101' => '1083'              , # G+1083  U+044D  afii10095,  CYRILLIC SMALL LETTER E
        '1102' => '1084'              , # G+1084  U+044E  afii10096,  CYRILLIC SMALL LETTER YU
        '1103' => '1085'              , # G+1085  U+044F  afii10097,  CYRILLIC SMALL LETTER YA
        '1105' => '1059'              , # G+1059  U+0451  afii10071,  CYRILLIC SMALL LETTER IO
        '8208' => '97'                , # G+97  U+2010  hyphentwo,  HYPHEN
        '8211' => '109'               , # G+109  U+2013  endash,  EN DASH
        '8212' => '110'               , # G+110  U+2014  emdash,  EM DASH
        '8213' => '110'               , # G+110  U+2015  afii00208,  HORIZONTAL BAR
        '8214' => '111'               , # G+111  U+2016  dblverticalbar,  DOUBLE VERTICAL LINE
        '8216' => '114'               , # G+114  U+2018  quoteleft,  LEFT SINGLE QUOTATION MARK
        '8217' => '115'               , # G+115  U+2019  quoteright,  RIGHT SINGLE QUOTATION MARK
        '8219' => '8238'              , # G+8238  U+201B  quotereversed,  SINGLE HIGH-REVERSED-9 QUOTATION MARK
        '8220' => '116'               , # G+116  U+201C  quotedblleft,  LEFT DOUBLE QUOTATION MARK
        '8221' => '117'               , # G+117  U+201D  quotedblright,  RIGHT DOUBLE QUOTATION MARK
        '8223' => '8237'              , # G+8237  U+201F  uni201F,  DOUBLE HIGH-REVERSED-9 QUOTATION MARK
        '8224' => '245'               , # G+245  U+2020  dagger,  DAGGER
        '8225' => '246'               , # G+246  U+2021  daggerdbl,  DOUBLE DAGGER
        '8226' => '8607'              , # G+8607  U+2022  bullet,  BULLET
        '8229' => '105'               , # G+105  U+2025  twodotenleader,  TWO DOT LEADER
        '8230' => '106'               , # G+106  U+2026  ellipsis,  HORIZONTAL ELLIPSIS
        '8240' => '216'               , # G+216  U+2030  perthousand,  PER MILLE SIGN
        '8242' => '139'               , # G+139  U+2032  minute,  PRIME
        '8243' => '140'               , # G+140  U+2033  second,  DOUBLE PRIME
        '8244' => '8582'              , # G+8582  U+2034  tprime,  TRIPLE PRIME
        '8245' => '9326'              , # G+9326  U+2035  primereversed,  REVERSED PRIME
        '8246' => '9324'              , # G+9324  U+2036  uni2036,  REVERSED DOUBLE PRIME
        '8249' => '8612'              , # G+8612  U+2039  guilsinglleft,  SINGLE LEFT-POINTING ANGLE QUOTATION MARK
        '8250' => '8613'              , # G+8613  U+203A  guilsinglright,  SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
        '8251' => '156'               , # G+156  U+203B  referencemark,  REFERENCE MARK
        '8252' => '8763'              , # G+8763  U+203C  exclamdbl,  DOUBLE EXCLAMATION MARK
        '8258' => '8599'              , # G+8599  U+2042  asterism,  ASTERISM
        '8308' => '845'               , # G+845  U+2074  foursuperior,  SUPERSCRIPT FOUR
        '8314' => '8239'              , # G+8239  U+207A  plussuperior,  SUPERSCRIPT PLUS SIGN
        '8315' => '8240'              , # G+8240  U+207B  uni207B,  SUPERSCRIPT MINUS
        '8316' => '8248'              , # G+8248  U+207C  equalsuperior,  SUPERSCRIPT EQUALS SIGN
        '8317' => '8250'              , # G+8250  U+207D  parenleftsuperior,  SUPERSCRIPT LEFT PARENTHESIS
        '8318' => '8251'              , # G+8251  U+207E  parenrightsuperior,  SUPERSCRIPT RIGHT PARENTHESIS
        '8319' => '846'               , # G+846  U+207F  nsuperior,  SUPERSCRIPT LATIN SMALL LETTER N
        '8321' => '847'               , # G+847  U+2081  oneinferior,  SUBSCRIPT ONE
        '8322' => '848'               , # G+848  U+2082  twoinferior,  SUBSCRIPT TWO
        '8323' => '849'               , # G+849  U+2083  threeinferior,  SUBSCRIPT THREE
        '8324' => '850'               , # G+850  U+2084  fourinferior,  SUBSCRIPT FOUR
        '8361' => '96'                , # G+96  U+20A9  won,  WON SIGN
        '8451' => '141'               , # G+141  U+2103  centigrade,  DEGREE CELSIUS
        '8457' => '215'               , # G+215  U+2109  fahrenheit,  DEGREE FAHRENHEIT
        '8467' => '590'               , # G+590  U+2113  afii61289,  SCRIPT SMALL L
        '8470' => '258'               , # G+258  U+2116  afii61352,  NUMERO SIGN
        '8481' => '263'               , # G+263  U+2121  telephone,  TELEPHONE SIGN
        '8482' => '260'               , # G+260  U+2122  trademark,  TRADE MARK SIGN
        '8486' => '643'               , # G+643  U+2126  Omega,  OHM SIGN
        '8491' => '142'               , # G+142  U+212B  angstrom,  ANGSTROM SIGN
        '8531' => '749'               , # G+749  U+2153  onethird,  VULGAR FRACTION ONE THIRD
        '8532' => '750'               , # G+750  U+2154  twothirds,  VULGAR FRACTION TWO THIRDS
        '8539' => '753'               , # G+753  U+215B  oneeighth,  VULGAR FRACTION ONE EIGHTH
        '8540' => '754'               , # G+754  U+215C  threeeighths,  VULGAR FRACTION THREE EIGHTHS
        '8541' => '755'               , # G+755  U+215D  fiveeighths,  VULGAR FRACTION FIVE EIGHTHS
        '8542' => '756'               , # G+756  U+215E  seveneighths,  VULGAR FRACTION SEVEN EIGHTHS
        '8544' => '461'               , # G+461  U+2160  Oneroman,  ROMAN NUMERAL ONE
        '8545' => '462'               , # G+462  U+2161  Tworoman,  ROMAN NUMERAL TWO
        '8546' => '463'               , # G+463  U+2162  Threeroman,  ROMAN NUMERAL THREE
        '8547' => '464'               , # G+464  U+2163  Fourroman,  ROMAN NUMERAL FOUR
        '8548' => '465'               , # G+465  U+2164  Fiveroman,  ROMAN NUMERAL FIVE
        '8549' => '466'               , # G+466  U+2165  Sixroman,  ROMAN NUMERAL SIX
        '8550' => '467'               , # G+467  U+2166  Sevenroman,  ROMAN NUMERAL SEVEN
        '8551' => '468'               , # G+468  U+2167  Eightroman,  ROMAN NUMERAL EIGHT
        '8552' => '469'               , # G+469  U+2168  Nineroman,  ROMAN NUMERAL NINE
        '8553' => '470'               , # G+470  U+2169  Tenroman,  ROMAN NUMERAL TEN
        '8560' => '451'               , # G+451  U+2170  oneroman,  SMALL ROMAN NUMERAL ONE
        '8561' => '452'               , # G+452  U+2171  tworoman,  SMALL ROMAN NUMERAL TWO
        '8562' => '453'               , # G+453  U+2172  threeroman,  SMALL ROMAN NUMERAL THREE
        '8563' => '454'               , # G+454  U+2173  fourroman,  SMALL ROMAN NUMERAL FOUR
        '8564' => '455'               , # G+455  U+2174  fiveroman,  SMALL ROMAN NUMERAL FIVE
        '8565' => '456'               , # G+456  U+2175  sixroman,  SMALL ROMAN NUMERAL SIX
        '8566' => '457'               , # G+457  U+2176  sevenroman,  SMALL ROMAN NUMERAL SEVEN
        '8567' => '458'               , # G+458  U+2177  eightroman,  SMALL ROMAN NUMERAL EIGHT
        '8568' => '459'               , # G+459  U+2178  nineroman,  SMALL ROMAN NUMERAL NINE
        '8569' => '460'               , # G+460  U+2179  tenroman,  SMALL ROMAN NUMERAL TEN
        '8592' => '171'               , # G+171  U+2190  arrowleft,  LEFTWARDS ARROW
        '8593' => '172'               , # G+172  U+2191  arrowup,  UPWARDS ARROW
        '8594' => '170'               , # G+170  U+2192  arrowright,  RIGHTWARDS ARROW
        '8595' => '173'               , # G+173  U+2193  arrowdown,  DOWNWARDS ARROW
        '8596' => '174'               , # G+174  U+2194  arrowboth,  LEFT RIGHT ARROW
        '8597' => '247'               , # G+247  U+2195  arrowupdn,  UP DOWN ARROW
        '8598' => '250'               , # G+250  U+2196  arrowupleft,  NORTH WEST ARROW
        '8599' => '248'               , # G+248  U+2197  arrowupright,  NORTH EAST ARROW
        '8600' => '251'               , # G+251  U+2198  arrowdownright,  SOUTH EAST ARROW
        '8601' => '249'               , # G+249  U+2199  arrowdownleft,  SOUTH WEST ARROW
        '8624' => '8868'              , # G+8868  U+21B0  lsh,  UPWARDS ARROW WITH TIP LEFTWARDS
        '8625' => '8865'              , # G+8865  U+21B1  rsh,  UPWARDS ARROW WITH TIP RIGHTWARDS
        '8626' => '8864'              , # G+8864  U+21B2  uni21B2,  DOWNWARDS ARROW WITH TIP LEFTWARDS
        '8627' => '8869'              , # G+8869  U+21B3  uni21B3,  DOWNWARDS ARROW WITH TIP RIGHTWARDS
        '8628' => '8867'              , # G+8867  U+21B4  uni21B4,  RIGHTWARDS ARROW WITH CORNER DOWNWARDS
        '8636' => '8884'              , # G+8884  U+21BC  harpoonleftbarbup,  LEFTWARDS HARPOON WITH BARB UPWARDS
        '8640' => '8885'              , # G+8885  U+21C0  harpoonrightbarbup,  RIGHTWARDS HARPOON WITH BARB UPWARDS
        '8644' => '8896'              , # G+8896  U+21C4  arrowrightoverleft,  RIGHTWARDS ARROW OVER LEFTWARDS ARROW
        '8645' => '8897'              , # G+8897  U+21C5  arrowupleftofdown,  UPWARDS ARROW LEFTWARDS OF DOWNWARDS ARROW
        '8653' => '8816'              , # G+8816  U+21CD  arrowleftdblstroke,  LEFTWARDS DOUBLE ARROW WITH STROKE
        '8655' => '8815'              , # G+8815  U+21CF  arrowrightdblstroke,  RIGHTWARDS DOUBLE ARROW WITH STROKE
        '8656' => '8814'              , # G+8814  U+21D0  arrowdblleft,  LEFTWARDS DOUBLE ARROW
        '8657' => '8854'              , # G+8854  U+21D1  arrowdblup,  UPWARDS DOUBLE ARROW
        '8658' => '195'               , # G+195  U+21D2  arrowdblright,  RIGHTWARDS DOUBLE ARROW
        '8659' => '8855'              , # G+8855  U+21D3  arrowdbldown,  DOWNWARDS DOUBLE ARROW
        '8660' => '196'               , # G+196  U+21D4  arrowdblboth,  LEFT RIGHT DOUBLE ARROW
        '8672' => '9190'              , # G+9190  U+21E0  arrowdashleft,  LEFTWARDS DASHED ARROW
        '8673' => '9192'              , # G+9192  U+21E1  arrowdashup,  UPWARDS DASHED ARROW
        '8674' => '9191'              , # G+9191  U+21E2  arrowdashright,  RIGHTWARDS DASHED ARROW
        '8675' => '9193'              , # G+9193  U+21E3  arrowdashdown,  DOWNWARDS DASHED ARROW
        '8678' => '9198'              , # G+9198  U+21E6  arrowleftwhite,  LEFTWARDS WHITE ARROW
        '8679' => '9200'              , # G+9200  U+21E7  arrowupwhite,  UPWARDS WHITE ARROW
        '8680' => '9199'              , # G+9199  U+21E8  arrowrightwhite,  RIGHTWARDS WHITE ARROW
        '8681' => '9201'              , # G+9201  U+21E9  arrowdownwhite,  DOWNWARDS WHITE ARROW
        '8704' => '197'               , # G+197  U+2200  universal,  FOR ALL
        '8706' => '151'               , # G+151  U+2202  partialdiff,  PARTIAL DIFFERENTIAL
        '8707' => '198'               , # G+198  U+2203  existential,  THERE EXISTS
        '8710' => '8715'              , # G+8715  U+2206  Delta,  INCREMENT
        '8711' => '152'               , # G+152  U+2207  gradient,  NABLA
        '8712' => '184'               , # G+184  U+2208  element,  ELEMENT OF
        '8713' => '8749'              , # G+8749  U+2209  notelement,  NOT AN ELEMENT OF
        '8715' => '185'               , # G+185  U+220B  suchthat,  CONTAINS AS MEMBER
        '8716' => '8750'              , # G+8750  U+220C  notcontains,  DOES NOT CONTAIN AS MEMBER
        '8719' => '213'               , # G+213  U+220F  product,  N-ARY PRODUCT
        '8721' => '212'               , # G+212  U+2211  summation,  N-ARY SUMMATION
        '8723' => '8726'              , # G+8726  U+2213  minusplus,  MINUS-OR-PLUS SIGN
        '8730' => '178'               , # G+178  U+221A  radical,  SQUARE ROOT
        '8733' => '180'               , # G+180  U+221D  proportional,  PROPORTIONAL TO
        '8734' => '136'               , # G+136  U+221E  infinity,  INFINITY
        '8735' => '8717'              , # G+8717  U+221F  orthogonal,  RIGHT ANGLE
        '8736' => '148'               , # G+148  U+2220  angle,  ANGLE
        '8738' => '8738'              , # G+8738  U+2222  angsph,  SPHERICAL ANGLE
        '8741' => '111'               , # G+111  U+2225  parallel,  PARALLEL TO
        '8742' => '8720'              , # G+8720  U+2226  notparallel,  NOT PARALLEL TO
        '8743' => '192'               , # G+192  U+2227  logicaland,  LOGICAL AND
        '8744' => '193'               , # G+193  U+2228  logicalor,  LOGICAL OR
        '8745' => '191'               , # G+191  U+2229  intersection,  INTERSECTION
        '8746' => '190'               , # G+190  U+222A  union,  UNION
        '8747' => '182'               , # G+182  U+222B  integral,  INTEGRAL
        '8748' => '183'               , # G+183  U+222C  dblintegral,  DOUBLE INTEGRAL
        '8750' => '211'               , # G+211  U+222E  contourintegral,  CONTOUR INTEGRAL
        '8756' => '137'               , # G+137  U+2234  therefore,  THEREFORE
        '8757' => '181'               , # G+181  U+2235  because,  BECAUSE
        '8758' => '210'               , # G+210  U+2236  ratio,  RATIO
        '8759' => '8321'              , # G+8321  U+2237  proportion,  PROPORTION
        '8764' => '113'               , # G+113  U+223C  similar,  TILDE OPERATOR
        '8765' => '179'               , # G+179  U+223D  reversedtilde,  REVERSED TILDE
        '8771' => '8500'              , # G+8500  U+2243  asymptoticallyequal,  ASYMPTOTICALLY EQUAL TO
        '8773' => '8499'              , # G+8499  U+2245  congruent,  APPROXIMATELY EQUAL TO
        '8776' => '8501'              , # G+8501  U+2248  approxequal,  ALMOST EQUAL TO
        '8784' => '8739'              , # G+8739  U+2250  approaches,  APPROACHES THE LIMIT
        '8785' => '8723'              , # G+8723  U+2251  geometricallyequal,  GEOMETRICALLY EQUAL TO
        '8786' => '154'               , # G+154  U+2252  approxequalorimage,  APPROXIMATELY EQUAL TO OR THE IMAGE OF
        '8787' => '8722'              , # G+8722  U+2253  imageorapproximatelyequal,  IMAGE OF OR APPROXIMATELY EQUAL TO
        '8794' => '8753'              , # G+8753  U+225A  uni225A,  EQUIANGULAR TO
        '8800' => '133'               , # G+133  U+2260  notequal,  NOT EQUAL TO
        '8801' => '153'               , # G+153  U+2261  equivalence,  IDENTICAL TO
        '8802' => '8734'              , # G+8734  U+2262  notidentical,  NOT IDENTICAL TO
        '8804' => '134'               , # G+134  U+2264  lessequal,  LESS-THAN OR EQUAL TO
        '8805' => '135'               , # G+135  U+2265  greaterequal,  GREATER-THAN OR EQUAL TO
        '8806' => '8724'              , # G+8724  U+2266  lessoverequal,  LESS-THAN OVER EQUAL TO
        '8807' => '8725'              , # G+8725  U+2267  greateroverequal,  GREATER-THAN OVER EQUAL TO
        '8810' => '176'               , # G+176  U+226A  muchless,  MUCH LESS-THAN
        '8811' => '177'               , # G+177  U+226B  muchgreater,  MUCH GREATER-THAN
        '8814' => '8745'              , # G+8745  U+226E  notless,  NOT LESS-THAN
        '8815' => '8746'              , # G+8746  U+226F  notgreater,  NOT GREATER-THAN
        '8816' => '8481'              , # G+8481  U+2270  notlessnorequal,  NEITHER LESS-THAN NOR EQUAL TO
        '8817' => '8482'              , # G+8482  U+2271  notgreaternorequal,  NEITHER GREATER-THAN NOR EQUAL TO
        '8818' => '8483'              , # G+8483  U+2272  lessorequivalent,  LESS-THAN OR EQUIVALENT TO
        '8819' => '8484'              , # G+8484  U+2273  greaterorequivalent,  GREATER-THAN OR EQUIVALENT TO
        '8822' => '8489'              , # G+8489  U+2276  lessorgreater,  LESS-THAN OR GREATER-THAN
        '8823' => '8490'              , # G+8490  U+2277  greaterorless,  GREATER-THAN OR LESS-THAN
        '8825' => '8491'              , # G+8491  U+2279  notgreaternorless,  NEITHER GREATER-THAN NOR LESS-THAN
        '8826' => '8475'              , # G+8475  U+227A  precedes,  PRECEDES
        '8827' => '8476'              , # G+8476  U+227B  succeeds,  SUCCEEDS
        '8832' => '8479'              , # G+8479  U+2280  notprecedes,  DOES NOT PRECEDE
        '8833' => '8480'              , # G+8480  U+2281  notsucceeds,  DOES NOT SUCCEED
        '8834' => '188'               , # G+188  U+2282  propersubset,  SUBSET OF
        '8835' => '189'               , # G+189  U+2283  propersuperset,  SUPERSET OF
        '8836' => '8748'              , # G+8748  U+2284  notsubset,  NOT A SUBSET OF
        '8837' => '8747'              , # G+8747  U+2285  notsuperset,  NOT A SUPERSET OF
        '8838' => '186'               , # G+186  U+2286  reflexsubset,  SUBSET OF OR EQUAL TO
        '8839' => '187'               , # G+187  U+2287  reflexsuperset,  SUPERSET OF OR EQUAL TO
        '8842' => '8486'              , # G+8486  U+228A  subsetnotequal,  SUBSET OF WITH NOT EQUAL TO
        '8843' => '8488'              , # G+8488  U+228B  supersetnotequal,  SUPERSET OF WITH NOT EQUAL TO
        '8853' => '8727'              , # G+8727  U+2295  circleplus,  CIRCLED PLUS
        '8854' => '8728'              , # G+8728  U+2296  minuscircle,  CIRCLED MINUS
        '8855' => '8729'              , # G+8729  U+2297  circlemultiply,  CIRCLED TIMES
        '8857' => '227'               , # G+227  U+2299  circleot,  CIRCLED DOT OPERATOR
        '8867' => '8742'              , # G+8742  U+22A3  tackleft,  LEFT TACK
        '8868' => '8503'              , # G+8503  U+22A4  tackdown,  DOWN TACK
        '8869' => '149'               , # G+149  U+22A5  perpendicular,  UP TACK
        '8891' => '8751'              , # G+8751  U+22BB  veebar,  XOR
        '8892' => '8752'              , # G+8752  U+22BC  barwed,  NAND
        '8910' => '8477'              , # G+8477  U+22CE  curlyor,  CURLY LOGICAL OR
        '8911' => '8478'              , # G+8478  U+22CF  curlyand,  CURLY LOGICAL AND
        '8922' => '8492'              , # G+8492  U+22DA  lessequalorgreater,  LESS-THAN EQUAL TO OR GREATER-THAN
        '8923' => '8493'              , # G+8493  U+22DB  greaterequalorless,  GREATER-THAN EQUAL TO OR LESS-THAN
        '8942' => '8320'              , # G+8320  U+22EE  ellipsisvertical,  VERTICAL ELLIPSIS
        '8943' => '106'               , # G+106  U+22EF  uni22EF,  MIDLINE HORIZONTAL ELLIPSIS
        '8966' => '8754'              , # G+8754  U+2306  Barwed,  PERSPECTIVE
        '8978' => '150'               , # G+150  U+2312  arc,  ARC
        '8980' => '8731'              , # G+8731  U+2314  uni2314,  SECTOR
        '9312' => '733'               , # G+733  U+2460  onecircle,  CIRCLED DIGIT ONE
        '9313' => '734'               , # G+734  U+2461  twocircle,  CIRCLED DIGIT TWO
        '9314' => '735'               , # G+735  U+2462  threecircle,  CIRCLED DIGIT THREE
        '9315' => '736'               , # G+736  U+2463  fourcircle,  CIRCLED DIGIT FOUR
        '9316' => '737'               , # G+737  U+2464  fivecircle,  CIRCLED DIGIT FIVE
        '9317' => '738'               , # G+738  U+2465  sixcircle,  CIRCLED DIGIT SIX
        '9318' => '739'               , # G+739  U+2466  sevencircle,  CIRCLED DIGIT SEVEN
        '9319' => '740'               , # G+740  U+2467  eightcircle,  CIRCLED DIGIT EIGHT
        '9320' => '741'               , # G+741  U+2468  ninecircle,  CIRCLED DIGIT NINE
        '9321' => '742'               , # G+742  U+2469  tencircle,  CIRCLED NUMBER TEN
        '9322' => '743'               , # G+743  U+246A  elevencircle,  CIRCLED NUMBER ELEVEN
        '9323' => '744'               , # G+744  U+246B  twelvecircle,  CIRCLED NUMBER TWELVE
        '9324' => '745'               , # G+745  U+246C  thirteencircle,  CIRCLED NUMBER THIRTEEN
        '9325' => '746'               , # G+746  U+246D  fourteencircle,  CIRCLED NUMBER FOURTEEN
        '9326' => '747'               , # G+747  U+246E  fifteencircle,  CIRCLED NUMBER FIFTEEN
        '9327' => '8791'              , # G+8791  U+246F  sixteencircle,  CIRCLED NUMBER SIXTEEN
        '9328' => '8792'              , # G+8792  U+2470  seventeencircle,  CIRCLED NUMBER SEVENTEEN
        '9329' => '8793'              , # G+8793  U+2471  eighteencircle,  CIRCLED NUMBER EIGHTEEN
        '9330' => '8794'              , # G+8794  U+2472  nineteencircle,  CIRCLED NUMBER NINETEEN
        '9331' => '8795'              , # G+8795  U+2473  twentycircle,  CIRCLED NUMBER TWENTY
        '9332' => '827'               , # G+827  U+2474  oneparen,  PARENTHESIZED DIGIT ONE
        '9333' => '828'               , # G+828  U+2475  twoparen,  PARENTHESIZED DIGIT TWO
        '9334' => '829'               , # G+829  U+2476  threeparen,  PARENTHESIZED DIGIT THREE
        '9335' => '830'               , # G+830  U+2477  fourparen,  PARENTHESIZED DIGIT FOUR
        '9336' => '831'               , # G+831  U+2478  fiveparen,  PARENTHESIZED DIGIT FIVE
        '9337' => '832'               , # G+832  U+2479  sixparen,  PARENTHESIZED DIGIT SIX
        '9338' => '833'               , # G+833  U+247A  sevenparen,  PARENTHESIZED DIGIT SEVEN
        '9339' => '834'               , # G+834  U+247B  eightparen,  PARENTHESIZED DIGIT EIGHT
        '9340' => '835'               , # G+835  U+247C  nineparen,  PARENTHESIZED DIGIT NINE
        '9341' => '836'               , # G+836  U+247D  tenparen,  PARENTHESIZED NUMBER TEN
        '9342' => '837'               , # G+837  U+247E  elevenparen,  PARENTHESIZED NUMBER ELEVEN
        '9343' => '838'               , # G+838  U+247F  twelveparen,  PARENTHESIZED NUMBER TWELVE
        '9344' => '839'               , # G+839  U+2480  thirteenparen,  PARENTHESIZED NUMBER THIRTEEN
        '9345' => '840'               , # G+840  U+2481  fourteenparen,  PARENTHESIZED NUMBER FOURTEEN
        '9346' => '841'               , # G+841  U+2482  fifteenparen,  PARENTHESIZED NUMBER FIFTEEN
        '9347' => '9042'              , # G+9042  U+2483  sixteenparen,  PARENTHESIZED NUMBER SIXTEEN
        '9348' => '9043'              , # G+9043  U+2484  seventeenparen,  PARENTHESIZED NUMBER SEVENTEEN
        '9349' => '9044'              , # G+9044  U+2485  eighteenparen,  PARENTHESIZED NUMBER EIGHTEEN
        '9350' => '9045'              , # G+9045  U+2486  nineteenparen,  PARENTHESIZED NUMBER NINETEEN
        '9351' => '9046'              , # G+9046  U+2487  twentyparen,  PARENTHESIZED NUMBER TWENTY
        '9372' => '801'               , # G+801  U+249C  aparen,  PARENTHESIZED LATIN SMALL LETTER A
        '9373' => '802'               , # G+802  U+249D  bparen,  PARENTHESIZED LATIN SMALL LETTER B
        '9374' => '803'               , # G+803  U+249E  cparen,  PARENTHESIZED LATIN SMALL LETTER C
        '9375' => '804'               , # G+804  U+249F  dparen,  PARENTHESIZED LATIN SMALL LETTER D
        '9376' => '805'               , # G+805  U+24A0  eparen,  PARENTHESIZED LATIN SMALL LETTER E
        '9377' => '806'               , # G+806  U+24A1  fparen,  PARENTHESIZED LATIN SMALL LETTER F
        '9378' => '807'               , # G+807  U+24A2  gparen,  PARENTHESIZED LATIN SMALL LETTER G
        '9379' => '808'               , # G+808  U+24A3  hparen,  PARENTHESIZED LATIN SMALL LETTER H
        '9380' => '809'               , # G+809  U+24A4  iparen,  PARENTHESIZED LATIN SMALL LETTER I
        '9381' => '810'               , # G+810  U+24A5  jparen,  PARENTHESIZED LATIN SMALL LETTER J
        '9382' => '811'               , # G+811  U+24A6  kparen,  PARENTHESIZED LATIN SMALL LETTER K
        '9383' => '812'               , # G+812  U+24A7  lparen,  PARENTHESIZED LATIN SMALL LETTER L
        '9384' => '813'               , # G+813  U+24A8  mparen,  PARENTHESIZED LATIN SMALL LETTER M
        '9385' => '814'               , # G+814  U+24A9  nparen,  PARENTHESIZED LATIN SMALL LETTER N
        '9386' => '815'               , # G+815  U+24AA  oparen,  PARENTHESIZED LATIN SMALL LETTER O
        '9387' => '816'               , # G+816  U+24AB  pparen,  PARENTHESIZED LATIN SMALL LETTER P
        '9388' => '817'               , # G+817  U+24AC  qparen,  PARENTHESIZED LATIN SMALL LETTER Q
        '9389' => '818'               , # G+818  U+24AD  rparen,  PARENTHESIZED LATIN SMALL LETTER R
        '9390' => '819'               , # G+819  U+24AE  sparen,  PARENTHESIZED LATIN SMALL LETTER S
        '9391' => '820'               , # G+820  U+24AF  tparen,  PARENTHESIZED LATIN SMALL LETTER T
        '9392' => '821'               , # G+821  U+24B0  uparen,  PARENTHESIZED LATIN SMALL LETTER U
        '9393' => '822'               , # G+822  U+24B1  vparen,  PARENTHESIZED LATIN SMALL LETTER V
        '9394' => '823'               , # G+823  U+24B2  wparen,  PARENTHESIZED LATIN SMALL LETTER W
        '9395' => '824'               , # G+824  U+24B3  xparen,  PARENTHESIZED LATIN SMALL LETTER X
        '9396' => '825'               , # G+825  U+24B4  yparen,  PARENTHESIZED LATIN SMALL LETTER Y
        '9397' => '826'               , # G+826  U+24B5  zparen,  PARENTHESIZED LATIN SMALL LETTER Z
        '9398' => '8388'              , # G+8388  U+24B6  Acircle,  CIRCLED LATIN CAPITAL LETTER A
        '9399' => '8389'              , # G+8389  U+24B7  Bcircle,  CIRCLED LATIN CAPITAL LETTER B
        '9400' => '8390'              , # G+8390  U+24B8  Ccircle,  CIRCLED LATIN CAPITAL LETTER C
        '9401' => '8391'              , # G+8391  U+24B9  Dcircle,  CIRCLED LATIN CAPITAL LETTER D
        '9402' => '8392'              , # G+8392  U+24BA  Ecircle,  CIRCLED LATIN CAPITAL LETTER E
        '9403' => '8393'              , # G+8393  U+24BB  Fcircle,  CIRCLED LATIN CAPITAL LETTER F
        '9404' => '8394'              , # G+8394  U+24BC  Gcircle,  CIRCLED LATIN CAPITAL LETTER G
        '9405' => '8395'              , # G+8395  U+24BD  Hcircle,  CIRCLED LATIN CAPITAL LETTER H
        '9406' => '8396'              , # G+8396  U+24BE  Icircle,  CIRCLED LATIN CAPITAL LETTER I
        '9407' => '8397'              , # G+8397  U+24BF  Jcircle,  CIRCLED LATIN CAPITAL LETTER J
        '9408' => '8398'              , # G+8398  U+24C0  Kcircle,  CIRCLED LATIN CAPITAL LETTER K
        '9409' => '8399'              , # G+8399  U+24C1  Lcircle,  CIRCLED LATIN CAPITAL LETTER L
        '9410' => '8400'              , # G+8400  U+24C2  Mcircle,  CIRCLED LATIN CAPITAL LETTER M
        '9411' => '8401'              , # G+8401  U+24C3  Ncircle,  CIRCLED LATIN CAPITAL LETTER N
        '9412' => '8402'              , # G+8402  U+24C4  Ocircle,  CIRCLED LATIN CAPITAL LETTER O
        '9413' => '8403'              , # G+8403  U+24C5  Pcircle,  CIRCLED LATIN CAPITAL LETTER P
        '9414' => '8404'              , # G+8404  U+24C6  Qcircle,  CIRCLED LATIN CAPITAL LETTER Q
        '9415' => '8405'              , # G+8405  U+24C7  Rcircle,  CIRCLED LATIN CAPITAL LETTER R
        '9416' => '8406'              , # G+8406  U+24C8  Scircle,  CIRCLED LATIN CAPITAL LETTER S
        '9417' => '8407'              , # G+8407  U+24C9  Tcircle,  CIRCLED LATIN CAPITAL LETTER T
        '9418' => '8408'              , # G+8408  U+24CA  Ucircle,  CIRCLED LATIN CAPITAL LETTER U
        '9419' => '8409'              , # G+8409  U+24CB  Vcircle,  CIRCLED LATIN CAPITAL LETTER V
        '9420' => '8410'              , # G+8410  U+24CC  Wcircle,  CIRCLED LATIN CAPITAL LETTER W
        '9421' => '8411'              , # G+8411  U+24CD  Xcircle,  CIRCLED LATIN CAPITAL LETTER X
        '9422' => '8412'              , # G+8412  U+24CE  Ycircle,  CIRCLED LATIN CAPITAL LETTER Y
        '9423' => '8413'              , # G+8413  U+24CF  Zcircle,  CIRCLED LATIN CAPITAL LETTER Z
        '9424' => '707'               , # G+707  U+24D0  acircle,  CIRCLED LATIN SMALL LETTER A
        '9425' => '708'               , # G+708  U+24D1  bcircle,  CIRCLED LATIN SMALL LETTER B
        '9426' => '709'               , # G+709  U+24D2  ccircle,  CIRCLED LATIN SMALL LETTER C
        '9427' => '710'               , # G+710  U+24D3  dcircle,  CIRCLED LATIN SMALL LETTER D
        '9428' => '711'               , # G+711  U+24D4  ecircle,  CIRCLED LATIN SMALL LETTER E
        '9429' => '712'               , # G+712  U+24D5  fcircle,  CIRCLED LATIN SMALL LETTER F
        '9430' => '713'               , # G+713  U+24D6  gcircle,  CIRCLED LATIN SMALL LETTER G
        '9431' => '714'               , # G+714  U+24D7  hcircle,  CIRCLED LATIN SMALL LETTER H
        '9432' => '715'               , # G+715  U+24D8  icircle,  CIRCLED LATIN SMALL LETTER I
        '9433' => '716'               , # G+716  U+24D9  jcircle,  CIRCLED LATIN SMALL LETTER J
        '9434' => '717'               , # G+717  U+24DA  kcircle,  CIRCLED LATIN SMALL LETTER K
        '9435' => '718'               , # G+718  U+24DB  lcircle,  CIRCLED LATIN SMALL LETTER L
        '9436' => '719'               , # G+719  U+24DC  mcircle,  CIRCLED LATIN SMALL LETTER M
        '9437' => '720'               , # G+720  U+24DD  ncircle,  CIRCLED LATIN SMALL LETTER N
        '9438' => '721'               , # G+721  U+24DE  ocircle,  CIRCLED LATIN SMALL LETTER O
        '9439' => '722'               , # G+722  U+24DF  pcircle,  CIRCLED LATIN SMALL LETTER P
        '9440' => '723'               , # G+723  U+24E0  qcircle,  CIRCLED LATIN SMALL LETTER Q
        '9441' => '724'               , # G+724  U+24E1  rcircle,  CIRCLED LATIN SMALL LETTER R
        '9442' => '725'               , # G+725  U+24E2  scircle,  CIRCLED LATIN SMALL LETTER S
        '9443' => '726'               , # G+726  U+24E3  tcircle,  CIRCLED LATIN SMALL LETTER T
        '9444' => '727'               , # G+727  U+24E4  ucircle,  CIRCLED LATIN SMALL LETTER U
        '9445' => '728'               , # G+728  U+24E5  vcircle,  CIRCLED LATIN SMALL LETTER V
        '9446' => '729'               , # G+729  U+24E6  wcircle,  CIRCLED LATIN SMALL LETTER W
        '9447' => '730'               , # G+730  U+24E7  xcircle,  CIRCLED LATIN SMALL LETTER X
        '9448' => '731'               , # G+731  U+24E8  ycircle,  CIRCLED LATIN SMALL LETTER Y
        '9449' => '732'               , # G+732  U+24E9  zcircle,  CIRCLED LATIN SMALL LETTER Z
        '9472' => '519'               , # G+519  U+2500  SF100000,  BOX DRAWINGS LIGHT HORIZONTAL
        '9473' => '530'               , # G+530  U+2501  uni2501,  BOX DRAWINGS HEAVY HORIZONTAL
        '9474' => '520'               , # G+520  U+2502  SF110000,  BOX DRAWINGS LIGHT VERTICAL
        '9475' => '531'               , # G+531  U+2503  uni2503,  BOX DRAWINGS HEAVY VERTICAL
        '9484' => '521'               , # G+521  U+250C  SF010000,  BOX DRAWINGS LIGHT DOWN AND RIGHT
        '9485' => '558'               , # G+558  U+250D  uni250D,  BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY
        '9486' => '557'               , # G+557  U+250E  uni250E,  BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT
        '9487' => '532'               , # G+532  U+250F  uni250F,  BOX DRAWINGS HEAVY DOWN AND RIGHT
        '9488' => '522'               , # G+522  U+2510  SF030000,  BOX DRAWINGS LIGHT DOWN AND LEFT
        '9489' => '552'               , # G+552  U+2511  uni2511,  BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY
        '9490' => '551'               , # G+551  U+2512  uni2512,  BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT
        '9491' => '533'               , # G+533  U+2513  uni2513,  BOX DRAWINGS HEAVY DOWN AND LEFT
        '9492' => '524'               , # G+524  U+2514  SF020000,  BOX DRAWINGS LIGHT UP AND RIGHT
        '9493' => '556'               , # G+556  U+2515  uni2515,  BOX DRAWINGS UP LIGHT AND RIGHT HEAVY
        '9494' => '555'               , # G+555  U+2516  uni2516,  BOX DRAWINGS UP HEAVY AND RIGHT LIGHT
        '9495' => '535'               , # G+535  U+2517  uni2517,  BOX DRAWINGS HEAVY UP AND RIGHT
        '9496' => '523'               , # G+523  U+2518  SF040000,  BOX DRAWINGS LIGHT UP AND LEFT
        '9497' => '554'               , # G+554  U+2519  uni2519,  BOX DRAWINGS UP LIGHT AND LEFT HEAVY
        '9498' => '553'               , # G+553  U+251A  uni251A,  BOX DRAWINGS UP HEAVY AND LEFT LIGHT
        '9499' => '534'               , # G+534  U+251B  uni251B,  BOX DRAWINGS HEAVY UP AND LEFT
        '9500' => '525'               , # G+525  U+251C  SF080000,  BOX DRAWINGS LIGHT VERTICAL AND RIGHT
        '9501' => '546'               , # G+546  U+251D  uni251D,  BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY
        '9502' => '559'               , # G+559  U+251E  uni251E,  BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT
        '9503' => '560'               , # G+560  U+251F  uni251F,  BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT
        '9504' => '541'               , # G+541  U+2520  uni2520,  BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT
        '9505' => '561'               , # G+561  U+2521  uni2521,  BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY
        '9506' => '562'               , # G+562  U+2522  uni2522,  BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY
        '9507' => '536'               , # G+536  U+2523  uni2523,  BOX DRAWINGS HEAVY VERTICAL AND RIGHT
        '9508' => '527'               , # G+527  U+2524  SF090000,  BOX DRAWINGS LIGHT VERTICAL AND LEFT
        '9509' => '548'               , # G+548  U+2525  uni2525,  BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY
        '9510' => '563'               , # G+563  U+2526  uni2526,  BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT
        '9511' => '564'               , # G+564  U+2527  uni2527,  BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT
        '9512' => '543'               , # G+543  U+2528  uni2528,  BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT
        '9513' => '565'               , # G+565  U+2529  uni2529,  BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY
        '9514' => '566'               , # G+566  U+252A  uni252A,  BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY
        '9515' => '538'               , # G+538  U+252B  uni252B,  BOX DRAWINGS HEAVY VERTICAL AND LEFT
        '9516' => '526'               , # G+526  U+252C  SF060000,  BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
        '9517' => '567'               , # G+567  U+252D  uni252D,  BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT
        '9518' => '568'               , # G+568  U+252E  uni252E,  BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT
        '9519' => '542'               , # G+542  U+252F  uni252F,  BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY
        '9520' => '547'               , # G+547  U+2530  uni2530,  BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT
        '9521' => '569'               , # G+569  U+2531  uni2531,  BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY
        '9522' => '570'               , # G+570  U+2532  uni2532,  BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY
        '9523' => '537'               , # G+537  U+2533  uni2533,  BOX DRAWINGS HEAVY DOWN AND HORIZONTAL
        '9524' => '528'               , # G+528  U+2534  SF070000,  BOX DRAWINGS LIGHT UP AND HORIZONTAL
        '9525' => '571'               , # G+571  U+2535  uni2535,  BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT
        '9526' => '572'               , # G+572  U+2536  uni2536,  BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT
        '9527' => '544'               , # G+544  U+2537  uni2537,  BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY
        '9528' => '549'               , # G+549  U+2538  uni2538,  BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT
        '9529' => '573'               , # G+573  U+2539  uni2539,  BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY
        '9530' => '574'               , # G+574  U+253A  uni253A,  BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY
        '9531' => '539'               , # G+539  U+253B  uni253B,  BOX DRAWINGS HEAVY UP AND HORIZONTAL
        '9532' => '529'               , # G+529  U+253C  SF050000,  BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
        '9533' => '575'               , # G+575  U+253D  uni253D,  BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT
        '9534' => '576'               , # G+576  U+253E  uni253E,  BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT
        '9535' => '545'               , # G+545  U+253F  uni253F,  BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY
        '9536' => '577'               , # G+577  U+2540  uni2540,  BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT
        '9537' => '578'               , # G+578  U+2541  uni2541,  BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT
        '9538' => '550'               , # G+550  U+2542  uni2542,  BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT
        '9539' => '579'               , # G+579  U+2543  uni2543,  BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT
        '9540' => '580'               , # G+580  U+2544  uni2544,  BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT
        '9541' => '581'               , # G+581  U+2545  uni2545,  BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT
        '9542' => '582'               , # G+582  U+2546  uni2546,  BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT
        '9543' => '583'               , # G+583  U+2547  uni2547,  BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY
        '9544' => '584'               , # G+584  U+2548  uni2548,  BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY
        '9545' => '585'               , # G+585  U+2549  uni2549,  BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY
        '9546' => '586'               , # G+586  U+254A  uni254A,  BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY
        '9547' => '540'               , # G+540  U+254B  uni254B,  BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL
        '9618' => '232'               , # G+232  U+2592  shade,  MEDIUM SHADE
        '9632' => '165'               , # G+165  U+25A0  filledbox,  BLACK SQUARE
        '9633' => '164'               , # G+164  U+25A1  H22073,  WHITE SQUARE
        '9635' => '229'               , # G+229  U+25A3  squarewhitewithsmallblack,  WHITE SQUARE CONTAINING BLACK SMALL SQUARE
        '9636' => '233'               , # G+233  U+25A4  squarehorizontalfill,  SQUARE WITH HORIZONTAL FILL
        '9637' => '234'               , # G+234  U+25A5  squareverticalfill,  SQUARE WITH VERTICAL FILL
        '9638' => '237'               , # G+237  U+25A6  squareorthogonalcrosshatchfill,  SQUARE WITH ORTHOGONAL CROSSHATCH FILL
        '9639' => '236'               , # G+236  U+25A7  squareupperlefttolowerrightfill,  SQUARE WITH UPPER LEFT TO LOWER RIGHT FILL
        '9640' => '235'               , # G+235  U+25A8  squareupperrighttolowerleftfill,  SQUARE WITH UPPER RIGHT TO LOWER LEFT FILL
        '9641' => '238'               , # G+238  U+25A9  squarediagonalcrosshatchfill,  SQUARE WITH DIAGONAL CROSSHATCH FILL
        '9649' => '8736'              , # G+8736  U+25B1  uni25B1,  WHITE PARALLELOGRAM
        '9650' => '167'               , # G+167  U+25B2  triagup,  BLACK UP-POINTING TRIANGLE
        '9651' => '166'               , # G+166  U+25B3  whiteuppointingtriangle,  WHITE UP-POINTING TRIANGLE
        '9653' => '8780'              , # G+8780  U+25B5  whiteuppointingsmalltriangle,  WHITE UP-POINTING SMALL TRIANGLE
        '9654' => '220'               , # G+220  U+25B6  blackrightpointingtriangle,  BLACK RIGHT-POINTING TRIANGLE
        '9655' => '219'               , # G+219  U+25B7  whiterightpointingtriangle,  WHITE RIGHT-POINTING TRIANGLE
        '9657' => '8781'              , # G+8781  U+25B9  whiterightpointingsmalltriangle,  WHITE RIGHT-POINTING SMALL TRIANGLE
        '9660' => '169'               , # G+169  U+25BC  triagdn,  BLACK DOWN-POINTING TRIANGLE
        '9661' => '168'               , # G+168  U+25BD  whitedownpointingtriangle,  WHITE DOWN-POINTING TRIANGLE
        '9663' => '8779'              , # G+8779  U+25BF  whitedownpointingsmalltriangle,  WHITE DOWN-POINTING SMALL TRIANGLE
        '9664' => '218'               , # G+218  U+25C0  blackleftpointingtriangle,  BLACK LEFT-POINTING TRIANGLE
        '9665' => '217'               , # G+217  U+25C1  whiteleftpointingtriangle,  WHITE LEFT-POINTING TRIANGLE
        '9667' => '8782'              , # G+8782  U+25C3  whiteleftpointingsmalltriangle,  WHITE LEFT-POINTING SMALL TRIANGLE
        '9670' => '163'               , # G+163  U+25C6  blackdiamond,  BLACK DIAMOND
        '9671' => '162'               , # G+162  U+25C7  whitediamond,  WHITE DIAMOND
        '9672' => '228'               , # G+228  U+25C8  whitediamondcontainingblacksmalldiamond,  WHITE DIAMOND CONTAINING BLACK SMALL DIAMOND
        '9673' => '227'               , # G+227  U+25C9  fisheye,  FISHEYE
        '9674' => '8787'              , # G+8787  U+25CA  lozenge,  LOZENGE
        '9675' => '159'               , # G+159  U+25CB  circle,  WHITE CIRCLE
        '9676' => '8639'              , # G+8639  U+25CC  dottedcircle,  DOTTED CIRCLE
        '9678' => '161'               , # G+161  U+25CE  bullseye,  BULLSEYE
        '9679' => '160'               , # G+160  U+25CF  H18533,  BLACK CIRCLE
        '9680' => '230'               , # G+230  U+25D0  circlewithlefthalfblack,  CIRCLE WITH LEFT HALF BLACK
        '9681' => '231'               , # G+231  U+25D1  circlewithrighthalfblack,  CIRCLE WITH RIGHT HALF BLACK
        '9702' => '8775'              , # G+8775  U+25E6  openbullet,  WHITE BULLET
        '9711' => '8633'              , # G+8633  U+25EF  largecircle,  LARGE CIRCLE
        '9733' => '158'               , # G+158  U+2605  blackstar,  BLACK STAR
        '9734' => '157'               , # G+157  U+2606  whitestar,  WHITE STAR
        '9742' => '241'               , # G+241  U+260E  telephoneblack,  BLACK TELEPHONE
        '9743' => '240'               , # G+240  U+260F  whitetelephone,  WHITE TELEPHONE
        '9756' => '242'               , # G+242  U+261C  pointingindexleftwhite,  WHITE LEFT POINTING INDEX
        '9757' => '9222'              , # G+9222  U+261D  pointingindexupwhite,  WHITE UP POINTING INDEX
        '9758' => '243'               , # G+243  U+261E  pointingindexrightwhite,  WHITE RIGHT POINTING INDEX
        '9759' => '9223'              , # G+9223  U+261F  pointingindexdownwhite,  WHITE DOWN POINTING INDEX
        '9775' => '8664'              , # G+8664  U+262F  yinyang,  YIN YANG
        '9792' => '147'               , # G+147  U+2640  female,  FEMALE SIGN
        '9794' => '146'               , # G+146  U+2642  male,  MALE SIGN
        '9824' => '222'               , # G+222  U+2660  spade,  BLACK SPADE SUIT
        '9825' => '223'               , # G+223  U+2661  heartsuitwhite,  WHITE HEART SUIT
        '9827' => '226'               , # G+226  U+2663  club,  BLACK CLUB SUIT
        '9828' => '221'               , # G+221  U+2664  spadesuitwhite,  WHITE SPADE SUIT
        '9829' => '224'               , # G+224  U+2665  heart,  BLACK HEART SUIT
        '9831' => '225'               , # G+225  U+2667  clubsuitwhite,  WHITE CLUB SUIT
        '9832' => '239'               , # G+239  U+2668  hotsprings,  HOT SPRINGS
        '9833' => '253'               , # G+253  U+2669  quarternote,  QUARTER NOTE
        '9834' => '254'               , # G+254  U+266A  musicalnote,  EIGHTH NOTE
        '9836' => '255'               , # G+255  U+266C  beamedsixteenthnotes,  BEAMED SIXTEENTH NOTES
        '9837' => '252'               , # G+252  U+266D  musicflatsign,  MUSIC FLAT SIGN
        '9839' => '8594'              , # G+8594  U+266F  musicsharpsign,  MUSIC SHARP SIGN
        '10006' => '8631'             , # G+8631  U+2716  a22,  HEAVY MULTIPLICATION X
        '10010' => '8630'             , # G+8630  U+271A  a26,  HEAVY GREEK CROSS
        '10045' => '8604'             , # G+8604  U+273D  a56,  HEAVY TEARDROP-SPOKED ASTERISK
        '10070' => '8637'             , # G+8637  U+2756  a79,  BLACK DIAMOND MINUS WHITE X
        '10102' => '8673'             , # G+8673  U+2776  a130,  DINGBAT NEGATIVE CIRCLED DIGIT ONE
        '10103' => '8674'             , # G+8674  U+2777  a131,  DINGBAT NEGATIVE CIRCLED DIGIT TWO
        '10104' => '8675'             , # G+8675  U+2778  a132,  DINGBAT NEGATIVE CIRCLED DIGIT THREE
        '10105' => '8676'             , # G+8676  U+2779  a133,  DINGBAT NEGATIVE CIRCLED DIGIT FOUR
        '10106' => '8677'             , # G+8677  U+277A  a134,  DINGBAT NEGATIVE CIRCLED DIGIT FIVE
        '10107' => '8678'             , # G+8678  U+277B  a135,  DINGBAT NEGATIVE CIRCLED DIGIT SIX
        '10108' => '8679'             , # G+8679  U+277C  a136,  DINGBAT NEGATIVE CIRCLED DIGIT SEVEN
        '10109' => '8680'             , # G+8680  U+277D  a137,  DINGBAT NEGATIVE CIRCLED DIGIT EIGHT
        '10110' => '8681'             , # G+8681  U+277E  a138,  DINGBAT NEGATIVE CIRCLED DIGIT NINE
        '10111' => '8682'             , # G+8682  U+277F  a139,  DINGBAT NEGATIVE CIRCLED NUMBER TEN
        '10122' => '8342'             , # G+8342  U+278A  onecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ONE
        '10123' => '8343'             , # G+8343  U+278B  twocircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT TWO
        '10124' => '8344'             , # G+8344  U+278C  threecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT THREE
        '10125' => '8345'             , # G+8345  U+278D  fourcircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FOUR
        '10126' => '8346'             , # G+8346  U+278E  fivecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FIVE
        '10127' => '8347'             , # G+8347  U+278F  sixcircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SIX
        '10128' => '8348'             , # G+8348  U+2790  sevencircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SEVEN
        '10129' => '8349'             , # G+8349  U+2791  eightcircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT EIGHT
        '10130' => '8350'             , # G+8350  U+2792  ninecircleinversesansserif,  DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT NINE
        '10131' => '8351'             , # G+8351  U+2793  a159,  DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN
        '12032' => '6460'             , # G+6460  U+2F00  uni2F00,  KANGXI RADICAL ONE
        '12036' => '6380'             , # G+6380  U+2F04  uni2F04,  KANGXI RADICAL SECOND
        '12038' => '6413'             , # G+6413  U+2F06  uni2F06,  KANGXI RADICAL TWO
        '12040' => '6443'             , # G+6443  U+2F08  uni2F08,  KANGXI RADICAL MAN
        '12042' => '6477'             , # G+6477  U+2F0A  uni2F0A,  KANGXI RADICAL ENTER
        '12043' => '7499'             , # G+7499  U+2F0B  uni2F0B,  KANGXI RADICAL EIGHT
        '12049' => '4270'             , # G+4270  U+2F11  uni2F11,  KANGXI RADICAL KNIFE
        '12050' => '4458'             , # G+4458  U+2F12  uni2F12,  KANGXI RADICAL POWER
        '12052' => '5183'             , # G+5183  U+2F14  uni2F14,  KANGXI RADICAL SPOON
        '12055' => '5772'             , # G+5772  U+2F17  uni2F17,  KANGXI RADICAL TEN
        '12056' => '5077'             , # G+5077  U+2F18  uni2F18,  KANGXI RADICAL DIVINATION
        '12060' => '6201'             , # G+6201  U+2F1C  uni2F1C,  KANGXI RADICAL AGAIN
        '12061' => '3901'             , # G+3901  U+2F1D  uni2F1D,  KANGXI RADICAL MOUTH
        '12063' => '7451'             , # G+7451  U+2F1F  uni2F1F,  KANGXI RADICAL EARTH
        '12064' => '5255'             , # G+5255  U+2F20  uni2F20,  KANGXI RADICAL SCHOLAR
        '12067' => '5403'             , # G+5403  U+2F23  uni2F23,  KANGXI RADICAL EVENING
        '12068' => '4253'             , # G+4253  U+2F24  uni2F24,  KANGXI RADICAL BIG
        '12069' => '4159'             , # G+4159  U+2F25  uni2F25,  KANGXI RADICAL WOMAN
        '12070' => '6488'             , # G+6488  U+2F26  uni2F26,  KANGXI RADICAL CHILD
        '12072' => '7247'             , # G+7247  U+2F28  uni2F28,  KANGXI RADICAL INCH
        '12073' => '5504'             , # G+5504  U+2F29  uni2F29,  KANGXI RADICAL SMALL
        '12075' => '5697'             , # G+5697  U+2F2B  uni2F2B,  KANGXI RADICAL CORPSE
        '12077' => '5305'             , # G+5305  U+2F2D  uni2F2D,  KANGXI RADICAL MOUNTAIN
        '12079' => '3791'             , # G+3791  U+2F2F  uni2F2F,  KANGXI RADICAL WORK
        '12080' => '4077'             , # G+4077  U+2F30  uni2F30,  KANGXI RADICAL ONESELF
        '12081' => '3604'             , # G+3604  U+2F31  uni2F31,  KANGXI RADICAL TURBAN
        '12082' => '3481'             , # G+3481  U+2F32  uni2F32,  KANGXI RADICAL DRY
        '12088' => '3964'             , # G+3964  U+2F38  uni2F38,  KANGXI RADICAL BOW
        '12092' => '5764'             , # G+5764  U+2F3C  uni2F3C,  KANGXI RADICAL HEART
        '12093' => '3804'             , # G+3804  U+2F3D  uni2F3D,  KANGXI RADICAL HALBERD
        '12094' => '7808'             , # G+7808  U+2F3E  uni2F3E,  KANGXI RADICAL DOOR
        '12095' => '5580'             , # G+5580  U+2F3F  uni2F3F,  KANGXI RADICAL HAND
        '12096' => '6951'             , # G+6951  U+2F40  uni2F40,  KANGXI RADICAL BRANCH
        '12098' => '4840'             , # G+4840  U+2F42  uni2F42,  KANGXI RADICAL SCRIPT
        '12099' => '4349'             , # G+4349  U+2F43  uni2F43,  KANGXI RADICAL DIPPER
        '12100' => '4027'             , # G+4027  U+2F44  uni2F44,  KANGXI RADICAL AXE
        '12101' => '4950'             , # G+4950  U+2F45  uni2F45,  KANGXI RADICAL SQUARE
        '12102' => '4819'             , # G+4819  U+2F46  uni2F46,  KANGXI RADICAL NOT
        '12103' => '6464'             , # G+6464  U+2F47  uni2F47,  KANGXI RADICAL SUN
        '12104' => '6125'             , # G+6125  U+2F48  uni2F48,  KANGXI RADICAL SAY
        '12105' => '6281'             , # G+6281  U+2F49  uni2F49,  KANGXI RADICAL MOON
        '12106' => '4788'             , # G+4788  U+2F4A  uni2F4A,  KANGXI RADICAL TREE
        '12107' => '8027'             , # G+8027  U+2F4B  uni2F4B,  KANGXI RADICAL LACK
        '12108' => '6956'             , # G+6956  U+2F4C  uni2F4C,  KANGXI RADICAL STOP
        '12111' => '4822'             , # G+4822  U+2F4F  uni2F4F,  KANGXI RADICAL DO NOT
        '12112' => '5196'             , # G+5196  U+2F50  uni2F50,  KANGXI RADICAL COMPARE
        '12113' => '4776'             , # G+4776  U+2F51  uni2F51,  KANGXI RADICAL FUR
        '12114' => '5774'             , # G+5774  U+2F52  uni2F52,  KANGXI RADICAL CLAN
        '12116' => '5587'             , # G+5587  U+2F54  uni2F54,  KANGXI RADICAL WATER
        '12117' => '7867'             , # G+7867  U+2F55  uni2F55,  KANGXI RADICAL FIRE
        '12118' => '6803'             , # G+6803  U+2F56  uni2F56,  KANGXI RADICAL CLAW
        '12119' => '5129'             , # G+5129  U+2F57  uni2F57,  KANGXI RADICAL FATHER
        '12120' => '7962'             , # G+7962  U+2F58  uni2F58,  KANGXI RADICAL DOUBLE X
        '12122' => '7521'             , # G+7521  U+2F5A  uni2F5A,  KANGXI RADICAL SLICE
        '12123' => '5782'             , # G+5782  U+2F5B  uni2F5B,  KANGXI RADICAL FANG
        '12124' => '6210'             , # G+6210  U+2F5C  uni2F5C,  KANGXI RADICAL COW
        '12125' => '3639'             , # G+3639  U+2F5D  uni2F5D,  KANGXI RADICAL DOG
        '12126' => '7741'             , # G+7741  U+2F5E  uni2F5E,  KANGXI RADICAL PROFOUND
        '12127' => '6081'             , # G+6081  U+2F5F  uni2F5F,  KANGXI RADICAL JADE
        '12128' => '3806'             , # G+3806  U+2F60  uni2F60,  KANGXI RADICAL MELON
        '12129' => '6100'             , # G+6100  U+2F61  uni2F61,  KANGXI RADICAL TILE
        '12130' => '3521'             , # G+3521  U+2F62  uni2F62,  KANGXI RADICAL SWEET
        '12131' => '5370'             , # G+5370  U+2F63  uni2F63,  KANGXI RADICAL LIFE
        '12132' => '6189'             , # G+6189  U+2F64  uni2F64,  KANGXI RADICAL USE
        '12133' => '6664'             , # G+6664  U+2F65  uni2F65,  KANGXI RADICAL FIELD
        '12134' => '7610'             , # G+7610  U+2F66  uni2F66,  KANGXI RADICAL BOLT OF CLOTH
        '12137' => '4993'             , # G+4993  U+2F69  uni2F69,  KANGXI RADICAL WHITE
        '12138' => '7600'             , # G+7600  U+2F6A  uni2F6A,  KANGXI RADICAL SKIN
        '12139' => '4755'             , # G+4755  U+2F6B  uni2F6B,  KANGXI RADICAL DISH
        '12140' => '4791'             , # G+4791  U+2F6C  uni2F6C,  KANGXI RADICAL EYE
        '12141' => '4781'             , # G+4781  U+2F6D  uni2F6D,  KANGXI RADICAL SPEAR
        '12142' => '5709'             , # G+5709  U+2F6E  uni2F6E,  KANGXI RADICAL ARROW
        '12143' => '5413'             , # G+5413  U+2F6F  uni2F6F,  KANGXI RADICAL STONE
        '12144' => '5710'             , # G+5710  U+2F70  uni2F70,  KANGXI RADICAL SPIRIT
        '12146' => '7870'             , # G+7870  U+2F72  uni2F72,  KANGXI RADICAL GRAIN
        '12147' => '7755'             , # G+7755  U+2F73  uni2F73,  KANGXI RADICAL CAVE
        '12148' => '4653'             , # G+4653  U+2F74  uni2F74,  KANGXI RADICAL STAND
        '12149' => '6900'             , # G+6900  U+2F75  uni2F75,  KANGXI RADICAL BAMBOO
        '12150' => '4863'             , # G+4863  U+2F76  uni2F76,  KANGXI RADICAL RICE
        '12152' => '5132'             , # G+5132  U+2F78  uni2F78,  KANGXI RADICAL JAR
        '12154' => '5896'             , # G+5896  U+2F7A  uni2F7A,  KANGXI RADICAL SHEEP
        '12155' => '6218'             , # G+6218  U+2F7B  uni2F7B,  KANGXI RADICAL FEATHER
        '12156' => '4519'             , # G+4519  U+2F7C  uni2F7C,  KANGXI RADICAL OLD
        '12157' => '6427'             , # G+6427  U+2F7D  uni2F7D,  KANGXI RADICAL AND
        '12159' => '6428'             , # G+6428  U+2F7F  uni2F7F,  KANGXI RADICAL EAR
        '12160' => '6368'             , # G+6368  U+2F80  uni2F80,  KANGXI RADICAL BRUSH
        '12161' => '6357'             , # G+6357  U+2F81  uni2F81,  KANGXI RADICAL MEAT
        '12162' => '5749'             , # G+5749  U+2F82  uni2F82,  KANGXI RADICAL MINISTER
        '12163' => '6502'             , # G+6502  U+2F83  uni2F83,  KANGXI RADICAL SELF
        '12164' => '6967'             , # G+6967  U+2F84  uni2F84,  KANGXI RADICAL ARRIVE
        '12165' => '3930'             , # G+3930  U+2F85  uni2F85,  KANGXI RADICAL MORTAR
        '12166' => '5456'             , # G+5456  U+2F86  uni2F86,  KANGXI RADICAL TONGUE
        '12167' => '7158'             , # G+7158  U+2F87  uni2F87,  KANGXI RADICAL OPPOSE
        '12168' => '6888'             , # G+6888  U+2F88  uni2F88,  KANGXI RADICAL BOAT
        '12169' => '3496'             , # G+3496  U+2F89  uni2F89,  KANGXI RADICAL STOPPING
        '12170' => '5368'             , # G+5368  U+2F8A  uni2F8A,  KANGXI RADICAL COLOR
        '12171' => '7232'             , # G+7232  U+2F8B  uni2F8B,  KANGXI RADICAL GRASS
        '12174' => '7756'             , # G+7756  U+2F8E  uni2F8E,  KANGXI RADICAL BLOOD
        '12175' => '7709'             , # G+7709  U+2F8F  uni2F8F,  KANGXI RADICAL WALK ENCLOSURE
        '12176' => '6409'             , # G+6409  U+2F90  uni2F90,  KANGXI RADICAL CLOTHES
        '12178' => '3644'             , # G+3644  U+2F92  uni2F92,  KANGXI RADICAL SEE
        '12179' => '3474'             , # G+3474  U+2F93  uni2F93,  KANGXI RADICAL HORN
        '12180' => '5921'             , # G+5921  U+2F94  uni2F94,  KANGXI RADICAL SPEECH
        '12181' => '3772'             , # G+3772  U+2F95  uni2F95,  KANGXI RADICAL VALLEY
        '12182' => '4355'             , # G+4355  U+2F96  uni2F96,  KANGXI RADICAL BEAN
        '12183' => '5718'             , # G+5718  U+2F97  uni2F97,  KANGXI RADICAL PIG
        '12185' => '7512'             , # G+7512  U+2F99  uni2F99,  KANGXI RADICAL SHELL
        '12186' => '6636'             , # G+6636  U+2F9A  uni2F9A,  KANGXI RADICAL RED
        '12187' => '6892'             , # G+6892  U+2F9B  uni2F9B,  KANGXI RADICAL RUN
        '12188' => '6830'             , # G+6830  U+2F9C  uni2F9C,  KANGXI RADICAL FOOT
        '12189' => '5755'             , # G+5755  U+2F9D  uni2F9D,  KANGXI RADICAL BODY
        '12190' => '7053'             , # G+7053  U+2F9E  uni2F9E,  KANGXI RADICAL CART
        '12191' => '5756'             , # G+5756  U+2F9F  uni2F9F,  KANGXI RADICAL BITTER
        '12192' => '7009'             , # G+7009  U+2FA0  uni2FA0,  KANGXI RADICAL MORNING
        '12194' => '6389'             , # G+6389  U+2FA2  uni2FA2,  KANGXI RADICAL CITY
        '12195' => '6352'             , # G+6352  U+2FA3  uni2FA3,  KANGXI RADICAL WINE
        '12197' => '4634'             , # G+4634  U+2FA5  uni2FA5,  KANGXI RADICAL VILLAGE
        '12198' => '4131'             , # G+4131  U+2FA6  uni2FA6,  KANGXI RADICAL GOLD
        '12199' => '6568'             , # G+6568  U+2FA7  uni2FA7,  KANGXI RADICAL LONG
        '12200' => '4846'             , # G+4846  U+2FA8  uni2FA8,  KANGXI RADICAL GATE
        '12201' => '5147'             , # G+5147  U+2FA9  uni2FA9,  KANGXI RADICAL MOUND
        '12204' => '6227'             , # G+6227  U+2FAC  uni2FAC,  KANGXI RADICAL RAIN
        '12205' => '7203'             , # G+7203  U+2FAD  uni2FAD,  KANGXI RADICAL BLUE
        '12206' => '5220'             , # G+5220  U+2FAE  uni2FAE,  KANGXI RADICAL WRONG
        '12207' => '4744'             , # G+4744  U+2FAF  uni2FAF,  KANGXI RADICAL FACE
        '12208' => '7733'             , # G+7733  U+2FB0  uni2FB0,  KANGXI RADICAL LEATHER
        '12209' => '6307'             , # G+6307  U+2FB1  uni2FB1,  KANGXI RADICAL TANNED LEATHER
        '12211' => '6385'             , # G+6385  U+2FB3  uni2FB3,  KANGXI RADICAL SOUND
        '12212' => '7757'             , # G+7757  U+2FB4  uni2FB4,  KANGXI RADICAL LEAF
        '12213' => '7595'             , # G+7595  U+2FB5  uni2FB5,  KANGXI RADICAL WIND
        '12214' => '5221'             , # G+5221  U+2FB6  uni2FB6,  KANGXI RADICAL FLY
        '12215' => '5733'             , # G+5733  U+2FB7  uni2FB7,  KANGXI RADICAL EAT
        '12216' => '5625'             , # G+5625  U+2FB8  uni2FB8,  KANGXI RADICAL HEAD
        '12217' => '7718'             , # G+7718  U+2FB9  uni2FB9,  KANGXI RADICAL FRAGRANT
        '12218' => '4661'             , # G+4661  U+2FBA  uni2FBA,  KANGXI RADICAL HORSE
        '12219' => '3785'             , # G+3785  U+2FBB  uni2FBB,  KANGXI RADICAL BONE
        '12220' => '3765'             , # G+3765  U+2FBC  uni2FBC,  KANGXI RADICAL TALL
        '12225' => '3993'             , # G+3993  U+2FC1  uni2FC1,  KANGXI RADICAL GHOST
        '12226' => '5910'             , # G+5910  U+2FC2  uni2FC2,  KANGXI RADICAL FISH
        '12227' => '6827'             , # G+6827  U+2FC3  uni2FC3,  KANGXI RADICAL BIRD
        '12228' => '4527'             , # G+4527  U+2FC4  uni2FC4,  KANGXI RADICAL SALT
        '12229' => '4533'             , # G+4533  U+2FC5  uni2FC5,  KANGXI RADICAL DEER
        '12230' => '4726'             , # G+4726  U+2FC6  uni2FC6,  KANGXI RADICAL WHEAT
        '12231' => '4663'             , # G+4663  U+2FC7  uni2FC7,  KANGXI RADICAL HEMP
        '12232' => '7927'             , # G+7927  U+2FC8  uni2FC8,  KANGXI RADICAL YELLOW
        '12233' => '5401'             , # G+5401  U+2FC9  uni2FC9,  KANGXI RADICAL MILLET
        '12234' => '8018'             , # G+8018  U+2FCA  uni2FCA,  KANGXI RADICAL BLACK
        '12237' => '6758'             , # G+6758  U+2FCD  uni2FCD,  KANGXI RADICAL TRIPOD
        '12238' => '3766'             , # G+3766  U+2FCE  uni2FCE,  KANGXI RADICAL DRUM
        '12239' => '5402'             , # G+5402  U+2FCF  uni2FCF,  KANGXI RADICAL RAT
        '12240' => '5222'             , # G+5222  U+2FD0  uni2FD0,  KANGXI RADICAL NOSE
        '12241' => '6781'             , # G+6781  U+2FD1  uni2FD1,  KANGXI RADICAL EVEN
        '12242' => '7359'             , # G+7359  U+2FD2  uni2FD2,  KANGXI RADICAL TOOTH
        '12243' => '4563'             , # G+4563  U+2FD3  uni2FD3,  KANGXI RADICAL DRAGON
        '12244' => '3946'             , # G+3946  U+2FD4  uni2FD4,  KANGXI RADICAL TURTLE
        '12288' => '101'              , # G+101  U+3000  ideographicspace,  IDEOGRAPHIC SPACE
        '12289' => '102'              , # G+102  U+3001  ideographiccomma,  IDEOGRAPHIC COMMA
        '12290' => '103'              , # G+103  U+3002  ideographicperiod,  IDEOGRAPHIC FULL STOP
        '12291' => '108'              , # G+108  U+3003  dittomark,  DITTO MARK
        '12296' => '120'              , # G+120  U+3008  anglebracketleft,  LEFT ANGLE BRACKET
        '12297' => '121'              , # G+121  U+3009  anglebracketright,  RIGHT ANGLE BRACKET
        '12298' => '122'              , # G+122  U+300A  dblanglebracketleft,  LEFT DOUBLE ANGLE BRACKET
        '12299' => '123'              , # G+123  U+300B  dblanglebracketright,  RIGHT DOUBLE ANGLE BRACKET
        '12300' => '124'              , # G+124  U+300C  cornerbracketleft,  LEFT CORNER BRACKET
        '12301' => '125'              , # G+125  U+300D  cornerbracketright,  RIGHT CORNER BRACKET
        '12302' => '126'              , # G+126  U+300E  whitecornerbracketleft,  LEFT WHITE CORNER BRACKET
        '12303' => '127'              , # G+127  U+300F  whitecornerbracketright,  RIGHT WHITE CORNER BRACKET
        '12304' => '128'              , # G+128  U+3010  blacklenticularbracketleft,  LEFT BLACK LENTICULAR BRACKET
        '12305' => '129'              , # G+129  U+3011  blacklenticularbracketright,  RIGHT BLACK LENTICULAR BRACKET
        '12306' => '8700'             , # G+8700  U+3012  postalmark,  POSTAL MARK
        '12307' => '175'              , # G+175  U+3013  getamark,  GETA MARK
        '12308' => '118'              , # G+118  U+3014  tortoiseshellbracketleft,  LEFT TORTOISE SHELL BRACKET
        '12309' => '119'              , # G+119  U+3015  tortoiseshellbracketright,  RIGHT TORTOISE SHELL BRACKET
        '12310' => '8219'             , # G+8219  U+3016  whitelenticularbracketleft,  LEFT WHITE LENTICULAR BRACKET
        '12311' => '8220'             , # G+8220  U+3017  whitelenticularbracketright,  RIGHT WHITE LENTICULAR BRACKET
        '12312' => '8221'             , # G+8221  U+3018  whitetortoiseshellbracketleft,  LEFT WHITE TORTOISE SHELL BRACKET
        '12313' => '8222'             , # G+8222  U+3019  whitetortoiseshellbracketright,  RIGHT WHITE TORTOISE SHELL BRACKET
        '12318' => '9322'             , # G+9322  U+301E  quotedblprime,  DOUBLE PRIME QUOTATION MARK
        '12319' => '9323'             , # G+9323  U+301F  uni301F,  LOW DOUBLE PRIME QUOTATION MARK
        '12320' => '8671'             , # G+8671  U+3020  postalmarkface,  POSTAL MARK FACE
        '12342' => '8701'             , # G+8701  U+3036  circlepostalmark,  CIRCLED POSTAL MARK
        '12353' => '851'              , # G+851  U+3041  asmallhiragana,  HIRAGANA LETTER SMALL A
        '12354' => '852'              , # G+852  U+3042  ahiragana,  HIRAGANA LETTER A
        '12355' => '853'              , # G+853  U+3043  ismallhiragana,  HIRAGANA LETTER SMALL I
        '12356' => '854'              , # G+854  U+3044  ihiragana,  HIRAGANA LETTER I
        '12357' => '855'              , # G+855  U+3045  usmallhiragana,  HIRAGANA LETTER SMALL U
        '12358' => '856'              , # G+856  U+3046  uhiragana,  HIRAGANA LETTER U
        '12359' => '857'              , # G+857  U+3047  esmallhiragana,  HIRAGANA LETTER SMALL E
        '12360' => '858'              , # G+858  U+3048  ehiragana,  HIRAGANA LETTER E
        '12361' => '859'              , # G+859  U+3049  osmallhiragana,  HIRAGANA LETTER SMALL O
        '12362' => '860'              , # G+860  U+304A  ohiragana,  HIRAGANA LETTER O
        '12363' => '861'              , # G+861  U+304B  kahiragana,  HIRAGANA LETTER KA
        '12364' => '862'              , # G+862  U+304C  gahiragana,  HIRAGANA LETTER GA
        '12365' => '863'              , # G+863  U+304D  kihiragana,  HIRAGANA LETTER KI
        '12366' => '864'              , # G+864  U+304E  gihiragana,  HIRAGANA LETTER GI
        '12367' => '865'              , # G+865  U+304F  kuhiragana,  HIRAGANA LETTER KU
        '12368' => '866'              , # G+866  U+3050  guhiragana,  HIRAGANA LETTER GU
        '12369' => '867'              , # G+867  U+3051  kehiragana,  HIRAGANA LETTER KE
        '12370' => '868'              , # G+868  U+3052  gehiragana,  HIRAGANA LETTER GE
        '12371' => '869'              , # G+869  U+3053  kohiragana,  HIRAGANA LETTER KO
        '12372' => '870'              , # G+870  U+3054  gohiragana,  HIRAGANA LETTER GO
        '12373' => '871'              , # G+871  U+3055  sahiragana,  HIRAGANA LETTER SA
        '12374' => '872'              , # G+872  U+3056  zahiragana,  HIRAGANA LETTER ZA
        '12375' => '873'              , # G+873  U+3057  sihiragana,  HIRAGANA LETTER SI
        '12376' => '874'              , # G+874  U+3058  zihiragana,  HIRAGANA LETTER ZI
        '12377' => '875'              , # G+875  U+3059  suhiragana,  HIRAGANA LETTER SU
        '12378' => '876'              , # G+876  U+305A  zuhiragana,  HIRAGANA LETTER ZU
        '12379' => '877'              , # G+877  U+305B  sehiragana,  HIRAGANA LETTER SE
        '12380' => '878'              , # G+878  U+305C  zehiragana,  HIRAGANA LETTER ZE
        '12381' => '879'              , # G+879  U+305D  sohiragana,  HIRAGANA LETTER SO
        '12382' => '880'              , # G+880  U+305E  zohiragana,  HIRAGANA LETTER ZO
        '12383' => '881'              , # G+881  U+305F  tahiragana,  HIRAGANA LETTER TA
        '12384' => '882'              , # G+882  U+3060  dahiragana,  HIRAGANA LETTER DA
        '12385' => '883'              , # G+883  U+3061  tihiragana,  HIRAGANA LETTER TI
        '12386' => '884'              , # G+884  U+3062  dihiragana,  HIRAGANA LETTER DI
        '12387' => '885'              , # G+885  U+3063  tusmallhiragana,  HIRAGANA LETTER SMALL TU
        '12388' => '886'              , # G+886  U+3064  tuhiragana,  HIRAGANA LETTER TU
        '12389' => '887'              , # G+887  U+3065  duhiragana,  HIRAGANA LETTER DU
        '12390' => '888'              , # G+888  U+3066  tehiragana,  HIRAGANA LETTER TE
        '12391' => '889'              , # G+889  U+3067  dehiragana,  HIRAGANA LETTER DE
        '12392' => '890'              , # G+890  U+3068  tohiragana,  HIRAGANA LETTER TO
        '12393' => '891'              , # G+891  U+3069  dohiragana,  HIRAGANA LETTER DO
        '12394' => '892'              , # G+892  U+306A  nahiragana,  HIRAGANA LETTER NA
        '12395' => '893'              , # G+893  U+306B  nihiragana,  HIRAGANA LETTER NI
        '12396' => '894'              , # G+894  U+306C  nuhiragana,  HIRAGANA LETTER NU
        '12397' => '895'              , # G+895  U+306D  nehiragana,  HIRAGANA LETTER NE
        '12398' => '896'              , # G+896  U+306E  nohiragana,  HIRAGANA LETTER NO
        '12399' => '897'              , # G+897  U+306F  hahiragana,  HIRAGANA LETTER HA
        '12400' => '898'              , # G+898  U+3070  bahiragana,  HIRAGANA LETTER BA
        '12401' => '899'              , # G+899  U+3071  pahiragana,  HIRAGANA LETTER PA
        '12402' => '900'              , # G+900  U+3072  hihiragana,  HIRAGANA LETTER HI
        '12403' => '901'              , # G+901  U+3073  bihiragana,  HIRAGANA LETTER BI
        '12404' => '902'              , # G+902  U+3074  pihiragana,  HIRAGANA LETTER PI
        '12405' => '903'              , # G+903  U+3075  huhiragana,  HIRAGANA LETTER HU
        '12406' => '904'              , # G+904  U+3076  buhiragana,  HIRAGANA LETTER BU
        '12407' => '905'              , # G+905  U+3077  puhiragana,  HIRAGANA LETTER PU
        '12408' => '906'              , # G+906  U+3078  hehiragana,  HIRAGANA LETTER HE
        '12409' => '907'              , # G+907  U+3079  behiragana,  HIRAGANA LETTER BE
        '12410' => '908'              , # G+908  U+307A  pehiragana,  HIRAGANA LETTER PE
        '12411' => '909'              , # G+909  U+307B  hohiragana,  HIRAGANA LETTER HO
        '12412' => '910'              , # G+910  U+307C  bohiragana,  HIRAGANA LETTER BO
        '12413' => '911'              , # G+911  U+307D  pohiragana,  HIRAGANA LETTER PO
        '12414' => '912'              , # G+912  U+307E  mahiragana,  HIRAGANA LETTER MA
        '12415' => '913'              , # G+913  U+307F  mihiragana,  HIRAGANA LETTER MI
        '12416' => '914'              , # G+914  U+3080  muhiragana,  HIRAGANA LETTER MU
        '12417' => '915'              , # G+915  U+3081  mehiragana,  HIRAGANA LETTER ME
        '12418' => '916'              , # G+916  U+3082  mohiragana,  HIRAGANA LETTER MO
        '12419' => '917'              , # G+917  U+3083  yasmallhiragana,  HIRAGANA LETTER SMALL YA
        '12420' => '918'              , # G+918  U+3084  yahiragana,  HIRAGANA LETTER YA
        '12421' => '919'              , # G+919  U+3085  yusmallhiragana,  HIRAGANA LETTER SMALL YU
        '12422' => '920'              , # G+920  U+3086  yuhiragana,  HIRAGANA LETTER YU
        '12423' => '921'              , # G+921  U+3087  yosmallhiragana,  HIRAGANA LETTER SMALL YO
        '12424' => '922'              , # G+922  U+3088  yohiragana,  HIRAGANA LETTER YO
        '12425' => '923'              , # G+923  U+3089  rahiragana,  HIRAGANA LETTER RA
        '12426' => '924'              , # G+924  U+308A  rihiragana,  HIRAGANA LETTER RI
        '12427' => '925'              , # G+925  U+308B  ruhiragana,  HIRAGANA LETTER RU
        '12428' => '926'              , # G+926  U+308C  rehiragana,  HIRAGANA LETTER RE
        '12429' => '927'              , # G+927  U+308D  rohiragana,  HIRAGANA LETTER RO
        '12430' => '928'              , # G+928  U+308E  wasmallhiragana,  HIRAGANA LETTER SMALL WA
        '12431' => '929'              , # G+929  U+308F  wahiragana,  HIRAGANA LETTER WA
        '12432' => '930'              , # G+930  U+3090  wihiragana,  HIRAGANA LETTER WI
        '12433' => '931'              , # G+931  U+3091  wehiragana,  HIRAGANA LETTER WE
        '12434' => '932'              , # G+932  U+3092  wohiragana,  HIRAGANA LETTER WO
        '12435' => '933'              , # G+933  U+3093  nhiragana,  HIRAGANA LETTER N
        '12449' => '934'              , # G+934  U+30A1  asmallkatakana,  KATAKANA LETTER SMALL A
        '12450' => '935'              , # G+935  U+30A2  akatakana,  KATAKANA LETTER A
        '12451' => '936'              , # G+936  U+30A3  ismallkatakana,  KATAKANA LETTER SMALL I
        '12452' => '937'              , # G+937  U+30A4  ikatakana,  KATAKANA LETTER I
        '12453' => '938'              , # G+938  U+30A5  usmallkatakana,  KATAKANA LETTER SMALL U
        '12454' => '939'              , # G+939  U+30A6  ukatakana,  KATAKANA LETTER U
        '12455' => '940'              , # G+940  U+30A7  esmallkatakana,  KATAKANA LETTER SMALL E
        '12456' => '941'              , # G+941  U+30A8  ekatakana,  KATAKANA LETTER E
        '12457' => '942'              , # G+942  U+30A9  osmallkatakana,  KATAKANA LETTER SMALL O
        '12458' => '943'              , # G+943  U+30AA  okatakana,  KATAKANA LETTER O
        '12459' => '944'              , # G+944  U+30AB  kakatakana,  KATAKANA LETTER KA
        '12460' => '945'              , # G+945  U+30AC  gakatakana,  KATAKANA LETTER GA
        '12461' => '946'              , # G+946  U+30AD  kikatakana,  KATAKANA LETTER KI
        '12462' => '947'              , # G+947  U+30AE  gikatakana,  KATAKANA LETTER GI
        '12463' => '948'              , # G+948  U+30AF  kukatakana,  KATAKANA LETTER KU
        '12464' => '949'              , # G+949  U+30B0  gukatakana,  KATAKANA LETTER GU
        '12465' => '950'              , # G+950  U+30B1  kekatakana,  KATAKANA LETTER KE
        '12466' => '951'              , # G+951  U+30B2  gekatakana,  KATAKANA LETTER GE
        '12467' => '952'              , # G+952  U+30B3  kokatakana,  KATAKANA LETTER KO
        '12468' => '953'              , # G+953  U+30B4  gokatakana,  KATAKANA LETTER GO
        '12469' => '954'              , # G+954  U+30B5  sakatakana,  KATAKANA LETTER SA
        '12470' => '955'              , # G+955  U+30B6  zakatakana,  KATAKANA LETTER ZA
        '12471' => '956'              , # G+956  U+30B7  sikatakana,  KATAKANA LETTER SI
        '12472' => '957'              , # G+957  U+30B8  zikatakana,  KATAKANA LETTER ZI
        '12473' => '958'              , # G+958  U+30B9  sukatakana,  KATAKANA LETTER SU
        '12474' => '959'              , # G+959  U+30BA  zukatakana,  KATAKANA LETTER ZU
        '12475' => '960'              , # G+960  U+30BB  sekatakana,  KATAKANA LETTER SE
        '12476' => '961'              , # G+961  U+30BC  zekatakana,  KATAKANA LETTER ZE
        '12477' => '962'              , # G+962  U+30BD  sokatakana,  KATAKANA LETTER SO
        '12478' => '963'              , # G+963  U+30BE  zokatakana,  KATAKANA LETTER ZO
        '12479' => '964'              , # G+964  U+30BF  takatakana,  KATAKANA LETTER TA
        '12480' => '965'              , # G+965  U+30C0  dakatakana,  KATAKANA LETTER DA
        '12481' => '966'              , # G+966  U+30C1  tikatakana,  KATAKANA LETTER TI
        '12482' => '967'              , # G+967  U+30C2  dikatakana,  KATAKANA LETTER DI
        '12483' => '968'              , # G+968  U+30C3  tusmallkatakana,  KATAKANA LETTER SMALL TU
        '12484' => '969'              , # G+969  U+30C4  tukatakana,  KATAKANA LETTER TU
        '12485' => '970'              , # G+970  U+30C5  dukatakana,  KATAKANA LETTER DU
        '12486' => '971'              , # G+971  U+30C6  tekatakana,  KATAKANA LETTER TE
        '12487' => '972'              , # G+972  U+30C7  dekatakana,  KATAKANA LETTER DE
        '12488' => '973'              , # G+973  U+30C8  tokatakana,  KATAKANA LETTER TO
        '12489' => '974'              , # G+974  U+30C9  dokatakana,  KATAKANA LETTER DO
        '12490' => '975'              , # G+975  U+30CA  nakatakana,  KATAKANA LETTER NA
        '12491' => '976'              , # G+976  U+30CB  nikatakana,  KATAKANA LETTER NI
        '12492' => '977'              , # G+977  U+30CC  nukatakana,  KATAKANA LETTER NU
        '12493' => '978'              , # G+978  U+30CD  nekatakana,  KATAKANA LETTER NE
        '12494' => '979'              , # G+979  U+30CE  nokatakana,  KATAKANA LETTER NO
        '12495' => '980'              , # G+980  U+30CF  hakatakana,  KATAKANA LETTER HA
        '12496' => '981'              , # G+981  U+30D0  bakatakana,  KATAKANA LETTER BA
        '12497' => '982'              , # G+982  U+30D1  pakatakana,  KATAKANA LETTER PA
        '12498' => '983'              , # G+983  U+30D2  hikatakana,  KATAKANA LETTER HI
        '12499' => '984'              , # G+984  U+30D3  bikatakana,  KATAKANA LETTER BI
        '12500' => '985'              , # G+985  U+30D4  pikatakana,  KATAKANA LETTER PI
        '12501' => '986'              , # G+986  U+30D5  hukatakana,  KATAKANA LETTER HU
        '12502' => '987'              , # G+987  U+30D6  bukatakana,  KATAKANA LETTER BU
        '12503' => '988'              , # G+988  U+30D7  pukatakana,  KATAKANA LETTER PU
        '12504' => '989'              , # G+989  U+30D8  hekatakana,  KATAKANA LETTER HE
        '12505' => '990'              , # G+990  U+30D9  bekatakana,  KATAKANA LETTER BE
        '12506' => '991'              , # G+991  U+30DA  pekatakana,  KATAKANA LETTER PE
        '12507' => '992'              , # G+992  U+30DB  hokatakana,  KATAKANA LETTER HO
        '12508' => '993'              , # G+993  U+30DC  bokatakana,  KATAKANA LETTER BO
        '12509' => '994'              , # G+994  U+30DD  pokatakana,  KATAKANA LETTER PO
        '12510' => '995'              , # G+995  U+30DE  makatakana,  KATAKANA LETTER MA
        '12511' => '996'              , # G+996  U+30DF  mikatakana,  KATAKANA LETTER MI
        '12512' => '997'              , # G+997  U+30E0  mukatakana,  KATAKANA LETTER MU
        '12513' => '998'              , # G+998  U+30E1  mekatakana,  KATAKANA LETTER ME
        '12514' => '999'              , # G+999  U+30E2  mokatakana,  KATAKANA LETTER MO
        '12515' => '1000'             , # G+1000  U+30E3  yasmallkatakana,  KATAKANA LETTER SMALL YA
        '12516' => '1001'             , # G+1001  U+30E4  yakatakana,  KATAKANA LETTER YA
        '12517' => '1002'             , # G+1002  U+30E5  yusmallkatakana,  KATAKANA LETTER SMALL YU
        '12518' => '1003'             , # G+1003  U+30E6  yukatakana,  KATAKANA LETTER YU
        '12519' => '1004'             , # G+1004  U+30E7  yosmallkatakana,  KATAKANA LETTER SMALL YO
        '12520' => '1005'             , # G+1005  U+30E8  yokatakana,  KATAKANA LETTER YO
        '12521' => '1006'             , # G+1006  U+30E9  rakatakana,  KATAKANA LETTER RA
        '12522' => '1007'             , # G+1007  U+30EA  rikatakana,  KATAKANA LETTER RI
        '12523' => '1008'             , # G+1008  U+30EB  rukatakana,  KATAKANA LETTER RU
        '12524' => '1009'             , # G+1009  U+30EC  rekatakana,  KATAKANA LETTER RE
        '12525' => '1010'             , # G+1010  U+30ED  rokatakana,  KATAKANA LETTER RO
        '12526' => '1011'             , # G+1011  U+30EE  wasmallkatakana,  KATAKANA LETTER SMALL WA
        '12527' => '1012'             , # G+1012  U+30EF  wakatakana,  KATAKANA LETTER WA
        '12528' => '1013'             , # G+1013  U+30F0  wikatakana,  KATAKANA LETTER WI
        '12529' => '1014'             , # G+1014  U+30F1  wekatakana,  KATAKANA LETTER WE
        '12530' => '1015'             , # G+1015  U+30F2  wokatakana,  KATAKANA LETTER WO
        '12531' => '1016'             , # G+1016  U+30F3  nkatakana,  KATAKANA LETTER N
        '12532' => '1017'             , # G+1017  U+30F4  vukatakana,  KATAKANA LETTER VU
        '12533' => '1018'             , # G+1018  U+30F5  kasmallkatakana,  KATAKANA LETTER SMALL KA
        '12534' => '1019'             , # G+1019  U+30F6  kesmallkatakana,  KATAKANA LETTER SMALL KE
        '12539' => '104'              , # G+104  U+30FB  dotkatakana,  KATAKANA MIDDLE DOT
        '12540' => '9330'             , # G+9330  U+30FC  prolongedkana,  KATAKANA-HIRAGANA PROLONGED SOUND MARK
        '12593' => '358'              , # G+358  U+3131  kiyeokkorean,  HANGUL LETTER KIYEOK
        '12594' => '359'              , # G+359  U+3132  ssangkiyeokkorean,  HANGUL LETTER SSANGKIYEOK
        '12595' => '360'              , # G+360  U+3133  kiyeoksioskorean,  HANGUL LETTER KIYEOK-SIOS
        '12596' => '361'              , # G+361  U+3134  nieunkorean,  HANGUL LETTER NIEUN
        '12597' => '362'              , # G+362  U+3135  nieuncieuckorean,  HANGUL LETTER NIEUN-CIEUC
        '12598' => '363'              , # G+363  U+3136  nieunhieuhkorean,  HANGUL LETTER NIEUN-HIEUH
        '12599' => '364'              , # G+364  U+3137  tikeutkorean,  HANGUL LETTER TIKEUT
        '12600' => '365'              , # G+365  U+3138  ssangtikeutkorean,  HANGUL LETTER SSANGTIKEUT
        '12601' => '366'              , # G+366  U+3139  rieulkorean,  HANGUL LETTER RIEUL
        '12602' => '367'              , # G+367  U+313A  rieulkiyeokkorean,  HANGUL LETTER RIEUL-KIYEOK
        '12603' => '368'              , # G+368  U+313B  rieulmieumkorean,  HANGUL LETTER RIEUL-MIEUM
        '12604' => '369'              , # G+369  U+313C  rieulpieupkorean,  HANGUL LETTER RIEUL-PIEUP
        '12605' => '370'              , # G+370  U+313D  rieulsioskorean,  HANGUL LETTER RIEUL-SIOS
        '12606' => '371'              , # G+371  U+313E  rieulthieuthkorean,  HANGUL LETTER RIEUL-THIEUTH
        '12607' => '372'              , # G+372  U+313F  rieulphieuphkorean,  HANGUL LETTER RIEUL-PHIEUPH
        '12608' => '373'              , # G+373  U+3140  rieulhieuhkorean,  HANGUL LETTER RIEUL-HIEUH
        '12609' => '374'              , # G+374  U+3141  mieumkorean,  HANGUL LETTER MIEUM
        '12610' => '375'              , # G+375  U+3142  pieupkorean,  HANGUL LETTER PIEUP
        '12611' => '376'              , # G+376  U+3143  ssangpieupkorean,  HANGUL LETTER SSANGPIEUP
        '12612' => '377'              , # G+377  U+3144  pieupsioskorean,  HANGUL LETTER PIEUP-SIOS
        '12613' => '378'              , # G+378  U+3145  sioskorean,  HANGUL LETTER SIOS
        '12614' => '379'              , # G+379  U+3146  ssangsioskorean,  HANGUL LETTER SSANGSIOS
        '12615' => '380'              , # G+380  U+3147  ieungkorean,  HANGUL LETTER IEUNG
        '12616' => '381'              , # G+381  U+3148  cieuckorean,  HANGUL LETTER CIEUC
        '12617' => '382'              , # G+382  U+3149  ssangcieuckorean,  HANGUL LETTER SSANGCIEUC
        '12618' => '383'              , # G+383  U+314A  chieuchkorean,  HANGUL LETTER CHIEUCH
        '12619' => '384'              , # G+384  U+314B  khieukhkorean,  HANGUL LETTER KHIEUKH
        '12620' => '385'              , # G+385  U+314C  thieuthkorean,  HANGUL LETTER THIEUTH
        '12621' => '386'              , # G+386  U+314D  phieuphkorean,  HANGUL LETTER PHIEUPH
        '12622' => '387'              , # G+387  U+314E  hieuhkorean,  HANGUL LETTER HIEUH
        '12623' => '388'              , # G+388  U+314F  akorean,  HANGUL LETTER A
        '12624' => '389'              , # G+389  U+3150  aekorean,  HANGUL LETTER AE
        '12625' => '390'              , # G+390  U+3151  yakorean,  HANGUL LETTER YA
        '12626' => '391'              , # G+391  U+3152  yaekorean,  HANGUL LETTER YAE
        '12627' => '392'              , # G+392  U+3153  eokorean,  HANGUL LETTER EO
        '12628' => '393'              , # G+393  U+3154  ekorean,  HANGUL LETTER E
        '12629' => '394'              , # G+394  U+3155  yeokorean,  HANGUL LETTER YEO
        '12630' => '395'              , # G+395  U+3156  yekorean,  HANGUL LETTER YE
        '12631' => '396'              , # G+396  U+3157  okorean,  HANGUL LETTER O
        '12632' => '397'              , # G+397  U+3158  wakorean,  HANGUL LETTER WA
        '12633' => '398'              , # G+398  U+3159  waekorean,  HANGUL LETTER WAE
        '12634' => '399'              , # G+399  U+315A  oekorean,  HANGUL LETTER OE
        '12635' => '400'              , # G+400  U+315B  yokorean,  HANGUL LETTER YO
        '12636' => '401'              , # G+401  U+315C  ukorean,  HANGUL LETTER U
        '12637' => '402'              , # G+402  U+315D  weokorean,  HANGUL LETTER WEO
        '12638' => '403'              , # G+403  U+315E  wekorean,  HANGUL LETTER WE
        '12639' => '404'              , # G+404  U+315F  wikorean,  HANGUL LETTER WI
        '12640' => '405'              , # G+405  U+3160  yukorean,  HANGUL LETTER YU
        '12641' => '406'              , # G+406  U+3161  eukorean,  HANGUL LETTER EU
        '12642' => '407'              , # G+407  U+3162  yikorean,  HANGUL LETTER YI
        '12643' => '408'              , # G+408  U+3163  ikorean,  HANGUL LETTER I
        '12644' => '101'              , # G+101  U+3164  hangulfiller,  HANGUL FILLER
        '12645' => '409'              , # G+409  U+3165  ssangnieunkorean,  HANGUL LETTER SSANGNIEUN
        '12646' => '410'              , # G+410  U+3166  nieuntikeutkorean,  HANGUL LETTER NIEUN-TIKEUT
        '12647' => '411'              , # G+411  U+3167  nieunsioskorean,  HANGUL LETTER NIEUN-SIOS
        '12648' => '412'              , # G+412  U+3168  nieunpansioskorean,  HANGUL LETTER NIEUN-PANSIOS
        '12649' => '413'              , # G+413  U+3169  rieulkiyeoksioskorean,  HANGUL LETTER RIEUL-KIYEOK-SIOS
        '12650' => '414'              , # G+414  U+316A  rieultikeutkorean,  HANGUL LETTER RIEUL-TIKEUT
        '12651' => '415'              , # G+415  U+316B  rieulpieupsioskorean,  HANGUL LETTER RIEUL-PIEUP-SIOS
        '12652' => '416'              , # G+416  U+316C  rieulpansioskorean,  HANGUL LETTER RIEUL-PANSIOS
        '12653' => '417'              , # G+417  U+316D  rieulyeorinhieuhkorean,  HANGUL LETTER RIEUL-YEORINHIEUH
        '12654' => '418'              , # G+418  U+316E  mieumpieupkorean,  HANGUL LETTER MIEUM-PIEUP
        '12655' => '419'              , # G+419  U+316F  mieumsioskorean,  HANGUL LETTER MIEUM-SIOS
        '12656' => '420'              , # G+420  U+3170  mieumpansioskorean,  HANGUL LETTER MIEUM-PANSIOS
        '12657' => '421'              , # G+421  U+3171  kapyeounmieumkorean,  HANGUL LETTER KAPYEOUNMIEUM
        '12658' => '422'              , # G+422  U+3172  pieupkiyeokkorean,  HANGUL LETTER PIEUP-KIYEOK
        '12659' => '423'              , # G+423  U+3173  pieuptikeutkorean,  HANGUL LETTER PIEUP-TIKEUT
        '12660' => '424'              , # G+424  U+3174  pieupsioskiyeokkorean,  HANGUL LETTER PIEUP-SIOS-KIYEOK
        '12661' => '425'              , # G+425  U+3175  pieupsiostikeutkorean,  HANGUL LETTER PIEUP-SIOS-TIKEUT
        '12662' => '426'              , # G+426  U+3176  pieupcieuckorean,  HANGUL LETTER PIEUP-CIEUC
        '12663' => '427'              , # G+427  U+3177  pieupthieuthkorean,  HANGUL LETTER PIEUP-THIEUTH
        '12664' => '428'              , # G+428  U+3178  kapyeounpieupkorean,  HANGUL LETTER KAPYEOUNPIEUP
        '12665' => '429'              , # G+429  U+3179  kapyeounssangpieupkorean,  HANGUL LETTER KAPYEOUNSSANGPIEUP
        '12666' => '430'              , # G+430  U+317A  sioskiyeokkorean,  HANGUL LETTER SIOS-KIYEOK
        '12667' => '431'              , # G+431  U+317B  siosnieunkorean,  HANGUL LETTER SIOS-NIEUN
        '12668' => '432'              , # G+432  U+317C  siostikeutkorean,  HANGUL LETTER SIOS-TIKEUT
        '12669' => '433'              , # G+433  U+317D  siospieupkorean,  HANGUL LETTER SIOS-PIEUP
        '12670' => '434'              , # G+434  U+317E  sioscieuckorean,  HANGUL LETTER SIOS-CIEUC
        '12671' => '435'              , # G+435  U+317F  pansioskorean,  HANGUL LETTER PANSIOS
        '12672' => '436'              , # G+436  U+3180  ssangieungkorean,  HANGUL LETTER SSANGIEUNG
        '12673' => '437'              , # G+437  U+3181  yesieungkorean,  HANGUL LETTER YESIEUNG
        '12674' => '438'              , # G+438  U+3182  yesieungsioskorean,  HANGUL LETTER YESIEUNG-SIOS
        '12675' => '439'              , # G+439  U+3183  yesieungpansioskorean,  HANGUL LETTER YESIEUNG-PANSIOS
        '12676' => '440'              , # G+440  U+3184  kapyeounphieuphkorean,  HANGUL LETTER KAPYEOUNPHIEUPH
        '12677' => '441'              , # G+441  U+3185  ssanghieuhkorean,  HANGUL LETTER SSANGHIEUH
        '12678' => '442'              , # G+442  U+3186  yeorinhieuhkorean,  HANGUL LETTER YEORINHIEUH
        '12679' => '443'              , # G+443  U+3187  yoyakorean,  HANGUL LETTER YO-YA
        '12680' => '444'              , # G+444  U+3188  yoyaekorean,  HANGUL LETTER YO-YAE
        '12681' => '445'              , # G+445  U+3189  yoikorean,  HANGUL LETTER YO-I
        '12682' => '446'              , # G+446  U+318A  yuyeokorean,  HANGUL LETTER YU-YEO
        '12683' => '447'              , # G+447  U+318B  yuyekorean,  HANGUL LETTER YU-YE
        '12684' => '448'              , # G+448  U+318C  yuikorean,  HANGUL LETTER YU-I
        '12685' => '449'              , # G+449  U+318D  araeakorean,  HANGUL LETTER ARAEA
        '12686' => '450'              , # G+450  U+318E  araeaekorean,  HANGUL LETTER ARAEAE
        '12800' => '773'              , # G+773  U+3200  kiyeokparenkorean,  PARENTHESIZED HANGUL KIYEOK
        '12801' => '774'              , # G+774  U+3201  nieunparenkorean,  PARENTHESIZED HANGUL NIEUN
        '12802' => '775'              , # G+775  U+3202  tikeutparenkorean,  PARENTHESIZED HANGUL TIKEUT
        '12803' => '776'              , # G+776  U+3203  rieulparenkorean,  PARENTHESIZED HANGUL RIEUL
        '12804' => '777'              , # G+777  U+3204  mieumparenkorean,  PARENTHESIZED HANGUL MIEUM
        '12805' => '778'              , # G+778  U+3205  pieupparenkorean,  PARENTHESIZED HANGUL PIEUP
        '12806' => '779'              , # G+779  U+3206  siosparenkorean,  PARENTHESIZED HANGUL SIOS
        '12807' => '780'              , # G+780  U+3207  ieungparenkorean,  PARENTHESIZED HANGUL IEUNG
        '12808' => '781'              , # G+781  U+3208  cieucparenkorean,  PARENTHESIZED HANGUL CIEUC
        '12809' => '782'              , # G+782  U+3209  chieuchparenkorean,  PARENTHESIZED HANGUL CHIEUCH
        '12810' => '783'              , # G+783  U+320A  khieukhparenkorean,  PARENTHESIZED HANGUL KHIEUKH
        '12811' => '784'              , # G+784  U+320B  thieuthparenkorean,  PARENTHESIZED HANGUL THIEUTH
        '12812' => '785'              , # G+785  U+320C  phieuphparenkorean,  PARENTHESIZED HANGUL PHIEUPH
        '12813' => '786'              , # G+786  U+320D  hieuhparenkorean,  PARENTHESIZED HANGUL HIEUH
        '12814' => '787'              , # G+787  U+320E  kiyeokaparenkorean,  PARENTHESIZED HANGUL KIYEOK A
        '12815' => '788'              , # G+788  U+320F  nieunaparenkorean,  PARENTHESIZED HANGUL NIEUN A
        '12816' => '789'              , # G+789  U+3210  tikeutaparenkorean,  PARENTHESIZED HANGUL TIKEUT A
        '12817' => '790'              , # G+790  U+3211  rieulaparenkorean,  PARENTHESIZED HANGUL RIEUL A
        '12818' => '791'              , # G+791  U+3212  mieumaparenkorean,  PARENTHESIZED HANGUL MIEUM A
        '12819' => '792'              , # G+792  U+3213  pieupaparenkorean,  PARENTHESIZED HANGUL PIEUP A
        '12820' => '793'              , # G+793  U+3214  siosaparenkorean,  PARENTHESIZED HANGUL SIOS A
        '12821' => '794'              , # G+794  U+3215  ieungaparenkorean,  PARENTHESIZED HANGUL IEUNG A
        '12822' => '795'              , # G+795  U+3216  cieucaparenkorean,  PARENTHESIZED HANGUL CIEUC A
        '12823' => '796'              , # G+796  U+3217  chieuchaparenkorean,  PARENTHESIZED HANGUL CHIEUCH A
        '12824' => '797'              , # G+797  U+3218  khieukhaparenkorean,  PARENTHESIZED HANGUL KHIEUKH A
        '12825' => '798'              , # G+798  U+3219  thieuthaparenkorean,  PARENTHESIZED HANGUL THIEUTH A
        '12826' => '799'              , # G+799  U+321A  phieuphaparenkorean,  PARENTHESIZED HANGUL PHIEUPH A
        '12827' => '800'              , # G+800  U+321B  hieuhaparenkorean,  PARENTHESIZED HANGUL HIEUH A
        '12828' => '257'              , # G+257  U+321C  cieucuparenkorean,  PARENTHESIZED HANGUL CIEUC U
        '12849' => '8788'             , # G+8788  U+3231  ideographicstockparen,  PARENTHESIZED IDEOGRAPH STOCK
        '12857' => '8789'             , # G+8789  U+3239  ideographicrepresentparen,  PARENTHESIZED IDEOGRAPH REPRESENT
        '12896' => '679'              , # G+679  U+3260  kiyeokcirclekorean,  CIRCLED HANGUL KIYEOK
        '12897' => '680'              , # G+680  U+3261  nieuncirclekorean,  CIRCLED HANGUL NIEUN
        '12898' => '681'              , # G+681  U+3262  tikeutcirclekorean,  CIRCLED HANGUL TIKEUT
        '12899' => '682'              , # G+682  U+3263  rieulcirclekorean,  CIRCLED HANGUL RIEUL
        '12900' => '683'              , # G+683  U+3264  mieumcirclekorean,  CIRCLED HANGUL MIEUM
        '12901' => '684'              , # G+684  U+3265  pieupcirclekorean,  CIRCLED HANGUL PIEUP
        '12902' => '685'              , # G+685  U+3266  sioscirclekorean,  CIRCLED HANGUL SIOS
        '12903' => '686'              , # G+686  U+3267  ieungcirclekorean,  CIRCLED HANGUL IEUNG
        '12904' => '687'              , # G+687  U+3268  cieuccirclekorean,  CIRCLED HANGUL CIEUC
        '12905' => '688'              , # G+688  U+3269  chieuchcirclekorean,  CIRCLED HANGUL CHIEUCH
        '12906' => '689'              , # G+689  U+326A  khieukhcirclekorean,  CIRCLED HANGUL KHIEUKH
        '12907' => '690'              , # G+690  U+326B  thieuthcirclekorean,  CIRCLED HANGUL THIEUTH
        '12908' => '691'              , # G+691  U+326C  phieuphcirclekorean,  CIRCLED HANGUL PHIEUPH
        '12909' => '692'              , # G+692  U+326D  hieuhcirclekorean,  CIRCLED HANGUL HIEUH
        '12910' => '693'              , # G+693  U+326E  kiyeokacirclekorean,  CIRCLED HANGUL KIYEOK A
        '12911' => '694'              , # G+694  U+326F  nieunacirclekorean,  CIRCLED HANGUL NIEUN A
        '12912' => '695'              , # G+695  U+3270  tikeutacirclekorean,  CIRCLED HANGUL TIKEUT A
        '12913' => '696'              , # G+696  U+3271  rieulacirclekorean,  CIRCLED HANGUL RIEUL A
        '12914' => '697'              , # G+697  U+3272  mieumacirclekorean,  CIRCLED HANGUL MIEUM A
        '12915' => '698'              , # G+698  U+3273  pieupacirclekorean,  CIRCLED HANGUL PIEUP A
        '12916' => '699'              , # G+699  U+3274  siosacirclekorean,  CIRCLED HANGUL SIOS A
        '12917' => '700'              , # G+700  U+3275  ieungacirclekorean,  CIRCLED HANGUL IEUNG A
        '12918' => '701'              , # G+701  U+3276  cieucacirclekorean,  CIRCLED HANGUL CIEUC A
        '12919' => '702'              , # G+702  U+3277  chieuchacirclekorean,  CIRCLED HANGUL CHIEUCH A
        '12920' => '703'              , # G+703  U+3278  khieukhacirclekorean,  CIRCLED HANGUL KHIEUKH A
        '12921' => '704'              , # G+704  U+3279  thieuthacirclekorean,  CIRCLED HANGUL THIEUTH A
        '12922' => '705'              , # G+705  U+327A  phieuphacirclekorean,  CIRCLED HANGUL PHIEUPH A
        '12923' => '706'              , # G+706  U+327B  hieuhacirclekorean,  CIRCLED HANGUL HIEUH A
        '12927' => '256'              , # G+256  U+327F  koreanstandardsymbol,  KOREAN STANDARD SYMBOL
        '12938' => '9301'             , # G+9301  U+328A  ideographmooncircle,  CIRCLED IDEOGRAPH MOON
        '12939' => '9302'             , # G+9302  U+328B  ideographfirecircle,  CIRCLED IDEOGRAPH FIRE
        '12940' => '9303'             , # G+9303  U+328C  ideographwatercircle,  CIRCLED IDEOGRAPH WATER
        '12941' => '9304'             , # G+9304  U+328D  ideographwoodcircle,  CIRCLED IDEOGRAPH WOOD
        '12942' => '9305'             , # G+9305  U+328E  ideographmetalcircle,  CIRCLED IDEOGRAPH METAL
        '12943' => '9306'             , # G+9306  U+328F  ideographearthcircle,  CIRCLED IDEOGRAPH EARTH
        '12944' => '9300'             , # G+9300  U+3290  ideographsuncircle,  CIRCLED IDEOGRAPH SUN
        '12948' => '9080'             , # G+9080  U+3294  ideographnamecircle,  CIRCLED IDEOGRAPH NAME
        '12958' => '8761'             , # G+8761  U+329E  ideographicprintcircle,  CIRCLED IDEOGRAPH PRINT
        '12965' => '9096'             , # G+9096  U+32A5  ideographiccentrecircle,  CIRCLED IDEOGRAPH CENTRE
        '13184' => '627'              , # G+627  U+3380  paampssquare,  SQUARE PA AMPS
        '13185' => '628'              , # G+628  U+3381  nasquare,  SQUARE NA
        '13186' => '629'              , # G+629  U+3382  muasquare,  SQUARE MU A
        '13187' => '630'              , # G+630  U+3383  masquare,  SQUARE MA
        '13188' => '631'              , # G+631  U+3384  kasquare,  SQUARE KA
        '13192' => '612'              , # G+612  U+3388  calsquare,  SQUARE CAL
        '13193' => '613'              , # G+613  U+3389  kcalsquare,  SQUARE KCAL
        '13194' => '646'              , # G+646  U+338A  pfsquare,  SQUARE PF
        '13195' => '647'              , # G+647  U+338B  nfsquare,  SQUARE NF
        '13196' => '648'              , # G+648  U+338C  mufsquare,  SQUARE MU F
        '13197' => '608'              , # G+608  U+338D  mugsquare,  SQUARE MU G
        '13198' => '609'              , # G+609  U+338E  squaremg,  SQUARE MG
        '13199' => '610'              , # G+610  U+338F  squarekg,  SQUARE KG
        '13200' => '638'              , # G+638  U+3390  Hzsquare,  SQUARE HZ
        '13201' => '639'              , # G+639  U+3391  khzsquare,  SQUARE KHZ
        '13202' => '640'              , # G+640  U+3392  mhzsquare,  SQUARE MHZ
        '13203' => '641'              , # G+641  U+3393  ghzsquare,  SQUARE GHZ
        '13204' => '642'              , # G+642  U+3394  thzsquare,  SQUARE THZ
        '13205' => '587'              , # G+587  U+3395  mulsquare,  SQUARE MU L
        '13206' => '588'              , # G+588  U+3396  mlsquare,  SQUARE ML
        '13207' => '589'              , # G+589  U+3397  dlsquare,  SQUARE DL
        '13208' => '591'              , # G+591  U+3398  klsquare,  SQUARE KL
        '13209' => '597'              , # G+597  U+3399  fmsquare,  SQUARE FM
        '13210' => '598'              , # G+598  U+339A  nmsquare,  SQUARE NM
        '13211' => '599'              , # G+599  U+339B  mumsquare,  SQUARE MU M
        '13212' => '600'              , # G+600  U+339C  squaremm,  SQUARE MM
        '13213' => '601'              , # G+601  U+339D  squarecm,  SQUARE CM
        '13214' => '602'              , # G+602  U+339E  squarekm,  SQUARE KM
        '13215' => '603'              , # G+603  U+339F  mmsquaredsquare,  SQUARE MM SQUARED
        '13216' => '604'              , # G+604  U+33A0  cmsquaredsquare,  SQUARE CM SQUARED
        '13217' => '605'              , # G+605  U+33A1  squaremsquared,  SQUARE M SQUARED
        '13218' => '606'              , # G+606  U+33A2  kmsquaredsquare,  SQUARE KM SQUARED
        '13219' => '593'              , # G+593  U+33A3  mmcubedsquare,  SQUARE MM CUBED
        '13220' => '594'              , # G+594  U+33A4  cmcubedsquare,  SQUARE CM CUBED
        '13221' => '595'              , # G+595  U+33A5  mcubedsquare,  SQUARE M CUBED
        '13222' => '596'              , # G+596  U+33A6  kmcubedsquare,  SQUARE KM CUBED
        '13223' => '615'              , # G+615  U+33A7  moverssquare,  SQUARE M OVER S
        '13224' => '616'              , # G+616  U+33A8  moverssquaredsquare,  SQUARE M OVER S SQUARED
        '13225' => '655'              , # G+655  U+33A9  pasquare,  SQUARE PA
        '13226' => '656'              , # G+656  U+33AA  kpasquare,  SQUARE KPA
        '13227' => '657'              , # G+657  U+33AB  mpasquare,  SQUARE MPA
        '13228' => '658'              , # G+658  U+33AC  gpasquare,  SQUARE GPA
        '13229' => '651'              , # G+651  U+33AD  radsquare,  SQUARE RAD
        '13230' => '652'              , # G+652  U+33AE  radoverssquare,  SQUARE RAD OVER S
        '13231' => '653'              , # G+653  U+33AF  radoverssquaredsquare,  SQUARE RAD OVER S SQUARED
        '13232' => '617'              , # G+617  U+33B0  pssquare,  SQUARE PS
        '13233' => '618'              , # G+618  U+33B1  nssquare,  SQUARE NS
        '13234' => '619'              , # G+619  U+33B2  mussquare,  SQUARE MU S
        '13235' => '620'              , # G+620  U+33B3  mssquare,  SQUARE MS
        '13236' => '621'              , # G+621  U+33B4  pvsquare,  SQUARE PV
        '13237' => '622'              , # G+622  U+33B5  nvsquare,  SQUARE NV
        '13238' => '623'              , # G+623  U+33B6  muvsquare,  SQUARE MU V
        '13239' => '624'              , # G+624  U+33B7  mvsquare,  SQUARE MV
        '13240' => '625'              , # G+625  U+33B8  kvsquare,  SQUARE KV
        '13241' => '626'              , # G+626  U+33B9  mvmegasquare,  SQUARE MV MEGA
        '13242' => '632'              , # G+632  U+33BA  pwsquare,  SQUARE PW
        '13243' => '633'              , # G+633  U+33BB  nwsquare,  SQUARE NW
        '13244' => '634'              , # G+634  U+33BC  muwsquare,  SQUARE MU W
        '13245' => '635'              , # G+635  U+33BD  mwsquare,  SQUARE MW
        '13246' => '636'              , # G+636  U+33BE  kwsquare,  SQUARE KW
        '13247' => '637'              , # G+637  U+33BF  mwmegasquare,  SQUARE MW MEGA
        '13248' => '644'              , # G+644  U+33C0  kohmsquare,  SQUARE K OHM
        '13249' => '645'              , # G+645  U+33C1  mohmsquare,  SQUARE M OHM
        '13250' => '261'              , # G+261  U+33C2  amsquare,  SQUARE AM
        '13251' => '662'              , # G+662  U+33C3  bqsquare,  SQUARE BQ
        '13252' => '592'              , # G+592  U+33C4  squarecc,  SQUARE CC
        '13253' => '650'              , # G+650  U+33C5  cdsquare,  SQUARE CD
        '13254' => '665'              , # G+665  U+33C6  coverkgsquare,  SQUARE C OVER KG
        '13255' => '259'              , # G+259  U+33C7  cosquare,  SQUARE CO
        '13256' => '614'              , # G+614  U+33C8  dbsquare,  SQUARE DB
        '13257' => '663'              , # G+663  U+33C9  gysquare,  SQUARE GY
        '13258' => '607'              , # G+607  U+33CA  hasquare,  SQUARE HA
        '13259' => '8790'             , # G+8790  U+33CB  HPsquare,  SQUARE HP
        '13263' => '611'              , # G+611  U+33CF  ktsquare,  SQUARE KT
        '13264' => '660'              , # G+660  U+33D0  lmsquare,  SQUARE LM
        '13267' => '661'              , # G+661  U+33D3  lxsquare,  SQUARE LX
        '13270' => '649'              , # G+649  U+33D6  molsquare,  SQUARE MOL
        '13272' => '262'              , # G+262  U+33D8  pmsquare,  SQUARE PM
        '13275' => '654'              , # G+654  U+33DB  srsquare,  SQUARE SR
        '13276' => '664'              , # G+664  U+33DC  svsquare,  SQUARE SV
        '13277' => '659'              , # G+659  U+33DD  wbsquare,  SQUARE WB
        '19968' => '6460'             , # G+6460  U+4E00  uni4E00,  CJK UNIFIED IDEOGRAPH-4E00
        '19969' => '6704'             , # G+6704  U+4E01  uni4E01,  CJK UNIFIED IDEOGRAPH-4E01
        '19971' => '7364'             , # G+7364  U+4E03  uni4E03,  CJK UNIFIED IDEOGRAPH-4E03
        '19975' => '4670'             , # G+4670  U+4E07  uni4E07,  CJK UNIFIED IDEOGRAPH-4E07
        '19976' => '6534'             , # G+6534  U+4E08  uni4E08,  CJK UNIFIED IDEOGRAPH-4E08
        '19977' => '5320'             , # G+5320  U+4E09  uni4E09,  CJK UNIFIED IDEOGRAPH-4E09
        '19978' => '5331'             , # G+5331  U+4E0A  uni4E0A,  CJK UNIFIED IDEOGRAPH-4E0A
        '19979' => '7616'             , # G+7616  U+4E0B  uni4E0B,  CJK UNIFIED IDEOGRAPH-4E0B
        '19981' => '5109'             , # G+5109  U+4E0D  uni4E0D,  CJK UNIFIED IDEOGRAPH-4E0D
        '19985' => '7288'             , # G+7288  U+4E11  uni4E11,  CJK UNIFIED IDEOGRAPH-4E11
        '19988' => '7041'             , # G+7041  U+4E14  uni4E14,  CJK UNIFIED IDEOGRAPH-4E14
        '19989' => '5181'             , # G+5181  U+4E15  uni4E15,  CJK UNIFIED IDEOGRAPH-4E15
        '19990' => '5492'             , # G+5492  U+4E16  uni4E16,  CJK UNIFIED IDEOGRAPH-4E16
        '19992' => '3893'             , # G+3893  U+4E18  uni4E18,  CJK UNIFIED IDEOGRAPH-4E18
        '19993' => '5041'             , # G+5041  U+4E19  uni4E19,  CJK UNIFIED IDEOGRAPH-4E19
        '19998' => '5682'             , # G+5682  U+4E1E  uni4E1E,  CJK UNIFIED IDEOGRAPH-4E1E
        '20013' => '6922'             , # G+6922  U+4E2D  uni4E2D,  CJK UNIFIED IDEOGRAPH-4E2D
        '20018' => '3802'             , # G+3802  U+4E32  uni4E32,  CJK UNIFIED IDEOGRAPH-4E32
        '20024' => '7882'             , # G+7882  U+4E38  uni4E38,  CJK UNIFIED IDEOGRAPH-4E38
        '20025' => '4192'             , # G+4192  U+4E39  uni4E39,  CJK UNIFIED IDEOGRAPH-4E39
        '20027' => '6860'             , # G+6860  U+4E3B  uni4E3B,  CJK UNIFIED IDEOGRAPH-4E3B
        '20034' => '6029'             , # G+6029  U+4E42  uni4E42,  CJK UNIFIED IDEOGRAPH-4E42
        '20035' => '4154'             , # G+4154  U+4E43  uni4E43,  CJK UNIFIED IDEOGRAPH-4E43
        '20037' => '3894'             , # G+3894  U+4E45  uni4E45,  CJK UNIFIED IDEOGRAPH-4E45
        '20043' => '6942'             , # G+6942  U+4E4B  uni4E4B,  CJK UNIFIED IDEOGRAPH-4E4B
        '20045' => '5241'             , # G+5241  U+4E4D  uni4E4D,  CJK UNIFIED IDEOGRAPH-4E4D
        '20046' => '7800'             , # G+7800  U+4E4E  uni4E4E,  CJK UNIFIED IDEOGRAPH-4E4E
        '20047' => '7614'             , # G+7614  U+4E4F  uni4E4F,  CJK UNIFIED IDEOGRAPH-4E4F
        '20054' => '3855'             , # G+3855  U+4E56  uni4E56,  CJK UNIFIED IDEOGRAPH-4E56
        '20056' => '5683'             , # G+5683  U+4E58  uni4E58,  CJK UNIFIED IDEOGRAPH-4E58
        '20057' => '6380'             , # G+6380  U+4E59  uni4E59,  CJK UNIFIED IDEOGRAPH-4E59
        '20061' => '3895'             , # G+3895  U+4E5D  uni4E5D,  CJK UNIFIED IDEOGRAPH-4E5D
        '20062' => '3613'             , # G+3613  U+4E5E  uni4E5E,  CJK UNIFIED IDEOGRAPH-4E5E
        '20063' => '5862'             , # G+5862  U+4E5F  uni4E5F,  CJK UNIFIED IDEOGRAPH-4E5F
        '20075' => '3500'             , # G+3500  U+4E6B  uni4E6B,  CJK UNIFIED IDEOGRAPH-4E6B
        '20077' => '4329'             , # G+4329  U+4E6D  uni4E6D,  CJK UNIFIED IDEOGRAPH-4E6D
        '20083' => '6309'             , # G+6309  U+4E73  uni4E73,  CJK UNIFIED IDEOGRAPH-4E73
        '20086' => '5092'             , # G+5092  U+4E76  uni4E76,  CJK UNIFIED IDEOGRAPH-4E76
        '20087' => '5315'             , # G+5315  U+4E77  uni4E77,  CJK UNIFIED IDEOGRAPH-4E77
        '20094' => '3601'             , # G+3601  U+4E7E  uni4E7E,  CJK UNIFIED IDEOGRAPH-4E7E
        '20098' => '4389'             , # G+4389  U+4E82  uni4E82,  CJK UNIFIED IDEOGRAPH-4E82
        '20102' => '4551'             , # G+4551  U+4E86  uni4E86,  CJK UNIFIED IDEOGRAPH-4E86
        '20104' => '5934'             , # G+5934  U+4E88  uni4E88,  CJK UNIFIED IDEOGRAPH-4E88
        '20107' => '5242'             , # G+5242  U+4E8B  uni4E8B,  CJK UNIFIED IDEOGRAPH-4E8B
        '20108' => '6413'             , # G+6413  U+4E8C  uni4E8C,  CJK UNIFIED IDEOGRAPH-4E8C
        '20110' => '6197'             , # G+6197  U+4E8E  uni4E8E,  CJK UNIFIED IDEOGRAPH-4E8E
        '20112' => '6252'             , # G+6252  U+4E90  uni4E90,  CJK UNIFIED IDEOGRAPH-4E90
        '20113' => '6238'             , # G+6238  U+4E91  uni4E91,  CJK UNIFIED IDEOGRAPH-4E91
        '20114' => '7801'             , # G+7801  U+4E92  uni4E92,  CJK UNIFIED IDEOGRAPH-4E92
        '20116' => '6049'             , # G+6049  U+4E94  uni4E94,  CJK UNIFIED IDEOGRAPH-4E94
        '20117' => '6705'             , # G+6705  U+4E95  uni4E95,  CJK UNIFIED IDEOGRAPH-4E95
        '20120' => '4058'             , # G+4058  U+4E98  uni4E98,  CJK UNIFIED IDEOGRAPH-4E98
        '20123' => '5243'             , # G+5243  U+4E9B  uni4E9B,  CJK UNIFIED IDEOGRAPH-4E9B
        '20126' => '5775'             , # G+5775  U+4E9E  uni4E9E,  CJK UNIFIED IDEOGRAPH-4E9E
        '20129' => '4696'             , # G+4696  U+4EA1  uni4EA1,  CJK UNIFIED IDEOGRAPH-4EA1
        '20130' => '7670'             , # G+7670  U+4EA2  uni4EA2,  CJK UNIFIED IDEOGRAPH-4EA2
        '20132' => '3868'             , # G+3868  U+4EA4  uni4EA4,  CJK UNIFIED IDEOGRAPH-4EA4
        '20133' => '7685'             , # G+7685  U+4EA5  uni4EA5,  CJK UNIFIED IDEOGRAPH-4EA5
        '20134' => '5947'             , # G+5947  U+4EA6  uni4EA6,  CJK UNIFIED IDEOGRAPH-4EA6
        '20136' => '7771'             , # G+7771  U+4EA8  uni4EA8,  CJK UNIFIED IDEOGRAPH-4EA8
        '20139' => '7710'             , # G+7710  U+4EAB  uni4EAB,  CJK UNIFIED IDEOGRAPH-4EAB
        '20140' => '3660'             , # G+3660  U+4EAC  uni4EAC,  CJK UNIFIED IDEOGRAPH-4EAC
        '20141' => '6706'             , # G+6706  U+4EAD  uni4EAD,  CJK UNIFIED IDEOGRAPH-4EAD
        '20142' => '4427'             , # G+4427  U+4EAE  uni4EAE,  CJK UNIFIED IDEOGRAPH-4EAE
        '20150' => '4193'             , # G+4193  U+4EB6  uni4EB6,  CJK UNIFIED IDEOGRAPH-4EB6
        '20154' => '6443'             , # G+6443  U+4EBA  uni4EBA,  CJK UNIFIED IDEOGRAPH-4EBA
        '20160' => '5771'             , # G+5771  U+4EC0  uni4EC0,  CJK UNIFIED IDEOGRAPH-4EC0
        '20161' => '6444'             , # G+6444  U+4EC1  uni4EC1,  CJK UNIFIED IDEOGRAPH-4EC1
        '20164' => '7331'             , # G+7331  U+4EC4  uni4EC4,  CJK UNIFIED IDEOGRAPH-4EC4
        '20167' => '3896'             , # G+3896  U+4EC7  uni4EC7,  CJK UNIFIED IDEOGRAPH-4EC7
        '20170' => '4038'             , # G+4038  U+4ECA  uni4ECA,  CJK UNIFIED IDEOGRAPH-4ECA
        '20171' => '3560'             , # G+3560  U+4ECB  uni4ECB,  CJK UNIFIED IDEOGRAPH-4ECB
        '20173' => '6479'             , # G+6479  U+4ECD  uni4ECD,  CJK UNIFIED IDEOGRAPH-4ECD
        '20180' => '6483'             , # G+6483  U+4ED4  uni4ED4,  CJK UNIFIED IDEOGRAPH-4ED4
        '20181' => '5244'             , # G+5244  U+4ED5  uni4ED5,  CJK UNIFIED IDEOGRAPH-4ED5
        '20182' => '7380'             , # G+7380  U+4ED6  uni4ED6,  CJK UNIFIED IDEOGRAPH-4ED6
        '20183' => '6535'             , # G+6535  U+4ED7  uni4ED7,  CJK UNIFIED IDEOGRAPH-4ED7
        '20184' => '5110'             , # G+5110  U+4ED8  uni4ED8,  CJK UNIFIED IDEOGRAPH-4ED8
        '20185' => '5418'             , # G+5418  U+4ED9  uni4ED9,  CJK UNIFIED IDEOGRAPH-4ED9
        '20189' => '4331'             , # G+4331  U+4EDD  uni4EDD,  CJK UNIFIED IDEOGRAPH-4EDD
        '20191' => '7148'             , # G+7148  U+4EDF  uni4EDF,  CJK UNIFIED IDEOGRAPH-4EDF
        '20195' => '4250'             , # G+4250  U+4EE3  uni4EE3,  CJK UNIFIED IDEOGRAPH-4EE3
        '20196' => '4489'             , # G+4489  U+4EE4  uni4EE4,  CJK UNIFIED IDEOGRAPH-4EE4
        '20197' => '6414'             , # G+6414  U+4EE5  uni4EE5,  CJK UNIFIED IDEOGRAPH-4EE5
        '20208' => '5833'             , # G+5833  U+4EF0  uni4EF0,  CJK UNIFIED IDEOGRAPH-4EF0
        '20210' => '6923'             , # G+6923  U+4EF2  uni4EF2,  CJK UNIFIED IDEOGRAPH-4EF2
        '20214' => '3602'             , # G+3602  U+4EF6  uni4EF6,  CJK UNIFIED IDEOGRAPH-4EF6
        '20215' => '3561'             , # G+3561  U+4EF7  uni4EF7,  CJK UNIFIED IDEOGRAPH-4EF7
        '20219' => '6469'             , # G+6469  U+4EFB  uni4EFB,  CJK UNIFIED IDEOGRAPH-4EFB
        '20225' => '4062'             , # G+4062  U+4F01  uni4F01,  CJK UNIFIED IDEOGRAPH-4F01
        '20233' => '7671'             , # G+7671  U+4F09  uni4F09,  CJK UNIFIED IDEOGRAPH-4F09
        '20234' => '6415'             , # G+6415  U+4F0A  uni4F0A,  CJK UNIFIED IDEOGRAPH-4F0A
        '20235' => '4051'             , # G+4051  U+4F0B  uni4F0B,  CJK UNIFIED IDEOGRAPH-4F0B
        '20237' => '6050'             , # G+6050  U+4F0D  uni4F0D,  CJK UNIFIED IDEOGRAPH-4F0D
        '20238' => '4063'             , # G+4063  U+4F0E  uni4F0E,  CJK UNIFIED IDEOGRAPH-4F0E
        '20239' => '5074'             , # G+5074  U+4F0F  uni4F0F,  CJK UNIFIED IDEOGRAPH-4F0F
        '20240' => '5005'             , # G+5005  U+4F10  uni4F10,  CJK UNIFIED IDEOGRAPH-4F10
        '20241' => '8005'             , # G+8005  U+4F11  uni4F11,  CJK UNIFIED IDEOGRAPH-4F11
        '20271' => '4988'             , # G+4988  U+4F2F  uni4F2F,  CJK UNIFIED IDEOGRAPH-4F2F
        '20276' => '4905'             , # G+4905  U+4F34  uni4F34,  CJK UNIFIED IDEOGRAPH-4F34
        '20278' => '4490'             , # G+4490  U+4F36  uni4F36,  CJK UNIFIED IDEOGRAPH-4F36
        '20280' => '5735'             , # G+5735  U+4F38  uni4F38,  CJK UNIFIED IDEOGRAPH-4F38
        '20282' => '5245'             , # G+5245  U+4F3A  uni4F3A,  CJK UNIFIED IDEOGRAPH-4F3A
        '20284' => '5246'             , # G+5246  U+4F3C  uni4F3C,  CJK UNIFIED IDEOGRAPH-4F3C
        '20285' => '3436'             , # G+3436  U+4F3D  uni4F3D,  CJK UNIFIED IDEOGRAPH-4F3D
        '20291' => '6643'             , # G+6643  U+4F43  uni4F43,  CJK UNIFIED IDEOGRAPH-4F43
        '20294' => '4194'             , # G+4194  U+4F46  uni4F46,  CJK UNIFIED IDEOGRAPH-4F46
        '20295' => '6591'             , # G+6591  U+4F47  uni4F47,  CJK UNIFIED IDEOGRAPH-4F47
        '20296' => '7544'             , # G+7544  U+4F48  uni4F48,  CJK UNIFIED IDEOGRAPH-4F48
        '20301' => '6284'             , # G+6284  U+4F4D  uni4F4D,  CJK UNIFIED IDEOGRAPH-4F4D
        '20302' => '6592'             , # G+6592  U+4F4E  uni4F4E,  CJK UNIFIED IDEOGRAPH-4F4E
        '20303' => '6861'             , # G+6861  U+4F4F  uni4F4F,  CJK UNIFIED IDEOGRAPH-4F4F
        '20304' => '6854'             , # G+6854  U+4F50  uni4F50,  CJK UNIFIED IDEOGRAPH-4F50
        '20305' => '6198'             , # G+6198  U+4F51  uni4F51,  CJK UNIFIED IDEOGRAPH-4F51
        '20309' => '7617'             , # G+7617  U+4F55  uni4F55,  CJK UNIFIED IDEOGRAPH-4F55
        '20313' => '5935'             , # G+5935  U+4F59  uni4F59,  CJK UNIFIED IDEOGRAPH-4F59
        '20314' => '6461'             , # G+6461  U+4F5A  uni4F5A,  CJK UNIFIED IDEOGRAPH-4F5A
        '20315' => '5171'             , # G+5171  U+4F5B  uni4F5B,  CJK UNIFIED IDEOGRAPH-4F5B
        '20316' => '6509'             , # G+6509  U+4F5C  uni4F5C,  CJK UNIFIED IDEOGRAPH-4F5C
        '20329' => '7502'             , # G+7502  U+4F69  uni4F69,  CJK UNIFIED IDEOGRAPH-4F69
        '20335' => '5879'             , # G+5879  U+4F6F  uni4F6F,  CJK UNIFIED IDEOGRAPH-4F6F
        '20336' => '4989'             , # G+4989  U+4F70  uni4F70,  CJK UNIFIED IDEOGRAPH-4F70
        '20339' => '3437'             , # G+3437  U+4F73  uni4F73,  CJK UNIFIED IDEOGRAPH-4F73
        '20342' => '4127'             , # G+4127  U+4F76  uni4F76,  CJK UNIFIED IDEOGRAPH-4F76
        '20346' => '6644'             , # G+6644  U+4F7A  uni4F7A,  CJK UNIFIED IDEOGRAPH-4F7A
        '20350' => '6462'             , # G+6462  U+4F7E  uni4F7E,  CJK UNIFIED IDEOGRAPH-4F7E
        '20351' => '5247'             , # G+5247  U+4F7F  uni4F7F,  CJK UNIFIED IDEOGRAPH-4F7F
        '20353' => '5736'             , # G+5736  U+4F81  uni4F81,  CJK UNIFIED IDEOGRAPH-4F81
        '20355' => '3476'             , # G+3476  U+4F83  uni4F83,  CJK UNIFIED IDEOGRAPH-4F83
        '20356' => '7015'             , # G+7015  U+4F84  uni4F84,  CJK UNIFIED IDEOGRAPH-4F84
        '20358' => '4420'             , # G+4420  U+4F86  uni4F86,  CJK UNIFIED IDEOGRAPH-4F86
        '20360' => '7336'             , # G+7336  U+4F88  uni4F88,  CJK UNIFIED IDEOGRAPH-4F88
        '20362' => '3839'             , # G+3839  U+4F8A  uni4F8A,  CJK UNIFIED IDEOGRAPH-4F8A
        '20363' => '4506'             , # G+4506  U+4F8B  uni4F8B,  CJK UNIFIED IDEOGRAPH-4F8B
        '20365' => '5692'             , # G+5692  U+4F8D  uni4F8D,  CJK UNIFIED IDEOGRAPH-4F8D
        '20367' => '6862'             , # G+6862  U+4F8F  uni4F8F,  CJK UNIFIED IDEOGRAPH-4F8F
        '20369' => '6310'             , # G+6310  U+4F91  uni4F91,  CJK UNIFIED IDEOGRAPH-4F91
        '20374' => '4594'             , # G+4594  U+4F96  uni4F96,  CJK UNIFIED IDEOGRAPH-4F96
        '20376' => '7042'             , # G+7042  U+4F98  uni4F98,  CJK UNIFIED IDEOGRAPH-4F98
        '20379' => '3786'             , # G+3786  U+4F9B  uni4F9B,  CJK UNIFIED IDEOGRAPH-4F9B
        '20381' => '6394'             , # G+6394  U+4F9D  uni4F9D,  CJK UNIFIED IDEOGRAPH-4F9D
        '20398' => '4764'             , # G+4764  U+4FAE  uni4FAE,  CJK UNIFIED IDEOGRAPH-4FAE
        '20399' => '7966'             , # G+7966  U+4FAF  uni4FAF,  CJK UNIFIED IDEOGRAPH-4FAF
        '20405' => '7367'             , # G+7367  U+4FB5  uni4FB5,  CJK UNIFIED IDEOGRAPH-4FB5
        '20406' => '4440'             , # G+4440  U+4FB6  uni4FB6,  CJK UNIFIED IDEOGRAPH-4FB6
        '20415' => '7518'             , # G+7518  U+4FBF  uni4FBF,  CJK UNIFIED IDEOGRAPH-4FBF
        '20418' => '3705'             , # G+3705  U+4FC2  uni4FC2,  CJK UNIFIED IDEOGRAPH-4FC2
        '20419' => '7241'             , # G+7241  U+4FC3  uni4FC3,  CJK UNIFIED IDEOGRAPH-4FC3
        '20420' => '5776'             , # G+5776  U+4FC4  uni4FC4,  CJK UNIFIED IDEOGRAPH-4FC4
        '20425' => '6051'             , # G+6051  U+4FC9  uni4FC9,  CJK UNIFIED IDEOGRAPH-4FC9
        '20426' => '6902'             , # G+6902  U+4FCA  uni4FCA,  CJK UNIFIED IDEOGRAPH-4FCA
        '20430' => '6782'             , # G+6782  U+4FCE  uni4FCE,  CJK UNIFIED IDEOGRAPH-4FCE
        '20433' => '6174'             , # G+6174  U+4FD1  uni4FD1,  CJK UNIFIED IDEOGRAPH-4FD1
        '20435' => '3661'             , # G+3661  U+4FD3  uni4FD3,  CJK UNIFIED IDEOGRAPH-4FD3
        '20436' => '7734'             , # G+7734  U+4FD4  uni4FD4,  CJK UNIFIED IDEOGRAPH-4FD4
        '20439' => '5537'             , # G+5537  U+4FD7  uni4FD7,  CJK UNIFIED IDEOGRAPH-4FD7
        '20442' => '4613'             , # G+4613  U+4FDA  uni4FDA,  CJK UNIFIED IDEOGRAPH-4FDA
        '20445' => '5058'             , # G+5058  U+4FDD  uni4FDD,  CJK UNIFIED IDEOGRAPH-4FDD
        '20447' => '5248'             , # G+5248  U+4FDF  uni4FDF,  CJK UNIFIED IDEOGRAPH-4FDF
        '20448' => '7759'             , # G+7759  U+4FE0  uni4FE0,  CJK UNIFIED IDEOGRAPH-4FE0
        '20449' => '5737'             , # G+5737  U+4FE1  uni4FE1,  CJK UNIFIED IDEOGRAPH-4FE1
        '20462' => '5567'             , # G+5567  U+4FEE  uni4FEE,  CJK UNIFIED IDEOGRAPH-4FEE
        '20463' => '5111'             , # G+5111  U+4FEF  uni4FEF,  CJK UNIFIED IDEOGRAPH-4FEF
        '20465' => '3897'             , # G+3897  U+4FF1  uni4FF1,  CJK UNIFIED IDEOGRAPH-4FF1
        '20467' => '4970'             , # G+4970  U+4FF3  uni4FF3,  CJK UNIFIED IDEOGRAPH-4FF3
        '20469' => '7576'             , # G+7576  U+4FF5  uni4FF5,  CJK UNIFIED IDEOGRAPH-4FF5
        '20472' => '5093'             , # G+5093  U+4FF8  uni4FF8,  CJK UNIFIED IDEOGRAPH-4FF8
        '20474' => '5925'             , # G+5925  U+4FFA  uni4FFA,  CJK UNIFIED IDEOGRAPH-4FFA
        '20482' => '5042'             , # G+5042  U+5002  uni5002,  CJK UNIFIED IDEOGRAPH-5002
        '20486' => '4428'             , # G+4428  U+5006  uni5006,  CJK UNIFIED IDEOGRAPH-5006
        '20489' => '7092'             , # G+7092  U+5009  uni5009,  CJK UNIFIED IDEOGRAPH-5009
        '20491' => '3562'             , # G+3562  U+500B  uni500B,  CJK UNIFIED IDEOGRAPH-500B
        '20493' => '4969'             , # G+4969  U+500D  uni500D,  CJK UNIFIED IDEOGRAPH-500D
        '20497' => '4836'             , # G+4836  U+5011  uni5011,  CJK UNIFIED IDEOGRAPH-5011
        '20498' => '4269'             , # G+4269  U+5012  uni5012,  CJK UNIFIED IDEOGRAPH-5012
        '20502' => '7705'             , # G+7705  U+5016  uni5016,  CJK UNIFIED IDEOGRAPH-5016
        '20505' => '7967'             , # G+7967  U+5019  uni5019,  CJK UNIFIED IDEOGRAPH-5019
        '20506' => '6395'             , # G+6395  U+501A  uni501A,  CJK UNIFIED IDEOGRAPH-501A
        '20508' => '7134'             , # G+7134  U+501C  uni501C,  CJK UNIFIED IDEOGRAPH-501C
        '20510' => '3662'             , # G+3662  U+501E  uni501E,  CJK UNIFIED IDEOGRAPH-501E
        '20511' => '7043'             , # G+7043  U+501F  uni501F,  CJK UNIFIED IDEOGRAPH-501F
        '20513' => '7093'             , # G+7093  U+5021  uni5021,  CJK UNIFIED IDEOGRAPH-5021
        '20515' => '4941'             , # G+4941  U+5023  uni5023,  CJK UNIFIED IDEOGRAPH-5023
        '20516' => '7337'             , # G+7337  U+5024  uni5024,  CJK UNIFIED IDEOGRAPH-5024
        '20518' => '3969'             , # G+3969  U+5026  uni5026,  CJK UNIFIED IDEOGRAPH-5026
        '20519' => '6837'             , # G+6837  U+5027  uni5027,  CJK UNIFIED IDEOGRAPH-5027
        '20520' => '3585'             , # G+3585  U+5028  uni5028,  CJK UNIFIED IDEOGRAPH-5028
        '20522' => '6030'             , # G+6030  U+502A  uni502A,  CJK UNIFIED IDEOGRAPH-502A
        '20523' => '4595'             , # G+4595  U+502B  uni502B,  CJK UNIFIED IDEOGRAPH-502B
        '20524' => '7394'             , # G+7394  U+502C  uni502C,  CJK UNIFIED IDEOGRAPH-502C
        '20525' => '6131'             , # G+6131  U+502D  uni502D,  CJK UNIFIED IDEOGRAPH-502D
        '20539' => '5863'             , # G+5863  U+503B  uni503B,  CJK UNIFIED IDEOGRAPH-503B
        '20547' => '5917'             , # G+5917  U+5043  uni5043,  CJK UNIFIED IDEOGRAPH-5043
        '20551' => '3438'             , # G+3438  U+5047  uni5047,  CJK UNIFIED IDEOGRAPH-5047
        '20552' => '3627'             , # G+3627  U+5048  uni5048,  CJK UNIFIED IDEOGRAPH-5048
        '20553' => '6285'             , # G+6285  U+5049  uni5049,  CJK UNIFIED IDEOGRAPH-5049
        '20559' => '7519'             , # G+7519  U+504F  uni504F,  CJK UNIFIED IDEOGRAPH-504F
        '20565' => '7686'             , # G+7686  U+5055  uni5055,  CJK UNIFIED IDEOGRAPH-5055
        '20570' => '6863'             , # G+6863  U+505A  uni505A,  CJK UNIFIED IDEOGRAPH-505A
        '20572' => '6707'             , # G+6707  U+505C  uni505C,  CJK UNIFIED IDEOGRAPH-505C
        '20581' => '3603'             , # G+3603  U+5065  uni5065,  CJK UNIFIED IDEOGRAPH-5065
        '20596' => '7330'             , # G+7330  U+5074  uni5074,  CJK UNIFIED IDEOGRAPH-5074
        '20597' => '6708'             , # G+6708  U+5075  uni5075,  CJK UNIFIED IDEOGRAPH-5075
        '20598' => '6199'             , # G+6199  U+5076  uni5076,  CJK UNIFIED IDEOGRAPH-5076
        '20600' => '7465'             , # G+7465  U+5078  uni5078,  CJK UNIFIED IDEOGRAPH-5078
        '20608' => '3856'             , # G+3856  U+5080  uni5080,  CJK UNIFIED IDEOGRAPH-5080
        '20613' => '5112'             , # G+5112  U+5085  uni5085,  CJK UNIFIED IDEOGRAPH-5085
        '20621' => '4942'             , # G+4942  U+508D  uni508D,  CJK UNIFIED IDEOGRAPH-508D
        '20625' => '3614'             , # G+3614  U+5091  uni5091,  CJK UNIFIED IDEOGRAPH-5091
        '20632' => '5303'             , # G+5303  U+5098  uni5098,  CJK UNIFIED IDEOGRAPH-5098
        '20633' => '5182'             , # G+5182  U+5099  uni5099,  CJK UNIFIED IDEOGRAPH-5099
        '20652' => '7262'             , # G+7262  U+50AC  uni50AC,  CJK UNIFIED IDEOGRAPH-50AC
        '20653' => '6175'             , # G+6175  U+50AD  uni50AD,  CJK UNIFIED IDEOGRAPH-50AD
        '20658' => '6052'             , # G+6052  U+50B2  uni50B2,  CJK UNIFIED IDEOGRAPH-50B2
        '20659' => '6645'             , # G+6645  U+50B3  uni50B3,  CJK UNIFIED IDEOGRAPH-50B3
        '20661' => '7114'             , # G+7114  U+50B5  uni50B5,  CJK UNIFIED IDEOGRAPH-50B5
        '20663' => '5332'             , # G+5332  U+50B7  uni50B7,  CJK UNIFIED IDEOGRAPH-50B7
        '20670' => '3663'             , # G+3663  U+50BE  uni50BE,  CJK UNIFIED IDEOGRAPH-50BE
        '20677' => '4023'             , # G+4023  U+50C5  uni50C5,  CJK UNIFIED IDEOGRAPH-50C5
        '20681' => '7177'             , # G+7177  U+50C9  uni50C9,  CJK UNIFIED IDEOGRAPH-50C9
        '20682' => '5419'             , # G+5419  U+50CA  uni50CA,  CJK UNIFIED IDEOGRAPH-50CA
        '20687' => '5333'             , # G+5333  U+50CF  uni50CF,  CJK UNIFIED IDEOGRAPH-50CF
        '20689' => '3869'             , # G+3869  U+50D1  uni50D1,  CJK UNIFIED IDEOGRAPH-50D1
        '20693' => '5075'             , # G+5075  U+50D5  uni50D5,  CJK UNIFIED IDEOGRAPH-50D5
        '20694' => '8035'             , # G+8035  U+50D6  uni50D6,  CJK UNIFIED IDEOGRAPH-50D6
        '20698' => '4552'             , # G+4552  U+50DA  uni50DA,  CJK UNIFIED IDEOGRAPH-50DA
        '20702' => '6286'             , # G+6286  U+50DE  uni50DE,  CJK UNIFIED IDEOGRAPH-50DE
        '20709' => '6140'             , # G+6140  U+50E5  uni50E5,  CJK UNIFIED IDEOGRAPH-50E5
        '20711' => '5684'             , # G+5684  U+50E7  uni50E7,  CJK UNIFIED IDEOGRAPH-50E7
        '20717' => '7082'             , # G+7082  U+50ED  uni50ED,  CJK UNIFIED IDEOGRAPH-50ED
        '20729' => '3439'             , # G+3439  U+50F9  uni50F9,  CJK UNIFIED IDEOGRAPH-50F9
        '20731' => '5020'             , # G+5020  U+50FB  uni50FB,  CJK UNIFIED IDEOGRAPH-50FB
        '20735' => '5249'             , # G+5249  U+50FF  uni50FF,  CJK UNIFIED IDEOGRAPH-50FF
        '20736' => '6396'             , # G+6396  U+5100  uni5100,  CJK UNIFIED IDEOGRAPH-5100
        '20737' => '6903'             , # G+6903  U+5101  uni5101,  CJK UNIFIED IDEOGRAPH-5101
        '20740' => '5912'             , # G+5912  U+5104  uni5104,  CJK UNIFIED IDEOGRAPH-5104
        '20742' => '3664'             , # G+3664  U+5106  uni5106,  CJK UNIFIED IDEOGRAPH-5106
        '20745' => '3617'             , # G+3617  U+5109  uni5109,  CJK UNIFIED IDEOGRAPH-5109
        '20754' => '6311'             , # G+6311  U+5112  uni5112,  CJK UNIFIED IDEOGRAPH-5112
        '20767' => '5334'             , # G+5334  U+511F  uni511F,  CJK UNIFIED IDEOGRAPH-511F
        '20769' => '4543'             , # G+4543  U+5121  uni5121,  CJK UNIFIED IDEOGRAPH-5121
        '20778' => '6200'             , # G+6200  U+512A  uni512A,  CJK UNIFIED IDEOGRAPH-512A
        '20786' => '6593'             , # G+6593  U+5132  uni5132,  CJK UNIFIED IDEOGRAPH-5132
        '20791' => '4441'             , # G+4441  U+5137  uni5137,  CJK UNIFIED IDEOGRAPH-5137
        '20794' => '4133'             , # G+4133  U+513A  uni513A,  CJK UNIFIED IDEOGRAPH-513A
        '20796' => '5926'             , # G+5926  U+513C  uni513C,  CJK UNIFIED IDEOGRAPH-513C
        '20800' => '6089'             , # G+6089  U+5140  uni5140,  CJK UNIFIED IDEOGRAPH-5140
        '20801' => '6359'             , # G+6359  U+5141  uni5141,  CJK UNIFIED IDEOGRAPH-5141
        '20803' => '6255'             , # G+6255  U+5143  uni5143,  CJK UNIFIED IDEOGRAPH-5143
        '20804' => '7772'             , # G+7772  U+5144  uni5144,  CJK UNIFIED IDEOGRAPH-5144
        '20805' => '7306'             , # G+7306  U+5145  uni5145,  CJK UNIFIED IDEOGRAPH-5145
        '20806' => '6783'             , # G+6783  U+5146  uni5146,  CJK UNIFIED IDEOGRAPH-5146
        '20807' => '8013'             , # G+8013  U+5147  uni5147,  CJK UNIFIED IDEOGRAPH-5147
        '20808' => '5420'             , # G+5420  U+5148  uni5148,  CJK UNIFIED IDEOGRAPH-5148
        '20809' => '3840'             , # G+3840  U+5149  uni5149,  CJK UNIFIED IDEOGRAPH-5149
        '20811' => '4016'             , # G+4016  U+514B  uni514B,  CJK UNIFIED IDEOGRAPH-514B
        '20812' => '7431'             , # G+7431  U+514C  uni514C,  CJK UNIFIED IDEOGRAPH-514C
        '20813' => '4735'             , # G+4735  U+514D  uni514D,  CJK UNIFIED IDEOGRAPH-514D
        '20814' => '7449'             , # G+7449  U+514E  uni514E,  CJK UNIFIED IDEOGRAPH-514E
        '20818' => '5777'             , # G+5777  U+5152  uni5152,  CJK UNIFIED IDEOGRAPH-5152
        '20828' => '4348'             , # G+4348  U+515C  uni515C,  CJK UNIFIED IDEOGRAPH-515C
        '20834' => '4059'             , # G+4059  U+5162  uni5162,  CJK UNIFIED IDEOGRAPH-5162
        '20837' => '6477'             , # G+6477  U+5165  uni5165,  CJK UNIFIED IDEOGRAPH-5165
        '20839' => '4155'             , # G+4155  U+5167  uni5167,  CJK UNIFIED IDEOGRAPH-5167
        '20840' => '6646'             , # G+6646  U+5168  uni5168,  CJK UNIFIED IDEOGRAPH-5168
        '20841' => '4429'             , # G+4429  U+5169  uni5169,  CJK UNIFIED IDEOGRAPH-5169
        '20842' => '6312'             , # G+6312  U+516A  uni516A,  CJK UNIFIED IDEOGRAPH-516A
        '20843' => '7499'             , # G+7499  U+516B  uni516B,  CJK UNIFIED IDEOGRAPH-516B
        '20844' => '3787'             , # G+3787  U+516C  uni516C,  CJK UNIFIED IDEOGRAPH-516C
        '20845' => '4591'             , # G+4591  U+516D  uni516D,  CJK UNIFIED IDEOGRAPH-516D
        '20846' => '7791'             , # G+7791  U+516E  uni516E,  CJK UNIFIED IDEOGRAPH-516E
        '20849' => '3788'             , # G+3788  U+5171  uni5171,  CJK UNIFIED IDEOGRAPH-5171
        '20853' => '5043'             , # G+5043  U+5175  uni5175,  CJK UNIFIED IDEOGRAPH-5175
        '20854' => '4064'             , # G+4064  U+5176  uni5176,  CJK UNIFIED IDEOGRAPH-5176
        '20855' => '3898'             , # G+3898  U+5177  uni5177,  CJK UNIFIED IDEOGRAPH-5177
        '20856' => '6647'             , # G+6647  U+5178  uni5178,  CJK UNIFIED IDEOGRAPH-5178
        '20860' => '3654'             , # G+3654  U+517C  uni517C,  CJK UNIFIED IDEOGRAPH-517C
        '20864' => '4065'             , # G+4065  U+5180  uni5180,  CJK UNIFIED IDEOGRAPH-5180
        '20870' => '5933'             , # G+5933  U+5186  uni5186,  CJK UNIFIED IDEOGRAPH-5186
        '20874' => '7126'             , # G+7126  U+518A  uni518A,  CJK UNIFIED IDEOGRAPH-518A
        '20877' => '6570'             , # G+6570  U+518D  uni518D,  CJK UNIFIED IDEOGRAPH-518D
        '20882' => '4765'             , # G+4765  U+5192  uni5192,  CJK UNIFIED IDEOGRAPH-5192
        '20885' => '4736'             , # G+4736  U+5195  uni5195,  CJK UNIFIED IDEOGRAPH-5195
        '20887' => '6176'             , # G+6176  U+5197  uni5197,  CJK UNIFIED IDEOGRAPH-5197
        '20896' => '3819'             , # G+3819  U+51A0  uni51A0,  CJK UNIFIED IDEOGRAPH-51A0
        '20901' => '4748'             , # G+4748  U+51A5  uni51A5,  CJK UNIFIED IDEOGRAPH-51A5
        '20906' => '4733'             , # G+4733  U+51AA  uni51AA,  CJK UNIFIED IDEOGRAPH-51AA
        '20908' => '4332'             , # G+4332  U+51AC  uni51AC,  CJK UNIFIED IDEOGRAPH-51AC
        '20918' => '5864'             , # G+5864  U+51B6  uni51B6,  CJK UNIFIED IDEOGRAPH-51B6
        '20919' => '4424'             , # G+4424  U+51B7  uni51B7,  CJK UNIFIED IDEOGRAPH-51B7
        '20925' => '4477'             , # G+4477  U+51BD  uni51BD,  CJK UNIFIED IDEOGRAPH-51BD
        '20932' => '7130'             , # G+7130  U+51C4  uni51C4,  CJK UNIFIED IDEOGRAPH-51C4
        '20934' => '6904'             , # G+6904  U+51C6  uni51C6,  CJK UNIFIED IDEOGRAPH-51C6
        '20937' => '4430'             , # G+4430  U+51C9  uni51C9,  CJK UNIFIED IDEOGRAPH-51C9
        '20939' => '6784'             , # G+6784  U+51CB  uni51CB,  CJK UNIFIED IDEOGRAPH-51CB
        '20940' => '4607'             , # G+4607  U+51CC  uni51CC,  CJK UNIFIED IDEOGRAPH-51CC
        '20941' => '4333'             , # G+4333  U+51CD  uni51CD,  CJK UNIFIED IDEOGRAPH-51CD
        '20956' => '4606'             , # G+4606  U+51DC  uni51DC,  CJK UNIFIED IDEOGRAPH-51DC
        '20957' => '6390'             , # G+6390  U+51DD  uni51DD,  CJK UNIFIED IDEOGRAPH-51DD
        '20958' => '8036'             , # G+8036  U+51DE  uni51DE,  CJK UNIFIED IDEOGRAPH-51DE
        '20961' => '5009'             , # G+5009  U+51E1  uni51E1,  CJK UNIFIED IDEOGRAPH-51E1
        '20976' => '7904'             , # G+7904  U+51F0  uni51F0,  CJK UNIFIED IDEOGRAPH-51F0
        '20977' => '3563'             , # G+3563  U+51F1  uni51F1,  CJK UNIFIED IDEOGRAPH-51F1
        '20982' => '8014'             , # G+8014  U+51F6  uni51F6,  CJK UNIFIED IDEOGRAPH-51F6
        '20984' => '7167'             , # G+7167  U+51F8  uni51F8,  CJK UNIFIED IDEOGRAPH-51F8
        '20985' => '6141'             , # G+6141  U+51F9  uni51F9,  CJK UNIFIED IDEOGRAPH-51F9
        '20986' => '7303'             , # G+7303  U+51FA  uni51FA,  CJK UNIFIED IDEOGRAPH-51FA
        '20989' => '7651'             , # G+7651  U+51FD  uni51FD,  CJK UNIFIED IDEOGRAPH-51FD
        '20992' => '4270'             , # G+4270  U+5200  uni5200,  CJK UNIFIED IDEOGRAPH-5200
        '20995' => '6445'             , # G+6445  U+5203  uni5203,  CJK UNIFIED IDEOGRAPH-5203
        '20998' => '5152'             , # G+5152  U+5206  uni5206,  CJK UNIFIED IDEOGRAPH-5206
        '20999' => '6684'             , # G+6684  U+5207  uni5207,  CJK UNIFIED IDEOGRAPH-5207
        '21000' => '6031'             , # G+6031  U+5208  uni5208,  CJK UNIFIED IDEOGRAPH-5208
        '21002' => '3477'             , # G+3477  U+520A  uni520A,  CJK UNIFIED IDEOGRAPH-520A
        '21006' => '4837'             , # G+4837  U+520E  uni520E,  CJK UNIFIED IDEOGRAPH-520E
        '21009' => '7773'             , # G+7773  U+5211  uni5211,  CJK UNIFIED IDEOGRAPH-5211
        '21015' => '4478'             , # G+4478  U+5217  uni5217,  CJK UNIFIED IDEOGRAPH-5217
        '21021' => '7214'             , # G+7214  U+521D  uni521D,  CJK UNIFIED IDEOGRAPH-521D
        '21028' => '7490'             , # G+7490  U+5224  uni5224,  CJK UNIFIED IDEOGRAPH-5224
        '21029' => '5037'             , # G+5037  U+5225  uni5225,  CJK UNIFIED IDEOGRAPH-5225
        '21033' => '4614'             , # G+4614  U+5229  uni5229,  CJK UNIFIED IDEOGRAPH-5229
        '21034' => '5304'             , # G+5304  U+522A  uni522A,  CJK UNIFIED IDEOGRAPH-522A
        '21038' => '3835'             , # G+3835  U+522E  uni522E,  CJK UNIFIED IDEOGRAPH-522E
        '21040' => '4271'             , # G+4271  U+5230  uni5230,  CJK UNIFIED IDEOGRAPH-5230
        '21046' => '6759'             , # G+6759  U+5236  uni5236,  CJK UNIFIED IDEOGRAPH-5236
        '21047' => '5561'             , # G+5561  U+5237  uni5237,  CJK UNIFIED IDEOGRAPH-5237
        '21048' => '3970'             , # G+3970  U+5238  uni5238,  CJK UNIFIED IDEOGRAPH-5238
        '21049' => '7077'             , # G+7077  U+5239  uni5239,  CJK UNIFIED IDEOGRAPH-5239
        '21050' => '6484'             , # G+6484  U+523A  uni523A,  CJK UNIFIED IDEOGRAPH-523A
        '21051' => '3465'             , # G+3465  U+523B  uni523B,  CJK UNIFIED IDEOGRAPH-523B
        '21059' => '7205'             , # G+7205  U+5243  uni5243,  CJK UNIFIED IDEOGRAPH-5243
        '21063' => '7360'             , # G+7360  U+5247  uni5247,  CJK UNIFIED IDEOGRAPH-5247
        '21066' => '5301'             , # G+5301  U+524A  uni524A,  CJK UNIFIED IDEOGRAPH-524A
        '21067' => '4017'             , # G+4017  U+524B  uni524B,  CJK UNIFIED IDEOGRAPH-524B
        '21068' => '4397'             , # G+4397  U+524C  uni524C,  CJK UNIFIED IDEOGRAPH-524C
        '21069' => '6648'             , # G+6648  U+524D  uni524D,  CJK UNIFIED IDEOGRAPH-524D
        '21076' => '7135'             , # G+7135  U+5254  uni5254,  CJK UNIFIED IDEOGRAPH-5254
        '21078' => '5113'             , # G+5113  U+5256  uni5256,  CJK UNIFIED IDEOGRAPH-5256
        '21083' => '3536'             , # G+3536  U+525B  uni525B,  CJK UNIFIED IDEOGRAPH-525B
        '21085' => '4886'             , # G+4886  U+525D  uni525D,  CJK UNIFIED IDEOGRAPH-525D
        '21089' => '5463'             , # G+5463  U+5261  uni5261,  CJK UNIFIED IDEOGRAPH-5261
        '21097' => '6480'             , # G+6480  U+5269  uni5269,  CJK UNIFIED IDEOGRAPH-5269
        '21098' => '6649'             , # G+6649  U+526A  uni526A,  CJK UNIFIED IDEOGRAPH-526A
        '21103' => '5114'             , # G+5114  U+526F  uni526F,  CJK UNIFIED IDEOGRAPH-526F
        '21106' => '7649'             , # G+7649  U+5272  uni5272,  CJK UNIFIED IDEOGRAPH-5272
        '21109' => '7094'             , # G+7094  U+5275  uni5275,  CJK UNIFIED IDEOGRAPH-5275
        '21117' => '7577'             , # G+7577  U+527D  uni527D,  CJK UNIFIED IDEOGRAPH-527D
        '21119' => '7215'             , # G+7215  U+527F  uni527F,  CJK UNIFIED IDEOGRAPH-527F
        '21123' => '7948'             , # G+7948  U+5283  uni5283,  CJK UNIFIED IDEOGRAPH-5283
        '21127' => '4018'             , # G+4018  U+5287  uni5287,  CJK UNIFIED IDEOGRAPH-5287
        '21128' => '5021'             , # G+5021  U+5288  uni5288,  CJK UNIFIED IDEOGRAPH-5288
        '21129' => '4577'             , # G+4577  U+5289  uni5289,  CJK UNIFIED IDEOGRAPH-5289
        '21133' => '3618'             , # G+3618  U+528D  uni528D,  CJK UNIFIED IDEOGRAPH-528D
        '21137' => '6760'             , # G+6760  U+5291  uni5291,  CJK UNIFIED IDEOGRAPH-5291
        '21138' => '3619'             , # G+3619  U+5292  uni5292,  CJK UNIFIED IDEOGRAPH-5292
        '21147' => '4458'             , # G+4458  U+529B  uni529B,  CJK UNIFIED IDEOGRAPH-529B
        '21151' => '3789'             , # G+3789  U+529F  uni529F,  CJK UNIFIED IDEOGRAPH-529F
        '21152' => '3440'             , # G+3440  U+52A0  uni52A0,  CJK UNIFIED IDEOGRAPH-52A0
        '21155' => '4479'             , # G+4479  U+52A3  uni52A3,  CJK UNIFIED IDEOGRAPH-52A3
        '21156' => '4024'             , # G+4024  U+52A4  uni52A4,  CJK UNIFIED IDEOGRAPH-52A4
        '21161' => '6785'             , # G+6785  U+52A9  uni52A9,  CJK UNIFIED IDEOGRAPH-52A9
        '21162' => '4169'             , # G+4169  U+52AA  uni52AA,  CJK UNIFIED IDEOGRAPH-52AA
        '21163' => '3624'             , # G+3624  U+52AB  uni52AB,  CJK UNIFIED IDEOGRAPH-52AB
        '21182' => '7703'             , # G+7703  U+52BE  uni52BE,  CJK UNIFIED IDEOGRAPH-52BE
        '21185' => '3665'             , # G+3665  U+52C1  uni52C1,  CJK UNIFIED IDEOGRAPH-52C1
        '21187' => '4930'             , # G+4930  U+52C3  uni52C3,  CJK UNIFIED IDEOGRAPH-52C3
        '21189' => '7361'             , # G+7361  U+52C5  uni52C5,  CJK UNIFIED IDEOGRAPH-52C5
        '21191' => '6177'             , # G+6177  U+52C7  uni52C7,  CJK UNIFIED IDEOGRAPH-52C7
        '21193' => '4737'             , # G+4737  U+52C9  uni52C9,  CJK UNIFIED IDEOGRAPH-52C9
        '21197' => '3666'             , # G+3666  U+52CD  uni52CD,  CJK UNIFIED IDEOGRAPH-52CD
        '21202' => '4604'             , # G+4604  U+52D2  uni52D2,  CJK UNIFIED IDEOGRAPH-52D2
        '21205' => '4334'             , # G+4334  U+52D5  uni52D5,  CJK UNIFIED IDEOGRAPH-52D5
        '21206' => '6229'             , # G+6229  U+52D6  uni52D6,  CJK UNIFIED IDEOGRAPH-52D6
        '21208' => '3510'             , # G+3510  U+52D8  uni52D8,  CJK UNIFIED IDEOGRAPH-52D8
        '21209' => '4812'             , # G+4812  U+52D9  uni52D9,  CJK UNIFIED IDEOGRAPH-52D9
        '21211' => '7979'             , # G+7979  U+52DB  uni52DB,  CJK UNIFIED IDEOGRAPH-52DB
        '21213' => '5685'             , # G+5685  U+52DD  uni52DD,  CJK UNIFIED IDEOGRAPH-52DD
        '21214' => '4511'             , # G+4511  U+52DE  uni52DE,  CJK UNIFIED IDEOGRAPH-52DE
        '21215' => '4766'             , # G+4766  U+52DF  uni52DF,  CJK UNIFIED IDEOGRAPH-52DF
        '21218' => '5493'             , # G+5493  U+52E2  uni52E2,  CJK UNIFIED IDEOGRAPH-52E2
        '21219' => '6619'             , # G+6619  U+52E3  uni52E3,  CJK UNIFIED IDEOGRAPH-52E3
        '21220' => '4025'             , # G+4025  U+52E4  uni52E4,  CJK UNIFIED IDEOGRAPH-52E4
        '21235' => '7980'             , # G+7980  U+52F3  uni52F3,  CJK UNIFIED IDEOGRAPH-52F3
        '21237' => '4442'             , # G+4442  U+52F5  uni52F5,  CJK UNIFIED IDEOGRAPH-52F5
        '21240' => '3971'             , # G+3971  U+52F8  uni52F8,  CJK UNIFIED IDEOGRAPH-52F8
        '21242' => '6510'             , # G+6510  U+52FA  uni52FA,  CJK UNIFIED IDEOGRAPH-52FA
        '21243' => '4009'             , # G+4009  U+52FB  uni52FB,  CJK UNIFIED IDEOGRAPH-52FB
        '21246' => '3899'             , # G+3899  U+52FE  uni52FE,  CJK UNIFIED IDEOGRAPH-52FE
        '21247' => '4848'             , # G+4848  U+52FF  uni52FF,  CJK UNIFIED IDEOGRAPH-52FF
        '21253' => '7545'             , # G+7545  U+5305  uni5305,  CJK UNIFIED IDEOGRAPH-5305
        '21256' => '8015'             , # G+8015  U+5308  uni5308,  CJK UNIFIED IDEOGRAPH-5308
        '21261' => '7546'             , # G+7546  U+530D  uni530D,  CJK UNIFIED IDEOGRAPH-530D
        '21263' => '7547'             , # G+7547  U+530F  uni530F,  CJK UNIFIED IDEOGRAPH-530F
        '21264' => '5076'             , # G+5076  U+5310  uni5310,  CJK UNIFIED IDEOGRAPH-5310
        '21269' => '5183'             , # G+5183  U+5315  uni5315,  CJK UNIFIED IDEOGRAPH-5315
        '21270' => '7863'             , # G+7863  U+5316  uni5316,  CJK UNIFIED IDEOGRAPH-5316
        '21271' => '5151'             , # G+5151  U+5317  uni5317,  CJK UNIFIED IDEOGRAPH-5317
        '21273' => '5693'             , # G+5693  U+5319  uni5319,  CJK UNIFIED IDEOGRAPH-5319
        '21280' => '6536'             , # G+6536  U+5320  uni5320,  CJK UNIFIED IDEOGRAPH-5320
        '21281' => '3841'             , # G+3841  U+5321  uni5321,  CJK UNIFIED IDEOGRAPH-5321
        '21283' => '3530'             , # G+3530  U+5323  uni5323,  CJK UNIFIED IDEOGRAPH-5323
        '21290' => '5184'             , # G+5184  U+532A  uni532A,  CJK UNIFIED IDEOGRAPH-532A
        '21295' => '7928'             , # G+7928  U+532F  uni532F,  CJK UNIFIED IDEOGRAPH-532F
        '21305' => '7604'             , # G+7604  U+5339  uni5339,  CJK UNIFIED IDEOGRAPH-5339
        '21311' => '4188'             , # G+4188  U+533F  uni533F,  CJK UNIFIED IDEOGRAPH-533F
        '21312' => '3900'             , # G+3900  U+5340  uni5340,  CJK UNIFIED IDEOGRAPH-5340
        '21313' => '5772'             , # G+5772  U+5341  uni5341,  CJK UNIFIED IDEOGRAPH-5341
        '21315' => '7149'             , # G+7149  U+5343  uni5343,  CJK UNIFIED IDEOGRAPH-5343
        '21316' => '6478'             , # G+6478  U+5344  twentyhangzhou,  CJK UNIFIED IDEOGRAPH-5344
        '21319' => '5686'             , # G+5686  U+5347  uni5347,  CJK UNIFIED IDEOGRAPH-5347
        '21320' => '6053'             , # G+6053  U+5348  uni5348,  CJK UNIFIED IDEOGRAPH-5348
        '21321' => '7994'             , # G+7994  U+5349  uni5349,  CJK UNIFIED IDEOGRAPH-5349
        '21322' => '4906'             , # G+4906  U+534A  uni534A,  CJK UNIFIED IDEOGRAPH-534A
        '21325' => '4671'             , # G+4671  U+534D  uni534D,  CJK UNIFIED IDEOGRAPH-534D
        '21329' => '5185'             , # G+5185  U+5351  uni5351,  CJK UNIFIED IDEOGRAPH-5351
        '21330' => '6834'             , # G+6834  U+5352  uni5352,  CJK UNIFIED IDEOGRAPH-5352
        '21331' => '7395'             , # G+7395  U+5353  uni5353,  CJK UNIFIED IDEOGRAPH-5353
        '21332' => '7760'             , # G+7760  U+5354  uni5354,  CJK UNIFIED IDEOGRAPH-5354
        '21335' => '4145'             , # G+4145  U+5357  uni5357,  CJK UNIFIED IDEOGRAPH-5357
        '21338' => '4887'             , # G+4887  U+535A  uni535A,  CJK UNIFIED IDEOGRAPH-535A
        '21340' => '5077'             , # G+5077  U+535C  uni535C,  CJK UNIFIED IDEOGRAPH-535C
        '21342' => '5031'             , # G+5031  U+535E  uni535E,  CJK UNIFIED IDEOGRAPH-535E
        '21344' => '6692'             , # G+6692  U+5360  uni5360,  CJK UNIFIED IDEOGRAPH-5360
        '21350' => '3852'             , # G+3852  U+5366  uni5366,  CJK UNIFIED IDEOGRAPH-5366
        '21352' => '5450'             , # G+5450  U+5368  uni5368,  CJK UNIFIED IDEOGRAPH-5368
        '21359' => '4800'             , # G+4800  U+536F  uni536F,  CJK UNIFIED IDEOGRAPH-536F
        '21360' => '6446'             , # G+6446  U+5370  uni5370,  CJK UNIFIED IDEOGRAPH-5370
        '21361' => '6287'             , # G+6287  U+5371  uni5371,  CJK UNIFIED IDEOGRAPH-5371
        '21364' => '3466'             , # G+3466  U+5374  uni5374,  CJK UNIFIED IDEOGRAPH-5374
        '21365' => '4390'             , # G+4390  U+5375  uni5375,  CJK UNIFIED IDEOGRAPH-5375
        '21367' => '3972'             , # G+3972  U+5377  uni5377,  CJK UNIFIED IDEOGRAPH-5377
        '21373' => '6926'             , # G+6926  U+537D  uni537D,  CJK UNIFIED IDEOGRAPH-537D
        '21375' => '3667'             , # G+3667  U+537F  uni537F,  CJK UNIFIED IDEOGRAPH-537F
        '21380' => '5851'             , # G+5851  U+5384  uni5384,  CJK UNIFIED IDEOGRAPH-5384
        '21395' => '5840'             , # G+5840  U+5393  uni5393,  CJK UNIFIED IDEOGRAPH-5393
        '21400' => '4615'             , # G+4615  U+5398  uni5398,  CJK UNIFIED IDEOGRAPH-5398
        '21402' => '7968'             , # G+7968  U+539A  uni539A,  CJK UNIFIED IDEOGRAPH-539A
        '21407' => '6256'             , # G+6256  U+539F  uni539F,  CJK UNIFIED IDEOGRAPH-539F
        '21408' => '7332'             , # G+7332  U+53A0  uni53A0,  CJK UNIFIED IDEOGRAPH-53A0
        '21413' => '3979'             , # G+3979  U+53A5  uni53A5,  CJK UNIFIED IDEOGRAPH-53A5
        '21414' => '7618'             , # G+7618  U+53A6  uni53A6,  CJK UNIFIED IDEOGRAPH-53A6
        '21421' => '5989'             , # G+5989  U+53AD  uni53AD,  CJK UNIFIED IDEOGRAPH-53AD
        '21435' => '3586'             , # G+3586  U+53BB  uni53BB,  CJK UNIFIED IDEOGRAPH-53BB
        '21443' => '7083'             , # G+7083  U+53C3  uni53C3,  CJK UNIFIED IDEOGRAPH-53C3
        '21448' => '6201'             , # G+6201  U+53C8  uni53C8,  CJK UNIFIED IDEOGRAPH-53C8
        '21449' => '7044'             , # G+7044  U+53C9  uni53C9,  CJK UNIFIED IDEOGRAPH-53C9
        '21450' => '4052'             , # G+4052  U+53CA  uni53CA,  CJK UNIFIED IDEOGRAPH-53CA
        '21451' => '6202'             , # G+6202  U+53CB  uni53CB,  CJK UNIFIED IDEOGRAPH-53CB
        '21453' => '4907'             , # G+4907  U+53CD  uni53CD,  CJK UNIFIED IDEOGRAPH-53CD
        '21460' => '5628'             , # G+5628  U+53D4  uni53D4,  CJK UNIFIED IDEOGRAPH-53D4
        '21462' => '7316'             , # G+7316  U+53D6  uni53D6,  CJK UNIFIED IDEOGRAPH-53D6
        '21463' => '5568'             , # G+5568  U+53D7  uni53D7,  CJK UNIFIED IDEOGRAPH-53D7
        '21467' => '4908'             , # G+4908  U+53DB  uni53DB,  CJK UNIFIED IDEOGRAPH-53DB
        '21473' => '6032'             , # G+6032  U+53E1  uni53E1,  CJK UNIFIED IDEOGRAPH-53E1
        '21474' => '7251'             , # G+7251  U+53E2  uni53E2,  CJK UNIFIED IDEOGRAPH-53E2
        '21475' => '3901'             , # G+3901  U+53E3  uni53E3,  CJK UNIFIED IDEOGRAPH-53E3
        '21476' => '3729'             , # G+3729  U+53E4  uni53E4,  CJK UNIFIED IDEOGRAPH-53E4
        '21477' => '3902'             , # G+3902  U+53E5  uni53E5,  CJK UNIFIED IDEOGRAPH-53E5
        '21481' => '3730'             , # G+3730  U+53E9  uni53E9,  CJK UNIFIED IDEOGRAPH-53E9
        '21482' => '6943'             , # G+6943  U+53EA  uni53EA,  CJK UNIFIED IDEOGRAPH-53EA
        '21483' => '3994'             , # G+3994  U+53EB  uni53EB,  CJK UNIFIED IDEOGRAPH-53EB
        '21484' => '5500'             , # G+5500  U+53EC  uni53EC,  CJK UNIFIED IDEOGRAPH-53EC
        '21485' => '7500'             , # G+7500  U+53ED  uni53ED,  CJK UNIFIED IDEOGRAPH-53ED
        '21487' => '3441'             , # G+3441  U+53EF  uni53EF,  CJK UNIFIED IDEOGRAPH-53EF
        '21488' => '7432'             , # G+7432  U+53F0  uni53F0,  CJK UNIFIED IDEOGRAPH-53F0
        '21489' => '7016'             , # G+7016  U+53F1  uni53F1,  CJK UNIFIED IDEOGRAPH-53F1
        '21490' => '5250'             , # G+5250  U+53F2  uni53F2,  CJK UNIFIED IDEOGRAPH-53F2
        '21491' => '6203'             , # G+6203  U+53F3  uni53F3,  CJK UNIFIED IDEOGRAPH-53F3
        '21496' => '5251'             , # G+5251  U+53F8  uni53F8,  CJK UNIFIED IDEOGRAPH-53F8
        '21507' => '8023'             , # G+8023  U+5403  uni5403,  CJK UNIFIED IDEOGRAPH-5403
        '21508' => '3467'             , # G+3467  U+5404  uni5404,  CJK UNIFIED IDEOGRAPH-5404
        '21512' => '7663'             , # G+7663  U+5408  uni5408,  CJK UNIFIED IDEOGRAPH-5408
        '21513' => '4128'             , # G+4128  U+5409  uni5409,  CJK UNIFIED IDEOGRAPH-5409
        '21514' => '6620'             , # G+6620  U+540A  uni540A,  CJK UNIFIED IDEOGRAPH-540A
        '21516' => '4335'             , # G+4335  U+540C  uni540C,  CJK UNIFIED IDEOGRAPH-540C
        '21517' => '4749'             , # G+4749  U+540D  uni540D,  CJK UNIFIED IDEOGRAPH-540D
        '21518' => '7969'             , # G+7969  U+540E  uni540E,  CJK UNIFIED IDEOGRAPH-540E
        '21519' => '4616'             , # G+4616  U+540F  uni540F,  CJK UNIFIED IDEOGRAPH-540F
        '21520' => '7450'             , # G+7450  U+5410  uni5410,  CJK UNIFIED IDEOGRAPH-5410
        '21521' => '7711'             , # G+7711  U+5411  uni5411,  CJK UNIFIED IDEOGRAPH-5411
        '21531' => '3953'             , # G+3953  U+541B  uni541B,  CJK UNIFIED IDEOGRAPH-541B
        '21533' => '4638'             , # G+4638  U+541D  uni541D,  CJK UNIFIED IDEOGRAPH-541D
        '21535' => '6381'             , # G+6381  U+541F  uni541F,  CJK UNIFIED IDEOGRAPH-541F
        '21536' => '7534'             , # G+7534  U+5420  uni5420,  CJK UNIFIED IDEOGRAPH-5420
        '21542' => '5115'             , # G+5115  U+5426  uni5426,  CJK UNIFIED IDEOGRAPH-5426
        '21545' => '5153'             , # G+5153  U+5429  uni5429,  CJK UNIFIED IDEOGRAPH-5429
        '21547' => '7652'             , # G+7652  U+542B  uni542B,  CJK UNIFIED IDEOGRAPH-542B
        '21555' => '6055'             , # G+6055  U+5433  uni5433,  CJK UNIFIED IDEOGRAPH-5433
        '21560' => '8030'             , # G+8030  U+5438  uni5438,  CJK UNIFIED IDEOGRAPH-5438
        '21561' => '7317'             , # G+7317  U+5439  uni5439,  CJK UNIFIED IDEOGRAPH-5439
        '21563' => '4838'             , # G+4838  U+543B  uni543B,  CJK UNIFIED IDEOGRAPH-543B
        '21564' => '7970'             , # G+7970  U+543C  uni543C,  CJK UNIFIED IDEOGRAPH-543C
        '21566' => '6054'             , # G+6054  U+543E  uni543E,  CJK UNIFIED IDEOGRAPH-543E
        '21570' => '4443'             , # G+4443  U+5442  uni5442,  CJK UNIFIED IDEOGRAPH-5442
        '21576' => '6709'             , # G+6709  U+5448  uni5448,  CJK UNIFIED IDEOGRAPH-5448
        '21578' => '3731'             , # G+3731  U+544A  uni544A,  CJK UNIFIED IDEOGRAPH-544A
        '21585' => '7408'             , # G+7408  U+5451  uni5451,  CJK UNIFIED IDEOGRAPH-5451
        '21608' => '6867'             , # G+6867  U+5468  uni5468,  CJK UNIFIED IDEOGRAPH-5468
        '21610' => '6866'             , # G+6866  U+546A  uni546A,  CJK UNIFIED IDEOGRAPH-546A
        '21617' => '3732'             , # G+3732  U+5471  uni5471,  CJK UNIFIED IDEOGRAPH-5471
        '21619' => '4851'             , # G+4851  U+5473  uni5473,  CJK UNIFIED IDEOGRAPH-5473
        '21621' => '3442'             , # G+3442  U+5475  uni5475,  CJK UNIFIED IDEOGRAPH-5475
        '21627' => '5738'             , # G+5738  U+547B  uni547B,  CJK UNIFIED IDEOGRAPH-547B
        '21628' => '7802'             , # G+7802  U+547C  uni547C,  CJK UNIFIED IDEOGRAPH-547C
        '21629' => '4750'             , # G+4750  U+547D  uni547D,  CJK UNIFIED IDEOGRAPH-547D
        '21632' => '6594'             , # G+6594  U+5480  uni5480,  CJK UNIFIED IDEOGRAPH-5480
        '21638' => '7548'             , # G+7548  U+5486  uni5486,  CJK UNIFIED IDEOGRAPH-5486
        '21644' => '7864'             , # G+7864  U+548C  uni548C,  CJK UNIFIED IDEOGRAPH-548C
        '21646' => '3903'             , # G+3903  U+548E  uni548E,  CJK UNIFIED IDEOGRAPH-548E
        '21648' => '5116'             , # G+5116  U+5490  uni5490,  CJK UNIFIED IDEOGRAPH-5490
        '21668' => '7381'             , # G+7381  U+54A4  uni54A4,  CJK UNIFIED IDEOGRAPH-54A4
        '21672' => '6485'             , # G+6485  U+54A8  uni54A8,  CJK UNIFIED IDEOGRAPH-54A8
        '21675' => '6944'             , # G+6944  U+54AB  uni54AB,  CJK UNIFIED IDEOGRAPH-54AB
        '21676' => '3870'             , # G+3870  U+54AC  uni54AC,  CJK UNIFIED IDEOGRAPH-54AC
        '21683' => '7687'             , # G+7687  U+54B3  uni54B3,  CJK UNIFIED IDEOGRAPH-54B3
        '21688' => '7653'             , # G+7653  U+54B8  uni54B8,  CJK UNIFIED IDEOGRAPH-54B8
        '21693' => '6447'             , # G+6447  U+54BD  uni54BD,  CJK UNIFIED IDEOGRAPH-54BD
        '21696' => '5841'             , # G+5841  U+54C0  uni54C0,  CJK UNIFIED IDEOGRAPH-54C0
        '21697' => '7590'             , # G+7590  U+54C1  uni54C1,  CJK UNIFIED IDEOGRAPH-54C1
        '21700' => '7853'             , # G+7853  U+54C4  uni54C4,  CJK UNIFIED IDEOGRAPH-54C4
        '21704' => '7664'             , # G+7664  U+54C8  uni54C8,  CJK UNIFIED IDEOGRAPH-54C8
        '21705' => '6571'             , # G+6571  U+54C9  uni54C9,  CJK UNIFIED IDEOGRAPH-54C9
        '21729' => '6257'             , # G+6257  U+54E1  uni54E1,  CJK UNIFIED IDEOGRAPH-54E1
        '21733' => '3443'             , # G+3443  U+54E5  uni54E5,  CJK UNIFIED IDEOGRAPH-54E5
        '21736' => '7216'             , # G+7216  U+54E8  uni54E8,  CJK UNIFIED IDEOGRAPH-54E8
        '21741' => '3767'             , # G+3767  U+54ED  uni54ED,  CJK UNIFIED IDEOGRAPH-54ED
        '21742' => '7953'             , # G+7953  U+54EE  uni54EE,  CJK UNIFIED IDEOGRAPH-54EE
        '21746' => '7168'             , # G+7168  U+54F2  uni54F2,  CJK UNIFIED IDEOGRAPH-54F2
        '21754' => '7549'             , # G+7549  U+54FA  uni54FA,  CJK UNIFIED IDEOGRAPH-54FA
        '21764' => '7503'             , # G+7503  U+5504  uni5504,  CJK UNIFIED IDEOGRAPH-5504
        '21766' => '5252'             , # G+5252  U+5506  uni5506,  CJK UNIFIED IDEOGRAPH-5506
        '21767' => '6980'             , # G+6980  U+5507  uni5507,  CJK UNIFIED IDEOGRAPH-5507
        '21774' => '4617'             , # G+4617  U+550E  uni550E,  CJK UNIFIED IDEOGRAPH-550E
        '21776' => '4239'             , # G+4239  U+5510  uni5510,  CJK UNIFIED IDEOGRAPH-5510
        '21788' => '4689'             , # G+4689  U+551C  uni551C,  CJK UNIFIED IDEOGRAPH-551C
        '21807' => '6313'             , # G+6313  U+552F  uni552F,  CJK UNIFIED IDEOGRAPH-552F
        '21809' => '7095'             , # G+7095  U+5531  uni5531,  CJK UNIFIED IDEOGRAPH-5531
        '21813' => '5821'             , # G+5821  U+5535  uni5535,  CJK UNIFIED IDEOGRAPH-5535
        '21822' => '7382'             , # G+7382  U+553E  uni553E,  CJK UNIFIED IDEOGRAPH-553E
        '21828' => '7396'             , # G+7396  U+5544  uni5544,  CJK UNIFIED IDEOGRAPH-5544
        '21830' => '5335'             , # G+5335  U+5546  uni5546,  CJK UNIFIED IDEOGRAPH-5546
        '21839' => '4839'             , # G+4839  U+554F  uni554F,  CJK UNIFIED IDEOGRAPH-554F
        '21843' => '3706'             , # G+3706  U+5553  uni5553,  CJK UNIFIED IDEOGRAPH-5553
        '21846' => '4217'             , # G+4217  U+5556  uni5556,  CJK UNIFIED IDEOGRAPH-5556
        '21854' => '5778'             , # G+5778  U+555E  uni555E,  CJK UNIFIED IDEOGRAPH-555E
        '21859' => '7654'             , # G+7654  U+5563  uni5563,  CJK UNIFIED IDEOGRAPH-5563
        '21884' => '6761'             , # G+6761  U+557C  uni557C,  CJK UNIFIED IDEOGRAPH-557C
        '21888' => '3579'             , # G+3579  U+5580  uni5580,  CJK UNIFIED IDEOGRAPH-5580
        '21892' => '5421'             , # G+5421  U+5584  uni5584,  CJK UNIFIED IDEOGRAPH-5584
        '21894' => '7169'             , # G+7169  U+5586  uni5586,  CJK UNIFIED IDEOGRAPH-5586
        '21895' => '4374'             , # G+4374  U+5587  uni5587,  CJK UNIFIED IDEOGRAPH-5587
        '21897' => '7971'             , # G+7971  U+5589  uni5589,  CJK UNIFIED IDEOGRAPH-5589
        '21898' => '7655'             , # G+7655  U+558A  uni558A,  CJK UNIFIED IDEOGRAPH-558A
        '21912' => '7150'             , # G+7150  U+5598  uni5598,  CJK UNIFIED IDEOGRAPH-5598
        '21913' => '7995'             , # G+7995  U+5599  uni5599,  CJK UNIFIED IDEOGRAPH-5599
        '21914' => '7883'             , # G+7883  U+559A  uni559A,  CJK UNIFIED IDEOGRAPH-559A
        '21916' => '8037'             , # G+8037  U+559C  uni559C,  CJK UNIFIED IDEOGRAPH-559C
        '21917' => '3501'             , # G+3501  U+559D  uni559D,  CJK UNIFIED IDEOGRAPH-559D
        '21927' => '7990'             , # G+7990  U+55A7  uni55A7,  CJK UNIFIED IDEOGRAPH-55A7
        '21929' => '6314'             , # G+6314  U+55A9  uni55A9,  CJK UNIFIED IDEOGRAPH-55A9
        '21930' => '5336'             , # G+5336  U+55AA  uni55AA,  CJK UNIFIED IDEOGRAPH-55AA
        '21931' => '4132'             , # G+4132  U+55AB  uni55AB,  CJK UNIFIED IDEOGRAPH-55AB
        '21932' => '3871'             , # G+3871  U+55AC  uni55AC,  CJK UNIFIED IDEOGRAPH-55AC
        '21934' => '4195'             , # G+4195  U+55AE  uni55AE,  CJK UNIFIED IDEOGRAPH-55AE
        '21957' => '7972'             , # G+7972  U+55C5  uni55C5,  CJK UNIFIED IDEOGRAPH-55C5
        '21959' => '5365'             , # G+5365  U+55C7  uni55C7,  CJK UNIFIED IDEOGRAPH-55C7
        '21972' => '6981'             , # G+6981  U+55D4  uni55D4,  CJK UNIFIED IDEOGRAPH-55D4
        '21978' => '6056'             , # G+6056  U+55DA  uni55DA,  CJK UNIFIED IDEOGRAPH-55DA
        '21980' => '4066'             , # G+4066  U+55DC  uni55DC,  CJK UNIFIED IDEOGRAPH-55DC
        '21983' => '7045'             , # G+7045  U+55DF  uni55DF,  CJK UNIFIED IDEOGRAPH-55DF
        '21987' => '5253'             , # G+5253  U+55E3  uni55E3,  CJK UNIFIED IDEOGRAPH-55E3
        '21988' => '7338'             , # G+7338  U+55E4  uni55E4,  CJK UNIFIED IDEOGRAPH-55E4
        '22013' => '5569'             , # G+5569  U+55FD  uni55FD,  CJK UNIFIED IDEOGRAPH-55FD
        '22014' => '6868'             , # G+6868  U+55FE  uni55FE,  CJK UNIFIED IDEOGRAPH-55FE
        '22022' => '7409'             , # G+7409  U+5606  uni5606,  CJK UNIFIED IDEOGRAPH-5606
        '22025' => '3444'             , # G+3444  U+5609  uni5609,  CJK UNIFIED IDEOGRAPH-5609
        '22036' => '3904'             , # G+3904  U+5614  uni5614,  CJK UNIFIED IDEOGRAPH-5614
        '22039' => '5337'             , # G+5337  U+5617  uni5617,  CJK UNIFIED IDEOGRAPH-5617
        '22063' => '5501'             , # G+5501  U+562F  uni562F,  CJK UNIFIED IDEOGRAPH-562F
        '22066' => '6786'             , # G+6786  U+5632  uni5632,  CJK UNIFIED IDEOGRAPH-5632
        '22068' => '7318'             , # G+7318  U+5634  uni5634,  CJK UNIFIED IDEOGRAPH-5634
        '22070' => '5694'             , # G+5694  U+5636  uni5636,  CJK UNIFIED IDEOGRAPH-5636
        '22099' => '7719'             , # G+7719  U+5653  uni5653,  CJK UNIFIED IDEOGRAPH-5653
        '22120' => '4067'             , # G+4067  U+5668  uni5668,  CJK UNIFIED IDEOGRAPH-5668
        '22123' => '8038'             , # G+8038  U+566B  uni566B,  CJK UNIFIED IDEOGRAPH-566B
        '22132' => '5154'             , # G+5154  U+5674  uni5674,  CJK UNIFIED IDEOGRAPH-5674
        '22150' => '7954'             , # G+7954  U+5686  uni5686,  CJK UNIFIED IDEOGRAPH-5686
        '22181' => '5956'             , # G+5956  U+56A5  uni56A5,  CJK UNIFIED IDEOGRAPH-56A5
        '22188' => '5223'             , # G+5223  U+56AC  uni56AC,  CJK UNIFIED IDEOGRAPH-56AC
        '22190' => '7712'             , # G+7712  U+56AE  uni56AE,  CJK UNIFIED IDEOGRAPH-56AE
        '22196' => '5927'             , # G+5927  U+56B4  uni56B4,  CJK UNIFIED IDEOGRAPH-56B4
        '22204' => '6511'             , # G+6511  U+56BC  uni56BC,  CJK UNIFIED IDEOGRAPH-56BC
        '22218' => '4152'             , # G+4152  U+56CA  uni56CA,  CJK UNIFIED IDEOGRAPH-56CA
        '22221' => '8039'             , # G+8039  U+56CD  uni56CD,  CJK UNIFIED IDEOGRAPH-56CD
        '22225' => '7242'             , # G+7242  U+56D1  uni56D1,  CJK UNIFIED IDEOGRAPH-56D1
        '22234' => '5570'             , # G+5570  U+56DA  uni56DA,  CJK UNIFIED IDEOGRAPH-56DA
        '22235' => '5254'             , # G+5254  U+56DB  uni56DB,  CJK UNIFIED IDEOGRAPH-56DB
        '22238' => '7929'             , # G+7929  U+56DE  uni56DE,  CJK UNIFIED IDEOGRAPH-56DE
        '22240' => '6448'             , # G+6448  U+56E0  uni56E0,  CJK UNIFIED IDEOGRAPH-56E0
        '22256' => '3774'             , # G+3774  U+56F0  uni56F0,  CJK UNIFIED IDEOGRAPH-56F0
        '22265' => '4491'             , # G+4491  U+56F9  uni56F9,  CJK UNIFIED IDEOGRAPH-56F9
        '22266' => '3733'             , # G+3733  U+56FA  uni56FA,  CJK UNIFIED IDEOGRAPH-56FA
        '22275' => '7550'             , # G+7550  U+5703  uni5703,  CJK UNIFIED IDEOGRAPH-5703
        '22276' => '5902'             , # G+5902  U+5704  uni5704,  CJK UNIFIED IDEOGRAPH-5704
        '22280' => '3973'             , # G+3973  U+5708  uni5708,  CJK UNIFIED IDEOGRAPH-5708
        '22283' => '3947'             , # G+3947  U+570B  uni570B,  CJK UNIFIED IDEOGRAPH-570B
        '22285' => '6288'             , # G+6288  U+570D  uni570D,  CJK UNIFIED IDEOGRAPH-570D
        '22290' => '6259'             , # G+6259  U+5712  uni5712,  CJK UNIFIED IDEOGRAPH-5712
        '22291' => '6258'             , # G+6258  U+5713  uni5713,  CJK UNIFIED IDEOGRAPH-5713
        '22294' => '4272'             , # G+4272  U+5716  uni5716,  CJK UNIFIED IDEOGRAPH-5716
        '22296' => '4196'             , # G+4196  U+5718  uni5718,  CJK UNIFIED IDEOGRAPH-5718
        '22303' => '7451'             , # G+7451  U+571F  uni571F,  CJK UNIFIED IDEOGRAPH-571F
        '22312' => '6572'             , # G+6572  U+5728  uni5728,  CJK UNIFIED IDEOGRAPH-5728
        '22317' => '3995'             , # G+3995  U+572D  uni572D,  CJK UNIFIED IDEOGRAPH-572D
        '22320' => '6945'             , # G+6945  U+5730  uni5730,  CJK UNIFIED IDEOGRAPH-5730
        '22331' => '4068'             , # G+4068  U+573B  uni573B,  CJK UNIFIED IDEOGRAPH-573B
        '22336' => '6946'             , # G+6946  U+5740  uni5740,  CJK UNIFIED IDEOGRAPH-5740
        '22338' => '7491'             , # G+7491  U+5742  uni5742,  CJK UNIFIED IDEOGRAPH-5742
        '22343' => '4010'             , # G+4010  U+5747  uni5747,  CJK UNIFIED IDEOGRAPH-5747
        '22346' => '4943'             , # G+4943  U+574A  uni574A,  CJK UNIFIED IDEOGRAPH-574A
        '22349' => '4218'             , # G+4218  U+574D  uni574D,  CJK UNIFIED IDEOGRAPH-574D
        '22350' => '3511'             , # G+3511  U+574E  uni574E,  CJK UNIFIED IDEOGRAPH-574E
        '22352' => '6855'             , # G+6855  U+5750  uni5750,  CJK UNIFIED IDEOGRAPH-5750
        '22353' => '3581'             , # G+3581  U+5751  uni5751,  CJK UNIFIED IDEOGRAPH-5751
        '22369' => '7474'             , # G+7474  U+5761  uni5761,  CJK UNIFIED IDEOGRAPH-5761
        '22372' => '3775'             , # G+3775  U+5764  uni5764,  CJK UNIFIED IDEOGRAPH-5764
        '22374' => '7410'             , # G+7410  U+5766  uni5766,  CJK UNIFIED IDEOGRAPH-5766
        '22378' => '7529'             , # G+7529  U+576A  uni576A,  CJK UNIFIED IDEOGRAPH-576A
        '22382' => '4252'             , # G+4252  U+576E  uni576E,  CJK UNIFIED IDEOGRAPH-576E
        '22384' => '3668'             , # G+3668  U+5770  uni5770,  CJK UNIFIED IDEOGRAPH-5770
        '22389' => '3905'             , # G+3905  U+5775  uni5775,  CJK UNIFIED IDEOGRAPH-5775
        '22396' => '7397'             , # G+7397  U+577C  uni577C,  CJK UNIFIED IDEOGRAPH-577C
        '22402' => '5571'             , # G+5571  U+5782  uni5782,  CJK UNIFIED IDEOGRAPH-5782
        '22408' => '4251'             , # G+4251  U+5788  uni5788,  CJK UNIFIED IDEOGRAPH-5788
        '22411' => '7774'             , # G+7774  U+578B  uni578B,  CJK UNIFIED IDEOGRAPH-578B
        '22419' => '7688'             , # G+7688  U+5793  uni5793,  CJK UNIFIED IDEOGRAPH-5793
        '22432' => '6373'             , # G+6373  U+57A0  uni57A0,  CJK UNIFIED IDEOGRAPH-57A0
        '22434' => '3906'             , # G+3906  U+57A2  uni57A2,  CJK UNIFIED IDEOGRAPH-57A2
        '22435' => '6260'             , # G+6260  U+57A3  uni57A3,  CJK UNIFIED IDEOGRAPH-57A3
        '22467' => '5842'             , # G+5842  U+57C3  uni57C3,  CJK UNIFIED IDEOGRAPH-57C3
        '22471' => '6178'             , # G+6178  U+57C7  uni57C7,  CJK UNIFIED IDEOGRAPH-57C7
        '22472' => '6905'             , # G+6905  U+57C8  uni57C8,  CJK UNIFIED IDEOGRAPH-57C8
        '22475' => '4708'             , # G+4708  U+57CB  uni57CB,  CJK UNIFIED IDEOGRAPH-57CB
        '22478' => '5474'             , # G+5474  U+57CE  uni57CE,  CJK UNIFIED IDEOGRAPH-57CE
        '22495' => '5948'             , # G+5948  U+57DF  uni57DF,  CJK UNIFIED IDEOGRAPH-57DF
        '22496' => '5117'             , # G+5117  U+57E0  uni57E0,  CJK UNIFIED IDEOGRAPH-57E0
        '22512' => '7115'             , # G+7115  U+57F0  uni57F0,  CJK UNIFIED IDEOGRAPH-57F0
        '22516' => '5720'             , # G+5720  U+57F4  uni57F4,  CJK UNIFIED IDEOGRAPH-57F4
        '22519' => '7032'             , # G+7032  U+57F7  uni57F7,  CJK UNIFIED IDEOGRAPH-57F7
        '22521' => '4971'             , # G+4971  U+57F9  uni57F9,  CJK UNIFIED IDEOGRAPH-57F9
        '22522' => '4069'             , # G+4069  U+57FA  uni57FA,  CJK UNIFIED IDEOGRAPH-57FA
        '22524' => '4070'             , # G+4070  U+57FC  uni57FC,  CJK UNIFIED IDEOGRAPH-57FC
        '22528' => '3959'             , # G+3959  U+5800  uni5800,  CJK UNIFIED IDEOGRAPH-5800
        '22530' => '4240'             , # G+4240  U+5802  uni5802,  CJK UNIFIED IDEOGRAPH-5802
        '22533' => '3637'             , # G+3637  U+5805  uni5805,  CJK UNIFIED IDEOGRAPH-5805
        '22534' => '7459'             , # G+7459  U+5806  uni5806,  CJK UNIFIED IDEOGRAPH-5806
        '22536' => '3537'             , # G+3537  U+5808  uni5808,  CJK UNIFIED IDEOGRAPH-5808
        '22537' => '6355'             , # G+6355  U+5809  uni5809,  CJK UNIFIED IDEOGRAPH-5809
        '22538' => '5793'             , # G+5793  U+580A  uni580A,  CJK UNIFIED IDEOGRAPH-580A
        '22558' => '7187'             , # G+7187  U+581E  uni581E,  CJK UNIFIED IDEOGRAPH-581E
        '22561' => '5059'             , # G+5059  U+5821  uni5821,  CJK UNIFIED IDEOGRAPH-5821
        '22564' => '6762'             , # G+6762  U+5824  uni5824,  CJK UNIFIED IDEOGRAPH-5824
        '22567' => '5957'             , # G+5957  U+5827  uni5827,  CJK UNIFIED IDEOGRAPH-5827
        '22570' => '3512'             , # G+3512  U+582A  uni582A,  CJK UNIFIED IDEOGRAPH-582A
        '22575' => '6142'             , # G+6142  U+582F  uni582F,  CJK UNIFIED IDEOGRAPH-582F
        '22576' => '5918'             , # G+5918  U+5830  uni5830,  CJK UNIFIED IDEOGRAPH-5830
        '22577' => '5060'             , # G+5060  U+5831  uni5831,  CJK UNIFIED IDEOGRAPH-5831
        '22580' => '6537'             , # G+6537  U+5834  uni5834,  CJK UNIFIED IDEOGRAPH-5834
        '22581' => '4273'             , # G+4273  U+5835  uni5835,  CJK UNIFIED IDEOGRAPH-5835
        '22586' => '3707'             , # G+3707  U+583A  uni583A,  CJK UNIFIED IDEOGRAPH-583A
        '22602' => '3857'             , # G+3857  U+584A  uni584A,  CJK UNIFIED IDEOGRAPH-584A
        '22603' => '6002'             , # G+6002  U+584B  uni584B,  CJK UNIFIED IDEOGRAPH-584B
        '22607' => '3564'             , # G+3564  U+584F  uni584F,  CJK UNIFIED IDEOGRAPH-584F
        '22609' => '5502'             , # G+5502  U+5851  uni5851,  CJK UNIFIED IDEOGRAPH-5851
        '22612' => '7424'             , # G+7424  U+5854  uni5854,  CJK UNIFIED IDEOGRAPH-5854
        '22615' => '4274'             , # G+4274  U+5857  uni5857,  CJK UNIFIED IDEOGRAPH-5857
        '22616' => '4241'             , # G+4241  U+5858  uni5858,  CJK UNIFIED IDEOGRAPH-5858
        '22618' => '7252'             , # G+7252  U+585A  uni585A,  CJK UNIFIED IDEOGRAPH-585A
        '22622' => '5362'             , # G+5362  U+585E  uni585E,  CJK UNIFIED IDEOGRAPH-585E
        '22625' => '6650'             , # G+6650  U+5861  uni5861,  CJK UNIFIED IDEOGRAPH-5861
        '22626' => '6057'             , # G+6057  U+5862  uni5862,  CJK UNIFIED IDEOGRAPH-5862
        '22628' => '7981'             , # G+7981  U+5864  uni5864,  CJK UNIFIED IDEOGRAPH-5864
        '22645' => '6982'             , # G+6982  U+5875  uni5875,  CJK UNIFIED IDEOGRAPH-5875
        '22649' => '7084'             , # G+7084  U+5879  uni5879,  CJK UNIFIED IDEOGRAPH-5879
        '22652' => '6651'             , # G+6651  U+587C  uni587C,  CJK UNIFIED IDEOGRAPH-587C
        '22654' => '5629'             , # G+5629  U+587E  uni587E,  CJK UNIFIED IDEOGRAPH-587E
        '22659' => '3669'             , # G+3669  U+5883  uni5883,  CJK UNIFIED IDEOGRAPH-5883
        '22661' => '5373'             , # G+5373  U+5885  uni5885,  CJK UNIFIED IDEOGRAPH-5885
        '22665' => '6179'             , # G+6179  U+5889  uni5889,  CJK UNIFIED IDEOGRAPH-5889
        '22675' => '4801'             , # G+4801  U+5893  uni5893,  CJK UNIFIED IDEOGRAPH-5893
        '22684' => '7265'             , # G+7265  U+589C  uni589C,  CJK UNIFIED IDEOGRAPH-589C
        '22686' => '6931'             , # G+6931  U+589E  uni589E,  CJK UNIFIED IDEOGRAPH-589E
        '22687' => '7720'             , # G+7720  U+589F  uni589F,  CJK UNIFIED IDEOGRAPH-589F
        '22696' => '4834'             , # G+4834  U+58A8  uni58A8,  CJK UNIFIED IDEOGRAPH-58A8
        '22697' => '4319'             , # G+4319  U+58A9  uni58A9,  CJK UNIFIED IDEOGRAPH-58A9
        '22702' => '7383'             , # G+7383  U+58AE  uni58AE,  CJK UNIFIED IDEOGRAPH-58AE
        '22707' => '5155'             , # G+5155  U+58B3  uni58B3,  CJK UNIFIED IDEOGRAPH-58B3
        '22714' => '6058'             , # G+6058  U+58BA  uni58BA,  CJK UNIFIED IDEOGRAPH-58BA
        '22715' => '6538'             , # G+6538  U+58BB  uni58BB,  CJK UNIFIED IDEOGRAPH-58BB
        '22718' => '3478'             , # G+3478  U+58BE  uni58BE,  CJK UNIFIED IDEOGRAPH-58BE
        '22721' => '5022'             , # G+5022  U+58C1  uni58C1,  CJK UNIFIED IDEOGRAPH-58C1
        '22725' => '6090'             , # G+6090  U+58C5  uni58C5,  CJK UNIFIED IDEOGRAPH-58C5
        '22727' => '4197'             , # G+4197  U+58C7  uni58C7,  CJK UNIFIED IDEOGRAPH-58C7
        '22734' => '7982'             , # G+7982  U+58CE  uni58CE,  CJK UNIFIED IDEOGRAPH-58CE
        '22737' => '7630'             , # G+7630  U+58D1  uni58D1,  CJK UNIFIED IDEOGRAPH-58D1
        '22739' => '5829'             , # G+5829  U+58D3  uni58D3,  CJK UNIFIED IDEOGRAPH-58D3
        '22741' => '7803'             , # G+7803  U+58D5  uni58D5,  CJK UNIFIED IDEOGRAPH-58D5
        '22744' => '4564'             , # G+4564  U+58D8  uni58D8,  CJK UNIFIED IDEOGRAPH-58D8
        '22745' => '3842'             , # G+3842  U+58D9  uni58D9,  CJK UNIFIED IDEOGRAPH-58D9
        '22750' => '3858'             , # G+3858  U+58DE  uni58DE,  CJK UNIFIED IDEOGRAPH-58DE
        '22751' => '4536'             , # G+4536  U+58DF  uni58DF,  CJK UNIFIED IDEOGRAPH-58DF
        '22756' => '5880'             , # G+5880  U+58E4  uni58E4,  CJK UNIFIED IDEOGRAPH-58E4
        '22763' => '5255'             , # G+5255  U+58EB  uni58EB,  CJK UNIFIED IDEOGRAPH-58EB
        '22764' => '6470'             , # G+6470  U+58EC  uni58EC,  CJK UNIFIED IDEOGRAPH-58EC
        '22767' => '6539'             , # G+6539  U+58EF  uni58EF,  CJK UNIFIED IDEOGRAPH-58EF
        '22777' => '6463'             , # G+6463  U+58F9  uni58F9,  CJK UNIFIED IDEOGRAPH-58F9
        '22778' => '7804'             , # G+7804  U+58FA  uni58FA,  CJK UNIFIED IDEOGRAPH-58FA
        '22779' => '5374'             , # G+5374  U+58FB  uni58FB,  CJK UNIFIED IDEOGRAPH-58FB
        '22781' => '5572'             , # G+5572  U+58FD  uni58FD,  CJK UNIFIED IDEOGRAPH-58FD
        '22799' => '7619'             , # G+7619  U+590F  uni590F,  CJK UNIFIED IDEOGRAPH-590F
        '22804' => '4071'             , # G+4071  U+5914  uni5914,  CJK UNIFIED IDEOGRAPH-5914
        '22805' => '5403'             , # G+5403  U+5915  uni5915,  CJK UNIFIED IDEOGRAPH-5915
        '22806' => '6135'             , # G+6135  U+5916  uni5916,  CJK UNIFIED IDEOGRAPH-5916
        '22809' => '5630'             , # G+5630  U+5919  uni5919,  CJK UNIFIED IDEOGRAPH-5919
        '22810' => '4190'             , # G+4190  U+591A  uni591A,  CJK UNIFIED IDEOGRAPH-591A
        '22812' => '5865'             , # G+5865  U+591C  uni591C,  CJK UNIFIED IDEOGRAPH-591C
        '22818' => '4797'             , # G+4797  U+5922  uni5922,  CJK UNIFIED IDEOGRAPH-5922
        '22823' => '4253'             , # G+4253  U+5927  uni5927,  CJK UNIFIED IDEOGRAPH-5927
        '22825' => '7151'             , # G+7151  U+5929  uni5929,  CJK UNIFIED IDEOGRAPH-5929
        '22826' => '7433'             , # G+7433  U+592A  uni592A,  CJK UNIFIED IDEOGRAPH-592A
        '22827' => '5118'             , # G+5118  U+592B  uni592B,  CJK UNIFIED IDEOGRAPH-592B
        '22829' => '6143'             , # G+6143  U+592D  uni592D,  CJK UNIFIED IDEOGRAPH-592D
        '22830' => '5834'             , # G+5834  U+592E  uni592E,  CJK UNIFIED IDEOGRAPH-592E
        '22833' => '5758'             , # G+5758  U+5931  uni5931,  CJK UNIFIED IDEOGRAPH-5931
        '22839' => '6416'             , # G+6416  U+5937  uni5937,  CJK UNIFIED IDEOGRAPH-5937
        '22846' => '7761'             , # G+7761  U+593E  uni593E,  CJK UNIFIED IDEOGRAPH-593E
        '22852' => '5928'             , # G+5928  U+5944  uni5944,  CJK UNIFIED IDEOGRAPH-5944
        '22855' => '4072'             , # G+4072  U+5947  uni5947,  CJK UNIFIED IDEOGRAPH-5947
        '22856' => '4156'             , # G+4156  U+5948  uni5948,  CJK UNIFIED IDEOGRAPH-5948
        '22857' => '5094'             , # G+5094  U+5949  uni5949,  CJK UNIFIED IDEOGRAPH-5949
        '22862' => '3996'             , # G+3996  U+594E  uni594E,  CJK UNIFIED IDEOGRAPH-594E
        '22863' => '6869'             , # G+6869  U+594F  uni594F,  CJK UNIFIED IDEOGRAPH-594F
        '22864' => '7884'             , # G+7884  U+5950  uni5950,  CJK UNIFIED IDEOGRAPH-5950
        '22865' => '3708'             , # G+3708  U+5951  uni5951,  CJK UNIFIED IDEOGRAPH-5951
        '22868' => '5156'             , # G+5156  U+5954  uni5954,  CJK UNIFIED IDEOGRAPH-5954
        '22869' => '7730'             , # G+7730  U+5955  uni5955,  CJK UNIFIED IDEOGRAPH-5955
        '22871' => '7466'             , # G+7466  U+5957  uni5957,  CJK UNIFIED IDEOGRAPH-5957
        '22874' => '7689'             , # G+7689  U+595A  uni595A,  CJK UNIFIED IDEOGRAPH-595A
        '22880' => '6652'             , # G+6652  U+5960  uni5960,  CJK UNIFIED IDEOGRAPH-5960
        '22882' => '5256'             , # G+5256  U+5962  uni5962,  CJK UNIFIED IDEOGRAPH-5962
        '22887' => '6059'             , # G+6059  U+5967  uni5967,  CJK UNIFIED IDEOGRAPH-5967
        '22890' => '7418'             , # G+7418  U+596A  uni596A,  CJK UNIFIED IDEOGRAPH-596A
        '22891' => '6360'             , # G+6360  U+596B  uni596B,  CJK UNIFIED IDEOGRAPH-596B
        '22892' => '6540'             , # G+6540  U+596C  uni596C,  CJK UNIFIED IDEOGRAPH-596C
        '22893' => '5404'             , # G+5404  U+596D  uni596D,  CJK UNIFIED IDEOGRAPH-596D
        '22894' => '5157'             , # G+5157  U+596E  uni596E,  CJK UNIFIED IDEOGRAPH-596E
        '22899' => '4159'             , # G+4159  U+5973  uni5973,  CJK UNIFIED IDEOGRAPH-5973
        '22900' => '4170'             , # G+4170  U+5974  uni5974,  CJK UNIFIED IDEOGRAPH-5974
        '22904' => '3479'             , # G+3479  U+5978  uni5978,  CJK UNIFIED IDEOGRAPH-5978
        '22909' => '7805'             , # G+7805  U+597D  uni597D,  CJK UNIFIED IDEOGRAPH-597D
        '22914' => '5936'             , # G+5936  U+5982  uni5982,  CJK UNIFIED IDEOGRAPH-5982
        '22915' => '5186'             , # G+5186  U+5983  uni5983,  CJK UNIFIED IDEOGRAPH-5983
        '22916' => '4697'             , # G+4697  U+5984  uni5984,  CJK UNIFIED IDEOGRAPH-5984
        '22922' => '6471'             , # G+6471  U+598A  uni598A,  CJK UNIFIED IDEOGRAPH-598A
        '22931' => '4073'             , # G+4073  U+5993  uni5993,  CJK UNIFIED IDEOGRAPH-5993
        '22934' => '6144'             , # G+6144  U+5996  uni5996,  CJK UNIFIED IDEOGRAPH-5996
        '22935' => '4039'             , # G+4039  U+5997  uni5997,  CJK UNIFIED IDEOGRAPH-5997
        '22937' => '4802'             , # G+4802  U+5999  uni5999,  CJK UNIFIED IDEOGRAPH-5999
        '22949' => '7384'             , # G+7384  U+59A5  uni59A5,  CJK UNIFIED IDEOGRAPH-59A5
        '22952' => '4944'             , # G+4944  U+59A8  uni59A8,  CJK UNIFIED IDEOGRAPH-59A8
        '22956' => '7467'             , # G+7467  U+59AC  uni59AC,  CJK UNIFIED IDEOGRAPH-59AC
        '22969' => '4709'             , # G+4709  U+59B9  uni59B9,  CJK UNIFIED IDEOGRAPH-59B9
        '22971' => '7131'             , # G+7131  U+59BB  uni59BB,  CJK UNIFIED IDEOGRAPH-59BB
        '22974' => '7188'             , # G+7188  U+59BE  uni59BE,  CJK UNIFIED IDEOGRAPH-59BE
        '22979' => '6710'             , # G+6710  U+59C3  uni59C3,  CJK UNIFIED IDEOGRAPH-59C3
        '22982' => '4767'             , # G+4767  U+59C6  uni59C6,  CJK UNIFIED IDEOGRAPH-59C6
        '22985' => '6486'             , # G+6486  U+59C9  uni59C9,  CJK UNIFIED IDEOGRAPH-59C9
        '22987' => '5695'             , # G+5695  U+59CB  uni59CB,  CJK UNIFIED IDEOGRAPH-59CB
        '22992' => '6595'             , # G+6595  U+59D0  uni59D0,  CJK UNIFIED IDEOGRAPH-59D0
        '22993' => '3734'             , # G+3734  U+59D1  uni59D1,  CJK UNIFIED IDEOGRAPH-59D1
        '22995' => '5475'             , # G+5475  U+59D3  uni59D3,  CJK UNIFIED IDEOGRAPH-59D3
        '22996' => '6289'             , # G+6289  U+59D4  uni59D4,  CJK UNIFIED IDEOGRAPH-59D4
        '23001' => '6472'             , # G+6472  U+59D9  uni59D9,  CJK UNIFIED IDEOGRAPH-59D9
        '23002' => '6145'             , # G+6145  U+59DA  uni59DA,  CJK UNIFIED IDEOGRAPH-59DA
        '23004' => '3538'             , # G+3538  U+59DC  uni59DC,  CJK UNIFIED IDEOGRAPH-59DC
        '23005' => '6864'             , # G+6864  U+59DD  uni59DD,  CJK UNIFIED IDEOGRAPH-59DD
        '23014' => '3480'             , # G+3480  U+59E6  uni59E6,  CJK UNIFIED IDEOGRAPH-59E6
        '23016' => '6417'             , # G+6417  U+59E8  uni59E8,  CJK UNIFIED IDEOGRAPH-59E8
        '23018' => '7017'             , # G+7017  U+59EA  uni59EA,  CJK UNIFIED IDEOGRAPH-59EA
        '23020' => '8040'             , # G+8040  U+59EC  uni59EC,  CJK UNIFIED IDEOGRAPH-59EC
        '23022' => '7672'             , # G+7672  U+59EE  uni59EE,  CJK UNIFIED IDEOGRAPH-59EE
        '23032' => '5958'             , # G+5958  U+59F8  uni59F8,  CJK UNIFIED IDEOGRAPH-59F8
        '23035' => '6449'             , # G+6449  U+59FB  uni59FB,  CJK UNIFIED IDEOGRAPH-59FB
        '23039' => '6487'             , # G+6487  U+59FF  uni59FF,  CJK UNIFIED IDEOGRAPH-59FF
        '23041' => '6290'             , # G+6290  U+5A01  uni5A01,  CJK UNIFIED IDEOGRAPH-5A01
        '23043' => '6132'             , # G+6132  U+5A03  uni5A03,  CJK UNIFIED IDEOGRAPH-5A03
        '23057' => '5257'             , # G+5257  U+5A11  uni5A11,  CJK UNIFIED IDEOGRAPH-5A11
        '23064' => '4153'             , # G+4153  U+5A18  uni5A18,  CJK UNIFIED IDEOGRAPH-5A18
        '23067' => '6060'             , # G+6060  U+5A1B  uni5A1B,  CJK UNIFIED IDEOGRAPH-5A1B
        '23068' => '4134'             , # G+4134  U+5A1C  uni5A1C,  CJK UNIFIED IDEOGRAPH-5A1C
        '23071' => '5959'             , # G+5959  U+5A1F  uni5A1F,  CJK UNIFIED IDEOGRAPH-5A1F
        '23072' => '5739'             , # G+5739  U+5A20  uni5A20,  CJK UNIFIED IDEOGRAPH-5A20
        '23077' => '5779'             , # G+5779  U+5A25  uni5A25,  CJK UNIFIED IDEOGRAPH-5A25
        '23081' => '4672'             , # G+4672  U+5A29  uni5A29,  CJK UNIFIED IDEOGRAPH-5A29
        '23094' => '7319'             , # G+7319  U+5A36  uni5A36,  CJK UNIFIED IDEOGRAPH-5A36
        '23100' => '7096'             , # G+7096  U+5A3C  uni5A3C,  CJK UNIFIED IDEOGRAPH-5A3C
        '23105' => '4565'             , # G+4565  U+5A41  uni5A41,  CJK UNIFIED IDEOGRAPH-5A41
        '23110' => '7475'             , # G+7475  U+5A46  uni5A46,  CJK UNIFIED IDEOGRAPH-5A46
        '23113' => '6107'             , # G+6107  U+5A49  uni5A49,  CJK UNIFIED IDEOGRAPH-5A49
        '23130' => '7844'             , # G+7844  U+5A5A  uni5A5A,  CJK UNIFIED IDEOGRAPH-5A5A
        '23138' => '5187'             , # G+5187  U+5A62  uni5A62,  CJK UNIFIED IDEOGRAPH-5A62
        '23142' => '5119'             , # G+5119  U+5A66  uni5A66,  CJK UNIFIED IDEOGRAPH-5A66
        '23186' => '4710'             , # G+4710  U+5A92  uni5A92,  CJK UNIFIED IDEOGRAPH-5A92
        '23194' => '4852'             , # G+4852  U+5A9A  uni5A9A,  CJK UNIFIED IDEOGRAPH-5A9A
        '23195' => '6261'             , # G+6261  U+5A9B  uni5A9B,  CJK UNIFIED IDEOGRAPH-5A9B
        '23204' => '5696'             , # G+5696  U+5AA4  uni5AA4,  CJK UNIFIED IDEOGRAPH-5AA4
        '23233' => '3445'             , # G+3445  U+5AC1  uni5AC1,  CJK UNIFIED IDEOGRAPH-5AC1
        '23234' => '5573'             , # G+5573  U+5AC2  uni5AC2,  CJK UNIFIED IDEOGRAPH-5AC2
        '23236' => '6262'             , # G+6262  U+5AC4  uni5AC4,  CJK UNIFIED IDEOGRAPH-5AC4
        '23241' => '7018'             , # G+7018  U+5AC9  uni5AC9,  CJK UNIFIED IDEOGRAPH-5AC9
        '23244' => '7758'             , # G+7758  U+5ACC  uni5ACC,  CJK UNIFIED IDEOGRAPH-5ACC
        '23265' => '6621'             , # G+6621  U+5AE1  uni5AE1,  CJK UNIFIED IDEOGRAPH-5AE1
        '23270' => '7673'             , # G+7673  U+5AE6  uni5AE6,  CJK UNIFIED IDEOGRAPH-5AE6
        '23273' => '4181'             , # G+4181  U+5AE9  uni5AE9,  CJK UNIFIED IDEOGRAPH-5AE9
        '23301' => '7865'             , # G+7865  U+5B05  uni5B05,  CJK UNIFIED IDEOGRAPH-5B05
        '23305' => '8041'             , # G+8041  U+5B09  uni5B09,  CJK UNIFIED IDEOGRAPH-5B09
        '23307' => '5422'             , # G+5422  U+5B0B  uni5B0B,  CJK UNIFIED IDEOGRAPH-5B0B
        '23308' => '3872'             , # G+3872  U+5B0C  uni5B0C,  CJK UNIFIED IDEOGRAPH-5B0C
        '23318' => '7535'             , # G+7535  U+5B16  uni5B16,  CJK UNIFIED IDEOGRAPH-5B16
        '23338' => '5224'             , # G+5224  U+5B2A  uni5B2A,  CJK UNIFIED IDEOGRAPH-5B2A
        '23360' => '5338'             , # G+5338  U+5B40  uni5B40,  CJK UNIFIED IDEOGRAPH-5B40
        '23363' => '5881'             , # G+5881  U+5B43  uni5B43,  CJK UNIFIED IDEOGRAPH-5B43
        '23376' => '6488'             , # G+6488  U+5B50  uni5B50,  CJK UNIFIED IDEOGRAPH-5B50
        '23377' => '7754'             , # G+7754  U+5B51  uni5B51,  CJK UNIFIED IDEOGRAPH-5B51
        '23380' => '3790'             , # G+3790  U+5B54  uni5B54,  CJK UNIFIED IDEOGRAPH-5B54
        '23381' => '6481'             , # G+6481  U+5B55  uni5B55,  CJK UNIFIED IDEOGRAPH-5B55
        '23383' => '6489'             , # G+6489  U+5B57  uni5B57,  CJK UNIFIED IDEOGRAPH-5B57
        '23384' => '6832'             , # G+6832  U+5B58  uni5B58,  CJK UNIFIED IDEOGRAPH-5B58
        '23386' => '5120'             , # G+5120  U+5B5A  uni5B5A,  CJK UNIFIED IDEOGRAPH-5B5A
        '23388' => '6490'             , # G+6490  U+5B5C  uni5B5C,  CJK UNIFIED IDEOGRAPH-5B5C
        '23389' => '7955'             , # G+7955  U+5B5D  uni5B5D,  CJK UNIFIED IDEOGRAPH-5B5D
        '23391' => '4727'             , # G+4727  U+5B5F  uni5B5F,  CJK UNIFIED IDEOGRAPH-5B5F
        '23395' => '3709'             , # G+3709  U+5B63  uni5B63,  CJK UNIFIED IDEOGRAPH-5B63
        '23396' => '3735'             , # G+3735  U+5B64  uni5B64,  CJK UNIFIED IDEOGRAPH-5B64
        '23401' => '7690'             , # G+7690  U+5B69  uni5B69,  CJK UNIFIED IDEOGRAPH-5B69
        '23403' => '5546'             , # G+5546  U+5B6B  uni5B6B,  CJK UNIFIED IDEOGRAPH-5B6B
        '23408' => '5631'             , # G+5631  U+5B70  uni5B70,  CJK UNIFIED IDEOGRAPH-5B70
        '23409' => '6522'             , # G+6522  U+5B71  uni5B71,  CJK UNIFIED IDEOGRAPH-5B71
        '23413' => '5121'             , # G+5121  U+5B75  uni5B75,  CJK UNIFIED IDEOGRAPH-5B75
        '23416' => '7631'             , # G+7631  U+5B78  uni5B78,  CJK UNIFIED IDEOGRAPH-5B78
        '23418' => '6315'             , # G+6315  U+5B7A  uni5B7A,  CJK UNIFIED IDEOGRAPH-5B7A
        '23420' => '5923'             , # G+5923  U+5B7C  uni5B7C,  CJK UNIFIED IDEOGRAPH-5B7C
        '23429' => '4266'             , # G+4266  U+5B85  uni5B85,  CJK UNIFIED IDEOGRAPH-5B85
        '23431' => '6204'             , # G+6204  U+5B87  uni5B87,  CJK UNIFIED IDEOGRAPH-5B87
        '23432' => '5574'             , # G+5574  U+5B88  uni5B88,  CJK UNIFIED IDEOGRAPH-5B88
        '23433' => '5807'             , # G+5807  U+5B89  uni5B89,  CJK UNIFIED IDEOGRAPH-5B89
        '23435' => '5553'             , # G+5553  U+5B8B  uni5B8B,  CJK UNIFIED IDEOGRAPH-5B8B
        '23436' => '6108'             , # G+6108  U+5B8C  uni5B8C,  CJK UNIFIED IDEOGRAPH-5B8C
        '23439' => '3864'             , # G+3864  U+5B8F  uni5B8F,  CJK UNIFIED IDEOGRAPH-5B8F
        '23443' => '5078'             , # G+5078  U+5B93  uni5B93,  CJK UNIFIED IDEOGRAPH-5B93
        '23445' => '7427'             , # G+7427  U+5B95  uni5B95,  CJK UNIFIED IDEOGRAPH-5B95
        '23446' => '7950'             , # G+7950  U+5B96  uni5B96,  CJK UNIFIED IDEOGRAPH-5B96
        '23447' => '6838'             , # G+6838  U+5B97  uni5B97,  CJK UNIFIED IDEOGRAPH-5B97
        '23448' => '3820'             , # G+3820  U+5B98  uni5B98,  CJK UNIFIED IDEOGRAPH-5B98
        '23449' => '6870'             , # G+6870  U+5B99  uni5B99,  CJK UNIFIED IDEOGRAPH-5B99
        '23450' => '6711'             , # G+6711  U+5B9A  uni5B9A,  CJK UNIFIED IDEOGRAPH-5B9A
        '23451' => '6109'             , # G+6109  U+5B9B  uni5B9B,  CJK UNIFIED IDEOGRAPH-5B9B
        '23452' => '6397'             , # G+6397  U+5B9C  uni5B9C,  CJK UNIFIED IDEOGRAPH-5B9C
        '23458' => '3580'             , # G+3580  U+5BA2  uni5BA2,  CJK UNIFIED IDEOGRAPH-5BA2
        '23459' => '5423'             , # G+5423  U+5BA3  uni5BA3,  CJK UNIFIED IDEOGRAPH-5BA3
        '23460' => '5759'             , # G+5759  U+5BA4  uni5BA4,  CJK UNIFIED IDEOGRAPH-5BA4
        '23461' => '6316'             , # G+6316  U+5BA5  uni5BA5,  CJK UNIFIED IDEOGRAPH-5BA5
        '23462' => '7885'             , # G+7885  U+5BA6  uni5BA6,  CJK UNIFIED IDEOGRAPH-5BA6
        '23468' => '5476'             , # G+5476  U+5BAC  uni5BAC,  CJK UNIFIED IDEOGRAPH-5BAC
        '23470' => '3963'             , # G+3963  U+5BAE  uni5BAE,  CJK UNIFIED IDEOGRAPH-5BAE
        '23472' => '6573'             , # G+6573  U+5BB0  uni5BB0,  CJK UNIFIED IDEOGRAPH-5BB0
        '23475' => '7691'             , # G+7691  U+5BB3  uni5BB3,  CJK UNIFIED IDEOGRAPH-5BB3
        '23476' => '5960'             , # G+5960  U+5BB4  uni5BB4,  CJK UNIFIED IDEOGRAPH-5BB4
        '23477' => '5503'             , # G+5503  U+5BB5  uni5BB5,  CJK UNIFIED IDEOGRAPH-5BB5
        '23478' => '3446'             , # G+3446  U+5BB6  uni5BB6,  CJK UNIFIED IDEOGRAPH-5BB6
        '23480' => '5740'             , # G+5740  U+5BB8  uni5BB8,  CJK UNIFIED IDEOGRAPH-5BB8
        '23481' => '6180'             , # G+6180  U+5BB9  uni5BB9,  CJK UNIFIED IDEOGRAPH-5BB9
        '23487' => '5632'             , # G+5632  U+5BBF  uni5BBF,  CJK UNIFIED IDEOGRAPH-5BBF
        '23488' => '7116'             , # G+7116  U+5BC0  uni5BC0,  CJK UNIFIED IDEOGRAPH-5BC0
        '23490' => '6622'             , # G+6622  U+5BC2  uni5BC2,  CJK UNIFIED IDEOGRAPH-5BC2
        '23491' => '6263'             , # G+6263  U+5BC3  uni5BC3,  CJK UNIFIED IDEOGRAPH-5BC3
        '23492' => '4074'             , # G+4074  U+5BC4  uni5BC4,  CJK UNIFIED IDEOGRAPH-5BC4
        '23493' => '6450'             , # G+6450  U+5BC5  uni5BC5,  CJK UNIFIED IDEOGRAPH-5BC5
        '23494' => '4883'             , # G+4883  U+5BC6  uni5BC6,  CJK UNIFIED IDEOGRAPH-5BC6
        '23495' => '3907'             , # G+3907  U+5BC7  uni5BC7,  CJK UNIFIED IDEOGRAPH-5BC7
        '23500' => '5122'             , # G+5122  U+5BCC  uni5BCC,  CJK UNIFIED IDEOGRAPH-5BCC
        '23504' => '4711'             , # G+4711  U+5BD0  uni5BD0,  CJK UNIFIED IDEOGRAPH-5BD0
        '23506' => '7635'             , # G+7635  U+5BD2  uni5BD2,  CJK UNIFIED IDEOGRAPH-5BD2
        '23507' => '6205'             , # G+6205  U+5BD3  uni5BD3,  CJK UNIFIED IDEOGRAPH-5BD3
        '23508' => '5721'             , # G+5721  U+5BD4  uni5BD4,  CJK UNIFIED IDEOGRAPH-5BD4
        '23511' => '4168'             , # G+4168  U+5BD7  uni5BD7,  CJK UNIFIED IDEOGRAPH-5BD7
        '23518' => '4664'             , # G+4664  U+5BDE  uni5BDE,  CJK UNIFIED IDEOGRAPH-5BDE
        '23519' => '7078'             , # G+7078  U+5BDF  uni5BDF,  CJK UNIFIED IDEOGRAPH-5BDF
        '23521' => '3803'             , # G+3803  U+5BE1  uni5BE1,  CJK UNIFIED IDEOGRAPH-5BE1
        '23522' => '7368'             , # G+7368  U+5BE2  uni5BE2,  CJK UNIFIED IDEOGRAPH-5BE2
        '23524' => '6061'             , # G+6061  U+5BE4  uni5BE4,  CJK UNIFIED IDEOGRAPH-5BE4
        '23525' => '6146'             , # G+6146  U+5BE5  uni5BE5,  CJK UNIFIED IDEOGRAPH-5BE5
        '23526' => '5760'             , # G+5760  U+5BE6  uni5BE6,  CJK UNIFIED IDEOGRAPH-5BE6
        '23527' => '4167'             , # G+4167  U+5BE7  uni5BE7,  CJK UNIFIED IDEOGRAPH-5BE7
        '23528' => '7117'             , # G+7117  U+5BE8  uni5BE8,  CJK UNIFIED IDEOGRAPH-5BE8
        '23529' => '5762'             , # G+5762  U+5BE9  uni5BE9,  CJK UNIFIED IDEOGRAPH-5BE9
        '23531' => '5258'             , # G+5258  U+5BEB  uni5BEB,  CJK UNIFIED IDEOGRAPH-5BEB
        '23532' => '3821'             , # G+3821  U+5BEC  uni5BEC,  CJK UNIFIED IDEOGRAPH-5BEC
        '23534' => '4553'             , # G+4553  U+5BEE  uni5BEE,  CJK UNIFIED IDEOGRAPH-5BEE
        '23535' => '6906'             , # G+6906  U+5BEF  uni5BEF,  CJK UNIFIED IDEOGRAPH-5BEF
        '23541' => '7253'             , # G+7253  U+5BF5  uni5BF5,  CJK UNIFIED IDEOGRAPH-5BF5
        '23542' => '5061'             , # G+5061  U+5BF6  uni5BF6,  CJK UNIFIED IDEOGRAPH-5BF6
        '23544' => '7247'             , # G+7247  U+5BF8  uni5BF8,  CJK UNIFIED IDEOGRAPH-5BF8
        '23546' => '5259'             , # G+5259  U+5BFA  uni5BFA,  CJK UNIFIED IDEOGRAPH-5BFA
        '23553' => '5095'             , # G+5095  U+5C01  uni5C01,  CJK UNIFIED IDEOGRAPH-5C01
        '23556' => '5260'             , # G+5260  U+5C04  uni5C04,  CJK UNIFIED IDEOGRAPH-5C04
        '23559' => '6541'             , # G+6541  U+5C07  uni5C07,  CJK UNIFIED IDEOGRAPH-5C07
        '23560' => '6653'             , # G+6653  U+5C08  uni5C08,  CJK UNIFIED IDEOGRAPH-5C08
        '23561' => '6291'             , # G+6291  U+5C09  uni5C09,  CJK UNIFIED IDEOGRAPH-5C09
        '23562' => '6833'             , # G+6833  U+5C0A  uni5C0A,  CJK UNIFIED IDEOGRAPH-5C0A
        '23563' => '5763'             , # G+5763  U+5C0B  uni5C0B,  CJK UNIFIED IDEOGRAPH-5C0B
        '23565' => '4254'             , # G+4254  U+5C0D  uni5C0D,  CJK UNIFIED IDEOGRAPH-5C0D
        '23566' => '4275'             , # G+4275  U+5C0E  uni5C0E,  CJK UNIFIED IDEOGRAPH-5C0E
        '23567' => '5504'             , # G+5504  U+5C0F  uni5C0F,  CJK UNIFIED IDEOGRAPH-5C0F
        '23569' => '5505'             , # G+5505  U+5C11  uni5C11,  CJK UNIFIED IDEOGRAPH-5C11
        '23574' => '7178'             , # G+7178  U+5C16  uni5C16,  CJK UNIFIED IDEOGRAPH-5C16
        '23577' => '5339'             , # G+5339  U+5C19  uni5C19,  CJK UNIFIED IDEOGRAPH-5C19
        '23588' => '6206'             , # G+6206  U+5C24  uni5C24,  CJK UNIFIED IDEOGRAPH-5C24
        '23592' => '4945'             , # G+4945  U+5C28  uni5C28,  CJK UNIFIED IDEOGRAPH-5C28
        '23601' => '7320'             , # G+7320  U+5C31  uni5C31,  CJK UNIFIED IDEOGRAPH-5C31
        '23608' => '5697'             , # G+5697  U+5C38  uni5C38,  CJK UNIFIED IDEOGRAPH-5C38
        '23609' => '6361'             , # G+6361  U+5C39  uni5C39,  CJK UNIFIED IDEOGRAPH-5C39
        '23610' => '7136'             , # G+7136  U+5C3A  uni5C3A,  CJK UNIFIED IDEOGRAPH-5C3A
        '23611' => '3736'             , # G+3736  U+5C3B  uni5C3B,  CJK UNIFIED IDEOGRAPH-5C3B
        '23612' => '4186'             , # G+4186  U+5C3C  uni5C3C,  CJK UNIFIED IDEOGRAPH-5C3C
        '23614' => '4853'             , # G+4853  U+5C3E  uni5C3E,  CJK UNIFIED IDEOGRAPH-5C3E
        '23615' => '4180'             , # G+4180  U+5C3F  uni5C3F,  CJK UNIFIED IDEOGRAPH-5C3F
        '23616' => '3948'             , # G+3948  U+5C40  uni5C40,  CJK UNIFIED IDEOGRAPH-5C40
        '23621' => '3587'             , # G+3587  U+5C45  uni5C45,  CJK UNIFIED IDEOGRAPH-5C45
        '23622' => '3710'             , # G+3710  U+5C46  uni5C46,  CJK UNIFIED IDEOGRAPH-5C46
        '23624' => '3960'             , # G+3960  U+5C48  uni5C48,  CJK UNIFIED IDEOGRAPH-5C48
        '23627' => '6078'             , # G+6078  U+5C4B  uni5C4B,  CJK UNIFIED IDEOGRAPH-5C4B
        '23629' => '5699'             , # G+5699  U+5C4D  uni5C4D,  CJK UNIFIED IDEOGRAPH-5C4D
        '23630' => '5698'             , # G+5698  U+5C4E  uni5C4E,  CJK UNIFIED IDEOGRAPH-5C4E
        '23633' => '5451'             , # G+5451  U+5C51  uni5C51,  CJK UNIFIED IDEOGRAPH-5C51
        '23637' => '6654'             , # G+6654  U+5C55  uni5C55,  CJK UNIFIED IDEOGRAPH-5C55
        '23643' => '5044'             , # G+5044  U+5C5B  uni5C5B,  CJK UNIFIED IDEOGRAPH-5C5B
        '23648' => '4276'             , # G+4276  U+5C60  uni5C60,  CJK UNIFIED IDEOGRAPH-5C60
        '23650' => '4566'             , # G+4566  U+5C62  uni5C62,  CJK UNIFIED IDEOGRAPH-5C62
        '23652' => '7335'             , # G+7335  U+5C64  uni5C64,  CJK UNIFIED IDEOGRAPH-5C64
        '23653' => '4618'             , # G+4618  U+5C65  uni5C65,  CJK UNIFIED IDEOGRAPH-5C65
        '23660' => '5538'             , # G+5538  U+5C6C  uni5C6C,  CJK UNIFIED IDEOGRAPH-5C6C
        '23663' => '4358'             , # G+4358  U+5C6F  uni5C6F,  CJK UNIFIED IDEOGRAPH-5C6F
        '23665' => '5305'             , # G+5305  U+5C71  uni5C71,  CJK UNIFIED IDEOGRAPH-5C71
        '23673' => '8024'             , # G+8024  U+5C79  uni5C79,  CJK UNIFIED IDEOGRAPH-5C79
        '23696' => '4075'             , # G+4075  U+5C90  uni5C90,  CJK UNIFIED IDEOGRAPH-5C90
        '23697' => '6527'             , # G+6527  U+5C91  uni5C91,  CJK UNIFIED IDEOGRAPH-5C91
        '23713' => '3539'             , # G+3539  U+5CA1  uni5CA1,  CJK UNIFIED IDEOGRAPH-5CA1
        '23721' => '5822'             , # G+5822  U+5CA9  uni5CA9,  CJK UNIFIED IDEOGRAPH-5CA9
        '23723' => '5575'             , # G+5575  U+5CAB  uni5CAB,  CJK UNIFIED IDEOGRAPH-5CAB
        '23724' => '3531'             , # G+3531  U+5CAC  uni5CAC,  CJK UNIFIED IDEOGRAPH-5CAC
        '23729' => '4255'             , # G+4255  U+5CB1  uni5CB1,  CJK UNIFIED IDEOGRAPH-5CB1
        '23731' => '5794'             , # G+5794  U+5CB3  uni5CB3,  CJK UNIFIED IDEOGRAPH-5CB3
        '23733' => '7806'             , # G+7806  U+5CB5  uni5CB5,  CJK UNIFIED IDEOGRAPH-5CB5
        '23735' => '4870'             , # G+4870  U+5CB7  uni5CB7,  CJK UNIFIED IDEOGRAPH-5CB7
        '23736' => '5808'             , # G+5808  U+5CB8  uni5CB8,  CJK UNIFIED IDEOGRAPH-5CB8
        '23738' => '4492'             , # G+4492  U+5CBA  uni5CBA,  CJK UNIFIED IDEOGRAPH-5CBA
        '23742' => '6693'             , # G+6693  U+5CBE  uni5CBE,  CJK UNIFIED IDEOGRAPH-5CBE
        '23744' => '5576'             , # G+5576  U+5CC0  uni5CC0,  CJK UNIFIED IDEOGRAPH-5CC0
        '23769' => '7339'             , # G+7339  U+5CD9  uni5CD9,  CJK UNIFIED IDEOGRAPH-5CD9
        '23776' => '5340'             , # G+5340  U+5CE0  uni5CE0,  CJK UNIFIED IDEOGRAPH-5CE0
        '23784' => '5780'             , # G+5780  U+5CE8  uni5CE8,  CJK UNIFIED IDEOGRAPH-5CE8
        '23791' => '5096'             , # G+5096  U+5CEF  uni5CEF,  CJK UNIFIED IDEOGRAPH-5CEF
        '23792' => '5097'             , # G+5097  U+5CF0  uni5CF0,  CJK UNIFIED IDEOGRAPH-5CF0
        '23796' => '7735'             , # G+7735  U+5CF4  uni5CF4,  CJK UNIFIED IDEOGRAPH-5CF4
        '23798' => '4277'             , # G+4277  U+5CF6  uni5CF6,  CJK UNIFIED IDEOGRAPH-5CF6
        '23803' => '6907'             , # G+6907  U+5CFB  uni5CFB,  CJK UNIFIED IDEOGRAPH-5CFB
        '23805' => '7762'             , # G+7762  U+5CFD  uni5CFD,  CJK UNIFIED IDEOGRAPH-5CFD
        '23815' => '5671'             , # G+5671  U+5D07  uni5D07,  CJK UNIFIED IDEOGRAPH-5D07
        '23821' => '4421'             , # G+4421  U+5D0D  uni5D0D,  CJK UNIFIED IDEOGRAPH-5D0D
        '23822' => '4076'             , # G+4076  U+5D0E  uni5D0E,  CJK UNIFIED IDEOGRAPH-5D0E
        '23825' => '3776'             , # G+3776  U+5D11  uni5D11,  CJK UNIFIED IDEOGRAPH-5D11
        '23828' => '7263'             , # G+7263  U+5D14  uni5D14,  CJK UNIFIED IDEOGRAPH-5D14
        '23830' => '5843'             , # G+5843  U+5D16  uni5D16,  CJK UNIFIED IDEOGRAPH-5D16
        '23831' => '3540'             , # G+3540  U+5D17  uni5D17,  CJK UNIFIED IDEOGRAPH-5D17
        '23833' => '4596'             , # G+4596  U+5D19  uni5D19,  CJK UNIFIED IDEOGRAPH-5D19
        '23847' => '5672'             , # G+5672  U+5D27  uni5D27,  CJK UNIFIED IDEOGRAPH-5D27
        '23849' => '5175'             , # G+5175  U+5D29  uni5D29,  CJK UNIFIED IDEOGRAPH-5D29
        '23883' => '4854'             , # G+4854  U+5D4B  uni5D4B,  CJK UNIFIED IDEOGRAPH-5D4B
        '23884' => '3513'             , # G+3513  U+5D4C  uni5D4C,  CJK UNIFIED IDEOGRAPH-5D4C
        '23888' => '4399'             , # G+4399  U+5D50  uni5D50,  CJK UNIFIED IDEOGRAPH-5D50
        '23913' => '5673'             , # G+5673  U+5D69  uni5D69,  CJK UNIFIED IDEOGRAPH-5D69
        '23916' => '6136'             , # G+6136  U+5D6C  uni5D6C,  CJK UNIFIED IDEOGRAPH-5D6C
        '23919' => '7046'             , # G+7046  U+5D6F  uni5D6F,  CJK UNIFIED IDEOGRAPH-5D6F
        '23943' => '3908'             , # G+3908  U+5D87  uni5D87,  CJK UNIFIED IDEOGRAPH-5D87
        '23947' => '4278'             , # G+4278  U+5D8B  uni5D8B,  CJK UNIFIED IDEOGRAPH-5D8B
        '23965' => '4365'             , # G+4365  U+5D9D  uni5D9D,  CJK UNIFIED IDEOGRAPH-5D9D
        '23968' => '3873'             , # G+3873  U+5DA0  uni5DA0,  CJK UNIFIED IDEOGRAPH-5DA0
        '23970' => '6147'             , # G+6147  U+5DA2  uni5DA2,  CJK UNIFIED IDEOGRAPH-5DA2
        '23978' => '5931'             , # G+5931  U+5DAA  uni5DAA,  CJK UNIFIED IDEOGRAPH-5DAA
        '23992' => '6003'             , # G+6003  U+5DB8  uni5DB8,  CJK UNIFIED IDEOGRAPH-5DB8
        '23994' => '4493'             , # G+4493  U+5DBA  uni5DBA,  CJK UNIFIED IDEOGRAPH-5DBA
        '23996' => '5375'             , # G+5375  U+5DBC  uni5DBC,  CJK UNIFIED IDEOGRAPH-5DBC
        '23997' => '5795'             , # G+5795  U+5DBD  uni5DBD,  CJK UNIFIED IDEOGRAPH-5DBD
        '24013' => '6137'             , # G+6137  U+5DCD  uni5DCD,  CJK UNIFIED IDEOGRAPH-5DCD
        '24018' => '4673'             , # G+4673  U+5DD2  uni5DD2,  CJK UNIFIED IDEOGRAPH-5DD2
        '24022' => '5823'             , # G+5823  U+5DD6  uni5DD6,  CJK UNIFIED IDEOGRAPH-5DD6
        '24029' => '7152'             , # G+7152  U+5DDD  uni5DDD,  CJK UNIFIED IDEOGRAPH-5DDD
        '24030' => '6871'             , # G+6871  U+5DDE  uni5DDE,  CJK UNIFIED IDEOGRAPH-5DDE
        '24033' => '5640'             , # G+5640  U+5DE1  uni5DE1,  CJK UNIFIED IDEOGRAPH-5DE1
        '24034' => '5506'             , # G+5506  U+5DE2  uni5DE2,  CJK UNIFIED IDEOGRAPH-5DE2
        '24037' => '3791'             , # G+3791  U+5DE5  uni5DE5,  CJK UNIFIED IDEOGRAPH-5DE5
        '24038' => '6856'             , # G+6856  U+5DE6  uni5DE6,  CJK UNIFIED IDEOGRAPH-5DE6
        '24039' => '3874'             , # G+3874  U+5DE7  uni5DE7,  CJK UNIFIED IDEOGRAPH-5DE7
        '24040' => '3588'             , # G+3588  U+5DE8  uni5DE8,  CJK UNIFIED IDEOGRAPH-5DE8
        '24043' => '4813'             , # G+4813  U+5DEB  uni5DEB,  CJK UNIFIED IDEOGRAPH-5DEB
        '24046' => '7047'             , # G+7047  U+5DEE  uni5DEE,  CJK UNIFIED IDEOGRAPH-5DEE
        '24049' => '4077'             , # G+4077  U+5DF1  uni5DF1,  CJK UNIFIED IDEOGRAPH-5DF1
        '24050' => '6418'             , # G+6418  U+5DF2  uni5DF2,  CJK UNIFIED IDEOGRAPH-5DF2
        '24051' => '5261'             , # G+5261  U+5DF3  uni5DF3,  CJK UNIFIED IDEOGRAPH-5DF3
        '24052' => '7476'             , # G+7476  U+5DF4  uni5DF4,  CJK UNIFIED IDEOGRAPH-5DF4
        '24055' => '7674'             , # G+7674  U+5DF7  uni5DF7,  CJK UNIFIED IDEOGRAPH-5DF7
        '24061' => '5547'             , # G+5547  U+5DFD  uni5DFD,  CJK UNIFIED IDEOGRAPH-5DFD
        '24062' => '3604'             , # G+3604  U+5DFE  uni5DFE,  CJK UNIFIED IDEOGRAPH-5DFE
        '24066' => '5700'             , # G+5700  U+5E02  uni5E02,  CJK UNIFIED IDEOGRAPH-5E02
        '24067' => '7551'             , # G+7551  U+5E03  uni5E03,  CJK UNIFIED IDEOGRAPH-5E03
        '24070' => '5010'             , # G+5010  U+5E06  uni5E06,  CJK UNIFIED IDEOGRAPH-5E06
        '24076' => '8042'             , # G+8042  U+5E0C  uni5E0C,  CJK UNIFIED IDEOGRAPH-5E0C
        '24081' => '7428'             , # G+7428  U+5E11  uni5E11,  CJK UNIFIED IDEOGRAPH-5E11
        '24086' => '7189'             , # G+7189  U+5E16  uni5E16,  CJK UNIFIED IDEOGRAPH-5E16
        '24089' => '7019'             , # G+7019  U+5E19  uni5E19,  CJK UNIFIED IDEOGRAPH-5E19
        '24091' => '4990'             , # G+4990  U+5E1B  uni5E1B,  CJK UNIFIED IDEOGRAPH-5E1B
        '24093' => '6763'             , # G+6763  U+5E1D  uni5E1D,  CJK UNIFIED IDEOGRAPH-5E1D
        '24101' => '5577'             , # G+5577  U+5E25  uni5E25,  CJK UNIFIED IDEOGRAPH-5E25
        '24107' => '5262'             , # G+5262  U+5E2B  uni5E2B,  CJK UNIFIED IDEOGRAPH-5E2B
        '24109' => '5405'             , # G+5405  U+5E2D  uni5E2D,  CJK UNIFIED IDEOGRAPH-5E2D
        '24115' => '6542'             , # G+6542  U+5E33  uni5E33,  CJK UNIFIED IDEOGRAPH-5E33
        '24118' => '4256'             , # G+4256  U+5E36  uni5E36,  CJK UNIFIED IDEOGRAPH-5E36
        '24120' => '5341'             , # G+5341  U+5E38  uni5E38,  CJK UNIFIED IDEOGRAPH-5E38
        '24125' => '4768'             , # G+4768  U+5E3D  uni5E3D,  CJK UNIFIED IDEOGRAPH-5E3D
        '24127' => '7973'             , # G+7973  U+5E3F  uni5E3F,  CJK UNIFIED IDEOGRAPH-5E3F
        '24128' => '6712'             , # G+6712  U+5E40  uni5E40,  CJK UNIFIED IDEOGRAPH-5E40
        '24132' => '5796'             , # G+5796  U+5E44  uni5E44,  CJK UNIFIED IDEOGRAPH-5E44
        '24133' => '7571'             , # G+7571  U+5E45  uni5E45,  CJK UNIFIED IDEOGRAPH-5E45
        '24135' => '4946'             , # G+4946  U+5E47  uni5E47,  CJK UNIFIED IDEOGRAPH-5E47
        '24140' => '7905'             , # G+7905  U+5E4C  uni5E4C,  CJK UNIFIED IDEOGRAPH-5E4C
        '24149' => '4665'             , # G+4665  U+5E55  uni5E55,  CJK UNIFIED IDEOGRAPH-5E55
        '24159' => '7340'             , # G+7340  U+5E5F  uni5E5F,  CJK UNIFIED IDEOGRAPH-5E5F
        '24161' => '4996'             , # G+4996  U+5E61  uni5E61,  CJK UNIFIED IDEOGRAPH-5E61
        '24162' => '4242'             , # G+4242  U+5E62  uni5E62,  CJK UNIFIED IDEOGRAPH-5E62
        '24163' => '7536'             , # G+7536  U+5E63  uni5E63,  CJK UNIFIED IDEOGRAPH-5E63
        '24178' => '3481'             , # G+3481  U+5E72  uni5E72,  CJK UNIFIED IDEOGRAPH-5E72
        '24179' => '7530'             , # G+7530  U+5E73  uni5E73,  CJK UNIFIED IDEOGRAPH-5E73
        '24180' => '4160'             , # G+4160  U+5E74  uni5E74,  CJK UNIFIED IDEOGRAPH-5E74
        '24183' => '5045'             , # G+5045  U+5E77  uni5E77,  CJK UNIFIED IDEOGRAPH-5E77
        '24184' => '7706'             , # G+7706  U+5E78  uni5E78,  CJK UNIFIED IDEOGRAPH-5E78
        '24185' => '3482'             , # G+3482  U+5E79  uni5E79,  CJK UNIFIED IDEOGRAPH-5E79
        '24187' => '7886'             , # G+7886  U+5E7B  uni5E7B,  CJK UNIFIED IDEOGRAPH-5E7B
        '24188' => '6317'             , # G+6317  U+5E7C  uni5E7C,  CJK UNIFIED IDEOGRAPH-5E7C
        '24189' => '6318'             , # G+6318  U+5E7D  uni5E7D,  CJK UNIFIED IDEOGRAPH-5E7D
        '24190' => '4078'             , # G+4078  U+5E7E  uni5E7E,  CJK UNIFIED IDEOGRAPH-5E7E
        '24196' => '6543'             , # G+6543  U+5E84  uni5E84,  CJK UNIFIED IDEOGRAPH-5E84
        '24199' => '5188'             , # G+5188  U+5E87  uni5E87,  CJK UNIFIED IDEOGRAPH-5E87
        '24202' => '5342'             , # G+5342  U+5E8A  uni5E8A,  CJK UNIFIED IDEOGRAPH-5E8A
        '24207' => '5376'             , # G+5376  U+5E8F  uni5E8F,  CJK UNIFIED IDEOGRAPH-5E8F
        '24213' => '6596'             , # G+6596  U+5E95  uni5E95,  CJK UNIFIED IDEOGRAPH-5E95
        '24215' => '6694'             , # G+6694  U+5E97  uni5E97,  CJK UNIFIED IDEOGRAPH-5E97
        '24218' => '3670'             , # G+3670  U+5E9A  uni5E9A,  CJK UNIFIED IDEOGRAPH-5E9A
        '24220' => '5123'             , # G+5123  U+5E9C  uni5E9C,  CJK UNIFIED IDEOGRAPH-5E9C
        '24224' => '5343'             , # G+5343  U+5EA0  uni5EA0,  CJK UNIFIED IDEOGRAPH-5EA0
        '24230' => '4279'             , # G+4279  U+5EA6  uni5EA6,  CJK UNIFIED IDEOGRAPH-5EA6
        '24231' => '6857'             , # G+6857  U+5EA7  uni5EA7,  CJK UNIFIED IDEOGRAPH-5EA7
        '24235' => '3737'             , # G+3737  U+5EAB  uni5EAB,  CJK UNIFIED IDEOGRAPH-5EAB
        '24237' => '6713'             , # G+6713  U+5EAD  uni5EAD,  CJK UNIFIED IDEOGRAPH-5EAD
        '24245' => '5824'             , # G+5824  U+5EB5  uni5EB5,  CJK UNIFIED IDEOGRAPH-5EB5
        '24246' => '5377'             , # G+5377  U+5EB6  uni5EB6,  CJK UNIFIED IDEOGRAPH-5EB6
        '24247' => '3541'             , # G+3541  U+5EB7  uni5EB7,  CJK UNIFIED IDEOGRAPH-5EB7
        '24248' => '6181'             , # G+6181  U+5EB8  uni5EB8,  CJK UNIFIED IDEOGRAPH-5EB8
        '24254' => '6319'             , # G+6319  U+5EBE  uni5EBE,  CJK UNIFIED IDEOGRAPH-5EBE
        '24258' => '5344'             , # G+5344  U+5EC2  uni5EC2,  CJK UNIFIED IDEOGRAPH-5EC2
        '24264' => '7620'             , # G+7620  U+5EC8  uni5EC8,  CJK UNIFIED IDEOGRAPH-5EC8
        '24265' => '4483'             , # G+4483  U+5EC9  uni5EC9,  CJK UNIFIED IDEOGRAPH-5EC9
        '24266' => '4412'             , # G+4412  U+5ECA  uni5ECA,  CJK UNIFIED IDEOGRAPH-5ECA
        '24272' => '3909'             , # G+3909  U+5ED0  uni5ED0,  CJK UNIFIED IDEOGRAPH-5ED0
        '24275' => '3815'             , # G+3815  U+5ED3  uni5ED3,  CJK UNIFIED IDEOGRAPH-5ED3
        '24278' => '4554'             , # G+4554  U+5ED6  uni5ED6,  CJK UNIFIED IDEOGRAPH-5ED6
        '24282' => '6872'             , # G+6872  U+5EDA  uni5EDA,  CJK UNIFIED IDEOGRAPH-5EDA
        '24283' => '6655'             , # G+6655  U+5EDB  uni5EDB,  CJK UNIFIED IDEOGRAPH-5EDB
        '24287' => '4803'             , # G+4803  U+5EDF  uni5EDF,  CJK UNIFIED IDEOGRAPH-5EDF
        '24288' => '7097'             , # G+7097  U+5EE0  uni5EE0,  CJK UNIFIED IDEOGRAPH-5EE0
        '24290' => '7537'             , # G+7537  U+5EE2  uni5EE2,  CJK UNIFIED IDEOGRAPH-5EE2
        '24291' => '3843'             , # G+3843  U+5EE3  uni5EE3,  CJK UNIFIED IDEOGRAPH-5EE3
        '24300' => '4444'             , # G+4444  U+5EEC  uni5EEC,  CJK UNIFIED IDEOGRAPH-5EEC
        '24307' => '7197'             , # G+7197  U+5EF3  uni5EF3,  CJK UNIFIED IDEOGRAPH-5EF3
        '24310' => '5961'             , # G+5961  U+5EF6  uni5EF6,  CJK UNIFIED IDEOGRAPH-5EF6
        '24311' => '6714'             , # G+6714  U+5EF7  uni5EF7,  CJK UNIFIED IDEOGRAPH-5EF7
        '24314' => '3605'             , # G+3605  U+5EFA  uni5EFA,  CJK UNIFIED IDEOGRAPH-5EFA
        '24315' => '7930'             , # G+7930  U+5EFB  uni5EFB,  CJK UNIFIED IDEOGRAPH-5EFB
        '24321' => '5032'             , # G+5032  U+5F01  uni5F01,  CJK UNIFIED IDEOGRAPH-5F01
        '24324' => '4537'             , # G+4537  U+5F04  uni5F04,  CJK UNIFIED IDEOGRAPH-5F04
        '24330' => '7538'             , # G+7538  U+5F0A  uni5F0A,  CJK UNIFIED IDEOGRAPH-5F0A
        '24335' => '5722'             , # G+5722  U+5F0F  uni5F0F,  CJK UNIFIED IDEOGRAPH-5F0F
        '24337' => '5701'             , # G+5701  U+5F11  uni5F11,  CJK UNIFIED IDEOGRAPH-5F11
        '24339' => '3964'             , # G+3964  U+5F13  uni5F13,  CJK UNIFIED IDEOGRAPH-5F13
        '24340' => '6787'             , # G+6787  U+5F14  uni5F14,  CJK UNIFIED IDEOGRAPH-5F14
        '24341' => '6451'             , # G+6451  U+5F15  uni5F15,  CJK UNIFIED IDEOGRAPH-5F15
        '24343' => '5172'             , # G+5172  U+5F17  uni5F17,  CJK UNIFIED IDEOGRAPH-5F17
        '24344' => '7854'             , # G+7854  U+5F18  uni5F18,  CJK UNIFIED IDEOGRAPH-5F18
        '24347' => '6419'             , # G+6419  U+5F1B  uni5F1B,  CJK UNIFIED IDEOGRAPH-5F1B
        '24351' => '6764'             , # G+6764  U+5F1F  uni5F1F,  CJK UNIFIED IDEOGRAPH-5F1F
        '24358' => '7736'             , # G+7736  U+5F26  uni5F26,  CJK UNIFIED IDEOGRAPH-5F26
        '24359' => '7807'             , # G+7807  U+5F27  uni5F27,  CJK UNIFIED IDEOGRAPH-5F27
        '24361' => '4171'             , # G+4171  U+5F29  uni5F29,  CJK UNIFIED IDEOGRAPH-5F29
        '24369' => '5872'             , # G+5872  U+5F31  uni5F31,  CJK UNIFIED IDEOGRAPH-5F31
        '24373' => '6544'             , # G+6544  U+5F35  uni5F35,  CJK UNIFIED IDEOGRAPH-5F35
        '24378' => '3542'             , # G+3542  U+5F3A  uni5F3A,  CJK UNIFIED IDEOGRAPH-5F3A
        '24380' => '7605'             , # G+7605  U+5F3C  uni5F3C,  CJK UNIFIED IDEOGRAPH-5F3C
        '24392' => '7411'             , # G+7411  U+5F48  uni5F48,  CJK UNIFIED IDEOGRAPH-5F48
        '24394' => '3543'             , # G+3543  U+5F4A  uni5F4A,  CJK UNIFIED IDEOGRAPH-5F4A
        '24396' => '4855'             , # G+4855  U+5F4C  uni5F4C,  CJK UNIFIED IDEOGRAPH-5F4C
        '24398' => '4674'             , # G+4674  U+5F4E  uni5F4E,  CJK UNIFIED IDEOGRAPH-5F4E
        '24406' => '4198'             , # G+4198  U+5F56  uni5F56,  CJK UNIFIED IDEOGRAPH-5F56
        '24407' => '7792'             , # G+7792  U+5F57  uni5F57,  CJK UNIFIED IDEOGRAPH-5F57
        '24409' => '7997'             , # G+7997  U+5F59  uni5F59,  CJK UNIFIED IDEOGRAPH-5F59
        '24411' => '6420'             , # G+6420  U+5F5B  uni5F5B,  CJK UNIFIED IDEOGRAPH-5F5B
        '24418' => '7775'             , # G+7775  U+5F62  uni5F62,  CJK UNIFIED IDEOGRAPH-5F62
        '24422' => '5919'             , # G+5919  U+5F66  uni5F66,  CJK UNIFIED IDEOGRAPH-5F66
        '24423' => '6230'             , # G+6230  U+5F67  uni5F67,  CJK UNIFIED IDEOGRAPH-5F67
        '24425' => '7118'             , # G+7118  U+5F69  uni5F69,  CJK UNIFIED IDEOGRAPH-5F69
        '24426' => '7578'             , # G+7578  U+5F6A  uni5F6A,  CJK UNIFIED IDEOGRAPH-5F6A
        '24427' => '6788'             , # G+6788  U+5F6B  uni5F6B,  CJK UNIFIED IDEOGRAPH-5F6B
        '24428' => '5225'             , # G+5225  U+5F6C  uni5F6C,  CJK UNIFIED IDEOGRAPH-5F6C
        '24429' => '7513'             , # G+7513  U+5F6D  uni5F6D,  CJK UNIFIED IDEOGRAPH-5F6D
        '24432' => '7098'             , # G+7098  U+5F70  uni5F70,  CJK UNIFIED IDEOGRAPH-5F70
        '24433' => '6004'             , # G+6004  U+5F71  uni5F71,  CJK UNIFIED IDEOGRAPH-5F71
        '24439' => '4947'             , # G+4947  U+5F77  uni5F77,  CJK UNIFIED IDEOGRAPH-5F77
        '24441' => '5949'             , # G+5949  U+5F79  uni5F79,  CJK UNIFIED IDEOGRAPH-5F79
        '24444' => '7597'             , # G+7597  U+5F7C  uni5F7C,  CJK UNIFIED IDEOGRAPH-5F7C
        '24447' => '5173'             , # G+5173  U+5F7F  uni5F7F,  CJK UNIFIED IDEOGRAPH-5F7F
        '24448' => '6126'             , # G+6126  U+5F80  uni5F80,  CJK UNIFIED IDEOGRAPH-5F80
        '24449' => '6715'             , # G+6715  U+5F81  uni5F81,  CJK UNIFIED IDEOGRAPH-5F81
        '24453' => '4257'             , # G+4257  U+5F85  uni5F85,  CJK UNIFIED IDEOGRAPH-5F85
        '24455' => '5641'             , # G+5641  U+5F87  uni5F87,  CJK UNIFIED IDEOGRAPH-5F87
        '24458' => '7931'             , # G+7931  U+5F8A  uni5F8A,  CJK UNIFIED IDEOGRAPH-5F8A
        '24459' => '4600'             , # G+4600  U+5F8B  uni5F8B,  CJK UNIFIED IDEOGRAPH-5F8B
        '24460' => '7974'             , # G+7974  U+5F8C  uni5F8C,  CJK UNIFIED IDEOGRAPH-5F8C
        '24464' => '5378'             , # G+5378  U+5F90  uni5F90,  CJK UNIFIED IDEOGRAPH-5F90
        '24465' => '3671'             , # G+3671  U+5F91  uni5F91,  CJK UNIFIED IDEOGRAPH-5F91
        '24466' => '4280'             , # G+4280  U+5F92  uni5F92,  CJK UNIFIED IDEOGRAPH-5F92
        '24471' => '4364'             , # G+4364  U+5F97  uni5F97,  CJK UNIFIED IDEOGRAPH-5F97
        '24472' => '4972'             , # G+4972  U+5F98  uni5F98,  CJK UNIFIED IDEOGRAPH-5F98
        '24473' => '5263'             , # G+5263  U+5F99  uni5F99,  CJK UNIFIED IDEOGRAPH-5F99
        '24478' => '6839'             , # G+6839  U+5F9E  uni5F9E,  CJK UNIFIED IDEOGRAPH-5F9E
        '24480' => '4422'             , # G+4422  U+5FA0  uni5FA0,  CJK UNIFIED IDEOGRAPH-5FA0
        '24481' => '5903'             , # G+5903  U+5FA1  uni5FA1,  CJK UNIFIED IDEOGRAPH-5FA1
        '24488' => '7906'             , # G+7906  U+5FA8  uni5FA8,  CJK UNIFIED IDEOGRAPH-5FA8
        '24489' => '5079'             , # G+5079  U+5FA9  uni5FA9,  CJK UNIFIED IDEOGRAPH-5FA9
        '24490' => '5642'             , # G+5642  U+5FAA  uni5FAA,  CJK UNIFIED IDEOGRAPH-5FAA
        '24494' => '4856'             , # G+4856  U+5FAE  uni5FAE,  CJK UNIFIED IDEOGRAPH-5FAE
        '24501' => '7038'             , # G+7038  U+5FB5  uni5FB5,  CJK UNIFIED IDEOGRAPH-5FB5
        '24503' => '4267'             , # G+4267  U+5FB7  uni5FB7,  CJK UNIFIED IDEOGRAPH-5FB7
        '24505' => '7170'             , # G+7170  U+5FB9  uni5FB9,  CJK UNIFIED IDEOGRAPH-5FB9
        '24509' => '7998'             , # G+7998  U+5FBD  uni5FBD,  CJK UNIFIED IDEOGRAPH-5FBD
        '24515' => '5764'             , # G+5764  U+5FC3  uni5FC3,  CJK UNIFIED IDEOGRAPH-5FC3
        '24517' => '7606'             , # G+7606  U+5FC5  uni5FC5,  CJK UNIFIED IDEOGRAPH-5FC5
        '24524' => '4079'             , # G+4079  U+5FCC  uni5FCC,  CJK UNIFIED IDEOGRAPH-5FCC
        '24525' => '6452'             , # G+6452  U+5FCD  uni5FCD,  CJK UNIFIED IDEOGRAPH-5FCD
        '24534' => '7248'             , # G+7248  U+5FD6  uni5FD6,  CJK UNIFIED IDEOGRAPH-5FD6
        '24535' => '6947'             , # G+6947  U+5FD7  uni5FD7,  CJK UNIFIED IDEOGRAPH-5FD7
        '24536' => '4698'             , # G+4698  U+5FD8  uni5FD8,  CJK UNIFIED IDEOGRAPH-5FD8
        '24537' => '4699'             , # G+4699  U+5FD9  uni5FD9,  CJK UNIFIED IDEOGRAPH-5FD9
        '24544' => '7307'             , # G+7307  U+5FE0  uni5FE0,  CJK UNIFIED IDEOGRAPH-5FE0
        '24555' => '7379'             , # G+7379  U+5FEB  uni5FEB,  CJK UNIFIED IDEOGRAPH-5FEB
        '24565' => '4163'             , # G+4163  U+5FF5  uni5FF5,  CJK UNIFIED IDEOGRAPH-5FF5
        '24573' => '7850'             , # G+7850  U+5FFD  uni5FFD,  CJK UNIFIED IDEOGRAPH-5FFD
        '24575' => '5158'             , # G+5158  U+5FFF  uni5FFF,  CJK UNIFIED IDEOGRAPH-5FFF
        '24591' => '5835'             , # G+5835  U+600F  uni600F,  CJK UNIFIED IDEOGRAPH-600F
        '24594' => '4172'             , # G+4172  U+6012  uni6012,  CJK UNIFIED IDEOGRAPH-6012
        '24598' => '7552'             , # G+7552  U+6016  uni6016,  CJK UNIFIED IDEOGRAPH-6016
        '24604' => '4494'             , # G+4494  U+601C  uni601C,  CJK UNIFIED IDEOGRAPH-601C
        '24605' => '5264'             , # G+5264  U+601D  uni601D,  CJK UNIFIED IDEOGRAPH-601D
        '24608' => '7434'             , # G+7434  U+6020  uni6020,  CJK UNIFIED IDEOGRAPH-6020
        '24609' => '6421'             , # G+6421  U+6021  uni6021,  CJK UNIFIED IDEOGRAPH-6021
        '24613' => '4053'             , # G+4053  U+6025  uni6025,  CJK UNIFIED IDEOGRAPH-6025
        '24615' => '5477'             , # G+5477  U+6027  uni6027,  CJK UNIFIED IDEOGRAPH-6027
        '24616' => '6264'             , # G+6264  U+6028  uni6028,  CJK UNIFIED IDEOGRAPH-6028
        '24618' => '3859'             , # G+3859  U+602A  uni602A,  CJK UNIFIED IDEOGRAPH-602A
        '24623' => '3625'             , # G+3625  U+602F  uni602F,  CJK UNIFIED IDEOGRAPH-602F
        '24641' => '6473'             , # G+6473  U+6041  uni6041,  CJK UNIFIED IDEOGRAPH-6041
        '24642' => '5643'             , # G+5643  U+6042  uni6042,  CJK UNIFIED IDEOGRAPH-6042
        '24643' => '5702'             , # G+5702  U+6043  uni6043,  CJK UNIFIED IDEOGRAPH-6043
        '24653' => '7907'             , # G+7907  U+604D  uni604D,  CJK UNIFIED IDEOGRAPH-604D
        '24656' => '3792'             , # G+3792  U+6050  uni6050,  CJK UNIFIED IDEOGRAPH-6050
        '24658' => '7675'             , # G+7675  U+6052  uni6052,  CJK UNIFIED IDEOGRAPH-6052
        '24661' => '5379'             , # G+5379  U+6055  uni6055,  CJK UNIFIED IDEOGRAPH-6055
        '24665' => '5882'             , # G+5882  U+6059  uni6059,  CJK UNIFIED IDEOGRAPH-6059
        '24669' => '3836'             , # G+3836  U+605D  uni605D,  CJK UNIFIED IDEOGRAPH-605D
        '24674' => '7932'             , # G+7932  U+6062  uni6062,  CJK UNIFIED IDEOGRAPH-6062
        '24675' => '6491'             , # G+6491  U+6063  uni6063,  CJK UNIFIED IDEOGRAPH-6063
        '24676' => '8010'             , # G+8010  U+6064  uni6064,  CJK UNIFIED IDEOGRAPH-6064
        '24677' => '7341'             , # G+7341  U+6065  uni6065,  CJK UNIFIED IDEOGRAPH-6065
        '24680' => '7636'             , # G+7636  U+6068  uni6068,  CJK UNIFIED IDEOGRAPH-6068
        '24681' => '6374'             , # G+6374  U+6069  uni6069,  CJK UNIFIED IDEOGRAPH-6069
        '24682' => '3468'             , # G+3468  U+606A  uni606A,  CJK UNIFIED IDEOGRAPH-606A
        '24684' => '4164'             , # G+4164  U+606C  uni606C,  CJK UNIFIED IDEOGRAPH-606C
        '24685' => '3793'             , # G+3793  U+606D  uni606D,  CJK UNIFIED IDEOGRAPH-606D
        '24687' => '5723'             , # G+5723  U+606F  uni606F,  CJK UNIFIED IDEOGRAPH-606F
        '24688' => '8031'             , # G+8031  U+6070  uni6070,  CJK UNIFIED IDEOGRAPH-6070
        '24709' => '5985'             , # G+5985  U+6085  uni6085,  CJK UNIFIED IDEOGRAPH-6085
        '24713' => '5761'             , # G+5761  U+6089  uni6089,  CJK UNIFIED IDEOGRAPH-6089
        '24716' => '6765'             , # G+6765  U+608C  uni608C,  CJK UNIFIED IDEOGRAPH-608C
        '24717' => '7637'             , # G+7637  U+608D  uni608D,  CJK UNIFIED IDEOGRAPH-608D
        '24724' => '7933'             , # G+7933  U+6094  uni6094,  CJK UNIFIED IDEOGRAPH-6094
        '24726' => '7504'             , # G+7504  U+6096  uni6096,  CJK UNIFIED IDEOGRAPH-6096
        '24730' => '5554'             , # G+5554  U+609A  uni609A,  CJK UNIFIED IDEOGRAPH-609A
        '24731' => '6656'             , # G+6656  U+609B  uni609B,  CJK UNIFIED IDEOGRAPH-609B
        '24735' => '6062'             , # G+6062  U+609F  uni609F,  CJK UNIFIED IDEOGRAPH-609F
        '24736' => '6320'             , # G+6320  U+60A0  uni60A0,  CJK UNIFIED IDEOGRAPH-60A0
        '24739' => '7887'             , # G+7887  U+60A3  uni60A3,  CJK UNIFIED IDEOGRAPH-60A3
        '24740' => '7254'             , # G+7254  U+60A4  uni60A4,  CJK UNIFIED IDEOGRAPH-60A4
        '24743' => '4619'             , # G+4619  U+60A7  uni60A7,  CJK UNIFIED IDEOGRAPH-60A7
        '24752' => '6840'             , # G+6840  U+60B0  uni60B0,  CJK UNIFIED IDEOGRAPH-60B0
        '24754' => '5189'             , # G+5189  U+60B2  uni60B2,  CJK UNIFIED IDEOGRAPH-60B2
        '24755' => '4268'             , # G+4268  U+60B3  uni60B3,  CJK UNIFIED IDEOGRAPH-60B3
        '24756' => '7312'             , # G+7312  U+60B4  uni60B4,  CJK UNIFIED IDEOGRAPH-60B4
        '24758' => '4871'             , # G+4871  U+60B6  uni60B6,  CJK UNIFIED IDEOGRAPH-60B6
        '24760' => '3711'             , # G+3711  U+60B8  uni60B8,  CJK UNIFIED IDEOGRAPH-60B8
        '24764' => '4281'             , # G+4281  U+60BC  uni60BC,  CJK UNIFIED IDEOGRAPH-60BC
        '24765' => '7132'             , # G+7132  U+60BD  uni60BD,  CJK UNIFIED IDEOGRAPH-60BD
        '24773' => '6716'             , # G+6716  U+60C5  uni60C5,  CJK UNIFIED IDEOGRAPH-60C5
        '24775' => '4320'             , # G+4320  U+60C7  uni60C7,  CJK UNIFIED IDEOGRAPH-60C7
        '24785' => '7841'             , # G+7841  U+60D1  uni60D1,  CJK UNIFIED IDEOGRAPH-60D1
        '24794' => '7851'             , # G+7851  U+60DA  uni60DA,  CJK UNIFIED IDEOGRAPH-60DA
        '24796' => '5406'             , # G+5406  U+60DC  uni60DC,  CJK UNIFIED IDEOGRAPH-60DC
        '24799' => '6321'             , # G+6321  U+60DF  uni60DF,  CJK UNIFIED IDEOGRAPH-60DF
        '24800' => '7793'             , # G+7793  U+60E0  uni60E0,  CJK UNIFIED IDEOGRAPH-60E0
        '24801' => '5797'             , # G+5797  U+60E1  uni60E1,  CJK UNIFIED IDEOGRAPH-60E1
        '24816' => '7385'             , # G+7385  U+60F0  uni60F0,  CJK UNIFIED IDEOGRAPH-60F0
        '24817' => '4178'             , # G+4178  U+60F1  uni60F1,  CJK UNIFIED IDEOGRAPH-60F1
        '24819' => '5345'             , # G+5345  U+60F3  uni60F3,  CJK UNIFIED IDEOGRAPH-60F3
        '24822' => '7908'             , # G+7908  U+60F6  uni60F6,  CJK UNIFIED IDEOGRAPH-60F6
        '24825' => '5866'             , # G+5866  U+60F9  uni60F9,  CJK UNIFIED IDEOGRAPH-60F9
        '24826' => '5478'             , # G+5478  U+60FA  uni60FA,  CJK UNIFIED IDEOGRAPH-60FA
        '24827' => '7333'             , # G+7333  U+60FB  uni60FB,  CJK UNIFIED IDEOGRAPH-60FB
        '24833' => '5578'             , # G+5578  U+6101  uni6101,  CJK UNIFIED IDEOGRAPH-6101
        '24838' => '3606'             , # G+3606  U+6106  uni6106,  CJK UNIFIED IDEOGRAPH-6106
        '24840' => '6322'             , # G+6322  U+6108  uni6108,  CJK UNIFIED IDEOGRAPH-6108
        '24841' => '6323'             , # G+6323  U+6109  uni6109,  CJK UNIFIED IDEOGRAPH-6109
        '24845' => '4872'             , # G+4872  U+610D  uni610D,  CJK UNIFIED IDEOGRAPH-610D
        '24846' => '7517'             , # G+7517  U+610E  uni610E,  CJK UNIFIED IDEOGRAPH-610E
        '24847' => '6398'             , # G+6398  U+610F  uni610F,  CJK UNIFIED IDEOGRAPH-610F
        '24853' => '5798'             , # G+5798  U+6115  uni6115,  CJK UNIFIED IDEOGRAPH-6115
        '24858' => '6207'             , # G+6207  U+611A  uni611A,  CJK UNIFIED IDEOGRAPH-611A
        '24859' => '5844'             , # G+5844  U+611B  uni611B,  CJK UNIFIED IDEOGRAPH-611B
        '24863' => '3514'             , # G+3514  U+611F  uni611F,  CJK UNIFIED IDEOGRAPH-611F
        '24871' => '3860'             , # G+3860  U+6127  uni6127,  CJK UNIFIED IDEOGRAPH-6127
        '24880' => '7909'             , # G+7909  U+6130  uni6130,  CJK UNIFIED IDEOGRAPH-6130
        '24884' => '7099'             , # G+7099  U+6134  uni6134,  CJK UNIFIED IDEOGRAPH-6134
        '24887' => '3565'             , # G+3565  U+6137  uni6137,  CJK UNIFIED IDEOGRAPH-6137
        '24892' => '5741'             , # G+5741  U+613C  uni613C,  CJK UNIFIED IDEOGRAPH-613C
        '24894' => '3566'             , # G+3566  U+613E  uni613E,  CJK UNIFIED IDEOGRAPH-613E
        '24895' => '6265'             , # G+6265  U+613F  uni613F,  CJK UNIFIED IDEOGRAPH-613F
        '24898' => '6182'             , # G+6182  U+6142  uni6142,  CJK UNIFIED IDEOGRAPH-6142
        '24900' => '4601'             , # G+4601  U+6144  uni6144,  CJK UNIFIED IDEOGRAPH-6144
        '24903' => '6375'             , # G+6375  U+6147  uni6147,  CJK UNIFIED IDEOGRAPH-6147
        '24904' => '6492'             , # G+6492  U+6148  uni6148,  CJK UNIFIED IDEOGRAPH-6148
        '24906' => '3655'             , # G+3655  U+614A  uni614A,  CJK UNIFIED IDEOGRAPH-614A
        '24907' => '7435'             , # G+7435  U+614B  uni614B,  CJK UNIFIED IDEOGRAPH-614B
        '24908' => '7910'             , # G+7910  U+614C  uni614C,  CJK UNIFIED IDEOGRAPH-614C
        '24915' => '7579'             , # G+7579  U+6153  uni6153,  CJK UNIFIED IDEOGRAPH-6153
        '24917' => '4769'             , # G+4769  U+6155  uni6155,  CJK UNIFIED IDEOGRAPH-6155
        '24920' => '7085'             , # G+7085  U+6158  uni6158,  CJK UNIFIED IDEOGRAPH-6158
        '24921' => '7086'             , # G+7086  U+6159  uni6159,  CJK UNIFIED IDEOGRAPH-6159
        '24925' => '7471'             , # G+7471  U+615D  uni615D,  CJK UNIFIED IDEOGRAPH-615D
        '24927' => '7453'             , # G+7453  U+615F  uni615F,  CJK UNIFIED IDEOGRAPH-615F
        '24930' => '4675'             , # G+4675  U+6162  uni6162,  CJK UNIFIED IDEOGRAPH-6162
        '24931' => '3822'             , # G+3822  U+6163  uni6163,  CJK UNIFIED IDEOGRAPH-6163
        '24932' => '3469'             , # G+3469  U+6164  uni6164,  CJK UNIFIED IDEOGRAPH-6164
        '24935' => '7794'             , # G+7794  U+6167  uni6167,  CJK UNIFIED IDEOGRAPH-6167
        '24936' => '3567'             , # G+3567  U+6168  uni6168,  CJK UNIFIED IDEOGRAPH-6168
        '24939' => '6841'             , # G+6841  U+616B  uni616B,  CJK UNIFIED IDEOGRAPH-616B
        '24942' => '4445'             , # G+4445  U+616E  uni616E,  CJK UNIFIED IDEOGRAPH-616E
        '24944' => '6292'             , # G+6292  U+6170  uni6170,  CJK UNIFIED IDEOGRAPH-6170
        '24950' => '3672'             , # G+3672  U+6176  uni6176,  CJK UNIFIED IDEOGRAPH-6176
        '24951' => '3544'             , # G+3544  U+6177  uni6177,  CJK UNIFIED IDEOGRAPH-6177
        '24957' => '7137'             , # G+7137  U+617D  uni617D,  CJK UNIFIED IDEOGRAPH-617D
        '24958' => '6168'             , # G+6168  U+617E  uni617E,  CJK UNIFIED IDEOGRAPH-617E
        '24961' => '7255'             , # G+7255  U+6181  uni6181,  CJK UNIFIED IDEOGRAPH-6181
        '24962' => '6208'             , # G+6208  U+6182  uni6182,  CJK UNIFIED IDEOGRAPH-6182
        '24970' => '5190'             , # G+5190  U+618A  uni618A,  CJK UNIFIED IDEOGRAPH-618A
        '24974' => '6932'             , # G+6932  U+618E  uni618E,  CJK UNIFIED IDEOGRAPH-618E
        '24976' => '4465'             , # G+4465  U+6190  uni6190,  CJK UNIFIED IDEOGRAPH-6190
        '24977' => '5237'             , # G+5237  U+6191  uni6191,  CJK UNIFIED IDEOGRAPH-6191
        '24980' => '7217'             , # G+7217  U+6194  uni6194,  CJK UNIFIED IDEOGRAPH-6194
        '24984' => '8044'             , # G+8044  U+6198  uni6198,  CJK UNIFIED IDEOGRAPH-6198
        '24985' => '8043'             , # G+8043  U+6199  uni6199,  CJK UNIFIED IDEOGRAPH-6199
        '24986' => '7412'             , # G+7412  U+619A  uni619A,  CJK UNIFIED IDEOGRAPH-619A
        '24996' => '5159'             , # G+5159  U+61A4  uni61A4,  CJK UNIFIED IDEOGRAPH-61A4
        '24999' => '4336'             , # G+4336  U+61A7  uni61A7,  CJK UNIFIED IDEOGRAPH-61A7
        '25001' => '3628'             , # G+3628  U+61A9  uni61A9,  CJK UNIFIED IDEOGRAPH-61A9
        '25003' => '4873'             , # G+4873  U+61AB  uni61AB,  CJK UNIFIED IDEOGRAPH-61AB
        '25004' => '3673'             , # G+3673  U+61AC  uni61AC,  CJK UNIFIED IDEOGRAPH-61AC
        '25006' => '4814'             , # G+4814  U+61AE  uni61AE,  CJK UNIFIED IDEOGRAPH-61AE
        '25010' => '7723'             , # G+7723  U+61B2  uni61B2,  CJK UNIFIED IDEOGRAPH-61B2
        '25014' => '5913'             , # G+5913  U+61B6  uni61B6,  CJK UNIFIED IDEOGRAPH-61B6
        '25018' => '4219'             , # G+4219  U+61BA  uni61BA,  CJK UNIFIED IDEOGRAPH-61BA
        '25022' => '3515'             , # G+3515  U+61BE  uni61BE,  CJK UNIFIED IDEOGRAPH-61BE
        '25027' => '4026'             , # G+4026  U+61C3  uni61C3,  CJK UNIFIED IDEOGRAPH-61C3
        '25031' => '3483'             , # G+3483  U+61C7  uni61C7,  CJK UNIFIED IDEOGRAPH-61C7
        '25032' => '7692'             , # G+7692  U+61C8  uni61C8,  CJK UNIFIED IDEOGRAPH-61C8
        '25033' => '6391'             , # G+6391  U+61C9  uni61C9,  CJK UNIFIED IDEOGRAPH-61C9
        '25034' => '6063'             , # G+6063  U+61CA  uni61CA,  CJK UNIFIED IDEOGRAPH-61CA
        '25035' => '4815'             , # G+4815  U+61CB  uni61CB,  CJK UNIFIED IDEOGRAPH-61CB
        '25062' => '4135'             , # G+4135  U+61E6  uni61E6,  CJK UNIFIED IDEOGRAPH-61E6
        '25074' => '7039'             , # G+7039  U+61F2  uni61F2,  CJK UNIFIED IDEOGRAPH-61F2
        '25078' => '4375'             , # G+4375  U+61F6  uni61F6,  CJK UNIFIED IDEOGRAPH-61F6
        '25079' => '7934'             , # G+7934  U+61F7  uni61F7,  CJK UNIFIED IDEOGRAPH-61F7
        '25080' => '7737'             , # G+7737  U+61F8  uni61F8,  CJK UNIFIED IDEOGRAPH-61F8
        '25082' => '7087'             , # G+7087  U+61FA  uni61FA,  CJK UNIFIED IDEOGRAPH-61FA
        '25084' => '3910'             , # G+3910  U+61FC  uni61FC,  CJK UNIFIED IDEOGRAPH-61FC
        '25087' => '6399'             , # G+6399  U+61FF  uni61FF,  CJK UNIFIED IDEOGRAPH-61FF
        '25088' => '4466'             , # G+4466  U+6200  uni6200,  CJK UNIFIED IDEOGRAPH-6200
        '25095' => '4243'             , # G+4243  U+6207  uni6207,  CJK UNIFIED IDEOGRAPH-6207
        '25096' => '3804'             , # G+3804  U+6208  uni6208,  CJK UNIFIED IDEOGRAPH-6208
        '25098' => '4816'             , # G+4816  U+620A  uni620A,  CJK UNIFIED IDEOGRAPH-620A
        '25100' => '5667'             , # G+5667  U+620C  uni620C,  CJK UNIFIED IDEOGRAPH-620C
        '25101' => '5579'             , # G+5579  U+620D  uni620D,  CJK UNIFIED IDEOGRAPH-620D
        '25102' => '6369'             , # G+6369  U+620E  uni620E,  CJK UNIFIED IDEOGRAPH-620E
        '25104' => '5479'             , # G+5479  U+6210  uni6210,  CJK UNIFIED IDEOGRAPH-6210
        '25105' => '5781'             , # G+5781  U+6211  uni6211,  CJK UNIFIED IDEOGRAPH-6211
        '25106' => '3712'             , # G+3712  U+6212  uni6212,  CJK UNIFIED IDEOGRAPH-6212
        '25110' => '7842'             , # G+7842  U+6216  uni6216,  CJK UNIFIED IDEOGRAPH-6216
        '25114' => '7138'             , # G+7138  U+621A  uni621A,  CJK UNIFIED IDEOGRAPH-621A
        '25119' => '4019'             , # G+4019  U+621F  uni621F,  CJK UNIFIED IDEOGRAPH-621F
        '25121' => '3516'             , # G+3516  U+6221  uni6221,  CJK UNIFIED IDEOGRAPH-6221
        '25130' => '6685'             , # G+6685  U+622A  uni622A,  CJK UNIFIED IDEOGRAPH-622A
        '25134' => '4592'             , # G+4592  U+622E  uni622E,  CJK UNIFIED IDEOGRAPH-622E
        '25136' => '6657'             , # G+6657  U+6230  uni6230,  CJK UNIFIED IDEOGRAPH-6230
        '25137' => '8045'             , # G+8045  U+6231  uni6231,  CJK UNIFIED IDEOGRAPH-6231
        '25140' => '4258'             , # G+4258  U+6234  uni6234,  CJK UNIFIED IDEOGRAPH-6234
        '25142' => '7808'             , # G+7808  U+6236  uni6236,  CJK UNIFIED IDEOGRAPH-6236
        '25150' => '4446'             , # G+4446  U+623E  uni623E,  CJK UNIFIED IDEOGRAPH-623E
        '25151' => '4948'             , # G+4948  U+623F  uni623F,  CJK UNIFIED IDEOGRAPH-623F
        '25152' => '5507'             , # G+5507  U+6240  uni6240,  CJK UNIFIED IDEOGRAPH-6240
        '25153' => '7520'             , # G+7520  U+6241  uni6241,  CJK UNIFIED IDEOGRAPH-6241
        '25159' => '5424'             , # G+5424  U+6247  uni6247,  CJK UNIFIED IDEOGRAPH-6247
        '25160' => '7809'             , # G+7809  U+6248  uni6248,  CJK UNIFIED IDEOGRAPH-6248
        '25161' => '5191'             , # G+5191  U+6249  uni6249,  CJK UNIFIED IDEOGRAPH-6249
        '25163' => '5580'             , # G+5580  U+624B  uni624B,  CJK UNIFIED IDEOGRAPH-624B
        '25165' => '6574'             , # G+6574  U+624D  uni624D,  CJK UNIFIED IDEOGRAPH-624D
        '25171' => '7386'             , # G+7386  U+6253  uni6253,  CJK UNIFIED IDEOGRAPH-6253
        '25176' => '7398'             , # G+7398  U+6258  uni6258,  CJK UNIFIED IDEOGRAPH-6258
        '25198' => '5160'             , # G+5160  U+626E  uni626E,  CJK UNIFIED IDEOGRAPH-626E
        '25201' => '4054'             , # G+4054  U+6271  uni6271,  CJK UNIFIED IDEOGRAPH-6271
        '25206' => '5124'             , # G+5124  U+6276  uni6276,  CJK UNIFIED IDEOGRAPH-6276
        '25209' => '5192'             , # G+5192  U+6279  uni6279,  CJK UNIFIED IDEOGRAPH-6279
        '25212' => '5852'             , # G+5852  U+627C  uni627C,  CJK UNIFIED IDEOGRAPH-627C
        '25215' => '5687'             , # G+5687  U+627F  uni627F,  CJK UNIFIED IDEOGRAPH-627F
        '25216' => '4080'             , # G+4080  U+6280  uni6280,  CJK UNIFIED IDEOGRAPH-6280
        '25220' => '7218'             , # G+7218  U+6284  uni6284,  CJK UNIFIED IDEOGRAPH-6284
        '25225' => '3648'             , # G+3648  U+6289  uni6289,  CJK UNIFIED IDEOGRAPH-6289
        '25226' => '7477'             , # G+7477  U+628A  uni628A,  CJK UNIFIED IDEOGRAPH-628A
        '25233' => '5914'             , # G+5914  U+6291  uni6291,  CJK UNIFIED IDEOGRAPH-6291
        '25234' => '5380'             , # G+5380  U+6292  uni6292,  CJK UNIFIED IDEOGRAPH-6292
        '25237' => '7468'             , # G+7468  U+6295  uni6295,  CJK UNIFIED IDEOGRAPH-6295
        '25239' => '7676'             , # G+7676  U+6297  uni6297,  CJK UNIFIED IDEOGRAPH-6297
        '25240' => '6686'             , # G+6686  U+6298  uni6298,  CJK UNIFIED IDEOGRAPH-6298
        '25243' => '7553'             , # G+7553  U+629B  uni629B,  CJK UNIFIED IDEOGRAPH-629B
        '25259' => '7598'             , # G+7598  U+62AB  uni62AB,  CJK UNIFIED IDEOGRAPH-62AB
        '25265' => '7554'             , # G+7554  U+62B1  uni62B1,  CJK UNIFIED IDEOGRAPH-62B1
        '25269' => '6597'             , # G+6597  U+62B5  uni62B5,  CJK UNIFIED IDEOGRAPH-62B5
        '25273' => '4690'             , # G+4690  U+62B9  uni62B9,  CJK UNIFIED IDEOGRAPH-62B9
        '25276' => '5830'             , # G+5830  U+62BC  uni62BC,  CJK UNIFIED IDEOGRAPH-62BC
        '25277' => '7266'             , # G+7266  U+62BD  uni62BD,  CJK UNIFIED IDEOGRAPH-62BD
        '25282' => '5174'             , # G+5174  U+62C2  uni62C2,  CJK UNIFIED IDEOGRAPH-62C2
        '25287' => '4817'             , # G+4817  U+62C7  uni62C7,  CJK UNIFIED IDEOGRAPH-62C7
        '25288' => '4165'             , # G+4165  U+62C8  uni62C8,  CJK UNIFIED IDEOGRAPH-62C8
        '25289' => '4409'             , # G+4409  U+62C9  uni62C9,  CJK UNIFIED IDEOGRAPH-62C9
        '25292' => '4909'             , # G+4909  U+62CC  uni62CC,  CJK UNIFIED IDEOGRAPH-62CC
        '25293' => '4888'             , # G+4888  U+62CD  uni62CD,  CJK UNIFIED IDEOGRAPH-62CD
        '25295' => '4136'             , # G+4136  U+62CF  uni62CF,  CJK UNIFIED IDEOGRAPH-62CF
        '25296' => '3861'             , # G+3861  U+62D0  uni62D0,  CJK UNIFIED IDEOGRAPH-62D0
        '25298' => '3589'             , # G+3589  U+62D2  uni62D2,  CJK UNIFIED IDEOGRAPH-62D2
        '25299' => '7139'             , # G+7139  U+62D3  uni62D3,  CJK UNIFIED IDEOGRAPH-62D3
        '25300' => '4931'             , # G+4931  U+62D4  uni62D4,  CJK UNIFIED IDEOGRAPH-62D4
        '25302' => '7387'             , # G+7387  U+62D6  uni62D6,  CJK UNIFIED IDEOGRAPH-62D6
        '25303' => '6148'             , # G+6148  U+62D7  uni62D7,  CJK UNIFIED IDEOGRAPH-62D7
        '25304' => '3911'             , # G+3911  U+62D8  uni62D8,  CJK UNIFIED IDEOGRAPH-62D8
        '25305' => '6835'             , # G+6835  U+62D9  uni62D9,  CJK UNIFIED IDEOGRAPH-62D9
        '25307' => '7219'             , # G+7219  U+62DB  uni62DB,  CJK UNIFIED IDEOGRAPH-62DB
        '25308' => '4973'             , # G+4973  U+62DC  uni62DC,  CJK UNIFIED IDEOGRAPH-62DC
        '25324' => '3837'             , # G+3837  U+62EC  uni62EC,  CJK UNIFIED IDEOGRAPH-62EC
        '25325' => '5724'             , # G+5724  U+62ED  uni62ED,  CJK UNIFIED IDEOGRAPH-62ED
        '25326' => '4129'             , # G+4129  U+62EE  uni62EE,  CJK UNIFIED IDEOGRAPH-62EE
        '25327' => '6934'             , # G+6934  U+62EF  uni62EF,  CJK UNIFIED IDEOGRAPH-62EF
        '25329' => '3794'             , # G+3794  U+62F1  uni62F1,  CJK UNIFIED IDEOGRAPH-62F1
        '25331' => '3974'             , # G+3974  U+62F3  uni62F3,  CJK UNIFIED IDEOGRAPH-62F3
        '25335' => '3738'             , # G+3738  U+62F7  uni62F7,  CJK UNIFIED IDEOGRAPH-62F7
        '25342' => '5678'             , # G+5678  U+62FE  uni62FE,  CJK UNIFIED IDEOGRAPH-62FE
        '25343' => '4137'             , # G+4137  U+62FF  uni62FF,  CJK UNIFIED IDEOGRAPH-62FF
        '25345' => '6948'             , # G+6948  U+6301  uni6301,  CJK UNIFIED IDEOGRAPH-6301
        '25351' => '6949'             , # G+6949  U+6307  uni6307,  CJK UNIFIED IDEOGRAPH-6307
        '25353' => '5809'             , # G+5809  U+6309  uni6309,  CJK UNIFIED IDEOGRAPH-6309
        '25361' => '4282'             , # G+4282  U+6311  uni6311,  CJK UNIFIED IDEOGRAPH-6311
        '25387' => '6858'             , # G+6858  U+632B  uni632B,  CJK UNIFIED IDEOGRAPH-632B
        '25391' => '6983'             , # G+6983  U+632F  uni632F,  CJK UNIFIED IDEOGRAPH-632F
        '25402' => '6717'             , # G+6717  U+633A  uni633A,  CJK UNIFIED IDEOGRAPH-633A
        '25403' => '5963'             , # G+5963  U+633B  uni633B,  CJK UNIFIED IDEOGRAPH-633B
        '25405' => '4676'             , # G+4676  U+633D  uni633D,  CJK UNIFIED IDEOGRAPH-633D
        '25406' => '7763'             , # G+7763  U+633E  uni633E,  CJK UNIFIED IDEOGRAPH-633E
        '25417' => '7055'             , # G+7055  U+6349  uni6349,  CJK UNIFIED IDEOGRAPH-6349
        '25420' => '7501'             , # G+7501  U+634C  uni634C,  CJK UNIFIED IDEOGRAPH-634C
        '25423' => '4143'             , # G+4143  U+634F  uni634F,  CJK UNIFIED IDEOGRAPH-634F
        '25424' => '5962'             , # G+5962  U+6350  uni6350,  CJK UNIFIED IDEOGRAPH-6350
        '25429' => '7555'             , # G+7555  U+6355  uni6355,  CJK UNIFIED IDEOGRAPH-6355
        '25447' => '5098'             , # G+5098  U+6367  uni6367,  CJK UNIFIED IDEOGRAPH-6367
        '25448' => '5265'             , # G+5265  U+6368  uni6368,  CJK UNIFIED IDEOGRAPH-6368
        '25454' => '3590'             , # G+3590  U+636E  uni636E,  CJK UNIFIED IDEOGRAPH-636E
        '25458' => '3975'             , # G+3975  U+6372  uni6372,  CJK UNIFIED IDEOGRAPH-6372
        '25463' => '7190'             , # G+7190  U+6377  uni6377,  CJK UNIFIED IDEOGRAPH-6377
        '25466' => '4144'             , # G+4144  U+637A  uni637A,  CJK UNIFIED IDEOGRAPH-637A
        '25467' => '4166'             , # G+4166  U+637B  uni637B,  CJK UNIFIED IDEOGRAPH-637B
        '25471' => '5381'             , # G+5381  U+637F  uni637F,  CJK UNIFIED IDEOGRAPH-637F
        '25475' => '5508'             , # G+5508  U+6383  uni6383,  CJK UNIFIED IDEOGRAPH-6383
        '25480' => '5581'             , # G+5581  U+6388  uni6388,  CJK UNIFIED IDEOGRAPH-6388
        '25481' => '4283'             , # G+4283  U+6389  uni6389,  CJK UNIFIED IDEOGRAPH-6389
        '25484' => '6545'             , # G+6545  U+638C  uni638C,  CJK UNIFIED IDEOGRAPH-638C
        '25490' => '4974'             , # G+4974  U+6392  uni6392,  CJK UNIFIED IDEOGRAPH-6392
        '25494' => '5853'             , # G+5853  U+6396  uni6396,  CJK UNIFIED IDEOGRAPH-6396
        '25496' => '3961'             , # G+3961  U+6398  uni6398,  CJK UNIFIED IDEOGRAPH-6398
        '25499' => '3853'             , # G+3853  U+639B  uni639B,  CJK UNIFIED IDEOGRAPH-639B
        '25504' => '4425'             , # G+4425  U+63A0  uni63A0,  CJK UNIFIED IDEOGRAPH-63A0
        '25505' => '7119'             , # G+7119  U+63A1  uni63A1,  CJK UNIFIED IDEOGRAPH-63A1
        '25506' => '7420'             , # G+7420  U+63A2  uni63A2,  CJK UNIFIED IDEOGRAPH-63A2
        '25509' => '6701'             , # G+6701  U+63A5  uni63A5,  CJK UNIFIED IDEOGRAPH-63A5
        '25511' => '3795'             , # G+3795  U+63A7  uni63A7,  CJK UNIFIED IDEOGRAPH-63A7
        '25512' => '7267'             , # G+7267  U+63A8  uni63A8,  CJK UNIFIED IDEOGRAPH-63A8
        '25513' => '5929'             , # G+5929  U+63A9  uni63A9,  CJK UNIFIED IDEOGRAPH-63A9
        '25514' => '6789'             , # G+6789  U+63AA  uni63AA,  CJK UNIFIED IDEOGRAPH-63AA
        '25536' => '3484'             , # G+3484  U+63C0  uni63C0,  CJK UNIFIED IDEOGRAPH-63C0
        '25540' => '6324'             , # G+6324  U+63C4  uni63C4,  CJK UNIFIED IDEOGRAPH-63C4
        '25542' => '3997'             , # G+3997  U+63C6  uni63C6,  CJK UNIFIED IDEOGRAPH-63C6
        '25551' => '4804'             , # G+4804  U+63CF  uni63CF,  CJK UNIFIED IDEOGRAPH-63CF
        '25552' => '6766'             , # G+6766  U+63D0  uni63D0,  CJK UNIFIED IDEOGRAPH-63D0
        '25558' => '6387'             , # G+6387  U+63D6  uni63D6,  CJK UNIFIED IDEOGRAPH-63D6
        '25562' => '5883'             , # G+5883  U+63DA  uni63DA,  CJK UNIFIED IDEOGRAPH-63DA
        '25563' => '7888'             , # G+7888  U+63DB  uni63DB,  CJK UNIFIED IDEOGRAPH-63DB
        '25569' => '5799'             , # G+5799  U+63E1  uni63E1,  CJK UNIFIED IDEOGRAPH-63E1
        '25581' => '3629'             , # G+3629  U+63ED  uni63ED,  CJK UNIFIED IDEOGRAPH-63ED
        '25582' => '7999'             , # G+7999  U+63EE  uni63EE,  CJK UNIFIED IDEOGRAPH-63EE
        '25588' => '6266'             , # G+6266  U+63F4  uni63F4,  CJK UNIFIED IDEOGRAPH-63F4
        '25590' => '5867'             , # G+5867  U+63F6  uni63F6,  CJK UNIFIED IDEOGRAPH-63F6
        '25591' => '5327'             , # G+5327  U+63F7  uni63F7,  CJK UNIFIED IDEOGRAPH-63F7
        '25613' => '5548'             , # G+5548  U+640D  uni640D,  CJK UNIFIED IDEOGRAPH-640D
        '25615' => '4889'             , # G+4889  U+640F  uni640F,  CJK UNIFIED IDEOGRAPH-640F
        '25620' => '5509'             , # G+5509  U+6414  uni6414,  CJK UNIFIED IDEOGRAPH-6414
        '25622' => '6149'             , # G+6149  U+6416  uni6416,  CJK UNIFIED IDEOGRAPH-6416
        '25623' => '4284'             , # G+4284  U+6417  uni6417,  CJK UNIFIED IDEOGRAPH-6417
        '25628' => '5582'             , # G+5582  U+641C  uni641C,  CJK UNIFIED IDEOGRAPH-641C
        '25634' => '6984'             , # G+6984  U+6422  uni6422,  CJK UNIFIED IDEOGRAPH-6422
        '25644' => '4910'             , # G+4910  U+642C  uni642C,  CJK UNIFIED IDEOGRAPH-642C
        '25645' => '7425'             , # G+7425  U+642D  uni642D,  CJK UNIFIED IDEOGRAPH-642D
        '25658' => '8006'             , # G+8006  U+643A  uni643A,  CJK UNIFIED IDEOGRAPH-643A
        '25662' => '7056'             , # G+7056  U+643E  uni643E,  CJK UNIFIED IDEOGRAPH-643E
        '25688' => '6623'             , # G+6623  U+6458  uni6458,  CJK UNIFIED IDEOGRAPH-6458
        '25696' => '7256'             , # G+7256  U+6460  uni6460,  CJK UNIFIED IDEOGRAPH-6460
        '25705' => '4656'             , # G+4656  U+6469  uni6469,  CJK UNIFIED IDEOGRAPH-6469
        '25711' => '6950'             , # G+6950  U+646F  uni646F,  CJK UNIFIED IDEOGRAPH-646F
        '25720' => '4770'             , # G+4770  U+6478  uni6478,  CJK UNIFIED IDEOGRAPH-6478
        '25721' => '4771'             , # G+4771  U+6479  uni6479,  CJK UNIFIED IDEOGRAPH-6479
        '25722' => '6702'             , # G+6702  U+647A  uni647A,  CJK UNIFIED IDEOGRAPH-647A
        '25736' => '4512'             , # G+4512  U+6488  uni6488,  CJK UNIFIED IDEOGRAPH-6488
        '25745' => '7447'             , # G+7447  U+6491  uni6491,  CJK UNIFIED IDEOGRAPH-6491
        '25746' => '5316'             , # G+5316  U+6492  uni6492,  CJK UNIFIED IDEOGRAPH-6492
        '25747' => '6150'             , # G+6150  U+6493  uni6493,  CJK UNIFIED IDEOGRAPH-6493
        '25754' => '4161'             , # G+4161  U+649A  uni649A,  CJK UNIFIED IDEOGRAPH-649A
        '25758' => '4244'             , # G+4244  U+649E  uni649E,  CJK UNIFIED IDEOGRAPH-649E
        '25764' => '7171'             , # G+7171  U+64A4  uni64A4,  CJK UNIFIED IDEOGRAPH-64A4
        '25765' => '4932'             , # G+4932  U+64A5  uni64A5,  CJK UNIFIED IDEOGRAPH-64A5
        '25771' => '4818'             , # G+4818  U+64AB  uni64AB,  CJK UNIFIED IDEOGRAPH-64AB
        '25773' => '7478'             , # G+7478  U+64AD  uni64AD,  CJK UNIFIED IDEOGRAPH-64AD
        '25774' => '7261'             , # G+7261  U+64AE  uni64AE,  CJK UNIFIED IDEOGRAPH-64AE
        '25776' => '7062'             , # G+7062  U+64B0  uni64B0,  CJK UNIFIED IDEOGRAPH-64B0
        '25778' => '4890'             , # G+4890  U+64B2  uni64B2,  CJK UNIFIED IDEOGRAPH-64B2
        '25787' => '4212'             , # G+4212  U+64BB  uni64BB,  CJK UNIFIED IDEOGRAPH-64BB
        '25793' => '6091'             , # G+6091  U+64C1  uni64C1,  CJK UNIFIED IDEOGRAPH-64C1
        '25796' => '4513'             , # G+4513  U+64C4  uni64C4,  CJK UNIFIED IDEOGRAPH-64C4
        '25797' => '7153'             , # G+7153  U+64C5  uni64C5,  CJK UNIFIED IDEOGRAPH-64C5
        '25799' => '7445'             , # G+7445  U+64C7  uni64C7,  CJK UNIFIED IDEOGRAPH-64C7
        '25802' => '3630'             , # G+3630  U+64CA  uni64CA,  CJK UNIFIED IDEOGRAPH-64CA
        '25805' => '6790'             , # G+6790  U+64CD  uni64CD,  CJK UNIFIED IDEOGRAPH-64CD
        '25806' => '3674'             , # G+3674  U+64CE  uni64CE,  CJK UNIFIED IDEOGRAPH-64CE
        '25810' => '4040'             , # G+4040  U+64D2  uni64D2,  CJK UNIFIED IDEOGRAPH-64D2
        '25812' => '4220'             , # G+4220  U+64D4  uni64D4,  CJK UNIFIED IDEOGRAPH-64D4
        '25816' => '5023'             , # G+5023  U+64D8  uni64D8,  CJK UNIFIED IDEOGRAPH-64D8
        '25818' => '3591'             , # G+3591  U+64DA  uni64DA,  CJK UNIFIED IDEOGRAPH-64DA
        '25825' => '4259'             , # G+4259  U+64E1  uni64E1,  CJK UNIFIED IDEOGRAPH-64E1
        '25826' => '7399'             , # G+7399  U+64E2  uni64E2,  CJK UNIFIED IDEOGRAPH-64E2
        '25829' => '4400'             , # G+4400  U+64E5  uni64E5,  CJK UNIFIED IDEOGRAPH-64E5
        '25830' => '7079'             , # G+7079  U+64E6  uni64E6,  CJK UNIFIED IDEOGRAPH-64E6
        '25831' => '3592'             , # G+3592  U+64E7  uni64E7,  CJK UNIFIED IDEOGRAPH-64E7
        '25836' => '6400'             , # G+6400  U+64EC  uni64EC,  CJK UNIFIED IDEOGRAPH-64EC
        '25842' => '7140'             , # G+7140  U+64F2  uni64F2,  CJK UNIFIED IDEOGRAPH-64F2
        '25844' => '7877'             , # G+7877  U+64F4  uni64F4,  CJK UNIFIED IDEOGRAPH-64F4
        '25850' => '7479'             , # G+7479  U+64FA  uni64FA,  CJK UNIFIED IDEOGRAPH-64FA
        '25854' => '6151'             , # G+6151  U+64FE  uni64FE,  CJK UNIFIED IDEOGRAPH-64FE
        '25856' => '4911'             , # G+4911  U+6500  uni6500,  CJK UNIFIED IDEOGRAPH-6500
        '25860' => '7448'             , # G+7448  U+6504  uni6504,  CJK UNIFIED IDEOGRAPH-6504
        '25880' => '5884'             , # G+5884  U+6518  uni6518,  CJK UNIFIED IDEOGRAPH-6518
        '25885' => '5471'             , # G+5471  U+651D  uni651D,  CJK UNIFIED IDEOGRAPH-651D
        '25891' => '4467'             , # G+4467  U+6523  uni6523,  CJK UNIFIED IDEOGRAPH-6523
        '25898' => '3875'             , # G+3875  U+652A  uni652A,  CJK UNIFIED IDEOGRAPH-652A
        '25899' => '7878'             , # G+7878  U+652B  uni652B,  CJK UNIFIED IDEOGRAPH-652B
        '25900' => '4401'             , # G+4401  U+652C  uni652C,  CJK UNIFIED IDEOGRAPH-652C
        '25903' => '6951'             , # G+6951  U+652F  uni652F,  CJK UNIFIED IDEOGRAPH-652F
        '25910' => '5583'             , # G+5583  U+6536  uni6536,  CJK UNIFIED IDEOGRAPH-6536
        '25911' => '3739'             , # G+3739  U+6537  uni6537,  CJK UNIFIED IDEOGRAPH-6537
        '25912' => '6325'             , # G+6325  U+6538  uni6538,  CJK UNIFIED IDEOGRAPH-6538
        '25913' => '3568'             , # G+3568  U+6539  uni6539,  CJK UNIFIED IDEOGRAPH-6539
        '25915' => '3796'             , # G+3796  U+653B  uni653B,  CJK UNIFIED IDEOGRAPH-653B
        '25918' => '4949'             , # G+4949  U+653E  uni653E,  CJK UNIFIED IDEOGRAPH-653E
        '25919' => '6718'             , # G+6718  U+653F  uni653F,  CJK UNIFIED IDEOGRAPH-653F
        '25925' => '3740'             , # G+3740  U+6545  uni6545,  CJK UNIFIED IDEOGRAPH-6545
        '25928' => '7956'             , # G+7956  U+6548  uni6548,  CJK UNIFIED IDEOGRAPH-6548
        '25933' => '5382'             , # G+5382  U+654D  uni654D,  CJK UNIFIED IDEOGRAPH-654D
        '25934' => '3876'             , # G+3876  U+654E  uni654E,  CJK UNIFIED IDEOGRAPH-654E
        '25935' => '4874'             , # G+4874  U+654F  uni654F,  CJK UNIFIED IDEOGRAPH-654F
        '25937' => '3912'             , # G+3912  U+6551  uni6551,  CJK UNIFIED IDEOGRAPH-6551
        '25942' => '6064'             , # G+6064  U+6556  uni6556,  CJK UNIFIED IDEOGRAPH-6556
        '25943' => '7505'             , # G+7505  U+6557  uni6557,  CJK UNIFIED IDEOGRAPH-6557
        '25950' => '7100'             , # G+7100  U+655E  uni655E,  CJK UNIFIED IDEOGRAPH-655E
        '25954' => '3517'             , # G+3517  U+6562  uni6562,  CJK UNIFIED IDEOGRAPH-6562
        '25955' => '5306'             , # G+5306  U+6563  uni6563,  CJK UNIFIED IDEOGRAPH-6563
        '25958' => '4321'             , # G+4321  U+6566  uni6566,  CJK UNIFIED IDEOGRAPH-6566
        '25964' => '3675'             , # G+3675  U+656C  uni656C,  CJK UNIFIED IDEOGRAPH-656C
        '25965' => '5885'             , # G+5885  U+656D  uni656D,  CJK UNIFIED IDEOGRAPH-656D
        '25970' => '3741'             , # G+3741  U+6572  uni6572,  CJK UNIFIED IDEOGRAPH-6572
        '25972' => '6719'             , # G+6719  U+6574  uni6574,  CJK UNIFIED IDEOGRAPH-6574
        '25973' => '6624'             , # G+6624  U+6575  uni6575,  CJK UNIFIED IDEOGRAPH-6575
        '25975' => '5125'             , # G+5125  U+6577  uni6577,  CJK UNIFIED IDEOGRAPH-6577
        '25976' => '5584'             , # G+5584  U+6578  uni6578,  CJK UNIFIED IDEOGRAPH-6578
        '25982' => '5425'             , # G+5425  U+657E  uni657E,  CJK UNIFIED IDEOGRAPH-657E
        '25986' => '4484'             , # G+4484  U+6582  uni6582,  CJK UNIFIED IDEOGRAPH-6582
        '25987' => '7539'             , # G+7539  U+6583  uni6583,  CJK UNIFIED IDEOGRAPH-6583
        '25989' => '7957'             , # G+7957  U+6585  uni6585,  CJK UNIFIED IDEOGRAPH-6585
        '25991' => '4840'             , # G+4840  U+6587  uni6587,  CJK UNIFIED IDEOGRAPH-6587
        '25996' => '5226'             , # G+5226  U+658C  uni658C,  CJK UNIFIED IDEOGRAPH-658C
        '26000' => '5193'             , # G+5193  U+6590  uni6590,  CJK UNIFIED IDEOGRAPH-6590
        '26001' => '4912'             , # G+4912  U+6591  uni6591,  CJK UNIFIED IDEOGRAPH-6591
        '26007' => '4349'             , # G+4349  U+6597  uni6597,  CJK UNIFIED IDEOGRAPH-6597
        '26009' => '4555'             , # G+4555  U+6599  uni6599,  CJK UNIFIED IDEOGRAPH-6599
        '26011' => '3768'             , # G+3768  U+659B  uni659B,  CJK UNIFIED IDEOGRAPH-659B
        '26012' => '5266'             , # G+5266  U+659C  uni659C,  CJK UNIFIED IDEOGRAPH-659C
        '26015' => '7030'             , # G+7030  U+659F  uni659F,  CJK UNIFIED IDEOGRAPH-659F
        '26017' => '5817'             , # G+5817  U+65A1  uni65A1,  CJK UNIFIED IDEOGRAPH-65A1
        '26020' => '4027'             , # G+4027  U+65A4  uni65A4,  CJK UNIFIED IDEOGRAPH-65A4
        '26021' => '7141'             , # G+7141  U+65A5  uni65A5,  CJK UNIFIED IDEOGRAPH-65A5
        '26023' => '5126'             , # G+5126  U+65A7  uni65A7,  CJK UNIFIED IDEOGRAPH-65A7
        '26027' => '6512'             , # G+6512  U+65AB  uni65AB,  CJK UNIFIED IDEOGRAPH-65AB
        '26028' => '7088'             , # G+7088  U+65AC  uni65AC,  CJK UNIFIED IDEOGRAPH-65AC
        '26031' => '5267'             , # G+5267  U+65AF  uni65AF,  CJK UNIFIED IDEOGRAPH-65AF
        '26032' => '5742'             , # G+5742  U+65B0  uni65B0,  CJK UNIFIED IDEOGRAPH-65B0
        '26039' => '4199'             , # G+4199  U+65B7  uni65B7,  CJK UNIFIED IDEOGRAPH-65B7
        '26041' => '4950'             , # G+4950  U+65B9  uni65B9,  CJK UNIFIED IDEOGRAPH-65B9
        '26044' => '5904'             , # G+5904  U+65BC  uni65BC,  CJK UNIFIED IDEOGRAPH-65BC
        '26045' => '5703'             , # G+5703  U+65BD  uni65BD,  CJK UNIFIED IDEOGRAPH-65BD
        '26049' => '4951'             , # G+4951  U+65C1  uni65C1,  CJK UNIFIED IDEOGRAPH-65C1
        '26053' => '4447'             , # G+4447  U+65C5  uni65C5,  CJK UNIFIED IDEOGRAPH-65C5
        '26059' => '5426'             , # G+5426  U+65CB  uni65CB,  CJK UNIFIED IDEOGRAPH-65CB
        '26060' => '6720'             , # G+6720  U+65CC  uni65CC,  CJK UNIFIED IDEOGRAPH-65CC
        '26063' => '6828'             , # G+6828  U+65CF  uni65CF,  CJK UNIFIED IDEOGRAPH-65CF
        '26066' => '4578'             , # G+4578  U+65D2  uni65D2,  CJK UNIFIED IDEOGRAPH-65D2
        '26071' => '4081'             , # G+4081  U+65D7  uni65D7,  CJK UNIFIED IDEOGRAPH-65D7
        '26080' => '4819'             , # G+4819  U+65E0  uni65E0,  CJK UNIFIED IDEOGRAPH-65E0
        '26083' => '4082'             , # G+4082  U+65E3  uni65E3,  CJK UNIFIED IDEOGRAPH-65E3
        '26085' => '6464'             , # G+6464  U+65E5  uni65E5,  CJK UNIFIED IDEOGRAPH-65E5
        '26086' => '4200'             , # G+4200  U+65E6  uni65E6,  CJK UNIFIED IDEOGRAPH-65E6
        '26088' => '6952'             , # G+6952  U+65E8  uni65E8,  CJK UNIFIED IDEOGRAPH-65E8
        '26089' => '6791'             , # G+6791  U+65E9  uni65E9,  CJK UNIFIED IDEOGRAPH-65E9
        '26092' => '5644'             , # G+5644  U+65EC  uni65EC,  CJK UNIFIED IDEOGRAPH-65EC
        '26093' => '6231'             , # G+6231  U+65ED  uni65ED,  CJK UNIFIED IDEOGRAPH-65ED
        '26097' => '7638'             , # G+7638  U+65F1  uni65F1,  CJK UNIFIED IDEOGRAPH-65F1
        '26100' => '6209'             , # G+6209  U+65F4  uni65F4,  CJK UNIFIED IDEOGRAPH-65F4
        '26106' => '6127'             , # G+6127  U+65FA  uni65FA,  CJK UNIFIED IDEOGRAPH-65FA
        '26107' => '4875'             , # G+4875  U+65FB  uni65FB,  CJK UNIFIED IDEOGRAPH-65FB
        '26108' => '4876'             , # G+4876  U+65FC  uni65FC,  CJK UNIFIED IDEOGRAPH-65FC
        '26109' => '4322'             , # G+4322  U+65FD  uni65FD,  CJK UNIFIED IDEOGRAPH-65FD
        '26111' => '6065'             , # G+6065  U+65FF  uni65FF,  CJK UNIFIED IDEOGRAPH-65FF
        '26118' => '3777'             , # G+3777  U+6606  uni6606,  CJK UNIFIED IDEOGRAPH-6606
        '26119' => '5688'             , # G+5688  U+6607  uni6607,  CJK UNIFIED IDEOGRAPH-6607
        '26121' => '4952'             , # G+4952  U+6609  uni6609,  CJK UNIFIED IDEOGRAPH-6609
        '26122' => '7810'             , # G+7810  U+660A  uni660A,  CJK UNIFIED IDEOGRAPH-660A
        '26124' => '7101'             , # G+7101  U+660C  uni660C,  CJK UNIFIED IDEOGRAPH-660C
        '26126' => '4751'             , # G+4751  U+660E  uni660E,  CJK UNIFIED IDEOGRAPH-660E
        '26127' => '7845'             , # G+7845  U+660F  uni660F,  CJK UNIFIED IDEOGRAPH-660F
        '26128' => '5161'             , # G+5161  U+6610  uni6610,  CJK UNIFIED IDEOGRAPH-6610
        '26129' => '4041'             , # G+4041  U+6611  uni6611,  CJK UNIFIED IDEOGRAPH-6611
        '26131' => '5950'             , # G+5950  U+6613  uni6613,  CJK UNIFIED IDEOGRAPH-6613
        '26132' => '5407'             , # G+5407  U+6614  uni6614,  CJK UNIFIED IDEOGRAPH-6614
        '26133' => '8019'             , # G+8019  U+6615  uni6615,  CJK UNIFIED IDEOGRAPH-6615
        '26142' => '5046'             , # G+5046  U+661E  uni661E,  CJK UNIFIED IDEOGRAPH-661E
        '26143' => '5480'             , # G+5480  U+661F  uni661F,  CJK UNIFIED IDEOGRAPH-661F
        '26144' => '6005'             , # G+6005  U+6620  uni6620,  CJK UNIFIED IDEOGRAPH-6620
        '26149' => '7300'             , # G+7300  U+6625  uni6625,  CJK UNIFIED IDEOGRAPH-6625
        '26151' => '4712'             , # G+4712  U+6627  uni6627,  CJK UNIFIED IDEOGRAPH-6627
        '26152' => '6513'             , # G+6513  U+6628  uni6628,  CJK UNIFIED IDEOGRAPH-6628
        '26157' => '5510'             , # G+5510  U+662D  uni662D,  CJK UNIFIED IDEOGRAPH-662D
        '26159' => '5704'             , # G+5704  U+662F  uni662F,  CJK UNIFIED IDEOGRAPH-662F
        '26160' => '7621'             , # G+7621  U+6630  uni6630,  CJK UNIFIED IDEOGRAPH-6630
        '26161' => '6232'             , # G+6232  U+6631  uni6631,  CJK UNIFIED IDEOGRAPH-6631
        '26164' => '4805'             , # G+4805  U+6634  uni6634,  CJK UNIFIED IDEOGRAPH-6634
        '26166' => '7102'             , # G+7102  U+6636  uni6636,  CJK UNIFIED IDEOGRAPH-6636
        '26170' => '5047'             , # G+5047  U+663A  uni663A,  CJK UNIFIED IDEOGRAPH-663A
        '26171' => '5836'             , # G+5836  U+663B  uni663B,  CJK UNIFIED IDEOGRAPH-663B
        '26177' => '6792'             , # G+6792  U+6641  uni6641,  CJK UNIFIED IDEOGRAPH-6641
        '26178' => '5705'             , # G+5705  U+6642  uni6642,  CJK UNIFIED IDEOGRAPH-6642
        '26179' => '7911'             , # G+7911  U+6643  uni6643,  CJK UNIFIED IDEOGRAPH-6643
        '26180' => '7912'             , # G+7912  U+6644  uni6644,  CJK UNIFIED IDEOGRAPH-6644
        '26185' => '6985'             , # G+6985  U+6649  uni6649,  CJK UNIFIED IDEOGRAPH-6649
        '26187' => '6986'             , # G+6986  U+664B  uni664B,  CJK UNIFIED IDEOGRAPH-664B
        '26191' => '5810'             , # G+5810  U+664F  uni664F,  CJK UNIFIED IDEOGRAPH-664F
        '26201' => '6908'             , # G+6908  U+6659  uni6659,  CJK UNIFIED IDEOGRAPH-6659
        '26203' => '7738'             , # G+7738  U+665B  uni665B,  CJK UNIFIED IDEOGRAPH-665B
        '26205' => '6873'             , # G+6873  U+665D  uni665D,  CJK UNIFIED IDEOGRAPH-665D
        '26206' => '8046'             , # G+8046  U+665E  uni665E,  CJK UNIFIED IDEOGRAPH-665E
        '26207' => '5481'             , # G+5481  U+665F  uni665F,  CJK UNIFIED IDEOGRAPH-665F
        '26212' => '6066'             , # G+6066  U+6664  uni6664,  CJK UNIFIED IDEOGRAPH-6664
        '26213' => '7890'             , # G+7890  U+6665  uni6665,  CJK UNIFIED IDEOGRAPH-6665
        '26214' => '7935'             , # G+7935  U+6666  uni6666,  CJK UNIFIED IDEOGRAPH-6666
        '26215' => '7811'             , # G+7811  U+6667  uni6667,  CJK UNIFIED IDEOGRAPH-6667
        '26216' => '5743'             , # G+5743  U+6668  uni6668,  CJK UNIFIED IDEOGRAPH-6668
        '26217' => '4677'             , # G+4677  U+6669  uni6669,  CJK UNIFIED IDEOGRAPH-6669
        '26219' => '7400'             , # G+7400  U+666B  uni666B,  CJK UNIFIED IDEOGRAPH-666B
        '26222' => '5062'             , # G+5062  U+666E  uni666E,  CJK UNIFIED IDEOGRAPH-666E
        '26223' => '3676'             , # G+3676  U+666F  uni666F,  CJK UNIFIED IDEOGRAPH-666F
        '26227' => '5408'             , # G+5408  U+6673  uni6673,  CJK UNIFIED IDEOGRAPH-6673
        '26228' => '7198'             , # G+7198  U+6674  uni6674,  CJK UNIFIED IDEOGRAPH-6674
        '26230' => '6721'             , # G+6721  U+6676  uni6676,  CJK UNIFIED IDEOGRAPH-6676
        '26231' => '3990'             , # G+3990  U+6677  uni6677,  CJK UNIFIED IDEOGRAPH-6677
        '26232' => '6722'             , # G+6722  U+6678  uni6678,  CJK UNIFIED IDEOGRAPH-6678
        '26234' => '6953'             , # G+6953  U+667A  uni667A,  CJK UNIFIED IDEOGRAPH-667A
        '26244' => '7991'             , # G+7991  U+6684  uni6684,  CJK UNIFIED IDEOGRAPH-6684
        '26247' => '3447'             , # G+3447  U+6687  uni6687,  CJK UNIFIED IDEOGRAPH-6687
        '26248' => '7988'             , # G+7988  U+6688  uni6688,  CJK UNIFIED IDEOGRAPH-6688
        '26249' => '8000'             , # G+8000  U+6689  uni6689,  CJK UNIFIED IDEOGRAPH-6689
        '26254' => '6006'             , # G+6006  U+668E  uni668E,  CJK UNIFIED IDEOGRAPH-668E
        '26256' => '6293'             , # G+6293  U+6690  uni6690,  CJK UNIFIED IDEOGRAPH-6690
        '26257' => '5383'             , # G+5383  U+6691  uni6691,  CJK UNIFIED IDEOGRAPH-6691
        '26262' => '4140'             , # G+4140  U+6696  uni6696,  CJK UNIFIED IDEOGRAPH-6696
        '26263' => '5825'             , # G+5825  U+6697  uni6697,  CJK UNIFIED IDEOGRAPH-6697
        '26264' => '5886'             , # G+5886  U+6698  uni6698,  CJK UNIFIED IDEOGRAPH-6698
        '26269' => '4752'             , # G+4752  U+669D  uni669D,  CJK UNIFIED IDEOGRAPH-669D
        '26272' => '3742'             , # G+3742  U+66A0  uni66A0,  CJK UNIFIED IDEOGRAPH-66A0
        '26274' => '7103'             , # G+7103  U+66A2  uni66A2,  CJK UNIFIED IDEOGRAPH-66A2
        '26283' => '6528'             , # G+6528  U+66AB  uni66AB,  CJK UNIFIED IDEOGRAPH-66AB
        '26286' => '4772'             , # G+4772  U+66AE  uni66AE,  CJK UNIFIED IDEOGRAPH-66AE
        '26290' => '6546'             , # G+6546  U+66B2  uni66B2,  CJK UNIFIED IDEOGRAPH-66B2
        '26291' => '7795'             , # G+7795  U+66B3  uni66B3,  CJK UNIFIED IDEOGRAPH-66B3
        '26292' => '7572'             , # G+7572  U+66B4  uni66B4,  CJK UNIFIED IDEOGRAPH-66B4
        '26297' => '5464'             , # G+5464  U+66B9  uni66B9,  CJK UNIFIED IDEOGRAPH-66B9
        '26299' => '3677'             , # G+3677  U+66BB  uni66BB,  CJK UNIFIED IDEOGRAPH-66BB
        '26302' => '4323'             , # G+4323  U+66BE  uni66BE,  CJK UNIFIED IDEOGRAPH-66BE
        '26308' => '5999'             , # G+5999  U+66C4  uni66C4,  CJK UNIFIED IDEOGRAPH-66C4
        '26310' => '4459'             , # G+4459  U+66C6  uni66C6,  CJK UNIFIED IDEOGRAPH-66C6
        '26311' => '4221'             , # G+4221  U+66C7  uni66C7,  CJK UNIFIED IDEOGRAPH-66C7
        '26313' => '7958'             , # G+7958  U+66C9  uni66C9,  CJK UNIFIED IDEOGRAPH-66C9
        '26326' => '5845'             , # G+5845  U+66D6  uni66D6,  CJK UNIFIED IDEOGRAPH-66D6
        '26329' => '5384'             , # G+5384  U+66D9  uni66D9,  CJK UNIFIED IDEOGRAPH-66D9
        '26332' => '6152'             , # G+6152  U+66DC  uni66DC,  CJK UNIFIED IDEOGRAPH-66DC
        '26333' => '7573'             , # G+7573  U+66DD  uni66DD,  CJK UNIFIED IDEOGRAPH-66DD
        '26336' => '3844'             , # G+3844  U+66E0  uni66E0,  CJK UNIFIED IDEOGRAPH-66E0
        '26342' => '8047'             , # G+8047  U+66E6  uni66E6,  CJK UNIFIED IDEOGRAPH-66E6
        '26352' => '6125'             , # G+6125  U+66F0  uni66F0,  CJK UNIFIED IDEOGRAPH-66F0
        '26354' => '3769'             , # G+3769  U+66F2  uni66F2,  CJK UNIFIED IDEOGRAPH-66F2
        '26355' => '6033'             , # G+6033  U+66F3  uni66F3,  CJK UNIFIED IDEOGRAPH-66F3
        '26356' => '3678'             , # G+3678  U+66F4  uni66F4,  CJK UNIFIED IDEOGRAPH-66F4
        '26359' => '3502'             , # G+3502  U+66F7  uni66F7,  CJK UNIFIED IDEOGRAPH-66F7
        '26360' => '5385'             , # G+5385  U+66F8  uni66F8,  CJK UNIFIED IDEOGRAPH-66F8
        '26361' => '6794'             , # G+6794  U+66F9  uni66F9,  CJK UNIFIED IDEOGRAPH-66F9
        '26362' => '6793'             , # G+6793  U+66FA  uni66FA,  CJK UNIFIED IDEOGRAPH-66FA
        '26364' => '4678'             , # G+4678  U+66FC  uni66FC,  CJK UNIFIED IDEOGRAPH-66FC
        '26366' => '6933'             , # G+6933  U+66FE  uni66FE,  CJK UNIFIED IDEOGRAPH-66FE
        '26367' => '7206'             , # G+7206  U+66FF  uni66FF,  CJK UNIFIED IDEOGRAPH-66FF
        '26368' => '7264'             , # G+7264  U+6700  uni6700,  CJK UNIFIED IDEOGRAPH-6700
        '26371' => '7936'             , # G+7936  U+6703  uni6703,  CJK UNIFIED IDEOGRAPH-6703
        '26376' => '6281'             , # G+6281  U+6708  uni6708,  CJK UNIFIED IDEOGRAPH-6708
        '26377' => '6326'             , # G+6326  U+6709  uni6709,  CJK UNIFIED IDEOGRAPH-6709
        '26379' => '5176'             , # G+5176  U+670B  uni670B,  CJK UNIFIED IDEOGRAPH-670B
        '26381' => '5080'             , # G+5080  U+670D  uni670D,  CJK UNIFIED IDEOGRAPH-670D
        '26388' => '5302'             , # G+5302  U+6714  uni6714,  CJK UNIFIED IDEOGRAPH-6714
        '26389' => '7031'             , # G+7031  U+6715  uni6715,  CJK UNIFIED IDEOGRAPH-6715
        '26391' => '4413'             , # G+4413  U+6717  uni6717,  CJK UNIFIED IDEOGRAPH-6717
        '26395' => '4700'             , # G+4700  U+671B  uni671B,  CJK UNIFIED IDEOGRAPH-671B
        '26397' => '6795'             , # G+6795  U+671D  uni671D,  CJK UNIFIED IDEOGRAPH-671D
        '26398' => '4083'             , # G+4083  U+671E  uni671E,  CJK UNIFIED IDEOGRAPH-671E
        '26399' => '4084'             , # G+4084  U+671F  uni671F,  CJK UNIFIED IDEOGRAPH-671F
        '26406' => '4798'             , # G+4798  U+6726  uni6726,  CJK UNIFIED IDEOGRAPH-6726
        '26407' => '4538'             , # G+4538  U+6727  uni6727,  CJK UNIFIED IDEOGRAPH-6727
        '26408' => '4788'             , # G+4788  U+6728  uni6728,  CJK UNIFIED IDEOGRAPH-6728
        '26410' => '4857'             , # G+4857  U+672A  uni672A,  CJK UNIFIED IDEOGRAPH-672A
        '26411' => '4691'             , # G+4691  U+672B  uni672B,  CJK UNIFIED IDEOGRAPH-672B
        '26412' => '5091'             , # G+5091  U+672C  uni672C,  CJK UNIFIED IDEOGRAPH-672C
        '26413' => '7080'             , # G+7080  U+672D  uni672D,  CJK UNIFIED IDEOGRAPH-672D
        '26414' => '7304'             , # G+7304  U+672E  uni672E,  CJK UNIFIED IDEOGRAPH-672E
        '26417' => '6874'             , # G+6874  U+6731  uni6731,  CJK UNIFIED IDEOGRAPH-6731
        '26420' => '4891'             , # G+4891  U+6734  uni6734,  CJK UNIFIED IDEOGRAPH-6734
        '26422' => '7388'             , # G+7388  U+6736  uni6736,  CJK UNIFIED IDEOGRAPH-6736
        '26426' => '3984'             , # G+3984  U+673A  uni673A,  CJK UNIFIED IDEOGRAPH-673A
        '26429' => '7975'             , # G+7975  U+673D  uni673D,  CJK UNIFIED IDEOGRAPH-673D
        '26438' => '3485'             , # G+3485  U+6746  uni6746,  CJK UNIFIED IDEOGRAPH-6746
        '26441' => '5321'             , # G+5321  U+6749  uni6749,  CJK UNIFIED IDEOGRAPH-6749
        '26446' => '4620'             , # G+4620  U+674E  uni674E,  CJK UNIFIED IDEOGRAPH-674E
        '26447' => '7707'             , # G+7707  U+674F  uni674F,  CJK UNIFIED IDEOGRAPH-674F
        '26448' => '6575'             , # G+6575  U+6750  uni6750,  CJK UNIFIED IDEOGRAPH-6750
        '26449' => '7249'             , # G+7249  U+6751  uni6751,  CJK UNIFIED IDEOGRAPH-6751
        '26451' => '7580'             , # G+7580  U+6753  uni6753,  CJK UNIFIED IDEOGRAPH-6753
        '26454' => '6547'             , # G+6547  U+6756  uni6756,  CJK UNIFIED IDEOGRAPH-6756
        '26460' => '4350'             , # G+4350  U+675C  uni675C,  CJK UNIFIED IDEOGRAPH-675C
        '26462' => '4085'             , # G+4085  U+675E  uni675E,  CJK UNIFIED IDEOGRAPH-675E
        '26463' => '5539'             , # G+5539  U+675F  uni675F,  CJK UNIFIED IDEOGRAPH-675F
        '26477' => '7677'             , # G+7677  U+676D  uni676D,  CJK UNIFIED IDEOGRAPH-676D
        '26479' => '4975'             , # G+4975  U+676F  uni676F,  CJK UNIFIED IDEOGRAPH-676F
        '26480' => '3615'             , # G+3615  U+6770  uni6770,  CJK UNIFIED IDEOGRAPH-6770
        '26481' => '4337'             , # G+4337  U+6771  uni6771,  CJK UNIFIED IDEOGRAPH-6771
        '26483' => '4806'             , # G+4806  U+6773  uni6773,  CJK UNIFIED IDEOGRAPH-6773
        '26485' => '6598'             , # G+6598  U+6775  uni6775,  CJK UNIFIED IDEOGRAPH-6775
        '26487' => '7480'             , # G+7480  U+6777  uni6777,  CJK UNIFIED IDEOGRAPH-6777
        '26491' => '4183'             , # G+4183  U+677B  uni677B,  CJK UNIFIED IDEOGRAPH-677B
        '26494' => '5555'             , # G+5555  U+677E  uni677E,  CJK UNIFIED IDEOGRAPH-677E
        '26495' => '7492'             , # G+7492  U+677F  uni677F,  CJK UNIFIED IDEOGRAPH-677F
        '26503' => '5194'             , # G+5194  U+6787  uni6787,  CJK UNIFIED IDEOGRAPH-6787
        '26505' => '6128'             , # G+6128  U+6789  uni6789,  CJK UNIFIED IDEOGRAPH-6789
        '26507' => '4953'             , # G+4953  U+678B  uni678B,  CJK UNIFIED IDEOGRAPH-678B
        '26511' => '4146'             , # G+4146  U+678F  uni678F,  CJK UNIFIED IDEOGRAPH-678F
        '26512' => '5409'             , # G+5409  U+6790  uni6790,  CJK UNIFIED IDEOGRAPH-6790
        '26515' => '4351'             , # G+4351  U+6793  uni6793,  CJK UNIFIED IDEOGRAPH-6793
        '26517' => '7369'             , # G+7369  U+6795  uni6795,  CJK UNIFIED IDEOGRAPH-6795
        '26519' => '4647'             , # G+4647  U+6797  uni6797,  CJK UNIFIED IDEOGRAPH-6797
        '26522' => '4713'             , # G+4713  U+679A  uni679A,  CJK UNIFIED IDEOGRAPH-679A
        '26524' => '3805'             , # G+3805  U+679C  uni679C,  CJK UNIFIED IDEOGRAPH-679C
        '26525' => '6954'             , # G+6954  U+679D  uni679D,  CJK UNIFIED IDEOGRAPH-679D
        '26543' => '3743'             , # G+3743  U+67AF  uni67AF,  CJK UNIFIED IDEOGRAPH-67AF
        '26544' => '7531'             , # G+7531  U+67B0  uni67B0,  CJK UNIFIED IDEOGRAPH-67B0
        '26547' => '6955'             , # G+6955  U+67B3  uni67B3,  CJK UNIFIED IDEOGRAPH-67B3
        '26550' => '3448'             , # G+3448  U+67B6  uni67B6,  CJK UNIFIED IDEOGRAPH-67B6
        '26551' => '3449'             , # G+3449  U+67B7  uni67B7,  CJK UNIFIED IDEOGRAPH-67B7
        '26552' => '3913'             , # G+3913  U+67B8  uni67B8,  CJK UNIFIED IDEOGRAPH-67B8
        '26558' => '5706'             , # G+5706  U+67BE  uni67BE,  CJK UNIFIED IDEOGRAPH-67BE
        '26564' => '5048'             , # G+5048  U+67C4  uni67C4,  CJK UNIFIED IDEOGRAPH-67C4
        '26575' => '4991'             , # G+4991  U+67CF  uni67CF,  CJK UNIFIED IDEOGRAPH-67CF
        '26576' => '4773'             , # G+4773  U+67D0  uni67D0,  CJK UNIFIED IDEOGRAPH-67D0
        '26577' => '3518'             , # G+3518  U+67D1  uni67D1,  CJK UNIFIED IDEOGRAPH-67D1
        '26578' => '7365'             , # G+7365  U+67D2  uni67D2,  CJK UNIFIED IDEOGRAPH-67D2
        '26579' => '5990'             , # G+5990  U+67D3  uni67D3,  CJK UNIFIED IDEOGRAPH-67D3
        '26580' => '6327'             , # G+6327  U+67D4  uni67D4,  CJK UNIFIED IDEOGRAPH-67D4
        '26586' => '6328'             , # G+6328  U+67DA  uni67DA,  CJK UNIFIED IDEOGRAPH-67DA
        '26589' => '7401'             , # G+7401  U+67DD  uni67DD,  CJK UNIFIED IDEOGRAPH-67DD
        '26601' => '3914'             , # G+3914  U+67E9  uni67E9,  CJK UNIFIED IDEOGRAPH-67E9
        '26604' => '3486'             , # G+3486  U+67EC  uni67EC,  CJK UNIFIED IDEOGRAPH-67EC
        '26607' => '3450'             , # G+3450  U+67EF  uni67EF,  CJK UNIFIED IDEOGRAPH-67EF
        '26608' => '4157'             , # G+4157  U+67F0  uni67F0,  CJK UNIFIED IDEOGRAPH-67F0
        '26609' => '6875'             , # G+6875  U+67F1  uni67F1,  CJK UNIFIED IDEOGRAPH-67F1
        '26611' => '4579'             , # G+4579  U+67F3  uni67F3,  CJK UNIFIED IDEOGRAPH-67F3
        '26612' => '5707'             , # G+5707  U+67F4  uni67F4,  CJK UNIFIED IDEOGRAPH-67F4
        '26613' => '7127'             , # G+7127  U+67F5  uni67F5,  CJK UNIFIED IDEOGRAPH-67F5
        '26614' => '5268'             , # G+5268  U+67F6  uni67F6,  CJK UNIFIED IDEOGRAPH-67F6
        '26619' => '5269'             , # G+5269  U+67FB  uni67FB,  CJK UNIFIED IDEOGRAPH-67FB
        '26622' => '6723'             , # G+6723  U+67FE  uni67FE,  CJK UNIFIED IDEOGRAPH-67FE
        '26642' => '5645'             , # G+5645  U+6812  uni6812,  CJK UNIFIED IDEOGRAPH-6812
        '26643' => '6658'             , # G+6658  U+6813  uni6813,  CJK UNIFIED IDEOGRAPH-6813
        '26646' => '5386'             , # G+5386  U+6816  uni6816,  CJK UNIFIED IDEOGRAPH-6816
        '26647' => '4602'             , # G+4602  U+6817  uni6817,  CJK UNIFIED IDEOGRAPH-6817
        '26657' => '3877'             , # G+3877  U+6821  uni6821,  CJK UNIFIED IDEOGRAPH-6821
        '26658' => '4992'             , # G+4992  U+6822  uni6822,  CJK UNIFIED IDEOGRAPH-6822
        '26666' => '6876'             , # G+6876  U+682A  uni682A,  CJK UNIFIED IDEOGRAPH-682A
        '26671' => '6233'             , # G+6233  U+682F  uni682F,  CJK UNIFIED IDEOGRAPH-682F
        '26680' => '7704'             , # G+7704  U+6838  uni6838,  CJK UNIFIED IDEOGRAPH-6838
        '26681' => '4028'             , # G+4028  U+6839  uni6839,  CJK UNIFIED IDEOGRAPH-6839
        '26684' => '3631'             , # G+3631  U+683C  uni683C,  CJK UNIFIED IDEOGRAPH-683C
        '26685' => '6576'             , # G+6576  U+683D  uni683D,  CJK UNIFIED IDEOGRAPH-683D
        '26688' => '3616'             , # G+3616  U+6840  uni6840,  CJK UNIFIED IDEOGRAPH-6840
        '26689' => '7678'             , # G+7678  U+6841  uni6841,  CJK UNIFIED IDEOGRAPH-6841
        '26690' => '3713'             , # G+3713  U+6842  uni6842,  CJK UNIFIED IDEOGRAPH-6842
        '26691' => '4285'             , # G+4285  U+6843  uni6843,  CJK UNIFIED IDEOGRAPH-6843
        '26696' => '5811'             , # G+5811  U+6848  uni6848,  CJK UNIFIED IDEOGRAPH-6848
        '26702' => '7020'             , # G+7020  U+684E  uni684E,  CJK UNIFIED IDEOGRAPH-684E
        '26704' => '4338'             , # G+4338  U+6850  uni6850,  CJK UNIFIED IDEOGRAPH-6850
        '26705' => '5346'             , # G+5346  U+6851  uni6851,  CJK UNIFIED IDEOGRAPH-6851
        '26707' => '7891'             , # G+7891  U+6853  uni6853,  CJK UNIFIED IDEOGRAPH-6853
        '26708' => '4130'             , # G+4130  U+6854  uni6854,  CJK UNIFIED IDEOGRAPH-6854
        '26733' => '6987'             , # G+6987  U+686D  uni686D,  CJK UNIFIED IDEOGRAPH-686D
        '26742' => '7454'             , # G+7454  U+6876  uni6876,  CJK UNIFIED IDEOGRAPH-6876
        '26751' => '3487'             , # G+3487  U+687F  uni687F,  CJK UNIFIED IDEOGRAPH-687F
        '26753' => '4431'             , # G+4431  U+6881  uni6881,  CJK UNIFIED IDEOGRAPH-6881
        '26757' => '4714'             , # G+4714  U+6885  uni6885,  CJK UNIFIED IDEOGRAPH-6885
        '26767' => '3770'             , # G+3770  U+688F  uni688F,  CJK UNIFIED IDEOGRAPH-688F
        '26771' => '6577'             , # G+6577  U+6893  uni6893,  CJK UNIFIED IDEOGRAPH-6893
        '26772' => '7342'             , # G+7342  U+6894  uni6894,  CJK UNIFIED IDEOGRAPH-6894
        '26775' => '3679'             , # G+3679  U+6897  uni6897,  CJK UNIFIED IDEOGRAPH-6897
        '26781' => '6796'             , # G+6796  U+689D  uni689D,  CJK UNIFIED IDEOGRAPH-689D
        '26783' => '7959'             , # G+7959  U+689F  uni689F,  CJK UNIFIED IDEOGRAPH-689F
        '26785' => '6110'             , # G+6110  U+68A1  uni68A1,  CJK UNIFIED IDEOGRAPH-68A1
        '26786' => '7220'             , # G+7220  U+68A2  uni68A2,  CJK UNIFIED IDEOGRAPH-68A2
        '26791' => '6067'             , # G+6067  U+68A7  uni68A7,  CJK UNIFIED IDEOGRAPH-68A7
        '26792' => '4621'             , # G+4621  U+68A8  uni68A8,  CJK UNIFIED IDEOGRAPH-68A8
        '26797' => '5270'             , # G+5270  U+68AD  uni68AD,  CJK UNIFIED IDEOGRAPH-68AD
        '26799' => '6767'             , # G+6767  U+68AF  uni68AF,  CJK UNIFIED IDEOGRAPH-68AF
        '26800' => '3714'             , # G+3714  U+68B0  uni68B0,  CJK UNIFIED IDEOGRAPH-68B0
        '26801' => '3778'             , # G+3778  U+68B1  uni68B1,  CJK UNIFIED IDEOGRAPH-68B1
        '26803' => '5511'             , # G+5511  U+68B3  uni68B3,  CJK UNIFIED IDEOGRAPH-68B3
        '26805' => '5011'             , # G+5011  U+68B5  uni68B5,  CJK UNIFIED IDEOGRAPH-68B5
        '26806' => '4858'             , # G+4858  U+68B6  uni68B6,  CJK UNIFIED IDEOGRAPH-68B6
        '26820' => '4087'             , # G+4087  U+68C4  uni68C4,  CJK UNIFIED IDEOGRAPH-68C4
        '26821' => '5049'             , # G+5049  U+68C5  uni68C5,  CJK UNIFIED IDEOGRAPH-68C5
        '26825' => '4738'             , # G+4738  U+68C9  uni68C9,  CJK UNIFIED IDEOGRAPH-68C9
        '26827' => '4086'             , # G+4086  U+68CB  uni68CB,  CJK UNIFIED IDEOGRAPH-68CB
        '26829' => '3779'             , # G+3779  U+68CD  uni68CD,  CJK UNIFIED IDEOGRAPH-68CD
        '26834' => '5099'             , # G+5099  U+68D2  uni68D2,  CJK UNIFIED IDEOGRAPH-68D2
        '26837' => '6842'             , # G+6842  U+68D5  uni68D5,  CJK UNIFIED IDEOGRAPH-68D5
        '26839' => '6797'             , # G+6797  U+68D7  uni68D7,  CJK UNIFIED IDEOGRAPH-68D7
        '26840' => '4020'             , # G+4020  U+68D8  uni68D8,  CJK UNIFIED IDEOGRAPH-68D8
        '26842' => '5177'             , # G+5177  U+68DA  uni68DA,  CJK UNIFIED IDEOGRAPH-68DA
        '26847' => '4339'             , # G+4339  U+68DF  uni68DF,  CJK UNIFIED IDEOGRAPH-68DF
        '26848' => '4245'             , # G+4245  U+68E0  uni68E0,  CJK UNIFIED IDEOGRAPH-68E0
        '26855' => '6523'             , # G+6523  U+68E7  uni68E7,  CJK UNIFIED IDEOGRAPH-68E7
        '26856' => '3715'             , # G+3715  U+68E8  uni68E8,  CJK UNIFIED IDEOGRAPH-68E8
        '26862' => '5322'             , # G+5322  U+68EE  uni68EE,  CJK UNIFIED IDEOGRAPH-68EE
        '26866' => '5387'             , # G+5387  U+68F2  uni68F2,  CJK UNIFIED IDEOGRAPH-68F2
        '26873' => '4286'             , # G+4286  U+68F9  uni68F9,  CJK UNIFIED IDEOGRAPH-68F9
        '26874' => '3823'             , # G+3823  U+68FA  uni68FA,  CJK UNIFIED IDEOGRAPH-68FA
        '26880' => '6111'             , # G+6111  U+6900  uni6900,  CJK UNIFIED IDEOGRAPH-6900
        '26885' => '6401'             , # G+6401  U+6905  uni6905,  CJK UNIFIED IDEOGRAPH-6905
        '26893' => '5725'             , # G+5725  U+690D  uni690D,  CJK UNIFIED IDEOGRAPH-690D
        '26894' => '7268'             , # G+7268  U+690E  uni690E,  CJK UNIFIED IDEOGRAPH-690E
        '26898' => '7221'             , # G+7221  U+6912  uni6912,  CJK UNIFIED IDEOGRAPH-6912
        '26919' => '4753'             , # G+4753  U+6927  uni6927,  CJK UNIFIED IDEOGRAPH-6927
        '26928' => '5868'             , # G+5868  U+6930  uni6930,  CJK UNIFIED IDEOGRAPH-6930
        '26941' => '5964'             , # G+5964  U+693D  uni693D,  CJK UNIFIED IDEOGRAPH-693D
        '26943' => '7301'             , # G+7301  U+693F  uni693F,  CJK UNIFIED IDEOGRAPH-693F
        '26954' => '5887'             , # G+5887  U+694A  uni694A,  CJK UNIFIED IDEOGRAPH-694A
        '26963' => '7592'             , # G+7592  U+6953  uni6953,  CJK UNIFIED IDEOGRAPH-6953
        '26964' => '5452'             , # G+5452  U+6954  uni6954,  CJK UNIFIED IDEOGRAPH-6954
        '26965' => '7389'             , # G+7389  U+6955  uni6955,  CJK UNIFIED IDEOGRAPH-6955
        '26967' => '3607'             , # G+3607  U+6957  uni6957,  CJK UNIFIED IDEOGRAPH-6957
        '26969' => '4820'             , # G+4820  U+6959  uni6959,  CJK UNIFIED IDEOGRAPH-6959
        '26970' => '7222'             , # G+7222  U+695A  uni695A,  CJK UNIFIED IDEOGRAPH-695A
        '26974' => '4608'             , # G+4608  U+695E  uni695E,  CJK UNIFIED IDEOGRAPH-695E
        '26976' => '4147'             , # G+4147  U+6960  uni6960,  CJK UNIFIED IDEOGRAPH-6960
        '26977' => '6329'             , # G+6329  U+6961  uni6961,  CJK UNIFIED IDEOGRAPH-6961
        '26978' => '6330'             , # G+6330  U+6962  uni6962,  CJK UNIFIED IDEOGRAPH-6962
        '26979' => '4859'             , # G+4859  U+6963  uni6963,  CJK UNIFIED IDEOGRAPH-6963
        '26984' => '6724'             , # G+6724  U+6968  uni6968,  CJK UNIFIED IDEOGRAPH-6968
        '26987' => '6928'             , # G+6928  U+696B  uni696B,  CJK UNIFIED IDEOGRAPH-696B
        '26989' => '5932'             , # G+5932  U+696D  uni696D,  CJK UNIFIED IDEOGRAPH-696D
        '26990' => '6599'             , # G+6599  U+696E  uni696E,  CJK UNIFIED IDEOGRAPH-696E
        '26991' => '5646'             , # G+5646  U+696F  uni696F,  CJK UNIFIED IDEOGRAPH-696F
        '26997' => '4021'             , # G+4021  U+6975  uni6975,  CJK UNIFIED IDEOGRAPH-6975
        '26999' => '7693'             , # G+7693  U+6977  uni6977,  CJK UNIFIED IDEOGRAPH-6977
        '27000' => '7269'             , # G+7269  U+6978  uni6978,  CJK UNIFIED IDEOGRAPH-6978
        '27001' => '6007'             , # G+6007  U+6979  uni6979,  CJK UNIFIED IDEOGRAPH-6979
        '27029' => '6183'             , # G+6183  U+6995  uni6995,  CJK UNIFIED IDEOGRAPH-6995
        '27035' => '6988'             , # G+6988  U+699B  uni699B,  CJK UNIFIED IDEOGRAPH-699B
        '27036' => '4954'             , # G+4954  U+699C  uni699C,  CJK UNIFIED IDEOGRAPH-699C
        '27045' => '7913'             , # G+7913  U+69A5  uni69A5,  CJK UNIFIED IDEOGRAPH-69A5
        '27047' => '5195'             , # G+5195  U+69A7  uni69A7,  CJK UNIFIED IDEOGRAPH-69A7
        '27054' => '6008'             , # G+6008  U+69AE  uni69AE,  CJK UNIFIED IDEOGRAPH-69AE
        '27060' => '4580'             , # G+4580  U+69B4  uni69B4,  CJK UNIFIED IDEOGRAPH-69B4
        '27067' => '7426'             , # G+7426  U+69BB  uni69BB,  CJK UNIFIED IDEOGRAPH-69BB
        '27073' => '3744'             , # G+3744  U+69C1  uni69C1,  CJK UNIFIED IDEOGRAPH-69C1
        '27075' => '4913'             , # G+4913  U+69C3  uni69C3,  CJK UNIFIED IDEOGRAPH-69C3
        '27083' => '3915'             , # G+3915  U+69CB  uni69CB,  CJK UNIFIED IDEOGRAPH-69CB
        '27084' => '7460'             , # G+7460  U+69CC  uni69CC,  CJK UNIFIED IDEOGRAPH-69CC
        '27085' => '7104'             , # G+7104  U+69CD  uni69CD,  CJK UNIFIED IDEOGRAPH-69CD
        '27088' => '3862'             , # G+3862  U+69D0  uni69D0,  CJK UNIFIED IDEOGRAPH-69D0
        '27112' => '3816'             , # G+3816  U+69E8  uni69E8,  CJK UNIFIED IDEOGRAPH-69E8
        '27114' => '3569'             , # G+3569  U+69EA  uni69EA,  CJK UNIFIED IDEOGRAPH-69EA
        '27131' => '3998'             , # G+3998  U+69FB  uni69FB,  CJK UNIFIED IDEOGRAPH-69FB
        '27133' => '6798'             , # G+6798  U+69FD  uni69FD,  CJK UNIFIED IDEOGRAPH-69FD
        '27135' => '4029'             , # G+4029  U+69FF  uni69FF,  CJK UNIFIED IDEOGRAPH-69FF
        '27138' => '5800'             , # G+5800  U+6A02  uni6A02,  CJK UNIFIED IDEOGRAPH-6A02
        '27146' => '4997'             , # G+4997  U+6A0A  uni6A0A,  CJK UNIFIED IDEOGRAPH-6A0A
        '27153' => '4432'             , # G+4432  U+6A11  uni6A11,  CJK UNIFIED IDEOGRAPH-6A11
        '27155' => '4567'             , # G+4567  U+6A13  uni6A13,  CJK UNIFIED IDEOGRAPH-6A13
        '27159' => '6600'             , # G+6600  U+6A17  uni6A17,  CJK UNIFIED IDEOGRAPH-6A17
        '27161' => '7581'             , # G+7581  U+6A19  uni6A19,  CJK UNIFIED IDEOGRAPH-6A19
        '27166' => '7270'             , # G+7270  U+6A1E  uni6A1E,  CJK UNIFIED IDEOGRAPH-6A1E
        '27167' => '6548'             , # G+6548  U+6A1F  uni6A1F,  CJK UNIFIED IDEOGRAPH-6A1F
        '27169' => '4774'             , # G+4774  U+6A21  uni6A21,  CJK UNIFIED IDEOGRAPH-6A21
        '27171' => '5888'             , # G+5888  U+6A23  uni6A23,  CJK UNIFIED IDEOGRAPH-6A23
        '27189' => '7223'             , # G+7223  U+6A35  uni6A35,  CJK UNIFIED IDEOGRAPH-6A35
        '27192' => '4892'             , # G+4892  U+6A38  uni6A38,  CJK UNIFIED IDEOGRAPH-6A38
        '27193' => '5585'             , # G+5585  U+6A39  uni6A39,  CJK UNIFIED IDEOGRAPH-6A39
        '27194' => '7866'             , # G+7866  U+6A3A  uni6A3A,  CJK UNIFIED IDEOGRAPH-6A3A
        '27197' => '6909'             , # G+6909  U+6A3D  uni6A3D,  CJK UNIFIED IDEOGRAPH-6A3D
        '27204' => '3519'             , # G+3519  U+6A44  uni6A44,  CJK UNIFIED IDEOGRAPH-6A44
        '27208' => '6153'             , # G+6153  U+6A48  uni6A48,  CJK UNIFIED IDEOGRAPH-6A48
        '27211' => '3878'             , # G+3878  U+6A4B  uni6A4B,  CJK UNIFIED IDEOGRAPH-6A4B
        '27218' => '6239'             , # G+6239  U+6A52  uni6A52,  CJK UNIFIED IDEOGRAPH-6A52
        '27219' => '5647'             , # G+5647  U+6A53  uni6A53,  CJK UNIFIED IDEOGRAPH-6A53
        '27224' => '4015'             , # G+4015  U+6A58  uni6A58,  CJK UNIFIED IDEOGRAPH-6A58
        '27225' => '4366'             , # G+4366  U+6A59  uni6A59,  CJK UNIFIED IDEOGRAPH-6A59
        '27231' => '4088'             , # G+4088  U+6A5F  uni6A5F,  CJK UNIFIED IDEOGRAPH-6A5F
        '27233' => '5347'             , # G+5347  U+6A61  uni6A61,  CJK UNIFIED IDEOGRAPH-6A61
        '27243' => '7951'             , # G+7951  U+6A6B  uni6A6B,  CJK UNIFIED IDEOGRAPH-6A6B
        '27264' => '4201'             , # G+4201  U+6A80  uni6A80,  CJK UNIFIED IDEOGRAPH-6A80
        '27268' => '3632'             , # G+3632  U+6A84  uni6A84,  CJK UNIFIED IDEOGRAPH-6A84
        '27273' => '6725'             , # G+6725  U+6A89  uni6A89,  CJK UNIFIED IDEOGRAPH-6A89
        '27277' => '5915'             , # G+5915  U+6A8D  uni6A8D,  CJK UNIFIED IDEOGRAPH-6A8D
        '27278' => '4042'             , # G+4042  U+6A8E  uni6A8E,  CJK UNIFIED IDEOGRAPH-6A8E
        '27287' => '5024'             , # G+5024  U+6A97  uni6A97,  CJK UNIFIED IDEOGRAPH-6A97
        '27292' => '7937'             , # G+7937  U+6A9C  uni6A9C,  CJK UNIFIED IDEOGRAPH-6A9C
        '27298' => '3620'             , # G+3620  U+6AA2  uni6AA2,  CJK UNIFIED IDEOGRAPH-6AA2
        '27299' => '6549'             , # G+6549  U+6AA3  uni6AA3,  CJK UNIFIED IDEOGRAPH-6AA3
        '27315' => '5227'             , # G+5227  U+6AB3  uni6AB3,  CJK UNIFIED IDEOGRAPH-6AB3
        '27323' => '7656'             , # G+7656  U+6ABB  uni6ABB,  CJK UNIFIED IDEOGRAPH-6ABB
        '27330' => '4287'             , # G+4287  U+6AC2  uni6AC2,  CJK UNIFIED IDEOGRAPH-6AC2
        '27331' => '3985'             , # G+3985  U+6AC3  uni6AC3,  CJK UNIFIED IDEOGRAPH-6AC3
        '27347' => '4514'             , # G+4514  U+6AD3  uni6AD3,  CJK UNIFIED IDEOGRAPH-6AD3
        '27354' => '4448'             , # G+4448  U+6ADA  uni6ADA,  CJK UNIFIED IDEOGRAPH-6ADA
        '27355' => '6927'             , # G+6927  U+6ADB  uni6ADB,  CJK UNIFIED IDEOGRAPH-6ADB
        '27382' => '7724'             , # G+7724  U+6AF6  uni6AF6,  CJK UNIFIED IDEOGRAPH-6AF6
        '27387' => '5858'             , # G+5858  U+6AFB  uni6AFB,  CJK UNIFIED IDEOGRAPH-6AFB
        '27396' => '4391'             , # G+4391  U+6B04  uni6B04,  CJK UNIFIED IDEOGRAPH-6B04
        '27402' => '3976'             , # G+3976  U+6B0A  uni6B0A,  CJK UNIFIED IDEOGRAPH-6B0A
        '27404' => '6550'             , # G+6550  U+6B0C  uni6B0C,  CJK UNIFIED IDEOGRAPH-6B0C
        '27410' => '4392'             , # G+4392  U+6B12  uni6B12,  CJK UNIFIED IDEOGRAPH-6B12
        '27414' => '4402'             , # G+4402  U+6B16  uni6B16,  CJK UNIFIED IDEOGRAPH-6B16
        '27424' => '8027'             , # G+8027  U+6B20  uni6B20,  CJK UNIFIED IDEOGRAPH-6B20
        '27425' => '7048'             , # G+7048  U+6B21  uni6B21,  CJK UNIFIED IDEOGRAPH-6B21
        '27427' => '8020'             , # G+8020  U+6B23  uni6B23,  CJK UNIFIED IDEOGRAPH-6B23
        '27442' => '6169'             , # G+6169  U+6B32  uni6B32,  CJK UNIFIED IDEOGRAPH-6B32
        '27450' => '4089'             , # G+4089  U+6B3A  uni6B3A,  CJK UNIFIED IDEOGRAPH-6B3A
        '27453' => '8028'             , # G+8028  U+6B3D  uni6B3D,  CJK UNIFIED IDEOGRAPH-6B3D
        '27454' => '3824'             , # G+3824  U+6B3E  uni6B3E,  CJK UNIFIED IDEOGRAPH-6B3E
        '27462' => '8029'             , # G+8029  U+6B46  uni6B46,  CJK UNIFIED IDEOGRAPH-6B46
        '27463' => '7727'             , # G+7727  U+6B47  uni6B47,  CJK UNIFIED IDEOGRAPH-6B47
        '27468' => '3451'             , # G+3451  U+6B4C  uni6B4C,  CJK UNIFIED IDEOGRAPH-6B4C
        '27470' => '7413'             , # G+7413  U+6B4E  uni6B4E,  CJK UNIFIED IDEOGRAPH-6B4E
        '27472' => '3916'             , # G+3916  U+6B50  uni6B50,  CJK UNIFIED IDEOGRAPH-6B50
        '27487' => '5937'             , # G+5937  U+6B5F  uni6B5F,  CJK UNIFIED IDEOGRAPH-6B5F
        '27489' => '7889'             , # G+7889  U+6B61  uni6B61,  CJK UNIFIED IDEOGRAPH-6B61
        '27490' => '6956'             , # G+6956  U+6B62  uni6B62,  CJK UNIFIED IDEOGRAPH-6B62
        '27491' => '6726'             , # G+6726  U+6B63  uni6B63,  CJK UNIFIED IDEOGRAPH-6B63
        '27492' => '7049'             , # G+7049  U+6B64  uni6B64,  CJK UNIFIED IDEOGRAPH-6B64
        '27493' => '5063'             , # G+5063  U+6B65  uni6B65,  CJK UNIFIED IDEOGRAPH-6B65
        '27494' => '4821'             , # G+4821  U+6B66  uni6B66,  CJK UNIFIED IDEOGRAPH-6B66
        '27498' => '6133'             , # G+6133  U+6B6A  uni6B6A,  CJK UNIFIED IDEOGRAPH-6B6A
        '27506' => '5494'             , # G+5494  U+6B72  uni6B72,  CJK UNIFIED IDEOGRAPH-6B72
        '27511' => '4460'             , # G+4460  U+6B77  uni6B77,  CJK UNIFIED IDEOGRAPH-6B77
        '27512' => '3991'             , # G+3991  U+6B78  uni6B78,  CJK UNIFIED IDEOGRAPH-6B78
        '27515' => '5271'             , # G+5271  U+6B7B  uni6B7B,  CJK UNIFIED IDEOGRAPH-6B7B
        '27519' => '4795'             , # G+4795  U+6B7F  uni6B7F,  CJK UNIFIED IDEOGRAPH-6B7F
        '27523' => '5837'             , # G+5837  U+6B83  uni6B83,  CJK UNIFIED IDEOGRAPH-6B83
        '27524' => '6989'             , # G+6989  U+6B84  uni6B84,  CJK UNIFIED IDEOGRAPH-6B84
        '27526' => '7436'             , # G+7436  U+6B86  uni6B86,  CJK UNIFIED IDEOGRAPH-6B86
        '27529' => '5648'             , # G+5648  U+6B89  uni6B89,  CJK UNIFIED IDEOGRAPH-6B89
        '27530' => '5586'             , # G+5586  U+6B8A  uni6B8A,  CJK UNIFIED IDEOGRAPH-6B8A
        '27542' => '5726'             , # G+5726  U+6B96  uni6B96,  CJK UNIFIED IDEOGRAPH-6B96
        '27544' => '6524'             , # G+6524  U+6B98  uni6B98,  CJK UNIFIED IDEOGRAPH-6B98
        '27550' => '6240'             , # G+6240  U+6B9E  uni6B9E,  CJK UNIFIED IDEOGRAPH-6B9E
        '27566' => '4485'             , # G+4485  U+6BAE  uni6BAE,  CJK UNIFIED IDEOGRAPH-6BAE
        '27567' => '5228'             , # G+5228  U+6BAF  uni6BAF,  CJK UNIFIED IDEOGRAPH-6BAF
        '27570' => '5465'             , # G+5465  U+6BB2  uni6BB2,  CJK UNIFIED IDEOGRAPH-6BB2
        '27573' => '4202'             , # G+4202  U+6BB5  uni6BB5,  CJK UNIFIED IDEOGRAPH-6BB5
        '27575' => '6376'             , # G+6376  U+6BB7  uni6BB7,  CJK UNIFIED IDEOGRAPH-6BB7
        '27578' => '5317'             , # G+5317  U+6BBA  uni6BBA,  CJK UNIFIED IDEOGRAPH-6BBA
        '27580' => '3470'             , # G+3470  U+6BBC  uni6BBC,  CJK UNIFIED IDEOGRAPH-6BBC
        '27583' => '6659'             , # G+6659  U+6BBF  uni6BBF,  CJK UNIFIED IDEOGRAPH-6BBF
        '27585' => '7996'             , # G+7996  U+6BC1  uni6BC1,  CJK UNIFIED IDEOGRAPH-6BC1
        '27589' => '6402'             , # G+6402  U+6BC5  uni6BC5,  CJK UNIFIED IDEOGRAPH-6BC5
        '27590' => '3917'             , # G+3917  U+6BC6  uni6BC6,  CJK UNIFIED IDEOGRAPH-6BC6
        '27595' => '4822'             , # G+4822  U+6BCB  uni6BCB,  CJK UNIFIED IDEOGRAPH-6BCB
        '27597' => '4775'             , # G+4775  U+6BCD  uni6BCD,  CJK UNIFIED IDEOGRAPH-6BCD
        '27599' => '4715'             , # G+4715  U+6BCF  uni6BCF,  CJK UNIFIED IDEOGRAPH-6BCF
        '27602' => '4309'             , # G+4309  U+6BD2  uni6BD2,  CJK UNIFIED IDEOGRAPH-6BD2
        '27603' => '6356'             , # G+6356  U+6BD3  uni6BD3,  CJK UNIFIED IDEOGRAPH-6BD3
        '27604' => '5196'             , # G+5196  U+6BD4  uni6BD4,  CJK UNIFIED IDEOGRAPH-6BD4
        '27606' => '5197'             , # G+5197  U+6BD6  uni6BD6,  CJK UNIFIED IDEOGRAPH-6BD6
        '27607' => '5198'             , # G+5198  U+6BD7  uni6BD7,  CJK UNIFIED IDEOGRAPH-6BD7
        '27608' => '5199'             , # G+5199  U+6BD8  uni6BD8,  CJK UNIFIED IDEOGRAPH-6BD8
        '27611' => '4776'             , # G+4776  U+6BDB  uni6BDB,  CJK UNIFIED IDEOGRAPH-6BDB
        '27627' => '7812'             , # G+7812  U+6BEB  uni6BEB,  CJK UNIFIED IDEOGRAPH-6BEB
        '27628' => '3918'             , # G+3918  U+6BEC  uni6BEC,  CJK UNIFIED IDEOGRAPH-6BEC
        '27656' => '6660'             , # G+6660  U+6C08  uni6C08,  CJK UNIFIED IDEOGRAPH-6C08
        '27663' => '5774'             , # G+5774  U+6C0F  uni6C0F,  CJK UNIFIED IDEOGRAPH-6C0F
        '27665' => '4877'             , # G+4877  U+6C11  uni6C11,  CJK UNIFIED IDEOGRAPH-6C11
        '27667' => '4728'             , # G+4728  U+6C13  uni6C13,  CJK UNIFIED IDEOGRAPH-6C13
        '27683' => '4090'             , # G+4090  U+6C23  uni6C23,  CJK UNIFIED IDEOGRAPH-6C23
        '27700' => '5587'             , # G+5587  U+6C34  uni6C34,  CJK UNIFIED IDEOGRAPH-6C34
        '27703' => '5238'             , # G+5238  U+6C37  uni6C37,  CJK UNIFIED IDEOGRAPH-6C37
        '27704' => '6009'             , # G+6009  U+6C38  uni6C38,  CJK UNIFIED IDEOGRAPH-6C38
        '27710' => '5012'             , # G+5012  U+6C3E  uni6C3E,  CJK UNIFIED IDEOGRAPH-6C3E
        '27712' => '6727'             , # G+6727  U+6C40  uni6C40,  CJK UNIFIED IDEOGRAPH-6C40
        '27713' => '6929'             , # G+6929  U+6C41  uni6C41,  CJK UNIFIED IDEOGRAPH-6C41
        '27714' => '3919'             , # G+3919  U+6C42  uni6C42,  CJK UNIFIED IDEOGRAPH-6C42
        '27726' => '5013'             , # G+5013  U+6C4E  uni6C4E,  CJK UNIFIED IDEOGRAPH-6C4E
        '27728' => '5410'             , # G+5410  U+6C50  uni6C50,  CJK UNIFIED IDEOGRAPH-6C50
        '27733' => '5307'             , # G+5307  U+6C55  uni6C55,  CJK UNIFIED IDEOGRAPH-6C55
        '27735' => '7639'             , # G+7639  U+6C57  uni6C57,  CJK UNIFIED IDEOGRAPH-6C57
        '27738' => '6068'             , # G+6068  U+6C5A  uni6C5A,  CJK UNIFIED IDEOGRAPH-6C5A
        '27741' => '5938'             , # G+5938  U+6C5D  uni6C5D,  CJK UNIFIED IDEOGRAPH-6C5D
        '27742' => '7855'             , # G+7855  U+6C5E  uni6C5E,  CJK UNIFIED IDEOGRAPH-6C5E
        '27743' => '3545'             , # G+3545  U+6C5F  uni6C5F,  CJK UNIFIED IDEOGRAPH-6C5F
        '27744' => '6957'             , # G+6957  U+6C60  uni6C60,  CJK UNIFIED IDEOGRAPH-6C60
        '27752' => '3784'             , # G+3784  U+6C68  uni6C68,  CJK UNIFIED IDEOGRAPH-6C68
        '27754' => '6129'             , # G+6129  U+6C6A  uni6C6A,  CJK UNIFIED IDEOGRAPH-6C6A
        '27757' => '6034'             , # G+6034  U+6C6D  uni6C6D,  CJK UNIFIED IDEOGRAPH-6C6D
        '27760' => '7437'             , # G+7437  U+6C70  uni6C70,  CJK UNIFIED IDEOGRAPH-6C70
        '27762' => '4055'             , # G+4055  U+6C72  uni6C72,  CJK UNIFIED IDEOGRAPH-6C72
        '27766' => '4841'             , # G+4841  U+6C76  uni6C76,  CJK UNIFIED IDEOGRAPH-6C76
        '27770' => '3649'             , # G+3649  U+6C7A  uni6C7A,  CJK UNIFIED IDEOGRAPH-6C7A
        '27773' => '4091'             , # G+4091  U+6C7D  uni6C7D,  CJK UNIFIED IDEOGRAPH-6C7D
        '27774' => '5162'             , # G+5162  U+6C7E  uni6C7E,  CJK UNIFIED IDEOGRAPH-6C7E
        '27777' => '5765'             , # G+5765  U+6C81  uni6C81,  CJK UNIFIED IDEOGRAPH-6C81
        '27778' => '4092'             , # G+4092  U+6C82  uni6C82,  CJK UNIFIED IDEOGRAPH-6C82
        '27779' => '6079'             , # G+6079  U+6C83  uni6C83,  CJK UNIFIED IDEOGRAPH-6C83
        '27781' => '6267'             , # G+6267  U+6C85  uni6C85,  CJK UNIFIED IDEOGRAPH-6C85
        '27782' => '7679'             , # G+7679  U+6C86  uni6C86,  CJK UNIFIED IDEOGRAPH-6C86
        '27783' => '5965'             , # G+5965  U+6C87  uni6C87,  CJK UNIFIED IDEOGRAPH-6C87
        '27784' => '7370'             , # G+7370  U+6C88  uni6C88,  CJK UNIFIED IDEOGRAPH-6C88
        '27788' => '4324'             , # G+4324  U+6C8C  uni6C8C,  CJK UNIFIED IDEOGRAPH-6C8C
        '27792' => '4789'             , # G+4789  U+6C90  uni6C90,  CJK UNIFIED IDEOGRAPH-6C90
        '27794' => '4796'             , # G+4796  U+6C92  uni6C92,  CJK UNIFIED IDEOGRAPH-6C92
        '27795' => '4234'             , # G+4234  U+6C93  uni6C93,  CJK UNIFIED IDEOGRAPH-6C93
        '27796' => '4739'             , # G+4739  U+6C94  uni6C94,  CJK UNIFIED IDEOGRAPH-6C94
        '27797' => '4849'             , # G+4849  U+6C95  uni6C95,  CJK UNIFIED IDEOGRAPH-6C95
        '27798' => '7308'             , # G+7308  U+6C96  uni6C96,  CJK UNIFIED IDEOGRAPH-6C96
        '27801' => '5272'             , # G+5272  U+6C99  uni6C99,  CJK UNIFIED IDEOGRAPH-6C99
        '27802' => '6958'             , # G+6958  U+6C9A  uni6C9A,  CJK UNIFIED IDEOGRAPH-6C9A
        '27803' => '7506'             , # G+7506  U+6C9B  uni6C9B,  CJK UNIFIED IDEOGRAPH-6C9B
        '27819' => '4692'             , # G+4692  U+6CAB  uni6CAB,  CJK UNIFIED IDEOGRAPH-6CAB
        '27822' => '6601'             , # G+6601  U+6CAE  uni6CAE,  CJK UNIFIED IDEOGRAPH-6CAE
        '27827' => '7622'             , # G+7622  U+6CB3  uni6CB3,  CJK UNIFIED IDEOGRAPH-6CB3
        '27832' => '5200'             , # G+5200  U+6CB8  uni6CB8,  CJK UNIFIED IDEOGRAPH-6CB8
        '27833' => '6331'             , # G+6331  U+6CB9  uni6CB9,  CJK UNIFIED IDEOGRAPH-6CB9
        '27835' => '7343'             , # G+7343  U+6CBB  uni6CBB,  CJK UNIFIED IDEOGRAPH-6CBB
        '27836' => '5512'             , # G+5512  U+6CBC  uni6CBC,  CJK UNIFIED IDEOGRAPH-6CBC
        '27837' => '3745'             , # G+3745  U+6CBD  uni6CBD,  CJK UNIFIED IDEOGRAPH-6CBD
        '27838' => '7179'             , # G+7179  U+6CBE  uni6CBE,  CJK UNIFIED IDEOGRAPH-6CBE
        '27839' => '5966'             , # G+5966  U+6CBF  uni6CBF,  CJK UNIFIED IDEOGRAPH-6CBF
        '27841' => '7914'             , # G+7914  U+6CC1  uni6CC1,  CJK UNIFIED IDEOGRAPH-6CC1
        '27842' => '7776'             , # G+7776  U+6CC2  uni6CC2,  CJK UNIFIED IDEOGRAPH-6CC2
        '27844' => '5453'             , # G+5453  U+6CC4  uni6CC4,  CJK UNIFIED IDEOGRAPH-6CC4
        '27849' => '7154'             , # G+7154  U+6CC9  uni6CC9,  CJK UNIFIED IDEOGRAPH-6CC9
        '27850' => '4893'             , # G+4893  U+6CCA  uni6CCA,  CJK UNIFIED IDEOGRAPH-6CCA
        '27852' => '7607'             , # G+7607  U+6CCC  uni6CCC,  CJK UNIFIED IDEOGRAPH-6CCC
        '27859' => '7856'             , # G+7856  U+6CD3  uni6CD3,  CJK UNIFIED IDEOGRAPH-6CD3
        '27861' => '5018'             , # G+5018  U+6CD5  uni6CD5,  CJK UNIFIED IDEOGRAPH-6CD5
        '27863' => '5273'             , # G+5273  U+6CD7  uni6CD7,  CJK UNIFIED IDEOGRAPH-6CD7
        '27867' => '5014'             , # G+5014  U+6CDB  uni6CDB,  CJK UNIFIED IDEOGRAPH-6CDB
        '27873' => '7556'             , # G+7556  U+6CE1  uni6CE1,  CJK UNIFIED IDEOGRAPH-6CE1
        '27874' => '7481'             , # G+7481  U+6CE2  uni6CE2,  CJK UNIFIED IDEOGRAPH-6CE2
        '27875' => '6388'             , # G+6388  U+6CE3  uni6CE3,  CJK UNIFIED IDEOGRAPH-6CE3
        '27877' => '4187'             , # G+4187  U+6CE5  uni6CE5,  CJK UNIFIED IDEOGRAPH-6CE5
        '27880' => '6877'             , # G+6877  U+6CE8  uni6CE8,  CJK UNIFIED IDEOGRAPH-6CE8
        '27883' => '7739'             , # G+7739  U+6CEB  uni6CEB,  CJK UNIFIED IDEOGRAPH-6CEB
        '27886' => '4914'             , # G+4914  U+6CEE  uni6CEE,  CJK UNIFIED IDEOGRAPH-6CEE
        '27887' => '4878'             , # G+4878  U+6CEF  uni6CEF,  CJK UNIFIED IDEOGRAPH-6CEF
        '27888' => '7438'             , # G+7438  U+6CF0  uni6CF0,  CJK UNIFIED IDEOGRAPH-6CF0
        '27891' => '6010'             , # G+6010  U+6CF3  uni6CF3,  CJK UNIFIED IDEOGRAPH-6CF3
        '27915' => '5889'             , # G+5889  U+6D0B  uni6D0B,  CJK UNIFIED IDEOGRAPH-6D0B
        '27916' => '4480'             , # G+4480  U+6D0C  uni6D0C,  CJK UNIFIED IDEOGRAPH-6D0C
        '27921' => '5064'             , # G+5064  U+6D11  uni6D11,  CJK UNIFIED IDEOGRAPH-6D11
        '27927' => '5495'             , # G+5495  U+6D17  uni6D17,  CJK UNIFIED IDEOGRAPH-6D17
        '27929' => '5588'             , # G+5588  U+6D19  uni6D19,  CJK UNIFIED IDEOGRAPH-6D19
        '27931' => '4382'             , # G+4382  U+6D1B  uni6D1B,  CJK UNIFIED IDEOGRAPH-6D1B
        '27934' => '4340'             , # G+4340  U+6D1E  uni6D1E,  CJK UNIFIED IDEOGRAPH-6D1E
        '27941' => '6990'             , # G+6990  U+6D25  uni6D25,  CJK UNIFIED IDEOGRAPH-6D25
        '27943' => '6332'             , # G+6332  U+6D27  uni6D27,  CJK UNIFIED IDEOGRAPH-6D27
        '27945' => '5454'             , # G+5454  U+6D29  uni6D29,  CJK UNIFIED IDEOGRAPH-6D29
        '27946' => '7857'             , # G+7857  U+6D2A  uni6D2A,  CJK UNIFIED IDEOGRAPH-6D2A
        '27954' => '6878'             , # G+6878  U+6D32  uni6D32,  CJK UNIFIED IDEOGRAPH-6D32
        '27957' => '5649'             , # G+5649  U+6D35  uni6D35,  CJK UNIFIED IDEOGRAPH-6D35
        '27958' => '8016'             , # G+8016  U+6D36  uni6D36,  CJK UNIFIED IDEOGRAPH-6D36
        '27960' => '3845'             , # G+3845  U+6D38  uni6D38,  CJK UNIFIED IDEOGRAPH-6D38
        '27961' => '6268'             , # G+6268  U+6D39  uni6D39,  CJK UNIFIED IDEOGRAPH-6D39
        '27963' => '7899'             , # G+7899  U+6D3B  uni6D3B,  CJK UNIFIED IDEOGRAPH-6D3B
        '27965' => '8032'             , # G+8032  U+6D3D  uni6D3D,  CJK UNIFIED IDEOGRAPH-6D3D
        '27966' => '7482'             , # G+7482  U+6D3E  uni6D3E,  CJK UNIFIED IDEOGRAPH-6D3E
        '27969' => '4581'             , # G+4581  U+6D41  uni6D41,  CJK UNIFIED IDEOGRAPH-6D41
        '27993' => '6687'             , # G+6687  U+6D59  uni6D59,  CJK UNIFIED IDEOGRAPH-6D59
        '27994' => '6910'             , # G+6910  U+6D5A  uni6D5A,  CJK UNIFIED IDEOGRAPH-6D5A
        '27996' => '5229'             , # G+5229  U+6D5C  uni6D5C,  CJK UNIFIED IDEOGRAPH-6D5C
        '28003' => '6112'             , # G+6112  U+6D63  uni6D63,  CJK UNIFIED IDEOGRAPH-6D63
        '28006' => '7557'             , # G+7557  U+6D66  uni6D66,  CJK UNIFIED IDEOGRAPH-6D66
        '28009' => '7813'             , # G+7813  U+6D69  uni6D69,  CJK UNIFIED IDEOGRAPH-6D69
        '28010' => '4414'             , # G+4414  U+6D6A  uni6D6A,  CJK UNIFIED IDEOGRAPH-6D6A
        '28012' => '4622'             , # G+4622  U+6D6C  uni6D6C,  CJK UNIFIED IDEOGRAPH-6D6C
        '28014' => '5127'             , # G+5127  U+6D6E  uni6D6E,  CJK UNIFIED IDEOGRAPH-6D6E
        '28020' => '6170'             , # G+6170  U+6D74  uni6D74,  CJK UNIFIED IDEOGRAPH-6D74
        '28023' => '7694'             , # G+7694  U+6D77  uni6D77,  CJK UNIFIED IDEOGRAPH-6D77
        '28024' => '7371'             , # G+7371  U+6D78  uni6D78,  CJK UNIFIED IDEOGRAPH-6D78
        '28025' => '7764'             , # G+7764  U+6D79  uni6D79,  CJK UNIFIED IDEOGRAPH-6D79
        '28031' => '7507'             , # G+7507  U+6D7F  uni6D7F,  CJK UNIFIED IDEOGRAPH-6D7F
        '28037' => '5986'             , # G+5986  U+6D85  uni6D85,  CJK UNIFIED IDEOGRAPH-6D85
        '28039' => '3680'             , # G+3680  U+6D87  uni6D87,  CJK UNIFIED IDEOGRAPH-6D87
        '28040' => '5513'             , # G+5513  U+6D88  uni6D88,  CJK UNIFIED IDEOGRAPH-6D88
        '28041' => '5472'             , # G+5472  U+6D89  uni6D89,  CJK UNIFIED IDEOGRAPH-6D89
        '28044' => '6184'             , # G+6184  U+6D8C  uni6D8C,  CJK UNIFIED IDEOGRAPH-6D8C
        '28045' => '7960'             , # G+7960  U+6D8D  uni6D8D,  CJK UNIFIED IDEOGRAPH-6D8D
        '28046' => '5967'             , # G+5967  U+6D8E  uni6D8E,  CJK UNIFIED IDEOGRAPH-6D8E
        '28049' => '5540'             , # G+5540  U+6D91  uni6D91,  CJK UNIFIED IDEOGRAPH-6D91
        '28051' => '5968'             , # G+5968  U+6D93  uni6D93,  CJK UNIFIED IDEOGRAPH-6D93
        '28053' => '7207'             , # G+7207  U+6D95  uni6D95,  CJK UNIFIED IDEOGRAPH-6D95
        '28079' => '5846'             , # G+5846  U+6DAF  uni6DAF,  CJK UNIFIED IDEOGRAPH-6DAF
        '28082' => '5854'             , # G+5854  U+6DB2  uni6DB2,  CJK UNIFIED IDEOGRAPH-6DB2
        '28085' => '7657'             , # G+7657  U+6DB5  uni6DB5,  CJK UNIFIED IDEOGRAPH-6DB5
        '28096' => '6728'             , # G+6728  U+6DC0  uni6DC0,  CJK UNIFIED IDEOGRAPH-6DC0
        '28099' => '3977'             , # G+3977  U+6DC3  uni6DC3,  CJK UNIFIED IDEOGRAPH-6DC3
        '28100' => '7344'             , # G+7344  U+6DC4  uni6DC4,  CJK UNIFIED IDEOGRAPH-6DC4
        '28101' => '5411'             , # G+5411  U+6DC5  uni6DC5,  CJK UNIFIED IDEOGRAPH-6DC5
        '28102' => '7961'             , # G+7961  U+6DC6  uni6DC6,  CJK UNIFIED IDEOGRAPH-6DC6
        '28103' => '4093'             , # G+4093  U+6DC7  uni6DC7,  CJK UNIFIED IDEOGRAPH-6DC7
        '28107' => '4648'             , # G+4648  U+6DCB  uni6DCB,  CJK UNIFIED IDEOGRAPH-6DCB
        '28111' => '7814'             , # G+7814  U+6DCF  uni6DCF,  CJK UNIFIED IDEOGRAPH-6DCF
        '28113' => '5633'             , # G+5633  U+6DD1  uni6DD1,  CJK UNIFIED IDEOGRAPH-6DD1
        '28120' => '4288'             , # G+4288  U+6DD8  uni6DD8,  CJK UNIFIED IDEOGRAPH-6DD8
        '28121' => '6843'             , # G+6843  U+6DD9  uni6DD9,  CJK UNIFIED IDEOGRAPH-6DD9
        '28122' => '4568'             , # G+4568  U+6DDA  uni6DDA,  CJK UNIFIED IDEOGRAPH-6DDA
        '28126' => '5556'             , # G+5556  U+6DDE  uni6DDE,  CJK UNIFIED IDEOGRAPH-6DDE
        '28129' => '4222'             , # G+4222  U+6DE1  uni6DE1,  CJK UNIFIED IDEOGRAPH-6DE1
        '28136' => '6729'             , # G+6729  U+6DE8  uni6DE8,  CJK UNIFIED IDEOGRAPH-6DE8
        '28138' => '4597'             , # G+4597  U+6DEA  uni6DEA,  CJK UNIFIED IDEOGRAPH-6DEA
        '28139' => '6382'             , # G+6382  U+6DEB  uni6DEB,  CJK UNIFIED IDEOGRAPH-6DEB
        '28142' => '7938'             , # G+7938  U+6DEE  uni6DEE,  CJK UNIFIED IDEOGRAPH-6DEE
        '28145' => '5766'             , # G+5766  U+6DF1  uni6DF1,  CJK UNIFIED IDEOGRAPH-6DF1
        '28147' => '5650'             , # G+5650  U+6DF3  uni6DF3,  CJK UNIFIED IDEOGRAPH-6DF3
        '28149' => '5969'             , # G+5969  U+6DF5  uni6DF5,  CJK UNIFIED IDEOGRAPH-6DF5
        '28151' => '7846'             , # G+7846  U+6DF7  uni6DF7,  CJK UNIFIED IDEOGRAPH-6DF7
        '28152' => '7199'             , # G+7199  U+6DF8  uni6DF8,  CJK UNIFIED IDEOGRAPH-6DF8
        '28153' => '5930'             , # G+5930  U+6DF9  uni6DF9,  CJK UNIFIED IDEOGRAPH-6DF9
        '28154' => '7155'             , # G+7155  U+6DFA  uni6DFA,  CJK UNIFIED IDEOGRAPH-6DFA
        '28155' => '7180'             , # G+7180  U+6DFB  uni6DFB,  CJK UNIFIED IDEOGRAPH-6DFB
        '28183' => '5323'             , # G+5323  U+6E17  uni6E17,  CJK UNIFIED IDEOGRAPH-6E17
        '28185' => '7892'             , # G+7892  U+6E19  uni6E19,  CJK UNIFIED IDEOGRAPH-6E19
        '28186' => '6602'             , # G+6602  U+6E1A  uni6E1A,  CJK UNIFIED IDEOGRAPH-6E1A
        '28187' => '3520'             , # G+3520  U+6E1B  uni6E1B,  CJK UNIFIED IDEOGRAPH-6E1B
        '28191' => '6730'             , # G+6730  U+6E1F  uni6E1F,  CJK UNIFIED IDEOGRAPH-6E1F
        '28192' => '3593'             , # G+3593  U+6E20  uni6E20,  CJK UNIFIED IDEOGRAPH-6E20
        '28193' => '4289'             , # G+4289  U+6E21  uni6E21,  CJK UNIFIED IDEOGRAPH-6E21
        '28195' => '5274'             , # G+5274  U+6E23  uni6E23,  CJK UNIFIED IDEOGRAPH-6E23
        '28196' => '4933'             , # G+4933  U+6E24  uni6E24,  CJK UNIFIED IDEOGRAPH-6E24
        '28197' => '5801'             , # G+5801  U+6E25  uni6E25,  CJK UNIFIED IDEOGRAPH-6E25
        '28198' => '6099'             , # G+6099  U+6E26  uni6E26,  CJK UNIFIED IDEOGRAPH-6E26
        '28203' => '5455'             , # G+5455  U+6E2B  uni6E2B,  CJK UNIFIED IDEOGRAPH-6E2B
        '28204' => '7334'             , # G+7334  U+6E2C  uni6E2C,  CJK UNIFIED IDEOGRAPH-6E2C
        '28205' => '6294'             , # G+6294  U+6E2D  uni6E2D,  CJK UNIFIED IDEOGRAPH-6E2D
        '28207' => '7680'             , # G+7680  U+6E2F  uni6E2F,  CJK UNIFIED IDEOGRAPH-6E2F
        '28210' => '5427'             , # G+5427  U+6E32  uni6E32,  CJK UNIFIED IDEOGRAPH-6E32
        '28212' => '3503'             , # G+3503  U+6E34  uni6E34,  CJK UNIFIED IDEOGRAPH-6E34
        '28214' => '6011'             , # G+6011  U+6E36  uni6E36,  CJK UNIFIED IDEOGRAPH-6E36
        '28216' => '6333'             , # G+6333  U+6E38  uni6E38,  CJK UNIFIED IDEOGRAPH-6E38
        '28218' => '4807'             , # G+4807  U+6E3A  uni6E3A,  CJK UNIFIED IDEOGRAPH-6E3A
        '28220' => '4860'             , # G+4860  U+6E3C  uni6E3C,  CJK UNIFIED IDEOGRAPH-6E3C
        '28221' => '6578'             , # G+6578  U+6E3D  uni6E3D,  CJK UNIFIED IDEOGRAPH-6E3D
        '28222' => '7847'             , # G+7847  U+6E3E  uni6E3E,  CJK UNIFIED IDEOGRAPH-6E3E
        '28227' => '4976'             , # G+4976  U+6E43  uni6E43,  CJK UNIFIED IDEOGRAPH-6E43
        '28228' => '4861'             , # G+4861  U+6E44  uni6E44,  CJK UNIFIED IDEOGRAPH-6E44
        '28234' => '6879'             , # G+6879  U+6E4A  uni6E4A,  CJK UNIFIED IDEOGRAPH-6E4A
        '28237' => '4203'             , # G+4203  U+6E4D  uni6E4D,  CJK UNIFIED IDEOGRAPH-6E4D
        '28246' => '7815'             , # G+7815  U+6E56  uni6E56,  CJK UNIFIED IDEOGRAPH-6E56
        '28248' => '5348'             , # G+5348  U+6E58  uni6E58,  CJK UNIFIED IDEOGRAPH-6E58
        '28251' => '4223'             , # G+4223  U+6E5B  uni6E5B,  CJK UNIFIED IDEOGRAPH-6E5B
        '28252' => '5727'             , # G+5727  U+6E5C  uni6E5C,  CJK UNIFIED IDEOGRAPH-6E5C
        '28254' => '6731'             , # G+6731  U+6E5E  uni6E5E,  CJK UNIFIED IDEOGRAPH-6E5E
        '28255' => '7915'             , # G+7915  U+6E5F  uni6E5F,  CJK UNIFIED IDEOGRAPH-6E5F
        '28263' => '6185'             , # G+6185  U+6E67  uni6E67,  CJK UNIFIED IDEOGRAPH-6E67
        '28267' => '7271'             , # G+7271  U+6E6B  uni6E6B,  CJK UNIFIED IDEOGRAPH-6E6B
        '28270' => '6453'             , # G+6453  U+6E6E  uni6E6E,  CJK UNIFIED IDEOGRAPH-6E6E
        '28271' => '7429'             , # G+7429  U+6E6F  uni6E6F,  CJK UNIFIED IDEOGRAPH-6E6F
        '28274' => '6269'             , # G+6269  U+6E72  uni6E72,  CJK UNIFIED IDEOGRAPH-6E72
        '28275' => '4148'             , # G+4148  U+6E73  uni6E73,  CJK UNIFIED IDEOGRAPH-6E73
        '28282' => '5065'             , # G+5065  U+6E7A  uni6E7A,  CJK UNIFIED IDEOGRAPH-6E7A
        '28304' => '6270'             , # G+6270  U+6E90  uni6E90,  CJK UNIFIED IDEOGRAPH-6E90
        '28310' => '6911'             , # G+6911  U+6E96  uni6E96,  CJK UNIFIED IDEOGRAPH-6E96
        '28316' => '4582'             , # G+4582  U+6E9C  uni6E9C,  CJK UNIFIED IDEOGRAPH-6E9C
        '28317' => '3920'             , # G+3920  U+6E9D  uni6E9D,  CJK UNIFIED IDEOGRAPH-6E9D
        '28319' => '4754'             , # G+4754  U+6E9F  uni6E9F,  CJK UNIFIED IDEOGRAPH-6E9F
        '28322' => '6465'             , # G+6465  U+6EA2  uni6EA2,  CJK UNIFIED IDEOGRAPH-6EA2
        '28325' => '5128'             , # G+5128  U+6EA5  uni6EA5,  CJK UNIFIED IDEOGRAPH-6EA5
        '28330' => '3716'             , # G+3716  U+6EAA  uni6EAA,  CJK UNIFIED IDEOGRAPH-6EAA
        '28331' => '6083'             , # G+6083  U+6EAB  uni6EAB,  CJK UNIFIED IDEOGRAPH-6EAB
        '28335' => '5514'             , # G+5514  U+6EAF  uni6EAF,  CJK UNIFIED IDEOGRAPH-6EAF
        '28337' => '6991'             , # G+6991  U+6EB1  uni6EB1,  CJK UNIFIED IDEOGRAPH-6EB1
        '28342' => '6186'             , # G+6186  U+6EB6  uni6EB6,  CJK UNIFIED IDEOGRAPH-6EB6
        '28346' => '4189'             , # G+4189  U+6EBA  uni6EBA,  CJK UNIFIED IDEOGRAPH-6EBA
        '28354' => '4955'             , # G+4955  U+6EC2  uni6EC2,  CJK UNIFIED IDEOGRAPH-6EC2
        '28356' => '7105'             , # G+7105  U+6EC4  uni6EC4,  CJK UNIFIED IDEOGRAPH-6EC4
        '28357' => '4746'             , # G+4746  U+6EC5  uni6EC5,  CJK UNIFIED IDEOGRAPH-6EC5
        '28361' => '7916'             , # G+7916  U+6EC9  uni6EC9,  CJK UNIFIED IDEOGRAPH-6EC9
        '28363' => '6493'             , # G+6493  U+6ECB  uni6ECB,  CJK UNIFIED IDEOGRAPH-6ECB
        '28364' => '7142'             , # G+7142  U+6ECC  uni6ECC,  CJK UNIFIED IDEOGRAPH-6ECC
        '28366' => '7777'             , # G+7777  U+6ECE  uni6ECE,  CJK UNIFIED IDEOGRAPH-6ECE
        '28369' => '7900'             , # G+7900  U+6ED1  uni6ED1,  CJK UNIFIED IDEOGRAPH-6ED1
        '28371' => '6579'             , # G+6579  U+6ED3  uni6ED3,  CJK UNIFIED IDEOGRAPH-6ED3
        '28372' => '4290'             , # G+4290  U+6ED4  uni6ED4,  CJK UNIFIED IDEOGRAPH-6ED4
        '28399' => '7208'             , # G+7208  U+6EEF  uni6EEF,  CJK UNIFIED IDEOGRAPH-6EEF
        '28404' => '6625'             , # G+6625  U+6EF4  uni6EF4,  CJK UNIFIED IDEOGRAPH-6EF4
        '28408' => '7816'             , # G+7816  U+6EF8  uni6EF8,  CJK UNIFIED IDEOGRAPH-6EF8
        '28414' => '3780'             , # G+3780  U+6EFE  uni6EFE,  CJK UNIFIED IDEOGRAPH-6EFE
        '28415' => '4679'             , # G+4679  U+6EFF  uni6EFF,  CJK UNIFIED IDEOGRAPH-6EFF
        '28417' => '5905'             , # G+5905  U+6F01  uni6F01,  CJK UNIFIED IDEOGRAPH-6F01
        '28418' => '7582'             , # G+7582  U+6F02  uni6F02,  CJK UNIFIED IDEOGRAPH-6F02
        '28422' => '7366'             , # G+7366  U+6F06  uni6F06,  CJK UNIFIED IDEOGRAPH-6F06
        '28431' => '4569'             , # G+4569  U+6F0F  uni6F0F,  CJK UNIFIED IDEOGRAPH-6F0F
        '28433' => '3570'             , # G+3570  U+6F11  uni6F11,  CJK UNIFIED IDEOGRAPH-6F11
        '28436' => '5970'             , # G+5970  U+6F14  uni6F14,  CJK UNIFIED IDEOGRAPH-6F14
        '28437' => '6799'             , # G+6799  U+6F15  uni6F15,  CJK UNIFIED IDEOGRAPH-6F15
        '28448' => '4666'             , # G+4666  U+6F20  uni6F20,  CJK UNIFIED IDEOGRAPH-6F20
        '28450' => '7640'             , # G+7640  U+6F22  uni6F22,  CJK UNIFIED IDEOGRAPH-6F22
        '28451' => '4468'             , # G+4468  U+6F23  uni6F23,  CJK UNIFIED IDEOGRAPH-6F23
        '28459' => '4680'             , # G+4680  U+6F2B  uni6F2B,  CJK UNIFIED IDEOGRAPH-6F2B
        '28460' => '6959'             , # G+6959  U+6F2C  uni6F2C,  CJK UNIFIED IDEOGRAPH-6F2C
        '28465' => '5589'             , # G+5589  U+6F31  uni6F31,  CJK UNIFIED IDEOGRAPH-6F31
        '28466' => '7106'             , # G+7106  U+6F32  uni6F32,  CJK UNIFIED IDEOGRAPH-6F32
        '28472' => '6695'             , # G+6695  U+6F38  uni6F38,  CJK UNIFIED IDEOGRAPH-6F38
        '28479' => '6551'             , # G+6551  U+6F3F  uni6F3F,  CJK UNIFIED IDEOGRAPH-6F3F
        '28481' => '6012'             , # G+6012  U+6F41  uni6F41,  CJK UNIFIED IDEOGRAPH-6F41
        '28497' => '4934'             , # G+4934  U+6F51  uni6F51,  CJK UNIFIED IDEOGRAPH-6F51
        '28500' => '3650'             , # G+3650  U+6F54  uni6F54,  CJK UNIFIED IDEOGRAPH-6F54
        '28503' => '7033'             , # G+7033  U+6F57  uni6F57,  CJK UNIFIED IDEOGRAPH-6F57
        '28504' => '4915'             , # G+4915  U+6F58  uni6F58,  CJK UNIFIED IDEOGRAPH-6F58
        '28506' => '5634'             , # G+5634  U+6F5A  uni6F5A,  CJK UNIFIED IDEOGRAPH-6F5A
        '28507' => '6529'             , # G+6529  U+6F5B  uni6F5B,  CJK UNIFIED IDEOGRAPH-6F5B
        '28510' => '4515'             , # G+4515  U+6F5E  uni6F5E,  CJK UNIFIED IDEOGRAPH-6F5E
        '28511' => '5412'             , # G+5412  U+6F5F  uni6F5F,  CJK UNIFIED IDEOGRAPH-6F5F
        '28514' => '7917'             , # G+7917  U+6F62  uni6F62,  CJK UNIFIED IDEOGRAPH-6F62
        '28516' => '6362'             , # G+6362  U+6F64  uni6F64,  CJK UNIFIED IDEOGRAPH-6F64
        '28525' => '4224'             , # G+4224  U+6F6D  uni6F6D,  CJK UNIFIED IDEOGRAPH-6F6D
        '28526' => '6800'             , # G+6800  U+6F6E  uni6F6E,  CJK UNIFIED IDEOGRAPH-6F6E
        '28528' => '3986'             , # G+3986  U+6F70  uni6F70,  CJK UNIFIED IDEOGRAPH-6F70
        '28538' => '6525'             , # G+6525  U+6F7A  uni6F7A,  CJK UNIFIED IDEOGRAPH-6F7A
        '28540' => '4341'             , # G+4341  U+6F7C  uni6F7C,  CJK UNIFIED IDEOGRAPH-6F7C
        '28541' => '5066'             , # G+5066  U+6F7D  uni6F7D,  CJK UNIFIED IDEOGRAPH-6F7D
        '28542' => '4639'             , # G+4639  U+6F7E  uni6F7E,  CJK UNIFIED IDEOGRAPH-6F7E
        '28545' => '5328'             , # G+5328  U+6F81  uni6F81,  CJK UNIFIED IDEOGRAPH-6F81
        '28548' => '7040'             , # G+7040  U+6F84  uni6F84,  CJK UNIFIED IDEOGRAPH-6F84
        '28552' => '7172'             , # G+7172  U+6F88  uni6F88,  CJK UNIFIED IDEOGRAPH-6F88
        '28557' => '6880'             , # G+6880  U+6F8D  uni6F8D,  CJK UNIFIED IDEOGRAPH-6F8D
        '28558' => '7514'             , # G+7514  U+6F8E  uni6F8E,  CJK UNIFIED IDEOGRAPH-6F8E
        '28560' => '6241'             , # G+6241  U+6F90  uni6F90,  CJK UNIFIED IDEOGRAPH-6F90
        '28564' => '7817'             , # G+7817  U+6F94  uni6F94,  CJK UNIFIED IDEOGRAPH-6F94
        '28567' => '3488'             , # G+3488  U+6F97  uni6F97,  CJK UNIFIED IDEOGRAPH-6F97
        '28579' => '7641'             , # G+7641  U+6FA3  uni6FA3,  CJK UNIFIED IDEOGRAPH-6FA3
        '28580' => '7446'             , # G+7446  U+6FA4  uni6FA4,  CJK UNIFIED IDEOGRAPH-6FA4
        '28583' => '4507'             , # G+4507  U+6FA7  uni6FA7,  CJK UNIFIED IDEOGRAPH-6FA7
        '28590' => '7939'             , # G+7939  U+6FAE  uni6FAE,  CJK UNIFIED IDEOGRAPH-6FAE
        '28591' => '7063'             , # G+7063  U+6FAF  uni6FAF,  CJK UNIFIED IDEOGRAPH-6FAF
        '28593' => '6661'             , # G+6661  U+6FB1  uni6FB1,  CJK UNIFIED IDEOGRAPH-6FB1
        '28595' => '6069'             , # G+6069  U+6FB3  uni6FB3,  CJK UNIFIED IDEOGRAPH-6FB3
        '28601' => '4225'             , # G+4225  U+6FB9  uni6FB9,  CJK UNIFIED IDEOGRAPH-6FB9
        '28606' => '4213'             , # G+4213  U+6FBE  uni6FBE,  CJK UNIFIED IDEOGRAPH-6FBE
        '28608' => '3633'             , # G+3633  U+6FC0  uni6FC0,  CJK UNIFIED IDEOGRAPH-6FC0
        '28609' => '7402'             , # G+7402  U+6FC1  uni6FC1,  CJK UNIFIED IDEOGRAPH-6FC1
        '28610' => '4486'             , # G+4486  U+6FC2  uni6FC2,  CJK UNIFIED IDEOGRAPH-6FC2
        '28611' => '4175'             , # G+4175  U+6FC3  uni6FC3,  CJK UNIFIED IDEOGRAPH-6FC3
        '28618' => '6035'             , # G+6035  U+6FCA  uni6FCA,  CJK UNIFIED IDEOGRAPH-6FCA
        '28629' => '5677'             , # G+5677  U+6FD5  uni6FD5,  CJK UNIFIED IDEOGRAPH-6FD5
        '28634' => '6013'             , # G+6013  U+6FDA  uni6FDA,  CJK UNIFIED IDEOGRAPH-6FDA
        '28639' => '6768'             , # G+6768  U+6FDF  uni6FDF,  CJK UNIFIED IDEOGRAPH-6FDF
        '28640' => '7818'             , # G+7818  U+6FE0  uni6FE0,  CJK UNIFIED IDEOGRAPH-6FE0
        '28641' => '6334'             , # G+6334  U+6FE1  uni6FE1,  CJK UNIFIED IDEOGRAPH-6FE1
        '28644' => '4291'             , # G+4291  U+6FE4  uni6FE4,  CJK UNIFIED IDEOGRAPH-6FE4
        '28649' => '7819'             , # G+7819  U+6FE9  uni6FE9,  CJK UNIFIED IDEOGRAPH-6FE9
        '28651' => '4403'             , # G+4403  U+6FEB  uni6FEB,  CJK UNIFIED IDEOGRAPH-6FEB
        '28652' => '6912'             , # G+6912  U+6FEC  uni6FEC,  CJK UNIFIED IDEOGRAPH-6FEC
        '28655' => '7403'             , # G+7403  U+6FEF  uni6FEF,  CJK UNIFIED IDEOGRAPH-6FEF
        '28657' => '5230'             , # G+5230  U+6FF1  uni6FF1,  CJK UNIFIED IDEOGRAPH-6FF1
        '28670' => '4449'             , # G+4449  U+6FFE  uni6FFE,  CJK UNIFIED IDEOGRAPH-6FFE
        '28673' => '5890'             , # G+5890  U+7001  uni7001,  CJK UNIFIED IDEOGRAPH-7001
        '28677' => '7778'             , # G+7778  U+7005  uni7005,  CJK UNIFIED IDEOGRAPH-7005
        '28678' => '4310'             , # G+4310  U+7006  uni7006,  CJK UNIFIED IDEOGRAPH-7006
        '28681' => '5275'             , # G+5275  U+7009  uni7009,  CJK UNIFIED IDEOGRAPH-7009
        '28683' => '5767'             , # G+5767  U+700B  uni700B,  CJK UNIFIED IDEOGRAPH-700B
        '28687' => '4583'             , # G+4583  U+700F  uni700F,  CJK UNIFIED IDEOGRAPH-700F
        '28689' => '7574'             , # G+7574  U+7011  uni7011,  CJK UNIFIED IDEOGRAPH-7011
        '28693' => '5231'             , # G+5231  U+7015  uni7015,  CJK UNIFIED IDEOGRAPH-7015
        '28696' => '4516'             , # G+4516  U+7018  uni7018,  CJK UNIFIED IDEOGRAPH-7018
        '28698' => '7642'             , # G+7642  U+701A  uni701A,  CJK UNIFIED IDEOGRAPH-701A
        '28699' => '6014'             , # G+6014  U+701B  uni701B,  CJK UNIFIED IDEOGRAPH-701B
        '28700' => '6370'             , # G+6370  U+701C  uni701C,  CJK UNIFIED IDEOGRAPH-701C
        '28701' => '4461'             , # G+4461  U+701D  uni701D,  CJK UNIFIED IDEOGRAPH-701D
        '28702' => '6732'             , # G+6732  U+701E  uni701E,  CJK UNIFIED IDEOGRAPH-701E
        '28703' => '5515'             , # G+5515  U+701F  uni701F,  CJK UNIFIED IDEOGRAPH-701F
        '28707' => '7695'             , # G+7695  U+7023  uni7023,  CJK UNIFIED IDEOGRAPH-7023
        '28711' => '4539'             , # G+4539  U+7027  uni7027,  CJK UNIFIED IDEOGRAPH-7027
        '28712' => '4544'             , # G+4544  U+7028  uni7028,  CJK UNIFIED IDEOGRAPH-7028
        '28719' => '6015'             , # G+6015  U+702F  uni702F,  CJK UNIFIED IDEOGRAPH-702F
        '28727' => '6437'             , # G+6437  U+7037  uni7037,  CJK UNIFIED IDEOGRAPH-7037
        '28734' => '4393'             , # G+4393  U+703E  uni703E,  CJK UNIFIED IDEOGRAPH-703E
        '28748' => '3825'             , # G+3825  U+704C  uni704C,  CJK UNIFIED IDEOGRAPH-704C
        '28752' => '7779'             , # G+7779  U+7050  uni7050,  CJK UNIFIED IDEOGRAPH-7050
        '28753' => '5562'             , # G+5562  U+7051  uni7051,  CJK UNIFIED IDEOGRAPH-7051
        '28760' => '7414'             , # G+7414  U+7058  uni7058,  CJK UNIFIED IDEOGRAPH-7058
        '28765' => '7820'             , # G+7820  U+705D  uni705D,  CJK UNIFIED IDEOGRAPH-705D
        '28771' => '4681'             , # G+4681  U+7063  uni7063,  CJK UNIFIED IDEOGRAPH-7063
        '28779' => '7867'             , # G+7867  U+706B  uni706B,  CJK UNIFIED IDEOGRAPH-706B
        '28784' => '7940'             , # G+7940  U+7070  uni7070,  CJK UNIFIED IDEOGRAPH-7070
        '28792' => '3921'             , # G+3921  U+7078  uni7078,  CJK UNIFIED IDEOGRAPH-7078
        '28796' => '6514'             , # G+6514  U+707C  uni707C,  CJK UNIFIED IDEOGRAPH-707C
        '28797' => '6580'             , # G+6580  U+707D  uni707D,  CJK UNIFIED IDEOGRAPH-707D
        '28805' => '3681'             , # G+3681  U+7085  uni7085,  CJK UNIFIED IDEOGRAPH-7085
        '28810' => '7321'             , # G+7321  U+708A  uni708A,  CJK UNIFIED IDEOGRAPH-708A
        '28814' => '5991'             , # G+5991  U+708E  uni708E,  CJK UNIFIED IDEOGRAPH-708E
        '28818' => '7224'             , # G+7224  U+7092  uni7092,  CJK UNIFIED IDEOGRAPH-7092
        '28824' => '8021'             , # G+8021  U+7098  uni7098,  CJK UNIFIED IDEOGRAPH-7098
        '28825' => '6494'             , # G+6494  U+7099  uni7099,  CJK UNIFIED IDEOGRAPH-7099
        '28826' => '3846'             , # G+3846  U+709A  uni709A,  CJK UNIFIED IDEOGRAPH-709A
        '28833' => '6733'             , # G+6733  U+70A1  uni70A1,  CJK UNIFIED IDEOGRAPH-70A1
        '28836' => '5516'             , # G+5516  U+70A4  uni70A4,  CJK UNIFIED IDEOGRAPH-70A4
        '28843' => '7740'             , # G+7740  U+70AB  uni70AB,  CJK UNIFIED IDEOGRAPH-70AB
        '28844' => '3594'             , # G+3594  U+70AC  uni70AC,  CJK UNIFIED IDEOGRAPH-70AC
        '28845' => '7415'             , # G+7415  U+70AD  uni70AD,  CJK UNIFIED IDEOGRAPH-70AD
        '28847' => '7780'             , # G+7780  U+70AF  uni70AF,  CJK UNIFIED IDEOGRAPH-70AF
        '28851' => '5050'             , # G+5050  U+70B3  uni70B3,  CJK UNIFIED IDEOGRAPH-70B3
        '28855' => '6881'             , # G+6881  U+70B7  uni70B7,  CJK UNIFIED IDEOGRAPH-70B7
        '28856' => '6515'             , # G+6515  U+70B8  uni70B8,  CJK UNIFIED IDEOGRAPH-70B8
        '28857' => '6696'             , # G+6696  U+70B9  uni70B9,  CJK UNIFIED IDEOGRAPH-70B9
        '28872' => '4481'             , # G+4481  U+70C8  uni70C8,  CJK UNIFIED IDEOGRAPH-70C8
        '28875' => '8007'             , # G+8007  U+70CB  uni70CB,  CJK UNIFIED IDEOGRAPH-70CB
        '28879' => '6070'             , # G+6070  U+70CF  uni70CF,  CJK UNIFIED IDEOGRAPH-70CF
        '28888' => '7858'             , # G+7858  U+70D8  uni70D8,  CJK UNIFIED IDEOGRAPH-70D8
        '28889' => '4383'             , # G+4383  U+70D9  uni70D9,  CJK UNIFIED IDEOGRAPH-70D9
        '28893' => '6935'             , # G+6935  U+70DD  uni70DD,  CJK UNIFIED IDEOGRAPH-70DD
        '28895' => '5971'             , # G+5971  U+70DF  uni70DF,  CJK UNIFIED IDEOGRAPH-70DF
        '28913' => '3682'             , # G+3682  U+70F1  uni70F1,  CJK UNIFIED IDEOGRAPH-70F1
        '28921' => '7515'             , # G+7515  U+70F9  uni70F9,  CJK UNIFIED IDEOGRAPH-70F9
        '28925' => '5100'             , # G+5100  U+70FD  uni70FD,  CJK UNIFIED IDEOGRAPH-70FD
        '28932' => '7983'             , # G+7983  U+7104  uni7104,  CJK UNIFIED IDEOGRAPH-7104
        '28937' => '5920'             , # G+5920  U+7109  uni7109,  CJK UNIFIED IDEOGRAPH-7109
        '28940' => '6913'             , # G+6913  U+710C  uni710C,  CJK UNIFIED IDEOGRAPH-710C
        '28953' => '4977'             , # G+4977  U+7119  uni7119,  CJK UNIFIED IDEOGRAPH-7119
        '28954' => '5163'             , # G+5163  U+711A  uni711A,  CJK UNIFIED IDEOGRAPH-711A
        '28958' => '4325'             , # G+4325  U+711E  uni711E,  CJK UNIFIED IDEOGRAPH-711E
        '28961' => '4823'             , # G+4823  U+7121  uni7121,  CJK UNIFIED IDEOGRAPH-7121
        '28966' => '7225'             , # G+7225  U+7126  uni7126,  CJK UNIFIED IDEOGRAPH-7126
        '28976' => '5992'             , # G+5992  U+7130  uni7130,  CJK UNIFIED IDEOGRAPH-7130
        '28982' => '5972'             , # G+5972  U+7136  uni7136,  CJK UNIFIED IDEOGRAPH-7136
        '28999' => '8001'             , # G+8001  U+7147  uni7147,  CJK UNIFIED IDEOGRAPH-7147
        '29001' => '4469'             , # G+4469  U+7149  uni7149,  CJK UNIFIED IDEOGRAPH-7149
        '29002' => '7992'             , # G+7992  U+714A  uni714A,  CJK UNIFIED IDEOGRAPH-714A
        '29004' => '7918'             , # G+7918  U+714C  uni714C,  CJK UNIFIED IDEOGRAPH-714C
        '29006' => '6662'             , # G+6662  U+714E  uni714E,  CJK UNIFIED IDEOGRAPH-714E
        '29008' => '6016'             , # G+6016  U+7150  uni7150,  CJK UNIFIED IDEOGRAPH-7150
        '29014' => '4141'             , # G+4141  U+7156  uni7156,  CJK UNIFIED IDEOGRAPH-7156
        '29017' => '5973'             , # G+5973  U+7159  uni7159,  CJK UNIFIED IDEOGRAPH-7159
        '29020' => '6234'             , # G+6234  U+715C  uni715C,  CJK UNIFIED IDEOGRAPH-715C
        '29022' => '5318'             , # G+5318  U+715E  uni715E,  CJK UNIFIED IDEOGRAPH-715E
        '29028' => '4716'             , # G+4716  U+7164  uni7164,  CJK UNIFIED IDEOGRAPH-7164
        '29029' => '7893'             , # G+7893  U+7165  uni7165,  CJK UNIFIED IDEOGRAPH-7165
        '29030' => '7976'             , # G+7976  U+7166  uni7166,  CJK UNIFIED IDEOGRAPH-7166
        '29031' => '6801'             , # G+6801  U+7167  uni7167,  CJK UNIFIED IDEOGRAPH-7167
        '29033' => '4998'             , # G+4998  U+7169  uni7169,  CJK UNIFIED IDEOGRAPH-7169
        '29036' => '5891'             , # G+5891  U+716C  uni716C,  CJK UNIFIED IDEOGRAPH-716C
        '29038' => '6495'             , # G+6495  U+716E  uni716E,  CJK UNIFIED IDEOGRAPH-716E
        '29053' => '5428'             , # G+5428  U+717D  uni717D,  CJK UNIFIED IDEOGRAPH-717D
        '29060' => '5728'             , # G+5728  U+7184  uni7184,  CJK UNIFIED IDEOGRAPH-7184
        '29065' => '6242'             , # G+6242  U+7189  uni7189,  CJK UNIFIED IDEOGRAPH-7189
        '29066' => '6253'             , # G+6253  U+718A  uni718A,  CJK UNIFIED IDEOGRAPH-718A
        '29071' => '7984'             , # G+7984  U+718F  uni718F,  CJK UNIFIED IDEOGRAPH-718F
        '29074' => '7781'             , # G+7781  U+7192  uni7192,  CJK UNIFIED IDEOGRAPH-7192
        '29076' => '6187'             , # G+6187  U+7194  uni7194,  CJK UNIFIED IDEOGRAPH-7194
        '29081' => '8048'             , # G+8048  U+7199  uni7199,  CJK UNIFIED IDEOGRAPH-7199
        '29087' => '5635'             , # G+5635  U+719F  uni719F,  CJK UNIFIED IDEOGRAPH-719F
        '29090' => '5101'             , # G+5101  U+71A2  uni71A2,  CJK UNIFIED IDEOGRAPH-71A2
        '29100' => '6071'             , # G+6071  U+71AC  uni71AC,  CJK UNIFIED IDEOGRAPH-71AC
        '29105' => '5987'             , # G+5987  U+71B1  uni71B1,  CJK UNIFIED IDEOGRAPH-71B1
        '29113' => '8049'             , # G+8049  U+71B9  uni71B9,  CJK UNIFIED IDEOGRAPH-71B9
        '29114' => '8050'             , # G+8050  U+71BA  uni71BA,  CJK UNIFIED IDEOGRAPH-71BA
        '29118' => '7345'             , # G+7345  U+71BE  uni71BE,  CJK UNIFIED IDEOGRAPH-71BE
        '29121' => '6000'             , # G+6000  U+71C1  uni71C1,  CJK UNIFIED IDEOGRAPH-71C1
        '29123' => '5974'             , # G+5974  U+71C3  uni71C3,  CJK UNIFIED IDEOGRAPH-71C3
        '29128' => '4367'             , # G+4367  U+71C8  uni71C8,  CJK UNIFIED IDEOGRAPH-71C8
        '29129' => '4326'             , # G+4326  U+71C9  uni71C9,  CJK UNIFIED IDEOGRAPH-71C9
        '29134' => '4556'             , # G+4556  U+71CE  uni71CE,  CJK UNIFIED IDEOGRAPH-71CE
        '29136' => '4640'             , # G+4640  U+71D0  uni71D0,  CJK UNIFIED IDEOGRAPH-71D0
        '29138' => '5517'             , # G+5517  U+71D2  uni71D2,  CJK UNIFIED IDEOGRAPH-71D2
        '29140' => '4999'             , # G+4999  U+71D4  uni71D4,  CJK UNIFIED IDEOGRAPH-71D4
        '29141' => '5975'             , # G+5975  U+71D5  uni71D5,  CJK UNIFIED IDEOGRAPH-71D5
        '29151' => '6017'             , # G+6017  U+71DF  uni71DF,  CJK UNIFIED IDEOGRAPH-71DF
        '29157' => '6802'             , # G+6802  U+71E5  uni71E5,  CJK UNIFIED IDEOGRAPH-71E5
        '29158' => '7064'             , # G+7064  U+71E6  uni71E6,  CJK UNIFIED IDEOGRAPH-71E6
        '29159' => '5590'             , # G+5590  U+71E7  uni71E7,  CJK UNIFIED IDEOGRAPH-71E7
        '29165' => '7243'             , # G+7243  U+71ED  uni71ED,  CJK UNIFIED IDEOGRAPH-71ED
        '29166' => '5473'             , # G+5473  U+71EE  uni71EE,  CJK UNIFIED IDEOGRAPH-71EE
        '29179' => '7985'             , # G+7985  U+71FB  uni71FB,  CJK UNIFIED IDEOGRAPH-71FB
        '29180' => '5744'             , # G+5744  U+71FC  uni71FC,  CJK UNIFIED IDEOGRAPH-71FC
        '29182' => '4292'             , # G+4292  U+71FE  uni71FE,  CJK UNIFIED IDEOGRAPH-71FE
        '29183' => '6154'             , # G+6154  U+71FF  uni71FF,  CJK UNIFIED IDEOGRAPH-71FF
        '29184' => '7731'             , # G+7731  U+7200  uni7200,  CJK UNIFIED IDEOGRAPH-7200
        '29190' => '7575'             , # G+7575  U+7206  uni7206,  CJK UNIFIED IDEOGRAPH-7206
        '29200' => '4517'             , # G+4517  U+7210  uni7210,  CJK UNIFIED IDEOGRAPH-7210
        '29211' => '4394'             , # G+4394  U+721B  uni721B,  CJK UNIFIED IDEOGRAPH-721B
        '29226' => '6803'             , # G+6803  U+722A  uni722A,  CJK UNIFIED IDEOGRAPH-722A
        '29228' => '7483'             , # G+7483  U+722C  uni722C,  CJK UNIFIED IDEOGRAPH-722C
        '29229' => '6587'             , # G+6587  U+722D  uni722D,  CJK UNIFIED IDEOGRAPH-722D
        '29232' => '6271'             , # G+6271  U+7230  uni7230,  CJK UNIFIED IDEOGRAPH-7230
        '29234' => '6295'             , # G+6295  U+7232  uni7232,  CJK UNIFIED IDEOGRAPH-7232
        '29237' => '6516'             , # G+6516  U+7235  uni7235,  CJK UNIFIED IDEOGRAPH-7235
        '29238' => '5129'             , # G+5129  U+7236  uni7236,  CJK UNIFIED IDEOGRAPH-7236
        '29242' => '5869'             , # G+5869  U+723A  uni723A,  CJK UNIFIED IDEOGRAPH-723A
        '29243' => '7962'             , # G+7962  U+723B  uni723B,  CJK UNIFIED IDEOGRAPH-723B
        '29245' => '5349'             , # G+5349  U+723D  uni723D,  CJK UNIFIED IDEOGRAPH-723D
        '29246' => '6422'             , # G+6422  U+723E  uni723E,  CJK UNIFIED IDEOGRAPH-723E
        '29248' => '5350'             , # G+5350  U+7240  uni7240,  CJK UNIFIED IDEOGRAPH-7240
        '29254' => '6552'             , # G+6552  U+7246  uni7246,  CJK UNIFIED IDEOGRAPH-7246
        '29255' => '7521'             , # G+7521  U+7247  uni7247,  CJK UNIFIED IDEOGRAPH-7247
        '29256' => '7493'             , # G+7493  U+7248  uni7248,  CJK UNIFIED IDEOGRAPH-7248
        '29260' => '7508'             , # G+7508  U+724C  uni724C,  CJK UNIFIED IDEOGRAPH-724C
        '29266' => '7191'             , # G+7191  U+7252  uni7252,  CJK UNIFIED IDEOGRAPH-7252
        '29272' => '4311'             , # G+4311  U+7258  uni7258,  CJK UNIFIED IDEOGRAPH-7258
        '29273' => '5782'             , # G+5782  U+7259  uni7259,  CJK UNIFIED IDEOGRAPH-7259
        '29275' => '6210'             , # G+6210  U+725B  uni725B,  CJK UNIFIED IDEOGRAPH-725B
        '29277' => '5232'             , # G+5232  U+725D  uni725D,  CJK UNIFIED IDEOGRAPH-725D
        '29279' => '4777'             , # G+4777  U+725F  uni725F,  CJK UNIFIED IDEOGRAPH-725F
        '29281' => '4778'             , # G+4778  U+7261  uni7261,  CJK UNIFIED IDEOGRAPH-7261
        '29282' => '4545'             , # G+4545  U+7262  uni7262,  CJK UNIFIED IDEOGRAPH-7262
        '29287' => '4790'             , # G+4790  U+7267  uni7267,  CJK UNIFIED IDEOGRAPH-7267
        '29289' => '4850'             , # G+4850  U+7269  uni7269,  CJK UNIFIED IDEOGRAPH-7269
        '29298' => '5369'             , # G+5369  U+7272  uni7272,  CJK UNIFIED IDEOGRAPH-7272
        '29305' => '7472'             , # G+7472  U+7279  uni7279,  CJK UNIFIED IDEOGRAPH-7279
        '29309' => '3638'             , # G+3638  U+727D  uni727D,  CJK UNIFIED IDEOGRAPH-727D
        '29312' => '5388'             , # G+5388  U+7280  uni7280,  CJK UNIFIED IDEOGRAPH-7280
        '29313' => '4623'             , # G+4623  U+7281  uni7281,  CJK UNIFIED IDEOGRAPH-7281
        '29346' => '4312'             , # G+4312  U+72A2  uni72A2,  CJK UNIFIED IDEOGRAPH-72A2
        '29351' => '8051'             , # G+8051  U+72A7  uni72A7,  CJK UNIFIED IDEOGRAPH-72A7
        '29356' => '3639'             , # G+3639  U+72AC  uni72AC,  CJK UNIFIED IDEOGRAPH-72AC
        '29359' => '5015'             , # G+5015  U+72AF  uni72AF,  CJK UNIFIED IDEOGRAPH-72AF
        '29376' => '5351'             , # G+5351  U+72C0  uni72C0,  CJK UNIFIED IDEOGRAPH-72C0
        '29378' => '3847'             , # G+3847  U+72C2  uni72C2,  CJK UNIFIED IDEOGRAPH-72C2
        '29380' => '6626'             , # G+6626  U+72C4  uni72C4,  CJK UNIFIED IDEOGRAPH-72C4
        '29390' => '5831'             , # G+5831  U+72CE  uni72CE,  CJK UNIFIED IDEOGRAPH-72CE
        '29392' => '7821'             , # G+7821  U+72D0  uni72D0,  CJK UNIFIED IDEOGRAPH-72D0
        '29399' => '3922'             , # G+3922  U+72D7  uni72D7,  CJK UNIFIED IDEOGRAPH-72D7
        '29401' => '6603'             , # G+6603  U+72D9  uni72D9,  CJK UNIFIED IDEOGRAPH-72D9
        '29409' => '3879'             , # G+3879  U+72E1  uni72E1,  CJK UNIFIED IDEOGRAPH-72E1
        '29417' => '5591'             , # G+5591  U+72E9  uni72E9,  CJK UNIFIED IDEOGRAPH-72E9
        '29432' => '4624'             , # G+4624  U+72F8  uni72F8,  CJK UNIFIED IDEOGRAPH-72F8
        '29433' => '7765'             , # G+7765  U+72F9  uni72F9,  CJK UNIFIED IDEOGRAPH-72F9
        '29436' => '4415'             , # G+4415  U+72FC  uni72FC,  CJK UNIFIED IDEOGRAPH-72FC
        '29437' => '7509'             , # G+7509  U+72FD  uni72FD,  CJK UNIFIED IDEOGRAPH-72FD
        '29450' => '6036'             , # G+6036  U+730A  uni730A,  CJK UNIFIED IDEOGRAPH-730A
        '29462' => '7107'             , # G+7107  U+7316  uni7316,  CJK UNIFIED IDEOGRAPH-7316
        '29467' => '4729'             , # G+4729  U+731B  uni731B,  CJK UNIFIED IDEOGRAPH-731B
        '29468' => '5708'             , # G+5708  U+731C  uni731C,  CJK UNIFIED IDEOGRAPH-731C
        '29469' => '6836'             , # G+6836  U+731D  uni731D,  CJK UNIFIED IDEOGRAPH-731D
        '29477' => '6138'             , # G+6138  U+7325  uni7325,  CJK UNIFIED IDEOGRAPH-7325
        '29481' => '5482'             , # G+5482  U+7329  uni7329,  CJK UNIFIED IDEOGRAPH-7329
        '29482' => '6604'             , # G+6604  U+732A  uni732A,  CJK UNIFIED IDEOGRAPH-732A
        '29483' => '4808'             , # G+4808  U+732B  uni732B,  CJK UNIFIED IDEOGRAPH-732B
        '29494' => '6335'             , # G+6335  U+7336  uni7336,  CJK UNIFIED IDEOGRAPH-7336
        '29495' => '6336'             , # G+6336  U+7337  uni7337,  CJK UNIFIED IDEOGRAPH-7337
        '29502' => '7901'             , # G+7901  U+733E  uni733E,  CJK UNIFIED IDEOGRAPH-733E
        '29503' => '6272'             , # G+6272  U+733F  uni733F,  CJK UNIFIED IDEOGRAPH-733F
        '29508' => '6080'             , # G+6080  U+7344  uni7344,  CJK UNIFIED IDEOGRAPH-7344
        '29509' => '5276'             , # G+5276  U+7345  uni7345,  CJK UNIFIED IDEOGRAPH-7345
        '29520' => '6553'             , # G+6553  U+7350  uni7350,  CJK UNIFIED IDEOGRAPH-7350
        '29522' => '6072'             , # G+6072  U+7352  uni7352,  CJK UNIFIED IDEOGRAPH-7352
        '29527' => '3980'             , # G+3980  U+7357  uni7357,  CJK UNIFIED IDEOGRAPH-7357
        '29544' => '4313'             , # G+4313  U+7368  uni7368,  CJK UNIFIED IDEOGRAPH-7368
        '29546' => '7941'             , # G+7941  U+736A  uni736A,  CJK UNIFIED IDEOGRAPH-736A
        '29552' => '6018'             , # G+6018  U+7370  uni7370,  CJK UNIFIED IDEOGRAPH-7370
        '29554' => '7949'             , # G+7949  U+7372  uni7372,  CJK UNIFIED IDEOGRAPH-7372
        '29557' => '4488'             , # G+4488  U+7375  uni7375,  CJK UNIFIED IDEOGRAPH-7375
        '29560' => '5592'             , # G+5592  U+7378  uni7378,  CJK UNIFIED IDEOGRAPH-7378
        '29562' => '4214'             , # G+4214  U+737A  uni737A,  CJK UNIFIED IDEOGRAPH-737A
        '29563' => '7725'             , # G+7725  U+737B  uni737B,  CJK UNIFIED IDEOGRAPH-737B
        '29572' => '7741'             , # G+7741  U+7384  uni7384,  CJK UNIFIED IDEOGRAPH-7384
        '29574' => '6496'             , # G+6496  U+7386  uni7386,  CJK UNIFIED IDEOGRAPH-7386
        '29575' => '5552'             , # G+5552  U+7387  uni7387,  CJK UNIFIED IDEOGRAPH-7387
        '29577' => '6081'             , # G+6081  U+7389  uni7389,  CJK UNIFIED IDEOGRAPH-7389
        '29579' => '6130'             , # G+6130  U+738B  uni738B,  CJK UNIFIED IDEOGRAPH-738B
        '29582' => '6734'             , # G+6734  U+738E  uni738E,  CJK UNIFIED IDEOGRAPH-738E
        '29588' => '7156'             , # G+7156  U+7394  uni7394,  CJK UNIFIED IDEOGRAPH-7394
        '29590' => '3923'             , # G+3923  U+7396  uni7396,  CJK UNIFIED IDEOGRAPH-7396
        '29591' => '6211'             , # G+6211  U+7397  uni7397,  CJK UNIFIED IDEOGRAPH-7397
        '29592' => '4094'             , # G+4094  U+7398  uni7398,  CJK UNIFIED IDEOGRAPH-7398
        '29599' => '4879'             , # G+4879  U+739F  uni739F,  CJK UNIFIED IDEOGRAPH-739F
        '29607' => '6363'             , # G+6363  U+73A7  uni73A7,  CJK UNIFIED IDEOGRAPH-73A7
        '29609' => '6113'             , # G+6113  U+73A9  uni73A9,  CJK UNIFIED IDEOGRAPH-73A9
        '29613' => '5233'             , # G+5233  U+73AD  uni73AD,  CJK UNIFIED IDEOGRAPH-73AD
        '29618' => '4495'             , # G+4495  U+73B2  uni73B2,  CJK UNIFIED IDEOGRAPH-73B2
        '29619' => '4260'             , # G+4260  U+73B3  uni73B3,  CJK UNIFIED IDEOGRAPH-73B3
        '29625' => '7742'             , # G+7742  U+73B9  uni73B9,  CJK UNIFIED IDEOGRAPH-73B9
        '29632' => '4894'             , # G+4894  U+73C0  uni73C0,  CJK UNIFIED IDEOGRAPH-73C0
        '29634' => '3452'             , # G+3452  U+73C2  uni73C2,  CJK UNIFIED IDEOGRAPH-73C2
        '29641' => '4880'             , # G+4880  U+73C9  uni73C9,  CJK UNIFIED IDEOGRAPH-73C9
        '29642' => '5308'             , # G+5308  U+73CA  uni73CA,  CJK UNIFIED IDEOGRAPH-73CA
        '29644' => '7608'             , # G+7608  U+73CC  uni73CC,  CJK UNIFIED IDEOGRAPH-73CC
        '29645' => '6992'             , # G+6992  U+73CD  uni73CD,  CJK UNIFIED IDEOGRAPH-73CD
        '29647' => '3471'             , # G+3471  U+73CF  uni73CF,  CJK UNIFIED IDEOGRAPH-73CF
        '29654' => '3848'             , # G+3848  U+73D6  uni73D6,  CJK UNIFIED IDEOGRAPH-73D6
        '29657' => '3797'             , # G+3797  U+73D9  uni73D9,  CJK UNIFIED IDEOGRAPH-73D9
        '29661' => '7977'             , # G+7977  U+73DD  uni73DD,  CJK UNIFIED IDEOGRAPH-73DD
        '29662' => '4384'             , # G+4384  U+73DE  uni73DE,  CJK UNIFIED IDEOGRAPH-73DE
        '29664' => '6882'             , # G+6882  U+73E0  uni73E0,  CJK UNIFIED IDEOGRAPH-73E0
        '29667' => '5651'             , # G+5651  U+73E3  uni73E3,  CJK UNIFIED IDEOGRAPH-73E3
        '29668' => '5067'             , # G+5067  U+73E4  uni73E4,  CJK UNIFIED IDEOGRAPH-73E4
        '29669' => '6423'             , # G+6423  U+73E5  uni73E5,  CJK UNIFIED IDEOGRAPH-73E5
        '29670' => '7713'             , # G+7713  U+73E6  uni73E6,  CJK UNIFIED IDEOGRAPH-73E6
        '29673' => '7782'             , # G+7782  U+73E9  uni73E9,  CJK UNIFIED IDEOGRAPH-73E9
        '29674' => '3999'             , # G+3999  U+73EA  uni73EA,  CJK UNIFIED IDEOGRAPH-73EA
        '29677' => '4916'             , # G+4916  U+73ED  uni73ED,  CJK UNIFIED IDEOGRAPH-73ED
        '29687' => '4824'             , # G+4824  U+73F7  uni73F7,  CJK UNIFIED IDEOGRAPH-73F7
        '29689' => '5483'             , # G+5483  U+73F9  uni73F9,  CJK UNIFIED IDEOGRAPH-73F9
        '29693' => '6735'             , # G+6735  U+73FD  uni73FD,  CJK UNIFIED IDEOGRAPH-73FD
        '29694' => '7743'             , # G+7743  U+73FE  uni73FE,  CJK UNIFIED IDEOGRAPH-73FE
        '29697' => '5429'             , # G+5429  U+7401  uni7401,  CJK UNIFIED IDEOGRAPH-7401
        '29699' => '3924'             , # G+3924  U+7403  uni7403,  CJK UNIFIED IDEOGRAPH-7403
        '29701' => '4416'             , # G+4416  U+7405  uni7405,  CJK UNIFIED IDEOGRAPH-7405
        '29702' => '4625'             , # G+4625  U+7406  uni7406,  CJK UNIFIED IDEOGRAPH-7406
        '29703' => '5593'             , # G+5593  U+7407  uni7407,  CJK UNIFIED IDEOGRAPH-7407
        '29705' => '4584'             , # G+4584  U+7409  uni7409,  CJK UNIFIED IDEOGRAPH-7409
        '29715' => '6114'             , # G+6114  U+7413  uni7413,  CJK UNIFIED IDEOGRAPH-7413
        '29723' => '7372'             , # G+7372  U+741B  uni741B,  CJK UNIFIED IDEOGRAPH-741B
        '29728' => '6663'             , # G+6663  U+7420  uni7420,  CJK UNIFIED IDEOGRAPH-7420
        '29729' => '5636'             , # G+5636  U+7421  uni7421,  CJK UNIFIED IDEOGRAPH-7421
        '29730' => '7404'             , # G+7404  U+7422  uni7422,  CJK UNIFIED IDEOGRAPH-7422
        '29733' => '7822'             , # G+7822  U+7425  uni7425,  CJK UNIFIED IDEOGRAPH-7425
        '29734' => '4095'             , # G+4095  U+7426  uni7426,  CJK UNIFIED IDEOGRAPH-7426
        '29736' => '3781'             , # G+3781  U+7428  uni7428,  CJK UNIFIED IDEOGRAPH-7428
        '29738' => '4096'             , # G+4096  U+742A  uni742A,  CJK UNIFIED IDEOGRAPH-742A
        '29739' => '5102'             , # G+5102  U+742B  uni742B,  CJK UNIFIED IDEOGRAPH-742B
        '29740' => '6115'             , # G+6115  U+742C  uni742C,  CJK UNIFIED IDEOGRAPH-742C
        '29742' => '6844'             , # G+6844  U+742E  uni742E,  CJK UNIFIED IDEOGRAPH-742E
        '29743' => '3826'             , # G+3826  U+742F  uni742F,  CJK UNIFIED IDEOGRAPH-742F
        '29744' => '5993'             , # G+5993  U+7430  uni7430,  CJK UNIFIED IDEOGRAPH-7430
        '29747' => '4649'             , # G+4649  U+7433  uni7433,  CJK UNIFIED IDEOGRAPH-7433
        '29748' => '4043'             , # G+4043  U+7434  uni7434,  CJK UNIFIED IDEOGRAPH-7434
        '29749' => '5201'             , # G+5201  U+7435  uni7435,  CJK UNIFIED IDEOGRAPH-7435
        '29750' => '7484'             , # G+7484  U+7436  uni7436,  CJK UNIFIED IDEOGRAPH-7436
        '29752' => '7405'             , # G+7405  U+7438  uni7438,  CJK UNIFIED IDEOGRAPH-7438
        '29754' => '5019'             , # G+5019  U+743A  uni743A,  CJK UNIFIED IDEOGRAPH-743A
        '29759' => '7848'             , # G+7848  U+743F  uni743F,  CJK UNIFIED IDEOGRAPH-743F
        '29760' => '6212'             , # G+6212  U+7440  uni7440,  CJK UNIFIED IDEOGRAPH-7440
        '29761' => '4779'             , # G+4779  U+7441  uni7441,  CJK UNIFIED IDEOGRAPH-7441
        '29763' => '7302'             , # G+7302  U+7443  uni7443,  CJK UNIFIED IDEOGRAPH-7443
        '29764' => '5430'             , # G+5430  U+7444  uni7444,  CJK UNIFIED IDEOGRAPH-7444
        '29771' => '6296'             , # G+6296  U+744B  uni744B,  CJK UNIFIED IDEOGRAPH-744B
        '29781' => '7623'             , # G+7623  U+7455  uni7455,  CJK UNIFIED IDEOGRAPH-7455
        '29783' => '6273'             , # G+6273  U+7457  uni7457,  CJK UNIFIED IDEOGRAPH-7457
        '29785' => '4173'             , # G+4173  U+7459  uni7459,  CJK UNIFIED IDEOGRAPH-7459
        '29786' => '7823'             , # G+7823  U+745A  uni745A,  CJK UNIFIED IDEOGRAPH-745A
        '29787' => '6019'             , # G+6019  U+745B  uni745B,  CJK UNIFIED IDEOGRAPH-745B
        '29788' => '6337'             , # G+6337  U+745C  uni745C,  CJK UNIFIED IDEOGRAPH-745C
        '29790' => '5389'             , # G+5389  U+745E  uni745E,  CJK UNIFIED IDEOGRAPH-745E
        '29791' => '5674'             , # G+5674  U+745F  uni745F,  CJK UNIFIED IDEOGRAPH-745F
        '29792' => '4585'             , # G+4585  U+7460  uni7460,  CJK UNIFIED IDEOGRAPH-7460
        '29794' => '6188'             , # G+6188  U+7462  uni7462,  CJK UNIFIED IDEOGRAPH-7462
        '29796' => '6155'             , # G+6155  U+7464  uni7464,  CJK UNIFIED IDEOGRAPH-7464
        '29797' => '6084'             , # G+6084  U+7465  uni7465,  CJK UNIFIED IDEOGRAPH-7465
        '29800' => '6993'             , # G+6993  U+7468  uni7468,  CJK UNIFIED IDEOGRAPH-7468
        '29801' => '7783'             , # G+7783  U+7469  uni7469,  CJK UNIFIED IDEOGRAPH-7469
        '29802' => '4657'             , # G+4657  U+746A  uni746A,  CJK UNIFIED IDEOGRAPH-746A
        '29807' => '4417'             , # G+4417  U+746F  uni746F,  CJK UNIFIED IDEOGRAPH-746F
        '29822' => '4030'             , # G+4030  U+747E  uni747E,  CJK UNIFIED IDEOGRAPH-747E
        '29826' => '4097'             , # G+4097  U+7482  uni7482,  CJK UNIFIED IDEOGRAPH-7482
        '29827' => '4626'             , # G+4626  U+7483  uni7483,  CJK UNIFIED IDEOGRAPH-7483
        '29831' => '5431'             , # G+5431  U+7487  uni7487,  CJK UNIFIED IDEOGRAPH-7487
        '29833' => '4470'             , # G+4470  U+7489  uni7489,  CJK UNIFIED IDEOGRAPH-7489
        '29835' => '6554'             , # G+6554  U+748B  uni748B,  CJK UNIFIED IDEOGRAPH-748B
        '29848' => '4641'             , # G+4641  U+7498  uni7498,  CJK UNIFIED IDEOGRAPH-7498
        '29852' => '7919'             , # G+7919  U+749C  uni749C,  CJK UNIFIED IDEOGRAPH-749C
        '29854' => '4895'             , # G+4895  U+749E  uni749E,  CJK UNIFIED IDEOGRAPH-749E
        '29855' => '3683'             , # G+3683  U+749F  uni749F,  CJK UNIFIED IDEOGRAPH-749F
        '29857' => '6994'             , # G+6994  U+74A1  uni74A1,  CJK UNIFIED IDEOGRAPH-74A1
        '29859' => '4098'             , # G+4098  U+74A3  uni74A3,  CJK UNIFIED IDEOGRAPH-74A3
        '29861' => '3684'             , # G+3684  U+74A5  uni74A5,  CJK UNIFIED IDEOGRAPH-74A5
        '29863' => '5025'             , # G+5025  U+74A7  uni74A7,  CJK UNIFIED IDEOGRAPH-74A7
        '29864' => '7065'             , # G+7065  U+74A8  uni74A8,  CJK UNIFIED IDEOGRAPH-74A8
        '29866' => '6804'             , # G+6804  U+74AA  uni74AA,  CJK UNIFIED IDEOGRAPH-74AA
        '29872' => '7894'             , # G+7894  U+74B0  uni74B0,  CJK UNIFIED IDEOGRAPH-74B0
        '29874' => '5594'             , # G+5594  U+74B2  uni74B2,  CJK UNIFIED IDEOGRAPH-74B2
        '29877' => '5939'             , # G+5939  U+74B5  uni74B5,  CJK UNIFIED IDEOGRAPH-74B5
        '29881' => '5637'             , # G+5637  U+74B9  uni74B9,  CJK UNIFIED IDEOGRAPH-74B9
        '29885' => '5363'             , # G+5363  U+74BD  uni74BD,  CJK UNIFIED IDEOGRAPH-74BD
        '29887' => '5432'             , # G+5432  U+74BF  uni74BF,  CJK UNIFIED IDEOGRAPH-74BF
        '29894' => '7021'             , # G+7021  U+74C6  uni74C6,  CJK UNIFIED IDEOGRAPH-74C6
        '29898' => '3685'             , # G+3685  U+74CA  uni74CA,  CJK UNIFIED IDEOGRAPH-74CA
        '29903' => '4540'             , # G+4540  U+74CF  uni74CF,  CJK UNIFIED IDEOGRAPH-74CF
        '29908' => '6020'             , # G+6020  U+74D4  uni74D4,  CJK UNIFIED IDEOGRAPH-74D4
        '29912' => '3827'             , # G+3827  U+74D8  uni74D8,  CJK UNIFIED IDEOGRAPH-74D8
        '29914' => '7066'             , # G+7066  U+74DA  uni74DA,  CJK UNIFIED IDEOGRAPH-74DA
        '29916' => '3806'             , # G+3806  U+74DC  uni74DC,  CJK UNIFIED IDEOGRAPH-74DC
        '29920' => '7824'             , # G+7824  U+74E0  uni74E0,  CJK UNIFIED IDEOGRAPH-74E0
        '29922' => '7583'             , # G+7583  U+74E2  uni74E2,  CJK UNIFIED IDEOGRAPH-74E2
        '29923' => '7494'             , # G+7494  U+74E3  uni74E3,  CJK UNIFIED IDEOGRAPH-74E3
        '29926' => '6100'             , # G+6100  U+74E6  uni74E6,  CJK UNIFIED IDEOGRAPH-74E6
        '29934' => '6092'             , # G+6092  U+74EE  uni74EE,  CJK UNIFIED IDEOGRAPH-74EE
        '29943' => '6497'             , # G+6497  U+74F7  uni74F7,  CJK UNIFIED IDEOGRAPH-74F7
        '29953' => '5051'             , # G+5051  U+7501  uni7501,  CJK UNIFIED IDEOGRAPH-7501
        '29956' => '3640'             , # G+3640  U+7504  uni7504,  CJK UNIFIED IDEOGRAPH-7504
        '29969' => '6936'             , # G+6936  U+7511  uni7511,  CJK UNIFIED IDEOGRAPH-7511
        '29973' => '6093'             , # G+6093  U+7515  uni7515,  CJK UNIFIED IDEOGRAPH-7515
        '29976' => '3521'             , # G+3521  U+7518  uni7518,  CJK UNIFIED IDEOGRAPH-7518
        '29978' => '5768'             , # G+5768  U+751A  uni751A,  CJK UNIFIED IDEOGRAPH-751A
        '29979' => '7181'             , # G+7181  U+751B  uni751B,  CJK UNIFIED IDEOGRAPH-751B
        '29983' => '5370'             , # G+5370  U+751F  uni751F,  CJK UNIFIED IDEOGRAPH-751F
        '29987' => '5309'             , # G+5309  U+7523  uni7523,  CJK UNIFIED IDEOGRAPH-7523
        '29989' => '5371'             , # G+5371  U+7525  uni7525,  CJK UNIFIED IDEOGRAPH-7525
        '29990' => '5518'             , # G+5518  U+7526  uni7526,  CJK UNIFIED IDEOGRAPH-7526
        '29992' => '6189'             , # G+6189  U+7528  uni7528,  CJK UNIFIED IDEOGRAPH-7528
        '29995' => '5068'             , # G+5068  U+752B  uni752B,  CJK UNIFIED IDEOGRAPH-752B
        '29996' => '6190'             , # G+6190  U+752C  uni752C,  CJK UNIFIED IDEOGRAPH-752C
        '30000' => '6664'             , # G+6664  U+7530  uni7530,  CJK UNIFIED IDEOGRAPH-7530
        '30001' => '6338'             , # G+6338  U+7531  uni7531,  CJK UNIFIED IDEOGRAPH-7531
        '30002' => '3532'             , # G+3532  U+7532  uni7532,  CJK UNIFIED IDEOGRAPH-7532
        '30003' => '5745'             , # G+5745  U+7533  uni7533,  CJK UNIFIED IDEOGRAPH-7533
        '30007' => '4149'             , # G+4149  U+7537  uni7537,  CJK UNIFIED IDEOGRAPH-7537
        '30008' => '6665'             , # G+6665  U+7538  uni7538,  CJK UNIFIED IDEOGRAPH-7538
        '30010' => '6736'             , # G+6736  U+753A  uni753A,  CJK UNIFIED IDEOGRAPH-753A
        '30023' => '4011'             , # G+4011  U+7547  uni7547,  CJK UNIFIED IDEOGRAPH-7547
        '30028' => '3717'             , # G+3717  U+754C  uni754C,  CJK UNIFIED IDEOGRAPH-754C
        '30031' => '6139'             , # G+6139  U+754F  uni754F,  CJK UNIFIED IDEOGRAPH-754F
        '30033' => '6666'             , # G+6666  U+7551  uni7551,  CJK UNIFIED IDEOGRAPH-7551
        '30035' => '4235'             , # G+4235  U+7553  uni7553,  CJK UNIFIED IDEOGRAPH-7553
        '30036' => '4917'             , # G+4917  U+7554  uni7554,  CJK UNIFIED IDEOGRAPH-7554
        '30041' => '4586'             , # G+4586  U+7559  uni7559,  CJK UNIFIED IDEOGRAPH-7559
        '30043' => '6995'             , # G+6995  U+755B  uni755B,  CJK UNIFIED IDEOGRAPH-755B
        '30044' => '7289'             , # G+7289  U+755C  uni755C,  CJK UNIFIED IDEOGRAPH-755C
        '30045' => '4825'             , # G+4825  U+755D  uni755D,  CJK UNIFIED IDEOGRAPH-755D
        '30050' => '7609'             , # G+7609  U+7562  uni7562,  CJK UNIFIED IDEOGRAPH-7562
        '30053' => '4426'             , # G+4426  U+7565  uni7565,  CJK UNIFIED IDEOGRAPH-7565
        '30054' => '8008'             , # G+8008  U+7566  uni7566,  CJK UNIFIED IDEOGRAPH-7566
        '30058' => '5000'             , # G+5000  U+756A  uni756A,  CJK UNIFIED IDEOGRAPH-756A
        '30063' => '6914'             , # G+6914  U+756F  uni756F,  CJK UNIFIED IDEOGRAPH-756F
        '30064' => '6424'             , # G+6424  U+7570  uni7570,  CJK UNIFIED IDEOGRAPH-7570
        '30069' => '7868'             , # G+7868  U+7575  uni7575,  CJK UNIFIED IDEOGRAPH-7575
        '30070' => '4246'             , # G+4246  U+7576  uni7576,  CJK UNIFIED IDEOGRAPH-7576
        '30072' => '4099'             , # G+4099  U+7578  uni7578,  CJK UNIFIED IDEOGRAPH-7578
        '30074' => '3546'             , # G+3546  U+757A  uni757A,  CJK UNIFIED IDEOGRAPH-757A
        '30079' => '4100'             , # G+4100  U+757F  uni757F,  CJK UNIFIED IDEOGRAPH-757F
        '30086' => '3547'             , # G+3547  U+7586  uni7586,  CJK UNIFIED IDEOGRAPH-7586
        '30087' => '6883'             , # G+6883  U+7587  uni7587,  CJK UNIFIED IDEOGRAPH-7587
        '30090' => '7192'             , # G+7192  U+758A  uni758A,  CJK UNIFIED IDEOGRAPH-758A
        '30091' => '7610'             , # G+7610  U+758B  uni758B,  CJK UNIFIED IDEOGRAPH-758B
        '30094' => '5520'             , # G+5520  U+758E  uni758E,  CJK UNIFIED IDEOGRAPH-758E
        '30095' => '5519'             , # G+5519  U+758F  uni758F,  CJK UNIFIED IDEOGRAPH-758F
        '30097' => '6403'             , # G+6403  U+7591  uni7591,  CJK UNIFIED IDEOGRAPH-7591
        '30109' => '5310'             , # G+5310  U+759D  uni759D,  CJK UNIFIED IDEOGRAPH-759D
        '30117' => '3571'             , # G+3571  U+75A5  uni75A5,  CJK UNIFIED IDEOGRAPH-75A5
        '30123' => '5951'             , # G+5951  U+75AB  uni75AB,  CJK UNIFIED IDEOGRAPH-75AB
        '30129' => '7558'             , # G+7558  U+75B1  uni75B1,  CJK UNIFIED IDEOGRAPH-75B1
        '30130' => '7599'             , # G+7599  U+75B2  uni75B2,  CJK UNIFIED IDEOGRAPH-75B2
        '30131' => '3522'             , # G+3522  U+75B3  uni75B3,  CJK UNIFIED IDEOGRAPH-75B3
        '30133' => '6498'             , # G+6498  U+75B5  uni75B5,  CJK UNIFIED IDEOGRAPH-75B5
        '30136' => '4215'             , # G+4215  U+75B8  uni75B8,  CJK UNIFIED IDEOGRAPH-75B8
        '30137' => '6996'             , # G+6996  U+75B9  uni75B9,  CJK UNIFIED IDEOGRAPH-75B9
        '30140' => '4342'             , # G+4342  U+75BC  uni75BC,  CJK UNIFIED IDEOGRAPH-75BC
        '30141' => '6605'             , # G+6605  U+75BD  uni75BD,  CJK UNIFIED IDEOGRAPH-75BD
        '30142' => '7022'             , # G+7022  U+75BE  uni75BE,  CJK UNIFIED IDEOGRAPH-75BE
        '30146' => '3453'             , # G+3453  U+75C2  uni75C2,  CJK UNIFIED IDEOGRAPH-75C2
        '30149' => '5052'             , # G+5052  U+75C5  uni75C5,  CJK UNIFIED IDEOGRAPH-75C5
        '30151' => '6937'             , # G+6937  U+75C7  uni75C7,  CJK UNIFIED IDEOGRAPH-75C7
        '30157' => '6425'             , # G+6425  U+75CD  uni75CD,  CJK UNIFIED IDEOGRAPH-75CD
        '30162' => '5892'             , # G+5892  U+75D2  uni75D2,  CJK UNIFIED IDEOGRAPH-75D2
        '30164' => '7346'             , # G+7346  U+75D4  uni75D4,  CJK UNIFIED IDEOGRAPH-75D4
        '30165' => '8022'             , # G+8022  U+75D5  uni75D5,  CJK UNIFIED IDEOGRAPH-75D5
        '30168' => '4352'             , # G+4352  U+75D8  uni75D8,  CJK UNIFIED IDEOGRAPH-75D8
        '30169' => '3686'             , # G+3686  U+75D9  uni75D9,  CJK UNIFIED IDEOGRAPH-75D9
        '30171' => '7455'             , # G+7455  U+75DB  uni75DB,  CJK UNIFIED IDEOGRAPH-75DB
        '30178' => '4627'             , # G+4627  U+75E2  uni75E2,  CJK UNIFIED IDEOGRAPH-75E2
        '30192' => '4226'             , # G+4226  U+75F0  uni75F0,  CJK UNIFIED IDEOGRAPH-75F0
        '30194' => '4658'             , # G+4658  U+75F2  uni75F2,  CJK UNIFIED IDEOGRAPH-75F2
        '30196' => '7347'             , # G+7347  U+75F4  uni75F4,  CJK UNIFIED IDEOGRAPH-75F4
        '30202' => '5202'             , # G+5202  U+75FA  uni75FA,  CJK UNIFIED IDEOGRAPH-75FA
        '30204' => '3746'             , # G+3746  U+75FC  uni75FC,  CJK UNIFIED IDEOGRAPH-75FC
        '30208' => '5906'             , # G+5906  U+7600  uni7600,  CJK UNIFIED IDEOGRAPH-7600
        '30221' => '5893'             , # G+5893  U+760D  uni760D,  CJK UNIFIED IDEOGRAPH-760D
        '30233' => '5521'             , # G+5521  U+7619  uni7619,  CJK UNIFIED IDEOGRAPH-7619
        '30239' => '6085'             , # G+6085  U+761F  uni761F,  CJK UNIFIED IDEOGRAPH-761F
        '30240' => '7143'             , # G+7143  U+7620  uni7620,  CJK UNIFIED IDEOGRAPH-7620
        '30241' => '7108'             , # G+7108  U+7621  uni7621,  CJK UNIFIED IDEOGRAPH-7621
        '30242' => '4918'             , # G+4918  U+7622  uni7622,  CJK UNIFIED IDEOGRAPH-7622
        '30244' => '4587'             , # G+4587  U+7624  uni7624,  CJK UNIFIED IDEOGRAPH-7624
        '30246' => '5595'             , # G+5595  U+7626  uni7626,  CJK UNIFIED IDEOGRAPH-7626
        '30267' => '4570'             , # G+4570  U+763B  uni763B,  CJK UNIFIED IDEOGRAPH-763B
        '30274' => '4557'             , # G+4557  U+7642  uni7642,  CJK UNIFIED IDEOGRAPH-7642
        '30284' => '5826'             , # G+5826  U+764C  uni764C,  CJK UNIFIED IDEOGRAPH-764C
        '30286' => '3489'             , # G+3489  U+764E  uni764E,  CJK UNIFIED IDEOGRAPH-764E
        '30290' => '6339'             , # G+6339  U+7652  uni7652,  CJK UNIFIED IDEOGRAPH-7652
        '30294' => '5026'             , # G+5026  U+7656  uni7656,  CJK UNIFIED IDEOGRAPH-7656
        '30305' => '7348'             , # G+7348  U+7661  uni7661,  CJK UNIFIED IDEOGRAPH-7661
        '30308' => '6688'             , # G+6688  U+7664  uni7664,  CJK UNIFIED IDEOGRAPH-7664
        '30313' => '4376'             , # G+4376  U+7669  uni7669,  CJK UNIFIED IDEOGRAPH-7669
        '30316' => '5433'             , # G+5433  U+766C  uni766C,  CJK UNIFIED IDEOGRAPH-766C
        '30320' => '6094'             , # G+6094  U+7670  uni7670,  CJK UNIFIED IDEOGRAPH-7670
        '30322' => '6667'             , # G+6667  U+7672  uni7672,  CJK UNIFIED IDEOGRAPH-7672
        '30328' => '3718'             , # G+3718  U+7678  uni7678,  CJK UNIFIED IDEOGRAPH-7678
        '30331' => '4368'             , # G+4368  U+767B  uni767B,  CJK UNIFIED IDEOGRAPH-767B
        '30332' => '4935'             , # G+4935  U+767C  uni767C,  CJK UNIFIED IDEOGRAPH-767C
        '30333' => '4993'             , # G+4993  U+767D  uni767D,  CJK UNIFIED IDEOGRAPH-767D
        '30334' => '4994'             , # G+4994  U+767E  uni767E,  CJK UNIFIED IDEOGRAPH-767E
        '30340' => '6627'             , # G+6627  U+7684  uni7684,  CJK UNIFIED IDEOGRAPH-7684
        '30342' => '3572'             , # G+3572  U+7686  uni7686,  CJK UNIFIED IDEOGRAPH-7686
        '30343' => '7920'             , # G+7920  U+7687  uni7687,  CJK UNIFIED IDEOGRAPH-7687
        '30350' => '3880'             , # G+3880  U+768E  uni768E,  CJK UNIFIED IDEOGRAPH-768E
        '30352' => '3747'             , # G+3747  U+7690  uni7690,  CJK UNIFIED IDEOGRAPH-7690
        '30355' => '7825'             , # G+7825  U+7693  uni7693,  CJK UNIFIED IDEOGRAPH-7693
        '30382' => '7600'             , # G+7600  U+76AE  uni76AE,  CJK UNIFIED IDEOGRAPH-76AE
        '30394' => '7272'             , # G+7272  U+76BA  uni76BA,  CJK UNIFIED IDEOGRAPH-76BA
        '30399' => '4755'             , # G+4755  U+76BF  uni76BF,  CJK UNIFIED IDEOGRAPH-76BF
        '30402' => '6213'             , # G+6213  U+76C2  uni76C2,  CJK UNIFIED IDEOGRAPH-76C2
        '30403' => '4978'             , # G+4978  U+76C3  uni76C3,  CJK UNIFIED IDEOGRAPH-76C3
        '30406' => '5164'             , # G+5164  U+76C6  uni76C6,  CJK UNIFIED IDEOGRAPH-76C6
        '30408' => '6021'             , # G+6021  U+76C8  uni76C8,  CJK UNIFIED IDEOGRAPH-76C8
        '30410' => '6438'             , # G+6438  U+76CA  uni76CA,  CJK UNIFIED IDEOGRAPH-76CA
        '30418' => '7665'             , # G+7665  U+76D2  uni76D2,  CJK UNIFIED IDEOGRAPH-76D2
        '30422' => '3573'             , # G+3573  U+76D6  uni76D6,  CJK UNIFIED IDEOGRAPH-76D6
        '30427' => '5484'             , # G+5484  U+76DB  uni76DB,  CJK UNIFIED IDEOGRAPH-76DB
        '30428' => '4293'             , # G+4293  U+76DC  uni76DC,  CJK UNIFIED IDEOGRAPH-76DC
        '30430' => '6526'             , # G+6526  U+76DE  uni76DE,  CJK UNIFIED IDEOGRAPH-76DE
        '30431' => '4731'             , # G+4731  U+76DF  uni76DF,  CJK UNIFIED IDEOGRAPH-76DF
        '30433' => '6997'             , # G+6997  U+76E1  uni76E1,  CJK UNIFIED IDEOGRAPH-76E1
        '30435' => '3523'             , # G+3523  U+76E3  uni76E3,  CJK UNIFIED IDEOGRAPH-76E3
        '30436' => '4919'             , # G+4919  U+76E4  uni76E4,  CJK UNIFIED IDEOGRAPH-76E4
        '30439' => '4518'             , # G+4518  U+76E7  uni76E7,  CJK UNIFIED IDEOGRAPH-76E7
        '30446' => '4791'             , # G+4791  U+76EE  uni76EE,  CJK UNIFIED IDEOGRAPH-76EE
        '30450' => '4730'             , # G+4730  U+76F2  uni76F2,  CJK UNIFIED IDEOGRAPH-76F2
        '30452' => '6975'             , # G+6975  U+76F4  uni76F4,  CJK UNIFIED IDEOGRAPH-76F4
        '30456' => '5352'             , # G+5352  U+76F8  uni76F8,  CJK UNIFIED IDEOGRAPH-76F8
        '30460' => '4920'             , # G+4920  U+76FC  uni76FC,  CJK UNIFIED IDEOGRAPH-76FC
        '30462' => '5652'             , # G+5652  U+76FE  uni76FE,  CJK UNIFIED IDEOGRAPH-76FE
        '30465' => '5485'             , # G+5485  U+7701  uni7701,  CJK UNIFIED IDEOGRAPH-7701
        '30468' => '4740'             , # G+4740  U+7704  uni7704,  CJK UNIFIED IDEOGRAPH-7704
        '30472' => '7421'             , # G+7421  U+7708  uni7708,  CJK UNIFIED IDEOGRAPH-7708
        '30473' => '4862'             , # G+4862  U+7709  uni7709,  CJK UNIFIED IDEOGRAPH-7709
        '30475' => '3490'             , # G+3490  U+770B  uni770B,  CJK UNIFIED IDEOGRAPH-770B
        '30494' => '6998'             , # G+6998  U+771E  uni771E,  CJK UNIFIED IDEOGRAPH-771E
        '30496' => '4741'             , # G+4741  U+7720  uni7720,  CJK UNIFIED IDEOGRAPH-7720
        '30505' => '7744'             , # G+7744  U+7729  uni7729,  CJK UNIFIED IDEOGRAPH-7729
        '30519' => '3978'             , # G+3978  U+7737  uni7737,  CJK UNIFIED IDEOGRAPH-7737
        '30520' => '4780'             , # G+4780  U+7738  uni7738,  CJK UNIFIED IDEOGRAPH-7738
        '30522' => '6805'             , # G+6805  U+773A  uni773A,  CJK UNIFIED IDEOGRAPH-773A
        '30524' => '5812'             , # G+5812  U+773C  uni773C,  CJK UNIFIED IDEOGRAPH-773C
        '30528' => '7057'             , # G+7057  U+7740  uni7740,  CJK UNIFIED IDEOGRAPH-7740
        '30541' => '7745'             , # G+7745  U+774D  uni774D,  CJK UNIFIED IDEOGRAPH-774D
        '30555' => '6737'             , # G+6737  U+775B  uni775B,  CJK UNIFIED IDEOGRAPH-775B
        '30561' => '5596'             , # G+5596  U+7761  uni7761,  CJK UNIFIED IDEOGRAPH-7761
        '30563' => '4314'             , # G+4314  U+7763  uni7763,  CJK UNIFIED IDEOGRAPH-7763
        '30566' => '4792'             , # G+4792  U+7766  uni7766,  CJK UNIFIED IDEOGRAPH-7766
        '30571' => '7193'             , # G+7193  U+776B  uni776B,  CJK UNIFIED IDEOGRAPH-776B
        '30585' => '4294'             , # G+4294  U+7779  uni7779,  CJK UNIFIED IDEOGRAPH-7779
        '30590' => '3748'             , # G+3748  U+777E  uni777E,  CJK UNIFIED IDEOGRAPH-777E
        '30591' => '6037'             , # G+6037  U+777F  uni777F,  CJK UNIFIED IDEOGRAPH-777F
        '30603' => '6999'             , # G+6999  U+778B  uni778B,  CJK UNIFIED IDEOGRAPH-778B
        '30609' => '4756'             , # G+4756  U+7791  uni7791,  CJK UNIFIED IDEOGRAPH-7791
        '30622' => '4682'             , # G+4682  U+779E  uni779E,  CJK UNIFIED IDEOGRAPH-779E
        '30629' => '5038'             , # G+5038  U+77A5  uni77A5,  CJK UNIFIED IDEOGRAPH-77A5
        '30636' => '5653'             , # G+5653  U+77AC  uni77AC,  CJK UNIFIED IDEOGRAPH-77AC
        '30637' => '4558'             , # G+4558  U+77AD  uni77AD,  CJK UNIFIED IDEOGRAPH-77AD
        '30640' => '3524'             , # G+3524  U+77B0  uni77B0,  CJK UNIFIED IDEOGRAPH-77B0
        '30643' => '4343'             , # G+4343  U+77B3  uni77B3,  CJK UNIFIED IDEOGRAPH-77B3
        '30651' => '7182'             , # G+7182  U+77BB  uni77BB,  CJK UNIFIED IDEOGRAPH-77BB
        '30652' => '3621'             , # G+3621  U+77BC  uni77BC,  CJK UNIFIED IDEOGRAPH-77BC
        '30655' => '3925'             , # G+3925  U+77BF  uni77BF,  CJK UNIFIED IDEOGRAPH-77BF
        '30679' => '7244'             , # G+7244  U+77D7  uni77D7,  CJK UNIFIED IDEOGRAPH-77D7
        '30683' => '4781'             , # G+4781  U+77DB  uni77DB,  CJK UNIFIED IDEOGRAPH-77DB
        '30684' => '4060'             , # G+4060  U+77DC  uni77DC,  CJK UNIFIED IDEOGRAPH-77DC
        '30690' => '5709'             , # G+5709  U+77E2  uni77E2,  CJK UNIFIED IDEOGRAPH-77E2
        '30691' => '6404'             , # G+6404  U+77E3  uni77E3,  CJK UNIFIED IDEOGRAPH-77E3
        '30693' => '6960'             , # G+6960  U+77E5  uni77E5,  CJK UNIFIED IDEOGRAPH-77E5
        '30697' => '3926'             , # G+3926  U+77E9  uni77E9,  CJK UNIFIED IDEOGRAPH-77E9
        '30701' => '4204'             , # G+4204  U+77ED  uni77ED,  CJK UNIFIED IDEOGRAPH-77ED
        '30702' => '6134'             , # G+6134  U+77EE  uni77EE,  CJK UNIFIED IDEOGRAPH-77EE
        '30703' => '3881'             , # G+3881  U+77EF  uni77EF,  CJK UNIFIED IDEOGRAPH-77EF
        '30707' => '5413'             , # G+5413  U+77F3  uni77F3,  CJK UNIFIED IDEOGRAPH-77F3
        '30722' => '5277'             , # G+5277  U+7802  uni7802,  CJK UNIFIED IDEOGRAPH-7802
        '30738' => '5203'             , # G+5203  U+7812  uni7812,  CJK UNIFIED IDEOGRAPH-7812
        '30757' => '6961'             , # G+6961  U+7825  uni7825,  CJK UNIFIED IDEOGRAPH-7825
        '30758' => '7120'             , # G+7120  U+7826  uni7826,  CJK UNIFIED IDEOGRAPH-7826
        '30759' => '7373'             , # G+7373  U+7827  uni7827,  CJK UNIFIED IDEOGRAPH-7827
        '30764' => '4652'             , # G+4652  U+782C  uni782C,  CJK UNIFIED IDEOGRAPH-782C
        '30770' => '7559'             , # G+7559  U+7832  uni7832,  CJK UNIFIED IDEOGRAPH-7832
        '30772' => '7485'             , # G+7485  U+7834  uni7834,  CJK UNIFIED IDEOGRAPH-7834
        '30789' => '4000'             , # G+4000  U+7845  uni7845,  CJK UNIFIED IDEOGRAPH-7845
        '30799' => '5976'             , # G+5976  U+784F  uni784F,  CJK UNIFIED IDEOGRAPH-784F
        '30813' => '7226'             , # G+7226  U+785D  uni785D,  CJK UNIFIED IDEOGRAPH-785D
        '30827' => '4588'             , # G+4588  U+786B  uni786B,  CJK UNIFIED IDEOGRAPH-786B
        '30828' => '3687'             , # G+3687  U+786C  uni786C,  CJK UNIFIED IDEOGRAPH-786C
        '30831' => '5977'             , # G+5977  U+786F  uni786F,  CJK UNIFIED IDEOGRAPH-786F
        '30844' => '5178'             , # G+5178  U+787C  uni787C,  CJK UNIFIED IDEOGRAPH-787C
        '30849' => '4101'             , # G+4101  U+7881  uni7881,  CJK UNIFIED IDEOGRAPH-7881
        '30855' => '6738'             , # G+6738  U+7887  uni7887,  CJK UNIFIED IDEOGRAPH-7887
        '30860' => '4528'             , # G+4528  U+788C  uni788C,  CJK UNIFIED IDEOGRAPH-788C
        '30861' => '5847'             , # G+5847  U+788D  uni788D,  CJK UNIFIED IDEOGRAPH-788D
        '30862' => '5563'             , # G+5563  U+788E  uni788E,  CJK UNIFIED IDEOGRAPH-788E
        '30865' => '5204'             , # G+5204  U+7891  uni7891,  CJK UNIFIED IDEOGRAPH-7891
        '30871' => '6116'             , # G+6116  U+7897  uni7897,  CJK UNIFIED IDEOGRAPH-7897
        '30883' => '3504'             , # G+3504  U+78A3  uni78A3,  CJK UNIFIED IDEOGRAPH-78A3
        '30887' => '5027'             , # G+5027  U+78A7  uni78A7,  CJK UNIFIED IDEOGRAPH-78A7
        '30889' => '5414'             , # G+5414  U+78A9  uni78A9,  CJK UNIFIED IDEOGRAPH-78A9
        '30906' => '7879'             , # G+7879  U+78BA  uni78BA,  CJK UNIFIED IDEOGRAPH-78BA
        '30907' => '7880'             , # G+7880  U+78BB  uni78BB,  CJK UNIFIED IDEOGRAPH-78BB
        '30908' => '4659'             , # G+4659  U+78BC  uni78BC,  CJK UNIFIED IDEOGRAPH-78BC
        '30913' => '6499'             , # G+6499  U+78C1  uni78C1,  CJK UNIFIED IDEOGRAPH-78C1
        '30917' => '4956'             , # G+4956  U+78C5  uni78C5,  CJK UNIFIED IDEOGRAPH-78C5
        '30922' => '4546'             , # G+4546  U+78CA  uni78CA,  CJK UNIFIED IDEOGRAPH-78CA
        '30923' => '7050'             , # G+7050  U+78CB  uni78CB,  CJK UNIFIED IDEOGRAPH-78CB
        '30926' => '3719'             , # G+3719  U+78CE  uni78CE,  CJK UNIFIED IDEOGRAPH-78CE
        '30928' => '4921'             , # G+4921  U+78D0  uni78D0,  CJK UNIFIED IDEOGRAPH-78D0
        '30952' => '4660'             , # G+4660  U+78E8  uni78E8,  CJK UNIFIED IDEOGRAPH-78E8
        '30956' => '3688'             , # G+3688  U+78EC  uni78EC,  CJK UNIFIED IDEOGRAPH-78EC
        '30959' => '4102'             , # G+4102  U+78EF  uni78EF,  CJK UNIFIED IDEOGRAPH-78EF
        '30965' => '3491'             , # G+3491  U+78F5  uni78F5,  CJK UNIFIED IDEOGRAPH-78F5
        '30971' => '4922'             , # G+4922  U+78FB  uni78FB,  CJK UNIFIED IDEOGRAPH-78FB
        '30977' => '7227'             , # G+7227  U+7901  uni7901,  CJK UNIFIED IDEOGRAPH-7901
        '30990' => '7228'             , # G+7228  U+790E  uni790E,  CJK UNIFIED IDEOGRAPH-790E
        '30998' => '5940'             , # G+5940  U+7916  uni7916,  CJK UNIFIED IDEOGRAPH-7916
        '31018' => '4450'             , # G+4450  U+792A  uni792A,  CJK UNIFIED IDEOGRAPH-792A
        '31019' => '4462'             , # G+4462  U+792B  uni792B,  CJK UNIFIED IDEOGRAPH-792B
        '31020' => '4923'             , # G+4923  U+792C  uni792C,  CJK UNIFIED IDEOGRAPH-792C
        '31034' => '5710'             , # G+5710  U+793A  uni793A,  CJK UNIFIED IDEOGRAPH-793A
        '31038' => '5278'             , # G+5278  U+793E  uni793E,  CJK UNIFIED IDEOGRAPH-793E
        '31040' => '5279'             , # G+5279  U+7940  uni7940,  CJK UNIFIED IDEOGRAPH-7940
        '31041' => '4103'             , # G+4103  U+7941  uni7941,  CJK UNIFIED IDEOGRAPH-7941
        '31047' => '4104'             , # G+4104  U+7947  uni7947,  CJK UNIFIED IDEOGRAPH-7947
        '31048' => '4105'             , # G+4105  U+7948  uni7948,  CJK UNIFIED IDEOGRAPH-7948
        '31049' => '6962'             , # G+6962  U+7949  uni7949,  CJK UNIFIED IDEOGRAPH-7949
        '31056' => '6214'             , # G+6214  U+7950  uni7950,  CJK UNIFIED IDEOGRAPH-7950
        '31062' => '6806'             , # G+6806  U+7956  uni7956,  CJK UNIFIED IDEOGRAPH-7956
        '31063' => '6963'             , # G+6963  U+7957  uni7957,  CJK UNIFIED IDEOGRAPH-7957
        '31066' => '6807'             , # G+6807  U+795A  uni795A,  CJK UNIFIED IDEOGRAPH-795A
        '31067' => '3595'             , # G+3595  U+795B  uni795B,  CJK UNIFIED IDEOGRAPH-795B
        '31068' => '7826'             , # G+7826  U+795C  uni795C,  CJK UNIFIED IDEOGRAPH-795C
        '31069' => '7290'             , # G+7290  U+795D  uni795D,  CJK UNIFIED IDEOGRAPH-795D
        '31070' => '5746'             , # G+5746  U+795E  uni795E,  CJK UNIFIED IDEOGRAPH-795E
        '31072' => '5280'             , # G+5280  U+7960  uni7960,  CJK UNIFIED IDEOGRAPH-7960
        '31077' => '5353'             , # G+5353  U+7965  uni7965,  CJK UNIFIED IDEOGRAPH-7965
        '31080' => '7584'             , # G+7584  U+7968  uni7968,  CJK UNIFIED IDEOGRAPH-7968
        '31085' => '6769'             , # G+6769  U+796D  uni796D,  CJK UNIFIED IDEOGRAPH-796D
        '31098' => '4106'             , # G+4106  U+797A  uni797A,  CJK UNIFIED IDEOGRAPH-797A
        '31103' => '4529'             , # G+4529  U+797F  uni797F,  CJK UNIFIED IDEOGRAPH-797F
        '31105' => '4044'             , # G+4044  U+7981  uni7981,  CJK UNIFIED IDEOGRAPH-7981
        '31117' => '7869'             , # G+7869  U+798D  uni798D,  CJK UNIFIED IDEOGRAPH-798D
        '31118' => '6739'             , # G+6739  U+798E  uni798E,  CJK UNIFIED IDEOGRAPH-798E
        '31119' => '5081'             , # G+5081  U+798F  uni798F,  CJK UNIFIED IDEOGRAPH-798F
        '31121' => '6215'             , # G+6215  U+7991  uni7991,  CJK UNIFIED IDEOGRAPH-7991
        '31142' => '5907'             , # G+5907  U+79A6  uni79A6,  CJK UNIFIED IDEOGRAPH-79A6
        '31143' => '8052'             , # G+8052  U+79A7  uni79A7,  CJK UNIFIED IDEOGRAPH-79A7
        '31146' => '5434'             , # G+5434  U+79AA  uni79AA,  CJK UNIFIED IDEOGRAPH-79AA
        '31150' => '4508'             , # G+4508  U+79AE  uni79AE,  CJK UNIFIED IDEOGRAPH-79AE
        '31153' => '4295'             , # G+4295  U+79B1  uni79B1,  CJK UNIFIED IDEOGRAPH-79B1
        '31155' => '5894'             , # G+5894  U+79B3  uni79B3,  CJK UNIFIED IDEOGRAPH-79B3
        '31161' => '6216'             , # G+6216  U+79B9  uni79B9,  CJK UNIFIED IDEOGRAPH-79B9
        '31165' => '4045'             , # G+4045  U+79BD  uni79BD,  CJK UNIFIED IDEOGRAPH-79BD
        '31166' => '7870'             , # G+7870  U+79BE  uni79BE,  CJK UNIFIED IDEOGRAPH-79BE
        '31167' => '4315'             , # G+4315  U+79BF  uni79BF,  CJK UNIFIED IDEOGRAPH-79BF
        '31168' => '5597'             , # G+5597  U+79C0  uni79C0,  CJK UNIFIED IDEOGRAPH-79C0
        '31169' => '5281'             , # G+5281  U+79C1  uni79C1,  CJK UNIFIED IDEOGRAPH-79C1
        '31177' => '5053'             , # G+5053  U+79C9  uni79C9,  CJK UNIFIED IDEOGRAPH-79C9
        '31178' => '4162'             , # G+4162  U+79CA  uni79CA,  CJK UNIFIED IDEOGRAPH-79CA
        '31179' => '7273'             , # G+7273  U+79CB  uni79CB,  CJK UNIFIED IDEOGRAPH-79CB
        '31185' => '3807'             , # G+3807  U+79D1  uni79D1,  CJK UNIFIED IDEOGRAPH-79D1
        '31186' => '7229'             , # G+7229  U+79D2  uni79D2,  CJK UNIFIED IDEOGRAPH-79D2
        '31189' => '5205'             , # G+5205  U+79D5  uni79D5,  CJK UNIFIED IDEOGRAPH-79D5
        '31192' => '5206'             , # G+5206  U+79D8  uni79D8,  CJK UNIFIED IDEOGRAPH-79D8
        '31199' => '6808'             , # G+6808  U+79DF  uni79DF,  CJK UNIFIED IDEOGRAPH-79DF
        '31204' => '7377'             , # G+7377  U+79E4  uni79E4,  CJK UNIFIED IDEOGRAPH-79E4
        '31206' => '7000'             , # G+7000  U+79E6  uni79E6,  CJK UNIFIED IDEOGRAPH-79E6
        '31207' => '5838'             , # G+5838  U+79E7  uni79E7,  CJK UNIFIED IDEOGRAPH-79E7
        '31209' => '7023'             , # G+7023  U+79E9  uni79E9,  CJK UNIFIED IDEOGRAPH-79E9
        '31227' => '6426'             , # G+6426  U+79FB  uni79FB,  CJK UNIFIED IDEOGRAPH-79FB
        '31232' => '8053'             , # G+8053  U+7A00  uni7A00,  CJK UNIFIED IDEOGRAPH-7A00
        '31237' => '5496'             , # G+5496  U+7A05  uni7A05,  CJK UNIFIED IDEOGRAPH-7A05
        '31240' => '3492'             , # G+3492  U+7A08  uni7A08,  CJK UNIFIED IDEOGRAPH-7A08
        '31243' => '6740'             , # G+6740  U+7A0B  uni7A0B,  CJK UNIFIED IDEOGRAPH-7A0B
        '31245' => '7230'             , # G+7230  U+7A0D  uni7A0D,  CJK UNIFIED IDEOGRAPH-7A0D
        '31252' => '6474'             , # G+6474  U+7A14  uni7A14,  CJK UNIFIED IDEOGRAPH-7A14
        '31255' => '7510'             , # G+7510  U+7A17  uni7A17,  CJK UNIFIED IDEOGRAPH-7A17
        '31257' => '6976'             , # G+6976  U+7A19  uni7A19,  CJK UNIFIED IDEOGRAPH-7A19
        '31258' => '7349'             , # G+7349  U+7A1A  uni7A1A,  CJK UNIFIED IDEOGRAPH-7A1A
        '31260' => '4609'             , # G+4609  U+7A1C  uni7A1C,  CJK UNIFIED IDEOGRAPH-7A1C
        '31263' => '7591'             , # G+7591  U+7A1F  uni7A1F,  CJK UNIFIED IDEOGRAPH-7A1F
        '31264' => '6809'             , # G+6809  U+7A20  uni7A20,  CJK UNIFIED IDEOGRAPH-7A20
        '31278' => '6845'             , # G+6845  U+7A2E  uni7A2E,  CJK UNIFIED IDEOGRAPH-7A2E
        '31281' => '7378'             , # G+7378  U+7A31  uni7A31,  CJK UNIFIED IDEOGRAPH-7A31
        '31286' => '6235'             , # G+6235  U+7A36  uni7A36,  CJK UNIFIED IDEOGRAPH-7A36
        '31287' => '6977'             , # G+6977  U+7A37  uni7A37,  CJK UNIFIED IDEOGRAPH-7A37
        '31291' => '4296'             , # G+4296  U+7A3B  uni7A3B,  CJK UNIFIED IDEOGRAPH-7A3B
        '31292' => '3454'             , # G+3454  U+7A3C  uni7A3C,  CJK UNIFIED IDEOGRAPH-7A3C
        '31293' => '3720'             , # G+3720  U+7A3D  uni7A3D,  CJK UNIFIED IDEOGRAPH-7A3D
        '31295' => '3749'             , # G+3749  U+7A3F  uni7A3F,  CJK UNIFIED IDEOGRAPH-7A3F
        '31296' => '3771'             , # G+3771  U+7A40  uni7A40,  CJK UNIFIED IDEOGRAPH-7A40
        '31302' => '4793'             , # G+4793  U+7A46  uni7A46,  CJK UNIFIED IDEOGRAPH-7A46
        '31305' => '7350'             , # G+7350  U+7A49  uni7A49,  CJK UNIFIED IDEOGRAPH-7A49
        '31309' => '6628'             , # G+6628  U+7A4D  uni7A4D,  CJK UNIFIED IDEOGRAPH-7A4D
        '31310' => '6022'             , # G+6022  U+7A4E  uni7A4E,  CJK UNIFIED IDEOGRAPH-7A4E
        '31319' => '5598'             , # G+5598  U+7A57  uni7A57,  CJK UNIFIED IDEOGRAPH-7A57
        '31329' => '5366'             , # G+5366  U+7A61  uni7A61,  CJK UNIFIED IDEOGRAPH-7A61
        '31330' => '6038'             , # G+6038  U+7A62  uni7A62,  CJK UNIFIED IDEOGRAPH-7A62
        '31337' => '6086'             , # G+6086  U+7A69  uni7A69,  CJK UNIFIED IDEOGRAPH-7A69
        '31339' => '7881'             , # G+7881  U+7A6B  uni7A6B,  CJK UNIFIED IDEOGRAPH-7A6B
        '31344' => '5895'             , # G+5895  U+7A70  uni7A70,  CJK UNIFIED IDEOGRAPH-7A70
        '31348' => '7755'             , # G+7755  U+7A74  uni7A74,  CJK UNIFIED IDEOGRAPH-7A74
        '31350' => '3927'             , # G+3927  U+7A76  uni7A76,  CJK UNIFIED IDEOGRAPH-7A76
        '31353' => '3965'             , # G+3965  U+7A79  uni7A79,  CJK UNIFIED IDEOGRAPH-7A79
        '31354' => '3798'             , # G+3798  U+7A7A  uni7A7A,  CJK UNIFIED IDEOGRAPH-7A7A
        '31357' => '6741'             , # G+6741  U+7A7D  uni7A7D,  CJK UNIFIED IDEOGRAPH-7A7D
        '31359' => '7157'             , # G+7157  U+7A7F  uni7A7F,  CJK UNIFIED IDEOGRAPH-7A7F
        '31361' => '4330'             , # G+4330  U+7A81  uni7A81,  CJK UNIFIED IDEOGRAPH-7A81
        '31364' => '7058'             , # G+7058  U+7A84  uni7A84,  CJK UNIFIED IDEOGRAPH-7A84
        '31368' => '6156'             , # G+6156  U+7A88  uni7A88,  CJK UNIFIED IDEOGRAPH-7A88
        '31378' => '7024'             , # G+7024  U+7A92  uni7A92,  CJK UNIFIED IDEOGRAPH-7A92
        '31379' => '7109'             , # G+7109  U+7A93  uni7A93,  CJK UNIFIED IDEOGRAPH-7A93
        '31381' => '6810'             , # G+6810  U+7A95  uni7A95,  CJK UNIFIED IDEOGRAPH-7A95
        '31384' => '3954'             , # G+3954  U+7A98  uni7A98,  CJK UNIFIED IDEOGRAPH-7A98
        '31391' => '3962'             , # G+3962  U+7A9F  uni7A9F,  CJK UNIFIED IDEOGRAPH-7A9F
        '31401' => '6101'             , # G+6101  U+7AA9  uni7AA9,  CJK UNIFIED IDEOGRAPH-7AA9
        '31402' => '6102'             , # G+6102  U+7AAA  uni7AAA,  CJK UNIFIED IDEOGRAPH-7AAA
        '31406' => '3966'             , # G+3966  U+7AAE  uni7AAE,  CJK UNIFIED IDEOGRAPH-7AAE
        '31407' => '6157'             , # G+6157  U+7AAF  uni7AAF,  CJK UNIFIED IDEOGRAPH-7AAF
        '31418' => '4001'             , # G+4001  U+7ABA  uni7ABA,  CJK UNIFIED IDEOGRAPH-7ABA
        '31428' => '7067'             , # G+7067  U+7AC4  uni7AC4,  CJK UNIFIED IDEOGRAPH-7AC4
        '31429' => '4002'             , # G+4002  U+7AC5  uni7AC5,  CJK UNIFIED IDEOGRAPH-7AC5
        '31431' => '4353'             , # G+4353  U+7AC7  uni7AC7,  CJK UNIFIED IDEOGRAPH-7AC7
        '31434' => '6689'             , # G+6689  U+7ACA  uni7ACA,  CJK UNIFIED IDEOGRAPH-7ACA
        '31435' => '4653'             , # G+4653  U+7ACB  uni7ACB,  CJK UNIFIED IDEOGRAPH-7ACB
        '31447' => '4809'             , # G+4809  U+7AD7  uni7AD7,  CJK UNIFIED IDEOGRAPH-7AD7
        '31449' => '7089'             , # G+7089  U+7AD9  uni7AD9,  CJK UNIFIED IDEOGRAPH-7AD9
        '31453' => '5054'             , # G+5054  U+7ADD  uni7ADD,  CJK UNIFIED IDEOGRAPH-7ADD
        '31455' => '3689'             , # G+3689  U+7ADF  uni7ADF,  CJK UNIFIED IDEOGRAPH-7ADF
        '31456' => '6555'             , # G+6555  U+7AE0  uni7AE0,  CJK UNIFIED IDEOGRAPH-7AE0
        '31459' => '6915'             , # G+6915  U+7AE3  uni7AE3,  CJK UNIFIED IDEOGRAPH-7AE3
        '31461' => '4344'             , # G+4344  U+7AE5  uni7AE5,  CJK UNIFIED IDEOGRAPH-7AE5
        '31466' => '5599'             , # G+5599  U+7AEA  uni7AEA,  CJK UNIFIED IDEOGRAPH-7AEA
        '31469' => '3505'             , # G+3505  U+7AED  uni7AED,  CJK UNIFIED IDEOGRAPH-7AED
        '31471' => '4205'             , # G+4205  U+7AEF  uni7AEF,  CJK UNIFIED IDEOGRAPH-7AEF
        '31478' => '3690'             , # G+3690  U+7AF6  uni7AF6,  CJK UNIFIED IDEOGRAPH-7AF6
        '31481' => '6900'             , # G+6900  U+7AF9  uni7AF9,  CJK UNIFIED IDEOGRAPH-7AF9
        '31482' => '7291'             , # G+7291  U+7AFA  uni7AFA,  CJK UNIFIED IDEOGRAPH-7AFA
        '31487' => '3493'             , # G+3493  U+7AFF  uni7AFF,  CJK UNIFIED IDEOGRAPH-7AFF
        '31503' => '7852'             , # G+7852  U+7B0F  uni7B0F,  CJK UNIFIED IDEOGRAPH-7B0F
        '31505' => '5522'             , # G+5522  U+7B11  uni7B11,  CJK UNIFIED IDEOGRAPH-7B11
        '31513' => '5372'             , # G+5372  U+7B19  uni7B19,  CJK UNIFIED IDEOGRAPH-7B19
        '31515' => '6629'             , # G+6629  U+7B1B  uni7B1B,  CJK UNIFIED IDEOGRAPH-7B1B
        '31518' => '7439'             , # G+7439  U+7B1E  uni7B1E,  CJK UNIFIED IDEOGRAPH-7B1E
        '31520' => '4654'             , # G+4654  U+7B20  uni7B20,  CJK UNIFIED IDEOGRAPH-7B20
        '31526' => '5130'             , # G+5130  U+7B26  uni7B26,  CJK UNIFIED IDEOGRAPH-7B26
        '31532' => '6770'             , # G+6770  U+7B2C  uni7B2C,  CJK UNIFIED IDEOGRAPH-7B2C
        '31533' => '4496'             , # G+4496  U+7B2D  uni7B2D,  CJK UNIFIED IDEOGRAPH-7B2D
        '31545' => '5497'             , # G+5497  U+7B39  uni7B39,  CJK UNIFIED IDEOGRAPH-7B39
        '31558' => '7611'             , # G+7611  U+7B46  uni7B46,  CJK UNIFIED IDEOGRAPH-7B46
        '31561' => '4369'             , # G+4369  U+7B49  uni7B49,  CJK UNIFIED IDEOGRAPH-7B49
        '31563' => '4031'             , # G+4031  U+7B4B  uni7B4B,  CJK UNIFIED IDEOGRAPH-7B4B
        '31564' => '6668'             , # G+6668  U+7B4C  uni7B4C,  CJK UNIFIED IDEOGRAPH-7B4C
        '31565' => '5654'             , # G+5654  U+7B4D  uni7B4D,  CJK UNIFIED IDEOGRAPH-7B4D
        '31567' => '5006'             , # G+5006  U+7B4F  uni7B4F,  CJK UNIFIED IDEOGRAPH-7B4F
        '31568' => '3849'             , # G+3849  U+7B50  uni7B50,  CJK UNIFIED IDEOGRAPH-7B50
        '31569' => '7292'             , # G+7292  U+7B51  uni7B51,  CJK UNIFIED IDEOGRAPH-7B51
        '31570' => '7456'             , # G+7456  U+7B52  uni7B52,  CJK UNIFIED IDEOGRAPH-7B52
        '31572' => '4236'             , # G+4236  U+7B54  uni7B54,  CJK UNIFIED IDEOGRAPH-7B54
        '31574' => '7128'             , # G+7128  U+7B56  uni7B56,  CJK UNIFIED IDEOGRAPH-7B56
        '31584' => '4012'             , # G+4012  U+7B60  uni7B60,  CJK UNIFIED IDEOGRAPH-7B60
        '31596' => '5486'             , # G+5486  U+7B6C  uni7B6C,  CJK UNIFIED IDEOGRAPH-7B6C
        '31598' => '5390'             , # G+5390  U+7B6E  uni7B6E,  CJK UNIFIED IDEOGRAPH-7B6E
        '31605' => '5978'             , # G+5978  U+7B75  uni7B75,  CJK UNIFIED IDEOGRAPH-7B75
        '31613' => '6073'             , # G+6073  U+7B7D  uni7B7D,  CJK UNIFIED IDEOGRAPH-7B7D
        '31623' => '3574'             , # G+3574  U+7B87  uni7B87,  CJK UNIFIED IDEOGRAPH-7B87
        '31627' => '6669'             , # G+6669  U+7B8B  uni7B8B,  CJK UNIFIED IDEOGRAPH-7B8B
        '31631' => '6588'             , # G+6588  U+7B8F  uni7B8F,  CJK UNIFIED IDEOGRAPH-7B8F
        '31636' => '4896'             , # G+4896  U+7B94  uni7B94,  CJK UNIFIED IDEOGRAPH-7B94
        '31637' => '4107'             , # G+4107  U+7B95  uni7B95,  CJK UNIFIED IDEOGRAPH-7B95
        '31639' => '5311'             , # G+5311  U+7B97  uni7B97,  CJK UNIFIED IDEOGRAPH-7B97
        '31642' => '7051'             , # G+7051  U+7B9A  uni7B9A,  CJK UNIFIED IDEOGRAPH-7B9A
        '31645' => '3656'             , # G+3656  U+7B9D  uni7B9D,  CJK UNIFIED IDEOGRAPH-7B9D
        '31649' => '3828'             , # G+3828  U+7BA1  uni7BA1,  CJK UNIFIED IDEOGRAPH-7BA1
        '31661' => '6670'             , # G+6670  U+7BAD  uni7BAD,  CJK UNIFIED IDEOGRAPH-7BAD
        '31665' => '5354'             , # G+5354  U+7BB1  uni7BB1,  CJK UNIFIED IDEOGRAPH-7BB1
        '31668' => '6530'             , # G+6530  U+7BB4  uni7BB4,  CJK UNIFIED IDEOGRAPH-7BB4
        '31672' => '6606'             , # G+6606  U+7BB8  uni7BB8,  CJK UNIFIED IDEOGRAPH-7BB8
        '31680' => '6690'             , # G+6690  U+7BC0  uni7BC0,  CJK UNIFIED IDEOGRAPH-7BC0
        '31681' => '7921'             , # G+7921  U+7BC1  uni7BC1,  CJK UNIFIED IDEOGRAPH-7BC1
        '31684' => '5016'             , # G+5016  U+7BC4  uni7BC4,  CJK UNIFIED IDEOGRAPH-7BC4
        '31686' => '6671'             , # G+6671  U+7BC6  uni7BC6,  CJK UNIFIED IDEOGRAPH-7BC6
        '31687' => '7522'             , # G+7522  U+7BC7  uni7BC7,  CJK UNIFIED IDEOGRAPH-7BC7
        '31689' => '7293'             , # G+7293  U+7BC9  uni7BC9,  CJK UNIFIED IDEOGRAPH-7BC9
        '31698' => '5729'             , # G+5729  U+7BD2  uni7BD2,  CJK UNIFIED IDEOGRAPH-7BD2
        '31712' => '5523'             , # G+5523  U+7BE0  uni7BE0,  CJK UNIFIED IDEOGRAPH-7BE0
        '31716' => '4316'             , # G+4316  U+7BE4  uni7BE4,  CJK UNIFIED IDEOGRAPH-7BE4
        '31721' => '5282'             , # G+5282  U+7BE9  uni7BE9,  CJK UNIFIED IDEOGRAPH-7BE9
        '31751' => '6829'             , # G+6829  U+7C07  uni7C07,  CJK UNIFIED IDEOGRAPH-7C07
        '31762' => '7068'             , # G+7068  U+7C12  uni7C12,  CJK UNIFIED IDEOGRAPH-7C12
        '31774' => '4206'             , # G+4206  U+7C1E  uni7C1E,  CJK UNIFIED IDEOGRAPH-7C1E
        '31777' => '3494'             , # G+3494  U+7C21  uni7C21,  CJK UNIFIED IDEOGRAPH-7C21
        '31783' => '7922'             , # G+7922  U+7C27  uni7C27,  CJK UNIFIED IDEOGRAPH-7C27
        '31786' => '6531'             , # G+6531  U+7C2A  uni7C2A,  CJK UNIFIED IDEOGRAPH-7C2A
        '31787' => '5524'             , # G+5524  U+7C2B  uni7C2B,  CJK UNIFIED IDEOGRAPH-7C2B
        '31805' => '7183'             , # G+7183  U+7C3D  uni7C3D,  CJK UNIFIED IDEOGRAPH-7C3D
        '31806' => '4487'             , # G+4487  U+7C3E  uni7C3E,  CJK UNIFIED IDEOGRAPH-7C3E
        '31807' => '5131'             , # G+5131  U+7C3F  uni7C3F,  CJK UNIFIED IDEOGRAPH-7C3F
        '31811' => '4404'             , # G+4404  U+7C43  uni7C43,  CJK UNIFIED IDEOGRAPH-7C43
        '31820' => '6884'             , # G+6884  U+7C4C  uni7C4C,  CJK UNIFIED IDEOGRAPH-7C4C
        '31821' => '6630'             , # G+6630  U+7C4D  uni7C4D,  CJK UNIFIED IDEOGRAPH-7C4D
        '31840' => '4541'             , # G+4541  U+7C60  uni7C60,  CJK UNIFIED IDEOGRAPH-7C60
        '31844' => '7184'             , # G+7184  U+7C64  uni7C64,  CJK UNIFIED IDEOGRAPH-7C64
        '31852' => '4628'             , # G+4628  U+7C6C  uni7C6C,  CJK UNIFIED IDEOGRAPH-7C6C
        '31859' => '4863'             , # G+4863  U+7C73  uni7C73,  CJK UNIFIED IDEOGRAPH-7C73
        '31875' => '5207'             , # G+5207  U+7C83  uni7C83,  CJK UNIFIED IDEOGRAPH-7C83
        '31881' => '5165'             , # G+5165  U+7C89  uni7C89,  CJK UNIFIED IDEOGRAPH-7C89
        '31890' => '4655'             , # G+4655  U+7C92  uni7C92,  CJK UNIFIED IDEOGRAPH-7C92
        '31893' => '4897'             , # G+4897  U+7C95  uni7C95,  CJK UNIFIED IDEOGRAPH-7C95
        '31895' => '6811'             , # G+6811  U+7C97  uni7C97,  CJK UNIFIED IDEOGRAPH-7C97
        '31896' => '6697'             , # G+6697  U+7C98  uni7C98,  CJK UNIFIED IDEOGRAPH-7C98
        '31903' => '5541'             , # G+5541  U+7C9F  uni7C9F,  CJK UNIFIED IDEOGRAPH-7C9F
        '31909' => '6901'             , # G+6901  U+7CA5  uni7CA5,  CJK UNIFIED IDEOGRAPH-7CA5
        '31911' => '6556'             , # G+6556  U+7CA7  uni7CA7,  CJK UNIFIED IDEOGRAPH-7CA7
        '31918' => '4433'             , # G+4433  U+7CAE  uni7CAE,  CJK UNIFIED IDEOGRAPH-7CAE
        '31921' => '4434'             , # G+4434  U+7CB1  uni7CB1,  CJK UNIFIED IDEOGRAPH-7CB1
        '31922' => '7070'             , # G+7070  U+7CB2  uni7CB2,  CJK UNIFIED IDEOGRAPH-7CB2
        '31923' => '3582'             , # G+3582  U+7CB3  uni7CB3,  CJK UNIFIED IDEOGRAPH-7CB3
        '31929' => '5600'             , # G+5600  U+7CB9  uni7CB9,  CJK UNIFIED IDEOGRAPH-7CB9
        '31934' => '6742'             , # G+6742  U+7CBE  uni7CBE,  CJK UNIFIED IDEOGRAPH-7CBE
        '31946' => '7827'             , # G+7827  U+7CCA  uni7CCA,  CJK UNIFIED IDEOGRAPH-7CCA
        '31958' => '4247'             , # G+4247  U+7CD6  uni7CD6,  CJK UNIFIED IDEOGRAPH-7CD6
        '31966' => '5166'             , # G+5166  U+7CDE  uni7CDE,  CJK UNIFIED IDEOGRAPH-7CDE
        '31967' => '6812'             , # G+6812  U+7CDF  uni7CDF,  CJK UNIFIED IDEOGRAPH-7CDF
        '31968' => '3548'             , # G+3548  U+7CE0  uni7CE0,  CJK UNIFIED IDEOGRAPH-7CE0
        '31975' => '4435'             , # G+4435  U+7CE7  uni7CE7,  CJK UNIFIED IDEOGRAPH-7CE7
        '31995' => '3721'             , # G+3721  U+7CFB  uni7CFB,  CJK UNIFIED IDEOGRAPH-7CFB
        '31998' => '4003'             , # G+4003  U+7CFE  uni7CFE,  CJK UNIFIED IDEOGRAPH-7CFE
        '32000' => '4108'             , # G+4108  U+7D00  uni7D00,  CJK UNIFIED IDEOGRAPH-7D00
        '32002' => '6885'             , # G+6885  U+7D02  uni7D02,  CJK UNIFIED IDEOGRAPH-7D02
        '32004' => '5873'             , # G+5873  U+7D04  uni7D04,  CJK UNIFIED IDEOGRAPH-7D04
        '32005' => '7859'             , # G+7859  U+7D05  uni7D05,  CJK UNIFIED IDEOGRAPH-7D05
        '32006' => '6217'             , # G+6217  U+7D06  uni7D06,  CJK UNIFIED IDEOGRAPH-7D06
        '32007' => '8025'             , # G+8025  U+7D07  uni7D07,  CJK UNIFIED IDEOGRAPH-7D07
        '32008' => '7895'             , # G+7895  U+7D08  uni7D08,  CJK UNIFIED IDEOGRAPH-7D08
        '32010' => '4842'             , # G+4842  U+7D0A  uni7D0A,  CJK UNIFIED IDEOGRAPH-7D0A
        '32011' => '4843'             , # G+4843  U+7D0B  uni7D0B,  CJK UNIFIED IDEOGRAPH-7D0B
        '32013' => '4150'             , # G+4150  U+7D0D  uni7D0D,  CJK UNIFIED IDEOGRAPH-7D0D
        '32016' => '4184'             , # G+4184  U+7D10  uni7D10,  CJK UNIFIED IDEOGRAPH-7D10
        '32020' => '5655'             , # G+5655  U+7D14  uni7D14,  CJK UNIFIED IDEOGRAPH-7D14
        '32023' => '5283'             , # G+5283  U+7D17  uni7D17,  CJK UNIFIED IDEOGRAPH-7D17
        '32024' => '3865'             , # G+3865  U+7D18  uni7D18,  CJK UNIFIED IDEOGRAPH-7D18
        '32025' => '6964'             , # G+6964  U+7D19  uni7D19,  CJK UNIFIED IDEOGRAPH-7D19
        '32026' => '4056'             , # G+4056  U+7D1A  uni7D1A,  CJK UNIFIED IDEOGRAPH-7D1A
        '32027' => '5167'             , # G+5167  U+7D1B  uni7D1B,  CJK UNIFIED IDEOGRAPH-7D1B
        '32032' => '5525'             , # G+5525  U+7D20  uni7D20,  CJK UNIFIED IDEOGRAPH-7D20
        '32033' => '4957'             , # G+4957  U+7D21  uni7D21,  CJK UNIFIED IDEOGRAPH-7D21
        '32034' => '5367'             , # G+5367  U+7D22  uni7D22,  CJK UNIFIED IDEOGRAPH-7D22
        '32043' => '6500'             , # G+6500  U+7D2B  uni7D2B,  CJK UNIFIED IDEOGRAPH-7D2B
        '32044' => '6886'             , # G+6886  U+7D2C  uni7D2C,  CJK UNIFIED IDEOGRAPH-7D2C
        '32046' => '7081'             , # G+7081  U+7D2E  uni7D2E,  CJK UNIFIED IDEOGRAPH-7D2E
        '32047' => '4571'             , # G+4571  U+7D2F  uni7D2F,  CJK UNIFIED IDEOGRAPH-7D2F
        '32048' => '5498'             , # G+5498  U+7D30  uni7D30,  CJK UNIFIED IDEOGRAPH-7D30
        '32051' => '5747'             , # G+5747  U+7D33  uni7D33,  CJK UNIFIED IDEOGRAPH-7D33
        '32053' => '6607'             , # G+6607  U+7D35  uni7D35,  CJK UNIFIED IDEOGRAPH-7D35
        '32057' => '5526'             , # G+5526  U+7D39  uni7D39,  CJK UNIFIED IDEOGRAPH-7D39
        '32058' => '3525'             , # G+3525  U+7D3A  uni7D3A,  CJK UNIFIED IDEOGRAPH-7D3A
        '32066' => '6846'             , # G+6846  U+7D42  uni7D42,  CJK UNIFIED IDEOGRAPH-7D42
        '32067' => '7746'             , # G+7746  U+7D43  uni7D43,  CJK UNIFIED IDEOGRAPH-7D43
        '32068' => '6813'             , # G+6813  U+7D44  uni7D44,  CJK UNIFIED IDEOGRAPH-7D44
        '32069' => '3691'             , # G+3691  U+7D45  uni7D45,  CJK UNIFIED IDEOGRAPH-7D45
        '32070' => '4924'             , # G+4924  U+7D46  uni7D46,  CJK UNIFIED IDEOGRAPH-7D46
        '32080' => '3651'             , # G+3651  U+7D50  uni7D50,  CJK UNIFIED IDEOGRAPH-7D50
        '32094' => '3882'             , # G+3882  U+7D5E  uni7D5E,  CJK UNIFIED IDEOGRAPH-7D5E
        '32097' => '4385'             , # G+4385  U+7D61  uni7D61,  CJK UNIFIED IDEOGRAPH-7D61
        '32098' => '7747'             , # G+7747  U+7D62  uni7D62,  CJK UNIFIED IDEOGRAPH-7D62
        '32102' => '4057'             , # G+4057  U+7D66  uni7D66,  CJK UNIFIED IDEOGRAPH-7D66
        '32104' => '6371'             , # G+6371  U+7D68  uni7D68,  CJK UNIFIED IDEOGRAPH-7D68
        '32106' => '6454'             , # G+6454  U+7D6A  uni7D6A,  CJK UNIFIED IDEOGRAPH-7D6A
        '32110' => '5391'             , # G+5391  U+7D6E  uni7D6E,  CJK UNIFIED IDEOGRAPH-7D6E
        '32113' => '7457'             , # G+7457  U+7D71  uni7D71,  CJK UNIFIED IDEOGRAPH-7D71
        '32114' => '5284'             , # G+5284  U+7D72  uni7D72,  CJK UNIFIED IDEOGRAPH-7D72
        '32115' => '3549'             , # G+3549  U+7D73  uni7D73,  CJK UNIFIED IDEOGRAPH-7D73
        '32118' => '6691'             , # G+6691  U+7D76  uni7D76,  CJK UNIFIED IDEOGRAPH-7D76
        '32121' => '3641'             , # G+3641  U+7D79  uni7D79,  CJK UNIFIED IDEOGRAPH-7D79
        '32127' => '3928'             , # G+3928  U+7D7F  uni7D7F,  CJK UNIFIED IDEOGRAPH-7D7F
        '32142' => '6743'             , # G+6743  U+7D8E  uni7D8E,  CJK UNIFIED IDEOGRAPH-7D8E
        '32143' => '5601'             , # G+5601  U+7D8F  uni7D8F,  CJK UNIFIED IDEOGRAPH-7D8F
        '32147' => '3692'             , # G+3692  U+7D93  uni7D93,  CJK UNIFIED IDEOGRAPH-7D93
        '32156' => '6847'             , # G+6847  U+7D9C  uni7D9C,  CJK UNIFIED IDEOGRAPH-7D9C
        '32160' => '4530'             , # G+4530  U+7DA0  uni7DA0,  CJK UNIFIED IDEOGRAPH-7DA0
        '32162' => '6887'             , # G+6887  U+7DA2  uni7DA2,  CJK UNIFIED IDEOGRAPH-7DA2
        '32172' => '5602'             , # G+5602  U+7DAC  uni7DAC,  CJK UNIFIED IDEOGRAPH-7DAC
        '32173' => '6340'             , # G+6340  U+7DAD  uni7DAD,  CJK UNIFIED IDEOGRAPH-7DAD
        '32177' => '3550'             , # G+3550  U+7DB1  uni7DB1,  CJK UNIFIED IDEOGRAPH-7DB1
        '32178' => '4701'             , # G+4701  U+7DB2  uni7DB2,  CJK UNIFIED IDEOGRAPH-7DB2
        '32180' => '7173'             , # G+7173  U+7DB4  uni7DB4,  CJK UNIFIED IDEOGRAPH-7DB4
        '32181' => '7121'             , # G+7121  U+7DB5  uni7DB5,  CJK UNIFIED IDEOGRAPH-7DB5
        '32184' => '4598'             , # G+4598  U+7DB8  uni7DB8,  CJK UNIFIED IDEOGRAPH-7DB8
        '32186' => '4109'             , # G+4109  U+7DBA  uni7DBA,  CJK UNIFIED IDEOGRAPH-7DBA
        '32187' => '7416'             , # G+7416  U+7DBB  uni7DBB,  CJK UNIFIED IDEOGRAPH-7DBB
        '32189' => '6517'             , # G+6517  U+7DBD  uni7DBD,  CJK UNIFIED IDEOGRAPH-7DBD
        '32190' => '4610'             , # G+4610  U+7DBE  uni7DBE,  CJK UNIFIED IDEOGRAPH-7DBE
        '32191' => '4742'             , # G+4742  U+7DBF  uni7DBF,  CJK UNIFIED IDEOGRAPH-7DBF
        '32199' => '7351'             , # G+7351  U+7DC7  uni7DC7,  CJK UNIFIED IDEOGRAPH-7DC7
        '32202' => '4126'             , # G+4126  U+7DCA  uni7DCA,  CJK UNIFIED IDEOGRAPH-7DCA
        '32203' => '5208'             , # G+5208  U+7DCB  uni7DCB,  CJK UNIFIED IDEOGRAPH-7DCB
        '32214' => '5392'             , # G+5392  U+7DD6  uni7DD6,  CJK UNIFIED IDEOGRAPH-7DD6
        '32216' => '7658'             , # G+7658  U+7DD8  uni7DD8,  CJK UNIFIED IDEOGRAPH-7DD8
        '32218' => '5435'             , # G+5435  U+7DDA  uni7DDA,  CJK UNIFIED IDEOGRAPH-7DDA
        '32221' => '7034'             , # G+7034  U+7DDD  uni7DDD,  CJK UNIFIED IDEOGRAPH-7DDD
        '32222' => '4207'             , # G+4207  U+7DDE  uni7DDE,  CJK UNIFIED IDEOGRAPH-7DDE
        '32224' => '7209'             , # G+7209  U+7DE0  uni7DE0,  CJK UNIFIED IDEOGRAPH-7DE0
        '32225' => '4881'             , # G+4881  U+7DE1  uni7DE1,  CJK UNIFIED IDEOGRAPH-7DE1
        '32227' => '5979'             , # G+5979  U+7DE3  uni7DE3,  CJK UNIFIED IDEOGRAPH-7DE3
        '32232' => '7523'             , # G+7523  U+7DE8  uni7DE8,  CJK UNIFIED IDEOGRAPH-7DE8
        '32233' => '6117'             , # G+6117  U+7DE9  uni7DE9,  CJK UNIFIED IDEOGRAPH-7DE9
        '32236' => '4743'             , # G+4743  U+7DEC  uni7DEC,  CJK UNIFIED IDEOGRAPH-7DEC
        '32239' => '6297'             , # G+6297  U+7DEF  uni7DEF,  CJK UNIFIED IDEOGRAPH-7DEF
        '32244' => '4471'             , # G+4471  U+7DF4  uni7DF4,  CJK UNIFIED IDEOGRAPH-7DF4
        '32251' => '7352'             , # G+7352  U+7DFB  uni7DFB,  CJK UNIFIED IDEOGRAPH-7DFB
        '32265' => '7001'             , # G+7001  U+7E09  uni7E09,  CJK UNIFIED IDEOGRAPH-7E09
        '32266' => '5855'             , # G+5855  U+7E0A  uni7E0A,  CJK UNIFIED IDEOGRAPH-7E0A
        '32277' => '6087'             , # G+6087  U+7E15  uni7E15,  CJK UNIFIED IDEOGRAPH-7E15
        '32283' => '4898'             , # G+4898  U+7E1B  uni7E1B,  CJK UNIFIED IDEOGRAPH-7E1B
        '32285' => '7002'             , # G+7002  U+7E1D  uni7E1D,  CJK UNIFIED IDEOGRAPH-7E1D
        '32286' => '7828'             , # G+7828  U+7E1E  uni7E1E,  CJK UNIFIED IDEOGRAPH-7E1E
        '32287' => '6171'             , # G+6171  U+7E1F  uni7E1F,  CJK UNIFIED IDEOGRAPH-7E1F
        '32289' => '6581'             , # G+6581  U+7E21  uni7E21,  CJK UNIFIED IDEOGRAPH-7E21
        '32291' => '7748'             , # G+7748  U+7E23  uni7E23,  CJK UNIFIED IDEOGRAPH-7E23
        '32299' => '5103'             , # G+5103  U+7E2B  uni7E2B,  CJK UNIFIED IDEOGRAPH-7E2B
        '32302' => '7294'             , # G+7294  U+7E2E  uni7E2E,  CJK UNIFIED IDEOGRAPH-7E2E
        '32303' => '5980'             , # G+5980  U+7E2F  uni7E2F,  CJK UNIFIED IDEOGRAPH-7E2F
        '32305' => '6848'             , # G+6848  U+7E31  uni7E31,  CJK UNIFIED IDEOGRAPH-7E31
        '32311' => '4572'             , # G+4572  U+7E37  uni7E37,  CJK UNIFIED IDEOGRAPH-7E37
        '32317' => '7257'             , # G+7257  U+7E3D  uni7E3D,  CJK UNIFIED IDEOGRAPH-7E3D
        '32318' => '6631'             , # G+6631  U+7E3E  uni7E3E,  CJK UNIFIED IDEOGRAPH-7E3E
        '32321' => '5001'             , # G+5001  U+7E41  uni7E41,  CJK UNIFIED IDEOGRAPH-7E41
        '32323' => '5179'             , # G+5179  U+7E43  uni7E43,  CJK UNIFIED IDEOGRAPH-7E43
        '32326' => '4826'             , # G+4826  U+7E46  uni7E46,  CJK UNIFIED IDEOGRAPH-7E46
        '32327' => '6158'             , # G+6158  U+7E47  uni7E47,  CJK UNIFIED IDEOGRAPH-7E47
        '32338' => '6938'             , # G+6938  U+7E52  uni7E52,  CJK UNIFIED IDEOGRAPH-7E52
        '32340' => '6978'             , # G+6978  U+7E54  uni7E54,  CJK UNIFIED IDEOGRAPH-7E54
        '32341' => '5436'             , # G+5436  U+7E55  uni7E55,  CJK UNIFIED IDEOGRAPH-7E55
        '32350' => '6159'             , # G+6159  U+7E5E  uni7E5E,  CJK UNIFIED IDEOGRAPH-7E5E
        '32353' => '5603'             , # G+5603  U+7E61  uni7E61,  CJK UNIFIED IDEOGRAPH-7E61
        '32361' => '5689'             , # G+5689  U+7E69  uni7E69,  CJK UNIFIED IDEOGRAPH-7E69
        '32362' => '7942'             , # G+7942  U+7E6A  uni7E6A,  CJK UNIFIED IDEOGRAPH-7E6A
        '32363' => '3722'             , # G+3722  U+7E6B  uni7E6B,  CJK UNIFIED IDEOGRAPH-7E6B
        '32365' => '3642'             , # G+3642  U+7E6D  uni7E6D,  CJK UNIFIED IDEOGRAPH-7E6D
        '32368' => '6814'             , # G+6814  U+7E70  uni7E70,  CJK UNIFIED IDEOGRAPH-7E70
        '32377' => '5952'             , # G+5952  U+7E79  uni7E79,  CJK UNIFIED IDEOGRAPH-7E79
        '32380' => '3723'             , # G+3723  U+7E7C  uni7E7C,  CJK UNIFIED IDEOGRAPH-7E7C
        '32386' => '7069'             , # G+7069  U+7E82  uni7E82,  CJK UNIFIED IDEOGRAPH-7E82
        '32396' => '5542'             , # G+5542  U+7E8C  uni7E8C,  CJK UNIFIED IDEOGRAPH-7E8C
        '32399' => '6672'             , # G+6672  U+7E8F  uni7E8F,  CJK UNIFIED IDEOGRAPH-7E8F
        '32403' => '6023'             , # G+6023  U+7E93  uni7E93,  CJK UNIFIED IDEOGRAPH-7E93
        '32406' => '5466'             , # G+5466  U+7E96  uni7E96,  CJK UNIFIED IDEOGRAPH-7E96
        '32408' => '7071'             , # G+7071  U+7E98  uni7E98,  CJK UNIFIED IDEOGRAPH-7E98
        '32411' => '4317'             , # G+4317  U+7E9B  uni7E9B,  CJK UNIFIED IDEOGRAPH-7E9B
        '32412' => '4405'             , # G+4405  U+7E9C  uni7E9C,  CJK UNIFIED IDEOGRAPH-7E9C
        '32566' => '5132'             , # G+5132  U+7F36  uni7F36,  CJK UNIFIED IDEOGRAPH-7F36
        '32568' => '7681'             , # G+7681  U+7F38  uni7F38,  CJK UNIFIED IDEOGRAPH-7F38
        '32570' => '3652'             , # G+3652  U+7F3A  uni7F3A,  CJK UNIFIED IDEOGRAPH-7F3A
        '32588' => '5859'             , # G+5859  U+7F4C  uni7F4C,  CJK UNIFIED IDEOGRAPH-7F4C
        '32592' => '3829'             , # G+3829  U+7F50  uni7F50,  CJK UNIFIED IDEOGRAPH-7F50
        '32596' => '4702'             , # G+4702  U+7F54  uni7F54,  CJK UNIFIED IDEOGRAPH-7F54
        '32597' => '7643'             , # G+7643  U+7F55  uni7F55,  CJK UNIFIED IDEOGRAPH-7F55
        '32618' => '6859'             , # G+6859  U+7F6A  uni7F6A,  CJK UNIFIED IDEOGRAPH-7F6A
        '32619' => '3854'             , # G+3854  U+7F6B  uni7F6B,  CJK UNIFIED IDEOGRAPH-7F6B
        '32622' => '7353'             , # G+7353  U+7F6E  uni7F6E,  CJK UNIFIED IDEOGRAPH-7F6E
        '32624' => '5007'             , # G+5007  U+7F70  uni7F70,  CJK UNIFIED IDEOGRAPH-7F70
        '32626' => '5393'             , # G+5393  U+7F72  uni7F72,  CJK UNIFIED IDEOGRAPH-7F72
        '32629' => '4717'             , # G+4717  U+7F75  uni7F75,  CJK UNIFIED IDEOGRAPH-7F75
        '32631' => '7486'             , # G+7486  U+7F77  uni7F77,  CJK UNIFIED IDEOGRAPH-7F77
        '32633' => '4629'             , # G+4629  U+7F79  uni7F79,  CJK UNIFIED IDEOGRAPH-7F79
        '32645' => '4377'             , # G+4377  U+7F85  uni7F85,  CJK UNIFIED IDEOGRAPH-7F85
        '32648' => '4110'             , # G+4110  U+7F88  uni7F88,  CJK UNIFIED IDEOGRAPH-7F88
        '32650' => '5896'             , # G+5896  U+7F8A  uni7F8A,  CJK UNIFIED IDEOGRAPH-7F8A
        '32652' => '3551'             , # G+3551  U+7F8C  uni7F8C,  CJK UNIFIED IDEOGRAPH-7F8C
        '32654' => '4864'             , # G+4864  U+7F8E  uni7F8E,  CJK UNIFIED IDEOGRAPH-7F8E
        '32660' => '3750'             , # G+3750  U+7F94  uni7F94,  CJK UNIFIED IDEOGRAPH-7F94
        '32666' => '4497'             , # G+4497  U+7F9A  uni7F9A,  CJK UNIFIED IDEOGRAPH-7F9A
        '32670' => '5604'             , # G+5604  U+7F9E  uni7F9E,  CJK UNIFIED IDEOGRAPH-7F9E
        '32676' => '3955'             , # G+3955  U+7FA4  uni7FA4,  CJK UNIFIED IDEOGRAPH-7FA4
        '32680' => '5437'             , # G+5437  U+7FA8  uni7FA8,  CJK UNIFIED IDEOGRAPH-7FA8
        '32681' => '6405'             , # G+6405  U+7FA9  uni7FA9,  CJK UNIFIED IDEOGRAPH-7FA9
        '32690' => '8054'             , # G+8054  U+7FB2  uni7FB2,  CJK UNIFIED IDEOGRAPH-7FB2
        '32696' => '4630'             , # G+4630  U+7FB8  uni7FB8,  CJK UNIFIED IDEOGRAPH-7FB8
        '32697' => '3583'             , # G+3583  U+7FB9  uni7FB9,  CJK UNIFIED IDEOGRAPH-7FB9
        '32701' => '6218'             , # G+6218  U+7FBD  uni7FBD,  CJK UNIFIED IDEOGRAPH-7FBD
        '32705' => '6095'             , # G+6095  U+7FC1  uni7FC1,  CJK UNIFIED IDEOGRAPH-7FC1
        '32709' => '5711'             , # G+5711  U+7FC5  uni7FC5,  CJK UNIFIED IDEOGRAPH-7FC5
        '32714' => '6439'             , # G+6439  U+7FCA  uni7FCA,  CJK UNIFIED IDEOGRAPH-7FCA
        '32716' => '6440'             , # G+6440  U+7FCC  uni7FCC,  CJK UNIFIED IDEOGRAPH-7FCC
        '32718' => '4498'             , # G+4498  U+7FCE  uni7FCE,  CJK UNIFIED IDEOGRAPH-7FCE
        '32722' => '5679'             , # G+5679  U+7FD2  uni7FD2,  CJK UNIFIED IDEOGRAPH-7FD2
        '32724' => '5355'             , # G+5355  U+7FD4  uni7FD4,  CJK UNIFIED IDEOGRAPH-7FD4
        '32725' => '8033'             , # G+8033  U+7FD5  uni7FD5,  CJK UNIFIED IDEOGRAPH-7FD5
        '32735' => '6632'             , # G+6632  U+7FDF  uni7FDF,  CJK UNIFIED IDEOGRAPH-7FDF
        '32736' => '7322'             , # G+7322  U+7FE0  uni7FE0,  CJK UNIFIED IDEOGRAPH-7FE0
        '32737' => '5209'             , # G+5209  U+7FE1  uni7FE1,  CJK UNIFIED IDEOGRAPH-7FE1
        '32745' => '7524'             , # G+7524  U+7FE9  uni7FE9,  CJK UNIFIED IDEOGRAPH-7FE9
        '32747' => '6118'             , # G+6118  U+7FEB  uni7FEB,  CJK UNIFIED IDEOGRAPH-7FEB
        '32752' => '7644'             , # G+7644  U+7FF0  uni7FF0,  CJK UNIFIED IDEOGRAPH-7FF0
        '32761' => '3883'             , # G+3883  U+7FF9  uni7FF9,  CJK UNIFIED IDEOGRAPH-7FF9
        '32764' => '6441'             , # G+6441  U+7FFC  uni7FFC,  CJK UNIFIED IDEOGRAPH-7FFC
        '32768' => '6160'             , # G+6160  U+8000  uni8000,  CJK UNIFIED IDEOGRAPH-8000
        '32769' => '4519'             , # G+4519  U+8001  uni8001,  CJK UNIFIED IDEOGRAPH-8001
        '32771' => '3751'             , # G+3751  U+8003  uni8003,  CJK UNIFIED IDEOGRAPH-8003
        '32773' => '6501'             , # G+6501  U+8005  uni8005,  CJK UNIFIED IDEOGRAPH-8005
        '32774' => '4111'             , # G+4111  U+8006  uni8006,  CJK UNIFIED IDEOGRAPH-8006
        '32777' => '3929'             , # G+3929  U+8009  uni8009,  CJK UNIFIED IDEOGRAPH-8009
        '32780' => '6427'             , # G+6427  U+800C  uni800C,  CJK UNIFIED IDEOGRAPH-800C
        '32784' => '4158'             , # G+4158  U+8010  uni8010,  CJK UNIFIED IDEOGRAPH-8010
        '32789' => '3693'             , # G+3693  U+8015  uni8015,  CJK UNIFIED IDEOGRAPH-8015
        '32791' => '4782'             , # G+4782  U+8017  uni8017,  CJK UNIFIED IDEOGRAPH-8017
        '32792' => '6243'             , # G+6243  U+8018  uni8018,  CJK UNIFIED IDEOGRAPH-8018
        '32813' => '4112'             , # G+4112  U+802D  uni802D,  CJK UNIFIED IDEOGRAPH-802D
        '32819' => '6428'             , # G+6428  U+8033  uni8033,  CJK UNIFIED IDEOGRAPH-8033
        '32822' => '5870'             , # G+5870  U+8036  uni8036,  CJK UNIFIED IDEOGRAPH-8036
        '32829' => '7422'             , # G+7422  U+803D  uni803D,  CJK UNIFIED IDEOGRAPH-803D
        '32831' => '3694'             , # G+3694  U+803F  uni803F,  CJK UNIFIED IDEOGRAPH-803F
        '32835' => '4227'             , # G+4227  U+8043  uni8043,  CJK UNIFIED IDEOGRAPH-8043
        '32838' => '4499'             , # G+4499  U+8046  uni8046,  CJK UNIFIED IDEOGRAPH-8046
        '32842' => '4559'             , # G+4559  U+804A  uni804A,  CJK UNIFIED IDEOGRAPH-804A
        '32854' => '5487'             , # G+5487  U+8056  uni8056,  CJK UNIFIED IDEOGRAPH-8056
        '32856' => '5239'             , # G+5239  U+8058  uni8058,  CJK UNIFIED IDEOGRAPH-8058
        '32858' => '7323'             , # G+7323  U+805A  uni805A,  CJK UNIFIED IDEOGRAPH-805A
        '32862' => '4844'             , # G+4844  U+805E  uni805E,  CJK UNIFIED IDEOGRAPH-805E
        '32879' => '4472'             , # G+4472  U+806F  uni806F,  CJK UNIFIED IDEOGRAPH-806F
        '32880' => '7258'             , # G+7258  U+8070  uni8070,  CJK UNIFIED IDEOGRAPH-8070
        '32882' => '5488'             , # G+5488  U+8072  uni8072,  CJK UNIFIED IDEOGRAPH-8072
        '32883' => '6191'             , # G+6191  U+8073  uni8073,  CJK UNIFIED IDEOGRAPH-8073
        '32887' => '6979'             , # G+6979  U+8077  uni8077,  CJK UNIFIED IDEOGRAPH-8077
        '32893' => '7200'             , # G+7200  U+807D  uni807D,  CJK UNIFIED IDEOGRAPH-807D
        '32894' => '4542'             , # G+4542  U+807E  uni807E,  CJK UNIFIED IDEOGRAPH-807E
        '32895' => '6368'             , # G+6368  U+807F  uni807F,  CJK UNIFIED IDEOGRAPH-807F
        '32900' => '6429'             , # G+6429  U+8084  uni8084,  CJK UNIFIED IDEOGRAPH-8084
        '32901' => '5638'             , # G+5638  U+8085  uni8085,  CJK UNIFIED IDEOGRAPH-8085
        '32902' => '5285'             , # G+5285  U+8086  uni8086,  CJK UNIFIED IDEOGRAPH-8086
        '32903' => '6815'             , # G+6815  U+8087  uni8087,  CJK UNIFIED IDEOGRAPH-8087
        '32905' => '6357'             , # G+6357  U+8089  uni8089,  CJK UNIFIED IDEOGRAPH-8089
        '32907' => '4605'             , # G+4605  U+808B  uni808B,  CJK UNIFIED IDEOGRAPH-808B
        '32908' => '4113'             , # G+4113  U+808C  uni808C,  CJK UNIFIED IDEOGRAPH-808C
        '32918' => '7231'             , # G+7231  U+8096  uni8096,  CJK UNIFIED IDEOGRAPH-8096
        '32923' => '7682'             , # G+7682  U+809B  uni809B,  CJK UNIFIED IDEOGRAPH-809B
        '32925' => '3495'             , # G+3495  U+809D  uni809D,  CJK UNIFIED IDEOGRAPH-809D
        '32929' => '3752'             , # G+3752  U+80A1  uni80A1,  CJK UNIFIED IDEOGRAPH-80A1
        '32930' => '6965'             , # G+6965  U+80A2  uni80A2,  CJK UNIFIED IDEOGRAPH-80A2
        '32933' => '5210'             , # G+5210  U+80A5  uni80A5,  CJK UNIFIED IDEOGRAPH-80A5
        '32937' => '3643'             , # G+3643  U+80A9  uni80A9,  CJK UNIFIED IDEOGRAPH-80A9
        '32938' => '4958'             , # G+4958  U+80AA  uni80AA,  CJK UNIFIED IDEOGRAPH-80AA
        '32943' => '4061'             , # G+4061  U+80AF  uni80AF,  CJK UNIFIED IDEOGRAPH-80AF
        '32945' => '3866'             , # G+3866  U+80B1  uni80B1,  CJK UNIFIED IDEOGRAPH-80B1
        '32946' => '6358'             , # G+6358  U+80B2  uni80B2,  CJK UNIFIED IDEOGRAPH-80B2
        '32948' => '7963'             , # G+7963  U+80B4  uni80B4,  CJK UNIFIED IDEOGRAPH-80B4
        '32954' => '7540'             , # G+7540  U+80BA  uni80BA,  CJK UNIFIED IDEOGRAPH-80BA
        '32963' => '6298'             , # G+6298  U+80C3  uni80C3,  CJK UNIFIED IDEOGRAPH-80C3
        '32964' => '6865'             , # G+6865  U+80C4  uni80C4,  CJK UNIFIED IDEOGRAPH-80C4
        '32972' => '4979'             , # G+4979  U+80CC  uni80CC,  CJK UNIFIED IDEOGRAPH-80CC
        '32974' => '7440'             , # G+7440  U+80CE  uni80CE,  CJK UNIFIED IDEOGRAPH-80CE
        '32986' => '4980'             , # G+4980  U+80DA  uni80DA,  CJK UNIFIED IDEOGRAPH-80DA
        '32987' => '3533'             , # G+3533  U+80DB  uni80DB,  CJK UNIFIED IDEOGRAPH-80DB
        '32990' => '7560'             , # G+7560  U+80DE  uni80DE,  CJK UNIFIED IDEOGRAPH-80DE
        '32993' => '7829'             , # G+7829  U+80E1  uni80E1,  CJK UNIFIED IDEOGRAPH-80E1
        '32996' => '6364'             , # G+6364  U+80E4  uni80E4,  CJK UNIFIED IDEOGRAPH-80E4
        '32997' => '5394'             , # G+5394  U+80E5  uni80E5,  CJK UNIFIED IDEOGRAPH-80E5
        '33009' => '3850'             , # G+3850  U+80F1  uni80F1,  CJK UNIFIED IDEOGRAPH-80F1
        '33012' => '4345'             , # G+4345  U+80F4  uni80F4,  CJK UNIFIED IDEOGRAPH-80F4
        '33016' => '8017'             , # G+8017  U+80F8  uni80F8,  CJK UNIFIED IDEOGRAPH-80F8
        '33021' => '4185'             , # G+4185  U+80FD  uni80FD,  CJK UNIFIED IDEOGRAPH-80FD
        '33026' => '6966'             , # G+6966  U+8102  uni8102,  CJK UNIFIED IDEOGRAPH-8102
        '33029' => '7766'             , # G+7766  U+8105  uni8105,  CJK UNIFIED IDEOGRAPH-8105
        '33030' => '7324'             , # G+7324  U+8106  uni8106,  CJK UNIFIED IDEOGRAPH-8106
        '33031' => '7767'             , # G+7767  U+8107  uni8107,  CJK UNIFIED IDEOGRAPH-8107
        '33032' => '4722'             , # G+4722  U+8108  uni8108,  CJK UNIFIED IDEOGRAPH-8108
        '33034' => '7144'             , # G+7144  U+810A  uni810A,  CJK UNIFIED IDEOGRAPH-810A
        '33048' => '6119'             , # G+6119  U+8118  uni8118,  CJK UNIFIED IDEOGRAPH-8118
        '33050' => '3472'             , # G+3472  U+811A  uni811A,  CJK UNIFIED IDEOGRAPH-811A
        '33051' => '3695'             , # G+3695  U+811B  uni811B,  CJK UNIFIED IDEOGRAPH-811B
        '33059' => '5656'             , # G+5656  U+8123  uni8123,  CJK UNIFIED IDEOGRAPH-8123
        '33065' => '5605'             , # G+5605  U+8129  uni8129,  CJK UNIFIED IDEOGRAPH-8129
        '33067' => '7419'             , # G+7419  U+812B  uni812B,  CJK UNIFIED IDEOGRAPH-812B
        '33071' => '7561'             , # G+7561  U+812F  uni812F,  CJK UNIFIED IDEOGRAPH-812F
        '33081' => '7110'             , # G+7110  U+8139  uni8139,  CJK UNIFIED IDEOGRAPH-8139
        '33086' => '5211'             , # G+5211  U+813E  uni813E,  CJK UNIFIED IDEOGRAPH-813E
        '33099' => '5856'             , # G+5856  U+814B  uni814B,  CJK UNIFIED IDEOGRAPH-814B
        '33102' => '5748'             , # G+5748  U+814E  uni814E,  CJK UNIFIED IDEOGRAPH-814E
        '33104' => '5133'             , # G+5133  U+8150  uni8150,  CJK UNIFIED IDEOGRAPH-8150
        '33105' => '5134'             , # G+5134  U+8151  uni8151,  CJK UNIFIED IDEOGRAPH-8151
        '33108' => '3552'             , # G+3552  U+8154  uni8154,  CJK UNIFIED IDEOGRAPH-8154
        '33109' => '6120'             , # G+6120  U+8155  uni8155,  CJK UNIFIED IDEOGRAPH-8155
        '33125' => '5489'             , # G+5489  U+8165  uni8165,  CJK UNIFIED IDEOGRAPH-8165
        '33126' => '4179'             , # G+4179  U+8166  uni8166,  CJK UNIFIED IDEOGRAPH-8166
        '33131' => '6849'             , # G+6849  U+816B  uni816B,  CJK UNIFIED IDEOGRAPH-816B
        '33136' => '6161'             , # G+6161  U+8170  uni8170,  CJK UNIFIED IDEOGRAPH-8170
        '33137' => '3608'             , # G+3608  U+8171  uni8171,  CJK UNIFIED IDEOGRAPH-8171
        '33144' => '6557'             , # G+6557  U+8178  uni8178,  CJK UNIFIED IDEOGRAPH-8178
        '33145' => '5082'             , # G+5082  U+8179  uni8179,  CJK UNIFIED IDEOGRAPH-8179
        '33146' => '5438'             , # G+5438  U+817A  uni817A,  CJK UNIFIED IDEOGRAPH-817A
        '33151' => '7461'             , # G+7461  U+817F  uni817F,  CJK UNIFIED IDEOGRAPH-817F
        '33152' => '4959'             , # G+4959  U+8180  uni8180,  CJK UNIFIED IDEOGRAPH-8180
        '33160' => '3634'             , # G+3634  U+8188  uni8188,  CJK UNIFIED IDEOGRAPH-8188
        '33162' => '4899'             , # G+4899  U+818A  uni818A,  CJK UNIFIED IDEOGRAPH-818A
        '33167' => '3753'             , # G+3753  U+818F  uni818F,  CJK UNIFIED IDEOGRAPH-818F
        '33178' => '5135'             , # G+5135  U+819A  uni819A,  CJK UNIFIED IDEOGRAPH-819A
        '33180' => '4667'             , # G+4667  U+819C  uni819C,  CJK UNIFIED IDEOGRAPH-819C
        '33181' => '5675'             , # G+5675  U+819D  uni819D,  CJK UNIFIED IDEOGRAPH-819D
        '33184' => '3884'             , # G+3884  U+81A0  uni81A0,  CJK UNIFIED IDEOGRAPH-81A0
        '33187' => '7025'             , # G+7025  U+81A3  uni81A3,  CJK UNIFIED IDEOGRAPH-81A3
        '33192' => '7516'             , # G+7516  U+81A8  uni81A8,  CJK UNIFIED IDEOGRAPH-81A8
        '33203' => '5439'             , # G+5439  U+81B3  uni81B3,  CJK UNIFIED IDEOGRAPH-81B3
        '33205' => '7313'             , # G+7313  U+81B5  uni81B5,  CJK UNIFIED IDEOGRAPH-81B5
        '33210' => '6392'             , # G+6392  U+81BA  uni81BA,  CJK UNIFIED IDEOGRAPH-81BA
        '33213' => '4228'             , # G+4228  U+81BD  uni81BD,  CJK UNIFIED IDEOGRAPH-81BD
        '33214' => '7943'             , # G+7943  U+81BE  uni81BE,  CJK UNIFIED IDEOGRAPH-81BE
        '33215' => '4176'             , # G+4176  U+81BF  uni81BF,  CJK UNIFIED IDEOGRAPH-81BF
        '33216' => '4359'             , # G+4359  U+81C0  uni81C0,  CJK UNIFIED IDEOGRAPH-81C0
        '33218' => '5212'             , # G+5212  U+81C2  uni81C2,  CJK UNIFIED IDEOGRAPH-81C2
        '33222' => '5916'             , # G+5916  U+81C6  uni81C6,  CJK UNIFIED IDEOGRAPH-81C6
        '33229' => '6771'             , # G+6771  U+81CD  uni81CD,  CJK UNIFIED IDEOGRAPH-81CD
        '33240' => '4410'             , # G+4410  U+81D8  uni81D8,  CJK UNIFIED IDEOGRAPH-81D8
        '33247' => '6558'             , # G+6558  U+81DF  uni81DF,  CJK UNIFIED IDEOGRAPH-81DF
        '33251' => '5749'             , # G+5749  U+81E3  uni81E3,  CJK UNIFIED IDEOGRAPH-81E3
        '33253' => '6103'             , # G+6103  U+81E5  uni81E5,  CJK UNIFIED IDEOGRAPH-81E5
        '33255' => '6559'             , # G+6559  U+81E7  uni81E7,  CJK UNIFIED IDEOGRAPH-81E7
        '33256' => '4650'             , # G+4650  U+81E8  uni81E8,  CJK UNIFIED IDEOGRAPH-81E8
        '33258' => '6502'             , # G+6502  U+81EA  uni81EA,  CJK UNIFIED IDEOGRAPH-81EA
        '33261' => '7325'             , # G+7325  U+81ED  uni81ED,  CJK UNIFIED IDEOGRAPH-81ED
        '33267' => '6967'             , # G+6967  U+81F3  uni81F3,  CJK UNIFIED IDEOGRAPH-81F3
        '33268' => '7354'             , # G+7354  U+81F4  uni81F4,  CJK UNIFIED IDEOGRAPH-81F4
        '33274' => '4261'             , # G+4261  U+81FA  uni81FA,  CJK UNIFIED IDEOGRAPH-81FA
        '33275' => '7003'             , # G+7003  U+81FB  uni81FB,  CJK UNIFIED IDEOGRAPH-81FB
        '33276' => '3930'             , # G+3930  U+81FC  uni81FC,  CJK UNIFIED IDEOGRAPH-81FC
        '33278' => '6341'             , # G+6341  U+81FE  uni81FE,  CJK UNIFIED IDEOGRAPH-81FE
        '33285' => '3931'             , # G+3931  U+8205  uni8205,  CJK UNIFIED IDEOGRAPH-8205
        '33287' => '5941'             , # G+5941  U+8207  uni8207,  CJK UNIFIED IDEOGRAPH-8207
        '33288' => '8034'             , # G+8034  U+8208  uni8208,  CJK UNIFIED IDEOGRAPH-8208
        '33290' => '3932'             , # G+3932  U+820A  uni820A,  CJK UNIFIED IDEOGRAPH-820A
        '33292' => '5456'             , # G+5456  U+820C  uni820C,  CJK UNIFIED IDEOGRAPH-820C
        '33293' => '5286'             , # G+5286  U+820D  uni820D,  CJK UNIFIED IDEOGRAPH-820D
        '33298' => '5395'             , # G+5395  U+8212  uni8212,  CJK UNIFIED IDEOGRAPH-8212
        '33307' => '7158'             , # G+7158  U+821B  uni821B,  CJK UNIFIED IDEOGRAPH-821B
        '33308' => '5657'             , # G+5657  U+821C  uni821C,  CJK UNIFIED IDEOGRAPH-821C
        '33310' => '4827'             , # G+4827  U+821E  uni821E,  CJK UNIFIED IDEOGRAPH-821E
        '33311' => '6888'             , # G+6888  U+821F  uni821F,  CJK UNIFIED IDEOGRAPH-821F
        '33313' => '3553'             , # G+3553  U+8221  uni8221,  CJK UNIFIED IDEOGRAPH-8221
        '33322' => '7683'             , # G+7683  U+822A  uni822A,  CJK UNIFIED IDEOGRAPH-822A
        '33323' => '4960'             , # G+4960  U+822B  uni822B,  CJK UNIFIED IDEOGRAPH-822B
        '33324' => '4925'             , # G+4925  U+822C  uni822C,  CJK UNIFIED IDEOGRAPH-822C
        '33333' => '7390'             , # G+7390  U+8235  uni8235,  CJK UNIFIED IDEOGRAPH-8235
        '33334' => '4900'             , # G+4900  U+8236  uni8236,  CJK UNIFIED IDEOGRAPH-8236
        '33335' => '7749'             , # G+7749  U+8237  uni8237,  CJK UNIFIED IDEOGRAPH-8237
        '33337' => '5440'             , # G+5440  U+8239  uni8239,  CJK UNIFIED IDEOGRAPH-8239
        '33344' => '5136'             , # G+5136  U+8240  uni8240,  CJK UNIFIED IDEOGRAPH-8240
        '33349' => '5942'             , # G+5942  U+8245  uni8245,  CJK UNIFIED IDEOGRAPH-8245
        '33351' => '6744'             , # G+6744  U+8247  uni8247,  CJK UNIFIED IDEOGRAPH-8247
        '33369' => '7111'             , # G+7111  U+8259  uni8259,  CJK UNIFIED IDEOGRAPH-8259
        '33380' => '6406'             , # G+6406  U+8264  uni8264,  CJK UNIFIED IDEOGRAPH-8264
        '33382' => '7659'             , # G+7659  U+8266  uni8266,  CJK UNIFIED IDEOGRAPH-8266
        '33390' => '3496'             , # G+3496  U+826E  uni826E,  CJK UNIFIED IDEOGRAPH-826E
        '33391' => '4436'             , # G+4436  U+826F  uni826F,  CJK UNIFIED IDEOGRAPH-826F
        '33393' => '3497'             , # G+3497  U+8271  uni8271,  CJK UNIFIED IDEOGRAPH-8271
        '33394' => '5368'             , # G+5368  U+8272  uni8272,  CJK UNIFIED IDEOGRAPH-8272
        '33398' => '5994'             , # G+5994  U+8276  uni8276,  CJK UNIFIED IDEOGRAPH-8276
        '33400' => '7232'             , # G+7232  U+8278  uni8278,  CJK UNIFIED IDEOGRAPH-8278
        '33406' => '5848'             , # G+5848  U+827E  uni827E,  CJK UNIFIED IDEOGRAPH-827E
        '33419' => '6219'             , # G+6219  U+828B  uni828B,  CJK UNIFIED IDEOGRAPH-828B
        '33421' => '6518'             , # G+6518  U+828D  uni828D,  CJK UNIFIED IDEOGRAPH-828D
        '33422' => '3967'             , # G+3967  U+828E  uni828E,  CJK UNIFIED IDEOGRAPH-828E
        '33426' => '4703'             , # G+4703  U+8292  uni8292,  CJK UNIFIED IDEOGRAPH-8292
        '33433' => '5137'             , # G+5137  U+8299  uni8299,  CJK UNIFIED IDEOGRAPH-8299
        '33434' => '4360'             , # G+4360  U+829A  uni829A,  CJK UNIFIED IDEOGRAPH-829A
        '33437' => '6968'             , # G+6968  U+829D  uni829D,  CJK UNIFIED IDEOGRAPH-829D
        '33439' => '5324'             , # G+5324  U+829F  uni829F,  CJK UNIFIED IDEOGRAPH-829F
        '33445' => '3575'             , # G+3575  U+82A5  uni82A5,  CJK UNIFIED IDEOGRAPH-82A5
        '33446' => '7830'             , # G+7830  U+82A6  uni82A6,  CJK UNIFIED IDEOGRAPH-82A6
        '33449' => '4046'             , # G+4046  U+82A9  uni82A9,  CJK UNIFIED IDEOGRAPH-82A9
        '33452' => '5168'             , # G+5168  U+82AC  uni82AC,  CJK UNIFIED IDEOGRAPH-82AC
        '33453' => '7487'             , # G+7487  U+82AD  uni82AD,  CJK UNIFIED IDEOGRAPH-82AD
        '33454' => '6039'             , # G+6039  U+82AE  uni82AE,  CJK UNIFIED IDEOGRAPH-82AE
        '33455' => '5769'             , # G+5769  U+82AF  uni82AF,  CJK UNIFIED IDEOGRAPH-82AF
        '33457' => '7871'             , # G+7871  U+82B1  uni82B1,  CJK UNIFIED IDEOGRAPH-82B1
        '33459' => '4961'             , # G+4961  U+82B3  uni82B3,  CJK UNIFIED IDEOGRAPH-82B3
        '33463' => '6969'             , # G+6969  U+82B7  uni82B7,  CJK UNIFIED IDEOGRAPH-82B7
        '33464' => '6244'             , # G+6244  U+82B8  uni82B8,  CJK UNIFIED IDEOGRAPH-82B8
        '33465' => '4032'             , # G+4032  U+82B9  uni82B9,  CJK UNIFIED IDEOGRAPH-82B9
        '33467' => '7274'             , # G+7274  U+82BB  uni82BB,  CJK UNIFIED IDEOGRAPH-82BB
        '33468' => '4783'             , # G+4783  U+82BC  uni82BC,  CJK UNIFIED IDEOGRAPH-82BC
        '33469' => '5783'             , # G+5783  U+82BD  uni82BD,  CJK UNIFIED IDEOGRAPH-82BD
        '33471' => '6482'             , # G+6482  U+82BF  uni82BF,  CJK UNIFIED IDEOGRAPH-82BF
        '33489' => '6274'             , # G+6274  U+82D1  uni82D1,  CJK UNIFIED IDEOGRAPH-82D1
        '33490' => '5995'             , # G+5995  U+82D2  uni82D2,  CJK UNIFIED IDEOGRAPH-82D2
        '33492' => '7441'             , # G+7441  U+82D4  uni82D4,  CJK UNIFIED IDEOGRAPH-82D4
        '33493' => '7233'             , # G+7233  U+82D5  uni82D5,  CJK UNIFIED IDEOGRAPH-82D5
        '33495' => '4810'             , # G+4810  U+82D7  uni82D7,  CJK UNIFIED IDEOGRAPH-82D7
        '33499' => '3455'             , # G+3455  U+82DB  uni82DB,  CJK UNIFIED IDEOGRAPH-82DB
        '33502' => '7562'             , # G+7562  U+82DE  uni82DE,  CJK UNIFIED IDEOGRAPH-82DE
        '33503' => '3933'             , # G+3933  U+82DF  uni82DF,  CJK UNIFIED IDEOGRAPH-82DF
        '33505' => '6430'             , # G+6430  U+82E1  uni82E1,  CJK UNIFIED IDEOGRAPH-82E1
        '33509' => '5874'             , # G+5874  U+82E5  uni82E5,  CJK UNIFIED IDEOGRAPH-82E5
        '33510' => '3754'             , # G+3754  U+82E6  uni82E6,  CJK UNIFIED IDEOGRAPH-82E6
        '33511' => '6608'             , # G+6608  U+82E7  uni82E7,  CJK UNIFIED IDEOGRAPH-82E7
        '33521' => '6024'             , # G+6024  U+82F1  uni82F1,  CJK UNIFIED IDEOGRAPH-82F1
        '33533' => '3755'             , # G+3755  U+82FD  uni82FD,  CJK UNIFIED IDEOGRAPH-82FD
        '33534' => '7612'             , # G+7612  U+82FE  uni82FE,  CJK UNIFIED IDEOGRAPH-82FE
        '33537' => '6921'             , # G+6921  U+8301  uni8301,  CJK UNIFIED IDEOGRAPH-8301
        '33538' => '4828'             , # G+4828  U+8302  uni8302,  CJK UNIFIED IDEOGRAPH-8302
        '33539' => '5017'             , # G+5017  U+8303  uni8303,  CJK UNIFIED IDEOGRAPH-8303
        '33540' => '3456'             , # G+3456  U+8304  uni8304,  CJK UNIFIED IDEOGRAPH-8304
        '33541' => '4784'             , # G+4784  U+8305  uni8305,  CJK UNIFIED IDEOGRAPH-8305
        '33545' => '4693'             , # G+4693  U+8309  uni8309,  CJK UNIFIED IDEOGRAPH-8309
        '33559' => '4757'             , # G+4757  U+8317  uni8317,  CJK UNIFIED IDEOGRAPH-8317
        '33576' => '6503'             , # G+6503  U+8328  uni8328,  CJK UNIFIED IDEOGRAPH-8328
        '33579' => '4704'             , # G+4704  U+832B  uni832B,  CJK UNIFIED IDEOGRAPH-832B
        '33583' => '5083'             , # G+5083  U+832F  uni832F,  CJK UNIFIED IDEOGRAPH-832F
        '33585' => '5606'             , # G+5606  U+8331  uni8331,  CJK UNIFIED IDEOGRAPH-8331
        '33588' => '7944'             , # G+7944  U+8334  uni8334,  CJK UNIFIED IDEOGRAPH-8334
        '33589' => '6455'             , # G+6455  U+8335  uni8335,  CJK UNIFIED IDEOGRAPH-8335
        '33590' => '4191'             , # G+4191  U+8336  uni8336,  CJK UNIFIED IDEOGRAPH-8336
        '33592' => '6192'             , # G+6192  U+8338  uni8338,  CJK UNIFIED IDEOGRAPH-8338
        '33593' => '5943'             , # G+5943  U+8339  uni8339,  CJK UNIFIED IDEOGRAPH-8339
        '33600' => '5658'             , # G+5658  U+8340  uni8340,  CJK UNIFIED IDEOGRAPH-8340
        '33607' => '7708'             , # G+7708  U+8347  uni8347,  CJK UNIFIED IDEOGRAPH-8347
        '33609' => '7234'             , # G+7234  U+8349  uni8349,  CJK UNIFIED IDEOGRAPH-8349
        '33610' => '7784'             , # G+7784  U+834A  uni834A,  CJK UNIFIED IDEOGRAPH-834A
        '33615' => '6475'             , # G+6475  U+834F  uni834F,  CJK UNIFIED IDEOGRAPH-834F
        '33617' => '6431'             , # G+6431  U+8351  uni8351,  CJK UNIFIED IDEOGRAPH-8351
        '33618' => '7923'             , # G+7923  U+8352  uni8352,  CJK UNIFIED IDEOGRAPH-8352
        '33651' => '4354'             , # G+4354  U+8373  uni8373,  CJK UNIFIED IDEOGRAPH-8373
        '33655' => '7624'             , # G+7624  U+8377  uni8377,  CJK UNIFIED IDEOGRAPH-8377
        '33659' => '6633'             , # G+6633  U+837B  uni837B,  CJK UNIFIED IDEOGRAPH-837B
        '33673' => '4631'             , # G+4631  U+8389  uni8389,  CJK UNIFIED IDEOGRAPH-8389
        '33674' => '6560'             , # G+6560  U+838A  uni838A,  CJK UNIFIED IDEOGRAPH-838A
        '33678' => '5287'             , # G+5287  U+838E  uni838E,  CJK UNIFIED IDEOGRAPH-838E
        '33686' => '3696'             , # G+3696  U+8396  uni8396,  CJK UNIFIED IDEOGRAPH-8396
        '33688' => '5750'             , # G+5750  U+8398  uni8398,  CJK UNIFIED IDEOGRAPH-8398
        '33694' => '6121'             , # G+6121  U+839E  uni839E,  CJK UNIFIED IDEOGRAPH-839E
        '33698' => '7768'             , # G+7768  U+83A2  uni83A2,  CJK UNIFIED IDEOGRAPH-83A2
        '33705' => '5138'             , # G+5138  U+83A9  uni83A9,  CJK UNIFIED IDEOGRAPH-83A9
        '33706' => '5784'             , # G+5784  U+83AA  uni83AA,  CJK UNIFIED IDEOGRAPH-83AA
        '33707' => '4668'             , # G+4668  U+83AB  uni83AB,  CJK UNIFIED IDEOGRAPH-83AB
        '33725' => '4705'             , # G+4705  U+83BD  uni83BD,  CJK UNIFIED IDEOGRAPH-83BD
        '33729' => '7201'             , # G+7201  U+83C1  uni83C1,  CJK UNIFIED IDEOGRAPH-83C1
        '33733' => '3830'             , # G+3830  U+83C5  uni83C5,  CJK UNIFIED IDEOGRAPH-83C5
        '33737' => '4531'             , # G+4531  U+83C9  uni83C9,  CJK UNIFIED IDEOGRAPH-83C9
        '33738' => '3949'             , # G+3949  U+83CA  uni83CA,  CJK UNIFIED IDEOGRAPH-83CA
        '33740' => '4013'             , # G+4013  U+83CC  uni83CC,  CJK UNIFIED IDEOGRAPH-83CC
        '33747' => '3808'             , # G+3808  U+83D3  uni83D3,  CJK UNIFIED IDEOGRAPH-83D3
        '33750' => '7112'             , # G+7112  U+83D6  uni83D6,  CJK UNIFIED IDEOGRAPH-83D6
        '33756' => '7122'             , # G+7122  U+83DC  uni83DC,  CJK UNIFIED IDEOGRAPH-83DC
        '33769' => '5069'             , # G+5069  U+83E9  uni83E9,  CJK UNIFIED IDEOGRAPH-83E9
        '33771' => '4033'             , # G+4033  U+83EB  uni83EB,  CJK UNIFIED IDEOGRAPH-83EB
        '33775' => '7872'             , # G+7872  U+83EF  uni83EF,  CJK UNIFIED IDEOGRAPH-83EF
        '33776' => '3756'             , # G+3756  U+83F0  uni83F0,  CJK UNIFIED IDEOGRAPH-83F0
        '33777' => '4611'             , # G+4611  U+83F1  uni83F1,  CJK UNIFIED IDEOGRAPH-83F1
        '33778' => '5213'             , # G+5213  U+83F2  uni83F2,  CJK UNIFIED IDEOGRAPH-83F2
        '33780' => '5827'             , # G+5827  U+83F4  uni83F4,  CJK UNIFIED IDEOGRAPH-83F4
        '33785' => '6609'             , # G+6609  U+83F9  uni83F9,  CJK UNIFIED IDEOGRAPH-83F9
        '33789' => '5639'             , # G+5639  U+83FD  uni83FD,  CJK UNIFIED IDEOGRAPH-83FD
        '33795' => '7314'             , # G+7314  U+8403  uni8403,  CJK UNIFIED IDEOGRAPH-8403
        '33796' => '4297'             , # G+4297  U+8404  uni8404,  CJK UNIFIED IDEOGRAPH-8404
        '33802' => '4423'             , # G+4423  U+840A  uni840A,  CJK UNIFIED IDEOGRAPH-840A
        '33804' => '4732'             , # G+4732  U+840C  uni840C,  CJK UNIFIED IDEOGRAPH-840C
        '33805' => '7532'             , # G+7532  U+840D  uni840D,  CJK UNIFIED IDEOGRAPH-840D
        '33806' => '6299'             , # G+6299  U+840E  uni840E,  CJK UNIFIED IDEOGRAPH-840E
        '33833' => '7275'             , # G+7275  U+8429  uni8429,  CJK UNIFIED IDEOGRAPH-8429
        '33836' => '4683'             , # G+4683  U+842C  uni842C,  CJK UNIFIED IDEOGRAPH-842C
        '33841' => '7993'             , # G+7993  U+8431  uni8431,  CJK UNIFIED IDEOGRAPH-8431
        '33848' => '6342'             , # G+6342  U+8438  uni8438,  CJK UNIFIED IDEOGRAPH-8438
        '33853' => '4386'             , # G+4386  U+843D  uni843D,  CJK UNIFIED IDEOGRAPH-843D
        '33865' => '6001'             , # G+6001  U+8449  uni8449,  CJK UNIFIED IDEOGRAPH-8449
        '33879' => '6610'             , # G+6610  U+8457  uni8457,  CJK UNIFIED IDEOGRAPH-8457
        '33883' => '3506'             , # G+3506  U+845B  uni845B,  CJK UNIFIED IDEOGRAPH-845B
        '33889' => '7563'             , # G+7563  U+8461  uni8461,  CJK UNIFIED IDEOGRAPH-8461
        '33891' => '4346'             , # G+4346  U+8463  uni8463,  CJK UNIFIED IDEOGRAPH-8463
        '33894' => '6300'             , # G+6300  U+8466  uni8466,  CJK UNIFIED IDEOGRAPH-8466
        '33899' => '7831'             , # G+7831  U+846B  uni846B,  CJK UNIFIED IDEOGRAPH-846B
        '33900' => '6561'             , # G+6561  U+846C  uni846C,  CJK UNIFIED IDEOGRAPH-846C
        '33903' => '5875'             , # G+5875  U+846F  uni846F,  CJK UNIFIED IDEOGRAPH-846F
        '33909' => '4004'             , # G+4004  U+8475  uni8475,  CJK UNIFIED IDEOGRAPH-8475
        '33914' => '6930'             , # G+6930  U+847A  uni847A,  CJK UNIFIED IDEOGRAPH-847A
        '33936' => '5607'             , # G+5607  U+8490  uni8490,  CJK UNIFIED IDEOGRAPH-8490
        '33940' => '5712'             , # G+5712  U+8494  uni8494,  CJK UNIFIED IDEOGRAPH-8494
        '33945' => '4799'             , # G+4799  U+8499  uni8499,  CJK UNIFIED IDEOGRAPH-8499
        '33948' => '5312'             , # G+5312  U+849C  uni849C,  CJK UNIFIED IDEOGRAPH-849C
        '33953' => '4962'             , # G+4962  U+84A1  uni84A1,  CJK UNIFIED IDEOGRAPH-84A1
        '33970' => '7564'             , # G+7564  U+84B2  uni84B2,  CJK UNIFIED IDEOGRAPH-84B2
        '33976' => '6939'             , # G+6939  U+84B8  uni84B8,  CJK UNIFIED IDEOGRAPH-84B8
        '33979' => '5876'             , # G+5876  U+84BB  uni84BB,  CJK UNIFIED IDEOGRAPH-84BB
        '33980' => '7113'             , # G+7113  U+84BC  uni84BC,  CJK UNIFIED IDEOGRAPH-84BC
        '33983' => '7832'             , # G+7832  U+84BF  uni84BF,  CJK UNIFIED IDEOGRAPH-84BF
        '33984' => '5549'             , # G+5549  U+84C0  uni84C0,  CJK UNIFIED IDEOGRAPH-84C0
        '33986' => '4758'             , # G+4758  U+84C2  uni84C2,  CJK UNIFIED IDEOGRAPH-84C2
        '33988' => '7295'             , # G+7295  U+84C4  uni84C4,  CJK UNIFIED IDEOGRAPH-84C4
        '33990' => '5415'             , # G+5415  U+84C6  uni84C6,  CJK UNIFIED IDEOGRAPH-84C6
        '33993' => '6193'             , # G+6193  U+84C9  uni84C9,  CJK UNIFIED IDEOGRAPH-84C9
        '33995' => '3576'             , # G+3576  U+84CB  uni84CB,  CJK UNIFIED IDEOGRAPH-84CB
        '33997' => '5713'             , # G+5713  U+84CD  uni84CD,  CJK UNIFIED IDEOGRAPH-84CD
        '34001' => '5288'             , # G+5288  U+84D1  uni84D1,  CJK UNIFIED IDEOGRAPH-84D1
        '34010' => '5608'             , # G+5608  U+84DA  uni84DA,  CJK UNIFIED IDEOGRAPH-84DA
        '34028' => '5104'             , # G+5104  U+84EC  uni84EC,  CJK UNIFIED IDEOGRAPH-84EC
        '34030' => '4473'             , # G+4473  U+84EE  uni84EE,  CJK UNIFIED IDEOGRAPH-84EE
        '34036' => '5659'             , # G+5659  U+84F4  uni84F4,  CJK UNIFIED IDEOGRAPH-84F4
        '34044' => '4560'             , # G+4560  U+84FC  uni84FC,  CJK UNIFIED IDEOGRAPH-84FC
        '34065' => '4747'             , # G+4747  U+8511  uni8511,  CJK UNIFIED IDEOGRAPH-8511
        '34067' => '4684'             , # G+4684  U+8513  uni8513,  CJK UNIFIED IDEOGRAPH-8513
        '34068' => '5084'             , # G+5084  U+8514  uni8514,  CJK UNIFIED IDEOGRAPH-8514
        '34071' => '6504'             , # G+6504  U+8517  uni8517,  CJK UNIFIED IDEOGRAPH-8517
        '34072' => '5325'             , # G+5325  U+8518  uni8518,  CJK UNIFIED IDEOGRAPH-8518
        '34074' => '6250'             , # G+6250  U+851A  uni851A,  CJK UNIFIED IDEOGRAPH-851A
        '34078' => '4573'             , # G+4573  U+851E  uni851E,  CJK UNIFIED IDEOGRAPH-851E
        '34081' => '7123'             , # G+7123  U+8521  uni8521,  CJK UNIFIED IDEOGRAPH-8521
        '34083' => '6562'             , # G+6562  U+8523  uni8523,  CJK UNIFIED IDEOGRAPH-8523
        '34085' => '7259'             , # G+7259  U+8525  uni8525,  CJK UNIFIED IDEOGRAPH-8525
        '34092' => '5527'             , # G+5527  U+852C  uni852C,  CJK UNIFIED IDEOGRAPH-852C
        '34093' => '6383'             , # G+6383  U+852D  uni852D,  CJK UNIFIED IDEOGRAPH-852D
        '34095' => '7004'             , # G+7004  U+852F  uni852F,  CJK UNIFIED IDEOGRAPH-852F
        '34109' => '7541'             , # G+7541  U+853D  uni853D,  CJK UNIFIED IDEOGRAPH-853D
        '34111' => '6301'             , # G+6301  U+853F  uni853F,  CJK UNIFIED IDEOGRAPH-853F
        '34113' => '4229'             , # G+4229  U+8541  uni8541,  CJK UNIFIED IDEOGRAPH-8541
        '34115' => '5002'             , # G+5002  U+8543  uni8543,  CJK UNIFIED IDEOGRAPH-8543
        '34121' => '7235'             , # G+7235  U+8549  uni8549,  CJK UNIFIED IDEOGRAPH-8549
        '34126' => '3885'             , # G+3885  U+854E  uni854E,  CJK UNIFIED IDEOGRAPH-854E
        '34131' => '6245'             , # G+6245  U+8553  uni8553,  CJK UNIFIED IDEOGRAPH-8553
        '34137' => '7796'             , # G+7796  U+8559  uni8559,  CJK UNIFIED IDEOGRAPH-8559
        '34147' => '5660'             , # G+5660  U+8563  uni8563,  CJK UNIFIED IDEOGRAPH-8563
        '34152' => '3981'             , # G+3981  U+8568  uni8568,  CJK UNIFIED IDEOGRAPH-8568
        '34153' => '7430'             , # G+7430  U+8569  uni8569,  CJK UNIFIED IDEOGRAPH-8569
        '34154' => '4829'             , # G+4829  U+856A  uni856A,  CJK UNIFIED IDEOGRAPH-856A
        '34157' => '5528'             , # G+5528  U+856D  uni856D,  CJK UNIFIED IDEOGRAPH-856D
        '34180' => '4901'             , # G+4901  U+8584  uni8584,  CJK UNIFIED IDEOGRAPH-8584
        '34183' => '4865'             , # G+4865  U+8587  uni8587,  CJK UNIFIED IDEOGRAPH-8587
        '34191' => '6407'             , # G+6407  U+858F  uni858F,  CJK UNIFIED IDEOGRAPH-858F
        '34193' => '3554'             , # G+3554  U+8591  uni8591,  CJK UNIFIED IDEOGRAPH-8591
        '34196' => '6563'             , # G+6563  U+8594  uni8594,  CJK UNIFIED IDEOGRAPH-8594
        '34203' => '5457'             , # G+5457  U+859B  uni859B,  CJK UNIFIED IDEOGRAPH-859B
        '34214' => '7159'             , # G+7159  U+85A6  uni85A6,  CJK UNIFIED IDEOGRAPH-85A6
        '34216' => '7989'             , # G+7989  U+85A8  uni85A8,  CJK UNIFIED IDEOGRAPH-85A8
        '34217' => '5319'             , # G+5319  U+85A9  uni85A9,  CJK UNIFIED IDEOGRAPH-85A9
        '34218' => '5751'             , # G+5751  U+85AA  uni85AA,  CJK UNIFIED IDEOGRAPH-85AA
        '34223' => '5396'             , # G+5396  U+85AF  uni85AF,  CJK UNIFIED IDEOGRAPH-85AF
        '34224' => '7986'             , # G+7986  U+85B0  uni85B0,  CJK UNIFIED IDEOGRAPH-85B0
        '34234' => '6772'             , # G+6772  U+85BA  uni85BA,  CJK UNIFIED IDEOGRAPH-85BA
        '34241' => '3757'             , # G+3757  U+85C1  uni85C1,  CJK UNIFIED IDEOGRAPH-85C1
        '34249' => '6505'             , # G+6505  U+85C9  uni85C9,  CJK UNIFIED IDEOGRAPH-85C9
        '34253' => '4406'             , # G+4406  U+85CD  uni85CD,  CJK UNIFIED IDEOGRAPH-85CD
        '34254' => '5752'             , # G+5752  U+85CE  uni85CE,  CJK UNIFIED IDEOGRAPH-85CE
        '34255' => '6564'             , # G+6564  U+85CF  uni85CF,  CJK UNIFIED IDEOGRAPH-85CF
        '34261' => '6220'             , # G+6220  U+85D5  uni85D5,  CJK UNIFIED IDEOGRAPH-85D5
        '34268' => '4451'             , # G+4451  U+85DC  uni85DC,  CJK UNIFIED IDEOGRAPH-85DC
        '34269' => '6040'             , # G+6040  U+85DD  uni85DD,  CJK UNIFIED IDEOGRAPH-85DD
        '34276' => '4370'             , # G+4370  U+85E4  uni85E4,  CJK UNIFIED IDEOGRAPH-85E4
        '34277' => '5877'             , # G+5877  U+85E5  uni85E5,  CJK UNIFIED IDEOGRAPH-85E5
        '34281' => '5003'             , # G+5003  U+85E9  uni85E9,  CJK UNIFIED IDEOGRAPH-85E9
        '34282' => '5609'             , # G+5609  U+85EA  uni85EA,  CJK UNIFIED IDEOGRAPH-85EA
        '34295' => '6611'             , # G+6611  U+85F7  uni85F7,  CJK UNIFIED IDEOGRAPH-85F7
        '34298' => '4642'             , # G+4642  U+85FA  uni85FA,  CJK UNIFIED IDEOGRAPH-85FA
        '34299' => '6816'             , # G+6816  U+85FB  uni85FB,  CJK UNIFIED IDEOGRAPH-85FB
        '34303' => '3817'             , # G+3817  U+85FF  uni85FF,  CJK UNIFIED IDEOGRAPH-85FF
        '34306' => '6041'             , # G+6041  U+8602  uni8602,  CJK UNIFIED IDEOGRAPH-8602
        '34310' => '4520'             , # G+4520  U+8606  uni8606,  CJK UNIFIED IDEOGRAPH-8606
        '34311' => '5529'             , # G+5529  U+8607  uni8607,  CJK UNIFIED IDEOGRAPH-8607
        '34314' => '6088'             , # G+6088  U+860A  uni860A,  CJK UNIFIED IDEOGRAPH-860A
        '34326' => '5924'             , # G+5924  U+8616  uni8616,  CJK UNIFIED IDEOGRAPH-8616
        '34327' => '5028'             , # G+5028  U+8617  uni8617,  CJK UNIFIED IDEOGRAPH-8617
        '34330' => '5441'             , # G+5441  U+861A  uni861A,  CJK UNIFIED IDEOGRAPH-861A
        '34349' => '4395'             , # G+4395  U+862D  uni862D,  CJK UNIFIED IDEOGRAPH-862D
        '34367' => '4378'             , # G+4378  U+863F  uni863F,  CJK UNIFIED IDEOGRAPH-863F
        '34382' => '7833'             , # G+7833  U+864E  uni864E,  CJK UNIFIED IDEOGRAPH-864E
        '34384' => '7632'             , # G+7632  U+8650  uni8650,  CJK UNIFIED IDEOGRAPH-8650
        '34388' => '3609'             , # G+3609  U+8654  uni8654,  CJK UNIFIED IDEOGRAPH-8654
        '34389' => '7133'             , # G+7133  U+8655  uni8655,  CJK UNIFIED IDEOGRAPH-8655
        '34395' => '7721'             , # G+7721  U+865B  uni865B,  CJK UNIFIED IDEOGRAPH-865B
        '34396' => '4521'             , # G+4521  U+865C  uni865C,  CJK UNIFIED IDEOGRAPH-865C
        '34398' => '6221'             , # G+6221  U+865E  uni865E,  CJK UNIFIED IDEOGRAPH-865E
        '34399' => '7834'             , # G+7834  U+865F  uni865F,  CJK UNIFIED IDEOGRAPH-865F
        '34407' => '8009'             , # G+8009  U+8667  uni8667,  CJK UNIFIED IDEOGRAPH-8667
        '34425' => '7860'             , # G+7860  U+8679  uni8679,  CJK UNIFIED IDEOGRAPH-8679
        '34442' => '4845'             , # G+4845  U+868A  uni868A,  CJK UNIFIED IDEOGRAPH-868A
        '34444' => '4963'             , # G+4963  U+868C  uni868C,  CJK UNIFIED IDEOGRAPH-868C
        '34451' => '6456'             , # G+6456  U+8693  uni8693,  CJK UNIFIED IDEOGRAPH-8693
        '34467' => '3799'             , # G+3799  U+86A3  uni86A3,  CJK UNIFIED IDEOGRAPH-86A3
        '34468' => '6817'             , # G+6817  U+86A4  uni86A4,  CJK UNIFIED IDEOGRAPH-86A4
        '34473' => '7355'             , # G+7355  U+86A9  uni86A9,  CJK UNIFIED IDEOGRAPH-86A9
        '34503' => '5289'             , # G+5289  U+86C7  uni86C7,  CJK UNIFIED IDEOGRAPH-86C7
        '34507' => '4208'             , # G+4208  U+86CB  uni86CB,  CJK UNIFIED IDEOGRAPH-86CB
        '34516' => '7945'             , # G+7945  U+86D4  uni86D4,  CJK UNIFIED IDEOGRAPH-86D4
        '34521' => '6104'             , # G+6104  U+86D9  uni86D9,  CJK UNIFIED IDEOGRAPH-86D9
        '34523' => '6889'             , # G+6889  U+86DB  uni86DB,  CJK UNIFIED IDEOGRAPH-86DB
        '34527' => '3886'             , # G+3886  U+86DF  uni86DF,  CJK UNIFIED IDEOGRAPH-86DF
        '34532' => '7666'             , # G+7666  U+86E4  uni86E4,  CJK UNIFIED IDEOGRAPH-86E4
        '34541' => '7026'             , # G+7026  U+86ED  uni86ED,  CJK UNIFIED IDEOGRAPH-86ED
        '34558' => '5785'             , # G+5785  U+86FE  uni86FE,  CJK UNIFIED IDEOGRAPH-86FE
        '34560' => '7245'             , # G+7245  U+8700  uni8700,  CJK UNIFIED IDEOGRAPH-8700
        '34562' => '5105'             , # G+5105  U+8702  uni8702,  CJK UNIFIED IDEOGRAPH-8702
        '34563' => '5753'             , # G+5753  U+8703  uni8703,  CJK UNIFIED IDEOGRAPH-8703
        '34568' => '6074'             , # G+6074  U+8708  uni8708,  CJK UNIFIED IDEOGRAPH-8708
        '34584' => '6970'             , # G+6970  U+8718  uni8718,  CJK UNIFIED IDEOGRAPH-8718
        '34586' => '5214'             , # G+5214  U+871A  uni871A,  CJK UNIFIED IDEOGRAPH-871A
        '34588' => '4884'             , # G+4884  U+871C  uni871C,  CJK UNIFIED IDEOGRAPH-871C
        '34638' => '3508'             , # G+3508  U+874E  uni874E,  CJK UNIFIED IDEOGRAPH-874E
        '34645' => '5730'             , # G+5730  U+8755  uni8755,  CJK UNIFIED IDEOGRAPH-8755
        '34647' => '7924'             , # G+7924  U+8757  uni8757,  CJK UNIFIED IDEOGRAPH-8757
        '34655' => '6302'             , # G+6302  U+875F  uni875F,  CJK UNIFIED IDEOGRAPH-875F
        '34662' => '7625'             , # G+7625  U+8766  uni8766,  CJK UNIFIED IDEOGRAPH-8766
        '34664' => '5676'             , # G+5676  U+8768  uni8768,  CJK UNIFIED IDEOGRAPH-8768
        '34676' => '7835'             , # G+7835  U+8774  uni8774,  CJK UNIFIED IDEOGRAPH-8774
        '34678' => '6703'             , # G+6703  U+8776  uni8776,  CJK UNIFIED IDEOGRAPH-8776
        '34680' => '6105'             , # G+6105  U+8778  uni8778,  CJK UNIFIED IDEOGRAPH-8778
        '34690' => '4418'             , # G+4418  U+8782  uni8782,  CJK UNIFIED IDEOGRAPH-8782
        '34701' => '6372'             , # G+6372  U+878D  uni878D,  CJK UNIFIED IDEOGRAPH-878D
        '34719' => '4759'             , # G+4759  U+879F  uni879F,  CJK UNIFIED IDEOGRAPH-879F
        '34722' => '7785'             , # G+7785  U+87A2  uni87A2,  CJK UNIFIED IDEOGRAPH-87A2
        '34739' => '4248'             , # G+4248  U+87B3  uni87B3,  CJK UNIFIED IDEOGRAPH-87B3
        '34746' => '4379'             , # G+4379  U+87BA  uni87BA,  CJK UNIFIED IDEOGRAPH-87BA
        '34756' => '7376'             , # G+7376  U+87C4  uni87C4,  CJK UNIFIED IDEOGRAPH-87C4
        '34784' => '4926'             , # G+4926  U+87E0  uni87E0,  CJK UNIFIED IDEOGRAPH-87E0
        '34796' => '5442'             , # G+5442  U+87EC  uni87EC,  CJK UNIFIED IDEOGRAPH-87EC
        '34799' => '6162'             , # G+6162  U+87EF  uni87EF,  CJK UNIFIED IDEOGRAPH-87EF
        '34802' => '7309'             , # G+7309  U+87F2  uni87F2,  CJK UNIFIED IDEOGRAPH-87F2
        '34809' => '7696'             , # G+7696  U+87F9  uni87F9,  CJK UNIFIED IDEOGRAPH-87F9
        '34811' => '6408'             , # G+6408  U+87FB  uni87FB,  CJK UNIFIED IDEOGRAPH-87FB
        '34814' => '5467'             , # G+5467  U+87FE  uni87FE,  CJK UNIFIED IDEOGRAPH-87FE
        '34821' => '5690'             , # G+5690  U+8805  uni8805,  CJK UNIFIED IDEOGRAPH-8805
        '34847' => '4411'             , # G+4411  U+881F  uni881F,  CJK UNIFIED IDEOGRAPH-881F
        '34850' => '6916'             , # G+6916  U+8822  uni8822,  CJK UNIFIED IDEOGRAPH-8822
        '34851' => '4452'             , # G+4452  U+8823  uni8823,  CJK UNIFIED IDEOGRAPH-8823
        '34865' => '3758'             , # G+3758  U+8831  uni8831,  CJK UNIFIED IDEOGRAPH-8831
        '34870' => '6532'             , # G+6532  U+8836  uni8836,  CJK UNIFIED IDEOGRAPH-8836
        '34875' => '4685'             , # G+4685  U+883B  uni883B,  CJK UNIFIED IDEOGRAPH-883B
        '34880' => '7756'             , # G+7756  U+8840  uni8840,  CJK UNIFIED IDEOGRAPH-8840
        '34886' => '6924'             , # G+6924  U+8846  uni8846,  CJK UNIFIED IDEOGRAPH-8846
        '34892' => '7709'             , # G+7709  U+884C  uni884C,  CJK UNIFIED IDEOGRAPH-884C
        '34893' => '5981'             , # G+5981  U+884D  uni884D,  CJK UNIFIED IDEOGRAPH-884D
        '34898' => '7750'             , # G+7750  U+8852  uni8852,  CJK UNIFIED IDEOGRAPH-8852
        '34899' => '5668'             , # G+5668  U+8853  uni8853,  CJK UNIFIED IDEOGRAPH-8853
        '34903' => '3457'             , # G+3457  U+8857  uni8857,  CJK UNIFIED IDEOGRAPH-8857
        '34905' => '5786'             , # G+5786  U+8859  uni8859,  CJK UNIFIED IDEOGRAPH-8859
        '34907' => '6303'             , # G+6303  U+885B  uni885B,  CJK UNIFIED IDEOGRAPH-885B
        '34909' => '7310'             , # G+7310  U+885D  uni885D,  CJK UNIFIED IDEOGRAPH-885D
        '34913' => '7786'             , # G+7786  U+8861  uni8861,  CJK UNIFIED IDEOGRAPH-8861
        '34914' => '3934'             , # G+3934  U+8862  uni8862,  CJK UNIFIED IDEOGRAPH-8862
        '34915' => '6409'             , # G+6409  U+8863  uni8863,  CJK UNIFIED IDEOGRAPH-8863
        '34920' => '7585'             , # G+7585  U+8868  uni8868,  CJK UNIFIED IDEOGRAPH-8868
        '34923' => '5326'             , # G+5326  U+886B  uni886B,  CJK UNIFIED IDEOGRAPH-886B
        '34928' => '5565'             , # G+5565  U+8870  uni8870,  CJK UNIFIED IDEOGRAPH-8870
        '34930' => '4151'             , # G+4151  U+8872  uni8872,  CJK UNIFIED IDEOGRAPH-8872
        '34935' => '7311'             , # G+7311  U+8877  uni8877,  CJK UNIFIED IDEOGRAPH-8877
        '34942' => '4047'             , # G+4047  U+887E  uni887E,  CJK UNIFIED IDEOGRAPH-887E
        '34943' => '4048'             , # G+4048  U+887F  uni887F,  CJK UNIFIED IDEOGRAPH-887F
        '34945' => '6275'             , # G+6275  U+8881  uni8881,  CJK UNIFIED IDEOGRAPH-8881
        '34946' => '4763'             , # G+4763  U+8882  uni8882,  CJK UNIFIED IDEOGRAPH-8882
        '34952' => '3458'             , # G+3458  U+8888  uni8888,  CJK UNIFIED IDEOGRAPH-8888
        '34955' => '4262'             , # G+4262  U+888B  uni888B,  CJK UNIFIED IDEOGRAPH-888B
        '34957' => '7565'             , # G+7565  U+888D  uni888D,  CJK UNIFIED IDEOGRAPH-888D
        '34962' => '4209'             , # G+4209  U+8892  uni8892,  CJK UNIFIED IDEOGRAPH-8892
        '34966' => '5610'             , # G+5610  U+8896  uni8896,  CJK UNIFIED IDEOGRAPH-8896
        '34967' => '7005'             , # G+7005  U+8897  uni8897,  CJK UNIFIED IDEOGRAPH-8897
        '34974' => '3782'             , # G+3782  U+889E  uni889E,  CJK UNIFIED IDEOGRAPH-889E
        '34987' => '7601'             , # G+7601  U+88AB  uni88AB,  CJK UNIFIED IDEOGRAPH-88AB
        '34996' => '3759'             , # G+3759  U+88B4  uni88B4,  CJK UNIFIED IDEOGRAPH-88B4
        '35009' => '6582'             , # G+6582  U+88C1  uni88C1,  CJK UNIFIED IDEOGRAPH-88C1
        '35010' => '4482'             , # G+4482  U+88C2  uni88C2,  CJK UNIFIED IDEOGRAPH-88C2
        '35023' => '4632'             , # G+4632  U+88CF  uni88CF,  CJK UNIFIED IDEOGRAPH-88CF
        '35028' => '6042'             , # G+6042  U+88D4  uni88D4,  CJK UNIFIED IDEOGRAPH-88D4
        '35029' => '6343'             , # G+6343  U+88D5  uni88D5,  CJK UNIFIED IDEOGRAPH-88D5
        '35033' => '3956'             , # G+3956  U+88D9  uni88D9,  CJK UNIFIED IDEOGRAPH-88D9
        '35036' => '5070'             , # G+5070  U+88DC  uni88DC,  CJK UNIFIED IDEOGRAPH-88DC
        '35037' => '6565'             , # G+6565  U+88DD  uni88DD,  CJK UNIFIED IDEOGRAPH-88DD
        '35039' => '5290'             , # G+5290  U+88DF  uni88DF,  CJK UNIFIED IDEOGRAPH-88DF
        '35041' => '4633'             , # G+4633  U+88E1  uni88E1,  CJK UNIFIED IDEOGRAPH-88E1
        '35048' => '5215'             , # G+5215  U+88E8  uni88E8,  CJK UNIFIED IDEOGRAPH-88E8
        '35059' => '5356'             , # G+5356  U+88F3  uni88F3,  CJK UNIFIED IDEOGRAPH-88F3
        '35060' => '4981'             , # G+4981  U+88F4  uni88F4,  CJK UNIFIED IDEOGRAPH-88F4
        '35061' => '4982'             , # G+4982  U+88F5  uni88F5,  CJK UNIFIED IDEOGRAPH-88F5
        '35064' => '4380'             , # G+4380  U+88F8  uni88F8,  CJK UNIFIED IDEOGRAPH-88F8
        '35069' => '6773'             , # G+6773  U+88FD  uni88FD,  CJK UNIFIED IDEOGRAPH-88FD
        '35079' => '5085'             , # G+5085  U+8907  uni8907,  CJK UNIFIED IDEOGRAPH-8907
        '35088' => '3507'             , # G+3507  U+8910  uni8910,  CJK UNIFIED IDEOGRAPH-8910
        '35090' => '7566'             , # G+7566  U+8912  uni8912,  CJK UNIFIED IDEOGRAPH-8912
        '35091' => '5071'             , # G+5071  U+8913  uni8913,  CJK UNIFIED IDEOGRAPH-8913
        '35096' => '6304'             , # G+6304  U+8918  uni8918,  CJK UNIFIED IDEOGRAPH-8918
        '35097' => '4983'             , # G+4983  U+8919  uni8919,  CJK UNIFIED IDEOGRAPH-8919
        '35109' => '6172'             , # G+6172  U+8925  uni8925,  CJK UNIFIED IDEOGRAPH-8925
        '35114' => '7462'             , # G+7462  U+892A  uni892A,  CJK UNIFIED IDEOGRAPH-892A
        '35126' => '5680'             , # G+5680  U+8936  uni8936,  CJK UNIFIED IDEOGRAPH-8936
        '35128' => '4574'             , # G+4574  U+8938  uni8938,  CJK UNIFIED IDEOGRAPH-8938
        '35131' => '5458'             , # G+5458  U+893B  uni893B,  CJK UNIFIED IDEOGRAPH-893B
        '35137' => '3555'             , # G+3555  U+8941  uni8941,  CJK UNIFIED IDEOGRAPH-8941
        '35140' => '5897'             , # G+5897  U+8944  uni8944,  CJK UNIFIED IDEOGRAPH-8944
        '35167' => '4049'             , # G+4049  U+895F  uni895F,  CJK UNIFIED IDEOGRAPH-895F
        '35172' => '4407'             , # G+4407  U+8964  uni8964,  CJK UNIFIED IDEOGRAPH-8964
        '35178' => '4694'             , # G+4694  U+896A  uni896A,  CJK UNIFIED IDEOGRAPH-896A
        '35186' => '5681'             , # G+5681  U+8972  uni8972,  CJK UNIFIED IDEOGRAPH-8972
        '35199' => '5397'             , # G+5397  U+897F  uni897F,  CJK UNIFIED IDEOGRAPH-897F
        '35201' => '6163'             , # G+6163  U+8981  uni8981,  CJK UNIFIED IDEOGRAPH-8981
        '35203' => '4230'             , # G+4230  U+8983  uni8983,  CJK UNIFIED IDEOGRAPH-8983
        '35206' => '5086'             , # G+5086  U+8986  uni8986,  CJK UNIFIED IDEOGRAPH-8986
        '35207' => '7511'             , # G+7511  U+8987  uni8987,  CJK UNIFIED IDEOGRAPH-8987
        '35211' => '3644'             , # G+3644  U+898B  uni898B,  CJK UNIFIED IDEOGRAPH-898B
        '35215' => '4005'             , # G+4005  U+898F  uni898F,  CJK UNIFIED IDEOGRAPH-898F
        '35219' => '4734'             , # G+4734  U+8993  uni8993,  CJK UNIFIED IDEOGRAPH-8993
        '35222' => '5714'             , # G+5714  U+8996  uni8996,  CJK UNIFIED IDEOGRAPH-8996
        '35233' => '3635'             , # G+3635  U+89A1  uni89A1,  CJK UNIFIED IDEOGRAPH-89A1
        '35241' => '4298'             , # G+4298  U+89A9  uni89A9,  CJK UNIFIED IDEOGRAPH-89A9
        '35242' => '7363'             , # G+7363  U+89AA  uni89AA,  CJK UNIFIED IDEOGRAPH-89AA
        '35250' => '4034'             , # G+4034  U+89B2  uni89B2,  CJK UNIFIED IDEOGRAPH-89B2
        '35258' => '3473'             , # G+3473  U+89BA  uni89BA,  CJK UNIFIED IDEOGRAPH-89BA
        '35261' => '4408'             , # G+4408  U+89BD  uni89BD,  CJK UNIFIED IDEOGRAPH-89BD
        '35264' => '3831'             , # G+3831  U+89C0  uni89C0,  CJK UNIFIED IDEOGRAPH-89C0
        '35282' => '3474'             , # G+3474  U+89D2  uni89D2,  CJK UNIFIED IDEOGRAPH-89D2
        '35299' => '7697'             , # G+7697  U+89E3  uni89E3,  CJK UNIFIED IDEOGRAPH-89E3
        '35316' => '5357'             , # G+5357  U+89F4  uni89F4,  CJK UNIFIED IDEOGRAPH-89F4
        '35320' => '7246'             , # G+7246  U+89F8  uni89F8,  CJK UNIFIED IDEOGRAPH-89F8
        '35328' => '5921'             , # G+5921  U+8A00  uni8A00,  CJK UNIFIED IDEOGRAPH-8A00
        '35330' => '6745'             , # G+6745  U+8A02  uni8A02,  CJK UNIFIED IDEOGRAPH-8A02
        '35331' => '5139'             , # G+5139  U+8A03  uni8A03,  CJK UNIFIED IDEOGRAPH-8A03
        '35336' => '3724'             , # G+3724  U+8A08  uni8A08,  CJK UNIFIED IDEOGRAPH-8A08
        '35338' => '5754'             , # G+5754  U+8A0A  uni8A0A,  CJK UNIFIED IDEOGRAPH-8A0A
        '35340' => '7861'             , # G+7861  U+8A0C  uni8A0C,  CJK UNIFIED IDEOGRAPH-8A0C
        '35342' => '7452'             , # G+7452  U+8A0E  uni8A0E,  CJK UNIFIED IDEOGRAPH-8A0E
        '35347' => '7987'             , # G+7987  U+8A13  uni8A13,  CJK UNIFIED IDEOGRAPH-8A13
        '35350' => '8026'             , # G+8026  U+8A16  uni8A16,  CJK UNIFIED IDEOGRAPH-8A16
        '35351' => '7406'             , # G+7406  U+8A17  uni8A17,  CJK UNIFIED IDEOGRAPH-8A17
        '35352' => '4114'             , # G+4114  U+8A18  uni8A18,  CJK UNIFIED IDEOGRAPH-8A18
        '35355' => '6106'             , # G+6106  U+8A1B  uni8A1B,  CJK UNIFIED IDEOGRAPH-8A1B
        '35357' => '5787'             , # G+5787  U+8A1D  uni8A1D,  CJK UNIFIED IDEOGRAPH-8A1D
        '35359' => '5557'             , # G+5557  U+8A1F  uni8A1F,  CJK UNIFIED IDEOGRAPH-8A1F
        '35363' => '3653'             , # G+3653  U+8A23  uni8A23,  CJK UNIFIED IDEOGRAPH-8A23
        '35365' => '4182'             , # G+4182  U+8A25  uni8A25,  CJK UNIFIED IDEOGRAPH-8A25
        '35370' => '4964'             , # G+4964  U+8A2A  uni8A2A,  CJK UNIFIED IDEOGRAPH-8A2A
        '35373' => '5459'             , # G+5459  U+8A2D  uni8A2D,  CJK UNIFIED IDEOGRAPH-8A2D
        '35377' => '7722'             , # G+7722  U+8A31  uni8A31,  CJK UNIFIED IDEOGRAPH-8A31
        '35380' => '5530'             , # G+5530  U+8A34  uni8A34,  CJK UNIFIED IDEOGRAPH-8A34
        '35382' => '3459'             , # G+3459  U+8A36  uni8A36,  CJK UNIFIED IDEOGRAPH-8A36
        '35386' => '7006'             , # G+7006  U+8A3A  uni8A3A,  CJK UNIFIED IDEOGRAPH-8A3A
        '35387' => '6890'             , # G+6890  U+8A3B  uni8A3B,  CJK UNIFIED IDEOGRAPH-8A3B
        '35408' => '5291'             , # G+5291  U+8A50  uni8A50,  CJK UNIFIED IDEOGRAPH-8A50
        '35412' => '6818'             , # G+6818  U+8A54  uni8A54,  CJK UNIFIED IDEOGRAPH-8A54
        '35413' => '7533'             , # G+7533  U+8A55  uni8A55,  CJK UNIFIED IDEOGRAPH-8A55
        '35419' => '6612'             , # G+6612  U+8A5B  uni8A5B,  CJK UNIFIED IDEOGRAPH-8A5B
        '35422' => '5292'             , # G+5292  U+8A5E  uni8A5E,  CJK UNIFIED IDEOGRAPH-8A5E
        '35424' => '6025'             , # G+6025  U+8A60  uni8A60,  CJK UNIFIED IDEOGRAPH-8A60
        '35426' => '5661'             , # G+5661  U+8A62  uni8A62,  CJK UNIFIED IDEOGRAPH-8A62
        '35427' => '6043'             , # G+6043  U+8A63  uni8A63,  CJK UNIFIED IDEOGRAPH-8A63
        '35430' => '5715'             , # G+5715  U+8A66  uni8A66,  CJK UNIFIED IDEOGRAPH-8A66
        '35433' => '5716'             , # G+5716  U+8A69  uni8A69,  CJK UNIFIED IDEOGRAPH-8A69
        '35437' => '3987'             , # G+3987  U+8A6D  uni8A6D,  CJK UNIFIED IDEOGRAPH-8A6D
        '35438' => '6673'             , # G+6673  U+8A6E  uni8A6E,  CJK UNIFIED IDEOGRAPH-8A6E
        '35440' => '8055'             , # G+8055  U+8A70  uni8A70,  CJK UNIFIED IDEOGRAPH-8A70
        '35441' => '7873'             , # G+7873  U+8A71  uni8A71,  CJK UNIFIED IDEOGRAPH-8A71
        '35442' => '7698'             , # G+7698  U+8A72  uni8A72,  CJK UNIFIED IDEOGRAPH-8A72
        '35443' => '5358'             , # G+5358  U+8A73  uni8A73,  CJK UNIFIED IDEOGRAPH-8A73
        '35445' => '5443'             , # G+5443  U+8A75  uni8A75,  CJK UNIFIED IDEOGRAPH-8A75
        '35449' => '7185'             , # G+7185  U+8A79  uni8A79,  CJK UNIFIED IDEOGRAPH-8A79
        '35461' => '6891'             , # G+6891  U+8A85  uni8A85,  CJK UNIFIED IDEOGRAPH-8A85
        '35463' => '3809'             , # G+3809  U+8A87  uni8A87,  CJK UNIFIED IDEOGRAPH-8A87
        '35468' => '6971'             , # G+6971  U+8A8C  uni8A8C,  CJK UNIFIED IDEOGRAPH-8A8C
        '35469' => '6457'             , # G+6457  U+8A8D  uni8A8D,  CJK UNIFIED IDEOGRAPH-8A8D
        '35475' => '5398'             , # G+5398  U+8A93  uni8A93,  CJK UNIFIED IDEOGRAPH-8A93
        '35477' => '7417'             , # G+7417  U+8A95  uni8A95,  CJK UNIFIED IDEOGRAPH-8A95
        '35480' => '6344'             , # G+6344  U+8A98  uni8A98,  CJK UNIFIED IDEOGRAPH-8A98
        '35486' => '5908'             , # G+5908  U+8A9E  uni8A9E,  CJK UNIFIED IDEOGRAPH-8A9E
        '35488' => '5490'             , # G+5490  U+8AA0  uni8AA0,  CJK UNIFIED IDEOGRAPH-8AA0
        '35489' => '3725'             , # G+3725  U+8AA1  uni8AA1,  CJK UNIFIED IDEOGRAPH-8AA1
        '35491' => '4830'             , # G+4830  U+8AA3  uni8AA3,  CJK UNIFIED IDEOGRAPH-8AA3
        '35492' => '6075'             , # G+6075  U+8AA4  uni8AA4,  CJK UNIFIED IDEOGRAPH-8AA4
        '35493' => '3760'             , # G+3760  U+8AA5  uni8AA5,  CJK UNIFIED IDEOGRAPH-8AA5
        '35494' => '5558'             , # G+5558  U+8AA6  uni8AA6,  CJK UNIFIED IDEOGRAPH-8AA6
        '35496' => '7946'             , # G+7946  U+8AA8  uni8AA8,  CJK UNIFIED IDEOGRAPH-8AA8
        '35498' => '5460'             , # G+5460  U+8AAA  uni8AAA,  CJK UNIFIED IDEOGRAPH-8AAA
        '35504' => '5611'             , # G+5611  U+8AB0  uni8AB0,  CJK UNIFIED IDEOGRAPH-8AB0
        '35506' => '3810'             , # G+3810  U+8AB2  uni8AB2,  CJK UNIFIED IDEOGRAPH-8AB2
        '35513' => '5216'             , # G+5216  U+8AB9  uni8AB9,  CJK UNIFIED IDEOGRAPH-8AB9
        '35516' => '6410'             , # G+6410  U+8ABC  uni8ABC,  CJK UNIFIED IDEOGRAPH-8ABC
        '35518' => '6377'             , # G+6377  U+8ABE  uni8ABE,  CJK UNIFIED IDEOGRAPH-8ABE
        '35519' => '6819'             , # G+6819  U+8ABF  uni8ABF,  CJK UNIFIED IDEOGRAPH-8ABF
        '35522' => '7186'             , # G+7186  U+8AC2  uni8AC2,  CJK UNIFIED IDEOGRAPH-8AC2
        '35524' => '5662'             , # G+5662  U+8AC4  uni8AC4,  CJK UNIFIED IDEOGRAPH-8AC4
        '35527' => '4231'             , # G+4231  U+8AC7  uni8AC7,  CJK UNIFIED IDEOGRAPH-8AC7
        '35531' => '7202'             , # G+7202  U+8ACB  uni8ACB,  CJK UNIFIED IDEOGRAPH-8ACB
        '35533' => '6589'             , # G+6589  U+8ACD  uni8ACD,  CJK UNIFIED IDEOGRAPH-8ACD
        '35535' => '7276'             , # G+7276  U+8ACF  uni8ACF,  CJK UNIFIED IDEOGRAPH-8ACF
        '35538' => '4437'             , # G+4437  U+8AD2  uni8AD2,  CJK UNIFIED IDEOGRAPH-8AD2
        '35542' => '4535'             , # G+4535  U+8AD6  uni8AD6,  CJK UNIFIED IDEOGRAPH-8AD6
        '35547' => '6345'             , # G+6345  U+8ADB  uni8ADB,  CJK UNIFIED IDEOGRAPH-8ADB
        '35548' => '7194'             , # G+7194  U+8ADC  uni8ADC,  CJK UNIFIED IDEOGRAPH-8ADC
        '35553' => '5717'             , # G+5717  U+8AE1  uni8AE1,  CJK UNIFIED IDEOGRAPH-8AE1
        '35558' => '7210'             , # G+7210  U+8AE6  uni8AE6,  CJK UNIFIED IDEOGRAPH-8AE6
        '35559' => '7699'             , # G+7699  U+8AE7  uni8AE7,  CJK UNIFIED IDEOGRAPH-8AE7
        '35562' => '6746'             , # G+6746  U+8AEA  uni8AEA,  CJK UNIFIED IDEOGRAPH-8AEA
        '35563' => '3498'             , # G+3498  U+8AEB  uni8AEB,  CJK UNIFIED IDEOGRAPH-8AEB
        '35565' => '6346'             , # G+6346  U+8AED  uni8AED,  CJK UNIFIED IDEOGRAPH-8AED
        '35566' => '6506'             , # G+6506  U+8AEE  uni8AEE,  CJK UNIFIED IDEOGRAPH-8AEE
        '35569' => '8002'             , # G+8002  U+8AF1  uni8AF1,  CJK UNIFIED IDEOGRAPH-8AF1
        '35574' => '5770'             , # G+5770  U+8AF6  uni8AF6,  CJK UNIFIED IDEOGRAPH-8AF6
        '35575' => '7593'             , # G+7593  U+8AF7  uni8AF7,  CJK UNIFIED IDEOGRAPH-8AF7
        '35576' => '6774'             , # G+6774  U+8AF8  uni8AF8,  CJK UNIFIED IDEOGRAPH-8AF8
        '35578' => '5922'             , # G+5922  U+8AFA  uni8AFA,  CJK UNIFIED IDEOGRAPH-8AFA
        '35582' => '4139'             , # G+4139  U+8AFE  uni8AFE,  CJK UNIFIED IDEOGRAPH-8AFE
        '35584' => '4785'             , # G+4785  U+8B00  uni8B00,  CJK UNIFIED IDEOGRAPH-8B00
        '35585' => '5818'             , # G+5818  U+8B01  uni8B01,  CJK UNIFIED IDEOGRAPH-8B01
        '35586' => '6305'             , # G+6305  U+8B02  uni8B02,  CJK UNIFIED IDEOGRAPH-8B02
        '35588' => '4371'             , # G+4371  U+8B04  uni8B04,  CJK UNIFIED IDEOGRAPH-8B04
        '35598' => '4866'             , # G+4866  U+8B0E  uni8B0E,  CJK UNIFIED IDEOGRAPH-8B0E
        '35600' => '4885'             , # G+4885  U+8B10  uni8B10,  CJK UNIFIED IDEOGRAPH-8B10
        '35604' => '7633'             , # G+7633  U+8B14  uni8B14,  CJK UNIFIED IDEOGRAPH-8B14
        '35606' => '5543'             , # G+5543  U+8B16  uni8B16,  CJK UNIFIED IDEOGRAPH-8B16
        '35607' => '4965'             , # G+4965  U+8B17  uni8B17,  CJK UNIFIED IDEOGRAPH-8B17
        '35609' => '3657'             , # G+3657  U+8B19  uni8B19,  CJK UNIFIED IDEOGRAPH-8B19
        '35610' => '6442'             , # G+6442  U+8B1A  uni8B1A,  CJK UNIFIED IDEOGRAPH-8B1A
        '35611' => '3556'             , # G+3556  U+8B1B  uni8B1B,  CJK UNIFIED IDEOGRAPH-8B1B
        '35613' => '5293'             , # G+5293  U+8B1D  uni8B1D,  CJK UNIFIED IDEOGRAPH-8B1D
        '35616' => '6164'             , # G+6164  U+8B20  uni8B20,  CJK UNIFIED IDEOGRAPH-8B20
        '35624' => '4786'             , # G+4786  U+8B28  uni8B28,  CJK UNIFIED IDEOGRAPH-8B28
        '35627' => '6634'             , # G+6634  U+8B2B  uni8B2B,  CJK UNIFIED IDEOGRAPH-8B2B
        '35628' => '4589'             , # G+4589  U+8B2C  uni8B2C,  CJK UNIFIED IDEOGRAPH-8B2C
        '35635' => '3935'             , # G+3935  U+8B33  uni8B33,  CJK UNIFIED IDEOGRAPH-8B33
        '35641' => '4035'             , # G+4035  U+8B39  uni8B39,  CJK UNIFIED IDEOGRAPH-8B39
        '35649' => '7874'             , # G+7874  U+8B41  uni8B41,  CJK UNIFIED IDEOGRAPH-8B41
        '35657' => '6940'             , # G+6940  U+8B49  uni8B49,  CJK UNIFIED IDEOGRAPH-8B49
        '35662' => '8011'             , # G+8011  U+8B4E  uni8B4E,  CJK UNIFIED IDEOGRAPH-8B4E
        '35663' => '4115'             , # G+4115  U+8B4F  uni8B4F,  CJK UNIFIED IDEOGRAPH-8B4F
        '35672' => '5731'             , # G+5731  U+8B58  uni8B58,  CJK UNIFIED IDEOGRAPH-8B58
        '35674' => '4232'             , # G+4232  U+8B5A  uni8B5A,  CJK UNIFIED IDEOGRAPH-8B5A
        '35676' => '5072'             , # G+5072  U+8B5C  uni8B5C,  CJK UNIFIED IDEOGRAPH-8B5C
        '35686' => '3697'             , # G+3697  U+8B66  uni8B66,  CJK UNIFIED IDEOGRAPH-8B66
        '35692' => '5217'             , # G+5217  U+8B6C  uni8B6C,  CJK UNIFIED IDEOGRAPH-8B6C
        '35695' => '5953'             , # G+5953  U+8B6F  uni8B6F,  CJK UNIFIED IDEOGRAPH-8B6F
        '35696' => '6411'             , # G+6411  U+8B70  uni8B70,  CJK UNIFIED IDEOGRAPH-8B70
        '35700' => '3645'             , # G+3645  U+8B74  uni8B74,  CJK UNIFIED IDEOGRAPH-8B74
        '35703' => '7836'             , # G+7836  U+8B77  uni8B77,  CJK UNIFIED IDEOGRAPH-8B77
        '35709' => '6044'             , # G+6044  U+8B7D  uni8B7D,  CJK UNIFIED IDEOGRAPH-8B7D
        '35712' => '4318'             , # G+4318  U+8B80  uni8B80,  CJK UNIFIED IDEOGRAPH-8B80
        '35722' => '5033'             , # G+5033  U+8B8A  uni8B8A,  CJK UNIFIED IDEOGRAPH-8B8A
        '35728' => '5612'             , # G+5612  U+8B90  uni8B90,  CJK UNIFIED IDEOGRAPH-8B90
        '35730' => '7090'             , # G+7090  U+8B92  uni8B92,  CJK UNIFIED IDEOGRAPH-8B92
        '35731' => '5898'             , # G+5898  U+8B93  uni8B93,  CJK UNIFIED IDEOGRAPH-8B93
        '35734' => '7091'             , # G+7091  U+8B96  uni8B96,  CJK UNIFIED IDEOGRAPH-8B96
        '35738' => '7072'             , # G+7072  U+8B9A  uni8B9A,  CJK UNIFIED IDEOGRAPH-8B9A
        '35895' => '3772'             , # G+3772  U+8C37  uni8C37,  CJK UNIFIED IDEOGRAPH-8C37
        '35903' => '3726'             , # G+3726  U+8C3F  uni8C3F,  CJK UNIFIED IDEOGRAPH-8C3F
        '35905' => '7902'             , # G+7902  U+8C41  uni8C41,  CJK UNIFIED IDEOGRAPH-8C41
        '35910' => '4355'             , # G+4355  U+8C46  uni8C46,  CJK UNIFIED IDEOGRAPH-8C46
        '35912' => '4116'             , # G+4116  U+8C48  uni8C48,  CJK UNIFIED IDEOGRAPH-8C48
        '35914' => '7594'             , # G+7594  U+8C4A  uni8C4A,  CJK UNIFIED IDEOGRAPH-8C4A
        '35916' => '6122'             , # G+6122  U+8C4C  uni8C4C,  CJK UNIFIED IDEOGRAPH-8C4C
        '35925' => '5718'             , # G+5718  U+8C55  uni8C55,  CJK UNIFIED IDEOGRAPH-8C55
        '35930' => '4327'             , # G+4327  U+8C5A  uni8C5A,  CJK UNIFIED IDEOGRAPH-8C5A
        '35937' => '5359'             , # G+5359  U+8C61  uni8C61,  CJK UNIFIED IDEOGRAPH-8C61
        '35946' => '7837'             , # G+7837  U+8C6A  uni8C6A,  CJK UNIFIED IDEOGRAPH-8C6A
        '35947' => '6045'             , # G+6045  U+8C6B  uni8C6B,  CJK UNIFIED IDEOGRAPH-8C6B
        '35961' => '7586'             , # G+7586  U+8C79  uni8C79,  CJK UNIFIED IDEOGRAPH-8C79
        '35962' => '5719'             , # G+5719  U+8C7A  uni8C7A,  CJK UNIFIED IDEOGRAPH-8C7A
        '35970' => '7236'             , # G+7236  U+8C82  uni8C82,  CJK UNIFIED IDEOGRAPH-8C82
        '35978' => '4723'             , # G+4723  U+8C8A  uni8C8A,  CJK UNIFIED IDEOGRAPH-8C8A
        '35980' => '4787'             , # G+4787  U+8C8C  uni8C8C,  CJK UNIFIED IDEOGRAPH-8C8C
        '35997' => '7512'             , # G+7512  U+8C9D  uni8C9D,  CJK UNIFIED IDEOGRAPH-8C9D
        '35998' => '6747'             , # G+6747  U+8C9E  uni8C9E,  CJK UNIFIED IDEOGRAPH-8C9E
        '36000' => '5140'             , # G+5140  U+8CA0  uni8CA0,  CJK UNIFIED IDEOGRAPH-8CA0
        '36001' => '6583'             , # G+6583  U+8CA1  uni8CA1,  CJK UNIFIED IDEOGRAPH-8CA1
        '36002' => '3800'             , # G+3800  U+8CA2  uni8CA2,  CJK UNIFIED IDEOGRAPH-8CA2
        '36007' => '5234'             , # G+5234  U+8CA7  uni8CA7,  CJK UNIFIED IDEOGRAPH-8CA7
        '36008' => '7875'             , # G+7875  U+8CA8  uni8CA8,  CJK UNIFIED IDEOGRAPH-8CA8
        '36009' => '7495'             , # G+7495  U+8CA9  uni8CA9,  CJK UNIFIED IDEOGRAPH-8CA9
        '36010' => '7423'             , # G+7423  U+8CAA  uni8CAA,  CJK UNIFIED IDEOGRAPH-8CAA
        '36011' => '3832'             , # G+3832  U+8CAB  uni8CAB,  CJK UNIFIED IDEOGRAPH-8CAB
        '36012' => '7129'             , # G+7129  U+8CAC  uni8CAC,  CJK UNIFIED IDEOGRAPH-8CAC
        '36015' => '6613'             , # G+6613  U+8CAF  uni8CAF,  CJK UNIFIED IDEOGRAPH-8CAF
        '36016' => '5499'             , # G+5499  U+8CB0  uni8CB0,  CJK UNIFIED IDEOGRAPH-8CB0
        '36019' => '6433'             , # G+6433  U+8CB3  uni8CB3,  CJK UNIFIED IDEOGRAPH-8CB3
        '36020' => '3992'             , # G+3992  U+8CB4  uni8CB4,  CJK UNIFIED IDEOGRAPH-8CB4
        '36022' => '7528'             , # G+7528  U+8CB6  uni8CB6,  CJK UNIFIED IDEOGRAPH-8CB6
        '36023' => '4718'             , # G+4718  U+8CB7  uni8CB7,  CJK UNIFIED IDEOGRAPH-8CB7
        '36024' => '4263'             , # G+4263  U+8CB8  uni8CB8,  CJK UNIFIED IDEOGRAPH-8CB8
        '36027' => '5218'             , # G+5218  U+8CBB  uni8CBB,  CJK UNIFIED IDEOGRAPH-8CBB
        '36028' => '7195'             , # G+7195  U+8CBC  uni8CBC,  CJK UNIFIED IDEOGRAPH-8CBC
        '36029' => '6432'             , # G+6432  U+8CBD  uni8CBD,  CJK UNIFIED IDEOGRAPH-8CBD
        '36031' => '4831'             , # G+4831  U+8CBF  uni8CBF,  CJK UNIFIED IDEOGRAPH-8CBF
        '36032' => '7626'             , # G+7626  U+8CC0  uni8CC0,  CJK UNIFIED IDEOGRAPH-8CC0
        '36033' => '5169'             , # G+5169  U+8CC1  uni8CC1,  CJK UNIFIED IDEOGRAPH-8CC1
        '36034' => '4547'             , # G+4547  U+8CC2  uni8CC2,  CJK UNIFIED IDEOGRAPH-8CC2
        '36035' => '6476'             , # G+6476  U+8CC3  uni8CC3,  CJK UNIFIED IDEOGRAPH-8CC3
        '36036' => '7947'             , # G+7947  U+8CC4  uni8CC4,  CJK UNIFIED IDEOGRAPH-8CC4
        '36039' => '6507'             , # G+6507  U+8CC7  uni8CC7,  CJK UNIFIED IDEOGRAPH-8CC7
        '36040' => '3460'             , # G+3460  U+8CC8  uni8CC8,  CJK UNIFIED IDEOGRAPH-8CC8
        '36042' => '6635'             , # G+6635  U+8CCA  uni8CCA,  CJK UNIFIED IDEOGRAPH-8CCA
        '36049' => '7007'             , # G+7007  U+8CD1  uni8CD1,  CJK UNIFIED IDEOGRAPH-8CD1
        '36051' => '5235'             , # G+5235  U+8CD3  uni8CD3,  CJK UNIFIED IDEOGRAPH-8CD3
        '36058' => '4548'             , # G+4548  U+8CDA  uni8CDA,  CJK UNIFIED IDEOGRAPH-8CDA
        '36060' => '5294'             , # G+5294  U+8CDC  uni8CDC,  CJK UNIFIED IDEOGRAPH-8CDC
        '36062' => '5360'             , # G+5360  U+8CDE  uni8CDE,  CJK UNIFIED IDEOGRAPH-8CDE
        '36064' => '4984'             , # G+4984  U+8CE0  uni8CE0,  CJK UNIFIED IDEOGRAPH-8CE0
        '36066' => '7751'             , # G+7751  U+8CE2  uni8CE2,  CJK UNIFIED IDEOGRAPH-8CE2
        '36067' => '4719'             , # G+4719  U+8CE3  uni8CE3,  CJK UNIFIED IDEOGRAPH-8CE3
        '36068' => '7160'             , # G+7160  U+8CE4  uni8CE4,  CJK UNIFIED IDEOGRAPH-8CE4
        '36070' => '5141'             , # G+5141  U+8CE6  uni8CE6,  CJK UNIFIED IDEOGRAPH-8CE6
        '36074' => '7027'             , # G+7027  U+8CEA  uni8CEA,  CJK UNIFIED IDEOGRAPH-8CEA
        '36077' => '4299'             , # G+4299  U+8CED  uni8CED,  CJK UNIFIED IDEOGRAPH-8CED
        '36084' => '4549'             , # G+4549  U+8CF4  uni8CF4,  CJK UNIFIED IDEOGRAPH-8CF4
        '36091' => '5142'             , # G+5142  U+8CFB  uni8CFB,  CJK UNIFIED IDEOGRAPH-8CFB
        '36092' => '3936'             , # G+3936  U+8CFC  uni8CFC,  CJK UNIFIED IDEOGRAPH-8CFC
        '36093' => '5364'             , # G+5364  U+8CFD  uni8CFD,  CJK UNIFIED IDEOGRAPH-8CFD
        '36100' => '6972'             , # G+6972  U+8D04  uni8D04,  CJK UNIFIED IDEOGRAPH-8D04
        '36101' => '7315'             , # G+7315  U+8D05  uni8D05,  CJK UNIFIED IDEOGRAPH-8D05
        '36103' => '6365'             , # G+6365  U+8D07  uni8D07,  CJK UNIFIED IDEOGRAPH-8D07
        '36104' => '6941'             , # G+6941  U+8D08  uni8D08,  CJK UNIFIED IDEOGRAPH-8D08
        '36106' => '7073'             , # G+7073  U+8D0A  uni8D0A,  CJK UNIFIED IDEOGRAPH-8D0A
        '36109' => '5468'             , # G+5468  U+8D0D  uni8D0D,  CJK UNIFIED IDEOGRAPH-8D0D
        '36115' => '6566'             , # G+6566  U+8D13  uni8D13,  CJK UNIFIED IDEOGRAPH-8D13
        '36118' => '5544'             , # G+5544  U+8D16  uni8D16,  CJK UNIFIED IDEOGRAPH-8D16
        '36196' => '6636'             , # G+6636  U+8D64  uni8D64,  CJK UNIFIED IDEOGRAPH-8D64
        '36198' => '5295'             , # G+5295  U+8D66  uni8D66,  CJK UNIFIED IDEOGRAPH-8D66
        '36203' => '7732'             , # G+7732  U+8D6B  uni8D6B,  CJK UNIFIED IDEOGRAPH-8D6B
        '36208' => '6892'             , # G+6892  U+8D70  uni8D70,  CJK UNIFIED IDEOGRAPH-8D70
        '36211' => '4006'             , # G+4006  U+8D73  uni8D73,  CJK UNIFIED IDEOGRAPH-8D73
        '36212' => '5143'             , # G+5143  U+8D74  uni8D74,  CJK UNIFIED IDEOGRAPH-8D74
        '36215' => '4117'             , # G+4117  U+8D77  uni8D77,  CJK UNIFIED IDEOGRAPH-8D77
        '36229' => '7237'             , # G+7237  U+8D85  uni8D85,  CJK UNIFIED IDEOGRAPH-8D85
        '36234' => '6282'             , # G+6282  U+8D8A  uni8D8A,  CJK UNIFIED IDEOGRAPH-8D8A
        '36249' => '6820'             , # G+6820  U+8D99  uni8D99,  CJK UNIFIED IDEOGRAPH-8D99
        '36259' => '7326'             , # G+7326  U+8DA3  uni8DA3,  CJK UNIFIED IDEOGRAPH-8DA3
        '36264' => '7277'             , # G+7277  U+8DA8  uni8DA8,  CJK UNIFIED IDEOGRAPH-8DA8
        '36275' => '6830'             , # G+6830  U+8DB3  uni8DB3,  CJK UNIFIED IDEOGRAPH-8DB3
        '36282' => '5144'             , # G+5144  U+8DBA  uni8DBA,  CJK UNIFIED IDEOGRAPH-8DBA
        '36286' => '6973'             , # G+6973  U+8DBE  uni8DBE,  CJK UNIFIED IDEOGRAPH-8DBE
        '36294' => '7442'             , # G+7442  U+8DC6  uni8DC6,  CJK UNIFIED IDEOGRAPH-8DC6
        '36299' => '4936'             , # G+4936  U+8DCB  uni8DCB,  CJK UNIFIED IDEOGRAPH-8DCB
        '36300' => '7028'             , # G+7028  U+8DCC  uni8DCC,  CJK UNIFIED IDEOGRAPH-8DCC
        '36303' => '3461'             , # G+3461  U+8DCF  uni8DCF,  CJK UNIFIED IDEOGRAPH-8DCF
        '36315' => '7488'             , # G+7488  U+8DDB  uni8DDB,  CJK UNIFIED IDEOGRAPH-8DDB
        '36317' => '3596'             , # G+3596  U+8DDD  uni8DDD,  CJK UNIFIED IDEOGRAPH-8DDD
        '36321' => '6637'             , # G+6637  U+8DE1  uni8DE1,  CJK UNIFIED IDEOGRAPH-8DE1
        '36323' => '5444'             , # G+5444  U+8DE3  uni8DE3,  CJK UNIFIED IDEOGRAPH-8DE3
        '36328' => '3811'             , # G+3811  U+8DE8  uni8DE8,  CJK UNIFIED IDEOGRAPH-8DE8
        '36335' => '4522'             , # G+4522  U+8DEF  uni8DEF,  CJK UNIFIED IDEOGRAPH-8DEF
        '36339' => '4300'             , # G+4300  U+8DF3  uni8DF3,  CJK UNIFIED IDEOGRAPH-8DF3
        '36362' => '6194'             , # G+6194  U+8E0A  uni8E0A,  CJK UNIFIED IDEOGRAPH-8E0A
        '36367' => '4237'             , # G+4237  U+8E0F  uni8E0F,  CJK UNIFIED IDEOGRAPH-8E0F
        '36368' => '7161'             , # G+7161  U+8E10  uni8E10,  CJK UNIFIED IDEOGRAPH-8E10
        '36382' => '3597'             , # G+3597  U+8E1E  uni8E1E,  CJK UNIFIED IDEOGRAPH-8E1E
        '36394' => '6850'             , # G+6850  U+8E2A  uni8E2A,  CJK UNIFIED IDEOGRAPH-8E2A
        '36400' => '6347'             , # G+6347  U+8E30  uni8E30,  CJK UNIFIED IDEOGRAPH-8E30
        '36405' => '6851'             , # G+6851  U+8E35  uni8E35,  CJK UNIFIED IDEOGRAPH-8E35
        '36418' => '6348'             , # G+6348  U+8E42  uni8E42,  CJK UNIFIED IDEOGRAPH-8E42
        '36420' => '6775'             , # G+6775  U+8E44  uni8E44,  CJK UNIFIED IDEOGRAPH-8E44
        '36423' => '3610'             , # G+3610  U+8E47  uni8E47,  CJK UNIFIED IDEOGRAPH-8E47
        '36424' => '4301'             , # G+4301  U+8E48  uni8E48,  CJK UNIFIED IDEOGRAPH-8E48
        '36425' => '7052'             , # G+7052  U+8E49  uni8E49,  CJK UNIFIED IDEOGRAPH-8E49
        '36426' => '7797'             , # G+7797  U+8E4A  uni8E4A,  CJK UNIFIED IDEOGRAPH-8E4A
        '36441' => '7296'             , # G+7296  U+8E59  uni8E59,  CJK UNIFIED IDEOGRAPH-8E59
        '36447' => '6638'             , # G+6638  U+8E5F  uni8E5F,  CJK UNIFIED IDEOGRAPH-8E5F
        '36448' => '7145'             , # G+7145  U+8E60  uni8E60,  CJK UNIFIED IDEOGRAPH-8E60
        '36468' => '7297'             , # G+7297  U+8E74  uni8E74,  CJK UNIFIED IDEOGRAPH-8E74
        '36470' => '3982'             , # G+3982  U+8E76  uni8E76,  CJK UNIFIED IDEOGRAPH-8E76
        '36481' => '6821'             , # G+6821  U+8E81  uni8E81,  CJK UNIFIED IDEOGRAPH-8E81
        '36487' => '6614'             , # G+6614  U+8E87  uni8E87,  CJK UNIFIED IDEOGRAPH-8E87
        '36490' => '6893'             , # G+6893  U+8E8A  uni8E8A,  CJK UNIFIED IDEOGRAPH-8E8A
        '36493' => '5878'             , # G+5878  U+8E8D  uni8E8D,  CJK UNIFIED IDEOGRAPH-8E8D
        '36522' => '4643'             , # G+4643  U+8EAA  uni8EAA,  CJK UNIFIED IDEOGRAPH-8EAA
        '36523' => '5755'             , # G+5755  U+8EAB  uni8EAB,  CJK UNIFIED IDEOGRAPH-8EAB
        '36524' => '3968'             , # G+3968  U+8EAC  uni8EAC,  CJK UNIFIED IDEOGRAPH-8EAC
        '36544' => '3937'             , # G+3937  U+8EC0  uni8EC0,  CJK UNIFIED IDEOGRAPH-8EC0
        '36554' => '7053'             , # G+7053  U+8ECA  uni8ECA,  CJK UNIFIED IDEOGRAPH-8ECA
        '36555' => '5819'             , # G+5819  U+8ECB  uni8ECB,  CJK UNIFIED IDEOGRAPH-8ECB
        '36556' => '3988'             , # G+3988  U+8ECC  uni8ECC,  CJK UNIFIED IDEOGRAPH-8ECC
        '36557' => '3957'             , # G+3957  U+8ECD  uni8ECD,  CJK UNIFIED IDEOGRAPH-8ECD
        '36562' => '7726'             , # G+7726  U+8ED2  uni8ED2,  CJK UNIFIED IDEOGRAPH-8ED2
        '36575' => '5982'             , # G+5982  U+8EDF  uni8EDF,  CJK UNIFIED IDEOGRAPH-8EDF
        '36587' => '7008'             , # G+7008  U+8EEB  uni8EEB,  CJK UNIFIED IDEOGRAPH-8EEB
        '36600' => '7298'             , # G+7298  U+8EF8  uni8EF8,  CJK UNIFIED IDEOGRAPH-8EF8
        '36603' => '3462'             , # G+3462  U+8EFB  uni8EFB,  CJK UNIFIED IDEOGRAPH-8EFB
        '36606' => '5732'             , # G+5732  U+8EFE  uni8EFE,  CJK UNIFIED IDEOGRAPH-8EFE
        '36611' => '3887'             , # G+3887  U+8F03  uni8F03,  CJK UNIFIED IDEOGRAPH-8F03
        '36613' => '4523'             , # G+4523  U+8F05  uni8F05,  CJK UNIFIED IDEOGRAPH-8F05
        '36617' => '6584'             , # G+6584  U+8F09  uni8F09,  CJK UNIFIED IDEOGRAPH-8F09
        '36626' => '7196'             , # G+7196  U+8F12  uni8F12,  CJK UNIFIED IDEOGRAPH-8F12
        '36627' => '4686'             , # G+4686  U+8F13  uni8F13,  CJK UNIFIED IDEOGRAPH-8F13
        '36628' => '5073'             , # G+5073  U+8F14  uni8F14,  CJK UNIFIED IDEOGRAPH-8F14
        '36629' => '3698'             , # G+3698  U+8F15  uni8F15,  CJK UNIFIED IDEOGRAPH-8F15
        '36635' => '4438'             , # G+4438  U+8F1B  uni8F1B,  CJK UNIFIED IDEOGRAPH-8F1B
        '36636' => '7356'             , # G+7356  U+8F1C  uni8F1C,  CJK UNIFIED IDEOGRAPH-8F1C
        '36637' => '8003'             , # G+8003  U+8F1D  uni8F1D,  CJK UNIFIED IDEOGRAPH-8F1D
        '36638' => '4706'             , # G+4706  U+8F1E  uni8F1E,  CJK UNIFIED IDEOGRAPH-8F1E
        '36639' => '7174'             , # G+7174  U+8F1F  uni8F1F,  CJK UNIFIED IDEOGRAPH-8F1F
        '36646' => '4474'             , # G+4474  U+8F26  uni8F26,  CJK UNIFIED IDEOGRAPH-8F26
        '36647' => '5055'             , # G+5055  U+8F27  uni8F27,  CJK UNIFIED IDEOGRAPH-8F27
        '36649' => '4985'             , # G+4985  U+8F29  uni8F29,  CJK UNIFIED IDEOGRAPH-8F29
        '36650' => '4599'             , # G+4599  U+8F2A  uni8F2A,  CJK UNIFIED IDEOGRAPH-8F2A
        '36655' => '7035'             , # G+7035  U+8F2F  uni8F2F,  CJK UNIFIED IDEOGRAPH-8F2F
        '36659' => '6894'             , # G+6894  U+8F33  uni8F33,  CJK UNIFIED IDEOGRAPH-8F33
        '36664' => '5613'             , # G+5613  U+8F38  uni8F38,  CJK UNIFIED IDEOGRAPH-8F38
        '36665' => '5087'             , # G+5087  U+8F39  uni8F39,  CJK UNIFIED IDEOGRAPH-8F39
        '36667' => '5088'             , # G+5088  U+8F3B  uni8F3B,  CJK UNIFIED IDEOGRAPH-8F3B
        '36670' => '6674'             , # G+6674  U+8F3E  uni8F3E,  CJK UNIFIED IDEOGRAPH-8F3E
        '36671' => '5944'             , # G+5944  U+8F3F  uni8F3F,  CJK UNIFIED IDEOGRAPH-8F3F
        '36676' => '7650'             , # G+7650  U+8F44  uni8F44,  CJK UNIFIED IDEOGRAPH-8F44
        '36677' => '6276'             , # G+6276  U+8F45  uni8F45,  CJK UNIFIED IDEOGRAPH-8F45
        '36681' => '6675'             , # G+6675  U+8F49  uni8F49,  CJK UNIFIED IDEOGRAPH-8F49
        '36685' => '7175'             , # G+7175  U+8F4D  uni8F4D,  CJK UNIFIED IDEOGRAPH-8F4D
        '36686' => '3888'             , # G+3888  U+8F4E  uni8F4E,  CJK UNIFIED IDEOGRAPH-8F4E
        '36701' => '5945'             , # G+5945  U+8F5D  uni8F5D,  CJK UNIFIED IDEOGRAPH-8F5D
        '36703' => '3867'             , # G+3867  U+8F5F  uni8F5F,  CJK UNIFIED IDEOGRAPH-8F5F
        '36706' => '4463'             , # G+4463  U+8F62  uni8F62,  CJK UNIFIED IDEOGRAPH-8F62
        '36763' => '5756'             , # G+5756  U+8F9B  uni8F9B,  CJK UNIFIED IDEOGRAPH-8F9B
        '36764' => '3761'             , # G+3761  U+8F9C  uni8F9C,  CJK UNIFIED IDEOGRAPH-8F9C
        '36771' => '4398'             , # G+4398  U+8FA3  uni8FA3,  CJK UNIFIED IDEOGRAPH-8FA3
        '36774' => '7496'             , # G+7496  U+8FA6  uni8FA6,  CJK UNIFIED IDEOGRAPH-8FA6
        '36776' => '5034'             , # G+5034  U+8FA8  uni8FA8,  CJK UNIFIED IDEOGRAPH-8FA8
        '36781' => '5296'             , # G+5296  U+8FAD  uni8FAD,  CJK UNIFIED IDEOGRAPH-8FAD
        '36783' => '5035'             , # G+5035  U+8FAF  uni8FAF,  CJK UNIFIED IDEOGRAPH-8FAF
        '36784' => '7009'             , # G+7009  U+8FB0  uni8FB0,  CJK UNIFIED IDEOGRAPH-8FB0
        '36785' => '6173'             , # G+6173  U+8FB1  uni8FB1,  CJK UNIFIED IDEOGRAPH-8FB1
        '36786' => '4177'             , # G+4177  U+8FB2  uni8FB2,  CJK UNIFIED IDEOGRAPH-8FB2
        '36802' => '6222'             , # G+6222  U+8FC2  uni8FC2,  CJK UNIFIED IDEOGRAPH-8FC2
        '36805' => '5757'             , # G+5757  U+8FC5  uni8FC5,  CJK UNIFIED IDEOGRAPH-8FC5
        '36814' => '6026'             , # G+6026  U+8FCE  uni8FCE,  CJK UNIFIED IDEOGRAPH-8FCE
        '36817' => '4036'             , # G+4036  U+8FD1  uni8FD1,  CJK UNIFIED IDEOGRAPH-8FD1
        '36820' => '4927'             , # G+4927  U+8FD4  uni8FD4,  CJK UNIFIED IDEOGRAPH-8FD4
        '36838' => '3463'             , # G+3463  U+8FE6  uni8FE6,  CJK UNIFIED IDEOGRAPH-8FE6
        '36842' => '6639'             , # G+6639  U+8FEA  uni8FEA,  CJK UNIFIED IDEOGRAPH-8FEA
        '36843' => '4902'             , # G+4902  U+8FEB  uni8FEB,  CJK UNIFIED IDEOGRAPH-8FEB
        '36845' => '7029'             , # G+7029  U+8FED  uni8FED,  CJK UNIFIED IDEOGRAPH-8FED
        '36848' => '5669'             , # G+5669  U+8FF0  uni8FF0,  CJK UNIFIED IDEOGRAPH-8FF0
        '36850' => '3626'             , # G+3626  U+8FF2  uni8FF2,  CJK UNIFIED IDEOGRAPH-8FF2
        '36855' => '4867'             , # G+4867  U+8FF7  uni8FF7,  CJK UNIFIED IDEOGRAPH-8FF7
        '36857' => '6640'             , # G+6640  U+8FF9  uni8FF9,  CJK UNIFIED IDEOGRAPH-8FF9
        '36861' => '7278'             , # G+7278  U+8FFD  uni8FFD,  CJK UNIFIED IDEOGRAPH-8FFD
        '36864' => '7463'             , # G+7463  U+9000  uni9000,  CJK UNIFIED IDEOGRAPH-9000
        '36865' => '5559'             , # G+5559  U+9001  uni9001,  CJK UNIFIED IDEOGRAPH-9001
        '36866' => '3838'             , # G+3838  U+9002  uni9002,  CJK UNIFIED IDEOGRAPH-9002
        '36867' => '4302'             , # G+4302  U+9003  uni9003,  CJK UNIFIED IDEOGRAPH-9003
        '36869' => '7978'             , # G+7978  U+9005  uni9005,  CJK UNIFIED IDEOGRAPH-9005
        '36870' => '5954'             , # G+5954  U+9006  uni9006,  CJK UNIFIED IDEOGRAPH-9006
        '36872' => '7787'             , # G+7787  U+9008  uni9008,  CJK UNIFIED IDEOGRAPH-9008
        '36875' => '7567'             , # G+7567  U+900B  uni900B,  CJK UNIFIED IDEOGRAPH-900B
        '36877' => '5531'             , # G+5531  U+900D  uni900D,  CJK UNIFIED IDEOGRAPH-900D
        '36879' => '7469'             , # G+7469  U+900F  uni900F,  CJK UNIFIED IDEOGRAPH-900F
        '36880' => '7299'             , # G+7299  U+9010  uni9010,  CJK UNIFIED IDEOGRAPH-9010
        '36881' => '3938'             , # G+3938  U+9011  uni9011,  CJK UNIFIED IDEOGRAPH-9011
        '36884' => '4303'             , # G+4303  U+9014  uni9014,  CJK UNIFIED IDEOGRAPH-9014
        '36885' => '3699'             , # G+3699  U+9015  uni9015,  CJK UNIFIED IDEOGRAPH-9015
        '36887' => '4356'             , # G+4356  U+9017  uni9017,  CJK UNIFIED IDEOGRAPH-9017
        '36889' => '6615'             , # G+6615  U+9019  uni9019,  CJK UNIFIED IDEOGRAPH-9019
        '36890' => '7458'             , # G+7458  U+901A  uni901A,  CJK UNIFIED IDEOGRAPH-901A
        '36893' => '5399'             , # G+5399  U+901D  uni901D,  CJK UNIFIED IDEOGRAPH-901D
        '36894' => '4500'             , # G+4500  U+901E  uni901E,  CJK UNIFIED IDEOGRAPH-901E
        '36895' => '5545'             , # G+5545  U+901F  uni901F,  CJK UNIFIED IDEOGRAPH-901F
        '36896' => '6822'             , # G+6822  U+9020  uni9020,  CJK UNIFIED IDEOGRAPH-9020
        '36897' => '6917'             , # G+6917  U+9021  uni9021,  CJK UNIFIED IDEOGRAPH-9021
        '36898' => '5106'             , # G+5106  U+9022  uni9022,  CJK UNIFIED IDEOGRAPH-9022
        '36899' => '4475'             , # G+4475  U+9023  uni9023,  CJK UNIFIED IDEOGRAPH-9023
        '36910' => '7211'             , # G+7211  U+902E  uni902E,  CJK UNIFIED IDEOGRAPH-902E
        '36913' => '6895'             , # G+6895  U+9031  uni9031,  CJK UNIFIED IDEOGRAPH-9031
        '36914' => '7010'             , # G+7010  U+9032  uni9032,  CJK UNIFIED IDEOGRAPH-9032
        '36917' => '4007'             , # G+4007  U+9035  uni9035,  CJK UNIFIED IDEOGRAPH-9035
        '36920' => '6466'             , # G+6466  U+9038  uni9038,  CJK UNIFIED IDEOGRAPH-9038
        '36924' => '7615'             , # G+7615  U+903C  uni903C,  CJK UNIFIED IDEOGRAPH-903C
        '36926' => '6350'             , # G+6350  U+903E  uni903E,  CJK UNIFIED IDEOGRAPH-903E
        '36929' => '4361'             , # G+4361  U+9041  uni9041,  CJK UNIFIED IDEOGRAPH-9041
        '36930' => '5614'             , # G+5614  U+9042  uni9042,  CJK UNIFIED IDEOGRAPH-9042
        '36935' => '6223'             , # G+6223  U+9047  uni9047,  CJK UNIFIED IDEOGRAPH-9047
        '36938' => '6349'             , # G+6349  U+904A  uni904A,  CJK UNIFIED IDEOGRAPH-904A
        '36939' => '6246'             , # G+6246  U+904B  uni904B,  CJK UNIFIED IDEOGRAPH-904B
        '36941' => '7525'             , # G+7525  U+904D  uni904D,  CJK UNIFIED IDEOGRAPH-904D
        '36942' => '3812'             , # G+3812  U+904E  uni904E,  CJK UNIFIED IDEOGRAPH-904E
        '36944' => '7627'             , # G+7627  U+9050  uni9050,  CJK UNIFIED IDEOGRAPH-9050
        '36945' => '7925'             , # G+7925  U+9051  uni9051,  CJK UNIFIED IDEOGRAPH-9051
        '36947' => '4304'             , # G+4304  U+9053  uni9053,  CJK UNIFIED IDEOGRAPH-9053
        '36948' => '4216'             , # G+4216  U+9054  uni9054,  CJK UNIFIED IDEOGRAPH-9054
        '36949' => '6306'             , # G+6306  U+9055  uni9055,  CJK UNIFIED IDEOGRAPH-9055
        '36953' => '6165'             , # G+6165  U+9059  uni9059,  CJK UNIFIED IDEOGRAPH-9059
        '36956' => '5550'             , # G+5550  U+905C  uni905C,  CJK UNIFIED IDEOGRAPH-905C
        '36957' => '4238'             , # G+4238  U+905D  uni905D,  CJK UNIFIED IDEOGRAPH-905D
        '36958' => '7212'             , # G+7212  U+905E  uni905E,  CJK UNIFIED IDEOGRAPH-905E
        '36960' => '6277'             , # G+6277  U+9060  uni9060,  CJK UNIFIED IDEOGRAPH-9060
        '36961' => '5532'             , # G+5532  U+9061  uni9061,  CJK UNIFIED IDEOGRAPH-9061
        '36963' => '3646'             , # G+3646  U+9063  uni9063,  CJK UNIFIED IDEOGRAPH-9063
        '36969' => '6641'             , # G+6641  U+9069  uni9069,  CJK UNIFIED IDEOGRAPH-9069
        '36973' => '6823'             , # G+6823  U+906D  uni906D,  CJK UNIFIED IDEOGRAPH-906D
        '36974' => '7054'             , # G+7054  U+906E  uni906E,  CJK UNIFIED IDEOGRAPH-906E
        '36975' => '4362'             , # G+4362  U+906F  uni906F,  CJK UNIFIED IDEOGRAPH-906F
        '36978' => '6974'             , # G+6974  U+9072  uni9072,  CJK UNIFIED IDEOGRAPH-9072
        '36981' => '6918'             , # G+6918  U+9075  uni9075,  CJK UNIFIED IDEOGRAPH-9075
        '36983' => '7162'             , # G+7162  U+9077  uni9077,  CJK UNIFIED IDEOGRAPH-9077
        '36984' => '5445'             , # G+5445  U+9078  uni9078,  CJK UNIFIED IDEOGRAPH-9078
        '36986' => '6351'             , # G+6351  U+907A  uni907A,  CJK UNIFIED IDEOGRAPH-907A
        '36988' => '4561'             , # G+4561  U+907C  uni907C,  CJK UNIFIED IDEOGRAPH-907C
        '36989' => '3598'             , # G+3598  U+907D  uni907D,  CJK UNIFIED IDEOGRAPH-907D
        '36991' => '7602'             , # G+7602  U+907F  uni907F,  CJK UNIFIED IDEOGRAPH-907F
        '36992' => '6166'             , # G+6166  U+9080  uni9080,  CJK UNIFIED IDEOGRAPH-9080
        '36993' => '4720'             , # G+4720  U+9081  uni9081,  CJK UNIFIED IDEOGRAPH-9081
        '36994' => '7700'             , # G+7700  U+9082  uni9082,  CJK UNIFIED IDEOGRAPH-9082
        '36995' => '5615'             , # G+5615  U+9083  uni9083,  CJK UNIFIED IDEOGRAPH-9083
        '36996' => '7896'             , # G+7896  U+9084  uni9084,  CJK UNIFIED IDEOGRAPH-9084
        '36999' => '6434'             , # G+6434  U+9087  uni9087,  CJK UNIFIED IDEOGRAPH-9087
        '37000' => '4669'             , # G+4669  U+9088  uni9088,  CJK UNIFIED IDEOGRAPH-9088
        '37002' => '5036'             , # G+5036  U+908A  uni908A,  CJK UNIFIED IDEOGRAPH-908A
        '37007' => '4381'             , # G+4381  U+908F  uni908F,  CJK UNIFIED IDEOGRAPH-908F
        '37009' => '6389'             , # G+6389  U+9091  uni9091,  CJK UNIFIED IDEOGRAPH-9091
        '37013' => '6096'             , # G+6096  U+9095  uni9095,  CJK UNIFIED IDEOGRAPH-9095
        '37017' => '4707'             , # G+4707  U+9099  uni9099,  CJK UNIFIED IDEOGRAPH-9099
        '37026' => '7788'             , # G+7788  U+90A2  uni90A2,  CJK UNIFIED IDEOGRAPH-90A2
        '37027' => '4138'             , # G+4138  U+90A3  uni90A3,  CJK UNIFIED IDEOGRAPH-90A3
        '37030' => '4966'             , # G+4966  U+90A6  uni90A6,  CJK UNIFIED IDEOGRAPH-90A6
        '37032' => '7250'             , # G+7250  U+90A8  uni90A8,  CJK UNIFIED IDEOGRAPH-90A8
        '37034' => '5297'             , # G+5297  U+90AA  uni90AA,  CJK UNIFIED IDEOGRAPH-90AA
        '37039' => '3526'             , # G+3526  U+90AF  uni90AF,  CJK UNIFIED IDEOGRAPH-90AF
        '37040' => '7443'             , # G+7443  U+90B0  uni90B0,  CJK UNIFIED IDEOGRAPH-90B0
        '37041' => '3939'             , # G+3939  U+90B1  uni90B1,  CJK UNIFIED IDEOGRAPH-90B1
        '37045' => '5533'             , # G+5533  U+90B5  uni90B5,  CJK UNIFIED IDEOGRAPH-90B5
        '37048' => '6616'             , # G+6616  U+90B8  uni90B8,  CJK UNIFIED IDEOGRAPH-90B8
        '37057' => '6236'             , # G+6236  U+90C1  uni90C1,  CJK UNIFIED IDEOGRAPH-90C1
        '37066' => '3889'             , # G+3889  U+90CA  uni90CA,  CJK UNIFIED IDEOGRAPH-90CA
        '37086' => '4419'             , # G+4419  U+90DE  uni90DE,  CJK UNIFIED IDEOGRAPH-90DE
        '37089' => '3958'             , # G+3958  U+90E1  uni90E1,  CJK UNIFIED IDEOGRAPH-90E1
        '37096' => '5145'             , # G+5145  U+90E8  uni90E8,  CJK UNIFIED IDEOGRAPH-90E8
        '37101' => '3818'             , # G+3818  U+90ED  uni90ED,  CJK UNIFIED IDEOGRAPH-90ED
        '37109' => '6224'             , # G+6224  U+90F5  uni90F5,  CJK UNIFIED IDEOGRAPH-90F5
        '37117' => '4305'             , # G+4305  U+90FD  uni90FD,  CJK UNIFIED IDEOGRAPH-90FD
        '37122' => '5802'             , # G+5802  U+9102  uni9102,  CJK UNIFIED IDEOGRAPH-9102
        '37138' => '7279'             , # G+7279  U+9112  uni9112,  CJK UNIFIED IDEOGRAPH-9112
        '37141' => '7714'             , # G+7714  U+9115  uni9115,  CJK UNIFIED IDEOGRAPH-9115
        '37145' => '5219'             , # G+5219  U+9119  uni9119,  CJK UNIFIED IDEOGRAPH-9119
        '37159' => '4372'             , # G+4372  U+9127  uni9127,  CJK UNIFIED IDEOGRAPH-9127
        '37165' => '6748'             , # G+6748  U+912D  uni912D,  CJK UNIFIED IDEOGRAPH-912D
        '37170' => '4210'             , # G+4210  U+9132  uni9132,  CJK UNIFIED IDEOGRAPH-9132
        '37193' => '6352'             , # G+6352  U+9149  uni9149,  CJK UNIFIED IDEOGRAPH-9149
        '37194' => '6749'             , # G+6749  U+914A  uni914A,  CJK UNIFIED IDEOGRAPH-914A
        '37195' => '7280'             , # G+7280  U+914B  uni914B,  CJK UNIFIED IDEOGRAPH-914B
        '37196' => '6519'             , # G+6519  U+914C  uni914C,  CJK UNIFIED IDEOGRAPH-914C
        '37197' => '4986'             , # G+4986  U+914D  uni914D,  CJK UNIFIED IDEOGRAPH-914D
        '37198' => '6896'             , # G+6896  U+914E  uni914E,  CJK UNIFIED IDEOGRAPH-914E
        '37202' => '6897'             , # G+6897  U+9152  uni9152,  CJK UNIFIED IDEOGRAPH-9152
        '37218' => '7238'             , # G+7238  U+9162  uni9162,  CJK UNIFIED IDEOGRAPH-9162
        '37225' => '4760'             , # G+4760  U+9169  uni9169,  CJK UNIFIED IDEOGRAPH-9169
        '37226' => '4387'             , # G+4387  U+916A  uni916A,  CJK UNIFIED IDEOGRAPH-916A
        '37228' => '5616'             , # G+5616  U+916C  uni916C,  CJK UNIFIED IDEOGRAPH-916C
        '37237' => '7964'             , # G+7964  U+9175  uni9175,  CJK UNIFIED IDEOGRAPH-9175
        '37239' => '7843'             , # G+7843  U+9177  uni9177,  CJK UNIFIED IDEOGRAPH-9177
        '37240' => '5313'             , # G+5313  U+9178  uni9178,  CJK UNIFIED IDEOGRAPH-9178
        '37255' => '5663'             , # G+5663  U+9187  uni9187,  CJK UNIFIED IDEOGRAPH-9187
        '37257' => '7327'             , # G+7327  U+9189  uni9189,  CJK UNIFIED IDEOGRAPH-9189
        '37259' => '7239'             , # G+7239  U+918B  uni918B,  CJK UNIFIED IDEOGRAPH-918B
        '37261' => '6776'             , # G+6776  U+918D  uni918D,  CJK UNIFIED IDEOGRAPH-918D
        '37266' => '5491'             , # G+5491  U+9192  uni9192,  CJK UNIFIED IDEOGRAPH-9192
        '37276' => '7281'             , # G+7281  U+919C  uni919C,  CJK UNIFIED IDEOGRAPH-919C
        '37291' => '6412'             , # G+6412  U+91AB  uni91AB,  CJK UNIFIED IDEOGRAPH-91AB
        '37292' => '6567'             , # G+6567  U+91AC  uni91AC,  CJK UNIFIED IDEOGRAPH-91AC
        '37294' => '7240'             , # G+7240  U+91AE  uni91AE,  CJK UNIFIED IDEOGRAPH-91AE
        '37295' => '7798'             , # G+7798  U+91AF  uni91AF,  CJK UNIFIED IDEOGRAPH-91AF
        '37297' => '4937'             , # G+4937  U+91B1  uni91B1,  CJK UNIFIED IDEOGRAPH-91B1
        '37300' => '4509'             , # G+4509  U+91B4  uni91B4,  CJK UNIFIED IDEOGRAPH-91B4
        '37301' => '3584'             , # G+3584  U+91B5  uni91B5,  CJK UNIFIED IDEOGRAPH-91B5
        '37312' => '5899'             , # G+5899  U+91C0  uni91C0,  CJK UNIFIED IDEOGRAPH-91C0
        '37319' => '7124'             , # G+7124  U+91C7  uni91C7,  CJK UNIFIED IDEOGRAPH-91C7
        '37321' => '6353'             , # G+6353  U+91C9  uni91C9,  CJK UNIFIED IDEOGRAPH-91C9
        '37323' => '5416'             , # G+5416  U+91CB  uni91CB,  CJK UNIFIED IDEOGRAPH-91CB
        '37324' => '4634'             , # G+4634  U+91CC  uni91CC,  CJK UNIFIED IDEOGRAPH-91CC
        '37325' => '6925'             , # G+6925  U+91CD  uni91CD,  CJK UNIFIED IDEOGRAPH-91CD
        '37326' => '5871'             , # G+5871  U+91CE  uni91CE,  CJK UNIFIED IDEOGRAPH-91CE
        '37327' => '4439'             , # G+4439  U+91CF  uni91CF,  CJK UNIFIED IDEOGRAPH-91CF
        '37328' => '4635'             , # G+4635  U+91D0  uni91D0,  CJK UNIFIED IDEOGRAPH-91D0
        '37329' => '4131'             , # G+4131  U+91D1  uni91D1,  CJK UNIFIED IDEOGRAPH-91D1
        '37335' => '5566'             , # G+5566  U+91D7  uni91D7,  CJK UNIFIED IDEOGRAPH-91D7
        '37336' => '6750'             , # G+6750  U+91D8  uni91D8,  CJK UNIFIED IDEOGRAPH-91D8
        '37340' => '5146'             , # G+5146  U+91DC  uni91DC,  CJK UNIFIED IDEOGRAPH-91DC
        '37341' => '7374'             , # G+7374  U+91DD  uni91DD,  CJK UNIFIED IDEOGRAPH-91DD
        '37347' => '6824'             , # G+6824  U+91E3  uni91E3,  CJK UNIFIED IDEOGRAPH-91E3
        '37351' => '7163'             , # G+7163  U+91E7  uni91E7,  CJK UNIFIED IDEOGRAPH-91E7
        '37354' => '6225'             , # G+6225  U+91EA  uni91EA,  CJK UNIFIED IDEOGRAPH-91EA
        '37365' => '7125'             , # G+7125  U+91F5  uni91F5,  CJK UNIFIED IDEOGRAPH-91F5
        '37389' => '4363'             , # G+4363  U+920D  uni920D,  CJK UNIFIED IDEOGRAPH-920D
        '37392' => '3622'             , # G+3622  U+9210  uni9210,  CJK UNIFIED IDEOGRAPH-9210
        '37393' => '7497'             , # G+7497  U+9211  uni9211,  CJK UNIFIED IDEOGRAPH-9211
        '37394' => '5329'             , # G+5329  U+9212  uni9212,  CJK UNIFIED IDEOGRAPH-9212
        '37399' => '6366'             , # G+6366  U+9217  uni9217,  CJK UNIFIED IDEOGRAPH-9217
        '37406' => '4014'             , # G+4014  U+921E  uni921E,  CJK UNIFIED IDEOGRAPH-921E
        '37428' => '4501'             , # G+4501  U+9234  uni9234,  CJK UNIFIED IDEOGRAPH-9234
        '37434' => '6082'             , # G+6082  U+923A  uni923A,  CJK UNIFIED IDEOGRAPH-923A
        '37439' => '6676'             , # G+6676  U+923F  uni923F,  CJK UNIFIED IDEOGRAPH-923F
        '37440' => '3534'             , # G+3534  U+9240  uni9240,  CJK UNIFIED IDEOGRAPH-9240
        '37445' => '3599'             , # G+3599  U+9245  uni9245,  CJK UNIFIED IDEOGRAPH-9245
        '37449' => '7752'             , # G+7752  U+9249  uni9249,  CJK UNIFIED IDEOGRAPH-9249
        '37463' => '3658'             , # G+3658  U+9257  uni9257,  CJK UNIFIED IDEOGRAPH-9257
        '37467' => '5983'             , # G+5983  U+925B  uni925B,  CJK UNIFIED IDEOGRAPH-925B
        '37470' => '6283'             , # G+6283  U+925E  uni925E,  CJK UNIFIED IDEOGRAPH-925E
        '37474' => '4938'             , # G+4938  U+9262  uni9262,  CJK UNIFIED IDEOGRAPH-9262
        '37476' => '3940'             , # G+3940  U+9264  uni9264,  CJK UNIFIED IDEOGRAPH-9264
        '37477' => '5670'             , # G+5670  U+9265  uni9265,  CJK UNIFIED IDEOGRAPH-9265
        '37478' => '6751'             , # G+6751  U+9266  uni9266,  CJK UNIFIED IDEOGRAPH-9266
        '37504' => '6378'             , # G+6378  U+9280  uni9280,  CJK UNIFIED IDEOGRAPH-9280
        '37507' => '7260'             , # G+7260  U+9283  uni9283,  CJK UNIFIED IDEOGRAPH-9283
        '37509' => '4347'             , # G+4347  U+9285  uni9285,  CJK UNIFIED IDEOGRAPH-9285
        '37521' => '5446'             , # G+5446  U+9291  uni9291,  CJK UNIFIED IDEOGRAPH-9291
        '37523' => '6677'             , # G+6677  U+9293  uni9293,  CJK UNIFIED IDEOGRAPH-9293
        '37526' => '5617'             , # G+5617  U+9296  uni9296,  CJK UNIFIED IDEOGRAPH-9296
        '37528' => '4761'             , # G+4761  U+9298  uni9298,  CJK UNIFIED IDEOGRAPH-9298
        '37532' => '7660'             , # G+7660  U+929C  uni929C,  CJK UNIFIED IDEOGRAPH-929C
        '37555' => '6046'             , # G+6046  U+92B3  uni92B3,  CJK UNIFIED IDEOGRAPH-92B3
        '37558' => '3941'             , # G+3941  U+92B6  uni92B6,  CJK UNIFIED IDEOGRAPH-92B6
        '37559' => '5534'             , # G+5534  U+92B7  uni92B7,  CJK UNIFIED IDEOGRAPH-92B7
        '37561' => '5618'             , # G+5618  U+92B9  uni92B9,  CJK UNIFIED IDEOGRAPH-92B9
        '37580' => '6752'             , # G+6752  U+92CC  uni92CC,  CJK UNIFIED IDEOGRAPH-92CC
        '37583' => '7769'             , # G+7769  U+92CF  uni92CF,  CJK UNIFIED IDEOGRAPH-92CF
        '37586' => '5107'             , # G+5107  U+92D2  uni92D2,  CJK UNIFIED IDEOGRAPH-92D2
        '37604' => '5400'             , # G+5400  U+92E4  uni92E4,  CJK UNIFIED IDEOGRAPH-92E4
        '37610' => '7568'             , # G+7568  U+92EA  uni92EA,  CJK UNIFIED IDEOGRAPH-92EA
        '37624' => '3600'             , # G+3600  U+92F8  uni92F8,  CJK UNIFIED IDEOGRAPH-92F8
        '37628' => '3557'             , # G+3557  U+92FC  uni92FC,  CJK UNIFIED IDEOGRAPH-92FC
        '37636' => '4532'             , # G+4532  U+9304  uni9304,  CJK UNIFIED IDEOGRAPH-9304
        '37648' => '7282'             , # G+7282  U+9310  uni9310,  CJK UNIFIED IDEOGRAPH-9310
        '37656' => '7283'             , # G+7283  U+9318  uni9318,  CJK UNIFIED IDEOGRAPH-9318
        '37658' => '6590'             , # G+6590  U+931A  uni931A,  CJK UNIFIED IDEOGRAPH-931A
        '37662' => '5664'             , # G+5664  U+931E  uni931E,  CJK UNIFIED IDEOGRAPH-931E
        '37663' => '4233'             , # G+4233  U+931F  uni931F,  CJK UNIFIED IDEOGRAPH-931F
        '37664' => '6753'             , # G+6753  U+9320  uni9320,  CJK UNIFIED IDEOGRAPH-9320
        '37665' => '4118'             , # G+4118  U+9321  uni9321,  CJK UNIFIED IDEOGRAPH-9321
        '37666' => '6678'             , # G+6678  U+9322  uni9322,  CJK UNIFIED IDEOGRAPH-9322
        '37668' => '4119'             , # G+4119  U+9324  uni9324,  CJK UNIFIED IDEOGRAPH-9324
        '37670' => '4050'             , # G+4050  U+9326  uni9326,  CJK UNIFIED IDEOGRAPH-9326
        '37672' => '4811'             , # G+4811  U+9328  uni9328,  CJK UNIFIED IDEOGRAPH-9328
        '37675' => '5417'             , # G+5417  U+932B  uni932B,  CJK UNIFIED IDEOGRAPH-932B
        '37678' => '3762'             , # G+3762  U+932E  uni932E,  CJK UNIFIED IDEOGRAPH-932E
        '37679' => '7059'             , # G+7059  U+932F  uni932F,  CJK UNIFIED IDEOGRAPH-932F
        '37704' => '6027'             , # G+6027  U+9348  uni9348,  CJK UNIFIED IDEOGRAPH-9348
        '37706' => '4476'             , # G+4476  U+934A  uni934A,  CJK UNIFIED IDEOGRAPH-934A
        '37707' => '3813'             , # G+3813  U+934B  uni934B,  CJK UNIFIED IDEOGRAPH-934B
        '37709' => '4306'             , # G+4306  U+934D  uni934D,  CJK UNIFIED IDEOGRAPH-934D
        '37716' => '5803'             , # G+5803  U+9354  uni9354,  CJK UNIFIED IDEOGRAPH-9354
        '37723' => '4211'             , # G+4211  U+935B  uni935B,  CJK UNIFIED IDEOGRAPH-935B
        '37742' => '6354'             , # G+6354  U+936E  uni936E,  CJK UNIFIED IDEOGRAPH-936E
        '37749' => '3611'             , # G+3611  U+9375  uni9375,  CJK UNIFIED IDEOGRAPH-9375
        '37756' => '7375'             , # G+7375  U+937C  uni937C,  CJK UNIFIED IDEOGRAPH-937C
        '37758' => '6852'             , # G+6852  U+937E  uni937E,  CJK UNIFIED IDEOGRAPH-937E
        '37772' => '3659'             , # G+3659  U+938C  uni938C,  CJK UNIFIED IDEOGRAPH-938C
        '37780' => '6195'             , # G+6195  U+9394  uni9394,  CJK UNIFIED IDEOGRAPH-9394
        '37782' => '5564'             , # G+5564  U+9396  uni9396,  CJK UNIFIED IDEOGRAPH-9396
        '37786' => '7284'             , # G+7284  U+939A  uni939A,  CJK UNIFIED IDEOGRAPH-939A
        '37795' => '7789'             , # G+7789  U+93A3  uni93A3,  CJK UNIFIED IDEOGRAPH-93A3
        '37799' => '3577'             , # G+3577  U+93A7  uni93A7,  CJK UNIFIED IDEOGRAPH-93A7
        '37804' => '7838'             , # G+7838  U+93AC  uni93AC,  CJK UNIFIED IDEOGRAPH-93AC
        '37805' => '7011'             , # G+7011  U+93AD  uni93AD,  CJK UNIFIED IDEOGRAPH-93AD
        '37808' => '6467'             , # G+6467  U+93B0  uni93B0,  CJK UNIFIED IDEOGRAPH-93B0
        '37827' => '6831'             , # G+6831  U+93C3  uni93C3,  CJK UNIFIED IDEOGRAPH-93C3
        '37841' => '6642'             , # G+6642  U+93D1  uni93D1,  CJK UNIFIED IDEOGRAPH-93D1
        '37854' => '6196'             , # G+6196  U+93DE  uni93DE,  CJK UNIFIED IDEOGRAPH-93DE
        '37857' => '3700'             , # G+3700  U+93E1  uni93E1,  CJK UNIFIED IDEOGRAPH-93E1
        '37860' => '4575'             , # G+4575  U+93E4  uni93E4,  CJK UNIFIED IDEOGRAPH-93E4
        '37878' => '7036'             , # G+7036  U+93F6  uni93F6,  CJK UNIFIED IDEOGRAPH-93F6
        '37892' => '7952'             , # G+7952  U+9404  uni9404,  CJK UNIFIED IDEOGRAPH-9404
        '37912' => '6853'             , # G+6853  U+9418  uni9418,  CJK UNIFIED IDEOGRAPH-9418
        '37925' => '5447'             , # G+5447  U+9425  uni9425,  CJK UNIFIED IDEOGRAPH-9425
        '37931' => '6679'             , # G+6679  U+942B  uni942B,  CJK UNIFIED IDEOGRAPH-942B
        '37941' => '7176'             , # G+7176  U+9435  uni9435,  CJK UNIFIED IDEOGRAPH-9435
        '37944' => '7407'             , # G+7407  U+9438  uni9438,  CJK UNIFIED IDEOGRAPH-9438
        '37956' => '6898'             , # G+6898  U+9444  uni9444,  CJK UNIFIED IDEOGRAPH-9444
        '37969' => '3527'             , # G+3527  U+9451  uni9451,  CJK UNIFIED IDEOGRAPH-9451
        '37970' => '3528'             , # G+3528  U+9452  uni9452,  CJK UNIFIED IDEOGRAPH-9452
        '37979' => '3851'             , # G+3851  U+945B  uni945B,  CJK UNIFIED IDEOGRAPH-945B
        '38013' => '7074'             , # G+7074  U+947D  uni947D,  CJK UNIFIED IDEOGRAPH-947D
        '38015' => '7060'             , # G+7060  U+947F  uni947F,  CJK UNIFIED IDEOGRAPH-947F
        '38263' => '6568'             , # G+6568  U+9577  uni9577,  CJK UNIFIED IDEOGRAPH-9577
        '38272' => '4846'             , # G+4846  U+9580  uni9580,  CJK UNIFIED IDEOGRAPH-9580
        '38275' => '5469'             , # G+5469  U+9583  uni9583,  CJK UNIFIED IDEOGRAPH-9583
        '38281' => '7542'             , # G+7542  U+9589  uni9589,  CJK UNIFIED IDEOGRAPH-9589
        '38283' => '3578'             , # G+3578  U+958B  uni958B,  CJK UNIFIED IDEOGRAPH-958B
        '38287' => '6367'             , # G+6367  U+958F  uni958F,  CJK UNIFIED IDEOGRAPH-958F
        '38289' => '7645'             , # G+7645  U+9591  uni9591,  CJK UNIFIED IDEOGRAPH-9591
        '38290' => '7646'             , # G+7646  U+9592  uni9592,  CJK UNIFIED IDEOGRAPH-9592
        '38291' => '3499'             , # G+3499  U+9593  uni9593,  CJK UNIFIED IDEOGRAPH-9593
        '38292' => '4882'             , # G+4882  U+9594  uni9594,  CJK UNIFIED IDEOGRAPH-9594
        '38296' => '3535'             , # G+3535  U+9598  uni9598,  CJK UNIFIED IDEOGRAPH-9598
        '38307' => '3475'             , # G+3475  U+95A3  uni95A3,  CJK UNIFIED IDEOGRAPH-95A3
        '38308' => '7667'             , # G+7667  U+95A4  uni95A4,  CJK UNIFIED IDEOGRAPH-95A4
        '38309' => '5008'             , # G+5008  U+95A5  uni95A5,  CJK UNIFIED IDEOGRAPH-95A5
        '38312' => '4008'             , # G+4008  U+95A8  uni95A8,  CJK UNIFIED IDEOGRAPH-95A8
        '38317' => '4453'             , # G+4453  U+95AD  uni95AD,  CJK UNIFIED IDEOGRAPH-95AD
        '38321' => '5988'             , # G+5988  U+95B1  uni95B1,  CJK UNIFIED IDEOGRAPH-95B1
        '38331' => '5996'             , # G+5996  U+95BB  uni95BB,  CJK UNIFIED IDEOGRAPH-95BB
        '38332' => '5820'             , # G+5820  U+95BC  uni95BC,  CJK UNIFIED IDEOGRAPH-95BC
        '38343' => '5828'             , # G+5828  U+95C7  uni95C7,  CJK UNIFIED IDEOGRAPH-95C7
        '38346' => '7903'             , # G+7903  U+95CA  uni95CA,  CJK UNIFIED IDEOGRAPH-95CA
        '38356' => '7668'             , # G+7668  U+95D4  uni95D4,  CJK UNIFIED IDEOGRAPH-95D4
        '38357' => '3983'             , # G+3983  U+95D5  uni95D5,  CJK UNIFIED IDEOGRAPH-95D5
        '38358' => '7473'             , # G+7473  U+95D6  uni95D6,  CJK UNIFIED IDEOGRAPH-95D6
        '38364' => '3833'             , # G+3833  U+95DC  uni95DC,  CJK UNIFIED IDEOGRAPH-95DC
        '38369' => '7164'             , # G+7164  U+95E1  uni95E1,  CJK UNIFIED IDEOGRAPH-95E1
        '38370' => '5029'             , # G+5029  U+95E2  uni95E2,  CJK UNIFIED IDEOGRAPH-95E2
        '38428' => '5147'             , # G+5147  U+961C  uni961C,  CJK UNIFIED IDEOGRAPH-961C
        '38433' => '7165'             , # G+7165  U+9621  uni9621,  CJK UNIFIED IDEOGRAPH-9621
        '38442' => '7498'             , # G+7498  U+962A  uni962A,  CJK UNIFIED IDEOGRAPH-962A
        '38446' => '6123'             , # G+6123  U+962E  uni962E,  CJK UNIFIED IDEOGRAPH-962E
        '38450' => '4967'             , # G+4967  U+9632  uni9632,  CJK UNIFIED IDEOGRAPH-9632
        '38459' => '6825'             , # G+6825  U+963B  uni963B,  CJK UNIFIED IDEOGRAPH-963B
        '38463' => '5788'             , # G+5788  U+963F  uni963F,  CJK UNIFIED IDEOGRAPH-963F
        '38464' => '7391'             , # G+7391  U+9640  uni9640,  CJK UNIFIED IDEOGRAPH-9640
        '38466' => '7603'             , # G+7603  U+9642  uni9642,  CJK UNIFIED IDEOGRAPH-9642
        '38468' => '5148'             , # G+5148  U+9644  uni9644,  CJK UNIFIED IDEOGRAPH-9644
        '38475' => '4576'             , # G+4576  U+964B  uni964B,  CJK UNIFIED IDEOGRAPH-964B
        '38476' => '4724'             , # G+4724  U+964C  uni964C,  CJK UNIFIED IDEOGRAPH-964C
        '38477' => '3558'             , # G+3558  U+964D  uni964D,  CJK UNIFIED IDEOGRAPH-964D
        '38480' => '7647'             , # G+7647  U+9650  uni9650,  CJK UNIFIED IDEOGRAPH-9650
        '38491' => '7543'             , # G+7543  U+965B  uni965B,  CJK UNIFIED IDEOGRAPH-965B
        '38492' => '7669'             , # G+7669  U+965C  uni965C,  CJK UNIFIED IDEOGRAPH-965C
        '38493' => '5470'             , # G+5470  U+965D  uni965D,  CJK UNIFIED IDEOGRAPH-965D
        '38494' => '5691'             , # G+5691  U+965E  uni965E,  CJK UNIFIED IDEOGRAPH-965E
        '38495' => '7146'             , # G+7146  U+965F  uni965F,  CJK UNIFIED IDEOGRAPH-965F
        '38498' => '6278'             , # G+6278  U+9662  uni9662,  CJK UNIFIED IDEOGRAPH-9662
        '38499' => '7012'             , # G+7012  U+9663  uni9663,  CJK UNIFIED IDEOGRAPH-9663
        '38500' => '6777'             , # G+6777  U+9664  uni9664,  CJK UNIFIED IDEOGRAPH-9664
        '38506' => '4987'             , # G+4987  U+966A  uni966A,  CJK UNIFIED IDEOGRAPH-966A
        '38512' => '6384'             , # G+6384  U+9670  uni9670,  CJK UNIFIED IDEOGRAPH-9670
        '38515' => '7013'             , # G+7013  U+9673  uni9673,  CJK UNIFIED IDEOGRAPH-9673
        '38517' => '4612'             , # G+4612  U+9675  uni9675,  CJK UNIFIED IDEOGRAPH-9675
        '38518' => '4307'             , # G+4307  U+9676  uni9676,  CJK UNIFIED IDEOGRAPH-9676
        '38519' => '7661'             , # G+7661  U+9677  uni9677,  CJK UNIFIED IDEOGRAPH-9677
        '38520' => '4593'             , # G+4593  U+9678  uni9678,  CJK UNIFIED IDEOGRAPH-9678
        '38525' => '5900'             , # G+5900  U+967D  uni967D,  CJK UNIFIED IDEOGRAPH-967D
        '38533' => '6226'             , # G+6226  U+9685  uni9685,  CJK UNIFIED IDEOGRAPH-9685
        '38534' => '4603'             , # G+4603  U+9686  uni9686,  CJK UNIFIED IDEOGRAPH-9686
        '38538' => '4264'             , # G+4264  U+968A  uni968A,  CJK UNIFIED IDEOGRAPH-968A
        '38539' => '5619'             , # G+5619  U+968B  uni968B,  CJK UNIFIED IDEOGRAPH-968B
        '38541' => '7926'             , # G+7926  U+968D  uni968D,  CJK UNIFIED IDEOGRAPH-968D
        '38542' => '3727'             , # G+3727  U+968E  uni968E,  CJK UNIFIED IDEOGRAPH-968E
        '38548' => '3636'             , # G+3636  U+9694  uni9694,  CJK UNIFIED IDEOGRAPH-9694
        '38549' => '6247'             , # G+6247  U+9695  uni9695,  CJK UNIFIED IDEOGRAPH-9695
        '38552' => '5849'             , # G+5849  U+9698  uni9698,  CJK UNIFIED IDEOGRAPH-9698
        '38553' => '4022'             , # G+4022  U+9699  uni9699,  CJK UNIFIED IDEOGRAPH-9699
        '38555' => '6778'             , # G+6778  U+969B  uni969B,  CJK UNIFIED IDEOGRAPH-969B
        '38556' => '6569'             , # G+6569  U+969C  uni969C,  CJK UNIFIED IDEOGRAPH-969C
        '38563' => '4644'             , # G+4644  U+96A3  uni96A3,  CJK UNIFIED IDEOGRAPH-96A3
        '38567' => '5620'             , # G+5620  U+96A7  uni96A7,  CJK UNIFIED IDEOGRAPH-96A7
        '38568' => '5621'             , # G+5621  U+96A8  uni96A8,  CJK UNIFIED IDEOGRAPH-96A8
        '38570' => '7728'             , # G+7728  U+96AA  uni96AA,  CJK UNIFIED IDEOGRAPH-96AA
        '38577' => '6379'             , # G+6379  U+96B1  uni96B1,  CJK UNIFIED IDEOGRAPH-96B1
        '38583' => '4510'             , # G+4510  U+96B7  uni96B7,  CJK UNIFIED IDEOGRAPH-96B7
        '38587' => '7147'             , # G+7147  U+96BB  uni96BB,  CJK UNIFIED IDEOGRAPH-96BB
        '38592' => '6520'             , # G+6520  U+96C0  uni96C0,  CJK UNIFIED IDEOGRAPH-96C0
        '38593' => '5813'             , # G+5813  U+96C1  uni96C1,  CJK UNIFIED IDEOGRAPH-96C1
        '38596' => '6254'             , # G+6254  U+96C4  uni96C4,  CJK UNIFIED IDEOGRAPH-96C4
        '38597' => '5789'             , # G+5789  U+96C5  uni96C5,  CJK UNIFIED IDEOGRAPH-96C5
        '38598' => '7037'             , # G+7037  U+96C6  uni96C6,  CJK UNIFIED IDEOGRAPH-96C6
        '38599' => '3763'             , # G+3763  U+96C7  uni96C7,  CJK UNIFIED IDEOGRAPH-96C7
        '38601' => '7357'             , # G+7357  U+96C9  uni96C9,  CJK UNIFIED IDEOGRAPH-96C9
        '38603' => '6919'             , # G+6919  U+96CB  uni96CB,  CJK UNIFIED IDEOGRAPH-96CB
        '38604' => '6508'             , # G+6508  U+96CC  uni96CC,  CJK UNIFIED IDEOGRAPH-96CC
        '38605' => '6097'             , # G+6097  U+96CD  uni96CD,  CJK UNIFIED IDEOGRAPH-96CD
        '38606' => '6617'             , # G+6617  U+96CE  uni96CE,  CJK UNIFIED IDEOGRAPH-96CE
        '38613' => '6826'             , # G+6826  U+96D5  uni96D5,  CJK UNIFIED IDEOGRAPH-96D5
        '38614' => '5622'             , # G+5622  U+96D6  uni96D6,  CJK UNIFIED IDEOGRAPH-96D6
        '38617' => '5773'             , # G+5773  U+96D9  uni96D9,  CJK UNIFIED IDEOGRAPH-96D9
        '38619' => '7285'             , # G+7285  U+96DB  uni96DB,  CJK UNIFIED IDEOGRAPH-96DB
        '38620' => '6533'             , # G+6533  U+96DC  uni96DC,  CJK UNIFIED IDEOGRAPH-96DC
        '38626' => '4636'             , # G+4636  U+96E2  uni96E2,  CJK UNIFIED IDEOGRAPH-96E2
        '38627' => '4142'             , # G+4142  U+96E3  uni96E3,  CJK UNIFIED IDEOGRAPH-96E3
        '38632' => '6227'             , # G+6227  U+96E8  uni96E8,  CJK UNIFIED IDEOGRAPH-96E8
        '38633' => '6228'             , # G+6228  U+96E9  uni96E9,  CJK UNIFIED IDEOGRAPH-96E9
        '38634' => '5461'             , # G+5461  U+96EA  uni96EA,  CJK UNIFIED IDEOGRAPH-96EA
        '38639' => '4847'             , # G+4847  U+96EF  uni96EF,  CJK UNIFIED IDEOGRAPH-96EF
        '38640' => '5170'             , # G+5170  U+96F0  uni96F0,  CJK UNIFIED IDEOGRAPH-96F0
        '38642' => '6248'             , # G+6248  U+96F2  uni96F2,  CJK UNIFIED IDEOGRAPH-96F2
        '38646' => '4502'             , # G+4502  U+96F6  uni96F6,  CJK UNIFIED IDEOGRAPH-96F6
        '38647' => '4550'             , # G+4550  U+96F7  uni96F7,  CJK UNIFIED IDEOGRAPH-96F7
        '38649' => '4903'             , # G+4903  U+96F9  uni96F9,  CJK UNIFIED IDEOGRAPH-96F9
        '38651' => '6680'             , # G+6680  U+96FB  uni96FB,  CJK UNIFIED IDEOGRAPH-96FB
        '38656' => '5623'             , # G+5623  U+9700  uni9700,  CJK UNIFIED IDEOGRAPH-9700
        '38662' => '6754'             , # G+6754  U+9706  uni9706,  CJK UNIFIED IDEOGRAPH-9706
        '38663' => '7014'             , # G+7014  U+9707  uni9707,  CJK UNIFIED IDEOGRAPH-9707
        '38673' => '6698'             , # G+6698  U+9711  uni9711,  CJK UNIFIED IDEOGRAPH-9711
        '38675' => '6047'             , # G+6047  U+9713  uni9713,  CJK UNIFIED IDEOGRAPH-9713
        '38678' => '4651'             , # G+4651  U+9716  uni9716,  CJK UNIFIED IDEOGRAPH-9716
        '38681' => '6028'             , # G+6028  U+9719  uni9719,  CJK UNIFIED IDEOGRAPH-9719
        '38684' => '5361'             , # G+5361  U+971C  uni971C,  CJK UNIFIED IDEOGRAPH-971C
        '38686' => '7628'             , # G+7628  U+971E  uni971E,  CJK UNIFIED IDEOGRAPH-971E
        '38695' => '4832'             , # G+4832  U+9727  uni9727,  CJK UNIFIED IDEOGRAPH-9727
        '38704' => '5314'             , # G+5314  U+9730  uni9730,  CJK UNIFIED IDEOGRAPH-9730
        '38706' => '4524'             , # G+4524  U+9732  uni9732,  CJK UNIFIED IDEOGRAPH-9732
        '38713' => '5030'             , # G+5030  U+9739  uni9739,  CJK UNIFIED IDEOGRAPH-9739
        '38717' => '6779'             , # G+6779  U+973D  uni973D,  CJK UNIFIED IDEOGRAPH-973D
        '38722' => '4464'             , # G+4464  U+9742  uni9742,  CJK UNIFIED IDEOGRAPH-9742
        '38724' => '5850'             , # G+5850  U+9744  uni9744,  CJK UNIFIED IDEOGRAPH-9744
        '38728' => '4503'             , # G+4503  U+9748  uni9748,  CJK UNIFIED IDEOGRAPH-9748
        '38737' => '7203'             , # G+7203  U+9751  uni9751,  CJK UNIFIED IDEOGRAPH-9751
        '38742' => '6755'             , # G+6755  U+9756  uni9756,  CJK UNIFIED IDEOGRAPH-9756
        '38748' => '6756'             , # G+6756  U+975C  uni975C,  CJK UNIFIED IDEOGRAPH-975C
        '38750' => '5220'             , # G+5220  U+975E  uni975E,  CJK UNIFIED IDEOGRAPH-975E
        '38753' => '4868'             , # G+4868  U+9761  uni9761,  CJK UNIFIED IDEOGRAPH-9761
        '38754' => '4744'             , # G+4744  U+9762  uni9762,  CJK UNIFIED IDEOGRAPH-9762
        '38761' => '7733'             , # G+7733  U+9769  uni9769,  CJK UNIFIED IDEOGRAPH-9769
        '38765' => '6458'             , # G+6458  U+976D  uni976D,  CJK UNIFIED IDEOGRAPH-976D
        '38772' => '7876'             , # G+7876  U+9774  uni9774,  CJK UNIFIED IDEOGRAPH-9774
        '38775' => '6459'             , # G+6459  U+9777  uni9777,  CJK UNIFIED IDEOGRAPH-9777
        '38778' => '4695'             , # G+4695  U+977A  uni977A,  CJK UNIFIED IDEOGRAPH-977A
        '38795' => '7799'             , # G+7799  U+978B  uni978B,  CJK UNIFIED IDEOGRAPH-978B
        '38797' => '5814'             , # G+5814  U+978D  uni978D,  CJK UNIFIED IDEOGRAPH-978D
        '38799' => '3801'             , # G+3801  U+978F  uni978F,  CJK UNIFIED IDEOGRAPH-978F
        '38816' => '3950'             , # G+3950  U+97A0  uni97A0,  CJK UNIFIED IDEOGRAPH-97A0
        '38824' => '3509'             , # G+3509  U+97A8  uni97A8,  CJK UNIFIED IDEOGRAPH-97A8
        '38827' => '3951'             , # G+3951  U+97AB  uni97AB,  CJK UNIFIED IDEOGRAPH-97AB
        '38829' => '7526'             , # G+7526  U+97AD  uni97AD,  CJK UNIFIED IDEOGRAPH-97AD
        '38854' => '7166'             , # G+7166  U+97C6  uni97C6,  CJK UNIFIED IDEOGRAPH-97C6
        '38859' => '6307'             , # G+6307  U+97CB  uni97CB,  CJK UNIFIED IDEOGRAPH-97CB
        '38867' => '7648'             , # G+7648  U+97D3  uni97D3,  CJK UNIFIED IDEOGRAPH-97D3
        '38876' => '4308'             , # G+4308  U+97DC  uni97DC,  CJK UNIFIED IDEOGRAPH-97DC
        '38899' => '6385'             , # G+6385  U+97F3  uni97F3,  CJK UNIFIED IDEOGRAPH-97F3
        '38902' => '5535'             , # G+5535  U+97F6  uni97F6,  CJK UNIFIED IDEOGRAPH-97F6
        '38907' => '6249'             , # G+6249  U+97FB  uni97FB,  CJK UNIFIED IDEOGRAPH-97FB
        '38911' => '7715'             , # G+7715  U+97FF  uni97FF,  CJK UNIFIED IDEOGRAPH-97FF
        '38912' => '7839'             , # G+7839  U+9800  uni9800,  CJK UNIFIED IDEOGRAPH-9800
        '38913' => '7757'             , # G+7757  U+9801  uni9801,  CJK UNIFIED IDEOGRAPH-9801
        '38914' => '6757'             , # G+6757  U+9802  uni9802,  CJK UNIFIED IDEOGRAPH-9802
        '38915' => '3701'             , # G+3701  U+9803  uni9803,  CJK UNIFIED IDEOGRAPH-9803
        '38917' => '7684'             , # G+7684  U+9805  uni9805,  CJK UNIFIED IDEOGRAPH-9805
        '38918' => '5665'             , # G+5665  U+9806  uni9806,  CJK UNIFIED IDEOGRAPH-9806
        '38920' => '5624'             , # G+5624  U+9808  uni9808,  CJK UNIFIED IDEOGRAPH-9808
        '38922' => '6237'             , # G+6237  U+980A  uni980A,  CJK UNIFIED IDEOGRAPH-980A
        '38924' => '5560'             , # G+5560  U+980C  uni980C,  CJK UNIFIED IDEOGRAPH-980C
        '38928' => '6048'             , # G+6048  U+9810  uni9810,  CJK UNIFIED IDEOGRAPH-9810
        '38929' => '6124'             , # G+6124  U+9811  uni9811,  CJK UNIFIED IDEOGRAPH-9811
        '38930' => '4928'             , # G+4928  U+9812  uni9812,  CJK UNIFIED IDEOGRAPH-9812
        '38931' => '4328'             , # G+4328  U+9813  uni9813,  CJK UNIFIED IDEOGRAPH-9813
        '38935' => '7489'             , # G+7489  U+9817  uni9817,  CJK UNIFIED IDEOGRAPH-9817
        '38936' => '4504'             , # G+4504  U+9818  uni9818,  CJK UNIFIED IDEOGRAPH-9818
        '38957' => '4357'             , # G+4357  U+982D  uni982D,  CJK UNIFIED IDEOGRAPH-982D
        '38960' => '7770'             , # G+7770  U+9830  uni9830,  CJK UNIFIED IDEOGRAPH-9830
        '38968' => '3702'             , # G+3702  U+9838  uni9838,  CJK UNIFIED IDEOGRAPH-9838
        '38969' => '7464'             , # G+7464  U+9839  uni9839,  CJK UNIFIED IDEOGRAPH-9839
        '38971' => '5236'             , # G+5236  U+983B  uni983B,  CJK UNIFIED IDEOGRAPH-983B
        '38982' => '3814'             , # G+3814  U+9846  uni9846,  CJK UNIFIED IDEOGRAPH-9846
        '38988' => '6780'             , # G+6780  U+984C  uni984C,  CJK UNIFIED IDEOGRAPH-984C
        '38989' => '5857'             , # G+5857  U+984D  uni984D,  CJK UNIFIED IDEOGRAPH-984D
        '38990' => '5804'             , # G+5804  U+984E  uni984E,  CJK UNIFIED IDEOGRAPH-984E
        '38996' => '5815'             , # G+5815  U+9854  uni9854,  CJK UNIFIED IDEOGRAPH-9854
        '39000' => '6279'             , # G+6279  U+9858  uni9858,  CJK UNIFIED IDEOGRAPH-9858
        '39002' => '6681'             , # G+6681  U+985A  uni985A,  CJK UNIFIED IDEOGRAPH-985A
        '39006' => '4590'             , # G+4590  U+985E  uni985E,  CJK UNIFIED IDEOGRAPH-985E
        '39013' => '7840'             , # G+7840  U+9865  uni9865,  CJK UNIFIED IDEOGRAPH-9865
        '39015' => '3764'             , # G+3764  U+9867  uni9867,  CJK UNIFIED IDEOGRAPH-9867
        '39019' => '6682'             , # G+6682  U+986B  uni986B,  CJK UNIFIED IDEOGRAPH-986B
        '39023' => '7753'             , # G+7753  U+986F  uni986F,  CJK UNIFIED IDEOGRAPH-986F
        '39080' => '7595'             , # G+7595  U+98A8  uni98A8,  CJK UNIFIED IDEOGRAPH-98A8
        '39087' => '5330'             , # G+5330  U+98AF  uni98AF,  CJK UNIFIED IDEOGRAPH-98AF
        '39089' => '7444'             , # G+7444  U+98B1  uni98B1,  CJK UNIFIED IDEOGRAPH-98B1
        '39108' => '7588'             , # G+7588  U+98C4  uni98C4,  CJK UNIFIED IDEOGRAPH-98C4
        '39111' => '7587'             , # G+7587  U+98C7  uni98C7,  CJK UNIFIED IDEOGRAPH-98C7
        '39131' => '5221'             , # G+5221  U+98DB  uni98DB,  CJK UNIFIED IDEOGRAPH-98DB
        '39132' => '5004'             , # G+5004  U+98DC  uni98DC,  CJK UNIFIED IDEOGRAPH-98DC
        '39135' => '5733'             , # G+5733  U+98DF  uni98DF,  CJK UNIFIED IDEOGRAPH-98DF
        '39137' => '5551'             , # G+5551  U+98E1  uni98E1,  CJK UNIFIED IDEOGRAPH-98E1
        '39138' => '4120'             , # G+4120  U+98E2  uni98E2,  CJK UNIFIED IDEOGRAPH-98E2
        '39149' => '7362'             , # G+7362  U+98ED  uni98ED,  CJK UNIFIED IDEOGRAPH-98ED
        '39150' => '6386'             , # G+6386  U+98EE  uni98EE,  CJK UNIFIED IDEOGRAPH-98EE
        '39151' => '4929'             , # G+4929  U+98EF  uni98EF,  CJK UNIFIED IDEOGRAPH-98EF
        '39156' => '6435'             , # G+6435  U+98F4  uni98F4,  CJK UNIFIED IDEOGRAPH-98F4
        '39164' => '5298'             , # G+5298  U+98FC  uni98FC,  CJK UNIFIED IDEOGRAPH-98FC
        '39165' => '7569'             , # G+7569  U+98FD  uni98FD,  CJK UNIFIED IDEOGRAPH-98FD
        '39166' => '5734'             , # G+5734  U+98FE  uni98FE,  CJK UNIFIED IDEOGRAPH-98FE
        '39171' => '3890'             , # G+3890  U+9903  uni9903,  CJK UNIFIED IDEOGRAPH-9903
        '39177' => '7716'             , # G+7716  U+9909  uni9909,  CJK UNIFIED IDEOGRAPH-9909
        '39178' => '5901'             , # G+5901  U+990A  uni990A,  CJK UNIFIED IDEOGRAPH-990A
        '39180' => '6436'             , # G+6436  U+990C  uni990C,  CJK UNIFIED IDEOGRAPH-990C
        '39184' => '7075'             , # G+7075  U+9910  uni9910,  CJK UNIFIED IDEOGRAPH-9910
        '39187' => '5790'             , # G+5790  U+9913  uni9913,  CJK UNIFIED IDEOGRAPH-9913
        '39192' => '5946'             , # G+5946  U+9918  uni9918,  CJK UNIFIED IDEOGRAPH-9918
        '39198' => '6683'             , # G+6683  U+991E  uni991E,  CJK UNIFIED IDEOGRAPH-991E
        '39200' => '5056'             , # G+5056  U+9920  uni9920,  CJK UNIFIED IDEOGRAPH-9920
        '39208' => '3834'             , # G+3834  U+9928  uni9928,  CJK UNIFIED IDEOGRAPH-9928
        '39237' => '4687'             , # G+4687  U+9945  uni9945,  CJK UNIFIED IDEOGRAPH-9945
        '39241' => '4037'             , # G+4037  U+9949  uni9949,  CJK UNIFIED IDEOGRAPH-9949
        '39243' => '3989'             , # G+3989  U+994B  uni994B,  CJK UNIFIED IDEOGRAPH-994B
        '39244' => '7076'             , # G+7076  U+994C  uni994C,  CJK UNIFIED IDEOGRAPH-994C
        '39245' => '5448'             , # G+5448  U+994D  uni994D,  CJK UNIFIED IDEOGRAPH-994D
        '39249' => '4121'             , # G+4121  U+9951  uni9951,  CJK UNIFIED IDEOGRAPH-9951
        '39250' => '6167'             , # G+6167  U+9952  uni9952,  CJK UNIFIED IDEOGRAPH-9952
        '39252' => '6098'             , # G+6098  U+9954  uni9954,  CJK UNIFIED IDEOGRAPH-9954
        '39255' => '7717'             , # G+7717  U+9957  uni9957,  CJK UNIFIED IDEOGRAPH-9957
        '39318' => '5625'             , # G+5625  U+9996  uni9996,  CJK UNIFIED IDEOGRAPH-9996
        '39321' => '7718'             , # G+7718  U+9999  uni9999,  CJK UNIFIED IDEOGRAPH-9999
        '39325' => '7613'             , # G+7613  U+999D  uni999D,  CJK UNIFIED IDEOGRAPH-999D
        '39333' => '5089'             , # G+5089  U+99A5  uni99A5,  CJK UNIFIED IDEOGRAPH-99A5
        '39336' => '7790'             , # G+7790  U+99A8  uni99A8,  CJK UNIFIED IDEOGRAPH-99A8
        '39340' => '4661'             , # G+4661  U+99AC  uni99AC,  CJK UNIFIED IDEOGRAPH-99AC
        '39341' => '5909'             , # G+5909  U+99AD  uni99AD,  CJK UNIFIED IDEOGRAPH-99AD
        '39342' => '7596'             , # G+7596  U+99AE  uni99AE,  CJK UNIFIED IDEOGRAPH-99AE
        '39345' => '7392'             , # G+7392  U+99B1  uni99B1,  CJK UNIFIED IDEOGRAPH-99B1
        '39347' => '7358'             , # G+7358  U+99B3  uni99B3,  CJK UNIFIED IDEOGRAPH-99B3
        '39348' => '5666'             , # G+5666  U+99B4  uni99B4,  CJK UNIFIED IDEOGRAPH-99B4
        '39353' => '6468'             , # G+6468  U+99B9  uni99B9,  CJK UNIFIED IDEOGRAPH-99B9
        '39361' => '4904'             , # G+4904  U+99C1  uni99C1,  CJK UNIFIED IDEOGRAPH-99C1
        '39376' => '6899'             , # G+6899  U+99D0  uni99D0,  CJK UNIFIED IDEOGRAPH-99D0
        '39377' => '4174'             , # G+4174  U+99D1  uni99D1,  CJK UNIFIED IDEOGRAPH-99D1
        '39378' => '3942'             , # G+3942  U+99D2  uni99D2,  CJK UNIFIED IDEOGRAPH-99D2
        '39381' => '3464'             , # G+3464  U+99D5  uni99D5,  CJK UNIFIED IDEOGRAPH-99D5
        '39385' => '5149'             , # G+5149  U+99D9  uni99D9,  CJK UNIFIED IDEOGRAPH-99D9
        '39389' => '7393'             , # G+7393  U+99DD  uni99DD,  CJK UNIFIED IDEOGRAPH-99DD
        '39391' => '5299'             , # G+5299  U+99DF  uni99DF,  CJK UNIFIED IDEOGRAPH-99DF
        '39405' => '7701'             , # G+7701  U+99ED  uni99ED,  CJK UNIFIED IDEOGRAPH-99ED
        '39409' => '4388'             , # G+4388  U+99F1  uni99F1,  CJK UNIFIED IDEOGRAPH-99F1
        '39423' => '6920'             , # G+6920  U+99FF  uni99FF,  CJK UNIFIED IDEOGRAPH-99FF
        '39425' => '5240'             , # G+5240  U+9A01  uni9A01,  CJK UNIFIED IDEOGRAPH-9A01
        '39432' => '5057'             , # G+5057  U+9A08  uni9A08,  CJK UNIFIED IDEOGRAPH-9A08
        '39438' => '4122'             , # G+4122  U+9A0E  uni9A0E,  CJK UNIFIED IDEOGRAPH-9A0E
        '39439' => '4123'             , # G+4123  U+9A0F  uni9A0F,  CJK UNIFIED IDEOGRAPH-9A0F
        '39449' => '7527'             , # G+7527  U+9A19  uni9A19,  CJK UNIFIED IDEOGRAPH-9A19
        '39467' => '3612'             , # G+3612  U+9A2B  uni9A2B,  CJK UNIFIED IDEOGRAPH-9A2B
        '39472' => '4373'             , # G+4373  U+9A30  uni9A30,  CJK UNIFIED IDEOGRAPH-9A30
        '39478' => '7286'             , # G+7286  U+9A36  uni9A36,  CJK UNIFIED IDEOGRAPH-9A36
        '39479' => '5536'             , # G+5536  U+9A37  uni9A37,  CJK UNIFIED IDEOGRAPH-9A37
        '39488' => '4725'             , # G+4725  U+9A40  uni9A40,  CJK UNIFIED IDEOGRAPH-9A40
        '39491' => '7589'             , # G+7589  U+9A43  uni9A43,  CJK UNIFIED IDEOGRAPH-9A43
        '39493' => '3943'             , # G+3943  U+9A45  uni9A45,  CJK UNIFIED IDEOGRAPH-9A45
        '39501' => '7965'             , # G+7965  U+9A4D  uni9A4D,  CJK UNIFIED IDEOGRAPH-9A4D
        '39509' => '3891'             , # G+3891  U+9A55  uni9A55,  CJK UNIFIED IDEOGRAPH-9A55
        '39511' => '7729'             , # G+7729  U+9A57  uni9A57,  CJK UNIFIED IDEOGRAPH-9A57
        '39514' => '3703'             , # G+3703  U+9A5A  uni9A5A,  CJK UNIFIED IDEOGRAPH-9A5A
        '39515' => '5955'             , # G+5955  U+9A5B  uni9A5B,  CJK UNIFIED IDEOGRAPH-9A5B
        '39519' => '7328'             , # G+7328  U+9A5F  uni9A5F,  CJK UNIFIED IDEOGRAPH-9A5F
        '39522' => '4454'             , # G+4454  U+9A62  uni9A62,  CJK UNIFIED IDEOGRAPH-9A62
        '39525' => '4124'             , # G+4124  U+9A65  uni9A65,  CJK UNIFIED IDEOGRAPH-9A65
        '39529' => '7897'             , # G+7897  U+9A69  uni9A69,  CJK UNIFIED IDEOGRAPH-9A69
        '39530' => '4455'             , # G+4455  U+9A6A  uni9A6A,  CJK UNIFIED IDEOGRAPH-9A6A
        '39592' => '3785'             , # G+3785  U+9AA8  uni9AA8,  CJK UNIFIED IDEOGRAPH-9AA8
        '39608' => '7702'             , # G+7702  U+9AB8  uni9AB8,  CJK UNIFIED IDEOGRAPH-9AB8
        '39635' => '5626'             , # G+5626  U+9AD3  uni9AD3,  CJK UNIFIED IDEOGRAPH-9AD3
        '39636' => '7213'             , # G+7213  U+9AD4  uni9AD4,  CJK UNIFIED IDEOGRAPH-9AD4
        '39640' => '3765'             , # G+3765  U+9AD8  uni9AD8,  CJK UNIFIED IDEOGRAPH-9AD8
        '39653' => '5997'             , # G+5997  U+9AE5  uni9AE5,  CJK UNIFIED IDEOGRAPH-9AE5
        '39662' => '4939'             , # G+4939  U+9AEE  uni9AEE,  CJK UNIFIED IDEOGRAPH-9AEE
        '39706' => '5627'             , # G+5627  U+9B1A  uni9B1A,  CJK UNIFIED IDEOGRAPH-9B1A
        '39719' => '4562'             , # G+4562  U+9B27  uni9B27,  CJK UNIFIED IDEOGRAPH-9B27
        '39722' => '7470'             , # G+7470  U+9B2A  uni9B2A,  CJK UNIFIED IDEOGRAPH-9B2A
        '39729' => '6251'             , # G+6251  U+9B31  uni9B31,  CJK UNIFIED IDEOGRAPH-9B31
        '39740' => '3993'             , # G+3993  U+9B3C  uni9B3C,  CJK UNIFIED IDEOGRAPH-9B3C
        '39745' => '3863'             , # G+3863  U+9B41  uni9B41,  CJK UNIFIED IDEOGRAPH-9B41
        '39746' => '7849'             , # G+7849  U+9B42  uni9B42,  CJK UNIFIED IDEOGRAPH-9B42
        '39747' => '4940'             , # G+4940  U+9B43  uni9B43,  CJK UNIFIED IDEOGRAPH-9B43
        '39748' => '4995'             , # G+4995  U+9B44  uni9B44,  CJK UNIFIED IDEOGRAPH-9B44
        '39749' => '4721'             , # G+4721  U+9B45  uni9B45,  CJK UNIFIED IDEOGRAPH-9B45
        '39759' => '6308'             , # G+6308  U+9B4F  uni9B4F,  CJK UNIFIED IDEOGRAPH-9B4F
        '39764' => '4662'             , # G+4662  U+9B54  uni9B54,  CJK UNIFIED IDEOGRAPH-9B54
        '39770' => '5910'             , # G+5910  U+9B5A  uni9B5A,  CJK UNIFIED IDEOGRAPH-9B5A
        '39791' => '4525'             , # G+4525  U+9B6F  uni9B6F,  CJK UNIFIED IDEOGRAPH-9B6F
        '39822' => '6699'             , # G+6699  U+9B8E  uni9B8E,  CJK UNIFIED IDEOGRAPH-9B8E
        '39825' => '7570'             , # G+7570  U+9B91  uni9B91,  CJK UNIFIED IDEOGRAPH-9B91
        '39839' => '5816'             , # G+5816  U+9B9F  uni9B9F,  CJK UNIFIED IDEOGRAPH-9B9F
        '39851' => '3892'             , # G+3892  U+9BAB  uni9BAB,  CJK UNIFIED IDEOGRAPH-9BAB
        '39854' => '5449'             , # G+5449  U+9BAE  uni9BAE,  CJK UNIFIED IDEOGRAPH-9BAE
        '39881' => '4637'             , # G+4637  U+9BC9  uni9BC9,  CJK UNIFIED IDEOGRAPH-9BC9
        '39894' => '7204'             , # G+7204  U+9BD6  uni9BD6,  CJK UNIFIED IDEOGRAPH-9BD6
        '39908' => '3783'             , # G+3783  U+9BE4  uni9BE4,  CJK UNIFIED IDEOGRAPH-9BE4
        '39912' => '3704'             , # G+3704  U+9BE8  uni9BE8,  CJK UNIFIED IDEOGRAPH-9BE8
        '39949' => '7287'             , # G+7287  U+9C0D  uni9C0D,  CJK UNIFIED IDEOGRAPH-9C0D
        '39952' => '5805'             , # G+5805  U+9C10  uni9C10,  CJK UNIFIED IDEOGRAPH-9C10
        '39954' => '5090'             , # G+5090  U+9C12  uni9C12,  CJK UNIFIED IDEOGRAPH-9C12
        '39957' => '7629'             , # G+7629  U+9C15  uni9C15,  CJK UNIFIED IDEOGRAPH-9C15
        '39973' => '7898'             , # G+7898  U+9C25  uni9C25,  CJK UNIFIED IDEOGRAPH-9C25
        '39986' => '6076'             , # G+6076  U+9C32  uni9C32,  CJK UNIFIED IDEOGRAPH-9C32
        '39995' => '4688'             , # G+4688  U+9C3B  uni9C3B,  CJK UNIFIED IDEOGRAPH-9C3B
        '40007' => '3559'             , # G+3559  U+9C47  uni9C47,  CJK UNIFIED IDEOGRAPH-9C47
        '40009' => '5039'             , # G+5039  U+9C49  uni9C49,  CJK UNIFIED IDEOGRAPH-9C49
        '40023' => '4645'             , # G+4645  U+9C57  uni9C57,  CJK UNIFIED IDEOGRAPH-9C57
        '40165' => '6827'             , # G+6827  U+9CE5  uni9CE5,  CJK UNIFIED IDEOGRAPH-9CE5
        '40167' => '5150'             , # G+5150  U+9CE7  uni9CE7,  CJK UNIFIED IDEOGRAPH-9CE7
        '40169' => '3944'             , # G+3944  U+9CE9  uni9CE9,  CJK UNIFIED IDEOGRAPH-9CE9
        '40179' => '5108'             , # G+5108  U+9CF3  uni9CF3,  CJK UNIFIED IDEOGRAPH-9CF3
        '40180' => '4762'             , # G+4762  U+9CF4  uni9CF4,  CJK UNIFIED IDEOGRAPH-9CF4
        '40182' => '5984'             , # G+5984  U+9CF6  uni9CF6,  CJK UNIFIED IDEOGRAPH-9CF6
        '40201' => '5791'             , # G+5791  U+9D09  uni9D09,  CJK UNIFIED IDEOGRAPH-9D09
        '40219' => '6280'             , # G+6280  U+9D1B  uni9D1B,  CJK UNIFIED IDEOGRAPH-9D1B
        '40230' => '5839'             , # G+5839  U+9D26  uni9D26,  CJK UNIFIED IDEOGRAPH-9D26
        '40232' => '5832'             , # G+5832  U+9D28  uni9D28,  CJK UNIFIED IDEOGRAPH-9D28
        '40251' => '7862'             , # G+7862  U+9D3B  uni9D3B,  CJK UNIFIED IDEOGRAPH-9D3B
        '40273' => '3647'             , # G+3647  U+9D51  uni9D51,  CJK UNIFIED IDEOGRAPH-9D51
        '40285' => '5792'             , # G+5792  U+9D5D  uni9D5D,  CJK UNIFIED IDEOGRAPH-9D5D
        '40288' => '3773'             , # G+3773  U+9D60  uni9D60,  CJK UNIFIED IDEOGRAPH-9D60
        '40289' => '4833'             , # G+4833  U+9D61  uni9D61,  CJK UNIFIED IDEOGRAPH-9D61
        '40300' => '5180'             , # G+5180  U+9D6C  uni9D6C,  CJK UNIFIED IDEOGRAPH-9D6C
        '40306' => '6521'             , # G+6521  U+9D72  uni9D72,  CJK UNIFIED IDEOGRAPH-9D72
        '40361' => '4794'             , # G+4794  U+9DA9  uni9DA9,  CJK UNIFIED IDEOGRAPH-9DA9
        '40367' => '5860'             , # G+5860  U+9DAF  uni9DAF,  CJK UNIFIED IDEOGRAPH-9DAF
        '40372' => '7634'             , # G+7634  U+9DB4  uni9DB4,  CJK UNIFIED IDEOGRAPH-9DB4
        '40388' => '3728'             , # G+3728  U+9DC4  uni9DC4,  CJK UNIFIED IDEOGRAPH-9DC4
        '40407' => '3945'             , # G+3945  U+9DD7  uni9DD7,  CJK UNIFIED IDEOGRAPH-9DD7
        '40434' => '7329'             , # G+7329  U+9DF2  uni9DF2,  CJK UNIFIED IDEOGRAPH-9DF2
        '40440' => '8012'             , # G+8012  U+9DF8  uni9DF8,  CJK UNIFIED IDEOGRAPH-9DF8
        '40441' => '6393'             , # G+6393  U+9DF9  uni9DF9,  CJK UNIFIED IDEOGRAPH-9DF9
        '40442' => '4526'             , # G+4526  U+9DFA  uni9DFA,  CJK UNIFIED IDEOGRAPH-9DFA
        '40474' => '5861'             , # G+5861  U+9E1A  uni9E1A,  CJK UNIFIED IDEOGRAPH-9E1A
        '40478' => '4396'             , # G+4396  U+9E1E  uni9E1E,  CJK UNIFIED IDEOGRAPH-9E1E
        '40565' => '4527'             , # G+4527  U+9E75  uni9E75,  CJK UNIFIED IDEOGRAPH-9E75
        '40569' => '7662'             , # G+7662  U+9E79  uni9E79,  CJK UNIFIED IDEOGRAPH-9E79
        '40573' => '5998'             , # G+5998  U+9E7D  uni9E7D,  CJK UNIFIED IDEOGRAPH-9E7D
        '40575' => '4533'             , # G+4533  U+9E7F  uni9E7F,  CJK UNIFIED IDEOGRAPH-9E7F
        '40594' => '4125'             , # G+4125  U+9E92  uni9E92,  CJK UNIFIED IDEOGRAPH-9E92
        '40595' => '4534'             , # G+4534  U+9E93  uni9E93,  CJK UNIFIED IDEOGRAPH-9E93
        '40599' => '4456'             , # G+4456  U+9E97  uni9E97,  CJK UNIFIED IDEOGRAPH-9E97
        '40605' => '5300'             , # G+5300  U+9E9D  uni9E9D,  CJK UNIFIED IDEOGRAPH-9E9D
        '40607' => '4646'             , # G+4646  U+9E9F  uni9E9F,  CJK UNIFIED IDEOGRAPH-9E9F
        '40613' => '4726'             , # G+4726  U+9EA5  uni9EA5,  CJK UNIFIED IDEOGRAPH-9EA5
        '40628' => '3952'             , # G+3952  U+9EB4  uni9EB4,  CJK UNIFIED IDEOGRAPH-9EB4
        '40629' => '4745'             , # G+4745  U+9EB5  uni9EB5,  CJK UNIFIED IDEOGRAPH-9EB5
        '40635' => '4663'             , # G+4663  U+9EBB  uni9EBB,  CJK UNIFIED IDEOGRAPH-9EBB
        '40638' => '8004'             , # G+8004  U+9EBE  uni9EBE,  CJK UNIFIED IDEOGRAPH-9EBE
        '40643' => '7927'             , # G+7927  U+9EC3  uni9EC3,  CJK UNIFIED IDEOGRAPH-9EC3
        '40653' => '5401'             , # G+5401  U+9ECD  uni9ECD,  CJK UNIFIED IDEOGRAPH-9ECD
        '40654' => '4457'             , # G+4457  U+9ECE  uni9ECE,  CJK UNIFIED IDEOGRAPH-9ECE
        '40657' => '8018'             , # G+8018  U+9ED1  uni9ED1,  CJK UNIFIED IDEOGRAPH-9ED1
        '40660' => '3623'             , # G+3623  U+9ED4  uni9ED4,  CJK UNIFIED IDEOGRAPH-9ED4
        '40664' => '4835'             , # G+4835  U+9ED8  uni9ED8,  CJK UNIFIED IDEOGRAPH-9ED8
        '40667' => '4265'             , # G+4265  U+9EDB  uni9EDB,  CJK UNIFIED IDEOGRAPH-9EDB
        '40668' => '7305'             , # G+7305  U+9EDC  uni9EDC,  CJK UNIFIED IDEOGRAPH-9EDC
        '40670' => '6700'             , # G+6700  U+9EDE  uni9EDE,  CJK UNIFIED IDEOGRAPH-9EDE
        '40680' => '4249'             , # G+4249  U+9EE8  uni9EE8,  CJK UNIFIED IDEOGRAPH-9EE8
        '40692' => '4869'             , # G+4869  U+9EF4  uni9EF4,  CJK UNIFIED IDEOGRAPH-9EF4
        '40711' => '6077'             , # G+6077  U+9F07  uni9F07,  CJK UNIFIED IDEOGRAPH-9F07
        '40712' => '5040'             , # G+5040  U+9F08  uni9F08,  CJK UNIFIED IDEOGRAPH-9F08
        '40718' => '6758'             , # G+6758  U+9F0E  uni9F0E,  CJK UNIFIED IDEOGRAPH-9F0E
        '40723' => '3766'             , # G+3766  U+9F13  uni9F13,  CJK UNIFIED IDEOGRAPH-9F13
        '40736' => '5402'             , # G+5402  U+9F20  uni9F20,  CJK UNIFIED IDEOGRAPH-9F20
        '40763' => '5222'             , # G+5222  U+9F3B  uni9F3B,  CJK UNIFIED IDEOGRAPH-9F3B
        '40778' => '6781'             , # G+6781  U+9F4A  uni9F4A,  CJK UNIFIED IDEOGRAPH-9F4A
        '40779' => '6585'             , # G+6585  U+9F4B  uni9F4B,  CJK UNIFIED IDEOGRAPH-9F4B
        '40782' => '6586'             , # G+6586  U+9F4E  uni9F4E,  CJK UNIFIED IDEOGRAPH-9F4E
        '40786' => '7359'             , # G+7359  U+9F52  uni9F52,  CJK UNIFIED IDEOGRAPH-9F52
        '40799' => '6618'             , # G+6618  U+9F5F  uni9F5F,  CJK UNIFIED IDEOGRAPH-9F5F
        '40801' => '4505'             , # G+4505  U+9F61  uni9F61,  CJK UNIFIED IDEOGRAPH-9F61
        '40807' => '5462'             , # G+5462  U+9F67  uni9F67,  CJK UNIFIED IDEOGRAPH-9F67
        '40810' => '7061'             , # G+7061  U+9F6A  uni9F6A,  CJK UNIFIED IDEOGRAPH-9F6A
        '40812' => '5911'             , # G+5911  U+9F6C  uni9F6C,  CJK UNIFIED IDEOGRAPH-9F6C
        '40823' => '5806'             , # G+5806  U+9F77  uni9F77,  CJK UNIFIED IDEOGRAPH-9F77
        '40845' => '4563'             , # G+4563  U+9F8D  uni9F8D,  CJK UNIFIED IDEOGRAPH-9F8D
        '40848' => '4968'             , # G+4968  U+9F90  uni9F90,  CJK UNIFIED IDEOGRAPH-9F90
        '40853' => '3529'             , # G+3529  U+9F95  uni9F95,  CJK UNIFIED IDEOGRAPH-9F95
        '40860' => '3946'             , # G+3946  U+9F9C  uni9F9C,  CJK UNIFIED IDEOGRAPH-9F9C
        '44032' => '1086'             , # G+1086  U+AC00  uniAC00,  HANGUL SYLLABLE-AC00
        '44033' => '1087'             , # G+1087  U+AC01  uniAC01,  HANGUL SYLLABLE-AC01
        '44034' => '9333'             , # G+9333  U+AC02  uniAC02,  HANGUL SYLLABLE-AC02
        '44035' => '9334'             , # G+9334  U+AC03  uniAC03,  HANGUL SYLLABLE-AC03
        '44036' => '1088'             , # G+1088  U+AC04  uniAC04,  HANGUL SYLLABLE-AC04
        '44037' => '9335'             , # G+9335  U+AC05  uniAC05,  HANGUL SYLLABLE-AC05
        '44038' => '9336'             , # G+9336  U+AC06  uniAC06,  HANGUL SYLLABLE-AC06
        '44039' => '1089'             , # G+1089  U+AC07  uniAC07,  HANGUL SYLLABLE-AC07
        '44040' => '1090'             , # G+1090  U+AC08  uniAC08,  HANGUL SYLLABLE-AC08
        '44041' => '1091'             , # G+1091  U+AC09  uniAC09,  HANGUL SYLLABLE-AC09
        '44042' => '1092'             , # G+1092  U+AC0A  uniAC0A,  HANGUL SYLLABLE-AC0A
        '44043' => '9337'             , # G+9337  U+AC0B  uniAC0B,  HANGUL SYLLABLE-AC0B
        '44044' => '9338'             , # G+9338  U+AC0C  uniAC0C,  HANGUL SYLLABLE-AC0C
        '44045' => '9339'             , # G+9339  U+AC0D  uniAC0D,  HANGUL SYLLABLE-AC0D
        '44046' => '9340'             , # G+9340  U+AC0E  uniAC0E,  HANGUL SYLLABLE-AC0E
        '44047' => '9341'             , # G+9341  U+AC0F  uniAC0F,  HANGUL SYLLABLE-AC0F
        '44048' => '1093'             , # G+1093  U+AC10  uniAC10,  HANGUL SYLLABLE-AC10
        '44049' => '1094'             , # G+1094  U+AC11  uniAC11,  HANGUL SYLLABLE-AC11
        '44050' => '1095'             , # G+1095  U+AC12  uniAC12,  HANGUL SYLLABLE-AC12
        '44051' => '1096'             , # G+1096  U+AC13  uniAC13,  HANGUL SYLLABLE-AC13
        '44052' => '1097'             , # G+1097  U+AC14  uniAC14,  HANGUL SYLLABLE-AC14
        '44053' => '1098'             , # G+1098  U+AC15  uniAC15,  HANGUL SYLLABLE-AC15
        '44054' => '1099'             , # G+1099  U+AC16  uniAC16,  HANGUL SYLLABLE-AC16
        '44055' => '1100'             , # G+1100  U+AC17  uniAC17,  HANGUL SYLLABLE-AC17
        '44056' => '9342'             , # G+9342  U+AC18  uniAC18,  HANGUL SYLLABLE-AC18
        '44057' => '1101'             , # G+1101  U+AC19  uniAC19,  HANGUL SYLLABLE-AC19
        '44058' => '1102'             , # G+1102  U+AC1A  uniAC1A,  HANGUL SYLLABLE-AC1A
        '44059' => '1103'             , # G+1103  U+AC1B  uniAC1B,  HANGUL SYLLABLE-AC1B
        '44060' => '1104'             , # G+1104  U+AC1C  uniAC1C,  HANGUL SYLLABLE-AC1C
        '44061' => '1105'             , # G+1105  U+AC1D  uniAC1D,  HANGUL SYLLABLE-AC1D
        '44062' => '9343'             , # G+9343  U+AC1E  uniAC1E,  HANGUL SYLLABLE-AC1E
        '44063' => '9344'             , # G+9344  U+AC1F  uniAC1F,  HANGUL SYLLABLE-AC1F
        '44064' => '1106'             , # G+1106  U+AC20  uniAC20,  HANGUL SYLLABLE-AC20
        '44065' => '9345'             , # G+9345  U+AC21  uniAC21,  HANGUL SYLLABLE-AC21
        '44066' => '9346'             , # G+9346  U+AC22  uniAC22,  HANGUL SYLLABLE-AC22
        '44067' => '9347'             , # G+9347  U+AC23  uniAC23,  HANGUL SYLLABLE-AC23
        '44068' => '1107'             , # G+1107  U+AC24  uniAC24,  HANGUL SYLLABLE-AC24
        '44069' => '9348'             , # G+9348  U+AC25  uniAC25,  HANGUL SYLLABLE-AC25
        '44070' => '9349'             , # G+9349  U+AC26  uniAC26,  HANGUL SYLLABLE-AC26
        '44071' => '9350'             , # G+9350  U+AC27  uniAC27,  HANGUL SYLLABLE-AC27
        '44072' => '9351'             , # G+9351  U+AC28  uniAC28,  HANGUL SYLLABLE-AC28
        '44073' => '9352'             , # G+9352  U+AC29  uniAC29,  HANGUL SYLLABLE-AC29
        '44074' => '9353'             , # G+9353  U+AC2A  uniAC2A,  HANGUL SYLLABLE-AC2A
        '44075' => '9354'             , # G+9354  U+AC2B  uniAC2B,  HANGUL SYLLABLE-AC2B
        '44076' => '1108'             , # G+1108  U+AC2C  uniAC2C,  HANGUL SYLLABLE-AC2C
        '44077' => '1109'             , # G+1109  U+AC2D  uniAC2D,  HANGUL SYLLABLE-AC2D
        '44078' => '9355'             , # G+9355  U+AC2E  uniAC2E,  HANGUL SYLLABLE-AC2E
        '44079' => '1110'             , # G+1110  U+AC2F  uniAC2F,  HANGUL SYLLABLE-AC2F
        '44080' => '1111'             , # G+1111  U+AC30  uniAC30,  HANGUL SYLLABLE-AC30
        '44081' => '1112'             , # G+1112  U+AC31  uniAC31,  HANGUL SYLLABLE-AC31
        '44082' => '9356'             , # G+9356  U+AC32  uniAC32,  HANGUL SYLLABLE-AC32
        '44083' => '9357'             , # G+9357  U+AC33  uniAC33,  HANGUL SYLLABLE-AC33
        '44084' => '9358'             , # G+9358  U+AC34  uniAC34,  HANGUL SYLLABLE-AC34
        '44085' => '9359'             , # G+9359  U+AC35  uniAC35,  HANGUL SYLLABLE-AC35
        '44086' => '9360'             , # G+9360  U+AC36  uniAC36,  HANGUL SYLLABLE-AC36
        '44087' => '9361'             , # G+9361  U+AC37  uniAC37,  HANGUL SYLLABLE-AC37
        '44088' => '1113'             , # G+1113  U+AC38  uniAC38,  HANGUL SYLLABLE-AC38
        '44089' => '1114'             , # G+1114  U+AC39  uniAC39,  HANGUL SYLLABLE-AC39
        '44090' => '9362'             , # G+9362  U+AC3A  uniAC3A,  HANGUL SYLLABLE-AC3A
        '44091' => '9363'             , # G+9363  U+AC3B  uniAC3B,  HANGUL SYLLABLE-AC3B
        '44092' => '1115'             , # G+1115  U+AC3C  uniAC3C,  HANGUL SYLLABLE-AC3C
        '44093' => '9364'             , # G+9364  U+AC3D  uniAC3D,  HANGUL SYLLABLE-AC3D
        '44094' => '9365'             , # G+9365  U+AC3E  uniAC3E,  HANGUL SYLLABLE-AC3E
        '44095' => '9366'             , # G+9366  U+AC3F  uniAC3F,  HANGUL SYLLABLE-AC3F
        '44096' => '1116'             , # G+1116  U+AC40  uniAC40,  HANGUL SYLLABLE-AC40
        '44097' => '9367'             , # G+9367  U+AC41  uniAC41,  HANGUL SYLLABLE-AC41
        '44098' => '9368'             , # G+9368  U+AC42  uniAC42,  HANGUL SYLLABLE-AC42
        '44099' => '9369'             , # G+9369  U+AC43  uniAC43,  HANGUL SYLLABLE-AC43
        '44100' => '9370'             , # G+9370  U+AC44  uniAC44,  HANGUL SYLLABLE-AC44
        '44101' => '9371'             , # G+9371  U+AC45  uniAC45,  HANGUL SYLLABLE-AC45
        '44102' => '9372'             , # G+9372  U+AC46  uniAC46,  HANGUL SYLLABLE-AC46
        '44103' => '9373'             , # G+9373  U+AC47  uniAC47,  HANGUL SYLLABLE-AC47
        '44104' => '9374'             , # G+9374  U+AC48  uniAC48,  HANGUL SYLLABLE-AC48
        '44105' => '9375'             , # G+9375  U+AC49  uniAC49,  HANGUL SYLLABLE-AC49
        '44106' => '9376'             , # G+9376  U+AC4A  uniAC4A,  HANGUL SYLLABLE-AC4A
        '44107' => '1117'             , # G+1117  U+AC4B  uniAC4B,  HANGUL SYLLABLE-AC4B
        '44108' => '9377'             , # G+9377  U+AC4C  uniAC4C,  HANGUL SYLLABLE-AC4C
        '44109' => '1118'             , # G+1118  U+AC4D  uniAC4D,  HANGUL SYLLABLE-AC4D
        '44110' => '9378'             , # G+9378  U+AC4E  uniAC4E,  HANGUL SYLLABLE-AC4E
        '44111' => '9379'             , # G+9379  U+AC4F  uniAC4F,  HANGUL SYLLABLE-AC4F
        '44112' => '9380'             , # G+9380  U+AC50  uniAC50,  HANGUL SYLLABLE-AC50
        '44113' => '9381'             , # G+9381  U+AC51  uniAC51,  HANGUL SYLLABLE-AC51
        '44114' => '9382'             , # G+9382  U+AC52  uniAC52,  HANGUL SYLLABLE-AC52
        '44115' => '9383'             , # G+9383  U+AC53  uniAC53,  HANGUL SYLLABLE-AC53
        '44116' => '1119'             , # G+1119  U+AC54  uniAC54,  HANGUL SYLLABLE-AC54
        '44117' => '9384'             , # G+9384  U+AC55  uniAC55,  HANGUL SYLLABLE-AC55
        '44118' => '9385'             , # G+9385  U+AC56  uniAC56,  HANGUL SYLLABLE-AC56
        '44119' => '9386'             , # G+9386  U+AC57  uniAC57,  HANGUL SYLLABLE-AC57
        '44120' => '1120'             , # G+1120  U+AC58  uniAC58,  HANGUL SYLLABLE-AC58
        '44121' => '9387'             , # G+9387  U+AC59  uniAC59,  HANGUL SYLLABLE-AC59
        '44122' => '9388'             , # G+9388  U+AC5A  uniAC5A,  HANGUL SYLLABLE-AC5A
        '44123' => '9389'             , # G+9389  U+AC5B  uniAC5B,  HANGUL SYLLABLE-AC5B
        '44124' => '1121'             , # G+1121  U+AC5C  uniAC5C,  HANGUL SYLLABLE-AC5C
        '44125' => '9390'             , # G+9390  U+AC5D  uniAC5D,  HANGUL SYLLABLE-AC5D
        '44126' => '9391'             , # G+9391  U+AC5E  uniAC5E,  HANGUL SYLLABLE-AC5E
        '44127' => '9392'             , # G+9392  U+AC5F  uniAC5F,  HANGUL SYLLABLE-AC5F
        '44128' => '9393'             , # G+9393  U+AC60  uniAC60,  HANGUL SYLLABLE-AC60
        '44129' => '9394'             , # G+9394  U+AC61  uniAC61,  HANGUL SYLLABLE-AC61
        '44130' => '9395'             , # G+9395  U+AC62  uniAC62,  HANGUL SYLLABLE-AC62
        '44131' => '9396'             , # G+9396  U+AC63  uniAC63,  HANGUL SYLLABLE-AC63
        '44132' => '9397'             , # G+9397  U+AC64  uniAC64,  HANGUL SYLLABLE-AC64
        '44133' => '9398'             , # G+9398  U+AC65  uniAC65,  HANGUL SYLLABLE-AC65
        '44134' => '9399'             , # G+9399  U+AC66  uniAC66,  HANGUL SYLLABLE-AC66
        '44135' => '9400'             , # G+9400  U+AC67  uniAC67,  HANGUL SYLLABLE-AC67
        '44136' => '9401'             , # G+9401  U+AC68  uniAC68,  HANGUL SYLLABLE-AC68
        '44137' => '9402'             , # G+9402  U+AC69  uniAC69,  HANGUL SYLLABLE-AC69
        '44138' => '9403'             , # G+9403  U+AC6A  uniAC6A,  HANGUL SYLLABLE-AC6A
        '44139' => '9404'             , # G+9404  U+AC6B  uniAC6B,  HANGUL SYLLABLE-AC6B
        '44140' => '9405'             , # G+9405  U+AC6C  uniAC6C,  HANGUL SYLLABLE-AC6C
        '44141' => '9406'             , # G+9406  U+AC6D  uniAC6D,  HANGUL SYLLABLE-AC6D
        '44142' => '9407'             , # G+9407  U+AC6E  uniAC6E,  HANGUL SYLLABLE-AC6E
        '44143' => '9408'             , # G+9408  U+AC6F  uniAC6F,  HANGUL SYLLABLE-AC6F
        '44144' => '1122'             , # G+1122  U+AC70  uniAC70,  HANGUL SYLLABLE-AC70
        '44145' => '1123'             , # G+1123  U+AC71  uniAC71,  HANGUL SYLLABLE-AC71
        '44146' => '9409'             , # G+9409  U+AC72  uniAC72,  HANGUL SYLLABLE-AC72
        '44147' => '9410'             , # G+9410  U+AC73  uniAC73,  HANGUL SYLLABLE-AC73
        '44148' => '1124'             , # G+1124  U+AC74  uniAC74,  HANGUL SYLLABLE-AC74
        '44149' => '9411'             , # G+9411  U+AC75  uniAC75,  HANGUL SYLLABLE-AC75
        '44150' => '9412'             , # G+9412  U+AC76  uniAC76,  HANGUL SYLLABLE-AC76
        '44151' => '1125'             , # G+1125  U+AC77  uniAC77,  HANGUL SYLLABLE-AC77
        '44152' => '1126'             , # G+1126  U+AC78  uniAC78,  HANGUL SYLLABLE-AC78
        '44153' => '9413'             , # G+9413  U+AC79  uniAC79,  HANGUL SYLLABLE-AC79
        '44154' => '1127'             , # G+1127  U+AC7A  uniAC7A,  HANGUL SYLLABLE-AC7A
        '44155' => '9414'             , # G+9414  U+AC7B  uniAC7B,  HANGUL SYLLABLE-AC7B
        '44156' => '9415'             , # G+9415  U+AC7C  uniAC7C,  HANGUL SYLLABLE-AC7C
        '44157' => '9416'             , # G+9416  U+AC7D  uniAC7D,  HANGUL SYLLABLE-AC7D
        '44158' => '9417'             , # G+9417  U+AC7E  uniAC7E,  HANGUL SYLLABLE-AC7E
        '44159' => '9418'             , # G+9418  U+AC7F  uniAC7F,  HANGUL SYLLABLE-AC7F
        '44160' => '1128'             , # G+1128  U+AC80  uniAC80,  HANGUL SYLLABLE-AC80
        '44161' => '1129'             , # G+1129  U+AC81  uniAC81,  HANGUL SYLLABLE-AC81
        '44162' => '9419'             , # G+9419  U+AC82  uniAC82,  HANGUL SYLLABLE-AC82
        '44163' => '1130'             , # G+1130  U+AC83  uniAC83,  HANGUL SYLLABLE-AC83
        '44164' => '1131'             , # G+1131  U+AC84  uniAC84,  HANGUL SYLLABLE-AC84
        '44165' => '1132'             , # G+1132  U+AC85  uniAC85,  HANGUL SYLLABLE-AC85
        '44166' => '1133'             , # G+1133  U+AC86  uniAC86,  HANGUL SYLLABLE-AC86
        '44167' => '9420'             , # G+9420  U+AC87  uniAC87,  HANGUL SYLLABLE-AC87
        '44168' => '9421'             , # G+9421  U+AC88  uniAC88,  HANGUL SYLLABLE-AC88
        '44169' => '1134'             , # G+1134  U+AC89  uniAC89,  HANGUL SYLLABLE-AC89
        '44170' => '1135'             , # G+1135  U+AC8A  uniAC8A,  HANGUL SYLLABLE-AC8A
        '44171' => '1136'             , # G+1136  U+AC8B  uniAC8B,  HANGUL SYLLABLE-AC8B
        '44172' => '1137'             , # G+1137  U+AC8C  uniAC8C,  HANGUL SYLLABLE-AC8C
        '44173' => '9422'             , # G+9422  U+AC8D  uniAC8D,  HANGUL SYLLABLE-AC8D
        '44174' => '9423'             , # G+9423  U+AC8E  uniAC8E,  HANGUL SYLLABLE-AC8E
        '44175' => '9424'             , # G+9424  U+AC8F  uniAC8F,  HANGUL SYLLABLE-AC8F
        '44176' => '1138'             , # G+1138  U+AC90  uniAC90,  HANGUL SYLLABLE-AC90
        '44177' => '9425'             , # G+9425  U+AC91  uniAC91,  HANGUL SYLLABLE-AC91
        '44178' => '9426'             , # G+9426  U+AC92  uniAC92,  HANGUL SYLLABLE-AC92
        '44179' => '9427'             , # G+9427  U+AC93  uniAC93,  HANGUL SYLLABLE-AC93
        '44180' => '1139'             , # G+1139  U+AC94  uniAC94,  HANGUL SYLLABLE-AC94
        '44181' => '9428'             , # G+9428  U+AC95  uniAC95,  HANGUL SYLLABLE-AC95
        '44182' => '9429'             , # G+9429  U+AC96  uniAC96,  HANGUL SYLLABLE-AC96
        '44183' => '9430'             , # G+9430  U+AC97  uniAC97,  HANGUL SYLLABLE-AC97
        '44184' => '9431'             , # G+9431  U+AC98  uniAC98,  HANGUL SYLLABLE-AC98
        '44185' => '9432'             , # G+9432  U+AC99  uniAC99,  HANGUL SYLLABLE-AC99
        '44186' => '9433'             , # G+9433  U+AC9A  uniAC9A,  HANGUL SYLLABLE-AC9A
        '44187' => '9434'             , # G+9434  U+AC9B  uniAC9B,  HANGUL SYLLABLE-AC9B
        '44188' => '1140'             , # G+1140  U+AC9C  uniAC9C,  HANGUL SYLLABLE-AC9C
        '44189' => '1141'             , # G+1141  U+AC9D  uniAC9D,  HANGUL SYLLABLE-AC9D
        '44190' => '9435'             , # G+9435  U+AC9E  uniAC9E,  HANGUL SYLLABLE-AC9E
        '44191' => '1142'             , # G+1142  U+AC9F  uniAC9F,  HANGUL SYLLABLE-AC9F
        '44192' => '1143'             , # G+1143  U+ACA0  uniACA0,  HANGUL SYLLABLE-ACA0
        '44193' => '1144'             , # G+1144  U+ACA1  uniACA1,  HANGUL SYLLABLE-ACA1
        '44194' => '9436'             , # G+9436  U+ACA2  uniACA2,  HANGUL SYLLABLE-ACA2
        '44195' => '9437'             , # G+9437  U+ACA3  uniACA3,  HANGUL SYLLABLE-ACA3
        '44196' => '9438'             , # G+9438  U+ACA4  uniACA4,  HANGUL SYLLABLE-ACA4
        '44197' => '9439'             , # G+9439  U+ACA5  uniACA5,  HANGUL SYLLABLE-ACA5
        '44198' => '9440'             , # G+9440  U+ACA6  uniACA6,  HANGUL SYLLABLE-ACA6
        '44199' => '9441'             , # G+9441  U+ACA7  uniACA7,  HANGUL SYLLABLE-ACA7
        '44200' => '1145'             , # G+1145  U+ACA8  uniACA8,  HANGUL SYLLABLE-ACA8
        '44201' => '1146'             , # G+1146  U+ACA9  uniACA9,  HANGUL SYLLABLE-ACA9
        '44202' => '1147'             , # G+1147  U+ACAA  uniACAA,  HANGUL SYLLABLE-ACAA
        '44203' => '9442'             , # G+9442  U+ACAB  uniACAB,  HANGUL SYLLABLE-ACAB
        '44204' => '1148'             , # G+1148  U+ACAC  uniACAC,  HANGUL SYLLABLE-ACAC
        '44205' => '9443'             , # G+9443  U+ACAD  uniACAD,  HANGUL SYLLABLE-ACAD
        '44206' => '9444'             , # G+9444  U+ACAE  uniACAE,  HANGUL SYLLABLE-ACAE
        '44207' => '1149'             , # G+1149  U+ACAF  uniACAF,  HANGUL SYLLABLE-ACAF
        '44208' => '1150'             , # G+1150  U+ACB0  uniACB0,  HANGUL SYLLABLE-ACB0
        '44209' => '9445'             , # G+9445  U+ACB1  uniACB1,  HANGUL SYLLABLE-ACB1
        '44210' => '9446'             , # G+9446  U+ACB2  uniACB2,  HANGUL SYLLABLE-ACB2
        '44211' => '9447'             , # G+9447  U+ACB3  uniACB3,  HANGUL SYLLABLE-ACB3
        '44212' => '9448'             , # G+9448  U+ACB4  uniACB4,  HANGUL SYLLABLE-ACB4
        '44213' => '9449'             , # G+9449  U+ACB5  uniACB5,  HANGUL SYLLABLE-ACB5
        '44214' => '9450'             , # G+9450  U+ACB6  uniACB6,  HANGUL SYLLABLE-ACB6
        '44215' => '9451'             , # G+9451  U+ACB7  uniACB7,  HANGUL SYLLABLE-ACB7
        '44216' => '1151'             , # G+1151  U+ACB8  uniACB8,  HANGUL SYLLABLE-ACB8
        '44217' => '1152'             , # G+1152  U+ACB9  uniACB9,  HANGUL SYLLABLE-ACB9
        '44218' => '9452'             , # G+9452  U+ACBA  uniACBA,  HANGUL SYLLABLE-ACBA
        '44219' => '1153'             , # G+1153  U+ACBB  uniACBB,  HANGUL SYLLABLE-ACBB
        '44220' => '1154'             , # G+1154  U+ACBC  uniACBC,  HANGUL SYLLABLE-ACBC
        '44221' => '1155'             , # G+1155  U+ACBD  uniACBD,  HANGUL SYLLABLE-ACBD
        '44222' => '9453'             , # G+9453  U+ACBE  uniACBE,  HANGUL SYLLABLE-ACBE
        '44223' => '9454'             , # G+9454  U+ACBF  uniACBF,  HANGUL SYLLABLE-ACBF
        '44224' => '9455'             , # G+9455  U+ACC0  uniACC0,  HANGUL SYLLABLE-ACC0
        '44225' => '1156'             , # G+1156  U+ACC1  uniACC1,  HANGUL SYLLABLE-ACC1
        '44226' => '9456'             , # G+9456  U+ACC2  uniACC2,  HANGUL SYLLABLE-ACC2
        '44227' => '9457'             , # G+9457  U+ACC3  uniACC3,  HANGUL SYLLABLE-ACC3
        '44228' => '1157'             , # G+1157  U+ACC4  uniACC4,  HANGUL SYLLABLE-ACC4
        '44229' => '9458'             , # G+9458  U+ACC5  uniACC5,  HANGUL SYLLABLE-ACC5
        '44230' => '9459'             , # G+9459  U+ACC6  uniACC6,  HANGUL SYLLABLE-ACC6
        '44231' => '9460'             , # G+9460  U+ACC7  uniACC7,  HANGUL SYLLABLE-ACC7
        '44232' => '1158'             , # G+1158  U+ACC8  uniACC8,  HANGUL SYLLABLE-ACC8
        '44233' => '9461'             , # G+9461  U+ACC9  uniACC9,  HANGUL SYLLABLE-ACC9
        '44234' => '9462'             , # G+9462  U+ACCA  uniACCA,  HANGUL SYLLABLE-ACCA
        '44235' => '9463'             , # G+9463  U+ACCB  uniACCB,  HANGUL SYLLABLE-ACCB
        '44236' => '1159'             , # G+1159  U+ACCC  uniACCC,  HANGUL SYLLABLE-ACCC
        '44237' => '9464'             , # G+9464  U+ACCD  uniACCD,  HANGUL SYLLABLE-ACCD
        '44238' => '9465'             , # G+9465  U+ACCE  uniACCE,  HANGUL SYLLABLE-ACCE
        '44239' => '9466'             , # G+9466  U+ACCF  uniACCF,  HANGUL SYLLABLE-ACCF
        '44240' => '9467'             , # G+9467  U+ACD0  uniACD0,  HANGUL SYLLABLE-ACD0
        '44241' => '9468'             , # G+9468  U+ACD1  uniACD1,  HANGUL SYLLABLE-ACD1
        '44242' => '9469'             , # G+9469  U+ACD2  uniACD2,  HANGUL SYLLABLE-ACD2
        '44243' => '9470'             , # G+9470  U+ACD3  uniACD3,  HANGUL SYLLABLE-ACD3
        '44244' => '9471'             , # G+9471  U+ACD4  uniACD4,  HANGUL SYLLABLE-ACD4
        '44245' => '1160'             , # G+1160  U+ACD5  uniACD5,  HANGUL SYLLABLE-ACD5
        '44246' => '9472'             , # G+9472  U+ACD6  uniACD6,  HANGUL SYLLABLE-ACD6
        '44247' => '1161'             , # G+1161  U+ACD7  uniACD7,  HANGUL SYLLABLE-ACD7
        '44248' => '9473'             , # G+9473  U+ACD8  uniACD8,  HANGUL SYLLABLE-ACD8
        '44249' => '9474'             , # G+9474  U+ACD9  uniACD9,  HANGUL SYLLABLE-ACD9
        '44250' => '9475'             , # G+9475  U+ACDA  uniACDA,  HANGUL SYLLABLE-ACDA
        '44251' => '9476'             , # G+9476  U+ACDB  uniACDB,  HANGUL SYLLABLE-ACDB
        '44252' => '9477'             , # G+9477  U+ACDC  uniACDC,  HANGUL SYLLABLE-ACDC
        '44253' => '9478'             , # G+9478  U+ACDD  uniACDD,  HANGUL SYLLABLE-ACDD
        '44254' => '9479'             , # G+9479  U+ACDE  uniACDE,  HANGUL SYLLABLE-ACDE
        '44255' => '9480'             , # G+9480  U+ACDF  uniACDF,  HANGUL SYLLABLE-ACDF
        '44256' => '1162'             , # G+1162  U+ACE0  uniACE0,  HANGUL SYLLABLE-ACE0
        '44257' => '1163'             , # G+1163  U+ACE1  uniACE1,  HANGUL SYLLABLE-ACE1
        '44258' => '9481'             , # G+9481  U+ACE2  uniACE2,  HANGUL SYLLABLE-ACE2
        '44259' => '9482'             , # G+9482  U+ACE3  uniACE3,  HANGUL SYLLABLE-ACE3
        '44260' => '1164'             , # G+1164  U+ACE4  uniACE4,  HANGUL SYLLABLE-ACE4
        '44261' => '9483'             , # G+9483  U+ACE5  uniACE5,  HANGUL SYLLABLE-ACE5
        '44262' => '9484'             , # G+9484  U+ACE6  uniACE6,  HANGUL SYLLABLE-ACE6
        '44263' => '1165'             , # G+1165  U+ACE7  uniACE7,  HANGUL SYLLABLE-ACE7
        '44264' => '1166'             , # G+1166  U+ACE8  uniACE8,  HANGUL SYLLABLE-ACE8
        '44265' => '9485'             , # G+9485  U+ACE9  uniACE9,  HANGUL SYLLABLE-ACE9
        '44266' => '1167'             , # G+1167  U+ACEA  uniACEA,  HANGUL SYLLABLE-ACEA
        '44267' => '9486'             , # G+9486  U+ACEB  uniACEB,  HANGUL SYLLABLE-ACEB
        '44268' => '1168'             , # G+1168  U+ACEC  uniACEC,  HANGUL SYLLABLE-ACEC
        '44269' => '9487'             , # G+9487  U+ACED  uniACED,  HANGUL SYLLABLE-ACED
        '44270' => '9488'             , # G+9488  U+ACEE  uniACEE,  HANGUL SYLLABLE-ACEE
        '44271' => '1169'             , # G+1169  U+ACEF  uniACEF,  HANGUL SYLLABLE-ACEF
        '44272' => '1170'             , # G+1170  U+ACF0  uniACF0,  HANGUL SYLLABLE-ACF0
        '44273' => '1171'             , # G+1171  U+ACF1  uniACF1,  HANGUL SYLLABLE-ACF1
        '44274' => '9489'             , # G+9489  U+ACF2  uniACF2,  HANGUL SYLLABLE-ACF2
        '44275' => '1172'             , # G+1172  U+ACF3  uniACF3,  HANGUL SYLLABLE-ACF3
        '44276' => '9490'             , # G+9490  U+ACF4  uniACF4,  HANGUL SYLLABLE-ACF4
        '44277' => '1173'             , # G+1173  U+ACF5  uniACF5,  HANGUL SYLLABLE-ACF5
        '44278' => '1174'             , # G+1174  U+ACF6  uniACF6,  HANGUL SYLLABLE-ACF6
        '44279' => '9491'             , # G+9491  U+ACF7  uniACF7,  HANGUL SYLLABLE-ACF7
        '44280' => '9492'             , # G+9492  U+ACF8  uniACF8,  HANGUL SYLLABLE-ACF8
        '44281' => '9493'             , # G+9493  U+ACF9  uniACF9,  HANGUL SYLLABLE-ACF9
        '44282' => '9494'             , # G+9494  U+ACFA  uniACFA,  HANGUL SYLLABLE-ACFA
        '44283' => '9495'             , # G+9495  U+ACFB  uniACFB,  HANGUL SYLLABLE-ACFB
        '44284' => '1175'             , # G+1175  U+ACFC  uniACFC,  HANGUL SYLLABLE-ACFC
        '44285' => '1176'             , # G+1176  U+ACFD  uniACFD,  HANGUL SYLLABLE-ACFD
        '44286' => '9496'             , # G+9496  U+ACFE  uniACFE,  HANGUL SYLLABLE-ACFE
        '44287' => '9497'             , # G+9497  U+ACFF  uniACFF,  HANGUL SYLLABLE-ACFF
        '44288' => '1177'             , # G+1177  U+AD00  uniAD00,  HANGUL SYLLABLE-AD00
        '44289' => '9498'             , # G+9498  U+AD01  uniAD01,  HANGUL SYLLABLE-AD01
        '44290' => '9499'             , # G+9499  U+AD02  uniAD02,  HANGUL SYLLABLE-AD02
        '44291' => '9500'             , # G+9500  U+AD03  uniAD03,  HANGUL SYLLABLE-AD03
        '44292' => '1178'             , # G+1178  U+AD04  uniAD04,  HANGUL SYLLABLE-AD04
        '44293' => '9501'             , # G+9501  U+AD05  uniAD05,  HANGUL SYLLABLE-AD05
        '44294' => '1179'             , # G+1179  U+AD06  uniAD06,  HANGUL SYLLABLE-AD06
        '44295' => '9502'             , # G+9502  U+AD07  uniAD07,  HANGUL SYLLABLE-AD07
        '44296' => '9503'             , # G+9503  U+AD08  uniAD08,  HANGUL SYLLABLE-AD08
        '44297' => '9504'             , # G+9504  U+AD09  uniAD09,  HANGUL SYLLABLE-AD09
        '44298' => '9505'             , # G+9505  U+AD0A  uniAD0A,  HANGUL SYLLABLE-AD0A
        '44299' => '9506'             , # G+9506  U+AD0B  uniAD0B,  HANGUL SYLLABLE-AD0B
        '44300' => '1180'             , # G+1180  U+AD0C  uniAD0C,  HANGUL SYLLABLE-AD0C
        '44301' => '1181'             , # G+1181  U+AD0D  uniAD0D,  HANGUL SYLLABLE-AD0D
        '44302' => '9507'             , # G+9507  U+AD0E  uniAD0E,  HANGUL SYLLABLE-AD0E
        '44303' => '1182'             , # G+1182  U+AD0F  uniAD0F,  HANGUL SYLLABLE-AD0F
        '44304' => '9508'             , # G+9508  U+AD10  uniAD10,  HANGUL SYLLABLE-AD10
        '44305' => '1183'             , # G+1183  U+AD11  uniAD11,  HANGUL SYLLABLE-AD11
        '44306' => '9509'             , # G+9509  U+AD12  uniAD12,  HANGUL SYLLABLE-AD12
        '44307' => '9510'             , # G+9510  U+AD13  uniAD13,  HANGUL SYLLABLE-AD13
        '44308' => '9511'             , # G+9511  U+AD14  uniAD14,  HANGUL SYLLABLE-AD14
        '44309' => '9512'             , # G+9512  U+AD15  uniAD15,  HANGUL SYLLABLE-AD15
        '44310' => '9513'             , # G+9513  U+AD16  uniAD16,  HANGUL SYLLABLE-AD16
        '44311' => '9514'             , # G+9514  U+AD17  uniAD17,  HANGUL SYLLABLE-AD17
        '44312' => '1184'             , # G+1184  U+AD18  uniAD18,  HANGUL SYLLABLE-AD18
        '44313' => '9515'             , # G+9515  U+AD19  uniAD19,  HANGUL SYLLABLE-AD19
        '44314' => '9516'             , # G+9516  U+AD1A  uniAD1A,  HANGUL SYLLABLE-AD1A
        '44315' => '9517'             , # G+9517  U+AD1B  uniAD1B,  HANGUL SYLLABLE-AD1B
        '44316' => '1185'             , # G+1185  U+AD1C  uniAD1C,  HANGUL SYLLABLE-AD1C
        '44317' => '9518'             , # G+9518  U+AD1D  uniAD1D,  HANGUL SYLLABLE-AD1D
        '44318' => '9519'             , # G+9519  U+AD1E  uniAD1E,  HANGUL SYLLABLE-AD1E
        '44319' => '9520'             , # G+9520  U+AD1F  uniAD1F,  HANGUL SYLLABLE-AD1F
        '44320' => '1186'             , # G+1186  U+AD20  uniAD20,  HANGUL SYLLABLE-AD20
        '44321' => '9521'             , # G+9521  U+AD21  uniAD21,  HANGUL SYLLABLE-AD21
        '44322' => '9522'             , # G+9522  U+AD22  uniAD22,  HANGUL SYLLABLE-AD22
        '44323' => '9523'             , # G+9523  U+AD23  uniAD23,  HANGUL SYLLABLE-AD23
        '44324' => '9524'             , # G+9524  U+AD24  uniAD24,  HANGUL SYLLABLE-AD24
        '44325' => '9525'             , # G+9525  U+AD25  uniAD25,  HANGUL SYLLABLE-AD25
        '44326' => '9526'             , # G+9526  U+AD26  uniAD26,  HANGUL SYLLABLE-AD26
        '44327' => '9527'             , # G+9527  U+AD27  uniAD27,  HANGUL SYLLABLE-AD27
        '44328' => '9528'             , # G+9528  U+AD28  uniAD28,  HANGUL SYLLABLE-AD28
        '44329' => '1187'             , # G+1187  U+AD29  uniAD29,  HANGUL SYLLABLE-AD29
        '44330' => '9529'             , # G+9529  U+AD2A  uniAD2A,  HANGUL SYLLABLE-AD2A
        '44331' => '9530'             , # G+9530  U+AD2B  uniAD2B,  HANGUL SYLLABLE-AD2B
        '44332' => '1188'             , # G+1188  U+AD2C  uniAD2C,  HANGUL SYLLABLE-AD2C
        '44333' => '1189'             , # G+1189  U+AD2D  uniAD2D,  HANGUL SYLLABLE-AD2D
        '44334' => '9531'             , # G+9531  U+AD2E  uniAD2E,  HANGUL SYLLABLE-AD2E
        '44335' => '9532'             , # G+9532  U+AD2F  uniAD2F,  HANGUL SYLLABLE-AD2F
        '44336' => '9533'             , # G+9533  U+AD30  uniAD30,  HANGUL SYLLABLE-AD30
        '44337' => '9534'             , # G+9534  U+AD31  uniAD31,  HANGUL SYLLABLE-AD31
        '44338' => '9535'             , # G+9535  U+AD32  uniAD32,  HANGUL SYLLABLE-AD32
        '44339' => '9536'             , # G+9536  U+AD33  uniAD33,  HANGUL SYLLABLE-AD33
        '44340' => '1190'             , # G+1190  U+AD34  uniAD34,  HANGUL SYLLABLE-AD34
        '44341' => '1191'             , # G+1191  U+AD35  uniAD35,  HANGUL SYLLABLE-AD35
        '44342' => '9537'             , # G+9537  U+AD36  uniAD36,  HANGUL SYLLABLE-AD36
        '44343' => '9538'             , # G+9538  U+AD37  uniAD37,  HANGUL SYLLABLE-AD37
        '44344' => '1192'             , # G+1192  U+AD38  uniAD38,  HANGUL SYLLABLE-AD38
        '44345' => '9539'             , # G+9539  U+AD39  uniAD39,  HANGUL SYLLABLE-AD39
        '44346' => '9540'             , # G+9540  U+AD3A  uniAD3A,  HANGUL SYLLABLE-AD3A
        '44347' => '9541'             , # G+9541  U+AD3B  uniAD3B,  HANGUL SYLLABLE-AD3B
        '44348' => '1193'             , # G+1193  U+AD3C  uniAD3C,  HANGUL SYLLABLE-AD3C
        '44349' => '9542'             , # G+9542  U+AD3D  uniAD3D,  HANGUL SYLLABLE-AD3D
        '44350' => '9543'             , # G+9543  U+AD3E  uniAD3E,  HANGUL SYLLABLE-AD3E
        '44351' => '9544'             , # G+9544  U+AD3F  uniAD3F,  HANGUL SYLLABLE-AD3F
        '44352' => '9545'             , # G+9545  U+AD40  uniAD40,  HANGUL SYLLABLE-AD40
        '44353' => '9546'             , # G+9546  U+AD41  uniAD41,  HANGUL SYLLABLE-AD41
        '44354' => '9547'             , # G+9547  U+AD42  uniAD42,  HANGUL SYLLABLE-AD42
        '44355' => '9548'             , # G+9548  U+AD43  uniAD43,  HANGUL SYLLABLE-AD43
        '44356' => '1194'             , # G+1194  U+AD44  uniAD44,  HANGUL SYLLABLE-AD44
        '44357' => '1195'             , # G+1195  U+AD45  uniAD45,  HANGUL SYLLABLE-AD45
        '44358' => '9549'             , # G+9549  U+AD46  uniAD46,  HANGUL SYLLABLE-AD46
        '44359' => '1196'             , # G+1196  U+AD47  uniAD47,  HANGUL SYLLABLE-AD47
        '44360' => '9550'             , # G+9550  U+AD48  uniAD48,  HANGUL SYLLABLE-AD48
        '44361' => '1197'             , # G+1197  U+AD49  uniAD49,  HANGUL SYLLABLE-AD49
        '44362' => '9551'             , # G+9551  U+AD4A  uniAD4A,  HANGUL SYLLABLE-AD4A
        '44363' => '9552'             , # G+9552  U+AD4B  uniAD4B,  HANGUL SYLLABLE-AD4B
        '44364' => '9553'             , # G+9553  U+AD4C  uniAD4C,  HANGUL SYLLABLE-AD4C
        '44365' => '9554'             , # G+9554  U+AD4D  uniAD4D,  HANGUL SYLLABLE-AD4D
        '44366' => '9555'             , # G+9555  U+AD4E  uniAD4E,  HANGUL SYLLABLE-AD4E
        '44367' => '9556'             , # G+9556  U+AD4F  uniAD4F,  HANGUL SYLLABLE-AD4F
        '44368' => '1198'             , # G+1198  U+AD50  uniAD50,  HANGUL SYLLABLE-AD50
        '44369' => '9557'             , # G+9557  U+AD51  uniAD51,  HANGUL SYLLABLE-AD51
        '44370' => '9558'             , # G+9558  U+AD52  uniAD52,  HANGUL SYLLABLE-AD52
        '44371' => '9559'             , # G+9559  U+AD53  uniAD53,  HANGUL SYLLABLE-AD53
        '44372' => '1199'             , # G+1199  U+AD54  uniAD54,  HANGUL SYLLABLE-AD54
        '44373' => '9560'             , # G+9560  U+AD55  uniAD55,  HANGUL SYLLABLE-AD55
        '44374' => '9561'             , # G+9561  U+AD56  uniAD56,  HANGUL SYLLABLE-AD56
        '44375' => '9562'             , # G+9562  U+AD57  uniAD57,  HANGUL SYLLABLE-AD57
        '44376' => '1200'             , # G+1200  U+AD58  uniAD58,  HANGUL SYLLABLE-AD58
        '44377' => '9563'             , # G+9563  U+AD59  uniAD59,  HANGUL SYLLABLE-AD59
        '44378' => '9564'             , # G+9564  U+AD5A  uniAD5A,  HANGUL SYLLABLE-AD5A
        '44379' => '9565'             , # G+9565  U+AD5B  uniAD5B,  HANGUL SYLLABLE-AD5B
        '44380' => '9566'             , # G+9566  U+AD5C  uniAD5C,  HANGUL SYLLABLE-AD5C
        '44381' => '9567'             , # G+9567  U+AD5D  uniAD5D,  HANGUL SYLLABLE-AD5D
        '44382' => '9568'             , # G+9568  U+AD5E  uniAD5E,  HANGUL SYLLABLE-AD5E
        '44383' => '9569'             , # G+9569  U+AD5F  uniAD5F,  HANGUL SYLLABLE-AD5F
        '44384' => '9570'             , # G+9570  U+AD60  uniAD60,  HANGUL SYLLABLE-AD60
        '44385' => '1201'             , # G+1201  U+AD61  uniAD61,  HANGUL SYLLABLE-AD61
        '44386' => '9571'             , # G+9571  U+AD62  uniAD62,  HANGUL SYLLABLE-AD62
        '44387' => '1202'             , # G+1202  U+AD63  uniAD63,  HANGUL SYLLABLE-AD63
        '44388' => '9572'             , # G+9572  U+AD64  uniAD64,  HANGUL SYLLABLE-AD64
        '44389' => '9573'             , # G+9573  U+AD65  uniAD65,  HANGUL SYLLABLE-AD65
        '44390' => '9574'             , # G+9574  U+AD66  uniAD66,  HANGUL SYLLABLE-AD66
        '44391' => '9575'             , # G+9575  U+AD67  uniAD67,  HANGUL SYLLABLE-AD67
        '44392' => '9576'             , # G+9576  U+AD68  uniAD68,  HANGUL SYLLABLE-AD68
        '44393' => '9577'             , # G+9577  U+AD69  uniAD69,  HANGUL SYLLABLE-AD69
        '44394' => '9578'             , # G+9578  U+AD6A  uniAD6A,  HANGUL SYLLABLE-AD6A
        '44395' => '9579'             , # G+9579  U+AD6B  uniAD6B,  HANGUL SYLLABLE-AD6B
        '44396' => '1203'             , # G+1203  U+AD6C  uniAD6C,  HANGUL SYLLABLE-AD6C
        '44397' => '1204'             , # G+1204  U+AD6D  uniAD6D,  HANGUL SYLLABLE-AD6D
        '44398' => '9580'             , # G+9580  U+AD6E  uniAD6E,  HANGUL SYLLABLE-AD6E
        '44399' => '9581'             , # G+9581  U+AD6F  uniAD6F,  HANGUL SYLLABLE-AD6F
        '44400' => '1205'             , # G+1205  U+AD70  uniAD70,  HANGUL SYLLABLE-AD70
        '44401' => '9582'             , # G+9582  U+AD71  uniAD71,  HANGUL SYLLABLE-AD71
        '44402' => '9583'             , # G+9583  U+AD72  uniAD72,  HANGUL SYLLABLE-AD72
        '44403' => '1206'             , # G+1206  U+AD73  uniAD73,  HANGUL SYLLABLE-AD73
        '44404' => '1207'             , # G+1207  U+AD74  uniAD74,  HANGUL SYLLABLE-AD74
        '44405' => '1208'             , # G+1208  U+AD75  uniAD75,  HANGUL SYLLABLE-AD75
        '44406' => '1209'             , # G+1209  U+AD76  uniAD76,  HANGUL SYLLABLE-AD76
        '44407' => '9584'             , # G+9584  U+AD77  uniAD77,  HANGUL SYLLABLE-AD77
        '44408' => '9585'             , # G+9585  U+AD78  uniAD78,  HANGUL SYLLABLE-AD78
        '44409' => '9586'             , # G+9586  U+AD79  uniAD79,  HANGUL SYLLABLE-AD79
        '44410' => '9587'             , # G+9587  U+AD7A  uniAD7A,  HANGUL SYLLABLE-AD7A
        '44411' => '1210'             , # G+1210  U+AD7B  uniAD7B,  HANGUL SYLLABLE-AD7B
        '44412' => '1211'             , # G+1211  U+AD7C  uniAD7C,  HANGUL SYLLABLE-AD7C
        '44413' => '1212'             , # G+1212  U+AD7D  uniAD7D,  HANGUL SYLLABLE-AD7D
        '44414' => '9588'             , # G+9588  U+AD7E  uniAD7E,  HANGUL SYLLABLE-AD7E
        '44415' => '1213'             , # G+1213  U+AD7F  uniAD7F,  HANGUL SYLLABLE-AD7F
        '44416' => '9589'             , # G+9589  U+AD80  uniAD80,  HANGUL SYLLABLE-AD80
        '44417' => '1214'             , # G+1214  U+AD81  uniAD81,  HANGUL SYLLABLE-AD81
        '44418' => '1215'             , # G+1215  U+AD82  uniAD82,  HANGUL SYLLABLE-AD82
        '44419' => '9590'             , # G+9590  U+AD83  uniAD83,  HANGUL SYLLABLE-AD83
        '44420' => '9591'             , # G+9591  U+AD84  uniAD84,  HANGUL SYLLABLE-AD84
        '44421' => '9592'             , # G+9592  U+AD85  uniAD85,  HANGUL SYLLABLE-AD85
        '44422' => '9593'             , # G+9593  U+AD86  uniAD86,  HANGUL SYLLABLE-AD86
        '44423' => '9594'             , # G+9594  U+AD87  uniAD87,  HANGUL SYLLABLE-AD87
        '44424' => '1216'             , # G+1216  U+AD88  uniAD88,  HANGUL SYLLABLE-AD88
        '44425' => '1217'             , # G+1217  U+AD89  uniAD89,  HANGUL SYLLABLE-AD89
        '44426' => '9595'             , # G+9595  U+AD8A  uniAD8A,  HANGUL SYLLABLE-AD8A
        '44427' => '9596'             , # G+9596  U+AD8B  uniAD8B,  HANGUL SYLLABLE-AD8B
        '44428' => '1218'             , # G+1218  U+AD8C  uniAD8C,  HANGUL SYLLABLE-AD8C
        '44429' => '9597'             , # G+9597  U+AD8D  uniAD8D,  HANGUL SYLLABLE-AD8D
        '44430' => '9598'             , # G+9598  U+AD8E  uniAD8E,  HANGUL SYLLABLE-AD8E
        '44431' => '9599'             , # G+9599  U+AD8F  uniAD8F,  HANGUL SYLLABLE-AD8F
        '44432' => '1219'             , # G+1219  U+AD90  uniAD90,  HANGUL SYLLABLE-AD90
        '44433' => '9600'             , # G+9600  U+AD91  uniAD91,  HANGUL SYLLABLE-AD91
        '44434' => '9601'             , # G+9601  U+AD92  uniAD92,  HANGUL SYLLABLE-AD92
        '44435' => '9602'             , # G+9602  U+AD93  uniAD93,  HANGUL SYLLABLE-AD93
        '44436' => '9603'             , # G+9603  U+AD94  uniAD94,  HANGUL SYLLABLE-AD94
        '44437' => '9604'             , # G+9604  U+AD95  uniAD95,  HANGUL SYLLABLE-AD95
        '44438' => '9605'             , # G+9605  U+AD96  uniAD96,  HANGUL SYLLABLE-AD96
        '44439' => '9606'             , # G+9606  U+AD97  uniAD97,  HANGUL SYLLABLE-AD97
        '44440' => '9607'             , # G+9607  U+AD98  uniAD98,  HANGUL SYLLABLE-AD98
        '44441' => '9608'             , # G+9608  U+AD99  uniAD99,  HANGUL SYLLABLE-AD99
        '44442' => '9609'             , # G+9609  U+AD9A  uniAD9A,  HANGUL SYLLABLE-AD9A
        '44443' => '9610'             , # G+9610  U+AD9B  uniAD9B,  HANGUL SYLLABLE-AD9B
        '44444' => '1220'             , # G+1220  U+AD9C  uniAD9C,  HANGUL SYLLABLE-AD9C
        '44445' => '1221'             , # G+1221  U+AD9D  uniAD9D,  HANGUL SYLLABLE-AD9D
        '44446' => '9611'             , # G+9611  U+AD9E  uniAD9E,  HANGUL SYLLABLE-AD9E
        '44447' => '9612'             , # G+9612  U+AD9F  uniAD9F,  HANGUL SYLLABLE-AD9F
        '44448' => '9613'             , # G+9613  U+ADA0  uniADA0,  HANGUL SYLLABLE-ADA0
        '44449' => '9614'             , # G+9614  U+ADA1  uniADA1,  HANGUL SYLLABLE-ADA1
        '44450' => '9615'             , # G+9615  U+ADA2  uniADA2,  HANGUL SYLLABLE-ADA2
        '44451' => '9616'             , # G+9616  U+ADA3  uniADA3,  HANGUL SYLLABLE-ADA3
        '44452' => '1222'             , # G+1222  U+ADA4  uniADA4,  HANGUL SYLLABLE-ADA4
        '44453' => '9617'             , # G+9617  U+ADA5  uniADA5,  HANGUL SYLLABLE-ADA5
        '44454' => '9618'             , # G+9618  U+ADA6  uniADA6,  HANGUL SYLLABLE-ADA6
        '44455' => '9619'             , # G+9619  U+ADA7  uniADA7,  HANGUL SYLLABLE-ADA7
        '44456' => '9620'             , # G+9620  U+ADA8  uniADA8,  HANGUL SYLLABLE-ADA8
        '44457' => '9621'             , # G+9621  U+ADA9  uniADA9,  HANGUL SYLLABLE-ADA9
        '44458' => '9622'             , # G+9622  U+ADAA  uniADAA,  HANGUL SYLLABLE-ADAA
        '44459' => '9623'             , # G+9623  U+ADAB  uniADAB,  HANGUL SYLLABLE-ADAB
        '44460' => '9624'             , # G+9624  U+ADAC  uniADAC,  HANGUL SYLLABLE-ADAC
        '44461' => '9625'             , # G+9625  U+ADAD  uniADAD,  HANGUL SYLLABLE-ADAD
        '44462' => '9626'             , # G+9626  U+ADAE  uniADAE,  HANGUL SYLLABLE-ADAE
        '44463' => '9627'             , # G+9627  U+ADAF  uniADAF,  HANGUL SYLLABLE-ADAF
        '44464' => '9628'             , # G+9628  U+ADB0  uniADB0,  HANGUL SYLLABLE-ADB0
        '44465' => '9629'             , # G+9629  U+ADB1  uniADB1,  HANGUL SYLLABLE-ADB1
        '44466' => '9630'             , # G+9630  U+ADB2  uniADB2,  HANGUL SYLLABLE-ADB2
        '44467' => '9631'             , # G+9631  U+ADB3  uniADB3,  HANGUL SYLLABLE-ADB3
        '44468' => '9632'             , # G+9632  U+ADB4  uniADB4,  HANGUL SYLLABLE-ADB4
        '44469' => '9633'             , # G+9633  U+ADB5  uniADB5,  HANGUL SYLLABLE-ADB5
        '44470' => '9634'             , # G+9634  U+ADB6  uniADB6,  HANGUL SYLLABLE-ADB6
        '44471' => '1223'             , # G+1223  U+ADB7  uniADB7,  HANGUL SYLLABLE-ADB7
        '44472' => '9635'             , # G+9635  U+ADB8  uniADB8,  HANGUL SYLLABLE-ADB8
        '44473' => '9636'             , # G+9636  U+ADB9  uniADB9,  HANGUL SYLLABLE-ADB9
        '44474' => '9637'             , # G+9637  U+ADBA  uniADBA,  HANGUL SYLLABLE-ADBA
        '44475' => '9638'             , # G+9638  U+ADBB  uniADBB,  HANGUL SYLLABLE-ADBB
        '44476' => '9639'             , # G+9639  U+ADBC  uniADBC,  HANGUL SYLLABLE-ADBC
        '44477' => '9640'             , # G+9640  U+ADBD  uniADBD,  HANGUL SYLLABLE-ADBD
        '44478' => '9641'             , # G+9641  U+ADBE  uniADBE,  HANGUL SYLLABLE-ADBE
        '44479' => '9642'             , # G+9642  U+ADBF  uniADBF,  HANGUL SYLLABLE-ADBF
        '44480' => '1224'             , # G+1224  U+ADC0  uniADC0,  HANGUL SYLLABLE-ADC0
        '44481' => '1225'             , # G+1225  U+ADC1  uniADC1,  HANGUL SYLLABLE-ADC1
        '44482' => '9643'             , # G+9643  U+ADC2  uniADC2,  HANGUL SYLLABLE-ADC2
        '44483' => '9644'             , # G+9644  U+ADC3  uniADC3,  HANGUL SYLLABLE-ADC3
        '44484' => '1226'             , # G+1226  U+ADC4  uniADC4,  HANGUL SYLLABLE-ADC4
        '44485' => '9645'             , # G+9645  U+ADC5  uniADC5,  HANGUL SYLLABLE-ADC5
        '44486' => '9646'             , # G+9646  U+ADC6  uniADC6,  HANGUL SYLLABLE-ADC6
        '44487' => '9647'             , # G+9647  U+ADC7  uniADC7,  HANGUL SYLLABLE-ADC7
        '44488' => '1227'             , # G+1227  U+ADC8  uniADC8,  HANGUL SYLLABLE-ADC8
        '44489' => '9648'             , # G+9648  U+ADC9  uniADC9,  HANGUL SYLLABLE-ADC9
        '44490' => '9649'             , # G+9649  U+ADCA  uniADCA,  HANGUL SYLLABLE-ADCA
        '44491' => '9650'             , # G+9650  U+ADCB  uniADCB,  HANGUL SYLLABLE-ADCB
        '44492' => '9651'             , # G+9651  U+ADCC  uniADCC,  HANGUL SYLLABLE-ADCC
        '44493' => '9652'             , # G+9652  U+ADCD  uniADCD,  HANGUL SYLLABLE-ADCD
        '44494' => '9653'             , # G+9653  U+ADCE  uniADCE,  HANGUL SYLLABLE-ADCE
        '44495' => '9654'             , # G+9654  U+ADCF  uniADCF,  HANGUL SYLLABLE-ADCF
        '44496' => '1228'             , # G+1228  U+ADD0  uniADD0,  HANGUL SYLLABLE-ADD0
        '44497' => '1229'             , # G+1229  U+ADD1  uniADD1,  HANGUL SYLLABLE-ADD1
        '44498' => '9655'             , # G+9655  U+ADD2  uniADD2,  HANGUL SYLLABLE-ADD2
        '44499' => '1230'             , # G+1230  U+ADD3  uniADD3,  HANGUL SYLLABLE-ADD3
        '44500' => '9656'             , # G+9656  U+ADD4  uniADD4,  HANGUL SYLLABLE-ADD4
        '44501' => '9657'             , # G+9657  U+ADD5  uniADD5,  HANGUL SYLLABLE-ADD5
        '44502' => '9658'             , # G+9658  U+ADD6  uniADD6,  HANGUL SYLLABLE-ADD6
        '44503' => '9659'             , # G+9659  U+ADD7  uniADD7,  HANGUL SYLLABLE-ADD7
        '44504' => '9660'             , # G+9660  U+ADD8  uniADD8,  HANGUL SYLLABLE-ADD8
        '44505' => '9661'             , # G+9661  U+ADD9  uniADD9,  HANGUL SYLLABLE-ADD9
        '44506' => '9662'             , # G+9662  U+ADDA  uniADDA,  HANGUL SYLLABLE-ADDA
        '44507' => '9663'             , # G+9663  U+ADDB  uniADDB,  HANGUL SYLLABLE-ADDB
        '44508' => '1231'             , # G+1231  U+ADDC  uniADDC,  HANGUL SYLLABLE-ADDC
        '44509' => '9664'             , # G+9664  U+ADDD  uniADDD,  HANGUL SYLLABLE-ADDD
        '44510' => '9665'             , # G+9665  U+ADDE  uniADDE,  HANGUL SYLLABLE-ADDE
        '44511' => '9666'             , # G+9666  U+ADDF  uniADDF,  HANGUL SYLLABLE-ADDF
        '44512' => '1232'             , # G+1232  U+ADE0  uniADE0,  HANGUL SYLLABLE-ADE0
        '44513' => '9667'             , # G+9667  U+ADE1  uniADE1,  HANGUL SYLLABLE-ADE1
        '44514' => '9668'             , # G+9668  U+ADE2  uniADE2,  HANGUL SYLLABLE-ADE2
        '44515' => '9669'             , # G+9669  U+ADE3  uniADE3,  HANGUL SYLLABLE-ADE3
        '44516' => '1233'             , # G+1233  U+ADE4  uniADE4,  HANGUL SYLLABLE-ADE4
        '44517' => '9670'             , # G+9670  U+ADE5  uniADE5,  HANGUL SYLLABLE-ADE5
        '44518' => '9671'             , # G+9671  U+ADE6  uniADE6,  HANGUL SYLLABLE-ADE6
        '44519' => '9672'             , # G+9672  U+ADE7  uniADE7,  HANGUL SYLLABLE-ADE7
        '44520' => '9673'             , # G+9673  U+ADE8  uniADE8,  HANGUL SYLLABLE-ADE8
        '44521' => '9674'             , # G+9674  U+ADE9  uniADE9,  HANGUL SYLLABLE-ADE9
        '44522' => '9675'             , # G+9675  U+ADEA  uniADEA,  HANGUL SYLLABLE-ADEA
        '44523' => '9676'             , # G+9676  U+ADEB  uniADEB,  HANGUL SYLLABLE-ADEB
        '44524' => '9677'             , # G+9677  U+ADEC  uniADEC,  HANGUL SYLLABLE-ADEC
        '44525' => '9678'             , # G+9678  U+ADED  uniADED,  HANGUL SYLLABLE-ADED
        '44526' => '9679'             , # G+9679  U+ADEE  uniADEE,  HANGUL SYLLABLE-ADEE
        '44527' => '9680'             , # G+9680  U+ADEF  uniADEF,  HANGUL SYLLABLE-ADEF
        '44528' => '9681'             , # G+9681  U+ADF0  uniADF0,  HANGUL SYLLABLE-ADF0
        '44529' => '9682'             , # G+9682  U+ADF1  uniADF1,  HANGUL SYLLABLE-ADF1
        '44530' => '9683'             , # G+9683  U+ADF2  uniADF2,  HANGUL SYLLABLE-ADF2
        '44531' => '9684'             , # G+9684  U+ADF3  uniADF3,  HANGUL SYLLABLE-ADF3
        '44532' => '9685'             , # G+9685  U+ADF4  uniADF4,  HANGUL SYLLABLE-ADF4
        '44533' => '9686'             , # G+9686  U+ADF5  uniADF5,  HANGUL SYLLABLE-ADF5
        '44534' => '9687'             , # G+9687  U+ADF6  uniADF6,  HANGUL SYLLABLE-ADF6
        '44535' => '9688'             , # G+9688  U+ADF7  uniADF7,  HANGUL SYLLABLE-ADF7
        '44536' => '1234'             , # G+1234  U+ADF8  uniADF8,  HANGUL SYLLABLE-ADF8
        '44537' => '1235'             , # G+1235  U+ADF9  uniADF9,  HANGUL SYLLABLE-ADF9
        '44538' => '9689'             , # G+9689  U+ADFA  uniADFA,  HANGUL SYLLABLE-ADFA
        '44539' => '9690'             , # G+9690  U+ADFB  uniADFB,  HANGUL SYLLABLE-ADFB
        '44540' => '1236'             , # G+1236  U+ADFC  uniADFC,  HANGUL SYLLABLE-ADFC
        '44541' => '9691'             , # G+9691  U+ADFD  uniADFD,  HANGUL SYLLABLE-ADFD
        '44542' => '9692'             , # G+9692  U+ADFE  uniADFE,  HANGUL SYLLABLE-ADFE
        '44543' => '1237'             , # G+1237  U+ADFF  uniADFF,  HANGUL SYLLABLE-ADFF
        '44544' => '1238'             , # G+1238  U+AE00  uniAE00,  HANGUL SYLLABLE-AE00
        '44545' => '1239'             , # G+1239  U+AE01  uniAE01,  HANGUL SYLLABLE-AE01
        '44546' => '9693'             , # G+9693  U+AE02  uniAE02,  HANGUL SYLLABLE-AE02
        '44547' => '9694'             , # G+9694  U+AE03  uniAE03,  HANGUL SYLLABLE-AE03
        '44548' => '9695'             , # G+9695  U+AE04  uniAE04,  HANGUL SYLLABLE-AE04
        '44549' => '9696'             , # G+9696  U+AE05  uniAE05,  HANGUL SYLLABLE-AE05
        '44550' => '9697'             , # G+9697  U+AE06  uniAE06,  HANGUL SYLLABLE-AE06
        '44551' => '9698'             , # G+9698  U+AE07  uniAE07,  HANGUL SYLLABLE-AE07
        '44552' => '1240'             , # G+1240  U+AE08  uniAE08,  HANGUL SYLLABLE-AE08
        '44553' => '1241'             , # G+1241  U+AE09  uniAE09,  HANGUL SYLLABLE-AE09
        '44554' => '9699'             , # G+9699  U+AE0A  uniAE0A,  HANGUL SYLLABLE-AE0A
        '44555' => '1242'             , # G+1242  U+AE0B  uniAE0B,  HANGUL SYLLABLE-AE0B
        '44556' => '9700'             , # G+9700  U+AE0C  uniAE0C,  HANGUL SYLLABLE-AE0C
        '44557' => '1243'             , # G+1243  U+AE0D  uniAE0D,  HANGUL SYLLABLE-AE0D
        '44558' => '9701'             , # G+9701  U+AE0E  uniAE0E,  HANGUL SYLLABLE-AE0E
        '44559' => '9702'             , # G+9702  U+AE0F  uniAE0F,  HANGUL SYLLABLE-AE0F
        '44560' => '9703'             , # G+9703  U+AE10  uniAE10,  HANGUL SYLLABLE-AE10
        '44561' => '9704'             , # G+9704  U+AE11  uniAE11,  HANGUL SYLLABLE-AE11
        '44562' => '9705'             , # G+9705  U+AE12  uniAE12,  HANGUL SYLLABLE-AE12
        '44563' => '9706'             , # G+9706  U+AE13  uniAE13,  HANGUL SYLLABLE-AE13
        '44564' => '1244'             , # G+1244  U+AE14  uniAE14,  HANGUL SYLLABLE-AE14
        '44565' => '9707'             , # G+9707  U+AE15  uniAE15,  HANGUL SYLLABLE-AE15
        '44566' => '9708'             , # G+9708  U+AE16  uniAE16,  HANGUL SYLLABLE-AE16
        '44567' => '9709'             , # G+9709  U+AE17  uniAE17,  HANGUL SYLLABLE-AE17
        '44568' => '9710'             , # G+9710  U+AE18  uniAE18,  HANGUL SYLLABLE-AE18
        '44569' => '9711'             , # G+9711  U+AE19  uniAE19,  HANGUL SYLLABLE-AE19
        '44570' => '9712'             , # G+9712  U+AE1A  uniAE1A,  HANGUL SYLLABLE-AE1A
        '44571' => '9713'             , # G+9713  U+AE1B  uniAE1B,  HANGUL SYLLABLE-AE1B
        '44572' => '9714'             , # G+9714  U+AE1C  uniAE1C,  HANGUL SYLLABLE-AE1C
        '44573' => '9715'             , # G+9715  U+AE1D  uniAE1D,  HANGUL SYLLABLE-AE1D
        '44574' => '9716'             , # G+9716  U+AE1E  uniAE1E,  HANGUL SYLLABLE-AE1E
        '44575' => '9717'             , # G+9717  U+AE1F  uniAE1F,  HANGUL SYLLABLE-AE1F
        '44576' => '9718'             , # G+9718  U+AE20  uniAE20,  HANGUL SYLLABLE-AE20
        '44577' => '9719'             , # G+9719  U+AE21  uniAE21,  HANGUL SYLLABLE-AE21
        '44578' => '9720'             , # G+9720  U+AE22  uniAE22,  HANGUL SYLLABLE-AE22
        '44579' => '9721'             , # G+9721  U+AE23  uniAE23,  HANGUL SYLLABLE-AE23
        '44580' => '9722'             , # G+9722  U+AE24  uniAE24,  HANGUL SYLLABLE-AE24
        '44581' => '9723'             , # G+9723  U+AE25  uniAE25,  HANGUL SYLLABLE-AE25
        '44582' => '9724'             , # G+9724  U+AE26  uniAE26,  HANGUL SYLLABLE-AE26
        '44583' => '9725'             , # G+9725  U+AE27  uniAE27,  HANGUL SYLLABLE-AE27
        '44584' => '9726'             , # G+9726  U+AE28  uniAE28,  HANGUL SYLLABLE-AE28
        '44585' => '9727'             , # G+9727  U+AE29  uniAE29,  HANGUL SYLLABLE-AE29
        '44586' => '9728'             , # G+9728  U+AE2A  uniAE2A,  HANGUL SYLLABLE-AE2A
        '44587' => '9729'             , # G+9729  U+AE2B  uniAE2B,  HANGUL SYLLABLE-AE2B
        '44588' => '9730'             , # G+9730  U+AE2C  uniAE2C,  HANGUL SYLLABLE-AE2C
        '44589' => '9731'             , # G+9731  U+AE2D  uniAE2D,  HANGUL SYLLABLE-AE2D
        '44590' => '9732'             , # G+9732  U+AE2E  uniAE2E,  HANGUL SYLLABLE-AE2E
        '44591' => '9733'             , # G+9733  U+AE2F  uniAE2F,  HANGUL SYLLABLE-AE2F
        '44592' => '1245'             , # G+1245  U+AE30  uniAE30,  HANGUL SYLLABLE-AE30
        '44593' => '1246'             , # G+1246  U+AE31  uniAE31,  HANGUL SYLLABLE-AE31
        '44594' => '9734'             , # G+9734  U+AE32  uniAE32,  HANGUL SYLLABLE-AE32
        '44595' => '9735'             , # G+9735  U+AE33  uniAE33,  HANGUL SYLLABLE-AE33
        '44596' => '1247'             , # G+1247  U+AE34  uniAE34,  HANGUL SYLLABLE-AE34
        '44597' => '9736'             , # G+9736  U+AE35  uniAE35,  HANGUL SYLLABLE-AE35
        '44598' => '9737'             , # G+9737  U+AE36  uniAE36,  HANGUL SYLLABLE-AE36
        '44599' => '1248'             , # G+1248  U+AE37  uniAE37,  HANGUL SYLLABLE-AE37
        '44600' => '1249'             , # G+1249  U+AE38  uniAE38,  HANGUL SYLLABLE-AE38
        '44601' => '9738'             , # G+9738  U+AE39  uniAE39,  HANGUL SYLLABLE-AE39
        '44602' => '1250'             , # G+1250  U+AE3A  uniAE3A,  HANGUL SYLLABLE-AE3A
        '44603' => '9739'             , # G+9739  U+AE3B  uniAE3B,  HANGUL SYLLABLE-AE3B
        '44604' => '9740'             , # G+9740  U+AE3C  uniAE3C,  HANGUL SYLLABLE-AE3C
        '44605' => '9741'             , # G+9741  U+AE3D  uniAE3D,  HANGUL SYLLABLE-AE3D
        '44606' => '9742'             , # G+9742  U+AE3E  uniAE3E,  HANGUL SYLLABLE-AE3E
        '44607' => '9743'             , # G+9743  U+AE3F  uniAE3F,  HANGUL SYLLABLE-AE3F
        '44608' => '1251'             , # G+1251  U+AE40  uniAE40,  HANGUL SYLLABLE-AE40
        '44609' => '1252'             , # G+1252  U+AE41  uniAE41,  HANGUL SYLLABLE-AE41
        '44610' => '9744'             , # G+9744  U+AE42  uniAE42,  HANGUL SYLLABLE-AE42
        '44611' => '1253'             , # G+1253  U+AE43  uniAE43,  HANGUL SYLLABLE-AE43
        '44612' => '9745'             , # G+9745  U+AE44  uniAE44,  HANGUL SYLLABLE-AE44
        '44613' => '1254'             , # G+1254  U+AE45  uniAE45,  HANGUL SYLLABLE-AE45
        '44614' => '1255'             , # G+1255  U+AE46  uniAE46,  HANGUL SYLLABLE-AE46
        '44615' => '9746'             , # G+9746  U+AE47  uniAE47,  HANGUL SYLLABLE-AE47
        '44616' => '9747'             , # G+9747  U+AE48  uniAE48,  HANGUL SYLLABLE-AE48
        '44617' => '9748'             , # G+9748  U+AE49  uniAE49,  HANGUL SYLLABLE-AE49
        '44618' => '1256'             , # G+1256  U+AE4A  uniAE4A,  HANGUL SYLLABLE-AE4A
        '44619' => '9749'             , # G+9749  U+AE4B  uniAE4B,  HANGUL SYLLABLE-AE4B
        '44620' => '1257'             , # G+1257  U+AE4C  uniAE4C,  HANGUL SYLLABLE-AE4C
        '44621' => '1258'             , # G+1258  U+AE4D  uniAE4D,  HANGUL SYLLABLE-AE4D
        '44622' => '1259'             , # G+1259  U+AE4E  uniAE4E,  HANGUL SYLLABLE-AE4E
        '44623' => '9750'             , # G+9750  U+AE4F  uniAE4F,  HANGUL SYLLABLE-AE4F
        '44624' => '1260'             , # G+1260  U+AE50  uniAE50,  HANGUL SYLLABLE-AE50
        '44625' => '9751'             , # G+9751  U+AE51  uniAE51,  HANGUL SYLLABLE-AE51
        '44626' => '9752'             , # G+9752  U+AE52  uniAE52,  HANGUL SYLLABLE-AE52
        '44627' => '9753'             , # G+9753  U+AE53  uniAE53,  HANGUL SYLLABLE-AE53
        '44628' => '1261'             , # G+1261  U+AE54  uniAE54,  HANGUL SYLLABLE-AE54
        '44629' => '9754'             , # G+9754  U+AE55  uniAE55,  HANGUL SYLLABLE-AE55
        '44630' => '1262'             , # G+1262  U+AE56  uniAE56,  HANGUL SYLLABLE-AE56
        '44631' => '9755'             , # G+9755  U+AE57  uniAE57,  HANGUL SYLLABLE-AE57
        '44632' => '9756'             , # G+9756  U+AE58  uniAE58,  HANGUL SYLLABLE-AE58
        '44633' => '9757'             , # G+9757  U+AE59  uniAE59,  HANGUL SYLLABLE-AE59
        '44634' => '9758'             , # G+9758  U+AE5A  uniAE5A,  HANGUL SYLLABLE-AE5A
        '44635' => '9759'             , # G+9759  U+AE5B  uniAE5B,  HANGUL SYLLABLE-AE5B
        '44636' => '1263'             , # G+1263  U+AE5C  uniAE5C,  HANGUL SYLLABLE-AE5C
        '44637' => '1264'             , # G+1264  U+AE5D  uniAE5D,  HANGUL SYLLABLE-AE5D
        '44638' => '9760'             , # G+9760  U+AE5E  uniAE5E,  HANGUL SYLLABLE-AE5E
        '44639' => '1265'             , # G+1265  U+AE5F  uniAE5F,  HANGUL SYLLABLE-AE5F
        '44640' => '1266'             , # G+1266  U+AE60  uniAE60,  HANGUL SYLLABLE-AE60
        '44641' => '1267'             , # G+1267  U+AE61  uniAE61,  HANGUL SYLLABLE-AE61
        '44642' => '9761'             , # G+9761  U+AE62  uniAE62,  HANGUL SYLLABLE-AE62
        '44643' => '9762'             , # G+9762  U+AE63  uniAE63,  HANGUL SYLLABLE-AE63
        '44644' => '9763'             , # G+9763  U+AE64  uniAE64,  HANGUL SYLLABLE-AE64
        '44645' => '1268'             , # G+1268  U+AE65  uniAE65,  HANGUL SYLLABLE-AE65
        '44646' => '9764'             , # G+9764  U+AE66  uniAE66,  HANGUL SYLLABLE-AE66
        '44647' => '9765'             , # G+9765  U+AE67  uniAE67,  HANGUL SYLLABLE-AE67
        '44648' => '1269'             , # G+1269  U+AE68  uniAE68,  HANGUL SYLLABLE-AE68
        '44649' => '1270'             , # G+1270  U+AE69  uniAE69,  HANGUL SYLLABLE-AE69
        '44650' => '9766'             , # G+9766  U+AE6A  uniAE6A,  HANGUL SYLLABLE-AE6A
        '44651' => '9767'             , # G+9767  U+AE6B  uniAE6B,  HANGUL SYLLABLE-AE6B
        '44652' => '1271'             , # G+1271  U+AE6C  uniAE6C,  HANGUL SYLLABLE-AE6C
        '44653' => '9768'             , # G+9768  U+AE6D  uniAE6D,  HANGUL SYLLABLE-AE6D
        '44654' => '9769'             , # G+9769  U+AE6E  uniAE6E,  HANGUL SYLLABLE-AE6E
        '44655' => '9770'             , # G+9770  U+AE6F  uniAE6F,  HANGUL SYLLABLE-AE6F
        '44656' => '1272'             , # G+1272  U+AE70  uniAE70,  HANGUL SYLLABLE-AE70
        '44657' => '9771'             , # G+9771  U+AE71  uniAE71,  HANGUL SYLLABLE-AE71
        '44658' => '9772'             , # G+9772  U+AE72  uniAE72,  HANGUL SYLLABLE-AE72
        '44659' => '9773'             , # G+9773  U+AE73  uniAE73,  HANGUL SYLLABLE-AE73
        '44660' => '9774'             , # G+9774  U+AE74  uniAE74,  HANGUL SYLLABLE-AE74
        '44661' => '9775'             , # G+9775  U+AE75  uniAE75,  HANGUL SYLLABLE-AE75
        '44662' => '9776'             , # G+9776  U+AE76  uniAE76,  HANGUL SYLLABLE-AE76
        '44663' => '9777'             , # G+9777  U+AE77  uniAE77,  HANGUL SYLLABLE-AE77
        '44664' => '1273'             , # G+1273  U+AE78  uniAE78,  HANGUL SYLLABLE-AE78
        '44665' => '1274'             , # G+1274  U+AE79  uniAE79,  HANGUL SYLLABLE-AE79
        '44666' => '9778'             , # G+9778  U+AE7A  uniAE7A,  HANGUL SYLLABLE-AE7A
        '44667' => '1275'             , # G+1275  U+AE7B  uniAE7B,  HANGUL SYLLABLE-AE7B
        '44668' => '1276'             , # G+1276  U+AE7C  uniAE7C,  HANGUL SYLLABLE-AE7C
        '44669' => '1277'             , # G+1277  U+AE7D  uniAE7D,  HANGUL SYLLABLE-AE7D
        '44670' => '9779'             , # G+9779  U+AE7E  uniAE7E,  HANGUL SYLLABLE-AE7E
        '44671' => '9780'             , # G+9780  U+AE7F  uniAE7F,  HANGUL SYLLABLE-AE7F
        '44672' => '9781'             , # G+9781  U+AE80  uniAE80,  HANGUL SYLLABLE-AE80
        '44673' => '9782'             , # G+9782  U+AE81  uniAE81,  HANGUL SYLLABLE-AE81
        '44674' => '9783'             , # G+9783  U+AE82  uniAE82,  HANGUL SYLLABLE-AE82
        '44675' => '9784'             , # G+9784  U+AE83  uniAE83,  HANGUL SYLLABLE-AE83
        '44676' => '1278'             , # G+1278  U+AE84  uniAE84,  HANGUL SYLLABLE-AE84
        '44677' => '1279'             , # G+1279  U+AE85  uniAE85,  HANGUL SYLLABLE-AE85
        '44678' => '9785'             , # G+9785  U+AE86  uniAE86,  HANGUL SYLLABLE-AE86
        '44679' => '9786'             , # G+9786  U+AE87  uniAE87,  HANGUL SYLLABLE-AE87
        '44680' => '9787'             , # G+9787  U+AE88  uniAE88,  HANGUL SYLLABLE-AE88
        '44681' => '9788'             , # G+9788  U+AE89  uniAE89,  HANGUL SYLLABLE-AE89
        '44682' => '9789'             , # G+9789  U+AE8A  uniAE8A,  HANGUL SYLLABLE-AE8A
        '44683' => '9790'             , # G+9790  U+AE8B  uniAE8B,  HANGUL SYLLABLE-AE8B
        '44684' => '1280'             , # G+1280  U+AE8C  uniAE8C,  HANGUL SYLLABLE-AE8C
        '44685' => '9791'             , # G+9791  U+AE8D  uniAE8D,  HANGUL SYLLABLE-AE8D
        '44686' => '9792'             , # G+9792  U+AE8E  uniAE8E,  HANGUL SYLLABLE-AE8E
        '44687' => '9793'             , # G+9793  U+AE8F  uniAE8F,  HANGUL SYLLABLE-AE8F
        '44688' => '9794'             , # G+9794  U+AE90  uniAE90,  HANGUL SYLLABLE-AE90
        '44689' => '9795'             , # G+9795  U+AE91  uniAE91,  HANGUL SYLLABLE-AE91
        '44690' => '9796'             , # G+9796  U+AE92  uniAE92,  HANGUL SYLLABLE-AE92
        '44691' => '9797'             , # G+9797  U+AE93  uniAE93,  HANGUL SYLLABLE-AE93
        '44692' => '9798'             , # G+9798  U+AE94  uniAE94,  HANGUL SYLLABLE-AE94
        '44693' => '9799'             , # G+9799  U+AE95  uniAE95,  HANGUL SYLLABLE-AE95
        '44694' => '9800'             , # G+9800  U+AE96  uniAE96,  HANGUL SYLLABLE-AE96
        '44695' => '9801'             , # G+9801  U+AE97  uniAE97,  HANGUL SYLLABLE-AE97
        '44696' => '9802'             , # G+9802  U+AE98  uniAE98,  HANGUL SYLLABLE-AE98
        '44697' => '9803'             , # G+9803  U+AE99  uniAE99,  HANGUL SYLLABLE-AE99
        '44698' => '9804'             , # G+9804  U+AE9A  uniAE9A,  HANGUL SYLLABLE-AE9A
        '44699' => '9805'             , # G+9805  U+AE9B  uniAE9B,  HANGUL SYLLABLE-AE9B
        '44700' => '9806'             , # G+9806  U+AE9C  uniAE9C,  HANGUL SYLLABLE-AE9C
        '44701' => '9807'             , # G+9807  U+AE9D  uniAE9D,  HANGUL SYLLABLE-AE9D
        '44702' => '9808'             , # G+9808  U+AE9E  uniAE9E,  HANGUL SYLLABLE-AE9E
        '44703' => '9809'             , # G+9809  U+AE9F  uniAE9F,  HANGUL SYLLABLE-AE9F
        '44704' => '9810'             , # G+9810  U+AEA0  uniAEA0,  HANGUL SYLLABLE-AEA0
        '44705' => '9811'             , # G+9811  U+AEA1  uniAEA1,  HANGUL SYLLABLE-AEA1
        '44706' => '9812'             , # G+9812  U+AEA2  uniAEA2,  HANGUL SYLLABLE-AEA2
        '44707' => '9813'             , # G+9813  U+AEA3  uniAEA3,  HANGUL SYLLABLE-AEA3
        '44708' => '9814'             , # G+9814  U+AEA4  uniAEA4,  HANGUL SYLLABLE-AEA4
        '44709' => '9815'             , # G+9815  U+AEA5  uniAEA5,  HANGUL SYLLABLE-AEA5
        '44710' => '9816'             , # G+9816  U+AEA6  uniAEA6,  HANGUL SYLLABLE-AEA6
        '44711' => '9817'             , # G+9817  U+AEA7  uniAEA7,  HANGUL SYLLABLE-AEA7
        '44712' => '9818'             , # G+9818  U+AEA8  uniAEA8,  HANGUL SYLLABLE-AEA8
        '44713' => '9819'             , # G+9819  U+AEA9  uniAEA9,  HANGUL SYLLABLE-AEA9
        '44714' => '9820'             , # G+9820  U+AEAA  uniAEAA,  HANGUL SYLLABLE-AEAA
        '44715' => '9821'             , # G+9821  U+AEAB  uniAEAB,  HANGUL SYLLABLE-AEAB
        '44716' => '9822'             , # G+9822  U+AEAC  uniAEAC,  HANGUL SYLLABLE-AEAC
        '44717' => '9823'             , # G+9823  U+AEAD  uniAEAD,  HANGUL SYLLABLE-AEAD
        '44718' => '9824'             , # G+9824  U+AEAE  uniAEAE,  HANGUL SYLLABLE-AEAE
        '44719' => '9825'             , # G+9825  U+AEAF  uniAEAF,  HANGUL SYLLABLE-AEAF
        '44720' => '9826'             , # G+9826  U+AEB0  uniAEB0,  HANGUL SYLLABLE-AEB0
        '44721' => '9827'             , # G+9827  U+AEB1  uniAEB1,  HANGUL SYLLABLE-AEB1
        '44722' => '9828'             , # G+9828  U+AEB2  uniAEB2,  HANGUL SYLLABLE-AEB2
        '44723' => '9829'             , # G+9829  U+AEB3  uniAEB3,  HANGUL SYLLABLE-AEB3
        '44724' => '9830'             , # G+9830  U+AEB4  uniAEB4,  HANGUL SYLLABLE-AEB4
        '44725' => '9831'             , # G+9831  U+AEB5  uniAEB5,  HANGUL SYLLABLE-AEB5
        '44726' => '9832'             , # G+9832  U+AEB6  uniAEB6,  HANGUL SYLLABLE-AEB6
        '44727' => '9833'             , # G+9833  U+AEB7  uniAEB7,  HANGUL SYLLABLE-AEB7
        '44728' => '9834'             , # G+9834  U+AEB8  uniAEB8,  HANGUL SYLLABLE-AEB8
        '44729' => '9835'             , # G+9835  U+AEB9  uniAEB9,  HANGUL SYLLABLE-AEB9
        '44730' => '9836'             , # G+9836  U+AEBA  uniAEBA,  HANGUL SYLLABLE-AEBA
        '44731' => '9837'             , # G+9837  U+AEBB  uniAEBB,  HANGUL SYLLABLE-AEBB
        '44732' => '1281'             , # G+1281  U+AEBC  uniAEBC,  HANGUL SYLLABLE-AEBC
        '44733' => '1282'             , # G+1282  U+AEBD  uniAEBD,  HANGUL SYLLABLE-AEBD
        '44734' => '1283'             , # G+1283  U+AEBE  uniAEBE,  HANGUL SYLLABLE-AEBE
        '44735' => '9838'             , # G+9838  U+AEBF  uniAEBF,  HANGUL SYLLABLE-AEBF
        '44736' => '1284'             , # G+1284  U+AEC0  uniAEC0,  HANGUL SYLLABLE-AEC0
        '44737' => '9839'             , # G+9839  U+AEC1  uniAEC1,  HANGUL SYLLABLE-AEC1
        '44738' => '9840'             , # G+9840  U+AEC2  uniAEC2,  HANGUL SYLLABLE-AEC2
        '44739' => '9841'             , # G+9841  U+AEC3  uniAEC3,  HANGUL SYLLABLE-AEC3
        '44740' => '1285'             , # G+1285  U+AEC4  uniAEC4,  HANGUL SYLLABLE-AEC4
        '44741' => '9842'             , # G+9842  U+AEC5  uniAEC5,  HANGUL SYLLABLE-AEC5
        '44742' => '9843'             , # G+9843  U+AEC6  uniAEC6,  HANGUL SYLLABLE-AEC6
        '44743' => '9844'             , # G+9844  U+AEC7  uniAEC7,  HANGUL SYLLABLE-AEC7
        '44744' => '9845'             , # G+9845  U+AEC8  uniAEC8,  HANGUL SYLLABLE-AEC8
        '44745' => '9846'             , # G+9846  U+AEC9  uniAEC9,  HANGUL SYLLABLE-AEC9
        '44746' => '9847'             , # G+9847  U+AECA  uniAECA,  HANGUL SYLLABLE-AECA
        '44747' => '9848'             , # G+9848  U+AECB  uniAECB,  HANGUL SYLLABLE-AECB
        '44748' => '1286'             , # G+1286  U+AECC  uniAECC,  HANGUL SYLLABLE-AECC
        '44749' => '1287'             , # G+1287  U+AECD  uniAECD,  HANGUL SYLLABLE-AECD
        '44750' => '9849'             , # G+9849  U+AECE  uniAECE,  HANGUL SYLLABLE-AECE
        '44751' => '1288'             , # G+1288  U+AECF  uniAECF,  HANGUL SYLLABLE-AECF
        '44752' => '1289'             , # G+1289  U+AED0  uniAED0,  HANGUL SYLLABLE-AED0
        '44753' => '1290'             , # G+1290  U+AED1  uniAED1,  HANGUL SYLLABLE-AED1
        '44754' => '9850'             , # G+9850  U+AED2  uniAED2,  HANGUL SYLLABLE-AED2
        '44755' => '9851'             , # G+9851  U+AED3  uniAED3,  HANGUL SYLLABLE-AED3
        '44756' => '9852'             , # G+9852  U+AED4  uniAED4,  HANGUL SYLLABLE-AED4
        '44757' => '9853'             , # G+9853  U+AED5  uniAED5,  HANGUL SYLLABLE-AED5
        '44758' => '9854'             , # G+9854  U+AED6  uniAED6,  HANGUL SYLLABLE-AED6
        '44759' => '9855'             , # G+9855  U+AED7  uniAED7,  HANGUL SYLLABLE-AED7
        '44760' => '1291'             , # G+1291  U+AED8  uniAED8,  HANGUL SYLLABLE-AED8
        '44761' => '1292'             , # G+1292  U+AED9  uniAED9,  HANGUL SYLLABLE-AED9
        '44762' => '9856'             , # G+9856  U+AEDA  uniAEDA,  HANGUL SYLLABLE-AEDA
        '44763' => '9857'             , # G+9857  U+AEDB  uniAEDB,  HANGUL SYLLABLE-AEDB
        '44764' => '1293'             , # G+1293  U+AEDC  uniAEDC,  HANGUL SYLLABLE-AEDC
        '44765' => '9858'             , # G+9858  U+AEDD  uniAEDD,  HANGUL SYLLABLE-AEDD
        '44766' => '9859'             , # G+9859  U+AEDE  uniAEDE,  HANGUL SYLLABLE-AEDE
        '44767' => '9860'             , # G+9860  U+AEDF  uniAEDF,  HANGUL SYLLABLE-AEDF
        '44768' => '9861'             , # G+9861  U+AEE0  uniAEE0,  HANGUL SYLLABLE-AEE0
        '44769' => '9862'             , # G+9862  U+AEE1  uniAEE1,  HANGUL SYLLABLE-AEE1
        '44770' => '9863'             , # G+9863  U+AEE2  uniAEE2,  HANGUL SYLLABLE-AEE2
        '44771' => '9864'             , # G+9864  U+AEE3  uniAEE3,  HANGUL SYLLABLE-AEE3
        '44772' => '9865'             , # G+9865  U+AEE4  uniAEE4,  HANGUL SYLLABLE-AEE4
        '44773' => '9866'             , # G+9866  U+AEE5  uniAEE5,  HANGUL SYLLABLE-AEE5
        '44774' => '9867'             , # G+9867  U+AEE6  uniAEE6,  HANGUL SYLLABLE-AEE6
        '44775' => '9868'             , # G+9868  U+AEE7  uniAEE7,  HANGUL SYLLABLE-AEE7
        '44776' => '1294'             , # G+1294  U+AEE8  uniAEE8,  HANGUL SYLLABLE-AEE8
        '44777' => '9869'             , # G+9869  U+AEE9  uniAEE9,  HANGUL SYLLABLE-AEE9
        '44778' => '9870'             , # G+9870  U+AEEA  uniAEEA,  HANGUL SYLLABLE-AEEA
        '44779' => '1295'             , # G+1295  U+AEEB  uniAEEB,  HANGUL SYLLABLE-AEEB
        '44780' => '9871'             , # G+9871  U+AEEC  uniAEEC,  HANGUL SYLLABLE-AEEC
        '44781' => '1296'             , # G+1296  U+AEED  uniAEED,  HANGUL SYLLABLE-AEED
        '44782' => '9872'             , # G+9872  U+AEEE  uniAEEE,  HANGUL SYLLABLE-AEEE
        '44783' => '9873'             , # G+9873  U+AEEF  uniAEEF,  HANGUL SYLLABLE-AEEF
        '44784' => '9874'             , # G+9874  U+AEF0  uniAEF0,  HANGUL SYLLABLE-AEF0
        '44785' => '9875'             , # G+9875  U+AEF1  uniAEF1,  HANGUL SYLLABLE-AEF1
        '44786' => '9876'             , # G+9876  U+AEF2  uniAEF2,  HANGUL SYLLABLE-AEF2
        '44787' => '9877'             , # G+9877  U+AEF3  uniAEF3,  HANGUL SYLLABLE-AEF3
        '44788' => '1297'             , # G+1297  U+AEF4  uniAEF4,  HANGUL SYLLABLE-AEF4
        '44789' => '9878'             , # G+9878  U+AEF5  uniAEF5,  HANGUL SYLLABLE-AEF5
        '44790' => '9879'             , # G+9879  U+AEF6  uniAEF6,  HANGUL SYLLABLE-AEF6
        '44791' => '9880'             , # G+9880  U+AEF7  uniAEF7,  HANGUL SYLLABLE-AEF7
        '44792' => '1298'             , # G+1298  U+AEF8  uniAEF8,  HANGUL SYLLABLE-AEF8
        '44793' => '9881'             , # G+9881  U+AEF9  uniAEF9,  HANGUL SYLLABLE-AEF9
        '44794' => '9882'             , # G+9882  U+AEFA  uniAEFA,  HANGUL SYLLABLE-AEFA
        '44795' => '9883'             , # G+9883  U+AEFB  uniAEFB,  HANGUL SYLLABLE-AEFB
        '44796' => '1299'             , # G+1299  U+AEFC  uniAEFC,  HANGUL SYLLABLE-AEFC
        '44797' => '9884'             , # G+9884  U+AEFD  uniAEFD,  HANGUL SYLLABLE-AEFD
        '44798' => '9885'             , # G+9885  U+AEFE  uniAEFE,  HANGUL SYLLABLE-AEFE
        '44799' => '9886'             , # G+9886  U+AEFF  uniAEFF,  HANGUL SYLLABLE-AEFF
        '44800' => '9887'             , # G+9887  U+AF00  uniAF00,  HANGUL SYLLABLE-AF00
        '44801' => '9888'             , # G+9888  U+AF01  uniAF01,  HANGUL SYLLABLE-AF01
        '44802' => '9889'             , # G+9889  U+AF02  uniAF02,  HANGUL SYLLABLE-AF02
        '44803' => '9890'             , # G+9890  U+AF03  uniAF03,  HANGUL SYLLABLE-AF03
        '44804' => '9891'             , # G+9891  U+AF04  uniAF04,  HANGUL SYLLABLE-AF04
        '44805' => '9892'             , # G+9892  U+AF05  uniAF05,  HANGUL SYLLABLE-AF05
        '44806' => '9893'             , # G+9893  U+AF06  uniAF06,  HANGUL SYLLABLE-AF06
        '44807' => '1300'             , # G+1300  U+AF07  uniAF07,  HANGUL SYLLABLE-AF07
        '44808' => '1301'             , # G+1301  U+AF08  uniAF08,  HANGUL SYLLABLE-AF08
        '44809' => '9894'             , # G+9894  U+AF09  uniAF09,  HANGUL SYLLABLE-AF09
        '44810' => '9895'             , # G+9895  U+AF0A  uniAF0A,  HANGUL SYLLABLE-AF0A
        '44811' => '9896'             , # G+9896  U+AF0B  uniAF0B,  HANGUL SYLLABLE-AF0B
        '44812' => '9897'             , # G+9897  U+AF0C  uniAF0C,  HANGUL SYLLABLE-AF0C
        '44813' => '1302'             , # G+1302  U+AF0D  uniAF0D,  HANGUL SYLLABLE-AF0D
        '44814' => '9898'             , # G+9898  U+AF0E  uniAF0E,  HANGUL SYLLABLE-AF0E
        '44815' => '9899'             , # G+9899  U+AF0F  uniAF0F,  HANGUL SYLLABLE-AF0F
        '44816' => '1303'             , # G+1303  U+AF10  uniAF10,  HANGUL SYLLABLE-AF10
        '44817' => '9900'             , # G+9900  U+AF11  uniAF11,  HANGUL SYLLABLE-AF11
        '44818' => '9901'             , # G+9901  U+AF12  uniAF12,  HANGUL SYLLABLE-AF12
        '44819' => '9902'             , # G+9902  U+AF13  uniAF13,  HANGUL SYLLABLE-AF13
        '44820' => '9903'             , # G+9903  U+AF14  uniAF14,  HANGUL SYLLABLE-AF14
        '44821' => '9904'             , # G+9904  U+AF15  uniAF15,  HANGUL SYLLABLE-AF15
        '44822' => '9905'             , # G+9905  U+AF16  uniAF16,  HANGUL SYLLABLE-AF16
        '44823' => '9906'             , # G+9906  U+AF17  uniAF17,  HANGUL SYLLABLE-AF17
        '44824' => '9907'             , # G+9907  U+AF18  uniAF18,  HANGUL SYLLABLE-AF18
        '44825' => '9908'             , # G+9908  U+AF19  uniAF19,  HANGUL SYLLABLE-AF19
        '44826' => '9909'             , # G+9909  U+AF1A  uniAF1A,  HANGUL SYLLABLE-AF1A
        '44827' => '9910'             , # G+9910  U+AF1B  uniAF1B,  HANGUL SYLLABLE-AF1B
        '44828' => '9911'             , # G+9911  U+AF1C  uniAF1C,  HANGUL SYLLABLE-AF1C
        '44829' => '9912'             , # G+9912  U+AF1D  uniAF1D,  HANGUL SYLLABLE-AF1D
        '44830' => '9913'             , # G+9913  U+AF1E  uniAF1E,  HANGUL SYLLABLE-AF1E
        '44831' => '9914'             , # G+9914  U+AF1F  uniAF1F,  HANGUL SYLLABLE-AF1F
        '44832' => '9915'             , # G+9915  U+AF20  uniAF20,  HANGUL SYLLABLE-AF20
        '44833' => '9916'             , # G+9916  U+AF21  uniAF21,  HANGUL SYLLABLE-AF21
        '44834' => '9917'             , # G+9917  U+AF22  uniAF22,  HANGUL SYLLABLE-AF22
        '44835' => '9918'             , # G+9918  U+AF23  uniAF23,  HANGUL SYLLABLE-AF23
        '44836' => '9919'             , # G+9919  U+AF24  uniAF24,  HANGUL SYLLABLE-AF24
        '44837' => '9920'             , # G+9920  U+AF25  uniAF25,  HANGUL SYLLABLE-AF25
        '44838' => '9921'             , # G+9921  U+AF26  uniAF26,  HANGUL SYLLABLE-AF26
        '44839' => '9922'             , # G+9922  U+AF27  uniAF27,  HANGUL SYLLABLE-AF27
        '44840' => '9923'             , # G+9923  U+AF28  uniAF28,  HANGUL SYLLABLE-AF28
        '44841' => '9924'             , # G+9924  U+AF29  uniAF29,  HANGUL SYLLABLE-AF29
        '44842' => '9925'             , # G+9925  U+AF2A  uniAF2A,  HANGUL SYLLABLE-AF2A
        '44843' => '9926'             , # G+9926  U+AF2B  uniAF2B,  HANGUL SYLLABLE-AF2B
        '44844' => '1304'             , # G+1304  U+AF2C  uniAF2C,  HANGUL SYLLABLE-AF2C
        '44845' => '1305'             , # G+1305  U+AF2D  uniAF2D,  HANGUL SYLLABLE-AF2D
        '44846' => '9927'             , # G+9927  U+AF2E  uniAF2E,  HANGUL SYLLABLE-AF2E
        '44847' => '9928'             , # G+9928  U+AF2F  uniAF2F,  HANGUL SYLLABLE-AF2F
        '44848' => '1306'             , # G+1306  U+AF30  uniAF30,  HANGUL SYLLABLE-AF30
        '44849' => '9929'             , # G+9929  U+AF31  uniAF31,  HANGUL SYLLABLE-AF31
        '44850' => '1307'             , # G+1307  U+AF32  uniAF32,  HANGUL SYLLABLE-AF32
        '44851' => '9930'             , # G+9930  U+AF33  uniAF33,  HANGUL SYLLABLE-AF33
        '44852' => '1308'             , # G+1308  U+AF34  uniAF34,  HANGUL SYLLABLE-AF34
        '44853' => '9931'             , # G+9931  U+AF35  uniAF35,  HANGUL SYLLABLE-AF35
        '44854' => '9932'             , # G+9932  U+AF36  uniAF36,  HANGUL SYLLABLE-AF36
        '44855' => '9933'             , # G+9933  U+AF37  uniAF37,  HANGUL SYLLABLE-AF37
        '44856' => '9934'             , # G+9934  U+AF38  uniAF38,  HANGUL SYLLABLE-AF38
        '44857' => '9935'             , # G+9935  U+AF39  uniAF39,  HANGUL SYLLABLE-AF39
        '44858' => '9936'             , # G+9936  U+AF3A  uniAF3A,  HANGUL SYLLABLE-AF3A
        '44859' => '9937'             , # G+9937  U+AF3B  uniAF3B,  HANGUL SYLLABLE-AF3B
        '44860' => '1309'             , # G+1309  U+AF3C  uniAF3C,  HANGUL SYLLABLE-AF3C
        '44861' => '1310'             , # G+1310  U+AF3D  uniAF3D,  HANGUL SYLLABLE-AF3D
        '44862' => '9938'             , # G+9938  U+AF3E  uniAF3E,  HANGUL SYLLABLE-AF3E
        '44863' => '1311'             , # G+1311  U+AF3F  uniAF3F,  HANGUL SYLLABLE-AF3F
        '44864' => '9939'             , # G+9939  U+AF40  uniAF40,  HANGUL SYLLABLE-AF40
        '44865' => '1312'             , # G+1312  U+AF41  uniAF41,  HANGUL SYLLABLE-AF41
        '44866' => '1313'             , # G+1313  U+AF42  uniAF42,  HANGUL SYLLABLE-AF42
        '44867' => '1314'             , # G+1314  U+AF43  uniAF43,  HANGUL SYLLABLE-AF43
        '44868' => '9940'             , # G+9940  U+AF44  uniAF44,  HANGUL SYLLABLE-AF44
        '44869' => '9941'             , # G+9941  U+AF45  uniAF45,  HANGUL SYLLABLE-AF45
        '44870' => '9942'             , # G+9942  U+AF46  uniAF46,  HANGUL SYLLABLE-AF46
        '44871' => '9943'             , # G+9943  U+AF47  uniAF47,  HANGUL SYLLABLE-AF47
        '44872' => '1315'             , # G+1315  U+AF48  uniAF48,  HANGUL SYLLABLE-AF48
        '44873' => '1316'             , # G+1316  U+AF49  uniAF49,  HANGUL SYLLABLE-AF49
        '44874' => '9944'             , # G+9944  U+AF4A  uniAF4A,  HANGUL SYLLABLE-AF4A
        '44875' => '9945'             , # G+9945  U+AF4B  uniAF4B,  HANGUL SYLLABLE-AF4B
        '44876' => '9946'             , # G+9946  U+AF4C  uniAF4C,  HANGUL SYLLABLE-AF4C
        '44877' => '9947'             , # G+9947  U+AF4D  uniAF4D,  HANGUL SYLLABLE-AF4D
        '44878' => '9948'             , # G+9948  U+AF4E  uniAF4E,  HANGUL SYLLABLE-AF4E
        '44879' => '9949'             , # G+9949  U+AF4F  uniAF4F,  HANGUL SYLLABLE-AF4F
        '44880' => '1317'             , # G+1317  U+AF50  uniAF50,  HANGUL SYLLABLE-AF50
        '44881' => '9950'             , # G+9950  U+AF51  uniAF51,  HANGUL SYLLABLE-AF51
        '44882' => '9951'             , # G+9951  U+AF52  uniAF52,  HANGUL SYLLABLE-AF52
        '44883' => '9952'             , # G+9952  U+AF53  uniAF53,  HANGUL SYLLABLE-AF53
        '44884' => '9953'             , # G+9953  U+AF54  uniAF54,  HANGUL SYLLABLE-AF54
        '44885' => '9954'             , # G+9954  U+AF55  uniAF55,  HANGUL SYLLABLE-AF55
        '44886' => '9955'             , # G+9955  U+AF56  uniAF56,  HANGUL SYLLABLE-AF56
        '44887' => '9956'             , # G+9956  U+AF57  uniAF57,  HANGUL SYLLABLE-AF57
        '44888' => '9957'             , # G+9957  U+AF58  uniAF58,  HANGUL SYLLABLE-AF58
        '44889' => '9958'             , # G+9958  U+AF59  uniAF59,  HANGUL SYLLABLE-AF59
        '44890' => '9959'             , # G+9959  U+AF5A  uniAF5A,  HANGUL SYLLABLE-AF5A
        '44891' => '9960'             , # G+9960  U+AF5B  uniAF5B,  HANGUL SYLLABLE-AF5B
        '44892' => '1318'             , # G+1318  U+AF5C  uniAF5C,  HANGUL SYLLABLE-AF5C
        '44893' => '1319'             , # G+1319  U+AF5D  uniAF5D,  HANGUL SYLLABLE-AF5D
        '44894' => '9961'             , # G+9961  U+AF5E  uniAF5E,  HANGUL SYLLABLE-AF5E
        '44895' => '9962'             , # G+9962  U+AF5F  uniAF5F,  HANGUL SYLLABLE-AF5F
        '44896' => '9963'             , # G+9963  U+AF60  uniAF60,  HANGUL SYLLABLE-AF60
        '44897' => '9964'             , # G+9964  U+AF61  uniAF61,  HANGUL SYLLABLE-AF61
        '44898' => '9965'             , # G+9965  U+AF62  uniAF62,  HANGUL SYLLABLE-AF62
        '44899' => '9966'             , # G+9966  U+AF63  uniAF63,  HANGUL SYLLABLE-AF63
        '44900' => '1320'             , # G+1320  U+AF64  uniAF64,  HANGUL SYLLABLE-AF64
        '44901' => '1321'             , # G+1321  U+AF65  uniAF65,  HANGUL SYLLABLE-AF65
        '44902' => '9967'             , # G+9967  U+AF66  uniAF66,  HANGUL SYLLABLE-AF66
        '44903' => '9968'             , # G+9968  U+AF67  uniAF67,  HANGUL SYLLABLE-AF67
        '44904' => '9969'             , # G+9969  U+AF68  uniAF68,  HANGUL SYLLABLE-AF68
        '44905' => '9970'             , # G+9970  U+AF69  uniAF69,  HANGUL SYLLABLE-AF69
        '44906' => '9971'             , # G+9971  U+AF6A  uniAF6A,  HANGUL SYLLABLE-AF6A
        '44907' => '9972'             , # G+9972  U+AF6B  uniAF6B,  HANGUL SYLLABLE-AF6B
        '44908' => '9973'             , # G+9973  U+AF6C  uniAF6C,  HANGUL SYLLABLE-AF6C
        '44909' => '9974'             , # G+9974  U+AF6D  uniAF6D,  HANGUL SYLLABLE-AF6D
        '44910' => '9975'             , # G+9975  U+AF6E  uniAF6E,  HANGUL SYLLABLE-AF6E
        '44911' => '9976'             , # G+9976  U+AF6F  uniAF6F,  HANGUL SYLLABLE-AF6F
        '44912' => '9977'             , # G+9977  U+AF70  uniAF70,  HANGUL SYLLABLE-AF70
        '44913' => '9978'             , # G+9978  U+AF71  uniAF71,  HANGUL SYLLABLE-AF71
        '44914' => '9979'             , # G+9979  U+AF72  uniAF72,  HANGUL SYLLABLE-AF72
        '44915' => '9980'             , # G+9980  U+AF73  uniAF73,  HANGUL SYLLABLE-AF73
        '44916' => '9981'             , # G+9981  U+AF74  uniAF74,  HANGUL SYLLABLE-AF74
        '44917' => '9982'             , # G+9982  U+AF75  uniAF75,  HANGUL SYLLABLE-AF75
        '44918' => '9983'             , # G+9983  U+AF76  uniAF76,  HANGUL SYLLABLE-AF76
        '44919' => '9984'             , # G+9984  U+AF77  uniAF77,  HANGUL SYLLABLE-AF77
        '44920' => '9985'             , # G+9985  U+AF78  uniAF78,  HANGUL SYLLABLE-AF78
        '44921' => '1322'             , # G+1322  U+AF79  uniAF79,  HANGUL SYLLABLE-AF79
        '44922' => '9986'             , # G+9986  U+AF7A  uniAF7A,  HANGUL SYLLABLE-AF7A
        '44923' => '9987'             , # G+9987  U+AF7B  uniAF7B,  HANGUL SYLLABLE-AF7B
        '44924' => '9988'             , # G+9988  U+AF7C  uniAF7C,  HANGUL SYLLABLE-AF7C
        '44925' => '9989'             , # G+9989  U+AF7D  uniAF7D,  HANGUL SYLLABLE-AF7D
        '44926' => '9990'             , # G+9990  U+AF7E  uniAF7E,  HANGUL SYLLABLE-AF7E
        '44927' => '9991'             , # G+9991  U+AF7F  uniAF7F,  HANGUL SYLLABLE-AF7F
        '44928' => '1323'             , # G+1323  U+AF80  uniAF80,  HANGUL SYLLABLE-AF80
        '44929' => '9992'             , # G+9992  U+AF81  uniAF81,  HANGUL SYLLABLE-AF81
        '44930' => '9993'             , # G+9993  U+AF82  uniAF82,  HANGUL SYLLABLE-AF82
        '44931' => '9994'             , # G+9994  U+AF83  uniAF83,  HANGUL SYLLABLE-AF83
        '44932' => '1324'             , # G+1324  U+AF84  uniAF84,  HANGUL SYLLABLE-AF84
        '44933' => '9995'             , # G+9995  U+AF85  uniAF85,  HANGUL SYLLABLE-AF85
        '44934' => '9996'             , # G+9996  U+AF86  uniAF86,  HANGUL SYLLABLE-AF86
        '44935' => '9997'             , # G+9997  U+AF87  uniAF87,  HANGUL SYLLABLE-AF87
        '44936' => '1325'             , # G+1325  U+AF88  uniAF88,  HANGUL SYLLABLE-AF88
        '44937' => '9998'             , # G+9998  U+AF89  uniAF89,  HANGUL SYLLABLE-AF89
        '44938' => '9999'             , # G+9999  U+AF8A  uniAF8A,  HANGUL SYLLABLE-AF8A
        '44939' => '10000'            , # G+10000  U+AF8B  uniAF8B,  HANGUL SYLLABLE-AF8B
        '44940' => '10001'            , # G+10001  U+AF8C  uniAF8C,  HANGUL SYLLABLE-AF8C
        '44941' => '10002'            , # G+10002  U+AF8D  uniAF8D,  HANGUL SYLLABLE-AF8D
        '44942' => '10003'            , # G+10003  U+AF8E  uniAF8E,  HANGUL SYLLABLE-AF8E
        '44943' => '10004'            , # G+10004  U+AF8F  uniAF8F,  HANGUL SYLLABLE-AF8F
        '44944' => '1326'             , # G+1326  U+AF90  uniAF90,  HANGUL SYLLABLE-AF90
        '44945' => '1327'             , # G+1327  U+AF91  uniAF91,  HANGUL SYLLABLE-AF91
        '44946' => '10005'            , # G+10005  U+AF92  uniAF92,  HANGUL SYLLABLE-AF92
        '44947' => '10006'            , # G+10006  U+AF93  uniAF93,  HANGUL SYLLABLE-AF93
        '44948' => '10007'            , # G+10007  U+AF94  uniAF94,  HANGUL SYLLABLE-AF94
        '44949' => '1328'             , # G+1328  U+AF95  uniAF95,  HANGUL SYLLABLE-AF95
        '44950' => '10008'            , # G+10008  U+AF96  uniAF96,  HANGUL SYLLABLE-AF96
        '44951' => '10009'            , # G+10009  U+AF97  uniAF97,  HANGUL SYLLABLE-AF97
        '44952' => '10010'            , # G+10010  U+AF98  uniAF98,  HANGUL SYLLABLE-AF98
        '44953' => '10011'            , # G+10011  U+AF99  uniAF99,  HANGUL SYLLABLE-AF99
        '44954' => '10012'            , # G+10012  U+AF9A  uniAF9A,  HANGUL SYLLABLE-AF9A
        '44955' => '10013'            , # G+10013  U+AF9B  uniAF9B,  HANGUL SYLLABLE-AF9B
        '44956' => '1329'             , # G+1329  U+AF9C  uniAF9C,  HANGUL SYLLABLE-AF9C
        '44957' => '10014'            , # G+10014  U+AF9D  uniAF9D,  HANGUL SYLLABLE-AF9D
        '44958' => '10015'            , # G+10015  U+AF9E  uniAF9E,  HANGUL SYLLABLE-AF9E
        '44959' => '10016'            , # G+10016  U+AF9F  uniAF9F,  HANGUL SYLLABLE-AF9F
        '44960' => '10017'            , # G+10017  U+AFA0  uniAFA0,  HANGUL SYLLABLE-AFA0
        '44961' => '10018'            , # G+10018  U+AFA1  uniAFA1,  HANGUL SYLLABLE-AFA1
        '44962' => '10019'            , # G+10019  U+AFA2  uniAFA2,  HANGUL SYLLABLE-AFA2
        '44963' => '10020'            , # G+10020  U+AFA3  uniAFA3,  HANGUL SYLLABLE-AFA3
        '44964' => '10021'            , # G+10021  U+AFA4  uniAFA4,  HANGUL SYLLABLE-AFA4
        '44965' => '10022'            , # G+10022  U+AFA5  uniAFA5,  HANGUL SYLLABLE-AFA5
        '44966' => '10023'            , # G+10023  U+AFA6  uniAFA6,  HANGUL SYLLABLE-AFA6
        '44967' => '10024'            , # G+10024  U+AFA7  uniAFA7,  HANGUL SYLLABLE-AFA7
        '44968' => '10025'            , # G+10025  U+AFA8  uniAFA8,  HANGUL SYLLABLE-AFA8
        '44969' => '10026'            , # G+10026  U+AFA9  uniAFA9,  HANGUL SYLLABLE-AFA9
        '44970' => '10027'            , # G+10027  U+AFAA  uniAFAA,  HANGUL SYLLABLE-AFAA
        '44971' => '10028'            , # G+10028  U+AFAB  uniAFAB,  HANGUL SYLLABLE-AFAB
        '44972' => '10029'            , # G+10029  U+AFAC  uniAFAC,  HANGUL SYLLABLE-AFAC
        '44973' => '10030'            , # G+10030  U+AFAD  uniAFAD,  HANGUL SYLLABLE-AFAD
        '44974' => '10031'            , # G+10031  U+AFAE  uniAFAE,  HANGUL SYLLABLE-AFAE
        '44975' => '10032'            , # G+10032  U+AFAF  uniAFAF,  HANGUL SYLLABLE-AFAF
        '44976' => '10033'            , # G+10033  U+AFB0  uniAFB0,  HANGUL SYLLABLE-AFB0
        '44977' => '10034'            , # G+10034  U+AFB1  uniAFB1,  HANGUL SYLLABLE-AFB1
        '44978' => '10035'            , # G+10035  U+AFB2  uniAFB2,  HANGUL SYLLABLE-AFB2
        '44979' => '10036'            , # G+10036  U+AFB3  uniAFB3,  HANGUL SYLLABLE-AFB3
        '44980' => '10037'            , # G+10037  U+AFB4  uniAFB4,  HANGUL SYLLABLE-AFB4
        '44981' => '10038'            , # G+10038  U+AFB5  uniAFB5,  HANGUL SYLLABLE-AFB5
        '44982' => '10039'            , # G+10039  U+AFB6  uniAFB6,  HANGUL SYLLABLE-AFB6
        '44983' => '10040'            , # G+10040  U+AFB7  uniAFB7,  HANGUL SYLLABLE-AFB7
        '44984' => '1330'             , # G+1330  U+AFB8  uniAFB8,  HANGUL SYLLABLE-AFB8
        '44985' => '1331'             , # G+1331  U+AFB9  uniAFB9,  HANGUL SYLLABLE-AFB9
        '44986' => '10041'            , # G+10041  U+AFBA  uniAFBA,  HANGUL SYLLABLE-AFBA
        '44987' => '10042'            , # G+10042  U+AFBB  uniAFBB,  HANGUL SYLLABLE-AFBB
        '44988' => '1332'             , # G+1332  U+AFBC  uniAFBC,  HANGUL SYLLABLE-AFBC
        '44989' => '10043'            , # G+10043  U+AFBD  uniAFBD,  HANGUL SYLLABLE-AFBD
        '44990' => '10044'            , # G+10044  U+AFBE  uniAFBE,  HANGUL SYLLABLE-AFBE
        '44991' => '10045'            , # G+10045  U+AFBF  uniAFBF,  HANGUL SYLLABLE-AFBF
        '44992' => '1333'             , # G+1333  U+AFC0  uniAFC0,  HANGUL SYLLABLE-AFC0
        '44993' => '10046'            , # G+10046  U+AFC1  uniAFC1,  HANGUL SYLLABLE-AFC1
        '44994' => '10047'            , # G+10047  U+AFC2  uniAFC2,  HANGUL SYLLABLE-AFC2
        '44995' => '10048'            , # G+10048  U+AFC3  uniAFC3,  HANGUL SYLLABLE-AFC3
        '44996' => '10049'            , # G+10049  U+AFC4  uniAFC4,  HANGUL SYLLABLE-AFC4
        '44997' => '10050'            , # G+10050  U+AFC5  uniAFC5,  HANGUL SYLLABLE-AFC5
        '44998' => '10051'            , # G+10051  U+AFC6  uniAFC6,  HANGUL SYLLABLE-AFC6
        '44999' => '1334'             , # G+1334  U+AFC7  uniAFC7,  HANGUL SYLLABLE-AFC7
        '45000' => '1335'             , # G+1335  U+AFC8  uniAFC8,  HANGUL SYLLABLE-AFC8
        '45001' => '1336'             , # G+1336  U+AFC9  uniAFC9,  HANGUL SYLLABLE-AFC9
        '45002' => '10052'            , # G+10052  U+AFCA  uniAFCA,  HANGUL SYLLABLE-AFCA
        '45003' => '1337'             , # G+1337  U+AFCB  uniAFCB,  HANGUL SYLLABLE-AFCB
        '45004' => '10053'            , # G+10053  U+AFCC  uniAFCC,  HANGUL SYLLABLE-AFCC
        '45005' => '1338'             , # G+1338  U+AFCD  uniAFCD,  HANGUL SYLLABLE-AFCD
        '45006' => '1339'             , # G+1339  U+AFCE  uniAFCE,  HANGUL SYLLABLE-AFCE
        '45007' => '10054'            , # G+10054  U+AFCF  uniAFCF,  HANGUL SYLLABLE-AFCF
        '45008' => '10055'            , # G+10055  U+AFD0  uniAFD0,  HANGUL SYLLABLE-AFD0
        '45009' => '10056'            , # G+10056  U+AFD1  uniAFD1,  HANGUL SYLLABLE-AFD1
        '45010' => '10057'            , # G+10057  U+AFD2  uniAFD2,  HANGUL SYLLABLE-AFD2
        '45011' => '10058'            , # G+10058  U+AFD3  uniAFD3,  HANGUL SYLLABLE-AFD3
        '45012' => '1340'             , # G+1340  U+AFD4  uniAFD4,  HANGUL SYLLABLE-AFD4
        '45013' => '10059'            , # G+10059  U+AFD5  uniAFD5,  HANGUL SYLLABLE-AFD5
        '45014' => '10060'            , # G+10060  U+AFD6  uniAFD6,  HANGUL SYLLABLE-AFD6
        '45015' => '10061'            , # G+10061  U+AFD7  uniAFD7,  HANGUL SYLLABLE-AFD7
        '45016' => '10062'            , # G+10062  U+AFD8  uniAFD8,  HANGUL SYLLABLE-AFD8
        '45017' => '10063'            , # G+10063  U+AFD9  uniAFD9,  HANGUL SYLLABLE-AFD9
        '45018' => '10064'            , # G+10064  U+AFDA  uniAFDA,  HANGUL SYLLABLE-AFDA
        '45019' => '10065'            , # G+10065  U+AFDB  uniAFDB,  HANGUL SYLLABLE-AFDB
        '45020' => '1341'             , # G+1341  U+AFDC  uniAFDC,  HANGUL SYLLABLE-AFDC
        '45021' => '10066'            , # G+10066  U+AFDD  uniAFDD,  HANGUL SYLLABLE-AFDD
        '45022' => '10067'            , # G+10067  U+AFDE  uniAFDE,  HANGUL SYLLABLE-AFDE
        '45023' => '10068'            , # G+10068  U+AFDF  uniAFDF,  HANGUL SYLLABLE-AFDF
        '45024' => '10069'            , # G+10069  U+AFE0  uniAFE0,  HANGUL SYLLABLE-AFE0
        '45025' => '10070'            , # G+10070  U+AFE1  uniAFE1,  HANGUL SYLLABLE-AFE1
        '45026' => '10071'            , # G+10071  U+AFE2  uniAFE2,  HANGUL SYLLABLE-AFE2
        '45027' => '10072'            , # G+10072  U+AFE3  uniAFE3,  HANGUL SYLLABLE-AFE3
        '45028' => '10073'            , # G+10073  U+AFE4  uniAFE4,  HANGUL SYLLABLE-AFE4
        '45029' => '10074'            , # G+10074  U+AFE5  uniAFE5,  HANGUL SYLLABLE-AFE5
        '45030' => '10075'            , # G+10075  U+AFE6  uniAFE6,  HANGUL SYLLABLE-AFE6
        '45031' => '10076'            , # G+10076  U+AFE7  uniAFE7,  HANGUL SYLLABLE-AFE7
        '45032' => '1342'             , # G+1342  U+AFE8  uniAFE8,  HANGUL SYLLABLE-AFE8
        '45033' => '1343'             , # G+1343  U+AFE9  uniAFE9,  HANGUL SYLLABLE-AFE9
        '45034' => '10077'            , # G+10077  U+AFEA  uniAFEA,  HANGUL SYLLABLE-AFEA
        '45035' => '10078'            , # G+10078  U+AFEB  uniAFEB,  HANGUL SYLLABLE-AFEB
        '45036' => '10079'            , # G+10079  U+AFEC  uniAFEC,  HANGUL SYLLABLE-AFEC
        '45037' => '10080'            , # G+10080  U+AFED  uniAFED,  HANGUL SYLLABLE-AFED
        '45038' => '10081'            , # G+10081  U+AFEE  uniAFEE,  HANGUL SYLLABLE-AFEE
        '45039' => '10082'            , # G+10082  U+AFEF  uniAFEF,  HANGUL SYLLABLE-AFEF
        '45040' => '1344'             , # G+1344  U+AFF0  uniAFF0,  HANGUL SYLLABLE-AFF0
        '45041' => '1345'             , # G+1345  U+AFF1  uniAFF1,  HANGUL SYLLABLE-AFF1
        '45042' => '10083'            , # G+10083  U+AFF2  uniAFF2,  HANGUL SYLLABLE-AFF2
        '45043' => '10084'            , # G+10084  U+AFF3  uniAFF3,  HANGUL SYLLABLE-AFF3
        '45044' => '1346'             , # G+1346  U+AFF4  uniAFF4,  HANGUL SYLLABLE-AFF4
        '45045' => '10085'            , # G+10085  U+AFF5  uniAFF5,  HANGUL SYLLABLE-AFF5
        '45046' => '10086'            , # G+10086  U+AFF6  uniAFF6,  HANGUL SYLLABLE-AFF6
        '45047' => '10087'            , # G+10087  U+AFF7  uniAFF7,  HANGUL SYLLABLE-AFF7
        '45048' => '1347'             , # G+1347  U+AFF8  uniAFF8,  HANGUL SYLLABLE-AFF8
        '45049' => '10088'            , # G+10088  U+AFF9  uniAFF9,  HANGUL SYLLABLE-AFF9
        '45050' => '10089'            , # G+10089  U+AFFA  uniAFFA,  HANGUL SYLLABLE-AFFA
        '45051' => '10090'            , # G+10090  U+AFFB  uniAFFB,  HANGUL SYLLABLE-AFFB
        '45052' => '10091'            , # G+10091  U+AFFC  uniAFFC,  HANGUL SYLLABLE-AFFC
        '45053' => '10092'            , # G+10092  U+AFFD  uniAFFD,  HANGUL SYLLABLE-AFFD
        '45054' => '10093'            , # G+10093  U+AFFE  uniAFFE,  HANGUL SYLLABLE-AFFE
        '45055' => '10094'            , # G+10094  U+AFFF  uniAFFF,  HANGUL SYLLABLE-AFFF
        '45056' => '1348'             , # G+1348  U+B000  uniB000,  HANGUL SYLLABLE-B000
        '45057' => '1349'             , # G+1349  U+B001  uniB001,  HANGUL SYLLABLE-B001
        '45058' => '10095'            , # G+10095  U+B002  uniB002,  HANGUL SYLLABLE-B002
        '45059' => '10096'            , # G+10096  U+B003  uniB003,  HANGUL SYLLABLE-B003
        '45060' => '1350'             , # G+1350  U+B004  uniB004,  HANGUL SYLLABLE-B004
        '45061' => '10097'            , # G+10097  U+B005  uniB005,  HANGUL SYLLABLE-B005
        '45062' => '10098'            , # G+10098  U+B006  uniB006,  HANGUL SYLLABLE-B006
        '45063' => '10099'            , # G+10099  U+B007  uniB007,  HANGUL SYLLABLE-B007
        '45064' => '10100'            , # G+10100  U+B008  uniB008,  HANGUL SYLLABLE-B008
        '45065' => '10101'            , # G+10101  U+B009  uniB009,  HANGUL SYLLABLE-B009
        '45066' => '10102'            , # G+10102  U+B00A  uniB00A,  HANGUL SYLLABLE-B00A
        '45067' => '10103'            , # G+10103  U+B00B  uniB00B,  HANGUL SYLLABLE-B00B
        '45068' => '1351'             , # G+1351  U+B00C  uniB00C,  HANGUL SYLLABLE-B00C
        '45069' => '10104'            , # G+10104  U+B00D  uniB00D,  HANGUL SYLLABLE-B00D
        '45070' => '10105'            , # G+10105  U+B00E  uniB00E,  HANGUL SYLLABLE-B00E
        '45071' => '10106'            , # G+10106  U+B00F  uniB00F,  HANGUL SYLLABLE-B00F
        '45072' => '1352'             , # G+1352  U+B010  uniB010,  HANGUL SYLLABLE-B010
        '45073' => '10107'            , # G+10107  U+B011  uniB011,  HANGUL SYLLABLE-B011
        '45074' => '10108'            , # G+10108  U+B012  uniB012,  HANGUL SYLLABLE-B012
        '45075' => '10109'            , # G+10109  U+B013  uniB013,  HANGUL SYLLABLE-B013
        '45076' => '1353'             , # G+1353  U+B014  uniB014,  HANGUL SYLLABLE-B014
        '45077' => '10110'            , # G+10110  U+B015  uniB015,  HANGUL SYLLABLE-B015
        '45078' => '10111'            , # G+10111  U+B016  uniB016,  HANGUL SYLLABLE-B016
        '45079' => '10112'            , # G+10112  U+B017  uniB017,  HANGUL SYLLABLE-B017
        '45080' => '10113'            , # G+10113  U+B018  uniB018,  HANGUL SYLLABLE-B018
        '45081' => '10114'            , # G+10114  U+B019  uniB019,  HANGUL SYLLABLE-B019
        '45082' => '10115'            , # G+10115  U+B01A  uniB01A,  HANGUL SYLLABLE-B01A
        '45083' => '10116'            , # G+10116  U+B01B  uniB01B,  HANGUL SYLLABLE-B01B
        '45084' => '1354'             , # G+1354  U+B01C  uniB01C,  HANGUL SYLLABLE-B01C
        '45085' => '1355'             , # G+1355  U+B01D  uniB01D,  HANGUL SYLLABLE-B01D
        '45086' => '10117'            , # G+10117  U+B01E  uniB01E,  HANGUL SYLLABLE-B01E
        '45087' => '10118'            , # G+10118  U+B01F  uniB01F,  HANGUL SYLLABLE-B01F
        '45088' => '10119'            , # G+10119  U+B020  uniB020,  HANGUL SYLLABLE-B020
        '45089' => '10120'            , # G+10120  U+B021  uniB021,  HANGUL SYLLABLE-B021
        '45090' => '10121'            , # G+10121  U+B022  uniB022,  HANGUL SYLLABLE-B022
        '45091' => '10122'            , # G+10122  U+B023  uniB023,  HANGUL SYLLABLE-B023
        '45092' => '10123'            , # G+10123  U+B024  uniB024,  HANGUL SYLLABLE-B024
        '45093' => '10124'            , # G+10124  U+B025  uniB025,  HANGUL SYLLABLE-B025
        '45094' => '10125'            , # G+10125  U+B026  uniB026,  HANGUL SYLLABLE-B026
        '45095' => '10126'            , # G+10126  U+B027  uniB027,  HANGUL SYLLABLE-B027
        '45096' => '1356'             , # G+1356  U+B028  uniB028,  HANGUL SYLLABLE-B028
        '45097' => '10127'            , # G+10127  U+B029  uniB029,  HANGUL SYLLABLE-B029
        '45098' => '10128'            , # G+10128  U+B02A  uniB02A,  HANGUL SYLLABLE-B02A
        '45099' => '10129'            , # G+10129  U+B02B  uniB02B,  HANGUL SYLLABLE-B02B
        '45100' => '10130'            , # G+10130  U+B02C  uniB02C,  HANGUL SYLLABLE-B02C
        '45101' => '10131'            , # G+10131  U+B02D  uniB02D,  HANGUL SYLLABLE-B02D
        '45102' => '10132'            , # G+10132  U+B02E  uniB02E,  HANGUL SYLLABLE-B02E
        '45103' => '10133'            , # G+10133  U+B02F  uniB02F,  HANGUL SYLLABLE-B02F
        '45104' => '10134'            , # G+10134  U+B030  uniB030,  HANGUL SYLLABLE-B030
        '45105' => '10135'            , # G+10135  U+B031  uniB031,  HANGUL SYLLABLE-B031
        '45106' => '10136'            , # G+10136  U+B032  uniB032,  HANGUL SYLLABLE-B032
        '45107' => '10137'            , # G+10137  U+B033  uniB033,  HANGUL SYLLABLE-B033
        '45108' => '10138'            , # G+10138  U+B034  uniB034,  HANGUL SYLLABLE-B034
        '45109' => '10139'            , # G+10139  U+B035  uniB035,  HANGUL SYLLABLE-B035
        '45110' => '10140'            , # G+10140  U+B036  uniB036,  HANGUL SYLLABLE-B036
        '45111' => '10141'            , # G+10141  U+B037  uniB037,  HANGUL SYLLABLE-B037
        '45112' => '10142'            , # G+10142  U+B038  uniB038,  HANGUL SYLLABLE-B038
        '45113' => '10143'            , # G+10143  U+B039  uniB039,  HANGUL SYLLABLE-B039
        '45114' => '10144'            , # G+10144  U+B03A  uniB03A,  HANGUL SYLLABLE-B03A
        '45115' => '10145'            , # G+10145  U+B03B  uniB03B,  HANGUL SYLLABLE-B03B
        '45116' => '10146'            , # G+10146  U+B03C  uniB03C,  HANGUL SYLLABLE-B03C
        '45117' => '10147'            , # G+10147  U+B03D  uniB03D,  HANGUL SYLLABLE-B03D
        '45118' => '10148'            , # G+10148  U+B03E  uniB03E,  HANGUL SYLLABLE-B03E
        '45119' => '10149'            , # G+10149  U+B03F  uniB03F,  HANGUL SYLLABLE-B03F
        '45120' => '10150'            , # G+10150  U+B040  uniB040,  HANGUL SYLLABLE-B040
        '45121' => '10151'            , # G+10151  U+B041  uniB041,  HANGUL SYLLABLE-B041
        '45122' => '10152'            , # G+10152  U+B042  uniB042,  HANGUL SYLLABLE-B042
        '45123' => '10153'            , # G+10153  U+B043  uniB043,  HANGUL SYLLABLE-B043
        '45124' => '1357'             , # G+1357  U+B044  uniB044,  HANGUL SYLLABLE-B044
        '45125' => '1358'             , # G+1358  U+B045  uniB045,  HANGUL SYLLABLE-B045
        '45126' => '10154'            , # G+10154  U+B046  uniB046,  HANGUL SYLLABLE-B046
        '45127' => '10155'            , # G+10155  U+B047  uniB047,  HANGUL SYLLABLE-B047
        '45128' => '1359'             , # G+1359  U+B048  uniB048,  HANGUL SYLLABLE-B048
        '45129' => '10156'            , # G+10156  U+B049  uniB049,  HANGUL SYLLABLE-B049
        '45130' => '1360'             , # G+1360  U+B04A  uniB04A,  HANGUL SYLLABLE-B04A
        '45131' => '10157'            , # G+10157  U+B04B  uniB04B,  HANGUL SYLLABLE-B04B
        '45132' => '1361'             , # G+1361  U+B04C  uniB04C,  HANGUL SYLLABLE-B04C
        '45133' => '10158'            , # G+10158  U+B04D  uniB04D,  HANGUL SYLLABLE-B04D
        '45134' => '1362'             , # G+1362  U+B04E  uniB04E,  HANGUL SYLLABLE-B04E
        '45135' => '10159'            , # G+10159  U+B04F  uniB04F,  HANGUL SYLLABLE-B04F
        '45136' => '10160'            , # G+10160  U+B050  uniB050,  HANGUL SYLLABLE-B050
        '45137' => '10161'            , # G+10161  U+B051  uniB051,  HANGUL SYLLABLE-B051
        '45138' => '10162'            , # G+10162  U+B052  uniB052,  HANGUL SYLLABLE-B052
        '45139' => '1363'             , # G+1363  U+B053  uniB053,  HANGUL SYLLABLE-B053
        '45140' => '1364'             , # G+1364  U+B054  uniB054,  HANGUL SYLLABLE-B054
        '45141' => '1365'             , # G+1365  U+B055  uniB055,  HANGUL SYLLABLE-B055
        '45142' => '10163'            , # G+10163  U+B056  uniB056,  HANGUL SYLLABLE-B056
        '45143' => '1366'             , # G+1366  U+B057  uniB057,  HANGUL SYLLABLE-B057
        '45144' => '10164'            , # G+10164  U+B058  uniB058,  HANGUL SYLLABLE-B058
        '45145' => '1367'             , # G+1367  U+B059  uniB059,  HANGUL SYLLABLE-B059
        '45146' => '10165'            , # G+10165  U+B05A  uniB05A,  HANGUL SYLLABLE-B05A
        '45147' => '10166'            , # G+10166  U+B05B  uniB05B,  HANGUL SYLLABLE-B05B
        '45148' => '10167'            , # G+10167  U+B05C  uniB05C,  HANGUL SYLLABLE-B05C
        '45149' => '1368'             , # G+1368  U+B05D  uniB05D,  HANGUL SYLLABLE-B05D
        '45150' => '10168'            , # G+10168  U+B05E  uniB05E,  HANGUL SYLLABLE-B05E
        '45151' => '10169'            , # G+10169  U+B05F  uniB05F,  HANGUL SYLLABLE-B05F
        '45152' => '10170'            , # G+10170  U+B060  uniB060,  HANGUL SYLLABLE-B060
        '45153' => '10171'            , # G+10171  U+B061  uniB061,  HANGUL SYLLABLE-B061
        '45154' => '10172'            , # G+10172  U+B062  uniB062,  HANGUL SYLLABLE-B062
        '45155' => '10173'            , # G+10173  U+B063  uniB063,  HANGUL SYLLABLE-B063
        '45156' => '10174'            , # G+10174  U+B064  uniB064,  HANGUL SYLLABLE-B064
        '45157' => '10175'            , # G+10175  U+B065  uniB065,  HANGUL SYLLABLE-B065
        '45158' => '10176'            , # G+10176  U+B066  uniB066,  HANGUL SYLLABLE-B066
        '45159' => '10177'            , # G+10177  U+B067  uniB067,  HANGUL SYLLABLE-B067
        '45160' => '10178'            , # G+10178  U+B068  uniB068,  HANGUL SYLLABLE-B068
        '45161' => '10179'            , # G+10179  U+B069  uniB069,  HANGUL SYLLABLE-B069
        '45162' => '10180'            , # G+10180  U+B06A  uniB06A,  HANGUL SYLLABLE-B06A
        '45163' => '10181'            , # G+10181  U+B06B  uniB06B,  HANGUL SYLLABLE-B06B
        '45164' => '10182'            , # G+10182  U+B06C  uniB06C,  HANGUL SYLLABLE-B06C
        '45165' => '10183'            , # G+10183  U+B06D  uniB06D,  HANGUL SYLLABLE-B06D
        '45166' => '10184'            , # G+10184  U+B06E  uniB06E,  HANGUL SYLLABLE-B06E
        '45167' => '10185'            , # G+10185  U+B06F  uniB06F,  HANGUL SYLLABLE-B06F
        '45168' => '10186'            , # G+10186  U+B070  uniB070,  HANGUL SYLLABLE-B070
        '45169' => '10187'            , # G+10187  U+B071  uniB071,  HANGUL SYLLABLE-B071
        '45170' => '10188'            , # G+10188  U+B072  uniB072,  HANGUL SYLLABLE-B072
        '45171' => '10189'            , # G+10189  U+B073  uniB073,  HANGUL SYLLABLE-B073
        '45172' => '10190'            , # G+10190  U+B074  uniB074,  HANGUL SYLLABLE-B074
        '45173' => '10191'            , # G+10191  U+B075  uniB075,  HANGUL SYLLABLE-B075
        '45174' => '10192'            , # G+10192  U+B076  uniB076,  HANGUL SYLLABLE-B076
        '45175' => '10193'            , # G+10193  U+B077  uniB077,  HANGUL SYLLABLE-B077
        '45176' => '10194'            , # G+10194  U+B078  uniB078,  HANGUL SYLLABLE-B078
        '45177' => '10195'            , # G+10195  U+B079  uniB079,  HANGUL SYLLABLE-B079
        '45178' => '10196'            , # G+10196  U+B07A  uniB07A,  HANGUL SYLLABLE-B07A
        '45179' => '10197'            , # G+10197  U+B07B  uniB07B,  HANGUL SYLLABLE-B07B
        '45180' => '1369'             , # G+1369  U+B07C  uniB07C,  HANGUL SYLLABLE-B07C
        '45181' => '1370'             , # G+1370  U+B07D  uniB07D,  HANGUL SYLLABLE-B07D
        '45182' => '10198'            , # G+10198  U+B07E  uniB07E,  HANGUL SYLLABLE-B07E
        '45183' => '10199'            , # G+10199  U+B07F  uniB07F,  HANGUL SYLLABLE-B07F
        '45184' => '1371'             , # G+1371  U+B080  uniB080,  HANGUL SYLLABLE-B080
        '45185' => '10200'            , # G+10200  U+B081  uniB081,  HANGUL SYLLABLE-B081
        '45186' => '10201'            , # G+10201  U+B082  uniB082,  HANGUL SYLLABLE-B082
        '45187' => '10202'            , # G+10202  U+B083  uniB083,  HANGUL SYLLABLE-B083
        '45188' => '1372'             , # G+1372  U+B084  uniB084,  HANGUL SYLLABLE-B084
        '45189' => '10203'            , # G+10203  U+B085  uniB085,  HANGUL SYLLABLE-B085
        '45190' => '10204'            , # G+10204  U+B086  uniB086,  HANGUL SYLLABLE-B086
        '45191' => '10205'            , # G+10205  U+B087  uniB087,  HANGUL SYLLABLE-B087
        '45192' => '10206'            , # G+10206  U+B088  uniB088,  HANGUL SYLLABLE-B088
        '45193' => '10207'            , # G+10207  U+B089  uniB089,  HANGUL SYLLABLE-B089
        '45194' => '10208'            , # G+10208  U+B08A  uniB08A,  HANGUL SYLLABLE-B08A
        '45195' => '10209'            , # G+10209  U+B08B  uniB08B,  HANGUL SYLLABLE-B08B
        '45196' => '1373'             , # G+1373  U+B08C  uniB08C,  HANGUL SYLLABLE-B08C
        '45197' => '1374'             , # G+1374  U+B08D  uniB08D,  HANGUL SYLLABLE-B08D
        '45198' => '10210'            , # G+10210  U+B08E  uniB08E,  HANGUL SYLLABLE-B08E
        '45199' => '1375'             , # G+1375  U+B08F  uniB08F,  HANGUL SYLLABLE-B08F
        '45200' => '10211'            , # G+10211  U+B090  uniB090,  HANGUL SYLLABLE-B090
        '45201' => '1376'             , # G+1376  U+B091  uniB091,  HANGUL SYLLABLE-B091
        '45202' => '10212'            , # G+10212  U+B092  uniB092,  HANGUL SYLLABLE-B092
        '45203' => '10213'            , # G+10213  U+B093  uniB093,  HANGUL SYLLABLE-B093
        '45204' => '10214'            , # G+10214  U+B094  uniB094,  HANGUL SYLLABLE-B094
        '45205' => '10215'            , # G+10215  U+B095  uniB095,  HANGUL SYLLABLE-B095
        '45206' => '10216'            , # G+10216  U+B096  uniB096,  HANGUL SYLLABLE-B096
        '45207' => '10217'            , # G+10217  U+B097  uniB097,  HANGUL SYLLABLE-B097
        '45208' => '1377'             , # G+1377  U+B098  uniB098,  HANGUL SYLLABLE-B098
        '45209' => '1378'             , # G+1378  U+B099  uniB099,  HANGUL SYLLABLE-B099
        '45210' => '1379'             , # G+1379  U+B09A  uniB09A,  HANGUL SYLLABLE-B09A
        '45211' => '10218'            , # G+10218  U+B09B  uniB09B,  HANGUL SYLLABLE-B09B
        '45212' => '1380'             , # G+1380  U+B09C  uniB09C,  HANGUL SYLLABLE-B09C
        '45213' => '10219'            , # G+10219  U+B09D  uniB09D,  HANGUL SYLLABLE-B09D
        '45214' => '10220'            , # G+10220  U+B09E  uniB09E,  HANGUL SYLLABLE-B09E
        '45215' => '1381'             , # G+1381  U+B09F  uniB09F,  HANGUL SYLLABLE-B09F
        '45216' => '1382'             , # G+1382  U+B0A0  uniB0A0,  HANGUL SYLLABLE-B0A0
        '45217' => '1383'             , # G+1383  U+B0A1  uniB0A1,  HANGUL SYLLABLE-B0A1
        '45218' => '1384'             , # G+1384  U+B0A2  uniB0A2,  HANGUL SYLLABLE-B0A2
        '45219' => '10221'            , # G+10221  U+B0A3  uniB0A3,  HANGUL SYLLABLE-B0A3
        '45220' => '10222'            , # G+10222  U+B0A4  uniB0A4,  HANGUL SYLLABLE-B0A4
        '45221' => '10223'            , # G+10223  U+B0A5  uniB0A5,  HANGUL SYLLABLE-B0A5
        '45222' => '10224'            , # G+10224  U+B0A6  uniB0A6,  HANGUL SYLLABLE-B0A6
        '45223' => '10225'            , # G+10225  U+B0A7  uniB0A7,  HANGUL SYLLABLE-B0A7
        '45224' => '1385'             , # G+1385  U+B0A8  uniB0A8,  HANGUL SYLLABLE-B0A8
        '45225' => '1386'             , # G+1386  U+B0A9  uniB0A9,  HANGUL SYLLABLE-B0A9
        '45226' => '10226'            , # G+10226  U+B0AA  uniB0AA,  HANGUL SYLLABLE-B0AA
        '45227' => '1387'             , # G+1387  U+B0AB  uniB0AB,  HANGUL SYLLABLE-B0AB
        '45228' => '1388'             , # G+1388  U+B0AC  uniB0AC,  HANGUL SYLLABLE-B0AC
        '45229' => '1389'             , # G+1389  U+B0AD  uniB0AD,  HANGUL SYLLABLE-B0AD
        '45230' => '1390'             , # G+1390  U+B0AE  uniB0AE,  HANGUL SYLLABLE-B0AE
        '45231' => '1391'             , # G+1391  U+B0AF  uniB0AF,  HANGUL SYLLABLE-B0AF
        '45232' => '10227'            , # G+10227  U+B0B0  uniB0B0,  HANGUL SYLLABLE-B0B0
        '45233' => '1392'             , # G+1392  U+B0B1  uniB0B1,  HANGUL SYLLABLE-B0B1
        '45234' => '10228'            , # G+10228  U+B0B2  uniB0B2,  HANGUL SYLLABLE-B0B2
        '45235' => '1393'             , # G+1393  U+B0B3  uniB0B3,  HANGUL SYLLABLE-B0B3
        '45236' => '1394'             , # G+1394  U+B0B4  uniB0B4,  HANGUL SYLLABLE-B0B4
        '45237' => '1395'             , # G+1395  U+B0B5  uniB0B5,  HANGUL SYLLABLE-B0B5
        '45238' => '10229'            , # G+10229  U+B0B6  uniB0B6,  HANGUL SYLLABLE-B0B6
        '45239' => '10230'            , # G+10230  U+B0B7  uniB0B7,  HANGUL SYLLABLE-B0B7
        '45240' => '1396'             , # G+1396  U+B0B8  uniB0B8,  HANGUL SYLLABLE-B0B8
        '45241' => '10231'            , # G+10231  U+B0B9  uniB0B9,  HANGUL SYLLABLE-B0B9
        '45242' => '10232'            , # G+10232  U+B0BA  uniB0BA,  HANGUL SYLLABLE-B0BA
        '45243' => '10233'            , # G+10233  U+B0BB  uniB0BB,  HANGUL SYLLABLE-B0BB
        '45244' => '1397'             , # G+1397  U+B0BC  uniB0BC,  HANGUL SYLLABLE-B0BC
        '45245' => '10234'            , # G+10234  U+B0BD  uniB0BD,  HANGUL SYLLABLE-B0BD
        '45246' => '10235'            , # G+10235  U+B0BE  uniB0BE,  HANGUL SYLLABLE-B0BE
        '45247' => '10236'            , # G+10236  U+B0BF  uniB0BF,  HANGUL SYLLABLE-B0BF
        '45248' => '10237'            , # G+10237  U+B0C0  uniB0C0,  HANGUL SYLLABLE-B0C0
        '45249' => '10238'            , # G+10238  U+B0C1  uniB0C1,  HANGUL SYLLABLE-B0C1
        '45250' => '10239'            , # G+10239  U+B0C2  uniB0C2,  HANGUL SYLLABLE-B0C2
        '45251' => '10240'            , # G+10240  U+B0C3  uniB0C3,  HANGUL SYLLABLE-B0C3
        '45252' => '1398'             , # G+1398  U+B0C4  uniB0C4,  HANGUL SYLLABLE-B0C4
        '45253' => '1399'             , # G+1399  U+B0C5  uniB0C5,  HANGUL SYLLABLE-B0C5
        '45254' => '10241'            , # G+10241  U+B0C6  uniB0C6,  HANGUL SYLLABLE-B0C6
        '45255' => '1400'             , # G+1400  U+B0C7  uniB0C7,  HANGUL SYLLABLE-B0C7
        '45256' => '1401'             , # G+1401  U+B0C8  uniB0C8,  HANGUL SYLLABLE-B0C8
        '45257' => '1402'             , # G+1402  U+B0C9  uniB0C9,  HANGUL SYLLABLE-B0C9
        '45258' => '10242'            , # G+10242  U+B0CA  uniB0CA,  HANGUL SYLLABLE-B0CA
        '45259' => '10243'            , # G+10243  U+B0CB  uniB0CB,  HANGUL SYLLABLE-B0CB
        '45260' => '10244'            , # G+10244  U+B0CC  uniB0CC,  HANGUL SYLLABLE-B0CC
        '45261' => '10245'            , # G+10245  U+B0CD  uniB0CD,  HANGUL SYLLABLE-B0CD
        '45262' => '10246'            , # G+10246  U+B0CE  uniB0CE,  HANGUL SYLLABLE-B0CE
        '45263' => '10247'            , # G+10247  U+B0CF  uniB0CF,  HANGUL SYLLABLE-B0CF
        '45264' => '1403'             , # G+1403  U+B0D0  uniB0D0,  HANGUL SYLLABLE-B0D0
        '45265' => '1404'             , # G+1404  U+B0D1  uniB0D1,  HANGUL SYLLABLE-B0D1
        '45266' => '10248'            , # G+10248  U+B0D2  uniB0D2,  HANGUL SYLLABLE-B0D2
        '45267' => '10249'            , # G+10249  U+B0D3  uniB0D3,  HANGUL SYLLABLE-B0D3
        '45268' => '1405'             , # G+1405  U+B0D4  uniB0D4,  HANGUL SYLLABLE-B0D4
        '45269' => '10250'            , # G+10250  U+B0D5  uniB0D5,  HANGUL SYLLABLE-B0D5
        '45270' => '10251'            , # G+10251  U+B0D6  uniB0D6,  HANGUL SYLLABLE-B0D6
        '45271' => '10252'            , # G+10252  U+B0D7  uniB0D7,  HANGUL SYLLABLE-B0D7
        '45272' => '1406'             , # G+1406  U+B0D8  uniB0D8,  HANGUL SYLLABLE-B0D8
        '45273' => '10253'            , # G+10253  U+B0D9  uniB0D9,  HANGUL SYLLABLE-B0D9
        '45274' => '10254'            , # G+10254  U+B0DA  uniB0DA,  HANGUL SYLLABLE-B0DA
        '45275' => '10255'            , # G+10255  U+B0DB  uniB0DB,  HANGUL SYLLABLE-B0DB
        '45276' => '10256'            , # G+10256  U+B0DC  uniB0DC,  HANGUL SYLLABLE-B0DC
        '45277' => '10257'            , # G+10257  U+B0DD  uniB0DD,  HANGUL SYLLABLE-B0DD
        '45278' => '10258'            , # G+10258  U+B0DE  uniB0DE,  HANGUL SYLLABLE-B0DE
        '45279' => '10259'            , # G+10259  U+B0DF  uniB0DF,  HANGUL SYLLABLE-B0DF
        '45280' => '1407'             , # G+1407  U+B0E0  uniB0E0,  HANGUL SYLLABLE-B0E0
        '45281' => '10260'            , # G+10260  U+B0E1  uniB0E1,  HANGUL SYLLABLE-B0E1
        '45282' => '10261'            , # G+10261  U+B0E2  uniB0E2,  HANGUL SYLLABLE-B0E2
        '45283' => '10262'            , # G+10262  U+B0E3  uniB0E3,  HANGUL SYLLABLE-B0E3
        '45284' => '10263'            , # G+10263  U+B0E4  uniB0E4,  HANGUL SYLLABLE-B0E4
        '45285' => '1408'             , # G+1408  U+B0E5  uniB0E5,  HANGUL SYLLABLE-B0E5
        '45286' => '10264'            , # G+10264  U+B0E6  uniB0E6,  HANGUL SYLLABLE-B0E6
        '45287' => '10265'            , # G+10265  U+B0E7  uniB0E7,  HANGUL SYLLABLE-B0E7
        '45288' => '10266'            , # G+10266  U+B0E8  uniB0E8,  HANGUL SYLLABLE-B0E8
        '45289' => '10267'            , # G+10267  U+B0E9  uniB0E9,  HANGUL SYLLABLE-B0E9
        '45290' => '10268'            , # G+10268  U+B0EA  uniB0EA,  HANGUL SYLLABLE-B0EA
        '45291' => '10269'            , # G+10269  U+B0EB  uniB0EB,  HANGUL SYLLABLE-B0EB
        '45292' => '10270'            , # G+10270  U+B0EC  uniB0EC,  HANGUL SYLLABLE-B0EC
        '45293' => '10271'            , # G+10271  U+B0ED  uniB0ED,  HANGUL SYLLABLE-B0ED
        '45294' => '10272'            , # G+10272  U+B0EE  uniB0EE,  HANGUL SYLLABLE-B0EE
        '45295' => '10273'            , # G+10273  U+B0EF  uniB0EF,  HANGUL SYLLABLE-B0EF
        '45296' => '10274'            , # G+10274  U+B0F0  uniB0F0,  HANGUL SYLLABLE-B0F0
        '45297' => '10275'            , # G+10275  U+B0F1  uniB0F1,  HANGUL SYLLABLE-B0F1
        '45298' => '10276'            , # G+10276  U+B0F2  uniB0F2,  HANGUL SYLLABLE-B0F2
        '45299' => '10277'            , # G+10277  U+B0F3  uniB0F3,  HANGUL SYLLABLE-B0F3
        '45300' => '10278'            , # G+10278  U+B0F4  uniB0F4,  HANGUL SYLLABLE-B0F4
        '45301' => '10279'            , # G+10279  U+B0F5  uniB0F5,  HANGUL SYLLABLE-B0F5
        '45302' => '10280'            , # G+10280  U+B0F6  uniB0F6,  HANGUL SYLLABLE-B0F6
        '45303' => '10281'            , # G+10281  U+B0F7  uniB0F7,  HANGUL SYLLABLE-B0F7
        '45304' => '10282'            , # G+10282  U+B0F8  uniB0F8,  HANGUL SYLLABLE-B0F8
        '45305' => '10283'            , # G+10283  U+B0F9  uniB0F9,  HANGUL SYLLABLE-B0F9
        '45306' => '10284'            , # G+10284  U+B0FA  uniB0FA,  HANGUL SYLLABLE-B0FA
        '45307' => '10285'            , # G+10285  U+B0FB  uniB0FB,  HANGUL SYLLABLE-B0FB
        '45308' => '10286'            , # G+10286  U+B0FC  uniB0FC,  HANGUL SYLLABLE-B0FC
        '45309' => '10287'            , # G+10287  U+B0FD  uniB0FD,  HANGUL SYLLABLE-B0FD
        '45310' => '10288'            , # G+10288  U+B0FE  uniB0FE,  HANGUL SYLLABLE-B0FE
        '45311' => '10289'            , # G+10289  U+B0FF  uniB0FF,  HANGUL SYLLABLE-B0FF
        '45312' => '10290'            , # G+10290  U+B100  uniB100,  HANGUL SYLLABLE-B100
        '45313' => '10291'            , # G+10291  U+B101  uniB101,  HANGUL SYLLABLE-B101
        '45314' => '10292'            , # G+10292  U+B102  uniB102,  HANGUL SYLLABLE-B102
        '45315' => '10293'            , # G+10293  U+B103  uniB103,  HANGUL SYLLABLE-B103
        '45316' => '10294'            , # G+10294  U+B104  uniB104,  HANGUL SYLLABLE-B104
        '45317' => '10295'            , # G+10295  U+B105  uniB105,  HANGUL SYLLABLE-B105
        '45318' => '10296'            , # G+10296  U+B106  uniB106,  HANGUL SYLLABLE-B106
        '45319' => '10297'            , # G+10297  U+B107  uniB107,  HANGUL SYLLABLE-B107
        '45320' => '1409'             , # G+1409  U+B108  uniB108,  HANGUL SYLLABLE-B108
        '45321' => '1410'             , # G+1410  U+B109  uniB109,  HANGUL SYLLABLE-B109
        '45322' => '10298'            , # G+10298  U+B10A  uniB10A,  HANGUL SYLLABLE-B10A
        '45323' => '1411'             , # G+1411  U+B10B  uniB10B,  HANGUL SYLLABLE-B10B
        '45324' => '1412'             , # G+1412  U+B10C  uniB10C,  HANGUL SYLLABLE-B10C
        '45325' => '10299'            , # G+10299  U+B10D  uniB10D,  HANGUL SYLLABLE-B10D
        '45326' => '10300'            , # G+10300  U+B10E  uniB10E,  HANGUL SYLLABLE-B10E
        '45327' => '10301'            , # G+10301  U+B10F  uniB10F,  HANGUL SYLLABLE-B10F
        '45328' => '1413'             , # G+1413  U+B110  uniB110,  HANGUL SYLLABLE-B110
        '45329' => '10302'            , # G+10302  U+B111  uniB111,  HANGUL SYLLABLE-B111
        '45330' => '1414'             , # G+1414  U+B112  uniB112,  HANGUL SYLLABLE-B112
        '45331' => '1415'             , # G+1415  U+B113  uniB113,  HANGUL SYLLABLE-B113
        '45332' => '10303'            , # G+10303  U+B114  uniB114,  HANGUL SYLLABLE-B114
        '45333' => '10304'            , # G+10304  U+B115  uniB115,  HANGUL SYLLABLE-B115
        '45334' => '10305'            , # G+10305  U+B116  uniB116,  HANGUL SYLLABLE-B116
        '45335' => '10306'            , # G+10306  U+B117  uniB117,  HANGUL SYLLABLE-B117
        '45336' => '1416'             , # G+1416  U+B118  uniB118,  HANGUL SYLLABLE-B118
        '45337' => '1417'             , # G+1417  U+B119  uniB119,  HANGUL SYLLABLE-B119
        '45338' => '10307'            , # G+10307  U+B11A  uniB11A,  HANGUL SYLLABLE-B11A
        '45339' => '1418'             , # G+1418  U+B11B  uniB11B,  HANGUL SYLLABLE-B11B
        '45340' => '1419'             , # G+1419  U+B11C  uniB11C,  HANGUL SYLLABLE-B11C
        '45341' => '1420'             , # G+1420  U+B11D  uniB11D,  HANGUL SYLLABLE-B11D
        '45342' => '10308'            , # G+10308  U+B11E  uniB11E,  HANGUL SYLLABLE-B11E
        '45343' => '10309'            , # G+10309  U+B11F  uniB11F,  HANGUL SYLLABLE-B11F
        '45344' => '10310'            , # G+10310  U+B120  uniB120,  HANGUL SYLLABLE-B120
        '45345' => '10311'            , # G+10311  U+B121  uniB121,  HANGUL SYLLABLE-B121
        '45346' => '10312'            , # G+10312  U+B122  uniB122,  HANGUL SYLLABLE-B122
        '45347' => '1421'             , # G+1421  U+B123  uniB123,  HANGUL SYLLABLE-B123
        '45348' => '1422'             , # G+1422  U+B124  uniB124,  HANGUL SYLLABLE-B124
        '45349' => '1423'             , # G+1423  U+B125  uniB125,  HANGUL SYLLABLE-B125
        '45350' => '10313'            , # G+10313  U+B126  uniB126,  HANGUL SYLLABLE-B126
        '45351' => '10314'            , # G+10314  U+B127  uniB127,  HANGUL SYLLABLE-B127
        '45352' => '1424'             , # G+1424  U+B128  uniB128,  HANGUL SYLLABLE-B128
        '45353' => '10315'            , # G+10315  U+B129  uniB129,  HANGUL SYLLABLE-B129
        '45354' => '10316'            , # G+10316  U+B12A  uniB12A,  HANGUL SYLLABLE-B12A
        '45355' => '10317'            , # G+10317  U+B12B  uniB12B,  HANGUL SYLLABLE-B12B
        '45356' => '1425'             , # G+1425  U+B12C  uniB12C,  HANGUL SYLLABLE-B12C
        '45357' => '10318'            , # G+10318  U+B12D  uniB12D,  HANGUL SYLLABLE-B12D
        '45358' => '10319'            , # G+10319  U+B12E  uniB12E,  HANGUL SYLLABLE-B12E
        '45359' => '10320'            , # G+10320  U+B12F  uniB12F,  HANGUL SYLLABLE-B12F
        '45360' => '10321'            , # G+10321  U+B130  uniB130,  HANGUL SYLLABLE-B130
        '45361' => '10322'            , # G+10322  U+B131  uniB131,  HANGUL SYLLABLE-B131
        '45362' => '10323'            , # G+10323  U+B132  uniB132,  HANGUL SYLLABLE-B132
        '45363' => '10324'            , # G+10324  U+B133  uniB133,  HANGUL SYLLABLE-B133
        '45364' => '1426'             , # G+1426  U+B134  uniB134,  HANGUL SYLLABLE-B134
        '45365' => '1427'             , # G+1427  U+B135  uniB135,  HANGUL SYLLABLE-B135
        '45366' => '10325'            , # G+10325  U+B136  uniB136,  HANGUL SYLLABLE-B136
        '45367' => '1428'             , # G+1428  U+B137  uniB137,  HANGUL SYLLABLE-B137
        '45368' => '1429'             , # G+1429  U+B138  uniB138,  HANGUL SYLLABLE-B138
        '45369' => '1430'             , # G+1430  U+B139  uniB139,  HANGUL SYLLABLE-B139
        '45370' => '10326'            , # G+10326  U+B13A  uniB13A,  HANGUL SYLLABLE-B13A
        '45371' => '10327'            , # G+10327  U+B13B  uniB13B,  HANGUL SYLLABLE-B13B
        '45372' => '10328'            , # G+10328  U+B13C  uniB13C,  HANGUL SYLLABLE-B13C
        '45373' => '10329'            , # G+10329  U+B13D  uniB13D,  HANGUL SYLLABLE-B13D
        '45374' => '10330'            , # G+10330  U+B13E  uniB13E,  HANGUL SYLLABLE-B13E
        '45375' => '10331'            , # G+10331  U+B13F  uniB13F,  HANGUL SYLLABLE-B13F
        '45376' => '1431'             , # G+1431  U+B140  uniB140,  HANGUL SYLLABLE-B140
        '45377' => '1432'             , # G+1432  U+B141  uniB141,  HANGUL SYLLABLE-B141
        '45378' => '10332'            , # G+10332  U+B142  uniB142,  HANGUL SYLLABLE-B142
        '45379' => '10333'            , # G+10333  U+B143  uniB143,  HANGUL SYLLABLE-B143
        '45380' => '1433'             , # G+1433  U+B144  uniB144,  HANGUL SYLLABLE-B144
        '45381' => '10334'            , # G+10334  U+B145  uniB145,  HANGUL SYLLABLE-B145
        '45382' => '10335'            , # G+10335  U+B146  uniB146,  HANGUL SYLLABLE-B146
        '45383' => '10336'            , # G+10336  U+B147  uniB147,  HANGUL SYLLABLE-B147
        '45384' => '1434'             , # G+1434  U+B148  uniB148,  HANGUL SYLLABLE-B148
        '45385' => '10337'            , # G+10337  U+B149  uniB149,  HANGUL SYLLABLE-B149
        '45386' => '10338'            , # G+10338  U+B14A  uniB14A,  HANGUL SYLLABLE-B14A
        '45387' => '10339'            , # G+10339  U+B14B  uniB14B,  HANGUL SYLLABLE-B14B
        '45388' => '10340'            , # G+10340  U+B14C  uniB14C,  HANGUL SYLLABLE-B14C
        '45389' => '10341'            , # G+10341  U+B14D  uniB14D,  HANGUL SYLLABLE-B14D
        '45390' => '10342'            , # G+10342  U+B14E  uniB14E,  HANGUL SYLLABLE-B14E
        '45391' => '10343'            , # G+10343  U+B14F  uniB14F,  HANGUL SYLLABLE-B14F
        '45392' => '1435'             , # G+1435  U+B150  uniB150,  HANGUL SYLLABLE-B150
        '45393' => '1436'             , # G+1436  U+B151  uniB151,  HANGUL SYLLABLE-B151
        '45394' => '10344'            , # G+10344  U+B152  uniB152,  HANGUL SYLLABLE-B152
        '45395' => '10345'            , # G+10345  U+B153  uniB153,  HANGUL SYLLABLE-B153
        '45396' => '1437'             , # G+1437  U+B154  uniB154,  HANGUL SYLLABLE-B154
        '45397' => '1438'             , # G+1438  U+B155  uniB155,  HANGUL SYLLABLE-B155
        '45398' => '10346'            , # G+10346  U+B156  uniB156,  HANGUL SYLLABLE-B156
        '45399' => '10347'            , # G+10347  U+B157  uniB157,  HANGUL SYLLABLE-B157
        '45400' => '1439'             , # G+1439  U+B158  uniB158,  HANGUL SYLLABLE-B158
        '45401' => '10348'            , # G+10348  U+B159  uniB159,  HANGUL SYLLABLE-B159
        '45402' => '10349'            , # G+10349  U+B15A  uniB15A,  HANGUL SYLLABLE-B15A
        '45403' => '10350'            , # G+10350  U+B15B  uniB15B,  HANGUL SYLLABLE-B15B
        '45404' => '1440'             , # G+1440  U+B15C  uniB15C,  HANGUL SYLLABLE-B15C
        '45405' => '10351'            , # G+10351  U+B15D  uniB15D,  HANGUL SYLLABLE-B15D
        '45406' => '10352'            , # G+10352  U+B15E  uniB15E,  HANGUL SYLLABLE-B15E
        '45407' => '10353'            , # G+10353  U+B15F  uniB15F,  HANGUL SYLLABLE-B15F
        '45408' => '1441'             , # G+1441  U+B160  uniB160,  HANGUL SYLLABLE-B160
        '45409' => '10354'            , # G+10354  U+B161  uniB161,  HANGUL SYLLABLE-B161
        '45410' => '10355'            , # G+10355  U+B162  uniB162,  HANGUL SYLLABLE-B162
        '45411' => '10356'            , # G+10356  U+B163  uniB163,  HANGUL SYLLABLE-B163
        '45412' => '10357'            , # G+10357  U+B164  uniB164,  HANGUL SYLLABLE-B164
        '45413' => '10358'            , # G+10358  U+B165  uniB165,  HANGUL SYLLABLE-B165
        '45414' => '10359'            , # G+10359  U+B166  uniB166,  HANGUL SYLLABLE-B166
        '45415' => '10360'            , # G+10360  U+B167  uniB167,  HANGUL SYLLABLE-B167
        '45416' => '10361'            , # G+10361  U+B168  uniB168,  HANGUL SYLLABLE-B168
        '45417' => '10362'            , # G+10362  U+B169  uniB169,  HANGUL SYLLABLE-B169
        '45418' => '10363'            , # G+10363  U+B16A  uniB16A,  HANGUL SYLLABLE-B16A
        '45419' => '10364'            , # G+10364  U+B16B  uniB16B,  HANGUL SYLLABLE-B16B
        '45420' => '10365'            , # G+10365  U+B16C  uniB16C,  HANGUL SYLLABLE-B16C
        '45421' => '10366'            , # G+10366  U+B16D  uniB16D,  HANGUL SYLLABLE-B16D
        '45422' => '10367'            , # G+10367  U+B16E  uniB16E,  HANGUL SYLLABLE-B16E
        '45423' => '10368'            , # G+10368  U+B16F  uniB16F,  HANGUL SYLLABLE-B16F
        '45424' => '10369'            , # G+10369  U+B170  uniB170,  HANGUL SYLLABLE-B170
        '45425' => '10370'            , # G+10370  U+B171  uniB171,  HANGUL SYLLABLE-B171
        '45426' => '10371'            , # G+10371  U+B172  uniB172,  HANGUL SYLLABLE-B172
        '45427' => '10372'            , # G+10372  U+B173  uniB173,  HANGUL SYLLABLE-B173
        '45428' => '10373'            , # G+10373  U+B174  uniB174,  HANGUL SYLLABLE-B174
        '45429' => '10374'            , # G+10374  U+B175  uniB175,  HANGUL SYLLABLE-B175
        '45430' => '10375'            , # G+10375  U+B176  uniB176,  HANGUL SYLLABLE-B176
        '45431' => '10376'            , # G+10376  U+B177  uniB177,  HANGUL SYLLABLE-B177
        '45432' => '1442'             , # G+1442  U+B178  uniB178,  HANGUL SYLLABLE-B178
        '45433' => '1443'             , # G+1443  U+B179  uniB179,  HANGUL SYLLABLE-B179
        '45434' => '10377'            , # G+10377  U+B17A  uniB17A,  HANGUL SYLLABLE-B17A
        '45435' => '10378'            , # G+10378  U+B17B  uniB17B,  HANGUL SYLLABLE-B17B
        '45436' => '1444'             , # G+1444  U+B17C  uniB17C,  HANGUL SYLLABLE-B17C
        '45437' => '10379'            , # G+10379  U+B17D  uniB17D,  HANGUL SYLLABLE-B17D
        '45438' => '10380'            , # G+10380  U+B17E  uniB17E,  HANGUL SYLLABLE-B17E
        '45439' => '10381'            , # G+10381  U+B17F  uniB17F,  HANGUL SYLLABLE-B17F
        '45440' => '1445'             , # G+1445  U+B180  uniB180,  HANGUL SYLLABLE-B180
        '45441' => '10382'            , # G+10382  U+B181  uniB181,  HANGUL SYLLABLE-B181
        '45442' => '1446'             , # G+1446  U+B182  uniB182,  HANGUL SYLLABLE-B182
        '45443' => '10383'            , # G+10383  U+B183  uniB183,  HANGUL SYLLABLE-B183
        '45444' => '10384'            , # G+10384  U+B184  uniB184,  HANGUL SYLLABLE-B184
        '45445' => '10385'            , # G+10385  U+B185  uniB185,  HANGUL SYLLABLE-B185
        '45446' => '10386'            , # G+10386  U+B186  uniB186,  HANGUL SYLLABLE-B186
        '45447' => '10387'            , # G+10387  U+B187  uniB187,  HANGUL SYLLABLE-B187
        '45448' => '1447'             , # G+1447  U+B188  uniB188,  HANGUL SYLLABLE-B188
        '45449' => '1448'             , # G+1448  U+B189  uniB189,  HANGUL SYLLABLE-B189
        '45450' => '10388'            , # G+10388  U+B18A  uniB18A,  HANGUL SYLLABLE-B18A
        '45451' => '1449'             , # G+1449  U+B18B  uniB18B,  HANGUL SYLLABLE-B18B
        '45452' => '10389'            , # G+10389  U+B18C  uniB18C,  HANGUL SYLLABLE-B18C
        '45453' => '1450'             , # G+1450  U+B18D  uniB18D,  HANGUL SYLLABLE-B18D
        '45454' => '10390'            , # G+10390  U+B18E  uniB18E,  HANGUL SYLLABLE-B18E
        '45455' => '10391'            , # G+10391  U+B18F  uniB18F,  HANGUL SYLLABLE-B18F
        '45456' => '10392'            , # G+10392  U+B190  uniB190,  HANGUL SYLLABLE-B190
        '45457' => '10393'            , # G+10393  U+B191  uniB191,  HANGUL SYLLABLE-B191
        '45458' => '1451'             , # G+1451  U+B192  uniB192,  HANGUL SYLLABLE-B192
        '45459' => '1452'             , # G+1452  U+B193  uniB193,  HANGUL SYLLABLE-B193
        '45460' => '1453'             , # G+1453  U+B194  uniB194,  HANGUL SYLLABLE-B194
        '45461' => '10394'            , # G+10394  U+B195  uniB195,  HANGUL SYLLABLE-B195
        '45462' => '10395'            , # G+10395  U+B196  uniB196,  HANGUL SYLLABLE-B196
        '45463' => '10396'            , # G+10396  U+B197  uniB197,  HANGUL SYLLABLE-B197
        '45464' => '1454'             , # G+1454  U+B198  uniB198,  HANGUL SYLLABLE-B198
        '45465' => '10397'            , # G+10397  U+B199  uniB199,  HANGUL SYLLABLE-B199
        '45466' => '10398'            , # G+10398  U+B19A  uniB19A,  HANGUL SYLLABLE-B19A
        '45467' => '10399'            , # G+10399  U+B19B  uniB19B,  HANGUL SYLLABLE-B19B
        '45468' => '1455'             , # G+1455  U+B19C  uniB19C,  HANGUL SYLLABLE-B19C
        '45469' => '10400'            , # G+10400  U+B19D  uniB19D,  HANGUL SYLLABLE-B19D
        '45470' => '10401'            , # G+10401  U+B19E  uniB19E,  HANGUL SYLLABLE-B19E
        '45471' => '10402'            , # G+10402  U+B19F  uniB19F,  HANGUL SYLLABLE-B19F
        '45472' => '10403'            , # G+10403  U+B1A0  uniB1A0,  HANGUL SYLLABLE-B1A0
        '45473' => '10404'            , # G+10404  U+B1A1  uniB1A1,  HANGUL SYLLABLE-B1A1
        '45474' => '10405'            , # G+10405  U+B1A2  uniB1A2,  HANGUL SYLLABLE-B1A2
        '45475' => '10406'            , # G+10406  U+B1A3  uniB1A3,  HANGUL SYLLABLE-B1A3
        '45476' => '10407'            , # G+10407  U+B1A4  uniB1A4,  HANGUL SYLLABLE-B1A4
        '45477' => '10408'            , # G+10408  U+B1A5  uniB1A5,  HANGUL SYLLABLE-B1A5
        '45478' => '10409'            , # G+10409  U+B1A6  uniB1A6,  HANGUL SYLLABLE-B1A6
        '45479' => '10410'            , # G+10410  U+B1A7  uniB1A7,  HANGUL SYLLABLE-B1A7
        '45480' => '1456'             , # G+1456  U+B1A8  uniB1A8,  HANGUL SYLLABLE-B1A8
        '45481' => '10411'            , # G+10411  U+B1A9  uniB1A9,  HANGUL SYLLABLE-B1A9
        '45482' => '10412'            , # G+10412  U+B1AA  uniB1AA,  HANGUL SYLLABLE-B1AA
        '45483' => '10413'            , # G+10413  U+B1AB  uniB1AB,  HANGUL SYLLABLE-B1AB
        '45484' => '10414'            , # G+10414  U+B1AC  uniB1AC,  HANGUL SYLLABLE-B1AC
        '45485' => '10415'            , # G+10415  U+B1AD  uniB1AD,  HANGUL SYLLABLE-B1AD
        '45486' => '10416'            , # G+10416  U+B1AE  uniB1AE,  HANGUL SYLLABLE-B1AE
        '45487' => '10417'            , # G+10417  U+B1AF  uniB1AF,  HANGUL SYLLABLE-B1AF
        '45488' => '10418'            , # G+10418  U+B1B0  uniB1B0,  HANGUL SYLLABLE-B1B0
        '45489' => '10419'            , # G+10419  U+B1B1  uniB1B1,  HANGUL SYLLABLE-B1B1
        '45490' => '10420'            , # G+10420  U+B1B2  uniB1B2,  HANGUL SYLLABLE-B1B2
        '45491' => '10421'            , # G+10421  U+B1B3  uniB1B3,  HANGUL SYLLABLE-B1B3
        '45492' => '10422'            , # G+10422  U+B1B4  uniB1B4,  HANGUL SYLLABLE-B1B4
        '45493' => '10423'            , # G+10423  U+B1B5  uniB1B5,  HANGUL SYLLABLE-B1B5
        '45494' => '10424'            , # G+10424  U+B1B6  uniB1B6,  HANGUL SYLLABLE-B1B6
        '45495' => '10425'            , # G+10425  U+B1B7  uniB1B7,  HANGUL SYLLABLE-B1B7
        '45496' => '10426'            , # G+10426  U+B1B8  uniB1B8,  HANGUL SYLLABLE-B1B8
        '45497' => '10427'            , # G+10427  U+B1B9  uniB1B9,  HANGUL SYLLABLE-B1B9
        '45498' => '10428'            , # G+10428  U+B1BA  uniB1BA,  HANGUL SYLLABLE-B1BA
        '45499' => '10429'            , # G+10429  U+B1BB  uniB1BB,  HANGUL SYLLABLE-B1BB
        '45500' => '10430'            , # G+10430  U+B1BC  uniB1BC,  HANGUL SYLLABLE-B1BC
        '45501' => '10431'            , # G+10431  U+B1BD  uniB1BD,  HANGUL SYLLABLE-B1BD
        '45502' => '10432'            , # G+10432  U+B1BE  uniB1BE,  HANGUL SYLLABLE-B1BE
        '45503' => '10433'            , # G+10433  U+B1BF  uniB1BF,  HANGUL SYLLABLE-B1BF
        '45504' => '10434'            , # G+10434  U+B1C0  uniB1C0,  HANGUL SYLLABLE-B1C0
        '45505' => '10435'            , # G+10435  U+B1C1  uniB1C1,  HANGUL SYLLABLE-B1C1
        '45506' => '10436'            , # G+10436  U+B1C2  uniB1C2,  HANGUL SYLLABLE-B1C2
        '45507' => '10437'            , # G+10437  U+B1C3  uniB1C3,  HANGUL SYLLABLE-B1C3
        '45508' => '10438'            , # G+10438  U+B1C4  uniB1C4,  HANGUL SYLLABLE-B1C4
        '45509' => '10439'            , # G+10439  U+B1C5  uniB1C5,  HANGUL SYLLABLE-B1C5
        '45510' => '10440'            , # G+10440  U+B1C6  uniB1C6,  HANGUL SYLLABLE-B1C6
        '45511' => '10441'            , # G+10441  U+B1C7  uniB1C7,  HANGUL SYLLABLE-B1C7
        '45512' => '10442'            , # G+10442  U+B1C8  uniB1C8,  HANGUL SYLLABLE-B1C8
        '45513' => '10443'            , # G+10443  U+B1C9  uniB1C9,  HANGUL SYLLABLE-B1C9
        '45514' => '10444'            , # G+10444  U+B1CA  uniB1CA,  HANGUL SYLLABLE-B1CA
        '45515' => '10445'            , # G+10445  U+B1CB  uniB1CB,  HANGUL SYLLABLE-B1CB
        '45516' => '1457'             , # G+1457  U+B1CC  uniB1CC,  HANGUL SYLLABLE-B1CC
        '45517' => '10446'            , # G+10446  U+B1CD  uniB1CD,  HANGUL SYLLABLE-B1CD
        '45518' => '10447'            , # G+10447  U+B1CE  uniB1CE,  HANGUL SYLLABLE-B1CE
        '45519' => '10448'            , # G+10448  U+B1CF  uniB1CF,  HANGUL SYLLABLE-B1CF
        '45520' => '1458'             , # G+1458  U+B1D0  uniB1D0,  HANGUL SYLLABLE-B1D0
        '45521' => '10449'            , # G+10449  U+B1D1  uniB1D1,  HANGUL SYLLABLE-B1D1
        '45522' => '10450'            , # G+10450  U+B1D2  uniB1D2,  HANGUL SYLLABLE-B1D2
        '45523' => '10451'            , # G+10451  U+B1D3  uniB1D3,  HANGUL SYLLABLE-B1D3
        '45524' => '1459'             , # G+1459  U+B1D4  uniB1D4,  HANGUL SYLLABLE-B1D4
        '45525' => '10452'            , # G+10452  U+B1D5  uniB1D5,  HANGUL SYLLABLE-B1D5
        '45526' => '10453'            , # G+10453  U+B1D6  uniB1D6,  HANGUL SYLLABLE-B1D6
        '45527' => '10454'            , # G+10454  U+B1D7  uniB1D7,  HANGUL SYLLABLE-B1D7
        '45528' => '10455'            , # G+10455  U+B1D8  uniB1D8,  HANGUL SYLLABLE-B1D8
        '45529' => '10456'            , # G+10456  U+B1D9  uniB1D9,  HANGUL SYLLABLE-B1D9
        '45530' => '10457'            , # G+10457  U+B1DA  uniB1DA,  HANGUL SYLLABLE-B1DA
        '45531' => '10458'            , # G+10458  U+B1DB  uniB1DB,  HANGUL SYLLABLE-B1DB
        '45532' => '1460'             , # G+1460  U+B1DC  uniB1DC,  HANGUL SYLLABLE-B1DC
        '45533' => '1461'             , # G+1461  U+B1DD  uniB1DD,  HANGUL SYLLABLE-B1DD
        '45534' => '10459'            , # G+10459  U+B1DE  uniB1DE,  HANGUL SYLLABLE-B1DE
        '45535' => '1462'             , # G+1462  U+B1DF  uniB1DF,  HANGUL SYLLABLE-B1DF
        '45536' => '10460'            , # G+10460  U+B1E0  uniB1E0,  HANGUL SYLLABLE-B1E0
        '45537' => '10461'            , # G+10461  U+B1E1  uniB1E1,  HANGUL SYLLABLE-B1E1
        '45538' => '10462'            , # G+10462  U+B1E2  uniB1E2,  HANGUL SYLLABLE-B1E2
        '45539' => '10463'            , # G+10463  U+B1E3  uniB1E3,  HANGUL SYLLABLE-B1E3
        '45540' => '10464'            , # G+10464  U+B1E4  uniB1E4,  HANGUL SYLLABLE-B1E4
        '45541' => '10465'            , # G+10465  U+B1E5  uniB1E5,  HANGUL SYLLABLE-B1E5
        '45542' => '10466'            , # G+10466  U+B1E6  uniB1E6,  HANGUL SYLLABLE-B1E6
        '45543' => '10467'            , # G+10467  U+B1E7  uniB1E7,  HANGUL SYLLABLE-B1E7
        '45544' => '1463'             , # G+1463  U+B1E8  uniB1E8,  HANGUL SYLLABLE-B1E8
        '45545' => '1464'             , # G+1464  U+B1E9  uniB1E9,  HANGUL SYLLABLE-B1E9
        '45546' => '10468'            , # G+10468  U+B1EA  uniB1EA,  HANGUL SYLLABLE-B1EA
        '45547' => '10469'            , # G+10469  U+B1EB  uniB1EB,  HANGUL SYLLABLE-B1EB
        '45548' => '1465'             , # G+1465  U+B1EC  uniB1EC,  HANGUL SYLLABLE-B1EC
        '45549' => '10470'            , # G+10470  U+B1ED  uniB1ED,  HANGUL SYLLABLE-B1ED
        '45550' => '10471'            , # G+10471  U+B1EE  uniB1EE,  HANGUL SYLLABLE-B1EE
        '45551' => '10472'            , # G+10472  U+B1EF  uniB1EF,  HANGUL SYLLABLE-B1EF
        '45552' => '1466'             , # G+1466  U+B1F0  uniB1F0,  HANGUL SYLLABLE-B1F0
        '45553' => '10473'            , # G+10473  U+B1F1  uniB1F1,  HANGUL SYLLABLE-B1F1
        '45554' => '10474'            , # G+10474  U+B1F2  uniB1F2,  HANGUL SYLLABLE-B1F2
        '45555' => '10475'            , # G+10475  U+B1F3  uniB1F3,  HANGUL SYLLABLE-B1F3
        '45556' => '10476'            , # G+10476  U+B1F4  uniB1F4,  HANGUL SYLLABLE-B1F4
        '45557' => '10477'            , # G+10477  U+B1F5  uniB1F5,  HANGUL SYLLABLE-B1F5
        '45558' => '10478'            , # G+10478  U+B1F6  uniB1F6,  HANGUL SYLLABLE-B1F6
        '45559' => '10479'            , # G+10479  U+B1F7  uniB1F7,  HANGUL SYLLABLE-B1F7
        '45560' => '10480'            , # G+10480  U+B1F8  uniB1F8,  HANGUL SYLLABLE-B1F8
        '45561' => '1467'             , # G+1467  U+B1F9  uniB1F9,  HANGUL SYLLABLE-B1F9
        '45562' => '10481'            , # G+10481  U+B1FA  uniB1FA,  HANGUL SYLLABLE-B1FA
        '45563' => '1468'             , # G+1468  U+B1FB  uniB1FB,  HANGUL SYLLABLE-B1FB
        '45564' => '10482'            , # G+10482  U+B1FC  uniB1FC,  HANGUL SYLLABLE-B1FC
        '45565' => '1469'             , # G+1469  U+B1FD  uniB1FD,  HANGUL SYLLABLE-B1FD
        '45566' => '10483'            , # G+10483  U+B1FE  uniB1FE,  HANGUL SYLLABLE-B1FE
        '45567' => '10484'            , # G+10484  U+B1FF  uniB1FF,  HANGUL SYLLABLE-B1FF
        '45568' => '10485'            , # G+10485  U+B200  uniB200,  HANGUL SYLLABLE-B200
        '45569' => '10486'            , # G+10486  U+B201  uniB201,  HANGUL SYLLABLE-B201
        '45570' => '10487'            , # G+10487  U+B202  uniB202,  HANGUL SYLLABLE-B202
        '45571' => '10488'            , # G+10488  U+B203  uniB203,  HANGUL SYLLABLE-B203
        '45572' => '1470'             , # G+1470  U+B204  uniB204,  HANGUL SYLLABLE-B204
        '45573' => '1471'             , # G+1471  U+B205  uniB205,  HANGUL SYLLABLE-B205
        '45574' => '10489'            , # G+10489  U+B206  uniB206,  HANGUL SYLLABLE-B206
        '45575' => '10490'            , # G+10490  U+B207  uniB207,  HANGUL SYLLABLE-B207
        '45576' => '1472'             , # G+1472  U+B208  uniB208,  HANGUL SYLLABLE-B208
        '45577' => '10491'            , # G+10491  U+B209  uniB209,  HANGUL SYLLABLE-B209
        '45578' => '10492'            , # G+10492  U+B20A  uniB20A,  HANGUL SYLLABLE-B20A
        '45579' => '1473'             , # G+1473  U+B20B  uniB20B,  HANGUL SYLLABLE-B20B
        '45580' => '1474'             , # G+1474  U+B20C  uniB20C,  HANGUL SYLLABLE-B20C
        '45581' => '10493'            , # G+10493  U+B20D  uniB20D,  HANGUL SYLLABLE-B20D
        '45582' => '10494'            , # G+10494  U+B20E  uniB20E,  HANGUL SYLLABLE-B20E
        '45583' => '10495'            , # G+10495  U+B20F  uniB20F,  HANGUL SYLLABLE-B20F
        '45584' => '10496'            , # G+10496  U+B210  uniB210,  HANGUL SYLLABLE-B210
        '45585' => '10497'            , # G+10497  U+B211  uniB211,  HANGUL SYLLABLE-B211
        '45586' => '10498'            , # G+10498  U+B212  uniB212,  HANGUL SYLLABLE-B212
        '45587' => '10499'            , # G+10499  U+B213  uniB213,  HANGUL SYLLABLE-B213
        '45588' => '1475'             , # G+1475  U+B214  uniB214,  HANGUL SYLLABLE-B214
        '45589' => '1476'             , # G+1476  U+B215  uniB215,  HANGUL SYLLABLE-B215
        '45590' => '10500'            , # G+10500  U+B216  uniB216,  HANGUL SYLLABLE-B216
        '45591' => '1477'             , # G+1477  U+B217  uniB217,  HANGUL SYLLABLE-B217
        '45592' => '10501'            , # G+10501  U+B218  uniB218,  HANGUL SYLLABLE-B218
        '45593' => '1478'             , # G+1478  U+B219  uniB219,  HANGUL SYLLABLE-B219
        '45594' => '10502'            , # G+10502  U+B21A  uniB21A,  HANGUL SYLLABLE-B21A
        '45595' => '10503'            , # G+10503  U+B21B  uniB21B,  HANGUL SYLLABLE-B21B
        '45596' => '10504'            , # G+10504  U+B21C  uniB21C,  HANGUL SYLLABLE-B21C
        '45597' => '10505'            , # G+10505  U+B21D  uniB21D,  HANGUL SYLLABLE-B21D
        '45598' => '10506'            , # G+10506  U+B21E  uniB21E,  HANGUL SYLLABLE-B21E
        '45599' => '10507'            , # G+10507  U+B21F  uniB21F,  HANGUL SYLLABLE-B21F
        '45600' => '1479'             , # G+1479  U+B220  uniB220,  HANGUL SYLLABLE-B220
        '45601' => '10508'            , # G+10508  U+B221  uniB221,  HANGUL SYLLABLE-B221
        '45602' => '10509'            , # G+10509  U+B222  uniB222,  HANGUL SYLLABLE-B222
        '45603' => '10510'            , # G+10510  U+B223  uniB223,  HANGUL SYLLABLE-B223
        '45604' => '10511'            , # G+10511  U+B224  uniB224,  HANGUL SYLLABLE-B224
        '45605' => '10512'            , # G+10512  U+B225  uniB225,  HANGUL SYLLABLE-B225
        '45606' => '10513'            , # G+10513  U+B226  uniB226,  HANGUL SYLLABLE-B226
        '45607' => '10514'            , # G+10514  U+B227  uniB227,  HANGUL SYLLABLE-B227
        '45608' => '10515'            , # G+10515  U+B228  uniB228,  HANGUL SYLLABLE-B228
        '45609' => '10516'            , # G+10516  U+B229  uniB229,  HANGUL SYLLABLE-B229
        '45610' => '10517'            , # G+10517  U+B22A  uniB22A,  HANGUL SYLLABLE-B22A
        '45611' => '10518'            , # G+10518  U+B22B  uniB22B,  HANGUL SYLLABLE-B22B
        '45612' => '10519'            , # G+10519  U+B22C  uniB22C,  HANGUL SYLLABLE-B22C
        '45613' => '10520'            , # G+10520  U+B22D  uniB22D,  HANGUL SYLLABLE-B22D
        '45614' => '10521'            , # G+10521  U+B22E  uniB22E,  HANGUL SYLLABLE-B22E
        '45615' => '10522'            , # G+10522  U+B22F  uniB22F,  HANGUL SYLLABLE-B22F
        '45616' => '10523'            , # G+10523  U+B230  uniB230,  HANGUL SYLLABLE-B230
        '45617' => '10524'            , # G+10524  U+B231  uniB231,  HANGUL SYLLABLE-B231
        '45618' => '10525'            , # G+10525  U+B232  uniB232,  HANGUL SYLLABLE-B232
        '45619' => '10526'            , # G+10526  U+B233  uniB233,  HANGUL SYLLABLE-B233
        '45620' => '1480'             , # G+1480  U+B234  uniB234,  HANGUL SYLLABLE-B234
        '45621' => '10527'            , # G+10527  U+B235  uniB235,  HANGUL SYLLABLE-B235
        '45622' => '10528'            , # G+10528  U+B236  uniB236,  HANGUL SYLLABLE-B236
        '45623' => '10529'            , # G+10529  U+B237  uniB237,  HANGUL SYLLABLE-B237
        '45624' => '10530'            , # G+10530  U+B238  uniB238,  HANGUL SYLLABLE-B238
        '45625' => '10531'            , # G+10531  U+B239  uniB239,  HANGUL SYLLABLE-B239
        '45626' => '10532'            , # G+10532  U+B23A  uniB23A,  HANGUL SYLLABLE-B23A
        '45627' => '10533'            , # G+10533  U+B23B  uniB23B,  HANGUL SYLLABLE-B23B
        '45628' => '1481'             , # G+1481  U+B23C  uniB23C,  HANGUL SYLLABLE-B23C
        '45629' => '10534'            , # G+10534  U+B23D  uniB23D,  HANGUL SYLLABLE-B23D
        '45630' => '10535'            , # G+10535  U+B23E  uniB23E,  HANGUL SYLLABLE-B23E
        '45631' => '10536'            , # G+10536  U+B23F  uniB23F,  HANGUL SYLLABLE-B23F
        '45632' => '10537'            , # G+10537  U+B240  uniB240,  HANGUL SYLLABLE-B240
        '45633' => '10538'            , # G+10538  U+B241  uniB241,  HANGUL SYLLABLE-B241
        '45634' => '10539'            , # G+10539  U+B242  uniB242,  HANGUL SYLLABLE-B242
        '45635' => '10540'            , # G+10540  U+B243  uniB243,  HANGUL SYLLABLE-B243
        '45636' => '10541'            , # G+10541  U+B244  uniB244,  HANGUL SYLLABLE-B244
        '45637' => '10542'            , # G+10542  U+B245  uniB245,  HANGUL SYLLABLE-B245
        '45638' => '10543'            , # G+10543  U+B246  uniB246,  HANGUL SYLLABLE-B246
        '45639' => '10544'            , # G+10544  U+B247  uniB247,  HANGUL SYLLABLE-B247
        '45640' => '10545'            , # G+10545  U+B248  uniB248,  HANGUL SYLLABLE-B248
        '45641' => '10546'            , # G+10546  U+B249  uniB249,  HANGUL SYLLABLE-B249
        '45642' => '10547'            , # G+10547  U+B24A  uniB24A,  HANGUL SYLLABLE-B24A
        '45643' => '10548'            , # G+10548  U+B24B  uniB24B,  HANGUL SYLLABLE-B24B
        '45644' => '10549'            , # G+10549  U+B24C  uniB24C,  HANGUL SYLLABLE-B24C
        '45645' => '10550'            , # G+10550  U+B24D  uniB24D,  HANGUL SYLLABLE-B24D
        '45646' => '10551'            , # G+10551  U+B24E  uniB24E,  HANGUL SYLLABLE-B24E
        '45647' => '10552'            , # G+10552  U+B24F  uniB24F,  HANGUL SYLLABLE-B24F
        '45648' => '10553'            , # G+10553  U+B250  uniB250,  HANGUL SYLLABLE-B250
        '45649' => '10554'            , # G+10554  U+B251  uniB251,  HANGUL SYLLABLE-B251
        '45650' => '10555'            , # G+10555  U+B252  uniB252,  HANGUL SYLLABLE-B252
        '45651' => '10556'            , # G+10556  U+B253  uniB253,  HANGUL SYLLABLE-B253
        '45652' => '10557'            , # G+10557  U+B254  uniB254,  HANGUL SYLLABLE-B254
        '45653' => '10558'            , # G+10558  U+B255  uniB255,  HANGUL SYLLABLE-B255
        '45654' => '10559'            , # G+10559  U+B256  uniB256,  HANGUL SYLLABLE-B256
        '45655' => '10560'            , # G+10560  U+B257  uniB257,  HANGUL SYLLABLE-B257
        '45656' => '1482'             , # G+1482  U+B258  uniB258,  HANGUL SYLLABLE-B258
        '45657' => '10561'            , # G+10561  U+B259  uniB259,  HANGUL SYLLABLE-B259
        '45658' => '10562'            , # G+10562  U+B25A  uniB25A,  HANGUL SYLLABLE-B25A
        '45659' => '10563'            , # G+10563  U+B25B  uniB25B,  HANGUL SYLLABLE-B25B
        '45660' => '1483'             , # G+1483  U+B25C  uniB25C,  HANGUL SYLLABLE-B25C
        '45661' => '10564'            , # G+10564  U+B25D  uniB25D,  HANGUL SYLLABLE-B25D
        '45662' => '10565'            , # G+10565  U+B25E  uniB25E,  HANGUL SYLLABLE-B25E
        '45663' => '10566'            , # G+10566  U+B25F  uniB25F,  HANGUL SYLLABLE-B25F
        '45664' => '1484'             , # G+1484  U+B260  uniB260,  HANGUL SYLLABLE-B260
        '45665' => '10567'            , # G+10567  U+B261  uniB261,  HANGUL SYLLABLE-B261
        '45666' => '10568'            , # G+10568  U+B262  uniB262,  HANGUL SYLLABLE-B262
        '45667' => '10569'            , # G+10569  U+B263  uniB263,  HANGUL SYLLABLE-B263
        '45668' => '10570'            , # G+10570  U+B264  uniB264,  HANGUL SYLLABLE-B264
        '45669' => '10571'            , # G+10571  U+B265  uniB265,  HANGUL SYLLABLE-B265
        '45670' => '10572'            , # G+10572  U+B266  uniB266,  HANGUL SYLLABLE-B266
        '45671' => '10573'            , # G+10573  U+B267  uniB267,  HANGUL SYLLABLE-B267
        '45672' => '1485'             , # G+1485  U+B268  uniB268,  HANGUL SYLLABLE-B268
        '45673' => '1486'             , # G+1486  U+B269  uniB269,  HANGUL SYLLABLE-B269
        '45674' => '10574'            , # G+10574  U+B26A  uniB26A,  HANGUL SYLLABLE-B26A
        '45675' => '10575'            , # G+10575  U+B26B  uniB26B,  HANGUL SYLLABLE-B26B
        '45676' => '10576'            , # G+10576  U+B26C  uniB26C,  HANGUL SYLLABLE-B26C
        '45677' => '10577'            , # G+10577  U+B26D  uniB26D,  HANGUL SYLLABLE-B26D
        '45678' => '10578'            , # G+10578  U+B26E  uniB26E,  HANGUL SYLLABLE-B26E
        '45679' => '10579'            , # G+10579  U+B26F  uniB26F,  HANGUL SYLLABLE-B26F
        '45680' => '10580'            , # G+10580  U+B270  uniB270,  HANGUL SYLLABLE-B270
        '45681' => '10581'            , # G+10581  U+B271  uniB271,  HANGUL SYLLABLE-B271
        '45682' => '10582'            , # G+10582  U+B272  uniB272,  HANGUL SYLLABLE-B272
        '45683' => '10583'            , # G+10583  U+B273  uniB273,  HANGUL SYLLABLE-B273
        '45684' => '1487'             , # G+1487  U+B274  uniB274,  HANGUL SYLLABLE-B274
        '45685' => '1488'             , # G+1488  U+B275  uniB275,  HANGUL SYLLABLE-B275
        '45686' => '10584'            , # G+10584  U+B276  uniB276,  HANGUL SYLLABLE-B276
        '45687' => '10585'            , # G+10585  U+B277  uniB277,  HANGUL SYLLABLE-B277
        '45688' => '10586'            , # G+10586  U+B278  uniB278,  HANGUL SYLLABLE-B278
        '45689' => '10587'            , # G+10587  U+B279  uniB279,  HANGUL SYLLABLE-B279
        '45690' => '10588'            , # G+10588  U+B27A  uniB27A,  HANGUL SYLLABLE-B27A
        '45691' => '10589'            , # G+10589  U+B27B  uniB27B,  HANGUL SYLLABLE-B27B
        '45692' => '1489'             , # G+1489  U+B27C  uniB27C,  HANGUL SYLLABLE-B27C
        '45693' => '10590'            , # G+10590  U+B27D  uniB27D,  HANGUL SYLLABLE-B27D
        '45694' => '10591'            , # G+10591  U+B27E  uniB27E,  HANGUL SYLLABLE-B27E
        '45695' => '10592'            , # G+10592  U+B27F  uniB27F,  HANGUL SYLLABLE-B27F
        '45696' => '10593'            , # G+10593  U+B280  uniB280,  HANGUL SYLLABLE-B280
        '45697' => '10594'            , # G+10594  U+B281  uniB281,  HANGUL SYLLABLE-B281
        '45698' => '10595'            , # G+10595  U+B282  uniB282,  HANGUL SYLLABLE-B282
        '45699' => '10596'            , # G+10596  U+B283  uniB283,  HANGUL SYLLABLE-B283
        '45700' => '1490'             , # G+1490  U+B284  uniB284,  HANGUL SYLLABLE-B284
        '45701' => '1491'             , # G+1491  U+B285  uniB285,  HANGUL SYLLABLE-B285
        '45702' => '10597'            , # G+10597  U+B286  uniB286,  HANGUL SYLLABLE-B286
        '45703' => '10598'            , # G+10598  U+B287  uniB287,  HANGUL SYLLABLE-B287
        '45704' => '10599'            , # G+10599  U+B288  uniB288,  HANGUL SYLLABLE-B288
        '45705' => '1492'             , # G+1492  U+B289  uniB289,  HANGUL SYLLABLE-B289
        '45706' => '10600'            , # G+10600  U+B28A  uniB28A,  HANGUL SYLLABLE-B28A
        '45707' => '10601'            , # G+10601  U+B28B  uniB28B,  HANGUL SYLLABLE-B28B
        '45708' => '10602'            , # G+10602  U+B28C  uniB28C,  HANGUL SYLLABLE-B28C
        '45709' => '10603'            , # G+10603  U+B28D  uniB28D,  HANGUL SYLLABLE-B28D
        '45710' => '10604'            , # G+10604  U+B28E  uniB28E,  HANGUL SYLLABLE-B28E
        '45711' => '10605'            , # G+10605  U+B28F  uniB28F,  HANGUL SYLLABLE-B28F
        '45712' => '1493'             , # G+1493  U+B290  uniB290,  HANGUL SYLLABLE-B290
        '45713' => '1494'             , # G+1494  U+B291  uniB291,  HANGUL SYLLABLE-B291
        '45714' => '10606'            , # G+10606  U+B292  uniB292,  HANGUL SYLLABLE-B292
        '45715' => '10607'            , # G+10607  U+B293  uniB293,  HANGUL SYLLABLE-B293
        '45716' => '1495'             , # G+1495  U+B294  uniB294,  HANGUL SYLLABLE-B294
        '45717' => '10608'            , # G+10608  U+B295  uniB295,  HANGUL SYLLABLE-B295
        '45718' => '10609'            , # G+10609  U+B296  uniB296,  HANGUL SYLLABLE-B296
        '45719' => '10610'            , # G+10610  U+B297  uniB297,  HANGUL SYLLABLE-B297
        '45720' => '1496'             , # G+1496  U+B298  uniB298,  HANGUL SYLLABLE-B298
        '45721' => '1497'             , # G+1497  U+B299  uniB299,  HANGUL SYLLABLE-B299
        '45722' => '1498'             , # G+1498  U+B29A  uniB29A,  HANGUL SYLLABLE-B29A
        '45723' => '10611'            , # G+10611  U+B29B  uniB29B,  HANGUL SYLLABLE-B29B
        '45724' => '10612'            , # G+10612  U+B29C  uniB29C,  HANGUL SYLLABLE-B29C
        '45725' => '10613'            , # G+10613  U+B29D  uniB29D,  HANGUL SYLLABLE-B29D
        '45726' => '10614'            , # G+10614  U+B29E  uniB29E,  HANGUL SYLLABLE-B29E
        '45727' => '10615'            , # G+10615  U+B29F  uniB29F,  HANGUL SYLLABLE-B29F
        '45728' => '1499'             , # G+1499  U+B2A0  uniB2A0,  HANGUL SYLLABLE-B2A0
        '45729' => '1500'             , # G+1500  U+B2A1  uniB2A1,  HANGUL SYLLABLE-B2A1
        '45730' => '10616'            , # G+10616  U+B2A2  uniB2A2,  HANGUL SYLLABLE-B2A2
        '45731' => '1501'             , # G+1501  U+B2A3  uniB2A3,  HANGUL SYLLABLE-B2A3
        '45732' => '10617'            , # G+10617  U+B2A4  uniB2A4,  HANGUL SYLLABLE-B2A4
        '45733' => '1502'             , # G+1502  U+B2A5  uniB2A5,  HANGUL SYLLABLE-B2A5
        '45734' => '1503'             , # G+1503  U+B2A6  uniB2A6,  HANGUL SYLLABLE-B2A6
        '45735' => '10618'            , # G+10618  U+B2A7  uniB2A7,  HANGUL SYLLABLE-B2A7
        '45736' => '10619'            , # G+10619  U+B2A8  uniB2A8,  HANGUL SYLLABLE-B2A8
        '45737' => '10620'            , # G+10620  U+B2A9  uniB2A9,  HANGUL SYLLABLE-B2A9
        '45738' => '1504'             , # G+1504  U+B2AA  uniB2AA,  HANGUL SYLLABLE-B2AA
        '45739' => '10621'            , # G+10621  U+B2AB  uniB2AB,  HANGUL SYLLABLE-B2AB
        '45740' => '1505'             , # G+1505  U+B2AC  uniB2AC,  HANGUL SYLLABLE-B2AC
        '45741' => '10622'            , # G+10622  U+B2AD  uniB2AD,  HANGUL SYLLABLE-B2AD
        '45742' => '10623'            , # G+10623  U+B2AE  uniB2AE,  HANGUL SYLLABLE-B2AE
        '45743' => '10624'            , # G+10624  U+B2AF  uniB2AF,  HANGUL SYLLABLE-B2AF
        '45744' => '1506'             , # G+1506  U+B2B0  uniB2B0,  HANGUL SYLLABLE-B2B0
        '45745' => '10625'            , # G+10625  U+B2B1  uniB2B1,  HANGUL SYLLABLE-B2B1
        '45746' => '10626'            , # G+10626  U+B2B2  uniB2B2,  HANGUL SYLLABLE-B2B2
        '45747' => '10627'            , # G+10627  U+B2B3  uniB2B3,  HANGUL SYLLABLE-B2B3
        '45748' => '1507'             , # G+1507  U+B2B4  uniB2B4,  HANGUL SYLLABLE-B2B4
        '45749' => '10628'            , # G+10628  U+B2B5  uniB2B5,  HANGUL SYLLABLE-B2B5
        '45750' => '10629'            , # G+10629  U+B2B6  uniB2B6,  HANGUL SYLLABLE-B2B6
        '45751' => '10630'            , # G+10630  U+B2B7  uniB2B7,  HANGUL SYLLABLE-B2B7
        '45752' => '10631'            , # G+10631  U+B2B8  uniB2B8,  HANGUL SYLLABLE-B2B8
        '45753' => '10632'            , # G+10632  U+B2B9  uniB2B9,  HANGUL SYLLABLE-B2B9
        '45754' => '10633'            , # G+10633  U+B2BA  uniB2BA,  HANGUL SYLLABLE-B2BA
        '45755' => '10634'            , # G+10634  U+B2BB  uniB2BB,  HANGUL SYLLABLE-B2BB
        '45756' => '10635'            , # G+10635  U+B2BC  uniB2BC,  HANGUL SYLLABLE-B2BC
        '45757' => '10636'            , # G+10636  U+B2BD  uniB2BD,  HANGUL SYLLABLE-B2BD
        '45758' => '10637'            , # G+10637  U+B2BE  uniB2BE,  HANGUL SYLLABLE-B2BE
        '45759' => '10638'            , # G+10638  U+B2BF  uniB2BF,  HANGUL SYLLABLE-B2BF
        '45760' => '10639'            , # G+10639  U+B2C0  uniB2C0,  HANGUL SYLLABLE-B2C0
        '45761' => '10640'            , # G+10640  U+B2C1  uniB2C1,  HANGUL SYLLABLE-B2C1
        '45762' => '10641'            , # G+10641  U+B2C2  uniB2C2,  HANGUL SYLLABLE-B2C2
        '45763' => '10642'            , # G+10642  U+B2C3  uniB2C3,  HANGUL SYLLABLE-B2C3
        '45764' => '10643'            , # G+10643  U+B2C4  uniB2C4,  HANGUL SYLLABLE-B2C4
        '45765' => '10644'            , # G+10644  U+B2C5  uniB2C5,  HANGUL SYLLABLE-B2C5
        '45766' => '10645'            , # G+10645  U+B2C6  uniB2C6,  HANGUL SYLLABLE-B2C6
        '45767' => '10646'            , # G+10646  U+B2C7  uniB2C7,  HANGUL SYLLABLE-B2C7
        '45768' => '1508'             , # G+1508  U+B2C8  uniB2C8,  HANGUL SYLLABLE-B2C8
        '45769' => '1509'             , # G+1509  U+B2C9  uniB2C9,  HANGUL SYLLABLE-B2C9
        '45770' => '10647'            , # G+10647  U+B2CA  uniB2CA,  HANGUL SYLLABLE-B2CA
        '45771' => '10648'            , # G+10648  U+B2CB  uniB2CB,  HANGUL SYLLABLE-B2CB
        '45772' => '1510'             , # G+1510  U+B2CC  uniB2CC,  HANGUL SYLLABLE-B2CC
        '45773' => '10649'            , # G+10649  U+B2CD  uniB2CD,  HANGUL SYLLABLE-B2CD
        '45774' => '10650'            , # G+10650  U+B2CE  uniB2CE,  HANGUL SYLLABLE-B2CE
        '45775' => '10651'            , # G+10651  U+B2CF  uniB2CF,  HANGUL SYLLABLE-B2CF
        '45776' => '1511'             , # G+1511  U+B2D0  uniB2D0,  HANGUL SYLLABLE-B2D0
        '45777' => '10652'            , # G+10652  U+B2D1  uniB2D1,  HANGUL SYLLABLE-B2D1
        '45778' => '1512'             , # G+1512  U+B2D2  uniB2D2,  HANGUL SYLLABLE-B2D2
        '45779' => '10653'            , # G+10653  U+B2D3  uniB2D3,  HANGUL SYLLABLE-B2D3
        '45780' => '10654'            , # G+10654  U+B2D4  uniB2D4,  HANGUL SYLLABLE-B2D4
        '45781' => '10655'            , # G+10655  U+B2D5  uniB2D5,  HANGUL SYLLABLE-B2D5
        '45782' => '10656'            , # G+10656  U+B2D6  uniB2D6,  HANGUL SYLLABLE-B2D6
        '45783' => '10657'            , # G+10657  U+B2D7  uniB2D7,  HANGUL SYLLABLE-B2D7
        '45784' => '1513'             , # G+1513  U+B2D8  uniB2D8,  HANGUL SYLLABLE-B2D8
        '45785' => '1514'             , # G+1514  U+B2D9  uniB2D9,  HANGUL SYLLABLE-B2D9
        '45786' => '10658'            , # G+10658  U+B2DA  uniB2DA,  HANGUL SYLLABLE-B2DA
        '45787' => '1515'             , # G+1515  U+B2DB  uniB2DB,  HANGUL SYLLABLE-B2DB
        '45788' => '10659'            , # G+10659  U+B2DC  uniB2DC,  HANGUL SYLLABLE-B2DC
        '45789' => '1516'             , # G+1516  U+B2DD  uniB2DD,  HANGUL SYLLABLE-B2DD
        '45790' => '10660'            , # G+10660  U+B2DE  uniB2DE,  HANGUL SYLLABLE-B2DE
        '45791' => '10661'            , # G+10661  U+B2DF  uniB2DF,  HANGUL SYLLABLE-B2DF
        '45792' => '10662'            , # G+10662  U+B2E0  uniB2E0,  HANGUL SYLLABLE-B2E0
        '45793' => '10663'            , # G+10663  U+B2E1  uniB2E1,  HANGUL SYLLABLE-B2E1
        '45794' => '1517'             , # G+1517  U+B2E2  uniB2E2,  HANGUL SYLLABLE-B2E2
        '45795' => '10664'            , # G+10664  U+B2E3  uniB2E3,  HANGUL SYLLABLE-B2E3
        '45796' => '1518'             , # G+1518  U+B2E4  uniB2E4,  HANGUL SYLLABLE-B2E4
        '45797' => '1519'             , # G+1519  U+B2E5  uniB2E5,  HANGUL SYLLABLE-B2E5
        '45798' => '1520'             , # G+1520  U+B2E6  uniB2E6,  HANGUL SYLLABLE-B2E6
        '45799' => '10665'            , # G+10665  U+B2E7  uniB2E7,  HANGUL SYLLABLE-B2E7
        '45800' => '1521'             , # G+1521  U+B2E8  uniB2E8,  HANGUL SYLLABLE-B2E8
        '45801' => '10666'            , # G+10666  U+B2E9  uniB2E9,  HANGUL SYLLABLE-B2E9
        '45802' => '10667'            , # G+10667  U+B2EA  uniB2EA,  HANGUL SYLLABLE-B2EA
        '45803' => '1522'             , # G+1522  U+B2EB  uniB2EB,  HANGUL SYLLABLE-B2EB
        '45804' => '1523'             , # G+1523  U+B2EC  uniB2EC,  HANGUL SYLLABLE-B2EC
        '45805' => '1524'             , # G+1524  U+B2ED  uniB2ED,  HANGUL SYLLABLE-B2ED
        '45806' => '1525'             , # G+1525  U+B2EE  uniB2EE,  HANGUL SYLLABLE-B2EE
        '45807' => '1526'             , # G+1526  U+B2EF  uniB2EF,  HANGUL SYLLABLE-B2EF
        '45808' => '10668'            , # G+10668  U+B2F0  uniB2F0,  HANGUL SYLLABLE-B2F0
        '45809' => '10669'            , # G+10669  U+B2F1  uniB2F1,  HANGUL SYLLABLE-B2F1
        '45810' => '10670'            , # G+10670  U+B2F2  uniB2F2,  HANGUL SYLLABLE-B2F2
        '45811' => '1527'             , # G+1527  U+B2F3  uniB2F3,  HANGUL SYLLABLE-B2F3
        '45812' => '1528'             , # G+1528  U+B2F4  uniB2F4,  HANGUL SYLLABLE-B2F4
        '45813' => '1529'             , # G+1529  U+B2F5  uniB2F5,  HANGUL SYLLABLE-B2F5
        '45814' => '10671'            , # G+10671  U+B2F6  uniB2F6,  HANGUL SYLLABLE-B2F6
        '45815' => '1530'             , # G+1530  U+B2F7  uniB2F7,  HANGUL SYLLABLE-B2F7
        '45816' => '1531'             , # G+1531  U+B2F8  uniB2F8,  HANGUL SYLLABLE-B2F8
        '45817' => '1532'             , # G+1532  U+B2F9  uniB2F9,  HANGUL SYLLABLE-B2F9
        '45818' => '1533'             , # G+1533  U+B2FA  uniB2FA,  HANGUL SYLLABLE-B2FA
        '45819' => '1534'             , # G+1534  U+B2FB  uniB2FB,  HANGUL SYLLABLE-B2FB
        '45820' => '10672'            , # G+10672  U+B2FC  uniB2FC,  HANGUL SYLLABLE-B2FC
        '45821' => '10673'            , # G+10673  U+B2FD  uniB2FD,  HANGUL SYLLABLE-B2FD
        '45822' => '10674'            , # G+10674  U+B2FE  uniB2FE,  HANGUL SYLLABLE-B2FE
        '45823' => '1535'             , # G+1535  U+B2FF  uniB2FF,  HANGUL SYLLABLE-B2FF
        '45824' => '1536'             , # G+1536  U+B300  uniB300,  HANGUL SYLLABLE-B300
        '45825' => '1537'             , # G+1537  U+B301  uniB301,  HANGUL SYLLABLE-B301
        '45826' => '10675'            , # G+10675  U+B302  uniB302,  HANGUL SYLLABLE-B302
        '45827' => '10676'            , # G+10676  U+B303  uniB303,  HANGUL SYLLABLE-B303
        '45828' => '1538'             , # G+1538  U+B304  uniB304,  HANGUL SYLLABLE-B304
        '45829' => '10677'            , # G+10677  U+B305  uniB305,  HANGUL SYLLABLE-B305
        '45830' => '10678'            , # G+10678  U+B306  uniB306,  HANGUL SYLLABLE-B306
        '45831' => '10679'            , # G+10679  U+B307  uniB307,  HANGUL SYLLABLE-B307
        '45832' => '1539'             , # G+1539  U+B308  uniB308,  HANGUL SYLLABLE-B308
        '45833' => '10680'            , # G+10680  U+B309  uniB309,  HANGUL SYLLABLE-B309
        '45834' => '10681'            , # G+10681  U+B30A  uniB30A,  HANGUL SYLLABLE-B30A
        '45835' => '10682'            , # G+10682  U+B30B  uniB30B,  HANGUL SYLLABLE-B30B
        '45836' => '10683'            , # G+10683  U+B30C  uniB30C,  HANGUL SYLLABLE-B30C
        '45837' => '10684'            , # G+10684  U+B30D  uniB30D,  HANGUL SYLLABLE-B30D
        '45838' => '10685'            , # G+10685  U+B30E  uniB30E,  HANGUL SYLLABLE-B30E
        '45839' => '10686'            , # G+10686  U+B30F  uniB30F,  HANGUL SYLLABLE-B30F
        '45840' => '1540'             , # G+1540  U+B310  uniB310,  HANGUL SYLLABLE-B310
        '45841' => '1541'             , # G+1541  U+B311  uniB311,  HANGUL SYLLABLE-B311
        '45842' => '10687'            , # G+10687  U+B312  uniB312,  HANGUL SYLLABLE-B312
        '45843' => '1542'             , # G+1542  U+B313  uniB313,  HANGUL SYLLABLE-B313
        '45844' => '1543'             , # G+1543  U+B314  uniB314,  HANGUL SYLLABLE-B314
        '45845' => '1544'             , # G+1544  U+B315  uniB315,  HANGUL SYLLABLE-B315
        '45846' => '10688'            , # G+10688  U+B316  uniB316,  HANGUL SYLLABLE-B316
        '45847' => '10689'            , # G+10689  U+B317  uniB317,  HANGUL SYLLABLE-B317
        '45848' => '10690'            , # G+10690  U+B318  uniB318,  HANGUL SYLLABLE-B318
        '45849' => '10691'            , # G+10691  U+B319  uniB319,  HANGUL SYLLABLE-B319
        '45850' => '10692'            , # G+10692  U+B31A  uniB31A,  HANGUL SYLLABLE-B31A
        '45851' => '10693'            , # G+10693  U+B31B  uniB31B,  HANGUL SYLLABLE-B31B
        '45852' => '1545'             , # G+1545  U+B31C  uniB31C,  HANGUL SYLLABLE-B31C
        '45853' => '10694'            , # G+10694  U+B31D  uniB31D,  HANGUL SYLLABLE-B31D
        '45854' => '10695'            , # G+10695  U+B31E  uniB31E,  HANGUL SYLLABLE-B31E
        '45855' => '10696'            , # G+10696  U+B31F  uniB31F,  HANGUL SYLLABLE-B31F
        '45856' => '10697'            , # G+10697  U+B320  uniB320,  HANGUL SYLLABLE-B320
        '45857' => '10698'            , # G+10698  U+B321  uniB321,  HANGUL SYLLABLE-B321
        '45858' => '10699'            , # G+10699  U+B322  uniB322,  HANGUL SYLLABLE-B322
        '45859' => '10700'            , # G+10700  U+B323  uniB323,  HANGUL SYLLABLE-B323
        '45860' => '10701'            , # G+10701  U+B324  uniB324,  HANGUL SYLLABLE-B324
        '45861' => '10702'            , # G+10702  U+B325  uniB325,  HANGUL SYLLABLE-B325
        '45862' => '10703'            , # G+10703  U+B326  uniB326,  HANGUL SYLLABLE-B326
        '45863' => '10704'            , # G+10704  U+B327  uniB327,  HANGUL SYLLABLE-B327
        '45864' => '10705'            , # G+10705  U+B328  uniB328,  HANGUL SYLLABLE-B328
        '45865' => '10706'            , # G+10706  U+B329  uniB329,  HANGUL SYLLABLE-B329
        '45866' => '10707'            , # G+10707  U+B32A  uniB32A,  HANGUL SYLLABLE-B32A
        '45867' => '10708'            , # G+10708  U+B32B  uniB32B,  HANGUL SYLLABLE-B32B
        '45868' => '10709'            , # G+10709  U+B32C  uniB32C,  HANGUL SYLLABLE-B32C
        '45869' => '10710'            , # G+10710  U+B32D  uniB32D,  HANGUL SYLLABLE-B32D
        '45870' => '10711'            , # G+10711  U+B32E  uniB32E,  HANGUL SYLLABLE-B32E
        '45871' => '10712'            , # G+10712  U+B32F  uniB32F,  HANGUL SYLLABLE-B32F
        '45872' => '10713'            , # G+10713  U+B330  uniB330,  HANGUL SYLLABLE-B330
        '45873' => '10714'            , # G+10714  U+B331  uniB331,  HANGUL SYLLABLE-B331
        '45874' => '10715'            , # G+10715  U+B332  uniB332,  HANGUL SYLLABLE-B332
        '45875' => '10716'            , # G+10716  U+B333  uniB333,  HANGUL SYLLABLE-B333
        '45876' => '10717'            , # G+10717  U+B334  uniB334,  HANGUL SYLLABLE-B334
        '45877' => '10718'            , # G+10718  U+B335  uniB335,  HANGUL SYLLABLE-B335
        '45878' => '10719'            , # G+10719  U+B336  uniB336,  HANGUL SYLLABLE-B336
        '45879' => '10720'            , # G+10720  U+B337  uniB337,  HANGUL SYLLABLE-B337
        '45880' => '10721'            , # G+10721  U+B338  uniB338,  HANGUL SYLLABLE-B338
        '45881' => '10722'            , # G+10722  U+B339  uniB339,  HANGUL SYLLABLE-B339
        '45882' => '10723'            , # G+10723  U+B33A  uniB33A,  HANGUL SYLLABLE-B33A
        '45883' => '10724'            , # G+10724  U+B33B  uniB33B,  HANGUL SYLLABLE-B33B
        '45884' => '10725'            , # G+10725  U+B33C  uniB33C,  HANGUL SYLLABLE-B33C
        '45885' => '10726'            , # G+10726  U+B33D  uniB33D,  HANGUL SYLLABLE-B33D
        '45886' => '10727'            , # G+10727  U+B33E  uniB33E,  HANGUL SYLLABLE-B33E
        '45887' => '10728'            , # G+10728  U+B33F  uniB33F,  HANGUL SYLLABLE-B33F
        '45888' => '10729'            , # G+10729  U+B340  uniB340,  HANGUL SYLLABLE-B340
        '45889' => '10730'            , # G+10730  U+B341  uniB341,  HANGUL SYLLABLE-B341
        '45890' => '10731'            , # G+10731  U+B342  uniB342,  HANGUL SYLLABLE-B342
        '45891' => '10732'            , # G+10732  U+B343  uniB343,  HANGUL SYLLABLE-B343
        '45892' => '10733'            , # G+10733  U+B344  uniB344,  HANGUL SYLLABLE-B344
        '45893' => '10734'            , # G+10734  U+B345  uniB345,  HANGUL SYLLABLE-B345
        '45894' => '10735'            , # G+10735  U+B346  uniB346,  HANGUL SYLLABLE-B346
        '45895' => '10736'            , # G+10736  U+B347  uniB347,  HANGUL SYLLABLE-B347
        '45896' => '10737'            , # G+10737  U+B348  uniB348,  HANGUL SYLLABLE-B348
        '45897' => '10738'            , # G+10738  U+B349  uniB349,  HANGUL SYLLABLE-B349
        '45898' => '10739'            , # G+10739  U+B34A  uniB34A,  HANGUL SYLLABLE-B34A
        '45899' => '10740'            , # G+10740  U+B34B  uniB34B,  HANGUL SYLLABLE-B34B
        '45900' => '10741'            , # G+10741  U+B34C  uniB34C,  HANGUL SYLLABLE-B34C
        '45901' => '10742'            , # G+10742  U+B34D  uniB34D,  HANGUL SYLLABLE-B34D
        '45902' => '10743'            , # G+10743  U+B34E  uniB34E,  HANGUL SYLLABLE-B34E
        '45903' => '10744'            , # G+10744  U+B34F  uniB34F,  HANGUL SYLLABLE-B34F
        '45904' => '10745'            , # G+10745  U+B350  uniB350,  HANGUL SYLLABLE-B350
        '45905' => '10746'            , # G+10746  U+B351  uniB351,  HANGUL SYLLABLE-B351
        '45906' => '10747'            , # G+10747  U+B352  uniB352,  HANGUL SYLLABLE-B352
        '45907' => '10748'            , # G+10748  U+B353  uniB353,  HANGUL SYLLABLE-B353
        '45908' => '1546'             , # G+1546  U+B354  uniB354,  HANGUL SYLLABLE-B354
        '45909' => '1547'             , # G+1547  U+B355  uniB355,  HANGUL SYLLABLE-B355
        '45910' => '1548'             , # G+1548  U+B356  uniB356,  HANGUL SYLLABLE-B356
        '45911' => '10749'            , # G+10749  U+B357  uniB357,  HANGUL SYLLABLE-B357
        '45912' => '1549'             , # G+1549  U+B358  uniB358,  HANGUL SYLLABLE-B358
        '45913' => '10750'            , # G+10750  U+B359  uniB359,  HANGUL SYLLABLE-B359
        '45914' => '10751'            , # G+10751  U+B35A  uniB35A,  HANGUL SYLLABLE-B35A
        '45915' => '1550'             , # G+1550  U+B35B  uniB35B,  HANGUL SYLLABLE-B35B
        '45916' => '1551'             , # G+1551  U+B35C  uniB35C,  HANGUL SYLLABLE-B35C
        '45917' => '10752'            , # G+10752  U+B35D  uniB35D,  HANGUL SYLLABLE-B35D
        '45918' => '1552'             , # G+1552  U+B35E  uniB35E,  HANGUL SYLLABLE-B35E
        '45919' => '1553'             , # G+1553  U+B35F  uniB35F,  HANGUL SYLLABLE-B35F
        '45920' => '10753'            , # G+10753  U+B360  uniB360,  HANGUL SYLLABLE-B360
        '45921' => '10754'            , # G+10754  U+B361  uniB361,  HANGUL SYLLABLE-B361
        '45922' => '10755'            , # G+10755  U+B362  uniB362,  HANGUL SYLLABLE-B362
        '45923' => '10756'            , # G+10756  U+B363  uniB363,  HANGUL SYLLABLE-B363
        '45924' => '1554'             , # G+1554  U+B364  uniB364,  HANGUL SYLLABLE-B364
        '45925' => '1555'             , # G+1555  U+B365  uniB365,  HANGUL SYLLABLE-B365
        '45926' => '10757'            , # G+10757  U+B366  uniB366,  HANGUL SYLLABLE-B366
        '45927' => '1556'             , # G+1556  U+B367  uniB367,  HANGUL SYLLABLE-B367
        '45928' => '10758'            , # G+10758  U+B368  uniB368,  HANGUL SYLLABLE-B368
        '45929' => '1557'             , # G+1557  U+B369  uniB369,  HANGUL SYLLABLE-B369
        '45930' => '10759'            , # G+10759  U+B36A  uniB36A,  HANGUL SYLLABLE-B36A
        '45931' => '1558'             , # G+1558  U+B36B  uniB36B,  HANGUL SYLLABLE-B36B
        '45932' => '10760'            , # G+10760  U+B36C  uniB36C,  HANGUL SYLLABLE-B36C
        '45933' => '10761'            , # G+10761  U+B36D  uniB36D,  HANGUL SYLLABLE-B36D
        '45934' => '1559'             , # G+1559  U+B36E  uniB36E,  HANGUL SYLLABLE-B36E
        '45935' => '10762'            , # G+10762  U+B36F  uniB36F,  HANGUL SYLLABLE-B36F
        '45936' => '1560'             , # G+1560  U+B370  uniB370,  HANGUL SYLLABLE-B370
        '45937' => '1561'             , # G+1561  U+B371  uniB371,  HANGUL SYLLABLE-B371
        '45938' => '10763'            , # G+10763  U+B372  uniB372,  HANGUL SYLLABLE-B372
        '45939' => '10764'            , # G+10764  U+B373  uniB373,  HANGUL SYLLABLE-B373
        '45940' => '1562'             , # G+1562  U+B374  uniB374,  HANGUL SYLLABLE-B374
        '45941' => '10765'            , # G+10765  U+B375  uniB375,  HANGUL SYLLABLE-B375
        '45942' => '10766'            , # G+10766  U+B376  uniB376,  HANGUL SYLLABLE-B376
        '45943' => '10767'            , # G+10767  U+B377  uniB377,  HANGUL SYLLABLE-B377
        '45944' => '1563'             , # G+1563  U+B378  uniB378,  HANGUL SYLLABLE-B378
        '45945' => '10768'            , # G+10768  U+B379  uniB379,  HANGUL SYLLABLE-B379
        '45946' => '10769'            , # G+10769  U+B37A  uniB37A,  HANGUL SYLLABLE-B37A
        '45947' => '10770'            , # G+10770  U+B37B  uniB37B,  HANGUL SYLLABLE-B37B
        '45948' => '10771'            , # G+10771  U+B37C  uniB37C,  HANGUL SYLLABLE-B37C
        '45949' => '10772'            , # G+10772  U+B37D  uniB37D,  HANGUL SYLLABLE-B37D
        '45950' => '10773'            , # G+10773  U+B37E  uniB37E,  HANGUL SYLLABLE-B37E
        '45951' => '10774'            , # G+10774  U+B37F  uniB37F,  HANGUL SYLLABLE-B37F
        '45952' => '1564'             , # G+1564  U+B380  uniB380,  HANGUL SYLLABLE-B380
        '45953' => '1565'             , # G+1565  U+B381  uniB381,  HANGUL SYLLABLE-B381
        '45954' => '10775'            , # G+10775  U+B382  uniB382,  HANGUL SYLLABLE-B382
        '45955' => '1566'             , # G+1566  U+B383  uniB383,  HANGUL SYLLABLE-B383
        '45956' => '1567'             , # G+1567  U+B384  uniB384,  HANGUL SYLLABLE-B384
        '45957' => '1568'             , # G+1568  U+B385  uniB385,  HANGUL SYLLABLE-B385
        '45958' => '10776'            , # G+10776  U+B386  uniB386,  HANGUL SYLLABLE-B386
        '45959' => '10777'            , # G+10777  U+B387  uniB387,  HANGUL SYLLABLE-B387
        '45960' => '10778'            , # G+10778  U+B388  uniB388,  HANGUL SYLLABLE-B388
        '45961' => '10779'            , # G+10779  U+B389  uniB389,  HANGUL SYLLABLE-B389
        '45962' => '10780'            , # G+10780  U+B38A  uniB38A,  HANGUL SYLLABLE-B38A
        '45963' => '10781'            , # G+10781  U+B38B  uniB38B,  HANGUL SYLLABLE-B38B
        '45964' => '1569'             , # G+1569  U+B38C  uniB38C,  HANGUL SYLLABLE-B38C
        '45965' => '10782'            , # G+10782  U+B38D  uniB38D,  HANGUL SYLLABLE-B38D
        '45966' => '10783'            , # G+10783  U+B38E  uniB38E,  HANGUL SYLLABLE-B38E
        '45967' => '10784'            , # G+10784  U+B38F  uniB38F,  HANGUL SYLLABLE-B38F
        '45968' => '1570'             , # G+1570  U+B390  uniB390,  HANGUL SYLLABLE-B390
        '45969' => '10785'            , # G+10785  U+B391  uniB391,  HANGUL SYLLABLE-B391
        '45970' => '10786'            , # G+10786  U+B392  uniB392,  HANGUL SYLLABLE-B392
        '45971' => '10787'            , # G+10787  U+B393  uniB393,  HANGUL SYLLABLE-B393
        '45972' => '1571'             , # G+1571  U+B394  uniB394,  HANGUL SYLLABLE-B394
        '45973' => '10788'            , # G+10788  U+B395  uniB395,  HANGUL SYLLABLE-B395
        '45974' => '10789'            , # G+10789  U+B396  uniB396,  HANGUL SYLLABLE-B396
        '45975' => '10790'            , # G+10790  U+B397  uniB397,  HANGUL SYLLABLE-B397
        '45976' => '10791'            , # G+10791  U+B398  uniB398,  HANGUL SYLLABLE-B398
        '45977' => '10792'            , # G+10792  U+B399  uniB399,  HANGUL SYLLABLE-B399
        '45978' => '10793'            , # G+10793  U+B39A  uniB39A,  HANGUL SYLLABLE-B39A
        '45979' => '10794'            , # G+10794  U+B39B  uniB39B,  HANGUL SYLLABLE-B39B
        '45980' => '10795'            , # G+10795  U+B39C  uniB39C,  HANGUL SYLLABLE-B39C
        '45981' => '10796'            , # G+10796  U+B39D  uniB39D,  HANGUL SYLLABLE-B39D
        '45982' => '10797'            , # G+10797  U+B39E  uniB39E,  HANGUL SYLLABLE-B39E
        '45983' => '10798'            , # G+10798  U+B39F  uniB39F,  HANGUL SYLLABLE-B39F
        '45984' => '1572'             , # G+1572  U+B3A0  uniB3A0,  HANGUL SYLLABLE-B3A0
        '45985' => '1573'             , # G+1573  U+B3A1  uniB3A1,  HANGUL SYLLABLE-B3A1
        '45986' => '10799'            , # G+10799  U+B3A2  uniB3A2,  HANGUL SYLLABLE-B3A2
        '45987' => '10800'            , # G+10800  U+B3A3  uniB3A3,  HANGUL SYLLABLE-B3A3
        '45988' => '10801'            , # G+10801  U+B3A4  uniB3A4,  HANGUL SYLLABLE-B3A4
        '45989' => '10802'            , # G+10802  U+B3A5  uniB3A5,  HANGUL SYLLABLE-B3A5
        '45990' => '10803'            , # G+10803  U+B3A6  uniB3A6,  HANGUL SYLLABLE-B3A6
        '45991' => '10804'            , # G+10804  U+B3A7  uniB3A7,  HANGUL SYLLABLE-B3A7
        '45992' => '1574'             , # G+1574  U+B3A8  uniB3A8,  HANGUL SYLLABLE-B3A8
        '45993' => '10805'            , # G+10805  U+B3A9  uniB3A9,  HANGUL SYLLABLE-B3A9
        '45994' => '10806'            , # G+10806  U+B3AA  uniB3AA,  HANGUL SYLLABLE-B3AA
        '45995' => '10807'            , # G+10807  U+B3AB  uniB3AB,  HANGUL SYLLABLE-B3AB
        '45996' => '1575'             , # G+1575  U+B3AC  uniB3AC,  HANGUL SYLLABLE-B3AC
        '45997' => '10808'            , # G+10808  U+B3AD  uniB3AD,  HANGUL SYLLABLE-B3AD
        '45998' => '10809'            , # G+10809  U+B3AE  uniB3AE,  HANGUL SYLLABLE-B3AE
        '45999' => '10810'            , # G+10810  U+B3AF  uniB3AF,  HANGUL SYLLABLE-B3AF
        '46000' => '10811'            , # G+10811  U+B3B0  uniB3B0,  HANGUL SYLLABLE-B3B0
        '46001' => '10812'            , # G+10812  U+B3B1  uniB3B1,  HANGUL SYLLABLE-B3B1
        '46002' => '10813'            , # G+10813  U+B3B2  uniB3B2,  HANGUL SYLLABLE-B3B2
        '46003' => '10814'            , # G+10814  U+B3B3  uniB3B3,  HANGUL SYLLABLE-B3B3
        '46004' => '10815'            , # G+10815  U+B3B4  uniB3B4,  HANGUL SYLLABLE-B3B4
        '46005' => '10816'            , # G+10816  U+B3B5  uniB3B5,  HANGUL SYLLABLE-B3B5
        '46006' => '10817'            , # G+10817  U+B3B6  uniB3B6,  HANGUL SYLLABLE-B3B6
        '46007' => '10818'            , # G+10818  U+B3B7  uniB3B7,  HANGUL SYLLABLE-B3B7
        '46008' => '10819'            , # G+10819  U+B3B8  uniB3B8,  HANGUL SYLLABLE-B3B8
        '46009' => '10820'            , # G+10820  U+B3B9  uniB3B9,  HANGUL SYLLABLE-B3B9
        '46010' => '10821'            , # G+10821  U+B3BA  uniB3BA,  HANGUL SYLLABLE-B3BA
        '46011' => '10822'            , # G+10822  U+B3BB  uniB3BB,  HANGUL SYLLABLE-B3BB
        '46012' => '10823'            , # G+10823  U+B3BC  uniB3BC,  HANGUL SYLLABLE-B3BC
        '46013' => '10824'            , # G+10824  U+B3BD  uniB3BD,  HANGUL SYLLABLE-B3BD
        '46014' => '10825'            , # G+10825  U+B3BE  uniB3BE,  HANGUL SYLLABLE-B3BE
        '46015' => '10826'            , # G+10826  U+B3BF  uniB3BF,  HANGUL SYLLABLE-B3BF
        '46016' => '10827'            , # G+10827  U+B3C0  uniB3C0,  HANGUL SYLLABLE-B3C0
        '46017' => '10828'            , # G+10828  U+B3C1  uniB3C1,  HANGUL SYLLABLE-B3C1
        '46018' => '10829'            , # G+10829  U+B3C2  uniB3C2,  HANGUL SYLLABLE-B3C2
        '46019' => '10830'            , # G+10830  U+B3C3  uniB3C3,  HANGUL SYLLABLE-B3C3
        '46020' => '1576'             , # G+1576  U+B3C4  uniB3C4,  HANGUL SYLLABLE-B3C4
        '46021' => '1577'             , # G+1577  U+B3C5  uniB3C5,  HANGUL SYLLABLE-B3C5
        '46022' => '10831'            , # G+10831  U+B3C6  uniB3C6,  HANGUL SYLLABLE-B3C6
        '46023' => '10832'            , # G+10832  U+B3C7  uniB3C7,  HANGUL SYLLABLE-B3C7
        '46024' => '1578'             , # G+1578  U+B3C8  uniB3C8,  HANGUL SYLLABLE-B3C8
        '46025' => '10833'            , # G+10833  U+B3C9  uniB3C9,  HANGUL SYLLABLE-B3C9
        '46026' => '10834'            , # G+10834  U+B3CA  uniB3CA,  HANGUL SYLLABLE-B3CA
        '46027' => '1579'             , # G+1579  U+B3CB  uniB3CB,  HANGUL SYLLABLE-B3CB
        '46028' => '1580'             , # G+1580  U+B3CC  uniB3CC,  HANGUL SYLLABLE-B3CC
        '46029' => '10835'            , # G+10835  U+B3CD  uniB3CD,  HANGUL SYLLABLE-B3CD
        '46030' => '1581'             , # G+1581  U+B3CE  uniB3CE,  HANGUL SYLLABLE-B3CE
        '46031' => '10836'            , # G+10836  U+B3CF  uniB3CF,  HANGUL SYLLABLE-B3CF
        '46032' => '1582'             , # G+1582  U+B3D0  uniB3D0,  HANGUL SYLLABLE-B3D0
        '46033' => '10837'            , # G+10837  U+B3D1  uniB3D1,  HANGUL SYLLABLE-B3D1
        '46034' => '10838'            , # G+10838  U+B3D2  uniB3D2,  HANGUL SYLLABLE-B3D2
        '46035' => '10839'            , # G+10839  U+B3D3  uniB3D3,  HANGUL SYLLABLE-B3D3
        '46036' => '1583'             , # G+1583  U+B3D4  uniB3D4,  HANGUL SYLLABLE-B3D4
        '46037' => '1584'             , # G+1584  U+B3D5  uniB3D5,  HANGUL SYLLABLE-B3D5
        '46038' => '10840'            , # G+10840  U+B3D6  uniB3D6,  HANGUL SYLLABLE-B3D6
        '46039' => '1585'             , # G+1585  U+B3D7  uniB3D7,  HANGUL SYLLABLE-B3D7
        '46040' => '10841'            , # G+10841  U+B3D8  uniB3D8,  HANGUL SYLLABLE-B3D8
        '46041' => '1586'             , # G+1586  U+B3D9  uniB3D9,  HANGUL SYLLABLE-B3D9
        '46042' => '10842'            , # G+10842  U+B3DA  uniB3DA,  HANGUL SYLLABLE-B3DA
        '46043' => '1587'             , # G+1587  U+B3DB  uniB3DB,  HANGUL SYLLABLE-B3DB
        '46044' => '10843'            , # G+10843  U+B3DC  uniB3DC,  HANGUL SYLLABLE-B3DC
        '46045' => '1588'             , # G+1588  U+B3DD  uniB3DD,  HANGUL SYLLABLE-B3DD
        '46046' => '10844'            , # G+10844  U+B3DE  uniB3DE,  HANGUL SYLLABLE-B3DE
        '46047' => '10845'            , # G+10845  U+B3DF  uniB3DF,  HANGUL SYLLABLE-B3DF
        '46048' => '1589'             , # G+1589  U+B3E0  uniB3E0,  HANGUL SYLLABLE-B3E0
        '46049' => '10846'            , # G+10846  U+B3E1  uniB3E1,  HANGUL SYLLABLE-B3E1
        '46050' => '10847'            , # G+10847  U+B3E2  uniB3E2,  HANGUL SYLLABLE-B3E2
        '46051' => '10848'            , # G+10848  U+B3E3  uniB3E3,  HANGUL SYLLABLE-B3E3
        '46052' => '1590'             , # G+1590  U+B3E4  uniB3E4,  HANGUL SYLLABLE-B3E4
        '46053' => '10849'            , # G+10849  U+B3E5  uniB3E5,  HANGUL SYLLABLE-B3E5
        '46054' => '10850'            , # G+10850  U+B3E6  uniB3E6,  HANGUL SYLLABLE-B3E6
        '46055' => '10851'            , # G+10851  U+B3E7  uniB3E7,  HANGUL SYLLABLE-B3E7
        '46056' => '1591'             , # G+1591  U+B3E8  uniB3E8,  HANGUL SYLLABLE-B3E8
        '46057' => '10852'            , # G+10852  U+B3E9  uniB3E9,  HANGUL SYLLABLE-B3E9
        '46058' => '10853'            , # G+10853  U+B3EA  uniB3EA,  HANGUL SYLLABLE-B3EA
        '46059' => '10854'            , # G+10854  U+B3EB  uniB3EB,  HANGUL SYLLABLE-B3EB
        '46060' => '10855'            , # G+10855  U+B3EC  uniB3EC,  HANGUL SYLLABLE-B3EC
        '46061' => '10856'            , # G+10856  U+B3ED  uniB3ED,  HANGUL SYLLABLE-B3ED
        '46062' => '10857'            , # G+10857  U+B3EE  uniB3EE,  HANGUL SYLLABLE-B3EE
        '46063' => '10858'            , # G+10858  U+B3EF  uniB3EF,  HANGUL SYLLABLE-B3EF
        '46064' => '10859'            , # G+10859  U+B3F0  uniB3F0,  HANGUL SYLLABLE-B3F0
        '46065' => '10860'            , # G+10860  U+B3F1  uniB3F1,  HANGUL SYLLABLE-B3F1
        '46066' => '10861'            , # G+10861  U+B3F2  uniB3F2,  HANGUL SYLLABLE-B3F2
        '46067' => '10862'            , # G+10862  U+B3F3  uniB3F3,  HANGUL SYLLABLE-B3F3
        '46068' => '10863'            , # G+10863  U+B3F4  uniB3F4,  HANGUL SYLLABLE-B3F4
        '46069' => '10864'            , # G+10864  U+B3F5  uniB3F5,  HANGUL SYLLABLE-B3F5
        '46070' => '10865'            , # G+10865  U+B3F6  uniB3F6,  HANGUL SYLLABLE-B3F6
        '46071' => '10866'            , # G+10866  U+B3F7  uniB3F7,  HANGUL SYLLABLE-B3F7
        '46072' => '10867'            , # G+10867  U+B3F8  uniB3F8,  HANGUL SYLLABLE-B3F8
        '46073' => '10868'            , # G+10868  U+B3F9  uniB3F9,  HANGUL SYLLABLE-B3F9
        '46074' => '10869'            , # G+10869  U+B3FA  uniB3FA,  HANGUL SYLLABLE-B3FA
        '46075' => '10870'            , # G+10870  U+B3FB  uniB3FB,  HANGUL SYLLABLE-B3FB
        '46076' => '1592'             , # G+1592  U+B3FC  uniB3FC,  HANGUL SYLLABLE-B3FC
        '46077' => '10871'            , # G+10871  U+B3FD  uniB3FD,  HANGUL SYLLABLE-B3FD
        '46078' => '10872'            , # G+10872  U+B3FE  uniB3FE,  HANGUL SYLLABLE-B3FE
        '46079' => '10873'            , # G+10873  U+B3FF  uniB3FF,  HANGUL SYLLABLE-B3FF
        '46080' => '10874'            , # G+10874  U+B400  uniB400,  HANGUL SYLLABLE-B400
        '46081' => '10875'            , # G+10875  U+B401  uniB401,  HANGUL SYLLABLE-B401
        '46082' => '10876'            , # G+10876  U+B402  uniB402,  HANGUL SYLLABLE-B402
        '46083' => '10877'            , # G+10877  U+B403  uniB403,  HANGUL SYLLABLE-B403
        '46084' => '10878'            , # G+10878  U+B404  uniB404,  HANGUL SYLLABLE-B404
        '46085' => '10879'            , # G+10879  U+B405  uniB405,  HANGUL SYLLABLE-B405
        '46086' => '10880'            , # G+10880  U+B406  uniB406,  HANGUL SYLLABLE-B406
        '46087' => '10881'            , # G+10881  U+B407  uniB407,  HANGUL SYLLABLE-B407
        '46088' => '10882'            , # G+10882  U+B408  uniB408,  HANGUL SYLLABLE-B408
        '46089' => '10883'            , # G+10883  U+B409  uniB409,  HANGUL SYLLABLE-B409
        '46090' => '10884'            , # G+10884  U+B40A  uniB40A,  HANGUL SYLLABLE-B40A
        '46091' => '10885'            , # G+10885  U+B40B  uniB40B,  HANGUL SYLLABLE-B40B
        '46092' => '10886'            , # G+10886  U+B40C  uniB40C,  HANGUL SYLLABLE-B40C
        '46093' => '10887'            , # G+10887  U+B40D  uniB40D,  HANGUL SYLLABLE-B40D
        '46094' => '10888'            , # G+10888  U+B40E  uniB40E,  HANGUL SYLLABLE-B40E
        '46095' => '10889'            , # G+10889  U+B40F  uniB40F,  HANGUL SYLLABLE-B40F
        '46096' => '1593'             , # G+1593  U+B410  uniB410,  HANGUL SYLLABLE-B410
        '46097' => '10890'            , # G+10890  U+B411  uniB411,  HANGUL SYLLABLE-B411
        '46098' => '10891'            , # G+10891  U+B412  uniB412,  HANGUL SYLLABLE-B412
        '46099' => '10892'            , # G+10892  U+B413  uniB413,  HANGUL SYLLABLE-B413
        '46100' => '10893'            , # G+10893  U+B414  uniB414,  HANGUL SYLLABLE-B414
        '46101' => '10894'            , # G+10894  U+B415  uniB415,  HANGUL SYLLABLE-B415
        '46102' => '10895'            , # G+10895  U+B416  uniB416,  HANGUL SYLLABLE-B416
        '46103' => '10896'            , # G+10896  U+B417  uniB417,  HANGUL SYLLABLE-B417
        '46104' => '1594'             , # G+1594  U+B418  uniB418,  HANGUL SYLLABLE-B418
        '46105' => '10897'            , # G+10897  U+B419  uniB419,  HANGUL SYLLABLE-B419
        '46106' => '10898'            , # G+10898  U+B41A  uniB41A,  HANGUL SYLLABLE-B41A
        '46107' => '10899'            , # G+10899  U+B41B  uniB41B,  HANGUL SYLLABLE-B41B
        '46108' => '1595'             , # G+1595  U+B41C  uniB41C,  HANGUL SYLLABLE-B41C
        '46109' => '10900'            , # G+10900  U+B41D  uniB41D,  HANGUL SYLLABLE-B41D
        '46110' => '10901'            , # G+10901  U+B41E  uniB41E,  HANGUL SYLLABLE-B41E
        '46111' => '10902'            , # G+10902  U+B41F  uniB41F,  HANGUL SYLLABLE-B41F
        '46112' => '1596'             , # G+1596  U+B420  uniB420,  HANGUL SYLLABLE-B420
        '46113' => '10903'            , # G+10903  U+B421  uniB421,  HANGUL SYLLABLE-B421
        '46114' => '10904'            , # G+10904  U+B422  uniB422,  HANGUL SYLLABLE-B422
        '46115' => '10905'            , # G+10905  U+B423  uniB423,  HANGUL SYLLABLE-B423
        '46116' => '10906'            , # G+10906  U+B424  uniB424,  HANGUL SYLLABLE-B424
        '46117' => '10907'            , # G+10907  U+B425  uniB425,  HANGUL SYLLABLE-B425
        '46118' => '10908'            , # G+10908  U+B426  uniB426,  HANGUL SYLLABLE-B426
        '46119' => '10909'            , # G+10909  U+B427  uniB427,  HANGUL SYLLABLE-B427
        '46120' => '1597'             , # G+1597  U+B428  uniB428,  HANGUL SYLLABLE-B428
        '46121' => '1598'             , # G+1598  U+B429  uniB429,  HANGUL SYLLABLE-B429
        '46122' => '10910'            , # G+10910  U+B42A  uniB42A,  HANGUL SYLLABLE-B42A
        '46123' => '1599'             , # G+1599  U+B42B  uniB42B,  HANGUL SYLLABLE-B42B
        '46124' => '10911'            , # G+10911  U+B42C  uniB42C,  HANGUL SYLLABLE-B42C
        '46125' => '10912'            , # G+10912  U+B42D  uniB42D,  HANGUL SYLLABLE-B42D
        '46126' => '10913'            , # G+10913  U+B42E  uniB42E,  HANGUL SYLLABLE-B42E
        '46127' => '10914'            , # G+10914  U+B42F  uniB42F,  HANGUL SYLLABLE-B42F
        '46128' => '10915'            , # G+10915  U+B430  uniB430,  HANGUL SYLLABLE-B430
        '46129' => '10916'            , # G+10916  U+B431  uniB431,  HANGUL SYLLABLE-B431
        '46130' => '10917'            , # G+10917  U+B432  uniB432,  HANGUL SYLLABLE-B432
        '46131' => '10918'            , # G+10918  U+B433  uniB433,  HANGUL SYLLABLE-B433
        '46132' => '1600'             , # G+1600  U+B434  uniB434,  HANGUL SYLLABLE-B434
        '46133' => '10919'            , # G+10919  U+B435  uniB435,  HANGUL SYLLABLE-B435
        '46134' => '10920'            , # G+10920  U+B436  uniB436,  HANGUL SYLLABLE-B436
        '46135' => '10921'            , # G+10921  U+B437  uniB437,  HANGUL SYLLABLE-B437
        '46136' => '10922'            , # G+10922  U+B438  uniB438,  HANGUL SYLLABLE-B438
        '46137' => '10923'            , # G+10923  U+B439  uniB439,  HANGUL SYLLABLE-B439
        '46138' => '10924'            , # G+10924  U+B43A  uniB43A,  HANGUL SYLLABLE-B43A
        '46139' => '10925'            , # G+10925  U+B43B  uniB43B,  HANGUL SYLLABLE-B43B
        '46140' => '10926'            , # G+10926  U+B43C  uniB43C,  HANGUL SYLLABLE-B43C
        '46141' => '10927'            , # G+10927  U+B43D  uniB43D,  HANGUL SYLLABLE-B43D
        '46142' => '10928'            , # G+10928  U+B43E  uniB43E,  HANGUL SYLLABLE-B43E
        '46143' => '10929'            , # G+10929  U+B43F  uniB43F,  HANGUL SYLLABLE-B43F
        '46144' => '10930'            , # G+10930  U+B440  uniB440,  HANGUL SYLLABLE-B440
        '46145' => '10931'            , # G+10931  U+B441  uniB441,  HANGUL SYLLABLE-B441
        '46146' => '10932'            , # G+10932  U+B442  uniB442,  HANGUL SYLLABLE-B442
        '46147' => '10933'            , # G+10933  U+B443  uniB443,  HANGUL SYLLABLE-B443
        '46148' => '10934'            , # G+10934  U+B444  uniB444,  HANGUL SYLLABLE-B444
        '46149' => '10935'            , # G+10935  U+B445  uniB445,  HANGUL SYLLABLE-B445
        '46150' => '10936'            , # G+10936  U+B446  uniB446,  HANGUL SYLLABLE-B446
        '46151' => '10937'            , # G+10937  U+B447  uniB447,  HANGUL SYLLABLE-B447
        '46152' => '10938'            , # G+10938  U+B448  uniB448,  HANGUL SYLLABLE-B448
        '46153' => '10939'            , # G+10939  U+B449  uniB449,  HANGUL SYLLABLE-B449
        '46154' => '10940'            , # G+10940  U+B44A  uniB44A,  HANGUL SYLLABLE-B44A
        '46155' => '10941'            , # G+10941  U+B44B  uniB44B,  HANGUL SYLLABLE-B44B
        '46156' => '10942'            , # G+10942  U+B44C  uniB44C,  HANGUL SYLLABLE-B44C
        '46157' => '10943'            , # G+10943  U+B44D  uniB44D,  HANGUL SYLLABLE-B44D
        '46158' => '10944'            , # G+10944  U+B44E  uniB44E,  HANGUL SYLLABLE-B44E
        '46159' => '10945'            , # G+10945  U+B44F  uniB44F,  HANGUL SYLLABLE-B44F
        '46160' => '1601'             , # G+1601  U+B450  uniB450,  HANGUL SYLLABLE-B450
        '46161' => '1602'             , # G+1602  U+B451  uniB451,  HANGUL SYLLABLE-B451
        '46162' => '10946'            , # G+10946  U+B452  uniB452,  HANGUL SYLLABLE-B452
        '46163' => '10947'            , # G+10947  U+B453  uniB453,  HANGUL SYLLABLE-B453
        '46164' => '1603'             , # G+1603  U+B454  uniB454,  HANGUL SYLLABLE-B454
        '46165' => '10948'            , # G+10948  U+B455  uniB455,  HANGUL SYLLABLE-B455
        '46166' => '10949'            , # G+10949  U+B456  uniB456,  HANGUL SYLLABLE-B456
        '46167' => '10950'            , # G+10950  U+B457  uniB457,  HANGUL SYLLABLE-B457
        '46168' => '1604'             , # G+1604  U+B458  uniB458,  HANGUL SYLLABLE-B458
        '46169' => '10951'            , # G+10951  U+B459  uniB459,  HANGUL SYLLABLE-B459
        '46170' => '10952'            , # G+10952  U+B45A  uniB45A,  HANGUL SYLLABLE-B45A
        '46171' => '10953'            , # G+10953  U+B45B  uniB45B,  HANGUL SYLLABLE-B45B
        '46172' => '10954'            , # G+10954  U+B45C  uniB45C,  HANGUL SYLLABLE-B45C
        '46173' => '10955'            , # G+10955  U+B45D  uniB45D,  HANGUL SYLLABLE-B45D
        '46174' => '10956'            , # G+10956  U+B45E  uniB45E,  HANGUL SYLLABLE-B45E
        '46175' => '10957'            , # G+10957  U+B45F  uniB45F,  HANGUL SYLLABLE-B45F
        '46176' => '1605'             , # G+1605  U+B460  uniB460,  HANGUL SYLLABLE-B460
        '46177' => '1606'             , # G+1606  U+B461  uniB461,  HANGUL SYLLABLE-B461
        '46178' => '10958'            , # G+10958  U+B462  uniB462,  HANGUL SYLLABLE-B462
        '46179' => '1607'             , # G+1607  U+B463  uniB463,  HANGUL SYLLABLE-B463
        '46180' => '10959'            , # G+10959  U+B464  uniB464,  HANGUL SYLLABLE-B464
        '46181' => '1608'             , # G+1608  U+B465  uniB465,  HANGUL SYLLABLE-B465
        '46182' => '10960'            , # G+10960  U+B466  uniB466,  HANGUL SYLLABLE-B466
        '46183' => '10961'            , # G+10961  U+B467  uniB467,  HANGUL SYLLABLE-B467
        '46184' => '10962'            , # G+10962  U+B468  uniB468,  HANGUL SYLLABLE-B468
        '46185' => '10963'            , # G+10963  U+B469  uniB469,  HANGUL SYLLABLE-B469
        '46186' => '10964'            , # G+10964  U+B46A  uniB46A,  HANGUL SYLLABLE-B46A
        '46187' => '10965'            , # G+10965  U+B46B  uniB46B,  HANGUL SYLLABLE-B46B
        '46188' => '1609'             , # G+1609  U+B46C  uniB46C,  HANGUL SYLLABLE-B46C
        '46189' => '10966'            , # G+10966  U+B46D  uniB46D,  HANGUL SYLLABLE-B46D
        '46190' => '10967'            , # G+10967  U+B46E  uniB46E,  HANGUL SYLLABLE-B46E
        '46191' => '10968'            , # G+10968  U+B46F  uniB46F,  HANGUL SYLLABLE-B46F
        '46192' => '10969'            , # G+10969  U+B470  uniB470,  HANGUL SYLLABLE-B470
        '46193' => '10970'            , # G+10970  U+B471  uniB471,  HANGUL SYLLABLE-B471
        '46194' => '10971'            , # G+10971  U+B472  uniB472,  HANGUL SYLLABLE-B472
        '46195' => '10972'            , # G+10972  U+B473  uniB473,  HANGUL SYLLABLE-B473
        '46196' => '10973'            , # G+10973  U+B474  uniB474,  HANGUL SYLLABLE-B474
        '46197' => '10974'            , # G+10974  U+B475  uniB475,  HANGUL SYLLABLE-B475
        '46198' => '10975'            , # G+10975  U+B476  uniB476,  HANGUL SYLLABLE-B476
        '46199' => '10976'            , # G+10976  U+B477  uniB477,  HANGUL SYLLABLE-B477
        '46200' => '10977'            , # G+10977  U+B478  uniB478,  HANGUL SYLLABLE-B478
        '46201' => '10978'            , # G+10978  U+B479  uniB479,  HANGUL SYLLABLE-B479
        '46202' => '10979'            , # G+10979  U+B47A  uniB47A,  HANGUL SYLLABLE-B47A
        '46203' => '10980'            , # G+10980  U+B47B  uniB47B,  HANGUL SYLLABLE-B47B
        '46204' => '10981'            , # G+10981  U+B47C  uniB47C,  HANGUL SYLLABLE-B47C
        '46205' => '10982'            , # G+10982  U+B47D  uniB47D,  HANGUL SYLLABLE-B47D
        '46206' => '10983'            , # G+10983  U+B47E  uniB47E,  HANGUL SYLLABLE-B47E
        '46207' => '10984'            , # G+10984  U+B47F  uniB47F,  HANGUL SYLLABLE-B47F
        '46208' => '1610'             , # G+1610  U+B480  uniB480,  HANGUL SYLLABLE-B480
        '46209' => '10985'            , # G+10985  U+B481  uniB481,  HANGUL SYLLABLE-B481
        '46210' => '10986'            , # G+10986  U+B482  uniB482,  HANGUL SYLLABLE-B482
        '46211' => '10987'            , # G+10987  U+B483  uniB483,  HANGUL SYLLABLE-B483
        '46212' => '10988'            , # G+10988  U+B484  uniB484,  HANGUL SYLLABLE-B484
        '46213' => '10989'            , # G+10989  U+B485  uniB485,  HANGUL SYLLABLE-B485
        '46214' => '10990'            , # G+10990  U+B486  uniB486,  HANGUL SYLLABLE-B486
        '46215' => '10991'            , # G+10991  U+B487  uniB487,  HANGUL SYLLABLE-B487
        '46216' => '1611'             , # G+1611  U+B488  uniB488,  HANGUL SYLLABLE-B488
        '46217' => '10992'            , # G+10992  U+B489  uniB489,  HANGUL SYLLABLE-B489
        '46218' => '10993'            , # G+10993  U+B48A  uniB48A,  HANGUL SYLLABLE-B48A
        '46219' => '10994'            , # G+10994  U+B48B  uniB48B,  HANGUL SYLLABLE-B48B
        '46220' => '10995'            , # G+10995  U+B48C  uniB48C,  HANGUL SYLLABLE-B48C
        '46221' => '10996'            , # G+10996  U+B48D  uniB48D,  HANGUL SYLLABLE-B48D
        '46222' => '10997'            , # G+10997  U+B48E  uniB48E,  HANGUL SYLLABLE-B48E
        '46223' => '10998'            , # G+10998  U+B48F  uniB48F,  HANGUL SYLLABLE-B48F
        '46224' => '10999'            , # G+10999  U+B490  uniB490,  HANGUL SYLLABLE-B490
        '46225' => '11000'            , # G+11000  U+B491  uniB491,  HANGUL SYLLABLE-B491
        '46226' => '11001'            , # G+11001  U+B492  uniB492,  HANGUL SYLLABLE-B492
        '46227' => '11002'            , # G+11002  U+B493  uniB493,  HANGUL SYLLABLE-B493
        '46228' => '11003'            , # G+11003  U+B494  uniB494,  HANGUL SYLLABLE-B494
        '46229' => '11004'            , # G+11004  U+B495  uniB495,  HANGUL SYLLABLE-B495
        '46230' => '11005'            , # G+11005  U+B496  uniB496,  HANGUL SYLLABLE-B496
        '46231' => '11006'            , # G+11006  U+B497  uniB497,  HANGUL SYLLABLE-B497
        '46232' => '11007'            , # G+11007  U+B498  uniB498,  HANGUL SYLLABLE-B498
        '46233' => '11008'            , # G+11008  U+B499  uniB499,  HANGUL SYLLABLE-B499
        '46234' => '11009'            , # G+11009  U+B49A  uniB49A,  HANGUL SYLLABLE-B49A
        '46235' => '11010'            , # G+11010  U+B49B  uniB49B,  HANGUL SYLLABLE-B49B
        '46236' => '11011'            , # G+11011  U+B49C  uniB49C,  HANGUL SYLLABLE-B49C
        '46237' => '1612'             , # G+1612  U+B49D  uniB49D,  HANGUL SYLLABLE-B49D
        '46238' => '11012'            , # G+11012  U+B49E  uniB49E,  HANGUL SYLLABLE-B49E
        '46239' => '11013'            , # G+11013  U+B49F  uniB49F,  HANGUL SYLLABLE-B49F
        '46240' => '11014'            , # G+11014  U+B4A0  uniB4A0,  HANGUL SYLLABLE-B4A0
        '46241' => '11015'            , # G+11015  U+B4A1  uniB4A1,  HANGUL SYLLABLE-B4A1
        '46242' => '11016'            , # G+11016  U+B4A2  uniB4A2,  HANGUL SYLLABLE-B4A2
        '46243' => '11017'            , # G+11017  U+B4A3  uniB4A3,  HANGUL SYLLABLE-B4A3
        '46244' => '1613'             , # G+1613  U+B4A4  uniB4A4,  HANGUL SYLLABLE-B4A4
        '46245' => '11018'            , # G+11018  U+B4A5  uniB4A5,  HANGUL SYLLABLE-B4A5
        '46246' => '11019'            , # G+11019  U+B4A6  uniB4A6,  HANGUL SYLLABLE-B4A6
        '46247' => '11020'            , # G+11020  U+B4A7  uniB4A7,  HANGUL SYLLABLE-B4A7
        '46248' => '1614'             , # G+1614  U+B4A8  uniB4A8,  HANGUL SYLLABLE-B4A8
        '46249' => '11021'            , # G+11021  U+B4A9  uniB4A9,  HANGUL SYLLABLE-B4A9
        '46250' => '11022'            , # G+11022  U+B4AA  uniB4AA,  HANGUL SYLLABLE-B4AA
        '46251' => '11023'            , # G+11023  U+B4AB  uniB4AB,  HANGUL SYLLABLE-B4AB
        '46252' => '1615'             , # G+1615  U+B4AC  uniB4AC,  HANGUL SYLLABLE-B4AC
        '46253' => '11024'            , # G+11024  U+B4AD  uniB4AD,  HANGUL SYLLABLE-B4AD
        '46254' => '11025'            , # G+11025  U+B4AE  uniB4AE,  HANGUL SYLLABLE-B4AE
        '46255' => '11026'            , # G+11026  U+B4AF  uniB4AF,  HANGUL SYLLABLE-B4AF
        '46256' => '11027'            , # G+11027  U+B4B0  uniB4B0,  HANGUL SYLLABLE-B4B0
        '46257' => '11028'            , # G+11028  U+B4B1  uniB4B1,  HANGUL SYLLABLE-B4B1
        '46258' => '11029'            , # G+11029  U+B4B2  uniB4B2,  HANGUL SYLLABLE-B4B2
        '46259' => '11030'            , # G+11030  U+B4B3  uniB4B3,  HANGUL SYLLABLE-B4B3
        '46260' => '11031'            , # G+11031  U+B4B4  uniB4B4,  HANGUL SYLLABLE-B4B4
        '46261' => '1616'             , # G+1616  U+B4B5  uniB4B5,  HANGUL SYLLABLE-B4B5
        '46262' => '11032'            , # G+11032  U+B4B6  uniB4B6,  HANGUL SYLLABLE-B4B6
        '46263' => '1617'             , # G+1617  U+B4B7  uniB4B7,  HANGUL SYLLABLE-B4B7
        '46264' => '11033'            , # G+11033  U+B4B8  uniB4B8,  HANGUL SYLLABLE-B4B8
        '46265' => '1618'             , # G+1618  U+B4B9  uniB4B9,  HANGUL SYLLABLE-B4B9
        '46266' => '11034'            , # G+11034  U+B4BA  uniB4BA,  HANGUL SYLLABLE-B4BA
        '46267' => '11035'            , # G+11035  U+B4BB  uniB4BB,  HANGUL SYLLABLE-B4BB
        '46268' => '11036'            , # G+11036  U+B4BC  uniB4BC,  HANGUL SYLLABLE-B4BC
        '46269' => '11037'            , # G+11037  U+B4BD  uniB4BD,  HANGUL SYLLABLE-B4BD
        '46270' => '11038'            , # G+11038  U+B4BE  uniB4BE,  HANGUL SYLLABLE-B4BE
        '46271' => '11039'            , # G+11039  U+B4BF  uniB4BF,  HANGUL SYLLABLE-B4BF
        '46272' => '1619'             , # G+1619  U+B4C0  uniB4C0,  HANGUL SYLLABLE-B4C0
        '46273' => '11040'            , # G+11040  U+B4C1  uniB4C1,  HANGUL SYLLABLE-B4C1
        '46274' => '11041'            , # G+11041  U+B4C2  uniB4C2,  HANGUL SYLLABLE-B4C2
        '46275' => '11042'            , # G+11042  U+B4C3  uniB4C3,  HANGUL SYLLABLE-B4C3
        '46276' => '1620'             , # G+1620  U+B4C4  uniB4C4,  HANGUL SYLLABLE-B4C4
        '46277' => '11043'            , # G+11043  U+B4C5  uniB4C5,  HANGUL SYLLABLE-B4C5
        '46278' => '11044'            , # G+11044  U+B4C6  uniB4C6,  HANGUL SYLLABLE-B4C6
        '46279' => '11045'            , # G+11045  U+B4C7  uniB4C7,  HANGUL SYLLABLE-B4C7
        '46280' => '1621'             , # G+1621  U+B4C8  uniB4C8,  HANGUL SYLLABLE-B4C8
        '46281' => '11046'            , # G+11046  U+B4C9  uniB4C9,  HANGUL SYLLABLE-B4C9
        '46282' => '11047'            , # G+11047  U+B4CA  uniB4CA,  HANGUL SYLLABLE-B4CA
        '46283' => '11048'            , # G+11048  U+B4CB  uniB4CB,  HANGUL SYLLABLE-B4CB
        '46284' => '11049'            , # G+11049  U+B4CC  uniB4CC,  HANGUL SYLLABLE-B4CC
        '46285' => '11050'            , # G+11050  U+B4CD  uniB4CD,  HANGUL SYLLABLE-B4CD
        '46286' => '11051'            , # G+11051  U+B4CE  uniB4CE,  HANGUL SYLLABLE-B4CE
        '46287' => '11052'            , # G+11052  U+B4CF  uniB4CF,  HANGUL SYLLABLE-B4CF
        '46288' => '1622'             , # G+1622  U+B4D0  uniB4D0,  HANGUL SYLLABLE-B4D0
        '46289' => '11053'            , # G+11053  U+B4D1  uniB4D1,  HANGUL SYLLABLE-B4D1
        '46290' => '11054'            , # G+11054  U+B4D2  uniB4D2,  HANGUL SYLLABLE-B4D2
        '46291' => '11055'            , # G+11055  U+B4D3  uniB4D3,  HANGUL SYLLABLE-B4D3
        '46292' => '11056'            , # G+11056  U+B4D4  uniB4D4,  HANGUL SYLLABLE-B4D4
        '46293' => '1623'             , # G+1623  U+B4D5  uniB4D5,  HANGUL SYLLABLE-B4D5
        '46294' => '11057'            , # G+11057  U+B4D6  uniB4D6,  HANGUL SYLLABLE-B4D6
        '46295' => '11058'            , # G+11058  U+B4D7  uniB4D7,  HANGUL SYLLABLE-B4D7
        '46296' => '11059'            , # G+11059  U+B4D8  uniB4D8,  HANGUL SYLLABLE-B4D8
        '46297' => '11060'            , # G+11060  U+B4D9  uniB4D9,  HANGUL SYLLABLE-B4D9
        '46298' => '11061'            , # G+11061  U+B4DA  uniB4DA,  HANGUL SYLLABLE-B4DA
        '46299' => '11062'            , # G+11062  U+B4DB  uniB4DB,  HANGUL SYLLABLE-B4DB
        '46300' => '1624'             , # G+1624  U+B4DC  uniB4DC,  HANGUL SYLLABLE-B4DC
        '46301' => '1625'             , # G+1625  U+B4DD  uniB4DD,  HANGUL SYLLABLE-B4DD
        '46302' => '11063'            , # G+11063  U+B4DE  uniB4DE,  HANGUL SYLLABLE-B4DE
        '46303' => '11064'            , # G+11064  U+B4DF  uniB4DF,  HANGUL SYLLABLE-B4DF
        '46304' => '1626'             , # G+1626  U+B4E0  uniB4E0,  HANGUL SYLLABLE-B4E0
        '46305' => '11065'            , # G+11065  U+B4E1  uniB4E1,  HANGUL SYLLABLE-B4E1
        '46306' => '11066'            , # G+11066  U+B4E2  uniB4E2,  HANGUL SYLLABLE-B4E2
        '46307' => '1627'             , # G+1627  U+B4E3  uniB4E3,  HANGUL SYLLABLE-B4E3
        '46308' => '1628'             , # G+1628  U+B4E4  uniB4E4,  HANGUL SYLLABLE-B4E4
        '46309' => '11067'            , # G+11067  U+B4E5  uniB4E5,  HANGUL SYLLABLE-B4E5
        '46310' => '1629'             , # G+1629  U+B4E6  uniB4E6,  HANGUL SYLLABLE-B4E6
        '46311' => '11068'            , # G+11068  U+B4E7  uniB4E7,  HANGUL SYLLABLE-B4E7
        '46312' => '11069'            , # G+11069  U+B4E8  uniB4E8,  HANGUL SYLLABLE-B4E8
        '46313' => '11070'            , # G+11070  U+B4E9  uniB4E9,  HANGUL SYLLABLE-B4E9
        '46314' => '11071'            , # G+11071  U+B4EA  uniB4EA,  HANGUL SYLLABLE-B4EA
        '46315' => '11072'            , # G+11072  U+B4EB  uniB4EB,  HANGUL SYLLABLE-B4EB
        '46316' => '1630'             , # G+1630  U+B4EC  uniB4EC,  HANGUL SYLLABLE-B4EC
        '46317' => '1631'             , # G+1631  U+B4ED  uniB4ED,  HANGUL SYLLABLE-B4ED
        '46318' => '11073'            , # G+11073  U+B4EE  uniB4EE,  HANGUL SYLLABLE-B4EE
        '46319' => '1632'             , # G+1632  U+B4EF  uniB4EF,  HANGUL SYLLABLE-B4EF
        '46320' => '11074'            , # G+11074  U+B4F0  uniB4F0,  HANGUL SYLLABLE-B4F0
        '46321' => '1633'             , # G+1633  U+B4F1  uniB4F1,  HANGUL SYLLABLE-B4F1
        '46322' => '11075'            , # G+11075  U+B4F2  uniB4F2,  HANGUL SYLLABLE-B4F2
        '46323' => '11076'            , # G+11076  U+B4F3  uniB4F3,  HANGUL SYLLABLE-B4F3
        '46324' => '11077'            , # G+11077  U+B4F4  uniB4F4,  HANGUL SYLLABLE-B4F4
        '46325' => '11078'            , # G+11078  U+B4F5  uniB4F5,  HANGUL SYLLABLE-B4F5
        '46326' => '11079'            , # G+11079  U+B4F6  uniB4F6,  HANGUL SYLLABLE-B4F6
        '46327' => '11080'            , # G+11080  U+B4F7  uniB4F7,  HANGUL SYLLABLE-B4F7
        '46328' => '1634'             , # G+1634  U+B4F8  uniB4F8,  HANGUL SYLLABLE-B4F8
        '46329' => '11081'            , # G+11081  U+B4F9  uniB4F9,  HANGUL SYLLABLE-B4F9
        '46330' => '11082'            , # G+11082  U+B4FA  uniB4FA,  HANGUL SYLLABLE-B4FA
        '46331' => '11083'            , # G+11083  U+B4FB  uniB4FB,  HANGUL SYLLABLE-B4FB
        '46332' => '11084'            , # G+11084  U+B4FC  uniB4FC,  HANGUL SYLLABLE-B4FC
        '46333' => '11085'            , # G+11085  U+B4FD  uniB4FD,  HANGUL SYLLABLE-B4FD
        '46334' => '11086'            , # G+11086  U+B4FE  uniB4FE,  HANGUL SYLLABLE-B4FE
        '46335' => '11087'            , # G+11087  U+B4FF  uniB4FF,  HANGUL SYLLABLE-B4FF
        '46336' => '11088'            , # G+11088  U+B500  uniB500,  HANGUL SYLLABLE-B500
        '46337' => '11089'            , # G+11089  U+B501  uniB501,  HANGUL SYLLABLE-B501
        '46338' => '11090'            , # G+11090  U+B502  uniB502,  HANGUL SYLLABLE-B502
        '46339' => '11091'            , # G+11091  U+B503  uniB503,  HANGUL SYLLABLE-B503
        '46340' => '11092'            , # G+11092  U+B504  uniB504,  HANGUL SYLLABLE-B504
        '46341' => '11093'            , # G+11093  U+B505  uniB505,  HANGUL SYLLABLE-B505
        '46342' => '11094'            , # G+11094  U+B506  uniB506,  HANGUL SYLLABLE-B506
        '46343' => '11095'            , # G+11095  U+B507  uniB507,  HANGUL SYLLABLE-B507
        '46344' => '11096'            , # G+11096  U+B508  uniB508,  HANGUL SYLLABLE-B508
        '46345' => '11097'            , # G+11097  U+B509  uniB509,  HANGUL SYLLABLE-B509
        '46346' => '11098'            , # G+11098  U+B50A  uniB50A,  HANGUL SYLLABLE-B50A
        '46347' => '11099'            , # G+11099  U+B50B  uniB50B,  HANGUL SYLLABLE-B50B
        '46348' => '11100'            , # G+11100  U+B50C  uniB50C,  HANGUL SYLLABLE-B50C
        '46349' => '11101'            , # G+11101  U+B50D  uniB50D,  HANGUL SYLLABLE-B50D
        '46350' => '11102'            , # G+11102  U+B50E  uniB50E,  HANGUL SYLLABLE-B50E
        '46351' => '11103'            , # G+11103  U+B50F  uniB50F,  HANGUL SYLLABLE-B50F
        '46352' => '11104'            , # G+11104  U+B510  uniB510,  HANGUL SYLLABLE-B510
        '46353' => '11105'            , # G+11105  U+B511  uniB511,  HANGUL SYLLABLE-B511
        '46354' => '11106'            , # G+11106  U+B512  uniB512,  HANGUL SYLLABLE-B512
        '46355' => '11107'            , # G+11107  U+B513  uniB513,  HANGUL SYLLABLE-B513
        '46356' => '1635'             , # G+1635  U+B514  uniB514,  HANGUL SYLLABLE-B514
        '46357' => '1636'             , # G+1636  U+B515  uniB515,  HANGUL SYLLABLE-B515
        '46358' => '11108'            , # G+11108  U+B516  uniB516,  HANGUL SYLLABLE-B516
        '46359' => '11109'            , # G+11109  U+B517  uniB517,  HANGUL SYLLABLE-B517
        '46360' => '1637'             , # G+1637  U+B518  uniB518,  HANGUL SYLLABLE-B518
        '46361' => '11110'            , # G+11110  U+B519  uniB519,  HANGUL SYLLABLE-B519
        '46362' => '11111'            , # G+11111  U+B51A  uniB51A,  HANGUL SYLLABLE-B51A
        '46363' => '1638'             , # G+1638  U+B51B  uniB51B,  HANGUL SYLLABLE-B51B
        '46364' => '1639'             , # G+1639  U+B51C  uniB51C,  HANGUL SYLLABLE-B51C
        '46365' => '11112'            , # G+11112  U+B51D  uniB51D,  HANGUL SYLLABLE-B51D
        '46366' => '11113'            , # G+11113  U+B51E  uniB51E,  HANGUL SYLLABLE-B51E
        '46367' => '11114'            , # G+11114  U+B51F  uniB51F,  HANGUL SYLLABLE-B51F
        '46368' => '11115'            , # G+11115  U+B520  uniB520,  HANGUL SYLLABLE-B520
        '46369' => '11116'            , # G+11116  U+B521  uniB521,  HANGUL SYLLABLE-B521
        '46370' => '11117'            , # G+11117  U+B522  uniB522,  HANGUL SYLLABLE-B522
        '46371' => '11118'            , # G+11118  U+B523  uniB523,  HANGUL SYLLABLE-B523
        '46372' => '1640'             , # G+1640  U+B524  uniB524,  HANGUL SYLLABLE-B524
        '46373' => '1641'             , # G+1641  U+B525  uniB525,  HANGUL SYLLABLE-B525
        '46374' => '11119'            , # G+11119  U+B526  uniB526,  HANGUL SYLLABLE-B526
        '46375' => '1642'             , # G+1642  U+B527  uniB527,  HANGUL SYLLABLE-B527
        '46376' => '1643'             , # G+1643  U+B528  uniB528,  HANGUL SYLLABLE-B528
        '46377' => '1644'             , # G+1644  U+B529  uniB529,  HANGUL SYLLABLE-B529
        '46378' => '1645'             , # G+1645  U+B52A  uniB52A,  HANGUL SYLLABLE-B52A
        '46379' => '11120'            , # G+11120  U+B52B  uniB52B,  HANGUL SYLLABLE-B52B
        '46380' => '11121'            , # G+11121  U+B52C  uniB52C,  HANGUL SYLLABLE-B52C
        '46381' => '11122'            , # G+11122  U+B52D  uniB52D,  HANGUL SYLLABLE-B52D
        '46382' => '11123'            , # G+11123  U+B52E  uniB52E,  HANGUL SYLLABLE-B52E
        '46383' => '11124'            , # G+11124  U+B52F  uniB52F,  HANGUL SYLLABLE-B52F
        '46384' => '1646'             , # G+1646  U+B530  uniB530,  HANGUL SYLLABLE-B530
        '46385' => '1647'             , # G+1647  U+B531  uniB531,  HANGUL SYLLABLE-B531
        '46386' => '11125'            , # G+11125  U+B532  uniB532,  HANGUL SYLLABLE-B532
        '46387' => '11126'            , # G+11126  U+B533  uniB533,  HANGUL SYLLABLE-B533
        '46388' => '1648'             , # G+1648  U+B534  uniB534,  HANGUL SYLLABLE-B534
        '46389' => '11127'            , # G+11127  U+B535  uniB535,  HANGUL SYLLABLE-B535
        '46390' => '11128'            , # G+11128  U+B536  uniB536,  HANGUL SYLLABLE-B536
        '46391' => '11129'            , # G+11129  U+B537  uniB537,  HANGUL SYLLABLE-B537
        '46392' => '1649'             , # G+1649  U+B538  uniB538,  HANGUL SYLLABLE-B538
        '46393' => '11130'            , # G+11130  U+B539  uniB539,  HANGUL SYLLABLE-B539
        '46394' => '11131'            , # G+11131  U+B53A  uniB53A,  HANGUL SYLLABLE-B53A
        '46395' => '11132'            , # G+11132  U+B53B  uniB53B,  HANGUL SYLLABLE-B53B
        '46396' => '11133'            , # G+11133  U+B53C  uniB53C,  HANGUL SYLLABLE-B53C
        '46397' => '11134'            , # G+11134  U+B53D  uniB53D,  HANGUL SYLLABLE-B53D
        '46398' => '11135'            , # G+11135  U+B53E  uniB53E,  HANGUL SYLLABLE-B53E
        '46399' => '11136'            , # G+11136  U+B53F  uniB53F,  HANGUL SYLLABLE-B53F
        '46400' => '1650'             , # G+1650  U+B540  uniB540,  HANGUL SYLLABLE-B540
        '46401' => '1651'             , # G+1651  U+B541  uniB541,  HANGUL SYLLABLE-B541
        '46402' => '11137'            , # G+11137  U+B542  uniB542,  HANGUL SYLLABLE-B542
        '46403' => '1652'             , # G+1652  U+B543  uniB543,  HANGUL SYLLABLE-B543
        '46404' => '1653'             , # G+1653  U+B544  uniB544,  HANGUL SYLLABLE-B544
        '46405' => '1654'             , # G+1654  U+B545  uniB545,  HANGUL SYLLABLE-B545
        '46406' => '11138'            , # G+11138  U+B546  uniB546,  HANGUL SYLLABLE-B546
        '46407' => '11139'            , # G+11139  U+B547  uniB547,  HANGUL SYLLABLE-B547
        '46408' => '11140'            , # G+11140  U+B548  uniB548,  HANGUL SYLLABLE-B548
        '46409' => '11141'            , # G+11141  U+B549  uniB549,  HANGUL SYLLABLE-B549
        '46410' => '11142'            , # G+11142  U+B54A  uniB54A,  HANGUL SYLLABLE-B54A
        '46411' => '1655'             , # G+1655  U+B54B  uniB54B,  HANGUL SYLLABLE-B54B
        '46412' => '1656'             , # G+1656  U+B54C  uniB54C,  HANGUL SYLLABLE-B54C
        '46413' => '1657'             , # G+1657  U+B54D  uniB54D,  HANGUL SYLLABLE-B54D
        '46414' => '11143'            , # G+11143  U+B54E  uniB54E,  HANGUL SYLLABLE-B54E
        '46415' => '11144'            , # G+11144  U+B54F  uniB54F,  HANGUL SYLLABLE-B54F
        '46416' => '1658'             , # G+1658  U+B550  uniB550,  HANGUL SYLLABLE-B550
        '46417' => '11145'            , # G+11145  U+B551  uniB551,  HANGUL SYLLABLE-B551
        '46418' => '11146'            , # G+11146  U+B552  uniB552,  HANGUL SYLLABLE-B552
        '46419' => '11147'            , # G+11147  U+B553  uniB553,  HANGUL SYLLABLE-B553
        '46420' => '1659'             , # G+1659  U+B554  uniB554,  HANGUL SYLLABLE-B554
        '46421' => '11148'            , # G+11148  U+B555  uniB555,  HANGUL SYLLABLE-B555
        '46422' => '11149'            , # G+11149  U+B556  uniB556,  HANGUL SYLLABLE-B556
        '46423' => '11150'            , # G+11150  U+B557  uniB557,  HANGUL SYLLABLE-B557
        '46424' => '11151'            , # G+11151  U+B558  uniB558,  HANGUL SYLLABLE-B558
        '46425' => '11152'            , # G+11152  U+B559  uniB559,  HANGUL SYLLABLE-B559
        '46426' => '11153'            , # G+11153  U+B55A  uniB55A,  HANGUL SYLLABLE-B55A
        '46427' => '11154'            , # G+11154  U+B55B  uniB55B,  HANGUL SYLLABLE-B55B
        '46428' => '1660'             , # G+1660  U+B55C  uniB55C,  HANGUL SYLLABLE-B55C
        '46429' => '1661'             , # G+1661  U+B55D  uniB55D,  HANGUL SYLLABLE-B55D
        '46430' => '11155'            , # G+11155  U+B55E  uniB55E,  HANGUL SYLLABLE-B55E
        '46431' => '1662'             , # G+1662  U+B55F  uniB55F,  HANGUL SYLLABLE-B55F
        '46432' => '1663'             , # G+1663  U+B560  uniB560,  HANGUL SYLLABLE-B560
        '46433' => '1664'             , # G+1664  U+B561  uniB561,  HANGUL SYLLABLE-B561
        '46434' => '11156'            , # G+11156  U+B562  uniB562,  HANGUL SYLLABLE-B562
        '46435' => '11157'            , # G+11157  U+B563  uniB563,  HANGUL SYLLABLE-B563
        '46436' => '11158'            , # G+11158  U+B564  uniB564,  HANGUL SYLLABLE-B564
        '46437' => '11159'            , # G+11159  U+B565  uniB565,  HANGUL SYLLABLE-B565
        '46438' => '11160'            , # G+11160  U+B566  uniB566,  HANGUL SYLLABLE-B566
        '46439' => '11161'            , # G+11161  U+B567  uniB567,  HANGUL SYLLABLE-B567
        '46440' => '11162'            , # G+11162  U+B568  uniB568,  HANGUL SYLLABLE-B568
        '46441' => '11163'            , # G+11163  U+B569  uniB569,  HANGUL SYLLABLE-B569
        '46442' => '11164'            , # G+11164  U+B56A  uniB56A,  HANGUL SYLLABLE-B56A
        '46443' => '11165'            , # G+11165  U+B56B  uniB56B,  HANGUL SYLLABLE-B56B
        '46444' => '11166'            , # G+11166  U+B56C  uniB56C,  HANGUL SYLLABLE-B56C
        '46445' => '11167'            , # G+11167  U+B56D  uniB56D,  HANGUL SYLLABLE-B56D
        '46446' => '11168'            , # G+11168  U+B56E  uniB56E,  HANGUL SYLLABLE-B56E
        '46447' => '11169'            , # G+11169  U+B56F  uniB56F,  HANGUL SYLLABLE-B56F
        '46448' => '11170'            , # G+11170  U+B570  uniB570,  HANGUL SYLLABLE-B570
        '46449' => '11171'            , # G+11171  U+B571  uniB571,  HANGUL SYLLABLE-B571
        '46450' => '11172'            , # G+11172  U+B572  uniB572,  HANGUL SYLLABLE-B572
        '46451' => '11173'            , # G+11173  U+B573  uniB573,  HANGUL SYLLABLE-B573
        '46452' => '11174'            , # G+11174  U+B574  uniB574,  HANGUL SYLLABLE-B574
        '46453' => '11175'            , # G+11175  U+B575  uniB575,  HANGUL SYLLABLE-B575
        '46454' => '11176'            , # G+11176  U+B576  uniB576,  HANGUL SYLLABLE-B576
        '46455' => '11177'            , # G+11177  U+B577  uniB577,  HANGUL SYLLABLE-B577
        '46456' => '11178'            , # G+11178  U+B578  uniB578,  HANGUL SYLLABLE-B578
        '46457' => '11179'            , # G+11179  U+B579  uniB579,  HANGUL SYLLABLE-B579
        '46458' => '11180'            , # G+11180  U+B57A  uniB57A,  HANGUL SYLLABLE-B57A
        '46459' => '11181'            , # G+11181  U+B57B  uniB57B,  HANGUL SYLLABLE-B57B
        '46460' => '11182'            , # G+11182  U+B57C  uniB57C,  HANGUL SYLLABLE-B57C
        '46461' => '11183'            , # G+11183  U+B57D  uniB57D,  HANGUL SYLLABLE-B57D
        '46462' => '11184'            , # G+11184  U+B57E  uniB57E,  HANGUL SYLLABLE-B57E
        '46463' => '11185'            , # G+11185  U+B57F  uniB57F,  HANGUL SYLLABLE-B57F
        '46464' => '11186'            , # G+11186  U+B580  uniB580,  HANGUL SYLLABLE-B580
        '46465' => '11187'            , # G+11187  U+B581  uniB581,  HANGUL SYLLABLE-B581
        '46466' => '11188'            , # G+11188  U+B582  uniB582,  HANGUL SYLLABLE-B582
        '46467' => '11189'            , # G+11189  U+B583  uniB583,  HANGUL SYLLABLE-B583
        '46468' => '11190'            , # G+11190  U+B584  uniB584,  HANGUL SYLLABLE-B584
        '46469' => '11191'            , # G+11191  U+B585  uniB585,  HANGUL SYLLABLE-B585
        '46470' => '11192'            , # G+11192  U+B586  uniB586,  HANGUL SYLLABLE-B586
        '46471' => '11193'            , # G+11193  U+B587  uniB587,  HANGUL SYLLABLE-B587
        '46472' => '11194'            , # G+11194  U+B588  uniB588,  HANGUL SYLLABLE-B588
        '46473' => '11195'            , # G+11195  U+B589  uniB589,  HANGUL SYLLABLE-B589
        '46474' => '11196'            , # G+11196  U+B58A  uniB58A,  HANGUL SYLLABLE-B58A
        '46475' => '11197'            , # G+11197  U+B58B  uniB58B,  HANGUL SYLLABLE-B58B
        '46476' => '11198'            , # G+11198  U+B58C  uniB58C,  HANGUL SYLLABLE-B58C
        '46477' => '11199'            , # G+11199  U+B58D  uniB58D,  HANGUL SYLLABLE-B58D
        '46478' => '11200'            , # G+11200  U+B58E  uniB58E,  HANGUL SYLLABLE-B58E
        '46479' => '11201'            , # G+11201  U+B58F  uniB58F,  HANGUL SYLLABLE-B58F
        '46480' => '11202'            , # G+11202  U+B590  uniB590,  HANGUL SYLLABLE-B590
        '46481' => '11203'            , # G+11203  U+B591  uniB591,  HANGUL SYLLABLE-B591
        '46482' => '11204'            , # G+11204  U+B592  uniB592,  HANGUL SYLLABLE-B592
        '46483' => '11205'            , # G+11205  U+B593  uniB593,  HANGUL SYLLABLE-B593
        '46484' => '11206'            , # G+11206  U+B594  uniB594,  HANGUL SYLLABLE-B594
        '46485' => '11207'            , # G+11207  U+B595  uniB595,  HANGUL SYLLABLE-B595
        '46486' => '11208'            , # G+11208  U+B596  uniB596,  HANGUL SYLLABLE-B596
        '46487' => '11209'            , # G+11209  U+B597  uniB597,  HANGUL SYLLABLE-B597
        '46488' => '11210'            , # G+11210  U+B598  uniB598,  HANGUL SYLLABLE-B598
        '46489' => '11211'            , # G+11211  U+B599  uniB599,  HANGUL SYLLABLE-B599
        '46490' => '11212'            , # G+11212  U+B59A  uniB59A,  HANGUL SYLLABLE-B59A
        '46491' => '11213'            , # G+11213  U+B59B  uniB59B,  HANGUL SYLLABLE-B59B
        '46492' => '11214'            , # G+11214  U+B59C  uniB59C,  HANGUL SYLLABLE-B59C
        '46493' => '11215'            , # G+11215  U+B59D  uniB59D,  HANGUL SYLLABLE-B59D
        '46494' => '11216'            , # G+11216  U+B59E  uniB59E,  HANGUL SYLLABLE-B59E
        '46495' => '11217'            , # G+11217  U+B59F  uniB59F,  HANGUL SYLLABLE-B59F
        '46496' => '1665'             , # G+1665  U+B5A0  uniB5A0,  HANGUL SYLLABLE-B5A0
        '46497' => '1666'             , # G+1666  U+B5A1  uniB5A1,  HANGUL SYLLABLE-B5A1
        '46498' => '11218'            , # G+11218  U+B5A2  uniB5A2,  HANGUL SYLLABLE-B5A2
        '46499' => '11219'            , # G+11219  U+B5A3  uniB5A3,  HANGUL SYLLABLE-B5A3
        '46500' => '1667'             , # G+1667  U+B5A4  uniB5A4,  HANGUL SYLLABLE-B5A4
        '46501' => '11220'            , # G+11220  U+B5A5  uniB5A5,  HANGUL SYLLABLE-B5A5
        '46502' => '11221'            , # G+11221  U+B5A6  uniB5A6,  HANGUL SYLLABLE-B5A6
        '46503' => '11222'            , # G+11222  U+B5A7  uniB5A7,  HANGUL SYLLABLE-B5A7
        '46504' => '1668'             , # G+1668  U+B5A8  uniB5A8,  HANGUL SYLLABLE-B5A8
        '46505' => '11223'            , # G+11223  U+B5A9  uniB5A9,  HANGUL SYLLABLE-B5A9
        '46506' => '1669'             , # G+1669  U+B5AA  uniB5AA,  HANGUL SYLLABLE-B5AA
        '46507' => '1670'             , # G+1670  U+B5AB  uniB5AB,  HANGUL SYLLABLE-B5AB
        '46508' => '11224'            , # G+11224  U+B5AC  uniB5AC,  HANGUL SYLLABLE-B5AC
        '46509' => '11225'            , # G+11225  U+B5AD  uniB5AD,  HANGUL SYLLABLE-B5AD
        '46510' => '11226'            , # G+11226  U+B5AE  uniB5AE,  HANGUL SYLLABLE-B5AE
        '46511' => '11227'            , # G+11227  U+B5AF  uniB5AF,  HANGUL SYLLABLE-B5AF
        '46512' => '1671'             , # G+1671  U+B5B0  uniB5B0,  HANGUL SYLLABLE-B5B0
        '46513' => '1672'             , # G+1672  U+B5B1  uniB5B1,  HANGUL SYLLABLE-B5B1
        '46514' => '11228'            , # G+11228  U+B5B2  uniB5B2,  HANGUL SYLLABLE-B5B2
        '46515' => '1673'             , # G+1673  U+B5B3  uniB5B3,  HANGUL SYLLABLE-B5B3
        '46516' => '1674'             , # G+1674  U+B5B4  uniB5B4,  HANGUL SYLLABLE-B5B4
        '46517' => '1675'             , # G+1675  U+B5B5  uniB5B5,  HANGUL SYLLABLE-B5B5
        '46518' => '11229'            , # G+11229  U+B5B6  uniB5B6,  HANGUL SYLLABLE-B5B6
        '46519' => '11230'            , # G+11230  U+B5B7  uniB5B7,  HANGUL SYLLABLE-B5B7
        '46520' => '11231'            , # G+11231  U+B5B8  uniB5B8,  HANGUL SYLLABLE-B5B8
        '46521' => '11232'            , # G+11232  U+B5B9  uniB5B9,  HANGUL SYLLABLE-B5B9
        '46522' => '11233'            , # G+11233  U+B5BA  uniB5BA,  HANGUL SYLLABLE-B5BA
        '46523' => '1676'             , # G+1676  U+B5BB  uniB5BB,  HANGUL SYLLABLE-B5BB
        '46524' => '1677'             , # G+1677  U+B5BC  uniB5BC,  HANGUL SYLLABLE-B5BC
        '46525' => '1678'             , # G+1678  U+B5BD  uniB5BD,  HANGUL SYLLABLE-B5BD
        '46526' => '11234'            , # G+11234  U+B5BE  uniB5BE,  HANGUL SYLLABLE-B5BE
        '46527' => '11235'            , # G+11235  U+B5BF  uniB5BF,  HANGUL SYLLABLE-B5BF
        '46528' => '1679'             , # G+1679  U+B5C0  uniB5C0,  HANGUL SYLLABLE-B5C0
        '46529' => '11236'            , # G+11236  U+B5C1  uniB5C1,  HANGUL SYLLABLE-B5C1
        '46530' => '11237'            , # G+11237  U+B5C2  uniB5C2,  HANGUL SYLLABLE-B5C2
        '46531' => '11238'            , # G+11238  U+B5C3  uniB5C3,  HANGUL SYLLABLE-B5C3
        '46532' => '1680'             , # G+1680  U+B5C4  uniB5C4,  HANGUL SYLLABLE-B5C4
        '46533' => '11239'            , # G+11239  U+B5C5  uniB5C5,  HANGUL SYLLABLE-B5C5
        '46534' => '11240'            , # G+11240  U+B5C6  uniB5C6,  HANGUL SYLLABLE-B5C6
        '46535' => '11241'            , # G+11241  U+B5C7  uniB5C7,  HANGUL SYLLABLE-B5C7
        '46536' => '11242'            , # G+11242  U+B5C8  uniB5C8,  HANGUL SYLLABLE-B5C8
        '46537' => '11243'            , # G+11243  U+B5C9  uniB5C9,  HANGUL SYLLABLE-B5C9
        '46538' => '11244'            , # G+11244  U+B5CA  uniB5CA,  HANGUL SYLLABLE-B5CA
        '46539' => '11245'            , # G+11245  U+B5CB  uniB5CB,  HANGUL SYLLABLE-B5CB
        '46540' => '1681'             , # G+1681  U+B5CC  uniB5CC,  HANGUL SYLLABLE-B5CC
        '46541' => '1682'             , # G+1682  U+B5CD  uniB5CD,  HANGUL SYLLABLE-B5CD
        '46542' => '11246'            , # G+11246  U+B5CE  uniB5CE,  HANGUL SYLLABLE-B5CE
        '46543' => '1683'             , # G+1683  U+B5CF  uniB5CF,  HANGUL SYLLABLE-B5CF
        '46544' => '1684'             , # G+1684  U+B5D0  uniB5D0,  HANGUL SYLLABLE-B5D0
        '46545' => '1685'             , # G+1685  U+B5D1  uniB5D1,  HANGUL SYLLABLE-B5D1
        '46546' => '11247'            , # G+11247  U+B5D2  uniB5D2,  HANGUL SYLLABLE-B5D2
        '46547' => '11248'            , # G+11248  U+B5D3  uniB5D3,  HANGUL SYLLABLE-B5D3
        '46548' => '11249'            , # G+11249  U+B5D4  uniB5D4,  HANGUL SYLLABLE-B5D4
        '46549' => '11250'            , # G+11250  U+B5D5  uniB5D5,  HANGUL SYLLABLE-B5D5
        '46550' => '11251'            , # G+11251  U+B5D6  uniB5D6,  HANGUL SYLLABLE-B5D6
        '46551' => '11252'            , # G+11252  U+B5D7  uniB5D7,  HANGUL SYLLABLE-B5D7
        '46552' => '1686'             , # G+1686  U+B5D8  uniB5D8,  HANGUL SYLLABLE-B5D8
        '46553' => '11253'            , # G+11253  U+B5D9  uniB5D9,  HANGUL SYLLABLE-B5D9
        '46554' => '11254'            , # G+11254  U+B5DA  uniB5DA,  HANGUL SYLLABLE-B5DA
        '46555' => '11255'            , # G+11255  U+B5DB  uniB5DB,  HANGUL SYLLABLE-B5DB
        '46556' => '11256'            , # G+11256  U+B5DC  uniB5DC,  HANGUL SYLLABLE-B5DC
        '46557' => '11257'            , # G+11257  U+B5DD  uniB5DD,  HANGUL SYLLABLE-B5DD
        '46558' => '11258'            , # G+11258  U+B5DE  uniB5DE,  HANGUL SYLLABLE-B5DE
        '46559' => '11259'            , # G+11259  U+B5DF  uniB5DF,  HANGUL SYLLABLE-B5DF
        '46560' => '11260'            , # G+11260  U+B5E0  uniB5E0,  HANGUL SYLLABLE-B5E0
        '46561' => '11261'            , # G+11261  U+B5E1  uniB5E1,  HANGUL SYLLABLE-B5E1
        '46562' => '11262'            , # G+11262  U+B5E2  uniB5E2,  HANGUL SYLLABLE-B5E2
        '46563' => '11263'            , # G+11263  U+B5E3  uniB5E3,  HANGUL SYLLABLE-B5E3
        '46564' => '11264'            , # G+11264  U+B5E4  uniB5E4,  HANGUL SYLLABLE-B5E4
        '46565' => '11265'            , # G+11265  U+B5E5  uniB5E5,  HANGUL SYLLABLE-B5E5
        '46566' => '11266'            , # G+11266  U+B5E6  uniB5E6,  HANGUL SYLLABLE-B5E6
        '46567' => '11267'            , # G+11267  U+B5E7  uniB5E7,  HANGUL SYLLABLE-B5E7
        '46568' => '11268'            , # G+11268  U+B5E8  uniB5E8,  HANGUL SYLLABLE-B5E8
        '46569' => '11269'            , # G+11269  U+B5E9  uniB5E9,  HANGUL SYLLABLE-B5E9
        '46570' => '11270'            , # G+11270  U+B5EA  uniB5EA,  HANGUL SYLLABLE-B5EA
        '46571' => '11271'            , # G+11271  U+B5EB  uniB5EB,  HANGUL SYLLABLE-B5EB
        '46572' => '1687'             , # G+1687  U+B5EC  uniB5EC,  HANGUL SYLLABLE-B5EC
        '46573' => '11272'            , # G+11272  U+B5ED  uniB5ED,  HANGUL SYLLABLE-B5ED
        '46574' => '11273'            , # G+11273  U+B5EE  uniB5EE,  HANGUL SYLLABLE-B5EE
        '46575' => '11274'            , # G+11274  U+B5EF  uniB5EF,  HANGUL SYLLABLE-B5EF
        '46576' => '11275'            , # G+11275  U+B5F0  uniB5F0,  HANGUL SYLLABLE-B5F0
        '46577' => '11276'            , # G+11276  U+B5F1  uniB5F1,  HANGUL SYLLABLE-B5F1
        '46578' => '11277'            , # G+11277  U+B5F2  uniB5F2,  HANGUL SYLLABLE-B5F2
        '46579' => '11278'            , # G+11278  U+B5F3  uniB5F3,  HANGUL SYLLABLE-B5F3
        '46580' => '11279'            , # G+11279  U+B5F4  uniB5F4,  HANGUL SYLLABLE-B5F4
        '46581' => '11280'            , # G+11280  U+B5F5  uniB5F5,  HANGUL SYLLABLE-B5F5
        '46582' => '11281'            , # G+11281  U+B5F6  uniB5F6,  HANGUL SYLLABLE-B5F6
        '46583' => '11282'            , # G+11282  U+B5F7  uniB5F7,  HANGUL SYLLABLE-B5F7
        '46584' => '11283'            , # G+11283  U+B5F8  uniB5F8,  HANGUL SYLLABLE-B5F8
        '46585' => '11284'            , # G+11284  U+B5F9  uniB5F9,  HANGUL SYLLABLE-B5F9
        '46586' => '11285'            , # G+11285  U+B5FA  uniB5FA,  HANGUL SYLLABLE-B5FA
        '46587' => '11286'            , # G+11286  U+B5FB  uniB5FB,  HANGUL SYLLABLE-B5FB
        '46588' => '11287'            , # G+11287  U+B5FC  uniB5FC,  HANGUL SYLLABLE-B5FC
        '46589' => '11288'            , # G+11288  U+B5FD  uniB5FD,  HANGUL SYLLABLE-B5FD
        '46590' => '11289'            , # G+11289  U+B5FE  uniB5FE,  HANGUL SYLLABLE-B5FE
        '46591' => '11290'            , # G+11290  U+B5FF  uniB5FF,  HANGUL SYLLABLE-B5FF
        '46592' => '11291'            , # G+11291  U+B600  uniB600,  HANGUL SYLLABLE-B600
        '46593' => '11292'            , # G+11292  U+B601  uniB601,  HANGUL SYLLABLE-B601
        '46594' => '11293'            , # G+11293  U+B602  uniB602,  HANGUL SYLLABLE-B602
        '46595' => '11294'            , # G+11294  U+B603  uniB603,  HANGUL SYLLABLE-B603
        '46596' => '11295'            , # G+11295  U+B604  uniB604,  HANGUL SYLLABLE-B604
        '46597' => '11296'            , # G+11296  U+B605  uniB605,  HANGUL SYLLABLE-B605
        '46598' => '11297'            , # G+11297  U+B606  uniB606,  HANGUL SYLLABLE-B606
        '46599' => '11298'            , # G+11298  U+B607  uniB607,  HANGUL SYLLABLE-B607
        '46600' => '11299'            , # G+11299  U+B608  uniB608,  HANGUL SYLLABLE-B608
        '46601' => '11300'            , # G+11300  U+B609  uniB609,  HANGUL SYLLABLE-B609
        '46602' => '11301'            , # G+11301  U+B60A  uniB60A,  HANGUL SYLLABLE-B60A
        '46603' => '11302'            , # G+11302  U+B60B  uniB60B,  HANGUL SYLLABLE-B60B
        '46604' => '11303'            , # G+11303  U+B60C  uniB60C,  HANGUL SYLLABLE-B60C
        '46605' => '11304'            , # G+11304  U+B60D  uniB60D,  HANGUL SYLLABLE-B60D
        '46606' => '11305'            , # G+11305  U+B60E  uniB60E,  HANGUL SYLLABLE-B60E
        '46607' => '11306'            , # G+11306  U+B60F  uniB60F,  HANGUL SYLLABLE-B60F
        '46608' => '1688'             , # G+1688  U+B610  uniB610,  HANGUL SYLLABLE-B610
        '46609' => '1689'             , # G+1689  U+B611  uniB611,  HANGUL SYLLABLE-B611
        '46610' => '11307'            , # G+11307  U+B612  uniB612,  HANGUL SYLLABLE-B612
        '46611' => '11308'            , # G+11308  U+B613  uniB613,  HANGUL SYLLABLE-B613
        '46612' => '1690'             , # G+1690  U+B614  uniB614,  HANGUL SYLLABLE-B614
        '46613' => '11309'            , # G+11309  U+B615  uniB615,  HANGUL SYLLABLE-B615
        '46614' => '11310'            , # G+11310  U+B616  uniB616,  HANGUL SYLLABLE-B616
        '46615' => '11311'            , # G+11311  U+B617  uniB617,  HANGUL SYLLABLE-B617
        '46616' => '1691'             , # G+1691  U+B618  uniB618,  HANGUL SYLLABLE-B618
        '46617' => '11312'            , # G+11312  U+B619  uniB619,  HANGUL SYLLABLE-B619
        '46618' => '11313'            , # G+11313  U+B61A  uniB61A,  HANGUL SYLLABLE-B61A
        '46619' => '11314'            , # G+11314  U+B61B  uniB61B,  HANGUL SYLLABLE-B61B
        '46620' => '11315'            , # G+11315  U+B61C  uniB61C,  HANGUL SYLLABLE-B61C
        '46621' => '11316'            , # G+11316  U+B61D  uniB61D,  HANGUL SYLLABLE-B61D
        '46622' => '11317'            , # G+11317  U+B61E  uniB61E,  HANGUL SYLLABLE-B61E
        '46623' => '11318'            , # G+11318  U+B61F  uniB61F,  HANGUL SYLLABLE-B61F
        '46624' => '11319'            , # G+11319  U+B620  uniB620,  HANGUL SYLLABLE-B620
        '46625' => '11320'            , # G+11320  U+B621  uniB621,  HANGUL SYLLABLE-B621
        '46626' => '11321'            , # G+11321  U+B622  uniB622,  HANGUL SYLLABLE-B622
        '46627' => '11322'            , # G+11322  U+B623  uniB623,  HANGUL SYLLABLE-B623
        '46628' => '11323'            , # G+11323  U+B624  uniB624,  HANGUL SYLLABLE-B624
        '46629' => '1692'             , # G+1692  U+B625  uniB625,  HANGUL SYLLABLE-B625
        '46630' => '11324'            , # G+11324  U+B626  uniB626,  HANGUL SYLLABLE-B626
        '46631' => '11325'            , # G+11325  U+B627  uniB627,  HANGUL SYLLABLE-B627
        '46632' => '11326'            , # G+11326  U+B628  uniB628,  HANGUL SYLLABLE-B628
        '46633' => '11327'            , # G+11327  U+B629  uniB629,  HANGUL SYLLABLE-B629
        '46634' => '11328'            , # G+11328  U+B62A  uniB62A,  HANGUL SYLLABLE-B62A
        '46635' => '11329'            , # G+11329  U+B62B  uniB62B,  HANGUL SYLLABLE-B62B
        '46636' => '1693'             , # G+1693  U+B62C  uniB62C,  HANGUL SYLLABLE-B62C
        '46637' => '11330'            , # G+11330  U+B62D  uniB62D,  HANGUL SYLLABLE-B62D
        '46638' => '11331'            , # G+11331  U+B62E  uniB62E,  HANGUL SYLLABLE-B62E
        '46639' => '11332'            , # G+11332  U+B62F  uniB62F,  HANGUL SYLLABLE-B62F
        '46640' => '11333'            , # G+11333  U+B630  uniB630,  HANGUL SYLLABLE-B630
        '46641' => '11334'            , # G+11334  U+B631  uniB631,  HANGUL SYLLABLE-B631
        '46642' => '11335'            , # G+11335  U+B632  uniB632,  HANGUL SYLLABLE-B632
        '46643' => '11336'            , # G+11336  U+B633  uniB633,  HANGUL SYLLABLE-B633
        '46644' => '1694'             , # G+1694  U+B634  uniB634,  HANGUL SYLLABLE-B634
        '46645' => '11337'            , # G+11337  U+B635  uniB635,  HANGUL SYLLABLE-B635
        '46646' => '11338'            , # G+11338  U+B636  uniB636,  HANGUL SYLLABLE-B636
        '46647' => '11339'            , # G+11339  U+B637  uniB637,  HANGUL SYLLABLE-B637
        '46648' => '11340'            , # G+11340  U+B638  uniB638,  HANGUL SYLLABLE-B638
        '46649' => '11341'            , # G+11341  U+B639  uniB639,  HANGUL SYLLABLE-B639
        '46650' => '11342'            , # G+11342  U+B63A  uniB63A,  HANGUL SYLLABLE-B63A
        '46651' => '11343'            , # G+11343  U+B63B  uniB63B,  HANGUL SYLLABLE-B63B
        '46652' => '11344'            , # G+11344  U+B63C  uniB63C,  HANGUL SYLLABLE-B63C
        '46653' => '11345'            , # G+11345  U+B63D  uniB63D,  HANGUL SYLLABLE-B63D
        '46654' => '11346'            , # G+11346  U+B63E  uniB63E,  HANGUL SYLLABLE-B63E
        '46655' => '11347'            , # G+11347  U+B63F  uniB63F,  HANGUL SYLLABLE-B63F
        '46656' => '11348'            , # G+11348  U+B640  uniB640,  HANGUL SYLLABLE-B640
        '46657' => '11349'            , # G+11349  U+B641  uniB641,  HANGUL SYLLABLE-B641
        '46658' => '11350'            , # G+11350  U+B642  uniB642,  HANGUL SYLLABLE-B642
        '46659' => '11351'            , # G+11351  U+B643  uniB643,  HANGUL SYLLABLE-B643
        '46660' => '11352'            , # G+11352  U+B644  uniB644,  HANGUL SYLLABLE-B644
        '46661' => '11353'            , # G+11353  U+B645  uniB645,  HANGUL SYLLABLE-B645
        '46662' => '11354'            , # G+11354  U+B646  uniB646,  HANGUL SYLLABLE-B646
        '46663' => '11355'            , # G+11355  U+B647  uniB647,  HANGUL SYLLABLE-B647
        '46664' => '1695'             , # G+1695  U+B648  uniB648,  HANGUL SYLLABLE-B648
        '46665' => '11356'            , # G+11356  U+B649  uniB649,  HANGUL SYLLABLE-B649
        '46666' => '11357'            , # G+11357  U+B64A  uniB64A,  HANGUL SYLLABLE-B64A
        '46667' => '11358'            , # G+11358  U+B64B  uniB64B,  HANGUL SYLLABLE-B64B
        '46668' => '11359'            , # G+11359  U+B64C  uniB64C,  HANGUL SYLLABLE-B64C
        '46669' => '11360'            , # G+11360  U+B64D  uniB64D,  HANGUL SYLLABLE-B64D
        '46670' => '11361'            , # G+11361  U+B64E  uniB64E,  HANGUL SYLLABLE-B64E
        '46671' => '11362'            , # G+11362  U+B64F  uniB64F,  HANGUL SYLLABLE-B64F
        '46672' => '11363'            , # G+11363  U+B650  uniB650,  HANGUL SYLLABLE-B650
        '46673' => '11364'            , # G+11364  U+B651  uniB651,  HANGUL SYLLABLE-B651
        '46674' => '11365'            , # G+11365  U+B652  uniB652,  HANGUL SYLLABLE-B652
        '46675' => '11366'            , # G+11366  U+B653  uniB653,  HANGUL SYLLABLE-B653
        '46676' => '11367'            , # G+11367  U+B654  uniB654,  HANGUL SYLLABLE-B654
        '46677' => '11368'            , # G+11368  U+B655  uniB655,  HANGUL SYLLABLE-B655
        '46678' => '11369'            , # G+11369  U+B656  uniB656,  HANGUL SYLLABLE-B656
        '46679' => '11370'            , # G+11370  U+B657  uniB657,  HANGUL SYLLABLE-B657
        '46680' => '11371'            , # G+11371  U+B658  uniB658,  HANGUL SYLLABLE-B658
        '46681' => '11372'            , # G+11372  U+B659  uniB659,  HANGUL SYLLABLE-B659
        '46682' => '11373'            , # G+11373  U+B65A  uniB65A,  HANGUL SYLLABLE-B65A
        '46683' => '11374'            , # G+11374  U+B65B  uniB65B,  HANGUL SYLLABLE-B65B
        '46684' => '11375'            , # G+11375  U+B65C  uniB65C,  HANGUL SYLLABLE-B65C
        '46685' => '11376'            , # G+11376  U+B65D  uniB65D,  HANGUL SYLLABLE-B65D
        '46686' => '11377'            , # G+11377  U+B65E  uniB65E,  HANGUL SYLLABLE-B65E
        '46687' => '11378'            , # G+11378  U+B65F  uniB65F,  HANGUL SYLLABLE-B65F
        '46688' => '11379'            , # G+11379  U+B660  uniB660,  HANGUL SYLLABLE-B660
        '46689' => '11380'            , # G+11380  U+B661  uniB661,  HANGUL SYLLABLE-B661
        '46690' => '11381'            , # G+11381  U+B662  uniB662,  HANGUL SYLLABLE-B662
        '46691' => '11382'            , # G+11382  U+B663  uniB663,  HANGUL SYLLABLE-B663
        '46692' => '1696'             , # G+1696  U+B664  uniB664,  HANGUL SYLLABLE-B664
        '46693' => '11383'            , # G+11383  U+B665  uniB665,  HANGUL SYLLABLE-B665
        '46694' => '11384'            , # G+11384  U+B666  uniB666,  HANGUL SYLLABLE-B666
        '46695' => '11385'            , # G+11385  U+B667  uniB667,  HANGUL SYLLABLE-B667
        '46696' => '1697'             , # G+1697  U+B668  uniB668,  HANGUL SYLLABLE-B668
        '46697' => '11386'            , # G+11386  U+B669  uniB669,  HANGUL SYLLABLE-B669
        '46698' => '11387'            , # G+11387  U+B66A  uniB66A,  HANGUL SYLLABLE-B66A
        '46699' => '11388'            , # G+11388  U+B66B  uniB66B,  HANGUL SYLLABLE-B66B
        '46700' => '11389'            , # G+11389  U+B66C  uniB66C,  HANGUL SYLLABLE-B66C
        '46701' => '11390'            , # G+11390  U+B66D  uniB66D,  HANGUL SYLLABLE-B66D
        '46702' => '11391'            , # G+11391  U+B66E  uniB66E,  HANGUL SYLLABLE-B66E
        '46703' => '11392'            , # G+11392  U+B66F  uniB66F,  HANGUL SYLLABLE-B66F
        '46704' => '11393'            , # G+11393  U+B670  uniB670,  HANGUL SYLLABLE-B670
        '46705' => '11394'            , # G+11394  U+B671  uniB671,  HANGUL SYLLABLE-B671
        '46706' => '11395'            , # G+11395  U+B672  uniB672,  HANGUL SYLLABLE-B672
        '46707' => '11396'            , # G+11396  U+B673  uniB673,  HANGUL SYLLABLE-B673
        '46708' => '11397'            , # G+11397  U+B674  uniB674,  HANGUL SYLLABLE-B674
        '46709' => '11398'            , # G+11398  U+B675  uniB675,  HANGUL SYLLABLE-B675
        '46710' => '11399'            , # G+11399  U+B676  uniB676,  HANGUL SYLLABLE-B676
        '46711' => '11400'            , # G+11400  U+B677  uniB677,  HANGUL SYLLABLE-B677
        '46712' => '11401'            , # G+11401  U+B678  uniB678,  HANGUL SYLLABLE-B678
        '46713' => '11402'            , # G+11402  U+B679  uniB679,  HANGUL SYLLABLE-B679
        '46714' => '11403'            , # G+11403  U+B67A  uniB67A,  HANGUL SYLLABLE-B67A
        '46715' => '11404'            , # G+11404  U+B67B  uniB67B,  HANGUL SYLLABLE-B67B
        '46716' => '11405'            , # G+11405  U+B67C  uniB67C,  HANGUL SYLLABLE-B67C
        '46717' => '11406'            , # G+11406  U+B67D  uniB67D,  HANGUL SYLLABLE-B67D
        '46718' => '11407'            , # G+11407  U+B67E  uniB67E,  HANGUL SYLLABLE-B67E
        '46719' => '11408'            , # G+11408  U+B67F  uniB67F,  HANGUL SYLLABLE-B67F
        '46720' => '11409'            , # G+11409  U+B680  uniB680,  HANGUL SYLLABLE-B680
        '46721' => '11410'            , # G+11410  U+B681  uniB681,  HANGUL SYLLABLE-B681
        '46722' => '11411'            , # G+11411  U+B682  uniB682,  HANGUL SYLLABLE-B682
        '46723' => '11412'            , # G+11412  U+B683  uniB683,  HANGUL SYLLABLE-B683
        '46724' => '11413'            , # G+11413  U+B684  uniB684,  HANGUL SYLLABLE-B684
        '46725' => '11414'            , # G+11414  U+B685  uniB685,  HANGUL SYLLABLE-B685
        '46726' => '11415'            , # G+11415  U+B686  uniB686,  HANGUL SYLLABLE-B686
        '46727' => '11416'            , # G+11416  U+B687  uniB687,  HANGUL SYLLABLE-B687
        '46728' => '11417'            , # G+11417  U+B688  uniB688,  HANGUL SYLLABLE-B688
        '46729' => '11418'            , # G+11418  U+B689  uniB689,  HANGUL SYLLABLE-B689
        '46730' => '11419'            , # G+11419  U+B68A  uniB68A,  HANGUL SYLLABLE-B68A
        '46731' => '11420'            , # G+11420  U+B68B  uniB68B,  HANGUL SYLLABLE-B68B
        '46732' => '11421'            , # G+11421  U+B68C  uniB68C,  HANGUL SYLLABLE-B68C
        '46733' => '11422'            , # G+11422  U+B68D  uniB68D,  HANGUL SYLLABLE-B68D
        '46734' => '11423'            , # G+11423  U+B68E  uniB68E,  HANGUL SYLLABLE-B68E
        '46735' => '11424'            , # G+11424  U+B68F  uniB68F,  HANGUL SYLLABLE-B68F
        '46736' => '11425'            , # G+11425  U+B690  uniB690,  HANGUL SYLLABLE-B690
        '46737' => '11426'            , # G+11426  U+B691  uniB691,  HANGUL SYLLABLE-B691
        '46738' => '11427'            , # G+11427  U+B692  uniB692,  HANGUL SYLLABLE-B692
        '46739' => '11428'            , # G+11428  U+B693  uniB693,  HANGUL SYLLABLE-B693
        '46740' => '11429'            , # G+11429  U+B694  uniB694,  HANGUL SYLLABLE-B694
        '46741' => '11430'            , # G+11430  U+B695  uniB695,  HANGUL SYLLABLE-B695
        '46742' => '11431'            , # G+11431  U+B696  uniB696,  HANGUL SYLLABLE-B696
        '46743' => '11432'            , # G+11432  U+B697  uniB697,  HANGUL SYLLABLE-B697
        '46744' => '11433'            , # G+11433  U+B698  uniB698,  HANGUL SYLLABLE-B698
        '46745' => '11434'            , # G+11434  U+B699  uniB699,  HANGUL SYLLABLE-B699
        '46746' => '11435'            , # G+11435  U+B69A  uniB69A,  HANGUL SYLLABLE-B69A
        '46747' => '11436'            , # G+11436  U+B69B  uniB69B,  HANGUL SYLLABLE-B69B
        '46748' => '1698'             , # G+1698  U+B69C  uniB69C,  HANGUL SYLLABLE-B69C
        '46749' => '1699'             , # G+1699  U+B69D  uniB69D,  HANGUL SYLLABLE-B69D
        '46750' => '11437'            , # G+11437  U+B69E  uniB69E,  HANGUL SYLLABLE-B69E
        '46751' => '11438'            , # G+11438  U+B69F  uniB69F,  HANGUL SYLLABLE-B69F
        '46752' => '1700'             , # G+1700  U+B6A0  uniB6A0,  HANGUL SYLLABLE-B6A0
        '46753' => '11439'            , # G+11439  U+B6A1  uniB6A1,  HANGUL SYLLABLE-B6A1
        '46754' => '11440'            , # G+11440  U+B6A2  uniB6A2,  HANGUL SYLLABLE-B6A2
        '46755' => '11441'            , # G+11441  U+B6A3  uniB6A3,  HANGUL SYLLABLE-B6A3
        '46756' => '1701'             , # G+1701  U+B6A4  uniB6A4,  HANGUL SYLLABLE-B6A4
        '46757' => '11442'            , # G+11442  U+B6A5  uniB6A5,  HANGUL SYLLABLE-B6A5
        '46758' => '11443'            , # G+11443  U+B6A6  uniB6A6,  HANGUL SYLLABLE-B6A6
        '46759' => '11444'            , # G+11444  U+B6A7  uniB6A7,  HANGUL SYLLABLE-B6A7
        '46760' => '11445'            , # G+11445  U+B6A8  uniB6A8,  HANGUL SYLLABLE-B6A8
        '46761' => '11446'            , # G+11446  U+B6A9  uniB6A9,  HANGUL SYLLABLE-B6A9
        '46762' => '11447'            , # G+11447  U+B6AA  uniB6AA,  HANGUL SYLLABLE-B6AA
        '46763' => '1702'             , # G+1702  U+B6AB  uniB6AB,  HANGUL SYLLABLE-B6AB
        '46764' => '1703'             , # G+1703  U+B6AC  uniB6AC,  HANGUL SYLLABLE-B6AC
        '46765' => '11448'            , # G+11448  U+B6AD  uniB6AD,  HANGUL SYLLABLE-B6AD
        '46766' => '11449'            , # G+11449  U+B6AE  uniB6AE,  HANGUL SYLLABLE-B6AE
        '46767' => '11450'            , # G+11450  U+B6AF  uniB6AF,  HANGUL SYLLABLE-B6AF
        '46768' => '11451'            , # G+11451  U+B6B0  uniB6B0,  HANGUL SYLLABLE-B6B0
        '46769' => '1704'             , # G+1704  U+B6B1  uniB6B1,  HANGUL SYLLABLE-B6B1
        '46770' => '11452'            , # G+11452  U+B6B2  uniB6B2,  HANGUL SYLLABLE-B6B2
        '46771' => '11453'            , # G+11453  U+B6B3  uniB6B3,  HANGUL SYLLABLE-B6B3
        '46772' => '11454'            , # G+11454  U+B6B4  uniB6B4,  HANGUL SYLLABLE-B6B4
        '46773' => '11455'            , # G+11455  U+B6B5  uniB6B5,  HANGUL SYLLABLE-B6B5
        '46774' => '11456'            , # G+11456  U+B6B6  uniB6B6,  HANGUL SYLLABLE-B6B6
        '46775' => '11457'            , # G+11457  U+B6B7  uniB6B7,  HANGUL SYLLABLE-B6B7
        '46776' => '11458'            , # G+11458  U+B6B8  uniB6B8,  HANGUL SYLLABLE-B6B8
        '46777' => '11459'            , # G+11459  U+B6B9  uniB6B9,  HANGUL SYLLABLE-B6B9
        '46778' => '11460'            , # G+11460  U+B6BA  uniB6BA,  HANGUL SYLLABLE-B6BA
        '46779' => '11461'            , # G+11461  U+B6BB  uniB6BB,  HANGUL SYLLABLE-B6BB
        '46780' => '11462'            , # G+11462  U+B6BC  uniB6BC,  HANGUL SYLLABLE-B6BC
        '46781' => '11463'            , # G+11463  U+B6BD  uniB6BD,  HANGUL SYLLABLE-B6BD
        '46782' => '11464'            , # G+11464  U+B6BE  uniB6BE,  HANGUL SYLLABLE-B6BE
        '46783' => '11465'            , # G+11465  U+B6BF  uniB6BF,  HANGUL SYLLABLE-B6BF
        '46784' => '11466'            , # G+11466  U+B6C0  uniB6C0,  HANGUL SYLLABLE-B6C0
        '46785' => '11467'            , # G+11467  U+B6C1  uniB6C1,  HANGUL SYLLABLE-B6C1
        '46786' => '11468'            , # G+11468  U+B6C2  uniB6C2,  HANGUL SYLLABLE-B6C2
        '46787' => '11469'            , # G+11469  U+B6C3  uniB6C3,  HANGUL SYLLABLE-B6C3
        '46788' => '11470'            , # G+11470  U+B6C4  uniB6C4,  HANGUL SYLLABLE-B6C4
        '46789' => '11471'            , # G+11471  U+B6C5  uniB6C5,  HANGUL SYLLABLE-B6C5
        '46790' => '11472'            , # G+11472  U+B6C6  uniB6C6,  HANGUL SYLLABLE-B6C6
        '46791' => '11473'            , # G+11473  U+B6C7  uniB6C7,  HANGUL SYLLABLE-B6C7
        '46792' => '11474'            , # G+11474  U+B6C8  uniB6C8,  HANGUL SYLLABLE-B6C8
        '46793' => '11475'            , # G+11475  U+B6C9  uniB6C9,  HANGUL SYLLABLE-B6C9
        '46794' => '11476'            , # G+11476  U+B6CA  uniB6CA,  HANGUL SYLLABLE-B6CA
        '46795' => '11477'            , # G+11477  U+B6CB  uniB6CB,  HANGUL SYLLABLE-B6CB
        '46796' => '11478'            , # G+11478  U+B6CC  uniB6CC,  HANGUL SYLLABLE-B6CC
        '46797' => '11479'            , # G+11479  U+B6CD  uniB6CD,  HANGUL SYLLABLE-B6CD
        '46798' => '11480'            , # G+11480  U+B6CE  uniB6CE,  HANGUL SYLLABLE-B6CE
        '46799' => '11481'            , # G+11481  U+B6CF  uniB6CF,  HANGUL SYLLABLE-B6CF
        '46800' => '11482'            , # G+11482  U+B6D0  uniB6D0,  HANGUL SYLLABLE-B6D0
        '46801' => '11483'            , # G+11483  U+B6D1  uniB6D1,  HANGUL SYLLABLE-B6D1
        '46802' => '11484'            , # G+11484  U+B6D2  uniB6D2,  HANGUL SYLLABLE-B6D2
        '46803' => '11485'            , # G+11485  U+B6D3  uniB6D3,  HANGUL SYLLABLE-B6D3
        '46804' => '1705'             , # G+1705  U+B6D4  uniB6D4,  HANGUL SYLLABLE-B6D4
        '46805' => '11486'            , # G+11486  U+B6D5  uniB6D5,  HANGUL SYLLABLE-B6D5
        '46806' => '11487'            , # G+11487  U+B6D6  uniB6D6,  HANGUL SYLLABLE-B6D6
        '46807' => '11488'            , # G+11488  U+B6D7  uniB6D7,  HANGUL SYLLABLE-B6D7
        '46808' => '11489'            , # G+11489  U+B6D8  uniB6D8,  HANGUL SYLLABLE-B6D8
        '46809' => '11490'            , # G+11490  U+B6D9  uniB6D9,  HANGUL SYLLABLE-B6D9
        '46810' => '11491'            , # G+11491  U+B6DA  uniB6DA,  HANGUL SYLLABLE-B6DA
        '46811' => '11492'            , # G+11492  U+B6DB  uniB6DB,  HANGUL SYLLABLE-B6DB
        '46812' => '11493'            , # G+11493  U+B6DC  uniB6DC,  HANGUL SYLLABLE-B6DC
        '46813' => '11494'            , # G+11494  U+B6DD  uniB6DD,  HANGUL SYLLABLE-B6DD
        '46814' => '11495'            , # G+11495  U+B6DE  uniB6DE,  HANGUL SYLLABLE-B6DE
        '46815' => '11496'            , # G+11496  U+B6DF  uniB6DF,  HANGUL SYLLABLE-B6DF
        '46816' => '11497'            , # G+11497  U+B6E0  uniB6E0,  HANGUL SYLLABLE-B6E0
        '46817' => '11498'            , # G+11498  U+B6E1  uniB6E1,  HANGUL SYLLABLE-B6E1
        '46818' => '11499'            , # G+11499  U+B6E2  uniB6E2,  HANGUL SYLLABLE-B6E2
        '46819' => '11500'            , # G+11500  U+B6E3  uniB6E3,  HANGUL SYLLABLE-B6E3
        '46820' => '11501'            , # G+11501  U+B6E4  uniB6E4,  HANGUL SYLLABLE-B6E4
        '46821' => '11502'            , # G+11502  U+B6E5  uniB6E5,  HANGUL SYLLABLE-B6E5
        '46822' => '11503'            , # G+11503  U+B6E6  uniB6E6,  HANGUL SYLLABLE-B6E6
        '46823' => '11504'            , # G+11504  U+B6E7  uniB6E7,  HANGUL SYLLABLE-B6E7
        '46824' => '11505'            , # G+11505  U+B6E8  uniB6E8,  HANGUL SYLLABLE-B6E8
        '46825' => '11506'            , # G+11506  U+B6E9  uniB6E9,  HANGUL SYLLABLE-B6E9
        '46826' => '11507'            , # G+11507  U+B6EA  uniB6EA,  HANGUL SYLLABLE-B6EA
        '46827' => '11508'            , # G+11508  U+B6EB  uniB6EB,  HANGUL SYLLABLE-B6EB
        '46828' => '11509'            , # G+11509  U+B6EC  uniB6EC,  HANGUL SYLLABLE-B6EC
        '46829' => '11510'            , # G+11510  U+B6ED  uniB6ED,  HANGUL SYLLABLE-B6ED
        '46830' => '11511'            , # G+11511  U+B6EE  uniB6EE,  HANGUL SYLLABLE-B6EE
        '46831' => '11512'            , # G+11512  U+B6EF  uniB6EF,  HANGUL SYLLABLE-B6EF
        '46832' => '1706'             , # G+1706  U+B6F0  uniB6F0,  HANGUL SYLLABLE-B6F0
        '46833' => '11513'            , # G+11513  U+B6F1  uniB6F1,  HANGUL SYLLABLE-B6F1
        '46834' => '11514'            , # G+11514  U+B6F2  uniB6F2,  HANGUL SYLLABLE-B6F2
        '46835' => '11515'            , # G+11515  U+B6F3  uniB6F3,  HANGUL SYLLABLE-B6F3
        '46836' => '1707'             , # G+1707  U+B6F4  uniB6F4,  HANGUL SYLLABLE-B6F4
        '46837' => '11516'            , # G+11516  U+B6F5  uniB6F5,  HANGUL SYLLABLE-B6F5
        '46838' => '11517'            , # G+11517  U+B6F6  uniB6F6,  HANGUL SYLLABLE-B6F6
        '46839' => '11518'            , # G+11518  U+B6F7  uniB6F7,  HANGUL SYLLABLE-B6F7
        '46840' => '1708'             , # G+1708  U+B6F8  uniB6F8,  HANGUL SYLLABLE-B6F8
        '46841' => '11519'            , # G+11519  U+B6F9  uniB6F9,  HANGUL SYLLABLE-B6F9
        '46842' => '11520'            , # G+11520  U+B6FA  uniB6FA,  HANGUL SYLLABLE-B6FA
        '46843' => '11521'            , # G+11521  U+B6FB  uniB6FB,  HANGUL SYLLABLE-B6FB
        '46844' => '11522'            , # G+11522  U+B6FC  uniB6FC,  HANGUL SYLLABLE-B6FC
        '46845' => '11523'            , # G+11523  U+B6FD  uniB6FD,  HANGUL SYLLABLE-B6FD
        '46846' => '11524'            , # G+11524  U+B6FE  uniB6FE,  HANGUL SYLLABLE-B6FE
        '46847' => '11525'            , # G+11525  U+B6FF  uniB6FF,  HANGUL SYLLABLE-B6FF
        '46848' => '1709'             , # G+1709  U+B700  uniB700,  HANGUL SYLLABLE-B700
        '46849' => '1710'             , # G+1710  U+B701  uniB701,  HANGUL SYLLABLE-B701
        '46850' => '11526'            , # G+11526  U+B702  uniB702,  HANGUL SYLLABLE-B702
        '46851' => '11527'            , # G+11527  U+B703  uniB703,  HANGUL SYLLABLE-B703
        '46852' => '11528'            , # G+11528  U+B704  uniB704,  HANGUL SYLLABLE-B704
        '46853' => '1711'             , # G+1711  U+B705  uniB705,  HANGUL SYLLABLE-B705
        '46854' => '11529'            , # G+11529  U+B706  uniB706,  HANGUL SYLLABLE-B706
        '46855' => '11530'            , # G+11530  U+B707  uniB707,  HANGUL SYLLABLE-B707
        '46856' => '11531'            , # G+11531  U+B708  uniB708,  HANGUL SYLLABLE-B708
        '46857' => '11532'            , # G+11532  U+B709  uniB709,  HANGUL SYLLABLE-B709
        '46858' => '11533'            , # G+11533  U+B70A  uniB70A,  HANGUL SYLLABLE-B70A
        '46859' => '11534'            , # G+11534  U+B70B  uniB70B,  HANGUL SYLLABLE-B70B
        '46860' => '11535'            , # G+11535  U+B70C  uniB70C,  HANGUL SYLLABLE-B70C
        '46861' => '11536'            , # G+11536  U+B70D  uniB70D,  HANGUL SYLLABLE-B70D
        '46862' => '11537'            , # G+11537  U+B70E  uniB70E,  HANGUL SYLLABLE-B70E
        '46863' => '11538'            , # G+11538  U+B70F  uniB70F,  HANGUL SYLLABLE-B70F
        '46864' => '11539'            , # G+11539  U+B710  uniB710,  HANGUL SYLLABLE-B710
        '46865' => '11540'            , # G+11540  U+B711  uniB711,  HANGUL SYLLABLE-B711
        '46866' => '11541'            , # G+11541  U+B712  uniB712,  HANGUL SYLLABLE-B712
        '46867' => '11542'            , # G+11542  U+B713  uniB713,  HANGUL SYLLABLE-B713
        '46868' => '11543'            , # G+11543  U+B714  uniB714,  HANGUL SYLLABLE-B714
        '46869' => '11544'            , # G+11544  U+B715  uniB715,  HANGUL SYLLABLE-B715
        '46870' => '11545'            , # G+11545  U+B716  uniB716,  HANGUL SYLLABLE-B716
        '46871' => '11546'            , # G+11546  U+B717  uniB717,  HANGUL SYLLABLE-B717
        '46872' => '11547'            , # G+11547  U+B718  uniB718,  HANGUL SYLLABLE-B718
        '46873' => '11548'            , # G+11548  U+B719  uniB719,  HANGUL SYLLABLE-B719
        '46874' => '11549'            , # G+11549  U+B71A  uniB71A,  HANGUL SYLLABLE-B71A
        '46875' => '11550'            , # G+11550  U+B71B  uniB71B,  HANGUL SYLLABLE-B71B
        '46876' => '11551'            , # G+11551  U+B71C  uniB71C,  HANGUL SYLLABLE-B71C
        '46877' => '11552'            , # G+11552  U+B71D  uniB71D,  HANGUL SYLLABLE-B71D
        '46878' => '11553'            , # G+11553  U+B71E  uniB71E,  HANGUL SYLLABLE-B71E
        '46879' => '11554'            , # G+11554  U+B71F  uniB71F,  HANGUL SYLLABLE-B71F
        '46880' => '11555'            , # G+11555  U+B720  uniB720,  HANGUL SYLLABLE-B720
        '46881' => '11556'            , # G+11556  U+B721  uniB721,  HANGUL SYLLABLE-B721
        '46882' => '11557'            , # G+11557  U+B722  uniB722,  HANGUL SYLLABLE-B722
        '46883' => '11558'            , # G+11558  U+B723  uniB723,  HANGUL SYLLABLE-B723
        '46884' => '11559'            , # G+11559  U+B724  uniB724,  HANGUL SYLLABLE-B724
        '46885' => '11560'            , # G+11560  U+B725  uniB725,  HANGUL SYLLABLE-B725
        '46886' => '11561'            , # G+11561  U+B726  uniB726,  HANGUL SYLLABLE-B726
        '46887' => '11562'            , # G+11562  U+B727  uniB727,  HANGUL SYLLABLE-B727
        '46888' => '1712'             , # G+1712  U+B728  uniB728,  HANGUL SYLLABLE-B728
        '46889' => '1713'             , # G+1713  U+B729  uniB729,  HANGUL SYLLABLE-B729
        '46890' => '11563'            , # G+11563  U+B72A  uniB72A,  HANGUL SYLLABLE-B72A
        '46891' => '11564'            , # G+11564  U+B72B  uniB72B,  HANGUL SYLLABLE-B72B
        '46892' => '1714'             , # G+1714  U+B72C  uniB72C,  HANGUL SYLLABLE-B72C
        '46893' => '11565'            , # G+11565  U+B72D  uniB72D,  HANGUL SYLLABLE-B72D
        '46894' => '11566'            , # G+11566  U+B72E  uniB72E,  HANGUL SYLLABLE-B72E
        '46895' => '1715'             , # G+1715  U+B72F  uniB72F,  HANGUL SYLLABLE-B72F
        '46896' => '1716'             , # G+1716  U+B730  uniB730,  HANGUL SYLLABLE-B730
        '46897' => '11567'            , # G+11567  U+B731  uniB731,  HANGUL SYLLABLE-B731
        '46898' => '11568'            , # G+11568  U+B732  uniB732,  HANGUL SYLLABLE-B732
        '46899' => '11569'            , # G+11569  U+B733  uniB733,  HANGUL SYLLABLE-B733
        '46900' => '11570'            , # G+11570  U+B734  uniB734,  HANGUL SYLLABLE-B734
        '46901' => '11571'            , # G+11571  U+B735  uniB735,  HANGUL SYLLABLE-B735
        '46902' => '11572'            , # G+11572  U+B736  uniB736,  HANGUL SYLLABLE-B736
        '46903' => '11573'            , # G+11573  U+B737  uniB737,  HANGUL SYLLABLE-B737
        '46904' => '1717'             , # G+1717  U+B738  uniB738,  HANGUL SYLLABLE-B738
        '46905' => '1718'             , # G+1718  U+B739  uniB739,  HANGUL SYLLABLE-B739
        '46906' => '11574'            , # G+11574  U+B73A  uniB73A,  HANGUL SYLLABLE-B73A
        '46907' => '1719'             , # G+1719  U+B73B  uniB73B,  HANGUL SYLLABLE-B73B
        '46908' => '11575'            , # G+11575  U+B73C  uniB73C,  HANGUL SYLLABLE-B73C
        '46909' => '11576'            , # G+11576  U+B73D  uniB73D,  HANGUL SYLLABLE-B73D
        '46910' => '11577'            , # G+11577  U+B73E  uniB73E,  HANGUL SYLLABLE-B73E
        '46911' => '11578'            , # G+11578  U+B73F  uniB73F,  HANGUL SYLLABLE-B73F
        '46912' => '11579'            , # G+11579  U+B740  uniB740,  HANGUL SYLLABLE-B740
        '46913' => '11580'            , # G+11580  U+B741  uniB741,  HANGUL SYLLABLE-B741
        '46914' => '11581'            , # G+11581  U+B742  uniB742,  HANGUL SYLLABLE-B742
        '46915' => '11582'            , # G+11582  U+B743  uniB743,  HANGUL SYLLABLE-B743
        '46916' => '1720'             , # G+1720  U+B744  uniB744,  HANGUL SYLLABLE-B744
        '46917' => '11583'            , # G+11583  U+B745  uniB745,  HANGUL SYLLABLE-B745
        '46918' => '11584'            , # G+11584  U+B746  uniB746,  HANGUL SYLLABLE-B746
        '46919' => '11585'            , # G+11585  U+B747  uniB747,  HANGUL SYLLABLE-B747
        '46920' => '1721'             , # G+1721  U+B748  uniB748,  HANGUL SYLLABLE-B748
        '46921' => '11586'            , # G+11586  U+B749  uniB749,  HANGUL SYLLABLE-B749
        '46922' => '11587'            , # G+11587  U+B74A  uniB74A,  HANGUL SYLLABLE-B74A
        '46923' => '11588'            , # G+11588  U+B74B  uniB74B,  HANGUL SYLLABLE-B74B
        '46924' => '1722'             , # G+1722  U+B74C  uniB74C,  HANGUL SYLLABLE-B74C
        '46925' => '11589'            , # G+11589  U+B74D  uniB74D,  HANGUL SYLLABLE-B74D
        '46926' => '11590'            , # G+11590  U+B74E  uniB74E,  HANGUL SYLLABLE-B74E
        '46927' => '11591'            , # G+11591  U+B74F  uniB74F,  HANGUL SYLLABLE-B74F
        '46928' => '11592'            , # G+11592  U+B750  uniB750,  HANGUL SYLLABLE-B750
        '46929' => '11593'            , # G+11593  U+B751  uniB751,  HANGUL SYLLABLE-B751
        '46930' => '11594'            , # G+11594  U+B752  uniB752,  HANGUL SYLLABLE-B752
        '46931' => '11595'            , # G+11595  U+B753  uniB753,  HANGUL SYLLABLE-B753
        '46932' => '1723'             , # G+1723  U+B754  uniB754,  HANGUL SYLLABLE-B754
        '46933' => '1724'             , # G+1724  U+B755  uniB755,  HANGUL SYLLABLE-B755
        '46934' => '11596'            , # G+11596  U+B756  uniB756,  HANGUL SYLLABLE-B756
        '46935' => '11597'            , # G+11597  U+B757  uniB757,  HANGUL SYLLABLE-B757
        '46936' => '11598'            , # G+11598  U+B758  uniB758,  HANGUL SYLLABLE-B758
        '46937' => '11599'            , # G+11599  U+B759  uniB759,  HANGUL SYLLABLE-B759
        '46938' => '11600'            , # G+11600  U+B75A  uniB75A,  HANGUL SYLLABLE-B75A
        '46939' => '11601'            , # G+11601  U+B75B  uniB75B,  HANGUL SYLLABLE-B75B
        '46940' => '11602'            , # G+11602  U+B75C  uniB75C,  HANGUL SYLLABLE-B75C
        '46941' => '11603'            , # G+11603  U+B75D  uniB75D,  HANGUL SYLLABLE-B75D
        '46942' => '11604'            , # G+11604  U+B75E  uniB75E,  HANGUL SYLLABLE-B75E
        '46943' => '11605'            , # G+11605  U+B75F  uniB75F,  HANGUL SYLLABLE-B75F
        '46944' => '1725'             , # G+1725  U+B760  uniB760,  HANGUL SYLLABLE-B760
        '46945' => '11606'            , # G+11606  U+B761  uniB761,  HANGUL SYLLABLE-B761
        '46946' => '11607'            , # G+11607  U+B762  uniB762,  HANGUL SYLLABLE-B762
        '46947' => '11608'            , # G+11608  U+B763  uniB763,  HANGUL SYLLABLE-B763
        '46948' => '1726'             , # G+1726  U+B764  uniB764,  HANGUL SYLLABLE-B764
        '46949' => '11609'            , # G+11609  U+B765  uniB765,  HANGUL SYLLABLE-B765
        '46950' => '11610'            , # G+11610  U+B766  uniB766,  HANGUL SYLLABLE-B766
        '46951' => '11611'            , # G+11611  U+B767  uniB767,  HANGUL SYLLABLE-B767
        '46952' => '1727'             , # G+1727  U+B768  uniB768,  HANGUL SYLLABLE-B768
        '46953' => '11612'            , # G+11612  U+B769  uniB769,  HANGUL SYLLABLE-B769
        '46954' => '11613'            , # G+11613  U+B76A  uniB76A,  HANGUL SYLLABLE-B76A
        '46955' => '11614'            , # G+11614  U+B76B  uniB76B,  HANGUL SYLLABLE-B76B
        '46956' => '11615'            , # G+11615  U+B76C  uniB76C,  HANGUL SYLLABLE-B76C
        '46957' => '11616'            , # G+11616  U+B76D  uniB76D,  HANGUL SYLLABLE-B76D
        '46958' => '11617'            , # G+11617  U+B76E  uniB76E,  HANGUL SYLLABLE-B76E
        '46959' => '11618'            , # G+11618  U+B76F  uniB76F,  HANGUL SYLLABLE-B76F
        '46960' => '1728'             , # G+1728  U+B770  uniB770,  HANGUL SYLLABLE-B770
        '46961' => '1729'             , # G+1729  U+B771  uniB771,  HANGUL SYLLABLE-B771
        '46962' => '11619'            , # G+11619  U+B772  uniB772,  HANGUL SYLLABLE-B772
        '46963' => '1730'             , # G+1730  U+B773  uniB773,  HANGUL SYLLABLE-B773
        '46964' => '11620'            , # G+11620  U+B774  uniB774,  HANGUL SYLLABLE-B774
        '46965' => '1731'             , # G+1731  U+B775  uniB775,  HANGUL SYLLABLE-B775
        '46966' => '11621'            , # G+11621  U+B776  uniB776,  HANGUL SYLLABLE-B776
        '46967' => '11622'            , # G+11622  U+B777  uniB777,  HANGUL SYLLABLE-B777
        '46968' => '11623'            , # G+11623  U+B778  uniB778,  HANGUL SYLLABLE-B778
        '46969' => '11624'            , # G+11624  U+B779  uniB779,  HANGUL SYLLABLE-B779
        '46970' => '11625'            , # G+11625  U+B77A  uniB77A,  HANGUL SYLLABLE-B77A
        '46971' => '11626'            , # G+11626  U+B77B  uniB77B,  HANGUL SYLLABLE-B77B
        '46972' => '1732'             , # G+1732  U+B77C  uniB77C,  HANGUL SYLLABLE-B77C
        '46973' => '1733'             , # G+1733  U+B77D  uniB77D,  HANGUL SYLLABLE-B77D
        '46974' => '11627'            , # G+11627  U+B77E  uniB77E,  HANGUL SYLLABLE-B77E
        '46975' => '11628'            , # G+11628  U+B77F  uniB77F,  HANGUL SYLLABLE-B77F
        '46976' => '1734'             , # G+1734  U+B780  uniB780,  HANGUL SYLLABLE-B780
        '46977' => '11629'            , # G+11629  U+B781  uniB781,  HANGUL SYLLABLE-B781
        '46978' => '11630'            , # G+11630  U+B782  uniB782,  HANGUL SYLLABLE-B782
        '46979' => '11631'            , # G+11631  U+B783  uniB783,  HANGUL SYLLABLE-B783
        '46980' => '1735'             , # G+1735  U+B784  uniB784,  HANGUL SYLLABLE-B784
        '46981' => '11632'            , # G+11632  U+B785  uniB785,  HANGUL SYLLABLE-B785
        '46982' => '11633'            , # G+11633  U+B786  uniB786,  HANGUL SYLLABLE-B786
        '46983' => '11634'            , # G+11634  U+B787  uniB787,  HANGUL SYLLABLE-B787
        '46984' => '11635'            , # G+11635  U+B788  uniB788,  HANGUL SYLLABLE-B788
        '46985' => '11636'            , # G+11636  U+B789  uniB789,  HANGUL SYLLABLE-B789
        '46986' => '11637'            , # G+11637  U+B78A  uniB78A,  HANGUL SYLLABLE-B78A
        '46987' => '11638'            , # G+11638  U+B78B  uniB78B,  HANGUL SYLLABLE-B78B
        '46988' => '1736'             , # G+1736  U+B78C  uniB78C,  HANGUL SYLLABLE-B78C
        '46989' => '1737'             , # G+1737  U+B78D  uniB78D,  HANGUL SYLLABLE-B78D
        '46990' => '11639'            , # G+11639  U+B78E  uniB78E,  HANGUL SYLLABLE-B78E
        '46991' => '1738'             , # G+1738  U+B78F  uniB78F,  HANGUL SYLLABLE-B78F
        '46992' => '1739'             , # G+1739  U+B790  uniB790,  HANGUL SYLLABLE-B790
        '46993' => '1740'             , # G+1740  U+B791  uniB791,  HANGUL SYLLABLE-B791
        '46994' => '1741'             , # G+1741  U+B792  uniB792,  HANGUL SYLLABLE-B792
        '46995' => '11640'            , # G+11640  U+B793  uniB793,  HANGUL SYLLABLE-B793
        '46996' => '11641'            , # G+11641  U+B794  uniB794,  HANGUL SYLLABLE-B794
        '46997' => '11642'            , # G+11642  U+B795  uniB795,  HANGUL SYLLABLE-B795
        '46998' => '1742'             , # G+1742  U+B796  uniB796,  HANGUL SYLLABLE-B796
        '46999' => '1743'             , # G+1743  U+B797  uniB797,  HANGUL SYLLABLE-B797
        '47000' => '1744'             , # G+1744  U+B798  uniB798,  HANGUL SYLLABLE-B798
        '47001' => '1745'             , # G+1745  U+B799  uniB799,  HANGUL SYLLABLE-B799
        '47002' => '11643'            , # G+11643  U+B79A  uniB79A,  HANGUL SYLLABLE-B79A
        '47003' => '11644'            , # G+11644  U+B79B  uniB79B,  HANGUL SYLLABLE-B79B
        '47004' => '1746'             , # G+1746  U+B79C  uniB79C,  HANGUL SYLLABLE-B79C
        '47005' => '11645'            , # G+11645  U+B79D  uniB79D,  HANGUL SYLLABLE-B79D
        '47006' => '11646'            , # G+11646  U+B79E  uniB79E,  HANGUL SYLLABLE-B79E
        '47007' => '11647'            , # G+11647  U+B79F  uniB79F,  HANGUL SYLLABLE-B79F
        '47008' => '1747'             , # G+1747  U+B7A0  uniB7A0,  HANGUL SYLLABLE-B7A0
        '47009' => '11648'            , # G+11648  U+B7A1  uniB7A1,  HANGUL SYLLABLE-B7A1
        '47010' => '11649'            , # G+11649  U+B7A2  uniB7A2,  HANGUL SYLLABLE-B7A2
        '47011' => '11650'            , # G+11650  U+B7A3  uniB7A3,  HANGUL SYLLABLE-B7A3
        '47012' => '11651'            , # G+11651  U+B7A4  uniB7A4,  HANGUL SYLLABLE-B7A4
        '47013' => '11652'            , # G+11652  U+B7A5  uniB7A5,  HANGUL SYLLABLE-B7A5
        '47014' => '11653'            , # G+11653  U+B7A6  uniB7A6,  HANGUL SYLLABLE-B7A6
        '47015' => '11654'            , # G+11654  U+B7A7  uniB7A7,  HANGUL SYLLABLE-B7A7
        '47016' => '1748'             , # G+1748  U+B7A8  uniB7A8,  HANGUL SYLLABLE-B7A8
        '47017' => '1749'             , # G+1749  U+B7A9  uniB7A9,  HANGUL SYLLABLE-B7A9
        '47018' => '11655'            , # G+11655  U+B7AA  uniB7AA,  HANGUL SYLLABLE-B7AA
        '47019' => '1750'             , # G+1750  U+B7AB  uniB7AB,  HANGUL SYLLABLE-B7AB
        '47020' => '1751'             , # G+1751  U+B7AC  uniB7AC,  HANGUL SYLLABLE-B7AC
        '47021' => '1752'             , # G+1752  U+B7AD  uniB7AD,  HANGUL SYLLABLE-B7AD
        '47022' => '11656'            , # G+11656  U+B7AE  uniB7AE,  HANGUL SYLLABLE-B7AE
        '47023' => '11657'            , # G+11657  U+B7AF  uniB7AF,  HANGUL SYLLABLE-B7AF
        '47024' => '11658'            , # G+11658  U+B7B0  uniB7B0,  HANGUL SYLLABLE-B7B0
        '47025' => '11659'            , # G+11659  U+B7B1  uniB7B1,  HANGUL SYLLABLE-B7B1
        '47026' => '11660'            , # G+11660  U+B7B2  uniB7B2,  HANGUL SYLLABLE-B7B2
        '47027' => '11661'            , # G+11661  U+B7B3  uniB7B3,  HANGUL SYLLABLE-B7B3
        '47028' => '1753'             , # G+1753  U+B7B4  uniB7B4,  HANGUL SYLLABLE-B7B4
        '47029' => '1754'             , # G+1754  U+B7B5  uniB7B5,  HANGUL SYLLABLE-B7B5
        '47030' => '11662'            , # G+11662  U+B7B6  uniB7B6,  HANGUL SYLLABLE-B7B6
        '47031' => '11663'            , # G+11663  U+B7B7  uniB7B7,  HANGUL SYLLABLE-B7B7
        '47032' => '1755'             , # G+1755  U+B7B8  uniB7B8,  HANGUL SYLLABLE-B7B8
        '47033' => '11664'            , # G+11664  U+B7B9  uniB7B9,  HANGUL SYLLABLE-B7B9
        '47034' => '11665'            , # G+11665  U+B7BA  uniB7BA,  HANGUL SYLLABLE-B7BA
        '47035' => '11666'            , # G+11666  U+B7BB  uniB7BB,  HANGUL SYLLABLE-B7BB
        '47036' => '11667'            , # G+11667  U+B7BC  uniB7BC,  HANGUL SYLLABLE-B7BC
        '47037' => '11668'            , # G+11668  U+B7BD  uniB7BD,  HANGUL SYLLABLE-B7BD
        '47038' => '11669'            , # G+11669  U+B7BE  uniB7BE,  HANGUL SYLLABLE-B7BE
        '47039' => '11670'            , # G+11670  U+B7BF  uniB7BF,  HANGUL SYLLABLE-B7BF
        '47040' => '11671'            , # G+11671  U+B7C0  uniB7C0,  HANGUL SYLLABLE-B7C0
        '47041' => '11672'            , # G+11672  U+B7C1  uniB7C1,  HANGUL SYLLABLE-B7C1
        '47042' => '11673'            , # G+11673  U+B7C2  uniB7C2,  HANGUL SYLLABLE-B7C2
        '47043' => '11674'            , # G+11674  U+B7C3  uniB7C3,  HANGUL SYLLABLE-B7C3
        '47044' => '11675'            , # G+11675  U+B7C4  uniB7C4,  HANGUL SYLLABLE-B7C4
        '47045' => '11676'            , # G+11676  U+B7C5  uniB7C5,  HANGUL SYLLABLE-B7C5
        '47046' => '11677'            , # G+11677  U+B7C6  uniB7C6,  HANGUL SYLLABLE-B7C6
        '47047' => '1756'             , # G+1756  U+B7C7  uniB7C7,  HANGUL SYLLABLE-B7C7
        '47048' => '11678'            , # G+11678  U+B7C8  uniB7C8,  HANGUL SYLLABLE-B7C8
        '47049' => '1757'             , # G+1757  U+B7C9  uniB7C9,  HANGUL SYLLABLE-B7C9
        '47050' => '11679'            , # G+11679  U+B7CA  uniB7CA,  HANGUL SYLLABLE-B7CA
        '47051' => '11680'            , # G+11680  U+B7CB  uniB7CB,  HANGUL SYLLABLE-B7CB
        '47052' => '11681'            , # G+11681  U+B7CC  uniB7CC,  HANGUL SYLLABLE-B7CC
        '47053' => '11682'            , # G+11682  U+B7CD  uniB7CD,  HANGUL SYLLABLE-B7CD
        '47054' => '11683'            , # G+11683  U+B7CE  uniB7CE,  HANGUL SYLLABLE-B7CE
        '47055' => '11684'            , # G+11684  U+B7CF  uniB7CF,  HANGUL SYLLABLE-B7CF
        '47056' => '11685'            , # G+11685  U+B7D0  uniB7D0,  HANGUL SYLLABLE-B7D0
        '47057' => '11686'            , # G+11686  U+B7D1  uniB7D1,  HANGUL SYLLABLE-B7D1
        '47058' => '11687'            , # G+11687  U+B7D2  uniB7D2,  HANGUL SYLLABLE-B7D2
        '47059' => '11688'            , # G+11688  U+B7D3  uniB7D3,  HANGUL SYLLABLE-B7D3
        '47060' => '11689'            , # G+11689  U+B7D4  uniB7D4,  HANGUL SYLLABLE-B7D4
        '47061' => '11690'            , # G+11690  U+B7D5  uniB7D5,  HANGUL SYLLABLE-B7D5
        '47062' => '11691'            , # G+11691  U+B7D6  uniB7D6,  HANGUL SYLLABLE-B7D6
        '47063' => '11692'            , # G+11692  U+B7D7  uniB7D7,  HANGUL SYLLABLE-B7D7
        '47064' => '11693'            , # G+11693  U+B7D8  uniB7D8,  HANGUL SYLLABLE-B7D8
        '47065' => '11694'            , # G+11694  U+B7D9  uniB7D9,  HANGUL SYLLABLE-B7D9
        '47066' => '11695'            , # G+11695  U+B7DA  uniB7DA,  HANGUL SYLLABLE-B7DA
        '47067' => '11696'            , # G+11696  U+B7DB  uniB7DB,  HANGUL SYLLABLE-B7DB
        '47068' => '11697'            , # G+11697  U+B7DC  uniB7DC,  HANGUL SYLLABLE-B7DC
        '47069' => '11698'            , # G+11698  U+B7DD  uniB7DD,  HANGUL SYLLABLE-B7DD
        '47070' => '11699'            , # G+11699  U+B7DE  uniB7DE,  HANGUL SYLLABLE-B7DE
        '47071' => '11700'            , # G+11700  U+B7DF  uniB7DF,  HANGUL SYLLABLE-B7DF
        '47072' => '11701'            , # G+11701  U+B7E0  uniB7E0,  HANGUL SYLLABLE-B7E0
        '47073' => '11702'            , # G+11702  U+B7E1  uniB7E1,  HANGUL SYLLABLE-B7E1
        '47074' => '11703'            , # G+11703  U+B7E2  uniB7E2,  HANGUL SYLLABLE-B7E2
        '47075' => '11704'            , # G+11704  U+B7E3  uniB7E3,  HANGUL SYLLABLE-B7E3
        '47076' => '11705'            , # G+11705  U+B7E4  uniB7E4,  HANGUL SYLLABLE-B7E4
        '47077' => '11706'            , # G+11706  U+B7E5  uniB7E5,  HANGUL SYLLABLE-B7E5
        '47078' => '11707'            , # G+11707  U+B7E6  uniB7E6,  HANGUL SYLLABLE-B7E6
        '47079' => '11708'            , # G+11708  U+B7E7  uniB7E7,  HANGUL SYLLABLE-B7E7
        '47080' => '11709'            , # G+11709  U+B7E8  uniB7E8,  HANGUL SYLLABLE-B7E8
        '47081' => '11710'            , # G+11710  U+B7E9  uniB7E9,  HANGUL SYLLABLE-B7E9
        '47082' => '11711'            , # G+11711  U+B7EA  uniB7EA,  HANGUL SYLLABLE-B7EA
        '47083' => '11712'            , # G+11712  U+B7EB  uniB7EB,  HANGUL SYLLABLE-B7EB
        '47084' => '1758'             , # G+1758  U+B7EC  uniB7EC,  HANGUL SYLLABLE-B7EC
        '47085' => '1759'             , # G+1759  U+B7ED  uniB7ED,  HANGUL SYLLABLE-B7ED
        '47086' => '11713'            , # G+11713  U+B7EE  uniB7EE,  HANGUL SYLLABLE-B7EE
        '47087' => '11714'            , # G+11714  U+B7EF  uniB7EF,  HANGUL SYLLABLE-B7EF
        '47088' => '1760'             , # G+1760  U+B7F0  uniB7F0,  HANGUL SYLLABLE-B7F0
        '47089' => '11715'            , # G+11715  U+B7F1  uniB7F1,  HANGUL SYLLABLE-B7F1
        '47090' => '11716'            , # G+11716  U+B7F2  uniB7F2,  HANGUL SYLLABLE-B7F2
        '47091' => '11717'            , # G+11717  U+B7F3  uniB7F3,  HANGUL SYLLABLE-B7F3
        '47092' => '1761'             , # G+1761  U+B7F4  uniB7F4,  HANGUL SYLLABLE-B7F4
        '47093' => '11718'            , # G+11718  U+B7F5  uniB7F5,  HANGUL SYLLABLE-B7F5
        '47094' => '11719'            , # G+11719  U+B7F6  uniB7F6,  HANGUL SYLLABLE-B7F6
        '47095' => '11720'            , # G+11720  U+B7F7  uniB7F7,  HANGUL SYLLABLE-B7F7
        '47096' => '11721'            , # G+11721  U+B7F8  uniB7F8,  HANGUL SYLLABLE-B7F8
        '47097' => '11722'            , # G+11722  U+B7F9  uniB7F9,  HANGUL SYLLABLE-B7F9
        '47098' => '11723'            , # G+11723  U+B7FA  uniB7FA,  HANGUL SYLLABLE-B7FA
        '47099' => '11724'            , # G+11724  U+B7FB  uniB7FB,  HANGUL SYLLABLE-B7FB
        '47100' => '1762'             , # G+1762  U+B7FC  uniB7FC,  HANGUL SYLLABLE-B7FC
        '47101' => '1763'             , # G+1763  U+B7FD  uniB7FD,  HANGUL SYLLABLE-B7FD
        '47102' => '11725'            , # G+11725  U+B7FE  uniB7FE,  HANGUL SYLLABLE-B7FE
        '47103' => '1764'             , # G+1764  U+B7FF  uniB7FF,  HANGUL SYLLABLE-B7FF
        '47104' => '1765'             , # G+1765  U+B800  uniB800,  HANGUL SYLLABLE-B800
        '47105' => '1766'             , # G+1766  U+B801  uniB801,  HANGUL SYLLABLE-B801
        '47106' => '11726'            , # G+11726  U+B802  uniB802,  HANGUL SYLLABLE-B802
        '47107' => '11727'            , # G+11727  U+B803  uniB803,  HANGUL SYLLABLE-B803
        '47108' => '11728'            , # G+11728  U+B804  uniB804,  HANGUL SYLLABLE-B804
        '47109' => '11729'            , # G+11729  U+B805  uniB805,  HANGUL SYLLABLE-B805
        '47110' => '11730'            , # G+11730  U+B806  uniB806,  HANGUL SYLLABLE-B806
        '47111' => '1767'             , # G+1767  U+B807  uniB807,  HANGUL SYLLABLE-B807
        '47112' => '1768'             , # G+1768  U+B808  uniB808,  HANGUL SYLLABLE-B808
        '47113' => '1769'             , # G+1769  U+B809  uniB809,  HANGUL SYLLABLE-B809
        '47114' => '11731'            , # G+11731  U+B80A  uniB80A,  HANGUL SYLLABLE-B80A
        '47115' => '11732'            , # G+11732  U+B80B  uniB80B,  HANGUL SYLLABLE-B80B
        '47116' => '1770'             , # G+1770  U+B80C  uniB80C,  HANGUL SYLLABLE-B80C
        '47117' => '11733'            , # G+11733  U+B80D  uniB80D,  HANGUL SYLLABLE-B80D
        '47118' => '11734'            , # G+11734  U+B80E  uniB80E,  HANGUL SYLLABLE-B80E
        '47119' => '11735'            , # G+11735  U+B80F  uniB80F,  HANGUL SYLLABLE-B80F
        '47120' => '1771'             , # G+1771  U+B810  uniB810,  HANGUL SYLLABLE-B810
        '47121' => '11736'            , # G+11736  U+B811  uniB811,  HANGUL SYLLABLE-B811
        '47122' => '11737'            , # G+11737  U+B812  uniB812,  HANGUL SYLLABLE-B812
        '47123' => '11738'            , # G+11738  U+B813  uniB813,  HANGUL SYLLABLE-B813
        '47124' => '11739'            , # G+11739  U+B814  uniB814,  HANGUL SYLLABLE-B814
        '47125' => '11740'            , # G+11740  U+B815  uniB815,  HANGUL SYLLABLE-B815
        '47126' => '11741'            , # G+11741  U+B816  uniB816,  HANGUL SYLLABLE-B816
        '47127' => '11742'            , # G+11742  U+B817  uniB817,  HANGUL SYLLABLE-B817
        '47128' => '1772'             , # G+1772  U+B818  uniB818,  HANGUL SYLLABLE-B818
        '47129' => '1773'             , # G+1773  U+B819  uniB819,  HANGUL SYLLABLE-B819
        '47130' => '11743'            , # G+11743  U+B81A  uniB81A,  HANGUL SYLLABLE-B81A
        '47131' => '1774'             , # G+1774  U+B81B  uniB81B,  HANGUL SYLLABLE-B81B
        '47132' => '11744'            , # G+11744  U+B81C  uniB81C,  HANGUL SYLLABLE-B81C
        '47133' => '1775'             , # G+1775  U+B81D  uniB81D,  HANGUL SYLLABLE-B81D
        '47134' => '11745'            , # G+11745  U+B81E  uniB81E,  HANGUL SYLLABLE-B81E
        '47135' => '11746'            , # G+11746  U+B81F  uniB81F,  HANGUL SYLLABLE-B81F
        '47136' => '11747'            , # G+11747  U+B820  uniB820,  HANGUL SYLLABLE-B820
        '47137' => '11748'            , # G+11748  U+B821  uniB821,  HANGUL SYLLABLE-B821
        '47138' => '11749'            , # G+11749  U+B822  uniB822,  HANGUL SYLLABLE-B822
        '47139' => '11750'            , # G+11750  U+B823  uniB823,  HANGUL SYLLABLE-B823
        '47140' => '1776'             , # G+1776  U+B824  uniB824,  HANGUL SYLLABLE-B824
        '47141' => '1777'             , # G+1777  U+B825  uniB825,  HANGUL SYLLABLE-B825
        '47142' => '11751'            , # G+11751  U+B826  uniB826,  HANGUL SYLLABLE-B826
        '47143' => '11752'            , # G+11752  U+B827  uniB827,  HANGUL SYLLABLE-B827
        '47144' => '1778'             , # G+1778  U+B828  uniB828,  HANGUL SYLLABLE-B828
        '47145' => '11753'            , # G+11753  U+B829  uniB829,  HANGUL SYLLABLE-B829
        '47146' => '11754'            , # G+11754  U+B82A  uniB82A,  HANGUL SYLLABLE-B82A
        '47147' => '11755'            , # G+11755  U+B82B  uniB82B,  HANGUL SYLLABLE-B82B
        '47148' => '1779'             , # G+1779  U+B82C  uniB82C,  HANGUL SYLLABLE-B82C
        '47149' => '11756'            , # G+11756  U+B82D  uniB82D,  HANGUL SYLLABLE-B82D
        '47150' => '11757'            , # G+11757  U+B82E  uniB82E,  HANGUL SYLLABLE-B82E
        '47151' => '11758'            , # G+11758  U+B82F  uniB82F,  HANGUL SYLLABLE-B82F
        '47152' => '11759'            , # G+11759  U+B830  uniB830,  HANGUL SYLLABLE-B830
        '47153' => '11760'            , # G+11760  U+B831  uniB831,  HANGUL SYLLABLE-B831
        '47154' => '11761'            , # G+11761  U+B832  uniB832,  HANGUL SYLLABLE-B832
        '47155' => '11762'            , # G+11762  U+B833  uniB833,  HANGUL SYLLABLE-B833
        '47156' => '1780'             , # G+1780  U+B834  uniB834,  HANGUL SYLLABLE-B834
        '47157' => '1781'             , # G+1781  U+B835  uniB835,  HANGUL SYLLABLE-B835
        '47158' => '11763'            , # G+11763  U+B836  uniB836,  HANGUL SYLLABLE-B836
        '47159' => '1782'             , # G+1782  U+B837  uniB837,  HANGUL SYLLABLE-B837
        '47160' => '1783'             , # G+1783  U+B838  uniB838,  HANGUL SYLLABLE-B838
        '47161' => '1784'             , # G+1784  U+B839  uniB839,  HANGUL SYLLABLE-B839
        '47162' => '11764'            , # G+11764  U+B83A  uniB83A,  HANGUL SYLLABLE-B83A
        '47163' => '11765'            , # G+11765  U+B83B  uniB83B,  HANGUL SYLLABLE-B83B
        '47164' => '11766'            , # G+11766  U+B83C  uniB83C,  HANGUL SYLLABLE-B83C
        '47165' => '11767'            , # G+11767  U+B83D  uniB83D,  HANGUL SYLLABLE-B83D
        '47166' => '11768'            , # G+11768  U+B83E  uniB83E,  HANGUL SYLLABLE-B83E
        '47167' => '11769'            , # G+11769  U+B83F  uniB83F,  HANGUL SYLLABLE-B83F
        '47168' => '1785'             , # G+1785  U+B840  uniB840,  HANGUL SYLLABLE-B840
        '47169' => '11770'            , # G+11770  U+B841  uniB841,  HANGUL SYLLABLE-B841
        '47170' => '11771'            , # G+11771  U+B842  uniB842,  HANGUL SYLLABLE-B842
        '47171' => '11772'            , # G+11772  U+B843  uniB843,  HANGUL SYLLABLE-B843
        '47172' => '1786'             , # G+1786  U+B844  uniB844,  HANGUL SYLLABLE-B844
        '47173' => '11773'            , # G+11773  U+B845  uniB845,  HANGUL SYLLABLE-B845
        '47174' => '11774'            , # G+11774  U+B846  uniB846,  HANGUL SYLLABLE-B846
        '47175' => '11775'            , # G+11775  U+B847  uniB847,  HANGUL SYLLABLE-B847
        '47176' => '11776'            , # G+11776  U+B848  uniB848,  HANGUL SYLLABLE-B848
        '47177' => '11777'            , # G+11777  U+B849  uniB849,  HANGUL SYLLABLE-B849
        '47178' => '11778'            , # G+11778  U+B84A  uniB84A,  HANGUL SYLLABLE-B84A
        '47179' => '11779'            , # G+11779  U+B84B  uniB84B,  HANGUL SYLLABLE-B84B
        '47180' => '11780'            , # G+11780  U+B84C  uniB84C,  HANGUL SYLLABLE-B84C
        '47181' => '11781'            , # G+11781  U+B84D  uniB84D,  HANGUL SYLLABLE-B84D
        '47182' => '11782'            , # G+11782  U+B84E  uniB84E,  HANGUL SYLLABLE-B84E
        '47183' => '11783'            , # G+11783  U+B84F  uniB84F,  HANGUL SYLLABLE-B84F
        '47184' => '11784'            , # G+11784  U+B850  uniB850,  HANGUL SYLLABLE-B850
        '47185' => '1787'             , # G+1787  U+B851  uniB851,  HANGUL SYLLABLE-B851
        '47186' => '11785'            , # G+11785  U+B852  uniB852,  HANGUL SYLLABLE-B852
        '47187' => '1788'             , # G+1788  U+B853  uniB853,  HANGUL SYLLABLE-B853
        '47188' => '11786'            , # G+11786  U+B854  uniB854,  HANGUL SYLLABLE-B854
        '47189' => '11787'            , # G+11787  U+B855  uniB855,  HANGUL SYLLABLE-B855
        '47190' => '11788'            , # G+11788  U+B856  uniB856,  HANGUL SYLLABLE-B856
        '47191' => '11789'            , # G+11789  U+B857  uniB857,  HANGUL SYLLABLE-B857
        '47192' => '11790'            , # G+11790  U+B858  uniB858,  HANGUL SYLLABLE-B858
        '47193' => '11791'            , # G+11791  U+B859  uniB859,  HANGUL SYLLABLE-B859
        '47194' => '11792'            , # G+11792  U+B85A  uniB85A,  HANGUL SYLLABLE-B85A
        '47195' => '11793'            , # G+11793  U+B85B  uniB85B,  HANGUL SYLLABLE-B85B
        '47196' => '1789'             , # G+1789  U+B85C  uniB85C,  HANGUL SYLLABLE-B85C
        '47197' => '1790'             , # G+1790  U+B85D  uniB85D,  HANGUL SYLLABLE-B85D
        '47198' => '11794'            , # G+11794  U+B85E  uniB85E,  HANGUL SYLLABLE-B85E
        '47199' => '11795'            , # G+11795  U+B85F  uniB85F,  HANGUL SYLLABLE-B85F
        '47200' => '1791'             , # G+1791  U+B860  uniB860,  HANGUL SYLLABLE-B860
        '47201' => '11796'            , # G+11796  U+B861  uniB861,  HANGUL SYLLABLE-B861
        '47202' => '11797'            , # G+11797  U+B862  uniB862,  HANGUL SYLLABLE-B862
        '47203' => '11798'            , # G+11798  U+B863  uniB863,  HANGUL SYLLABLE-B863
        '47204' => '1792'             , # G+1792  U+B864  uniB864,  HANGUL SYLLABLE-B864
        '47205' => '11799'            , # G+11799  U+B865  uniB865,  HANGUL SYLLABLE-B865
        '47206' => '11800'            , # G+11800  U+B866  uniB866,  HANGUL SYLLABLE-B866
        '47207' => '11801'            , # G+11801  U+B867  uniB867,  HANGUL SYLLABLE-B867
        '47208' => '11802'            , # G+11802  U+B868  uniB868,  HANGUL SYLLABLE-B868
        '47209' => '11803'            , # G+11803  U+B869  uniB869,  HANGUL SYLLABLE-B869
        '47210' => '11804'            , # G+11804  U+B86A  uniB86A,  HANGUL SYLLABLE-B86A
        '47211' => '11805'            , # G+11805  U+B86B  uniB86B,  HANGUL SYLLABLE-B86B
        '47212' => '1793'             , # G+1793  U+B86C  uniB86C,  HANGUL SYLLABLE-B86C
        '47213' => '1794'             , # G+1794  U+B86D  uniB86D,  HANGUL SYLLABLE-B86D
        '47214' => '11806'            , # G+11806  U+B86E  uniB86E,  HANGUL SYLLABLE-B86E
        '47215' => '1795'             , # G+1795  U+B86F  uniB86F,  HANGUL SYLLABLE-B86F
        '47216' => '11807'            , # G+11807  U+B870  uniB870,  HANGUL SYLLABLE-B870
        '47217' => '1796'             , # G+1796  U+B871  uniB871,  HANGUL SYLLABLE-B871
        '47218' => '11808'            , # G+11808  U+B872  uniB872,  HANGUL SYLLABLE-B872
        '47219' => '11809'            , # G+11809  U+B873  uniB873,  HANGUL SYLLABLE-B873
        '47220' => '11810'            , # G+11810  U+B874  uniB874,  HANGUL SYLLABLE-B874
        '47221' => '11811'            , # G+11811  U+B875  uniB875,  HANGUL SYLLABLE-B875
        '47222' => '11812'            , # G+11812  U+B876  uniB876,  HANGUL SYLLABLE-B876
        '47223' => '11813'            , # G+11813  U+B877  uniB877,  HANGUL SYLLABLE-B877
        '47224' => '1797'             , # G+1797  U+B878  uniB878,  HANGUL SYLLABLE-B878
        '47225' => '11814'            , # G+11814  U+B879  uniB879,  HANGUL SYLLABLE-B879
        '47226' => '11815'            , # G+11815  U+B87A  uniB87A,  HANGUL SYLLABLE-B87A
        '47227' => '11816'            , # G+11816  U+B87B  uniB87B,  HANGUL SYLLABLE-B87B
        '47228' => '1798'             , # G+1798  U+B87C  uniB87C,  HANGUL SYLLABLE-B87C
        '47229' => '11817'            , # G+11817  U+B87D  uniB87D,  HANGUL SYLLABLE-B87D
        '47230' => '11818'            , # G+11818  U+B87E  uniB87E,  HANGUL SYLLABLE-B87E
        '47231' => '11819'            , # G+11819  U+B87F  uniB87F,  HANGUL SYLLABLE-B87F
        '47232' => '11820'            , # G+11820  U+B880  uniB880,  HANGUL SYLLABLE-B880
        '47233' => '11821'            , # G+11821  U+B881  uniB881,  HANGUL SYLLABLE-B881
        '47234' => '11822'            , # G+11822  U+B882  uniB882,  HANGUL SYLLABLE-B882
        '47235' => '11823'            , # G+11823  U+B883  uniB883,  HANGUL SYLLABLE-B883
        '47236' => '11824'            , # G+11824  U+B884  uniB884,  HANGUL SYLLABLE-B884
        '47237' => '11825'            , # G+11825  U+B885  uniB885,  HANGUL SYLLABLE-B885
        '47238' => '11826'            , # G+11826  U+B886  uniB886,  HANGUL SYLLABLE-B886
        '47239' => '11827'            , # G+11827  U+B887  uniB887,  HANGUL SYLLABLE-B887
        '47240' => '11828'            , # G+11828  U+B888  uniB888,  HANGUL SYLLABLE-B888
        '47241' => '11829'            , # G+11829  U+B889  uniB889,  HANGUL SYLLABLE-B889
        '47242' => '11830'            , # G+11830  U+B88A  uniB88A,  HANGUL SYLLABLE-B88A
        '47243' => '11831'            , # G+11831  U+B88B  uniB88B,  HANGUL SYLLABLE-B88B
        '47244' => '11832'            , # G+11832  U+B88C  uniB88C,  HANGUL SYLLABLE-B88C
        '47245' => '1799'             , # G+1799  U+B88D  uniB88D,  HANGUL SYLLABLE-B88D
        '47246' => '11833'            , # G+11833  U+B88E  uniB88E,  HANGUL SYLLABLE-B88E
        '47247' => '11834'            , # G+11834  U+B88F  uniB88F,  HANGUL SYLLABLE-B88F
        '47248' => '11835'            , # G+11835  U+B890  uniB890,  HANGUL SYLLABLE-B890
        '47249' => '11836'            , # G+11836  U+B891  uniB891,  HANGUL SYLLABLE-B891
        '47250' => '11837'            , # G+11837  U+B892  uniB892,  HANGUL SYLLABLE-B892
        '47251' => '11838'            , # G+11838  U+B893  uniB893,  HANGUL SYLLABLE-B893
        '47252' => '11839'            , # G+11839  U+B894  uniB894,  HANGUL SYLLABLE-B894
        '47253' => '11840'            , # G+11840  U+B895  uniB895,  HANGUL SYLLABLE-B895
        '47254' => '11841'            , # G+11841  U+B896  uniB896,  HANGUL SYLLABLE-B896
        '47255' => '11842'            , # G+11842  U+B897  uniB897,  HANGUL SYLLABLE-B897
        '47256' => '11843'            , # G+11843  U+B898  uniB898,  HANGUL SYLLABLE-B898
        '47257' => '11844'            , # G+11844  U+B899  uniB899,  HANGUL SYLLABLE-B899
        '47258' => '11845'            , # G+11845  U+B89A  uniB89A,  HANGUL SYLLABLE-B89A
        '47259' => '11846'            , # G+11846  U+B89B  uniB89B,  HANGUL SYLLABLE-B89B
        '47260' => '11847'            , # G+11847  U+B89C  uniB89C,  HANGUL SYLLABLE-B89C
        '47261' => '11848'            , # G+11848  U+B89D  uniB89D,  HANGUL SYLLABLE-B89D
        '47262' => '11849'            , # G+11849  U+B89E  uniB89E,  HANGUL SYLLABLE-B89E
        '47263' => '11850'            , # G+11850  U+B89F  uniB89F,  HANGUL SYLLABLE-B89F
        '47264' => '11851'            , # G+11851  U+B8A0  uniB8A0,  HANGUL SYLLABLE-B8A0
        '47265' => '11852'            , # G+11852  U+B8A1  uniB8A1,  HANGUL SYLLABLE-B8A1
        '47266' => '11853'            , # G+11853  U+B8A2  uniB8A2,  HANGUL SYLLABLE-B8A2
        '47267' => '11854'            , # G+11854  U+B8A3  uniB8A3,  HANGUL SYLLABLE-B8A3
        '47268' => '11855'            , # G+11855  U+B8A4  uniB8A4,  HANGUL SYLLABLE-B8A4
        '47269' => '11856'            , # G+11856  U+B8A5  uniB8A5,  HANGUL SYLLABLE-B8A5
        '47270' => '11857'            , # G+11857  U+B8A6  uniB8A6,  HANGUL SYLLABLE-B8A6
        '47271' => '11858'            , # G+11858  U+B8A7  uniB8A7,  HANGUL SYLLABLE-B8A7
        '47272' => '1800'             , # G+1800  U+B8A8  uniB8A8,  HANGUL SYLLABLE-B8A8
        '47273' => '11859'            , # G+11859  U+B8A9  uniB8A9,  HANGUL SYLLABLE-B8A9
        '47274' => '11860'            , # G+11860  U+B8AA  uniB8AA,  HANGUL SYLLABLE-B8AA
        '47275' => '11861'            , # G+11861  U+B8AB  uniB8AB,  HANGUL SYLLABLE-B8AB
        '47276' => '11862'            , # G+11862  U+B8AC  uniB8AC,  HANGUL SYLLABLE-B8AC
        '47277' => '11863'            , # G+11863  U+B8AD  uniB8AD,  HANGUL SYLLABLE-B8AD
        '47278' => '11864'            , # G+11864  U+B8AE  uniB8AE,  HANGUL SYLLABLE-B8AE
        '47279' => '11865'            , # G+11865  U+B8AF  uniB8AF,  HANGUL SYLLABLE-B8AF
        '47280' => '1801'             , # G+1801  U+B8B0  uniB8B0,  HANGUL SYLLABLE-B8B0
        '47281' => '11866'            , # G+11866  U+B8B1  uniB8B1,  HANGUL SYLLABLE-B8B1
        '47282' => '11867'            , # G+11867  U+B8B2  uniB8B2,  HANGUL SYLLABLE-B8B2
        '47283' => '11868'            , # G+11868  U+B8B3  uniB8B3,  HANGUL SYLLABLE-B8B3
        '47284' => '1802'             , # G+1802  U+B8B4  uniB8B4,  HANGUL SYLLABLE-B8B4
        '47285' => '11869'            , # G+11869  U+B8B5  uniB8B5,  HANGUL SYLLABLE-B8B5
        '47286' => '11870'            , # G+11870  U+B8B6  uniB8B6,  HANGUL SYLLABLE-B8B6
        '47287' => '11871'            , # G+11871  U+B8B7  uniB8B7,  HANGUL SYLLABLE-B8B7
        '47288' => '1803'             , # G+1803  U+B8B8  uniB8B8,  HANGUL SYLLABLE-B8B8
        '47289' => '11872'            , # G+11872  U+B8B9  uniB8B9,  HANGUL SYLLABLE-B8B9
        '47290' => '11873'            , # G+11873  U+B8BA  uniB8BA,  HANGUL SYLLABLE-B8BA
        '47291' => '11874'            , # G+11874  U+B8BB  uniB8BB,  HANGUL SYLLABLE-B8BB
        '47292' => '11875'            , # G+11875  U+B8BC  uniB8BC,  HANGUL SYLLABLE-B8BC
        '47293' => '11876'            , # G+11876  U+B8BD  uniB8BD,  HANGUL SYLLABLE-B8BD
        '47294' => '11877'            , # G+11877  U+B8BE  uniB8BE,  HANGUL SYLLABLE-B8BE
        '47295' => '11878'            , # G+11878  U+B8BF  uniB8BF,  HANGUL SYLLABLE-B8BF
        '47296' => '1804'             , # G+1804  U+B8C0  uniB8C0,  HANGUL SYLLABLE-B8C0
        '47297' => '1805'             , # G+1805  U+B8C1  uniB8C1,  HANGUL SYLLABLE-B8C1
        '47298' => '11879'            , # G+11879  U+B8C2  uniB8C2,  HANGUL SYLLABLE-B8C2
        '47299' => '1806'             , # G+1806  U+B8C3  uniB8C3,  HANGUL SYLLABLE-B8C3
        '47300' => '11880'            , # G+11880  U+B8C4  uniB8C4,  HANGUL SYLLABLE-B8C4
        '47301' => '1807'             , # G+1807  U+B8C5  uniB8C5,  HANGUL SYLLABLE-B8C5
        '47302' => '11881'            , # G+11881  U+B8C6  uniB8C6,  HANGUL SYLLABLE-B8C6
        '47303' => '11882'            , # G+11882  U+B8C7  uniB8C7,  HANGUL SYLLABLE-B8C7
        '47304' => '11883'            , # G+11883  U+B8C8  uniB8C8,  HANGUL SYLLABLE-B8C8
        '47305' => '11884'            , # G+11884  U+B8C9  uniB8C9,  HANGUL SYLLABLE-B8C9
        '47306' => '11885'            , # G+11885  U+B8CA  uniB8CA,  HANGUL SYLLABLE-B8CA
        '47307' => '11886'            , # G+11886  U+B8CB  uniB8CB,  HANGUL SYLLABLE-B8CB
        '47308' => '1808'             , # G+1808  U+B8CC  uniB8CC,  HANGUL SYLLABLE-B8CC
        '47309' => '11887'            , # G+11887  U+B8CD  uniB8CD,  HANGUL SYLLABLE-B8CD
        '47310' => '11888'            , # G+11888  U+B8CE  uniB8CE,  HANGUL SYLLABLE-B8CE
        '47311' => '11889'            , # G+11889  U+B8CF  uniB8CF,  HANGUL SYLLABLE-B8CF
        '47312' => '1809'             , # G+1809  U+B8D0  uniB8D0,  HANGUL SYLLABLE-B8D0
        '47313' => '11890'            , # G+11890  U+B8D1  uniB8D1,  HANGUL SYLLABLE-B8D1
        '47314' => '11891'            , # G+11891  U+B8D2  uniB8D2,  HANGUL SYLLABLE-B8D2
        '47315' => '11892'            , # G+11892  U+B8D3  uniB8D3,  HANGUL SYLLABLE-B8D3
        '47316' => '1810'             , # G+1810  U+B8D4  uniB8D4,  HANGUL SYLLABLE-B8D4
        '47317' => '11893'            , # G+11893  U+B8D5  uniB8D5,  HANGUL SYLLABLE-B8D5
        '47318' => '11894'            , # G+11894  U+B8D6  uniB8D6,  HANGUL SYLLABLE-B8D6
        '47319' => '11895'            , # G+11895  U+B8D7  uniB8D7,  HANGUL SYLLABLE-B8D7
        '47320' => '11896'            , # G+11896  U+B8D8  uniB8D8,  HANGUL SYLLABLE-B8D8
        '47321' => '11897'            , # G+11897  U+B8D9  uniB8D9,  HANGUL SYLLABLE-B8D9
        '47322' => '11898'            , # G+11898  U+B8DA  uniB8DA,  HANGUL SYLLABLE-B8DA
        '47323' => '11899'            , # G+11899  U+B8DB  uniB8DB,  HANGUL SYLLABLE-B8DB
        '47324' => '11900'            , # G+11900  U+B8DC  uniB8DC,  HANGUL SYLLABLE-B8DC
        '47325' => '1811'             , # G+1811  U+B8DD  uniB8DD,  HANGUL SYLLABLE-B8DD
        '47326' => '11901'            , # G+11901  U+B8DE  uniB8DE,  HANGUL SYLLABLE-B8DE
        '47327' => '1812'             , # G+1812  U+B8DF  uniB8DF,  HANGUL SYLLABLE-B8DF
        '47328' => '11902'            , # G+11902  U+B8E0  uniB8E0,  HANGUL SYLLABLE-B8E0
        '47329' => '1813'             , # G+1813  U+B8E1  uniB8E1,  HANGUL SYLLABLE-B8E1
        '47330' => '11903'            , # G+11903  U+B8E2  uniB8E2,  HANGUL SYLLABLE-B8E2
        '47331' => '11904'            , # G+11904  U+B8E3  uniB8E3,  HANGUL SYLLABLE-B8E3
        '47332' => '11905'            , # G+11905  U+B8E4  uniB8E4,  HANGUL SYLLABLE-B8E4
        '47333' => '11906'            , # G+11906  U+B8E5  uniB8E5,  HANGUL SYLLABLE-B8E5
        '47334' => '11907'            , # G+11907  U+B8E6  uniB8E6,  HANGUL SYLLABLE-B8E6
        '47335' => '11908'            , # G+11908  U+B8E7  uniB8E7,  HANGUL SYLLABLE-B8E7
        '47336' => '1814'             , # G+1814  U+B8E8  uniB8E8,  HANGUL SYLLABLE-B8E8
        '47337' => '1815'             , # G+1815  U+B8E9  uniB8E9,  HANGUL SYLLABLE-B8E9
        '47338' => '11909'            , # G+11909  U+B8EA  uniB8EA,  HANGUL SYLLABLE-B8EA
        '47339' => '11910'            , # G+11910  U+B8EB  uniB8EB,  HANGUL SYLLABLE-B8EB
        '47340' => '1816'             , # G+1816  U+B8EC  uniB8EC,  HANGUL SYLLABLE-B8EC
        '47341' => '11911'            , # G+11911  U+B8ED  uniB8ED,  HANGUL SYLLABLE-B8ED
        '47342' => '11912'            , # G+11912  U+B8EE  uniB8EE,  HANGUL SYLLABLE-B8EE
        '47343' => '11913'            , # G+11913  U+B8EF  uniB8EF,  HANGUL SYLLABLE-B8EF
        '47344' => '1817'             , # G+1817  U+B8F0  uniB8F0,  HANGUL SYLLABLE-B8F0
        '47345' => '11914'            , # G+11914  U+B8F1  uniB8F1,  HANGUL SYLLABLE-B8F1
        '47346' => '11915'            , # G+11915  U+B8F2  uniB8F2,  HANGUL SYLLABLE-B8F2
        '47347' => '11916'            , # G+11916  U+B8F3  uniB8F3,  HANGUL SYLLABLE-B8F3
        '47348' => '11917'            , # G+11917  U+B8F4  uniB8F4,  HANGUL SYLLABLE-B8F4
        '47349' => '11918'            , # G+11918  U+B8F5  uniB8F5,  HANGUL SYLLABLE-B8F5
        '47350' => '11919'            , # G+11919  U+B8F6  uniB8F6,  HANGUL SYLLABLE-B8F6
        '47351' => '11920'            , # G+11920  U+B8F7  uniB8F7,  HANGUL SYLLABLE-B8F7
        '47352' => '1818'             , # G+1818  U+B8F8  uniB8F8,  HANGUL SYLLABLE-B8F8
        '47353' => '1819'             , # G+1819  U+B8F9  uniB8F9,  HANGUL SYLLABLE-B8F9
        '47354' => '11921'            , # G+11921  U+B8FA  uniB8FA,  HANGUL SYLLABLE-B8FA
        '47355' => '1820'             , # G+1820  U+B8FB  uniB8FB,  HANGUL SYLLABLE-B8FB
        '47356' => '11922'            , # G+11922  U+B8FC  uniB8FC,  HANGUL SYLLABLE-B8FC
        '47357' => '1821'             , # G+1821  U+B8FD  uniB8FD,  HANGUL SYLLABLE-B8FD
        '47358' => '11923'            , # G+11923  U+B8FE  uniB8FE,  HANGUL SYLLABLE-B8FE
        '47359' => '11924'            , # G+11924  U+B8FF  uniB8FF,  HANGUL SYLLABLE-B8FF
        '47360' => '11925'            , # G+11925  U+B900  uniB900,  HANGUL SYLLABLE-B900
        '47361' => '11926'            , # G+11926  U+B901  uniB901,  HANGUL SYLLABLE-B901
        '47362' => '11927'            , # G+11927  U+B902  uniB902,  HANGUL SYLLABLE-B902
        '47363' => '11928'            , # G+11928  U+B903  uniB903,  HANGUL SYLLABLE-B903
        '47364' => '1822'             , # G+1822  U+B904  uniB904,  HANGUL SYLLABLE-B904
        '47365' => '11929'            , # G+11929  U+B905  uniB905,  HANGUL SYLLABLE-B905
        '47366' => '11930'            , # G+11930  U+B906  uniB906,  HANGUL SYLLABLE-B906
        '47367' => '11931'            , # G+11931  U+B907  uniB907,  HANGUL SYLLABLE-B907
        '47368' => '11932'            , # G+11932  U+B908  uniB908,  HANGUL SYLLABLE-B908
        '47369' => '11933'            , # G+11933  U+B909  uniB909,  HANGUL SYLLABLE-B909
        '47370' => '11934'            , # G+11934  U+B90A  uniB90A,  HANGUL SYLLABLE-B90A
        '47371' => '11935'            , # G+11935  U+B90B  uniB90B,  HANGUL SYLLABLE-B90B
        '47372' => '11936'            , # G+11936  U+B90C  uniB90C,  HANGUL SYLLABLE-B90C
        '47373' => '11937'            , # G+11937  U+B90D  uniB90D,  HANGUL SYLLABLE-B90D
        '47374' => '11938'            , # G+11938  U+B90E  uniB90E,  HANGUL SYLLABLE-B90E
        '47375' => '11939'            , # G+11939  U+B90F  uniB90F,  HANGUL SYLLABLE-B90F
        '47376' => '11940'            , # G+11940  U+B910  uniB910,  HANGUL SYLLABLE-B910
        '47377' => '11941'            , # G+11941  U+B911  uniB911,  HANGUL SYLLABLE-B911
        '47378' => '11942'            , # G+11942  U+B912  uniB912,  HANGUL SYLLABLE-B912
        '47379' => '11943'            , # G+11943  U+B913  uniB913,  HANGUL SYLLABLE-B913
        '47380' => '11944'            , # G+11944  U+B914  uniB914,  HANGUL SYLLABLE-B914
        '47381' => '11945'            , # G+11945  U+B915  uniB915,  HANGUL SYLLABLE-B915
        '47382' => '11946'            , # G+11946  U+B916  uniB916,  HANGUL SYLLABLE-B916
        '47383' => '11947'            , # G+11947  U+B917  uniB917,  HANGUL SYLLABLE-B917
        '47384' => '1823'             , # G+1823  U+B918  uniB918,  HANGUL SYLLABLE-B918
        '47385' => '11948'            , # G+11948  U+B919  uniB919,  HANGUL SYLLABLE-B919
        '47386' => '11949'            , # G+11949  U+B91A  uniB91A,  HANGUL SYLLABLE-B91A
        '47387' => '11950'            , # G+11950  U+B91B  uniB91B,  HANGUL SYLLABLE-B91B
        '47388' => '11951'            , # G+11951  U+B91C  uniB91C,  HANGUL SYLLABLE-B91C
        '47389' => '11952'            , # G+11952  U+B91D  uniB91D,  HANGUL SYLLABLE-B91D
        '47390' => '11953'            , # G+11953  U+B91E  uniB91E,  HANGUL SYLLABLE-B91E
        '47391' => '11954'            , # G+11954  U+B91F  uniB91F,  HANGUL SYLLABLE-B91F
        '47392' => '1824'             , # G+1824  U+B920  uniB920,  HANGUL SYLLABLE-B920
        '47393' => '11955'            , # G+11955  U+B921  uniB921,  HANGUL SYLLABLE-B921
        '47394' => '11956'            , # G+11956  U+B922  uniB922,  HANGUL SYLLABLE-B922
        '47395' => '11957'            , # G+11957  U+B923  uniB923,  HANGUL SYLLABLE-B923
        '47396' => '11958'            , # G+11958  U+B924  uniB924,  HANGUL SYLLABLE-B924
        '47397' => '11959'            , # G+11959  U+B925  uniB925,  HANGUL SYLLABLE-B925
        '47398' => '11960'            , # G+11960  U+B926  uniB926,  HANGUL SYLLABLE-B926
        '47399' => '11961'            , # G+11961  U+B927  uniB927,  HANGUL SYLLABLE-B927
        '47400' => '11962'            , # G+11962  U+B928  uniB928,  HANGUL SYLLABLE-B928
        '47401' => '11963'            , # G+11963  U+B929  uniB929,  HANGUL SYLLABLE-B929
        '47402' => '11964'            , # G+11964  U+B92A  uniB92A,  HANGUL SYLLABLE-B92A
        '47403' => '11965'            , # G+11965  U+B92B  uniB92B,  HANGUL SYLLABLE-B92B
        '47404' => '11966'            , # G+11966  U+B92C  uniB92C,  HANGUL SYLLABLE-B92C
        '47405' => '11967'            , # G+11967  U+B92D  uniB92D,  HANGUL SYLLABLE-B92D
        '47406' => '11968'            , # G+11968  U+B92E  uniB92E,  HANGUL SYLLABLE-B92E
        '47407' => '11969'            , # G+11969  U+B92F  uniB92F,  HANGUL SYLLABLE-B92F
        '47408' => '11970'            , # G+11970  U+B930  uniB930,  HANGUL SYLLABLE-B930
        '47409' => '11971'            , # G+11971  U+B931  uniB931,  HANGUL SYLLABLE-B931
        '47410' => '11972'            , # G+11972  U+B932  uniB932,  HANGUL SYLLABLE-B932
        '47411' => '11973'            , # G+11973  U+B933  uniB933,  HANGUL SYLLABLE-B933
        '47412' => '11974'            , # G+11974  U+B934  uniB934,  HANGUL SYLLABLE-B934
        '47413' => '11975'            , # G+11975  U+B935  uniB935,  HANGUL SYLLABLE-B935
        '47414' => '11976'            , # G+11976  U+B936  uniB936,  HANGUL SYLLABLE-B936
        '47415' => '11977'            , # G+11977  U+B937  uniB937,  HANGUL SYLLABLE-B937
        '47416' => '11978'            , # G+11978  U+B938  uniB938,  HANGUL SYLLABLE-B938
        '47417' => '11979'            , # G+11979  U+B939  uniB939,  HANGUL SYLLABLE-B939
        '47418' => '11980'            , # G+11980  U+B93A  uniB93A,  HANGUL SYLLABLE-B93A
        '47419' => '11981'            , # G+11981  U+B93B  uniB93B,  HANGUL SYLLABLE-B93B
        '47420' => '1825'             , # G+1825  U+B93C  uniB93C,  HANGUL SYLLABLE-B93C
        '47421' => '1826'             , # G+1826  U+B93D  uniB93D,  HANGUL SYLLABLE-B93D
        '47422' => '11982'            , # G+11982  U+B93E  uniB93E,  HANGUL SYLLABLE-B93E
        '47423' => '11983'            , # G+11983  U+B93F  uniB93F,  HANGUL SYLLABLE-B93F
        '47424' => '1827'             , # G+1827  U+B940  uniB940,  HANGUL SYLLABLE-B940
        '47425' => '11984'            , # G+11984  U+B941  uniB941,  HANGUL SYLLABLE-B941
        '47426' => '11985'            , # G+11985  U+B942  uniB942,  HANGUL SYLLABLE-B942
        '47427' => '11986'            , # G+11986  U+B943  uniB943,  HANGUL SYLLABLE-B943
        '47428' => '1828'             , # G+1828  U+B944  uniB944,  HANGUL SYLLABLE-B944
        '47429' => '11987'            , # G+11987  U+B945  uniB945,  HANGUL SYLLABLE-B945
        '47430' => '11988'            , # G+11988  U+B946  uniB946,  HANGUL SYLLABLE-B946
        '47431' => '11989'            , # G+11989  U+B947  uniB947,  HANGUL SYLLABLE-B947
        '47432' => '11990'            , # G+11990  U+B948  uniB948,  HANGUL SYLLABLE-B948
        '47433' => '11991'            , # G+11991  U+B949  uniB949,  HANGUL SYLLABLE-B949
        '47434' => '11992'            , # G+11992  U+B94A  uniB94A,  HANGUL SYLLABLE-B94A
        '47435' => '11993'            , # G+11993  U+B94B  uniB94B,  HANGUL SYLLABLE-B94B
        '47436' => '1829'             , # G+1829  U+B94C  uniB94C,  HANGUL SYLLABLE-B94C
        '47437' => '11994'            , # G+11994  U+B94D  uniB94D,  HANGUL SYLLABLE-B94D
        '47438' => '11995'            , # G+11995  U+B94E  uniB94E,  HANGUL SYLLABLE-B94E
        '47439' => '1830'             , # G+1830  U+B94F  uniB94F,  HANGUL SYLLABLE-B94F
        '47440' => '11996'            , # G+11996  U+B950  uniB950,  HANGUL SYLLABLE-B950
        '47441' => '1831'             , # G+1831  U+B951  uniB951,  HANGUL SYLLABLE-B951
        '47442' => '11997'            , # G+11997  U+B952  uniB952,  HANGUL SYLLABLE-B952
        '47443' => '11998'            , # G+11998  U+B953  uniB953,  HANGUL SYLLABLE-B953
        '47444' => '11999'            , # G+11999  U+B954  uniB954,  HANGUL SYLLABLE-B954
        '47445' => '12000'            , # G+12000  U+B955  uniB955,  HANGUL SYLLABLE-B955
        '47446' => '12001'            , # G+12001  U+B956  uniB956,  HANGUL SYLLABLE-B956
        '47447' => '12002'            , # G+12002  U+B957  uniB957,  HANGUL SYLLABLE-B957
        '47448' => '1832'             , # G+1832  U+B958  uniB958,  HANGUL SYLLABLE-B958
        '47449' => '1833'             , # G+1833  U+B959  uniB959,  HANGUL SYLLABLE-B959
        '47450' => '12003'            , # G+12003  U+B95A  uniB95A,  HANGUL SYLLABLE-B95A
        '47451' => '12004'            , # G+12004  U+B95B  uniB95B,  HANGUL SYLLABLE-B95B
        '47452' => '1834'             , # G+1834  U+B95C  uniB95C,  HANGUL SYLLABLE-B95C
        '47453' => '12005'            , # G+12005  U+B95D  uniB95D,  HANGUL SYLLABLE-B95D
        '47454' => '12006'            , # G+12006  U+B95E  uniB95E,  HANGUL SYLLABLE-B95E
        '47455' => '12007'            , # G+12007  U+B95F  uniB95F,  HANGUL SYLLABLE-B95F
        '47456' => '1835'             , # G+1835  U+B960  uniB960,  HANGUL SYLLABLE-B960
        '47457' => '12008'            , # G+12008  U+B961  uniB961,  HANGUL SYLLABLE-B961
        '47458' => '12009'            , # G+12009  U+B962  uniB962,  HANGUL SYLLABLE-B962
        '47459' => '12010'            , # G+12010  U+B963  uniB963,  HANGUL SYLLABLE-B963
        '47460' => '12011'            , # G+12011  U+B964  uniB964,  HANGUL SYLLABLE-B964
        '47461' => '12012'            , # G+12012  U+B965  uniB965,  HANGUL SYLLABLE-B965
        '47462' => '12013'            , # G+12013  U+B966  uniB966,  HANGUL SYLLABLE-B966
        '47463' => '12014'            , # G+12014  U+B967  uniB967,  HANGUL SYLLABLE-B967
        '47464' => '1836'             , # G+1836  U+B968  uniB968,  HANGUL SYLLABLE-B968
        '47465' => '1837'             , # G+1837  U+B969  uniB969,  HANGUL SYLLABLE-B969
        '47466' => '12015'            , # G+12015  U+B96A  uniB96A,  HANGUL SYLLABLE-B96A
        '47467' => '1838'             , # G+1838  U+B96B  uniB96B,  HANGUL SYLLABLE-B96B
        '47468' => '12016'            , # G+12016  U+B96C  uniB96C,  HANGUL SYLLABLE-B96C
        '47469' => '1839'             , # G+1839  U+B96D  uniB96D,  HANGUL SYLLABLE-B96D
        '47470' => '12017'            , # G+12017  U+B96E  uniB96E,  HANGUL SYLLABLE-B96E
        '47471' => '12018'            , # G+12018  U+B96F  uniB96F,  HANGUL SYLLABLE-B96F
        '47472' => '12019'            , # G+12019  U+B970  uniB970,  HANGUL SYLLABLE-B970
        '47473' => '12020'            , # G+12020  U+B971  uniB971,  HANGUL SYLLABLE-B971
        '47474' => '12021'            , # G+12021  U+B972  uniB972,  HANGUL SYLLABLE-B972
        '47475' => '12022'            , # G+12022  U+B973  uniB973,  HANGUL SYLLABLE-B973
        '47476' => '1840'             , # G+1840  U+B974  uniB974,  HANGUL SYLLABLE-B974
        '47477' => '1841'             , # G+1841  U+B975  uniB975,  HANGUL SYLLABLE-B975
        '47478' => '12023'            , # G+12023  U+B976  uniB976,  HANGUL SYLLABLE-B976
        '47479' => '12024'            , # G+12024  U+B977  uniB977,  HANGUL SYLLABLE-B977
        '47480' => '1842'             , # G+1842  U+B978  uniB978,  HANGUL SYLLABLE-B978
        '47481' => '12025'            , # G+12025  U+B979  uniB979,  HANGUL SYLLABLE-B979
        '47482' => '12026'            , # G+12026  U+B97A  uniB97A,  HANGUL SYLLABLE-B97A
        '47483' => '12027'            , # G+12027  U+B97B  uniB97B,  HANGUL SYLLABLE-B97B
        '47484' => '1843'             , # G+1843  U+B97C  uniB97C,  HANGUL SYLLABLE-B97C
        '47485' => '12028'            , # G+12028  U+B97D  uniB97D,  HANGUL SYLLABLE-B97D
        '47486' => '12029'            , # G+12029  U+B97E  uniB97E,  HANGUL SYLLABLE-B97E
        '47487' => '12030'            , # G+12030  U+B97F  uniB97F,  HANGUL SYLLABLE-B97F
        '47488' => '12031'            , # G+12031  U+B980  uniB980,  HANGUL SYLLABLE-B980
        '47489' => '12032'            , # G+12032  U+B981  uniB981,  HANGUL SYLLABLE-B981
        '47490' => '12033'            , # G+12033  U+B982  uniB982,  HANGUL SYLLABLE-B982
        '47491' => '12034'            , # G+12034  U+B983  uniB983,  HANGUL SYLLABLE-B983
        '47492' => '1844'             , # G+1844  U+B984  uniB984,  HANGUL SYLLABLE-B984
        '47493' => '1845'             , # G+1845  U+B985  uniB985,  HANGUL SYLLABLE-B985
        '47494' => '12035'            , # G+12035  U+B986  uniB986,  HANGUL SYLLABLE-B986
        '47495' => '1846'             , # G+1846  U+B987  uniB987,  HANGUL SYLLABLE-B987
        '47496' => '12036'            , # G+12036  U+B988  uniB988,  HANGUL SYLLABLE-B988
        '47497' => '1847'             , # G+1847  U+B989  uniB989,  HANGUL SYLLABLE-B989
        '47498' => '1848'             , # G+1848  U+B98A  uniB98A,  HANGUL SYLLABLE-B98A
        '47499' => '12037'            , # G+12037  U+B98B  uniB98B,  HANGUL SYLLABLE-B98B
        '47500' => '12038'            , # G+12038  U+B98C  uniB98C,  HANGUL SYLLABLE-B98C
        '47501' => '1849'             , # G+1849  U+B98D  uniB98D,  HANGUL SYLLABLE-B98D
        '47502' => '1850'             , # G+1850  U+B98E  uniB98E,  HANGUL SYLLABLE-B98E
        '47503' => '12039'            , # G+12039  U+B98F  uniB98F,  HANGUL SYLLABLE-B98F
        '47504' => '12040'            , # G+12040  U+B990  uniB990,  HANGUL SYLLABLE-B990
        '47505' => '12041'            , # G+12041  U+B991  uniB991,  HANGUL SYLLABLE-B991
        '47506' => '12042'            , # G+12042  U+B992  uniB992,  HANGUL SYLLABLE-B992
        '47507' => '12043'            , # G+12043  U+B993  uniB993,  HANGUL SYLLABLE-B993
        '47508' => '12044'            , # G+12044  U+B994  uniB994,  HANGUL SYLLABLE-B994
        '47509' => '12045'            , # G+12045  U+B995  uniB995,  HANGUL SYLLABLE-B995
        '47510' => '12046'            , # G+12046  U+B996  uniB996,  HANGUL SYLLABLE-B996
        '47511' => '12047'            , # G+12047  U+B997  uniB997,  HANGUL SYLLABLE-B997
        '47512' => '12048'            , # G+12048  U+B998  uniB998,  HANGUL SYLLABLE-B998
        '47513' => '12049'            , # G+12049  U+B999  uniB999,  HANGUL SYLLABLE-B999
        '47514' => '12050'            , # G+12050  U+B99A  uniB99A,  HANGUL SYLLABLE-B99A
        '47515' => '12051'            , # G+12051  U+B99B  uniB99B,  HANGUL SYLLABLE-B99B
        '47516' => '12052'            , # G+12052  U+B99C  uniB99C,  HANGUL SYLLABLE-B99C
        '47517' => '12053'            , # G+12053  U+B99D  uniB99D,  HANGUL SYLLABLE-B99D
        '47518' => '12054'            , # G+12054  U+B99E  uniB99E,  HANGUL SYLLABLE-B99E
        '47519' => '12055'            , # G+12055  U+B99F  uniB99F,  HANGUL SYLLABLE-B99F
        '47520' => '12056'            , # G+12056  U+B9A0  uniB9A0,  HANGUL SYLLABLE-B9A0
        '47521' => '12057'            , # G+12057  U+B9A1  uniB9A1,  HANGUL SYLLABLE-B9A1
        '47522' => '12058'            , # G+12058  U+B9A2  uniB9A2,  HANGUL SYLLABLE-B9A2
        '47523' => '12059'            , # G+12059  U+B9A3  uniB9A3,  HANGUL SYLLABLE-B9A3
        '47524' => '12060'            , # G+12060  U+B9A4  uniB9A4,  HANGUL SYLLABLE-B9A4
        '47525' => '12061'            , # G+12061  U+B9A5  uniB9A5,  HANGUL SYLLABLE-B9A5
        '47526' => '12062'            , # G+12062  U+B9A6  uniB9A6,  HANGUL SYLLABLE-B9A6
        '47527' => '12063'            , # G+12063  U+B9A7  uniB9A7,  HANGUL SYLLABLE-B9A7
        '47528' => '12064'            , # G+12064  U+B9A8  uniB9A8,  HANGUL SYLLABLE-B9A8
        '47529' => '12065'            , # G+12065  U+B9A9  uniB9A9,  HANGUL SYLLABLE-B9A9
        '47530' => '12066'            , # G+12066  U+B9AA  uniB9AA,  HANGUL SYLLABLE-B9AA
        '47531' => '12067'            , # G+12067  U+B9AB  uniB9AB,  HANGUL SYLLABLE-B9AB
        '47532' => '1851'             , # G+1851  U+B9AC  uniB9AC,  HANGUL SYLLABLE-B9AC
        '47533' => '1852'             , # G+1852  U+B9AD  uniB9AD,  HANGUL SYLLABLE-B9AD
        '47534' => '12068'            , # G+12068  U+B9AE  uniB9AE,  HANGUL SYLLABLE-B9AE
        '47535' => '12069'            , # G+12069  U+B9AF  uniB9AF,  HANGUL SYLLABLE-B9AF
        '47536' => '1853'             , # G+1853  U+B9B0  uniB9B0,  HANGUL SYLLABLE-B9B0
        '47537' => '12070'            , # G+12070  U+B9B1  uniB9B1,  HANGUL SYLLABLE-B9B1
        '47538' => '12071'            , # G+12071  U+B9B2  uniB9B2,  HANGUL SYLLABLE-B9B2
        '47539' => '12072'            , # G+12072  U+B9B3  uniB9B3,  HANGUL SYLLABLE-B9B3
        '47540' => '1854'             , # G+1854  U+B9B4  uniB9B4,  HANGUL SYLLABLE-B9B4
        '47541' => '12073'            , # G+12073  U+B9B5  uniB9B5,  HANGUL SYLLABLE-B9B5
        '47542' => '12074'            , # G+12074  U+B9B6  uniB9B6,  HANGUL SYLLABLE-B9B6
        '47543' => '12075'            , # G+12075  U+B9B7  uniB9B7,  HANGUL SYLLABLE-B9B7
        '47544' => '12076'            , # G+12076  U+B9B8  uniB9B8,  HANGUL SYLLABLE-B9B8
        '47545' => '12077'            , # G+12077  U+B9B9  uniB9B9,  HANGUL SYLLABLE-B9B9
        '47546' => '12078'            , # G+12078  U+B9BA  uniB9BA,  HANGUL SYLLABLE-B9BA
        '47547' => '12079'            , # G+12079  U+B9BB  uniB9BB,  HANGUL SYLLABLE-B9BB
        '47548' => '1855'             , # G+1855  U+B9BC  uniB9BC,  HANGUL SYLLABLE-B9BC
        '47549' => '1856'             , # G+1856  U+B9BD  uniB9BD,  HANGUL SYLLABLE-B9BD
        '47550' => '12080'            , # G+12080  U+B9BE  uniB9BE,  HANGUL SYLLABLE-B9BE
        '47551' => '1857'             , # G+1857  U+B9BF  uniB9BF,  HANGUL SYLLABLE-B9BF
        '47552' => '12081'            , # G+12081  U+B9C0  uniB9C0,  HANGUL SYLLABLE-B9C0
        '47553' => '1858'             , # G+1858  U+B9C1  uniB9C1,  HANGUL SYLLABLE-B9C1
        '47554' => '12082'            , # G+12082  U+B9C2  uniB9C2,  HANGUL SYLLABLE-B9C2
        '47555' => '12083'            , # G+12083  U+B9C3  uniB9C3,  HANGUL SYLLABLE-B9C3
        '47556' => '12084'            , # G+12084  U+B9C4  uniB9C4,  HANGUL SYLLABLE-B9C4
        '47557' => '12085'            , # G+12085  U+B9C5  uniB9C5,  HANGUL SYLLABLE-B9C5
        '47558' => '12086'            , # G+12086  U+B9C6  uniB9C6,  HANGUL SYLLABLE-B9C6
        '47559' => '12087'            , # G+12087  U+B9C7  uniB9C7,  HANGUL SYLLABLE-B9C7
        '47560' => '1859'             , # G+1859  U+B9C8  uniB9C8,  HANGUL SYLLABLE-B9C8
        '47561' => '1860'             , # G+1860  U+B9C9  uniB9C9,  HANGUL SYLLABLE-B9C9
        '47562' => '12088'            , # G+12088  U+B9CA  uniB9CA,  HANGUL SYLLABLE-B9CA
        '47563' => '12089'            , # G+12089  U+B9CB  uniB9CB,  HANGUL SYLLABLE-B9CB
        '47564' => '1861'             , # G+1861  U+B9CC  uniB9CC,  HANGUL SYLLABLE-B9CC
        '47565' => '12090'            , # G+12090  U+B9CD  uniB9CD,  HANGUL SYLLABLE-B9CD
        '47566' => '1862'             , # G+1862  U+B9CE  uniB9CE,  HANGUL SYLLABLE-B9CE
        '47567' => '1863'             , # G+1863  U+B9CF  uniB9CF,  HANGUL SYLLABLE-B9CF
        '47568' => '1864'             , # G+1864  U+B9D0  uniB9D0,  HANGUL SYLLABLE-B9D0
        '47569' => '1865'             , # G+1865  U+B9D1  uniB9D1,  HANGUL SYLLABLE-B9D1
        '47570' => '1866'             , # G+1866  U+B9D2  uniB9D2,  HANGUL SYLLABLE-B9D2
        '47571' => '12091'            , # G+12091  U+B9D3  uniB9D3,  HANGUL SYLLABLE-B9D3
        '47572' => '12092'            , # G+12092  U+B9D4  uniB9D4,  HANGUL SYLLABLE-B9D4
        '47573' => '12093'            , # G+12093  U+B9D5  uniB9D5,  HANGUL SYLLABLE-B9D5
        '47574' => '12094'            , # G+12094  U+B9D6  uniB9D6,  HANGUL SYLLABLE-B9D6
        '47575' => '12095'            , # G+12095  U+B9D7  uniB9D7,  HANGUL SYLLABLE-B9D7
        '47576' => '1867'             , # G+1867  U+B9D8  uniB9D8,  HANGUL SYLLABLE-B9D8
        '47577' => '1868'             , # G+1868  U+B9D9  uniB9D9,  HANGUL SYLLABLE-B9D9
        '47578' => '12096'            , # G+12096  U+B9DA  uniB9DA,  HANGUL SYLLABLE-B9DA
        '47579' => '1869'             , # G+1869  U+B9DB  uniB9DB,  HANGUL SYLLABLE-B9DB
        '47580' => '12097'            , # G+12097  U+B9DC  uniB9DC,  HANGUL SYLLABLE-B9DC
        '47581' => '1870'             , # G+1870  U+B9DD  uniB9DD,  HANGUL SYLLABLE-B9DD
        '47582' => '1871'             , # G+1871  U+B9DE  uniB9DE,  HANGUL SYLLABLE-B9DE
        '47583' => '12098'            , # G+12098  U+B9DF  uniB9DF,  HANGUL SYLLABLE-B9DF
        '47584' => '12099'            , # G+12099  U+B9E0  uniB9E0,  HANGUL SYLLABLE-B9E0
        '47585' => '1872'             , # G+1872  U+B9E1  uniB9E1,  HANGUL SYLLABLE-B9E1
        '47586' => '12100'            , # G+12100  U+B9E2  uniB9E2,  HANGUL SYLLABLE-B9E2
        '47587' => '1873'             , # G+1873  U+B9E3  uniB9E3,  HANGUL SYLLABLE-B9E3
        '47588' => '1874'             , # G+1874  U+B9E4  uniB9E4,  HANGUL SYLLABLE-B9E4
        '47589' => '1875'             , # G+1875  U+B9E5  uniB9E5,  HANGUL SYLLABLE-B9E5
        '47590' => '12101'            , # G+12101  U+B9E6  uniB9E6,  HANGUL SYLLABLE-B9E6
        '47591' => '12102'            , # G+12102  U+B9E7  uniB9E7,  HANGUL SYLLABLE-B9E7
        '47592' => '1876'             , # G+1876  U+B9E8  uniB9E8,  HANGUL SYLLABLE-B9E8
        '47593' => '12103'            , # G+12103  U+B9E9  uniB9E9,  HANGUL SYLLABLE-B9E9
        '47594' => '12104'            , # G+12104  U+B9EA  uniB9EA,  HANGUL SYLLABLE-B9EA
        '47595' => '12105'            , # G+12105  U+B9EB  uniB9EB,  HANGUL SYLLABLE-B9EB
        '47596' => '1877'             , # G+1877  U+B9EC  uniB9EC,  HANGUL SYLLABLE-B9EC
        '47597' => '12106'            , # G+12106  U+B9ED  uniB9ED,  HANGUL SYLLABLE-B9ED
        '47598' => '12107'            , # G+12107  U+B9EE  uniB9EE,  HANGUL SYLLABLE-B9EE
        '47599' => '12108'            , # G+12108  U+B9EF  uniB9EF,  HANGUL SYLLABLE-B9EF
        '47600' => '12109'            , # G+12109  U+B9F0  uniB9F0,  HANGUL SYLLABLE-B9F0
        '47601' => '12110'            , # G+12110  U+B9F1  uniB9F1,  HANGUL SYLLABLE-B9F1
        '47602' => '12111'            , # G+12111  U+B9F2  uniB9F2,  HANGUL SYLLABLE-B9F2
        '47603' => '12112'            , # G+12112  U+B9F3  uniB9F3,  HANGUL SYLLABLE-B9F3
        '47604' => '1878'             , # G+1878  U+B9F4  uniB9F4,  HANGUL SYLLABLE-B9F4
        '47605' => '1879'             , # G+1879  U+B9F5  uniB9F5,  HANGUL SYLLABLE-B9F5
        '47606' => '12113'            , # G+12113  U+B9F6  uniB9F6,  HANGUL SYLLABLE-B9F6
        '47607' => '1880'             , # G+1880  U+B9F7  uniB9F7,  HANGUL SYLLABLE-B9F7
        '47608' => '1881'             , # G+1881  U+B9F8  uniB9F8,  HANGUL SYLLABLE-B9F8
        '47609' => '1882'             , # G+1882  U+B9F9  uniB9F9,  HANGUL SYLLABLE-B9F9
        '47610' => '1883'             , # G+1883  U+B9FA  uniB9FA,  HANGUL SYLLABLE-B9FA
        '47611' => '12114'            , # G+12114  U+B9FB  uniB9FB,  HANGUL SYLLABLE-B9FB
        '47612' => '12115'            , # G+12115  U+B9FC  uniB9FC,  HANGUL SYLLABLE-B9FC
        '47613' => '12116'            , # G+12116  U+B9FD  uniB9FD,  HANGUL SYLLABLE-B9FD
        '47614' => '12117'            , # G+12117  U+B9FE  uniB9FE,  HANGUL SYLLABLE-B9FE
        '47615' => '12118'            , # G+12118  U+B9FF  uniB9FF,  HANGUL SYLLABLE-B9FF
        '47616' => '1884'             , # G+1884  U+BA00  uniBA00,  HANGUL SYLLABLE-BA00
        '47617' => '1885'             , # G+1885  U+BA01  uniBA01,  HANGUL SYLLABLE-BA01
        '47618' => '12119'            , # G+12119  U+BA02  uniBA02,  HANGUL SYLLABLE-BA02
        '47619' => '12120'            , # G+12120  U+BA03  uniBA03,  HANGUL SYLLABLE-BA03
        '47620' => '12121'            , # G+12121  U+BA04  uniBA04,  HANGUL SYLLABLE-BA04
        '47621' => '12122'            , # G+12122  U+BA05  uniBA05,  HANGUL SYLLABLE-BA05
        '47622' => '12123'            , # G+12123  U+BA06  uniBA06,  HANGUL SYLLABLE-BA06
        '47623' => '12124'            , # G+12124  U+BA07  uniBA07,  HANGUL SYLLABLE-BA07
        '47624' => '1886'             , # G+1886  U+BA08  uniBA08,  HANGUL SYLLABLE-BA08
        '47625' => '12125'            , # G+12125  U+BA09  uniBA09,  HANGUL SYLLABLE-BA09
        '47626' => '12126'            , # G+12126  U+BA0A  uniBA0A,  HANGUL SYLLABLE-BA0A
        '47627' => '12127'            , # G+12127  U+BA0B  uniBA0B,  HANGUL SYLLABLE-BA0B
        '47628' => '12128'            , # G+12128  U+BA0C  uniBA0C,  HANGUL SYLLABLE-BA0C
        '47629' => '12129'            , # G+12129  U+BA0D  uniBA0D,  HANGUL SYLLABLE-BA0D
        '47630' => '12130'            , # G+12130  U+BA0E  uniBA0E,  HANGUL SYLLABLE-BA0E
        '47631' => '12131'            , # G+12131  U+BA0F  uniBA0F,  HANGUL SYLLABLE-BA0F
        '47632' => '12132'            , # G+12132  U+BA10  uniBA10,  HANGUL SYLLABLE-BA10
        '47633' => '12133'            , # G+12133  U+BA11  uniBA11,  HANGUL SYLLABLE-BA11
        '47634' => '12134'            , # G+12134  U+BA12  uniBA12,  HANGUL SYLLABLE-BA12
        '47635' => '12135'            , # G+12135  U+BA13  uniBA13,  HANGUL SYLLABLE-BA13
        '47636' => '12136'            , # G+12136  U+BA14  uniBA14,  HANGUL SYLLABLE-BA14
        '47637' => '1887'             , # G+1887  U+BA15  uniBA15,  HANGUL SYLLABLE-BA15
        '47638' => '12137'            , # G+12137  U+BA16  uniBA16,  HANGUL SYLLABLE-BA16
        '47639' => '12138'            , # G+12138  U+BA17  uniBA17,  HANGUL SYLLABLE-BA17
        '47640' => '12139'            , # G+12139  U+BA18  uniBA18,  HANGUL SYLLABLE-BA18
        '47641' => '12140'            , # G+12140  U+BA19  uniBA19,  HANGUL SYLLABLE-BA19
        '47642' => '12141'            , # G+12141  U+BA1A  uniBA1A,  HANGUL SYLLABLE-BA1A
        '47643' => '12142'            , # G+12142  U+BA1B  uniBA1B,  HANGUL SYLLABLE-BA1B
        '47644' => '12143'            , # G+12143  U+BA1C  uniBA1C,  HANGUL SYLLABLE-BA1C
        '47645' => '12144'            , # G+12144  U+BA1D  uniBA1D,  HANGUL SYLLABLE-BA1D
        '47646' => '12145'            , # G+12145  U+BA1E  uniBA1E,  HANGUL SYLLABLE-BA1E
        '47647' => '12146'            , # G+12146  U+BA1F  uniBA1F,  HANGUL SYLLABLE-BA1F
        '47648' => '12147'            , # G+12147  U+BA20  uniBA20,  HANGUL SYLLABLE-BA20
        '47649' => '12148'            , # G+12148  U+BA21  uniBA21,  HANGUL SYLLABLE-BA21
        '47650' => '12149'            , # G+12149  U+BA22  uniBA22,  HANGUL SYLLABLE-BA22
        '47651' => '12150'            , # G+12150  U+BA23  uniBA23,  HANGUL SYLLABLE-BA23
        '47652' => '12151'            , # G+12151  U+BA24  uniBA24,  HANGUL SYLLABLE-BA24
        '47653' => '12152'            , # G+12152  U+BA25  uniBA25,  HANGUL SYLLABLE-BA25
        '47654' => '12153'            , # G+12153  U+BA26  uniBA26,  HANGUL SYLLABLE-BA26
        '47655' => '12154'            , # G+12154  U+BA27  uniBA27,  HANGUL SYLLABLE-BA27
        '47656' => '12155'            , # G+12155  U+BA28  uniBA28,  HANGUL SYLLABLE-BA28
        '47657' => '12156'            , # G+12156  U+BA29  uniBA29,  HANGUL SYLLABLE-BA29
        '47658' => '12157'            , # G+12157  U+BA2A  uniBA2A,  HANGUL SYLLABLE-BA2A
        '47659' => '12158'            , # G+12158  U+BA2B  uniBA2B,  HANGUL SYLLABLE-BA2B
        '47660' => '12159'            , # G+12159  U+BA2C  uniBA2C,  HANGUL SYLLABLE-BA2C
        '47661' => '12160'            , # G+12160  U+BA2D  uniBA2D,  HANGUL SYLLABLE-BA2D
        '47662' => '12161'            , # G+12161  U+BA2E  uniBA2E,  HANGUL SYLLABLE-BA2E
        '47663' => '12162'            , # G+12162  U+BA2F  uniBA2F,  HANGUL SYLLABLE-BA2F
        '47664' => '12163'            , # G+12163  U+BA30  uniBA30,  HANGUL SYLLABLE-BA30
        '47665' => '12164'            , # G+12164  U+BA31  uniBA31,  HANGUL SYLLABLE-BA31
        '47666' => '12165'            , # G+12165  U+BA32  uniBA32,  HANGUL SYLLABLE-BA32
        '47667' => '12166'            , # G+12166  U+BA33  uniBA33,  HANGUL SYLLABLE-BA33
        '47668' => '12167'            , # G+12167  U+BA34  uniBA34,  HANGUL SYLLABLE-BA34
        '47669' => '12168'            , # G+12168  U+BA35  uniBA35,  HANGUL SYLLABLE-BA35
        '47670' => '12169'            , # G+12169  U+BA36  uniBA36,  HANGUL SYLLABLE-BA36
        '47671' => '12170'            , # G+12170  U+BA37  uniBA37,  HANGUL SYLLABLE-BA37
        '47672' => '1888'             , # G+1888  U+BA38  uniBA38,  HANGUL SYLLABLE-BA38
        '47673' => '1889'             , # G+1889  U+BA39  uniBA39,  HANGUL SYLLABLE-BA39
        '47674' => '12171'            , # G+12171  U+BA3A  uniBA3A,  HANGUL SYLLABLE-BA3A
        '47675' => '12172'            , # G+12172  U+BA3B  uniBA3B,  HANGUL SYLLABLE-BA3B
        '47676' => '1890'             , # G+1890  U+BA3C  uniBA3C,  HANGUL SYLLABLE-BA3C
        '47677' => '12173'            , # G+12173  U+BA3D  uniBA3D,  HANGUL SYLLABLE-BA3D
        '47678' => '12174'            , # G+12174  U+BA3E  uniBA3E,  HANGUL SYLLABLE-BA3E
        '47679' => '12175'            , # G+12175  U+BA3F  uniBA3F,  HANGUL SYLLABLE-BA3F
        '47680' => '1891'             , # G+1891  U+BA40  uniBA40,  HANGUL SYLLABLE-BA40
        '47681' => '12176'            , # G+12176  U+BA41  uniBA41,  HANGUL SYLLABLE-BA41
        '47682' => '1892'             , # G+1892  U+BA42  uniBA42,  HANGUL SYLLABLE-BA42
        '47683' => '12177'            , # G+12177  U+BA43  uniBA43,  HANGUL SYLLABLE-BA43
        '47684' => '12178'            , # G+12178  U+BA44  uniBA44,  HANGUL SYLLABLE-BA44
        '47685' => '12179'            , # G+12179  U+BA45  uniBA45,  HANGUL SYLLABLE-BA45
        '47686' => '12180'            , # G+12180  U+BA46  uniBA46,  HANGUL SYLLABLE-BA46
        '47687' => '12181'            , # G+12181  U+BA47  uniBA47,  HANGUL SYLLABLE-BA47
        '47688' => '1893'             , # G+1893  U+BA48  uniBA48,  HANGUL SYLLABLE-BA48
        '47689' => '1894'             , # G+1894  U+BA49  uniBA49,  HANGUL SYLLABLE-BA49
        '47690' => '12182'            , # G+12182  U+BA4A  uniBA4A,  HANGUL SYLLABLE-BA4A
        '47691' => '1895'             , # G+1895  U+BA4B  uniBA4B,  HANGUL SYLLABLE-BA4B
        '47692' => '12183'            , # G+12183  U+BA4C  uniBA4C,  HANGUL SYLLABLE-BA4C
        '47693' => '1896'             , # G+1896  U+BA4D  uniBA4D,  HANGUL SYLLABLE-BA4D
        '47694' => '1897'             , # G+1897  U+BA4E  uniBA4E,  HANGUL SYLLABLE-BA4E
        '47695' => '12184'            , # G+12184  U+BA4F  uniBA4F,  HANGUL SYLLABLE-BA4F
        '47696' => '12185'            , # G+12185  U+BA50  uniBA50,  HANGUL SYLLABLE-BA50
        '47697' => '12186'            , # G+12186  U+BA51  uniBA51,  HANGUL SYLLABLE-BA51
        '47698' => '12187'            , # G+12187  U+BA52  uniBA52,  HANGUL SYLLABLE-BA52
        '47699' => '1898'             , # G+1898  U+BA53  uniBA53,  HANGUL SYLLABLE-BA53
        '47700' => '1899'             , # G+1899  U+BA54  uniBA54,  HANGUL SYLLABLE-BA54
        '47701' => '1900'             , # G+1900  U+BA55  uniBA55,  HANGUL SYLLABLE-BA55
        '47702' => '12188'            , # G+12188  U+BA56  uniBA56,  HANGUL SYLLABLE-BA56
        '47703' => '12189'            , # G+12189  U+BA57  uniBA57,  HANGUL SYLLABLE-BA57
        '47704' => '1901'             , # G+1901  U+BA58  uniBA58,  HANGUL SYLLABLE-BA58
        '47705' => '12190'            , # G+12190  U+BA59  uniBA59,  HANGUL SYLLABLE-BA59
        '47706' => '12191'            , # G+12191  U+BA5A  uniBA5A,  HANGUL SYLLABLE-BA5A
        '47707' => '12192'            , # G+12192  U+BA5B  uniBA5B,  HANGUL SYLLABLE-BA5B
        '47708' => '1902'             , # G+1902  U+BA5C  uniBA5C,  HANGUL SYLLABLE-BA5C
        '47709' => '12193'            , # G+12193  U+BA5D  uniBA5D,  HANGUL SYLLABLE-BA5D
        '47710' => '12194'            , # G+12194  U+BA5E  uniBA5E,  HANGUL SYLLABLE-BA5E
        '47711' => '12195'            , # G+12195  U+BA5F  uniBA5F,  HANGUL SYLLABLE-BA5F
        '47712' => '12196'            , # G+12196  U+BA60  uniBA60,  HANGUL SYLLABLE-BA60
        '47713' => '12197'            , # G+12197  U+BA61  uniBA61,  HANGUL SYLLABLE-BA61
        '47714' => '12198'            , # G+12198  U+BA62  uniBA62,  HANGUL SYLLABLE-BA62
        '47715' => '12199'            , # G+12199  U+BA63  uniBA63,  HANGUL SYLLABLE-BA63
        '47716' => '1903'             , # G+1903  U+BA64  uniBA64,  HANGUL SYLLABLE-BA64
        '47717' => '1904'             , # G+1904  U+BA65  uniBA65,  HANGUL SYLLABLE-BA65
        '47718' => '12200'            , # G+12200  U+BA66  uniBA66,  HANGUL SYLLABLE-BA66
        '47719' => '1905'             , # G+1905  U+BA67  uniBA67,  HANGUL SYLLABLE-BA67
        '47720' => '1906'             , # G+1906  U+BA68  uniBA68,  HANGUL SYLLABLE-BA68
        '47721' => '1907'             , # G+1907  U+BA69  uniBA69,  HANGUL SYLLABLE-BA69
        '47722' => '12201'            , # G+12201  U+BA6A  uniBA6A,  HANGUL SYLLABLE-BA6A
        '47723' => '12202'            , # G+12202  U+BA6B  uniBA6B,  HANGUL SYLLABLE-BA6B
        '47724' => '12203'            , # G+12203  U+BA6C  uniBA6C,  HANGUL SYLLABLE-BA6C
        '47725' => '12204'            , # G+12204  U+BA6D  uniBA6D,  HANGUL SYLLABLE-BA6D
        '47726' => '12205'            , # G+12205  U+BA6E  uniBA6E,  HANGUL SYLLABLE-BA6E
        '47727' => '12206'            , # G+12206  U+BA6F  uniBA6F,  HANGUL SYLLABLE-BA6F
        '47728' => '1908'             , # G+1908  U+BA70  uniBA70,  HANGUL SYLLABLE-BA70
        '47729' => '1909'             , # G+1909  U+BA71  uniBA71,  HANGUL SYLLABLE-BA71
        '47730' => '12207'            , # G+12207  U+BA72  uniBA72,  HANGUL SYLLABLE-BA72
        '47731' => '12208'            , # G+12208  U+BA73  uniBA73,  HANGUL SYLLABLE-BA73
        '47732' => '1910'             , # G+1910  U+BA74  uniBA74,  HANGUL SYLLABLE-BA74
        '47733' => '12209'            , # G+12209  U+BA75  uniBA75,  HANGUL SYLLABLE-BA75
        '47734' => '12210'            , # G+12210  U+BA76  uniBA76,  HANGUL SYLLABLE-BA76
        '47735' => '12211'            , # G+12211  U+BA77  uniBA77,  HANGUL SYLLABLE-BA77
        '47736' => '1911'             , # G+1911  U+BA78  uniBA78,  HANGUL SYLLABLE-BA78
        '47737' => '12212'            , # G+12212  U+BA79  uniBA79,  HANGUL SYLLABLE-BA79
        '47738' => '12213'            , # G+12213  U+BA7A  uniBA7A,  HANGUL SYLLABLE-BA7A
        '47739' => '12214'            , # G+12214  U+BA7B  uniBA7B,  HANGUL SYLLABLE-BA7B
        '47740' => '12215'            , # G+12215  U+BA7C  uniBA7C,  HANGUL SYLLABLE-BA7C
        '47741' => '12216'            , # G+12216  U+BA7D  uniBA7D,  HANGUL SYLLABLE-BA7D
        '47742' => '12217'            , # G+12217  U+BA7E  uniBA7E,  HANGUL SYLLABLE-BA7E
        '47743' => '12218'            , # G+12218  U+BA7F  uniBA7F,  HANGUL SYLLABLE-BA7F
        '47744' => '12219'            , # G+12219  U+BA80  uniBA80,  HANGUL SYLLABLE-BA80
        '47745' => '12220'            , # G+12220  U+BA81  uniBA81,  HANGUL SYLLABLE-BA81
        '47746' => '12221'            , # G+12221  U+BA82  uniBA82,  HANGUL SYLLABLE-BA82
        '47747' => '1912'             , # G+1912  U+BA83  uniBA83,  HANGUL SYLLABLE-BA83
        '47748' => '1913'             , # G+1913  U+BA84  uniBA84,  HANGUL SYLLABLE-BA84
        '47749' => '1914'             , # G+1914  U+BA85  uniBA85,  HANGUL SYLLABLE-BA85
        '47750' => '12222'            , # G+12222  U+BA86  uniBA86,  HANGUL SYLLABLE-BA86
        '47751' => '1915'             , # G+1915  U+BA87  uniBA87,  HANGUL SYLLABLE-BA87
        '47752' => '12223'            , # G+12223  U+BA88  uniBA88,  HANGUL SYLLABLE-BA88
        '47753' => '12224'            , # G+12224  U+BA89  uniBA89,  HANGUL SYLLABLE-BA89
        '47754' => '12225'            , # G+12225  U+BA8A  uniBA8A,  HANGUL SYLLABLE-BA8A
        '47755' => '12226'            , # G+12226  U+BA8B  uniBA8B,  HANGUL SYLLABLE-BA8B
        '47756' => '1916'             , # G+1916  U+BA8C  uniBA8C,  HANGUL SYLLABLE-BA8C
        '47757' => '12227'            , # G+12227  U+BA8D  uniBA8D,  HANGUL SYLLABLE-BA8D
        '47758' => '12228'            , # G+12228  U+BA8E  uniBA8E,  HANGUL SYLLABLE-BA8E
        '47759' => '12229'            , # G+12229  U+BA8F  uniBA8F,  HANGUL SYLLABLE-BA8F
        '47760' => '12230'            , # G+12230  U+BA90  uniBA90,  HANGUL SYLLABLE-BA90
        '47761' => '12231'            , # G+12231  U+BA91  uniBA91,  HANGUL SYLLABLE-BA91
        '47762' => '12232'            , # G+12232  U+BA92  uniBA92,  HANGUL SYLLABLE-BA92
        '47763' => '12233'            , # G+12233  U+BA93  uniBA93,  HANGUL SYLLABLE-BA93
        '47764' => '12234'            , # G+12234  U+BA94  uniBA94,  HANGUL SYLLABLE-BA94
        '47765' => '12235'            , # G+12235  U+BA95  uniBA95,  HANGUL SYLLABLE-BA95
        '47766' => '12236'            , # G+12236  U+BA96  uniBA96,  HANGUL SYLLABLE-BA96
        '47767' => '12237'            , # G+12237  U+BA97  uniBA97,  HANGUL SYLLABLE-BA97
        '47768' => '12238'            , # G+12238  U+BA98  uniBA98,  HANGUL SYLLABLE-BA98
        '47769' => '12239'            , # G+12239  U+BA99  uniBA99,  HANGUL SYLLABLE-BA99
        '47770' => '12240'            , # G+12240  U+BA9A  uniBA9A,  HANGUL SYLLABLE-BA9A
        '47771' => '12241'            , # G+12241  U+BA9B  uniBA9B,  HANGUL SYLLABLE-BA9B
        '47772' => '12242'            , # G+12242  U+BA9C  uniBA9C,  HANGUL SYLLABLE-BA9C
        '47773' => '12243'            , # G+12243  U+BA9D  uniBA9D,  HANGUL SYLLABLE-BA9D
        '47774' => '12244'            , # G+12244  U+BA9E  uniBA9E,  HANGUL SYLLABLE-BA9E
        '47775' => '12245'            , # G+12245  U+BA9F  uniBA9F,  HANGUL SYLLABLE-BA9F
        '47776' => '12246'            , # G+12246  U+BAA0  uniBAA0,  HANGUL SYLLABLE-BAA0
        '47777' => '12247'            , # G+12247  U+BAA1  uniBAA1,  HANGUL SYLLABLE-BAA1
        '47778' => '12248'            , # G+12248  U+BAA2  uniBAA2,  HANGUL SYLLABLE-BAA2
        '47779' => '12249'            , # G+12249  U+BAA3  uniBAA3,  HANGUL SYLLABLE-BAA3
        '47780' => '12250'            , # G+12250  U+BAA4  uniBAA4,  HANGUL SYLLABLE-BAA4
        '47781' => '12251'            , # G+12251  U+BAA5  uniBAA5,  HANGUL SYLLABLE-BAA5
        '47782' => '12252'            , # G+12252  U+BAA6  uniBAA6,  HANGUL SYLLABLE-BAA6
        '47783' => '12253'            , # G+12253  U+BAA7  uniBAA7,  HANGUL SYLLABLE-BAA7
        '47784' => '1917'             , # G+1917  U+BAA8  uniBAA8,  HANGUL SYLLABLE-BAA8
        '47785' => '1918'             , # G+1918  U+BAA9  uniBAA9,  HANGUL SYLLABLE-BAA9
        '47786' => '12254'            , # G+12254  U+BAAA  uniBAAA,  HANGUL SYLLABLE-BAAA
        '47787' => '1919'             , # G+1919  U+BAAB  uniBAAB,  HANGUL SYLLABLE-BAAB
        '47788' => '1920'             , # G+1920  U+BAAC  uniBAAC,  HANGUL SYLLABLE-BAAC
        '47789' => '12255'            , # G+12255  U+BAAD  uniBAAD,  HANGUL SYLLABLE-BAAD
        '47790' => '12256'            , # G+12256  U+BAAE  uniBAAE,  HANGUL SYLLABLE-BAAE
        '47791' => '12257'            , # G+12257  U+BAAF  uniBAAF,  HANGUL SYLLABLE-BAAF
        '47792' => '1921'             , # G+1921  U+BAB0  uniBAB0,  HANGUL SYLLABLE-BAB0
        '47793' => '12258'            , # G+12258  U+BAB1  uniBAB1,  HANGUL SYLLABLE-BAB1
        '47794' => '1922'             , # G+1922  U+BAB2  uniBAB2,  HANGUL SYLLABLE-BAB2
        '47795' => '12259'            , # G+12259  U+BAB3  uniBAB3,  HANGUL SYLLABLE-BAB3
        '47796' => '12260'            , # G+12260  U+BAB4  uniBAB4,  HANGUL SYLLABLE-BAB4
        '47797' => '12261'            , # G+12261  U+BAB5  uniBAB5,  HANGUL SYLLABLE-BAB5
        '47798' => '12262'            , # G+12262  U+BAB6  uniBAB6,  HANGUL SYLLABLE-BAB6
        '47799' => '12263'            , # G+12263  U+BAB7  uniBAB7,  HANGUL SYLLABLE-BAB7
        '47800' => '1923'             , # G+1923  U+BAB8  uniBAB8,  HANGUL SYLLABLE-BAB8
        '47801' => '1924'             , # G+1924  U+BAB9  uniBAB9,  HANGUL SYLLABLE-BAB9
        '47802' => '12264'            , # G+12264  U+BABA  uniBABA,  HANGUL SYLLABLE-BABA
        '47803' => '1925'             , # G+1925  U+BABB  uniBABB,  HANGUL SYLLABLE-BABB
        '47804' => '12265'            , # G+12265  U+BABC  uniBABC,  HANGUL SYLLABLE-BABC
        '47805' => '1926'             , # G+1926  U+BABD  uniBABD,  HANGUL SYLLABLE-BABD
        '47806' => '12266'            , # G+12266  U+BABE  uniBABE,  HANGUL SYLLABLE-BABE
        '47807' => '12267'            , # G+12267  U+BABF  uniBABF,  HANGUL SYLLABLE-BABF
        '47808' => '12268'            , # G+12268  U+BAC0  uniBAC0,  HANGUL SYLLABLE-BAC0
        '47809' => '12269'            , # G+12269  U+BAC1  uniBAC1,  HANGUL SYLLABLE-BAC1
        '47810' => '12270'            , # G+12270  U+BAC2  uniBAC2,  HANGUL SYLLABLE-BAC2
        '47811' => '12271'            , # G+12271  U+BAC3  uniBAC3,  HANGUL SYLLABLE-BAC3
        '47812' => '1927'             , # G+1927  U+BAC4  uniBAC4,  HANGUL SYLLABLE-BAC4
        '47813' => '12272'            , # G+12272  U+BAC5  uniBAC5,  HANGUL SYLLABLE-BAC5
        '47814' => '12273'            , # G+12273  U+BAC6  uniBAC6,  HANGUL SYLLABLE-BAC6
        '47815' => '12274'            , # G+12274  U+BAC7  uniBAC7,  HANGUL SYLLABLE-BAC7
        '47816' => '1928'             , # G+1928  U+BAC8  uniBAC8,  HANGUL SYLLABLE-BAC8
        '47817' => '12275'            , # G+12275  U+BAC9  uniBAC9,  HANGUL SYLLABLE-BAC9
        '47818' => '12276'            , # G+12276  U+BACA  uniBACA,  HANGUL SYLLABLE-BACA
        '47819' => '12277'            , # G+12277  U+BACB  uniBACB,  HANGUL SYLLABLE-BACB
        '47820' => '12278'            , # G+12278  U+BACC  uniBACC,  HANGUL SYLLABLE-BACC
        '47821' => '12279'            , # G+12279  U+BACD  uniBACD,  HANGUL SYLLABLE-BACD
        '47822' => '12280'            , # G+12280  U+BACE  uniBACE,  HANGUL SYLLABLE-BACE
        '47823' => '12281'            , # G+12281  U+BACF  uniBACF,  HANGUL SYLLABLE-BACF
        '47824' => '12282'            , # G+12282  U+BAD0  uniBAD0,  HANGUL SYLLABLE-BAD0
        '47825' => '12283'            , # G+12283  U+BAD1  uniBAD1,  HANGUL SYLLABLE-BAD1
        '47826' => '12284'            , # G+12284  U+BAD2  uniBAD2,  HANGUL SYLLABLE-BAD2
        '47827' => '12285'            , # G+12285  U+BAD3  uniBAD3,  HANGUL SYLLABLE-BAD3
        '47828' => '12286'            , # G+12286  U+BAD4  uniBAD4,  HANGUL SYLLABLE-BAD4
        '47829' => '12287'            , # G+12287  U+BAD5  uniBAD5,  HANGUL SYLLABLE-BAD5
        '47830' => '12288'            , # G+12288  U+BAD6  uniBAD6,  HANGUL SYLLABLE-BAD6
        '47831' => '12289'            , # G+12289  U+BAD7  uniBAD7,  HANGUL SYLLABLE-BAD7
        '47832' => '1929'             , # G+1929  U+BAD8  uniBAD8,  HANGUL SYLLABLE-BAD8
        '47833' => '1930'             , # G+1930  U+BAD9  uniBAD9,  HANGUL SYLLABLE-BAD9
        '47834' => '12290'            , # G+12290  U+BADA  uniBADA,  HANGUL SYLLABLE-BADA
        '47835' => '12291'            , # G+12291  U+BADB  uniBADB,  HANGUL SYLLABLE-BADB
        '47836' => '12292'            , # G+12292  U+BADC  uniBADC,  HANGUL SYLLABLE-BADC
        '47837' => '12293'            , # G+12293  U+BADD  uniBADD,  HANGUL SYLLABLE-BADD
        '47838' => '12294'            , # G+12294  U+BADE  uniBADE,  HANGUL SYLLABLE-BADE
        '47839' => '12295'            , # G+12295  U+BADF  uniBADF,  HANGUL SYLLABLE-BADF
        '47840' => '12296'            , # G+12296  U+BAE0  uniBAE0,  HANGUL SYLLABLE-BAE0
        '47841' => '12297'            , # G+12297  U+BAE1  uniBAE1,  HANGUL SYLLABLE-BAE1
        '47842' => '12298'            , # G+12298  U+BAE2  uniBAE2,  HANGUL SYLLABLE-BAE2
        '47843' => '12299'            , # G+12299  U+BAE3  uniBAE3,  HANGUL SYLLABLE-BAE3
        '47844' => '12300'            , # G+12300  U+BAE4  uniBAE4,  HANGUL SYLLABLE-BAE4
        '47845' => '12301'            , # G+12301  U+BAE5  uniBAE5,  HANGUL SYLLABLE-BAE5
        '47846' => '12302'            , # G+12302  U+BAE6  uniBAE6,  HANGUL SYLLABLE-BAE6
        '47847' => '12303'            , # G+12303  U+BAE7  uniBAE7,  HANGUL SYLLABLE-BAE7
        '47848' => '12304'            , # G+12304  U+BAE8  uniBAE8,  HANGUL SYLLABLE-BAE8
        '47849' => '12305'            , # G+12305  U+BAE9  uniBAE9,  HANGUL SYLLABLE-BAE9
        '47850' => '12306'            , # G+12306  U+BAEA  uniBAEA,  HANGUL SYLLABLE-BAEA
        '47851' => '12307'            , # G+12307  U+BAEB  uniBAEB,  HANGUL SYLLABLE-BAEB
        '47852' => '12308'            , # G+12308  U+BAEC  uniBAEC,  HANGUL SYLLABLE-BAEC
        '47853' => '12309'            , # G+12309  U+BAED  uniBAED,  HANGUL SYLLABLE-BAED
        '47854' => '12310'            , # G+12310  U+BAEE  uniBAEE,  HANGUL SYLLABLE-BAEE
        '47855' => '12311'            , # G+12311  U+BAEF  uniBAEF,  HANGUL SYLLABLE-BAEF
        '47856' => '12312'            , # G+12312  U+BAF0  uniBAF0,  HANGUL SYLLABLE-BAF0
        '47857' => '12313'            , # G+12313  U+BAF1  uniBAF1,  HANGUL SYLLABLE-BAF1
        '47858' => '12314'            , # G+12314  U+BAF2  uniBAF2,  HANGUL SYLLABLE-BAF2
        '47859' => '12315'            , # G+12315  U+BAF3  uniBAF3,  HANGUL SYLLABLE-BAF3
        '47860' => '12316'            , # G+12316  U+BAF4  uniBAF4,  HANGUL SYLLABLE-BAF4
        '47861' => '12317'            , # G+12317  U+BAF5  uniBAF5,  HANGUL SYLLABLE-BAF5
        '47862' => '12318'            , # G+12318  U+BAF6  uniBAF6,  HANGUL SYLLABLE-BAF6
        '47863' => '12319'            , # G+12319  U+BAF7  uniBAF7,  HANGUL SYLLABLE-BAF7
        '47864' => '12320'            , # G+12320  U+BAF8  uniBAF8,  HANGUL SYLLABLE-BAF8
        '47865' => '12321'            , # G+12321  U+BAF9  uniBAF9,  HANGUL SYLLABLE-BAF9
        '47866' => '12322'            , # G+12322  U+BAFA  uniBAFA,  HANGUL SYLLABLE-BAFA
        '47867' => '12323'            , # G+12323  U+BAFB  uniBAFB,  HANGUL SYLLABLE-BAFB
        '47868' => '1931'             , # G+1931  U+BAFC  uniBAFC,  HANGUL SYLLABLE-BAFC
        '47869' => '12324'            , # G+12324  U+BAFD  uniBAFD,  HANGUL SYLLABLE-BAFD
        '47870' => '12325'            , # G+12325  U+BAFE  uniBAFE,  HANGUL SYLLABLE-BAFE
        '47871' => '12326'            , # G+12326  U+BAFF  uniBAFF,  HANGUL SYLLABLE-BAFF
        '47872' => '1932'             , # G+1932  U+BB00  uniBB00,  HANGUL SYLLABLE-BB00
        '47873' => '12327'            , # G+12327  U+BB01  uniBB01,  HANGUL SYLLABLE-BB01
        '47874' => '12328'            , # G+12328  U+BB02  uniBB02,  HANGUL SYLLABLE-BB02
        '47875' => '12329'            , # G+12329  U+BB03  uniBB03,  HANGUL SYLLABLE-BB03
        '47876' => '1933'             , # G+1933  U+BB04  uniBB04,  HANGUL SYLLABLE-BB04
        '47877' => '12330'            , # G+12330  U+BB05  uniBB05,  HANGUL SYLLABLE-BB05
        '47878' => '12331'            , # G+12331  U+BB06  uniBB06,  HANGUL SYLLABLE-BB06
        '47879' => '12332'            , # G+12332  U+BB07  uniBB07,  HANGUL SYLLABLE-BB07
        '47880' => '12333'            , # G+12333  U+BB08  uniBB08,  HANGUL SYLLABLE-BB08
        '47881' => '12334'            , # G+12334  U+BB09  uniBB09,  HANGUL SYLLABLE-BB09
        '47882' => '12335'            , # G+12335  U+BB0A  uniBB0A,  HANGUL SYLLABLE-BB0A
        '47883' => '12336'            , # G+12336  U+BB0B  uniBB0B,  HANGUL SYLLABLE-BB0B
        '47884' => '12337'            , # G+12337  U+BB0C  uniBB0C,  HANGUL SYLLABLE-BB0C
        '47885' => '1934'             , # G+1934  U+BB0D  uniBB0D,  HANGUL SYLLABLE-BB0D
        '47886' => '12338'            , # G+12338  U+BB0E  uniBB0E,  HANGUL SYLLABLE-BB0E
        '47887' => '1935'             , # G+1935  U+BB0F  uniBB0F,  HANGUL SYLLABLE-BB0F
        '47888' => '12339'            , # G+12339  U+BB10  uniBB10,  HANGUL SYLLABLE-BB10
        '47889' => '1936'             , # G+1936  U+BB11  uniBB11,  HANGUL SYLLABLE-BB11
        '47890' => '12340'            , # G+12340  U+BB12  uniBB12,  HANGUL SYLLABLE-BB12
        '47891' => '12341'            , # G+12341  U+BB13  uniBB13,  HANGUL SYLLABLE-BB13
        '47892' => '12342'            , # G+12342  U+BB14  uniBB14,  HANGUL SYLLABLE-BB14
        '47893' => '12343'            , # G+12343  U+BB15  uniBB15,  HANGUL SYLLABLE-BB15
        '47894' => '12344'            , # G+12344  U+BB16  uniBB16,  HANGUL SYLLABLE-BB16
        '47895' => '12345'            , # G+12345  U+BB17  uniBB17,  HANGUL SYLLABLE-BB17
        '47896' => '1937'             , # G+1937  U+BB18  uniBB18,  HANGUL SYLLABLE-BB18
        '47897' => '12346'            , # G+12346  U+BB19  uniBB19,  HANGUL SYLLABLE-BB19
        '47898' => '12347'            , # G+12347  U+BB1A  uniBB1A,  HANGUL SYLLABLE-BB1A
        '47899' => '12348'            , # G+12348  U+BB1B  uniBB1B,  HANGUL SYLLABLE-BB1B
        '47900' => '1938'             , # G+1938  U+BB1C  uniBB1C,  HANGUL SYLLABLE-BB1C
        '47901' => '12349'            , # G+12349  U+BB1D  uniBB1D,  HANGUL SYLLABLE-BB1D
        '47902' => '12350'            , # G+12350  U+BB1E  uniBB1E,  HANGUL SYLLABLE-BB1E
        '47903' => '12351'            , # G+12351  U+BB1F  uniBB1F,  HANGUL SYLLABLE-BB1F
        '47904' => '1939'             , # G+1939  U+BB20  uniBB20,  HANGUL SYLLABLE-BB20
        '47905' => '12352'            , # G+12352  U+BB21  uniBB21,  HANGUL SYLLABLE-BB21
        '47906' => '12353'            , # G+12353  U+BB22  uniBB22,  HANGUL SYLLABLE-BB22
        '47907' => '12354'            , # G+12354  U+BB23  uniBB23,  HANGUL SYLLABLE-BB23
        '47908' => '12355'            , # G+12355  U+BB24  uniBB24,  HANGUL SYLLABLE-BB24
        '47909' => '12356'            , # G+12356  U+BB25  uniBB25,  HANGUL SYLLABLE-BB25
        '47910' => '12357'            , # G+12357  U+BB26  uniBB26,  HANGUL SYLLABLE-BB26
        '47911' => '12358'            , # G+12358  U+BB27  uniBB27,  HANGUL SYLLABLE-BB27
        '47912' => '12359'            , # G+12359  U+BB28  uniBB28,  HANGUL SYLLABLE-BB28
        '47913' => '1940'             , # G+1940  U+BB29  uniBB29,  HANGUL SYLLABLE-BB29
        '47914' => '12360'            , # G+12360  U+BB2A  uniBB2A,  HANGUL SYLLABLE-BB2A
        '47915' => '1941'             , # G+1941  U+BB2B  uniBB2B,  HANGUL SYLLABLE-BB2B
        '47916' => '12361'            , # G+12361  U+BB2C  uniBB2C,  HANGUL SYLLABLE-BB2C
        '47917' => '12362'            , # G+12362  U+BB2D  uniBB2D,  HANGUL SYLLABLE-BB2D
        '47918' => '12363'            , # G+12363  U+BB2E  uniBB2E,  HANGUL SYLLABLE-BB2E
        '47919' => '12364'            , # G+12364  U+BB2F  uniBB2F,  HANGUL SYLLABLE-BB2F
        '47920' => '12365'            , # G+12365  U+BB30  uniBB30,  HANGUL SYLLABLE-BB30
        '47921' => '12366'            , # G+12366  U+BB31  uniBB31,  HANGUL SYLLABLE-BB31
        '47922' => '12367'            , # G+12367  U+BB32  uniBB32,  HANGUL SYLLABLE-BB32
        '47923' => '12368'            , # G+12368  U+BB33  uniBB33,  HANGUL SYLLABLE-BB33
        '47924' => '1942'             , # G+1942  U+BB34  uniBB34,  HANGUL SYLLABLE-BB34
        '47925' => '1943'             , # G+1943  U+BB35  uniBB35,  HANGUL SYLLABLE-BB35
        '47926' => '1944'             , # G+1944  U+BB36  uniBB36,  HANGUL SYLLABLE-BB36
        '47927' => '12369'            , # G+12369  U+BB37  uniBB37,  HANGUL SYLLABLE-BB37
        '47928' => '1945'             , # G+1945  U+BB38  uniBB38,  HANGUL SYLLABLE-BB38
        '47929' => '12370'            , # G+12370  U+BB39  uniBB39,  HANGUL SYLLABLE-BB39
        '47930' => '12371'            , # G+12371  U+BB3A  uniBB3A,  HANGUL SYLLABLE-BB3A
        '47931' => '1946'             , # G+1946  U+BB3B  uniBB3B,  HANGUL SYLLABLE-BB3B
        '47932' => '1947'             , # G+1947  U+BB3C  uniBB3C,  HANGUL SYLLABLE-BB3C
        '47933' => '1948'             , # G+1948  U+BB3D  uniBB3D,  HANGUL SYLLABLE-BB3D
        '47934' => '1949'             , # G+1949  U+BB3E  uniBB3E,  HANGUL SYLLABLE-BB3E
        '47935' => '12372'            , # G+12372  U+BB3F  uniBB3F,  HANGUL SYLLABLE-BB3F
        '47936' => '12373'            , # G+12373  U+BB40  uniBB40,  HANGUL SYLLABLE-BB40
        '47937' => '12374'            , # G+12374  U+BB41  uniBB41,  HANGUL SYLLABLE-BB41
        '47938' => '12375'            , # G+12375  U+BB42  uniBB42,  HANGUL SYLLABLE-BB42
        '47939' => '12376'            , # G+12376  U+BB43  uniBB43,  HANGUL SYLLABLE-BB43
        '47940' => '1950'             , # G+1950  U+BB44  uniBB44,  HANGUL SYLLABLE-BB44
        '47941' => '1951'             , # G+1951  U+BB45  uniBB45,  HANGUL SYLLABLE-BB45
        '47942' => '12377'            , # G+12377  U+BB46  uniBB46,  HANGUL SYLLABLE-BB46
        '47943' => '1952'             , # G+1952  U+BB47  uniBB47,  HANGUL SYLLABLE-BB47
        '47944' => '12378'            , # G+12378  U+BB48  uniBB48,  HANGUL SYLLABLE-BB48
        '47945' => '1953'             , # G+1953  U+BB49  uniBB49,  HANGUL SYLLABLE-BB49
        '47946' => '12379'            , # G+12379  U+BB4A  uniBB4A,  HANGUL SYLLABLE-BB4A
        '47947' => '12380'            , # G+12380  U+BB4B  uniBB4B,  HANGUL SYLLABLE-BB4B
        '47948' => '12381'            , # G+12381  U+BB4C  uniBB4C,  HANGUL SYLLABLE-BB4C
        '47949' => '1954'             , # G+1954  U+BB4D  uniBB4D,  HANGUL SYLLABLE-BB4D
        '47950' => '12382'            , # G+12382  U+BB4E  uniBB4E,  HANGUL SYLLABLE-BB4E
        '47951' => '1955'             , # G+1955  U+BB4F  uniBB4F,  HANGUL SYLLABLE-BB4F
        '47952' => '1956'             , # G+1956  U+BB50  uniBB50,  HANGUL SYLLABLE-BB50
        '47953' => '12383'            , # G+12383  U+BB51  uniBB51,  HANGUL SYLLABLE-BB51
        '47954' => '12384'            , # G+12384  U+BB52  uniBB52,  HANGUL SYLLABLE-BB52
        '47955' => '12385'            , # G+12385  U+BB53  uniBB53,  HANGUL SYLLABLE-BB53
        '47956' => '1957'             , # G+1957  U+BB54  uniBB54,  HANGUL SYLLABLE-BB54
        '47957' => '12386'            , # G+12386  U+BB55  uniBB55,  HANGUL SYLLABLE-BB55
        '47958' => '12387'            , # G+12387  U+BB56  uniBB56,  HANGUL SYLLABLE-BB56
        '47959' => '12388'            , # G+12388  U+BB57  uniBB57,  HANGUL SYLLABLE-BB57
        '47960' => '1958'             , # G+1958  U+BB58  uniBB58,  HANGUL SYLLABLE-BB58
        '47961' => '12389'            , # G+12389  U+BB59  uniBB59,  HANGUL SYLLABLE-BB59
        '47962' => '12390'            , # G+12390  U+BB5A  uniBB5A,  HANGUL SYLLABLE-BB5A
        '47963' => '12391'            , # G+12391  U+BB5B  uniBB5B,  HANGUL SYLLABLE-BB5B
        '47964' => '12392'            , # G+12392  U+BB5C  uniBB5C,  HANGUL SYLLABLE-BB5C
        '47965' => '12393'            , # G+12393  U+BB5D  uniBB5D,  HANGUL SYLLABLE-BB5D
        '47966' => '12394'            , # G+12394  U+BB5E  uniBB5E,  HANGUL SYLLABLE-BB5E
        '47967' => '12395'            , # G+12395  U+BB5F  uniBB5F,  HANGUL SYLLABLE-BB5F
        '47968' => '12396'            , # G+12396  U+BB60  uniBB60,  HANGUL SYLLABLE-BB60
        '47969' => '1959'             , # G+1959  U+BB61  uniBB61,  HANGUL SYLLABLE-BB61
        '47970' => '12397'            , # G+12397  U+BB62  uniBB62,  HANGUL SYLLABLE-BB62
        '47971' => '1960'             , # G+1960  U+BB63  uniBB63,  HANGUL SYLLABLE-BB63
        '47972' => '12398'            , # G+12398  U+BB64  uniBB64,  HANGUL SYLLABLE-BB64
        '47973' => '12399'            , # G+12399  U+BB65  uniBB65,  HANGUL SYLLABLE-BB65
        '47974' => '12400'            , # G+12400  U+BB66  uniBB66,  HANGUL SYLLABLE-BB66
        '47975' => '12401'            , # G+12401  U+BB67  uniBB67,  HANGUL SYLLABLE-BB67
        '47976' => '12402'            , # G+12402  U+BB68  uniBB68,  HANGUL SYLLABLE-BB68
        '47977' => '12403'            , # G+12403  U+BB69  uniBB69,  HANGUL SYLLABLE-BB69
        '47978' => '12404'            , # G+12404  U+BB6A  uniBB6A,  HANGUL SYLLABLE-BB6A
        '47979' => '12405'            , # G+12405  U+BB6B  uniBB6B,  HANGUL SYLLABLE-BB6B
        '47980' => '1961'             , # G+1961  U+BB6C  uniBB6C,  HANGUL SYLLABLE-BB6C
        '47981' => '12406'            , # G+12406  U+BB6D  uniBB6D,  HANGUL SYLLABLE-BB6D
        '47982' => '12407'            , # G+12407  U+BB6E  uniBB6E,  HANGUL SYLLABLE-BB6E
        '47983' => '12408'            , # G+12408  U+BB6F  uniBB6F,  HANGUL SYLLABLE-BB6F
        '47984' => '12409'            , # G+12409  U+BB70  uniBB70,  HANGUL SYLLABLE-BB70
        '47985' => '12410'            , # G+12410  U+BB71  uniBB71,  HANGUL SYLLABLE-BB71
        '47986' => '12411'            , # G+12411  U+BB72  uniBB72,  HANGUL SYLLABLE-BB72
        '47987' => '12412'            , # G+12412  U+BB73  uniBB73,  HANGUL SYLLABLE-BB73
        '47988' => '12413'            , # G+12413  U+BB74  uniBB74,  HANGUL SYLLABLE-BB74
        '47989' => '12414'            , # G+12414  U+BB75  uniBB75,  HANGUL SYLLABLE-BB75
        '47990' => '12415'            , # G+12415  U+BB76  uniBB76,  HANGUL SYLLABLE-BB76
        '47991' => '12416'            , # G+12416  U+BB77  uniBB77,  HANGUL SYLLABLE-BB77
        '47992' => '12417'            , # G+12417  U+BB78  uniBB78,  HANGUL SYLLABLE-BB78
        '47993' => '12418'            , # G+12418  U+BB79  uniBB79,  HANGUL SYLLABLE-BB79
        '47994' => '12419'            , # G+12419  U+BB7A  uniBB7A,  HANGUL SYLLABLE-BB7A
        '47995' => '12420'            , # G+12420  U+BB7B  uniBB7B,  HANGUL SYLLABLE-BB7B
        '47996' => '12421'            , # G+12421  U+BB7C  uniBB7C,  HANGUL SYLLABLE-BB7C
        '47997' => '12422'            , # G+12422  U+BB7D  uniBB7D,  HANGUL SYLLABLE-BB7D
        '47998' => '12423'            , # G+12423  U+BB7E  uniBB7E,  HANGUL SYLLABLE-BB7E
        '47999' => '12424'            , # G+12424  U+BB7F  uniBB7F,  HANGUL SYLLABLE-BB7F
        '48000' => '12425'            , # G+12425  U+BB80  uniBB80,  HANGUL SYLLABLE-BB80
        '48001' => '12426'            , # G+12426  U+BB81  uniBB81,  HANGUL SYLLABLE-BB81
        '48002' => '12427'            , # G+12427  U+BB82  uniBB82,  HANGUL SYLLABLE-BB82
        '48003' => '12428'            , # G+12428  U+BB83  uniBB83,  HANGUL SYLLABLE-BB83
        '48004' => '12429'            , # G+12429  U+BB84  uniBB84,  HANGUL SYLLABLE-BB84
        '48005' => '12430'            , # G+12430  U+BB85  uniBB85,  HANGUL SYLLABLE-BB85
        '48006' => '12431'            , # G+12431  U+BB86  uniBB86,  HANGUL SYLLABLE-BB86
        '48007' => '12432'            , # G+12432  U+BB87  uniBB87,  HANGUL SYLLABLE-BB87
        '48008' => '1962'             , # G+1962  U+BB88  uniBB88,  HANGUL SYLLABLE-BB88
        '48009' => '12433'            , # G+12433  U+BB89  uniBB89,  HANGUL SYLLABLE-BB89
        '48010' => '12434'            , # G+12434  U+BB8A  uniBB8A,  HANGUL SYLLABLE-BB8A
        '48011' => '12435'            , # G+12435  U+BB8B  uniBB8B,  HANGUL SYLLABLE-BB8B
        '48012' => '1963'             , # G+1963  U+BB8C  uniBB8C,  HANGUL SYLLABLE-BB8C
        '48013' => '12436'            , # G+12436  U+BB8D  uniBB8D,  HANGUL SYLLABLE-BB8D
        '48014' => '12437'            , # G+12437  U+BB8E  uniBB8E,  HANGUL SYLLABLE-BB8E
        '48015' => '12438'            , # G+12438  U+BB8F  uniBB8F,  HANGUL SYLLABLE-BB8F
        '48016' => '1964'             , # G+1964  U+BB90  uniBB90,  HANGUL SYLLABLE-BB90
        '48017' => '12439'            , # G+12439  U+BB91  uniBB91,  HANGUL SYLLABLE-BB91
        '48018' => '12440'            , # G+12440  U+BB92  uniBB92,  HANGUL SYLLABLE-BB92
        '48019' => '12441'            , # G+12441  U+BB93  uniBB93,  HANGUL SYLLABLE-BB93
        '48020' => '12442'            , # G+12442  U+BB94  uniBB94,  HANGUL SYLLABLE-BB94
        '48021' => '12443'            , # G+12443  U+BB95  uniBB95,  HANGUL SYLLABLE-BB95
        '48022' => '12444'            , # G+12444  U+BB96  uniBB96,  HANGUL SYLLABLE-BB96
        '48023' => '12445'            , # G+12445  U+BB97  uniBB97,  HANGUL SYLLABLE-BB97
        '48024' => '12446'            , # G+12446  U+BB98  uniBB98,  HANGUL SYLLABLE-BB98
        '48025' => '12447'            , # G+12447  U+BB99  uniBB99,  HANGUL SYLLABLE-BB99
        '48026' => '12448'            , # G+12448  U+BB9A  uniBB9A,  HANGUL SYLLABLE-BB9A
        '48027' => '12449'            , # G+12449  U+BB9B  uniBB9B,  HANGUL SYLLABLE-BB9B
        '48028' => '12450'            , # G+12450  U+BB9C  uniBB9C,  HANGUL SYLLABLE-BB9C
        '48029' => '12451'            , # G+12451  U+BB9D  uniBB9D,  HANGUL SYLLABLE-BB9D
        '48030' => '12452'            , # G+12452  U+BB9E  uniBB9E,  HANGUL SYLLABLE-BB9E
        '48031' => '12453'            , # G+12453  U+BB9F  uniBB9F,  HANGUL SYLLABLE-BB9F
        '48032' => '12454'            , # G+12454  U+BBA0  uniBBA0,  HANGUL SYLLABLE-BBA0
        '48033' => '12455'            , # G+12455  U+BBA1  uniBBA1,  HANGUL SYLLABLE-BBA1
        '48034' => '12456'            , # G+12456  U+BBA2  uniBBA2,  HANGUL SYLLABLE-BBA2
        '48035' => '12457'            , # G+12457  U+BBA3  uniBBA3,  HANGUL SYLLABLE-BBA3
        '48036' => '1965'             , # G+1965  U+BBA4  uniBBA4,  HANGUL SYLLABLE-BBA4
        '48037' => '12458'            , # G+12458  U+BBA5  uniBBA5,  HANGUL SYLLABLE-BBA5
        '48038' => '12459'            , # G+12459  U+BBA6  uniBBA6,  HANGUL SYLLABLE-BBA6
        '48039' => '12460'            , # G+12460  U+BBA7  uniBBA7,  HANGUL SYLLABLE-BBA7
        '48040' => '1966'             , # G+1966  U+BBA8  uniBBA8,  HANGUL SYLLABLE-BBA8
        '48041' => '12461'            , # G+12461  U+BBA9  uniBBA9,  HANGUL SYLLABLE-BBA9
        '48042' => '12462'            , # G+12462  U+BBAA  uniBBAA,  HANGUL SYLLABLE-BBAA
        '48043' => '12463'            , # G+12463  U+BBAB  uniBBAB,  HANGUL SYLLABLE-BBAB
        '48044' => '1967'             , # G+1967  U+BBAC  uniBBAC,  HANGUL SYLLABLE-BBAC
        '48045' => '12464'            , # G+12464  U+BBAD  uniBBAD,  HANGUL SYLLABLE-BBAD
        '48046' => '12465'            , # G+12465  U+BBAE  uniBBAE,  HANGUL SYLLABLE-BBAE
        '48047' => '12466'            , # G+12466  U+BBAF  uniBBAF,  HANGUL SYLLABLE-BBAF
        '48048' => '12467'            , # G+12467  U+BBB0  uniBBB0,  HANGUL SYLLABLE-BBB0
        '48049' => '12468'            , # G+12468  U+BBB1  uniBBB1,  HANGUL SYLLABLE-BBB1
        '48050' => '12469'            , # G+12469  U+BBB2  uniBBB2,  HANGUL SYLLABLE-BBB2
        '48051' => '12470'            , # G+12470  U+BBB3  uniBBB3,  HANGUL SYLLABLE-BBB3
        '48052' => '1968'             , # G+1968  U+BBB4  uniBBB4,  HANGUL SYLLABLE-BBB4
        '48053' => '12471'            , # G+12471  U+BBB5  uniBBB5,  HANGUL SYLLABLE-BBB5
        '48054' => '12472'            , # G+12472  U+BBB6  uniBBB6,  HANGUL SYLLABLE-BBB6
        '48055' => '1969'             , # G+1969  U+BBB7  uniBBB7,  HANGUL SYLLABLE-BBB7
        '48056' => '12473'            , # G+12473  U+BBB8  uniBBB8,  HANGUL SYLLABLE-BBB8
        '48057' => '12474'            , # G+12474  U+BBB9  uniBBB9,  HANGUL SYLLABLE-BBB9
        '48058' => '12475'            , # G+12475  U+BBBA  uniBBBA,  HANGUL SYLLABLE-BBBA
        '48059' => '12476'            , # G+12476  U+BBBB  uniBBBB,  HANGUL SYLLABLE-BBBB
        '48060' => '12477'            , # G+12477  U+BBBC  uniBBBC,  HANGUL SYLLABLE-BBBC
        '48061' => '12478'            , # G+12478  U+BBBD  uniBBBD,  HANGUL SYLLABLE-BBBD
        '48062' => '12479'            , # G+12479  U+BBBE  uniBBBE,  HANGUL SYLLABLE-BBBE
        '48063' => '12480'            , # G+12480  U+BBBF  uniBBBF,  HANGUL SYLLABLE-BBBF
        '48064' => '1970'             , # G+1970  U+BBC0  uniBBC0,  HANGUL SYLLABLE-BBC0
        '48065' => '12481'            , # G+12481  U+BBC1  uniBBC1,  HANGUL SYLLABLE-BBC1
        '48066' => '12482'            , # G+12482  U+BBC2  uniBBC2,  HANGUL SYLLABLE-BBC2
        '48067' => '12483'            , # G+12483  U+BBC3  uniBBC3,  HANGUL SYLLABLE-BBC3
        '48068' => '1971'             , # G+1971  U+BBC4  uniBBC4,  HANGUL SYLLABLE-BBC4
        '48069' => '12484'            , # G+12484  U+BBC5  uniBBC5,  HANGUL SYLLABLE-BBC5
        '48070' => '12485'            , # G+12485  U+BBC6  uniBBC6,  HANGUL SYLLABLE-BBC6
        '48071' => '12486'            , # G+12486  U+BBC7  uniBBC7,  HANGUL SYLLABLE-BBC7
        '48072' => '1972'             , # G+1972  U+BBC8  uniBBC8,  HANGUL SYLLABLE-BBC8
        '48073' => '12487'            , # G+12487  U+BBC9  uniBBC9,  HANGUL SYLLABLE-BBC9
        '48074' => '12488'            , # G+12488  U+BBCA  uniBBCA,  HANGUL SYLLABLE-BBCA
        '48075' => '12489'            , # G+12489  U+BBCB  uniBBCB,  HANGUL SYLLABLE-BBCB
        '48076' => '12490'            , # G+12490  U+BBCC  uniBBCC,  HANGUL SYLLABLE-BBCC
        '48077' => '12491'            , # G+12491  U+BBCD  uniBBCD,  HANGUL SYLLABLE-BBCD
        '48078' => '12492'            , # G+12492  U+BBCE  uniBBCE,  HANGUL SYLLABLE-BBCE
        '48079' => '12493'            , # G+12493  U+BBCF  uniBBCF,  HANGUL SYLLABLE-BBCF
        '48080' => '1973'             , # G+1973  U+BBD0  uniBBD0,  HANGUL SYLLABLE-BBD0
        '48081' => '12494'            , # G+12494  U+BBD1  uniBBD1,  HANGUL SYLLABLE-BBD1
        '48082' => '12495'            , # G+12495  U+BBD2  uniBBD2,  HANGUL SYLLABLE-BBD2
        '48083' => '1974'             , # G+1974  U+BBD3  uniBBD3,  HANGUL SYLLABLE-BBD3
        '48084' => '12496'            , # G+12496  U+BBD4  uniBBD4,  HANGUL SYLLABLE-BBD4
        '48085' => '12497'            , # G+12497  U+BBD5  uniBBD5,  HANGUL SYLLABLE-BBD5
        '48086' => '12498'            , # G+12498  U+BBD6  uniBBD6,  HANGUL SYLLABLE-BBD6
        '48087' => '12499'            , # G+12499  U+BBD7  uniBBD7,  HANGUL SYLLABLE-BBD7
        '48088' => '12500'            , # G+12500  U+BBD8  uniBBD8,  HANGUL SYLLABLE-BBD8
        '48089' => '12501'            , # G+12501  U+BBD9  uniBBD9,  HANGUL SYLLABLE-BBD9
        '48090' => '12502'            , # G+12502  U+BBDA  uniBBDA,  HANGUL SYLLABLE-BBDA
        '48091' => '12503'            , # G+12503  U+BBDB  uniBBDB,  HANGUL SYLLABLE-BBDB
        '48092' => '12504'            , # G+12504  U+BBDC  uniBBDC,  HANGUL SYLLABLE-BBDC
        '48093' => '12505'            , # G+12505  U+BBDD  uniBBDD,  HANGUL SYLLABLE-BBDD
        '48094' => '12506'            , # G+12506  U+BBDE  uniBBDE,  HANGUL SYLLABLE-BBDE
        '48095' => '12507'            , # G+12507  U+BBDF  uniBBDF,  HANGUL SYLLABLE-BBDF
        '48096' => '12508'            , # G+12508  U+BBE0  uniBBE0,  HANGUL SYLLABLE-BBE0
        '48097' => '12509'            , # G+12509  U+BBE1  uniBBE1,  HANGUL SYLLABLE-BBE1
        '48098' => '12510'            , # G+12510  U+BBE2  uniBBE2,  HANGUL SYLLABLE-BBE2
        '48099' => '12511'            , # G+12511  U+BBE3  uniBBE3,  HANGUL SYLLABLE-BBE3
        '48100' => '12512'            , # G+12512  U+BBE4  uniBBE4,  HANGUL SYLLABLE-BBE4
        '48101' => '12513'            , # G+12513  U+BBE5  uniBBE5,  HANGUL SYLLABLE-BBE5
        '48102' => '12514'            , # G+12514  U+BBE6  uniBBE6,  HANGUL SYLLABLE-BBE6
        '48103' => '12515'            , # G+12515  U+BBE7  uniBBE7,  HANGUL SYLLABLE-BBE7
        '48104' => '12516'            , # G+12516  U+BBE8  uniBBE8,  HANGUL SYLLABLE-BBE8
        '48105' => '12517'            , # G+12517  U+BBE9  uniBBE9,  HANGUL SYLLABLE-BBE9
        '48106' => '12518'            , # G+12518  U+BBEA  uniBBEA,  HANGUL SYLLABLE-BBEA
        '48107' => '12519'            , # G+12519  U+BBEB  uniBBEB,  HANGUL SYLLABLE-BBEB
        '48108' => '12520'            , # G+12520  U+BBEC  uniBBEC,  HANGUL SYLLABLE-BBEC
        '48109' => '12521'            , # G+12521  U+BBED  uniBBED,  HANGUL SYLLABLE-BBED
        '48110' => '12522'            , # G+12522  U+BBEE  uniBBEE,  HANGUL SYLLABLE-BBEE
        '48111' => '12523'            , # G+12523  U+BBEF  uniBBEF,  HANGUL SYLLABLE-BBEF
        '48112' => '12524'            , # G+12524  U+BBF0  uniBBF0,  HANGUL SYLLABLE-BBF0
        '48113' => '12525'            , # G+12525  U+BBF1  uniBBF1,  HANGUL SYLLABLE-BBF1
        '48114' => '12526'            , # G+12526  U+BBF2  uniBBF2,  HANGUL SYLLABLE-BBF2
        '48115' => '12527'            , # G+12527  U+BBF3  uniBBF3,  HANGUL SYLLABLE-BBF3
        '48116' => '12528'            , # G+12528  U+BBF4  uniBBF4,  HANGUL SYLLABLE-BBF4
        '48117' => '12529'            , # G+12529  U+BBF5  uniBBF5,  HANGUL SYLLABLE-BBF5
        '48118' => '12530'            , # G+12530  U+BBF6  uniBBF6,  HANGUL SYLLABLE-BBF6
        '48119' => '12531'            , # G+12531  U+BBF7  uniBBF7,  HANGUL SYLLABLE-BBF7
        '48120' => '1975'             , # G+1975  U+BBF8  uniBBF8,  HANGUL SYLLABLE-BBF8
        '48121' => '1976'             , # G+1976  U+BBF9  uniBBF9,  HANGUL SYLLABLE-BBF9
        '48122' => '12532'            , # G+12532  U+BBFA  uniBBFA,  HANGUL SYLLABLE-BBFA
        '48123' => '12533'            , # G+12533  U+BBFB  uniBBFB,  HANGUL SYLLABLE-BBFB
        '48124' => '1977'             , # G+1977  U+BBFC  uniBBFC,  HANGUL SYLLABLE-BBFC
        '48125' => '12534'            , # G+12534  U+BBFD  uniBBFD,  HANGUL SYLLABLE-BBFD
        '48126' => '12535'            , # G+12535  U+BBFE  uniBBFE,  HANGUL SYLLABLE-BBFE
        '48127' => '1978'             , # G+1978  U+BBFF  uniBBFF,  HANGUL SYLLABLE-BBFF
        '48128' => '1979'             , # G+1979  U+BC00  uniBC00,  HANGUL SYLLABLE-BC00
        '48129' => '12536'            , # G+12536  U+BC01  uniBC01,  HANGUL SYLLABLE-BC01
        '48130' => '1980'             , # G+1980  U+BC02  uniBC02,  HANGUL SYLLABLE-BC02
        '48131' => '12537'            , # G+12537  U+BC03  uniBC03,  HANGUL SYLLABLE-BC03
        '48132' => '12538'            , # G+12538  U+BC04  uniBC04,  HANGUL SYLLABLE-BC04
        '48133' => '12539'            , # G+12539  U+BC05  uniBC05,  HANGUL SYLLABLE-BC05
        '48134' => '12540'            , # G+12540  U+BC06  uniBC06,  HANGUL SYLLABLE-BC06
        '48135' => '12541'            , # G+12541  U+BC07  uniBC07,  HANGUL SYLLABLE-BC07
        '48136' => '1981'             , # G+1981  U+BC08  uniBC08,  HANGUL SYLLABLE-BC08
        '48137' => '1982'             , # G+1982  U+BC09  uniBC09,  HANGUL SYLLABLE-BC09
        '48138' => '12542'            , # G+12542  U+BC0A  uniBC0A,  HANGUL SYLLABLE-BC0A
        '48139' => '1983'             , # G+1983  U+BC0B  uniBC0B,  HANGUL SYLLABLE-BC0B
        '48140' => '1984'             , # G+1984  U+BC0C  uniBC0C,  HANGUL SYLLABLE-BC0C
        '48141' => '1985'             , # G+1985  U+BC0D  uniBC0D,  HANGUL SYLLABLE-BC0D
        '48142' => '12543'            , # G+12543  U+BC0E  uniBC0E,  HANGUL SYLLABLE-BC0E
        '48143' => '1986'             , # G+1986  U+BC0F  uniBC0F,  HANGUL SYLLABLE-BC0F
        '48144' => '12544'            , # G+12544  U+BC10  uniBC10,  HANGUL SYLLABLE-BC10
        '48145' => '1987'             , # G+1987  U+BC11  uniBC11,  HANGUL SYLLABLE-BC11
        '48146' => '12545'            , # G+12545  U+BC12  uniBC12,  HANGUL SYLLABLE-BC12
        '48147' => '12546'            , # G+12546  U+BC13  uniBC13,  HANGUL SYLLABLE-BC13
        '48148' => '1988'             , # G+1988  U+BC14  uniBC14,  HANGUL SYLLABLE-BC14
        '48149' => '1989'             , # G+1989  U+BC15  uniBC15,  HANGUL SYLLABLE-BC15
        '48150' => '1990'             , # G+1990  U+BC16  uniBC16,  HANGUL SYLLABLE-BC16
        '48151' => '1991'             , # G+1991  U+BC17  uniBC17,  HANGUL SYLLABLE-BC17
        '48152' => '1992'             , # G+1992  U+BC18  uniBC18,  HANGUL SYLLABLE-BC18
        '48153' => '12547'            , # G+12547  U+BC19  uniBC19,  HANGUL SYLLABLE-BC19
        '48154' => '12548'            , # G+12548  U+BC1A  uniBC1A,  HANGUL SYLLABLE-BC1A
        '48155' => '1993'             , # G+1993  U+BC1B  uniBC1B,  HANGUL SYLLABLE-BC1B
        '48156' => '1994'             , # G+1994  U+BC1C  uniBC1C,  HANGUL SYLLABLE-BC1C
        '48157' => '1995'             , # G+1995  U+BC1D  uniBC1D,  HANGUL SYLLABLE-BC1D
        '48158' => '1996'             , # G+1996  U+BC1E  uniBC1E,  HANGUL SYLLABLE-BC1E
        '48159' => '1997'             , # G+1997  U+BC1F  uniBC1F,  HANGUL SYLLABLE-BC1F
        '48160' => '12549'            , # G+12549  U+BC20  uniBC20,  HANGUL SYLLABLE-BC20
        '48161' => '12550'            , # G+12550  U+BC21  uniBC21,  HANGUL SYLLABLE-BC21
        '48162' => '12551'            , # G+12551  U+BC22  uniBC22,  HANGUL SYLLABLE-BC22
        '48163' => '12552'            , # G+12552  U+BC23  uniBC23,  HANGUL SYLLABLE-BC23
        '48164' => '1998'             , # G+1998  U+BC24  uniBC24,  HANGUL SYLLABLE-BC24
        '48165' => '1999'             , # G+1999  U+BC25  uniBC25,  HANGUL SYLLABLE-BC25
        '48166' => '12553'            , # G+12553  U+BC26  uniBC26,  HANGUL SYLLABLE-BC26
        '48167' => '2000'             , # G+2000  U+BC27  uniBC27,  HANGUL SYLLABLE-BC27
        '48168' => '12554'            , # G+12554  U+BC28  uniBC28,  HANGUL SYLLABLE-BC28
        '48169' => '2001'             , # G+2001  U+BC29  uniBC29,  HANGUL SYLLABLE-BC29
        '48170' => '12555'            , # G+12555  U+BC2A  uniBC2A,  HANGUL SYLLABLE-BC2A
        '48171' => '12556'            , # G+12556  U+BC2B  uniBC2B,  HANGUL SYLLABLE-BC2B
        '48172' => '12557'            , # G+12557  U+BC2C  uniBC2C,  HANGUL SYLLABLE-BC2C
        '48173' => '2002'             , # G+2002  U+BC2D  uniBC2D,  HANGUL SYLLABLE-BC2D
        '48174' => '12558'            , # G+12558  U+BC2E  uniBC2E,  HANGUL SYLLABLE-BC2E
        '48175' => '12559'            , # G+12559  U+BC2F  uniBC2F,  HANGUL SYLLABLE-BC2F
        '48176' => '2003'             , # G+2003  U+BC30  uniBC30,  HANGUL SYLLABLE-BC30
        '48177' => '2004'             , # G+2004  U+BC31  uniBC31,  HANGUL SYLLABLE-BC31
        '48178' => '12560'            , # G+12560  U+BC32  uniBC32,  HANGUL SYLLABLE-BC32
        '48179' => '12561'            , # G+12561  U+BC33  uniBC33,  HANGUL SYLLABLE-BC33
        '48180' => '2005'             , # G+2005  U+BC34  uniBC34,  HANGUL SYLLABLE-BC34
        '48181' => '12562'            , # G+12562  U+BC35  uniBC35,  HANGUL SYLLABLE-BC35
        '48182' => '12563'            , # G+12563  U+BC36  uniBC36,  HANGUL SYLLABLE-BC36
        '48183' => '12564'            , # G+12564  U+BC37  uniBC37,  HANGUL SYLLABLE-BC37
        '48184' => '2006'             , # G+2006  U+BC38  uniBC38,  HANGUL SYLLABLE-BC38
        '48185' => '12565'            , # G+12565  U+BC39  uniBC39,  HANGUL SYLLABLE-BC39
        '48186' => '12566'            , # G+12566  U+BC3A  uniBC3A,  HANGUL SYLLABLE-BC3A
        '48187' => '12567'            , # G+12567  U+BC3B  uniBC3B,  HANGUL SYLLABLE-BC3B
        '48188' => '12568'            , # G+12568  U+BC3C  uniBC3C,  HANGUL SYLLABLE-BC3C
        '48189' => '12569'            , # G+12569  U+BC3D  uniBC3D,  HANGUL SYLLABLE-BC3D
        '48190' => '12570'            , # G+12570  U+BC3E  uniBC3E,  HANGUL SYLLABLE-BC3E
        '48191' => '12571'            , # G+12571  U+BC3F  uniBC3F,  HANGUL SYLLABLE-BC3F
        '48192' => '2007'             , # G+2007  U+BC40  uniBC40,  HANGUL SYLLABLE-BC40
        '48193' => '2008'             , # G+2008  U+BC41  uniBC41,  HANGUL SYLLABLE-BC41
        '48194' => '12572'            , # G+12572  U+BC42  uniBC42,  HANGUL SYLLABLE-BC42
        '48195' => '2009'             , # G+2009  U+BC43  uniBC43,  HANGUL SYLLABLE-BC43
        '48196' => '2010'             , # G+2010  U+BC44  uniBC44,  HANGUL SYLLABLE-BC44
        '48197' => '2011'             , # G+2011  U+BC45  uniBC45,  HANGUL SYLLABLE-BC45
        '48198' => '12573'            , # G+12573  U+BC46  uniBC46,  HANGUL SYLLABLE-BC46
        '48199' => '12574'            , # G+12574  U+BC47  uniBC47,  HANGUL SYLLABLE-BC47
        '48200' => '12575'            , # G+12575  U+BC48  uniBC48,  HANGUL SYLLABLE-BC48
        '48201' => '2012'             , # G+2012  U+BC49  uniBC49,  HANGUL SYLLABLE-BC49
        '48202' => '12576'            , # G+12576  U+BC4A  uniBC4A,  HANGUL SYLLABLE-BC4A
        '48203' => '12577'            , # G+12577  U+BC4B  uniBC4B,  HANGUL SYLLABLE-BC4B
        '48204' => '2013'             , # G+2013  U+BC4C  uniBC4C,  HANGUL SYLLABLE-BC4C
        '48205' => '2014'             , # G+2014  U+BC4D  uniBC4D,  HANGUL SYLLABLE-BC4D
        '48206' => '12578'            , # G+12578  U+BC4E  uniBC4E,  HANGUL SYLLABLE-BC4E
        '48207' => '12579'            , # G+12579  U+BC4F  uniBC4F,  HANGUL SYLLABLE-BC4F
        '48208' => '2015'             , # G+2015  U+BC50  uniBC50,  HANGUL SYLLABLE-BC50
        '48209' => '12580'            , # G+12580  U+BC51  uniBC51,  HANGUL SYLLABLE-BC51
        '48210' => '12581'            , # G+12581  U+BC52  uniBC52,  HANGUL SYLLABLE-BC52
        '48211' => '12582'            , # G+12582  U+BC53  uniBC53,  HANGUL SYLLABLE-BC53
        '48212' => '12583'            , # G+12583  U+BC54  uniBC54,  HANGUL SYLLABLE-BC54
        '48213' => '12584'            , # G+12584  U+BC55  uniBC55,  HANGUL SYLLABLE-BC55
        '48214' => '12585'            , # G+12585  U+BC56  uniBC56,  HANGUL SYLLABLE-BC56
        '48215' => '12586'            , # G+12586  U+BC57  uniBC57,  HANGUL SYLLABLE-BC57
        '48216' => '12587'            , # G+12587  U+BC58  uniBC58,  HANGUL SYLLABLE-BC58
        '48217' => '12588'            , # G+12588  U+BC59  uniBC59,  HANGUL SYLLABLE-BC59
        '48218' => '12589'            , # G+12589  U+BC5A  uniBC5A,  HANGUL SYLLABLE-BC5A
        '48219' => '12590'            , # G+12590  U+BC5B  uniBC5B,  HANGUL SYLLABLE-BC5B
        '48220' => '12591'            , # G+12591  U+BC5C  uniBC5C,  HANGUL SYLLABLE-BC5C
        '48221' => '2016'             , # G+2016  U+BC5D  uniBC5D,  HANGUL SYLLABLE-BC5D
        '48222' => '12592'            , # G+12592  U+BC5E  uniBC5E,  HANGUL SYLLABLE-BC5E
        '48223' => '12593'            , # G+12593  U+BC5F  uniBC5F,  HANGUL SYLLABLE-BC5F
        '48224' => '12594'            , # G+12594  U+BC60  uniBC60,  HANGUL SYLLABLE-BC60
        '48225' => '12595'            , # G+12595  U+BC61  uniBC61,  HANGUL SYLLABLE-BC61
        '48226' => '12596'            , # G+12596  U+BC62  uniBC62,  HANGUL SYLLABLE-BC62
        '48227' => '12597'            , # G+12597  U+BC63  uniBC63,  HANGUL SYLLABLE-BC63
        '48228' => '12598'            , # G+12598  U+BC64  uniBC64,  HANGUL SYLLABLE-BC64
        '48229' => '12599'            , # G+12599  U+BC65  uniBC65,  HANGUL SYLLABLE-BC65
        '48230' => '12600'            , # G+12600  U+BC66  uniBC66,  HANGUL SYLLABLE-BC66
        '48231' => '12601'            , # G+12601  U+BC67  uniBC67,  HANGUL SYLLABLE-BC67
        '48232' => '12602'            , # G+12602  U+BC68  uniBC68,  HANGUL SYLLABLE-BC68
        '48233' => '12603'            , # G+12603  U+BC69  uniBC69,  HANGUL SYLLABLE-BC69
        '48234' => '12604'            , # G+12604  U+BC6A  uniBC6A,  HANGUL SYLLABLE-BC6A
        '48235' => '12605'            , # G+12605  U+BC6B  uniBC6B,  HANGUL SYLLABLE-BC6B
        '48236' => '12606'            , # G+12606  U+BC6C  uniBC6C,  HANGUL SYLLABLE-BC6C
        '48237' => '12607'            , # G+12607  U+BC6D  uniBC6D,  HANGUL SYLLABLE-BC6D
        '48238' => '12608'            , # G+12608  U+BC6E  uniBC6E,  HANGUL SYLLABLE-BC6E
        '48239' => '12609'            , # G+12609  U+BC6F  uniBC6F,  HANGUL SYLLABLE-BC6F
        '48240' => '12610'            , # G+12610  U+BC70  uniBC70,  HANGUL SYLLABLE-BC70
        '48241' => '12611'            , # G+12611  U+BC71  uniBC71,  HANGUL SYLLABLE-BC71
        '48242' => '12612'            , # G+12612  U+BC72  uniBC72,  HANGUL SYLLABLE-BC72
        '48243' => '12613'            , # G+12613  U+BC73  uniBC73,  HANGUL SYLLABLE-BC73
        '48244' => '12614'            , # G+12614  U+BC74  uniBC74,  HANGUL SYLLABLE-BC74
        '48245' => '12615'            , # G+12615  U+BC75  uniBC75,  HANGUL SYLLABLE-BC75
        '48246' => '12616'            , # G+12616  U+BC76  uniBC76,  HANGUL SYLLABLE-BC76
        '48247' => '12617'            , # G+12617  U+BC77  uniBC77,  HANGUL SYLLABLE-BC77
        '48248' => '12618'            , # G+12618  U+BC78  uniBC78,  HANGUL SYLLABLE-BC78
        '48249' => '12619'            , # G+12619  U+BC79  uniBC79,  HANGUL SYLLABLE-BC79
        '48250' => '12620'            , # G+12620  U+BC7A  uniBC7A,  HANGUL SYLLABLE-BC7A
        '48251' => '12621'            , # G+12621  U+BC7B  uniBC7B,  HANGUL SYLLABLE-BC7B
        '48252' => '12622'            , # G+12622  U+BC7C  uniBC7C,  HANGUL SYLLABLE-BC7C
        '48253' => '12623'            , # G+12623  U+BC7D  uniBC7D,  HANGUL SYLLABLE-BC7D
        '48254' => '12624'            , # G+12624  U+BC7E  uniBC7E,  HANGUL SYLLABLE-BC7E
        '48255' => '12625'            , # G+12625  U+BC7F  uniBC7F,  HANGUL SYLLABLE-BC7F
        '48256' => '12626'            , # G+12626  U+BC80  uniBC80,  HANGUL SYLLABLE-BC80
        '48257' => '12627'            , # G+12627  U+BC81  uniBC81,  HANGUL SYLLABLE-BC81
        '48258' => '12628'            , # G+12628  U+BC82  uniBC82,  HANGUL SYLLABLE-BC82
        '48259' => '12629'            , # G+12629  U+BC83  uniBC83,  HANGUL SYLLABLE-BC83
        '48260' => '2017'             , # G+2017  U+BC84  uniBC84,  HANGUL SYLLABLE-BC84
        '48261' => '2018'             , # G+2018  U+BC85  uniBC85,  HANGUL SYLLABLE-BC85
        '48262' => '12630'            , # G+12630  U+BC86  uniBC86,  HANGUL SYLLABLE-BC86
        '48263' => '12631'            , # G+12631  U+BC87  uniBC87,  HANGUL SYLLABLE-BC87
        '48264' => '2019'             , # G+2019  U+BC88  uniBC88,  HANGUL SYLLABLE-BC88
        '48265' => '12632'            , # G+12632  U+BC89  uniBC89,  HANGUL SYLLABLE-BC89
        '48266' => '12633'            , # G+12633  U+BC8A  uniBC8A,  HANGUL SYLLABLE-BC8A
        '48267' => '2020'             , # G+2020  U+BC8B  uniBC8B,  HANGUL SYLLABLE-BC8B
        '48268' => '2021'             , # G+2021  U+BC8C  uniBC8C,  HANGUL SYLLABLE-BC8C
        '48269' => '12634'            , # G+12634  U+BC8D  uniBC8D,  HANGUL SYLLABLE-BC8D
        '48270' => '2022'             , # G+2022  U+BC8E  uniBC8E,  HANGUL SYLLABLE-BC8E
        '48271' => '12635'            , # G+12635  U+BC8F  uniBC8F,  HANGUL SYLLABLE-BC8F
        '48272' => '12636'            , # G+12636  U+BC90  uniBC90,  HANGUL SYLLABLE-BC90
        '48273' => '12637'            , # G+12637  U+BC91  uniBC91,  HANGUL SYLLABLE-BC91
        '48274' => '12638'            , # G+12638  U+BC92  uniBC92,  HANGUL SYLLABLE-BC92
        '48275' => '12639'            , # G+12639  U+BC93  uniBC93,  HANGUL SYLLABLE-BC93
        '48276' => '2023'             , # G+2023  U+BC94  uniBC94,  HANGUL SYLLABLE-BC94
        '48277' => '2024'             , # G+2024  U+BC95  uniBC95,  HANGUL SYLLABLE-BC95
        '48278' => '12640'            , # G+12640  U+BC96  uniBC96,  HANGUL SYLLABLE-BC96
        '48279' => '2025'             , # G+2025  U+BC97  uniBC97,  HANGUL SYLLABLE-BC97
        '48280' => '12641'            , # G+12641  U+BC98  uniBC98,  HANGUL SYLLABLE-BC98
        '48281' => '2026'             , # G+2026  U+BC99  uniBC99,  HANGUL SYLLABLE-BC99
        '48282' => '2027'             , # G+2027  U+BC9A  uniBC9A,  HANGUL SYLLABLE-BC9A
        '48283' => '12642'            , # G+12642  U+BC9B  uniBC9B,  HANGUL SYLLABLE-BC9B
        '48284' => '12643'            , # G+12643  U+BC9C  uniBC9C,  HANGUL SYLLABLE-BC9C
        '48285' => '12644'            , # G+12644  U+BC9D  uniBC9D,  HANGUL SYLLABLE-BC9D
        '48286' => '12645'            , # G+12645  U+BC9E  uniBC9E,  HANGUL SYLLABLE-BC9E
        '48287' => '12646'            , # G+12646  U+BC9F  uniBC9F,  HANGUL SYLLABLE-BC9F
        '48288' => '2028'             , # G+2028  U+BCA0  uniBCA0,  HANGUL SYLLABLE-BCA0
        '48289' => '2029'             , # G+2029  U+BCA1  uniBCA1,  HANGUL SYLLABLE-BCA1
        '48290' => '12647'            , # G+12647  U+BCA2  uniBCA2,  HANGUL SYLLABLE-BCA2
        '48291' => '12648'            , # G+12648  U+BCA3  uniBCA3,  HANGUL SYLLABLE-BCA3
        '48292' => '2030'             , # G+2030  U+BCA4  uniBCA4,  HANGUL SYLLABLE-BCA4
        '48293' => '12649'            , # G+12649  U+BCA5  uniBCA5,  HANGUL SYLLABLE-BCA5
        '48294' => '12650'            , # G+12650  U+BCA6  uniBCA6,  HANGUL SYLLABLE-BCA6
        '48295' => '2031'             , # G+2031  U+BCA7  uniBCA7,  HANGUL SYLLABLE-BCA7
        '48296' => '2032'             , # G+2032  U+BCA8  uniBCA8,  HANGUL SYLLABLE-BCA8
        '48297' => '12651'            , # G+12651  U+BCA9  uniBCA9,  HANGUL SYLLABLE-BCA9
        '48298' => '12652'            , # G+12652  U+BCAA  uniBCAA,  HANGUL SYLLABLE-BCAA
        '48299' => '12653'            , # G+12653  U+BCAB  uniBCAB,  HANGUL SYLLABLE-BCAB
        '48300' => '12654'            , # G+12654  U+BCAC  uniBCAC,  HANGUL SYLLABLE-BCAC
        '48301' => '12655'            , # G+12655  U+BCAD  uniBCAD,  HANGUL SYLLABLE-BCAD
        '48302' => '12656'            , # G+12656  U+BCAE  uniBCAE,  HANGUL SYLLABLE-BCAE
        '48303' => '12657'            , # G+12657  U+BCAF  uniBCAF,  HANGUL SYLLABLE-BCAF
        '48304' => '2033'             , # G+2033  U+BCB0  uniBCB0,  HANGUL SYLLABLE-BCB0
        '48305' => '2034'             , # G+2034  U+BCB1  uniBCB1,  HANGUL SYLLABLE-BCB1
        '48306' => '12658'            , # G+12658  U+BCB2  uniBCB2,  HANGUL SYLLABLE-BCB2
        '48307' => '2035'             , # G+2035  U+BCB3  uniBCB3,  HANGUL SYLLABLE-BCB3
        '48308' => '2036'             , # G+2036  U+BCB4  uniBCB4,  HANGUL SYLLABLE-BCB4
        '48309' => '2037'             , # G+2037  U+BCB5  uniBCB5,  HANGUL SYLLABLE-BCB5
        '48310' => '12659'            , # G+12659  U+BCB6  uniBCB6,  HANGUL SYLLABLE-BCB6
        '48311' => '12660'            , # G+12660  U+BCB7  uniBCB7,  HANGUL SYLLABLE-BCB7
        '48312' => '12661'            , # G+12661  U+BCB8  uniBCB8,  HANGUL SYLLABLE-BCB8
        '48313' => '12662'            , # G+12662  U+BCB9  uniBCB9,  HANGUL SYLLABLE-BCB9
        '48314' => '12663'            , # G+12663  U+BCBA  uniBCBA,  HANGUL SYLLABLE-BCBA
        '48315' => '12664'            , # G+12664  U+BCBB  uniBCBB,  HANGUL SYLLABLE-BCBB
        '48316' => '2038'             , # G+2038  U+BCBC  uniBCBC,  HANGUL SYLLABLE-BCBC
        '48317' => '2039'             , # G+2039  U+BCBD  uniBCBD,  HANGUL SYLLABLE-BCBD
        '48318' => '12665'            , # G+12665  U+BCBE  uniBCBE,  HANGUL SYLLABLE-BCBE
        '48319' => '12666'            , # G+12666  U+BCBF  uniBCBF,  HANGUL SYLLABLE-BCBF
        '48320' => '2040'             , # G+2040  U+BCC0  uniBCC0,  HANGUL SYLLABLE-BCC0
        '48321' => '12667'            , # G+12667  U+BCC1  uniBCC1,  HANGUL SYLLABLE-BCC1
        '48322' => '12668'            , # G+12668  U+BCC2  uniBCC2,  HANGUL SYLLABLE-BCC2
        '48323' => '12669'            , # G+12669  U+BCC3  uniBCC3,  HANGUL SYLLABLE-BCC3
        '48324' => '2041'             , # G+2041  U+BCC4  uniBCC4,  HANGUL SYLLABLE-BCC4
        '48325' => '12670'            , # G+12670  U+BCC5  uniBCC5,  HANGUL SYLLABLE-BCC5
        '48326' => '12671'            , # G+12671  U+BCC6  uniBCC6,  HANGUL SYLLABLE-BCC6
        '48327' => '12672'            , # G+12672  U+BCC7  uniBCC7,  HANGUL SYLLABLE-BCC7
        '48328' => '12673'            , # G+12673  U+BCC8  uniBCC8,  HANGUL SYLLABLE-BCC8
        '48329' => '12674'            , # G+12674  U+BCC9  uniBCC9,  HANGUL SYLLABLE-BCC9
        '48330' => '12675'            , # G+12675  U+BCCA  uniBCCA,  HANGUL SYLLABLE-BCCA
        '48331' => '12676'            , # G+12676  U+BCCB  uniBCCB,  HANGUL SYLLABLE-BCCB
        '48332' => '12677'            , # G+12677  U+BCCC  uniBCCC,  HANGUL SYLLABLE-BCCC
        '48333' => '2042'             , # G+2042  U+BCCD  uniBCCD,  HANGUL SYLLABLE-BCCD
        '48334' => '12678'            , # G+12678  U+BCCE  uniBCCE,  HANGUL SYLLABLE-BCCE
        '48335' => '2043'             , # G+2043  U+BCCF  uniBCCF,  HANGUL SYLLABLE-BCCF
        '48336' => '2044'             , # G+2044  U+BCD0  uniBCD0,  HANGUL SYLLABLE-BCD0
        '48337' => '2045'             , # G+2045  U+BCD1  uniBCD1,  HANGUL SYLLABLE-BCD1
        '48338' => '12679'            , # G+12679  U+BCD2  uniBCD2,  HANGUL SYLLABLE-BCD2
        '48339' => '12680'            , # G+12680  U+BCD3  uniBCD3,  HANGUL SYLLABLE-BCD3
        '48340' => '12681'            , # G+12681  U+BCD4  uniBCD4,  HANGUL SYLLABLE-BCD4
        '48341' => '2046'             , # G+2046  U+BCD5  uniBCD5,  HANGUL SYLLABLE-BCD5
        '48342' => '12682'            , # G+12682  U+BCD6  uniBCD6,  HANGUL SYLLABLE-BCD6
        '48343' => '12683'            , # G+12683  U+BCD7  uniBCD7,  HANGUL SYLLABLE-BCD7
        '48344' => '2047'             , # G+2047  U+BCD8  uniBCD8,  HANGUL SYLLABLE-BCD8
        '48345' => '12684'            , # G+12684  U+BCD9  uniBCD9,  HANGUL SYLLABLE-BCD9
        '48346' => '12685'            , # G+12685  U+BCDA  uniBCDA,  HANGUL SYLLABLE-BCDA
        '48347' => '12686'            , # G+12686  U+BCDB  uniBCDB,  HANGUL SYLLABLE-BCDB
        '48348' => '2048'             , # G+2048  U+BCDC  uniBCDC,  HANGUL SYLLABLE-BCDC
        '48349' => '12687'            , # G+12687  U+BCDD  uniBCDD,  HANGUL SYLLABLE-BCDD
        '48350' => '12688'            , # G+12688  U+BCDE  uniBCDE,  HANGUL SYLLABLE-BCDE
        '48351' => '12689'            , # G+12689  U+BCDF  uniBCDF,  HANGUL SYLLABLE-BCDF
        '48352' => '12690'            , # G+12690  U+BCE0  uniBCE0,  HANGUL SYLLABLE-BCE0
        '48353' => '12691'            , # G+12691  U+BCE1  uniBCE1,  HANGUL SYLLABLE-BCE1
        '48354' => '12692'            , # G+12692  U+BCE2  uniBCE2,  HANGUL SYLLABLE-BCE2
        '48355' => '12693'            , # G+12693  U+BCE3  uniBCE3,  HANGUL SYLLABLE-BCE3
        '48356' => '12694'            , # G+12694  U+BCE4  uniBCE4,  HANGUL SYLLABLE-BCE4
        '48357' => '12695'            , # G+12695  U+BCE5  uniBCE5,  HANGUL SYLLABLE-BCE5
        '48358' => '12696'            , # G+12696  U+BCE6  uniBCE6,  HANGUL SYLLABLE-BCE6
        '48359' => '12697'            , # G+12697  U+BCE7  uniBCE7,  HANGUL SYLLABLE-BCE7
        '48360' => '12698'            , # G+12698  U+BCE8  uniBCE8,  HANGUL SYLLABLE-BCE8
        '48361' => '12699'            , # G+12699  U+BCE9  uniBCE9,  HANGUL SYLLABLE-BCE9
        '48362' => '12700'            , # G+12700  U+BCEA  uniBCEA,  HANGUL SYLLABLE-BCEA
        '48363' => '12701'            , # G+12701  U+BCEB  uniBCEB,  HANGUL SYLLABLE-BCEB
        '48364' => '12702'            , # G+12702  U+BCEC  uniBCEC,  HANGUL SYLLABLE-BCEC
        '48365' => '12703'            , # G+12703  U+BCED  uniBCED,  HANGUL SYLLABLE-BCED
        '48366' => '12704'            , # G+12704  U+BCEE  uniBCEE,  HANGUL SYLLABLE-BCEE
        '48367' => '12705'            , # G+12705  U+BCEF  uniBCEF,  HANGUL SYLLABLE-BCEF
        '48368' => '12706'            , # G+12706  U+BCF0  uniBCF0,  HANGUL SYLLABLE-BCF0
        '48369' => '12707'            , # G+12707  U+BCF1  uniBCF1,  HANGUL SYLLABLE-BCF1
        '48370' => '12708'            , # G+12708  U+BCF2  uniBCF2,  HANGUL SYLLABLE-BCF2
        '48371' => '12709'            , # G+12709  U+BCF3  uniBCF3,  HANGUL SYLLABLE-BCF3
        '48372' => '2049'             , # G+2049  U+BCF4  uniBCF4,  HANGUL SYLLABLE-BCF4
        '48373' => '2050'             , # G+2050  U+BCF5  uniBCF5,  HANGUL SYLLABLE-BCF5
        '48374' => '2051'             , # G+2051  U+BCF6  uniBCF6,  HANGUL SYLLABLE-BCF6
        '48375' => '12710'            , # G+12710  U+BCF7  uniBCF7,  HANGUL SYLLABLE-BCF7
        '48376' => '2052'             , # G+2052  U+BCF8  uniBCF8,  HANGUL SYLLABLE-BCF8
        '48377' => '12711'            , # G+12711  U+BCF9  uniBCF9,  HANGUL SYLLABLE-BCF9
        '48378' => '12712'            , # G+12712  U+BCFA  uniBCFA,  HANGUL SYLLABLE-BCFA
        '48379' => '12713'            , # G+12713  U+BCFB  uniBCFB,  HANGUL SYLLABLE-BCFB
        '48380' => '2053'             , # G+2053  U+BCFC  uniBCFC,  HANGUL SYLLABLE-BCFC
        '48381' => '12714'            , # G+12714  U+BCFD  uniBCFD,  HANGUL SYLLABLE-BCFD
        '48382' => '12715'            , # G+12715  U+BCFE  uniBCFE,  HANGUL SYLLABLE-BCFE
        '48383' => '12716'            , # G+12716  U+BCFF  uniBCFF,  HANGUL SYLLABLE-BCFF
        '48384' => '12717'            , # G+12717  U+BD00  uniBD00,  HANGUL SYLLABLE-BD00
        '48385' => '12718'            , # G+12718  U+BD01  uniBD01,  HANGUL SYLLABLE-BD01
        '48386' => '12719'            , # G+12719  U+BD02  uniBD02,  HANGUL SYLLABLE-BD02
        '48387' => '12720'            , # G+12720  U+BD03  uniBD03,  HANGUL SYLLABLE-BD03
        '48388' => '2054'             , # G+2054  U+BD04  uniBD04,  HANGUL SYLLABLE-BD04
        '48389' => '2055'             , # G+2055  U+BD05  uniBD05,  HANGUL SYLLABLE-BD05
        '48390' => '12721'            , # G+12721  U+BD06  uniBD06,  HANGUL SYLLABLE-BD06
        '48391' => '2056'             , # G+2056  U+BD07  uniBD07,  HANGUL SYLLABLE-BD07
        '48392' => '12722'            , # G+12722  U+BD08  uniBD08,  HANGUL SYLLABLE-BD08
        '48393' => '2057'             , # G+2057  U+BD09  uniBD09,  HANGUL SYLLABLE-BD09
        '48394' => '12723'            , # G+12723  U+BD0A  uniBD0A,  HANGUL SYLLABLE-BD0A
        '48395' => '12724'            , # G+12724  U+BD0B  uniBD0B,  HANGUL SYLLABLE-BD0B
        '48396' => '12725'            , # G+12725  U+BD0C  uniBD0C,  HANGUL SYLLABLE-BD0C
        '48397' => '12726'            , # G+12726  U+BD0D  uniBD0D,  HANGUL SYLLABLE-BD0D
        '48398' => '12727'            , # G+12727  U+BD0E  uniBD0E,  HANGUL SYLLABLE-BD0E
        '48399' => '12728'            , # G+12728  U+BD0F  uniBD0F,  HANGUL SYLLABLE-BD0F
        '48400' => '2058'             , # G+2058  U+BD10  uniBD10,  HANGUL SYLLABLE-BD10
        '48401' => '12729'            , # G+12729  U+BD11  uniBD11,  HANGUL SYLLABLE-BD11
        '48402' => '12730'            , # G+12730  U+BD12  uniBD12,  HANGUL SYLLABLE-BD12
        '48403' => '12731'            , # G+12731  U+BD13  uniBD13,  HANGUL SYLLABLE-BD13
        '48404' => '2059'             , # G+2059  U+BD14  uniBD14,  HANGUL SYLLABLE-BD14
        '48405' => '12732'            , # G+12732  U+BD15  uniBD15,  HANGUL SYLLABLE-BD15
        '48406' => '12733'            , # G+12733  U+BD16  uniBD16,  HANGUL SYLLABLE-BD16
        '48407' => '12734'            , # G+12734  U+BD17  uniBD17,  HANGUL SYLLABLE-BD17
        '48408' => '12735'            , # G+12735  U+BD18  uniBD18,  HANGUL SYLLABLE-BD18
        '48409' => '12736'            , # G+12736  U+BD19  uniBD19,  HANGUL SYLLABLE-BD19
        '48410' => '12737'            , # G+12737  U+BD1A  uniBD1A,  HANGUL SYLLABLE-BD1A
        '48411' => '12738'            , # G+12738  U+BD1B  uniBD1B,  HANGUL SYLLABLE-BD1B
        '48412' => '12739'            , # G+12739  U+BD1C  uniBD1C,  HANGUL SYLLABLE-BD1C
        '48413' => '12740'            , # G+12740  U+BD1D  uniBD1D,  HANGUL SYLLABLE-BD1D
        '48414' => '12741'            , # G+12741  U+BD1E  uniBD1E,  HANGUL SYLLABLE-BD1E
        '48415' => '12742'            , # G+12742  U+BD1F  uniBD1F,  HANGUL SYLLABLE-BD1F
        '48416' => '12743'            , # G+12743  U+BD20  uniBD20,  HANGUL SYLLABLE-BD20
        '48417' => '12744'            , # G+12744  U+BD21  uniBD21,  HANGUL SYLLABLE-BD21
        '48418' => '12745'            , # G+12745  U+BD22  uniBD22,  HANGUL SYLLABLE-BD22
        '48419' => '12746'            , # G+12746  U+BD23  uniBD23,  HANGUL SYLLABLE-BD23
        '48420' => '2060'             , # G+2060  U+BD24  uniBD24,  HANGUL SYLLABLE-BD24
        '48421' => '12747'            , # G+12747  U+BD25  uniBD25,  HANGUL SYLLABLE-BD25
        '48422' => '12748'            , # G+12748  U+BD26  uniBD26,  HANGUL SYLLABLE-BD26
        '48423' => '12749'            , # G+12749  U+BD27  uniBD27,  HANGUL SYLLABLE-BD27
        '48424' => '12750'            , # G+12750  U+BD28  uniBD28,  HANGUL SYLLABLE-BD28
        '48425' => '12751'            , # G+12751  U+BD29  uniBD29,  HANGUL SYLLABLE-BD29
        '48426' => '12752'            , # G+12752  U+BD2A  uniBD2A,  HANGUL SYLLABLE-BD2A
        '48427' => '12753'            , # G+12753  U+BD2B  uniBD2B,  HANGUL SYLLABLE-BD2B
        '48428' => '2061'             , # G+2061  U+BD2C  uniBD2C,  HANGUL SYLLABLE-BD2C
        '48429' => '12754'            , # G+12754  U+BD2D  uniBD2D,  HANGUL SYLLABLE-BD2D
        '48430' => '12755'            , # G+12755  U+BD2E  uniBD2E,  HANGUL SYLLABLE-BD2E
        '48431' => '12756'            , # G+12756  U+BD2F  uniBD2F,  HANGUL SYLLABLE-BD2F
        '48432' => '12757'            , # G+12757  U+BD30  uniBD30,  HANGUL SYLLABLE-BD30
        '48433' => '12758'            , # G+12758  U+BD31  uniBD31,  HANGUL SYLLABLE-BD31
        '48434' => '12759'            , # G+12759  U+BD32  uniBD32,  HANGUL SYLLABLE-BD32
        '48435' => '12760'            , # G+12760  U+BD33  uniBD33,  HANGUL SYLLABLE-BD33
        '48436' => '12761'            , # G+12761  U+BD34  uniBD34,  HANGUL SYLLABLE-BD34
        '48437' => '12762'            , # G+12762  U+BD35  uniBD35,  HANGUL SYLLABLE-BD35
        '48438' => '12763'            , # G+12763  U+BD36  uniBD36,  HANGUL SYLLABLE-BD36
        '48439' => '12764'            , # G+12764  U+BD37  uniBD37,  HANGUL SYLLABLE-BD37
        '48440' => '12765'            , # G+12765  U+BD38  uniBD38,  HANGUL SYLLABLE-BD38
        '48441' => '12766'            , # G+12766  U+BD39  uniBD39,  HANGUL SYLLABLE-BD39
        '48442' => '12767'            , # G+12767  U+BD3A  uniBD3A,  HANGUL SYLLABLE-BD3A
        '48443' => '12768'            , # G+12768  U+BD3B  uniBD3B,  HANGUL SYLLABLE-BD3B
        '48444' => '12769'            , # G+12769  U+BD3C  uniBD3C,  HANGUL SYLLABLE-BD3C
        '48445' => '12770'            , # G+12770  U+BD3D  uniBD3D,  HANGUL SYLLABLE-BD3D
        '48446' => '12771'            , # G+12771  U+BD3E  uniBD3E,  HANGUL SYLLABLE-BD3E
        '48447' => '12772'            , # G+12772  U+BD3F  uniBD3F,  HANGUL SYLLABLE-BD3F
        '48448' => '2062'             , # G+2062  U+BD40  uniBD40,  HANGUL SYLLABLE-BD40
        '48449' => '12773'            , # G+12773  U+BD41  uniBD41,  HANGUL SYLLABLE-BD41
        '48450' => '12774'            , # G+12774  U+BD42  uniBD42,  HANGUL SYLLABLE-BD42
        '48451' => '12775'            , # G+12775  U+BD43  uniBD43,  HANGUL SYLLABLE-BD43
        '48452' => '12776'            , # G+12776  U+BD44  uniBD44,  HANGUL SYLLABLE-BD44
        '48453' => '12777'            , # G+12777  U+BD45  uniBD45,  HANGUL SYLLABLE-BD45
        '48454' => '12778'            , # G+12778  U+BD46  uniBD46,  HANGUL SYLLABLE-BD46
        '48455' => '12779'            , # G+12779  U+BD47  uniBD47,  HANGUL SYLLABLE-BD47
        '48456' => '2063'             , # G+2063  U+BD48  uniBD48,  HANGUL SYLLABLE-BD48
        '48457' => '2064'             , # G+2064  U+BD49  uniBD49,  HANGUL SYLLABLE-BD49
        '48458' => '12780'            , # G+12780  U+BD4A  uniBD4A,  HANGUL SYLLABLE-BD4A
        '48459' => '12781'            , # G+12781  U+BD4B  uniBD4B,  HANGUL SYLLABLE-BD4B
        '48460' => '2065'             , # G+2065  U+BD4C  uniBD4C,  HANGUL SYLLABLE-BD4C
        '48461' => '12782'            , # G+12782  U+BD4D  uniBD4D,  HANGUL SYLLABLE-BD4D
        '48462' => '12783'            , # G+12783  U+BD4E  uniBD4E,  HANGUL SYLLABLE-BD4E
        '48463' => '12784'            , # G+12784  U+BD4F  uniBD4F,  HANGUL SYLLABLE-BD4F
        '48464' => '2066'             , # G+2066  U+BD50  uniBD50,  HANGUL SYLLABLE-BD50
        '48465' => '12785'            , # G+12785  U+BD51  uniBD51,  HANGUL SYLLABLE-BD51
        '48466' => '12786'            , # G+12786  U+BD52  uniBD52,  HANGUL SYLLABLE-BD52
        '48467' => '12787'            , # G+12787  U+BD53  uniBD53,  HANGUL SYLLABLE-BD53
        '48468' => '12788'            , # G+12788  U+BD54  uniBD54,  HANGUL SYLLABLE-BD54
        '48469' => '12789'            , # G+12789  U+BD55  uniBD55,  HANGUL SYLLABLE-BD55
        '48470' => '12790'            , # G+12790  U+BD56  uniBD56,  HANGUL SYLLABLE-BD56
        '48471' => '12791'            , # G+12791  U+BD57  uniBD57,  HANGUL SYLLABLE-BD57
        '48472' => '2067'             , # G+2067  U+BD58  uniBD58,  HANGUL SYLLABLE-BD58
        '48473' => '2068'             , # G+2068  U+BD59  uniBD59,  HANGUL SYLLABLE-BD59
        '48474' => '12792'            , # G+12792  U+BD5A  uniBD5A,  HANGUL SYLLABLE-BD5A
        '48475' => '12793'            , # G+12793  U+BD5B  uniBD5B,  HANGUL SYLLABLE-BD5B
        '48476' => '12794'            , # G+12794  U+BD5C  uniBD5C,  HANGUL SYLLABLE-BD5C
        '48477' => '12795'            , # G+12795  U+BD5D  uniBD5D,  HANGUL SYLLABLE-BD5D
        '48478' => '12796'            , # G+12796  U+BD5E  uniBD5E,  HANGUL SYLLABLE-BD5E
        '48479' => '12797'            , # G+12797  U+BD5F  uniBD5F,  HANGUL SYLLABLE-BD5F
        '48480' => '12798'            , # G+12798  U+BD60  uniBD60,  HANGUL SYLLABLE-BD60
        '48481' => '12799'            , # G+12799  U+BD61  uniBD61,  HANGUL SYLLABLE-BD61
        '48482' => '12800'            , # G+12800  U+BD62  uniBD62,  HANGUL SYLLABLE-BD62
        '48483' => '12801'            , # G+12801  U+BD63  uniBD63,  HANGUL SYLLABLE-BD63
        '48484' => '2069'             , # G+2069  U+BD64  uniBD64,  HANGUL SYLLABLE-BD64
        '48485' => '12802'            , # G+12802  U+BD65  uniBD65,  HANGUL SYLLABLE-BD65
        '48486' => '12803'            , # G+12803  U+BD66  uniBD66,  HANGUL SYLLABLE-BD66
        '48487' => '12804'            , # G+12804  U+BD67  uniBD67,  HANGUL SYLLABLE-BD67
        '48488' => '2070'             , # G+2070  U+BD68  uniBD68,  HANGUL SYLLABLE-BD68
        '48489' => '12805'            , # G+12805  U+BD69  uniBD69,  HANGUL SYLLABLE-BD69
        '48490' => '12806'            , # G+12806  U+BD6A  uniBD6A,  HANGUL SYLLABLE-BD6A
        '48491' => '12807'            , # G+12807  U+BD6B  uniBD6B,  HANGUL SYLLABLE-BD6B
        '48492' => '12808'            , # G+12808  U+BD6C  uniBD6C,  HANGUL SYLLABLE-BD6C
        '48493' => '12809'            , # G+12809  U+BD6D  uniBD6D,  HANGUL SYLLABLE-BD6D
        '48494' => '12810'            , # G+12810  U+BD6E  uniBD6E,  HANGUL SYLLABLE-BD6E
        '48495' => '12811'            , # G+12811  U+BD6F  uniBD6F,  HANGUL SYLLABLE-BD6F
        '48496' => '12812'            , # G+12812  U+BD70  uniBD70,  HANGUL SYLLABLE-BD70
        '48497' => '12813'            , # G+12813  U+BD71  uniBD71,  HANGUL SYLLABLE-BD71
        '48498' => '12814'            , # G+12814  U+BD72  uniBD72,  HANGUL SYLLABLE-BD72
        '48499' => '12815'            , # G+12815  U+BD73  uniBD73,  HANGUL SYLLABLE-BD73
        '48500' => '12816'            , # G+12816  U+BD74  uniBD74,  HANGUL SYLLABLE-BD74
        '48501' => '12817'            , # G+12817  U+BD75  uniBD75,  HANGUL SYLLABLE-BD75
        '48502' => '12818'            , # G+12818  U+BD76  uniBD76,  HANGUL SYLLABLE-BD76
        '48503' => '12819'            , # G+12819  U+BD77  uniBD77,  HANGUL SYLLABLE-BD77
        '48504' => '12820'            , # G+12820  U+BD78  uniBD78,  HANGUL SYLLABLE-BD78
        '48505' => '12821'            , # G+12821  U+BD79  uniBD79,  HANGUL SYLLABLE-BD79
        '48506' => '12822'            , # G+12822  U+BD7A  uniBD7A,  HANGUL SYLLABLE-BD7A
        '48507' => '12823'            , # G+12823  U+BD7B  uniBD7B,  HANGUL SYLLABLE-BD7B
        '48508' => '12824'            , # G+12824  U+BD7C  uniBD7C,  HANGUL SYLLABLE-BD7C
        '48509' => '12825'            , # G+12825  U+BD7D  uniBD7D,  HANGUL SYLLABLE-BD7D
        '48510' => '12826'            , # G+12826  U+BD7E  uniBD7E,  HANGUL SYLLABLE-BD7E
        '48511' => '12827'            , # G+12827  U+BD7F  uniBD7F,  HANGUL SYLLABLE-BD7F
        '48512' => '2071'             , # G+2071  U+BD80  uniBD80,  HANGUL SYLLABLE-BD80
        '48513' => '2072'             , # G+2072  U+BD81  uniBD81,  HANGUL SYLLABLE-BD81
        '48514' => '12828'            , # G+12828  U+BD82  uniBD82,  HANGUL SYLLABLE-BD82
        '48515' => '12829'            , # G+12829  U+BD83  uniBD83,  HANGUL SYLLABLE-BD83
        '48516' => '2073'             , # G+2073  U+BD84  uniBD84,  HANGUL SYLLABLE-BD84
        '48517' => '12830'            , # G+12830  U+BD85  uniBD85,  HANGUL SYLLABLE-BD85
        '48518' => '12831'            , # G+12831  U+BD86  uniBD86,  HANGUL SYLLABLE-BD86
        '48519' => '2074'             , # G+2074  U+BD87  uniBD87,  HANGUL SYLLABLE-BD87
        '48520' => '2075'             , # G+2075  U+BD88  uniBD88,  HANGUL SYLLABLE-BD88
        '48521' => '2076'             , # G+2076  U+BD89  uniBD89,  HANGUL SYLLABLE-BD89
        '48522' => '2077'             , # G+2077  U+BD8A  uniBD8A,  HANGUL SYLLABLE-BD8A
        '48523' => '12832'            , # G+12832  U+BD8B  uniBD8B,  HANGUL SYLLABLE-BD8B
        '48524' => '12833'            , # G+12833  U+BD8C  uniBD8C,  HANGUL SYLLABLE-BD8C
        '48525' => '12834'            , # G+12834  U+BD8D  uniBD8D,  HANGUL SYLLABLE-BD8D
        '48526' => '12835'            , # G+12835  U+BD8E  uniBD8E,  HANGUL SYLLABLE-BD8E
        '48527' => '12836'            , # G+12836  U+BD8F  uniBD8F,  HANGUL SYLLABLE-BD8F
        '48528' => '2078'             , # G+2078  U+BD90  uniBD90,  HANGUL SYLLABLE-BD90
        '48529' => '2079'             , # G+2079  U+BD91  uniBD91,  HANGUL SYLLABLE-BD91
        '48530' => '12837'            , # G+12837  U+BD92  uniBD92,  HANGUL SYLLABLE-BD92
        '48531' => '2080'             , # G+2080  U+BD93  uniBD93,  HANGUL SYLLABLE-BD93
        '48532' => '12838'            , # G+12838  U+BD94  uniBD94,  HANGUL SYLLABLE-BD94
        '48533' => '2081'             , # G+2081  U+BD95  uniBD95,  HANGUL SYLLABLE-BD95
        '48534' => '12839'            , # G+12839  U+BD96  uniBD96,  HANGUL SYLLABLE-BD96
        '48535' => '12840'            , # G+12840  U+BD97  uniBD97,  HANGUL SYLLABLE-BD97
        '48536' => '12841'            , # G+12841  U+BD98  uniBD98,  HANGUL SYLLABLE-BD98
        '48537' => '2082'             , # G+2082  U+BD99  uniBD99,  HANGUL SYLLABLE-BD99
        '48538' => '2083'             , # G+2083  U+BD9A  uniBD9A,  HANGUL SYLLABLE-BD9A
        '48539' => '12842'            , # G+12842  U+BD9B  uniBD9B,  HANGUL SYLLABLE-BD9B
        '48540' => '2084'             , # G+2084  U+BD9C  uniBD9C,  HANGUL SYLLABLE-BD9C
        '48541' => '12843'            , # G+12843  U+BD9D  uniBD9D,  HANGUL SYLLABLE-BD9D
        '48542' => '12844'            , # G+12844  U+BD9E  uniBD9E,  HANGUL SYLLABLE-BD9E
        '48543' => '12845'            , # G+12845  U+BD9F  uniBD9F,  HANGUL SYLLABLE-BD9F
        '48544' => '12846'            , # G+12846  U+BDA0  uniBDA0,  HANGUL SYLLABLE-BDA0
        '48545' => '12847'            , # G+12847  U+BDA1  uniBDA1,  HANGUL SYLLABLE-BDA1
        '48546' => '12848'            , # G+12848  U+BDA2  uniBDA2,  HANGUL SYLLABLE-BDA2
        '48547' => '12849'            , # G+12849  U+BDA3  uniBDA3,  HANGUL SYLLABLE-BDA3
        '48548' => '2085'             , # G+2085  U+BDA4  uniBDA4,  HANGUL SYLLABLE-BDA4
        '48549' => '12850'            , # G+12850  U+BDA5  uniBDA5,  HANGUL SYLLABLE-BDA5
        '48550' => '12851'            , # G+12851  U+BDA6  uniBDA6,  HANGUL SYLLABLE-BDA6
        '48551' => '12852'            , # G+12852  U+BDA7  uniBDA7,  HANGUL SYLLABLE-BDA7
        '48552' => '12853'            , # G+12853  U+BDA8  uniBDA8,  HANGUL SYLLABLE-BDA8
        '48553' => '12854'            , # G+12854  U+BDA9  uniBDA9,  HANGUL SYLLABLE-BDA9
        '48554' => '12855'            , # G+12855  U+BDAA  uniBDAA,  HANGUL SYLLABLE-BDAA
        '48555' => '12856'            , # G+12856  U+BDAB  uniBDAB,  HANGUL SYLLABLE-BDAB
        '48556' => '12857'            , # G+12857  U+BDAC  uniBDAC,  HANGUL SYLLABLE-BDAC
        '48557' => '12858'            , # G+12858  U+BDAD  uniBDAD,  HANGUL SYLLABLE-BDAD
        '48558' => '12859'            , # G+12859  U+BDAE  uniBDAE,  HANGUL SYLLABLE-BDAE
        '48559' => '12860'            , # G+12860  U+BDAF  uniBDAF,  HANGUL SYLLABLE-BDAF
        '48560' => '2086'             , # G+2086  U+BDB0  uniBDB0,  HANGUL SYLLABLE-BDB0
        '48561' => '12861'            , # G+12861  U+BDB1  uniBDB1,  HANGUL SYLLABLE-BDB1
        '48562' => '12862'            , # G+12862  U+BDB2  uniBDB2,  HANGUL SYLLABLE-BDB2
        '48563' => '12863'            , # G+12863  U+BDB3  uniBDB3,  HANGUL SYLLABLE-BDB3
        '48564' => '12864'            , # G+12864  U+BDB4  uniBDB4,  HANGUL SYLLABLE-BDB4
        '48565' => '12865'            , # G+12865  U+BDB5  uniBDB5,  HANGUL SYLLABLE-BDB5
        '48566' => '12866'            , # G+12866  U+BDB6  uniBDB6,  HANGUL SYLLABLE-BDB6
        '48567' => '12867'            , # G+12867  U+BDB7  uniBDB7,  HANGUL SYLLABLE-BDB7
        '48568' => '2087'             , # G+2087  U+BDB8  uniBDB8,  HANGUL SYLLABLE-BDB8
        '48569' => '12868'            , # G+12868  U+BDB9  uniBDB9,  HANGUL SYLLABLE-BDB9
        '48570' => '12869'            , # G+12869  U+BDBA  uniBDBA,  HANGUL SYLLABLE-BDBA
        '48571' => '12870'            , # G+12870  U+BDBB  uniBDBB,  HANGUL SYLLABLE-BDBB
        '48572' => '12871'            , # G+12871  U+BDBC  uniBDBC,  HANGUL SYLLABLE-BDBC
        '48573' => '12872'            , # G+12872  U+BDBD  uniBDBD,  HANGUL SYLLABLE-BDBD
        '48574' => '12873'            , # G+12873  U+BDBE  uniBDBE,  HANGUL SYLLABLE-BDBE
        '48575' => '12874'            , # G+12874  U+BDBF  uniBDBF,  HANGUL SYLLABLE-BDBF
        '48576' => '12875'            , # G+12875  U+BDC0  uniBDC0,  HANGUL SYLLABLE-BDC0
        '48577' => '12876'            , # G+12876  U+BDC1  uniBDC1,  HANGUL SYLLABLE-BDC1
        '48578' => '12877'            , # G+12877  U+BDC2  uniBDC2,  HANGUL SYLLABLE-BDC2
        '48579' => '12878'            , # G+12878  U+BDC3  uniBDC3,  HANGUL SYLLABLE-BDC3
        '48580' => '12879'            , # G+12879  U+BDC4  uniBDC4,  HANGUL SYLLABLE-BDC4
        '48581' => '12880'            , # G+12880  U+BDC5  uniBDC5,  HANGUL SYLLABLE-BDC5
        '48582' => '12881'            , # G+12881  U+BDC6  uniBDC6,  HANGUL SYLLABLE-BDC6
        '48583' => '12882'            , # G+12882  U+BDC7  uniBDC7,  HANGUL SYLLABLE-BDC7
        '48584' => '12883'            , # G+12883  U+BDC8  uniBDC8,  HANGUL SYLLABLE-BDC8
        '48585' => '12884'            , # G+12884  U+BDC9  uniBDC9,  HANGUL SYLLABLE-BDC9
        '48586' => '12885'            , # G+12885  U+BDCA  uniBDCA,  HANGUL SYLLABLE-BDCA
        '48587' => '12886'            , # G+12886  U+BDCB  uniBDCB,  HANGUL SYLLABLE-BDCB
        '48588' => '12887'            , # G+12887  U+BDCC  uniBDCC,  HANGUL SYLLABLE-BDCC
        '48589' => '12888'            , # G+12888  U+BDCD  uniBDCD,  HANGUL SYLLABLE-BDCD
        '48590' => '12889'            , # G+12889  U+BDCE  uniBDCE,  HANGUL SYLLABLE-BDCE
        '48591' => '12890'            , # G+12890  U+BDCF  uniBDCF,  HANGUL SYLLABLE-BDCF
        '48592' => '12891'            , # G+12891  U+BDD0  uniBDD0,  HANGUL SYLLABLE-BDD0
        '48593' => '12892'            , # G+12892  U+BDD1  uniBDD1,  HANGUL SYLLABLE-BDD1
        '48594' => '12893'            , # G+12893  U+BDD2  uniBDD2,  HANGUL SYLLABLE-BDD2
        '48595' => '12894'            , # G+12894  U+BDD3  uniBDD3,  HANGUL SYLLABLE-BDD3
        '48596' => '2088'             , # G+2088  U+BDD4  uniBDD4,  HANGUL SYLLABLE-BDD4
        '48597' => '2089'             , # G+2089  U+BDD5  uniBDD5,  HANGUL SYLLABLE-BDD5
        '48598' => '12895'            , # G+12895  U+BDD6  uniBDD6,  HANGUL SYLLABLE-BDD6
        '48599' => '12896'            , # G+12896  U+BDD7  uniBDD7,  HANGUL SYLLABLE-BDD7
        '48600' => '2090'             , # G+2090  U+BDD8  uniBDD8,  HANGUL SYLLABLE-BDD8
        '48601' => '12897'            , # G+12897  U+BDD9  uniBDD9,  HANGUL SYLLABLE-BDD9
        '48602' => '12898'            , # G+12898  U+BDDA  uniBDDA,  HANGUL SYLLABLE-BDDA
        '48603' => '12899'            , # G+12899  U+BDDB  uniBDDB,  HANGUL SYLLABLE-BDDB
        '48604' => '2091'             , # G+2091  U+BDDC  uniBDDC,  HANGUL SYLLABLE-BDDC
        '48605' => '12900'            , # G+12900  U+BDDD  uniBDDD,  HANGUL SYLLABLE-BDDD
        '48606' => '12901'            , # G+12901  U+BDDE  uniBDDE,  HANGUL SYLLABLE-BDDE
        '48607' => '12902'            , # G+12902  U+BDDF  uniBDDF,  HANGUL SYLLABLE-BDDF
        '48608' => '12903'            , # G+12903  U+BDE0  uniBDE0,  HANGUL SYLLABLE-BDE0
        '48609' => '12904'            , # G+12904  U+BDE1  uniBDE1,  HANGUL SYLLABLE-BDE1
        '48610' => '12905'            , # G+12905  U+BDE2  uniBDE2,  HANGUL SYLLABLE-BDE2
        '48611' => '12906'            , # G+12906  U+BDE3  uniBDE3,  HANGUL SYLLABLE-BDE3
        '48612' => '12907'            , # G+12907  U+BDE4  uniBDE4,  HANGUL SYLLABLE-BDE4
        '48613' => '12908'            , # G+12908  U+BDE5  uniBDE5,  HANGUL SYLLABLE-BDE5
        '48614' => '12909'            , # G+12909  U+BDE6  uniBDE6,  HANGUL SYLLABLE-BDE6
        '48615' => '12910'            , # G+12910  U+BDE7  uniBDE7,  HANGUL SYLLABLE-BDE7
        '48616' => '12911'            , # G+12911  U+BDE8  uniBDE8,  HANGUL SYLLABLE-BDE8
        '48617' => '2092'             , # G+2092  U+BDE9  uniBDE9,  HANGUL SYLLABLE-BDE9
        '48618' => '12912'            , # G+12912  U+BDEA  uniBDEA,  HANGUL SYLLABLE-BDEA
        '48619' => '12913'            , # G+12913  U+BDEB  uniBDEB,  HANGUL SYLLABLE-BDEB
        '48620' => '12914'            , # G+12914  U+BDEC  uniBDEC,  HANGUL SYLLABLE-BDEC
        '48621' => '12915'            , # G+12915  U+BDED  uniBDED,  HANGUL SYLLABLE-BDED
        '48622' => '12916'            , # G+12916  U+BDEE  uniBDEE,  HANGUL SYLLABLE-BDEE
        '48623' => '12917'            , # G+12917  U+BDEF  uniBDEF,  HANGUL SYLLABLE-BDEF
        '48624' => '2093'             , # G+2093  U+BDF0  uniBDF0,  HANGUL SYLLABLE-BDF0
        '48625' => '12918'            , # G+12918  U+BDF1  uniBDF1,  HANGUL SYLLABLE-BDF1
        '48626' => '12919'            , # G+12919  U+BDF2  uniBDF2,  HANGUL SYLLABLE-BDF2
        '48627' => '12920'            , # G+12920  U+BDF3  uniBDF3,  HANGUL SYLLABLE-BDF3
        '48628' => '2094'             , # G+2094  U+BDF4  uniBDF4,  HANGUL SYLLABLE-BDF4
        '48629' => '12921'            , # G+12921  U+BDF5  uniBDF5,  HANGUL SYLLABLE-BDF5
        '48630' => '12922'            , # G+12922  U+BDF6  uniBDF6,  HANGUL SYLLABLE-BDF6
        '48631' => '12923'            , # G+12923  U+BDF7  uniBDF7,  HANGUL SYLLABLE-BDF7
        '48632' => '2095'             , # G+2095  U+BDF8  uniBDF8,  HANGUL SYLLABLE-BDF8
        '48633' => '12924'            , # G+12924  U+BDF9  uniBDF9,  HANGUL SYLLABLE-BDF9
        '48634' => '12925'            , # G+12925  U+BDFA  uniBDFA,  HANGUL SYLLABLE-BDFA
        '48635' => '12926'            , # G+12926  U+BDFB  uniBDFB,  HANGUL SYLLABLE-BDFB
        '48636' => '12927'            , # G+12927  U+BDFC  uniBDFC,  HANGUL SYLLABLE-BDFC
        '48637' => '12928'            , # G+12928  U+BDFD  uniBDFD,  HANGUL SYLLABLE-BDFD
        '48638' => '12929'            , # G+12929  U+BDFE  uniBDFE,  HANGUL SYLLABLE-BDFE
        '48639' => '12930'            , # G+12930  U+BDFF  uniBDFF,  HANGUL SYLLABLE-BDFF
        '48640' => '2096'             , # G+2096  U+BE00  uniBE00,  HANGUL SYLLABLE-BE00
        '48641' => '12931'            , # G+12931  U+BE01  uniBE01,  HANGUL SYLLABLE-BE01
        '48642' => '12932'            , # G+12932  U+BE02  uniBE02,  HANGUL SYLLABLE-BE02
        '48643' => '2097'             , # G+2097  U+BE03  uniBE03,  HANGUL SYLLABLE-BE03
        '48644' => '12933'            , # G+12933  U+BE04  uniBE04,  HANGUL SYLLABLE-BE04
        '48645' => '2098'             , # G+2098  U+BE05  uniBE05,  HANGUL SYLLABLE-BE05
        '48646' => '12934'            , # G+12934  U+BE06  uniBE06,  HANGUL SYLLABLE-BE06
        '48647' => '12935'            , # G+12935  U+BE07  uniBE07,  HANGUL SYLLABLE-BE07
        '48648' => '12936'            , # G+12936  U+BE08  uniBE08,  HANGUL SYLLABLE-BE08
        '48649' => '12937'            , # G+12937  U+BE09  uniBE09,  HANGUL SYLLABLE-BE09
        '48650' => '12938'            , # G+12938  U+BE0A  uniBE0A,  HANGUL SYLLABLE-BE0A
        '48651' => '12939'            , # G+12939  U+BE0B  uniBE0B,  HANGUL SYLLABLE-BE0B
        '48652' => '2099'             , # G+2099  U+BE0C  uniBE0C,  HANGUL SYLLABLE-BE0C
        '48653' => '2100'             , # G+2100  U+BE0D  uniBE0D,  HANGUL SYLLABLE-BE0D
        '48654' => '12940'            , # G+12940  U+BE0E  uniBE0E,  HANGUL SYLLABLE-BE0E
        '48655' => '12941'            , # G+12941  U+BE0F  uniBE0F,  HANGUL SYLLABLE-BE0F
        '48656' => '2101'             , # G+2101  U+BE10  uniBE10,  HANGUL SYLLABLE-BE10
        '48657' => '12942'            , # G+12942  U+BE11  uniBE11,  HANGUL SYLLABLE-BE11
        '48658' => '12943'            , # G+12943  U+BE12  uniBE12,  HANGUL SYLLABLE-BE12
        '48659' => '12944'            , # G+12944  U+BE13  uniBE13,  HANGUL SYLLABLE-BE13
        '48660' => '2102'             , # G+2102  U+BE14  uniBE14,  HANGUL SYLLABLE-BE14
        '48661' => '12945'            , # G+12945  U+BE15  uniBE15,  HANGUL SYLLABLE-BE15
        '48662' => '12946'            , # G+12946  U+BE16  uniBE16,  HANGUL SYLLABLE-BE16
        '48663' => '12947'            , # G+12947  U+BE17  uniBE17,  HANGUL SYLLABLE-BE17
        '48664' => '12948'            , # G+12948  U+BE18  uniBE18,  HANGUL SYLLABLE-BE18
        '48665' => '12949'            , # G+12949  U+BE19  uniBE19,  HANGUL SYLLABLE-BE19
        '48666' => '12950'            , # G+12950  U+BE1A  uniBE1A,  HANGUL SYLLABLE-BE1A
        '48667' => '12951'            , # G+12951  U+BE1B  uniBE1B,  HANGUL SYLLABLE-BE1B
        '48668' => '2103'             , # G+2103  U+BE1C  uniBE1C,  HANGUL SYLLABLE-BE1C
        '48669' => '2104'             , # G+2104  U+BE1D  uniBE1D,  HANGUL SYLLABLE-BE1D
        '48670' => '12952'            , # G+12952  U+BE1E  uniBE1E,  HANGUL SYLLABLE-BE1E
        '48671' => '2105'             , # G+2105  U+BE1F  uniBE1F,  HANGUL SYLLABLE-BE1F
        '48672' => '12953'            , # G+12953  U+BE20  uniBE20,  HANGUL SYLLABLE-BE20
        '48673' => '12954'            , # G+12954  U+BE21  uniBE21,  HANGUL SYLLABLE-BE21
        '48674' => '12955'            , # G+12955  U+BE22  uniBE22,  HANGUL SYLLABLE-BE22
        '48675' => '12956'            , # G+12956  U+BE23  uniBE23,  HANGUL SYLLABLE-BE23
        '48676' => '12957'            , # G+12957  U+BE24  uniBE24,  HANGUL SYLLABLE-BE24
        '48677' => '12958'            , # G+12958  U+BE25  uniBE25,  HANGUL SYLLABLE-BE25
        '48678' => '12959'            , # G+12959  U+BE26  uniBE26,  HANGUL SYLLABLE-BE26
        '48679' => '12960'            , # G+12960  U+BE27  uniBE27,  HANGUL SYLLABLE-BE27
        '48680' => '12961'            , # G+12961  U+BE28  uniBE28,  HANGUL SYLLABLE-BE28
        '48681' => '12962'            , # G+12962  U+BE29  uniBE29,  HANGUL SYLLABLE-BE29
        '48682' => '12963'            , # G+12963  U+BE2A  uniBE2A,  HANGUL SYLLABLE-BE2A
        '48683' => '12964'            , # G+12964  U+BE2B  uniBE2B,  HANGUL SYLLABLE-BE2B
        '48684' => '12965'            , # G+12965  U+BE2C  uniBE2C,  HANGUL SYLLABLE-BE2C
        '48685' => '12966'            , # G+12966  U+BE2D  uniBE2D,  HANGUL SYLLABLE-BE2D
        '48686' => '12967'            , # G+12967  U+BE2E  uniBE2E,  HANGUL SYLLABLE-BE2E
        '48687' => '12968'            , # G+12968  U+BE2F  uniBE2F,  HANGUL SYLLABLE-BE2F
        '48688' => '12969'            , # G+12969  U+BE30  uniBE30,  HANGUL SYLLABLE-BE30
        '48689' => '12970'            , # G+12970  U+BE31  uniBE31,  HANGUL SYLLABLE-BE31
        '48690' => '12971'            , # G+12971  U+BE32  uniBE32,  HANGUL SYLLABLE-BE32
        '48691' => '12972'            , # G+12972  U+BE33  uniBE33,  HANGUL SYLLABLE-BE33
        '48692' => '12973'            , # G+12973  U+BE34  uniBE34,  HANGUL SYLLABLE-BE34
        '48693' => '12974'            , # G+12974  U+BE35  uniBE35,  HANGUL SYLLABLE-BE35
        '48694' => '12975'            , # G+12975  U+BE36  uniBE36,  HANGUL SYLLABLE-BE36
        '48695' => '12976'            , # G+12976  U+BE37  uniBE37,  HANGUL SYLLABLE-BE37
        '48696' => '12977'            , # G+12977  U+BE38  uniBE38,  HANGUL SYLLABLE-BE38
        '48697' => '12978'            , # G+12978  U+BE39  uniBE39,  HANGUL SYLLABLE-BE39
        '48698' => '12979'            , # G+12979  U+BE3A  uniBE3A,  HANGUL SYLLABLE-BE3A
        '48699' => '12980'            , # G+12980  U+BE3B  uniBE3B,  HANGUL SYLLABLE-BE3B
        '48700' => '12981'            , # G+12981  U+BE3C  uniBE3C,  HANGUL SYLLABLE-BE3C
        '48701' => '12982'            , # G+12982  U+BE3D  uniBE3D,  HANGUL SYLLABLE-BE3D
        '48702' => '12983'            , # G+12983  U+BE3E  uniBE3E,  HANGUL SYLLABLE-BE3E
        '48703' => '12984'            , # G+12984  U+BE3F  uniBE3F,  HANGUL SYLLABLE-BE3F
        '48704' => '12985'            , # G+12985  U+BE40  uniBE40,  HANGUL SYLLABLE-BE40
        '48705' => '12986'            , # G+12986  U+BE41  uniBE41,  HANGUL SYLLABLE-BE41
        '48706' => '12987'            , # G+12987  U+BE42  uniBE42,  HANGUL SYLLABLE-BE42
        '48707' => '12988'            , # G+12988  U+BE43  uniBE43,  HANGUL SYLLABLE-BE43
        '48708' => '2106'             , # G+2106  U+BE44  uniBE44,  HANGUL SYLLABLE-BE44
        '48709' => '2107'             , # G+2107  U+BE45  uniBE45,  HANGUL SYLLABLE-BE45
        '48710' => '12989'            , # G+12989  U+BE46  uniBE46,  HANGUL SYLLABLE-BE46
        '48711' => '12990'            , # G+12990  U+BE47  uniBE47,  HANGUL SYLLABLE-BE47
        '48712' => '2108'             , # G+2108  U+BE48  uniBE48,  HANGUL SYLLABLE-BE48
        '48713' => '12991'            , # G+12991  U+BE49  uniBE49,  HANGUL SYLLABLE-BE49
        '48714' => '12992'            , # G+12992  U+BE4A  uniBE4A,  HANGUL SYLLABLE-BE4A
        '48715' => '12993'            , # G+12993  U+BE4B  uniBE4B,  HANGUL SYLLABLE-BE4B
        '48716' => '2109'             , # G+2109  U+BE4C  uniBE4C,  HANGUL SYLLABLE-BE4C
        '48717' => '12994'            , # G+12994  U+BE4D  uniBE4D,  HANGUL SYLLABLE-BE4D
        '48718' => '2110'             , # G+2110  U+BE4E  uniBE4E,  HANGUL SYLLABLE-BE4E
        '48719' => '12995'            , # G+12995  U+BE4F  uniBE4F,  HANGUL SYLLABLE-BE4F
        '48720' => '12996'            , # G+12996  U+BE50  uniBE50,  HANGUL SYLLABLE-BE50
        '48721' => '12997'            , # G+12997  U+BE51  uniBE51,  HANGUL SYLLABLE-BE51
        '48722' => '12998'            , # G+12998  U+BE52  uniBE52,  HANGUL SYLLABLE-BE52
        '48723' => '12999'            , # G+12999  U+BE53  uniBE53,  HANGUL SYLLABLE-BE53
        '48724' => '2111'             , # G+2111  U+BE54  uniBE54,  HANGUL SYLLABLE-BE54
        '48725' => '2112'             , # G+2112  U+BE55  uniBE55,  HANGUL SYLLABLE-BE55
        '48726' => '13000'            , # G+13000  U+BE56  uniBE56,  HANGUL SYLLABLE-BE56
        '48727' => '2113'             , # G+2113  U+BE57  uniBE57,  HANGUL SYLLABLE-BE57
        '48728' => '13001'            , # G+13001  U+BE58  uniBE58,  HANGUL SYLLABLE-BE58
        '48729' => '2114'             , # G+2114  U+BE59  uniBE59,  HANGUL SYLLABLE-BE59
        '48730' => '2115'             , # G+2115  U+BE5A  uniBE5A,  HANGUL SYLLABLE-BE5A
        '48731' => '2116'             , # G+2116  U+BE5B  uniBE5B,  HANGUL SYLLABLE-BE5B
        '48732' => '13002'            , # G+13002  U+BE5C  uniBE5C,  HANGUL SYLLABLE-BE5C
        '48733' => '13003'            , # G+13003  U+BE5D  uniBE5D,  HANGUL SYLLABLE-BE5D
        '48734' => '13004'            , # G+13004  U+BE5E  uniBE5E,  HANGUL SYLLABLE-BE5E
        '48735' => '13005'            , # G+13005  U+BE5F  uniBE5F,  HANGUL SYLLABLE-BE5F
        '48736' => '2117'             , # G+2117  U+BE60  uniBE60,  HANGUL SYLLABLE-BE60
        '48737' => '2118'             , # G+2118  U+BE61  uniBE61,  HANGUL SYLLABLE-BE61
        '48738' => '13006'            , # G+13006  U+BE62  uniBE62,  HANGUL SYLLABLE-BE62
        '48739' => '13007'            , # G+13007  U+BE63  uniBE63,  HANGUL SYLLABLE-BE63
        '48740' => '2119'             , # G+2119  U+BE64  uniBE64,  HANGUL SYLLABLE-BE64
        '48741' => '13008'            , # G+13008  U+BE65  uniBE65,  HANGUL SYLLABLE-BE65
        '48742' => '13009'            , # G+13009  U+BE66  uniBE66,  HANGUL SYLLABLE-BE66
        '48743' => '13010'            , # G+13010  U+BE67  uniBE67,  HANGUL SYLLABLE-BE67
        '48744' => '2120'             , # G+2120  U+BE68  uniBE68,  HANGUL SYLLABLE-BE68
        '48745' => '13011'            , # G+13011  U+BE69  uniBE69,  HANGUL SYLLABLE-BE69
        '48746' => '2121'             , # G+2121  U+BE6A  uniBE6A,  HANGUL SYLLABLE-BE6A
        '48747' => '13012'            , # G+13012  U+BE6B  uniBE6B,  HANGUL SYLLABLE-BE6B
        '48748' => '13013'            , # G+13013  U+BE6C  uniBE6C,  HANGUL SYLLABLE-BE6C
        '48749' => '13014'            , # G+13014  U+BE6D  uniBE6D,  HANGUL SYLLABLE-BE6D
        '48750' => '13015'            , # G+13015  U+BE6E  uniBE6E,  HANGUL SYLLABLE-BE6E
        '48751' => '13016'            , # G+13016  U+BE6F  uniBE6F,  HANGUL SYLLABLE-BE6F
        '48752' => '2122'             , # G+2122  U+BE70  uniBE70,  HANGUL SYLLABLE-BE70
        '48753' => '2123'             , # G+2123  U+BE71  uniBE71,  HANGUL SYLLABLE-BE71
        '48754' => '13017'            , # G+13017  U+BE72  uniBE72,  HANGUL SYLLABLE-BE72
        '48755' => '2124'             , # G+2124  U+BE73  uniBE73,  HANGUL SYLLABLE-BE73
        '48756' => '2125'             , # G+2125  U+BE74  uniBE74,  HANGUL SYLLABLE-BE74
        '48757' => '2126'             , # G+2126  U+BE75  uniBE75,  HANGUL SYLLABLE-BE75
        '48758' => '13018'            , # G+13018  U+BE76  uniBE76,  HANGUL SYLLABLE-BE76
        '48759' => '13019'            , # G+13019  U+BE77  uniBE77,  HANGUL SYLLABLE-BE77
        '48760' => '13020'            , # G+13020  U+BE78  uniBE78,  HANGUL SYLLABLE-BE78
        '48761' => '13021'            , # G+13021  U+BE79  uniBE79,  HANGUL SYLLABLE-BE79
        '48762' => '13022'            , # G+13022  U+BE7A  uniBE7A,  HANGUL SYLLABLE-BE7A
        '48763' => '2127'             , # G+2127  U+BE7B  uniBE7B,  HANGUL SYLLABLE-BE7B
        '48764' => '2128'             , # G+2128  U+BE7C  uniBE7C,  HANGUL SYLLABLE-BE7C
        '48765' => '2129'             , # G+2129  U+BE7D  uniBE7D,  HANGUL SYLLABLE-BE7D
        '48766' => '13023'            , # G+13023  U+BE7E  uniBE7E,  HANGUL SYLLABLE-BE7E
        '48767' => '13024'            , # G+13024  U+BE7F  uniBE7F,  HANGUL SYLLABLE-BE7F
        '48768' => '2130'             , # G+2130  U+BE80  uniBE80,  HANGUL SYLLABLE-BE80
        '48769' => '13025'            , # G+13025  U+BE81  uniBE81,  HANGUL SYLLABLE-BE81
        '48770' => '13026'            , # G+13026  U+BE82  uniBE82,  HANGUL SYLLABLE-BE82
        '48771' => '13027'            , # G+13027  U+BE83  uniBE83,  HANGUL SYLLABLE-BE83
        '48772' => '2131'             , # G+2131  U+BE84  uniBE84,  HANGUL SYLLABLE-BE84
        '48773' => '13028'            , # G+13028  U+BE85  uniBE85,  HANGUL SYLLABLE-BE85
        '48774' => '13029'            , # G+13029  U+BE86  uniBE86,  HANGUL SYLLABLE-BE86
        '48775' => '13030'            , # G+13030  U+BE87  uniBE87,  HANGUL SYLLABLE-BE87
        '48776' => '13031'            , # G+13031  U+BE88  uniBE88,  HANGUL SYLLABLE-BE88
        '48777' => '13032'            , # G+13032  U+BE89  uniBE89,  HANGUL SYLLABLE-BE89
        '48778' => '13033'            , # G+13033  U+BE8A  uniBE8A,  HANGUL SYLLABLE-BE8A
        '48779' => '13034'            , # G+13034  U+BE8B  uniBE8B,  HANGUL SYLLABLE-BE8B
        '48780' => '2132'             , # G+2132  U+BE8C  uniBE8C,  HANGUL SYLLABLE-BE8C
        '48781' => '2133'             , # G+2133  U+BE8D  uniBE8D,  HANGUL SYLLABLE-BE8D
        '48782' => '13035'            , # G+13035  U+BE8E  uniBE8E,  HANGUL SYLLABLE-BE8E
        '48783' => '2134'             , # G+2134  U+BE8F  uniBE8F,  HANGUL SYLLABLE-BE8F
        '48784' => '2135'             , # G+2135  U+BE90  uniBE90,  HANGUL SYLLABLE-BE90
        '48785' => '2136'             , # G+2136  U+BE91  uniBE91,  HANGUL SYLLABLE-BE91
        '48786' => '13036'            , # G+13036  U+BE92  uniBE92,  HANGUL SYLLABLE-BE92
        '48787' => '13037'            , # G+13037  U+BE93  uniBE93,  HANGUL SYLLABLE-BE93
        '48788' => '13038'            , # G+13038  U+BE94  uniBE94,  HANGUL SYLLABLE-BE94
        '48789' => '13039'            , # G+13039  U+BE95  uniBE95,  HANGUL SYLLABLE-BE95
        '48790' => '13040'            , # G+13040  U+BE96  uniBE96,  HANGUL SYLLABLE-BE96
        '48791' => '13041'            , # G+13041  U+BE97  uniBE97,  HANGUL SYLLABLE-BE97
        '48792' => '2137'             , # G+2137  U+BE98  uniBE98,  HANGUL SYLLABLE-BE98
        '48793' => '2138'             , # G+2138  U+BE99  uniBE99,  HANGUL SYLLABLE-BE99
        '48794' => '13042'            , # G+13042  U+BE9A  uniBE9A,  HANGUL SYLLABLE-BE9A
        '48795' => '13043'            , # G+13043  U+BE9B  uniBE9B,  HANGUL SYLLABLE-BE9B
        '48796' => '13044'            , # G+13044  U+BE9C  uniBE9C,  HANGUL SYLLABLE-BE9C
        '48797' => '13045'            , # G+13045  U+BE9D  uniBE9D,  HANGUL SYLLABLE-BE9D
        '48798' => '13046'            , # G+13046  U+BE9E  uniBE9E,  HANGUL SYLLABLE-BE9E
        '48799' => '13047'            , # G+13047  U+BE9F  uniBE9F,  HANGUL SYLLABLE-BE9F
        '48800' => '13048'            , # G+13048  U+BEA0  uniBEA0,  HANGUL SYLLABLE-BEA0
        '48801' => '13049'            , # G+13049  U+BEA1  uniBEA1,  HANGUL SYLLABLE-BEA1
        '48802' => '13050'            , # G+13050  U+BEA2  uniBEA2,  HANGUL SYLLABLE-BEA2
        '48803' => '13051'            , # G+13051  U+BEA3  uniBEA3,  HANGUL SYLLABLE-BEA3
        '48804' => '13052'            , # G+13052  U+BEA4  uniBEA4,  HANGUL SYLLABLE-BEA4
        '48805' => '13053'            , # G+13053  U+BEA5  uniBEA5,  HANGUL SYLLABLE-BEA5
        '48806' => '13054'            , # G+13054  U+BEA6  uniBEA6,  HANGUL SYLLABLE-BEA6
        '48807' => '13055'            , # G+13055  U+BEA7  uniBEA7,  HANGUL SYLLABLE-BEA7
        '48808' => '2139'             , # G+2139  U+BEA8  uniBEA8,  HANGUL SYLLABLE-BEA8
        '48809' => '13056'            , # G+13056  U+BEA9  uniBEA9,  HANGUL SYLLABLE-BEA9
        '48810' => '13057'            , # G+13057  U+BEAA  uniBEAA,  HANGUL SYLLABLE-BEAA
        '48811' => '13058'            , # G+13058  U+BEAB  uniBEAB,  HANGUL SYLLABLE-BEAB
        '48812' => '13059'            , # G+13059  U+BEAC  uniBEAC,  HANGUL SYLLABLE-BEAC
        '48813' => '13060'            , # G+13060  U+BEAD  uniBEAD,  HANGUL SYLLABLE-BEAD
        '48814' => '13061'            , # G+13061  U+BEAE  uniBEAE,  HANGUL SYLLABLE-BEAE
        '48815' => '13062'            , # G+13062  U+BEAF  uniBEAF,  HANGUL SYLLABLE-BEAF
        '48816' => '13063'            , # G+13063  U+BEB0  uniBEB0,  HANGUL SYLLABLE-BEB0
        '48817' => '13064'            , # G+13064  U+BEB1  uniBEB1,  HANGUL SYLLABLE-BEB1
        '48818' => '13065'            , # G+13065  U+BEB2  uniBEB2,  HANGUL SYLLABLE-BEB2
        '48819' => '13066'            , # G+13066  U+BEB3  uniBEB3,  HANGUL SYLLABLE-BEB3
        '48820' => '13067'            , # G+13067  U+BEB4  uniBEB4,  HANGUL SYLLABLE-BEB4
        '48821' => '13068'            , # G+13068  U+BEB5  uniBEB5,  HANGUL SYLLABLE-BEB5
        '48822' => '13069'            , # G+13069  U+BEB6  uniBEB6,  HANGUL SYLLABLE-BEB6
        '48823' => '13070'            , # G+13070  U+BEB7  uniBEB7,  HANGUL SYLLABLE-BEB7
        '48824' => '13071'            , # G+13071  U+BEB8  uniBEB8,  HANGUL SYLLABLE-BEB8
        '48825' => '13072'            , # G+13072  U+BEB9  uniBEB9,  HANGUL SYLLABLE-BEB9
        '48826' => '13073'            , # G+13073  U+BEBA  uniBEBA,  HANGUL SYLLABLE-BEBA
        '48827' => '13074'            , # G+13074  U+BEBB  uniBEBB,  HANGUL SYLLABLE-BEBB
        '48828' => '13075'            , # G+13075  U+BEBC  uniBEBC,  HANGUL SYLLABLE-BEBC
        '48829' => '13076'            , # G+13076  U+BEBD  uniBEBD,  HANGUL SYLLABLE-BEBD
        '48830' => '13077'            , # G+13077  U+BEBE  uniBEBE,  HANGUL SYLLABLE-BEBE
        '48831' => '13078'            , # G+13078  U+BEBF  uniBEBF,  HANGUL SYLLABLE-BEBF
        '48832' => '13079'            , # G+13079  U+BEC0  uniBEC0,  HANGUL SYLLABLE-BEC0
        '48833' => '13080'            , # G+13080  U+BEC1  uniBEC1,  HANGUL SYLLABLE-BEC1
        '48834' => '13081'            , # G+13081  U+BEC2  uniBEC2,  HANGUL SYLLABLE-BEC2
        '48835' => '13082'            , # G+13082  U+BEC3  uniBEC3,  HANGUL SYLLABLE-BEC3
        '48836' => '13083'            , # G+13083  U+BEC4  uniBEC4,  HANGUL SYLLABLE-BEC4
        '48837' => '13084'            , # G+13084  U+BEC5  uniBEC5,  HANGUL SYLLABLE-BEC5
        '48838' => '13085'            , # G+13085  U+BEC6  uniBEC6,  HANGUL SYLLABLE-BEC6
        '48839' => '13086'            , # G+13086  U+BEC7  uniBEC7,  HANGUL SYLLABLE-BEC7
        '48840' => '13087'            , # G+13087  U+BEC8  uniBEC8,  HANGUL SYLLABLE-BEC8
        '48841' => '13088'            , # G+13088  U+BEC9  uniBEC9,  HANGUL SYLLABLE-BEC9
        '48842' => '13089'            , # G+13089  U+BECA  uniBECA,  HANGUL SYLLABLE-BECA
        '48843' => '13090'            , # G+13090  U+BECB  uniBECB,  HANGUL SYLLABLE-BECB
        '48844' => '13091'            , # G+13091  U+BECC  uniBECC,  HANGUL SYLLABLE-BECC
        '48845' => '13092'            , # G+13092  U+BECD  uniBECD,  HANGUL SYLLABLE-BECD
        '48846' => '13093'            , # G+13093  U+BECE  uniBECE,  HANGUL SYLLABLE-BECE
        '48847' => '13094'            , # G+13094  U+BECF  uniBECF,  HANGUL SYLLABLE-BECF
        '48848' => '2140'             , # G+2140  U+BED0  uniBED0,  HANGUL SYLLABLE-BED0
        '48849' => '2141'             , # G+2141  U+BED1  uniBED1,  HANGUL SYLLABLE-BED1
        '48850' => '13095'            , # G+13095  U+BED2  uniBED2,  HANGUL SYLLABLE-BED2
        '48851' => '13096'            , # G+13096  U+BED3  uniBED3,  HANGUL SYLLABLE-BED3
        '48852' => '2142'             , # G+2142  U+BED4  uniBED4,  HANGUL SYLLABLE-BED4
        '48853' => '13097'            , # G+13097  U+BED5  uniBED5,  HANGUL SYLLABLE-BED5
        '48854' => '13098'            , # G+13098  U+BED6  uniBED6,  HANGUL SYLLABLE-BED6
        '48855' => '2143'             , # G+2143  U+BED7  uniBED7,  HANGUL SYLLABLE-BED7
        '48856' => '2144'             , # G+2144  U+BED8  uniBED8,  HANGUL SYLLABLE-BED8
        '48857' => '13099'            , # G+13099  U+BED9  uniBED9,  HANGUL SYLLABLE-BED9
        '48858' => '13100'            , # G+13100  U+BEDA  uniBEDA,  HANGUL SYLLABLE-BEDA
        '48859' => '13101'            , # G+13101  U+BEDB  uniBEDB,  HANGUL SYLLABLE-BEDB
        '48860' => '13102'            , # G+13102  U+BEDC  uniBEDC,  HANGUL SYLLABLE-BEDC
        '48861' => '13103'            , # G+13103  U+BEDD  uniBEDD,  HANGUL SYLLABLE-BEDD
        '48862' => '13104'            , # G+13104  U+BEDE  uniBEDE,  HANGUL SYLLABLE-BEDE
        '48863' => '13105'            , # G+13105  U+BEDF  uniBEDF,  HANGUL SYLLABLE-BEDF
        '48864' => '2145'             , # G+2145  U+BEE0  uniBEE0,  HANGUL SYLLABLE-BEE0
        '48865' => '13106'            , # G+13106  U+BEE1  uniBEE1,  HANGUL SYLLABLE-BEE1
        '48866' => '13107'            , # G+13107  U+BEE2  uniBEE2,  HANGUL SYLLABLE-BEE2
        '48867' => '2146'             , # G+2146  U+BEE3  uniBEE3,  HANGUL SYLLABLE-BEE3
        '48868' => '2147'             , # G+2147  U+BEE4  uniBEE4,  HANGUL SYLLABLE-BEE4
        '48869' => '2148'             , # G+2148  U+BEE5  uniBEE5,  HANGUL SYLLABLE-BEE5
        '48870' => '13108'            , # G+13108  U+BEE6  uniBEE6,  HANGUL SYLLABLE-BEE6
        '48871' => '13109'            , # G+13109  U+BEE7  uniBEE7,  HANGUL SYLLABLE-BEE7
        '48872' => '13110'            , # G+13110  U+BEE8  uniBEE8,  HANGUL SYLLABLE-BEE8
        '48873' => '13111'            , # G+13111  U+BEE9  uniBEE9,  HANGUL SYLLABLE-BEE9
        '48874' => '13112'            , # G+13112  U+BEEA  uniBEEA,  HANGUL SYLLABLE-BEEA
        '48875' => '13113'            , # G+13113  U+BEEB  uniBEEB,  HANGUL SYLLABLE-BEEB
        '48876' => '2149'             , # G+2149  U+BEEC  uniBEEC,  HANGUL SYLLABLE-BEEC
        '48877' => '13114'            , # G+13114  U+BEED  uniBEED,  HANGUL SYLLABLE-BEED
        '48878' => '13115'            , # G+13115  U+BEEE  uniBEEE,  HANGUL SYLLABLE-BEEE
        '48879' => '13116'            , # G+13116  U+BEEF  uniBEEF,  HANGUL SYLLABLE-BEEF
        '48880' => '13117'            , # G+13117  U+BEF0  uniBEF0,  HANGUL SYLLABLE-BEF0
        '48881' => '13118'            , # G+13118  U+BEF1  uniBEF1,  HANGUL SYLLABLE-BEF1
        '48882' => '13119'            , # G+13119  U+BEF2  uniBEF2,  HANGUL SYLLABLE-BEF2
        '48883' => '13120'            , # G+13120  U+BEF3  uniBEF3,  HANGUL SYLLABLE-BEF3
        '48884' => '13121'            , # G+13121  U+BEF4  uniBEF4,  HANGUL SYLLABLE-BEF4
        '48885' => '13122'            , # G+13122  U+BEF5  uniBEF5,  HANGUL SYLLABLE-BEF5
        '48886' => '13123'            , # G+13123  U+BEF6  uniBEF6,  HANGUL SYLLABLE-BEF6
        '48887' => '13124'            , # G+13124  U+BEF7  uniBEF7,  HANGUL SYLLABLE-BEF7
        '48888' => '13125'            , # G+13125  U+BEF8  uniBEF8,  HANGUL SYLLABLE-BEF8
        '48889' => '13126'            , # G+13126  U+BEF9  uniBEF9,  HANGUL SYLLABLE-BEF9
        '48890' => '13127'            , # G+13127  U+BEFA  uniBEFA,  HANGUL SYLLABLE-BEFA
        '48891' => '13128'            , # G+13128  U+BEFB  uniBEFB,  HANGUL SYLLABLE-BEFB
        '48892' => '13129'            , # G+13129  U+BEFC  uniBEFC,  HANGUL SYLLABLE-BEFC
        '48893' => '13130'            , # G+13130  U+BEFD  uniBEFD,  HANGUL SYLLABLE-BEFD
        '48894' => '13131'            , # G+13131  U+BEFE  uniBEFE,  HANGUL SYLLABLE-BEFE
        '48895' => '13132'            , # G+13132  U+BEFF  uniBEFF,  HANGUL SYLLABLE-BEFF
        '48896' => '13133'            , # G+13133  U+BF00  uniBF00,  HANGUL SYLLABLE-BF00
        '48897' => '2150'             , # G+2150  U+BF01  uniBF01,  HANGUL SYLLABLE-BF01
        '48898' => '13134'            , # G+13134  U+BF02  uniBF02,  HANGUL SYLLABLE-BF02
        '48899' => '13135'            , # G+13135  U+BF03  uniBF03,  HANGUL SYLLABLE-BF03
        '48900' => '13136'            , # G+13136  U+BF04  uniBF04,  HANGUL SYLLABLE-BF04
        '48901' => '13137'            , # G+13137  U+BF05  uniBF05,  HANGUL SYLLABLE-BF05
        '48902' => '13138'            , # G+13138  U+BF06  uniBF06,  HANGUL SYLLABLE-BF06
        '48903' => '13139'            , # G+13139  U+BF07  uniBF07,  HANGUL SYLLABLE-BF07
        '48904' => '2151'             , # G+2151  U+BF08  uniBF08,  HANGUL SYLLABLE-BF08
        '48905' => '2152'             , # G+2152  U+BF09  uniBF09,  HANGUL SYLLABLE-BF09
        '48906' => '13140'            , # G+13140  U+BF0A  uniBF0A,  HANGUL SYLLABLE-BF0A
        '48907' => '13141'            , # G+13141  U+BF0B  uniBF0B,  HANGUL SYLLABLE-BF0B
        '48908' => '13142'            , # G+13142  U+BF0C  uniBF0C,  HANGUL SYLLABLE-BF0C
        '48909' => '13143'            , # G+13143  U+BF0D  uniBF0D,  HANGUL SYLLABLE-BF0D
        '48910' => '13144'            , # G+13144  U+BF0E  uniBF0E,  HANGUL SYLLABLE-BF0E
        '48911' => '13145'            , # G+13145  U+BF0F  uniBF0F,  HANGUL SYLLABLE-BF0F
        '48912' => '13146'            , # G+13146  U+BF10  uniBF10,  HANGUL SYLLABLE-BF10
        '48913' => '13147'            , # G+13147  U+BF11  uniBF11,  HANGUL SYLLABLE-BF11
        '48914' => '13148'            , # G+13148  U+BF12  uniBF12,  HANGUL SYLLABLE-BF12
        '48915' => '13149'            , # G+13149  U+BF13  uniBF13,  HANGUL SYLLABLE-BF13
        '48916' => '13150'            , # G+13150  U+BF14  uniBF14,  HANGUL SYLLABLE-BF14
        '48917' => '13151'            , # G+13151  U+BF15  uniBF15,  HANGUL SYLLABLE-BF15
        '48918' => '13152'            , # G+13152  U+BF16  uniBF16,  HANGUL SYLLABLE-BF16
        '48919' => '13153'            , # G+13153  U+BF17  uniBF17,  HANGUL SYLLABLE-BF17
        '48920' => '2153'             , # G+2153  U+BF18  uniBF18,  HANGUL SYLLABLE-BF18
        '48921' => '2154'             , # G+2154  U+BF19  uniBF19,  HANGUL SYLLABLE-BF19
        '48922' => '13154'            , # G+13154  U+BF1A  uniBF1A,  HANGUL SYLLABLE-BF1A
        '48923' => '2155'             , # G+2155  U+BF1B  uniBF1B,  HANGUL SYLLABLE-BF1B
        '48924' => '2156'             , # G+2156  U+BF1C  uniBF1C,  HANGUL SYLLABLE-BF1C
        '48925' => '2157'             , # G+2157  U+BF1D  uniBF1D,  HANGUL SYLLABLE-BF1D
        '48926' => '13155'            , # G+13155  U+BF1E  uniBF1E,  HANGUL SYLLABLE-BF1E
        '48927' => '13156'            , # G+13156  U+BF1F  uniBF1F,  HANGUL SYLLABLE-BF1F
        '48928' => '13157'            , # G+13157  U+BF20  uniBF20,  HANGUL SYLLABLE-BF20
        '48929' => '13158'            , # G+13158  U+BF21  uniBF21,  HANGUL SYLLABLE-BF21
        '48930' => '13159'            , # G+13159  U+BF22  uniBF22,  HANGUL SYLLABLE-BF22
        '48931' => '13160'            , # G+13160  U+BF23  uniBF23,  HANGUL SYLLABLE-BF23
        '48932' => '13161'            , # G+13161  U+BF24  uniBF24,  HANGUL SYLLABLE-BF24
        '48933' => '13162'            , # G+13162  U+BF25  uniBF25,  HANGUL SYLLABLE-BF25
        '48934' => '13163'            , # G+13163  U+BF26  uniBF26,  HANGUL SYLLABLE-BF26
        '48935' => '13164'            , # G+13164  U+BF27  uniBF27,  HANGUL SYLLABLE-BF27
        '48936' => '13165'            , # G+13165  U+BF28  uniBF28,  HANGUL SYLLABLE-BF28
        '48937' => '13166'            , # G+13166  U+BF29  uniBF29,  HANGUL SYLLABLE-BF29
        '48938' => '13167'            , # G+13167  U+BF2A  uniBF2A,  HANGUL SYLLABLE-BF2A
        '48939' => '13168'            , # G+13168  U+BF2B  uniBF2B,  HANGUL SYLLABLE-BF2B
        '48940' => '13169'            , # G+13169  U+BF2C  uniBF2C,  HANGUL SYLLABLE-BF2C
        '48941' => '13170'            , # G+13170  U+BF2D  uniBF2D,  HANGUL SYLLABLE-BF2D
        '48942' => '13171'            , # G+13171  U+BF2E  uniBF2E,  HANGUL SYLLABLE-BF2E
        '48943' => '13172'            , # G+13172  U+BF2F  uniBF2F,  HANGUL SYLLABLE-BF2F
        '48944' => '13173'            , # G+13173  U+BF30  uniBF30,  HANGUL SYLLABLE-BF30
        '48945' => '13174'            , # G+13174  U+BF31  uniBF31,  HANGUL SYLLABLE-BF31
        '48946' => '13175'            , # G+13175  U+BF32  uniBF32,  HANGUL SYLLABLE-BF32
        '48947' => '13176'            , # G+13176  U+BF33  uniBF33,  HANGUL SYLLABLE-BF33
        '48948' => '13177'            , # G+13177  U+BF34  uniBF34,  HANGUL SYLLABLE-BF34
        '48949' => '13178'            , # G+13178  U+BF35  uniBF35,  HANGUL SYLLABLE-BF35
        '48950' => '13179'            , # G+13179  U+BF36  uniBF36,  HANGUL SYLLABLE-BF36
        '48951' => '13180'            , # G+13180  U+BF37  uniBF37,  HANGUL SYLLABLE-BF37
        '48952' => '13181'            , # G+13181  U+BF38  uniBF38,  HANGUL SYLLABLE-BF38
        '48953' => '13182'            , # G+13182  U+BF39  uniBF39,  HANGUL SYLLABLE-BF39
        '48954' => '13183'            , # G+13183  U+BF3A  uniBF3A,  HANGUL SYLLABLE-BF3A
        '48955' => '13184'            , # G+13184  U+BF3B  uniBF3B,  HANGUL SYLLABLE-BF3B
        '48956' => '13185'            , # G+13185  U+BF3C  uniBF3C,  HANGUL SYLLABLE-BF3C
        '48957' => '13186'            , # G+13186  U+BF3D  uniBF3D,  HANGUL SYLLABLE-BF3D
        '48958' => '13187'            , # G+13187  U+BF3E  uniBF3E,  HANGUL SYLLABLE-BF3E
        '48959' => '13188'            , # G+13188  U+BF3F  uniBF3F,  HANGUL SYLLABLE-BF3F
        '48960' => '2158'             , # G+2158  U+BF40  uniBF40,  HANGUL SYLLABLE-BF40
        '48961' => '2159'             , # G+2159  U+BF41  uniBF41,  HANGUL SYLLABLE-BF41
        '48962' => '13189'            , # G+13189  U+BF42  uniBF42,  HANGUL SYLLABLE-BF42
        '48963' => '13190'            , # G+13190  U+BF43  uniBF43,  HANGUL SYLLABLE-BF43
        '48964' => '2160'             , # G+2160  U+BF44  uniBF44,  HANGUL SYLLABLE-BF44
        '48965' => '13191'            , # G+13191  U+BF45  uniBF45,  HANGUL SYLLABLE-BF45
        '48966' => '13192'            , # G+13192  U+BF46  uniBF46,  HANGUL SYLLABLE-BF46
        '48967' => '13193'            , # G+13193  U+BF47  uniBF47,  HANGUL SYLLABLE-BF47
        '48968' => '2161'             , # G+2161  U+BF48  uniBF48,  HANGUL SYLLABLE-BF48
        '48969' => '13194'            , # G+13194  U+BF49  uniBF49,  HANGUL SYLLABLE-BF49
        '48970' => '13195'            , # G+13195  U+BF4A  uniBF4A,  HANGUL SYLLABLE-BF4A
        '48971' => '13196'            , # G+13196  U+BF4B  uniBF4B,  HANGUL SYLLABLE-BF4B
        '48972' => '13197'            , # G+13197  U+BF4C  uniBF4C,  HANGUL SYLLABLE-BF4C
        '48973' => '13198'            , # G+13198  U+BF4D  uniBF4D,  HANGUL SYLLABLE-BF4D
        '48974' => '13199'            , # G+13199  U+BF4E  uniBF4E,  HANGUL SYLLABLE-BF4E
        '48975' => '13200'            , # G+13200  U+BF4F  uniBF4F,  HANGUL SYLLABLE-BF4F
        '48976' => '2162'             , # G+2162  U+BF50  uniBF50,  HANGUL SYLLABLE-BF50
        '48977' => '2163'             , # G+2163  U+BF51  uniBF51,  HANGUL SYLLABLE-BF51
        '48978' => '13201'            , # G+13201  U+BF52  uniBF52,  HANGUL SYLLABLE-BF52
        '48979' => '13202'            , # G+13202  U+BF53  uniBF53,  HANGUL SYLLABLE-BF53
        '48980' => '13203'            , # G+13203  U+BF54  uniBF54,  HANGUL SYLLABLE-BF54
        '48981' => '2164'             , # G+2164  U+BF55  uniBF55,  HANGUL SYLLABLE-BF55
        '48982' => '13204'            , # G+13204  U+BF56  uniBF56,  HANGUL SYLLABLE-BF56
        '48983' => '13205'            , # G+13205  U+BF57  uniBF57,  HANGUL SYLLABLE-BF57
        '48984' => '13206'            , # G+13206  U+BF58  uniBF58,  HANGUL SYLLABLE-BF58
        '48985' => '13207'            , # G+13207  U+BF59  uniBF59,  HANGUL SYLLABLE-BF59
        '48986' => '13208'            , # G+13208  U+BF5A  uniBF5A,  HANGUL SYLLABLE-BF5A
        '48987' => '13209'            , # G+13209  U+BF5B  uniBF5B,  HANGUL SYLLABLE-BF5B
        '48988' => '13210'            , # G+13210  U+BF5C  uniBF5C,  HANGUL SYLLABLE-BF5C
        '48989' => '13211'            , # G+13211  U+BF5D  uniBF5D,  HANGUL SYLLABLE-BF5D
        '48990' => '13212'            , # G+13212  U+BF5E  uniBF5E,  HANGUL SYLLABLE-BF5E
        '48991' => '13213'            , # G+13213  U+BF5F  uniBF5F,  HANGUL SYLLABLE-BF5F
        '48992' => '13214'            , # G+13214  U+BF60  uniBF60,  HANGUL SYLLABLE-BF60
        '48993' => '13215'            , # G+13215  U+BF61  uniBF61,  HANGUL SYLLABLE-BF61
        '48994' => '13216'            , # G+13216  U+BF62  uniBF62,  HANGUL SYLLABLE-BF62
        '48995' => '13217'            , # G+13217  U+BF63  uniBF63,  HANGUL SYLLABLE-BF63
        '48996' => '13218'            , # G+13218  U+BF64  uniBF64,  HANGUL SYLLABLE-BF64
        '48997' => '13219'            , # G+13219  U+BF65  uniBF65,  HANGUL SYLLABLE-BF65
        '48998' => '13220'            , # G+13220  U+BF66  uniBF66,  HANGUL SYLLABLE-BF66
        '48999' => '13221'            , # G+13221  U+BF67  uniBF67,  HANGUL SYLLABLE-BF67
        '49000' => '13222'            , # G+13222  U+BF68  uniBF68,  HANGUL SYLLABLE-BF68
        '49001' => '13223'            , # G+13223  U+BF69  uniBF69,  HANGUL SYLLABLE-BF69
        '49002' => '13224'            , # G+13224  U+BF6A  uniBF6A,  HANGUL SYLLABLE-BF6A
        '49003' => '13225'            , # G+13225  U+BF6B  uniBF6B,  HANGUL SYLLABLE-BF6B
        '49004' => '13226'            , # G+13226  U+BF6C  uniBF6C,  HANGUL SYLLABLE-BF6C
        '49005' => '13227'            , # G+13227  U+BF6D  uniBF6D,  HANGUL SYLLABLE-BF6D
        '49006' => '13228'            , # G+13228  U+BF6E  uniBF6E,  HANGUL SYLLABLE-BF6E
        '49007' => '13229'            , # G+13229  U+BF6F  uniBF6F,  HANGUL SYLLABLE-BF6F
        '49008' => '13230'            , # G+13230  U+BF70  uniBF70,  HANGUL SYLLABLE-BF70
        '49009' => '13231'            , # G+13231  U+BF71  uniBF71,  HANGUL SYLLABLE-BF71
        '49010' => '13232'            , # G+13232  U+BF72  uniBF72,  HANGUL SYLLABLE-BF72
        '49011' => '13233'            , # G+13233  U+BF73  uniBF73,  HANGUL SYLLABLE-BF73
        '49012' => '13234'            , # G+13234  U+BF74  uniBF74,  HANGUL SYLLABLE-BF74
        '49013' => '13235'            , # G+13235  U+BF75  uniBF75,  HANGUL SYLLABLE-BF75
        '49014' => '13236'            , # G+13236  U+BF76  uniBF76,  HANGUL SYLLABLE-BF76
        '49015' => '13237'            , # G+13237  U+BF77  uniBF77,  HANGUL SYLLABLE-BF77
        '49016' => '13238'            , # G+13238  U+BF78  uniBF78,  HANGUL SYLLABLE-BF78
        '49017' => '13239'            , # G+13239  U+BF79  uniBF79,  HANGUL SYLLABLE-BF79
        '49018' => '13240'            , # G+13240  U+BF7A  uniBF7A,  HANGUL SYLLABLE-BF7A
        '49019' => '13241'            , # G+13241  U+BF7B  uniBF7B,  HANGUL SYLLABLE-BF7B
        '49020' => '13242'            , # G+13242  U+BF7C  uniBF7C,  HANGUL SYLLABLE-BF7C
        '49021' => '13243'            , # G+13243  U+BF7D  uniBF7D,  HANGUL SYLLABLE-BF7D
        '49022' => '13244'            , # G+13244  U+BF7E  uniBF7E,  HANGUL SYLLABLE-BF7E
        '49023' => '13245'            , # G+13245  U+BF7F  uniBF7F,  HANGUL SYLLABLE-BF7F
        '49024' => '13246'            , # G+13246  U+BF80  uniBF80,  HANGUL SYLLABLE-BF80
        '49025' => '13247'            , # G+13247  U+BF81  uniBF81,  HANGUL SYLLABLE-BF81
        '49026' => '13248'            , # G+13248  U+BF82  uniBF82,  HANGUL SYLLABLE-BF82
        '49027' => '13249'            , # G+13249  U+BF83  uniBF83,  HANGUL SYLLABLE-BF83
        '49028' => '13250'            , # G+13250  U+BF84  uniBF84,  HANGUL SYLLABLE-BF84
        '49029' => '13251'            , # G+13251  U+BF85  uniBF85,  HANGUL SYLLABLE-BF85
        '49030' => '13252'            , # G+13252  U+BF86  uniBF86,  HANGUL SYLLABLE-BF86
        '49031' => '13253'            , # G+13253  U+BF87  uniBF87,  HANGUL SYLLABLE-BF87
        '49032' => '13254'            , # G+13254  U+BF88  uniBF88,  HANGUL SYLLABLE-BF88
        '49033' => '13255'            , # G+13255  U+BF89  uniBF89,  HANGUL SYLLABLE-BF89
        '49034' => '13256'            , # G+13256  U+BF8A  uniBF8A,  HANGUL SYLLABLE-BF8A
        '49035' => '13257'            , # G+13257  U+BF8B  uniBF8B,  HANGUL SYLLABLE-BF8B
        '49036' => '13258'            , # G+13258  U+BF8C  uniBF8C,  HANGUL SYLLABLE-BF8C
        '49037' => '13259'            , # G+13259  U+BF8D  uniBF8D,  HANGUL SYLLABLE-BF8D
        '49038' => '13260'            , # G+13260  U+BF8E  uniBF8E,  HANGUL SYLLABLE-BF8E
        '49039' => '13261'            , # G+13261  U+BF8F  uniBF8F,  HANGUL SYLLABLE-BF8F
        '49040' => '13262'            , # G+13262  U+BF90  uniBF90,  HANGUL SYLLABLE-BF90
        '49041' => '13263'            , # G+13263  U+BF91  uniBF91,  HANGUL SYLLABLE-BF91
        '49042' => '13264'            , # G+13264  U+BF92  uniBF92,  HANGUL SYLLABLE-BF92
        '49043' => '13265'            , # G+13265  U+BF93  uniBF93,  HANGUL SYLLABLE-BF93
        '49044' => '2165'             , # G+2165  U+BF94  uniBF94,  HANGUL SYLLABLE-BF94
        '49045' => '13266'            , # G+13266  U+BF95  uniBF95,  HANGUL SYLLABLE-BF95
        '49046' => '13267'            , # G+13267  U+BF96  uniBF96,  HANGUL SYLLABLE-BF96
        '49047' => '13268'            , # G+13268  U+BF97  uniBF97,  HANGUL SYLLABLE-BF97
        '49048' => '13269'            , # G+13269  U+BF98  uniBF98,  HANGUL SYLLABLE-BF98
        '49049' => '13270'            , # G+13270  U+BF99  uniBF99,  HANGUL SYLLABLE-BF99
        '49050' => '13271'            , # G+13271  U+BF9A  uniBF9A,  HANGUL SYLLABLE-BF9A
        '49051' => '13272'            , # G+13272  U+BF9B  uniBF9B,  HANGUL SYLLABLE-BF9B
        '49052' => '13273'            , # G+13273  U+BF9C  uniBF9C,  HANGUL SYLLABLE-BF9C
        '49053' => '13274'            , # G+13274  U+BF9D  uniBF9D,  HANGUL SYLLABLE-BF9D
        '49054' => '13275'            , # G+13275  U+BF9E  uniBF9E,  HANGUL SYLLABLE-BF9E
        '49055' => '13276'            , # G+13276  U+BF9F  uniBF9F,  HANGUL SYLLABLE-BF9F
        '49056' => '13277'            , # G+13277  U+BFA0  uniBFA0,  HANGUL SYLLABLE-BFA0
        '49057' => '13278'            , # G+13278  U+BFA1  uniBFA1,  HANGUL SYLLABLE-BFA1
        '49058' => '13279'            , # G+13279  U+BFA2  uniBFA2,  HANGUL SYLLABLE-BFA2
        '49059' => '13280'            , # G+13280  U+BFA3  uniBFA3,  HANGUL SYLLABLE-BFA3
        '49060' => '13281'            , # G+13281  U+BFA4  uniBFA4,  HANGUL SYLLABLE-BFA4
        '49061' => '13282'            , # G+13282  U+BFA5  uniBFA5,  HANGUL SYLLABLE-BFA5
        '49062' => '13283'            , # G+13283  U+BFA6  uniBFA6,  HANGUL SYLLABLE-BFA6
        '49063' => '13284'            , # G+13284  U+BFA7  uniBFA7,  HANGUL SYLLABLE-BFA7
        '49064' => '13285'            , # G+13285  U+BFA8  uniBFA8,  HANGUL SYLLABLE-BFA8
        '49065' => '13286'            , # G+13286  U+BFA9  uniBFA9,  HANGUL SYLLABLE-BFA9
        '49066' => '13287'            , # G+13287  U+BFAA  uniBFAA,  HANGUL SYLLABLE-BFAA
        '49067' => '13288'            , # G+13288  U+BFAB  uniBFAB,  HANGUL SYLLABLE-BFAB
        '49068' => '13289'            , # G+13289  U+BFAC  uniBFAC,  HANGUL SYLLABLE-BFAC
        '49069' => '13290'            , # G+13290  U+BFAD  uniBFAD,  HANGUL SYLLABLE-BFAD
        '49070' => '13291'            , # G+13291  U+BFAE  uniBFAE,  HANGUL SYLLABLE-BFAE
        '49071' => '13292'            , # G+13292  U+BFAF  uniBFAF,  HANGUL SYLLABLE-BFAF
        '49072' => '2166'             , # G+2166  U+BFB0  uniBFB0,  HANGUL SYLLABLE-BFB0
        '49073' => '13293'            , # G+13293  U+BFB1  uniBFB1,  HANGUL SYLLABLE-BFB1
        '49074' => '13294'            , # G+13294  U+BFB2  uniBFB2,  HANGUL SYLLABLE-BFB2
        '49075' => '13295'            , # G+13295  U+BFB3  uniBFB3,  HANGUL SYLLABLE-BFB3
        '49076' => '13296'            , # G+13296  U+BFB4  uniBFB4,  HANGUL SYLLABLE-BFB4
        '49077' => '13297'            , # G+13297  U+BFB5  uniBFB5,  HANGUL SYLLABLE-BFB5
        '49078' => '13298'            , # G+13298  U+BFB6  uniBFB6,  HANGUL SYLLABLE-BFB6
        '49079' => '13299'            , # G+13299  U+BFB7  uniBFB7,  HANGUL SYLLABLE-BFB7
        '49080' => '13300'            , # G+13300  U+BFB8  uniBFB8,  HANGUL SYLLABLE-BFB8
        '49081' => '13301'            , # G+13301  U+BFB9  uniBFB9,  HANGUL SYLLABLE-BFB9
        '49082' => '13302'            , # G+13302  U+BFBA  uniBFBA,  HANGUL SYLLABLE-BFBA
        '49083' => '13303'            , # G+13303  U+BFBB  uniBFBB,  HANGUL SYLLABLE-BFBB
        '49084' => '13304'            , # G+13304  U+BFBC  uniBFBC,  HANGUL SYLLABLE-BFBC
        '49085' => '13305'            , # G+13305  U+BFBD  uniBFBD,  HANGUL SYLLABLE-BFBD
        '49086' => '13306'            , # G+13306  U+BFBE  uniBFBE,  HANGUL SYLLABLE-BFBE
        '49087' => '13307'            , # G+13307  U+BFBF  uniBFBF,  HANGUL SYLLABLE-BFBF
        '49088' => '13308'            , # G+13308  U+BFC0  uniBFC0,  HANGUL SYLLABLE-BFC0
        '49089' => '13309'            , # G+13309  U+BFC1  uniBFC1,  HANGUL SYLLABLE-BFC1
        '49090' => '13310'            , # G+13310  U+BFC2  uniBFC2,  HANGUL SYLLABLE-BFC2
        '49091' => '13311'            , # G+13311  U+BFC3  uniBFC3,  HANGUL SYLLABLE-BFC3
        '49092' => '13312'            , # G+13312  U+BFC4  uniBFC4,  HANGUL SYLLABLE-BFC4
        '49093' => '2167'             , # G+2167  U+BFC5  uniBFC5,  HANGUL SYLLABLE-BFC5
        '49094' => '13313'            , # G+13313  U+BFC6  uniBFC6,  HANGUL SYLLABLE-BFC6
        '49095' => '13314'            , # G+13314  U+BFC7  uniBFC7,  HANGUL SYLLABLE-BFC7
        '49096' => '13315'            , # G+13315  U+BFC8  uniBFC8,  HANGUL SYLLABLE-BFC8
        '49097' => '13316'            , # G+13316  U+BFC9  uniBFC9,  HANGUL SYLLABLE-BFC9
        '49098' => '13317'            , # G+13317  U+BFCA  uniBFCA,  HANGUL SYLLABLE-BFCA
        '49099' => '13318'            , # G+13318  U+BFCB  uniBFCB,  HANGUL SYLLABLE-BFCB
        '49100' => '2168'             , # G+2168  U+BFCC  uniBFCC,  HANGUL SYLLABLE-BFCC
        '49101' => '2169'             , # G+2169  U+BFCD  uniBFCD,  HANGUL SYLLABLE-BFCD
        '49102' => '13319'            , # G+13319  U+BFCE  uniBFCE,  HANGUL SYLLABLE-BFCE
        '49103' => '13320'            , # G+13320  U+BFCF  uniBFCF,  HANGUL SYLLABLE-BFCF
        '49104' => '2170'             , # G+2170  U+BFD0  uniBFD0,  HANGUL SYLLABLE-BFD0
        '49105' => '13321'            , # G+13321  U+BFD1  uniBFD1,  HANGUL SYLLABLE-BFD1
        '49106' => '13322'            , # G+13322  U+BFD2  uniBFD2,  HANGUL SYLLABLE-BFD2
        '49107' => '13323'            , # G+13323  U+BFD3  uniBFD3,  HANGUL SYLLABLE-BFD3
        '49108' => '2171'             , # G+2171  U+BFD4  uniBFD4,  HANGUL SYLLABLE-BFD4
        '49109' => '13324'            , # G+13324  U+BFD5  uniBFD5,  HANGUL SYLLABLE-BFD5
        '49110' => '13325'            , # G+13325  U+BFD6  uniBFD6,  HANGUL SYLLABLE-BFD6
        '49111' => '13326'            , # G+13326  U+BFD7  uniBFD7,  HANGUL SYLLABLE-BFD7
        '49112' => '13327'            , # G+13327  U+BFD8  uniBFD8,  HANGUL SYLLABLE-BFD8
        '49113' => '13328'            , # G+13328  U+BFD9  uniBFD9,  HANGUL SYLLABLE-BFD9
        '49114' => '13329'            , # G+13329  U+BFDA  uniBFDA,  HANGUL SYLLABLE-BFDA
        '49115' => '13330'            , # G+13330  U+BFDB  uniBFDB,  HANGUL SYLLABLE-BFDB
        '49116' => '2172'             , # G+2172  U+BFDC  uniBFDC,  HANGUL SYLLABLE-BFDC
        '49117' => '13331'            , # G+13331  U+BFDD  uniBFDD,  HANGUL SYLLABLE-BFDD
        '49118' => '13332'            , # G+13332  U+BFDE  uniBFDE,  HANGUL SYLLABLE-BFDE
        '49119' => '2173'             , # G+2173  U+BFDF  uniBFDF,  HANGUL SYLLABLE-BFDF
        '49120' => '13333'            , # G+13333  U+BFE0  uniBFE0,  HANGUL SYLLABLE-BFE0
        '49121' => '2174'             , # G+2174  U+BFE1  uniBFE1,  HANGUL SYLLABLE-BFE1
        '49122' => '13334'            , # G+13334  U+BFE2  uniBFE2,  HANGUL SYLLABLE-BFE2
        '49123' => '13335'            , # G+13335  U+BFE3  uniBFE3,  HANGUL SYLLABLE-BFE3
        '49124' => '13336'            , # G+13336  U+BFE4  uniBFE4,  HANGUL SYLLABLE-BFE4
        '49125' => '13337'            , # G+13337  U+BFE5  uniBFE5,  HANGUL SYLLABLE-BFE5
        '49126' => '13338'            , # G+13338  U+BFE6  uniBFE6,  HANGUL SYLLABLE-BFE6
        '49127' => '13339'            , # G+13339  U+BFE7  uniBFE7,  HANGUL SYLLABLE-BFE7
        '49128' => '13340'            , # G+13340  U+BFE8  uniBFE8,  HANGUL SYLLABLE-BFE8
        '49129' => '13341'            , # G+13341  U+BFE9  uniBFE9,  HANGUL SYLLABLE-BFE9
        '49130' => '13342'            , # G+13342  U+BFEA  uniBFEA,  HANGUL SYLLABLE-BFEA
        '49131' => '13343'            , # G+13343  U+BFEB  uniBFEB,  HANGUL SYLLABLE-BFEB
        '49132' => '13344'            , # G+13344  U+BFEC  uniBFEC,  HANGUL SYLLABLE-BFEC
        '49133' => '13345'            , # G+13345  U+BFED  uniBFED,  HANGUL SYLLABLE-BFED
        '49134' => '13346'            , # G+13346  U+BFEE  uniBFEE,  HANGUL SYLLABLE-BFEE
        '49135' => '13347'            , # G+13347  U+BFEF  uniBFEF,  HANGUL SYLLABLE-BFEF
        '49136' => '13348'            , # G+13348  U+BFF0  uniBFF0,  HANGUL SYLLABLE-BFF0
        '49137' => '13349'            , # G+13349  U+BFF1  uniBFF1,  HANGUL SYLLABLE-BFF1
        '49138' => '13350'            , # G+13350  U+BFF2  uniBFF2,  HANGUL SYLLABLE-BFF2
        '49139' => '13351'            , # G+13351  U+BFF3  uniBFF3,  HANGUL SYLLABLE-BFF3
        '49140' => '13352'            , # G+13352  U+BFF4  uniBFF4,  HANGUL SYLLABLE-BFF4
        '49141' => '13353'            , # G+13353  U+BFF5  uniBFF5,  HANGUL SYLLABLE-BFF5
        '49142' => '13354'            , # G+13354  U+BFF6  uniBFF6,  HANGUL SYLLABLE-BFF6
        '49143' => '13355'            , # G+13355  U+BFF7  uniBFF7,  HANGUL SYLLABLE-BFF7
        '49144' => '13356'            , # G+13356  U+BFF8  uniBFF8,  HANGUL SYLLABLE-BFF8
        '49145' => '13357'            , # G+13357  U+BFF9  uniBFF9,  HANGUL SYLLABLE-BFF9
        '49146' => '13358'            , # G+13358  U+BFFA  uniBFFA,  HANGUL SYLLABLE-BFFA
        '49147' => '13359'            , # G+13359  U+BFFB  uniBFFB,  HANGUL SYLLABLE-BFFB
        '49148' => '13360'            , # G+13360  U+BFFC  uniBFFC,  HANGUL SYLLABLE-BFFC
        '49149' => '13361'            , # G+13361  U+BFFD  uniBFFD,  HANGUL SYLLABLE-BFFD
        '49150' => '13362'            , # G+13362  U+BFFE  uniBFFE,  HANGUL SYLLABLE-BFFE
        '49151' => '13363'            , # G+13363  U+BFFF  uniBFFF,  HANGUL SYLLABLE-BFFF
        '49152' => '13364'            , # G+13364  U+C000  uniC000,  HANGUL SYLLABLE-C000
        '49153' => '13365'            , # G+13365  U+C001  uniC001,  HANGUL SYLLABLE-C001
        '49154' => '13366'            , # G+13366  U+C002  uniC002,  HANGUL SYLLABLE-C002
        '49155' => '13367'            , # G+13367  U+C003  uniC003,  HANGUL SYLLABLE-C003
        '49156' => '13368'            , # G+13368  U+C004  uniC004,  HANGUL SYLLABLE-C004
        '49157' => '13369'            , # G+13369  U+C005  uniC005,  HANGUL SYLLABLE-C005
        '49158' => '13370'            , # G+13370  U+C006  uniC006,  HANGUL SYLLABLE-C006
        '49159' => '13371'            , # G+13371  U+C007  uniC007,  HANGUL SYLLABLE-C007
        '49160' => '13372'            , # G+13372  U+C008  uniC008,  HANGUL SYLLABLE-C008
        '49161' => '13373'            , # G+13373  U+C009  uniC009,  HANGUL SYLLABLE-C009
        '49162' => '13374'            , # G+13374  U+C00A  uniC00A,  HANGUL SYLLABLE-C00A
        '49163' => '13375'            , # G+13375  U+C00B  uniC00B,  HANGUL SYLLABLE-C00B
        '49164' => '13376'            , # G+13376  U+C00C  uniC00C,  HANGUL SYLLABLE-C00C
        '49165' => '13377'            , # G+13377  U+C00D  uniC00D,  HANGUL SYLLABLE-C00D
        '49166' => '13378'            , # G+13378  U+C00E  uniC00E,  HANGUL SYLLABLE-C00E
        '49167' => '13379'            , # G+13379  U+C00F  uniC00F,  HANGUL SYLLABLE-C00F
        '49168' => '13380'            , # G+13380  U+C010  uniC010,  HANGUL SYLLABLE-C010
        '49169' => '13381'            , # G+13381  U+C011  uniC011,  HANGUL SYLLABLE-C011
        '49170' => '13382'            , # G+13382  U+C012  uniC012,  HANGUL SYLLABLE-C012
        '49171' => '13383'            , # G+13383  U+C013  uniC013,  HANGUL SYLLABLE-C013
        '49172' => '13384'            , # G+13384  U+C014  uniC014,  HANGUL SYLLABLE-C014
        '49173' => '13385'            , # G+13385  U+C015  uniC015,  HANGUL SYLLABLE-C015
        '49174' => '13386'            , # G+13386  U+C016  uniC016,  HANGUL SYLLABLE-C016
        '49175' => '13387'            , # G+13387  U+C017  uniC017,  HANGUL SYLLABLE-C017
        '49176' => '13388'            , # G+13388  U+C018  uniC018,  HANGUL SYLLABLE-C018
        '49177' => '13389'            , # G+13389  U+C019  uniC019,  HANGUL SYLLABLE-C019
        '49178' => '13390'            , # G+13390  U+C01A  uniC01A,  HANGUL SYLLABLE-C01A
        '49179' => '13391'            , # G+13391  U+C01B  uniC01B,  HANGUL SYLLABLE-C01B
        '49180' => '13392'            , # G+13392  U+C01C  uniC01C,  HANGUL SYLLABLE-C01C
        '49181' => '13393'            , # G+13393  U+C01D  uniC01D,  HANGUL SYLLABLE-C01D
        '49182' => '13394'            , # G+13394  U+C01E  uniC01E,  HANGUL SYLLABLE-C01E
        '49183' => '13395'            , # G+13395  U+C01F  uniC01F,  HANGUL SYLLABLE-C01F
        '49184' => '13396'            , # G+13396  U+C020  uniC020,  HANGUL SYLLABLE-C020
        '49185' => '13397'            , # G+13397  U+C021  uniC021,  HANGUL SYLLABLE-C021
        '49186' => '13398'            , # G+13398  U+C022  uniC022,  HANGUL SYLLABLE-C022
        '49187' => '13399'            , # G+13399  U+C023  uniC023,  HANGUL SYLLABLE-C023
        '49188' => '13400'            , # G+13400  U+C024  uniC024,  HANGUL SYLLABLE-C024
        '49189' => '13401'            , # G+13401  U+C025  uniC025,  HANGUL SYLLABLE-C025
        '49190' => '13402'            , # G+13402  U+C026  uniC026,  HANGUL SYLLABLE-C026
        '49191' => '13403'            , # G+13403  U+C027  uniC027,  HANGUL SYLLABLE-C027
        '49192' => '13404'            , # G+13404  U+C028  uniC028,  HANGUL SYLLABLE-C028
        '49193' => '13405'            , # G+13405  U+C029  uniC029,  HANGUL SYLLABLE-C029
        '49194' => '13406'            , # G+13406  U+C02A  uniC02A,  HANGUL SYLLABLE-C02A
        '49195' => '13407'            , # G+13407  U+C02B  uniC02B,  HANGUL SYLLABLE-C02B
        '49196' => '13408'            , # G+13408  U+C02C  uniC02C,  HANGUL SYLLABLE-C02C
        '49197' => '13409'            , # G+13409  U+C02D  uniC02D,  HANGUL SYLLABLE-C02D
        '49198' => '13410'            , # G+13410  U+C02E  uniC02E,  HANGUL SYLLABLE-C02E
        '49199' => '13411'            , # G+13411  U+C02F  uniC02F,  HANGUL SYLLABLE-C02F
        '49200' => '13412'            , # G+13412  U+C030  uniC030,  HANGUL SYLLABLE-C030
        '49201' => '13413'            , # G+13413  U+C031  uniC031,  HANGUL SYLLABLE-C031
        '49202' => '13414'            , # G+13414  U+C032  uniC032,  HANGUL SYLLABLE-C032
        '49203' => '13415'            , # G+13415  U+C033  uniC033,  HANGUL SYLLABLE-C033
        '49204' => '13416'            , # G+13416  U+C034  uniC034,  HANGUL SYLLABLE-C034
        '49205' => '13417'            , # G+13417  U+C035  uniC035,  HANGUL SYLLABLE-C035
        '49206' => '13418'            , # G+13418  U+C036  uniC036,  HANGUL SYLLABLE-C036
        '49207' => '13419'            , # G+13419  U+C037  uniC037,  HANGUL SYLLABLE-C037
        '49208' => '13420'            , # G+13420  U+C038  uniC038,  HANGUL SYLLABLE-C038
        '49209' => '13421'            , # G+13421  U+C039  uniC039,  HANGUL SYLLABLE-C039
        '49210' => '13422'            , # G+13422  U+C03A  uniC03A,  HANGUL SYLLABLE-C03A
        '49211' => '13423'            , # G+13423  U+C03B  uniC03B,  HANGUL SYLLABLE-C03B
        '49212' => '2175'             , # G+2175  U+C03C  uniC03C,  HANGUL SYLLABLE-C03C
        '49213' => '13424'            , # G+13424  U+C03D  uniC03D,  HANGUL SYLLABLE-C03D
        '49214' => '13425'            , # G+13425  U+C03E  uniC03E,  HANGUL SYLLABLE-C03E
        '49215' => '13426'            , # G+13426  U+C03F  uniC03F,  HANGUL SYLLABLE-C03F
        '49216' => '13427'            , # G+13427  U+C040  uniC040,  HANGUL SYLLABLE-C040
        '49217' => '13428'            , # G+13428  U+C041  uniC041,  HANGUL SYLLABLE-C041
        '49218' => '13429'            , # G+13429  U+C042  uniC042,  HANGUL SYLLABLE-C042
        '49219' => '13430'            , # G+13430  U+C043  uniC043,  HANGUL SYLLABLE-C043
        '49220' => '13431'            , # G+13431  U+C044  uniC044,  HANGUL SYLLABLE-C044
        '49221' => '13432'            , # G+13432  U+C045  uniC045,  HANGUL SYLLABLE-C045
        '49222' => '13433'            , # G+13433  U+C046  uniC046,  HANGUL SYLLABLE-C046
        '49223' => '13434'            , # G+13434  U+C047  uniC047,  HANGUL SYLLABLE-C047
        '49224' => '13435'            , # G+13435  U+C048  uniC048,  HANGUL SYLLABLE-C048
        '49225' => '13436'            , # G+13436  U+C049  uniC049,  HANGUL SYLLABLE-C049
        '49226' => '13437'            , # G+13437  U+C04A  uniC04A,  HANGUL SYLLABLE-C04A
        '49227' => '13438'            , # G+13438  U+C04B  uniC04B,  HANGUL SYLLABLE-C04B
        '49228' => '13439'            , # G+13439  U+C04C  uniC04C,  HANGUL SYLLABLE-C04C
        '49229' => '13440'            , # G+13440  U+C04D  uniC04D,  HANGUL SYLLABLE-C04D
        '49230' => '13441'            , # G+13441  U+C04E  uniC04E,  HANGUL SYLLABLE-C04E
        '49231' => '13442'            , # G+13442  U+C04F  uniC04F,  HANGUL SYLLABLE-C04F
        '49232' => '13443'            , # G+13443  U+C050  uniC050,  HANGUL SYLLABLE-C050
        '49233' => '2176'             , # G+2176  U+C051  uniC051,  HANGUL SYLLABLE-C051
        '49234' => '13444'            , # G+13444  U+C052  uniC052,  HANGUL SYLLABLE-C052
        '49235' => '13445'            , # G+13445  U+C053  uniC053,  HANGUL SYLLABLE-C053
        '49236' => '13446'            , # G+13446  U+C054  uniC054,  HANGUL SYLLABLE-C054
        '49237' => '13447'            , # G+13447  U+C055  uniC055,  HANGUL SYLLABLE-C055
        '49238' => '13448'            , # G+13448  U+C056  uniC056,  HANGUL SYLLABLE-C056
        '49239' => '13449'            , # G+13449  U+C057  uniC057,  HANGUL SYLLABLE-C057
        '49240' => '2177'             , # G+2177  U+C058  uniC058,  HANGUL SYLLABLE-C058
        '49241' => '13450'            , # G+13450  U+C059  uniC059,  HANGUL SYLLABLE-C059
        '49242' => '13451'            , # G+13451  U+C05A  uniC05A,  HANGUL SYLLABLE-C05A
        '49243' => '13452'            , # G+13452  U+C05B  uniC05B,  HANGUL SYLLABLE-C05B
        '49244' => '2178'             , # G+2178  U+C05C  uniC05C,  HANGUL SYLLABLE-C05C
        '49245' => '13453'            , # G+13453  U+C05D  uniC05D,  HANGUL SYLLABLE-C05D
        '49246' => '13454'            , # G+13454  U+C05E  uniC05E,  HANGUL SYLLABLE-C05E
        '49247' => '13455'            , # G+13455  U+C05F  uniC05F,  HANGUL SYLLABLE-C05F
        '49248' => '2179'             , # G+2179  U+C060  uniC060,  HANGUL SYLLABLE-C060
        '49249' => '13456'            , # G+13456  U+C061  uniC061,  HANGUL SYLLABLE-C061
        '49250' => '13457'            , # G+13457  U+C062  uniC062,  HANGUL SYLLABLE-C062
        '49251' => '13458'            , # G+13458  U+C063  uniC063,  HANGUL SYLLABLE-C063
        '49252' => '13459'            , # G+13459  U+C064  uniC064,  HANGUL SYLLABLE-C064
        '49253' => '13460'            , # G+13460  U+C065  uniC065,  HANGUL SYLLABLE-C065
        '49254' => '13461'            , # G+13461  U+C066  uniC066,  HANGUL SYLLABLE-C066
        '49255' => '13462'            , # G+13462  U+C067  uniC067,  HANGUL SYLLABLE-C067
        '49256' => '2180'             , # G+2180  U+C068  uniC068,  HANGUL SYLLABLE-C068
        '49257' => '2181'             , # G+2181  U+C069  uniC069,  HANGUL SYLLABLE-C069
        '49258' => '13463'            , # G+13463  U+C06A  uniC06A,  HANGUL SYLLABLE-C06A
        '49259' => '13464'            , # G+13464  U+C06B  uniC06B,  HANGUL SYLLABLE-C06B
        '49260' => '13465'            , # G+13465  U+C06C  uniC06C,  HANGUL SYLLABLE-C06C
        '49261' => '13466'            , # G+13466  U+C06D  uniC06D,  HANGUL SYLLABLE-C06D
        '49262' => '13467'            , # G+13467  U+C06E  uniC06E,  HANGUL SYLLABLE-C06E
        '49263' => '13468'            , # G+13468  U+C06F  uniC06F,  HANGUL SYLLABLE-C06F
        '49264' => '13469'            , # G+13469  U+C070  uniC070,  HANGUL SYLLABLE-C070
        '49265' => '13470'            , # G+13470  U+C071  uniC071,  HANGUL SYLLABLE-C071
        '49266' => '13471'            , # G+13471  U+C072  uniC072,  HANGUL SYLLABLE-C072
        '49267' => '13472'            , # G+13472  U+C073  uniC073,  HANGUL SYLLABLE-C073
        '49268' => '13473'            , # G+13473  U+C074  uniC074,  HANGUL SYLLABLE-C074
        '49269' => '13474'            , # G+13474  U+C075  uniC075,  HANGUL SYLLABLE-C075
        '49270' => '13475'            , # G+13475  U+C076  uniC076,  HANGUL SYLLABLE-C076
        '49271' => '13476'            , # G+13476  U+C077  uniC077,  HANGUL SYLLABLE-C077
        '49272' => '13477'            , # G+13477  U+C078  uniC078,  HANGUL SYLLABLE-C078
        '49273' => '13478'            , # G+13478  U+C079  uniC079,  HANGUL SYLLABLE-C079
        '49274' => '13479'            , # G+13479  U+C07A  uniC07A,  HANGUL SYLLABLE-C07A
        '49275' => '13480'            , # G+13480  U+C07B  uniC07B,  HANGUL SYLLABLE-C07B
        '49276' => '13481'            , # G+13481  U+C07C  uniC07C,  HANGUL SYLLABLE-C07C
        '49277' => '13482'            , # G+13482  U+C07D  uniC07D,  HANGUL SYLLABLE-C07D
        '49278' => '13483'            , # G+13483  U+C07E  uniC07E,  HANGUL SYLLABLE-C07E
        '49279' => '13484'            , # G+13484  U+C07F  uniC07F,  HANGUL SYLLABLE-C07F
        '49280' => '13485'            , # G+13485  U+C080  uniC080,  HANGUL SYLLABLE-C080
        '49281' => '13486'            , # G+13486  U+C081  uniC081,  HANGUL SYLLABLE-C081
        '49282' => '13487'            , # G+13487  U+C082  uniC082,  HANGUL SYLLABLE-C082
        '49283' => '13488'            , # G+13488  U+C083  uniC083,  HANGUL SYLLABLE-C083
        '49284' => '13489'            , # G+13489  U+C084  uniC084,  HANGUL SYLLABLE-C084
        '49285' => '13490'            , # G+13490  U+C085  uniC085,  HANGUL SYLLABLE-C085
        '49286' => '13491'            , # G+13491  U+C086  uniC086,  HANGUL SYLLABLE-C086
        '49287' => '13492'            , # G+13492  U+C087  uniC087,  HANGUL SYLLABLE-C087
        '49288' => '13493'            , # G+13493  U+C088  uniC088,  HANGUL SYLLABLE-C088
        '49289' => '13494'            , # G+13494  U+C089  uniC089,  HANGUL SYLLABLE-C089
        '49290' => '13495'            , # G+13495  U+C08A  uniC08A,  HANGUL SYLLABLE-C08A
        '49291' => '13496'            , # G+13496  U+C08B  uniC08B,  HANGUL SYLLABLE-C08B
        '49292' => '13497'            , # G+13497  U+C08C  uniC08C,  HANGUL SYLLABLE-C08C
        '49293' => '13498'            , # G+13498  U+C08D  uniC08D,  HANGUL SYLLABLE-C08D
        '49294' => '13499'            , # G+13499  U+C08E  uniC08E,  HANGUL SYLLABLE-C08E
        '49295' => '13500'            , # G+13500  U+C08F  uniC08F,  HANGUL SYLLABLE-C08F
        '49296' => '2182'             , # G+2182  U+C090  uniC090,  HANGUL SYLLABLE-C090
        '49297' => '2183'             , # G+2183  U+C091  uniC091,  HANGUL SYLLABLE-C091
        '49298' => '13501'            , # G+13501  U+C092  uniC092,  HANGUL SYLLABLE-C092
        '49299' => '13502'            , # G+13502  U+C093  uniC093,  HANGUL SYLLABLE-C093
        '49300' => '2184'             , # G+2184  U+C094  uniC094,  HANGUL SYLLABLE-C094
        '49301' => '13503'            , # G+13503  U+C095  uniC095,  HANGUL SYLLABLE-C095
        '49302' => '13504'            , # G+13504  U+C096  uniC096,  HANGUL SYLLABLE-C096
        '49303' => '13505'            , # G+13505  U+C097  uniC097,  HANGUL SYLLABLE-C097
        '49304' => '2185'             , # G+2185  U+C098  uniC098,  HANGUL SYLLABLE-C098
        '49305' => '13506'            , # G+13506  U+C099  uniC099,  HANGUL SYLLABLE-C099
        '49306' => '13507'            , # G+13507  U+C09A  uniC09A,  HANGUL SYLLABLE-C09A
        '49307' => '13508'            , # G+13508  U+C09B  uniC09B,  HANGUL SYLLABLE-C09B
        '49308' => '13509'            , # G+13509  U+C09C  uniC09C,  HANGUL SYLLABLE-C09C
        '49309' => '13510'            , # G+13510  U+C09D  uniC09D,  HANGUL SYLLABLE-C09D
        '49310' => '13511'            , # G+13511  U+C09E  uniC09E,  HANGUL SYLLABLE-C09E
        '49311' => '13512'            , # G+13512  U+C09F  uniC09F,  HANGUL SYLLABLE-C09F
        '49312' => '2186'             , # G+2186  U+C0A0  uniC0A0,  HANGUL SYLLABLE-C0A0
        '49313' => '2187'             , # G+2187  U+C0A1  uniC0A1,  HANGUL SYLLABLE-C0A1
        '49314' => '13513'            , # G+13513  U+C0A2  uniC0A2,  HANGUL SYLLABLE-C0A2
        '49315' => '2188'             , # G+2188  U+C0A3  uniC0A3,  HANGUL SYLLABLE-C0A3
        '49316' => '13514'            , # G+13514  U+C0A4  uniC0A4,  HANGUL SYLLABLE-C0A4
        '49317' => '2189'             , # G+2189  U+C0A5  uniC0A5,  HANGUL SYLLABLE-C0A5
        '49318' => '13515'            , # G+13515  U+C0A6  uniC0A6,  HANGUL SYLLABLE-C0A6
        '49319' => '13516'            , # G+13516  U+C0A7  uniC0A7,  HANGUL SYLLABLE-C0A7
        '49320' => '13517'            , # G+13517  U+C0A8  uniC0A8,  HANGUL SYLLABLE-C0A8
        '49321' => '13518'            , # G+13518  U+C0A9  uniC0A9,  HANGUL SYLLABLE-C0A9
        '49322' => '13519'            , # G+13519  U+C0AA  uniC0AA,  HANGUL SYLLABLE-C0AA
        '49323' => '13520'            , # G+13520  U+C0AB  uniC0AB,  HANGUL SYLLABLE-C0AB
        '49324' => '2190'             , # G+2190  U+C0AC  uniC0AC,  HANGUL SYLLABLE-C0AC
        '49325' => '2191'             , # G+2191  U+C0AD  uniC0AD,  HANGUL SYLLABLE-C0AD
        '49326' => '13521'            , # G+13521  U+C0AE  uniC0AE,  HANGUL SYLLABLE-C0AE
        '49327' => '2192'             , # G+2192  U+C0AF  uniC0AF,  HANGUL SYLLABLE-C0AF
        '49328' => '2193'             , # G+2193  U+C0B0  uniC0B0,  HANGUL SYLLABLE-C0B0
        '49329' => '13522'            , # G+13522  U+C0B1  uniC0B1,  HANGUL SYLLABLE-C0B1
        '49330' => '13523'            , # G+13523  U+C0B2  uniC0B2,  HANGUL SYLLABLE-C0B2
        '49331' => '2194'             , # G+2194  U+C0B3  uniC0B3,  HANGUL SYLLABLE-C0B3
        '49332' => '2195'             , # G+2195  U+C0B4  uniC0B4,  HANGUL SYLLABLE-C0B4
        '49333' => '2196'             , # G+2196  U+C0B5  uniC0B5,  HANGUL SYLLABLE-C0B5
        '49334' => '2197'             , # G+2197  U+C0B6  uniC0B6,  HANGUL SYLLABLE-C0B6
        '49335' => '13524'            , # G+13524  U+C0B7  uniC0B7,  HANGUL SYLLABLE-C0B7
        '49336' => '13525'            , # G+13525  U+C0B8  uniC0B8,  HANGUL SYLLABLE-C0B8
        '49337' => '13526'            , # G+13526  U+C0B9  uniC0B9,  HANGUL SYLLABLE-C0B9
        '49338' => '13527'            , # G+13527  U+C0BA  uniC0BA,  HANGUL SYLLABLE-C0BA
        '49339' => '13528'            , # G+13528  U+C0BB  uniC0BB,  HANGUL SYLLABLE-C0BB
        '49340' => '2198'             , # G+2198  U+C0BC  uniC0BC,  HANGUL SYLLABLE-C0BC
        '49341' => '2199'             , # G+2199  U+C0BD  uniC0BD,  HANGUL SYLLABLE-C0BD
        '49342' => '13529'            , # G+13529  U+C0BE  uniC0BE,  HANGUL SYLLABLE-C0BE
        '49343' => '2200'             , # G+2200  U+C0BF  uniC0BF,  HANGUL SYLLABLE-C0BF
        '49344' => '2201'             , # G+2201  U+C0C0  uniC0C0,  HANGUL SYLLABLE-C0C0
        '49345' => '2202'             , # G+2202  U+C0C1  uniC0C1,  HANGUL SYLLABLE-C0C1
        '49346' => '13530'            , # G+13530  U+C0C2  uniC0C2,  HANGUL SYLLABLE-C0C2
        '49347' => '13531'            , # G+13531  U+C0C3  uniC0C3,  HANGUL SYLLABLE-C0C3
        '49348' => '13532'            , # G+13532  U+C0C4  uniC0C4,  HANGUL SYLLABLE-C0C4
        '49349' => '2203'             , # G+2203  U+C0C5  uniC0C5,  HANGUL SYLLABLE-C0C5
        '49350' => '13533'            , # G+13533  U+C0C6  uniC0C6,  HANGUL SYLLABLE-C0C6
        '49351' => '13534'            , # G+13534  U+C0C7  uniC0C7,  HANGUL SYLLABLE-C0C7
        '49352' => '2204'             , # G+2204  U+C0C8  uniC0C8,  HANGUL SYLLABLE-C0C8
        '49353' => '2205'             , # G+2205  U+C0C9  uniC0C9,  HANGUL SYLLABLE-C0C9
        '49354' => '13535'            , # G+13535  U+C0CA  uniC0CA,  HANGUL SYLLABLE-C0CA
        '49355' => '13536'            , # G+13536  U+C0CB  uniC0CB,  HANGUL SYLLABLE-C0CB
        '49356' => '2206'             , # G+2206  U+C0CC  uniC0CC,  HANGUL SYLLABLE-C0CC
        '49357' => '13537'            , # G+13537  U+C0CD  uniC0CD,  HANGUL SYLLABLE-C0CD
        '49358' => '13538'            , # G+13538  U+C0CE  uniC0CE,  HANGUL SYLLABLE-C0CE
        '49359' => '13539'            , # G+13539  U+C0CF  uniC0CF,  HANGUL SYLLABLE-C0CF
        '49360' => '2207'             , # G+2207  U+C0D0  uniC0D0,  HANGUL SYLLABLE-C0D0
        '49361' => '13540'            , # G+13540  U+C0D1  uniC0D1,  HANGUL SYLLABLE-C0D1
        '49362' => '13541'            , # G+13541  U+C0D2  uniC0D2,  HANGUL SYLLABLE-C0D2
        '49363' => '13542'            , # G+13542  U+C0D3  uniC0D3,  HANGUL SYLLABLE-C0D3
        '49364' => '13543'            , # G+13543  U+C0D4  uniC0D4,  HANGUL SYLLABLE-C0D4
        '49365' => '13544'            , # G+13544  U+C0D5  uniC0D5,  HANGUL SYLLABLE-C0D5
        '49366' => '13545'            , # G+13545  U+C0D6  uniC0D6,  HANGUL SYLLABLE-C0D6
        '49367' => '13546'            , # G+13546  U+C0D7  uniC0D7,  HANGUL SYLLABLE-C0D7
        '49368' => '2208'             , # G+2208  U+C0D8  uniC0D8,  HANGUL SYLLABLE-C0D8
        '49369' => '2209'             , # G+2209  U+C0D9  uniC0D9,  HANGUL SYLLABLE-C0D9
        '49370' => '13547'            , # G+13547  U+C0DA  uniC0DA,  HANGUL SYLLABLE-C0DA
        '49371' => '2210'             , # G+2210  U+C0DB  uniC0DB,  HANGUL SYLLABLE-C0DB
        '49372' => '2211'             , # G+2211  U+C0DC  uniC0DC,  HANGUL SYLLABLE-C0DC
        '49373' => '2212'             , # G+2212  U+C0DD  uniC0DD,  HANGUL SYLLABLE-C0DD
        '49374' => '13548'            , # G+13548  U+C0DE  uniC0DE,  HANGUL SYLLABLE-C0DE
        '49375' => '13549'            , # G+13549  U+C0DF  uniC0DF,  HANGUL SYLLABLE-C0DF
        '49376' => '13550'            , # G+13550  U+C0E0  uniC0E0,  HANGUL SYLLABLE-C0E0
        '49377' => '13551'            , # G+13551  U+C0E1  uniC0E1,  HANGUL SYLLABLE-C0E1
        '49378' => '13552'            , # G+13552  U+C0E2  uniC0E2,  HANGUL SYLLABLE-C0E2
        '49379' => '13553'            , # G+13553  U+C0E3  uniC0E3,  HANGUL SYLLABLE-C0E3
        '49380' => '2213'             , # G+2213  U+C0E4  uniC0E4,  HANGUL SYLLABLE-C0E4
        '49381' => '2214'             , # G+2214  U+C0E5  uniC0E5,  HANGUL SYLLABLE-C0E5
        '49382' => '13554'            , # G+13554  U+C0E6  uniC0E6,  HANGUL SYLLABLE-C0E6
        '49383' => '13555'            , # G+13555  U+C0E7  uniC0E7,  HANGUL SYLLABLE-C0E7
        '49384' => '2215'             , # G+2215  U+C0E8  uniC0E8,  HANGUL SYLLABLE-C0E8
        '49385' => '13556'            , # G+13556  U+C0E9  uniC0E9,  HANGUL SYLLABLE-C0E9
        '49386' => '13557'            , # G+13557  U+C0EA  uniC0EA,  HANGUL SYLLABLE-C0EA
        '49387' => '13558'            , # G+13558  U+C0EB  uniC0EB,  HANGUL SYLLABLE-C0EB
        '49388' => '2216'             , # G+2216  U+C0EC  uniC0EC,  HANGUL SYLLABLE-C0EC
        '49389' => '13559'            , # G+13559  U+C0ED  uniC0ED,  HANGUL SYLLABLE-C0ED
        '49390' => '13560'            , # G+13560  U+C0EE  uniC0EE,  HANGUL SYLLABLE-C0EE
        '49391' => '13561'            , # G+13561  U+C0EF  uniC0EF,  HANGUL SYLLABLE-C0EF
        '49392' => '13562'            , # G+13562  U+C0F0  uniC0F0,  HANGUL SYLLABLE-C0F0
        '49393' => '13563'            , # G+13563  U+C0F1  uniC0F1,  HANGUL SYLLABLE-C0F1
        '49394' => '13564'            , # G+13564  U+C0F2  uniC0F2,  HANGUL SYLLABLE-C0F2
        '49395' => '13565'            , # G+13565  U+C0F3  uniC0F3,  HANGUL SYLLABLE-C0F3
        '49396' => '2217'             , # G+2217  U+C0F4  uniC0F4,  HANGUL SYLLABLE-C0F4
        '49397' => '2218'             , # G+2218  U+C0F5  uniC0F5,  HANGUL SYLLABLE-C0F5
        '49398' => '13566'            , # G+13566  U+C0F6  uniC0F6,  HANGUL SYLLABLE-C0F6
        '49399' => '2219'             , # G+2219  U+C0F7  uniC0F7,  HANGUL SYLLABLE-C0F7
        '49400' => '13567'            , # G+13567  U+C0F8  uniC0F8,  HANGUL SYLLABLE-C0F8
        '49401' => '2220'             , # G+2220  U+C0F9  uniC0F9,  HANGUL SYLLABLE-C0F9
        '49402' => '13568'            , # G+13568  U+C0FA  uniC0FA,  HANGUL SYLLABLE-C0FA
        '49403' => '13569'            , # G+13569  U+C0FB  uniC0FB,  HANGUL SYLLABLE-C0FB
        '49404' => '13570'            , # G+13570  U+C0FC  uniC0FC,  HANGUL SYLLABLE-C0FC
        '49405' => '13571'            , # G+13571  U+C0FD  uniC0FD,  HANGUL SYLLABLE-C0FD
        '49406' => '13572'            , # G+13572  U+C0FE  uniC0FE,  HANGUL SYLLABLE-C0FE
        '49407' => '13573'            , # G+13573  U+C0FF  uniC0FF,  HANGUL SYLLABLE-C0FF
        '49408' => '2221'             , # G+2221  U+C100  uniC100,  HANGUL SYLLABLE-C100
        '49409' => '13574'            , # G+13574  U+C101  uniC101,  HANGUL SYLLABLE-C101
        '49410' => '13575'            , # G+13575  U+C102  uniC102,  HANGUL SYLLABLE-C102
        '49411' => '13576'            , # G+13576  U+C103  uniC103,  HANGUL SYLLABLE-C103
        '49412' => '2222'             , # G+2222  U+C104  uniC104,  HANGUL SYLLABLE-C104
        '49413' => '13577'            , # G+13577  U+C105  uniC105,  HANGUL SYLLABLE-C105
        '49414' => '13578'            , # G+13578  U+C106  uniC106,  HANGUL SYLLABLE-C106
        '49415' => '13579'            , # G+13579  U+C107  uniC107,  HANGUL SYLLABLE-C107
        '49416' => '2223'             , # G+2223  U+C108  uniC108,  HANGUL SYLLABLE-C108
        '49417' => '13580'            , # G+13580  U+C109  uniC109,  HANGUL SYLLABLE-C109
        '49418' => '13581'            , # G+13581  U+C10A  uniC10A,  HANGUL SYLLABLE-C10A
        '49419' => '13582'            , # G+13582  U+C10B  uniC10B,  HANGUL SYLLABLE-C10B
        '49420' => '13583'            , # G+13583  U+C10C  uniC10C,  HANGUL SYLLABLE-C10C
        '49421' => '13584'            , # G+13584  U+C10D  uniC10D,  HANGUL SYLLABLE-C10D
        '49422' => '13585'            , # G+13585  U+C10E  uniC10E,  HANGUL SYLLABLE-C10E
        '49423' => '13586'            , # G+13586  U+C10F  uniC10F,  HANGUL SYLLABLE-C10F
        '49424' => '2224'             , # G+2224  U+C110  uniC110,  HANGUL SYLLABLE-C110
        '49425' => '13587'            , # G+13587  U+C111  uniC111,  HANGUL SYLLABLE-C111
        '49426' => '13588'            , # G+13588  U+C112  uniC112,  HANGUL SYLLABLE-C112
        '49427' => '13589'            , # G+13589  U+C113  uniC113,  HANGUL SYLLABLE-C113
        '49428' => '13590'            , # G+13590  U+C114  uniC114,  HANGUL SYLLABLE-C114
        '49429' => '2225'             , # G+2225  U+C115  uniC115,  HANGUL SYLLABLE-C115
        '49430' => '13591'            , # G+13591  U+C116  uniC116,  HANGUL SYLLABLE-C116
        '49431' => '13592'            , # G+13592  U+C117  uniC117,  HANGUL SYLLABLE-C117
        '49432' => '13593'            , # G+13593  U+C118  uniC118,  HANGUL SYLLABLE-C118
        '49433' => '13594'            , # G+13594  U+C119  uniC119,  HANGUL SYLLABLE-C119
        '49434' => '13595'            , # G+13595  U+C11A  uniC11A,  HANGUL SYLLABLE-C11A
        '49435' => '13596'            , # G+13596  U+C11B  uniC11B,  HANGUL SYLLABLE-C11B
        '49436' => '2226'             , # G+2226  U+C11C  uniC11C,  HANGUL SYLLABLE-C11C
        '49437' => '2227'             , # G+2227  U+C11D  uniC11D,  HANGUL SYLLABLE-C11D
        '49438' => '2228'             , # G+2228  U+C11E  uniC11E,  HANGUL SYLLABLE-C11E
        '49439' => '2229'             , # G+2229  U+C11F  uniC11F,  HANGUL SYLLABLE-C11F
        '49440' => '2230'             , # G+2230  U+C120  uniC120,  HANGUL SYLLABLE-C120
        '49441' => '13597'            , # G+13597  U+C121  uniC121,  HANGUL SYLLABLE-C121
        '49442' => '13598'            , # G+13598  U+C122  uniC122,  HANGUL SYLLABLE-C122
        '49443' => '2231'             , # G+2231  U+C123  uniC123,  HANGUL SYLLABLE-C123
        '49444' => '2232'             , # G+2232  U+C124  uniC124,  HANGUL SYLLABLE-C124
        '49445' => '13599'            , # G+13599  U+C125  uniC125,  HANGUL SYLLABLE-C125
        '49446' => '2233'             , # G+2233  U+C126  uniC126,  HANGUL SYLLABLE-C126
        '49447' => '2234'             , # G+2234  U+C127  uniC127,  HANGUL SYLLABLE-C127
        '49448' => '13600'            , # G+13600  U+C128  uniC128,  HANGUL SYLLABLE-C128
        '49449' => '13601'            , # G+13601  U+C129  uniC129,  HANGUL SYLLABLE-C129
        '49450' => '13602'            , # G+13602  U+C12A  uniC12A,  HANGUL SYLLABLE-C12A
        '49451' => '13603'            , # G+13603  U+C12B  uniC12B,  HANGUL SYLLABLE-C12B
        '49452' => '2235'             , # G+2235  U+C12C  uniC12C,  HANGUL SYLLABLE-C12C
        '49453' => '2236'             , # G+2236  U+C12D  uniC12D,  HANGUL SYLLABLE-C12D
        '49454' => '13604'            , # G+13604  U+C12E  uniC12E,  HANGUL SYLLABLE-C12E
        '49455' => '2237'             , # G+2237  U+C12F  uniC12F,  HANGUL SYLLABLE-C12F
        '49456' => '2238'             , # G+2238  U+C130  uniC130,  HANGUL SYLLABLE-C130
        '49457' => '2239'             , # G+2239  U+C131  uniC131,  HANGUL SYLLABLE-C131
        '49458' => '13605'            , # G+13605  U+C132  uniC132,  HANGUL SYLLABLE-C132
        '49459' => '13606'            , # G+13606  U+C133  uniC133,  HANGUL SYLLABLE-C133
        '49460' => '13607'            , # G+13607  U+C134  uniC134,  HANGUL SYLLABLE-C134
        '49461' => '13608'            , # G+13608  U+C135  uniC135,  HANGUL SYLLABLE-C135
        '49462' => '2240'             , # G+2240  U+C136  uniC136,  HANGUL SYLLABLE-C136
        '49463' => '13609'            , # G+13609  U+C137  uniC137,  HANGUL SYLLABLE-C137
        '49464' => '2241'             , # G+2241  U+C138  uniC138,  HANGUL SYLLABLE-C138
        '49465' => '2242'             , # G+2242  U+C139  uniC139,  HANGUL SYLLABLE-C139
        '49466' => '13610'            , # G+13610  U+C13A  uniC13A,  HANGUL SYLLABLE-C13A
        '49467' => '13611'            , # G+13611  U+C13B  uniC13B,  HANGUL SYLLABLE-C13B
        '49468' => '2243'             , # G+2243  U+C13C  uniC13C,  HANGUL SYLLABLE-C13C
        '49469' => '13612'            , # G+13612  U+C13D  uniC13D,  HANGUL SYLLABLE-C13D
        '49470' => '13613'            , # G+13613  U+C13E  uniC13E,  HANGUL SYLLABLE-C13E
        '49471' => '13614'            , # G+13614  U+C13F  uniC13F,  HANGUL SYLLABLE-C13F
        '49472' => '2244'             , # G+2244  U+C140  uniC140,  HANGUL SYLLABLE-C140
        '49473' => '13615'            , # G+13615  U+C141  uniC141,  HANGUL SYLLABLE-C141
        '49474' => '13616'            , # G+13616  U+C142  uniC142,  HANGUL SYLLABLE-C142
        '49475' => '13617'            , # G+13617  U+C143  uniC143,  HANGUL SYLLABLE-C143
        '49476' => '13618'            , # G+13618  U+C144  uniC144,  HANGUL SYLLABLE-C144
        '49477' => '13619'            , # G+13619  U+C145  uniC145,  HANGUL SYLLABLE-C145
        '49478' => '13620'            , # G+13620  U+C146  uniC146,  HANGUL SYLLABLE-C146
        '49479' => '13621'            , # G+13621  U+C147  uniC147,  HANGUL SYLLABLE-C147
        '49480' => '2245'             , # G+2245  U+C148  uniC148,  HANGUL SYLLABLE-C148
        '49481' => '2246'             , # G+2246  U+C149  uniC149,  HANGUL SYLLABLE-C149
        '49482' => '13622'            , # G+13622  U+C14A  uniC14A,  HANGUL SYLLABLE-C14A
        '49483' => '2247'             , # G+2247  U+C14B  uniC14B,  HANGUL SYLLABLE-C14B
        '49484' => '2248'             , # G+2248  U+C14C  uniC14C,  HANGUL SYLLABLE-C14C
        '49485' => '2249'             , # G+2249  U+C14D  uniC14D,  HANGUL SYLLABLE-C14D
        '49486' => '13623'            , # G+13623  U+C14E  uniC14E,  HANGUL SYLLABLE-C14E
        '49487' => '13624'            , # G+13624  U+C14F  uniC14F,  HANGUL SYLLABLE-C14F
        '49488' => '13625'            , # G+13625  U+C150  uniC150,  HANGUL SYLLABLE-C150
        '49489' => '13626'            , # G+13626  U+C151  uniC151,  HANGUL SYLLABLE-C151
        '49490' => '13627'            , # G+13627  U+C152  uniC152,  HANGUL SYLLABLE-C152
        '49491' => '13628'            , # G+13628  U+C153  uniC153,  HANGUL SYLLABLE-C153
        '49492' => '2250'             , # G+2250  U+C154  uniC154,  HANGUL SYLLABLE-C154
        '49493' => '2251'             , # G+2251  U+C155  uniC155,  HANGUL SYLLABLE-C155
        '49494' => '13629'            , # G+13629  U+C156  uniC156,  HANGUL SYLLABLE-C156
        '49495' => '13630'            , # G+13630  U+C157  uniC157,  HANGUL SYLLABLE-C157
        '49496' => '2252'             , # G+2252  U+C158  uniC158,  HANGUL SYLLABLE-C158
        '49497' => '13631'            , # G+13631  U+C159  uniC159,  HANGUL SYLLABLE-C159
        '49498' => '13632'            , # G+13632  U+C15A  uniC15A,  HANGUL SYLLABLE-C15A
        '49499' => '13633'            , # G+13633  U+C15B  uniC15B,  HANGUL SYLLABLE-C15B
        '49500' => '2253'             , # G+2253  U+C15C  uniC15C,  HANGUL SYLLABLE-C15C
        '49501' => '13634'            , # G+13634  U+C15D  uniC15D,  HANGUL SYLLABLE-C15D
        '49502' => '13635'            , # G+13635  U+C15E  uniC15E,  HANGUL SYLLABLE-C15E
        '49503' => '13636'            , # G+13636  U+C15F  uniC15F,  HANGUL SYLLABLE-C15F
        '49504' => '13637'            , # G+13637  U+C160  uniC160,  HANGUL SYLLABLE-C160
        '49505' => '13638'            , # G+13638  U+C161  uniC161,  HANGUL SYLLABLE-C161
        '49506' => '13639'            , # G+13639  U+C162  uniC162,  HANGUL SYLLABLE-C162
        '49507' => '13640'            , # G+13640  U+C163  uniC163,  HANGUL SYLLABLE-C163
        '49508' => '2254'             , # G+2254  U+C164  uniC164,  HANGUL SYLLABLE-C164
        '49509' => '2255'             , # G+2255  U+C165  uniC165,  HANGUL SYLLABLE-C165
        '49510' => '13641'            , # G+13641  U+C166  uniC166,  HANGUL SYLLABLE-C166
        '49511' => '2256'             , # G+2256  U+C167  uniC167,  HANGUL SYLLABLE-C167
        '49512' => '2257'             , # G+2257  U+C168  uniC168,  HANGUL SYLLABLE-C168
        '49513' => '2258'             , # G+2258  U+C169  uniC169,  HANGUL SYLLABLE-C169
        '49514' => '13642'            , # G+13642  U+C16A  uniC16A,  HANGUL SYLLABLE-C16A
        '49515' => '13643'            , # G+13643  U+C16B  uniC16B,  HANGUL SYLLABLE-C16B
        '49516' => '13644'            , # G+13644  U+C16C  uniC16C,  HANGUL SYLLABLE-C16C
        '49517' => '13645'            , # G+13645  U+C16D  uniC16D,  HANGUL SYLLABLE-C16D
        '49518' => '13646'            , # G+13646  U+C16E  uniC16E,  HANGUL SYLLABLE-C16E
        '49519' => '13647'            , # G+13647  U+C16F  uniC16F,  HANGUL SYLLABLE-C16F
        '49520' => '2259'             , # G+2259  U+C170  uniC170,  HANGUL SYLLABLE-C170
        '49521' => '13648'            , # G+13648  U+C171  uniC171,  HANGUL SYLLABLE-C171
        '49522' => '13649'            , # G+13649  U+C172  uniC172,  HANGUL SYLLABLE-C172
        '49523' => '13650'            , # G+13650  U+C173  uniC173,  HANGUL SYLLABLE-C173
        '49524' => '2260'             , # G+2260  U+C174  uniC174,  HANGUL SYLLABLE-C174
        '49525' => '13651'            , # G+13651  U+C175  uniC175,  HANGUL SYLLABLE-C175
        '49526' => '13652'            , # G+13652  U+C176  uniC176,  HANGUL SYLLABLE-C176
        '49527' => '13653'            , # G+13653  U+C177  uniC177,  HANGUL SYLLABLE-C177
        '49528' => '2261'             , # G+2261  U+C178  uniC178,  HANGUL SYLLABLE-C178
        '49529' => '13654'            , # G+13654  U+C179  uniC179,  HANGUL SYLLABLE-C179
        '49530' => '13655'            , # G+13655  U+C17A  uniC17A,  HANGUL SYLLABLE-C17A
        '49531' => '13656'            , # G+13656  U+C17B  uniC17B,  HANGUL SYLLABLE-C17B
        '49532' => '13657'            , # G+13657  U+C17C  uniC17C,  HANGUL SYLLABLE-C17C
        '49533' => '13658'            , # G+13658  U+C17D  uniC17D,  HANGUL SYLLABLE-C17D
        '49534' => '13659'            , # G+13659  U+C17E  uniC17E,  HANGUL SYLLABLE-C17E
        '49535' => '13660'            , # G+13660  U+C17F  uniC17F,  HANGUL SYLLABLE-C17F
        '49536' => '13661'            , # G+13661  U+C180  uniC180,  HANGUL SYLLABLE-C180
        '49537' => '13662'            , # G+13662  U+C181  uniC181,  HANGUL SYLLABLE-C181
        '49538' => '13663'            , # G+13663  U+C182  uniC182,  HANGUL SYLLABLE-C182
        '49539' => '13664'            , # G+13664  U+C183  uniC183,  HANGUL SYLLABLE-C183
        '49540' => '13665'            , # G+13665  U+C184  uniC184,  HANGUL SYLLABLE-C184
        '49541' => '2262'             , # G+2262  U+C185  uniC185,  HANGUL SYLLABLE-C185
        '49542' => '13666'            , # G+13666  U+C186  uniC186,  HANGUL SYLLABLE-C186
        '49543' => '13667'            , # G+13667  U+C187  uniC187,  HANGUL SYLLABLE-C187
        '49544' => '13668'            , # G+13668  U+C188  uniC188,  HANGUL SYLLABLE-C188
        '49545' => '13669'            , # G+13669  U+C189  uniC189,  HANGUL SYLLABLE-C189
        '49546' => '13670'            , # G+13670  U+C18A  uniC18A,  HANGUL SYLLABLE-C18A
        '49547' => '13671'            , # G+13671  U+C18B  uniC18B,  HANGUL SYLLABLE-C18B
        '49548' => '2263'             , # G+2263  U+C18C  uniC18C,  HANGUL SYLLABLE-C18C
        '49549' => '2264'             , # G+2264  U+C18D  uniC18D,  HANGUL SYLLABLE-C18D
        '49550' => '2265'             , # G+2265  U+C18E  uniC18E,  HANGUL SYLLABLE-C18E
        '49551' => '13672'            , # G+13672  U+C18F  uniC18F,  HANGUL SYLLABLE-C18F
        '49552' => '2266'             , # G+2266  U+C190  uniC190,  HANGUL SYLLABLE-C190
        '49553' => '13673'            , # G+13673  U+C191  uniC191,  HANGUL SYLLABLE-C191
        '49554' => '13674'            , # G+13674  U+C192  uniC192,  HANGUL SYLLABLE-C192
        '49555' => '13675'            , # G+13675  U+C193  uniC193,  HANGUL SYLLABLE-C193
        '49556' => '2267'             , # G+2267  U+C194  uniC194,  HANGUL SYLLABLE-C194
        '49557' => '13676'            , # G+13676  U+C195  uniC195,  HANGUL SYLLABLE-C195
        '49558' => '2268'             , # G+2268  U+C196  uniC196,  HANGUL SYLLABLE-C196
        '49559' => '13677'            , # G+13677  U+C197  uniC197,  HANGUL SYLLABLE-C197
        '49560' => '13678'            , # G+13678  U+C198  uniC198,  HANGUL SYLLABLE-C198
        '49561' => '13679'            , # G+13679  U+C199  uniC199,  HANGUL SYLLABLE-C199
        '49562' => '13680'            , # G+13680  U+C19A  uniC19A,  HANGUL SYLLABLE-C19A
        '49563' => '13681'            , # G+13681  U+C19B  uniC19B,  HANGUL SYLLABLE-C19B
        '49564' => '2269'             , # G+2269  U+C19C  uniC19C,  HANGUL SYLLABLE-C19C
        '49565' => '2270'             , # G+2270  U+C19D  uniC19D,  HANGUL SYLLABLE-C19D
        '49566' => '13682'            , # G+13682  U+C19E  uniC19E,  HANGUL SYLLABLE-C19E
        '49567' => '2271'             , # G+2271  U+C19F  uniC19F,  HANGUL SYLLABLE-C19F
        '49568' => '13683'            , # G+13683  U+C1A0  uniC1A0,  HANGUL SYLLABLE-C1A0
        '49569' => '2272'             , # G+2272  U+C1A1  uniC1A1,  HANGUL SYLLABLE-C1A1
        '49570' => '13684'            , # G+13684  U+C1A2  uniC1A2,  HANGUL SYLLABLE-C1A2
        '49571' => '13685'            , # G+13685  U+C1A3  uniC1A3,  HANGUL SYLLABLE-C1A3
        '49572' => '13686'            , # G+13686  U+C1A4  uniC1A4,  HANGUL SYLLABLE-C1A4
        '49573' => '2273'             , # G+2273  U+C1A5  uniC1A5,  HANGUL SYLLABLE-C1A5
        '49574' => '13687'            , # G+13687  U+C1A6  uniC1A6,  HANGUL SYLLABLE-C1A6
        '49575' => '13688'            , # G+13688  U+C1A7  uniC1A7,  HANGUL SYLLABLE-C1A7
        '49576' => '2274'             , # G+2274  U+C1A8  uniC1A8,  HANGUL SYLLABLE-C1A8
        '49577' => '2275'             , # G+2275  U+C1A9  uniC1A9,  HANGUL SYLLABLE-C1A9
        '49578' => '13689'            , # G+13689  U+C1AA  uniC1AA,  HANGUL SYLLABLE-C1AA
        '49579' => '13690'            , # G+13690  U+C1AB  uniC1AB,  HANGUL SYLLABLE-C1AB
        '49580' => '2276'             , # G+2276  U+C1AC  uniC1AC,  HANGUL SYLLABLE-C1AC
        '49581' => '13691'            , # G+13691  U+C1AD  uniC1AD,  HANGUL SYLLABLE-C1AD
        '49582' => '13692'            , # G+13692  U+C1AE  uniC1AE,  HANGUL SYLLABLE-C1AE
        '49583' => '13693'            , # G+13693  U+C1AF  uniC1AF,  HANGUL SYLLABLE-C1AF
        '49584' => '2277'             , # G+2277  U+C1B0  uniC1B0,  HANGUL SYLLABLE-C1B0
        '49585' => '13694'            , # G+13694  U+C1B1  uniC1B1,  HANGUL SYLLABLE-C1B1
        '49586' => '13695'            , # G+13695  U+C1B2  uniC1B2,  HANGUL SYLLABLE-C1B2
        '49587' => '13696'            , # G+13696  U+C1B3  uniC1B3,  HANGUL SYLLABLE-C1B3
        '49588' => '13697'            , # G+13697  U+C1B4  uniC1B4,  HANGUL SYLLABLE-C1B4
        '49589' => '13698'            , # G+13698  U+C1B5  uniC1B5,  HANGUL SYLLABLE-C1B5
        '49590' => '13699'            , # G+13699  U+C1B6  uniC1B6,  HANGUL SYLLABLE-C1B6
        '49591' => '13700'            , # G+13700  U+C1B7  uniC1B7,  HANGUL SYLLABLE-C1B7
        '49592' => '13701'            , # G+13701  U+C1B8  uniC1B8,  HANGUL SYLLABLE-C1B8
        '49593' => '13702'            , # G+13702  U+C1B9  uniC1B9,  HANGUL SYLLABLE-C1B9
        '49594' => '13703'            , # G+13703  U+C1BA  uniC1BA,  HANGUL SYLLABLE-C1BA
        '49595' => '13704'            , # G+13704  U+C1BB  uniC1BB,  HANGUL SYLLABLE-C1BB
        '49596' => '13705'            , # G+13705  U+C1BC  uniC1BC,  HANGUL SYLLABLE-C1BC
        '49597' => '2278'             , # G+2278  U+C1BD  uniC1BD,  HANGUL SYLLABLE-C1BD
        '49598' => '13706'            , # G+13706  U+C1BE  uniC1BE,  HANGUL SYLLABLE-C1BE
        '49599' => '13707'            , # G+13707  U+C1BF  uniC1BF,  HANGUL SYLLABLE-C1BF
        '49600' => '13708'            , # G+13708  U+C1C0  uniC1C0,  HANGUL SYLLABLE-C1C0
        '49601' => '13709'            , # G+13709  U+C1C1  uniC1C1,  HANGUL SYLLABLE-C1C1
        '49602' => '13710'            , # G+13710  U+C1C2  uniC1C2,  HANGUL SYLLABLE-C1C2
        '49603' => '13711'            , # G+13711  U+C1C3  uniC1C3,  HANGUL SYLLABLE-C1C3
        '49604' => '2279'             , # G+2279  U+C1C4  uniC1C4,  HANGUL SYLLABLE-C1C4
        '49605' => '13712'            , # G+13712  U+C1C5  uniC1C5,  HANGUL SYLLABLE-C1C5
        '49606' => '13713'            , # G+13713  U+C1C6  uniC1C6,  HANGUL SYLLABLE-C1C6
        '49607' => '13714'            , # G+13714  U+C1C7  uniC1C7,  HANGUL SYLLABLE-C1C7
        '49608' => '2280'             , # G+2280  U+C1C8  uniC1C8,  HANGUL SYLLABLE-C1C8
        '49609' => '13715'            , # G+13715  U+C1C9  uniC1C9,  HANGUL SYLLABLE-C1C9
        '49610' => '13716'            , # G+13716  U+C1CA  uniC1CA,  HANGUL SYLLABLE-C1CA
        '49611' => '13717'            , # G+13717  U+C1CB  uniC1CB,  HANGUL SYLLABLE-C1CB
        '49612' => '2281'             , # G+2281  U+C1CC  uniC1CC,  HANGUL SYLLABLE-C1CC
        '49613' => '13718'            , # G+13718  U+C1CD  uniC1CD,  HANGUL SYLLABLE-C1CD
        '49614' => '13719'            , # G+13719  U+C1CE  uniC1CE,  HANGUL SYLLABLE-C1CE
        '49615' => '13720'            , # G+13720  U+C1CF  uniC1CF,  HANGUL SYLLABLE-C1CF
        '49616' => '13721'            , # G+13721  U+C1D0  uniC1D0,  HANGUL SYLLABLE-C1D0
        '49617' => '13722'            , # G+13722  U+C1D1  uniC1D1,  HANGUL SYLLABLE-C1D1
        '49618' => '13723'            , # G+13723  U+C1D2  uniC1D2,  HANGUL SYLLABLE-C1D2
        '49619' => '13724'            , # G+13724  U+C1D3  uniC1D3,  HANGUL SYLLABLE-C1D3
        '49620' => '2282'             , # G+2282  U+C1D4  uniC1D4,  HANGUL SYLLABLE-C1D4
        '49621' => '13725'            , # G+13725  U+C1D5  uniC1D5,  HANGUL SYLLABLE-C1D5
        '49622' => '13726'            , # G+13726  U+C1D6  uniC1D6,  HANGUL SYLLABLE-C1D6
        '49623' => '2283'             , # G+2283  U+C1D7  uniC1D7,  HANGUL SYLLABLE-C1D7
        '49624' => '2284'             , # G+2284  U+C1D8  uniC1D8,  HANGUL SYLLABLE-C1D8
        '49625' => '13727'            , # G+13727  U+C1D9  uniC1D9,  HANGUL SYLLABLE-C1D9
        '49626' => '13728'            , # G+13728  U+C1DA  uniC1DA,  HANGUL SYLLABLE-C1DA
        '49627' => '13729'            , # G+13729  U+C1DB  uniC1DB,  HANGUL SYLLABLE-C1DB
        '49628' => '13730'            , # G+13730  U+C1DC  uniC1DC,  HANGUL SYLLABLE-C1DC
        '49629' => '13731'            , # G+13731  U+C1DD  uniC1DD,  HANGUL SYLLABLE-C1DD
        '49630' => '13732'            , # G+13732  U+C1DE  uniC1DE,  HANGUL SYLLABLE-C1DE
        '49631' => '13733'            , # G+13733  U+C1DF  uniC1DF,  HANGUL SYLLABLE-C1DF
        '49632' => '2285'             , # G+2285  U+C1E0  uniC1E0,  HANGUL SYLLABLE-C1E0
        '49633' => '13734'            , # G+13734  U+C1E1  uniC1E1,  HANGUL SYLLABLE-C1E1
        '49634' => '13735'            , # G+13735  U+C1E2  uniC1E2,  HANGUL SYLLABLE-C1E2
        '49635' => '13736'            , # G+13736  U+C1E3  uniC1E3,  HANGUL SYLLABLE-C1E3
        '49636' => '2286'             , # G+2286  U+C1E4  uniC1E4,  HANGUL SYLLABLE-C1E4
        '49637' => '13737'            , # G+13737  U+C1E5  uniC1E5,  HANGUL SYLLABLE-C1E5
        '49638' => '13738'            , # G+13738  U+C1E6  uniC1E6,  HANGUL SYLLABLE-C1E6
        '49639' => '13739'            , # G+13739  U+C1E7  uniC1E7,  HANGUL SYLLABLE-C1E7
        '49640' => '2287'             , # G+2287  U+C1E8  uniC1E8,  HANGUL SYLLABLE-C1E8
        '49641' => '13740'            , # G+13740  U+C1E9  uniC1E9,  HANGUL SYLLABLE-C1E9
        '49642' => '13741'            , # G+13741  U+C1EA  uniC1EA,  HANGUL SYLLABLE-C1EA
        '49643' => '13742'            , # G+13742  U+C1EB  uniC1EB,  HANGUL SYLLABLE-C1EB
        '49644' => '13743'            , # G+13743  U+C1EC  uniC1EC,  HANGUL SYLLABLE-C1EC
        '49645' => '13744'            , # G+13744  U+C1ED  uniC1ED,  HANGUL SYLLABLE-C1ED
        '49646' => '13745'            , # G+13745  U+C1EE  uniC1EE,  HANGUL SYLLABLE-C1EE
        '49647' => '13746'            , # G+13746  U+C1EF  uniC1EF,  HANGUL SYLLABLE-C1EF
        '49648' => '2288'             , # G+2288  U+C1F0  uniC1F0,  HANGUL SYLLABLE-C1F0
        '49649' => '2289'             , # G+2289  U+C1F1  uniC1F1,  HANGUL SYLLABLE-C1F1
        '49650' => '13747'            , # G+13747  U+C1F2  uniC1F2,  HANGUL SYLLABLE-C1F2
        '49651' => '2290'             , # G+2290  U+C1F3  uniC1F3,  HANGUL SYLLABLE-C1F3
        '49652' => '13748'            , # G+13748  U+C1F4  uniC1F4,  HANGUL SYLLABLE-C1F4
        '49653' => '13749'            , # G+13749  U+C1F5  uniC1F5,  HANGUL SYLLABLE-C1F5
        '49654' => '13750'            , # G+13750  U+C1F6  uniC1F6,  HANGUL SYLLABLE-C1F6
        '49655' => '13751'            , # G+13751  U+C1F7  uniC1F7,  HANGUL SYLLABLE-C1F7
        '49656' => '13752'            , # G+13752  U+C1F8  uniC1F8,  HANGUL SYLLABLE-C1F8
        '49657' => '13753'            , # G+13753  U+C1F9  uniC1F9,  HANGUL SYLLABLE-C1F9
        '49658' => '13754'            , # G+13754  U+C1FA  uniC1FA,  HANGUL SYLLABLE-C1FA
        '49659' => '13755'            , # G+13755  U+C1FB  uniC1FB,  HANGUL SYLLABLE-C1FB
        '49660' => '2291'             , # G+2291  U+C1FC  uniC1FC,  HANGUL SYLLABLE-C1FC
        '49661' => '2292'             , # G+2292  U+C1FD  uniC1FD,  HANGUL SYLLABLE-C1FD
        '49662' => '13756'            , # G+13756  U+C1FE  uniC1FE,  HANGUL SYLLABLE-C1FE
        '49663' => '13757'            , # G+13757  U+C1FF  uniC1FF,  HANGUL SYLLABLE-C1FF
        '49664' => '2293'             , # G+2293  U+C200  uniC200,  HANGUL SYLLABLE-C200
        '49665' => '13758'            , # G+13758  U+C201  uniC201,  HANGUL SYLLABLE-C201
        '49666' => '13759'            , # G+13759  U+C202  uniC202,  HANGUL SYLLABLE-C202
        '49667' => '13760'            , # G+13760  U+C203  uniC203,  HANGUL SYLLABLE-C203
        '49668' => '2294'             , # G+2294  U+C204  uniC204,  HANGUL SYLLABLE-C204
        '49669' => '13761'            , # G+13761  U+C205  uniC205,  HANGUL SYLLABLE-C205
        '49670' => '13762'            , # G+13762  U+C206  uniC206,  HANGUL SYLLABLE-C206
        '49671' => '13763'            , # G+13763  U+C207  uniC207,  HANGUL SYLLABLE-C207
        '49672' => '13764'            , # G+13764  U+C208  uniC208,  HANGUL SYLLABLE-C208
        '49673' => '13765'            , # G+13765  U+C209  uniC209,  HANGUL SYLLABLE-C209
        '49674' => '13766'            , # G+13766  U+C20A  uniC20A,  HANGUL SYLLABLE-C20A
        '49675' => '13767'            , # G+13767  U+C20B  uniC20B,  HANGUL SYLLABLE-C20B
        '49676' => '2295'             , # G+2295  U+C20C  uniC20C,  HANGUL SYLLABLE-C20C
        '49677' => '2296'             , # G+2296  U+C20D  uniC20D,  HANGUL SYLLABLE-C20D
        '49678' => '13768'            , # G+13768  U+C20E  uniC20E,  HANGUL SYLLABLE-C20E
        '49679' => '2297'             , # G+2297  U+C20F  uniC20F,  HANGUL SYLLABLE-C20F
        '49680' => '13769'            , # G+13769  U+C210  uniC210,  HANGUL SYLLABLE-C210
        '49681' => '2298'             , # G+2298  U+C211  uniC211,  HANGUL SYLLABLE-C211
        '49682' => '13770'            , # G+13770  U+C212  uniC212,  HANGUL SYLLABLE-C212
        '49683' => '13771'            , # G+13771  U+C213  uniC213,  HANGUL SYLLABLE-C213
        '49684' => '13772'            , # G+13772  U+C214  uniC214,  HANGUL SYLLABLE-C214
        '49685' => '13773'            , # G+13773  U+C215  uniC215,  HANGUL SYLLABLE-C215
        '49686' => '13774'            , # G+13774  U+C216  uniC216,  HANGUL SYLLABLE-C216
        '49687' => '13775'            , # G+13775  U+C217  uniC217,  HANGUL SYLLABLE-C217
        '49688' => '2299'             , # G+2299  U+C218  uniC218,  HANGUL SYLLABLE-C218
        '49689' => '2300'             , # G+2300  U+C219  uniC219,  HANGUL SYLLABLE-C219
        '49690' => '13776'            , # G+13776  U+C21A  uniC21A,  HANGUL SYLLABLE-C21A
        '49691' => '13777'            , # G+13777  U+C21B  uniC21B,  HANGUL SYLLABLE-C21B
        '49692' => '2301'             , # G+2301  U+C21C  uniC21C,  HANGUL SYLLABLE-C21C
        '49693' => '13778'            , # G+13778  U+C21D  uniC21D,  HANGUL SYLLABLE-C21D
        '49694' => '13779'            , # G+13779  U+C21E  uniC21E,  HANGUL SYLLABLE-C21E
        '49695' => '2302'             , # G+2302  U+C21F  uniC21F,  HANGUL SYLLABLE-C21F
        '49696' => '2303'             , # G+2303  U+C220  uniC220,  HANGUL SYLLABLE-C220
        '49697' => '13780'            , # G+13780  U+C221  uniC221,  HANGUL SYLLABLE-C221
        '49698' => '13781'            , # G+13781  U+C222  uniC222,  HANGUL SYLLABLE-C222
        '49699' => '13782'            , # G+13782  U+C223  uniC223,  HANGUL SYLLABLE-C223
        '49700' => '13783'            , # G+13783  U+C224  uniC224,  HANGUL SYLLABLE-C224
        '49701' => '13784'            , # G+13784  U+C225  uniC225,  HANGUL SYLLABLE-C225
        '49702' => '13785'            , # G+13785  U+C226  uniC226,  HANGUL SYLLABLE-C226
        '49703' => '13786'            , # G+13786  U+C227  uniC227,  HANGUL SYLLABLE-C227
        '49704' => '2304'             , # G+2304  U+C228  uniC228,  HANGUL SYLLABLE-C228
        '49705' => '2305'             , # G+2305  U+C229  uniC229,  HANGUL SYLLABLE-C229
        '49706' => '13787'            , # G+13787  U+C22A  uniC22A,  HANGUL SYLLABLE-C22A
        '49707' => '2306'             , # G+2306  U+C22B  uniC22B,  HANGUL SYLLABLE-C22B
        '49708' => '13788'            , # G+13788  U+C22C  uniC22C,  HANGUL SYLLABLE-C22C
        '49709' => '2307'             , # G+2307  U+C22D  uniC22D,  HANGUL SYLLABLE-C22D
        '49710' => '13789'            , # G+13789  U+C22E  uniC22E,  HANGUL SYLLABLE-C22E
        '49711' => '2308'             , # G+2308  U+C22F  uniC22F,  HANGUL SYLLABLE-C22F
        '49712' => '13790'            , # G+13790  U+C230  uniC230,  HANGUL SYLLABLE-C230
        '49713' => '2309'             , # G+2309  U+C231  uniC231,  HANGUL SYLLABLE-C231
        '49714' => '2310'             , # G+2310  U+C232  uniC232,  HANGUL SYLLABLE-C232
        '49715' => '13791'            , # G+13791  U+C233  uniC233,  HANGUL SYLLABLE-C233
        '49716' => '2311'             , # G+2311  U+C234  uniC234,  HANGUL SYLLABLE-C234
        '49717' => '13792'            , # G+13792  U+C235  uniC235,  HANGUL SYLLABLE-C235
        '49718' => '13793'            , # G+13793  U+C236  uniC236,  HANGUL SYLLABLE-C236
        '49719' => '13794'            , # G+13794  U+C237  uniC237,  HANGUL SYLLABLE-C237
        '49720' => '13795'            , # G+13795  U+C238  uniC238,  HANGUL SYLLABLE-C238
        '49721' => '13796'            , # G+13796  U+C239  uniC239,  HANGUL SYLLABLE-C239
        '49722' => '13797'            , # G+13797  U+C23A  uniC23A,  HANGUL SYLLABLE-C23A
        '49723' => '13798'            , # G+13798  U+C23B  uniC23B,  HANGUL SYLLABLE-C23B
        '49724' => '13799'            , # G+13799  U+C23C  uniC23C,  HANGUL SYLLABLE-C23C
        '49725' => '13800'            , # G+13800  U+C23D  uniC23D,  HANGUL SYLLABLE-C23D
        '49726' => '13801'            , # G+13801  U+C23E  uniC23E,  HANGUL SYLLABLE-C23E
        '49727' => '13802'            , # G+13802  U+C23F  uniC23F,  HANGUL SYLLABLE-C23F
        '49728' => '13803'            , # G+13803  U+C240  uniC240,  HANGUL SYLLABLE-C240
        '49729' => '13804'            , # G+13804  U+C241  uniC241,  HANGUL SYLLABLE-C241
        '49730' => '13805'            , # G+13805  U+C242  uniC242,  HANGUL SYLLABLE-C242
        '49731' => '13806'            , # G+13806  U+C243  uniC243,  HANGUL SYLLABLE-C243
        '49732' => '13807'            , # G+13807  U+C244  uniC244,  HANGUL SYLLABLE-C244
        '49733' => '13808'            , # G+13808  U+C245  uniC245,  HANGUL SYLLABLE-C245
        '49734' => '13809'            , # G+13809  U+C246  uniC246,  HANGUL SYLLABLE-C246
        '49735' => '13810'            , # G+13810  U+C247  uniC247,  HANGUL SYLLABLE-C247
        '49736' => '2312'             , # G+2312  U+C248  uniC248,  HANGUL SYLLABLE-C248
        '49737' => '13811'            , # G+13811  U+C249  uniC249,  HANGUL SYLLABLE-C249
        '49738' => '13812'            , # G+13812  U+C24A  uniC24A,  HANGUL SYLLABLE-C24A
        '49739' => '13813'            , # G+13813  U+C24B  uniC24B,  HANGUL SYLLABLE-C24B
        '49740' => '13814'            , # G+13814  U+C24C  uniC24C,  HANGUL SYLLABLE-C24C
        '49741' => '13815'            , # G+13815  U+C24D  uniC24D,  HANGUL SYLLABLE-C24D
        '49742' => '13816'            , # G+13816  U+C24E  uniC24E,  HANGUL SYLLABLE-C24E
        '49743' => '13817'            , # G+13817  U+C24F  uniC24F,  HANGUL SYLLABLE-C24F
        '49744' => '2313'             , # G+2313  U+C250  uniC250,  HANGUL SYLLABLE-C250
        '49745' => '2314'             , # G+2314  U+C251  uniC251,  HANGUL SYLLABLE-C251
        '49746' => '13818'            , # G+13818  U+C252  uniC252,  HANGUL SYLLABLE-C252
        '49747' => '13819'            , # G+13819  U+C253  uniC253,  HANGUL SYLLABLE-C253
        '49748' => '2315'             , # G+2315  U+C254  uniC254,  HANGUL SYLLABLE-C254
        '49749' => '13820'            , # G+13820  U+C255  uniC255,  HANGUL SYLLABLE-C255
        '49750' => '13821'            , # G+13821  U+C256  uniC256,  HANGUL SYLLABLE-C256
        '49751' => '13822'            , # G+13822  U+C257  uniC257,  HANGUL SYLLABLE-C257
        '49752' => '2316'             , # G+2316  U+C258  uniC258,  HANGUL SYLLABLE-C258
        '49753' => '13823'            , # G+13823  U+C259  uniC259,  HANGUL SYLLABLE-C259
        '49754' => '13824'            , # G+13824  U+C25A  uniC25A,  HANGUL SYLLABLE-C25A
        '49755' => '13825'            , # G+13825  U+C25B  uniC25B,  HANGUL SYLLABLE-C25B
        '49756' => '13826'            , # G+13826  U+C25C  uniC25C,  HANGUL SYLLABLE-C25C
        '49757' => '13827'            , # G+13827  U+C25D  uniC25D,  HANGUL SYLLABLE-C25D
        '49758' => '13828'            , # G+13828  U+C25E  uniC25E,  HANGUL SYLLABLE-C25E
        '49759' => '13829'            , # G+13829  U+C25F  uniC25F,  HANGUL SYLLABLE-C25F
        '49760' => '2317'             , # G+2317  U+C260  uniC260,  HANGUL SYLLABLE-C260
        '49761' => '13830'            , # G+13830  U+C261  uniC261,  HANGUL SYLLABLE-C261
        '49762' => '13831'            , # G+13831  U+C262  uniC262,  HANGUL SYLLABLE-C262
        '49763' => '13832'            , # G+13832  U+C263  uniC263,  HANGUL SYLLABLE-C263
        '49764' => '13833'            , # G+13833  U+C264  uniC264,  HANGUL SYLLABLE-C264
        '49765' => '2318'             , # G+2318  U+C265  uniC265,  HANGUL SYLLABLE-C265
        '49766' => '13834'            , # G+13834  U+C266  uniC266,  HANGUL SYLLABLE-C266
        '49767' => '13835'            , # G+13835  U+C267  uniC267,  HANGUL SYLLABLE-C267
        '49768' => '13836'            , # G+13836  U+C268  uniC268,  HANGUL SYLLABLE-C268
        '49769' => '13837'            , # G+13837  U+C269  uniC269,  HANGUL SYLLABLE-C269
        '49770' => '13838'            , # G+13838  U+C26A  uniC26A,  HANGUL SYLLABLE-C26A
        '49771' => '13839'            , # G+13839  U+C26B  uniC26B,  HANGUL SYLLABLE-C26B
        '49772' => '2319'             , # G+2319  U+C26C  uniC26C,  HANGUL SYLLABLE-C26C
        '49773' => '2320'             , # G+2320  U+C26D  uniC26D,  HANGUL SYLLABLE-C26D
        '49774' => '13840'            , # G+13840  U+C26E  uniC26E,  HANGUL SYLLABLE-C26E
        '49775' => '13841'            , # G+13841  U+C26F  uniC26F,  HANGUL SYLLABLE-C26F
        '49776' => '2321'             , # G+2321  U+C270  uniC270,  HANGUL SYLLABLE-C270
        '49777' => '13842'            , # G+13842  U+C271  uniC271,  HANGUL SYLLABLE-C271
        '49778' => '13843'            , # G+13843  U+C272  uniC272,  HANGUL SYLLABLE-C272
        '49779' => '13844'            , # G+13844  U+C273  uniC273,  HANGUL SYLLABLE-C273
        '49780' => '2322'             , # G+2322  U+C274  uniC274,  HANGUL SYLLABLE-C274
        '49781' => '13845'            , # G+13845  U+C275  uniC275,  HANGUL SYLLABLE-C275
        '49782' => '13846'            , # G+13846  U+C276  uniC276,  HANGUL SYLLABLE-C276
        '49783' => '13847'            , # G+13847  U+C277  uniC277,  HANGUL SYLLABLE-C277
        '49784' => '13848'            , # G+13848  U+C278  uniC278,  HANGUL SYLLABLE-C278
        '49785' => '13849'            , # G+13849  U+C279  uniC279,  HANGUL SYLLABLE-C279
        '49786' => '13850'            , # G+13850  U+C27A  uniC27A,  HANGUL SYLLABLE-C27A
        '49787' => '13851'            , # G+13851  U+C27B  uniC27B,  HANGUL SYLLABLE-C27B
        '49788' => '2323'             , # G+2323  U+C27C  uniC27C,  HANGUL SYLLABLE-C27C
        '49789' => '2324'             , # G+2324  U+C27D  uniC27D,  HANGUL SYLLABLE-C27D
        '49790' => '13852'            , # G+13852  U+C27E  uniC27E,  HANGUL SYLLABLE-C27E
        '49791' => '2325'             , # G+2325  U+C27F  uniC27F,  HANGUL SYLLABLE-C27F
        '49792' => '13853'            , # G+13853  U+C280  uniC280,  HANGUL SYLLABLE-C280
        '49793' => '2326'             , # G+2326  U+C281  uniC281,  HANGUL SYLLABLE-C281
        '49794' => '13854'            , # G+13854  U+C282  uniC282,  HANGUL SYLLABLE-C282
        '49795' => '13855'            , # G+13855  U+C283  uniC283,  HANGUL SYLLABLE-C283
        '49796' => '13856'            , # G+13856  U+C284  uniC284,  HANGUL SYLLABLE-C284
        '49797' => '13857'            , # G+13857  U+C285  uniC285,  HANGUL SYLLABLE-C285
        '49798' => '13858'            , # G+13858  U+C286  uniC286,  HANGUL SYLLABLE-C286
        '49799' => '13859'            , # G+13859  U+C287  uniC287,  HANGUL SYLLABLE-C287
        '49800' => '2327'             , # G+2327  U+C288  uniC288,  HANGUL SYLLABLE-C288
        '49801' => '2328'             , # G+2328  U+C289  uniC289,  HANGUL SYLLABLE-C289
        '49802' => '13860'            , # G+13860  U+C28A  uniC28A,  HANGUL SYLLABLE-C28A
        '49803' => '13861'            , # G+13861  U+C28B  uniC28B,  HANGUL SYLLABLE-C28B
        '49804' => '13862'            , # G+13862  U+C28C  uniC28C,  HANGUL SYLLABLE-C28C
        '49805' => '13863'            , # G+13863  U+C28D  uniC28D,  HANGUL SYLLABLE-C28D
        '49806' => '13864'            , # G+13864  U+C28E  uniC28E,  HANGUL SYLLABLE-C28E
        '49807' => '13865'            , # G+13865  U+C28F  uniC28F,  HANGUL SYLLABLE-C28F
        '49808' => '2329'             , # G+2329  U+C290  uniC290,  HANGUL SYLLABLE-C290
        '49809' => '13866'            , # G+13866  U+C291  uniC291,  HANGUL SYLLABLE-C291
        '49810' => '13867'            , # G+13867  U+C292  uniC292,  HANGUL SYLLABLE-C292
        '49811' => '13868'            , # G+13868  U+C293  uniC293,  HANGUL SYLLABLE-C293
        '49812' => '13869'            , # G+13869  U+C294  uniC294,  HANGUL SYLLABLE-C294
        '49813' => '13870'            , # G+13870  U+C295  uniC295,  HANGUL SYLLABLE-C295
        '49814' => '13871'            , # G+13871  U+C296  uniC296,  HANGUL SYLLABLE-C296
        '49815' => '13872'            , # G+13872  U+C297  uniC297,  HANGUL SYLLABLE-C297
        '49816' => '2330'             , # G+2330  U+C298  uniC298,  HANGUL SYLLABLE-C298
        '49817' => '13873'            , # G+13873  U+C299  uniC299,  HANGUL SYLLABLE-C299
        '49818' => '13874'            , # G+13874  U+C29A  uniC29A,  HANGUL SYLLABLE-C29A
        '49819' => '2331'             , # G+2331  U+C29B  uniC29B,  HANGUL SYLLABLE-C29B
        '49820' => '13875'            , # G+13875  U+C29C  uniC29C,  HANGUL SYLLABLE-C29C
        '49821' => '2332'             , # G+2332  U+C29D  uniC29D,  HANGUL SYLLABLE-C29D
        '49822' => '13876'            , # G+13876  U+C29E  uniC29E,  HANGUL SYLLABLE-C29E
        '49823' => '13877'            , # G+13877  U+C29F  uniC29F,  HANGUL SYLLABLE-C29F
        '49824' => '13878'            , # G+13878  U+C2A0  uniC2A0,  HANGUL SYLLABLE-C2A0
        '49825' => '13879'            , # G+13879  U+C2A1  uniC2A1,  HANGUL SYLLABLE-C2A1
        '49826' => '13880'            , # G+13880  U+C2A2  uniC2A2,  HANGUL SYLLABLE-C2A2
        '49827' => '13881'            , # G+13881  U+C2A3  uniC2A3,  HANGUL SYLLABLE-C2A3
        '49828' => '2333'             , # G+2333  U+C2A4  uniC2A4,  HANGUL SYLLABLE-C2A4
        '49829' => '2334'             , # G+2334  U+C2A5  uniC2A5,  HANGUL SYLLABLE-C2A5
        '49830' => '13882'            , # G+13882  U+C2A6  uniC2A6,  HANGUL SYLLABLE-C2A6
        '49831' => '13883'            , # G+13883  U+C2A7  uniC2A7,  HANGUL SYLLABLE-C2A7
        '49832' => '2335'             , # G+2335  U+C2A8  uniC2A8,  HANGUL SYLLABLE-C2A8
        '49833' => '13884'            , # G+13884  U+C2A9  uniC2A9,  HANGUL SYLLABLE-C2A9
        '49834' => '13885'            , # G+13885  U+C2AA  uniC2AA,  HANGUL SYLLABLE-C2AA
        '49835' => '13886'            , # G+13886  U+C2AB  uniC2AB,  HANGUL SYLLABLE-C2AB
        '49836' => '2336'             , # G+2336  U+C2AC  uniC2AC,  HANGUL SYLLABLE-C2AC
        '49837' => '2337'             , # G+2337  U+C2AD  uniC2AD,  HANGUL SYLLABLE-C2AD
        '49838' => '13887'            , # G+13887  U+C2AE  uniC2AE,  HANGUL SYLLABLE-C2AE
        '49839' => '13888'            , # G+13888  U+C2AF  uniC2AF,  HANGUL SYLLABLE-C2AF
        '49840' => '13889'            , # G+13889  U+C2B0  uniC2B0,  HANGUL SYLLABLE-C2B0
        '49841' => '13890'            , # G+13890  U+C2B1  uniC2B1,  HANGUL SYLLABLE-C2B1
        '49842' => '13891'            , # G+13891  U+C2B2  uniC2B2,  HANGUL SYLLABLE-C2B2
        '49843' => '13892'            , # G+13892  U+C2B3  uniC2B3,  HANGUL SYLLABLE-C2B3
        '49844' => '2338'             , # G+2338  U+C2B4  uniC2B4,  HANGUL SYLLABLE-C2B4
        '49845' => '2339'             , # G+2339  U+C2B5  uniC2B5,  HANGUL SYLLABLE-C2B5
        '49846' => '13893'            , # G+13893  U+C2B6  uniC2B6,  HANGUL SYLLABLE-C2B6
        '49847' => '2340'             , # G+2340  U+C2B7  uniC2B7,  HANGUL SYLLABLE-C2B7
        '49848' => '13894'            , # G+13894  U+C2B8  uniC2B8,  HANGUL SYLLABLE-C2B8
        '49849' => '2341'             , # G+2341  U+C2B9  uniC2B9,  HANGUL SYLLABLE-C2B9
        '49850' => '13895'            , # G+13895  U+C2BA  uniC2BA,  HANGUL SYLLABLE-C2BA
        '49851' => '13896'            , # G+13896  U+C2BB  uniC2BB,  HANGUL SYLLABLE-C2BB
        '49852' => '13897'            , # G+13897  U+C2BC  uniC2BC,  HANGUL SYLLABLE-C2BC
        '49853' => '13898'            , # G+13898  U+C2BD  uniC2BD,  HANGUL SYLLABLE-C2BD
        '49854' => '13899'            , # G+13899  U+C2BE  uniC2BE,  HANGUL SYLLABLE-C2BE
        '49855' => '13900'            , # G+13900  U+C2BF  uniC2BF,  HANGUL SYLLABLE-C2BF
        '49856' => '13901'            , # G+13901  U+C2C0  uniC2C0,  HANGUL SYLLABLE-C2C0
        '49857' => '13902'            , # G+13902  U+C2C1  uniC2C1,  HANGUL SYLLABLE-C2C1
        '49858' => '13903'            , # G+13903  U+C2C2  uniC2C2,  HANGUL SYLLABLE-C2C2
        '49859' => '13904'            , # G+13904  U+C2C3  uniC2C3,  HANGUL SYLLABLE-C2C3
        '49860' => '13905'            , # G+13905  U+C2C4  uniC2C4,  HANGUL SYLLABLE-C2C4
        '49861' => '13906'            , # G+13906  U+C2C5  uniC2C5,  HANGUL SYLLABLE-C2C5
        '49862' => '13907'            , # G+13907  U+C2C6  uniC2C6,  HANGUL SYLLABLE-C2C6
        '49863' => '13908'            , # G+13908  U+C2C7  uniC2C7,  HANGUL SYLLABLE-C2C7
        '49864' => '13909'            , # G+13909  U+C2C8  uniC2C8,  HANGUL SYLLABLE-C2C8
        '49865' => '13910'            , # G+13910  U+C2C9  uniC2C9,  HANGUL SYLLABLE-C2C9
        '49866' => '13911'            , # G+13911  U+C2CA  uniC2CA,  HANGUL SYLLABLE-C2CA
        '49867' => '13912'            , # G+13912  U+C2CB  uniC2CB,  HANGUL SYLLABLE-C2CB
        '49868' => '13913'            , # G+13913  U+C2CC  uniC2CC,  HANGUL SYLLABLE-C2CC
        '49869' => '13914'            , # G+13914  U+C2CD  uniC2CD,  HANGUL SYLLABLE-C2CD
        '49870' => '13915'            , # G+13915  U+C2CE  uniC2CE,  HANGUL SYLLABLE-C2CE
        '49871' => '13916'            , # G+13916  U+C2CF  uniC2CF,  HANGUL SYLLABLE-C2CF
        '49872' => '13917'            , # G+13917  U+C2D0  uniC2D0,  HANGUL SYLLABLE-C2D0
        '49873' => '13918'            , # G+13918  U+C2D1  uniC2D1,  HANGUL SYLLABLE-C2D1
        '49874' => '13919'            , # G+13919  U+C2D2  uniC2D2,  HANGUL SYLLABLE-C2D2
        '49875' => '13920'            , # G+13920  U+C2D3  uniC2D3,  HANGUL SYLLABLE-C2D3
        '49876' => '13921'            , # G+13921  U+C2D4  uniC2D4,  HANGUL SYLLABLE-C2D4
        '49877' => '13922'            , # G+13922  U+C2D5  uniC2D5,  HANGUL SYLLABLE-C2D5
        '49878' => '13923'            , # G+13923  U+C2D6  uniC2D6,  HANGUL SYLLABLE-C2D6
        '49879' => '13924'            , # G+13924  U+C2D7  uniC2D7,  HANGUL SYLLABLE-C2D7
        '49880' => '13925'            , # G+13925  U+C2D8  uniC2D8,  HANGUL SYLLABLE-C2D8
        '49881' => '13926'            , # G+13926  U+C2D9  uniC2D9,  HANGUL SYLLABLE-C2D9
        '49882' => '13927'            , # G+13927  U+C2DA  uniC2DA,  HANGUL SYLLABLE-C2DA
        '49883' => '13928'            , # G+13928  U+C2DB  uniC2DB,  HANGUL SYLLABLE-C2DB
        '49884' => '2342'             , # G+2342  U+C2DC  uniC2DC,  HANGUL SYLLABLE-C2DC
        '49885' => '2343'             , # G+2343  U+C2DD  uniC2DD,  HANGUL SYLLABLE-C2DD
        '49886' => '13929'            , # G+13929  U+C2DE  uniC2DE,  HANGUL SYLLABLE-C2DE
        '49887' => '13930'            , # G+13930  U+C2DF  uniC2DF,  HANGUL SYLLABLE-C2DF
        '49888' => '2344'             , # G+2344  U+C2E0  uniC2E0,  HANGUL SYLLABLE-C2E0
        '49889' => '13931'            , # G+13931  U+C2E1  uniC2E1,  HANGUL SYLLABLE-C2E1
        '49890' => '13932'            , # G+13932  U+C2E2  uniC2E2,  HANGUL SYLLABLE-C2E2
        '49891' => '2345'             , # G+2345  U+C2E3  uniC2E3,  HANGUL SYLLABLE-C2E3
        '49892' => '2346'             , # G+2346  U+C2E4  uniC2E4,  HANGUL SYLLABLE-C2E4
        '49893' => '13933'            , # G+13933  U+C2E5  uniC2E5,  HANGUL SYLLABLE-C2E5
        '49894' => '13934'            , # G+13934  U+C2E6  uniC2E6,  HANGUL SYLLABLE-C2E6
        '49895' => '13935'            , # G+13935  U+C2E7  uniC2E7,  HANGUL SYLLABLE-C2E7
        '49896' => '13936'            , # G+13936  U+C2E8  uniC2E8,  HANGUL SYLLABLE-C2E8
        '49897' => '13937'            , # G+13937  U+C2E9  uniC2E9,  HANGUL SYLLABLE-C2E9
        '49898' => '13938'            , # G+13938  U+C2EA  uniC2EA,  HANGUL SYLLABLE-C2EA
        '49899' => '2347'             , # G+2347  U+C2EB  uniC2EB,  HANGUL SYLLABLE-C2EB
        '49900' => '2348'             , # G+2348  U+C2EC  uniC2EC,  HANGUL SYLLABLE-C2EC
        '49901' => '2349'             , # G+2349  U+C2ED  uniC2ED,  HANGUL SYLLABLE-C2ED
        '49902' => '13939'            , # G+13939  U+C2EE  uniC2EE,  HANGUL SYLLABLE-C2EE
        '49903' => '2350'             , # G+2350  U+C2EF  uniC2EF,  HANGUL SYLLABLE-C2EF
        '49904' => '13940'            , # G+13940  U+C2F0  uniC2F0,  HANGUL SYLLABLE-C2F0
        '49905' => '2351'             , # G+2351  U+C2F1  uniC2F1,  HANGUL SYLLABLE-C2F1
        '49906' => '13941'            , # G+13941  U+C2F2  uniC2F2,  HANGUL SYLLABLE-C2F2
        '49907' => '13942'            , # G+13942  U+C2F3  uniC2F3,  HANGUL SYLLABLE-C2F3
        '49908' => '13943'            , # G+13943  U+C2F4  uniC2F4,  HANGUL SYLLABLE-C2F4
        '49909' => '13944'            , # G+13944  U+C2F5  uniC2F5,  HANGUL SYLLABLE-C2F5
        '49910' => '2352'             , # G+2352  U+C2F6  uniC2F6,  HANGUL SYLLABLE-C2F6
        '49911' => '13945'            , # G+13945  U+C2F7  uniC2F7,  HANGUL SYLLABLE-C2F7
        '49912' => '2353'             , # G+2353  U+C2F8  uniC2F8,  HANGUL SYLLABLE-C2F8
        '49913' => '2354'             , # G+2354  U+C2F9  uniC2F9,  HANGUL SYLLABLE-C2F9
        '49914' => '13946'            , # G+13946  U+C2FA  uniC2FA,  HANGUL SYLLABLE-C2FA
        '49915' => '2355'             , # G+2355  U+C2FB  uniC2FB,  HANGUL SYLLABLE-C2FB
        '49916' => '2356'             , # G+2356  U+C2FC  uniC2FC,  HANGUL SYLLABLE-C2FC
        '49917' => '13947'            , # G+13947  U+C2FD  uniC2FD,  HANGUL SYLLABLE-C2FD
        '49918' => '13948'            , # G+13948  U+C2FE  uniC2FE,  HANGUL SYLLABLE-C2FE
        '49919' => '13949'            , # G+13949  U+C2FF  uniC2FF,  HANGUL SYLLABLE-C2FF
        '49920' => '2357'             , # G+2357  U+C300  uniC300,  HANGUL SYLLABLE-C300
        '49921' => '13950'            , # G+13950  U+C301  uniC301,  HANGUL SYLLABLE-C301
        '49922' => '13951'            , # G+13951  U+C302  uniC302,  HANGUL SYLLABLE-C302
        '49923' => '13952'            , # G+13952  U+C303  uniC303,  HANGUL SYLLABLE-C303
        '49924' => '13953'            , # G+13953  U+C304  uniC304,  HANGUL SYLLABLE-C304
        '49925' => '13954'            , # G+13954  U+C305  uniC305,  HANGUL SYLLABLE-C305
        '49926' => '13955'            , # G+13955  U+C306  uniC306,  HANGUL SYLLABLE-C306
        '49927' => '13956'            , # G+13956  U+C307  uniC307,  HANGUL SYLLABLE-C307
        '49928' => '2358'             , # G+2358  U+C308  uniC308,  HANGUL SYLLABLE-C308
        '49929' => '2359'             , # G+2359  U+C309  uniC309,  HANGUL SYLLABLE-C309
        '49930' => '13957'            , # G+13957  U+C30A  uniC30A,  HANGUL SYLLABLE-C30A
        '49931' => '13958'            , # G+13958  U+C30B  uniC30B,  HANGUL SYLLABLE-C30B
        '49932' => '2360'             , # G+2360  U+C30C  uniC30C,  HANGUL SYLLABLE-C30C
        '49933' => '2361'             , # G+2361  U+C30D  uniC30D,  HANGUL SYLLABLE-C30D
        '49934' => '13959'            , # G+13959  U+C30E  uniC30E,  HANGUL SYLLABLE-C30E
        '49935' => '13960'            , # G+13960  U+C30F  uniC30F,  HANGUL SYLLABLE-C30F
        '49936' => '13961'            , # G+13961  U+C310  uniC310,  HANGUL SYLLABLE-C310
        '49937' => '13962'            , # G+13962  U+C311  uniC311,  HANGUL SYLLABLE-C311
        '49938' => '13963'            , # G+13963  U+C312  uniC312,  HANGUL SYLLABLE-C312
        '49939' => '2362'             , # G+2362  U+C313  uniC313,  HANGUL SYLLABLE-C313
        '49940' => '2363'             , # G+2363  U+C314  uniC314,  HANGUL SYLLABLE-C314
        '49941' => '2364'             , # G+2364  U+C315  uniC315,  HANGUL SYLLABLE-C315
        '49942' => '13964'            , # G+13964  U+C316  uniC316,  HANGUL SYLLABLE-C316
        '49943' => '13965'            , # G+13965  U+C317  uniC317,  HANGUL SYLLABLE-C317
        '49944' => '2365'             , # G+2365  U+C318  uniC318,  HANGUL SYLLABLE-C318
        '49945' => '13966'            , # G+13966  U+C319  uniC319,  HANGUL SYLLABLE-C319
        '49946' => '13967'            , # G+13967  U+C31A  uniC31A,  HANGUL SYLLABLE-C31A
        '49947' => '13968'            , # G+13968  U+C31B  uniC31B,  HANGUL SYLLABLE-C31B
        '49948' => '2366'             , # G+2366  U+C31C  uniC31C,  HANGUL SYLLABLE-C31C
        '49949' => '13969'            , # G+13969  U+C31D  uniC31D,  HANGUL SYLLABLE-C31D
        '49950' => '13970'            , # G+13970  U+C31E  uniC31E,  HANGUL SYLLABLE-C31E
        '49951' => '13971'            , # G+13971  U+C31F  uniC31F,  HANGUL SYLLABLE-C31F
        '49952' => '13972'            , # G+13972  U+C320  uniC320,  HANGUL SYLLABLE-C320
        '49953' => '13973'            , # G+13973  U+C321  uniC321,  HANGUL SYLLABLE-C321
        '49954' => '13974'            , # G+13974  U+C322  uniC322,  HANGUL SYLLABLE-C322
        '49955' => '13975'            , # G+13975  U+C323  uniC323,  HANGUL SYLLABLE-C323
        '49956' => '2367'             , # G+2367  U+C324  uniC324,  HANGUL SYLLABLE-C324
        '49957' => '2368'             , # G+2368  U+C325  uniC325,  HANGUL SYLLABLE-C325
        '49958' => '13976'            , # G+13976  U+C326  uniC326,  HANGUL SYLLABLE-C326
        '49959' => '13977'            , # G+13977  U+C327  uniC327,  HANGUL SYLLABLE-C327
        '49960' => '2369'             , # G+2369  U+C328  uniC328,  HANGUL SYLLABLE-C328
        '49961' => '2370'             , # G+2370  U+C329  uniC329,  HANGUL SYLLABLE-C329
        '49962' => '13978'            , # G+13978  U+C32A  uniC32A,  HANGUL SYLLABLE-C32A
        '49963' => '13979'            , # G+13979  U+C32B  uniC32B,  HANGUL SYLLABLE-C32B
        '49964' => '13980'            , # G+13980  U+C32C  uniC32C,  HANGUL SYLLABLE-C32C
        '49965' => '13981'            , # G+13981  U+C32D  uniC32D,  HANGUL SYLLABLE-C32D
        '49966' => '13982'            , # G+13982  U+C32E  uniC32E,  HANGUL SYLLABLE-C32E
        '49967' => '13983'            , # G+13983  U+C32F  uniC32F,  HANGUL SYLLABLE-C32F
        '49968' => '13984'            , # G+13984  U+C330  uniC330,  HANGUL SYLLABLE-C330
        '49969' => '13985'            , # G+13985  U+C331  uniC331,  HANGUL SYLLABLE-C331
        '49970' => '13986'            , # G+13986  U+C332  uniC332,  HANGUL SYLLABLE-C332
        '49971' => '13987'            , # G+13987  U+C333  uniC333,  HANGUL SYLLABLE-C333
        '49972' => '13988'            , # G+13988  U+C334  uniC334,  HANGUL SYLLABLE-C334
        '49973' => '13989'            , # G+13989  U+C335  uniC335,  HANGUL SYLLABLE-C335
        '49974' => '13990'            , # G+13990  U+C336  uniC336,  HANGUL SYLLABLE-C336
        '49975' => '13991'            , # G+13991  U+C337  uniC337,  HANGUL SYLLABLE-C337
        '49976' => '13992'            , # G+13992  U+C338  uniC338,  HANGUL SYLLABLE-C338
        '49977' => '13993'            , # G+13993  U+C339  uniC339,  HANGUL SYLLABLE-C339
        '49978' => '13994'            , # G+13994  U+C33A  uniC33A,  HANGUL SYLLABLE-C33A
        '49979' => '13995'            , # G+13995  U+C33B  uniC33B,  HANGUL SYLLABLE-C33B
        '49980' => '13996'            , # G+13996  U+C33C  uniC33C,  HANGUL SYLLABLE-C33C
        '49981' => '13997'            , # G+13997  U+C33D  uniC33D,  HANGUL SYLLABLE-C33D
        '49982' => '13998'            , # G+13998  U+C33E  uniC33E,  HANGUL SYLLABLE-C33E
        '49983' => '13999'            , # G+13999  U+C33F  uniC33F,  HANGUL SYLLABLE-C33F
        '49984' => '14000'            , # G+14000  U+C340  uniC340,  HANGUL SYLLABLE-C340
        '49985' => '14001'            , # G+14001  U+C341  uniC341,  HANGUL SYLLABLE-C341
        '49986' => '14002'            , # G+14002  U+C342  uniC342,  HANGUL SYLLABLE-C342
        '49987' => '14003'            , # G+14003  U+C343  uniC343,  HANGUL SYLLABLE-C343
        '49988' => '14004'            , # G+14004  U+C344  uniC344,  HANGUL SYLLABLE-C344
        '49989' => '2371'             , # G+2371  U+C345  uniC345,  HANGUL SYLLABLE-C345
        '49990' => '14005'            , # G+14005  U+C346  uniC346,  HANGUL SYLLABLE-C346
        '49991' => '14006'            , # G+14006  U+C347  uniC347,  HANGUL SYLLABLE-C347
        '49992' => '14007'            , # G+14007  U+C348  uniC348,  HANGUL SYLLABLE-C348
        '49993' => '14008'            , # G+14008  U+C349  uniC349,  HANGUL SYLLABLE-C349
        '49994' => '14009'            , # G+14009  U+C34A  uniC34A,  HANGUL SYLLABLE-C34A
        '49995' => '14010'            , # G+14010  U+C34B  uniC34B,  HANGUL SYLLABLE-C34B
        '49996' => '14011'            , # G+14011  U+C34C  uniC34C,  HANGUL SYLLABLE-C34C
        '49997' => '14012'            , # G+14012  U+C34D  uniC34D,  HANGUL SYLLABLE-C34D
        '49998' => '14013'            , # G+14013  U+C34E  uniC34E,  HANGUL SYLLABLE-C34E
        '49999' => '14014'            , # G+14014  U+C34F  uniC34F,  HANGUL SYLLABLE-C34F
        '50000' => '14015'            , # G+14015  U+C350  uniC350,  HANGUL SYLLABLE-C350
        '50001' => '14016'            , # G+14016  U+C351  uniC351,  HANGUL SYLLABLE-C351
        '50002' => '14017'            , # G+14017  U+C352  uniC352,  HANGUL SYLLABLE-C352
        '50003' => '14018'            , # G+14018  U+C353  uniC353,  HANGUL SYLLABLE-C353
        '50004' => '14019'            , # G+14019  U+C354  uniC354,  HANGUL SYLLABLE-C354
        '50005' => '14020'            , # G+14020  U+C355  uniC355,  HANGUL SYLLABLE-C355
        '50006' => '14021'            , # G+14021  U+C356  uniC356,  HANGUL SYLLABLE-C356
        '50007' => '14022'            , # G+14022  U+C357  uniC357,  HANGUL SYLLABLE-C357
        '50008' => '14023'            , # G+14023  U+C358  uniC358,  HANGUL SYLLABLE-C358
        '50009' => '14024'            , # G+14024  U+C359  uniC359,  HANGUL SYLLABLE-C359
        '50010' => '14025'            , # G+14025  U+C35A  uniC35A,  HANGUL SYLLABLE-C35A
        '50011' => '14026'            , # G+14026  U+C35B  uniC35B,  HANGUL SYLLABLE-C35B
        '50012' => '14027'            , # G+14027  U+C35C  uniC35C,  HANGUL SYLLABLE-C35C
        '50013' => '14028'            , # G+14028  U+C35D  uniC35D,  HANGUL SYLLABLE-C35D
        '50014' => '14029'            , # G+14029  U+C35E  uniC35E,  HANGUL SYLLABLE-C35E
        '50015' => '14030'            , # G+14030  U+C35F  uniC35F,  HANGUL SYLLABLE-C35F
        '50016' => '14031'            , # G+14031  U+C360  uniC360,  HANGUL SYLLABLE-C360
        '50017' => '14032'            , # G+14032  U+C361  uniC361,  HANGUL SYLLABLE-C361
        '50018' => '14033'            , # G+14033  U+C362  uniC362,  HANGUL SYLLABLE-C362
        '50019' => '14034'            , # G+14034  U+C363  uniC363,  HANGUL SYLLABLE-C363
        '50020' => '14035'            , # G+14035  U+C364  uniC364,  HANGUL SYLLABLE-C364
        '50021' => '14036'            , # G+14036  U+C365  uniC365,  HANGUL SYLLABLE-C365
        '50022' => '14037'            , # G+14037  U+C366  uniC366,  HANGUL SYLLABLE-C366
        '50023' => '14038'            , # G+14038  U+C367  uniC367,  HANGUL SYLLABLE-C367
        '50024' => '2372'             , # G+2372  U+C368  uniC368,  HANGUL SYLLABLE-C368
        '50025' => '2373'             , # G+2373  U+C369  uniC369,  HANGUL SYLLABLE-C369
        '50026' => '14039'            , # G+14039  U+C36A  uniC36A,  HANGUL SYLLABLE-C36A
        '50027' => '14040'            , # G+14040  U+C36B  uniC36B,  HANGUL SYLLABLE-C36B
        '50028' => '2374'             , # G+2374  U+C36C  uniC36C,  HANGUL SYLLABLE-C36C
        '50029' => '14041'            , # G+14041  U+C36D  uniC36D,  HANGUL SYLLABLE-C36D
        '50030' => '14042'            , # G+14042  U+C36E  uniC36E,  HANGUL SYLLABLE-C36E
        '50031' => '14043'            , # G+14043  U+C36F  uniC36F,  HANGUL SYLLABLE-C36F
        '50032' => '2375'             , # G+2375  U+C370  uniC370,  HANGUL SYLLABLE-C370
        '50033' => '14044'            , # G+14044  U+C371  uniC371,  HANGUL SYLLABLE-C371
        '50034' => '2376'             , # G+2376  U+C372  uniC372,  HANGUL SYLLABLE-C372
        '50035' => '14045'            , # G+14045  U+C373  uniC373,  HANGUL SYLLABLE-C373
        '50036' => '14046'            , # G+14046  U+C374  uniC374,  HANGUL SYLLABLE-C374
        '50037' => '14047'            , # G+14047  U+C375  uniC375,  HANGUL SYLLABLE-C375
        '50038' => '14048'            , # G+14048  U+C376  uniC376,  HANGUL SYLLABLE-C376
        '50039' => '14049'            , # G+14049  U+C377  uniC377,  HANGUL SYLLABLE-C377
        '50040' => '2377'             , # G+2377  U+C378  uniC378,  HANGUL SYLLABLE-C378
        '50041' => '2378'             , # G+2378  U+C379  uniC379,  HANGUL SYLLABLE-C379
        '50042' => '14050'            , # G+14050  U+C37A  uniC37A,  HANGUL SYLLABLE-C37A
        '50043' => '14051'            , # G+14051  U+C37B  uniC37B,  HANGUL SYLLABLE-C37B
        '50044' => '2379'             , # G+2379  U+C37C  uniC37C,  HANGUL SYLLABLE-C37C
        '50045' => '2380'             , # G+2380  U+C37D  uniC37D,  HANGUL SYLLABLE-C37D
        '50046' => '14052'            , # G+14052  U+C37E  uniC37E,  HANGUL SYLLABLE-C37E
        '50047' => '14053'            , # G+14053  U+C37F  uniC37F,  HANGUL SYLLABLE-C37F
        '50048' => '14054'            , # G+14054  U+C380  uniC380,  HANGUL SYLLABLE-C380
        '50049' => '14055'            , # G+14055  U+C381  uniC381,  HANGUL SYLLABLE-C381
        '50050' => '14056'            , # G+14056  U+C382  uniC382,  HANGUL SYLLABLE-C382
        '50051' => '14057'            , # G+14057  U+C383  uniC383,  HANGUL SYLLABLE-C383
        '50052' => '2381'             , # G+2381  U+C384  uniC384,  HANGUL SYLLABLE-C384
        '50053' => '14058'            , # G+14058  U+C385  uniC385,  HANGUL SYLLABLE-C385
        '50054' => '14059'            , # G+14059  U+C386  uniC386,  HANGUL SYLLABLE-C386
        '50055' => '14060'            , # G+14060  U+C387  uniC387,  HANGUL SYLLABLE-C387
        '50056' => '2382'             , # G+2382  U+C388  uniC388,  HANGUL SYLLABLE-C388
        '50057' => '14061'            , # G+14061  U+C389  uniC389,  HANGUL SYLLABLE-C389
        '50058' => '14062'            , # G+14062  U+C38A  uniC38A,  HANGUL SYLLABLE-C38A
        '50059' => '14063'            , # G+14063  U+C38B  uniC38B,  HANGUL SYLLABLE-C38B
        '50060' => '2383'             , # G+2383  U+C38C  uniC38C,  HANGUL SYLLABLE-C38C
        '50061' => '14064'            , # G+14064  U+C38D  uniC38D,  HANGUL SYLLABLE-C38D
        '50062' => '14065'            , # G+14065  U+C38E  uniC38E,  HANGUL SYLLABLE-C38E
        '50063' => '14066'            , # G+14066  U+C38F  uniC38F,  HANGUL SYLLABLE-C38F
        '50064' => '14067'            , # G+14067  U+C390  uniC390,  HANGUL SYLLABLE-C390
        '50065' => '14068'            , # G+14068  U+C391  uniC391,  HANGUL SYLLABLE-C391
        '50066' => '14069'            , # G+14069  U+C392  uniC392,  HANGUL SYLLABLE-C392
        '50067' => '14070'            , # G+14070  U+C393  uniC393,  HANGUL SYLLABLE-C393
        '50068' => '14071'            , # G+14071  U+C394  uniC394,  HANGUL SYLLABLE-C394
        '50069' => '14072'            , # G+14072  U+C395  uniC395,  HANGUL SYLLABLE-C395
        '50070' => '14073'            , # G+14073  U+C396  uniC396,  HANGUL SYLLABLE-C396
        '50071' => '14074'            , # G+14074  U+C397  uniC397,  HANGUL SYLLABLE-C397
        '50072' => '14075'            , # G+14075  U+C398  uniC398,  HANGUL SYLLABLE-C398
        '50073' => '14076'            , # G+14076  U+C399  uniC399,  HANGUL SYLLABLE-C399
        '50074' => '14077'            , # G+14077  U+C39A  uniC39A,  HANGUL SYLLABLE-C39A
        '50075' => '14078'            , # G+14078  U+C39B  uniC39B,  HANGUL SYLLABLE-C39B
        '50076' => '14079'            , # G+14079  U+C39C  uniC39C,  HANGUL SYLLABLE-C39C
        '50077' => '14080'            , # G+14080  U+C39D  uniC39D,  HANGUL SYLLABLE-C39D
        '50078' => '14081'            , # G+14081  U+C39E  uniC39E,  HANGUL SYLLABLE-C39E
        '50079' => '14082'            , # G+14082  U+C39F  uniC39F,  HANGUL SYLLABLE-C39F
        '50080' => '14083'            , # G+14083  U+C3A0  uniC3A0,  HANGUL SYLLABLE-C3A0
        '50081' => '14084'            , # G+14084  U+C3A1  uniC3A1,  HANGUL SYLLABLE-C3A1
        '50082' => '14085'            , # G+14085  U+C3A2  uniC3A2,  HANGUL SYLLABLE-C3A2
        '50083' => '14086'            , # G+14086  U+C3A3  uniC3A3,  HANGUL SYLLABLE-C3A3
        '50084' => '14087'            , # G+14087  U+C3A4  uniC3A4,  HANGUL SYLLABLE-C3A4
        '50085' => '14088'            , # G+14088  U+C3A5  uniC3A5,  HANGUL SYLLABLE-C3A5
        '50086' => '14089'            , # G+14089  U+C3A6  uniC3A6,  HANGUL SYLLABLE-C3A6
        '50087' => '14090'            , # G+14090  U+C3A7  uniC3A7,  HANGUL SYLLABLE-C3A7
        '50088' => '14091'            , # G+14091  U+C3A8  uniC3A8,  HANGUL SYLLABLE-C3A8
        '50089' => '14092'            , # G+14092  U+C3A9  uniC3A9,  HANGUL SYLLABLE-C3A9
        '50090' => '14093'            , # G+14093  U+C3AA  uniC3AA,  HANGUL SYLLABLE-C3AA
        '50091' => '14094'            , # G+14094  U+C3AB  uniC3AB,  HANGUL SYLLABLE-C3AB
        '50092' => '14095'            , # G+14095  U+C3AC  uniC3AC,  HANGUL SYLLABLE-C3AC
        '50093' => '14096'            , # G+14096  U+C3AD  uniC3AD,  HANGUL SYLLABLE-C3AD
        '50094' => '14097'            , # G+14097  U+C3AE  uniC3AE,  HANGUL SYLLABLE-C3AE
        '50095' => '14098'            , # G+14098  U+C3AF  uniC3AF,  HANGUL SYLLABLE-C3AF
        '50096' => '14099'            , # G+14099  U+C3B0  uniC3B0,  HANGUL SYLLABLE-C3B0
        '50097' => '14100'            , # G+14100  U+C3B1  uniC3B1,  HANGUL SYLLABLE-C3B1
        '50098' => '14101'            , # G+14101  U+C3B2  uniC3B2,  HANGUL SYLLABLE-C3B2
        '50099' => '14102'            , # G+14102  U+C3B3  uniC3B3,  HANGUL SYLLABLE-C3B3
        '50100' => '14103'            , # G+14103  U+C3B4  uniC3B4,  HANGUL SYLLABLE-C3B4
        '50101' => '14104'            , # G+14104  U+C3B5  uniC3B5,  HANGUL SYLLABLE-C3B5
        '50102' => '14105'            , # G+14105  U+C3B6  uniC3B6,  HANGUL SYLLABLE-C3B6
        '50103' => '14106'            , # G+14106  U+C3B7  uniC3B7,  HANGUL SYLLABLE-C3B7
        '50104' => '14107'            , # G+14107  U+C3B8  uniC3B8,  HANGUL SYLLABLE-C3B8
        '50105' => '14108'            , # G+14108  U+C3B9  uniC3B9,  HANGUL SYLLABLE-C3B9
        '50106' => '14109'            , # G+14109  U+C3BA  uniC3BA,  HANGUL SYLLABLE-C3BA
        '50107' => '14110'            , # G+14110  U+C3BB  uniC3BB,  HANGUL SYLLABLE-C3BB
        '50108' => '14111'            , # G+14111  U+C3BC  uniC3BC,  HANGUL SYLLABLE-C3BC
        '50109' => '14112'            , # G+14112  U+C3BD  uniC3BD,  HANGUL SYLLABLE-C3BD
        '50110' => '14113'            , # G+14113  U+C3BE  uniC3BE,  HANGUL SYLLABLE-C3BE
        '50111' => '14114'            , # G+14114  U+C3BF  uniC3BF,  HANGUL SYLLABLE-C3BF
        '50112' => '2384'             , # G+2384  U+C3C0  uniC3C0,  HANGUL SYLLABLE-C3C0
        '50113' => '14115'            , # G+14115  U+C3C1  uniC3C1,  HANGUL SYLLABLE-C3C1
        '50114' => '14116'            , # G+14116  U+C3C2  uniC3C2,  HANGUL SYLLABLE-C3C2
        '50115' => '14117'            , # G+14117  U+C3C3  uniC3C3,  HANGUL SYLLABLE-C3C3
        '50116' => '14118'            , # G+14118  U+C3C4  uniC3C4,  HANGUL SYLLABLE-C3C4
        '50117' => '14119'            , # G+14119  U+C3C5  uniC3C5,  HANGUL SYLLABLE-C3C5
        '50118' => '14120'            , # G+14120  U+C3C6  uniC3C6,  HANGUL SYLLABLE-C3C6
        '50119' => '14121'            , # G+14121  U+C3C7  uniC3C7,  HANGUL SYLLABLE-C3C7
        '50120' => '14122'            , # G+14122  U+C3C8  uniC3C8,  HANGUL SYLLABLE-C3C8
        '50121' => '14123'            , # G+14123  U+C3C9  uniC3C9,  HANGUL SYLLABLE-C3C9
        '50122' => '14124'            , # G+14124  U+C3CA  uniC3CA,  HANGUL SYLLABLE-C3CA
        '50123' => '14125'            , # G+14125  U+C3CB  uniC3CB,  HANGUL SYLLABLE-C3CB
        '50124' => '14126'            , # G+14126  U+C3CC  uniC3CC,  HANGUL SYLLABLE-C3CC
        '50125' => '14127'            , # G+14127  U+C3CD  uniC3CD,  HANGUL SYLLABLE-C3CD
        '50126' => '14128'            , # G+14128  U+C3CE  uniC3CE,  HANGUL SYLLABLE-C3CE
        '50127' => '14129'            , # G+14129  U+C3CF  uniC3CF,  HANGUL SYLLABLE-C3CF
        '50128' => '14130'            , # G+14130  U+C3D0  uniC3D0,  HANGUL SYLLABLE-C3D0
        '50129' => '14131'            , # G+14131  U+C3D1  uniC3D1,  HANGUL SYLLABLE-C3D1
        '50130' => '14132'            , # G+14132  U+C3D2  uniC3D2,  HANGUL SYLLABLE-C3D2
        '50131' => '14133'            , # G+14133  U+C3D3  uniC3D3,  HANGUL SYLLABLE-C3D3
        '50132' => '14134'            , # G+14134  U+C3D4  uniC3D4,  HANGUL SYLLABLE-C3D4
        '50133' => '14135'            , # G+14135  U+C3D5  uniC3D5,  HANGUL SYLLABLE-C3D5
        '50134' => '14136'            , # G+14136  U+C3D6  uniC3D6,  HANGUL SYLLABLE-C3D6
        '50135' => '14137'            , # G+14137  U+C3D7  uniC3D7,  HANGUL SYLLABLE-C3D7
        '50136' => '2385'             , # G+2385  U+C3D8  uniC3D8,  HANGUL SYLLABLE-C3D8
        '50137' => '2386'             , # G+2386  U+C3D9  uniC3D9,  HANGUL SYLLABLE-C3D9
        '50138' => '14138'            , # G+14138  U+C3DA  uniC3DA,  HANGUL SYLLABLE-C3DA
        '50139' => '14139'            , # G+14139  U+C3DB  uniC3DB,  HANGUL SYLLABLE-C3DB
        '50140' => '2387'             , # G+2387  U+C3DC  uniC3DC,  HANGUL SYLLABLE-C3DC
        '50141' => '14140'            , # G+14140  U+C3DD  uniC3DD,  HANGUL SYLLABLE-C3DD
        '50142' => '14141'            , # G+14141  U+C3DE  uniC3DE,  HANGUL SYLLABLE-C3DE
        '50143' => '2388'             , # G+2388  U+C3DF  uniC3DF,  HANGUL SYLLABLE-C3DF
        '50144' => '2389'             , # G+2389  U+C3E0  uniC3E0,  HANGUL SYLLABLE-C3E0
        '50145' => '14142'            , # G+14142  U+C3E1  uniC3E1,  HANGUL SYLLABLE-C3E1
        '50146' => '2390'             , # G+2390  U+C3E2  uniC3E2,  HANGUL SYLLABLE-C3E2
        '50147' => '14143'            , # G+14143  U+C3E3  uniC3E3,  HANGUL SYLLABLE-C3E3
        '50148' => '14144'            , # G+14144  U+C3E4  uniC3E4,  HANGUL SYLLABLE-C3E4
        '50149' => '14145'            , # G+14145  U+C3E5  uniC3E5,  HANGUL SYLLABLE-C3E5
        '50150' => '14146'            , # G+14146  U+C3E6  uniC3E6,  HANGUL SYLLABLE-C3E6
        '50151' => '14147'            , # G+14147  U+C3E7  uniC3E7,  HANGUL SYLLABLE-C3E7
        '50152' => '2391'             , # G+2391  U+C3E8  uniC3E8,  HANGUL SYLLABLE-C3E8
        '50153' => '2392'             , # G+2392  U+C3E9  uniC3E9,  HANGUL SYLLABLE-C3E9
        '50154' => '14148'            , # G+14148  U+C3EA  uniC3EA,  HANGUL SYLLABLE-C3EA
        '50155' => '14149'            , # G+14149  U+C3EB  uniC3EB,  HANGUL SYLLABLE-C3EB
        '50156' => '14150'            , # G+14150  U+C3EC  uniC3EC,  HANGUL SYLLABLE-C3EC
        '50157' => '2393'             , # G+2393  U+C3ED  uniC3ED,  HANGUL SYLLABLE-C3ED
        '50158' => '14151'            , # G+14151  U+C3EE  uniC3EE,  HANGUL SYLLABLE-C3EE
        '50159' => '14152'            , # G+14152  U+C3EF  uniC3EF,  HANGUL SYLLABLE-C3EF
        '50160' => '14153'            , # G+14153  U+C3F0  uniC3F0,  HANGUL SYLLABLE-C3F0
        '50161' => '14154'            , # G+14154  U+C3F1  uniC3F1,  HANGUL SYLLABLE-C3F1
        '50162' => '14155'            , # G+14155  U+C3F2  uniC3F2,  HANGUL SYLLABLE-C3F2
        '50163' => '14156'            , # G+14156  U+C3F3  uniC3F3,  HANGUL SYLLABLE-C3F3
        '50164' => '2394'             , # G+2394  U+C3F4  uniC3F4,  HANGUL SYLLABLE-C3F4
        '50165' => '2395'             , # G+2395  U+C3F5  uniC3F5,  HANGUL SYLLABLE-C3F5
        '50166' => '14157'            , # G+14157  U+C3F6  uniC3F6,  HANGUL SYLLABLE-C3F6
        '50167' => '14158'            , # G+14158  U+C3F7  uniC3F7,  HANGUL SYLLABLE-C3F7
        '50168' => '2396'             , # G+2396  U+C3F8  uniC3F8,  HANGUL SYLLABLE-C3F8
        '50169' => '14159'            , # G+14159  U+C3F9  uniC3F9,  HANGUL SYLLABLE-C3F9
        '50170' => '14160'            , # G+14160  U+C3FA  uniC3FA,  HANGUL SYLLABLE-C3FA
        '50171' => '14161'            , # G+14161  U+C3FB  uniC3FB,  HANGUL SYLLABLE-C3FB
        '50172' => '14162'            , # G+14162  U+C3FC  uniC3FC,  HANGUL SYLLABLE-C3FC
        '50173' => '14163'            , # G+14163  U+C3FD  uniC3FD,  HANGUL SYLLABLE-C3FD
        '50174' => '14164'            , # G+14164  U+C3FE  uniC3FE,  HANGUL SYLLABLE-C3FE
        '50175' => '14165'            , # G+14165  U+C3FF  uniC3FF,  HANGUL SYLLABLE-C3FF
        '50176' => '14166'            , # G+14166  U+C400  uniC400,  HANGUL SYLLABLE-C400
        '50177' => '14167'            , # G+14167  U+C401  uniC401,  HANGUL SYLLABLE-C401
        '50178' => '14168'            , # G+14168  U+C402  uniC402,  HANGUL SYLLABLE-C402
        '50179' => '14169'            , # G+14169  U+C403  uniC403,  HANGUL SYLLABLE-C403
        '50180' => '14170'            , # G+14170  U+C404  uniC404,  HANGUL SYLLABLE-C404
        '50181' => '14171'            , # G+14171  U+C405  uniC405,  HANGUL SYLLABLE-C405
        '50182' => '14172'            , # G+14172  U+C406  uniC406,  HANGUL SYLLABLE-C406
        '50183' => '14173'            , # G+14173  U+C407  uniC407,  HANGUL SYLLABLE-C407
        '50184' => '2397'             , # G+2397  U+C408  uniC408,  HANGUL SYLLABLE-C408
        '50185' => '14174'            , # G+14174  U+C409  uniC409,  HANGUL SYLLABLE-C409
        '50186' => '14175'            , # G+14175  U+C40A  uniC40A,  HANGUL SYLLABLE-C40A
        '50187' => '14176'            , # G+14176  U+C40B  uniC40B,  HANGUL SYLLABLE-C40B
        '50188' => '14177'            , # G+14177  U+C40C  uniC40C,  HANGUL SYLLABLE-C40C
        '50189' => '14178'            , # G+14178  U+C40D  uniC40D,  HANGUL SYLLABLE-C40D
        '50190' => '14179'            , # G+14179  U+C40E  uniC40E,  HANGUL SYLLABLE-C40E
        '50191' => '14180'            , # G+14180  U+C40F  uniC40F,  HANGUL SYLLABLE-C40F
        '50192' => '2398'             , # G+2398  U+C410  uniC410,  HANGUL SYLLABLE-C410
        '50193' => '14181'            , # G+14181  U+C411  uniC411,  HANGUL SYLLABLE-C411
        '50194' => '14182'            , # G+14182  U+C412  uniC412,  HANGUL SYLLABLE-C412
        '50195' => '14183'            , # G+14183  U+C413  uniC413,  HANGUL SYLLABLE-C413
        '50196' => '14184'            , # G+14184  U+C414  uniC414,  HANGUL SYLLABLE-C414
        '50197' => '14185'            , # G+14185  U+C415  uniC415,  HANGUL SYLLABLE-C415
        '50198' => '14186'            , # G+14186  U+C416  uniC416,  HANGUL SYLLABLE-C416
        '50199' => '14187'            , # G+14187  U+C417  uniC417,  HANGUL SYLLABLE-C417
        '50200' => '14188'            , # G+14188  U+C418  uniC418,  HANGUL SYLLABLE-C418
        '50201' => '14189'            , # G+14189  U+C419  uniC419,  HANGUL SYLLABLE-C419
        '50202' => '14190'            , # G+14190  U+C41A  uniC41A,  HANGUL SYLLABLE-C41A
        '50203' => '14191'            , # G+14191  U+C41B  uniC41B,  HANGUL SYLLABLE-C41B
        '50204' => '14192'            , # G+14192  U+C41C  uniC41C,  HANGUL SYLLABLE-C41C
        '50205' => '14193'            , # G+14193  U+C41D  uniC41D,  HANGUL SYLLABLE-C41D
        '50206' => '14194'            , # G+14194  U+C41E  uniC41E,  HANGUL SYLLABLE-C41E
        '50207' => '14195'            , # G+14195  U+C41F  uniC41F,  HANGUL SYLLABLE-C41F
        '50208' => '14196'            , # G+14196  U+C420  uniC420,  HANGUL SYLLABLE-C420
        '50209' => '14197'            , # G+14197  U+C421  uniC421,  HANGUL SYLLABLE-C421
        '50210' => '14198'            , # G+14198  U+C422  uniC422,  HANGUL SYLLABLE-C422
        '50211' => '14199'            , # G+14199  U+C423  uniC423,  HANGUL SYLLABLE-C423
        '50212' => '2399'             , # G+2399  U+C424  uniC424,  HANGUL SYLLABLE-C424
        '50213' => '14200'            , # G+14200  U+C425  uniC425,  HANGUL SYLLABLE-C425
        '50214' => '14201'            , # G+14201  U+C426  uniC426,  HANGUL SYLLABLE-C426
        '50215' => '14202'            , # G+14202  U+C427  uniC427,  HANGUL SYLLABLE-C427
        '50216' => '14203'            , # G+14203  U+C428  uniC428,  HANGUL SYLLABLE-C428
        '50217' => '14204'            , # G+14204  U+C429  uniC429,  HANGUL SYLLABLE-C429
        '50218' => '14205'            , # G+14205  U+C42A  uniC42A,  HANGUL SYLLABLE-C42A
        '50219' => '14206'            , # G+14206  U+C42B  uniC42B,  HANGUL SYLLABLE-C42B
        '50220' => '2400'             , # G+2400  U+C42C  uniC42C,  HANGUL SYLLABLE-C42C
        '50221' => '14207'            , # G+14207  U+C42D  uniC42D,  HANGUL SYLLABLE-C42D
        '50222' => '14208'            , # G+14208  U+C42E  uniC42E,  HANGUL SYLLABLE-C42E
        '50223' => '14209'            , # G+14209  U+C42F  uniC42F,  HANGUL SYLLABLE-C42F
        '50224' => '2401'             , # G+2401  U+C430  uniC430,  HANGUL SYLLABLE-C430
        '50225' => '14210'            , # G+14210  U+C431  uniC431,  HANGUL SYLLABLE-C431
        '50226' => '14211'            , # G+14211  U+C432  uniC432,  HANGUL SYLLABLE-C432
        '50227' => '14212'            , # G+14212  U+C433  uniC433,  HANGUL SYLLABLE-C433
        '50228' => '2402'             , # G+2402  U+C434  uniC434,  HANGUL SYLLABLE-C434
        '50229' => '14213'            , # G+14213  U+C435  uniC435,  HANGUL SYLLABLE-C435
        '50230' => '14214'            , # G+14214  U+C436  uniC436,  HANGUL SYLLABLE-C436
        '50231' => '14215'            , # G+14215  U+C437  uniC437,  HANGUL SYLLABLE-C437
        '50232' => '14216'            , # G+14216  U+C438  uniC438,  HANGUL SYLLABLE-C438
        '50233' => '14217'            , # G+14217  U+C439  uniC439,  HANGUL SYLLABLE-C439
        '50234' => '14218'            , # G+14218  U+C43A  uniC43A,  HANGUL SYLLABLE-C43A
        '50235' => '14219'            , # G+14219  U+C43B  uniC43B,  HANGUL SYLLABLE-C43B
        '50236' => '2403'             , # G+2403  U+C43C  uniC43C,  HANGUL SYLLABLE-C43C
        '50237' => '2404'             , # G+2404  U+C43D  uniC43D,  HANGUL SYLLABLE-C43D
        '50238' => '14220'            , # G+14220  U+C43E  uniC43E,  HANGUL SYLLABLE-C43E
        '50239' => '14221'            , # G+14221  U+C43F  uniC43F,  HANGUL SYLLABLE-C43F
        '50240' => '14222'            , # G+14222  U+C440  uniC440,  HANGUL SYLLABLE-C440
        '50241' => '14223'            , # G+14223  U+C441  uniC441,  HANGUL SYLLABLE-C441
        '50242' => '14224'            , # G+14224  U+C442  uniC442,  HANGUL SYLLABLE-C442
        '50243' => '14225'            , # G+14225  U+C443  uniC443,  HANGUL SYLLABLE-C443
        '50244' => '14226'            , # G+14226  U+C444  uniC444,  HANGUL SYLLABLE-C444
        '50245' => '14227'            , # G+14227  U+C445  uniC445,  HANGUL SYLLABLE-C445
        '50246' => '14228'            , # G+14228  U+C446  uniC446,  HANGUL SYLLABLE-C446
        '50247' => '14229'            , # G+14229  U+C447  uniC447,  HANGUL SYLLABLE-C447
        '50248' => '2405'             , # G+2405  U+C448  uniC448,  HANGUL SYLLABLE-C448
        '50249' => '14230'            , # G+14230  U+C449  uniC449,  HANGUL SYLLABLE-C449
        '50250' => '14231'            , # G+14231  U+C44A  uniC44A,  HANGUL SYLLABLE-C44A
        '50251' => '14232'            , # G+14232  U+C44B  uniC44B,  HANGUL SYLLABLE-C44B
        '50252' => '14233'            , # G+14233  U+C44C  uniC44C,  HANGUL SYLLABLE-C44C
        '50253' => '14234'            , # G+14234  U+C44D  uniC44D,  HANGUL SYLLABLE-C44D
        '50254' => '14235'            , # G+14235  U+C44E  uniC44E,  HANGUL SYLLABLE-C44E
        '50255' => '14236'            , # G+14236  U+C44F  uniC44F,  HANGUL SYLLABLE-C44F
        '50256' => '14237'            , # G+14237  U+C450  uniC450,  HANGUL SYLLABLE-C450
        '50257' => '14238'            , # G+14238  U+C451  uniC451,  HANGUL SYLLABLE-C451
        '50258' => '14239'            , # G+14239  U+C452  uniC452,  HANGUL SYLLABLE-C452
        '50259' => '14240'            , # G+14240  U+C453  uniC453,  HANGUL SYLLABLE-C453
        '50260' => '14241'            , # G+14241  U+C454  uniC454,  HANGUL SYLLABLE-C454
        '50261' => '14242'            , # G+14242  U+C455  uniC455,  HANGUL SYLLABLE-C455
        '50262' => '14243'            , # G+14243  U+C456  uniC456,  HANGUL SYLLABLE-C456
        '50263' => '14244'            , # G+14244  U+C457  uniC457,  HANGUL SYLLABLE-C457
        '50264' => '14245'            , # G+14245  U+C458  uniC458,  HANGUL SYLLABLE-C458
        '50265' => '14246'            , # G+14246  U+C459  uniC459,  HANGUL SYLLABLE-C459
        '50266' => '14247'            , # G+14247  U+C45A  uniC45A,  HANGUL SYLLABLE-C45A
        '50267' => '14248'            , # G+14248  U+C45B  uniC45B,  HANGUL SYLLABLE-C45B
        '50268' => '14249'            , # G+14249  U+C45C  uniC45C,  HANGUL SYLLABLE-C45C
        '50269' => '14250'            , # G+14250  U+C45D  uniC45D,  HANGUL SYLLABLE-C45D
        '50270' => '14251'            , # G+14251  U+C45E  uniC45E,  HANGUL SYLLABLE-C45E
        '50271' => '14252'            , # G+14252  U+C45F  uniC45F,  HANGUL SYLLABLE-C45F
        '50272' => '14253'            , # G+14253  U+C460  uniC460,  HANGUL SYLLABLE-C460
        '50273' => '14254'            , # G+14254  U+C461  uniC461,  HANGUL SYLLABLE-C461
        '50274' => '14255'            , # G+14255  U+C462  uniC462,  HANGUL SYLLABLE-C462
        '50275' => '14256'            , # G+14256  U+C463  uniC463,  HANGUL SYLLABLE-C463
        '50276' => '2406'             , # G+2406  U+C464  uniC464,  HANGUL SYLLABLE-C464
        '50277' => '2407'             , # G+2407  U+C465  uniC465,  HANGUL SYLLABLE-C465
        '50278' => '14257'            , # G+14257  U+C466  uniC466,  HANGUL SYLLABLE-C466
        '50279' => '14258'            , # G+14258  U+C467  uniC467,  HANGUL SYLLABLE-C467
        '50280' => '2408'             , # G+2408  U+C468  uniC468,  HANGUL SYLLABLE-C468
        '50281' => '14259'            , # G+14259  U+C469  uniC469,  HANGUL SYLLABLE-C469
        '50282' => '14260'            , # G+14260  U+C46A  uniC46A,  HANGUL SYLLABLE-C46A
        '50283' => '14261'            , # G+14261  U+C46B  uniC46B,  HANGUL SYLLABLE-C46B
        '50284' => '2409'             , # G+2409  U+C46C  uniC46C,  HANGUL SYLLABLE-C46C
        '50285' => '14262'            , # G+14262  U+C46D  uniC46D,  HANGUL SYLLABLE-C46D
        '50286' => '14263'            , # G+14263  U+C46E  uniC46E,  HANGUL SYLLABLE-C46E
        '50287' => '14264'            , # G+14264  U+C46F  uniC46F,  HANGUL SYLLABLE-C46F
        '50288' => '14265'            , # G+14265  U+C470  uniC470,  HANGUL SYLLABLE-C470
        '50289' => '14266'            , # G+14266  U+C471  uniC471,  HANGUL SYLLABLE-C471
        '50290' => '14267'            , # G+14267  U+C472  uniC472,  HANGUL SYLLABLE-C472
        '50291' => '14268'            , # G+14268  U+C473  uniC473,  HANGUL SYLLABLE-C473
        '50292' => '2410'             , # G+2410  U+C474  uniC474,  HANGUL SYLLABLE-C474
        '50293' => '2411'             , # G+2411  U+C475  uniC475,  HANGUL SYLLABLE-C475
        '50294' => '14269'            , # G+14269  U+C476  uniC476,  HANGUL SYLLABLE-C476
        '50295' => '14270'            , # G+14270  U+C477  uniC477,  HANGUL SYLLABLE-C477
        '50296' => '14271'            , # G+14271  U+C478  uniC478,  HANGUL SYLLABLE-C478
        '50297' => '2412'             , # G+2412  U+C479  uniC479,  HANGUL SYLLABLE-C479
        '50298' => '14272'            , # G+14272  U+C47A  uniC47A,  HANGUL SYLLABLE-C47A
        '50299' => '14273'            , # G+14273  U+C47B  uniC47B,  HANGUL SYLLABLE-C47B
        '50300' => '14274'            , # G+14274  U+C47C  uniC47C,  HANGUL SYLLABLE-C47C
        '50301' => '14275'            , # G+14275  U+C47D  uniC47D,  HANGUL SYLLABLE-C47D
        '50302' => '14276'            , # G+14276  U+C47E  uniC47E,  HANGUL SYLLABLE-C47E
        '50303' => '14277'            , # G+14277  U+C47F  uniC47F,  HANGUL SYLLABLE-C47F
        '50304' => '2413'             , # G+2413  U+C480  uniC480,  HANGUL SYLLABLE-C480
        '50305' => '14278'            , # G+14278  U+C481  uniC481,  HANGUL SYLLABLE-C481
        '50306' => '14279'            , # G+14279  U+C482  uniC482,  HANGUL SYLLABLE-C482
        '50307' => '14280'            , # G+14280  U+C483  uniC483,  HANGUL SYLLABLE-C483
        '50308' => '14281'            , # G+14281  U+C484  uniC484,  HANGUL SYLLABLE-C484
        '50309' => '14282'            , # G+14282  U+C485  uniC485,  HANGUL SYLLABLE-C485
        '50310' => '14283'            , # G+14283  U+C486  uniC486,  HANGUL SYLLABLE-C486
        '50311' => '14284'            , # G+14284  U+C487  uniC487,  HANGUL SYLLABLE-C487
        '50312' => '14285'            , # G+14285  U+C488  uniC488,  HANGUL SYLLABLE-C488
        '50313' => '14286'            , # G+14286  U+C489  uniC489,  HANGUL SYLLABLE-C489
        '50314' => '14287'            , # G+14287  U+C48A  uniC48A,  HANGUL SYLLABLE-C48A
        '50315' => '14288'            , # G+14288  U+C48B  uniC48B,  HANGUL SYLLABLE-C48B
        '50316' => '14289'            , # G+14289  U+C48C  uniC48C,  HANGUL SYLLABLE-C48C
        '50317' => '14290'            , # G+14290  U+C48D  uniC48D,  HANGUL SYLLABLE-C48D
        '50318' => '14291'            , # G+14291  U+C48E  uniC48E,  HANGUL SYLLABLE-C48E
        '50319' => '14292'            , # G+14292  U+C48F  uniC48F,  HANGUL SYLLABLE-C48F
        '50320' => '14293'            , # G+14293  U+C490  uniC490,  HANGUL SYLLABLE-C490
        '50321' => '14294'            , # G+14294  U+C491  uniC491,  HANGUL SYLLABLE-C491
        '50322' => '14295'            , # G+14295  U+C492  uniC492,  HANGUL SYLLABLE-C492
        '50323' => '14296'            , # G+14296  U+C493  uniC493,  HANGUL SYLLABLE-C493
        '50324' => '2414'             , # G+2414  U+C494  uniC494,  HANGUL SYLLABLE-C494
        '50325' => '14297'            , # G+14297  U+C495  uniC495,  HANGUL SYLLABLE-C495
        '50326' => '14298'            , # G+14298  U+C496  uniC496,  HANGUL SYLLABLE-C496
        '50327' => '14299'            , # G+14299  U+C497  uniC497,  HANGUL SYLLABLE-C497
        '50328' => '14300'            , # G+14300  U+C498  uniC498,  HANGUL SYLLABLE-C498
        '50329' => '14301'            , # G+14301  U+C499  uniC499,  HANGUL SYLLABLE-C499
        '50330' => '14302'            , # G+14302  U+C49A  uniC49A,  HANGUL SYLLABLE-C49A
        '50331' => '14303'            , # G+14303  U+C49B  uniC49B,  HANGUL SYLLABLE-C49B
        '50332' => '2415'             , # G+2415  U+C49C  uniC49C,  HANGUL SYLLABLE-C49C
        '50333' => '14304'            , # G+14304  U+C49D  uniC49D,  HANGUL SYLLABLE-C49D
        '50334' => '14305'            , # G+14305  U+C49E  uniC49E,  HANGUL SYLLABLE-C49E
        '50335' => '14306'            , # G+14306  U+C49F  uniC49F,  HANGUL SYLLABLE-C49F
        '50336' => '14307'            , # G+14307  U+C4A0  uniC4A0,  HANGUL SYLLABLE-C4A0
        '50337' => '14308'            , # G+14308  U+C4A1  uniC4A1,  HANGUL SYLLABLE-C4A1
        '50338' => '14309'            , # G+14309  U+C4A2  uniC4A2,  HANGUL SYLLABLE-C4A2
        '50339' => '14310'            , # G+14310  U+C4A3  uniC4A3,  HANGUL SYLLABLE-C4A3
        '50340' => '14311'            , # G+14311  U+C4A4  uniC4A4,  HANGUL SYLLABLE-C4A4
        '50341' => '14312'            , # G+14312  U+C4A5  uniC4A5,  HANGUL SYLLABLE-C4A5
        '50342' => '14313'            , # G+14313  U+C4A6  uniC4A6,  HANGUL SYLLABLE-C4A6
        '50343' => '14314'            , # G+14314  U+C4A7  uniC4A7,  HANGUL SYLLABLE-C4A7
        '50344' => '14315'            , # G+14315  U+C4A8  uniC4A8,  HANGUL SYLLABLE-C4A8
        '50345' => '14316'            , # G+14316  U+C4A9  uniC4A9,  HANGUL SYLLABLE-C4A9
        '50346' => '14317'            , # G+14317  U+C4AA  uniC4AA,  HANGUL SYLLABLE-C4AA
        '50347' => '14318'            , # G+14318  U+C4AB  uniC4AB,  HANGUL SYLLABLE-C4AB
        '50348' => '14319'            , # G+14319  U+C4AC  uniC4AC,  HANGUL SYLLABLE-C4AC
        '50349' => '14320'            , # G+14320  U+C4AD  uniC4AD,  HANGUL SYLLABLE-C4AD
        '50350' => '14321'            , # G+14321  U+C4AE  uniC4AE,  HANGUL SYLLABLE-C4AE
        '50351' => '14322'            , # G+14322  U+C4AF  uniC4AF,  HANGUL SYLLABLE-C4AF
        '50352' => '14323'            , # G+14323  U+C4B0  uniC4B0,  HANGUL SYLLABLE-C4B0
        '50353' => '14324'            , # G+14324  U+C4B1  uniC4B1,  HANGUL SYLLABLE-C4B1
        '50354' => '14325'            , # G+14325  U+C4B2  uniC4B2,  HANGUL SYLLABLE-C4B2
        '50355' => '14326'            , # G+14326  U+C4B3  uniC4B3,  HANGUL SYLLABLE-C4B3
        '50356' => '14327'            , # G+14327  U+C4B4  uniC4B4,  HANGUL SYLLABLE-C4B4
        '50357' => '14328'            , # G+14328  U+C4B5  uniC4B5,  HANGUL SYLLABLE-C4B5
        '50358' => '14329'            , # G+14329  U+C4B6  uniC4B6,  HANGUL SYLLABLE-C4B6
        '50359' => '14330'            , # G+14330  U+C4B7  uniC4B7,  HANGUL SYLLABLE-C4B7
        '50360' => '2416'             , # G+2416  U+C4B8  uniC4B8,  HANGUL SYLLABLE-C4B8
        '50361' => '14331'            , # G+14331  U+C4B9  uniC4B9,  HANGUL SYLLABLE-C4B9
        '50362' => '14332'            , # G+14332  U+C4BA  uniC4BA,  HANGUL SYLLABLE-C4BA
        '50363' => '14333'            , # G+14333  U+C4BB  uniC4BB,  HANGUL SYLLABLE-C4BB
        '50364' => '2417'             , # G+2417  U+C4BC  uniC4BC,  HANGUL SYLLABLE-C4BC
        '50365' => '14334'            , # G+14334  U+C4BD  uniC4BD,  HANGUL SYLLABLE-C4BD
        '50366' => '14335'            , # G+14335  U+C4BE  uniC4BE,  HANGUL SYLLABLE-C4BE
        '50367' => '14336'            , # G+14336  U+C4BF  uniC4BF,  HANGUL SYLLABLE-C4BF
        '50368' => '14337'            , # G+14337  U+C4C0  uniC4C0,  HANGUL SYLLABLE-C4C0
        '50369' => '14338'            , # G+14338  U+C4C1  uniC4C1,  HANGUL SYLLABLE-C4C1
        '50370' => '14339'            , # G+14339  U+C4C2  uniC4C2,  HANGUL SYLLABLE-C4C2
        '50371' => '14340'            , # G+14340  U+C4C3  uniC4C3,  HANGUL SYLLABLE-C4C3
        '50372' => '14341'            , # G+14341  U+C4C4  uniC4C4,  HANGUL SYLLABLE-C4C4
        '50373' => '14342'            , # G+14342  U+C4C5  uniC4C5,  HANGUL SYLLABLE-C4C5
        '50374' => '14343'            , # G+14343  U+C4C6  uniC4C6,  HANGUL SYLLABLE-C4C6
        '50375' => '14344'            , # G+14344  U+C4C7  uniC4C7,  HANGUL SYLLABLE-C4C7
        '50376' => '14345'            , # G+14345  U+C4C8  uniC4C8,  HANGUL SYLLABLE-C4C8
        '50377' => '14346'            , # G+14346  U+C4C9  uniC4C9,  HANGUL SYLLABLE-C4C9
        '50378' => '14347'            , # G+14347  U+C4CA  uniC4CA,  HANGUL SYLLABLE-C4CA
        '50379' => '14348'            , # G+14348  U+C4CB  uniC4CB,  HANGUL SYLLABLE-C4CB
        '50380' => '14349'            , # G+14349  U+C4CC  uniC4CC,  HANGUL SYLLABLE-C4CC
        '50381' => '14350'            , # G+14350  U+C4CD  uniC4CD,  HANGUL SYLLABLE-C4CD
        '50382' => '14351'            , # G+14351  U+C4CE  uniC4CE,  HANGUL SYLLABLE-C4CE
        '50383' => '14352'            , # G+14352  U+C4CF  uniC4CF,  HANGUL SYLLABLE-C4CF
        '50384' => '14353'            , # G+14353  U+C4D0  uniC4D0,  HANGUL SYLLABLE-C4D0
        '50385' => '14354'            , # G+14354  U+C4D1  uniC4D1,  HANGUL SYLLABLE-C4D1
        '50386' => '14355'            , # G+14355  U+C4D2  uniC4D2,  HANGUL SYLLABLE-C4D2
        '50387' => '14356'            , # G+14356  U+C4D3  uniC4D3,  HANGUL SYLLABLE-C4D3
        '50388' => '14357'            , # G+14357  U+C4D4  uniC4D4,  HANGUL SYLLABLE-C4D4
        '50389' => '14358'            , # G+14358  U+C4D5  uniC4D5,  HANGUL SYLLABLE-C4D5
        '50390' => '14359'            , # G+14359  U+C4D6  uniC4D6,  HANGUL SYLLABLE-C4D6
        '50391' => '14360'            , # G+14360  U+C4D7  uniC4D7,  HANGUL SYLLABLE-C4D7
        '50392' => '14361'            , # G+14361  U+C4D8  uniC4D8,  HANGUL SYLLABLE-C4D8
        '50393' => '14362'            , # G+14362  U+C4D9  uniC4D9,  HANGUL SYLLABLE-C4D9
        '50394' => '14363'            , # G+14363  U+C4DA  uniC4DA,  HANGUL SYLLABLE-C4DA
        '50395' => '14364'            , # G+14364  U+C4DB  uniC4DB,  HANGUL SYLLABLE-C4DB
        '50396' => '14365'            , # G+14365  U+C4DC  uniC4DC,  HANGUL SYLLABLE-C4DC
        '50397' => '14366'            , # G+14366  U+C4DD  uniC4DD,  HANGUL SYLLABLE-C4DD
        '50398' => '14367'            , # G+14367  U+C4DE  uniC4DE,  HANGUL SYLLABLE-C4DE
        '50399' => '14368'            , # G+14368  U+C4DF  uniC4DF,  HANGUL SYLLABLE-C4DF
        '50400' => '14369'            , # G+14369  U+C4E0  uniC4E0,  HANGUL SYLLABLE-C4E0
        '50401' => '14370'            , # G+14370  U+C4E1  uniC4E1,  HANGUL SYLLABLE-C4E1
        '50402' => '14371'            , # G+14371  U+C4E2  uniC4E2,  HANGUL SYLLABLE-C4E2
        '50403' => '14372'            , # G+14372  U+C4E3  uniC4E3,  HANGUL SYLLABLE-C4E3
        '50404' => '14373'            , # G+14373  U+C4E4  uniC4E4,  HANGUL SYLLABLE-C4E4
        '50405' => '14374'            , # G+14374  U+C4E5  uniC4E5,  HANGUL SYLLABLE-C4E5
        '50406' => '14375'            , # G+14375  U+C4E6  uniC4E6,  HANGUL SYLLABLE-C4E6
        '50407' => '14376'            , # G+14376  U+C4E7  uniC4E7,  HANGUL SYLLABLE-C4E7
        '50408' => '14377'            , # G+14377  U+C4E8  uniC4E8,  HANGUL SYLLABLE-C4E8
        '50409' => '2418'             , # G+2418  U+C4E9  uniC4E9,  HANGUL SYLLABLE-C4E9
        '50410' => '14378'            , # G+14378  U+C4EA  uniC4EA,  HANGUL SYLLABLE-C4EA
        '50411' => '14379'            , # G+14379  U+C4EB  uniC4EB,  HANGUL SYLLABLE-C4EB
        '50412' => '14380'            , # G+14380  U+C4EC  uniC4EC,  HANGUL SYLLABLE-C4EC
        '50413' => '14381'            , # G+14381  U+C4ED  uniC4ED,  HANGUL SYLLABLE-C4ED
        '50414' => '14382'            , # G+14382  U+C4EE  uniC4EE,  HANGUL SYLLABLE-C4EE
        '50415' => '14383'            , # G+14383  U+C4EF  uniC4EF,  HANGUL SYLLABLE-C4EF
        '50416' => '2419'             , # G+2419  U+C4F0  uniC4F0,  HANGUL SYLLABLE-C4F0
        '50417' => '2420'             , # G+2420  U+C4F1  uniC4F1,  HANGUL SYLLABLE-C4F1
        '50418' => '14384'            , # G+14384  U+C4F2  uniC4F2,  HANGUL SYLLABLE-C4F2
        '50419' => '14385'            , # G+14385  U+C4F3  uniC4F3,  HANGUL SYLLABLE-C4F3
        '50420' => '2421'             , # G+2421  U+C4F4  uniC4F4,  HANGUL SYLLABLE-C4F4
        '50421' => '14386'            , # G+14386  U+C4F5  uniC4F5,  HANGUL SYLLABLE-C4F5
        '50422' => '14387'            , # G+14387  U+C4F6  uniC4F6,  HANGUL SYLLABLE-C4F6
        '50423' => '14388'            , # G+14388  U+C4F7  uniC4F7,  HANGUL SYLLABLE-C4F7
        '50424' => '2422'             , # G+2422  U+C4F8  uniC4F8,  HANGUL SYLLABLE-C4F8
        '50425' => '14389'            , # G+14389  U+C4F9  uniC4F9,  HANGUL SYLLABLE-C4F9
        '50426' => '2423'             , # G+2423  U+C4FA  uniC4FA,  HANGUL SYLLABLE-C4FA
        '50427' => '14390'            , # G+14390  U+C4FB  uniC4FB,  HANGUL SYLLABLE-C4FB
        '50428' => '14391'            , # G+14391  U+C4FC  uniC4FC,  HANGUL SYLLABLE-C4FC
        '50429' => '14392'            , # G+14392  U+C4FD  uniC4FD,  HANGUL SYLLABLE-C4FD
        '50430' => '14393'            , # G+14393  U+C4FE  uniC4FE,  HANGUL SYLLABLE-C4FE
        '50431' => '2424'             , # G+2424  U+C4FF  uniC4FF,  HANGUL SYLLABLE-C4FF
        '50432' => '2425'             , # G+2425  U+C500  uniC500,  HANGUL SYLLABLE-C500
        '50433' => '2426'             , # G+2426  U+C501  uniC501,  HANGUL SYLLABLE-C501
        '50434' => '14394'            , # G+14394  U+C502  uniC502,  HANGUL SYLLABLE-C502
        '50435' => '14395'            , # G+14395  U+C503  uniC503,  HANGUL SYLLABLE-C503
        '50436' => '14396'            , # G+14396  U+C504  uniC504,  HANGUL SYLLABLE-C504
        '50437' => '14397'            , # G+14397  U+C505  uniC505,  HANGUL SYLLABLE-C505
        '50438' => '14398'            , # G+14398  U+C506  uniC506,  HANGUL SYLLABLE-C506
        '50439' => '14399'            , # G+14399  U+C507  uniC507,  HANGUL SYLLABLE-C507
        '50440' => '14400'            , # G+14400  U+C508  uniC508,  HANGUL SYLLABLE-C508
        '50441' => '14401'            , # G+14401  U+C509  uniC509,  HANGUL SYLLABLE-C509
        '50442' => '14402'            , # G+14402  U+C50A  uniC50A,  HANGUL SYLLABLE-C50A
        '50443' => '14403'            , # G+14403  U+C50B  uniC50B,  HANGUL SYLLABLE-C50B
        '50444' => '2427'             , # G+2427  U+C50C  uniC50C,  HANGUL SYLLABLE-C50C
        '50445' => '14404'            , # G+14404  U+C50D  uniC50D,  HANGUL SYLLABLE-C50D
        '50446' => '14405'            , # G+14405  U+C50E  uniC50E,  HANGUL SYLLABLE-C50E
        '50447' => '14406'            , # G+14406  U+C50F  uniC50F,  HANGUL SYLLABLE-C50F
        '50448' => '2428'             , # G+2428  U+C510  uniC510,  HANGUL SYLLABLE-C510
        '50449' => '14407'            , # G+14407  U+C511  uniC511,  HANGUL SYLLABLE-C511
        '50450' => '14408'            , # G+14408  U+C512  uniC512,  HANGUL SYLLABLE-C512
        '50451' => '14409'            , # G+14409  U+C513  uniC513,  HANGUL SYLLABLE-C513
        '50452' => '2429'             , # G+2429  U+C514  uniC514,  HANGUL SYLLABLE-C514
        '50453' => '14410'            , # G+14410  U+C515  uniC515,  HANGUL SYLLABLE-C515
        '50454' => '14411'            , # G+14411  U+C516  uniC516,  HANGUL SYLLABLE-C516
        '50455' => '14412'            , # G+14412  U+C517  uniC517,  HANGUL SYLLABLE-C517
        '50456' => '14413'            , # G+14413  U+C518  uniC518,  HANGUL SYLLABLE-C518
        '50457' => '14414'            , # G+14414  U+C519  uniC519,  HANGUL SYLLABLE-C519
        '50458' => '14415'            , # G+14415  U+C51A  uniC51A,  HANGUL SYLLABLE-C51A
        '50459' => '14416'            , # G+14416  U+C51B  uniC51B,  HANGUL SYLLABLE-C51B
        '50460' => '2430'             , # G+2430  U+C51C  uniC51C,  HANGUL SYLLABLE-C51C
        '50461' => '14417'            , # G+14417  U+C51D  uniC51D,  HANGUL SYLLABLE-C51D
        '50462' => '14418'            , # G+14418  U+C51E  uniC51E,  HANGUL SYLLABLE-C51E
        '50463' => '14419'            , # G+14419  U+C51F  uniC51F,  HANGUL SYLLABLE-C51F
        '50464' => '14420'            , # G+14420  U+C520  uniC520,  HANGUL SYLLABLE-C520
        '50465' => '14421'            , # G+14421  U+C521  uniC521,  HANGUL SYLLABLE-C521
        '50466' => '14422'            , # G+14422  U+C522  uniC522,  HANGUL SYLLABLE-C522
        '50467' => '14423'            , # G+14423  U+C523  uniC523,  HANGUL SYLLABLE-C523
        '50468' => '14424'            , # G+14424  U+C524  uniC524,  HANGUL SYLLABLE-C524
        '50469' => '14425'            , # G+14425  U+C525  uniC525,  HANGUL SYLLABLE-C525
        '50470' => '14426'            , # G+14426  U+C526  uniC526,  HANGUL SYLLABLE-C526
        '50471' => '14427'            , # G+14427  U+C527  uniC527,  HANGUL SYLLABLE-C527
        '50472' => '2431'             , # G+2431  U+C528  uniC528,  HANGUL SYLLABLE-C528
        '50473' => '2432'             , # G+2432  U+C529  uniC529,  HANGUL SYLLABLE-C529
        '50474' => '14428'            , # G+14428  U+C52A  uniC52A,  HANGUL SYLLABLE-C52A
        '50475' => '14429'            , # G+14429  U+C52B  uniC52B,  HANGUL SYLLABLE-C52B
        '50476' => '2433'             , # G+2433  U+C52C  uniC52C,  HANGUL SYLLABLE-C52C
        '50477' => '14430'            , # G+14430  U+C52D  uniC52D,  HANGUL SYLLABLE-C52D
        '50478' => '14431'            , # G+14431  U+C52E  uniC52E,  HANGUL SYLLABLE-C52E
        '50479' => '14432'            , # G+14432  U+C52F  uniC52F,  HANGUL SYLLABLE-C52F
        '50480' => '2434'             , # G+2434  U+C530  uniC530,  HANGUL SYLLABLE-C530
        '50481' => '14433'            , # G+14433  U+C531  uniC531,  HANGUL SYLLABLE-C531
        '50482' => '14434'            , # G+14434  U+C532  uniC532,  HANGUL SYLLABLE-C532
        '50483' => '14435'            , # G+14435  U+C533  uniC533,  HANGUL SYLLABLE-C533
        '50484' => '14436'            , # G+14436  U+C534  uniC534,  HANGUL SYLLABLE-C534
        '50485' => '14437'            , # G+14437  U+C535  uniC535,  HANGUL SYLLABLE-C535
        '50486' => '14438'            , # G+14438  U+C536  uniC536,  HANGUL SYLLABLE-C536
        '50487' => '14439'            , # G+14439  U+C537  uniC537,  HANGUL SYLLABLE-C537
        '50488' => '2435'             , # G+2435  U+C538  uniC538,  HANGUL SYLLABLE-C538
        '50489' => '2436'             , # G+2436  U+C539  uniC539,  HANGUL SYLLABLE-C539
        '50490' => '14440'            , # G+14440  U+C53A  uniC53A,  HANGUL SYLLABLE-C53A
        '50491' => '2437'             , # G+2437  U+C53B  uniC53B,  HANGUL SYLLABLE-C53B
        '50492' => '14441'            , # G+14441  U+C53C  uniC53C,  HANGUL SYLLABLE-C53C
        '50493' => '2438'             , # G+2438  U+C53D  uniC53D,  HANGUL SYLLABLE-C53D
        '50494' => '14442'            , # G+14442  U+C53E  uniC53E,  HANGUL SYLLABLE-C53E
        '50495' => '14443'            , # G+14443  U+C53F  uniC53F,  HANGUL SYLLABLE-C53F
        '50496' => '14444'            , # G+14444  U+C540  uniC540,  HANGUL SYLLABLE-C540
        '50497' => '14445'            , # G+14445  U+C541  uniC541,  HANGUL SYLLABLE-C541
        '50498' => '14446'            , # G+14446  U+C542  uniC542,  HANGUL SYLLABLE-C542
        '50499' => '14447'            , # G+14447  U+C543  uniC543,  HANGUL SYLLABLE-C543
        '50500' => '2439'             , # G+2439  U+C544  uniC544,  HANGUL SYLLABLE-C544
        '50501' => '2440'             , # G+2440  U+C545  uniC545,  HANGUL SYLLABLE-C545
        '50502' => '14448'            , # G+14448  U+C546  uniC546,  HANGUL SYLLABLE-C546
        '50503' => '14449'            , # G+14449  U+C547  uniC547,  HANGUL SYLLABLE-C547
        '50504' => '2441'             , # G+2441  U+C548  uniC548,  HANGUL SYLLABLE-C548
        '50505' => '2442'             , # G+2442  U+C549  uniC549,  HANGUL SYLLABLE-C549
        '50506' => '2443'             , # G+2443  U+C54A  uniC54A,  HANGUL SYLLABLE-C54A
        '50507' => '14450'            , # G+14450  U+C54B  uniC54B,  HANGUL SYLLABLE-C54B
        '50508' => '2444'             , # G+2444  U+C54C  uniC54C,  HANGUL SYLLABLE-C54C
        '50509' => '2445'             , # G+2445  U+C54D  uniC54D,  HANGUL SYLLABLE-C54D
        '50510' => '2446'             , # G+2446  U+C54E  uniC54E,  HANGUL SYLLABLE-C54E
        '50511' => '14451'            , # G+14451  U+C54F  uniC54F,  HANGUL SYLLABLE-C54F
        '50512' => '14452'            , # G+14452  U+C550  uniC550,  HANGUL SYLLABLE-C550
        '50513' => '14453'            , # G+14453  U+C551  uniC551,  HANGUL SYLLABLE-C551
        '50514' => '14454'            , # G+14454  U+C552  uniC552,  HANGUL SYLLABLE-C552
        '50515' => '2447'             , # G+2447  U+C553  uniC553,  HANGUL SYLLABLE-C553
        '50516' => '2448'             , # G+2448  U+C554  uniC554,  HANGUL SYLLABLE-C554
        '50517' => '2449'             , # G+2449  U+C555  uniC555,  HANGUL SYLLABLE-C555
        '50518' => '14455'            , # G+14455  U+C556  uniC556,  HANGUL SYLLABLE-C556
        '50519' => '2450'             , # G+2450  U+C557  uniC557,  HANGUL SYLLABLE-C557
        '50520' => '2451'             , # G+2451  U+C558  uniC558,  HANGUL SYLLABLE-C558
        '50521' => '2452'             , # G+2452  U+C559  uniC559,  HANGUL SYLLABLE-C559
        '50522' => '14456'            , # G+14456  U+C55A  uniC55A,  HANGUL SYLLABLE-C55A
        '50523' => '14457'            , # G+14457  U+C55B  uniC55B,  HANGUL SYLLABLE-C55B
        '50524' => '14458'            , # G+14458  U+C55C  uniC55C,  HANGUL SYLLABLE-C55C
        '50525' => '2453'             , # G+2453  U+C55D  uniC55D,  HANGUL SYLLABLE-C55D
        '50526' => '2454'             , # G+2454  U+C55E  uniC55E,  HANGUL SYLLABLE-C55E
        '50527' => '14459'            , # G+14459  U+C55F  uniC55F,  HANGUL SYLLABLE-C55F
        '50528' => '2455'             , # G+2455  U+C560  uniC560,  HANGUL SYLLABLE-C560
        '50529' => '2456'             , # G+2456  U+C561  uniC561,  HANGUL SYLLABLE-C561
        '50530' => '14460'            , # G+14460  U+C562  uniC562,  HANGUL SYLLABLE-C562
        '50531' => '14461'            , # G+14461  U+C563  uniC563,  HANGUL SYLLABLE-C563
        '50532' => '2457'             , # G+2457  U+C564  uniC564,  HANGUL SYLLABLE-C564
        '50533' => '14462'            , # G+14462  U+C565  uniC565,  HANGUL SYLLABLE-C565
        '50534' => '14463'            , # G+14463  U+C566  uniC566,  HANGUL SYLLABLE-C566
        '50535' => '14464'            , # G+14464  U+C567  uniC567,  HANGUL SYLLABLE-C567
        '50536' => '2458'             , # G+2458  U+C568  uniC568,  HANGUL SYLLABLE-C568
        '50537' => '14465'            , # G+14465  U+C569  uniC569,  HANGUL SYLLABLE-C569
        '50538' => '14466'            , # G+14466  U+C56A  uniC56A,  HANGUL SYLLABLE-C56A
        '50539' => '14467'            , # G+14467  U+C56B  uniC56B,  HANGUL SYLLABLE-C56B
        '50540' => '14468'            , # G+14468  U+C56C  uniC56C,  HANGUL SYLLABLE-C56C
        '50541' => '14469'            , # G+14469  U+C56D  uniC56D,  HANGUL SYLLABLE-C56D
        '50542' => '14470'            , # G+14470  U+C56E  uniC56E,  HANGUL SYLLABLE-C56E
        '50543' => '14471'            , # G+14471  U+C56F  uniC56F,  HANGUL SYLLABLE-C56F
        '50544' => '2459'             , # G+2459  U+C570  uniC570,  HANGUL SYLLABLE-C570
        '50545' => '2460'             , # G+2460  U+C571  uniC571,  HANGUL SYLLABLE-C571
        '50546' => '14472'            , # G+14472  U+C572  uniC572,  HANGUL SYLLABLE-C572
        '50547' => '2461'             , # G+2461  U+C573  uniC573,  HANGUL SYLLABLE-C573
        '50548' => '2462'             , # G+2462  U+C574  uniC574,  HANGUL SYLLABLE-C574
        '50549' => '2463'             , # G+2463  U+C575  uniC575,  HANGUL SYLLABLE-C575
        '50550' => '14473'            , # G+14473  U+C576  uniC576,  HANGUL SYLLABLE-C576
        '50551' => '14474'            , # G+14474  U+C577  uniC577,  HANGUL SYLLABLE-C577
        '50552' => '14475'            , # G+14475  U+C578  uniC578,  HANGUL SYLLABLE-C578
        '50553' => '14476'            , # G+14476  U+C579  uniC579,  HANGUL SYLLABLE-C579
        '50554' => '14477'            , # G+14477  U+C57A  uniC57A,  HANGUL SYLLABLE-C57A
        '50555' => '14478'            , # G+14478  U+C57B  uniC57B,  HANGUL SYLLABLE-C57B
        '50556' => '2464'             , # G+2464  U+C57C  uniC57C,  HANGUL SYLLABLE-C57C
        '50557' => '2465'             , # G+2465  U+C57D  uniC57D,  HANGUL SYLLABLE-C57D
        '50558' => '14479'            , # G+14479  U+C57E  uniC57E,  HANGUL SYLLABLE-C57E
        '50559' => '14480'            , # G+14480  U+C57F  uniC57F,  HANGUL SYLLABLE-C57F
        '50560' => '2466'             , # G+2466  U+C580  uniC580,  HANGUL SYLLABLE-C580
        '50561' => '14481'            , # G+14481  U+C581  uniC581,  HANGUL SYLLABLE-C581
        '50562' => '14482'            , # G+14482  U+C582  uniC582,  HANGUL SYLLABLE-C582
        '50563' => '14483'            , # G+14483  U+C583  uniC583,  HANGUL SYLLABLE-C583
        '50564' => '2467'             , # G+2467  U+C584  uniC584,  HANGUL SYLLABLE-C584
        '50565' => '14484'            , # G+14484  U+C585  uniC585,  HANGUL SYLLABLE-C585
        '50566' => '14485'            , # G+14485  U+C586  uniC586,  HANGUL SYLLABLE-C586
        '50567' => '2468'             , # G+2468  U+C587  uniC587,  HANGUL SYLLABLE-C587
        '50568' => '14486'            , # G+14486  U+C588  uniC588,  HANGUL SYLLABLE-C588
        '50569' => '14487'            , # G+14487  U+C589  uniC589,  HANGUL SYLLABLE-C589
        '50570' => '14488'            , # G+14488  U+C58A  uniC58A,  HANGUL SYLLABLE-C58A
        '50571' => '14489'            , # G+14489  U+C58B  uniC58B,  HANGUL SYLLABLE-C58B
        '50572' => '2469'             , # G+2469  U+C58C  uniC58C,  HANGUL SYLLABLE-C58C
        '50573' => '2470'             , # G+2470  U+C58D  uniC58D,  HANGUL SYLLABLE-C58D
        '50574' => '14490'            , # G+14490  U+C58E  uniC58E,  HANGUL SYLLABLE-C58E
        '50575' => '2471'             , # G+2471  U+C58F  uniC58F,  HANGUL SYLLABLE-C58F
        '50576' => '14491'            , # G+14491  U+C590  uniC590,  HANGUL SYLLABLE-C590
        '50577' => '2472'             , # G+2472  U+C591  uniC591,  HANGUL SYLLABLE-C591
        '50578' => '14492'            , # G+14492  U+C592  uniC592,  HANGUL SYLLABLE-C592
        '50579' => '14493'            , # G+14493  U+C593  uniC593,  HANGUL SYLLABLE-C593
        '50580' => '14494'            , # G+14494  U+C594  uniC594,  HANGUL SYLLABLE-C594
        '50581' => '2473'             , # G+2473  U+C595  uniC595,  HANGUL SYLLABLE-C595
        '50582' => '14495'            , # G+14495  U+C596  uniC596,  HANGUL SYLLABLE-C596
        '50583' => '2474'             , # G+2474  U+C597  uniC597,  HANGUL SYLLABLE-C597
        '50584' => '2475'             , # G+2475  U+C598  uniC598,  HANGUL SYLLABLE-C598
        '50585' => '14496'            , # G+14496  U+C599  uniC599,  HANGUL SYLLABLE-C599
        '50586' => '14497'            , # G+14497  U+C59A  uniC59A,  HANGUL SYLLABLE-C59A
        '50587' => '14498'            , # G+14498  U+C59B  uniC59B,  HANGUL SYLLABLE-C59B
        '50588' => '2476'             , # G+2476  U+C59C  uniC59C,  HANGUL SYLLABLE-C59C
        '50589' => '14499'            , # G+14499  U+C59D  uniC59D,  HANGUL SYLLABLE-C59D
        '50590' => '14500'            , # G+14500  U+C59E  uniC59E,  HANGUL SYLLABLE-C59E
        '50591' => '14501'            , # G+14501  U+C59F  uniC59F,  HANGUL SYLLABLE-C59F
        '50592' => '2477'             , # G+2477  U+C5A0  uniC5A0,  HANGUL SYLLABLE-C5A0
        '50593' => '14502'            , # G+14502  U+C5A1  uniC5A1,  HANGUL SYLLABLE-C5A1
        '50594' => '14503'            , # G+14503  U+C5A2  uniC5A2,  HANGUL SYLLABLE-C5A2
        '50595' => '14504'            , # G+14504  U+C5A3  uniC5A3,  HANGUL SYLLABLE-C5A3
        '50596' => '14505'            , # G+14505  U+C5A4  uniC5A4,  HANGUL SYLLABLE-C5A4
        '50597' => '14506'            , # G+14506  U+C5A5  uniC5A5,  HANGUL SYLLABLE-C5A5
        '50598' => '14507'            , # G+14507  U+C5A6  uniC5A6,  HANGUL SYLLABLE-C5A6
        '50599' => '14508'            , # G+14508  U+C5A7  uniC5A7,  HANGUL SYLLABLE-C5A7
        '50600' => '14509'            , # G+14509  U+C5A8  uniC5A8,  HANGUL SYLLABLE-C5A8
        '50601' => '2478'             , # G+2478  U+C5A9  uniC5A9,  HANGUL SYLLABLE-C5A9
        '50602' => '14510'            , # G+14510  U+C5AA  uniC5AA,  HANGUL SYLLABLE-C5AA
        '50603' => '14511'            , # G+14511  U+C5AB  uniC5AB,  HANGUL SYLLABLE-C5AB
        '50604' => '14512'            , # G+14512  U+C5AC  uniC5AC,  HANGUL SYLLABLE-C5AC
        '50605' => '14513'            , # G+14513  U+C5AD  uniC5AD,  HANGUL SYLLABLE-C5AD
        '50606' => '14514'            , # G+14514  U+C5AE  uniC5AE,  HANGUL SYLLABLE-C5AE
        '50607' => '14515'            , # G+14515  U+C5AF  uniC5AF,  HANGUL SYLLABLE-C5AF
        '50608' => '14516'            , # G+14516  U+C5B0  uniC5B0,  HANGUL SYLLABLE-C5B0
        '50609' => '14517'            , # G+14517  U+C5B1  uniC5B1,  HANGUL SYLLABLE-C5B1
        '50610' => '14518'            , # G+14518  U+C5B2  uniC5B2,  HANGUL SYLLABLE-C5B2
        '50611' => '14519'            , # G+14519  U+C5B3  uniC5B3,  HANGUL SYLLABLE-C5B3
        '50612' => '2479'             , # G+2479  U+C5B4  uniC5B4,  HANGUL SYLLABLE-C5B4
        '50613' => '2480'             , # G+2480  U+C5B5  uniC5B5,  HANGUL SYLLABLE-C5B5
        '50614' => '14520'            , # G+14520  U+C5B6  uniC5B6,  HANGUL SYLLABLE-C5B6
        '50615' => '14521'            , # G+14521  U+C5B7  uniC5B7,  HANGUL SYLLABLE-C5B7
        '50616' => '2481'             , # G+2481  U+C5B8  uniC5B8,  HANGUL SYLLABLE-C5B8
        '50617' => '2482'             , # G+2482  U+C5B9  uniC5B9,  HANGUL SYLLABLE-C5B9
        '50618' => '14522'            , # G+14522  U+C5BA  uniC5BA,  HANGUL SYLLABLE-C5BA
        '50619' => '2483'             , # G+2483  U+C5BB  uniC5BB,  HANGUL SYLLABLE-C5BB
        '50620' => '2484'             , # G+2484  U+C5BC  uniC5BC,  HANGUL SYLLABLE-C5BC
        '50621' => '2485'             , # G+2485  U+C5BD  uniC5BD,  HANGUL SYLLABLE-C5BD
        '50622' => '2486'             , # G+2486  U+C5BE  uniC5BE,  HANGUL SYLLABLE-C5BE
        '50623' => '14523'            , # G+14523  U+C5BF  uniC5BF,  HANGUL SYLLABLE-C5BF
        '50624' => '14524'            , # G+14524  U+C5C0  uniC5C0,  HANGUL SYLLABLE-C5C0
        '50625' => '14525'            , # G+14525  U+C5C1  uniC5C1,  HANGUL SYLLABLE-C5C1
        '50626' => '14526'            , # G+14526  U+C5C2  uniC5C2,  HANGUL SYLLABLE-C5C2
        '50627' => '14527'            , # G+14527  U+C5C3  uniC5C3,  HANGUL SYLLABLE-C5C3
        '50628' => '2487'             , # G+2487  U+C5C4  uniC5C4,  HANGUL SYLLABLE-C5C4
        '50629' => '2488'             , # G+2488  U+C5C5  uniC5C5,  HANGUL SYLLABLE-C5C5
        '50630' => '2489'             , # G+2489  U+C5C6  uniC5C6,  HANGUL SYLLABLE-C5C6
        '50631' => '2490'             , # G+2490  U+C5C7  uniC5C7,  HANGUL SYLLABLE-C5C7
        '50632' => '2491'             , # G+2491  U+C5C8  uniC5C8,  HANGUL SYLLABLE-C5C8
        '50633' => '2492'             , # G+2492  U+C5C9  uniC5C9,  HANGUL SYLLABLE-C5C9
        '50634' => '2493'             , # G+2493  U+C5CA  uniC5CA,  HANGUL SYLLABLE-C5CA
        '50635' => '14528'            , # G+14528  U+C5CB  uniC5CB,  HANGUL SYLLABLE-C5CB
        '50636' => '2494'             , # G+2494  U+C5CC  uniC5CC,  HANGUL SYLLABLE-C5CC
        '50637' => '14529'            , # G+14529  U+C5CD  uniC5CD,  HANGUL SYLLABLE-C5CD
        '50638' => '2495'             , # G+2495  U+C5CE  uniC5CE,  HANGUL SYLLABLE-C5CE
        '50639' => '14530'            , # G+14530  U+C5CF  uniC5CF,  HANGUL SYLLABLE-C5CF
        '50640' => '2496'             , # G+2496  U+C5D0  uniC5D0,  HANGUL SYLLABLE-C5D0
        '50641' => '2497'             , # G+2497  U+C5D1  uniC5D1,  HANGUL SYLLABLE-C5D1
        '50642' => '14531'            , # G+14531  U+C5D2  uniC5D2,  HANGUL SYLLABLE-C5D2
        '50643' => '14532'            , # G+14532  U+C5D3  uniC5D3,  HANGUL SYLLABLE-C5D3
        '50644' => '2498'             , # G+2498  U+C5D4  uniC5D4,  HANGUL SYLLABLE-C5D4
        '50645' => '14533'            , # G+14533  U+C5D5  uniC5D5,  HANGUL SYLLABLE-C5D5
        '50646' => '14534'            , # G+14534  U+C5D6  uniC5D6,  HANGUL SYLLABLE-C5D6
        '50647' => '14535'            , # G+14535  U+C5D7  uniC5D7,  HANGUL SYLLABLE-C5D7
        '50648' => '2499'             , # G+2499  U+C5D8  uniC5D8,  HANGUL SYLLABLE-C5D8
        '50649' => '14536'            , # G+14536  U+C5D9  uniC5D9,  HANGUL SYLLABLE-C5D9
        '50650' => '14537'            , # G+14537  U+C5DA  uniC5DA,  HANGUL SYLLABLE-C5DA
        '50651' => '14538'            , # G+14538  U+C5DB  uniC5DB,  HANGUL SYLLABLE-C5DB
        '50652' => '14539'            , # G+14539  U+C5DC  uniC5DC,  HANGUL SYLLABLE-C5DC
        '50653' => '14540'            , # G+14540  U+C5DD  uniC5DD,  HANGUL SYLLABLE-C5DD
        '50654' => '14541'            , # G+14541  U+C5DE  uniC5DE,  HANGUL SYLLABLE-C5DE
        '50655' => '14542'            , # G+14542  U+C5DF  uniC5DF,  HANGUL SYLLABLE-C5DF
        '50656' => '2500'             , # G+2500  U+C5E0  uniC5E0,  HANGUL SYLLABLE-C5E0
        '50657' => '2501'             , # G+2501  U+C5E1  uniC5E1,  HANGUL SYLLABLE-C5E1
        '50658' => '14543'            , # G+14543  U+C5E2  uniC5E2,  HANGUL SYLLABLE-C5E2
        '50659' => '2502'             , # G+2502  U+C5E3  uniC5E3,  HANGUL SYLLABLE-C5E3
        '50660' => '14544'            , # G+14544  U+C5E4  uniC5E4,  HANGUL SYLLABLE-C5E4
        '50661' => '2503'             , # G+2503  U+C5E5  uniC5E5,  HANGUL SYLLABLE-C5E5
        '50662' => '14545'            , # G+14545  U+C5E6  uniC5E6,  HANGUL SYLLABLE-C5E6
        '50663' => '14546'            , # G+14546  U+C5E7  uniC5E7,  HANGUL SYLLABLE-C5E7
        '50664' => '14547'            , # G+14547  U+C5E8  uniC5E8,  HANGUL SYLLABLE-C5E8
        '50665' => '14548'            , # G+14548  U+C5E9  uniC5E9,  HANGUL SYLLABLE-C5E9
        '50666' => '14549'            , # G+14549  U+C5EA  uniC5EA,  HANGUL SYLLABLE-C5EA
        '50667' => '14550'            , # G+14550  U+C5EB  uniC5EB,  HANGUL SYLLABLE-C5EB
        '50668' => '2504'             , # G+2504  U+C5EC  uniC5EC,  HANGUL SYLLABLE-C5EC
        '50669' => '2505'             , # G+2505  U+C5ED  uniC5ED,  HANGUL SYLLABLE-C5ED
        '50670' => '2506'             , # G+2506  U+C5EE  uniC5EE,  HANGUL SYLLABLE-C5EE
        '50671' => '14551'            , # G+14551  U+C5EF  uniC5EF,  HANGUL SYLLABLE-C5EF
        '50672' => '2507'             , # G+2507  U+C5F0  uniC5F0,  HANGUL SYLLABLE-C5F0
        '50673' => '14552'            , # G+14552  U+C5F1  uniC5F1,  HANGUL SYLLABLE-C5F1
        '50674' => '14553'            , # G+14553  U+C5F2  uniC5F2,  HANGUL SYLLABLE-C5F2
        '50675' => '14554'            , # G+14554  U+C5F3  uniC5F3,  HANGUL SYLLABLE-C5F3
        '50676' => '2508'             , # G+2508  U+C5F4  uniC5F4,  HANGUL SYLLABLE-C5F4
        '50677' => '14555'            , # G+14555  U+C5F5  uniC5F5,  HANGUL SYLLABLE-C5F5
        '50678' => '2509'             , # G+2509  U+C5F6  uniC5F6,  HANGUL SYLLABLE-C5F6
        '50679' => '2510'             , # G+2510  U+C5F7  uniC5F7,  HANGUL SYLLABLE-C5F7
        '50680' => '14556'            , # G+14556  U+C5F8  uniC5F8,  HANGUL SYLLABLE-C5F8
        '50681' => '14557'            , # G+14557  U+C5F9  uniC5F9,  HANGUL SYLLABLE-C5F9
        '50682' => '14558'            , # G+14558  U+C5FA  uniC5FA,  HANGUL SYLLABLE-C5FA
        '50683' => '14559'            , # G+14559  U+C5FB  uniC5FB,  HANGUL SYLLABLE-C5FB
        '50684' => '2511'             , # G+2511  U+C5FC  uniC5FC,  HANGUL SYLLABLE-C5FC
        '50685' => '2512'             , # G+2512  U+C5FD  uniC5FD,  HANGUL SYLLABLE-C5FD
        '50686' => '2513'             , # G+2513  U+C5FE  uniC5FE,  HANGUL SYLLABLE-C5FE
        '50687' => '2514'             , # G+2514  U+C5FF  uniC5FF,  HANGUL SYLLABLE-C5FF
        '50688' => '2515'             , # G+2515  U+C600  uniC600,  HANGUL SYLLABLE-C600
        '50689' => '2516'             , # G+2516  U+C601  uniC601,  HANGUL SYLLABLE-C601
        '50690' => '14560'            , # G+14560  U+C602  uniC602,  HANGUL SYLLABLE-C602
        '50691' => '14561'            , # G+14561  U+C603  uniC603,  HANGUL SYLLABLE-C603
        '50692' => '14562'            , # G+14562  U+C604  uniC604,  HANGUL SYLLABLE-C604
        '50693' => '2517'             , # G+2517  U+C605  uniC605,  HANGUL SYLLABLE-C605
        '50694' => '2518'             , # G+2518  U+C606  uniC606,  HANGUL SYLLABLE-C606
        '50695' => '2519'             , # G+2519  U+C607  uniC607,  HANGUL SYLLABLE-C607
        '50696' => '2520'             , # G+2520  U+C608  uniC608,  HANGUL SYLLABLE-C608
        '50697' => '14563'            , # G+14563  U+C609  uniC609,  HANGUL SYLLABLE-C609
        '50698' => '14564'            , # G+14564  U+C60A  uniC60A,  HANGUL SYLLABLE-C60A
        '50699' => '14565'            , # G+14565  U+C60B  uniC60B,  HANGUL SYLLABLE-C60B
        '50700' => '2521'             , # G+2521  U+C60C  uniC60C,  HANGUL SYLLABLE-C60C
        '50701' => '14566'            , # G+14566  U+C60D  uniC60D,  HANGUL SYLLABLE-C60D
        '50702' => '14567'            , # G+14567  U+C60E  uniC60E,  HANGUL SYLLABLE-C60E
        '50703' => '14568'            , # G+14568  U+C60F  uniC60F,  HANGUL SYLLABLE-C60F
        '50704' => '2522'             , # G+2522  U+C610  uniC610,  HANGUL SYLLABLE-C610
        '50705' => '14569'            , # G+14569  U+C611  uniC611,  HANGUL SYLLABLE-C611
        '50706' => '14570'            , # G+14570  U+C612  uniC612,  HANGUL SYLLABLE-C612
        '50707' => '14571'            , # G+14571  U+C613  uniC613,  HANGUL SYLLABLE-C613
        '50708' => '14572'            , # G+14572  U+C614  uniC614,  HANGUL SYLLABLE-C614
        '50709' => '14573'            , # G+14573  U+C615  uniC615,  HANGUL SYLLABLE-C615
        '50710' => '14574'            , # G+14574  U+C616  uniC616,  HANGUL SYLLABLE-C616
        '50711' => '14575'            , # G+14575  U+C617  uniC617,  HANGUL SYLLABLE-C617
        '50712' => '2523'             , # G+2523  U+C618  uniC618,  HANGUL SYLLABLE-C618
        '50713' => '2524'             , # G+2524  U+C619  uniC619,  HANGUL SYLLABLE-C619
        '50714' => '14576'            , # G+14576  U+C61A  uniC61A,  HANGUL SYLLABLE-C61A
        '50715' => '2525'             , # G+2525  U+C61B  uniC61B,  HANGUL SYLLABLE-C61B
        '50716' => '2526'             , # G+2526  U+C61C  uniC61C,  HANGUL SYLLABLE-C61C
        '50717' => '14577'            , # G+14577  U+C61D  uniC61D,  HANGUL SYLLABLE-C61D
        '50718' => '14578'            , # G+14578  U+C61E  uniC61E,  HANGUL SYLLABLE-C61E
        '50719' => '14579'            , # G+14579  U+C61F  uniC61F,  HANGUL SYLLABLE-C61F
        '50720' => '14580'            , # G+14580  U+C620  uniC620,  HANGUL SYLLABLE-C620
        '50721' => '14581'            , # G+14581  U+C621  uniC621,  HANGUL SYLLABLE-C621
        '50722' => '14582'            , # G+14582  U+C622  uniC622,  HANGUL SYLLABLE-C622
        '50723' => '14583'            , # G+14583  U+C623  uniC623,  HANGUL SYLLABLE-C623
        '50724' => '2527'             , # G+2527  U+C624  uniC624,  HANGUL SYLLABLE-C624
        '50725' => '2528'             , # G+2528  U+C625  uniC625,  HANGUL SYLLABLE-C625
        '50726' => '14584'            , # G+14584  U+C626  uniC626,  HANGUL SYLLABLE-C626
        '50727' => '14585'            , # G+14585  U+C627  uniC627,  HANGUL SYLLABLE-C627
        '50728' => '2529'             , # G+2529  U+C628  uniC628,  HANGUL SYLLABLE-C628
        '50729' => '14586'            , # G+14586  U+C629  uniC629,  HANGUL SYLLABLE-C629
        '50730' => '14587'            , # G+14587  U+C62A  uniC62A,  HANGUL SYLLABLE-C62A
        '50731' => '14588'            , # G+14588  U+C62B  uniC62B,  HANGUL SYLLABLE-C62B
        '50732' => '2530'             , # G+2530  U+C62C  uniC62C,  HANGUL SYLLABLE-C62C
        '50733' => '2531'             , # G+2531  U+C62D  uniC62D,  HANGUL SYLLABLE-C62D
        '50734' => '2532'             , # G+2532  U+C62E  uniC62E,  HANGUL SYLLABLE-C62E
        '50735' => '14589'            , # G+14589  U+C62F  uniC62F,  HANGUL SYLLABLE-C62F
        '50736' => '2533'             , # G+2533  U+C630  uniC630,  HANGUL SYLLABLE-C630
        '50737' => '14590'            , # G+14590  U+C631  uniC631,  HANGUL SYLLABLE-C631
        '50738' => '14591'            , # G+14591  U+C632  uniC632,  HANGUL SYLLABLE-C632
        '50739' => '2534'             , # G+2534  U+C633  uniC633,  HANGUL SYLLABLE-C633
        '50740' => '2535'             , # G+2535  U+C634  uniC634,  HANGUL SYLLABLE-C634
        '50741' => '2536'             , # G+2536  U+C635  uniC635,  HANGUL SYLLABLE-C635
        '50742' => '14592'            , # G+14592  U+C636  uniC636,  HANGUL SYLLABLE-C636
        '50743' => '2537'             , # G+2537  U+C637  uniC637,  HANGUL SYLLABLE-C637
        '50744' => '14593'            , # G+14593  U+C638  uniC638,  HANGUL SYLLABLE-C638
        '50745' => '2538'             , # G+2538  U+C639  uniC639,  HANGUL SYLLABLE-C639
        '50746' => '14594'            , # G+14594  U+C63A  uniC63A,  HANGUL SYLLABLE-C63A
        '50747' => '2539'             , # G+2539  U+C63B  uniC63B,  HANGUL SYLLABLE-C63B
        '50748' => '14595'            , # G+14595  U+C63C  uniC63C,  HANGUL SYLLABLE-C63C
        '50749' => '14596'            , # G+14596  U+C63D  uniC63D,  HANGUL SYLLABLE-C63D
        '50750' => '14597'            , # G+14597  U+C63E  uniC63E,  HANGUL SYLLABLE-C63E
        '50751' => '14598'            , # G+14598  U+C63F  uniC63F,  HANGUL SYLLABLE-C63F
        '50752' => '2540'             , # G+2540  U+C640  uniC640,  HANGUL SYLLABLE-C640
        '50753' => '2541'             , # G+2541  U+C641  uniC641,  HANGUL SYLLABLE-C641
        '50754' => '14599'            , # G+14599  U+C642  uniC642,  HANGUL SYLLABLE-C642
        '50755' => '14600'            , # G+14600  U+C643  uniC643,  HANGUL SYLLABLE-C643
        '50756' => '2542'             , # G+2542  U+C644  uniC644,  HANGUL SYLLABLE-C644
        '50757' => '14601'            , # G+14601  U+C645  uniC645,  HANGUL SYLLABLE-C645
        '50758' => '14602'            , # G+14602  U+C646  uniC646,  HANGUL SYLLABLE-C646
        '50759' => '14603'            , # G+14603  U+C647  uniC647,  HANGUL SYLLABLE-C647
        '50760' => '2543'             , # G+2543  U+C648  uniC648,  HANGUL SYLLABLE-C648
        '50761' => '14604'            , # G+14604  U+C649  uniC649,  HANGUL SYLLABLE-C649
        '50762' => '14605'            , # G+14605  U+C64A  uniC64A,  HANGUL SYLLABLE-C64A
        '50763' => '14606'            , # G+14606  U+C64B  uniC64B,  HANGUL SYLLABLE-C64B
        '50764' => '14607'            , # G+14607  U+C64C  uniC64C,  HANGUL SYLLABLE-C64C
        '50765' => '14608'            , # G+14608  U+C64D  uniC64D,  HANGUL SYLLABLE-C64D
        '50766' => '14609'            , # G+14609  U+C64E  uniC64E,  HANGUL SYLLABLE-C64E
        '50767' => '14610'            , # G+14610  U+C64F  uniC64F,  HANGUL SYLLABLE-C64F
        '50768' => '2544'             , # G+2544  U+C650  uniC650,  HANGUL SYLLABLE-C650
        '50769' => '2545'             , # G+2545  U+C651  uniC651,  HANGUL SYLLABLE-C651
        '50770' => '14611'            , # G+14611  U+C652  uniC652,  HANGUL SYLLABLE-C652
        '50771' => '2546'             , # G+2546  U+C653  uniC653,  HANGUL SYLLABLE-C653
        '50772' => '2547'             , # G+2547  U+C654  uniC654,  HANGUL SYLLABLE-C654
        '50773' => '2548'             , # G+2548  U+C655  uniC655,  HANGUL SYLLABLE-C655
        '50774' => '14612'            , # G+14612  U+C656  uniC656,  HANGUL SYLLABLE-C656
        '50775' => '14613'            , # G+14613  U+C657  uniC657,  HANGUL SYLLABLE-C657
        '50776' => '14614'            , # G+14614  U+C658  uniC658,  HANGUL SYLLABLE-C658
        '50777' => '14615'            , # G+14615  U+C659  uniC659,  HANGUL SYLLABLE-C659
        '50778' => '14616'            , # G+14616  U+C65A  uniC65A,  HANGUL SYLLABLE-C65A
        '50779' => '14617'            , # G+14617  U+C65B  uniC65B,  HANGUL SYLLABLE-C65B
        '50780' => '2549'             , # G+2549  U+C65C  uniC65C,  HANGUL SYLLABLE-C65C
        '50781' => '2550'             , # G+2550  U+C65D  uniC65D,  HANGUL SYLLABLE-C65D
        '50782' => '14618'            , # G+14618  U+C65E  uniC65E,  HANGUL SYLLABLE-C65E
        '50783' => '14619'            , # G+14619  U+C65F  uniC65F,  HANGUL SYLLABLE-C65F
        '50784' => '2551'             , # G+2551  U+C660  uniC660,  HANGUL SYLLABLE-C660
        '50785' => '14620'            , # G+14620  U+C661  uniC661,  HANGUL SYLLABLE-C661
        '50786' => '14621'            , # G+14621  U+C662  uniC662,  HANGUL SYLLABLE-C662
        '50787' => '14622'            , # G+14622  U+C663  uniC663,  HANGUL SYLLABLE-C663
        '50788' => '14623'            , # G+14623  U+C664  uniC664,  HANGUL SYLLABLE-C664
        '50789' => '14624'            , # G+14624  U+C665  uniC665,  HANGUL SYLLABLE-C665
        '50790' => '14625'            , # G+14625  U+C666  uniC666,  HANGUL SYLLABLE-C666
        '50791' => '14626'            , # G+14626  U+C667  uniC667,  HANGUL SYLLABLE-C667
        '50792' => '14627'            , # G+14627  U+C668  uniC668,  HANGUL SYLLABLE-C668
        '50793' => '14628'            , # G+14628  U+C669  uniC669,  HANGUL SYLLABLE-C669
        '50794' => '14629'            , # G+14629  U+C66A  uniC66A,  HANGUL SYLLABLE-C66A
        '50795' => '14630'            , # G+14630  U+C66B  uniC66B,  HANGUL SYLLABLE-C66B
        '50796' => '2552'             , # G+2552  U+C66C  uniC66C,  HANGUL SYLLABLE-C66C
        '50797' => '14631'            , # G+14631  U+C66D  uniC66D,  HANGUL SYLLABLE-C66D
        '50798' => '14632'            , # G+14632  U+C66E  uniC66E,  HANGUL SYLLABLE-C66E
        '50799' => '2553'             , # G+2553  U+C66F  uniC66F,  HANGUL SYLLABLE-C66F
        '50800' => '14633'            , # G+14633  U+C670  uniC670,  HANGUL SYLLABLE-C670
        '50801' => '2554'             , # G+2554  U+C671  uniC671,  HANGUL SYLLABLE-C671
        '50802' => '14634'            , # G+14634  U+C672  uniC672,  HANGUL SYLLABLE-C672
        '50803' => '14635'            , # G+14635  U+C673  uniC673,  HANGUL SYLLABLE-C673
        '50804' => '14636'            , # G+14636  U+C674  uniC674,  HANGUL SYLLABLE-C674
        '50805' => '14637'            , # G+14637  U+C675  uniC675,  HANGUL SYLLABLE-C675
        '50806' => '14638'            , # G+14638  U+C676  uniC676,  HANGUL SYLLABLE-C676
        '50807' => '14639'            , # G+14639  U+C677  uniC677,  HANGUL SYLLABLE-C677
        '50808' => '2555'             , # G+2555  U+C678  uniC678,  HANGUL SYLLABLE-C678
        '50809' => '2556'             , # G+2556  U+C679  uniC679,  HANGUL SYLLABLE-C679
        '50810' => '14640'            , # G+14640  U+C67A  uniC67A,  HANGUL SYLLABLE-C67A
        '50811' => '14641'            , # G+14641  U+C67B  uniC67B,  HANGUL SYLLABLE-C67B
        '50812' => '2557'             , # G+2557  U+C67C  uniC67C,  HANGUL SYLLABLE-C67C
        '50813' => '14642'            , # G+14642  U+C67D  uniC67D,  HANGUL SYLLABLE-C67D
        '50814' => '14643'            , # G+14643  U+C67E  uniC67E,  HANGUL SYLLABLE-C67E
        '50815' => '14644'            , # G+14644  U+C67F  uniC67F,  HANGUL SYLLABLE-C67F
        '50816' => '2558'             , # G+2558  U+C680  uniC680,  HANGUL SYLLABLE-C680
        '50817' => '14645'            , # G+14645  U+C681  uniC681,  HANGUL SYLLABLE-C681
        '50818' => '14646'            , # G+14646  U+C682  uniC682,  HANGUL SYLLABLE-C682
        '50819' => '14647'            , # G+14647  U+C683  uniC683,  HANGUL SYLLABLE-C683
        '50820' => '14648'            , # G+14648  U+C684  uniC684,  HANGUL SYLLABLE-C684
        '50821' => '14649'            , # G+14649  U+C685  uniC685,  HANGUL SYLLABLE-C685
        '50822' => '14650'            , # G+14650  U+C686  uniC686,  HANGUL SYLLABLE-C686
        '50823' => '14651'            , # G+14651  U+C687  uniC687,  HANGUL SYLLABLE-C687
        '50824' => '2559'             , # G+2559  U+C688  uniC688,  HANGUL SYLLABLE-C688
        '50825' => '2560'             , # G+2560  U+C689  uniC689,  HANGUL SYLLABLE-C689
        '50826' => '14652'            , # G+14652  U+C68A  uniC68A,  HANGUL SYLLABLE-C68A
        '50827' => '2561'             , # G+2561  U+C68B  uniC68B,  HANGUL SYLLABLE-C68B
        '50828' => '14653'            , # G+14653  U+C68C  uniC68C,  HANGUL SYLLABLE-C68C
        '50829' => '2562'             , # G+2562  U+C68D  uniC68D,  HANGUL SYLLABLE-C68D
        '50830' => '14654'            , # G+14654  U+C68E  uniC68E,  HANGUL SYLLABLE-C68E
        '50831' => '14655'            , # G+14655  U+C68F  uniC68F,  HANGUL SYLLABLE-C68F
        '50832' => '14656'            , # G+14656  U+C690  uniC690,  HANGUL SYLLABLE-C690
        '50833' => '14657'            , # G+14657  U+C691  uniC691,  HANGUL SYLLABLE-C691
        '50834' => '14658'            , # G+14658  U+C692  uniC692,  HANGUL SYLLABLE-C692
        '50835' => '14659'            , # G+14659  U+C693  uniC693,  HANGUL SYLLABLE-C693
        '50836' => '2563'             , # G+2563  U+C694  uniC694,  HANGUL SYLLABLE-C694
        '50837' => '2564'             , # G+2564  U+C695  uniC695,  HANGUL SYLLABLE-C695
        '50838' => '14660'            , # G+14660  U+C696  uniC696,  HANGUL SYLLABLE-C696
        '50839' => '14661'            , # G+14661  U+C697  uniC697,  HANGUL SYLLABLE-C697
        '50840' => '2565'             , # G+2565  U+C698  uniC698,  HANGUL SYLLABLE-C698
        '50841' => '14662'            , # G+14662  U+C699  uniC699,  HANGUL SYLLABLE-C699
        '50842' => '14663'            , # G+14663  U+C69A  uniC69A,  HANGUL SYLLABLE-C69A
        '50843' => '14664'            , # G+14664  U+C69B  uniC69B,  HANGUL SYLLABLE-C69B
        '50844' => '2566'             , # G+2566  U+C69C  uniC69C,  HANGUL SYLLABLE-C69C
        '50845' => '14665'            , # G+14665  U+C69D  uniC69D,  HANGUL SYLLABLE-C69D
        '50846' => '14666'            , # G+14666  U+C69E  uniC69E,  HANGUL SYLLABLE-C69E
        '50847' => '14667'            , # G+14667  U+C69F  uniC69F,  HANGUL SYLLABLE-C69F
        '50848' => '14668'            , # G+14668  U+C6A0  uniC6A0,  HANGUL SYLLABLE-C6A0
        '50849' => '14669'            , # G+14669  U+C6A1  uniC6A1,  HANGUL SYLLABLE-C6A1
        '50850' => '14670'            , # G+14670  U+C6A2  uniC6A2,  HANGUL SYLLABLE-C6A2
        '50851' => '14671'            , # G+14671  U+C6A3  uniC6A3,  HANGUL SYLLABLE-C6A3
        '50852' => '2567'             , # G+2567  U+C6A4  uniC6A4,  HANGUL SYLLABLE-C6A4
        '50853' => '2568'             , # G+2568  U+C6A5  uniC6A5,  HANGUL SYLLABLE-C6A5
        '50854' => '14672'            , # G+14672  U+C6A6  uniC6A6,  HANGUL SYLLABLE-C6A6
        '50855' => '2569'             , # G+2569  U+C6A7  uniC6A7,  HANGUL SYLLABLE-C6A7
        '50856' => '14673'            , # G+14673  U+C6A8  uniC6A8,  HANGUL SYLLABLE-C6A8
        '50857' => '2570'             , # G+2570  U+C6A9  uniC6A9,  HANGUL SYLLABLE-C6A9
        '50858' => '14674'            , # G+14674  U+C6AA  uniC6AA,  HANGUL SYLLABLE-C6AA
        '50859' => '14675'            , # G+14675  U+C6AB  uniC6AB,  HANGUL SYLLABLE-C6AB
        '50860' => '14676'            , # G+14676  U+C6AC  uniC6AC,  HANGUL SYLLABLE-C6AC
        '50861' => '14677'            , # G+14677  U+C6AD  uniC6AD,  HANGUL SYLLABLE-C6AD
        '50862' => '14678'            , # G+14678  U+C6AE  uniC6AE,  HANGUL SYLLABLE-C6AE
        '50863' => '14679'            , # G+14679  U+C6AF  uniC6AF,  HANGUL SYLLABLE-C6AF
        '50864' => '2571'             , # G+2571  U+C6B0  uniC6B0,  HANGUL SYLLABLE-C6B0
        '50865' => '2572'             , # G+2572  U+C6B1  uniC6B1,  HANGUL SYLLABLE-C6B1
        '50866' => '14680'            , # G+14680  U+C6B2  uniC6B2,  HANGUL SYLLABLE-C6B2
        '50867' => '14681'            , # G+14681  U+C6B3  uniC6B3,  HANGUL SYLLABLE-C6B3
        '50868' => '2573'             , # G+2573  U+C6B4  uniC6B4,  HANGUL SYLLABLE-C6B4
        '50869' => '14682'            , # G+14682  U+C6B5  uniC6B5,  HANGUL SYLLABLE-C6B5
        '50870' => '14683'            , # G+14683  U+C6B6  uniC6B6,  HANGUL SYLLABLE-C6B6
        '50871' => '14684'            , # G+14684  U+C6B7  uniC6B7,  HANGUL SYLLABLE-C6B7
        '50872' => '2574'             , # G+2574  U+C6B8  uniC6B8,  HANGUL SYLLABLE-C6B8
        '50873' => '2575'             , # G+2575  U+C6B9  uniC6B9,  HANGUL SYLLABLE-C6B9
        '50874' => '2576'             , # G+2576  U+C6BA  uniC6BA,  HANGUL SYLLABLE-C6BA
        '50875' => '14685'            , # G+14685  U+C6BB  uniC6BB,  HANGUL SYLLABLE-C6BB
        '50876' => '14686'            , # G+14686  U+C6BC  uniC6BC,  HANGUL SYLLABLE-C6BC
        '50877' => '14687'            , # G+14687  U+C6BD  uniC6BD,  HANGUL SYLLABLE-C6BD
        '50878' => '14688'            , # G+14688  U+C6BE  uniC6BE,  HANGUL SYLLABLE-C6BE
        '50879' => '14689'            , # G+14689  U+C6BF  uniC6BF,  HANGUL SYLLABLE-C6BF
        '50880' => '2577'             , # G+2577  U+C6C0  uniC6C0,  HANGUL SYLLABLE-C6C0
        '50881' => '2578'             , # G+2578  U+C6C1  uniC6C1,  HANGUL SYLLABLE-C6C1
        '50882' => '14690'            , # G+14690  U+C6C2  uniC6C2,  HANGUL SYLLABLE-C6C2
        '50883' => '2579'             , # G+2579  U+C6C3  uniC6C3,  HANGUL SYLLABLE-C6C3
        '50884' => '14691'            , # G+14691  U+C6C4  uniC6C4,  HANGUL SYLLABLE-C6C4
        '50885' => '2580'             , # G+2580  U+C6C5  uniC6C5,  HANGUL SYLLABLE-C6C5
        '50886' => '14692'            , # G+14692  U+C6C6  uniC6C6,  HANGUL SYLLABLE-C6C6
        '50887' => '14693'            , # G+14693  U+C6C7  uniC6C7,  HANGUL SYLLABLE-C6C7
        '50888' => '14694'            , # G+14694  U+C6C8  uniC6C8,  HANGUL SYLLABLE-C6C8
        '50889' => '14695'            , # G+14695  U+C6C9  uniC6C9,  HANGUL SYLLABLE-C6C9
        '50890' => '14696'            , # G+14696  U+C6CA  uniC6CA,  HANGUL SYLLABLE-C6CA
        '50891' => '14697'            , # G+14697  U+C6CB  uniC6CB,  HANGUL SYLLABLE-C6CB
        '50892' => '2581'             , # G+2581  U+C6CC  uniC6CC,  HANGUL SYLLABLE-C6CC
        '50893' => '2582'             , # G+2582  U+C6CD  uniC6CD,  HANGUL SYLLABLE-C6CD
        '50894' => '14698'            , # G+14698  U+C6CE  uniC6CE,  HANGUL SYLLABLE-C6CE
        '50895' => '14699'            , # G+14699  U+C6CF  uniC6CF,  HANGUL SYLLABLE-C6CF
        '50896' => '2583'             , # G+2583  U+C6D0  uniC6D0,  HANGUL SYLLABLE-C6D0
        '50897' => '14700'            , # G+14700  U+C6D1  uniC6D1,  HANGUL SYLLABLE-C6D1
        '50898' => '14701'            , # G+14701  U+C6D2  uniC6D2,  HANGUL SYLLABLE-C6D2
        '50899' => '14702'            , # G+14702  U+C6D3  uniC6D3,  HANGUL SYLLABLE-C6D3
        '50900' => '2584'             , # G+2584  U+C6D4  uniC6D4,  HANGUL SYLLABLE-C6D4
        '50901' => '14703'            , # G+14703  U+C6D5  uniC6D5,  HANGUL SYLLABLE-C6D5
        '50902' => '14704'            , # G+14704  U+C6D6  uniC6D6,  HANGUL SYLLABLE-C6D6
        '50903' => '14705'            , # G+14705  U+C6D7  uniC6D7,  HANGUL SYLLABLE-C6D7
        '50904' => '14706'            , # G+14706  U+C6D8  uniC6D8,  HANGUL SYLLABLE-C6D8
        '50905' => '14707'            , # G+14707  U+C6D9  uniC6D9,  HANGUL SYLLABLE-C6D9
        '50906' => '14708'            , # G+14708  U+C6DA  uniC6DA,  HANGUL SYLLABLE-C6DA
        '50907' => '14709'            , # G+14709  U+C6DB  uniC6DB,  HANGUL SYLLABLE-C6DB
        '50908' => '2585'             , # G+2585  U+C6DC  uniC6DC,  HANGUL SYLLABLE-C6DC
        '50909' => '2586'             , # G+2586  U+C6DD  uniC6DD,  HANGUL SYLLABLE-C6DD
        '50910' => '14710'            , # G+14710  U+C6DE  uniC6DE,  HANGUL SYLLABLE-C6DE
        '50911' => '14711'            , # G+14711  U+C6DF  uniC6DF,  HANGUL SYLLABLE-C6DF
        '50912' => '2587'             , # G+2587  U+C6E0  uniC6E0,  HANGUL SYLLABLE-C6E0
        '50913' => '2588'             , # G+2588  U+C6E1  uniC6E1,  HANGUL SYLLABLE-C6E1
        '50914' => '14712'            , # G+14712  U+C6E2  uniC6E2,  HANGUL SYLLABLE-C6E2
        '50915' => '14713'            , # G+14713  U+C6E3  uniC6E3,  HANGUL SYLLABLE-C6E3
        '50916' => '14714'            , # G+14714  U+C6E4  uniC6E4,  HANGUL SYLLABLE-C6E4
        '50917' => '14715'            , # G+14715  U+C6E5  uniC6E5,  HANGUL SYLLABLE-C6E5
        '50918' => '14716'            , # G+14716  U+C6E6  uniC6E6,  HANGUL SYLLABLE-C6E6
        '50919' => '14717'            , # G+14717  U+C6E7  uniC6E7,  HANGUL SYLLABLE-C6E7
        '50920' => '2589'             , # G+2589  U+C6E8  uniC6E8,  HANGUL SYLLABLE-C6E8
        '50921' => '2590'             , # G+2590  U+C6E9  uniC6E9,  HANGUL SYLLABLE-C6E9
        '50922' => '14718'            , # G+14718  U+C6EA  uniC6EA,  HANGUL SYLLABLE-C6EA
        '50923' => '14719'            , # G+14719  U+C6EB  uniC6EB,  HANGUL SYLLABLE-C6EB
        '50924' => '2591'             , # G+2591  U+C6EC  uniC6EC,  HANGUL SYLLABLE-C6EC
        '50925' => '14720'            , # G+14720  U+C6ED  uniC6ED,  HANGUL SYLLABLE-C6ED
        '50926' => '14721'            , # G+14721  U+C6EE  uniC6EE,  HANGUL SYLLABLE-C6EE
        '50927' => '14722'            , # G+14722  U+C6EF  uniC6EF,  HANGUL SYLLABLE-C6EF
        '50928' => '2592'             , # G+2592  U+C6F0  uniC6F0,  HANGUL SYLLABLE-C6F0
        '50929' => '14723'            , # G+14723  U+C6F1  uniC6F1,  HANGUL SYLLABLE-C6F1
        '50930' => '14724'            , # G+14724  U+C6F2  uniC6F2,  HANGUL SYLLABLE-C6F2
        '50931' => '14725'            , # G+14725  U+C6F3  uniC6F3,  HANGUL SYLLABLE-C6F3
        '50932' => '14726'            , # G+14726  U+C6F4  uniC6F4,  HANGUL SYLLABLE-C6F4
        '50933' => '14727'            , # G+14727  U+C6F5  uniC6F5,  HANGUL SYLLABLE-C6F5
        '50934' => '14728'            , # G+14728  U+C6F6  uniC6F6,  HANGUL SYLLABLE-C6F6
        '50935' => '14729'            , # G+14729  U+C6F7  uniC6F7,  HANGUL SYLLABLE-C6F7
        '50936' => '2593'             , # G+2593  U+C6F8  uniC6F8,  HANGUL SYLLABLE-C6F8
        '50937' => '2594'             , # G+2594  U+C6F9  uniC6F9,  HANGUL SYLLABLE-C6F9
        '50938' => '14730'            , # G+14730  U+C6FA  uniC6FA,  HANGUL SYLLABLE-C6FA
        '50939' => '14731'            , # G+14731  U+C6FB  uniC6FB,  HANGUL SYLLABLE-C6FB
        '50940' => '14732'            , # G+14732  U+C6FC  uniC6FC,  HANGUL SYLLABLE-C6FC
        '50941' => '2595'             , # G+2595  U+C6FD  uniC6FD,  HANGUL SYLLABLE-C6FD
        '50942' => '14733'            , # G+14733  U+C6FE  uniC6FE,  HANGUL SYLLABLE-C6FE
        '50943' => '14734'            , # G+14734  U+C6FF  uniC6FF,  HANGUL SYLLABLE-C6FF
        '50944' => '14735'            , # G+14735  U+C700  uniC700,  HANGUL SYLLABLE-C700
        '50945' => '14736'            , # G+14736  U+C701  uniC701,  HANGUL SYLLABLE-C701
        '50946' => '14737'            , # G+14737  U+C702  uniC702,  HANGUL SYLLABLE-C702
        '50947' => '14738'            , # G+14738  U+C703  uniC703,  HANGUL SYLLABLE-C703
        '50948' => '2596'             , # G+2596  U+C704  uniC704,  HANGUL SYLLABLE-C704
        '50949' => '2597'             , # G+2597  U+C705  uniC705,  HANGUL SYLLABLE-C705
        '50950' => '14739'            , # G+14739  U+C706  uniC706,  HANGUL SYLLABLE-C706
        '50951' => '14740'            , # G+14740  U+C707  uniC707,  HANGUL SYLLABLE-C707
        '50952' => '2598'             , # G+2598  U+C708  uniC708,  HANGUL SYLLABLE-C708
        '50953' => '14741'            , # G+14741  U+C709  uniC709,  HANGUL SYLLABLE-C709
        '50954' => '14742'            , # G+14742  U+C70A  uniC70A,  HANGUL SYLLABLE-C70A
        '50955' => '14743'            , # G+14743  U+C70B  uniC70B,  HANGUL SYLLABLE-C70B
        '50956' => '2599'             , # G+2599  U+C70C  uniC70C,  HANGUL SYLLABLE-C70C
        '50957' => '14744'            , # G+14744  U+C70D  uniC70D,  HANGUL SYLLABLE-C70D
        '50958' => '14745'            , # G+14745  U+C70E  uniC70E,  HANGUL SYLLABLE-C70E
        '50959' => '14746'            , # G+14746  U+C70F  uniC70F,  HANGUL SYLLABLE-C70F
        '50960' => '14747'            , # G+14747  U+C710  uniC710,  HANGUL SYLLABLE-C710
        '50961' => '14748'            , # G+14748  U+C711  uniC711,  HANGUL SYLLABLE-C711
        '50962' => '14749'            , # G+14749  U+C712  uniC712,  HANGUL SYLLABLE-C712
        '50963' => '14750'            , # G+14750  U+C713  uniC713,  HANGUL SYLLABLE-C713
        '50964' => '2600'             , # G+2600  U+C714  uniC714,  HANGUL SYLLABLE-C714
        '50965' => '2601'             , # G+2601  U+C715  uniC715,  HANGUL SYLLABLE-C715
        '50966' => '14751'            , # G+14751  U+C716  uniC716,  HANGUL SYLLABLE-C716
        '50967' => '2602'             , # G+2602  U+C717  uniC717,  HANGUL SYLLABLE-C717
        '50968' => '14752'            , # G+14752  U+C718  uniC718,  HANGUL SYLLABLE-C718
        '50969' => '2603'             , # G+2603  U+C719  uniC719,  HANGUL SYLLABLE-C719
        '50970' => '14753'            , # G+14753  U+C71A  uniC71A,  HANGUL SYLLABLE-C71A
        '50971' => '14754'            , # G+14754  U+C71B  uniC71B,  HANGUL SYLLABLE-C71B
        '50972' => '14755'            , # G+14755  U+C71C  uniC71C,  HANGUL SYLLABLE-C71C
        '50973' => '14756'            , # G+14756  U+C71D  uniC71D,  HANGUL SYLLABLE-C71D
        '50974' => '14757'            , # G+14757  U+C71E  uniC71E,  HANGUL SYLLABLE-C71E
        '50975' => '14758'            , # G+14758  U+C71F  uniC71F,  HANGUL SYLLABLE-C71F
        '50976' => '2604'             , # G+2604  U+C720  uniC720,  HANGUL SYLLABLE-C720
        '50977' => '2605'             , # G+2605  U+C721  uniC721,  HANGUL SYLLABLE-C721
        '50978' => '14759'            , # G+14759  U+C722  uniC722,  HANGUL SYLLABLE-C722
        '50979' => '14760'            , # G+14760  U+C723  uniC723,  HANGUL SYLLABLE-C723
        '50980' => '2606'             , # G+2606  U+C724  uniC724,  HANGUL SYLLABLE-C724
        '50981' => '14761'            , # G+14761  U+C725  uniC725,  HANGUL SYLLABLE-C725
        '50982' => '14762'            , # G+14762  U+C726  uniC726,  HANGUL SYLLABLE-C726
        '50983' => '14763'            , # G+14763  U+C727  uniC727,  HANGUL SYLLABLE-C727
        '50984' => '2607'             , # G+2607  U+C728  uniC728,  HANGUL SYLLABLE-C728
        '50985' => '14764'            , # G+14764  U+C729  uniC729,  HANGUL SYLLABLE-C729
        '50986' => '14765'            , # G+14765  U+C72A  uniC72A,  HANGUL SYLLABLE-C72A
        '50987' => '14766'            , # G+14766  U+C72B  uniC72B,  HANGUL SYLLABLE-C72B
        '50988' => '14767'            , # G+14767  U+C72C  uniC72C,  HANGUL SYLLABLE-C72C
        '50989' => '14768'            , # G+14768  U+C72D  uniC72D,  HANGUL SYLLABLE-C72D
        '50990' => '14769'            , # G+14769  U+C72E  uniC72E,  HANGUL SYLLABLE-C72E
        '50991' => '14770'            , # G+14770  U+C72F  uniC72F,  HANGUL SYLLABLE-C72F
        '50992' => '2608'             , # G+2608  U+C730  uniC730,  HANGUL SYLLABLE-C730
        '50993' => '2609'             , # G+2609  U+C731  uniC731,  HANGUL SYLLABLE-C731
        '50994' => '14771'            , # G+14771  U+C732  uniC732,  HANGUL SYLLABLE-C732
        '50995' => '2610'             , # G+2610  U+C733  uniC733,  HANGUL SYLLABLE-C733
        '50996' => '14772'            , # G+14772  U+C734  uniC734,  HANGUL SYLLABLE-C734
        '50997' => '2611'             , # G+2611  U+C735  uniC735,  HANGUL SYLLABLE-C735
        '50998' => '14773'            , # G+14773  U+C736  uniC736,  HANGUL SYLLABLE-C736
        '50999' => '2612'             , # G+2612  U+C737  uniC737,  HANGUL SYLLABLE-C737
        '51000' => '14774'            , # G+14774  U+C738  uniC738,  HANGUL SYLLABLE-C738
        '51001' => '14775'            , # G+14775  U+C739  uniC739,  HANGUL SYLLABLE-C739
        '51002' => '14776'            , # G+14776  U+C73A  uniC73A,  HANGUL SYLLABLE-C73A
        '51003' => '14777'            , # G+14777  U+C73B  uniC73B,  HANGUL SYLLABLE-C73B
        '51004' => '2613'             , # G+2613  U+C73C  uniC73C,  HANGUL SYLLABLE-C73C
        '51005' => '2614'             , # G+2614  U+C73D  uniC73D,  HANGUL SYLLABLE-C73D
        '51006' => '14778'            , # G+14778  U+C73E  uniC73E,  HANGUL SYLLABLE-C73E
        '51007' => '14779'            , # G+14779  U+C73F  uniC73F,  HANGUL SYLLABLE-C73F
        '51008' => '2615'             , # G+2615  U+C740  uniC740,  HANGUL SYLLABLE-C740
        '51009' => '14780'            , # G+14780  U+C741  uniC741,  HANGUL SYLLABLE-C741
        '51010' => '14781'            , # G+14781  U+C742  uniC742,  HANGUL SYLLABLE-C742
        '51011' => '14782'            , # G+14782  U+C743  uniC743,  HANGUL SYLLABLE-C743
        '51012' => '2616'             , # G+2616  U+C744  uniC744,  HANGUL SYLLABLE-C744
        '51013' => '14783'            , # G+14783  U+C745  uniC745,  HANGUL SYLLABLE-C745
        '51014' => '14784'            , # G+14784  U+C746  uniC746,  HANGUL SYLLABLE-C746
        '51015' => '14785'            , # G+14785  U+C747  uniC747,  HANGUL SYLLABLE-C747
        '51016' => '14786'            , # G+14786  U+C748  uniC748,  HANGUL SYLLABLE-C748
        '51017' => '14787'            , # G+14787  U+C749  uniC749,  HANGUL SYLLABLE-C749
        '51018' => '2617'             , # G+2617  U+C74A  uniC74A,  HANGUL SYLLABLE-C74A
        '51019' => '14788'            , # G+14788  U+C74B  uniC74B,  HANGUL SYLLABLE-C74B
        '51020' => '2618'             , # G+2618  U+C74C  uniC74C,  HANGUL SYLLABLE-C74C
        '51021' => '2619'             , # G+2619  U+C74D  uniC74D,  HANGUL SYLLABLE-C74D
        '51022' => '14789'            , # G+14789  U+C74E  uniC74E,  HANGUL SYLLABLE-C74E
        '51023' => '2620'             , # G+2620  U+C74F  uniC74F,  HANGUL SYLLABLE-C74F
        '51024' => '14790'            , # G+14790  U+C750  uniC750,  HANGUL SYLLABLE-C750
        '51025' => '2621'             , # G+2621  U+C751  uniC751,  HANGUL SYLLABLE-C751
        '51026' => '2622'             , # G+2622  U+C752  uniC752,  HANGUL SYLLABLE-C752
        '51027' => '2623'             , # G+2623  U+C753  uniC753,  HANGUL SYLLABLE-C753
        '51028' => '2624'             , # G+2624  U+C754  uniC754,  HANGUL SYLLABLE-C754
        '51029' => '2625'             , # G+2625  U+C755  uniC755,  HANGUL SYLLABLE-C755
        '51030' => '2626'             , # G+2626  U+C756  uniC756,  HANGUL SYLLABLE-C756
        '51031' => '2627'             , # G+2627  U+C757  uniC757,  HANGUL SYLLABLE-C757
        '51032' => '2628'             , # G+2628  U+C758  uniC758,  HANGUL SYLLABLE-C758
        '51033' => '14791'            , # G+14791  U+C759  uniC759,  HANGUL SYLLABLE-C759
        '51034' => '14792'            , # G+14792  U+C75A  uniC75A,  HANGUL SYLLABLE-C75A
        '51035' => '14793'            , # G+14793  U+C75B  uniC75B,  HANGUL SYLLABLE-C75B
        '51036' => '2629'             , # G+2629  U+C75C  uniC75C,  HANGUL SYLLABLE-C75C
        '51037' => '14794'            , # G+14794  U+C75D  uniC75D,  HANGUL SYLLABLE-C75D
        '51038' => '14795'            , # G+14795  U+C75E  uniC75E,  HANGUL SYLLABLE-C75E
        '51039' => '14796'            , # G+14796  U+C75F  uniC75F,  HANGUL SYLLABLE-C75F
        '51040' => '2630'             , # G+2630  U+C760  uniC760,  HANGUL SYLLABLE-C760
        '51041' => '14797'            , # G+14797  U+C761  uniC761,  HANGUL SYLLABLE-C761
        '51042' => '14798'            , # G+14798  U+C762  uniC762,  HANGUL SYLLABLE-C762
        '51043' => '14799'            , # G+14799  U+C763  uniC763,  HANGUL SYLLABLE-C763
        '51044' => '14800'            , # G+14800  U+C764  uniC764,  HANGUL SYLLABLE-C764
        '51045' => '14801'            , # G+14801  U+C765  uniC765,  HANGUL SYLLABLE-C765
        '51046' => '14802'            , # G+14802  U+C766  uniC766,  HANGUL SYLLABLE-C766
        '51047' => '14803'            , # G+14803  U+C767  uniC767,  HANGUL SYLLABLE-C767
        '51048' => '2631'             , # G+2631  U+C768  uniC768,  HANGUL SYLLABLE-C768
        '51049' => '14804'            , # G+14804  U+C769  uniC769,  HANGUL SYLLABLE-C769
        '51050' => '14805'            , # G+14805  U+C76A  uniC76A,  HANGUL SYLLABLE-C76A
        '51051' => '2632'             , # G+2632  U+C76B  uniC76B,  HANGUL SYLLABLE-C76B
        '51052' => '14806'            , # G+14806  U+C76C  uniC76C,  HANGUL SYLLABLE-C76C
        '51053' => '14807'            , # G+14807  U+C76D  uniC76D,  HANGUL SYLLABLE-C76D
        '51054' => '14808'            , # G+14808  U+C76E  uniC76E,  HANGUL SYLLABLE-C76E
        '51055' => '14809'            , # G+14809  U+C76F  uniC76F,  HANGUL SYLLABLE-C76F
        '51056' => '14810'            , # G+14810  U+C770  uniC770,  HANGUL SYLLABLE-C770
        '51057' => '14811'            , # G+14811  U+C771  uniC771,  HANGUL SYLLABLE-C771
        '51058' => '14812'            , # G+14812  U+C772  uniC772,  HANGUL SYLLABLE-C772
        '51059' => '14813'            , # G+14813  U+C773  uniC773,  HANGUL SYLLABLE-C773
        '51060' => '2633'             , # G+2633  U+C774  uniC774,  HANGUL SYLLABLE-C774
        '51061' => '2634'             , # G+2634  U+C775  uniC775,  HANGUL SYLLABLE-C775
        '51062' => '14814'            , # G+14814  U+C776  uniC776,  HANGUL SYLLABLE-C776
        '51063' => '14815'            , # G+14815  U+C777  uniC777,  HANGUL SYLLABLE-C777
        '51064' => '2635'             , # G+2635  U+C778  uniC778,  HANGUL SYLLABLE-C778
        '51065' => '14816'            , # G+14816  U+C779  uniC779,  HANGUL SYLLABLE-C779
        '51066' => '14817'            , # G+14817  U+C77A  uniC77A,  HANGUL SYLLABLE-C77A
        '51067' => '14818'            , # G+14818  U+C77B  uniC77B,  HANGUL SYLLABLE-C77B
        '51068' => '2636'             , # G+2636  U+C77C  uniC77C,  HANGUL SYLLABLE-C77C
        '51069' => '2637'             , # G+2637  U+C77D  uniC77D,  HANGUL SYLLABLE-C77D
        '51070' => '2638'             , # G+2638  U+C77E  uniC77E,  HANGUL SYLLABLE-C77E
        '51071' => '14819'            , # G+14819  U+C77F  uniC77F,  HANGUL SYLLABLE-C77F
        '51072' => '14820'            , # G+14820  U+C780  uniC780,  HANGUL SYLLABLE-C780
        '51073' => '14821'            , # G+14821  U+C781  uniC781,  HANGUL SYLLABLE-C781
        '51074' => '14822'            , # G+14822  U+C782  uniC782,  HANGUL SYLLABLE-C782
        '51075' => '2639'             , # G+2639  U+C783  uniC783,  HANGUL SYLLABLE-C783
        '51076' => '2640'             , # G+2640  U+C784  uniC784,  HANGUL SYLLABLE-C784
        '51077' => '2641'             , # G+2641  U+C785  uniC785,  HANGUL SYLLABLE-C785
        '51078' => '14823'            , # G+14823  U+C786  uniC786,  HANGUL SYLLABLE-C786
        '51079' => '2642'             , # G+2642  U+C787  uniC787,  HANGUL SYLLABLE-C787
        '51080' => '2643'             , # G+2643  U+C788  uniC788,  HANGUL SYLLABLE-C788
        '51081' => '2644'             , # G+2644  U+C789  uniC789,  HANGUL SYLLABLE-C789
        '51082' => '2645'             , # G+2645  U+C78A  uniC78A,  HANGUL SYLLABLE-C78A
        '51083' => '14824'            , # G+14824  U+C78B  uniC78B,  HANGUL SYLLABLE-C78B
        '51084' => '14825'            , # G+14825  U+C78C  uniC78C,  HANGUL SYLLABLE-C78C
        '51085' => '14826'            , # G+14826  U+C78D  uniC78D,  HANGUL SYLLABLE-C78D
        '51086' => '2646'             , # G+2646  U+C78E  uniC78E,  HANGUL SYLLABLE-C78E
        '51087' => '14827'            , # G+14827  U+C78F  uniC78F,  HANGUL SYLLABLE-C78F
        '51088' => '2647'             , # G+2647  U+C790  uniC790,  HANGUL SYLLABLE-C790
        '51089' => '2648'             , # G+2648  U+C791  uniC791,  HANGUL SYLLABLE-C791
        '51090' => '14828'            , # G+14828  U+C792  uniC792,  HANGUL SYLLABLE-C792
        '51091' => '14829'            , # G+14829  U+C793  uniC793,  HANGUL SYLLABLE-C793
        '51092' => '2649'             , # G+2649  U+C794  uniC794,  HANGUL SYLLABLE-C794
        '51093' => '14830'            , # G+14830  U+C795  uniC795,  HANGUL SYLLABLE-C795
        '51094' => '2650'             , # G+2650  U+C796  uniC796,  HANGUL SYLLABLE-C796
        '51095' => '2651'             , # G+2651  U+C797  uniC797,  HANGUL SYLLABLE-C797
        '51096' => '2652'             , # G+2652  U+C798  uniC798,  HANGUL SYLLABLE-C798
        '51097' => '14831'            , # G+14831  U+C799  uniC799,  HANGUL SYLLABLE-C799
        '51098' => '2653'             , # G+2653  U+C79A  uniC79A,  HANGUL SYLLABLE-C79A
        '51099' => '14832'            , # G+14832  U+C79B  uniC79B,  HANGUL SYLLABLE-C79B
        '51100' => '14833'            , # G+14833  U+C79C  uniC79C,  HANGUL SYLLABLE-C79C
        '51101' => '14834'            , # G+14834  U+C79D  uniC79D,  HANGUL SYLLABLE-C79D
        '51102' => '14835'            , # G+14835  U+C79E  uniC79E,  HANGUL SYLLABLE-C79E
        '51103' => '14836'            , # G+14836  U+C79F  uniC79F,  HANGUL SYLLABLE-C79F
        '51104' => '2654'             , # G+2654  U+C7A0  uniC7A0,  HANGUL SYLLABLE-C7A0
        '51105' => '2655'             , # G+2655  U+C7A1  uniC7A1,  HANGUL SYLLABLE-C7A1
        '51106' => '14837'            , # G+14837  U+C7A2  uniC7A2,  HANGUL SYLLABLE-C7A2
        '51107' => '2656'             , # G+2656  U+C7A3  uniC7A3,  HANGUL SYLLABLE-C7A3
        '51108' => '2657'             , # G+2657  U+C7A4  uniC7A4,  HANGUL SYLLABLE-C7A4
        '51109' => '2658'             , # G+2658  U+C7A5  uniC7A5,  HANGUL SYLLABLE-C7A5
        '51110' => '2659'             , # G+2659  U+C7A6  uniC7A6,  HANGUL SYLLABLE-C7A6
        '51111' => '14838'            , # G+14838  U+C7A7  uniC7A7,  HANGUL SYLLABLE-C7A7
        '51112' => '14839'            , # G+14839  U+C7A8  uniC7A8,  HANGUL SYLLABLE-C7A8
        '51113' => '14840'            , # G+14840  U+C7A9  uniC7A9,  HANGUL SYLLABLE-C7A9
        '51114' => '14841'            , # G+14841  U+C7AA  uniC7AA,  HANGUL SYLLABLE-C7AA
        '51115' => '14842'            , # G+14842  U+C7AB  uniC7AB,  HANGUL SYLLABLE-C7AB
        '51116' => '2660'             , # G+2660  U+C7AC  uniC7AC,  HANGUL SYLLABLE-C7AC
        '51117' => '2661'             , # G+2661  U+C7AD  uniC7AD,  HANGUL SYLLABLE-C7AD
        '51118' => '14843'            , # G+14843  U+C7AE  uniC7AE,  HANGUL SYLLABLE-C7AE
        '51119' => '14844'            , # G+14844  U+C7AF  uniC7AF,  HANGUL SYLLABLE-C7AF
        '51120' => '2662'             , # G+2662  U+C7B0  uniC7B0,  HANGUL SYLLABLE-C7B0
        '51121' => '14845'            , # G+14845  U+C7B1  uniC7B1,  HANGUL SYLLABLE-C7B1
        '51122' => '14846'            , # G+14846  U+C7B2  uniC7B2,  HANGUL SYLLABLE-C7B2
        '51123' => '14847'            , # G+14847  U+C7B3  uniC7B3,  HANGUL SYLLABLE-C7B3
        '51124' => '2663'             , # G+2663  U+C7B4  uniC7B4,  HANGUL SYLLABLE-C7B4
        '51125' => '14848'            , # G+14848  U+C7B5  uniC7B5,  HANGUL SYLLABLE-C7B5
        '51126' => '14849'            , # G+14849  U+C7B6  uniC7B6,  HANGUL SYLLABLE-C7B6
        '51127' => '14850'            , # G+14850  U+C7B7  uniC7B7,  HANGUL SYLLABLE-C7B7
        '51128' => '14851'            , # G+14851  U+C7B8  uniC7B8,  HANGUL SYLLABLE-C7B8
        '51129' => '14852'            , # G+14852  U+C7B9  uniC7B9,  HANGUL SYLLABLE-C7B9
        '51130' => '14853'            , # G+14853  U+C7BA  uniC7BA,  HANGUL SYLLABLE-C7BA
        '51131' => '14854'            , # G+14854  U+C7BB  uniC7BB,  HANGUL SYLLABLE-C7BB
        '51132' => '2664'             , # G+2664  U+C7BC  uniC7BC,  HANGUL SYLLABLE-C7BC
        '51133' => '2665'             , # G+2665  U+C7BD  uniC7BD,  HANGUL SYLLABLE-C7BD
        '51134' => '14855'            , # G+14855  U+C7BE  uniC7BE,  HANGUL SYLLABLE-C7BE
        '51135' => '2666'             , # G+2666  U+C7BF  uniC7BF,  HANGUL SYLLABLE-C7BF
        '51136' => '2667'             , # G+2667  U+C7C0  uniC7C0,  HANGUL SYLLABLE-C7C0
        '51137' => '2668'             , # G+2668  U+C7C1  uniC7C1,  HANGUL SYLLABLE-C7C1
        '51138' => '14856'            , # G+14856  U+C7C2  uniC7C2,  HANGUL SYLLABLE-C7C2
        '51139' => '14857'            , # G+14857  U+C7C3  uniC7C3,  HANGUL SYLLABLE-C7C3
        '51140' => '14858'            , # G+14858  U+C7C4  uniC7C4,  HANGUL SYLLABLE-C7C4
        '51141' => '14859'            , # G+14859  U+C7C5  uniC7C5,  HANGUL SYLLABLE-C7C5
        '51142' => '14860'            , # G+14860  U+C7C6  uniC7C6,  HANGUL SYLLABLE-C7C6
        '51143' => '14861'            , # G+14861  U+C7C7  uniC7C7,  HANGUL SYLLABLE-C7C7
        '51144' => '2669'             , # G+2669  U+C7C8  uniC7C8,  HANGUL SYLLABLE-C7C8
        '51145' => '2670'             , # G+2670  U+C7C9  uniC7C9,  HANGUL SYLLABLE-C7C9
        '51146' => '14862'            , # G+14862  U+C7CA  uniC7CA,  HANGUL SYLLABLE-C7CA
        '51147' => '14863'            , # G+14863  U+C7CB  uniC7CB,  HANGUL SYLLABLE-C7CB
        '51148' => '2671'             , # G+2671  U+C7CC  uniC7CC,  HANGUL SYLLABLE-C7CC
        '51149' => '14864'            , # G+14864  U+C7CD  uniC7CD,  HANGUL SYLLABLE-C7CD
        '51150' => '2672'             , # G+2672  U+C7CE  uniC7CE,  HANGUL SYLLABLE-C7CE
        '51151' => '14865'            , # G+14865  U+C7CF  uniC7CF,  HANGUL SYLLABLE-C7CF
        '51152' => '2673'             , # G+2673  U+C7D0  uniC7D0,  HANGUL SYLLABLE-C7D0
        '51153' => '14866'            , # G+14866  U+C7D1  uniC7D1,  HANGUL SYLLABLE-C7D1
        '51154' => '14867'            , # G+14867  U+C7D2  uniC7D2,  HANGUL SYLLABLE-C7D2
        '51155' => '14868'            , # G+14868  U+C7D3  uniC7D3,  HANGUL SYLLABLE-C7D3
        '51156' => '14869'            , # G+14869  U+C7D4  uniC7D4,  HANGUL SYLLABLE-C7D4
        '51157' => '14870'            , # G+14870  U+C7D5  uniC7D5,  HANGUL SYLLABLE-C7D5
        '51158' => '14871'            , # G+14871  U+C7D6  uniC7D6,  HANGUL SYLLABLE-C7D6
        '51159' => '14872'            , # G+14872  U+C7D7  uniC7D7,  HANGUL SYLLABLE-C7D7
        '51160' => '2674'             , # G+2674  U+C7D8  uniC7D8,  HANGUL SYLLABLE-C7D8
        '51161' => '14873'            , # G+14873  U+C7D9  uniC7D9,  HANGUL SYLLABLE-C7D9
        '51162' => '14874'            , # G+14874  U+C7DA  uniC7DA,  HANGUL SYLLABLE-C7DA
        '51163' => '14875'            , # G+14875  U+C7DB  uniC7DB,  HANGUL SYLLABLE-C7DB
        '51164' => '14876'            , # G+14876  U+C7DC  uniC7DC,  HANGUL SYLLABLE-C7DC
        '51165' => '2675'             , # G+2675  U+C7DD  uniC7DD,  HANGUL SYLLABLE-C7DD
        '51166' => '14877'            , # G+14877  U+C7DE  uniC7DE,  HANGUL SYLLABLE-C7DE
        '51167' => '14878'            , # G+14878  U+C7DF  uniC7DF,  HANGUL SYLLABLE-C7DF
        '51168' => '14879'            , # G+14879  U+C7E0  uniC7E0,  HANGUL SYLLABLE-C7E0
        '51169' => '14880'            , # G+14880  U+C7E1  uniC7E1,  HANGUL SYLLABLE-C7E1
        '51170' => '14881'            , # G+14881  U+C7E2  uniC7E2,  HANGUL SYLLABLE-C7E2
        '51171' => '14882'            , # G+14882  U+C7E3  uniC7E3,  HANGUL SYLLABLE-C7E3
        '51172' => '2676'             , # G+2676  U+C7E4  uniC7E4,  HANGUL SYLLABLE-C7E4
        '51173' => '14883'            , # G+14883  U+C7E5  uniC7E5,  HANGUL SYLLABLE-C7E5
        '51174' => '14884'            , # G+14884  U+C7E6  uniC7E6,  HANGUL SYLLABLE-C7E6
        '51175' => '14885'            , # G+14885  U+C7E7  uniC7E7,  HANGUL SYLLABLE-C7E7
        '51176' => '2677'             , # G+2677  U+C7E8  uniC7E8,  HANGUL SYLLABLE-C7E8
        '51177' => '14886'            , # G+14886  U+C7E9  uniC7E9,  HANGUL SYLLABLE-C7E9
        '51178' => '14887'            , # G+14887  U+C7EA  uniC7EA,  HANGUL SYLLABLE-C7EA
        '51179' => '14888'            , # G+14888  U+C7EB  uniC7EB,  HANGUL SYLLABLE-C7EB
        '51180' => '2678'             , # G+2678  U+C7EC  uniC7EC,  HANGUL SYLLABLE-C7EC
        '51181' => '14889'            , # G+14889  U+C7ED  uniC7ED,  HANGUL SYLLABLE-C7ED
        '51182' => '14890'            , # G+14890  U+C7EE  uniC7EE,  HANGUL SYLLABLE-C7EE
        '51183' => '14891'            , # G+14891  U+C7EF  uniC7EF,  HANGUL SYLLABLE-C7EF
        '51184' => '14892'            , # G+14892  U+C7F0  uniC7F0,  HANGUL SYLLABLE-C7F0
        '51185' => '14893'            , # G+14893  U+C7F1  uniC7F1,  HANGUL SYLLABLE-C7F1
        '51186' => '14894'            , # G+14894  U+C7F2  uniC7F2,  HANGUL SYLLABLE-C7F2
        '51187' => '14895'            , # G+14895  U+C7F3  uniC7F3,  HANGUL SYLLABLE-C7F3
        '51188' => '14896'            , # G+14896  U+C7F4  uniC7F4,  HANGUL SYLLABLE-C7F4
        '51189' => '14897'            , # G+14897  U+C7F5  uniC7F5,  HANGUL SYLLABLE-C7F5
        '51190' => '14898'            , # G+14898  U+C7F6  uniC7F6,  HANGUL SYLLABLE-C7F6
        '51191' => '14899'            , # G+14899  U+C7F7  uniC7F7,  HANGUL SYLLABLE-C7F7
        '51192' => '14900'            , # G+14900  U+C7F8  uniC7F8,  HANGUL SYLLABLE-C7F8
        '51193' => '14901'            , # G+14901  U+C7F9  uniC7F9,  HANGUL SYLLABLE-C7F9
        '51194' => '14902'            , # G+14902  U+C7FA  uniC7FA,  HANGUL SYLLABLE-C7FA
        '51195' => '14903'            , # G+14903  U+C7FB  uniC7FB,  HANGUL SYLLABLE-C7FB
        '51196' => '14904'            , # G+14904  U+C7FC  uniC7FC,  HANGUL SYLLABLE-C7FC
        '51197' => '14905'            , # G+14905  U+C7FD  uniC7FD,  HANGUL SYLLABLE-C7FD
        '51198' => '14906'            , # G+14906  U+C7FE  uniC7FE,  HANGUL SYLLABLE-C7FE
        '51199' => '14907'            , # G+14907  U+C7FF  uniC7FF,  HANGUL SYLLABLE-C7FF
        '51200' => '2679'             , # G+2679  U+C800  uniC800,  HANGUL SYLLABLE-C800
        '51201' => '2680'             , # G+2680  U+C801  uniC801,  HANGUL SYLLABLE-C801
        '51202' => '14908'            , # G+14908  U+C802  uniC802,  HANGUL SYLLABLE-C802
        '51203' => '14909'            , # G+14909  U+C803  uniC803,  HANGUL SYLLABLE-C803
        '51204' => '2681'             , # G+2681  U+C804  uniC804,  HANGUL SYLLABLE-C804
        '51205' => '14910'            , # G+14910  U+C805  uniC805,  HANGUL SYLLABLE-C805
        '51206' => '14911'            , # G+14911  U+C806  uniC806,  HANGUL SYLLABLE-C806
        '51207' => '14912'            , # G+14912  U+C807  uniC807,  HANGUL SYLLABLE-C807
        '51208' => '2682'             , # G+2682  U+C808  uniC808,  HANGUL SYLLABLE-C808
        '51209' => '14913'            , # G+14913  U+C809  uniC809,  HANGUL SYLLABLE-C809
        '51210' => '2683'             , # G+2683  U+C80A  uniC80A,  HANGUL SYLLABLE-C80A
        '51211' => '14914'            , # G+14914  U+C80B  uniC80B,  HANGUL SYLLABLE-C80B
        '51212' => '14915'            , # G+14915  U+C80C  uniC80C,  HANGUL SYLLABLE-C80C
        '51213' => '14916'            , # G+14916  U+C80D  uniC80D,  HANGUL SYLLABLE-C80D
        '51214' => '14917'            , # G+14917  U+C80E  uniC80E,  HANGUL SYLLABLE-C80E
        '51215' => '14918'            , # G+14918  U+C80F  uniC80F,  HANGUL SYLLABLE-C80F
        '51216' => '2684'             , # G+2684  U+C810  uniC810,  HANGUL SYLLABLE-C810
        '51217' => '2685'             , # G+2685  U+C811  uniC811,  HANGUL SYLLABLE-C811
        '51218' => '14919'            , # G+14919  U+C812  uniC812,  HANGUL SYLLABLE-C812
        '51219' => '2686'             , # G+2686  U+C813  uniC813,  HANGUL SYLLABLE-C813
        '51220' => '14920'            , # G+14920  U+C814  uniC814,  HANGUL SYLLABLE-C814
        '51221' => '2687'             , # G+2687  U+C815  uniC815,  HANGUL SYLLABLE-C815
        '51222' => '2688'             , # G+2688  U+C816  uniC816,  HANGUL SYLLABLE-C816
        '51223' => '14921'            , # G+14921  U+C817  uniC817,  HANGUL SYLLABLE-C817
        '51224' => '14922'            , # G+14922  U+C818  uniC818,  HANGUL SYLLABLE-C818
        '51225' => '14923'            , # G+14923  U+C819  uniC819,  HANGUL SYLLABLE-C819
        '51226' => '14924'            , # G+14924  U+C81A  uniC81A,  HANGUL SYLLABLE-C81A
        '51227' => '14925'            , # G+14925  U+C81B  uniC81B,  HANGUL SYLLABLE-C81B
        '51228' => '2689'             , # G+2689  U+C81C  uniC81C,  HANGUL SYLLABLE-C81C
        '51229' => '2690'             , # G+2690  U+C81D  uniC81D,  HANGUL SYLLABLE-C81D
        '51230' => '14926'            , # G+14926  U+C81E  uniC81E,  HANGUL SYLLABLE-C81E
        '51231' => '14927'            , # G+14927  U+C81F  uniC81F,  HANGUL SYLLABLE-C81F
        '51232' => '2691'             , # G+2691  U+C820  uniC820,  HANGUL SYLLABLE-C820
        '51233' => '14928'            , # G+14928  U+C821  uniC821,  HANGUL SYLLABLE-C821
        '51234' => '14929'            , # G+14929  U+C822  uniC822,  HANGUL SYLLABLE-C822
        '51235' => '14930'            , # G+14930  U+C823  uniC823,  HANGUL SYLLABLE-C823
        '51236' => '2692'             , # G+2692  U+C824  uniC824,  HANGUL SYLLABLE-C824
        '51237' => '14931'            , # G+14931  U+C825  uniC825,  HANGUL SYLLABLE-C825
        '51238' => '14932'            , # G+14932  U+C826  uniC826,  HANGUL SYLLABLE-C826
        '51239' => '14933'            , # G+14933  U+C827  uniC827,  HANGUL SYLLABLE-C827
        '51240' => '14934'            , # G+14934  U+C828  uniC828,  HANGUL SYLLABLE-C828
        '51241' => '14935'            , # G+14935  U+C829  uniC829,  HANGUL SYLLABLE-C829
        '51242' => '14936'            , # G+14936  U+C82A  uniC82A,  HANGUL SYLLABLE-C82A
        '51243' => '14937'            , # G+14937  U+C82B  uniC82B,  HANGUL SYLLABLE-C82B
        '51244' => '2693'             , # G+2693  U+C82C  uniC82C,  HANGUL SYLLABLE-C82C
        '51245' => '2694'             , # G+2694  U+C82D  uniC82D,  HANGUL SYLLABLE-C82D
        '51246' => '14938'            , # G+14938  U+C82E  uniC82E,  HANGUL SYLLABLE-C82E
        '51247' => '2695'             , # G+2695  U+C82F  uniC82F,  HANGUL SYLLABLE-C82F
        '51248' => '14939'            , # G+14939  U+C830  uniC830,  HANGUL SYLLABLE-C830
        '51249' => '2696'             , # G+2696  U+C831  uniC831,  HANGUL SYLLABLE-C831
        '51250' => '14940'            , # G+14940  U+C832  uniC832,  HANGUL SYLLABLE-C832
        '51251' => '14941'            , # G+14941  U+C833  uniC833,  HANGUL SYLLABLE-C833
        '51252' => '14942'            , # G+14942  U+C834  uniC834,  HANGUL SYLLABLE-C834
        '51253' => '14943'            , # G+14943  U+C835  uniC835,  HANGUL SYLLABLE-C835
        '51254' => '14944'            , # G+14944  U+C836  uniC836,  HANGUL SYLLABLE-C836
        '51255' => '14945'            , # G+14945  U+C837  uniC837,  HANGUL SYLLABLE-C837
        '51256' => '2697'             , # G+2697  U+C838  uniC838,  HANGUL SYLLABLE-C838
        '51257' => '14946'            , # G+14946  U+C839  uniC839,  HANGUL SYLLABLE-C839
        '51258' => '14947'            , # G+14947  U+C83A  uniC83A,  HANGUL SYLLABLE-C83A
        '51259' => '14948'            , # G+14948  U+C83B  uniC83B,  HANGUL SYLLABLE-C83B
        '51260' => '2698'             , # G+2698  U+C83C  uniC83C,  HANGUL SYLLABLE-C83C
        '51261' => '14949'            , # G+14949  U+C83D  uniC83D,  HANGUL SYLLABLE-C83D
        '51262' => '14950'            , # G+14950  U+C83E  uniC83E,  HANGUL SYLLABLE-C83E
        '51263' => '14951'            , # G+14951  U+C83F  uniC83F,  HANGUL SYLLABLE-C83F
        '51264' => '2699'             , # G+2699  U+C840  uniC840,  HANGUL SYLLABLE-C840
        '51265' => '14952'            , # G+14952  U+C841  uniC841,  HANGUL SYLLABLE-C841
        '51266' => '14953'            , # G+14953  U+C842  uniC842,  HANGUL SYLLABLE-C842
        '51267' => '14954'            , # G+14954  U+C843  uniC843,  HANGUL SYLLABLE-C843
        '51268' => '14955'            , # G+14955  U+C844  uniC844,  HANGUL SYLLABLE-C844
        '51269' => '14956'            , # G+14956  U+C845  uniC845,  HANGUL SYLLABLE-C845
        '51270' => '14957'            , # G+14957  U+C846  uniC846,  HANGUL SYLLABLE-C846
        '51271' => '14958'            , # G+14958  U+C847  uniC847,  HANGUL SYLLABLE-C847
        '51272' => '2700'             , # G+2700  U+C848  uniC848,  HANGUL SYLLABLE-C848
        '51273' => '2701'             , # G+2701  U+C849  uniC849,  HANGUL SYLLABLE-C849
        '51274' => '14959'            , # G+14959  U+C84A  uniC84A,  HANGUL SYLLABLE-C84A
        '51275' => '14960'            , # G+14960  U+C84B  uniC84B,  HANGUL SYLLABLE-C84B
        '51276' => '2702'             , # G+2702  U+C84C  uniC84C,  HANGUL SYLLABLE-C84C
        '51277' => '2703'             , # G+2703  U+C84D  uniC84D,  HANGUL SYLLABLE-C84D
        '51278' => '14961'            , # G+14961  U+C84E  uniC84E,  HANGUL SYLLABLE-C84E
        '51279' => '14962'            , # G+14962  U+C84F  uniC84F,  HANGUL SYLLABLE-C84F
        '51280' => '14963'            , # G+14963  U+C850  uniC850,  HANGUL SYLLABLE-C850
        '51281' => '14964'            , # G+14964  U+C851  uniC851,  HANGUL SYLLABLE-C851
        '51282' => '14965'            , # G+14965  U+C852  uniC852,  HANGUL SYLLABLE-C852
        '51283' => '14966'            , # G+14966  U+C853  uniC853,  HANGUL SYLLABLE-C853
        '51284' => '2704'             , # G+2704  U+C854  uniC854,  HANGUL SYLLABLE-C854
        '51285' => '14967'            , # G+14967  U+C855  uniC855,  HANGUL SYLLABLE-C855
        '51286' => '14968'            , # G+14968  U+C856  uniC856,  HANGUL SYLLABLE-C856
        '51287' => '14969'            , # G+14969  U+C857  uniC857,  HANGUL SYLLABLE-C857
        '51288' => '14970'            , # G+14970  U+C858  uniC858,  HANGUL SYLLABLE-C858
        '51289' => '14971'            , # G+14971  U+C859  uniC859,  HANGUL SYLLABLE-C859
        '51290' => '14972'            , # G+14972  U+C85A  uniC85A,  HANGUL SYLLABLE-C85A
        '51291' => '14973'            , # G+14973  U+C85B  uniC85B,  HANGUL SYLLABLE-C85B
        '51292' => '14974'            , # G+14974  U+C85C  uniC85C,  HANGUL SYLLABLE-C85C
        '51293' => '14975'            , # G+14975  U+C85D  uniC85D,  HANGUL SYLLABLE-C85D
        '51294' => '14976'            , # G+14976  U+C85E  uniC85E,  HANGUL SYLLABLE-C85E
        '51295' => '14977'            , # G+14977  U+C85F  uniC85F,  HANGUL SYLLABLE-C85F
        '51296' => '14978'            , # G+14978  U+C860  uniC860,  HANGUL SYLLABLE-C860
        '51297' => '14979'            , # G+14979  U+C861  uniC861,  HANGUL SYLLABLE-C861
        '51298' => '14980'            , # G+14980  U+C862  uniC862,  HANGUL SYLLABLE-C862
        '51299' => '14981'            , # G+14981  U+C863  uniC863,  HANGUL SYLLABLE-C863
        '51300' => '14982'            , # G+14982  U+C864  uniC864,  HANGUL SYLLABLE-C864
        '51301' => '14983'            , # G+14983  U+C865  uniC865,  HANGUL SYLLABLE-C865
        '51302' => '14984'            , # G+14984  U+C866  uniC866,  HANGUL SYLLABLE-C866
        '51303' => '14985'            , # G+14985  U+C867  uniC867,  HANGUL SYLLABLE-C867
        '51304' => '14986'            , # G+14986  U+C868  uniC868,  HANGUL SYLLABLE-C868
        '51305' => '14987'            , # G+14987  U+C869  uniC869,  HANGUL SYLLABLE-C869
        '51306' => '14988'            , # G+14988  U+C86A  uniC86A,  HANGUL SYLLABLE-C86A
        '51307' => '14989'            , # G+14989  U+C86B  uniC86B,  HANGUL SYLLABLE-C86B
        '51308' => '14990'            , # G+14990  U+C86C  uniC86C,  HANGUL SYLLABLE-C86C
        '51309' => '14991'            , # G+14991  U+C86D  uniC86D,  HANGUL SYLLABLE-C86D
        '51310' => '14992'            , # G+14992  U+C86E  uniC86E,  HANGUL SYLLABLE-C86E
        '51311' => '14993'            , # G+14993  U+C86F  uniC86F,  HANGUL SYLLABLE-C86F
        '51312' => '2705'             , # G+2705  U+C870  uniC870,  HANGUL SYLLABLE-C870
        '51313' => '2706'             , # G+2706  U+C871  uniC871,  HANGUL SYLLABLE-C871
        '51314' => '14994'            , # G+14994  U+C872  uniC872,  HANGUL SYLLABLE-C872
        '51315' => '14995'            , # G+14995  U+C873  uniC873,  HANGUL SYLLABLE-C873
        '51316' => '2707'             , # G+2707  U+C874  uniC874,  HANGUL SYLLABLE-C874
        '51317' => '14996'            , # G+14996  U+C875  uniC875,  HANGUL SYLLABLE-C875
        '51318' => '14997'            , # G+14997  U+C876  uniC876,  HANGUL SYLLABLE-C876
        '51319' => '14998'            , # G+14998  U+C877  uniC877,  HANGUL SYLLABLE-C877
        '51320' => '2708'             , # G+2708  U+C878  uniC878,  HANGUL SYLLABLE-C878
        '51321' => '14999'            , # G+14999  U+C879  uniC879,  HANGUL SYLLABLE-C879
        '51322' => '2709'             , # G+2709  U+C87A  uniC87A,  HANGUL SYLLABLE-C87A
        '51323' => '15000'            , # G+15000  U+C87B  uniC87B,  HANGUL SYLLABLE-C87B
        '51324' => '15001'            , # G+15001  U+C87C  uniC87C,  HANGUL SYLLABLE-C87C
        '51325' => '15002'            , # G+15002  U+C87D  uniC87D,  HANGUL SYLLABLE-C87D
        '51326' => '15003'            , # G+15003  U+C87E  uniC87E,  HANGUL SYLLABLE-C87E
        '51327' => '15004'            , # G+15004  U+C87F  uniC87F,  HANGUL SYLLABLE-C87F
        '51328' => '2710'             , # G+2710  U+C880  uniC880,  HANGUL SYLLABLE-C880
        '51329' => '2711'             , # G+2711  U+C881  uniC881,  HANGUL SYLLABLE-C881
        '51330' => '15005'            , # G+15005  U+C882  uniC882,  HANGUL SYLLABLE-C882
        '51331' => '2712'             , # G+2712  U+C883  uniC883,  HANGUL SYLLABLE-C883
        '51332' => '15006'            , # G+15006  U+C884  uniC884,  HANGUL SYLLABLE-C884
        '51333' => '2713'             , # G+2713  U+C885  uniC885,  HANGUL SYLLABLE-C885
        '51334' => '2714'             , # G+2714  U+C886  uniC886,  HANGUL SYLLABLE-C886
        '51335' => '2715'             , # G+2715  U+C887  uniC887,  HANGUL SYLLABLE-C887
        '51336' => '15007'            , # G+15007  U+C888  uniC888,  HANGUL SYLLABLE-C888
        '51337' => '15008'            , # G+15008  U+C889  uniC889,  HANGUL SYLLABLE-C889
        '51338' => '15009'            , # G+15009  U+C88A  uniC88A,  HANGUL SYLLABLE-C88A
        '51339' => '2716'             , # G+2716  U+C88B  uniC88B,  HANGUL SYLLABLE-C88B
        '51340' => '2717'             , # G+2717  U+C88C  uniC88C,  HANGUL SYLLABLE-C88C
        '51341' => '2718'             , # G+2718  U+C88D  uniC88D,  HANGUL SYLLABLE-C88D
        '51342' => '15010'            , # G+15010  U+C88E  uniC88E,  HANGUL SYLLABLE-C88E
        '51343' => '15011'            , # G+15011  U+C88F  uniC88F,  HANGUL SYLLABLE-C88F
        '51344' => '15012'            , # G+15012  U+C890  uniC890,  HANGUL SYLLABLE-C890
        '51345' => '15013'            , # G+15013  U+C891  uniC891,  HANGUL SYLLABLE-C891
        '51346' => '15014'            , # G+15014  U+C892  uniC892,  HANGUL SYLLABLE-C892
        '51347' => '15015'            , # G+15015  U+C893  uniC893,  HANGUL SYLLABLE-C893
        '51348' => '2719'             , # G+2719  U+C894  uniC894,  HANGUL SYLLABLE-C894
        '51349' => '15016'            , # G+15016  U+C895  uniC895,  HANGUL SYLLABLE-C895
        '51350' => '15017'            , # G+15017  U+C896  uniC896,  HANGUL SYLLABLE-C896
        '51351' => '15018'            , # G+15018  U+C897  uniC897,  HANGUL SYLLABLE-C897
        '51352' => '15019'            , # G+15019  U+C898  uniC898,  HANGUL SYLLABLE-C898
        '51353' => '15020'            , # G+15020  U+C899  uniC899,  HANGUL SYLLABLE-C899
        '51354' => '15021'            , # G+15021  U+C89A  uniC89A,  HANGUL SYLLABLE-C89A
        '51355' => '15022'            , # G+15022  U+C89B  uniC89B,  HANGUL SYLLABLE-C89B
        '51356' => '15023'            , # G+15023  U+C89C  uniC89C,  HANGUL SYLLABLE-C89C
        '51357' => '2720'             , # G+2720  U+C89D  uniC89D,  HANGUL SYLLABLE-C89D
        '51358' => '15024'            , # G+15024  U+C89E  uniC89E,  HANGUL SYLLABLE-C89E
        '51359' => '2721'             , # G+2721  U+C89F  uniC89F,  HANGUL SYLLABLE-C89F
        '51360' => '15025'            , # G+15025  U+C8A0  uniC8A0,  HANGUL SYLLABLE-C8A0
        '51361' => '2722'             , # G+2722  U+C8A1  uniC8A1,  HANGUL SYLLABLE-C8A1
        '51362' => '15026'            , # G+15026  U+C8A2  uniC8A2,  HANGUL SYLLABLE-C8A2
        '51363' => '15027'            , # G+15027  U+C8A3  uniC8A3,  HANGUL SYLLABLE-C8A3
        '51364' => '15028'            , # G+15028  U+C8A4  uniC8A4,  HANGUL SYLLABLE-C8A4
        '51365' => '15029'            , # G+15029  U+C8A5  uniC8A5,  HANGUL SYLLABLE-C8A5
        '51366' => '15030'            , # G+15030  U+C8A6  uniC8A6,  HANGUL SYLLABLE-C8A6
        '51367' => '15031'            , # G+15031  U+C8A7  uniC8A7,  HANGUL SYLLABLE-C8A7
        '51368' => '2723'             , # G+2723  U+C8A8  uniC8A8,  HANGUL SYLLABLE-C8A8
        '51369' => '15032'            , # G+15032  U+C8A9  uniC8A9,  HANGUL SYLLABLE-C8A9
        '51370' => '15033'            , # G+15033  U+C8AA  uniC8AA,  HANGUL SYLLABLE-C8AA
        '51371' => '15034'            , # G+15034  U+C8AB  uniC8AB,  HANGUL SYLLABLE-C8AB
        '51372' => '15035'            , # G+15035  U+C8AC  uniC8AC,  HANGUL SYLLABLE-C8AC
        '51373' => '15036'            , # G+15036  U+C8AD  uniC8AD,  HANGUL SYLLABLE-C8AD
        '51374' => '15037'            , # G+15037  U+C8AE  uniC8AE,  HANGUL SYLLABLE-C8AE
        '51375' => '15038'            , # G+15038  U+C8AF  uniC8AF,  HANGUL SYLLABLE-C8AF
        '51376' => '15039'            , # G+15039  U+C8B0  uniC8B0,  HANGUL SYLLABLE-C8B0
        '51377' => '15040'            , # G+15040  U+C8B1  uniC8B1,  HANGUL SYLLABLE-C8B1
        '51378' => '15041'            , # G+15041  U+C8B2  uniC8B2,  HANGUL SYLLABLE-C8B2
        '51379' => '15042'            , # G+15042  U+C8B3  uniC8B3,  HANGUL SYLLABLE-C8B3
        '51380' => '15043'            , # G+15043  U+C8B4  uniC8B4,  HANGUL SYLLABLE-C8B4
        '51381' => '15044'            , # G+15044  U+C8B5  uniC8B5,  HANGUL SYLLABLE-C8B5
        '51382' => '15045'            , # G+15045  U+C8B6  uniC8B6,  HANGUL SYLLABLE-C8B6
        '51383' => '15046'            , # G+15046  U+C8B7  uniC8B7,  HANGUL SYLLABLE-C8B7
        '51384' => '15047'            , # G+15047  U+C8B8  uniC8B8,  HANGUL SYLLABLE-C8B8
        '51385' => '15048'            , # G+15048  U+C8B9  uniC8B9,  HANGUL SYLLABLE-C8B9
        '51386' => '15049'            , # G+15049  U+C8BA  uniC8BA,  HANGUL SYLLABLE-C8BA
        '51387' => '15050'            , # G+15050  U+C8BB  uniC8BB,  HANGUL SYLLABLE-C8BB
        '51388' => '2724'             , # G+2724  U+C8BC  uniC8BC,  HANGUL SYLLABLE-C8BC
        '51389' => '2725'             , # G+2725  U+C8BD  uniC8BD,  HANGUL SYLLABLE-C8BD
        '51390' => '15051'            , # G+15051  U+C8BE  uniC8BE,  HANGUL SYLLABLE-C8BE
        '51391' => '15052'            , # G+15052  U+C8BF  uniC8BF,  HANGUL SYLLABLE-C8BF
        '51392' => '15053'            , # G+15053  U+C8C0  uniC8C0,  HANGUL SYLLABLE-C8C0
        '51393' => '15054'            , # G+15054  U+C8C1  uniC8C1,  HANGUL SYLLABLE-C8C1
        '51394' => '15055'            , # G+15055  U+C8C2  uniC8C2,  HANGUL SYLLABLE-C8C2
        '51395' => '15056'            , # G+15056  U+C8C3  uniC8C3,  HANGUL SYLLABLE-C8C3
        '51396' => '2726'             , # G+2726  U+C8C4  uniC8C4,  HANGUL SYLLABLE-C8C4
        '51397' => '15057'            , # G+15057  U+C8C5  uniC8C5,  HANGUL SYLLABLE-C8C5
        '51398' => '15058'            , # G+15058  U+C8C6  uniC8C6,  HANGUL SYLLABLE-C8C6
        '51399' => '15059'            , # G+15059  U+C8C7  uniC8C7,  HANGUL SYLLABLE-C8C7
        '51400' => '2727'             , # G+2727  U+C8C8  uniC8C8,  HANGUL SYLLABLE-C8C8
        '51401' => '15060'            , # G+15060  U+C8C9  uniC8C9,  HANGUL SYLLABLE-C8C9
        '51402' => '15061'            , # G+15061  U+C8CA  uniC8CA,  HANGUL SYLLABLE-C8CA
        '51403' => '15062'            , # G+15062  U+C8CB  uniC8CB,  HANGUL SYLLABLE-C8CB
        '51404' => '2728'             , # G+2728  U+C8CC  uniC8CC,  HANGUL SYLLABLE-C8CC
        '51405' => '15063'            , # G+15063  U+C8CD  uniC8CD,  HANGUL SYLLABLE-C8CD
        '51406' => '15064'            , # G+15064  U+C8CE  uniC8CE,  HANGUL SYLLABLE-C8CE
        '51407' => '15065'            , # G+15065  U+C8CF  uniC8CF,  HANGUL SYLLABLE-C8CF
        '51408' => '15066'            , # G+15066  U+C8D0  uniC8D0,  HANGUL SYLLABLE-C8D0
        '51409' => '15067'            , # G+15067  U+C8D1  uniC8D1,  HANGUL SYLLABLE-C8D1
        '51410' => '15068'            , # G+15068  U+C8D2  uniC8D2,  HANGUL SYLLABLE-C8D2
        '51411' => '15069'            , # G+15069  U+C8D3  uniC8D3,  HANGUL SYLLABLE-C8D3
        '51412' => '2729'             , # G+2729  U+C8D4  uniC8D4,  HANGUL SYLLABLE-C8D4
        '51413' => '2730'             , # G+2730  U+C8D5  uniC8D5,  HANGUL SYLLABLE-C8D5
        '51414' => '15070'            , # G+15070  U+C8D6  uniC8D6,  HANGUL SYLLABLE-C8D6
        '51415' => '2731'             , # G+2731  U+C8D7  uniC8D7,  HANGUL SYLLABLE-C8D7
        '51416' => '15071'            , # G+15071  U+C8D8  uniC8D8,  HANGUL SYLLABLE-C8D8
        '51417' => '2732'             , # G+2732  U+C8D9  uniC8D9,  HANGUL SYLLABLE-C8D9
        '51418' => '15072'            , # G+15072  U+C8DA  uniC8DA,  HANGUL SYLLABLE-C8DA
        '51419' => '15073'            , # G+15073  U+C8DB  uniC8DB,  HANGUL SYLLABLE-C8DB
        '51420' => '15074'            , # G+15074  U+C8DC  uniC8DC,  HANGUL SYLLABLE-C8DC
        '51421' => '15075'            , # G+15075  U+C8DD  uniC8DD,  HANGUL SYLLABLE-C8DD
        '51422' => '15076'            , # G+15076  U+C8DE  uniC8DE,  HANGUL SYLLABLE-C8DE
        '51423' => '15077'            , # G+15077  U+C8DF  uniC8DF,  HANGUL SYLLABLE-C8DF
        '51424' => '2733'             , # G+2733  U+C8E0  uniC8E0,  HANGUL SYLLABLE-C8E0
        '51425' => '2734'             , # G+2734  U+C8E1  uniC8E1,  HANGUL SYLLABLE-C8E1
        '51426' => '15078'            , # G+15078  U+C8E2  uniC8E2,  HANGUL SYLLABLE-C8E2
        '51427' => '15079'            , # G+15079  U+C8E3  uniC8E3,  HANGUL SYLLABLE-C8E3
        '51428' => '2735'             , # G+2735  U+C8E4  uniC8E4,  HANGUL SYLLABLE-C8E4
        '51429' => '15080'            , # G+15080  U+C8E5  uniC8E5,  HANGUL SYLLABLE-C8E5
        '51430' => '15081'            , # G+15081  U+C8E6  uniC8E6,  HANGUL SYLLABLE-C8E6
        '51431' => '15082'            , # G+15082  U+C8E7  uniC8E7,  HANGUL SYLLABLE-C8E7
        '51432' => '15083'            , # G+15083  U+C8E8  uniC8E8,  HANGUL SYLLABLE-C8E8
        '51433' => '15084'            , # G+15084  U+C8E9  uniC8E9,  HANGUL SYLLABLE-C8E9
        '51434' => '15085'            , # G+15085  U+C8EA  uniC8EA,  HANGUL SYLLABLE-C8EA
        '51435' => '15086'            , # G+15086  U+C8EB  uniC8EB,  HANGUL SYLLABLE-C8EB
        '51436' => '15087'            , # G+15087  U+C8EC  uniC8EC,  HANGUL SYLLABLE-C8EC
        '51437' => '15088'            , # G+15088  U+C8ED  uniC8ED,  HANGUL SYLLABLE-C8ED
        '51438' => '15089'            , # G+15089  U+C8EE  uniC8EE,  HANGUL SYLLABLE-C8EE
        '51439' => '15090'            , # G+15090  U+C8EF  uniC8EF,  HANGUL SYLLABLE-C8EF
        '51440' => '15091'            , # G+15091  U+C8F0  uniC8F0,  HANGUL SYLLABLE-C8F0
        '51441' => '15092'            , # G+15092  U+C8F1  uniC8F1,  HANGUL SYLLABLE-C8F1
        '51442' => '15093'            , # G+15093  U+C8F2  uniC8F2,  HANGUL SYLLABLE-C8F2
        '51443' => '15094'            , # G+15094  U+C8F3  uniC8F3,  HANGUL SYLLABLE-C8F3
        '51444' => '15095'            , # G+15095  U+C8F4  uniC8F4,  HANGUL SYLLABLE-C8F4
        '51445' => '2736'             , # G+2736  U+C8F5  uniC8F5,  HANGUL SYLLABLE-C8F5
        '51446' => '15096'            , # G+15096  U+C8F6  uniC8F6,  HANGUL SYLLABLE-C8F6
        '51447' => '15097'            , # G+15097  U+C8F7  uniC8F7,  HANGUL SYLLABLE-C8F7
        '51448' => '15098'            , # G+15098  U+C8F8  uniC8F8,  HANGUL SYLLABLE-C8F8
        '51449' => '15099'            , # G+15099  U+C8F9  uniC8F9,  HANGUL SYLLABLE-C8F9
        '51450' => '15100'            , # G+15100  U+C8FA  uniC8FA,  HANGUL SYLLABLE-C8FA
        '51451' => '15101'            , # G+15101  U+C8FB  uniC8FB,  HANGUL SYLLABLE-C8FB
        '51452' => '2737'             , # G+2737  U+C8FC  uniC8FC,  HANGUL SYLLABLE-C8FC
        '51453' => '2738'             , # G+2738  U+C8FD  uniC8FD,  HANGUL SYLLABLE-C8FD
        '51454' => '15102'            , # G+15102  U+C8FE  uniC8FE,  HANGUL SYLLABLE-C8FE
        '51455' => '15103'            , # G+15103  U+C8FF  uniC8FF,  HANGUL SYLLABLE-C8FF
        '51456' => '2739'             , # G+2739  U+C900  uniC900,  HANGUL SYLLABLE-C900
        '51457' => '15104'            , # G+15104  U+C901  uniC901,  HANGUL SYLLABLE-C901
        '51458' => '15105'            , # G+15105  U+C902  uniC902,  HANGUL SYLLABLE-C902
        '51459' => '15106'            , # G+15106  U+C903  uniC903,  HANGUL SYLLABLE-C903
        '51460' => '2740'             , # G+2740  U+C904  uniC904,  HANGUL SYLLABLE-C904
        '51461' => '2741'             , # G+2741  U+C905  uniC905,  HANGUL SYLLABLE-C905
        '51462' => '2742'             , # G+2742  U+C906  uniC906,  HANGUL SYLLABLE-C906
        '51463' => '15107'            , # G+15107  U+C907  uniC907,  HANGUL SYLLABLE-C907
        '51464' => '15108'            , # G+15108  U+C908  uniC908,  HANGUL SYLLABLE-C908
        '51465' => '15109'            , # G+15109  U+C909  uniC909,  HANGUL SYLLABLE-C909
        '51466' => '15110'            , # G+15110  U+C90A  uniC90A,  HANGUL SYLLABLE-C90A
        '51467' => '15111'            , # G+15111  U+C90B  uniC90B,  HANGUL SYLLABLE-C90B
        '51468' => '2743'             , # G+2743  U+C90C  uniC90C,  HANGUL SYLLABLE-C90C
        '51469' => '2744'             , # G+2744  U+C90D  uniC90D,  HANGUL SYLLABLE-C90D
        '51470' => '15112'            , # G+15112  U+C90E  uniC90E,  HANGUL SYLLABLE-C90E
        '51471' => '2745'             , # G+2745  U+C90F  uniC90F,  HANGUL SYLLABLE-C90F
        '51472' => '15113'            , # G+15113  U+C910  uniC910,  HANGUL SYLLABLE-C910
        '51473' => '2746'             , # G+2746  U+C911  uniC911,  HANGUL SYLLABLE-C911
        '51474' => '15114'            , # G+15114  U+C912  uniC912,  HANGUL SYLLABLE-C912
        '51475' => '15115'            , # G+15115  U+C913  uniC913,  HANGUL SYLLABLE-C913
        '51476' => '15116'            , # G+15116  U+C914  uniC914,  HANGUL SYLLABLE-C914
        '51477' => '15117'            , # G+15117  U+C915  uniC915,  HANGUL SYLLABLE-C915
        '51478' => '15118'            , # G+15118  U+C916  uniC916,  HANGUL SYLLABLE-C916
        '51479' => '15119'            , # G+15119  U+C917  uniC917,  HANGUL SYLLABLE-C917
        '51480' => '2747'             , # G+2747  U+C918  uniC918,  HANGUL SYLLABLE-C918
        '51481' => '15120'            , # G+15120  U+C919  uniC919,  HANGUL SYLLABLE-C919
        '51482' => '15121'            , # G+15121  U+C91A  uniC91A,  HANGUL SYLLABLE-C91A
        '51483' => '15122'            , # G+15122  U+C91B  uniC91B,  HANGUL SYLLABLE-C91B
        '51484' => '15123'            , # G+15123  U+C91C  uniC91C,  HANGUL SYLLABLE-C91C
        '51485' => '15124'            , # G+15124  U+C91D  uniC91D,  HANGUL SYLLABLE-C91D
        '51486' => '15125'            , # G+15125  U+C91E  uniC91E,  HANGUL SYLLABLE-C91E
        '51487' => '15126'            , # G+15126  U+C91F  uniC91F,  HANGUL SYLLABLE-C91F
        '51488' => '15127'            , # G+15127  U+C920  uniC920,  HANGUL SYLLABLE-C920
        '51489' => '15128'            , # G+15128  U+C921  uniC921,  HANGUL SYLLABLE-C921
        '51490' => '15129'            , # G+15129  U+C922  uniC922,  HANGUL SYLLABLE-C922
        '51491' => '15130'            , # G+15130  U+C923  uniC923,  HANGUL SYLLABLE-C923
        '51492' => '15131'            , # G+15131  U+C924  uniC924,  HANGUL SYLLABLE-C924
        '51493' => '15132'            , # G+15132  U+C925  uniC925,  HANGUL SYLLABLE-C925
        '51494' => '15133'            , # G+15133  U+C926  uniC926,  HANGUL SYLLABLE-C926
        '51495' => '15134'            , # G+15134  U+C927  uniC927,  HANGUL SYLLABLE-C927
        '51496' => '15135'            , # G+15135  U+C928  uniC928,  HANGUL SYLLABLE-C928
        '51497' => '15136'            , # G+15136  U+C929  uniC929,  HANGUL SYLLABLE-C929
        '51498' => '15137'            , # G+15137  U+C92A  uniC92A,  HANGUL SYLLABLE-C92A
        '51499' => '15138'            , # G+15138  U+C92B  uniC92B,  HANGUL SYLLABLE-C92B
        '51500' => '2748'             , # G+2748  U+C92C  uniC92C,  HANGUL SYLLABLE-C92C
        '51501' => '15139'            , # G+15139  U+C92D  uniC92D,  HANGUL SYLLABLE-C92D
        '51502' => '15140'            , # G+15140  U+C92E  uniC92E,  HANGUL SYLLABLE-C92E
        '51503' => '15141'            , # G+15141  U+C92F  uniC92F,  HANGUL SYLLABLE-C92F
        '51504' => '15142'            , # G+15142  U+C930  uniC930,  HANGUL SYLLABLE-C930
        '51505' => '15143'            , # G+15143  U+C931  uniC931,  HANGUL SYLLABLE-C931
        '51506' => '15144'            , # G+15144  U+C932  uniC932,  HANGUL SYLLABLE-C932
        '51507' => '15145'            , # G+15145  U+C933  uniC933,  HANGUL SYLLABLE-C933
        '51508' => '2749'             , # G+2749  U+C934  uniC934,  HANGUL SYLLABLE-C934
        '51509' => '15146'            , # G+15146  U+C935  uniC935,  HANGUL SYLLABLE-C935
        '51510' => '15147'            , # G+15147  U+C936  uniC936,  HANGUL SYLLABLE-C936
        '51511' => '15148'            , # G+15148  U+C937  uniC937,  HANGUL SYLLABLE-C937
        '51512' => '15149'            , # G+15149  U+C938  uniC938,  HANGUL SYLLABLE-C938
        '51513' => '15150'            , # G+15150  U+C939  uniC939,  HANGUL SYLLABLE-C939
        '51514' => '15151'            , # G+15151  U+C93A  uniC93A,  HANGUL SYLLABLE-C93A
        '51515' => '15152'            , # G+15152  U+C93B  uniC93B,  HANGUL SYLLABLE-C93B
        '51516' => '15153'            , # G+15153  U+C93C  uniC93C,  HANGUL SYLLABLE-C93C
        '51517' => '15154'            , # G+15154  U+C93D  uniC93D,  HANGUL SYLLABLE-C93D
        '51518' => '15155'            , # G+15155  U+C93E  uniC93E,  HANGUL SYLLABLE-C93E
        '51519' => '15156'            , # G+15156  U+C93F  uniC93F,  HANGUL SYLLABLE-C93F
        '51520' => '15157'            , # G+15157  U+C940  uniC940,  HANGUL SYLLABLE-C940
        '51521' => '15158'            , # G+15158  U+C941  uniC941,  HANGUL SYLLABLE-C941
        '51522' => '15159'            , # G+15159  U+C942  uniC942,  HANGUL SYLLABLE-C942
        '51523' => '15160'            , # G+15160  U+C943  uniC943,  HANGUL SYLLABLE-C943
        '51524' => '15161'            , # G+15161  U+C944  uniC944,  HANGUL SYLLABLE-C944
        '51525' => '15162'            , # G+15162  U+C945  uniC945,  HANGUL SYLLABLE-C945
        '51526' => '15163'            , # G+15163  U+C946  uniC946,  HANGUL SYLLABLE-C946
        '51527' => '15164'            , # G+15164  U+C947  uniC947,  HANGUL SYLLABLE-C947
        '51528' => '15165'            , # G+15165  U+C948  uniC948,  HANGUL SYLLABLE-C948
        '51529' => '15166'            , # G+15166  U+C949  uniC949,  HANGUL SYLLABLE-C949
        '51530' => '15167'            , # G+15167  U+C94A  uniC94A,  HANGUL SYLLABLE-C94A
        '51531' => '15168'            , # G+15168  U+C94B  uniC94B,  HANGUL SYLLABLE-C94B
        '51532' => '15169'            , # G+15169  U+C94C  uniC94C,  HANGUL SYLLABLE-C94C
        '51533' => '15170'            , # G+15170  U+C94D  uniC94D,  HANGUL SYLLABLE-C94D
        '51534' => '15171'            , # G+15171  U+C94E  uniC94E,  HANGUL SYLLABLE-C94E
        '51535' => '15172'            , # G+15172  U+C94F  uniC94F,  HANGUL SYLLABLE-C94F
        '51536' => '2750'             , # G+2750  U+C950  uniC950,  HANGUL SYLLABLE-C950
        '51537' => '2751'             , # G+2751  U+C951  uniC951,  HANGUL SYLLABLE-C951
        '51538' => '15173'            , # G+15173  U+C952  uniC952,  HANGUL SYLLABLE-C952
        '51539' => '15174'            , # G+15174  U+C953  uniC953,  HANGUL SYLLABLE-C953
        '51540' => '2752'             , # G+2752  U+C954  uniC954,  HANGUL SYLLABLE-C954
        '51541' => '15175'            , # G+15175  U+C955  uniC955,  HANGUL SYLLABLE-C955
        '51542' => '15176'            , # G+15176  U+C956  uniC956,  HANGUL SYLLABLE-C956
        '51543' => '15177'            , # G+15177  U+C957  uniC957,  HANGUL SYLLABLE-C957
        '51544' => '2753'             , # G+2753  U+C958  uniC958,  HANGUL SYLLABLE-C958
        '51545' => '15178'            , # G+15178  U+C959  uniC959,  HANGUL SYLLABLE-C959
        '51546' => '15179'            , # G+15179  U+C95A  uniC95A,  HANGUL SYLLABLE-C95A
        '51547' => '15180'            , # G+15180  U+C95B  uniC95B,  HANGUL SYLLABLE-C95B
        '51548' => '15181'            , # G+15181  U+C95C  uniC95C,  HANGUL SYLLABLE-C95C
        '51549' => '15182'            , # G+15182  U+C95D  uniC95D,  HANGUL SYLLABLE-C95D
        '51550' => '15183'            , # G+15183  U+C95E  uniC95E,  HANGUL SYLLABLE-C95E
        '51551' => '15184'            , # G+15184  U+C95F  uniC95F,  HANGUL SYLLABLE-C95F
        '51552' => '2754'             , # G+2754  U+C960  uniC960,  HANGUL SYLLABLE-C960
        '51553' => '2755'             , # G+2755  U+C961  uniC961,  HANGUL SYLLABLE-C961
        '51554' => '15185'            , # G+15185  U+C962  uniC962,  HANGUL SYLLABLE-C962
        '51555' => '2756'             , # G+2756  U+C963  uniC963,  HANGUL SYLLABLE-C963
        '51556' => '15186'            , # G+15186  U+C964  uniC964,  HANGUL SYLLABLE-C964
        '51557' => '15187'            , # G+15187  U+C965  uniC965,  HANGUL SYLLABLE-C965
        '51558' => '15188'            , # G+15188  U+C966  uniC966,  HANGUL SYLLABLE-C966
        '51559' => '15189'            , # G+15189  U+C967  uniC967,  HANGUL SYLLABLE-C967
        '51560' => '15190'            , # G+15190  U+C968  uniC968,  HANGUL SYLLABLE-C968
        '51561' => '15191'            , # G+15191  U+C969  uniC969,  HANGUL SYLLABLE-C969
        '51562' => '15192'            , # G+15192  U+C96A  uniC96A,  HANGUL SYLLABLE-C96A
        '51563' => '15193'            , # G+15193  U+C96B  uniC96B,  HANGUL SYLLABLE-C96B
        '51564' => '2757'             , # G+2757  U+C96C  uniC96C,  HANGUL SYLLABLE-C96C
        '51565' => '15194'            , # G+15194  U+C96D  uniC96D,  HANGUL SYLLABLE-C96D
        '51566' => '15195'            , # G+15195  U+C96E  uniC96E,  HANGUL SYLLABLE-C96E
        '51567' => '15196'            , # G+15196  U+C96F  uniC96F,  HANGUL SYLLABLE-C96F
        '51568' => '2758'             , # G+2758  U+C970  uniC970,  HANGUL SYLLABLE-C970
        '51569' => '15197'            , # G+15197  U+C971  uniC971,  HANGUL SYLLABLE-C971
        '51570' => '15198'            , # G+15198  U+C972  uniC972,  HANGUL SYLLABLE-C972
        '51571' => '15199'            , # G+15199  U+C973  uniC973,  HANGUL SYLLABLE-C973
        '51572' => '2759'             , # G+2759  U+C974  uniC974,  HANGUL SYLLABLE-C974
        '51573' => '15200'            , # G+15200  U+C975  uniC975,  HANGUL SYLLABLE-C975
        '51574' => '15201'            , # G+15201  U+C976  uniC976,  HANGUL SYLLABLE-C976
        '51575' => '15202'            , # G+15202  U+C977  uniC977,  HANGUL SYLLABLE-C977
        '51576' => '15203'            , # G+15203  U+C978  uniC978,  HANGUL SYLLABLE-C978
        '51577' => '15204'            , # G+15204  U+C979  uniC979,  HANGUL SYLLABLE-C979
        '51578' => '15205'            , # G+15205  U+C97A  uniC97A,  HANGUL SYLLABLE-C97A
        '51579' => '15206'            , # G+15206  U+C97B  uniC97B,  HANGUL SYLLABLE-C97B
        '51580' => '2760'             , # G+2760  U+C97C  uniC97C,  HANGUL SYLLABLE-C97C
        '51581' => '15207'            , # G+15207  U+C97D  uniC97D,  HANGUL SYLLABLE-C97D
        '51582' => '15208'            , # G+15208  U+C97E  uniC97E,  HANGUL SYLLABLE-C97E
        '51583' => '15209'            , # G+15209  U+C97F  uniC97F,  HANGUL SYLLABLE-C97F
        '51584' => '15210'            , # G+15210  U+C980  uniC980,  HANGUL SYLLABLE-C980
        '51585' => '15211'            , # G+15211  U+C981  uniC981,  HANGUL SYLLABLE-C981
        '51586' => '15212'            , # G+15212  U+C982  uniC982,  HANGUL SYLLABLE-C982
        '51587' => '15213'            , # G+15213  U+C983  uniC983,  HANGUL SYLLABLE-C983
        '51588' => '15214'            , # G+15214  U+C984  uniC984,  HANGUL SYLLABLE-C984
        '51589' => '15215'            , # G+15215  U+C985  uniC985,  HANGUL SYLLABLE-C985
        '51590' => '15216'            , # G+15216  U+C986  uniC986,  HANGUL SYLLABLE-C986
        '51591' => '15217'            , # G+15217  U+C987  uniC987,  HANGUL SYLLABLE-C987
        '51592' => '2761'             , # G+2761  U+C988  uniC988,  HANGUL SYLLABLE-C988
        '51593' => '2762'             , # G+2762  U+C989  uniC989,  HANGUL SYLLABLE-C989
        '51594' => '15218'            , # G+15218  U+C98A  uniC98A,  HANGUL SYLLABLE-C98A
        '51595' => '15219'            , # G+15219  U+C98B  uniC98B,  HANGUL SYLLABLE-C98B
        '51596' => '2763'             , # G+2763  U+C98C  uniC98C,  HANGUL SYLLABLE-C98C
        '51597' => '15220'            , # G+15220  U+C98D  uniC98D,  HANGUL SYLLABLE-C98D
        '51598' => '15221'            , # G+15221  U+C98E  uniC98E,  HANGUL SYLLABLE-C98E
        '51599' => '15222'            , # G+15222  U+C98F  uniC98F,  HANGUL SYLLABLE-C98F
        '51600' => '2764'             , # G+2764  U+C990  uniC990,  HANGUL SYLLABLE-C990
        '51601' => '15223'            , # G+15223  U+C991  uniC991,  HANGUL SYLLABLE-C991
        '51602' => '15224'            , # G+15224  U+C992  uniC992,  HANGUL SYLLABLE-C992
        '51603' => '15225'            , # G+15225  U+C993  uniC993,  HANGUL SYLLABLE-C993
        '51604' => '15226'            , # G+15226  U+C994  uniC994,  HANGUL SYLLABLE-C994
        '51605' => '15227'            , # G+15227  U+C995  uniC995,  HANGUL SYLLABLE-C995
        '51606' => '15228'            , # G+15228  U+C996  uniC996,  HANGUL SYLLABLE-C996
        '51607' => '15229'            , # G+15229  U+C997  uniC997,  HANGUL SYLLABLE-C997
        '51608' => '2765'             , # G+2765  U+C998  uniC998,  HANGUL SYLLABLE-C998
        '51609' => '2766'             , # G+2766  U+C999  uniC999,  HANGUL SYLLABLE-C999
        '51610' => '15230'            , # G+15230  U+C99A  uniC99A,  HANGUL SYLLABLE-C99A
        '51611' => '2767'             , # G+2767  U+C99B  uniC99B,  HANGUL SYLLABLE-C99B
        '51612' => '15231'            , # G+15231  U+C99C  uniC99C,  HANGUL SYLLABLE-C99C
        '51613' => '2768'             , # G+2768  U+C99D  uniC99D,  HANGUL SYLLABLE-C99D
        '51614' => '15232'            , # G+15232  U+C99E  uniC99E,  HANGUL SYLLABLE-C99E
        '51615' => '15233'            , # G+15233  U+C99F  uniC99F,  HANGUL SYLLABLE-C99F
        '51616' => '15234'            , # G+15234  U+C9A0  uniC9A0,  HANGUL SYLLABLE-C9A0
        '51617' => '15235'            , # G+15235  U+C9A1  uniC9A1,  HANGUL SYLLABLE-C9A1
        '51618' => '15236'            , # G+15236  U+C9A2  uniC9A2,  HANGUL SYLLABLE-C9A2
        '51619' => '15237'            , # G+15237  U+C9A3  uniC9A3,  HANGUL SYLLABLE-C9A3
        '51620' => '15238'            , # G+15238  U+C9A4  uniC9A4,  HANGUL SYLLABLE-C9A4
        '51621' => '15239'            , # G+15239  U+C9A5  uniC9A5,  HANGUL SYLLABLE-C9A5
        '51622' => '15240'            , # G+15240  U+C9A6  uniC9A6,  HANGUL SYLLABLE-C9A6
        '51623' => '15241'            , # G+15241  U+C9A7  uniC9A7,  HANGUL SYLLABLE-C9A7
        '51624' => '15242'            , # G+15242  U+C9A8  uniC9A8,  HANGUL SYLLABLE-C9A8
        '51625' => '15243'            , # G+15243  U+C9A9  uniC9A9,  HANGUL SYLLABLE-C9A9
        '51626' => '15244'            , # G+15244  U+C9AA  uniC9AA,  HANGUL SYLLABLE-C9AA
        '51627' => '15245'            , # G+15245  U+C9AB  uniC9AB,  HANGUL SYLLABLE-C9AB
        '51628' => '15246'            , # G+15246  U+C9AC  uniC9AC,  HANGUL SYLLABLE-C9AC
        '51629' => '15247'            , # G+15247  U+C9AD  uniC9AD,  HANGUL SYLLABLE-C9AD
        '51630' => '15248'            , # G+15248  U+C9AE  uniC9AE,  HANGUL SYLLABLE-C9AE
        '51631' => '15249'            , # G+15249  U+C9AF  uniC9AF,  HANGUL SYLLABLE-C9AF
        '51632' => '15250'            , # G+15250  U+C9B0  uniC9B0,  HANGUL SYLLABLE-C9B0
        '51633' => '15251'            , # G+15251  U+C9B1  uniC9B1,  HANGUL SYLLABLE-C9B1
        '51634' => '15252'            , # G+15252  U+C9B2  uniC9B2,  HANGUL SYLLABLE-C9B2
        '51635' => '15253'            , # G+15253  U+C9B3  uniC9B3,  HANGUL SYLLABLE-C9B3
        '51636' => '15254'            , # G+15254  U+C9B4  uniC9B4,  HANGUL SYLLABLE-C9B4
        '51637' => '15255'            , # G+15255  U+C9B5  uniC9B5,  HANGUL SYLLABLE-C9B5
        '51638' => '15256'            , # G+15256  U+C9B6  uniC9B6,  HANGUL SYLLABLE-C9B6
        '51639' => '15257'            , # G+15257  U+C9B7  uniC9B7,  HANGUL SYLLABLE-C9B7
        '51640' => '15258'            , # G+15258  U+C9B8  uniC9B8,  HANGUL SYLLABLE-C9B8
        '51641' => '15259'            , # G+15259  U+C9B9  uniC9B9,  HANGUL SYLLABLE-C9B9
        '51642' => '15260'            , # G+15260  U+C9BA  uniC9BA,  HANGUL SYLLABLE-C9BA
        '51643' => '15261'            , # G+15261  U+C9BB  uniC9BB,  HANGUL SYLLABLE-C9BB
        '51644' => '15262'            , # G+15262  U+C9BC  uniC9BC,  HANGUL SYLLABLE-C9BC
        '51645' => '15263'            , # G+15263  U+C9BD  uniC9BD,  HANGUL SYLLABLE-C9BD
        '51646' => '15264'            , # G+15264  U+C9BE  uniC9BE,  HANGUL SYLLABLE-C9BE
        '51647' => '15265'            , # G+15265  U+C9BF  uniC9BF,  HANGUL SYLLABLE-C9BF
        '51648' => '2769'             , # G+2769  U+C9C0  uniC9C0,  HANGUL SYLLABLE-C9C0
        '51649' => '2770'             , # G+2770  U+C9C1  uniC9C1,  HANGUL SYLLABLE-C9C1
        '51650' => '15266'            , # G+15266  U+C9C2  uniC9C2,  HANGUL SYLLABLE-C9C2
        '51651' => '15267'            , # G+15267  U+C9C3  uniC9C3,  HANGUL SYLLABLE-C9C3
        '51652' => '2771'             , # G+2771  U+C9C4  uniC9C4,  HANGUL SYLLABLE-C9C4
        '51653' => '15268'            , # G+15268  U+C9C5  uniC9C5,  HANGUL SYLLABLE-C9C5
        '51654' => '15269'            , # G+15269  U+C9C6  uniC9C6,  HANGUL SYLLABLE-C9C6
        '51655' => '2772'             , # G+2772  U+C9C7  uniC9C7,  HANGUL SYLLABLE-C9C7
        '51656' => '2773'             , # G+2773  U+C9C8  uniC9C8,  HANGUL SYLLABLE-C9C8
        '51657' => '15270'            , # G+15270  U+C9C9  uniC9C9,  HANGUL SYLLABLE-C9C9
        '51658' => '2774'             , # G+2774  U+C9CA  uniC9CA,  HANGUL SYLLABLE-C9CA
        '51659' => '15271'            , # G+15271  U+C9CB  uniC9CB,  HANGUL SYLLABLE-C9CB
        '51660' => '15272'            , # G+15272  U+C9CC  uniC9CC,  HANGUL SYLLABLE-C9CC
        '51661' => '15273'            , # G+15273  U+C9CD  uniC9CD,  HANGUL SYLLABLE-C9CD
        '51662' => '15274'            , # G+15274  U+C9CE  uniC9CE,  HANGUL SYLLABLE-C9CE
        '51663' => '15275'            , # G+15275  U+C9CF  uniC9CF,  HANGUL SYLLABLE-C9CF
        '51664' => '2775'             , # G+2775  U+C9D0  uniC9D0,  HANGUL SYLLABLE-C9D0
        '51665' => '2776'             , # G+2776  U+C9D1  uniC9D1,  HANGUL SYLLABLE-C9D1
        '51666' => '15276'            , # G+15276  U+C9D2  uniC9D2,  HANGUL SYLLABLE-C9D2
        '51667' => '2777'             , # G+2777  U+C9D3  uniC9D3,  HANGUL SYLLABLE-C9D3
        '51668' => '15277'            , # G+15277  U+C9D4  uniC9D4,  HANGUL SYLLABLE-C9D4
        '51669' => '2778'             , # G+2778  U+C9D5  uniC9D5,  HANGUL SYLLABLE-C9D5
        '51670' => '2779'             , # G+2779  U+C9D6  uniC9D6,  HANGUL SYLLABLE-C9D6
        '51671' => '15278'            , # G+15278  U+C9D7  uniC9D7,  HANGUL SYLLABLE-C9D7
        '51672' => '15279'            , # G+15279  U+C9D8  uniC9D8,  HANGUL SYLLABLE-C9D8
        '51673' => '2780'             , # G+2780  U+C9D9  uniC9D9,  HANGUL SYLLABLE-C9D9
        '51674' => '2781'             , # G+2781  U+C9DA  uniC9DA,  HANGUL SYLLABLE-C9DA
        '51675' => '15280'            , # G+15280  U+C9DB  uniC9DB,  HANGUL SYLLABLE-C9DB
        '51676' => '2782'             , # G+2782  U+C9DC  uniC9DC,  HANGUL SYLLABLE-C9DC
        '51677' => '2783'             , # G+2783  U+C9DD  uniC9DD,  HANGUL SYLLABLE-C9DD
        '51678' => '15281'            , # G+15281  U+C9DE  uniC9DE,  HANGUL SYLLABLE-C9DE
        '51679' => '15282'            , # G+15282  U+C9DF  uniC9DF,  HANGUL SYLLABLE-C9DF
        '51680' => '2784'             , # G+2784  U+C9E0  uniC9E0,  HANGUL SYLLABLE-C9E0
        '51681' => '15283'            , # G+15283  U+C9E1  uniC9E1,  HANGUL SYLLABLE-C9E1
        '51682' => '2785'             , # G+2785  U+C9E2  uniC9E2,  HANGUL SYLLABLE-C9E2
        '51683' => '15284'            , # G+15284  U+C9E3  uniC9E3,  HANGUL SYLLABLE-C9E3
        '51684' => '2786'             , # G+2786  U+C9E4  uniC9E4,  HANGUL SYLLABLE-C9E4
        '51685' => '15285'            , # G+15285  U+C9E5  uniC9E5,  HANGUL SYLLABLE-C9E5
        '51686' => '15286'            , # G+15286  U+C9E6  uniC9E6,  HANGUL SYLLABLE-C9E6
        '51687' => '2787'             , # G+2787  U+C9E7  uniC9E7,  HANGUL SYLLABLE-C9E7
        '51688' => '15287'            , # G+15287  U+C9E8  uniC9E8,  HANGUL SYLLABLE-C9E8
        '51689' => '15288'            , # G+15288  U+C9E9  uniC9E9,  HANGUL SYLLABLE-C9E9
        '51690' => '15289'            , # G+15289  U+C9EA  uniC9EA,  HANGUL SYLLABLE-C9EA
        '51691' => '15290'            , # G+15290  U+C9EB  uniC9EB,  HANGUL SYLLABLE-C9EB
        '51692' => '2788'             , # G+2788  U+C9EC  uniC9EC,  HANGUL SYLLABLE-C9EC
        '51693' => '2789'             , # G+2789  U+C9ED  uniC9ED,  HANGUL SYLLABLE-C9ED
        '51694' => '15291'            , # G+15291  U+C9EE  uniC9EE,  HANGUL SYLLABLE-C9EE
        '51695' => '2790'             , # G+2790  U+C9EF  uniC9EF,  HANGUL SYLLABLE-C9EF
        '51696' => '2791'             , # G+2791  U+C9F0  uniC9F0,  HANGUL SYLLABLE-C9F0
        '51697' => '2792'             , # G+2792  U+C9F1  uniC9F1,  HANGUL SYLLABLE-C9F1
        '51698' => '15292'            , # G+15292  U+C9F2  uniC9F2,  HANGUL SYLLABLE-C9F2
        '51699' => '15293'            , # G+15293  U+C9F3  uniC9F3,  HANGUL SYLLABLE-C9F3
        '51700' => '15294'            , # G+15294  U+C9F4  uniC9F4,  HANGUL SYLLABLE-C9F4
        '51701' => '15295'            , # G+15295  U+C9F5  uniC9F5,  HANGUL SYLLABLE-C9F5
        '51702' => '15296'            , # G+15296  U+C9F6  uniC9F6,  HANGUL SYLLABLE-C9F6
        '51703' => '15297'            , # G+15297  U+C9F7  uniC9F7,  HANGUL SYLLABLE-C9F7
        '51704' => '2793'             , # G+2793  U+C9F8  uniC9F8,  HANGUL SYLLABLE-C9F8
        '51705' => '2794'             , # G+2794  U+C9F9  uniC9F9,  HANGUL SYLLABLE-C9F9
        '51706' => '15298'            , # G+15298  U+C9FA  uniC9FA,  HANGUL SYLLABLE-C9FA
        '51707' => '15299'            , # G+15299  U+C9FB  uniC9FB,  HANGUL SYLLABLE-C9FB
        '51708' => '2795'             , # G+2795  U+C9FC  uniC9FC,  HANGUL SYLLABLE-C9FC
        '51709' => '15300'            , # G+15300  U+C9FD  uniC9FD,  HANGUL SYLLABLE-C9FD
        '51710' => '15301'            , # G+15301  U+C9FE  uniC9FE,  HANGUL SYLLABLE-C9FE
        '51711' => '15302'            , # G+15302  U+C9FF  uniC9FF,  HANGUL SYLLABLE-C9FF
        '51712' => '2796'             , # G+2796  U+CA00  uniCA00,  HANGUL SYLLABLE-CA00
        '51713' => '15303'            , # G+15303  U+CA01  uniCA01,  HANGUL SYLLABLE-CA01
        '51714' => '15304'            , # G+15304  U+CA02  uniCA02,  HANGUL SYLLABLE-CA02
        '51715' => '15305'            , # G+15305  U+CA03  uniCA03,  HANGUL SYLLABLE-CA03
        '51716' => '15306'            , # G+15306  U+CA04  uniCA04,  HANGUL SYLLABLE-CA04
        '51717' => '15307'            , # G+15307  U+CA05  uniCA05,  HANGUL SYLLABLE-CA05
        '51718' => '15308'            , # G+15308  U+CA06  uniCA06,  HANGUL SYLLABLE-CA06
        '51719' => '15309'            , # G+15309  U+CA07  uniCA07,  HANGUL SYLLABLE-CA07
        '51720' => '2797'             , # G+2797  U+CA08  uniCA08,  HANGUL SYLLABLE-CA08
        '51721' => '2798'             , # G+2798  U+CA09  uniCA09,  HANGUL SYLLABLE-CA09
        '51722' => '15310'            , # G+15310  U+CA0A  uniCA0A,  HANGUL SYLLABLE-CA0A
        '51723' => '2799'             , # G+2799  U+CA0B  uniCA0B,  HANGUL SYLLABLE-CA0B
        '51724' => '2800'             , # G+2800  U+CA0C  uniCA0C,  HANGUL SYLLABLE-CA0C
        '51725' => '2801'             , # G+2801  U+CA0D  uniCA0D,  HANGUL SYLLABLE-CA0D
        '51726' => '15311'            , # G+15311  U+CA0E  uniCA0E,  HANGUL SYLLABLE-CA0E
        '51727' => '15312'            , # G+15312  U+CA0F  uniCA0F,  HANGUL SYLLABLE-CA0F
        '51728' => '15313'            , # G+15313  U+CA10  uniCA10,  HANGUL SYLLABLE-CA10
        '51729' => '15314'            , # G+15314  U+CA11  uniCA11,  HANGUL SYLLABLE-CA11
        '51730' => '15315'            , # G+15315  U+CA12  uniCA12,  HANGUL SYLLABLE-CA12
        '51731' => '15316'            , # G+15316  U+CA13  uniCA13,  HANGUL SYLLABLE-CA13
        '51732' => '2802'             , # G+2802  U+CA14  uniCA14,  HANGUL SYLLABLE-CA14
        '51733' => '15317'            , # G+15317  U+CA15  uniCA15,  HANGUL SYLLABLE-CA15
        '51734' => '15318'            , # G+15318  U+CA16  uniCA16,  HANGUL SYLLABLE-CA16
        '51735' => '15319'            , # G+15319  U+CA17  uniCA17,  HANGUL SYLLABLE-CA17
        '51736' => '2803'             , # G+2803  U+CA18  uniCA18,  HANGUL SYLLABLE-CA18
        '51737' => '15320'            , # G+15320  U+CA19  uniCA19,  HANGUL SYLLABLE-CA19
        '51738' => '15321'            , # G+15321  U+CA1A  uniCA1A,  HANGUL SYLLABLE-CA1A
        '51739' => '15322'            , # G+15322  U+CA1B  uniCA1B,  HANGUL SYLLABLE-CA1B
        '51740' => '15323'            , # G+15323  U+CA1C  uniCA1C,  HANGUL SYLLABLE-CA1C
        '51741' => '15324'            , # G+15324  U+CA1D  uniCA1D,  HANGUL SYLLABLE-CA1D
        '51742' => '15325'            , # G+15325  U+CA1E  uniCA1E,  HANGUL SYLLABLE-CA1E
        '51743' => '15326'            , # G+15326  U+CA1F  uniCA1F,  HANGUL SYLLABLE-CA1F
        '51744' => '15327'            , # G+15327  U+CA20  uniCA20,  HANGUL SYLLABLE-CA20
        '51745' => '15328'            , # G+15328  U+CA21  uniCA21,  HANGUL SYLLABLE-CA21
        '51746' => '15329'            , # G+15329  U+CA22  uniCA22,  HANGUL SYLLABLE-CA22
        '51747' => '15330'            , # G+15330  U+CA23  uniCA23,  HANGUL SYLLABLE-CA23
        '51748' => '15331'            , # G+15331  U+CA24  uniCA24,  HANGUL SYLLABLE-CA24
        '51749' => '15332'            , # G+15332  U+CA25  uniCA25,  HANGUL SYLLABLE-CA25
        '51750' => '15333'            , # G+15333  U+CA26  uniCA26,  HANGUL SYLLABLE-CA26
        '51751' => '15334'            , # G+15334  U+CA27  uniCA27,  HANGUL SYLLABLE-CA27
        '51752' => '15335'            , # G+15335  U+CA28  uniCA28,  HANGUL SYLLABLE-CA28
        '51753' => '2804'             , # G+2804  U+CA29  uniCA29,  HANGUL SYLLABLE-CA29
        '51754' => '15336'            , # G+15336  U+CA2A  uniCA2A,  HANGUL SYLLABLE-CA2A
        '51755' => '15337'            , # G+15337  U+CA2B  uniCA2B,  HANGUL SYLLABLE-CA2B
        '51756' => '15338'            , # G+15338  U+CA2C  uniCA2C,  HANGUL SYLLABLE-CA2C
        '51757' => '15339'            , # G+15339  U+CA2D  uniCA2D,  HANGUL SYLLABLE-CA2D
        '51758' => '15340'            , # G+15340  U+CA2E  uniCA2E,  HANGUL SYLLABLE-CA2E
        '51759' => '15341'            , # G+15341  U+CA2F  uniCA2F,  HANGUL SYLLABLE-CA2F
        '51760' => '15342'            , # G+15342  U+CA30  uniCA30,  HANGUL SYLLABLE-CA30
        '51761' => '15343'            , # G+15343  U+CA31  uniCA31,  HANGUL SYLLABLE-CA31
        '51762' => '15344'            , # G+15344  U+CA32  uniCA32,  HANGUL SYLLABLE-CA32
        '51763' => '15345'            , # G+15345  U+CA33  uniCA33,  HANGUL SYLLABLE-CA33
        '51764' => '15346'            , # G+15346  U+CA34  uniCA34,  HANGUL SYLLABLE-CA34
        '51765' => '15347'            , # G+15347  U+CA35  uniCA35,  HANGUL SYLLABLE-CA35
        '51766' => '15348'            , # G+15348  U+CA36  uniCA36,  HANGUL SYLLABLE-CA36
        '51767' => '15349'            , # G+15349  U+CA37  uniCA37,  HANGUL SYLLABLE-CA37
        '51768' => '15350'            , # G+15350  U+CA38  uniCA38,  HANGUL SYLLABLE-CA38
        '51769' => '15351'            , # G+15351  U+CA39  uniCA39,  HANGUL SYLLABLE-CA39
        '51770' => '15352'            , # G+15352  U+CA3A  uniCA3A,  HANGUL SYLLABLE-CA3A
        '51771' => '15353'            , # G+15353  U+CA3B  uniCA3B,  HANGUL SYLLABLE-CA3B
        '51772' => '15354'            , # G+15354  U+CA3C  uniCA3C,  HANGUL SYLLABLE-CA3C
        '51773' => '15355'            , # G+15355  U+CA3D  uniCA3D,  HANGUL SYLLABLE-CA3D
        '51774' => '15356'            , # G+15356  U+CA3E  uniCA3E,  HANGUL SYLLABLE-CA3E
        '51775' => '15357'            , # G+15357  U+CA3F  uniCA3F,  HANGUL SYLLABLE-CA3F
        '51776' => '15358'            , # G+15358  U+CA40  uniCA40,  HANGUL SYLLABLE-CA40
        '51777' => '15359'            , # G+15359  U+CA41  uniCA41,  HANGUL SYLLABLE-CA41
        '51778' => '15360'            , # G+15360  U+CA42  uniCA42,  HANGUL SYLLABLE-CA42
        '51779' => '15361'            , # G+15361  U+CA43  uniCA43,  HANGUL SYLLABLE-CA43
        '51780' => '15362'            , # G+15362  U+CA44  uniCA44,  HANGUL SYLLABLE-CA44
        '51781' => '15363'            , # G+15363  U+CA45  uniCA45,  HANGUL SYLLABLE-CA45
        '51782' => '15364'            , # G+15364  U+CA46  uniCA46,  HANGUL SYLLABLE-CA46
        '51783' => '15365'            , # G+15365  U+CA47  uniCA47,  HANGUL SYLLABLE-CA47
        '51784' => '15366'            , # G+15366  U+CA48  uniCA48,  HANGUL SYLLABLE-CA48
        '51785' => '15367'            , # G+15367  U+CA49  uniCA49,  HANGUL SYLLABLE-CA49
        '51786' => '15368'            , # G+15368  U+CA4A  uniCA4A,  HANGUL SYLLABLE-CA4A
        '51787' => '15369'            , # G+15369  U+CA4B  uniCA4B,  HANGUL SYLLABLE-CA4B
        '51788' => '2805'             , # G+2805  U+CA4C  uniCA4C,  HANGUL SYLLABLE-CA4C
        '51789' => '2806'             , # G+2806  U+CA4D  uniCA4D,  HANGUL SYLLABLE-CA4D
        '51790' => '15370'            , # G+15370  U+CA4E  uniCA4E,  HANGUL SYLLABLE-CA4E
        '51791' => '15371'            , # G+15371  U+CA4F  uniCA4F,  HANGUL SYLLABLE-CA4F
        '51792' => '2807'             , # G+2807  U+CA50  uniCA50,  HANGUL SYLLABLE-CA50
        '51793' => '15372'            , # G+15372  U+CA51  uniCA51,  HANGUL SYLLABLE-CA51
        '51794' => '15373'            , # G+15373  U+CA52  uniCA52,  HANGUL SYLLABLE-CA52
        '51795' => '15374'            , # G+15374  U+CA53  uniCA53,  HANGUL SYLLABLE-CA53
        '51796' => '2808'             , # G+2808  U+CA54  uniCA54,  HANGUL SYLLABLE-CA54
        '51797' => '15375'            , # G+15375  U+CA55  uniCA55,  HANGUL SYLLABLE-CA55
        '51798' => '15376'            , # G+15376  U+CA56  uniCA56,  HANGUL SYLLABLE-CA56
        '51799' => '15377'            , # G+15377  U+CA57  uniCA57,  HANGUL SYLLABLE-CA57
        '51800' => '15378'            , # G+15378  U+CA58  uniCA58,  HANGUL SYLLABLE-CA58
        '51801' => '15379'            , # G+15379  U+CA59  uniCA59,  HANGUL SYLLABLE-CA59
        '51802' => '15380'            , # G+15380  U+CA5A  uniCA5A,  HANGUL SYLLABLE-CA5A
        '51803' => '15381'            , # G+15381  U+CA5B  uniCA5B,  HANGUL SYLLABLE-CA5B
        '51804' => '2809'             , # G+2809  U+CA5C  uniCA5C,  HANGUL SYLLABLE-CA5C
        '51805' => '2810'             , # G+2810  U+CA5D  uniCA5D,  HANGUL SYLLABLE-CA5D
        '51806' => '15382'            , # G+15382  U+CA5E  uniCA5E,  HANGUL SYLLABLE-CA5E
        '51807' => '2811'             , # G+2811  U+CA5F  uniCA5F,  HANGUL SYLLABLE-CA5F
        '51808' => '2812'             , # G+2812  U+CA60  uniCA60,  HANGUL SYLLABLE-CA60
        '51809' => '2813'             , # G+2813  U+CA61  uniCA61,  HANGUL SYLLABLE-CA61
        '51810' => '15383'            , # G+15383  U+CA62  uniCA62,  HANGUL SYLLABLE-CA62
        '51811' => '15384'            , # G+15384  U+CA63  uniCA63,  HANGUL SYLLABLE-CA63
        '51812' => '15385'            , # G+15385  U+CA64  uniCA64,  HANGUL SYLLABLE-CA64
        '51813' => '15386'            , # G+15386  U+CA65  uniCA65,  HANGUL SYLLABLE-CA65
        '51814' => '15387'            , # G+15387  U+CA66  uniCA66,  HANGUL SYLLABLE-CA66
        '51815' => '15388'            , # G+15388  U+CA67  uniCA67,  HANGUL SYLLABLE-CA67
        '51816' => '2814'             , # G+2814  U+CA68  uniCA68,  HANGUL SYLLABLE-CA68
        '51817' => '15389'            , # G+15389  U+CA69  uniCA69,  HANGUL SYLLABLE-CA69
        '51818' => '15390'            , # G+15390  U+CA6A  uniCA6A,  HANGUL SYLLABLE-CA6A
        '51819' => '15391'            , # G+15391  U+CA6B  uniCA6B,  HANGUL SYLLABLE-CA6B
        '51820' => '15392'            , # G+15392  U+CA6C  uniCA6C,  HANGUL SYLLABLE-CA6C
        '51821' => '15393'            , # G+15393  U+CA6D  uniCA6D,  HANGUL SYLLABLE-CA6D
        '51822' => '15394'            , # G+15394  U+CA6E  uniCA6E,  HANGUL SYLLABLE-CA6E
        '51823' => '15395'            , # G+15395  U+CA6F  uniCA6F,  HANGUL SYLLABLE-CA6F
        '51824' => '15396'            , # G+15396  U+CA70  uniCA70,  HANGUL SYLLABLE-CA70
        '51825' => '15397'            , # G+15397  U+CA71  uniCA71,  HANGUL SYLLABLE-CA71
        '51826' => '15398'            , # G+15398  U+CA72  uniCA72,  HANGUL SYLLABLE-CA72
        '51827' => '15399'            , # G+15399  U+CA73  uniCA73,  HANGUL SYLLABLE-CA73
        '51828' => '15400'            , # G+15400  U+CA74  uniCA74,  HANGUL SYLLABLE-CA74
        '51829' => '15401'            , # G+15401  U+CA75  uniCA75,  HANGUL SYLLABLE-CA75
        '51830' => '15402'            , # G+15402  U+CA76  uniCA76,  HANGUL SYLLABLE-CA76
        '51831' => '15403'            , # G+15403  U+CA77  uniCA77,  HANGUL SYLLABLE-CA77
        '51832' => '15404'            , # G+15404  U+CA78  uniCA78,  HANGUL SYLLABLE-CA78
        '51833' => '15405'            , # G+15405  U+CA79  uniCA79,  HANGUL SYLLABLE-CA79
        '51834' => '15406'            , # G+15406  U+CA7A  uniCA7A,  HANGUL SYLLABLE-CA7A
        '51835' => '15407'            , # G+15407  U+CA7B  uniCA7B,  HANGUL SYLLABLE-CA7B
        '51836' => '15408'            , # G+15408  U+CA7C  uniCA7C,  HANGUL SYLLABLE-CA7C
        '51837' => '2815'             , # G+2815  U+CA7D  uniCA7D,  HANGUL SYLLABLE-CA7D
        '51838' => '15409'            , # G+15409  U+CA7E  uniCA7E,  HANGUL SYLLABLE-CA7E
        '51839' => '15410'            , # G+15410  U+CA7F  uniCA7F,  HANGUL SYLLABLE-CA7F
        '51840' => '15411'            , # G+15411  U+CA80  uniCA80,  HANGUL SYLLABLE-CA80
        '51841' => '15412'            , # G+15412  U+CA81  uniCA81,  HANGUL SYLLABLE-CA81
        '51842' => '15413'            , # G+15413  U+CA82  uniCA82,  HANGUL SYLLABLE-CA82
        '51843' => '15414'            , # G+15414  U+CA83  uniCA83,  HANGUL SYLLABLE-CA83
        '51844' => '2816'             , # G+2816  U+CA84  uniCA84,  HANGUL SYLLABLE-CA84
        '51845' => '15415'            , # G+15415  U+CA85  uniCA85,  HANGUL SYLLABLE-CA85
        '51846' => '15416'            , # G+15416  U+CA86  uniCA86,  HANGUL SYLLABLE-CA86
        '51847' => '15417'            , # G+15417  U+CA87  uniCA87,  HANGUL SYLLABLE-CA87
        '51848' => '15418'            , # G+15418  U+CA88  uniCA88,  HANGUL SYLLABLE-CA88
        '51849' => '15419'            , # G+15419  U+CA89  uniCA89,  HANGUL SYLLABLE-CA89
        '51850' => '15420'            , # G+15420  U+CA8A  uniCA8A,  HANGUL SYLLABLE-CA8A
        '51851' => '15421'            , # G+15421  U+CA8B  uniCA8B,  HANGUL SYLLABLE-CA8B
        '51852' => '15422'            , # G+15422  U+CA8C  uniCA8C,  HANGUL SYLLABLE-CA8C
        '51853' => '15423'            , # G+15423  U+CA8D  uniCA8D,  HANGUL SYLLABLE-CA8D
        '51854' => '15424'            , # G+15424  U+CA8E  uniCA8E,  HANGUL SYLLABLE-CA8E
        '51855' => '15425'            , # G+15425  U+CA8F  uniCA8F,  HANGUL SYLLABLE-CA8F
        '51856' => '15426'            , # G+15426  U+CA90  uniCA90,  HANGUL SYLLABLE-CA90
        '51857' => '15427'            , # G+15427  U+CA91  uniCA91,  HANGUL SYLLABLE-CA91
        '51858' => '15428'            , # G+15428  U+CA92  uniCA92,  HANGUL SYLLABLE-CA92
        '51859' => '15429'            , # G+15429  U+CA93  uniCA93,  HANGUL SYLLABLE-CA93
        '51860' => '15430'            , # G+15430  U+CA94  uniCA94,  HANGUL SYLLABLE-CA94
        '51861' => '15431'            , # G+15431  U+CA95  uniCA95,  HANGUL SYLLABLE-CA95
        '51862' => '15432'            , # G+15432  U+CA96  uniCA96,  HANGUL SYLLABLE-CA96
        '51863' => '15433'            , # G+15433  U+CA97  uniCA97,  HANGUL SYLLABLE-CA97
        '51864' => '2817'             , # G+2817  U+CA98  uniCA98,  HANGUL SYLLABLE-CA98
        '51865' => '15434'            , # G+15434  U+CA99  uniCA99,  HANGUL SYLLABLE-CA99
        '51866' => '15435'            , # G+15435  U+CA9A  uniCA9A,  HANGUL SYLLABLE-CA9A
        '51867' => '15436'            , # G+15436  U+CA9B  uniCA9B,  HANGUL SYLLABLE-CA9B
        '51868' => '15437'            , # G+15437  U+CA9C  uniCA9C,  HANGUL SYLLABLE-CA9C
        '51869' => '15438'            , # G+15438  U+CA9D  uniCA9D,  HANGUL SYLLABLE-CA9D
        '51870' => '15439'            , # G+15439  U+CA9E  uniCA9E,  HANGUL SYLLABLE-CA9E
        '51871' => '15440'            , # G+15440  U+CA9F  uniCA9F,  HANGUL SYLLABLE-CA9F
        '51872' => '15441'            , # G+15441  U+CAA0  uniCAA0,  HANGUL SYLLABLE-CAA0
        '51873' => '15442'            , # G+15442  U+CAA1  uniCAA1,  HANGUL SYLLABLE-CAA1
        '51874' => '15443'            , # G+15443  U+CAA2  uniCAA2,  HANGUL SYLLABLE-CAA2
        '51875' => '15444'            , # G+15444  U+CAA3  uniCAA3,  HANGUL SYLLABLE-CAA3
        '51876' => '15445'            , # G+15445  U+CAA4  uniCAA4,  HANGUL SYLLABLE-CAA4
        '51877' => '15446'            , # G+15446  U+CAA5  uniCAA5,  HANGUL SYLLABLE-CAA5
        '51878' => '15447'            , # G+15447  U+CAA6  uniCAA6,  HANGUL SYLLABLE-CAA6
        '51879' => '15448'            , # G+15448  U+CAA7  uniCAA7,  HANGUL SYLLABLE-CAA7
        '51880' => '15449'            , # G+15449  U+CAA8  uniCAA8,  HANGUL SYLLABLE-CAA8
        '51881' => '15450'            , # G+15450  U+CAA9  uniCAA9,  HANGUL SYLLABLE-CAA9
        '51882' => '15451'            , # G+15451  U+CAAA  uniCAAA,  HANGUL SYLLABLE-CAAA
        '51883' => '15452'            , # G+15452  U+CAAB  uniCAAB,  HANGUL SYLLABLE-CAAB
        '51884' => '15453'            , # G+15453  U+CAAC  uniCAAC,  HANGUL SYLLABLE-CAAC
        '51885' => '15454'            , # G+15454  U+CAAD  uniCAAD,  HANGUL SYLLABLE-CAAD
        '51886' => '15455'            , # G+15455  U+CAAE  uniCAAE,  HANGUL SYLLABLE-CAAE
        '51887' => '15456'            , # G+15456  U+CAAF  uniCAAF,  HANGUL SYLLABLE-CAAF
        '51888' => '15457'            , # G+15457  U+CAB0  uniCAB0,  HANGUL SYLLABLE-CAB0
        '51889' => '15458'            , # G+15458  U+CAB1  uniCAB1,  HANGUL SYLLABLE-CAB1
        '51890' => '15459'            , # G+15459  U+CAB2  uniCAB2,  HANGUL SYLLABLE-CAB2
        '51891' => '15460'            , # G+15460  U+CAB3  uniCAB3,  HANGUL SYLLABLE-CAB3
        '51892' => '15461'            , # G+15461  U+CAB4  uniCAB4,  HANGUL SYLLABLE-CAB4
        '51893' => '15462'            , # G+15462  U+CAB5  uniCAB5,  HANGUL SYLLABLE-CAB5
        '51894' => '15463'            , # G+15463  U+CAB6  uniCAB6,  HANGUL SYLLABLE-CAB6
        '51895' => '15464'            , # G+15464  U+CAB7  uniCAB7,  HANGUL SYLLABLE-CAB7
        '51896' => '15465'            , # G+15465  U+CAB8  uniCAB8,  HANGUL SYLLABLE-CAB8
        '51897' => '15466'            , # G+15466  U+CAB9  uniCAB9,  HANGUL SYLLABLE-CAB9
        '51898' => '15467'            , # G+15467  U+CABA  uniCABA,  HANGUL SYLLABLE-CABA
        '51899' => '15468'            , # G+15468  U+CABB  uniCABB,  HANGUL SYLLABLE-CABB
        '51900' => '2818'             , # G+2818  U+CABC  uniCABC,  HANGUL SYLLABLE-CABC
        '51901' => '2819'             , # G+2819  U+CABD  uniCABD,  HANGUL SYLLABLE-CABD
        '51902' => '15469'            , # G+15469  U+CABE  uniCABE,  HANGUL SYLLABLE-CABE
        '51903' => '15470'            , # G+15470  U+CABF  uniCABF,  HANGUL SYLLABLE-CABF
        '51904' => '2820'             , # G+2820  U+CAC0  uniCAC0,  HANGUL SYLLABLE-CAC0
        '51905' => '15471'            , # G+15471  U+CAC1  uniCAC1,  HANGUL SYLLABLE-CAC1
        '51906' => '15472'            , # G+15472  U+CAC2  uniCAC2,  HANGUL SYLLABLE-CAC2
        '51907' => '15473'            , # G+15473  U+CAC3  uniCAC3,  HANGUL SYLLABLE-CAC3
        '51908' => '2821'             , # G+2821  U+CAC4  uniCAC4,  HANGUL SYLLABLE-CAC4
        '51909' => '15474'            , # G+15474  U+CAC5  uniCAC5,  HANGUL SYLLABLE-CAC5
        '51910' => '15475'            , # G+15475  U+CAC6  uniCAC6,  HANGUL SYLLABLE-CAC6
        '51911' => '15476'            , # G+15476  U+CAC7  uniCAC7,  HANGUL SYLLABLE-CAC7
        '51912' => '15477'            , # G+15477  U+CAC8  uniCAC8,  HANGUL SYLLABLE-CAC8
        '51913' => '15478'            , # G+15478  U+CAC9  uniCAC9,  HANGUL SYLLABLE-CAC9
        '51914' => '15479'            , # G+15479  U+CACA  uniCACA,  HANGUL SYLLABLE-CACA
        '51915' => '15480'            , # G+15480  U+CACB  uniCACB,  HANGUL SYLLABLE-CACB
        '51916' => '2822'             , # G+2822  U+CACC  uniCACC,  HANGUL SYLLABLE-CACC
        '51917' => '2823'             , # G+2823  U+CACD  uniCACD,  HANGUL SYLLABLE-CACD
        '51918' => '15481'            , # G+15481  U+CACE  uniCACE,  HANGUL SYLLABLE-CACE
        '51919' => '2824'             , # G+2824  U+CACF  uniCACF,  HANGUL SYLLABLE-CACF
        '51920' => '15482'            , # G+15482  U+CAD0  uniCAD0,  HANGUL SYLLABLE-CAD0
        '51921' => '2825'             , # G+2825  U+CAD1  uniCAD1,  HANGUL SYLLABLE-CAD1
        '51922' => '15483'            , # G+15483  U+CAD2  uniCAD2,  HANGUL SYLLABLE-CAD2
        '51923' => '2826'             , # G+2826  U+CAD3  uniCAD3,  HANGUL SYLLABLE-CAD3
        '51924' => '15484'            , # G+15484  U+CAD4  uniCAD4,  HANGUL SYLLABLE-CAD4
        '51925' => '15485'            , # G+15485  U+CAD5  uniCAD5,  HANGUL SYLLABLE-CAD5
        '51926' => '15486'            , # G+15486  U+CAD6  uniCAD6,  HANGUL SYLLABLE-CAD6
        '51927' => '15487'            , # G+15487  U+CAD7  uniCAD7,  HANGUL SYLLABLE-CAD7
        '51928' => '2827'             , # G+2827  U+CAD8  uniCAD8,  HANGUL SYLLABLE-CAD8
        '51929' => '2828'             , # G+2828  U+CAD9  uniCAD9,  HANGUL SYLLABLE-CAD9
        '51930' => '15488'            , # G+15488  U+CADA  uniCADA,  HANGUL SYLLABLE-CADA
        '51931' => '15489'            , # G+15489  U+CADB  uniCADB,  HANGUL SYLLABLE-CADB
        '51932' => '15490'            , # G+15490  U+CADC  uniCADC,  HANGUL SYLLABLE-CADC
        '51933' => '15491'            , # G+15491  U+CADD  uniCADD,  HANGUL SYLLABLE-CADD
        '51934' => '15492'            , # G+15492  U+CADE  uniCADE,  HANGUL SYLLABLE-CADE
        '51935' => '15493'            , # G+15493  U+CADF  uniCADF,  HANGUL SYLLABLE-CADF
        '51936' => '2829'             , # G+2829  U+CAE0  uniCAE0,  HANGUL SYLLABLE-CAE0
        '51937' => '15494'            , # G+15494  U+CAE1  uniCAE1,  HANGUL SYLLABLE-CAE1
        '51938' => '15495'            , # G+15495  U+CAE2  uniCAE2,  HANGUL SYLLABLE-CAE2
        '51939' => '15496'            , # G+15496  U+CAE3  uniCAE3,  HANGUL SYLLABLE-CAE3
        '51940' => '15497'            , # G+15497  U+CAE4  uniCAE4,  HANGUL SYLLABLE-CAE4
        '51941' => '15498'            , # G+15498  U+CAE5  uniCAE5,  HANGUL SYLLABLE-CAE5
        '51942' => '15499'            , # G+15499  U+CAE6  uniCAE6,  HANGUL SYLLABLE-CAE6
        '51943' => '15500'            , # G+15500  U+CAE7  uniCAE7,  HANGUL SYLLABLE-CAE7
        '51944' => '15501'            , # G+15501  U+CAE8  uniCAE8,  HANGUL SYLLABLE-CAE8
        '51945' => '15502'            , # G+15502  U+CAE9  uniCAE9,  HANGUL SYLLABLE-CAE9
        '51946' => '15503'            , # G+15503  U+CAEA  uniCAEA,  HANGUL SYLLABLE-CAEA
        '51947' => '15504'            , # G+15504  U+CAEB  uniCAEB,  HANGUL SYLLABLE-CAEB
        '51948' => '2830'             , # G+2830  U+CAEC  uniCAEC,  HANGUL SYLLABLE-CAEC
        '51949' => '15505'            , # G+15505  U+CAED  uniCAED,  HANGUL SYLLABLE-CAED
        '51950' => '15506'            , # G+15506  U+CAEE  uniCAEE,  HANGUL SYLLABLE-CAEE
        '51951' => '15507'            , # G+15507  U+CAEF  uniCAEF,  HANGUL SYLLABLE-CAEF
        '51952' => '15508'            , # G+15508  U+CAF0  uniCAF0,  HANGUL SYLLABLE-CAF0
        '51953' => '15509'            , # G+15509  U+CAF1  uniCAF1,  HANGUL SYLLABLE-CAF1
        '51954' => '15510'            , # G+15510  U+CAF2  uniCAF2,  HANGUL SYLLABLE-CAF2
        '51955' => '15511'            , # G+15511  U+CAF3  uniCAF3,  HANGUL SYLLABLE-CAF3
        '51956' => '2831'             , # G+2831  U+CAF4  uniCAF4,  HANGUL SYLLABLE-CAF4
        '51957' => '15512'            , # G+15512  U+CAF5  uniCAF5,  HANGUL SYLLABLE-CAF5
        '51958' => '15513'            , # G+15513  U+CAF6  uniCAF6,  HANGUL SYLLABLE-CAF6
        '51959' => '15514'            , # G+15514  U+CAF7  uniCAF7,  HANGUL SYLLABLE-CAF7
        '51960' => '15515'            , # G+15515  U+CAF8  uniCAF8,  HANGUL SYLLABLE-CAF8
        '51961' => '15516'            , # G+15516  U+CAF9  uniCAF9,  HANGUL SYLLABLE-CAF9
        '51962' => '15517'            , # G+15517  U+CAFA  uniCAFA,  HANGUL SYLLABLE-CAFA
        '51963' => '15518'            , # G+15518  U+CAFB  uniCAFB,  HANGUL SYLLABLE-CAFB
        '51964' => '15519'            , # G+15519  U+CAFC  uniCAFC,  HANGUL SYLLABLE-CAFC
        '51965' => '15520'            , # G+15520  U+CAFD  uniCAFD,  HANGUL SYLLABLE-CAFD
        '51966' => '15521'            , # G+15521  U+CAFE  uniCAFE,  HANGUL SYLLABLE-CAFE
        '51967' => '15522'            , # G+15522  U+CAFF  uniCAFF,  HANGUL SYLLABLE-CAFF
        '51968' => '15523'            , # G+15523  U+CB00  uniCB00,  HANGUL SYLLABLE-CB00
        '51969' => '15524'            , # G+15524  U+CB01  uniCB01,  HANGUL SYLLABLE-CB01
        '51970' => '15525'            , # G+15525  U+CB02  uniCB02,  HANGUL SYLLABLE-CB02
        '51971' => '15526'            , # G+15526  U+CB03  uniCB03,  HANGUL SYLLABLE-CB03
        '51972' => '15527'            , # G+15527  U+CB04  uniCB04,  HANGUL SYLLABLE-CB04
        '51973' => '15528'            , # G+15528  U+CB05  uniCB05,  HANGUL SYLLABLE-CB05
        '51974' => '15529'            , # G+15529  U+CB06  uniCB06,  HANGUL SYLLABLE-CB06
        '51975' => '15530'            , # G+15530  U+CB07  uniCB07,  HANGUL SYLLABLE-CB07
        '51976' => '2832'             , # G+2832  U+CB08  uniCB08,  HANGUL SYLLABLE-CB08
        '51977' => '15531'            , # G+15531  U+CB09  uniCB09,  HANGUL SYLLABLE-CB09
        '51978' => '15532'            , # G+15532  U+CB0A  uniCB0A,  HANGUL SYLLABLE-CB0A
        '51979' => '15533'            , # G+15533  U+CB0B  uniCB0B,  HANGUL SYLLABLE-CB0B
        '51980' => '15534'            , # G+15534  U+CB0C  uniCB0C,  HANGUL SYLLABLE-CB0C
        '51981' => '15535'            , # G+15535  U+CB0D  uniCB0D,  HANGUL SYLLABLE-CB0D
        '51982' => '15536'            , # G+15536  U+CB0E  uniCB0E,  HANGUL SYLLABLE-CB0E
        '51983' => '15537'            , # G+15537  U+CB0F  uniCB0F,  HANGUL SYLLABLE-CB0F
        '51984' => '2833'             , # G+2833  U+CB10  uniCB10,  HANGUL SYLLABLE-CB10
        '51985' => '15538'            , # G+15538  U+CB11  uniCB11,  HANGUL SYLLABLE-CB11
        '51986' => '15539'            , # G+15539  U+CB12  uniCB12,  HANGUL SYLLABLE-CB12
        '51987' => '15540'            , # G+15540  U+CB13  uniCB13,  HANGUL SYLLABLE-CB13
        '51988' => '2834'             , # G+2834  U+CB14  uniCB14,  HANGUL SYLLABLE-CB14
        '51989' => '15541'            , # G+15541  U+CB15  uniCB15,  HANGUL SYLLABLE-CB15
        '51990' => '15542'            , # G+15542  U+CB16  uniCB16,  HANGUL SYLLABLE-CB16
        '51991' => '15543'            , # G+15543  U+CB17  uniCB17,  HANGUL SYLLABLE-CB17
        '51992' => '2835'             , # G+2835  U+CB18  uniCB18,  HANGUL SYLLABLE-CB18
        '51993' => '15544'            , # G+15544  U+CB19  uniCB19,  HANGUL SYLLABLE-CB19
        '51994' => '15545'            , # G+15545  U+CB1A  uniCB1A,  HANGUL SYLLABLE-CB1A
        '51995' => '15546'            , # G+15546  U+CB1B  uniCB1B,  HANGUL SYLLABLE-CB1B
        '51996' => '15547'            , # G+15547  U+CB1C  uniCB1C,  HANGUL SYLLABLE-CB1C
        '51997' => '15548'            , # G+15548  U+CB1D  uniCB1D,  HANGUL SYLLABLE-CB1D
        '51998' => '15549'            , # G+15549  U+CB1E  uniCB1E,  HANGUL SYLLABLE-CB1E
        '51999' => '15550'            , # G+15550  U+CB1F  uniCB1F,  HANGUL SYLLABLE-CB1F
        '52000' => '2836'             , # G+2836  U+CB20  uniCB20,  HANGUL SYLLABLE-CB20
        '52001' => '2837'             , # G+2837  U+CB21  uniCB21,  HANGUL SYLLABLE-CB21
        '52002' => '15551'            , # G+15551  U+CB22  uniCB22,  HANGUL SYLLABLE-CB22
        '52003' => '15552'            , # G+15552  U+CB23  uniCB23,  HANGUL SYLLABLE-CB23
        '52004' => '15553'            , # G+15553  U+CB24  uniCB24,  HANGUL SYLLABLE-CB24
        '52005' => '15554'            , # G+15554  U+CB25  uniCB25,  HANGUL SYLLABLE-CB25
        '52006' => '15555'            , # G+15555  U+CB26  uniCB26,  HANGUL SYLLABLE-CB26
        '52007' => '15556'            , # G+15556  U+CB27  uniCB27,  HANGUL SYLLABLE-CB27
        '52008' => '15557'            , # G+15557  U+CB28  uniCB28,  HANGUL SYLLABLE-CB28
        '52009' => '15558'            , # G+15558  U+CB29  uniCB29,  HANGUL SYLLABLE-CB29
        '52010' => '15559'            , # G+15559  U+CB2A  uniCB2A,  HANGUL SYLLABLE-CB2A
        '52011' => '15560'            , # G+15560  U+CB2B  uniCB2B,  HANGUL SYLLABLE-CB2B
        '52012' => '15561'            , # G+15561  U+CB2C  uniCB2C,  HANGUL SYLLABLE-CB2C
        '52013' => '15562'            , # G+15562  U+CB2D  uniCB2D,  HANGUL SYLLABLE-CB2D
        '52014' => '15563'            , # G+15563  U+CB2E  uniCB2E,  HANGUL SYLLABLE-CB2E
        '52015' => '15564'            , # G+15564  U+CB2F  uniCB2F,  HANGUL SYLLABLE-CB2F
        '52016' => '15565'            , # G+15565  U+CB30  uniCB30,  HANGUL SYLLABLE-CB30
        '52017' => '15566'            , # G+15566  U+CB31  uniCB31,  HANGUL SYLLABLE-CB31
        '52018' => '15567'            , # G+15567  U+CB32  uniCB32,  HANGUL SYLLABLE-CB32
        '52019' => '15568'            , # G+15568  U+CB33  uniCB33,  HANGUL SYLLABLE-CB33
        '52020' => '15569'            , # G+15569  U+CB34  uniCB34,  HANGUL SYLLABLE-CB34
        '52021' => '15570'            , # G+15570  U+CB35  uniCB35,  HANGUL SYLLABLE-CB35
        '52022' => '15571'            , # G+15571  U+CB36  uniCB36,  HANGUL SYLLABLE-CB36
        '52023' => '15572'            , # G+15572  U+CB37  uniCB37,  HANGUL SYLLABLE-CB37
        '52024' => '15573'            , # G+15573  U+CB38  uniCB38,  HANGUL SYLLABLE-CB38
        '52025' => '15574'            , # G+15574  U+CB39  uniCB39,  HANGUL SYLLABLE-CB39
        '52026' => '15575'            , # G+15575  U+CB3A  uniCB3A,  HANGUL SYLLABLE-CB3A
        '52027' => '15576'            , # G+15576  U+CB3B  uniCB3B,  HANGUL SYLLABLE-CB3B
        '52028' => '15577'            , # G+15577  U+CB3C  uniCB3C,  HANGUL SYLLABLE-CB3C
        '52029' => '15578'            , # G+15578  U+CB3D  uniCB3D,  HANGUL SYLLABLE-CB3D
        '52030' => '15579'            , # G+15579  U+CB3E  uniCB3E,  HANGUL SYLLABLE-CB3E
        '52031' => '15580'            , # G+15580  U+CB3F  uniCB3F,  HANGUL SYLLABLE-CB3F
        '52032' => '15581'            , # G+15581  U+CB40  uniCB40,  HANGUL SYLLABLE-CB40
        '52033' => '2838'             , # G+2838  U+CB41  uniCB41,  HANGUL SYLLABLE-CB41
        '52034' => '15582'            , # G+15582  U+CB42  uniCB42,  HANGUL SYLLABLE-CB42
        '52035' => '15583'            , # G+15583  U+CB43  uniCB43,  HANGUL SYLLABLE-CB43
        '52036' => '15584'            , # G+15584  U+CB44  uniCB44,  HANGUL SYLLABLE-CB44
        '52037' => '15585'            , # G+15585  U+CB45  uniCB45,  HANGUL SYLLABLE-CB45
        '52038' => '15586'            , # G+15586  U+CB46  uniCB46,  HANGUL SYLLABLE-CB46
        '52039' => '15587'            , # G+15587  U+CB47  uniCB47,  HANGUL SYLLABLE-CB47
        '52040' => '2839'             , # G+2839  U+CB48  uniCB48,  HANGUL SYLLABLE-CB48
        '52041' => '2840'             , # G+2840  U+CB49  uniCB49,  HANGUL SYLLABLE-CB49
        '52042' => '15588'            , # G+15588  U+CB4A  uniCB4A,  HANGUL SYLLABLE-CB4A
        '52043' => '15589'            , # G+15589  U+CB4B  uniCB4B,  HANGUL SYLLABLE-CB4B
        '52044' => '2841'             , # G+2841  U+CB4C  uniCB4C,  HANGUL SYLLABLE-CB4C
        '52045' => '15590'            , # G+15590  U+CB4D  uniCB4D,  HANGUL SYLLABLE-CB4D
        '52046' => '15591'            , # G+15591  U+CB4E  uniCB4E,  HANGUL SYLLABLE-CB4E
        '52047' => '15592'            , # G+15592  U+CB4F  uniCB4F,  HANGUL SYLLABLE-CB4F
        '52048' => '2842'             , # G+2842  U+CB50  uniCB50,  HANGUL SYLLABLE-CB50
        '52049' => '15593'            , # G+15593  U+CB51  uniCB51,  HANGUL SYLLABLE-CB51
        '52050' => '15594'            , # G+15594  U+CB52  uniCB52,  HANGUL SYLLABLE-CB52
        '52051' => '15595'            , # G+15595  U+CB53  uniCB53,  HANGUL SYLLABLE-CB53
        '52052' => '15596'            , # G+15596  U+CB54  uniCB54,  HANGUL SYLLABLE-CB54
        '52053' => '15597'            , # G+15597  U+CB55  uniCB55,  HANGUL SYLLABLE-CB55
        '52054' => '15598'            , # G+15598  U+CB56  uniCB56,  HANGUL SYLLABLE-CB56
        '52055' => '15599'            , # G+15599  U+CB57  uniCB57,  HANGUL SYLLABLE-CB57
        '52056' => '2843'             , # G+2843  U+CB58  uniCB58,  HANGUL SYLLABLE-CB58
        '52057' => '2844'             , # G+2844  U+CB59  uniCB59,  HANGUL SYLLABLE-CB59
        '52058' => '15600'            , # G+15600  U+CB5A  uniCB5A,  HANGUL SYLLABLE-CB5A
        '52059' => '15601'            , # G+15601  U+CB5B  uniCB5B,  HANGUL SYLLABLE-CB5B
        '52060' => '15602'            , # G+15602  U+CB5C  uniCB5C,  HANGUL SYLLABLE-CB5C
        '52061' => '2845'             , # G+2845  U+CB5D  uniCB5D,  HANGUL SYLLABLE-CB5D
        '52062' => '15603'            , # G+15603  U+CB5E  uniCB5E,  HANGUL SYLLABLE-CB5E
        '52063' => '15604'            , # G+15604  U+CB5F  uniCB5F,  HANGUL SYLLABLE-CB5F
        '52064' => '15605'            , # G+15605  U+CB60  uniCB60,  HANGUL SYLLABLE-CB60
        '52065' => '15606'            , # G+15606  U+CB61  uniCB61,  HANGUL SYLLABLE-CB61
        '52066' => '15607'            , # G+15607  U+CB62  uniCB62,  HANGUL SYLLABLE-CB62
        '52067' => '15608'            , # G+15608  U+CB63  uniCB63,  HANGUL SYLLABLE-CB63
        '52068' => '2846'             , # G+2846  U+CB64  uniCB64,  HANGUL SYLLABLE-CB64
        '52069' => '15609'            , # G+15609  U+CB65  uniCB65,  HANGUL SYLLABLE-CB65
        '52070' => '15610'            , # G+15610  U+CB66  uniCB66,  HANGUL SYLLABLE-CB66
        '52071' => '15611'            , # G+15611  U+CB67  uniCB67,  HANGUL SYLLABLE-CB67
        '52072' => '15612'            , # G+15612  U+CB68  uniCB68,  HANGUL SYLLABLE-CB68
        '52073' => '15613'            , # G+15613  U+CB69  uniCB69,  HANGUL SYLLABLE-CB69
        '52074' => '15614'            , # G+15614  U+CB6A  uniCB6A,  HANGUL SYLLABLE-CB6A
        '52075' => '15615'            , # G+15615  U+CB6B  uniCB6B,  HANGUL SYLLABLE-CB6B
        '52076' => '15616'            , # G+15616  U+CB6C  uniCB6C,  HANGUL SYLLABLE-CB6C
        '52077' => '15617'            , # G+15617  U+CB6D  uniCB6D,  HANGUL SYLLABLE-CB6D
        '52078' => '15618'            , # G+15618  U+CB6E  uniCB6E,  HANGUL SYLLABLE-CB6E
        '52079' => '15619'            , # G+15619  U+CB6F  uniCB6F,  HANGUL SYLLABLE-CB6F
        '52080' => '15620'            , # G+15620  U+CB70  uniCB70,  HANGUL SYLLABLE-CB70
        '52081' => '15621'            , # G+15621  U+CB71  uniCB71,  HANGUL SYLLABLE-CB71
        '52082' => '15622'            , # G+15622  U+CB72  uniCB72,  HANGUL SYLLABLE-CB72
        '52083' => '15623'            , # G+15623  U+CB73  uniCB73,  HANGUL SYLLABLE-CB73
        '52084' => '15624'            , # G+15624  U+CB74  uniCB74,  HANGUL SYLLABLE-CB74
        '52085' => '15625'            , # G+15625  U+CB75  uniCB75,  HANGUL SYLLABLE-CB75
        '52086' => '15626'            , # G+15626  U+CB76  uniCB76,  HANGUL SYLLABLE-CB76
        '52087' => '15627'            , # G+15627  U+CB77  uniCB77,  HANGUL SYLLABLE-CB77
        '52088' => '2847'             , # G+2847  U+CB78  uniCB78,  HANGUL SYLLABLE-CB78
        '52089' => '2848'             , # G+2848  U+CB79  uniCB79,  HANGUL SYLLABLE-CB79
        '52090' => '15628'            , # G+15628  U+CB7A  uniCB7A,  HANGUL SYLLABLE-CB7A
        '52091' => '15629'            , # G+15629  U+CB7B  uniCB7B,  HANGUL SYLLABLE-CB7B
        '52092' => '15630'            , # G+15630  U+CB7C  uniCB7C,  HANGUL SYLLABLE-CB7C
        '52093' => '15631'            , # G+15631  U+CB7D  uniCB7D,  HANGUL SYLLABLE-CB7D
        '52094' => '15632'            , # G+15632  U+CB7E  uniCB7E,  HANGUL SYLLABLE-CB7E
        '52095' => '15633'            , # G+15633  U+CB7F  uniCB7F,  HANGUL SYLLABLE-CB7F
        '52096' => '15634'            , # G+15634  U+CB80  uniCB80,  HANGUL SYLLABLE-CB80
        '52097' => '15635'            , # G+15635  U+CB81  uniCB81,  HANGUL SYLLABLE-CB81
        '52098' => '15636'            , # G+15636  U+CB82  uniCB82,  HANGUL SYLLABLE-CB82
        '52099' => '15637'            , # G+15637  U+CB83  uniCB83,  HANGUL SYLLABLE-CB83
        '52100' => '15638'            , # G+15638  U+CB84  uniCB84,  HANGUL SYLLABLE-CB84
        '52101' => '15639'            , # G+15639  U+CB85  uniCB85,  HANGUL SYLLABLE-CB85
        '52102' => '15640'            , # G+15640  U+CB86  uniCB86,  HANGUL SYLLABLE-CB86
        '52103' => '15641'            , # G+15641  U+CB87  uniCB87,  HANGUL SYLLABLE-CB87
        '52104' => '15642'            , # G+15642  U+CB88  uniCB88,  HANGUL SYLLABLE-CB88
        '52105' => '15643'            , # G+15643  U+CB89  uniCB89,  HANGUL SYLLABLE-CB89
        '52106' => '15644'            , # G+15644  U+CB8A  uniCB8A,  HANGUL SYLLABLE-CB8A
        '52107' => '15645'            , # G+15645  U+CB8B  uniCB8B,  HANGUL SYLLABLE-CB8B
        '52108' => '15646'            , # G+15646  U+CB8C  uniCB8C,  HANGUL SYLLABLE-CB8C
        '52109' => '15647'            , # G+15647  U+CB8D  uniCB8D,  HANGUL SYLLABLE-CB8D
        '52110' => '15648'            , # G+15648  U+CB8E  uniCB8E,  HANGUL SYLLABLE-CB8E
        '52111' => '15649'            , # G+15649  U+CB8F  uniCB8F,  HANGUL SYLLABLE-CB8F
        '52112' => '15650'            , # G+15650  U+CB90  uniCB90,  HANGUL SYLLABLE-CB90
        '52113' => '15651'            , # G+15651  U+CB91  uniCB91,  HANGUL SYLLABLE-CB91
        '52114' => '15652'            , # G+15652  U+CB92  uniCB92,  HANGUL SYLLABLE-CB92
        '52115' => '15653'            , # G+15653  U+CB93  uniCB93,  HANGUL SYLLABLE-CB93
        '52116' => '15654'            , # G+15654  U+CB94  uniCB94,  HANGUL SYLLABLE-CB94
        '52117' => '15655'            , # G+15655  U+CB95  uniCB95,  HANGUL SYLLABLE-CB95
        '52118' => '15656'            , # G+15656  U+CB96  uniCB96,  HANGUL SYLLABLE-CB96
        '52119' => '15657'            , # G+15657  U+CB97  uniCB97,  HANGUL SYLLABLE-CB97
        '52120' => '15658'            , # G+15658  U+CB98  uniCB98,  HANGUL SYLLABLE-CB98
        '52121' => '15659'            , # G+15659  U+CB99  uniCB99,  HANGUL SYLLABLE-CB99
        '52122' => '15660'            , # G+15660  U+CB9A  uniCB9A,  HANGUL SYLLABLE-CB9A
        '52123' => '15661'            , # G+15661  U+CB9B  uniCB9B,  HANGUL SYLLABLE-CB9B
        '52124' => '2849'             , # G+2849  U+CB9C  uniCB9C,  HANGUL SYLLABLE-CB9C
        '52125' => '15662'            , # G+15662  U+CB9D  uniCB9D,  HANGUL SYLLABLE-CB9D
        '52126' => '15663'            , # G+15663  U+CB9E  uniCB9E,  HANGUL SYLLABLE-CB9E
        '52127' => '15664'            , # G+15664  U+CB9F  uniCB9F,  HANGUL SYLLABLE-CB9F
        '52128' => '15665'            , # G+15665  U+CBA0  uniCBA0,  HANGUL SYLLABLE-CBA0
        '52129' => '15666'            , # G+15666  U+CBA1  uniCBA1,  HANGUL SYLLABLE-CBA1
        '52130' => '15667'            , # G+15667  U+CBA2  uniCBA2,  HANGUL SYLLABLE-CBA2
        '52131' => '15668'            , # G+15668  U+CBA3  uniCBA3,  HANGUL SYLLABLE-CBA3
        '52132' => '15669'            , # G+15669  U+CBA4  uniCBA4,  HANGUL SYLLABLE-CBA4
        '52133' => '15670'            , # G+15670  U+CBA5  uniCBA5,  HANGUL SYLLABLE-CBA5
        '52134' => '15671'            , # G+15671  U+CBA6  uniCBA6,  HANGUL SYLLABLE-CBA6
        '52135' => '15672'            , # G+15672  U+CBA7  uniCBA7,  HANGUL SYLLABLE-CBA7
        '52136' => '15673'            , # G+15673  U+CBA8  uniCBA8,  HANGUL SYLLABLE-CBA8
        '52137' => '15674'            , # G+15674  U+CBA9  uniCBA9,  HANGUL SYLLABLE-CBA9
        '52138' => '15675'            , # G+15675  U+CBAA  uniCBAA,  HANGUL SYLLABLE-CBAA
        '52139' => '15676'            , # G+15676  U+CBAB  uniCBAB,  HANGUL SYLLABLE-CBAB
        '52140' => '15677'            , # G+15677  U+CBAC  uniCBAC,  HANGUL SYLLABLE-CBAC
        '52141' => '15678'            , # G+15678  U+CBAD  uniCBAD,  HANGUL SYLLABLE-CBAD
        '52142' => '15679'            , # G+15679  U+CBAE  uniCBAE,  HANGUL SYLLABLE-CBAE
        '52143' => '15680'            , # G+15680  U+CBAF  uniCBAF,  HANGUL SYLLABLE-CBAF
        '52144' => '15681'            , # G+15681  U+CBB0  uniCBB0,  HANGUL SYLLABLE-CBB0
        '52145' => '15682'            , # G+15682  U+CBB1  uniCBB1,  HANGUL SYLLABLE-CBB1
        '52146' => '15683'            , # G+15683  U+CBB2  uniCBB2,  HANGUL SYLLABLE-CBB2
        '52147' => '15684'            , # G+15684  U+CBB3  uniCBB3,  HANGUL SYLLABLE-CBB3
        '52148' => '15685'            , # G+15685  U+CBB4  uniCBB4,  HANGUL SYLLABLE-CBB4
        '52149' => '15686'            , # G+15686  U+CBB5  uniCBB5,  HANGUL SYLLABLE-CBB5
        '52150' => '15687'            , # G+15687  U+CBB6  uniCBB6,  HANGUL SYLLABLE-CBB6
        '52151' => '15688'            , # G+15688  U+CBB7  uniCBB7,  HANGUL SYLLABLE-CBB7
        '52152' => '2850'             , # G+2850  U+CBB8  uniCBB8,  HANGUL SYLLABLE-CBB8
        '52153' => '15689'            , # G+15689  U+CBB9  uniCBB9,  HANGUL SYLLABLE-CBB9
        '52154' => '15690'            , # G+15690  U+CBBA  uniCBBA,  HANGUL SYLLABLE-CBBA
        '52155' => '15691'            , # G+15691  U+CBBB  uniCBBB,  HANGUL SYLLABLE-CBBB
        '52156' => '15692'            , # G+15692  U+CBBC  uniCBBC,  HANGUL SYLLABLE-CBBC
        '52157' => '15693'            , # G+15693  U+CBBD  uniCBBD,  HANGUL SYLLABLE-CBBD
        '52158' => '15694'            , # G+15694  U+CBBE  uniCBBE,  HANGUL SYLLABLE-CBBE
        '52159' => '15695'            , # G+15695  U+CBBF  uniCBBF,  HANGUL SYLLABLE-CBBF
        '52160' => '15696'            , # G+15696  U+CBC0  uniCBC0,  HANGUL SYLLABLE-CBC0
        '52161' => '15697'            , # G+15697  U+CBC1  uniCBC1,  HANGUL SYLLABLE-CBC1
        '52162' => '15698'            , # G+15698  U+CBC2  uniCBC2,  HANGUL SYLLABLE-CBC2
        '52163' => '15699'            , # G+15699  U+CBC3  uniCBC3,  HANGUL SYLLABLE-CBC3
        '52164' => '15700'            , # G+15700  U+CBC4  uniCBC4,  HANGUL SYLLABLE-CBC4
        '52165' => '15701'            , # G+15701  U+CBC5  uniCBC5,  HANGUL SYLLABLE-CBC5
        '52166' => '15702'            , # G+15702  U+CBC6  uniCBC6,  HANGUL SYLLABLE-CBC6
        '52167' => '15703'            , # G+15703  U+CBC7  uniCBC7,  HANGUL SYLLABLE-CBC7
        '52168' => '15704'            , # G+15704  U+CBC8  uniCBC8,  HANGUL SYLLABLE-CBC8
        '52169' => '15705'            , # G+15705  U+CBC9  uniCBC9,  HANGUL SYLLABLE-CBC9
        '52170' => '15706'            , # G+15706  U+CBCA  uniCBCA,  HANGUL SYLLABLE-CBCA
        '52171' => '15707'            , # G+15707  U+CBCB  uniCBCB,  HANGUL SYLLABLE-CBCB
        '52172' => '15708'            , # G+15708  U+CBCC  uniCBCC,  HANGUL SYLLABLE-CBCC
        '52173' => '15709'            , # G+15709  U+CBCD  uniCBCD,  HANGUL SYLLABLE-CBCD
        '52174' => '15710'            , # G+15710  U+CBCE  uniCBCE,  HANGUL SYLLABLE-CBCE
        '52175' => '15711'            , # G+15711  U+CBCF  uniCBCF,  HANGUL SYLLABLE-CBCF
        '52176' => '15712'            , # G+15712  U+CBD0  uniCBD0,  HANGUL SYLLABLE-CBD0
        '52177' => '15713'            , # G+15713  U+CBD1  uniCBD1,  HANGUL SYLLABLE-CBD1
        '52178' => '15714'            , # G+15714  U+CBD2  uniCBD2,  HANGUL SYLLABLE-CBD2
        '52179' => '15715'            , # G+15715  U+CBD3  uniCBD3,  HANGUL SYLLABLE-CBD3
        '52180' => '2851'             , # G+2851  U+CBD4  uniCBD4,  HANGUL SYLLABLE-CBD4
        '52181' => '15716'            , # G+15716  U+CBD5  uniCBD5,  HANGUL SYLLABLE-CBD5
        '52182' => '15717'            , # G+15717  U+CBD6  uniCBD6,  HANGUL SYLLABLE-CBD6
        '52183' => '15718'            , # G+15718  U+CBD7  uniCBD7,  HANGUL SYLLABLE-CBD7
        '52184' => '15719'            , # G+15719  U+CBD8  uniCBD8,  HANGUL SYLLABLE-CBD8
        '52185' => '15720'            , # G+15720  U+CBD9  uniCBD9,  HANGUL SYLLABLE-CBD9
        '52186' => '15721'            , # G+15721  U+CBDA  uniCBDA,  HANGUL SYLLABLE-CBDA
        '52187' => '15722'            , # G+15722  U+CBDB  uniCBDB,  HANGUL SYLLABLE-CBDB
        '52188' => '15723'            , # G+15723  U+CBDC  uniCBDC,  HANGUL SYLLABLE-CBDC
        '52189' => '15724'            , # G+15724  U+CBDD  uniCBDD,  HANGUL SYLLABLE-CBDD
        '52190' => '15725'            , # G+15725  U+CBDE  uniCBDE,  HANGUL SYLLABLE-CBDE
        '52191' => '15726'            , # G+15726  U+CBDF  uniCBDF,  HANGUL SYLLABLE-CBDF
        '52192' => '15727'            , # G+15727  U+CBE0  uniCBE0,  HANGUL SYLLABLE-CBE0
        '52193' => '15728'            , # G+15728  U+CBE1  uniCBE1,  HANGUL SYLLABLE-CBE1
        '52194' => '15729'            , # G+15729  U+CBE2  uniCBE2,  HANGUL SYLLABLE-CBE2
        '52195' => '15730'            , # G+15730  U+CBE3  uniCBE3,  HANGUL SYLLABLE-CBE3
        '52196' => '2852'             , # G+2852  U+CBE4  uniCBE4,  HANGUL SYLLABLE-CBE4
        '52197' => '15731'            , # G+15731  U+CBE5  uniCBE5,  HANGUL SYLLABLE-CBE5
        '52198' => '15732'            , # G+15732  U+CBE6  uniCBE6,  HANGUL SYLLABLE-CBE6
        '52199' => '2853'             , # G+2853  U+CBE7  uniCBE7,  HANGUL SYLLABLE-CBE7
        '52200' => '15733'            , # G+15733  U+CBE8  uniCBE8,  HANGUL SYLLABLE-CBE8
        '52201' => '2854'             , # G+2854  U+CBE9  uniCBE9,  HANGUL SYLLABLE-CBE9
        '52202' => '15734'            , # G+15734  U+CBEA  uniCBEA,  HANGUL SYLLABLE-CBEA
        '52203' => '15735'            , # G+15735  U+CBEB  uniCBEB,  HANGUL SYLLABLE-CBEB
        '52204' => '15736'            , # G+15736  U+CBEC  uniCBEC,  HANGUL SYLLABLE-CBEC
        '52205' => '15737'            , # G+15737  U+CBED  uniCBED,  HANGUL SYLLABLE-CBED
        '52206' => '15738'            , # G+15738  U+CBEE  uniCBEE,  HANGUL SYLLABLE-CBEE
        '52207' => '15739'            , # G+15739  U+CBEF  uniCBEF,  HANGUL SYLLABLE-CBEF
        '52208' => '15740'            , # G+15740  U+CBF0  uniCBF0,  HANGUL SYLLABLE-CBF0
        '52209' => '15741'            , # G+15741  U+CBF1  uniCBF1,  HANGUL SYLLABLE-CBF1
        '52210' => '15742'            , # G+15742  U+CBF2  uniCBF2,  HANGUL SYLLABLE-CBF2
        '52211' => '15743'            , # G+15743  U+CBF3  uniCBF3,  HANGUL SYLLABLE-CBF3
        '52212' => '15744'            , # G+15744  U+CBF4  uniCBF4,  HANGUL SYLLABLE-CBF4
        '52213' => '15745'            , # G+15745  U+CBF5  uniCBF5,  HANGUL SYLLABLE-CBF5
        '52214' => '15746'            , # G+15746  U+CBF6  uniCBF6,  HANGUL SYLLABLE-CBF6
        '52215' => '15747'            , # G+15747  U+CBF7  uniCBF7,  HANGUL SYLLABLE-CBF7
        '52216' => '15748'            , # G+15748  U+CBF8  uniCBF8,  HANGUL SYLLABLE-CBF8
        '52217' => '15749'            , # G+15749  U+CBF9  uniCBF9,  HANGUL SYLLABLE-CBF9
        '52218' => '15750'            , # G+15750  U+CBFA  uniCBFA,  HANGUL SYLLABLE-CBFA
        '52219' => '15751'            , # G+15751  U+CBFB  uniCBFB,  HANGUL SYLLABLE-CBFB
        '52220' => '15752'            , # G+15752  U+CBFC  uniCBFC,  HANGUL SYLLABLE-CBFC
        '52221' => '15753'            , # G+15753  U+CBFD  uniCBFD,  HANGUL SYLLABLE-CBFD
        '52222' => '15754'            , # G+15754  U+CBFE  uniCBFE,  HANGUL SYLLABLE-CBFE
        '52223' => '15755'            , # G+15755  U+CBFF  uniCBFF,  HANGUL SYLLABLE-CBFF
        '52224' => '15756'            , # G+15756  U+CC00  uniCC00,  HANGUL SYLLABLE-CC00
        '52225' => '15757'            , # G+15757  U+CC01  uniCC01,  HANGUL SYLLABLE-CC01
        '52226' => '15758'            , # G+15758  U+CC02  uniCC02,  HANGUL SYLLABLE-CC02
        '52227' => '15759'            , # G+15759  U+CC03  uniCC03,  HANGUL SYLLABLE-CC03
        '52228' => '15760'            , # G+15760  U+CC04  uniCC04,  HANGUL SYLLABLE-CC04
        '52229' => '15761'            , # G+15761  U+CC05  uniCC05,  HANGUL SYLLABLE-CC05
        '52230' => '15762'            , # G+15762  U+CC06  uniCC06,  HANGUL SYLLABLE-CC06
        '52231' => '15763'            , # G+15763  U+CC07  uniCC07,  HANGUL SYLLABLE-CC07
        '52232' => '15764'            , # G+15764  U+CC08  uniCC08,  HANGUL SYLLABLE-CC08
        '52233' => '15765'            , # G+15765  U+CC09  uniCC09,  HANGUL SYLLABLE-CC09
        '52234' => '15766'            , # G+15766  U+CC0A  uniCC0A,  HANGUL SYLLABLE-CC0A
        '52235' => '15767'            , # G+15767  U+CC0B  uniCC0B,  HANGUL SYLLABLE-CC0B
        '52236' => '2855'             , # G+2855  U+CC0C  uniCC0C,  HANGUL SYLLABLE-CC0C
        '52237' => '2856'             , # G+2856  U+CC0D  uniCC0D,  HANGUL SYLLABLE-CC0D
        '52238' => '15768'            , # G+15768  U+CC0E  uniCC0E,  HANGUL SYLLABLE-CC0E
        '52239' => '15769'            , # G+15769  U+CC0F  uniCC0F,  HANGUL SYLLABLE-CC0F
        '52240' => '2857'             , # G+2857  U+CC10  uniCC10,  HANGUL SYLLABLE-CC10
        '52241' => '15770'            , # G+15770  U+CC11  uniCC11,  HANGUL SYLLABLE-CC11
        '52242' => '15771'            , # G+15771  U+CC12  uniCC12,  HANGUL SYLLABLE-CC12
        '52243' => '15772'            , # G+15772  U+CC13  uniCC13,  HANGUL SYLLABLE-CC13
        '52244' => '2858'             , # G+2858  U+CC14  uniCC14,  HANGUL SYLLABLE-CC14
        '52245' => '15773'            , # G+15773  U+CC15  uniCC15,  HANGUL SYLLABLE-CC15
        '52246' => '15774'            , # G+15774  U+CC16  uniCC16,  HANGUL SYLLABLE-CC16
        '52247' => '15775'            , # G+15775  U+CC17  uniCC17,  HANGUL SYLLABLE-CC17
        '52248' => '15776'            , # G+15776  U+CC18  uniCC18,  HANGUL SYLLABLE-CC18
        '52249' => '15777'            , # G+15777  U+CC19  uniCC19,  HANGUL SYLLABLE-CC19
        '52250' => '15778'            , # G+15778  U+CC1A  uniCC1A,  HANGUL SYLLABLE-CC1A
        '52251' => '15779'            , # G+15779  U+CC1B  uniCC1B,  HANGUL SYLLABLE-CC1B
        '52252' => '2859'             , # G+2859  U+CC1C  uniCC1C,  HANGUL SYLLABLE-CC1C
        '52253' => '2860'             , # G+2860  U+CC1D  uniCC1D,  HANGUL SYLLABLE-CC1D
        '52254' => '15780'            , # G+15780  U+CC1E  uniCC1E,  HANGUL SYLLABLE-CC1E
        '52255' => '15781'            , # G+15781  U+CC1F  uniCC1F,  HANGUL SYLLABLE-CC1F
        '52256' => '15782'            , # G+15782  U+CC20  uniCC20,  HANGUL SYLLABLE-CC20
        '52257' => '2861'             , # G+2861  U+CC21  uniCC21,  HANGUL SYLLABLE-CC21
        '52258' => '2862'             , # G+2862  U+CC22  uniCC22,  HANGUL SYLLABLE-CC22
        '52259' => '15783'            , # G+15783  U+CC23  uniCC23,  HANGUL SYLLABLE-CC23
        '52260' => '15784'            , # G+15784  U+CC24  uniCC24,  HANGUL SYLLABLE-CC24
        '52261' => '15785'            , # G+15785  U+CC25  uniCC25,  HANGUL SYLLABLE-CC25
        '52262' => '15786'            , # G+15786  U+CC26  uniCC26,  HANGUL SYLLABLE-CC26
        '52263' => '2863'             , # G+2863  U+CC27  uniCC27,  HANGUL SYLLABLE-CC27
        '52264' => '2864'             , # G+2864  U+CC28  uniCC28,  HANGUL SYLLABLE-CC28
        '52265' => '2865'             , # G+2865  U+CC29  uniCC29,  HANGUL SYLLABLE-CC29
        '52266' => '15787'            , # G+15787  U+CC2A  uniCC2A,  HANGUL SYLLABLE-CC2A
        '52267' => '15788'            , # G+15788  U+CC2B  uniCC2B,  HANGUL SYLLABLE-CC2B
        '52268' => '2866'             , # G+2866  U+CC2C  uniCC2C,  HANGUL SYLLABLE-CC2C
        '52269' => '15789'            , # G+15789  U+CC2D  uniCC2D,  HANGUL SYLLABLE-CC2D
        '52270' => '2867'             , # G+2867  U+CC2E  uniCC2E,  HANGUL SYLLABLE-CC2E
        '52271' => '15790'            , # G+15790  U+CC2F  uniCC2F,  HANGUL SYLLABLE-CC2F
        '52272' => '2868'             , # G+2868  U+CC30  uniCC30,  HANGUL SYLLABLE-CC30
        '52273' => '15791'            , # G+15791  U+CC31  uniCC31,  HANGUL SYLLABLE-CC31
        '52274' => '15792'            , # G+15792  U+CC32  uniCC32,  HANGUL SYLLABLE-CC32
        '52275' => '15793'            , # G+15793  U+CC33  uniCC33,  HANGUL SYLLABLE-CC33
        '52276' => '15794'            , # G+15794  U+CC34  uniCC34,  HANGUL SYLLABLE-CC34
        '52277' => '15795'            , # G+15795  U+CC35  uniCC35,  HANGUL SYLLABLE-CC35
        '52278' => '15796'            , # G+15796  U+CC36  uniCC36,  HANGUL SYLLABLE-CC36
        '52279' => '15797'            , # G+15797  U+CC37  uniCC37,  HANGUL SYLLABLE-CC37
        '52280' => '2869'             , # G+2869  U+CC38  uniCC38,  HANGUL SYLLABLE-CC38
        '52281' => '2870'             , # G+2870  U+CC39  uniCC39,  HANGUL SYLLABLE-CC39
        '52282' => '15798'            , # G+15798  U+CC3A  uniCC3A,  HANGUL SYLLABLE-CC3A
        '52283' => '2871'             , # G+2871  U+CC3B  uniCC3B,  HANGUL SYLLABLE-CC3B
        '52284' => '2872'             , # G+2872  U+CC3C  uniCC3C,  HANGUL SYLLABLE-CC3C
        '52285' => '2873'             , # G+2873  U+CC3D  uniCC3D,  HANGUL SYLLABLE-CC3D
        '52286' => '2874'             , # G+2874  U+CC3E  uniCC3E,  HANGUL SYLLABLE-CC3E
        '52287' => '15799'            , # G+15799  U+CC3F  uniCC3F,  HANGUL SYLLABLE-CC3F
        '52288' => '15800'            , # G+15800  U+CC40  uniCC40,  HANGUL SYLLABLE-CC40
        '52289' => '15801'            , # G+15801  U+CC41  uniCC41,  HANGUL SYLLABLE-CC41
        '52290' => '15802'            , # G+15802  U+CC42  uniCC42,  HANGUL SYLLABLE-CC42
        '52291' => '15803'            , # G+15803  U+CC43  uniCC43,  HANGUL SYLLABLE-CC43
        '52292' => '2875'             , # G+2875  U+CC44  uniCC44,  HANGUL SYLLABLE-CC44
        '52293' => '2876'             , # G+2876  U+CC45  uniCC45,  HANGUL SYLLABLE-CC45
        '52294' => '15804'            , # G+15804  U+CC46  uniCC46,  HANGUL SYLLABLE-CC46
        '52295' => '15805'            , # G+15805  U+CC47  uniCC47,  HANGUL SYLLABLE-CC47
        '52296' => '2877'             , # G+2877  U+CC48  uniCC48,  HANGUL SYLLABLE-CC48
        '52297' => '15806'            , # G+15806  U+CC49  uniCC49,  HANGUL SYLLABLE-CC49
        '52298' => '15807'            , # G+15807  U+CC4A  uniCC4A,  HANGUL SYLLABLE-CC4A
        '52299' => '15808'            , # G+15808  U+CC4B  uniCC4B,  HANGUL SYLLABLE-CC4B
        '52300' => '2878'             , # G+2878  U+CC4C  uniCC4C,  HANGUL SYLLABLE-CC4C
        '52301' => '15809'            , # G+15809  U+CC4D  uniCC4D,  HANGUL SYLLABLE-CC4D
        '52302' => '15810'            , # G+15810  U+CC4E  uniCC4E,  HANGUL SYLLABLE-CC4E
        '52303' => '15811'            , # G+15811  U+CC4F  uniCC4F,  HANGUL SYLLABLE-CC4F
        '52304' => '15812'            , # G+15812  U+CC50  uniCC50,  HANGUL SYLLABLE-CC50
        '52305' => '15813'            , # G+15813  U+CC51  uniCC51,  HANGUL SYLLABLE-CC51
        '52306' => '15814'            , # G+15814  U+CC52  uniCC52,  HANGUL SYLLABLE-CC52
        '52307' => '15815'            , # G+15815  U+CC53  uniCC53,  HANGUL SYLLABLE-CC53
        '52308' => '2879'             , # G+2879  U+CC54  uniCC54,  HANGUL SYLLABLE-CC54
        '52309' => '2880'             , # G+2880  U+CC55  uniCC55,  HANGUL SYLLABLE-CC55
        '52310' => '15816'            , # G+15816  U+CC56  uniCC56,  HANGUL SYLLABLE-CC56
        '52311' => '2881'             , # G+2881  U+CC57  uniCC57,  HANGUL SYLLABLE-CC57
        '52312' => '2882'             , # G+2882  U+CC58  uniCC58,  HANGUL SYLLABLE-CC58
        '52313' => '2883'             , # G+2883  U+CC59  uniCC59,  HANGUL SYLLABLE-CC59
        '52314' => '15817'            , # G+15817  U+CC5A  uniCC5A,  HANGUL SYLLABLE-CC5A
        '52315' => '15818'            , # G+15818  U+CC5B  uniCC5B,  HANGUL SYLLABLE-CC5B
        '52316' => '15819'            , # G+15819  U+CC5C  uniCC5C,  HANGUL SYLLABLE-CC5C
        '52317' => '15820'            , # G+15820  U+CC5D  uniCC5D,  HANGUL SYLLABLE-CC5D
        '52318' => '15821'            , # G+15821  U+CC5E  uniCC5E,  HANGUL SYLLABLE-CC5E
        '52319' => '15822'            , # G+15822  U+CC5F  uniCC5F,  HANGUL SYLLABLE-CC5F
        '52320' => '2884'             , # G+2884  U+CC60  uniCC60,  HANGUL SYLLABLE-CC60
        '52321' => '15823'            , # G+15823  U+CC61  uniCC61,  HANGUL SYLLABLE-CC61
        '52322' => '15824'            , # G+15824  U+CC62  uniCC62,  HANGUL SYLLABLE-CC62
        '52323' => '15825'            , # G+15825  U+CC63  uniCC63,  HANGUL SYLLABLE-CC63
        '52324' => '2885'             , # G+2885  U+CC64  uniCC64,  HANGUL SYLLABLE-CC64
        '52325' => '15826'            , # G+15826  U+CC65  uniCC65,  HANGUL SYLLABLE-CC65
        '52326' => '2886'             , # G+2886  U+CC66  uniCC66,  HANGUL SYLLABLE-CC66
        '52327' => '15827'            , # G+15827  U+CC67  uniCC67,  HANGUL SYLLABLE-CC67
        '52328' => '2887'             , # G+2887  U+CC68  uniCC68,  HANGUL SYLLABLE-CC68
        '52329' => '15828'            , # G+15828  U+CC69  uniCC69,  HANGUL SYLLABLE-CC69
        '52330' => '15829'            , # G+15829  U+CC6A  uniCC6A,  HANGUL SYLLABLE-CC6A
        '52331' => '15830'            , # G+15830  U+CC6B  uniCC6B,  HANGUL SYLLABLE-CC6B
        '52332' => '15831'            , # G+15831  U+CC6C  uniCC6C,  HANGUL SYLLABLE-CC6C
        '52333' => '15832'            , # G+15832  U+CC6D  uniCC6D,  HANGUL SYLLABLE-CC6D
        '52334' => '15833'            , # G+15833  U+CC6E  uniCC6E,  HANGUL SYLLABLE-CC6E
        '52335' => '15834'            , # G+15834  U+CC6F  uniCC6F,  HANGUL SYLLABLE-CC6F
        '52336' => '2888'             , # G+2888  U+CC70  uniCC70,  HANGUL SYLLABLE-CC70
        '52337' => '15835'            , # G+15835  U+CC71  uniCC71,  HANGUL SYLLABLE-CC71
        '52338' => '15836'            , # G+15836  U+CC72  uniCC72,  HANGUL SYLLABLE-CC72
        '52339' => '15837'            , # G+15837  U+CC73  uniCC73,  HANGUL SYLLABLE-CC73
        '52340' => '15838'            , # G+15838  U+CC74  uniCC74,  HANGUL SYLLABLE-CC74
        '52341' => '2889'             , # G+2889  U+CC75  uniCC75,  HANGUL SYLLABLE-CC75
        '52342' => '15839'            , # G+15839  U+CC76  uniCC76,  HANGUL SYLLABLE-CC76
        '52343' => '15840'            , # G+15840  U+CC77  uniCC77,  HANGUL SYLLABLE-CC77
        '52344' => '15841'            , # G+15841  U+CC78  uniCC78,  HANGUL SYLLABLE-CC78
        '52345' => '15842'            , # G+15842  U+CC79  uniCC79,  HANGUL SYLLABLE-CC79
        '52346' => '15843'            , # G+15843  U+CC7A  uniCC7A,  HANGUL SYLLABLE-CC7A
        '52347' => '15844'            , # G+15844  U+CC7B  uniCC7B,  HANGUL SYLLABLE-CC7B
        '52348' => '15845'            , # G+15845  U+CC7C  uniCC7C,  HANGUL SYLLABLE-CC7C
        '52349' => '15846'            , # G+15846  U+CC7D  uniCC7D,  HANGUL SYLLABLE-CC7D
        '52350' => '15847'            , # G+15847  U+CC7E  uniCC7E,  HANGUL SYLLABLE-CC7E
        '52351' => '15848'            , # G+15848  U+CC7F  uniCC7F,  HANGUL SYLLABLE-CC7F
        '52352' => '15849'            , # G+15849  U+CC80  uniCC80,  HANGUL SYLLABLE-CC80
        '52353' => '15850'            , # G+15850  U+CC81  uniCC81,  HANGUL SYLLABLE-CC81
        '52354' => '15851'            , # G+15851  U+CC82  uniCC82,  HANGUL SYLLABLE-CC82
        '52355' => '15852'            , # G+15852  U+CC83  uniCC83,  HANGUL SYLLABLE-CC83
        '52356' => '15853'            , # G+15853  U+CC84  uniCC84,  HANGUL SYLLABLE-CC84
        '52357' => '15854'            , # G+15854  U+CC85  uniCC85,  HANGUL SYLLABLE-CC85
        '52358' => '15855'            , # G+15855  U+CC86  uniCC86,  HANGUL SYLLABLE-CC86
        '52359' => '15856'            , # G+15856  U+CC87  uniCC87,  HANGUL SYLLABLE-CC87
        '52360' => '15857'            , # G+15857  U+CC88  uniCC88,  HANGUL SYLLABLE-CC88
        '52361' => '15858'            , # G+15858  U+CC89  uniCC89,  HANGUL SYLLABLE-CC89
        '52362' => '15859'            , # G+15859  U+CC8A  uniCC8A,  HANGUL SYLLABLE-CC8A
        '52363' => '15860'            , # G+15860  U+CC8B  uniCC8B,  HANGUL SYLLABLE-CC8B
        '52364' => '15861'            , # G+15861  U+CC8C  uniCC8C,  HANGUL SYLLABLE-CC8C
        '52365' => '15862'            , # G+15862  U+CC8D  uniCC8D,  HANGUL SYLLABLE-CC8D
        '52366' => '15863'            , # G+15863  U+CC8E  uniCC8E,  HANGUL SYLLABLE-CC8E
        '52367' => '15864'            , # G+15864  U+CC8F  uniCC8F,  HANGUL SYLLABLE-CC8F
        '52368' => '15865'            , # G+15865  U+CC90  uniCC90,  HANGUL SYLLABLE-CC90
        '52369' => '15866'            , # G+15866  U+CC91  uniCC91,  HANGUL SYLLABLE-CC91
        '52370' => '15867'            , # G+15867  U+CC92  uniCC92,  HANGUL SYLLABLE-CC92
        '52371' => '15868'            , # G+15868  U+CC93  uniCC93,  HANGUL SYLLABLE-CC93
        '52372' => '15869'            , # G+15869  U+CC94  uniCC94,  HANGUL SYLLABLE-CC94
        '52373' => '15870'            , # G+15870  U+CC95  uniCC95,  HANGUL SYLLABLE-CC95
        '52374' => '15871'            , # G+15871  U+CC96  uniCC96,  HANGUL SYLLABLE-CC96
        '52375' => '15872'            , # G+15872  U+CC97  uniCC97,  HANGUL SYLLABLE-CC97
        '52376' => '2890'             , # G+2890  U+CC98  uniCC98,  HANGUL SYLLABLE-CC98
        '52377' => '2891'             , # G+2891  U+CC99  uniCC99,  HANGUL SYLLABLE-CC99
        '52378' => '15873'            , # G+15873  U+CC9A  uniCC9A,  HANGUL SYLLABLE-CC9A
        '52379' => '15874'            , # G+15874  U+CC9B  uniCC9B,  HANGUL SYLLABLE-CC9B
        '52380' => '2892'             , # G+2892  U+CC9C  uniCC9C,  HANGUL SYLLABLE-CC9C
        '52381' => '15875'            , # G+15875  U+CC9D  uniCC9D,  HANGUL SYLLABLE-CC9D
        '52382' => '15876'            , # G+15876  U+CC9E  uniCC9E,  HANGUL SYLLABLE-CC9E
        '52383' => '15877'            , # G+15877  U+CC9F  uniCC9F,  HANGUL SYLLABLE-CC9F
        '52384' => '2893'             , # G+2893  U+CCA0  uniCCA0,  HANGUL SYLLABLE-CCA0
        '52385' => '15878'            , # G+15878  U+CCA1  uniCCA1,  HANGUL SYLLABLE-CCA1
        '52386' => '15879'            , # G+15879  U+CCA2  uniCCA2,  HANGUL SYLLABLE-CCA2
        '52387' => '15880'            , # G+15880  U+CCA3  uniCCA3,  HANGUL SYLLABLE-CCA3
        '52388' => '15881'            , # G+15881  U+CCA4  uniCCA4,  HANGUL SYLLABLE-CCA4
        '52389' => '15882'            , # G+15882  U+CCA5  uniCCA5,  HANGUL SYLLABLE-CCA5
        '52390' => '15883'            , # G+15883  U+CCA6  uniCCA6,  HANGUL SYLLABLE-CCA6
        '52391' => '15884'            , # G+15884  U+CCA7  uniCCA7,  HANGUL SYLLABLE-CCA7
        '52392' => '2894'             , # G+2894  U+CCA8  uniCCA8,  HANGUL SYLLABLE-CCA8
        '52393' => '2895'             , # G+2895  U+CCA9  uniCCA9,  HANGUL SYLLABLE-CCA9
        '52394' => '15885'            , # G+15885  U+CCAA  uniCCAA,  HANGUL SYLLABLE-CCAA
        '52395' => '2896'             , # G+2896  U+CCAB  uniCCAB,  HANGUL SYLLABLE-CCAB
        '52396' => '2897'             , # G+2897  U+CCAC  uniCCAC,  HANGUL SYLLABLE-CCAC
        '52397' => '2898'             , # G+2898  U+CCAD  uniCCAD,  HANGUL SYLLABLE-CCAD
        '52398' => '15886'            , # G+15886  U+CCAE  uniCCAE,  HANGUL SYLLABLE-CCAE
        '52399' => '15887'            , # G+15887  U+CCAF  uniCCAF,  HANGUL SYLLABLE-CCAF
        '52400' => '15888'            , # G+15888  U+CCB0  uniCCB0,  HANGUL SYLLABLE-CCB0
        '52401' => '15889'            , # G+15889  U+CCB1  uniCCB1,  HANGUL SYLLABLE-CCB1
        '52402' => '15890'            , # G+15890  U+CCB2  uniCCB2,  HANGUL SYLLABLE-CCB2
        '52403' => '15891'            , # G+15891  U+CCB3  uniCCB3,  HANGUL SYLLABLE-CCB3
        '52404' => '2899'             , # G+2899  U+CCB4  uniCCB4,  HANGUL SYLLABLE-CCB4
        '52405' => '2900'             , # G+2900  U+CCB5  uniCCB5,  HANGUL SYLLABLE-CCB5
        '52406' => '15892'            , # G+15892  U+CCB6  uniCCB6,  HANGUL SYLLABLE-CCB6
        '52407' => '15893'            , # G+15893  U+CCB7  uniCCB7,  HANGUL SYLLABLE-CCB7
        '52408' => '2901'             , # G+2901  U+CCB8  uniCCB8,  HANGUL SYLLABLE-CCB8
        '52409' => '15894'            , # G+15894  U+CCB9  uniCCB9,  HANGUL SYLLABLE-CCB9
        '52410' => '15895'            , # G+15895  U+CCBA  uniCCBA,  HANGUL SYLLABLE-CCBA
        '52411' => '15896'            , # G+15896  U+CCBB  uniCCBB,  HANGUL SYLLABLE-CCBB
        '52412' => '2902'             , # G+2902  U+CCBC  uniCCBC,  HANGUL SYLLABLE-CCBC
        '52413' => '15897'            , # G+15897  U+CCBD  uniCCBD,  HANGUL SYLLABLE-CCBD
        '52414' => '15898'            , # G+15898  U+CCBE  uniCCBE,  HANGUL SYLLABLE-CCBE
        '52415' => '15899'            , # G+15899  U+CCBF  uniCCBF,  HANGUL SYLLABLE-CCBF
        '52416' => '15900'            , # G+15900  U+CCC0  uniCCC0,  HANGUL SYLLABLE-CCC0
        '52417' => '15901'            , # G+15901  U+CCC1  uniCCC1,  HANGUL SYLLABLE-CCC1
        '52418' => '15902'            , # G+15902  U+CCC2  uniCCC2,  HANGUL SYLLABLE-CCC2
        '52419' => '15903'            , # G+15903  U+CCC3  uniCCC3,  HANGUL SYLLABLE-CCC3
        '52420' => '2903'             , # G+2903  U+CCC4  uniCCC4,  HANGUL SYLLABLE-CCC4
        '52421' => '2904'             , # G+2904  U+CCC5  uniCCC5,  HANGUL SYLLABLE-CCC5
        '52422' => '15904'            , # G+15904  U+CCC6  uniCCC6,  HANGUL SYLLABLE-CCC6
        '52423' => '2905'             , # G+2905  U+CCC7  uniCCC7,  HANGUL SYLLABLE-CCC7
        '52424' => '15905'            , # G+15905  U+CCC8  uniCCC8,  HANGUL SYLLABLE-CCC8
        '52425' => '2906'             , # G+2906  U+CCC9  uniCCC9,  HANGUL SYLLABLE-CCC9
        '52426' => '15906'            , # G+15906  U+CCCA  uniCCCA,  HANGUL SYLLABLE-CCCA
        '52427' => '15907'            , # G+15907  U+CCCB  uniCCCB,  HANGUL SYLLABLE-CCCB
        '52428' => '15908'            , # G+15908  U+CCCC  uniCCCC,  HANGUL SYLLABLE-CCCC
        '52429' => '15909'            , # G+15909  U+CCCD  uniCCCD,  HANGUL SYLLABLE-CCCD
        '52430' => '15910'            , # G+15910  U+CCCE  uniCCCE,  HANGUL SYLLABLE-CCCE
        '52431' => '15911'            , # G+15911  U+CCCF  uniCCCF,  HANGUL SYLLABLE-CCCF
        '52432' => '2907'             , # G+2907  U+CCD0  uniCCD0,  HANGUL SYLLABLE-CCD0
        '52433' => '15912'            , # G+15912  U+CCD1  uniCCD1,  HANGUL SYLLABLE-CCD1
        '52434' => '15913'            , # G+15913  U+CCD2  uniCCD2,  HANGUL SYLLABLE-CCD2
        '52435' => '15914'            , # G+15914  U+CCD3  uniCCD3,  HANGUL SYLLABLE-CCD3
        '52436' => '2908'             , # G+2908  U+CCD4  uniCCD4,  HANGUL SYLLABLE-CCD4
        '52437' => '15915'            , # G+15915  U+CCD5  uniCCD5,  HANGUL SYLLABLE-CCD5
        '52438' => '15916'            , # G+15916  U+CCD6  uniCCD6,  HANGUL SYLLABLE-CCD6
        '52439' => '15917'            , # G+15917  U+CCD7  uniCCD7,  HANGUL SYLLABLE-CCD7
        '52440' => '15918'            , # G+15918  U+CCD8  uniCCD8,  HANGUL SYLLABLE-CCD8
        '52441' => '15919'            , # G+15919  U+CCD9  uniCCD9,  HANGUL SYLLABLE-CCD9
        '52442' => '15920'            , # G+15920  U+CCDA  uniCCDA,  HANGUL SYLLABLE-CCDA
        '52443' => '15921'            , # G+15921  U+CCDB  uniCCDB,  HANGUL SYLLABLE-CCDB
        '52444' => '15922'            , # G+15922  U+CCDC  uniCCDC,  HANGUL SYLLABLE-CCDC
        '52445' => '15923'            , # G+15923  U+CCDD  uniCCDD,  HANGUL SYLLABLE-CCDD
        '52446' => '15924'            , # G+15924  U+CCDE  uniCCDE,  HANGUL SYLLABLE-CCDE
        '52447' => '15925'            , # G+15925  U+CCDF  uniCCDF,  HANGUL SYLLABLE-CCDF
        '52448' => '15926'            , # G+15926  U+CCE0  uniCCE0,  HANGUL SYLLABLE-CCE0
        '52449' => '15927'            , # G+15927  U+CCE1  uniCCE1,  HANGUL SYLLABLE-CCE1
        '52450' => '15928'            , # G+15928  U+CCE2  uniCCE2,  HANGUL SYLLABLE-CCE2
        '52451' => '15929'            , # G+15929  U+CCE3  uniCCE3,  HANGUL SYLLABLE-CCE3
        '52452' => '2909'             , # G+2909  U+CCE4  uniCCE4,  HANGUL SYLLABLE-CCE4
        '52453' => '15930'            , # G+15930  U+CCE5  uniCCE5,  HANGUL SYLLABLE-CCE5
        '52454' => '15931'            , # G+15931  U+CCE6  uniCCE6,  HANGUL SYLLABLE-CCE6
        '52455' => '15932'            , # G+15932  U+CCE7  uniCCE7,  HANGUL SYLLABLE-CCE7
        '52456' => '15933'            , # G+15933  U+CCE8  uniCCE8,  HANGUL SYLLABLE-CCE8
        '52457' => '15934'            , # G+15934  U+CCE9  uniCCE9,  HANGUL SYLLABLE-CCE9
        '52458' => '15935'            , # G+15935  U+CCEA  uniCCEA,  HANGUL SYLLABLE-CCEA
        '52459' => '15936'            , # G+15936  U+CCEB  uniCCEB,  HANGUL SYLLABLE-CCEB
        '52460' => '2910'             , # G+2910  U+CCEC  uniCCEC,  HANGUL SYLLABLE-CCEC
        '52461' => '15937'            , # G+15937  U+CCED  uniCCED,  HANGUL SYLLABLE-CCED
        '52462' => '15938'            , # G+15938  U+CCEE  uniCCEE,  HANGUL SYLLABLE-CCEE
        '52463' => '15939'            , # G+15939  U+CCEF  uniCCEF,  HANGUL SYLLABLE-CCEF
        '52464' => '2911'             , # G+2911  U+CCF0  uniCCF0,  HANGUL SYLLABLE-CCF0
        '52465' => '15940'            , # G+15940  U+CCF1  uniCCF1,  HANGUL SYLLABLE-CCF1
        '52466' => '15941'            , # G+15941  U+CCF2  uniCCF2,  HANGUL SYLLABLE-CCF2
        '52467' => '15942'            , # G+15942  U+CCF3  uniCCF3,  HANGUL SYLLABLE-CCF3
        '52468' => '15943'            , # G+15943  U+CCF4  uniCCF4,  HANGUL SYLLABLE-CCF4
        '52469' => '15944'            , # G+15944  U+CCF5  uniCCF5,  HANGUL SYLLABLE-CCF5
        '52470' => '15945'            , # G+15945  U+CCF6  uniCCF6,  HANGUL SYLLABLE-CCF6
        '52471' => '15946'            , # G+15946  U+CCF7  uniCCF7,  HANGUL SYLLABLE-CCF7
        '52472' => '15947'            , # G+15947  U+CCF8  uniCCF8,  HANGUL SYLLABLE-CCF8
        '52473' => '15948'            , # G+15948  U+CCF9  uniCCF9,  HANGUL SYLLABLE-CCF9
        '52474' => '15949'            , # G+15949  U+CCFA  uniCCFA,  HANGUL SYLLABLE-CCFA
        '52475' => '15950'            , # G+15950  U+CCFB  uniCCFB,  HANGUL SYLLABLE-CCFB
        '52476' => '15951'            , # G+15951  U+CCFC  uniCCFC,  HANGUL SYLLABLE-CCFC
        '52477' => '15952'            , # G+15952  U+CCFD  uniCCFD,  HANGUL SYLLABLE-CCFD
        '52478' => '15953'            , # G+15953  U+CCFE  uniCCFE,  HANGUL SYLLABLE-CCFE
        '52479' => '15954'            , # G+15954  U+CCFF  uniCCFF,  HANGUL SYLLABLE-CCFF
        '52480' => '15955'            , # G+15955  U+CD00  uniCD00,  HANGUL SYLLABLE-CD00
        '52481' => '2912'             , # G+2912  U+CD01  uniCD01,  HANGUL SYLLABLE-CD01
        '52482' => '15956'            , # G+15956  U+CD02  uniCD02,  HANGUL SYLLABLE-CD02
        '52483' => '15957'            , # G+15957  U+CD03  uniCD03,  HANGUL SYLLABLE-CD03
        '52484' => '15958'            , # G+15958  U+CD04  uniCD04,  HANGUL SYLLABLE-CD04
        '52485' => '15959'            , # G+15959  U+CD05  uniCD05,  HANGUL SYLLABLE-CD05
        '52486' => '15960'            , # G+15960  U+CD06  uniCD06,  HANGUL SYLLABLE-CD06
        '52487' => '15961'            , # G+15961  U+CD07  uniCD07,  HANGUL SYLLABLE-CD07
        '52488' => '2913'             , # G+2913  U+CD08  uniCD08,  HANGUL SYLLABLE-CD08
        '52489' => '2914'             , # G+2914  U+CD09  uniCD09,  HANGUL SYLLABLE-CD09
        '52490' => '15962'            , # G+15962  U+CD0A  uniCD0A,  HANGUL SYLLABLE-CD0A
        '52491' => '15963'            , # G+15963  U+CD0B  uniCD0B,  HANGUL SYLLABLE-CD0B
        '52492' => '2915'             , # G+2915  U+CD0C  uniCD0C,  HANGUL SYLLABLE-CD0C
        '52493' => '15964'            , # G+15964  U+CD0D  uniCD0D,  HANGUL SYLLABLE-CD0D
        '52494' => '15965'            , # G+15965  U+CD0E  uniCD0E,  HANGUL SYLLABLE-CD0E
        '52495' => '15966'            , # G+15966  U+CD0F  uniCD0F,  HANGUL SYLLABLE-CD0F
        '52496' => '2916'             , # G+2916  U+CD10  uniCD10,  HANGUL SYLLABLE-CD10
        '52497' => '15967'            , # G+15967  U+CD11  uniCD11,  HANGUL SYLLABLE-CD11
        '52498' => '15968'            , # G+15968  U+CD12  uniCD12,  HANGUL SYLLABLE-CD12
        '52499' => '15969'            , # G+15969  U+CD13  uniCD13,  HANGUL SYLLABLE-CD13
        '52500' => '15970'            , # G+15970  U+CD14  uniCD14,  HANGUL SYLLABLE-CD14
        '52501' => '15971'            , # G+15971  U+CD15  uniCD15,  HANGUL SYLLABLE-CD15
        '52502' => '15972'            , # G+15972  U+CD16  uniCD16,  HANGUL SYLLABLE-CD16
        '52503' => '15973'            , # G+15973  U+CD17  uniCD17,  HANGUL SYLLABLE-CD17
        '52504' => '2917'             , # G+2917  U+CD18  uniCD18,  HANGUL SYLLABLE-CD18
        '52505' => '2918'             , # G+2918  U+CD19  uniCD19,  HANGUL SYLLABLE-CD19
        '52506' => '15974'            , # G+15974  U+CD1A  uniCD1A,  HANGUL SYLLABLE-CD1A
        '52507' => '2919'             , # G+2919  U+CD1B  uniCD1B,  HANGUL SYLLABLE-CD1B
        '52508' => '15975'            , # G+15975  U+CD1C  uniCD1C,  HANGUL SYLLABLE-CD1C
        '52509' => '2920'             , # G+2920  U+CD1D  uniCD1D,  HANGUL SYLLABLE-CD1D
        '52510' => '15976'            , # G+15976  U+CD1E  uniCD1E,  HANGUL SYLLABLE-CD1E
        '52511' => '15977'            , # G+15977  U+CD1F  uniCD1F,  HANGUL SYLLABLE-CD1F
        '52512' => '15978'            , # G+15978  U+CD20  uniCD20,  HANGUL SYLLABLE-CD20
        '52513' => '15979'            , # G+15979  U+CD21  uniCD21,  HANGUL SYLLABLE-CD21
        '52514' => '15980'            , # G+15980  U+CD22  uniCD22,  HANGUL SYLLABLE-CD22
        '52515' => '15981'            , # G+15981  U+CD23  uniCD23,  HANGUL SYLLABLE-CD23
        '52516' => '2921'             , # G+2921  U+CD24  uniCD24,  HANGUL SYLLABLE-CD24
        '52517' => '15982'            , # G+15982  U+CD25  uniCD25,  HANGUL SYLLABLE-CD25
        '52518' => '15983'            , # G+15983  U+CD26  uniCD26,  HANGUL SYLLABLE-CD26
        '52519' => '15984'            , # G+15984  U+CD27  uniCD27,  HANGUL SYLLABLE-CD27
        '52520' => '2922'             , # G+2922  U+CD28  uniCD28,  HANGUL SYLLABLE-CD28
        '52521' => '15985'            , # G+15985  U+CD29  uniCD29,  HANGUL SYLLABLE-CD29
        '52522' => '15986'            , # G+15986  U+CD2A  uniCD2A,  HANGUL SYLLABLE-CD2A
        '52523' => '15987'            , # G+15987  U+CD2B  uniCD2B,  HANGUL SYLLABLE-CD2B
        '52524' => '2923'             , # G+2923  U+CD2C  uniCD2C,  HANGUL SYLLABLE-CD2C
        '52525' => '15988'            , # G+15988  U+CD2D  uniCD2D,  HANGUL SYLLABLE-CD2D
        '52526' => '15989'            , # G+15989  U+CD2E  uniCD2E,  HANGUL SYLLABLE-CD2E
        '52527' => '15990'            , # G+15990  U+CD2F  uniCD2F,  HANGUL SYLLABLE-CD2F
        '52528' => '15991'            , # G+15991  U+CD30  uniCD30,  HANGUL SYLLABLE-CD30
        '52529' => '15992'            , # G+15992  U+CD31  uniCD31,  HANGUL SYLLABLE-CD31
        '52530' => '15993'            , # G+15993  U+CD32  uniCD32,  HANGUL SYLLABLE-CD32
        '52531' => '15994'            , # G+15994  U+CD33  uniCD33,  HANGUL SYLLABLE-CD33
        '52532' => '15995'            , # G+15995  U+CD34  uniCD34,  HANGUL SYLLABLE-CD34
        '52533' => '15996'            , # G+15996  U+CD35  uniCD35,  HANGUL SYLLABLE-CD35
        '52534' => '15997'            , # G+15997  U+CD36  uniCD36,  HANGUL SYLLABLE-CD36
        '52535' => '15998'            , # G+15998  U+CD37  uniCD37,  HANGUL SYLLABLE-CD37
        '52536' => '15999'            , # G+15999  U+CD38  uniCD38,  HANGUL SYLLABLE-CD38
        '52537' => '2924'             , # G+2924  U+CD39  uniCD39,  HANGUL SYLLABLE-CD39
        '52538' => '16000'            , # G+16000  U+CD3A  uniCD3A,  HANGUL SYLLABLE-CD3A
        '52539' => '16001'            , # G+16001  U+CD3B  uniCD3B,  HANGUL SYLLABLE-CD3B
        '52540' => '16002'            , # G+16002  U+CD3C  uniCD3C,  HANGUL SYLLABLE-CD3C
        '52541' => '16003'            , # G+16003  U+CD3D  uniCD3D,  HANGUL SYLLABLE-CD3D
        '52542' => '16004'            , # G+16004  U+CD3E  uniCD3E,  HANGUL SYLLABLE-CD3E
        '52543' => '16005'            , # G+16005  U+CD3F  uniCD3F,  HANGUL SYLLABLE-CD3F
        '52544' => '16006'            , # G+16006  U+CD40  uniCD40,  HANGUL SYLLABLE-CD40
        '52545' => '16007'            , # G+16007  U+CD41  uniCD41,  HANGUL SYLLABLE-CD41
        '52546' => '16008'            , # G+16008  U+CD42  uniCD42,  HANGUL SYLLABLE-CD42
        '52547' => '16009'            , # G+16009  U+CD43  uniCD43,  HANGUL SYLLABLE-CD43
        '52548' => '16010'            , # G+16010  U+CD44  uniCD44,  HANGUL SYLLABLE-CD44
        '52549' => '16011'            , # G+16011  U+CD45  uniCD45,  HANGUL SYLLABLE-CD45
        '52550' => '16012'            , # G+16012  U+CD46  uniCD46,  HANGUL SYLLABLE-CD46
        '52551' => '16013'            , # G+16013  U+CD47  uniCD47,  HANGUL SYLLABLE-CD47
        '52552' => '16014'            , # G+16014  U+CD48  uniCD48,  HANGUL SYLLABLE-CD48
        '52553' => '16015'            , # G+16015  U+CD49  uniCD49,  HANGUL SYLLABLE-CD49
        '52554' => '16016'            , # G+16016  U+CD4A  uniCD4A,  HANGUL SYLLABLE-CD4A
        '52555' => '16017'            , # G+16017  U+CD4B  uniCD4B,  HANGUL SYLLABLE-CD4B
        '52556' => '16018'            , # G+16018  U+CD4C  uniCD4C,  HANGUL SYLLABLE-CD4C
        '52557' => '16019'            , # G+16019  U+CD4D  uniCD4D,  HANGUL SYLLABLE-CD4D
        '52558' => '16020'            , # G+16020  U+CD4E  uniCD4E,  HANGUL SYLLABLE-CD4E
        '52559' => '16021'            , # G+16021  U+CD4F  uniCD4F,  HANGUL SYLLABLE-CD4F
        '52560' => '16022'            , # G+16022  U+CD50  uniCD50,  HANGUL SYLLABLE-CD50
        '52561' => '16023'            , # G+16023  U+CD51  uniCD51,  HANGUL SYLLABLE-CD51
        '52562' => '16024'            , # G+16024  U+CD52  uniCD52,  HANGUL SYLLABLE-CD52
        '52563' => '16025'            , # G+16025  U+CD53  uniCD53,  HANGUL SYLLABLE-CD53
        '52564' => '16026'            , # G+16026  U+CD54  uniCD54,  HANGUL SYLLABLE-CD54
        '52565' => '16027'            , # G+16027  U+CD55  uniCD55,  HANGUL SYLLABLE-CD55
        '52566' => '16028'            , # G+16028  U+CD56  uniCD56,  HANGUL SYLLABLE-CD56
        '52567' => '16029'            , # G+16029  U+CD57  uniCD57,  HANGUL SYLLABLE-CD57
        '52568' => '16030'            , # G+16030  U+CD58  uniCD58,  HANGUL SYLLABLE-CD58
        '52569' => '16031'            , # G+16031  U+CD59  uniCD59,  HANGUL SYLLABLE-CD59
        '52570' => '16032'            , # G+16032  U+CD5A  uniCD5A,  HANGUL SYLLABLE-CD5A
        '52571' => '16033'            , # G+16033  U+CD5B  uniCD5B,  HANGUL SYLLABLE-CD5B
        '52572' => '2925'             , # G+2925  U+CD5C  uniCD5C,  HANGUL SYLLABLE-CD5C
        '52573' => '16034'            , # G+16034  U+CD5D  uniCD5D,  HANGUL SYLLABLE-CD5D
        '52574' => '16035'            , # G+16035  U+CD5E  uniCD5E,  HANGUL SYLLABLE-CD5E
        '52575' => '16036'            , # G+16036  U+CD5F  uniCD5F,  HANGUL SYLLABLE-CD5F
        '52576' => '2926'             , # G+2926  U+CD60  uniCD60,  HANGUL SYLLABLE-CD60
        '52577' => '16037'            , # G+16037  U+CD61  uniCD61,  HANGUL SYLLABLE-CD61
        '52578' => '16038'            , # G+16038  U+CD62  uniCD62,  HANGUL SYLLABLE-CD62
        '52579' => '16039'            , # G+16039  U+CD63  uniCD63,  HANGUL SYLLABLE-CD63
        '52580' => '2927'             , # G+2927  U+CD64  uniCD64,  HANGUL SYLLABLE-CD64
        '52581' => '16040'            , # G+16040  U+CD65  uniCD65,  HANGUL SYLLABLE-CD65
        '52582' => '16041'            , # G+16041  U+CD66  uniCD66,  HANGUL SYLLABLE-CD66
        '52583' => '16042'            , # G+16042  U+CD67  uniCD67,  HANGUL SYLLABLE-CD67
        '52584' => '16043'            , # G+16043  U+CD68  uniCD68,  HANGUL SYLLABLE-CD68
        '52585' => '16044'            , # G+16044  U+CD69  uniCD69,  HANGUL SYLLABLE-CD69
        '52586' => '16045'            , # G+16045  U+CD6A  uniCD6A,  HANGUL SYLLABLE-CD6A
        '52587' => '16046'            , # G+16046  U+CD6B  uniCD6B,  HANGUL SYLLABLE-CD6B
        '52588' => '2928'             , # G+2928  U+CD6C  uniCD6C,  HANGUL SYLLABLE-CD6C
        '52589' => '2929'             , # G+2929  U+CD6D  uniCD6D,  HANGUL SYLLABLE-CD6D
        '52590' => '16047'            , # G+16047  U+CD6E  uniCD6E,  HANGUL SYLLABLE-CD6E
        '52591' => '2930'             , # G+2930  U+CD6F  uniCD6F,  HANGUL SYLLABLE-CD6F
        '52592' => '16048'            , # G+16048  U+CD70  uniCD70,  HANGUL SYLLABLE-CD70
        '52593' => '2931'             , # G+2931  U+CD71  uniCD71,  HANGUL SYLLABLE-CD71
        '52594' => '16049'            , # G+16049  U+CD72  uniCD72,  HANGUL SYLLABLE-CD72
        '52595' => '16050'            , # G+16050  U+CD73  uniCD73,  HANGUL SYLLABLE-CD73
        '52596' => '16051'            , # G+16051  U+CD74  uniCD74,  HANGUL SYLLABLE-CD74
        '52597' => '16052'            , # G+16052  U+CD75  uniCD75,  HANGUL SYLLABLE-CD75
        '52598' => '16053'            , # G+16053  U+CD76  uniCD76,  HANGUL SYLLABLE-CD76
        '52599' => '16054'            , # G+16054  U+CD77  uniCD77,  HANGUL SYLLABLE-CD77
        '52600' => '2932'             , # G+2932  U+CD78  uniCD78,  HANGUL SYLLABLE-CD78
        '52601' => '16055'            , # G+16055  U+CD79  uniCD79,  HANGUL SYLLABLE-CD79
        '52602' => '16056'            , # G+16056  U+CD7A  uniCD7A,  HANGUL SYLLABLE-CD7A
        '52603' => '16057'            , # G+16057  U+CD7B  uniCD7B,  HANGUL SYLLABLE-CD7B
        '52604' => '16058'            , # G+16058  U+CD7C  uniCD7C,  HANGUL SYLLABLE-CD7C
        '52605' => '16059'            , # G+16059  U+CD7D  uniCD7D,  HANGUL SYLLABLE-CD7D
        '52606' => '16060'            , # G+16060  U+CD7E  uniCD7E,  HANGUL SYLLABLE-CD7E
        '52607' => '16061'            , # G+16061  U+CD7F  uniCD7F,  HANGUL SYLLABLE-CD7F
        '52608' => '16062'            , # G+16062  U+CD80  uniCD80,  HANGUL SYLLABLE-CD80
        '52609' => '16063'            , # G+16063  U+CD81  uniCD81,  HANGUL SYLLABLE-CD81
        '52610' => '16064'            , # G+16064  U+CD82  uniCD82,  HANGUL SYLLABLE-CD82
        '52611' => '16065'            , # G+16065  U+CD83  uniCD83,  HANGUL SYLLABLE-CD83
        '52612' => '16066'            , # G+16066  U+CD84  uniCD84,  HANGUL SYLLABLE-CD84
        '52613' => '16067'            , # G+16067  U+CD85  uniCD85,  HANGUL SYLLABLE-CD85
        '52614' => '16068'            , # G+16068  U+CD86  uniCD86,  HANGUL SYLLABLE-CD86
        '52615' => '16069'            , # G+16069  U+CD87  uniCD87,  HANGUL SYLLABLE-CD87
        '52616' => '2933'             , # G+2933  U+CD88  uniCD88,  HANGUL SYLLABLE-CD88
        '52617' => '16070'            , # G+16070  U+CD89  uniCD89,  HANGUL SYLLABLE-CD89
        '52618' => '16071'            , # G+16071  U+CD8A  uniCD8A,  HANGUL SYLLABLE-CD8A
        '52619' => '16072'            , # G+16072  U+CD8B  uniCD8B,  HANGUL SYLLABLE-CD8B
        '52620' => '16073'            , # G+16073  U+CD8C  uniCD8C,  HANGUL SYLLABLE-CD8C
        '52621' => '16074'            , # G+16074  U+CD8D  uniCD8D,  HANGUL SYLLABLE-CD8D
        '52622' => '16075'            , # G+16075  U+CD8E  uniCD8E,  HANGUL SYLLABLE-CD8E
        '52623' => '16076'            , # G+16076  U+CD8F  uniCD8F,  HANGUL SYLLABLE-CD8F
        '52624' => '16077'            , # G+16077  U+CD90  uniCD90,  HANGUL SYLLABLE-CD90
        '52625' => '16078'            , # G+16078  U+CD91  uniCD91,  HANGUL SYLLABLE-CD91
        '52626' => '16079'            , # G+16079  U+CD92  uniCD92,  HANGUL SYLLABLE-CD92
        '52627' => '16080'            , # G+16080  U+CD93  uniCD93,  HANGUL SYLLABLE-CD93
        '52628' => '2934'             , # G+2934  U+CD94  uniCD94,  HANGUL SYLLABLE-CD94
        '52629' => '2935'             , # G+2935  U+CD95  uniCD95,  HANGUL SYLLABLE-CD95
        '52630' => '16081'            , # G+16081  U+CD96  uniCD96,  HANGUL SYLLABLE-CD96
        '52631' => '16082'            , # G+16082  U+CD97  uniCD97,  HANGUL SYLLABLE-CD97
        '52632' => '2936'             , # G+2936  U+CD98  uniCD98,  HANGUL SYLLABLE-CD98
        '52633' => '16083'            , # G+16083  U+CD99  uniCD99,  HANGUL SYLLABLE-CD99
        '52634' => '16084'            , # G+16084  U+CD9A  uniCD9A,  HANGUL SYLLABLE-CD9A
        '52635' => '16085'            , # G+16085  U+CD9B  uniCD9B,  HANGUL SYLLABLE-CD9B
        '52636' => '2937'             , # G+2937  U+CD9C  uniCD9C,  HANGUL SYLLABLE-CD9C
        '52637' => '16086'            , # G+16086  U+CD9D  uniCD9D,  HANGUL SYLLABLE-CD9D
        '52638' => '16087'            , # G+16087  U+CD9E  uniCD9E,  HANGUL SYLLABLE-CD9E
        '52639' => '16088'            , # G+16088  U+CD9F  uniCD9F,  HANGUL SYLLABLE-CD9F
        '52640' => '16089'            , # G+16089  U+CDA0  uniCDA0,  HANGUL SYLLABLE-CDA0
        '52641' => '16090'            , # G+16090  U+CDA1  uniCDA1,  HANGUL SYLLABLE-CDA1
        '52642' => '16091'            , # G+16091  U+CDA2  uniCDA2,  HANGUL SYLLABLE-CDA2
        '52643' => '16092'            , # G+16092  U+CDA3  uniCDA3,  HANGUL SYLLABLE-CDA3
        '52644' => '2938'             , # G+2938  U+CDA4  uniCDA4,  HANGUL SYLLABLE-CDA4
        '52645' => '2939'             , # G+2939  U+CDA5  uniCDA5,  HANGUL SYLLABLE-CDA5
        '52646' => '16093'            , # G+16093  U+CDA6  uniCDA6,  HANGUL SYLLABLE-CDA6
        '52647' => '2940'             , # G+2940  U+CDA7  uniCDA7,  HANGUL SYLLABLE-CDA7
        '52648' => '16094'            , # G+16094  U+CDA8  uniCDA8,  HANGUL SYLLABLE-CDA8
        '52649' => '2941'             , # G+2941  U+CDA9  uniCDA9,  HANGUL SYLLABLE-CDA9
        '52650' => '16095'            , # G+16095  U+CDAA  uniCDAA,  HANGUL SYLLABLE-CDAA
        '52651' => '16096'            , # G+16096  U+CDAB  uniCDAB,  HANGUL SYLLABLE-CDAB
        '52652' => '16097'            , # G+16097  U+CDAC  uniCDAC,  HANGUL SYLLABLE-CDAC
        '52653' => '16098'            , # G+16098  U+CDAD  uniCDAD,  HANGUL SYLLABLE-CDAD
        '52654' => '16099'            , # G+16099  U+CDAE  uniCDAE,  HANGUL SYLLABLE-CDAE
        '52655' => '16100'            , # G+16100  U+CDAF  uniCDAF,  HANGUL SYLLABLE-CDAF
        '52656' => '2942'             , # G+2942  U+CDB0  uniCDB0,  HANGUL SYLLABLE-CDB0
        '52657' => '16101'            , # G+16101  U+CDB1  uniCDB1,  HANGUL SYLLABLE-CDB1
        '52658' => '16102'            , # G+16102  U+CDB2  uniCDB2,  HANGUL SYLLABLE-CDB2
        '52659' => '16103'            , # G+16103  U+CDB3  uniCDB3,  HANGUL SYLLABLE-CDB3
        '52660' => '16104'            , # G+16104  U+CDB4  uniCDB4,  HANGUL SYLLABLE-CDB4
        '52661' => '16105'            , # G+16105  U+CDB5  uniCDB5,  HANGUL SYLLABLE-CDB5
        '52662' => '16106'            , # G+16106  U+CDB6  uniCDB6,  HANGUL SYLLABLE-CDB6
        '52663' => '16107'            , # G+16107  U+CDB7  uniCDB7,  HANGUL SYLLABLE-CDB7
        '52664' => '16108'            , # G+16108  U+CDB8  uniCDB8,  HANGUL SYLLABLE-CDB8
        '52665' => '16109'            , # G+16109  U+CDB9  uniCDB9,  HANGUL SYLLABLE-CDB9
        '52666' => '16110'            , # G+16110  U+CDBA  uniCDBA,  HANGUL SYLLABLE-CDBA
        '52667' => '16111'            , # G+16111  U+CDBB  uniCDBB,  HANGUL SYLLABLE-CDBB
        '52668' => '16112'            , # G+16112  U+CDBC  uniCDBC,  HANGUL SYLLABLE-CDBC
        '52669' => '16113'            , # G+16113  U+CDBD  uniCDBD,  HANGUL SYLLABLE-CDBD
        '52670' => '16114'            , # G+16114  U+CDBE  uniCDBE,  HANGUL SYLLABLE-CDBE
        '52671' => '16115'            , # G+16115  U+CDBF  uniCDBF,  HANGUL SYLLABLE-CDBF
        '52672' => '16116'            , # G+16116  U+CDC0  uniCDC0,  HANGUL SYLLABLE-CDC0
        '52673' => '16117'            , # G+16117  U+CDC1  uniCDC1,  HANGUL SYLLABLE-CDC1
        '52674' => '16118'            , # G+16118  U+CDC2  uniCDC2,  HANGUL SYLLABLE-CDC2
        '52675' => '16119'            , # G+16119  U+CDC3  uniCDC3,  HANGUL SYLLABLE-CDC3
        '52676' => '2943'             , # G+2943  U+CDC4  uniCDC4,  HANGUL SYLLABLE-CDC4
        '52677' => '16120'            , # G+16120  U+CDC5  uniCDC5,  HANGUL SYLLABLE-CDC5
        '52678' => '16121'            , # G+16121  U+CDC6  uniCDC6,  HANGUL SYLLABLE-CDC6
        '52679' => '16122'            , # G+16122  U+CDC7  uniCDC7,  HANGUL SYLLABLE-CDC7
        '52680' => '16123'            , # G+16123  U+CDC8  uniCDC8,  HANGUL SYLLABLE-CDC8
        '52681' => '16124'            , # G+16124  U+CDC9  uniCDC9,  HANGUL SYLLABLE-CDC9
        '52682' => '16125'            , # G+16125  U+CDCA  uniCDCA,  HANGUL SYLLABLE-CDCA
        '52683' => '16126'            , # G+16126  U+CDCB  uniCDCB,  HANGUL SYLLABLE-CDCB
        '52684' => '2944'             , # G+2944  U+CDCC  uniCDCC,  HANGUL SYLLABLE-CDCC
        '52685' => '16127'            , # G+16127  U+CDCD  uniCDCD,  HANGUL SYLLABLE-CDCD
        '52686' => '16128'            , # G+16128  U+CDCE  uniCDCE,  HANGUL SYLLABLE-CDCE
        '52687' => '16129'            , # G+16129  U+CDCF  uniCDCF,  HANGUL SYLLABLE-CDCF
        '52688' => '2945'             , # G+2945  U+CDD0  uniCDD0,  HANGUL SYLLABLE-CDD0
        '52689' => '16130'            , # G+16130  U+CDD1  uniCDD1,  HANGUL SYLLABLE-CDD1
        '52690' => '16131'            , # G+16131  U+CDD2  uniCDD2,  HANGUL SYLLABLE-CDD2
        '52691' => '16132'            , # G+16132  U+CDD3  uniCDD3,  HANGUL SYLLABLE-CDD3
        '52692' => '16133'            , # G+16133  U+CDD4  uniCDD4,  HANGUL SYLLABLE-CDD4
        '52693' => '16134'            , # G+16134  U+CDD5  uniCDD5,  HANGUL SYLLABLE-CDD5
        '52694' => '16135'            , # G+16135  U+CDD6  uniCDD6,  HANGUL SYLLABLE-CDD6
        '52695' => '16136'            , # G+16136  U+CDD7  uniCDD7,  HANGUL SYLLABLE-CDD7
        '52696' => '16137'            , # G+16137  U+CDD8  uniCDD8,  HANGUL SYLLABLE-CDD8
        '52697' => '16138'            , # G+16138  U+CDD9  uniCDD9,  HANGUL SYLLABLE-CDD9
        '52698' => '16139'            , # G+16139  U+CDDA  uniCDDA,  HANGUL SYLLABLE-CDDA
        '52699' => '16140'            , # G+16140  U+CDDB  uniCDDB,  HANGUL SYLLABLE-CDDB
        '52700' => '16141'            , # G+16141  U+CDDC  uniCDDC,  HANGUL SYLLABLE-CDDC
        '52701' => '16142'            , # G+16142  U+CDDD  uniCDDD,  HANGUL SYLLABLE-CDDD
        '52702' => '16143'            , # G+16143  U+CDDE  uniCDDE,  HANGUL SYLLABLE-CDDE
        '52703' => '16144'            , # G+16144  U+CDDF  uniCDDF,  HANGUL SYLLABLE-CDDF
        '52704' => '16145'            , # G+16145  U+CDE0  uniCDE0,  HANGUL SYLLABLE-CDE0
        '52705' => '16146'            , # G+16146  U+CDE1  uniCDE1,  HANGUL SYLLABLE-CDE1
        '52706' => '16147'            , # G+16147  U+CDE2  uniCDE2,  HANGUL SYLLABLE-CDE2
        '52707' => '16148'            , # G+16148  U+CDE3  uniCDE3,  HANGUL SYLLABLE-CDE3
        '52708' => '16149'            , # G+16149  U+CDE4  uniCDE4,  HANGUL SYLLABLE-CDE4
        '52709' => '16150'            , # G+16150  U+CDE5  uniCDE5,  HANGUL SYLLABLE-CDE5
        '52710' => '16151'            , # G+16151  U+CDE6  uniCDE6,  HANGUL SYLLABLE-CDE6
        '52711' => '16152'            , # G+16152  U+CDE7  uniCDE7,  HANGUL SYLLABLE-CDE7
        '52712' => '2946'             , # G+2946  U+CDE8  uniCDE8,  HANGUL SYLLABLE-CDE8
        '52713' => '16153'            , # G+16153  U+CDE9  uniCDE9,  HANGUL SYLLABLE-CDE9
        '52714' => '16154'            , # G+16154  U+CDEA  uniCDEA,  HANGUL SYLLABLE-CDEA
        '52715' => '16155'            , # G+16155  U+CDEB  uniCDEB,  HANGUL SYLLABLE-CDEB
        '52716' => '2947'             , # G+2947  U+CDEC  uniCDEC,  HANGUL SYLLABLE-CDEC
        '52717' => '16156'            , # G+16156  U+CDED  uniCDED,  HANGUL SYLLABLE-CDED
        '52718' => '16157'            , # G+16157  U+CDEE  uniCDEE,  HANGUL SYLLABLE-CDEE
        '52719' => '16158'            , # G+16158  U+CDEF  uniCDEF,  HANGUL SYLLABLE-CDEF
        '52720' => '2948'             , # G+2948  U+CDF0  uniCDF0,  HANGUL SYLLABLE-CDF0
        '52721' => '16159'            , # G+16159  U+CDF1  uniCDF1,  HANGUL SYLLABLE-CDF1
        '52722' => '16160'            , # G+16160  U+CDF2  uniCDF2,  HANGUL SYLLABLE-CDF2
        '52723' => '16161'            , # G+16161  U+CDF3  uniCDF3,  HANGUL SYLLABLE-CDF3
        '52724' => '16162'            , # G+16162  U+CDF4  uniCDF4,  HANGUL SYLLABLE-CDF4
        '52725' => '16163'            , # G+16163  U+CDF5  uniCDF5,  HANGUL SYLLABLE-CDF5
        '52726' => '16164'            , # G+16164  U+CDF6  uniCDF6,  HANGUL SYLLABLE-CDF6
        '52727' => '16165'            , # G+16165  U+CDF7  uniCDF7,  HANGUL SYLLABLE-CDF7
        '52728' => '2949'             , # G+2949  U+CDF8  uniCDF8,  HANGUL SYLLABLE-CDF8
        '52729' => '2950'             , # G+2950  U+CDF9  uniCDF9,  HANGUL SYLLABLE-CDF9
        '52730' => '16166'            , # G+16166  U+CDFA  uniCDFA,  HANGUL SYLLABLE-CDFA
        '52731' => '2951'             , # G+2951  U+CDFB  uniCDFB,  HANGUL SYLLABLE-CDFB
        '52732' => '16167'            , # G+16167  U+CDFC  uniCDFC,  HANGUL SYLLABLE-CDFC
        '52733' => '2952'             , # G+2952  U+CDFD  uniCDFD,  HANGUL SYLLABLE-CDFD
        '52734' => '16168'            , # G+16168  U+CDFE  uniCDFE,  HANGUL SYLLABLE-CDFE
        '52735' => '16169'            , # G+16169  U+CDFF  uniCDFF,  HANGUL SYLLABLE-CDFF
        '52736' => '16170'            , # G+16170  U+CE00  uniCE00,  HANGUL SYLLABLE-CE00
        '52737' => '16171'            , # G+16171  U+CE01  uniCE01,  HANGUL SYLLABLE-CE01
        '52738' => '16172'            , # G+16172  U+CE02  uniCE02,  HANGUL SYLLABLE-CE02
        '52739' => '16173'            , # G+16173  U+CE03  uniCE03,  HANGUL SYLLABLE-CE03
        '52740' => '2953'             , # G+2953  U+CE04  uniCE04,  HANGUL SYLLABLE-CE04
        '52741' => '16174'            , # G+16174  U+CE05  uniCE05,  HANGUL SYLLABLE-CE05
        '52742' => '16175'            , # G+16175  U+CE06  uniCE06,  HANGUL SYLLABLE-CE06
        '52743' => '16176'            , # G+16176  U+CE07  uniCE07,  HANGUL SYLLABLE-CE07
        '52744' => '2954'             , # G+2954  U+CE08  uniCE08,  HANGUL SYLLABLE-CE08
        '52745' => '16177'            , # G+16177  U+CE09  uniCE09,  HANGUL SYLLABLE-CE09
        '52746' => '16178'            , # G+16178  U+CE0A  uniCE0A,  HANGUL SYLLABLE-CE0A
        '52747' => '16179'            , # G+16179  U+CE0B  uniCE0B,  HANGUL SYLLABLE-CE0B
        '52748' => '2955'             , # G+2955  U+CE0C  uniCE0C,  HANGUL SYLLABLE-CE0C
        '52749' => '16180'            , # G+16180  U+CE0D  uniCE0D,  HANGUL SYLLABLE-CE0D
        '52750' => '16181'            , # G+16181  U+CE0E  uniCE0E,  HANGUL SYLLABLE-CE0E
        '52751' => '16182'            , # G+16182  U+CE0F  uniCE0F,  HANGUL SYLLABLE-CE0F
        '52752' => '16183'            , # G+16183  U+CE10  uniCE10,  HANGUL SYLLABLE-CE10
        '52753' => '16184'            , # G+16184  U+CE11  uniCE11,  HANGUL SYLLABLE-CE11
        '52754' => '16185'            , # G+16185  U+CE12  uniCE12,  HANGUL SYLLABLE-CE12
        '52755' => '16186'            , # G+16186  U+CE13  uniCE13,  HANGUL SYLLABLE-CE13
        '52756' => '2956'             , # G+2956  U+CE14  uniCE14,  HANGUL SYLLABLE-CE14
        '52757' => '16187'            , # G+16187  U+CE15  uniCE15,  HANGUL SYLLABLE-CE15
        '52758' => '16188'            , # G+16188  U+CE16  uniCE16,  HANGUL SYLLABLE-CE16
        '52759' => '16189'            , # G+16189  U+CE17  uniCE17,  HANGUL SYLLABLE-CE17
        '52760' => '16190'            , # G+16190  U+CE18  uniCE18,  HANGUL SYLLABLE-CE18
        '52761' => '2957'             , # G+2957  U+CE19  uniCE19,  HANGUL SYLLABLE-CE19
        '52762' => '16191'            , # G+16191  U+CE1A  uniCE1A,  HANGUL SYLLABLE-CE1A
        '52763' => '16192'            , # G+16192  U+CE1B  uniCE1B,  HANGUL SYLLABLE-CE1B
        '52764' => '16193'            , # G+16193  U+CE1C  uniCE1C,  HANGUL SYLLABLE-CE1C
        '52765' => '16194'            , # G+16194  U+CE1D  uniCE1D,  HANGUL SYLLABLE-CE1D
        '52766' => '16195'            , # G+16195  U+CE1E  uniCE1E,  HANGUL SYLLABLE-CE1E
        '52767' => '16196'            , # G+16196  U+CE1F  uniCE1F,  HANGUL SYLLABLE-CE1F
        '52768' => '2958'             , # G+2958  U+CE20  uniCE20,  HANGUL SYLLABLE-CE20
        '52769' => '2959'             , # G+2959  U+CE21  uniCE21,  HANGUL SYLLABLE-CE21
        '52770' => '16197'            , # G+16197  U+CE22  uniCE22,  HANGUL SYLLABLE-CE22
        '52771' => '16198'            , # G+16198  U+CE23  uniCE23,  HANGUL SYLLABLE-CE23
        '52772' => '2960'             , # G+2960  U+CE24  uniCE24,  HANGUL SYLLABLE-CE24
        '52773' => '16199'            , # G+16199  U+CE25  uniCE25,  HANGUL SYLLABLE-CE25
        '52774' => '16200'            , # G+16200  U+CE26  uniCE26,  HANGUL SYLLABLE-CE26
        '52775' => '16201'            , # G+16201  U+CE27  uniCE27,  HANGUL SYLLABLE-CE27
        '52776' => '2961'             , # G+2961  U+CE28  uniCE28,  HANGUL SYLLABLE-CE28
        '52777' => '16202'            , # G+16202  U+CE29  uniCE29,  HANGUL SYLLABLE-CE29
        '52778' => '16203'            , # G+16203  U+CE2A  uniCE2A,  HANGUL SYLLABLE-CE2A
        '52779' => '16204'            , # G+16204  U+CE2B  uniCE2B,  HANGUL SYLLABLE-CE2B
        '52780' => '16205'            , # G+16205  U+CE2C  uniCE2C,  HANGUL SYLLABLE-CE2C
        '52781' => '16206'            , # G+16206  U+CE2D  uniCE2D,  HANGUL SYLLABLE-CE2D
        '52782' => '16207'            , # G+16207  U+CE2E  uniCE2E,  HANGUL SYLLABLE-CE2E
        '52783' => '16208'            , # G+16208  U+CE2F  uniCE2F,  HANGUL SYLLABLE-CE2F
        '52784' => '2962'             , # G+2962  U+CE30  uniCE30,  HANGUL SYLLABLE-CE30
        '52785' => '2963'             , # G+2963  U+CE31  uniCE31,  HANGUL SYLLABLE-CE31
        '52786' => '16209'            , # G+16209  U+CE32  uniCE32,  HANGUL SYLLABLE-CE32
        '52787' => '2964'             , # G+2964  U+CE33  uniCE33,  HANGUL SYLLABLE-CE33
        '52788' => '16210'            , # G+16210  U+CE34  uniCE34,  HANGUL SYLLABLE-CE34
        '52789' => '2965'             , # G+2965  U+CE35  uniCE35,  HANGUL SYLLABLE-CE35
        '52790' => '16211'            , # G+16211  U+CE36  uniCE36,  HANGUL SYLLABLE-CE36
        '52791' => '16212'            , # G+16212  U+CE37  uniCE37,  HANGUL SYLLABLE-CE37
        '52792' => '16213'            , # G+16213  U+CE38  uniCE38,  HANGUL SYLLABLE-CE38
        '52793' => '16214'            , # G+16214  U+CE39  uniCE39,  HANGUL SYLLABLE-CE39
        '52794' => '16215'            , # G+16215  U+CE3A  uniCE3A,  HANGUL SYLLABLE-CE3A
        '52795' => '16216'            , # G+16216  U+CE3B  uniCE3B,  HANGUL SYLLABLE-CE3B
        '52796' => '16217'            , # G+16217  U+CE3C  uniCE3C,  HANGUL SYLLABLE-CE3C
        '52797' => '16218'            , # G+16218  U+CE3D  uniCE3D,  HANGUL SYLLABLE-CE3D
        '52798' => '16219'            , # G+16219  U+CE3E  uniCE3E,  HANGUL SYLLABLE-CE3E
        '52799' => '16220'            , # G+16220  U+CE3F  uniCE3F,  HANGUL SYLLABLE-CE3F
        '52800' => '16221'            , # G+16221  U+CE40  uniCE40,  HANGUL SYLLABLE-CE40
        '52801' => '16222'            , # G+16222  U+CE41  uniCE41,  HANGUL SYLLABLE-CE41
        '52802' => '16223'            , # G+16223  U+CE42  uniCE42,  HANGUL SYLLABLE-CE42
        '52803' => '16224'            , # G+16224  U+CE43  uniCE43,  HANGUL SYLLABLE-CE43
        '52804' => '16225'            , # G+16225  U+CE44  uniCE44,  HANGUL SYLLABLE-CE44
        '52805' => '16226'            , # G+16226  U+CE45  uniCE45,  HANGUL SYLLABLE-CE45
        '52806' => '16227'            , # G+16227  U+CE46  uniCE46,  HANGUL SYLLABLE-CE46
        '52807' => '16228'            , # G+16228  U+CE47  uniCE47,  HANGUL SYLLABLE-CE47
        '52808' => '16229'            , # G+16229  U+CE48  uniCE48,  HANGUL SYLLABLE-CE48
        '52809' => '16230'            , # G+16230  U+CE49  uniCE49,  HANGUL SYLLABLE-CE49
        '52810' => '16231'            , # G+16231  U+CE4A  uniCE4A,  HANGUL SYLLABLE-CE4A
        '52811' => '16232'            , # G+16232  U+CE4B  uniCE4B,  HANGUL SYLLABLE-CE4B
        '52812' => '16233'            , # G+16233  U+CE4C  uniCE4C,  HANGUL SYLLABLE-CE4C
        '52813' => '16234'            , # G+16234  U+CE4D  uniCE4D,  HANGUL SYLLABLE-CE4D
        '52814' => '16235'            , # G+16235  U+CE4E  uniCE4E,  HANGUL SYLLABLE-CE4E
        '52815' => '16236'            , # G+16236  U+CE4F  uniCE4F,  HANGUL SYLLABLE-CE4F
        '52816' => '16237'            , # G+16237  U+CE50  uniCE50,  HANGUL SYLLABLE-CE50
        '52817' => '16238'            , # G+16238  U+CE51  uniCE51,  HANGUL SYLLABLE-CE51
        '52818' => '16239'            , # G+16239  U+CE52  uniCE52,  HANGUL SYLLABLE-CE52
        '52819' => '16240'            , # G+16240  U+CE53  uniCE53,  HANGUL SYLLABLE-CE53
        '52820' => '16241'            , # G+16241  U+CE54  uniCE54,  HANGUL SYLLABLE-CE54
        '52821' => '16242'            , # G+16242  U+CE55  uniCE55,  HANGUL SYLLABLE-CE55
        '52822' => '16243'            , # G+16243  U+CE56  uniCE56,  HANGUL SYLLABLE-CE56
        '52823' => '16244'            , # G+16244  U+CE57  uniCE57,  HANGUL SYLLABLE-CE57
        '52824' => '2966'             , # G+2966  U+CE58  uniCE58,  HANGUL SYLLABLE-CE58
        '52825' => '2967'             , # G+2967  U+CE59  uniCE59,  HANGUL SYLLABLE-CE59
        '52826' => '16245'            , # G+16245  U+CE5A  uniCE5A,  HANGUL SYLLABLE-CE5A
        '52827' => '16246'            , # G+16246  U+CE5B  uniCE5B,  HANGUL SYLLABLE-CE5B
        '52828' => '2968'             , # G+2968  U+CE5C  uniCE5C,  HANGUL SYLLABLE-CE5C
        '52829' => '16247'            , # G+16247  U+CE5D  uniCE5D,  HANGUL SYLLABLE-CE5D
        '52830' => '16248'            , # G+16248  U+CE5E  uniCE5E,  HANGUL SYLLABLE-CE5E
        '52831' => '2969'             , # G+2969  U+CE5F  uniCE5F,  HANGUL SYLLABLE-CE5F
        '52832' => '2970'             , # G+2970  U+CE60  uniCE60,  HANGUL SYLLABLE-CE60
        '52833' => '2971'             , # G+2971  U+CE61  uniCE61,  HANGUL SYLLABLE-CE61
        '52834' => '16249'            , # G+16249  U+CE62  uniCE62,  HANGUL SYLLABLE-CE62
        '52835' => '16250'            , # G+16250  U+CE63  uniCE63,  HANGUL SYLLABLE-CE63
        '52836' => '16251'            , # G+16251  U+CE64  uniCE64,  HANGUL SYLLABLE-CE64
        '52837' => '16252'            , # G+16252  U+CE65  uniCE65,  HANGUL SYLLABLE-CE65
        '52838' => '16253'            , # G+16253  U+CE66  uniCE66,  HANGUL SYLLABLE-CE66
        '52839' => '16254'            , # G+16254  U+CE67  uniCE67,  HANGUL SYLLABLE-CE67
        '52840' => '2972'             , # G+2972  U+CE68  uniCE68,  HANGUL SYLLABLE-CE68
        '52841' => '2973'             , # G+2973  U+CE69  uniCE69,  HANGUL SYLLABLE-CE69
        '52842' => '16255'            , # G+16255  U+CE6A  uniCE6A,  HANGUL SYLLABLE-CE6A
        '52843' => '2974'             , # G+2974  U+CE6B  uniCE6B,  HANGUL SYLLABLE-CE6B
        '52844' => '16256'            , # G+16256  U+CE6C  uniCE6C,  HANGUL SYLLABLE-CE6C
        '52845' => '2975'             , # G+2975  U+CE6D  uniCE6D,  HANGUL SYLLABLE-CE6D
        '52846' => '16257'            , # G+16257  U+CE6E  uniCE6E,  HANGUL SYLLABLE-CE6E
        '52847' => '16258'            , # G+16258  U+CE6F  uniCE6F,  HANGUL SYLLABLE-CE6F
        '52848' => '16259'            , # G+16259  U+CE70  uniCE70,  HANGUL SYLLABLE-CE70
        '52849' => '16260'            , # G+16260  U+CE71  uniCE71,  HANGUL SYLLABLE-CE71
        '52850' => '16261'            , # G+16261  U+CE72  uniCE72,  HANGUL SYLLABLE-CE72
        '52851' => '16262'            , # G+16262  U+CE73  uniCE73,  HANGUL SYLLABLE-CE73
        '52852' => '2976'             , # G+2976  U+CE74  uniCE74,  HANGUL SYLLABLE-CE74
        '52853' => '2977'             , # G+2977  U+CE75  uniCE75,  HANGUL SYLLABLE-CE75
        '52854' => '16263'            , # G+16263  U+CE76  uniCE76,  HANGUL SYLLABLE-CE76
        '52855' => '16264'            , # G+16264  U+CE77  uniCE77,  HANGUL SYLLABLE-CE77
        '52856' => '2978'             , # G+2978  U+CE78  uniCE78,  HANGUL SYLLABLE-CE78
        '52857' => '16265'            , # G+16265  U+CE79  uniCE79,  HANGUL SYLLABLE-CE79
        '52858' => '16266'            , # G+16266  U+CE7A  uniCE7A,  HANGUL SYLLABLE-CE7A
        '52859' => '16267'            , # G+16267  U+CE7B  uniCE7B,  HANGUL SYLLABLE-CE7B
        '52860' => '2979'             , # G+2979  U+CE7C  uniCE7C,  HANGUL SYLLABLE-CE7C
        '52861' => '16268'            , # G+16268  U+CE7D  uniCE7D,  HANGUL SYLLABLE-CE7D
        '52862' => '16269'            , # G+16269  U+CE7E  uniCE7E,  HANGUL SYLLABLE-CE7E
        '52863' => '16270'            , # G+16270  U+CE7F  uniCE7F,  HANGUL SYLLABLE-CE7F
        '52864' => '16271'            , # G+16271  U+CE80  uniCE80,  HANGUL SYLLABLE-CE80
        '52865' => '16272'            , # G+16272  U+CE81  uniCE81,  HANGUL SYLLABLE-CE81
        '52866' => '16273'            , # G+16273  U+CE82  uniCE82,  HANGUL SYLLABLE-CE82
        '52867' => '16274'            , # G+16274  U+CE83  uniCE83,  HANGUL SYLLABLE-CE83
        '52868' => '2980'             , # G+2980  U+CE84  uniCE84,  HANGUL SYLLABLE-CE84
        '52869' => '2981'             , # G+2981  U+CE85  uniCE85,  HANGUL SYLLABLE-CE85
        '52870' => '16275'            , # G+16275  U+CE86  uniCE86,  HANGUL SYLLABLE-CE86
        '52871' => '2982'             , # G+2982  U+CE87  uniCE87,  HANGUL SYLLABLE-CE87
        '52872' => '16276'            , # G+16276  U+CE88  uniCE88,  HANGUL SYLLABLE-CE88
        '52873' => '2983'             , # G+2983  U+CE89  uniCE89,  HANGUL SYLLABLE-CE89
        '52874' => '16277'            , # G+16277  U+CE8A  uniCE8A,  HANGUL SYLLABLE-CE8A
        '52875' => '16278'            , # G+16278  U+CE8B  uniCE8B,  HANGUL SYLLABLE-CE8B
        '52876' => '16279'            , # G+16279  U+CE8C  uniCE8C,  HANGUL SYLLABLE-CE8C
        '52877' => '16280'            , # G+16280  U+CE8D  uniCE8D,  HANGUL SYLLABLE-CE8D
        '52878' => '16281'            , # G+16281  U+CE8E  uniCE8E,  HANGUL SYLLABLE-CE8E
        '52879' => '16282'            , # G+16282  U+CE8F  uniCE8F,  HANGUL SYLLABLE-CE8F
        '52880' => '2984'             , # G+2984  U+CE90  uniCE90,  HANGUL SYLLABLE-CE90
        '52881' => '2985'             , # G+2985  U+CE91  uniCE91,  HANGUL SYLLABLE-CE91
        '52882' => '16283'            , # G+16283  U+CE92  uniCE92,  HANGUL SYLLABLE-CE92
        '52883' => '16284'            , # G+16284  U+CE93  uniCE93,  HANGUL SYLLABLE-CE93
        '52884' => '2986'             , # G+2986  U+CE94  uniCE94,  HANGUL SYLLABLE-CE94
        '52885' => '16285'            , # G+16285  U+CE95  uniCE95,  HANGUL SYLLABLE-CE95
        '52886' => '16286'            , # G+16286  U+CE96  uniCE96,  HANGUL SYLLABLE-CE96
        '52887' => '16287'            , # G+16287  U+CE97  uniCE97,  HANGUL SYLLABLE-CE97
        '52888' => '2987'             , # G+2987  U+CE98  uniCE98,  HANGUL SYLLABLE-CE98
        '52889' => '16288'            , # G+16288  U+CE99  uniCE99,  HANGUL SYLLABLE-CE99
        '52890' => '16289'            , # G+16289  U+CE9A  uniCE9A,  HANGUL SYLLABLE-CE9A
        '52891' => '16290'            , # G+16290  U+CE9B  uniCE9B,  HANGUL SYLLABLE-CE9B
        '52892' => '16291'            , # G+16291  U+CE9C  uniCE9C,  HANGUL SYLLABLE-CE9C
        '52893' => '16292'            , # G+16292  U+CE9D  uniCE9D,  HANGUL SYLLABLE-CE9D
        '52894' => '16293'            , # G+16293  U+CE9E  uniCE9E,  HANGUL SYLLABLE-CE9E
        '52895' => '16294'            , # G+16294  U+CE9F  uniCE9F,  HANGUL SYLLABLE-CE9F
        '52896' => '2988'             , # G+2988  U+CEA0  uniCEA0,  HANGUL SYLLABLE-CEA0
        '52897' => '2989'             , # G+2989  U+CEA1  uniCEA1,  HANGUL SYLLABLE-CEA1
        '52898' => '16295'            , # G+16295  U+CEA2  uniCEA2,  HANGUL SYLLABLE-CEA2
        '52899' => '2990'             , # G+2990  U+CEA3  uniCEA3,  HANGUL SYLLABLE-CEA3
        '52900' => '2991'             , # G+2991  U+CEA4  uniCEA4,  HANGUL SYLLABLE-CEA4
        '52901' => '2992'             , # G+2992  U+CEA5  uniCEA5,  HANGUL SYLLABLE-CEA5
        '52902' => '16296'            , # G+16296  U+CEA6  uniCEA6,  HANGUL SYLLABLE-CEA6
        '52903' => '16297'            , # G+16297  U+CEA7  uniCEA7,  HANGUL SYLLABLE-CEA7
        '52904' => '16298'            , # G+16298  U+CEA8  uniCEA8,  HANGUL SYLLABLE-CEA8
        '52905' => '16299'            , # G+16299  U+CEA9  uniCEA9,  HANGUL SYLLABLE-CEA9
        '52906' => '16300'            , # G+16300  U+CEAA  uniCEAA,  HANGUL SYLLABLE-CEAA
        '52907' => '16301'            , # G+16301  U+CEAB  uniCEAB,  HANGUL SYLLABLE-CEAB
        '52908' => '2993'             , # G+2993  U+CEAC  uniCEAC,  HANGUL SYLLABLE-CEAC
        '52909' => '2994'             , # G+2994  U+CEAD  uniCEAD,  HANGUL SYLLABLE-CEAD
        '52910' => '16302'            , # G+16302  U+CEAE  uniCEAE,  HANGUL SYLLABLE-CEAE
        '52911' => '16303'            , # G+16303  U+CEAF  uniCEAF,  HANGUL SYLLABLE-CEAF
        '52912' => '16304'            , # G+16304  U+CEB0  uniCEB0,  HANGUL SYLLABLE-CEB0
        '52913' => '16305'            , # G+16305  U+CEB1  uniCEB1,  HANGUL SYLLABLE-CEB1
        '52914' => '16306'            , # G+16306  U+CEB2  uniCEB2,  HANGUL SYLLABLE-CEB2
        '52915' => '16307'            , # G+16307  U+CEB3  uniCEB3,  HANGUL SYLLABLE-CEB3
        '52916' => '16308'            , # G+16308  U+CEB4  uniCEB4,  HANGUL SYLLABLE-CEB4
        '52917' => '16309'            , # G+16309  U+CEB5  uniCEB5,  HANGUL SYLLABLE-CEB5
        '52918' => '16310'            , # G+16310  U+CEB6  uniCEB6,  HANGUL SYLLABLE-CEB6
        '52919' => '16311'            , # G+16311  U+CEB7  uniCEB7,  HANGUL SYLLABLE-CEB7
        '52920' => '16312'            , # G+16312  U+CEB8  uniCEB8,  HANGUL SYLLABLE-CEB8
        '52921' => '16313'            , # G+16313  U+CEB9  uniCEB9,  HANGUL SYLLABLE-CEB9
        '52922' => '16314'            , # G+16314  U+CEBA  uniCEBA,  HANGUL SYLLABLE-CEBA
        '52923' => '16315'            , # G+16315  U+CEBB  uniCEBB,  HANGUL SYLLABLE-CEBB
        '52924' => '16316'            , # G+16316  U+CEBC  uniCEBC,  HANGUL SYLLABLE-CEBC
        '52925' => '16317'            , # G+16317  U+CEBD  uniCEBD,  HANGUL SYLLABLE-CEBD
        '52926' => '16318'            , # G+16318  U+CEBE  uniCEBE,  HANGUL SYLLABLE-CEBE
        '52927' => '16319'            , # G+16319  U+CEBF  uniCEBF,  HANGUL SYLLABLE-CEBF
        '52928' => '16320'            , # G+16320  U+CEC0  uniCEC0,  HANGUL SYLLABLE-CEC0
        '52929' => '2995'             , # G+2995  U+CEC1  uniCEC1,  HANGUL SYLLABLE-CEC1
        '52930' => '16321'            , # G+16321  U+CEC2  uniCEC2,  HANGUL SYLLABLE-CEC2
        '52931' => '16322'            , # G+16322  U+CEC3  uniCEC3,  HANGUL SYLLABLE-CEC3
        '52932' => '16323'            , # G+16323  U+CEC4  uniCEC4,  HANGUL SYLLABLE-CEC4
        '52933' => '16324'            , # G+16324  U+CEC5  uniCEC5,  HANGUL SYLLABLE-CEC5
        '52934' => '16325'            , # G+16325  U+CEC6  uniCEC6,  HANGUL SYLLABLE-CEC6
        '52935' => '16326'            , # G+16326  U+CEC7  uniCEC7,  HANGUL SYLLABLE-CEC7
        '52936' => '16327'            , # G+16327  U+CEC8  uniCEC8,  HANGUL SYLLABLE-CEC8
        '52937' => '16328'            , # G+16328  U+CEC9  uniCEC9,  HANGUL SYLLABLE-CEC9
        '52938' => '16329'            , # G+16329  U+CECA  uniCECA,  HANGUL SYLLABLE-CECA
        '52939' => '16330'            , # G+16330  U+CECB  uniCECB,  HANGUL SYLLABLE-CECB
        '52940' => '16331'            , # G+16331  U+CECC  uniCECC,  HANGUL SYLLABLE-CECC
        '52941' => '16332'            , # G+16332  U+CECD  uniCECD,  HANGUL SYLLABLE-CECD
        '52942' => '16333'            , # G+16333  U+CECE  uniCECE,  HANGUL SYLLABLE-CECE
        '52943' => '16334'            , # G+16334  U+CECF  uniCECF,  HANGUL SYLLABLE-CECF
        '52944' => '16335'            , # G+16335  U+CED0  uniCED0,  HANGUL SYLLABLE-CED0
        '52945' => '16336'            , # G+16336  U+CED1  uniCED1,  HANGUL SYLLABLE-CED1
        '52946' => '16337'            , # G+16337  U+CED2  uniCED2,  HANGUL SYLLABLE-CED2
        '52947' => '16338'            , # G+16338  U+CED3  uniCED3,  HANGUL SYLLABLE-CED3
        '52948' => '16339'            , # G+16339  U+CED4  uniCED4,  HANGUL SYLLABLE-CED4
        '52949' => '16340'            , # G+16340  U+CED5  uniCED5,  HANGUL SYLLABLE-CED5
        '52950' => '16341'            , # G+16341  U+CED6  uniCED6,  HANGUL SYLLABLE-CED6
        '52951' => '16342'            , # G+16342  U+CED7  uniCED7,  HANGUL SYLLABLE-CED7
        '52952' => '16343'            , # G+16343  U+CED8  uniCED8,  HANGUL SYLLABLE-CED8
        '52953' => '16344'            , # G+16344  U+CED9  uniCED9,  HANGUL SYLLABLE-CED9
        '52954' => '16345'            , # G+16345  U+CEDA  uniCEDA,  HANGUL SYLLABLE-CEDA
        '52955' => '16346'            , # G+16346  U+CEDB  uniCEDB,  HANGUL SYLLABLE-CEDB
        '52956' => '16347'            , # G+16347  U+CEDC  uniCEDC,  HANGUL SYLLABLE-CEDC
        '52957' => '16348'            , # G+16348  U+CEDD  uniCEDD,  HANGUL SYLLABLE-CEDD
        '52958' => '16349'            , # G+16349  U+CEDE  uniCEDE,  HANGUL SYLLABLE-CEDE
        '52959' => '16350'            , # G+16350  U+CEDF  uniCEDF,  HANGUL SYLLABLE-CEDF
        '52960' => '16351'            , # G+16351  U+CEE0  uniCEE0,  HANGUL SYLLABLE-CEE0
        '52961' => '16352'            , # G+16352  U+CEE1  uniCEE1,  HANGUL SYLLABLE-CEE1
        '52962' => '16353'            , # G+16353  U+CEE2  uniCEE2,  HANGUL SYLLABLE-CEE2
        '52963' => '16354'            , # G+16354  U+CEE3  uniCEE3,  HANGUL SYLLABLE-CEE3
        '52964' => '2996'             , # G+2996  U+CEE4  uniCEE4,  HANGUL SYLLABLE-CEE4
        '52965' => '2997'             , # G+2997  U+CEE5  uniCEE5,  HANGUL SYLLABLE-CEE5
        '52966' => '16355'            , # G+16355  U+CEE6  uniCEE6,  HANGUL SYLLABLE-CEE6
        '52967' => '16356'            , # G+16356  U+CEE7  uniCEE7,  HANGUL SYLLABLE-CEE7
        '52968' => '2998'             , # G+2998  U+CEE8  uniCEE8,  HANGUL SYLLABLE-CEE8
        '52969' => '16357'            , # G+16357  U+CEE9  uniCEE9,  HANGUL SYLLABLE-CEE9
        '52970' => '16358'            , # G+16358  U+CEEA  uniCEEA,  HANGUL SYLLABLE-CEEA
        '52971' => '2999'             , # G+2999  U+CEEB  uniCEEB,  HANGUL SYLLABLE-CEEB
        '52972' => '3000'             , # G+3000  U+CEEC  uniCEEC,  HANGUL SYLLABLE-CEEC
        '52973' => '16359'            , # G+16359  U+CEED  uniCEED,  HANGUL SYLLABLE-CEED
        '52974' => '16360'            , # G+16360  U+CEEE  uniCEEE,  HANGUL SYLLABLE-CEEE
        '52975' => '16361'            , # G+16361  U+CEEF  uniCEEF,  HANGUL SYLLABLE-CEEF
        '52976' => '16362'            , # G+16362  U+CEF0  uniCEF0,  HANGUL SYLLABLE-CEF0
        '52977' => '16363'            , # G+16363  U+CEF1  uniCEF1,  HANGUL SYLLABLE-CEF1
        '52978' => '16364'            , # G+16364  U+CEF2  uniCEF2,  HANGUL SYLLABLE-CEF2
        '52979' => '16365'            , # G+16365  U+CEF3  uniCEF3,  HANGUL SYLLABLE-CEF3
        '52980' => '3001'             , # G+3001  U+CEF4  uniCEF4,  HANGUL SYLLABLE-CEF4
        '52981' => '3002'             , # G+3002  U+CEF5  uniCEF5,  HANGUL SYLLABLE-CEF5
        '52982' => '16366'            , # G+16366  U+CEF6  uniCEF6,  HANGUL SYLLABLE-CEF6
        '52983' => '3003'             , # G+3003  U+CEF7  uniCEF7,  HANGUL SYLLABLE-CEF7
        '52984' => '3004'             , # G+3004  U+CEF8  uniCEF8,  HANGUL SYLLABLE-CEF8
        '52985' => '3005'             , # G+3005  U+CEF9  uniCEF9,  HANGUL SYLLABLE-CEF9
        '52986' => '16367'            , # G+16367  U+CEFA  uniCEFA,  HANGUL SYLLABLE-CEFA
        '52987' => '16368'            , # G+16368  U+CEFB  uniCEFB,  HANGUL SYLLABLE-CEFB
        '52988' => '16369'            , # G+16369  U+CEFC  uniCEFC,  HANGUL SYLLABLE-CEFC
        '52989' => '16370'            , # G+16370  U+CEFD  uniCEFD,  HANGUL SYLLABLE-CEFD
        '52990' => '16371'            , # G+16371  U+CEFE  uniCEFE,  HANGUL SYLLABLE-CEFE
        '52991' => '16372'            , # G+16372  U+CEFF  uniCEFF,  HANGUL SYLLABLE-CEFF
        '52992' => '3006'             , # G+3006  U+CF00  uniCF00,  HANGUL SYLLABLE-CF00
        '52993' => '3007'             , # G+3007  U+CF01  uniCF01,  HANGUL SYLLABLE-CF01
        '52994' => '16373'            , # G+16373  U+CF02  uniCF02,  HANGUL SYLLABLE-CF02
        '52995' => '16374'            , # G+16374  U+CF03  uniCF03,  HANGUL SYLLABLE-CF03
        '52996' => '3008'             , # G+3008  U+CF04  uniCF04,  HANGUL SYLLABLE-CF04
        '52997' => '16375'            , # G+16375  U+CF05  uniCF05,  HANGUL SYLLABLE-CF05
        '52998' => '16376'            , # G+16376  U+CF06  uniCF06,  HANGUL SYLLABLE-CF06
        '52999' => '16377'            , # G+16377  U+CF07  uniCF07,  HANGUL SYLLABLE-CF07
        '53000' => '3009'             , # G+3009  U+CF08  uniCF08,  HANGUL SYLLABLE-CF08
        '53001' => '16378'            , # G+16378  U+CF09  uniCF09,  HANGUL SYLLABLE-CF09
        '53002' => '16379'            , # G+16379  U+CF0A  uniCF0A,  HANGUL SYLLABLE-CF0A
        '53003' => '16380'            , # G+16380  U+CF0B  uniCF0B,  HANGUL SYLLABLE-CF0B
        '53004' => '16381'            , # G+16381  U+CF0C  uniCF0C,  HANGUL SYLLABLE-CF0C
        '53005' => '16382'            , # G+16382  U+CF0D  uniCF0D,  HANGUL SYLLABLE-CF0D
        '53006' => '16383'            , # G+16383  U+CF0E  uniCF0E,  HANGUL SYLLABLE-CF0E
        '53007' => '16384'            , # G+16384  U+CF0F  uniCF0F,  HANGUL SYLLABLE-CF0F
        '53008' => '3010'             , # G+3010  U+CF10  uniCF10,  HANGUL SYLLABLE-CF10
        '53009' => '3011'             , # G+3011  U+CF11  uniCF11,  HANGUL SYLLABLE-CF11
        '53010' => '16385'            , # G+16385  U+CF12  uniCF12,  HANGUL SYLLABLE-CF12
        '53011' => '3012'             , # G+3012  U+CF13  uniCF13,  HANGUL SYLLABLE-CF13
        '53012' => '16386'            , # G+16386  U+CF14  uniCF14,  HANGUL SYLLABLE-CF14
        '53013' => '3013'             , # G+3013  U+CF15  uniCF15,  HANGUL SYLLABLE-CF15
        '53014' => '16387'            , # G+16387  U+CF16  uniCF16,  HANGUL SYLLABLE-CF16
        '53015' => '16388'            , # G+16388  U+CF17  uniCF17,  HANGUL SYLLABLE-CF17
        '53016' => '16389'            , # G+16389  U+CF18  uniCF18,  HANGUL SYLLABLE-CF18
        '53017' => '16390'            , # G+16390  U+CF19  uniCF19,  HANGUL SYLLABLE-CF19
        '53018' => '16391'            , # G+16391  U+CF1A  uniCF1A,  HANGUL SYLLABLE-CF1A
        '53019' => '16392'            , # G+16392  U+CF1B  uniCF1B,  HANGUL SYLLABLE-CF1B
        '53020' => '3014'             , # G+3014  U+CF1C  uniCF1C,  HANGUL SYLLABLE-CF1C
        '53021' => '16393'            , # G+16393  U+CF1D  uniCF1D,  HANGUL SYLLABLE-CF1D
        '53022' => '16394'            , # G+16394  U+CF1E  uniCF1E,  HANGUL SYLLABLE-CF1E
        '53023' => '16395'            , # G+16395  U+CF1F  uniCF1F,  HANGUL SYLLABLE-CF1F
        '53024' => '3015'             , # G+3015  U+CF20  uniCF20,  HANGUL SYLLABLE-CF20
        '53025' => '16396'            , # G+16396  U+CF21  uniCF21,  HANGUL SYLLABLE-CF21
        '53026' => '16397'            , # G+16397  U+CF22  uniCF22,  HANGUL SYLLABLE-CF22
        '53027' => '16398'            , # G+16398  U+CF23  uniCF23,  HANGUL SYLLABLE-CF23
        '53028' => '3016'             , # G+3016  U+CF24  uniCF24,  HANGUL SYLLABLE-CF24
        '53029' => '16399'            , # G+16399  U+CF25  uniCF25,  HANGUL SYLLABLE-CF25
        '53030' => '16400'            , # G+16400  U+CF26  uniCF26,  HANGUL SYLLABLE-CF26
        '53031' => '16401'            , # G+16401  U+CF27  uniCF27,  HANGUL SYLLABLE-CF27
        '53032' => '16402'            , # G+16402  U+CF28  uniCF28,  HANGUL SYLLABLE-CF28
        '53033' => '16403'            , # G+16403  U+CF29  uniCF29,  HANGUL SYLLABLE-CF29
        '53034' => '16404'            , # G+16404  U+CF2A  uniCF2A,  HANGUL SYLLABLE-CF2A
        '53035' => '16405'            , # G+16405  U+CF2B  uniCF2B,  HANGUL SYLLABLE-CF2B
        '53036' => '3017'             , # G+3017  U+CF2C  uniCF2C,  HANGUL SYLLABLE-CF2C
        '53037' => '3018'             , # G+3018  U+CF2D  uniCF2D,  HANGUL SYLLABLE-CF2D
        '53038' => '16406'            , # G+16406  U+CF2E  uniCF2E,  HANGUL SYLLABLE-CF2E
        '53039' => '3019'             , # G+3019  U+CF2F  uniCF2F,  HANGUL SYLLABLE-CF2F
        '53040' => '3020'             , # G+3020  U+CF30  uniCF30,  HANGUL SYLLABLE-CF30
        '53041' => '3021'             , # G+3021  U+CF31  uniCF31,  HANGUL SYLLABLE-CF31
        '53042' => '16407'            , # G+16407  U+CF32  uniCF32,  HANGUL SYLLABLE-CF32
        '53043' => '16408'            , # G+16408  U+CF33  uniCF33,  HANGUL SYLLABLE-CF33
        '53044' => '16409'            , # G+16409  U+CF34  uniCF34,  HANGUL SYLLABLE-CF34
        '53045' => '16410'            , # G+16410  U+CF35  uniCF35,  HANGUL SYLLABLE-CF35
        '53046' => '16411'            , # G+16411  U+CF36  uniCF36,  HANGUL SYLLABLE-CF36
        '53047' => '16412'            , # G+16412  U+CF37  uniCF37,  HANGUL SYLLABLE-CF37
        '53048' => '3022'             , # G+3022  U+CF38  uniCF38,  HANGUL SYLLABLE-CF38
        '53049' => '16413'            , # G+16413  U+CF39  uniCF39,  HANGUL SYLLABLE-CF39
        '53050' => '16414'            , # G+16414  U+CF3A  uniCF3A,  HANGUL SYLLABLE-CF3A
        '53051' => '16415'            , # G+16415  U+CF3B  uniCF3B,  HANGUL SYLLABLE-CF3B
        '53052' => '16416'            , # G+16416  U+CF3C  uniCF3C,  HANGUL SYLLABLE-CF3C
        '53053' => '16417'            , # G+16417  U+CF3D  uniCF3D,  HANGUL SYLLABLE-CF3D
        '53054' => '16418'            , # G+16418  U+CF3E  uniCF3E,  HANGUL SYLLABLE-CF3E
        '53055' => '16419'            , # G+16419  U+CF3F  uniCF3F,  HANGUL SYLLABLE-CF3F
        '53056' => '16420'            , # G+16420  U+CF40  uniCF40,  HANGUL SYLLABLE-CF40
        '53057' => '16421'            , # G+16421  U+CF41  uniCF41,  HANGUL SYLLABLE-CF41
        '53058' => '16422'            , # G+16422  U+CF42  uniCF42,  HANGUL SYLLABLE-CF42
        '53059' => '16423'            , # G+16423  U+CF43  uniCF43,  HANGUL SYLLABLE-CF43
        '53060' => '16424'            , # G+16424  U+CF44  uniCF44,  HANGUL SYLLABLE-CF44
        '53061' => '16425'            , # G+16425  U+CF45  uniCF45,  HANGUL SYLLABLE-CF45
        '53062' => '16426'            , # G+16426  U+CF46  uniCF46,  HANGUL SYLLABLE-CF46
        '53063' => '16427'            , # G+16427  U+CF47  uniCF47,  HANGUL SYLLABLE-CF47
        '53064' => '16428'            , # G+16428  U+CF48  uniCF48,  HANGUL SYLLABLE-CF48
        '53065' => '16429'            , # G+16429  U+CF49  uniCF49,  HANGUL SYLLABLE-CF49
        '53066' => '16430'            , # G+16430  U+CF4A  uniCF4A,  HANGUL SYLLABLE-CF4A
        '53067' => '16431'            , # G+16431  U+CF4B  uniCF4B,  HANGUL SYLLABLE-CF4B
        '53068' => '16432'            , # G+16432  U+CF4C  uniCF4C,  HANGUL SYLLABLE-CF4C
        '53069' => '16433'            , # G+16433  U+CF4D  uniCF4D,  HANGUL SYLLABLE-CF4D
        '53070' => '16434'            , # G+16434  U+CF4E  uniCF4E,  HANGUL SYLLABLE-CF4E
        '53071' => '16435'            , # G+16435  U+CF4F  uniCF4F,  HANGUL SYLLABLE-CF4F
        '53072' => '16436'            , # G+16436  U+CF50  uniCF50,  HANGUL SYLLABLE-CF50
        '53073' => '16437'            , # G+16437  U+CF51  uniCF51,  HANGUL SYLLABLE-CF51
        '53074' => '16438'            , # G+16438  U+CF52  uniCF52,  HANGUL SYLLABLE-CF52
        '53075' => '16439'            , # G+16439  U+CF53  uniCF53,  HANGUL SYLLABLE-CF53
        '53076' => '3023'             , # G+3023  U+CF54  uniCF54,  HANGUL SYLLABLE-CF54
        '53077' => '3024'             , # G+3024  U+CF55  uniCF55,  HANGUL SYLLABLE-CF55
        '53078' => '16440'            , # G+16440  U+CF56  uniCF56,  HANGUL SYLLABLE-CF56
        '53079' => '16441'            , # G+16441  U+CF57  uniCF57,  HANGUL SYLLABLE-CF57
        '53080' => '3025'             , # G+3025  U+CF58  uniCF58,  HANGUL SYLLABLE-CF58
        '53081' => '16442'            , # G+16442  U+CF59  uniCF59,  HANGUL SYLLABLE-CF59
        '53082' => '16443'            , # G+16443  U+CF5A  uniCF5A,  HANGUL SYLLABLE-CF5A
        '53083' => '16444'            , # G+16444  U+CF5B  uniCF5B,  HANGUL SYLLABLE-CF5B
        '53084' => '3026'             , # G+3026  U+CF5C  uniCF5C,  HANGUL SYLLABLE-CF5C
        '53085' => '16445'            , # G+16445  U+CF5D  uniCF5D,  HANGUL SYLLABLE-CF5D
        '53086' => '16446'            , # G+16446  U+CF5E  uniCF5E,  HANGUL SYLLABLE-CF5E
        '53087' => '16447'            , # G+16447  U+CF5F  uniCF5F,  HANGUL SYLLABLE-CF5F
        '53088' => '16448'            , # G+16448  U+CF60  uniCF60,  HANGUL SYLLABLE-CF60
        '53089' => '16449'            , # G+16449  U+CF61  uniCF61,  HANGUL SYLLABLE-CF61
        '53090' => '16450'            , # G+16450  U+CF62  uniCF62,  HANGUL SYLLABLE-CF62
        '53091' => '16451'            , # G+16451  U+CF63  uniCF63,  HANGUL SYLLABLE-CF63
        '53092' => '3027'             , # G+3027  U+CF64  uniCF64,  HANGUL SYLLABLE-CF64
        '53093' => '3028'             , # G+3028  U+CF65  uniCF65,  HANGUL SYLLABLE-CF65
        '53094' => '16452'            , # G+16452  U+CF66  uniCF66,  HANGUL SYLLABLE-CF66
        '53095' => '3029'             , # G+3029  U+CF67  uniCF67,  HANGUL SYLLABLE-CF67
        '53096' => '16453'            , # G+16453  U+CF68  uniCF68,  HANGUL SYLLABLE-CF68
        '53097' => '3030'             , # G+3030  U+CF69  uniCF69,  HANGUL SYLLABLE-CF69
        '53098' => '16454'            , # G+16454  U+CF6A  uniCF6A,  HANGUL SYLLABLE-CF6A
        '53099' => '16455'            , # G+16455  U+CF6B  uniCF6B,  HANGUL SYLLABLE-CF6B
        '53100' => '16456'            , # G+16456  U+CF6C  uniCF6C,  HANGUL SYLLABLE-CF6C
        '53101' => '16457'            , # G+16457  U+CF6D  uniCF6D,  HANGUL SYLLABLE-CF6D
        '53102' => '16458'            , # G+16458  U+CF6E  uniCF6E,  HANGUL SYLLABLE-CF6E
        '53103' => '16459'            , # G+16459  U+CF6F  uniCF6F,  HANGUL SYLLABLE-CF6F
        '53104' => '3031'             , # G+3031  U+CF70  uniCF70,  HANGUL SYLLABLE-CF70
        '53105' => '3032'             , # G+3032  U+CF71  uniCF71,  HANGUL SYLLABLE-CF71
        '53106' => '16460'            , # G+16460  U+CF72  uniCF72,  HANGUL SYLLABLE-CF72
        '53107' => '16461'            , # G+16461  U+CF73  uniCF73,  HANGUL SYLLABLE-CF73
        '53108' => '3033'             , # G+3033  U+CF74  uniCF74,  HANGUL SYLLABLE-CF74
        '53109' => '16462'            , # G+16462  U+CF75  uniCF75,  HANGUL SYLLABLE-CF75
        '53110' => '16463'            , # G+16463  U+CF76  uniCF76,  HANGUL SYLLABLE-CF76
        '53111' => '16464'            , # G+16464  U+CF77  uniCF77,  HANGUL SYLLABLE-CF77
        '53112' => '3034'             , # G+3034  U+CF78  uniCF78,  HANGUL SYLLABLE-CF78
        '53113' => '16465'            , # G+16465  U+CF79  uniCF79,  HANGUL SYLLABLE-CF79
        '53114' => '16466'            , # G+16466  U+CF7A  uniCF7A,  HANGUL SYLLABLE-CF7A
        '53115' => '16467'            , # G+16467  U+CF7B  uniCF7B,  HANGUL SYLLABLE-CF7B
        '53116' => '16468'            , # G+16468  U+CF7C  uniCF7C,  HANGUL SYLLABLE-CF7C
        '53117' => '16469'            , # G+16469  U+CF7D  uniCF7D,  HANGUL SYLLABLE-CF7D
        '53118' => '16470'            , # G+16470  U+CF7E  uniCF7E,  HANGUL SYLLABLE-CF7E
        '53119' => '16471'            , # G+16471  U+CF7F  uniCF7F,  HANGUL SYLLABLE-CF7F
        '53120' => '3035'             , # G+3035  U+CF80  uniCF80,  HANGUL SYLLABLE-CF80
        '53121' => '16472'            , # G+16472  U+CF81  uniCF81,  HANGUL SYLLABLE-CF81
        '53122' => '16473'            , # G+16473  U+CF82  uniCF82,  HANGUL SYLLABLE-CF82
        '53123' => '16474'            , # G+16474  U+CF83  uniCF83,  HANGUL SYLLABLE-CF83
        '53124' => '16475'            , # G+16475  U+CF84  uniCF84,  HANGUL SYLLABLE-CF84
        '53125' => '3036'             , # G+3036  U+CF85  uniCF85,  HANGUL SYLLABLE-CF85
        '53126' => '16476'            , # G+16476  U+CF86  uniCF86,  HANGUL SYLLABLE-CF86
        '53127' => '16477'            , # G+16477  U+CF87  uniCF87,  HANGUL SYLLABLE-CF87
        '53128' => '16478'            , # G+16478  U+CF88  uniCF88,  HANGUL SYLLABLE-CF88
        '53129' => '16479'            , # G+16479  U+CF89  uniCF89,  HANGUL SYLLABLE-CF89
        '53130' => '16480'            , # G+16480  U+CF8A  uniCF8A,  HANGUL SYLLABLE-CF8A
        '53131' => '16481'            , # G+16481  U+CF8B  uniCF8B,  HANGUL SYLLABLE-CF8B
        '53132' => '3037'             , # G+3037  U+CF8C  uniCF8C,  HANGUL SYLLABLE-CF8C
        '53133' => '16482'            , # G+16482  U+CF8D  uniCF8D,  HANGUL SYLLABLE-CF8D
        '53134' => '16483'            , # G+16483  U+CF8E  uniCF8E,  HANGUL SYLLABLE-CF8E
        '53135' => '16484'            , # G+16484  U+CF8F  uniCF8F,  HANGUL SYLLABLE-CF8F
        '53136' => '16485'            , # G+16485  U+CF90  uniCF90,  HANGUL SYLLABLE-CF90
        '53137' => '16486'            , # G+16486  U+CF91  uniCF91,  HANGUL SYLLABLE-CF91
        '53138' => '16487'            , # G+16487  U+CF92  uniCF92,  HANGUL SYLLABLE-CF92
        '53139' => '16488'            , # G+16488  U+CF93  uniCF93,  HANGUL SYLLABLE-CF93
        '53140' => '16489'            , # G+16489  U+CF94  uniCF94,  HANGUL SYLLABLE-CF94
        '53141' => '16490'            , # G+16490  U+CF95  uniCF95,  HANGUL SYLLABLE-CF95
        '53142' => '16491'            , # G+16491  U+CF96  uniCF96,  HANGUL SYLLABLE-CF96
        '53143' => '16492'            , # G+16492  U+CF97  uniCF97,  HANGUL SYLLABLE-CF97
        '53144' => '16493'            , # G+16493  U+CF98  uniCF98,  HANGUL SYLLABLE-CF98
        '53145' => '16494'            , # G+16494  U+CF99  uniCF99,  HANGUL SYLLABLE-CF99
        '53146' => '16495'            , # G+16495  U+CF9A  uniCF9A,  HANGUL SYLLABLE-CF9A
        '53147' => '16496'            , # G+16496  U+CF9B  uniCF9B,  HANGUL SYLLABLE-CF9B
        '53148' => '16497'            , # G+16497  U+CF9C  uniCF9C,  HANGUL SYLLABLE-CF9C
        '53149' => '16498'            , # G+16498  U+CF9D  uniCF9D,  HANGUL SYLLABLE-CF9D
        '53150' => '16499'            , # G+16499  U+CF9E  uniCF9E,  HANGUL SYLLABLE-CF9E
        '53151' => '16500'            , # G+16500  U+CF9F  uniCF9F,  HANGUL SYLLABLE-CF9F
        '53152' => '16501'            , # G+16501  U+CFA0  uniCFA0,  HANGUL SYLLABLE-CFA0
        '53153' => '3038'             , # G+3038  U+CFA1  uniCFA1,  HANGUL SYLLABLE-CFA1
        '53154' => '16502'            , # G+16502  U+CFA2  uniCFA2,  HANGUL SYLLABLE-CFA2
        '53155' => '16503'            , # G+16503  U+CFA3  uniCFA3,  HANGUL SYLLABLE-CFA3
        '53156' => '16504'            , # G+16504  U+CFA4  uniCFA4,  HANGUL SYLLABLE-CFA4
        '53157' => '16505'            , # G+16505  U+CFA5  uniCFA5,  HANGUL SYLLABLE-CFA5
        '53158' => '16506'            , # G+16506  U+CFA6  uniCFA6,  HANGUL SYLLABLE-CFA6
        '53159' => '16507'            , # G+16507  U+CFA7  uniCFA7,  HANGUL SYLLABLE-CFA7
        '53160' => '3039'             , # G+3039  U+CFA8  uniCFA8,  HANGUL SYLLABLE-CFA8
        '53161' => '16508'            , # G+16508  U+CFA9  uniCFA9,  HANGUL SYLLABLE-CFA9
        '53162' => '16509'            , # G+16509  U+CFAA  uniCFAA,  HANGUL SYLLABLE-CFAA
        '53163' => '16510'            , # G+16510  U+CFAB  uniCFAB,  HANGUL SYLLABLE-CFAB
        '53164' => '16511'            , # G+16511  U+CFAC  uniCFAC,  HANGUL SYLLABLE-CFAC
        '53165' => '16512'            , # G+16512  U+CFAD  uniCFAD,  HANGUL SYLLABLE-CFAD
        '53166' => '16513'            , # G+16513  U+CFAE  uniCFAE,  HANGUL SYLLABLE-CFAE
        '53167' => '16514'            , # G+16514  U+CFAF  uniCFAF,  HANGUL SYLLABLE-CFAF
        '53168' => '3040'             , # G+3040  U+CFB0  uniCFB0,  HANGUL SYLLABLE-CFB0
        '53169' => '16515'            , # G+16515  U+CFB1  uniCFB1,  HANGUL SYLLABLE-CFB1
        '53170' => '16516'            , # G+16516  U+CFB2  uniCFB2,  HANGUL SYLLABLE-CFB2
        '53171' => '16517'            , # G+16517  U+CFB3  uniCFB3,  HANGUL SYLLABLE-CFB3
        '53172' => '16518'            , # G+16518  U+CFB4  uniCFB4,  HANGUL SYLLABLE-CFB4
        '53173' => '16519'            , # G+16519  U+CFB5  uniCFB5,  HANGUL SYLLABLE-CFB5
        '53174' => '16520'            , # G+16520  U+CFB6  uniCFB6,  HANGUL SYLLABLE-CFB6
        '53175' => '16521'            , # G+16521  U+CFB7  uniCFB7,  HANGUL SYLLABLE-CFB7
        '53176' => '16522'            , # G+16522  U+CFB8  uniCFB8,  HANGUL SYLLABLE-CFB8
        '53177' => '16523'            , # G+16523  U+CFB9  uniCFB9,  HANGUL SYLLABLE-CFB9
        '53178' => '16524'            , # G+16524  U+CFBA  uniCFBA,  HANGUL SYLLABLE-CFBA
        '53179' => '16525'            , # G+16525  U+CFBB  uniCFBB,  HANGUL SYLLABLE-CFBB
        '53180' => '16526'            , # G+16526  U+CFBC  uniCFBC,  HANGUL SYLLABLE-CFBC
        '53181' => '16527'            , # G+16527  U+CFBD  uniCFBD,  HANGUL SYLLABLE-CFBD
        '53182' => '16528'            , # G+16528  U+CFBE  uniCFBE,  HANGUL SYLLABLE-CFBE
        '53183' => '16529'            , # G+16529  U+CFBF  uniCFBF,  HANGUL SYLLABLE-CFBF
        '53184' => '16530'            , # G+16530  U+CFC0  uniCFC0,  HANGUL SYLLABLE-CFC0
        '53185' => '16531'            , # G+16531  U+CFC1  uniCFC1,  HANGUL SYLLABLE-CFC1
        '53186' => '16532'            , # G+16532  U+CFC2  uniCFC2,  HANGUL SYLLABLE-CFC2
        '53187' => '16533'            , # G+16533  U+CFC3  uniCFC3,  HANGUL SYLLABLE-CFC3
        '53188' => '3041'             , # G+3041  U+CFC4  uniCFC4,  HANGUL SYLLABLE-CFC4
        '53189' => '16534'            , # G+16534  U+CFC5  uniCFC5,  HANGUL SYLLABLE-CFC5
        '53190' => '16535'            , # G+16535  U+CFC6  uniCFC6,  HANGUL SYLLABLE-CFC6
        '53191' => '16536'            , # G+16536  U+CFC7  uniCFC7,  HANGUL SYLLABLE-CFC7
        '53192' => '16537'            , # G+16537  U+CFC8  uniCFC8,  HANGUL SYLLABLE-CFC8
        '53193' => '16538'            , # G+16538  U+CFC9  uniCFC9,  HANGUL SYLLABLE-CFC9
        '53194' => '16539'            , # G+16539  U+CFCA  uniCFCA,  HANGUL SYLLABLE-CFCA
        '53195' => '16540'            , # G+16540  U+CFCB  uniCFCB,  HANGUL SYLLABLE-CFCB
        '53196' => '16541'            , # G+16541  U+CFCC  uniCFCC,  HANGUL SYLLABLE-CFCC
        '53197' => '16542'            , # G+16542  U+CFCD  uniCFCD,  HANGUL SYLLABLE-CFCD
        '53198' => '16543'            , # G+16543  U+CFCE  uniCFCE,  HANGUL SYLLABLE-CFCE
        '53199' => '16544'            , # G+16544  U+CFCF  uniCFCF,  HANGUL SYLLABLE-CFCF
        '53200' => '16545'            , # G+16545  U+CFD0  uniCFD0,  HANGUL SYLLABLE-CFD0
        '53201' => '16546'            , # G+16546  U+CFD1  uniCFD1,  HANGUL SYLLABLE-CFD1
        '53202' => '16547'            , # G+16547  U+CFD2  uniCFD2,  HANGUL SYLLABLE-CFD2
        '53203' => '16548'            , # G+16548  U+CFD3  uniCFD3,  HANGUL SYLLABLE-CFD3
        '53204' => '16549'            , # G+16549  U+CFD4  uniCFD4,  HANGUL SYLLABLE-CFD4
        '53205' => '16550'            , # G+16550  U+CFD5  uniCFD5,  HANGUL SYLLABLE-CFD5
        '53206' => '16551'            , # G+16551  U+CFD6  uniCFD6,  HANGUL SYLLABLE-CFD6
        '53207' => '16552'            , # G+16552  U+CFD7  uniCFD7,  HANGUL SYLLABLE-CFD7
        '53208' => '16553'            , # G+16553  U+CFD8  uniCFD8,  HANGUL SYLLABLE-CFD8
        '53209' => '16554'            , # G+16554  U+CFD9  uniCFD9,  HANGUL SYLLABLE-CFD9
        '53210' => '16555'            , # G+16555  U+CFDA  uniCFDA,  HANGUL SYLLABLE-CFDA
        '53211' => '16556'            , # G+16556  U+CFDB  uniCFDB,  HANGUL SYLLABLE-CFDB
        '53212' => '16557'            , # G+16557  U+CFDC  uniCFDC,  HANGUL SYLLABLE-CFDC
        '53213' => '16558'            , # G+16558  U+CFDD  uniCFDD,  HANGUL SYLLABLE-CFDD
        '53214' => '16559'            , # G+16559  U+CFDE  uniCFDE,  HANGUL SYLLABLE-CFDE
        '53215' => '16560'            , # G+16560  U+CFDF  uniCFDF,  HANGUL SYLLABLE-CFDF
        '53216' => '3042'             , # G+3042  U+CFE0  uniCFE0,  HANGUL SYLLABLE-CFE0
        '53217' => '3043'             , # G+3043  U+CFE1  uniCFE1,  HANGUL SYLLABLE-CFE1
        '53218' => '16561'            , # G+16561  U+CFE2  uniCFE2,  HANGUL SYLLABLE-CFE2
        '53219' => '16562'            , # G+16562  U+CFE3  uniCFE3,  HANGUL SYLLABLE-CFE3
        '53220' => '3044'             , # G+3044  U+CFE4  uniCFE4,  HANGUL SYLLABLE-CFE4
        '53221' => '16563'            , # G+16563  U+CFE5  uniCFE5,  HANGUL SYLLABLE-CFE5
        '53222' => '16564'            , # G+16564  U+CFE6  uniCFE6,  HANGUL SYLLABLE-CFE6
        '53223' => '16565'            , # G+16565  U+CFE7  uniCFE7,  HANGUL SYLLABLE-CFE7
        '53224' => '3045'             , # G+3045  U+CFE8  uniCFE8,  HANGUL SYLLABLE-CFE8
        '53225' => '16566'            , # G+16566  U+CFE9  uniCFE9,  HANGUL SYLLABLE-CFE9
        '53226' => '16567'            , # G+16567  U+CFEA  uniCFEA,  HANGUL SYLLABLE-CFEA
        '53227' => '16568'            , # G+16568  U+CFEB  uniCFEB,  HANGUL SYLLABLE-CFEB
        '53228' => '16569'            , # G+16569  U+CFEC  uniCFEC,  HANGUL SYLLABLE-CFEC
        '53229' => '16570'            , # G+16570  U+CFED  uniCFED,  HANGUL SYLLABLE-CFED
        '53230' => '16571'            , # G+16571  U+CFEE  uniCFEE,  HANGUL SYLLABLE-CFEE
        '53231' => '16572'            , # G+16572  U+CFEF  uniCFEF,  HANGUL SYLLABLE-CFEF
        '53232' => '3046'             , # G+3046  U+CFF0  uniCFF0,  HANGUL SYLLABLE-CFF0
        '53233' => '3047'             , # G+3047  U+CFF1  uniCFF1,  HANGUL SYLLABLE-CFF1
        '53234' => '16573'            , # G+16573  U+CFF2  uniCFF2,  HANGUL SYLLABLE-CFF2
        '53235' => '3048'             , # G+3048  U+CFF3  uniCFF3,  HANGUL SYLLABLE-CFF3
        '53236' => '16574'            , # G+16574  U+CFF4  uniCFF4,  HANGUL SYLLABLE-CFF4
        '53237' => '3049'             , # G+3049  U+CFF5  uniCFF5,  HANGUL SYLLABLE-CFF5
        '53238' => '16575'            , # G+16575  U+CFF6  uniCFF6,  HANGUL SYLLABLE-CFF6
        '53239' => '16576'            , # G+16576  U+CFF7  uniCFF7,  HANGUL SYLLABLE-CFF7
        '53240' => '16577'            , # G+16577  U+CFF8  uniCFF8,  HANGUL SYLLABLE-CFF8
        '53241' => '16578'            , # G+16578  U+CFF9  uniCFF9,  HANGUL SYLLABLE-CFF9
        '53242' => '16579'            , # G+16579  U+CFFA  uniCFFA,  HANGUL SYLLABLE-CFFA
        '53243' => '16580'            , # G+16580  U+CFFB  uniCFFB,  HANGUL SYLLABLE-CFFB
        '53244' => '3050'             , # G+3050  U+CFFC  uniCFFC,  HANGUL SYLLABLE-CFFC
        '53245' => '16581'            , # G+16581  U+CFFD  uniCFFD,  HANGUL SYLLABLE-CFFD
        '53246' => '16582'            , # G+16582  U+CFFE  uniCFFE,  HANGUL SYLLABLE-CFFE
        '53247' => '16583'            , # G+16583  U+CFFF  uniCFFF,  HANGUL SYLLABLE-CFFF
        '53248' => '3051'             , # G+3051  U+D000  uniD000,  HANGUL SYLLABLE-D000
        '53249' => '16584'            , # G+16584  U+D001  uniD001,  HANGUL SYLLABLE-D001
        '53250' => '16585'            , # G+16585  U+D002  uniD002,  HANGUL SYLLABLE-D002
        '53251' => '16586'            , # G+16586  U+D003  uniD003,  HANGUL SYLLABLE-D003
        '53252' => '3052'             , # G+3052  U+D004  uniD004,  HANGUL SYLLABLE-D004
        '53253' => '16587'            , # G+16587  U+D005  uniD005,  HANGUL SYLLABLE-D005
        '53254' => '16588'            , # G+16588  U+D006  uniD006,  HANGUL SYLLABLE-D006
        '53255' => '16589'            , # G+16589  U+D007  uniD007,  HANGUL SYLLABLE-D007
        '53256' => '16590'            , # G+16590  U+D008  uniD008,  HANGUL SYLLABLE-D008
        '53257' => '16591'            , # G+16591  U+D009  uniD009,  HANGUL SYLLABLE-D009
        '53258' => '16592'            , # G+16592  U+D00A  uniD00A,  HANGUL SYLLABLE-D00A
        '53259' => '16593'            , # G+16593  U+D00B  uniD00B,  HANGUL SYLLABLE-D00B
        '53260' => '16594'            , # G+16594  U+D00C  uniD00C,  HANGUL SYLLABLE-D00C
        '53261' => '16595'            , # G+16595  U+D00D  uniD00D,  HANGUL SYLLABLE-D00D
        '53262' => '16596'            , # G+16596  U+D00E  uniD00E,  HANGUL SYLLABLE-D00E
        '53263' => '16597'            , # G+16597  U+D00F  uniD00F,  HANGUL SYLLABLE-D00F
        '53264' => '16598'            , # G+16598  U+D010  uniD010,  HANGUL SYLLABLE-D010
        '53265' => '3053'             , # G+3053  U+D011  uniD011,  HANGUL SYLLABLE-D011
        '53266' => '16599'            , # G+16599  U+D012  uniD012,  HANGUL SYLLABLE-D012
        '53267' => '16600'            , # G+16600  U+D013  uniD013,  HANGUL SYLLABLE-D013
        '53268' => '16601'            , # G+16601  U+D014  uniD014,  HANGUL SYLLABLE-D014
        '53269' => '16602'            , # G+16602  U+D015  uniD015,  HANGUL SYLLABLE-D015
        '53270' => '16603'            , # G+16603  U+D016  uniD016,  HANGUL SYLLABLE-D016
        '53271' => '16604'            , # G+16604  U+D017  uniD017,  HANGUL SYLLABLE-D017
        '53272' => '3054'             , # G+3054  U+D018  uniD018,  HANGUL SYLLABLE-D018
        '53273' => '16605'            , # G+16605  U+D019  uniD019,  HANGUL SYLLABLE-D019
        '53274' => '16606'            , # G+16606  U+D01A  uniD01A,  HANGUL SYLLABLE-D01A
        '53275' => '16607'            , # G+16607  U+D01B  uniD01B,  HANGUL SYLLABLE-D01B
        '53276' => '16608'            , # G+16608  U+D01C  uniD01C,  HANGUL SYLLABLE-D01C
        '53277' => '16609'            , # G+16609  U+D01D  uniD01D,  HANGUL SYLLABLE-D01D
        '53278' => '16610'            , # G+16610  U+D01E  uniD01E,  HANGUL SYLLABLE-D01E
        '53279' => '16611'            , # G+16611  U+D01F  uniD01F,  HANGUL SYLLABLE-D01F
        '53280' => '16612'            , # G+16612  U+D020  uniD020,  HANGUL SYLLABLE-D020
        '53281' => '16613'            , # G+16613  U+D021  uniD021,  HANGUL SYLLABLE-D021
        '53282' => '16614'            , # G+16614  U+D022  uniD022,  HANGUL SYLLABLE-D022
        '53283' => '16615'            , # G+16615  U+D023  uniD023,  HANGUL SYLLABLE-D023
        '53284' => '16616'            , # G+16616  U+D024  uniD024,  HANGUL SYLLABLE-D024
        '53285' => '16617'            , # G+16617  U+D025  uniD025,  HANGUL SYLLABLE-D025
        '53286' => '16618'            , # G+16618  U+D026  uniD026,  HANGUL SYLLABLE-D026
        '53287' => '16619'            , # G+16619  U+D027  uniD027,  HANGUL SYLLABLE-D027
        '53288' => '16620'            , # G+16620  U+D028  uniD028,  HANGUL SYLLABLE-D028
        '53289' => '16621'            , # G+16621  U+D029  uniD029,  HANGUL SYLLABLE-D029
        '53290' => '16622'            , # G+16622  U+D02A  uniD02A,  HANGUL SYLLABLE-D02A
        '53291' => '16623'            , # G+16623  U+D02B  uniD02B,  HANGUL SYLLABLE-D02B
        '53292' => '16624'            , # G+16624  U+D02C  uniD02C,  HANGUL SYLLABLE-D02C
        '53293' => '3055'             , # G+3055  U+D02D  uniD02D,  HANGUL SYLLABLE-D02D
        '53294' => '16625'            , # G+16625  U+D02E  uniD02E,  HANGUL SYLLABLE-D02E
        '53295' => '16626'            , # G+16626  U+D02F  uniD02F,  HANGUL SYLLABLE-D02F
        '53296' => '16627'            , # G+16627  U+D030  uniD030,  HANGUL SYLLABLE-D030
        '53297' => '16628'            , # G+16628  U+D031  uniD031,  HANGUL SYLLABLE-D031
        '53298' => '16629'            , # G+16629  U+D032  uniD032,  HANGUL SYLLABLE-D032
        '53299' => '16630'            , # G+16630  U+D033  uniD033,  HANGUL SYLLABLE-D033
        '53300' => '3056'             , # G+3056  U+D034  uniD034,  HANGUL SYLLABLE-D034
        '53301' => '3057'             , # G+3057  U+D035  uniD035,  HANGUL SYLLABLE-D035
        '53302' => '16631'            , # G+16631  U+D036  uniD036,  HANGUL SYLLABLE-D036
        '53303' => '16632'            , # G+16632  U+D037  uniD037,  HANGUL SYLLABLE-D037
        '53304' => '3058'             , # G+3058  U+D038  uniD038,  HANGUL SYLLABLE-D038
        '53305' => '16633'            , # G+16633  U+D039  uniD039,  HANGUL SYLLABLE-D039
        '53306' => '16634'            , # G+16634  U+D03A  uniD03A,  HANGUL SYLLABLE-D03A
        '53307' => '16635'            , # G+16635  U+D03B  uniD03B,  HANGUL SYLLABLE-D03B
        '53308' => '3059'             , # G+3059  U+D03C  uniD03C,  HANGUL SYLLABLE-D03C
        '53309' => '16636'            , # G+16636  U+D03D  uniD03D,  HANGUL SYLLABLE-D03D
        '53310' => '16637'            , # G+16637  U+D03E  uniD03E,  HANGUL SYLLABLE-D03E
        '53311' => '16638'            , # G+16638  U+D03F  uniD03F,  HANGUL SYLLABLE-D03F
        '53312' => '16639'            , # G+16639  U+D040  uniD040,  HANGUL SYLLABLE-D040
        '53313' => '16640'            , # G+16640  U+D041  uniD041,  HANGUL SYLLABLE-D041
        '53314' => '16641'            , # G+16641  U+D042  uniD042,  HANGUL SYLLABLE-D042
        '53315' => '16642'            , # G+16642  U+D043  uniD043,  HANGUL SYLLABLE-D043
        '53316' => '3060'             , # G+3060  U+D044  uniD044,  HANGUL SYLLABLE-D044
        '53317' => '3061'             , # G+3061  U+D045  uniD045,  HANGUL SYLLABLE-D045
        '53318' => '16643'            , # G+16643  U+D046  uniD046,  HANGUL SYLLABLE-D046
        '53319' => '3062'             , # G+3062  U+D047  uniD047,  HANGUL SYLLABLE-D047
        '53320' => '16644'            , # G+16644  U+D048  uniD048,  HANGUL SYLLABLE-D048
        '53321' => '3063'             , # G+3063  U+D049  uniD049,  HANGUL SYLLABLE-D049
        '53322' => '16645'            , # G+16645  U+D04A  uniD04A,  HANGUL SYLLABLE-D04A
        '53323' => '16646'            , # G+16646  U+D04B  uniD04B,  HANGUL SYLLABLE-D04B
        '53324' => '16647'            , # G+16647  U+D04C  uniD04C,  HANGUL SYLLABLE-D04C
        '53325' => '16648'            , # G+16648  U+D04D  uniD04D,  HANGUL SYLLABLE-D04D
        '53326' => '16649'            , # G+16649  U+D04E  uniD04E,  HANGUL SYLLABLE-D04E
        '53327' => '16650'            , # G+16650  U+D04F  uniD04F,  HANGUL SYLLABLE-D04F
        '53328' => '3064'             , # G+3064  U+D050  uniD050,  HANGUL SYLLABLE-D050
        '53329' => '16651'            , # G+16651  U+D051  uniD051,  HANGUL SYLLABLE-D051
        '53330' => '16652'            , # G+16652  U+D052  uniD052,  HANGUL SYLLABLE-D052
        '53331' => '16653'            , # G+16653  U+D053  uniD053,  HANGUL SYLLABLE-D053
        '53332' => '3065'             , # G+3065  U+D054  uniD054,  HANGUL SYLLABLE-D054
        '53333' => '16654'            , # G+16654  U+D055  uniD055,  HANGUL SYLLABLE-D055
        '53334' => '16655'            , # G+16655  U+D056  uniD056,  HANGUL SYLLABLE-D056
        '53335' => '16656'            , # G+16656  U+D057  uniD057,  HANGUL SYLLABLE-D057
        '53336' => '3066'             , # G+3066  U+D058  uniD058,  HANGUL SYLLABLE-D058
        '53337' => '16657'            , # G+16657  U+D059  uniD059,  HANGUL SYLLABLE-D059
        '53338' => '16658'            , # G+16658  U+D05A  uniD05A,  HANGUL SYLLABLE-D05A
        '53339' => '16659'            , # G+16659  U+D05B  uniD05B,  HANGUL SYLLABLE-D05B
        '53340' => '16660'            , # G+16660  U+D05C  uniD05C,  HANGUL SYLLABLE-D05C
        '53341' => '16661'            , # G+16661  U+D05D  uniD05D,  HANGUL SYLLABLE-D05D
        '53342' => '16662'            , # G+16662  U+D05E  uniD05E,  HANGUL SYLLABLE-D05E
        '53343' => '16663'            , # G+16663  U+D05F  uniD05F,  HANGUL SYLLABLE-D05F
        '53344' => '3067'             , # G+3067  U+D060  uniD060,  HANGUL SYLLABLE-D060
        '53345' => '16664'            , # G+16664  U+D061  uniD061,  HANGUL SYLLABLE-D061
        '53346' => '16665'            , # G+16665  U+D062  uniD062,  HANGUL SYLLABLE-D062
        '53347' => '16666'            , # G+16666  U+D063  uniD063,  HANGUL SYLLABLE-D063
        '53348' => '16667'            , # G+16667  U+D064  uniD064,  HANGUL SYLLABLE-D064
        '53349' => '16668'            , # G+16668  U+D065  uniD065,  HANGUL SYLLABLE-D065
        '53350' => '16669'            , # G+16669  U+D066  uniD066,  HANGUL SYLLABLE-D066
        '53351' => '16670'            , # G+16670  U+D067  uniD067,  HANGUL SYLLABLE-D067
        '53352' => '16671'            , # G+16671  U+D068  uniD068,  HANGUL SYLLABLE-D068
        '53353' => '16672'            , # G+16672  U+D069  uniD069,  HANGUL SYLLABLE-D069
        '53354' => '16673'            , # G+16673  U+D06A  uniD06A,  HANGUL SYLLABLE-D06A
        '53355' => '16674'            , # G+16674  U+D06B  uniD06B,  HANGUL SYLLABLE-D06B
        '53356' => '3068'             , # G+3068  U+D06C  uniD06C,  HANGUL SYLLABLE-D06C
        '53357' => '3069'             , # G+3069  U+D06D  uniD06D,  HANGUL SYLLABLE-D06D
        '53358' => '16675'            , # G+16675  U+D06E  uniD06E,  HANGUL SYLLABLE-D06E
        '53359' => '16676'            , # G+16676  U+D06F  uniD06F,  HANGUL SYLLABLE-D06F
        '53360' => '3070'             , # G+3070  U+D070  uniD070,  HANGUL SYLLABLE-D070
        '53361' => '16677'            , # G+16677  U+D071  uniD071,  HANGUL SYLLABLE-D071
        '53362' => '16678'            , # G+16678  U+D072  uniD072,  HANGUL SYLLABLE-D072
        '53363' => '16679'            , # G+16679  U+D073  uniD073,  HANGUL SYLLABLE-D073
        '53364' => '3071'             , # G+3071  U+D074  uniD074,  HANGUL SYLLABLE-D074
        '53365' => '16680'            , # G+16680  U+D075  uniD075,  HANGUL SYLLABLE-D075
        '53366' => '16681'            , # G+16681  U+D076  uniD076,  HANGUL SYLLABLE-D076
        '53367' => '16682'            , # G+16682  U+D077  uniD077,  HANGUL SYLLABLE-D077
        '53368' => '16683'            , # G+16683  U+D078  uniD078,  HANGUL SYLLABLE-D078
        '53369' => '16684'            , # G+16684  U+D079  uniD079,  HANGUL SYLLABLE-D079
        '53370' => '16685'            , # G+16685  U+D07A  uniD07A,  HANGUL SYLLABLE-D07A
        '53371' => '16686'            , # G+16686  U+D07B  uniD07B,  HANGUL SYLLABLE-D07B
        '53372' => '3072'             , # G+3072  U+D07C  uniD07C,  HANGUL SYLLABLE-D07C
        '53373' => '3073'             , # G+3073  U+D07D  uniD07D,  HANGUL SYLLABLE-D07D
        '53374' => '16687'            , # G+16687  U+D07E  uniD07E,  HANGUL SYLLABLE-D07E
        '53375' => '16688'            , # G+16688  U+D07F  uniD07F,  HANGUL SYLLABLE-D07F
        '53376' => '16689'            , # G+16689  U+D080  uniD080,  HANGUL SYLLABLE-D080
        '53377' => '3074'             , # G+3074  U+D081  uniD081,  HANGUL SYLLABLE-D081
        '53378' => '16690'            , # G+16690  U+D082  uniD082,  HANGUL SYLLABLE-D082
        '53379' => '16691'            , # G+16691  U+D083  uniD083,  HANGUL SYLLABLE-D083
        '53380' => '16692'            , # G+16692  U+D084  uniD084,  HANGUL SYLLABLE-D084
        '53381' => '16693'            , # G+16693  U+D085  uniD085,  HANGUL SYLLABLE-D085
        '53382' => '16694'            , # G+16694  U+D086  uniD086,  HANGUL SYLLABLE-D086
        '53383' => '16695'            , # G+16695  U+D087  uniD087,  HANGUL SYLLABLE-D087
        '53384' => '16696'            , # G+16696  U+D088  uniD088,  HANGUL SYLLABLE-D088
        '53385' => '16697'            , # G+16697  U+D089  uniD089,  HANGUL SYLLABLE-D089
        '53386' => '16698'            , # G+16698  U+D08A  uniD08A,  HANGUL SYLLABLE-D08A
        '53387' => '16699'            , # G+16699  U+D08B  uniD08B,  HANGUL SYLLABLE-D08B
        '53388' => '16700'            , # G+16700  U+D08C  uniD08C,  HANGUL SYLLABLE-D08C
        '53389' => '16701'            , # G+16701  U+D08D  uniD08D,  HANGUL SYLLABLE-D08D
        '53390' => '16702'            , # G+16702  U+D08E  uniD08E,  HANGUL SYLLABLE-D08E
        '53391' => '16703'            , # G+16703  U+D08F  uniD08F,  HANGUL SYLLABLE-D08F
        '53392' => '16704'            , # G+16704  U+D090  uniD090,  HANGUL SYLLABLE-D090
        '53393' => '16705'            , # G+16705  U+D091  uniD091,  HANGUL SYLLABLE-D091
        '53394' => '16706'            , # G+16706  U+D092  uniD092,  HANGUL SYLLABLE-D092
        '53395' => '16707'            , # G+16707  U+D093  uniD093,  HANGUL SYLLABLE-D093
        '53396' => '16708'            , # G+16708  U+D094  uniD094,  HANGUL SYLLABLE-D094
        '53397' => '16709'            , # G+16709  U+D095  uniD095,  HANGUL SYLLABLE-D095
        '53398' => '16710'            , # G+16710  U+D096  uniD096,  HANGUL SYLLABLE-D096
        '53399' => '16711'            , # G+16711  U+D097  uniD097,  HANGUL SYLLABLE-D097
        '53400' => '16712'            , # G+16712  U+D098  uniD098,  HANGUL SYLLABLE-D098
        '53401' => '16713'            , # G+16713  U+D099  uniD099,  HANGUL SYLLABLE-D099
        '53402' => '16714'            , # G+16714  U+D09A  uniD09A,  HANGUL SYLLABLE-D09A
        '53403' => '16715'            , # G+16715  U+D09B  uniD09B,  HANGUL SYLLABLE-D09B
        '53404' => '16716'            , # G+16716  U+D09C  uniD09C,  HANGUL SYLLABLE-D09C
        '53405' => '16717'            , # G+16717  U+D09D  uniD09D,  HANGUL SYLLABLE-D09D
        '53406' => '16718'            , # G+16718  U+D09E  uniD09E,  HANGUL SYLLABLE-D09E
        '53407' => '16719'            , # G+16719  U+D09F  uniD09F,  HANGUL SYLLABLE-D09F
        '53408' => '16720'            , # G+16720  U+D0A0  uniD0A0,  HANGUL SYLLABLE-D0A0
        '53409' => '16721'            , # G+16721  U+D0A1  uniD0A1,  HANGUL SYLLABLE-D0A1
        '53410' => '16722'            , # G+16722  U+D0A2  uniD0A2,  HANGUL SYLLABLE-D0A2
        '53411' => '16723'            , # G+16723  U+D0A3  uniD0A3,  HANGUL SYLLABLE-D0A3
        '53412' => '3075'             , # G+3075  U+D0A4  uniD0A4,  HANGUL SYLLABLE-D0A4
        '53413' => '3076'             , # G+3076  U+D0A5  uniD0A5,  HANGUL SYLLABLE-D0A5
        '53414' => '16724'            , # G+16724  U+D0A6  uniD0A6,  HANGUL SYLLABLE-D0A6
        '53415' => '16725'            , # G+16725  U+D0A7  uniD0A7,  HANGUL SYLLABLE-D0A7
        '53416' => '3077'             , # G+3077  U+D0A8  uniD0A8,  HANGUL SYLLABLE-D0A8
        '53417' => '16726'            , # G+16726  U+D0A9  uniD0A9,  HANGUL SYLLABLE-D0A9
        '53418' => '16727'            , # G+16727  U+D0AA  uniD0AA,  HANGUL SYLLABLE-D0AA
        '53419' => '16728'            , # G+16728  U+D0AB  uniD0AB,  HANGUL SYLLABLE-D0AB
        '53420' => '3078'             , # G+3078  U+D0AC  uniD0AC,  HANGUL SYLLABLE-D0AC
        '53421' => '16729'            , # G+16729  U+D0AD  uniD0AD,  HANGUL SYLLABLE-D0AD
        '53422' => '16730'            , # G+16730  U+D0AE  uniD0AE,  HANGUL SYLLABLE-D0AE
        '53423' => '16731'            , # G+16731  U+D0AF  uniD0AF,  HANGUL SYLLABLE-D0AF
        '53424' => '16732'            , # G+16732  U+D0B0  uniD0B0,  HANGUL SYLLABLE-D0B0
        '53425' => '16733'            , # G+16733  U+D0B1  uniD0B1,  HANGUL SYLLABLE-D0B1
        '53426' => '16734'            , # G+16734  U+D0B2  uniD0B2,  HANGUL SYLLABLE-D0B2
        '53427' => '16735'            , # G+16735  U+D0B3  uniD0B3,  HANGUL SYLLABLE-D0B3
        '53428' => '3079'             , # G+3079  U+D0B4  uniD0B4,  HANGUL SYLLABLE-D0B4
        '53429' => '3080'             , # G+3080  U+D0B5  uniD0B5,  HANGUL SYLLABLE-D0B5
        '53430' => '16736'            , # G+16736  U+D0B6  uniD0B6,  HANGUL SYLLABLE-D0B6
        '53431' => '3081'             , # G+3081  U+D0B7  uniD0B7,  HANGUL SYLLABLE-D0B7
        '53432' => '16737'            , # G+16737  U+D0B8  uniD0B8,  HANGUL SYLLABLE-D0B8
        '53433' => '3082'             , # G+3082  U+D0B9  uniD0B9,  HANGUL SYLLABLE-D0B9
        '53434' => '16738'            , # G+16738  U+D0BA  uniD0BA,  HANGUL SYLLABLE-D0BA
        '53435' => '16739'            , # G+16739  U+D0BB  uniD0BB,  HANGUL SYLLABLE-D0BB
        '53436' => '16740'            , # G+16740  U+D0BC  uniD0BC,  HANGUL SYLLABLE-D0BC
        '53437' => '16741'            , # G+16741  U+D0BD  uniD0BD,  HANGUL SYLLABLE-D0BD
        '53438' => '16742'            , # G+16742  U+D0BE  uniD0BE,  HANGUL SYLLABLE-D0BE
        '53439' => '16743'            , # G+16743  U+D0BF  uniD0BF,  HANGUL SYLLABLE-D0BF
        '53440' => '3083'             , # G+3083  U+D0C0  uniD0C0,  HANGUL SYLLABLE-D0C0
        '53441' => '3084'             , # G+3084  U+D0C1  uniD0C1,  HANGUL SYLLABLE-D0C1
        '53442' => '16744'            , # G+16744  U+D0C2  uniD0C2,  HANGUL SYLLABLE-D0C2
        '53443' => '16745'            , # G+16745  U+D0C3  uniD0C3,  HANGUL SYLLABLE-D0C3
        '53444' => '3085'             , # G+3085  U+D0C4  uniD0C4,  HANGUL SYLLABLE-D0C4
        '53445' => '16746'            , # G+16746  U+D0C5  uniD0C5,  HANGUL SYLLABLE-D0C5
        '53446' => '16747'            , # G+16747  U+D0C6  uniD0C6,  HANGUL SYLLABLE-D0C6
        '53447' => '16748'            , # G+16748  U+D0C7  uniD0C7,  HANGUL SYLLABLE-D0C7
        '53448' => '3086'             , # G+3086  U+D0C8  uniD0C8,  HANGUL SYLLABLE-D0C8
        '53449' => '3087'             , # G+3087  U+D0C9  uniD0C9,  HANGUL SYLLABLE-D0C9
        '53450' => '16749'            , # G+16749  U+D0CA  uniD0CA,  HANGUL SYLLABLE-D0CA
        '53451' => '16750'            , # G+16750  U+D0CB  uniD0CB,  HANGUL SYLLABLE-D0CB
        '53452' => '16751'            , # G+16751  U+D0CC  uniD0CC,  HANGUL SYLLABLE-D0CC
        '53453' => '16752'            , # G+16752  U+D0CD  uniD0CD,  HANGUL SYLLABLE-D0CD
        '53454' => '16753'            , # G+16753  U+D0CE  uniD0CE,  HANGUL SYLLABLE-D0CE
        '53455' => '16754'            , # G+16754  U+D0CF  uniD0CF,  HANGUL SYLLABLE-D0CF
        '53456' => '3088'             , # G+3088  U+D0D0  uniD0D0,  HANGUL SYLLABLE-D0D0
        '53457' => '3089'             , # G+3089  U+D0D1  uniD0D1,  HANGUL SYLLABLE-D0D1
        '53458' => '16755'            , # G+16755  U+D0D2  uniD0D2,  HANGUL SYLLABLE-D0D2
        '53459' => '3090'             , # G+3090  U+D0D3  uniD0D3,  HANGUL SYLLABLE-D0D3
        '53460' => '3091'             , # G+3091  U+D0D4  uniD0D4,  HANGUL SYLLABLE-D0D4
        '53461' => '3092'             , # G+3092  U+D0D5  uniD0D5,  HANGUL SYLLABLE-D0D5
        '53462' => '16756'            , # G+16756  U+D0D6  uniD0D6,  HANGUL SYLLABLE-D0D6
        '53463' => '16757'            , # G+16757  U+D0D7  uniD0D7,  HANGUL SYLLABLE-D0D7
        '53464' => '16758'            , # G+16758  U+D0D8  uniD0D8,  HANGUL SYLLABLE-D0D8
        '53465' => '16759'            , # G+16759  U+D0D9  uniD0D9,  HANGUL SYLLABLE-D0D9
        '53466' => '16760'            , # G+16760  U+D0DA  uniD0DA,  HANGUL SYLLABLE-D0DA
        '53467' => '16761'            , # G+16761  U+D0DB  uniD0DB,  HANGUL SYLLABLE-D0DB
        '53468' => '3093'             , # G+3093  U+D0DC  uniD0DC,  HANGUL SYLLABLE-D0DC
        '53469' => '3094'             , # G+3094  U+D0DD  uniD0DD,  HANGUL SYLLABLE-D0DD
        '53470' => '16762'            , # G+16762  U+D0DE  uniD0DE,  HANGUL SYLLABLE-D0DE
        '53471' => '16763'            , # G+16763  U+D0DF  uniD0DF,  HANGUL SYLLABLE-D0DF
        '53472' => '3095'             , # G+3095  U+D0E0  uniD0E0,  HANGUL SYLLABLE-D0E0
        '53473' => '16764'            , # G+16764  U+D0E1  uniD0E1,  HANGUL SYLLABLE-D0E1
        '53474' => '16765'            , # G+16765  U+D0E2  uniD0E2,  HANGUL SYLLABLE-D0E2
        '53475' => '16766'            , # G+16766  U+D0E3  uniD0E3,  HANGUL SYLLABLE-D0E3
        '53476' => '3096'             , # G+3096  U+D0E4  uniD0E4,  HANGUL SYLLABLE-D0E4
        '53477' => '16767'            , # G+16767  U+D0E5  uniD0E5,  HANGUL SYLLABLE-D0E5
        '53478' => '16768'            , # G+16768  U+D0E6  uniD0E6,  HANGUL SYLLABLE-D0E6
        '53479' => '16769'            , # G+16769  U+D0E7  uniD0E7,  HANGUL SYLLABLE-D0E7
        '53480' => '16770'            , # G+16770  U+D0E8  uniD0E8,  HANGUL SYLLABLE-D0E8
        '53481' => '16771'            , # G+16771  U+D0E9  uniD0E9,  HANGUL SYLLABLE-D0E9
        '53482' => '16772'            , # G+16772  U+D0EA  uniD0EA,  HANGUL SYLLABLE-D0EA
        '53483' => '16773'            , # G+16773  U+D0EB  uniD0EB,  HANGUL SYLLABLE-D0EB
        '53484' => '3097'             , # G+3097  U+D0EC  uniD0EC,  HANGUL SYLLABLE-D0EC
        '53485' => '3098'             , # G+3098  U+D0ED  uniD0ED,  HANGUL SYLLABLE-D0ED
        '53486' => '16774'            , # G+16774  U+D0EE  uniD0EE,  HANGUL SYLLABLE-D0EE
        '53487' => '3099'             , # G+3099  U+D0EF  uniD0EF,  HANGUL SYLLABLE-D0EF
        '53488' => '3100'             , # G+3100  U+D0F0  uniD0F0,  HANGUL SYLLABLE-D0F0
        '53489' => '3101'             , # G+3101  U+D0F1  uniD0F1,  HANGUL SYLLABLE-D0F1
        '53490' => '16775'            , # G+16775  U+D0F2  uniD0F2,  HANGUL SYLLABLE-D0F2
        '53491' => '16776'            , # G+16776  U+D0F3  uniD0F3,  HANGUL SYLLABLE-D0F3
        '53492' => '16777'            , # G+16777  U+D0F4  uniD0F4,  HANGUL SYLLABLE-D0F4
        '53493' => '16778'            , # G+16778  U+D0F5  uniD0F5,  HANGUL SYLLABLE-D0F5
        '53494' => '16779'            , # G+16779  U+D0F6  uniD0F6,  HANGUL SYLLABLE-D0F6
        '53495' => '16780'            , # G+16780  U+D0F7  uniD0F7,  HANGUL SYLLABLE-D0F7
        '53496' => '3102'             , # G+3102  U+D0F8  uniD0F8,  HANGUL SYLLABLE-D0F8
        '53497' => '16781'            , # G+16781  U+D0F9  uniD0F9,  HANGUL SYLLABLE-D0F9
        '53498' => '16782'            , # G+16782  U+D0FA  uniD0FA,  HANGUL SYLLABLE-D0FA
        '53499' => '16783'            , # G+16783  U+D0FB  uniD0FB,  HANGUL SYLLABLE-D0FB
        '53500' => '16784'            , # G+16784  U+D0FC  uniD0FC,  HANGUL SYLLABLE-D0FC
        '53501' => '16785'            , # G+16785  U+D0FD  uniD0FD,  HANGUL SYLLABLE-D0FD
        '53502' => '16786'            , # G+16786  U+D0FE  uniD0FE,  HANGUL SYLLABLE-D0FE
        '53503' => '16787'            , # G+16787  U+D0FF  uniD0FF,  HANGUL SYLLABLE-D0FF
        '53504' => '16788'            , # G+16788  U+D100  uniD100,  HANGUL SYLLABLE-D100
        '53505' => '16789'            , # G+16789  U+D101  uniD101,  HANGUL SYLLABLE-D101
        '53506' => '16790'            , # G+16790  U+D102  uniD102,  HANGUL SYLLABLE-D102
        '53507' => '16791'            , # G+16791  U+D103  uniD103,  HANGUL SYLLABLE-D103
        '53508' => '16792'            , # G+16792  U+D104  uniD104,  HANGUL SYLLABLE-D104
        '53509' => '16793'            , # G+16793  U+D105  uniD105,  HANGUL SYLLABLE-D105
        '53510' => '16794'            , # G+16794  U+D106  uniD106,  HANGUL SYLLABLE-D106
        '53511' => '16795'            , # G+16795  U+D107  uniD107,  HANGUL SYLLABLE-D107
        '53512' => '16796'            , # G+16796  U+D108  uniD108,  HANGUL SYLLABLE-D108
        '53513' => '16797'            , # G+16797  U+D109  uniD109,  HANGUL SYLLABLE-D109
        '53514' => '16798'            , # G+16798  U+D10A  uniD10A,  HANGUL SYLLABLE-D10A
        '53515' => '16799'            , # G+16799  U+D10B  uniD10B,  HANGUL SYLLABLE-D10B
        '53516' => '16800'            , # G+16800  U+D10C  uniD10C,  HANGUL SYLLABLE-D10C
        '53517' => '3103'             , # G+3103  U+D10D  uniD10D,  HANGUL SYLLABLE-D10D
        '53518' => '16801'            , # G+16801  U+D10E  uniD10E,  HANGUL SYLLABLE-D10E
        '53519' => '16802'            , # G+16802  U+D10F  uniD10F,  HANGUL SYLLABLE-D10F
        '53520' => '16803'            , # G+16803  U+D110  uniD110,  HANGUL SYLLABLE-D110
        '53521' => '16804'            , # G+16804  U+D111  uniD111,  HANGUL SYLLABLE-D111
        '53522' => '16805'            , # G+16805  U+D112  uniD112,  HANGUL SYLLABLE-D112
        '53523' => '16806'            , # G+16806  U+D113  uniD113,  HANGUL SYLLABLE-D113
        '53524' => '16807'            , # G+16807  U+D114  uniD114,  HANGUL SYLLABLE-D114
        '53525' => '16808'            , # G+16808  U+D115  uniD115,  HANGUL SYLLABLE-D115
        '53526' => '16809'            , # G+16809  U+D116  uniD116,  HANGUL SYLLABLE-D116
        '53527' => '16810'            , # G+16810  U+D117  uniD117,  HANGUL SYLLABLE-D117
        '53528' => '16811'            , # G+16811  U+D118  uniD118,  HANGUL SYLLABLE-D118
        '53529' => '16812'            , # G+16812  U+D119  uniD119,  HANGUL SYLLABLE-D119
        '53530' => '16813'            , # G+16813  U+D11A  uniD11A,  HANGUL SYLLABLE-D11A
        '53531' => '16814'            , # G+16814  U+D11B  uniD11B,  HANGUL SYLLABLE-D11B
        '53532' => '16815'            , # G+16815  U+D11C  uniD11C,  HANGUL SYLLABLE-D11C
        '53533' => '16816'            , # G+16816  U+D11D  uniD11D,  HANGUL SYLLABLE-D11D
        '53534' => '16817'            , # G+16817  U+D11E  uniD11E,  HANGUL SYLLABLE-D11E
        '53535' => '16818'            , # G+16818  U+D11F  uniD11F,  HANGUL SYLLABLE-D11F
        '53536' => '16819'            , # G+16819  U+D120  uniD120,  HANGUL SYLLABLE-D120
        '53537' => '16820'            , # G+16820  U+D121  uniD121,  HANGUL SYLLABLE-D121
        '53538' => '16821'            , # G+16821  U+D122  uniD122,  HANGUL SYLLABLE-D122
        '53539' => '16822'            , # G+16822  U+D123  uniD123,  HANGUL SYLLABLE-D123
        '53540' => '16823'            , # G+16823  U+D124  uniD124,  HANGUL SYLLABLE-D124
        '53541' => '16824'            , # G+16824  U+D125  uniD125,  HANGUL SYLLABLE-D125
        '53542' => '16825'            , # G+16825  U+D126  uniD126,  HANGUL SYLLABLE-D126
        '53543' => '16826'            , # G+16826  U+D127  uniD127,  HANGUL SYLLABLE-D127
        '53544' => '16827'            , # G+16827  U+D128  uniD128,  HANGUL SYLLABLE-D128
        '53545' => '16828'            , # G+16828  U+D129  uniD129,  HANGUL SYLLABLE-D129
        '53546' => '16829'            , # G+16829  U+D12A  uniD12A,  HANGUL SYLLABLE-D12A
        '53547' => '16830'            , # G+16830  U+D12B  uniD12B,  HANGUL SYLLABLE-D12B
        '53548' => '16831'            , # G+16831  U+D12C  uniD12C,  HANGUL SYLLABLE-D12C
        '53549' => '16832'            , # G+16832  U+D12D  uniD12D,  HANGUL SYLLABLE-D12D
        '53550' => '16833'            , # G+16833  U+D12E  uniD12E,  HANGUL SYLLABLE-D12E
        '53551' => '16834'            , # G+16834  U+D12F  uniD12F,  HANGUL SYLLABLE-D12F
        '53552' => '3104'             , # G+3104  U+D130  uniD130,  HANGUL SYLLABLE-D130
        '53553' => '3105'             , # G+3105  U+D131  uniD131,  HANGUL SYLLABLE-D131
        '53554' => '16835'            , # G+16835  U+D132  uniD132,  HANGUL SYLLABLE-D132
        '53555' => '16836'            , # G+16836  U+D133  uniD133,  HANGUL SYLLABLE-D133
        '53556' => '3106'             , # G+3106  U+D134  uniD134,  HANGUL SYLLABLE-D134
        '53557' => '16837'            , # G+16837  U+D135  uniD135,  HANGUL SYLLABLE-D135
        '53558' => '16838'            , # G+16838  U+D136  uniD136,  HANGUL SYLLABLE-D136
        '53559' => '16839'            , # G+16839  U+D137  uniD137,  HANGUL SYLLABLE-D137
        '53560' => '3107'             , # G+3107  U+D138  uniD138,  HANGUL SYLLABLE-D138
        '53561' => '16840'            , # G+16840  U+D139  uniD139,  HANGUL SYLLABLE-D139
        '53562' => '3108'             , # G+3108  U+D13A  uniD13A,  HANGUL SYLLABLE-D13A
        '53563' => '16841'            , # G+16841  U+D13B  uniD13B,  HANGUL SYLLABLE-D13B
        '53564' => '16842'            , # G+16842  U+D13C  uniD13C,  HANGUL SYLLABLE-D13C
        '53565' => '16843'            , # G+16843  U+D13D  uniD13D,  HANGUL SYLLABLE-D13D
        '53566' => '16844'            , # G+16844  U+D13E  uniD13E,  HANGUL SYLLABLE-D13E
        '53567' => '16845'            , # G+16845  U+D13F  uniD13F,  HANGUL SYLLABLE-D13F
        '53568' => '3109'             , # G+3109  U+D140  uniD140,  HANGUL SYLLABLE-D140
        '53569' => '3110'             , # G+3110  U+D141  uniD141,  HANGUL SYLLABLE-D141
        '53570' => '16846'            , # G+16846  U+D142  uniD142,  HANGUL SYLLABLE-D142
        '53571' => '3111'             , # G+3111  U+D143  uniD143,  HANGUL SYLLABLE-D143
        '53572' => '3112'             , # G+3112  U+D144  uniD144,  HANGUL SYLLABLE-D144
        '53573' => '3113'             , # G+3113  U+D145  uniD145,  HANGUL SYLLABLE-D145
        '53574' => '16847'            , # G+16847  U+D146  uniD146,  HANGUL SYLLABLE-D146
        '53575' => '16848'            , # G+16848  U+D147  uniD147,  HANGUL SYLLABLE-D147
        '53576' => '16849'            , # G+16849  U+D148  uniD148,  HANGUL SYLLABLE-D148
        '53577' => '16850'            , # G+16850  U+D149  uniD149,  HANGUL SYLLABLE-D149
        '53578' => '16851'            , # G+16851  U+D14A  uniD14A,  HANGUL SYLLABLE-D14A
        '53579' => '16852'            , # G+16852  U+D14B  uniD14B,  HANGUL SYLLABLE-D14B
        '53580' => '3114'             , # G+3114  U+D14C  uniD14C,  HANGUL SYLLABLE-D14C
        '53581' => '3115'             , # G+3115  U+D14D  uniD14D,  HANGUL SYLLABLE-D14D
        '53582' => '16853'            , # G+16853  U+D14E  uniD14E,  HANGUL SYLLABLE-D14E
        '53583' => '16854'            , # G+16854  U+D14F  uniD14F,  HANGUL SYLLABLE-D14F
        '53584' => '3116'             , # G+3116  U+D150  uniD150,  HANGUL SYLLABLE-D150
        '53585' => '16855'            , # G+16855  U+D151  uniD151,  HANGUL SYLLABLE-D151
        '53586' => '16856'            , # G+16856  U+D152  uniD152,  HANGUL SYLLABLE-D152
        '53587' => '16857'            , # G+16857  U+D153  uniD153,  HANGUL SYLLABLE-D153
        '53588' => '3117'             , # G+3117  U+D154  uniD154,  HANGUL SYLLABLE-D154
        '53589' => '16858'            , # G+16858  U+D155  uniD155,  HANGUL SYLLABLE-D155
        '53590' => '16859'            , # G+16859  U+D156  uniD156,  HANGUL SYLLABLE-D156
        '53591' => '16860'            , # G+16860  U+D157  uniD157,  HANGUL SYLLABLE-D157
        '53592' => '16861'            , # G+16861  U+D158  uniD158,  HANGUL SYLLABLE-D158
        '53593' => '16862'            , # G+16862  U+D159  uniD159,  HANGUL SYLLABLE-D159
        '53594' => '16863'            , # G+16863  U+D15A  uniD15A,  HANGUL SYLLABLE-D15A
        '53595' => '16864'            , # G+16864  U+D15B  uniD15B,  HANGUL SYLLABLE-D15B
        '53596' => '3118'             , # G+3118  U+D15C  uniD15C,  HANGUL SYLLABLE-D15C
        '53597' => '3119'             , # G+3119  U+D15D  uniD15D,  HANGUL SYLLABLE-D15D
        '53598' => '16865'            , # G+16865  U+D15E  uniD15E,  HANGUL SYLLABLE-D15E
        '53599' => '3120'             , # G+3120  U+D15F  uniD15F,  HANGUL SYLLABLE-D15F
        '53600' => '16866'            , # G+16866  U+D160  uniD160,  HANGUL SYLLABLE-D160
        '53601' => '3121'             , # G+3121  U+D161  uniD161,  HANGUL SYLLABLE-D161
        '53602' => '16867'            , # G+16867  U+D162  uniD162,  HANGUL SYLLABLE-D162
        '53603' => '16868'            , # G+16868  U+D163  uniD163,  HANGUL SYLLABLE-D163
        '53604' => '16869'            , # G+16869  U+D164  uniD164,  HANGUL SYLLABLE-D164
        '53605' => '16870'            , # G+16870  U+D165  uniD165,  HANGUL SYLLABLE-D165
        '53606' => '16871'            , # G+16871  U+D166  uniD166,  HANGUL SYLLABLE-D166
        '53607' => '16872'            , # G+16872  U+D167  uniD167,  HANGUL SYLLABLE-D167
        '53608' => '3122'             , # G+3122  U+D168  uniD168,  HANGUL SYLLABLE-D168
        '53609' => '16873'            , # G+16873  U+D169  uniD169,  HANGUL SYLLABLE-D169
        '53610' => '16874'            , # G+16874  U+D16A  uniD16A,  HANGUL SYLLABLE-D16A
        '53611' => '16875'            , # G+16875  U+D16B  uniD16B,  HANGUL SYLLABLE-D16B
        '53612' => '3123'             , # G+3123  U+D16C  uniD16C,  HANGUL SYLLABLE-D16C
        '53613' => '16876'            , # G+16876  U+D16D  uniD16D,  HANGUL SYLLABLE-D16D
        '53614' => '16877'            , # G+16877  U+D16E  uniD16E,  HANGUL SYLLABLE-D16E
        '53615' => '16878'            , # G+16878  U+D16F  uniD16F,  HANGUL SYLLABLE-D16F
        '53616' => '16879'            , # G+16879  U+D170  uniD170,  HANGUL SYLLABLE-D170
        '53617' => '16880'            , # G+16880  U+D171  uniD171,  HANGUL SYLLABLE-D171
        '53618' => '16881'            , # G+16881  U+D172  uniD172,  HANGUL SYLLABLE-D172
        '53619' => '16882'            , # G+16882  U+D173  uniD173,  HANGUL SYLLABLE-D173
        '53620' => '16883'            , # G+16883  U+D174  uniD174,  HANGUL SYLLABLE-D174
        '53621' => '16884'            , # G+16884  U+D175  uniD175,  HANGUL SYLLABLE-D175
        '53622' => '16885'            , # G+16885  U+D176  uniD176,  HANGUL SYLLABLE-D176
        '53623' => '16886'            , # G+16886  U+D177  uniD177,  HANGUL SYLLABLE-D177
        '53624' => '16887'            , # G+16887  U+D178  uniD178,  HANGUL SYLLABLE-D178
        '53625' => '16888'            , # G+16888  U+D179  uniD179,  HANGUL SYLLABLE-D179
        '53626' => '16889'            , # G+16889  U+D17A  uniD17A,  HANGUL SYLLABLE-D17A
        '53627' => '16890'            , # G+16890  U+D17B  uniD17B,  HANGUL SYLLABLE-D17B
        '53628' => '3124'             , # G+3124  U+D17C  uniD17C,  HANGUL SYLLABLE-D17C
        '53629' => '16891'            , # G+16891  U+D17D  uniD17D,  HANGUL SYLLABLE-D17D
        '53630' => '16892'            , # G+16892  U+D17E  uniD17E,  HANGUL SYLLABLE-D17E
        '53631' => '16893'            , # G+16893  U+D17F  uniD17F,  HANGUL SYLLABLE-D17F
        '53632' => '16894'            , # G+16894  U+D180  uniD180,  HANGUL SYLLABLE-D180
        '53633' => '16895'            , # G+16895  U+D181  uniD181,  HANGUL SYLLABLE-D181
        '53634' => '16896'            , # G+16896  U+D182  uniD182,  HANGUL SYLLABLE-D182
        '53635' => '16897'            , # G+16897  U+D183  uniD183,  HANGUL SYLLABLE-D183
        '53636' => '3125'             , # G+3125  U+D184  uniD184,  HANGUL SYLLABLE-D184
        '53637' => '16898'            , # G+16898  U+D185  uniD185,  HANGUL SYLLABLE-D185
        '53638' => '16899'            , # G+16899  U+D186  uniD186,  HANGUL SYLLABLE-D186
        '53639' => '16900'            , # G+16900  U+D187  uniD187,  HANGUL SYLLABLE-D187
        '53640' => '3126'             , # G+3126  U+D188  uniD188,  HANGUL SYLLABLE-D188
        '53641' => '16901'            , # G+16901  U+D189  uniD189,  HANGUL SYLLABLE-D189
        '53642' => '16902'            , # G+16902  U+D18A  uniD18A,  HANGUL SYLLABLE-D18A
        '53643' => '16903'            , # G+16903  U+D18B  uniD18B,  HANGUL SYLLABLE-D18B
        '53644' => '16904'            , # G+16904  U+D18C  uniD18C,  HANGUL SYLLABLE-D18C
        '53645' => '16905'            , # G+16905  U+D18D  uniD18D,  HANGUL SYLLABLE-D18D
        '53646' => '16906'            , # G+16906  U+D18E  uniD18E,  HANGUL SYLLABLE-D18E
        '53647' => '16907'            , # G+16907  U+D18F  uniD18F,  HANGUL SYLLABLE-D18F
        '53648' => '16908'            , # G+16908  U+D190  uniD190,  HANGUL SYLLABLE-D190
        '53649' => '16909'            , # G+16909  U+D191  uniD191,  HANGUL SYLLABLE-D191
        '53650' => '16910'            , # G+16910  U+D192  uniD192,  HANGUL SYLLABLE-D192
        '53651' => '16911'            , # G+16911  U+D193  uniD193,  HANGUL SYLLABLE-D193
        '53652' => '16912'            , # G+16912  U+D194  uniD194,  HANGUL SYLLABLE-D194
        '53653' => '16913'            , # G+16913  U+D195  uniD195,  HANGUL SYLLABLE-D195
        '53654' => '16914'            , # G+16914  U+D196  uniD196,  HANGUL SYLLABLE-D196
        '53655' => '16915'            , # G+16915  U+D197  uniD197,  HANGUL SYLLABLE-D197
        '53656' => '16916'            , # G+16916  U+D198  uniD198,  HANGUL SYLLABLE-D198
        '53657' => '16917'            , # G+16917  U+D199  uniD199,  HANGUL SYLLABLE-D199
        '53658' => '16918'            , # G+16918  U+D19A  uniD19A,  HANGUL SYLLABLE-D19A
        '53659' => '16919'            , # G+16919  U+D19B  uniD19B,  HANGUL SYLLABLE-D19B
        '53660' => '16920'            , # G+16920  U+D19C  uniD19C,  HANGUL SYLLABLE-D19C
        '53661' => '16921'            , # G+16921  U+D19D  uniD19D,  HANGUL SYLLABLE-D19D
        '53662' => '16922'            , # G+16922  U+D19E  uniD19E,  HANGUL SYLLABLE-D19E
        '53663' => '16923'            , # G+16923  U+D19F  uniD19F,  HANGUL SYLLABLE-D19F
        '53664' => '3127'             , # G+3127  U+D1A0  uniD1A0,  HANGUL SYLLABLE-D1A0
        '53665' => '3128'             , # G+3128  U+D1A1  uniD1A1,  HANGUL SYLLABLE-D1A1
        '53666' => '16924'            , # G+16924  U+D1A2  uniD1A2,  HANGUL SYLLABLE-D1A2
        '53667' => '16925'            , # G+16925  U+D1A3  uniD1A3,  HANGUL SYLLABLE-D1A3
        '53668' => '3129'             , # G+3129  U+D1A4  uniD1A4,  HANGUL SYLLABLE-D1A4
        '53669' => '16926'            , # G+16926  U+D1A5  uniD1A5,  HANGUL SYLLABLE-D1A5
        '53670' => '16927'            , # G+16927  U+D1A6  uniD1A6,  HANGUL SYLLABLE-D1A6
        '53671' => '16928'            , # G+16928  U+D1A7  uniD1A7,  HANGUL SYLLABLE-D1A7
        '53672' => '3130'             , # G+3130  U+D1A8  uniD1A8,  HANGUL SYLLABLE-D1A8
        '53673' => '16929'            , # G+16929  U+D1A9  uniD1A9,  HANGUL SYLLABLE-D1A9
        '53674' => '16930'            , # G+16930  U+D1AA  uniD1AA,  HANGUL SYLLABLE-D1AA
        '53675' => '16931'            , # G+16931  U+D1AB  uniD1AB,  HANGUL SYLLABLE-D1AB
        '53676' => '16932'            , # G+16932  U+D1AC  uniD1AC,  HANGUL SYLLABLE-D1AC
        '53677' => '16933'            , # G+16933  U+D1AD  uniD1AD,  HANGUL SYLLABLE-D1AD
        '53678' => '16934'            , # G+16934  U+D1AE  uniD1AE,  HANGUL SYLLABLE-D1AE
        '53679' => '16935'            , # G+16935  U+D1AF  uniD1AF,  HANGUL SYLLABLE-D1AF
        '53680' => '3131'             , # G+3131  U+D1B0  uniD1B0,  HANGUL SYLLABLE-D1B0
        '53681' => '3132'             , # G+3132  U+D1B1  uniD1B1,  HANGUL SYLLABLE-D1B1
        '53682' => '16936'            , # G+16936  U+D1B2  uniD1B2,  HANGUL SYLLABLE-D1B2
        '53683' => '3133'             , # G+3133  U+D1B3  uniD1B3,  HANGUL SYLLABLE-D1B3
        '53684' => '16937'            , # G+16937  U+D1B4  uniD1B4,  HANGUL SYLLABLE-D1B4
        '53685' => '3134'             , # G+3134  U+D1B5  uniD1B5,  HANGUL SYLLABLE-D1B5
        '53686' => '16938'            , # G+16938  U+D1B6  uniD1B6,  HANGUL SYLLABLE-D1B6
        '53687' => '16939'            , # G+16939  U+D1B7  uniD1B7,  HANGUL SYLLABLE-D1B7
        '53688' => '16940'            , # G+16940  U+D1B8  uniD1B8,  HANGUL SYLLABLE-D1B8
        '53689' => '16941'            , # G+16941  U+D1B9  uniD1B9,  HANGUL SYLLABLE-D1B9
        '53690' => '3135'             , # G+3135  U+D1BA  uniD1BA,  HANGUL SYLLABLE-D1BA
        '53691' => '16942'            , # G+16942  U+D1BB  uniD1BB,  HANGUL SYLLABLE-D1BB
        '53692' => '3136'             , # G+3136  U+D1BC  uniD1BC,  HANGUL SYLLABLE-D1BC
        '53693' => '16943'            , # G+16943  U+D1BD  uniD1BD,  HANGUL SYLLABLE-D1BD
        '53694' => '16944'            , # G+16944  U+D1BE  uniD1BE,  HANGUL SYLLABLE-D1BE
        '53695' => '16945'            , # G+16945  U+D1BF  uniD1BF,  HANGUL SYLLABLE-D1BF
        '53696' => '3137'             , # G+3137  U+D1C0  uniD1C0,  HANGUL SYLLABLE-D1C0
        '53697' => '16946'            , # G+16946  U+D1C1  uniD1C1,  HANGUL SYLLABLE-D1C1
        '53698' => '16947'            , # G+16947  U+D1C2  uniD1C2,  HANGUL SYLLABLE-D1C2
        '53699' => '16948'            , # G+16948  U+D1C3  uniD1C3,  HANGUL SYLLABLE-D1C3
        '53700' => '16949'            , # G+16949  U+D1C4  uniD1C4,  HANGUL SYLLABLE-D1C4
        '53701' => '16950'            , # G+16950  U+D1C5  uniD1C5,  HANGUL SYLLABLE-D1C5
        '53702' => '16951'            , # G+16951  U+D1C6  uniD1C6,  HANGUL SYLLABLE-D1C6
        '53703' => '16952'            , # G+16952  U+D1C7  uniD1C7,  HANGUL SYLLABLE-D1C7
        '53704' => '16953'            , # G+16953  U+D1C8  uniD1C8,  HANGUL SYLLABLE-D1C8
        '53705' => '16954'            , # G+16954  U+D1C9  uniD1C9,  HANGUL SYLLABLE-D1C9
        '53706' => '16955'            , # G+16955  U+D1CA  uniD1CA,  HANGUL SYLLABLE-D1CA
        '53707' => '16956'            , # G+16956  U+D1CB  uniD1CB,  HANGUL SYLLABLE-D1CB
        '53708' => '16957'            , # G+16957  U+D1CC  uniD1CC,  HANGUL SYLLABLE-D1CC
        '53709' => '16958'            , # G+16958  U+D1CD  uniD1CD,  HANGUL SYLLABLE-D1CD
        '53710' => '16959'            , # G+16959  U+D1CE  uniD1CE,  HANGUL SYLLABLE-D1CE
        '53711' => '16960'            , # G+16960  U+D1CF  uniD1CF,  HANGUL SYLLABLE-D1CF
        '53712' => '16961'            , # G+16961  U+D1D0  uniD1D0,  HANGUL SYLLABLE-D1D0
        '53713' => '16962'            , # G+16962  U+D1D1  uniD1D1,  HANGUL SYLLABLE-D1D1
        '53714' => '16963'            , # G+16963  U+D1D2  uniD1D2,  HANGUL SYLLABLE-D1D2
        '53715' => '16964'            , # G+16964  U+D1D3  uniD1D3,  HANGUL SYLLABLE-D1D3
        '53716' => '16965'            , # G+16965  U+D1D4  uniD1D4,  HANGUL SYLLABLE-D1D4
        '53717' => '16966'            , # G+16966  U+D1D5  uniD1D5,  HANGUL SYLLABLE-D1D5
        '53718' => '16967'            , # G+16967  U+D1D6  uniD1D6,  HANGUL SYLLABLE-D1D6
        '53719' => '16968'            , # G+16968  U+D1D7  uniD1D7,  HANGUL SYLLABLE-D1D7
        '53720' => '3138'             , # G+3138  U+D1D8  uniD1D8,  HANGUL SYLLABLE-D1D8
        '53721' => '16969'            , # G+16969  U+D1D9  uniD1D9,  HANGUL SYLLABLE-D1D9
        '53722' => '16970'            , # G+16970  U+D1DA  uniD1DA,  HANGUL SYLLABLE-D1DA
        '53723' => '16971'            , # G+16971  U+D1DB  uniD1DB,  HANGUL SYLLABLE-D1DB
        '53724' => '16972'            , # G+16972  U+D1DC  uniD1DC,  HANGUL SYLLABLE-D1DC
        '53725' => '16973'            , # G+16973  U+D1DD  uniD1DD,  HANGUL SYLLABLE-D1DD
        '53726' => '16974'            , # G+16974  U+D1DE  uniD1DE,  HANGUL SYLLABLE-D1DE
        '53727' => '16975'            , # G+16975  U+D1DF  uniD1DF,  HANGUL SYLLABLE-D1DF
        '53728' => '16976'            , # G+16976  U+D1E0  uniD1E0,  HANGUL SYLLABLE-D1E0
        '53729' => '16977'            , # G+16977  U+D1E1  uniD1E1,  HANGUL SYLLABLE-D1E1
        '53730' => '16978'            , # G+16978  U+D1E2  uniD1E2,  HANGUL SYLLABLE-D1E2
        '53731' => '16979'            , # G+16979  U+D1E3  uniD1E3,  HANGUL SYLLABLE-D1E3
        '53732' => '16980'            , # G+16980  U+D1E4  uniD1E4,  HANGUL SYLLABLE-D1E4
        '53733' => '16981'            , # G+16981  U+D1E5  uniD1E5,  HANGUL SYLLABLE-D1E5
        '53734' => '16982'            , # G+16982  U+D1E6  uniD1E6,  HANGUL SYLLABLE-D1E6
        '53735' => '16983'            , # G+16983  U+D1E7  uniD1E7,  HANGUL SYLLABLE-D1E7
        '53736' => '16984'            , # G+16984  U+D1E8  uniD1E8,  HANGUL SYLLABLE-D1E8
        '53737' => '16985'            , # G+16985  U+D1E9  uniD1E9,  HANGUL SYLLABLE-D1E9
        '53738' => '16986'            , # G+16986  U+D1EA  uniD1EA,  HANGUL SYLLABLE-D1EA
        '53739' => '16987'            , # G+16987  U+D1EB  uniD1EB,  HANGUL SYLLABLE-D1EB
        '53740' => '16988'            , # G+16988  U+D1EC  uniD1EC,  HANGUL SYLLABLE-D1EC
        '53741' => '16989'            , # G+16989  U+D1ED  uniD1ED,  HANGUL SYLLABLE-D1ED
        '53742' => '16990'            , # G+16990  U+D1EE  uniD1EE,  HANGUL SYLLABLE-D1EE
        '53743' => '16991'            , # G+16991  U+D1EF  uniD1EF,  HANGUL SYLLABLE-D1EF
        '53744' => '16992'            , # G+16992  U+D1F0  uniD1F0,  HANGUL SYLLABLE-D1F0
        '53745' => '16993'            , # G+16993  U+D1F1  uniD1F1,  HANGUL SYLLABLE-D1F1
        '53746' => '16994'            , # G+16994  U+D1F2  uniD1F2,  HANGUL SYLLABLE-D1F2
        '53747' => '16995'            , # G+16995  U+D1F3  uniD1F3,  HANGUL SYLLABLE-D1F3
        '53748' => '3139'             , # G+3139  U+D1F4  uniD1F4,  HANGUL SYLLABLE-D1F4
        '53749' => '16996'            , # G+16996  U+D1F5  uniD1F5,  HANGUL SYLLABLE-D1F5
        '53750' => '16997'            , # G+16997  U+D1F6  uniD1F6,  HANGUL SYLLABLE-D1F6
        '53751' => '16998'            , # G+16998  U+D1F7  uniD1F7,  HANGUL SYLLABLE-D1F7
        '53752' => '3140'             , # G+3140  U+D1F8  uniD1F8,  HANGUL SYLLABLE-D1F8
        '53753' => '16999'            , # G+16999  U+D1F9  uniD1F9,  HANGUL SYLLABLE-D1F9
        '53754' => '17000'            , # G+17000  U+D1FA  uniD1FA,  HANGUL SYLLABLE-D1FA
        '53755' => '17001'            , # G+17001  U+D1FB  uniD1FB,  HANGUL SYLLABLE-D1FB
        '53756' => '17002'            , # G+17002  U+D1FC  uniD1FC,  HANGUL SYLLABLE-D1FC
        '53757' => '17003'            , # G+17003  U+D1FD  uniD1FD,  HANGUL SYLLABLE-D1FD
        '53758' => '17004'            , # G+17004  U+D1FE  uniD1FE,  HANGUL SYLLABLE-D1FE
        '53759' => '17005'            , # G+17005  U+D1FF  uniD1FF,  HANGUL SYLLABLE-D1FF
        '53760' => '17006'            , # G+17006  U+D200  uniD200,  HANGUL SYLLABLE-D200
        '53761' => '17007'            , # G+17007  U+D201  uniD201,  HANGUL SYLLABLE-D201
        '53762' => '17008'            , # G+17008  U+D202  uniD202,  HANGUL SYLLABLE-D202
        '53763' => '17009'            , # G+17009  U+D203  uniD203,  HANGUL SYLLABLE-D203
        '53764' => '17010'            , # G+17010  U+D204  uniD204,  HANGUL SYLLABLE-D204
        '53765' => '17011'            , # G+17011  U+D205  uniD205,  HANGUL SYLLABLE-D205
        '53766' => '17012'            , # G+17012  U+D206  uniD206,  HANGUL SYLLABLE-D206
        '53767' => '3141'             , # G+3141  U+D207  uniD207,  HANGUL SYLLABLE-D207
        '53768' => '17013'            , # G+17013  U+D208  uniD208,  HANGUL SYLLABLE-D208
        '53769' => '3142'             , # G+3142  U+D209  uniD209,  HANGUL SYLLABLE-D209
        '53770' => '17014'            , # G+17014  U+D20A  uniD20A,  HANGUL SYLLABLE-D20A
        '53771' => '17015'            , # G+17015  U+D20B  uniD20B,  HANGUL SYLLABLE-D20B
        '53772' => '17016'            , # G+17016  U+D20C  uniD20C,  HANGUL SYLLABLE-D20C
        '53773' => '17017'            , # G+17017  U+D20D  uniD20D,  HANGUL SYLLABLE-D20D
        '53774' => '17018'            , # G+17018  U+D20E  uniD20E,  HANGUL SYLLABLE-D20E
        '53775' => '17019'            , # G+17019  U+D20F  uniD20F,  HANGUL SYLLABLE-D20F
        '53776' => '3143'             , # G+3143  U+D210  uniD210,  HANGUL SYLLABLE-D210
        '53777' => '17020'            , # G+17020  U+D211  uniD211,  HANGUL SYLLABLE-D211
        '53778' => '17021'            , # G+17021  U+D212  uniD212,  HANGUL SYLLABLE-D212
        '53779' => '17022'            , # G+17022  U+D213  uniD213,  HANGUL SYLLABLE-D213
        '53780' => '17023'            , # G+17023  U+D214  uniD214,  HANGUL SYLLABLE-D214
        '53781' => '17024'            , # G+17024  U+D215  uniD215,  HANGUL SYLLABLE-D215
        '53782' => '17025'            , # G+17025  U+D216  uniD216,  HANGUL SYLLABLE-D216
        '53783' => '17026'            , # G+17026  U+D217  uniD217,  HANGUL SYLLABLE-D217
        '53784' => '17027'            , # G+17027  U+D218  uniD218,  HANGUL SYLLABLE-D218
        '53785' => '17028'            , # G+17028  U+D219  uniD219,  HANGUL SYLLABLE-D219
        '53786' => '17029'            , # G+17029  U+D21A  uniD21A,  HANGUL SYLLABLE-D21A
        '53787' => '17030'            , # G+17030  U+D21B  uniD21B,  HANGUL SYLLABLE-D21B
        '53788' => '17031'            , # G+17031  U+D21C  uniD21C,  HANGUL SYLLABLE-D21C
        '53789' => '17032'            , # G+17032  U+D21D  uniD21D,  HANGUL SYLLABLE-D21D
        '53790' => '17033'            , # G+17033  U+D21E  uniD21E,  HANGUL SYLLABLE-D21E
        '53791' => '17034'            , # G+17034  U+D21F  uniD21F,  HANGUL SYLLABLE-D21F
        '53792' => '17035'            , # G+17035  U+D220  uniD220,  HANGUL SYLLABLE-D220
        '53793' => '17036'            , # G+17036  U+D221  uniD221,  HANGUL SYLLABLE-D221
        '53794' => '17037'            , # G+17037  U+D222  uniD222,  HANGUL SYLLABLE-D222
        '53795' => '17038'            , # G+17038  U+D223  uniD223,  HANGUL SYLLABLE-D223
        '53796' => '17039'            , # G+17039  U+D224  uniD224,  HANGUL SYLLABLE-D224
        '53797' => '17040'            , # G+17040  U+D225  uniD225,  HANGUL SYLLABLE-D225
        '53798' => '17041'            , # G+17041  U+D226  uniD226,  HANGUL SYLLABLE-D226
        '53799' => '17042'            , # G+17042  U+D227  uniD227,  HANGUL SYLLABLE-D227
        '53800' => '17043'            , # G+17043  U+D228  uniD228,  HANGUL SYLLABLE-D228
        '53801' => '17044'            , # G+17044  U+D229  uniD229,  HANGUL SYLLABLE-D229
        '53802' => '17045'            , # G+17045  U+D22A  uniD22A,  HANGUL SYLLABLE-D22A
        '53803' => '17046'            , # G+17046  U+D22B  uniD22B,  HANGUL SYLLABLE-D22B
        '53804' => '3144'             , # G+3144  U+D22C  uniD22C,  HANGUL SYLLABLE-D22C
        '53805' => '3145'             , # G+3145  U+D22D  uniD22D,  HANGUL SYLLABLE-D22D
        '53806' => '17047'            , # G+17047  U+D22E  uniD22E,  HANGUL SYLLABLE-D22E
        '53807' => '17048'            , # G+17048  U+D22F  uniD22F,  HANGUL SYLLABLE-D22F
        '53808' => '3146'             , # G+3146  U+D230  uniD230,  HANGUL SYLLABLE-D230
        '53809' => '17049'            , # G+17049  U+D231  uniD231,  HANGUL SYLLABLE-D231
        '53810' => '17050'            , # G+17050  U+D232  uniD232,  HANGUL SYLLABLE-D232
        '53811' => '17051'            , # G+17051  U+D233  uniD233,  HANGUL SYLLABLE-D233
        '53812' => '3147'             , # G+3147  U+D234  uniD234,  HANGUL SYLLABLE-D234
        '53813' => '17052'            , # G+17052  U+D235  uniD235,  HANGUL SYLLABLE-D235
        '53814' => '17053'            , # G+17053  U+D236  uniD236,  HANGUL SYLLABLE-D236
        '53815' => '17054'            , # G+17054  U+D237  uniD237,  HANGUL SYLLABLE-D237
        '53816' => '17055'            , # G+17055  U+D238  uniD238,  HANGUL SYLLABLE-D238
        '53817' => '17056'            , # G+17056  U+D239  uniD239,  HANGUL SYLLABLE-D239
        '53818' => '17057'            , # G+17057  U+D23A  uniD23A,  HANGUL SYLLABLE-D23A
        '53819' => '17058'            , # G+17058  U+D23B  uniD23B,  HANGUL SYLLABLE-D23B
        '53820' => '3148'             , # G+3148  U+D23C  uniD23C,  HANGUL SYLLABLE-D23C
        '53821' => '3149'             , # G+3149  U+D23D  uniD23D,  HANGUL SYLLABLE-D23D
        '53822' => '17059'            , # G+17059  U+D23E  uniD23E,  HANGUL SYLLABLE-D23E
        '53823' => '3150'             , # G+3150  U+D23F  uniD23F,  HANGUL SYLLABLE-D23F
        '53824' => '17060'            , # G+17060  U+D240  uniD240,  HANGUL SYLLABLE-D240
        '53825' => '3151'             , # G+3151  U+D241  uniD241,  HANGUL SYLLABLE-D241
        '53826' => '17061'            , # G+17061  U+D242  uniD242,  HANGUL SYLLABLE-D242
        '53827' => '17062'            , # G+17062  U+D243  uniD243,  HANGUL SYLLABLE-D243
        '53828' => '17063'            , # G+17063  U+D244  uniD244,  HANGUL SYLLABLE-D244
        '53829' => '17064'            , # G+17064  U+D245  uniD245,  HANGUL SYLLABLE-D245
        '53830' => '17065'            , # G+17065  U+D246  uniD246,  HANGUL SYLLABLE-D246
        '53831' => '17066'            , # G+17066  U+D247  uniD247,  HANGUL SYLLABLE-D247
        '53832' => '3152'             , # G+3152  U+D248  uniD248,  HANGUL SYLLABLE-D248
        '53833' => '17067'            , # G+17067  U+D249  uniD249,  HANGUL SYLLABLE-D249
        '53834' => '17068'            , # G+17068  U+D24A  uniD24A,  HANGUL SYLLABLE-D24A
        '53835' => '17069'            , # G+17069  U+D24B  uniD24B,  HANGUL SYLLABLE-D24B
        '53836' => '17070'            , # G+17070  U+D24C  uniD24C,  HANGUL SYLLABLE-D24C
        '53837' => '17071'            , # G+17071  U+D24D  uniD24D,  HANGUL SYLLABLE-D24D
        '53838' => '17072'            , # G+17072  U+D24E  uniD24E,  HANGUL SYLLABLE-D24E
        '53839' => '17073'            , # G+17073  U+D24F  uniD24F,  HANGUL SYLLABLE-D24F
        '53840' => '17074'            , # G+17074  U+D250  uniD250,  HANGUL SYLLABLE-D250
        '53841' => '17075'            , # G+17075  U+D251  uniD251,  HANGUL SYLLABLE-D251
        '53842' => '17076'            , # G+17076  U+D252  uniD252,  HANGUL SYLLABLE-D252
        '53843' => '17077'            , # G+17077  U+D253  uniD253,  HANGUL SYLLABLE-D253
        '53844' => '17078'            , # G+17078  U+D254  uniD254,  HANGUL SYLLABLE-D254
        '53845' => '17079'            , # G+17079  U+D255  uniD255,  HANGUL SYLLABLE-D255
        '53846' => '17080'            , # G+17080  U+D256  uniD256,  HANGUL SYLLABLE-D256
        '53847' => '17081'            , # G+17081  U+D257  uniD257,  HANGUL SYLLABLE-D257
        '53848' => '17082'            , # G+17082  U+D258  uniD258,  HANGUL SYLLABLE-D258
        '53849' => '17083'            , # G+17083  U+D259  uniD259,  HANGUL SYLLABLE-D259
        '53850' => '17084'            , # G+17084  U+D25A  uniD25A,  HANGUL SYLLABLE-D25A
        '53851' => '17085'            , # G+17085  U+D25B  uniD25B,  HANGUL SYLLABLE-D25B
        '53852' => '3153'             , # G+3153  U+D25C  uniD25C,  HANGUL SYLLABLE-D25C
        '53853' => '17086'            , # G+17086  U+D25D  uniD25D,  HANGUL SYLLABLE-D25D
        '53854' => '17087'            , # G+17087  U+D25E  uniD25E,  HANGUL SYLLABLE-D25E
        '53855' => '17088'            , # G+17088  U+D25F  uniD25F,  HANGUL SYLLABLE-D25F
        '53856' => '17089'            , # G+17089  U+D260  uniD260,  HANGUL SYLLABLE-D260
        '53857' => '17090'            , # G+17090  U+D261  uniD261,  HANGUL SYLLABLE-D261
        '53858' => '17091'            , # G+17091  U+D262  uniD262,  HANGUL SYLLABLE-D262
        '53859' => '17092'            , # G+17092  U+D263  uniD263,  HANGUL SYLLABLE-D263
        '53860' => '3154'             , # G+3154  U+D264  uniD264,  HANGUL SYLLABLE-D264
        '53861' => '17093'            , # G+17093  U+D265  uniD265,  HANGUL SYLLABLE-D265
        '53862' => '17094'            , # G+17094  U+D266  uniD266,  HANGUL SYLLABLE-D266
        '53863' => '17095'            , # G+17095  U+D267  uniD267,  HANGUL SYLLABLE-D267
        '53864' => '17096'            , # G+17096  U+D268  uniD268,  HANGUL SYLLABLE-D268
        '53865' => '17097'            , # G+17097  U+D269  uniD269,  HANGUL SYLLABLE-D269
        '53866' => '17098'            , # G+17098  U+D26A  uniD26A,  HANGUL SYLLABLE-D26A
        '53867' => '17099'            , # G+17099  U+D26B  uniD26B,  HANGUL SYLLABLE-D26B
        '53868' => '17100'            , # G+17100  U+D26C  uniD26C,  HANGUL SYLLABLE-D26C
        '53869' => '17101'            , # G+17101  U+D26D  uniD26D,  HANGUL SYLLABLE-D26D
        '53870' => '17102'            , # G+17102  U+D26E  uniD26E,  HANGUL SYLLABLE-D26E
        '53871' => '17103'            , # G+17103  U+D26F  uniD26F,  HANGUL SYLLABLE-D26F
        '53872' => '17104'            , # G+17104  U+D270  uniD270,  HANGUL SYLLABLE-D270
        '53873' => '17105'            , # G+17105  U+D271  uniD271,  HANGUL SYLLABLE-D271
        '53874' => '17106'            , # G+17106  U+D272  uniD272,  HANGUL SYLLABLE-D272
        '53875' => '17107'            , # G+17107  U+D273  uniD273,  HANGUL SYLLABLE-D273
        '53876' => '17108'            , # G+17108  U+D274  uniD274,  HANGUL SYLLABLE-D274
        '53877' => '17109'            , # G+17109  U+D275  uniD275,  HANGUL SYLLABLE-D275
        '53878' => '17110'            , # G+17110  U+D276  uniD276,  HANGUL SYLLABLE-D276
        '53879' => '17111'            , # G+17111  U+D277  uniD277,  HANGUL SYLLABLE-D277
        '53880' => '17112'            , # G+17112  U+D278  uniD278,  HANGUL SYLLABLE-D278
        '53881' => '17113'            , # G+17113  U+D279  uniD279,  HANGUL SYLLABLE-D279
        '53882' => '17114'            , # G+17114  U+D27A  uniD27A,  HANGUL SYLLABLE-D27A
        '53883' => '17115'            , # G+17115  U+D27B  uniD27B,  HANGUL SYLLABLE-D27B
        '53884' => '17116'            , # G+17116  U+D27C  uniD27C,  HANGUL SYLLABLE-D27C
        '53885' => '17117'            , # G+17117  U+D27D  uniD27D,  HANGUL SYLLABLE-D27D
        '53886' => '17118'            , # G+17118  U+D27E  uniD27E,  HANGUL SYLLABLE-D27E
        '53887' => '17119'            , # G+17119  U+D27F  uniD27F,  HANGUL SYLLABLE-D27F
        '53888' => '3155'             , # G+3155  U+D280  uniD280,  HANGUL SYLLABLE-D280
        '53889' => '3156'             , # G+3156  U+D281  uniD281,  HANGUL SYLLABLE-D281
        '53890' => '17120'            , # G+17120  U+D282  uniD282,  HANGUL SYLLABLE-D282
        '53891' => '17121'            , # G+17121  U+D283  uniD283,  HANGUL SYLLABLE-D283
        '53892' => '3157'             , # G+3157  U+D284  uniD284,  HANGUL SYLLABLE-D284
        '53893' => '17122'            , # G+17122  U+D285  uniD285,  HANGUL SYLLABLE-D285
        '53894' => '17123'            , # G+17123  U+D286  uniD286,  HANGUL SYLLABLE-D286
        '53895' => '17124'            , # G+17124  U+D287  uniD287,  HANGUL SYLLABLE-D287
        '53896' => '3158'             , # G+3158  U+D288  uniD288,  HANGUL SYLLABLE-D288
        '53897' => '17125'            , # G+17125  U+D289  uniD289,  HANGUL SYLLABLE-D289
        '53898' => '17126'            , # G+17126  U+D28A  uniD28A,  HANGUL SYLLABLE-D28A
        '53899' => '17127'            , # G+17127  U+D28B  uniD28B,  HANGUL SYLLABLE-D28B
        '53900' => '17128'            , # G+17128  U+D28C  uniD28C,  HANGUL SYLLABLE-D28C
        '53901' => '17129'            , # G+17129  U+D28D  uniD28D,  HANGUL SYLLABLE-D28D
        '53902' => '17130'            , # G+17130  U+D28E  uniD28E,  HANGUL SYLLABLE-D28E
        '53903' => '17131'            , # G+17131  U+D28F  uniD28F,  HANGUL SYLLABLE-D28F
        '53904' => '3159'             , # G+3159  U+D290  uniD290,  HANGUL SYLLABLE-D290
        '53905' => '3160'             , # G+3160  U+D291  uniD291,  HANGUL SYLLABLE-D291
        '53906' => '17132'            , # G+17132  U+D292  uniD292,  HANGUL SYLLABLE-D292
        '53907' => '17133'            , # G+17133  U+D293  uniD293,  HANGUL SYLLABLE-D293
        '53908' => '17134'            , # G+17134  U+D294  uniD294,  HANGUL SYLLABLE-D294
        '53909' => '3161'             , # G+3161  U+D295  uniD295,  HANGUL SYLLABLE-D295
        '53910' => '17135'            , # G+17135  U+D296  uniD296,  HANGUL SYLLABLE-D296
        '53911' => '17136'            , # G+17136  U+D297  uniD297,  HANGUL SYLLABLE-D297
        '53912' => '17137'            , # G+17137  U+D298  uniD298,  HANGUL SYLLABLE-D298
        '53913' => '17138'            , # G+17138  U+D299  uniD299,  HANGUL SYLLABLE-D299
        '53914' => '17139'            , # G+17139  U+D29A  uniD29A,  HANGUL SYLLABLE-D29A
        '53915' => '17140'            , # G+17140  U+D29B  uniD29B,  HANGUL SYLLABLE-D29B
        '53916' => '3162'             , # G+3162  U+D29C  uniD29C,  HANGUL SYLLABLE-D29C
        '53917' => '17141'            , # G+17141  U+D29D  uniD29D,  HANGUL SYLLABLE-D29D
        '53918' => '17142'            , # G+17142  U+D29E  uniD29E,  HANGUL SYLLABLE-D29E
        '53919' => '17143'            , # G+17143  U+D29F  uniD29F,  HANGUL SYLLABLE-D29F
        '53920' => '3163'             , # G+3163  U+D2A0  uniD2A0,  HANGUL SYLLABLE-D2A0
        '53921' => '17144'            , # G+17144  U+D2A1  uniD2A1,  HANGUL SYLLABLE-D2A1
        '53922' => '17145'            , # G+17145  U+D2A2  uniD2A2,  HANGUL SYLLABLE-D2A2
        '53923' => '17146'            , # G+17146  U+D2A3  uniD2A3,  HANGUL SYLLABLE-D2A3
        '53924' => '3164'             , # G+3164  U+D2A4  uniD2A4,  HANGUL SYLLABLE-D2A4
        '53925' => '17147'            , # G+17147  U+D2A5  uniD2A5,  HANGUL SYLLABLE-D2A5
        '53926' => '17148'            , # G+17148  U+D2A6  uniD2A6,  HANGUL SYLLABLE-D2A6
        '53927' => '17149'            , # G+17149  U+D2A7  uniD2A7,  HANGUL SYLLABLE-D2A7
        '53928' => '17150'            , # G+17150  U+D2A8  uniD2A8,  HANGUL SYLLABLE-D2A8
        '53929' => '17151'            , # G+17151  U+D2A9  uniD2A9,  HANGUL SYLLABLE-D2A9
        '53930' => '17152'            , # G+17152  U+D2AA  uniD2AA,  HANGUL SYLLABLE-D2AA
        '53931' => '17153'            , # G+17153  U+D2AB  uniD2AB,  HANGUL SYLLABLE-D2AB
        '53932' => '3165'             , # G+3165  U+D2AC  uniD2AC,  HANGUL SYLLABLE-D2AC
        '53933' => '17154'            , # G+17154  U+D2AD  uniD2AD,  HANGUL SYLLABLE-D2AD
        '53934' => '17155'            , # G+17155  U+D2AE  uniD2AE,  HANGUL SYLLABLE-D2AE
        '53935' => '17156'            , # G+17156  U+D2AF  uniD2AF,  HANGUL SYLLABLE-D2AF
        '53936' => '17157'            , # G+17157  U+D2B0  uniD2B0,  HANGUL SYLLABLE-D2B0
        '53937' => '3166'             , # G+3166  U+D2B1  uniD2B1,  HANGUL SYLLABLE-D2B1
        '53938' => '17158'            , # G+17158  U+D2B2  uniD2B2,  HANGUL SYLLABLE-D2B2
        '53939' => '17159'            , # G+17159  U+D2B3  uniD2B3,  HANGUL SYLLABLE-D2B3
        '53940' => '17160'            , # G+17160  U+D2B4  uniD2B4,  HANGUL SYLLABLE-D2B4
        '53941' => '17161'            , # G+17161  U+D2B5  uniD2B5,  HANGUL SYLLABLE-D2B5
        '53942' => '17162'            , # G+17162  U+D2B6  uniD2B6,  HANGUL SYLLABLE-D2B6
        '53943' => '17163'            , # G+17163  U+D2B7  uniD2B7,  HANGUL SYLLABLE-D2B7
        '53944' => '3167'             , # G+3167  U+D2B8  uniD2B8,  HANGUL SYLLABLE-D2B8
        '53945' => '3168'             , # G+3168  U+D2B9  uniD2B9,  HANGUL SYLLABLE-D2B9
        '53946' => '17164'            , # G+17164  U+D2BA  uniD2BA,  HANGUL SYLLABLE-D2BA
        '53947' => '17165'            , # G+17165  U+D2BB  uniD2BB,  HANGUL SYLLABLE-D2BB
        '53948' => '3169'             , # G+3169  U+D2BC  uniD2BC,  HANGUL SYLLABLE-D2BC
        '53949' => '17166'            , # G+17166  U+D2BD  uniD2BD,  HANGUL SYLLABLE-D2BD
        '53950' => '17167'            , # G+17167  U+D2BE  uniD2BE,  HANGUL SYLLABLE-D2BE
        '53951' => '3170'             , # G+3170  U+D2BF  uniD2BF,  HANGUL SYLLABLE-D2BF
        '53952' => '3171'             , # G+3171  U+D2C0  uniD2C0,  HANGUL SYLLABLE-D2C0
        '53953' => '17168'            , # G+17168  U+D2C1  uniD2C1,  HANGUL SYLLABLE-D2C1
        '53954' => '3172'             , # G+3172  U+D2C2  uniD2C2,  HANGUL SYLLABLE-D2C2
        '53955' => '17169'            , # G+17169  U+D2C3  uniD2C3,  HANGUL SYLLABLE-D2C3
        '53956' => '17170'            , # G+17170  U+D2C4  uniD2C4,  HANGUL SYLLABLE-D2C4
        '53957' => '17171'            , # G+17171  U+D2C5  uniD2C5,  HANGUL SYLLABLE-D2C5
        '53958' => '17172'            , # G+17172  U+D2C6  uniD2C6,  HANGUL SYLLABLE-D2C6
        '53959' => '17173'            , # G+17173  U+D2C7  uniD2C7,  HANGUL SYLLABLE-D2C7
        '53960' => '3173'             , # G+3173  U+D2C8  uniD2C8,  HANGUL SYLLABLE-D2C8
        '53961' => '3174'             , # G+3174  U+D2C9  uniD2C9,  HANGUL SYLLABLE-D2C9
        '53962' => '17174'            , # G+17174  U+D2CA  uniD2CA,  HANGUL SYLLABLE-D2CA
        '53963' => '3175'             , # G+3175  U+D2CB  uniD2CB,  HANGUL SYLLABLE-D2CB
        '53964' => '17175'            , # G+17175  U+D2CC  uniD2CC,  HANGUL SYLLABLE-D2CC
        '53965' => '17176'            , # G+17176  U+D2CD  uniD2CD,  HANGUL SYLLABLE-D2CD
        '53966' => '17177'            , # G+17177  U+D2CE  uniD2CE,  HANGUL SYLLABLE-D2CE
        '53967' => '17178'            , # G+17178  U+D2CF  uniD2CF,  HANGUL SYLLABLE-D2CF
        '53968' => '17179'            , # G+17179  U+D2D0  uniD2D0,  HANGUL SYLLABLE-D2D0
        '53969' => '17180'            , # G+17180  U+D2D1  uniD2D1,  HANGUL SYLLABLE-D2D1
        '53970' => '17181'            , # G+17181  U+D2D2  uniD2D2,  HANGUL SYLLABLE-D2D2
        '53971' => '17182'            , # G+17182  U+D2D3  uniD2D3,  HANGUL SYLLABLE-D2D3
        '53972' => '3176'             , # G+3176  U+D2D4  uniD2D4,  HANGUL SYLLABLE-D2D4
        '53973' => '17183'            , # G+17183  U+D2D5  uniD2D5,  HANGUL SYLLABLE-D2D5
        '53974' => '17184'            , # G+17184  U+D2D6  uniD2D6,  HANGUL SYLLABLE-D2D6
        '53975' => '17185'            , # G+17185  U+D2D7  uniD2D7,  HANGUL SYLLABLE-D2D7
        '53976' => '3177'             , # G+3177  U+D2D8  uniD2D8,  HANGUL SYLLABLE-D2D8
        '53977' => '17186'            , # G+17186  U+D2D9  uniD2D9,  HANGUL SYLLABLE-D2D9
        '53978' => '17187'            , # G+17187  U+D2DA  uniD2DA,  HANGUL SYLLABLE-D2DA
        '53979' => '17188'            , # G+17188  U+D2DB  uniD2DB,  HANGUL SYLLABLE-D2DB
        '53980' => '3178'             , # G+3178  U+D2DC  uniD2DC,  HANGUL SYLLABLE-D2DC
        '53981' => '17189'            , # G+17189  U+D2DD  uniD2DD,  HANGUL SYLLABLE-D2DD
        '53982' => '17190'            , # G+17190  U+D2DE  uniD2DE,  HANGUL SYLLABLE-D2DE
        '53983' => '17191'            , # G+17191  U+D2DF  uniD2DF,  HANGUL SYLLABLE-D2DF
        '53984' => '17192'            , # G+17192  U+D2E0  uniD2E0,  HANGUL SYLLABLE-D2E0
        '53985' => '17193'            , # G+17193  U+D2E1  uniD2E1,  HANGUL SYLLABLE-D2E1
        '53986' => '17194'            , # G+17194  U+D2E2  uniD2E2,  HANGUL SYLLABLE-D2E2
        '53987' => '17195'            , # G+17195  U+D2E3  uniD2E3,  HANGUL SYLLABLE-D2E3
        '53988' => '3179'             , # G+3179  U+D2E4  uniD2E4,  HANGUL SYLLABLE-D2E4
        '53989' => '3180'             , # G+3180  U+D2E5  uniD2E5,  HANGUL SYLLABLE-D2E5
        '53990' => '17196'            , # G+17196  U+D2E6  uniD2E6,  HANGUL SYLLABLE-D2E6
        '53991' => '17197'            , # G+17197  U+D2E7  uniD2E7,  HANGUL SYLLABLE-D2E7
        '53992' => '17198'            , # G+17198  U+D2E8  uniD2E8,  HANGUL SYLLABLE-D2E8
        '53993' => '17199'            , # G+17199  U+D2E9  uniD2E9,  HANGUL SYLLABLE-D2E9
        '53994' => '17200'            , # G+17200  U+D2EA  uniD2EA,  HANGUL SYLLABLE-D2EA
        '53995' => '17201'            , # G+17201  U+D2EB  uniD2EB,  HANGUL SYLLABLE-D2EB
        '53996' => '17202'            , # G+17202  U+D2EC  uniD2EC,  HANGUL SYLLABLE-D2EC
        '53997' => '17203'            , # G+17203  U+D2ED  uniD2ED,  HANGUL SYLLABLE-D2ED
        '53998' => '17204'            , # G+17204  U+D2EE  uniD2EE,  HANGUL SYLLABLE-D2EE
        '53999' => '17205'            , # G+17205  U+D2EF  uniD2EF,  HANGUL SYLLABLE-D2EF
        '54000' => '3181'             , # G+3181  U+D2F0  uniD2F0,  HANGUL SYLLABLE-D2F0
        '54001' => '3182'             , # G+3182  U+D2F1  uniD2F1,  HANGUL SYLLABLE-D2F1
        '54002' => '17206'            , # G+17206  U+D2F2  uniD2F2,  HANGUL SYLLABLE-D2F2
        '54003' => '17207'            , # G+17207  U+D2F3  uniD2F3,  HANGUL SYLLABLE-D2F3
        '54004' => '3183'             , # G+3183  U+D2F4  uniD2F4,  HANGUL SYLLABLE-D2F4
        '54005' => '17208'            , # G+17208  U+D2F5  uniD2F5,  HANGUL SYLLABLE-D2F5
        '54006' => '17209'            , # G+17209  U+D2F6  uniD2F6,  HANGUL SYLLABLE-D2F6
        '54007' => '17210'            , # G+17210  U+D2F7  uniD2F7,  HANGUL SYLLABLE-D2F7
        '54008' => '3184'             , # G+3184  U+D2F8  uniD2F8,  HANGUL SYLLABLE-D2F8
        '54009' => '17211'            , # G+17211  U+D2F9  uniD2F9,  HANGUL SYLLABLE-D2F9
        '54010' => '17212'            , # G+17212  U+D2FA  uniD2FA,  HANGUL SYLLABLE-D2FA
        '54011' => '17213'            , # G+17213  U+D2FB  uniD2FB,  HANGUL SYLLABLE-D2FB
        '54012' => '17214'            , # G+17214  U+D2FC  uniD2FC,  HANGUL SYLLABLE-D2FC
        '54013' => '17215'            , # G+17215  U+D2FD  uniD2FD,  HANGUL SYLLABLE-D2FD
        '54014' => '17216'            , # G+17216  U+D2FE  uniD2FE,  HANGUL SYLLABLE-D2FE
        '54015' => '17217'            , # G+17217  U+D2FF  uniD2FF,  HANGUL SYLLABLE-D2FF
        '54016' => '3185'             , # G+3185  U+D300  uniD300,  HANGUL SYLLABLE-D300
        '54017' => '3186'             , # G+3186  U+D301  uniD301,  HANGUL SYLLABLE-D301
        '54018' => '17218'            , # G+17218  U+D302  uniD302,  HANGUL SYLLABLE-D302
        '54019' => '3187'             , # G+3187  U+D303  uniD303,  HANGUL SYLLABLE-D303
        '54020' => '17219'            , # G+17219  U+D304  uniD304,  HANGUL SYLLABLE-D304
        '54021' => '3188'             , # G+3188  U+D305  uniD305,  HANGUL SYLLABLE-D305
        '54022' => '17220'            , # G+17220  U+D306  uniD306,  HANGUL SYLLABLE-D306
        '54023' => '17221'            , # G+17221  U+D307  uniD307,  HANGUL SYLLABLE-D307
        '54024' => '17222'            , # G+17222  U+D308  uniD308,  HANGUL SYLLABLE-D308
        '54025' => '17223'            , # G+17223  U+D309  uniD309,  HANGUL SYLLABLE-D309
        '54026' => '17224'            , # G+17224  U+D30A  uniD30A,  HANGUL SYLLABLE-D30A
        '54027' => '17225'            , # G+17225  U+D30B  uniD30B,  HANGUL SYLLABLE-D30B
        '54028' => '3189'             , # G+3189  U+D30C  uniD30C,  HANGUL SYLLABLE-D30C
        '54029' => '3190'             , # G+3190  U+D30D  uniD30D,  HANGUL SYLLABLE-D30D
        '54030' => '3191'             , # G+3191  U+D30E  uniD30E,  HANGUL SYLLABLE-D30E
        '54031' => '17226'            , # G+17226  U+D30F  uniD30F,  HANGUL SYLLABLE-D30F
        '54032' => '3192'             , # G+3192  U+D310  uniD310,  HANGUL SYLLABLE-D310
        '54033' => '17227'            , # G+17227  U+D311  uniD311,  HANGUL SYLLABLE-D311
        '54034' => '17228'            , # G+17228  U+D312  uniD312,  HANGUL SYLLABLE-D312
        '54035' => '17229'            , # G+17229  U+D313  uniD313,  HANGUL SYLLABLE-D313
        '54036' => '3193'             , # G+3193  U+D314  uniD314,  HANGUL SYLLABLE-D314
        '54037' => '17230'            , # G+17230  U+D315  uniD315,  HANGUL SYLLABLE-D315
        '54038' => '3194'             , # G+3194  U+D316  uniD316,  HANGUL SYLLABLE-D316
        '54039' => '17231'            , # G+17231  U+D317  uniD317,  HANGUL SYLLABLE-D317
        '54040' => '17232'            , # G+17232  U+D318  uniD318,  HANGUL SYLLABLE-D318
        '54041' => '17233'            , # G+17233  U+D319  uniD319,  HANGUL SYLLABLE-D319
        '54042' => '17234'            , # G+17234  U+D31A  uniD31A,  HANGUL SYLLABLE-D31A
        '54043' => '17235'            , # G+17235  U+D31B  uniD31B,  HANGUL SYLLABLE-D31B
        '54044' => '3195'             , # G+3195  U+D31C  uniD31C,  HANGUL SYLLABLE-D31C
        '54045' => '3196'             , # G+3196  U+D31D  uniD31D,  HANGUL SYLLABLE-D31D
        '54046' => '17236'            , # G+17236  U+D31E  uniD31E,  HANGUL SYLLABLE-D31E
        '54047' => '3197'             , # G+3197  U+D31F  uniD31F,  HANGUL SYLLABLE-D31F
        '54048' => '3198'             , # G+3198  U+D320  uniD320,  HANGUL SYLLABLE-D320
        '54049' => '3199'             , # G+3199  U+D321  uniD321,  HANGUL SYLLABLE-D321
        '54050' => '17237'            , # G+17237  U+D322  uniD322,  HANGUL SYLLABLE-D322
        '54051' => '17238'            , # G+17238  U+D323  uniD323,  HANGUL SYLLABLE-D323
        '54052' => '17239'            , # G+17239  U+D324  uniD324,  HANGUL SYLLABLE-D324
        '54053' => '3200'             , # G+3200  U+D325  uniD325,  HANGUL SYLLABLE-D325
        '54054' => '17240'            , # G+17240  U+D326  uniD326,  HANGUL SYLLABLE-D326
        '54055' => '17241'            , # G+17241  U+D327  uniD327,  HANGUL SYLLABLE-D327
        '54056' => '3201'             , # G+3201  U+D328  uniD328,  HANGUL SYLLABLE-D328
        '54057' => '3202'             , # G+3202  U+D329  uniD329,  HANGUL SYLLABLE-D329
        '54058' => '17242'            , # G+17242  U+D32A  uniD32A,  HANGUL SYLLABLE-D32A
        '54059' => '17243'            , # G+17243  U+D32B  uniD32B,  HANGUL SYLLABLE-D32B
        '54060' => '3203'             , # G+3203  U+D32C  uniD32C,  HANGUL SYLLABLE-D32C
        '54061' => '17244'            , # G+17244  U+D32D  uniD32D,  HANGUL SYLLABLE-D32D
        '54062' => '17245'            , # G+17245  U+D32E  uniD32E,  HANGUL SYLLABLE-D32E
        '54063' => '17246'            , # G+17246  U+D32F  uniD32F,  HANGUL SYLLABLE-D32F
        '54064' => '3204'             , # G+3204  U+D330  uniD330,  HANGUL SYLLABLE-D330
        '54065' => '17247'            , # G+17247  U+D331  uniD331,  HANGUL SYLLABLE-D331
        '54066' => '17248'            , # G+17248  U+D332  uniD332,  HANGUL SYLLABLE-D332
        '54067' => '17249'            , # G+17249  U+D333  uniD333,  HANGUL SYLLABLE-D333
        '54068' => '17250'            , # G+17250  U+D334  uniD334,  HANGUL SYLLABLE-D334
        '54069' => '17251'            , # G+17251  U+D335  uniD335,  HANGUL SYLLABLE-D335
        '54070' => '17252'            , # G+17252  U+D336  uniD336,  HANGUL SYLLABLE-D336
        '54071' => '17253'            , # G+17253  U+D337  uniD337,  HANGUL SYLLABLE-D337
        '54072' => '3205'             , # G+3205  U+D338  uniD338,  HANGUL SYLLABLE-D338
        '54073' => '3206'             , # G+3206  U+D339  uniD339,  HANGUL SYLLABLE-D339
        '54074' => '17254'            , # G+17254  U+D33A  uniD33A,  HANGUL SYLLABLE-D33A
        '54075' => '3207'             , # G+3207  U+D33B  uniD33B,  HANGUL SYLLABLE-D33B
        '54076' => '3208'             , # G+3208  U+D33C  uniD33C,  HANGUL SYLLABLE-D33C
        '54077' => '3209'             , # G+3209  U+D33D  uniD33D,  HANGUL SYLLABLE-D33D
        '54078' => '17255'            , # G+17255  U+D33E  uniD33E,  HANGUL SYLLABLE-D33E
        '54079' => '17256'            , # G+17256  U+D33F  uniD33F,  HANGUL SYLLABLE-D33F
        '54080' => '17257'            , # G+17257  U+D340  uniD340,  HANGUL SYLLABLE-D340
        '54081' => '17258'            , # G+17258  U+D341  uniD341,  HANGUL SYLLABLE-D341
        '54082' => '17259'            , # G+17259  U+D342  uniD342,  HANGUL SYLLABLE-D342
        '54083' => '17260'            , # G+17260  U+D343  uniD343,  HANGUL SYLLABLE-D343
        '54084' => '3210'             , # G+3210  U+D344  uniD344,  HANGUL SYLLABLE-D344
        '54085' => '3211'             , # G+3211  U+D345  uniD345,  HANGUL SYLLABLE-D345
        '54086' => '17261'            , # G+17261  U+D346  uniD346,  HANGUL SYLLABLE-D346
        '54087' => '17262'            , # G+17262  U+D347  uniD347,  HANGUL SYLLABLE-D347
        '54088' => '17263'            , # G+17263  U+D348  uniD348,  HANGUL SYLLABLE-D348
        '54089' => '17264'            , # G+17264  U+D349  uniD349,  HANGUL SYLLABLE-D349
        '54090' => '17265'            , # G+17265  U+D34A  uniD34A,  HANGUL SYLLABLE-D34A
        '54091' => '17266'            , # G+17266  U+D34B  uniD34B,  HANGUL SYLLABLE-D34B
        '54092' => '17267'            , # G+17267  U+D34C  uniD34C,  HANGUL SYLLABLE-D34C
        '54093' => '17268'            , # G+17268  U+D34D  uniD34D,  HANGUL SYLLABLE-D34D
        '54094' => '17269'            , # G+17269  U+D34E  uniD34E,  HANGUL SYLLABLE-D34E
        '54095' => '17270'            , # G+17270  U+D34F  uniD34F,  HANGUL SYLLABLE-D34F
        '54096' => '17271'            , # G+17271  U+D350  uniD350,  HANGUL SYLLABLE-D350
        '54097' => '17272'            , # G+17272  U+D351  uniD351,  HANGUL SYLLABLE-D351
        '54098' => '17273'            , # G+17273  U+D352  uniD352,  HANGUL SYLLABLE-D352
        '54099' => '17274'            , # G+17274  U+D353  uniD353,  HANGUL SYLLABLE-D353
        '54100' => '17275'            , # G+17275  U+D354  uniD354,  HANGUL SYLLABLE-D354
        '54101' => '17276'            , # G+17276  U+D355  uniD355,  HANGUL SYLLABLE-D355
        '54102' => '17277'            , # G+17277  U+D356  uniD356,  HANGUL SYLLABLE-D356
        '54103' => '17278'            , # G+17278  U+D357  uniD357,  HANGUL SYLLABLE-D357
        '54104' => '17279'            , # G+17279  U+D358  uniD358,  HANGUL SYLLABLE-D358
        '54105' => '17280'            , # G+17280  U+D359  uniD359,  HANGUL SYLLABLE-D359
        '54106' => '17281'            , # G+17281  U+D35A  uniD35A,  HANGUL SYLLABLE-D35A
        '54107' => '17282'            , # G+17282  U+D35B  uniD35B,  HANGUL SYLLABLE-D35B
        '54108' => '17283'            , # G+17283  U+D35C  uniD35C,  HANGUL SYLLABLE-D35C
        '54109' => '17284'            , # G+17284  U+D35D  uniD35D,  HANGUL SYLLABLE-D35D
        '54110' => '17285'            , # G+17285  U+D35E  uniD35E,  HANGUL SYLLABLE-D35E
        '54111' => '17286'            , # G+17286  U+D35F  uniD35F,  HANGUL SYLLABLE-D35F
        '54112' => '17287'            , # G+17287  U+D360  uniD360,  HANGUL SYLLABLE-D360
        '54113' => '17288'            , # G+17288  U+D361  uniD361,  HANGUL SYLLABLE-D361
        '54114' => '17289'            , # G+17289  U+D362  uniD362,  HANGUL SYLLABLE-D362
        '54115' => '17290'            , # G+17290  U+D363  uniD363,  HANGUL SYLLABLE-D363
        '54116' => '17291'            , # G+17291  U+D364  uniD364,  HANGUL SYLLABLE-D364
        '54117' => '17292'            , # G+17292  U+D365  uniD365,  HANGUL SYLLABLE-D365
        '54118' => '17293'            , # G+17293  U+D366  uniD366,  HANGUL SYLLABLE-D366
        '54119' => '17294'            , # G+17294  U+D367  uniD367,  HANGUL SYLLABLE-D367
        '54120' => '17295'            , # G+17295  U+D368  uniD368,  HANGUL SYLLABLE-D368
        '54121' => '17296'            , # G+17296  U+D369  uniD369,  HANGUL SYLLABLE-D369
        '54122' => '17297'            , # G+17297  U+D36A  uniD36A,  HANGUL SYLLABLE-D36A
        '54123' => '17298'            , # G+17298  U+D36B  uniD36B,  HANGUL SYLLABLE-D36B
        '54124' => '17299'            , # G+17299  U+D36C  uniD36C,  HANGUL SYLLABLE-D36C
        '54125' => '17300'            , # G+17300  U+D36D  uniD36D,  HANGUL SYLLABLE-D36D
        '54126' => '17301'            , # G+17301  U+D36E  uniD36E,  HANGUL SYLLABLE-D36E
        '54127' => '17302'            , # G+17302  U+D36F  uniD36F,  HANGUL SYLLABLE-D36F
        '54128' => '17303'            , # G+17303  U+D370  uniD370,  HANGUL SYLLABLE-D370
        '54129' => '17304'            , # G+17304  U+D371  uniD371,  HANGUL SYLLABLE-D371
        '54130' => '17305'            , # G+17305  U+D372  uniD372,  HANGUL SYLLABLE-D372
        '54131' => '17306'            , # G+17306  U+D373  uniD373,  HANGUL SYLLABLE-D373
        '54132' => '17307'            , # G+17307  U+D374  uniD374,  HANGUL SYLLABLE-D374
        '54133' => '17308'            , # G+17308  U+D375  uniD375,  HANGUL SYLLABLE-D375
        '54134' => '17309'            , # G+17309  U+D376  uniD376,  HANGUL SYLLABLE-D376
        '54135' => '17310'            , # G+17310  U+D377  uniD377,  HANGUL SYLLABLE-D377
        '54136' => '17311'            , # G+17311  U+D378  uniD378,  HANGUL SYLLABLE-D378
        '54137' => '17312'            , # G+17312  U+D379  uniD379,  HANGUL SYLLABLE-D379
        '54138' => '17313'            , # G+17313  U+D37A  uniD37A,  HANGUL SYLLABLE-D37A
        '54139' => '17314'            , # G+17314  U+D37B  uniD37B,  HANGUL SYLLABLE-D37B
        '54140' => '3212'             , # G+3212  U+D37C  uniD37C,  HANGUL SYLLABLE-D37C
        '54141' => '3213'             , # G+3213  U+D37D  uniD37D,  HANGUL SYLLABLE-D37D
        '54142' => '17315'            , # G+17315  U+D37E  uniD37E,  HANGUL SYLLABLE-D37E
        '54143' => '17316'            , # G+17316  U+D37F  uniD37F,  HANGUL SYLLABLE-D37F
        '54144' => '3214'             , # G+3214  U+D380  uniD380,  HANGUL SYLLABLE-D380
        '54145' => '17317'            , # G+17317  U+D381  uniD381,  HANGUL SYLLABLE-D381
        '54146' => '17318'            , # G+17318  U+D382  uniD382,  HANGUL SYLLABLE-D382
        '54147' => '17319'            , # G+17319  U+D383  uniD383,  HANGUL SYLLABLE-D383
        '54148' => '3215'             , # G+3215  U+D384  uniD384,  HANGUL SYLLABLE-D384
        '54149' => '17320'            , # G+17320  U+D385  uniD385,  HANGUL SYLLABLE-D385
        '54150' => '17321'            , # G+17321  U+D386  uniD386,  HANGUL SYLLABLE-D386
        '54151' => '17322'            , # G+17322  U+D387  uniD387,  HANGUL SYLLABLE-D387
        '54152' => '17323'            , # G+17323  U+D388  uniD388,  HANGUL SYLLABLE-D388
        '54153' => '17324'            , # G+17324  U+D389  uniD389,  HANGUL SYLLABLE-D389
        '54154' => '17325'            , # G+17325  U+D38A  uniD38A,  HANGUL SYLLABLE-D38A
        '54155' => '17326'            , # G+17326  U+D38B  uniD38B,  HANGUL SYLLABLE-D38B
        '54156' => '3216'             , # G+3216  U+D38C  uniD38C,  HANGUL SYLLABLE-D38C
        '54157' => '3217'             , # G+3217  U+D38D  uniD38D,  HANGUL SYLLABLE-D38D
        '54158' => '17327'            , # G+17327  U+D38E  uniD38E,  HANGUL SYLLABLE-D38E
        '54159' => '3218'             , # G+3218  U+D38F  uniD38F,  HANGUL SYLLABLE-D38F
        '54160' => '3219'             , # G+3219  U+D390  uniD390,  HANGUL SYLLABLE-D390
        '54161' => '3220'             , # G+3220  U+D391  uniD391,  HANGUL SYLLABLE-D391
        '54162' => '17328'            , # G+17328  U+D392  uniD392,  HANGUL SYLLABLE-D392
        '54163' => '17329'            , # G+17329  U+D393  uniD393,  HANGUL SYLLABLE-D393
        '54164' => '17330'            , # G+17330  U+D394  uniD394,  HANGUL SYLLABLE-D394
        '54165' => '17331'            , # G+17331  U+D395  uniD395,  HANGUL SYLLABLE-D395
        '54166' => '17332'            , # G+17332  U+D396  uniD396,  HANGUL SYLLABLE-D396
        '54167' => '17333'            , # G+17333  U+D397  uniD397,  HANGUL SYLLABLE-D397
        '54168' => '3221'             , # G+3221  U+D398  uniD398,  HANGUL SYLLABLE-D398
        '54169' => '3222'             , # G+3222  U+D399  uniD399,  HANGUL SYLLABLE-D399
        '54170' => '17334'            , # G+17334  U+D39A  uniD39A,  HANGUL SYLLABLE-D39A
        '54171' => '17335'            , # G+17335  U+D39B  uniD39B,  HANGUL SYLLABLE-D39B
        '54172' => '3223'             , # G+3223  U+D39C  uniD39C,  HANGUL SYLLABLE-D39C
        '54173' => '17336'            , # G+17336  U+D39D  uniD39D,  HANGUL SYLLABLE-D39D
        '54174' => '17337'            , # G+17337  U+D39E  uniD39E,  HANGUL SYLLABLE-D39E
        '54175' => '17338'            , # G+17338  U+D39F  uniD39F,  HANGUL SYLLABLE-D39F
        '54176' => '3224'             , # G+3224  U+D3A0  uniD3A0,  HANGUL SYLLABLE-D3A0
        '54177' => '17339'            , # G+17339  U+D3A1  uniD3A1,  HANGUL SYLLABLE-D3A1
        '54178' => '17340'            , # G+17340  U+D3A2  uniD3A2,  HANGUL SYLLABLE-D3A2
        '54179' => '17341'            , # G+17341  U+D3A3  uniD3A3,  HANGUL SYLLABLE-D3A3
        '54180' => '17342'            , # G+17342  U+D3A4  uniD3A4,  HANGUL SYLLABLE-D3A4
        '54181' => '17343'            , # G+17343  U+D3A5  uniD3A5,  HANGUL SYLLABLE-D3A5
        '54182' => '17344'            , # G+17344  U+D3A6  uniD3A6,  HANGUL SYLLABLE-D3A6
        '54183' => '17345'            , # G+17345  U+D3A7  uniD3A7,  HANGUL SYLLABLE-D3A7
        '54184' => '3225'             , # G+3225  U+D3A8  uniD3A8,  HANGUL SYLLABLE-D3A8
        '54185' => '3226'             , # G+3226  U+D3A9  uniD3A9,  HANGUL SYLLABLE-D3A9
        '54186' => '17346'            , # G+17346  U+D3AA  uniD3AA,  HANGUL SYLLABLE-D3AA
        '54187' => '3227'             , # G+3227  U+D3AB  uniD3AB,  HANGUL SYLLABLE-D3AB
        '54188' => '17347'            , # G+17347  U+D3AC  uniD3AC,  HANGUL SYLLABLE-D3AC
        '54189' => '3228'             , # G+3228  U+D3AD  uniD3AD,  HANGUL SYLLABLE-D3AD
        '54190' => '17348'            , # G+17348  U+D3AE  uniD3AE,  HANGUL SYLLABLE-D3AE
        '54191' => '17349'            , # G+17349  U+D3AF  uniD3AF,  HANGUL SYLLABLE-D3AF
        '54192' => '17350'            , # G+17350  U+D3B0  uniD3B0,  HANGUL SYLLABLE-D3B0
        '54193' => '17351'            , # G+17351  U+D3B1  uniD3B1,  HANGUL SYLLABLE-D3B1
        '54194' => '17352'            , # G+17352  U+D3B2  uniD3B2,  HANGUL SYLLABLE-D3B2
        '54195' => '17353'            , # G+17353  U+D3B3  uniD3B3,  HANGUL SYLLABLE-D3B3
        '54196' => '3229'             , # G+3229  U+D3B4  uniD3B4,  HANGUL SYLLABLE-D3B4
        '54197' => '17354'            , # G+17354  U+D3B5  uniD3B5,  HANGUL SYLLABLE-D3B5
        '54198' => '17355'            , # G+17355  U+D3B6  uniD3B6,  HANGUL SYLLABLE-D3B6
        '54199' => '17356'            , # G+17356  U+D3B7  uniD3B7,  HANGUL SYLLABLE-D3B7
        '54200' => '3230'             , # G+3230  U+D3B8  uniD3B8,  HANGUL SYLLABLE-D3B8
        '54201' => '17357'            , # G+17357  U+D3B9  uniD3B9,  HANGUL SYLLABLE-D3B9
        '54202' => '17358'            , # G+17358  U+D3BA  uniD3BA,  HANGUL SYLLABLE-D3BA
        '54203' => '17359'            , # G+17359  U+D3BB  uniD3BB,  HANGUL SYLLABLE-D3BB
        '54204' => '3231'             , # G+3231  U+D3BC  uniD3BC,  HANGUL SYLLABLE-D3BC
        '54205' => '17360'            , # G+17360  U+D3BD  uniD3BD,  HANGUL SYLLABLE-D3BD
        '54206' => '17361'            , # G+17361  U+D3BE  uniD3BE,  HANGUL SYLLABLE-D3BE
        '54207' => '17362'            , # G+17362  U+D3BF  uniD3BF,  HANGUL SYLLABLE-D3BF
        '54208' => '17363'            , # G+17363  U+D3C0  uniD3C0,  HANGUL SYLLABLE-D3C0
        '54209' => '17364'            , # G+17364  U+D3C1  uniD3C1,  HANGUL SYLLABLE-D3C1
        '54210' => '17365'            , # G+17365  U+D3C2  uniD3C2,  HANGUL SYLLABLE-D3C2
        '54211' => '17366'            , # G+17366  U+D3C3  uniD3C3,  HANGUL SYLLABLE-D3C3
        '54212' => '3232'             , # G+3232  U+D3C4  uniD3C4,  HANGUL SYLLABLE-D3C4
        '54213' => '3233'             , # G+3233  U+D3C5  uniD3C5,  HANGUL SYLLABLE-D3C5
        '54214' => '17367'            , # G+17367  U+D3C6  uniD3C6,  HANGUL SYLLABLE-D3C6
        '54215' => '17368'            , # G+17368  U+D3C7  uniD3C7,  HANGUL SYLLABLE-D3C7
        '54216' => '3234'             , # G+3234  U+D3C8  uniD3C8,  HANGUL SYLLABLE-D3C8
        '54217' => '3235'             , # G+3235  U+D3C9  uniD3C9,  HANGUL SYLLABLE-D3C9
        '54218' => '17369'            , # G+17369  U+D3CA  uniD3CA,  HANGUL SYLLABLE-D3CA
        '54219' => '17370'            , # G+17370  U+D3CB  uniD3CB,  HANGUL SYLLABLE-D3CB
        '54220' => '17371'            , # G+17371  U+D3CC  uniD3CC,  HANGUL SYLLABLE-D3CC
        '54221' => '17372'            , # G+17372  U+D3CD  uniD3CD,  HANGUL SYLLABLE-D3CD
        '54222' => '17373'            , # G+17373  U+D3CE  uniD3CE,  HANGUL SYLLABLE-D3CE
        '54223' => '17374'            , # G+17374  U+D3CF  uniD3CF,  HANGUL SYLLABLE-D3CF
        '54224' => '3236'             , # G+3236  U+D3D0  uniD3D0,  HANGUL SYLLABLE-D3D0
        '54225' => '17375'            , # G+17375  U+D3D1  uniD3D1,  HANGUL SYLLABLE-D3D1
        '54226' => '17376'            , # G+17376  U+D3D2  uniD3D2,  HANGUL SYLLABLE-D3D2
        '54227' => '17377'            , # G+17377  U+D3D3  uniD3D3,  HANGUL SYLLABLE-D3D3
        '54228' => '17378'            , # G+17378  U+D3D4  uniD3D4,  HANGUL SYLLABLE-D3D4
        '54229' => '17379'            , # G+17379  U+D3D5  uniD3D5,  HANGUL SYLLABLE-D3D5
        '54230' => '17380'            , # G+17380  U+D3D6  uniD3D6,  HANGUL SYLLABLE-D3D6
        '54231' => '17381'            , # G+17381  U+D3D7  uniD3D7,  HANGUL SYLLABLE-D3D7
        '54232' => '3237'             , # G+3237  U+D3D8  uniD3D8,  HANGUL SYLLABLE-D3D8
        '54233' => '17382'            , # G+17382  U+D3D9  uniD3D9,  HANGUL SYLLABLE-D3D9
        '54234' => '17383'            , # G+17383  U+D3DA  uniD3DA,  HANGUL SYLLABLE-D3DA
        '54235' => '17384'            , # G+17384  U+D3DB  uniD3DB,  HANGUL SYLLABLE-D3DB
        '54236' => '17385'            , # G+17385  U+D3DC  uniD3DC,  HANGUL SYLLABLE-D3DC
        '54237' => '17386'            , # G+17386  U+D3DD  uniD3DD,  HANGUL SYLLABLE-D3DD
        '54238' => '17387'            , # G+17387  U+D3DE  uniD3DE,  HANGUL SYLLABLE-D3DE
        '54239' => '17388'            , # G+17388  U+D3DF  uniD3DF,  HANGUL SYLLABLE-D3DF
        '54240' => '17389'            , # G+17389  U+D3E0  uniD3E0,  HANGUL SYLLABLE-D3E0
        '54241' => '3238'             , # G+3238  U+D3E1  uniD3E1,  HANGUL SYLLABLE-D3E1
        '54242' => '17390'            , # G+17390  U+D3E2  uniD3E2,  HANGUL SYLLABLE-D3E2
        '54243' => '3239'             , # G+3239  U+D3E3  uniD3E3,  HANGUL SYLLABLE-D3E3
        '54244' => '17391'            , # G+17391  U+D3E4  uniD3E4,  HANGUL SYLLABLE-D3E4
        '54245' => '17392'            , # G+17392  U+D3E5  uniD3E5,  HANGUL SYLLABLE-D3E5
        '54246' => '17393'            , # G+17393  U+D3E6  uniD3E6,  HANGUL SYLLABLE-D3E6
        '54247' => '17394'            , # G+17394  U+D3E7  uniD3E7,  HANGUL SYLLABLE-D3E7
        '54248' => '17395'            , # G+17395  U+D3E8  uniD3E8,  HANGUL SYLLABLE-D3E8
        '54249' => '17396'            , # G+17396  U+D3E9  uniD3E9,  HANGUL SYLLABLE-D3E9
        '54250' => '17397'            , # G+17397  U+D3EA  uniD3EA,  HANGUL SYLLABLE-D3EA
        '54251' => '17398'            , # G+17398  U+D3EB  uniD3EB,  HANGUL SYLLABLE-D3EB
        '54252' => '3240'             , # G+3240  U+D3EC  uniD3EC,  HANGUL SYLLABLE-D3EC
        '54253' => '3241'             , # G+3241  U+D3ED  uniD3ED,  HANGUL SYLLABLE-D3ED
        '54254' => '17399'            , # G+17399  U+D3EE  uniD3EE,  HANGUL SYLLABLE-D3EE
        '54255' => '17400'            , # G+17400  U+D3EF  uniD3EF,  HANGUL SYLLABLE-D3EF
        '54256' => '3242'             , # G+3242  U+D3F0  uniD3F0,  HANGUL SYLLABLE-D3F0
        '54257' => '17401'            , # G+17401  U+D3F1  uniD3F1,  HANGUL SYLLABLE-D3F1
        '54258' => '17402'            , # G+17402  U+D3F2  uniD3F2,  HANGUL SYLLABLE-D3F2
        '54259' => '17403'            , # G+17403  U+D3F3  uniD3F3,  HANGUL SYLLABLE-D3F3
        '54260' => '3243'             , # G+3243  U+D3F4  uniD3F4,  HANGUL SYLLABLE-D3F4
        '54261' => '17404'            , # G+17404  U+D3F5  uniD3F5,  HANGUL SYLLABLE-D3F5
        '54262' => '17405'            , # G+17405  U+D3F6  uniD3F6,  HANGUL SYLLABLE-D3F6
        '54263' => '17406'            , # G+17406  U+D3F7  uniD3F7,  HANGUL SYLLABLE-D3F7
        '54264' => '17407'            , # G+17407  U+D3F8  uniD3F8,  HANGUL SYLLABLE-D3F8
        '54265' => '17408'            , # G+17408  U+D3F9  uniD3F9,  HANGUL SYLLABLE-D3F9
        '54266' => '17409'            , # G+17409  U+D3FA  uniD3FA,  HANGUL SYLLABLE-D3FA
        '54267' => '17410'            , # G+17410  U+D3FB  uniD3FB,  HANGUL SYLLABLE-D3FB
        '54268' => '3244'             , # G+3244  U+D3FC  uniD3FC,  HANGUL SYLLABLE-D3FC
        '54269' => '3245'             , # G+3245  U+D3FD  uniD3FD,  HANGUL SYLLABLE-D3FD
        '54270' => '17411'            , # G+17411  U+D3FE  uniD3FE,  HANGUL SYLLABLE-D3FE
        '54271' => '3246'             , # G+3246  U+D3FF  uniD3FF,  HANGUL SYLLABLE-D3FF
        '54272' => '17412'            , # G+17412  U+D400  uniD400,  HANGUL SYLLABLE-D400
        '54273' => '3247'             , # G+3247  U+D401  uniD401,  HANGUL SYLLABLE-D401
        '54274' => '17413'            , # G+17413  U+D402  uniD402,  HANGUL SYLLABLE-D402
        '54275' => '17414'            , # G+17414  U+D403  uniD403,  HANGUL SYLLABLE-D403
        '54276' => '17415'            , # G+17415  U+D404  uniD404,  HANGUL SYLLABLE-D404
        '54277' => '17416'            , # G+17416  U+D405  uniD405,  HANGUL SYLLABLE-D405
        '54278' => '17417'            , # G+17417  U+D406  uniD406,  HANGUL SYLLABLE-D406
        '54279' => '17418'            , # G+17418  U+D407  uniD407,  HANGUL SYLLABLE-D407
        '54280' => '3248'             , # G+3248  U+D408  uniD408,  HANGUL SYLLABLE-D408
        '54281' => '17419'            , # G+17419  U+D409  uniD409,  HANGUL SYLLABLE-D409
        '54282' => '17420'            , # G+17420  U+D40A  uniD40A,  HANGUL SYLLABLE-D40A
        '54283' => '17421'            , # G+17421  U+D40B  uniD40B,  HANGUL SYLLABLE-D40B
        '54284' => '17422'            , # G+17422  U+D40C  uniD40C,  HANGUL SYLLABLE-D40C
        '54285' => '17423'            , # G+17423  U+D40D  uniD40D,  HANGUL SYLLABLE-D40D
        '54286' => '17424'            , # G+17424  U+D40E  uniD40E,  HANGUL SYLLABLE-D40E
        '54287' => '17425'            , # G+17425  U+D40F  uniD40F,  HANGUL SYLLABLE-D40F
        '54288' => '17426'            , # G+17426  U+D410  uniD410,  HANGUL SYLLABLE-D410
        '54289' => '17427'            , # G+17427  U+D411  uniD411,  HANGUL SYLLABLE-D411
        '54290' => '17428'            , # G+17428  U+D412  uniD412,  HANGUL SYLLABLE-D412
        '54291' => '17429'            , # G+17429  U+D413  uniD413,  HANGUL SYLLABLE-D413
        '54292' => '17430'            , # G+17430  U+D414  uniD414,  HANGUL SYLLABLE-D414
        '54293' => '17431'            , # G+17431  U+D415  uniD415,  HANGUL SYLLABLE-D415
        '54294' => '17432'            , # G+17432  U+D416  uniD416,  HANGUL SYLLABLE-D416
        '54295' => '17433'            , # G+17433  U+D417  uniD417,  HANGUL SYLLABLE-D417
        '54296' => '17434'            , # G+17434  U+D418  uniD418,  HANGUL SYLLABLE-D418
        '54297' => '17435'            , # G+17435  U+D419  uniD419,  HANGUL SYLLABLE-D419
        '54298' => '17436'            , # G+17436  U+D41A  uniD41A,  HANGUL SYLLABLE-D41A
        '54299' => '17437'            , # G+17437  U+D41B  uniD41B,  HANGUL SYLLABLE-D41B
        '54300' => '17438'            , # G+17438  U+D41C  uniD41C,  HANGUL SYLLABLE-D41C
        '54301' => '3249'             , # G+3249  U+D41D  uniD41D,  HANGUL SYLLABLE-D41D
        '54302' => '17439'            , # G+17439  U+D41E  uniD41E,  HANGUL SYLLABLE-D41E
        '54303' => '17440'            , # G+17440  U+D41F  uniD41F,  HANGUL SYLLABLE-D41F
        '54304' => '17441'            , # G+17441  U+D420  uniD420,  HANGUL SYLLABLE-D420
        '54305' => '17442'            , # G+17442  U+D421  uniD421,  HANGUL SYLLABLE-D421
        '54306' => '17443'            , # G+17443  U+D422  uniD422,  HANGUL SYLLABLE-D422
        '54307' => '17444'            , # G+17444  U+D423  uniD423,  HANGUL SYLLABLE-D423
        '54308' => '17445'            , # G+17445  U+D424  uniD424,  HANGUL SYLLABLE-D424
        '54309' => '17446'            , # G+17446  U+D425  uniD425,  HANGUL SYLLABLE-D425
        '54310' => '17447'            , # G+17447  U+D426  uniD426,  HANGUL SYLLABLE-D426
        '54311' => '17448'            , # G+17448  U+D427  uniD427,  HANGUL SYLLABLE-D427
        '54312' => '17449'            , # G+17449  U+D428  uniD428,  HANGUL SYLLABLE-D428
        '54313' => '17450'            , # G+17450  U+D429  uniD429,  HANGUL SYLLABLE-D429
        '54314' => '17451'            , # G+17451  U+D42A  uniD42A,  HANGUL SYLLABLE-D42A
        '54315' => '17452'            , # G+17452  U+D42B  uniD42B,  HANGUL SYLLABLE-D42B
        '54316' => '17453'            , # G+17453  U+D42C  uniD42C,  HANGUL SYLLABLE-D42C
        '54317' => '17454'            , # G+17454  U+D42D  uniD42D,  HANGUL SYLLABLE-D42D
        '54318' => '17455'            , # G+17455  U+D42E  uniD42E,  HANGUL SYLLABLE-D42E
        '54319' => '17456'            , # G+17456  U+D42F  uniD42F,  HANGUL SYLLABLE-D42F
        '54320' => '17457'            , # G+17457  U+D430  uniD430,  HANGUL SYLLABLE-D430
        '54321' => '17458'            , # G+17458  U+D431  uniD431,  HANGUL SYLLABLE-D431
        '54322' => '17459'            , # G+17459  U+D432  uniD432,  HANGUL SYLLABLE-D432
        '54323' => '17460'            , # G+17460  U+D433  uniD433,  HANGUL SYLLABLE-D433
        '54324' => '17461'            , # G+17461  U+D434  uniD434,  HANGUL SYLLABLE-D434
        '54325' => '17462'            , # G+17462  U+D435  uniD435,  HANGUL SYLLABLE-D435
        '54326' => '17463'            , # G+17463  U+D436  uniD436,  HANGUL SYLLABLE-D436
        '54327' => '17464'            , # G+17464  U+D437  uniD437,  HANGUL SYLLABLE-D437
        '54328' => '17465'            , # G+17465  U+D438  uniD438,  HANGUL SYLLABLE-D438
        '54329' => '17466'            , # G+17466  U+D439  uniD439,  HANGUL SYLLABLE-D439
        '54330' => '17467'            , # G+17467  U+D43A  uniD43A,  HANGUL SYLLABLE-D43A
        '54331' => '17468'            , # G+17468  U+D43B  uniD43B,  HANGUL SYLLABLE-D43B
        '54332' => '17469'            , # G+17469  U+D43C  uniD43C,  HANGUL SYLLABLE-D43C
        '54333' => '17470'            , # G+17470  U+D43D  uniD43D,  HANGUL SYLLABLE-D43D
        '54334' => '17471'            , # G+17471  U+D43E  uniD43E,  HANGUL SYLLABLE-D43E
        '54335' => '17472'            , # G+17472  U+D43F  uniD43F,  HANGUL SYLLABLE-D43F
        '54336' => '3250'             , # G+3250  U+D440  uniD440,  HANGUL SYLLABLE-D440
        '54337' => '17473'            , # G+17473  U+D441  uniD441,  HANGUL SYLLABLE-D441
        '54338' => '17474'            , # G+17474  U+D442  uniD442,  HANGUL SYLLABLE-D442
        '54339' => '17475'            , # G+17475  U+D443  uniD443,  HANGUL SYLLABLE-D443
        '54340' => '3251'             , # G+3251  U+D444  uniD444,  HANGUL SYLLABLE-D444
        '54341' => '17476'            , # G+17476  U+D445  uniD445,  HANGUL SYLLABLE-D445
        '54342' => '17477'            , # G+17477  U+D446  uniD446,  HANGUL SYLLABLE-D446
        '54343' => '17478'            , # G+17478  U+D447  uniD447,  HANGUL SYLLABLE-D447
        '54344' => '17479'            , # G+17479  U+D448  uniD448,  HANGUL SYLLABLE-D448
        '54345' => '17480'            , # G+17480  U+D449  uniD449,  HANGUL SYLLABLE-D449
        '54346' => '17481'            , # G+17481  U+D44A  uniD44A,  HANGUL SYLLABLE-D44A
        '54347' => '17482'            , # G+17482  U+D44B  uniD44B,  HANGUL SYLLABLE-D44B
        '54348' => '17483'            , # G+17483  U+D44C  uniD44C,  HANGUL SYLLABLE-D44C
        '54349' => '17484'            , # G+17484  U+D44D  uniD44D,  HANGUL SYLLABLE-D44D
        '54350' => '17485'            , # G+17485  U+D44E  uniD44E,  HANGUL SYLLABLE-D44E
        '54351' => '17486'            , # G+17486  U+D44F  uniD44F,  HANGUL SYLLABLE-D44F
        '54352' => '17487'            , # G+17487  U+D450  uniD450,  HANGUL SYLLABLE-D450
        '54353' => '17488'            , # G+17488  U+D451  uniD451,  HANGUL SYLLABLE-D451
        '54354' => '17489'            , # G+17489  U+D452  uniD452,  HANGUL SYLLABLE-D452
        '54355' => '17490'            , # G+17490  U+D453  uniD453,  HANGUL SYLLABLE-D453
        '54356' => '17491'            , # G+17491  U+D454  uniD454,  HANGUL SYLLABLE-D454
        '54357' => '17492'            , # G+17492  U+D455  uniD455,  HANGUL SYLLABLE-D455
        '54358' => '17493'            , # G+17493  U+D456  uniD456,  HANGUL SYLLABLE-D456
        '54359' => '17494'            , # G+17494  U+D457  uniD457,  HANGUL SYLLABLE-D457
        '54360' => '17495'            , # G+17495  U+D458  uniD458,  HANGUL SYLLABLE-D458
        '54361' => '17496'            , # G+17496  U+D459  uniD459,  HANGUL SYLLABLE-D459
        '54362' => '17497'            , # G+17497  U+D45A  uniD45A,  HANGUL SYLLABLE-D45A
        '54363' => '17498'            , # G+17498  U+D45B  uniD45B,  HANGUL SYLLABLE-D45B
        '54364' => '3252'             , # G+3252  U+D45C  uniD45C,  HANGUL SYLLABLE-D45C
        '54365' => '17499'            , # G+17499  U+D45D  uniD45D,  HANGUL SYLLABLE-D45D
        '54366' => '17500'            , # G+17500  U+D45E  uniD45E,  HANGUL SYLLABLE-D45E
        '54367' => '17501'            , # G+17501  U+D45F  uniD45F,  HANGUL SYLLABLE-D45F
        '54368' => '3253'             , # G+3253  U+D460  uniD460,  HANGUL SYLLABLE-D460
        '54369' => '17502'            , # G+17502  U+D461  uniD461,  HANGUL SYLLABLE-D461
        '54370' => '17503'            , # G+17503  U+D462  uniD462,  HANGUL SYLLABLE-D462
        '54371' => '17504'            , # G+17504  U+D463  uniD463,  HANGUL SYLLABLE-D463
        '54372' => '3254'             , # G+3254  U+D464  uniD464,  HANGUL SYLLABLE-D464
        '54373' => '17505'            , # G+17505  U+D465  uniD465,  HANGUL SYLLABLE-D465
        '54374' => '17506'            , # G+17506  U+D466  uniD466,  HANGUL SYLLABLE-D466
        '54375' => '17507'            , # G+17507  U+D467  uniD467,  HANGUL SYLLABLE-D467
        '54376' => '17508'            , # G+17508  U+D468  uniD468,  HANGUL SYLLABLE-D468
        '54377' => '17509'            , # G+17509  U+D469  uniD469,  HANGUL SYLLABLE-D469
        '54378' => '17510'            , # G+17510  U+D46A  uniD46A,  HANGUL SYLLABLE-D46A
        '54379' => '17511'            , # G+17511  U+D46B  uniD46B,  HANGUL SYLLABLE-D46B
        '54380' => '17512'            , # G+17512  U+D46C  uniD46C,  HANGUL SYLLABLE-D46C
        '54381' => '3255'             , # G+3255  U+D46D  uniD46D,  HANGUL SYLLABLE-D46D
        '54382' => '17513'            , # G+17513  U+D46E  uniD46E,  HANGUL SYLLABLE-D46E
        '54383' => '3256'             , # G+3256  U+D46F  uniD46F,  HANGUL SYLLABLE-D46F
        '54384' => '17514'            , # G+17514  U+D470  uniD470,  HANGUL SYLLABLE-D470
        '54385' => '17515'            , # G+17515  U+D471  uniD471,  HANGUL SYLLABLE-D471
        '54386' => '17516'            , # G+17516  U+D472  uniD472,  HANGUL SYLLABLE-D472
        '54387' => '17517'            , # G+17517  U+D473  uniD473,  HANGUL SYLLABLE-D473
        '54388' => '17518'            , # G+17518  U+D474  uniD474,  HANGUL SYLLABLE-D474
        '54389' => '17519'            , # G+17519  U+D475  uniD475,  HANGUL SYLLABLE-D475
        '54390' => '17520'            , # G+17520  U+D476  uniD476,  HANGUL SYLLABLE-D476
        '54391' => '17521'            , # G+17521  U+D477  uniD477,  HANGUL SYLLABLE-D477
        '54392' => '3257'             , # G+3257  U+D478  uniD478,  HANGUL SYLLABLE-D478
        '54393' => '3258'             , # G+3258  U+D479  uniD479,  HANGUL SYLLABLE-D479
        '54394' => '17522'            , # G+17522  U+D47A  uniD47A,  HANGUL SYLLABLE-D47A
        '54395' => '17523'            , # G+17523  U+D47B  uniD47B,  HANGUL SYLLABLE-D47B
        '54396' => '3259'             , # G+3259  U+D47C  uniD47C,  HANGUL SYLLABLE-D47C
        '54397' => '17524'            , # G+17524  U+D47D  uniD47D,  HANGUL SYLLABLE-D47D
        '54398' => '17525'            , # G+17525  U+D47E  uniD47E,  HANGUL SYLLABLE-D47E
        '54399' => '3260'             , # G+3260  U+D47F  uniD47F,  HANGUL SYLLABLE-D47F
        '54400' => '3261'             , # G+3261  U+D480  uniD480,  HANGUL SYLLABLE-D480
        '54401' => '17526'            , # G+17526  U+D481  uniD481,  HANGUL SYLLABLE-D481
        '54402' => '3262'             , # G+3262  U+D482  uniD482,  HANGUL SYLLABLE-D482
        '54403' => '17527'            , # G+17527  U+D483  uniD483,  HANGUL SYLLABLE-D483
        '54404' => '17528'            , # G+17528  U+D484  uniD484,  HANGUL SYLLABLE-D484
        '54405' => '17529'            , # G+17529  U+D485  uniD485,  HANGUL SYLLABLE-D485
        '54406' => '17530'            , # G+17530  U+D486  uniD486,  HANGUL SYLLABLE-D486
        '54407' => '17531'            , # G+17531  U+D487  uniD487,  HANGUL SYLLABLE-D487
        '54408' => '3263'             , # G+3263  U+D488  uniD488,  HANGUL SYLLABLE-D488
        '54409' => '3264'             , # G+3264  U+D489  uniD489,  HANGUL SYLLABLE-D489
        '54410' => '17532'            , # G+17532  U+D48A  uniD48A,  HANGUL SYLLABLE-D48A
        '54411' => '3265'             , # G+3265  U+D48B  uniD48B,  HANGUL SYLLABLE-D48B
        '54412' => '17533'            , # G+17533  U+D48C  uniD48C,  HANGUL SYLLABLE-D48C
        '54413' => '3266'             , # G+3266  U+D48D  uniD48D,  HANGUL SYLLABLE-D48D
        '54414' => '17534'            , # G+17534  U+D48E  uniD48E,  HANGUL SYLLABLE-D48E
        '54415' => '17535'            , # G+17535  U+D48F  uniD48F,  HANGUL SYLLABLE-D48F
        '54416' => '17536'            , # G+17536  U+D490  uniD490,  HANGUL SYLLABLE-D490
        '54417' => '17537'            , # G+17537  U+D491  uniD491,  HANGUL SYLLABLE-D491
        '54418' => '17538'            , # G+17538  U+D492  uniD492,  HANGUL SYLLABLE-D492
        '54419' => '17539'            , # G+17539  U+D493  uniD493,  HANGUL SYLLABLE-D493
        '54420' => '3267'             , # G+3267  U+D494  uniD494,  HANGUL SYLLABLE-D494
        '54421' => '17540'            , # G+17540  U+D495  uniD495,  HANGUL SYLLABLE-D495
        '54422' => '17541'            , # G+17541  U+D496  uniD496,  HANGUL SYLLABLE-D496
        '54423' => '17542'            , # G+17542  U+D497  uniD497,  HANGUL SYLLABLE-D497
        '54424' => '17543'            , # G+17543  U+D498  uniD498,  HANGUL SYLLABLE-D498
        '54425' => '17544'            , # G+17544  U+D499  uniD499,  HANGUL SYLLABLE-D499
        '54426' => '17545'            , # G+17545  U+D49A  uniD49A,  HANGUL SYLLABLE-D49A
        '54427' => '17546'            , # G+17546  U+D49B  uniD49B,  HANGUL SYLLABLE-D49B
        '54428' => '17547'            , # G+17547  U+D49C  uniD49C,  HANGUL SYLLABLE-D49C
        '54429' => '17548'            , # G+17548  U+D49D  uniD49D,  HANGUL SYLLABLE-D49D
        '54430' => '17549'            , # G+17549  U+D49E  uniD49E,  HANGUL SYLLABLE-D49E
        '54431' => '17550'            , # G+17550  U+D49F  uniD49F,  HANGUL SYLLABLE-D49F
        '54432' => '17551'            , # G+17551  U+D4A0  uniD4A0,  HANGUL SYLLABLE-D4A0
        '54433' => '17552'            , # G+17552  U+D4A1  uniD4A1,  HANGUL SYLLABLE-D4A1
        '54434' => '17553'            , # G+17553  U+D4A2  uniD4A2,  HANGUL SYLLABLE-D4A2
        '54435' => '17554'            , # G+17554  U+D4A3  uniD4A3,  HANGUL SYLLABLE-D4A3
        '54436' => '17555'            , # G+17555  U+D4A4  uniD4A4,  HANGUL SYLLABLE-D4A4
        '54437' => '17556'            , # G+17556  U+D4A5  uniD4A5,  HANGUL SYLLABLE-D4A5
        '54438' => '17557'            , # G+17557  U+D4A6  uniD4A6,  HANGUL SYLLABLE-D4A6
        '54439' => '17558'            , # G+17558  U+D4A7  uniD4A7,  HANGUL SYLLABLE-D4A7
        '54440' => '17559'            , # G+17559  U+D4A8  uniD4A8,  HANGUL SYLLABLE-D4A8
        '54441' => '3268'             , # G+3268  U+D4A9  uniD4A9,  HANGUL SYLLABLE-D4A9
        '54442' => '17560'            , # G+17560  U+D4AA  uniD4AA,  HANGUL SYLLABLE-D4AA
        '54443' => '17561'            , # G+17561  U+D4AB  uniD4AB,  HANGUL SYLLABLE-D4AB
        '54444' => '17562'            , # G+17562  U+D4AC  uniD4AC,  HANGUL SYLLABLE-D4AC
        '54445' => '17563'            , # G+17563  U+D4AD  uniD4AD,  HANGUL SYLLABLE-D4AD
        '54446' => '17564'            , # G+17564  U+D4AE  uniD4AE,  HANGUL SYLLABLE-D4AE
        '54447' => '17565'            , # G+17565  U+D4AF  uniD4AF,  HANGUL SYLLABLE-D4AF
        '54448' => '17566'            , # G+17566  U+D4B0  uniD4B0,  HANGUL SYLLABLE-D4B0
        '54449' => '17567'            , # G+17567  U+D4B1  uniD4B1,  HANGUL SYLLABLE-D4B1
        '54450' => '17568'            , # G+17568  U+D4B2  uniD4B2,  HANGUL SYLLABLE-D4B2
        '54451' => '17569'            , # G+17569  U+D4B3  uniD4B3,  HANGUL SYLLABLE-D4B3
        '54452' => '17570'            , # G+17570  U+D4B4  uniD4B4,  HANGUL SYLLABLE-D4B4
        '54453' => '17571'            , # G+17571  U+D4B5  uniD4B5,  HANGUL SYLLABLE-D4B5
        '54454' => '17572'            , # G+17572  U+D4B6  uniD4B6,  HANGUL SYLLABLE-D4B6
        '54455' => '17573'            , # G+17573  U+D4B7  uniD4B7,  HANGUL SYLLABLE-D4B7
        '54456' => '17574'            , # G+17574  U+D4B8  uniD4B8,  HANGUL SYLLABLE-D4B8
        '54457' => '17575'            , # G+17575  U+D4B9  uniD4B9,  HANGUL SYLLABLE-D4B9
        '54458' => '17576'            , # G+17576  U+D4BA  uniD4BA,  HANGUL SYLLABLE-D4BA
        '54459' => '17577'            , # G+17577  U+D4BB  uniD4BB,  HANGUL SYLLABLE-D4BB
        '54460' => '17578'            , # G+17578  U+D4BC  uniD4BC,  HANGUL SYLLABLE-D4BC
        '54461' => '17579'            , # G+17579  U+D4BD  uniD4BD,  HANGUL SYLLABLE-D4BD
        '54462' => '17580'            , # G+17580  U+D4BE  uniD4BE,  HANGUL SYLLABLE-D4BE
        '54463' => '17581'            , # G+17581  U+D4BF  uniD4BF,  HANGUL SYLLABLE-D4BF
        '54464' => '17582'            , # G+17582  U+D4C0  uniD4C0,  HANGUL SYLLABLE-D4C0
        '54465' => '17583'            , # G+17583  U+D4C1  uniD4C1,  HANGUL SYLLABLE-D4C1
        '54466' => '17584'            , # G+17584  U+D4C2  uniD4C2,  HANGUL SYLLABLE-D4C2
        '54467' => '17585'            , # G+17585  U+D4C3  uniD4C3,  HANGUL SYLLABLE-D4C3
        '54468' => '17586'            , # G+17586  U+D4C4  uniD4C4,  HANGUL SYLLABLE-D4C4
        '54469' => '17587'            , # G+17587  U+D4C5  uniD4C5,  HANGUL SYLLABLE-D4C5
        '54470' => '17588'            , # G+17588  U+D4C6  uniD4C6,  HANGUL SYLLABLE-D4C6
        '54471' => '17589'            , # G+17589  U+D4C7  uniD4C7,  HANGUL SYLLABLE-D4C7
        '54472' => '17590'            , # G+17590  U+D4C8  uniD4C8,  HANGUL SYLLABLE-D4C8
        '54473' => '17591'            , # G+17591  U+D4C9  uniD4C9,  HANGUL SYLLABLE-D4C9
        '54474' => '17592'            , # G+17592  U+D4CA  uniD4CA,  HANGUL SYLLABLE-D4CA
        '54475' => '17593'            , # G+17593  U+D4CB  uniD4CB,  HANGUL SYLLABLE-D4CB
        '54476' => '3269'             , # G+3269  U+D4CC  uniD4CC,  HANGUL SYLLABLE-D4CC
        '54477' => '17594'            , # G+17594  U+D4CD  uniD4CD,  HANGUL SYLLABLE-D4CD
        '54478' => '17595'            , # G+17595  U+D4CE  uniD4CE,  HANGUL SYLLABLE-D4CE
        '54479' => '17596'            , # G+17596  U+D4CF  uniD4CF,  HANGUL SYLLABLE-D4CF
        '54480' => '3270'             , # G+3270  U+D4D0  uniD4D0,  HANGUL SYLLABLE-D4D0
        '54481' => '17597'            , # G+17597  U+D4D1  uniD4D1,  HANGUL SYLLABLE-D4D1
        '54482' => '17598'            , # G+17598  U+D4D2  uniD4D2,  HANGUL SYLLABLE-D4D2
        '54483' => '17599'            , # G+17599  U+D4D3  uniD4D3,  HANGUL SYLLABLE-D4D3
        '54484' => '3271'             , # G+3271  U+D4D4  uniD4D4,  HANGUL SYLLABLE-D4D4
        '54485' => '17600'            , # G+17600  U+D4D5  uniD4D5,  HANGUL SYLLABLE-D4D5
        '54486' => '17601'            , # G+17601  U+D4D6  uniD4D6,  HANGUL SYLLABLE-D4D6
        '54487' => '17602'            , # G+17602  U+D4D7  uniD4D7,  HANGUL SYLLABLE-D4D7
        '54488' => '17603'            , # G+17603  U+D4D8  uniD4D8,  HANGUL SYLLABLE-D4D8
        '54489' => '17604'            , # G+17604  U+D4D9  uniD4D9,  HANGUL SYLLABLE-D4D9
        '54490' => '17605'            , # G+17605  U+D4DA  uniD4DA,  HANGUL SYLLABLE-D4DA
        '54491' => '17606'            , # G+17606  U+D4DB  uniD4DB,  HANGUL SYLLABLE-D4DB
        '54492' => '3272'             , # G+3272  U+D4DC  uniD4DC,  HANGUL SYLLABLE-D4DC
        '54493' => '17607'            , # G+17607  U+D4DD  uniD4DD,  HANGUL SYLLABLE-D4DD
        '54494' => '17608'            , # G+17608  U+D4DE  uniD4DE,  HANGUL SYLLABLE-D4DE
        '54495' => '3273'             , # G+3273  U+D4DF  uniD4DF,  HANGUL SYLLABLE-D4DF
        '54496' => '17609'            , # G+17609  U+D4E0  uniD4E0,  HANGUL SYLLABLE-D4E0
        '54497' => '17610'            , # G+17610  U+D4E1  uniD4E1,  HANGUL SYLLABLE-D4E1
        '54498' => '17611'            , # G+17611  U+D4E2  uniD4E2,  HANGUL SYLLABLE-D4E2
        '54499' => '17612'            , # G+17612  U+D4E3  uniD4E3,  HANGUL SYLLABLE-D4E3
        '54500' => '17613'            , # G+17613  U+D4E4  uniD4E4,  HANGUL SYLLABLE-D4E4
        '54501' => '17614'            , # G+17614  U+D4E5  uniD4E5,  HANGUL SYLLABLE-D4E5
        '54502' => '17615'            , # G+17615  U+D4E6  uniD4E6,  HANGUL SYLLABLE-D4E6
        '54503' => '17616'            , # G+17616  U+D4E7  uniD4E7,  HANGUL SYLLABLE-D4E7
        '54504' => '3274'             , # G+3274  U+D4E8  uniD4E8,  HANGUL SYLLABLE-D4E8
        '54505' => '17617'            , # G+17617  U+D4E9  uniD4E9,  HANGUL SYLLABLE-D4E9
        '54506' => '17618'            , # G+17618  U+D4EA  uniD4EA,  HANGUL SYLLABLE-D4EA
        '54507' => '17619'            , # G+17619  U+D4EB  uniD4EB,  HANGUL SYLLABLE-D4EB
        '54508' => '3275'             , # G+3275  U+D4EC  uniD4EC,  HANGUL SYLLABLE-D4EC
        '54509' => '17620'            , # G+17620  U+D4ED  uniD4ED,  HANGUL SYLLABLE-D4ED
        '54510' => '17621'            , # G+17621  U+D4EE  uniD4EE,  HANGUL SYLLABLE-D4EE
        '54511' => '17622'            , # G+17622  U+D4EF  uniD4EF,  HANGUL SYLLABLE-D4EF
        '54512' => '3276'             , # G+3276  U+D4F0  uniD4F0,  HANGUL SYLLABLE-D4F0
        '54513' => '17623'            , # G+17623  U+D4F1  uniD4F1,  HANGUL SYLLABLE-D4F1
        '54514' => '17624'            , # G+17624  U+D4F2  uniD4F2,  HANGUL SYLLABLE-D4F2
        '54515' => '17625'            , # G+17625  U+D4F3  uniD4F3,  HANGUL SYLLABLE-D4F3
        '54516' => '17626'            , # G+17626  U+D4F4  uniD4F4,  HANGUL SYLLABLE-D4F4
        '54517' => '17627'            , # G+17627  U+D4F5  uniD4F5,  HANGUL SYLLABLE-D4F5
        '54518' => '17628'            , # G+17628  U+D4F6  uniD4F6,  HANGUL SYLLABLE-D4F6
        '54519' => '17629'            , # G+17629  U+D4F7  uniD4F7,  HANGUL SYLLABLE-D4F7
        '54520' => '3277'             , # G+3277  U+D4F8  uniD4F8,  HANGUL SYLLABLE-D4F8
        '54521' => '17630'            , # G+17630  U+D4F9  uniD4F9,  HANGUL SYLLABLE-D4F9
        '54522' => '17631'            , # G+17631  U+D4FA  uniD4FA,  HANGUL SYLLABLE-D4FA
        '54523' => '3278'             , # G+3278  U+D4FB  uniD4FB,  HANGUL SYLLABLE-D4FB
        '54524' => '17632'            , # G+17632  U+D4FC  uniD4FC,  HANGUL SYLLABLE-D4FC
        '54525' => '3279'             , # G+3279  U+D4FD  uniD4FD,  HANGUL SYLLABLE-D4FD
        '54526' => '17633'            , # G+17633  U+D4FE  uniD4FE,  HANGUL SYLLABLE-D4FE
        '54527' => '17634'            , # G+17634  U+D4FF  uniD4FF,  HANGUL SYLLABLE-D4FF
        '54528' => '17635'            , # G+17635  U+D500  uniD500,  HANGUL SYLLABLE-D500
        '54529' => '17636'            , # G+17636  U+D501  uniD501,  HANGUL SYLLABLE-D501
        '54530' => '17637'            , # G+17637  U+D502  uniD502,  HANGUL SYLLABLE-D502
        '54531' => '17638'            , # G+17638  U+D503  uniD503,  HANGUL SYLLABLE-D503
        '54532' => '3280'             , # G+3280  U+D504  uniD504,  HANGUL SYLLABLE-D504
        '54533' => '17639'            , # G+17639  U+D505  uniD505,  HANGUL SYLLABLE-D505
        '54534' => '17640'            , # G+17640  U+D506  uniD506,  HANGUL SYLLABLE-D506
        '54535' => '17641'            , # G+17641  U+D507  uniD507,  HANGUL SYLLABLE-D507
        '54536' => '3281'             , # G+3281  U+D508  uniD508,  HANGUL SYLLABLE-D508
        '54537' => '17642'            , # G+17642  U+D509  uniD509,  HANGUL SYLLABLE-D509
        '54538' => '17643'            , # G+17643  U+D50A  uniD50A,  HANGUL SYLLABLE-D50A
        '54539' => '17644'            , # G+17644  U+D50B  uniD50B,  HANGUL SYLLABLE-D50B
        '54540' => '3282'             , # G+3282  U+D50C  uniD50C,  HANGUL SYLLABLE-D50C
        '54541' => '17645'            , # G+17645  U+D50D  uniD50D,  HANGUL SYLLABLE-D50D
        '54542' => '17646'            , # G+17646  U+D50E  uniD50E,  HANGUL SYLLABLE-D50E
        '54543' => '17647'            , # G+17647  U+D50F  uniD50F,  HANGUL SYLLABLE-D50F
        '54544' => '17648'            , # G+17648  U+D510  uniD510,  HANGUL SYLLABLE-D510
        '54545' => '17649'            , # G+17649  U+D511  uniD511,  HANGUL SYLLABLE-D511
        '54546' => '17650'            , # G+17650  U+D512  uniD512,  HANGUL SYLLABLE-D512
        '54547' => '17651'            , # G+17651  U+D513  uniD513,  HANGUL SYLLABLE-D513
        '54548' => '3283'             , # G+3283  U+D514  uniD514,  HANGUL SYLLABLE-D514
        '54549' => '3284'             , # G+3284  U+D515  uniD515,  HANGUL SYLLABLE-D515
        '54550' => '17652'            , # G+17652  U+D516  uniD516,  HANGUL SYLLABLE-D516
        '54551' => '3285'             , # G+3285  U+D517  uniD517,  HANGUL SYLLABLE-D517
        '54552' => '17653'            , # G+17653  U+D518  uniD518,  HANGUL SYLLABLE-D518
        '54553' => '17654'            , # G+17654  U+D519  uniD519,  HANGUL SYLLABLE-D519
        '54554' => '17655'            , # G+17655  U+D51A  uniD51A,  HANGUL SYLLABLE-D51A
        '54555' => '17656'            , # G+17656  U+D51B  uniD51B,  HANGUL SYLLABLE-D51B
        '54556' => '17657'            , # G+17657  U+D51C  uniD51C,  HANGUL SYLLABLE-D51C
        '54557' => '17658'            , # G+17658  U+D51D  uniD51D,  HANGUL SYLLABLE-D51D
        '54558' => '17659'            , # G+17659  U+D51E  uniD51E,  HANGUL SYLLABLE-D51E
        '54559' => '17660'            , # G+17660  U+D51F  uniD51F,  HANGUL SYLLABLE-D51F
        '54560' => '17661'            , # G+17661  U+D520  uniD520,  HANGUL SYLLABLE-D520
        '54561' => '17662'            , # G+17662  U+D521  uniD521,  HANGUL SYLLABLE-D521
        '54562' => '17663'            , # G+17663  U+D522  uniD522,  HANGUL SYLLABLE-D522
        '54563' => '17664'            , # G+17664  U+D523  uniD523,  HANGUL SYLLABLE-D523
        '54564' => '17665'            , # G+17665  U+D524  uniD524,  HANGUL SYLLABLE-D524
        '54565' => '17666'            , # G+17666  U+D525  uniD525,  HANGUL SYLLABLE-D525
        '54566' => '17667'            , # G+17667  U+D526  uniD526,  HANGUL SYLLABLE-D526
        '54567' => '17668'            , # G+17668  U+D527  uniD527,  HANGUL SYLLABLE-D527
        '54568' => '17669'            , # G+17669  U+D528  uniD528,  HANGUL SYLLABLE-D528
        '54569' => '17670'            , # G+17670  U+D529  uniD529,  HANGUL SYLLABLE-D529
        '54570' => '17671'            , # G+17671  U+D52A  uniD52A,  HANGUL SYLLABLE-D52A
        '54571' => '17672'            , # G+17672  U+D52B  uniD52B,  HANGUL SYLLABLE-D52B
        '54572' => '17673'            , # G+17673  U+D52C  uniD52C,  HANGUL SYLLABLE-D52C
        '54573' => '17674'            , # G+17674  U+D52D  uniD52D,  HANGUL SYLLABLE-D52D
        '54574' => '17675'            , # G+17675  U+D52E  uniD52E,  HANGUL SYLLABLE-D52E
        '54575' => '17676'            , # G+17676  U+D52F  uniD52F,  HANGUL SYLLABLE-D52F
        '54576' => '17677'            , # G+17677  U+D530  uniD530,  HANGUL SYLLABLE-D530
        '54577' => '17678'            , # G+17678  U+D531  uniD531,  HANGUL SYLLABLE-D531
        '54578' => '17679'            , # G+17679  U+D532  uniD532,  HANGUL SYLLABLE-D532
        '54579' => '17680'            , # G+17680  U+D533  uniD533,  HANGUL SYLLABLE-D533
        '54580' => '17681'            , # G+17681  U+D534  uniD534,  HANGUL SYLLABLE-D534
        '54581' => '17682'            , # G+17682  U+D535  uniD535,  HANGUL SYLLABLE-D535
        '54582' => '17683'            , # G+17683  U+D536  uniD536,  HANGUL SYLLABLE-D536
        '54583' => '17684'            , # G+17684  U+D537  uniD537,  HANGUL SYLLABLE-D537
        '54584' => '17685'            , # G+17685  U+D538  uniD538,  HANGUL SYLLABLE-D538
        '54585' => '17686'            , # G+17686  U+D539  uniD539,  HANGUL SYLLABLE-D539
        '54586' => '17687'            , # G+17687  U+D53A  uniD53A,  HANGUL SYLLABLE-D53A
        '54587' => '17688'            , # G+17688  U+D53B  uniD53B,  HANGUL SYLLABLE-D53B
        '54588' => '3286'             , # G+3286  U+D53C  uniD53C,  HANGUL SYLLABLE-D53C
        '54589' => '3287'             , # G+3287  U+D53D  uniD53D,  HANGUL SYLLABLE-D53D
        '54590' => '17689'            , # G+17689  U+D53E  uniD53E,  HANGUL SYLLABLE-D53E
        '54591' => '17690'            , # G+17690  U+D53F  uniD53F,  HANGUL SYLLABLE-D53F
        '54592' => '3288'             , # G+3288  U+D540  uniD540,  HANGUL SYLLABLE-D540
        '54593' => '17691'            , # G+17691  U+D541  uniD541,  HANGUL SYLLABLE-D541
        '54594' => '17692'            , # G+17692  U+D542  uniD542,  HANGUL SYLLABLE-D542
        '54595' => '17693'            , # G+17693  U+D543  uniD543,  HANGUL SYLLABLE-D543
        '54596' => '3289'             , # G+3289  U+D544  uniD544,  HANGUL SYLLABLE-D544
        '54597' => '17694'            , # G+17694  U+D545  uniD545,  HANGUL SYLLABLE-D545
        '54598' => '17695'            , # G+17695  U+D546  uniD546,  HANGUL SYLLABLE-D546
        '54599' => '17696'            , # G+17696  U+D547  uniD547,  HANGUL SYLLABLE-D547
        '54600' => '17697'            , # G+17697  U+D548  uniD548,  HANGUL SYLLABLE-D548
        '54601' => '17698'            , # G+17698  U+D549  uniD549,  HANGUL SYLLABLE-D549
        '54602' => '17699'            , # G+17699  U+D54A  uniD54A,  HANGUL SYLLABLE-D54A
        '54603' => '17700'            , # G+17700  U+D54B  uniD54B,  HANGUL SYLLABLE-D54B
        '54604' => '3290'             , # G+3290  U+D54C  uniD54C,  HANGUL SYLLABLE-D54C
        '54605' => '3291'             , # G+3291  U+D54D  uniD54D,  HANGUL SYLLABLE-D54D
        '54606' => '17701'            , # G+17701  U+D54E  uniD54E,  HANGUL SYLLABLE-D54E
        '54607' => '3292'             , # G+3292  U+D54F  uniD54F,  HANGUL SYLLABLE-D54F
        '54608' => '17702'            , # G+17702  U+D550  uniD550,  HANGUL SYLLABLE-D550
        '54609' => '3293'             , # G+3293  U+D551  uniD551,  HANGUL SYLLABLE-D551
        '54610' => '17703'            , # G+17703  U+D552  uniD552,  HANGUL SYLLABLE-D552
        '54611' => '17704'            , # G+17704  U+D553  uniD553,  HANGUL SYLLABLE-D553
        '54612' => '17705'            , # G+17705  U+D554  uniD554,  HANGUL SYLLABLE-D554
        '54613' => '17706'            , # G+17706  U+D555  uniD555,  HANGUL SYLLABLE-D555
        '54614' => '17707'            , # G+17707  U+D556  uniD556,  HANGUL SYLLABLE-D556
        '54615' => '17708'            , # G+17708  U+D557  uniD557,  HANGUL SYLLABLE-D557
        '54616' => '3294'             , # G+3294  U+D558  uniD558,  HANGUL SYLLABLE-D558
        '54617' => '3295'             , # G+3295  U+D559  uniD559,  HANGUL SYLLABLE-D559
        '54618' => '17709'            , # G+17709  U+D55A  uniD55A,  HANGUL SYLLABLE-D55A
        '54619' => '17710'            , # G+17710  U+D55B  uniD55B,  HANGUL SYLLABLE-D55B
        '54620' => '3296'             , # G+3296  U+D55C  uniD55C,  HANGUL SYLLABLE-D55C
        '54621' => '17711'            , # G+17711  U+D55D  uniD55D,  HANGUL SYLLABLE-D55D
        '54622' => '17712'            , # G+17712  U+D55E  uniD55E,  HANGUL SYLLABLE-D55E
        '54623' => '17713'            , # G+17713  U+D55F  uniD55F,  HANGUL SYLLABLE-D55F
        '54624' => '3297'             , # G+3297  U+D560  uniD560,  HANGUL SYLLABLE-D560
        '54625' => '17714'            , # G+17714  U+D561  uniD561,  HANGUL SYLLABLE-D561
        '54626' => '17715'            , # G+17715  U+D562  uniD562,  HANGUL SYLLABLE-D562
        '54627' => '17716'            , # G+17716  U+D563  uniD563,  HANGUL SYLLABLE-D563
        '54628' => '17717'            , # G+17717  U+D564  uniD564,  HANGUL SYLLABLE-D564
        '54629' => '3298'             , # G+3298  U+D565  uniD565,  HANGUL SYLLABLE-D565
        '54630' => '17718'            , # G+17718  U+D566  uniD566,  HANGUL SYLLABLE-D566
        '54631' => '17719'            , # G+17719  U+D567  uniD567,  HANGUL SYLLABLE-D567
        '54632' => '3299'             , # G+3299  U+D568  uniD568,  HANGUL SYLLABLE-D568
        '54633' => '3300'             , # G+3300  U+D569  uniD569,  HANGUL SYLLABLE-D569
        '54634' => '17720'            , # G+17720  U+D56A  uniD56A,  HANGUL SYLLABLE-D56A
        '54635' => '3301'             , # G+3301  U+D56B  uniD56B,  HANGUL SYLLABLE-D56B
        '54636' => '17721'            , # G+17721  U+D56C  uniD56C,  HANGUL SYLLABLE-D56C
        '54637' => '3302'             , # G+3302  U+D56D  uniD56D,  HANGUL SYLLABLE-D56D
        '54638' => '17722'            , # G+17722  U+D56E  uniD56E,  HANGUL SYLLABLE-D56E
        '54639' => '17723'            , # G+17723  U+D56F  uniD56F,  HANGUL SYLLABLE-D56F
        '54640' => '17724'            , # G+17724  U+D570  uniD570,  HANGUL SYLLABLE-D570
        '54641' => '17725'            , # G+17725  U+D571  uniD571,  HANGUL SYLLABLE-D571
        '54642' => '17726'            , # G+17726  U+D572  uniD572,  HANGUL SYLLABLE-D572
        '54643' => '17727'            , # G+17727  U+D573  uniD573,  HANGUL SYLLABLE-D573
        '54644' => '3303'             , # G+3303  U+D574  uniD574,  HANGUL SYLLABLE-D574
        '54645' => '3304'             , # G+3304  U+D575  uniD575,  HANGUL SYLLABLE-D575
        '54646' => '17728'            , # G+17728  U+D576  uniD576,  HANGUL SYLLABLE-D576
        '54647' => '17729'            , # G+17729  U+D577  uniD577,  HANGUL SYLLABLE-D577
        '54648' => '3305'             , # G+3305  U+D578  uniD578,  HANGUL SYLLABLE-D578
        '54649' => '17730'            , # G+17730  U+D579  uniD579,  HANGUL SYLLABLE-D579
        '54650' => '17731'            , # G+17731  U+D57A  uniD57A,  HANGUL SYLLABLE-D57A
        '54651' => '17732'            , # G+17732  U+D57B  uniD57B,  HANGUL SYLLABLE-D57B
        '54652' => '3306'             , # G+3306  U+D57C  uniD57C,  HANGUL SYLLABLE-D57C
        '54653' => '17733'            , # G+17733  U+D57D  uniD57D,  HANGUL SYLLABLE-D57D
        '54654' => '17734'            , # G+17734  U+D57E  uniD57E,  HANGUL SYLLABLE-D57E
        '54655' => '17735'            , # G+17735  U+D57F  uniD57F,  HANGUL SYLLABLE-D57F
        '54656' => '17736'            , # G+17736  U+D580  uniD580,  HANGUL SYLLABLE-D580
        '54657' => '17737'            , # G+17737  U+D581  uniD581,  HANGUL SYLLABLE-D581
        '54658' => '17738'            , # G+17738  U+D582  uniD582,  HANGUL SYLLABLE-D582
        '54659' => '17739'            , # G+17739  U+D583  uniD583,  HANGUL SYLLABLE-D583
        '54660' => '3307'             , # G+3307  U+D584  uniD584,  HANGUL SYLLABLE-D584
        '54661' => '3308'             , # G+3308  U+D585  uniD585,  HANGUL SYLLABLE-D585
        '54662' => '17740'            , # G+17740  U+D586  uniD586,  HANGUL SYLLABLE-D586
        '54663' => '3309'             , # G+3309  U+D587  uniD587,  HANGUL SYLLABLE-D587
        '54664' => '3310'             , # G+3310  U+D588  uniD588,  HANGUL SYLLABLE-D588
        '54665' => '3311'             , # G+3311  U+D589  uniD589,  HANGUL SYLLABLE-D589
        '54666' => '17741'            , # G+17741  U+D58A  uniD58A,  HANGUL SYLLABLE-D58A
        '54667' => '17742'            , # G+17742  U+D58B  uniD58B,  HANGUL SYLLABLE-D58B
        '54668' => '17743'            , # G+17743  U+D58C  uniD58C,  HANGUL SYLLABLE-D58C
        '54669' => '17744'            , # G+17744  U+D58D  uniD58D,  HANGUL SYLLABLE-D58D
        '54670' => '17745'            , # G+17745  U+D58E  uniD58E,  HANGUL SYLLABLE-D58E
        '54671' => '17746'            , # G+17746  U+D58F  uniD58F,  HANGUL SYLLABLE-D58F
        '54672' => '3312'             , # G+3312  U+D590  uniD590,  HANGUL SYLLABLE-D590
        '54673' => '17747'            , # G+17747  U+D591  uniD591,  HANGUL SYLLABLE-D591
        '54674' => '17748'            , # G+17748  U+D592  uniD592,  HANGUL SYLLABLE-D592
        '54675' => '17749'            , # G+17749  U+D593  uniD593,  HANGUL SYLLABLE-D593
        '54676' => '17750'            , # G+17750  U+D594  uniD594,  HANGUL SYLLABLE-D594
        '54677' => '17751'            , # G+17751  U+D595  uniD595,  HANGUL SYLLABLE-D595
        '54678' => '17752'            , # G+17752  U+D596  uniD596,  HANGUL SYLLABLE-D596
        '54679' => '17753'            , # G+17753  U+D597  uniD597,  HANGUL SYLLABLE-D597
        '54680' => '17754'            , # G+17754  U+D598  uniD598,  HANGUL SYLLABLE-D598
        '54681' => '17755'            , # G+17755  U+D599  uniD599,  HANGUL SYLLABLE-D599
        '54682' => '17756'            , # G+17756  U+D59A  uniD59A,  HANGUL SYLLABLE-D59A
        '54683' => '17757'            , # G+17757  U+D59B  uniD59B,  HANGUL SYLLABLE-D59B
        '54684' => '17758'            , # G+17758  U+D59C  uniD59C,  HANGUL SYLLABLE-D59C
        '54685' => '17759'            , # G+17759  U+D59D  uniD59D,  HANGUL SYLLABLE-D59D
        '54686' => '17760'            , # G+17760  U+D59E  uniD59E,  HANGUL SYLLABLE-D59E
        '54687' => '17761'            , # G+17761  U+D59F  uniD59F,  HANGUL SYLLABLE-D59F
        '54688' => '17762'            , # G+17762  U+D5A0  uniD5A0,  HANGUL SYLLABLE-D5A0
        '54689' => '17763'            , # G+17763  U+D5A1  uniD5A1,  HANGUL SYLLABLE-D5A1
        '54690' => '17764'            , # G+17764  U+D5A2  uniD5A2,  HANGUL SYLLABLE-D5A2
        '54691' => '17765'            , # G+17765  U+D5A3  uniD5A3,  HANGUL SYLLABLE-D5A3
        '54692' => '17766'            , # G+17766  U+D5A4  uniD5A4,  HANGUL SYLLABLE-D5A4
        '54693' => '3313'             , # G+3313  U+D5A5  uniD5A5,  HANGUL SYLLABLE-D5A5
        '54694' => '17767'            , # G+17767  U+D5A6  uniD5A6,  HANGUL SYLLABLE-D5A6
        '54695' => '17768'            , # G+17768  U+D5A7  uniD5A7,  HANGUL SYLLABLE-D5A7
        '54696' => '17769'            , # G+17769  U+D5A8  uniD5A8,  HANGUL SYLLABLE-D5A8
        '54697' => '17770'            , # G+17770  U+D5A9  uniD5A9,  HANGUL SYLLABLE-D5A9
        '54698' => '17771'            , # G+17771  U+D5AA  uniD5AA,  HANGUL SYLLABLE-D5AA
        '54699' => '17772'            , # G+17772  U+D5AB  uniD5AB,  HANGUL SYLLABLE-D5AB
        '54700' => '17773'            , # G+17773  U+D5AC  uniD5AC,  HANGUL SYLLABLE-D5AC
        '54701' => '17774'            , # G+17774  U+D5AD  uniD5AD,  HANGUL SYLLABLE-D5AD
        '54702' => '17775'            , # G+17775  U+D5AE  uniD5AE,  HANGUL SYLLABLE-D5AE
        '54703' => '17776'            , # G+17776  U+D5AF  uniD5AF,  HANGUL SYLLABLE-D5AF
        '54704' => '17777'            , # G+17777  U+D5B0  uniD5B0,  HANGUL SYLLABLE-D5B0
        '54705' => '17778'            , # G+17778  U+D5B1  uniD5B1,  HANGUL SYLLABLE-D5B1
        '54706' => '17779'            , # G+17779  U+D5B2  uniD5B2,  HANGUL SYLLABLE-D5B2
        '54707' => '17780'            , # G+17780  U+D5B3  uniD5B3,  HANGUL SYLLABLE-D5B3
        '54708' => '17781'            , # G+17781  U+D5B4  uniD5B4,  HANGUL SYLLABLE-D5B4
        '54709' => '17782'            , # G+17782  U+D5B5  uniD5B5,  HANGUL SYLLABLE-D5B5
        '54710' => '17783'            , # G+17783  U+D5B6  uniD5B6,  HANGUL SYLLABLE-D5B6
        '54711' => '17784'            , # G+17784  U+D5B7  uniD5B7,  HANGUL SYLLABLE-D5B7
        '54712' => '17785'            , # G+17785  U+D5B8  uniD5B8,  HANGUL SYLLABLE-D5B8
        '54713' => '17786'            , # G+17786  U+D5B9  uniD5B9,  HANGUL SYLLABLE-D5B9
        '54714' => '17787'            , # G+17787  U+D5BA  uniD5BA,  HANGUL SYLLABLE-D5BA
        '54715' => '17788'            , # G+17788  U+D5BB  uniD5BB,  HANGUL SYLLABLE-D5BB
        '54716' => '17789'            , # G+17789  U+D5BC  uniD5BC,  HANGUL SYLLABLE-D5BC
        '54717' => '17790'            , # G+17790  U+D5BD  uniD5BD,  HANGUL SYLLABLE-D5BD
        '54718' => '17791'            , # G+17791  U+D5BE  uniD5BE,  HANGUL SYLLABLE-D5BE
        '54719' => '17792'            , # G+17792  U+D5BF  uniD5BF,  HANGUL SYLLABLE-D5BF
        '54720' => '17793'            , # G+17793  U+D5C0  uniD5C0,  HANGUL SYLLABLE-D5C0
        '54721' => '17794'            , # G+17794  U+D5C1  uniD5C1,  HANGUL SYLLABLE-D5C1
        '54722' => '17795'            , # G+17795  U+D5C2  uniD5C2,  HANGUL SYLLABLE-D5C2
        '54723' => '17796'            , # G+17796  U+D5C3  uniD5C3,  HANGUL SYLLABLE-D5C3
        '54724' => '17797'            , # G+17797  U+D5C4  uniD5C4,  HANGUL SYLLABLE-D5C4
        '54725' => '17798'            , # G+17798  U+D5C5  uniD5C5,  HANGUL SYLLABLE-D5C5
        '54726' => '17799'            , # G+17799  U+D5C6  uniD5C6,  HANGUL SYLLABLE-D5C6
        '54727' => '17800'            , # G+17800  U+D5C7  uniD5C7,  HANGUL SYLLABLE-D5C7
        '54728' => '3314'             , # G+3314  U+D5C8  uniD5C8,  HANGUL SYLLABLE-D5C8
        '54729' => '3315'             , # G+3315  U+D5C9  uniD5C9,  HANGUL SYLLABLE-D5C9
        '54730' => '17801'            , # G+17801  U+D5CA  uniD5CA,  HANGUL SYLLABLE-D5CA
        '54731' => '17802'            , # G+17802  U+D5CB  uniD5CB,  HANGUL SYLLABLE-D5CB
        '54732' => '3316'             , # G+3316  U+D5CC  uniD5CC,  HANGUL SYLLABLE-D5CC
        '54733' => '17803'            , # G+17803  U+D5CD  uniD5CD,  HANGUL SYLLABLE-D5CD
        '54734' => '17804'            , # G+17804  U+D5CE  uniD5CE,  HANGUL SYLLABLE-D5CE
        '54735' => '17805'            , # G+17805  U+D5CF  uniD5CF,  HANGUL SYLLABLE-D5CF
        '54736' => '3317'             , # G+3317  U+D5D0  uniD5D0,  HANGUL SYLLABLE-D5D0
        '54737' => '17806'            , # G+17806  U+D5D1  uniD5D1,  HANGUL SYLLABLE-D5D1
        '54738' => '3318'             , # G+3318  U+D5D2  uniD5D2,  HANGUL SYLLABLE-D5D2
        '54739' => '17807'            , # G+17807  U+D5D3  uniD5D3,  HANGUL SYLLABLE-D5D3
        '54740' => '17808'            , # G+17808  U+D5D4  uniD5D4,  HANGUL SYLLABLE-D5D4
        '54741' => '17809'            , # G+17809  U+D5D5  uniD5D5,  HANGUL SYLLABLE-D5D5
        '54742' => '17810'            , # G+17810  U+D5D6  uniD5D6,  HANGUL SYLLABLE-D5D6
        '54743' => '17811'            , # G+17811  U+D5D7  uniD5D7,  HANGUL SYLLABLE-D5D7
        '54744' => '3319'             , # G+3319  U+D5D8  uniD5D8,  HANGUL SYLLABLE-D5D8
        '54745' => '3320'             , # G+3320  U+D5D9  uniD5D9,  HANGUL SYLLABLE-D5D9
        '54746' => '17812'            , # G+17812  U+D5DA  uniD5DA,  HANGUL SYLLABLE-D5DA
        '54747' => '3321'             , # G+3321  U+D5DB  uniD5DB,  HANGUL SYLLABLE-D5DB
        '54748' => '17813'            , # G+17813  U+D5DC  uniD5DC,  HANGUL SYLLABLE-D5DC
        '54749' => '3322'             , # G+3322  U+D5DD  uniD5DD,  HANGUL SYLLABLE-D5DD
        '54750' => '17814'            , # G+17814  U+D5DE  uniD5DE,  HANGUL SYLLABLE-D5DE
        '54751' => '17815'            , # G+17815  U+D5DF  uniD5DF,  HANGUL SYLLABLE-D5DF
        '54752' => '17816'            , # G+17816  U+D5E0  uniD5E0,  HANGUL SYLLABLE-D5E0
        '54753' => '17817'            , # G+17817  U+D5E1  uniD5E1,  HANGUL SYLLABLE-D5E1
        '54754' => '17818'            , # G+17818  U+D5E2  uniD5E2,  HANGUL SYLLABLE-D5E2
        '54755' => '17819'            , # G+17819  U+D5E3  uniD5E3,  HANGUL SYLLABLE-D5E3
        '54756' => '3323'             , # G+3323  U+D5E4  uniD5E4,  HANGUL SYLLABLE-D5E4
        '54757' => '3324'             , # G+3324  U+D5E5  uniD5E5,  HANGUL SYLLABLE-D5E5
        '54758' => '17820'            , # G+17820  U+D5E6  uniD5E6,  HANGUL SYLLABLE-D5E6
        '54759' => '17821'            , # G+17821  U+D5E7  uniD5E7,  HANGUL SYLLABLE-D5E7
        '54760' => '3325'             , # G+3325  U+D5E8  uniD5E8,  HANGUL SYLLABLE-D5E8
        '54761' => '17822'            , # G+17822  U+D5E9  uniD5E9,  HANGUL SYLLABLE-D5E9
        '54762' => '17823'            , # G+17823  U+D5EA  uniD5EA,  HANGUL SYLLABLE-D5EA
        '54763' => '17824'            , # G+17824  U+D5EB  uniD5EB,  HANGUL SYLLABLE-D5EB
        '54764' => '3326'             , # G+3326  U+D5EC  uniD5EC,  HANGUL SYLLABLE-D5EC
        '54765' => '17825'            , # G+17825  U+D5ED  uniD5ED,  HANGUL SYLLABLE-D5ED
        '54766' => '17826'            , # G+17826  U+D5EE  uniD5EE,  HANGUL SYLLABLE-D5EE
        '54767' => '17827'            , # G+17827  U+D5EF  uniD5EF,  HANGUL SYLLABLE-D5EF
        '54768' => '17828'            , # G+17828  U+D5F0  uniD5F0,  HANGUL SYLLABLE-D5F0
        '54769' => '17829'            , # G+17829  U+D5F1  uniD5F1,  HANGUL SYLLABLE-D5F1
        '54770' => '17830'            , # G+17830  U+D5F2  uniD5F2,  HANGUL SYLLABLE-D5F2
        '54771' => '17831'            , # G+17831  U+D5F3  uniD5F3,  HANGUL SYLLABLE-D5F3
        '54772' => '3327'             , # G+3327  U+D5F4  uniD5F4,  HANGUL SYLLABLE-D5F4
        '54773' => '3328'             , # G+3328  U+D5F5  uniD5F5,  HANGUL SYLLABLE-D5F5
        '54774' => '17832'            , # G+17832  U+D5F6  uniD5F6,  HANGUL SYLLABLE-D5F6
        '54775' => '3329'             , # G+3329  U+D5F7  uniD5F7,  HANGUL SYLLABLE-D5F7
        '54776' => '17833'            , # G+17833  U+D5F8  uniD5F8,  HANGUL SYLLABLE-D5F8
        '54777' => '3330'             , # G+3330  U+D5F9  uniD5F9,  HANGUL SYLLABLE-D5F9
        '54778' => '17834'            , # G+17834  U+D5FA  uniD5FA,  HANGUL SYLLABLE-D5FA
        '54779' => '17835'            , # G+17835  U+D5FB  uniD5FB,  HANGUL SYLLABLE-D5FB
        '54780' => '17836'            , # G+17836  U+D5FC  uniD5FC,  HANGUL SYLLABLE-D5FC
        '54781' => '17837'            , # G+17837  U+D5FD  uniD5FD,  HANGUL SYLLABLE-D5FD
        '54782' => '17838'            , # G+17838  U+D5FE  uniD5FE,  HANGUL SYLLABLE-D5FE
        '54783' => '17839'            , # G+17839  U+D5FF  uniD5FF,  HANGUL SYLLABLE-D5FF
        '54784' => '3331'             , # G+3331  U+D600  uniD600,  HANGUL SYLLABLE-D600
        '54785' => '3332'             , # G+3332  U+D601  uniD601,  HANGUL SYLLABLE-D601
        '54786' => '17840'            , # G+17840  U+D602  uniD602,  HANGUL SYLLABLE-D602
        '54787' => '17841'            , # G+17841  U+D603  uniD603,  HANGUL SYLLABLE-D603
        '54788' => '3333'             , # G+3333  U+D604  uniD604,  HANGUL SYLLABLE-D604
        '54789' => '17842'            , # G+17842  U+D605  uniD605,  HANGUL SYLLABLE-D605
        '54790' => '17843'            , # G+17843  U+D606  uniD606,  HANGUL SYLLABLE-D606
        '54791' => '17844'            , # G+17844  U+D607  uniD607,  HANGUL SYLLABLE-D607
        '54792' => '3334'             , # G+3334  U+D608  uniD608,  HANGUL SYLLABLE-D608
        '54793' => '17845'            , # G+17845  U+D609  uniD609,  HANGUL SYLLABLE-D609
        '54794' => '17846'            , # G+17846  U+D60A  uniD60A,  HANGUL SYLLABLE-D60A
        '54795' => '17847'            , # G+17847  U+D60B  uniD60B,  HANGUL SYLLABLE-D60B
        '54796' => '17848'            , # G+17848  U+D60C  uniD60C,  HANGUL SYLLABLE-D60C
        '54797' => '17849'            , # G+17849  U+D60D  uniD60D,  HANGUL SYLLABLE-D60D
        '54798' => '17850'            , # G+17850  U+D60E  uniD60E,  HANGUL SYLLABLE-D60E
        '54799' => '17851'            , # G+17851  U+D60F  uniD60F,  HANGUL SYLLABLE-D60F
        '54800' => '3335'             , # G+3335  U+D610  uniD610,  HANGUL SYLLABLE-D610
        '54801' => '3336'             , # G+3336  U+D611  uniD611,  HANGUL SYLLABLE-D611
        '54802' => '17852'            , # G+17852  U+D612  uniD612,  HANGUL SYLLABLE-D612
        '54803' => '3337'             , # G+3337  U+D613  uniD613,  HANGUL SYLLABLE-D613
        '54804' => '3338'             , # G+3338  U+D614  uniD614,  HANGUL SYLLABLE-D614
        '54805' => '3339'             , # G+3339  U+D615  uniD615,  HANGUL SYLLABLE-D615
        '54806' => '17853'            , # G+17853  U+D616  uniD616,  HANGUL SYLLABLE-D616
        '54807' => '17854'            , # G+17854  U+D617  uniD617,  HANGUL SYLLABLE-D617
        '54808' => '17855'            , # G+17855  U+D618  uniD618,  HANGUL SYLLABLE-D618
        '54809' => '17856'            , # G+17856  U+D619  uniD619,  HANGUL SYLLABLE-D619
        '54810' => '17857'            , # G+17857  U+D61A  uniD61A,  HANGUL SYLLABLE-D61A
        '54811' => '17858'            , # G+17858  U+D61B  uniD61B,  HANGUL SYLLABLE-D61B
        '54812' => '3340'             , # G+3340  U+D61C  uniD61C,  HANGUL SYLLABLE-D61C
        '54813' => '17859'            , # G+17859  U+D61D  uniD61D,  HANGUL SYLLABLE-D61D
        '54814' => '17860'            , # G+17860  U+D61E  uniD61E,  HANGUL SYLLABLE-D61E
        '54815' => '17861'            , # G+17861  U+D61F  uniD61F,  HANGUL SYLLABLE-D61F
        '54816' => '3341'             , # G+3341  U+D620  uniD620,  HANGUL SYLLABLE-D620
        '54817' => '17862'            , # G+17862  U+D621  uniD621,  HANGUL SYLLABLE-D621
        '54818' => '17863'            , # G+17863  U+D622  uniD622,  HANGUL SYLLABLE-D622
        '54819' => '17864'            , # G+17864  U+D623  uniD623,  HANGUL SYLLABLE-D623
        '54820' => '3342'             , # G+3342  U+D624  uniD624,  HANGUL SYLLABLE-D624
        '54821' => '17865'            , # G+17865  U+D625  uniD625,  HANGUL SYLLABLE-D625
        '54822' => '17866'            , # G+17866  U+D626  uniD626,  HANGUL SYLLABLE-D626
        '54823' => '17867'            , # G+17867  U+D627  uniD627,  HANGUL SYLLABLE-D627
        '54824' => '17868'            , # G+17868  U+D628  uniD628,  HANGUL SYLLABLE-D628
        '54825' => '17869'            , # G+17869  U+D629  uniD629,  HANGUL SYLLABLE-D629
        '54826' => '17870'            , # G+17870  U+D62A  uniD62A,  HANGUL SYLLABLE-D62A
        '54827' => '17871'            , # G+17871  U+D62B  uniD62B,  HANGUL SYLLABLE-D62B
        '54828' => '17872'            , # G+17872  U+D62C  uniD62C,  HANGUL SYLLABLE-D62C
        '54829' => '3343'             , # G+3343  U+D62D  uniD62D,  HANGUL SYLLABLE-D62D
        '54830' => '17873'            , # G+17873  U+D62E  uniD62E,  HANGUL SYLLABLE-D62E
        '54831' => '17874'            , # G+17874  U+D62F  uniD62F,  HANGUL SYLLABLE-D62F
        '54832' => '17875'            , # G+17875  U+D630  uniD630,  HANGUL SYLLABLE-D630
        '54833' => '17876'            , # G+17876  U+D631  uniD631,  HANGUL SYLLABLE-D631
        '54834' => '17877'            , # G+17877  U+D632  uniD632,  HANGUL SYLLABLE-D632
        '54835' => '17878'            , # G+17878  U+D633  uniD633,  HANGUL SYLLABLE-D633
        '54836' => '17879'            , # G+17879  U+D634  uniD634,  HANGUL SYLLABLE-D634
        '54837' => '17880'            , # G+17880  U+D635  uniD635,  HANGUL SYLLABLE-D635
        '54838' => '17881'            , # G+17881  U+D636  uniD636,  HANGUL SYLLABLE-D636
        '54839' => '17882'            , # G+17882  U+D637  uniD637,  HANGUL SYLLABLE-D637
        '54840' => '3344'             , # G+3344  U+D638  uniD638,  HANGUL SYLLABLE-D638
        '54841' => '3345'             , # G+3345  U+D639  uniD639,  HANGUL SYLLABLE-D639
        '54842' => '17883'            , # G+17883  U+D63A  uniD63A,  HANGUL SYLLABLE-D63A
        '54843' => '17884'            , # G+17884  U+D63B  uniD63B,  HANGUL SYLLABLE-D63B
        '54844' => '3346'             , # G+3346  U+D63C  uniD63C,  HANGUL SYLLABLE-D63C
        '54845' => '17885'            , # G+17885  U+D63D  uniD63D,  HANGUL SYLLABLE-D63D
        '54846' => '17886'            , # G+17886  U+D63E  uniD63E,  HANGUL SYLLABLE-D63E
        '54847' => '17887'            , # G+17887  U+D63F  uniD63F,  HANGUL SYLLABLE-D63F
        '54848' => '3347'             , # G+3347  U+D640  uniD640,  HANGUL SYLLABLE-D640
        '54849' => '17888'            , # G+17888  U+D641  uniD641,  HANGUL SYLLABLE-D641
        '54850' => '17889'            , # G+17889  U+D642  uniD642,  HANGUL SYLLABLE-D642
        '54851' => '17890'            , # G+17890  U+D643  uniD643,  HANGUL SYLLABLE-D643
        '54852' => '17891'            , # G+17891  U+D644  uniD644,  HANGUL SYLLABLE-D644
        '54853' => '3348'             , # G+3348  U+D645  uniD645,  HANGUL SYLLABLE-D645
        '54854' => '17892'            , # G+17892  U+D646  uniD646,  HANGUL SYLLABLE-D646
        '54855' => '17893'            , # G+17893  U+D647  uniD647,  HANGUL SYLLABLE-D647
        '54856' => '3349'             , # G+3349  U+D648  uniD648,  HANGUL SYLLABLE-D648
        '54857' => '3350'             , # G+3350  U+D649  uniD649,  HANGUL SYLLABLE-D649
        '54858' => '17894'            , # G+17894  U+D64A  uniD64A,  HANGUL SYLLABLE-D64A
        '54859' => '3351'             , # G+3351  U+D64B  uniD64B,  HANGUL SYLLABLE-D64B
        '54860' => '17895'            , # G+17895  U+D64C  uniD64C,  HANGUL SYLLABLE-D64C
        '54861' => '3352'             , # G+3352  U+D64D  uniD64D,  HANGUL SYLLABLE-D64D
        '54862' => '17896'            , # G+17896  U+D64E  uniD64E,  HANGUL SYLLABLE-D64E
        '54863' => '17897'            , # G+17897  U+D64F  uniD64F,  HANGUL SYLLABLE-D64F
        '54864' => '17898'            , # G+17898  U+D650  uniD650,  HANGUL SYLLABLE-D650
        '54865' => '3353'             , # G+3353  U+D651  uniD651,  HANGUL SYLLABLE-D651
        '54866' => '17899'            , # G+17899  U+D652  uniD652,  HANGUL SYLLABLE-D652
        '54867' => '17900'            , # G+17900  U+D653  uniD653,  HANGUL SYLLABLE-D653
        '54868' => '3354'             , # G+3354  U+D654  uniD654,  HANGUL SYLLABLE-D654
        '54869' => '3355'             , # G+3355  U+D655  uniD655,  HANGUL SYLLABLE-D655
        '54870' => '17901'            , # G+17901  U+D656  uniD656,  HANGUL SYLLABLE-D656
        '54871' => '17902'            , # G+17902  U+D657  uniD657,  HANGUL SYLLABLE-D657
        '54872' => '3356'             , # G+3356  U+D658  uniD658,  HANGUL SYLLABLE-D658
        '54873' => '17903'            , # G+17903  U+D659  uniD659,  HANGUL SYLLABLE-D659
        '54874' => '17904'            , # G+17904  U+D65A  uniD65A,  HANGUL SYLLABLE-D65A
        '54875' => '17905'            , # G+17905  U+D65B  uniD65B,  HANGUL SYLLABLE-D65B
        '54876' => '3357'             , # G+3357  U+D65C  uniD65C,  HANGUL SYLLABLE-D65C
        '54877' => '17906'            , # G+17906  U+D65D  uniD65D,  HANGUL SYLLABLE-D65D
        '54878' => '17907'            , # G+17907  U+D65E  uniD65E,  HANGUL SYLLABLE-D65E
        '54879' => '17908'            , # G+17908  U+D65F  uniD65F,  HANGUL SYLLABLE-D65F
        '54880' => '17909'            , # G+17909  U+D660  uniD660,  HANGUL SYLLABLE-D660
        '54881' => '17910'            , # G+17910  U+D661  uniD661,  HANGUL SYLLABLE-D661
        '54882' => '17911'            , # G+17911  U+D662  uniD662,  HANGUL SYLLABLE-D662
        '54883' => '17912'            , # G+17912  U+D663  uniD663,  HANGUL SYLLABLE-D663
        '54884' => '17913'            , # G+17913  U+D664  uniD664,  HANGUL SYLLABLE-D664
        '54885' => '17914'            , # G+17914  U+D665  uniD665,  HANGUL SYLLABLE-D665
        '54886' => '17915'            , # G+17915  U+D666  uniD666,  HANGUL SYLLABLE-D666
        '54887' => '3358'             , # G+3358  U+D667  uniD667,  HANGUL SYLLABLE-D667
        '54888' => '17916'            , # G+17916  U+D668  uniD668,  HANGUL SYLLABLE-D668
        '54889' => '3359'             , # G+3359  U+D669  uniD669,  HANGUL SYLLABLE-D669
        '54890' => '17917'            , # G+17917  U+D66A  uniD66A,  HANGUL SYLLABLE-D66A
        '54891' => '17918'            , # G+17918  U+D66B  uniD66B,  HANGUL SYLLABLE-D66B
        '54892' => '17919'            , # G+17919  U+D66C  uniD66C,  HANGUL SYLLABLE-D66C
        '54893' => '17920'            , # G+17920  U+D66D  uniD66D,  HANGUL SYLLABLE-D66D
        '54894' => '17921'            , # G+17921  U+D66E  uniD66E,  HANGUL SYLLABLE-D66E
        '54895' => '17922'            , # G+17922  U+D66F  uniD66F,  HANGUL SYLLABLE-D66F
        '54896' => '3360'             , # G+3360  U+D670  uniD670,  HANGUL SYLLABLE-D670
        '54897' => '3361'             , # G+3361  U+D671  uniD671,  HANGUL SYLLABLE-D671
        '54898' => '17923'            , # G+17923  U+D672  uniD672,  HANGUL SYLLABLE-D672
        '54899' => '17924'            , # G+17924  U+D673  uniD673,  HANGUL SYLLABLE-D673
        '54900' => '3362'             , # G+3362  U+D674  uniD674,  HANGUL SYLLABLE-D674
        '54901' => '17925'            , # G+17925  U+D675  uniD675,  HANGUL SYLLABLE-D675
        '54902' => '17926'            , # G+17926  U+D676  uniD676,  HANGUL SYLLABLE-D676
        '54903' => '17927'            , # G+17927  U+D677  uniD677,  HANGUL SYLLABLE-D677
        '54904' => '17928'            , # G+17928  U+D678  uniD678,  HANGUL SYLLABLE-D678
        '54905' => '17929'            , # G+17929  U+D679  uniD679,  HANGUL SYLLABLE-D679
        '54906' => '17930'            , # G+17930  U+D67A  uniD67A,  HANGUL SYLLABLE-D67A
        '54907' => '17931'            , # G+17931  U+D67B  uniD67B,  HANGUL SYLLABLE-D67B
        '54908' => '17932'            , # G+17932  U+D67C  uniD67C,  HANGUL SYLLABLE-D67C
        '54909' => '17933'            , # G+17933  U+D67D  uniD67D,  HANGUL SYLLABLE-D67D
        '54910' => '17934'            , # G+17934  U+D67E  uniD67E,  HANGUL SYLLABLE-D67E
        '54911' => '17935'            , # G+17935  U+D67F  uniD67F,  HANGUL SYLLABLE-D67F
        '54912' => '17936'            , # G+17936  U+D680  uniD680,  HANGUL SYLLABLE-D680
        '54913' => '17937'            , # G+17937  U+D681  uniD681,  HANGUL SYLLABLE-D681
        '54914' => '17938'            , # G+17938  U+D682  uniD682,  HANGUL SYLLABLE-D682
        '54915' => '3363'             , # G+3363  U+D683  uniD683,  HANGUL SYLLABLE-D683
        '54916' => '17939'            , # G+17939  U+D684  uniD684,  HANGUL SYLLABLE-D684
        '54917' => '3364'             , # G+3364  U+D685  uniD685,  HANGUL SYLLABLE-D685
        '54918' => '17940'            , # G+17940  U+D686  uniD686,  HANGUL SYLLABLE-D686
        '54919' => '17941'            , # G+17941  U+D687  uniD687,  HANGUL SYLLABLE-D687
        '54920' => '17942'            , # G+17942  U+D688  uniD688,  HANGUL SYLLABLE-D688
        '54921' => '17943'            , # G+17943  U+D689  uniD689,  HANGUL SYLLABLE-D689
        '54922' => '17944'            , # G+17944  U+D68A  uniD68A,  HANGUL SYLLABLE-D68A
        '54923' => '17945'            , # G+17945  U+D68B  uniD68B,  HANGUL SYLLABLE-D68B
        '54924' => '3365'             , # G+3365  U+D68C  uniD68C,  HANGUL SYLLABLE-D68C
        '54925' => '3366'             , # G+3366  U+D68D  uniD68D,  HANGUL SYLLABLE-D68D
        '54926' => '17946'            , # G+17946  U+D68E  uniD68E,  HANGUL SYLLABLE-D68E
        '54927' => '17947'            , # G+17947  U+D68F  uniD68F,  HANGUL SYLLABLE-D68F
        '54928' => '3367'             , # G+3367  U+D690  uniD690,  HANGUL SYLLABLE-D690
        '54929' => '17948'            , # G+17948  U+D691  uniD691,  HANGUL SYLLABLE-D691
        '54930' => '17949'            , # G+17949  U+D692  uniD692,  HANGUL SYLLABLE-D692
        '54931' => '17950'            , # G+17950  U+D693  uniD693,  HANGUL SYLLABLE-D693
        '54932' => '3368'             , # G+3368  U+D694  uniD694,  HANGUL SYLLABLE-D694
        '54933' => '17951'            , # G+17951  U+D695  uniD695,  HANGUL SYLLABLE-D695
        '54934' => '17952'            , # G+17952  U+D696  uniD696,  HANGUL SYLLABLE-D696
        '54935' => '17953'            , # G+17953  U+D697  uniD697,  HANGUL SYLLABLE-D697
        '54936' => '17954'            , # G+17954  U+D698  uniD698,  HANGUL SYLLABLE-D698
        '54937' => '17955'            , # G+17955  U+D699  uniD699,  HANGUL SYLLABLE-D699
        '54938' => '17956'            , # G+17956  U+D69A  uniD69A,  HANGUL SYLLABLE-D69A
        '54939' => '17957'            , # G+17957  U+D69B  uniD69B,  HANGUL SYLLABLE-D69B
        '54940' => '17958'            , # G+17958  U+D69C  uniD69C,  HANGUL SYLLABLE-D69C
        '54941' => '3369'             , # G+3369  U+D69D  uniD69D,  HANGUL SYLLABLE-D69D
        '54942' => '17959'            , # G+17959  U+D69E  uniD69E,  HANGUL SYLLABLE-D69E
        '54943' => '3370'             , # G+3370  U+D69F  uniD69F,  HANGUL SYLLABLE-D69F
        '54944' => '17960'            , # G+17960  U+D6A0  uniD6A0,  HANGUL SYLLABLE-D6A0
        '54945' => '3371'             , # G+3371  U+D6A1  uniD6A1,  HANGUL SYLLABLE-D6A1
        '54946' => '17961'            , # G+17961  U+D6A2  uniD6A2,  HANGUL SYLLABLE-D6A2
        '54947' => '17962'            , # G+17962  U+D6A3  uniD6A3,  HANGUL SYLLABLE-D6A3
        '54948' => '17963'            , # G+17963  U+D6A4  uniD6A4,  HANGUL SYLLABLE-D6A4
        '54949' => '17964'            , # G+17964  U+D6A5  uniD6A5,  HANGUL SYLLABLE-D6A5
        '54950' => '17965'            , # G+17965  U+D6A6  uniD6A6,  HANGUL SYLLABLE-D6A6
        '54951' => '17966'            , # G+17966  U+D6A7  uniD6A7,  HANGUL SYLLABLE-D6A7
        '54952' => '3372'             , # G+3372  U+D6A8  uniD6A8,  HANGUL SYLLABLE-D6A8
        '54953' => '17967'            , # G+17967  U+D6A9  uniD6A9,  HANGUL SYLLABLE-D6A9
        '54954' => '17968'            , # G+17968  U+D6AA  uniD6AA,  HANGUL SYLLABLE-D6AA
        '54955' => '17969'            , # G+17969  U+D6AB  uniD6AB,  HANGUL SYLLABLE-D6AB
        '54956' => '3373'             , # G+3373  U+D6AC  uniD6AC,  HANGUL SYLLABLE-D6AC
        '54957' => '17970'            , # G+17970  U+D6AD  uniD6AD,  HANGUL SYLLABLE-D6AD
        '54958' => '17971'            , # G+17971  U+D6AE  uniD6AE,  HANGUL SYLLABLE-D6AE
        '54959' => '17972'            , # G+17972  U+D6AF  uniD6AF,  HANGUL SYLLABLE-D6AF
        '54960' => '3374'             , # G+3374  U+D6B0  uniD6B0,  HANGUL SYLLABLE-D6B0
        '54961' => '17973'            , # G+17973  U+D6B1  uniD6B1,  HANGUL SYLLABLE-D6B1
        '54962' => '17974'            , # G+17974  U+D6B2  uniD6B2,  HANGUL SYLLABLE-D6B2
        '54963' => '17975'            , # G+17975  U+D6B3  uniD6B3,  HANGUL SYLLABLE-D6B3
        '54964' => '17976'            , # G+17976  U+D6B4  uniD6B4,  HANGUL SYLLABLE-D6B4
        '54965' => '17977'            , # G+17977  U+D6B5  uniD6B5,  HANGUL SYLLABLE-D6B5
        '54966' => '17978'            , # G+17978  U+D6B6  uniD6B6,  HANGUL SYLLABLE-D6B6
        '54967' => '17979'            , # G+17979  U+D6B7  uniD6B7,  HANGUL SYLLABLE-D6B7
        '54968' => '17980'            , # G+17980  U+D6B8  uniD6B8,  HANGUL SYLLABLE-D6B8
        '54969' => '3375'             , # G+3375  U+D6B9  uniD6B9,  HANGUL SYLLABLE-D6B9
        '54970' => '17981'            , # G+17981  U+D6BA  uniD6BA,  HANGUL SYLLABLE-D6BA
        '54971' => '3376'             , # G+3376  U+D6BB  uniD6BB,  HANGUL SYLLABLE-D6BB
        '54972' => '17982'            , # G+17982  U+D6BC  uniD6BC,  HANGUL SYLLABLE-D6BC
        '54973' => '17983'            , # G+17983  U+D6BD  uniD6BD,  HANGUL SYLLABLE-D6BD
        '54974' => '17984'            , # G+17984  U+D6BE  uniD6BE,  HANGUL SYLLABLE-D6BE
        '54975' => '17985'            , # G+17985  U+D6BF  uniD6BF,  HANGUL SYLLABLE-D6BF
        '54976' => '17986'            , # G+17986  U+D6C0  uniD6C0,  HANGUL SYLLABLE-D6C0
        '54977' => '17987'            , # G+17987  U+D6C1  uniD6C1,  HANGUL SYLLABLE-D6C1
        '54978' => '17988'            , # G+17988  U+D6C2  uniD6C2,  HANGUL SYLLABLE-D6C2
        '54979' => '17989'            , # G+17989  U+D6C3  uniD6C3,  HANGUL SYLLABLE-D6C3
        '54980' => '3377'             , # G+3377  U+D6C4  uniD6C4,  HANGUL SYLLABLE-D6C4
        '54981' => '3378'             , # G+3378  U+D6C5  uniD6C5,  HANGUL SYLLABLE-D6C5
        '54982' => '17990'            , # G+17990  U+D6C6  uniD6C6,  HANGUL SYLLABLE-D6C6
        '54983' => '17991'            , # G+17991  U+D6C7  uniD6C7,  HANGUL SYLLABLE-D6C7
        '54984' => '3379'             , # G+3379  U+D6C8  uniD6C8,  HANGUL SYLLABLE-D6C8
        '54985' => '17992'            , # G+17992  U+D6C9  uniD6C9,  HANGUL SYLLABLE-D6C9
        '54986' => '17993'            , # G+17993  U+D6CA  uniD6CA,  HANGUL SYLLABLE-D6CA
        '54987' => '17994'            , # G+17994  U+D6CB  uniD6CB,  HANGUL SYLLABLE-D6CB
        '54988' => '3380'             , # G+3380  U+D6CC  uniD6CC,  HANGUL SYLLABLE-D6CC
        '54989' => '17995'            , # G+17995  U+D6CD  uniD6CD,  HANGUL SYLLABLE-D6CD
        '54990' => '17996'            , # G+17996  U+D6CE  uniD6CE,  HANGUL SYLLABLE-D6CE
        '54991' => '17997'            , # G+17997  U+D6CF  uniD6CF,  HANGUL SYLLABLE-D6CF
        '54992' => '17998'            , # G+17998  U+D6D0  uniD6D0,  HANGUL SYLLABLE-D6D0
        '54993' => '3381'             , # G+3381  U+D6D1  uniD6D1,  HANGUL SYLLABLE-D6D1
        '54994' => '17999'            , # G+17999  U+D6D2  uniD6D2,  HANGUL SYLLABLE-D6D2
        '54995' => '18000'            , # G+18000  U+D6D3  uniD6D3,  HANGUL SYLLABLE-D6D3
        '54996' => '3382'             , # G+3382  U+D6D4  uniD6D4,  HANGUL SYLLABLE-D6D4
        '54997' => '18001'            , # G+18001  U+D6D5  uniD6D5,  HANGUL SYLLABLE-D6D5
        '54998' => '18002'            , # G+18002  U+D6D6  uniD6D6,  HANGUL SYLLABLE-D6D6
        '54999' => '3383'             , # G+3383  U+D6D7  uniD6D7,  HANGUL SYLLABLE-D6D7
        '55000' => '18003'            , # G+18003  U+D6D8  uniD6D8,  HANGUL SYLLABLE-D6D8
        '55001' => '3384'             , # G+3384  U+D6D9  uniD6D9,  HANGUL SYLLABLE-D6D9
        '55002' => '18004'            , # G+18004  U+D6DA  uniD6DA,  HANGUL SYLLABLE-D6DA
        '55003' => '18005'            , # G+18005  U+D6DB  uniD6DB,  HANGUL SYLLABLE-D6DB
        '55004' => '18006'            , # G+18006  U+D6DC  uniD6DC,  HANGUL SYLLABLE-D6DC
        '55005' => '18007'            , # G+18007  U+D6DD  uniD6DD,  HANGUL SYLLABLE-D6DD
        '55006' => '18008'            , # G+18008  U+D6DE  uniD6DE,  HANGUL SYLLABLE-D6DE
        '55007' => '18009'            , # G+18009  U+D6DF  uniD6DF,  HANGUL SYLLABLE-D6DF
        '55008' => '3385'             , # G+3385  U+D6E0  uniD6E0,  HANGUL SYLLABLE-D6E0
        '55009' => '18010'            , # G+18010  U+D6E1  uniD6E1,  HANGUL SYLLABLE-D6E1
        '55010' => '18011'            , # G+18011  U+D6E2  uniD6E2,  HANGUL SYLLABLE-D6E2
        '55011' => '18012'            , # G+18012  U+D6E3  uniD6E3,  HANGUL SYLLABLE-D6E3
        '55012' => '3386'             , # G+3386  U+D6E4  uniD6E4,  HANGUL SYLLABLE-D6E4
        '55013' => '18013'            , # G+18013  U+D6E5  uniD6E5,  HANGUL SYLLABLE-D6E5
        '55014' => '18014'            , # G+18014  U+D6E6  uniD6E6,  HANGUL SYLLABLE-D6E6
        '55015' => '18015'            , # G+18015  U+D6E7  uniD6E7,  HANGUL SYLLABLE-D6E7
        '55016' => '3387'             , # G+3387  U+D6E8  uniD6E8,  HANGUL SYLLABLE-D6E8
        '55017' => '18016'            , # G+18016  U+D6E9  uniD6E9,  HANGUL SYLLABLE-D6E9
        '55018' => '18017'            , # G+18017  U+D6EA  uniD6EA,  HANGUL SYLLABLE-D6EA
        '55019' => '18018'            , # G+18018  U+D6EB  uniD6EB,  HANGUL SYLLABLE-D6EB
        '55020' => '18019'            , # G+18019  U+D6EC  uniD6EC,  HANGUL SYLLABLE-D6EC
        '55021' => '18020'            , # G+18020  U+D6ED  uniD6ED,  HANGUL SYLLABLE-D6ED
        '55022' => '18021'            , # G+18021  U+D6EE  uniD6EE,  HANGUL SYLLABLE-D6EE
        '55023' => '18022'            , # G+18022  U+D6EF  uniD6EF,  HANGUL SYLLABLE-D6EF
        '55024' => '3388'             , # G+3388  U+D6F0  uniD6F0,  HANGUL SYLLABLE-D6F0
        '55025' => '18023'            , # G+18023  U+D6F1  uniD6F1,  HANGUL SYLLABLE-D6F1
        '55026' => '18024'            , # G+18024  U+D6F2  uniD6F2,  HANGUL SYLLABLE-D6F2
        '55027' => '18025'            , # G+18025  U+D6F3  uniD6F3,  HANGUL SYLLABLE-D6F3
        '55028' => '18026'            , # G+18026  U+D6F4  uniD6F4,  HANGUL SYLLABLE-D6F4
        '55029' => '3389'             , # G+3389  U+D6F5  uniD6F5,  HANGUL SYLLABLE-D6F5
        '55030' => '18027'            , # G+18027  U+D6F6  uniD6F6,  HANGUL SYLLABLE-D6F6
        '55031' => '18028'            , # G+18028  U+D6F7  uniD6F7,  HANGUL SYLLABLE-D6F7
        '55032' => '18029'            , # G+18029  U+D6F8  uniD6F8,  HANGUL SYLLABLE-D6F8
        '55033' => '18030'            , # G+18030  U+D6F9  uniD6F9,  HANGUL SYLLABLE-D6F9
        '55034' => '18031'            , # G+18031  U+D6FA  uniD6FA,  HANGUL SYLLABLE-D6FA
        '55035' => '18032'            , # G+18032  U+D6FB  uniD6FB,  HANGUL SYLLABLE-D6FB
        '55036' => '3390'             , # G+3390  U+D6FC  uniD6FC,  HANGUL SYLLABLE-D6FC
        '55037' => '3391'             , # G+3391  U+D6FD  uniD6FD,  HANGUL SYLLABLE-D6FD
        '55038' => '18033'            , # G+18033  U+D6FE  uniD6FE,  HANGUL SYLLABLE-D6FE
        '55039' => '18034'            , # G+18034  U+D6FF  uniD6FF,  HANGUL SYLLABLE-D6FF
        '55040' => '3392'             , # G+3392  U+D700  uniD700,  HANGUL SYLLABLE-D700
        '55041' => '18035'            , # G+18035  U+D701  uniD701,  HANGUL SYLLABLE-D701
        '55042' => '18036'            , # G+18036  U+D702  uniD702,  HANGUL SYLLABLE-D702
        '55043' => '18037'            , # G+18037  U+D703  uniD703,  HANGUL SYLLABLE-D703
        '55044' => '3393'             , # G+3393  U+D704  uniD704,  HANGUL SYLLABLE-D704
        '55045' => '18038'            , # G+18038  U+D705  uniD705,  HANGUL SYLLABLE-D705
        '55046' => '18039'            , # G+18039  U+D706  uniD706,  HANGUL SYLLABLE-D706
        '55047' => '18040'            , # G+18040  U+D707  uniD707,  HANGUL SYLLABLE-D707
        '55048' => '18041'            , # G+18041  U+D708  uniD708,  HANGUL SYLLABLE-D708
        '55049' => '18042'            , # G+18042  U+D709  uniD709,  HANGUL SYLLABLE-D709
        '55050' => '18043'            , # G+18043  U+D70A  uniD70A,  HANGUL SYLLABLE-D70A
        '55051' => '18044'            , # G+18044  U+D70B  uniD70B,  HANGUL SYLLABLE-D70B
        '55052' => '18045'            , # G+18045  U+D70C  uniD70C,  HANGUL SYLLABLE-D70C
        '55053' => '18046'            , # G+18046  U+D70D  uniD70D,  HANGUL SYLLABLE-D70D
        '55054' => '18047'            , # G+18047  U+D70E  uniD70E,  HANGUL SYLLABLE-D70E
        '55055' => '18048'            , # G+18048  U+D70F  uniD70F,  HANGUL SYLLABLE-D70F
        '55056' => '18049'            , # G+18049  U+D710  uniD710,  HANGUL SYLLABLE-D710
        '55057' => '3394'             , # G+3394  U+D711  uniD711,  HANGUL SYLLABLE-D711
        '55058' => '18050'            , # G+18050  U+D712  uniD712,  HANGUL SYLLABLE-D712
        '55059' => '18051'            , # G+18051  U+D713  uniD713,  HANGUL SYLLABLE-D713
        '55060' => '18052'            , # G+18052  U+D714  uniD714,  HANGUL SYLLABLE-D714
        '55061' => '18053'            , # G+18053  U+D715  uniD715,  HANGUL SYLLABLE-D715
        '55062' => '18054'            , # G+18054  U+D716  uniD716,  HANGUL SYLLABLE-D716
        '55063' => '18055'            , # G+18055  U+D717  uniD717,  HANGUL SYLLABLE-D717
        '55064' => '3395'             , # G+3395  U+D718  uniD718,  HANGUL SYLLABLE-D718
        '55065' => '3396'             , # G+3396  U+D719  uniD719,  HANGUL SYLLABLE-D719
        '55066' => '18056'            , # G+18056  U+D71A  uniD71A,  HANGUL SYLLABLE-D71A
        '55067' => '18057'            , # G+18057  U+D71B  uniD71B,  HANGUL SYLLABLE-D71B
        '55068' => '3397'             , # G+3397  U+D71C  uniD71C,  HANGUL SYLLABLE-D71C
        '55069' => '18058'            , # G+18058  U+D71D  uniD71D,  HANGUL SYLLABLE-D71D
        '55070' => '18059'            , # G+18059  U+D71E  uniD71E,  HANGUL SYLLABLE-D71E
        '55071' => '18060'            , # G+18060  U+D71F  uniD71F,  HANGUL SYLLABLE-D71F
        '55072' => '3398'             , # G+3398  U+D720  uniD720,  HANGUL SYLLABLE-D720
        '55073' => '18061'            , # G+18061  U+D721  uniD721,  HANGUL SYLLABLE-D721
        '55074' => '18062'            , # G+18062  U+D722  uniD722,  HANGUL SYLLABLE-D722
        '55075' => '18063'            , # G+18063  U+D723  uniD723,  HANGUL SYLLABLE-D723
        '55076' => '18064'            , # G+18064  U+D724  uniD724,  HANGUL SYLLABLE-D724
        '55077' => '18065'            , # G+18065  U+D725  uniD725,  HANGUL SYLLABLE-D725
        '55078' => '18066'            , # G+18066  U+D726  uniD726,  HANGUL SYLLABLE-D726
        '55079' => '18067'            , # G+18067  U+D727  uniD727,  HANGUL SYLLABLE-D727
        '55080' => '3399'             , # G+3399  U+D728  uniD728,  HANGUL SYLLABLE-D728
        '55081' => '3400'             , # G+3400  U+D729  uniD729,  HANGUL SYLLABLE-D729
        '55082' => '18068'            , # G+18068  U+D72A  uniD72A,  HANGUL SYLLABLE-D72A
        '55083' => '3401'             , # G+3401  U+D72B  uniD72B,  HANGUL SYLLABLE-D72B
        '55084' => '18069'            , # G+18069  U+D72C  uniD72C,  HANGUL SYLLABLE-D72C
        '55085' => '3402'             , # G+3402  U+D72D  uniD72D,  HANGUL SYLLABLE-D72D
        '55086' => '18070'            , # G+18070  U+D72E  uniD72E,  HANGUL SYLLABLE-D72E
        '55087' => '18071'            , # G+18071  U+D72F  uniD72F,  HANGUL SYLLABLE-D72F
        '55088' => '18072'            , # G+18072  U+D730  uniD730,  HANGUL SYLLABLE-D730
        '55089' => '18073'            , # G+18073  U+D731  uniD731,  HANGUL SYLLABLE-D731
        '55090' => '18074'            , # G+18074  U+D732  uniD732,  HANGUL SYLLABLE-D732
        '55091' => '18075'            , # G+18075  U+D733  uniD733,  HANGUL SYLLABLE-D733
        '55092' => '3403'             , # G+3403  U+D734  uniD734,  HANGUL SYLLABLE-D734
        '55093' => '3404'             , # G+3404  U+D735  uniD735,  HANGUL SYLLABLE-D735
        '55094' => '18076'            , # G+18076  U+D736  uniD736,  HANGUL SYLLABLE-D736
        '55095' => '18077'            , # G+18077  U+D737  uniD737,  HANGUL SYLLABLE-D737
        '55096' => '3405'             , # G+3405  U+D738  uniD738,  HANGUL SYLLABLE-D738
        '55097' => '18078'            , # G+18078  U+D739  uniD739,  HANGUL SYLLABLE-D739
        '55098' => '18079'            , # G+18079  U+D73A  uniD73A,  HANGUL SYLLABLE-D73A
        '55099' => '18080'            , # G+18080  U+D73B  uniD73B,  HANGUL SYLLABLE-D73B
        '55100' => '3406'             , # G+3406  U+D73C  uniD73C,  HANGUL SYLLABLE-D73C
        '55101' => '18081'            , # G+18081  U+D73D  uniD73D,  HANGUL SYLLABLE-D73D
        '55102' => '18082'            , # G+18082  U+D73E  uniD73E,  HANGUL SYLLABLE-D73E
        '55103' => '18083'            , # G+18083  U+D73F  uniD73F,  HANGUL SYLLABLE-D73F
        '55104' => '18084'            , # G+18084  U+D740  uniD740,  HANGUL SYLLABLE-D740
        '55105' => '18085'            , # G+18085  U+D741  uniD741,  HANGUL SYLLABLE-D741
        '55106' => '18086'            , # G+18086  U+D742  uniD742,  HANGUL SYLLABLE-D742
        '55107' => '18087'            , # G+18087  U+D743  uniD743,  HANGUL SYLLABLE-D743
        '55108' => '3407'             , # G+3407  U+D744  uniD744,  HANGUL SYLLABLE-D744
        '55109' => '18088'            , # G+18088  U+D745  uniD745,  HANGUL SYLLABLE-D745
        '55110' => '18089'            , # G+18089  U+D746  uniD746,  HANGUL SYLLABLE-D746
        '55111' => '3408'             , # G+3408  U+D747  uniD747,  HANGUL SYLLABLE-D747
        '55112' => '18090'            , # G+18090  U+D748  uniD748,  HANGUL SYLLABLE-D748
        '55113' => '3409'             , # G+3409  U+D749  uniD749,  HANGUL SYLLABLE-D749
        '55114' => '18091'            , # G+18091  U+D74A  uniD74A,  HANGUL SYLLABLE-D74A
        '55115' => '18092'            , # G+18092  U+D74B  uniD74B,  HANGUL SYLLABLE-D74B
        '55116' => '18093'            , # G+18093  U+D74C  uniD74C,  HANGUL SYLLABLE-D74C
        '55117' => '18094'            , # G+18094  U+D74D  uniD74D,  HANGUL SYLLABLE-D74D
        '55118' => '18095'            , # G+18095  U+D74E  uniD74E,  HANGUL SYLLABLE-D74E
        '55119' => '18096'            , # G+18096  U+D74F  uniD74F,  HANGUL SYLLABLE-D74F
        '55120' => '3410'             , # G+3410  U+D750  uniD750,  HANGUL SYLLABLE-D750
        '55121' => '3411'             , # G+3411  U+D751  uniD751,  HANGUL SYLLABLE-D751
        '55122' => '18097'            , # G+18097  U+D752  uniD752,  HANGUL SYLLABLE-D752
        '55123' => '18098'            , # G+18098  U+D753  uniD753,  HANGUL SYLLABLE-D753
        '55124' => '3412'             , # G+3412  U+D754  uniD754,  HANGUL SYLLABLE-D754
        '55125' => '18099'            , # G+18099  U+D755  uniD755,  HANGUL SYLLABLE-D755
        '55126' => '3413'             , # G+3413  U+D756  uniD756,  HANGUL SYLLABLE-D756
        '55127' => '3414'             , # G+3414  U+D757  uniD757,  HANGUL SYLLABLE-D757
        '55128' => '3415'             , # G+3415  U+D758  uniD758,  HANGUL SYLLABLE-D758
        '55129' => '3416'             , # G+3416  U+D759  uniD759,  HANGUL SYLLABLE-D759
        '55130' => '18100'            , # G+18100  U+D75A  uniD75A,  HANGUL SYLLABLE-D75A
        '55131' => '18101'            , # G+18101  U+D75B  uniD75B,  HANGUL SYLLABLE-D75B
        '55132' => '18102'            , # G+18102  U+D75C  uniD75C,  HANGUL SYLLABLE-D75C
        '55133' => '18103'            , # G+18103  U+D75D  uniD75D,  HANGUL SYLLABLE-D75D
        '55134' => '18104'            , # G+18104  U+D75E  uniD75E,  HANGUL SYLLABLE-D75E
        '55135' => '18105'            , # G+18105  U+D75F  uniD75F,  HANGUL SYLLABLE-D75F
        '55136' => '3417'             , # G+3417  U+D760  uniD760,  HANGUL SYLLABLE-D760
        '55137' => '3418'             , # G+3418  U+D761  uniD761,  HANGUL SYLLABLE-D761
        '55138' => '18106'            , # G+18106  U+D762  uniD762,  HANGUL SYLLABLE-D762
        '55139' => '3419'             , # G+3419  U+D763  uniD763,  HANGUL SYLLABLE-D763
        '55140' => '18107'            , # G+18107  U+D764  uniD764,  HANGUL SYLLABLE-D764
        '55141' => '3420'             , # G+3420  U+D765  uniD765,  HANGUL SYLLABLE-D765
        '55142' => '18108'            , # G+18108  U+D766  uniD766,  HANGUL SYLLABLE-D766
        '55143' => '18109'            , # G+18109  U+D767  uniD767,  HANGUL SYLLABLE-D767
        '55144' => '18110'            , # G+18110  U+D768  uniD768,  HANGUL SYLLABLE-D768
        '55145' => '3421'             , # G+3421  U+D769  uniD769,  HANGUL SYLLABLE-D769
        '55146' => '18111'            , # G+18111  U+D76A  uniD76A,  HANGUL SYLLABLE-D76A
        '55147' => '18112'            , # G+18112  U+D76B  uniD76B,  HANGUL SYLLABLE-D76B
        '55148' => '3422'             , # G+3422  U+D76C  uniD76C,  HANGUL SYLLABLE-D76C
        '55149' => '18113'            , # G+18113  U+D76D  uniD76D,  HANGUL SYLLABLE-D76D
        '55150' => '18114'            , # G+18114  U+D76E  uniD76E,  HANGUL SYLLABLE-D76E
        '55151' => '18115'            , # G+18115  U+D76F  uniD76F,  HANGUL SYLLABLE-D76F
        '55152' => '3423'             , # G+3423  U+D770  uniD770,  HANGUL SYLLABLE-D770
        '55153' => '18116'            , # G+18116  U+D771  uniD771,  HANGUL SYLLABLE-D771
        '55154' => '18117'            , # G+18117  U+D772  uniD772,  HANGUL SYLLABLE-D772
        '55155' => '18118'            , # G+18118  U+D773  uniD773,  HANGUL SYLLABLE-D773
        '55156' => '3424'             , # G+3424  U+D774  uniD774,  HANGUL SYLLABLE-D774
        '55157' => '18119'            , # G+18119  U+D775  uniD775,  HANGUL SYLLABLE-D775
        '55158' => '18120'            , # G+18120  U+D776  uniD776,  HANGUL SYLLABLE-D776
        '55159' => '18121'            , # G+18121  U+D777  uniD777,  HANGUL SYLLABLE-D777
        '55160' => '18122'            , # G+18122  U+D778  uniD778,  HANGUL SYLLABLE-D778
        '55161' => '18123'            , # G+18123  U+D779  uniD779,  HANGUL SYLLABLE-D779
        '55162' => '18124'            , # G+18124  U+D77A  uniD77A,  HANGUL SYLLABLE-D77A
        '55163' => '18125'            , # G+18125  U+D77B  uniD77B,  HANGUL SYLLABLE-D77B
        '55164' => '3425'             , # G+3425  U+D77C  uniD77C,  HANGUL SYLLABLE-D77C
        '55165' => '3426'             , # G+3426  U+D77D  uniD77D,  HANGUL SYLLABLE-D77D
        '55166' => '18126'            , # G+18126  U+D77E  uniD77E,  HANGUL SYLLABLE-D77E
        '55167' => '18127'            , # G+18127  U+D77F  uniD77F,  HANGUL SYLLABLE-D77F
        '55168' => '18128'            , # G+18128  U+D780  uniD780,  HANGUL SYLLABLE-D780
        '55169' => '3427'             , # G+3427  U+D781  uniD781,  HANGUL SYLLABLE-D781
        '55170' => '18129'            , # G+18129  U+D782  uniD782,  HANGUL SYLLABLE-D782
        '55171' => '18130'            , # G+18130  U+D783  uniD783,  HANGUL SYLLABLE-D783
        '55172' => '18131'            , # G+18131  U+D784  uniD784,  HANGUL SYLLABLE-D784
        '55173' => '18132'            , # G+18132  U+D785  uniD785,  HANGUL SYLLABLE-D785
        '55174' => '18133'            , # G+18133  U+D786  uniD786,  HANGUL SYLLABLE-D786
        '55175' => '18134'            , # G+18134  U+D787  uniD787,  HANGUL SYLLABLE-D787
        '55176' => '3428'             , # G+3428  U+D788  uniD788,  HANGUL SYLLABLE-D788
        '55177' => '3429'             , # G+3429  U+D789  uniD789,  HANGUL SYLLABLE-D789
        '55178' => '18135'            , # G+18135  U+D78A  uniD78A,  HANGUL SYLLABLE-D78A
        '55179' => '18136'            , # G+18136  U+D78B  uniD78B,  HANGUL SYLLABLE-D78B
        '55180' => '3430'             , # G+3430  U+D78C  uniD78C,  HANGUL SYLLABLE-D78C
        '55181' => '18137'            , # G+18137  U+D78D  uniD78D,  HANGUL SYLLABLE-D78D
        '55182' => '18138'            , # G+18138  U+D78E  uniD78E,  HANGUL SYLLABLE-D78E
        '55183' => '18139'            , # G+18139  U+D78F  uniD78F,  HANGUL SYLLABLE-D78F
        '55184' => '3431'             , # G+3431  U+D790  uniD790,  HANGUL SYLLABLE-D790
        '55185' => '18140'            , # G+18140  U+D791  uniD791,  HANGUL SYLLABLE-D791
        '55186' => '18141'            , # G+18141  U+D792  uniD792,  HANGUL SYLLABLE-D792
        '55187' => '18142'            , # G+18142  U+D793  uniD793,  HANGUL SYLLABLE-D793
        '55188' => '18143'            , # G+18143  U+D794  uniD794,  HANGUL SYLLABLE-D794
        '55189' => '18144'            , # G+18144  U+D795  uniD795,  HANGUL SYLLABLE-D795
        '55190' => '18145'            , # G+18145  U+D796  uniD796,  HANGUL SYLLABLE-D796
        '55191' => '18146'            , # G+18146  U+D797  uniD797,  HANGUL SYLLABLE-D797
        '55192' => '3432'             , # G+3432  U+D798  uniD798,  HANGUL SYLLABLE-D798
        '55193' => '3433'             , # G+3433  U+D799  uniD799,  HANGUL SYLLABLE-D799
        '55194' => '18147'            , # G+18147  U+D79A  uniD79A,  HANGUL SYLLABLE-D79A
        '55195' => '3434'             , # G+3434  U+D79B  uniD79B,  HANGUL SYLLABLE-D79B
        '55196' => '18148'            , # G+18148  U+D79C  uniD79C,  HANGUL SYLLABLE-D79C
        '55197' => '3435'             , # G+3435  U+D79D  uniD79D,  HANGUL SYLLABLE-D79D
        '55198' => '18149'            , # G+18149  U+D79E  uniD79E,  HANGUL SYLLABLE-D79E
        '55199' => '18150'            , # G+18150  U+D79F  uniD79F,  HANGUL SYLLABLE-D79F
        '55200' => '18151'            , # G+18151  U+D7A0  uniD7A0,  HANGUL SYLLABLE-D7A0
        '55201' => '18152'            , # G+18152  U+D7A1  uniD7A1,  HANGUL SYLLABLE-D7A1
        '55202' => '18153'            , # G+18153  U+D7A2  uniD7A2,  HANGUL SYLLABLE-D7A2
        '55203' => '18154'            , # G+18154  U+D7A3  uniD7A3,  HANGUL SYLLABLE-D7A3
        '63744' => '4116'             , # G+4116  U+F900  uniF900,  CJK COMPATIBILITY IDEOGRAPH-F900
        '63745' => '3678'             , # G+3678  U+F901  uniF901,  CJK COMPATIBILITY IDEOGRAPH-F901
        '63746' => '7053'             , # G+7053  U+F902  uniF902,  CJK COMPATIBILITY IDEOGRAPH-F902
        '63747' => '3460'             , # G+3460  U+F903  uniF903,  CJK COMPATIBILITY IDEOGRAPH-F903
        '63748' => '7900'             , # G+7900  U+F904  uniF904,  CJK COMPATIBILITY IDEOGRAPH-F904
        '63749' => '3802'             , # G+3802  U+F905  uniF905,  CJK COMPATIBILITY IDEOGRAPH-F905
        '63750' => '3902'             , # G+3902  U+F906  uniF906,  CJK COMPATIBILITY IDEOGRAPH-F906
        '63751' => '3946'             , # G+3946  U+F907  uniF907,  CJK COMPATIBILITY IDEOGRAPH-F907
        '63752' => '3946'             , # G+3946  U+F908  uniF908,  CJK COMPATIBILITY IDEOGRAPH-F908
        '63753' => '3708'             , # G+3708  U+F909  uniF909,  CJK COMPATIBILITY IDEOGRAPH-F909
        '63754' => '4131'             , # G+4131  U+F90A  uniF90A,  CJK COMPATIBILITY IDEOGRAPH-F90A
        '63755' => '4374'             , # G+4374  U+F90B  uniF90B,  CJK COMPATIBILITY IDEOGRAPH-F90B
        '63756' => '4156'             , # G+4156  U+F90C  uniF90C,  CJK COMPATIBILITY IDEOGRAPH-F90C
        '63757' => '4375'             , # G+4375  U+F90D  uniF90D,  CJK COMPATIBILITY IDEOGRAPH-F90D
        '63758' => '4376'             , # G+4376  U+F90E  uniF90E,  CJK COMPATIBILITY IDEOGRAPH-F90E
        '63759' => '4377'             , # G+4377  U+F90F  uniF90F,  CJK COMPATIBILITY IDEOGRAPH-F90F
        '63760' => '4378'             , # G+4378  U+F910  uniF910,  CJK COMPATIBILITY IDEOGRAPH-F910
        '63761' => '4379'             , # G+4379  U+F911  uniF911,  CJK COMPATIBILITY IDEOGRAPH-F911
        '63762' => '4380'             , # G+4380  U+F912  uniF912,  CJK COMPATIBILITY IDEOGRAPH-F912
        '63763' => '4381'             , # G+4381  U+F913  uniF913,  CJK COMPATIBILITY IDEOGRAPH-F913
        '63764' => '5800'             , # G+5800  U+F914  uniF914,  CJK COMPATIBILITY IDEOGRAPH-F914
        '63765' => '4382'             , # G+4382  U+F915  uniF915,  CJK COMPATIBILITY IDEOGRAPH-F915
        '63766' => '4383'             , # G+4383  U+F916  uniF916,  CJK COMPATIBILITY IDEOGRAPH-F916
        '63767' => '4384'             , # G+4384  U+F917  uniF917,  CJK COMPATIBILITY IDEOGRAPH-F917
        '63768' => '4386'             , # G+4386  U+F918  uniF918,  CJK COMPATIBILITY IDEOGRAPH-F918
        '63769' => '4387'             , # G+4387  U+F919  uniF919,  CJK COMPATIBILITY IDEOGRAPH-F919
        '63770' => '4388'             , # G+4388  U+F91A  uniF91A,  CJK COMPATIBILITY IDEOGRAPH-F91A
        '63771' => '4389'             , # G+4389  U+F91B  uniF91B,  CJK COMPATIBILITY IDEOGRAPH-F91B
        '63772' => '4390'             , # G+4390  U+F91C  uniF91C,  CJK COMPATIBILITY IDEOGRAPH-F91C
        '63773' => '4391'             , # G+4391  U+F91D  uniF91D,  CJK COMPATIBILITY IDEOGRAPH-F91D
        '63774' => '4394'             , # G+4394  U+F91E  uniF91E,  CJK COMPATIBILITY IDEOGRAPH-F91E
        '63775' => '4395'             , # G+4395  U+F91F  uniF91F,  CJK COMPATIBILITY IDEOGRAPH-F91F
        '63776' => '4396'             , # G+4396  U+F920  uniF920,  CJK COMPATIBILITY IDEOGRAPH-F920
        '63777' => '4399'             , # G+4399  U+F921  uniF921,  CJK COMPATIBILITY IDEOGRAPH-F921
        '63778' => '4403'             , # G+4403  U+F922  uniF922,  CJK COMPATIBILITY IDEOGRAPH-F922
        '63779' => '4406'             , # G+4406  U+F923  uniF923,  CJK COMPATIBILITY IDEOGRAPH-F923
        '63780' => '4407'             , # G+4407  U+F924  uniF924,  CJK COMPATIBILITY IDEOGRAPH-F924
        '63781' => '4409'             , # G+4409  U+F925  uniF925,  CJK COMPATIBILITY IDEOGRAPH-F925
        '63782' => '4410'             , # G+4410  U+F926  uniF926,  CJK COMPATIBILITY IDEOGRAPH-F926
        '63783' => '4411'             , # G+4411  U+F927  uniF927,  CJK COMPATIBILITY IDEOGRAPH-F927
        '63784' => '4412'             , # G+4412  U+F928  uniF928,  CJK COMPATIBILITY IDEOGRAPH-F928
        '63785' => '4413'             , # G+4413  U+F929  uniF929,  CJK COMPATIBILITY IDEOGRAPH-F929
        '63786' => '4414'             , # G+4414  U+F92A  uniF92A,  CJK COMPATIBILITY IDEOGRAPH-F92A
        '63787' => '4415'             , # G+4415  U+F92B  uniF92B,  CJK COMPATIBILITY IDEOGRAPH-F92B
        '63788' => '4419'             , # G+4419  U+F92C  uniF92C,  CJK COMPATIBILITY IDEOGRAPH-F92C
        '63789' => '4420'             , # G+4420  U+F92D  uniF92D,  CJK COMPATIBILITY IDEOGRAPH-F92D
        '63790' => '4424'             , # G+4424  U+F92E  uniF92E,  CJK COMPATIBILITY IDEOGRAPH-F92E
        '63791' => '4511'             , # G+4511  U+F92F  uniF92F,  CJK COMPATIBILITY IDEOGRAPH-F92F
        '63792' => '4513'             , # G+4513  U+F930  uniF930,  CJK COMPATIBILITY IDEOGRAPH-F930
        '63793' => '4514'             , # G+4514  U+F931  uniF931,  CJK COMPATIBILITY IDEOGRAPH-F931
        '63794' => '4517'             , # G+4517  U+F932  uniF932,  CJK COMPATIBILITY IDEOGRAPH-F932
        '63795' => '4518'             , # G+4518  U+F933  uniF933,  CJK COMPATIBILITY IDEOGRAPH-F933
        '63796' => '4519'             , # G+4519  U+F934  uniF934,  CJK COMPATIBILITY IDEOGRAPH-F934
        '63797' => '4520'             , # G+4520  U+F935  uniF935,  CJK COMPATIBILITY IDEOGRAPH-F935
        '63798' => '4521'             , # G+4521  U+F936  uniF936,  CJK COMPATIBILITY IDEOGRAPH-F936
        '63799' => '4522'             , # G+4522  U+F937  uniF937,  CJK COMPATIBILITY IDEOGRAPH-F937
        '63800' => '4524'             , # G+4524  U+F938  uniF938,  CJK COMPATIBILITY IDEOGRAPH-F938
        '63801' => '4525'             , # G+4525  U+F939  uniF939,  CJK COMPATIBILITY IDEOGRAPH-F939
        '63802' => '4526'             , # G+4526  U+F93A  uniF93A,  CJK COMPATIBILITY IDEOGRAPH-F93A
        '63803' => '4528'             , # G+4528  U+F93B  uniF93B,  CJK COMPATIBILITY IDEOGRAPH-F93B
        '63804' => '4529'             , # G+4529  U+F93C  uniF93C,  CJK COMPATIBILITY IDEOGRAPH-F93C
        '63805' => '4530'             , # G+4530  U+F93D  uniF93D,  CJK COMPATIBILITY IDEOGRAPH-F93D
        '63806' => '4531'             , # G+4531  U+F93E  uniF93E,  CJK COMPATIBILITY IDEOGRAPH-F93E
        '63807' => '4532'             , # G+4532  U+F93F  uniF93F,  CJK COMPATIBILITY IDEOGRAPH-F93F
        '63808' => '4533'             , # G+4533  U+F940  uniF940,  CJK COMPATIBILITY IDEOGRAPH-F940
        '63809' => '4535'             , # G+4535  U+F941  uniF941,  CJK COMPATIBILITY IDEOGRAPH-F941
        '63810' => '4536'             , # G+4536  U+F942  uniF942,  CJK COMPATIBILITY IDEOGRAPH-F942
        '63811' => '4537'             , # G+4537  U+F943  uniF943,  CJK COMPATIBILITY IDEOGRAPH-F943
        '63812' => '4541'             , # G+4541  U+F944  uniF944,  CJK COMPATIBILITY IDEOGRAPH-F944
        '63813' => '4542'             , # G+4542  U+F945  uniF945,  CJK COMPATIBILITY IDEOGRAPH-F945
        '63814' => '4545'             , # G+4545  U+F946  uniF946,  CJK COMPATIBILITY IDEOGRAPH-F946
        '63815' => '4546'             , # G+4546  U+F947  uniF947,  CJK COMPATIBILITY IDEOGRAPH-F947
        '63816' => '4547'             , # G+4547  U+F948  uniF948,  CJK COMPATIBILITY IDEOGRAPH-F948
        '63817' => '4550'             , # G+4550  U+F949  uniF949,  CJK COMPATIBILITY IDEOGRAPH-F949
        '63818' => '4564'             , # G+4564  U+F94A  uniF94A,  CJK COMPATIBILITY IDEOGRAPH-F94A
        '63819' => '4566'             , # G+4566  U+F94B  uniF94B,  CJK COMPATIBILITY IDEOGRAPH-F94B
        '63820' => '4567'             , # G+4567  U+F94C  uniF94C,  CJK COMPATIBILITY IDEOGRAPH-F94C
        '63821' => '4568'             , # G+4568  U+F94D  uniF94D,  CJK COMPATIBILITY IDEOGRAPH-F94D
        '63822' => '4569'             , # G+4569  U+F94E  uniF94E,  CJK COMPATIBILITY IDEOGRAPH-F94E
        '63823' => '4571'             , # G+4571  U+F94F  uniF94F,  CJK COMPATIBILITY IDEOGRAPH-F94F
        '63824' => '4572'             , # G+4572  U+F950  uniF950,  CJK COMPATIBILITY IDEOGRAPH-F950
        '63825' => '4576'             , # G+4576  U+F951  uniF951,  CJK COMPATIBILITY IDEOGRAPH-F951
        '63826' => '4604'             , # G+4604  U+F952  uniF952,  CJK COMPATIBILITY IDEOGRAPH-F952
        '63827' => '4605'             , # G+4605  U+F953  uniF953,  CJK COMPATIBILITY IDEOGRAPH-F953
        '63828' => '4606'             , # G+4606  U+F954  uniF954,  CJK COMPATIBILITY IDEOGRAPH-F954
        '63829' => '4607'             , # G+4607  U+F955  uniF955,  CJK COMPATIBILITY IDEOGRAPH-F955
        '63830' => '4609'             , # G+4609  U+F956  uniF956,  CJK COMPATIBILITY IDEOGRAPH-F956
        '63831' => '4610'             , # G+4610  U+F957  uniF957,  CJK COMPATIBILITY IDEOGRAPH-F957
        '63832' => '4611'             , # G+4611  U+F958  uniF958,  CJK COMPATIBILITY IDEOGRAPH-F958
        '63833' => '4612'             , # G+4612  U+F959  uniF959,  CJK COMPATIBILITY IDEOGRAPH-F959
        '63834' => '4318'             , # G+4318  U+F95A  uniF95A,  CJK COMPATIBILITY IDEOGRAPH-F95A
        '63835' => '4136'             , # G+4136  U+F95B  uniF95B,  CJK COMPATIBILITY IDEOGRAPH-F95B
        '63836' => '5800'             , # G+5800  U+F95C  uniF95C,  CJK COMPATIBILITY IDEOGRAPH-F95C
        '63837' => '4139'             , # G+4139  U+F95D  uniF95D,  CJK COMPATIBILITY IDEOGRAPH-F95D
        '63838' => '4192'             , # G+4192  U+F95E  uniF95E,  CJK COMPATIBILITY IDEOGRAPH-F95E
        '63839' => '4167'             , # G+4167  U+F95F  uniF95F,  CJK COMPATIBILITY IDEOGRAPH-F95F
        '63840' => '4172'             , # G+4172  U+F960  uniF960,  CJK COMPATIBILITY IDEOGRAPH-F960
        '63841' => '5552'             , # G+5552  U+F961  uniF961,  CJK COMPATIBILITY IDEOGRAPH-F961
        '63842' => '6424'             , # G+6424  U+F962  uniF962,  CJK COMPATIBILITY IDEOGRAPH-F962
        '63843' => '5151'             , # G+5151  U+F963  uniF963,  CJK COMPATIBILITY IDEOGRAPH-F963
        '63844' => '4922'             , # G+4922  U+F964  uniF964,  CJK COMPATIBILITY IDEOGRAPH-F964
        '63845' => '7518'             , # G+7518  U+F965  uniF965,  CJK COMPATIBILITY IDEOGRAPH-F965
        '63846' => '5079'             , # G+5079  U+F966  uniF966,  CJK COMPATIBILITY IDEOGRAPH-F966
        '63847' => '5109'             , # G+5109  U+F967  uniF967,  CJK COMPATIBILITY IDEOGRAPH-F967
        '63848' => '7607'             , # G+7607  U+F968  uniF968,  CJK COMPATIBILITY IDEOGRAPH-F968
        '63849' => '5584'             , # G+5584  U+F969  uniF969,  CJK COMPATIBILITY IDEOGRAPH-F969
        '63850' => '5367'             , # G+5367  U+F96A  uniF96A,  CJK COMPATIBILITY IDEOGRAPH-F96A
        '63851' => '7083'             , # G+7083  U+F96B  uniF96B,  CJK COMPATIBILITY IDEOGRAPH-F96B
        '63852' => '5362'             , # G+5362  U+F96C  uniF96C,  CJK COMPATIBILITY IDEOGRAPH-F96C
        '63853' => '5485'             , # G+5485  U+F96D  uniF96D,  CJK COMPATIBILITY IDEOGRAPH-F96D
        '63854' => '6001'             , # G+6001  U+F96E  uniF96E,  CJK COMPATIBILITY IDEOGRAPH-F96E
        '63855' => '5460'             , # G+5460  U+F96F  uniF96F,  CJK COMPATIBILITY IDEOGRAPH-F96F
        '63856' => '5317'             , # G+5317  U+F970  uniF970,  CJK COMPATIBILITY IDEOGRAPH-F970
        '63857' => '7009'             , # G+7009  U+F971  uniF971,  CJK COMPATIBILITY IDEOGRAPH-F971
        '63858' => '7370'             , # G+7370  U+F972  uniF972,  CJK COMPATIBILITY IDEOGRAPH-F972
        '63859' => '5678'             , # G+5678  U+F973  uniF973,  CJK COMPATIBILITY IDEOGRAPH-F973
        '63860' => '5874'             , # G+5874  U+F974  uniF974,  CJK COMPATIBILITY IDEOGRAPH-F974
        '63861' => '4425'             , # G+4425  U+F975  uniF975,  CJK COMPATIBILITY IDEOGRAPH-F975
        '63862' => '4426'             , # G+4426  U+F976  uniF976,  CJK COMPATIBILITY IDEOGRAPH-F976
        '63863' => '4427'             , # G+4427  U+F977  uniF977,  CJK COMPATIBILITY IDEOGRAPH-F977
        '63864' => '4429'             , # G+4429  U+F978  uniF978,  CJK COMPATIBILITY IDEOGRAPH-F978
        '63865' => '4430'             , # G+4430  U+F979  uniF979,  CJK COMPATIBILITY IDEOGRAPH-F979
        '63866' => '4431'             , # G+4431  U+F97A  uniF97A,  CJK COMPATIBILITY IDEOGRAPH-F97A
        '63867' => '4435'             , # G+4435  U+F97B  uniF97B,  CJK COMPATIBILITY IDEOGRAPH-F97B
        '63868' => '4436'             , # G+4436  U+F97C  uniF97C,  CJK COMPATIBILITY IDEOGRAPH-F97C
        '63869' => '4437'             , # G+4437  U+F97D  uniF97D,  CJK COMPATIBILITY IDEOGRAPH-F97D
        '63870' => '4439'             , # G+4439  U+F97E  uniF97E,  CJK COMPATIBILITY IDEOGRAPH-F97E
        '63871' => '4442'             , # G+4442  U+F97F  uniF97F,  CJK COMPATIBILITY IDEOGRAPH-F97F
        '63872' => '4443'             , # G+4443  U+F980  uniF980,  CJK COMPATIBILITY IDEOGRAPH-F980
        '63873' => '4159'             , # G+4159  U+F981  uniF981,  CJK COMPATIBILITY IDEOGRAPH-F981
        '63874' => '4444'             , # G+4444  U+F982  uniF982,  CJK COMPATIBILITY IDEOGRAPH-F982
        '63875' => '4447'             , # G+4447  U+F983  uniF983,  CJK COMPATIBILITY IDEOGRAPH-F983
        '63876' => '4449'             , # G+4449  U+F984  uniF984,  CJK COMPATIBILITY IDEOGRAPH-F984
        '63877' => '4450'             , # G+4450  U+F985  uniF985,  CJK COMPATIBILITY IDEOGRAPH-F985
        '63878' => '4453'             , # G+4453  U+F986  uniF986,  CJK COMPATIBILITY IDEOGRAPH-F986
        '63879' => '4455'             , # G+4455  U+F987  uniF987,  CJK COMPATIBILITY IDEOGRAPH-F987
        '63880' => '4456'             , # G+4456  U+F988  uniF988,  CJK COMPATIBILITY IDEOGRAPH-F988
        '63881' => '4457'             , # G+4457  U+F989  uniF989,  CJK COMPATIBILITY IDEOGRAPH-F989
        '63882' => '4458'             , # G+4458  U+F98A  uniF98A,  CJK COMPATIBILITY IDEOGRAPH-F98A
        '63883' => '4459'             , # G+4459  U+F98B  uniF98B,  CJK COMPATIBILITY IDEOGRAPH-F98B
        '63884' => '4460'             , # G+4460  U+F98C  uniF98C,  CJK COMPATIBILITY IDEOGRAPH-F98C
        '63885' => '4463'             , # G+4463  U+F98D  uniF98D,  CJK COMPATIBILITY IDEOGRAPH-F98D
        '63886' => '4160'             , # G+4160  U+F98E  uniF98E,  CJK COMPATIBILITY IDEOGRAPH-F98E
        '63887' => '4465'             , # G+4465  U+F98F  uniF98F,  CJK COMPATIBILITY IDEOGRAPH-F98F
        '63888' => '4466'             , # G+4466  U+F990  uniF990,  CJK COMPATIBILITY IDEOGRAPH-F990
        '63889' => '4161'             , # G+4161  U+F991  uniF991,  CJK COMPATIBILITY IDEOGRAPH-F991
        '63890' => '4468'             , # G+4468  U+F992  uniF992,  CJK COMPATIBILITY IDEOGRAPH-F992
        '63891' => '4469'             , # G+4469  U+F993  uniF993,  CJK COMPATIBILITY IDEOGRAPH-F993
        '63892' => '4470'             , # G+4470  U+F994  uniF994,  CJK COMPATIBILITY IDEOGRAPH-F994
        '63893' => '4162'             , # G+4162  U+F995  uniF995,  CJK COMPATIBILITY IDEOGRAPH-F995
        '63894' => '4471'             , # G+4471  U+F996  uniF996,  CJK COMPATIBILITY IDEOGRAPH-F996
        '63895' => '4472'             , # G+4472  U+F997  uniF997,  CJK COMPATIBILITY IDEOGRAPH-F997
        '63896' => '4474'             , # G+4474  U+F998  uniF998,  CJK COMPATIBILITY IDEOGRAPH-F998
        '63897' => '4473'             , # G+4473  U+F999  uniF999,  CJK COMPATIBILITY IDEOGRAPH-F999
        '63898' => '4475'             , # G+4475  U+F99A  uniF99A,  CJK COMPATIBILITY IDEOGRAPH-F99A
        '63899' => '4476'             , # G+4476  U+F99B  uniF99B,  CJK COMPATIBILITY IDEOGRAPH-F99B
        '63900' => '4478'             , # G+4478  U+F99C  uniF99C,  CJK COMPATIBILITY IDEOGRAPH-F99C
        '63901' => '4479'             , # G+4479  U+F99D  uniF99D,  CJK COMPATIBILITY IDEOGRAPH-F99D
        '63902' => '6447'             , # G+6447  U+F99E  uniF99E,  CJK COMPATIBILITY IDEOGRAPH-F99E
        '63903' => '4481'             , # G+4481  U+F99F  uniF99F,  CJK COMPATIBILITY IDEOGRAPH-F99F
        '63904' => '4482'             , # G+4482  U+F9A0  uniF9A0,  CJK COMPATIBILITY IDEOGRAPH-F9A0
        '63905' => '5460'             , # G+5460  U+F9A1  uniF9A1,  CJK COMPATIBILITY IDEOGRAPH-F9A1
        '63906' => '4483'             , # G+4483  U+F9A2  uniF9A2,  CJK COMPATIBILITY IDEOGRAPH-F9A2
        '63907' => '4163'             , # G+4163  U+F9A3  uniF9A3,  CJK COMPATIBILITY IDEOGRAPH-F9A3
        '63908' => '4166'             , # G+4166  U+F9A4  uniF9A4,  CJK COMPATIBILITY IDEOGRAPH-F9A4
        '63909' => '4485'             , # G+4485  U+F9A5  uniF9A5,  CJK COMPATIBILITY IDEOGRAPH-F9A5
        '63910' => '4487'             , # G+4487  U+F9A6  uniF9A6,  CJK COMPATIBILITY IDEOGRAPH-F9A6
        '63911' => '4488'             , # G+4488  U+F9A7  uniF9A7,  CJK COMPATIBILITY IDEOGRAPH-F9A7
        '63912' => '4489'             , # G+4489  U+F9A8  uniF9A8,  CJK COMPATIBILITY IDEOGRAPH-F9A8
        '63913' => '4491'             , # G+4491  U+F9A9  uniF9A9,  CJK COMPATIBILITY IDEOGRAPH-F9A9
        '63914' => '4167'             , # G+4167  U+F9AA  uniF9AA,  CJK COMPATIBILITY IDEOGRAPH-F9AA
        '63915' => '4493'             , # G+4493  U+F9AB  uniF9AB,  CJK COMPATIBILITY IDEOGRAPH-F9AB
        '63916' => '4494'             , # G+4494  U+F9AC  uniF9AC,  CJK COMPATIBILITY IDEOGRAPH-F9AC
        '63917' => '4495'             , # G+4495  U+F9AD  uniF9AD,  CJK COMPATIBILITY IDEOGRAPH-F9AD
        '63918' => '7783'             , # G+7783  U+F9AE  uniF9AE,  CJK COMPATIBILITY IDEOGRAPH-F9AE
        '63919' => '4497'             , # G+4497  U+F9AF  uniF9AF,  CJK COMPATIBILITY IDEOGRAPH-F9AF
        '63920' => '4499'             , # G+4499  U+F9B0  uniF9B0,  CJK COMPATIBILITY IDEOGRAPH-F9B0
        '63921' => '4501'             , # G+4501  U+F9B1  uniF9B1,  CJK COMPATIBILITY IDEOGRAPH-F9B1
        '63922' => '4502'             , # G+4502  U+F9B2  uniF9B2,  CJK COMPATIBILITY IDEOGRAPH-F9B2
        '63923' => '4503'             , # G+4503  U+F9B3  uniF9B3,  CJK COMPATIBILITY IDEOGRAPH-F9B3
        '63924' => '4504'             , # G+4504  U+F9B4  uniF9B4,  CJK COMPATIBILITY IDEOGRAPH-F9B4
        '63925' => '4506'             , # G+4506  U+F9B5  uniF9B5,  CJK COMPATIBILITY IDEOGRAPH-F9B5
        '63926' => '4508'             , # G+4508  U+F9B6  uniF9B6,  CJK COMPATIBILITY IDEOGRAPH-F9B6
        '63927' => '4509'             , # G+4509  U+F9B7  uniF9B7,  CJK COMPATIBILITY IDEOGRAPH-F9B7
        '63928' => '4510'             , # G+4510  U+F9B8  uniF9B8,  CJK COMPATIBILITY IDEOGRAPH-F9B8
        '63929' => '5797'             , # G+5797  U+F9B9  uniF9B9,  CJK COMPATIBILITY IDEOGRAPH-F9B9
        '63930' => '4551'             , # G+4551  U+F9BA  uniF9BA,  CJK COMPATIBILITY IDEOGRAPH-F9BA
        '63931' => '4552'             , # G+4552  U+F9BB  uniF9BB,  CJK COMPATIBILITY IDEOGRAPH-F9BB
        '63932' => '4553'             , # G+4553  U+F9BC  uniF9BC,  CJK COMPATIBILITY IDEOGRAPH-F9BC
        '63933' => '4180'             , # G+4180  U+F9BD  uniF9BD,  CJK COMPATIBILITY IDEOGRAPH-F9BD
        '63934' => '4555'             , # G+4555  U+F9BE  uniF9BE,  CJK COMPATIBILITY IDEOGRAPH-F9BE
        '63935' => '5800'             , # G+5800  U+F9BF  uniF9BF,  CJK COMPATIBILITY IDEOGRAPH-F9BF
        '63936' => '4556'             , # G+4556  U+F9C0  uniF9C0,  CJK COMPATIBILITY IDEOGRAPH-F9C0
        '63937' => '4557'             , # G+4557  U+F9C1  uniF9C1,  CJK COMPATIBILITY IDEOGRAPH-F9C1
        '63938' => '4560'             , # G+4560  U+F9C2  uniF9C2,  CJK COMPATIBILITY IDEOGRAPH-F9C2
        '63939' => '4561'             , # G+4561  U+F9C3  uniF9C3,  CJK COMPATIBILITY IDEOGRAPH-F9C3
        '63940' => '4563'             , # G+4563  U+F9C4  uniF9C4,  CJK COMPATIBILITY IDEOGRAPH-F9C4
        '63941' => '7988'             , # G+7988  U+F9C5  uniF9C5,  CJK COMPATIBILITY IDEOGRAPH-F9C5
        '63942' => '6123'             , # G+6123  U+F9C6  uniF9C6,  CJK COMPATIBILITY IDEOGRAPH-F9C6
        '63943' => '4577'             , # G+4577  U+F9C7  uniF9C7,  CJK COMPATIBILITY IDEOGRAPH-F9C7
        '63944' => '4183'             , # G+4183  U+F9C8  uniF9C8,  CJK COMPATIBILITY IDEOGRAPH-F9C8
        '63945' => '4579'             , # G+4579  U+F9C9  uniF9C9,  CJK COMPATIBILITY IDEOGRAPH-F9C9
        '63946' => '4581'             , # G+4581  U+F9CA  uniF9CA,  CJK COMPATIBILITY IDEOGRAPH-F9CA
        '63947' => '4582'             , # G+4582  U+F9CB  uniF9CB,  CJK COMPATIBILITY IDEOGRAPH-F9CB
        '63948' => '4584'             , # G+4584  U+F9CC  uniF9CC,  CJK COMPATIBILITY IDEOGRAPH-F9CC
        '63949' => '4586'             , # G+4586  U+F9CD  uniF9CD,  CJK COMPATIBILITY IDEOGRAPH-F9CD
        '63950' => '4588'             , # G+4588  U+F9CE  uniF9CE,  CJK COMPATIBILITY IDEOGRAPH-F9CE
        '63951' => '4184'             , # G+4184  U+F9CF  uniF9CF,  CJK COMPATIBILITY IDEOGRAPH-F9CF
        '63952' => '4590'             , # G+4590  U+F9D0  uniF9D0,  CJK COMPATIBILITY IDEOGRAPH-F9D0
        '63953' => '4591'             , # G+4591  U+F9D1  uniF9D1,  CJK COMPATIBILITY IDEOGRAPH-F9D1
        '63954' => '4592'             , # G+4592  U+F9D2  uniF9D2,  CJK COMPATIBILITY IDEOGRAPH-F9D2
        '63955' => '4593'             , # G+4593  U+F9D3  uniF9D3,  CJK COMPATIBILITY IDEOGRAPH-F9D3
        '63956' => '4595'             , # G+4595  U+F9D4  uniF9D4,  CJK COMPATIBILITY IDEOGRAPH-F9D4
        '63957' => '4596'             , # G+4596  U+F9D5  uniF9D5,  CJK COMPATIBILITY IDEOGRAPH-F9D5
        '63958' => '4597'             , # G+4597  U+F9D6  uniF9D6,  CJK COMPATIBILITY IDEOGRAPH-F9D6
        '63959' => '4599'             , # G+4599  U+F9D7  uniF9D7,  CJK COMPATIBILITY IDEOGRAPH-F9D7
        '63960' => '4600'             , # G+4600  U+F9D8  uniF9D8,  CJK COMPATIBILITY IDEOGRAPH-F9D8
        '63961' => '4601'             , # G+4601  U+F9D9  uniF9D9,  CJK COMPATIBILITY IDEOGRAPH-F9D9
        '63962' => '4602'             , # G+4602  U+F9DA  uniF9DA,  CJK COMPATIBILITY IDEOGRAPH-F9DA
        '63963' => '5552'             , # G+5552  U+F9DB  uniF9DB,  CJK COMPATIBILITY IDEOGRAPH-F9DB
        '63964' => '4603'             , # G+4603  U+F9DC  uniF9DC,  CJK COMPATIBILITY IDEOGRAPH-F9DC
        '63965' => '4614'             , # G+4614  U+F9DD  uniF9DD,  CJK COMPATIBILITY IDEOGRAPH-F9DD
        '63966' => '4616'             , # G+4616  U+F9DE  uniF9DE,  CJK COMPATIBILITY IDEOGRAPH-F9DE
        '63967' => '4618'             , # G+4618  U+F9DF  uniF9DF,  CJK COMPATIBILITY IDEOGRAPH-F9DF
        '63968' => '5950'             , # G+5950  U+F9E0  uniF9E0,  CJK COMPATIBILITY IDEOGRAPH-F9E0
        '63969' => '4620'             , # G+4620  U+F9E1  uniF9E1,  CJK COMPATIBILITY IDEOGRAPH-F9E1
        '63970' => '4621'             , # G+4621  U+F9E2  uniF9E2,  CJK COMPATIBILITY IDEOGRAPH-F9E2
        '63971' => '4187'             , # G+4187  U+F9E3  uniF9E3,  CJK COMPATIBILITY IDEOGRAPH-F9E3
        '63972' => '4625'             , # G+4625  U+F9E4  uniF9E4,  CJK COMPATIBILITY IDEOGRAPH-F9E4
        '63973' => '4627'             , # G+4627  U+F9E5  uniF9E5,  CJK COMPATIBILITY IDEOGRAPH-F9E5
        '63974' => '4629'             , # G+4629  U+F9E6  uniF9E6,  CJK COMPATIBILITY IDEOGRAPH-F9E6
        '63975' => '4632'             , # G+4632  U+F9E7  uniF9E7,  CJK COMPATIBILITY IDEOGRAPH-F9E7
        '63976' => '4633'             , # G+4633  U+F9E8  uniF9E8,  CJK COMPATIBILITY IDEOGRAPH-F9E8
        '63977' => '4634'             , # G+4634  U+F9E9  uniF9E9,  CJK COMPATIBILITY IDEOGRAPH-F9E9
        '63978' => '4636'             , # G+4636  U+F9EA  uniF9EA,  CJK COMPATIBILITY IDEOGRAPH-F9EA
        '63979' => '4188'             , # G+4188  U+F9EB  uniF9EB,  CJK COMPATIBILITY IDEOGRAPH-F9EB
        '63980' => '4189'             , # G+4189  U+F9EC  uniF9EC,  CJK COMPATIBILITY IDEOGRAPH-F9EC
        '63981' => '4638'             , # G+4638  U+F9ED  uniF9ED,  CJK COMPATIBILITY IDEOGRAPH-F9ED
        '63982' => '4640'             , # G+4640  U+F9EE  uniF9EE,  CJK COMPATIBILITY IDEOGRAPH-F9EE
        '63983' => '4641'             , # G+4641  U+F9EF  uniF9EF,  CJK COMPATIBILITY IDEOGRAPH-F9EF
        '63984' => '4642'             , # G+4642  U+F9F0  uniF9F0,  CJK COMPATIBILITY IDEOGRAPH-F9F0
        '63985' => '4644'             , # G+4644  U+F9F1  uniF9F1,  CJK COMPATIBILITY IDEOGRAPH-F9F1
        '63986' => '4645'             , # G+4645  U+F9F2  uniF9F2,  CJK COMPATIBILITY IDEOGRAPH-F9F2
        '63987' => '4646'             , # G+4646  U+F9F3  uniF9F3,  CJK COMPATIBILITY IDEOGRAPH-F9F3
        '63988' => '4647'             , # G+4647  U+F9F4  uniF9F4,  CJK COMPATIBILITY IDEOGRAPH-F9F4
        '63989' => '4648'             , # G+4648  U+F9F5  uniF9F5,  CJK COMPATIBILITY IDEOGRAPH-F9F5
        '63990' => '4650'             , # G+4650  U+F9F6  uniF9F6,  CJK COMPATIBILITY IDEOGRAPH-F9F6
        '63991' => '4653'             , # G+4653  U+F9F7  uniF9F7,  CJK COMPATIBILITY IDEOGRAPH-F9F7
        '63992' => '4654'             , # G+4654  U+F9F8  uniF9F8,  CJK COMPATIBILITY IDEOGRAPH-F9F8
        '63993' => '4655'             , # G+4655  U+F9F9  uniF9F9,  CJK COMPATIBILITY IDEOGRAPH-F9F9
        '63994' => '5351'             , # G+5351  U+F9FA  uniF9FA,  CJK COMPATIBILITY IDEOGRAPH-F9FA
        '63995' => '6494'             , # G+6494  U+F9FB  uniF9FB,  CJK COMPATIBILITY IDEOGRAPH-F9FB
        '63996' => '5731'             , # G+5731  U+F9FC  uniF9FC,  CJK COMPATIBILITY IDEOGRAPH-F9FC
        '63997' => '5771'             , # G+5771  U+F9FD  uniF9FD,  CJK COMPATIBILITY IDEOGRAPH-F9FD
        '63998' => '4191'             , # G+4191  U+F9FE  uniF9FE,  CJK COMPATIBILITY IDEOGRAPH-F9FE
        '63999' => '6484'             , # G+6484  U+F9FF  uniF9FF,  CJK COMPATIBILITY IDEOGRAPH-F9FF
        '64000' => '6684'             , # G+6684  U+FA00  uniFA00,  CJK COMPATIBILITY IDEOGRAPH-FA00
        '64001' => '4279'             , # G+4279  U+FA01  uniFA01,  CJK COMPATIBILITY IDEOGRAPH-FA01
        '64002' => '7139'             , # G+7139  U+FA02  uniFA02,  CJK COMPATIBILITY IDEOGRAPH-FA02
        '64003' => '4247'             , # G+4247  U+FA03  uniFA03,  CJK COMPATIBILITY IDEOGRAPH-FA03
        '64004' => '4266'             , # G+4266  U+FA04  uniFA04,  CJK COMPATIBILITY IDEOGRAPH-FA04
        '64005' => '4340'             , # G+4340  U+FA05  uniFA05,  CJK COMPATIBILITY IDEOGRAPH-FA05
        '64006' => '7572'             , # G+7572  U+FA06  uniFA06,  CJK COMPATIBILITY IDEOGRAPH-FA06
        '64007' => '5088'             , # G+5088  U+FA07  uniFA07,  CJK COMPATIBILITY IDEOGRAPH-FA07
        '64008' => '7709'             , # G+7709  U+FA08  uniFA08,  CJK COMPATIBILITY IDEOGRAPH-FA08
        '64009' => '3558'             , # G+3558  U+FA09  uniFA09,  CJK COMPATIBILITY IDEOGRAPH-FA09
        '64010' => '3644'             , # G+3644  U+FA0A  uniFA0A,  CJK COMPATIBILITY IDEOGRAPH-FA0A
        '64011' => '3815'             , # G+3815  U+FA0B  uniFA0B,  CJK COMPATIBILITY IDEOGRAPH-FA0B
        '65281' => '264'              , # G+264  U+FF01  exclammonospace,  FULLWIDTH EXCLAMATION MARK
        '65282' => '265'              , # G+265  U+FF02  quotedblmonospace,  FULLWIDTH QUOTATION MARK
        '65283' => '266'              , # G+266  U+FF03  numbersignmonospace,  FULLWIDTH NUMBER SIGN
        '65284' => '267'              , # G+267  U+FF04  dollarmonospace,  FULLWIDTH DOLLAR SIGN
        '65285' => '268'              , # G+268  U+FF05  percentmonospace,  FULLWIDTH PERCENT SIGN
        '65286' => '269'              , # G+269  U+FF06  ampersandmonospace,  FULLWIDTH AMPERSAND
        '65287' => '270'              , # G+270  U+FF07  quotesinglemonospace,  FULLWIDTH APOSTROPHE
        '65288' => '271'              , # G+271  U+FF08  parenleftmonospace,  FULLWIDTH LEFT PARENTHESIS
        '65289' => '272'              , # G+272  U+FF09  parenrightmonospace,  FULLWIDTH RIGHT PARENTHESIS
        '65290' => '273'              , # G+273  U+FF0A  asteriskmonospace,  FULLWIDTH ASTERISK
        '65291' => '274'              , # G+274  U+FF0B  plusmonospace,  FULLWIDTH PLUS SIGN
        '65292' => '275'              , # G+275  U+FF0C  commamonospace,  FULLWIDTH COMMA
        '65293' => '276'              , # G+276  U+FF0D  hyphenmonospace,  FULLWIDTH HYPHEN-MINUS
        '65294' => '277'              , # G+277  U+FF0E  periodmonospace,  FULLWIDTH FULL STOP
        '65295' => '278'              , # G+278  U+FF0F  slashmonospace,  FULLWIDTH SOLIDUS
        '65296' => '279'              , # G+279  U+FF10  zeromonospace,  FULLWIDTH DIGIT ZERO
        '65297' => '280'              , # G+280  U+FF11  onemonospace,  FULLWIDTH DIGIT ONE
        '65298' => '281'              , # G+281  U+FF12  twomonospace,  FULLWIDTH DIGIT TWO
        '65299' => '282'              , # G+282  U+FF13  threemonospace,  FULLWIDTH DIGIT THREE
        '65300' => '283'              , # G+283  U+FF14  fourmonospace,  FULLWIDTH DIGIT FOUR
        '65301' => '284'              , # G+284  U+FF15  fivemonospace,  FULLWIDTH DIGIT FIVE
        '65302' => '285'              , # G+285  U+FF16  sixmonospace,  FULLWIDTH DIGIT SIX
        '65303' => '286'              , # G+286  U+FF17  sevenmonospace,  FULLWIDTH DIGIT SEVEN
        '65304' => '287'              , # G+287  U+FF18  eightmonospace,  FULLWIDTH DIGIT EIGHT
        '65305' => '288'              , # G+288  U+FF19  ninemonospace,  FULLWIDTH DIGIT NINE
        '65306' => '289'              , # G+289  U+FF1A  colonmonospace,  FULLWIDTH COLON
        '65307' => '290'              , # G+290  U+FF1B  semicolonmonospace,  FULLWIDTH SEMICOLON
        '65308' => '291'              , # G+291  U+FF1C  lessmonospace,  FULLWIDTH LESS-THAN SIGN
        '65309' => '292'              , # G+292  U+FF1D  equalmonospace,  FULLWIDTH EQUALS SIGN
        '65310' => '293'              , # G+293  U+FF1E  greatermonospace,  FULLWIDTH GREATER-THAN SIGN
        '65311' => '294'              , # G+294  U+FF1F  questionmonospace,  FULLWIDTH QUESTION MARK
        '65312' => '295'              , # G+295  U+FF20  atmonospace,  FULLWIDTH COMMERCIAL AT
        '65313' => '296'              , # G+296  U+FF21  Amonospace,  FULLWIDTH LATIN CAPITAL LETTER A
        '65314' => '297'              , # G+297  U+FF22  Bmonospace,  FULLWIDTH LATIN CAPITAL LETTER B
        '65315' => '298'              , # G+298  U+FF23  Cmonospace,  FULLWIDTH LATIN CAPITAL LETTER C
        '65316' => '299'              , # G+299  U+FF24  Dmonospace,  FULLWIDTH LATIN CAPITAL LETTER D
        '65317' => '300'              , # G+300  U+FF25  Emonospace,  FULLWIDTH LATIN CAPITAL LETTER E
        '65318' => '301'              , # G+301  U+FF26  Fmonospace,  FULLWIDTH LATIN CAPITAL LETTER F
        '65319' => '302'              , # G+302  U+FF27  Gmonospace,  FULLWIDTH LATIN CAPITAL LETTER G
        '65320' => '303'              , # G+303  U+FF28  Hmonospace,  FULLWIDTH LATIN CAPITAL LETTER H
        '65321' => '304'              , # G+304  U+FF29  Imonospace,  FULLWIDTH LATIN CAPITAL LETTER I
        '65322' => '305'              , # G+305  U+FF2A  Jmonospace,  FULLWIDTH LATIN CAPITAL LETTER J
        '65323' => '306'              , # G+306  U+FF2B  Kmonospace,  FULLWIDTH LATIN CAPITAL LETTER K
        '65324' => '307'              , # G+307  U+FF2C  Lmonospace,  FULLWIDTH LATIN CAPITAL LETTER L
        '65325' => '308'              , # G+308  U+FF2D  Mmonospace,  FULLWIDTH LATIN CAPITAL LETTER M
        '65326' => '309'              , # G+309  U+FF2E  Nmonospace,  FULLWIDTH LATIN CAPITAL LETTER N
        '65327' => '310'              , # G+310  U+FF2F  Omonospace,  FULLWIDTH LATIN CAPITAL LETTER O
        '65328' => '311'              , # G+311  U+FF30  Pmonospace,  FULLWIDTH LATIN CAPITAL LETTER P
        '65329' => '312'              , # G+312  U+FF31  Qmonospace,  FULLWIDTH LATIN CAPITAL LETTER Q
        '65330' => '313'              , # G+313  U+FF32  Rmonospace,  FULLWIDTH LATIN CAPITAL LETTER R
        '65331' => '314'              , # G+314  U+FF33  Smonospace,  FULLWIDTH LATIN CAPITAL LETTER S
        '65332' => '315'              , # G+315  U+FF34  Tmonospace,  FULLWIDTH LATIN CAPITAL LETTER T
        '65333' => '316'              , # G+316  U+FF35  Umonospace,  FULLWIDTH LATIN CAPITAL LETTER U
        '65334' => '317'              , # G+317  U+FF36  Vmonospace,  FULLWIDTH LATIN CAPITAL LETTER V
        '65335' => '318'              , # G+318  U+FF37  Wmonospace,  FULLWIDTH LATIN CAPITAL LETTER W
        '65336' => '319'              , # G+319  U+FF38  Xmonospace,  FULLWIDTH LATIN CAPITAL LETTER X
        '65337' => '320'              , # G+320  U+FF39  Ymonospace,  FULLWIDTH LATIN CAPITAL LETTER Y
        '65338' => '321'              , # G+321  U+FF3A  Zmonospace,  FULLWIDTH LATIN CAPITAL LETTER Z
        '65339' => '322'              , # G+322  U+FF3B  bracketleftmonospace,  FULLWIDTH LEFT SQUARE BRACKET
        '65340' => '112'              , # G+112  U+FF3C  backslashmonospace,  FULLWIDTH REVERSE SOLIDUS
        '65341' => '324'              , # G+324  U+FF3D  bracketrightmonospace,  FULLWIDTH RIGHT SQUARE BRACKET
        '65342' => '325'              , # G+325  U+FF3E  asciicircummonospace,  FULLWIDTH CIRCUMFLEX ACCENT
        '65343' => '326'              , # G+326  U+FF3F  underscoremonospace,  FULLWIDTH LOW LINE
        '65344' => '327'              , # G+327  U+FF40  gravemonospace,  FULLWIDTH GRAVE ACCENT
        '65345' => '328'              , # G+328  U+FF41  amonospace,  FULLWIDTH LATIN SMALL LETTER A
        '65346' => '329'              , # G+329  U+FF42  bmonospace,  FULLWIDTH LATIN SMALL LETTER B
        '65347' => '330'              , # G+330  U+FF43  cmonospace,  FULLWIDTH LATIN SMALL LETTER C
        '65348' => '331'              , # G+331  U+FF44  dmonospace,  FULLWIDTH LATIN SMALL LETTER D
        '65349' => '332'              , # G+332  U+FF45  emonospace,  FULLWIDTH LATIN SMALL LETTER E
        '65350' => '333'              , # G+333  U+FF46  fmonospace,  FULLWIDTH LATIN SMALL LETTER F
        '65351' => '334'              , # G+334  U+FF47  gmonospace,  FULLWIDTH LATIN SMALL LETTER G
        '65352' => '335'              , # G+335  U+FF48  hmonospace,  FULLWIDTH LATIN SMALL LETTER H
        '65353' => '336'              , # G+336  U+FF49  imonospace,  FULLWIDTH LATIN SMALL LETTER I
        '65354' => '337'              , # G+337  U+FF4A  jmonospace,  FULLWIDTH LATIN SMALL LETTER J
        '65355' => '338'              , # G+338  U+FF4B  kmonospace,  FULLWIDTH LATIN SMALL LETTER K
        '65356' => '339'              , # G+339  U+FF4C  lmonospace,  FULLWIDTH LATIN SMALL LETTER L
        '65357' => '340'              , # G+340  U+FF4D  mmonospace,  FULLWIDTH LATIN SMALL LETTER M
        '65358' => '341'              , # G+341  U+FF4E  nmonospace,  FULLWIDTH LATIN SMALL LETTER N
        '65359' => '342'              , # G+342  U+FF4F  omonospace,  FULLWIDTH LATIN SMALL LETTER O
        '65360' => '343'              , # G+343  U+FF50  pmonospace,  FULLWIDTH LATIN SMALL LETTER P
        '65361' => '344'              , # G+344  U+FF51  qmonospace,  FULLWIDTH LATIN SMALL LETTER Q
        '65362' => '345'              , # G+345  U+FF52  rmonospace,  FULLWIDTH LATIN SMALL LETTER R
        '65363' => '346'              , # G+346  U+FF53  smonospace,  FULLWIDTH LATIN SMALL LETTER S
        '65364' => '347'              , # G+347  U+FF54  tmonospace,  FULLWIDTH LATIN SMALL LETTER T
        '65365' => '348'              , # G+348  U+FF55  umonospace,  FULLWIDTH LATIN SMALL LETTER U
        '65366' => '349'              , # G+349  U+FF56  vmonospace,  FULLWIDTH LATIN SMALL LETTER V
        '65367' => '350'              , # G+350  U+FF57  wmonospace,  FULLWIDTH LATIN SMALL LETTER W
        '65368' => '351'              , # G+351  U+FF58  xmonospace,  FULLWIDTH LATIN SMALL LETTER X
        '65369' => '352'              , # G+352  U+FF59  ymonospace,  FULLWIDTH LATIN SMALL LETTER Y
        '65370' => '353'              , # G+353  U+FF5A  zmonospace,  FULLWIDTH LATIN SMALL LETTER Z
        '65371' => '354'              , # G+354  U+FF5B  braceleftmonospace,  FULLWIDTH LEFT CURLY BRACKET
        '65372' => '355'              , # G+355  U+FF5C  barmonospace,  FULLWIDTH VERTICAL LINE
        '65373' => '356'              , # G+356  U+FF5D  bracerightmonospace,  FULLWIDTH RIGHT CURLY BRACKET
        '65374' => '200'              , # G+200  U+FF5E  asciitildemonospace,  FULLWIDTH TILDE
        '65504' => '143'              , # G+143  U+FFE0  centmonospace,  FULLWIDTH CENT SIGN
        '65505' => '144'              , # G+144  U+FFE1  sterlingmonospace,  FULLWIDTH POUND SIGN
        '65506' => '194'              , # G+194  U+FFE2  uniFFE2,  FULLWIDTH NOT SIGN
        '65507' => '357'              , # G+357  U+FFE3  macronmonospace,  FULLWIDTH MACRON
        '65509' => '145'              , # G+145  U+FFE5  yenmonospace,  FULLWIDTH YEN SIGN
        '65510' => '323'              , # G+323  U+FFE6  wonmonospace,  FULLWIDTH WON SIGN
        '65533' => '358'              , # G+358  U+FFFD  uniFFFD,  REPLACEMENT CHARACTER
    }
};

__END__

=head1 HISTORY

    $Log: korean.pm,v $
    Revision 1.7  2004/11/22 01:05:08  fredo
    updated to korea1-2 spec

    Revision 1.6  2004/06/15 09:14:45  fredo
    removed cr+lf

    Revision 1.5  2004/06/07 19:44:40  fredo
    cleaned out cr+lf for lf

    Revision 1.4  2004/02/23 23:34:16  fredo
    fixed unicode mapping based on
    adobe's utf32/utf16 specification

    Revision 1.3  2003/12/08 13:05:45  Administrator
    corrected to proper licencing statement

    Revision 1.2  2003/11/30 17:31:17  Administrator
    merged into default

    Revision 1.1.1.1.2.2  2003/11/30 16:56:47  Administrator
    merged into default

    Revision 1.1.1.1.2.1  2003/11/30 14:21:40  Administrator
    added CVS id/log


=cut